// This file was auto-generated by unicode_tablegen.
#pragma once

#include <libunicode/codepoint_properties.h>

#include <array>
#include <cstdint>

namespace unicode::precompiled
{

extern std::array<uint8_t, 4352> const stage1;
extern std::array<uint16_t, 54016> const stage2;
extern std::array<codepoint_properties, 1524> const properties;
extern std::array<uint8_t, 4352> const names_stage1;
extern std::array<uint16_t, 1114112> const names_stage2;
extern std::array<std::string_view, 45773> const names_stage3;

} // end namespace unicode::precompiled
