/**
* This file is part of the "libunicode" project
*   Copyright (c) 2020-2021 Christian Parpart <christian@parpart.family>
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include <libunicode/ucd.h>
#include <libunicode/ucd_private.h>

#include <algorithm>
#include <array>
#include <span>
#include <string_view>

namespace unicode
{

namespace tables
{
    // clang-format off
    auto static const Plane = std::array<Prop<::unicode::Plane>, 17>{ // {{{
        Prop<::unicode::Plane> { { 0x0000, 0xFFFF }, unicode::Plane::Basic_Multilingual_Plane }, // Plane 0 BMP
        Prop<::unicode::Plane> { { 0x10000, 0x1FFFF }, unicode::Plane::Supplementary_Multilingual_Plane }, // Plane 1 SMP
        Prop<::unicode::Plane> { { 0x20000, 0x2FFFF }, unicode::Plane::Supplementary_Ideographic_Plane }, // Plane 2 SIP
        Prop<::unicode::Plane> { { 0x30000, 0x3FFFF }, unicode::Plane::Tertiary_Ideographic_Plane }, // Plane 3 TIP
        Prop<::unicode::Plane> { { 0x40000, 0x4FFFF }, unicode::Plane::Unassigned }, // Plane 4
        Prop<::unicode::Plane> { { 0x50000, 0x5FFFF }, unicode::Plane::Unassigned }, // Plane 5
        Prop<::unicode::Plane> { { 0x60000, 0x6FFFF }, unicode::Plane::Unassigned }, // Plane 6
        Prop<::unicode::Plane> { { 0x70000, 0x7FFFF }, unicode::Plane::Unassigned }, // Plane 7
        Prop<::unicode::Plane> { { 0x80000, 0x8FFFF }, unicode::Plane::Unassigned }, // Plane 8
        Prop<::unicode::Plane> { { 0x90000, 0x9FFFF }, unicode::Plane::Unassigned }, // Plane 9
        Prop<::unicode::Plane> { { 0xA0000, 0xAFFFF }, unicode::Plane::Unassigned }, // Plane 10
        Prop<::unicode::Plane> { { 0xB0000, 0xBFFFF }, unicode::Plane::Unassigned }, // Plane 11
        Prop<::unicode::Plane> { { 0xC0000, 0xCFFFF }, unicode::Plane::Unassigned }, // Plane 12
        Prop<::unicode::Plane> { { 0xD0000, 0xDFFFF }, unicode::Plane::Unassigned }, // Plane 13
        Prop<::unicode::Plane> { { 0xE0000, 0xEFFFF }, unicode::Plane::Supplementary_Special_purpose_Plane }, // Plane 14 SSP
        Prop<::unicode::Plane> { { 0xF0000, 0xFFFFF }, unicode::Plane::Supplementary_Private_Use_Area_Plane }, // Plane 15 SPUA-A
        Prop<::unicode::Plane> { { 0x100000, 0x10FFFF }, unicode::Plane::Supplementary_Private_Use_Area_Plane }, // Plane 16 SPUA-B
    }; // }}}
    // clang-format off
} // namespace tables

Plane plane(char32_t codepoint) noexcept {
    return search(tables::Plane, codepoint).value_or(Plane::Unassigned);
}

namespace tables {
auto static const Alphabetic = std::array<Interval, 1202>{ // {{{
    Interval{ 0x0041, 0x005A }, // L&  [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Interval{ 0x0061, 0x007A }, // L&  [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Interval{ 0x00AA, 0x00AA }, // Lo       FEMININE ORDINAL INDICATOR
    Interval{ 0x00B5, 0x00B5 }, // L&       MICRO SIGN
    Interval{ 0x00BA, 0x00BA }, // Lo       MASCULINE ORDINAL INDICATOR
    Interval{ 0x00C0, 0x00D6 }, // L&  [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x00D8, 0x00F6 }, // L&  [31] LATIN CAPITAL LETTER O WITH STROKE..LATIN SMALL LETTER O WITH DIAERESIS
    Interval{ 0x00F8, 0x01BA }, // L& [195] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER EZH WITH TAIL
    Interval{ 0x01BB, 0x01BB }, // Lo       LATIN LETTER TWO WITH STROKE
    Interval{ 0x01BC, 0x01BF }, // L&   [4] LATIN CAPITAL LETTER TONE FIVE..LATIN LETTER WYNN
    Interval{ 0x01C0, 0x01C3 }, // Lo   [4] LATIN LETTER DENTAL CLICK..LATIN LETTER RETROFLEX CLICK
    Interval{ 0x01C4, 0x0293 }, // L& [208] LATIN CAPITAL LETTER DZ WITH CARON..LATIN SMALL LETTER EZH WITH CURL
    Interval{ 0x0294, 0x0295 }, // Lo   [2] LATIN LETTER GLOTTAL STOP..LATIN LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0x0296, 0x02AF }, // L&  [26] LATIN LETTER INVERTED GLOTTAL STOP..LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    Interval{ 0x02B0, 0x02C1 }, // Lm  [18] MODIFIER LETTER SMALL H..MODIFIER LETTER REVERSED GLOTTAL STOP
    Interval{ 0x02C6, 0x02D1 }, // Lm  [12] MODIFIER LETTER CIRCUMFLEX ACCENT..MODIFIER LETTER HALF TRIANGULAR COLON
    Interval{ 0x02E0, 0x02E4 }, // Lm   [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    Interval{ 0x02EC, 0x02EC }, // Lm       MODIFIER LETTER VOICING
    Interval{ 0x02EE, 0x02EE }, // Lm       MODIFIER LETTER DOUBLE APOSTROPHE
    Interval{ 0x0345, 0x0345 }, // Mn       COMBINING GREEK YPOGEGRAMMENI
    Interval{ 0x0363, 0x036F }, // Mn  [13] COMBINING LATIN SMALL LETTER A..COMBINING LATIN SMALL LETTER X
    Interval{ 0x0370, 0x0373 }, // L&   [4] GREEK CAPITAL LETTER HETA..GREEK SMALL LETTER ARCHAIC SAMPI
    Interval{ 0x0374, 0x0374 }, // Lm       GREEK NUMERAL SIGN
    Interval{ 0x0376, 0x0377 }, // L&   [2] GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA..GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037A, 0x037A }, // Lm       GREEK YPOGEGRAMMENI
    Interval{ 0x037B, 0x037D }, // L&   [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Interval{ 0x037F, 0x037F }, // L&       GREEK CAPITAL LETTER YOT
    Interval{ 0x0386, 0x0386 }, // L&       GREEK CAPITAL LETTER ALPHA WITH TONOS
    Interval{ 0x0388, 0x038A }, // L&   [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Interval{ 0x038C, 0x038C }, // L&       GREEK CAPITAL LETTER OMICRON WITH TONOS
    Interval{ 0x038E, 0x03A1 }, // L&  [20] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER RHO
    Interval{ 0x03A3, 0x03F5 }, // L&  [83] GREEK CAPITAL LETTER SIGMA..GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F7, 0x0481 }, // L& [139] GREEK CAPITAL LETTER SHO..CYRILLIC SMALL LETTER KOPPA
    Interval{ 0x048A, 0x052F }, // L& [166] CYRILLIC CAPITAL LETTER SHORT I WITH TAIL..CYRILLIC SMALL LETTER EL WITH DESCENDER
    Interval{ 0x0531, 0x0556 }, // L&  [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Interval{ 0x0559, 0x0559 }, // Lm       ARMENIAN MODIFIER LETTER LEFT HALF RING
    Interval{ 0x0560, 0x0588 }, // L&  [41] ARMENIAN SMALL LETTER TURNED AYB..ARMENIAN SMALL LETTER YI WITH STROKE
    Interval{ 0x05B0, 0x05BD }, // Mn  [14] HEBREW POINT SHEVA..HEBREW POINT METEG
    Interval{ 0x05BF, 0x05BF }, // Mn       HEBREW POINT RAFE
    Interval{ 0x05C1, 0x05C2 }, // Mn   [2] HEBREW POINT SHIN DOT..HEBREW POINT SIN DOT
    Interval{ 0x05C4, 0x05C5 }, // Mn   [2] HEBREW MARK UPPER DOT..HEBREW MARK LOWER DOT
    Interval{ 0x05C7, 0x05C7 }, // Mn       HEBREW POINT QAMATS QATAN
    Interval{ 0x05D0, 0x05EA }, // Lo  [27] HEBREW LETTER ALEF..HEBREW LETTER TAV
    Interval{ 0x05EF, 0x05F2 }, // Lo   [4] HEBREW YOD TRIANGLE..HEBREW LIGATURE YIDDISH DOUBLE YOD
    Interval{ 0x0610, 0x061A }, // Mn  [11] ARABIC SIGN SALLALLAHOU ALAYHE WASSALLAM..ARABIC SMALL KASRA
    Interval{ 0x0620, 0x063F }, // Lo  [32] ARABIC LETTER KASHMIRI YEH..ARABIC LETTER FARSI YEH WITH THREE DOTS ABOVE
    Interval{ 0x0640, 0x0640 }, // Lm       ARABIC TATWEEL
    Interval{ 0x0641, 0x064A }, // Lo  [10] ARABIC LETTER FEH..ARABIC LETTER YEH
    Interval{ 0x064B, 0x0657 }, // Mn  [13] ARABIC FATHATAN..ARABIC INVERTED DAMMA
    Interval{ 0x0659, 0x065F }, // Mn   [7] ARABIC ZWARAKAY..ARABIC WAVY HAMZA BELOW
    Interval{ 0x066E, 0x066F }, // Lo   [2] ARABIC LETTER DOTLESS BEH..ARABIC LETTER DOTLESS QAF
    Interval{ 0x0670, 0x0670 }, // Mn       ARABIC LETTER SUPERSCRIPT ALEF
    Interval{ 0x0671, 0x06D3 }, // Lo  [99] ARABIC LETTER ALEF WASLA..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE
    Interval{ 0x06D5, 0x06D5 }, // Lo       ARABIC LETTER AE
    Interval{ 0x06D6, 0x06DC }, // Mn   [7] ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA..ARABIC SMALL HIGH SEEN
    Interval{ 0x06E1, 0x06E4 }, // Mn   [4] ARABIC SMALL HIGH DOTLESS HEAD OF KHAH..ARABIC SMALL HIGH MADDA
    Interval{ 0x06E5, 0x06E6 }, // Lm   [2] ARABIC SMALL WAW..ARABIC SMALL YEH
    Interval{ 0x06E7, 0x06E8 }, // Mn   [2] ARABIC SMALL HIGH YEH..ARABIC SMALL HIGH NOON
    Interval{ 0x06ED, 0x06ED }, // Mn       ARABIC SMALL LOW MEEM
    Interval{ 0x06EE, 0x06EF }, // Lo   [2] ARABIC LETTER DAL WITH INVERTED V..ARABIC LETTER REH WITH INVERTED V
    Interval{ 0x06FA, 0x06FC }, // Lo   [3] ARABIC LETTER SHEEN WITH DOT BELOW..ARABIC LETTER GHAIN WITH DOT BELOW
    Interval{ 0x06FF, 0x06FF }, // Lo       ARABIC LETTER HEH WITH INVERTED V
    Interval{ 0x0710, 0x0710 }, // Lo       SYRIAC LETTER ALAPH
    Interval{ 0x0711, 0x0711 }, // Mn       SYRIAC LETTER SUPERSCRIPT ALAPH
    Interval{ 0x0712, 0x072F }, // Lo  [30] SYRIAC LETTER BETH..SYRIAC LETTER PERSIAN DHALATH
    Interval{ 0x0730, 0x073F }, // Mn  [16] SYRIAC PTHAHA ABOVE..SYRIAC RWAHA
    Interval{ 0x074D, 0x07A5 }, // Lo  [89] SYRIAC LETTER SOGDIAN ZHAIN..THAANA LETTER WAAVU
    Interval{ 0x07A6, 0x07B0 }, // Mn  [11] THAANA ABAFILI..THAANA SUKUN
    Interval{ 0x07B1, 0x07B1 }, // Lo       THAANA LETTER NAA
    Interval{ 0x07CA, 0x07EA }, // Lo  [33] NKO LETTER A..NKO LETTER JONA RA
    Interval{ 0x07F4, 0x07F5 }, // Lm   [2] NKO HIGH TONE APOSTROPHE..NKO LOW TONE APOSTROPHE
    Interval{ 0x07FA, 0x07FA }, // Lm       NKO LAJANYALAN
    Interval{ 0x0800, 0x0815 }, // Lo  [22] SAMARITAN LETTER ALAF..SAMARITAN LETTER TAAF
    Interval{ 0x0816, 0x0817 }, // Mn   [2] SAMARITAN MARK IN..SAMARITAN MARK IN-ALAF
    Interval{ 0x081A, 0x081A }, // Lm       SAMARITAN MODIFIER LETTER EPENTHETIC YUT
    Interval{ 0x081B, 0x0823 }, // Mn   [9] SAMARITAN MARK EPENTHETIC YUT..SAMARITAN VOWEL SIGN A
    Interval{ 0x0824, 0x0824 }, // Lm       SAMARITAN MODIFIER LETTER SHORT A
    Interval{ 0x0825, 0x0827 }, // Mn   [3] SAMARITAN VOWEL SIGN SHORT A..SAMARITAN VOWEL SIGN U
    Interval{ 0x0828, 0x0828 }, // Lm       SAMARITAN MODIFIER LETTER I
    Interval{ 0x0829, 0x082C }, // Mn   [4] SAMARITAN VOWEL SIGN LONG I..SAMARITAN VOWEL SIGN SUKUN
    Interval{ 0x0840, 0x0858 }, // Lo  [25] MANDAIC LETTER HALQA..MANDAIC LETTER AIN
    Interval{ 0x0860, 0x086A }, // Lo  [11] SYRIAC LETTER MALAYALAM NGA..SYRIAC LETTER MALAYALAM SSA
    Interval{ 0x0870, 0x0887 }, // Lo  [24] ARABIC LETTER ALEF WITH ATTACHED FATHA..ARABIC BASELINE ROUND DOT
    Interval{ 0x0889, 0x088F }, // Lo   [7] ARABIC LETTER NOON WITH INVERTED SMALL V..ARABIC LETTER NOON WITH RING ABOVE
    Interval{ 0x0897, 0x0897 }, // Mn       ARABIC PEPET
    Interval{ 0x08A0, 0x08C8 }, // Lo  [41] ARABIC LETTER BEH WITH SMALL V BELOW..ARABIC LETTER GRAF
    Interval{ 0x08C9, 0x08C9 }, // Lm       ARABIC SMALL FARSI YEH
    Interval{ 0x08D4, 0x08DF }, // Mn  [12] ARABIC SMALL HIGH WORD AR-RUB..ARABIC SMALL HIGH WORD WAQFA
    Interval{ 0x08E3, 0x08E9 }, // Mn   [7] ARABIC TURNED DAMMA BELOW..ARABIC CURLY KASRATAN
    Interval{ 0x08F0, 0x0902 }, // Mn  [19] ARABIC OPEN FATHATAN..DEVANAGARI SIGN ANUSVARA
    Interval{ 0x0903, 0x0903 }, // Mc       DEVANAGARI SIGN VISARGA
    Interval{ 0x0904, 0x0939 }, // Lo  [54] DEVANAGARI LETTER SHORT A..DEVANAGARI LETTER HA
    Interval{ 0x093A, 0x093A }, // Mn       DEVANAGARI VOWEL SIGN OE
    Interval{ 0x093B, 0x093B }, // Mc       DEVANAGARI VOWEL SIGN OOE
    Interval{ 0x093D, 0x093D }, // Lo       DEVANAGARI SIGN AVAGRAHA
    Interval{ 0x093E, 0x0940 }, // Mc   [3] DEVANAGARI VOWEL SIGN AA..DEVANAGARI VOWEL SIGN II
    Interval{ 0x0941, 0x0948 }, // Mn   [8] DEVANAGARI VOWEL SIGN U..DEVANAGARI VOWEL SIGN AI
    Interval{ 0x0949, 0x094C }, // Mc   [4] DEVANAGARI VOWEL SIGN CANDRA O..DEVANAGARI VOWEL SIGN AU
    Interval{ 0x094E, 0x094F }, // Mc   [2] DEVANAGARI VOWEL SIGN PRISHTHAMATRA E..DEVANAGARI VOWEL SIGN AW
    Interval{ 0x0950, 0x0950 }, // Lo       DEVANAGARI OM
    Interval{ 0x0955, 0x0957 }, // Mn   [3] DEVANAGARI VOWEL SIGN CANDRA LONG E..DEVANAGARI VOWEL SIGN UUE
    Interval{ 0x0958, 0x0961 }, // Lo  [10] DEVANAGARI LETTER QA..DEVANAGARI LETTER VOCALIC LL
    Interval{ 0x0962, 0x0963 }, // Mn   [2] DEVANAGARI VOWEL SIGN VOCALIC L..DEVANAGARI VOWEL SIGN VOCALIC LL
    Interval{ 0x0971, 0x0971 }, // Lm       DEVANAGARI SIGN HIGH SPACING DOT
    Interval{ 0x0972, 0x0980 }, // Lo  [15] DEVANAGARI LETTER CANDRA A..BENGALI ANJI
    Interval{ 0x0981, 0x0981 }, // Mn       BENGALI SIGN CANDRABINDU
    Interval{ 0x0982, 0x0983 }, // Mc   [2] BENGALI SIGN ANUSVARA..BENGALI SIGN VISARGA
    Interval{ 0x0985, 0x098C }, // Lo   [8] BENGALI LETTER A..BENGALI LETTER VOCALIC L
    Interval{ 0x098F, 0x0990 }, // Lo   [2] BENGALI LETTER E..BENGALI LETTER AI
    Interval{ 0x0993, 0x09A8 }, // Lo  [22] BENGALI LETTER O..BENGALI LETTER NA
    Interval{ 0x09AA, 0x09B0 }, // Lo   [7] BENGALI LETTER PA..BENGALI LETTER RA
    Interval{ 0x09B2, 0x09B2 }, // Lo       BENGALI LETTER LA
    Interval{ 0x09B6, 0x09B9 }, // Lo   [4] BENGALI LETTER SHA..BENGALI LETTER HA
    Interval{ 0x09BD, 0x09BD }, // Lo       BENGALI SIGN AVAGRAHA
    Interval{ 0x09BE, 0x09C0 }, // Mc   [3] BENGALI VOWEL SIGN AA..BENGALI VOWEL SIGN II
    Interval{ 0x09C1, 0x09C4 }, // Mn   [4] BENGALI VOWEL SIGN U..BENGALI VOWEL SIGN VOCALIC RR
    Interval{ 0x09C7, 0x09C8 }, // Mc   [2] BENGALI VOWEL SIGN E..BENGALI VOWEL SIGN AI
    Interval{ 0x09CB, 0x09CC }, // Mc   [2] BENGALI VOWEL SIGN O..BENGALI VOWEL SIGN AU
    Interval{ 0x09CE, 0x09CE }, // Lo       BENGALI LETTER KHANDA TA
    Interval{ 0x09D7, 0x09D7 }, // Mc       BENGALI AU LENGTH MARK
    Interval{ 0x09DC, 0x09DD }, // Lo   [2] BENGALI LETTER RRA..BENGALI LETTER RHA
    Interval{ 0x09DF, 0x09E1 }, // Lo   [3] BENGALI LETTER YYA..BENGALI LETTER VOCALIC LL
    Interval{ 0x09E2, 0x09E3 }, // Mn   [2] BENGALI VOWEL SIGN VOCALIC L..BENGALI VOWEL SIGN VOCALIC LL
    Interval{ 0x09F0, 0x09F1 }, // Lo   [2] BENGALI LETTER RA WITH MIDDLE DIAGONAL..BENGALI LETTER RA WITH LOWER DIAGONAL
    Interval{ 0x09FC, 0x09FC }, // Lo       BENGALI LETTER VEDIC ANUSVARA
    Interval{ 0x0A01, 0x0A02 }, // Mn   [2] GURMUKHI SIGN ADAK BINDI..GURMUKHI SIGN BINDI
    Interval{ 0x0A03, 0x0A03 }, // Mc       GURMUKHI SIGN VISARGA
    Interval{ 0x0A05, 0x0A0A }, // Lo   [6] GURMUKHI LETTER A..GURMUKHI LETTER UU
    Interval{ 0x0A0F, 0x0A10 }, // Lo   [2] GURMUKHI LETTER EE..GURMUKHI LETTER AI
    Interval{ 0x0A13, 0x0A28 }, // Lo  [22] GURMUKHI LETTER OO..GURMUKHI LETTER NA
    Interval{ 0x0A2A, 0x0A30 }, // Lo   [7] GURMUKHI LETTER PA..GURMUKHI LETTER RA
    Interval{ 0x0A32, 0x0A33 }, // Lo   [2] GURMUKHI LETTER LA..GURMUKHI LETTER LLA
    Interval{ 0x0A35, 0x0A36 }, // Lo   [2] GURMUKHI LETTER VA..GURMUKHI LETTER SHA
    Interval{ 0x0A38, 0x0A39 }, // Lo   [2] GURMUKHI LETTER SA..GURMUKHI LETTER HA
    Interval{ 0x0A3E, 0x0A40 }, // Mc   [3] GURMUKHI VOWEL SIGN AA..GURMUKHI VOWEL SIGN II
    Interval{ 0x0A41, 0x0A42 }, // Mn   [2] GURMUKHI VOWEL SIGN U..GURMUKHI VOWEL SIGN UU
    Interval{ 0x0A47, 0x0A48 }, // Mn   [2] GURMUKHI VOWEL SIGN EE..GURMUKHI VOWEL SIGN AI
    Interval{ 0x0A4B, 0x0A4C }, // Mn   [2] GURMUKHI VOWEL SIGN OO..GURMUKHI VOWEL SIGN AU
    Interval{ 0x0A51, 0x0A51 }, // Mn       GURMUKHI SIGN UDAAT
    Interval{ 0x0A59, 0x0A5C }, // Lo   [4] GURMUKHI LETTER KHHA..GURMUKHI LETTER RRA
    Interval{ 0x0A5E, 0x0A5E }, // Lo       GURMUKHI LETTER FA
    Interval{ 0x0A70, 0x0A71 }, // Mn   [2] GURMUKHI TIPPI..GURMUKHI ADDAK
    Interval{ 0x0A72, 0x0A74 }, // Lo   [3] GURMUKHI IRI..GURMUKHI EK ONKAR
    Interval{ 0x0A75, 0x0A75 }, // Mn       GURMUKHI SIGN YAKASH
    Interval{ 0x0A81, 0x0A82 }, // Mn   [2] GUJARATI SIGN CANDRABINDU..GUJARATI SIGN ANUSVARA
    Interval{ 0x0A83, 0x0A83 }, // Mc       GUJARATI SIGN VISARGA
    Interval{ 0x0A85, 0x0A8D }, // Lo   [9] GUJARATI LETTER A..GUJARATI VOWEL CANDRA E
    Interval{ 0x0A8F, 0x0A91 }, // Lo   [3] GUJARATI LETTER E..GUJARATI VOWEL CANDRA O
    Interval{ 0x0A93, 0x0AA8 }, // Lo  [22] GUJARATI LETTER O..GUJARATI LETTER NA
    Interval{ 0x0AAA, 0x0AB0 }, // Lo   [7] GUJARATI LETTER PA..GUJARATI LETTER RA
    Interval{ 0x0AB2, 0x0AB3 }, // Lo   [2] GUJARATI LETTER LA..GUJARATI LETTER LLA
    Interval{ 0x0AB5, 0x0AB9 }, // Lo   [5] GUJARATI LETTER VA..GUJARATI LETTER HA
    Interval{ 0x0ABD, 0x0ABD }, // Lo       GUJARATI SIGN AVAGRAHA
    Interval{ 0x0ABE, 0x0AC0 }, // Mc   [3] GUJARATI VOWEL SIGN AA..GUJARATI VOWEL SIGN II
    Interval{ 0x0AC1, 0x0AC5 }, // Mn   [5] GUJARATI VOWEL SIGN U..GUJARATI VOWEL SIGN CANDRA E
    Interval{ 0x0AC7, 0x0AC8 }, // Mn   [2] GUJARATI VOWEL SIGN E..GUJARATI VOWEL SIGN AI
    Interval{ 0x0AC9, 0x0AC9 }, // Mc       GUJARATI VOWEL SIGN CANDRA O
    Interval{ 0x0ACB, 0x0ACC }, // Mc   [2] GUJARATI VOWEL SIGN O..GUJARATI VOWEL SIGN AU
    Interval{ 0x0AD0, 0x0AD0 }, // Lo       GUJARATI OM
    Interval{ 0x0AE0, 0x0AE1 }, // Lo   [2] GUJARATI LETTER VOCALIC RR..GUJARATI LETTER VOCALIC LL
    Interval{ 0x0AE2, 0x0AE3 }, // Mn   [2] GUJARATI VOWEL SIGN VOCALIC L..GUJARATI VOWEL SIGN VOCALIC LL
    Interval{ 0x0AF9, 0x0AF9 }, // Lo       GUJARATI LETTER ZHA
    Interval{ 0x0AFA, 0x0AFC }, // Mn   [3] GUJARATI SIGN SUKUN..GUJARATI SIGN MADDAH
    Interval{ 0x0B01, 0x0B01 }, // Mn       ORIYA SIGN CANDRABINDU
    Interval{ 0x0B02, 0x0B03 }, // Mc   [2] ORIYA SIGN ANUSVARA..ORIYA SIGN VISARGA
    Interval{ 0x0B05, 0x0B0C }, // Lo   [8] ORIYA LETTER A..ORIYA LETTER VOCALIC L
    Interval{ 0x0B0F, 0x0B10 }, // Lo   [2] ORIYA LETTER E..ORIYA LETTER AI
    Interval{ 0x0B13, 0x0B28 }, // Lo  [22] ORIYA LETTER O..ORIYA LETTER NA
    Interval{ 0x0B2A, 0x0B30 }, // Lo   [7] ORIYA LETTER PA..ORIYA LETTER RA
    Interval{ 0x0B32, 0x0B33 }, // Lo   [2] ORIYA LETTER LA..ORIYA LETTER LLA
    Interval{ 0x0B35, 0x0B39 }, // Lo   [5] ORIYA LETTER VA..ORIYA LETTER HA
    Interval{ 0x0B3D, 0x0B3D }, // Lo       ORIYA SIGN AVAGRAHA
    Interval{ 0x0B3E, 0x0B3E }, // Mc       ORIYA VOWEL SIGN AA
    Interval{ 0x0B3F, 0x0B3F }, // Mn       ORIYA VOWEL SIGN I
    Interval{ 0x0B40, 0x0B40 }, // Mc       ORIYA VOWEL SIGN II
    Interval{ 0x0B41, 0x0B44 }, // Mn   [4] ORIYA VOWEL SIGN U..ORIYA VOWEL SIGN VOCALIC RR
    Interval{ 0x0B47, 0x0B48 }, // Mc   [2] ORIYA VOWEL SIGN E..ORIYA VOWEL SIGN AI
    Interval{ 0x0B4B, 0x0B4C }, // Mc   [2] ORIYA VOWEL SIGN O..ORIYA VOWEL SIGN AU
    Interval{ 0x0B56, 0x0B56 }, // Mn       ORIYA AI LENGTH MARK
    Interval{ 0x0B57, 0x0B57 }, // Mc       ORIYA AU LENGTH MARK
    Interval{ 0x0B5C, 0x0B5D }, // Lo   [2] ORIYA LETTER RRA..ORIYA LETTER RHA
    Interval{ 0x0B5F, 0x0B61 }, // Lo   [3] ORIYA LETTER YYA..ORIYA LETTER VOCALIC LL
    Interval{ 0x0B62, 0x0B63 }, // Mn   [2] ORIYA VOWEL SIGN VOCALIC L..ORIYA VOWEL SIGN VOCALIC LL
    Interval{ 0x0B71, 0x0B71 }, // Lo       ORIYA LETTER WA
    Interval{ 0x0B82, 0x0B82 }, // Mn       TAMIL SIGN ANUSVARA
    Interval{ 0x0B83, 0x0B83 }, // Lo       TAMIL SIGN VISARGA
    Interval{ 0x0B85, 0x0B8A }, // Lo   [6] TAMIL LETTER A..TAMIL LETTER UU
    Interval{ 0x0B8E, 0x0B90 }, // Lo   [3] TAMIL LETTER E..TAMIL LETTER AI
    Interval{ 0x0B92, 0x0B95 }, // Lo   [4] TAMIL LETTER O..TAMIL LETTER KA
    Interval{ 0x0B99, 0x0B9A }, // Lo   [2] TAMIL LETTER NGA..TAMIL LETTER CA
    Interval{ 0x0B9C, 0x0B9C }, // Lo       TAMIL LETTER JA
    Interval{ 0x0B9E, 0x0B9F }, // Lo   [2] TAMIL LETTER NYA..TAMIL LETTER TTA
    Interval{ 0x0BA3, 0x0BA4 }, // Lo   [2] TAMIL LETTER NNA..TAMIL LETTER TA
    Interval{ 0x0BA8, 0x0BAA }, // Lo   [3] TAMIL LETTER NA..TAMIL LETTER PA
    Interval{ 0x0BAE, 0x0BB9 }, // Lo  [12] TAMIL LETTER MA..TAMIL LETTER HA
    Interval{ 0x0BBE, 0x0BBF }, // Mc   [2] TAMIL VOWEL SIGN AA..TAMIL VOWEL SIGN I
    Interval{ 0x0BC0, 0x0BC0 }, // Mn       TAMIL VOWEL SIGN II
    Interval{ 0x0BC1, 0x0BC2 }, // Mc   [2] TAMIL VOWEL SIGN U..TAMIL VOWEL SIGN UU
    Interval{ 0x0BC6, 0x0BC8 }, // Mc   [3] TAMIL VOWEL SIGN E..TAMIL VOWEL SIGN AI
    Interval{ 0x0BCA, 0x0BCC }, // Mc   [3] TAMIL VOWEL SIGN O..TAMIL VOWEL SIGN AU
    Interval{ 0x0BD0, 0x0BD0 }, // Lo       TAMIL OM
    Interval{ 0x0BD7, 0x0BD7 }, // Mc       TAMIL AU LENGTH MARK
    Interval{ 0x0C00, 0x0C00 }, // Mn       TELUGU SIGN COMBINING CANDRABINDU ABOVE
    Interval{ 0x0C01, 0x0C03 }, // Mc   [3] TELUGU SIGN CANDRABINDU..TELUGU SIGN VISARGA
    Interval{ 0x0C04, 0x0C04 }, // Mn       TELUGU SIGN COMBINING ANUSVARA ABOVE
    Interval{ 0x0C05, 0x0C0C }, // Lo   [8] TELUGU LETTER A..TELUGU LETTER VOCALIC L
    Interval{ 0x0C0E, 0x0C10 }, // Lo   [3] TELUGU LETTER E..TELUGU LETTER AI
    Interval{ 0x0C12, 0x0C28 }, // Lo  [23] TELUGU LETTER O..TELUGU LETTER NA
    Interval{ 0x0C2A, 0x0C39 }, // Lo  [16] TELUGU LETTER PA..TELUGU LETTER HA
    Interval{ 0x0C3D, 0x0C3D }, // Lo       TELUGU SIGN AVAGRAHA
    Interval{ 0x0C3E, 0x0C40 }, // Mn   [3] TELUGU VOWEL SIGN AA..TELUGU VOWEL SIGN II
    Interval{ 0x0C41, 0x0C44 }, // Mc   [4] TELUGU VOWEL SIGN U..TELUGU VOWEL SIGN VOCALIC RR
    Interval{ 0x0C46, 0x0C48 }, // Mn   [3] TELUGU VOWEL SIGN E..TELUGU VOWEL SIGN AI
    Interval{ 0x0C4A, 0x0C4C }, // Mn   [3] TELUGU VOWEL SIGN O..TELUGU VOWEL SIGN AU
    Interval{ 0x0C55, 0x0C56 }, // Mn   [2] TELUGU LENGTH MARK..TELUGU AI LENGTH MARK
    Interval{ 0x0C58, 0x0C5A }, // Lo   [3] TELUGU LETTER TSA..TELUGU LETTER RRRA
    Interval{ 0x0C5C, 0x0C5D }, // Lo   [2] TELUGU ARCHAIC SHRII..TELUGU LETTER NAKAARA POLLU
    Interval{ 0x0C60, 0x0C61 }, // Lo   [2] TELUGU LETTER VOCALIC RR..TELUGU LETTER VOCALIC LL
    Interval{ 0x0C62, 0x0C63 }, // Mn   [2] TELUGU VOWEL SIGN VOCALIC L..TELUGU VOWEL SIGN VOCALIC LL
    Interval{ 0x0C80, 0x0C80 }, // Lo       KANNADA SIGN SPACING CANDRABINDU
    Interval{ 0x0C81, 0x0C81 }, // Mn       KANNADA SIGN CANDRABINDU
    Interval{ 0x0C82, 0x0C83 }, // Mc   [2] KANNADA SIGN ANUSVARA..KANNADA SIGN VISARGA
    Interval{ 0x0C85, 0x0C8C }, // Lo   [8] KANNADA LETTER A..KANNADA LETTER VOCALIC L
    Interval{ 0x0C8E, 0x0C90 }, // Lo   [3] KANNADA LETTER E..KANNADA LETTER AI
    Interval{ 0x0C92, 0x0CA8 }, // Lo  [23] KANNADA LETTER O..KANNADA LETTER NA
    Interval{ 0x0CAA, 0x0CB3 }, // Lo  [10] KANNADA LETTER PA..KANNADA LETTER LLA
    Interval{ 0x0CB5, 0x0CB9 }, // Lo   [5] KANNADA LETTER VA..KANNADA LETTER HA
    Interval{ 0x0CBD, 0x0CBD }, // Lo       KANNADA SIGN AVAGRAHA
    Interval{ 0x0CBE, 0x0CBE }, // Mc       KANNADA VOWEL SIGN AA
    Interval{ 0x0CBF, 0x0CBF }, // Mn       KANNADA VOWEL SIGN I
    Interval{ 0x0CC0, 0x0CC4 }, // Mc   [5] KANNADA VOWEL SIGN II..KANNADA VOWEL SIGN VOCALIC RR
    Interval{ 0x0CC6, 0x0CC6 }, // Mn       KANNADA VOWEL SIGN E
    Interval{ 0x0CC7, 0x0CC8 }, // Mc   [2] KANNADA VOWEL SIGN EE..KANNADA VOWEL SIGN AI
    Interval{ 0x0CCA, 0x0CCB }, // Mc   [2] KANNADA VOWEL SIGN O..KANNADA VOWEL SIGN OO
    Interval{ 0x0CCC, 0x0CCC }, // Mn       KANNADA VOWEL SIGN AU
    Interval{ 0x0CD5, 0x0CD6 }, // Mc   [2] KANNADA LENGTH MARK..KANNADA AI LENGTH MARK
    Interval{ 0x0CDC, 0x0CDE }, // Lo   [3] KANNADA ARCHAIC SHRII..KANNADA LETTER FA
    Interval{ 0x0CE0, 0x0CE1 }, // Lo   [2] KANNADA LETTER VOCALIC RR..KANNADA LETTER VOCALIC LL
    Interval{ 0x0CE2, 0x0CE3 }, // Mn   [2] KANNADA VOWEL SIGN VOCALIC L..KANNADA VOWEL SIGN VOCALIC LL
    Interval{ 0x0CF1, 0x0CF2 }, // Lo   [2] KANNADA SIGN JIHVAMULIYA..KANNADA SIGN UPADHMANIYA
    Interval{ 0x0CF3, 0x0CF3 }, // Mc       KANNADA SIGN COMBINING ANUSVARA ABOVE RIGHT
    Interval{ 0x0D00, 0x0D01 }, // Mn   [2] MALAYALAM SIGN COMBINING ANUSVARA ABOVE..MALAYALAM SIGN CANDRABINDU
    Interval{ 0x0D02, 0x0D03 }, // Mc   [2] MALAYALAM SIGN ANUSVARA..MALAYALAM SIGN VISARGA
    Interval{ 0x0D04, 0x0D0C }, // Lo   [9] MALAYALAM LETTER VEDIC ANUSVARA..MALAYALAM LETTER VOCALIC L
    Interval{ 0x0D0E, 0x0D10 }, // Lo   [3] MALAYALAM LETTER E..MALAYALAM LETTER AI
    Interval{ 0x0D12, 0x0D3A }, // Lo  [41] MALAYALAM LETTER O..MALAYALAM LETTER TTTA
    Interval{ 0x0D3D, 0x0D3D }, // Lo       MALAYALAM SIGN AVAGRAHA
    Interval{ 0x0D3E, 0x0D40 }, // Mc   [3] MALAYALAM VOWEL SIGN AA..MALAYALAM VOWEL SIGN II
    Interval{ 0x0D41, 0x0D44 }, // Mn   [4] MALAYALAM VOWEL SIGN U..MALAYALAM VOWEL SIGN VOCALIC RR
    Interval{ 0x0D46, 0x0D48 }, // Mc   [3] MALAYALAM VOWEL SIGN E..MALAYALAM VOWEL SIGN AI
    Interval{ 0x0D4A, 0x0D4C }, // Mc   [3] MALAYALAM VOWEL SIGN O..MALAYALAM VOWEL SIGN AU
    Interval{ 0x0D4E, 0x0D4E }, // Lo       MALAYALAM LETTER DOT REPH
    Interval{ 0x0D54, 0x0D56 }, // Lo   [3] MALAYALAM LETTER CHILLU M..MALAYALAM LETTER CHILLU LLL
    Interval{ 0x0D57, 0x0D57 }, // Mc       MALAYALAM AU LENGTH MARK
    Interval{ 0x0D5F, 0x0D61 }, // Lo   [3] MALAYALAM LETTER ARCHAIC II..MALAYALAM LETTER VOCALIC LL
    Interval{ 0x0D62, 0x0D63 }, // Mn   [2] MALAYALAM VOWEL SIGN VOCALIC L..MALAYALAM VOWEL SIGN VOCALIC LL
    Interval{ 0x0D7A, 0x0D7F }, // Lo   [6] MALAYALAM LETTER CHILLU NN..MALAYALAM LETTER CHILLU K
    Interval{ 0x0D81, 0x0D81 }, // Mn       SINHALA SIGN CANDRABINDU
    Interval{ 0x0D82, 0x0D83 }, // Mc   [2] SINHALA SIGN ANUSVARAYA..SINHALA SIGN VISARGAYA
    Interval{ 0x0D85, 0x0D96 }, // Lo  [18] SINHALA LETTER AYANNA..SINHALA LETTER AUYANNA
    Interval{ 0x0D9A, 0x0DB1 }, // Lo  [24] SINHALA LETTER ALPAPRAANA KAYANNA..SINHALA LETTER DANTAJA NAYANNA
    Interval{ 0x0DB3, 0x0DBB }, // Lo   [9] SINHALA LETTER SANYAKA DAYANNA..SINHALA LETTER RAYANNA
    Interval{ 0x0DBD, 0x0DBD }, // Lo       SINHALA LETTER DANTAJA LAYANNA
    Interval{ 0x0DC0, 0x0DC6 }, // Lo   [7] SINHALA LETTER VAYANNA..SINHALA LETTER FAYANNA
    Interval{ 0x0DCF, 0x0DD1 }, // Mc   [3] SINHALA VOWEL SIGN AELA-PILLA..SINHALA VOWEL SIGN DIGA AEDA-PILLA
    Interval{ 0x0DD2, 0x0DD4 }, // Mn   [3] SINHALA VOWEL SIGN KETTI IS-PILLA..SINHALA VOWEL SIGN KETTI PAA-PILLA
    Interval{ 0x0DD6, 0x0DD6 }, // Mn       SINHALA VOWEL SIGN DIGA PAA-PILLA
    Interval{ 0x0DD8, 0x0DDF }, // Mc   [8] SINHALA VOWEL SIGN GAETTA-PILLA..SINHALA VOWEL SIGN GAYANUKITTA
    Interval{ 0x0DF2, 0x0DF3 }, // Mc   [2] SINHALA VOWEL SIGN DIGA GAETTA-PILLA..SINHALA VOWEL SIGN DIGA GAYANUKITTA
    Interval{ 0x0E01, 0x0E30 }, // Lo  [48] THAI CHARACTER KO KAI..THAI CHARACTER SARA A
    Interval{ 0x0E31, 0x0E31 }, // Mn       THAI CHARACTER MAI HAN-AKAT
    Interval{ 0x0E32, 0x0E33 }, // Lo   [2] THAI CHARACTER SARA AA..THAI CHARACTER SARA AM
    Interval{ 0x0E34, 0x0E3A }, // Mn   [7] THAI CHARACTER SARA I..THAI CHARACTER PHINTHU
    Interval{ 0x0E40, 0x0E45 }, // Lo   [6] THAI CHARACTER SARA E..THAI CHARACTER LAKKHANGYAO
    Interval{ 0x0E46, 0x0E46 }, // Lm       THAI CHARACTER MAIYAMOK
    Interval{ 0x0E4D, 0x0E4D }, // Mn       THAI CHARACTER NIKHAHIT
    Interval{ 0x0E81, 0x0E82 }, // Lo   [2] LAO LETTER KO..LAO LETTER KHO SUNG
    Interval{ 0x0E84, 0x0E84 }, // Lo       LAO LETTER KHO TAM
    Interval{ 0x0E86, 0x0E8A }, // Lo   [5] LAO LETTER PALI GHA..LAO LETTER SO TAM
    Interval{ 0x0E8C, 0x0EA3 }, // Lo  [24] LAO LETTER PALI JHA..LAO LETTER LO LING
    Interval{ 0x0EA5, 0x0EA5 }, // Lo       LAO LETTER LO LOOT
    Interval{ 0x0EA7, 0x0EB0 }, // Lo  [10] LAO LETTER WO..LAO VOWEL SIGN A
    Interval{ 0x0EB1, 0x0EB1 }, // Mn       LAO VOWEL SIGN MAI KAN
    Interval{ 0x0EB2, 0x0EB3 }, // Lo   [2] LAO VOWEL SIGN AA..LAO VOWEL SIGN AM
    Interval{ 0x0EB4, 0x0EB9 }, // Mn   [6] LAO VOWEL SIGN I..LAO VOWEL SIGN UU
    Interval{ 0x0EBB, 0x0EBC }, // Mn   [2] LAO VOWEL SIGN MAI KON..LAO SEMIVOWEL SIGN LO
    Interval{ 0x0EBD, 0x0EBD }, // Lo       LAO SEMIVOWEL SIGN NYO
    Interval{ 0x0EC0, 0x0EC4 }, // Lo   [5] LAO VOWEL SIGN E..LAO VOWEL SIGN AI
    Interval{ 0x0EC6, 0x0EC6 }, // Lm       LAO KO LA
    Interval{ 0x0ECD, 0x0ECD }, // Mn       LAO NIGGAHITA
    Interval{ 0x0EDC, 0x0EDF }, // Lo   [4] LAO HO NO..LAO LETTER KHMU NYO
    Interval{ 0x0F00, 0x0F00 }, // Lo       TIBETAN SYLLABLE OM
    Interval{ 0x0F40, 0x0F47 }, // Lo   [8] TIBETAN LETTER KA..TIBETAN LETTER JA
    Interval{ 0x0F49, 0x0F6C }, // Lo  [36] TIBETAN LETTER NYA..TIBETAN LETTER RRA
    Interval{ 0x0F71, 0x0F7E }, // Mn  [14] TIBETAN VOWEL SIGN AA..TIBETAN SIGN RJES SU NGA RO
    Interval{ 0x0F7F, 0x0F7F }, // Mc       TIBETAN SIGN RNAM BCAD
    Interval{ 0x0F80, 0x0F83 }, // Mn   [4] TIBETAN VOWEL SIGN REVERSED I..TIBETAN SIGN SNA LDAN
    Interval{ 0x0F88, 0x0F8C }, // Lo   [5] TIBETAN SIGN LCE TSA CAN..TIBETAN SIGN INVERTED MCHU CAN
    Interval{ 0x0F8D, 0x0F97 }, // Mn  [11] TIBETAN SUBJOINED SIGN LCE TSA CAN..TIBETAN SUBJOINED LETTER JA
    Interval{ 0x0F99, 0x0FBC }, // Mn  [36] TIBETAN SUBJOINED LETTER NYA..TIBETAN SUBJOINED LETTER FIXED-FORM RA
    Interval{ 0x1000, 0x102A }, // Lo  [43] MYANMAR LETTER KA..MYANMAR LETTER AU
    Interval{ 0x102B, 0x102C }, // Mc   [2] MYANMAR VOWEL SIGN TALL AA..MYANMAR VOWEL SIGN AA
    Interval{ 0x102D, 0x1030 }, // Mn   [4] MYANMAR VOWEL SIGN I..MYANMAR VOWEL SIGN UU
    Interval{ 0x1031, 0x1031 }, // Mc       MYANMAR VOWEL SIGN E
    Interval{ 0x1032, 0x1036 }, // Mn   [5] MYANMAR VOWEL SIGN AI..MYANMAR SIGN ANUSVARA
    Interval{ 0x1038, 0x1038 }, // Mc       MYANMAR SIGN VISARGA
    Interval{ 0x103B, 0x103C }, // Mc   [2] MYANMAR CONSONANT SIGN MEDIAL YA..MYANMAR CONSONANT SIGN MEDIAL RA
    Interval{ 0x103D, 0x103E }, // Mn   [2] MYANMAR CONSONANT SIGN MEDIAL WA..MYANMAR CONSONANT SIGN MEDIAL HA
    Interval{ 0x103F, 0x103F }, // Lo       MYANMAR LETTER GREAT SA
    Interval{ 0x1050, 0x1055 }, // Lo   [6] MYANMAR LETTER SHA..MYANMAR LETTER VOCALIC LL
    Interval{ 0x1056, 0x1057 }, // Mc   [2] MYANMAR VOWEL SIGN VOCALIC R..MYANMAR VOWEL SIGN VOCALIC RR
    Interval{ 0x1058, 0x1059 }, // Mn   [2] MYANMAR VOWEL SIGN VOCALIC L..MYANMAR VOWEL SIGN VOCALIC LL
    Interval{ 0x105A, 0x105D }, // Lo   [4] MYANMAR LETTER MON NGA..MYANMAR LETTER MON BBE
    Interval{ 0x105E, 0x1060 }, // Mn   [3] MYANMAR CONSONANT SIGN MON MEDIAL NA..MYANMAR CONSONANT SIGN MON MEDIAL LA
    Interval{ 0x1061, 0x1061 }, // Lo       MYANMAR LETTER SGAW KAREN SHA
    Interval{ 0x1062, 0x1064 }, // Mc   [3] MYANMAR VOWEL SIGN SGAW KAREN EU..MYANMAR TONE MARK SGAW KAREN KE PHO
    Interval{ 0x1065, 0x1066 }, // Lo   [2] MYANMAR LETTER WESTERN PWO KAREN THA..MYANMAR LETTER WESTERN PWO KAREN PWA
    Interval{ 0x1067, 0x106D }, // Mc   [7] MYANMAR VOWEL SIGN WESTERN PWO KAREN EU..MYANMAR SIGN WESTERN PWO KAREN TONE-5
    Interval{ 0x106E, 0x1070 }, // Lo   [3] MYANMAR LETTER EASTERN PWO KAREN NNA..MYANMAR LETTER EASTERN PWO KAREN GHWA
    Interval{ 0x1071, 0x1074 }, // Mn   [4] MYANMAR VOWEL SIGN GEBA KAREN I..MYANMAR VOWEL SIGN KAYAH EE
    Interval{ 0x1075, 0x1081 }, // Lo  [13] MYANMAR LETTER SHAN KA..MYANMAR LETTER SHAN HA
    Interval{ 0x1082, 0x1082 }, // Mn       MYANMAR CONSONANT SIGN SHAN MEDIAL WA
    Interval{ 0x1083, 0x1084 }, // Mc   [2] MYANMAR VOWEL SIGN SHAN AA..MYANMAR VOWEL SIGN SHAN E
    Interval{ 0x1085, 0x1086 }, // Mn   [2] MYANMAR VOWEL SIGN SHAN E ABOVE..MYANMAR VOWEL SIGN SHAN FINAL Y
    Interval{ 0x1087, 0x108C }, // Mc   [6] MYANMAR SIGN SHAN TONE-2..MYANMAR SIGN SHAN COUNCIL TONE-3
    Interval{ 0x108D, 0x108D }, // Mn       MYANMAR SIGN SHAN COUNCIL EMPHATIC TONE
    Interval{ 0x108E, 0x108E }, // Lo       MYANMAR LETTER RUMAI PALAUNG FA
    Interval{ 0x108F, 0x108F }, // Mc       MYANMAR SIGN RUMAI PALAUNG TONE-5
    Interval{ 0x109A, 0x109C }, // Mc   [3] MYANMAR SIGN KHAMTI TONE-1..MYANMAR VOWEL SIGN AITON A
    Interval{ 0x109D, 0x109D }, // Mn       MYANMAR VOWEL SIGN AITON AI
    Interval{ 0x10A0, 0x10C5 }, // L&  [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Interval{ 0x10C7, 0x10C7 }, // L&       GEORGIAN CAPITAL LETTER YN
    Interval{ 0x10CD, 0x10CD }, // L&       GEORGIAN CAPITAL LETTER AEN
    Interval{ 0x10D0, 0x10FA }, // L&  [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Interval{ 0x10FC, 0x10FC }, // Lm       MODIFIER LETTER GEORGIAN NAR
    Interval{ 0x10FD, 0x10FF }, // L&   [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Interval{ 0x1100, 0x1248 }, // Lo [329] HANGUL CHOSEONG KIYEOK..ETHIOPIC SYLLABLE QWA
    Interval{ 0x124A, 0x124D }, // Lo   [4] ETHIOPIC SYLLABLE QWI..ETHIOPIC SYLLABLE QWE
    Interval{ 0x1250, 0x1256 }, // Lo   [7] ETHIOPIC SYLLABLE QHA..ETHIOPIC SYLLABLE QHO
    Interval{ 0x1258, 0x1258 }, // Lo       ETHIOPIC SYLLABLE QHWA
    Interval{ 0x125A, 0x125D }, // Lo   [4] ETHIOPIC SYLLABLE QHWI..ETHIOPIC SYLLABLE QHWE
    Interval{ 0x1260, 0x1288 }, // Lo  [41] ETHIOPIC SYLLABLE BA..ETHIOPIC SYLLABLE XWA
    Interval{ 0x128A, 0x128D }, // Lo   [4] ETHIOPIC SYLLABLE XWI..ETHIOPIC SYLLABLE XWE
    Interval{ 0x1290, 0x12B0 }, // Lo  [33] ETHIOPIC SYLLABLE NA..ETHIOPIC SYLLABLE KWA
    Interval{ 0x12B2, 0x12B5 }, // Lo   [4] ETHIOPIC SYLLABLE KWI..ETHIOPIC SYLLABLE KWE
    Interval{ 0x12B8, 0x12BE }, // Lo   [7] ETHIOPIC SYLLABLE KXA..ETHIOPIC SYLLABLE KXO
    Interval{ 0x12C0, 0x12C0 }, // Lo       ETHIOPIC SYLLABLE KXWA
    Interval{ 0x12C2, 0x12C5 }, // Lo   [4] ETHIOPIC SYLLABLE KXWI..ETHIOPIC SYLLABLE KXWE
    Interval{ 0x12C8, 0x12D6 }, // Lo  [15] ETHIOPIC SYLLABLE WA..ETHIOPIC SYLLABLE PHARYNGEAL O
    Interval{ 0x12D8, 0x1310 }, // Lo  [57] ETHIOPIC SYLLABLE ZA..ETHIOPIC SYLLABLE GWA
    Interval{ 0x1312, 0x1315 }, // Lo   [4] ETHIOPIC SYLLABLE GWI..ETHIOPIC SYLLABLE GWE
    Interval{ 0x1318, 0x135A }, // Lo  [67] ETHIOPIC SYLLABLE GGA..ETHIOPIC SYLLABLE FYA
    Interval{ 0x1380, 0x138F }, // Lo  [16] ETHIOPIC SYLLABLE SEBATBEIT MWA..ETHIOPIC SYLLABLE PWE
    Interval{ 0x13A0, 0x13F5 }, // L&  [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Interval{ 0x13F8, 0x13FD }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1401, 0x166C }, // Lo [620] CANADIAN SYLLABICS E..CANADIAN SYLLABICS CARRIER TTSA
    Interval{ 0x166F, 0x167F }, // Lo  [17] CANADIAN SYLLABICS QAI..CANADIAN SYLLABICS BLACKFOOT W
    Interval{ 0x1681, 0x169A }, // Lo  [26] OGHAM LETTER BEITH..OGHAM LETTER PEITH
    Interval{ 0x16A0, 0x16EA }, // Lo  [75] RUNIC LETTER FEHU FEOH FE F..RUNIC LETTER X
    Interval{ 0x16EE, 0x16F0 }, // Nl   [3] RUNIC ARLAUG SYMBOL..RUNIC BELGTHOR SYMBOL
    Interval{ 0x16F1, 0x16F8 }, // Lo   [8] RUNIC LETTER K..RUNIC LETTER FRANKS CASKET AESC
    Interval{ 0x1700, 0x1711 }, // Lo  [18] TAGALOG LETTER A..TAGALOG LETTER HA
    Interval{ 0x1712, 0x1713 }, // Mn   [2] TAGALOG VOWEL SIGN I..TAGALOG VOWEL SIGN U
    Interval{ 0x171F, 0x1731 }, // Lo  [19] TAGALOG LETTER ARCHAIC RA..HANUNOO LETTER HA
    Interval{ 0x1732, 0x1733 }, // Mn   [2] HANUNOO VOWEL SIGN I..HANUNOO VOWEL SIGN U
    Interval{ 0x1740, 0x1751 }, // Lo  [18] BUHID LETTER A..BUHID LETTER HA
    Interval{ 0x1752, 0x1753 }, // Mn   [2] BUHID VOWEL SIGN I..BUHID VOWEL SIGN U
    Interval{ 0x1760, 0x176C }, // Lo  [13] TAGBANWA LETTER A..TAGBANWA LETTER YA
    Interval{ 0x176E, 0x1770 }, // Lo   [3] TAGBANWA LETTER LA..TAGBANWA LETTER SA
    Interval{ 0x1772, 0x1773 }, // Mn   [2] TAGBANWA VOWEL SIGN I..TAGBANWA VOWEL SIGN U
    Interval{ 0x1780, 0x17B3 }, // Lo  [52] KHMER LETTER KA..KHMER INDEPENDENT VOWEL QAU
    Interval{ 0x17B6, 0x17B6 }, // Mc       KHMER VOWEL SIGN AA
    Interval{ 0x17B7, 0x17BD }, // Mn   [7] KHMER VOWEL SIGN I..KHMER VOWEL SIGN UA
    Interval{ 0x17BE, 0x17C5 }, // Mc   [8] KHMER VOWEL SIGN OE..KHMER VOWEL SIGN AU
    Interval{ 0x17C6, 0x17C6 }, // Mn       KHMER SIGN NIKAHIT
    Interval{ 0x17C7, 0x17C8 }, // Mc   [2] KHMER SIGN REAHMUK..KHMER SIGN YUUKALEAPINTU
    Interval{ 0x17D7, 0x17D7 }, // Lm       KHMER SIGN LEK TOO
    Interval{ 0x17DC, 0x17DC }, // Lo       KHMER SIGN AVAKRAHASANYA
    Interval{ 0x1820, 0x1842 }, // Lo  [35] MONGOLIAN LETTER A..MONGOLIAN LETTER CHI
    Interval{ 0x1843, 0x1843 }, // Lm       MONGOLIAN LETTER TODO LONG VOWEL SIGN
    Interval{ 0x1844, 0x1878 }, // Lo  [53] MONGOLIAN LETTER TODO E..MONGOLIAN LETTER CHA WITH TWO DOTS
    Interval{ 0x1880, 0x1884 }, // Lo   [5] MONGOLIAN LETTER ALI GALI ANUSVARA ONE..MONGOLIAN LETTER ALI GALI INVERTED UBADAMA
    Interval{ 0x1885, 0x1886 }, // Mn   [2] MONGOLIAN LETTER ALI GALI BALUDA..MONGOLIAN LETTER ALI GALI THREE BALUDA
    Interval{ 0x1887, 0x18A8 }, // Lo  [34] MONGOLIAN LETTER ALI GALI A..MONGOLIAN LETTER MANCHU ALI GALI BHA
    Interval{ 0x18A9, 0x18A9 }, // Mn       MONGOLIAN LETTER ALI GALI DAGALGA
    Interval{ 0x18AA, 0x18AA }, // Lo       MONGOLIAN LETTER MANCHU ALI GALI LHA
    Interval{ 0x18B0, 0x18F5 }, // Lo  [70] CANADIAN SYLLABICS OY..CANADIAN SYLLABICS CARRIER DENTAL S
    Interval{ 0x1900, 0x191E }, // Lo  [31] LIMBU VOWEL-CARRIER LETTER..LIMBU LETTER TRA
    Interval{ 0x1920, 0x1922 }, // Mn   [3] LIMBU VOWEL SIGN A..LIMBU VOWEL SIGN U
    Interval{ 0x1923, 0x1926 }, // Mc   [4] LIMBU VOWEL SIGN EE..LIMBU VOWEL SIGN AU
    Interval{ 0x1927, 0x1928 }, // Mn   [2] LIMBU VOWEL SIGN E..LIMBU VOWEL SIGN O
    Interval{ 0x1929, 0x192B }, // Mc   [3] LIMBU SUBJOINED LETTER YA..LIMBU SUBJOINED LETTER WA
    Interval{ 0x1930, 0x1931 }, // Mc   [2] LIMBU SMALL LETTER KA..LIMBU SMALL LETTER NGA
    Interval{ 0x1932, 0x1932 }, // Mn       LIMBU SMALL LETTER ANUSVARA
    Interval{ 0x1933, 0x1938 }, // Mc   [6] LIMBU SMALL LETTER TA..LIMBU SMALL LETTER LA
    Interval{ 0x1950, 0x196D }, // Lo  [30] TAI LE LETTER KA..TAI LE LETTER AI
    Interval{ 0x1970, 0x1974 }, // Lo   [5] TAI LE LETTER TONE-2..TAI LE LETTER TONE-6
    Interval{ 0x1980, 0x19AB }, // Lo  [44] NEW TAI LUE LETTER HIGH QA..NEW TAI LUE LETTER LOW SUA
    Interval{ 0x19B0, 0x19C9 }, // Lo  [26] NEW TAI LUE VOWEL SIGN VOWEL SHORTENER..NEW TAI LUE TONE MARK-2
    Interval{ 0x1A00, 0x1A16 }, // Lo  [23] BUGINESE LETTER KA..BUGINESE LETTER HA
    Interval{ 0x1A17, 0x1A18 }, // Mn   [2] BUGINESE VOWEL SIGN I..BUGINESE VOWEL SIGN U
    Interval{ 0x1A19, 0x1A1A }, // Mc   [2] BUGINESE VOWEL SIGN E..BUGINESE VOWEL SIGN O
    Interval{ 0x1A1B, 0x1A1B }, // Mn       BUGINESE VOWEL SIGN AE
    Interval{ 0x1A20, 0x1A54 }, // Lo  [53] TAI THAM LETTER HIGH KA..TAI THAM LETTER GREAT SA
    Interval{ 0x1A55, 0x1A55 }, // Mc       TAI THAM CONSONANT SIGN MEDIAL RA
    Interval{ 0x1A56, 0x1A56 }, // Mn       TAI THAM CONSONANT SIGN MEDIAL LA
    Interval{ 0x1A57, 0x1A57 }, // Mc       TAI THAM CONSONANT SIGN LA TANG LAI
    Interval{ 0x1A58, 0x1A5E }, // Mn   [7] TAI THAM SIGN MAI KANG LAI..TAI THAM CONSONANT SIGN SA
    Interval{ 0x1A61, 0x1A61 }, // Mc       TAI THAM VOWEL SIGN A
    Interval{ 0x1A62, 0x1A62 }, // Mn       TAI THAM VOWEL SIGN MAI SAT
    Interval{ 0x1A63, 0x1A64 }, // Mc   [2] TAI THAM VOWEL SIGN AA..TAI THAM VOWEL SIGN TALL AA
    Interval{ 0x1A65, 0x1A6C }, // Mn   [8] TAI THAM VOWEL SIGN I..TAI THAM VOWEL SIGN OA BELOW
    Interval{ 0x1A6D, 0x1A72 }, // Mc   [6] TAI THAM VOWEL SIGN OY..TAI THAM VOWEL SIGN THAM AI
    Interval{ 0x1A73, 0x1A74 }, // Mn   [2] TAI THAM VOWEL SIGN OA ABOVE..TAI THAM SIGN MAI KANG
    Interval{ 0x1AA7, 0x1AA7 }, // Lm       TAI THAM SIGN MAI YAMOK
    Interval{ 0x1ABF, 0x1AC0 }, // Mn   [2] COMBINING LATIN SMALL LETTER W BELOW..COMBINING LATIN SMALL LETTER TURNED W BELOW
    Interval{ 0x1ACC, 0x1ACE }, // Mn   [3] COMBINING LATIN SMALL LETTER INSULAR G..COMBINING LATIN SMALL LETTER INSULAR T
    Interval{ 0x1B00, 0x1B03 }, // Mn   [4] BALINESE SIGN ULU RICEM..BALINESE SIGN SURANG
    Interval{ 0x1B04, 0x1B04 }, // Mc       BALINESE SIGN BISAH
    Interval{ 0x1B05, 0x1B33 }, // Lo  [47] BALINESE LETTER AKARA..BALINESE LETTER HA
    Interval{ 0x1B35, 0x1B35 }, // Mc       BALINESE VOWEL SIGN TEDUNG
    Interval{ 0x1B36, 0x1B3A }, // Mn   [5] BALINESE VOWEL SIGN ULU..BALINESE VOWEL SIGN RA REPA
    Interval{ 0x1B3B, 0x1B3B }, // Mc       BALINESE VOWEL SIGN RA REPA TEDUNG
    Interval{ 0x1B3C, 0x1B3C }, // Mn       BALINESE VOWEL SIGN LA LENGA
    Interval{ 0x1B3D, 0x1B41 }, // Mc   [5] BALINESE VOWEL SIGN LA LENGA TEDUNG..BALINESE VOWEL SIGN TALING REPA TEDUNG
    Interval{ 0x1B42, 0x1B42 }, // Mn       BALINESE VOWEL SIGN PEPET
    Interval{ 0x1B43, 0x1B43 }, // Mc       BALINESE VOWEL SIGN PEPET TEDUNG
    Interval{ 0x1B45, 0x1B4C }, // Lo   [8] BALINESE LETTER KAF SASAK..BALINESE LETTER ARCHAIC JNYA
    Interval{ 0x1B80, 0x1B81 }, // Mn   [2] SUNDANESE SIGN PANYECEK..SUNDANESE SIGN PANGLAYAR
    Interval{ 0x1B82, 0x1B82 }, // Mc       SUNDANESE SIGN PANGWISAD
    Interval{ 0x1B83, 0x1BA0 }, // Lo  [30] SUNDANESE LETTER A..SUNDANESE LETTER HA
    Interval{ 0x1BA1, 0x1BA1 }, // Mc       SUNDANESE CONSONANT SIGN PAMINGKAL
    Interval{ 0x1BA2, 0x1BA5 }, // Mn   [4] SUNDANESE CONSONANT SIGN PANYAKRA..SUNDANESE VOWEL SIGN PANYUKU
    Interval{ 0x1BA6, 0x1BA7 }, // Mc   [2] SUNDANESE VOWEL SIGN PANAELAENG..SUNDANESE VOWEL SIGN PANOLONG
    Interval{ 0x1BA8, 0x1BA9 }, // Mn   [2] SUNDANESE VOWEL SIGN PAMEPET..SUNDANESE VOWEL SIGN PANEULEUNG
    Interval{ 0x1BAC, 0x1BAD }, // Mn   [2] SUNDANESE CONSONANT SIGN PASANGAN MA..SUNDANESE CONSONANT SIGN PASANGAN WA
    Interval{ 0x1BAE, 0x1BAF }, // Lo   [2] SUNDANESE LETTER KHA..SUNDANESE LETTER SYA
    Interval{ 0x1BBA, 0x1BE5 }, // Lo  [44] SUNDANESE AVAGRAHA..BATAK LETTER U
    Interval{ 0x1BE7, 0x1BE7 }, // Mc       BATAK VOWEL SIGN E
    Interval{ 0x1BE8, 0x1BE9 }, // Mn   [2] BATAK VOWEL SIGN PAKPAK E..BATAK VOWEL SIGN EE
    Interval{ 0x1BEA, 0x1BEC }, // Mc   [3] BATAK VOWEL SIGN I..BATAK VOWEL SIGN O
    Interval{ 0x1BED, 0x1BED }, // Mn       BATAK VOWEL SIGN KARO O
    Interval{ 0x1BEE, 0x1BEE }, // Mc       BATAK VOWEL SIGN U
    Interval{ 0x1BEF, 0x1BF1 }, // Mn   [3] BATAK VOWEL SIGN U FOR SIMALUNGUN SA..BATAK CONSONANT SIGN H
    Interval{ 0x1C00, 0x1C23 }, // Lo  [36] LEPCHA LETTER KA..LEPCHA LETTER A
    Interval{ 0x1C24, 0x1C2B }, // Mc   [8] LEPCHA SUBJOINED LETTER YA..LEPCHA VOWEL SIGN UU
    Interval{ 0x1C2C, 0x1C33 }, // Mn   [8] LEPCHA VOWEL SIGN E..LEPCHA CONSONANT SIGN T
    Interval{ 0x1C34, 0x1C35 }, // Mc   [2] LEPCHA CONSONANT SIGN NYIN-DO..LEPCHA CONSONANT SIGN KANG
    Interval{ 0x1C36, 0x1C36 }, // Mn       LEPCHA SIGN RAN
    Interval{ 0x1C4D, 0x1C4F }, // Lo   [3] LEPCHA LETTER TTA..LEPCHA LETTER DDA
    Interval{ 0x1C5A, 0x1C77 }, // Lo  [30] OL CHIKI LETTER LA..OL CHIKI LETTER OH
    Interval{ 0x1C78, 0x1C7D }, // Lm   [6] OL CHIKI MU TTUDDAG..OL CHIKI AHAD
    Interval{ 0x1C80, 0x1C8A }, // L&  [11] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER TJE
    Interval{ 0x1C90, 0x1CBA }, // L&  [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Interval{ 0x1CBD, 0x1CBF }, // L&   [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Interval{ 0x1CE9, 0x1CEC }, // Lo   [4] VEDIC SIGN ANUSVARA ANTARGOMUKHA..VEDIC SIGN ANUSVARA VAMAGOMUKHA WITH TAIL
    Interval{ 0x1CEE, 0x1CF3 }, // Lo   [6] VEDIC SIGN HEXIFORM LONG ANUSVARA..VEDIC SIGN ROTATED ARDHAVISARGA
    Interval{ 0x1CF5, 0x1CF6 }, // Lo   [2] VEDIC SIGN JIHVAMULIYA..VEDIC SIGN UPADHMANIYA
    Interval{ 0x1CFA, 0x1CFA }, // Lo       VEDIC SIGN DOUBLE ANUSVARA ANTARGOMUKHA
    Interval{ 0x1D00, 0x1D2B }, // L&  [44] LATIN LETTER SMALL CAPITAL A..CYRILLIC LETTER SMALL CAPITAL EL
    Interval{ 0x1D2C, 0x1D6A }, // Lm  [63] MODIFIER LETTER CAPITAL A..GREEK SUBSCRIPT SMALL LETTER CHI
    Interval{ 0x1D6B, 0x1D77 }, // L&  [13] LATIN SMALL LETTER UE..LATIN SMALL LETTER TURNED G
    Interval{ 0x1D78, 0x1D78 }, // Lm       MODIFIER LETTER CYRILLIC EN
    Interval{ 0x1D79, 0x1D9A }, // L&  [34] LATIN SMALL LETTER INSULAR G..LATIN SMALL LETTER EZH WITH RETROFLEX HOOK
    Interval{ 0x1D9B, 0x1DBF }, // Lm  [37] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL THETA
    Interval{ 0x1DD3, 0x1DF4 }, // Mn  [34] COMBINING LATIN SMALL LETTER FLATTENED OPEN A ABOVE..COMBINING LATIN SMALL LETTER U WITH DIAERESIS
    Interval{ 0x1E00, 0x1F15 }, // L& [278] LATIN CAPITAL LETTER A WITH RING BELOW..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F18, 0x1F1D }, // L&   [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F20, 0x1F45 }, // L&  [38] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F48, 0x1F4D }, // L&   [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F50, 0x1F57 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F59, 0x1F59 }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA
    Interval{ 0x1F5B, 0x1F5B }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Interval{ 0x1F5D, 0x1F5D }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Interval{ 0x1F5F, 0x1F7D }, // L&  [31] GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI..GREEK SMALL LETTER OMEGA WITH OXIA
    Interval{ 0x1F80, 0x1FB4 }, // L&  [53] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB6, 0x1FBC }, // L&   [7] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Interval{ 0x1FBE, 0x1FBE }, // L&       GREEK PROSGEGRAMMENI
    Interval{ 0x1FC2, 0x1FC4 }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC6, 0x1FCC }, // L&   [7] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Interval{ 0x1FD0, 0x1FD3 }, // L&   [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Interval{ 0x1FD6, 0x1FDB }, // L&   [6] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK CAPITAL LETTER IOTA WITH OXIA
    Interval{ 0x1FE0, 0x1FEC }, // L&  [13] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Interval{ 0x1FF2, 0x1FF4 }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF6, 0x1FFC }, // L&   [7] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Interval{ 0x2071, 0x2071 }, // Lm       SUPERSCRIPT LATIN SMALL LETTER I
    Interval{ 0x207F, 0x207F }, // Lm       SUPERSCRIPT LATIN SMALL LETTER N
    Interval{ 0x2090, 0x209C }, // Lm  [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Interval{ 0x2102, 0x2102 }, // L&       DOUBLE-STRUCK CAPITAL C
    Interval{ 0x2107, 0x2107 }, // L&       EULER CONSTANT
    Interval{ 0x210A, 0x2113 }, // L&  [10] SCRIPT SMALL G..SCRIPT SMALL L
    Interval{ 0x2115, 0x2115 }, // L&       DOUBLE-STRUCK CAPITAL N
    Interval{ 0x2119, 0x211D }, // L&   [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    Interval{ 0x2124, 0x2124 }, // L&       DOUBLE-STRUCK CAPITAL Z
    Interval{ 0x2126, 0x2126 }, // L&       OHM SIGN
    Interval{ 0x2128, 0x2128 }, // L&       BLACK-LETTER CAPITAL Z
    Interval{ 0x212A, 0x212D }, // L&   [4] KELVIN SIGN..BLACK-LETTER CAPITAL C
    Interval{ 0x212F, 0x2134 }, // L&   [6] SCRIPT SMALL E..SCRIPT SMALL O
    Interval{ 0x2135, 0x2138 }, // Lo   [4] ALEF SYMBOL..DALET SYMBOL
    Interval{ 0x2139, 0x2139 }, // L&       INFORMATION SOURCE
    Interval{ 0x213C, 0x213F }, // L&   [4] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK CAPITAL PI
    Interval{ 0x2145, 0x2149 }, // L&   [5] DOUBLE-STRUCK ITALIC CAPITAL D..DOUBLE-STRUCK ITALIC SMALL J
    Interval{ 0x214E, 0x214E }, // L&       TURNED SMALL F
    Interval{ 0x2160, 0x2182 }, // Nl  [35] ROMAN NUMERAL ONE..ROMAN NUMERAL TEN THOUSAND
    Interval{ 0x2183, 0x2184 }, // L&   [2] ROMAN NUMERAL REVERSED ONE HUNDRED..LATIN SMALL LETTER REVERSED C
    Interval{ 0x2185, 0x2188 }, // Nl   [4] ROMAN NUMERAL SIX LATE FORM..ROMAN NUMERAL ONE HUNDRED THOUSAND
    Interval{ 0x24B6, 0x24E9 }, // So  [52] CIRCLED LATIN CAPITAL LETTER A..CIRCLED LATIN SMALL LETTER Z
    Interval{ 0x2C00, 0x2C7B }, // L& [124] GLAGOLITIC CAPITAL LETTER AZU..LATIN LETTER SMALL CAPITAL TURNED E
    Interval{ 0x2C7C, 0x2C7D }, // Lm   [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    Interval{ 0x2C7E, 0x2CE4 }, // L& [103] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC SYMBOL KAI
    Interval{ 0x2CEB, 0x2CEE }, // L&   [4] COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI..COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF2, 0x2CF3 }, // L&   [2] COPTIC CAPITAL LETTER BOHAIRIC KHEI..COPTIC SMALL LETTER BOHAIRIC KHEI
    Interval{ 0x2D00, 0x2D25 }, // L&  [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Interval{ 0x2D27, 0x2D27 }, // L&       GEORGIAN SMALL LETTER YN
    Interval{ 0x2D2D, 0x2D2D }, // L&       GEORGIAN SMALL LETTER AEN
    Interval{ 0x2D30, 0x2D67 }, // Lo  [56] TIFINAGH LETTER YA..TIFINAGH LETTER YO
    Interval{ 0x2D6F, 0x2D6F }, // Lm       TIFINAGH MODIFIER LETTER LABIALIZATION MARK
    Interval{ 0x2D80, 0x2D96 }, // Lo  [23] ETHIOPIC SYLLABLE LOA..ETHIOPIC SYLLABLE GGWE
    Interval{ 0x2DA0, 0x2DA6 }, // Lo   [7] ETHIOPIC SYLLABLE SSA..ETHIOPIC SYLLABLE SSO
    Interval{ 0x2DA8, 0x2DAE }, // Lo   [7] ETHIOPIC SYLLABLE CCA..ETHIOPIC SYLLABLE CCO
    Interval{ 0x2DB0, 0x2DB6 }, // Lo   [7] ETHIOPIC SYLLABLE ZZA..ETHIOPIC SYLLABLE ZZO
    Interval{ 0x2DB8, 0x2DBE }, // Lo   [7] ETHIOPIC SYLLABLE CCHA..ETHIOPIC SYLLABLE CCHO
    Interval{ 0x2DC0, 0x2DC6 }, // Lo   [7] ETHIOPIC SYLLABLE QYA..ETHIOPIC SYLLABLE QYO
    Interval{ 0x2DC8, 0x2DCE }, // Lo   [7] ETHIOPIC SYLLABLE KYA..ETHIOPIC SYLLABLE KYO
    Interval{ 0x2DD0, 0x2DD6 }, // Lo   [7] ETHIOPIC SYLLABLE XYA..ETHIOPIC SYLLABLE XYO
    Interval{ 0x2DD8, 0x2DDE }, // Lo   [7] ETHIOPIC SYLLABLE GYA..ETHIOPIC SYLLABLE GYO
    Interval{ 0x2DE0, 0x2DFF }, // Mn  [32] COMBINING CYRILLIC LETTER BE..COMBINING CYRILLIC LETTER IOTIFIED BIG YUS
    Interval{ 0x2E2F, 0x2E2F }, // Lm       VERTICAL TILDE
    Interval{ 0x3005, 0x3005 }, // Lm       IDEOGRAPHIC ITERATION MARK
    Interval{ 0x3006, 0x3006 }, // Lo       IDEOGRAPHIC CLOSING MARK
    Interval{ 0x3007, 0x3007 }, // Nl       IDEOGRAPHIC NUMBER ZERO
    Interval{ 0x3021, 0x3029 }, // Nl   [9] HANGZHOU NUMERAL ONE..HANGZHOU NUMERAL NINE
    Interval{ 0x3031, 0x3035 }, // Lm   [5] VERTICAL KANA REPEAT MARK..VERTICAL KANA REPEAT MARK LOWER HALF
    Interval{ 0x3038, 0x303A }, // Nl   [3] HANGZHOU NUMERAL TEN..HANGZHOU NUMERAL THIRTY
    Interval{ 0x303B, 0x303B }, // Lm       VERTICAL IDEOGRAPHIC ITERATION MARK
    Interval{ 0x303C, 0x303C }, // Lo       MASU MARK
    Interval{ 0x3041, 0x3096 }, // Lo  [86] HIRAGANA LETTER SMALL A..HIRAGANA LETTER SMALL KE
    Interval{ 0x309D, 0x309E }, // Lm   [2] HIRAGANA ITERATION MARK..HIRAGANA VOICED ITERATION MARK
    Interval{ 0x309F, 0x309F }, // Lo       HIRAGANA DIGRAPH YORI
    Interval{ 0x30A1, 0x30FA }, // Lo  [90] KATAKANA LETTER SMALL A..KATAKANA LETTER VO
    Interval{ 0x30FC, 0x30FE }, // Lm   [3] KATAKANA-HIRAGANA PROLONGED SOUND MARK..KATAKANA VOICED ITERATION MARK
    Interval{ 0x30FF, 0x30FF }, // Lo       KATAKANA DIGRAPH KOTO
    Interval{ 0x3105, 0x312F }, // Lo  [43] BOPOMOFO LETTER B..BOPOMOFO LETTER NN
    Interval{ 0x3131, 0x318E }, // Lo  [94] HANGUL LETTER KIYEOK..HANGUL LETTER ARAEAE
    Interval{ 0x31A0, 0x31BF }, // Lo  [32] BOPOMOFO LETTER BU..BOPOMOFO LETTER AH
    Interval{ 0x31F0, 0x31FF }, // Lo  [16] KATAKANA LETTER SMALL KU..KATAKANA LETTER SMALL RO
    Interval{ 0x3400, 0x4DBF }, // Lo [6592] CJK UNIFIED IDEOGRAPH-3400..CJK UNIFIED IDEOGRAPH-4DBF
    Interval{ 0x4E00, 0xA014 }, // Lo [21013] CJK UNIFIED IDEOGRAPH-4E00..YI SYLLABLE E
    Interval{ 0xA015, 0xA015 }, // Lm       YI SYLLABLE WU
    Interval{ 0xA016, 0xA48C }, // Lo [1143] YI SYLLABLE BIT..YI SYLLABLE YYR
    Interval{ 0xA4D0, 0xA4F7 }, // Lo  [40] LISU LETTER BA..LISU LETTER OE
    Interval{ 0xA4F8, 0xA4FD }, // Lm   [6] LISU LETTER TONE MYA TI..LISU LETTER TONE MYA JEU
    Interval{ 0xA500, 0xA60B }, // Lo [268] VAI SYLLABLE EE..VAI SYLLABLE NG
    Interval{ 0xA60C, 0xA60C }, // Lm       VAI SYLLABLE LENGTHENER
    Interval{ 0xA610, 0xA61F }, // Lo  [16] VAI SYLLABLE NDOLE FA..VAI SYMBOL JONG
    Interval{ 0xA62A, 0xA62B }, // Lo   [2] VAI SYLLABLE NDOLE MA..VAI SYLLABLE NDOLE DO
    Interval{ 0xA640, 0xA66D }, // L&  [46] CYRILLIC CAPITAL LETTER ZEMLYA..CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA66E, 0xA66E }, // Lo       CYRILLIC LETTER MULTIOCULAR O
    Interval{ 0xA674, 0xA67B }, // Mn   [8] COMBINING CYRILLIC LETTER UKRAINIAN IE..COMBINING CYRILLIC LETTER OMEGA
    Interval{ 0xA67F, 0xA67F }, // Lm       CYRILLIC PAYEROK
    Interval{ 0xA680, 0xA69B }, // L&  [28] CYRILLIC CAPITAL LETTER DWE..CYRILLIC SMALL LETTER CROSSED O
    Interval{ 0xA69C, 0xA69D }, // Lm   [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    Interval{ 0xA69E, 0xA69F }, // Mn   [2] COMBINING CYRILLIC LETTER EF..COMBINING CYRILLIC LETTER IOTIFIED E
    Interval{ 0xA6A0, 0xA6E5 }, // Lo  [70] BAMUM LETTER A..BAMUM LETTER KI
    Interval{ 0xA6E6, 0xA6EF }, // Nl  [10] BAMUM LETTER MO..BAMUM LETTER KOGHOM
    Interval{ 0xA717, 0xA71F }, // Lm   [9] MODIFIER LETTER DOT VERTICAL BAR..MODIFIER LETTER LOW INVERTED EXCLAMATION MARK
    Interval{ 0xA722, 0xA76F }, // L&  [78] LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF..LATIN SMALL LETTER CON
    Interval{ 0xA770, 0xA770 }, // Lm       MODIFIER LETTER US
    Interval{ 0xA771, 0xA787 }, // L&  [23] LATIN SMALL LETTER DUM..LATIN SMALL LETTER INSULAR T
    Interval{ 0xA788, 0xA788 }, // Lm       MODIFIER LETTER LOW CIRCUMFLEX ACCENT
    Interval{ 0xA78B, 0xA78E }, // L&   [4] LATIN CAPITAL LETTER SALTILLO..LATIN SMALL LETTER L WITH RETROFLEX HOOK AND BELT
    Interval{ 0xA78F, 0xA78F }, // Lo       LATIN LETTER SINOLOGICAL DOT
    Interval{ 0xA790, 0xA7DC }, // L&  [77] LATIN CAPITAL LETTER N WITH DESCENDER..LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Interval{ 0xA7F1, 0xA7F4 }, // Lm   [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    Interval{ 0xA7F5, 0xA7F6 }, // L&   [2] LATIN CAPITAL LETTER REVERSED HALF H..LATIN SMALL LETTER REVERSED HALF H
    Interval{ 0xA7F7, 0xA7F7 }, // Lo       LATIN EPIGRAPHIC LETTER SIDEWAYS I
    Interval{ 0xA7F8, 0xA7F9 }, // Lm   [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    Interval{ 0xA7FA, 0xA7FA }, // L&       LATIN LETTER SMALL CAPITAL TURNED M
    Interval{ 0xA7FB, 0xA801 }, // Lo   [7] LATIN EPIGRAPHIC LETTER REVERSED F..SYLOTI NAGRI LETTER I
    Interval{ 0xA802, 0xA802 }, // Mn       SYLOTI NAGRI SIGN DVISVARA
    Interval{ 0xA803, 0xA805 }, // Lo   [3] SYLOTI NAGRI LETTER U..SYLOTI NAGRI LETTER O
    Interval{ 0xA807, 0xA80A }, // Lo   [4] SYLOTI NAGRI LETTER KO..SYLOTI NAGRI LETTER GHO
    Interval{ 0xA80B, 0xA80B }, // Mn       SYLOTI NAGRI SIGN ANUSVARA
    Interval{ 0xA80C, 0xA822 }, // Lo  [23] SYLOTI NAGRI LETTER CO..SYLOTI NAGRI LETTER HO
    Interval{ 0xA823, 0xA824 }, // Mc   [2] SYLOTI NAGRI VOWEL SIGN A..SYLOTI NAGRI VOWEL SIGN I
    Interval{ 0xA825, 0xA826 }, // Mn   [2] SYLOTI NAGRI VOWEL SIGN U..SYLOTI NAGRI VOWEL SIGN E
    Interval{ 0xA827, 0xA827 }, // Mc       SYLOTI NAGRI VOWEL SIGN OO
    Interval{ 0xA840, 0xA873 }, // Lo  [52] PHAGS-PA LETTER KA..PHAGS-PA LETTER CANDRABINDU
    Interval{ 0xA880, 0xA881 }, // Mc   [2] SAURASHTRA SIGN ANUSVARA..SAURASHTRA SIGN VISARGA
    Interval{ 0xA882, 0xA8B3 }, // Lo  [50] SAURASHTRA LETTER A..SAURASHTRA LETTER LLA
    Interval{ 0xA8B4, 0xA8C3 }, // Mc  [16] SAURASHTRA CONSONANT SIGN HAARU..SAURASHTRA VOWEL SIGN AU
    Interval{ 0xA8C5, 0xA8C5 }, // Mn       SAURASHTRA SIGN CANDRABINDU
    Interval{ 0xA8F2, 0xA8F7 }, // Lo   [6] DEVANAGARI SIGN SPACING CANDRABINDU..DEVANAGARI SIGN CANDRABINDU AVAGRAHA
    Interval{ 0xA8FB, 0xA8FB }, // Lo       DEVANAGARI HEADSTROKE
    Interval{ 0xA8FD, 0xA8FE }, // Lo   [2] DEVANAGARI JAIN OM..DEVANAGARI LETTER AY
    Interval{ 0xA8FF, 0xA8FF }, // Mn       DEVANAGARI VOWEL SIGN AY
    Interval{ 0xA90A, 0xA925 }, // Lo  [28] KAYAH LI LETTER KA..KAYAH LI LETTER OO
    Interval{ 0xA926, 0xA92A }, // Mn   [5] KAYAH LI VOWEL UE..KAYAH LI VOWEL O
    Interval{ 0xA930, 0xA946 }, // Lo  [23] REJANG LETTER KA..REJANG LETTER A
    Interval{ 0xA947, 0xA951 }, // Mn  [11] REJANG VOWEL SIGN I..REJANG CONSONANT SIGN R
    Interval{ 0xA952, 0xA952 }, // Mc       REJANG CONSONANT SIGN H
    Interval{ 0xA960, 0xA97C }, // Lo  [29] HANGUL CHOSEONG TIKEUT-MIEUM..HANGUL CHOSEONG SSANGYEORINHIEUH
    Interval{ 0xA980, 0xA982 }, // Mn   [3] JAVANESE SIGN PANYANGGA..JAVANESE SIGN LAYAR
    Interval{ 0xA983, 0xA983 }, // Mc       JAVANESE SIGN WIGNYAN
    Interval{ 0xA984, 0xA9B2 }, // Lo  [47] JAVANESE LETTER A..JAVANESE LETTER HA
    Interval{ 0xA9B4, 0xA9B5 }, // Mc   [2] JAVANESE VOWEL SIGN TARUNG..JAVANESE VOWEL SIGN TOLONG
    Interval{ 0xA9B6, 0xA9B9 }, // Mn   [4] JAVANESE VOWEL SIGN WULU..JAVANESE VOWEL SIGN SUKU MENDUT
    Interval{ 0xA9BA, 0xA9BB }, // Mc   [2] JAVANESE VOWEL SIGN TALING..JAVANESE VOWEL SIGN DIRGA MURE
    Interval{ 0xA9BC, 0xA9BD }, // Mn   [2] JAVANESE VOWEL SIGN PEPET..JAVANESE CONSONANT SIGN KERET
    Interval{ 0xA9BE, 0xA9BF }, // Mc   [2] JAVANESE CONSONANT SIGN PENGKAL..JAVANESE CONSONANT SIGN CAKRA
    Interval{ 0xA9CF, 0xA9CF }, // Lm       JAVANESE PANGRANGKEP
    Interval{ 0xA9E0, 0xA9E4 }, // Lo   [5] MYANMAR LETTER SHAN GHA..MYANMAR LETTER SHAN BHA
    Interval{ 0xA9E5, 0xA9E5 }, // Mn       MYANMAR SIGN SHAN SAW
    Interval{ 0xA9E6, 0xA9E6 }, // Lm       MYANMAR MODIFIER LETTER SHAN REDUPLICATION
    Interval{ 0xA9E7, 0xA9EF }, // Lo   [9] MYANMAR LETTER TAI LAING NYA..MYANMAR LETTER TAI LAING NNA
    Interval{ 0xA9FA, 0xA9FE }, // Lo   [5] MYANMAR LETTER TAI LAING LLA..MYANMAR LETTER TAI LAING BHA
    Interval{ 0xAA00, 0xAA28 }, // Lo  [41] CHAM LETTER A..CHAM LETTER HA
    Interval{ 0xAA29, 0xAA2E }, // Mn   [6] CHAM VOWEL SIGN AA..CHAM VOWEL SIGN OE
    Interval{ 0xAA2F, 0xAA30 }, // Mc   [2] CHAM VOWEL SIGN O..CHAM VOWEL SIGN AI
    Interval{ 0xAA31, 0xAA32 }, // Mn   [2] CHAM VOWEL SIGN AU..CHAM VOWEL SIGN UE
    Interval{ 0xAA33, 0xAA34 }, // Mc   [2] CHAM CONSONANT SIGN YA..CHAM CONSONANT SIGN RA
    Interval{ 0xAA35, 0xAA36 }, // Mn   [2] CHAM CONSONANT SIGN LA..CHAM CONSONANT SIGN WA
    Interval{ 0xAA40, 0xAA42 }, // Lo   [3] CHAM LETTER FINAL K..CHAM LETTER FINAL NG
    Interval{ 0xAA43, 0xAA43 }, // Mn       CHAM CONSONANT SIGN FINAL NG
    Interval{ 0xAA44, 0xAA4B }, // Lo   [8] CHAM LETTER FINAL CH..CHAM LETTER FINAL SS
    Interval{ 0xAA4C, 0xAA4C }, // Mn       CHAM CONSONANT SIGN FINAL M
    Interval{ 0xAA4D, 0xAA4D }, // Mc       CHAM CONSONANT SIGN FINAL H
    Interval{ 0xAA60, 0xAA6F }, // Lo  [16] MYANMAR LETTER KHAMTI GA..MYANMAR LETTER KHAMTI FA
    Interval{ 0xAA70, 0xAA70 }, // Lm       MYANMAR MODIFIER LETTER KHAMTI REDUPLICATION
    Interval{ 0xAA71, 0xAA76 }, // Lo   [6] MYANMAR LETTER KHAMTI XA..MYANMAR LOGOGRAM KHAMTI HM
    Interval{ 0xAA7A, 0xAA7A }, // Lo       MYANMAR LETTER AITON RA
    Interval{ 0xAA7B, 0xAA7B }, // Mc       MYANMAR SIGN PAO KAREN TONE
    Interval{ 0xAA7C, 0xAA7C }, // Mn       MYANMAR SIGN TAI LAING TONE-2
    Interval{ 0xAA7D, 0xAA7D }, // Mc       MYANMAR SIGN TAI LAING TONE-5
    Interval{ 0xAA7E, 0xAAAF }, // Lo  [50] MYANMAR LETTER SHWE PALAUNG CHA..TAI VIET LETTER HIGH O
    Interval{ 0xAAB0, 0xAAB0 }, // Mn       TAI VIET MAI KANG
    Interval{ 0xAAB1, 0xAAB1 }, // Lo       TAI VIET VOWEL AA
    Interval{ 0xAAB2, 0xAAB4 }, // Mn   [3] TAI VIET VOWEL I..TAI VIET VOWEL U
    Interval{ 0xAAB5, 0xAAB6 }, // Lo   [2] TAI VIET VOWEL E..TAI VIET VOWEL O
    Interval{ 0xAAB7, 0xAAB8 }, // Mn   [2] TAI VIET MAI KHIT..TAI VIET VOWEL IA
    Interval{ 0xAAB9, 0xAABD }, // Lo   [5] TAI VIET VOWEL UEA..TAI VIET VOWEL AN
    Interval{ 0xAABE, 0xAABE }, // Mn       TAI VIET VOWEL AM
    Interval{ 0xAAC0, 0xAAC0 }, // Lo       TAI VIET TONE MAI NUENG
    Interval{ 0xAAC2, 0xAAC2 }, // Lo       TAI VIET TONE MAI SONG
    Interval{ 0xAADB, 0xAADC }, // Lo   [2] TAI VIET SYMBOL KON..TAI VIET SYMBOL NUENG
    Interval{ 0xAADD, 0xAADD }, // Lm       TAI VIET SYMBOL SAM
    Interval{ 0xAAE0, 0xAAEA }, // Lo  [11] MEETEI MAYEK LETTER E..MEETEI MAYEK LETTER SSA
    Interval{ 0xAAEB, 0xAAEB }, // Mc       MEETEI MAYEK VOWEL SIGN II
    Interval{ 0xAAEC, 0xAAED }, // Mn   [2] MEETEI MAYEK VOWEL SIGN UU..MEETEI MAYEK VOWEL SIGN AAI
    Interval{ 0xAAEE, 0xAAEF }, // Mc   [2] MEETEI MAYEK VOWEL SIGN AU..MEETEI MAYEK VOWEL SIGN AAU
    Interval{ 0xAAF2, 0xAAF2 }, // Lo       MEETEI MAYEK ANJI
    Interval{ 0xAAF3, 0xAAF4 }, // Lm   [2] MEETEI MAYEK SYLLABLE REPETITION MARK..MEETEI MAYEK WORD REPETITION MARK
    Interval{ 0xAAF5, 0xAAF5 }, // Mc       MEETEI MAYEK VOWEL SIGN VISARGA
    Interval{ 0xAB01, 0xAB06 }, // Lo   [6] ETHIOPIC SYLLABLE TTHU..ETHIOPIC SYLLABLE TTHO
    Interval{ 0xAB09, 0xAB0E }, // Lo   [6] ETHIOPIC SYLLABLE DDHU..ETHIOPIC SYLLABLE DDHO
    Interval{ 0xAB11, 0xAB16 }, // Lo   [6] ETHIOPIC SYLLABLE DZU..ETHIOPIC SYLLABLE DZO
    Interval{ 0xAB20, 0xAB26 }, // Lo   [7] ETHIOPIC SYLLABLE CCHHA..ETHIOPIC SYLLABLE CCHHO
    Interval{ 0xAB28, 0xAB2E }, // Lo   [7] ETHIOPIC SYLLABLE BBA..ETHIOPIC SYLLABLE BBO
    Interval{ 0xAB30, 0xAB5A }, // L&  [43] LATIN SMALL LETTER BARRED ALPHA..LATIN SMALL LETTER Y WITH SHORT RIGHT LEG
    Interval{ 0xAB5C, 0xAB5F }, // Lm   [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    Interval{ 0xAB60, 0xAB68 }, // L&   [9] LATIN SMALL LETTER SAKHA YAT..LATIN SMALL LETTER TURNED R WITH MIDDLE TILDE
    Interval{ 0xAB69, 0xAB69 }, // Lm       MODIFIER LETTER SMALL TURNED W
    Interval{ 0xAB70, 0xABBF }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xABC0, 0xABE2 }, // Lo  [35] MEETEI MAYEK LETTER KOK..MEETEI MAYEK LETTER I LONSUM
    Interval{ 0xABE3, 0xABE4 }, // Mc   [2] MEETEI MAYEK VOWEL SIGN ONAP..MEETEI MAYEK VOWEL SIGN INAP
    Interval{ 0xABE5, 0xABE5 }, // Mn       MEETEI MAYEK VOWEL SIGN ANAP
    Interval{ 0xABE6, 0xABE7 }, // Mc   [2] MEETEI MAYEK VOWEL SIGN YENAP..MEETEI MAYEK VOWEL SIGN SOUNAP
    Interval{ 0xABE8, 0xABE8 }, // Mn       MEETEI MAYEK VOWEL SIGN UNAP
    Interval{ 0xABE9, 0xABEA }, // Mc   [2] MEETEI MAYEK VOWEL SIGN CHEINAP..MEETEI MAYEK VOWEL SIGN NUNG
    Interval{ 0xAC00, 0xD7A3 }, // Lo [11172] HANGUL SYLLABLE GA..HANGUL SYLLABLE HIH
    Interval{ 0xD7B0, 0xD7C6 }, // Lo  [23] HANGUL JUNGSEONG O-YEO..HANGUL JUNGSEONG ARAEA-E
    Interval{ 0xD7CB, 0xD7FB }, // Lo  [49] HANGUL JONGSEONG NIEUN-RIEUL..HANGUL JONGSEONG PHIEUPH-THIEUTH
    Interval{ 0xF900, 0xFA6D }, // Lo [366] CJK COMPATIBILITY IDEOGRAPH-F900..CJK COMPATIBILITY IDEOGRAPH-FA6D
    Interval{ 0xFA70, 0xFAD9 }, // Lo [106] CJK COMPATIBILITY IDEOGRAPH-FA70..CJK COMPATIBILITY IDEOGRAPH-FAD9
    Interval{ 0xFB00, 0xFB06 }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFB1D, 0xFB1D }, // Lo       HEBREW LETTER YOD WITH HIRIQ
    Interval{ 0xFB1E, 0xFB1E }, // Mn       HEBREW POINT JUDEO-SPANISH VARIKA
    Interval{ 0xFB1F, 0xFB28 }, // Lo  [10] HEBREW LIGATURE YIDDISH YOD YOD PATAH..HEBREW LETTER WIDE TAV
    Interval{ 0xFB2A, 0xFB36 }, // Lo  [13] HEBREW LETTER SHIN WITH SHIN DOT..HEBREW LETTER ZAYIN WITH DAGESH
    Interval{ 0xFB38, 0xFB3C }, // Lo   [5] HEBREW LETTER TET WITH DAGESH..HEBREW LETTER LAMED WITH DAGESH
    Interval{ 0xFB3E, 0xFB3E }, // Lo       HEBREW LETTER MEM WITH DAGESH
    Interval{ 0xFB40, 0xFB41 }, // Lo   [2] HEBREW LETTER NUN WITH DAGESH..HEBREW LETTER SAMEKH WITH DAGESH
    Interval{ 0xFB43, 0xFB44 }, // Lo   [2] HEBREW LETTER FINAL PE WITH DAGESH..HEBREW LETTER PE WITH DAGESH
    Interval{ 0xFB46, 0xFBB1 }, // Lo [108] HEBREW LETTER TSADI WITH DAGESH..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE FINAL FORM
    Interval{ 0xFBD3, 0xFD3D }, // Lo [363] ARABIC LETTER NG ISOLATED FORM..ARABIC LIGATURE ALEF WITH FATHATAN ISOLATED FORM
    Interval{ 0xFD50, 0xFD8F }, // Lo  [64] ARABIC LIGATURE TEH WITH JEEM WITH MEEM INITIAL FORM..ARABIC LIGATURE MEEM WITH KHAH WITH MEEM INITIAL FORM
    Interval{ 0xFD92, 0xFDC7 }, // Lo  [54] ARABIC LIGATURE MEEM WITH JEEM WITH KHAH INITIAL FORM..ARABIC LIGATURE NOON WITH JEEM WITH YEH FINAL FORM
    Interval{ 0xFDF0, 0xFDFB }, // Lo  [12] ARABIC LIGATURE SALLA USED AS KORANIC STOP SIGN ISOLATED FORM..ARABIC LIGATURE JALLAJALALOUHOU
    Interval{ 0xFE70, 0xFE74 }, // Lo   [5] ARABIC FATHATAN ISOLATED FORM..ARABIC KASRATAN ISOLATED FORM
    Interval{ 0xFE76, 0xFEFC }, // Lo [135] ARABIC FATHA ISOLATED FORM..ARABIC LIGATURE LAM WITH ALEF FINAL FORM
    Interval{ 0xFF21, 0xFF3A }, // L&  [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Interval{ 0xFF41, 0xFF5A }, // L&  [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Interval{ 0xFF66, 0xFF6F }, // Lo  [10] HALFWIDTH KATAKANA LETTER WO..HALFWIDTH KATAKANA LETTER SMALL TU
    Interval{ 0xFF70, 0xFF70 }, // Lm       HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Interval{ 0xFF71, 0xFF9D }, // Lo  [45] HALFWIDTH KATAKANA LETTER A..HALFWIDTH KATAKANA LETTER N
    Interval{ 0xFF9E, 0xFF9F }, // Lm   [2] HALFWIDTH KATAKANA VOICED SOUND MARK..HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK
    Interval{ 0xFFA0, 0xFFBE }, // Lo  [31] HALFWIDTH HANGUL FILLER..HALFWIDTH HANGUL LETTER HIEUH
    Interval{ 0xFFC2, 0xFFC7 }, // Lo   [6] HALFWIDTH HANGUL LETTER A..HALFWIDTH HANGUL LETTER E
    Interval{ 0xFFCA, 0xFFCF }, // Lo   [6] HALFWIDTH HANGUL LETTER YEO..HALFWIDTH HANGUL LETTER OE
    Interval{ 0xFFD2, 0xFFD7 }, // Lo   [6] HALFWIDTH HANGUL LETTER YO..HALFWIDTH HANGUL LETTER YU
    Interval{ 0xFFDA, 0xFFDC }, // Lo   [3] HALFWIDTH HANGUL LETTER EU..HALFWIDTH HANGUL LETTER I
    Interval{ 0x10000, 0x1000B }, // Lo  [12] LINEAR B SYLLABLE B008 A..LINEAR B SYLLABLE B046 JE
    Interval{ 0x1000D, 0x10026 }, // Lo  [26] LINEAR B SYLLABLE B036 JO..LINEAR B SYLLABLE B032 QO
    Interval{ 0x10028, 0x1003A }, // Lo  [19] LINEAR B SYLLABLE B060 RA..LINEAR B SYLLABLE B042 WO
    Interval{ 0x1003C, 0x1003D }, // Lo   [2] LINEAR B SYLLABLE B017 ZA..LINEAR B SYLLABLE B074 ZE
    Interval{ 0x1003F, 0x1004D }, // Lo  [15] LINEAR B SYLLABLE B020 ZO..LINEAR B SYLLABLE B091 TWO
    Interval{ 0x10050, 0x1005D }, // Lo  [14] LINEAR B SYMBOL B018..LINEAR B SYMBOL B089
    Interval{ 0x10080, 0x100FA }, // Lo [123] LINEAR B IDEOGRAM B100 MAN..LINEAR B IDEOGRAM VESSEL B305
    Interval{ 0x10140, 0x10174 }, // Nl  [53] GREEK ACROPHONIC ATTIC ONE QUARTER..GREEK ACROPHONIC STRATIAN FIFTY MNAS
    Interval{ 0x10280, 0x1029C }, // Lo  [29] LYCIAN LETTER A..LYCIAN LETTER X
    Interval{ 0x102A0, 0x102D0 }, // Lo  [49] CARIAN LETTER A..CARIAN LETTER UUU3
    Interval{ 0x10300, 0x1031F }, // Lo  [32] OLD ITALIC LETTER A..OLD ITALIC LETTER ESS
    Interval{ 0x1032D, 0x10340 }, // Lo  [20] OLD ITALIC LETTER YE..GOTHIC LETTER PAIRTHRA
    Interval{ 0x10341, 0x10341 }, // Nl       GOTHIC LETTER NINETY
    Interval{ 0x10342, 0x10349 }, // Lo   [8] GOTHIC LETTER RAIDA..GOTHIC LETTER OTHAL
    Interval{ 0x1034A, 0x1034A }, // Nl       GOTHIC LETTER NINE HUNDRED
    Interval{ 0x10350, 0x10375 }, // Lo  [38] OLD PERMIC LETTER AN..OLD PERMIC LETTER IA
    Interval{ 0x10376, 0x1037A }, // Mn   [5] COMBINING OLD PERMIC LETTER AN..COMBINING OLD PERMIC LETTER SII
    Interval{ 0x10380, 0x1039D }, // Lo  [30] UGARITIC LETTER ALPA..UGARITIC LETTER SSU
    Interval{ 0x103A0, 0x103C3 }, // Lo  [36] OLD PERSIAN SIGN A..OLD PERSIAN SIGN HA
    Interval{ 0x103C8, 0x103CF }, // Lo   [8] OLD PERSIAN SIGN AURAMAZDAA..OLD PERSIAN SIGN BUUMISH
    Interval{ 0x103D1, 0x103D5 }, // Nl   [5] OLD PERSIAN NUMBER ONE..OLD PERSIAN NUMBER HUNDRED
    Interval{ 0x10400, 0x1044F }, // L&  [80] DESERET CAPITAL LETTER LONG I..DESERET SMALL LETTER EW
    Interval{ 0x10450, 0x1049D }, // Lo  [78] SHAVIAN LETTER PEEP..OSMANYA LETTER OO
    Interval{ 0x104B0, 0x104D3 }, // L&  [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Interval{ 0x104D8, 0x104FB }, // L&  [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Interval{ 0x10500, 0x10527 }, // Lo  [40] ELBASAN LETTER A..ELBASAN LETTER KHE
    Interval{ 0x10530, 0x10563 }, // Lo  [52] CAUCASIAN ALBANIAN LETTER ALT..CAUCASIAN ALBANIAN LETTER KIW
    Interval{ 0x10570, 0x1057A }, // L&  [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Interval{ 0x1057C, 0x1058A }, // L&  [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Interval{ 0x1058C, 0x10592 }, // L&   [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Interval{ 0x10594, 0x10595 }, // L&   [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Interval{ 0x10597, 0x105A1 }, // L&  [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Interval{ 0x105A3, 0x105B1 }, // L&  [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Interval{ 0x105B3, 0x105B9 }, // L&   [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Interval{ 0x105BB, 0x105BC }, // L&   [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Interval{ 0x105C0, 0x105F3 }, // Lo  [52] TODHRI LETTER A..TODHRI LETTER OO
    Interval{ 0x10600, 0x10736 }, // Lo [311] LINEAR A SIGN AB001..LINEAR A SIGN A664
    Interval{ 0x10740, 0x10755 }, // Lo  [22] LINEAR A SIGN A701 A..LINEAR A SIGN A732 JE
    Interval{ 0x10760, 0x10767 }, // Lo   [8] LINEAR A SIGN A800..LINEAR A SIGN A807
    Interval{ 0x10780, 0x10785 }, // Lm   [6] MODIFIER LETTER SMALL CAPITAL AA..MODIFIER LETTER SMALL B WITH HOOK
    Interval{ 0x10787, 0x107B0 }, // Lm  [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Interval{ 0x107B2, 0x107BA }, // Lm   [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Interval{ 0x10800, 0x10805 }, // Lo   [6] CYPRIOT SYLLABLE A..CYPRIOT SYLLABLE JA
    Interval{ 0x10808, 0x10808 }, // Lo       CYPRIOT SYLLABLE JO
    Interval{ 0x1080A, 0x10835 }, // Lo  [44] CYPRIOT SYLLABLE KA..CYPRIOT SYLLABLE WO
    Interval{ 0x10837, 0x10838 }, // Lo   [2] CYPRIOT SYLLABLE XA..CYPRIOT SYLLABLE XE
    Interval{ 0x1083C, 0x1083C }, // Lo       CYPRIOT SYLLABLE ZA
    Interval{ 0x1083F, 0x10855 }, // Lo  [23] CYPRIOT SYLLABLE ZO..IMPERIAL ARAMAIC LETTER TAW
    Interval{ 0x10860, 0x10876 }, // Lo  [23] PALMYRENE LETTER ALEPH..PALMYRENE LETTER TAW
    Interval{ 0x10880, 0x1089E }, // Lo  [31] NABATAEAN LETTER FINAL ALEPH..NABATAEAN LETTER TAW
    Interval{ 0x108E0, 0x108F2 }, // Lo  [19] HATRAN LETTER ALEPH..HATRAN LETTER QOPH
    Interval{ 0x108F4, 0x108F5 }, // Lo   [2] HATRAN LETTER SHIN..HATRAN LETTER TAW
    Interval{ 0x10900, 0x10915 }, // Lo  [22] PHOENICIAN LETTER ALF..PHOENICIAN LETTER TAU
    Interval{ 0x10920, 0x10939 }, // Lo  [26] LYDIAN LETTER A..LYDIAN LETTER C
    Interval{ 0x10940, 0x10959 }, // Lo  [26] SIDETIC LETTER N01..SIDETIC LETTER N26
    Interval{ 0x10980, 0x109B7 }, // Lo  [56] MEROITIC HIEROGLYPHIC LETTER A..MEROITIC CURSIVE LETTER DA
    Interval{ 0x109BE, 0x109BF }, // Lo   [2] MEROITIC CURSIVE LOGOGRAM RMT..MEROITIC CURSIVE LOGOGRAM IMN
    Interval{ 0x10A00, 0x10A00 }, // Lo       KHAROSHTHI LETTER A
    Interval{ 0x10A01, 0x10A03 }, // Mn   [3] KHAROSHTHI VOWEL SIGN I..KHAROSHTHI VOWEL SIGN VOCALIC R
    Interval{ 0x10A05, 0x10A06 }, // Mn   [2] KHAROSHTHI VOWEL SIGN E..KHAROSHTHI VOWEL SIGN O
    Interval{ 0x10A0C, 0x10A0F }, // Mn   [4] KHAROSHTHI VOWEL LENGTH MARK..KHAROSHTHI SIGN VISARGA
    Interval{ 0x10A10, 0x10A13 }, // Lo   [4] KHAROSHTHI LETTER KA..KHAROSHTHI LETTER GHA
    Interval{ 0x10A15, 0x10A17 }, // Lo   [3] KHAROSHTHI LETTER CA..KHAROSHTHI LETTER JA
    Interval{ 0x10A19, 0x10A35 }, // Lo  [29] KHAROSHTHI LETTER NYA..KHAROSHTHI LETTER VHA
    Interval{ 0x10A60, 0x10A7C }, // Lo  [29] OLD SOUTH ARABIAN LETTER HE..OLD SOUTH ARABIAN LETTER THETH
    Interval{ 0x10A80, 0x10A9C }, // Lo  [29] OLD NORTH ARABIAN LETTER HEH..OLD NORTH ARABIAN LETTER ZAH
    Interval{ 0x10AC0, 0x10AC7 }, // Lo   [8] MANICHAEAN LETTER ALEPH..MANICHAEAN LETTER WAW
    Interval{ 0x10AC9, 0x10AE4 }, // Lo  [28] MANICHAEAN LETTER ZAYIN..MANICHAEAN LETTER TAW
    Interval{ 0x10B00, 0x10B35 }, // Lo  [54] AVESTAN LETTER A..AVESTAN LETTER HE
    Interval{ 0x10B40, 0x10B55 }, // Lo  [22] INSCRIPTIONAL PARTHIAN LETTER ALEPH..INSCRIPTIONAL PARTHIAN LETTER TAW
    Interval{ 0x10B60, 0x10B72 }, // Lo  [19] INSCRIPTIONAL PAHLAVI LETTER ALEPH..INSCRIPTIONAL PAHLAVI LETTER TAW
    Interval{ 0x10B80, 0x10B91 }, // Lo  [18] PSALTER PAHLAVI LETTER ALEPH..PSALTER PAHLAVI LETTER TAW
    Interval{ 0x10C00, 0x10C48 }, // Lo  [73] OLD TURKIC LETTER ORKHON A..OLD TURKIC LETTER ORKHON BASH
    Interval{ 0x10C80, 0x10CB2 }, // L&  [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Interval{ 0x10CC0, 0x10CF2 }, // L&  [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Interval{ 0x10D00, 0x10D23 }, // Lo  [36] HANIFI ROHINGYA LETTER A..HANIFI ROHINGYA MARK NA KHONNA
    Interval{ 0x10D24, 0x10D27 }, // Mn   [4] HANIFI ROHINGYA SIGN HARBAHAY..HANIFI ROHINGYA SIGN TASSI
    Interval{ 0x10D4A, 0x10D4D }, // Lo   [4] GARAY VOWEL SIGN A..GARAY VOWEL SIGN EE
    Interval{ 0x10D4E, 0x10D4E }, // Lm       GARAY VOWEL LENGTH MARK
    Interval{ 0x10D4F, 0x10D4F }, // Lo       GARAY SUKUN
    Interval{ 0x10D50, 0x10D65 }, // L&  [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Interval{ 0x10D69, 0x10D69 }, // Mn       GARAY VOWEL SIGN E
    Interval{ 0x10D6F, 0x10D6F }, // Lm       GARAY REDUPLICATION MARK
    Interval{ 0x10D70, 0x10D85 }, // L&  [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Interval{ 0x10E80, 0x10EA9 }, // Lo  [42] YEZIDI LETTER ELIF..YEZIDI LETTER ET
    Interval{ 0x10EAB, 0x10EAC }, // Mn   [2] YEZIDI COMBINING HAMZA MARK..YEZIDI COMBINING MADDA MARK
    Interval{ 0x10EB0, 0x10EB1 }, // Lo   [2] YEZIDI LETTER LAM WITH DOT ABOVE..YEZIDI LETTER YOT WITH CIRCUMFLEX ABOVE
    Interval{ 0x10EC2, 0x10EC4 }, // Lo   [3] ARABIC LETTER DAL WITH TWO DOTS VERTICALLY BELOW..ARABIC LETTER KAF WITH TWO DOTS VERTICALLY BELOW
    Interval{ 0x10EC5, 0x10EC5 }, // Lm       ARABIC SMALL YEH BARREE WITH TWO DOTS BELOW
    Interval{ 0x10EC6, 0x10EC7 }, // Lo   [2] ARABIC LETTER THIN NOON..ARABIC LETTER YEH WITH FOUR DOTS BELOW
    Interval{ 0x10EFA, 0x10EFC }, // Mn   [3] ARABIC DOUBLE VERTICAL BAR BELOW..ARABIC COMBINING ALEF OVERLAY
    Interval{ 0x10F00, 0x10F1C }, // Lo  [29] OLD SOGDIAN LETTER ALEPH..OLD SOGDIAN LETTER FINAL TAW WITH VERTICAL TAIL
    Interval{ 0x10F27, 0x10F27 }, // Lo       OLD SOGDIAN LIGATURE AYIN-DALETH
    Interval{ 0x10F30, 0x10F45 }, // Lo  [22] SOGDIAN LETTER ALEPH..SOGDIAN INDEPENDENT SHIN
    Interval{ 0x10F70, 0x10F81 }, // Lo  [18] OLD UYGHUR LETTER ALEPH..OLD UYGHUR LETTER LESH
    Interval{ 0x10FB0, 0x10FC4 }, // Lo  [21] CHORASMIAN LETTER ALEPH..CHORASMIAN LETTER TAW
    Interval{ 0x10FE0, 0x10FF6 }, // Lo  [23] ELYMAIC LETTER ALEPH..ELYMAIC LIGATURE ZAYIN-YODH
    Interval{ 0x11000, 0x11000 }, // Mc       BRAHMI SIGN CANDRABINDU
    Interval{ 0x11001, 0x11001 }, // Mn       BRAHMI SIGN ANUSVARA
    Interval{ 0x11002, 0x11002 }, // Mc       BRAHMI SIGN VISARGA
    Interval{ 0x11003, 0x11037 }, // Lo  [53] BRAHMI SIGN JIHVAMULIYA..BRAHMI LETTER OLD TAMIL NNNA
    Interval{ 0x11038, 0x11045 }, // Mn  [14] BRAHMI VOWEL SIGN AA..BRAHMI VOWEL SIGN AU
    Interval{ 0x11071, 0x11072 }, // Lo   [2] BRAHMI LETTER OLD TAMIL SHORT E..BRAHMI LETTER OLD TAMIL SHORT O
    Interval{ 0x11073, 0x11074 }, // Mn   [2] BRAHMI VOWEL SIGN OLD TAMIL SHORT E..BRAHMI VOWEL SIGN OLD TAMIL SHORT O
    Interval{ 0x11075, 0x11075 }, // Lo       BRAHMI LETTER OLD TAMIL LLA
    Interval{ 0x11080, 0x11081 }, // Mn   [2] KAITHI SIGN CANDRABINDU..KAITHI SIGN ANUSVARA
    Interval{ 0x11082, 0x11082 }, // Mc       KAITHI SIGN VISARGA
    Interval{ 0x11083, 0x110AF }, // Lo  [45] KAITHI LETTER A..KAITHI LETTER HA
    Interval{ 0x110B0, 0x110B2 }, // Mc   [3] KAITHI VOWEL SIGN AA..KAITHI VOWEL SIGN II
    Interval{ 0x110B3, 0x110B6 }, // Mn   [4] KAITHI VOWEL SIGN U..KAITHI VOWEL SIGN AI
    Interval{ 0x110B7, 0x110B8 }, // Mc   [2] KAITHI VOWEL SIGN O..KAITHI VOWEL SIGN AU
    Interval{ 0x110C2, 0x110C2 }, // Mn       KAITHI VOWEL SIGN VOCALIC R
    Interval{ 0x110D0, 0x110E8 }, // Lo  [25] SORA SOMPENG LETTER SAH..SORA SOMPENG LETTER MAE
    Interval{ 0x11100, 0x11102 }, // Mn   [3] CHAKMA SIGN CANDRABINDU..CHAKMA SIGN VISARGA
    Interval{ 0x11103, 0x11126 }, // Lo  [36] CHAKMA LETTER AA..CHAKMA LETTER HAA
    Interval{ 0x11127, 0x1112B }, // Mn   [5] CHAKMA VOWEL SIGN A..CHAKMA VOWEL SIGN UU
    Interval{ 0x1112C, 0x1112C }, // Mc       CHAKMA VOWEL SIGN E
    Interval{ 0x1112D, 0x11132 }, // Mn   [6] CHAKMA VOWEL SIGN AI..CHAKMA AU MARK
    Interval{ 0x11144, 0x11144 }, // Lo       CHAKMA LETTER LHAA
    Interval{ 0x11145, 0x11146 }, // Mc   [2] CHAKMA VOWEL SIGN AA..CHAKMA VOWEL SIGN EI
    Interval{ 0x11147, 0x11147 }, // Lo       CHAKMA LETTER VAA
    Interval{ 0x11150, 0x11172 }, // Lo  [35] MAHAJANI LETTER A..MAHAJANI LETTER RRA
    Interval{ 0x11176, 0x11176 }, // Lo       MAHAJANI LIGATURE SHRI
    Interval{ 0x11180, 0x11181 }, // Mn   [2] SHARADA SIGN CANDRABINDU..SHARADA SIGN ANUSVARA
    Interval{ 0x11182, 0x11182 }, // Mc       SHARADA SIGN VISARGA
    Interval{ 0x11183, 0x111B2 }, // Lo  [48] SHARADA LETTER A..SHARADA LETTER HA
    Interval{ 0x111B3, 0x111B5 }, // Mc   [3] SHARADA VOWEL SIGN AA..SHARADA VOWEL SIGN II
    Interval{ 0x111B6, 0x111BE }, // Mn   [9] SHARADA VOWEL SIGN U..SHARADA VOWEL SIGN O
    Interval{ 0x111BF, 0x111BF }, // Mc       SHARADA VOWEL SIGN AU
    Interval{ 0x111C1, 0x111C4 }, // Lo   [4] SHARADA SIGN AVAGRAHA..SHARADA OM
    Interval{ 0x111CE, 0x111CE }, // Mc       SHARADA VOWEL SIGN PRISHTHAMATRA E
    Interval{ 0x111CF, 0x111CF }, // Mn       SHARADA SIGN INVERTED CANDRABINDU
    Interval{ 0x111DA, 0x111DA }, // Lo       SHARADA EKAM
    Interval{ 0x111DC, 0x111DC }, // Lo       SHARADA HEADSTROKE
    Interval{ 0x11200, 0x11211 }, // Lo  [18] KHOJKI LETTER A..KHOJKI LETTER JJA
    Interval{ 0x11213, 0x1122B }, // Lo  [25] KHOJKI LETTER NYA..KHOJKI LETTER LLA
    Interval{ 0x1122C, 0x1122E }, // Mc   [3] KHOJKI VOWEL SIGN AA..KHOJKI VOWEL SIGN II
    Interval{ 0x1122F, 0x11231 }, // Mn   [3] KHOJKI VOWEL SIGN U..KHOJKI VOWEL SIGN AI
    Interval{ 0x11232, 0x11233 }, // Mc   [2] KHOJKI VOWEL SIGN O..KHOJKI VOWEL SIGN AU
    Interval{ 0x11234, 0x11234 }, // Mn       KHOJKI SIGN ANUSVARA
    Interval{ 0x11237, 0x11237 }, // Mn       KHOJKI SIGN SHADDA
    Interval{ 0x1123E, 0x1123E }, // Mn       KHOJKI SIGN SUKUN
    Interval{ 0x1123F, 0x11240 }, // Lo   [2] KHOJKI LETTER QA..KHOJKI LETTER SHORT I
    Interval{ 0x11241, 0x11241 }, // Mn       KHOJKI VOWEL SIGN VOCALIC R
    Interval{ 0x11280, 0x11286 }, // Lo   [7] MULTANI LETTER A..MULTANI LETTER GA
    Interval{ 0x11288, 0x11288 }, // Lo       MULTANI LETTER GHA
    Interval{ 0x1128A, 0x1128D }, // Lo   [4] MULTANI LETTER CA..MULTANI LETTER JJA
    Interval{ 0x1128F, 0x1129D }, // Lo  [15] MULTANI LETTER NYA..MULTANI LETTER BA
    Interval{ 0x1129F, 0x112A8 }, // Lo  [10] MULTANI LETTER BHA..MULTANI LETTER RHA
    Interval{ 0x112B0, 0x112DE }, // Lo  [47] KHUDAWADI LETTER A..KHUDAWADI LETTER HA
    Interval{ 0x112DF, 0x112DF }, // Mn       KHUDAWADI SIGN ANUSVARA
    Interval{ 0x112E0, 0x112E2 }, // Mc   [3] KHUDAWADI VOWEL SIGN AA..KHUDAWADI VOWEL SIGN II
    Interval{ 0x112E3, 0x112E8 }, // Mn   [6] KHUDAWADI VOWEL SIGN U..KHUDAWADI VOWEL SIGN AU
    Interval{ 0x11300, 0x11301 }, // Mn   [2] GRANTHA SIGN COMBINING ANUSVARA ABOVE..GRANTHA SIGN CANDRABINDU
    Interval{ 0x11302, 0x11303 }, // Mc   [2] GRANTHA SIGN ANUSVARA..GRANTHA SIGN VISARGA
    Interval{ 0x11305, 0x1130C }, // Lo   [8] GRANTHA LETTER A..GRANTHA LETTER VOCALIC L
    Interval{ 0x1130F, 0x11310 }, // Lo   [2] GRANTHA LETTER EE..GRANTHA LETTER AI
    Interval{ 0x11313, 0x11328 }, // Lo  [22] GRANTHA LETTER OO..GRANTHA LETTER NA
    Interval{ 0x1132A, 0x11330 }, // Lo   [7] GRANTHA LETTER PA..GRANTHA LETTER RA
    Interval{ 0x11332, 0x11333 }, // Lo   [2] GRANTHA LETTER LA..GRANTHA LETTER LLA
    Interval{ 0x11335, 0x11339 }, // Lo   [5] GRANTHA LETTER VA..GRANTHA LETTER HA
    Interval{ 0x1133D, 0x1133D }, // Lo       GRANTHA SIGN AVAGRAHA
    Interval{ 0x1133E, 0x1133F }, // Mc   [2] GRANTHA VOWEL SIGN AA..GRANTHA VOWEL SIGN I
    Interval{ 0x11340, 0x11340 }, // Mn       GRANTHA VOWEL SIGN II
    Interval{ 0x11341, 0x11344 }, // Mc   [4] GRANTHA VOWEL SIGN U..GRANTHA VOWEL SIGN VOCALIC RR
    Interval{ 0x11347, 0x11348 }, // Mc   [2] GRANTHA VOWEL SIGN EE..GRANTHA VOWEL SIGN AI
    Interval{ 0x1134B, 0x1134C }, // Mc   [2] GRANTHA VOWEL SIGN OO..GRANTHA VOWEL SIGN AU
    Interval{ 0x11350, 0x11350 }, // Lo       GRANTHA OM
    Interval{ 0x11357, 0x11357 }, // Mc       GRANTHA AU LENGTH MARK
    Interval{ 0x1135D, 0x11361 }, // Lo   [5] GRANTHA SIGN PLUTA..GRANTHA LETTER VOCALIC LL
    Interval{ 0x11362, 0x11363 }, // Mc   [2] GRANTHA VOWEL SIGN VOCALIC L..GRANTHA VOWEL SIGN VOCALIC LL
    Interval{ 0x11380, 0x11389 }, // Lo  [10] TULU-TIGALARI LETTER A..TULU-TIGALARI LETTER VOCALIC LL
    Interval{ 0x1138B, 0x1138B }, // Lo       TULU-TIGALARI LETTER EE
    Interval{ 0x1138E, 0x1138E }, // Lo       TULU-TIGALARI LETTER AI
    Interval{ 0x11390, 0x113B5 }, // Lo  [38] TULU-TIGALARI LETTER OO..TULU-TIGALARI LETTER LLLA
    Interval{ 0x113B7, 0x113B7 }, // Lo       TULU-TIGALARI SIGN AVAGRAHA
    Interval{ 0x113B8, 0x113BA }, // Mc   [3] TULU-TIGALARI VOWEL SIGN AA..TULU-TIGALARI VOWEL SIGN II
    Interval{ 0x113BB, 0x113C0 }, // Mn   [6] TULU-TIGALARI VOWEL SIGN U..TULU-TIGALARI VOWEL SIGN VOCALIC LL
    Interval{ 0x113C2, 0x113C2 }, // Mc       TULU-TIGALARI VOWEL SIGN EE
    Interval{ 0x113C5, 0x113C5 }, // Mc       TULU-TIGALARI VOWEL SIGN AI
    Interval{ 0x113C7, 0x113CA }, // Mc   [4] TULU-TIGALARI VOWEL SIGN OO..TULU-TIGALARI SIGN CANDRA ANUNASIKA
    Interval{ 0x113CC, 0x113CD }, // Mc   [2] TULU-TIGALARI SIGN ANUSVARA..TULU-TIGALARI SIGN VISARGA
    Interval{ 0x113D1, 0x113D1 }, // Lo       TULU-TIGALARI REPHA
    Interval{ 0x113D3, 0x113D3 }, // Lo       TULU-TIGALARI SIGN PLUTA
    Interval{ 0x11400, 0x11434 }, // Lo  [53] NEWA LETTER A..NEWA LETTER HA
    Interval{ 0x11435, 0x11437 }, // Mc   [3] NEWA VOWEL SIGN AA..NEWA VOWEL SIGN II
    Interval{ 0x11438, 0x1143F }, // Mn   [8] NEWA VOWEL SIGN U..NEWA VOWEL SIGN AI
    Interval{ 0x11440, 0x11441 }, // Mc   [2] NEWA VOWEL SIGN O..NEWA VOWEL SIGN AU
    Interval{ 0x11443, 0x11444 }, // Mn   [2] NEWA SIGN CANDRABINDU..NEWA SIGN ANUSVARA
    Interval{ 0x11445, 0x11445 }, // Mc       NEWA SIGN VISARGA
    Interval{ 0x11447, 0x1144A }, // Lo   [4] NEWA SIGN AVAGRAHA..NEWA SIDDHI
    Interval{ 0x1145F, 0x11461 }, // Lo   [3] NEWA LETTER VEDIC ANUSVARA..NEWA SIGN UPADHMANIYA
    Interval{ 0x11480, 0x114AF }, // Lo  [48] TIRHUTA ANJI..TIRHUTA LETTER HA
    Interval{ 0x114B0, 0x114B2 }, // Mc   [3] TIRHUTA VOWEL SIGN AA..TIRHUTA VOWEL SIGN II
    Interval{ 0x114B3, 0x114B8 }, // Mn   [6] TIRHUTA VOWEL SIGN U..TIRHUTA VOWEL SIGN VOCALIC LL
    Interval{ 0x114B9, 0x114B9 }, // Mc       TIRHUTA VOWEL SIGN E
    Interval{ 0x114BA, 0x114BA }, // Mn       TIRHUTA VOWEL SIGN SHORT E
    Interval{ 0x114BB, 0x114BE }, // Mc   [4] TIRHUTA VOWEL SIGN AI..TIRHUTA VOWEL SIGN AU
    Interval{ 0x114BF, 0x114C0 }, // Mn   [2] TIRHUTA SIGN CANDRABINDU..TIRHUTA SIGN ANUSVARA
    Interval{ 0x114C1, 0x114C1 }, // Mc       TIRHUTA SIGN VISARGA
    Interval{ 0x114C4, 0x114C5 }, // Lo   [2] TIRHUTA SIGN AVAGRAHA..TIRHUTA GVANG
    Interval{ 0x114C7, 0x114C7 }, // Lo       TIRHUTA OM
    Interval{ 0x11580, 0x115AE }, // Lo  [47] SIDDHAM LETTER A..SIDDHAM LETTER HA
    Interval{ 0x115AF, 0x115B1 }, // Mc   [3] SIDDHAM VOWEL SIGN AA..SIDDHAM VOWEL SIGN II
    Interval{ 0x115B2, 0x115B5 }, // Mn   [4] SIDDHAM VOWEL SIGN U..SIDDHAM VOWEL SIGN VOCALIC RR
    Interval{ 0x115B8, 0x115BB }, // Mc   [4] SIDDHAM VOWEL SIGN E..SIDDHAM VOWEL SIGN AU
    Interval{ 0x115BC, 0x115BD }, // Mn   [2] SIDDHAM SIGN CANDRABINDU..SIDDHAM SIGN ANUSVARA
    Interval{ 0x115BE, 0x115BE }, // Mc       SIDDHAM SIGN VISARGA
    Interval{ 0x115D8, 0x115DB }, // Lo   [4] SIDDHAM LETTER THREE-CIRCLE ALTERNATE I..SIDDHAM LETTER ALTERNATE U
    Interval{ 0x115DC, 0x115DD }, // Mn   [2] SIDDHAM VOWEL SIGN ALTERNATE U..SIDDHAM VOWEL SIGN ALTERNATE UU
    Interval{ 0x11600, 0x1162F }, // Lo  [48] MODI LETTER A..MODI LETTER LLA
    Interval{ 0x11630, 0x11632 }, // Mc   [3] MODI VOWEL SIGN AA..MODI VOWEL SIGN II
    Interval{ 0x11633, 0x1163A }, // Mn   [8] MODI VOWEL SIGN U..MODI VOWEL SIGN AI
    Interval{ 0x1163B, 0x1163C }, // Mc   [2] MODI VOWEL SIGN O..MODI VOWEL SIGN AU
    Interval{ 0x1163D, 0x1163D }, // Mn       MODI SIGN ANUSVARA
    Interval{ 0x1163E, 0x1163E }, // Mc       MODI SIGN VISARGA
    Interval{ 0x11640, 0x11640 }, // Mn       MODI SIGN ARDHACANDRA
    Interval{ 0x11644, 0x11644 }, // Lo       MODI SIGN HUVA
    Interval{ 0x11680, 0x116AA }, // Lo  [43] TAKRI LETTER A..TAKRI LETTER RRA
    Interval{ 0x116AB, 0x116AB }, // Mn       TAKRI SIGN ANUSVARA
    Interval{ 0x116AC, 0x116AC }, // Mc       TAKRI SIGN VISARGA
    Interval{ 0x116AD, 0x116AD }, // Mn       TAKRI VOWEL SIGN AA
    Interval{ 0x116AE, 0x116AF }, // Mc   [2] TAKRI VOWEL SIGN I..TAKRI VOWEL SIGN II
    Interval{ 0x116B0, 0x116B5 }, // Mn   [6] TAKRI VOWEL SIGN U..TAKRI VOWEL SIGN AU
    Interval{ 0x116B8, 0x116B8 }, // Lo       TAKRI LETTER ARCHAIC KHA
    Interval{ 0x11700, 0x1171A }, // Lo  [27] AHOM LETTER KA..AHOM LETTER ALTERNATE BA
    Interval{ 0x1171D, 0x1171D }, // Mn       AHOM CONSONANT SIGN MEDIAL LA
    Interval{ 0x1171E, 0x1171E }, // Mc       AHOM CONSONANT SIGN MEDIAL RA
    Interval{ 0x1171F, 0x1171F }, // Mn       AHOM CONSONANT SIGN MEDIAL LIGATING RA
    Interval{ 0x11720, 0x11721 }, // Mc   [2] AHOM VOWEL SIGN A..AHOM VOWEL SIGN AA
    Interval{ 0x11722, 0x11725 }, // Mn   [4] AHOM VOWEL SIGN I..AHOM VOWEL SIGN UU
    Interval{ 0x11726, 0x11726 }, // Mc       AHOM VOWEL SIGN E
    Interval{ 0x11727, 0x1172A }, // Mn   [4] AHOM VOWEL SIGN AW..AHOM VOWEL SIGN AM
    Interval{ 0x11740, 0x11746 }, // Lo   [7] AHOM LETTER CA..AHOM LETTER LLA
    Interval{ 0x11800, 0x1182B }, // Lo  [44] DOGRA LETTER A..DOGRA LETTER RRA
    Interval{ 0x1182C, 0x1182E }, // Mc   [3] DOGRA VOWEL SIGN AA..DOGRA VOWEL SIGN II
    Interval{ 0x1182F, 0x11837 }, // Mn   [9] DOGRA VOWEL SIGN U..DOGRA SIGN ANUSVARA
    Interval{ 0x11838, 0x11838 }, // Mc       DOGRA SIGN VISARGA
    Interval{ 0x118A0, 0x118DF }, // L&  [64] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Interval{ 0x118FF, 0x11906 }, // Lo   [8] WARANG CITI OM..DIVES AKURU LETTER E
    Interval{ 0x11909, 0x11909 }, // Lo       DIVES AKURU LETTER O
    Interval{ 0x1190C, 0x11913 }, // Lo   [8] DIVES AKURU LETTER KA..DIVES AKURU LETTER JA
    Interval{ 0x11915, 0x11916 }, // Lo   [2] DIVES AKURU LETTER NYA..DIVES AKURU LETTER TTA
    Interval{ 0x11918, 0x1192F }, // Lo  [24] DIVES AKURU LETTER DDA..DIVES AKURU LETTER ZA
    Interval{ 0x11930, 0x11935 }, // Mc   [6] DIVES AKURU VOWEL SIGN AA..DIVES AKURU VOWEL SIGN E
    Interval{ 0x11937, 0x11938 }, // Mc   [2] DIVES AKURU VOWEL SIGN AI..DIVES AKURU VOWEL SIGN O
    Interval{ 0x1193B, 0x1193C }, // Mn   [2] DIVES AKURU SIGN ANUSVARA..DIVES AKURU SIGN CANDRABINDU
    Interval{ 0x1193F, 0x1193F }, // Lo       DIVES AKURU PREFIXED NASAL SIGN
    Interval{ 0x11940, 0x11940 }, // Mc       DIVES AKURU MEDIAL YA
    Interval{ 0x11941, 0x11941 }, // Lo       DIVES AKURU INITIAL RA
    Interval{ 0x11942, 0x11942 }, // Mc       DIVES AKURU MEDIAL RA
    Interval{ 0x119A0, 0x119A7 }, // Lo   [8] NANDINAGARI LETTER A..NANDINAGARI LETTER VOCALIC RR
    Interval{ 0x119AA, 0x119D0 }, // Lo  [39] NANDINAGARI LETTER E..NANDINAGARI LETTER RRA
    Interval{ 0x119D1, 0x119D3 }, // Mc   [3] NANDINAGARI VOWEL SIGN AA..NANDINAGARI VOWEL SIGN II
    Interval{ 0x119D4, 0x119D7 }, // Mn   [4] NANDINAGARI VOWEL SIGN U..NANDINAGARI VOWEL SIGN VOCALIC RR
    Interval{ 0x119DA, 0x119DB }, // Mn   [2] NANDINAGARI VOWEL SIGN E..NANDINAGARI VOWEL SIGN AI
    Interval{ 0x119DC, 0x119DF }, // Mc   [4] NANDINAGARI VOWEL SIGN O..NANDINAGARI SIGN VISARGA
    Interval{ 0x119E1, 0x119E1 }, // Lo       NANDINAGARI SIGN AVAGRAHA
    Interval{ 0x119E3, 0x119E3 }, // Lo       NANDINAGARI HEADSTROKE
    Interval{ 0x119E4, 0x119E4 }, // Mc       NANDINAGARI VOWEL SIGN PRISHTHAMATRA E
    Interval{ 0x11A00, 0x11A00 }, // Lo       ZANABAZAR SQUARE LETTER A
    Interval{ 0x11A01, 0x11A0A }, // Mn  [10] ZANABAZAR SQUARE VOWEL SIGN I..ZANABAZAR SQUARE VOWEL LENGTH MARK
    Interval{ 0x11A0B, 0x11A32 }, // Lo  [40] ZANABAZAR SQUARE LETTER KA..ZANABAZAR SQUARE LETTER KSSA
    Interval{ 0x11A35, 0x11A38 }, // Mn   [4] ZANABAZAR SQUARE SIGN CANDRABINDU..ZANABAZAR SQUARE SIGN ANUSVARA
    Interval{ 0x11A39, 0x11A39 }, // Mc       ZANABAZAR SQUARE SIGN VISARGA
    Interval{ 0x11A3A, 0x11A3A }, // Lo       ZANABAZAR SQUARE CLUSTER-INITIAL LETTER RA
    Interval{ 0x11A3B, 0x11A3E }, // Mn   [4] ZANABAZAR SQUARE CLUSTER-FINAL LETTER YA..ZANABAZAR SQUARE CLUSTER-FINAL LETTER VA
    Interval{ 0x11A50, 0x11A50 }, // Lo       SOYOMBO LETTER A
    Interval{ 0x11A51, 0x11A56 }, // Mn   [6] SOYOMBO VOWEL SIGN I..SOYOMBO VOWEL SIGN OE
    Interval{ 0x11A57, 0x11A58 }, // Mc   [2] SOYOMBO VOWEL SIGN AI..SOYOMBO VOWEL SIGN AU
    Interval{ 0x11A59, 0x11A5B }, // Mn   [3] SOYOMBO VOWEL SIGN VOCALIC R..SOYOMBO VOWEL LENGTH MARK
    Interval{ 0x11A5C, 0x11A89 }, // Lo  [46] SOYOMBO LETTER KA..SOYOMBO CLUSTER-INITIAL LETTER SA
    Interval{ 0x11A8A, 0x11A96 }, // Mn  [13] SOYOMBO FINAL CONSONANT SIGN G..SOYOMBO SIGN ANUSVARA
    Interval{ 0x11A97, 0x11A97 }, // Mc       SOYOMBO SIGN VISARGA
    Interval{ 0x11A9D, 0x11A9D }, // Lo       SOYOMBO MARK PLUTA
    Interval{ 0x11AB0, 0x11AF8 }, // Lo  [73] CANADIAN SYLLABICS NATTILIK HI..PAU CIN HAU GLOTTAL STOP FINAL
    Interval{ 0x11B60, 0x11B60 }, // Mn       SHARADA VOWEL SIGN OE
    Interval{ 0x11B61, 0x11B61 }, // Mc       SHARADA VOWEL SIGN OOE
    Interval{ 0x11B62, 0x11B64 }, // Mn   [3] SHARADA VOWEL SIGN UE..SHARADA VOWEL SIGN SHORT E
    Interval{ 0x11B65, 0x11B65 }, // Mc       SHARADA VOWEL SIGN SHORT O
    Interval{ 0x11B66, 0x11B66 }, // Mn       SHARADA VOWEL SIGN CANDRA E
    Interval{ 0x11B67, 0x11B67 }, // Mc       SHARADA VOWEL SIGN CANDRA O
    Interval{ 0x11BC0, 0x11BE0 }, // Lo  [33] SUNUWAR LETTER DEVI..SUNUWAR LETTER KLOKO
    Interval{ 0x11C00, 0x11C08 }, // Lo   [9] BHAIKSUKI LETTER A..BHAIKSUKI LETTER VOCALIC L
    Interval{ 0x11C0A, 0x11C2E }, // Lo  [37] BHAIKSUKI LETTER E..BHAIKSUKI LETTER HA
    Interval{ 0x11C2F, 0x11C2F }, // Mc       BHAIKSUKI VOWEL SIGN AA
    Interval{ 0x11C30, 0x11C36 }, // Mn   [7] BHAIKSUKI VOWEL SIGN I..BHAIKSUKI VOWEL SIGN VOCALIC L
    Interval{ 0x11C38, 0x11C3D }, // Mn   [6] BHAIKSUKI VOWEL SIGN E..BHAIKSUKI SIGN ANUSVARA
    Interval{ 0x11C3E, 0x11C3E }, // Mc       BHAIKSUKI SIGN VISARGA
    Interval{ 0x11C40, 0x11C40 }, // Lo       BHAIKSUKI SIGN AVAGRAHA
    Interval{ 0x11C72, 0x11C8F }, // Lo  [30] MARCHEN LETTER KA..MARCHEN LETTER A
    Interval{ 0x11C92, 0x11CA7 }, // Mn  [22] MARCHEN SUBJOINED LETTER KA..MARCHEN SUBJOINED LETTER ZA
    Interval{ 0x11CA9, 0x11CA9 }, // Mc       MARCHEN SUBJOINED LETTER YA
    Interval{ 0x11CAA, 0x11CB0 }, // Mn   [7] MARCHEN SUBJOINED LETTER RA..MARCHEN VOWEL SIGN AA
    Interval{ 0x11CB1, 0x11CB1 }, // Mc       MARCHEN VOWEL SIGN I
    Interval{ 0x11CB2, 0x11CB3 }, // Mn   [2] MARCHEN VOWEL SIGN U..MARCHEN VOWEL SIGN E
    Interval{ 0x11CB4, 0x11CB4 }, // Mc       MARCHEN VOWEL SIGN O
    Interval{ 0x11CB5, 0x11CB6 }, // Mn   [2] MARCHEN SIGN ANUSVARA..MARCHEN SIGN CANDRABINDU
    Interval{ 0x11D00, 0x11D06 }, // Lo   [7] MASARAM GONDI LETTER A..MASARAM GONDI LETTER E
    Interval{ 0x11D08, 0x11D09 }, // Lo   [2] MASARAM GONDI LETTER AI..MASARAM GONDI LETTER O
    Interval{ 0x11D0B, 0x11D30 }, // Lo  [38] MASARAM GONDI LETTER AU..MASARAM GONDI LETTER TRA
    Interval{ 0x11D31, 0x11D36 }, // Mn   [6] MASARAM GONDI VOWEL SIGN AA..MASARAM GONDI VOWEL SIGN VOCALIC R
    Interval{ 0x11D3A, 0x11D3A }, // Mn       MASARAM GONDI VOWEL SIGN E
    Interval{ 0x11D3C, 0x11D3D }, // Mn   [2] MASARAM GONDI VOWEL SIGN AI..MASARAM GONDI VOWEL SIGN O
    Interval{ 0x11D3F, 0x11D41 }, // Mn   [3] MASARAM GONDI VOWEL SIGN AU..MASARAM GONDI SIGN VISARGA
    Interval{ 0x11D43, 0x11D43 }, // Mn       MASARAM GONDI SIGN CANDRA
    Interval{ 0x11D46, 0x11D46 }, // Lo       MASARAM GONDI REPHA
    Interval{ 0x11D47, 0x11D47 }, // Mn       MASARAM GONDI RA-KARA
    Interval{ 0x11D60, 0x11D65 }, // Lo   [6] GUNJALA GONDI LETTER A..GUNJALA GONDI LETTER UU
    Interval{ 0x11D67, 0x11D68 }, // Lo   [2] GUNJALA GONDI LETTER EE..GUNJALA GONDI LETTER AI
    Interval{ 0x11D6A, 0x11D89 }, // Lo  [32] GUNJALA GONDI LETTER OO..GUNJALA GONDI LETTER SA
    Interval{ 0x11D8A, 0x11D8E }, // Mc   [5] GUNJALA GONDI VOWEL SIGN AA..GUNJALA GONDI VOWEL SIGN UU
    Interval{ 0x11D90, 0x11D91 }, // Mn   [2] GUNJALA GONDI VOWEL SIGN EE..GUNJALA GONDI VOWEL SIGN AI
    Interval{ 0x11D93, 0x11D94 }, // Mc   [2] GUNJALA GONDI VOWEL SIGN OO..GUNJALA GONDI VOWEL SIGN AU
    Interval{ 0x11D95, 0x11D95 }, // Mn       GUNJALA GONDI SIGN ANUSVARA
    Interval{ 0x11D96, 0x11D96 }, // Mc       GUNJALA GONDI SIGN VISARGA
    Interval{ 0x11D98, 0x11D98 }, // Lo       GUNJALA GONDI OM
    Interval{ 0x11DB0, 0x11DD8 }, // Lo  [41] TOLONG SIKI LETTER I..TOLONG SIKI LETTER RRH
    Interval{ 0x11DD9, 0x11DD9 }, // Lm       TOLONG SIKI SIGN SELA
    Interval{ 0x11DDA, 0x11DDB }, // Lo   [2] TOLONG SIKI SIGN HECAKA..TOLONG SIKI UNGGA
    Interval{ 0x11EE0, 0x11EF2 }, // Lo  [19] MAKASAR LETTER KA..MAKASAR ANGKA
    Interval{ 0x11EF3, 0x11EF4 }, // Mn   [2] MAKASAR VOWEL SIGN I..MAKASAR VOWEL SIGN U
    Interval{ 0x11EF5, 0x11EF6 }, // Mc   [2] MAKASAR VOWEL SIGN E..MAKASAR VOWEL SIGN O
    Interval{ 0x11F00, 0x11F01 }, // Mn   [2] KAWI SIGN CANDRABINDU..KAWI SIGN ANUSVARA
    Interval{ 0x11F02, 0x11F02 }, // Lo       KAWI SIGN REPHA
    Interval{ 0x11F03, 0x11F03 }, // Mc       KAWI SIGN VISARGA
    Interval{ 0x11F04, 0x11F10 }, // Lo  [13] KAWI LETTER A..KAWI LETTER O
    Interval{ 0x11F12, 0x11F33 }, // Lo  [34] KAWI LETTER KA..KAWI LETTER JNYA
    Interval{ 0x11F34, 0x11F35 }, // Mc   [2] KAWI VOWEL SIGN AA..KAWI VOWEL SIGN ALTERNATE AA
    Interval{ 0x11F36, 0x11F3A }, // Mn   [5] KAWI VOWEL SIGN I..KAWI VOWEL SIGN VOCALIC R
    Interval{ 0x11F3E, 0x11F3F }, // Mc   [2] KAWI VOWEL SIGN E..KAWI VOWEL SIGN AI
    Interval{ 0x11F40, 0x11F40 }, // Mn       KAWI VOWEL SIGN EU
    Interval{ 0x11FB0, 0x11FB0 }, // Lo       LISU LETTER YHA
    Interval{ 0x12000, 0x12399 }, // Lo [922] CUNEIFORM SIGN A..CUNEIFORM SIGN U U
    Interval{ 0x12400, 0x1246E }, // Nl [111] CUNEIFORM NUMERIC SIGN TWO ASH..CUNEIFORM NUMERIC SIGN NINE U VARIANT FORM
    Interval{ 0x12480, 0x12543 }, // Lo [196] CUNEIFORM SIGN AB TIMES NUN TENU..CUNEIFORM SIGN ZU5 TIMES THREE DISH TENU
    Interval{ 0x12F90, 0x12FF0 }, // Lo  [97] CYPRO-MINOAN SIGN CM001..CYPRO-MINOAN SIGN CM114
    Interval{ 0x13000, 0x1342F }, // Lo [1072] EGYPTIAN HIEROGLYPH A001..EGYPTIAN HIEROGLYPH V011D
    Interval{ 0x13441, 0x13446 }, // Lo   [6] EGYPTIAN HIEROGLYPH FULL BLANK..EGYPTIAN HIEROGLYPH WIDE LOST SIGN
    Interval{ 0x13460, 0x143FA }, // Lo [3995] EGYPTIAN HIEROGLYPH-13460..EGYPTIAN HIEROGLYPH-143FA
    Interval{ 0x14400, 0x14646 }, // Lo [583] ANATOLIAN HIEROGLYPH A001..ANATOLIAN HIEROGLYPH A530
    Interval{ 0x16100, 0x1611D }, // Lo  [30] GURUNG KHEMA LETTER A..GURUNG KHEMA LETTER SA
    Interval{ 0x1611E, 0x16129 }, // Mn  [12] GURUNG KHEMA VOWEL SIGN AA..GURUNG KHEMA VOWEL LENGTH MARK
    Interval{ 0x1612A, 0x1612C }, // Mc   [3] GURUNG KHEMA CONSONANT SIGN MEDIAL YA..GURUNG KHEMA CONSONANT SIGN MEDIAL HA
    Interval{ 0x1612D, 0x1612E }, // Mn   [2] GURUNG KHEMA SIGN ANUSVARA..GURUNG KHEMA CONSONANT SIGN MEDIAL RA
    Interval{ 0x16800, 0x16A38 }, // Lo [569] BAMUM LETTER PHASE-A NGKUE MFON..BAMUM LETTER PHASE-F VUEQ
    Interval{ 0x16A40, 0x16A5E }, // Lo  [31] MRO LETTER TA..MRO LETTER TEK
    Interval{ 0x16A70, 0x16ABE }, // Lo  [79] TANGSA LETTER OZ..TANGSA LETTER ZA
    Interval{ 0x16AD0, 0x16AED }, // Lo  [30] BASSA VAH LETTER ENNI..BASSA VAH LETTER I
    Interval{ 0x16B00, 0x16B2F }, // Lo  [48] PAHAWH HMONG VOWEL KEEB..PAHAWH HMONG CONSONANT CAU
    Interval{ 0x16B40, 0x16B43 }, // Lm   [4] PAHAWH HMONG SIGN VOS SEEV..PAHAWH HMONG SIGN IB YAM
    Interval{ 0x16B63, 0x16B77 }, // Lo  [21] PAHAWH HMONG SIGN VOS LUB..PAHAWH HMONG SIGN CIM NRES TOS
    Interval{ 0x16B7D, 0x16B8F }, // Lo  [19] PAHAWH HMONG CLAN SIGN TSHEEJ..PAHAWH HMONG CLAN SIGN VWJ
    Interval{ 0x16D40, 0x16D42 }, // Lm   [3] KIRAT RAI SIGN ANUSVARA..KIRAT RAI SIGN VISARGA
    Interval{ 0x16D43, 0x16D6A }, // Lo  [40] KIRAT RAI LETTER A..KIRAT RAI VOWEL SIGN AU
    Interval{ 0x16D6B, 0x16D6C }, // Lm   [2] KIRAT RAI SIGN VIRAMA..KIRAT RAI SIGN SAAT
    Interval{ 0x16E40, 0x16E7F }, // L&  [64] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Interval{ 0x16EA0, 0x16EB8 }, // L&  [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Interval{ 0x16EBB, 0x16ED3 }, // L&  [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Interval{ 0x16F00, 0x16F4A }, // Lo  [75] MIAO LETTER PA..MIAO LETTER RTE
    Interval{ 0x16F4F, 0x16F4F }, // Mn       MIAO SIGN CONSONANT MODIFIER BAR
    Interval{ 0x16F50, 0x16F50 }, // Lo       MIAO LETTER NASALIZATION
    Interval{ 0x16F51, 0x16F87 }, // Mc  [55] MIAO SIGN ASPIRATION..MIAO VOWEL SIGN UI
    Interval{ 0x16F8F, 0x16F92 }, // Mn   [4] MIAO TONE RIGHT..MIAO TONE BELOW
    Interval{ 0x16F93, 0x16F9F }, // Lm  [13] MIAO LETTER TONE-2..MIAO LETTER REFORMED TONE-8
    Interval{ 0x16FE0, 0x16FE1 }, // Lm   [2] TANGUT ITERATION MARK..NUSHU ITERATION MARK
    Interval{ 0x16FE3, 0x16FE3 }, // Lm       OLD CHINESE ITERATION MARK
    Interval{ 0x16FF0, 0x16FF1 }, // Mc   [2] VIETNAMESE ALTERNATE READING MARK CA..VIETNAMESE ALTERNATE READING MARK NHAY
    Interval{ 0x16FF2, 0x16FF3 }, // Lm   [2] CHINESE SMALL SIMPLIFIED ER..CHINESE SMALL TRADITIONAL ER
    Interval{ 0x16FF4, 0x16FF6 }, // Nl   [3] YANGQIN SIGN SLOW ONE BEAT..YANGQIN SIGN SLOW TWO BEATS
    Interval{ 0x17000, 0x18CD5 }, // Lo [7382] TANGUT IDEOGRAPH-17000..KHITAN SMALL SCRIPT CHARACTER-18CD5
    Interval{ 0x18CFF, 0x18D1E }, // Lo  [32] KHITAN SMALL SCRIPT CHARACTER-18CFF..TANGUT IDEOGRAPH-18D1E
    Interval{ 0x18D80, 0x18DF2 }, // Lo [115] TANGUT COMPONENT-769..TANGUT COMPONENT-883
    Interval{ 0x1AFF0, 0x1AFF3 }, // Lm   [4] KATAKANA LETTER MINNAN TONE-2..KATAKANA LETTER MINNAN TONE-5
    Interval{ 0x1AFF5, 0x1AFFB }, // Lm   [7] KATAKANA LETTER MINNAN TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-5
    Interval{ 0x1AFFD, 0x1AFFE }, // Lm   [2] KATAKANA LETTER MINNAN NASALIZED TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-8
    Interval{ 0x1B000, 0x1B122 }, // Lo [291] KATAKANA LETTER ARCHAIC E..KATAKANA LETTER ARCHAIC WU
    Interval{ 0x1B132, 0x1B132 }, // Lo       HIRAGANA LETTER SMALL KO
    Interval{ 0x1B150, 0x1B152 }, // Lo   [3] HIRAGANA LETTER SMALL WI..HIRAGANA LETTER SMALL WO
    Interval{ 0x1B155, 0x1B155 }, // Lo       KATAKANA LETTER SMALL KO
    Interval{ 0x1B164, 0x1B167 }, // Lo   [4] KATAKANA LETTER SMALL WI..KATAKANA LETTER SMALL N
    Interval{ 0x1B170, 0x1B2FB }, // Lo [396] NUSHU CHARACTER-1B170..NUSHU CHARACTER-1B2FB
    Interval{ 0x1BC00, 0x1BC6A }, // Lo [107] DUPLOYAN LETTER H..DUPLOYAN LETTER VOCALIC M
    Interval{ 0x1BC70, 0x1BC7C }, // Lo  [13] DUPLOYAN AFFIX LEFT HORIZONTAL SECANT..DUPLOYAN AFFIX ATTACHED TANGENT HOOK
    Interval{ 0x1BC80, 0x1BC88 }, // Lo   [9] DUPLOYAN AFFIX HIGH ACUTE..DUPLOYAN AFFIX HIGH VERTICAL
    Interval{ 0x1BC90, 0x1BC99 }, // Lo  [10] DUPLOYAN AFFIX LOW ACUTE..DUPLOYAN AFFIX LOW ARROW
    Interval{ 0x1BC9E, 0x1BC9E }, // Mn       DUPLOYAN DOUBLE MARK
    Interval{ 0x1D400, 0x1D454 }, // L&  [85] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL ITALIC SMALL G
    Interval{ 0x1D456, 0x1D49C }, // L&  [71] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL SCRIPT CAPITAL A
    Interval{ 0x1D49E, 0x1D49F }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Interval{ 0x1D4A2, 0x1D4A2 }, // L&       MATHEMATICAL SCRIPT CAPITAL G
    Interval{ 0x1D4A5, 0x1D4A6 }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Interval{ 0x1D4A9, 0x1D4AC }, // L&   [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Interval{ 0x1D4AE, 0x1D4B9 }, // L&  [12] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT SMALL D
    Interval{ 0x1D4BB, 0x1D4BB }, // L&       MATHEMATICAL SCRIPT SMALL F
    Interval{ 0x1D4BD, 0x1D4C3 }, // L&   [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Interval{ 0x1D4C5, 0x1D505 }, // L&  [65] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL FRAKTUR CAPITAL B
    Interval{ 0x1D507, 0x1D50A }, // L&   [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Interval{ 0x1D50D, 0x1D514 }, // L&   [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Interval{ 0x1D516, 0x1D51C }, // L&   [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Interval{ 0x1D51E, 0x1D539 }, // L&  [28] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Interval{ 0x1D53B, 0x1D53E }, // L&   [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Interval{ 0x1D540, 0x1D544 }, // L&   [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Interval{ 0x1D546, 0x1D546 }, // L&       MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Interval{ 0x1D54A, 0x1D550 }, // L&   [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Interval{ 0x1D552, 0x1D6A5 }, // L& [340] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    Interval{ 0x1D6A8, 0x1D6C0 }, // L&  [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    Interval{ 0x1D6C2, 0x1D6DA }, // L&  [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    Interval{ 0x1D6DC, 0x1D6FA }, // L&  [31] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL ITALIC CAPITAL OMEGA
    Interval{ 0x1D6FC, 0x1D714 }, // L&  [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    Interval{ 0x1D716, 0x1D734 }, // L&  [31] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D736, 0x1D74E }, // L&  [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D750, 0x1D76E }, // L&  [31] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    Interval{ 0x1D770, 0x1D788 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    Interval{ 0x1D78A, 0x1D7A8 }, // L&  [31] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D7AA, 0x1D7C2 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D7C4, 0x1D7CB }, // L&   [8] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD SMALL DIGAMMA
    Interval{ 0x1DF00, 0x1DF09 }, // L&  [10] LATIN SMALL LETTER FENG DIGRAPH WITH TRILL..LATIN SMALL LETTER T WITH HOOK AND RETROFLEX HOOK
    Interval{ 0x1DF0A, 0x1DF0A }, // Lo       LATIN LETTER RETROFLEX CLICK WITH RETROFLEX HOOK
    Interval{ 0x1DF0B, 0x1DF1E }, // L&  [20] LATIN SMALL LETTER ESH WITH DOUBLE BAR..LATIN SMALL LETTER S WITH CURL
    Interval{ 0x1DF25, 0x1DF2A }, // L&   [6] LATIN SMALL LETTER D WITH MID-HEIGHT LEFT HOOK..LATIN SMALL LETTER T WITH MID-HEIGHT LEFT HOOK
    Interval{ 0x1E000, 0x1E006 }, // Mn   [7] COMBINING GLAGOLITIC LETTER AZU..COMBINING GLAGOLITIC LETTER ZHIVETE
    Interval{ 0x1E008, 0x1E018 }, // Mn  [17] COMBINING GLAGOLITIC LETTER ZEMLJA..COMBINING GLAGOLITIC LETTER HERU
    Interval{ 0x1E01B, 0x1E021 }, // Mn   [7] COMBINING GLAGOLITIC LETTER SHTA..COMBINING GLAGOLITIC LETTER YATI
    Interval{ 0x1E023, 0x1E024 }, // Mn   [2] COMBINING GLAGOLITIC LETTER YU..COMBINING GLAGOLITIC LETTER SMALL YUS
    Interval{ 0x1E026, 0x1E02A }, // Mn   [5] COMBINING GLAGOLITIC LETTER YO..COMBINING GLAGOLITIC LETTER FITA
    Interval{ 0x1E030, 0x1E06D }, // Lm  [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Interval{ 0x1E08F, 0x1E08F }, // Mn       COMBINING CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
    Interval{ 0x1E100, 0x1E12C }, // Lo  [45] NYIAKENG PUACHUE HMONG LETTER MA..NYIAKENG PUACHUE HMONG LETTER W
    Interval{ 0x1E137, 0x1E13D }, // Lm   [7] NYIAKENG PUACHUE HMONG SIGN FOR PERSON..NYIAKENG PUACHUE HMONG SYLLABLE LENGTHENER
    Interval{ 0x1E14E, 0x1E14E }, // Lo       NYIAKENG PUACHUE HMONG LOGOGRAM NYAJ
    Interval{ 0x1E290, 0x1E2AD }, // Lo  [30] TOTO LETTER PA..TOTO LETTER A
    Interval{ 0x1E2C0, 0x1E2EB }, // Lo  [44] WANCHO LETTER AA..WANCHO LETTER YIH
    Interval{ 0x1E4D0, 0x1E4EA }, // Lo  [27] NAG MUNDARI LETTER O..NAG MUNDARI LETTER ELL
    Interval{ 0x1E4EB, 0x1E4EB }, // Lm       NAG MUNDARI SIGN OJOD
    Interval{ 0x1E5D0, 0x1E5ED }, // Lo  [30] OL ONAL LETTER O..OL ONAL LETTER EG
    Interval{ 0x1E5F0, 0x1E5F0 }, // Lo       OL ONAL SIGN HODDOND
    Interval{ 0x1E6C0, 0x1E6DE }, // Lo  [31] TAI YO LETTER LOW KO..TAI YO LETTER HIGH KVO
    Interval{ 0x1E6E0, 0x1E6E2 }, // Lo   [3] TAI YO LETTER AA..TAI YO LETTER UE
    Interval{ 0x1E6E3, 0x1E6E3 }, // Mn       TAI YO SIGN UE
    Interval{ 0x1E6E4, 0x1E6E5 }, // Lo   [2] TAI YO LETTER U..TAI YO LETTER AE
    Interval{ 0x1E6E6, 0x1E6E6 }, // Mn       TAI YO SIGN AU
    Interval{ 0x1E6E7, 0x1E6ED }, // Lo   [7] TAI YO LETTER O..TAI YO LETTER AUE
    Interval{ 0x1E6EE, 0x1E6EF }, // Mn   [2] TAI YO SIGN AY..TAI YO SIGN ANG
    Interval{ 0x1E6F0, 0x1E6F4 }, // Lo   [5] TAI YO LETTER AN..TAI YO LETTER AP
    Interval{ 0x1E6F5, 0x1E6F5 }, // Mn       TAI YO SIGN OM
    Interval{ 0x1E6FE, 0x1E6FE }, // Lo       TAI YO SYMBOL MUEANG
    Interval{ 0x1E6FF, 0x1E6FF }, // Lm       TAI YO XAM LAI
    Interval{ 0x1E7E0, 0x1E7E6 }, // Lo   [7] ETHIOPIC SYLLABLE HHYA..ETHIOPIC SYLLABLE HHYO
    Interval{ 0x1E7E8, 0x1E7EB }, // Lo   [4] ETHIOPIC SYLLABLE GURAGE HHWA..ETHIOPIC SYLLABLE HHWE
    Interval{ 0x1E7ED, 0x1E7EE }, // Lo   [2] ETHIOPIC SYLLABLE GURAGE MWI..ETHIOPIC SYLLABLE GURAGE MWEE
    Interval{ 0x1E7F0, 0x1E7FE }, // Lo  [15] ETHIOPIC SYLLABLE GURAGE QWI..ETHIOPIC SYLLABLE GURAGE PWEE
    Interval{ 0x1E800, 0x1E8C4 }, // Lo [197] MENDE KIKAKUI SYLLABLE M001 KI..MENDE KIKAKUI SYLLABLE M060 NYON
    Interval{ 0x1E900, 0x1E943 }, // L&  [68] ADLAM CAPITAL LETTER ALIF..ADLAM SMALL LETTER SHA
    Interval{ 0x1E947, 0x1E947 }, // Mn       ADLAM HAMZA
    Interval{ 0x1E94B, 0x1E94B }, // Lm       ADLAM NASALIZATION MARK
    Interval{ 0x1EE00, 0x1EE03 }, // Lo   [4] ARABIC MATHEMATICAL ALEF..ARABIC MATHEMATICAL DAL
    Interval{ 0x1EE05, 0x1EE1F }, // Lo  [27] ARABIC MATHEMATICAL WAW..ARABIC MATHEMATICAL DOTLESS QAF
    Interval{ 0x1EE21, 0x1EE22 }, // Lo   [2] ARABIC MATHEMATICAL INITIAL BEH..ARABIC MATHEMATICAL INITIAL JEEM
    Interval{ 0x1EE24, 0x1EE24 }, // Lo       ARABIC MATHEMATICAL INITIAL HEH
    Interval{ 0x1EE27, 0x1EE27 }, // Lo       ARABIC MATHEMATICAL INITIAL HAH
    Interval{ 0x1EE29, 0x1EE32 }, // Lo  [10] ARABIC MATHEMATICAL INITIAL YEH..ARABIC MATHEMATICAL INITIAL QAF
    Interval{ 0x1EE34, 0x1EE37 }, // Lo   [4] ARABIC MATHEMATICAL INITIAL SHEEN..ARABIC MATHEMATICAL INITIAL KHAH
    Interval{ 0x1EE39, 0x1EE39 }, // Lo       ARABIC MATHEMATICAL INITIAL DAD
    Interval{ 0x1EE3B, 0x1EE3B }, // Lo       ARABIC MATHEMATICAL INITIAL GHAIN
    Interval{ 0x1EE42, 0x1EE42 }, // Lo       ARABIC MATHEMATICAL TAILED JEEM
    Interval{ 0x1EE47, 0x1EE47 }, // Lo       ARABIC MATHEMATICAL TAILED HAH
    Interval{ 0x1EE49, 0x1EE49 }, // Lo       ARABIC MATHEMATICAL TAILED YEH
    Interval{ 0x1EE4B, 0x1EE4B }, // Lo       ARABIC MATHEMATICAL TAILED LAM
    Interval{ 0x1EE4D, 0x1EE4F }, // Lo   [3] ARABIC MATHEMATICAL TAILED NOON..ARABIC MATHEMATICAL TAILED AIN
    Interval{ 0x1EE51, 0x1EE52 }, // Lo   [2] ARABIC MATHEMATICAL TAILED SAD..ARABIC MATHEMATICAL TAILED QAF
    Interval{ 0x1EE54, 0x1EE54 }, // Lo       ARABIC MATHEMATICAL TAILED SHEEN
    Interval{ 0x1EE57, 0x1EE57 }, // Lo       ARABIC MATHEMATICAL TAILED KHAH
    Interval{ 0x1EE59, 0x1EE59 }, // Lo       ARABIC MATHEMATICAL TAILED DAD
    Interval{ 0x1EE5B, 0x1EE5B }, // Lo       ARABIC MATHEMATICAL TAILED GHAIN
    Interval{ 0x1EE5D, 0x1EE5D }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS NOON
    Interval{ 0x1EE5F, 0x1EE5F }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS QAF
    Interval{ 0x1EE61, 0x1EE62 }, // Lo   [2] ARABIC MATHEMATICAL STRETCHED BEH..ARABIC MATHEMATICAL STRETCHED JEEM
    Interval{ 0x1EE64, 0x1EE64 }, // Lo       ARABIC MATHEMATICAL STRETCHED HEH
    Interval{ 0x1EE67, 0x1EE6A }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED HAH..ARABIC MATHEMATICAL STRETCHED KAF
    Interval{ 0x1EE6C, 0x1EE72 }, // Lo   [7] ARABIC MATHEMATICAL STRETCHED MEEM..ARABIC MATHEMATICAL STRETCHED QAF
    Interval{ 0x1EE74, 0x1EE77 }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED SHEEN..ARABIC MATHEMATICAL STRETCHED KHAH
    Interval{ 0x1EE79, 0x1EE7C }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED DAD..ARABIC MATHEMATICAL STRETCHED DOTLESS BEH
    Interval{ 0x1EE7E, 0x1EE7E }, // Lo       ARABIC MATHEMATICAL STRETCHED DOTLESS FEH
    Interval{ 0x1EE80, 0x1EE89 }, // Lo  [10] ARABIC MATHEMATICAL LOOPED ALEF..ARABIC MATHEMATICAL LOOPED YEH
    Interval{ 0x1EE8B, 0x1EE9B }, // Lo  [17] ARABIC MATHEMATICAL LOOPED LAM..ARABIC MATHEMATICAL LOOPED GHAIN
    Interval{ 0x1EEA1, 0x1EEA3 }, // Lo   [3] ARABIC MATHEMATICAL DOUBLE-STRUCK BEH..ARABIC MATHEMATICAL DOUBLE-STRUCK DAL
    Interval{ 0x1EEA5, 0x1EEA9 }, // Lo   [5] ARABIC MATHEMATICAL DOUBLE-STRUCK WAW..ARABIC MATHEMATICAL DOUBLE-STRUCK YEH
    Interval{ 0x1EEAB, 0x1EEBB }, // Lo  [17] ARABIC MATHEMATICAL DOUBLE-STRUCK LAM..ARABIC MATHEMATICAL DOUBLE-STRUCK GHAIN
    Interval{ 0x1F130, 0x1F149 }, // So  [26] SQUARED LATIN CAPITAL LETTER A..SQUARED LATIN CAPITAL LETTER Z
    Interval{ 0x1F150, 0x1F169 }, // So  [26] NEGATIVE CIRCLED LATIN CAPITAL LETTER A..NEGATIVE CIRCLED LATIN CAPITAL LETTER Z
    Interval{ 0x1F170, 0x1F189 }, // So  [26] NEGATIVE SQUARED LATIN CAPITAL LETTER A..NEGATIVE SQUARED LATIN CAPITAL LETTER Z
    Interval{ 0x20000, 0x2A6DF }, // Lo [42720] CJK UNIFIED IDEOGRAPH-20000..CJK UNIFIED IDEOGRAPH-2A6DF
    Interval{ 0x2A700, 0x2B81D }, // Lo [4382] CJK UNIFIED IDEOGRAPH-2A700..CJK UNIFIED IDEOGRAPH-2B81D
    Interval{ 0x2B820, 0x2CEAD }, // Lo [5774] CJK UNIFIED IDEOGRAPH-2B820..CJK UNIFIED IDEOGRAPH-2CEAD
    Interval{ 0x2CEB0, 0x2EBE0 }, // Lo [7473] CJK UNIFIED IDEOGRAPH-2CEB0..CJK UNIFIED IDEOGRAPH-2EBE0
    Interval{ 0x2EBF0, 0x2EE5D }, // Lo [622] CJK UNIFIED IDEOGRAPH-2EBF0..CJK UNIFIED IDEOGRAPH-2EE5D
    Interval{ 0x2F800, 0x2FA1D }, // Lo [542] CJK COMPATIBILITY IDEOGRAPH-2F800..CJK COMPATIBILITY IDEOGRAPH-2FA1D
    Interval{ 0x30000, 0x3134A }, // Lo [4939] CJK UNIFIED IDEOGRAPH-30000..CJK UNIFIED IDEOGRAPH-3134A
    Interval{ 0x31350, 0x33479 }, // Lo [8490] CJK UNIFIED IDEOGRAPH-31350..CJK UNIFIED IDEOGRAPH-33479
}; // }}}
auto static const Case_Ignorable = std::array<Interval, 518>{ // {{{
    Interval{ 0x0027, 0x0027 }, // Po       APOSTROPHE
    Interval{ 0x002E, 0x002E }, // Po       FULL STOP
    Interval{ 0x003A, 0x003A }, // Po       COLON
    Interval{ 0x005E, 0x005E }, // Sk       CIRCUMFLEX ACCENT
    Interval{ 0x0060, 0x0060 }, // Sk       GRAVE ACCENT
    Interval{ 0x00A8, 0x00A8 }, // Sk       DIAERESIS
    Interval{ 0x00AD, 0x00AD }, // Cf       SOFT HYPHEN
    Interval{ 0x00AF, 0x00AF }, // Sk       MACRON
    Interval{ 0x00B4, 0x00B4 }, // Sk       ACUTE ACCENT
    Interval{ 0x00B7, 0x00B7 }, // Po       MIDDLE DOT
    Interval{ 0x00B8, 0x00B8 }, // Sk       CEDILLA
    Interval{ 0x02B0, 0x02C1 }, // Lm  [18] MODIFIER LETTER SMALL H..MODIFIER LETTER REVERSED GLOTTAL STOP
    Interval{ 0x02C2, 0x02C5 }, // Sk   [4] MODIFIER LETTER LEFT ARROWHEAD..MODIFIER LETTER DOWN ARROWHEAD
    Interval{ 0x02C6, 0x02D1 }, // Lm  [12] MODIFIER LETTER CIRCUMFLEX ACCENT..MODIFIER LETTER HALF TRIANGULAR COLON
    Interval{ 0x02D2, 0x02DF }, // Sk  [14] MODIFIER LETTER CENTRED RIGHT HALF RING..MODIFIER LETTER CROSS ACCENT
    Interval{ 0x02E0, 0x02E4 }, // Lm   [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    Interval{ 0x02E5, 0x02EB }, // Sk   [7] MODIFIER LETTER EXTRA-HIGH TONE BAR..MODIFIER LETTER YANG DEPARTING TONE MARK
    Interval{ 0x02EC, 0x02EC }, // Lm       MODIFIER LETTER VOICING
    Interval{ 0x02ED, 0x02ED }, // Sk       MODIFIER LETTER UNASPIRATED
    Interval{ 0x02EE, 0x02EE }, // Lm       MODIFIER LETTER DOUBLE APOSTROPHE
    Interval{ 0x02EF, 0x02FF }, // Sk  [17] MODIFIER LETTER LOW DOWN ARROWHEAD..MODIFIER LETTER LOW LEFT ARROW
    Interval{ 0x0300, 0x036F }, // Mn [112] COMBINING GRAVE ACCENT..COMBINING LATIN SMALL LETTER X
    Interval{ 0x0374, 0x0374 }, // Lm       GREEK NUMERAL SIGN
    Interval{ 0x0375, 0x0375 }, // Sk       GREEK LOWER NUMERAL SIGN
    Interval{ 0x037A, 0x037A }, // Lm       GREEK YPOGEGRAMMENI
    Interval{ 0x0384, 0x0385 }, // Sk   [2] GREEK TONOS..GREEK DIALYTIKA TONOS
    Interval{ 0x0387, 0x0387 }, // Po       GREEK ANO TELEIA
    Interval{ 0x0483, 0x0487 }, // Mn   [5] COMBINING CYRILLIC TITLO..COMBINING CYRILLIC POKRYTIE
    Interval{ 0x0488, 0x0489 }, // Me   [2] COMBINING CYRILLIC HUNDRED THOUSANDS SIGN..COMBINING CYRILLIC MILLIONS SIGN
    Interval{ 0x0559, 0x0559 }, // Lm       ARMENIAN MODIFIER LETTER LEFT HALF RING
    Interval{ 0x055F, 0x055F }, // Po       ARMENIAN ABBREVIATION MARK
    Interval{ 0x0591, 0x05BD }, // Mn  [45] HEBREW ACCENT ETNAHTA..HEBREW POINT METEG
    Interval{ 0x05BF, 0x05BF }, // Mn       HEBREW POINT RAFE
    Interval{ 0x05C1, 0x05C2 }, // Mn   [2] HEBREW POINT SHIN DOT..HEBREW POINT SIN DOT
    Interval{ 0x05C4, 0x05C5 }, // Mn   [2] HEBREW MARK UPPER DOT..HEBREW MARK LOWER DOT
    Interval{ 0x05C7, 0x05C7 }, // Mn       HEBREW POINT QAMATS QATAN
    Interval{ 0x05F4, 0x05F4 }, // Po       HEBREW PUNCTUATION GERSHAYIM
    Interval{ 0x0600, 0x0605 }, // Cf   [6] ARABIC NUMBER SIGN..ARABIC NUMBER MARK ABOVE
    Interval{ 0x0610, 0x061A }, // Mn  [11] ARABIC SIGN SALLALLAHOU ALAYHE WASSALLAM..ARABIC SMALL KASRA
    Interval{ 0x061C, 0x061C }, // Cf       ARABIC LETTER MARK
    Interval{ 0x0640, 0x0640 }, // Lm       ARABIC TATWEEL
    Interval{ 0x064B, 0x065F }, // Mn  [21] ARABIC FATHATAN..ARABIC WAVY HAMZA BELOW
    Interval{ 0x0670, 0x0670 }, // Mn       ARABIC LETTER SUPERSCRIPT ALEF
    Interval{ 0x06D6, 0x06DC }, // Mn   [7] ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA..ARABIC SMALL HIGH SEEN
    Interval{ 0x06DD, 0x06DD }, // Cf       ARABIC END OF AYAH
    Interval{ 0x06DF, 0x06E4 }, // Mn   [6] ARABIC SMALL HIGH ROUNDED ZERO..ARABIC SMALL HIGH MADDA
    Interval{ 0x06E5, 0x06E6 }, // Lm   [2] ARABIC SMALL WAW..ARABIC SMALL YEH
    Interval{ 0x06E7, 0x06E8 }, // Mn   [2] ARABIC SMALL HIGH YEH..ARABIC SMALL HIGH NOON
    Interval{ 0x06EA, 0x06ED }, // Mn   [4] ARABIC EMPTY CENTRE LOW STOP..ARABIC SMALL LOW MEEM
    Interval{ 0x070F, 0x070F }, // Cf       SYRIAC ABBREVIATION MARK
    Interval{ 0x0711, 0x0711 }, // Mn       SYRIAC LETTER SUPERSCRIPT ALAPH
    Interval{ 0x0730, 0x074A }, // Mn  [27] SYRIAC PTHAHA ABOVE..SYRIAC BARREKH
    Interval{ 0x07A6, 0x07B0 }, // Mn  [11] THAANA ABAFILI..THAANA SUKUN
    Interval{ 0x07EB, 0x07F3 }, // Mn   [9] NKO COMBINING SHORT HIGH TONE..NKO COMBINING DOUBLE DOT ABOVE
    Interval{ 0x07F4, 0x07F5 }, // Lm   [2] NKO HIGH TONE APOSTROPHE..NKO LOW TONE APOSTROPHE
    Interval{ 0x07FA, 0x07FA }, // Lm       NKO LAJANYALAN
    Interval{ 0x07FD, 0x07FD }, // Mn       NKO DANTAYALAN
    Interval{ 0x0816, 0x0819 }, // Mn   [4] SAMARITAN MARK IN..SAMARITAN MARK DAGESH
    Interval{ 0x081A, 0x081A }, // Lm       SAMARITAN MODIFIER LETTER EPENTHETIC YUT
    Interval{ 0x081B, 0x0823 }, // Mn   [9] SAMARITAN MARK EPENTHETIC YUT..SAMARITAN VOWEL SIGN A
    Interval{ 0x0824, 0x0824 }, // Lm       SAMARITAN MODIFIER LETTER SHORT A
    Interval{ 0x0825, 0x0827 }, // Mn   [3] SAMARITAN VOWEL SIGN SHORT A..SAMARITAN VOWEL SIGN U
    Interval{ 0x0828, 0x0828 }, // Lm       SAMARITAN MODIFIER LETTER I
    Interval{ 0x0829, 0x082D }, // Mn   [5] SAMARITAN VOWEL SIGN LONG I..SAMARITAN MARK NEQUDAA
    Interval{ 0x0859, 0x085B }, // Mn   [3] MANDAIC AFFRICATION MARK..MANDAIC GEMINATION MARK
    Interval{ 0x0888, 0x0888 }, // Sk       ARABIC RAISED ROUND DOT
    Interval{ 0x0890, 0x0891 }, // Cf   [2] ARABIC POUND MARK ABOVE..ARABIC PIASTRE MARK ABOVE
    Interval{ 0x0897, 0x089F }, // Mn   [9] ARABIC PEPET..ARABIC HALF MADDA OVER MADDA
    Interval{ 0x08C9, 0x08C9 }, // Lm       ARABIC SMALL FARSI YEH
    Interval{ 0x08CA, 0x08E1 }, // Mn  [24] ARABIC SMALL HIGH FARSI YEH..ARABIC SMALL HIGH SIGN SAFHA
    Interval{ 0x08E2, 0x08E2 }, // Cf       ARABIC DISPUTED END OF AYAH
    Interval{ 0x08E3, 0x0902 }, // Mn  [32] ARABIC TURNED DAMMA BELOW..DEVANAGARI SIGN ANUSVARA
    Interval{ 0x093A, 0x093A }, // Mn       DEVANAGARI VOWEL SIGN OE
    Interval{ 0x093C, 0x093C }, // Mn       DEVANAGARI SIGN NUKTA
    Interval{ 0x0941, 0x0948 }, // Mn   [8] DEVANAGARI VOWEL SIGN U..DEVANAGARI VOWEL SIGN AI
    Interval{ 0x094D, 0x094D }, // Mn       DEVANAGARI SIGN VIRAMA
    Interval{ 0x0951, 0x0957 }, // Mn   [7] DEVANAGARI STRESS SIGN UDATTA..DEVANAGARI VOWEL SIGN UUE
    Interval{ 0x0962, 0x0963 }, // Mn   [2] DEVANAGARI VOWEL SIGN VOCALIC L..DEVANAGARI VOWEL SIGN VOCALIC LL
    Interval{ 0x0971, 0x0971 }, // Lm       DEVANAGARI SIGN HIGH SPACING DOT
    Interval{ 0x0981, 0x0981 }, // Mn       BENGALI SIGN CANDRABINDU
    Interval{ 0x09BC, 0x09BC }, // Mn       BENGALI SIGN NUKTA
    Interval{ 0x09C1, 0x09C4 }, // Mn   [4] BENGALI VOWEL SIGN U..BENGALI VOWEL SIGN VOCALIC RR
    Interval{ 0x09CD, 0x09CD }, // Mn       BENGALI SIGN VIRAMA
    Interval{ 0x09E2, 0x09E3 }, // Mn   [2] BENGALI VOWEL SIGN VOCALIC L..BENGALI VOWEL SIGN VOCALIC LL
    Interval{ 0x09FE, 0x09FE }, // Mn       BENGALI SANDHI MARK
    Interval{ 0x0A01, 0x0A02 }, // Mn   [2] GURMUKHI SIGN ADAK BINDI..GURMUKHI SIGN BINDI
    Interval{ 0x0A3C, 0x0A3C }, // Mn       GURMUKHI SIGN NUKTA
    Interval{ 0x0A41, 0x0A42 }, // Mn   [2] GURMUKHI VOWEL SIGN U..GURMUKHI VOWEL SIGN UU
    Interval{ 0x0A47, 0x0A48 }, // Mn   [2] GURMUKHI VOWEL SIGN EE..GURMUKHI VOWEL SIGN AI
    Interval{ 0x0A4B, 0x0A4D }, // Mn   [3] GURMUKHI VOWEL SIGN OO..GURMUKHI SIGN VIRAMA
    Interval{ 0x0A51, 0x0A51 }, // Mn       GURMUKHI SIGN UDAAT
    Interval{ 0x0A70, 0x0A71 }, // Mn   [2] GURMUKHI TIPPI..GURMUKHI ADDAK
    Interval{ 0x0A75, 0x0A75 }, // Mn       GURMUKHI SIGN YAKASH
    Interval{ 0x0A81, 0x0A82 }, // Mn   [2] GUJARATI SIGN CANDRABINDU..GUJARATI SIGN ANUSVARA
    Interval{ 0x0ABC, 0x0ABC }, // Mn       GUJARATI SIGN NUKTA
    Interval{ 0x0AC1, 0x0AC5 }, // Mn   [5] GUJARATI VOWEL SIGN U..GUJARATI VOWEL SIGN CANDRA E
    Interval{ 0x0AC7, 0x0AC8 }, // Mn   [2] GUJARATI VOWEL SIGN E..GUJARATI VOWEL SIGN AI
    Interval{ 0x0ACD, 0x0ACD }, // Mn       GUJARATI SIGN VIRAMA
    Interval{ 0x0AE2, 0x0AE3 }, // Mn   [2] GUJARATI VOWEL SIGN VOCALIC L..GUJARATI VOWEL SIGN VOCALIC LL
    Interval{ 0x0AFA, 0x0AFF }, // Mn   [6] GUJARATI SIGN SUKUN..GUJARATI SIGN TWO-CIRCLE NUKTA ABOVE
    Interval{ 0x0B01, 0x0B01 }, // Mn       ORIYA SIGN CANDRABINDU
    Interval{ 0x0B3C, 0x0B3C }, // Mn       ORIYA SIGN NUKTA
    Interval{ 0x0B3F, 0x0B3F }, // Mn       ORIYA VOWEL SIGN I
    Interval{ 0x0B41, 0x0B44 }, // Mn   [4] ORIYA VOWEL SIGN U..ORIYA VOWEL SIGN VOCALIC RR
    Interval{ 0x0B4D, 0x0B4D }, // Mn       ORIYA SIGN VIRAMA
    Interval{ 0x0B55, 0x0B56 }, // Mn   [2] ORIYA SIGN OVERLINE..ORIYA AI LENGTH MARK
    Interval{ 0x0B62, 0x0B63 }, // Mn   [2] ORIYA VOWEL SIGN VOCALIC L..ORIYA VOWEL SIGN VOCALIC LL
    Interval{ 0x0B82, 0x0B82 }, // Mn       TAMIL SIGN ANUSVARA
    Interval{ 0x0BC0, 0x0BC0 }, // Mn       TAMIL VOWEL SIGN II
    Interval{ 0x0BCD, 0x0BCD }, // Mn       TAMIL SIGN VIRAMA
    Interval{ 0x0C00, 0x0C00 }, // Mn       TELUGU SIGN COMBINING CANDRABINDU ABOVE
    Interval{ 0x0C04, 0x0C04 }, // Mn       TELUGU SIGN COMBINING ANUSVARA ABOVE
    Interval{ 0x0C3C, 0x0C3C }, // Mn       TELUGU SIGN NUKTA
    Interval{ 0x0C3E, 0x0C40 }, // Mn   [3] TELUGU VOWEL SIGN AA..TELUGU VOWEL SIGN II
    Interval{ 0x0C46, 0x0C48 }, // Mn   [3] TELUGU VOWEL SIGN E..TELUGU VOWEL SIGN AI
    Interval{ 0x0C4A, 0x0C4D }, // Mn   [4] TELUGU VOWEL SIGN O..TELUGU SIGN VIRAMA
    Interval{ 0x0C55, 0x0C56 }, // Mn   [2] TELUGU LENGTH MARK..TELUGU AI LENGTH MARK
    Interval{ 0x0C62, 0x0C63 }, // Mn   [2] TELUGU VOWEL SIGN VOCALIC L..TELUGU VOWEL SIGN VOCALIC LL
    Interval{ 0x0C81, 0x0C81 }, // Mn       KANNADA SIGN CANDRABINDU
    Interval{ 0x0CBC, 0x0CBC }, // Mn       KANNADA SIGN NUKTA
    Interval{ 0x0CBF, 0x0CBF }, // Mn       KANNADA VOWEL SIGN I
    Interval{ 0x0CC6, 0x0CC6 }, // Mn       KANNADA VOWEL SIGN E
    Interval{ 0x0CCC, 0x0CCD }, // Mn   [2] KANNADA VOWEL SIGN AU..KANNADA SIGN VIRAMA
    Interval{ 0x0CE2, 0x0CE3 }, // Mn   [2] KANNADA VOWEL SIGN VOCALIC L..KANNADA VOWEL SIGN VOCALIC LL
    Interval{ 0x0D00, 0x0D01 }, // Mn   [2] MALAYALAM SIGN COMBINING ANUSVARA ABOVE..MALAYALAM SIGN CANDRABINDU
    Interval{ 0x0D3B, 0x0D3C }, // Mn   [2] MALAYALAM SIGN VERTICAL BAR VIRAMA..MALAYALAM SIGN CIRCULAR VIRAMA
    Interval{ 0x0D41, 0x0D44 }, // Mn   [4] MALAYALAM VOWEL SIGN U..MALAYALAM VOWEL SIGN VOCALIC RR
    Interval{ 0x0D4D, 0x0D4D }, // Mn       MALAYALAM SIGN VIRAMA
    Interval{ 0x0D62, 0x0D63 }, // Mn   [2] MALAYALAM VOWEL SIGN VOCALIC L..MALAYALAM VOWEL SIGN VOCALIC LL
    Interval{ 0x0D81, 0x0D81 }, // Mn       SINHALA SIGN CANDRABINDU
    Interval{ 0x0DCA, 0x0DCA }, // Mn       SINHALA SIGN AL-LAKUNA
    Interval{ 0x0DD2, 0x0DD4 }, // Mn   [3] SINHALA VOWEL SIGN KETTI IS-PILLA..SINHALA VOWEL SIGN KETTI PAA-PILLA
    Interval{ 0x0DD6, 0x0DD6 }, // Mn       SINHALA VOWEL SIGN DIGA PAA-PILLA
    Interval{ 0x0E31, 0x0E31 }, // Mn       THAI CHARACTER MAI HAN-AKAT
    Interval{ 0x0E34, 0x0E3A }, // Mn   [7] THAI CHARACTER SARA I..THAI CHARACTER PHINTHU
    Interval{ 0x0E46, 0x0E46 }, // Lm       THAI CHARACTER MAIYAMOK
    Interval{ 0x0E47, 0x0E4E }, // Mn   [8] THAI CHARACTER MAITAIKHU..THAI CHARACTER YAMAKKAN
    Interval{ 0x0EB1, 0x0EB1 }, // Mn       LAO VOWEL SIGN MAI KAN
    Interval{ 0x0EB4, 0x0EBC }, // Mn   [9] LAO VOWEL SIGN I..LAO SEMIVOWEL SIGN LO
    Interval{ 0x0EC6, 0x0EC6 }, // Lm       LAO KO LA
    Interval{ 0x0EC8, 0x0ECE }, // Mn   [7] LAO TONE MAI EK..LAO YAMAKKAN
    Interval{ 0x0F18, 0x0F19 }, // Mn   [2] TIBETAN ASTROLOGICAL SIGN -KHYUD PA..TIBETAN ASTROLOGICAL SIGN SDONG TSHUGS
    Interval{ 0x0F35, 0x0F35 }, // Mn       TIBETAN MARK NGAS BZUNG NYI ZLA
    Interval{ 0x0F37, 0x0F37 }, // Mn       TIBETAN MARK NGAS BZUNG SGOR RTAGS
    Interval{ 0x0F39, 0x0F39 }, // Mn       TIBETAN MARK TSA -PHRU
    Interval{ 0x0F71, 0x0F7E }, // Mn  [14] TIBETAN VOWEL SIGN AA..TIBETAN SIGN RJES SU NGA RO
    Interval{ 0x0F80, 0x0F84 }, // Mn   [5] TIBETAN VOWEL SIGN REVERSED I..TIBETAN MARK HALANTA
    Interval{ 0x0F86, 0x0F87 }, // Mn   [2] TIBETAN SIGN LCI RTAGS..TIBETAN SIGN YANG RTAGS
    Interval{ 0x0F8D, 0x0F97 }, // Mn  [11] TIBETAN SUBJOINED SIGN LCE TSA CAN..TIBETAN SUBJOINED LETTER JA
    Interval{ 0x0F99, 0x0FBC }, // Mn  [36] TIBETAN SUBJOINED LETTER NYA..TIBETAN SUBJOINED LETTER FIXED-FORM RA
    Interval{ 0x0FC6, 0x0FC6 }, // Mn       TIBETAN SYMBOL PADMA GDAN
    Interval{ 0x102D, 0x1030 }, // Mn   [4] MYANMAR VOWEL SIGN I..MYANMAR VOWEL SIGN UU
    Interval{ 0x1032, 0x1037 }, // Mn   [6] MYANMAR VOWEL SIGN AI..MYANMAR SIGN DOT BELOW
    Interval{ 0x1039, 0x103A }, // Mn   [2] MYANMAR SIGN VIRAMA..MYANMAR SIGN ASAT
    Interval{ 0x103D, 0x103E }, // Mn   [2] MYANMAR CONSONANT SIGN MEDIAL WA..MYANMAR CONSONANT SIGN MEDIAL HA
    Interval{ 0x1058, 0x1059 }, // Mn   [2] MYANMAR VOWEL SIGN VOCALIC L..MYANMAR VOWEL SIGN VOCALIC LL
    Interval{ 0x105E, 0x1060 }, // Mn   [3] MYANMAR CONSONANT SIGN MON MEDIAL NA..MYANMAR CONSONANT SIGN MON MEDIAL LA
    Interval{ 0x1071, 0x1074 }, // Mn   [4] MYANMAR VOWEL SIGN GEBA KAREN I..MYANMAR VOWEL SIGN KAYAH EE
    Interval{ 0x1082, 0x1082 }, // Mn       MYANMAR CONSONANT SIGN SHAN MEDIAL WA
    Interval{ 0x1085, 0x1086 }, // Mn   [2] MYANMAR VOWEL SIGN SHAN E ABOVE..MYANMAR VOWEL SIGN SHAN FINAL Y
    Interval{ 0x108D, 0x108D }, // Mn       MYANMAR SIGN SHAN COUNCIL EMPHATIC TONE
    Interval{ 0x109D, 0x109D }, // Mn       MYANMAR VOWEL SIGN AITON AI
    Interval{ 0x10FC, 0x10FC }, // Lm       MODIFIER LETTER GEORGIAN NAR
    Interval{ 0x135D, 0x135F }, // Mn   [3] ETHIOPIC COMBINING GEMINATION AND VOWEL LENGTH MARK..ETHIOPIC COMBINING GEMINATION MARK
    Interval{ 0x1712, 0x1714 }, // Mn   [3] TAGALOG VOWEL SIGN I..TAGALOG SIGN VIRAMA
    Interval{ 0x1732, 0x1733 }, // Mn   [2] HANUNOO VOWEL SIGN I..HANUNOO VOWEL SIGN U
    Interval{ 0x1752, 0x1753 }, // Mn   [2] BUHID VOWEL SIGN I..BUHID VOWEL SIGN U
    Interval{ 0x1772, 0x1773 }, // Mn   [2] TAGBANWA VOWEL SIGN I..TAGBANWA VOWEL SIGN U
    Interval{ 0x17B4, 0x17B5 }, // Mn   [2] KHMER VOWEL INHERENT AQ..KHMER VOWEL INHERENT AA
    Interval{ 0x17B7, 0x17BD }, // Mn   [7] KHMER VOWEL SIGN I..KHMER VOWEL SIGN UA
    Interval{ 0x17C6, 0x17C6 }, // Mn       KHMER SIGN NIKAHIT
    Interval{ 0x17C9, 0x17D3 }, // Mn  [11] KHMER SIGN MUUSIKATOAN..KHMER SIGN BATHAMASAT
    Interval{ 0x17D7, 0x17D7 }, // Lm       KHMER SIGN LEK TOO
    Interval{ 0x17DD, 0x17DD }, // Mn       KHMER SIGN ATTHACAN
    Interval{ 0x180B, 0x180D }, // Mn   [3] MONGOLIAN FREE VARIATION SELECTOR ONE..MONGOLIAN FREE VARIATION SELECTOR THREE
    Interval{ 0x180E, 0x180E }, // Cf       MONGOLIAN VOWEL SEPARATOR
    Interval{ 0x180F, 0x180F }, // Mn       MONGOLIAN FREE VARIATION SELECTOR FOUR
    Interval{ 0x1843, 0x1843 }, // Lm       MONGOLIAN LETTER TODO LONG VOWEL SIGN
    Interval{ 0x1885, 0x1886 }, // Mn   [2] MONGOLIAN LETTER ALI GALI BALUDA..MONGOLIAN LETTER ALI GALI THREE BALUDA
    Interval{ 0x18A9, 0x18A9 }, // Mn       MONGOLIAN LETTER ALI GALI DAGALGA
    Interval{ 0x1920, 0x1922 }, // Mn   [3] LIMBU VOWEL SIGN A..LIMBU VOWEL SIGN U
    Interval{ 0x1927, 0x1928 }, // Mn   [2] LIMBU VOWEL SIGN E..LIMBU VOWEL SIGN O
    Interval{ 0x1932, 0x1932 }, // Mn       LIMBU SMALL LETTER ANUSVARA
    Interval{ 0x1939, 0x193B }, // Mn   [3] LIMBU SIGN MUKPHRENG..LIMBU SIGN SA-I
    Interval{ 0x1A17, 0x1A18 }, // Mn   [2] BUGINESE VOWEL SIGN I..BUGINESE VOWEL SIGN U
    Interval{ 0x1A1B, 0x1A1B }, // Mn       BUGINESE VOWEL SIGN AE
    Interval{ 0x1A56, 0x1A56 }, // Mn       TAI THAM CONSONANT SIGN MEDIAL LA
    Interval{ 0x1A58, 0x1A5E }, // Mn   [7] TAI THAM SIGN MAI KANG LAI..TAI THAM CONSONANT SIGN SA
    Interval{ 0x1A60, 0x1A60 }, // Mn       TAI THAM SIGN SAKOT
    Interval{ 0x1A62, 0x1A62 }, // Mn       TAI THAM VOWEL SIGN MAI SAT
    Interval{ 0x1A65, 0x1A6C }, // Mn   [8] TAI THAM VOWEL SIGN I..TAI THAM VOWEL SIGN OA BELOW
    Interval{ 0x1A73, 0x1A7C }, // Mn  [10] TAI THAM VOWEL SIGN OA ABOVE..TAI THAM SIGN KHUEN-LUE KARAN
    Interval{ 0x1A7F, 0x1A7F }, // Mn       TAI THAM COMBINING CRYPTOGRAMMIC DOT
    Interval{ 0x1AA7, 0x1AA7 }, // Lm       TAI THAM SIGN MAI YAMOK
    Interval{ 0x1AB0, 0x1ABD }, // Mn  [14] COMBINING DOUBLED CIRCUMFLEX ACCENT..COMBINING PARENTHESES BELOW
    Interval{ 0x1ABE, 0x1ABE }, // Me       COMBINING PARENTHESES OVERLAY
    Interval{ 0x1ABF, 0x1ADD }, // Mn  [31] COMBINING LATIN SMALL LETTER W BELOW..COMBINING DOT-AND-RING BELOW
    Interval{ 0x1AE0, 0x1AEB }, // Mn  [12] COMBINING LEFT TACK ABOVE..COMBINING DOUBLE RIGHTWARDS ARROW ABOVE
    Interval{ 0x1B00, 0x1B03 }, // Mn   [4] BALINESE SIGN ULU RICEM..BALINESE SIGN SURANG
    Interval{ 0x1B34, 0x1B34 }, // Mn       BALINESE SIGN REREKAN
    Interval{ 0x1B36, 0x1B3A }, // Mn   [5] BALINESE VOWEL SIGN ULU..BALINESE VOWEL SIGN RA REPA
    Interval{ 0x1B3C, 0x1B3C }, // Mn       BALINESE VOWEL SIGN LA LENGA
    Interval{ 0x1B42, 0x1B42 }, // Mn       BALINESE VOWEL SIGN PEPET
    Interval{ 0x1B6B, 0x1B73 }, // Mn   [9] BALINESE MUSICAL SYMBOL COMBINING TEGEH..BALINESE MUSICAL SYMBOL COMBINING GONG
    Interval{ 0x1B80, 0x1B81 }, // Mn   [2] SUNDANESE SIGN PANYECEK..SUNDANESE SIGN PANGLAYAR
    Interval{ 0x1BA2, 0x1BA5 }, // Mn   [4] SUNDANESE CONSONANT SIGN PANYAKRA..SUNDANESE VOWEL SIGN PANYUKU
    Interval{ 0x1BA8, 0x1BA9 }, // Mn   [2] SUNDANESE VOWEL SIGN PAMEPET..SUNDANESE VOWEL SIGN PANEULEUNG
    Interval{ 0x1BAB, 0x1BAD }, // Mn   [3] SUNDANESE SIGN VIRAMA..SUNDANESE CONSONANT SIGN PASANGAN WA
    Interval{ 0x1BE6, 0x1BE6 }, // Mn       BATAK SIGN TOMPI
    Interval{ 0x1BE8, 0x1BE9 }, // Mn   [2] BATAK VOWEL SIGN PAKPAK E..BATAK VOWEL SIGN EE
    Interval{ 0x1BED, 0x1BED }, // Mn       BATAK VOWEL SIGN KARO O
    Interval{ 0x1BEF, 0x1BF1 }, // Mn   [3] BATAK VOWEL SIGN U FOR SIMALUNGUN SA..BATAK CONSONANT SIGN H
    Interval{ 0x1C2C, 0x1C33 }, // Mn   [8] LEPCHA VOWEL SIGN E..LEPCHA CONSONANT SIGN T
    Interval{ 0x1C36, 0x1C37 }, // Mn   [2] LEPCHA SIGN RAN..LEPCHA SIGN NUKTA
    Interval{ 0x1C78, 0x1C7D }, // Lm   [6] OL CHIKI MU TTUDDAG..OL CHIKI AHAD
    Interval{ 0x1CD0, 0x1CD2 }, // Mn   [3] VEDIC TONE KARSHANA..VEDIC TONE PRENKHA
    Interval{ 0x1CD4, 0x1CE0 }, // Mn  [13] VEDIC SIGN YAJURVEDIC MIDLINE SVARITA..VEDIC TONE RIGVEDIC KASHMIRI INDEPENDENT SVARITA
    Interval{ 0x1CE2, 0x1CE8 }, // Mn   [7] VEDIC SIGN VISARGA SVARITA..VEDIC SIGN VISARGA ANUDATTA WITH TAIL
    Interval{ 0x1CED, 0x1CED }, // Mn       VEDIC SIGN TIRYAK
    Interval{ 0x1CF4, 0x1CF4 }, // Mn       VEDIC TONE CANDRA ABOVE
    Interval{ 0x1CF8, 0x1CF9 }, // Mn   [2] VEDIC TONE RING ABOVE..VEDIC TONE DOUBLE RING ABOVE
    Interval{ 0x1D2C, 0x1D6A }, // Lm  [63] MODIFIER LETTER CAPITAL A..GREEK SUBSCRIPT SMALL LETTER CHI
    Interval{ 0x1D78, 0x1D78 }, // Lm       MODIFIER LETTER CYRILLIC EN
    Interval{ 0x1D9B, 0x1DBF }, // Lm  [37] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL THETA
    Interval{ 0x1DC0, 0x1DFF }, // Mn  [64] COMBINING DOTTED GRAVE ACCENT..COMBINING RIGHT ARROWHEAD AND DOWN ARROWHEAD BELOW
    Interval{ 0x1FBD, 0x1FBD }, // Sk       GREEK KORONIS
    Interval{ 0x1FBF, 0x1FC1 }, // Sk   [3] GREEK PSILI..GREEK DIALYTIKA AND PERISPOMENI
    Interval{ 0x1FCD, 0x1FCF }, // Sk   [3] GREEK PSILI AND VARIA..GREEK PSILI AND PERISPOMENI
    Interval{ 0x1FDD, 0x1FDF }, // Sk   [3] GREEK DASIA AND VARIA..GREEK DASIA AND PERISPOMENI
    Interval{ 0x1FED, 0x1FEF }, // Sk   [3] GREEK DIALYTIKA AND VARIA..GREEK VARIA
    Interval{ 0x1FFD, 0x1FFE }, // Sk   [2] GREEK OXIA..GREEK DASIA
    Interval{ 0x200B, 0x200F }, // Cf   [5] ZERO WIDTH SPACE..RIGHT-TO-LEFT MARK
    Interval{ 0x2018, 0x2018 }, // Pi       LEFT SINGLE QUOTATION MARK
    Interval{ 0x2019, 0x2019 }, // Pf       RIGHT SINGLE QUOTATION MARK
    Interval{ 0x2024, 0x2024 }, // Po       ONE DOT LEADER
    Interval{ 0x2027, 0x2027 }, // Po       HYPHENATION POINT
    Interval{ 0x202A, 0x202E }, // Cf   [5] LEFT-TO-RIGHT EMBEDDING..RIGHT-TO-LEFT OVERRIDE
    Interval{ 0x2060, 0x2064 }, // Cf   [5] WORD JOINER..INVISIBLE PLUS
    Interval{ 0x2066, 0x206F }, // Cf  [10] LEFT-TO-RIGHT ISOLATE..NOMINAL DIGIT SHAPES
    Interval{ 0x2071, 0x2071 }, // Lm       SUPERSCRIPT LATIN SMALL LETTER I
    Interval{ 0x207F, 0x207F }, // Lm       SUPERSCRIPT LATIN SMALL LETTER N
    Interval{ 0x2090, 0x209C }, // Lm  [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Interval{ 0x20D0, 0x20DC }, // Mn  [13] COMBINING LEFT HARPOON ABOVE..COMBINING FOUR DOTS ABOVE
    Interval{ 0x20DD, 0x20E0 }, // Me   [4] COMBINING ENCLOSING CIRCLE..COMBINING ENCLOSING CIRCLE BACKSLASH
    Interval{ 0x20E1, 0x20E1 }, // Mn       COMBINING LEFT RIGHT ARROW ABOVE
    Interval{ 0x20E2, 0x20E4 }, // Me   [3] COMBINING ENCLOSING SCREEN..COMBINING ENCLOSING UPWARD POINTING TRIANGLE
    Interval{ 0x20E5, 0x20F0 }, // Mn  [12] COMBINING REVERSE SOLIDUS OVERLAY..COMBINING ASTERISK ABOVE
    Interval{ 0x2C7C, 0x2C7D }, // Lm   [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    Interval{ 0x2CEF, 0x2CF1 }, // Mn   [3] COPTIC COMBINING NI ABOVE..COPTIC COMBINING SPIRITUS LENIS
    Interval{ 0x2D6F, 0x2D6F }, // Lm       TIFINAGH MODIFIER LETTER LABIALIZATION MARK
    Interval{ 0x2D7F, 0x2D7F }, // Mn       TIFINAGH CONSONANT JOINER
    Interval{ 0x2DE0, 0x2DFF }, // Mn  [32] COMBINING CYRILLIC LETTER BE..COMBINING CYRILLIC LETTER IOTIFIED BIG YUS
    Interval{ 0x2E2F, 0x2E2F }, // Lm       VERTICAL TILDE
    Interval{ 0x3005, 0x3005 }, // Lm       IDEOGRAPHIC ITERATION MARK
    Interval{ 0x302A, 0x302D }, // Mn   [4] IDEOGRAPHIC LEVEL TONE MARK..IDEOGRAPHIC ENTERING TONE MARK
    Interval{ 0x3031, 0x3035 }, // Lm   [5] VERTICAL KANA REPEAT MARK..VERTICAL KANA REPEAT MARK LOWER HALF
    Interval{ 0x303B, 0x303B }, // Lm       VERTICAL IDEOGRAPHIC ITERATION MARK
    Interval{ 0x3099, 0x309A }, // Mn   [2] COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK..COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Interval{ 0x309B, 0x309C }, // Sk   [2] KATAKANA-HIRAGANA VOICED SOUND MARK..KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Interval{ 0x309D, 0x309E }, // Lm   [2] HIRAGANA ITERATION MARK..HIRAGANA VOICED ITERATION MARK
    Interval{ 0x30FC, 0x30FE }, // Lm   [3] KATAKANA-HIRAGANA PROLONGED SOUND MARK..KATAKANA VOICED ITERATION MARK
    Interval{ 0xA015, 0xA015 }, // Lm       YI SYLLABLE WU
    Interval{ 0xA4F8, 0xA4FD }, // Lm   [6] LISU LETTER TONE MYA TI..LISU LETTER TONE MYA JEU
    Interval{ 0xA60C, 0xA60C }, // Lm       VAI SYLLABLE LENGTHENER
    Interval{ 0xA66F, 0xA66F }, // Mn       COMBINING CYRILLIC VZMET
    Interval{ 0xA670, 0xA672 }, // Me   [3] COMBINING CYRILLIC TEN MILLIONS SIGN..COMBINING CYRILLIC THOUSAND MILLIONS SIGN
    Interval{ 0xA674, 0xA67D }, // Mn  [10] COMBINING CYRILLIC LETTER UKRAINIAN IE..COMBINING CYRILLIC PAYEROK
    Interval{ 0xA67F, 0xA67F }, // Lm       CYRILLIC PAYEROK
    Interval{ 0xA69C, 0xA69D }, // Lm   [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    Interval{ 0xA69E, 0xA69F }, // Mn   [2] COMBINING CYRILLIC LETTER EF..COMBINING CYRILLIC LETTER IOTIFIED E
    Interval{ 0xA6F0, 0xA6F1 }, // Mn   [2] BAMUM COMBINING MARK KOQNDON..BAMUM COMBINING MARK TUKWENTIS
    Interval{ 0xA700, 0xA716 }, // Sk  [23] MODIFIER LETTER CHINESE TONE YIN PING..MODIFIER LETTER EXTRA-LOW LEFT-STEM TONE BAR
    Interval{ 0xA717, 0xA71F }, // Lm   [9] MODIFIER LETTER DOT VERTICAL BAR..MODIFIER LETTER LOW INVERTED EXCLAMATION MARK
    Interval{ 0xA720, 0xA721 }, // Sk   [2] MODIFIER LETTER STRESS AND HIGH TONE..MODIFIER LETTER STRESS AND LOW TONE
    Interval{ 0xA770, 0xA770 }, // Lm       MODIFIER LETTER US
    Interval{ 0xA788, 0xA788 }, // Lm       MODIFIER LETTER LOW CIRCUMFLEX ACCENT
    Interval{ 0xA789, 0xA78A }, // Sk   [2] MODIFIER LETTER COLON..MODIFIER LETTER SHORT EQUALS SIGN
    Interval{ 0xA7F1, 0xA7F4 }, // Lm   [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    Interval{ 0xA7F8, 0xA7F9 }, // Lm   [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    Interval{ 0xA802, 0xA802 }, // Mn       SYLOTI NAGRI SIGN DVISVARA
    Interval{ 0xA806, 0xA806 }, // Mn       SYLOTI NAGRI SIGN HASANTA
    Interval{ 0xA80B, 0xA80B }, // Mn       SYLOTI NAGRI SIGN ANUSVARA
    Interval{ 0xA825, 0xA826 }, // Mn   [2] SYLOTI NAGRI VOWEL SIGN U..SYLOTI NAGRI VOWEL SIGN E
    Interval{ 0xA82C, 0xA82C }, // Mn       SYLOTI NAGRI SIGN ALTERNATE HASANTA
    Interval{ 0xA8C4, 0xA8C5 }, // Mn   [2] SAURASHTRA SIGN VIRAMA..SAURASHTRA SIGN CANDRABINDU
    Interval{ 0xA8E0, 0xA8F1 }, // Mn  [18] COMBINING DEVANAGARI DIGIT ZERO..COMBINING DEVANAGARI SIGN AVAGRAHA
    Interval{ 0xA8FF, 0xA8FF }, // Mn       DEVANAGARI VOWEL SIGN AY
    Interval{ 0xA926, 0xA92D }, // Mn   [8] KAYAH LI VOWEL UE..KAYAH LI TONE CALYA PLOPHU
    Interval{ 0xA947, 0xA951 }, // Mn  [11] REJANG VOWEL SIGN I..REJANG CONSONANT SIGN R
    Interval{ 0xA980, 0xA982 }, // Mn   [3] JAVANESE SIGN PANYANGGA..JAVANESE SIGN LAYAR
    Interval{ 0xA9B3, 0xA9B3 }, // Mn       JAVANESE SIGN CECAK TELU
    Interval{ 0xA9B6, 0xA9B9 }, // Mn   [4] JAVANESE VOWEL SIGN WULU..JAVANESE VOWEL SIGN SUKU MENDUT
    Interval{ 0xA9BC, 0xA9BD }, // Mn   [2] JAVANESE VOWEL SIGN PEPET..JAVANESE CONSONANT SIGN KERET
    Interval{ 0xA9CF, 0xA9CF }, // Lm       JAVANESE PANGRANGKEP
    Interval{ 0xA9E5, 0xA9E5 }, // Mn       MYANMAR SIGN SHAN SAW
    Interval{ 0xA9E6, 0xA9E6 }, // Lm       MYANMAR MODIFIER LETTER SHAN REDUPLICATION
    Interval{ 0xAA29, 0xAA2E }, // Mn   [6] CHAM VOWEL SIGN AA..CHAM VOWEL SIGN OE
    Interval{ 0xAA31, 0xAA32 }, // Mn   [2] CHAM VOWEL SIGN AU..CHAM VOWEL SIGN UE
    Interval{ 0xAA35, 0xAA36 }, // Mn   [2] CHAM CONSONANT SIGN LA..CHAM CONSONANT SIGN WA
    Interval{ 0xAA43, 0xAA43 }, // Mn       CHAM CONSONANT SIGN FINAL NG
    Interval{ 0xAA4C, 0xAA4C }, // Mn       CHAM CONSONANT SIGN FINAL M
    Interval{ 0xAA70, 0xAA70 }, // Lm       MYANMAR MODIFIER LETTER KHAMTI REDUPLICATION
    Interval{ 0xAA7C, 0xAA7C }, // Mn       MYANMAR SIGN TAI LAING TONE-2
    Interval{ 0xAAB0, 0xAAB0 }, // Mn       TAI VIET MAI KANG
    Interval{ 0xAAB2, 0xAAB4 }, // Mn   [3] TAI VIET VOWEL I..TAI VIET VOWEL U
    Interval{ 0xAAB7, 0xAAB8 }, // Mn   [2] TAI VIET MAI KHIT..TAI VIET VOWEL IA
    Interval{ 0xAABE, 0xAABF }, // Mn   [2] TAI VIET VOWEL AM..TAI VIET TONE MAI EK
    Interval{ 0xAAC1, 0xAAC1 }, // Mn       TAI VIET TONE MAI THO
    Interval{ 0xAADD, 0xAADD }, // Lm       TAI VIET SYMBOL SAM
    Interval{ 0xAAEC, 0xAAED }, // Mn   [2] MEETEI MAYEK VOWEL SIGN UU..MEETEI MAYEK VOWEL SIGN AAI
    Interval{ 0xAAF3, 0xAAF4 }, // Lm   [2] MEETEI MAYEK SYLLABLE REPETITION MARK..MEETEI MAYEK WORD REPETITION MARK
    Interval{ 0xAAF6, 0xAAF6 }, // Mn       MEETEI MAYEK VIRAMA
    Interval{ 0xAB5B, 0xAB5B }, // Sk       MODIFIER BREVE WITH INVERTED BREVE
    Interval{ 0xAB5C, 0xAB5F }, // Lm   [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    Interval{ 0xAB69, 0xAB69 }, // Lm       MODIFIER LETTER SMALL TURNED W
    Interval{ 0xAB6A, 0xAB6B }, // Sk   [2] MODIFIER LETTER LEFT TACK..MODIFIER LETTER RIGHT TACK
    Interval{ 0xABE5, 0xABE5 }, // Mn       MEETEI MAYEK VOWEL SIGN ANAP
    Interval{ 0xABE8, 0xABE8 }, // Mn       MEETEI MAYEK VOWEL SIGN UNAP
    Interval{ 0xABED, 0xABED }, // Mn       MEETEI MAYEK APUN IYEK
    Interval{ 0xFB1E, 0xFB1E }, // Mn       HEBREW POINT JUDEO-SPANISH VARIKA
    Interval{ 0xFBB2, 0xFBC2 }, // Sk  [17] ARABIC SYMBOL DOT ABOVE..ARABIC SYMBOL WASLA ABOVE
    Interval{ 0xFE00, 0xFE0F }, // Mn  [16] VARIATION SELECTOR-1..VARIATION SELECTOR-16
    Interval{ 0xFE13, 0xFE13 }, // Po       PRESENTATION FORM FOR VERTICAL COLON
    Interval{ 0xFE20, 0xFE2F }, // Mn  [16] COMBINING LIGATURE LEFT HALF..COMBINING CYRILLIC TITLO RIGHT HALF
    Interval{ 0xFE52, 0xFE52 }, // Po       SMALL FULL STOP
    Interval{ 0xFE55, 0xFE55 }, // Po       SMALL COLON
    Interval{ 0xFEFF, 0xFEFF }, // Cf       ZERO WIDTH NO-BREAK SPACE
    Interval{ 0xFF07, 0xFF07 }, // Po       FULLWIDTH APOSTROPHE
    Interval{ 0xFF0E, 0xFF0E }, // Po       FULLWIDTH FULL STOP
    Interval{ 0xFF1A, 0xFF1A }, // Po       FULLWIDTH COLON
    Interval{ 0xFF3E, 0xFF3E }, // Sk       FULLWIDTH CIRCUMFLEX ACCENT
    Interval{ 0xFF40, 0xFF40 }, // Sk       FULLWIDTH GRAVE ACCENT
    Interval{ 0xFF70, 0xFF70 }, // Lm       HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Interval{ 0xFF9E, 0xFF9F }, // Lm   [2] HALFWIDTH KATAKANA VOICED SOUND MARK..HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK
    Interval{ 0xFFE3, 0xFFE3 }, // Sk       FULLWIDTH MACRON
    Interval{ 0xFFF9, 0xFFFB }, // Cf   [3] INTERLINEAR ANNOTATION ANCHOR..INTERLINEAR ANNOTATION TERMINATOR
    Interval{ 0x101FD, 0x101FD }, // Mn       PHAISTOS DISC SIGN COMBINING OBLIQUE STROKE
    Interval{ 0x102E0, 0x102E0 }, // Mn       COPTIC EPACT THOUSANDS MARK
    Interval{ 0x10376, 0x1037A }, // Mn   [5] COMBINING OLD PERMIC LETTER AN..COMBINING OLD PERMIC LETTER SII
    Interval{ 0x10780, 0x10785 }, // Lm   [6] MODIFIER LETTER SMALL CAPITAL AA..MODIFIER LETTER SMALL B WITH HOOK
    Interval{ 0x10787, 0x107B0 }, // Lm  [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Interval{ 0x107B2, 0x107BA }, // Lm   [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Interval{ 0x10A01, 0x10A03 }, // Mn   [3] KHAROSHTHI VOWEL SIGN I..KHAROSHTHI VOWEL SIGN VOCALIC R
    Interval{ 0x10A05, 0x10A06 }, // Mn   [2] KHAROSHTHI VOWEL SIGN E..KHAROSHTHI VOWEL SIGN O
    Interval{ 0x10A0C, 0x10A0F }, // Mn   [4] KHAROSHTHI VOWEL LENGTH MARK..KHAROSHTHI SIGN VISARGA
    Interval{ 0x10A38, 0x10A3A }, // Mn   [3] KHAROSHTHI SIGN BAR ABOVE..KHAROSHTHI SIGN DOT BELOW
    Interval{ 0x10A3F, 0x10A3F }, // Mn       KHAROSHTHI VIRAMA
    Interval{ 0x10AE5, 0x10AE6 }, // Mn   [2] MANICHAEAN ABBREVIATION MARK ABOVE..MANICHAEAN ABBREVIATION MARK BELOW
    Interval{ 0x10D24, 0x10D27 }, // Mn   [4] HANIFI ROHINGYA SIGN HARBAHAY..HANIFI ROHINGYA SIGN TASSI
    Interval{ 0x10D4E, 0x10D4E }, // Lm       GARAY VOWEL LENGTH MARK
    Interval{ 0x10D69, 0x10D6D }, // Mn   [5] GARAY VOWEL SIGN E..GARAY CONSONANT NASALIZATION MARK
    Interval{ 0x10D6F, 0x10D6F }, // Lm       GARAY REDUPLICATION MARK
    Interval{ 0x10EAB, 0x10EAC }, // Mn   [2] YEZIDI COMBINING HAMZA MARK..YEZIDI COMBINING MADDA MARK
    Interval{ 0x10EC5, 0x10EC5 }, // Lm       ARABIC SMALL YEH BARREE WITH TWO DOTS BELOW
    Interval{ 0x10EFA, 0x10EFF }, // Mn   [6] ARABIC DOUBLE VERTICAL BAR BELOW..ARABIC SMALL LOW WORD MADDA
    Interval{ 0x10F46, 0x10F50 }, // Mn  [11] SOGDIAN COMBINING DOT BELOW..SOGDIAN COMBINING STROKE BELOW
    Interval{ 0x10F82, 0x10F85 }, // Mn   [4] OLD UYGHUR COMBINING DOT ABOVE..OLD UYGHUR COMBINING TWO DOTS BELOW
    Interval{ 0x11001, 0x11001 }, // Mn       BRAHMI SIGN ANUSVARA
    Interval{ 0x11038, 0x11046 }, // Mn  [15] BRAHMI VOWEL SIGN AA..BRAHMI VIRAMA
    Interval{ 0x11070, 0x11070 }, // Mn       BRAHMI SIGN OLD TAMIL VIRAMA
    Interval{ 0x11073, 0x11074 }, // Mn   [2] BRAHMI VOWEL SIGN OLD TAMIL SHORT E..BRAHMI VOWEL SIGN OLD TAMIL SHORT O
    Interval{ 0x1107F, 0x11081 }, // Mn   [3] BRAHMI NUMBER JOINER..KAITHI SIGN ANUSVARA
    Interval{ 0x110B3, 0x110B6 }, // Mn   [4] KAITHI VOWEL SIGN U..KAITHI VOWEL SIGN AI
    Interval{ 0x110B9, 0x110BA }, // Mn   [2] KAITHI SIGN VIRAMA..KAITHI SIGN NUKTA
    Interval{ 0x110BD, 0x110BD }, // Cf       KAITHI NUMBER SIGN
    Interval{ 0x110C2, 0x110C2 }, // Mn       KAITHI VOWEL SIGN VOCALIC R
    Interval{ 0x110CD, 0x110CD }, // Cf       KAITHI NUMBER SIGN ABOVE
    Interval{ 0x11100, 0x11102 }, // Mn   [3] CHAKMA SIGN CANDRABINDU..CHAKMA SIGN VISARGA
    Interval{ 0x11127, 0x1112B }, // Mn   [5] CHAKMA VOWEL SIGN A..CHAKMA VOWEL SIGN UU
    Interval{ 0x1112D, 0x11134 }, // Mn   [8] CHAKMA VOWEL SIGN AI..CHAKMA MAAYYAA
    Interval{ 0x11173, 0x11173 }, // Mn       MAHAJANI SIGN NUKTA
    Interval{ 0x11180, 0x11181 }, // Mn   [2] SHARADA SIGN CANDRABINDU..SHARADA SIGN ANUSVARA
    Interval{ 0x111B6, 0x111BE }, // Mn   [9] SHARADA VOWEL SIGN U..SHARADA VOWEL SIGN O
    Interval{ 0x111C9, 0x111CC }, // Mn   [4] SHARADA SANDHI MARK..SHARADA EXTRA SHORT VOWEL MARK
    Interval{ 0x111CF, 0x111CF }, // Mn       SHARADA SIGN INVERTED CANDRABINDU
    Interval{ 0x1122F, 0x11231 }, // Mn   [3] KHOJKI VOWEL SIGN U..KHOJKI VOWEL SIGN AI
    Interval{ 0x11234, 0x11234 }, // Mn       KHOJKI SIGN ANUSVARA
    Interval{ 0x11236, 0x11237 }, // Mn   [2] KHOJKI SIGN NUKTA..KHOJKI SIGN SHADDA
    Interval{ 0x1123E, 0x1123E }, // Mn       KHOJKI SIGN SUKUN
    Interval{ 0x11241, 0x11241 }, // Mn       KHOJKI VOWEL SIGN VOCALIC R
    Interval{ 0x112DF, 0x112DF }, // Mn       KHUDAWADI SIGN ANUSVARA
    Interval{ 0x112E3, 0x112EA }, // Mn   [8] KHUDAWADI VOWEL SIGN U..KHUDAWADI SIGN VIRAMA
    Interval{ 0x11300, 0x11301 }, // Mn   [2] GRANTHA SIGN COMBINING ANUSVARA ABOVE..GRANTHA SIGN CANDRABINDU
    Interval{ 0x1133B, 0x1133C }, // Mn   [2] COMBINING BINDU BELOW..GRANTHA SIGN NUKTA
    Interval{ 0x11340, 0x11340 }, // Mn       GRANTHA VOWEL SIGN II
    Interval{ 0x11366, 0x1136C }, // Mn   [7] COMBINING GRANTHA DIGIT ZERO..COMBINING GRANTHA DIGIT SIX
    Interval{ 0x11370, 0x11374 }, // Mn   [5] COMBINING GRANTHA LETTER A..COMBINING GRANTHA LETTER PA
    Interval{ 0x113BB, 0x113C0 }, // Mn   [6] TULU-TIGALARI VOWEL SIGN U..TULU-TIGALARI VOWEL SIGN VOCALIC LL
    Interval{ 0x113CE, 0x113CE }, // Mn       TULU-TIGALARI SIGN VIRAMA
    Interval{ 0x113D0, 0x113D0 }, // Mn       TULU-TIGALARI CONJOINER
    Interval{ 0x113D2, 0x113D2 }, // Mn       TULU-TIGALARI GEMINATION MARK
    Interval{ 0x113E1, 0x113E2 }, // Mn   [2] TULU-TIGALARI VEDIC TONE SVARITA..TULU-TIGALARI VEDIC TONE ANUDATTA
    Interval{ 0x11438, 0x1143F }, // Mn   [8] NEWA VOWEL SIGN U..NEWA VOWEL SIGN AI
    Interval{ 0x11442, 0x11444 }, // Mn   [3] NEWA SIGN VIRAMA..NEWA SIGN ANUSVARA
    Interval{ 0x11446, 0x11446 }, // Mn       NEWA SIGN NUKTA
    Interval{ 0x1145E, 0x1145E }, // Mn       NEWA SANDHI MARK
    Interval{ 0x114B3, 0x114B8 }, // Mn   [6] TIRHUTA VOWEL SIGN U..TIRHUTA VOWEL SIGN VOCALIC LL
    Interval{ 0x114BA, 0x114BA }, // Mn       TIRHUTA VOWEL SIGN SHORT E
    Interval{ 0x114BF, 0x114C0 }, // Mn   [2] TIRHUTA SIGN CANDRABINDU..TIRHUTA SIGN ANUSVARA
    Interval{ 0x114C2, 0x114C3 }, // Mn   [2] TIRHUTA SIGN VIRAMA..TIRHUTA SIGN NUKTA
    Interval{ 0x115B2, 0x115B5 }, // Mn   [4] SIDDHAM VOWEL SIGN U..SIDDHAM VOWEL SIGN VOCALIC RR
    Interval{ 0x115BC, 0x115BD }, // Mn   [2] SIDDHAM SIGN CANDRABINDU..SIDDHAM SIGN ANUSVARA
    Interval{ 0x115BF, 0x115C0 }, // Mn   [2] SIDDHAM SIGN VIRAMA..SIDDHAM SIGN NUKTA
    Interval{ 0x115DC, 0x115DD }, // Mn   [2] SIDDHAM VOWEL SIGN ALTERNATE U..SIDDHAM VOWEL SIGN ALTERNATE UU
    Interval{ 0x11633, 0x1163A }, // Mn   [8] MODI VOWEL SIGN U..MODI VOWEL SIGN AI
    Interval{ 0x1163D, 0x1163D }, // Mn       MODI SIGN ANUSVARA
    Interval{ 0x1163F, 0x11640 }, // Mn   [2] MODI SIGN VIRAMA..MODI SIGN ARDHACANDRA
    Interval{ 0x116AB, 0x116AB }, // Mn       TAKRI SIGN ANUSVARA
    Interval{ 0x116AD, 0x116AD }, // Mn       TAKRI VOWEL SIGN AA
    Interval{ 0x116B0, 0x116B5 }, // Mn   [6] TAKRI VOWEL SIGN U..TAKRI VOWEL SIGN AU
    Interval{ 0x116B7, 0x116B7 }, // Mn       TAKRI SIGN NUKTA
    Interval{ 0x1171D, 0x1171D }, // Mn       AHOM CONSONANT SIGN MEDIAL LA
    Interval{ 0x1171F, 0x1171F }, // Mn       AHOM CONSONANT SIGN MEDIAL LIGATING RA
    Interval{ 0x11722, 0x11725 }, // Mn   [4] AHOM VOWEL SIGN I..AHOM VOWEL SIGN UU
    Interval{ 0x11727, 0x1172B }, // Mn   [5] AHOM VOWEL SIGN AW..AHOM SIGN KILLER
    Interval{ 0x1182F, 0x11837 }, // Mn   [9] DOGRA VOWEL SIGN U..DOGRA SIGN ANUSVARA
    Interval{ 0x11839, 0x1183A }, // Mn   [2] DOGRA SIGN VIRAMA..DOGRA SIGN NUKTA
    Interval{ 0x1193B, 0x1193C }, // Mn   [2] DIVES AKURU SIGN ANUSVARA..DIVES AKURU SIGN CANDRABINDU
    Interval{ 0x1193E, 0x1193E }, // Mn       DIVES AKURU VIRAMA
    Interval{ 0x11943, 0x11943 }, // Mn       DIVES AKURU SIGN NUKTA
    Interval{ 0x119D4, 0x119D7 }, // Mn   [4] NANDINAGARI VOWEL SIGN U..NANDINAGARI VOWEL SIGN VOCALIC RR
    Interval{ 0x119DA, 0x119DB }, // Mn   [2] NANDINAGARI VOWEL SIGN E..NANDINAGARI VOWEL SIGN AI
    Interval{ 0x119E0, 0x119E0 }, // Mn       NANDINAGARI SIGN VIRAMA
    Interval{ 0x11A01, 0x11A0A }, // Mn  [10] ZANABAZAR SQUARE VOWEL SIGN I..ZANABAZAR SQUARE VOWEL LENGTH MARK
    Interval{ 0x11A33, 0x11A38 }, // Mn   [6] ZANABAZAR SQUARE FINAL CONSONANT MARK..ZANABAZAR SQUARE SIGN ANUSVARA
    Interval{ 0x11A3B, 0x11A3E }, // Mn   [4] ZANABAZAR SQUARE CLUSTER-FINAL LETTER YA..ZANABAZAR SQUARE CLUSTER-FINAL LETTER VA
    Interval{ 0x11A47, 0x11A47 }, // Mn       ZANABAZAR SQUARE SUBJOINER
    Interval{ 0x11A51, 0x11A56 }, // Mn   [6] SOYOMBO VOWEL SIGN I..SOYOMBO VOWEL SIGN OE
    Interval{ 0x11A59, 0x11A5B }, // Mn   [3] SOYOMBO VOWEL SIGN VOCALIC R..SOYOMBO VOWEL LENGTH MARK
    Interval{ 0x11A8A, 0x11A96 }, // Mn  [13] SOYOMBO FINAL CONSONANT SIGN G..SOYOMBO SIGN ANUSVARA
    Interval{ 0x11A98, 0x11A99 }, // Mn   [2] SOYOMBO GEMINATION MARK..SOYOMBO SUBJOINER
    Interval{ 0x11B60, 0x11B60 }, // Mn       SHARADA VOWEL SIGN OE
    Interval{ 0x11B62, 0x11B64 }, // Mn   [3] SHARADA VOWEL SIGN UE..SHARADA VOWEL SIGN SHORT E
    Interval{ 0x11B66, 0x11B66 }, // Mn       SHARADA VOWEL SIGN CANDRA E
    Interval{ 0x11C30, 0x11C36 }, // Mn   [7] BHAIKSUKI VOWEL SIGN I..BHAIKSUKI VOWEL SIGN VOCALIC L
    Interval{ 0x11C38, 0x11C3D }, // Mn   [6] BHAIKSUKI VOWEL SIGN E..BHAIKSUKI SIGN ANUSVARA
    Interval{ 0x11C3F, 0x11C3F }, // Mn       BHAIKSUKI SIGN VIRAMA
    Interval{ 0x11C92, 0x11CA7 }, // Mn  [22] MARCHEN SUBJOINED LETTER KA..MARCHEN SUBJOINED LETTER ZA
    Interval{ 0x11CAA, 0x11CB0 }, // Mn   [7] MARCHEN SUBJOINED LETTER RA..MARCHEN VOWEL SIGN AA
    Interval{ 0x11CB2, 0x11CB3 }, // Mn   [2] MARCHEN VOWEL SIGN U..MARCHEN VOWEL SIGN E
    Interval{ 0x11CB5, 0x11CB6 }, // Mn   [2] MARCHEN SIGN ANUSVARA..MARCHEN SIGN CANDRABINDU
    Interval{ 0x11D31, 0x11D36 }, // Mn   [6] MASARAM GONDI VOWEL SIGN AA..MASARAM GONDI VOWEL SIGN VOCALIC R
    Interval{ 0x11D3A, 0x11D3A }, // Mn       MASARAM GONDI VOWEL SIGN E
    Interval{ 0x11D3C, 0x11D3D }, // Mn   [2] MASARAM GONDI VOWEL SIGN AI..MASARAM GONDI VOWEL SIGN O
    Interval{ 0x11D3F, 0x11D45 }, // Mn   [7] MASARAM GONDI VOWEL SIGN AU..MASARAM GONDI VIRAMA
    Interval{ 0x11D47, 0x11D47 }, // Mn       MASARAM GONDI RA-KARA
    Interval{ 0x11D90, 0x11D91 }, // Mn   [2] GUNJALA GONDI VOWEL SIGN EE..GUNJALA GONDI VOWEL SIGN AI
    Interval{ 0x11D95, 0x11D95 }, // Mn       GUNJALA GONDI SIGN ANUSVARA
    Interval{ 0x11D97, 0x11D97 }, // Mn       GUNJALA GONDI VIRAMA
    Interval{ 0x11DD9, 0x11DD9 }, // Lm       TOLONG SIKI SIGN SELA
    Interval{ 0x11EF3, 0x11EF4 }, // Mn   [2] MAKASAR VOWEL SIGN I..MAKASAR VOWEL SIGN U
    Interval{ 0x11F00, 0x11F01 }, // Mn   [2] KAWI SIGN CANDRABINDU..KAWI SIGN ANUSVARA
    Interval{ 0x11F36, 0x11F3A }, // Mn   [5] KAWI VOWEL SIGN I..KAWI VOWEL SIGN VOCALIC R
    Interval{ 0x11F40, 0x11F40 }, // Mn       KAWI VOWEL SIGN EU
    Interval{ 0x11F42, 0x11F42 }, // Mn       KAWI CONJOINER
    Interval{ 0x11F5A, 0x11F5A }, // Mn       KAWI SIGN NUKTA
    Interval{ 0x13430, 0x1343F }, // Cf  [16] EGYPTIAN HIEROGLYPH VERTICAL JOINER..EGYPTIAN HIEROGLYPH END WALLED ENCLOSURE
    Interval{ 0x13440, 0x13440 }, // Mn       EGYPTIAN HIEROGLYPH MIRROR HORIZONTALLY
    Interval{ 0x13447, 0x13455 }, // Mn  [15] EGYPTIAN HIEROGLYPH MODIFIER DAMAGED AT TOP START..EGYPTIAN HIEROGLYPH MODIFIER DAMAGED
    Interval{ 0x1611E, 0x16129 }, // Mn  [12] GURUNG KHEMA VOWEL SIGN AA..GURUNG KHEMA VOWEL LENGTH MARK
    Interval{ 0x1612D, 0x1612F }, // Mn   [3] GURUNG KHEMA SIGN ANUSVARA..GURUNG KHEMA SIGN THOLHOMA
    Interval{ 0x16AF0, 0x16AF4 }, // Mn   [5] BASSA VAH COMBINING HIGH TONE..BASSA VAH COMBINING HIGH-LOW TONE
    Interval{ 0x16B30, 0x16B36 }, // Mn   [7] PAHAWH HMONG MARK CIM TUB..PAHAWH HMONG MARK CIM TAUM
    Interval{ 0x16B40, 0x16B43 }, // Lm   [4] PAHAWH HMONG SIGN VOS SEEV..PAHAWH HMONG SIGN IB YAM
    Interval{ 0x16D40, 0x16D42 }, // Lm   [3] KIRAT RAI SIGN ANUSVARA..KIRAT RAI SIGN VISARGA
    Interval{ 0x16D6B, 0x16D6C }, // Lm   [2] KIRAT RAI SIGN VIRAMA..KIRAT RAI SIGN SAAT
    Interval{ 0x16F4F, 0x16F4F }, // Mn       MIAO SIGN CONSONANT MODIFIER BAR
    Interval{ 0x16F8F, 0x16F92 }, // Mn   [4] MIAO TONE RIGHT..MIAO TONE BELOW
    Interval{ 0x16F93, 0x16F9F }, // Lm  [13] MIAO LETTER TONE-2..MIAO LETTER REFORMED TONE-8
    Interval{ 0x16FE0, 0x16FE1 }, // Lm   [2] TANGUT ITERATION MARK..NUSHU ITERATION MARK
    Interval{ 0x16FE3, 0x16FE3 }, // Lm       OLD CHINESE ITERATION MARK
    Interval{ 0x16FE4, 0x16FE4 }, // Mn       KHITAN SMALL SCRIPT FILLER
    Interval{ 0x16FF2, 0x16FF3 }, // Lm   [2] CHINESE SMALL SIMPLIFIED ER..CHINESE SMALL TRADITIONAL ER
    Interval{ 0x1AFF0, 0x1AFF3 }, // Lm   [4] KATAKANA LETTER MINNAN TONE-2..KATAKANA LETTER MINNAN TONE-5
    Interval{ 0x1AFF5, 0x1AFFB }, // Lm   [7] KATAKANA LETTER MINNAN TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-5
    Interval{ 0x1AFFD, 0x1AFFE }, // Lm   [2] KATAKANA LETTER MINNAN NASALIZED TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-8
    Interval{ 0x1BC9D, 0x1BC9E }, // Mn   [2] DUPLOYAN THICK LETTER SELECTOR..DUPLOYAN DOUBLE MARK
    Interval{ 0x1BCA0, 0x1BCA3 }, // Cf   [4] SHORTHAND FORMAT LETTER OVERLAP..SHORTHAND FORMAT UP STEP
    Interval{ 0x1CF00, 0x1CF2D }, // Mn  [46] ZNAMENNY COMBINING MARK GORAZDO NIZKO S KRYZHEM ON LEFT..ZNAMENNY COMBINING MARK KRYZH ON LEFT
    Interval{ 0x1CF30, 0x1CF46 }, // Mn  [23] ZNAMENNY COMBINING TONAL RANGE MARK MRACHNO..ZNAMENNY PRIZNAK MODIFIER ROG
    Interval{ 0x1D167, 0x1D169 }, // Mn   [3] MUSICAL SYMBOL COMBINING TREMOLO-1..MUSICAL SYMBOL COMBINING TREMOLO-3
    Interval{ 0x1D173, 0x1D17A }, // Cf   [8] MUSICAL SYMBOL BEGIN BEAM..MUSICAL SYMBOL END PHRASE
    Interval{ 0x1D17B, 0x1D182 }, // Mn   [8] MUSICAL SYMBOL COMBINING ACCENT..MUSICAL SYMBOL COMBINING LOURE
    Interval{ 0x1D185, 0x1D18B }, // Mn   [7] MUSICAL SYMBOL COMBINING DOIT..MUSICAL SYMBOL COMBINING TRIPLE TONGUE
    Interval{ 0x1D1AA, 0x1D1AD }, // Mn   [4] MUSICAL SYMBOL COMBINING DOWN BOW..MUSICAL SYMBOL COMBINING SNAP PIZZICATO
    Interval{ 0x1D242, 0x1D244 }, // Mn   [3] COMBINING GREEK MUSICAL TRISEME..COMBINING GREEK MUSICAL PENTASEME
    Interval{ 0x1DA00, 0x1DA36 }, // Mn  [55] SIGNWRITING HEAD RIM..SIGNWRITING AIR SUCKING IN
    Interval{ 0x1DA3B, 0x1DA6C }, // Mn  [50] SIGNWRITING MOUTH CLOSED NEUTRAL..SIGNWRITING EXCITEMENT
    Interval{ 0x1DA75, 0x1DA75 }, // Mn       SIGNWRITING UPPER BODY TILTING FROM HIP JOINTS
    Interval{ 0x1DA84, 0x1DA84 }, // Mn       SIGNWRITING LOCATION HEAD NECK
    Interval{ 0x1DA9B, 0x1DA9F }, // Mn   [5] SIGNWRITING FILL MODIFIER-2..SIGNWRITING FILL MODIFIER-6
    Interval{ 0x1DAA1, 0x1DAAF }, // Mn  [15] SIGNWRITING ROTATION MODIFIER-2..SIGNWRITING ROTATION MODIFIER-16
    Interval{ 0x1E000, 0x1E006 }, // Mn   [7] COMBINING GLAGOLITIC LETTER AZU..COMBINING GLAGOLITIC LETTER ZHIVETE
    Interval{ 0x1E008, 0x1E018 }, // Mn  [17] COMBINING GLAGOLITIC LETTER ZEMLJA..COMBINING GLAGOLITIC LETTER HERU
    Interval{ 0x1E01B, 0x1E021 }, // Mn   [7] COMBINING GLAGOLITIC LETTER SHTA..COMBINING GLAGOLITIC LETTER YATI
    Interval{ 0x1E023, 0x1E024 }, // Mn   [2] COMBINING GLAGOLITIC LETTER YU..COMBINING GLAGOLITIC LETTER SMALL YUS
    Interval{ 0x1E026, 0x1E02A }, // Mn   [5] COMBINING GLAGOLITIC LETTER YO..COMBINING GLAGOLITIC LETTER FITA
    Interval{ 0x1E030, 0x1E06D }, // Lm  [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Interval{ 0x1E08F, 0x1E08F }, // Mn       COMBINING CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
    Interval{ 0x1E130, 0x1E136 }, // Mn   [7] NYIAKENG PUACHUE HMONG TONE-B..NYIAKENG PUACHUE HMONG TONE-D
    Interval{ 0x1E137, 0x1E13D }, // Lm   [7] NYIAKENG PUACHUE HMONG SIGN FOR PERSON..NYIAKENG PUACHUE HMONG SYLLABLE LENGTHENER
    Interval{ 0x1E2AE, 0x1E2AE }, // Mn       TOTO SIGN RISING TONE
    Interval{ 0x1E2EC, 0x1E2EF }, // Mn   [4] WANCHO TONE TUP..WANCHO TONE KOINI
    Interval{ 0x1E4EB, 0x1E4EB }, // Lm       NAG MUNDARI SIGN OJOD
    Interval{ 0x1E4EC, 0x1E4EF }, // Mn   [4] NAG MUNDARI SIGN MUHOR..NAG MUNDARI SIGN SUTUH
    Interval{ 0x1E5EE, 0x1E5EF }, // Mn   [2] OL ONAL SIGN MU..OL ONAL SIGN IKIR
    Interval{ 0x1E6E3, 0x1E6E3 }, // Mn       TAI YO SIGN UE
    Interval{ 0x1E6E6, 0x1E6E6 }, // Mn       TAI YO SIGN AU
    Interval{ 0x1E6EE, 0x1E6EF }, // Mn   [2] TAI YO SIGN AY..TAI YO SIGN ANG
    Interval{ 0x1E6F5, 0x1E6F5 }, // Mn       TAI YO SIGN OM
    Interval{ 0x1E6FF, 0x1E6FF }, // Lm       TAI YO XAM LAI
    Interval{ 0x1E8D0, 0x1E8D6 }, // Mn   [7] MENDE KIKAKUI COMBINING NUMBER TEENS..MENDE KIKAKUI COMBINING NUMBER MILLIONS
    Interval{ 0x1E944, 0x1E94A }, // Mn   [7] ADLAM ALIF LENGTHENER..ADLAM NUKTA
    Interval{ 0x1E94B, 0x1E94B }, // Lm       ADLAM NASALIZATION MARK
    Interval{ 0x1F3FB, 0x1F3FF }, // Sk   [5] EMOJI MODIFIER FITZPATRICK TYPE-1-2..EMOJI MODIFIER FITZPATRICK TYPE-6
    Interval{ 0xE0001, 0xE0001 }, // Cf       LANGUAGE TAG
    Interval{ 0xE0020, 0xE007F }, // Cf  [96] TAG SPACE..CANCEL TAG
    Interval{ 0xE0100, 0xE01EF }, // Mn [240] VARIATION SELECTOR-17..VARIATION SELECTOR-256
}; // }}}
auto static const Cased = std::array<Interval, 175>{ // {{{
    Interval{ 0x0041, 0x005A }, // L&  [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Interval{ 0x0061, 0x007A }, // L&  [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Interval{ 0x00AA, 0x00AA }, // Lo       FEMININE ORDINAL INDICATOR
    Interval{ 0x00B5, 0x00B5 }, // L&       MICRO SIGN
    Interval{ 0x00BA, 0x00BA }, // Lo       MASCULINE ORDINAL INDICATOR
    Interval{ 0x00C0, 0x00D6 }, // L&  [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x00D8, 0x00F6 }, // L&  [31] LATIN CAPITAL LETTER O WITH STROKE..LATIN SMALL LETTER O WITH DIAERESIS
    Interval{ 0x00F8, 0x01BA }, // L& [195] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER EZH WITH TAIL
    Interval{ 0x01BC, 0x01BF }, // L&   [4] LATIN CAPITAL LETTER TONE FIVE..LATIN LETTER WYNN
    Interval{ 0x01C4, 0x0293 }, // L& [208] LATIN CAPITAL LETTER DZ WITH CARON..LATIN SMALL LETTER EZH WITH CURL
    Interval{ 0x0296, 0x02AF }, // L&  [26] LATIN LETTER INVERTED GLOTTAL STOP..LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    Interval{ 0x02B0, 0x02B8 }, // Lm   [9] MODIFIER LETTER SMALL H..MODIFIER LETTER SMALL Y
    Interval{ 0x02C0, 0x02C1 }, // Lm   [2] MODIFIER LETTER GLOTTAL STOP..MODIFIER LETTER REVERSED GLOTTAL STOP
    Interval{ 0x02E0, 0x02E4 }, // Lm   [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    Interval{ 0x0345, 0x0345 }, // Mn       COMBINING GREEK YPOGEGRAMMENI
    Interval{ 0x0370, 0x0373 }, // L&   [4] GREEK CAPITAL LETTER HETA..GREEK SMALL LETTER ARCHAIC SAMPI
    Interval{ 0x0376, 0x0377 }, // L&   [2] GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA..GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037A, 0x037A }, // Lm       GREEK YPOGEGRAMMENI
    Interval{ 0x037B, 0x037D }, // L&   [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Interval{ 0x037F, 0x037F }, // L&       GREEK CAPITAL LETTER YOT
    Interval{ 0x0386, 0x0386 }, // L&       GREEK CAPITAL LETTER ALPHA WITH TONOS
    Interval{ 0x0388, 0x038A }, // L&   [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Interval{ 0x038C, 0x038C }, // L&       GREEK CAPITAL LETTER OMICRON WITH TONOS
    Interval{ 0x038E, 0x03A1 }, // L&  [20] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER RHO
    Interval{ 0x03A3, 0x03F5 }, // L&  [83] GREEK CAPITAL LETTER SIGMA..GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F7, 0x0481 }, // L& [139] GREEK CAPITAL LETTER SHO..CYRILLIC SMALL LETTER KOPPA
    Interval{ 0x048A, 0x052F }, // L& [166] CYRILLIC CAPITAL LETTER SHORT I WITH TAIL..CYRILLIC SMALL LETTER EL WITH DESCENDER
    Interval{ 0x0531, 0x0556 }, // L&  [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Interval{ 0x0560, 0x0588 }, // L&  [41] ARMENIAN SMALL LETTER TURNED AYB..ARMENIAN SMALL LETTER YI WITH STROKE
    Interval{ 0x10A0, 0x10C5 }, // L&  [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Interval{ 0x10C7, 0x10C7 }, // L&       GEORGIAN CAPITAL LETTER YN
    Interval{ 0x10CD, 0x10CD }, // L&       GEORGIAN CAPITAL LETTER AEN
    Interval{ 0x10D0, 0x10FA }, // L&  [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Interval{ 0x10FC, 0x10FC }, // Lm       MODIFIER LETTER GEORGIAN NAR
    Interval{ 0x10FD, 0x10FF }, // L&   [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Interval{ 0x13A0, 0x13F5 }, // L&  [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Interval{ 0x13F8, 0x13FD }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1C80, 0x1C8A }, // L&  [11] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER TJE
    Interval{ 0x1C90, 0x1CBA }, // L&  [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Interval{ 0x1CBD, 0x1CBF }, // L&   [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Interval{ 0x1D00, 0x1D2B }, // L&  [44] LATIN LETTER SMALL CAPITAL A..CYRILLIC LETTER SMALL CAPITAL EL
    Interval{ 0x1D2C, 0x1D6A }, // Lm  [63] MODIFIER LETTER CAPITAL A..GREEK SUBSCRIPT SMALL LETTER CHI
    Interval{ 0x1D6B, 0x1D77 }, // L&  [13] LATIN SMALL LETTER UE..LATIN SMALL LETTER TURNED G
    Interval{ 0x1D78, 0x1D78 }, // Lm       MODIFIER LETTER CYRILLIC EN
    Interval{ 0x1D79, 0x1D9A }, // L&  [34] LATIN SMALL LETTER INSULAR G..LATIN SMALL LETTER EZH WITH RETROFLEX HOOK
    Interval{ 0x1D9B, 0x1DBF }, // Lm  [37] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL THETA
    Interval{ 0x1E00, 0x1F15 }, // L& [278] LATIN CAPITAL LETTER A WITH RING BELOW..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F18, 0x1F1D }, // L&   [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F20, 0x1F45 }, // L&  [38] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F48, 0x1F4D }, // L&   [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F50, 0x1F57 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F59, 0x1F59 }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA
    Interval{ 0x1F5B, 0x1F5B }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Interval{ 0x1F5D, 0x1F5D }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Interval{ 0x1F5F, 0x1F7D }, // L&  [31] GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI..GREEK SMALL LETTER OMEGA WITH OXIA
    Interval{ 0x1F80, 0x1FB4 }, // L&  [53] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB6, 0x1FBC }, // L&   [7] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Interval{ 0x1FBE, 0x1FBE }, // L&       GREEK PROSGEGRAMMENI
    Interval{ 0x1FC2, 0x1FC4 }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC6, 0x1FCC }, // L&   [7] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Interval{ 0x1FD0, 0x1FD3 }, // L&   [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Interval{ 0x1FD6, 0x1FDB }, // L&   [6] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK CAPITAL LETTER IOTA WITH OXIA
    Interval{ 0x1FE0, 0x1FEC }, // L&  [13] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Interval{ 0x1FF2, 0x1FF4 }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF6, 0x1FFC }, // L&   [7] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Interval{ 0x2071, 0x2071 }, // Lm       SUPERSCRIPT LATIN SMALL LETTER I
    Interval{ 0x207F, 0x207F }, // Lm       SUPERSCRIPT LATIN SMALL LETTER N
    Interval{ 0x2090, 0x209C }, // Lm  [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Interval{ 0x2102, 0x2102 }, // L&       DOUBLE-STRUCK CAPITAL C
    Interval{ 0x2107, 0x2107 }, // L&       EULER CONSTANT
    Interval{ 0x210A, 0x2113 }, // L&  [10] SCRIPT SMALL G..SCRIPT SMALL L
    Interval{ 0x2115, 0x2115 }, // L&       DOUBLE-STRUCK CAPITAL N
    Interval{ 0x2119, 0x211D }, // L&   [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    Interval{ 0x2124, 0x2124 }, // L&       DOUBLE-STRUCK CAPITAL Z
    Interval{ 0x2126, 0x2126 }, // L&       OHM SIGN
    Interval{ 0x2128, 0x2128 }, // L&       BLACK-LETTER CAPITAL Z
    Interval{ 0x212A, 0x212D }, // L&   [4] KELVIN SIGN..BLACK-LETTER CAPITAL C
    Interval{ 0x212F, 0x2134 }, // L&   [6] SCRIPT SMALL E..SCRIPT SMALL O
    Interval{ 0x2139, 0x2139 }, // L&       INFORMATION SOURCE
    Interval{ 0x213C, 0x213F }, // L&   [4] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK CAPITAL PI
    Interval{ 0x2145, 0x2149 }, // L&   [5] DOUBLE-STRUCK ITALIC CAPITAL D..DOUBLE-STRUCK ITALIC SMALL J
    Interval{ 0x214E, 0x214E }, // L&       TURNED SMALL F
    Interval{ 0x2160, 0x217F }, // Nl  [32] ROMAN NUMERAL ONE..SMALL ROMAN NUMERAL ONE THOUSAND
    Interval{ 0x2183, 0x2184 }, // L&   [2] ROMAN NUMERAL REVERSED ONE HUNDRED..LATIN SMALL LETTER REVERSED C
    Interval{ 0x24B6, 0x24E9 }, // So  [52] CIRCLED LATIN CAPITAL LETTER A..CIRCLED LATIN SMALL LETTER Z
    Interval{ 0x2C00, 0x2C7B }, // L& [124] GLAGOLITIC CAPITAL LETTER AZU..LATIN LETTER SMALL CAPITAL TURNED E
    Interval{ 0x2C7C, 0x2C7D }, // Lm   [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    Interval{ 0x2C7E, 0x2CE4 }, // L& [103] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC SYMBOL KAI
    Interval{ 0x2CEB, 0x2CEE }, // L&   [4] COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI..COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF2, 0x2CF3 }, // L&   [2] COPTIC CAPITAL LETTER BOHAIRIC KHEI..COPTIC SMALL LETTER BOHAIRIC KHEI
    Interval{ 0x2D00, 0x2D25 }, // L&  [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Interval{ 0x2D27, 0x2D27 }, // L&       GEORGIAN SMALL LETTER YN
    Interval{ 0x2D2D, 0x2D2D }, // L&       GEORGIAN SMALL LETTER AEN
    Interval{ 0xA640, 0xA66D }, // L&  [46] CYRILLIC CAPITAL LETTER ZEMLYA..CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA680, 0xA69B }, // L&  [28] CYRILLIC CAPITAL LETTER DWE..CYRILLIC SMALL LETTER CROSSED O
    Interval{ 0xA69C, 0xA69D }, // Lm   [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    Interval{ 0xA722, 0xA76F }, // L&  [78] LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF..LATIN SMALL LETTER CON
    Interval{ 0xA770, 0xA770 }, // Lm       MODIFIER LETTER US
    Interval{ 0xA771, 0xA787 }, // L&  [23] LATIN SMALL LETTER DUM..LATIN SMALL LETTER INSULAR T
    Interval{ 0xA78B, 0xA78E }, // L&   [4] LATIN CAPITAL LETTER SALTILLO..LATIN SMALL LETTER L WITH RETROFLEX HOOK AND BELT
    Interval{ 0xA790, 0xA7DC }, // L&  [77] LATIN CAPITAL LETTER N WITH DESCENDER..LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Interval{ 0xA7F1, 0xA7F4 }, // Lm   [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    Interval{ 0xA7F5, 0xA7F6 }, // L&   [2] LATIN CAPITAL LETTER REVERSED HALF H..LATIN SMALL LETTER REVERSED HALF H
    Interval{ 0xA7F8, 0xA7F9 }, // Lm   [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    Interval{ 0xA7FA, 0xA7FA }, // L&       LATIN LETTER SMALL CAPITAL TURNED M
    Interval{ 0xAB30, 0xAB5A }, // L&  [43] LATIN SMALL LETTER BARRED ALPHA..LATIN SMALL LETTER Y WITH SHORT RIGHT LEG
    Interval{ 0xAB5C, 0xAB5F }, // Lm   [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    Interval{ 0xAB60, 0xAB68 }, // L&   [9] LATIN SMALL LETTER SAKHA YAT..LATIN SMALL LETTER TURNED R WITH MIDDLE TILDE
    Interval{ 0xAB69, 0xAB69 }, // Lm       MODIFIER LETTER SMALL TURNED W
    Interval{ 0xAB70, 0xABBF }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xFB00, 0xFB06 }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFF21, 0xFF3A }, // L&  [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Interval{ 0xFF41, 0xFF5A }, // L&  [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Interval{ 0x10400, 0x1044F }, // L&  [80] DESERET CAPITAL LETTER LONG I..DESERET SMALL LETTER EW
    Interval{ 0x104B0, 0x104D3 }, // L&  [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Interval{ 0x104D8, 0x104FB }, // L&  [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Interval{ 0x10570, 0x1057A }, // L&  [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Interval{ 0x1057C, 0x1058A }, // L&  [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Interval{ 0x1058C, 0x10592 }, // L&   [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Interval{ 0x10594, 0x10595 }, // L&   [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Interval{ 0x10597, 0x105A1 }, // L&  [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Interval{ 0x105A3, 0x105B1 }, // L&  [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Interval{ 0x105B3, 0x105B9 }, // L&   [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Interval{ 0x105BB, 0x105BC }, // L&   [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Interval{ 0x10780, 0x10780 }, // Lm       MODIFIER LETTER SMALL CAPITAL AA
    Interval{ 0x10783, 0x10785 }, // Lm   [3] MODIFIER LETTER SMALL AE..MODIFIER LETTER SMALL B WITH HOOK
    Interval{ 0x10787, 0x107B0 }, // Lm  [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Interval{ 0x107B2, 0x107BA }, // Lm   [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Interval{ 0x10C80, 0x10CB2 }, // L&  [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Interval{ 0x10CC0, 0x10CF2 }, // L&  [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Interval{ 0x10D50, 0x10D65 }, // L&  [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Interval{ 0x10D70, 0x10D85 }, // L&  [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Interval{ 0x118A0, 0x118DF }, // L&  [64] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Interval{ 0x16E40, 0x16E7F }, // L&  [64] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Interval{ 0x16EA0, 0x16EB8 }, // L&  [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Interval{ 0x16EBB, 0x16ED3 }, // L&  [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Interval{ 0x1D400, 0x1D454 }, // L&  [85] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL ITALIC SMALL G
    Interval{ 0x1D456, 0x1D49C }, // L&  [71] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL SCRIPT CAPITAL A
    Interval{ 0x1D49E, 0x1D49F }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Interval{ 0x1D4A2, 0x1D4A2 }, // L&       MATHEMATICAL SCRIPT CAPITAL G
    Interval{ 0x1D4A5, 0x1D4A6 }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Interval{ 0x1D4A9, 0x1D4AC }, // L&   [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Interval{ 0x1D4AE, 0x1D4B9 }, // L&  [12] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT SMALL D
    Interval{ 0x1D4BB, 0x1D4BB }, // L&       MATHEMATICAL SCRIPT SMALL F
    Interval{ 0x1D4BD, 0x1D4C3 }, // L&   [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Interval{ 0x1D4C5, 0x1D505 }, // L&  [65] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL FRAKTUR CAPITAL B
    Interval{ 0x1D507, 0x1D50A }, // L&   [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Interval{ 0x1D50D, 0x1D514 }, // L&   [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Interval{ 0x1D516, 0x1D51C }, // L&   [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Interval{ 0x1D51E, 0x1D539 }, // L&  [28] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Interval{ 0x1D53B, 0x1D53E }, // L&   [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Interval{ 0x1D540, 0x1D544 }, // L&   [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Interval{ 0x1D546, 0x1D546 }, // L&       MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Interval{ 0x1D54A, 0x1D550 }, // L&   [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Interval{ 0x1D552, 0x1D6A5 }, // L& [340] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    Interval{ 0x1D6A8, 0x1D6C0 }, // L&  [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    Interval{ 0x1D6C2, 0x1D6DA }, // L&  [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    Interval{ 0x1D6DC, 0x1D6FA }, // L&  [31] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL ITALIC CAPITAL OMEGA
    Interval{ 0x1D6FC, 0x1D714 }, // L&  [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    Interval{ 0x1D716, 0x1D734 }, // L&  [31] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D736, 0x1D74E }, // L&  [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D750, 0x1D76E }, // L&  [31] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    Interval{ 0x1D770, 0x1D788 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    Interval{ 0x1D78A, 0x1D7A8 }, // L&  [31] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D7AA, 0x1D7C2 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D7C4, 0x1D7CB }, // L&   [8] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD SMALL DIGAMMA
    Interval{ 0x1DF00, 0x1DF09 }, // L&  [10] LATIN SMALL LETTER FENG DIGRAPH WITH TRILL..LATIN SMALL LETTER T WITH HOOK AND RETROFLEX HOOK
    Interval{ 0x1DF0B, 0x1DF1E }, // L&  [20] LATIN SMALL LETTER ESH WITH DOUBLE BAR..LATIN SMALL LETTER S WITH CURL
    Interval{ 0x1DF25, 0x1DF2A }, // L&   [6] LATIN SMALL LETTER D WITH MID-HEIGHT LEFT HOOK..LATIN SMALL LETTER T WITH MID-HEIGHT LEFT HOOK
    Interval{ 0x1E030, 0x1E06D }, // Lm  [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Interval{ 0x1E900, 0x1E943 }, // L&  [68] ADLAM CAPITAL LETTER ALIF..ADLAM SMALL LETTER SHA
    Interval{ 0x1F130, 0x1F149 }, // So  [26] SQUARED LATIN CAPITAL LETTER A..SQUARED LATIN CAPITAL LETTER Z
    Interval{ 0x1F150, 0x1F169 }, // So  [26] NEGATIVE CIRCLED LATIN CAPITAL LETTER A..NEGATIVE CIRCLED LATIN CAPITAL LETTER Z
    Interval{ 0x1F170, 0x1F189 }, // So  [26] NEGATIVE SQUARED LATIN CAPITAL LETTER A..NEGATIVE SQUARED LATIN CAPITAL LETTER Z
}; // }}}
auto static const Changes_When_Casefolded = std::array<Interval, 630>{ // {{{
    Interval{ 0x0041, 0x005A }, // L&  [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Interval{ 0x00B5, 0x00B5 }, // L&       MICRO SIGN
    Interval{ 0x00C0, 0x00D6 }, // L&  [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x00D8, 0x00DF }, // L&   [8] LATIN CAPITAL LETTER O WITH STROKE..LATIN SMALL LETTER SHARP S
    Interval{ 0x0100, 0x0100 }, // L&       LATIN CAPITAL LETTER A WITH MACRON
    Interval{ 0x0102, 0x0102 }, // L&       LATIN CAPITAL LETTER A WITH BREVE
    Interval{ 0x0104, 0x0104 }, // L&       LATIN CAPITAL LETTER A WITH OGONEK
    Interval{ 0x0106, 0x0106 }, // L&       LATIN CAPITAL LETTER C WITH ACUTE
    Interval{ 0x0108, 0x0108 }, // L&       LATIN CAPITAL LETTER C WITH CIRCUMFLEX
    Interval{ 0x010A, 0x010A }, // L&       LATIN CAPITAL LETTER C WITH DOT ABOVE
    Interval{ 0x010C, 0x010C }, // L&       LATIN CAPITAL LETTER C WITH CARON
    Interval{ 0x010E, 0x010E }, // L&       LATIN CAPITAL LETTER D WITH CARON
    Interval{ 0x0110, 0x0110 }, // L&       LATIN CAPITAL LETTER D WITH STROKE
    Interval{ 0x0112, 0x0112 }, // L&       LATIN CAPITAL LETTER E WITH MACRON
    Interval{ 0x0114, 0x0114 }, // L&       LATIN CAPITAL LETTER E WITH BREVE
    Interval{ 0x0116, 0x0116 }, // L&       LATIN CAPITAL LETTER E WITH DOT ABOVE
    Interval{ 0x0118, 0x0118 }, // L&       LATIN CAPITAL LETTER E WITH OGONEK
    Interval{ 0x011A, 0x011A }, // L&       LATIN CAPITAL LETTER E WITH CARON
    Interval{ 0x011C, 0x011C }, // L&       LATIN CAPITAL LETTER G WITH CIRCUMFLEX
    Interval{ 0x011E, 0x011E }, // L&       LATIN CAPITAL LETTER G WITH BREVE
    Interval{ 0x0120, 0x0120 }, // L&       LATIN CAPITAL LETTER G WITH DOT ABOVE
    Interval{ 0x0122, 0x0122 }, // L&       LATIN CAPITAL LETTER G WITH CEDILLA
    Interval{ 0x0124, 0x0124 }, // L&       LATIN CAPITAL LETTER H WITH CIRCUMFLEX
    Interval{ 0x0126, 0x0126 }, // L&       LATIN CAPITAL LETTER H WITH STROKE
    Interval{ 0x0128, 0x0128 }, // L&       LATIN CAPITAL LETTER I WITH TILDE
    Interval{ 0x012A, 0x012A }, // L&       LATIN CAPITAL LETTER I WITH MACRON
    Interval{ 0x012C, 0x012C }, // L&       LATIN CAPITAL LETTER I WITH BREVE
    Interval{ 0x012E, 0x012E }, // L&       LATIN CAPITAL LETTER I WITH OGONEK
    Interval{ 0x0130, 0x0130 }, // L&       LATIN CAPITAL LETTER I WITH DOT ABOVE
    Interval{ 0x0132, 0x0132 }, // L&       LATIN CAPITAL LIGATURE IJ
    Interval{ 0x0134, 0x0134 }, // L&       LATIN CAPITAL LETTER J WITH CIRCUMFLEX
    Interval{ 0x0136, 0x0136 }, // L&       LATIN CAPITAL LETTER K WITH CEDILLA
    Interval{ 0x0139, 0x0139 }, // L&       LATIN CAPITAL LETTER L WITH ACUTE
    Interval{ 0x013B, 0x013B }, // L&       LATIN CAPITAL LETTER L WITH CEDILLA
    Interval{ 0x013D, 0x013D }, // L&       LATIN CAPITAL LETTER L WITH CARON
    Interval{ 0x013F, 0x013F }, // L&       LATIN CAPITAL LETTER L WITH MIDDLE DOT
    Interval{ 0x0141, 0x0141 }, // L&       LATIN CAPITAL LETTER L WITH STROKE
    Interval{ 0x0143, 0x0143 }, // L&       LATIN CAPITAL LETTER N WITH ACUTE
    Interval{ 0x0145, 0x0145 }, // L&       LATIN CAPITAL LETTER N WITH CEDILLA
    Interval{ 0x0147, 0x0147 }, // L&       LATIN CAPITAL LETTER N WITH CARON
    Interval{ 0x0149, 0x014A }, // L&   [2] LATIN SMALL LETTER N PRECEDED BY APOSTROPHE..LATIN CAPITAL LETTER ENG
    Interval{ 0x014C, 0x014C }, // L&       LATIN CAPITAL LETTER O WITH MACRON
    Interval{ 0x014E, 0x014E }, // L&       LATIN CAPITAL LETTER O WITH BREVE
    Interval{ 0x0150, 0x0150 }, // L&       LATIN CAPITAL LETTER O WITH DOUBLE ACUTE
    Interval{ 0x0152, 0x0152 }, // L&       LATIN CAPITAL LIGATURE OE
    Interval{ 0x0154, 0x0154 }, // L&       LATIN CAPITAL LETTER R WITH ACUTE
    Interval{ 0x0156, 0x0156 }, // L&       LATIN CAPITAL LETTER R WITH CEDILLA
    Interval{ 0x0158, 0x0158 }, // L&       LATIN CAPITAL LETTER R WITH CARON
    Interval{ 0x015A, 0x015A }, // L&       LATIN CAPITAL LETTER S WITH ACUTE
    Interval{ 0x015C, 0x015C }, // L&       LATIN CAPITAL LETTER S WITH CIRCUMFLEX
    Interval{ 0x015E, 0x015E }, // L&       LATIN CAPITAL LETTER S WITH CEDILLA
    Interval{ 0x0160, 0x0160 }, // L&       LATIN CAPITAL LETTER S WITH CARON
    Interval{ 0x0162, 0x0162 }, // L&       LATIN CAPITAL LETTER T WITH CEDILLA
    Interval{ 0x0164, 0x0164 }, // L&       LATIN CAPITAL LETTER T WITH CARON
    Interval{ 0x0166, 0x0166 }, // L&       LATIN CAPITAL LETTER T WITH STROKE
    Interval{ 0x0168, 0x0168 }, // L&       LATIN CAPITAL LETTER U WITH TILDE
    Interval{ 0x016A, 0x016A }, // L&       LATIN CAPITAL LETTER U WITH MACRON
    Interval{ 0x016C, 0x016C }, // L&       LATIN CAPITAL LETTER U WITH BREVE
    Interval{ 0x016E, 0x016E }, // L&       LATIN CAPITAL LETTER U WITH RING ABOVE
    Interval{ 0x0170, 0x0170 }, // L&       LATIN CAPITAL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x0172, 0x0172 }, // L&       LATIN CAPITAL LETTER U WITH OGONEK
    Interval{ 0x0174, 0x0174 }, // L&       LATIN CAPITAL LETTER W WITH CIRCUMFLEX
    Interval{ 0x0176, 0x0176 }, // L&       LATIN CAPITAL LETTER Y WITH CIRCUMFLEX
    Interval{ 0x0178, 0x0179 }, // L&   [2] LATIN CAPITAL LETTER Y WITH DIAERESIS..LATIN CAPITAL LETTER Z WITH ACUTE
    Interval{ 0x017B, 0x017B }, // L&       LATIN CAPITAL LETTER Z WITH DOT ABOVE
    Interval{ 0x017D, 0x017D }, // L&       LATIN CAPITAL LETTER Z WITH CARON
    Interval{ 0x017F, 0x017F }, // L&       LATIN SMALL LETTER LONG S
    Interval{ 0x0181, 0x0182 }, // L&   [2] LATIN CAPITAL LETTER B WITH HOOK..LATIN CAPITAL LETTER B WITH TOPBAR
    Interval{ 0x0184, 0x0184 }, // L&       LATIN CAPITAL LETTER TONE SIX
    Interval{ 0x0186, 0x0187 }, // L&   [2] LATIN CAPITAL LETTER OPEN O..LATIN CAPITAL LETTER C WITH HOOK
    Interval{ 0x0189, 0x018B }, // L&   [3] LATIN CAPITAL LETTER AFRICAN D..LATIN CAPITAL LETTER D WITH TOPBAR
    Interval{ 0x018E, 0x0191 }, // L&   [4] LATIN CAPITAL LETTER REVERSED E..LATIN CAPITAL LETTER F WITH HOOK
    Interval{ 0x0193, 0x0194 }, // L&   [2] LATIN CAPITAL LETTER G WITH HOOK..LATIN CAPITAL LETTER GAMMA
    Interval{ 0x0196, 0x0198 }, // L&   [3] LATIN CAPITAL LETTER IOTA..LATIN CAPITAL LETTER K WITH HOOK
    Interval{ 0x019C, 0x019D }, // L&   [2] LATIN CAPITAL LETTER TURNED M..LATIN CAPITAL LETTER N WITH LEFT HOOK
    Interval{ 0x019F, 0x01A0 }, // L&   [2] LATIN CAPITAL LETTER O WITH MIDDLE TILDE..LATIN CAPITAL LETTER O WITH HORN
    Interval{ 0x01A2, 0x01A2 }, // L&       LATIN CAPITAL LETTER OI
    Interval{ 0x01A4, 0x01A4 }, // L&       LATIN CAPITAL LETTER P WITH HOOK
    Interval{ 0x01A6, 0x01A7 }, // L&   [2] LATIN LETTER YR..LATIN CAPITAL LETTER TONE TWO
    Interval{ 0x01A9, 0x01A9 }, // L&       LATIN CAPITAL LETTER ESH
    Interval{ 0x01AC, 0x01AC }, // L&       LATIN CAPITAL LETTER T WITH HOOK
    Interval{ 0x01AE, 0x01AF }, // L&   [2] LATIN CAPITAL LETTER T WITH RETROFLEX HOOK..LATIN CAPITAL LETTER U WITH HORN
    Interval{ 0x01B1, 0x01B3 }, // L&   [3] LATIN CAPITAL LETTER UPSILON..LATIN CAPITAL LETTER Y WITH HOOK
    Interval{ 0x01B5, 0x01B5 }, // L&       LATIN CAPITAL LETTER Z WITH STROKE
    Interval{ 0x01B7, 0x01B8 }, // L&   [2] LATIN CAPITAL LETTER EZH..LATIN CAPITAL LETTER EZH REVERSED
    Interval{ 0x01BC, 0x01BC }, // L&       LATIN CAPITAL LETTER TONE FIVE
    Interval{ 0x01C4, 0x01C5 }, // L&   [2] LATIN CAPITAL LETTER DZ WITH CARON..LATIN CAPITAL LETTER D WITH SMALL LETTER Z WITH CARON
    Interval{ 0x01C7, 0x01C8 }, // L&   [2] LATIN CAPITAL LETTER LJ..LATIN CAPITAL LETTER L WITH SMALL LETTER J
    Interval{ 0x01CA, 0x01CB }, // L&   [2] LATIN CAPITAL LETTER NJ..LATIN CAPITAL LETTER N WITH SMALL LETTER J
    Interval{ 0x01CD, 0x01CD }, // L&       LATIN CAPITAL LETTER A WITH CARON
    Interval{ 0x01CF, 0x01CF }, // L&       LATIN CAPITAL LETTER I WITH CARON
    Interval{ 0x01D1, 0x01D1 }, // L&       LATIN CAPITAL LETTER O WITH CARON
    Interval{ 0x01D3, 0x01D3 }, // L&       LATIN CAPITAL LETTER U WITH CARON
    Interval{ 0x01D5, 0x01D5 }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS AND MACRON
    Interval{ 0x01D7, 0x01D7 }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS AND ACUTE
    Interval{ 0x01D9, 0x01D9 }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS AND CARON
    Interval{ 0x01DB, 0x01DB }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS AND GRAVE
    Interval{ 0x01DE, 0x01DE }, // L&       LATIN CAPITAL LETTER A WITH DIAERESIS AND MACRON
    Interval{ 0x01E0, 0x01E0 }, // L&       LATIN CAPITAL LETTER A WITH DOT ABOVE AND MACRON
    Interval{ 0x01E2, 0x01E2 }, // L&       LATIN CAPITAL LETTER AE WITH MACRON
    Interval{ 0x01E4, 0x01E4 }, // L&       LATIN CAPITAL LETTER G WITH STROKE
    Interval{ 0x01E6, 0x01E6 }, // L&       LATIN CAPITAL LETTER G WITH CARON
    Interval{ 0x01E8, 0x01E8 }, // L&       LATIN CAPITAL LETTER K WITH CARON
    Interval{ 0x01EA, 0x01EA }, // L&       LATIN CAPITAL LETTER O WITH OGONEK
    Interval{ 0x01EC, 0x01EC }, // L&       LATIN CAPITAL LETTER O WITH OGONEK AND MACRON
    Interval{ 0x01EE, 0x01EE }, // L&       LATIN CAPITAL LETTER EZH WITH CARON
    Interval{ 0x01F1, 0x01F2 }, // L&   [2] LATIN CAPITAL LETTER DZ..LATIN CAPITAL LETTER D WITH SMALL LETTER Z
    Interval{ 0x01F4, 0x01F4 }, // L&       LATIN CAPITAL LETTER G WITH ACUTE
    Interval{ 0x01F6, 0x01F8 }, // L&   [3] LATIN CAPITAL LETTER HWAIR..LATIN CAPITAL LETTER N WITH GRAVE
    Interval{ 0x01FA, 0x01FA }, // L&       LATIN CAPITAL LETTER A WITH RING ABOVE AND ACUTE
    Interval{ 0x01FC, 0x01FC }, // L&       LATIN CAPITAL LETTER AE WITH ACUTE
    Interval{ 0x01FE, 0x01FE }, // L&       LATIN CAPITAL LETTER O WITH STROKE AND ACUTE
    Interval{ 0x0200, 0x0200 }, // L&       LATIN CAPITAL LETTER A WITH DOUBLE GRAVE
    Interval{ 0x0202, 0x0202 }, // L&       LATIN CAPITAL LETTER A WITH INVERTED BREVE
    Interval{ 0x0204, 0x0204 }, // L&       LATIN CAPITAL LETTER E WITH DOUBLE GRAVE
    Interval{ 0x0206, 0x0206 }, // L&       LATIN CAPITAL LETTER E WITH INVERTED BREVE
    Interval{ 0x0208, 0x0208 }, // L&       LATIN CAPITAL LETTER I WITH DOUBLE GRAVE
    Interval{ 0x020A, 0x020A }, // L&       LATIN CAPITAL LETTER I WITH INVERTED BREVE
    Interval{ 0x020C, 0x020C }, // L&       LATIN CAPITAL LETTER O WITH DOUBLE GRAVE
    Interval{ 0x020E, 0x020E }, // L&       LATIN CAPITAL LETTER O WITH INVERTED BREVE
    Interval{ 0x0210, 0x0210 }, // L&       LATIN CAPITAL LETTER R WITH DOUBLE GRAVE
    Interval{ 0x0212, 0x0212 }, // L&       LATIN CAPITAL LETTER R WITH INVERTED BREVE
    Interval{ 0x0214, 0x0214 }, // L&       LATIN CAPITAL LETTER U WITH DOUBLE GRAVE
    Interval{ 0x0216, 0x0216 }, // L&       LATIN CAPITAL LETTER U WITH INVERTED BREVE
    Interval{ 0x0218, 0x0218 }, // L&       LATIN CAPITAL LETTER S WITH COMMA BELOW
    Interval{ 0x021A, 0x021A }, // L&       LATIN CAPITAL LETTER T WITH COMMA BELOW
    Interval{ 0x021C, 0x021C }, // L&       LATIN CAPITAL LETTER YOGH
    Interval{ 0x021E, 0x021E }, // L&       LATIN CAPITAL LETTER H WITH CARON
    Interval{ 0x0220, 0x0220 }, // L&       LATIN CAPITAL LETTER N WITH LONG RIGHT LEG
    Interval{ 0x0222, 0x0222 }, // L&       LATIN CAPITAL LETTER OU
    Interval{ 0x0224, 0x0224 }, // L&       LATIN CAPITAL LETTER Z WITH HOOK
    Interval{ 0x0226, 0x0226 }, // L&       LATIN CAPITAL LETTER A WITH DOT ABOVE
    Interval{ 0x0228, 0x0228 }, // L&       LATIN CAPITAL LETTER E WITH CEDILLA
    Interval{ 0x022A, 0x022A }, // L&       LATIN CAPITAL LETTER O WITH DIAERESIS AND MACRON
    Interval{ 0x022C, 0x022C }, // L&       LATIN CAPITAL LETTER O WITH TILDE AND MACRON
    Interval{ 0x022E, 0x022E }, // L&       LATIN CAPITAL LETTER O WITH DOT ABOVE
    Interval{ 0x0230, 0x0230 }, // L&       LATIN CAPITAL LETTER O WITH DOT ABOVE AND MACRON
    Interval{ 0x0232, 0x0232 }, // L&       LATIN CAPITAL LETTER Y WITH MACRON
    Interval{ 0x023A, 0x023B }, // L&   [2] LATIN CAPITAL LETTER A WITH STROKE..LATIN CAPITAL LETTER C WITH STROKE
    Interval{ 0x023D, 0x023E }, // L&   [2] LATIN CAPITAL LETTER L WITH BAR..LATIN CAPITAL LETTER T WITH DIAGONAL STROKE
    Interval{ 0x0241, 0x0241 }, // L&       LATIN CAPITAL LETTER GLOTTAL STOP
    Interval{ 0x0243, 0x0246 }, // L&   [4] LATIN CAPITAL LETTER B WITH STROKE..LATIN CAPITAL LETTER E WITH STROKE
    Interval{ 0x0248, 0x0248 }, // L&       LATIN CAPITAL LETTER J WITH STROKE
    Interval{ 0x024A, 0x024A }, // L&       LATIN CAPITAL LETTER SMALL Q WITH HOOK TAIL
    Interval{ 0x024C, 0x024C }, // L&       LATIN CAPITAL LETTER R WITH STROKE
    Interval{ 0x024E, 0x024E }, // L&       LATIN CAPITAL LETTER Y WITH STROKE
    Interval{ 0x0345, 0x0345 }, // Mn       COMBINING GREEK YPOGEGRAMMENI
    Interval{ 0x0370, 0x0370 }, // L&       GREEK CAPITAL LETTER HETA
    Interval{ 0x0372, 0x0372 }, // L&       GREEK CAPITAL LETTER ARCHAIC SAMPI
    Interval{ 0x0376, 0x0376 }, // L&       GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037F, 0x037F }, // L&       GREEK CAPITAL LETTER YOT
    Interval{ 0x0386, 0x0386 }, // L&       GREEK CAPITAL LETTER ALPHA WITH TONOS
    Interval{ 0x0388, 0x038A }, // L&   [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Interval{ 0x038C, 0x038C }, // L&       GREEK CAPITAL LETTER OMICRON WITH TONOS
    Interval{ 0x038E, 0x038F }, // L&   [2] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER OMEGA WITH TONOS
    Interval{ 0x0391, 0x03A1 }, // L&  [17] GREEK CAPITAL LETTER ALPHA..GREEK CAPITAL LETTER RHO
    Interval{ 0x03A3, 0x03AB }, // L&   [9] GREEK CAPITAL LETTER SIGMA..GREEK CAPITAL LETTER UPSILON WITH DIALYTIKA
    Interval{ 0x03C2, 0x03C2 }, // L&       GREEK SMALL LETTER FINAL SIGMA
    Interval{ 0x03CF, 0x03D1 }, // L&   [3] GREEK CAPITAL KAI SYMBOL..GREEK THETA SYMBOL
    Interval{ 0x03D5, 0x03D6 }, // L&   [2] GREEK PHI SYMBOL..GREEK PI SYMBOL
    Interval{ 0x03D8, 0x03D8 }, // L&       GREEK LETTER ARCHAIC KOPPA
    Interval{ 0x03DA, 0x03DA }, // L&       GREEK LETTER STIGMA
    Interval{ 0x03DC, 0x03DC }, // L&       GREEK LETTER DIGAMMA
    Interval{ 0x03DE, 0x03DE }, // L&       GREEK LETTER KOPPA
    Interval{ 0x03E0, 0x03E0 }, // L&       GREEK LETTER SAMPI
    Interval{ 0x03E2, 0x03E2 }, // L&       COPTIC CAPITAL LETTER SHEI
    Interval{ 0x03E4, 0x03E4 }, // L&       COPTIC CAPITAL LETTER FEI
    Interval{ 0x03E6, 0x03E6 }, // L&       COPTIC CAPITAL LETTER KHEI
    Interval{ 0x03E8, 0x03E8 }, // L&       COPTIC CAPITAL LETTER HORI
    Interval{ 0x03EA, 0x03EA }, // L&       COPTIC CAPITAL LETTER GANGIA
    Interval{ 0x03EC, 0x03EC }, // L&       COPTIC CAPITAL LETTER SHIMA
    Interval{ 0x03EE, 0x03EE }, // L&       COPTIC CAPITAL LETTER DEI
    Interval{ 0x03F0, 0x03F1 }, // L&   [2] GREEK KAPPA SYMBOL..GREEK RHO SYMBOL
    Interval{ 0x03F4, 0x03F5 }, // L&   [2] GREEK CAPITAL THETA SYMBOL..GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F7, 0x03F7 }, // L&       GREEK CAPITAL LETTER SHO
    Interval{ 0x03F9, 0x03FA }, // L&   [2] GREEK CAPITAL LUNATE SIGMA SYMBOL..GREEK CAPITAL LETTER SAN
    Interval{ 0x03FD, 0x042F }, // L&  [51] GREEK CAPITAL REVERSED LUNATE SIGMA SYMBOL..CYRILLIC CAPITAL LETTER YA
    Interval{ 0x0460, 0x0460 }, // L&       CYRILLIC CAPITAL LETTER OMEGA
    Interval{ 0x0462, 0x0462 }, // L&       CYRILLIC CAPITAL LETTER YAT
    Interval{ 0x0464, 0x0464 }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED E
    Interval{ 0x0466, 0x0466 }, // L&       CYRILLIC CAPITAL LETTER LITTLE YUS
    Interval{ 0x0468, 0x0468 }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED LITTLE YUS
    Interval{ 0x046A, 0x046A }, // L&       CYRILLIC CAPITAL LETTER BIG YUS
    Interval{ 0x046C, 0x046C }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED BIG YUS
    Interval{ 0x046E, 0x046E }, // L&       CYRILLIC CAPITAL LETTER KSI
    Interval{ 0x0470, 0x0470 }, // L&       CYRILLIC CAPITAL LETTER PSI
    Interval{ 0x0472, 0x0472 }, // L&       CYRILLIC CAPITAL LETTER FITA
    Interval{ 0x0474, 0x0474 }, // L&       CYRILLIC CAPITAL LETTER IZHITSA
    Interval{ 0x0476, 0x0476 }, // L&       CYRILLIC CAPITAL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT
    Interval{ 0x0478, 0x0478 }, // L&       CYRILLIC CAPITAL LETTER UK
    Interval{ 0x047A, 0x047A }, // L&       CYRILLIC CAPITAL LETTER ROUND OMEGA
    Interval{ 0x047C, 0x047C }, // L&       CYRILLIC CAPITAL LETTER OMEGA WITH TITLO
    Interval{ 0x047E, 0x047E }, // L&       CYRILLIC CAPITAL LETTER OT
    Interval{ 0x0480, 0x0480 }, // L&       CYRILLIC CAPITAL LETTER KOPPA
    Interval{ 0x048A, 0x048A }, // L&       CYRILLIC CAPITAL LETTER SHORT I WITH TAIL
    Interval{ 0x048C, 0x048C }, // L&       CYRILLIC CAPITAL LETTER SEMISOFT SIGN
    Interval{ 0x048E, 0x048E }, // L&       CYRILLIC CAPITAL LETTER ER WITH TICK
    Interval{ 0x0490, 0x0490 }, // L&       CYRILLIC CAPITAL LETTER GHE WITH UPTURN
    Interval{ 0x0492, 0x0492 }, // L&       CYRILLIC CAPITAL LETTER GHE WITH STROKE
    Interval{ 0x0494, 0x0494 }, // L&       CYRILLIC CAPITAL LETTER GHE WITH MIDDLE HOOK
    Interval{ 0x0496, 0x0496 }, // L&       CYRILLIC CAPITAL LETTER ZHE WITH DESCENDER
    Interval{ 0x0498, 0x0498 }, // L&       CYRILLIC CAPITAL LETTER ZE WITH DESCENDER
    Interval{ 0x049A, 0x049A }, // L&       CYRILLIC CAPITAL LETTER KA WITH DESCENDER
    Interval{ 0x049C, 0x049C }, // L&       CYRILLIC CAPITAL LETTER KA WITH VERTICAL STROKE
    Interval{ 0x049E, 0x049E }, // L&       CYRILLIC CAPITAL LETTER KA WITH STROKE
    Interval{ 0x04A0, 0x04A0 }, // L&       CYRILLIC CAPITAL LETTER BASHKIR KA
    Interval{ 0x04A2, 0x04A2 }, // L&       CYRILLIC CAPITAL LETTER EN WITH DESCENDER
    Interval{ 0x04A4, 0x04A4 }, // L&       CYRILLIC CAPITAL LIGATURE EN GHE
    Interval{ 0x04A6, 0x04A6 }, // L&       CYRILLIC CAPITAL LETTER PE WITH MIDDLE HOOK
    Interval{ 0x04A8, 0x04A8 }, // L&       CYRILLIC CAPITAL LETTER ABKHASIAN HA
    Interval{ 0x04AA, 0x04AA }, // L&       CYRILLIC CAPITAL LETTER ES WITH DESCENDER
    Interval{ 0x04AC, 0x04AC }, // L&       CYRILLIC CAPITAL LETTER TE WITH DESCENDER
    Interval{ 0x04AE, 0x04AE }, // L&       CYRILLIC CAPITAL LETTER STRAIGHT U
    Interval{ 0x04B0, 0x04B0 }, // L&       CYRILLIC CAPITAL LETTER STRAIGHT U WITH STROKE
    Interval{ 0x04B2, 0x04B2 }, // L&       CYRILLIC CAPITAL LETTER HA WITH DESCENDER
    Interval{ 0x04B4, 0x04B4 }, // L&       CYRILLIC CAPITAL LIGATURE TE TSE
    Interval{ 0x04B6, 0x04B6 }, // L&       CYRILLIC CAPITAL LETTER CHE WITH DESCENDER
    Interval{ 0x04B8, 0x04B8 }, // L&       CYRILLIC CAPITAL LETTER CHE WITH VERTICAL STROKE
    Interval{ 0x04BA, 0x04BA }, // L&       CYRILLIC CAPITAL LETTER SHHA
    Interval{ 0x04BC, 0x04BC }, // L&       CYRILLIC CAPITAL LETTER ABKHASIAN CHE
    Interval{ 0x04BE, 0x04BE }, // L&       CYRILLIC CAPITAL LETTER ABKHASIAN CHE WITH DESCENDER
    Interval{ 0x04C0, 0x04C1 }, // L&   [2] CYRILLIC LETTER PALOCHKA..CYRILLIC CAPITAL LETTER ZHE WITH BREVE
    Interval{ 0x04C3, 0x04C3 }, // L&       CYRILLIC CAPITAL LETTER KA WITH HOOK
    Interval{ 0x04C5, 0x04C5 }, // L&       CYRILLIC CAPITAL LETTER EL WITH TAIL
    Interval{ 0x04C7, 0x04C7 }, // L&       CYRILLIC CAPITAL LETTER EN WITH HOOK
    Interval{ 0x04C9, 0x04C9 }, // L&       CYRILLIC CAPITAL LETTER EN WITH TAIL
    Interval{ 0x04CB, 0x04CB }, // L&       CYRILLIC CAPITAL LETTER KHAKASSIAN CHE
    Interval{ 0x04CD, 0x04CD }, // L&       CYRILLIC CAPITAL LETTER EM WITH TAIL
    Interval{ 0x04D0, 0x04D0 }, // L&       CYRILLIC CAPITAL LETTER A WITH BREVE
    Interval{ 0x04D2, 0x04D2 }, // L&       CYRILLIC CAPITAL LETTER A WITH DIAERESIS
    Interval{ 0x04D4, 0x04D4 }, // L&       CYRILLIC CAPITAL LIGATURE A IE
    Interval{ 0x04D6, 0x04D6 }, // L&       CYRILLIC CAPITAL LETTER IE WITH BREVE
    Interval{ 0x04D8, 0x04D8 }, // L&       CYRILLIC CAPITAL LETTER SCHWA
    Interval{ 0x04DA, 0x04DA }, // L&       CYRILLIC CAPITAL LETTER SCHWA WITH DIAERESIS
    Interval{ 0x04DC, 0x04DC }, // L&       CYRILLIC CAPITAL LETTER ZHE WITH DIAERESIS
    Interval{ 0x04DE, 0x04DE }, // L&       CYRILLIC CAPITAL LETTER ZE WITH DIAERESIS
    Interval{ 0x04E0, 0x04E0 }, // L&       CYRILLIC CAPITAL LETTER ABKHASIAN DZE
    Interval{ 0x04E2, 0x04E2 }, // L&       CYRILLIC CAPITAL LETTER I WITH MACRON
    Interval{ 0x04E4, 0x04E4 }, // L&       CYRILLIC CAPITAL LETTER I WITH DIAERESIS
    Interval{ 0x04E6, 0x04E6 }, // L&       CYRILLIC CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x04E8, 0x04E8 }, // L&       CYRILLIC CAPITAL LETTER BARRED O
    Interval{ 0x04EA, 0x04EA }, // L&       CYRILLIC CAPITAL LETTER BARRED O WITH DIAERESIS
    Interval{ 0x04EC, 0x04EC }, // L&       CYRILLIC CAPITAL LETTER E WITH DIAERESIS
    Interval{ 0x04EE, 0x04EE }, // L&       CYRILLIC CAPITAL LETTER U WITH MACRON
    Interval{ 0x04F0, 0x04F0 }, // L&       CYRILLIC CAPITAL LETTER U WITH DIAERESIS
    Interval{ 0x04F2, 0x04F2 }, // L&       CYRILLIC CAPITAL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x04F4, 0x04F4 }, // L&       CYRILLIC CAPITAL LETTER CHE WITH DIAERESIS
    Interval{ 0x04F6, 0x04F6 }, // L&       CYRILLIC CAPITAL LETTER GHE WITH DESCENDER
    Interval{ 0x04F8, 0x04F8 }, // L&       CYRILLIC CAPITAL LETTER YERU WITH DIAERESIS
    Interval{ 0x04FA, 0x04FA }, // L&       CYRILLIC CAPITAL LETTER GHE WITH STROKE AND HOOK
    Interval{ 0x04FC, 0x04FC }, // L&       CYRILLIC CAPITAL LETTER HA WITH HOOK
    Interval{ 0x04FE, 0x04FE }, // L&       CYRILLIC CAPITAL LETTER HA WITH STROKE
    Interval{ 0x0500, 0x0500 }, // L&       CYRILLIC CAPITAL LETTER KOMI DE
    Interval{ 0x0502, 0x0502 }, // L&       CYRILLIC CAPITAL LETTER KOMI DJE
    Interval{ 0x0504, 0x0504 }, // L&       CYRILLIC CAPITAL LETTER KOMI ZJE
    Interval{ 0x0506, 0x0506 }, // L&       CYRILLIC CAPITAL LETTER KOMI DZJE
    Interval{ 0x0508, 0x0508 }, // L&       CYRILLIC CAPITAL LETTER KOMI LJE
    Interval{ 0x050A, 0x050A }, // L&       CYRILLIC CAPITAL LETTER KOMI NJE
    Interval{ 0x050C, 0x050C }, // L&       CYRILLIC CAPITAL LETTER KOMI SJE
    Interval{ 0x050E, 0x050E }, // L&       CYRILLIC CAPITAL LETTER KOMI TJE
    Interval{ 0x0510, 0x0510 }, // L&       CYRILLIC CAPITAL LETTER REVERSED ZE
    Interval{ 0x0512, 0x0512 }, // L&       CYRILLIC CAPITAL LETTER EL WITH HOOK
    Interval{ 0x0514, 0x0514 }, // L&       CYRILLIC CAPITAL LETTER LHA
    Interval{ 0x0516, 0x0516 }, // L&       CYRILLIC CAPITAL LETTER RHA
    Interval{ 0x0518, 0x0518 }, // L&       CYRILLIC CAPITAL LETTER YAE
    Interval{ 0x051A, 0x051A }, // L&       CYRILLIC CAPITAL LETTER QA
    Interval{ 0x051C, 0x051C }, // L&       CYRILLIC CAPITAL LETTER WE
    Interval{ 0x051E, 0x051E }, // L&       CYRILLIC CAPITAL LETTER ALEUT KA
    Interval{ 0x0520, 0x0520 }, // L&       CYRILLIC CAPITAL LETTER EL WITH MIDDLE HOOK
    Interval{ 0x0522, 0x0522 }, // L&       CYRILLIC CAPITAL LETTER EN WITH MIDDLE HOOK
    Interval{ 0x0524, 0x0524 }, // L&       CYRILLIC CAPITAL LETTER PE WITH DESCENDER
    Interval{ 0x0526, 0x0526 }, // L&       CYRILLIC CAPITAL LETTER SHHA WITH DESCENDER
    Interval{ 0x0528, 0x0528 }, // L&       CYRILLIC CAPITAL LETTER EN WITH LEFT HOOK
    Interval{ 0x052A, 0x052A }, // L&       CYRILLIC CAPITAL LETTER DZZHE
    Interval{ 0x052C, 0x052C }, // L&       CYRILLIC CAPITAL LETTER DCHE
    Interval{ 0x052E, 0x052E }, // L&       CYRILLIC CAPITAL LETTER EL WITH DESCENDER
    Interval{ 0x0531, 0x0556 }, // L&  [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Interval{ 0x0587, 0x0587 }, // L&       ARMENIAN SMALL LIGATURE ECH YIWN
    Interval{ 0x10A0, 0x10C5 }, // L&  [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Interval{ 0x10C7, 0x10C7 }, // L&       GEORGIAN CAPITAL LETTER YN
    Interval{ 0x10CD, 0x10CD }, // L&       GEORGIAN CAPITAL LETTER AEN
    Interval{ 0x13F8, 0x13FD }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1C80, 0x1C89 }, // L&  [10] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC CAPITAL LETTER TJE
    Interval{ 0x1C90, 0x1CBA }, // L&  [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Interval{ 0x1CBD, 0x1CBF }, // L&   [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Interval{ 0x1E00, 0x1E00 }, // L&       LATIN CAPITAL LETTER A WITH RING BELOW
    Interval{ 0x1E02, 0x1E02 }, // L&       LATIN CAPITAL LETTER B WITH DOT ABOVE
    Interval{ 0x1E04, 0x1E04 }, // L&       LATIN CAPITAL LETTER B WITH DOT BELOW
    Interval{ 0x1E06, 0x1E06 }, // L&       LATIN CAPITAL LETTER B WITH LINE BELOW
    Interval{ 0x1E08, 0x1E08 }, // L&       LATIN CAPITAL LETTER C WITH CEDILLA AND ACUTE
    Interval{ 0x1E0A, 0x1E0A }, // L&       LATIN CAPITAL LETTER D WITH DOT ABOVE
    Interval{ 0x1E0C, 0x1E0C }, // L&       LATIN CAPITAL LETTER D WITH DOT BELOW
    Interval{ 0x1E0E, 0x1E0E }, // L&       LATIN CAPITAL LETTER D WITH LINE BELOW
    Interval{ 0x1E10, 0x1E10 }, // L&       LATIN CAPITAL LETTER D WITH CEDILLA
    Interval{ 0x1E12, 0x1E12 }, // L&       LATIN CAPITAL LETTER D WITH CIRCUMFLEX BELOW
    Interval{ 0x1E14, 0x1E14 }, // L&       LATIN CAPITAL LETTER E WITH MACRON AND GRAVE
    Interval{ 0x1E16, 0x1E16 }, // L&       LATIN CAPITAL LETTER E WITH MACRON AND ACUTE
    Interval{ 0x1E18, 0x1E18 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX BELOW
    Interval{ 0x1E1A, 0x1E1A }, // L&       LATIN CAPITAL LETTER E WITH TILDE BELOW
    Interval{ 0x1E1C, 0x1E1C }, // L&       LATIN CAPITAL LETTER E WITH CEDILLA AND BREVE
    Interval{ 0x1E1E, 0x1E1E }, // L&       LATIN CAPITAL LETTER F WITH DOT ABOVE
    Interval{ 0x1E20, 0x1E20 }, // L&       LATIN CAPITAL LETTER G WITH MACRON
    Interval{ 0x1E22, 0x1E22 }, // L&       LATIN CAPITAL LETTER H WITH DOT ABOVE
    Interval{ 0x1E24, 0x1E24 }, // L&       LATIN CAPITAL LETTER H WITH DOT BELOW
    Interval{ 0x1E26, 0x1E26 }, // L&       LATIN CAPITAL LETTER H WITH DIAERESIS
    Interval{ 0x1E28, 0x1E28 }, // L&       LATIN CAPITAL LETTER H WITH CEDILLA
    Interval{ 0x1E2A, 0x1E2A }, // L&       LATIN CAPITAL LETTER H WITH BREVE BELOW
    Interval{ 0x1E2C, 0x1E2C }, // L&       LATIN CAPITAL LETTER I WITH TILDE BELOW
    Interval{ 0x1E2E, 0x1E2E }, // L&       LATIN CAPITAL LETTER I WITH DIAERESIS AND ACUTE
    Interval{ 0x1E30, 0x1E30 }, // L&       LATIN CAPITAL LETTER K WITH ACUTE
    Interval{ 0x1E32, 0x1E32 }, // L&       LATIN CAPITAL LETTER K WITH DOT BELOW
    Interval{ 0x1E34, 0x1E34 }, // L&       LATIN CAPITAL LETTER K WITH LINE BELOW
    Interval{ 0x1E36, 0x1E36 }, // L&       LATIN CAPITAL LETTER L WITH DOT BELOW
    Interval{ 0x1E38, 0x1E38 }, // L&       LATIN CAPITAL LETTER L WITH DOT BELOW AND MACRON
    Interval{ 0x1E3A, 0x1E3A }, // L&       LATIN CAPITAL LETTER L WITH LINE BELOW
    Interval{ 0x1E3C, 0x1E3C }, // L&       LATIN CAPITAL LETTER L WITH CIRCUMFLEX BELOW
    Interval{ 0x1E3E, 0x1E3E }, // L&       LATIN CAPITAL LETTER M WITH ACUTE
    Interval{ 0x1E40, 0x1E40 }, // L&       LATIN CAPITAL LETTER M WITH DOT ABOVE
    Interval{ 0x1E42, 0x1E42 }, // L&       LATIN CAPITAL LETTER M WITH DOT BELOW
    Interval{ 0x1E44, 0x1E44 }, // L&       LATIN CAPITAL LETTER N WITH DOT ABOVE
    Interval{ 0x1E46, 0x1E46 }, // L&       LATIN CAPITAL LETTER N WITH DOT BELOW
    Interval{ 0x1E48, 0x1E48 }, // L&       LATIN CAPITAL LETTER N WITH LINE BELOW
    Interval{ 0x1E4A, 0x1E4A }, // L&       LATIN CAPITAL LETTER N WITH CIRCUMFLEX BELOW
    Interval{ 0x1E4C, 0x1E4C }, // L&       LATIN CAPITAL LETTER O WITH TILDE AND ACUTE
    Interval{ 0x1E4E, 0x1E4E }, // L&       LATIN CAPITAL LETTER O WITH TILDE AND DIAERESIS
    Interval{ 0x1E50, 0x1E50 }, // L&       LATIN CAPITAL LETTER O WITH MACRON AND GRAVE
    Interval{ 0x1E52, 0x1E52 }, // L&       LATIN CAPITAL LETTER O WITH MACRON AND ACUTE
    Interval{ 0x1E54, 0x1E54 }, // L&       LATIN CAPITAL LETTER P WITH ACUTE
    Interval{ 0x1E56, 0x1E56 }, // L&       LATIN CAPITAL LETTER P WITH DOT ABOVE
    Interval{ 0x1E58, 0x1E58 }, // L&       LATIN CAPITAL LETTER R WITH DOT ABOVE
    Interval{ 0x1E5A, 0x1E5A }, // L&       LATIN CAPITAL LETTER R WITH DOT BELOW
    Interval{ 0x1E5C, 0x1E5C }, // L&       LATIN CAPITAL LETTER R WITH DOT BELOW AND MACRON
    Interval{ 0x1E5E, 0x1E5E }, // L&       LATIN CAPITAL LETTER R WITH LINE BELOW
    Interval{ 0x1E60, 0x1E60 }, // L&       LATIN CAPITAL LETTER S WITH DOT ABOVE
    Interval{ 0x1E62, 0x1E62 }, // L&       LATIN CAPITAL LETTER S WITH DOT BELOW
    Interval{ 0x1E64, 0x1E64 }, // L&       LATIN CAPITAL LETTER S WITH ACUTE AND DOT ABOVE
    Interval{ 0x1E66, 0x1E66 }, // L&       LATIN CAPITAL LETTER S WITH CARON AND DOT ABOVE
    Interval{ 0x1E68, 0x1E68 }, // L&       LATIN CAPITAL LETTER S WITH DOT BELOW AND DOT ABOVE
    Interval{ 0x1E6A, 0x1E6A }, // L&       LATIN CAPITAL LETTER T WITH DOT ABOVE
    Interval{ 0x1E6C, 0x1E6C }, // L&       LATIN CAPITAL LETTER T WITH DOT BELOW
    Interval{ 0x1E6E, 0x1E6E }, // L&       LATIN CAPITAL LETTER T WITH LINE BELOW
    Interval{ 0x1E70, 0x1E70 }, // L&       LATIN CAPITAL LETTER T WITH CIRCUMFLEX BELOW
    Interval{ 0x1E72, 0x1E72 }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS BELOW
    Interval{ 0x1E74, 0x1E74 }, // L&       LATIN CAPITAL LETTER U WITH TILDE BELOW
    Interval{ 0x1E76, 0x1E76 }, // L&       LATIN CAPITAL LETTER U WITH CIRCUMFLEX BELOW
    Interval{ 0x1E78, 0x1E78 }, // L&       LATIN CAPITAL LETTER U WITH TILDE AND ACUTE
    Interval{ 0x1E7A, 0x1E7A }, // L&       LATIN CAPITAL LETTER U WITH MACRON AND DIAERESIS
    Interval{ 0x1E7C, 0x1E7C }, // L&       LATIN CAPITAL LETTER V WITH TILDE
    Interval{ 0x1E7E, 0x1E7E }, // L&       LATIN CAPITAL LETTER V WITH DOT BELOW
    Interval{ 0x1E80, 0x1E80 }, // L&       LATIN CAPITAL LETTER W WITH GRAVE
    Interval{ 0x1E82, 0x1E82 }, // L&       LATIN CAPITAL LETTER W WITH ACUTE
    Interval{ 0x1E84, 0x1E84 }, // L&       LATIN CAPITAL LETTER W WITH DIAERESIS
    Interval{ 0x1E86, 0x1E86 }, // L&       LATIN CAPITAL LETTER W WITH DOT ABOVE
    Interval{ 0x1E88, 0x1E88 }, // L&       LATIN CAPITAL LETTER W WITH DOT BELOW
    Interval{ 0x1E8A, 0x1E8A }, // L&       LATIN CAPITAL LETTER X WITH DOT ABOVE
    Interval{ 0x1E8C, 0x1E8C }, // L&       LATIN CAPITAL LETTER X WITH DIAERESIS
    Interval{ 0x1E8E, 0x1E8E }, // L&       LATIN CAPITAL LETTER Y WITH DOT ABOVE
    Interval{ 0x1E90, 0x1E90 }, // L&       LATIN CAPITAL LETTER Z WITH CIRCUMFLEX
    Interval{ 0x1E92, 0x1E92 }, // L&       LATIN CAPITAL LETTER Z WITH DOT BELOW
    Interval{ 0x1E94, 0x1E94 }, // L&       LATIN CAPITAL LETTER Z WITH LINE BELOW
    Interval{ 0x1E9A, 0x1E9B }, // L&   [2] LATIN SMALL LETTER A WITH RIGHT HALF RING..LATIN SMALL LETTER LONG S WITH DOT ABOVE
    Interval{ 0x1E9E, 0x1E9E }, // L&       LATIN CAPITAL LETTER SHARP S
    Interval{ 0x1EA0, 0x1EA0 }, // L&       LATIN CAPITAL LETTER A WITH DOT BELOW
    Interval{ 0x1EA2, 0x1EA2 }, // L&       LATIN CAPITAL LETTER A WITH HOOK ABOVE
    Interval{ 0x1EA4, 0x1EA4 }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EA6, 0x1EA6 }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EA8, 0x1EA8 }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EAA, 0x1EAA }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EAC, 0x1EAC }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EAE, 0x1EAE }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND ACUTE
    Interval{ 0x1EB0, 0x1EB0 }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND GRAVE
    Interval{ 0x1EB2, 0x1EB2 }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND HOOK ABOVE
    Interval{ 0x1EB4, 0x1EB4 }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND TILDE
    Interval{ 0x1EB6, 0x1EB6 }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND DOT BELOW
    Interval{ 0x1EB8, 0x1EB8 }, // L&       LATIN CAPITAL LETTER E WITH DOT BELOW
    Interval{ 0x1EBA, 0x1EBA }, // L&       LATIN CAPITAL LETTER E WITH HOOK ABOVE
    Interval{ 0x1EBC, 0x1EBC }, // L&       LATIN CAPITAL LETTER E WITH TILDE
    Interval{ 0x1EBE, 0x1EBE }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EC0, 0x1EC0 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EC2, 0x1EC2 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EC4, 0x1EC4 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EC6, 0x1EC6 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EC8, 0x1EC8 }, // L&       LATIN CAPITAL LETTER I WITH HOOK ABOVE
    Interval{ 0x1ECA, 0x1ECA }, // L&       LATIN CAPITAL LETTER I WITH DOT BELOW
    Interval{ 0x1ECC, 0x1ECC }, // L&       LATIN CAPITAL LETTER O WITH DOT BELOW
    Interval{ 0x1ECE, 0x1ECE }, // L&       LATIN CAPITAL LETTER O WITH HOOK ABOVE
    Interval{ 0x1ED0, 0x1ED0 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1ED2, 0x1ED2 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1ED4, 0x1ED4 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1ED6, 0x1ED6 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1ED8, 0x1ED8 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EDA, 0x1EDA }, // L&       LATIN CAPITAL LETTER O WITH HORN AND ACUTE
    Interval{ 0x1EDC, 0x1EDC }, // L&       LATIN CAPITAL LETTER O WITH HORN AND GRAVE
    Interval{ 0x1EDE, 0x1EDE }, // L&       LATIN CAPITAL LETTER O WITH HORN AND HOOK ABOVE
    Interval{ 0x1EE0, 0x1EE0 }, // L&       LATIN CAPITAL LETTER O WITH HORN AND TILDE
    Interval{ 0x1EE2, 0x1EE2 }, // L&       LATIN CAPITAL LETTER O WITH HORN AND DOT BELOW
    Interval{ 0x1EE4, 0x1EE4 }, // L&       LATIN CAPITAL LETTER U WITH DOT BELOW
    Interval{ 0x1EE6, 0x1EE6 }, // L&       LATIN CAPITAL LETTER U WITH HOOK ABOVE
    Interval{ 0x1EE8, 0x1EE8 }, // L&       LATIN CAPITAL LETTER U WITH HORN AND ACUTE
    Interval{ 0x1EEA, 0x1EEA }, // L&       LATIN CAPITAL LETTER U WITH HORN AND GRAVE
    Interval{ 0x1EEC, 0x1EEC }, // L&       LATIN CAPITAL LETTER U WITH HORN AND HOOK ABOVE
    Interval{ 0x1EEE, 0x1EEE }, // L&       LATIN CAPITAL LETTER U WITH HORN AND TILDE
    Interval{ 0x1EF0, 0x1EF0 }, // L&       LATIN CAPITAL LETTER U WITH HORN AND DOT BELOW
    Interval{ 0x1EF2, 0x1EF2 }, // L&       LATIN CAPITAL LETTER Y WITH GRAVE
    Interval{ 0x1EF4, 0x1EF4 }, // L&       LATIN CAPITAL LETTER Y WITH DOT BELOW
    Interval{ 0x1EF6, 0x1EF6 }, // L&       LATIN CAPITAL LETTER Y WITH HOOK ABOVE
    Interval{ 0x1EF8, 0x1EF8 }, // L&       LATIN CAPITAL LETTER Y WITH TILDE
    Interval{ 0x1EFA, 0x1EFA }, // L&       LATIN CAPITAL LETTER MIDDLE-WELSH LL
    Interval{ 0x1EFC, 0x1EFC }, // L&       LATIN CAPITAL LETTER MIDDLE-WELSH V
    Interval{ 0x1EFE, 0x1EFE }, // L&       LATIN CAPITAL LETTER Y WITH LOOP
    Interval{ 0x1F08, 0x1F0F }, // L&   [8] GREEK CAPITAL LETTER ALPHA WITH PSILI..GREEK CAPITAL LETTER ALPHA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F18, 0x1F1D }, // L&   [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F28, 0x1F2F }, // L&   [8] GREEK CAPITAL LETTER ETA WITH PSILI..GREEK CAPITAL LETTER ETA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F38, 0x1F3F }, // L&   [8] GREEK CAPITAL LETTER IOTA WITH PSILI..GREEK CAPITAL LETTER IOTA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F48, 0x1F4D }, // L&   [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F59, 0x1F59 }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA
    Interval{ 0x1F5B, 0x1F5B }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Interval{ 0x1F5D, 0x1F5D }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Interval{ 0x1F5F, 0x1F5F }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F68, 0x1F6F }, // L&   [8] GREEK CAPITAL LETTER OMEGA WITH PSILI..GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F80, 0x1FAF }, // L&  [48] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI
    Interval{ 0x1FB2, 0x1FB4 }, // L&   [3] GREEK SMALL LETTER ALPHA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB7, 0x1FBC }, // L&   [6] GREEK SMALL LETTER ALPHA WITH PERISPOMENI AND YPOGEGRAMMENI..GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Interval{ 0x1FC2, 0x1FC4 }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC7, 0x1FCC }, // L&   [6] GREEK SMALL LETTER ETA WITH PERISPOMENI AND YPOGEGRAMMENI..GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Interval{ 0x1FD8, 0x1FDB }, // L&   [4] GREEK CAPITAL LETTER IOTA WITH VRACHY..GREEK CAPITAL LETTER IOTA WITH OXIA
    Interval{ 0x1FE8, 0x1FEC }, // L&   [5] GREEK CAPITAL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Interval{ 0x1FF2, 0x1FF4 }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF7, 0x1FFC }, // L&   [6] GREEK SMALL LETTER OMEGA WITH PERISPOMENI AND YPOGEGRAMMENI..GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Interval{ 0x2126, 0x2126 }, // L&       OHM SIGN
    Interval{ 0x212A, 0x212B }, // L&   [2] KELVIN SIGN..ANGSTROM SIGN
    Interval{ 0x2132, 0x2132 }, // L&       TURNED CAPITAL F
    Interval{ 0x2160, 0x216F }, // Nl  [16] ROMAN NUMERAL ONE..ROMAN NUMERAL ONE THOUSAND
    Interval{ 0x2183, 0x2183 }, // L&       ROMAN NUMERAL REVERSED ONE HUNDRED
    Interval{ 0x24B6, 0x24CF }, // So  [26] CIRCLED LATIN CAPITAL LETTER A..CIRCLED LATIN CAPITAL LETTER Z
    Interval{ 0x2C00, 0x2C2F }, // L&  [48] GLAGOLITIC CAPITAL LETTER AZU..GLAGOLITIC CAPITAL LETTER CAUDATE CHRIVI
    Interval{ 0x2C60, 0x2C60 }, // L&       LATIN CAPITAL LETTER L WITH DOUBLE BAR
    Interval{ 0x2C62, 0x2C64 }, // L&   [3] LATIN CAPITAL LETTER L WITH MIDDLE TILDE..LATIN CAPITAL LETTER R WITH TAIL
    Interval{ 0x2C67, 0x2C67 }, // L&       LATIN CAPITAL LETTER H WITH DESCENDER
    Interval{ 0x2C69, 0x2C69 }, // L&       LATIN CAPITAL LETTER K WITH DESCENDER
    Interval{ 0x2C6B, 0x2C6B }, // L&       LATIN CAPITAL LETTER Z WITH DESCENDER
    Interval{ 0x2C6D, 0x2C70 }, // L&   [4] LATIN CAPITAL LETTER ALPHA..LATIN CAPITAL LETTER TURNED ALPHA
    Interval{ 0x2C72, 0x2C72 }, // L&       LATIN CAPITAL LETTER W WITH HOOK
    Interval{ 0x2C75, 0x2C75 }, // L&       LATIN CAPITAL LETTER HALF H
    Interval{ 0x2C7E, 0x2C80 }, // L&   [3] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC CAPITAL LETTER ALFA
    Interval{ 0x2C82, 0x2C82 }, // L&       COPTIC CAPITAL LETTER VIDA
    Interval{ 0x2C84, 0x2C84 }, // L&       COPTIC CAPITAL LETTER GAMMA
    Interval{ 0x2C86, 0x2C86 }, // L&       COPTIC CAPITAL LETTER DALDA
    Interval{ 0x2C88, 0x2C88 }, // L&       COPTIC CAPITAL LETTER EIE
    Interval{ 0x2C8A, 0x2C8A }, // L&       COPTIC CAPITAL LETTER SOU
    Interval{ 0x2C8C, 0x2C8C }, // L&       COPTIC CAPITAL LETTER ZATA
    Interval{ 0x2C8E, 0x2C8E }, // L&       COPTIC CAPITAL LETTER HATE
    Interval{ 0x2C90, 0x2C90 }, // L&       COPTIC CAPITAL LETTER THETHE
    Interval{ 0x2C92, 0x2C92 }, // L&       COPTIC CAPITAL LETTER IAUDA
    Interval{ 0x2C94, 0x2C94 }, // L&       COPTIC CAPITAL LETTER KAPA
    Interval{ 0x2C96, 0x2C96 }, // L&       COPTIC CAPITAL LETTER LAULA
    Interval{ 0x2C98, 0x2C98 }, // L&       COPTIC CAPITAL LETTER MI
    Interval{ 0x2C9A, 0x2C9A }, // L&       COPTIC CAPITAL LETTER NI
    Interval{ 0x2C9C, 0x2C9C }, // L&       COPTIC CAPITAL LETTER KSI
    Interval{ 0x2C9E, 0x2C9E }, // L&       COPTIC CAPITAL LETTER O
    Interval{ 0x2CA0, 0x2CA0 }, // L&       COPTIC CAPITAL LETTER PI
    Interval{ 0x2CA2, 0x2CA2 }, // L&       COPTIC CAPITAL LETTER RO
    Interval{ 0x2CA4, 0x2CA4 }, // L&       COPTIC CAPITAL LETTER SIMA
    Interval{ 0x2CA6, 0x2CA6 }, // L&       COPTIC CAPITAL LETTER TAU
    Interval{ 0x2CA8, 0x2CA8 }, // L&       COPTIC CAPITAL LETTER UA
    Interval{ 0x2CAA, 0x2CAA }, // L&       COPTIC CAPITAL LETTER FI
    Interval{ 0x2CAC, 0x2CAC }, // L&       COPTIC CAPITAL LETTER KHI
    Interval{ 0x2CAE, 0x2CAE }, // L&       COPTIC CAPITAL LETTER PSI
    Interval{ 0x2CB0, 0x2CB0 }, // L&       COPTIC CAPITAL LETTER OOU
    Interval{ 0x2CB2, 0x2CB2 }, // L&       COPTIC CAPITAL LETTER DIALECT-P ALEF
    Interval{ 0x2CB4, 0x2CB4 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC AIN
    Interval{ 0x2CB6, 0x2CB6 }, // L&       COPTIC CAPITAL LETTER CRYPTOGRAMMIC EIE
    Interval{ 0x2CB8, 0x2CB8 }, // L&       COPTIC CAPITAL LETTER DIALECT-P KAPA
    Interval{ 0x2CBA, 0x2CBA }, // L&       COPTIC CAPITAL LETTER DIALECT-P NI
    Interval{ 0x2CBC, 0x2CBC }, // L&       COPTIC CAPITAL LETTER CRYPTOGRAMMIC NI
    Interval{ 0x2CBE, 0x2CBE }, // L&       COPTIC CAPITAL LETTER OLD COPTIC OOU
    Interval{ 0x2CC0, 0x2CC0 }, // L&       COPTIC CAPITAL LETTER SAMPI
    Interval{ 0x2CC2, 0x2CC2 }, // L&       COPTIC CAPITAL LETTER CROSSED SHEI
    Interval{ 0x2CC4, 0x2CC4 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC SHEI
    Interval{ 0x2CC6, 0x2CC6 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC ESH
    Interval{ 0x2CC8, 0x2CC8 }, // L&       COPTIC CAPITAL LETTER AKHMIMIC KHEI
    Interval{ 0x2CCA, 0x2CCA }, // L&       COPTIC CAPITAL LETTER DIALECT-P HORI
    Interval{ 0x2CCC, 0x2CCC }, // L&       COPTIC CAPITAL LETTER OLD COPTIC HORI
    Interval{ 0x2CCE, 0x2CCE }, // L&       COPTIC CAPITAL LETTER OLD COPTIC HA
    Interval{ 0x2CD0, 0x2CD0 }, // L&       COPTIC CAPITAL LETTER L-SHAPED HA
    Interval{ 0x2CD2, 0x2CD2 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC HEI
    Interval{ 0x2CD4, 0x2CD4 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC HAT
    Interval{ 0x2CD6, 0x2CD6 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC GANGIA
    Interval{ 0x2CD8, 0x2CD8 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC DJA
    Interval{ 0x2CDA, 0x2CDA }, // L&       COPTIC CAPITAL LETTER OLD COPTIC SHIMA
    Interval{ 0x2CDC, 0x2CDC }, // L&       COPTIC CAPITAL LETTER OLD NUBIAN SHIMA
    Interval{ 0x2CDE, 0x2CDE }, // L&       COPTIC CAPITAL LETTER OLD NUBIAN NGI
    Interval{ 0x2CE0, 0x2CE0 }, // L&       COPTIC CAPITAL LETTER OLD NUBIAN NYI
    Interval{ 0x2CE2, 0x2CE2 }, // L&       COPTIC CAPITAL LETTER OLD NUBIAN WAU
    Interval{ 0x2CEB, 0x2CEB }, // L&       COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI
    Interval{ 0x2CED, 0x2CED }, // L&       COPTIC CAPITAL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF2, 0x2CF2 }, // L&       COPTIC CAPITAL LETTER BOHAIRIC KHEI
    Interval{ 0xA640, 0xA640 }, // L&       CYRILLIC CAPITAL LETTER ZEMLYA
    Interval{ 0xA642, 0xA642 }, // L&       CYRILLIC CAPITAL LETTER DZELO
    Interval{ 0xA644, 0xA644 }, // L&       CYRILLIC CAPITAL LETTER REVERSED DZE
    Interval{ 0xA646, 0xA646 }, // L&       CYRILLIC CAPITAL LETTER IOTA
    Interval{ 0xA648, 0xA648 }, // L&       CYRILLIC CAPITAL LETTER DJERV
    Interval{ 0xA64A, 0xA64A }, // L&       CYRILLIC CAPITAL LETTER MONOGRAPH UK
    Interval{ 0xA64C, 0xA64C }, // L&       CYRILLIC CAPITAL LETTER BROAD OMEGA
    Interval{ 0xA64E, 0xA64E }, // L&       CYRILLIC CAPITAL LETTER NEUTRAL YER
    Interval{ 0xA650, 0xA650 }, // L&       CYRILLIC CAPITAL LETTER YERU WITH BACK YER
    Interval{ 0xA652, 0xA652 }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED YAT
    Interval{ 0xA654, 0xA654 }, // L&       CYRILLIC CAPITAL LETTER REVERSED YU
    Interval{ 0xA656, 0xA656 }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED A
    Interval{ 0xA658, 0xA658 }, // L&       CYRILLIC CAPITAL LETTER CLOSED LITTLE YUS
    Interval{ 0xA65A, 0xA65A }, // L&       CYRILLIC CAPITAL LETTER BLENDED YUS
    Interval{ 0xA65C, 0xA65C }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED CLOSED LITTLE YUS
    Interval{ 0xA65E, 0xA65E }, // L&       CYRILLIC CAPITAL LETTER YN
    Interval{ 0xA660, 0xA660 }, // L&       CYRILLIC CAPITAL LETTER REVERSED TSE
    Interval{ 0xA662, 0xA662 }, // L&       CYRILLIC CAPITAL LETTER SOFT DE
    Interval{ 0xA664, 0xA664 }, // L&       CYRILLIC CAPITAL LETTER SOFT EL
    Interval{ 0xA666, 0xA666 }, // L&       CYRILLIC CAPITAL LETTER SOFT EM
    Interval{ 0xA668, 0xA668 }, // L&       CYRILLIC CAPITAL LETTER MONOCULAR O
    Interval{ 0xA66A, 0xA66A }, // L&       CYRILLIC CAPITAL LETTER BINOCULAR O
    Interval{ 0xA66C, 0xA66C }, // L&       CYRILLIC CAPITAL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA680, 0xA680 }, // L&       CYRILLIC CAPITAL LETTER DWE
    Interval{ 0xA682, 0xA682 }, // L&       CYRILLIC CAPITAL LETTER DZWE
    Interval{ 0xA684, 0xA684 }, // L&       CYRILLIC CAPITAL LETTER ZHWE
    Interval{ 0xA686, 0xA686 }, // L&       CYRILLIC CAPITAL LETTER CCHE
    Interval{ 0xA688, 0xA688 }, // L&       CYRILLIC CAPITAL LETTER DZZE
    Interval{ 0xA68A, 0xA68A }, // L&       CYRILLIC CAPITAL LETTER TE WITH MIDDLE HOOK
    Interval{ 0xA68C, 0xA68C }, // L&       CYRILLIC CAPITAL LETTER TWE
    Interval{ 0xA68E, 0xA68E }, // L&       CYRILLIC CAPITAL LETTER TSWE
    Interval{ 0xA690, 0xA690 }, // L&       CYRILLIC CAPITAL LETTER TSSE
    Interval{ 0xA692, 0xA692 }, // L&       CYRILLIC CAPITAL LETTER TCHE
    Interval{ 0xA694, 0xA694 }, // L&       CYRILLIC CAPITAL LETTER HWE
    Interval{ 0xA696, 0xA696 }, // L&       CYRILLIC CAPITAL LETTER SHWE
    Interval{ 0xA698, 0xA698 }, // L&       CYRILLIC CAPITAL LETTER DOUBLE O
    Interval{ 0xA69A, 0xA69A }, // L&       CYRILLIC CAPITAL LETTER CROSSED O
    Interval{ 0xA722, 0xA722 }, // L&       LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF
    Interval{ 0xA724, 0xA724 }, // L&       LATIN CAPITAL LETTER EGYPTOLOGICAL AIN
    Interval{ 0xA726, 0xA726 }, // L&       LATIN CAPITAL LETTER HENG
    Interval{ 0xA728, 0xA728 }, // L&       LATIN CAPITAL LETTER TZ
    Interval{ 0xA72A, 0xA72A }, // L&       LATIN CAPITAL LETTER TRESILLO
    Interval{ 0xA72C, 0xA72C }, // L&       LATIN CAPITAL LETTER CUATRILLO
    Interval{ 0xA72E, 0xA72E }, // L&       LATIN CAPITAL LETTER CUATRILLO WITH COMMA
    Interval{ 0xA732, 0xA732 }, // L&       LATIN CAPITAL LETTER AA
    Interval{ 0xA734, 0xA734 }, // L&       LATIN CAPITAL LETTER AO
    Interval{ 0xA736, 0xA736 }, // L&       LATIN CAPITAL LETTER AU
    Interval{ 0xA738, 0xA738 }, // L&       LATIN CAPITAL LETTER AV
    Interval{ 0xA73A, 0xA73A }, // L&       LATIN CAPITAL LETTER AV WITH HORIZONTAL BAR
    Interval{ 0xA73C, 0xA73C }, // L&       LATIN CAPITAL LETTER AY
    Interval{ 0xA73E, 0xA73E }, // L&       LATIN CAPITAL LETTER REVERSED C WITH DOT
    Interval{ 0xA740, 0xA740 }, // L&       LATIN CAPITAL LETTER K WITH STROKE
    Interval{ 0xA742, 0xA742 }, // L&       LATIN CAPITAL LETTER K WITH DIAGONAL STROKE
    Interval{ 0xA744, 0xA744 }, // L&       LATIN CAPITAL LETTER K WITH STROKE AND DIAGONAL STROKE
    Interval{ 0xA746, 0xA746 }, // L&       LATIN CAPITAL LETTER BROKEN L
    Interval{ 0xA748, 0xA748 }, // L&       LATIN CAPITAL LETTER L WITH HIGH STROKE
    Interval{ 0xA74A, 0xA74A }, // L&       LATIN CAPITAL LETTER O WITH LONG STROKE OVERLAY
    Interval{ 0xA74C, 0xA74C }, // L&       LATIN CAPITAL LETTER O WITH LOOP
    Interval{ 0xA74E, 0xA74E }, // L&       LATIN CAPITAL LETTER OO
    Interval{ 0xA750, 0xA750 }, // L&       LATIN CAPITAL LETTER P WITH STROKE THROUGH DESCENDER
    Interval{ 0xA752, 0xA752 }, // L&       LATIN CAPITAL LETTER P WITH FLOURISH
    Interval{ 0xA754, 0xA754 }, // L&       LATIN CAPITAL LETTER P WITH SQUIRREL TAIL
    Interval{ 0xA756, 0xA756 }, // L&       LATIN CAPITAL LETTER Q WITH STROKE THROUGH DESCENDER
    Interval{ 0xA758, 0xA758 }, // L&       LATIN CAPITAL LETTER Q WITH DIAGONAL STROKE
    Interval{ 0xA75A, 0xA75A }, // L&       LATIN CAPITAL LETTER R ROTUNDA
    Interval{ 0xA75C, 0xA75C }, // L&       LATIN CAPITAL LETTER RUM ROTUNDA
    Interval{ 0xA75E, 0xA75E }, // L&       LATIN CAPITAL LETTER V WITH DIAGONAL STROKE
    Interval{ 0xA760, 0xA760 }, // L&       LATIN CAPITAL LETTER VY
    Interval{ 0xA762, 0xA762 }, // L&       LATIN CAPITAL LETTER VISIGOTHIC Z
    Interval{ 0xA764, 0xA764 }, // L&       LATIN CAPITAL LETTER THORN WITH STROKE
    Interval{ 0xA766, 0xA766 }, // L&       LATIN CAPITAL LETTER THORN WITH STROKE THROUGH DESCENDER
    Interval{ 0xA768, 0xA768 }, // L&       LATIN CAPITAL LETTER VEND
    Interval{ 0xA76A, 0xA76A }, // L&       LATIN CAPITAL LETTER ET
    Interval{ 0xA76C, 0xA76C }, // L&       LATIN CAPITAL LETTER IS
    Interval{ 0xA76E, 0xA76E }, // L&       LATIN CAPITAL LETTER CON
    Interval{ 0xA779, 0xA779 }, // L&       LATIN CAPITAL LETTER INSULAR D
    Interval{ 0xA77B, 0xA77B }, // L&       LATIN CAPITAL LETTER INSULAR F
    Interval{ 0xA77D, 0xA77E }, // L&   [2] LATIN CAPITAL LETTER INSULAR G..LATIN CAPITAL LETTER TURNED INSULAR G
    Interval{ 0xA780, 0xA780 }, // L&       LATIN CAPITAL LETTER TURNED L
    Interval{ 0xA782, 0xA782 }, // L&       LATIN CAPITAL LETTER INSULAR R
    Interval{ 0xA784, 0xA784 }, // L&       LATIN CAPITAL LETTER INSULAR S
    Interval{ 0xA786, 0xA786 }, // L&       LATIN CAPITAL LETTER INSULAR T
    Interval{ 0xA78B, 0xA78B }, // L&       LATIN CAPITAL LETTER SALTILLO
    Interval{ 0xA78D, 0xA78D }, // L&       LATIN CAPITAL LETTER TURNED H
    Interval{ 0xA790, 0xA790 }, // L&       LATIN CAPITAL LETTER N WITH DESCENDER
    Interval{ 0xA792, 0xA792 }, // L&       LATIN CAPITAL LETTER C WITH BAR
    Interval{ 0xA796, 0xA796 }, // L&       LATIN CAPITAL LETTER B WITH FLOURISH
    Interval{ 0xA798, 0xA798 }, // L&       LATIN CAPITAL LETTER F WITH STROKE
    Interval{ 0xA79A, 0xA79A }, // L&       LATIN CAPITAL LETTER VOLAPUK AE
    Interval{ 0xA79C, 0xA79C }, // L&       LATIN CAPITAL LETTER VOLAPUK OE
    Interval{ 0xA79E, 0xA79E }, // L&       LATIN CAPITAL LETTER VOLAPUK UE
    Interval{ 0xA7A0, 0xA7A0 }, // L&       LATIN CAPITAL LETTER G WITH OBLIQUE STROKE
    Interval{ 0xA7A2, 0xA7A2 }, // L&       LATIN CAPITAL LETTER K WITH OBLIQUE STROKE
    Interval{ 0xA7A4, 0xA7A4 }, // L&       LATIN CAPITAL LETTER N WITH OBLIQUE STROKE
    Interval{ 0xA7A6, 0xA7A6 }, // L&       LATIN CAPITAL LETTER R WITH OBLIQUE STROKE
    Interval{ 0xA7A8, 0xA7A8 }, // L&       LATIN CAPITAL LETTER S WITH OBLIQUE STROKE
    Interval{ 0xA7AA, 0xA7AE }, // L&   [5] LATIN CAPITAL LETTER H WITH HOOK..LATIN CAPITAL LETTER SMALL CAPITAL I
    Interval{ 0xA7B0, 0xA7B4 }, // L&   [5] LATIN CAPITAL LETTER TURNED K..LATIN CAPITAL LETTER BETA
    Interval{ 0xA7B6, 0xA7B6 }, // L&       LATIN CAPITAL LETTER OMEGA
    Interval{ 0xA7B8, 0xA7B8 }, // L&       LATIN CAPITAL LETTER U WITH STROKE
    Interval{ 0xA7BA, 0xA7BA }, // L&       LATIN CAPITAL LETTER GLOTTAL A
    Interval{ 0xA7BC, 0xA7BC }, // L&       LATIN CAPITAL LETTER GLOTTAL I
    Interval{ 0xA7BE, 0xA7BE }, // L&       LATIN CAPITAL LETTER GLOTTAL U
    Interval{ 0xA7C0, 0xA7C0 }, // L&       LATIN CAPITAL LETTER OLD POLISH O
    Interval{ 0xA7C2, 0xA7C2 }, // L&       LATIN CAPITAL LETTER ANGLICANA W
    Interval{ 0xA7C4, 0xA7C7 }, // L&   [4] LATIN CAPITAL LETTER C WITH PALATAL HOOK..LATIN CAPITAL LETTER D WITH SHORT STROKE OVERLAY
    Interval{ 0xA7C9, 0xA7C9 }, // L&       LATIN CAPITAL LETTER S WITH SHORT STROKE OVERLAY
    Interval{ 0xA7CB, 0xA7CC }, // L&   [2] LATIN CAPITAL LETTER RAMS HORN..LATIN CAPITAL LETTER S WITH DIAGONAL STROKE
    Interval{ 0xA7CE, 0xA7CE }, // L&       LATIN CAPITAL LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0xA7D0, 0xA7D0 }, // L&       LATIN CAPITAL LETTER CLOSED INSULAR G
    Interval{ 0xA7D2, 0xA7D2 }, // L&       LATIN CAPITAL LETTER DOUBLE THORN
    Interval{ 0xA7D4, 0xA7D4 }, // L&       LATIN CAPITAL LETTER DOUBLE WYNN
    Interval{ 0xA7D6, 0xA7D6 }, // L&       LATIN CAPITAL LETTER MIDDLE SCOTS S
    Interval{ 0xA7D8, 0xA7D8 }, // L&       LATIN CAPITAL LETTER SIGMOID S
    Interval{ 0xA7DA, 0xA7DA }, // L&       LATIN CAPITAL LETTER LAMBDA
    Interval{ 0xA7DC, 0xA7DC }, // L&       LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Interval{ 0xA7F5, 0xA7F5 }, // L&       LATIN CAPITAL LETTER REVERSED HALF H
    Interval{ 0xAB70, 0xABBF }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xFB00, 0xFB06 }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFF21, 0xFF3A }, // L&  [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Interval{ 0x10400, 0x10427 }, // L&  [40] DESERET CAPITAL LETTER LONG I..DESERET CAPITAL LETTER EW
    Interval{ 0x104B0, 0x104D3 }, // L&  [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Interval{ 0x10570, 0x1057A }, // L&  [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Interval{ 0x1057C, 0x1058A }, // L&  [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Interval{ 0x1058C, 0x10592 }, // L&   [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Interval{ 0x10594, 0x10595 }, // L&   [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Interval{ 0x10C80, 0x10CB2 }, // L&  [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Interval{ 0x10D50, 0x10D65 }, // L&  [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Interval{ 0x118A0, 0x118BF }, // L&  [32] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI CAPITAL LETTER VIYO
    Interval{ 0x16E40, 0x16E5F }, // L&  [32] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN CAPITAL LETTER Y
    Interval{ 0x16EA0, 0x16EB8 }, // L&  [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Interval{ 0x1E900, 0x1E921 }, // L&  [34] ADLAM CAPITAL LETTER ALIF..ADLAM CAPITAL LETTER SHA
}; // }}}
auto static const Changes_When_Casemapped = std::array<Interval, 131>{ // {{{
    Interval{ 0x0041, 0x005A }, // L&  [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Interval{ 0x0061, 0x007A }, // L&  [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Interval{ 0x00B5, 0x00B5 }, // L&       MICRO SIGN
    Interval{ 0x00C0, 0x00D6 }, // L&  [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x00D8, 0x00F6 }, // L&  [31] LATIN CAPITAL LETTER O WITH STROKE..LATIN SMALL LETTER O WITH DIAERESIS
    Interval{ 0x00F8, 0x0137 }, // L&  [64] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER K WITH CEDILLA
    Interval{ 0x0139, 0x018C }, // L&  [84] LATIN CAPITAL LETTER L WITH ACUTE..LATIN SMALL LETTER D WITH TOPBAR
    Interval{ 0x018E, 0x01A9 }, // L&  [28] LATIN CAPITAL LETTER REVERSED E..LATIN CAPITAL LETTER ESH
    Interval{ 0x01AC, 0x01B9 }, // L&  [14] LATIN CAPITAL LETTER T WITH HOOK..LATIN SMALL LETTER EZH REVERSED
    Interval{ 0x01BC, 0x01BD }, // L&   [2] LATIN CAPITAL LETTER TONE FIVE..LATIN SMALL LETTER TONE FIVE
    Interval{ 0x01BF, 0x01BF }, // L&       LATIN LETTER WYNN
    Interval{ 0x01C4, 0x0220 }, // L&  [93] LATIN CAPITAL LETTER DZ WITH CARON..LATIN CAPITAL LETTER N WITH LONG RIGHT LEG
    Interval{ 0x0222, 0x0233 }, // L&  [18] LATIN CAPITAL LETTER OU..LATIN SMALL LETTER Y WITH MACRON
    Interval{ 0x023A, 0x0254 }, // L&  [27] LATIN CAPITAL LETTER A WITH STROKE..LATIN SMALL LETTER OPEN O
    Interval{ 0x0256, 0x0257 }, // L&   [2] LATIN SMALL LETTER D WITH TAIL..LATIN SMALL LETTER D WITH HOOK
    Interval{ 0x0259, 0x0259 }, // L&       LATIN SMALL LETTER SCHWA
    Interval{ 0x025B, 0x025C }, // L&   [2] LATIN SMALL LETTER OPEN E..LATIN SMALL LETTER REVERSED OPEN E
    Interval{ 0x0260, 0x0261 }, // L&   [2] LATIN SMALL LETTER G WITH HOOK..LATIN SMALL LETTER SCRIPT G
    Interval{ 0x0263, 0x0266 }, // L&   [4] LATIN SMALL LETTER GAMMA..LATIN SMALL LETTER H WITH HOOK
    Interval{ 0x0268, 0x026C }, // L&   [5] LATIN SMALL LETTER I WITH STROKE..LATIN SMALL LETTER L WITH BELT
    Interval{ 0x026F, 0x026F }, // L&       LATIN SMALL LETTER TURNED M
    Interval{ 0x0271, 0x0272 }, // L&   [2] LATIN SMALL LETTER M WITH HOOK..LATIN SMALL LETTER N WITH LEFT HOOK
    Interval{ 0x0275, 0x0275 }, // L&       LATIN SMALL LETTER BARRED O
    Interval{ 0x027D, 0x027D }, // L&       LATIN SMALL LETTER R WITH TAIL
    Interval{ 0x0280, 0x0280 }, // L&       LATIN LETTER SMALL CAPITAL R
    Interval{ 0x0282, 0x0283 }, // L&   [2] LATIN SMALL LETTER S WITH HOOK..LATIN SMALL LETTER ESH
    Interval{ 0x0287, 0x028C }, // L&   [6] LATIN SMALL LETTER TURNED T..LATIN SMALL LETTER TURNED V
    Interval{ 0x0292, 0x0292 }, // L&       LATIN SMALL LETTER EZH
    Interval{ 0x029D, 0x029E }, // L&   [2] LATIN SMALL LETTER J WITH CROSSED-TAIL..LATIN SMALL LETTER TURNED K
    Interval{ 0x0345, 0x0345 }, // Mn       COMBINING GREEK YPOGEGRAMMENI
    Interval{ 0x0370, 0x0373 }, // L&   [4] GREEK CAPITAL LETTER HETA..GREEK SMALL LETTER ARCHAIC SAMPI
    Interval{ 0x0376, 0x0377 }, // L&   [2] GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA..GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037B, 0x037D }, // L&   [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Interval{ 0x037F, 0x037F }, // L&       GREEK CAPITAL LETTER YOT
    Interval{ 0x0386, 0x0386 }, // L&       GREEK CAPITAL LETTER ALPHA WITH TONOS
    Interval{ 0x0388, 0x038A }, // L&   [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Interval{ 0x038C, 0x038C }, // L&       GREEK CAPITAL LETTER OMICRON WITH TONOS
    Interval{ 0x038E, 0x03A1 }, // L&  [20] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER RHO
    Interval{ 0x03A3, 0x03D1 }, // L&  [47] GREEK CAPITAL LETTER SIGMA..GREEK THETA SYMBOL
    Interval{ 0x03D5, 0x03F5 }, // L&  [33] GREEK PHI SYMBOL..GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F7, 0x03FB }, // L&   [5] GREEK CAPITAL LETTER SHO..GREEK SMALL LETTER SAN
    Interval{ 0x03FD, 0x0481 }, // L& [133] GREEK CAPITAL REVERSED LUNATE SIGMA SYMBOL..CYRILLIC SMALL LETTER KOPPA
    Interval{ 0x048A, 0x052F }, // L& [166] CYRILLIC CAPITAL LETTER SHORT I WITH TAIL..CYRILLIC SMALL LETTER EL WITH DESCENDER
    Interval{ 0x0531, 0x0556 }, // L&  [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Interval{ 0x0561, 0x0587 }, // L&  [39] ARMENIAN SMALL LETTER AYB..ARMENIAN SMALL LIGATURE ECH YIWN
    Interval{ 0x10A0, 0x10C5 }, // L&  [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Interval{ 0x10C7, 0x10C7 }, // L&       GEORGIAN CAPITAL LETTER YN
    Interval{ 0x10CD, 0x10CD }, // L&       GEORGIAN CAPITAL LETTER AEN
    Interval{ 0x10D0, 0x10FA }, // L&  [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Interval{ 0x10FD, 0x10FF }, // L&   [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Interval{ 0x13A0, 0x13F5 }, // L&  [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Interval{ 0x13F8, 0x13FD }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1C80, 0x1C8A }, // L&  [11] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER TJE
    Interval{ 0x1C90, 0x1CBA }, // L&  [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Interval{ 0x1CBD, 0x1CBF }, // L&   [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Interval{ 0x1D79, 0x1D79 }, // L&       LATIN SMALL LETTER INSULAR G
    Interval{ 0x1D7D, 0x1D7D }, // L&       LATIN SMALL LETTER P WITH STROKE
    Interval{ 0x1D8E, 0x1D8E }, // L&       LATIN SMALL LETTER Z WITH PALATAL HOOK
    Interval{ 0x1E00, 0x1E9B }, // L& [156] LATIN CAPITAL LETTER A WITH RING BELOW..LATIN SMALL LETTER LONG S WITH DOT ABOVE
    Interval{ 0x1E9E, 0x1E9E }, // L&       LATIN CAPITAL LETTER SHARP S
    Interval{ 0x1EA0, 0x1F15 }, // L& [118] LATIN CAPITAL LETTER A WITH DOT BELOW..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F18, 0x1F1D }, // L&   [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F20, 0x1F45 }, // L&  [38] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F48, 0x1F4D }, // L&   [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F50, 0x1F57 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F59, 0x1F59 }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA
    Interval{ 0x1F5B, 0x1F5B }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Interval{ 0x1F5D, 0x1F5D }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Interval{ 0x1F5F, 0x1F7D }, // L&  [31] GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI..GREEK SMALL LETTER OMEGA WITH OXIA
    Interval{ 0x1F80, 0x1FB4 }, // L&  [53] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB6, 0x1FBC }, // L&   [7] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Interval{ 0x1FBE, 0x1FBE }, // L&       GREEK PROSGEGRAMMENI
    Interval{ 0x1FC2, 0x1FC4 }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC6, 0x1FCC }, // L&   [7] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Interval{ 0x1FD0, 0x1FD3 }, // L&   [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Interval{ 0x1FD6, 0x1FDB }, // L&   [6] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK CAPITAL LETTER IOTA WITH OXIA
    Interval{ 0x1FE0, 0x1FEC }, // L&  [13] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Interval{ 0x1FF2, 0x1FF4 }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF6, 0x1FFC }, // L&   [7] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Interval{ 0x2126, 0x2126 }, // L&       OHM SIGN
    Interval{ 0x212A, 0x212B }, // L&   [2] KELVIN SIGN..ANGSTROM SIGN
    Interval{ 0x2132, 0x2132 }, // L&       TURNED CAPITAL F
    Interval{ 0x214E, 0x214E }, // L&       TURNED SMALL F
    Interval{ 0x2160, 0x217F }, // Nl  [32] ROMAN NUMERAL ONE..SMALL ROMAN NUMERAL ONE THOUSAND
    Interval{ 0x2183, 0x2184 }, // L&   [2] ROMAN NUMERAL REVERSED ONE HUNDRED..LATIN SMALL LETTER REVERSED C
    Interval{ 0x24B6, 0x24E9 }, // So  [52] CIRCLED LATIN CAPITAL LETTER A..CIRCLED LATIN SMALL LETTER Z
    Interval{ 0x2C00, 0x2C70 }, // L& [113] GLAGOLITIC CAPITAL LETTER AZU..LATIN CAPITAL LETTER TURNED ALPHA
    Interval{ 0x2C72, 0x2C73 }, // L&   [2] LATIN CAPITAL LETTER W WITH HOOK..LATIN SMALL LETTER W WITH HOOK
    Interval{ 0x2C75, 0x2C76 }, // L&   [2] LATIN CAPITAL LETTER HALF H..LATIN SMALL LETTER HALF H
    Interval{ 0x2C7E, 0x2CE3 }, // L& [102] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC SMALL LETTER OLD NUBIAN WAU
    Interval{ 0x2CEB, 0x2CEE }, // L&   [4] COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI..COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF2, 0x2CF3 }, // L&   [2] COPTIC CAPITAL LETTER BOHAIRIC KHEI..COPTIC SMALL LETTER BOHAIRIC KHEI
    Interval{ 0x2D00, 0x2D25 }, // L&  [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Interval{ 0x2D27, 0x2D27 }, // L&       GEORGIAN SMALL LETTER YN
    Interval{ 0x2D2D, 0x2D2D }, // L&       GEORGIAN SMALL LETTER AEN
    Interval{ 0xA640, 0xA66D }, // L&  [46] CYRILLIC CAPITAL LETTER ZEMLYA..CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA680, 0xA69B }, // L&  [28] CYRILLIC CAPITAL LETTER DWE..CYRILLIC SMALL LETTER CROSSED O
    Interval{ 0xA722, 0xA72F }, // L&  [14] LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF..LATIN SMALL LETTER CUATRILLO WITH COMMA
    Interval{ 0xA732, 0xA76F }, // L&  [62] LATIN CAPITAL LETTER AA..LATIN SMALL LETTER CON
    Interval{ 0xA779, 0xA787 }, // L&  [15] LATIN CAPITAL LETTER INSULAR D..LATIN SMALL LETTER INSULAR T
    Interval{ 0xA78B, 0xA78D }, // L&   [3] LATIN CAPITAL LETTER SALTILLO..LATIN CAPITAL LETTER TURNED H
    Interval{ 0xA790, 0xA794 }, // L&   [5] LATIN CAPITAL LETTER N WITH DESCENDER..LATIN SMALL LETTER C WITH PALATAL HOOK
    Interval{ 0xA796, 0xA7AE }, // L&  [25] LATIN CAPITAL LETTER B WITH FLOURISH..LATIN CAPITAL LETTER SMALL CAPITAL I
    Interval{ 0xA7B0, 0xA7DC }, // L&  [45] LATIN CAPITAL LETTER TURNED K..LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Interval{ 0xA7F5, 0xA7F6 }, // L&   [2] LATIN CAPITAL LETTER REVERSED HALF H..LATIN SMALL LETTER REVERSED HALF H
    Interval{ 0xAB53, 0xAB53 }, // L&       LATIN SMALL LETTER CHI
    Interval{ 0xAB70, 0xABBF }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xFB00, 0xFB06 }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFF21, 0xFF3A }, // L&  [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Interval{ 0xFF41, 0xFF5A }, // L&  [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Interval{ 0x10400, 0x1044F }, // L&  [80] DESERET CAPITAL LETTER LONG I..DESERET SMALL LETTER EW
    Interval{ 0x104B0, 0x104D3 }, // L&  [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Interval{ 0x104D8, 0x104FB }, // L&  [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Interval{ 0x10570, 0x1057A }, // L&  [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Interval{ 0x1057C, 0x1058A }, // L&  [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Interval{ 0x1058C, 0x10592 }, // L&   [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Interval{ 0x10594, 0x10595 }, // L&   [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Interval{ 0x10597, 0x105A1 }, // L&  [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Interval{ 0x105A3, 0x105B1 }, // L&  [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Interval{ 0x105B3, 0x105B9 }, // L&   [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Interval{ 0x105BB, 0x105BC }, // L&   [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Interval{ 0x10C80, 0x10CB2 }, // L&  [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Interval{ 0x10CC0, 0x10CF2 }, // L&  [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Interval{ 0x10D50, 0x10D65 }, // L&  [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Interval{ 0x10D70, 0x10D85 }, // L&  [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Interval{ 0x118A0, 0x118DF }, // L&  [64] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Interval{ 0x16E40, 0x16E7F }, // L&  [64] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Interval{ 0x16EA0, 0x16EB8 }, // L&  [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Interval{ 0x16EBB, 0x16ED3 }, // L&  [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Interval{ 0x1E900, 0x1E943 }, // L&  [68] ADLAM CAPITAL LETTER ALIF..ADLAM SMALL LETTER SHA
}; // }}}
auto static const Changes_When_Lowercased = std::array<Interval, 618>{ // {{{
    Interval{ 0x0041, 0x005A }, // L&  [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Interval{ 0x00C0, 0x00D6 }, // L&  [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x00D8, 0x00DE }, // L&   [7] LATIN CAPITAL LETTER O WITH STROKE..LATIN CAPITAL LETTER THORN
    Interval{ 0x0100, 0x0100 }, // L&       LATIN CAPITAL LETTER A WITH MACRON
    Interval{ 0x0102, 0x0102 }, // L&       LATIN CAPITAL LETTER A WITH BREVE
    Interval{ 0x0104, 0x0104 }, // L&       LATIN CAPITAL LETTER A WITH OGONEK
    Interval{ 0x0106, 0x0106 }, // L&       LATIN CAPITAL LETTER C WITH ACUTE
    Interval{ 0x0108, 0x0108 }, // L&       LATIN CAPITAL LETTER C WITH CIRCUMFLEX
    Interval{ 0x010A, 0x010A }, // L&       LATIN CAPITAL LETTER C WITH DOT ABOVE
    Interval{ 0x010C, 0x010C }, // L&       LATIN CAPITAL LETTER C WITH CARON
    Interval{ 0x010E, 0x010E }, // L&       LATIN CAPITAL LETTER D WITH CARON
    Interval{ 0x0110, 0x0110 }, // L&       LATIN CAPITAL LETTER D WITH STROKE
    Interval{ 0x0112, 0x0112 }, // L&       LATIN CAPITAL LETTER E WITH MACRON
    Interval{ 0x0114, 0x0114 }, // L&       LATIN CAPITAL LETTER E WITH BREVE
    Interval{ 0x0116, 0x0116 }, // L&       LATIN CAPITAL LETTER E WITH DOT ABOVE
    Interval{ 0x0118, 0x0118 }, // L&       LATIN CAPITAL LETTER E WITH OGONEK
    Interval{ 0x011A, 0x011A }, // L&       LATIN CAPITAL LETTER E WITH CARON
    Interval{ 0x011C, 0x011C }, // L&       LATIN CAPITAL LETTER G WITH CIRCUMFLEX
    Interval{ 0x011E, 0x011E }, // L&       LATIN CAPITAL LETTER G WITH BREVE
    Interval{ 0x0120, 0x0120 }, // L&       LATIN CAPITAL LETTER G WITH DOT ABOVE
    Interval{ 0x0122, 0x0122 }, // L&       LATIN CAPITAL LETTER G WITH CEDILLA
    Interval{ 0x0124, 0x0124 }, // L&       LATIN CAPITAL LETTER H WITH CIRCUMFLEX
    Interval{ 0x0126, 0x0126 }, // L&       LATIN CAPITAL LETTER H WITH STROKE
    Interval{ 0x0128, 0x0128 }, // L&       LATIN CAPITAL LETTER I WITH TILDE
    Interval{ 0x012A, 0x012A }, // L&       LATIN CAPITAL LETTER I WITH MACRON
    Interval{ 0x012C, 0x012C }, // L&       LATIN CAPITAL LETTER I WITH BREVE
    Interval{ 0x012E, 0x012E }, // L&       LATIN CAPITAL LETTER I WITH OGONEK
    Interval{ 0x0130, 0x0130 }, // L&       LATIN CAPITAL LETTER I WITH DOT ABOVE
    Interval{ 0x0132, 0x0132 }, // L&       LATIN CAPITAL LIGATURE IJ
    Interval{ 0x0134, 0x0134 }, // L&       LATIN CAPITAL LETTER J WITH CIRCUMFLEX
    Interval{ 0x0136, 0x0136 }, // L&       LATIN CAPITAL LETTER K WITH CEDILLA
    Interval{ 0x0139, 0x0139 }, // L&       LATIN CAPITAL LETTER L WITH ACUTE
    Interval{ 0x013B, 0x013B }, // L&       LATIN CAPITAL LETTER L WITH CEDILLA
    Interval{ 0x013D, 0x013D }, // L&       LATIN CAPITAL LETTER L WITH CARON
    Interval{ 0x013F, 0x013F }, // L&       LATIN CAPITAL LETTER L WITH MIDDLE DOT
    Interval{ 0x0141, 0x0141 }, // L&       LATIN CAPITAL LETTER L WITH STROKE
    Interval{ 0x0143, 0x0143 }, // L&       LATIN CAPITAL LETTER N WITH ACUTE
    Interval{ 0x0145, 0x0145 }, // L&       LATIN CAPITAL LETTER N WITH CEDILLA
    Interval{ 0x0147, 0x0147 }, // L&       LATIN CAPITAL LETTER N WITH CARON
    Interval{ 0x014A, 0x014A }, // L&       LATIN CAPITAL LETTER ENG
    Interval{ 0x014C, 0x014C }, // L&       LATIN CAPITAL LETTER O WITH MACRON
    Interval{ 0x014E, 0x014E }, // L&       LATIN CAPITAL LETTER O WITH BREVE
    Interval{ 0x0150, 0x0150 }, // L&       LATIN CAPITAL LETTER O WITH DOUBLE ACUTE
    Interval{ 0x0152, 0x0152 }, // L&       LATIN CAPITAL LIGATURE OE
    Interval{ 0x0154, 0x0154 }, // L&       LATIN CAPITAL LETTER R WITH ACUTE
    Interval{ 0x0156, 0x0156 }, // L&       LATIN CAPITAL LETTER R WITH CEDILLA
    Interval{ 0x0158, 0x0158 }, // L&       LATIN CAPITAL LETTER R WITH CARON
    Interval{ 0x015A, 0x015A }, // L&       LATIN CAPITAL LETTER S WITH ACUTE
    Interval{ 0x015C, 0x015C }, // L&       LATIN CAPITAL LETTER S WITH CIRCUMFLEX
    Interval{ 0x015E, 0x015E }, // L&       LATIN CAPITAL LETTER S WITH CEDILLA
    Interval{ 0x0160, 0x0160 }, // L&       LATIN CAPITAL LETTER S WITH CARON
    Interval{ 0x0162, 0x0162 }, // L&       LATIN CAPITAL LETTER T WITH CEDILLA
    Interval{ 0x0164, 0x0164 }, // L&       LATIN CAPITAL LETTER T WITH CARON
    Interval{ 0x0166, 0x0166 }, // L&       LATIN CAPITAL LETTER T WITH STROKE
    Interval{ 0x0168, 0x0168 }, // L&       LATIN CAPITAL LETTER U WITH TILDE
    Interval{ 0x016A, 0x016A }, // L&       LATIN CAPITAL LETTER U WITH MACRON
    Interval{ 0x016C, 0x016C }, // L&       LATIN CAPITAL LETTER U WITH BREVE
    Interval{ 0x016E, 0x016E }, // L&       LATIN CAPITAL LETTER U WITH RING ABOVE
    Interval{ 0x0170, 0x0170 }, // L&       LATIN CAPITAL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x0172, 0x0172 }, // L&       LATIN CAPITAL LETTER U WITH OGONEK
    Interval{ 0x0174, 0x0174 }, // L&       LATIN CAPITAL LETTER W WITH CIRCUMFLEX
    Interval{ 0x0176, 0x0176 }, // L&       LATIN CAPITAL LETTER Y WITH CIRCUMFLEX
    Interval{ 0x0178, 0x0179 }, // L&   [2] LATIN CAPITAL LETTER Y WITH DIAERESIS..LATIN CAPITAL LETTER Z WITH ACUTE
    Interval{ 0x017B, 0x017B }, // L&       LATIN CAPITAL LETTER Z WITH DOT ABOVE
    Interval{ 0x017D, 0x017D }, // L&       LATIN CAPITAL LETTER Z WITH CARON
    Interval{ 0x0181, 0x0182 }, // L&   [2] LATIN CAPITAL LETTER B WITH HOOK..LATIN CAPITAL LETTER B WITH TOPBAR
    Interval{ 0x0184, 0x0184 }, // L&       LATIN CAPITAL LETTER TONE SIX
    Interval{ 0x0186, 0x0187 }, // L&   [2] LATIN CAPITAL LETTER OPEN O..LATIN CAPITAL LETTER C WITH HOOK
    Interval{ 0x0189, 0x018B }, // L&   [3] LATIN CAPITAL LETTER AFRICAN D..LATIN CAPITAL LETTER D WITH TOPBAR
    Interval{ 0x018E, 0x0191 }, // L&   [4] LATIN CAPITAL LETTER REVERSED E..LATIN CAPITAL LETTER F WITH HOOK
    Interval{ 0x0193, 0x0194 }, // L&   [2] LATIN CAPITAL LETTER G WITH HOOK..LATIN CAPITAL LETTER GAMMA
    Interval{ 0x0196, 0x0198 }, // L&   [3] LATIN CAPITAL LETTER IOTA..LATIN CAPITAL LETTER K WITH HOOK
    Interval{ 0x019C, 0x019D }, // L&   [2] LATIN CAPITAL LETTER TURNED M..LATIN CAPITAL LETTER N WITH LEFT HOOK
    Interval{ 0x019F, 0x01A0 }, // L&   [2] LATIN CAPITAL LETTER O WITH MIDDLE TILDE..LATIN CAPITAL LETTER O WITH HORN
    Interval{ 0x01A2, 0x01A2 }, // L&       LATIN CAPITAL LETTER OI
    Interval{ 0x01A4, 0x01A4 }, // L&       LATIN CAPITAL LETTER P WITH HOOK
    Interval{ 0x01A6, 0x01A7 }, // L&   [2] LATIN LETTER YR..LATIN CAPITAL LETTER TONE TWO
    Interval{ 0x01A9, 0x01A9 }, // L&       LATIN CAPITAL LETTER ESH
    Interval{ 0x01AC, 0x01AC }, // L&       LATIN CAPITAL LETTER T WITH HOOK
    Interval{ 0x01AE, 0x01AF }, // L&   [2] LATIN CAPITAL LETTER T WITH RETROFLEX HOOK..LATIN CAPITAL LETTER U WITH HORN
    Interval{ 0x01B1, 0x01B3 }, // L&   [3] LATIN CAPITAL LETTER UPSILON..LATIN CAPITAL LETTER Y WITH HOOK
    Interval{ 0x01B5, 0x01B5 }, // L&       LATIN CAPITAL LETTER Z WITH STROKE
    Interval{ 0x01B7, 0x01B8 }, // L&   [2] LATIN CAPITAL LETTER EZH..LATIN CAPITAL LETTER EZH REVERSED
    Interval{ 0x01BC, 0x01BC }, // L&       LATIN CAPITAL LETTER TONE FIVE
    Interval{ 0x01C4, 0x01C5 }, // L&   [2] LATIN CAPITAL LETTER DZ WITH CARON..LATIN CAPITAL LETTER D WITH SMALL LETTER Z WITH CARON
    Interval{ 0x01C7, 0x01C8 }, // L&   [2] LATIN CAPITAL LETTER LJ..LATIN CAPITAL LETTER L WITH SMALL LETTER J
    Interval{ 0x01CA, 0x01CB }, // L&   [2] LATIN CAPITAL LETTER NJ..LATIN CAPITAL LETTER N WITH SMALL LETTER J
    Interval{ 0x01CD, 0x01CD }, // L&       LATIN CAPITAL LETTER A WITH CARON
    Interval{ 0x01CF, 0x01CF }, // L&       LATIN CAPITAL LETTER I WITH CARON
    Interval{ 0x01D1, 0x01D1 }, // L&       LATIN CAPITAL LETTER O WITH CARON
    Interval{ 0x01D3, 0x01D3 }, // L&       LATIN CAPITAL LETTER U WITH CARON
    Interval{ 0x01D5, 0x01D5 }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS AND MACRON
    Interval{ 0x01D7, 0x01D7 }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS AND ACUTE
    Interval{ 0x01D9, 0x01D9 }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS AND CARON
    Interval{ 0x01DB, 0x01DB }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS AND GRAVE
    Interval{ 0x01DE, 0x01DE }, // L&       LATIN CAPITAL LETTER A WITH DIAERESIS AND MACRON
    Interval{ 0x01E0, 0x01E0 }, // L&       LATIN CAPITAL LETTER A WITH DOT ABOVE AND MACRON
    Interval{ 0x01E2, 0x01E2 }, // L&       LATIN CAPITAL LETTER AE WITH MACRON
    Interval{ 0x01E4, 0x01E4 }, // L&       LATIN CAPITAL LETTER G WITH STROKE
    Interval{ 0x01E6, 0x01E6 }, // L&       LATIN CAPITAL LETTER G WITH CARON
    Interval{ 0x01E8, 0x01E8 }, // L&       LATIN CAPITAL LETTER K WITH CARON
    Interval{ 0x01EA, 0x01EA }, // L&       LATIN CAPITAL LETTER O WITH OGONEK
    Interval{ 0x01EC, 0x01EC }, // L&       LATIN CAPITAL LETTER O WITH OGONEK AND MACRON
    Interval{ 0x01EE, 0x01EE }, // L&       LATIN CAPITAL LETTER EZH WITH CARON
    Interval{ 0x01F1, 0x01F2 }, // L&   [2] LATIN CAPITAL LETTER DZ..LATIN CAPITAL LETTER D WITH SMALL LETTER Z
    Interval{ 0x01F4, 0x01F4 }, // L&       LATIN CAPITAL LETTER G WITH ACUTE
    Interval{ 0x01F6, 0x01F8 }, // L&   [3] LATIN CAPITAL LETTER HWAIR..LATIN CAPITAL LETTER N WITH GRAVE
    Interval{ 0x01FA, 0x01FA }, // L&       LATIN CAPITAL LETTER A WITH RING ABOVE AND ACUTE
    Interval{ 0x01FC, 0x01FC }, // L&       LATIN CAPITAL LETTER AE WITH ACUTE
    Interval{ 0x01FE, 0x01FE }, // L&       LATIN CAPITAL LETTER O WITH STROKE AND ACUTE
    Interval{ 0x0200, 0x0200 }, // L&       LATIN CAPITAL LETTER A WITH DOUBLE GRAVE
    Interval{ 0x0202, 0x0202 }, // L&       LATIN CAPITAL LETTER A WITH INVERTED BREVE
    Interval{ 0x0204, 0x0204 }, // L&       LATIN CAPITAL LETTER E WITH DOUBLE GRAVE
    Interval{ 0x0206, 0x0206 }, // L&       LATIN CAPITAL LETTER E WITH INVERTED BREVE
    Interval{ 0x0208, 0x0208 }, // L&       LATIN CAPITAL LETTER I WITH DOUBLE GRAVE
    Interval{ 0x020A, 0x020A }, // L&       LATIN CAPITAL LETTER I WITH INVERTED BREVE
    Interval{ 0x020C, 0x020C }, // L&       LATIN CAPITAL LETTER O WITH DOUBLE GRAVE
    Interval{ 0x020E, 0x020E }, // L&       LATIN CAPITAL LETTER O WITH INVERTED BREVE
    Interval{ 0x0210, 0x0210 }, // L&       LATIN CAPITAL LETTER R WITH DOUBLE GRAVE
    Interval{ 0x0212, 0x0212 }, // L&       LATIN CAPITAL LETTER R WITH INVERTED BREVE
    Interval{ 0x0214, 0x0214 }, // L&       LATIN CAPITAL LETTER U WITH DOUBLE GRAVE
    Interval{ 0x0216, 0x0216 }, // L&       LATIN CAPITAL LETTER U WITH INVERTED BREVE
    Interval{ 0x0218, 0x0218 }, // L&       LATIN CAPITAL LETTER S WITH COMMA BELOW
    Interval{ 0x021A, 0x021A }, // L&       LATIN CAPITAL LETTER T WITH COMMA BELOW
    Interval{ 0x021C, 0x021C }, // L&       LATIN CAPITAL LETTER YOGH
    Interval{ 0x021E, 0x021E }, // L&       LATIN CAPITAL LETTER H WITH CARON
    Interval{ 0x0220, 0x0220 }, // L&       LATIN CAPITAL LETTER N WITH LONG RIGHT LEG
    Interval{ 0x0222, 0x0222 }, // L&       LATIN CAPITAL LETTER OU
    Interval{ 0x0224, 0x0224 }, // L&       LATIN CAPITAL LETTER Z WITH HOOK
    Interval{ 0x0226, 0x0226 }, // L&       LATIN CAPITAL LETTER A WITH DOT ABOVE
    Interval{ 0x0228, 0x0228 }, // L&       LATIN CAPITAL LETTER E WITH CEDILLA
    Interval{ 0x022A, 0x022A }, // L&       LATIN CAPITAL LETTER O WITH DIAERESIS AND MACRON
    Interval{ 0x022C, 0x022C }, // L&       LATIN CAPITAL LETTER O WITH TILDE AND MACRON
    Interval{ 0x022E, 0x022E }, // L&       LATIN CAPITAL LETTER O WITH DOT ABOVE
    Interval{ 0x0230, 0x0230 }, // L&       LATIN CAPITAL LETTER O WITH DOT ABOVE AND MACRON
    Interval{ 0x0232, 0x0232 }, // L&       LATIN CAPITAL LETTER Y WITH MACRON
    Interval{ 0x023A, 0x023B }, // L&   [2] LATIN CAPITAL LETTER A WITH STROKE..LATIN CAPITAL LETTER C WITH STROKE
    Interval{ 0x023D, 0x023E }, // L&   [2] LATIN CAPITAL LETTER L WITH BAR..LATIN CAPITAL LETTER T WITH DIAGONAL STROKE
    Interval{ 0x0241, 0x0241 }, // L&       LATIN CAPITAL LETTER GLOTTAL STOP
    Interval{ 0x0243, 0x0246 }, // L&   [4] LATIN CAPITAL LETTER B WITH STROKE..LATIN CAPITAL LETTER E WITH STROKE
    Interval{ 0x0248, 0x0248 }, // L&       LATIN CAPITAL LETTER J WITH STROKE
    Interval{ 0x024A, 0x024A }, // L&       LATIN CAPITAL LETTER SMALL Q WITH HOOK TAIL
    Interval{ 0x024C, 0x024C }, // L&       LATIN CAPITAL LETTER R WITH STROKE
    Interval{ 0x024E, 0x024E }, // L&       LATIN CAPITAL LETTER Y WITH STROKE
    Interval{ 0x0370, 0x0370 }, // L&       GREEK CAPITAL LETTER HETA
    Interval{ 0x0372, 0x0372 }, // L&       GREEK CAPITAL LETTER ARCHAIC SAMPI
    Interval{ 0x0376, 0x0376 }, // L&       GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037F, 0x037F }, // L&       GREEK CAPITAL LETTER YOT
    Interval{ 0x0386, 0x0386 }, // L&       GREEK CAPITAL LETTER ALPHA WITH TONOS
    Interval{ 0x0388, 0x038A }, // L&   [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Interval{ 0x038C, 0x038C }, // L&       GREEK CAPITAL LETTER OMICRON WITH TONOS
    Interval{ 0x038E, 0x038F }, // L&   [2] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER OMEGA WITH TONOS
    Interval{ 0x0391, 0x03A1 }, // L&  [17] GREEK CAPITAL LETTER ALPHA..GREEK CAPITAL LETTER RHO
    Interval{ 0x03A3, 0x03AB }, // L&   [9] GREEK CAPITAL LETTER SIGMA..GREEK CAPITAL LETTER UPSILON WITH DIALYTIKA
    Interval{ 0x03CF, 0x03CF }, // L&       GREEK CAPITAL KAI SYMBOL
    Interval{ 0x03D8, 0x03D8 }, // L&       GREEK LETTER ARCHAIC KOPPA
    Interval{ 0x03DA, 0x03DA }, // L&       GREEK LETTER STIGMA
    Interval{ 0x03DC, 0x03DC }, // L&       GREEK LETTER DIGAMMA
    Interval{ 0x03DE, 0x03DE }, // L&       GREEK LETTER KOPPA
    Interval{ 0x03E0, 0x03E0 }, // L&       GREEK LETTER SAMPI
    Interval{ 0x03E2, 0x03E2 }, // L&       COPTIC CAPITAL LETTER SHEI
    Interval{ 0x03E4, 0x03E4 }, // L&       COPTIC CAPITAL LETTER FEI
    Interval{ 0x03E6, 0x03E6 }, // L&       COPTIC CAPITAL LETTER KHEI
    Interval{ 0x03E8, 0x03E8 }, // L&       COPTIC CAPITAL LETTER HORI
    Interval{ 0x03EA, 0x03EA }, // L&       COPTIC CAPITAL LETTER GANGIA
    Interval{ 0x03EC, 0x03EC }, // L&       COPTIC CAPITAL LETTER SHIMA
    Interval{ 0x03EE, 0x03EE }, // L&       COPTIC CAPITAL LETTER DEI
    Interval{ 0x03F4, 0x03F4 }, // L&       GREEK CAPITAL THETA SYMBOL
    Interval{ 0x03F7, 0x03F7 }, // L&       GREEK CAPITAL LETTER SHO
    Interval{ 0x03F9, 0x03FA }, // L&   [2] GREEK CAPITAL LUNATE SIGMA SYMBOL..GREEK CAPITAL LETTER SAN
    Interval{ 0x03FD, 0x042F }, // L&  [51] GREEK CAPITAL REVERSED LUNATE SIGMA SYMBOL..CYRILLIC CAPITAL LETTER YA
    Interval{ 0x0460, 0x0460 }, // L&       CYRILLIC CAPITAL LETTER OMEGA
    Interval{ 0x0462, 0x0462 }, // L&       CYRILLIC CAPITAL LETTER YAT
    Interval{ 0x0464, 0x0464 }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED E
    Interval{ 0x0466, 0x0466 }, // L&       CYRILLIC CAPITAL LETTER LITTLE YUS
    Interval{ 0x0468, 0x0468 }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED LITTLE YUS
    Interval{ 0x046A, 0x046A }, // L&       CYRILLIC CAPITAL LETTER BIG YUS
    Interval{ 0x046C, 0x046C }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED BIG YUS
    Interval{ 0x046E, 0x046E }, // L&       CYRILLIC CAPITAL LETTER KSI
    Interval{ 0x0470, 0x0470 }, // L&       CYRILLIC CAPITAL LETTER PSI
    Interval{ 0x0472, 0x0472 }, // L&       CYRILLIC CAPITAL LETTER FITA
    Interval{ 0x0474, 0x0474 }, // L&       CYRILLIC CAPITAL LETTER IZHITSA
    Interval{ 0x0476, 0x0476 }, // L&       CYRILLIC CAPITAL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT
    Interval{ 0x0478, 0x0478 }, // L&       CYRILLIC CAPITAL LETTER UK
    Interval{ 0x047A, 0x047A }, // L&       CYRILLIC CAPITAL LETTER ROUND OMEGA
    Interval{ 0x047C, 0x047C }, // L&       CYRILLIC CAPITAL LETTER OMEGA WITH TITLO
    Interval{ 0x047E, 0x047E }, // L&       CYRILLIC CAPITAL LETTER OT
    Interval{ 0x0480, 0x0480 }, // L&       CYRILLIC CAPITAL LETTER KOPPA
    Interval{ 0x048A, 0x048A }, // L&       CYRILLIC CAPITAL LETTER SHORT I WITH TAIL
    Interval{ 0x048C, 0x048C }, // L&       CYRILLIC CAPITAL LETTER SEMISOFT SIGN
    Interval{ 0x048E, 0x048E }, // L&       CYRILLIC CAPITAL LETTER ER WITH TICK
    Interval{ 0x0490, 0x0490 }, // L&       CYRILLIC CAPITAL LETTER GHE WITH UPTURN
    Interval{ 0x0492, 0x0492 }, // L&       CYRILLIC CAPITAL LETTER GHE WITH STROKE
    Interval{ 0x0494, 0x0494 }, // L&       CYRILLIC CAPITAL LETTER GHE WITH MIDDLE HOOK
    Interval{ 0x0496, 0x0496 }, // L&       CYRILLIC CAPITAL LETTER ZHE WITH DESCENDER
    Interval{ 0x0498, 0x0498 }, // L&       CYRILLIC CAPITAL LETTER ZE WITH DESCENDER
    Interval{ 0x049A, 0x049A }, // L&       CYRILLIC CAPITAL LETTER KA WITH DESCENDER
    Interval{ 0x049C, 0x049C }, // L&       CYRILLIC CAPITAL LETTER KA WITH VERTICAL STROKE
    Interval{ 0x049E, 0x049E }, // L&       CYRILLIC CAPITAL LETTER KA WITH STROKE
    Interval{ 0x04A0, 0x04A0 }, // L&       CYRILLIC CAPITAL LETTER BASHKIR KA
    Interval{ 0x04A2, 0x04A2 }, // L&       CYRILLIC CAPITAL LETTER EN WITH DESCENDER
    Interval{ 0x04A4, 0x04A4 }, // L&       CYRILLIC CAPITAL LIGATURE EN GHE
    Interval{ 0x04A6, 0x04A6 }, // L&       CYRILLIC CAPITAL LETTER PE WITH MIDDLE HOOK
    Interval{ 0x04A8, 0x04A8 }, // L&       CYRILLIC CAPITAL LETTER ABKHASIAN HA
    Interval{ 0x04AA, 0x04AA }, // L&       CYRILLIC CAPITAL LETTER ES WITH DESCENDER
    Interval{ 0x04AC, 0x04AC }, // L&       CYRILLIC CAPITAL LETTER TE WITH DESCENDER
    Interval{ 0x04AE, 0x04AE }, // L&       CYRILLIC CAPITAL LETTER STRAIGHT U
    Interval{ 0x04B0, 0x04B0 }, // L&       CYRILLIC CAPITAL LETTER STRAIGHT U WITH STROKE
    Interval{ 0x04B2, 0x04B2 }, // L&       CYRILLIC CAPITAL LETTER HA WITH DESCENDER
    Interval{ 0x04B4, 0x04B4 }, // L&       CYRILLIC CAPITAL LIGATURE TE TSE
    Interval{ 0x04B6, 0x04B6 }, // L&       CYRILLIC CAPITAL LETTER CHE WITH DESCENDER
    Interval{ 0x04B8, 0x04B8 }, // L&       CYRILLIC CAPITAL LETTER CHE WITH VERTICAL STROKE
    Interval{ 0x04BA, 0x04BA }, // L&       CYRILLIC CAPITAL LETTER SHHA
    Interval{ 0x04BC, 0x04BC }, // L&       CYRILLIC CAPITAL LETTER ABKHASIAN CHE
    Interval{ 0x04BE, 0x04BE }, // L&       CYRILLIC CAPITAL LETTER ABKHASIAN CHE WITH DESCENDER
    Interval{ 0x04C0, 0x04C1 }, // L&   [2] CYRILLIC LETTER PALOCHKA..CYRILLIC CAPITAL LETTER ZHE WITH BREVE
    Interval{ 0x04C3, 0x04C3 }, // L&       CYRILLIC CAPITAL LETTER KA WITH HOOK
    Interval{ 0x04C5, 0x04C5 }, // L&       CYRILLIC CAPITAL LETTER EL WITH TAIL
    Interval{ 0x04C7, 0x04C7 }, // L&       CYRILLIC CAPITAL LETTER EN WITH HOOK
    Interval{ 0x04C9, 0x04C9 }, // L&       CYRILLIC CAPITAL LETTER EN WITH TAIL
    Interval{ 0x04CB, 0x04CB }, // L&       CYRILLIC CAPITAL LETTER KHAKASSIAN CHE
    Interval{ 0x04CD, 0x04CD }, // L&       CYRILLIC CAPITAL LETTER EM WITH TAIL
    Interval{ 0x04D0, 0x04D0 }, // L&       CYRILLIC CAPITAL LETTER A WITH BREVE
    Interval{ 0x04D2, 0x04D2 }, // L&       CYRILLIC CAPITAL LETTER A WITH DIAERESIS
    Interval{ 0x04D4, 0x04D4 }, // L&       CYRILLIC CAPITAL LIGATURE A IE
    Interval{ 0x04D6, 0x04D6 }, // L&       CYRILLIC CAPITAL LETTER IE WITH BREVE
    Interval{ 0x04D8, 0x04D8 }, // L&       CYRILLIC CAPITAL LETTER SCHWA
    Interval{ 0x04DA, 0x04DA }, // L&       CYRILLIC CAPITAL LETTER SCHWA WITH DIAERESIS
    Interval{ 0x04DC, 0x04DC }, // L&       CYRILLIC CAPITAL LETTER ZHE WITH DIAERESIS
    Interval{ 0x04DE, 0x04DE }, // L&       CYRILLIC CAPITAL LETTER ZE WITH DIAERESIS
    Interval{ 0x04E0, 0x04E0 }, // L&       CYRILLIC CAPITAL LETTER ABKHASIAN DZE
    Interval{ 0x04E2, 0x04E2 }, // L&       CYRILLIC CAPITAL LETTER I WITH MACRON
    Interval{ 0x04E4, 0x04E4 }, // L&       CYRILLIC CAPITAL LETTER I WITH DIAERESIS
    Interval{ 0x04E6, 0x04E6 }, // L&       CYRILLIC CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x04E8, 0x04E8 }, // L&       CYRILLIC CAPITAL LETTER BARRED O
    Interval{ 0x04EA, 0x04EA }, // L&       CYRILLIC CAPITAL LETTER BARRED O WITH DIAERESIS
    Interval{ 0x04EC, 0x04EC }, // L&       CYRILLIC CAPITAL LETTER E WITH DIAERESIS
    Interval{ 0x04EE, 0x04EE }, // L&       CYRILLIC CAPITAL LETTER U WITH MACRON
    Interval{ 0x04F0, 0x04F0 }, // L&       CYRILLIC CAPITAL LETTER U WITH DIAERESIS
    Interval{ 0x04F2, 0x04F2 }, // L&       CYRILLIC CAPITAL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x04F4, 0x04F4 }, // L&       CYRILLIC CAPITAL LETTER CHE WITH DIAERESIS
    Interval{ 0x04F6, 0x04F6 }, // L&       CYRILLIC CAPITAL LETTER GHE WITH DESCENDER
    Interval{ 0x04F8, 0x04F8 }, // L&       CYRILLIC CAPITAL LETTER YERU WITH DIAERESIS
    Interval{ 0x04FA, 0x04FA }, // L&       CYRILLIC CAPITAL LETTER GHE WITH STROKE AND HOOK
    Interval{ 0x04FC, 0x04FC }, // L&       CYRILLIC CAPITAL LETTER HA WITH HOOK
    Interval{ 0x04FE, 0x04FE }, // L&       CYRILLIC CAPITAL LETTER HA WITH STROKE
    Interval{ 0x0500, 0x0500 }, // L&       CYRILLIC CAPITAL LETTER KOMI DE
    Interval{ 0x0502, 0x0502 }, // L&       CYRILLIC CAPITAL LETTER KOMI DJE
    Interval{ 0x0504, 0x0504 }, // L&       CYRILLIC CAPITAL LETTER KOMI ZJE
    Interval{ 0x0506, 0x0506 }, // L&       CYRILLIC CAPITAL LETTER KOMI DZJE
    Interval{ 0x0508, 0x0508 }, // L&       CYRILLIC CAPITAL LETTER KOMI LJE
    Interval{ 0x050A, 0x050A }, // L&       CYRILLIC CAPITAL LETTER KOMI NJE
    Interval{ 0x050C, 0x050C }, // L&       CYRILLIC CAPITAL LETTER KOMI SJE
    Interval{ 0x050E, 0x050E }, // L&       CYRILLIC CAPITAL LETTER KOMI TJE
    Interval{ 0x0510, 0x0510 }, // L&       CYRILLIC CAPITAL LETTER REVERSED ZE
    Interval{ 0x0512, 0x0512 }, // L&       CYRILLIC CAPITAL LETTER EL WITH HOOK
    Interval{ 0x0514, 0x0514 }, // L&       CYRILLIC CAPITAL LETTER LHA
    Interval{ 0x0516, 0x0516 }, // L&       CYRILLIC CAPITAL LETTER RHA
    Interval{ 0x0518, 0x0518 }, // L&       CYRILLIC CAPITAL LETTER YAE
    Interval{ 0x051A, 0x051A }, // L&       CYRILLIC CAPITAL LETTER QA
    Interval{ 0x051C, 0x051C }, // L&       CYRILLIC CAPITAL LETTER WE
    Interval{ 0x051E, 0x051E }, // L&       CYRILLIC CAPITAL LETTER ALEUT KA
    Interval{ 0x0520, 0x0520 }, // L&       CYRILLIC CAPITAL LETTER EL WITH MIDDLE HOOK
    Interval{ 0x0522, 0x0522 }, // L&       CYRILLIC CAPITAL LETTER EN WITH MIDDLE HOOK
    Interval{ 0x0524, 0x0524 }, // L&       CYRILLIC CAPITAL LETTER PE WITH DESCENDER
    Interval{ 0x0526, 0x0526 }, // L&       CYRILLIC CAPITAL LETTER SHHA WITH DESCENDER
    Interval{ 0x0528, 0x0528 }, // L&       CYRILLIC CAPITAL LETTER EN WITH LEFT HOOK
    Interval{ 0x052A, 0x052A }, // L&       CYRILLIC CAPITAL LETTER DZZHE
    Interval{ 0x052C, 0x052C }, // L&       CYRILLIC CAPITAL LETTER DCHE
    Interval{ 0x052E, 0x052E }, // L&       CYRILLIC CAPITAL LETTER EL WITH DESCENDER
    Interval{ 0x0531, 0x0556 }, // L&  [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Interval{ 0x10A0, 0x10C5 }, // L&  [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Interval{ 0x10C7, 0x10C7 }, // L&       GEORGIAN CAPITAL LETTER YN
    Interval{ 0x10CD, 0x10CD }, // L&       GEORGIAN CAPITAL LETTER AEN
    Interval{ 0x13A0, 0x13F5 }, // L&  [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Interval{ 0x1C89, 0x1C89 }, // L&       CYRILLIC CAPITAL LETTER TJE
    Interval{ 0x1C90, 0x1CBA }, // L&  [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Interval{ 0x1CBD, 0x1CBF }, // L&   [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Interval{ 0x1E00, 0x1E00 }, // L&       LATIN CAPITAL LETTER A WITH RING BELOW
    Interval{ 0x1E02, 0x1E02 }, // L&       LATIN CAPITAL LETTER B WITH DOT ABOVE
    Interval{ 0x1E04, 0x1E04 }, // L&       LATIN CAPITAL LETTER B WITH DOT BELOW
    Interval{ 0x1E06, 0x1E06 }, // L&       LATIN CAPITAL LETTER B WITH LINE BELOW
    Interval{ 0x1E08, 0x1E08 }, // L&       LATIN CAPITAL LETTER C WITH CEDILLA AND ACUTE
    Interval{ 0x1E0A, 0x1E0A }, // L&       LATIN CAPITAL LETTER D WITH DOT ABOVE
    Interval{ 0x1E0C, 0x1E0C }, // L&       LATIN CAPITAL LETTER D WITH DOT BELOW
    Interval{ 0x1E0E, 0x1E0E }, // L&       LATIN CAPITAL LETTER D WITH LINE BELOW
    Interval{ 0x1E10, 0x1E10 }, // L&       LATIN CAPITAL LETTER D WITH CEDILLA
    Interval{ 0x1E12, 0x1E12 }, // L&       LATIN CAPITAL LETTER D WITH CIRCUMFLEX BELOW
    Interval{ 0x1E14, 0x1E14 }, // L&       LATIN CAPITAL LETTER E WITH MACRON AND GRAVE
    Interval{ 0x1E16, 0x1E16 }, // L&       LATIN CAPITAL LETTER E WITH MACRON AND ACUTE
    Interval{ 0x1E18, 0x1E18 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX BELOW
    Interval{ 0x1E1A, 0x1E1A }, // L&       LATIN CAPITAL LETTER E WITH TILDE BELOW
    Interval{ 0x1E1C, 0x1E1C }, // L&       LATIN CAPITAL LETTER E WITH CEDILLA AND BREVE
    Interval{ 0x1E1E, 0x1E1E }, // L&       LATIN CAPITAL LETTER F WITH DOT ABOVE
    Interval{ 0x1E20, 0x1E20 }, // L&       LATIN CAPITAL LETTER G WITH MACRON
    Interval{ 0x1E22, 0x1E22 }, // L&       LATIN CAPITAL LETTER H WITH DOT ABOVE
    Interval{ 0x1E24, 0x1E24 }, // L&       LATIN CAPITAL LETTER H WITH DOT BELOW
    Interval{ 0x1E26, 0x1E26 }, // L&       LATIN CAPITAL LETTER H WITH DIAERESIS
    Interval{ 0x1E28, 0x1E28 }, // L&       LATIN CAPITAL LETTER H WITH CEDILLA
    Interval{ 0x1E2A, 0x1E2A }, // L&       LATIN CAPITAL LETTER H WITH BREVE BELOW
    Interval{ 0x1E2C, 0x1E2C }, // L&       LATIN CAPITAL LETTER I WITH TILDE BELOW
    Interval{ 0x1E2E, 0x1E2E }, // L&       LATIN CAPITAL LETTER I WITH DIAERESIS AND ACUTE
    Interval{ 0x1E30, 0x1E30 }, // L&       LATIN CAPITAL LETTER K WITH ACUTE
    Interval{ 0x1E32, 0x1E32 }, // L&       LATIN CAPITAL LETTER K WITH DOT BELOW
    Interval{ 0x1E34, 0x1E34 }, // L&       LATIN CAPITAL LETTER K WITH LINE BELOW
    Interval{ 0x1E36, 0x1E36 }, // L&       LATIN CAPITAL LETTER L WITH DOT BELOW
    Interval{ 0x1E38, 0x1E38 }, // L&       LATIN CAPITAL LETTER L WITH DOT BELOW AND MACRON
    Interval{ 0x1E3A, 0x1E3A }, // L&       LATIN CAPITAL LETTER L WITH LINE BELOW
    Interval{ 0x1E3C, 0x1E3C }, // L&       LATIN CAPITAL LETTER L WITH CIRCUMFLEX BELOW
    Interval{ 0x1E3E, 0x1E3E }, // L&       LATIN CAPITAL LETTER M WITH ACUTE
    Interval{ 0x1E40, 0x1E40 }, // L&       LATIN CAPITAL LETTER M WITH DOT ABOVE
    Interval{ 0x1E42, 0x1E42 }, // L&       LATIN CAPITAL LETTER M WITH DOT BELOW
    Interval{ 0x1E44, 0x1E44 }, // L&       LATIN CAPITAL LETTER N WITH DOT ABOVE
    Interval{ 0x1E46, 0x1E46 }, // L&       LATIN CAPITAL LETTER N WITH DOT BELOW
    Interval{ 0x1E48, 0x1E48 }, // L&       LATIN CAPITAL LETTER N WITH LINE BELOW
    Interval{ 0x1E4A, 0x1E4A }, // L&       LATIN CAPITAL LETTER N WITH CIRCUMFLEX BELOW
    Interval{ 0x1E4C, 0x1E4C }, // L&       LATIN CAPITAL LETTER O WITH TILDE AND ACUTE
    Interval{ 0x1E4E, 0x1E4E }, // L&       LATIN CAPITAL LETTER O WITH TILDE AND DIAERESIS
    Interval{ 0x1E50, 0x1E50 }, // L&       LATIN CAPITAL LETTER O WITH MACRON AND GRAVE
    Interval{ 0x1E52, 0x1E52 }, // L&       LATIN CAPITAL LETTER O WITH MACRON AND ACUTE
    Interval{ 0x1E54, 0x1E54 }, // L&       LATIN CAPITAL LETTER P WITH ACUTE
    Interval{ 0x1E56, 0x1E56 }, // L&       LATIN CAPITAL LETTER P WITH DOT ABOVE
    Interval{ 0x1E58, 0x1E58 }, // L&       LATIN CAPITAL LETTER R WITH DOT ABOVE
    Interval{ 0x1E5A, 0x1E5A }, // L&       LATIN CAPITAL LETTER R WITH DOT BELOW
    Interval{ 0x1E5C, 0x1E5C }, // L&       LATIN CAPITAL LETTER R WITH DOT BELOW AND MACRON
    Interval{ 0x1E5E, 0x1E5E }, // L&       LATIN CAPITAL LETTER R WITH LINE BELOW
    Interval{ 0x1E60, 0x1E60 }, // L&       LATIN CAPITAL LETTER S WITH DOT ABOVE
    Interval{ 0x1E62, 0x1E62 }, // L&       LATIN CAPITAL LETTER S WITH DOT BELOW
    Interval{ 0x1E64, 0x1E64 }, // L&       LATIN CAPITAL LETTER S WITH ACUTE AND DOT ABOVE
    Interval{ 0x1E66, 0x1E66 }, // L&       LATIN CAPITAL LETTER S WITH CARON AND DOT ABOVE
    Interval{ 0x1E68, 0x1E68 }, // L&       LATIN CAPITAL LETTER S WITH DOT BELOW AND DOT ABOVE
    Interval{ 0x1E6A, 0x1E6A }, // L&       LATIN CAPITAL LETTER T WITH DOT ABOVE
    Interval{ 0x1E6C, 0x1E6C }, // L&       LATIN CAPITAL LETTER T WITH DOT BELOW
    Interval{ 0x1E6E, 0x1E6E }, // L&       LATIN CAPITAL LETTER T WITH LINE BELOW
    Interval{ 0x1E70, 0x1E70 }, // L&       LATIN CAPITAL LETTER T WITH CIRCUMFLEX BELOW
    Interval{ 0x1E72, 0x1E72 }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS BELOW
    Interval{ 0x1E74, 0x1E74 }, // L&       LATIN CAPITAL LETTER U WITH TILDE BELOW
    Interval{ 0x1E76, 0x1E76 }, // L&       LATIN CAPITAL LETTER U WITH CIRCUMFLEX BELOW
    Interval{ 0x1E78, 0x1E78 }, // L&       LATIN CAPITAL LETTER U WITH TILDE AND ACUTE
    Interval{ 0x1E7A, 0x1E7A }, // L&       LATIN CAPITAL LETTER U WITH MACRON AND DIAERESIS
    Interval{ 0x1E7C, 0x1E7C }, // L&       LATIN CAPITAL LETTER V WITH TILDE
    Interval{ 0x1E7E, 0x1E7E }, // L&       LATIN CAPITAL LETTER V WITH DOT BELOW
    Interval{ 0x1E80, 0x1E80 }, // L&       LATIN CAPITAL LETTER W WITH GRAVE
    Interval{ 0x1E82, 0x1E82 }, // L&       LATIN CAPITAL LETTER W WITH ACUTE
    Interval{ 0x1E84, 0x1E84 }, // L&       LATIN CAPITAL LETTER W WITH DIAERESIS
    Interval{ 0x1E86, 0x1E86 }, // L&       LATIN CAPITAL LETTER W WITH DOT ABOVE
    Interval{ 0x1E88, 0x1E88 }, // L&       LATIN CAPITAL LETTER W WITH DOT BELOW
    Interval{ 0x1E8A, 0x1E8A }, // L&       LATIN CAPITAL LETTER X WITH DOT ABOVE
    Interval{ 0x1E8C, 0x1E8C }, // L&       LATIN CAPITAL LETTER X WITH DIAERESIS
    Interval{ 0x1E8E, 0x1E8E }, // L&       LATIN CAPITAL LETTER Y WITH DOT ABOVE
    Interval{ 0x1E90, 0x1E90 }, // L&       LATIN CAPITAL LETTER Z WITH CIRCUMFLEX
    Interval{ 0x1E92, 0x1E92 }, // L&       LATIN CAPITAL LETTER Z WITH DOT BELOW
    Interval{ 0x1E94, 0x1E94 }, // L&       LATIN CAPITAL LETTER Z WITH LINE BELOW
    Interval{ 0x1E9E, 0x1E9E }, // L&       LATIN CAPITAL LETTER SHARP S
    Interval{ 0x1EA0, 0x1EA0 }, // L&       LATIN CAPITAL LETTER A WITH DOT BELOW
    Interval{ 0x1EA2, 0x1EA2 }, // L&       LATIN CAPITAL LETTER A WITH HOOK ABOVE
    Interval{ 0x1EA4, 0x1EA4 }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EA6, 0x1EA6 }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EA8, 0x1EA8 }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EAA, 0x1EAA }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EAC, 0x1EAC }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EAE, 0x1EAE }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND ACUTE
    Interval{ 0x1EB0, 0x1EB0 }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND GRAVE
    Interval{ 0x1EB2, 0x1EB2 }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND HOOK ABOVE
    Interval{ 0x1EB4, 0x1EB4 }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND TILDE
    Interval{ 0x1EB6, 0x1EB6 }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND DOT BELOW
    Interval{ 0x1EB8, 0x1EB8 }, // L&       LATIN CAPITAL LETTER E WITH DOT BELOW
    Interval{ 0x1EBA, 0x1EBA }, // L&       LATIN CAPITAL LETTER E WITH HOOK ABOVE
    Interval{ 0x1EBC, 0x1EBC }, // L&       LATIN CAPITAL LETTER E WITH TILDE
    Interval{ 0x1EBE, 0x1EBE }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EC0, 0x1EC0 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EC2, 0x1EC2 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EC4, 0x1EC4 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EC6, 0x1EC6 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EC8, 0x1EC8 }, // L&       LATIN CAPITAL LETTER I WITH HOOK ABOVE
    Interval{ 0x1ECA, 0x1ECA }, // L&       LATIN CAPITAL LETTER I WITH DOT BELOW
    Interval{ 0x1ECC, 0x1ECC }, // L&       LATIN CAPITAL LETTER O WITH DOT BELOW
    Interval{ 0x1ECE, 0x1ECE }, // L&       LATIN CAPITAL LETTER O WITH HOOK ABOVE
    Interval{ 0x1ED0, 0x1ED0 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1ED2, 0x1ED2 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1ED4, 0x1ED4 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1ED6, 0x1ED6 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1ED8, 0x1ED8 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EDA, 0x1EDA }, // L&       LATIN CAPITAL LETTER O WITH HORN AND ACUTE
    Interval{ 0x1EDC, 0x1EDC }, // L&       LATIN CAPITAL LETTER O WITH HORN AND GRAVE
    Interval{ 0x1EDE, 0x1EDE }, // L&       LATIN CAPITAL LETTER O WITH HORN AND HOOK ABOVE
    Interval{ 0x1EE0, 0x1EE0 }, // L&       LATIN CAPITAL LETTER O WITH HORN AND TILDE
    Interval{ 0x1EE2, 0x1EE2 }, // L&       LATIN CAPITAL LETTER O WITH HORN AND DOT BELOW
    Interval{ 0x1EE4, 0x1EE4 }, // L&       LATIN CAPITAL LETTER U WITH DOT BELOW
    Interval{ 0x1EE6, 0x1EE6 }, // L&       LATIN CAPITAL LETTER U WITH HOOK ABOVE
    Interval{ 0x1EE8, 0x1EE8 }, // L&       LATIN CAPITAL LETTER U WITH HORN AND ACUTE
    Interval{ 0x1EEA, 0x1EEA }, // L&       LATIN CAPITAL LETTER U WITH HORN AND GRAVE
    Interval{ 0x1EEC, 0x1EEC }, // L&       LATIN CAPITAL LETTER U WITH HORN AND HOOK ABOVE
    Interval{ 0x1EEE, 0x1EEE }, // L&       LATIN CAPITAL LETTER U WITH HORN AND TILDE
    Interval{ 0x1EF0, 0x1EF0 }, // L&       LATIN CAPITAL LETTER U WITH HORN AND DOT BELOW
    Interval{ 0x1EF2, 0x1EF2 }, // L&       LATIN CAPITAL LETTER Y WITH GRAVE
    Interval{ 0x1EF4, 0x1EF4 }, // L&       LATIN CAPITAL LETTER Y WITH DOT BELOW
    Interval{ 0x1EF6, 0x1EF6 }, // L&       LATIN CAPITAL LETTER Y WITH HOOK ABOVE
    Interval{ 0x1EF8, 0x1EF8 }, // L&       LATIN CAPITAL LETTER Y WITH TILDE
    Interval{ 0x1EFA, 0x1EFA }, // L&       LATIN CAPITAL LETTER MIDDLE-WELSH LL
    Interval{ 0x1EFC, 0x1EFC }, // L&       LATIN CAPITAL LETTER MIDDLE-WELSH V
    Interval{ 0x1EFE, 0x1EFE }, // L&       LATIN CAPITAL LETTER Y WITH LOOP
    Interval{ 0x1F08, 0x1F0F }, // L&   [8] GREEK CAPITAL LETTER ALPHA WITH PSILI..GREEK CAPITAL LETTER ALPHA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F18, 0x1F1D }, // L&   [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F28, 0x1F2F }, // L&   [8] GREEK CAPITAL LETTER ETA WITH PSILI..GREEK CAPITAL LETTER ETA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F38, 0x1F3F }, // L&   [8] GREEK CAPITAL LETTER IOTA WITH PSILI..GREEK CAPITAL LETTER IOTA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F48, 0x1F4D }, // L&   [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F59, 0x1F59 }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA
    Interval{ 0x1F5B, 0x1F5B }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Interval{ 0x1F5D, 0x1F5D }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Interval{ 0x1F5F, 0x1F5F }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F68, 0x1F6F }, // L&   [8] GREEK CAPITAL LETTER OMEGA WITH PSILI..GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F88, 0x1F8F }, // L&   [8] GREEK CAPITAL LETTER ALPHA WITH PSILI AND PROSGEGRAMMENI..GREEK CAPITAL LETTER ALPHA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI
    Interval{ 0x1F98, 0x1F9F }, // L&   [8] GREEK CAPITAL LETTER ETA WITH PSILI AND PROSGEGRAMMENI..GREEK CAPITAL LETTER ETA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI
    Interval{ 0x1FA8, 0x1FAF }, // L&   [8] GREEK CAPITAL LETTER OMEGA WITH PSILI AND PROSGEGRAMMENI..GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI
    Interval{ 0x1FB8, 0x1FBC }, // L&   [5] GREEK CAPITAL LETTER ALPHA WITH VRACHY..GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Interval{ 0x1FC8, 0x1FCC }, // L&   [5] GREEK CAPITAL LETTER EPSILON WITH VARIA..GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Interval{ 0x1FD8, 0x1FDB }, // L&   [4] GREEK CAPITAL LETTER IOTA WITH VRACHY..GREEK CAPITAL LETTER IOTA WITH OXIA
    Interval{ 0x1FE8, 0x1FEC }, // L&   [5] GREEK CAPITAL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Interval{ 0x1FF8, 0x1FFC }, // L&   [5] GREEK CAPITAL LETTER OMICRON WITH VARIA..GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Interval{ 0x2126, 0x2126 }, // L&       OHM SIGN
    Interval{ 0x212A, 0x212B }, // L&   [2] KELVIN SIGN..ANGSTROM SIGN
    Interval{ 0x2132, 0x2132 }, // L&       TURNED CAPITAL F
    Interval{ 0x2160, 0x216F }, // Nl  [16] ROMAN NUMERAL ONE..ROMAN NUMERAL ONE THOUSAND
    Interval{ 0x2183, 0x2183 }, // L&       ROMAN NUMERAL REVERSED ONE HUNDRED
    Interval{ 0x24B6, 0x24CF }, // So  [26] CIRCLED LATIN CAPITAL LETTER A..CIRCLED LATIN CAPITAL LETTER Z
    Interval{ 0x2C00, 0x2C2F }, // L&  [48] GLAGOLITIC CAPITAL LETTER AZU..GLAGOLITIC CAPITAL LETTER CAUDATE CHRIVI
    Interval{ 0x2C60, 0x2C60 }, // L&       LATIN CAPITAL LETTER L WITH DOUBLE BAR
    Interval{ 0x2C62, 0x2C64 }, // L&   [3] LATIN CAPITAL LETTER L WITH MIDDLE TILDE..LATIN CAPITAL LETTER R WITH TAIL
    Interval{ 0x2C67, 0x2C67 }, // L&       LATIN CAPITAL LETTER H WITH DESCENDER
    Interval{ 0x2C69, 0x2C69 }, // L&       LATIN CAPITAL LETTER K WITH DESCENDER
    Interval{ 0x2C6B, 0x2C6B }, // L&       LATIN CAPITAL LETTER Z WITH DESCENDER
    Interval{ 0x2C6D, 0x2C70 }, // L&   [4] LATIN CAPITAL LETTER ALPHA..LATIN CAPITAL LETTER TURNED ALPHA
    Interval{ 0x2C72, 0x2C72 }, // L&       LATIN CAPITAL LETTER W WITH HOOK
    Interval{ 0x2C75, 0x2C75 }, // L&       LATIN CAPITAL LETTER HALF H
    Interval{ 0x2C7E, 0x2C80 }, // L&   [3] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC CAPITAL LETTER ALFA
    Interval{ 0x2C82, 0x2C82 }, // L&       COPTIC CAPITAL LETTER VIDA
    Interval{ 0x2C84, 0x2C84 }, // L&       COPTIC CAPITAL LETTER GAMMA
    Interval{ 0x2C86, 0x2C86 }, // L&       COPTIC CAPITAL LETTER DALDA
    Interval{ 0x2C88, 0x2C88 }, // L&       COPTIC CAPITAL LETTER EIE
    Interval{ 0x2C8A, 0x2C8A }, // L&       COPTIC CAPITAL LETTER SOU
    Interval{ 0x2C8C, 0x2C8C }, // L&       COPTIC CAPITAL LETTER ZATA
    Interval{ 0x2C8E, 0x2C8E }, // L&       COPTIC CAPITAL LETTER HATE
    Interval{ 0x2C90, 0x2C90 }, // L&       COPTIC CAPITAL LETTER THETHE
    Interval{ 0x2C92, 0x2C92 }, // L&       COPTIC CAPITAL LETTER IAUDA
    Interval{ 0x2C94, 0x2C94 }, // L&       COPTIC CAPITAL LETTER KAPA
    Interval{ 0x2C96, 0x2C96 }, // L&       COPTIC CAPITAL LETTER LAULA
    Interval{ 0x2C98, 0x2C98 }, // L&       COPTIC CAPITAL LETTER MI
    Interval{ 0x2C9A, 0x2C9A }, // L&       COPTIC CAPITAL LETTER NI
    Interval{ 0x2C9C, 0x2C9C }, // L&       COPTIC CAPITAL LETTER KSI
    Interval{ 0x2C9E, 0x2C9E }, // L&       COPTIC CAPITAL LETTER O
    Interval{ 0x2CA0, 0x2CA0 }, // L&       COPTIC CAPITAL LETTER PI
    Interval{ 0x2CA2, 0x2CA2 }, // L&       COPTIC CAPITAL LETTER RO
    Interval{ 0x2CA4, 0x2CA4 }, // L&       COPTIC CAPITAL LETTER SIMA
    Interval{ 0x2CA6, 0x2CA6 }, // L&       COPTIC CAPITAL LETTER TAU
    Interval{ 0x2CA8, 0x2CA8 }, // L&       COPTIC CAPITAL LETTER UA
    Interval{ 0x2CAA, 0x2CAA }, // L&       COPTIC CAPITAL LETTER FI
    Interval{ 0x2CAC, 0x2CAC }, // L&       COPTIC CAPITAL LETTER KHI
    Interval{ 0x2CAE, 0x2CAE }, // L&       COPTIC CAPITAL LETTER PSI
    Interval{ 0x2CB0, 0x2CB0 }, // L&       COPTIC CAPITAL LETTER OOU
    Interval{ 0x2CB2, 0x2CB2 }, // L&       COPTIC CAPITAL LETTER DIALECT-P ALEF
    Interval{ 0x2CB4, 0x2CB4 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC AIN
    Interval{ 0x2CB6, 0x2CB6 }, // L&       COPTIC CAPITAL LETTER CRYPTOGRAMMIC EIE
    Interval{ 0x2CB8, 0x2CB8 }, // L&       COPTIC CAPITAL LETTER DIALECT-P KAPA
    Interval{ 0x2CBA, 0x2CBA }, // L&       COPTIC CAPITAL LETTER DIALECT-P NI
    Interval{ 0x2CBC, 0x2CBC }, // L&       COPTIC CAPITAL LETTER CRYPTOGRAMMIC NI
    Interval{ 0x2CBE, 0x2CBE }, // L&       COPTIC CAPITAL LETTER OLD COPTIC OOU
    Interval{ 0x2CC0, 0x2CC0 }, // L&       COPTIC CAPITAL LETTER SAMPI
    Interval{ 0x2CC2, 0x2CC2 }, // L&       COPTIC CAPITAL LETTER CROSSED SHEI
    Interval{ 0x2CC4, 0x2CC4 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC SHEI
    Interval{ 0x2CC6, 0x2CC6 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC ESH
    Interval{ 0x2CC8, 0x2CC8 }, // L&       COPTIC CAPITAL LETTER AKHMIMIC KHEI
    Interval{ 0x2CCA, 0x2CCA }, // L&       COPTIC CAPITAL LETTER DIALECT-P HORI
    Interval{ 0x2CCC, 0x2CCC }, // L&       COPTIC CAPITAL LETTER OLD COPTIC HORI
    Interval{ 0x2CCE, 0x2CCE }, // L&       COPTIC CAPITAL LETTER OLD COPTIC HA
    Interval{ 0x2CD0, 0x2CD0 }, // L&       COPTIC CAPITAL LETTER L-SHAPED HA
    Interval{ 0x2CD2, 0x2CD2 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC HEI
    Interval{ 0x2CD4, 0x2CD4 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC HAT
    Interval{ 0x2CD6, 0x2CD6 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC GANGIA
    Interval{ 0x2CD8, 0x2CD8 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC DJA
    Interval{ 0x2CDA, 0x2CDA }, // L&       COPTIC CAPITAL LETTER OLD COPTIC SHIMA
    Interval{ 0x2CDC, 0x2CDC }, // L&       COPTIC CAPITAL LETTER OLD NUBIAN SHIMA
    Interval{ 0x2CDE, 0x2CDE }, // L&       COPTIC CAPITAL LETTER OLD NUBIAN NGI
    Interval{ 0x2CE0, 0x2CE0 }, // L&       COPTIC CAPITAL LETTER OLD NUBIAN NYI
    Interval{ 0x2CE2, 0x2CE2 }, // L&       COPTIC CAPITAL LETTER OLD NUBIAN WAU
    Interval{ 0x2CEB, 0x2CEB }, // L&       COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI
    Interval{ 0x2CED, 0x2CED }, // L&       COPTIC CAPITAL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF2, 0x2CF2 }, // L&       COPTIC CAPITAL LETTER BOHAIRIC KHEI
    Interval{ 0xA640, 0xA640 }, // L&       CYRILLIC CAPITAL LETTER ZEMLYA
    Interval{ 0xA642, 0xA642 }, // L&       CYRILLIC CAPITAL LETTER DZELO
    Interval{ 0xA644, 0xA644 }, // L&       CYRILLIC CAPITAL LETTER REVERSED DZE
    Interval{ 0xA646, 0xA646 }, // L&       CYRILLIC CAPITAL LETTER IOTA
    Interval{ 0xA648, 0xA648 }, // L&       CYRILLIC CAPITAL LETTER DJERV
    Interval{ 0xA64A, 0xA64A }, // L&       CYRILLIC CAPITAL LETTER MONOGRAPH UK
    Interval{ 0xA64C, 0xA64C }, // L&       CYRILLIC CAPITAL LETTER BROAD OMEGA
    Interval{ 0xA64E, 0xA64E }, // L&       CYRILLIC CAPITAL LETTER NEUTRAL YER
    Interval{ 0xA650, 0xA650 }, // L&       CYRILLIC CAPITAL LETTER YERU WITH BACK YER
    Interval{ 0xA652, 0xA652 }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED YAT
    Interval{ 0xA654, 0xA654 }, // L&       CYRILLIC CAPITAL LETTER REVERSED YU
    Interval{ 0xA656, 0xA656 }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED A
    Interval{ 0xA658, 0xA658 }, // L&       CYRILLIC CAPITAL LETTER CLOSED LITTLE YUS
    Interval{ 0xA65A, 0xA65A }, // L&       CYRILLIC CAPITAL LETTER BLENDED YUS
    Interval{ 0xA65C, 0xA65C }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED CLOSED LITTLE YUS
    Interval{ 0xA65E, 0xA65E }, // L&       CYRILLIC CAPITAL LETTER YN
    Interval{ 0xA660, 0xA660 }, // L&       CYRILLIC CAPITAL LETTER REVERSED TSE
    Interval{ 0xA662, 0xA662 }, // L&       CYRILLIC CAPITAL LETTER SOFT DE
    Interval{ 0xA664, 0xA664 }, // L&       CYRILLIC CAPITAL LETTER SOFT EL
    Interval{ 0xA666, 0xA666 }, // L&       CYRILLIC CAPITAL LETTER SOFT EM
    Interval{ 0xA668, 0xA668 }, // L&       CYRILLIC CAPITAL LETTER MONOCULAR O
    Interval{ 0xA66A, 0xA66A }, // L&       CYRILLIC CAPITAL LETTER BINOCULAR O
    Interval{ 0xA66C, 0xA66C }, // L&       CYRILLIC CAPITAL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA680, 0xA680 }, // L&       CYRILLIC CAPITAL LETTER DWE
    Interval{ 0xA682, 0xA682 }, // L&       CYRILLIC CAPITAL LETTER DZWE
    Interval{ 0xA684, 0xA684 }, // L&       CYRILLIC CAPITAL LETTER ZHWE
    Interval{ 0xA686, 0xA686 }, // L&       CYRILLIC CAPITAL LETTER CCHE
    Interval{ 0xA688, 0xA688 }, // L&       CYRILLIC CAPITAL LETTER DZZE
    Interval{ 0xA68A, 0xA68A }, // L&       CYRILLIC CAPITAL LETTER TE WITH MIDDLE HOOK
    Interval{ 0xA68C, 0xA68C }, // L&       CYRILLIC CAPITAL LETTER TWE
    Interval{ 0xA68E, 0xA68E }, // L&       CYRILLIC CAPITAL LETTER TSWE
    Interval{ 0xA690, 0xA690 }, // L&       CYRILLIC CAPITAL LETTER TSSE
    Interval{ 0xA692, 0xA692 }, // L&       CYRILLIC CAPITAL LETTER TCHE
    Interval{ 0xA694, 0xA694 }, // L&       CYRILLIC CAPITAL LETTER HWE
    Interval{ 0xA696, 0xA696 }, // L&       CYRILLIC CAPITAL LETTER SHWE
    Interval{ 0xA698, 0xA698 }, // L&       CYRILLIC CAPITAL LETTER DOUBLE O
    Interval{ 0xA69A, 0xA69A }, // L&       CYRILLIC CAPITAL LETTER CROSSED O
    Interval{ 0xA722, 0xA722 }, // L&       LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF
    Interval{ 0xA724, 0xA724 }, // L&       LATIN CAPITAL LETTER EGYPTOLOGICAL AIN
    Interval{ 0xA726, 0xA726 }, // L&       LATIN CAPITAL LETTER HENG
    Interval{ 0xA728, 0xA728 }, // L&       LATIN CAPITAL LETTER TZ
    Interval{ 0xA72A, 0xA72A }, // L&       LATIN CAPITAL LETTER TRESILLO
    Interval{ 0xA72C, 0xA72C }, // L&       LATIN CAPITAL LETTER CUATRILLO
    Interval{ 0xA72E, 0xA72E }, // L&       LATIN CAPITAL LETTER CUATRILLO WITH COMMA
    Interval{ 0xA732, 0xA732 }, // L&       LATIN CAPITAL LETTER AA
    Interval{ 0xA734, 0xA734 }, // L&       LATIN CAPITAL LETTER AO
    Interval{ 0xA736, 0xA736 }, // L&       LATIN CAPITAL LETTER AU
    Interval{ 0xA738, 0xA738 }, // L&       LATIN CAPITAL LETTER AV
    Interval{ 0xA73A, 0xA73A }, // L&       LATIN CAPITAL LETTER AV WITH HORIZONTAL BAR
    Interval{ 0xA73C, 0xA73C }, // L&       LATIN CAPITAL LETTER AY
    Interval{ 0xA73E, 0xA73E }, // L&       LATIN CAPITAL LETTER REVERSED C WITH DOT
    Interval{ 0xA740, 0xA740 }, // L&       LATIN CAPITAL LETTER K WITH STROKE
    Interval{ 0xA742, 0xA742 }, // L&       LATIN CAPITAL LETTER K WITH DIAGONAL STROKE
    Interval{ 0xA744, 0xA744 }, // L&       LATIN CAPITAL LETTER K WITH STROKE AND DIAGONAL STROKE
    Interval{ 0xA746, 0xA746 }, // L&       LATIN CAPITAL LETTER BROKEN L
    Interval{ 0xA748, 0xA748 }, // L&       LATIN CAPITAL LETTER L WITH HIGH STROKE
    Interval{ 0xA74A, 0xA74A }, // L&       LATIN CAPITAL LETTER O WITH LONG STROKE OVERLAY
    Interval{ 0xA74C, 0xA74C }, // L&       LATIN CAPITAL LETTER O WITH LOOP
    Interval{ 0xA74E, 0xA74E }, // L&       LATIN CAPITAL LETTER OO
    Interval{ 0xA750, 0xA750 }, // L&       LATIN CAPITAL LETTER P WITH STROKE THROUGH DESCENDER
    Interval{ 0xA752, 0xA752 }, // L&       LATIN CAPITAL LETTER P WITH FLOURISH
    Interval{ 0xA754, 0xA754 }, // L&       LATIN CAPITAL LETTER P WITH SQUIRREL TAIL
    Interval{ 0xA756, 0xA756 }, // L&       LATIN CAPITAL LETTER Q WITH STROKE THROUGH DESCENDER
    Interval{ 0xA758, 0xA758 }, // L&       LATIN CAPITAL LETTER Q WITH DIAGONAL STROKE
    Interval{ 0xA75A, 0xA75A }, // L&       LATIN CAPITAL LETTER R ROTUNDA
    Interval{ 0xA75C, 0xA75C }, // L&       LATIN CAPITAL LETTER RUM ROTUNDA
    Interval{ 0xA75E, 0xA75E }, // L&       LATIN CAPITAL LETTER V WITH DIAGONAL STROKE
    Interval{ 0xA760, 0xA760 }, // L&       LATIN CAPITAL LETTER VY
    Interval{ 0xA762, 0xA762 }, // L&       LATIN CAPITAL LETTER VISIGOTHIC Z
    Interval{ 0xA764, 0xA764 }, // L&       LATIN CAPITAL LETTER THORN WITH STROKE
    Interval{ 0xA766, 0xA766 }, // L&       LATIN CAPITAL LETTER THORN WITH STROKE THROUGH DESCENDER
    Interval{ 0xA768, 0xA768 }, // L&       LATIN CAPITAL LETTER VEND
    Interval{ 0xA76A, 0xA76A }, // L&       LATIN CAPITAL LETTER ET
    Interval{ 0xA76C, 0xA76C }, // L&       LATIN CAPITAL LETTER IS
    Interval{ 0xA76E, 0xA76E }, // L&       LATIN CAPITAL LETTER CON
    Interval{ 0xA779, 0xA779 }, // L&       LATIN CAPITAL LETTER INSULAR D
    Interval{ 0xA77B, 0xA77B }, // L&       LATIN CAPITAL LETTER INSULAR F
    Interval{ 0xA77D, 0xA77E }, // L&   [2] LATIN CAPITAL LETTER INSULAR G..LATIN CAPITAL LETTER TURNED INSULAR G
    Interval{ 0xA780, 0xA780 }, // L&       LATIN CAPITAL LETTER TURNED L
    Interval{ 0xA782, 0xA782 }, // L&       LATIN CAPITAL LETTER INSULAR R
    Interval{ 0xA784, 0xA784 }, // L&       LATIN CAPITAL LETTER INSULAR S
    Interval{ 0xA786, 0xA786 }, // L&       LATIN CAPITAL LETTER INSULAR T
    Interval{ 0xA78B, 0xA78B }, // L&       LATIN CAPITAL LETTER SALTILLO
    Interval{ 0xA78D, 0xA78D }, // L&       LATIN CAPITAL LETTER TURNED H
    Interval{ 0xA790, 0xA790 }, // L&       LATIN CAPITAL LETTER N WITH DESCENDER
    Interval{ 0xA792, 0xA792 }, // L&       LATIN CAPITAL LETTER C WITH BAR
    Interval{ 0xA796, 0xA796 }, // L&       LATIN CAPITAL LETTER B WITH FLOURISH
    Interval{ 0xA798, 0xA798 }, // L&       LATIN CAPITAL LETTER F WITH STROKE
    Interval{ 0xA79A, 0xA79A }, // L&       LATIN CAPITAL LETTER VOLAPUK AE
    Interval{ 0xA79C, 0xA79C }, // L&       LATIN CAPITAL LETTER VOLAPUK OE
    Interval{ 0xA79E, 0xA79E }, // L&       LATIN CAPITAL LETTER VOLAPUK UE
    Interval{ 0xA7A0, 0xA7A0 }, // L&       LATIN CAPITAL LETTER G WITH OBLIQUE STROKE
    Interval{ 0xA7A2, 0xA7A2 }, // L&       LATIN CAPITAL LETTER K WITH OBLIQUE STROKE
    Interval{ 0xA7A4, 0xA7A4 }, // L&       LATIN CAPITAL LETTER N WITH OBLIQUE STROKE
    Interval{ 0xA7A6, 0xA7A6 }, // L&       LATIN CAPITAL LETTER R WITH OBLIQUE STROKE
    Interval{ 0xA7A8, 0xA7A8 }, // L&       LATIN CAPITAL LETTER S WITH OBLIQUE STROKE
    Interval{ 0xA7AA, 0xA7AE }, // L&   [5] LATIN CAPITAL LETTER H WITH HOOK..LATIN CAPITAL LETTER SMALL CAPITAL I
    Interval{ 0xA7B0, 0xA7B4 }, // L&   [5] LATIN CAPITAL LETTER TURNED K..LATIN CAPITAL LETTER BETA
    Interval{ 0xA7B6, 0xA7B6 }, // L&       LATIN CAPITAL LETTER OMEGA
    Interval{ 0xA7B8, 0xA7B8 }, // L&       LATIN CAPITAL LETTER U WITH STROKE
    Interval{ 0xA7BA, 0xA7BA }, // L&       LATIN CAPITAL LETTER GLOTTAL A
    Interval{ 0xA7BC, 0xA7BC }, // L&       LATIN CAPITAL LETTER GLOTTAL I
    Interval{ 0xA7BE, 0xA7BE }, // L&       LATIN CAPITAL LETTER GLOTTAL U
    Interval{ 0xA7C0, 0xA7C0 }, // L&       LATIN CAPITAL LETTER OLD POLISH O
    Interval{ 0xA7C2, 0xA7C2 }, // L&       LATIN CAPITAL LETTER ANGLICANA W
    Interval{ 0xA7C4, 0xA7C7 }, // L&   [4] LATIN CAPITAL LETTER C WITH PALATAL HOOK..LATIN CAPITAL LETTER D WITH SHORT STROKE OVERLAY
    Interval{ 0xA7C9, 0xA7C9 }, // L&       LATIN CAPITAL LETTER S WITH SHORT STROKE OVERLAY
    Interval{ 0xA7CB, 0xA7CC }, // L&   [2] LATIN CAPITAL LETTER RAMS HORN..LATIN CAPITAL LETTER S WITH DIAGONAL STROKE
    Interval{ 0xA7CE, 0xA7CE }, // L&       LATIN CAPITAL LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0xA7D0, 0xA7D0 }, // L&       LATIN CAPITAL LETTER CLOSED INSULAR G
    Interval{ 0xA7D2, 0xA7D2 }, // L&       LATIN CAPITAL LETTER DOUBLE THORN
    Interval{ 0xA7D4, 0xA7D4 }, // L&       LATIN CAPITAL LETTER DOUBLE WYNN
    Interval{ 0xA7D6, 0xA7D6 }, // L&       LATIN CAPITAL LETTER MIDDLE SCOTS S
    Interval{ 0xA7D8, 0xA7D8 }, // L&       LATIN CAPITAL LETTER SIGMOID S
    Interval{ 0xA7DA, 0xA7DA }, // L&       LATIN CAPITAL LETTER LAMBDA
    Interval{ 0xA7DC, 0xA7DC }, // L&       LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Interval{ 0xA7F5, 0xA7F5 }, // L&       LATIN CAPITAL LETTER REVERSED HALF H
    Interval{ 0xFF21, 0xFF3A }, // L&  [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Interval{ 0x10400, 0x10427 }, // L&  [40] DESERET CAPITAL LETTER LONG I..DESERET CAPITAL LETTER EW
    Interval{ 0x104B0, 0x104D3 }, // L&  [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Interval{ 0x10570, 0x1057A }, // L&  [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Interval{ 0x1057C, 0x1058A }, // L&  [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Interval{ 0x1058C, 0x10592 }, // L&   [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Interval{ 0x10594, 0x10595 }, // L&   [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Interval{ 0x10C80, 0x10CB2 }, // L&  [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Interval{ 0x10D50, 0x10D65 }, // L&  [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Interval{ 0x118A0, 0x118BF }, // L&  [32] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI CAPITAL LETTER VIYO
    Interval{ 0x16E40, 0x16E5F }, // L&  [32] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN CAPITAL LETTER Y
    Interval{ 0x16EA0, 0x16EB8 }, // L&  [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Interval{ 0x1E900, 0x1E921 }, // L&  [34] ADLAM CAPITAL LETTER ALIF..ADLAM CAPITAL LETTER SHA
}; // }}}
auto static const Changes_When_Titlecased = std::array<Interval, 633>{ // {{{
    Interval{ 0x0061, 0x007A }, // L&  [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Interval{ 0x00B5, 0x00B5 }, // L&       MICRO SIGN
    Interval{ 0x00DF, 0x00F6 }, // L&  [24] LATIN SMALL LETTER SHARP S..LATIN SMALL LETTER O WITH DIAERESIS
    Interval{ 0x00F8, 0x00FF }, // L&   [8] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER Y WITH DIAERESIS
    Interval{ 0x0101, 0x0101 }, // L&       LATIN SMALL LETTER A WITH MACRON
    Interval{ 0x0103, 0x0103 }, // L&       LATIN SMALL LETTER A WITH BREVE
    Interval{ 0x0105, 0x0105 }, // L&       LATIN SMALL LETTER A WITH OGONEK
    Interval{ 0x0107, 0x0107 }, // L&       LATIN SMALL LETTER C WITH ACUTE
    Interval{ 0x0109, 0x0109 }, // L&       LATIN SMALL LETTER C WITH CIRCUMFLEX
    Interval{ 0x010B, 0x010B }, // L&       LATIN SMALL LETTER C WITH DOT ABOVE
    Interval{ 0x010D, 0x010D }, // L&       LATIN SMALL LETTER C WITH CARON
    Interval{ 0x010F, 0x010F }, // L&       LATIN SMALL LETTER D WITH CARON
    Interval{ 0x0111, 0x0111 }, // L&       LATIN SMALL LETTER D WITH STROKE
    Interval{ 0x0113, 0x0113 }, // L&       LATIN SMALL LETTER E WITH MACRON
    Interval{ 0x0115, 0x0115 }, // L&       LATIN SMALL LETTER E WITH BREVE
    Interval{ 0x0117, 0x0117 }, // L&       LATIN SMALL LETTER E WITH DOT ABOVE
    Interval{ 0x0119, 0x0119 }, // L&       LATIN SMALL LETTER E WITH OGONEK
    Interval{ 0x011B, 0x011B }, // L&       LATIN SMALL LETTER E WITH CARON
    Interval{ 0x011D, 0x011D }, // L&       LATIN SMALL LETTER G WITH CIRCUMFLEX
    Interval{ 0x011F, 0x011F }, // L&       LATIN SMALL LETTER G WITH BREVE
    Interval{ 0x0121, 0x0121 }, // L&       LATIN SMALL LETTER G WITH DOT ABOVE
    Interval{ 0x0123, 0x0123 }, // L&       LATIN SMALL LETTER G WITH CEDILLA
    Interval{ 0x0125, 0x0125 }, // L&       LATIN SMALL LETTER H WITH CIRCUMFLEX
    Interval{ 0x0127, 0x0127 }, // L&       LATIN SMALL LETTER H WITH STROKE
    Interval{ 0x0129, 0x0129 }, // L&       LATIN SMALL LETTER I WITH TILDE
    Interval{ 0x012B, 0x012B }, // L&       LATIN SMALL LETTER I WITH MACRON
    Interval{ 0x012D, 0x012D }, // L&       LATIN SMALL LETTER I WITH BREVE
    Interval{ 0x012F, 0x012F }, // L&       LATIN SMALL LETTER I WITH OGONEK
    Interval{ 0x0131, 0x0131 }, // L&       LATIN SMALL LETTER DOTLESS I
    Interval{ 0x0133, 0x0133 }, // L&       LATIN SMALL LIGATURE IJ
    Interval{ 0x0135, 0x0135 }, // L&       LATIN SMALL LETTER J WITH CIRCUMFLEX
    Interval{ 0x0137, 0x0137 }, // L&       LATIN SMALL LETTER K WITH CEDILLA
    Interval{ 0x013A, 0x013A }, // L&       LATIN SMALL LETTER L WITH ACUTE
    Interval{ 0x013C, 0x013C }, // L&       LATIN SMALL LETTER L WITH CEDILLA
    Interval{ 0x013E, 0x013E }, // L&       LATIN SMALL LETTER L WITH CARON
    Interval{ 0x0140, 0x0140 }, // L&       LATIN SMALL LETTER L WITH MIDDLE DOT
    Interval{ 0x0142, 0x0142 }, // L&       LATIN SMALL LETTER L WITH STROKE
    Interval{ 0x0144, 0x0144 }, // L&       LATIN SMALL LETTER N WITH ACUTE
    Interval{ 0x0146, 0x0146 }, // L&       LATIN SMALL LETTER N WITH CEDILLA
    Interval{ 0x0148, 0x0149 }, // L&   [2] LATIN SMALL LETTER N WITH CARON..LATIN SMALL LETTER N PRECEDED BY APOSTROPHE
    Interval{ 0x014B, 0x014B }, // L&       LATIN SMALL LETTER ENG
    Interval{ 0x014D, 0x014D }, // L&       LATIN SMALL LETTER O WITH MACRON
    Interval{ 0x014F, 0x014F }, // L&       LATIN SMALL LETTER O WITH BREVE
    Interval{ 0x0151, 0x0151 }, // L&       LATIN SMALL LETTER O WITH DOUBLE ACUTE
    Interval{ 0x0153, 0x0153 }, // L&       LATIN SMALL LIGATURE OE
    Interval{ 0x0155, 0x0155 }, // L&       LATIN SMALL LETTER R WITH ACUTE
    Interval{ 0x0157, 0x0157 }, // L&       LATIN SMALL LETTER R WITH CEDILLA
    Interval{ 0x0159, 0x0159 }, // L&       LATIN SMALL LETTER R WITH CARON
    Interval{ 0x015B, 0x015B }, // L&       LATIN SMALL LETTER S WITH ACUTE
    Interval{ 0x015D, 0x015D }, // L&       LATIN SMALL LETTER S WITH CIRCUMFLEX
    Interval{ 0x015F, 0x015F }, // L&       LATIN SMALL LETTER S WITH CEDILLA
    Interval{ 0x0161, 0x0161 }, // L&       LATIN SMALL LETTER S WITH CARON
    Interval{ 0x0163, 0x0163 }, // L&       LATIN SMALL LETTER T WITH CEDILLA
    Interval{ 0x0165, 0x0165 }, // L&       LATIN SMALL LETTER T WITH CARON
    Interval{ 0x0167, 0x0167 }, // L&       LATIN SMALL LETTER T WITH STROKE
    Interval{ 0x0169, 0x0169 }, // L&       LATIN SMALL LETTER U WITH TILDE
    Interval{ 0x016B, 0x016B }, // L&       LATIN SMALL LETTER U WITH MACRON
    Interval{ 0x016D, 0x016D }, // L&       LATIN SMALL LETTER U WITH BREVE
    Interval{ 0x016F, 0x016F }, // L&       LATIN SMALL LETTER U WITH RING ABOVE
    Interval{ 0x0171, 0x0171 }, // L&       LATIN SMALL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x0173, 0x0173 }, // L&       LATIN SMALL LETTER U WITH OGONEK
    Interval{ 0x0175, 0x0175 }, // L&       LATIN SMALL LETTER W WITH CIRCUMFLEX
    Interval{ 0x0177, 0x0177 }, // L&       LATIN SMALL LETTER Y WITH CIRCUMFLEX
    Interval{ 0x017A, 0x017A }, // L&       LATIN SMALL LETTER Z WITH ACUTE
    Interval{ 0x017C, 0x017C }, // L&       LATIN SMALL LETTER Z WITH DOT ABOVE
    Interval{ 0x017E, 0x0180 }, // L&   [3] LATIN SMALL LETTER Z WITH CARON..LATIN SMALL LETTER B WITH STROKE
    Interval{ 0x0183, 0x0183 }, // L&       LATIN SMALL LETTER B WITH TOPBAR
    Interval{ 0x0185, 0x0185 }, // L&       LATIN SMALL LETTER TONE SIX
    Interval{ 0x0188, 0x0188 }, // L&       LATIN SMALL LETTER C WITH HOOK
    Interval{ 0x018C, 0x018C }, // L&       LATIN SMALL LETTER D WITH TOPBAR
    Interval{ 0x0192, 0x0192 }, // L&       LATIN SMALL LETTER F WITH HOOK
    Interval{ 0x0195, 0x0195 }, // L&       LATIN SMALL LETTER HV
    Interval{ 0x0199, 0x019B }, // L&   [3] LATIN SMALL LETTER K WITH HOOK..LATIN SMALL LETTER LAMBDA WITH STROKE
    Interval{ 0x019E, 0x019E }, // L&       LATIN SMALL LETTER N WITH LONG RIGHT LEG
    Interval{ 0x01A1, 0x01A1 }, // L&       LATIN SMALL LETTER O WITH HORN
    Interval{ 0x01A3, 0x01A3 }, // L&       LATIN SMALL LETTER OI
    Interval{ 0x01A5, 0x01A5 }, // L&       LATIN SMALL LETTER P WITH HOOK
    Interval{ 0x01A8, 0x01A8 }, // L&       LATIN SMALL LETTER TONE TWO
    Interval{ 0x01AD, 0x01AD }, // L&       LATIN SMALL LETTER T WITH HOOK
    Interval{ 0x01B0, 0x01B0 }, // L&       LATIN SMALL LETTER U WITH HORN
    Interval{ 0x01B4, 0x01B4 }, // L&       LATIN SMALL LETTER Y WITH HOOK
    Interval{ 0x01B6, 0x01B6 }, // L&       LATIN SMALL LETTER Z WITH STROKE
    Interval{ 0x01B9, 0x01B9 }, // L&       LATIN SMALL LETTER EZH REVERSED
    Interval{ 0x01BD, 0x01BD }, // L&       LATIN SMALL LETTER TONE FIVE
    Interval{ 0x01BF, 0x01BF }, // L&       LATIN LETTER WYNN
    Interval{ 0x01C4, 0x01C4 }, // L&       LATIN CAPITAL LETTER DZ WITH CARON
    Interval{ 0x01C6, 0x01C7 }, // L&   [2] LATIN SMALL LETTER DZ WITH CARON..LATIN CAPITAL LETTER LJ
    Interval{ 0x01C9, 0x01CA }, // L&   [2] LATIN SMALL LETTER LJ..LATIN CAPITAL LETTER NJ
    Interval{ 0x01CC, 0x01CC }, // L&       LATIN SMALL LETTER NJ
    Interval{ 0x01CE, 0x01CE }, // L&       LATIN SMALL LETTER A WITH CARON
    Interval{ 0x01D0, 0x01D0 }, // L&       LATIN SMALL LETTER I WITH CARON
    Interval{ 0x01D2, 0x01D2 }, // L&       LATIN SMALL LETTER O WITH CARON
    Interval{ 0x01D4, 0x01D4 }, // L&       LATIN SMALL LETTER U WITH CARON
    Interval{ 0x01D6, 0x01D6 }, // L&       LATIN SMALL LETTER U WITH DIAERESIS AND MACRON
    Interval{ 0x01D8, 0x01D8 }, // L&       LATIN SMALL LETTER U WITH DIAERESIS AND ACUTE
    Interval{ 0x01DA, 0x01DA }, // L&       LATIN SMALL LETTER U WITH DIAERESIS AND CARON
    Interval{ 0x01DC, 0x01DD }, // L&   [2] LATIN SMALL LETTER U WITH DIAERESIS AND GRAVE..LATIN SMALL LETTER TURNED E
    Interval{ 0x01DF, 0x01DF }, // L&       LATIN SMALL LETTER A WITH DIAERESIS AND MACRON
    Interval{ 0x01E1, 0x01E1 }, // L&       LATIN SMALL LETTER A WITH DOT ABOVE AND MACRON
    Interval{ 0x01E3, 0x01E3 }, // L&       LATIN SMALL LETTER AE WITH MACRON
    Interval{ 0x01E5, 0x01E5 }, // L&       LATIN SMALL LETTER G WITH STROKE
    Interval{ 0x01E7, 0x01E7 }, // L&       LATIN SMALL LETTER G WITH CARON
    Interval{ 0x01E9, 0x01E9 }, // L&       LATIN SMALL LETTER K WITH CARON
    Interval{ 0x01EB, 0x01EB }, // L&       LATIN SMALL LETTER O WITH OGONEK
    Interval{ 0x01ED, 0x01ED }, // L&       LATIN SMALL LETTER O WITH OGONEK AND MACRON
    Interval{ 0x01EF, 0x01F1 }, // L&   [3] LATIN SMALL LETTER EZH WITH CARON..LATIN CAPITAL LETTER DZ
    Interval{ 0x01F3, 0x01F3 }, // L&       LATIN SMALL LETTER DZ
    Interval{ 0x01F5, 0x01F5 }, // L&       LATIN SMALL LETTER G WITH ACUTE
    Interval{ 0x01F9, 0x01F9 }, // L&       LATIN SMALL LETTER N WITH GRAVE
    Interval{ 0x01FB, 0x01FB }, // L&       LATIN SMALL LETTER A WITH RING ABOVE AND ACUTE
    Interval{ 0x01FD, 0x01FD }, // L&       LATIN SMALL LETTER AE WITH ACUTE
    Interval{ 0x01FF, 0x01FF }, // L&       LATIN SMALL LETTER O WITH STROKE AND ACUTE
    Interval{ 0x0201, 0x0201 }, // L&       LATIN SMALL LETTER A WITH DOUBLE GRAVE
    Interval{ 0x0203, 0x0203 }, // L&       LATIN SMALL LETTER A WITH INVERTED BREVE
    Interval{ 0x0205, 0x0205 }, // L&       LATIN SMALL LETTER E WITH DOUBLE GRAVE
    Interval{ 0x0207, 0x0207 }, // L&       LATIN SMALL LETTER E WITH INVERTED BREVE
    Interval{ 0x0209, 0x0209 }, // L&       LATIN SMALL LETTER I WITH DOUBLE GRAVE
    Interval{ 0x020B, 0x020B }, // L&       LATIN SMALL LETTER I WITH INVERTED BREVE
    Interval{ 0x020D, 0x020D }, // L&       LATIN SMALL LETTER O WITH DOUBLE GRAVE
    Interval{ 0x020F, 0x020F }, // L&       LATIN SMALL LETTER O WITH INVERTED BREVE
    Interval{ 0x0211, 0x0211 }, // L&       LATIN SMALL LETTER R WITH DOUBLE GRAVE
    Interval{ 0x0213, 0x0213 }, // L&       LATIN SMALL LETTER R WITH INVERTED BREVE
    Interval{ 0x0215, 0x0215 }, // L&       LATIN SMALL LETTER U WITH DOUBLE GRAVE
    Interval{ 0x0217, 0x0217 }, // L&       LATIN SMALL LETTER U WITH INVERTED BREVE
    Interval{ 0x0219, 0x0219 }, // L&       LATIN SMALL LETTER S WITH COMMA BELOW
    Interval{ 0x021B, 0x021B }, // L&       LATIN SMALL LETTER T WITH COMMA BELOW
    Interval{ 0x021D, 0x021D }, // L&       LATIN SMALL LETTER YOGH
    Interval{ 0x021F, 0x021F }, // L&       LATIN SMALL LETTER H WITH CARON
    Interval{ 0x0223, 0x0223 }, // L&       LATIN SMALL LETTER OU
    Interval{ 0x0225, 0x0225 }, // L&       LATIN SMALL LETTER Z WITH HOOK
    Interval{ 0x0227, 0x0227 }, // L&       LATIN SMALL LETTER A WITH DOT ABOVE
    Interval{ 0x0229, 0x0229 }, // L&       LATIN SMALL LETTER E WITH CEDILLA
    Interval{ 0x022B, 0x022B }, // L&       LATIN SMALL LETTER O WITH DIAERESIS AND MACRON
    Interval{ 0x022D, 0x022D }, // L&       LATIN SMALL LETTER O WITH TILDE AND MACRON
    Interval{ 0x022F, 0x022F }, // L&       LATIN SMALL LETTER O WITH DOT ABOVE
    Interval{ 0x0231, 0x0231 }, // L&       LATIN SMALL LETTER O WITH DOT ABOVE AND MACRON
    Interval{ 0x0233, 0x0233 }, // L&       LATIN SMALL LETTER Y WITH MACRON
    Interval{ 0x023C, 0x023C }, // L&       LATIN SMALL LETTER C WITH STROKE
    Interval{ 0x023F, 0x0240 }, // L&   [2] LATIN SMALL LETTER S WITH SWASH TAIL..LATIN SMALL LETTER Z WITH SWASH TAIL
    Interval{ 0x0242, 0x0242 }, // L&       LATIN SMALL LETTER GLOTTAL STOP
    Interval{ 0x0247, 0x0247 }, // L&       LATIN SMALL LETTER E WITH STROKE
    Interval{ 0x0249, 0x0249 }, // L&       LATIN SMALL LETTER J WITH STROKE
    Interval{ 0x024B, 0x024B }, // L&       LATIN SMALL LETTER Q WITH HOOK TAIL
    Interval{ 0x024D, 0x024D }, // L&       LATIN SMALL LETTER R WITH STROKE
    Interval{ 0x024F, 0x0254 }, // L&   [6] LATIN SMALL LETTER Y WITH STROKE..LATIN SMALL LETTER OPEN O
    Interval{ 0x0256, 0x0257 }, // L&   [2] LATIN SMALL LETTER D WITH TAIL..LATIN SMALL LETTER D WITH HOOK
    Interval{ 0x0259, 0x0259 }, // L&       LATIN SMALL LETTER SCHWA
    Interval{ 0x025B, 0x025C }, // L&   [2] LATIN SMALL LETTER OPEN E..LATIN SMALL LETTER REVERSED OPEN E
    Interval{ 0x0260, 0x0261 }, // L&   [2] LATIN SMALL LETTER G WITH HOOK..LATIN SMALL LETTER SCRIPT G
    Interval{ 0x0263, 0x0266 }, // L&   [4] LATIN SMALL LETTER GAMMA..LATIN SMALL LETTER H WITH HOOK
    Interval{ 0x0268, 0x026C }, // L&   [5] LATIN SMALL LETTER I WITH STROKE..LATIN SMALL LETTER L WITH BELT
    Interval{ 0x026F, 0x026F }, // L&       LATIN SMALL LETTER TURNED M
    Interval{ 0x0271, 0x0272 }, // L&   [2] LATIN SMALL LETTER M WITH HOOK..LATIN SMALL LETTER N WITH LEFT HOOK
    Interval{ 0x0275, 0x0275 }, // L&       LATIN SMALL LETTER BARRED O
    Interval{ 0x027D, 0x027D }, // L&       LATIN SMALL LETTER R WITH TAIL
    Interval{ 0x0280, 0x0280 }, // L&       LATIN LETTER SMALL CAPITAL R
    Interval{ 0x0282, 0x0283 }, // L&   [2] LATIN SMALL LETTER S WITH HOOK..LATIN SMALL LETTER ESH
    Interval{ 0x0287, 0x028C }, // L&   [6] LATIN SMALL LETTER TURNED T..LATIN SMALL LETTER TURNED V
    Interval{ 0x0292, 0x0292 }, // L&       LATIN SMALL LETTER EZH
    Interval{ 0x029D, 0x029E }, // L&   [2] LATIN SMALL LETTER J WITH CROSSED-TAIL..LATIN SMALL LETTER TURNED K
    Interval{ 0x0345, 0x0345 }, // Mn       COMBINING GREEK YPOGEGRAMMENI
    Interval{ 0x0371, 0x0371 }, // L&       GREEK SMALL LETTER HETA
    Interval{ 0x0373, 0x0373 }, // L&       GREEK SMALL LETTER ARCHAIC SAMPI
    Interval{ 0x0377, 0x0377 }, // L&       GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037B, 0x037D }, // L&   [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Interval{ 0x0390, 0x0390 }, // L&       GREEK SMALL LETTER IOTA WITH DIALYTIKA AND TONOS
    Interval{ 0x03AC, 0x03CE }, // L&  [35] GREEK SMALL LETTER ALPHA WITH TONOS..GREEK SMALL LETTER OMEGA WITH TONOS
    Interval{ 0x03D0, 0x03D1 }, // L&   [2] GREEK BETA SYMBOL..GREEK THETA SYMBOL
    Interval{ 0x03D5, 0x03D7 }, // L&   [3] GREEK PHI SYMBOL..GREEK KAI SYMBOL
    Interval{ 0x03D9, 0x03D9 }, // L&       GREEK SMALL LETTER ARCHAIC KOPPA
    Interval{ 0x03DB, 0x03DB }, // L&       GREEK SMALL LETTER STIGMA
    Interval{ 0x03DD, 0x03DD }, // L&       GREEK SMALL LETTER DIGAMMA
    Interval{ 0x03DF, 0x03DF }, // L&       GREEK SMALL LETTER KOPPA
    Interval{ 0x03E1, 0x03E1 }, // L&       GREEK SMALL LETTER SAMPI
    Interval{ 0x03E3, 0x03E3 }, // L&       COPTIC SMALL LETTER SHEI
    Interval{ 0x03E5, 0x03E5 }, // L&       COPTIC SMALL LETTER FEI
    Interval{ 0x03E7, 0x03E7 }, // L&       COPTIC SMALL LETTER KHEI
    Interval{ 0x03E9, 0x03E9 }, // L&       COPTIC SMALL LETTER HORI
    Interval{ 0x03EB, 0x03EB }, // L&       COPTIC SMALL LETTER GANGIA
    Interval{ 0x03ED, 0x03ED }, // L&       COPTIC SMALL LETTER SHIMA
    Interval{ 0x03EF, 0x03F3 }, // L&   [5] COPTIC SMALL LETTER DEI..GREEK LETTER YOT
    Interval{ 0x03F5, 0x03F5 }, // L&       GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F8, 0x03F8 }, // L&       GREEK SMALL LETTER SHO
    Interval{ 0x03FB, 0x03FB }, // L&       GREEK SMALL LETTER SAN
    Interval{ 0x0430, 0x045F }, // L&  [48] CYRILLIC SMALL LETTER A..CYRILLIC SMALL LETTER DZHE
    Interval{ 0x0461, 0x0461 }, // L&       CYRILLIC SMALL LETTER OMEGA
    Interval{ 0x0463, 0x0463 }, // L&       CYRILLIC SMALL LETTER YAT
    Interval{ 0x0465, 0x0465 }, // L&       CYRILLIC SMALL LETTER IOTIFIED E
    Interval{ 0x0467, 0x0467 }, // L&       CYRILLIC SMALL LETTER LITTLE YUS
    Interval{ 0x0469, 0x0469 }, // L&       CYRILLIC SMALL LETTER IOTIFIED LITTLE YUS
    Interval{ 0x046B, 0x046B }, // L&       CYRILLIC SMALL LETTER BIG YUS
    Interval{ 0x046D, 0x046D }, // L&       CYRILLIC SMALL LETTER IOTIFIED BIG YUS
    Interval{ 0x046F, 0x046F }, // L&       CYRILLIC SMALL LETTER KSI
    Interval{ 0x0471, 0x0471 }, // L&       CYRILLIC SMALL LETTER PSI
    Interval{ 0x0473, 0x0473 }, // L&       CYRILLIC SMALL LETTER FITA
    Interval{ 0x0475, 0x0475 }, // L&       CYRILLIC SMALL LETTER IZHITSA
    Interval{ 0x0477, 0x0477 }, // L&       CYRILLIC SMALL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT
    Interval{ 0x0479, 0x0479 }, // L&       CYRILLIC SMALL LETTER UK
    Interval{ 0x047B, 0x047B }, // L&       CYRILLIC SMALL LETTER ROUND OMEGA
    Interval{ 0x047D, 0x047D }, // L&       CYRILLIC SMALL LETTER OMEGA WITH TITLO
    Interval{ 0x047F, 0x047F }, // L&       CYRILLIC SMALL LETTER OT
    Interval{ 0x0481, 0x0481 }, // L&       CYRILLIC SMALL LETTER KOPPA
    Interval{ 0x048B, 0x048B }, // L&       CYRILLIC SMALL LETTER SHORT I WITH TAIL
    Interval{ 0x048D, 0x048D }, // L&       CYRILLIC SMALL LETTER SEMISOFT SIGN
    Interval{ 0x048F, 0x048F }, // L&       CYRILLIC SMALL LETTER ER WITH TICK
    Interval{ 0x0491, 0x0491 }, // L&       CYRILLIC SMALL LETTER GHE WITH UPTURN
    Interval{ 0x0493, 0x0493 }, // L&       CYRILLIC SMALL LETTER GHE WITH STROKE
    Interval{ 0x0495, 0x0495 }, // L&       CYRILLIC SMALL LETTER GHE WITH MIDDLE HOOK
    Interval{ 0x0497, 0x0497 }, // L&       CYRILLIC SMALL LETTER ZHE WITH DESCENDER
    Interval{ 0x0499, 0x0499 }, // L&       CYRILLIC SMALL LETTER ZE WITH DESCENDER
    Interval{ 0x049B, 0x049B }, // L&       CYRILLIC SMALL LETTER KA WITH DESCENDER
    Interval{ 0x049D, 0x049D }, // L&       CYRILLIC SMALL LETTER KA WITH VERTICAL STROKE
    Interval{ 0x049F, 0x049F }, // L&       CYRILLIC SMALL LETTER KA WITH STROKE
    Interval{ 0x04A1, 0x04A1 }, // L&       CYRILLIC SMALL LETTER BASHKIR KA
    Interval{ 0x04A3, 0x04A3 }, // L&       CYRILLIC SMALL LETTER EN WITH DESCENDER
    Interval{ 0x04A5, 0x04A5 }, // L&       CYRILLIC SMALL LIGATURE EN GHE
    Interval{ 0x04A7, 0x04A7 }, // L&       CYRILLIC SMALL LETTER PE WITH MIDDLE HOOK
    Interval{ 0x04A9, 0x04A9 }, // L&       CYRILLIC SMALL LETTER ABKHASIAN HA
    Interval{ 0x04AB, 0x04AB }, // L&       CYRILLIC SMALL LETTER ES WITH DESCENDER
    Interval{ 0x04AD, 0x04AD }, // L&       CYRILLIC SMALL LETTER TE WITH DESCENDER
    Interval{ 0x04AF, 0x04AF }, // L&       CYRILLIC SMALL LETTER STRAIGHT U
    Interval{ 0x04B1, 0x04B1 }, // L&       CYRILLIC SMALL LETTER STRAIGHT U WITH STROKE
    Interval{ 0x04B3, 0x04B3 }, // L&       CYRILLIC SMALL LETTER HA WITH DESCENDER
    Interval{ 0x04B5, 0x04B5 }, // L&       CYRILLIC SMALL LIGATURE TE TSE
    Interval{ 0x04B7, 0x04B7 }, // L&       CYRILLIC SMALL LETTER CHE WITH DESCENDER
    Interval{ 0x04B9, 0x04B9 }, // L&       CYRILLIC SMALL LETTER CHE WITH VERTICAL STROKE
    Interval{ 0x04BB, 0x04BB }, // L&       CYRILLIC SMALL LETTER SHHA
    Interval{ 0x04BD, 0x04BD }, // L&       CYRILLIC SMALL LETTER ABKHASIAN CHE
    Interval{ 0x04BF, 0x04BF }, // L&       CYRILLIC SMALL LETTER ABKHASIAN CHE WITH DESCENDER
    Interval{ 0x04C2, 0x04C2 }, // L&       CYRILLIC SMALL LETTER ZHE WITH BREVE
    Interval{ 0x04C4, 0x04C4 }, // L&       CYRILLIC SMALL LETTER KA WITH HOOK
    Interval{ 0x04C6, 0x04C6 }, // L&       CYRILLIC SMALL LETTER EL WITH TAIL
    Interval{ 0x04C8, 0x04C8 }, // L&       CYRILLIC SMALL LETTER EN WITH HOOK
    Interval{ 0x04CA, 0x04CA }, // L&       CYRILLIC SMALL LETTER EN WITH TAIL
    Interval{ 0x04CC, 0x04CC }, // L&       CYRILLIC SMALL LETTER KHAKASSIAN CHE
    Interval{ 0x04CE, 0x04CF }, // L&   [2] CYRILLIC SMALL LETTER EM WITH TAIL..CYRILLIC SMALL LETTER PALOCHKA
    Interval{ 0x04D1, 0x04D1 }, // L&       CYRILLIC SMALL LETTER A WITH BREVE
    Interval{ 0x04D3, 0x04D3 }, // L&       CYRILLIC SMALL LETTER A WITH DIAERESIS
    Interval{ 0x04D5, 0x04D5 }, // L&       CYRILLIC SMALL LIGATURE A IE
    Interval{ 0x04D7, 0x04D7 }, // L&       CYRILLIC SMALL LETTER IE WITH BREVE
    Interval{ 0x04D9, 0x04D9 }, // L&       CYRILLIC SMALL LETTER SCHWA
    Interval{ 0x04DB, 0x04DB }, // L&       CYRILLIC SMALL LETTER SCHWA WITH DIAERESIS
    Interval{ 0x04DD, 0x04DD }, // L&       CYRILLIC SMALL LETTER ZHE WITH DIAERESIS
    Interval{ 0x04DF, 0x04DF }, // L&       CYRILLIC SMALL LETTER ZE WITH DIAERESIS
    Interval{ 0x04E1, 0x04E1 }, // L&       CYRILLIC SMALL LETTER ABKHASIAN DZE
    Interval{ 0x04E3, 0x04E3 }, // L&       CYRILLIC SMALL LETTER I WITH MACRON
    Interval{ 0x04E5, 0x04E5 }, // L&       CYRILLIC SMALL LETTER I WITH DIAERESIS
    Interval{ 0x04E7, 0x04E7 }, // L&       CYRILLIC SMALL LETTER O WITH DIAERESIS
    Interval{ 0x04E9, 0x04E9 }, // L&       CYRILLIC SMALL LETTER BARRED O
    Interval{ 0x04EB, 0x04EB }, // L&       CYRILLIC SMALL LETTER BARRED O WITH DIAERESIS
    Interval{ 0x04ED, 0x04ED }, // L&       CYRILLIC SMALL LETTER E WITH DIAERESIS
    Interval{ 0x04EF, 0x04EF }, // L&       CYRILLIC SMALL LETTER U WITH MACRON
    Interval{ 0x04F1, 0x04F1 }, // L&       CYRILLIC SMALL LETTER U WITH DIAERESIS
    Interval{ 0x04F3, 0x04F3 }, // L&       CYRILLIC SMALL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x04F5, 0x04F5 }, // L&       CYRILLIC SMALL LETTER CHE WITH DIAERESIS
    Interval{ 0x04F7, 0x04F7 }, // L&       CYRILLIC SMALL LETTER GHE WITH DESCENDER
    Interval{ 0x04F9, 0x04F9 }, // L&       CYRILLIC SMALL LETTER YERU WITH DIAERESIS
    Interval{ 0x04FB, 0x04FB }, // L&       CYRILLIC SMALL LETTER GHE WITH STROKE AND HOOK
    Interval{ 0x04FD, 0x04FD }, // L&       CYRILLIC SMALL LETTER HA WITH HOOK
    Interval{ 0x04FF, 0x04FF }, // L&       CYRILLIC SMALL LETTER HA WITH STROKE
    Interval{ 0x0501, 0x0501 }, // L&       CYRILLIC SMALL LETTER KOMI DE
    Interval{ 0x0503, 0x0503 }, // L&       CYRILLIC SMALL LETTER KOMI DJE
    Interval{ 0x0505, 0x0505 }, // L&       CYRILLIC SMALL LETTER KOMI ZJE
    Interval{ 0x0507, 0x0507 }, // L&       CYRILLIC SMALL LETTER KOMI DZJE
    Interval{ 0x0509, 0x0509 }, // L&       CYRILLIC SMALL LETTER KOMI LJE
    Interval{ 0x050B, 0x050B }, // L&       CYRILLIC SMALL LETTER KOMI NJE
    Interval{ 0x050D, 0x050D }, // L&       CYRILLIC SMALL LETTER KOMI SJE
    Interval{ 0x050F, 0x050F }, // L&       CYRILLIC SMALL LETTER KOMI TJE
    Interval{ 0x0511, 0x0511 }, // L&       CYRILLIC SMALL LETTER REVERSED ZE
    Interval{ 0x0513, 0x0513 }, // L&       CYRILLIC SMALL LETTER EL WITH HOOK
    Interval{ 0x0515, 0x0515 }, // L&       CYRILLIC SMALL LETTER LHA
    Interval{ 0x0517, 0x0517 }, // L&       CYRILLIC SMALL LETTER RHA
    Interval{ 0x0519, 0x0519 }, // L&       CYRILLIC SMALL LETTER YAE
    Interval{ 0x051B, 0x051B }, // L&       CYRILLIC SMALL LETTER QA
    Interval{ 0x051D, 0x051D }, // L&       CYRILLIC SMALL LETTER WE
    Interval{ 0x051F, 0x051F }, // L&       CYRILLIC SMALL LETTER ALEUT KA
    Interval{ 0x0521, 0x0521 }, // L&       CYRILLIC SMALL LETTER EL WITH MIDDLE HOOK
    Interval{ 0x0523, 0x0523 }, // L&       CYRILLIC SMALL LETTER EN WITH MIDDLE HOOK
    Interval{ 0x0525, 0x0525 }, // L&       CYRILLIC SMALL LETTER PE WITH DESCENDER
    Interval{ 0x0527, 0x0527 }, // L&       CYRILLIC SMALL LETTER SHHA WITH DESCENDER
    Interval{ 0x0529, 0x0529 }, // L&       CYRILLIC SMALL LETTER EN WITH LEFT HOOK
    Interval{ 0x052B, 0x052B }, // L&       CYRILLIC SMALL LETTER DZZHE
    Interval{ 0x052D, 0x052D }, // L&       CYRILLIC SMALL LETTER DCHE
    Interval{ 0x052F, 0x052F }, // L&       CYRILLIC SMALL LETTER EL WITH DESCENDER
    Interval{ 0x0561, 0x0587 }, // L&  [39] ARMENIAN SMALL LETTER AYB..ARMENIAN SMALL LIGATURE ECH YIWN
    Interval{ 0x13F8, 0x13FD }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1C80, 0x1C88 }, // L&   [9] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER UNBLENDED UK
    Interval{ 0x1C8A, 0x1C8A }, // L&       CYRILLIC SMALL LETTER TJE
    Interval{ 0x1D79, 0x1D79 }, // L&       LATIN SMALL LETTER INSULAR G
    Interval{ 0x1D7D, 0x1D7D }, // L&       LATIN SMALL LETTER P WITH STROKE
    Interval{ 0x1D8E, 0x1D8E }, // L&       LATIN SMALL LETTER Z WITH PALATAL HOOK
    Interval{ 0x1E01, 0x1E01 }, // L&       LATIN SMALL LETTER A WITH RING BELOW
    Interval{ 0x1E03, 0x1E03 }, // L&       LATIN SMALL LETTER B WITH DOT ABOVE
    Interval{ 0x1E05, 0x1E05 }, // L&       LATIN SMALL LETTER B WITH DOT BELOW
    Interval{ 0x1E07, 0x1E07 }, // L&       LATIN SMALL LETTER B WITH LINE BELOW
    Interval{ 0x1E09, 0x1E09 }, // L&       LATIN SMALL LETTER C WITH CEDILLA AND ACUTE
    Interval{ 0x1E0B, 0x1E0B }, // L&       LATIN SMALL LETTER D WITH DOT ABOVE
    Interval{ 0x1E0D, 0x1E0D }, // L&       LATIN SMALL LETTER D WITH DOT BELOW
    Interval{ 0x1E0F, 0x1E0F }, // L&       LATIN SMALL LETTER D WITH LINE BELOW
    Interval{ 0x1E11, 0x1E11 }, // L&       LATIN SMALL LETTER D WITH CEDILLA
    Interval{ 0x1E13, 0x1E13 }, // L&       LATIN SMALL LETTER D WITH CIRCUMFLEX BELOW
    Interval{ 0x1E15, 0x1E15 }, // L&       LATIN SMALL LETTER E WITH MACRON AND GRAVE
    Interval{ 0x1E17, 0x1E17 }, // L&       LATIN SMALL LETTER E WITH MACRON AND ACUTE
    Interval{ 0x1E19, 0x1E19 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX BELOW
    Interval{ 0x1E1B, 0x1E1B }, // L&       LATIN SMALL LETTER E WITH TILDE BELOW
    Interval{ 0x1E1D, 0x1E1D }, // L&       LATIN SMALL LETTER E WITH CEDILLA AND BREVE
    Interval{ 0x1E1F, 0x1E1F }, // L&       LATIN SMALL LETTER F WITH DOT ABOVE
    Interval{ 0x1E21, 0x1E21 }, // L&       LATIN SMALL LETTER G WITH MACRON
    Interval{ 0x1E23, 0x1E23 }, // L&       LATIN SMALL LETTER H WITH DOT ABOVE
    Interval{ 0x1E25, 0x1E25 }, // L&       LATIN SMALL LETTER H WITH DOT BELOW
    Interval{ 0x1E27, 0x1E27 }, // L&       LATIN SMALL LETTER H WITH DIAERESIS
    Interval{ 0x1E29, 0x1E29 }, // L&       LATIN SMALL LETTER H WITH CEDILLA
    Interval{ 0x1E2B, 0x1E2B }, // L&       LATIN SMALL LETTER H WITH BREVE BELOW
    Interval{ 0x1E2D, 0x1E2D }, // L&       LATIN SMALL LETTER I WITH TILDE BELOW
    Interval{ 0x1E2F, 0x1E2F }, // L&       LATIN SMALL LETTER I WITH DIAERESIS AND ACUTE
    Interval{ 0x1E31, 0x1E31 }, // L&       LATIN SMALL LETTER K WITH ACUTE
    Interval{ 0x1E33, 0x1E33 }, // L&       LATIN SMALL LETTER K WITH DOT BELOW
    Interval{ 0x1E35, 0x1E35 }, // L&       LATIN SMALL LETTER K WITH LINE BELOW
    Interval{ 0x1E37, 0x1E37 }, // L&       LATIN SMALL LETTER L WITH DOT BELOW
    Interval{ 0x1E39, 0x1E39 }, // L&       LATIN SMALL LETTER L WITH DOT BELOW AND MACRON
    Interval{ 0x1E3B, 0x1E3B }, // L&       LATIN SMALL LETTER L WITH LINE BELOW
    Interval{ 0x1E3D, 0x1E3D }, // L&       LATIN SMALL LETTER L WITH CIRCUMFLEX BELOW
    Interval{ 0x1E3F, 0x1E3F }, // L&       LATIN SMALL LETTER M WITH ACUTE
    Interval{ 0x1E41, 0x1E41 }, // L&       LATIN SMALL LETTER M WITH DOT ABOVE
    Interval{ 0x1E43, 0x1E43 }, // L&       LATIN SMALL LETTER M WITH DOT BELOW
    Interval{ 0x1E45, 0x1E45 }, // L&       LATIN SMALL LETTER N WITH DOT ABOVE
    Interval{ 0x1E47, 0x1E47 }, // L&       LATIN SMALL LETTER N WITH DOT BELOW
    Interval{ 0x1E49, 0x1E49 }, // L&       LATIN SMALL LETTER N WITH LINE BELOW
    Interval{ 0x1E4B, 0x1E4B }, // L&       LATIN SMALL LETTER N WITH CIRCUMFLEX BELOW
    Interval{ 0x1E4D, 0x1E4D }, // L&       LATIN SMALL LETTER O WITH TILDE AND ACUTE
    Interval{ 0x1E4F, 0x1E4F }, // L&       LATIN SMALL LETTER O WITH TILDE AND DIAERESIS
    Interval{ 0x1E51, 0x1E51 }, // L&       LATIN SMALL LETTER O WITH MACRON AND GRAVE
    Interval{ 0x1E53, 0x1E53 }, // L&       LATIN SMALL LETTER O WITH MACRON AND ACUTE
    Interval{ 0x1E55, 0x1E55 }, // L&       LATIN SMALL LETTER P WITH ACUTE
    Interval{ 0x1E57, 0x1E57 }, // L&       LATIN SMALL LETTER P WITH DOT ABOVE
    Interval{ 0x1E59, 0x1E59 }, // L&       LATIN SMALL LETTER R WITH DOT ABOVE
    Interval{ 0x1E5B, 0x1E5B }, // L&       LATIN SMALL LETTER R WITH DOT BELOW
    Interval{ 0x1E5D, 0x1E5D }, // L&       LATIN SMALL LETTER R WITH DOT BELOW AND MACRON
    Interval{ 0x1E5F, 0x1E5F }, // L&       LATIN SMALL LETTER R WITH LINE BELOW
    Interval{ 0x1E61, 0x1E61 }, // L&       LATIN SMALL LETTER S WITH DOT ABOVE
    Interval{ 0x1E63, 0x1E63 }, // L&       LATIN SMALL LETTER S WITH DOT BELOW
    Interval{ 0x1E65, 0x1E65 }, // L&       LATIN SMALL LETTER S WITH ACUTE AND DOT ABOVE
    Interval{ 0x1E67, 0x1E67 }, // L&       LATIN SMALL LETTER S WITH CARON AND DOT ABOVE
    Interval{ 0x1E69, 0x1E69 }, // L&       LATIN SMALL LETTER S WITH DOT BELOW AND DOT ABOVE
    Interval{ 0x1E6B, 0x1E6B }, // L&       LATIN SMALL LETTER T WITH DOT ABOVE
    Interval{ 0x1E6D, 0x1E6D }, // L&       LATIN SMALL LETTER T WITH DOT BELOW
    Interval{ 0x1E6F, 0x1E6F }, // L&       LATIN SMALL LETTER T WITH LINE BELOW
    Interval{ 0x1E71, 0x1E71 }, // L&       LATIN SMALL LETTER T WITH CIRCUMFLEX BELOW
    Interval{ 0x1E73, 0x1E73 }, // L&       LATIN SMALL LETTER U WITH DIAERESIS BELOW
    Interval{ 0x1E75, 0x1E75 }, // L&       LATIN SMALL LETTER U WITH TILDE BELOW
    Interval{ 0x1E77, 0x1E77 }, // L&       LATIN SMALL LETTER U WITH CIRCUMFLEX BELOW
    Interval{ 0x1E79, 0x1E79 }, // L&       LATIN SMALL LETTER U WITH TILDE AND ACUTE
    Interval{ 0x1E7B, 0x1E7B }, // L&       LATIN SMALL LETTER U WITH MACRON AND DIAERESIS
    Interval{ 0x1E7D, 0x1E7D }, // L&       LATIN SMALL LETTER V WITH TILDE
    Interval{ 0x1E7F, 0x1E7F }, // L&       LATIN SMALL LETTER V WITH DOT BELOW
    Interval{ 0x1E81, 0x1E81 }, // L&       LATIN SMALL LETTER W WITH GRAVE
    Interval{ 0x1E83, 0x1E83 }, // L&       LATIN SMALL LETTER W WITH ACUTE
    Interval{ 0x1E85, 0x1E85 }, // L&       LATIN SMALL LETTER W WITH DIAERESIS
    Interval{ 0x1E87, 0x1E87 }, // L&       LATIN SMALL LETTER W WITH DOT ABOVE
    Interval{ 0x1E89, 0x1E89 }, // L&       LATIN SMALL LETTER W WITH DOT BELOW
    Interval{ 0x1E8B, 0x1E8B }, // L&       LATIN SMALL LETTER X WITH DOT ABOVE
    Interval{ 0x1E8D, 0x1E8D }, // L&       LATIN SMALL LETTER X WITH DIAERESIS
    Interval{ 0x1E8F, 0x1E8F }, // L&       LATIN SMALL LETTER Y WITH DOT ABOVE
    Interval{ 0x1E91, 0x1E91 }, // L&       LATIN SMALL LETTER Z WITH CIRCUMFLEX
    Interval{ 0x1E93, 0x1E93 }, // L&       LATIN SMALL LETTER Z WITH DOT BELOW
    Interval{ 0x1E95, 0x1E9B }, // L&   [7] LATIN SMALL LETTER Z WITH LINE BELOW..LATIN SMALL LETTER LONG S WITH DOT ABOVE
    Interval{ 0x1EA1, 0x1EA1 }, // L&       LATIN SMALL LETTER A WITH DOT BELOW
    Interval{ 0x1EA3, 0x1EA3 }, // L&       LATIN SMALL LETTER A WITH HOOK ABOVE
    Interval{ 0x1EA5, 0x1EA5 }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EA7, 0x1EA7 }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EA9, 0x1EA9 }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EAB, 0x1EAB }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EAD, 0x1EAD }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EAF, 0x1EAF }, // L&       LATIN SMALL LETTER A WITH BREVE AND ACUTE
    Interval{ 0x1EB1, 0x1EB1 }, // L&       LATIN SMALL LETTER A WITH BREVE AND GRAVE
    Interval{ 0x1EB3, 0x1EB3 }, // L&       LATIN SMALL LETTER A WITH BREVE AND HOOK ABOVE
    Interval{ 0x1EB5, 0x1EB5 }, // L&       LATIN SMALL LETTER A WITH BREVE AND TILDE
    Interval{ 0x1EB7, 0x1EB7 }, // L&       LATIN SMALL LETTER A WITH BREVE AND DOT BELOW
    Interval{ 0x1EB9, 0x1EB9 }, // L&       LATIN SMALL LETTER E WITH DOT BELOW
    Interval{ 0x1EBB, 0x1EBB }, // L&       LATIN SMALL LETTER E WITH HOOK ABOVE
    Interval{ 0x1EBD, 0x1EBD }, // L&       LATIN SMALL LETTER E WITH TILDE
    Interval{ 0x1EBF, 0x1EBF }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EC1, 0x1EC1 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EC3, 0x1EC3 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EC5, 0x1EC5 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EC7, 0x1EC7 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EC9, 0x1EC9 }, // L&       LATIN SMALL LETTER I WITH HOOK ABOVE
    Interval{ 0x1ECB, 0x1ECB }, // L&       LATIN SMALL LETTER I WITH DOT BELOW
    Interval{ 0x1ECD, 0x1ECD }, // L&       LATIN SMALL LETTER O WITH DOT BELOW
    Interval{ 0x1ECF, 0x1ECF }, // L&       LATIN SMALL LETTER O WITH HOOK ABOVE
    Interval{ 0x1ED1, 0x1ED1 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1ED3, 0x1ED3 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1ED5, 0x1ED5 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1ED7, 0x1ED7 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1ED9, 0x1ED9 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EDB, 0x1EDB }, // L&       LATIN SMALL LETTER O WITH HORN AND ACUTE
    Interval{ 0x1EDD, 0x1EDD }, // L&       LATIN SMALL LETTER O WITH HORN AND GRAVE
    Interval{ 0x1EDF, 0x1EDF }, // L&       LATIN SMALL LETTER O WITH HORN AND HOOK ABOVE
    Interval{ 0x1EE1, 0x1EE1 }, // L&       LATIN SMALL LETTER O WITH HORN AND TILDE
    Interval{ 0x1EE3, 0x1EE3 }, // L&       LATIN SMALL LETTER O WITH HORN AND DOT BELOW
    Interval{ 0x1EE5, 0x1EE5 }, // L&       LATIN SMALL LETTER U WITH DOT BELOW
    Interval{ 0x1EE7, 0x1EE7 }, // L&       LATIN SMALL LETTER U WITH HOOK ABOVE
    Interval{ 0x1EE9, 0x1EE9 }, // L&       LATIN SMALL LETTER U WITH HORN AND ACUTE
    Interval{ 0x1EEB, 0x1EEB }, // L&       LATIN SMALL LETTER U WITH HORN AND GRAVE
    Interval{ 0x1EED, 0x1EED }, // L&       LATIN SMALL LETTER U WITH HORN AND HOOK ABOVE
    Interval{ 0x1EEF, 0x1EEF }, // L&       LATIN SMALL LETTER U WITH HORN AND TILDE
    Interval{ 0x1EF1, 0x1EF1 }, // L&       LATIN SMALL LETTER U WITH HORN AND DOT BELOW
    Interval{ 0x1EF3, 0x1EF3 }, // L&       LATIN SMALL LETTER Y WITH GRAVE
    Interval{ 0x1EF5, 0x1EF5 }, // L&       LATIN SMALL LETTER Y WITH DOT BELOW
    Interval{ 0x1EF7, 0x1EF7 }, // L&       LATIN SMALL LETTER Y WITH HOOK ABOVE
    Interval{ 0x1EF9, 0x1EF9 }, // L&       LATIN SMALL LETTER Y WITH TILDE
    Interval{ 0x1EFB, 0x1EFB }, // L&       LATIN SMALL LETTER MIDDLE-WELSH LL
    Interval{ 0x1EFD, 0x1EFD }, // L&       LATIN SMALL LETTER MIDDLE-WELSH V
    Interval{ 0x1EFF, 0x1F07 }, // L&   [9] LATIN SMALL LETTER Y WITH LOOP..GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F10, 0x1F15 }, // L&   [6] GREEK SMALL LETTER EPSILON WITH PSILI..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F20, 0x1F27 }, // L&   [8] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F30, 0x1F37 }, // L&   [8] GREEK SMALL LETTER IOTA WITH PSILI..GREEK SMALL LETTER IOTA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F40, 0x1F45 }, // L&   [6] GREEK SMALL LETTER OMICRON WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F50, 0x1F57 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F60, 0x1F67 }, // L&   [8] GREEK SMALL LETTER OMEGA WITH PSILI..GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F70, 0x1F7D }, // L&  [14] GREEK SMALL LETTER ALPHA WITH VARIA..GREEK SMALL LETTER OMEGA WITH OXIA
    Interval{ 0x1F80, 0x1F87 }, // L&   [8] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1F90, 0x1F97 }, // L&   [8] GREEK SMALL LETTER ETA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FA0, 0x1FA7 }, // L&   [8] GREEK SMALL LETTER OMEGA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FB0, 0x1FB4 }, // L&   [5] GREEK SMALL LETTER ALPHA WITH VRACHY..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB6, 0x1FB7 }, // L&   [2] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK SMALL LETTER ALPHA WITH PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FBE, 0x1FBE }, // L&       GREEK PROSGEGRAMMENI
    Interval{ 0x1FC2, 0x1FC4 }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC6, 0x1FC7 }, // L&   [2] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK SMALL LETTER ETA WITH PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FD0, 0x1FD3 }, // L&   [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Interval{ 0x1FD6, 0x1FD7 }, // L&   [2] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND PERISPOMENI
    Interval{ 0x1FE0, 0x1FE7 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND PERISPOMENI
    Interval{ 0x1FF2, 0x1FF4 }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF6, 0x1FF7 }, // L&   [2] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK SMALL LETTER OMEGA WITH PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x214E, 0x214E }, // L&       TURNED SMALL F
    Interval{ 0x2170, 0x217F }, // Nl  [16] SMALL ROMAN NUMERAL ONE..SMALL ROMAN NUMERAL ONE THOUSAND
    Interval{ 0x2184, 0x2184 }, // L&       LATIN SMALL LETTER REVERSED C
    Interval{ 0x24D0, 0x24E9 }, // So  [26] CIRCLED LATIN SMALL LETTER A..CIRCLED LATIN SMALL LETTER Z
    Interval{ 0x2C30, 0x2C5F }, // L&  [48] GLAGOLITIC SMALL LETTER AZU..GLAGOLITIC SMALL LETTER CAUDATE CHRIVI
    Interval{ 0x2C61, 0x2C61 }, // L&       LATIN SMALL LETTER L WITH DOUBLE BAR
    Interval{ 0x2C65, 0x2C66 }, // L&   [2] LATIN SMALL LETTER A WITH STROKE..LATIN SMALL LETTER T WITH DIAGONAL STROKE
    Interval{ 0x2C68, 0x2C68 }, // L&       LATIN SMALL LETTER H WITH DESCENDER
    Interval{ 0x2C6A, 0x2C6A }, // L&       LATIN SMALL LETTER K WITH DESCENDER
    Interval{ 0x2C6C, 0x2C6C }, // L&       LATIN SMALL LETTER Z WITH DESCENDER
    Interval{ 0x2C73, 0x2C73 }, // L&       LATIN SMALL LETTER W WITH HOOK
    Interval{ 0x2C76, 0x2C76 }, // L&       LATIN SMALL LETTER HALF H
    Interval{ 0x2C81, 0x2C81 }, // L&       COPTIC SMALL LETTER ALFA
    Interval{ 0x2C83, 0x2C83 }, // L&       COPTIC SMALL LETTER VIDA
    Interval{ 0x2C85, 0x2C85 }, // L&       COPTIC SMALL LETTER GAMMA
    Interval{ 0x2C87, 0x2C87 }, // L&       COPTIC SMALL LETTER DALDA
    Interval{ 0x2C89, 0x2C89 }, // L&       COPTIC SMALL LETTER EIE
    Interval{ 0x2C8B, 0x2C8B }, // L&       COPTIC SMALL LETTER SOU
    Interval{ 0x2C8D, 0x2C8D }, // L&       COPTIC SMALL LETTER ZATA
    Interval{ 0x2C8F, 0x2C8F }, // L&       COPTIC SMALL LETTER HATE
    Interval{ 0x2C91, 0x2C91 }, // L&       COPTIC SMALL LETTER THETHE
    Interval{ 0x2C93, 0x2C93 }, // L&       COPTIC SMALL LETTER IAUDA
    Interval{ 0x2C95, 0x2C95 }, // L&       COPTIC SMALL LETTER KAPA
    Interval{ 0x2C97, 0x2C97 }, // L&       COPTIC SMALL LETTER LAULA
    Interval{ 0x2C99, 0x2C99 }, // L&       COPTIC SMALL LETTER MI
    Interval{ 0x2C9B, 0x2C9B }, // L&       COPTIC SMALL LETTER NI
    Interval{ 0x2C9D, 0x2C9D }, // L&       COPTIC SMALL LETTER KSI
    Interval{ 0x2C9F, 0x2C9F }, // L&       COPTIC SMALL LETTER O
    Interval{ 0x2CA1, 0x2CA1 }, // L&       COPTIC SMALL LETTER PI
    Interval{ 0x2CA3, 0x2CA3 }, // L&       COPTIC SMALL LETTER RO
    Interval{ 0x2CA5, 0x2CA5 }, // L&       COPTIC SMALL LETTER SIMA
    Interval{ 0x2CA7, 0x2CA7 }, // L&       COPTIC SMALL LETTER TAU
    Interval{ 0x2CA9, 0x2CA9 }, // L&       COPTIC SMALL LETTER UA
    Interval{ 0x2CAB, 0x2CAB }, // L&       COPTIC SMALL LETTER FI
    Interval{ 0x2CAD, 0x2CAD }, // L&       COPTIC SMALL LETTER KHI
    Interval{ 0x2CAF, 0x2CAF }, // L&       COPTIC SMALL LETTER PSI
    Interval{ 0x2CB1, 0x2CB1 }, // L&       COPTIC SMALL LETTER OOU
    Interval{ 0x2CB3, 0x2CB3 }, // L&       COPTIC SMALL LETTER DIALECT-P ALEF
    Interval{ 0x2CB5, 0x2CB5 }, // L&       COPTIC SMALL LETTER OLD COPTIC AIN
    Interval{ 0x2CB7, 0x2CB7 }, // L&       COPTIC SMALL LETTER CRYPTOGRAMMIC EIE
    Interval{ 0x2CB9, 0x2CB9 }, // L&       COPTIC SMALL LETTER DIALECT-P KAPA
    Interval{ 0x2CBB, 0x2CBB }, // L&       COPTIC SMALL LETTER DIALECT-P NI
    Interval{ 0x2CBD, 0x2CBD }, // L&       COPTIC SMALL LETTER CRYPTOGRAMMIC NI
    Interval{ 0x2CBF, 0x2CBF }, // L&       COPTIC SMALL LETTER OLD COPTIC OOU
    Interval{ 0x2CC1, 0x2CC1 }, // L&       COPTIC SMALL LETTER SAMPI
    Interval{ 0x2CC3, 0x2CC3 }, // L&       COPTIC SMALL LETTER CROSSED SHEI
    Interval{ 0x2CC5, 0x2CC5 }, // L&       COPTIC SMALL LETTER OLD COPTIC SHEI
    Interval{ 0x2CC7, 0x2CC7 }, // L&       COPTIC SMALL LETTER OLD COPTIC ESH
    Interval{ 0x2CC9, 0x2CC9 }, // L&       COPTIC SMALL LETTER AKHMIMIC KHEI
    Interval{ 0x2CCB, 0x2CCB }, // L&       COPTIC SMALL LETTER DIALECT-P HORI
    Interval{ 0x2CCD, 0x2CCD }, // L&       COPTIC SMALL LETTER OLD COPTIC HORI
    Interval{ 0x2CCF, 0x2CCF }, // L&       COPTIC SMALL LETTER OLD COPTIC HA
    Interval{ 0x2CD1, 0x2CD1 }, // L&       COPTIC SMALL LETTER L-SHAPED HA
    Interval{ 0x2CD3, 0x2CD3 }, // L&       COPTIC SMALL LETTER OLD COPTIC HEI
    Interval{ 0x2CD5, 0x2CD5 }, // L&       COPTIC SMALL LETTER OLD COPTIC HAT
    Interval{ 0x2CD7, 0x2CD7 }, // L&       COPTIC SMALL LETTER OLD COPTIC GANGIA
    Interval{ 0x2CD9, 0x2CD9 }, // L&       COPTIC SMALL LETTER OLD COPTIC DJA
    Interval{ 0x2CDB, 0x2CDB }, // L&       COPTIC SMALL LETTER OLD COPTIC SHIMA
    Interval{ 0x2CDD, 0x2CDD }, // L&       COPTIC SMALL LETTER OLD NUBIAN SHIMA
    Interval{ 0x2CDF, 0x2CDF }, // L&       COPTIC SMALL LETTER OLD NUBIAN NGI
    Interval{ 0x2CE1, 0x2CE1 }, // L&       COPTIC SMALL LETTER OLD NUBIAN NYI
    Interval{ 0x2CE3, 0x2CE3 }, // L&       COPTIC SMALL LETTER OLD NUBIAN WAU
    Interval{ 0x2CEC, 0x2CEC }, // L&       COPTIC SMALL LETTER CRYPTOGRAMMIC SHEI
    Interval{ 0x2CEE, 0x2CEE }, // L&       COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF3, 0x2CF3 }, // L&       COPTIC SMALL LETTER BOHAIRIC KHEI
    Interval{ 0x2D00, 0x2D25 }, // L&  [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Interval{ 0x2D27, 0x2D27 }, // L&       GEORGIAN SMALL LETTER YN
    Interval{ 0x2D2D, 0x2D2D }, // L&       GEORGIAN SMALL LETTER AEN
    Interval{ 0xA641, 0xA641 }, // L&       CYRILLIC SMALL LETTER ZEMLYA
    Interval{ 0xA643, 0xA643 }, // L&       CYRILLIC SMALL LETTER DZELO
    Interval{ 0xA645, 0xA645 }, // L&       CYRILLIC SMALL LETTER REVERSED DZE
    Interval{ 0xA647, 0xA647 }, // L&       CYRILLIC SMALL LETTER IOTA
    Interval{ 0xA649, 0xA649 }, // L&       CYRILLIC SMALL LETTER DJERV
    Interval{ 0xA64B, 0xA64B }, // L&       CYRILLIC SMALL LETTER MONOGRAPH UK
    Interval{ 0xA64D, 0xA64D }, // L&       CYRILLIC SMALL LETTER BROAD OMEGA
    Interval{ 0xA64F, 0xA64F }, // L&       CYRILLIC SMALL LETTER NEUTRAL YER
    Interval{ 0xA651, 0xA651 }, // L&       CYRILLIC SMALL LETTER YERU WITH BACK YER
    Interval{ 0xA653, 0xA653 }, // L&       CYRILLIC SMALL LETTER IOTIFIED YAT
    Interval{ 0xA655, 0xA655 }, // L&       CYRILLIC SMALL LETTER REVERSED YU
    Interval{ 0xA657, 0xA657 }, // L&       CYRILLIC SMALL LETTER IOTIFIED A
    Interval{ 0xA659, 0xA659 }, // L&       CYRILLIC SMALL LETTER CLOSED LITTLE YUS
    Interval{ 0xA65B, 0xA65B }, // L&       CYRILLIC SMALL LETTER BLENDED YUS
    Interval{ 0xA65D, 0xA65D }, // L&       CYRILLIC SMALL LETTER IOTIFIED CLOSED LITTLE YUS
    Interval{ 0xA65F, 0xA65F }, // L&       CYRILLIC SMALL LETTER YN
    Interval{ 0xA661, 0xA661 }, // L&       CYRILLIC SMALL LETTER REVERSED TSE
    Interval{ 0xA663, 0xA663 }, // L&       CYRILLIC SMALL LETTER SOFT DE
    Interval{ 0xA665, 0xA665 }, // L&       CYRILLIC SMALL LETTER SOFT EL
    Interval{ 0xA667, 0xA667 }, // L&       CYRILLIC SMALL LETTER SOFT EM
    Interval{ 0xA669, 0xA669 }, // L&       CYRILLIC SMALL LETTER MONOCULAR O
    Interval{ 0xA66B, 0xA66B }, // L&       CYRILLIC SMALL LETTER BINOCULAR O
    Interval{ 0xA66D, 0xA66D }, // L&       CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA681, 0xA681 }, // L&       CYRILLIC SMALL LETTER DWE
    Interval{ 0xA683, 0xA683 }, // L&       CYRILLIC SMALL LETTER DZWE
    Interval{ 0xA685, 0xA685 }, // L&       CYRILLIC SMALL LETTER ZHWE
    Interval{ 0xA687, 0xA687 }, // L&       CYRILLIC SMALL LETTER CCHE
    Interval{ 0xA689, 0xA689 }, // L&       CYRILLIC SMALL LETTER DZZE
    Interval{ 0xA68B, 0xA68B }, // L&       CYRILLIC SMALL LETTER TE WITH MIDDLE HOOK
    Interval{ 0xA68D, 0xA68D }, // L&       CYRILLIC SMALL LETTER TWE
    Interval{ 0xA68F, 0xA68F }, // L&       CYRILLIC SMALL LETTER TSWE
    Interval{ 0xA691, 0xA691 }, // L&       CYRILLIC SMALL LETTER TSSE
    Interval{ 0xA693, 0xA693 }, // L&       CYRILLIC SMALL LETTER TCHE
    Interval{ 0xA695, 0xA695 }, // L&       CYRILLIC SMALL LETTER HWE
    Interval{ 0xA697, 0xA697 }, // L&       CYRILLIC SMALL LETTER SHWE
    Interval{ 0xA699, 0xA699 }, // L&       CYRILLIC SMALL LETTER DOUBLE O
    Interval{ 0xA69B, 0xA69B }, // L&       CYRILLIC SMALL LETTER CROSSED O
    Interval{ 0xA723, 0xA723 }, // L&       LATIN SMALL LETTER EGYPTOLOGICAL ALEF
    Interval{ 0xA725, 0xA725 }, // L&       LATIN SMALL LETTER EGYPTOLOGICAL AIN
    Interval{ 0xA727, 0xA727 }, // L&       LATIN SMALL LETTER HENG
    Interval{ 0xA729, 0xA729 }, // L&       LATIN SMALL LETTER TZ
    Interval{ 0xA72B, 0xA72B }, // L&       LATIN SMALL LETTER TRESILLO
    Interval{ 0xA72D, 0xA72D }, // L&       LATIN SMALL LETTER CUATRILLO
    Interval{ 0xA72F, 0xA72F }, // L&       LATIN SMALL LETTER CUATRILLO WITH COMMA
    Interval{ 0xA733, 0xA733 }, // L&       LATIN SMALL LETTER AA
    Interval{ 0xA735, 0xA735 }, // L&       LATIN SMALL LETTER AO
    Interval{ 0xA737, 0xA737 }, // L&       LATIN SMALL LETTER AU
    Interval{ 0xA739, 0xA739 }, // L&       LATIN SMALL LETTER AV
    Interval{ 0xA73B, 0xA73B }, // L&       LATIN SMALL LETTER AV WITH HORIZONTAL BAR
    Interval{ 0xA73D, 0xA73D }, // L&       LATIN SMALL LETTER AY
    Interval{ 0xA73F, 0xA73F }, // L&       LATIN SMALL LETTER REVERSED C WITH DOT
    Interval{ 0xA741, 0xA741 }, // L&       LATIN SMALL LETTER K WITH STROKE
    Interval{ 0xA743, 0xA743 }, // L&       LATIN SMALL LETTER K WITH DIAGONAL STROKE
    Interval{ 0xA745, 0xA745 }, // L&       LATIN SMALL LETTER K WITH STROKE AND DIAGONAL STROKE
    Interval{ 0xA747, 0xA747 }, // L&       LATIN SMALL LETTER BROKEN L
    Interval{ 0xA749, 0xA749 }, // L&       LATIN SMALL LETTER L WITH HIGH STROKE
    Interval{ 0xA74B, 0xA74B }, // L&       LATIN SMALL LETTER O WITH LONG STROKE OVERLAY
    Interval{ 0xA74D, 0xA74D }, // L&       LATIN SMALL LETTER O WITH LOOP
    Interval{ 0xA74F, 0xA74F }, // L&       LATIN SMALL LETTER OO
    Interval{ 0xA751, 0xA751 }, // L&       LATIN SMALL LETTER P WITH STROKE THROUGH DESCENDER
    Interval{ 0xA753, 0xA753 }, // L&       LATIN SMALL LETTER P WITH FLOURISH
    Interval{ 0xA755, 0xA755 }, // L&       LATIN SMALL LETTER P WITH SQUIRREL TAIL
    Interval{ 0xA757, 0xA757 }, // L&       LATIN SMALL LETTER Q WITH STROKE THROUGH DESCENDER
    Interval{ 0xA759, 0xA759 }, // L&       LATIN SMALL LETTER Q WITH DIAGONAL STROKE
    Interval{ 0xA75B, 0xA75B }, // L&       LATIN SMALL LETTER R ROTUNDA
    Interval{ 0xA75D, 0xA75D }, // L&       LATIN SMALL LETTER RUM ROTUNDA
    Interval{ 0xA75F, 0xA75F }, // L&       LATIN SMALL LETTER V WITH DIAGONAL STROKE
    Interval{ 0xA761, 0xA761 }, // L&       LATIN SMALL LETTER VY
    Interval{ 0xA763, 0xA763 }, // L&       LATIN SMALL LETTER VISIGOTHIC Z
    Interval{ 0xA765, 0xA765 }, // L&       LATIN SMALL LETTER THORN WITH STROKE
    Interval{ 0xA767, 0xA767 }, // L&       LATIN SMALL LETTER THORN WITH STROKE THROUGH DESCENDER
    Interval{ 0xA769, 0xA769 }, // L&       LATIN SMALL LETTER VEND
    Interval{ 0xA76B, 0xA76B }, // L&       LATIN SMALL LETTER ET
    Interval{ 0xA76D, 0xA76D }, // L&       LATIN SMALL LETTER IS
    Interval{ 0xA76F, 0xA76F }, // L&       LATIN SMALL LETTER CON
    Interval{ 0xA77A, 0xA77A }, // L&       LATIN SMALL LETTER INSULAR D
    Interval{ 0xA77C, 0xA77C }, // L&       LATIN SMALL LETTER INSULAR F
    Interval{ 0xA77F, 0xA77F }, // L&       LATIN SMALL LETTER TURNED INSULAR G
    Interval{ 0xA781, 0xA781 }, // L&       LATIN SMALL LETTER TURNED L
    Interval{ 0xA783, 0xA783 }, // L&       LATIN SMALL LETTER INSULAR R
    Interval{ 0xA785, 0xA785 }, // L&       LATIN SMALL LETTER INSULAR S
    Interval{ 0xA787, 0xA787 }, // L&       LATIN SMALL LETTER INSULAR T
    Interval{ 0xA78C, 0xA78C }, // L&       LATIN SMALL LETTER SALTILLO
    Interval{ 0xA791, 0xA791 }, // L&       LATIN SMALL LETTER N WITH DESCENDER
    Interval{ 0xA793, 0xA794 }, // L&   [2] LATIN SMALL LETTER C WITH BAR..LATIN SMALL LETTER C WITH PALATAL HOOK
    Interval{ 0xA797, 0xA797 }, // L&       LATIN SMALL LETTER B WITH FLOURISH
    Interval{ 0xA799, 0xA799 }, // L&       LATIN SMALL LETTER F WITH STROKE
    Interval{ 0xA79B, 0xA79B }, // L&       LATIN SMALL LETTER VOLAPUK AE
    Interval{ 0xA79D, 0xA79D }, // L&       LATIN SMALL LETTER VOLAPUK OE
    Interval{ 0xA79F, 0xA79F }, // L&       LATIN SMALL LETTER VOLAPUK UE
    Interval{ 0xA7A1, 0xA7A1 }, // L&       LATIN SMALL LETTER G WITH OBLIQUE STROKE
    Interval{ 0xA7A3, 0xA7A3 }, // L&       LATIN SMALL LETTER K WITH OBLIQUE STROKE
    Interval{ 0xA7A5, 0xA7A5 }, // L&       LATIN SMALL LETTER N WITH OBLIQUE STROKE
    Interval{ 0xA7A7, 0xA7A7 }, // L&       LATIN SMALL LETTER R WITH OBLIQUE STROKE
    Interval{ 0xA7A9, 0xA7A9 }, // L&       LATIN SMALL LETTER S WITH OBLIQUE STROKE
    Interval{ 0xA7B5, 0xA7B5 }, // L&       LATIN SMALL LETTER BETA
    Interval{ 0xA7B7, 0xA7B7 }, // L&       LATIN SMALL LETTER OMEGA
    Interval{ 0xA7B9, 0xA7B9 }, // L&       LATIN SMALL LETTER U WITH STROKE
    Interval{ 0xA7BB, 0xA7BB }, // L&       LATIN SMALL LETTER GLOTTAL A
    Interval{ 0xA7BD, 0xA7BD }, // L&       LATIN SMALL LETTER GLOTTAL I
    Interval{ 0xA7BF, 0xA7BF }, // L&       LATIN SMALL LETTER GLOTTAL U
    Interval{ 0xA7C1, 0xA7C1 }, // L&       LATIN SMALL LETTER OLD POLISH O
    Interval{ 0xA7C3, 0xA7C3 }, // L&       LATIN SMALL LETTER ANGLICANA W
    Interval{ 0xA7C8, 0xA7C8 }, // L&       LATIN SMALL LETTER D WITH SHORT STROKE OVERLAY
    Interval{ 0xA7CA, 0xA7CA }, // L&       LATIN SMALL LETTER S WITH SHORT STROKE OVERLAY
    Interval{ 0xA7CD, 0xA7CD }, // L&       LATIN SMALL LETTER S WITH DIAGONAL STROKE
    Interval{ 0xA7CF, 0xA7CF }, // L&       LATIN SMALL LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0xA7D1, 0xA7D1 }, // L&       LATIN SMALL LETTER CLOSED INSULAR G
    Interval{ 0xA7D3, 0xA7D3 }, // L&       LATIN SMALL LETTER DOUBLE THORN
    Interval{ 0xA7D5, 0xA7D5 }, // L&       LATIN SMALL LETTER DOUBLE WYNN
    Interval{ 0xA7D7, 0xA7D7 }, // L&       LATIN SMALL LETTER MIDDLE SCOTS S
    Interval{ 0xA7D9, 0xA7D9 }, // L&       LATIN SMALL LETTER SIGMOID S
    Interval{ 0xA7DB, 0xA7DB }, // L&       LATIN SMALL LETTER LAMBDA
    Interval{ 0xA7F6, 0xA7F6 }, // L&       LATIN SMALL LETTER REVERSED HALF H
    Interval{ 0xAB53, 0xAB53 }, // L&       LATIN SMALL LETTER CHI
    Interval{ 0xAB70, 0xABBF }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xFB00, 0xFB06 }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFF41, 0xFF5A }, // L&  [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Interval{ 0x10428, 0x1044F }, // L&  [40] DESERET SMALL LETTER LONG I..DESERET SMALL LETTER EW
    Interval{ 0x104D8, 0x104FB }, // L&  [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Interval{ 0x10597, 0x105A1 }, // L&  [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Interval{ 0x105A3, 0x105B1 }, // L&  [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Interval{ 0x105B3, 0x105B9 }, // L&   [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Interval{ 0x105BB, 0x105BC }, // L&   [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Interval{ 0x10CC0, 0x10CF2 }, // L&  [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Interval{ 0x10D70, 0x10D85 }, // L&  [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Interval{ 0x118C0, 0x118DF }, // L&  [32] WARANG CITI SMALL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Interval{ 0x16E60, 0x16E7F }, // L&  [32] MEDEFAIDRIN SMALL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Interval{ 0x16EBB, 0x16ED3 }, // L&  [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Interval{ 0x1E922, 0x1E943 }, // L&  [34] ADLAM SMALL LETTER ALIF..ADLAM SMALL LETTER SHA
}; // }}}
auto static const Changes_When_Uppercased = std::array<Interval, 634>{ // {{{
    Interval{ 0x0061, 0x007A }, // L&  [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Interval{ 0x00B5, 0x00B5 }, // L&       MICRO SIGN
    Interval{ 0x00DF, 0x00F6 }, // L&  [24] LATIN SMALL LETTER SHARP S..LATIN SMALL LETTER O WITH DIAERESIS
    Interval{ 0x00F8, 0x00FF }, // L&   [8] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER Y WITH DIAERESIS
    Interval{ 0x0101, 0x0101 }, // L&       LATIN SMALL LETTER A WITH MACRON
    Interval{ 0x0103, 0x0103 }, // L&       LATIN SMALL LETTER A WITH BREVE
    Interval{ 0x0105, 0x0105 }, // L&       LATIN SMALL LETTER A WITH OGONEK
    Interval{ 0x0107, 0x0107 }, // L&       LATIN SMALL LETTER C WITH ACUTE
    Interval{ 0x0109, 0x0109 }, // L&       LATIN SMALL LETTER C WITH CIRCUMFLEX
    Interval{ 0x010B, 0x010B }, // L&       LATIN SMALL LETTER C WITH DOT ABOVE
    Interval{ 0x010D, 0x010D }, // L&       LATIN SMALL LETTER C WITH CARON
    Interval{ 0x010F, 0x010F }, // L&       LATIN SMALL LETTER D WITH CARON
    Interval{ 0x0111, 0x0111 }, // L&       LATIN SMALL LETTER D WITH STROKE
    Interval{ 0x0113, 0x0113 }, // L&       LATIN SMALL LETTER E WITH MACRON
    Interval{ 0x0115, 0x0115 }, // L&       LATIN SMALL LETTER E WITH BREVE
    Interval{ 0x0117, 0x0117 }, // L&       LATIN SMALL LETTER E WITH DOT ABOVE
    Interval{ 0x0119, 0x0119 }, // L&       LATIN SMALL LETTER E WITH OGONEK
    Interval{ 0x011B, 0x011B }, // L&       LATIN SMALL LETTER E WITH CARON
    Interval{ 0x011D, 0x011D }, // L&       LATIN SMALL LETTER G WITH CIRCUMFLEX
    Interval{ 0x011F, 0x011F }, // L&       LATIN SMALL LETTER G WITH BREVE
    Interval{ 0x0121, 0x0121 }, // L&       LATIN SMALL LETTER G WITH DOT ABOVE
    Interval{ 0x0123, 0x0123 }, // L&       LATIN SMALL LETTER G WITH CEDILLA
    Interval{ 0x0125, 0x0125 }, // L&       LATIN SMALL LETTER H WITH CIRCUMFLEX
    Interval{ 0x0127, 0x0127 }, // L&       LATIN SMALL LETTER H WITH STROKE
    Interval{ 0x0129, 0x0129 }, // L&       LATIN SMALL LETTER I WITH TILDE
    Interval{ 0x012B, 0x012B }, // L&       LATIN SMALL LETTER I WITH MACRON
    Interval{ 0x012D, 0x012D }, // L&       LATIN SMALL LETTER I WITH BREVE
    Interval{ 0x012F, 0x012F }, // L&       LATIN SMALL LETTER I WITH OGONEK
    Interval{ 0x0131, 0x0131 }, // L&       LATIN SMALL LETTER DOTLESS I
    Interval{ 0x0133, 0x0133 }, // L&       LATIN SMALL LIGATURE IJ
    Interval{ 0x0135, 0x0135 }, // L&       LATIN SMALL LETTER J WITH CIRCUMFLEX
    Interval{ 0x0137, 0x0137 }, // L&       LATIN SMALL LETTER K WITH CEDILLA
    Interval{ 0x013A, 0x013A }, // L&       LATIN SMALL LETTER L WITH ACUTE
    Interval{ 0x013C, 0x013C }, // L&       LATIN SMALL LETTER L WITH CEDILLA
    Interval{ 0x013E, 0x013E }, // L&       LATIN SMALL LETTER L WITH CARON
    Interval{ 0x0140, 0x0140 }, // L&       LATIN SMALL LETTER L WITH MIDDLE DOT
    Interval{ 0x0142, 0x0142 }, // L&       LATIN SMALL LETTER L WITH STROKE
    Interval{ 0x0144, 0x0144 }, // L&       LATIN SMALL LETTER N WITH ACUTE
    Interval{ 0x0146, 0x0146 }, // L&       LATIN SMALL LETTER N WITH CEDILLA
    Interval{ 0x0148, 0x0149 }, // L&   [2] LATIN SMALL LETTER N WITH CARON..LATIN SMALL LETTER N PRECEDED BY APOSTROPHE
    Interval{ 0x014B, 0x014B }, // L&       LATIN SMALL LETTER ENG
    Interval{ 0x014D, 0x014D }, // L&       LATIN SMALL LETTER O WITH MACRON
    Interval{ 0x014F, 0x014F }, // L&       LATIN SMALL LETTER O WITH BREVE
    Interval{ 0x0151, 0x0151 }, // L&       LATIN SMALL LETTER O WITH DOUBLE ACUTE
    Interval{ 0x0153, 0x0153 }, // L&       LATIN SMALL LIGATURE OE
    Interval{ 0x0155, 0x0155 }, // L&       LATIN SMALL LETTER R WITH ACUTE
    Interval{ 0x0157, 0x0157 }, // L&       LATIN SMALL LETTER R WITH CEDILLA
    Interval{ 0x0159, 0x0159 }, // L&       LATIN SMALL LETTER R WITH CARON
    Interval{ 0x015B, 0x015B }, // L&       LATIN SMALL LETTER S WITH ACUTE
    Interval{ 0x015D, 0x015D }, // L&       LATIN SMALL LETTER S WITH CIRCUMFLEX
    Interval{ 0x015F, 0x015F }, // L&       LATIN SMALL LETTER S WITH CEDILLA
    Interval{ 0x0161, 0x0161 }, // L&       LATIN SMALL LETTER S WITH CARON
    Interval{ 0x0163, 0x0163 }, // L&       LATIN SMALL LETTER T WITH CEDILLA
    Interval{ 0x0165, 0x0165 }, // L&       LATIN SMALL LETTER T WITH CARON
    Interval{ 0x0167, 0x0167 }, // L&       LATIN SMALL LETTER T WITH STROKE
    Interval{ 0x0169, 0x0169 }, // L&       LATIN SMALL LETTER U WITH TILDE
    Interval{ 0x016B, 0x016B }, // L&       LATIN SMALL LETTER U WITH MACRON
    Interval{ 0x016D, 0x016D }, // L&       LATIN SMALL LETTER U WITH BREVE
    Interval{ 0x016F, 0x016F }, // L&       LATIN SMALL LETTER U WITH RING ABOVE
    Interval{ 0x0171, 0x0171 }, // L&       LATIN SMALL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x0173, 0x0173 }, // L&       LATIN SMALL LETTER U WITH OGONEK
    Interval{ 0x0175, 0x0175 }, // L&       LATIN SMALL LETTER W WITH CIRCUMFLEX
    Interval{ 0x0177, 0x0177 }, // L&       LATIN SMALL LETTER Y WITH CIRCUMFLEX
    Interval{ 0x017A, 0x017A }, // L&       LATIN SMALL LETTER Z WITH ACUTE
    Interval{ 0x017C, 0x017C }, // L&       LATIN SMALL LETTER Z WITH DOT ABOVE
    Interval{ 0x017E, 0x0180 }, // L&   [3] LATIN SMALL LETTER Z WITH CARON..LATIN SMALL LETTER B WITH STROKE
    Interval{ 0x0183, 0x0183 }, // L&       LATIN SMALL LETTER B WITH TOPBAR
    Interval{ 0x0185, 0x0185 }, // L&       LATIN SMALL LETTER TONE SIX
    Interval{ 0x0188, 0x0188 }, // L&       LATIN SMALL LETTER C WITH HOOK
    Interval{ 0x018C, 0x018C }, // L&       LATIN SMALL LETTER D WITH TOPBAR
    Interval{ 0x0192, 0x0192 }, // L&       LATIN SMALL LETTER F WITH HOOK
    Interval{ 0x0195, 0x0195 }, // L&       LATIN SMALL LETTER HV
    Interval{ 0x0199, 0x019B }, // L&   [3] LATIN SMALL LETTER K WITH HOOK..LATIN SMALL LETTER LAMBDA WITH STROKE
    Interval{ 0x019E, 0x019E }, // L&       LATIN SMALL LETTER N WITH LONG RIGHT LEG
    Interval{ 0x01A1, 0x01A1 }, // L&       LATIN SMALL LETTER O WITH HORN
    Interval{ 0x01A3, 0x01A3 }, // L&       LATIN SMALL LETTER OI
    Interval{ 0x01A5, 0x01A5 }, // L&       LATIN SMALL LETTER P WITH HOOK
    Interval{ 0x01A8, 0x01A8 }, // L&       LATIN SMALL LETTER TONE TWO
    Interval{ 0x01AD, 0x01AD }, // L&       LATIN SMALL LETTER T WITH HOOK
    Interval{ 0x01B0, 0x01B0 }, // L&       LATIN SMALL LETTER U WITH HORN
    Interval{ 0x01B4, 0x01B4 }, // L&       LATIN SMALL LETTER Y WITH HOOK
    Interval{ 0x01B6, 0x01B6 }, // L&       LATIN SMALL LETTER Z WITH STROKE
    Interval{ 0x01B9, 0x01B9 }, // L&       LATIN SMALL LETTER EZH REVERSED
    Interval{ 0x01BD, 0x01BD }, // L&       LATIN SMALL LETTER TONE FIVE
    Interval{ 0x01BF, 0x01BF }, // L&       LATIN LETTER WYNN
    Interval{ 0x01C5, 0x01C6 }, // L&   [2] LATIN CAPITAL LETTER D WITH SMALL LETTER Z WITH CARON..LATIN SMALL LETTER DZ WITH CARON
    Interval{ 0x01C8, 0x01C9 }, // L&   [2] LATIN CAPITAL LETTER L WITH SMALL LETTER J..LATIN SMALL LETTER LJ
    Interval{ 0x01CB, 0x01CC }, // L&   [2] LATIN CAPITAL LETTER N WITH SMALL LETTER J..LATIN SMALL LETTER NJ
    Interval{ 0x01CE, 0x01CE }, // L&       LATIN SMALL LETTER A WITH CARON
    Interval{ 0x01D0, 0x01D0 }, // L&       LATIN SMALL LETTER I WITH CARON
    Interval{ 0x01D2, 0x01D2 }, // L&       LATIN SMALL LETTER O WITH CARON
    Interval{ 0x01D4, 0x01D4 }, // L&       LATIN SMALL LETTER U WITH CARON
    Interval{ 0x01D6, 0x01D6 }, // L&       LATIN SMALL LETTER U WITH DIAERESIS AND MACRON
    Interval{ 0x01D8, 0x01D8 }, // L&       LATIN SMALL LETTER U WITH DIAERESIS AND ACUTE
    Interval{ 0x01DA, 0x01DA }, // L&       LATIN SMALL LETTER U WITH DIAERESIS AND CARON
    Interval{ 0x01DC, 0x01DD }, // L&   [2] LATIN SMALL LETTER U WITH DIAERESIS AND GRAVE..LATIN SMALL LETTER TURNED E
    Interval{ 0x01DF, 0x01DF }, // L&       LATIN SMALL LETTER A WITH DIAERESIS AND MACRON
    Interval{ 0x01E1, 0x01E1 }, // L&       LATIN SMALL LETTER A WITH DOT ABOVE AND MACRON
    Interval{ 0x01E3, 0x01E3 }, // L&       LATIN SMALL LETTER AE WITH MACRON
    Interval{ 0x01E5, 0x01E5 }, // L&       LATIN SMALL LETTER G WITH STROKE
    Interval{ 0x01E7, 0x01E7 }, // L&       LATIN SMALL LETTER G WITH CARON
    Interval{ 0x01E9, 0x01E9 }, // L&       LATIN SMALL LETTER K WITH CARON
    Interval{ 0x01EB, 0x01EB }, // L&       LATIN SMALL LETTER O WITH OGONEK
    Interval{ 0x01ED, 0x01ED }, // L&       LATIN SMALL LETTER O WITH OGONEK AND MACRON
    Interval{ 0x01EF, 0x01F0 }, // L&   [2] LATIN SMALL LETTER EZH WITH CARON..LATIN SMALL LETTER J WITH CARON
    Interval{ 0x01F2, 0x01F3 }, // L&   [2] LATIN CAPITAL LETTER D WITH SMALL LETTER Z..LATIN SMALL LETTER DZ
    Interval{ 0x01F5, 0x01F5 }, // L&       LATIN SMALL LETTER G WITH ACUTE
    Interval{ 0x01F9, 0x01F9 }, // L&       LATIN SMALL LETTER N WITH GRAVE
    Interval{ 0x01FB, 0x01FB }, // L&       LATIN SMALL LETTER A WITH RING ABOVE AND ACUTE
    Interval{ 0x01FD, 0x01FD }, // L&       LATIN SMALL LETTER AE WITH ACUTE
    Interval{ 0x01FF, 0x01FF }, // L&       LATIN SMALL LETTER O WITH STROKE AND ACUTE
    Interval{ 0x0201, 0x0201 }, // L&       LATIN SMALL LETTER A WITH DOUBLE GRAVE
    Interval{ 0x0203, 0x0203 }, // L&       LATIN SMALL LETTER A WITH INVERTED BREVE
    Interval{ 0x0205, 0x0205 }, // L&       LATIN SMALL LETTER E WITH DOUBLE GRAVE
    Interval{ 0x0207, 0x0207 }, // L&       LATIN SMALL LETTER E WITH INVERTED BREVE
    Interval{ 0x0209, 0x0209 }, // L&       LATIN SMALL LETTER I WITH DOUBLE GRAVE
    Interval{ 0x020B, 0x020B }, // L&       LATIN SMALL LETTER I WITH INVERTED BREVE
    Interval{ 0x020D, 0x020D }, // L&       LATIN SMALL LETTER O WITH DOUBLE GRAVE
    Interval{ 0x020F, 0x020F }, // L&       LATIN SMALL LETTER O WITH INVERTED BREVE
    Interval{ 0x0211, 0x0211 }, // L&       LATIN SMALL LETTER R WITH DOUBLE GRAVE
    Interval{ 0x0213, 0x0213 }, // L&       LATIN SMALL LETTER R WITH INVERTED BREVE
    Interval{ 0x0215, 0x0215 }, // L&       LATIN SMALL LETTER U WITH DOUBLE GRAVE
    Interval{ 0x0217, 0x0217 }, // L&       LATIN SMALL LETTER U WITH INVERTED BREVE
    Interval{ 0x0219, 0x0219 }, // L&       LATIN SMALL LETTER S WITH COMMA BELOW
    Interval{ 0x021B, 0x021B }, // L&       LATIN SMALL LETTER T WITH COMMA BELOW
    Interval{ 0x021D, 0x021D }, // L&       LATIN SMALL LETTER YOGH
    Interval{ 0x021F, 0x021F }, // L&       LATIN SMALL LETTER H WITH CARON
    Interval{ 0x0223, 0x0223 }, // L&       LATIN SMALL LETTER OU
    Interval{ 0x0225, 0x0225 }, // L&       LATIN SMALL LETTER Z WITH HOOK
    Interval{ 0x0227, 0x0227 }, // L&       LATIN SMALL LETTER A WITH DOT ABOVE
    Interval{ 0x0229, 0x0229 }, // L&       LATIN SMALL LETTER E WITH CEDILLA
    Interval{ 0x022B, 0x022B }, // L&       LATIN SMALL LETTER O WITH DIAERESIS AND MACRON
    Interval{ 0x022D, 0x022D }, // L&       LATIN SMALL LETTER O WITH TILDE AND MACRON
    Interval{ 0x022F, 0x022F }, // L&       LATIN SMALL LETTER O WITH DOT ABOVE
    Interval{ 0x0231, 0x0231 }, // L&       LATIN SMALL LETTER O WITH DOT ABOVE AND MACRON
    Interval{ 0x0233, 0x0233 }, // L&       LATIN SMALL LETTER Y WITH MACRON
    Interval{ 0x023C, 0x023C }, // L&       LATIN SMALL LETTER C WITH STROKE
    Interval{ 0x023F, 0x0240 }, // L&   [2] LATIN SMALL LETTER S WITH SWASH TAIL..LATIN SMALL LETTER Z WITH SWASH TAIL
    Interval{ 0x0242, 0x0242 }, // L&       LATIN SMALL LETTER GLOTTAL STOP
    Interval{ 0x0247, 0x0247 }, // L&       LATIN SMALL LETTER E WITH STROKE
    Interval{ 0x0249, 0x0249 }, // L&       LATIN SMALL LETTER J WITH STROKE
    Interval{ 0x024B, 0x024B }, // L&       LATIN SMALL LETTER Q WITH HOOK TAIL
    Interval{ 0x024D, 0x024D }, // L&       LATIN SMALL LETTER R WITH STROKE
    Interval{ 0x024F, 0x0254 }, // L&   [6] LATIN SMALL LETTER Y WITH STROKE..LATIN SMALL LETTER OPEN O
    Interval{ 0x0256, 0x0257 }, // L&   [2] LATIN SMALL LETTER D WITH TAIL..LATIN SMALL LETTER D WITH HOOK
    Interval{ 0x0259, 0x0259 }, // L&       LATIN SMALL LETTER SCHWA
    Interval{ 0x025B, 0x025C }, // L&   [2] LATIN SMALL LETTER OPEN E..LATIN SMALL LETTER REVERSED OPEN E
    Interval{ 0x0260, 0x0261 }, // L&   [2] LATIN SMALL LETTER G WITH HOOK..LATIN SMALL LETTER SCRIPT G
    Interval{ 0x0263, 0x0266 }, // L&   [4] LATIN SMALL LETTER GAMMA..LATIN SMALL LETTER H WITH HOOK
    Interval{ 0x0268, 0x026C }, // L&   [5] LATIN SMALL LETTER I WITH STROKE..LATIN SMALL LETTER L WITH BELT
    Interval{ 0x026F, 0x026F }, // L&       LATIN SMALL LETTER TURNED M
    Interval{ 0x0271, 0x0272 }, // L&   [2] LATIN SMALL LETTER M WITH HOOK..LATIN SMALL LETTER N WITH LEFT HOOK
    Interval{ 0x0275, 0x0275 }, // L&       LATIN SMALL LETTER BARRED O
    Interval{ 0x027D, 0x027D }, // L&       LATIN SMALL LETTER R WITH TAIL
    Interval{ 0x0280, 0x0280 }, // L&       LATIN LETTER SMALL CAPITAL R
    Interval{ 0x0282, 0x0283 }, // L&   [2] LATIN SMALL LETTER S WITH HOOK..LATIN SMALL LETTER ESH
    Interval{ 0x0287, 0x028C }, // L&   [6] LATIN SMALL LETTER TURNED T..LATIN SMALL LETTER TURNED V
    Interval{ 0x0292, 0x0292 }, // L&       LATIN SMALL LETTER EZH
    Interval{ 0x029D, 0x029E }, // L&   [2] LATIN SMALL LETTER J WITH CROSSED-TAIL..LATIN SMALL LETTER TURNED K
    Interval{ 0x0345, 0x0345 }, // Mn       COMBINING GREEK YPOGEGRAMMENI
    Interval{ 0x0371, 0x0371 }, // L&       GREEK SMALL LETTER HETA
    Interval{ 0x0373, 0x0373 }, // L&       GREEK SMALL LETTER ARCHAIC SAMPI
    Interval{ 0x0377, 0x0377 }, // L&       GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037B, 0x037D }, // L&   [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Interval{ 0x0390, 0x0390 }, // L&       GREEK SMALL LETTER IOTA WITH DIALYTIKA AND TONOS
    Interval{ 0x03AC, 0x03CE }, // L&  [35] GREEK SMALL LETTER ALPHA WITH TONOS..GREEK SMALL LETTER OMEGA WITH TONOS
    Interval{ 0x03D0, 0x03D1 }, // L&   [2] GREEK BETA SYMBOL..GREEK THETA SYMBOL
    Interval{ 0x03D5, 0x03D7 }, // L&   [3] GREEK PHI SYMBOL..GREEK KAI SYMBOL
    Interval{ 0x03D9, 0x03D9 }, // L&       GREEK SMALL LETTER ARCHAIC KOPPA
    Interval{ 0x03DB, 0x03DB }, // L&       GREEK SMALL LETTER STIGMA
    Interval{ 0x03DD, 0x03DD }, // L&       GREEK SMALL LETTER DIGAMMA
    Interval{ 0x03DF, 0x03DF }, // L&       GREEK SMALL LETTER KOPPA
    Interval{ 0x03E1, 0x03E1 }, // L&       GREEK SMALL LETTER SAMPI
    Interval{ 0x03E3, 0x03E3 }, // L&       COPTIC SMALL LETTER SHEI
    Interval{ 0x03E5, 0x03E5 }, // L&       COPTIC SMALL LETTER FEI
    Interval{ 0x03E7, 0x03E7 }, // L&       COPTIC SMALL LETTER KHEI
    Interval{ 0x03E9, 0x03E9 }, // L&       COPTIC SMALL LETTER HORI
    Interval{ 0x03EB, 0x03EB }, // L&       COPTIC SMALL LETTER GANGIA
    Interval{ 0x03ED, 0x03ED }, // L&       COPTIC SMALL LETTER SHIMA
    Interval{ 0x03EF, 0x03F3 }, // L&   [5] COPTIC SMALL LETTER DEI..GREEK LETTER YOT
    Interval{ 0x03F5, 0x03F5 }, // L&       GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F8, 0x03F8 }, // L&       GREEK SMALL LETTER SHO
    Interval{ 0x03FB, 0x03FB }, // L&       GREEK SMALL LETTER SAN
    Interval{ 0x0430, 0x045F }, // L&  [48] CYRILLIC SMALL LETTER A..CYRILLIC SMALL LETTER DZHE
    Interval{ 0x0461, 0x0461 }, // L&       CYRILLIC SMALL LETTER OMEGA
    Interval{ 0x0463, 0x0463 }, // L&       CYRILLIC SMALL LETTER YAT
    Interval{ 0x0465, 0x0465 }, // L&       CYRILLIC SMALL LETTER IOTIFIED E
    Interval{ 0x0467, 0x0467 }, // L&       CYRILLIC SMALL LETTER LITTLE YUS
    Interval{ 0x0469, 0x0469 }, // L&       CYRILLIC SMALL LETTER IOTIFIED LITTLE YUS
    Interval{ 0x046B, 0x046B }, // L&       CYRILLIC SMALL LETTER BIG YUS
    Interval{ 0x046D, 0x046D }, // L&       CYRILLIC SMALL LETTER IOTIFIED BIG YUS
    Interval{ 0x046F, 0x046F }, // L&       CYRILLIC SMALL LETTER KSI
    Interval{ 0x0471, 0x0471 }, // L&       CYRILLIC SMALL LETTER PSI
    Interval{ 0x0473, 0x0473 }, // L&       CYRILLIC SMALL LETTER FITA
    Interval{ 0x0475, 0x0475 }, // L&       CYRILLIC SMALL LETTER IZHITSA
    Interval{ 0x0477, 0x0477 }, // L&       CYRILLIC SMALL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT
    Interval{ 0x0479, 0x0479 }, // L&       CYRILLIC SMALL LETTER UK
    Interval{ 0x047B, 0x047B }, // L&       CYRILLIC SMALL LETTER ROUND OMEGA
    Interval{ 0x047D, 0x047D }, // L&       CYRILLIC SMALL LETTER OMEGA WITH TITLO
    Interval{ 0x047F, 0x047F }, // L&       CYRILLIC SMALL LETTER OT
    Interval{ 0x0481, 0x0481 }, // L&       CYRILLIC SMALL LETTER KOPPA
    Interval{ 0x048B, 0x048B }, // L&       CYRILLIC SMALL LETTER SHORT I WITH TAIL
    Interval{ 0x048D, 0x048D }, // L&       CYRILLIC SMALL LETTER SEMISOFT SIGN
    Interval{ 0x048F, 0x048F }, // L&       CYRILLIC SMALL LETTER ER WITH TICK
    Interval{ 0x0491, 0x0491 }, // L&       CYRILLIC SMALL LETTER GHE WITH UPTURN
    Interval{ 0x0493, 0x0493 }, // L&       CYRILLIC SMALL LETTER GHE WITH STROKE
    Interval{ 0x0495, 0x0495 }, // L&       CYRILLIC SMALL LETTER GHE WITH MIDDLE HOOK
    Interval{ 0x0497, 0x0497 }, // L&       CYRILLIC SMALL LETTER ZHE WITH DESCENDER
    Interval{ 0x0499, 0x0499 }, // L&       CYRILLIC SMALL LETTER ZE WITH DESCENDER
    Interval{ 0x049B, 0x049B }, // L&       CYRILLIC SMALL LETTER KA WITH DESCENDER
    Interval{ 0x049D, 0x049D }, // L&       CYRILLIC SMALL LETTER KA WITH VERTICAL STROKE
    Interval{ 0x049F, 0x049F }, // L&       CYRILLIC SMALL LETTER KA WITH STROKE
    Interval{ 0x04A1, 0x04A1 }, // L&       CYRILLIC SMALL LETTER BASHKIR KA
    Interval{ 0x04A3, 0x04A3 }, // L&       CYRILLIC SMALL LETTER EN WITH DESCENDER
    Interval{ 0x04A5, 0x04A5 }, // L&       CYRILLIC SMALL LIGATURE EN GHE
    Interval{ 0x04A7, 0x04A7 }, // L&       CYRILLIC SMALL LETTER PE WITH MIDDLE HOOK
    Interval{ 0x04A9, 0x04A9 }, // L&       CYRILLIC SMALL LETTER ABKHASIAN HA
    Interval{ 0x04AB, 0x04AB }, // L&       CYRILLIC SMALL LETTER ES WITH DESCENDER
    Interval{ 0x04AD, 0x04AD }, // L&       CYRILLIC SMALL LETTER TE WITH DESCENDER
    Interval{ 0x04AF, 0x04AF }, // L&       CYRILLIC SMALL LETTER STRAIGHT U
    Interval{ 0x04B1, 0x04B1 }, // L&       CYRILLIC SMALL LETTER STRAIGHT U WITH STROKE
    Interval{ 0x04B3, 0x04B3 }, // L&       CYRILLIC SMALL LETTER HA WITH DESCENDER
    Interval{ 0x04B5, 0x04B5 }, // L&       CYRILLIC SMALL LIGATURE TE TSE
    Interval{ 0x04B7, 0x04B7 }, // L&       CYRILLIC SMALL LETTER CHE WITH DESCENDER
    Interval{ 0x04B9, 0x04B9 }, // L&       CYRILLIC SMALL LETTER CHE WITH VERTICAL STROKE
    Interval{ 0x04BB, 0x04BB }, // L&       CYRILLIC SMALL LETTER SHHA
    Interval{ 0x04BD, 0x04BD }, // L&       CYRILLIC SMALL LETTER ABKHASIAN CHE
    Interval{ 0x04BF, 0x04BF }, // L&       CYRILLIC SMALL LETTER ABKHASIAN CHE WITH DESCENDER
    Interval{ 0x04C2, 0x04C2 }, // L&       CYRILLIC SMALL LETTER ZHE WITH BREVE
    Interval{ 0x04C4, 0x04C4 }, // L&       CYRILLIC SMALL LETTER KA WITH HOOK
    Interval{ 0x04C6, 0x04C6 }, // L&       CYRILLIC SMALL LETTER EL WITH TAIL
    Interval{ 0x04C8, 0x04C8 }, // L&       CYRILLIC SMALL LETTER EN WITH HOOK
    Interval{ 0x04CA, 0x04CA }, // L&       CYRILLIC SMALL LETTER EN WITH TAIL
    Interval{ 0x04CC, 0x04CC }, // L&       CYRILLIC SMALL LETTER KHAKASSIAN CHE
    Interval{ 0x04CE, 0x04CF }, // L&   [2] CYRILLIC SMALL LETTER EM WITH TAIL..CYRILLIC SMALL LETTER PALOCHKA
    Interval{ 0x04D1, 0x04D1 }, // L&       CYRILLIC SMALL LETTER A WITH BREVE
    Interval{ 0x04D3, 0x04D3 }, // L&       CYRILLIC SMALL LETTER A WITH DIAERESIS
    Interval{ 0x04D5, 0x04D5 }, // L&       CYRILLIC SMALL LIGATURE A IE
    Interval{ 0x04D7, 0x04D7 }, // L&       CYRILLIC SMALL LETTER IE WITH BREVE
    Interval{ 0x04D9, 0x04D9 }, // L&       CYRILLIC SMALL LETTER SCHWA
    Interval{ 0x04DB, 0x04DB }, // L&       CYRILLIC SMALL LETTER SCHWA WITH DIAERESIS
    Interval{ 0x04DD, 0x04DD }, // L&       CYRILLIC SMALL LETTER ZHE WITH DIAERESIS
    Interval{ 0x04DF, 0x04DF }, // L&       CYRILLIC SMALL LETTER ZE WITH DIAERESIS
    Interval{ 0x04E1, 0x04E1 }, // L&       CYRILLIC SMALL LETTER ABKHASIAN DZE
    Interval{ 0x04E3, 0x04E3 }, // L&       CYRILLIC SMALL LETTER I WITH MACRON
    Interval{ 0x04E5, 0x04E5 }, // L&       CYRILLIC SMALL LETTER I WITH DIAERESIS
    Interval{ 0x04E7, 0x04E7 }, // L&       CYRILLIC SMALL LETTER O WITH DIAERESIS
    Interval{ 0x04E9, 0x04E9 }, // L&       CYRILLIC SMALL LETTER BARRED O
    Interval{ 0x04EB, 0x04EB }, // L&       CYRILLIC SMALL LETTER BARRED O WITH DIAERESIS
    Interval{ 0x04ED, 0x04ED }, // L&       CYRILLIC SMALL LETTER E WITH DIAERESIS
    Interval{ 0x04EF, 0x04EF }, // L&       CYRILLIC SMALL LETTER U WITH MACRON
    Interval{ 0x04F1, 0x04F1 }, // L&       CYRILLIC SMALL LETTER U WITH DIAERESIS
    Interval{ 0x04F3, 0x04F3 }, // L&       CYRILLIC SMALL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x04F5, 0x04F5 }, // L&       CYRILLIC SMALL LETTER CHE WITH DIAERESIS
    Interval{ 0x04F7, 0x04F7 }, // L&       CYRILLIC SMALL LETTER GHE WITH DESCENDER
    Interval{ 0x04F9, 0x04F9 }, // L&       CYRILLIC SMALL LETTER YERU WITH DIAERESIS
    Interval{ 0x04FB, 0x04FB }, // L&       CYRILLIC SMALL LETTER GHE WITH STROKE AND HOOK
    Interval{ 0x04FD, 0x04FD }, // L&       CYRILLIC SMALL LETTER HA WITH HOOK
    Interval{ 0x04FF, 0x04FF }, // L&       CYRILLIC SMALL LETTER HA WITH STROKE
    Interval{ 0x0501, 0x0501 }, // L&       CYRILLIC SMALL LETTER KOMI DE
    Interval{ 0x0503, 0x0503 }, // L&       CYRILLIC SMALL LETTER KOMI DJE
    Interval{ 0x0505, 0x0505 }, // L&       CYRILLIC SMALL LETTER KOMI ZJE
    Interval{ 0x0507, 0x0507 }, // L&       CYRILLIC SMALL LETTER KOMI DZJE
    Interval{ 0x0509, 0x0509 }, // L&       CYRILLIC SMALL LETTER KOMI LJE
    Interval{ 0x050B, 0x050B }, // L&       CYRILLIC SMALL LETTER KOMI NJE
    Interval{ 0x050D, 0x050D }, // L&       CYRILLIC SMALL LETTER KOMI SJE
    Interval{ 0x050F, 0x050F }, // L&       CYRILLIC SMALL LETTER KOMI TJE
    Interval{ 0x0511, 0x0511 }, // L&       CYRILLIC SMALL LETTER REVERSED ZE
    Interval{ 0x0513, 0x0513 }, // L&       CYRILLIC SMALL LETTER EL WITH HOOK
    Interval{ 0x0515, 0x0515 }, // L&       CYRILLIC SMALL LETTER LHA
    Interval{ 0x0517, 0x0517 }, // L&       CYRILLIC SMALL LETTER RHA
    Interval{ 0x0519, 0x0519 }, // L&       CYRILLIC SMALL LETTER YAE
    Interval{ 0x051B, 0x051B }, // L&       CYRILLIC SMALL LETTER QA
    Interval{ 0x051D, 0x051D }, // L&       CYRILLIC SMALL LETTER WE
    Interval{ 0x051F, 0x051F }, // L&       CYRILLIC SMALL LETTER ALEUT KA
    Interval{ 0x0521, 0x0521 }, // L&       CYRILLIC SMALL LETTER EL WITH MIDDLE HOOK
    Interval{ 0x0523, 0x0523 }, // L&       CYRILLIC SMALL LETTER EN WITH MIDDLE HOOK
    Interval{ 0x0525, 0x0525 }, // L&       CYRILLIC SMALL LETTER PE WITH DESCENDER
    Interval{ 0x0527, 0x0527 }, // L&       CYRILLIC SMALL LETTER SHHA WITH DESCENDER
    Interval{ 0x0529, 0x0529 }, // L&       CYRILLIC SMALL LETTER EN WITH LEFT HOOK
    Interval{ 0x052B, 0x052B }, // L&       CYRILLIC SMALL LETTER DZZHE
    Interval{ 0x052D, 0x052D }, // L&       CYRILLIC SMALL LETTER DCHE
    Interval{ 0x052F, 0x052F }, // L&       CYRILLIC SMALL LETTER EL WITH DESCENDER
    Interval{ 0x0561, 0x0587 }, // L&  [39] ARMENIAN SMALL LETTER AYB..ARMENIAN SMALL LIGATURE ECH YIWN
    Interval{ 0x10D0, 0x10FA }, // L&  [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Interval{ 0x10FD, 0x10FF }, // L&   [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Interval{ 0x13F8, 0x13FD }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1C80, 0x1C88 }, // L&   [9] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER UNBLENDED UK
    Interval{ 0x1C8A, 0x1C8A }, // L&       CYRILLIC SMALL LETTER TJE
    Interval{ 0x1D79, 0x1D79 }, // L&       LATIN SMALL LETTER INSULAR G
    Interval{ 0x1D7D, 0x1D7D }, // L&       LATIN SMALL LETTER P WITH STROKE
    Interval{ 0x1D8E, 0x1D8E }, // L&       LATIN SMALL LETTER Z WITH PALATAL HOOK
    Interval{ 0x1E01, 0x1E01 }, // L&       LATIN SMALL LETTER A WITH RING BELOW
    Interval{ 0x1E03, 0x1E03 }, // L&       LATIN SMALL LETTER B WITH DOT ABOVE
    Interval{ 0x1E05, 0x1E05 }, // L&       LATIN SMALL LETTER B WITH DOT BELOW
    Interval{ 0x1E07, 0x1E07 }, // L&       LATIN SMALL LETTER B WITH LINE BELOW
    Interval{ 0x1E09, 0x1E09 }, // L&       LATIN SMALL LETTER C WITH CEDILLA AND ACUTE
    Interval{ 0x1E0B, 0x1E0B }, // L&       LATIN SMALL LETTER D WITH DOT ABOVE
    Interval{ 0x1E0D, 0x1E0D }, // L&       LATIN SMALL LETTER D WITH DOT BELOW
    Interval{ 0x1E0F, 0x1E0F }, // L&       LATIN SMALL LETTER D WITH LINE BELOW
    Interval{ 0x1E11, 0x1E11 }, // L&       LATIN SMALL LETTER D WITH CEDILLA
    Interval{ 0x1E13, 0x1E13 }, // L&       LATIN SMALL LETTER D WITH CIRCUMFLEX BELOW
    Interval{ 0x1E15, 0x1E15 }, // L&       LATIN SMALL LETTER E WITH MACRON AND GRAVE
    Interval{ 0x1E17, 0x1E17 }, // L&       LATIN SMALL LETTER E WITH MACRON AND ACUTE
    Interval{ 0x1E19, 0x1E19 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX BELOW
    Interval{ 0x1E1B, 0x1E1B }, // L&       LATIN SMALL LETTER E WITH TILDE BELOW
    Interval{ 0x1E1D, 0x1E1D }, // L&       LATIN SMALL LETTER E WITH CEDILLA AND BREVE
    Interval{ 0x1E1F, 0x1E1F }, // L&       LATIN SMALL LETTER F WITH DOT ABOVE
    Interval{ 0x1E21, 0x1E21 }, // L&       LATIN SMALL LETTER G WITH MACRON
    Interval{ 0x1E23, 0x1E23 }, // L&       LATIN SMALL LETTER H WITH DOT ABOVE
    Interval{ 0x1E25, 0x1E25 }, // L&       LATIN SMALL LETTER H WITH DOT BELOW
    Interval{ 0x1E27, 0x1E27 }, // L&       LATIN SMALL LETTER H WITH DIAERESIS
    Interval{ 0x1E29, 0x1E29 }, // L&       LATIN SMALL LETTER H WITH CEDILLA
    Interval{ 0x1E2B, 0x1E2B }, // L&       LATIN SMALL LETTER H WITH BREVE BELOW
    Interval{ 0x1E2D, 0x1E2D }, // L&       LATIN SMALL LETTER I WITH TILDE BELOW
    Interval{ 0x1E2F, 0x1E2F }, // L&       LATIN SMALL LETTER I WITH DIAERESIS AND ACUTE
    Interval{ 0x1E31, 0x1E31 }, // L&       LATIN SMALL LETTER K WITH ACUTE
    Interval{ 0x1E33, 0x1E33 }, // L&       LATIN SMALL LETTER K WITH DOT BELOW
    Interval{ 0x1E35, 0x1E35 }, // L&       LATIN SMALL LETTER K WITH LINE BELOW
    Interval{ 0x1E37, 0x1E37 }, // L&       LATIN SMALL LETTER L WITH DOT BELOW
    Interval{ 0x1E39, 0x1E39 }, // L&       LATIN SMALL LETTER L WITH DOT BELOW AND MACRON
    Interval{ 0x1E3B, 0x1E3B }, // L&       LATIN SMALL LETTER L WITH LINE BELOW
    Interval{ 0x1E3D, 0x1E3D }, // L&       LATIN SMALL LETTER L WITH CIRCUMFLEX BELOW
    Interval{ 0x1E3F, 0x1E3F }, // L&       LATIN SMALL LETTER M WITH ACUTE
    Interval{ 0x1E41, 0x1E41 }, // L&       LATIN SMALL LETTER M WITH DOT ABOVE
    Interval{ 0x1E43, 0x1E43 }, // L&       LATIN SMALL LETTER M WITH DOT BELOW
    Interval{ 0x1E45, 0x1E45 }, // L&       LATIN SMALL LETTER N WITH DOT ABOVE
    Interval{ 0x1E47, 0x1E47 }, // L&       LATIN SMALL LETTER N WITH DOT BELOW
    Interval{ 0x1E49, 0x1E49 }, // L&       LATIN SMALL LETTER N WITH LINE BELOW
    Interval{ 0x1E4B, 0x1E4B }, // L&       LATIN SMALL LETTER N WITH CIRCUMFLEX BELOW
    Interval{ 0x1E4D, 0x1E4D }, // L&       LATIN SMALL LETTER O WITH TILDE AND ACUTE
    Interval{ 0x1E4F, 0x1E4F }, // L&       LATIN SMALL LETTER O WITH TILDE AND DIAERESIS
    Interval{ 0x1E51, 0x1E51 }, // L&       LATIN SMALL LETTER O WITH MACRON AND GRAVE
    Interval{ 0x1E53, 0x1E53 }, // L&       LATIN SMALL LETTER O WITH MACRON AND ACUTE
    Interval{ 0x1E55, 0x1E55 }, // L&       LATIN SMALL LETTER P WITH ACUTE
    Interval{ 0x1E57, 0x1E57 }, // L&       LATIN SMALL LETTER P WITH DOT ABOVE
    Interval{ 0x1E59, 0x1E59 }, // L&       LATIN SMALL LETTER R WITH DOT ABOVE
    Interval{ 0x1E5B, 0x1E5B }, // L&       LATIN SMALL LETTER R WITH DOT BELOW
    Interval{ 0x1E5D, 0x1E5D }, // L&       LATIN SMALL LETTER R WITH DOT BELOW AND MACRON
    Interval{ 0x1E5F, 0x1E5F }, // L&       LATIN SMALL LETTER R WITH LINE BELOW
    Interval{ 0x1E61, 0x1E61 }, // L&       LATIN SMALL LETTER S WITH DOT ABOVE
    Interval{ 0x1E63, 0x1E63 }, // L&       LATIN SMALL LETTER S WITH DOT BELOW
    Interval{ 0x1E65, 0x1E65 }, // L&       LATIN SMALL LETTER S WITH ACUTE AND DOT ABOVE
    Interval{ 0x1E67, 0x1E67 }, // L&       LATIN SMALL LETTER S WITH CARON AND DOT ABOVE
    Interval{ 0x1E69, 0x1E69 }, // L&       LATIN SMALL LETTER S WITH DOT BELOW AND DOT ABOVE
    Interval{ 0x1E6B, 0x1E6B }, // L&       LATIN SMALL LETTER T WITH DOT ABOVE
    Interval{ 0x1E6D, 0x1E6D }, // L&       LATIN SMALL LETTER T WITH DOT BELOW
    Interval{ 0x1E6F, 0x1E6F }, // L&       LATIN SMALL LETTER T WITH LINE BELOW
    Interval{ 0x1E71, 0x1E71 }, // L&       LATIN SMALL LETTER T WITH CIRCUMFLEX BELOW
    Interval{ 0x1E73, 0x1E73 }, // L&       LATIN SMALL LETTER U WITH DIAERESIS BELOW
    Interval{ 0x1E75, 0x1E75 }, // L&       LATIN SMALL LETTER U WITH TILDE BELOW
    Interval{ 0x1E77, 0x1E77 }, // L&       LATIN SMALL LETTER U WITH CIRCUMFLEX BELOW
    Interval{ 0x1E79, 0x1E79 }, // L&       LATIN SMALL LETTER U WITH TILDE AND ACUTE
    Interval{ 0x1E7B, 0x1E7B }, // L&       LATIN SMALL LETTER U WITH MACRON AND DIAERESIS
    Interval{ 0x1E7D, 0x1E7D }, // L&       LATIN SMALL LETTER V WITH TILDE
    Interval{ 0x1E7F, 0x1E7F }, // L&       LATIN SMALL LETTER V WITH DOT BELOW
    Interval{ 0x1E81, 0x1E81 }, // L&       LATIN SMALL LETTER W WITH GRAVE
    Interval{ 0x1E83, 0x1E83 }, // L&       LATIN SMALL LETTER W WITH ACUTE
    Interval{ 0x1E85, 0x1E85 }, // L&       LATIN SMALL LETTER W WITH DIAERESIS
    Interval{ 0x1E87, 0x1E87 }, // L&       LATIN SMALL LETTER W WITH DOT ABOVE
    Interval{ 0x1E89, 0x1E89 }, // L&       LATIN SMALL LETTER W WITH DOT BELOW
    Interval{ 0x1E8B, 0x1E8B }, // L&       LATIN SMALL LETTER X WITH DOT ABOVE
    Interval{ 0x1E8D, 0x1E8D }, // L&       LATIN SMALL LETTER X WITH DIAERESIS
    Interval{ 0x1E8F, 0x1E8F }, // L&       LATIN SMALL LETTER Y WITH DOT ABOVE
    Interval{ 0x1E91, 0x1E91 }, // L&       LATIN SMALL LETTER Z WITH CIRCUMFLEX
    Interval{ 0x1E93, 0x1E93 }, // L&       LATIN SMALL LETTER Z WITH DOT BELOW
    Interval{ 0x1E95, 0x1E9B }, // L&   [7] LATIN SMALL LETTER Z WITH LINE BELOW..LATIN SMALL LETTER LONG S WITH DOT ABOVE
    Interval{ 0x1EA1, 0x1EA1 }, // L&       LATIN SMALL LETTER A WITH DOT BELOW
    Interval{ 0x1EA3, 0x1EA3 }, // L&       LATIN SMALL LETTER A WITH HOOK ABOVE
    Interval{ 0x1EA5, 0x1EA5 }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EA7, 0x1EA7 }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EA9, 0x1EA9 }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EAB, 0x1EAB }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EAD, 0x1EAD }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EAF, 0x1EAF }, // L&       LATIN SMALL LETTER A WITH BREVE AND ACUTE
    Interval{ 0x1EB1, 0x1EB1 }, // L&       LATIN SMALL LETTER A WITH BREVE AND GRAVE
    Interval{ 0x1EB3, 0x1EB3 }, // L&       LATIN SMALL LETTER A WITH BREVE AND HOOK ABOVE
    Interval{ 0x1EB5, 0x1EB5 }, // L&       LATIN SMALL LETTER A WITH BREVE AND TILDE
    Interval{ 0x1EB7, 0x1EB7 }, // L&       LATIN SMALL LETTER A WITH BREVE AND DOT BELOW
    Interval{ 0x1EB9, 0x1EB9 }, // L&       LATIN SMALL LETTER E WITH DOT BELOW
    Interval{ 0x1EBB, 0x1EBB }, // L&       LATIN SMALL LETTER E WITH HOOK ABOVE
    Interval{ 0x1EBD, 0x1EBD }, // L&       LATIN SMALL LETTER E WITH TILDE
    Interval{ 0x1EBF, 0x1EBF }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EC1, 0x1EC1 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EC3, 0x1EC3 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EC5, 0x1EC5 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EC7, 0x1EC7 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EC9, 0x1EC9 }, // L&       LATIN SMALL LETTER I WITH HOOK ABOVE
    Interval{ 0x1ECB, 0x1ECB }, // L&       LATIN SMALL LETTER I WITH DOT BELOW
    Interval{ 0x1ECD, 0x1ECD }, // L&       LATIN SMALL LETTER O WITH DOT BELOW
    Interval{ 0x1ECF, 0x1ECF }, // L&       LATIN SMALL LETTER O WITH HOOK ABOVE
    Interval{ 0x1ED1, 0x1ED1 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1ED3, 0x1ED3 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1ED5, 0x1ED5 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1ED7, 0x1ED7 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1ED9, 0x1ED9 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EDB, 0x1EDB }, // L&       LATIN SMALL LETTER O WITH HORN AND ACUTE
    Interval{ 0x1EDD, 0x1EDD }, // L&       LATIN SMALL LETTER O WITH HORN AND GRAVE
    Interval{ 0x1EDF, 0x1EDF }, // L&       LATIN SMALL LETTER O WITH HORN AND HOOK ABOVE
    Interval{ 0x1EE1, 0x1EE1 }, // L&       LATIN SMALL LETTER O WITH HORN AND TILDE
    Interval{ 0x1EE3, 0x1EE3 }, // L&       LATIN SMALL LETTER O WITH HORN AND DOT BELOW
    Interval{ 0x1EE5, 0x1EE5 }, // L&       LATIN SMALL LETTER U WITH DOT BELOW
    Interval{ 0x1EE7, 0x1EE7 }, // L&       LATIN SMALL LETTER U WITH HOOK ABOVE
    Interval{ 0x1EE9, 0x1EE9 }, // L&       LATIN SMALL LETTER U WITH HORN AND ACUTE
    Interval{ 0x1EEB, 0x1EEB }, // L&       LATIN SMALL LETTER U WITH HORN AND GRAVE
    Interval{ 0x1EED, 0x1EED }, // L&       LATIN SMALL LETTER U WITH HORN AND HOOK ABOVE
    Interval{ 0x1EEF, 0x1EEF }, // L&       LATIN SMALL LETTER U WITH HORN AND TILDE
    Interval{ 0x1EF1, 0x1EF1 }, // L&       LATIN SMALL LETTER U WITH HORN AND DOT BELOW
    Interval{ 0x1EF3, 0x1EF3 }, // L&       LATIN SMALL LETTER Y WITH GRAVE
    Interval{ 0x1EF5, 0x1EF5 }, // L&       LATIN SMALL LETTER Y WITH DOT BELOW
    Interval{ 0x1EF7, 0x1EF7 }, // L&       LATIN SMALL LETTER Y WITH HOOK ABOVE
    Interval{ 0x1EF9, 0x1EF9 }, // L&       LATIN SMALL LETTER Y WITH TILDE
    Interval{ 0x1EFB, 0x1EFB }, // L&       LATIN SMALL LETTER MIDDLE-WELSH LL
    Interval{ 0x1EFD, 0x1EFD }, // L&       LATIN SMALL LETTER MIDDLE-WELSH V
    Interval{ 0x1EFF, 0x1F07 }, // L&   [9] LATIN SMALL LETTER Y WITH LOOP..GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F10, 0x1F15 }, // L&   [6] GREEK SMALL LETTER EPSILON WITH PSILI..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F20, 0x1F27 }, // L&   [8] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F30, 0x1F37 }, // L&   [8] GREEK SMALL LETTER IOTA WITH PSILI..GREEK SMALL LETTER IOTA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F40, 0x1F45 }, // L&   [6] GREEK SMALL LETTER OMICRON WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F50, 0x1F57 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F60, 0x1F67 }, // L&   [8] GREEK SMALL LETTER OMEGA WITH PSILI..GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F70, 0x1F7D }, // L&  [14] GREEK SMALL LETTER ALPHA WITH VARIA..GREEK SMALL LETTER OMEGA WITH OXIA
    Interval{ 0x1F80, 0x1FB4 }, // L&  [53] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB6, 0x1FB7 }, // L&   [2] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK SMALL LETTER ALPHA WITH PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FBC, 0x1FBC }, // L&       GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Interval{ 0x1FBE, 0x1FBE }, // L&       GREEK PROSGEGRAMMENI
    Interval{ 0x1FC2, 0x1FC4 }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC6, 0x1FC7 }, // L&   [2] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK SMALL LETTER ETA WITH PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FCC, 0x1FCC }, // L&       GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Interval{ 0x1FD0, 0x1FD3 }, // L&   [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Interval{ 0x1FD6, 0x1FD7 }, // L&   [2] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND PERISPOMENI
    Interval{ 0x1FE0, 0x1FE7 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND PERISPOMENI
    Interval{ 0x1FF2, 0x1FF4 }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF6, 0x1FF7 }, // L&   [2] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK SMALL LETTER OMEGA WITH PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FFC, 0x1FFC }, // L&       GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Interval{ 0x214E, 0x214E }, // L&       TURNED SMALL F
    Interval{ 0x2170, 0x217F }, // Nl  [16] SMALL ROMAN NUMERAL ONE..SMALL ROMAN NUMERAL ONE THOUSAND
    Interval{ 0x2184, 0x2184 }, // L&       LATIN SMALL LETTER REVERSED C
    Interval{ 0x24D0, 0x24E9 }, // So  [26] CIRCLED LATIN SMALL LETTER A..CIRCLED LATIN SMALL LETTER Z
    Interval{ 0x2C30, 0x2C5F }, // L&  [48] GLAGOLITIC SMALL LETTER AZU..GLAGOLITIC SMALL LETTER CAUDATE CHRIVI
    Interval{ 0x2C61, 0x2C61 }, // L&       LATIN SMALL LETTER L WITH DOUBLE BAR
    Interval{ 0x2C65, 0x2C66 }, // L&   [2] LATIN SMALL LETTER A WITH STROKE..LATIN SMALL LETTER T WITH DIAGONAL STROKE
    Interval{ 0x2C68, 0x2C68 }, // L&       LATIN SMALL LETTER H WITH DESCENDER
    Interval{ 0x2C6A, 0x2C6A }, // L&       LATIN SMALL LETTER K WITH DESCENDER
    Interval{ 0x2C6C, 0x2C6C }, // L&       LATIN SMALL LETTER Z WITH DESCENDER
    Interval{ 0x2C73, 0x2C73 }, // L&       LATIN SMALL LETTER W WITH HOOK
    Interval{ 0x2C76, 0x2C76 }, // L&       LATIN SMALL LETTER HALF H
    Interval{ 0x2C81, 0x2C81 }, // L&       COPTIC SMALL LETTER ALFA
    Interval{ 0x2C83, 0x2C83 }, // L&       COPTIC SMALL LETTER VIDA
    Interval{ 0x2C85, 0x2C85 }, // L&       COPTIC SMALL LETTER GAMMA
    Interval{ 0x2C87, 0x2C87 }, // L&       COPTIC SMALL LETTER DALDA
    Interval{ 0x2C89, 0x2C89 }, // L&       COPTIC SMALL LETTER EIE
    Interval{ 0x2C8B, 0x2C8B }, // L&       COPTIC SMALL LETTER SOU
    Interval{ 0x2C8D, 0x2C8D }, // L&       COPTIC SMALL LETTER ZATA
    Interval{ 0x2C8F, 0x2C8F }, // L&       COPTIC SMALL LETTER HATE
    Interval{ 0x2C91, 0x2C91 }, // L&       COPTIC SMALL LETTER THETHE
    Interval{ 0x2C93, 0x2C93 }, // L&       COPTIC SMALL LETTER IAUDA
    Interval{ 0x2C95, 0x2C95 }, // L&       COPTIC SMALL LETTER KAPA
    Interval{ 0x2C97, 0x2C97 }, // L&       COPTIC SMALL LETTER LAULA
    Interval{ 0x2C99, 0x2C99 }, // L&       COPTIC SMALL LETTER MI
    Interval{ 0x2C9B, 0x2C9B }, // L&       COPTIC SMALL LETTER NI
    Interval{ 0x2C9D, 0x2C9D }, // L&       COPTIC SMALL LETTER KSI
    Interval{ 0x2C9F, 0x2C9F }, // L&       COPTIC SMALL LETTER O
    Interval{ 0x2CA1, 0x2CA1 }, // L&       COPTIC SMALL LETTER PI
    Interval{ 0x2CA3, 0x2CA3 }, // L&       COPTIC SMALL LETTER RO
    Interval{ 0x2CA5, 0x2CA5 }, // L&       COPTIC SMALL LETTER SIMA
    Interval{ 0x2CA7, 0x2CA7 }, // L&       COPTIC SMALL LETTER TAU
    Interval{ 0x2CA9, 0x2CA9 }, // L&       COPTIC SMALL LETTER UA
    Interval{ 0x2CAB, 0x2CAB }, // L&       COPTIC SMALL LETTER FI
    Interval{ 0x2CAD, 0x2CAD }, // L&       COPTIC SMALL LETTER KHI
    Interval{ 0x2CAF, 0x2CAF }, // L&       COPTIC SMALL LETTER PSI
    Interval{ 0x2CB1, 0x2CB1 }, // L&       COPTIC SMALL LETTER OOU
    Interval{ 0x2CB3, 0x2CB3 }, // L&       COPTIC SMALL LETTER DIALECT-P ALEF
    Interval{ 0x2CB5, 0x2CB5 }, // L&       COPTIC SMALL LETTER OLD COPTIC AIN
    Interval{ 0x2CB7, 0x2CB7 }, // L&       COPTIC SMALL LETTER CRYPTOGRAMMIC EIE
    Interval{ 0x2CB9, 0x2CB9 }, // L&       COPTIC SMALL LETTER DIALECT-P KAPA
    Interval{ 0x2CBB, 0x2CBB }, // L&       COPTIC SMALL LETTER DIALECT-P NI
    Interval{ 0x2CBD, 0x2CBD }, // L&       COPTIC SMALL LETTER CRYPTOGRAMMIC NI
    Interval{ 0x2CBF, 0x2CBF }, // L&       COPTIC SMALL LETTER OLD COPTIC OOU
    Interval{ 0x2CC1, 0x2CC1 }, // L&       COPTIC SMALL LETTER SAMPI
    Interval{ 0x2CC3, 0x2CC3 }, // L&       COPTIC SMALL LETTER CROSSED SHEI
    Interval{ 0x2CC5, 0x2CC5 }, // L&       COPTIC SMALL LETTER OLD COPTIC SHEI
    Interval{ 0x2CC7, 0x2CC7 }, // L&       COPTIC SMALL LETTER OLD COPTIC ESH
    Interval{ 0x2CC9, 0x2CC9 }, // L&       COPTIC SMALL LETTER AKHMIMIC KHEI
    Interval{ 0x2CCB, 0x2CCB }, // L&       COPTIC SMALL LETTER DIALECT-P HORI
    Interval{ 0x2CCD, 0x2CCD }, // L&       COPTIC SMALL LETTER OLD COPTIC HORI
    Interval{ 0x2CCF, 0x2CCF }, // L&       COPTIC SMALL LETTER OLD COPTIC HA
    Interval{ 0x2CD1, 0x2CD1 }, // L&       COPTIC SMALL LETTER L-SHAPED HA
    Interval{ 0x2CD3, 0x2CD3 }, // L&       COPTIC SMALL LETTER OLD COPTIC HEI
    Interval{ 0x2CD5, 0x2CD5 }, // L&       COPTIC SMALL LETTER OLD COPTIC HAT
    Interval{ 0x2CD7, 0x2CD7 }, // L&       COPTIC SMALL LETTER OLD COPTIC GANGIA
    Interval{ 0x2CD9, 0x2CD9 }, // L&       COPTIC SMALL LETTER OLD COPTIC DJA
    Interval{ 0x2CDB, 0x2CDB }, // L&       COPTIC SMALL LETTER OLD COPTIC SHIMA
    Interval{ 0x2CDD, 0x2CDD }, // L&       COPTIC SMALL LETTER OLD NUBIAN SHIMA
    Interval{ 0x2CDF, 0x2CDF }, // L&       COPTIC SMALL LETTER OLD NUBIAN NGI
    Interval{ 0x2CE1, 0x2CE1 }, // L&       COPTIC SMALL LETTER OLD NUBIAN NYI
    Interval{ 0x2CE3, 0x2CE3 }, // L&       COPTIC SMALL LETTER OLD NUBIAN WAU
    Interval{ 0x2CEC, 0x2CEC }, // L&       COPTIC SMALL LETTER CRYPTOGRAMMIC SHEI
    Interval{ 0x2CEE, 0x2CEE }, // L&       COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF3, 0x2CF3 }, // L&       COPTIC SMALL LETTER BOHAIRIC KHEI
    Interval{ 0x2D00, 0x2D25 }, // L&  [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Interval{ 0x2D27, 0x2D27 }, // L&       GEORGIAN SMALL LETTER YN
    Interval{ 0x2D2D, 0x2D2D }, // L&       GEORGIAN SMALL LETTER AEN
    Interval{ 0xA641, 0xA641 }, // L&       CYRILLIC SMALL LETTER ZEMLYA
    Interval{ 0xA643, 0xA643 }, // L&       CYRILLIC SMALL LETTER DZELO
    Interval{ 0xA645, 0xA645 }, // L&       CYRILLIC SMALL LETTER REVERSED DZE
    Interval{ 0xA647, 0xA647 }, // L&       CYRILLIC SMALL LETTER IOTA
    Interval{ 0xA649, 0xA649 }, // L&       CYRILLIC SMALL LETTER DJERV
    Interval{ 0xA64B, 0xA64B }, // L&       CYRILLIC SMALL LETTER MONOGRAPH UK
    Interval{ 0xA64D, 0xA64D }, // L&       CYRILLIC SMALL LETTER BROAD OMEGA
    Interval{ 0xA64F, 0xA64F }, // L&       CYRILLIC SMALL LETTER NEUTRAL YER
    Interval{ 0xA651, 0xA651 }, // L&       CYRILLIC SMALL LETTER YERU WITH BACK YER
    Interval{ 0xA653, 0xA653 }, // L&       CYRILLIC SMALL LETTER IOTIFIED YAT
    Interval{ 0xA655, 0xA655 }, // L&       CYRILLIC SMALL LETTER REVERSED YU
    Interval{ 0xA657, 0xA657 }, // L&       CYRILLIC SMALL LETTER IOTIFIED A
    Interval{ 0xA659, 0xA659 }, // L&       CYRILLIC SMALL LETTER CLOSED LITTLE YUS
    Interval{ 0xA65B, 0xA65B }, // L&       CYRILLIC SMALL LETTER BLENDED YUS
    Interval{ 0xA65D, 0xA65D }, // L&       CYRILLIC SMALL LETTER IOTIFIED CLOSED LITTLE YUS
    Interval{ 0xA65F, 0xA65F }, // L&       CYRILLIC SMALL LETTER YN
    Interval{ 0xA661, 0xA661 }, // L&       CYRILLIC SMALL LETTER REVERSED TSE
    Interval{ 0xA663, 0xA663 }, // L&       CYRILLIC SMALL LETTER SOFT DE
    Interval{ 0xA665, 0xA665 }, // L&       CYRILLIC SMALL LETTER SOFT EL
    Interval{ 0xA667, 0xA667 }, // L&       CYRILLIC SMALL LETTER SOFT EM
    Interval{ 0xA669, 0xA669 }, // L&       CYRILLIC SMALL LETTER MONOCULAR O
    Interval{ 0xA66B, 0xA66B }, // L&       CYRILLIC SMALL LETTER BINOCULAR O
    Interval{ 0xA66D, 0xA66D }, // L&       CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA681, 0xA681 }, // L&       CYRILLIC SMALL LETTER DWE
    Interval{ 0xA683, 0xA683 }, // L&       CYRILLIC SMALL LETTER DZWE
    Interval{ 0xA685, 0xA685 }, // L&       CYRILLIC SMALL LETTER ZHWE
    Interval{ 0xA687, 0xA687 }, // L&       CYRILLIC SMALL LETTER CCHE
    Interval{ 0xA689, 0xA689 }, // L&       CYRILLIC SMALL LETTER DZZE
    Interval{ 0xA68B, 0xA68B }, // L&       CYRILLIC SMALL LETTER TE WITH MIDDLE HOOK
    Interval{ 0xA68D, 0xA68D }, // L&       CYRILLIC SMALL LETTER TWE
    Interval{ 0xA68F, 0xA68F }, // L&       CYRILLIC SMALL LETTER TSWE
    Interval{ 0xA691, 0xA691 }, // L&       CYRILLIC SMALL LETTER TSSE
    Interval{ 0xA693, 0xA693 }, // L&       CYRILLIC SMALL LETTER TCHE
    Interval{ 0xA695, 0xA695 }, // L&       CYRILLIC SMALL LETTER HWE
    Interval{ 0xA697, 0xA697 }, // L&       CYRILLIC SMALL LETTER SHWE
    Interval{ 0xA699, 0xA699 }, // L&       CYRILLIC SMALL LETTER DOUBLE O
    Interval{ 0xA69B, 0xA69B }, // L&       CYRILLIC SMALL LETTER CROSSED O
    Interval{ 0xA723, 0xA723 }, // L&       LATIN SMALL LETTER EGYPTOLOGICAL ALEF
    Interval{ 0xA725, 0xA725 }, // L&       LATIN SMALL LETTER EGYPTOLOGICAL AIN
    Interval{ 0xA727, 0xA727 }, // L&       LATIN SMALL LETTER HENG
    Interval{ 0xA729, 0xA729 }, // L&       LATIN SMALL LETTER TZ
    Interval{ 0xA72B, 0xA72B }, // L&       LATIN SMALL LETTER TRESILLO
    Interval{ 0xA72D, 0xA72D }, // L&       LATIN SMALL LETTER CUATRILLO
    Interval{ 0xA72F, 0xA72F }, // L&       LATIN SMALL LETTER CUATRILLO WITH COMMA
    Interval{ 0xA733, 0xA733 }, // L&       LATIN SMALL LETTER AA
    Interval{ 0xA735, 0xA735 }, // L&       LATIN SMALL LETTER AO
    Interval{ 0xA737, 0xA737 }, // L&       LATIN SMALL LETTER AU
    Interval{ 0xA739, 0xA739 }, // L&       LATIN SMALL LETTER AV
    Interval{ 0xA73B, 0xA73B }, // L&       LATIN SMALL LETTER AV WITH HORIZONTAL BAR
    Interval{ 0xA73D, 0xA73D }, // L&       LATIN SMALL LETTER AY
    Interval{ 0xA73F, 0xA73F }, // L&       LATIN SMALL LETTER REVERSED C WITH DOT
    Interval{ 0xA741, 0xA741 }, // L&       LATIN SMALL LETTER K WITH STROKE
    Interval{ 0xA743, 0xA743 }, // L&       LATIN SMALL LETTER K WITH DIAGONAL STROKE
    Interval{ 0xA745, 0xA745 }, // L&       LATIN SMALL LETTER K WITH STROKE AND DIAGONAL STROKE
    Interval{ 0xA747, 0xA747 }, // L&       LATIN SMALL LETTER BROKEN L
    Interval{ 0xA749, 0xA749 }, // L&       LATIN SMALL LETTER L WITH HIGH STROKE
    Interval{ 0xA74B, 0xA74B }, // L&       LATIN SMALL LETTER O WITH LONG STROKE OVERLAY
    Interval{ 0xA74D, 0xA74D }, // L&       LATIN SMALL LETTER O WITH LOOP
    Interval{ 0xA74F, 0xA74F }, // L&       LATIN SMALL LETTER OO
    Interval{ 0xA751, 0xA751 }, // L&       LATIN SMALL LETTER P WITH STROKE THROUGH DESCENDER
    Interval{ 0xA753, 0xA753 }, // L&       LATIN SMALL LETTER P WITH FLOURISH
    Interval{ 0xA755, 0xA755 }, // L&       LATIN SMALL LETTER P WITH SQUIRREL TAIL
    Interval{ 0xA757, 0xA757 }, // L&       LATIN SMALL LETTER Q WITH STROKE THROUGH DESCENDER
    Interval{ 0xA759, 0xA759 }, // L&       LATIN SMALL LETTER Q WITH DIAGONAL STROKE
    Interval{ 0xA75B, 0xA75B }, // L&       LATIN SMALL LETTER R ROTUNDA
    Interval{ 0xA75D, 0xA75D }, // L&       LATIN SMALL LETTER RUM ROTUNDA
    Interval{ 0xA75F, 0xA75F }, // L&       LATIN SMALL LETTER V WITH DIAGONAL STROKE
    Interval{ 0xA761, 0xA761 }, // L&       LATIN SMALL LETTER VY
    Interval{ 0xA763, 0xA763 }, // L&       LATIN SMALL LETTER VISIGOTHIC Z
    Interval{ 0xA765, 0xA765 }, // L&       LATIN SMALL LETTER THORN WITH STROKE
    Interval{ 0xA767, 0xA767 }, // L&       LATIN SMALL LETTER THORN WITH STROKE THROUGH DESCENDER
    Interval{ 0xA769, 0xA769 }, // L&       LATIN SMALL LETTER VEND
    Interval{ 0xA76B, 0xA76B }, // L&       LATIN SMALL LETTER ET
    Interval{ 0xA76D, 0xA76D }, // L&       LATIN SMALL LETTER IS
    Interval{ 0xA76F, 0xA76F }, // L&       LATIN SMALL LETTER CON
    Interval{ 0xA77A, 0xA77A }, // L&       LATIN SMALL LETTER INSULAR D
    Interval{ 0xA77C, 0xA77C }, // L&       LATIN SMALL LETTER INSULAR F
    Interval{ 0xA77F, 0xA77F }, // L&       LATIN SMALL LETTER TURNED INSULAR G
    Interval{ 0xA781, 0xA781 }, // L&       LATIN SMALL LETTER TURNED L
    Interval{ 0xA783, 0xA783 }, // L&       LATIN SMALL LETTER INSULAR R
    Interval{ 0xA785, 0xA785 }, // L&       LATIN SMALL LETTER INSULAR S
    Interval{ 0xA787, 0xA787 }, // L&       LATIN SMALL LETTER INSULAR T
    Interval{ 0xA78C, 0xA78C }, // L&       LATIN SMALL LETTER SALTILLO
    Interval{ 0xA791, 0xA791 }, // L&       LATIN SMALL LETTER N WITH DESCENDER
    Interval{ 0xA793, 0xA794 }, // L&   [2] LATIN SMALL LETTER C WITH BAR..LATIN SMALL LETTER C WITH PALATAL HOOK
    Interval{ 0xA797, 0xA797 }, // L&       LATIN SMALL LETTER B WITH FLOURISH
    Interval{ 0xA799, 0xA799 }, // L&       LATIN SMALL LETTER F WITH STROKE
    Interval{ 0xA79B, 0xA79B }, // L&       LATIN SMALL LETTER VOLAPUK AE
    Interval{ 0xA79D, 0xA79D }, // L&       LATIN SMALL LETTER VOLAPUK OE
    Interval{ 0xA79F, 0xA79F }, // L&       LATIN SMALL LETTER VOLAPUK UE
    Interval{ 0xA7A1, 0xA7A1 }, // L&       LATIN SMALL LETTER G WITH OBLIQUE STROKE
    Interval{ 0xA7A3, 0xA7A3 }, // L&       LATIN SMALL LETTER K WITH OBLIQUE STROKE
    Interval{ 0xA7A5, 0xA7A5 }, // L&       LATIN SMALL LETTER N WITH OBLIQUE STROKE
    Interval{ 0xA7A7, 0xA7A7 }, // L&       LATIN SMALL LETTER R WITH OBLIQUE STROKE
    Interval{ 0xA7A9, 0xA7A9 }, // L&       LATIN SMALL LETTER S WITH OBLIQUE STROKE
    Interval{ 0xA7B5, 0xA7B5 }, // L&       LATIN SMALL LETTER BETA
    Interval{ 0xA7B7, 0xA7B7 }, // L&       LATIN SMALL LETTER OMEGA
    Interval{ 0xA7B9, 0xA7B9 }, // L&       LATIN SMALL LETTER U WITH STROKE
    Interval{ 0xA7BB, 0xA7BB }, // L&       LATIN SMALL LETTER GLOTTAL A
    Interval{ 0xA7BD, 0xA7BD }, // L&       LATIN SMALL LETTER GLOTTAL I
    Interval{ 0xA7BF, 0xA7BF }, // L&       LATIN SMALL LETTER GLOTTAL U
    Interval{ 0xA7C1, 0xA7C1 }, // L&       LATIN SMALL LETTER OLD POLISH O
    Interval{ 0xA7C3, 0xA7C3 }, // L&       LATIN SMALL LETTER ANGLICANA W
    Interval{ 0xA7C8, 0xA7C8 }, // L&       LATIN SMALL LETTER D WITH SHORT STROKE OVERLAY
    Interval{ 0xA7CA, 0xA7CA }, // L&       LATIN SMALL LETTER S WITH SHORT STROKE OVERLAY
    Interval{ 0xA7CD, 0xA7CD }, // L&       LATIN SMALL LETTER S WITH DIAGONAL STROKE
    Interval{ 0xA7CF, 0xA7CF }, // L&       LATIN SMALL LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0xA7D1, 0xA7D1 }, // L&       LATIN SMALL LETTER CLOSED INSULAR G
    Interval{ 0xA7D3, 0xA7D3 }, // L&       LATIN SMALL LETTER DOUBLE THORN
    Interval{ 0xA7D5, 0xA7D5 }, // L&       LATIN SMALL LETTER DOUBLE WYNN
    Interval{ 0xA7D7, 0xA7D7 }, // L&       LATIN SMALL LETTER MIDDLE SCOTS S
    Interval{ 0xA7D9, 0xA7D9 }, // L&       LATIN SMALL LETTER SIGMOID S
    Interval{ 0xA7DB, 0xA7DB }, // L&       LATIN SMALL LETTER LAMBDA
    Interval{ 0xA7F6, 0xA7F6 }, // L&       LATIN SMALL LETTER REVERSED HALF H
    Interval{ 0xAB53, 0xAB53 }, // L&       LATIN SMALL LETTER CHI
    Interval{ 0xAB70, 0xABBF }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xFB00, 0xFB06 }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFF41, 0xFF5A }, // L&  [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Interval{ 0x10428, 0x1044F }, // L&  [40] DESERET SMALL LETTER LONG I..DESERET SMALL LETTER EW
    Interval{ 0x104D8, 0x104FB }, // L&  [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Interval{ 0x10597, 0x105A1 }, // L&  [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Interval{ 0x105A3, 0x105B1 }, // L&  [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Interval{ 0x105B3, 0x105B9 }, // L&   [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Interval{ 0x105BB, 0x105BC }, // L&   [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Interval{ 0x10CC0, 0x10CF2 }, // L&  [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Interval{ 0x10D70, 0x10D85 }, // L&  [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Interval{ 0x118C0, 0x118DF }, // L&  [32] WARANG CITI SMALL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Interval{ 0x16E60, 0x16E7F }, // L&  [32] MEDEFAIDRIN SMALL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Interval{ 0x16EBB, 0x16ED3 }, // L&  [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Interval{ 0x1E922, 0x1E943 }, // L&  [34] ADLAM SMALL LETTER ALIF..ADLAM SMALL LETTER SHA
}; // }}}
auto static const Default_Ignorable_Code_Point = std::array<Interval, 27>{ // {{{
    Interval{ 0x00AD, 0x00AD }, // Cf       SOFT HYPHEN
    Interval{ 0x034F, 0x034F }, // Mn       COMBINING GRAPHEME JOINER
    Interval{ 0x061C, 0x061C }, // Cf       ARABIC LETTER MARK
    Interval{ 0x115F, 0x1160 }, // Lo   [2] HANGUL CHOSEONG FILLER..HANGUL JUNGSEONG FILLER
    Interval{ 0x17B4, 0x17B5 }, // Mn   [2] KHMER VOWEL INHERENT AQ..KHMER VOWEL INHERENT AA
    Interval{ 0x180B, 0x180D }, // Mn   [3] MONGOLIAN FREE VARIATION SELECTOR ONE..MONGOLIAN FREE VARIATION SELECTOR THREE
    Interval{ 0x180E, 0x180E }, // Cf       MONGOLIAN VOWEL SEPARATOR
    Interval{ 0x180F, 0x180F }, // Mn       MONGOLIAN FREE VARIATION SELECTOR FOUR
    Interval{ 0x200B, 0x200F }, // Cf   [5] ZERO WIDTH SPACE..RIGHT-TO-LEFT MARK
    Interval{ 0x202A, 0x202E }, // Cf   [5] LEFT-TO-RIGHT EMBEDDING..RIGHT-TO-LEFT OVERRIDE
    Interval{ 0x2060, 0x2064 }, // Cf   [5] WORD JOINER..INVISIBLE PLUS
    Interval{ 0x2065, 0x2065 }, // Cn       <reserved-2065>
    Interval{ 0x2066, 0x206F }, // Cf  [10] LEFT-TO-RIGHT ISOLATE..NOMINAL DIGIT SHAPES
    Interval{ 0x3164, 0x3164 }, // Lo       HANGUL FILLER
    Interval{ 0xFE00, 0xFE0F }, // Mn  [16] VARIATION SELECTOR-1..VARIATION SELECTOR-16
    Interval{ 0xFEFF, 0xFEFF }, // Cf       ZERO WIDTH NO-BREAK SPACE
    Interval{ 0xFFA0, 0xFFA0 }, // Lo       HALFWIDTH HANGUL FILLER
    Interval{ 0xFFF0, 0xFFF8 }, // Cn   [9] <reserved-FFF0>..<reserved-FFF8>
    Interval{ 0x1BCA0, 0x1BCA3 }, // Cf   [4] SHORTHAND FORMAT LETTER OVERLAP..SHORTHAND FORMAT UP STEP
    Interval{ 0x1D173, 0x1D17A }, // Cf   [8] MUSICAL SYMBOL BEGIN BEAM..MUSICAL SYMBOL END PHRASE
    Interval{ 0xE0000, 0xE0000 }, // Cn       <reserved-E0000>
    Interval{ 0xE0001, 0xE0001 }, // Cf       LANGUAGE TAG
    Interval{ 0xE0002, 0xE001F }, // Cn  [30] <reserved-E0002>..<reserved-E001F>
    Interval{ 0xE0020, 0xE007F }, // Cf  [96] TAG SPACE..CANCEL TAG
    Interval{ 0xE0080, 0xE00FF }, // Cn [128] <reserved-E0080>..<reserved-E00FF>
    Interval{ 0xE0100, 0xE01EF }, // Mn [240] VARIATION SELECTOR-17..VARIATION SELECTOR-256
    Interval{ 0xE01F0, 0xE0FFF }, // Cn [3600] <reserved-E01F0>..<reserved-E0FFF>
}; // }}}
auto static const Grapheme_Base = std::array<Interval, 1799>{ // {{{
    Interval{ 0x0020, 0x0020 }, // Zs       SPACE
    Interval{ 0x0021, 0x0023 }, // Po   [3] EXCLAMATION MARK..NUMBER SIGN
    Interval{ 0x0024, 0x0024 }, // Sc       DOLLAR SIGN
    Interval{ 0x0025, 0x0027 }, // Po   [3] PERCENT SIGN..APOSTROPHE
    Interval{ 0x0028, 0x0028 }, // Ps       LEFT PARENTHESIS
    Interval{ 0x0029, 0x0029 }, // Pe       RIGHT PARENTHESIS
    Interval{ 0x002A, 0x002A }, // Po       ASTERISK
    Interval{ 0x002B, 0x002B }, // Sm       PLUS SIGN
    Interval{ 0x002C, 0x002C }, // Po       COMMA
    Interval{ 0x002D, 0x002D }, // Pd       HYPHEN-MINUS
    Interval{ 0x002E, 0x002F }, // Po   [2] FULL STOP..SOLIDUS
    Interval{ 0x0030, 0x0039 }, // Nd  [10] DIGIT ZERO..DIGIT NINE
    Interval{ 0x003A, 0x003B }, // Po   [2] COLON..SEMICOLON
    Interval{ 0x003C, 0x003E }, // Sm   [3] LESS-THAN SIGN..GREATER-THAN SIGN
    Interval{ 0x003F, 0x0040 }, // Po   [2] QUESTION MARK..COMMERCIAL AT
    Interval{ 0x0041, 0x005A }, // L&  [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Interval{ 0x005B, 0x005B }, // Ps       LEFT SQUARE BRACKET
    Interval{ 0x005C, 0x005C }, // Po       REVERSE SOLIDUS
    Interval{ 0x005D, 0x005D }, // Pe       RIGHT SQUARE BRACKET
    Interval{ 0x005E, 0x005E }, // Sk       CIRCUMFLEX ACCENT
    Interval{ 0x005F, 0x005F }, // Pc       LOW LINE
    Interval{ 0x0060, 0x0060 }, // Sk       GRAVE ACCENT
    Interval{ 0x0061, 0x007A }, // L&  [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Interval{ 0x007B, 0x007B }, // Ps       LEFT CURLY BRACKET
    Interval{ 0x007C, 0x007C }, // Sm       VERTICAL LINE
    Interval{ 0x007D, 0x007D }, // Pe       RIGHT CURLY BRACKET
    Interval{ 0x007E, 0x007E }, // Sm       TILDE
    Interval{ 0x00A0, 0x00A0 }, // Zs       NO-BREAK SPACE
    Interval{ 0x00A1, 0x00A1 }, // Po       INVERTED EXCLAMATION MARK
    Interval{ 0x00A2, 0x00A5 }, // Sc   [4] CENT SIGN..YEN SIGN
    Interval{ 0x00A6, 0x00A6 }, // So       BROKEN BAR
    Interval{ 0x00A7, 0x00A7 }, // Po       SECTION SIGN
    Interval{ 0x00A8, 0x00A8 }, // Sk       DIAERESIS
    Interval{ 0x00A9, 0x00A9 }, // So       COPYRIGHT SIGN
    Interval{ 0x00AA, 0x00AA }, // Lo       FEMININE ORDINAL INDICATOR
    Interval{ 0x00AB, 0x00AB }, // Pi       LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    Interval{ 0x00AC, 0x00AC }, // Sm       NOT SIGN
    Interval{ 0x00AE, 0x00AE }, // So       REGISTERED SIGN
    Interval{ 0x00AF, 0x00AF }, // Sk       MACRON
    Interval{ 0x00B0, 0x00B0 }, // So       DEGREE SIGN
    Interval{ 0x00B1, 0x00B1 }, // Sm       PLUS-MINUS SIGN
    Interval{ 0x00B2, 0x00B3 }, // No   [2] SUPERSCRIPT TWO..SUPERSCRIPT THREE
    Interval{ 0x00B4, 0x00B4 }, // Sk       ACUTE ACCENT
    Interval{ 0x00B5, 0x00B5 }, // L&       MICRO SIGN
    Interval{ 0x00B6, 0x00B7 }, // Po   [2] PILCROW SIGN..MIDDLE DOT
    Interval{ 0x00B8, 0x00B8 }, // Sk       CEDILLA
    Interval{ 0x00B9, 0x00B9 }, // No       SUPERSCRIPT ONE
    Interval{ 0x00BA, 0x00BA }, // Lo       MASCULINE ORDINAL INDICATOR
    Interval{ 0x00BB, 0x00BB }, // Pf       RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    Interval{ 0x00BC, 0x00BE }, // No   [3] VULGAR FRACTION ONE QUARTER..VULGAR FRACTION THREE QUARTERS
    Interval{ 0x00BF, 0x00BF }, // Po       INVERTED QUESTION MARK
    Interval{ 0x00C0, 0x00D6 }, // L&  [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x00D7, 0x00D7 }, // Sm       MULTIPLICATION SIGN
    Interval{ 0x00D8, 0x00F6 }, // L&  [31] LATIN CAPITAL LETTER O WITH STROKE..LATIN SMALL LETTER O WITH DIAERESIS
    Interval{ 0x00F7, 0x00F7 }, // Sm       DIVISION SIGN
    Interval{ 0x00F8, 0x01BA }, // L& [195] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER EZH WITH TAIL
    Interval{ 0x01BB, 0x01BB }, // Lo       LATIN LETTER TWO WITH STROKE
    Interval{ 0x01BC, 0x01BF }, // L&   [4] LATIN CAPITAL LETTER TONE FIVE..LATIN LETTER WYNN
    Interval{ 0x01C0, 0x01C3 }, // Lo   [4] LATIN LETTER DENTAL CLICK..LATIN LETTER RETROFLEX CLICK
    Interval{ 0x01C4, 0x0293 }, // L& [208] LATIN CAPITAL LETTER DZ WITH CARON..LATIN SMALL LETTER EZH WITH CURL
    Interval{ 0x0294, 0x0295 }, // Lo   [2] LATIN LETTER GLOTTAL STOP..LATIN LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0x0296, 0x02AF }, // L&  [26] LATIN LETTER INVERTED GLOTTAL STOP..LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    Interval{ 0x02B0, 0x02C1 }, // Lm  [18] MODIFIER LETTER SMALL H..MODIFIER LETTER REVERSED GLOTTAL STOP
    Interval{ 0x02C2, 0x02C5 }, // Sk   [4] MODIFIER LETTER LEFT ARROWHEAD..MODIFIER LETTER DOWN ARROWHEAD
    Interval{ 0x02C6, 0x02D1 }, // Lm  [12] MODIFIER LETTER CIRCUMFLEX ACCENT..MODIFIER LETTER HALF TRIANGULAR COLON
    Interval{ 0x02D2, 0x02DF }, // Sk  [14] MODIFIER LETTER CENTRED RIGHT HALF RING..MODIFIER LETTER CROSS ACCENT
    Interval{ 0x02E0, 0x02E4 }, // Lm   [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    Interval{ 0x02E5, 0x02EB }, // Sk   [7] MODIFIER LETTER EXTRA-HIGH TONE BAR..MODIFIER LETTER YANG DEPARTING TONE MARK
    Interval{ 0x02EC, 0x02EC }, // Lm       MODIFIER LETTER VOICING
    Interval{ 0x02ED, 0x02ED }, // Sk       MODIFIER LETTER UNASPIRATED
    Interval{ 0x02EE, 0x02EE }, // Lm       MODIFIER LETTER DOUBLE APOSTROPHE
    Interval{ 0x02EF, 0x02FF }, // Sk  [17] MODIFIER LETTER LOW DOWN ARROWHEAD..MODIFIER LETTER LOW LEFT ARROW
    Interval{ 0x0370, 0x0373 }, // L&   [4] GREEK CAPITAL LETTER HETA..GREEK SMALL LETTER ARCHAIC SAMPI
    Interval{ 0x0374, 0x0374 }, // Lm       GREEK NUMERAL SIGN
    Interval{ 0x0375, 0x0375 }, // Sk       GREEK LOWER NUMERAL SIGN
    Interval{ 0x0376, 0x0377 }, // L&   [2] GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA..GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037A, 0x037A }, // Lm       GREEK YPOGEGRAMMENI
    Interval{ 0x037B, 0x037D }, // L&   [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Interval{ 0x037E, 0x037E }, // Po       GREEK QUESTION MARK
    Interval{ 0x037F, 0x037F }, // L&       GREEK CAPITAL LETTER YOT
    Interval{ 0x0384, 0x0385 }, // Sk   [2] GREEK TONOS..GREEK DIALYTIKA TONOS
    Interval{ 0x0386, 0x0386 }, // L&       GREEK CAPITAL LETTER ALPHA WITH TONOS
    Interval{ 0x0387, 0x0387 }, // Po       GREEK ANO TELEIA
    Interval{ 0x0388, 0x038A }, // L&   [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Interval{ 0x038C, 0x038C }, // L&       GREEK CAPITAL LETTER OMICRON WITH TONOS
    Interval{ 0x038E, 0x03A1 }, // L&  [20] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER RHO
    Interval{ 0x03A3, 0x03F5 }, // L&  [83] GREEK CAPITAL LETTER SIGMA..GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F6, 0x03F6 }, // Sm       GREEK REVERSED LUNATE EPSILON SYMBOL
    Interval{ 0x03F7, 0x0481 }, // L& [139] GREEK CAPITAL LETTER SHO..CYRILLIC SMALL LETTER KOPPA
    Interval{ 0x0482, 0x0482 }, // So       CYRILLIC THOUSANDS SIGN
    Interval{ 0x048A, 0x052F }, // L& [166] CYRILLIC CAPITAL LETTER SHORT I WITH TAIL..CYRILLIC SMALL LETTER EL WITH DESCENDER
    Interval{ 0x0531, 0x0556 }, // L&  [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Interval{ 0x0559, 0x0559 }, // Lm       ARMENIAN MODIFIER LETTER LEFT HALF RING
    Interval{ 0x055A, 0x055F }, // Po   [6] ARMENIAN APOSTROPHE..ARMENIAN ABBREVIATION MARK
    Interval{ 0x0560, 0x0588 }, // L&  [41] ARMENIAN SMALL LETTER TURNED AYB..ARMENIAN SMALL LETTER YI WITH STROKE
    Interval{ 0x0589, 0x0589 }, // Po       ARMENIAN FULL STOP
    Interval{ 0x058A, 0x058A }, // Pd       ARMENIAN HYPHEN
    Interval{ 0x058D, 0x058E }, // So   [2] RIGHT-FACING ARMENIAN ETERNITY SIGN..LEFT-FACING ARMENIAN ETERNITY SIGN
    Interval{ 0x058F, 0x058F }, // Sc       ARMENIAN DRAM SIGN
    Interval{ 0x05BE, 0x05BE }, // Pd       HEBREW PUNCTUATION MAQAF
    Interval{ 0x05C0, 0x05C0 }, // Po       HEBREW PUNCTUATION PASEQ
    Interval{ 0x05C3, 0x05C3 }, // Po       HEBREW PUNCTUATION SOF PASUQ
    Interval{ 0x05C6, 0x05C6 }, // Po       HEBREW PUNCTUATION NUN HAFUKHA
    Interval{ 0x05D0, 0x05EA }, // Lo  [27] HEBREW LETTER ALEF..HEBREW LETTER TAV
    Interval{ 0x05EF, 0x05F2 }, // Lo   [4] HEBREW YOD TRIANGLE..HEBREW LIGATURE YIDDISH DOUBLE YOD
    Interval{ 0x05F3, 0x05F4 }, // Po   [2] HEBREW PUNCTUATION GERESH..HEBREW PUNCTUATION GERSHAYIM
    Interval{ 0x0606, 0x0608 }, // Sm   [3] ARABIC-INDIC CUBE ROOT..ARABIC RAY
    Interval{ 0x0609, 0x060A }, // Po   [2] ARABIC-INDIC PER MILLE SIGN..ARABIC-INDIC PER TEN THOUSAND SIGN
    Interval{ 0x060B, 0x060B }, // Sc       AFGHANI SIGN
    Interval{ 0x060C, 0x060D }, // Po   [2] ARABIC COMMA..ARABIC DATE SEPARATOR
    Interval{ 0x060E, 0x060F }, // So   [2] ARABIC POETIC VERSE SIGN..ARABIC SIGN MISRA
    Interval{ 0x061B, 0x061B }, // Po       ARABIC SEMICOLON
    Interval{ 0x061D, 0x061F }, // Po   [3] ARABIC END OF TEXT MARK..ARABIC QUESTION MARK
    Interval{ 0x0620, 0x063F }, // Lo  [32] ARABIC LETTER KASHMIRI YEH..ARABIC LETTER FARSI YEH WITH THREE DOTS ABOVE
    Interval{ 0x0640, 0x0640 }, // Lm       ARABIC TATWEEL
    Interval{ 0x0641, 0x064A }, // Lo  [10] ARABIC LETTER FEH..ARABIC LETTER YEH
    Interval{ 0x0660, 0x0669 }, // Nd  [10] ARABIC-INDIC DIGIT ZERO..ARABIC-INDIC DIGIT NINE
    Interval{ 0x066A, 0x066D }, // Po   [4] ARABIC PERCENT SIGN..ARABIC FIVE POINTED STAR
    Interval{ 0x066E, 0x066F }, // Lo   [2] ARABIC LETTER DOTLESS BEH..ARABIC LETTER DOTLESS QAF
    Interval{ 0x0671, 0x06D3 }, // Lo  [99] ARABIC LETTER ALEF WASLA..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE
    Interval{ 0x06D4, 0x06D4 }, // Po       ARABIC FULL STOP
    Interval{ 0x06D5, 0x06D5 }, // Lo       ARABIC LETTER AE
    Interval{ 0x06DE, 0x06DE }, // So       ARABIC START OF RUB EL HIZB
    Interval{ 0x06E5, 0x06E6 }, // Lm   [2] ARABIC SMALL WAW..ARABIC SMALL YEH
    Interval{ 0x06E9, 0x06E9 }, // So       ARABIC PLACE OF SAJDAH
    Interval{ 0x06EE, 0x06EF }, // Lo   [2] ARABIC LETTER DAL WITH INVERTED V..ARABIC LETTER REH WITH INVERTED V
    Interval{ 0x06F0, 0x06F9 }, // Nd  [10] EXTENDED ARABIC-INDIC DIGIT ZERO..EXTENDED ARABIC-INDIC DIGIT NINE
    Interval{ 0x06FA, 0x06FC }, // Lo   [3] ARABIC LETTER SHEEN WITH DOT BELOW..ARABIC LETTER GHAIN WITH DOT BELOW
    Interval{ 0x06FD, 0x06FE }, // So   [2] ARABIC SIGN SINDHI AMPERSAND..ARABIC SIGN SINDHI POSTPOSITION MEN
    Interval{ 0x06FF, 0x06FF }, // Lo       ARABIC LETTER HEH WITH INVERTED V
    Interval{ 0x0700, 0x070D }, // Po  [14] SYRIAC END OF PARAGRAPH..SYRIAC HARKLEAN ASTERISCUS
    Interval{ 0x0710, 0x0710 }, // Lo       SYRIAC LETTER ALAPH
    Interval{ 0x0712, 0x072F }, // Lo  [30] SYRIAC LETTER BETH..SYRIAC LETTER PERSIAN DHALATH
    Interval{ 0x074D, 0x07A5 }, // Lo  [89] SYRIAC LETTER SOGDIAN ZHAIN..THAANA LETTER WAAVU
    Interval{ 0x07B1, 0x07B1 }, // Lo       THAANA LETTER NAA
    Interval{ 0x07C0, 0x07C9 }, // Nd  [10] NKO DIGIT ZERO..NKO DIGIT NINE
    Interval{ 0x07CA, 0x07EA }, // Lo  [33] NKO LETTER A..NKO LETTER JONA RA
    Interval{ 0x07F4, 0x07F5 }, // Lm   [2] NKO HIGH TONE APOSTROPHE..NKO LOW TONE APOSTROPHE
    Interval{ 0x07F6, 0x07F6 }, // So       NKO SYMBOL OO DENNEN
    Interval{ 0x07F7, 0x07F9 }, // Po   [3] NKO SYMBOL GBAKURUNEN..NKO EXCLAMATION MARK
    Interval{ 0x07FA, 0x07FA }, // Lm       NKO LAJANYALAN
    Interval{ 0x07FE, 0x07FF }, // Sc   [2] NKO DOROME SIGN..NKO TAMAN SIGN
    Interval{ 0x0800, 0x0815 }, // Lo  [22] SAMARITAN LETTER ALAF..SAMARITAN LETTER TAAF
    Interval{ 0x081A, 0x081A }, // Lm       SAMARITAN MODIFIER LETTER EPENTHETIC YUT
    Interval{ 0x0824, 0x0824 }, // Lm       SAMARITAN MODIFIER LETTER SHORT A
    Interval{ 0x0828, 0x0828 }, // Lm       SAMARITAN MODIFIER LETTER I
    Interval{ 0x0830, 0x083E }, // Po  [15] SAMARITAN PUNCTUATION NEQUDAA..SAMARITAN PUNCTUATION ANNAAU
    Interval{ 0x0840, 0x0858 }, // Lo  [25] MANDAIC LETTER HALQA..MANDAIC LETTER AIN
    Interval{ 0x085E, 0x085E }, // Po       MANDAIC PUNCTUATION
    Interval{ 0x0860, 0x086A }, // Lo  [11] SYRIAC LETTER MALAYALAM NGA..SYRIAC LETTER MALAYALAM SSA
    Interval{ 0x0870, 0x0887 }, // Lo  [24] ARABIC LETTER ALEF WITH ATTACHED FATHA..ARABIC BASELINE ROUND DOT
    Interval{ 0x0888, 0x0888 }, // Sk       ARABIC RAISED ROUND DOT
    Interval{ 0x0889, 0x088F }, // Lo   [7] ARABIC LETTER NOON WITH INVERTED SMALL V..ARABIC LETTER NOON WITH RING ABOVE
    Interval{ 0x08A0, 0x08C8 }, // Lo  [41] ARABIC LETTER BEH WITH SMALL V BELOW..ARABIC LETTER GRAF
    Interval{ 0x08C9, 0x08C9 }, // Lm       ARABIC SMALL FARSI YEH
    Interval{ 0x0903, 0x0903 }, // Mc       DEVANAGARI SIGN VISARGA
    Interval{ 0x0904, 0x0939 }, // Lo  [54] DEVANAGARI LETTER SHORT A..DEVANAGARI LETTER HA
    Interval{ 0x093B, 0x093B }, // Mc       DEVANAGARI VOWEL SIGN OOE
    Interval{ 0x093D, 0x093D }, // Lo       DEVANAGARI SIGN AVAGRAHA
    Interval{ 0x093E, 0x0940 }, // Mc   [3] DEVANAGARI VOWEL SIGN AA..DEVANAGARI VOWEL SIGN II
    Interval{ 0x0949, 0x094C }, // Mc   [4] DEVANAGARI VOWEL SIGN CANDRA O..DEVANAGARI VOWEL SIGN AU
    Interval{ 0x094E, 0x094F }, // Mc   [2] DEVANAGARI VOWEL SIGN PRISHTHAMATRA E..DEVANAGARI VOWEL SIGN AW
    Interval{ 0x0950, 0x0950 }, // Lo       DEVANAGARI OM
    Interval{ 0x0958, 0x0961 }, // Lo  [10] DEVANAGARI LETTER QA..DEVANAGARI LETTER VOCALIC LL
    Interval{ 0x0964, 0x0965 }, // Po   [2] DEVANAGARI DANDA..DEVANAGARI DOUBLE DANDA
    Interval{ 0x0966, 0x096F }, // Nd  [10] DEVANAGARI DIGIT ZERO..DEVANAGARI DIGIT NINE
    Interval{ 0x0970, 0x0970 }, // Po       DEVANAGARI ABBREVIATION SIGN
    Interval{ 0x0971, 0x0971 }, // Lm       DEVANAGARI SIGN HIGH SPACING DOT
    Interval{ 0x0972, 0x0980 }, // Lo  [15] DEVANAGARI LETTER CANDRA A..BENGALI ANJI
    Interval{ 0x0982, 0x0983 }, // Mc   [2] BENGALI SIGN ANUSVARA..BENGALI SIGN VISARGA
    Interval{ 0x0985, 0x098C }, // Lo   [8] BENGALI LETTER A..BENGALI LETTER VOCALIC L
    Interval{ 0x098F, 0x0990 }, // Lo   [2] BENGALI LETTER E..BENGALI LETTER AI
    Interval{ 0x0993, 0x09A8 }, // Lo  [22] BENGALI LETTER O..BENGALI LETTER NA
    Interval{ 0x09AA, 0x09B0 }, // Lo   [7] BENGALI LETTER PA..BENGALI LETTER RA
    Interval{ 0x09B2, 0x09B2 }, // Lo       BENGALI LETTER LA
    Interval{ 0x09B6, 0x09B9 }, // Lo   [4] BENGALI LETTER SHA..BENGALI LETTER HA
    Interval{ 0x09BD, 0x09BD }, // Lo       BENGALI SIGN AVAGRAHA
    Interval{ 0x09BF, 0x09C0 }, // Mc   [2] BENGALI VOWEL SIGN I..BENGALI VOWEL SIGN II
    Interval{ 0x09C7, 0x09C8 }, // Mc   [2] BENGALI VOWEL SIGN E..BENGALI VOWEL SIGN AI
    Interval{ 0x09CB, 0x09CC }, // Mc   [2] BENGALI VOWEL SIGN O..BENGALI VOWEL SIGN AU
    Interval{ 0x09CE, 0x09CE }, // Lo       BENGALI LETTER KHANDA TA
    Interval{ 0x09DC, 0x09DD }, // Lo   [2] BENGALI LETTER RRA..BENGALI LETTER RHA
    Interval{ 0x09DF, 0x09E1 }, // Lo   [3] BENGALI LETTER YYA..BENGALI LETTER VOCALIC LL
    Interval{ 0x09E6, 0x09EF }, // Nd  [10] BENGALI DIGIT ZERO..BENGALI DIGIT NINE
    Interval{ 0x09F0, 0x09F1 }, // Lo   [2] BENGALI LETTER RA WITH MIDDLE DIAGONAL..BENGALI LETTER RA WITH LOWER DIAGONAL
    Interval{ 0x09F2, 0x09F3 }, // Sc   [2] BENGALI RUPEE MARK..BENGALI RUPEE SIGN
    Interval{ 0x09F4, 0x09F9 }, // No   [6] BENGALI CURRENCY NUMERATOR ONE..BENGALI CURRENCY DENOMINATOR SIXTEEN
    Interval{ 0x09FA, 0x09FA }, // So       BENGALI ISSHAR
    Interval{ 0x09FB, 0x09FB }, // Sc       BENGALI GANDA MARK
    Interval{ 0x09FC, 0x09FC }, // Lo       BENGALI LETTER VEDIC ANUSVARA
    Interval{ 0x09FD, 0x09FD }, // Po       BENGALI ABBREVIATION SIGN
    Interval{ 0x0A03, 0x0A03 }, // Mc       GURMUKHI SIGN VISARGA
    Interval{ 0x0A05, 0x0A0A }, // Lo   [6] GURMUKHI LETTER A..GURMUKHI LETTER UU
    Interval{ 0x0A0F, 0x0A10 }, // Lo   [2] GURMUKHI LETTER EE..GURMUKHI LETTER AI
    Interval{ 0x0A13, 0x0A28 }, // Lo  [22] GURMUKHI LETTER OO..GURMUKHI LETTER NA
    Interval{ 0x0A2A, 0x0A30 }, // Lo   [7] GURMUKHI LETTER PA..GURMUKHI LETTER RA
    Interval{ 0x0A32, 0x0A33 }, // Lo   [2] GURMUKHI LETTER LA..GURMUKHI LETTER LLA
    Interval{ 0x0A35, 0x0A36 }, // Lo   [2] GURMUKHI LETTER VA..GURMUKHI LETTER SHA
    Interval{ 0x0A38, 0x0A39 }, // Lo   [2] GURMUKHI LETTER SA..GURMUKHI LETTER HA
    Interval{ 0x0A3E, 0x0A40 }, // Mc   [3] GURMUKHI VOWEL SIGN AA..GURMUKHI VOWEL SIGN II
    Interval{ 0x0A59, 0x0A5C }, // Lo   [4] GURMUKHI LETTER KHHA..GURMUKHI LETTER RRA
    Interval{ 0x0A5E, 0x0A5E }, // Lo       GURMUKHI LETTER FA
    Interval{ 0x0A66, 0x0A6F }, // Nd  [10] GURMUKHI DIGIT ZERO..GURMUKHI DIGIT NINE
    Interval{ 0x0A72, 0x0A74 }, // Lo   [3] GURMUKHI IRI..GURMUKHI EK ONKAR
    Interval{ 0x0A76, 0x0A76 }, // Po       GURMUKHI ABBREVIATION SIGN
    Interval{ 0x0A83, 0x0A83 }, // Mc       GUJARATI SIGN VISARGA
    Interval{ 0x0A85, 0x0A8D }, // Lo   [9] GUJARATI LETTER A..GUJARATI VOWEL CANDRA E
    Interval{ 0x0A8F, 0x0A91 }, // Lo   [3] GUJARATI LETTER E..GUJARATI VOWEL CANDRA O
    Interval{ 0x0A93, 0x0AA8 }, // Lo  [22] GUJARATI LETTER O..GUJARATI LETTER NA
    Interval{ 0x0AAA, 0x0AB0 }, // Lo   [7] GUJARATI LETTER PA..GUJARATI LETTER RA
    Interval{ 0x0AB2, 0x0AB3 }, // Lo   [2] GUJARATI LETTER LA..GUJARATI LETTER LLA
    Interval{ 0x0AB5, 0x0AB9 }, // Lo   [5] GUJARATI LETTER VA..GUJARATI LETTER HA
    Interval{ 0x0ABD, 0x0ABD }, // Lo       GUJARATI SIGN AVAGRAHA
    Interval{ 0x0ABE, 0x0AC0 }, // Mc   [3] GUJARATI VOWEL SIGN AA..GUJARATI VOWEL SIGN II
    Interval{ 0x0AC9, 0x0AC9 }, // Mc       GUJARATI VOWEL SIGN CANDRA O
    Interval{ 0x0ACB, 0x0ACC }, // Mc   [2] GUJARATI VOWEL SIGN O..GUJARATI VOWEL SIGN AU
    Interval{ 0x0AD0, 0x0AD0 }, // Lo       GUJARATI OM
    Interval{ 0x0AE0, 0x0AE1 }, // Lo   [2] GUJARATI LETTER VOCALIC RR..GUJARATI LETTER VOCALIC LL
    Interval{ 0x0AE6, 0x0AEF }, // Nd  [10] GUJARATI DIGIT ZERO..GUJARATI DIGIT NINE
    Interval{ 0x0AF0, 0x0AF0 }, // Po       GUJARATI ABBREVIATION SIGN
    Interval{ 0x0AF1, 0x0AF1 }, // Sc       GUJARATI RUPEE SIGN
    Interval{ 0x0AF9, 0x0AF9 }, // Lo       GUJARATI LETTER ZHA
    Interval{ 0x0B02, 0x0B03 }, // Mc   [2] ORIYA SIGN ANUSVARA..ORIYA SIGN VISARGA
    Interval{ 0x0B05, 0x0B0C }, // Lo   [8] ORIYA LETTER A..ORIYA LETTER VOCALIC L
    Interval{ 0x0B0F, 0x0B10 }, // Lo   [2] ORIYA LETTER E..ORIYA LETTER AI
    Interval{ 0x0B13, 0x0B28 }, // Lo  [22] ORIYA LETTER O..ORIYA LETTER NA
    Interval{ 0x0B2A, 0x0B30 }, // Lo   [7] ORIYA LETTER PA..ORIYA LETTER RA
    Interval{ 0x0B32, 0x0B33 }, // Lo   [2] ORIYA LETTER LA..ORIYA LETTER LLA
    Interval{ 0x0B35, 0x0B39 }, // Lo   [5] ORIYA LETTER VA..ORIYA LETTER HA
    Interval{ 0x0B3D, 0x0B3D }, // Lo       ORIYA SIGN AVAGRAHA
    Interval{ 0x0B40, 0x0B40 }, // Mc       ORIYA VOWEL SIGN II
    Interval{ 0x0B47, 0x0B48 }, // Mc   [2] ORIYA VOWEL SIGN E..ORIYA VOWEL SIGN AI
    Interval{ 0x0B4B, 0x0B4C }, // Mc   [2] ORIYA VOWEL SIGN O..ORIYA VOWEL SIGN AU
    Interval{ 0x0B5C, 0x0B5D }, // Lo   [2] ORIYA LETTER RRA..ORIYA LETTER RHA
    Interval{ 0x0B5F, 0x0B61 }, // Lo   [3] ORIYA LETTER YYA..ORIYA LETTER VOCALIC LL
    Interval{ 0x0B66, 0x0B6F }, // Nd  [10] ORIYA DIGIT ZERO..ORIYA DIGIT NINE
    Interval{ 0x0B70, 0x0B70 }, // So       ORIYA ISSHAR
    Interval{ 0x0B71, 0x0B71 }, // Lo       ORIYA LETTER WA
    Interval{ 0x0B72, 0x0B77 }, // No   [6] ORIYA FRACTION ONE QUARTER..ORIYA FRACTION THREE SIXTEENTHS
    Interval{ 0x0B83, 0x0B83 }, // Lo       TAMIL SIGN VISARGA
    Interval{ 0x0B85, 0x0B8A }, // Lo   [6] TAMIL LETTER A..TAMIL LETTER UU
    Interval{ 0x0B8E, 0x0B90 }, // Lo   [3] TAMIL LETTER E..TAMIL LETTER AI
    Interval{ 0x0B92, 0x0B95 }, // Lo   [4] TAMIL LETTER O..TAMIL LETTER KA
    Interval{ 0x0B99, 0x0B9A }, // Lo   [2] TAMIL LETTER NGA..TAMIL LETTER CA
    Interval{ 0x0B9C, 0x0B9C }, // Lo       TAMIL LETTER JA
    Interval{ 0x0B9E, 0x0B9F }, // Lo   [2] TAMIL LETTER NYA..TAMIL LETTER TTA
    Interval{ 0x0BA3, 0x0BA4 }, // Lo   [2] TAMIL LETTER NNA..TAMIL LETTER TA
    Interval{ 0x0BA8, 0x0BAA }, // Lo   [3] TAMIL LETTER NA..TAMIL LETTER PA
    Interval{ 0x0BAE, 0x0BB9 }, // Lo  [12] TAMIL LETTER MA..TAMIL LETTER HA
    Interval{ 0x0BBF, 0x0BBF }, // Mc       TAMIL VOWEL SIGN I
    Interval{ 0x0BC1, 0x0BC2 }, // Mc   [2] TAMIL VOWEL SIGN U..TAMIL VOWEL SIGN UU
    Interval{ 0x0BC6, 0x0BC8 }, // Mc   [3] TAMIL VOWEL SIGN E..TAMIL VOWEL SIGN AI
    Interval{ 0x0BCA, 0x0BCC }, // Mc   [3] TAMIL VOWEL SIGN O..TAMIL VOWEL SIGN AU
    Interval{ 0x0BD0, 0x0BD0 }, // Lo       TAMIL OM
    Interval{ 0x0BE6, 0x0BEF }, // Nd  [10] TAMIL DIGIT ZERO..TAMIL DIGIT NINE
    Interval{ 0x0BF0, 0x0BF2 }, // No   [3] TAMIL NUMBER TEN..TAMIL NUMBER ONE THOUSAND
    Interval{ 0x0BF3, 0x0BF8 }, // So   [6] TAMIL DAY SIGN..TAMIL AS ABOVE SIGN
    Interval{ 0x0BF9, 0x0BF9 }, // Sc       TAMIL RUPEE SIGN
    Interval{ 0x0BFA, 0x0BFA }, // So       TAMIL NUMBER SIGN
    Interval{ 0x0C01, 0x0C03 }, // Mc   [3] TELUGU SIGN CANDRABINDU..TELUGU SIGN VISARGA
    Interval{ 0x0C05, 0x0C0C }, // Lo   [8] TELUGU LETTER A..TELUGU LETTER VOCALIC L
    Interval{ 0x0C0E, 0x0C10 }, // Lo   [3] TELUGU LETTER E..TELUGU LETTER AI
    Interval{ 0x0C12, 0x0C28 }, // Lo  [23] TELUGU LETTER O..TELUGU LETTER NA
    Interval{ 0x0C2A, 0x0C39 }, // Lo  [16] TELUGU LETTER PA..TELUGU LETTER HA
    Interval{ 0x0C3D, 0x0C3D }, // Lo       TELUGU SIGN AVAGRAHA
    Interval{ 0x0C41, 0x0C44 }, // Mc   [4] TELUGU VOWEL SIGN U..TELUGU VOWEL SIGN VOCALIC RR
    Interval{ 0x0C58, 0x0C5A }, // Lo   [3] TELUGU LETTER TSA..TELUGU LETTER RRRA
    Interval{ 0x0C5C, 0x0C5D }, // Lo   [2] TELUGU ARCHAIC SHRII..TELUGU LETTER NAKAARA POLLU
    Interval{ 0x0C60, 0x0C61 }, // Lo   [2] TELUGU LETTER VOCALIC RR..TELUGU LETTER VOCALIC LL
    Interval{ 0x0C66, 0x0C6F }, // Nd  [10] TELUGU DIGIT ZERO..TELUGU DIGIT NINE
    Interval{ 0x0C77, 0x0C77 }, // Po       TELUGU SIGN SIDDHAM
    Interval{ 0x0C78, 0x0C7E }, // No   [7] TELUGU FRACTION DIGIT ZERO FOR ODD POWERS OF FOUR..TELUGU FRACTION DIGIT THREE FOR EVEN POWERS OF FOUR
    Interval{ 0x0C7F, 0x0C7F }, // So       TELUGU SIGN TUUMU
    Interval{ 0x0C80, 0x0C80 }, // Lo       KANNADA SIGN SPACING CANDRABINDU
    Interval{ 0x0C82, 0x0C83 }, // Mc   [2] KANNADA SIGN ANUSVARA..KANNADA SIGN VISARGA
    Interval{ 0x0C84, 0x0C84 }, // Po       KANNADA SIGN SIDDHAM
    Interval{ 0x0C85, 0x0C8C }, // Lo   [8] KANNADA LETTER A..KANNADA LETTER VOCALIC L
    Interval{ 0x0C8E, 0x0C90 }, // Lo   [3] KANNADA LETTER E..KANNADA LETTER AI
    Interval{ 0x0C92, 0x0CA8 }, // Lo  [23] KANNADA LETTER O..KANNADA LETTER NA
    Interval{ 0x0CAA, 0x0CB3 }, // Lo  [10] KANNADA LETTER PA..KANNADA LETTER LLA
    Interval{ 0x0CB5, 0x0CB9 }, // Lo   [5] KANNADA LETTER VA..KANNADA LETTER HA
    Interval{ 0x0CBD, 0x0CBD }, // Lo       KANNADA SIGN AVAGRAHA
    Interval{ 0x0CBE, 0x0CBE }, // Mc       KANNADA VOWEL SIGN AA
    Interval{ 0x0CC1, 0x0CC1 }, // Mc       KANNADA VOWEL SIGN U
    Interval{ 0x0CC3, 0x0CC4 }, // Mc   [2] KANNADA VOWEL SIGN VOCALIC R..KANNADA VOWEL SIGN VOCALIC RR
    Interval{ 0x0CDC, 0x0CDE }, // Lo   [3] KANNADA ARCHAIC SHRII..KANNADA LETTER FA
    Interval{ 0x0CE0, 0x0CE1 }, // Lo   [2] KANNADA LETTER VOCALIC RR..KANNADA LETTER VOCALIC LL
    Interval{ 0x0CE6, 0x0CEF }, // Nd  [10] KANNADA DIGIT ZERO..KANNADA DIGIT NINE
    Interval{ 0x0CF1, 0x0CF2 }, // Lo   [2] KANNADA SIGN JIHVAMULIYA..KANNADA SIGN UPADHMANIYA
    Interval{ 0x0CF3, 0x0CF3 }, // Mc       KANNADA SIGN COMBINING ANUSVARA ABOVE RIGHT
    Interval{ 0x0D02, 0x0D03 }, // Mc   [2] MALAYALAM SIGN ANUSVARA..MALAYALAM SIGN VISARGA
    Interval{ 0x0D04, 0x0D0C }, // Lo   [9] MALAYALAM LETTER VEDIC ANUSVARA..MALAYALAM LETTER VOCALIC L
    Interval{ 0x0D0E, 0x0D10 }, // Lo   [3] MALAYALAM LETTER E..MALAYALAM LETTER AI
    Interval{ 0x0D12, 0x0D3A }, // Lo  [41] MALAYALAM LETTER O..MALAYALAM LETTER TTTA
    Interval{ 0x0D3D, 0x0D3D }, // Lo       MALAYALAM SIGN AVAGRAHA
    Interval{ 0x0D3F, 0x0D40 }, // Mc   [2] MALAYALAM VOWEL SIGN I..MALAYALAM VOWEL SIGN II
    Interval{ 0x0D46, 0x0D48 }, // Mc   [3] MALAYALAM VOWEL SIGN E..MALAYALAM VOWEL SIGN AI
    Interval{ 0x0D4A, 0x0D4C }, // Mc   [3] MALAYALAM VOWEL SIGN O..MALAYALAM VOWEL SIGN AU
    Interval{ 0x0D4E, 0x0D4E }, // Lo       MALAYALAM LETTER DOT REPH
    Interval{ 0x0D4F, 0x0D4F }, // So       MALAYALAM SIGN PARA
    Interval{ 0x0D54, 0x0D56 }, // Lo   [3] MALAYALAM LETTER CHILLU M..MALAYALAM LETTER CHILLU LLL
    Interval{ 0x0D58, 0x0D5E }, // No   [7] MALAYALAM FRACTION ONE ONE-HUNDRED-AND-SIXTIETH..MALAYALAM FRACTION ONE FIFTH
    Interval{ 0x0D5F, 0x0D61 }, // Lo   [3] MALAYALAM LETTER ARCHAIC II..MALAYALAM LETTER VOCALIC LL
    Interval{ 0x0D66, 0x0D6F }, // Nd  [10] MALAYALAM DIGIT ZERO..MALAYALAM DIGIT NINE
    Interval{ 0x0D70, 0x0D78 }, // No   [9] MALAYALAM NUMBER TEN..MALAYALAM FRACTION THREE SIXTEENTHS
    Interval{ 0x0D79, 0x0D79 }, // So       MALAYALAM DATE MARK
    Interval{ 0x0D7A, 0x0D7F }, // Lo   [6] MALAYALAM LETTER CHILLU NN..MALAYALAM LETTER CHILLU K
    Interval{ 0x0D82, 0x0D83 }, // Mc   [2] SINHALA SIGN ANUSVARAYA..SINHALA SIGN VISARGAYA
    Interval{ 0x0D85, 0x0D96 }, // Lo  [18] SINHALA LETTER AYANNA..SINHALA LETTER AUYANNA
    Interval{ 0x0D9A, 0x0DB1 }, // Lo  [24] SINHALA LETTER ALPAPRAANA KAYANNA..SINHALA LETTER DANTAJA NAYANNA
    Interval{ 0x0DB3, 0x0DBB }, // Lo   [9] SINHALA LETTER SANYAKA DAYANNA..SINHALA LETTER RAYANNA
    Interval{ 0x0DBD, 0x0DBD }, // Lo       SINHALA LETTER DANTAJA LAYANNA
    Interval{ 0x0DC0, 0x0DC6 }, // Lo   [7] SINHALA LETTER VAYANNA..SINHALA LETTER FAYANNA
    Interval{ 0x0DD0, 0x0DD1 }, // Mc   [2] SINHALA VOWEL SIGN KETTI AEDA-PILLA..SINHALA VOWEL SIGN DIGA AEDA-PILLA
    Interval{ 0x0DD8, 0x0DDE }, // Mc   [7] SINHALA VOWEL SIGN GAETTA-PILLA..SINHALA VOWEL SIGN KOMBUVA HAA GAYANUKITTA
    Interval{ 0x0DE6, 0x0DEF }, // Nd  [10] SINHALA LITH DIGIT ZERO..SINHALA LITH DIGIT NINE
    Interval{ 0x0DF2, 0x0DF3 }, // Mc   [2] SINHALA VOWEL SIGN DIGA GAETTA-PILLA..SINHALA VOWEL SIGN DIGA GAYANUKITTA
    Interval{ 0x0DF4, 0x0DF4 }, // Po       SINHALA PUNCTUATION KUNDDALIYA
    Interval{ 0x0E01, 0x0E30 }, // Lo  [48] THAI CHARACTER KO KAI..THAI CHARACTER SARA A
    Interval{ 0x0E32, 0x0E33 }, // Lo   [2] THAI CHARACTER SARA AA..THAI CHARACTER SARA AM
    Interval{ 0x0E3F, 0x0E3F }, // Sc       THAI CURRENCY SYMBOL BAHT
    Interval{ 0x0E40, 0x0E45 }, // Lo   [6] THAI CHARACTER SARA E..THAI CHARACTER LAKKHANGYAO
    Interval{ 0x0E46, 0x0E46 }, // Lm       THAI CHARACTER MAIYAMOK
    Interval{ 0x0E4F, 0x0E4F }, // Po       THAI CHARACTER FONGMAN
    Interval{ 0x0E50, 0x0E59 }, // Nd  [10] THAI DIGIT ZERO..THAI DIGIT NINE
    Interval{ 0x0E5A, 0x0E5B }, // Po   [2] THAI CHARACTER ANGKHANKHU..THAI CHARACTER KHOMUT
    Interval{ 0x0E81, 0x0E82 }, // Lo   [2] LAO LETTER KO..LAO LETTER KHO SUNG
    Interval{ 0x0E84, 0x0E84 }, // Lo       LAO LETTER KHO TAM
    Interval{ 0x0E86, 0x0E8A }, // Lo   [5] LAO LETTER PALI GHA..LAO LETTER SO TAM
    Interval{ 0x0E8C, 0x0EA3 }, // Lo  [24] LAO LETTER PALI JHA..LAO LETTER LO LING
    Interval{ 0x0EA5, 0x0EA5 }, // Lo       LAO LETTER LO LOOT
    Interval{ 0x0EA7, 0x0EB0 }, // Lo  [10] LAO LETTER WO..LAO VOWEL SIGN A
    Interval{ 0x0EB2, 0x0EB3 }, // Lo   [2] LAO VOWEL SIGN AA..LAO VOWEL SIGN AM
    Interval{ 0x0EBD, 0x0EBD }, // Lo       LAO SEMIVOWEL SIGN NYO
    Interval{ 0x0EC0, 0x0EC4 }, // Lo   [5] LAO VOWEL SIGN E..LAO VOWEL SIGN AI
    Interval{ 0x0EC6, 0x0EC6 }, // Lm       LAO KO LA
    Interval{ 0x0ED0, 0x0ED9 }, // Nd  [10] LAO DIGIT ZERO..LAO DIGIT NINE
    Interval{ 0x0EDC, 0x0EDF }, // Lo   [4] LAO HO NO..LAO LETTER KHMU NYO
    Interval{ 0x0F00, 0x0F00 }, // Lo       TIBETAN SYLLABLE OM
    Interval{ 0x0F01, 0x0F03 }, // So   [3] TIBETAN MARK GTER YIG MGO TRUNCATED A..TIBETAN MARK GTER YIG MGO -UM GTER TSHEG MA
    Interval{ 0x0F04, 0x0F12 }, // Po  [15] TIBETAN MARK INITIAL YIG MGO MDUN MA..TIBETAN MARK RGYA GRAM SHAD
    Interval{ 0x0F13, 0x0F13 }, // So       TIBETAN MARK CARET -DZUD RTAGS ME LONG CAN
    Interval{ 0x0F14, 0x0F14 }, // Po       TIBETAN MARK GTER TSHEG
    Interval{ 0x0F15, 0x0F17 }, // So   [3] TIBETAN LOGOTYPE SIGN CHAD RTAGS..TIBETAN ASTROLOGICAL SIGN SGRA GCAN -CHAR RTAGS
    Interval{ 0x0F1A, 0x0F1F }, // So   [6] TIBETAN SIGN RDEL DKAR GCIG..TIBETAN SIGN RDEL DKAR RDEL NAG
    Interval{ 0x0F20, 0x0F29 }, // Nd  [10] TIBETAN DIGIT ZERO..TIBETAN DIGIT NINE
    Interval{ 0x0F2A, 0x0F33 }, // No  [10] TIBETAN DIGIT HALF ONE..TIBETAN DIGIT HALF ZERO
    Interval{ 0x0F34, 0x0F34 }, // So       TIBETAN MARK BSDUS RTAGS
    Interval{ 0x0F36, 0x0F36 }, // So       TIBETAN MARK CARET -DZUD RTAGS BZHI MIG CAN
    Interval{ 0x0F38, 0x0F38 }, // So       TIBETAN MARK CHE MGO
    Interval{ 0x0F3A, 0x0F3A }, // Ps       TIBETAN MARK GUG RTAGS GYON
    Interval{ 0x0F3B, 0x0F3B }, // Pe       TIBETAN MARK GUG RTAGS GYAS
    Interval{ 0x0F3C, 0x0F3C }, // Ps       TIBETAN MARK ANG KHANG GYON
    Interval{ 0x0F3D, 0x0F3D }, // Pe       TIBETAN MARK ANG KHANG GYAS
    Interval{ 0x0F3E, 0x0F3F }, // Mc   [2] TIBETAN SIGN YAR TSHES..TIBETAN SIGN MAR TSHES
    Interval{ 0x0F40, 0x0F47 }, // Lo   [8] TIBETAN LETTER KA..TIBETAN LETTER JA
    Interval{ 0x0F49, 0x0F6C }, // Lo  [36] TIBETAN LETTER NYA..TIBETAN LETTER RRA
    Interval{ 0x0F7F, 0x0F7F }, // Mc       TIBETAN SIGN RNAM BCAD
    Interval{ 0x0F85, 0x0F85 }, // Po       TIBETAN MARK PALUTA
    Interval{ 0x0F88, 0x0F8C }, // Lo   [5] TIBETAN SIGN LCE TSA CAN..TIBETAN SIGN INVERTED MCHU CAN
    Interval{ 0x0FBE, 0x0FC5 }, // So   [8] TIBETAN KU RU KHA..TIBETAN SYMBOL RDO RJE
    Interval{ 0x0FC7, 0x0FCC }, // So   [6] TIBETAN SYMBOL RDO RJE RGYA GRAM..TIBETAN SYMBOL NOR BU BZHI -KHYIL
    Interval{ 0x0FCE, 0x0FCF }, // So   [2] TIBETAN SIGN RDEL NAG RDEL DKAR..TIBETAN SIGN RDEL NAG GSUM
    Interval{ 0x0FD0, 0x0FD4 }, // Po   [5] TIBETAN MARK BSKA- SHOG GI MGO RGYAN..TIBETAN MARK CLOSING BRDA RNYING YIG MGO SGAB MA
    Interval{ 0x0FD5, 0x0FD8 }, // So   [4] RIGHT-FACING SVASTI SIGN..LEFT-FACING SVASTI SIGN WITH DOTS
    Interval{ 0x0FD9, 0x0FDA }, // Po   [2] TIBETAN MARK LEADING MCHAN RTAGS..TIBETAN MARK TRAILING MCHAN RTAGS
    Interval{ 0x1000, 0x102A }, // Lo  [43] MYANMAR LETTER KA..MYANMAR LETTER AU
    Interval{ 0x102B, 0x102C }, // Mc   [2] MYANMAR VOWEL SIGN TALL AA..MYANMAR VOWEL SIGN AA
    Interval{ 0x1031, 0x1031 }, // Mc       MYANMAR VOWEL SIGN E
    Interval{ 0x1038, 0x1038 }, // Mc       MYANMAR SIGN VISARGA
    Interval{ 0x103B, 0x103C }, // Mc   [2] MYANMAR CONSONANT SIGN MEDIAL YA..MYANMAR CONSONANT SIGN MEDIAL RA
    Interval{ 0x103F, 0x103F }, // Lo       MYANMAR LETTER GREAT SA
    Interval{ 0x1040, 0x1049 }, // Nd  [10] MYANMAR DIGIT ZERO..MYANMAR DIGIT NINE
    Interval{ 0x104A, 0x104F }, // Po   [6] MYANMAR SIGN LITTLE SECTION..MYANMAR SYMBOL GENITIVE
    Interval{ 0x1050, 0x1055 }, // Lo   [6] MYANMAR LETTER SHA..MYANMAR LETTER VOCALIC LL
    Interval{ 0x1056, 0x1057 }, // Mc   [2] MYANMAR VOWEL SIGN VOCALIC R..MYANMAR VOWEL SIGN VOCALIC RR
    Interval{ 0x105A, 0x105D }, // Lo   [4] MYANMAR LETTER MON NGA..MYANMAR LETTER MON BBE
    Interval{ 0x1061, 0x1061 }, // Lo       MYANMAR LETTER SGAW KAREN SHA
    Interval{ 0x1062, 0x1064 }, // Mc   [3] MYANMAR VOWEL SIGN SGAW KAREN EU..MYANMAR TONE MARK SGAW KAREN KE PHO
    Interval{ 0x1065, 0x1066 }, // Lo   [2] MYANMAR LETTER WESTERN PWO KAREN THA..MYANMAR LETTER WESTERN PWO KAREN PWA
    Interval{ 0x1067, 0x106D }, // Mc   [7] MYANMAR VOWEL SIGN WESTERN PWO KAREN EU..MYANMAR SIGN WESTERN PWO KAREN TONE-5
    Interval{ 0x106E, 0x1070 }, // Lo   [3] MYANMAR LETTER EASTERN PWO KAREN NNA..MYANMAR LETTER EASTERN PWO KAREN GHWA
    Interval{ 0x1075, 0x1081 }, // Lo  [13] MYANMAR LETTER SHAN KA..MYANMAR LETTER SHAN HA
    Interval{ 0x1083, 0x1084 }, // Mc   [2] MYANMAR VOWEL SIGN SHAN AA..MYANMAR VOWEL SIGN SHAN E
    Interval{ 0x1087, 0x108C }, // Mc   [6] MYANMAR SIGN SHAN TONE-2..MYANMAR SIGN SHAN COUNCIL TONE-3
    Interval{ 0x108E, 0x108E }, // Lo       MYANMAR LETTER RUMAI PALAUNG FA
    Interval{ 0x108F, 0x108F }, // Mc       MYANMAR SIGN RUMAI PALAUNG TONE-5
    Interval{ 0x1090, 0x1099 }, // Nd  [10] MYANMAR SHAN DIGIT ZERO..MYANMAR SHAN DIGIT NINE
    Interval{ 0x109A, 0x109C }, // Mc   [3] MYANMAR SIGN KHAMTI TONE-1..MYANMAR VOWEL SIGN AITON A
    Interval{ 0x109E, 0x109F }, // So   [2] MYANMAR SYMBOL SHAN ONE..MYANMAR SYMBOL SHAN EXCLAMATION
    Interval{ 0x10A0, 0x10C5 }, // L&  [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Interval{ 0x10C7, 0x10C7 }, // L&       GEORGIAN CAPITAL LETTER YN
    Interval{ 0x10CD, 0x10CD }, // L&       GEORGIAN CAPITAL LETTER AEN
    Interval{ 0x10D0, 0x10FA }, // L&  [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Interval{ 0x10FB, 0x10FB }, // Po       GEORGIAN PARAGRAPH SEPARATOR
    Interval{ 0x10FC, 0x10FC }, // Lm       MODIFIER LETTER GEORGIAN NAR
    Interval{ 0x10FD, 0x10FF }, // L&   [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Interval{ 0x1100, 0x1248 }, // Lo [329] HANGUL CHOSEONG KIYEOK..ETHIOPIC SYLLABLE QWA
    Interval{ 0x124A, 0x124D }, // Lo   [4] ETHIOPIC SYLLABLE QWI..ETHIOPIC SYLLABLE QWE
    Interval{ 0x1250, 0x1256 }, // Lo   [7] ETHIOPIC SYLLABLE QHA..ETHIOPIC SYLLABLE QHO
    Interval{ 0x1258, 0x1258 }, // Lo       ETHIOPIC SYLLABLE QHWA
    Interval{ 0x125A, 0x125D }, // Lo   [4] ETHIOPIC SYLLABLE QHWI..ETHIOPIC SYLLABLE QHWE
    Interval{ 0x1260, 0x1288 }, // Lo  [41] ETHIOPIC SYLLABLE BA..ETHIOPIC SYLLABLE XWA
    Interval{ 0x128A, 0x128D }, // Lo   [4] ETHIOPIC SYLLABLE XWI..ETHIOPIC SYLLABLE XWE
    Interval{ 0x1290, 0x12B0 }, // Lo  [33] ETHIOPIC SYLLABLE NA..ETHIOPIC SYLLABLE KWA
    Interval{ 0x12B2, 0x12B5 }, // Lo   [4] ETHIOPIC SYLLABLE KWI..ETHIOPIC SYLLABLE KWE
    Interval{ 0x12B8, 0x12BE }, // Lo   [7] ETHIOPIC SYLLABLE KXA..ETHIOPIC SYLLABLE KXO
    Interval{ 0x12C0, 0x12C0 }, // Lo       ETHIOPIC SYLLABLE KXWA
    Interval{ 0x12C2, 0x12C5 }, // Lo   [4] ETHIOPIC SYLLABLE KXWI..ETHIOPIC SYLLABLE KXWE
    Interval{ 0x12C8, 0x12D6 }, // Lo  [15] ETHIOPIC SYLLABLE WA..ETHIOPIC SYLLABLE PHARYNGEAL O
    Interval{ 0x12D8, 0x1310 }, // Lo  [57] ETHIOPIC SYLLABLE ZA..ETHIOPIC SYLLABLE GWA
    Interval{ 0x1312, 0x1315 }, // Lo   [4] ETHIOPIC SYLLABLE GWI..ETHIOPIC SYLLABLE GWE
    Interval{ 0x1318, 0x135A }, // Lo  [67] ETHIOPIC SYLLABLE GGA..ETHIOPIC SYLLABLE FYA
    Interval{ 0x1360, 0x1368 }, // Po   [9] ETHIOPIC SECTION MARK..ETHIOPIC PARAGRAPH SEPARATOR
    Interval{ 0x1369, 0x137C }, // No  [20] ETHIOPIC DIGIT ONE..ETHIOPIC NUMBER TEN THOUSAND
    Interval{ 0x1380, 0x138F }, // Lo  [16] ETHIOPIC SYLLABLE SEBATBEIT MWA..ETHIOPIC SYLLABLE PWE
    Interval{ 0x1390, 0x1399 }, // So  [10] ETHIOPIC TONAL MARK YIZET..ETHIOPIC TONAL MARK KURT
    Interval{ 0x13A0, 0x13F5 }, // L&  [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Interval{ 0x13F8, 0x13FD }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1400, 0x1400 }, // Pd       CANADIAN SYLLABICS HYPHEN
    Interval{ 0x1401, 0x166C }, // Lo [620] CANADIAN SYLLABICS E..CANADIAN SYLLABICS CARRIER TTSA
    Interval{ 0x166D, 0x166D }, // So       CANADIAN SYLLABICS CHI SIGN
    Interval{ 0x166E, 0x166E }, // Po       CANADIAN SYLLABICS FULL STOP
    Interval{ 0x166F, 0x167F }, // Lo  [17] CANADIAN SYLLABICS QAI..CANADIAN SYLLABICS BLACKFOOT W
    Interval{ 0x1680, 0x1680 }, // Zs       OGHAM SPACE MARK
    Interval{ 0x1681, 0x169A }, // Lo  [26] OGHAM LETTER BEITH..OGHAM LETTER PEITH
    Interval{ 0x169B, 0x169B }, // Ps       OGHAM FEATHER MARK
    Interval{ 0x169C, 0x169C }, // Pe       OGHAM REVERSED FEATHER MARK
    Interval{ 0x16A0, 0x16EA }, // Lo  [75] RUNIC LETTER FEHU FEOH FE F..RUNIC LETTER X
    Interval{ 0x16EB, 0x16ED }, // Po   [3] RUNIC SINGLE PUNCTUATION..RUNIC CROSS PUNCTUATION
    Interval{ 0x16EE, 0x16F0 }, // Nl   [3] RUNIC ARLAUG SYMBOL..RUNIC BELGTHOR SYMBOL
    Interval{ 0x16F1, 0x16F8 }, // Lo   [8] RUNIC LETTER K..RUNIC LETTER FRANKS CASKET AESC
    Interval{ 0x1700, 0x1711 }, // Lo  [18] TAGALOG LETTER A..TAGALOG LETTER HA
    Interval{ 0x171F, 0x1731 }, // Lo  [19] TAGALOG LETTER ARCHAIC RA..HANUNOO LETTER HA
    Interval{ 0x1735, 0x1736 }, // Po   [2] PHILIPPINE SINGLE PUNCTUATION..PHILIPPINE DOUBLE PUNCTUATION
    Interval{ 0x1740, 0x1751 }, // Lo  [18] BUHID LETTER A..BUHID LETTER HA
    Interval{ 0x1760, 0x176C }, // Lo  [13] TAGBANWA LETTER A..TAGBANWA LETTER YA
    Interval{ 0x176E, 0x1770 }, // Lo   [3] TAGBANWA LETTER LA..TAGBANWA LETTER SA
    Interval{ 0x1780, 0x17B3 }, // Lo  [52] KHMER LETTER KA..KHMER INDEPENDENT VOWEL QAU
    Interval{ 0x17B6, 0x17B6 }, // Mc       KHMER VOWEL SIGN AA
    Interval{ 0x17BE, 0x17C5 }, // Mc   [8] KHMER VOWEL SIGN OE..KHMER VOWEL SIGN AU
    Interval{ 0x17C7, 0x17C8 }, // Mc   [2] KHMER SIGN REAHMUK..KHMER SIGN YUUKALEAPINTU
    Interval{ 0x17D4, 0x17D6 }, // Po   [3] KHMER SIGN KHAN..KHMER SIGN CAMNUC PII KUUH
    Interval{ 0x17D7, 0x17D7 }, // Lm       KHMER SIGN LEK TOO
    Interval{ 0x17D8, 0x17DA }, // Po   [3] KHMER SIGN BEYYAL..KHMER SIGN KOOMUUT
    Interval{ 0x17DB, 0x17DB }, // Sc       KHMER CURRENCY SYMBOL RIEL
    Interval{ 0x17DC, 0x17DC }, // Lo       KHMER SIGN AVAKRAHASANYA
    Interval{ 0x17E0, 0x17E9 }, // Nd  [10] KHMER DIGIT ZERO..KHMER DIGIT NINE
    Interval{ 0x17F0, 0x17F9 }, // No  [10] KHMER SYMBOL LEK ATTAK SON..KHMER SYMBOL LEK ATTAK PRAM-BUON
    Interval{ 0x1800, 0x1805 }, // Po   [6] MONGOLIAN BIRGA..MONGOLIAN FOUR DOTS
    Interval{ 0x1806, 0x1806 }, // Pd       MONGOLIAN TODO SOFT HYPHEN
    Interval{ 0x1807, 0x180A }, // Po   [4] MONGOLIAN SIBE SYLLABLE BOUNDARY MARKER..MONGOLIAN NIRUGU
    Interval{ 0x1810, 0x1819 }, // Nd  [10] MONGOLIAN DIGIT ZERO..MONGOLIAN DIGIT NINE
    Interval{ 0x1820, 0x1842 }, // Lo  [35] MONGOLIAN LETTER A..MONGOLIAN LETTER CHI
    Interval{ 0x1843, 0x1843 }, // Lm       MONGOLIAN LETTER TODO LONG VOWEL SIGN
    Interval{ 0x1844, 0x1878 }, // Lo  [53] MONGOLIAN LETTER TODO E..MONGOLIAN LETTER CHA WITH TWO DOTS
    Interval{ 0x1880, 0x1884 }, // Lo   [5] MONGOLIAN LETTER ALI GALI ANUSVARA ONE..MONGOLIAN LETTER ALI GALI INVERTED UBADAMA
    Interval{ 0x1887, 0x18A8 }, // Lo  [34] MONGOLIAN LETTER ALI GALI A..MONGOLIAN LETTER MANCHU ALI GALI BHA
    Interval{ 0x18AA, 0x18AA }, // Lo       MONGOLIAN LETTER MANCHU ALI GALI LHA
    Interval{ 0x18B0, 0x18F5 }, // Lo  [70] CANADIAN SYLLABICS OY..CANADIAN SYLLABICS CARRIER DENTAL S
    Interval{ 0x1900, 0x191E }, // Lo  [31] LIMBU VOWEL-CARRIER LETTER..LIMBU LETTER TRA
    Interval{ 0x1923, 0x1926 }, // Mc   [4] LIMBU VOWEL SIGN EE..LIMBU VOWEL SIGN AU
    Interval{ 0x1929, 0x192B }, // Mc   [3] LIMBU SUBJOINED LETTER YA..LIMBU SUBJOINED LETTER WA
    Interval{ 0x1930, 0x1931 }, // Mc   [2] LIMBU SMALL LETTER KA..LIMBU SMALL LETTER NGA
    Interval{ 0x1933, 0x1938 }, // Mc   [6] LIMBU SMALL LETTER TA..LIMBU SMALL LETTER LA
    Interval{ 0x1940, 0x1940 }, // So       LIMBU SIGN LOO
    Interval{ 0x1944, 0x1945 }, // Po   [2] LIMBU EXCLAMATION MARK..LIMBU QUESTION MARK
    Interval{ 0x1946, 0x194F }, // Nd  [10] LIMBU DIGIT ZERO..LIMBU DIGIT NINE
    Interval{ 0x1950, 0x196D }, // Lo  [30] TAI LE LETTER KA..TAI LE LETTER AI
    Interval{ 0x1970, 0x1974 }, // Lo   [5] TAI LE LETTER TONE-2..TAI LE LETTER TONE-6
    Interval{ 0x1980, 0x19AB }, // Lo  [44] NEW TAI LUE LETTER HIGH QA..NEW TAI LUE LETTER LOW SUA
    Interval{ 0x19B0, 0x19C9 }, // Lo  [26] NEW TAI LUE VOWEL SIGN VOWEL SHORTENER..NEW TAI LUE TONE MARK-2
    Interval{ 0x19D0, 0x19D9 }, // Nd  [10] NEW TAI LUE DIGIT ZERO..NEW TAI LUE DIGIT NINE
    Interval{ 0x19DA, 0x19DA }, // No       NEW TAI LUE THAM DIGIT ONE
    Interval{ 0x19DE, 0x19FF }, // So  [34] NEW TAI LUE SIGN LAE..KHMER SYMBOL DAP-PRAM ROC
    Interval{ 0x1A00, 0x1A16 }, // Lo  [23] BUGINESE LETTER KA..BUGINESE LETTER HA
    Interval{ 0x1A19, 0x1A1A }, // Mc   [2] BUGINESE VOWEL SIGN E..BUGINESE VOWEL SIGN O
    Interval{ 0x1A1E, 0x1A1F }, // Po   [2] BUGINESE PALLAWA..BUGINESE END OF SECTION
    Interval{ 0x1A20, 0x1A54 }, // Lo  [53] TAI THAM LETTER HIGH KA..TAI THAM LETTER GREAT SA
    Interval{ 0x1A55, 0x1A55 }, // Mc       TAI THAM CONSONANT SIGN MEDIAL RA
    Interval{ 0x1A57, 0x1A57 }, // Mc       TAI THAM CONSONANT SIGN LA TANG LAI
    Interval{ 0x1A61, 0x1A61 }, // Mc       TAI THAM VOWEL SIGN A
    Interval{ 0x1A63, 0x1A64 }, // Mc   [2] TAI THAM VOWEL SIGN AA..TAI THAM VOWEL SIGN TALL AA
    Interval{ 0x1A6D, 0x1A72 }, // Mc   [6] TAI THAM VOWEL SIGN OY..TAI THAM VOWEL SIGN THAM AI
    Interval{ 0x1A80, 0x1A89 }, // Nd  [10] TAI THAM HORA DIGIT ZERO..TAI THAM HORA DIGIT NINE
    Interval{ 0x1A90, 0x1A99 }, // Nd  [10] TAI THAM THAM DIGIT ZERO..TAI THAM THAM DIGIT NINE
    Interval{ 0x1AA0, 0x1AA6 }, // Po   [7] TAI THAM SIGN WIANG..TAI THAM SIGN REVERSED ROTATED RANA
    Interval{ 0x1AA7, 0x1AA7 }, // Lm       TAI THAM SIGN MAI YAMOK
    Interval{ 0x1AA8, 0x1AAD }, // Po   [6] TAI THAM SIGN KAAN..TAI THAM SIGN CAANG
    Interval{ 0x1B04, 0x1B04 }, // Mc       BALINESE SIGN BISAH
    Interval{ 0x1B05, 0x1B33 }, // Lo  [47] BALINESE LETTER AKARA..BALINESE LETTER HA
    Interval{ 0x1B3E, 0x1B41 }, // Mc   [4] BALINESE VOWEL SIGN TALING..BALINESE VOWEL SIGN TALING REPA TEDUNG
    Interval{ 0x1B45, 0x1B4C }, // Lo   [8] BALINESE LETTER KAF SASAK..BALINESE LETTER ARCHAIC JNYA
    Interval{ 0x1B4E, 0x1B4F }, // Po   [2] BALINESE INVERTED CARIK SIKI..BALINESE INVERTED CARIK PAREREN
    Interval{ 0x1B50, 0x1B59 }, // Nd  [10] BALINESE DIGIT ZERO..BALINESE DIGIT NINE
    Interval{ 0x1B5A, 0x1B60 }, // Po   [7] BALINESE PANTI..BALINESE PAMENENG
    Interval{ 0x1B61, 0x1B6A }, // So  [10] BALINESE MUSICAL SYMBOL DONG..BALINESE MUSICAL SYMBOL DANG GEDE
    Interval{ 0x1B74, 0x1B7C }, // So   [9] BALINESE MUSICAL SYMBOL RIGHT-HAND OPEN DUG..BALINESE MUSICAL SYMBOL LEFT-HAND OPEN PING
    Interval{ 0x1B7D, 0x1B7F }, // Po   [3] BALINESE PANTI LANTANG..BALINESE PANTI BAWAK
    Interval{ 0x1B82, 0x1B82 }, // Mc       SUNDANESE SIGN PANGWISAD
    Interval{ 0x1B83, 0x1BA0 }, // Lo  [30] SUNDANESE LETTER A..SUNDANESE LETTER HA
    Interval{ 0x1BA1, 0x1BA1 }, // Mc       SUNDANESE CONSONANT SIGN PAMINGKAL
    Interval{ 0x1BA6, 0x1BA7 }, // Mc   [2] SUNDANESE VOWEL SIGN PANAELAENG..SUNDANESE VOWEL SIGN PANOLONG
    Interval{ 0x1BAE, 0x1BAF }, // Lo   [2] SUNDANESE LETTER KHA..SUNDANESE LETTER SYA
    Interval{ 0x1BB0, 0x1BB9 }, // Nd  [10] SUNDANESE DIGIT ZERO..SUNDANESE DIGIT NINE
    Interval{ 0x1BBA, 0x1BE5 }, // Lo  [44] SUNDANESE AVAGRAHA..BATAK LETTER U
    Interval{ 0x1BE7, 0x1BE7 }, // Mc       BATAK VOWEL SIGN E
    Interval{ 0x1BEA, 0x1BEC }, // Mc   [3] BATAK VOWEL SIGN I..BATAK VOWEL SIGN O
    Interval{ 0x1BEE, 0x1BEE }, // Mc       BATAK VOWEL SIGN U
    Interval{ 0x1BFC, 0x1BFF }, // Po   [4] BATAK SYMBOL BINDU NA METEK..BATAK SYMBOL BINDU PANGOLAT
    Interval{ 0x1C00, 0x1C23 }, // Lo  [36] LEPCHA LETTER KA..LEPCHA LETTER A
    Interval{ 0x1C24, 0x1C2B }, // Mc   [8] LEPCHA SUBJOINED LETTER YA..LEPCHA VOWEL SIGN UU
    Interval{ 0x1C34, 0x1C35 }, // Mc   [2] LEPCHA CONSONANT SIGN NYIN-DO..LEPCHA CONSONANT SIGN KANG
    Interval{ 0x1C3B, 0x1C3F }, // Po   [5] LEPCHA PUNCTUATION TA-ROL..LEPCHA PUNCTUATION TSHOOK
    Interval{ 0x1C40, 0x1C49 }, // Nd  [10] LEPCHA DIGIT ZERO..LEPCHA DIGIT NINE
    Interval{ 0x1C4D, 0x1C4F }, // Lo   [3] LEPCHA LETTER TTA..LEPCHA LETTER DDA
    Interval{ 0x1C50, 0x1C59 }, // Nd  [10] OL CHIKI DIGIT ZERO..OL CHIKI DIGIT NINE
    Interval{ 0x1C5A, 0x1C77 }, // Lo  [30] OL CHIKI LETTER LA..OL CHIKI LETTER OH
    Interval{ 0x1C78, 0x1C7D }, // Lm   [6] OL CHIKI MU TTUDDAG..OL CHIKI AHAD
    Interval{ 0x1C7E, 0x1C7F }, // Po   [2] OL CHIKI PUNCTUATION MUCAAD..OL CHIKI PUNCTUATION DOUBLE MUCAAD
    Interval{ 0x1C80, 0x1C8A }, // L&  [11] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER TJE
    Interval{ 0x1C90, 0x1CBA }, // L&  [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Interval{ 0x1CBD, 0x1CBF }, // L&   [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Interval{ 0x1CC0, 0x1CC7 }, // Po   [8] SUNDANESE PUNCTUATION BINDU SURYA..SUNDANESE PUNCTUATION BINDU BA SATANGA
    Interval{ 0x1CD3, 0x1CD3 }, // Po       VEDIC SIGN NIHSHVASA
    Interval{ 0x1CE1, 0x1CE1 }, // Mc       VEDIC TONE ATHARVAVEDIC INDEPENDENT SVARITA
    Interval{ 0x1CE9, 0x1CEC }, // Lo   [4] VEDIC SIGN ANUSVARA ANTARGOMUKHA..VEDIC SIGN ANUSVARA VAMAGOMUKHA WITH TAIL
    Interval{ 0x1CEE, 0x1CF3 }, // Lo   [6] VEDIC SIGN HEXIFORM LONG ANUSVARA..VEDIC SIGN ROTATED ARDHAVISARGA
    Interval{ 0x1CF5, 0x1CF6 }, // Lo   [2] VEDIC SIGN JIHVAMULIYA..VEDIC SIGN UPADHMANIYA
    Interval{ 0x1CF7, 0x1CF7 }, // Mc       VEDIC SIGN ATIKRAMA
    Interval{ 0x1CFA, 0x1CFA }, // Lo       VEDIC SIGN DOUBLE ANUSVARA ANTARGOMUKHA
    Interval{ 0x1D00, 0x1D2B }, // L&  [44] LATIN LETTER SMALL CAPITAL A..CYRILLIC LETTER SMALL CAPITAL EL
    Interval{ 0x1D2C, 0x1D6A }, // Lm  [63] MODIFIER LETTER CAPITAL A..GREEK SUBSCRIPT SMALL LETTER CHI
    Interval{ 0x1D6B, 0x1D77 }, // L&  [13] LATIN SMALL LETTER UE..LATIN SMALL LETTER TURNED G
    Interval{ 0x1D78, 0x1D78 }, // Lm       MODIFIER LETTER CYRILLIC EN
    Interval{ 0x1D79, 0x1D9A }, // L&  [34] LATIN SMALL LETTER INSULAR G..LATIN SMALL LETTER EZH WITH RETROFLEX HOOK
    Interval{ 0x1D9B, 0x1DBF }, // Lm  [37] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL THETA
    Interval{ 0x1E00, 0x1F15 }, // L& [278] LATIN CAPITAL LETTER A WITH RING BELOW..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F18, 0x1F1D }, // L&   [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F20, 0x1F45 }, // L&  [38] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F48, 0x1F4D }, // L&   [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F50, 0x1F57 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F59, 0x1F59 }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA
    Interval{ 0x1F5B, 0x1F5B }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Interval{ 0x1F5D, 0x1F5D }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Interval{ 0x1F5F, 0x1F7D }, // L&  [31] GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI..GREEK SMALL LETTER OMEGA WITH OXIA
    Interval{ 0x1F80, 0x1FB4 }, // L&  [53] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB6, 0x1FBC }, // L&   [7] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Interval{ 0x1FBD, 0x1FBD }, // Sk       GREEK KORONIS
    Interval{ 0x1FBE, 0x1FBE }, // L&       GREEK PROSGEGRAMMENI
    Interval{ 0x1FBF, 0x1FC1 }, // Sk   [3] GREEK PSILI..GREEK DIALYTIKA AND PERISPOMENI
    Interval{ 0x1FC2, 0x1FC4 }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC6, 0x1FCC }, // L&   [7] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Interval{ 0x1FCD, 0x1FCF }, // Sk   [3] GREEK PSILI AND VARIA..GREEK PSILI AND PERISPOMENI
    Interval{ 0x1FD0, 0x1FD3 }, // L&   [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Interval{ 0x1FD6, 0x1FDB }, // L&   [6] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK CAPITAL LETTER IOTA WITH OXIA
    Interval{ 0x1FDD, 0x1FDF }, // Sk   [3] GREEK DASIA AND VARIA..GREEK DASIA AND PERISPOMENI
    Interval{ 0x1FE0, 0x1FEC }, // L&  [13] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Interval{ 0x1FED, 0x1FEF }, // Sk   [3] GREEK DIALYTIKA AND VARIA..GREEK VARIA
    Interval{ 0x1FF2, 0x1FF4 }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF6, 0x1FFC }, // L&   [7] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Interval{ 0x1FFD, 0x1FFE }, // Sk   [2] GREEK OXIA..GREEK DASIA
    Interval{ 0x2000, 0x200A }, // Zs  [11] EN QUAD..HAIR SPACE
    Interval{ 0x2010, 0x2015 }, // Pd   [6] HYPHEN..HORIZONTAL BAR
    Interval{ 0x2016, 0x2017 }, // Po   [2] DOUBLE VERTICAL LINE..DOUBLE LOW LINE
    Interval{ 0x2018, 0x2018 }, // Pi       LEFT SINGLE QUOTATION MARK
    Interval{ 0x2019, 0x2019 }, // Pf       RIGHT SINGLE QUOTATION MARK
    Interval{ 0x201A, 0x201A }, // Ps       SINGLE LOW-9 QUOTATION MARK
    Interval{ 0x201B, 0x201C }, // Pi   [2] SINGLE HIGH-REVERSED-9 QUOTATION MARK..LEFT DOUBLE QUOTATION MARK
    Interval{ 0x201D, 0x201D }, // Pf       RIGHT DOUBLE QUOTATION MARK
    Interval{ 0x201E, 0x201E }, // Ps       DOUBLE LOW-9 QUOTATION MARK
    Interval{ 0x201F, 0x201F }, // Pi       DOUBLE HIGH-REVERSED-9 QUOTATION MARK
    Interval{ 0x2020, 0x2027 }, // Po   [8] DAGGER..HYPHENATION POINT
    Interval{ 0x202F, 0x202F }, // Zs       NARROW NO-BREAK SPACE
    Interval{ 0x2030, 0x2038 }, // Po   [9] PER MILLE SIGN..CARET
    Interval{ 0x2039, 0x2039 }, // Pi       SINGLE LEFT-POINTING ANGLE QUOTATION MARK
    Interval{ 0x203A, 0x203A }, // Pf       SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
    Interval{ 0x203B, 0x203E }, // Po   [4] REFERENCE MARK..OVERLINE
    Interval{ 0x203F, 0x2040 }, // Pc   [2] UNDERTIE..CHARACTER TIE
    Interval{ 0x2041, 0x2043 }, // Po   [3] CARET INSERTION POINT..HYPHEN BULLET
    Interval{ 0x2044, 0x2044 }, // Sm       FRACTION SLASH
    Interval{ 0x2045, 0x2045 }, // Ps       LEFT SQUARE BRACKET WITH QUILL
    Interval{ 0x2046, 0x2046 }, // Pe       RIGHT SQUARE BRACKET WITH QUILL
    Interval{ 0x2047, 0x2051 }, // Po  [11] DOUBLE QUESTION MARK..TWO ASTERISKS ALIGNED VERTICALLY
    Interval{ 0x2052, 0x2052 }, // Sm       COMMERCIAL MINUS SIGN
    Interval{ 0x2053, 0x2053 }, // Po       SWUNG DASH
    Interval{ 0x2054, 0x2054 }, // Pc       INVERTED UNDERTIE
    Interval{ 0x2055, 0x205E }, // Po  [10] FLOWER PUNCTUATION MARK..VERTICAL FOUR DOTS
    Interval{ 0x205F, 0x205F }, // Zs       MEDIUM MATHEMATICAL SPACE
    Interval{ 0x2070, 0x2070 }, // No       SUPERSCRIPT ZERO
    Interval{ 0x2071, 0x2071 }, // Lm       SUPERSCRIPT LATIN SMALL LETTER I
    Interval{ 0x2074, 0x2079 }, // No   [6] SUPERSCRIPT FOUR..SUPERSCRIPT NINE
    Interval{ 0x207A, 0x207C }, // Sm   [3] SUPERSCRIPT PLUS SIGN..SUPERSCRIPT EQUALS SIGN
    Interval{ 0x207D, 0x207D }, // Ps       SUPERSCRIPT LEFT PARENTHESIS
    Interval{ 0x207E, 0x207E }, // Pe       SUPERSCRIPT RIGHT PARENTHESIS
    Interval{ 0x207F, 0x207F }, // Lm       SUPERSCRIPT LATIN SMALL LETTER N
    Interval{ 0x2080, 0x2089 }, // No  [10] SUBSCRIPT ZERO..SUBSCRIPT NINE
    Interval{ 0x208A, 0x208C }, // Sm   [3] SUBSCRIPT PLUS SIGN..SUBSCRIPT EQUALS SIGN
    Interval{ 0x208D, 0x208D }, // Ps       SUBSCRIPT LEFT PARENTHESIS
    Interval{ 0x208E, 0x208E }, // Pe       SUBSCRIPT RIGHT PARENTHESIS
    Interval{ 0x2090, 0x209C }, // Lm  [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Interval{ 0x20A0, 0x20C1 }, // Sc  [34] EURO-CURRENCY SIGN..SAUDI RIYAL SIGN
    Interval{ 0x2100, 0x2101 }, // So   [2] ACCOUNT OF..ADDRESSED TO THE SUBJECT
    Interval{ 0x2102, 0x2102 }, // L&       DOUBLE-STRUCK CAPITAL C
    Interval{ 0x2103, 0x2106 }, // So   [4] DEGREE CELSIUS..CADA UNA
    Interval{ 0x2107, 0x2107 }, // L&       EULER CONSTANT
    Interval{ 0x2108, 0x2109 }, // So   [2] SCRUPLE..DEGREE FAHRENHEIT
    Interval{ 0x210A, 0x2113 }, // L&  [10] SCRIPT SMALL G..SCRIPT SMALL L
    Interval{ 0x2114, 0x2114 }, // So       L B BAR SYMBOL
    Interval{ 0x2115, 0x2115 }, // L&       DOUBLE-STRUCK CAPITAL N
    Interval{ 0x2116, 0x2117 }, // So   [2] NUMERO SIGN..SOUND RECORDING COPYRIGHT
    Interval{ 0x2118, 0x2118 }, // Sm       SCRIPT CAPITAL P
    Interval{ 0x2119, 0x211D }, // L&   [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    Interval{ 0x211E, 0x2123 }, // So   [6] PRESCRIPTION TAKE..VERSICLE
    Interval{ 0x2124, 0x2124 }, // L&       DOUBLE-STRUCK CAPITAL Z
    Interval{ 0x2125, 0x2125 }, // So       OUNCE SIGN
    Interval{ 0x2126, 0x2126 }, // L&       OHM SIGN
    Interval{ 0x2127, 0x2127 }, // So       INVERTED OHM SIGN
    Interval{ 0x2128, 0x2128 }, // L&       BLACK-LETTER CAPITAL Z
    Interval{ 0x2129, 0x2129 }, // So       TURNED GREEK SMALL LETTER IOTA
    Interval{ 0x212A, 0x212D }, // L&   [4] KELVIN SIGN..BLACK-LETTER CAPITAL C
    Interval{ 0x212E, 0x212E }, // So       ESTIMATED SYMBOL
    Interval{ 0x212F, 0x2134 }, // L&   [6] SCRIPT SMALL E..SCRIPT SMALL O
    Interval{ 0x2135, 0x2138 }, // Lo   [4] ALEF SYMBOL..DALET SYMBOL
    Interval{ 0x2139, 0x2139 }, // L&       INFORMATION SOURCE
    Interval{ 0x213A, 0x213B }, // So   [2] ROTATED CAPITAL Q..FACSIMILE SIGN
    Interval{ 0x213C, 0x213F }, // L&   [4] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK CAPITAL PI
    Interval{ 0x2140, 0x2144 }, // Sm   [5] DOUBLE-STRUCK N-ARY SUMMATION..TURNED SANS-SERIF CAPITAL Y
    Interval{ 0x2145, 0x2149 }, // L&   [5] DOUBLE-STRUCK ITALIC CAPITAL D..DOUBLE-STRUCK ITALIC SMALL J
    Interval{ 0x214A, 0x214A }, // So       PROPERTY LINE
    Interval{ 0x214B, 0x214B }, // Sm       TURNED AMPERSAND
    Interval{ 0x214C, 0x214D }, // So   [2] PER SIGN..AKTIESELSKAB
    Interval{ 0x214E, 0x214E }, // L&       TURNED SMALL F
    Interval{ 0x214F, 0x214F }, // So       SYMBOL FOR SAMARITAN SOURCE
    Interval{ 0x2150, 0x215F }, // No  [16] VULGAR FRACTION ONE SEVENTH..FRACTION NUMERATOR ONE
    Interval{ 0x2160, 0x2182 }, // Nl  [35] ROMAN NUMERAL ONE..ROMAN NUMERAL TEN THOUSAND
    Interval{ 0x2183, 0x2184 }, // L&   [2] ROMAN NUMERAL REVERSED ONE HUNDRED..LATIN SMALL LETTER REVERSED C
    Interval{ 0x2185, 0x2188 }, // Nl   [4] ROMAN NUMERAL SIX LATE FORM..ROMAN NUMERAL ONE HUNDRED THOUSAND
    Interval{ 0x2189, 0x2189 }, // No       VULGAR FRACTION ZERO THIRDS
    Interval{ 0x218A, 0x218B }, // So   [2] TURNED DIGIT TWO..TURNED DIGIT THREE
    Interval{ 0x2190, 0x2194 }, // Sm   [5] LEFTWARDS ARROW..LEFT RIGHT ARROW
    Interval{ 0x2195, 0x2199 }, // So   [5] UP DOWN ARROW..SOUTH WEST ARROW
    Interval{ 0x219A, 0x219B }, // Sm   [2] LEFTWARDS ARROW WITH STROKE..RIGHTWARDS ARROW WITH STROKE
    Interval{ 0x219C, 0x219F }, // So   [4] LEFTWARDS WAVE ARROW..UPWARDS TWO HEADED ARROW
    Interval{ 0x21A0, 0x21A0 }, // Sm       RIGHTWARDS TWO HEADED ARROW
    Interval{ 0x21A1, 0x21A2 }, // So   [2] DOWNWARDS TWO HEADED ARROW..LEFTWARDS ARROW WITH TAIL
    Interval{ 0x21A3, 0x21A3 }, // Sm       RIGHTWARDS ARROW WITH TAIL
    Interval{ 0x21A4, 0x21A5 }, // So   [2] LEFTWARDS ARROW FROM BAR..UPWARDS ARROW FROM BAR
    Interval{ 0x21A6, 0x21A6 }, // Sm       RIGHTWARDS ARROW FROM BAR
    Interval{ 0x21A7, 0x21AD }, // So   [7] DOWNWARDS ARROW FROM BAR..LEFT RIGHT WAVE ARROW
    Interval{ 0x21AE, 0x21AE }, // Sm       LEFT RIGHT ARROW WITH STROKE
    Interval{ 0x21AF, 0x21CD }, // So  [31] DOWNWARDS ZIGZAG ARROW..LEFTWARDS DOUBLE ARROW WITH STROKE
    Interval{ 0x21CE, 0x21CF }, // Sm   [2] LEFT RIGHT DOUBLE ARROW WITH STROKE..RIGHTWARDS DOUBLE ARROW WITH STROKE
    Interval{ 0x21D0, 0x21D1 }, // So   [2] LEFTWARDS DOUBLE ARROW..UPWARDS DOUBLE ARROW
    Interval{ 0x21D2, 0x21D2 }, // Sm       RIGHTWARDS DOUBLE ARROW
    Interval{ 0x21D3, 0x21D3 }, // So       DOWNWARDS DOUBLE ARROW
    Interval{ 0x21D4, 0x21D4 }, // Sm       LEFT RIGHT DOUBLE ARROW
    Interval{ 0x21D5, 0x21F3 }, // So  [31] UP DOWN DOUBLE ARROW..UP DOWN WHITE ARROW
    Interval{ 0x21F4, 0x22FF }, // Sm [268] RIGHT ARROW WITH SMALL CIRCLE..Z NOTATION BAG MEMBERSHIP
    Interval{ 0x2300, 0x2307 }, // So   [8] DIAMETER SIGN..WAVY LINE
    Interval{ 0x2308, 0x2308 }, // Ps       LEFT CEILING
    Interval{ 0x2309, 0x2309 }, // Pe       RIGHT CEILING
    Interval{ 0x230A, 0x230A }, // Ps       LEFT FLOOR
    Interval{ 0x230B, 0x230B }, // Pe       RIGHT FLOOR
    Interval{ 0x230C, 0x231F }, // So  [20] BOTTOM RIGHT CROP..BOTTOM RIGHT CORNER
    Interval{ 0x2320, 0x2321 }, // Sm   [2] TOP HALF INTEGRAL..BOTTOM HALF INTEGRAL
    Interval{ 0x2322, 0x2328 }, // So   [7] FROWN..KEYBOARD
    Interval{ 0x2329, 0x2329 }, // Ps       LEFT-POINTING ANGLE BRACKET
    Interval{ 0x232A, 0x232A }, // Pe       RIGHT-POINTING ANGLE BRACKET
    Interval{ 0x232B, 0x237B }, // So  [81] ERASE TO THE LEFT..NOT CHECK MARK
    Interval{ 0x237C, 0x237C }, // Sm       RIGHT ANGLE WITH DOWNWARDS ZIGZAG ARROW
    Interval{ 0x237D, 0x239A }, // So  [30] SHOULDERED OPEN BOX..CLEAR SCREEN SYMBOL
    Interval{ 0x239B, 0x23B3 }, // Sm  [25] LEFT PARENTHESIS UPPER HOOK..SUMMATION BOTTOM
    Interval{ 0x23B4, 0x23DB }, // So  [40] TOP SQUARE BRACKET..FUSE
    Interval{ 0x23DC, 0x23E1 }, // Sm   [6] TOP PARENTHESIS..BOTTOM TORTOISE SHELL BRACKET
    Interval{ 0x23E2, 0x2429 }, // So  [72] WHITE TRAPEZIUM..SYMBOL FOR DELETE MEDIUM SHADE FORM
    Interval{ 0x2440, 0x244A }, // So  [11] OCR HOOK..OCR DOUBLE BACKSLASH
    Interval{ 0x2460, 0x249B }, // No  [60] CIRCLED DIGIT ONE..NUMBER TWENTY FULL STOP
    Interval{ 0x249C, 0x24E9 }, // So  [78] PARENTHESIZED LATIN SMALL LETTER A..CIRCLED LATIN SMALL LETTER Z
    Interval{ 0x24EA, 0x24FF }, // No  [22] CIRCLED DIGIT ZERO..NEGATIVE CIRCLED DIGIT ZERO
    Interval{ 0x2500, 0x25B6 }, // So [183] BOX DRAWINGS LIGHT HORIZONTAL..BLACK RIGHT-POINTING TRIANGLE
    Interval{ 0x25B7, 0x25B7 }, // Sm       WHITE RIGHT-POINTING TRIANGLE
    Interval{ 0x25B8, 0x25C0 }, // So   [9] BLACK RIGHT-POINTING SMALL TRIANGLE..BLACK LEFT-POINTING TRIANGLE
    Interval{ 0x25C1, 0x25C1 }, // Sm       WHITE LEFT-POINTING TRIANGLE
    Interval{ 0x25C2, 0x25F7 }, // So  [54] BLACK LEFT-POINTING SMALL TRIANGLE..WHITE CIRCLE WITH UPPER RIGHT QUADRANT
    Interval{ 0x25F8, 0x25FF }, // Sm   [8] UPPER LEFT TRIANGLE..LOWER RIGHT TRIANGLE
    Interval{ 0x2600, 0x266E }, // So [111] BLACK SUN WITH RAYS..MUSIC NATURAL SIGN
    Interval{ 0x266F, 0x266F }, // Sm       MUSIC SHARP SIGN
    Interval{ 0x2670, 0x2767 }, // So [248] WEST SYRIAC CROSS..ROTATED FLORAL HEART BULLET
    Interval{ 0x2768, 0x2768 }, // Ps       MEDIUM LEFT PARENTHESIS ORNAMENT
    Interval{ 0x2769, 0x2769 }, // Pe       MEDIUM RIGHT PARENTHESIS ORNAMENT
    Interval{ 0x276A, 0x276A }, // Ps       MEDIUM FLATTENED LEFT PARENTHESIS ORNAMENT
    Interval{ 0x276B, 0x276B }, // Pe       MEDIUM FLATTENED RIGHT PARENTHESIS ORNAMENT
    Interval{ 0x276C, 0x276C }, // Ps       MEDIUM LEFT-POINTING ANGLE BRACKET ORNAMENT
    Interval{ 0x276D, 0x276D }, // Pe       MEDIUM RIGHT-POINTING ANGLE BRACKET ORNAMENT
    Interval{ 0x276E, 0x276E }, // Ps       HEAVY LEFT-POINTING ANGLE QUOTATION MARK ORNAMENT
    Interval{ 0x276F, 0x276F }, // Pe       HEAVY RIGHT-POINTING ANGLE QUOTATION MARK ORNAMENT
    Interval{ 0x2770, 0x2770 }, // Ps       HEAVY LEFT-POINTING ANGLE BRACKET ORNAMENT
    Interval{ 0x2771, 0x2771 }, // Pe       HEAVY RIGHT-POINTING ANGLE BRACKET ORNAMENT
    Interval{ 0x2772, 0x2772 }, // Ps       LIGHT LEFT TORTOISE SHELL BRACKET ORNAMENT
    Interval{ 0x2773, 0x2773 }, // Pe       LIGHT RIGHT TORTOISE SHELL BRACKET ORNAMENT
    Interval{ 0x2774, 0x2774 }, // Ps       MEDIUM LEFT CURLY BRACKET ORNAMENT
    Interval{ 0x2775, 0x2775 }, // Pe       MEDIUM RIGHT CURLY BRACKET ORNAMENT
    Interval{ 0x2776, 0x2793 }, // No  [30] DINGBAT NEGATIVE CIRCLED DIGIT ONE..DINGBAT NEGATIVE CIRCLED SANS-SERIF NUMBER TEN
    Interval{ 0x2794, 0x27BF }, // So  [44] HEAVY WIDE-HEADED RIGHTWARDS ARROW..DOUBLE CURLY LOOP
    Interval{ 0x27C0, 0x27C4 }, // Sm   [5] THREE DIMENSIONAL ANGLE..OPEN SUPERSET
    Interval{ 0x27C5, 0x27C5 }, // Ps       LEFT S-SHAPED BAG DELIMITER
    Interval{ 0x27C6, 0x27C6 }, // Pe       RIGHT S-SHAPED BAG DELIMITER
    Interval{ 0x27C7, 0x27E5 }, // Sm  [31] OR WITH DOT INSIDE..WHITE SQUARE WITH RIGHTWARDS TICK
    Interval{ 0x27E6, 0x27E6 }, // Ps       MATHEMATICAL LEFT WHITE SQUARE BRACKET
    Interval{ 0x27E7, 0x27E7 }, // Pe       MATHEMATICAL RIGHT WHITE SQUARE BRACKET
    Interval{ 0x27E8, 0x27E8 }, // Ps       MATHEMATICAL LEFT ANGLE BRACKET
    Interval{ 0x27E9, 0x27E9 }, // Pe       MATHEMATICAL RIGHT ANGLE BRACKET
    Interval{ 0x27EA, 0x27EA }, // Ps       MATHEMATICAL LEFT DOUBLE ANGLE BRACKET
    Interval{ 0x27EB, 0x27EB }, // Pe       MATHEMATICAL RIGHT DOUBLE ANGLE BRACKET
    Interval{ 0x27EC, 0x27EC }, // Ps       MATHEMATICAL LEFT WHITE TORTOISE SHELL BRACKET
    Interval{ 0x27ED, 0x27ED }, // Pe       MATHEMATICAL RIGHT WHITE TORTOISE SHELL BRACKET
    Interval{ 0x27EE, 0x27EE }, // Ps       MATHEMATICAL LEFT FLATTENED PARENTHESIS
    Interval{ 0x27EF, 0x27EF }, // Pe       MATHEMATICAL RIGHT FLATTENED PARENTHESIS
    Interval{ 0x27F0, 0x27FF }, // Sm  [16] UPWARDS QUADRUPLE ARROW..LONG RIGHTWARDS SQUIGGLE ARROW
    Interval{ 0x2800, 0x28FF }, // So [256] BRAILLE PATTERN BLANK..BRAILLE PATTERN DOTS-12345678
    Interval{ 0x2900, 0x2982 }, // Sm [131] RIGHTWARDS TWO-HEADED ARROW WITH VERTICAL STROKE..Z NOTATION TYPE COLON
    Interval{ 0x2983, 0x2983 }, // Ps       LEFT WHITE CURLY BRACKET
    Interval{ 0x2984, 0x2984 }, // Pe       RIGHT WHITE CURLY BRACKET
    Interval{ 0x2985, 0x2985 }, // Ps       LEFT WHITE PARENTHESIS
    Interval{ 0x2986, 0x2986 }, // Pe       RIGHT WHITE PARENTHESIS
    Interval{ 0x2987, 0x2987 }, // Ps       Z NOTATION LEFT IMAGE BRACKET
    Interval{ 0x2988, 0x2988 }, // Pe       Z NOTATION RIGHT IMAGE BRACKET
    Interval{ 0x2989, 0x2989 }, // Ps       Z NOTATION LEFT BINDING BRACKET
    Interval{ 0x298A, 0x298A }, // Pe       Z NOTATION RIGHT BINDING BRACKET
    Interval{ 0x298B, 0x298B }, // Ps       LEFT SQUARE BRACKET WITH UNDERBAR
    Interval{ 0x298C, 0x298C }, // Pe       RIGHT SQUARE BRACKET WITH UNDERBAR
    Interval{ 0x298D, 0x298D }, // Ps       LEFT SQUARE BRACKET WITH TICK IN TOP CORNER
    Interval{ 0x298E, 0x298E }, // Pe       RIGHT SQUARE BRACKET WITH TICK IN BOTTOM CORNER
    Interval{ 0x298F, 0x298F }, // Ps       LEFT SQUARE BRACKET WITH TICK IN BOTTOM CORNER
    Interval{ 0x2990, 0x2990 }, // Pe       RIGHT SQUARE BRACKET WITH TICK IN TOP CORNER
    Interval{ 0x2991, 0x2991 }, // Ps       LEFT ANGLE BRACKET WITH DOT
    Interval{ 0x2992, 0x2992 }, // Pe       RIGHT ANGLE BRACKET WITH DOT
    Interval{ 0x2993, 0x2993 }, // Ps       LEFT ARC LESS-THAN BRACKET
    Interval{ 0x2994, 0x2994 }, // Pe       RIGHT ARC GREATER-THAN BRACKET
    Interval{ 0x2995, 0x2995 }, // Ps       DOUBLE LEFT ARC GREATER-THAN BRACKET
    Interval{ 0x2996, 0x2996 }, // Pe       DOUBLE RIGHT ARC LESS-THAN BRACKET
    Interval{ 0x2997, 0x2997 }, // Ps       LEFT BLACK TORTOISE SHELL BRACKET
    Interval{ 0x2998, 0x2998 }, // Pe       RIGHT BLACK TORTOISE SHELL BRACKET
    Interval{ 0x2999, 0x29D7 }, // Sm  [63] DOTTED FENCE..BLACK HOURGLASS
    Interval{ 0x29D8, 0x29D8 }, // Ps       LEFT WIGGLY FENCE
    Interval{ 0x29D9, 0x29D9 }, // Pe       RIGHT WIGGLY FENCE
    Interval{ 0x29DA, 0x29DA }, // Ps       LEFT DOUBLE WIGGLY FENCE
    Interval{ 0x29DB, 0x29DB }, // Pe       RIGHT DOUBLE WIGGLY FENCE
    Interval{ 0x29DC, 0x29FB }, // Sm  [32] INCOMPLETE INFINITY..TRIPLE PLUS
    Interval{ 0x29FC, 0x29FC }, // Ps       LEFT-POINTING CURVED ANGLE BRACKET
    Interval{ 0x29FD, 0x29FD }, // Pe       RIGHT-POINTING CURVED ANGLE BRACKET
    Interval{ 0x29FE, 0x2AFF }, // Sm [258] TINY..N-ARY WHITE VERTICAL BAR
    Interval{ 0x2B00, 0x2B2F }, // So  [48] NORTH EAST WHITE ARROW..WHITE VERTICAL ELLIPSE
    Interval{ 0x2B30, 0x2B44 }, // Sm  [21] LEFT ARROW WITH SMALL CIRCLE..RIGHTWARDS ARROW THROUGH SUPERSET
    Interval{ 0x2B45, 0x2B46 }, // So   [2] LEFTWARDS QUADRUPLE ARROW..RIGHTWARDS QUADRUPLE ARROW
    Interval{ 0x2B47, 0x2B4C }, // Sm   [6] REVERSE TILDE OPERATOR ABOVE RIGHTWARDS ARROW..RIGHTWARDS ARROW ABOVE REVERSE TILDE OPERATOR
    Interval{ 0x2B4D, 0x2B73 }, // So  [39] DOWNWARDS TRIANGLE-HEADED ZIGZAG ARROW..DOWNWARDS TRIANGLE-HEADED ARROW TO BAR
    Interval{ 0x2B76, 0x2BFF }, // So [138] NORTH WEST TRIANGLE-HEADED ARROW TO BAR..HELLSCHREIBER PAUSE SYMBOL
    Interval{ 0x2C00, 0x2C7B }, // L& [124] GLAGOLITIC CAPITAL LETTER AZU..LATIN LETTER SMALL CAPITAL TURNED E
    Interval{ 0x2C7C, 0x2C7D }, // Lm   [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    Interval{ 0x2C7E, 0x2CE4 }, // L& [103] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC SYMBOL KAI
    Interval{ 0x2CE5, 0x2CEA }, // So   [6] COPTIC SYMBOL MI RO..COPTIC SYMBOL SHIMA SIMA
    Interval{ 0x2CEB, 0x2CEE }, // L&   [4] COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI..COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF2, 0x2CF3 }, // L&   [2] COPTIC CAPITAL LETTER BOHAIRIC KHEI..COPTIC SMALL LETTER BOHAIRIC KHEI
    Interval{ 0x2CF9, 0x2CFC }, // Po   [4] COPTIC OLD NUBIAN FULL STOP..COPTIC OLD NUBIAN VERSE DIVIDER
    Interval{ 0x2CFD, 0x2CFD }, // No       COPTIC FRACTION ONE HALF
    Interval{ 0x2CFE, 0x2CFF }, // Po   [2] COPTIC FULL STOP..COPTIC MORPHOLOGICAL DIVIDER
    Interval{ 0x2D00, 0x2D25 }, // L&  [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Interval{ 0x2D27, 0x2D27 }, // L&       GEORGIAN SMALL LETTER YN
    Interval{ 0x2D2D, 0x2D2D }, // L&       GEORGIAN SMALL LETTER AEN
    Interval{ 0x2D30, 0x2D67 }, // Lo  [56] TIFINAGH LETTER YA..TIFINAGH LETTER YO
    Interval{ 0x2D6F, 0x2D6F }, // Lm       TIFINAGH MODIFIER LETTER LABIALIZATION MARK
    Interval{ 0x2D70, 0x2D70 }, // Po       TIFINAGH SEPARATOR MARK
    Interval{ 0x2D80, 0x2D96 }, // Lo  [23] ETHIOPIC SYLLABLE LOA..ETHIOPIC SYLLABLE GGWE
    Interval{ 0x2DA0, 0x2DA6 }, // Lo   [7] ETHIOPIC SYLLABLE SSA..ETHIOPIC SYLLABLE SSO
    Interval{ 0x2DA8, 0x2DAE }, // Lo   [7] ETHIOPIC SYLLABLE CCA..ETHIOPIC SYLLABLE CCO
    Interval{ 0x2DB0, 0x2DB6 }, // Lo   [7] ETHIOPIC SYLLABLE ZZA..ETHIOPIC SYLLABLE ZZO
    Interval{ 0x2DB8, 0x2DBE }, // Lo   [7] ETHIOPIC SYLLABLE CCHA..ETHIOPIC SYLLABLE CCHO
    Interval{ 0x2DC0, 0x2DC6 }, // Lo   [7] ETHIOPIC SYLLABLE QYA..ETHIOPIC SYLLABLE QYO
    Interval{ 0x2DC8, 0x2DCE }, // Lo   [7] ETHIOPIC SYLLABLE KYA..ETHIOPIC SYLLABLE KYO
    Interval{ 0x2DD0, 0x2DD6 }, // Lo   [7] ETHIOPIC SYLLABLE XYA..ETHIOPIC SYLLABLE XYO
    Interval{ 0x2DD8, 0x2DDE }, // Lo   [7] ETHIOPIC SYLLABLE GYA..ETHIOPIC SYLLABLE GYO
    Interval{ 0x2E00, 0x2E01 }, // Po   [2] RIGHT ANGLE SUBSTITUTION MARKER..RIGHT ANGLE DOTTED SUBSTITUTION MARKER
    Interval{ 0x2E02, 0x2E02 }, // Pi       LEFT SUBSTITUTION BRACKET
    Interval{ 0x2E03, 0x2E03 }, // Pf       RIGHT SUBSTITUTION BRACKET
    Interval{ 0x2E04, 0x2E04 }, // Pi       LEFT DOTTED SUBSTITUTION BRACKET
    Interval{ 0x2E05, 0x2E05 }, // Pf       RIGHT DOTTED SUBSTITUTION BRACKET
    Interval{ 0x2E06, 0x2E08 }, // Po   [3] RAISED INTERPOLATION MARKER..DOTTED TRANSPOSITION MARKER
    Interval{ 0x2E09, 0x2E09 }, // Pi       LEFT TRANSPOSITION BRACKET
    Interval{ 0x2E0A, 0x2E0A }, // Pf       RIGHT TRANSPOSITION BRACKET
    Interval{ 0x2E0B, 0x2E0B }, // Po       RAISED SQUARE
    Interval{ 0x2E0C, 0x2E0C }, // Pi       LEFT RAISED OMISSION BRACKET
    Interval{ 0x2E0D, 0x2E0D }, // Pf       RIGHT RAISED OMISSION BRACKET
    Interval{ 0x2E0E, 0x2E16 }, // Po   [9] EDITORIAL CORONIS..DOTTED RIGHT-POINTING ANGLE
    Interval{ 0x2E17, 0x2E17 }, // Pd       DOUBLE OBLIQUE HYPHEN
    Interval{ 0x2E18, 0x2E19 }, // Po   [2] INVERTED INTERROBANG..PALM BRANCH
    Interval{ 0x2E1A, 0x2E1A }, // Pd       HYPHEN WITH DIAERESIS
    Interval{ 0x2E1B, 0x2E1B }, // Po       TILDE WITH RING ABOVE
    Interval{ 0x2E1C, 0x2E1C }, // Pi       LEFT LOW PARAPHRASE BRACKET
    Interval{ 0x2E1D, 0x2E1D }, // Pf       RIGHT LOW PARAPHRASE BRACKET
    Interval{ 0x2E1E, 0x2E1F }, // Po   [2] TILDE WITH DOT ABOVE..TILDE WITH DOT BELOW
    Interval{ 0x2E20, 0x2E20 }, // Pi       LEFT VERTICAL BAR WITH QUILL
    Interval{ 0x2E21, 0x2E21 }, // Pf       RIGHT VERTICAL BAR WITH QUILL
    Interval{ 0x2E22, 0x2E22 }, // Ps       TOP LEFT HALF BRACKET
    Interval{ 0x2E23, 0x2E23 }, // Pe       TOP RIGHT HALF BRACKET
    Interval{ 0x2E24, 0x2E24 }, // Ps       BOTTOM LEFT HALF BRACKET
    Interval{ 0x2E25, 0x2E25 }, // Pe       BOTTOM RIGHT HALF BRACKET
    Interval{ 0x2E26, 0x2E26 }, // Ps       LEFT SIDEWAYS U BRACKET
    Interval{ 0x2E27, 0x2E27 }, // Pe       RIGHT SIDEWAYS U BRACKET
    Interval{ 0x2E28, 0x2E28 }, // Ps       LEFT DOUBLE PARENTHESIS
    Interval{ 0x2E29, 0x2E29 }, // Pe       RIGHT DOUBLE PARENTHESIS
    Interval{ 0x2E2A, 0x2E2E }, // Po   [5] TWO DOTS OVER ONE DOT PUNCTUATION..REVERSED QUESTION MARK
    Interval{ 0x2E2F, 0x2E2F }, // Lm       VERTICAL TILDE
    Interval{ 0x2E30, 0x2E39 }, // Po  [10] RING POINT..TOP HALF SECTION SIGN
    Interval{ 0x2E3A, 0x2E3B }, // Pd   [2] TWO-EM DASH..THREE-EM DASH
    Interval{ 0x2E3C, 0x2E3F }, // Po   [4] STENOGRAPHIC FULL STOP..CAPITULUM
    Interval{ 0x2E40, 0x2E40 }, // Pd       DOUBLE HYPHEN
    Interval{ 0x2E41, 0x2E41 }, // Po       REVERSED COMMA
    Interval{ 0x2E42, 0x2E42 }, // Ps       DOUBLE LOW-REVERSED-9 QUOTATION MARK
    Interval{ 0x2E43, 0x2E4F }, // Po  [13] DASH WITH LEFT UPTURN..CORNISH VERSE DIVIDER
    Interval{ 0x2E50, 0x2E51 }, // So   [2] CROSS PATTY WITH RIGHT CROSSBAR..CROSS PATTY WITH LEFT CROSSBAR
    Interval{ 0x2E52, 0x2E54 }, // Po   [3] TIRONIAN SIGN CAPITAL ET..MEDIEVAL QUESTION MARK
    Interval{ 0x2E55, 0x2E55 }, // Ps       LEFT SQUARE BRACKET WITH STROKE
    Interval{ 0x2E56, 0x2E56 }, // Pe       RIGHT SQUARE BRACKET WITH STROKE
    Interval{ 0x2E57, 0x2E57 }, // Ps       LEFT SQUARE BRACKET WITH DOUBLE STROKE
    Interval{ 0x2E58, 0x2E58 }, // Pe       RIGHT SQUARE BRACKET WITH DOUBLE STROKE
    Interval{ 0x2E59, 0x2E59 }, // Ps       TOP HALF LEFT PARENTHESIS
    Interval{ 0x2E5A, 0x2E5A }, // Pe       TOP HALF RIGHT PARENTHESIS
    Interval{ 0x2E5B, 0x2E5B }, // Ps       BOTTOM HALF LEFT PARENTHESIS
    Interval{ 0x2E5C, 0x2E5C }, // Pe       BOTTOM HALF RIGHT PARENTHESIS
    Interval{ 0x2E5D, 0x2E5D }, // Pd       OBLIQUE HYPHEN
    Interval{ 0x2E80, 0x2E99 }, // So  [26] CJK RADICAL REPEAT..CJK RADICAL RAP
    Interval{ 0x2E9B, 0x2EF3 }, // So  [89] CJK RADICAL CHOKE..CJK RADICAL C-SIMPLIFIED TURTLE
    Interval{ 0x2F00, 0x2FD5 }, // So [214] KANGXI RADICAL ONE..KANGXI RADICAL FLUTE
    Interval{ 0x2FF0, 0x2FFF }, // So  [16] IDEOGRAPHIC DESCRIPTION CHARACTER LEFT TO RIGHT..IDEOGRAPHIC DESCRIPTION CHARACTER ROTATION
    Interval{ 0x3000, 0x3000 }, // Zs       IDEOGRAPHIC SPACE
    Interval{ 0x3001, 0x3003 }, // Po   [3] IDEOGRAPHIC COMMA..DITTO MARK
    Interval{ 0x3004, 0x3004 }, // So       JAPANESE INDUSTRIAL STANDARD SYMBOL
    Interval{ 0x3005, 0x3005 }, // Lm       IDEOGRAPHIC ITERATION MARK
    Interval{ 0x3006, 0x3006 }, // Lo       IDEOGRAPHIC CLOSING MARK
    Interval{ 0x3007, 0x3007 }, // Nl       IDEOGRAPHIC NUMBER ZERO
    Interval{ 0x3008, 0x3008 }, // Ps       LEFT ANGLE BRACKET
    Interval{ 0x3009, 0x3009 }, // Pe       RIGHT ANGLE BRACKET
    Interval{ 0x300A, 0x300A }, // Ps       LEFT DOUBLE ANGLE BRACKET
    Interval{ 0x300B, 0x300B }, // Pe       RIGHT DOUBLE ANGLE BRACKET
    Interval{ 0x300C, 0x300C }, // Ps       LEFT CORNER BRACKET
    Interval{ 0x300D, 0x300D }, // Pe       RIGHT CORNER BRACKET
    Interval{ 0x300E, 0x300E }, // Ps       LEFT WHITE CORNER BRACKET
    Interval{ 0x300F, 0x300F }, // Pe       RIGHT WHITE CORNER BRACKET
    Interval{ 0x3010, 0x3010 }, // Ps       LEFT BLACK LENTICULAR BRACKET
    Interval{ 0x3011, 0x3011 }, // Pe       RIGHT BLACK LENTICULAR BRACKET
    Interval{ 0x3012, 0x3013 }, // So   [2] POSTAL MARK..GETA MARK
    Interval{ 0x3014, 0x3014 }, // Ps       LEFT TORTOISE SHELL BRACKET
    Interval{ 0x3015, 0x3015 }, // Pe       RIGHT TORTOISE SHELL BRACKET
    Interval{ 0x3016, 0x3016 }, // Ps       LEFT WHITE LENTICULAR BRACKET
    Interval{ 0x3017, 0x3017 }, // Pe       RIGHT WHITE LENTICULAR BRACKET
    Interval{ 0x3018, 0x3018 }, // Ps       LEFT WHITE TORTOISE SHELL BRACKET
    Interval{ 0x3019, 0x3019 }, // Pe       RIGHT WHITE TORTOISE SHELL BRACKET
    Interval{ 0x301A, 0x301A }, // Ps       LEFT WHITE SQUARE BRACKET
    Interval{ 0x301B, 0x301B }, // Pe       RIGHT WHITE SQUARE BRACKET
    Interval{ 0x301C, 0x301C }, // Pd       WAVE DASH
    Interval{ 0x301D, 0x301D }, // Ps       REVERSED DOUBLE PRIME QUOTATION MARK
    Interval{ 0x301E, 0x301F }, // Pe   [2] DOUBLE PRIME QUOTATION MARK..LOW DOUBLE PRIME QUOTATION MARK
    Interval{ 0x3020, 0x3020 }, // So       POSTAL MARK FACE
    Interval{ 0x3021, 0x3029 }, // Nl   [9] HANGZHOU NUMERAL ONE..HANGZHOU NUMERAL NINE
    Interval{ 0x3030, 0x3030 }, // Pd       WAVY DASH
    Interval{ 0x3031, 0x3035 }, // Lm   [5] VERTICAL KANA REPEAT MARK..VERTICAL KANA REPEAT MARK LOWER HALF
    Interval{ 0x3036, 0x3037 }, // So   [2] CIRCLED POSTAL MARK..IDEOGRAPHIC TELEGRAPH LINE FEED SEPARATOR SYMBOL
    Interval{ 0x3038, 0x303A }, // Nl   [3] HANGZHOU NUMERAL TEN..HANGZHOU NUMERAL THIRTY
    Interval{ 0x303B, 0x303B }, // Lm       VERTICAL IDEOGRAPHIC ITERATION MARK
    Interval{ 0x303C, 0x303C }, // Lo       MASU MARK
    Interval{ 0x303D, 0x303D }, // Po       PART ALTERNATION MARK
    Interval{ 0x303E, 0x303F }, // So   [2] IDEOGRAPHIC VARIATION INDICATOR..IDEOGRAPHIC HALF FILL SPACE
    Interval{ 0x3041, 0x3096 }, // Lo  [86] HIRAGANA LETTER SMALL A..HIRAGANA LETTER SMALL KE
    Interval{ 0x309B, 0x309C }, // Sk   [2] KATAKANA-HIRAGANA VOICED SOUND MARK..KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Interval{ 0x309D, 0x309E }, // Lm   [2] HIRAGANA ITERATION MARK..HIRAGANA VOICED ITERATION MARK
    Interval{ 0x309F, 0x309F }, // Lo       HIRAGANA DIGRAPH YORI
    Interval{ 0x30A0, 0x30A0 }, // Pd       KATAKANA-HIRAGANA DOUBLE HYPHEN
    Interval{ 0x30A1, 0x30FA }, // Lo  [90] KATAKANA LETTER SMALL A..KATAKANA LETTER VO
    Interval{ 0x30FB, 0x30FB }, // Po       KATAKANA MIDDLE DOT
    Interval{ 0x30FC, 0x30FE }, // Lm   [3] KATAKANA-HIRAGANA PROLONGED SOUND MARK..KATAKANA VOICED ITERATION MARK
    Interval{ 0x30FF, 0x30FF }, // Lo       KATAKANA DIGRAPH KOTO
    Interval{ 0x3105, 0x312F }, // Lo  [43] BOPOMOFO LETTER B..BOPOMOFO LETTER NN
    Interval{ 0x3131, 0x318E }, // Lo  [94] HANGUL LETTER KIYEOK..HANGUL LETTER ARAEAE
    Interval{ 0x3190, 0x3191 }, // So   [2] IDEOGRAPHIC ANNOTATION LINKING MARK..IDEOGRAPHIC ANNOTATION REVERSE MARK
    Interval{ 0x3192, 0x3195 }, // No   [4] IDEOGRAPHIC ANNOTATION ONE MARK..IDEOGRAPHIC ANNOTATION FOUR MARK
    Interval{ 0x3196, 0x319F }, // So  [10] IDEOGRAPHIC ANNOTATION TOP MARK..IDEOGRAPHIC ANNOTATION MAN MARK
    Interval{ 0x31A0, 0x31BF }, // Lo  [32] BOPOMOFO LETTER BU..BOPOMOFO LETTER AH
    Interval{ 0x31C0, 0x31E5 }, // So  [38] CJK STROKE T..CJK STROKE SZP
    Interval{ 0x31EF, 0x31EF }, // So       IDEOGRAPHIC DESCRIPTION CHARACTER SUBTRACTION
    Interval{ 0x31F0, 0x31FF }, // Lo  [16] KATAKANA LETTER SMALL KU..KATAKANA LETTER SMALL RO
    Interval{ 0x3200, 0x321E }, // So  [31] PARENTHESIZED HANGUL KIYEOK..PARENTHESIZED KOREAN CHARACTER O HU
    Interval{ 0x3220, 0x3229 }, // No  [10] PARENTHESIZED IDEOGRAPH ONE..PARENTHESIZED IDEOGRAPH TEN
    Interval{ 0x322A, 0x3247 }, // So  [30] PARENTHESIZED IDEOGRAPH MOON..CIRCLED IDEOGRAPH KOTO
    Interval{ 0x3248, 0x324F }, // No   [8] CIRCLED NUMBER TEN ON BLACK SQUARE..CIRCLED NUMBER EIGHTY ON BLACK SQUARE
    Interval{ 0x3250, 0x3250 }, // So       PARTNERSHIP SIGN
    Interval{ 0x3251, 0x325F }, // No  [15] CIRCLED NUMBER TWENTY ONE..CIRCLED NUMBER THIRTY FIVE
    Interval{ 0x3260, 0x327F }, // So  [32] CIRCLED HANGUL KIYEOK..KOREAN STANDARD SYMBOL
    Interval{ 0x3280, 0x3289 }, // No  [10] CIRCLED IDEOGRAPH ONE..CIRCLED IDEOGRAPH TEN
    Interval{ 0x328A, 0x32B0 }, // So  [39] CIRCLED IDEOGRAPH MOON..CIRCLED IDEOGRAPH NIGHT
    Interval{ 0x32B1, 0x32BF }, // No  [15] CIRCLED NUMBER THIRTY SIX..CIRCLED NUMBER FIFTY
    Interval{ 0x32C0, 0x33FF }, // So [320] IDEOGRAPHIC TELEGRAPH SYMBOL FOR JANUARY..SQUARE GAL
    Interval{ 0x3400, 0x4DBF }, // Lo [6592] CJK UNIFIED IDEOGRAPH-3400..CJK UNIFIED IDEOGRAPH-4DBF
    Interval{ 0x4DC0, 0x4DFF }, // So  [64] HEXAGRAM FOR THE CREATIVE HEAVEN..HEXAGRAM FOR BEFORE COMPLETION
    Interval{ 0x4E00, 0xA014 }, // Lo [21013] CJK UNIFIED IDEOGRAPH-4E00..YI SYLLABLE E
    Interval{ 0xA015, 0xA015 }, // Lm       YI SYLLABLE WU
    Interval{ 0xA016, 0xA48C }, // Lo [1143] YI SYLLABLE BIT..YI SYLLABLE YYR
    Interval{ 0xA490, 0xA4C6 }, // So  [55] YI RADICAL QOT..YI RADICAL KE
    Interval{ 0xA4D0, 0xA4F7 }, // Lo  [40] LISU LETTER BA..LISU LETTER OE
    Interval{ 0xA4F8, 0xA4FD }, // Lm   [6] LISU LETTER TONE MYA TI..LISU LETTER TONE MYA JEU
    Interval{ 0xA4FE, 0xA4FF }, // Po   [2] LISU PUNCTUATION COMMA..LISU PUNCTUATION FULL STOP
    Interval{ 0xA500, 0xA60B }, // Lo [268] VAI SYLLABLE EE..VAI SYLLABLE NG
    Interval{ 0xA60C, 0xA60C }, // Lm       VAI SYLLABLE LENGTHENER
    Interval{ 0xA60D, 0xA60F }, // Po   [3] VAI COMMA..VAI QUESTION MARK
    Interval{ 0xA610, 0xA61F }, // Lo  [16] VAI SYLLABLE NDOLE FA..VAI SYMBOL JONG
    Interval{ 0xA620, 0xA629 }, // Nd  [10] VAI DIGIT ZERO..VAI DIGIT NINE
    Interval{ 0xA62A, 0xA62B }, // Lo   [2] VAI SYLLABLE NDOLE MA..VAI SYLLABLE NDOLE DO
    Interval{ 0xA640, 0xA66D }, // L&  [46] CYRILLIC CAPITAL LETTER ZEMLYA..CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA66E, 0xA66E }, // Lo       CYRILLIC LETTER MULTIOCULAR O
    Interval{ 0xA673, 0xA673 }, // Po       SLAVONIC ASTERISK
    Interval{ 0xA67E, 0xA67E }, // Po       CYRILLIC KAVYKA
    Interval{ 0xA67F, 0xA67F }, // Lm       CYRILLIC PAYEROK
    Interval{ 0xA680, 0xA69B }, // L&  [28] CYRILLIC CAPITAL LETTER DWE..CYRILLIC SMALL LETTER CROSSED O
    Interval{ 0xA69C, 0xA69D }, // Lm   [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    Interval{ 0xA6A0, 0xA6E5 }, // Lo  [70] BAMUM LETTER A..BAMUM LETTER KI
    Interval{ 0xA6E6, 0xA6EF }, // Nl  [10] BAMUM LETTER MO..BAMUM LETTER KOGHOM
    Interval{ 0xA6F2, 0xA6F7 }, // Po   [6] BAMUM NJAEMLI..BAMUM QUESTION MARK
    Interval{ 0xA700, 0xA716 }, // Sk  [23] MODIFIER LETTER CHINESE TONE YIN PING..MODIFIER LETTER EXTRA-LOW LEFT-STEM TONE BAR
    Interval{ 0xA717, 0xA71F }, // Lm   [9] MODIFIER LETTER DOT VERTICAL BAR..MODIFIER LETTER LOW INVERTED EXCLAMATION MARK
    Interval{ 0xA720, 0xA721 }, // Sk   [2] MODIFIER LETTER STRESS AND HIGH TONE..MODIFIER LETTER STRESS AND LOW TONE
    Interval{ 0xA722, 0xA76F }, // L&  [78] LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF..LATIN SMALL LETTER CON
    Interval{ 0xA770, 0xA770 }, // Lm       MODIFIER LETTER US
    Interval{ 0xA771, 0xA787 }, // L&  [23] LATIN SMALL LETTER DUM..LATIN SMALL LETTER INSULAR T
    Interval{ 0xA788, 0xA788 }, // Lm       MODIFIER LETTER LOW CIRCUMFLEX ACCENT
    Interval{ 0xA789, 0xA78A }, // Sk   [2] MODIFIER LETTER COLON..MODIFIER LETTER SHORT EQUALS SIGN
    Interval{ 0xA78B, 0xA78E }, // L&   [4] LATIN CAPITAL LETTER SALTILLO..LATIN SMALL LETTER L WITH RETROFLEX HOOK AND BELT
    Interval{ 0xA78F, 0xA78F }, // Lo       LATIN LETTER SINOLOGICAL DOT
    Interval{ 0xA790, 0xA7DC }, // L&  [77] LATIN CAPITAL LETTER N WITH DESCENDER..LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Interval{ 0xA7F1, 0xA7F4 }, // Lm   [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    Interval{ 0xA7F5, 0xA7F6 }, // L&   [2] LATIN CAPITAL LETTER REVERSED HALF H..LATIN SMALL LETTER REVERSED HALF H
    Interval{ 0xA7F7, 0xA7F7 }, // Lo       LATIN EPIGRAPHIC LETTER SIDEWAYS I
    Interval{ 0xA7F8, 0xA7F9 }, // Lm   [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    Interval{ 0xA7FA, 0xA7FA }, // L&       LATIN LETTER SMALL CAPITAL TURNED M
    Interval{ 0xA7FB, 0xA801 }, // Lo   [7] LATIN EPIGRAPHIC LETTER REVERSED F..SYLOTI NAGRI LETTER I
    Interval{ 0xA803, 0xA805 }, // Lo   [3] SYLOTI NAGRI LETTER U..SYLOTI NAGRI LETTER O
    Interval{ 0xA807, 0xA80A }, // Lo   [4] SYLOTI NAGRI LETTER KO..SYLOTI NAGRI LETTER GHO
    Interval{ 0xA80C, 0xA822 }, // Lo  [23] SYLOTI NAGRI LETTER CO..SYLOTI NAGRI LETTER HO
    Interval{ 0xA823, 0xA824 }, // Mc   [2] SYLOTI NAGRI VOWEL SIGN A..SYLOTI NAGRI VOWEL SIGN I
    Interval{ 0xA827, 0xA827 }, // Mc       SYLOTI NAGRI VOWEL SIGN OO
    Interval{ 0xA828, 0xA82B }, // So   [4] SYLOTI NAGRI POETRY MARK-1..SYLOTI NAGRI POETRY MARK-4
    Interval{ 0xA830, 0xA835 }, // No   [6] NORTH INDIC FRACTION ONE QUARTER..NORTH INDIC FRACTION THREE SIXTEENTHS
    Interval{ 0xA836, 0xA837 }, // So   [2] NORTH INDIC QUARTER MARK..NORTH INDIC PLACEHOLDER MARK
    Interval{ 0xA838, 0xA838 }, // Sc       NORTH INDIC RUPEE MARK
    Interval{ 0xA839, 0xA839 }, // So       NORTH INDIC QUANTITY MARK
    Interval{ 0xA840, 0xA873 }, // Lo  [52] PHAGS-PA LETTER KA..PHAGS-PA LETTER CANDRABINDU
    Interval{ 0xA874, 0xA877 }, // Po   [4] PHAGS-PA SINGLE HEAD MARK..PHAGS-PA MARK DOUBLE SHAD
    Interval{ 0xA880, 0xA881 }, // Mc   [2] SAURASHTRA SIGN ANUSVARA..SAURASHTRA SIGN VISARGA
    Interval{ 0xA882, 0xA8B3 }, // Lo  [50] SAURASHTRA LETTER A..SAURASHTRA LETTER LLA
    Interval{ 0xA8B4, 0xA8C3 }, // Mc  [16] SAURASHTRA CONSONANT SIGN HAARU..SAURASHTRA VOWEL SIGN AU
    Interval{ 0xA8CE, 0xA8CF }, // Po   [2] SAURASHTRA DANDA..SAURASHTRA DOUBLE DANDA
    Interval{ 0xA8D0, 0xA8D9 }, // Nd  [10] SAURASHTRA DIGIT ZERO..SAURASHTRA DIGIT NINE
    Interval{ 0xA8F2, 0xA8F7 }, // Lo   [6] DEVANAGARI SIGN SPACING CANDRABINDU..DEVANAGARI SIGN CANDRABINDU AVAGRAHA
    Interval{ 0xA8F8, 0xA8FA }, // Po   [3] DEVANAGARI SIGN PUSHPIKA..DEVANAGARI CARET
    Interval{ 0xA8FB, 0xA8FB }, // Lo       DEVANAGARI HEADSTROKE
    Interval{ 0xA8FC, 0xA8FC }, // Po       DEVANAGARI SIGN SIDDHAM
    Interval{ 0xA8FD, 0xA8FE }, // Lo   [2] DEVANAGARI JAIN OM..DEVANAGARI LETTER AY
    Interval{ 0xA900, 0xA909 }, // Nd  [10] KAYAH LI DIGIT ZERO..KAYAH LI DIGIT NINE
    Interval{ 0xA90A, 0xA925 }, // Lo  [28] KAYAH LI LETTER KA..KAYAH LI LETTER OO
    Interval{ 0xA92E, 0xA92F }, // Po   [2] KAYAH LI SIGN CWI..KAYAH LI SIGN SHYA
    Interval{ 0xA930, 0xA946 }, // Lo  [23] REJANG LETTER KA..REJANG LETTER A
    Interval{ 0xA952, 0xA952 }, // Mc       REJANG CONSONANT SIGN H
    Interval{ 0xA95F, 0xA95F }, // Po       REJANG SECTION MARK
    Interval{ 0xA960, 0xA97C }, // Lo  [29] HANGUL CHOSEONG TIKEUT-MIEUM..HANGUL CHOSEONG SSANGYEORINHIEUH
    Interval{ 0xA983, 0xA983 }, // Mc       JAVANESE SIGN WIGNYAN
    Interval{ 0xA984, 0xA9B2 }, // Lo  [47] JAVANESE LETTER A..JAVANESE LETTER HA
    Interval{ 0xA9B4, 0xA9B5 }, // Mc   [2] JAVANESE VOWEL SIGN TARUNG..JAVANESE VOWEL SIGN TOLONG
    Interval{ 0xA9BA, 0xA9BB }, // Mc   [2] JAVANESE VOWEL SIGN TALING..JAVANESE VOWEL SIGN DIRGA MURE
    Interval{ 0xA9BE, 0xA9BF }, // Mc   [2] JAVANESE CONSONANT SIGN PENGKAL..JAVANESE CONSONANT SIGN CAKRA
    Interval{ 0xA9C1, 0xA9CD }, // Po  [13] JAVANESE LEFT RERENGGAN..JAVANESE TURNED PADA PISELEH
    Interval{ 0xA9CF, 0xA9CF }, // Lm       JAVANESE PANGRANGKEP
    Interval{ 0xA9D0, 0xA9D9 }, // Nd  [10] JAVANESE DIGIT ZERO..JAVANESE DIGIT NINE
    Interval{ 0xA9DE, 0xA9DF }, // Po   [2] JAVANESE PADA TIRTA TUMETES..JAVANESE PADA ISEN-ISEN
    Interval{ 0xA9E0, 0xA9E4 }, // Lo   [5] MYANMAR LETTER SHAN GHA..MYANMAR LETTER SHAN BHA
    Interval{ 0xA9E6, 0xA9E6 }, // Lm       MYANMAR MODIFIER LETTER SHAN REDUPLICATION
    Interval{ 0xA9E7, 0xA9EF }, // Lo   [9] MYANMAR LETTER TAI LAING NYA..MYANMAR LETTER TAI LAING NNA
    Interval{ 0xA9F0, 0xA9F9 }, // Nd  [10] MYANMAR TAI LAING DIGIT ZERO..MYANMAR TAI LAING DIGIT NINE
    Interval{ 0xA9FA, 0xA9FE }, // Lo   [5] MYANMAR LETTER TAI LAING LLA..MYANMAR LETTER TAI LAING BHA
    Interval{ 0xAA00, 0xAA28 }, // Lo  [41] CHAM LETTER A..CHAM LETTER HA
    Interval{ 0xAA2F, 0xAA30 }, // Mc   [2] CHAM VOWEL SIGN O..CHAM VOWEL SIGN AI
    Interval{ 0xAA33, 0xAA34 }, // Mc   [2] CHAM CONSONANT SIGN YA..CHAM CONSONANT SIGN RA
    Interval{ 0xAA40, 0xAA42 }, // Lo   [3] CHAM LETTER FINAL K..CHAM LETTER FINAL NG
    Interval{ 0xAA44, 0xAA4B }, // Lo   [8] CHAM LETTER FINAL CH..CHAM LETTER FINAL SS
    Interval{ 0xAA4D, 0xAA4D }, // Mc       CHAM CONSONANT SIGN FINAL H
    Interval{ 0xAA50, 0xAA59 }, // Nd  [10] CHAM DIGIT ZERO..CHAM DIGIT NINE
    Interval{ 0xAA5C, 0xAA5F }, // Po   [4] CHAM PUNCTUATION SPIRAL..CHAM PUNCTUATION TRIPLE DANDA
    Interval{ 0xAA60, 0xAA6F }, // Lo  [16] MYANMAR LETTER KHAMTI GA..MYANMAR LETTER KHAMTI FA
    Interval{ 0xAA70, 0xAA70 }, // Lm       MYANMAR MODIFIER LETTER KHAMTI REDUPLICATION
    Interval{ 0xAA71, 0xAA76 }, // Lo   [6] MYANMAR LETTER KHAMTI XA..MYANMAR LOGOGRAM KHAMTI HM
    Interval{ 0xAA77, 0xAA79 }, // So   [3] MYANMAR SYMBOL AITON EXCLAMATION..MYANMAR SYMBOL AITON TWO
    Interval{ 0xAA7A, 0xAA7A }, // Lo       MYANMAR LETTER AITON RA
    Interval{ 0xAA7B, 0xAA7B }, // Mc       MYANMAR SIGN PAO KAREN TONE
    Interval{ 0xAA7D, 0xAA7D }, // Mc       MYANMAR SIGN TAI LAING TONE-5
    Interval{ 0xAA7E, 0xAAAF }, // Lo  [50] MYANMAR LETTER SHWE PALAUNG CHA..TAI VIET LETTER HIGH O
    Interval{ 0xAAB1, 0xAAB1 }, // Lo       TAI VIET VOWEL AA
    Interval{ 0xAAB5, 0xAAB6 }, // Lo   [2] TAI VIET VOWEL E..TAI VIET VOWEL O
    Interval{ 0xAAB9, 0xAABD }, // Lo   [5] TAI VIET VOWEL UEA..TAI VIET VOWEL AN
    Interval{ 0xAAC0, 0xAAC0 }, // Lo       TAI VIET TONE MAI NUENG
    Interval{ 0xAAC2, 0xAAC2 }, // Lo       TAI VIET TONE MAI SONG
    Interval{ 0xAADB, 0xAADC }, // Lo   [2] TAI VIET SYMBOL KON..TAI VIET SYMBOL NUENG
    Interval{ 0xAADD, 0xAADD }, // Lm       TAI VIET SYMBOL SAM
    Interval{ 0xAADE, 0xAADF }, // Po   [2] TAI VIET SYMBOL HO HOI..TAI VIET SYMBOL KOI KOI
    Interval{ 0xAAE0, 0xAAEA }, // Lo  [11] MEETEI MAYEK LETTER E..MEETEI MAYEK LETTER SSA
    Interval{ 0xAAEB, 0xAAEB }, // Mc       MEETEI MAYEK VOWEL SIGN II
    Interval{ 0xAAEE, 0xAAEF }, // Mc   [2] MEETEI MAYEK VOWEL SIGN AU..MEETEI MAYEK VOWEL SIGN AAU
    Interval{ 0xAAF0, 0xAAF1 }, // Po   [2] MEETEI MAYEK CHEIKHAN..MEETEI MAYEK AHANG KHUDAM
    Interval{ 0xAAF2, 0xAAF2 }, // Lo       MEETEI MAYEK ANJI
    Interval{ 0xAAF3, 0xAAF4 }, // Lm   [2] MEETEI MAYEK SYLLABLE REPETITION MARK..MEETEI MAYEK WORD REPETITION MARK
    Interval{ 0xAAF5, 0xAAF5 }, // Mc       MEETEI MAYEK VOWEL SIGN VISARGA
    Interval{ 0xAB01, 0xAB06 }, // Lo   [6] ETHIOPIC SYLLABLE TTHU..ETHIOPIC SYLLABLE TTHO
    Interval{ 0xAB09, 0xAB0E }, // Lo   [6] ETHIOPIC SYLLABLE DDHU..ETHIOPIC SYLLABLE DDHO
    Interval{ 0xAB11, 0xAB16 }, // Lo   [6] ETHIOPIC SYLLABLE DZU..ETHIOPIC SYLLABLE DZO
    Interval{ 0xAB20, 0xAB26 }, // Lo   [7] ETHIOPIC SYLLABLE CCHHA..ETHIOPIC SYLLABLE CCHHO
    Interval{ 0xAB28, 0xAB2E }, // Lo   [7] ETHIOPIC SYLLABLE BBA..ETHIOPIC SYLLABLE BBO
    Interval{ 0xAB30, 0xAB5A }, // L&  [43] LATIN SMALL LETTER BARRED ALPHA..LATIN SMALL LETTER Y WITH SHORT RIGHT LEG
    Interval{ 0xAB5B, 0xAB5B }, // Sk       MODIFIER BREVE WITH INVERTED BREVE
    Interval{ 0xAB5C, 0xAB5F }, // Lm   [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    Interval{ 0xAB60, 0xAB68 }, // L&   [9] LATIN SMALL LETTER SAKHA YAT..LATIN SMALL LETTER TURNED R WITH MIDDLE TILDE
    Interval{ 0xAB69, 0xAB69 }, // Lm       MODIFIER LETTER SMALL TURNED W
    Interval{ 0xAB6A, 0xAB6B }, // Sk   [2] MODIFIER LETTER LEFT TACK..MODIFIER LETTER RIGHT TACK
    Interval{ 0xAB70, 0xABBF }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xABC0, 0xABE2 }, // Lo  [35] MEETEI MAYEK LETTER KOK..MEETEI MAYEK LETTER I LONSUM
    Interval{ 0xABE3, 0xABE4 }, // Mc   [2] MEETEI MAYEK VOWEL SIGN ONAP..MEETEI MAYEK VOWEL SIGN INAP
    Interval{ 0xABE6, 0xABE7 }, // Mc   [2] MEETEI MAYEK VOWEL SIGN YENAP..MEETEI MAYEK VOWEL SIGN SOUNAP
    Interval{ 0xABE9, 0xABEA }, // Mc   [2] MEETEI MAYEK VOWEL SIGN CHEINAP..MEETEI MAYEK VOWEL SIGN NUNG
    Interval{ 0xABEB, 0xABEB }, // Po       MEETEI MAYEK CHEIKHEI
    Interval{ 0xABEC, 0xABEC }, // Mc       MEETEI MAYEK LUM IYEK
    Interval{ 0xABF0, 0xABF9 }, // Nd  [10] MEETEI MAYEK DIGIT ZERO..MEETEI MAYEK DIGIT NINE
    Interval{ 0xAC00, 0xD7A3 }, // Lo [11172] HANGUL SYLLABLE GA..HANGUL SYLLABLE HIH
    Interval{ 0xD7B0, 0xD7C6 }, // Lo  [23] HANGUL JUNGSEONG O-YEO..HANGUL JUNGSEONG ARAEA-E
    Interval{ 0xD7CB, 0xD7FB }, // Lo  [49] HANGUL JONGSEONG NIEUN-RIEUL..HANGUL JONGSEONG PHIEUPH-THIEUTH
    Interval{ 0xF900, 0xFA6D }, // Lo [366] CJK COMPATIBILITY IDEOGRAPH-F900..CJK COMPATIBILITY IDEOGRAPH-FA6D
    Interval{ 0xFA70, 0xFAD9 }, // Lo [106] CJK COMPATIBILITY IDEOGRAPH-FA70..CJK COMPATIBILITY IDEOGRAPH-FAD9
    Interval{ 0xFB00, 0xFB06 }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFB1D, 0xFB1D }, // Lo       HEBREW LETTER YOD WITH HIRIQ
    Interval{ 0xFB1F, 0xFB28 }, // Lo  [10] HEBREW LIGATURE YIDDISH YOD YOD PATAH..HEBREW LETTER WIDE TAV
    Interval{ 0xFB29, 0xFB29 }, // Sm       HEBREW LETTER ALTERNATIVE PLUS SIGN
    Interval{ 0xFB2A, 0xFB36 }, // Lo  [13] HEBREW LETTER SHIN WITH SHIN DOT..HEBREW LETTER ZAYIN WITH DAGESH
    Interval{ 0xFB38, 0xFB3C }, // Lo   [5] HEBREW LETTER TET WITH DAGESH..HEBREW LETTER LAMED WITH DAGESH
    Interval{ 0xFB3E, 0xFB3E }, // Lo       HEBREW LETTER MEM WITH DAGESH
    Interval{ 0xFB40, 0xFB41 }, // Lo   [2] HEBREW LETTER NUN WITH DAGESH..HEBREW LETTER SAMEKH WITH DAGESH
    Interval{ 0xFB43, 0xFB44 }, // Lo   [2] HEBREW LETTER FINAL PE WITH DAGESH..HEBREW LETTER PE WITH DAGESH
    Interval{ 0xFB46, 0xFBB1 }, // Lo [108] HEBREW LETTER TSADI WITH DAGESH..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE FINAL FORM
    Interval{ 0xFBB2, 0xFBC2 }, // Sk  [17] ARABIC SYMBOL DOT ABOVE..ARABIC SYMBOL WASLA ABOVE
    Interval{ 0xFBC3, 0xFBD2 }, // So  [16] ARABIC LIGATURE JALLA WA-ALAA..ARABIC LIGATURE ALAYHI AR-RAHMAH
    Interval{ 0xFBD3, 0xFD3D }, // Lo [363] ARABIC LETTER NG ISOLATED FORM..ARABIC LIGATURE ALEF WITH FATHATAN ISOLATED FORM
    Interval{ 0xFD3E, 0xFD3E }, // Pe       ORNATE LEFT PARENTHESIS
    Interval{ 0xFD3F, 0xFD3F }, // Ps       ORNATE RIGHT PARENTHESIS
    Interval{ 0xFD40, 0xFD4F }, // So  [16] ARABIC LIGATURE RAHIMAHU ALLAAH..ARABIC LIGATURE RAHIMAHUM ALLAAH
    Interval{ 0xFD50, 0xFD8F }, // Lo  [64] ARABIC LIGATURE TEH WITH JEEM WITH MEEM INITIAL FORM..ARABIC LIGATURE MEEM WITH KHAH WITH MEEM INITIAL FORM
    Interval{ 0xFD90, 0xFD91 }, // So   [2] ARABIC LIGATURE RAHMATU ALLAAHI ALAYH..ARABIC LIGATURE RAHMATU ALLAAHI ALAYHAA
    Interval{ 0xFD92, 0xFDC7 }, // Lo  [54] ARABIC LIGATURE MEEM WITH JEEM WITH KHAH INITIAL FORM..ARABIC LIGATURE NOON WITH JEEM WITH YEH FINAL FORM
    Interval{ 0xFDC8, 0xFDCF }, // So   [8] ARABIC LIGATURE RAHIMAHU ALLAAH TAAALAA..ARABIC LIGATURE SALAAMUHU ALAYNAA
    Interval{ 0xFDF0, 0xFDFB }, // Lo  [12] ARABIC LIGATURE SALLA USED AS KORANIC STOP SIGN ISOLATED FORM..ARABIC LIGATURE JALLAJALALOUHOU
    Interval{ 0xFDFC, 0xFDFC }, // Sc       RIAL SIGN
    Interval{ 0xFDFD, 0xFDFF }, // So   [3] ARABIC LIGATURE BISMILLAH AR-RAHMAN AR-RAHEEM..ARABIC LIGATURE AZZA WA JALL
    Interval{ 0xFE10, 0xFE16 }, // Po   [7] PRESENTATION FORM FOR VERTICAL COMMA..PRESENTATION FORM FOR VERTICAL QUESTION MARK
    Interval{ 0xFE17, 0xFE17 }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT WHITE LENTICULAR BRACKET
    Interval{ 0xFE18, 0xFE18 }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT WHITE LENTICULAR BRAKCET
    Interval{ 0xFE19, 0xFE19 }, // Po       PRESENTATION FORM FOR VERTICAL HORIZONTAL ELLIPSIS
    Interval{ 0xFE30, 0xFE30 }, // Po       PRESENTATION FORM FOR VERTICAL TWO DOT LEADER
    Interval{ 0xFE31, 0xFE32 }, // Pd   [2] PRESENTATION FORM FOR VERTICAL EM DASH..PRESENTATION FORM FOR VERTICAL EN DASH
    Interval{ 0xFE33, 0xFE34 }, // Pc   [2] PRESENTATION FORM FOR VERTICAL LOW LINE..PRESENTATION FORM FOR VERTICAL WAVY LOW LINE
    Interval{ 0xFE35, 0xFE35 }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT PARENTHESIS
    Interval{ 0xFE36, 0xFE36 }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT PARENTHESIS
    Interval{ 0xFE37, 0xFE37 }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT CURLY BRACKET
    Interval{ 0xFE38, 0xFE38 }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT CURLY BRACKET
    Interval{ 0xFE39, 0xFE39 }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT TORTOISE SHELL BRACKET
    Interval{ 0xFE3A, 0xFE3A }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT TORTOISE SHELL BRACKET
    Interval{ 0xFE3B, 0xFE3B }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT BLACK LENTICULAR BRACKET
    Interval{ 0xFE3C, 0xFE3C }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT BLACK LENTICULAR BRACKET
    Interval{ 0xFE3D, 0xFE3D }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT DOUBLE ANGLE BRACKET
    Interval{ 0xFE3E, 0xFE3E }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT DOUBLE ANGLE BRACKET
    Interval{ 0xFE3F, 0xFE3F }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT ANGLE BRACKET
    Interval{ 0xFE40, 0xFE40 }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT ANGLE BRACKET
    Interval{ 0xFE41, 0xFE41 }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT CORNER BRACKET
    Interval{ 0xFE42, 0xFE42 }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT CORNER BRACKET
    Interval{ 0xFE43, 0xFE43 }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT WHITE CORNER BRACKET
    Interval{ 0xFE44, 0xFE44 }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT WHITE CORNER BRACKET
    Interval{ 0xFE45, 0xFE46 }, // Po   [2] SESAME DOT..WHITE SESAME DOT
    Interval{ 0xFE47, 0xFE47 }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT SQUARE BRACKET
    Interval{ 0xFE48, 0xFE48 }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT SQUARE BRACKET
    Interval{ 0xFE49, 0xFE4C }, // Po   [4] DASHED OVERLINE..DOUBLE WAVY OVERLINE
    Interval{ 0xFE4D, 0xFE4F }, // Pc   [3] DASHED LOW LINE..WAVY LOW LINE
    Interval{ 0xFE50, 0xFE52 }, // Po   [3] SMALL COMMA..SMALL FULL STOP
    Interval{ 0xFE54, 0xFE57 }, // Po   [4] SMALL SEMICOLON..SMALL EXCLAMATION MARK
    Interval{ 0xFE58, 0xFE58 }, // Pd       SMALL EM DASH
    Interval{ 0xFE59, 0xFE59 }, // Ps       SMALL LEFT PARENTHESIS
    Interval{ 0xFE5A, 0xFE5A }, // Pe       SMALL RIGHT PARENTHESIS
    Interval{ 0xFE5B, 0xFE5B }, // Ps       SMALL LEFT CURLY BRACKET
    Interval{ 0xFE5C, 0xFE5C }, // Pe       SMALL RIGHT CURLY BRACKET
    Interval{ 0xFE5D, 0xFE5D }, // Ps       SMALL LEFT TORTOISE SHELL BRACKET
    Interval{ 0xFE5E, 0xFE5E }, // Pe       SMALL RIGHT TORTOISE SHELL BRACKET
    Interval{ 0xFE5F, 0xFE61 }, // Po   [3] SMALL NUMBER SIGN..SMALL ASTERISK
    Interval{ 0xFE62, 0xFE62 }, // Sm       SMALL PLUS SIGN
    Interval{ 0xFE63, 0xFE63 }, // Pd       SMALL HYPHEN-MINUS
    Interval{ 0xFE64, 0xFE66 }, // Sm   [3] SMALL LESS-THAN SIGN..SMALL EQUALS SIGN
    Interval{ 0xFE68, 0xFE68 }, // Po       SMALL REVERSE SOLIDUS
    Interval{ 0xFE69, 0xFE69 }, // Sc       SMALL DOLLAR SIGN
    Interval{ 0xFE6A, 0xFE6B }, // Po   [2] SMALL PERCENT SIGN..SMALL COMMERCIAL AT
    Interval{ 0xFE70, 0xFE74 }, // Lo   [5] ARABIC FATHATAN ISOLATED FORM..ARABIC KASRATAN ISOLATED FORM
    Interval{ 0xFE76, 0xFEFC }, // Lo [135] ARABIC FATHA ISOLATED FORM..ARABIC LIGATURE LAM WITH ALEF FINAL FORM
    Interval{ 0xFF01, 0xFF03 }, // Po   [3] FULLWIDTH EXCLAMATION MARK..FULLWIDTH NUMBER SIGN
    Interval{ 0xFF04, 0xFF04 }, // Sc       FULLWIDTH DOLLAR SIGN
    Interval{ 0xFF05, 0xFF07 }, // Po   [3] FULLWIDTH PERCENT SIGN..FULLWIDTH APOSTROPHE
    Interval{ 0xFF08, 0xFF08 }, // Ps       FULLWIDTH LEFT PARENTHESIS
    Interval{ 0xFF09, 0xFF09 }, // Pe       FULLWIDTH RIGHT PARENTHESIS
    Interval{ 0xFF0A, 0xFF0A }, // Po       FULLWIDTH ASTERISK
    Interval{ 0xFF0B, 0xFF0B }, // Sm       FULLWIDTH PLUS SIGN
    Interval{ 0xFF0C, 0xFF0C }, // Po       FULLWIDTH COMMA
    Interval{ 0xFF0D, 0xFF0D }, // Pd       FULLWIDTH HYPHEN-MINUS
    Interval{ 0xFF0E, 0xFF0F }, // Po   [2] FULLWIDTH FULL STOP..FULLWIDTH SOLIDUS
    Interval{ 0xFF10, 0xFF19 }, // Nd  [10] FULLWIDTH DIGIT ZERO..FULLWIDTH DIGIT NINE
    Interval{ 0xFF1A, 0xFF1B }, // Po   [2] FULLWIDTH COLON..FULLWIDTH SEMICOLON
    Interval{ 0xFF1C, 0xFF1E }, // Sm   [3] FULLWIDTH LESS-THAN SIGN..FULLWIDTH GREATER-THAN SIGN
    Interval{ 0xFF1F, 0xFF20 }, // Po   [2] FULLWIDTH QUESTION MARK..FULLWIDTH COMMERCIAL AT
    Interval{ 0xFF21, 0xFF3A }, // L&  [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Interval{ 0xFF3B, 0xFF3B }, // Ps       FULLWIDTH LEFT SQUARE BRACKET
    Interval{ 0xFF3C, 0xFF3C }, // Po       FULLWIDTH REVERSE SOLIDUS
    Interval{ 0xFF3D, 0xFF3D }, // Pe       FULLWIDTH RIGHT SQUARE BRACKET
    Interval{ 0xFF3E, 0xFF3E }, // Sk       FULLWIDTH CIRCUMFLEX ACCENT
    Interval{ 0xFF3F, 0xFF3F }, // Pc       FULLWIDTH LOW LINE
    Interval{ 0xFF40, 0xFF40 }, // Sk       FULLWIDTH GRAVE ACCENT
    Interval{ 0xFF41, 0xFF5A }, // L&  [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Interval{ 0xFF5B, 0xFF5B }, // Ps       FULLWIDTH LEFT CURLY BRACKET
    Interval{ 0xFF5C, 0xFF5C }, // Sm       FULLWIDTH VERTICAL LINE
    Interval{ 0xFF5D, 0xFF5D }, // Pe       FULLWIDTH RIGHT CURLY BRACKET
    Interval{ 0xFF5E, 0xFF5E }, // Sm       FULLWIDTH TILDE
    Interval{ 0xFF5F, 0xFF5F }, // Ps       FULLWIDTH LEFT WHITE PARENTHESIS
    Interval{ 0xFF60, 0xFF60 }, // Pe       FULLWIDTH RIGHT WHITE PARENTHESIS
    Interval{ 0xFF61, 0xFF61 }, // Po       HALFWIDTH IDEOGRAPHIC FULL STOP
    Interval{ 0xFF62, 0xFF62 }, // Ps       HALFWIDTH LEFT CORNER BRACKET
    Interval{ 0xFF63, 0xFF63 }, // Pe       HALFWIDTH RIGHT CORNER BRACKET
    Interval{ 0xFF64, 0xFF65 }, // Po   [2] HALFWIDTH IDEOGRAPHIC COMMA..HALFWIDTH KATAKANA MIDDLE DOT
    Interval{ 0xFF66, 0xFF6F }, // Lo  [10] HALFWIDTH KATAKANA LETTER WO..HALFWIDTH KATAKANA LETTER SMALL TU
    Interval{ 0xFF70, 0xFF70 }, // Lm       HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Interval{ 0xFF71, 0xFF9D }, // Lo  [45] HALFWIDTH KATAKANA LETTER A..HALFWIDTH KATAKANA LETTER N
    Interval{ 0xFFA0, 0xFFBE }, // Lo  [31] HALFWIDTH HANGUL FILLER..HALFWIDTH HANGUL LETTER HIEUH
    Interval{ 0xFFC2, 0xFFC7 }, // Lo   [6] HALFWIDTH HANGUL LETTER A..HALFWIDTH HANGUL LETTER E
    Interval{ 0xFFCA, 0xFFCF }, // Lo   [6] HALFWIDTH HANGUL LETTER YEO..HALFWIDTH HANGUL LETTER OE
    Interval{ 0xFFD2, 0xFFD7 }, // Lo   [6] HALFWIDTH HANGUL LETTER YO..HALFWIDTH HANGUL LETTER YU
    Interval{ 0xFFDA, 0xFFDC }, // Lo   [3] HALFWIDTH HANGUL LETTER EU..HALFWIDTH HANGUL LETTER I
    Interval{ 0xFFE0, 0xFFE1 }, // Sc   [2] FULLWIDTH CENT SIGN..FULLWIDTH POUND SIGN
    Interval{ 0xFFE2, 0xFFE2 }, // Sm       FULLWIDTH NOT SIGN
    Interval{ 0xFFE3, 0xFFE3 }, // Sk       FULLWIDTH MACRON
    Interval{ 0xFFE4, 0xFFE4 }, // So       FULLWIDTH BROKEN BAR
    Interval{ 0xFFE5, 0xFFE6 }, // Sc   [2] FULLWIDTH YEN SIGN..FULLWIDTH WON SIGN
    Interval{ 0xFFE8, 0xFFE8 }, // So       HALFWIDTH FORMS LIGHT VERTICAL
    Interval{ 0xFFE9, 0xFFEC }, // Sm   [4] HALFWIDTH LEFTWARDS ARROW..HALFWIDTH DOWNWARDS ARROW
    Interval{ 0xFFED, 0xFFEE }, // So   [2] HALFWIDTH BLACK SQUARE..HALFWIDTH WHITE CIRCLE
    Interval{ 0xFFFC, 0xFFFD }, // So   [2] OBJECT REPLACEMENT CHARACTER..REPLACEMENT CHARACTER
    Interval{ 0x10000, 0x1000B }, // Lo  [12] LINEAR B SYLLABLE B008 A..LINEAR B SYLLABLE B046 JE
    Interval{ 0x1000D, 0x10026 }, // Lo  [26] LINEAR B SYLLABLE B036 JO..LINEAR B SYLLABLE B032 QO
    Interval{ 0x10028, 0x1003A }, // Lo  [19] LINEAR B SYLLABLE B060 RA..LINEAR B SYLLABLE B042 WO
    Interval{ 0x1003C, 0x1003D }, // Lo   [2] LINEAR B SYLLABLE B017 ZA..LINEAR B SYLLABLE B074 ZE
    Interval{ 0x1003F, 0x1004D }, // Lo  [15] LINEAR B SYLLABLE B020 ZO..LINEAR B SYLLABLE B091 TWO
    Interval{ 0x10050, 0x1005D }, // Lo  [14] LINEAR B SYMBOL B018..LINEAR B SYMBOL B089
    Interval{ 0x10080, 0x100FA }, // Lo [123] LINEAR B IDEOGRAM B100 MAN..LINEAR B IDEOGRAM VESSEL B305
    Interval{ 0x10100, 0x10102 }, // Po   [3] AEGEAN WORD SEPARATOR LINE..AEGEAN CHECK MARK
    Interval{ 0x10107, 0x10133 }, // No  [45] AEGEAN NUMBER ONE..AEGEAN NUMBER NINETY THOUSAND
    Interval{ 0x10137, 0x1013F }, // So   [9] AEGEAN WEIGHT BASE UNIT..AEGEAN MEASURE THIRD SUBUNIT
    Interval{ 0x10140, 0x10174 }, // Nl  [53] GREEK ACROPHONIC ATTIC ONE QUARTER..GREEK ACROPHONIC STRATIAN FIFTY MNAS
    Interval{ 0x10175, 0x10178 }, // No   [4] GREEK ONE HALF SIGN..GREEK THREE QUARTERS SIGN
    Interval{ 0x10179, 0x10189 }, // So  [17] GREEK YEAR SIGN..GREEK TRYBLION BASE SIGN
    Interval{ 0x1018A, 0x1018B }, // No   [2] GREEK ZERO SIGN..GREEK ONE QUARTER SIGN
    Interval{ 0x1018C, 0x1018E }, // So   [3] GREEK SINUSOID SIGN..NOMISMA SIGN
    Interval{ 0x10190, 0x1019C }, // So  [13] ROMAN SEXTANS SIGN..ASCIA SYMBOL
    Interval{ 0x101A0, 0x101A0 }, // So       GREEK SYMBOL TAU RHO
    Interval{ 0x101D0, 0x101FC }, // So  [45] PHAISTOS DISC SIGN PEDESTRIAN..PHAISTOS DISC SIGN WAVY BAND
    Interval{ 0x10280, 0x1029C }, // Lo  [29] LYCIAN LETTER A..LYCIAN LETTER X
    Interval{ 0x102A0, 0x102D0 }, // Lo  [49] CARIAN LETTER A..CARIAN LETTER UUU3
    Interval{ 0x102E1, 0x102FB }, // No  [27] COPTIC EPACT DIGIT ONE..COPTIC EPACT NUMBER NINE HUNDRED
    Interval{ 0x10300, 0x1031F }, // Lo  [32] OLD ITALIC LETTER A..OLD ITALIC LETTER ESS
    Interval{ 0x10320, 0x10323 }, // No   [4] OLD ITALIC NUMERAL ONE..OLD ITALIC NUMERAL FIFTY
    Interval{ 0x1032D, 0x10340 }, // Lo  [20] OLD ITALIC LETTER YE..GOTHIC LETTER PAIRTHRA
    Interval{ 0x10341, 0x10341 }, // Nl       GOTHIC LETTER NINETY
    Interval{ 0x10342, 0x10349 }, // Lo   [8] GOTHIC LETTER RAIDA..GOTHIC LETTER OTHAL
    Interval{ 0x1034A, 0x1034A }, // Nl       GOTHIC LETTER NINE HUNDRED
    Interval{ 0x10350, 0x10375 }, // Lo  [38] OLD PERMIC LETTER AN..OLD PERMIC LETTER IA
    Interval{ 0x10380, 0x1039D }, // Lo  [30] UGARITIC LETTER ALPA..UGARITIC LETTER SSU
    Interval{ 0x1039F, 0x1039F }, // Po       UGARITIC WORD DIVIDER
    Interval{ 0x103A0, 0x103C3 }, // Lo  [36] OLD PERSIAN SIGN A..OLD PERSIAN SIGN HA
    Interval{ 0x103C8, 0x103CF }, // Lo   [8] OLD PERSIAN SIGN AURAMAZDAA..OLD PERSIAN SIGN BUUMISH
    Interval{ 0x103D0, 0x103D0 }, // Po       OLD PERSIAN WORD DIVIDER
    Interval{ 0x103D1, 0x103D5 }, // Nl   [5] OLD PERSIAN NUMBER ONE..OLD PERSIAN NUMBER HUNDRED
    Interval{ 0x10400, 0x1044F }, // L&  [80] DESERET CAPITAL LETTER LONG I..DESERET SMALL LETTER EW
    Interval{ 0x10450, 0x1049D }, // Lo  [78] SHAVIAN LETTER PEEP..OSMANYA LETTER OO
    Interval{ 0x104A0, 0x104A9 }, // Nd  [10] OSMANYA DIGIT ZERO..OSMANYA DIGIT NINE
    Interval{ 0x104B0, 0x104D3 }, // L&  [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Interval{ 0x104D8, 0x104FB }, // L&  [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Interval{ 0x10500, 0x10527 }, // Lo  [40] ELBASAN LETTER A..ELBASAN LETTER KHE
    Interval{ 0x10530, 0x10563 }, // Lo  [52] CAUCASIAN ALBANIAN LETTER ALT..CAUCASIAN ALBANIAN LETTER KIW
    Interval{ 0x1056F, 0x1056F }, // Po       CAUCASIAN ALBANIAN CITATION MARK
    Interval{ 0x10570, 0x1057A }, // L&  [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Interval{ 0x1057C, 0x1058A }, // L&  [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Interval{ 0x1058C, 0x10592 }, // L&   [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Interval{ 0x10594, 0x10595 }, // L&   [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Interval{ 0x10597, 0x105A1 }, // L&  [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Interval{ 0x105A3, 0x105B1 }, // L&  [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Interval{ 0x105B3, 0x105B9 }, // L&   [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Interval{ 0x105BB, 0x105BC }, // L&   [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Interval{ 0x105C0, 0x105F3 }, // Lo  [52] TODHRI LETTER A..TODHRI LETTER OO
    Interval{ 0x10600, 0x10736 }, // Lo [311] LINEAR A SIGN AB001..LINEAR A SIGN A664
    Interval{ 0x10740, 0x10755 }, // Lo  [22] LINEAR A SIGN A701 A..LINEAR A SIGN A732 JE
    Interval{ 0x10760, 0x10767 }, // Lo   [8] LINEAR A SIGN A800..LINEAR A SIGN A807
    Interval{ 0x10780, 0x10785 }, // Lm   [6] MODIFIER LETTER SMALL CAPITAL AA..MODIFIER LETTER SMALL B WITH HOOK
    Interval{ 0x10787, 0x107B0 }, // Lm  [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Interval{ 0x107B2, 0x107BA }, // Lm   [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Interval{ 0x10800, 0x10805 }, // Lo   [6] CYPRIOT SYLLABLE A..CYPRIOT SYLLABLE JA
    Interval{ 0x10808, 0x10808 }, // Lo       CYPRIOT SYLLABLE JO
    Interval{ 0x1080A, 0x10835 }, // Lo  [44] CYPRIOT SYLLABLE KA..CYPRIOT SYLLABLE WO
    Interval{ 0x10837, 0x10838 }, // Lo   [2] CYPRIOT SYLLABLE XA..CYPRIOT SYLLABLE XE
    Interval{ 0x1083C, 0x1083C }, // Lo       CYPRIOT SYLLABLE ZA
    Interval{ 0x1083F, 0x10855 }, // Lo  [23] CYPRIOT SYLLABLE ZO..IMPERIAL ARAMAIC LETTER TAW
    Interval{ 0x10857, 0x10857 }, // Po       IMPERIAL ARAMAIC SECTION SIGN
    Interval{ 0x10858, 0x1085F }, // No   [8] IMPERIAL ARAMAIC NUMBER ONE..IMPERIAL ARAMAIC NUMBER TEN THOUSAND
    Interval{ 0x10860, 0x10876 }, // Lo  [23] PALMYRENE LETTER ALEPH..PALMYRENE LETTER TAW
    Interval{ 0x10877, 0x10878 }, // So   [2] PALMYRENE LEFT-POINTING FLEURON..PALMYRENE RIGHT-POINTING FLEURON
    Interval{ 0x10879, 0x1087F }, // No   [7] PALMYRENE NUMBER ONE..PALMYRENE NUMBER TWENTY
    Interval{ 0x10880, 0x1089E }, // Lo  [31] NABATAEAN LETTER FINAL ALEPH..NABATAEAN LETTER TAW
    Interval{ 0x108A7, 0x108AF }, // No   [9] NABATAEAN NUMBER ONE..NABATAEAN NUMBER ONE HUNDRED
    Interval{ 0x108E0, 0x108F2 }, // Lo  [19] HATRAN LETTER ALEPH..HATRAN LETTER QOPH
    Interval{ 0x108F4, 0x108F5 }, // Lo   [2] HATRAN LETTER SHIN..HATRAN LETTER TAW
    Interval{ 0x108FB, 0x108FF }, // No   [5] HATRAN NUMBER ONE..HATRAN NUMBER ONE HUNDRED
    Interval{ 0x10900, 0x10915 }, // Lo  [22] PHOENICIAN LETTER ALF..PHOENICIAN LETTER TAU
    Interval{ 0x10916, 0x1091B }, // No   [6] PHOENICIAN NUMBER ONE..PHOENICIAN NUMBER THREE
    Interval{ 0x1091F, 0x1091F }, // Po       PHOENICIAN WORD SEPARATOR
    Interval{ 0x10920, 0x10939 }, // Lo  [26] LYDIAN LETTER A..LYDIAN LETTER C
    Interval{ 0x1093F, 0x1093F }, // Po       LYDIAN TRIANGULAR MARK
    Interval{ 0x10940, 0x10959 }, // Lo  [26] SIDETIC LETTER N01..SIDETIC LETTER N26
    Interval{ 0x10980, 0x109B7 }, // Lo  [56] MEROITIC HIEROGLYPHIC LETTER A..MEROITIC CURSIVE LETTER DA
    Interval{ 0x109BC, 0x109BD }, // No   [2] MEROITIC CURSIVE FRACTION ELEVEN TWELFTHS..MEROITIC CURSIVE FRACTION ONE HALF
    Interval{ 0x109BE, 0x109BF }, // Lo   [2] MEROITIC CURSIVE LOGOGRAM RMT..MEROITIC CURSIVE LOGOGRAM IMN
    Interval{ 0x109C0, 0x109CF }, // No  [16] MEROITIC CURSIVE NUMBER ONE..MEROITIC CURSIVE NUMBER SEVENTY
    Interval{ 0x109D2, 0x109FF }, // No  [46] MEROITIC CURSIVE NUMBER ONE HUNDRED..MEROITIC CURSIVE FRACTION TEN TWELFTHS
    Interval{ 0x10A00, 0x10A00 }, // Lo       KHAROSHTHI LETTER A
    Interval{ 0x10A10, 0x10A13 }, // Lo   [4] KHAROSHTHI LETTER KA..KHAROSHTHI LETTER GHA
    Interval{ 0x10A15, 0x10A17 }, // Lo   [3] KHAROSHTHI LETTER CA..KHAROSHTHI LETTER JA
    Interval{ 0x10A19, 0x10A35 }, // Lo  [29] KHAROSHTHI LETTER NYA..KHAROSHTHI LETTER VHA
    Interval{ 0x10A40, 0x10A48 }, // No   [9] KHAROSHTHI DIGIT ONE..KHAROSHTHI FRACTION ONE HALF
    Interval{ 0x10A50, 0x10A58 }, // Po   [9] KHAROSHTHI PUNCTUATION DOT..KHAROSHTHI PUNCTUATION LINES
    Interval{ 0x10A60, 0x10A7C }, // Lo  [29] OLD SOUTH ARABIAN LETTER HE..OLD SOUTH ARABIAN LETTER THETH
    Interval{ 0x10A7D, 0x10A7E }, // No   [2] OLD SOUTH ARABIAN NUMBER ONE..OLD SOUTH ARABIAN NUMBER FIFTY
    Interval{ 0x10A7F, 0x10A7F }, // Po       OLD SOUTH ARABIAN NUMERIC INDICATOR
    Interval{ 0x10A80, 0x10A9C }, // Lo  [29] OLD NORTH ARABIAN LETTER HEH..OLD NORTH ARABIAN LETTER ZAH
    Interval{ 0x10A9D, 0x10A9F }, // No   [3] OLD NORTH ARABIAN NUMBER ONE..OLD NORTH ARABIAN NUMBER TWENTY
    Interval{ 0x10AC0, 0x10AC7 }, // Lo   [8] MANICHAEAN LETTER ALEPH..MANICHAEAN LETTER WAW
    Interval{ 0x10AC8, 0x10AC8 }, // So       MANICHAEAN SIGN UD
    Interval{ 0x10AC9, 0x10AE4 }, // Lo  [28] MANICHAEAN LETTER ZAYIN..MANICHAEAN LETTER TAW
    Interval{ 0x10AEB, 0x10AEF }, // No   [5] MANICHAEAN NUMBER ONE..MANICHAEAN NUMBER ONE HUNDRED
    Interval{ 0x10AF0, 0x10AF6 }, // Po   [7] MANICHAEAN PUNCTUATION STAR..MANICHAEAN PUNCTUATION LINE FILLER
    Interval{ 0x10B00, 0x10B35 }, // Lo  [54] AVESTAN LETTER A..AVESTAN LETTER HE
    Interval{ 0x10B39, 0x10B3F }, // Po   [7] AVESTAN ABBREVIATION MARK..LARGE ONE RING OVER TWO RINGS PUNCTUATION
    Interval{ 0x10B40, 0x10B55 }, // Lo  [22] INSCRIPTIONAL PARTHIAN LETTER ALEPH..INSCRIPTIONAL PARTHIAN LETTER TAW
    Interval{ 0x10B58, 0x10B5F }, // No   [8] INSCRIPTIONAL PARTHIAN NUMBER ONE..INSCRIPTIONAL PARTHIAN NUMBER ONE THOUSAND
    Interval{ 0x10B60, 0x10B72 }, // Lo  [19] INSCRIPTIONAL PAHLAVI LETTER ALEPH..INSCRIPTIONAL PAHLAVI LETTER TAW
    Interval{ 0x10B78, 0x10B7F }, // No   [8] INSCRIPTIONAL PAHLAVI NUMBER ONE..INSCRIPTIONAL PAHLAVI NUMBER ONE THOUSAND
    Interval{ 0x10B80, 0x10B91 }, // Lo  [18] PSALTER PAHLAVI LETTER ALEPH..PSALTER PAHLAVI LETTER TAW
    Interval{ 0x10B99, 0x10B9C }, // Po   [4] PSALTER PAHLAVI SECTION MARK..PSALTER PAHLAVI FOUR DOTS WITH DOT
    Interval{ 0x10BA9, 0x10BAF }, // No   [7] PSALTER PAHLAVI NUMBER ONE..PSALTER PAHLAVI NUMBER ONE HUNDRED
    Interval{ 0x10C00, 0x10C48 }, // Lo  [73] OLD TURKIC LETTER ORKHON A..OLD TURKIC LETTER ORKHON BASH
    Interval{ 0x10C80, 0x10CB2 }, // L&  [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Interval{ 0x10CC0, 0x10CF2 }, // L&  [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Interval{ 0x10CFA, 0x10CFF }, // No   [6] OLD HUNGARIAN NUMBER ONE..OLD HUNGARIAN NUMBER ONE THOUSAND
    Interval{ 0x10D00, 0x10D23 }, // Lo  [36] HANIFI ROHINGYA LETTER A..HANIFI ROHINGYA MARK NA KHONNA
    Interval{ 0x10D30, 0x10D39 }, // Nd  [10] HANIFI ROHINGYA DIGIT ZERO..HANIFI ROHINGYA DIGIT NINE
    Interval{ 0x10D40, 0x10D49 }, // Nd  [10] GARAY DIGIT ZERO..GARAY DIGIT NINE
    Interval{ 0x10D4A, 0x10D4D }, // Lo   [4] GARAY VOWEL SIGN A..GARAY VOWEL SIGN EE
    Interval{ 0x10D4E, 0x10D4E }, // Lm       GARAY VOWEL LENGTH MARK
    Interval{ 0x10D4F, 0x10D4F }, // Lo       GARAY SUKUN
    Interval{ 0x10D50, 0x10D65 }, // L&  [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Interval{ 0x10D6E, 0x10D6E }, // Pd       GARAY HYPHEN
    Interval{ 0x10D6F, 0x10D6F }, // Lm       GARAY REDUPLICATION MARK
    Interval{ 0x10D70, 0x10D85 }, // L&  [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Interval{ 0x10D8E, 0x10D8F }, // Sm   [2] GARAY PLUS SIGN..GARAY MINUS SIGN
    Interval{ 0x10E60, 0x10E7E }, // No  [31] RUMI DIGIT ONE..RUMI FRACTION TWO THIRDS
    Interval{ 0x10E80, 0x10EA9 }, // Lo  [42] YEZIDI LETTER ELIF..YEZIDI LETTER ET
    Interval{ 0x10EAD, 0x10EAD }, // Pd       YEZIDI HYPHENATION MARK
    Interval{ 0x10EB0, 0x10EB1 }, // Lo   [2] YEZIDI LETTER LAM WITH DOT ABOVE..YEZIDI LETTER YOT WITH CIRCUMFLEX ABOVE
    Interval{ 0x10EC2, 0x10EC4 }, // Lo   [3] ARABIC LETTER DAL WITH TWO DOTS VERTICALLY BELOW..ARABIC LETTER KAF WITH TWO DOTS VERTICALLY BELOW
    Interval{ 0x10EC5, 0x10EC5 }, // Lm       ARABIC SMALL YEH BARREE WITH TWO DOTS BELOW
    Interval{ 0x10EC6, 0x10EC7 }, // Lo   [2] ARABIC LETTER THIN NOON..ARABIC LETTER YEH WITH FOUR DOTS BELOW
    Interval{ 0x10ED0, 0x10ED0 }, // Po       ARABIC BIBLICAL END OF VERSE
    Interval{ 0x10ED1, 0x10ED8 }, // So   [8] ARABIC LIGATURE ALAYHAA AS-SALAATU WAS-SALAAM..ARABIC LIGATURE NAWWARA ALLAAHU MARQADAH
    Interval{ 0x10F00, 0x10F1C }, // Lo  [29] OLD SOGDIAN LETTER ALEPH..OLD SOGDIAN LETTER FINAL TAW WITH VERTICAL TAIL
    Interval{ 0x10F1D, 0x10F26 }, // No  [10] OLD SOGDIAN NUMBER ONE..OLD SOGDIAN FRACTION ONE HALF
    Interval{ 0x10F27, 0x10F27 }, // Lo       OLD SOGDIAN LIGATURE AYIN-DALETH
    Interval{ 0x10F30, 0x10F45 }, // Lo  [22] SOGDIAN LETTER ALEPH..SOGDIAN INDEPENDENT SHIN
    Interval{ 0x10F51, 0x10F54 }, // No   [4] SOGDIAN NUMBER ONE..SOGDIAN NUMBER ONE HUNDRED
    Interval{ 0x10F55, 0x10F59 }, // Po   [5] SOGDIAN PUNCTUATION TWO VERTICAL BARS..SOGDIAN PUNCTUATION HALF CIRCLE WITH DOT
    Interval{ 0x10F70, 0x10F81 }, // Lo  [18] OLD UYGHUR LETTER ALEPH..OLD UYGHUR LETTER LESH
    Interval{ 0x10F86, 0x10F89 }, // Po   [4] OLD UYGHUR PUNCTUATION BAR..OLD UYGHUR PUNCTUATION FOUR DOTS
    Interval{ 0x10FB0, 0x10FC4 }, // Lo  [21] CHORASMIAN LETTER ALEPH..CHORASMIAN LETTER TAW
    Interval{ 0x10FC5, 0x10FCB }, // No   [7] CHORASMIAN NUMBER ONE..CHORASMIAN NUMBER ONE HUNDRED
    Interval{ 0x10FE0, 0x10FF6 }, // Lo  [23] ELYMAIC LETTER ALEPH..ELYMAIC LIGATURE ZAYIN-YODH
    Interval{ 0x11000, 0x11000 }, // Mc       BRAHMI SIGN CANDRABINDU
    Interval{ 0x11002, 0x11002 }, // Mc       BRAHMI SIGN VISARGA
    Interval{ 0x11003, 0x11037 }, // Lo  [53] BRAHMI SIGN JIHVAMULIYA..BRAHMI LETTER OLD TAMIL NNNA
    Interval{ 0x11047, 0x1104D }, // Po   [7] BRAHMI DANDA..BRAHMI PUNCTUATION LOTUS
    Interval{ 0x11052, 0x11065 }, // No  [20] BRAHMI NUMBER ONE..BRAHMI NUMBER ONE THOUSAND
    Interval{ 0x11066, 0x1106F }, // Nd  [10] BRAHMI DIGIT ZERO..BRAHMI DIGIT NINE
    Interval{ 0x11071, 0x11072 }, // Lo   [2] BRAHMI LETTER OLD TAMIL SHORT E..BRAHMI LETTER OLD TAMIL SHORT O
    Interval{ 0x11075, 0x11075 }, // Lo       BRAHMI LETTER OLD TAMIL LLA
    Interval{ 0x11082, 0x11082 }, // Mc       KAITHI SIGN VISARGA
    Interval{ 0x11083, 0x110AF }, // Lo  [45] KAITHI LETTER A..KAITHI LETTER HA
    Interval{ 0x110B0, 0x110B2 }, // Mc   [3] KAITHI VOWEL SIGN AA..KAITHI VOWEL SIGN II
    Interval{ 0x110B7, 0x110B8 }, // Mc   [2] KAITHI VOWEL SIGN O..KAITHI VOWEL SIGN AU
    Interval{ 0x110BB, 0x110BC }, // Po   [2] KAITHI ABBREVIATION SIGN..KAITHI ENUMERATION SIGN
    Interval{ 0x110BE, 0x110C1 }, // Po   [4] KAITHI SECTION MARK..KAITHI DOUBLE DANDA
    Interval{ 0x110D0, 0x110E8 }, // Lo  [25] SORA SOMPENG LETTER SAH..SORA SOMPENG LETTER MAE
    Interval{ 0x110F0, 0x110F9 }, // Nd  [10] SORA SOMPENG DIGIT ZERO..SORA SOMPENG DIGIT NINE
    Interval{ 0x11103, 0x11126 }, // Lo  [36] CHAKMA LETTER AA..CHAKMA LETTER HAA
    Interval{ 0x1112C, 0x1112C }, // Mc       CHAKMA VOWEL SIGN E
    Interval{ 0x11136, 0x1113F }, // Nd  [10] CHAKMA DIGIT ZERO..CHAKMA DIGIT NINE
    Interval{ 0x11140, 0x11143 }, // Po   [4] CHAKMA SECTION MARK..CHAKMA QUESTION MARK
    Interval{ 0x11144, 0x11144 }, // Lo       CHAKMA LETTER LHAA
    Interval{ 0x11145, 0x11146 }, // Mc   [2] CHAKMA VOWEL SIGN AA..CHAKMA VOWEL SIGN EI
    Interval{ 0x11147, 0x11147 }, // Lo       CHAKMA LETTER VAA
    Interval{ 0x11150, 0x11172 }, // Lo  [35] MAHAJANI LETTER A..MAHAJANI LETTER RRA
    Interval{ 0x11174, 0x11175 }, // Po   [2] MAHAJANI ABBREVIATION SIGN..MAHAJANI SECTION MARK
    Interval{ 0x11176, 0x11176 }, // Lo       MAHAJANI LIGATURE SHRI
    Interval{ 0x11182, 0x11182 }, // Mc       SHARADA SIGN VISARGA
    Interval{ 0x11183, 0x111B2 }, // Lo  [48] SHARADA LETTER A..SHARADA LETTER HA
    Interval{ 0x111B3, 0x111B5 }, // Mc   [3] SHARADA VOWEL SIGN AA..SHARADA VOWEL SIGN II
    Interval{ 0x111BF, 0x111BF }, // Mc       SHARADA VOWEL SIGN AU
    Interval{ 0x111C1, 0x111C4 }, // Lo   [4] SHARADA SIGN AVAGRAHA..SHARADA OM
    Interval{ 0x111C5, 0x111C8 }, // Po   [4] SHARADA DANDA..SHARADA SEPARATOR
    Interval{ 0x111CD, 0x111CD }, // Po       SHARADA SUTRA MARK
    Interval{ 0x111CE, 0x111CE }, // Mc       SHARADA VOWEL SIGN PRISHTHAMATRA E
    Interval{ 0x111D0, 0x111D9 }, // Nd  [10] SHARADA DIGIT ZERO..SHARADA DIGIT NINE
    Interval{ 0x111DA, 0x111DA }, // Lo       SHARADA EKAM
    Interval{ 0x111DB, 0x111DB }, // Po       SHARADA SIGN SIDDHAM
    Interval{ 0x111DC, 0x111DC }, // Lo       SHARADA HEADSTROKE
    Interval{ 0x111DD, 0x111DF }, // Po   [3] SHARADA CONTINUATION SIGN..SHARADA SECTION MARK-2
    Interval{ 0x111E1, 0x111F4 }, // No  [20] SINHALA ARCHAIC DIGIT ONE..SINHALA ARCHAIC NUMBER ONE THOUSAND
    Interval{ 0x11200, 0x11211 }, // Lo  [18] KHOJKI LETTER A..KHOJKI LETTER JJA
    Interval{ 0x11213, 0x1122B }, // Lo  [25] KHOJKI LETTER NYA..KHOJKI LETTER LLA
    Interval{ 0x1122C, 0x1122E }, // Mc   [3] KHOJKI VOWEL SIGN AA..KHOJKI VOWEL SIGN II
    Interval{ 0x11232, 0x11233 }, // Mc   [2] KHOJKI VOWEL SIGN O..KHOJKI VOWEL SIGN AU
    Interval{ 0x11238, 0x1123D }, // Po   [6] KHOJKI DANDA..KHOJKI ABBREVIATION SIGN
    Interval{ 0x1123F, 0x11240 }, // Lo   [2] KHOJKI LETTER QA..KHOJKI LETTER SHORT I
    Interval{ 0x11280, 0x11286 }, // Lo   [7] MULTANI LETTER A..MULTANI LETTER GA
    Interval{ 0x11288, 0x11288 }, // Lo       MULTANI LETTER GHA
    Interval{ 0x1128A, 0x1128D }, // Lo   [4] MULTANI LETTER CA..MULTANI LETTER JJA
    Interval{ 0x1128F, 0x1129D }, // Lo  [15] MULTANI LETTER NYA..MULTANI LETTER BA
    Interval{ 0x1129F, 0x112A8 }, // Lo  [10] MULTANI LETTER BHA..MULTANI LETTER RHA
    Interval{ 0x112A9, 0x112A9 }, // Po       MULTANI SECTION MARK
    Interval{ 0x112B0, 0x112DE }, // Lo  [47] KHUDAWADI LETTER A..KHUDAWADI LETTER HA
    Interval{ 0x112E0, 0x112E2 }, // Mc   [3] KHUDAWADI VOWEL SIGN AA..KHUDAWADI VOWEL SIGN II
    Interval{ 0x112F0, 0x112F9 }, // Nd  [10] KHUDAWADI DIGIT ZERO..KHUDAWADI DIGIT NINE
    Interval{ 0x11302, 0x11303 }, // Mc   [2] GRANTHA SIGN ANUSVARA..GRANTHA SIGN VISARGA
    Interval{ 0x11305, 0x1130C }, // Lo   [8] GRANTHA LETTER A..GRANTHA LETTER VOCALIC L
    Interval{ 0x1130F, 0x11310 }, // Lo   [2] GRANTHA LETTER EE..GRANTHA LETTER AI
    Interval{ 0x11313, 0x11328 }, // Lo  [22] GRANTHA LETTER OO..GRANTHA LETTER NA
    Interval{ 0x1132A, 0x11330 }, // Lo   [7] GRANTHA LETTER PA..GRANTHA LETTER RA
    Interval{ 0x11332, 0x11333 }, // Lo   [2] GRANTHA LETTER LA..GRANTHA LETTER LLA
    Interval{ 0x11335, 0x11339 }, // Lo   [5] GRANTHA LETTER VA..GRANTHA LETTER HA
    Interval{ 0x1133D, 0x1133D }, // Lo       GRANTHA SIGN AVAGRAHA
    Interval{ 0x1133F, 0x1133F }, // Mc       GRANTHA VOWEL SIGN I
    Interval{ 0x11341, 0x11344 }, // Mc   [4] GRANTHA VOWEL SIGN U..GRANTHA VOWEL SIGN VOCALIC RR
    Interval{ 0x11347, 0x11348 }, // Mc   [2] GRANTHA VOWEL SIGN EE..GRANTHA VOWEL SIGN AI
    Interval{ 0x1134B, 0x1134C }, // Mc   [2] GRANTHA VOWEL SIGN OO..GRANTHA VOWEL SIGN AU
    Interval{ 0x11350, 0x11350 }, // Lo       GRANTHA OM
    Interval{ 0x1135D, 0x11361 }, // Lo   [5] GRANTHA SIGN PLUTA..GRANTHA LETTER VOCALIC LL
    Interval{ 0x11362, 0x11363 }, // Mc   [2] GRANTHA VOWEL SIGN VOCALIC L..GRANTHA VOWEL SIGN VOCALIC LL
    Interval{ 0x11380, 0x11389 }, // Lo  [10] TULU-TIGALARI LETTER A..TULU-TIGALARI LETTER VOCALIC LL
    Interval{ 0x1138B, 0x1138B }, // Lo       TULU-TIGALARI LETTER EE
    Interval{ 0x1138E, 0x1138E }, // Lo       TULU-TIGALARI LETTER AI
    Interval{ 0x11390, 0x113B5 }, // Lo  [38] TULU-TIGALARI LETTER OO..TULU-TIGALARI LETTER LLLA
    Interval{ 0x113B7, 0x113B7 }, // Lo       TULU-TIGALARI SIGN AVAGRAHA
    Interval{ 0x113B9, 0x113BA }, // Mc   [2] TULU-TIGALARI VOWEL SIGN I..TULU-TIGALARI VOWEL SIGN II
    Interval{ 0x113CA, 0x113CA }, // Mc       TULU-TIGALARI SIGN CANDRA ANUNASIKA
    Interval{ 0x113CC, 0x113CD }, // Mc   [2] TULU-TIGALARI SIGN ANUSVARA..TULU-TIGALARI SIGN VISARGA
    Interval{ 0x113D1, 0x113D1 }, // Lo       TULU-TIGALARI REPHA
    Interval{ 0x113D3, 0x113D3 }, // Lo       TULU-TIGALARI SIGN PLUTA
    Interval{ 0x113D4, 0x113D5 }, // Po   [2] TULU-TIGALARI DANDA..TULU-TIGALARI DOUBLE DANDA
    Interval{ 0x113D7, 0x113D8 }, // Po   [2] TULU-TIGALARI SIGN OM PUSHPIKA..TULU-TIGALARI SIGN SHRII PUSHPIKA
    Interval{ 0x11400, 0x11434 }, // Lo  [53] NEWA LETTER A..NEWA LETTER HA
    Interval{ 0x11435, 0x11437 }, // Mc   [3] NEWA VOWEL SIGN AA..NEWA VOWEL SIGN II
    Interval{ 0x11440, 0x11441 }, // Mc   [2] NEWA VOWEL SIGN O..NEWA VOWEL SIGN AU
    Interval{ 0x11445, 0x11445 }, // Mc       NEWA SIGN VISARGA
    Interval{ 0x11447, 0x1144A }, // Lo   [4] NEWA SIGN AVAGRAHA..NEWA SIDDHI
    Interval{ 0x1144B, 0x1144F }, // Po   [5] NEWA DANDA..NEWA ABBREVIATION SIGN
    Interval{ 0x11450, 0x11459 }, // Nd  [10] NEWA DIGIT ZERO..NEWA DIGIT NINE
    Interval{ 0x1145A, 0x1145B }, // Po   [2] NEWA DOUBLE COMMA..NEWA PLACEHOLDER MARK
    Interval{ 0x1145D, 0x1145D }, // Po       NEWA INSERTION SIGN
    Interval{ 0x1145F, 0x11461 }, // Lo   [3] NEWA LETTER VEDIC ANUSVARA..NEWA SIGN UPADHMANIYA
    Interval{ 0x11480, 0x114AF }, // Lo  [48] TIRHUTA ANJI..TIRHUTA LETTER HA
    Interval{ 0x114B1, 0x114B2 }, // Mc   [2] TIRHUTA VOWEL SIGN I..TIRHUTA VOWEL SIGN II
    Interval{ 0x114B9, 0x114B9 }, // Mc       TIRHUTA VOWEL SIGN E
    Interval{ 0x114BB, 0x114BC }, // Mc   [2] TIRHUTA VOWEL SIGN AI..TIRHUTA VOWEL SIGN O
    Interval{ 0x114BE, 0x114BE }, // Mc       TIRHUTA VOWEL SIGN AU
    Interval{ 0x114C1, 0x114C1 }, // Mc       TIRHUTA SIGN VISARGA
    Interval{ 0x114C4, 0x114C5 }, // Lo   [2] TIRHUTA SIGN AVAGRAHA..TIRHUTA GVANG
    Interval{ 0x114C6, 0x114C6 }, // Po       TIRHUTA ABBREVIATION SIGN
    Interval{ 0x114C7, 0x114C7 }, // Lo       TIRHUTA OM
    Interval{ 0x114D0, 0x114D9 }, // Nd  [10] TIRHUTA DIGIT ZERO..TIRHUTA DIGIT NINE
    Interval{ 0x11580, 0x115AE }, // Lo  [47] SIDDHAM LETTER A..SIDDHAM LETTER HA
    Interval{ 0x115B0, 0x115B1 }, // Mc   [2] SIDDHAM VOWEL SIGN I..SIDDHAM VOWEL SIGN II
    Interval{ 0x115B8, 0x115BB }, // Mc   [4] SIDDHAM VOWEL SIGN E..SIDDHAM VOWEL SIGN AU
    Interval{ 0x115BE, 0x115BE }, // Mc       SIDDHAM SIGN VISARGA
    Interval{ 0x115C1, 0x115D7 }, // Po  [23] SIDDHAM SIGN SIDDHAM..SIDDHAM SECTION MARK WITH CIRCLES AND FOUR ENCLOSURES
    Interval{ 0x115D8, 0x115DB }, // Lo   [4] SIDDHAM LETTER THREE-CIRCLE ALTERNATE I..SIDDHAM LETTER ALTERNATE U
    Interval{ 0x11600, 0x1162F }, // Lo  [48] MODI LETTER A..MODI LETTER LLA
    Interval{ 0x11630, 0x11632 }, // Mc   [3] MODI VOWEL SIGN AA..MODI VOWEL SIGN II
    Interval{ 0x1163B, 0x1163C }, // Mc   [2] MODI VOWEL SIGN O..MODI VOWEL SIGN AU
    Interval{ 0x1163E, 0x1163E }, // Mc       MODI SIGN VISARGA
    Interval{ 0x11641, 0x11643 }, // Po   [3] MODI DANDA..MODI ABBREVIATION SIGN
    Interval{ 0x11644, 0x11644 }, // Lo       MODI SIGN HUVA
    Interval{ 0x11650, 0x11659 }, // Nd  [10] MODI DIGIT ZERO..MODI DIGIT NINE
    Interval{ 0x11660, 0x1166C }, // Po  [13] MONGOLIAN BIRGA WITH ORNAMENT..MONGOLIAN TURNED SWIRL BIRGA WITH DOUBLE ORNAMENT
    Interval{ 0x11680, 0x116AA }, // Lo  [43] TAKRI LETTER A..TAKRI LETTER RRA
    Interval{ 0x116AC, 0x116AC }, // Mc       TAKRI SIGN VISARGA
    Interval{ 0x116AE, 0x116AF }, // Mc   [2] TAKRI VOWEL SIGN I..TAKRI VOWEL SIGN II
    Interval{ 0x116B8, 0x116B8 }, // Lo       TAKRI LETTER ARCHAIC KHA
    Interval{ 0x116B9, 0x116B9 }, // Po       TAKRI ABBREVIATION SIGN
    Interval{ 0x116C0, 0x116C9 }, // Nd  [10] TAKRI DIGIT ZERO..TAKRI DIGIT NINE
    Interval{ 0x116D0, 0x116E3 }, // Nd  [20] MYANMAR PAO DIGIT ZERO..MYANMAR EASTERN PWO KAREN DIGIT NINE
    Interval{ 0x11700, 0x1171A }, // Lo  [27] AHOM LETTER KA..AHOM LETTER ALTERNATE BA
    Interval{ 0x1171E, 0x1171E }, // Mc       AHOM CONSONANT SIGN MEDIAL RA
    Interval{ 0x11720, 0x11721 }, // Mc   [2] AHOM VOWEL SIGN A..AHOM VOWEL SIGN AA
    Interval{ 0x11726, 0x11726 }, // Mc       AHOM VOWEL SIGN E
    Interval{ 0x11730, 0x11739 }, // Nd  [10] AHOM DIGIT ZERO..AHOM DIGIT NINE
    Interval{ 0x1173A, 0x1173B }, // No   [2] AHOM NUMBER TEN..AHOM NUMBER TWENTY
    Interval{ 0x1173C, 0x1173E }, // Po   [3] AHOM SIGN SMALL SECTION..AHOM SIGN RULAI
    Interval{ 0x1173F, 0x1173F }, // So       AHOM SYMBOL VI
    Interval{ 0x11740, 0x11746 }, // Lo   [7] AHOM LETTER CA..AHOM LETTER LLA
    Interval{ 0x11800, 0x1182B }, // Lo  [44] DOGRA LETTER A..DOGRA LETTER RRA
    Interval{ 0x1182C, 0x1182E }, // Mc   [3] DOGRA VOWEL SIGN AA..DOGRA VOWEL SIGN II
    Interval{ 0x11838, 0x11838 }, // Mc       DOGRA SIGN VISARGA
    Interval{ 0x1183B, 0x1183B }, // Po       DOGRA ABBREVIATION SIGN
    Interval{ 0x118A0, 0x118DF }, // L&  [64] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Interval{ 0x118E0, 0x118E9 }, // Nd  [10] WARANG CITI DIGIT ZERO..WARANG CITI DIGIT NINE
    Interval{ 0x118EA, 0x118F2 }, // No   [9] WARANG CITI NUMBER TEN..WARANG CITI NUMBER NINETY
    Interval{ 0x118FF, 0x11906 }, // Lo   [8] WARANG CITI OM..DIVES AKURU LETTER E
    Interval{ 0x11909, 0x11909 }, // Lo       DIVES AKURU LETTER O
    Interval{ 0x1190C, 0x11913 }, // Lo   [8] DIVES AKURU LETTER KA..DIVES AKURU LETTER JA
    Interval{ 0x11915, 0x11916 }, // Lo   [2] DIVES AKURU LETTER NYA..DIVES AKURU LETTER TTA
    Interval{ 0x11918, 0x1192F }, // Lo  [24] DIVES AKURU LETTER DDA..DIVES AKURU LETTER ZA
    Interval{ 0x11931, 0x11935 }, // Mc   [5] DIVES AKURU VOWEL SIGN I..DIVES AKURU VOWEL SIGN E
    Interval{ 0x11937, 0x11938 }, // Mc   [2] DIVES AKURU VOWEL SIGN AI..DIVES AKURU VOWEL SIGN O
    Interval{ 0x1193F, 0x1193F }, // Lo       DIVES AKURU PREFIXED NASAL SIGN
    Interval{ 0x11940, 0x11940 }, // Mc       DIVES AKURU MEDIAL YA
    Interval{ 0x11941, 0x11941 }, // Lo       DIVES AKURU INITIAL RA
    Interval{ 0x11942, 0x11942 }, // Mc       DIVES AKURU MEDIAL RA
    Interval{ 0x11944, 0x11946 }, // Po   [3] DIVES AKURU DOUBLE DANDA..DIVES AKURU END OF TEXT MARK
    Interval{ 0x11950, 0x11959 }, // Nd  [10] DIVES AKURU DIGIT ZERO..DIVES AKURU DIGIT NINE
    Interval{ 0x119A0, 0x119A7 }, // Lo   [8] NANDINAGARI LETTER A..NANDINAGARI LETTER VOCALIC RR
    Interval{ 0x119AA, 0x119D0 }, // Lo  [39] NANDINAGARI LETTER E..NANDINAGARI LETTER RRA
    Interval{ 0x119D1, 0x119D3 }, // Mc   [3] NANDINAGARI VOWEL SIGN AA..NANDINAGARI VOWEL SIGN II
    Interval{ 0x119DC, 0x119DF }, // Mc   [4] NANDINAGARI VOWEL SIGN O..NANDINAGARI SIGN VISARGA
    Interval{ 0x119E1, 0x119E1 }, // Lo       NANDINAGARI SIGN AVAGRAHA
    Interval{ 0x119E2, 0x119E2 }, // Po       NANDINAGARI SIGN SIDDHAM
    Interval{ 0x119E3, 0x119E3 }, // Lo       NANDINAGARI HEADSTROKE
    Interval{ 0x119E4, 0x119E4 }, // Mc       NANDINAGARI VOWEL SIGN PRISHTHAMATRA E
    Interval{ 0x11A00, 0x11A00 }, // Lo       ZANABAZAR SQUARE LETTER A
    Interval{ 0x11A0B, 0x11A32 }, // Lo  [40] ZANABAZAR SQUARE LETTER KA..ZANABAZAR SQUARE LETTER KSSA
    Interval{ 0x11A39, 0x11A39 }, // Mc       ZANABAZAR SQUARE SIGN VISARGA
    Interval{ 0x11A3A, 0x11A3A }, // Lo       ZANABAZAR SQUARE CLUSTER-INITIAL LETTER RA
    Interval{ 0x11A3F, 0x11A46 }, // Po   [8] ZANABAZAR SQUARE INITIAL HEAD MARK..ZANABAZAR SQUARE CLOSING DOUBLE-LINED HEAD MARK
    Interval{ 0x11A50, 0x11A50 }, // Lo       SOYOMBO LETTER A
    Interval{ 0x11A57, 0x11A58 }, // Mc   [2] SOYOMBO VOWEL SIGN AI..SOYOMBO VOWEL SIGN AU
    Interval{ 0x11A5C, 0x11A89 }, // Lo  [46] SOYOMBO LETTER KA..SOYOMBO CLUSTER-INITIAL LETTER SA
    Interval{ 0x11A97, 0x11A97 }, // Mc       SOYOMBO SIGN VISARGA
    Interval{ 0x11A9A, 0x11A9C }, // Po   [3] SOYOMBO MARK TSHEG..SOYOMBO MARK DOUBLE SHAD
    Interval{ 0x11A9D, 0x11A9D }, // Lo       SOYOMBO MARK PLUTA
    Interval{ 0x11A9E, 0x11AA2 }, // Po   [5] SOYOMBO HEAD MARK WITH MOON AND SUN AND TRIPLE FLAME..SOYOMBO TERMINAL MARK-2
    Interval{ 0x11AB0, 0x11AF8 }, // Lo  [73] CANADIAN SYLLABICS NATTILIK HI..PAU CIN HAU GLOTTAL STOP FINAL
    Interval{ 0x11B00, 0x11B09 }, // Po  [10] DEVANAGARI HEAD MARK..DEVANAGARI SIGN MINDU
    Interval{ 0x11B61, 0x11B61 }, // Mc       SHARADA VOWEL SIGN OOE
    Interval{ 0x11B65, 0x11B65 }, // Mc       SHARADA VOWEL SIGN SHORT O
    Interval{ 0x11B67, 0x11B67 }, // Mc       SHARADA VOWEL SIGN CANDRA O
    Interval{ 0x11BC0, 0x11BE0 }, // Lo  [33] SUNUWAR LETTER DEVI..SUNUWAR LETTER KLOKO
    Interval{ 0x11BE1, 0x11BE1 }, // Po       SUNUWAR SIGN PVO
    Interval{ 0x11BF0, 0x11BF9 }, // Nd  [10] SUNUWAR DIGIT ZERO..SUNUWAR DIGIT NINE
    Interval{ 0x11C00, 0x11C08 }, // Lo   [9] BHAIKSUKI LETTER A..BHAIKSUKI LETTER VOCALIC L
    Interval{ 0x11C0A, 0x11C2E }, // Lo  [37] BHAIKSUKI LETTER E..BHAIKSUKI LETTER HA
    Interval{ 0x11C2F, 0x11C2F }, // Mc       BHAIKSUKI VOWEL SIGN AA
    Interval{ 0x11C3E, 0x11C3E }, // Mc       BHAIKSUKI SIGN VISARGA
    Interval{ 0x11C40, 0x11C40 }, // Lo       BHAIKSUKI SIGN AVAGRAHA
    Interval{ 0x11C41, 0x11C45 }, // Po   [5] BHAIKSUKI DANDA..BHAIKSUKI GAP FILLER-2
    Interval{ 0x11C50, 0x11C59 }, // Nd  [10] BHAIKSUKI DIGIT ZERO..BHAIKSUKI DIGIT NINE
    Interval{ 0x11C5A, 0x11C6C }, // No  [19] BHAIKSUKI NUMBER ONE..BHAIKSUKI HUNDREDS UNIT MARK
    Interval{ 0x11C70, 0x11C71 }, // Po   [2] MARCHEN HEAD MARK..MARCHEN MARK SHAD
    Interval{ 0x11C72, 0x11C8F }, // Lo  [30] MARCHEN LETTER KA..MARCHEN LETTER A
    Interval{ 0x11CA9, 0x11CA9 }, // Mc       MARCHEN SUBJOINED LETTER YA
    Interval{ 0x11CB1, 0x11CB1 }, // Mc       MARCHEN VOWEL SIGN I
    Interval{ 0x11CB4, 0x11CB4 }, // Mc       MARCHEN VOWEL SIGN O
    Interval{ 0x11D00, 0x11D06 }, // Lo   [7] MASARAM GONDI LETTER A..MASARAM GONDI LETTER E
    Interval{ 0x11D08, 0x11D09 }, // Lo   [2] MASARAM GONDI LETTER AI..MASARAM GONDI LETTER O
    Interval{ 0x11D0B, 0x11D30 }, // Lo  [38] MASARAM GONDI LETTER AU..MASARAM GONDI LETTER TRA
    Interval{ 0x11D46, 0x11D46 }, // Lo       MASARAM GONDI REPHA
    Interval{ 0x11D50, 0x11D59 }, // Nd  [10] MASARAM GONDI DIGIT ZERO..MASARAM GONDI DIGIT NINE
    Interval{ 0x11D60, 0x11D65 }, // Lo   [6] GUNJALA GONDI LETTER A..GUNJALA GONDI LETTER UU
    Interval{ 0x11D67, 0x11D68 }, // Lo   [2] GUNJALA GONDI LETTER EE..GUNJALA GONDI LETTER AI
    Interval{ 0x11D6A, 0x11D89 }, // Lo  [32] GUNJALA GONDI LETTER OO..GUNJALA GONDI LETTER SA
    Interval{ 0x11D8A, 0x11D8E }, // Mc   [5] GUNJALA GONDI VOWEL SIGN AA..GUNJALA GONDI VOWEL SIGN UU
    Interval{ 0x11D93, 0x11D94 }, // Mc   [2] GUNJALA GONDI VOWEL SIGN OO..GUNJALA GONDI VOWEL SIGN AU
    Interval{ 0x11D96, 0x11D96 }, // Mc       GUNJALA GONDI SIGN VISARGA
    Interval{ 0x11D98, 0x11D98 }, // Lo       GUNJALA GONDI OM
    Interval{ 0x11DA0, 0x11DA9 }, // Nd  [10] GUNJALA GONDI DIGIT ZERO..GUNJALA GONDI DIGIT NINE
    Interval{ 0x11DB0, 0x11DD8 }, // Lo  [41] TOLONG SIKI LETTER I..TOLONG SIKI LETTER RRH
    Interval{ 0x11DD9, 0x11DD9 }, // Lm       TOLONG SIKI SIGN SELA
    Interval{ 0x11DDA, 0x11DDB }, // Lo   [2] TOLONG SIKI SIGN HECAKA..TOLONG SIKI UNGGA
    Interval{ 0x11DE0, 0x11DE9 }, // Nd  [10] TOLONG SIKI DIGIT ZERO..TOLONG SIKI DIGIT NINE
    Interval{ 0x11EE0, 0x11EF2 }, // Lo  [19] MAKASAR LETTER KA..MAKASAR ANGKA
    Interval{ 0x11EF5, 0x11EF6 }, // Mc   [2] MAKASAR VOWEL SIGN E..MAKASAR VOWEL SIGN O
    Interval{ 0x11EF7, 0x11EF8 }, // Po   [2] MAKASAR PASSIMBANG..MAKASAR END OF SECTION
    Interval{ 0x11F02, 0x11F02 }, // Lo       KAWI SIGN REPHA
    Interval{ 0x11F03, 0x11F03 }, // Mc       KAWI SIGN VISARGA
    Interval{ 0x11F04, 0x11F10 }, // Lo  [13] KAWI LETTER A..KAWI LETTER O
    Interval{ 0x11F12, 0x11F33 }, // Lo  [34] KAWI LETTER KA..KAWI LETTER JNYA
    Interval{ 0x11F34, 0x11F35 }, // Mc   [2] KAWI VOWEL SIGN AA..KAWI VOWEL SIGN ALTERNATE AA
    Interval{ 0x11F3E, 0x11F3F }, // Mc   [2] KAWI VOWEL SIGN E..KAWI VOWEL SIGN AI
    Interval{ 0x11F43, 0x11F4F }, // Po  [13] KAWI DANDA..KAWI PUNCTUATION CLOSING SPIRAL
    Interval{ 0x11F50, 0x11F59 }, // Nd  [10] KAWI DIGIT ZERO..KAWI DIGIT NINE
    Interval{ 0x11FB0, 0x11FB0 }, // Lo       LISU LETTER YHA
    Interval{ 0x11FC0, 0x11FD4 }, // No  [21] TAMIL FRACTION ONE THREE-HUNDRED-AND-TWENTIETH..TAMIL FRACTION DOWNSCALING FACTOR KIIZH
    Interval{ 0x11FD5, 0x11FDC }, // So   [8] TAMIL SIGN NEL..TAMIL SIGN MUKKURUNI
    Interval{ 0x11FDD, 0x11FE0 }, // Sc   [4] TAMIL SIGN KAACU..TAMIL SIGN VARAAKAN
    Interval{ 0x11FE1, 0x11FF1 }, // So  [17] TAMIL SIGN PAARAM..TAMIL SIGN VAKAIYARAA
    Interval{ 0x11FFF, 0x11FFF }, // Po       TAMIL PUNCTUATION END OF TEXT
    Interval{ 0x12000, 0x12399 }, // Lo [922] CUNEIFORM SIGN A..CUNEIFORM SIGN U U
    Interval{ 0x12400, 0x1246E }, // Nl [111] CUNEIFORM NUMERIC SIGN TWO ASH..CUNEIFORM NUMERIC SIGN NINE U VARIANT FORM
    Interval{ 0x12470, 0x12474 }, // Po   [5] CUNEIFORM PUNCTUATION SIGN OLD ASSYRIAN WORD DIVIDER..CUNEIFORM PUNCTUATION SIGN DIAGONAL QUADCOLON
    Interval{ 0x12480, 0x12543 }, // Lo [196] CUNEIFORM SIGN AB TIMES NUN TENU..CUNEIFORM SIGN ZU5 TIMES THREE DISH TENU
    Interval{ 0x12F90, 0x12FF0 }, // Lo  [97] CYPRO-MINOAN SIGN CM001..CYPRO-MINOAN SIGN CM114
    Interval{ 0x12FF1, 0x12FF2 }, // Po   [2] CYPRO-MINOAN SIGN CM301..CYPRO-MINOAN SIGN CM302
    Interval{ 0x13000, 0x1342F }, // Lo [1072] EGYPTIAN HIEROGLYPH A001..EGYPTIAN HIEROGLYPH V011D
    Interval{ 0x13441, 0x13446 }, // Lo   [6] EGYPTIAN HIEROGLYPH FULL BLANK..EGYPTIAN HIEROGLYPH WIDE LOST SIGN
    Interval{ 0x13460, 0x143FA }, // Lo [3995] EGYPTIAN HIEROGLYPH-13460..EGYPTIAN HIEROGLYPH-143FA
    Interval{ 0x14400, 0x14646 }, // Lo [583] ANATOLIAN HIEROGLYPH A001..ANATOLIAN HIEROGLYPH A530
    Interval{ 0x16100, 0x1611D }, // Lo  [30] GURUNG KHEMA LETTER A..GURUNG KHEMA LETTER SA
    Interval{ 0x1612A, 0x1612C }, // Mc   [3] GURUNG KHEMA CONSONANT SIGN MEDIAL YA..GURUNG KHEMA CONSONANT SIGN MEDIAL HA
    Interval{ 0x16130, 0x16139 }, // Nd  [10] GURUNG KHEMA DIGIT ZERO..GURUNG KHEMA DIGIT NINE
    Interval{ 0x16800, 0x16A38 }, // Lo [569] BAMUM LETTER PHASE-A NGKUE MFON..BAMUM LETTER PHASE-F VUEQ
    Interval{ 0x16A40, 0x16A5E }, // Lo  [31] MRO LETTER TA..MRO LETTER TEK
    Interval{ 0x16A60, 0x16A69 }, // Nd  [10] MRO DIGIT ZERO..MRO DIGIT NINE
    Interval{ 0x16A6E, 0x16A6F }, // Po   [2] MRO DANDA..MRO DOUBLE DANDA
    Interval{ 0x16A70, 0x16ABE }, // Lo  [79] TANGSA LETTER OZ..TANGSA LETTER ZA
    Interval{ 0x16AC0, 0x16AC9 }, // Nd  [10] TANGSA DIGIT ZERO..TANGSA DIGIT NINE
    Interval{ 0x16AD0, 0x16AED }, // Lo  [30] BASSA VAH LETTER ENNI..BASSA VAH LETTER I
    Interval{ 0x16AF5, 0x16AF5 }, // Po       BASSA VAH FULL STOP
    Interval{ 0x16B00, 0x16B2F }, // Lo  [48] PAHAWH HMONG VOWEL KEEB..PAHAWH HMONG CONSONANT CAU
    Interval{ 0x16B37, 0x16B3B }, // Po   [5] PAHAWH HMONG SIGN VOS THOM..PAHAWH HMONG SIGN VOS FEEM
    Interval{ 0x16B3C, 0x16B3F }, // So   [4] PAHAWH HMONG SIGN XYEEM NTXIV..PAHAWH HMONG SIGN XYEEM FAIB
    Interval{ 0x16B40, 0x16B43 }, // Lm   [4] PAHAWH HMONG SIGN VOS SEEV..PAHAWH HMONG SIGN IB YAM
    Interval{ 0x16B44, 0x16B44 }, // Po       PAHAWH HMONG SIGN XAUS
    Interval{ 0x16B45, 0x16B45 }, // So       PAHAWH HMONG SIGN CIM TSOV ROG
    Interval{ 0x16B50, 0x16B59 }, // Nd  [10] PAHAWH HMONG DIGIT ZERO..PAHAWH HMONG DIGIT NINE
    Interval{ 0x16B5B, 0x16B61 }, // No   [7] PAHAWH HMONG NUMBER TENS..PAHAWH HMONG NUMBER TRILLIONS
    Interval{ 0x16B63, 0x16B77 }, // Lo  [21] PAHAWH HMONG SIGN VOS LUB..PAHAWH HMONG SIGN CIM NRES TOS
    Interval{ 0x16B7D, 0x16B8F }, // Lo  [19] PAHAWH HMONG CLAN SIGN TSHEEJ..PAHAWH HMONG CLAN SIGN VWJ
    Interval{ 0x16D40, 0x16D42 }, // Lm   [3] KIRAT RAI SIGN ANUSVARA..KIRAT RAI SIGN VISARGA
    Interval{ 0x16D43, 0x16D6A }, // Lo  [40] KIRAT RAI LETTER A..KIRAT RAI VOWEL SIGN AU
    Interval{ 0x16D6B, 0x16D6C }, // Lm   [2] KIRAT RAI SIGN VIRAMA..KIRAT RAI SIGN SAAT
    Interval{ 0x16D6D, 0x16D6F }, // Po   [3] KIRAT RAI SIGN YUPI..KIRAT RAI DOUBLE DANDA
    Interval{ 0x16D70, 0x16D79 }, // Nd  [10] KIRAT RAI DIGIT ZERO..KIRAT RAI DIGIT NINE
    Interval{ 0x16E40, 0x16E7F }, // L&  [64] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Interval{ 0x16E80, 0x16E96 }, // No  [23] MEDEFAIDRIN DIGIT ZERO..MEDEFAIDRIN DIGIT THREE ALTERNATE FORM
    Interval{ 0x16E97, 0x16E9A }, // Po   [4] MEDEFAIDRIN COMMA..MEDEFAIDRIN EXCLAMATION OH
    Interval{ 0x16EA0, 0x16EB8 }, // L&  [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Interval{ 0x16EBB, 0x16ED3 }, // L&  [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Interval{ 0x16F00, 0x16F4A }, // Lo  [75] MIAO LETTER PA..MIAO LETTER RTE
    Interval{ 0x16F50, 0x16F50 }, // Lo       MIAO LETTER NASALIZATION
    Interval{ 0x16F51, 0x16F87 }, // Mc  [55] MIAO SIGN ASPIRATION..MIAO VOWEL SIGN UI
    Interval{ 0x16F93, 0x16F9F }, // Lm  [13] MIAO LETTER TONE-2..MIAO LETTER REFORMED TONE-8
    Interval{ 0x16FE0, 0x16FE1 }, // Lm   [2] TANGUT ITERATION MARK..NUSHU ITERATION MARK
    Interval{ 0x16FE2, 0x16FE2 }, // Po       OLD CHINESE HOOK MARK
    Interval{ 0x16FE3, 0x16FE3 }, // Lm       OLD CHINESE ITERATION MARK
    Interval{ 0x16FF2, 0x16FF3 }, // Lm   [2] CHINESE SMALL SIMPLIFIED ER..CHINESE SMALL TRADITIONAL ER
    Interval{ 0x16FF4, 0x16FF6 }, // Nl   [3] YANGQIN SIGN SLOW ONE BEAT..YANGQIN SIGN SLOW TWO BEATS
    Interval{ 0x17000, 0x18CD5 }, // Lo [7382] TANGUT IDEOGRAPH-17000..KHITAN SMALL SCRIPT CHARACTER-18CD5
    Interval{ 0x18CFF, 0x18D1E }, // Lo  [32] KHITAN SMALL SCRIPT CHARACTER-18CFF..TANGUT IDEOGRAPH-18D1E
    Interval{ 0x18D80, 0x18DF2 }, // Lo [115] TANGUT COMPONENT-769..TANGUT COMPONENT-883
    Interval{ 0x1AFF0, 0x1AFF3 }, // Lm   [4] KATAKANA LETTER MINNAN TONE-2..KATAKANA LETTER MINNAN TONE-5
    Interval{ 0x1AFF5, 0x1AFFB }, // Lm   [7] KATAKANA LETTER MINNAN TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-5
    Interval{ 0x1AFFD, 0x1AFFE }, // Lm   [2] KATAKANA LETTER MINNAN NASALIZED TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-8
    Interval{ 0x1B000, 0x1B122 }, // Lo [291] KATAKANA LETTER ARCHAIC E..KATAKANA LETTER ARCHAIC WU
    Interval{ 0x1B132, 0x1B132 }, // Lo       HIRAGANA LETTER SMALL KO
    Interval{ 0x1B150, 0x1B152 }, // Lo   [3] HIRAGANA LETTER SMALL WI..HIRAGANA LETTER SMALL WO
    Interval{ 0x1B155, 0x1B155 }, // Lo       KATAKANA LETTER SMALL KO
    Interval{ 0x1B164, 0x1B167 }, // Lo   [4] KATAKANA LETTER SMALL WI..KATAKANA LETTER SMALL N
    Interval{ 0x1B170, 0x1B2FB }, // Lo [396] NUSHU CHARACTER-1B170..NUSHU CHARACTER-1B2FB
    Interval{ 0x1BC00, 0x1BC6A }, // Lo [107] DUPLOYAN LETTER H..DUPLOYAN LETTER VOCALIC M
    Interval{ 0x1BC70, 0x1BC7C }, // Lo  [13] DUPLOYAN AFFIX LEFT HORIZONTAL SECANT..DUPLOYAN AFFIX ATTACHED TANGENT HOOK
    Interval{ 0x1BC80, 0x1BC88 }, // Lo   [9] DUPLOYAN AFFIX HIGH ACUTE..DUPLOYAN AFFIX HIGH VERTICAL
    Interval{ 0x1BC90, 0x1BC99 }, // Lo  [10] DUPLOYAN AFFIX LOW ACUTE..DUPLOYAN AFFIX LOW ARROW
    Interval{ 0x1BC9C, 0x1BC9C }, // So       DUPLOYAN SIGN O WITH CROSS
    Interval{ 0x1BC9F, 0x1BC9F }, // Po       DUPLOYAN PUNCTUATION CHINOOK FULL STOP
    Interval{ 0x1CC00, 0x1CCEF }, // So [240] UP-POINTING GO-KART..OUTLINED LATIN CAPITAL LETTER Z
    Interval{ 0x1CCF0, 0x1CCF9 }, // Nd  [10] OUTLINED DIGIT ZERO..OUTLINED DIGIT NINE
    Interval{ 0x1CCFA, 0x1CCFC }, // So   [3] SNAKE SYMBOL..NOSE SYMBOL
    Interval{ 0x1CD00, 0x1CEB3 }, // So [436] BLOCK OCTANT-3..BLACK RIGHT TRIANGLE CARET
    Interval{ 0x1CEBA, 0x1CED0 }, // So  [23] FRAGILE SYMBOL..LEUKOTHEA
    Interval{ 0x1CEE0, 0x1CEEF }, // So  [16] GEOMANTIC FIGURE POPULUS..GEOMANTIC FIGURE VIA
    Interval{ 0x1CEF0, 0x1CEF0 }, // Sm       MEDIUM SMALL WHITE CIRCLE WITH HORIZONTAL BAR
    Interval{ 0x1CF50, 0x1CFC3 }, // So [116] ZNAMENNY NEUME KRYUK..ZNAMENNY NEUME PAUK
    Interval{ 0x1D000, 0x1D0F5 }, // So [246] BYZANTINE MUSICAL SYMBOL PSILI..BYZANTINE MUSICAL SYMBOL GORGON NEO KATO
    Interval{ 0x1D100, 0x1D126 }, // So  [39] MUSICAL SYMBOL SINGLE BARLINE..MUSICAL SYMBOL DRUM CLEF-2
    Interval{ 0x1D129, 0x1D164 }, // So  [60] MUSICAL SYMBOL MULTIPLE MEASURE REST..MUSICAL SYMBOL ONE HUNDRED TWENTY-EIGHTH NOTE
    Interval{ 0x1D16A, 0x1D16C }, // So   [3] MUSICAL SYMBOL FINGERED TREMOLO-1..MUSICAL SYMBOL FINGERED TREMOLO-3
    Interval{ 0x1D183, 0x1D184 }, // So   [2] MUSICAL SYMBOL ARPEGGIATO UP..MUSICAL SYMBOL ARPEGGIATO DOWN
    Interval{ 0x1D18C, 0x1D1A9 }, // So  [30] MUSICAL SYMBOL RINFORZANDO..MUSICAL SYMBOL DEGREE SLASH
    Interval{ 0x1D1AE, 0x1D1EA }, // So  [61] MUSICAL SYMBOL PEDAL MARK..MUSICAL SYMBOL KORON
    Interval{ 0x1D200, 0x1D241 }, // So  [66] GREEK VOCAL NOTATION SYMBOL-1..GREEK INSTRUMENTAL NOTATION SYMBOL-54
    Interval{ 0x1D245, 0x1D245 }, // So       GREEK MUSICAL LEIMMA
    Interval{ 0x1D2C0, 0x1D2D3 }, // No  [20] KAKTOVIK NUMERAL ZERO..KAKTOVIK NUMERAL NINETEEN
    Interval{ 0x1D2E0, 0x1D2F3 }, // No  [20] MAYAN NUMERAL ZERO..MAYAN NUMERAL NINETEEN
    Interval{ 0x1D300, 0x1D356 }, // So  [87] MONOGRAM FOR EARTH..TETRAGRAM FOR FOSTERING
    Interval{ 0x1D360, 0x1D378 }, // No  [25] COUNTING ROD UNIT DIGIT ONE..TALLY MARK FIVE
    Interval{ 0x1D400, 0x1D454 }, // L&  [85] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL ITALIC SMALL G
    Interval{ 0x1D456, 0x1D49C }, // L&  [71] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL SCRIPT CAPITAL A
    Interval{ 0x1D49E, 0x1D49F }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Interval{ 0x1D4A2, 0x1D4A2 }, // L&       MATHEMATICAL SCRIPT CAPITAL G
    Interval{ 0x1D4A5, 0x1D4A6 }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Interval{ 0x1D4A9, 0x1D4AC }, // L&   [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Interval{ 0x1D4AE, 0x1D4B9 }, // L&  [12] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT SMALL D
    Interval{ 0x1D4BB, 0x1D4BB }, // L&       MATHEMATICAL SCRIPT SMALL F
    Interval{ 0x1D4BD, 0x1D4C3 }, // L&   [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Interval{ 0x1D4C5, 0x1D505 }, // L&  [65] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL FRAKTUR CAPITAL B
    Interval{ 0x1D507, 0x1D50A }, // L&   [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Interval{ 0x1D50D, 0x1D514 }, // L&   [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Interval{ 0x1D516, 0x1D51C }, // L&   [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Interval{ 0x1D51E, 0x1D539 }, // L&  [28] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Interval{ 0x1D53B, 0x1D53E }, // L&   [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Interval{ 0x1D540, 0x1D544 }, // L&   [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Interval{ 0x1D546, 0x1D546 }, // L&       MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Interval{ 0x1D54A, 0x1D550 }, // L&   [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Interval{ 0x1D552, 0x1D6A5 }, // L& [340] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    Interval{ 0x1D6A8, 0x1D6C0 }, // L&  [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    Interval{ 0x1D6C1, 0x1D6C1 }, // Sm       MATHEMATICAL BOLD NABLA
    Interval{ 0x1D6C2, 0x1D6DA }, // L&  [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    Interval{ 0x1D6DB, 0x1D6DB }, // Sm       MATHEMATICAL BOLD PARTIAL DIFFERENTIAL
    Interval{ 0x1D6DC, 0x1D6FA }, // L&  [31] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL ITALIC CAPITAL OMEGA
    Interval{ 0x1D6FB, 0x1D6FB }, // Sm       MATHEMATICAL ITALIC NABLA
    Interval{ 0x1D6FC, 0x1D714 }, // L&  [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    Interval{ 0x1D715, 0x1D715 }, // Sm       MATHEMATICAL ITALIC PARTIAL DIFFERENTIAL
    Interval{ 0x1D716, 0x1D734 }, // L&  [31] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D735, 0x1D735 }, // Sm       MATHEMATICAL BOLD ITALIC NABLA
    Interval{ 0x1D736, 0x1D74E }, // L&  [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D74F, 0x1D74F }, // Sm       MATHEMATICAL BOLD ITALIC PARTIAL DIFFERENTIAL
    Interval{ 0x1D750, 0x1D76E }, // L&  [31] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    Interval{ 0x1D76F, 0x1D76F }, // Sm       MATHEMATICAL SANS-SERIF BOLD NABLA
    Interval{ 0x1D770, 0x1D788 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    Interval{ 0x1D789, 0x1D789 }, // Sm       MATHEMATICAL SANS-SERIF BOLD PARTIAL DIFFERENTIAL
    Interval{ 0x1D78A, 0x1D7A8 }, // L&  [31] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D7A9, 0x1D7A9 }, // Sm       MATHEMATICAL SANS-SERIF BOLD ITALIC NABLA
    Interval{ 0x1D7AA, 0x1D7C2 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D7C3, 0x1D7C3 }, // Sm       MATHEMATICAL SANS-SERIF BOLD ITALIC PARTIAL DIFFERENTIAL
    Interval{ 0x1D7C4, 0x1D7CB }, // L&   [8] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD SMALL DIGAMMA
    Interval{ 0x1D7CE, 0x1D7FF }, // Nd  [50] MATHEMATICAL BOLD DIGIT ZERO..MATHEMATICAL MONOSPACE DIGIT NINE
    Interval{ 0x1D800, 0x1D9FF }, // So [512] SIGNWRITING HAND-FIST INDEX..SIGNWRITING HEAD
    Interval{ 0x1DA37, 0x1DA3A }, // So   [4] SIGNWRITING AIR BLOW SMALL ROTATIONS..SIGNWRITING BREATH EXHALE
    Interval{ 0x1DA6D, 0x1DA74 }, // So   [8] SIGNWRITING SHOULDER HIP SPINE..SIGNWRITING TORSO-FLOORPLANE TWISTING
    Interval{ 0x1DA76, 0x1DA83 }, // So  [14] SIGNWRITING LIMB COMBINATION..SIGNWRITING LOCATION DEPTH
    Interval{ 0x1DA85, 0x1DA86 }, // So   [2] SIGNWRITING LOCATION TORSO..SIGNWRITING LOCATION LIMBS DIGITS
    Interval{ 0x1DA87, 0x1DA8B }, // Po   [5] SIGNWRITING COMMA..SIGNWRITING PARENTHESIS
    Interval{ 0x1DF00, 0x1DF09 }, // L&  [10] LATIN SMALL LETTER FENG DIGRAPH WITH TRILL..LATIN SMALL LETTER T WITH HOOK AND RETROFLEX HOOK
    Interval{ 0x1DF0A, 0x1DF0A }, // Lo       LATIN LETTER RETROFLEX CLICK WITH RETROFLEX HOOK
    Interval{ 0x1DF0B, 0x1DF1E }, // L&  [20] LATIN SMALL LETTER ESH WITH DOUBLE BAR..LATIN SMALL LETTER S WITH CURL
    Interval{ 0x1DF25, 0x1DF2A }, // L&   [6] LATIN SMALL LETTER D WITH MID-HEIGHT LEFT HOOK..LATIN SMALL LETTER T WITH MID-HEIGHT LEFT HOOK
    Interval{ 0x1E030, 0x1E06D }, // Lm  [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Interval{ 0x1E100, 0x1E12C }, // Lo  [45] NYIAKENG PUACHUE HMONG LETTER MA..NYIAKENG PUACHUE HMONG LETTER W
    Interval{ 0x1E137, 0x1E13D }, // Lm   [7] NYIAKENG PUACHUE HMONG SIGN FOR PERSON..NYIAKENG PUACHUE HMONG SYLLABLE LENGTHENER
    Interval{ 0x1E140, 0x1E149 }, // Nd  [10] NYIAKENG PUACHUE HMONG DIGIT ZERO..NYIAKENG PUACHUE HMONG DIGIT NINE
    Interval{ 0x1E14E, 0x1E14E }, // Lo       NYIAKENG PUACHUE HMONG LOGOGRAM NYAJ
    Interval{ 0x1E14F, 0x1E14F }, // So       NYIAKENG PUACHUE HMONG CIRCLED CA
    Interval{ 0x1E290, 0x1E2AD }, // Lo  [30] TOTO LETTER PA..TOTO LETTER A
    Interval{ 0x1E2C0, 0x1E2EB }, // Lo  [44] WANCHO LETTER AA..WANCHO LETTER YIH
    Interval{ 0x1E2F0, 0x1E2F9 }, // Nd  [10] WANCHO DIGIT ZERO..WANCHO DIGIT NINE
    Interval{ 0x1E2FF, 0x1E2FF }, // Sc       WANCHO NGUN SIGN
    Interval{ 0x1E4D0, 0x1E4EA }, // Lo  [27] NAG MUNDARI LETTER O..NAG MUNDARI LETTER ELL
    Interval{ 0x1E4EB, 0x1E4EB }, // Lm       NAG MUNDARI SIGN OJOD
    Interval{ 0x1E4F0, 0x1E4F9 }, // Nd  [10] NAG MUNDARI DIGIT ZERO..NAG MUNDARI DIGIT NINE
    Interval{ 0x1E5D0, 0x1E5ED }, // Lo  [30] OL ONAL LETTER O..OL ONAL LETTER EG
    Interval{ 0x1E5F0, 0x1E5F0 }, // Lo       OL ONAL SIGN HODDOND
    Interval{ 0x1E5F1, 0x1E5FA }, // Nd  [10] OL ONAL DIGIT ZERO..OL ONAL DIGIT NINE
    Interval{ 0x1E5FF, 0x1E5FF }, // Po       OL ONAL ABBREVIATION SIGN
    Interval{ 0x1E6C0, 0x1E6DE }, // Lo  [31] TAI YO LETTER LOW KO..TAI YO LETTER HIGH KVO
    Interval{ 0x1E6E0, 0x1E6E2 }, // Lo   [3] TAI YO LETTER AA..TAI YO LETTER UE
    Interval{ 0x1E6E4, 0x1E6E5 }, // Lo   [2] TAI YO LETTER U..TAI YO LETTER AE
    Interval{ 0x1E6E7, 0x1E6ED }, // Lo   [7] TAI YO LETTER O..TAI YO LETTER AUE
    Interval{ 0x1E6F0, 0x1E6F4 }, // Lo   [5] TAI YO LETTER AN..TAI YO LETTER AP
    Interval{ 0x1E6FE, 0x1E6FE }, // Lo       TAI YO SYMBOL MUEANG
    Interval{ 0x1E6FF, 0x1E6FF }, // Lm       TAI YO XAM LAI
    Interval{ 0x1E7E0, 0x1E7E6 }, // Lo   [7] ETHIOPIC SYLLABLE HHYA..ETHIOPIC SYLLABLE HHYO
    Interval{ 0x1E7E8, 0x1E7EB }, // Lo   [4] ETHIOPIC SYLLABLE GURAGE HHWA..ETHIOPIC SYLLABLE HHWE
    Interval{ 0x1E7ED, 0x1E7EE }, // Lo   [2] ETHIOPIC SYLLABLE GURAGE MWI..ETHIOPIC SYLLABLE GURAGE MWEE
    Interval{ 0x1E7F0, 0x1E7FE }, // Lo  [15] ETHIOPIC SYLLABLE GURAGE QWI..ETHIOPIC SYLLABLE GURAGE PWEE
    Interval{ 0x1E800, 0x1E8C4 }, // Lo [197] MENDE KIKAKUI SYLLABLE M001 KI..MENDE KIKAKUI SYLLABLE M060 NYON
    Interval{ 0x1E8C7, 0x1E8CF }, // No   [9] MENDE KIKAKUI DIGIT ONE..MENDE KIKAKUI DIGIT NINE
    Interval{ 0x1E900, 0x1E943 }, // L&  [68] ADLAM CAPITAL LETTER ALIF..ADLAM SMALL LETTER SHA
    Interval{ 0x1E94B, 0x1E94B }, // Lm       ADLAM NASALIZATION MARK
    Interval{ 0x1E950, 0x1E959 }, // Nd  [10] ADLAM DIGIT ZERO..ADLAM DIGIT NINE
    Interval{ 0x1E95E, 0x1E95F }, // Po   [2] ADLAM INITIAL EXCLAMATION MARK..ADLAM INITIAL QUESTION MARK
    Interval{ 0x1EC71, 0x1ECAB }, // No  [59] INDIC SIYAQ NUMBER ONE..INDIC SIYAQ NUMBER PREFIXED NINE
    Interval{ 0x1ECAC, 0x1ECAC }, // So       INDIC SIYAQ PLACEHOLDER
    Interval{ 0x1ECAD, 0x1ECAF }, // No   [3] INDIC SIYAQ FRACTION ONE QUARTER..INDIC SIYAQ FRACTION THREE QUARTERS
    Interval{ 0x1ECB0, 0x1ECB0 }, // Sc       INDIC SIYAQ RUPEE MARK
    Interval{ 0x1ECB1, 0x1ECB4 }, // No   [4] INDIC SIYAQ NUMBER ALTERNATE ONE..INDIC SIYAQ ALTERNATE LAKH MARK
    Interval{ 0x1ED01, 0x1ED2D }, // No  [45] OTTOMAN SIYAQ NUMBER ONE..OTTOMAN SIYAQ NUMBER NINETY THOUSAND
    Interval{ 0x1ED2E, 0x1ED2E }, // So       OTTOMAN SIYAQ MARRATAN
    Interval{ 0x1ED2F, 0x1ED3D }, // No  [15] OTTOMAN SIYAQ ALTERNATE NUMBER TWO..OTTOMAN SIYAQ FRACTION ONE SIXTH
    Interval{ 0x1EE00, 0x1EE03 }, // Lo   [4] ARABIC MATHEMATICAL ALEF..ARABIC MATHEMATICAL DAL
    Interval{ 0x1EE05, 0x1EE1F }, // Lo  [27] ARABIC MATHEMATICAL WAW..ARABIC MATHEMATICAL DOTLESS QAF
    Interval{ 0x1EE21, 0x1EE22 }, // Lo   [2] ARABIC MATHEMATICAL INITIAL BEH..ARABIC MATHEMATICAL INITIAL JEEM
    Interval{ 0x1EE24, 0x1EE24 }, // Lo       ARABIC MATHEMATICAL INITIAL HEH
    Interval{ 0x1EE27, 0x1EE27 }, // Lo       ARABIC MATHEMATICAL INITIAL HAH
    Interval{ 0x1EE29, 0x1EE32 }, // Lo  [10] ARABIC MATHEMATICAL INITIAL YEH..ARABIC MATHEMATICAL INITIAL QAF
    Interval{ 0x1EE34, 0x1EE37 }, // Lo   [4] ARABIC MATHEMATICAL INITIAL SHEEN..ARABIC MATHEMATICAL INITIAL KHAH
    Interval{ 0x1EE39, 0x1EE39 }, // Lo       ARABIC MATHEMATICAL INITIAL DAD
    Interval{ 0x1EE3B, 0x1EE3B }, // Lo       ARABIC MATHEMATICAL INITIAL GHAIN
    Interval{ 0x1EE42, 0x1EE42 }, // Lo       ARABIC MATHEMATICAL TAILED JEEM
    Interval{ 0x1EE47, 0x1EE47 }, // Lo       ARABIC MATHEMATICAL TAILED HAH
    Interval{ 0x1EE49, 0x1EE49 }, // Lo       ARABIC MATHEMATICAL TAILED YEH
    Interval{ 0x1EE4B, 0x1EE4B }, // Lo       ARABIC MATHEMATICAL TAILED LAM
    Interval{ 0x1EE4D, 0x1EE4F }, // Lo   [3] ARABIC MATHEMATICAL TAILED NOON..ARABIC MATHEMATICAL TAILED AIN
    Interval{ 0x1EE51, 0x1EE52 }, // Lo   [2] ARABIC MATHEMATICAL TAILED SAD..ARABIC MATHEMATICAL TAILED QAF
    Interval{ 0x1EE54, 0x1EE54 }, // Lo       ARABIC MATHEMATICAL TAILED SHEEN
    Interval{ 0x1EE57, 0x1EE57 }, // Lo       ARABIC MATHEMATICAL TAILED KHAH
    Interval{ 0x1EE59, 0x1EE59 }, // Lo       ARABIC MATHEMATICAL TAILED DAD
    Interval{ 0x1EE5B, 0x1EE5B }, // Lo       ARABIC MATHEMATICAL TAILED GHAIN
    Interval{ 0x1EE5D, 0x1EE5D }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS NOON
    Interval{ 0x1EE5F, 0x1EE5F }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS QAF
    Interval{ 0x1EE61, 0x1EE62 }, // Lo   [2] ARABIC MATHEMATICAL STRETCHED BEH..ARABIC MATHEMATICAL STRETCHED JEEM
    Interval{ 0x1EE64, 0x1EE64 }, // Lo       ARABIC MATHEMATICAL STRETCHED HEH
    Interval{ 0x1EE67, 0x1EE6A }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED HAH..ARABIC MATHEMATICAL STRETCHED KAF
    Interval{ 0x1EE6C, 0x1EE72 }, // Lo   [7] ARABIC MATHEMATICAL STRETCHED MEEM..ARABIC MATHEMATICAL STRETCHED QAF
    Interval{ 0x1EE74, 0x1EE77 }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED SHEEN..ARABIC MATHEMATICAL STRETCHED KHAH
    Interval{ 0x1EE79, 0x1EE7C }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED DAD..ARABIC MATHEMATICAL STRETCHED DOTLESS BEH
    Interval{ 0x1EE7E, 0x1EE7E }, // Lo       ARABIC MATHEMATICAL STRETCHED DOTLESS FEH
    Interval{ 0x1EE80, 0x1EE89 }, // Lo  [10] ARABIC MATHEMATICAL LOOPED ALEF..ARABIC MATHEMATICAL LOOPED YEH
    Interval{ 0x1EE8B, 0x1EE9B }, // Lo  [17] ARABIC MATHEMATICAL LOOPED LAM..ARABIC MATHEMATICAL LOOPED GHAIN
    Interval{ 0x1EEA1, 0x1EEA3 }, // Lo   [3] ARABIC MATHEMATICAL DOUBLE-STRUCK BEH..ARABIC MATHEMATICAL DOUBLE-STRUCK DAL
    Interval{ 0x1EEA5, 0x1EEA9 }, // Lo   [5] ARABIC MATHEMATICAL DOUBLE-STRUCK WAW..ARABIC MATHEMATICAL DOUBLE-STRUCK YEH
    Interval{ 0x1EEAB, 0x1EEBB }, // Lo  [17] ARABIC MATHEMATICAL DOUBLE-STRUCK LAM..ARABIC MATHEMATICAL DOUBLE-STRUCK GHAIN
    Interval{ 0x1EEF0, 0x1EEF1 }, // Sm   [2] ARABIC MATHEMATICAL OPERATOR MEEM WITH HAH WITH TATWEEL..ARABIC MATHEMATICAL OPERATOR HAH WITH DAL
    Interval{ 0x1F000, 0x1F02B }, // So  [44] MAHJONG TILE EAST WIND..MAHJONG TILE BACK
    Interval{ 0x1F030, 0x1F093 }, // So [100] DOMINO TILE HORIZONTAL BACK..DOMINO TILE VERTICAL-06-06
    Interval{ 0x1F0A0, 0x1F0AE }, // So  [15] PLAYING CARD BACK..PLAYING CARD KING OF SPADES
    Interval{ 0x1F0B1, 0x1F0BF }, // So  [15] PLAYING CARD ACE OF HEARTS..PLAYING CARD RED JOKER
    Interval{ 0x1F0C1, 0x1F0CF }, // So  [15] PLAYING CARD ACE OF DIAMONDS..PLAYING CARD BLACK JOKER
    Interval{ 0x1F0D1, 0x1F0F5 }, // So  [37] PLAYING CARD ACE OF CLUBS..PLAYING CARD TRUMP-21
    Interval{ 0x1F100, 0x1F10C }, // No  [13] DIGIT ZERO FULL STOP..DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT ZERO
    Interval{ 0x1F10D, 0x1F1AD }, // So [161] CIRCLED ZERO WITH SLASH..MASK WORK SYMBOL
    Interval{ 0x1F1E6, 0x1F202 }, // So  [29] REGIONAL INDICATOR SYMBOL LETTER A..SQUARED KATAKANA SA
    Interval{ 0x1F210, 0x1F23B }, // So  [44] SQUARED CJK UNIFIED IDEOGRAPH-624B..SQUARED CJK UNIFIED IDEOGRAPH-914D
    Interval{ 0x1F240, 0x1F248 }, // So   [9] TORTOISE SHELL BRACKETED CJK UNIFIED IDEOGRAPH-672C..TORTOISE SHELL BRACKETED CJK UNIFIED IDEOGRAPH-6557
    Interval{ 0x1F250, 0x1F251 }, // So   [2] CIRCLED IDEOGRAPH ADVANTAGE..CIRCLED IDEOGRAPH ACCEPT
    Interval{ 0x1F260, 0x1F265 }, // So   [6] ROUNDED SYMBOL FOR FU..ROUNDED SYMBOL FOR CAI
    Interval{ 0x1F300, 0x1F3FA }, // So [251] CYCLONE..AMPHORA
    Interval{ 0x1F3FB, 0x1F3FF }, // Sk   [5] EMOJI MODIFIER FITZPATRICK TYPE-1-2..EMOJI MODIFIER FITZPATRICK TYPE-6
    Interval{ 0x1F400, 0x1F6D8 }, // So [729] RAT..LANDSLIDE
    Interval{ 0x1F6DC, 0x1F6EC }, // So  [17] WIRELESS..AIRPLANE ARRIVING
    Interval{ 0x1F6F0, 0x1F6FC }, // So  [13] SATELLITE..ROLLER SKATE
    Interval{ 0x1F700, 0x1F7D9 }, // So [218] ALCHEMICAL SYMBOL FOR QUINTESSENCE..NINE POINTED WHITE STAR
    Interval{ 0x1F7E0, 0x1F7EB }, // So  [12] LARGE ORANGE CIRCLE..LARGE BROWN SQUARE
    Interval{ 0x1F7F0, 0x1F7F0 }, // So       HEAVY EQUALS SIGN
    Interval{ 0x1F800, 0x1F80B }, // So  [12] LEFTWARDS ARROW WITH SMALL TRIANGLE ARROWHEAD..DOWNWARDS ARROW WITH LARGE TRIANGLE ARROWHEAD
    Interval{ 0x1F810, 0x1F847 }, // So  [56] LEFTWARDS ARROW WITH SMALL EQUILATERAL ARROWHEAD..DOWNWARDS HEAVY ARROW
    Interval{ 0x1F850, 0x1F859 }, // So  [10] LEFTWARDS SANS-SERIF ARROW..UP DOWN SANS-SERIF ARROW
    Interval{ 0x1F860, 0x1F887 }, // So  [40] WIDE-HEADED LEFTWARDS LIGHT BARB ARROW..WIDE-HEADED SOUTH WEST VERY HEAVY BARB ARROW
    Interval{ 0x1F890, 0x1F8AD }, // So  [30] LEFTWARDS TRIANGLE ARROWHEAD..WHITE ARROW SHAFT WIDTH TWO THIRDS
    Interval{ 0x1F8B0, 0x1F8BB }, // So  [12] ARROW POINTING UPWARDS THEN NORTH WEST..SOUTH WEST ARROW FROM BAR
    Interval{ 0x1F8C0, 0x1F8C1 }, // So   [2] LEFTWARDS ARROW FROM DOWNWARDS ARROW..RIGHTWARDS ARROW FROM DOWNWARDS ARROW
    Interval{ 0x1F8D0, 0x1F8D8 }, // Sm   [9] LONG RIGHTWARDS ARROW OVER LONG LEFTWARDS ARROW..LONG LEFT RIGHT ARROW WITH DEPENDENT LOBE
    Interval{ 0x1F900, 0x1FA57 }, // So [344] CIRCLED CROSS FORMEE WITH FOUR DOTS..BLACK CHESS ALFIL
    Interval{ 0x1FA60, 0x1FA6D }, // So  [14] XIANGQI RED GENERAL..XIANGQI BLACK SOLDIER
    Interval{ 0x1FA70, 0x1FA7C }, // So  [13] BALLET SHOES..CRUTCH
    Interval{ 0x1FA80, 0x1FA8A }, // So  [11] YO-YO..TROMBONE
    Interval{ 0x1FA8E, 0x1FAC6 }, // So  [57] TREASURE CHEST..FINGERPRINT
    Interval{ 0x1FAC8, 0x1FAC8 }, // So       HAIRY CREATURE
    Interval{ 0x1FACD, 0x1FADC }, // So  [16] ORCA..ROOT VEGETABLE
    Interval{ 0x1FADF, 0x1FAEA }, // So  [12] SPLATTER..DISTORTED FACE
    Interval{ 0x1FAEF, 0x1FAF8 }, // So  [10] FIGHT CLOUD..RIGHTWARDS PUSHING HAND
    Interval{ 0x1FB00, 0x1FB92 }, // So [147] BLOCK SEXTANT-1..UPPER HALF INVERSE MEDIUM SHADE AND LOWER HALF BLOCK
    Interval{ 0x1FB94, 0x1FBEF }, // So  [92] LEFT HALF INVERSE MEDIUM SHADE AND RIGHT HALF BLOCK..TOP LEFT JUSTIFIED LOWER RIGHT QUARTER BLACK CIRCLE
    Interval{ 0x1FBF0, 0x1FBF9 }, // Nd  [10] SEGMENTED DIGIT ZERO..SEGMENTED DIGIT NINE
    Interval{ 0x1FBFA, 0x1FBFA }, // So       ALARM BELL SYMBOL
    Interval{ 0x20000, 0x2A6DF }, // Lo [42720] CJK UNIFIED IDEOGRAPH-20000..CJK UNIFIED IDEOGRAPH-2A6DF
    Interval{ 0x2A700, 0x2B81D }, // Lo [4382] CJK UNIFIED IDEOGRAPH-2A700..CJK UNIFIED IDEOGRAPH-2B81D
    Interval{ 0x2B820, 0x2CEAD }, // Lo [5774] CJK UNIFIED IDEOGRAPH-2B820..CJK UNIFIED IDEOGRAPH-2CEAD
    Interval{ 0x2CEB0, 0x2EBE0 }, // Lo [7473] CJK UNIFIED IDEOGRAPH-2CEB0..CJK UNIFIED IDEOGRAPH-2EBE0
    Interval{ 0x2EBF0, 0x2EE5D }, // Lo [622] CJK UNIFIED IDEOGRAPH-2EBF0..CJK UNIFIED IDEOGRAPH-2EE5D
    Interval{ 0x2F800, 0x2FA1D }, // Lo [542] CJK COMPATIBILITY IDEOGRAPH-2F800..CJK COMPATIBILITY IDEOGRAPH-2FA1D
    Interval{ 0x30000, 0x3134A }, // Lo [4939] CJK UNIFIED IDEOGRAPH-30000..CJK UNIFIED IDEOGRAPH-3134A
    Interval{ 0x31350, 0x33479 }, // Lo [8490] CJK UNIFIED IDEOGRAPH-31350..CJK UNIFIED IDEOGRAPH-33479
}; // }}}
auto static const Grapheme_Extend = std::array<Interval, 419>{ // {{{
    Interval{ 0x0300, 0x036F }, // Mn [112] COMBINING GRAVE ACCENT..COMBINING LATIN SMALL LETTER X
    Interval{ 0x0483, 0x0487 }, // Mn   [5] COMBINING CYRILLIC TITLO..COMBINING CYRILLIC POKRYTIE
    Interval{ 0x0488, 0x0489 }, // Me   [2] COMBINING CYRILLIC HUNDRED THOUSANDS SIGN..COMBINING CYRILLIC MILLIONS SIGN
    Interval{ 0x0591, 0x05BD }, // Mn  [45] HEBREW ACCENT ETNAHTA..HEBREW POINT METEG
    Interval{ 0x05BF, 0x05BF }, // Mn       HEBREW POINT RAFE
    Interval{ 0x05C1, 0x05C2 }, // Mn   [2] HEBREW POINT SHIN DOT..HEBREW POINT SIN DOT
    Interval{ 0x05C4, 0x05C5 }, // Mn   [2] HEBREW MARK UPPER DOT..HEBREW MARK LOWER DOT
    Interval{ 0x05C7, 0x05C7 }, // Mn       HEBREW POINT QAMATS QATAN
    Interval{ 0x0610, 0x061A }, // Mn  [11] ARABIC SIGN SALLALLAHOU ALAYHE WASSALLAM..ARABIC SMALL KASRA
    Interval{ 0x064B, 0x065F }, // Mn  [21] ARABIC FATHATAN..ARABIC WAVY HAMZA BELOW
    Interval{ 0x0670, 0x0670 }, // Mn       ARABIC LETTER SUPERSCRIPT ALEF
    Interval{ 0x06D6, 0x06DC }, // Mn   [7] ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA..ARABIC SMALL HIGH SEEN
    Interval{ 0x06DF, 0x06E4 }, // Mn   [6] ARABIC SMALL HIGH ROUNDED ZERO..ARABIC SMALL HIGH MADDA
    Interval{ 0x06E7, 0x06E8 }, // Mn   [2] ARABIC SMALL HIGH YEH..ARABIC SMALL HIGH NOON
    Interval{ 0x06EA, 0x06ED }, // Mn   [4] ARABIC EMPTY CENTRE LOW STOP..ARABIC SMALL LOW MEEM
    Interval{ 0x0711, 0x0711 }, // Mn       SYRIAC LETTER SUPERSCRIPT ALAPH
    Interval{ 0x0730, 0x074A }, // Mn  [27] SYRIAC PTHAHA ABOVE..SYRIAC BARREKH
    Interval{ 0x07A6, 0x07B0 }, // Mn  [11] THAANA ABAFILI..THAANA SUKUN
    Interval{ 0x07EB, 0x07F3 }, // Mn   [9] NKO COMBINING SHORT HIGH TONE..NKO COMBINING DOUBLE DOT ABOVE
    Interval{ 0x07FD, 0x07FD }, // Mn       NKO DANTAYALAN
    Interval{ 0x0816, 0x0819 }, // Mn   [4] SAMARITAN MARK IN..SAMARITAN MARK DAGESH
    Interval{ 0x081B, 0x0823 }, // Mn   [9] SAMARITAN MARK EPENTHETIC YUT..SAMARITAN VOWEL SIGN A
    Interval{ 0x0825, 0x0827 }, // Mn   [3] SAMARITAN VOWEL SIGN SHORT A..SAMARITAN VOWEL SIGN U
    Interval{ 0x0829, 0x082D }, // Mn   [5] SAMARITAN VOWEL SIGN LONG I..SAMARITAN MARK NEQUDAA
    Interval{ 0x0859, 0x085B }, // Mn   [3] MANDAIC AFFRICATION MARK..MANDAIC GEMINATION MARK
    Interval{ 0x0897, 0x089F }, // Mn   [9] ARABIC PEPET..ARABIC HALF MADDA OVER MADDA
    Interval{ 0x08CA, 0x08E1 }, // Mn  [24] ARABIC SMALL HIGH FARSI YEH..ARABIC SMALL HIGH SIGN SAFHA
    Interval{ 0x08E3, 0x0902 }, // Mn  [32] ARABIC TURNED DAMMA BELOW..DEVANAGARI SIGN ANUSVARA
    Interval{ 0x093A, 0x093A }, // Mn       DEVANAGARI VOWEL SIGN OE
    Interval{ 0x093C, 0x093C }, // Mn       DEVANAGARI SIGN NUKTA
    Interval{ 0x0941, 0x0948 }, // Mn   [8] DEVANAGARI VOWEL SIGN U..DEVANAGARI VOWEL SIGN AI
    Interval{ 0x094D, 0x094D }, // Mn       DEVANAGARI SIGN VIRAMA
    Interval{ 0x0951, 0x0957 }, // Mn   [7] DEVANAGARI STRESS SIGN UDATTA..DEVANAGARI VOWEL SIGN UUE
    Interval{ 0x0962, 0x0963 }, // Mn   [2] DEVANAGARI VOWEL SIGN VOCALIC L..DEVANAGARI VOWEL SIGN VOCALIC LL
    Interval{ 0x0981, 0x0981 }, // Mn       BENGALI SIGN CANDRABINDU
    Interval{ 0x09BC, 0x09BC }, // Mn       BENGALI SIGN NUKTA
    Interval{ 0x09BE, 0x09BE }, // Mc       BENGALI VOWEL SIGN AA
    Interval{ 0x09C1, 0x09C4 }, // Mn   [4] BENGALI VOWEL SIGN U..BENGALI VOWEL SIGN VOCALIC RR
    Interval{ 0x09CD, 0x09CD }, // Mn       BENGALI SIGN VIRAMA
    Interval{ 0x09D7, 0x09D7 }, // Mc       BENGALI AU LENGTH MARK
    Interval{ 0x09E2, 0x09E3 }, // Mn   [2] BENGALI VOWEL SIGN VOCALIC L..BENGALI VOWEL SIGN VOCALIC LL
    Interval{ 0x09FE, 0x09FE }, // Mn       BENGALI SANDHI MARK
    Interval{ 0x0A01, 0x0A02 }, // Mn   [2] GURMUKHI SIGN ADAK BINDI..GURMUKHI SIGN BINDI
    Interval{ 0x0A3C, 0x0A3C }, // Mn       GURMUKHI SIGN NUKTA
    Interval{ 0x0A41, 0x0A42 }, // Mn   [2] GURMUKHI VOWEL SIGN U..GURMUKHI VOWEL SIGN UU
    Interval{ 0x0A47, 0x0A48 }, // Mn   [2] GURMUKHI VOWEL SIGN EE..GURMUKHI VOWEL SIGN AI
    Interval{ 0x0A4B, 0x0A4D }, // Mn   [3] GURMUKHI VOWEL SIGN OO..GURMUKHI SIGN VIRAMA
    Interval{ 0x0A51, 0x0A51 }, // Mn       GURMUKHI SIGN UDAAT
    Interval{ 0x0A70, 0x0A71 }, // Mn   [2] GURMUKHI TIPPI..GURMUKHI ADDAK
    Interval{ 0x0A75, 0x0A75 }, // Mn       GURMUKHI SIGN YAKASH
    Interval{ 0x0A81, 0x0A82 }, // Mn   [2] GUJARATI SIGN CANDRABINDU..GUJARATI SIGN ANUSVARA
    Interval{ 0x0ABC, 0x0ABC }, // Mn       GUJARATI SIGN NUKTA
    Interval{ 0x0AC1, 0x0AC5 }, // Mn   [5] GUJARATI VOWEL SIGN U..GUJARATI VOWEL SIGN CANDRA E
    Interval{ 0x0AC7, 0x0AC8 }, // Mn   [2] GUJARATI VOWEL SIGN E..GUJARATI VOWEL SIGN AI
    Interval{ 0x0ACD, 0x0ACD }, // Mn       GUJARATI SIGN VIRAMA
    Interval{ 0x0AE2, 0x0AE3 }, // Mn   [2] GUJARATI VOWEL SIGN VOCALIC L..GUJARATI VOWEL SIGN VOCALIC LL
    Interval{ 0x0AFA, 0x0AFF }, // Mn   [6] GUJARATI SIGN SUKUN..GUJARATI SIGN TWO-CIRCLE NUKTA ABOVE
    Interval{ 0x0B01, 0x0B01 }, // Mn       ORIYA SIGN CANDRABINDU
    Interval{ 0x0B3C, 0x0B3C }, // Mn       ORIYA SIGN NUKTA
    Interval{ 0x0B3E, 0x0B3E }, // Mc       ORIYA VOWEL SIGN AA
    Interval{ 0x0B3F, 0x0B3F }, // Mn       ORIYA VOWEL SIGN I
    Interval{ 0x0B41, 0x0B44 }, // Mn   [4] ORIYA VOWEL SIGN U..ORIYA VOWEL SIGN VOCALIC RR
    Interval{ 0x0B4D, 0x0B4D }, // Mn       ORIYA SIGN VIRAMA
    Interval{ 0x0B55, 0x0B56 }, // Mn   [2] ORIYA SIGN OVERLINE..ORIYA AI LENGTH MARK
    Interval{ 0x0B57, 0x0B57 }, // Mc       ORIYA AU LENGTH MARK
    Interval{ 0x0B62, 0x0B63 }, // Mn   [2] ORIYA VOWEL SIGN VOCALIC L..ORIYA VOWEL SIGN VOCALIC LL
    Interval{ 0x0B82, 0x0B82 }, // Mn       TAMIL SIGN ANUSVARA
    Interval{ 0x0BBE, 0x0BBE }, // Mc       TAMIL VOWEL SIGN AA
    Interval{ 0x0BC0, 0x0BC0 }, // Mn       TAMIL VOWEL SIGN II
    Interval{ 0x0BCD, 0x0BCD }, // Mn       TAMIL SIGN VIRAMA
    Interval{ 0x0BD7, 0x0BD7 }, // Mc       TAMIL AU LENGTH MARK
    Interval{ 0x0C00, 0x0C00 }, // Mn       TELUGU SIGN COMBINING CANDRABINDU ABOVE
    Interval{ 0x0C04, 0x0C04 }, // Mn       TELUGU SIGN COMBINING ANUSVARA ABOVE
    Interval{ 0x0C3C, 0x0C3C }, // Mn       TELUGU SIGN NUKTA
    Interval{ 0x0C3E, 0x0C40 }, // Mn   [3] TELUGU VOWEL SIGN AA..TELUGU VOWEL SIGN II
    Interval{ 0x0C46, 0x0C48 }, // Mn   [3] TELUGU VOWEL SIGN E..TELUGU VOWEL SIGN AI
    Interval{ 0x0C4A, 0x0C4D }, // Mn   [4] TELUGU VOWEL SIGN O..TELUGU SIGN VIRAMA
    Interval{ 0x0C55, 0x0C56 }, // Mn   [2] TELUGU LENGTH MARK..TELUGU AI LENGTH MARK
    Interval{ 0x0C62, 0x0C63 }, // Mn   [2] TELUGU VOWEL SIGN VOCALIC L..TELUGU VOWEL SIGN VOCALIC LL
    Interval{ 0x0C81, 0x0C81 }, // Mn       KANNADA SIGN CANDRABINDU
    Interval{ 0x0CBC, 0x0CBC }, // Mn       KANNADA SIGN NUKTA
    Interval{ 0x0CBF, 0x0CBF }, // Mn       KANNADA VOWEL SIGN I
    Interval{ 0x0CC0, 0x0CC0 }, // Mc       KANNADA VOWEL SIGN II
    Interval{ 0x0CC2, 0x0CC2 }, // Mc       KANNADA VOWEL SIGN UU
    Interval{ 0x0CC6, 0x0CC6 }, // Mn       KANNADA VOWEL SIGN E
    Interval{ 0x0CC7, 0x0CC8 }, // Mc   [2] KANNADA VOWEL SIGN EE..KANNADA VOWEL SIGN AI
    Interval{ 0x0CCA, 0x0CCB }, // Mc   [2] KANNADA VOWEL SIGN O..KANNADA VOWEL SIGN OO
    Interval{ 0x0CCC, 0x0CCD }, // Mn   [2] KANNADA VOWEL SIGN AU..KANNADA SIGN VIRAMA
    Interval{ 0x0CD5, 0x0CD6 }, // Mc   [2] KANNADA LENGTH MARK..KANNADA AI LENGTH MARK
    Interval{ 0x0CE2, 0x0CE3 }, // Mn   [2] KANNADA VOWEL SIGN VOCALIC L..KANNADA VOWEL SIGN VOCALIC LL
    Interval{ 0x0D00, 0x0D01 }, // Mn   [2] MALAYALAM SIGN COMBINING ANUSVARA ABOVE..MALAYALAM SIGN CANDRABINDU
    Interval{ 0x0D3B, 0x0D3C }, // Mn   [2] MALAYALAM SIGN VERTICAL BAR VIRAMA..MALAYALAM SIGN CIRCULAR VIRAMA
    Interval{ 0x0D3E, 0x0D3E }, // Mc       MALAYALAM VOWEL SIGN AA
    Interval{ 0x0D41, 0x0D44 }, // Mn   [4] MALAYALAM VOWEL SIGN U..MALAYALAM VOWEL SIGN VOCALIC RR
    Interval{ 0x0D4D, 0x0D4D }, // Mn       MALAYALAM SIGN VIRAMA
    Interval{ 0x0D57, 0x0D57 }, // Mc       MALAYALAM AU LENGTH MARK
    Interval{ 0x0D62, 0x0D63 }, // Mn   [2] MALAYALAM VOWEL SIGN VOCALIC L..MALAYALAM VOWEL SIGN VOCALIC LL
    Interval{ 0x0D81, 0x0D81 }, // Mn       SINHALA SIGN CANDRABINDU
    Interval{ 0x0DCA, 0x0DCA }, // Mn       SINHALA SIGN AL-LAKUNA
    Interval{ 0x0DCF, 0x0DCF }, // Mc       SINHALA VOWEL SIGN AELA-PILLA
    Interval{ 0x0DD2, 0x0DD4 }, // Mn   [3] SINHALA VOWEL SIGN KETTI IS-PILLA..SINHALA VOWEL SIGN KETTI PAA-PILLA
    Interval{ 0x0DD6, 0x0DD6 }, // Mn       SINHALA VOWEL SIGN DIGA PAA-PILLA
    Interval{ 0x0DDF, 0x0DDF }, // Mc       SINHALA VOWEL SIGN GAYANUKITTA
    Interval{ 0x0E31, 0x0E31 }, // Mn       THAI CHARACTER MAI HAN-AKAT
    Interval{ 0x0E34, 0x0E3A }, // Mn   [7] THAI CHARACTER SARA I..THAI CHARACTER PHINTHU
    Interval{ 0x0E47, 0x0E4E }, // Mn   [8] THAI CHARACTER MAITAIKHU..THAI CHARACTER YAMAKKAN
    Interval{ 0x0EB1, 0x0EB1 }, // Mn       LAO VOWEL SIGN MAI KAN
    Interval{ 0x0EB4, 0x0EBC }, // Mn   [9] LAO VOWEL SIGN I..LAO SEMIVOWEL SIGN LO
    Interval{ 0x0EC8, 0x0ECE }, // Mn   [7] LAO TONE MAI EK..LAO YAMAKKAN
    Interval{ 0x0F18, 0x0F19 }, // Mn   [2] TIBETAN ASTROLOGICAL SIGN -KHYUD PA..TIBETAN ASTROLOGICAL SIGN SDONG TSHUGS
    Interval{ 0x0F35, 0x0F35 }, // Mn       TIBETAN MARK NGAS BZUNG NYI ZLA
    Interval{ 0x0F37, 0x0F37 }, // Mn       TIBETAN MARK NGAS BZUNG SGOR RTAGS
    Interval{ 0x0F39, 0x0F39 }, // Mn       TIBETAN MARK TSA -PHRU
    Interval{ 0x0F71, 0x0F7E }, // Mn  [14] TIBETAN VOWEL SIGN AA..TIBETAN SIGN RJES SU NGA RO
    Interval{ 0x0F80, 0x0F84 }, // Mn   [5] TIBETAN VOWEL SIGN REVERSED I..TIBETAN MARK HALANTA
    Interval{ 0x0F86, 0x0F87 }, // Mn   [2] TIBETAN SIGN LCI RTAGS..TIBETAN SIGN YANG RTAGS
    Interval{ 0x0F8D, 0x0F97 }, // Mn  [11] TIBETAN SUBJOINED SIGN LCE TSA CAN..TIBETAN SUBJOINED LETTER JA
    Interval{ 0x0F99, 0x0FBC }, // Mn  [36] TIBETAN SUBJOINED LETTER NYA..TIBETAN SUBJOINED LETTER FIXED-FORM RA
    Interval{ 0x0FC6, 0x0FC6 }, // Mn       TIBETAN SYMBOL PADMA GDAN
    Interval{ 0x102D, 0x1030 }, // Mn   [4] MYANMAR VOWEL SIGN I..MYANMAR VOWEL SIGN UU
    Interval{ 0x1032, 0x1037 }, // Mn   [6] MYANMAR VOWEL SIGN AI..MYANMAR SIGN DOT BELOW
    Interval{ 0x1039, 0x103A }, // Mn   [2] MYANMAR SIGN VIRAMA..MYANMAR SIGN ASAT
    Interval{ 0x103D, 0x103E }, // Mn   [2] MYANMAR CONSONANT SIGN MEDIAL WA..MYANMAR CONSONANT SIGN MEDIAL HA
    Interval{ 0x1058, 0x1059 }, // Mn   [2] MYANMAR VOWEL SIGN VOCALIC L..MYANMAR VOWEL SIGN VOCALIC LL
    Interval{ 0x105E, 0x1060 }, // Mn   [3] MYANMAR CONSONANT SIGN MON MEDIAL NA..MYANMAR CONSONANT SIGN MON MEDIAL LA
    Interval{ 0x1071, 0x1074 }, // Mn   [4] MYANMAR VOWEL SIGN GEBA KAREN I..MYANMAR VOWEL SIGN KAYAH EE
    Interval{ 0x1082, 0x1082 }, // Mn       MYANMAR CONSONANT SIGN SHAN MEDIAL WA
    Interval{ 0x1085, 0x1086 }, // Mn   [2] MYANMAR VOWEL SIGN SHAN E ABOVE..MYANMAR VOWEL SIGN SHAN FINAL Y
    Interval{ 0x108D, 0x108D }, // Mn       MYANMAR SIGN SHAN COUNCIL EMPHATIC TONE
    Interval{ 0x109D, 0x109D }, // Mn       MYANMAR VOWEL SIGN AITON AI
    Interval{ 0x135D, 0x135F }, // Mn   [3] ETHIOPIC COMBINING GEMINATION AND VOWEL LENGTH MARK..ETHIOPIC COMBINING GEMINATION MARK
    Interval{ 0x1712, 0x1714 }, // Mn   [3] TAGALOG VOWEL SIGN I..TAGALOG SIGN VIRAMA
    Interval{ 0x1715, 0x1715 }, // Mc       TAGALOG SIGN PAMUDPOD
    Interval{ 0x1732, 0x1733 }, // Mn   [2] HANUNOO VOWEL SIGN I..HANUNOO VOWEL SIGN U
    Interval{ 0x1734, 0x1734 }, // Mc       HANUNOO SIGN PAMUDPOD
    Interval{ 0x1752, 0x1753 }, // Mn   [2] BUHID VOWEL SIGN I..BUHID VOWEL SIGN U
    Interval{ 0x1772, 0x1773 }, // Mn   [2] TAGBANWA VOWEL SIGN I..TAGBANWA VOWEL SIGN U
    Interval{ 0x17B4, 0x17B5 }, // Mn   [2] KHMER VOWEL INHERENT AQ..KHMER VOWEL INHERENT AA
    Interval{ 0x17B7, 0x17BD }, // Mn   [7] KHMER VOWEL SIGN I..KHMER VOWEL SIGN UA
    Interval{ 0x17C6, 0x17C6 }, // Mn       KHMER SIGN NIKAHIT
    Interval{ 0x17C9, 0x17D3 }, // Mn  [11] KHMER SIGN MUUSIKATOAN..KHMER SIGN BATHAMASAT
    Interval{ 0x17DD, 0x17DD }, // Mn       KHMER SIGN ATTHACAN
    Interval{ 0x180B, 0x180D }, // Mn   [3] MONGOLIAN FREE VARIATION SELECTOR ONE..MONGOLIAN FREE VARIATION SELECTOR THREE
    Interval{ 0x180F, 0x180F }, // Mn       MONGOLIAN FREE VARIATION SELECTOR FOUR
    Interval{ 0x1885, 0x1886 }, // Mn   [2] MONGOLIAN LETTER ALI GALI BALUDA..MONGOLIAN LETTER ALI GALI THREE BALUDA
    Interval{ 0x18A9, 0x18A9 }, // Mn       MONGOLIAN LETTER ALI GALI DAGALGA
    Interval{ 0x1920, 0x1922 }, // Mn   [3] LIMBU VOWEL SIGN A..LIMBU VOWEL SIGN U
    Interval{ 0x1927, 0x1928 }, // Mn   [2] LIMBU VOWEL SIGN E..LIMBU VOWEL SIGN O
    Interval{ 0x1932, 0x1932 }, // Mn       LIMBU SMALL LETTER ANUSVARA
    Interval{ 0x1939, 0x193B }, // Mn   [3] LIMBU SIGN MUKPHRENG..LIMBU SIGN SA-I
    Interval{ 0x1A17, 0x1A18 }, // Mn   [2] BUGINESE VOWEL SIGN I..BUGINESE VOWEL SIGN U
    Interval{ 0x1A1B, 0x1A1B }, // Mn       BUGINESE VOWEL SIGN AE
    Interval{ 0x1A56, 0x1A56 }, // Mn       TAI THAM CONSONANT SIGN MEDIAL LA
    Interval{ 0x1A58, 0x1A5E }, // Mn   [7] TAI THAM SIGN MAI KANG LAI..TAI THAM CONSONANT SIGN SA
    Interval{ 0x1A60, 0x1A60 }, // Mn       TAI THAM SIGN SAKOT
    Interval{ 0x1A62, 0x1A62 }, // Mn       TAI THAM VOWEL SIGN MAI SAT
    Interval{ 0x1A65, 0x1A6C }, // Mn   [8] TAI THAM VOWEL SIGN I..TAI THAM VOWEL SIGN OA BELOW
    Interval{ 0x1A73, 0x1A7C }, // Mn  [10] TAI THAM VOWEL SIGN OA ABOVE..TAI THAM SIGN KHUEN-LUE KARAN
    Interval{ 0x1A7F, 0x1A7F }, // Mn       TAI THAM COMBINING CRYPTOGRAMMIC DOT
    Interval{ 0x1AB0, 0x1ABD }, // Mn  [14] COMBINING DOUBLED CIRCUMFLEX ACCENT..COMBINING PARENTHESES BELOW
    Interval{ 0x1ABE, 0x1ABE }, // Me       COMBINING PARENTHESES OVERLAY
    Interval{ 0x1ABF, 0x1ADD }, // Mn  [31] COMBINING LATIN SMALL LETTER W BELOW..COMBINING DOT-AND-RING BELOW
    Interval{ 0x1AE0, 0x1AEB }, // Mn  [12] COMBINING LEFT TACK ABOVE..COMBINING DOUBLE RIGHTWARDS ARROW ABOVE
    Interval{ 0x1B00, 0x1B03 }, // Mn   [4] BALINESE SIGN ULU RICEM..BALINESE SIGN SURANG
    Interval{ 0x1B34, 0x1B34 }, // Mn       BALINESE SIGN REREKAN
    Interval{ 0x1B35, 0x1B35 }, // Mc       BALINESE VOWEL SIGN TEDUNG
    Interval{ 0x1B36, 0x1B3A }, // Mn   [5] BALINESE VOWEL SIGN ULU..BALINESE VOWEL SIGN RA REPA
    Interval{ 0x1B3B, 0x1B3B }, // Mc       BALINESE VOWEL SIGN RA REPA TEDUNG
    Interval{ 0x1B3C, 0x1B3C }, // Mn       BALINESE VOWEL SIGN LA LENGA
    Interval{ 0x1B3D, 0x1B3D }, // Mc       BALINESE VOWEL SIGN LA LENGA TEDUNG
    Interval{ 0x1B42, 0x1B42 }, // Mn       BALINESE VOWEL SIGN PEPET
    Interval{ 0x1B43, 0x1B44 }, // Mc   [2] BALINESE VOWEL SIGN PEPET TEDUNG..BALINESE ADEG ADEG
    Interval{ 0x1B6B, 0x1B73 }, // Mn   [9] BALINESE MUSICAL SYMBOL COMBINING TEGEH..BALINESE MUSICAL SYMBOL COMBINING GONG
    Interval{ 0x1B80, 0x1B81 }, // Mn   [2] SUNDANESE SIGN PANYECEK..SUNDANESE SIGN PANGLAYAR
    Interval{ 0x1BA2, 0x1BA5 }, // Mn   [4] SUNDANESE CONSONANT SIGN PANYAKRA..SUNDANESE VOWEL SIGN PANYUKU
    Interval{ 0x1BA8, 0x1BA9 }, // Mn   [2] SUNDANESE VOWEL SIGN PAMEPET..SUNDANESE VOWEL SIGN PANEULEUNG
    Interval{ 0x1BAA, 0x1BAA }, // Mc       SUNDANESE SIGN PAMAAEH
    Interval{ 0x1BAB, 0x1BAD }, // Mn   [3] SUNDANESE SIGN VIRAMA..SUNDANESE CONSONANT SIGN PASANGAN WA
    Interval{ 0x1BE6, 0x1BE6 }, // Mn       BATAK SIGN TOMPI
    Interval{ 0x1BE8, 0x1BE9 }, // Mn   [2] BATAK VOWEL SIGN PAKPAK E..BATAK VOWEL SIGN EE
    Interval{ 0x1BED, 0x1BED }, // Mn       BATAK VOWEL SIGN KARO O
    Interval{ 0x1BEF, 0x1BF1 }, // Mn   [3] BATAK VOWEL SIGN U FOR SIMALUNGUN SA..BATAK CONSONANT SIGN H
    Interval{ 0x1BF2, 0x1BF3 }, // Mc   [2] BATAK PANGOLAT..BATAK PANONGONAN
    Interval{ 0x1C2C, 0x1C33 }, // Mn   [8] LEPCHA VOWEL SIGN E..LEPCHA CONSONANT SIGN T
    Interval{ 0x1C36, 0x1C37 }, // Mn   [2] LEPCHA SIGN RAN..LEPCHA SIGN NUKTA
    Interval{ 0x1CD0, 0x1CD2 }, // Mn   [3] VEDIC TONE KARSHANA..VEDIC TONE PRENKHA
    Interval{ 0x1CD4, 0x1CE0 }, // Mn  [13] VEDIC SIGN YAJURVEDIC MIDLINE SVARITA..VEDIC TONE RIGVEDIC KASHMIRI INDEPENDENT SVARITA
    Interval{ 0x1CE2, 0x1CE8 }, // Mn   [7] VEDIC SIGN VISARGA SVARITA..VEDIC SIGN VISARGA ANUDATTA WITH TAIL
    Interval{ 0x1CED, 0x1CED }, // Mn       VEDIC SIGN TIRYAK
    Interval{ 0x1CF4, 0x1CF4 }, // Mn       VEDIC TONE CANDRA ABOVE
    Interval{ 0x1CF8, 0x1CF9 }, // Mn   [2] VEDIC TONE RING ABOVE..VEDIC TONE DOUBLE RING ABOVE
    Interval{ 0x1DC0, 0x1DFF }, // Mn  [64] COMBINING DOTTED GRAVE ACCENT..COMBINING RIGHT ARROWHEAD AND DOWN ARROWHEAD BELOW
    Interval{ 0x200C, 0x200C }, // Cf       ZERO WIDTH NON-JOINER
    Interval{ 0x20D0, 0x20DC }, // Mn  [13] COMBINING LEFT HARPOON ABOVE..COMBINING FOUR DOTS ABOVE
    Interval{ 0x20DD, 0x20E0 }, // Me   [4] COMBINING ENCLOSING CIRCLE..COMBINING ENCLOSING CIRCLE BACKSLASH
    Interval{ 0x20E1, 0x20E1 }, // Mn       COMBINING LEFT RIGHT ARROW ABOVE
    Interval{ 0x20E2, 0x20E4 }, // Me   [3] COMBINING ENCLOSING SCREEN..COMBINING ENCLOSING UPWARD POINTING TRIANGLE
    Interval{ 0x20E5, 0x20F0 }, // Mn  [12] COMBINING REVERSE SOLIDUS OVERLAY..COMBINING ASTERISK ABOVE
    Interval{ 0x2CEF, 0x2CF1 }, // Mn   [3] COPTIC COMBINING NI ABOVE..COPTIC COMBINING SPIRITUS LENIS
    Interval{ 0x2D7F, 0x2D7F }, // Mn       TIFINAGH CONSONANT JOINER
    Interval{ 0x2DE0, 0x2DFF }, // Mn  [32] COMBINING CYRILLIC LETTER BE..COMBINING CYRILLIC LETTER IOTIFIED BIG YUS
    Interval{ 0x302A, 0x302D }, // Mn   [4] IDEOGRAPHIC LEVEL TONE MARK..IDEOGRAPHIC ENTERING TONE MARK
    Interval{ 0x302E, 0x302F }, // Mc   [2] HANGUL SINGLE DOT TONE MARK..HANGUL DOUBLE DOT TONE MARK
    Interval{ 0x3099, 0x309A }, // Mn   [2] COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK..COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Interval{ 0xA66F, 0xA66F }, // Mn       COMBINING CYRILLIC VZMET
    Interval{ 0xA670, 0xA672 }, // Me   [3] COMBINING CYRILLIC TEN MILLIONS SIGN..COMBINING CYRILLIC THOUSAND MILLIONS SIGN
    Interval{ 0xA674, 0xA67D }, // Mn  [10] COMBINING CYRILLIC LETTER UKRAINIAN IE..COMBINING CYRILLIC PAYEROK
    Interval{ 0xA69E, 0xA69F }, // Mn   [2] COMBINING CYRILLIC LETTER EF..COMBINING CYRILLIC LETTER IOTIFIED E
    Interval{ 0xA6F0, 0xA6F1 }, // Mn   [2] BAMUM COMBINING MARK KOQNDON..BAMUM COMBINING MARK TUKWENTIS
    Interval{ 0xA802, 0xA802 }, // Mn       SYLOTI NAGRI SIGN DVISVARA
    Interval{ 0xA806, 0xA806 }, // Mn       SYLOTI NAGRI SIGN HASANTA
    Interval{ 0xA80B, 0xA80B }, // Mn       SYLOTI NAGRI SIGN ANUSVARA
    Interval{ 0xA825, 0xA826 }, // Mn   [2] SYLOTI NAGRI VOWEL SIGN U..SYLOTI NAGRI VOWEL SIGN E
    Interval{ 0xA82C, 0xA82C }, // Mn       SYLOTI NAGRI SIGN ALTERNATE HASANTA
    Interval{ 0xA8C4, 0xA8C5 }, // Mn   [2] SAURASHTRA SIGN VIRAMA..SAURASHTRA SIGN CANDRABINDU
    Interval{ 0xA8E0, 0xA8F1 }, // Mn  [18] COMBINING DEVANAGARI DIGIT ZERO..COMBINING DEVANAGARI SIGN AVAGRAHA
    Interval{ 0xA8FF, 0xA8FF }, // Mn       DEVANAGARI VOWEL SIGN AY
    Interval{ 0xA926, 0xA92D }, // Mn   [8] KAYAH LI VOWEL UE..KAYAH LI TONE CALYA PLOPHU
    Interval{ 0xA947, 0xA951 }, // Mn  [11] REJANG VOWEL SIGN I..REJANG CONSONANT SIGN R
    Interval{ 0xA953, 0xA953 }, // Mc       REJANG VIRAMA
    Interval{ 0xA980, 0xA982 }, // Mn   [3] JAVANESE SIGN PANYANGGA..JAVANESE SIGN LAYAR
    Interval{ 0xA9B3, 0xA9B3 }, // Mn       JAVANESE SIGN CECAK TELU
    Interval{ 0xA9B6, 0xA9B9 }, // Mn   [4] JAVANESE VOWEL SIGN WULU..JAVANESE VOWEL SIGN SUKU MENDUT
    Interval{ 0xA9BC, 0xA9BD }, // Mn   [2] JAVANESE VOWEL SIGN PEPET..JAVANESE CONSONANT SIGN KERET
    Interval{ 0xA9C0, 0xA9C0 }, // Mc       JAVANESE PANGKON
    Interval{ 0xA9E5, 0xA9E5 }, // Mn       MYANMAR SIGN SHAN SAW
    Interval{ 0xAA29, 0xAA2E }, // Mn   [6] CHAM VOWEL SIGN AA..CHAM VOWEL SIGN OE
    Interval{ 0xAA31, 0xAA32 }, // Mn   [2] CHAM VOWEL SIGN AU..CHAM VOWEL SIGN UE
    Interval{ 0xAA35, 0xAA36 }, // Mn   [2] CHAM CONSONANT SIGN LA..CHAM CONSONANT SIGN WA
    Interval{ 0xAA43, 0xAA43 }, // Mn       CHAM CONSONANT SIGN FINAL NG
    Interval{ 0xAA4C, 0xAA4C }, // Mn       CHAM CONSONANT SIGN FINAL M
    Interval{ 0xAA7C, 0xAA7C }, // Mn       MYANMAR SIGN TAI LAING TONE-2
    Interval{ 0xAAB0, 0xAAB0 }, // Mn       TAI VIET MAI KANG
    Interval{ 0xAAB2, 0xAAB4 }, // Mn   [3] TAI VIET VOWEL I..TAI VIET VOWEL U
    Interval{ 0xAAB7, 0xAAB8 }, // Mn   [2] TAI VIET MAI KHIT..TAI VIET VOWEL IA
    Interval{ 0xAABE, 0xAABF }, // Mn   [2] TAI VIET VOWEL AM..TAI VIET TONE MAI EK
    Interval{ 0xAAC1, 0xAAC1 }, // Mn       TAI VIET TONE MAI THO
    Interval{ 0xAAEC, 0xAAED }, // Mn   [2] MEETEI MAYEK VOWEL SIGN UU..MEETEI MAYEK VOWEL SIGN AAI
    Interval{ 0xAAF6, 0xAAF6 }, // Mn       MEETEI MAYEK VIRAMA
    Interval{ 0xABE5, 0xABE5 }, // Mn       MEETEI MAYEK VOWEL SIGN ANAP
    Interval{ 0xABE8, 0xABE8 }, // Mn       MEETEI MAYEK VOWEL SIGN UNAP
    Interval{ 0xABED, 0xABED }, // Mn       MEETEI MAYEK APUN IYEK
    Interval{ 0xFB1E, 0xFB1E }, // Mn       HEBREW POINT JUDEO-SPANISH VARIKA
    Interval{ 0xFE00, 0xFE0F }, // Mn  [16] VARIATION SELECTOR-1..VARIATION SELECTOR-16
    Interval{ 0xFE20, 0xFE2F }, // Mn  [16] COMBINING LIGATURE LEFT HALF..COMBINING CYRILLIC TITLO RIGHT HALF
    Interval{ 0xFF9E, 0xFF9F }, // Lm   [2] HALFWIDTH KATAKANA VOICED SOUND MARK..HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK
    Interval{ 0x101FD, 0x101FD }, // Mn       PHAISTOS DISC SIGN COMBINING OBLIQUE STROKE
    Interval{ 0x102E0, 0x102E0 }, // Mn       COPTIC EPACT THOUSANDS MARK
    Interval{ 0x10376, 0x1037A }, // Mn   [5] COMBINING OLD PERMIC LETTER AN..COMBINING OLD PERMIC LETTER SII
    Interval{ 0x10A01, 0x10A03 }, // Mn   [3] KHAROSHTHI VOWEL SIGN I..KHAROSHTHI VOWEL SIGN VOCALIC R
    Interval{ 0x10A05, 0x10A06 }, // Mn   [2] KHAROSHTHI VOWEL SIGN E..KHAROSHTHI VOWEL SIGN O
    Interval{ 0x10A0C, 0x10A0F }, // Mn   [4] KHAROSHTHI VOWEL LENGTH MARK..KHAROSHTHI SIGN VISARGA
    Interval{ 0x10A38, 0x10A3A }, // Mn   [3] KHAROSHTHI SIGN BAR ABOVE..KHAROSHTHI SIGN DOT BELOW
    Interval{ 0x10A3F, 0x10A3F }, // Mn       KHAROSHTHI VIRAMA
    Interval{ 0x10AE5, 0x10AE6 }, // Mn   [2] MANICHAEAN ABBREVIATION MARK ABOVE..MANICHAEAN ABBREVIATION MARK BELOW
    Interval{ 0x10D24, 0x10D27 }, // Mn   [4] HANIFI ROHINGYA SIGN HARBAHAY..HANIFI ROHINGYA SIGN TASSI
    Interval{ 0x10D69, 0x10D6D }, // Mn   [5] GARAY VOWEL SIGN E..GARAY CONSONANT NASALIZATION MARK
    Interval{ 0x10EAB, 0x10EAC }, // Mn   [2] YEZIDI COMBINING HAMZA MARK..YEZIDI COMBINING MADDA MARK
    Interval{ 0x10EFA, 0x10EFF }, // Mn   [6] ARABIC DOUBLE VERTICAL BAR BELOW..ARABIC SMALL LOW WORD MADDA
    Interval{ 0x10F46, 0x10F50 }, // Mn  [11] SOGDIAN COMBINING DOT BELOW..SOGDIAN COMBINING STROKE BELOW
    Interval{ 0x10F82, 0x10F85 }, // Mn   [4] OLD UYGHUR COMBINING DOT ABOVE..OLD UYGHUR COMBINING TWO DOTS BELOW
    Interval{ 0x11001, 0x11001 }, // Mn       BRAHMI SIGN ANUSVARA
    Interval{ 0x11038, 0x11046 }, // Mn  [15] BRAHMI VOWEL SIGN AA..BRAHMI VIRAMA
    Interval{ 0x11070, 0x11070 }, // Mn       BRAHMI SIGN OLD TAMIL VIRAMA
    Interval{ 0x11073, 0x11074 }, // Mn   [2] BRAHMI VOWEL SIGN OLD TAMIL SHORT E..BRAHMI VOWEL SIGN OLD TAMIL SHORT O
    Interval{ 0x1107F, 0x11081 }, // Mn   [3] BRAHMI NUMBER JOINER..KAITHI SIGN ANUSVARA
    Interval{ 0x110B3, 0x110B6 }, // Mn   [4] KAITHI VOWEL SIGN U..KAITHI VOWEL SIGN AI
    Interval{ 0x110B9, 0x110BA }, // Mn   [2] KAITHI SIGN VIRAMA..KAITHI SIGN NUKTA
    Interval{ 0x110C2, 0x110C2 }, // Mn       KAITHI VOWEL SIGN VOCALIC R
    Interval{ 0x11100, 0x11102 }, // Mn   [3] CHAKMA SIGN CANDRABINDU..CHAKMA SIGN VISARGA
    Interval{ 0x11127, 0x1112B }, // Mn   [5] CHAKMA VOWEL SIGN A..CHAKMA VOWEL SIGN UU
    Interval{ 0x1112D, 0x11134 }, // Mn   [8] CHAKMA VOWEL SIGN AI..CHAKMA MAAYYAA
    Interval{ 0x11173, 0x11173 }, // Mn       MAHAJANI SIGN NUKTA
    Interval{ 0x11180, 0x11181 }, // Mn   [2] SHARADA SIGN CANDRABINDU..SHARADA SIGN ANUSVARA
    Interval{ 0x111B6, 0x111BE }, // Mn   [9] SHARADA VOWEL SIGN U..SHARADA VOWEL SIGN O
    Interval{ 0x111C0, 0x111C0 }, // Mc       SHARADA SIGN VIRAMA
    Interval{ 0x111C9, 0x111CC }, // Mn   [4] SHARADA SANDHI MARK..SHARADA EXTRA SHORT VOWEL MARK
    Interval{ 0x111CF, 0x111CF }, // Mn       SHARADA SIGN INVERTED CANDRABINDU
    Interval{ 0x1122F, 0x11231 }, // Mn   [3] KHOJKI VOWEL SIGN U..KHOJKI VOWEL SIGN AI
    Interval{ 0x11234, 0x11234 }, // Mn       KHOJKI SIGN ANUSVARA
    Interval{ 0x11235, 0x11235 }, // Mc       KHOJKI SIGN VIRAMA
    Interval{ 0x11236, 0x11237 }, // Mn   [2] KHOJKI SIGN NUKTA..KHOJKI SIGN SHADDA
    Interval{ 0x1123E, 0x1123E }, // Mn       KHOJKI SIGN SUKUN
    Interval{ 0x11241, 0x11241 }, // Mn       KHOJKI VOWEL SIGN VOCALIC R
    Interval{ 0x112DF, 0x112DF }, // Mn       KHUDAWADI SIGN ANUSVARA
    Interval{ 0x112E3, 0x112EA }, // Mn   [8] KHUDAWADI VOWEL SIGN U..KHUDAWADI SIGN VIRAMA
    Interval{ 0x11300, 0x11301 }, // Mn   [2] GRANTHA SIGN COMBINING ANUSVARA ABOVE..GRANTHA SIGN CANDRABINDU
    Interval{ 0x1133B, 0x1133C }, // Mn   [2] COMBINING BINDU BELOW..GRANTHA SIGN NUKTA
    Interval{ 0x1133E, 0x1133E }, // Mc       GRANTHA VOWEL SIGN AA
    Interval{ 0x11340, 0x11340 }, // Mn       GRANTHA VOWEL SIGN II
    Interval{ 0x1134D, 0x1134D }, // Mc       GRANTHA SIGN VIRAMA
    Interval{ 0x11357, 0x11357 }, // Mc       GRANTHA AU LENGTH MARK
    Interval{ 0x11366, 0x1136C }, // Mn   [7] COMBINING GRANTHA DIGIT ZERO..COMBINING GRANTHA DIGIT SIX
    Interval{ 0x11370, 0x11374 }, // Mn   [5] COMBINING GRANTHA LETTER A..COMBINING GRANTHA LETTER PA
    Interval{ 0x113B8, 0x113B8 }, // Mc       TULU-TIGALARI VOWEL SIGN AA
    Interval{ 0x113BB, 0x113C0 }, // Mn   [6] TULU-TIGALARI VOWEL SIGN U..TULU-TIGALARI VOWEL SIGN VOCALIC LL
    Interval{ 0x113C2, 0x113C2 }, // Mc       TULU-TIGALARI VOWEL SIGN EE
    Interval{ 0x113C5, 0x113C5 }, // Mc       TULU-TIGALARI VOWEL SIGN AI
    Interval{ 0x113C7, 0x113C9 }, // Mc   [3] TULU-TIGALARI VOWEL SIGN OO..TULU-TIGALARI AU LENGTH MARK
    Interval{ 0x113CE, 0x113CE }, // Mn       TULU-TIGALARI SIGN VIRAMA
    Interval{ 0x113CF, 0x113CF }, // Mc       TULU-TIGALARI SIGN LOOPED VIRAMA
    Interval{ 0x113D0, 0x113D0 }, // Mn       TULU-TIGALARI CONJOINER
    Interval{ 0x113D2, 0x113D2 }, // Mn       TULU-TIGALARI GEMINATION MARK
    Interval{ 0x113E1, 0x113E2 }, // Mn   [2] TULU-TIGALARI VEDIC TONE SVARITA..TULU-TIGALARI VEDIC TONE ANUDATTA
    Interval{ 0x11438, 0x1143F }, // Mn   [8] NEWA VOWEL SIGN U..NEWA VOWEL SIGN AI
    Interval{ 0x11442, 0x11444 }, // Mn   [3] NEWA SIGN VIRAMA..NEWA SIGN ANUSVARA
    Interval{ 0x11446, 0x11446 }, // Mn       NEWA SIGN NUKTA
    Interval{ 0x1145E, 0x1145E }, // Mn       NEWA SANDHI MARK
    Interval{ 0x114B0, 0x114B0 }, // Mc       TIRHUTA VOWEL SIGN AA
    Interval{ 0x114B3, 0x114B8 }, // Mn   [6] TIRHUTA VOWEL SIGN U..TIRHUTA VOWEL SIGN VOCALIC LL
    Interval{ 0x114BA, 0x114BA }, // Mn       TIRHUTA VOWEL SIGN SHORT E
    Interval{ 0x114BD, 0x114BD }, // Mc       TIRHUTA VOWEL SIGN SHORT O
    Interval{ 0x114BF, 0x114C0 }, // Mn   [2] TIRHUTA SIGN CANDRABINDU..TIRHUTA SIGN ANUSVARA
    Interval{ 0x114C2, 0x114C3 }, // Mn   [2] TIRHUTA SIGN VIRAMA..TIRHUTA SIGN NUKTA
    Interval{ 0x115AF, 0x115AF }, // Mc       SIDDHAM VOWEL SIGN AA
    Interval{ 0x115B2, 0x115B5 }, // Mn   [4] SIDDHAM VOWEL SIGN U..SIDDHAM VOWEL SIGN VOCALIC RR
    Interval{ 0x115BC, 0x115BD }, // Mn   [2] SIDDHAM SIGN CANDRABINDU..SIDDHAM SIGN ANUSVARA
    Interval{ 0x115BF, 0x115C0 }, // Mn   [2] SIDDHAM SIGN VIRAMA..SIDDHAM SIGN NUKTA
    Interval{ 0x115DC, 0x115DD }, // Mn   [2] SIDDHAM VOWEL SIGN ALTERNATE U..SIDDHAM VOWEL SIGN ALTERNATE UU
    Interval{ 0x11633, 0x1163A }, // Mn   [8] MODI VOWEL SIGN U..MODI VOWEL SIGN AI
    Interval{ 0x1163D, 0x1163D }, // Mn       MODI SIGN ANUSVARA
    Interval{ 0x1163F, 0x11640 }, // Mn   [2] MODI SIGN VIRAMA..MODI SIGN ARDHACANDRA
    Interval{ 0x116AB, 0x116AB }, // Mn       TAKRI SIGN ANUSVARA
    Interval{ 0x116AD, 0x116AD }, // Mn       TAKRI VOWEL SIGN AA
    Interval{ 0x116B0, 0x116B5 }, // Mn   [6] TAKRI VOWEL SIGN U..TAKRI VOWEL SIGN AU
    Interval{ 0x116B6, 0x116B6 }, // Mc       TAKRI SIGN VIRAMA
    Interval{ 0x116B7, 0x116B7 }, // Mn       TAKRI SIGN NUKTA
    Interval{ 0x1171D, 0x1171D }, // Mn       AHOM CONSONANT SIGN MEDIAL LA
    Interval{ 0x1171F, 0x1171F }, // Mn       AHOM CONSONANT SIGN MEDIAL LIGATING RA
    Interval{ 0x11722, 0x11725 }, // Mn   [4] AHOM VOWEL SIGN I..AHOM VOWEL SIGN UU
    Interval{ 0x11727, 0x1172B }, // Mn   [5] AHOM VOWEL SIGN AW..AHOM SIGN KILLER
    Interval{ 0x1182F, 0x11837 }, // Mn   [9] DOGRA VOWEL SIGN U..DOGRA SIGN ANUSVARA
    Interval{ 0x11839, 0x1183A }, // Mn   [2] DOGRA SIGN VIRAMA..DOGRA SIGN NUKTA
    Interval{ 0x11930, 0x11930 }, // Mc       DIVES AKURU VOWEL SIGN AA
    Interval{ 0x1193B, 0x1193C }, // Mn   [2] DIVES AKURU SIGN ANUSVARA..DIVES AKURU SIGN CANDRABINDU
    Interval{ 0x1193D, 0x1193D }, // Mc       DIVES AKURU SIGN HALANTA
    Interval{ 0x1193E, 0x1193E }, // Mn       DIVES AKURU VIRAMA
    Interval{ 0x11943, 0x11943 }, // Mn       DIVES AKURU SIGN NUKTA
    Interval{ 0x119D4, 0x119D7 }, // Mn   [4] NANDINAGARI VOWEL SIGN U..NANDINAGARI VOWEL SIGN VOCALIC RR
    Interval{ 0x119DA, 0x119DB }, // Mn   [2] NANDINAGARI VOWEL SIGN E..NANDINAGARI VOWEL SIGN AI
    Interval{ 0x119E0, 0x119E0 }, // Mn       NANDINAGARI SIGN VIRAMA
    Interval{ 0x11A01, 0x11A0A }, // Mn  [10] ZANABAZAR SQUARE VOWEL SIGN I..ZANABAZAR SQUARE VOWEL LENGTH MARK
    Interval{ 0x11A33, 0x11A38 }, // Mn   [6] ZANABAZAR SQUARE FINAL CONSONANT MARK..ZANABAZAR SQUARE SIGN ANUSVARA
    Interval{ 0x11A3B, 0x11A3E }, // Mn   [4] ZANABAZAR SQUARE CLUSTER-FINAL LETTER YA..ZANABAZAR SQUARE CLUSTER-FINAL LETTER VA
    Interval{ 0x11A47, 0x11A47 }, // Mn       ZANABAZAR SQUARE SUBJOINER
    Interval{ 0x11A51, 0x11A56 }, // Mn   [6] SOYOMBO VOWEL SIGN I..SOYOMBO VOWEL SIGN OE
    Interval{ 0x11A59, 0x11A5B }, // Mn   [3] SOYOMBO VOWEL SIGN VOCALIC R..SOYOMBO VOWEL LENGTH MARK
    Interval{ 0x11A8A, 0x11A96 }, // Mn  [13] SOYOMBO FINAL CONSONANT SIGN G..SOYOMBO SIGN ANUSVARA
    Interval{ 0x11A98, 0x11A99 }, // Mn   [2] SOYOMBO GEMINATION MARK..SOYOMBO SUBJOINER
    Interval{ 0x11B60, 0x11B60 }, // Mn       SHARADA VOWEL SIGN OE
    Interval{ 0x11B62, 0x11B64 }, // Mn   [3] SHARADA VOWEL SIGN UE..SHARADA VOWEL SIGN SHORT E
    Interval{ 0x11B66, 0x11B66 }, // Mn       SHARADA VOWEL SIGN CANDRA E
    Interval{ 0x11C30, 0x11C36 }, // Mn   [7] BHAIKSUKI VOWEL SIGN I..BHAIKSUKI VOWEL SIGN VOCALIC L
    Interval{ 0x11C38, 0x11C3D }, // Mn   [6] BHAIKSUKI VOWEL SIGN E..BHAIKSUKI SIGN ANUSVARA
    Interval{ 0x11C3F, 0x11C3F }, // Mn       BHAIKSUKI SIGN VIRAMA
    Interval{ 0x11C92, 0x11CA7 }, // Mn  [22] MARCHEN SUBJOINED LETTER KA..MARCHEN SUBJOINED LETTER ZA
    Interval{ 0x11CAA, 0x11CB0 }, // Mn   [7] MARCHEN SUBJOINED LETTER RA..MARCHEN VOWEL SIGN AA
    Interval{ 0x11CB2, 0x11CB3 }, // Mn   [2] MARCHEN VOWEL SIGN U..MARCHEN VOWEL SIGN E
    Interval{ 0x11CB5, 0x11CB6 }, // Mn   [2] MARCHEN SIGN ANUSVARA..MARCHEN SIGN CANDRABINDU
    Interval{ 0x11D31, 0x11D36 }, // Mn   [6] MASARAM GONDI VOWEL SIGN AA..MASARAM GONDI VOWEL SIGN VOCALIC R
    Interval{ 0x11D3A, 0x11D3A }, // Mn       MASARAM GONDI VOWEL SIGN E
    Interval{ 0x11D3C, 0x11D3D }, // Mn   [2] MASARAM GONDI VOWEL SIGN AI..MASARAM GONDI VOWEL SIGN O
    Interval{ 0x11D3F, 0x11D45 }, // Mn   [7] MASARAM GONDI VOWEL SIGN AU..MASARAM GONDI VIRAMA
    Interval{ 0x11D47, 0x11D47 }, // Mn       MASARAM GONDI RA-KARA
    Interval{ 0x11D90, 0x11D91 }, // Mn   [2] GUNJALA GONDI VOWEL SIGN EE..GUNJALA GONDI VOWEL SIGN AI
    Interval{ 0x11D95, 0x11D95 }, // Mn       GUNJALA GONDI SIGN ANUSVARA
    Interval{ 0x11D97, 0x11D97 }, // Mn       GUNJALA GONDI VIRAMA
    Interval{ 0x11EF3, 0x11EF4 }, // Mn   [2] MAKASAR VOWEL SIGN I..MAKASAR VOWEL SIGN U
    Interval{ 0x11F00, 0x11F01 }, // Mn   [2] KAWI SIGN CANDRABINDU..KAWI SIGN ANUSVARA
    Interval{ 0x11F36, 0x11F3A }, // Mn   [5] KAWI VOWEL SIGN I..KAWI VOWEL SIGN VOCALIC R
    Interval{ 0x11F40, 0x11F40 }, // Mn       KAWI VOWEL SIGN EU
    Interval{ 0x11F41, 0x11F41 }, // Mc       KAWI SIGN KILLER
    Interval{ 0x11F42, 0x11F42 }, // Mn       KAWI CONJOINER
    Interval{ 0x11F5A, 0x11F5A }, // Mn       KAWI SIGN NUKTA
    Interval{ 0x13440, 0x13440 }, // Mn       EGYPTIAN HIEROGLYPH MIRROR HORIZONTALLY
    Interval{ 0x13447, 0x13455 }, // Mn  [15] EGYPTIAN HIEROGLYPH MODIFIER DAMAGED AT TOP START..EGYPTIAN HIEROGLYPH MODIFIER DAMAGED
    Interval{ 0x1611E, 0x16129 }, // Mn  [12] GURUNG KHEMA VOWEL SIGN AA..GURUNG KHEMA VOWEL LENGTH MARK
    Interval{ 0x1612D, 0x1612F }, // Mn   [3] GURUNG KHEMA SIGN ANUSVARA..GURUNG KHEMA SIGN THOLHOMA
    Interval{ 0x16AF0, 0x16AF4 }, // Mn   [5] BASSA VAH COMBINING HIGH TONE..BASSA VAH COMBINING HIGH-LOW TONE
    Interval{ 0x16B30, 0x16B36 }, // Mn   [7] PAHAWH HMONG MARK CIM TUB..PAHAWH HMONG MARK CIM TAUM
    Interval{ 0x16F4F, 0x16F4F }, // Mn       MIAO SIGN CONSONANT MODIFIER BAR
    Interval{ 0x16F8F, 0x16F92 }, // Mn   [4] MIAO TONE RIGHT..MIAO TONE BELOW
    Interval{ 0x16FE4, 0x16FE4 }, // Mn       KHITAN SMALL SCRIPT FILLER
    Interval{ 0x16FF0, 0x16FF1 }, // Mc   [2] VIETNAMESE ALTERNATE READING MARK CA..VIETNAMESE ALTERNATE READING MARK NHAY
    Interval{ 0x1BC9D, 0x1BC9E }, // Mn   [2] DUPLOYAN THICK LETTER SELECTOR..DUPLOYAN DOUBLE MARK
    Interval{ 0x1CF00, 0x1CF2D }, // Mn  [46] ZNAMENNY COMBINING MARK GORAZDO NIZKO S KRYZHEM ON LEFT..ZNAMENNY COMBINING MARK KRYZH ON LEFT
    Interval{ 0x1CF30, 0x1CF46 }, // Mn  [23] ZNAMENNY COMBINING TONAL RANGE MARK MRACHNO..ZNAMENNY PRIZNAK MODIFIER ROG
    Interval{ 0x1D165, 0x1D166 }, // Mc   [2] MUSICAL SYMBOL COMBINING STEM..MUSICAL SYMBOL COMBINING SPRECHGESANG STEM
    Interval{ 0x1D167, 0x1D169 }, // Mn   [3] MUSICAL SYMBOL COMBINING TREMOLO-1..MUSICAL SYMBOL COMBINING TREMOLO-3
    Interval{ 0x1D16D, 0x1D172 }, // Mc   [6] MUSICAL SYMBOL COMBINING AUGMENTATION DOT..MUSICAL SYMBOL COMBINING FLAG-5
    Interval{ 0x1D17B, 0x1D182 }, // Mn   [8] MUSICAL SYMBOL COMBINING ACCENT..MUSICAL SYMBOL COMBINING LOURE
    Interval{ 0x1D185, 0x1D18B }, // Mn   [7] MUSICAL SYMBOL COMBINING DOIT..MUSICAL SYMBOL COMBINING TRIPLE TONGUE
    Interval{ 0x1D1AA, 0x1D1AD }, // Mn   [4] MUSICAL SYMBOL COMBINING DOWN BOW..MUSICAL SYMBOL COMBINING SNAP PIZZICATO
    Interval{ 0x1D242, 0x1D244 }, // Mn   [3] COMBINING GREEK MUSICAL TRISEME..COMBINING GREEK MUSICAL PENTASEME
    Interval{ 0x1DA00, 0x1DA36 }, // Mn  [55] SIGNWRITING HEAD RIM..SIGNWRITING AIR SUCKING IN
    Interval{ 0x1DA3B, 0x1DA6C }, // Mn  [50] SIGNWRITING MOUTH CLOSED NEUTRAL..SIGNWRITING EXCITEMENT
    Interval{ 0x1DA75, 0x1DA75 }, // Mn       SIGNWRITING UPPER BODY TILTING FROM HIP JOINTS
    Interval{ 0x1DA84, 0x1DA84 }, // Mn       SIGNWRITING LOCATION HEAD NECK
    Interval{ 0x1DA9B, 0x1DA9F }, // Mn   [5] SIGNWRITING FILL MODIFIER-2..SIGNWRITING FILL MODIFIER-6
    Interval{ 0x1DAA1, 0x1DAAF }, // Mn  [15] SIGNWRITING ROTATION MODIFIER-2..SIGNWRITING ROTATION MODIFIER-16
    Interval{ 0x1E000, 0x1E006 }, // Mn   [7] COMBINING GLAGOLITIC LETTER AZU..COMBINING GLAGOLITIC LETTER ZHIVETE
    Interval{ 0x1E008, 0x1E018 }, // Mn  [17] COMBINING GLAGOLITIC LETTER ZEMLJA..COMBINING GLAGOLITIC LETTER HERU
    Interval{ 0x1E01B, 0x1E021 }, // Mn   [7] COMBINING GLAGOLITIC LETTER SHTA..COMBINING GLAGOLITIC LETTER YATI
    Interval{ 0x1E023, 0x1E024 }, // Mn   [2] COMBINING GLAGOLITIC LETTER YU..COMBINING GLAGOLITIC LETTER SMALL YUS
    Interval{ 0x1E026, 0x1E02A }, // Mn   [5] COMBINING GLAGOLITIC LETTER YO..COMBINING GLAGOLITIC LETTER FITA
    Interval{ 0x1E08F, 0x1E08F }, // Mn       COMBINING CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
    Interval{ 0x1E130, 0x1E136 }, // Mn   [7] NYIAKENG PUACHUE HMONG TONE-B..NYIAKENG PUACHUE HMONG TONE-D
    Interval{ 0x1E2AE, 0x1E2AE }, // Mn       TOTO SIGN RISING TONE
    Interval{ 0x1E2EC, 0x1E2EF }, // Mn   [4] WANCHO TONE TUP..WANCHO TONE KOINI
    Interval{ 0x1E4EC, 0x1E4EF }, // Mn   [4] NAG MUNDARI SIGN MUHOR..NAG MUNDARI SIGN SUTUH
    Interval{ 0x1E5EE, 0x1E5EF }, // Mn   [2] OL ONAL SIGN MU..OL ONAL SIGN IKIR
    Interval{ 0x1E6E3, 0x1E6E3 }, // Mn       TAI YO SIGN UE
    Interval{ 0x1E6E6, 0x1E6E6 }, // Mn       TAI YO SIGN AU
    Interval{ 0x1E6EE, 0x1E6EF }, // Mn   [2] TAI YO SIGN AY..TAI YO SIGN ANG
    Interval{ 0x1E6F5, 0x1E6F5 }, // Mn       TAI YO SIGN OM
    Interval{ 0x1E8D0, 0x1E8D6 }, // Mn   [7] MENDE KIKAKUI COMBINING NUMBER TEENS..MENDE KIKAKUI COMBINING NUMBER MILLIONS
    Interval{ 0x1E944, 0x1E94A }, // Mn   [7] ADLAM ALIF LENGTHENER..ADLAM NUKTA
    Interval{ 0xE0020, 0xE007F }, // Cf  [96] TAG SPACE..CANCEL TAG
    Interval{ 0xE0100, 0xE01EF }, // Mn [240] VARIATION SELECTOR-17..VARIATION SELECTOR-256
}; // }}}
auto static const Grapheme_Link = std::array<Interval, 64>{ // {{{
    Interval{ 0x094D, 0x094D }, // Mn       DEVANAGARI SIGN VIRAMA
    Interval{ 0x09CD, 0x09CD }, // Mn       BENGALI SIGN VIRAMA
    Interval{ 0x0A4D, 0x0A4D }, // Mn       GURMUKHI SIGN VIRAMA
    Interval{ 0x0ACD, 0x0ACD }, // Mn       GUJARATI SIGN VIRAMA
    Interval{ 0x0B4D, 0x0B4D }, // Mn       ORIYA SIGN VIRAMA
    Interval{ 0x0BCD, 0x0BCD }, // Mn       TAMIL SIGN VIRAMA
    Interval{ 0x0C4D, 0x0C4D }, // Mn       TELUGU SIGN VIRAMA
    Interval{ 0x0CCD, 0x0CCD }, // Mn       KANNADA SIGN VIRAMA
    Interval{ 0x0D3B, 0x0D3C }, // Mn   [2] MALAYALAM SIGN VERTICAL BAR VIRAMA..MALAYALAM SIGN CIRCULAR VIRAMA
    Interval{ 0x0D4D, 0x0D4D }, // Mn       MALAYALAM SIGN VIRAMA
    Interval{ 0x0DCA, 0x0DCA }, // Mn       SINHALA SIGN AL-LAKUNA
    Interval{ 0x0E3A, 0x0E3A }, // Mn       THAI CHARACTER PHINTHU
    Interval{ 0x0EBA, 0x0EBA }, // Mn       LAO SIGN PALI VIRAMA
    Interval{ 0x0F84, 0x0F84 }, // Mn       TIBETAN MARK HALANTA
    Interval{ 0x1039, 0x103A }, // Mn   [2] MYANMAR SIGN VIRAMA..MYANMAR SIGN ASAT
    Interval{ 0x1714, 0x1714 }, // Mn       TAGALOG SIGN VIRAMA
    Interval{ 0x1715, 0x1715 }, // Mc       TAGALOG SIGN PAMUDPOD
    Interval{ 0x1734, 0x1734 }, // Mc       HANUNOO SIGN PAMUDPOD
    Interval{ 0x17D2, 0x17D2 }, // Mn       KHMER SIGN COENG
    Interval{ 0x1A60, 0x1A60 }, // Mn       TAI THAM SIGN SAKOT
    Interval{ 0x1B44, 0x1B44 }, // Mc       BALINESE ADEG ADEG
    Interval{ 0x1BAA, 0x1BAA }, // Mc       SUNDANESE SIGN PAMAAEH
    Interval{ 0x1BAB, 0x1BAB }, // Mn       SUNDANESE SIGN VIRAMA
    Interval{ 0x1BF2, 0x1BF3 }, // Mc   [2] BATAK PANGOLAT..BATAK PANONGONAN
    Interval{ 0x2D7F, 0x2D7F }, // Mn       TIFINAGH CONSONANT JOINER
    Interval{ 0xA806, 0xA806 }, // Mn       SYLOTI NAGRI SIGN HASANTA
    Interval{ 0xA82C, 0xA82C }, // Mn       SYLOTI NAGRI SIGN ALTERNATE HASANTA
    Interval{ 0xA8C4, 0xA8C4 }, // Mn       SAURASHTRA SIGN VIRAMA
    Interval{ 0xA953, 0xA953 }, // Mc       REJANG VIRAMA
    Interval{ 0xA9C0, 0xA9C0 }, // Mc       JAVANESE PANGKON
    Interval{ 0xAAF6, 0xAAF6 }, // Mn       MEETEI MAYEK VIRAMA
    Interval{ 0xABED, 0xABED }, // Mn       MEETEI MAYEK APUN IYEK
    Interval{ 0x10A3F, 0x10A3F }, // Mn       KHAROSHTHI VIRAMA
    Interval{ 0x11046, 0x11046 }, // Mn       BRAHMI VIRAMA
    Interval{ 0x11070, 0x11070 }, // Mn       BRAHMI SIGN OLD TAMIL VIRAMA
    Interval{ 0x1107F, 0x1107F }, // Mn       BRAHMI NUMBER JOINER
    Interval{ 0x110B9, 0x110B9 }, // Mn       KAITHI SIGN VIRAMA
    Interval{ 0x11133, 0x11134 }, // Mn   [2] CHAKMA VIRAMA..CHAKMA MAAYYAA
    Interval{ 0x111C0, 0x111C0 }, // Mc       SHARADA SIGN VIRAMA
    Interval{ 0x11235, 0x11235 }, // Mc       KHOJKI SIGN VIRAMA
    Interval{ 0x112EA, 0x112EA }, // Mn       KHUDAWADI SIGN VIRAMA
    Interval{ 0x1134D, 0x1134D }, // Mc       GRANTHA SIGN VIRAMA
    Interval{ 0x113CE, 0x113CE }, // Mn       TULU-TIGALARI SIGN VIRAMA
    Interval{ 0x113CF, 0x113CF }, // Mc       TULU-TIGALARI SIGN LOOPED VIRAMA
    Interval{ 0x113D0, 0x113D0 }, // Mn       TULU-TIGALARI CONJOINER
    Interval{ 0x11442, 0x11442 }, // Mn       NEWA SIGN VIRAMA
    Interval{ 0x114C2, 0x114C2 }, // Mn       TIRHUTA SIGN VIRAMA
    Interval{ 0x115BF, 0x115BF }, // Mn       SIDDHAM SIGN VIRAMA
    Interval{ 0x1163F, 0x1163F }, // Mn       MODI SIGN VIRAMA
    Interval{ 0x116B6, 0x116B6 }, // Mc       TAKRI SIGN VIRAMA
    Interval{ 0x1172B, 0x1172B }, // Mn       AHOM SIGN KILLER
    Interval{ 0x11839, 0x11839 }, // Mn       DOGRA SIGN VIRAMA
    Interval{ 0x1193D, 0x1193D }, // Mc       DIVES AKURU SIGN HALANTA
    Interval{ 0x1193E, 0x1193E }, // Mn       DIVES AKURU VIRAMA
    Interval{ 0x119E0, 0x119E0 }, // Mn       NANDINAGARI SIGN VIRAMA
    Interval{ 0x11A34, 0x11A34 }, // Mn       ZANABAZAR SQUARE SIGN VIRAMA
    Interval{ 0x11A47, 0x11A47 }, // Mn       ZANABAZAR SQUARE SUBJOINER
    Interval{ 0x11A99, 0x11A99 }, // Mn       SOYOMBO SUBJOINER
    Interval{ 0x11C3F, 0x11C3F }, // Mn       BHAIKSUKI SIGN VIRAMA
    Interval{ 0x11D44, 0x11D45 }, // Mn   [2] MASARAM GONDI SIGN HALANTA..MASARAM GONDI VIRAMA
    Interval{ 0x11D97, 0x11D97 }, // Mn       GUNJALA GONDI VIRAMA
    Interval{ 0x11F41, 0x11F41 }, // Mc       KAWI SIGN KILLER
    Interval{ 0x11F42, 0x11F42 }, // Mn       KAWI CONJOINER
    Interval{ 0x1612F, 0x1612F }, // Mn       GURUNG KHEMA SIGN THOLHOMA
}; // }}}
auto static const ID_Continue = std::array<Interval, 1418>{ // {{{
    Interval{ 0x0030, 0x0039 }, // Nd  [10] DIGIT ZERO..DIGIT NINE
    Interval{ 0x0041, 0x005A }, // L&  [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Interval{ 0x005F, 0x005F }, // Pc       LOW LINE
    Interval{ 0x0061, 0x007A }, // L&  [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Interval{ 0x00AA, 0x00AA }, // Lo       FEMININE ORDINAL INDICATOR
    Interval{ 0x00B5, 0x00B5 }, // L&       MICRO SIGN
    Interval{ 0x00B7, 0x00B7 }, // Po       MIDDLE DOT
    Interval{ 0x00BA, 0x00BA }, // Lo       MASCULINE ORDINAL INDICATOR
    Interval{ 0x00C0, 0x00D6 }, // L&  [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x00D8, 0x00F6 }, // L&  [31] LATIN CAPITAL LETTER O WITH STROKE..LATIN SMALL LETTER O WITH DIAERESIS
    Interval{ 0x00F8, 0x01BA }, // L& [195] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER EZH WITH TAIL
    Interval{ 0x01BB, 0x01BB }, // Lo       LATIN LETTER TWO WITH STROKE
    Interval{ 0x01BC, 0x01BF }, // L&   [4] LATIN CAPITAL LETTER TONE FIVE..LATIN LETTER WYNN
    Interval{ 0x01C0, 0x01C3 }, // Lo   [4] LATIN LETTER DENTAL CLICK..LATIN LETTER RETROFLEX CLICK
    Interval{ 0x01C4, 0x0293 }, // L& [208] LATIN CAPITAL LETTER DZ WITH CARON..LATIN SMALL LETTER EZH WITH CURL
    Interval{ 0x0294, 0x0295 }, // Lo   [2] LATIN LETTER GLOTTAL STOP..LATIN LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0x0296, 0x02AF }, // L&  [26] LATIN LETTER INVERTED GLOTTAL STOP..LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    Interval{ 0x02B0, 0x02C1 }, // Lm  [18] MODIFIER LETTER SMALL H..MODIFIER LETTER REVERSED GLOTTAL STOP
    Interval{ 0x02C6, 0x02D1 }, // Lm  [12] MODIFIER LETTER CIRCUMFLEX ACCENT..MODIFIER LETTER HALF TRIANGULAR COLON
    Interval{ 0x02E0, 0x02E4 }, // Lm   [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    Interval{ 0x02EC, 0x02EC }, // Lm       MODIFIER LETTER VOICING
    Interval{ 0x02EE, 0x02EE }, // Lm       MODIFIER LETTER DOUBLE APOSTROPHE
    Interval{ 0x0300, 0x036F }, // Mn [112] COMBINING GRAVE ACCENT..COMBINING LATIN SMALL LETTER X
    Interval{ 0x0370, 0x0373 }, // L&   [4] GREEK CAPITAL LETTER HETA..GREEK SMALL LETTER ARCHAIC SAMPI
    Interval{ 0x0374, 0x0374 }, // Lm       GREEK NUMERAL SIGN
    Interval{ 0x0376, 0x0377 }, // L&   [2] GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA..GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037A, 0x037A }, // Lm       GREEK YPOGEGRAMMENI
    Interval{ 0x037B, 0x037D }, // L&   [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Interval{ 0x037F, 0x037F }, // L&       GREEK CAPITAL LETTER YOT
    Interval{ 0x0386, 0x0386 }, // L&       GREEK CAPITAL LETTER ALPHA WITH TONOS
    Interval{ 0x0387, 0x0387 }, // Po       GREEK ANO TELEIA
    Interval{ 0x0388, 0x038A }, // L&   [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Interval{ 0x038C, 0x038C }, // L&       GREEK CAPITAL LETTER OMICRON WITH TONOS
    Interval{ 0x038E, 0x03A1 }, // L&  [20] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER RHO
    Interval{ 0x03A3, 0x03F5 }, // L&  [83] GREEK CAPITAL LETTER SIGMA..GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F7, 0x0481 }, // L& [139] GREEK CAPITAL LETTER SHO..CYRILLIC SMALL LETTER KOPPA
    Interval{ 0x0483, 0x0487 }, // Mn   [5] COMBINING CYRILLIC TITLO..COMBINING CYRILLIC POKRYTIE
    Interval{ 0x048A, 0x052F }, // L& [166] CYRILLIC CAPITAL LETTER SHORT I WITH TAIL..CYRILLIC SMALL LETTER EL WITH DESCENDER
    Interval{ 0x0531, 0x0556 }, // L&  [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Interval{ 0x0559, 0x0559 }, // Lm       ARMENIAN MODIFIER LETTER LEFT HALF RING
    Interval{ 0x0560, 0x0588 }, // L&  [41] ARMENIAN SMALL LETTER TURNED AYB..ARMENIAN SMALL LETTER YI WITH STROKE
    Interval{ 0x0591, 0x05BD }, // Mn  [45] HEBREW ACCENT ETNAHTA..HEBREW POINT METEG
    Interval{ 0x05BF, 0x05BF }, // Mn       HEBREW POINT RAFE
    Interval{ 0x05C1, 0x05C2 }, // Mn   [2] HEBREW POINT SHIN DOT..HEBREW POINT SIN DOT
    Interval{ 0x05C4, 0x05C5 }, // Mn   [2] HEBREW MARK UPPER DOT..HEBREW MARK LOWER DOT
    Interval{ 0x05C7, 0x05C7 }, // Mn       HEBREW POINT QAMATS QATAN
    Interval{ 0x05D0, 0x05EA }, // Lo  [27] HEBREW LETTER ALEF..HEBREW LETTER TAV
    Interval{ 0x05EF, 0x05F2 }, // Lo   [4] HEBREW YOD TRIANGLE..HEBREW LIGATURE YIDDISH DOUBLE YOD
    Interval{ 0x0610, 0x061A }, // Mn  [11] ARABIC SIGN SALLALLAHOU ALAYHE WASSALLAM..ARABIC SMALL KASRA
    Interval{ 0x0620, 0x063F }, // Lo  [32] ARABIC LETTER KASHMIRI YEH..ARABIC LETTER FARSI YEH WITH THREE DOTS ABOVE
    Interval{ 0x0640, 0x0640 }, // Lm       ARABIC TATWEEL
    Interval{ 0x0641, 0x064A }, // Lo  [10] ARABIC LETTER FEH..ARABIC LETTER YEH
    Interval{ 0x064B, 0x065F }, // Mn  [21] ARABIC FATHATAN..ARABIC WAVY HAMZA BELOW
    Interval{ 0x0660, 0x0669 }, // Nd  [10] ARABIC-INDIC DIGIT ZERO..ARABIC-INDIC DIGIT NINE
    Interval{ 0x066E, 0x066F }, // Lo   [2] ARABIC LETTER DOTLESS BEH..ARABIC LETTER DOTLESS QAF
    Interval{ 0x0670, 0x0670 }, // Mn       ARABIC LETTER SUPERSCRIPT ALEF
    Interval{ 0x0671, 0x06D3 }, // Lo  [99] ARABIC LETTER ALEF WASLA..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE
    Interval{ 0x06D5, 0x06D5 }, // Lo       ARABIC LETTER AE
    Interval{ 0x06D6, 0x06DC }, // Mn   [7] ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA..ARABIC SMALL HIGH SEEN
    Interval{ 0x06DF, 0x06E4 }, // Mn   [6] ARABIC SMALL HIGH ROUNDED ZERO..ARABIC SMALL HIGH MADDA
    Interval{ 0x06E5, 0x06E6 }, // Lm   [2] ARABIC SMALL WAW..ARABIC SMALL YEH
    Interval{ 0x06E7, 0x06E8 }, // Mn   [2] ARABIC SMALL HIGH YEH..ARABIC SMALL HIGH NOON
    Interval{ 0x06EA, 0x06ED }, // Mn   [4] ARABIC EMPTY CENTRE LOW STOP..ARABIC SMALL LOW MEEM
    Interval{ 0x06EE, 0x06EF }, // Lo   [2] ARABIC LETTER DAL WITH INVERTED V..ARABIC LETTER REH WITH INVERTED V
    Interval{ 0x06F0, 0x06F9 }, // Nd  [10] EXTENDED ARABIC-INDIC DIGIT ZERO..EXTENDED ARABIC-INDIC DIGIT NINE
    Interval{ 0x06FA, 0x06FC }, // Lo   [3] ARABIC LETTER SHEEN WITH DOT BELOW..ARABIC LETTER GHAIN WITH DOT BELOW
    Interval{ 0x06FF, 0x06FF }, // Lo       ARABIC LETTER HEH WITH INVERTED V
    Interval{ 0x0710, 0x0710 }, // Lo       SYRIAC LETTER ALAPH
    Interval{ 0x0711, 0x0711 }, // Mn       SYRIAC LETTER SUPERSCRIPT ALAPH
    Interval{ 0x0712, 0x072F }, // Lo  [30] SYRIAC LETTER BETH..SYRIAC LETTER PERSIAN DHALATH
    Interval{ 0x0730, 0x074A }, // Mn  [27] SYRIAC PTHAHA ABOVE..SYRIAC BARREKH
    Interval{ 0x074D, 0x07A5 }, // Lo  [89] SYRIAC LETTER SOGDIAN ZHAIN..THAANA LETTER WAAVU
    Interval{ 0x07A6, 0x07B0 }, // Mn  [11] THAANA ABAFILI..THAANA SUKUN
    Interval{ 0x07B1, 0x07B1 }, // Lo       THAANA LETTER NAA
    Interval{ 0x07C0, 0x07C9 }, // Nd  [10] NKO DIGIT ZERO..NKO DIGIT NINE
    Interval{ 0x07CA, 0x07EA }, // Lo  [33] NKO LETTER A..NKO LETTER JONA RA
    Interval{ 0x07EB, 0x07F3 }, // Mn   [9] NKO COMBINING SHORT HIGH TONE..NKO COMBINING DOUBLE DOT ABOVE
    Interval{ 0x07F4, 0x07F5 }, // Lm   [2] NKO HIGH TONE APOSTROPHE..NKO LOW TONE APOSTROPHE
    Interval{ 0x07FA, 0x07FA }, // Lm       NKO LAJANYALAN
    Interval{ 0x07FD, 0x07FD }, // Mn       NKO DANTAYALAN
    Interval{ 0x0800, 0x0815 }, // Lo  [22] SAMARITAN LETTER ALAF..SAMARITAN LETTER TAAF
    Interval{ 0x0816, 0x0819 }, // Mn   [4] SAMARITAN MARK IN..SAMARITAN MARK DAGESH
    Interval{ 0x081A, 0x081A }, // Lm       SAMARITAN MODIFIER LETTER EPENTHETIC YUT
    Interval{ 0x081B, 0x0823 }, // Mn   [9] SAMARITAN MARK EPENTHETIC YUT..SAMARITAN VOWEL SIGN A
    Interval{ 0x0824, 0x0824 }, // Lm       SAMARITAN MODIFIER LETTER SHORT A
    Interval{ 0x0825, 0x0827 }, // Mn   [3] SAMARITAN VOWEL SIGN SHORT A..SAMARITAN VOWEL SIGN U
    Interval{ 0x0828, 0x0828 }, // Lm       SAMARITAN MODIFIER LETTER I
    Interval{ 0x0829, 0x082D }, // Mn   [5] SAMARITAN VOWEL SIGN LONG I..SAMARITAN MARK NEQUDAA
    Interval{ 0x0840, 0x0858 }, // Lo  [25] MANDAIC LETTER HALQA..MANDAIC LETTER AIN
    Interval{ 0x0859, 0x085B }, // Mn   [3] MANDAIC AFFRICATION MARK..MANDAIC GEMINATION MARK
    Interval{ 0x0860, 0x086A }, // Lo  [11] SYRIAC LETTER MALAYALAM NGA..SYRIAC LETTER MALAYALAM SSA
    Interval{ 0x0870, 0x0887 }, // Lo  [24] ARABIC LETTER ALEF WITH ATTACHED FATHA..ARABIC BASELINE ROUND DOT
    Interval{ 0x0889, 0x088F }, // Lo   [7] ARABIC LETTER NOON WITH INVERTED SMALL V..ARABIC LETTER NOON WITH RING ABOVE
    Interval{ 0x0897, 0x089F }, // Mn   [9] ARABIC PEPET..ARABIC HALF MADDA OVER MADDA
    Interval{ 0x08A0, 0x08C8 }, // Lo  [41] ARABIC LETTER BEH WITH SMALL V BELOW..ARABIC LETTER GRAF
    Interval{ 0x08C9, 0x08C9 }, // Lm       ARABIC SMALL FARSI YEH
    Interval{ 0x08CA, 0x08E1 }, // Mn  [24] ARABIC SMALL HIGH FARSI YEH..ARABIC SMALL HIGH SIGN SAFHA
    Interval{ 0x08E3, 0x0902 }, // Mn  [32] ARABIC TURNED DAMMA BELOW..DEVANAGARI SIGN ANUSVARA
    Interval{ 0x0903, 0x0903 }, // Mc       DEVANAGARI SIGN VISARGA
    Interval{ 0x0904, 0x0939 }, // Lo  [54] DEVANAGARI LETTER SHORT A..DEVANAGARI LETTER HA
    Interval{ 0x093A, 0x093A }, // Mn       DEVANAGARI VOWEL SIGN OE
    Interval{ 0x093B, 0x093B }, // Mc       DEVANAGARI VOWEL SIGN OOE
    Interval{ 0x093C, 0x093C }, // Mn       DEVANAGARI SIGN NUKTA
    Interval{ 0x093D, 0x093D }, // Lo       DEVANAGARI SIGN AVAGRAHA
    Interval{ 0x093E, 0x0940 }, // Mc   [3] DEVANAGARI VOWEL SIGN AA..DEVANAGARI VOWEL SIGN II
    Interval{ 0x0941, 0x0948 }, // Mn   [8] DEVANAGARI VOWEL SIGN U..DEVANAGARI VOWEL SIGN AI
    Interval{ 0x0949, 0x094C }, // Mc   [4] DEVANAGARI VOWEL SIGN CANDRA O..DEVANAGARI VOWEL SIGN AU
    Interval{ 0x094D, 0x094D }, // Mn       DEVANAGARI SIGN VIRAMA
    Interval{ 0x094E, 0x094F }, // Mc   [2] DEVANAGARI VOWEL SIGN PRISHTHAMATRA E..DEVANAGARI VOWEL SIGN AW
    Interval{ 0x0950, 0x0950 }, // Lo       DEVANAGARI OM
    Interval{ 0x0951, 0x0957 }, // Mn   [7] DEVANAGARI STRESS SIGN UDATTA..DEVANAGARI VOWEL SIGN UUE
    Interval{ 0x0958, 0x0961 }, // Lo  [10] DEVANAGARI LETTER QA..DEVANAGARI LETTER VOCALIC LL
    Interval{ 0x0962, 0x0963 }, // Mn   [2] DEVANAGARI VOWEL SIGN VOCALIC L..DEVANAGARI VOWEL SIGN VOCALIC LL
    Interval{ 0x0966, 0x096F }, // Nd  [10] DEVANAGARI DIGIT ZERO..DEVANAGARI DIGIT NINE
    Interval{ 0x0971, 0x0971 }, // Lm       DEVANAGARI SIGN HIGH SPACING DOT
    Interval{ 0x0972, 0x0980 }, // Lo  [15] DEVANAGARI LETTER CANDRA A..BENGALI ANJI
    Interval{ 0x0981, 0x0981 }, // Mn       BENGALI SIGN CANDRABINDU
    Interval{ 0x0982, 0x0983 }, // Mc   [2] BENGALI SIGN ANUSVARA..BENGALI SIGN VISARGA
    Interval{ 0x0985, 0x098C }, // Lo   [8] BENGALI LETTER A..BENGALI LETTER VOCALIC L
    Interval{ 0x098F, 0x0990 }, // Lo   [2] BENGALI LETTER E..BENGALI LETTER AI
    Interval{ 0x0993, 0x09A8 }, // Lo  [22] BENGALI LETTER O..BENGALI LETTER NA
    Interval{ 0x09AA, 0x09B0 }, // Lo   [7] BENGALI LETTER PA..BENGALI LETTER RA
    Interval{ 0x09B2, 0x09B2 }, // Lo       BENGALI LETTER LA
    Interval{ 0x09B6, 0x09B9 }, // Lo   [4] BENGALI LETTER SHA..BENGALI LETTER HA
    Interval{ 0x09BC, 0x09BC }, // Mn       BENGALI SIGN NUKTA
    Interval{ 0x09BD, 0x09BD }, // Lo       BENGALI SIGN AVAGRAHA
    Interval{ 0x09BE, 0x09C0 }, // Mc   [3] BENGALI VOWEL SIGN AA..BENGALI VOWEL SIGN II
    Interval{ 0x09C1, 0x09C4 }, // Mn   [4] BENGALI VOWEL SIGN U..BENGALI VOWEL SIGN VOCALIC RR
    Interval{ 0x09C7, 0x09C8 }, // Mc   [2] BENGALI VOWEL SIGN E..BENGALI VOWEL SIGN AI
    Interval{ 0x09CB, 0x09CC }, // Mc   [2] BENGALI VOWEL SIGN O..BENGALI VOWEL SIGN AU
    Interval{ 0x09CD, 0x09CD }, // Mn       BENGALI SIGN VIRAMA
    Interval{ 0x09CE, 0x09CE }, // Lo       BENGALI LETTER KHANDA TA
    Interval{ 0x09D7, 0x09D7 }, // Mc       BENGALI AU LENGTH MARK
    Interval{ 0x09DC, 0x09DD }, // Lo   [2] BENGALI LETTER RRA..BENGALI LETTER RHA
    Interval{ 0x09DF, 0x09E1 }, // Lo   [3] BENGALI LETTER YYA..BENGALI LETTER VOCALIC LL
    Interval{ 0x09E2, 0x09E3 }, // Mn   [2] BENGALI VOWEL SIGN VOCALIC L..BENGALI VOWEL SIGN VOCALIC LL
    Interval{ 0x09E6, 0x09EF }, // Nd  [10] BENGALI DIGIT ZERO..BENGALI DIGIT NINE
    Interval{ 0x09F0, 0x09F1 }, // Lo   [2] BENGALI LETTER RA WITH MIDDLE DIAGONAL..BENGALI LETTER RA WITH LOWER DIAGONAL
    Interval{ 0x09FC, 0x09FC }, // Lo       BENGALI LETTER VEDIC ANUSVARA
    Interval{ 0x09FE, 0x09FE }, // Mn       BENGALI SANDHI MARK
    Interval{ 0x0A01, 0x0A02 }, // Mn   [2] GURMUKHI SIGN ADAK BINDI..GURMUKHI SIGN BINDI
    Interval{ 0x0A03, 0x0A03 }, // Mc       GURMUKHI SIGN VISARGA
    Interval{ 0x0A05, 0x0A0A }, // Lo   [6] GURMUKHI LETTER A..GURMUKHI LETTER UU
    Interval{ 0x0A0F, 0x0A10 }, // Lo   [2] GURMUKHI LETTER EE..GURMUKHI LETTER AI
    Interval{ 0x0A13, 0x0A28 }, // Lo  [22] GURMUKHI LETTER OO..GURMUKHI LETTER NA
    Interval{ 0x0A2A, 0x0A30 }, // Lo   [7] GURMUKHI LETTER PA..GURMUKHI LETTER RA
    Interval{ 0x0A32, 0x0A33 }, // Lo   [2] GURMUKHI LETTER LA..GURMUKHI LETTER LLA
    Interval{ 0x0A35, 0x0A36 }, // Lo   [2] GURMUKHI LETTER VA..GURMUKHI LETTER SHA
    Interval{ 0x0A38, 0x0A39 }, // Lo   [2] GURMUKHI LETTER SA..GURMUKHI LETTER HA
    Interval{ 0x0A3C, 0x0A3C }, // Mn       GURMUKHI SIGN NUKTA
    Interval{ 0x0A3E, 0x0A40 }, // Mc   [3] GURMUKHI VOWEL SIGN AA..GURMUKHI VOWEL SIGN II
    Interval{ 0x0A41, 0x0A42 }, // Mn   [2] GURMUKHI VOWEL SIGN U..GURMUKHI VOWEL SIGN UU
    Interval{ 0x0A47, 0x0A48 }, // Mn   [2] GURMUKHI VOWEL SIGN EE..GURMUKHI VOWEL SIGN AI
    Interval{ 0x0A4B, 0x0A4D }, // Mn   [3] GURMUKHI VOWEL SIGN OO..GURMUKHI SIGN VIRAMA
    Interval{ 0x0A51, 0x0A51 }, // Mn       GURMUKHI SIGN UDAAT
    Interval{ 0x0A59, 0x0A5C }, // Lo   [4] GURMUKHI LETTER KHHA..GURMUKHI LETTER RRA
    Interval{ 0x0A5E, 0x0A5E }, // Lo       GURMUKHI LETTER FA
    Interval{ 0x0A66, 0x0A6F }, // Nd  [10] GURMUKHI DIGIT ZERO..GURMUKHI DIGIT NINE
    Interval{ 0x0A70, 0x0A71 }, // Mn   [2] GURMUKHI TIPPI..GURMUKHI ADDAK
    Interval{ 0x0A72, 0x0A74 }, // Lo   [3] GURMUKHI IRI..GURMUKHI EK ONKAR
    Interval{ 0x0A75, 0x0A75 }, // Mn       GURMUKHI SIGN YAKASH
    Interval{ 0x0A81, 0x0A82 }, // Mn   [2] GUJARATI SIGN CANDRABINDU..GUJARATI SIGN ANUSVARA
    Interval{ 0x0A83, 0x0A83 }, // Mc       GUJARATI SIGN VISARGA
    Interval{ 0x0A85, 0x0A8D }, // Lo   [9] GUJARATI LETTER A..GUJARATI VOWEL CANDRA E
    Interval{ 0x0A8F, 0x0A91 }, // Lo   [3] GUJARATI LETTER E..GUJARATI VOWEL CANDRA O
    Interval{ 0x0A93, 0x0AA8 }, // Lo  [22] GUJARATI LETTER O..GUJARATI LETTER NA
    Interval{ 0x0AAA, 0x0AB0 }, // Lo   [7] GUJARATI LETTER PA..GUJARATI LETTER RA
    Interval{ 0x0AB2, 0x0AB3 }, // Lo   [2] GUJARATI LETTER LA..GUJARATI LETTER LLA
    Interval{ 0x0AB5, 0x0AB9 }, // Lo   [5] GUJARATI LETTER VA..GUJARATI LETTER HA
    Interval{ 0x0ABC, 0x0ABC }, // Mn       GUJARATI SIGN NUKTA
    Interval{ 0x0ABD, 0x0ABD }, // Lo       GUJARATI SIGN AVAGRAHA
    Interval{ 0x0ABE, 0x0AC0 }, // Mc   [3] GUJARATI VOWEL SIGN AA..GUJARATI VOWEL SIGN II
    Interval{ 0x0AC1, 0x0AC5 }, // Mn   [5] GUJARATI VOWEL SIGN U..GUJARATI VOWEL SIGN CANDRA E
    Interval{ 0x0AC7, 0x0AC8 }, // Mn   [2] GUJARATI VOWEL SIGN E..GUJARATI VOWEL SIGN AI
    Interval{ 0x0AC9, 0x0AC9 }, // Mc       GUJARATI VOWEL SIGN CANDRA O
    Interval{ 0x0ACB, 0x0ACC }, // Mc   [2] GUJARATI VOWEL SIGN O..GUJARATI VOWEL SIGN AU
    Interval{ 0x0ACD, 0x0ACD }, // Mn       GUJARATI SIGN VIRAMA
    Interval{ 0x0AD0, 0x0AD0 }, // Lo       GUJARATI OM
    Interval{ 0x0AE0, 0x0AE1 }, // Lo   [2] GUJARATI LETTER VOCALIC RR..GUJARATI LETTER VOCALIC LL
    Interval{ 0x0AE2, 0x0AE3 }, // Mn   [2] GUJARATI VOWEL SIGN VOCALIC L..GUJARATI VOWEL SIGN VOCALIC LL
    Interval{ 0x0AE6, 0x0AEF }, // Nd  [10] GUJARATI DIGIT ZERO..GUJARATI DIGIT NINE
    Interval{ 0x0AF9, 0x0AF9 }, // Lo       GUJARATI LETTER ZHA
    Interval{ 0x0AFA, 0x0AFF }, // Mn   [6] GUJARATI SIGN SUKUN..GUJARATI SIGN TWO-CIRCLE NUKTA ABOVE
    Interval{ 0x0B01, 0x0B01 }, // Mn       ORIYA SIGN CANDRABINDU
    Interval{ 0x0B02, 0x0B03 }, // Mc   [2] ORIYA SIGN ANUSVARA..ORIYA SIGN VISARGA
    Interval{ 0x0B05, 0x0B0C }, // Lo   [8] ORIYA LETTER A..ORIYA LETTER VOCALIC L
    Interval{ 0x0B0F, 0x0B10 }, // Lo   [2] ORIYA LETTER E..ORIYA LETTER AI
    Interval{ 0x0B13, 0x0B28 }, // Lo  [22] ORIYA LETTER O..ORIYA LETTER NA
    Interval{ 0x0B2A, 0x0B30 }, // Lo   [7] ORIYA LETTER PA..ORIYA LETTER RA
    Interval{ 0x0B32, 0x0B33 }, // Lo   [2] ORIYA LETTER LA..ORIYA LETTER LLA
    Interval{ 0x0B35, 0x0B39 }, // Lo   [5] ORIYA LETTER VA..ORIYA LETTER HA
    Interval{ 0x0B3C, 0x0B3C }, // Mn       ORIYA SIGN NUKTA
    Interval{ 0x0B3D, 0x0B3D }, // Lo       ORIYA SIGN AVAGRAHA
    Interval{ 0x0B3E, 0x0B3E }, // Mc       ORIYA VOWEL SIGN AA
    Interval{ 0x0B3F, 0x0B3F }, // Mn       ORIYA VOWEL SIGN I
    Interval{ 0x0B40, 0x0B40 }, // Mc       ORIYA VOWEL SIGN II
    Interval{ 0x0B41, 0x0B44 }, // Mn   [4] ORIYA VOWEL SIGN U..ORIYA VOWEL SIGN VOCALIC RR
    Interval{ 0x0B47, 0x0B48 }, // Mc   [2] ORIYA VOWEL SIGN E..ORIYA VOWEL SIGN AI
    Interval{ 0x0B4B, 0x0B4C }, // Mc   [2] ORIYA VOWEL SIGN O..ORIYA VOWEL SIGN AU
    Interval{ 0x0B4D, 0x0B4D }, // Mn       ORIYA SIGN VIRAMA
    Interval{ 0x0B55, 0x0B56 }, // Mn   [2] ORIYA SIGN OVERLINE..ORIYA AI LENGTH MARK
    Interval{ 0x0B57, 0x0B57 }, // Mc       ORIYA AU LENGTH MARK
    Interval{ 0x0B5C, 0x0B5D }, // Lo   [2] ORIYA LETTER RRA..ORIYA LETTER RHA
    Interval{ 0x0B5F, 0x0B61 }, // Lo   [3] ORIYA LETTER YYA..ORIYA LETTER VOCALIC LL
    Interval{ 0x0B62, 0x0B63 }, // Mn   [2] ORIYA VOWEL SIGN VOCALIC L..ORIYA VOWEL SIGN VOCALIC LL
    Interval{ 0x0B66, 0x0B6F }, // Nd  [10] ORIYA DIGIT ZERO..ORIYA DIGIT NINE
    Interval{ 0x0B71, 0x0B71 }, // Lo       ORIYA LETTER WA
    Interval{ 0x0B82, 0x0B82 }, // Mn       TAMIL SIGN ANUSVARA
    Interval{ 0x0B83, 0x0B83 }, // Lo       TAMIL SIGN VISARGA
    Interval{ 0x0B85, 0x0B8A }, // Lo   [6] TAMIL LETTER A..TAMIL LETTER UU
    Interval{ 0x0B8E, 0x0B90 }, // Lo   [3] TAMIL LETTER E..TAMIL LETTER AI
    Interval{ 0x0B92, 0x0B95 }, // Lo   [4] TAMIL LETTER O..TAMIL LETTER KA
    Interval{ 0x0B99, 0x0B9A }, // Lo   [2] TAMIL LETTER NGA..TAMIL LETTER CA
    Interval{ 0x0B9C, 0x0B9C }, // Lo       TAMIL LETTER JA
    Interval{ 0x0B9E, 0x0B9F }, // Lo   [2] TAMIL LETTER NYA..TAMIL LETTER TTA
    Interval{ 0x0BA3, 0x0BA4 }, // Lo   [2] TAMIL LETTER NNA..TAMIL LETTER TA
    Interval{ 0x0BA8, 0x0BAA }, // Lo   [3] TAMIL LETTER NA..TAMIL LETTER PA
    Interval{ 0x0BAE, 0x0BB9 }, // Lo  [12] TAMIL LETTER MA..TAMIL LETTER HA
    Interval{ 0x0BBE, 0x0BBF }, // Mc   [2] TAMIL VOWEL SIGN AA..TAMIL VOWEL SIGN I
    Interval{ 0x0BC0, 0x0BC0 }, // Mn       TAMIL VOWEL SIGN II
    Interval{ 0x0BC1, 0x0BC2 }, // Mc   [2] TAMIL VOWEL SIGN U..TAMIL VOWEL SIGN UU
    Interval{ 0x0BC6, 0x0BC8 }, // Mc   [3] TAMIL VOWEL SIGN E..TAMIL VOWEL SIGN AI
    Interval{ 0x0BCA, 0x0BCC }, // Mc   [3] TAMIL VOWEL SIGN O..TAMIL VOWEL SIGN AU
    Interval{ 0x0BCD, 0x0BCD }, // Mn       TAMIL SIGN VIRAMA
    Interval{ 0x0BD0, 0x0BD0 }, // Lo       TAMIL OM
    Interval{ 0x0BD7, 0x0BD7 }, // Mc       TAMIL AU LENGTH MARK
    Interval{ 0x0BE6, 0x0BEF }, // Nd  [10] TAMIL DIGIT ZERO..TAMIL DIGIT NINE
    Interval{ 0x0C00, 0x0C00 }, // Mn       TELUGU SIGN COMBINING CANDRABINDU ABOVE
    Interval{ 0x0C01, 0x0C03 }, // Mc   [3] TELUGU SIGN CANDRABINDU..TELUGU SIGN VISARGA
    Interval{ 0x0C04, 0x0C04 }, // Mn       TELUGU SIGN COMBINING ANUSVARA ABOVE
    Interval{ 0x0C05, 0x0C0C }, // Lo   [8] TELUGU LETTER A..TELUGU LETTER VOCALIC L
    Interval{ 0x0C0E, 0x0C10 }, // Lo   [3] TELUGU LETTER E..TELUGU LETTER AI
    Interval{ 0x0C12, 0x0C28 }, // Lo  [23] TELUGU LETTER O..TELUGU LETTER NA
    Interval{ 0x0C2A, 0x0C39 }, // Lo  [16] TELUGU LETTER PA..TELUGU LETTER HA
    Interval{ 0x0C3C, 0x0C3C }, // Mn       TELUGU SIGN NUKTA
    Interval{ 0x0C3D, 0x0C3D }, // Lo       TELUGU SIGN AVAGRAHA
    Interval{ 0x0C3E, 0x0C40 }, // Mn   [3] TELUGU VOWEL SIGN AA..TELUGU VOWEL SIGN II
    Interval{ 0x0C41, 0x0C44 }, // Mc   [4] TELUGU VOWEL SIGN U..TELUGU VOWEL SIGN VOCALIC RR
    Interval{ 0x0C46, 0x0C48 }, // Mn   [3] TELUGU VOWEL SIGN E..TELUGU VOWEL SIGN AI
    Interval{ 0x0C4A, 0x0C4D }, // Mn   [4] TELUGU VOWEL SIGN O..TELUGU SIGN VIRAMA
    Interval{ 0x0C55, 0x0C56 }, // Mn   [2] TELUGU LENGTH MARK..TELUGU AI LENGTH MARK
    Interval{ 0x0C58, 0x0C5A }, // Lo   [3] TELUGU LETTER TSA..TELUGU LETTER RRRA
    Interval{ 0x0C5C, 0x0C5D }, // Lo   [2] TELUGU ARCHAIC SHRII..TELUGU LETTER NAKAARA POLLU
    Interval{ 0x0C60, 0x0C61 }, // Lo   [2] TELUGU LETTER VOCALIC RR..TELUGU LETTER VOCALIC LL
    Interval{ 0x0C62, 0x0C63 }, // Mn   [2] TELUGU VOWEL SIGN VOCALIC L..TELUGU VOWEL SIGN VOCALIC LL
    Interval{ 0x0C66, 0x0C6F }, // Nd  [10] TELUGU DIGIT ZERO..TELUGU DIGIT NINE
    Interval{ 0x0C80, 0x0C80 }, // Lo       KANNADA SIGN SPACING CANDRABINDU
    Interval{ 0x0C81, 0x0C81 }, // Mn       KANNADA SIGN CANDRABINDU
    Interval{ 0x0C82, 0x0C83 }, // Mc   [2] KANNADA SIGN ANUSVARA..KANNADA SIGN VISARGA
    Interval{ 0x0C85, 0x0C8C }, // Lo   [8] KANNADA LETTER A..KANNADA LETTER VOCALIC L
    Interval{ 0x0C8E, 0x0C90 }, // Lo   [3] KANNADA LETTER E..KANNADA LETTER AI
    Interval{ 0x0C92, 0x0CA8 }, // Lo  [23] KANNADA LETTER O..KANNADA LETTER NA
    Interval{ 0x0CAA, 0x0CB3 }, // Lo  [10] KANNADA LETTER PA..KANNADA LETTER LLA
    Interval{ 0x0CB5, 0x0CB9 }, // Lo   [5] KANNADA LETTER VA..KANNADA LETTER HA
    Interval{ 0x0CBC, 0x0CBC }, // Mn       KANNADA SIGN NUKTA
    Interval{ 0x0CBD, 0x0CBD }, // Lo       KANNADA SIGN AVAGRAHA
    Interval{ 0x0CBE, 0x0CBE }, // Mc       KANNADA VOWEL SIGN AA
    Interval{ 0x0CBF, 0x0CBF }, // Mn       KANNADA VOWEL SIGN I
    Interval{ 0x0CC0, 0x0CC4 }, // Mc   [5] KANNADA VOWEL SIGN II..KANNADA VOWEL SIGN VOCALIC RR
    Interval{ 0x0CC6, 0x0CC6 }, // Mn       KANNADA VOWEL SIGN E
    Interval{ 0x0CC7, 0x0CC8 }, // Mc   [2] KANNADA VOWEL SIGN EE..KANNADA VOWEL SIGN AI
    Interval{ 0x0CCA, 0x0CCB }, // Mc   [2] KANNADA VOWEL SIGN O..KANNADA VOWEL SIGN OO
    Interval{ 0x0CCC, 0x0CCD }, // Mn   [2] KANNADA VOWEL SIGN AU..KANNADA SIGN VIRAMA
    Interval{ 0x0CD5, 0x0CD6 }, // Mc   [2] KANNADA LENGTH MARK..KANNADA AI LENGTH MARK
    Interval{ 0x0CDC, 0x0CDE }, // Lo   [3] KANNADA ARCHAIC SHRII..KANNADA LETTER FA
    Interval{ 0x0CE0, 0x0CE1 }, // Lo   [2] KANNADA LETTER VOCALIC RR..KANNADA LETTER VOCALIC LL
    Interval{ 0x0CE2, 0x0CE3 }, // Mn   [2] KANNADA VOWEL SIGN VOCALIC L..KANNADA VOWEL SIGN VOCALIC LL
    Interval{ 0x0CE6, 0x0CEF }, // Nd  [10] KANNADA DIGIT ZERO..KANNADA DIGIT NINE
    Interval{ 0x0CF1, 0x0CF2 }, // Lo   [2] KANNADA SIGN JIHVAMULIYA..KANNADA SIGN UPADHMANIYA
    Interval{ 0x0CF3, 0x0CF3 }, // Mc       KANNADA SIGN COMBINING ANUSVARA ABOVE RIGHT
    Interval{ 0x0D00, 0x0D01 }, // Mn   [2] MALAYALAM SIGN COMBINING ANUSVARA ABOVE..MALAYALAM SIGN CANDRABINDU
    Interval{ 0x0D02, 0x0D03 }, // Mc   [2] MALAYALAM SIGN ANUSVARA..MALAYALAM SIGN VISARGA
    Interval{ 0x0D04, 0x0D0C }, // Lo   [9] MALAYALAM LETTER VEDIC ANUSVARA..MALAYALAM LETTER VOCALIC L
    Interval{ 0x0D0E, 0x0D10 }, // Lo   [3] MALAYALAM LETTER E..MALAYALAM LETTER AI
    Interval{ 0x0D12, 0x0D3A }, // Lo  [41] MALAYALAM LETTER O..MALAYALAM LETTER TTTA
    Interval{ 0x0D3B, 0x0D3C }, // Mn   [2] MALAYALAM SIGN VERTICAL BAR VIRAMA..MALAYALAM SIGN CIRCULAR VIRAMA
    Interval{ 0x0D3D, 0x0D3D }, // Lo       MALAYALAM SIGN AVAGRAHA
    Interval{ 0x0D3E, 0x0D40 }, // Mc   [3] MALAYALAM VOWEL SIGN AA..MALAYALAM VOWEL SIGN II
    Interval{ 0x0D41, 0x0D44 }, // Mn   [4] MALAYALAM VOWEL SIGN U..MALAYALAM VOWEL SIGN VOCALIC RR
    Interval{ 0x0D46, 0x0D48 }, // Mc   [3] MALAYALAM VOWEL SIGN E..MALAYALAM VOWEL SIGN AI
    Interval{ 0x0D4A, 0x0D4C }, // Mc   [3] MALAYALAM VOWEL SIGN O..MALAYALAM VOWEL SIGN AU
    Interval{ 0x0D4D, 0x0D4D }, // Mn       MALAYALAM SIGN VIRAMA
    Interval{ 0x0D4E, 0x0D4E }, // Lo       MALAYALAM LETTER DOT REPH
    Interval{ 0x0D54, 0x0D56 }, // Lo   [3] MALAYALAM LETTER CHILLU M..MALAYALAM LETTER CHILLU LLL
    Interval{ 0x0D57, 0x0D57 }, // Mc       MALAYALAM AU LENGTH MARK
    Interval{ 0x0D5F, 0x0D61 }, // Lo   [3] MALAYALAM LETTER ARCHAIC II..MALAYALAM LETTER VOCALIC LL
    Interval{ 0x0D62, 0x0D63 }, // Mn   [2] MALAYALAM VOWEL SIGN VOCALIC L..MALAYALAM VOWEL SIGN VOCALIC LL
    Interval{ 0x0D66, 0x0D6F }, // Nd  [10] MALAYALAM DIGIT ZERO..MALAYALAM DIGIT NINE
    Interval{ 0x0D7A, 0x0D7F }, // Lo   [6] MALAYALAM LETTER CHILLU NN..MALAYALAM LETTER CHILLU K
    Interval{ 0x0D81, 0x0D81 }, // Mn       SINHALA SIGN CANDRABINDU
    Interval{ 0x0D82, 0x0D83 }, // Mc   [2] SINHALA SIGN ANUSVARAYA..SINHALA SIGN VISARGAYA
    Interval{ 0x0D85, 0x0D96 }, // Lo  [18] SINHALA LETTER AYANNA..SINHALA LETTER AUYANNA
    Interval{ 0x0D9A, 0x0DB1 }, // Lo  [24] SINHALA LETTER ALPAPRAANA KAYANNA..SINHALA LETTER DANTAJA NAYANNA
    Interval{ 0x0DB3, 0x0DBB }, // Lo   [9] SINHALA LETTER SANYAKA DAYANNA..SINHALA LETTER RAYANNA
    Interval{ 0x0DBD, 0x0DBD }, // Lo       SINHALA LETTER DANTAJA LAYANNA
    Interval{ 0x0DC0, 0x0DC6 }, // Lo   [7] SINHALA LETTER VAYANNA..SINHALA LETTER FAYANNA
    Interval{ 0x0DCA, 0x0DCA }, // Mn       SINHALA SIGN AL-LAKUNA
    Interval{ 0x0DCF, 0x0DD1 }, // Mc   [3] SINHALA VOWEL SIGN AELA-PILLA..SINHALA VOWEL SIGN DIGA AEDA-PILLA
    Interval{ 0x0DD2, 0x0DD4 }, // Mn   [3] SINHALA VOWEL SIGN KETTI IS-PILLA..SINHALA VOWEL SIGN KETTI PAA-PILLA
    Interval{ 0x0DD6, 0x0DD6 }, // Mn       SINHALA VOWEL SIGN DIGA PAA-PILLA
    Interval{ 0x0DD8, 0x0DDF }, // Mc   [8] SINHALA VOWEL SIGN GAETTA-PILLA..SINHALA VOWEL SIGN GAYANUKITTA
    Interval{ 0x0DE6, 0x0DEF }, // Nd  [10] SINHALA LITH DIGIT ZERO..SINHALA LITH DIGIT NINE
    Interval{ 0x0DF2, 0x0DF3 }, // Mc   [2] SINHALA VOWEL SIGN DIGA GAETTA-PILLA..SINHALA VOWEL SIGN DIGA GAYANUKITTA
    Interval{ 0x0E01, 0x0E30 }, // Lo  [48] THAI CHARACTER KO KAI..THAI CHARACTER SARA A
    Interval{ 0x0E31, 0x0E31 }, // Mn       THAI CHARACTER MAI HAN-AKAT
    Interval{ 0x0E32, 0x0E33 }, // Lo   [2] THAI CHARACTER SARA AA..THAI CHARACTER SARA AM
    Interval{ 0x0E34, 0x0E3A }, // Mn   [7] THAI CHARACTER SARA I..THAI CHARACTER PHINTHU
    Interval{ 0x0E40, 0x0E45 }, // Lo   [6] THAI CHARACTER SARA E..THAI CHARACTER LAKKHANGYAO
    Interval{ 0x0E46, 0x0E46 }, // Lm       THAI CHARACTER MAIYAMOK
    Interval{ 0x0E47, 0x0E4E }, // Mn   [8] THAI CHARACTER MAITAIKHU..THAI CHARACTER YAMAKKAN
    Interval{ 0x0E50, 0x0E59 }, // Nd  [10] THAI DIGIT ZERO..THAI DIGIT NINE
    Interval{ 0x0E81, 0x0E82 }, // Lo   [2] LAO LETTER KO..LAO LETTER KHO SUNG
    Interval{ 0x0E84, 0x0E84 }, // Lo       LAO LETTER KHO TAM
    Interval{ 0x0E86, 0x0E8A }, // Lo   [5] LAO LETTER PALI GHA..LAO LETTER SO TAM
    Interval{ 0x0E8C, 0x0EA3 }, // Lo  [24] LAO LETTER PALI JHA..LAO LETTER LO LING
    Interval{ 0x0EA5, 0x0EA5 }, // Lo       LAO LETTER LO LOOT
    Interval{ 0x0EA7, 0x0EB0 }, // Lo  [10] LAO LETTER WO..LAO VOWEL SIGN A
    Interval{ 0x0EB1, 0x0EB1 }, // Mn       LAO VOWEL SIGN MAI KAN
    Interval{ 0x0EB2, 0x0EB3 }, // Lo   [2] LAO VOWEL SIGN AA..LAO VOWEL SIGN AM
    Interval{ 0x0EB4, 0x0EBC }, // Mn   [9] LAO VOWEL SIGN I..LAO SEMIVOWEL SIGN LO
    Interval{ 0x0EBD, 0x0EBD }, // Lo       LAO SEMIVOWEL SIGN NYO
    Interval{ 0x0EC0, 0x0EC4 }, // Lo   [5] LAO VOWEL SIGN E..LAO VOWEL SIGN AI
    Interval{ 0x0EC6, 0x0EC6 }, // Lm       LAO KO LA
    Interval{ 0x0EC8, 0x0ECE }, // Mn   [7] LAO TONE MAI EK..LAO YAMAKKAN
    Interval{ 0x0ED0, 0x0ED9 }, // Nd  [10] LAO DIGIT ZERO..LAO DIGIT NINE
    Interval{ 0x0EDC, 0x0EDF }, // Lo   [4] LAO HO NO..LAO LETTER KHMU NYO
    Interval{ 0x0F00, 0x0F00 }, // Lo       TIBETAN SYLLABLE OM
    Interval{ 0x0F18, 0x0F19 }, // Mn   [2] TIBETAN ASTROLOGICAL SIGN -KHYUD PA..TIBETAN ASTROLOGICAL SIGN SDONG TSHUGS
    Interval{ 0x0F20, 0x0F29 }, // Nd  [10] TIBETAN DIGIT ZERO..TIBETAN DIGIT NINE
    Interval{ 0x0F35, 0x0F35 }, // Mn       TIBETAN MARK NGAS BZUNG NYI ZLA
    Interval{ 0x0F37, 0x0F37 }, // Mn       TIBETAN MARK NGAS BZUNG SGOR RTAGS
    Interval{ 0x0F39, 0x0F39 }, // Mn       TIBETAN MARK TSA -PHRU
    Interval{ 0x0F3E, 0x0F3F }, // Mc   [2] TIBETAN SIGN YAR TSHES..TIBETAN SIGN MAR TSHES
    Interval{ 0x0F40, 0x0F47 }, // Lo   [8] TIBETAN LETTER KA..TIBETAN LETTER JA
    Interval{ 0x0F49, 0x0F6C }, // Lo  [36] TIBETAN LETTER NYA..TIBETAN LETTER RRA
    Interval{ 0x0F71, 0x0F7E }, // Mn  [14] TIBETAN VOWEL SIGN AA..TIBETAN SIGN RJES SU NGA RO
    Interval{ 0x0F7F, 0x0F7F }, // Mc       TIBETAN SIGN RNAM BCAD
    Interval{ 0x0F80, 0x0F84 }, // Mn   [5] TIBETAN VOWEL SIGN REVERSED I..TIBETAN MARK HALANTA
    Interval{ 0x0F86, 0x0F87 }, // Mn   [2] TIBETAN SIGN LCI RTAGS..TIBETAN SIGN YANG RTAGS
    Interval{ 0x0F88, 0x0F8C }, // Lo   [5] TIBETAN SIGN LCE TSA CAN..TIBETAN SIGN INVERTED MCHU CAN
    Interval{ 0x0F8D, 0x0F97 }, // Mn  [11] TIBETAN SUBJOINED SIGN LCE TSA CAN..TIBETAN SUBJOINED LETTER JA
    Interval{ 0x0F99, 0x0FBC }, // Mn  [36] TIBETAN SUBJOINED LETTER NYA..TIBETAN SUBJOINED LETTER FIXED-FORM RA
    Interval{ 0x0FC6, 0x0FC6 }, // Mn       TIBETAN SYMBOL PADMA GDAN
    Interval{ 0x1000, 0x102A }, // Lo  [43] MYANMAR LETTER KA..MYANMAR LETTER AU
    Interval{ 0x102B, 0x102C }, // Mc   [2] MYANMAR VOWEL SIGN TALL AA..MYANMAR VOWEL SIGN AA
    Interval{ 0x102D, 0x1030 }, // Mn   [4] MYANMAR VOWEL SIGN I..MYANMAR VOWEL SIGN UU
    Interval{ 0x1031, 0x1031 }, // Mc       MYANMAR VOWEL SIGN E
    Interval{ 0x1032, 0x1037 }, // Mn   [6] MYANMAR VOWEL SIGN AI..MYANMAR SIGN DOT BELOW
    Interval{ 0x1038, 0x1038 }, // Mc       MYANMAR SIGN VISARGA
    Interval{ 0x1039, 0x103A }, // Mn   [2] MYANMAR SIGN VIRAMA..MYANMAR SIGN ASAT
    Interval{ 0x103B, 0x103C }, // Mc   [2] MYANMAR CONSONANT SIGN MEDIAL YA..MYANMAR CONSONANT SIGN MEDIAL RA
    Interval{ 0x103D, 0x103E }, // Mn   [2] MYANMAR CONSONANT SIGN MEDIAL WA..MYANMAR CONSONANT SIGN MEDIAL HA
    Interval{ 0x103F, 0x103F }, // Lo       MYANMAR LETTER GREAT SA
    Interval{ 0x1040, 0x1049 }, // Nd  [10] MYANMAR DIGIT ZERO..MYANMAR DIGIT NINE
    Interval{ 0x1050, 0x1055 }, // Lo   [6] MYANMAR LETTER SHA..MYANMAR LETTER VOCALIC LL
    Interval{ 0x1056, 0x1057 }, // Mc   [2] MYANMAR VOWEL SIGN VOCALIC R..MYANMAR VOWEL SIGN VOCALIC RR
    Interval{ 0x1058, 0x1059 }, // Mn   [2] MYANMAR VOWEL SIGN VOCALIC L..MYANMAR VOWEL SIGN VOCALIC LL
    Interval{ 0x105A, 0x105D }, // Lo   [4] MYANMAR LETTER MON NGA..MYANMAR LETTER MON BBE
    Interval{ 0x105E, 0x1060 }, // Mn   [3] MYANMAR CONSONANT SIGN MON MEDIAL NA..MYANMAR CONSONANT SIGN MON MEDIAL LA
    Interval{ 0x1061, 0x1061 }, // Lo       MYANMAR LETTER SGAW KAREN SHA
    Interval{ 0x1062, 0x1064 }, // Mc   [3] MYANMAR VOWEL SIGN SGAW KAREN EU..MYANMAR TONE MARK SGAW KAREN KE PHO
    Interval{ 0x1065, 0x1066 }, // Lo   [2] MYANMAR LETTER WESTERN PWO KAREN THA..MYANMAR LETTER WESTERN PWO KAREN PWA
    Interval{ 0x1067, 0x106D }, // Mc   [7] MYANMAR VOWEL SIGN WESTERN PWO KAREN EU..MYANMAR SIGN WESTERN PWO KAREN TONE-5
    Interval{ 0x106E, 0x1070 }, // Lo   [3] MYANMAR LETTER EASTERN PWO KAREN NNA..MYANMAR LETTER EASTERN PWO KAREN GHWA
    Interval{ 0x1071, 0x1074 }, // Mn   [4] MYANMAR VOWEL SIGN GEBA KAREN I..MYANMAR VOWEL SIGN KAYAH EE
    Interval{ 0x1075, 0x1081 }, // Lo  [13] MYANMAR LETTER SHAN KA..MYANMAR LETTER SHAN HA
    Interval{ 0x1082, 0x1082 }, // Mn       MYANMAR CONSONANT SIGN SHAN MEDIAL WA
    Interval{ 0x1083, 0x1084 }, // Mc   [2] MYANMAR VOWEL SIGN SHAN AA..MYANMAR VOWEL SIGN SHAN E
    Interval{ 0x1085, 0x1086 }, // Mn   [2] MYANMAR VOWEL SIGN SHAN E ABOVE..MYANMAR VOWEL SIGN SHAN FINAL Y
    Interval{ 0x1087, 0x108C }, // Mc   [6] MYANMAR SIGN SHAN TONE-2..MYANMAR SIGN SHAN COUNCIL TONE-3
    Interval{ 0x108D, 0x108D }, // Mn       MYANMAR SIGN SHAN COUNCIL EMPHATIC TONE
    Interval{ 0x108E, 0x108E }, // Lo       MYANMAR LETTER RUMAI PALAUNG FA
    Interval{ 0x108F, 0x108F }, // Mc       MYANMAR SIGN RUMAI PALAUNG TONE-5
    Interval{ 0x1090, 0x1099 }, // Nd  [10] MYANMAR SHAN DIGIT ZERO..MYANMAR SHAN DIGIT NINE
    Interval{ 0x109A, 0x109C }, // Mc   [3] MYANMAR SIGN KHAMTI TONE-1..MYANMAR VOWEL SIGN AITON A
    Interval{ 0x109D, 0x109D }, // Mn       MYANMAR VOWEL SIGN AITON AI
    Interval{ 0x10A0, 0x10C5 }, // L&  [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Interval{ 0x10C7, 0x10C7 }, // L&       GEORGIAN CAPITAL LETTER YN
    Interval{ 0x10CD, 0x10CD }, // L&       GEORGIAN CAPITAL LETTER AEN
    Interval{ 0x10D0, 0x10FA }, // L&  [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Interval{ 0x10FC, 0x10FC }, // Lm       MODIFIER LETTER GEORGIAN NAR
    Interval{ 0x10FD, 0x10FF }, // L&   [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Interval{ 0x1100, 0x1248 }, // Lo [329] HANGUL CHOSEONG KIYEOK..ETHIOPIC SYLLABLE QWA
    Interval{ 0x124A, 0x124D }, // Lo   [4] ETHIOPIC SYLLABLE QWI..ETHIOPIC SYLLABLE QWE
    Interval{ 0x1250, 0x1256 }, // Lo   [7] ETHIOPIC SYLLABLE QHA..ETHIOPIC SYLLABLE QHO
    Interval{ 0x1258, 0x1258 }, // Lo       ETHIOPIC SYLLABLE QHWA
    Interval{ 0x125A, 0x125D }, // Lo   [4] ETHIOPIC SYLLABLE QHWI..ETHIOPIC SYLLABLE QHWE
    Interval{ 0x1260, 0x1288 }, // Lo  [41] ETHIOPIC SYLLABLE BA..ETHIOPIC SYLLABLE XWA
    Interval{ 0x128A, 0x128D }, // Lo   [4] ETHIOPIC SYLLABLE XWI..ETHIOPIC SYLLABLE XWE
    Interval{ 0x1290, 0x12B0 }, // Lo  [33] ETHIOPIC SYLLABLE NA..ETHIOPIC SYLLABLE KWA
    Interval{ 0x12B2, 0x12B5 }, // Lo   [4] ETHIOPIC SYLLABLE KWI..ETHIOPIC SYLLABLE KWE
    Interval{ 0x12B8, 0x12BE }, // Lo   [7] ETHIOPIC SYLLABLE KXA..ETHIOPIC SYLLABLE KXO
    Interval{ 0x12C0, 0x12C0 }, // Lo       ETHIOPIC SYLLABLE KXWA
    Interval{ 0x12C2, 0x12C5 }, // Lo   [4] ETHIOPIC SYLLABLE KXWI..ETHIOPIC SYLLABLE KXWE
    Interval{ 0x12C8, 0x12D6 }, // Lo  [15] ETHIOPIC SYLLABLE WA..ETHIOPIC SYLLABLE PHARYNGEAL O
    Interval{ 0x12D8, 0x1310 }, // Lo  [57] ETHIOPIC SYLLABLE ZA..ETHIOPIC SYLLABLE GWA
    Interval{ 0x1312, 0x1315 }, // Lo   [4] ETHIOPIC SYLLABLE GWI..ETHIOPIC SYLLABLE GWE
    Interval{ 0x1318, 0x135A }, // Lo  [67] ETHIOPIC SYLLABLE GGA..ETHIOPIC SYLLABLE FYA
    Interval{ 0x135D, 0x135F }, // Mn   [3] ETHIOPIC COMBINING GEMINATION AND VOWEL LENGTH MARK..ETHIOPIC COMBINING GEMINATION MARK
    Interval{ 0x1369, 0x1371 }, // No   [9] ETHIOPIC DIGIT ONE..ETHIOPIC DIGIT NINE
    Interval{ 0x1380, 0x138F }, // Lo  [16] ETHIOPIC SYLLABLE SEBATBEIT MWA..ETHIOPIC SYLLABLE PWE
    Interval{ 0x13A0, 0x13F5 }, // L&  [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Interval{ 0x13F8, 0x13FD }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1401, 0x166C }, // Lo [620] CANADIAN SYLLABICS E..CANADIAN SYLLABICS CARRIER TTSA
    Interval{ 0x166F, 0x167F }, // Lo  [17] CANADIAN SYLLABICS QAI..CANADIAN SYLLABICS BLACKFOOT W
    Interval{ 0x1681, 0x169A }, // Lo  [26] OGHAM LETTER BEITH..OGHAM LETTER PEITH
    Interval{ 0x16A0, 0x16EA }, // Lo  [75] RUNIC LETTER FEHU FEOH FE F..RUNIC LETTER X
    Interval{ 0x16EE, 0x16F0 }, // Nl   [3] RUNIC ARLAUG SYMBOL..RUNIC BELGTHOR SYMBOL
    Interval{ 0x16F1, 0x16F8 }, // Lo   [8] RUNIC LETTER K..RUNIC LETTER FRANKS CASKET AESC
    Interval{ 0x1700, 0x1711 }, // Lo  [18] TAGALOG LETTER A..TAGALOG LETTER HA
    Interval{ 0x1712, 0x1714 }, // Mn   [3] TAGALOG VOWEL SIGN I..TAGALOG SIGN VIRAMA
    Interval{ 0x1715, 0x1715 }, // Mc       TAGALOG SIGN PAMUDPOD
    Interval{ 0x171F, 0x1731 }, // Lo  [19] TAGALOG LETTER ARCHAIC RA..HANUNOO LETTER HA
    Interval{ 0x1732, 0x1733 }, // Mn   [2] HANUNOO VOWEL SIGN I..HANUNOO VOWEL SIGN U
    Interval{ 0x1734, 0x1734 }, // Mc       HANUNOO SIGN PAMUDPOD
    Interval{ 0x1740, 0x1751 }, // Lo  [18] BUHID LETTER A..BUHID LETTER HA
    Interval{ 0x1752, 0x1753 }, // Mn   [2] BUHID VOWEL SIGN I..BUHID VOWEL SIGN U
    Interval{ 0x1760, 0x176C }, // Lo  [13] TAGBANWA LETTER A..TAGBANWA LETTER YA
    Interval{ 0x176E, 0x1770 }, // Lo   [3] TAGBANWA LETTER LA..TAGBANWA LETTER SA
    Interval{ 0x1772, 0x1773 }, // Mn   [2] TAGBANWA VOWEL SIGN I..TAGBANWA VOWEL SIGN U
    Interval{ 0x1780, 0x17B3 }, // Lo  [52] KHMER LETTER KA..KHMER INDEPENDENT VOWEL QAU
    Interval{ 0x17B4, 0x17B5 }, // Mn   [2] KHMER VOWEL INHERENT AQ..KHMER VOWEL INHERENT AA
    Interval{ 0x17B6, 0x17B6 }, // Mc       KHMER VOWEL SIGN AA
    Interval{ 0x17B7, 0x17BD }, // Mn   [7] KHMER VOWEL SIGN I..KHMER VOWEL SIGN UA
    Interval{ 0x17BE, 0x17C5 }, // Mc   [8] KHMER VOWEL SIGN OE..KHMER VOWEL SIGN AU
    Interval{ 0x17C6, 0x17C6 }, // Mn       KHMER SIGN NIKAHIT
    Interval{ 0x17C7, 0x17C8 }, // Mc   [2] KHMER SIGN REAHMUK..KHMER SIGN YUUKALEAPINTU
    Interval{ 0x17C9, 0x17D3 }, // Mn  [11] KHMER SIGN MUUSIKATOAN..KHMER SIGN BATHAMASAT
    Interval{ 0x17D7, 0x17D7 }, // Lm       KHMER SIGN LEK TOO
    Interval{ 0x17DC, 0x17DC }, // Lo       KHMER SIGN AVAKRAHASANYA
    Interval{ 0x17DD, 0x17DD }, // Mn       KHMER SIGN ATTHACAN
    Interval{ 0x17E0, 0x17E9 }, // Nd  [10] KHMER DIGIT ZERO..KHMER DIGIT NINE
    Interval{ 0x180B, 0x180D }, // Mn   [3] MONGOLIAN FREE VARIATION SELECTOR ONE..MONGOLIAN FREE VARIATION SELECTOR THREE
    Interval{ 0x180F, 0x180F }, // Mn       MONGOLIAN FREE VARIATION SELECTOR FOUR
    Interval{ 0x1810, 0x1819 }, // Nd  [10] MONGOLIAN DIGIT ZERO..MONGOLIAN DIGIT NINE
    Interval{ 0x1820, 0x1842 }, // Lo  [35] MONGOLIAN LETTER A..MONGOLIAN LETTER CHI
    Interval{ 0x1843, 0x1843 }, // Lm       MONGOLIAN LETTER TODO LONG VOWEL SIGN
    Interval{ 0x1844, 0x1878 }, // Lo  [53] MONGOLIAN LETTER TODO E..MONGOLIAN LETTER CHA WITH TWO DOTS
    Interval{ 0x1880, 0x1884 }, // Lo   [5] MONGOLIAN LETTER ALI GALI ANUSVARA ONE..MONGOLIAN LETTER ALI GALI INVERTED UBADAMA
    Interval{ 0x1885, 0x1886 }, // Mn   [2] MONGOLIAN LETTER ALI GALI BALUDA..MONGOLIAN LETTER ALI GALI THREE BALUDA
    Interval{ 0x1887, 0x18A8 }, // Lo  [34] MONGOLIAN LETTER ALI GALI A..MONGOLIAN LETTER MANCHU ALI GALI BHA
    Interval{ 0x18A9, 0x18A9 }, // Mn       MONGOLIAN LETTER ALI GALI DAGALGA
    Interval{ 0x18AA, 0x18AA }, // Lo       MONGOLIAN LETTER MANCHU ALI GALI LHA
    Interval{ 0x18B0, 0x18F5 }, // Lo  [70] CANADIAN SYLLABICS OY..CANADIAN SYLLABICS CARRIER DENTAL S
    Interval{ 0x1900, 0x191E }, // Lo  [31] LIMBU VOWEL-CARRIER LETTER..LIMBU LETTER TRA
    Interval{ 0x1920, 0x1922 }, // Mn   [3] LIMBU VOWEL SIGN A..LIMBU VOWEL SIGN U
    Interval{ 0x1923, 0x1926 }, // Mc   [4] LIMBU VOWEL SIGN EE..LIMBU VOWEL SIGN AU
    Interval{ 0x1927, 0x1928 }, // Mn   [2] LIMBU VOWEL SIGN E..LIMBU VOWEL SIGN O
    Interval{ 0x1929, 0x192B }, // Mc   [3] LIMBU SUBJOINED LETTER YA..LIMBU SUBJOINED LETTER WA
    Interval{ 0x1930, 0x1931 }, // Mc   [2] LIMBU SMALL LETTER KA..LIMBU SMALL LETTER NGA
    Interval{ 0x1932, 0x1932 }, // Mn       LIMBU SMALL LETTER ANUSVARA
    Interval{ 0x1933, 0x1938 }, // Mc   [6] LIMBU SMALL LETTER TA..LIMBU SMALL LETTER LA
    Interval{ 0x1939, 0x193B }, // Mn   [3] LIMBU SIGN MUKPHRENG..LIMBU SIGN SA-I
    Interval{ 0x1946, 0x194F }, // Nd  [10] LIMBU DIGIT ZERO..LIMBU DIGIT NINE
    Interval{ 0x1950, 0x196D }, // Lo  [30] TAI LE LETTER KA..TAI LE LETTER AI
    Interval{ 0x1970, 0x1974 }, // Lo   [5] TAI LE LETTER TONE-2..TAI LE LETTER TONE-6
    Interval{ 0x1980, 0x19AB }, // Lo  [44] NEW TAI LUE LETTER HIGH QA..NEW TAI LUE LETTER LOW SUA
    Interval{ 0x19B0, 0x19C9 }, // Lo  [26] NEW TAI LUE VOWEL SIGN VOWEL SHORTENER..NEW TAI LUE TONE MARK-2
    Interval{ 0x19D0, 0x19D9 }, // Nd  [10] NEW TAI LUE DIGIT ZERO..NEW TAI LUE DIGIT NINE
    Interval{ 0x19DA, 0x19DA }, // No       NEW TAI LUE THAM DIGIT ONE
    Interval{ 0x1A00, 0x1A16 }, // Lo  [23] BUGINESE LETTER KA..BUGINESE LETTER HA
    Interval{ 0x1A17, 0x1A18 }, // Mn   [2] BUGINESE VOWEL SIGN I..BUGINESE VOWEL SIGN U
    Interval{ 0x1A19, 0x1A1A }, // Mc   [2] BUGINESE VOWEL SIGN E..BUGINESE VOWEL SIGN O
    Interval{ 0x1A1B, 0x1A1B }, // Mn       BUGINESE VOWEL SIGN AE
    Interval{ 0x1A20, 0x1A54 }, // Lo  [53] TAI THAM LETTER HIGH KA..TAI THAM LETTER GREAT SA
    Interval{ 0x1A55, 0x1A55 }, // Mc       TAI THAM CONSONANT SIGN MEDIAL RA
    Interval{ 0x1A56, 0x1A56 }, // Mn       TAI THAM CONSONANT SIGN MEDIAL LA
    Interval{ 0x1A57, 0x1A57 }, // Mc       TAI THAM CONSONANT SIGN LA TANG LAI
    Interval{ 0x1A58, 0x1A5E }, // Mn   [7] TAI THAM SIGN MAI KANG LAI..TAI THAM CONSONANT SIGN SA
    Interval{ 0x1A60, 0x1A60 }, // Mn       TAI THAM SIGN SAKOT
    Interval{ 0x1A61, 0x1A61 }, // Mc       TAI THAM VOWEL SIGN A
    Interval{ 0x1A62, 0x1A62 }, // Mn       TAI THAM VOWEL SIGN MAI SAT
    Interval{ 0x1A63, 0x1A64 }, // Mc   [2] TAI THAM VOWEL SIGN AA..TAI THAM VOWEL SIGN TALL AA
    Interval{ 0x1A65, 0x1A6C }, // Mn   [8] TAI THAM VOWEL SIGN I..TAI THAM VOWEL SIGN OA BELOW
    Interval{ 0x1A6D, 0x1A72 }, // Mc   [6] TAI THAM VOWEL SIGN OY..TAI THAM VOWEL SIGN THAM AI
    Interval{ 0x1A73, 0x1A7C }, // Mn  [10] TAI THAM VOWEL SIGN OA ABOVE..TAI THAM SIGN KHUEN-LUE KARAN
    Interval{ 0x1A7F, 0x1A7F }, // Mn       TAI THAM COMBINING CRYPTOGRAMMIC DOT
    Interval{ 0x1A80, 0x1A89 }, // Nd  [10] TAI THAM HORA DIGIT ZERO..TAI THAM HORA DIGIT NINE
    Interval{ 0x1A90, 0x1A99 }, // Nd  [10] TAI THAM THAM DIGIT ZERO..TAI THAM THAM DIGIT NINE
    Interval{ 0x1AA7, 0x1AA7 }, // Lm       TAI THAM SIGN MAI YAMOK
    Interval{ 0x1AB0, 0x1ABD }, // Mn  [14] COMBINING DOUBLED CIRCUMFLEX ACCENT..COMBINING PARENTHESES BELOW
    Interval{ 0x1ABF, 0x1ADD }, // Mn  [31] COMBINING LATIN SMALL LETTER W BELOW..COMBINING DOT-AND-RING BELOW
    Interval{ 0x1AE0, 0x1AEB }, // Mn  [12] COMBINING LEFT TACK ABOVE..COMBINING DOUBLE RIGHTWARDS ARROW ABOVE
    Interval{ 0x1B00, 0x1B03 }, // Mn   [4] BALINESE SIGN ULU RICEM..BALINESE SIGN SURANG
    Interval{ 0x1B04, 0x1B04 }, // Mc       BALINESE SIGN BISAH
    Interval{ 0x1B05, 0x1B33 }, // Lo  [47] BALINESE LETTER AKARA..BALINESE LETTER HA
    Interval{ 0x1B34, 0x1B34 }, // Mn       BALINESE SIGN REREKAN
    Interval{ 0x1B35, 0x1B35 }, // Mc       BALINESE VOWEL SIGN TEDUNG
    Interval{ 0x1B36, 0x1B3A }, // Mn   [5] BALINESE VOWEL SIGN ULU..BALINESE VOWEL SIGN RA REPA
    Interval{ 0x1B3B, 0x1B3B }, // Mc       BALINESE VOWEL SIGN RA REPA TEDUNG
    Interval{ 0x1B3C, 0x1B3C }, // Mn       BALINESE VOWEL SIGN LA LENGA
    Interval{ 0x1B3D, 0x1B41 }, // Mc   [5] BALINESE VOWEL SIGN LA LENGA TEDUNG..BALINESE VOWEL SIGN TALING REPA TEDUNG
    Interval{ 0x1B42, 0x1B42 }, // Mn       BALINESE VOWEL SIGN PEPET
    Interval{ 0x1B43, 0x1B44 }, // Mc   [2] BALINESE VOWEL SIGN PEPET TEDUNG..BALINESE ADEG ADEG
    Interval{ 0x1B45, 0x1B4C }, // Lo   [8] BALINESE LETTER KAF SASAK..BALINESE LETTER ARCHAIC JNYA
    Interval{ 0x1B50, 0x1B59 }, // Nd  [10] BALINESE DIGIT ZERO..BALINESE DIGIT NINE
    Interval{ 0x1B6B, 0x1B73 }, // Mn   [9] BALINESE MUSICAL SYMBOL COMBINING TEGEH..BALINESE MUSICAL SYMBOL COMBINING GONG
    Interval{ 0x1B80, 0x1B81 }, // Mn   [2] SUNDANESE SIGN PANYECEK..SUNDANESE SIGN PANGLAYAR
    Interval{ 0x1B82, 0x1B82 }, // Mc       SUNDANESE SIGN PANGWISAD
    Interval{ 0x1B83, 0x1BA0 }, // Lo  [30] SUNDANESE LETTER A..SUNDANESE LETTER HA
    Interval{ 0x1BA1, 0x1BA1 }, // Mc       SUNDANESE CONSONANT SIGN PAMINGKAL
    Interval{ 0x1BA2, 0x1BA5 }, // Mn   [4] SUNDANESE CONSONANT SIGN PANYAKRA..SUNDANESE VOWEL SIGN PANYUKU
    Interval{ 0x1BA6, 0x1BA7 }, // Mc   [2] SUNDANESE VOWEL SIGN PANAELAENG..SUNDANESE VOWEL SIGN PANOLONG
    Interval{ 0x1BA8, 0x1BA9 }, // Mn   [2] SUNDANESE VOWEL SIGN PAMEPET..SUNDANESE VOWEL SIGN PANEULEUNG
    Interval{ 0x1BAA, 0x1BAA }, // Mc       SUNDANESE SIGN PAMAAEH
    Interval{ 0x1BAB, 0x1BAD }, // Mn   [3] SUNDANESE SIGN VIRAMA..SUNDANESE CONSONANT SIGN PASANGAN WA
    Interval{ 0x1BAE, 0x1BAF }, // Lo   [2] SUNDANESE LETTER KHA..SUNDANESE LETTER SYA
    Interval{ 0x1BB0, 0x1BB9 }, // Nd  [10] SUNDANESE DIGIT ZERO..SUNDANESE DIGIT NINE
    Interval{ 0x1BBA, 0x1BE5 }, // Lo  [44] SUNDANESE AVAGRAHA..BATAK LETTER U
    Interval{ 0x1BE6, 0x1BE6 }, // Mn       BATAK SIGN TOMPI
    Interval{ 0x1BE7, 0x1BE7 }, // Mc       BATAK VOWEL SIGN E
    Interval{ 0x1BE8, 0x1BE9 }, // Mn   [2] BATAK VOWEL SIGN PAKPAK E..BATAK VOWEL SIGN EE
    Interval{ 0x1BEA, 0x1BEC }, // Mc   [3] BATAK VOWEL SIGN I..BATAK VOWEL SIGN O
    Interval{ 0x1BED, 0x1BED }, // Mn       BATAK VOWEL SIGN KARO O
    Interval{ 0x1BEE, 0x1BEE }, // Mc       BATAK VOWEL SIGN U
    Interval{ 0x1BEF, 0x1BF1 }, // Mn   [3] BATAK VOWEL SIGN U FOR SIMALUNGUN SA..BATAK CONSONANT SIGN H
    Interval{ 0x1BF2, 0x1BF3 }, // Mc   [2] BATAK PANGOLAT..BATAK PANONGONAN
    Interval{ 0x1C00, 0x1C23 }, // Lo  [36] LEPCHA LETTER KA..LEPCHA LETTER A
    Interval{ 0x1C24, 0x1C2B }, // Mc   [8] LEPCHA SUBJOINED LETTER YA..LEPCHA VOWEL SIGN UU
    Interval{ 0x1C2C, 0x1C33 }, // Mn   [8] LEPCHA VOWEL SIGN E..LEPCHA CONSONANT SIGN T
    Interval{ 0x1C34, 0x1C35 }, // Mc   [2] LEPCHA CONSONANT SIGN NYIN-DO..LEPCHA CONSONANT SIGN KANG
    Interval{ 0x1C36, 0x1C37 }, // Mn   [2] LEPCHA SIGN RAN..LEPCHA SIGN NUKTA
    Interval{ 0x1C40, 0x1C49 }, // Nd  [10] LEPCHA DIGIT ZERO..LEPCHA DIGIT NINE
    Interval{ 0x1C4D, 0x1C4F }, // Lo   [3] LEPCHA LETTER TTA..LEPCHA LETTER DDA
    Interval{ 0x1C50, 0x1C59 }, // Nd  [10] OL CHIKI DIGIT ZERO..OL CHIKI DIGIT NINE
    Interval{ 0x1C5A, 0x1C77 }, // Lo  [30] OL CHIKI LETTER LA..OL CHIKI LETTER OH
    Interval{ 0x1C78, 0x1C7D }, // Lm   [6] OL CHIKI MU TTUDDAG..OL CHIKI AHAD
    Interval{ 0x1C80, 0x1C8A }, // L&  [11] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER TJE
    Interval{ 0x1C90, 0x1CBA }, // L&  [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Interval{ 0x1CBD, 0x1CBF }, // L&   [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Interval{ 0x1CD0, 0x1CD2 }, // Mn   [3] VEDIC TONE KARSHANA..VEDIC TONE PRENKHA
    Interval{ 0x1CD4, 0x1CE0 }, // Mn  [13] VEDIC SIGN YAJURVEDIC MIDLINE SVARITA..VEDIC TONE RIGVEDIC KASHMIRI INDEPENDENT SVARITA
    Interval{ 0x1CE1, 0x1CE1 }, // Mc       VEDIC TONE ATHARVAVEDIC INDEPENDENT SVARITA
    Interval{ 0x1CE2, 0x1CE8 }, // Mn   [7] VEDIC SIGN VISARGA SVARITA..VEDIC SIGN VISARGA ANUDATTA WITH TAIL
    Interval{ 0x1CE9, 0x1CEC }, // Lo   [4] VEDIC SIGN ANUSVARA ANTARGOMUKHA..VEDIC SIGN ANUSVARA VAMAGOMUKHA WITH TAIL
    Interval{ 0x1CED, 0x1CED }, // Mn       VEDIC SIGN TIRYAK
    Interval{ 0x1CEE, 0x1CF3 }, // Lo   [6] VEDIC SIGN HEXIFORM LONG ANUSVARA..VEDIC SIGN ROTATED ARDHAVISARGA
    Interval{ 0x1CF4, 0x1CF4 }, // Mn       VEDIC TONE CANDRA ABOVE
    Interval{ 0x1CF5, 0x1CF6 }, // Lo   [2] VEDIC SIGN JIHVAMULIYA..VEDIC SIGN UPADHMANIYA
    Interval{ 0x1CF7, 0x1CF7 }, // Mc       VEDIC SIGN ATIKRAMA
    Interval{ 0x1CF8, 0x1CF9 }, // Mn   [2] VEDIC TONE RING ABOVE..VEDIC TONE DOUBLE RING ABOVE
    Interval{ 0x1CFA, 0x1CFA }, // Lo       VEDIC SIGN DOUBLE ANUSVARA ANTARGOMUKHA
    Interval{ 0x1D00, 0x1D2B }, // L&  [44] LATIN LETTER SMALL CAPITAL A..CYRILLIC LETTER SMALL CAPITAL EL
    Interval{ 0x1D2C, 0x1D6A }, // Lm  [63] MODIFIER LETTER CAPITAL A..GREEK SUBSCRIPT SMALL LETTER CHI
    Interval{ 0x1D6B, 0x1D77 }, // L&  [13] LATIN SMALL LETTER UE..LATIN SMALL LETTER TURNED G
    Interval{ 0x1D78, 0x1D78 }, // Lm       MODIFIER LETTER CYRILLIC EN
    Interval{ 0x1D79, 0x1D9A }, // L&  [34] LATIN SMALL LETTER INSULAR G..LATIN SMALL LETTER EZH WITH RETROFLEX HOOK
    Interval{ 0x1D9B, 0x1DBF }, // Lm  [37] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL THETA
    Interval{ 0x1DC0, 0x1DFF }, // Mn  [64] COMBINING DOTTED GRAVE ACCENT..COMBINING RIGHT ARROWHEAD AND DOWN ARROWHEAD BELOW
    Interval{ 0x1E00, 0x1F15 }, // L& [278] LATIN CAPITAL LETTER A WITH RING BELOW..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F18, 0x1F1D }, // L&   [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F20, 0x1F45 }, // L&  [38] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F48, 0x1F4D }, // L&   [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F50, 0x1F57 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F59, 0x1F59 }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA
    Interval{ 0x1F5B, 0x1F5B }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Interval{ 0x1F5D, 0x1F5D }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Interval{ 0x1F5F, 0x1F7D }, // L&  [31] GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI..GREEK SMALL LETTER OMEGA WITH OXIA
    Interval{ 0x1F80, 0x1FB4 }, // L&  [53] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB6, 0x1FBC }, // L&   [7] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Interval{ 0x1FBE, 0x1FBE }, // L&       GREEK PROSGEGRAMMENI
    Interval{ 0x1FC2, 0x1FC4 }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC6, 0x1FCC }, // L&   [7] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Interval{ 0x1FD0, 0x1FD3 }, // L&   [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Interval{ 0x1FD6, 0x1FDB }, // L&   [6] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK CAPITAL LETTER IOTA WITH OXIA
    Interval{ 0x1FE0, 0x1FEC }, // L&  [13] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Interval{ 0x1FF2, 0x1FF4 }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF6, 0x1FFC }, // L&   [7] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Interval{ 0x200C, 0x200D }, // Cf   [2] ZERO WIDTH NON-JOINER..ZERO WIDTH JOINER
    Interval{ 0x203F, 0x2040 }, // Pc   [2] UNDERTIE..CHARACTER TIE
    Interval{ 0x2054, 0x2054 }, // Pc       INVERTED UNDERTIE
    Interval{ 0x2071, 0x2071 }, // Lm       SUPERSCRIPT LATIN SMALL LETTER I
    Interval{ 0x207F, 0x207F }, // Lm       SUPERSCRIPT LATIN SMALL LETTER N
    Interval{ 0x2090, 0x209C }, // Lm  [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Interval{ 0x20D0, 0x20DC }, // Mn  [13] COMBINING LEFT HARPOON ABOVE..COMBINING FOUR DOTS ABOVE
    Interval{ 0x20E1, 0x20E1 }, // Mn       COMBINING LEFT RIGHT ARROW ABOVE
    Interval{ 0x20E5, 0x20F0 }, // Mn  [12] COMBINING REVERSE SOLIDUS OVERLAY..COMBINING ASTERISK ABOVE
    Interval{ 0x2102, 0x2102 }, // L&       DOUBLE-STRUCK CAPITAL C
    Interval{ 0x2107, 0x2107 }, // L&       EULER CONSTANT
    Interval{ 0x210A, 0x2113 }, // L&  [10] SCRIPT SMALL G..SCRIPT SMALL L
    Interval{ 0x2115, 0x2115 }, // L&       DOUBLE-STRUCK CAPITAL N
    Interval{ 0x2118, 0x2118 }, // Sm       SCRIPT CAPITAL P
    Interval{ 0x2119, 0x211D }, // L&   [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    Interval{ 0x2124, 0x2124 }, // L&       DOUBLE-STRUCK CAPITAL Z
    Interval{ 0x2126, 0x2126 }, // L&       OHM SIGN
    Interval{ 0x2128, 0x2128 }, // L&       BLACK-LETTER CAPITAL Z
    Interval{ 0x212A, 0x212D }, // L&   [4] KELVIN SIGN..BLACK-LETTER CAPITAL C
    Interval{ 0x212E, 0x212E }, // So       ESTIMATED SYMBOL
    Interval{ 0x212F, 0x2134 }, // L&   [6] SCRIPT SMALL E..SCRIPT SMALL O
    Interval{ 0x2135, 0x2138 }, // Lo   [4] ALEF SYMBOL..DALET SYMBOL
    Interval{ 0x2139, 0x2139 }, // L&       INFORMATION SOURCE
    Interval{ 0x213C, 0x213F }, // L&   [4] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK CAPITAL PI
    Interval{ 0x2145, 0x2149 }, // L&   [5] DOUBLE-STRUCK ITALIC CAPITAL D..DOUBLE-STRUCK ITALIC SMALL J
    Interval{ 0x214E, 0x214E }, // L&       TURNED SMALL F
    Interval{ 0x2160, 0x2182 }, // Nl  [35] ROMAN NUMERAL ONE..ROMAN NUMERAL TEN THOUSAND
    Interval{ 0x2183, 0x2184 }, // L&   [2] ROMAN NUMERAL REVERSED ONE HUNDRED..LATIN SMALL LETTER REVERSED C
    Interval{ 0x2185, 0x2188 }, // Nl   [4] ROMAN NUMERAL SIX LATE FORM..ROMAN NUMERAL ONE HUNDRED THOUSAND
    Interval{ 0x2C00, 0x2C7B }, // L& [124] GLAGOLITIC CAPITAL LETTER AZU..LATIN LETTER SMALL CAPITAL TURNED E
    Interval{ 0x2C7C, 0x2C7D }, // Lm   [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    Interval{ 0x2C7E, 0x2CE4 }, // L& [103] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC SYMBOL KAI
    Interval{ 0x2CEB, 0x2CEE }, // L&   [4] COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI..COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CEF, 0x2CF1 }, // Mn   [3] COPTIC COMBINING NI ABOVE..COPTIC COMBINING SPIRITUS LENIS
    Interval{ 0x2CF2, 0x2CF3 }, // L&   [2] COPTIC CAPITAL LETTER BOHAIRIC KHEI..COPTIC SMALL LETTER BOHAIRIC KHEI
    Interval{ 0x2D00, 0x2D25 }, // L&  [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Interval{ 0x2D27, 0x2D27 }, // L&       GEORGIAN SMALL LETTER YN
    Interval{ 0x2D2D, 0x2D2D }, // L&       GEORGIAN SMALL LETTER AEN
    Interval{ 0x2D30, 0x2D67 }, // Lo  [56] TIFINAGH LETTER YA..TIFINAGH LETTER YO
    Interval{ 0x2D6F, 0x2D6F }, // Lm       TIFINAGH MODIFIER LETTER LABIALIZATION MARK
    Interval{ 0x2D7F, 0x2D7F }, // Mn       TIFINAGH CONSONANT JOINER
    Interval{ 0x2D80, 0x2D96 }, // Lo  [23] ETHIOPIC SYLLABLE LOA..ETHIOPIC SYLLABLE GGWE
    Interval{ 0x2DA0, 0x2DA6 }, // Lo   [7] ETHIOPIC SYLLABLE SSA..ETHIOPIC SYLLABLE SSO
    Interval{ 0x2DA8, 0x2DAE }, // Lo   [7] ETHIOPIC SYLLABLE CCA..ETHIOPIC SYLLABLE CCO
    Interval{ 0x2DB0, 0x2DB6 }, // Lo   [7] ETHIOPIC SYLLABLE ZZA..ETHIOPIC SYLLABLE ZZO
    Interval{ 0x2DB8, 0x2DBE }, // Lo   [7] ETHIOPIC SYLLABLE CCHA..ETHIOPIC SYLLABLE CCHO
    Interval{ 0x2DC0, 0x2DC6 }, // Lo   [7] ETHIOPIC SYLLABLE QYA..ETHIOPIC SYLLABLE QYO
    Interval{ 0x2DC8, 0x2DCE }, // Lo   [7] ETHIOPIC SYLLABLE KYA..ETHIOPIC SYLLABLE KYO
    Interval{ 0x2DD0, 0x2DD6 }, // Lo   [7] ETHIOPIC SYLLABLE XYA..ETHIOPIC SYLLABLE XYO
    Interval{ 0x2DD8, 0x2DDE }, // Lo   [7] ETHIOPIC SYLLABLE GYA..ETHIOPIC SYLLABLE GYO
    Interval{ 0x2DE0, 0x2DFF }, // Mn  [32] COMBINING CYRILLIC LETTER BE..COMBINING CYRILLIC LETTER IOTIFIED BIG YUS
    Interval{ 0x3005, 0x3005 }, // Lm       IDEOGRAPHIC ITERATION MARK
    Interval{ 0x3006, 0x3006 }, // Lo       IDEOGRAPHIC CLOSING MARK
    Interval{ 0x3007, 0x3007 }, // Nl       IDEOGRAPHIC NUMBER ZERO
    Interval{ 0x3021, 0x3029 }, // Nl   [9] HANGZHOU NUMERAL ONE..HANGZHOU NUMERAL NINE
    Interval{ 0x302A, 0x302D }, // Mn   [4] IDEOGRAPHIC LEVEL TONE MARK..IDEOGRAPHIC ENTERING TONE MARK
    Interval{ 0x302E, 0x302F }, // Mc   [2] HANGUL SINGLE DOT TONE MARK..HANGUL DOUBLE DOT TONE MARK
    Interval{ 0x3031, 0x3035 }, // Lm   [5] VERTICAL KANA REPEAT MARK..VERTICAL KANA REPEAT MARK LOWER HALF
    Interval{ 0x3038, 0x303A }, // Nl   [3] HANGZHOU NUMERAL TEN..HANGZHOU NUMERAL THIRTY
    Interval{ 0x303B, 0x303B }, // Lm       VERTICAL IDEOGRAPHIC ITERATION MARK
    Interval{ 0x303C, 0x303C }, // Lo       MASU MARK
    Interval{ 0x3041, 0x3096 }, // Lo  [86] HIRAGANA LETTER SMALL A..HIRAGANA LETTER SMALL KE
    Interval{ 0x3099, 0x309A }, // Mn   [2] COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK..COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Interval{ 0x309B, 0x309C }, // Sk   [2] KATAKANA-HIRAGANA VOICED SOUND MARK..KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Interval{ 0x309D, 0x309E }, // Lm   [2] HIRAGANA ITERATION MARK..HIRAGANA VOICED ITERATION MARK
    Interval{ 0x309F, 0x309F }, // Lo       HIRAGANA DIGRAPH YORI
    Interval{ 0x30A1, 0x30FA }, // Lo  [90] KATAKANA LETTER SMALL A..KATAKANA LETTER VO
    Interval{ 0x30FB, 0x30FB }, // Po       KATAKANA MIDDLE DOT
    Interval{ 0x30FC, 0x30FE }, // Lm   [3] KATAKANA-HIRAGANA PROLONGED SOUND MARK..KATAKANA VOICED ITERATION MARK
    Interval{ 0x30FF, 0x30FF }, // Lo       KATAKANA DIGRAPH KOTO
    Interval{ 0x3105, 0x312F }, // Lo  [43] BOPOMOFO LETTER B..BOPOMOFO LETTER NN
    Interval{ 0x3131, 0x318E }, // Lo  [94] HANGUL LETTER KIYEOK..HANGUL LETTER ARAEAE
    Interval{ 0x31A0, 0x31BF }, // Lo  [32] BOPOMOFO LETTER BU..BOPOMOFO LETTER AH
    Interval{ 0x31F0, 0x31FF }, // Lo  [16] KATAKANA LETTER SMALL KU..KATAKANA LETTER SMALL RO
    Interval{ 0x3400, 0x4DBF }, // Lo [6592] CJK UNIFIED IDEOGRAPH-3400..CJK UNIFIED IDEOGRAPH-4DBF
    Interval{ 0x4E00, 0xA014 }, // Lo [21013] CJK UNIFIED IDEOGRAPH-4E00..YI SYLLABLE E
    Interval{ 0xA015, 0xA015 }, // Lm       YI SYLLABLE WU
    Interval{ 0xA016, 0xA48C }, // Lo [1143] YI SYLLABLE BIT..YI SYLLABLE YYR
    Interval{ 0xA4D0, 0xA4F7 }, // Lo  [40] LISU LETTER BA..LISU LETTER OE
    Interval{ 0xA4F8, 0xA4FD }, // Lm   [6] LISU LETTER TONE MYA TI..LISU LETTER TONE MYA JEU
    Interval{ 0xA500, 0xA60B }, // Lo [268] VAI SYLLABLE EE..VAI SYLLABLE NG
    Interval{ 0xA60C, 0xA60C }, // Lm       VAI SYLLABLE LENGTHENER
    Interval{ 0xA610, 0xA61F }, // Lo  [16] VAI SYLLABLE NDOLE FA..VAI SYMBOL JONG
    Interval{ 0xA620, 0xA629 }, // Nd  [10] VAI DIGIT ZERO..VAI DIGIT NINE
    Interval{ 0xA62A, 0xA62B }, // Lo   [2] VAI SYLLABLE NDOLE MA..VAI SYLLABLE NDOLE DO
    Interval{ 0xA640, 0xA66D }, // L&  [46] CYRILLIC CAPITAL LETTER ZEMLYA..CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA66E, 0xA66E }, // Lo       CYRILLIC LETTER MULTIOCULAR O
    Interval{ 0xA66F, 0xA66F }, // Mn       COMBINING CYRILLIC VZMET
    Interval{ 0xA674, 0xA67D }, // Mn  [10] COMBINING CYRILLIC LETTER UKRAINIAN IE..COMBINING CYRILLIC PAYEROK
    Interval{ 0xA67F, 0xA67F }, // Lm       CYRILLIC PAYEROK
    Interval{ 0xA680, 0xA69B }, // L&  [28] CYRILLIC CAPITAL LETTER DWE..CYRILLIC SMALL LETTER CROSSED O
    Interval{ 0xA69C, 0xA69D }, // Lm   [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    Interval{ 0xA69E, 0xA69F }, // Mn   [2] COMBINING CYRILLIC LETTER EF..COMBINING CYRILLIC LETTER IOTIFIED E
    Interval{ 0xA6A0, 0xA6E5 }, // Lo  [70] BAMUM LETTER A..BAMUM LETTER KI
    Interval{ 0xA6E6, 0xA6EF }, // Nl  [10] BAMUM LETTER MO..BAMUM LETTER KOGHOM
    Interval{ 0xA6F0, 0xA6F1 }, // Mn   [2] BAMUM COMBINING MARK KOQNDON..BAMUM COMBINING MARK TUKWENTIS
    Interval{ 0xA717, 0xA71F }, // Lm   [9] MODIFIER LETTER DOT VERTICAL BAR..MODIFIER LETTER LOW INVERTED EXCLAMATION MARK
    Interval{ 0xA722, 0xA76F }, // L&  [78] LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF..LATIN SMALL LETTER CON
    Interval{ 0xA770, 0xA770 }, // Lm       MODIFIER LETTER US
    Interval{ 0xA771, 0xA787 }, // L&  [23] LATIN SMALL LETTER DUM..LATIN SMALL LETTER INSULAR T
    Interval{ 0xA788, 0xA788 }, // Lm       MODIFIER LETTER LOW CIRCUMFLEX ACCENT
    Interval{ 0xA78B, 0xA78E }, // L&   [4] LATIN CAPITAL LETTER SALTILLO..LATIN SMALL LETTER L WITH RETROFLEX HOOK AND BELT
    Interval{ 0xA78F, 0xA78F }, // Lo       LATIN LETTER SINOLOGICAL DOT
    Interval{ 0xA790, 0xA7DC }, // L&  [77] LATIN CAPITAL LETTER N WITH DESCENDER..LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Interval{ 0xA7F1, 0xA7F4 }, // Lm   [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    Interval{ 0xA7F5, 0xA7F6 }, // L&   [2] LATIN CAPITAL LETTER REVERSED HALF H..LATIN SMALL LETTER REVERSED HALF H
    Interval{ 0xA7F7, 0xA7F7 }, // Lo       LATIN EPIGRAPHIC LETTER SIDEWAYS I
    Interval{ 0xA7F8, 0xA7F9 }, // Lm   [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    Interval{ 0xA7FA, 0xA7FA }, // L&       LATIN LETTER SMALL CAPITAL TURNED M
    Interval{ 0xA7FB, 0xA801 }, // Lo   [7] LATIN EPIGRAPHIC LETTER REVERSED F..SYLOTI NAGRI LETTER I
    Interval{ 0xA802, 0xA802 }, // Mn       SYLOTI NAGRI SIGN DVISVARA
    Interval{ 0xA803, 0xA805 }, // Lo   [3] SYLOTI NAGRI LETTER U..SYLOTI NAGRI LETTER O
    Interval{ 0xA806, 0xA806 }, // Mn       SYLOTI NAGRI SIGN HASANTA
    Interval{ 0xA807, 0xA80A }, // Lo   [4] SYLOTI NAGRI LETTER KO..SYLOTI NAGRI LETTER GHO
    Interval{ 0xA80B, 0xA80B }, // Mn       SYLOTI NAGRI SIGN ANUSVARA
    Interval{ 0xA80C, 0xA822 }, // Lo  [23] SYLOTI NAGRI LETTER CO..SYLOTI NAGRI LETTER HO
    Interval{ 0xA823, 0xA824 }, // Mc   [2] SYLOTI NAGRI VOWEL SIGN A..SYLOTI NAGRI VOWEL SIGN I
    Interval{ 0xA825, 0xA826 }, // Mn   [2] SYLOTI NAGRI VOWEL SIGN U..SYLOTI NAGRI VOWEL SIGN E
    Interval{ 0xA827, 0xA827 }, // Mc       SYLOTI NAGRI VOWEL SIGN OO
    Interval{ 0xA82C, 0xA82C }, // Mn       SYLOTI NAGRI SIGN ALTERNATE HASANTA
    Interval{ 0xA840, 0xA873 }, // Lo  [52] PHAGS-PA LETTER KA..PHAGS-PA LETTER CANDRABINDU
    Interval{ 0xA880, 0xA881 }, // Mc   [2] SAURASHTRA SIGN ANUSVARA..SAURASHTRA SIGN VISARGA
    Interval{ 0xA882, 0xA8B3 }, // Lo  [50] SAURASHTRA LETTER A..SAURASHTRA LETTER LLA
    Interval{ 0xA8B4, 0xA8C3 }, // Mc  [16] SAURASHTRA CONSONANT SIGN HAARU..SAURASHTRA VOWEL SIGN AU
    Interval{ 0xA8C4, 0xA8C5 }, // Mn   [2] SAURASHTRA SIGN VIRAMA..SAURASHTRA SIGN CANDRABINDU
    Interval{ 0xA8D0, 0xA8D9 }, // Nd  [10] SAURASHTRA DIGIT ZERO..SAURASHTRA DIGIT NINE
    Interval{ 0xA8E0, 0xA8F1 }, // Mn  [18] COMBINING DEVANAGARI DIGIT ZERO..COMBINING DEVANAGARI SIGN AVAGRAHA
    Interval{ 0xA8F2, 0xA8F7 }, // Lo   [6] DEVANAGARI SIGN SPACING CANDRABINDU..DEVANAGARI SIGN CANDRABINDU AVAGRAHA
    Interval{ 0xA8FB, 0xA8FB }, // Lo       DEVANAGARI HEADSTROKE
    Interval{ 0xA8FD, 0xA8FE }, // Lo   [2] DEVANAGARI JAIN OM..DEVANAGARI LETTER AY
    Interval{ 0xA8FF, 0xA8FF }, // Mn       DEVANAGARI VOWEL SIGN AY
    Interval{ 0xA900, 0xA909 }, // Nd  [10] KAYAH LI DIGIT ZERO..KAYAH LI DIGIT NINE
    Interval{ 0xA90A, 0xA925 }, // Lo  [28] KAYAH LI LETTER KA..KAYAH LI LETTER OO
    Interval{ 0xA926, 0xA92D }, // Mn   [8] KAYAH LI VOWEL UE..KAYAH LI TONE CALYA PLOPHU
    Interval{ 0xA930, 0xA946 }, // Lo  [23] REJANG LETTER KA..REJANG LETTER A
    Interval{ 0xA947, 0xA951 }, // Mn  [11] REJANG VOWEL SIGN I..REJANG CONSONANT SIGN R
    Interval{ 0xA952, 0xA953 }, // Mc   [2] REJANG CONSONANT SIGN H..REJANG VIRAMA
    Interval{ 0xA960, 0xA97C }, // Lo  [29] HANGUL CHOSEONG TIKEUT-MIEUM..HANGUL CHOSEONG SSANGYEORINHIEUH
    Interval{ 0xA980, 0xA982 }, // Mn   [3] JAVANESE SIGN PANYANGGA..JAVANESE SIGN LAYAR
    Interval{ 0xA983, 0xA983 }, // Mc       JAVANESE SIGN WIGNYAN
    Interval{ 0xA984, 0xA9B2 }, // Lo  [47] JAVANESE LETTER A..JAVANESE LETTER HA
    Interval{ 0xA9B3, 0xA9B3 }, // Mn       JAVANESE SIGN CECAK TELU
    Interval{ 0xA9B4, 0xA9B5 }, // Mc   [2] JAVANESE VOWEL SIGN TARUNG..JAVANESE VOWEL SIGN TOLONG
    Interval{ 0xA9B6, 0xA9B9 }, // Mn   [4] JAVANESE VOWEL SIGN WULU..JAVANESE VOWEL SIGN SUKU MENDUT
    Interval{ 0xA9BA, 0xA9BB }, // Mc   [2] JAVANESE VOWEL SIGN TALING..JAVANESE VOWEL SIGN DIRGA MURE
    Interval{ 0xA9BC, 0xA9BD }, // Mn   [2] JAVANESE VOWEL SIGN PEPET..JAVANESE CONSONANT SIGN KERET
    Interval{ 0xA9BE, 0xA9C0 }, // Mc   [3] JAVANESE CONSONANT SIGN PENGKAL..JAVANESE PANGKON
    Interval{ 0xA9CF, 0xA9CF }, // Lm       JAVANESE PANGRANGKEP
    Interval{ 0xA9D0, 0xA9D9 }, // Nd  [10] JAVANESE DIGIT ZERO..JAVANESE DIGIT NINE
    Interval{ 0xA9E0, 0xA9E4 }, // Lo   [5] MYANMAR LETTER SHAN GHA..MYANMAR LETTER SHAN BHA
    Interval{ 0xA9E5, 0xA9E5 }, // Mn       MYANMAR SIGN SHAN SAW
    Interval{ 0xA9E6, 0xA9E6 }, // Lm       MYANMAR MODIFIER LETTER SHAN REDUPLICATION
    Interval{ 0xA9E7, 0xA9EF }, // Lo   [9] MYANMAR LETTER TAI LAING NYA..MYANMAR LETTER TAI LAING NNA
    Interval{ 0xA9F0, 0xA9F9 }, // Nd  [10] MYANMAR TAI LAING DIGIT ZERO..MYANMAR TAI LAING DIGIT NINE
    Interval{ 0xA9FA, 0xA9FE }, // Lo   [5] MYANMAR LETTER TAI LAING LLA..MYANMAR LETTER TAI LAING BHA
    Interval{ 0xAA00, 0xAA28 }, // Lo  [41] CHAM LETTER A..CHAM LETTER HA
    Interval{ 0xAA29, 0xAA2E }, // Mn   [6] CHAM VOWEL SIGN AA..CHAM VOWEL SIGN OE
    Interval{ 0xAA2F, 0xAA30 }, // Mc   [2] CHAM VOWEL SIGN O..CHAM VOWEL SIGN AI
    Interval{ 0xAA31, 0xAA32 }, // Mn   [2] CHAM VOWEL SIGN AU..CHAM VOWEL SIGN UE
    Interval{ 0xAA33, 0xAA34 }, // Mc   [2] CHAM CONSONANT SIGN YA..CHAM CONSONANT SIGN RA
    Interval{ 0xAA35, 0xAA36 }, // Mn   [2] CHAM CONSONANT SIGN LA..CHAM CONSONANT SIGN WA
    Interval{ 0xAA40, 0xAA42 }, // Lo   [3] CHAM LETTER FINAL K..CHAM LETTER FINAL NG
    Interval{ 0xAA43, 0xAA43 }, // Mn       CHAM CONSONANT SIGN FINAL NG
    Interval{ 0xAA44, 0xAA4B }, // Lo   [8] CHAM LETTER FINAL CH..CHAM LETTER FINAL SS
    Interval{ 0xAA4C, 0xAA4C }, // Mn       CHAM CONSONANT SIGN FINAL M
    Interval{ 0xAA4D, 0xAA4D }, // Mc       CHAM CONSONANT SIGN FINAL H
    Interval{ 0xAA50, 0xAA59 }, // Nd  [10] CHAM DIGIT ZERO..CHAM DIGIT NINE
    Interval{ 0xAA60, 0xAA6F }, // Lo  [16] MYANMAR LETTER KHAMTI GA..MYANMAR LETTER KHAMTI FA
    Interval{ 0xAA70, 0xAA70 }, // Lm       MYANMAR MODIFIER LETTER KHAMTI REDUPLICATION
    Interval{ 0xAA71, 0xAA76 }, // Lo   [6] MYANMAR LETTER KHAMTI XA..MYANMAR LOGOGRAM KHAMTI HM
    Interval{ 0xAA7A, 0xAA7A }, // Lo       MYANMAR LETTER AITON RA
    Interval{ 0xAA7B, 0xAA7B }, // Mc       MYANMAR SIGN PAO KAREN TONE
    Interval{ 0xAA7C, 0xAA7C }, // Mn       MYANMAR SIGN TAI LAING TONE-2
    Interval{ 0xAA7D, 0xAA7D }, // Mc       MYANMAR SIGN TAI LAING TONE-5
    Interval{ 0xAA7E, 0xAAAF }, // Lo  [50] MYANMAR LETTER SHWE PALAUNG CHA..TAI VIET LETTER HIGH O
    Interval{ 0xAAB0, 0xAAB0 }, // Mn       TAI VIET MAI KANG
    Interval{ 0xAAB1, 0xAAB1 }, // Lo       TAI VIET VOWEL AA
    Interval{ 0xAAB2, 0xAAB4 }, // Mn   [3] TAI VIET VOWEL I..TAI VIET VOWEL U
    Interval{ 0xAAB5, 0xAAB6 }, // Lo   [2] TAI VIET VOWEL E..TAI VIET VOWEL O
    Interval{ 0xAAB7, 0xAAB8 }, // Mn   [2] TAI VIET MAI KHIT..TAI VIET VOWEL IA
    Interval{ 0xAAB9, 0xAABD }, // Lo   [5] TAI VIET VOWEL UEA..TAI VIET VOWEL AN
    Interval{ 0xAABE, 0xAABF }, // Mn   [2] TAI VIET VOWEL AM..TAI VIET TONE MAI EK
    Interval{ 0xAAC0, 0xAAC0 }, // Lo       TAI VIET TONE MAI NUENG
    Interval{ 0xAAC1, 0xAAC1 }, // Mn       TAI VIET TONE MAI THO
    Interval{ 0xAAC2, 0xAAC2 }, // Lo       TAI VIET TONE MAI SONG
    Interval{ 0xAADB, 0xAADC }, // Lo   [2] TAI VIET SYMBOL KON..TAI VIET SYMBOL NUENG
    Interval{ 0xAADD, 0xAADD }, // Lm       TAI VIET SYMBOL SAM
    Interval{ 0xAAE0, 0xAAEA }, // Lo  [11] MEETEI MAYEK LETTER E..MEETEI MAYEK LETTER SSA
    Interval{ 0xAAEB, 0xAAEB }, // Mc       MEETEI MAYEK VOWEL SIGN II
    Interval{ 0xAAEC, 0xAAED }, // Mn   [2] MEETEI MAYEK VOWEL SIGN UU..MEETEI MAYEK VOWEL SIGN AAI
    Interval{ 0xAAEE, 0xAAEF }, // Mc   [2] MEETEI MAYEK VOWEL SIGN AU..MEETEI MAYEK VOWEL SIGN AAU
    Interval{ 0xAAF2, 0xAAF2 }, // Lo       MEETEI MAYEK ANJI
    Interval{ 0xAAF3, 0xAAF4 }, // Lm   [2] MEETEI MAYEK SYLLABLE REPETITION MARK..MEETEI MAYEK WORD REPETITION MARK
    Interval{ 0xAAF5, 0xAAF5 }, // Mc       MEETEI MAYEK VOWEL SIGN VISARGA
    Interval{ 0xAAF6, 0xAAF6 }, // Mn       MEETEI MAYEK VIRAMA
    Interval{ 0xAB01, 0xAB06 }, // Lo   [6] ETHIOPIC SYLLABLE TTHU..ETHIOPIC SYLLABLE TTHO
    Interval{ 0xAB09, 0xAB0E }, // Lo   [6] ETHIOPIC SYLLABLE DDHU..ETHIOPIC SYLLABLE DDHO
    Interval{ 0xAB11, 0xAB16 }, // Lo   [6] ETHIOPIC SYLLABLE DZU..ETHIOPIC SYLLABLE DZO
    Interval{ 0xAB20, 0xAB26 }, // Lo   [7] ETHIOPIC SYLLABLE CCHHA..ETHIOPIC SYLLABLE CCHHO
    Interval{ 0xAB28, 0xAB2E }, // Lo   [7] ETHIOPIC SYLLABLE BBA..ETHIOPIC SYLLABLE BBO
    Interval{ 0xAB30, 0xAB5A }, // L&  [43] LATIN SMALL LETTER BARRED ALPHA..LATIN SMALL LETTER Y WITH SHORT RIGHT LEG
    Interval{ 0xAB5C, 0xAB5F }, // Lm   [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    Interval{ 0xAB60, 0xAB68 }, // L&   [9] LATIN SMALL LETTER SAKHA YAT..LATIN SMALL LETTER TURNED R WITH MIDDLE TILDE
    Interval{ 0xAB69, 0xAB69 }, // Lm       MODIFIER LETTER SMALL TURNED W
    Interval{ 0xAB70, 0xABBF }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xABC0, 0xABE2 }, // Lo  [35] MEETEI MAYEK LETTER KOK..MEETEI MAYEK LETTER I LONSUM
    Interval{ 0xABE3, 0xABE4 }, // Mc   [2] MEETEI MAYEK VOWEL SIGN ONAP..MEETEI MAYEK VOWEL SIGN INAP
    Interval{ 0xABE5, 0xABE5 }, // Mn       MEETEI MAYEK VOWEL SIGN ANAP
    Interval{ 0xABE6, 0xABE7 }, // Mc   [2] MEETEI MAYEK VOWEL SIGN YENAP..MEETEI MAYEK VOWEL SIGN SOUNAP
    Interval{ 0xABE8, 0xABE8 }, // Mn       MEETEI MAYEK VOWEL SIGN UNAP
    Interval{ 0xABE9, 0xABEA }, // Mc   [2] MEETEI MAYEK VOWEL SIGN CHEINAP..MEETEI MAYEK VOWEL SIGN NUNG
    Interval{ 0xABEC, 0xABEC }, // Mc       MEETEI MAYEK LUM IYEK
    Interval{ 0xABED, 0xABED }, // Mn       MEETEI MAYEK APUN IYEK
    Interval{ 0xABF0, 0xABF9 }, // Nd  [10] MEETEI MAYEK DIGIT ZERO..MEETEI MAYEK DIGIT NINE
    Interval{ 0xAC00, 0xD7A3 }, // Lo [11172] HANGUL SYLLABLE GA..HANGUL SYLLABLE HIH
    Interval{ 0xD7B0, 0xD7C6 }, // Lo  [23] HANGUL JUNGSEONG O-YEO..HANGUL JUNGSEONG ARAEA-E
    Interval{ 0xD7CB, 0xD7FB }, // Lo  [49] HANGUL JONGSEONG NIEUN-RIEUL..HANGUL JONGSEONG PHIEUPH-THIEUTH
    Interval{ 0xF900, 0xFA6D }, // Lo [366] CJK COMPATIBILITY IDEOGRAPH-F900..CJK COMPATIBILITY IDEOGRAPH-FA6D
    Interval{ 0xFA70, 0xFAD9 }, // Lo [106] CJK COMPATIBILITY IDEOGRAPH-FA70..CJK COMPATIBILITY IDEOGRAPH-FAD9
    Interval{ 0xFB00, 0xFB06 }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFB1D, 0xFB1D }, // Lo       HEBREW LETTER YOD WITH HIRIQ
    Interval{ 0xFB1E, 0xFB1E }, // Mn       HEBREW POINT JUDEO-SPANISH VARIKA
    Interval{ 0xFB1F, 0xFB28 }, // Lo  [10] HEBREW LIGATURE YIDDISH YOD YOD PATAH..HEBREW LETTER WIDE TAV
    Interval{ 0xFB2A, 0xFB36 }, // Lo  [13] HEBREW LETTER SHIN WITH SHIN DOT..HEBREW LETTER ZAYIN WITH DAGESH
    Interval{ 0xFB38, 0xFB3C }, // Lo   [5] HEBREW LETTER TET WITH DAGESH..HEBREW LETTER LAMED WITH DAGESH
    Interval{ 0xFB3E, 0xFB3E }, // Lo       HEBREW LETTER MEM WITH DAGESH
    Interval{ 0xFB40, 0xFB41 }, // Lo   [2] HEBREW LETTER NUN WITH DAGESH..HEBREW LETTER SAMEKH WITH DAGESH
    Interval{ 0xFB43, 0xFB44 }, // Lo   [2] HEBREW LETTER FINAL PE WITH DAGESH..HEBREW LETTER PE WITH DAGESH
    Interval{ 0xFB46, 0xFBB1 }, // Lo [108] HEBREW LETTER TSADI WITH DAGESH..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE FINAL FORM
    Interval{ 0xFBD3, 0xFD3D }, // Lo [363] ARABIC LETTER NG ISOLATED FORM..ARABIC LIGATURE ALEF WITH FATHATAN ISOLATED FORM
    Interval{ 0xFD50, 0xFD8F }, // Lo  [64] ARABIC LIGATURE TEH WITH JEEM WITH MEEM INITIAL FORM..ARABIC LIGATURE MEEM WITH KHAH WITH MEEM INITIAL FORM
    Interval{ 0xFD92, 0xFDC7 }, // Lo  [54] ARABIC LIGATURE MEEM WITH JEEM WITH KHAH INITIAL FORM..ARABIC LIGATURE NOON WITH JEEM WITH YEH FINAL FORM
    Interval{ 0xFDF0, 0xFDFB }, // Lo  [12] ARABIC LIGATURE SALLA USED AS KORANIC STOP SIGN ISOLATED FORM..ARABIC LIGATURE JALLAJALALOUHOU
    Interval{ 0xFE00, 0xFE0F }, // Mn  [16] VARIATION SELECTOR-1..VARIATION SELECTOR-16
    Interval{ 0xFE20, 0xFE2F }, // Mn  [16] COMBINING LIGATURE LEFT HALF..COMBINING CYRILLIC TITLO RIGHT HALF
    Interval{ 0xFE33, 0xFE34 }, // Pc   [2] PRESENTATION FORM FOR VERTICAL LOW LINE..PRESENTATION FORM FOR VERTICAL WAVY LOW LINE
    Interval{ 0xFE4D, 0xFE4F }, // Pc   [3] DASHED LOW LINE..WAVY LOW LINE
    Interval{ 0xFE70, 0xFE74 }, // Lo   [5] ARABIC FATHATAN ISOLATED FORM..ARABIC KASRATAN ISOLATED FORM
    Interval{ 0xFE76, 0xFEFC }, // Lo [135] ARABIC FATHA ISOLATED FORM..ARABIC LIGATURE LAM WITH ALEF FINAL FORM
    Interval{ 0xFF10, 0xFF19 }, // Nd  [10] FULLWIDTH DIGIT ZERO..FULLWIDTH DIGIT NINE
    Interval{ 0xFF21, 0xFF3A }, // L&  [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Interval{ 0xFF3F, 0xFF3F }, // Pc       FULLWIDTH LOW LINE
    Interval{ 0xFF41, 0xFF5A }, // L&  [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Interval{ 0xFF65, 0xFF65 }, // Po       HALFWIDTH KATAKANA MIDDLE DOT
    Interval{ 0xFF66, 0xFF6F }, // Lo  [10] HALFWIDTH KATAKANA LETTER WO..HALFWIDTH KATAKANA LETTER SMALL TU
    Interval{ 0xFF70, 0xFF70 }, // Lm       HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Interval{ 0xFF71, 0xFF9D }, // Lo  [45] HALFWIDTH KATAKANA LETTER A..HALFWIDTH KATAKANA LETTER N
    Interval{ 0xFF9E, 0xFF9F }, // Lm   [2] HALFWIDTH KATAKANA VOICED SOUND MARK..HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK
    Interval{ 0xFFA0, 0xFFBE }, // Lo  [31] HALFWIDTH HANGUL FILLER..HALFWIDTH HANGUL LETTER HIEUH
    Interval{ 0xFFC2, 0xFFC7 }, // Lo   [6] HALFWIDTH HANGUL LETTER A..HALFWIDTH HANGUL LETTER E
    Interval{ 0xFFCA, 0xFFCF }, // Lo   [6] HALFWIDTH HANGUL LETTER YEO..HALFWIDTH HANGUL LETTER OE
    Interval{ 0xFFD2, 0xFFD7 }, // Lo   [6] HALFWIDTH HANGUL LETTER YO..HALFWIDTH HANGUL LETTER YU
    Interval{ 0xFFDA, 0xFFDC }, // Lo   [3] HALFWIDTH HANGUL LETTER EU..HALFWIDTH HANGUL LETTER I
    Interval{ 0x10000, 0x1000B }, // Lo  [12] LINEAR B SYLLABLE B008 A..LINEAR B SYLLABLE B046 JE
    Interval{ 0x1000D, 0x10026 }, // Lo  [26] LINEAR B SYLLABLE B036 JO..LINEAR B SYLLABLE B032 QO
    Interval{ 0x10028, 0x1003A }, // Lo  [19] LINEAR B SYLLABLE B060 RA..LINEAR B SYLLABLE B042 WO
    Interval{ 0x1003C, 0x1003D }, // Lo   [2] LINEAR B SYLLABLE B017 ZA..LINEAR B SYLLABLE B074 ZE
    Interval{ 0x1003F, 0x1004D }, // Lo  [15] LINEAR B SYLLABLE B020 ZO..LINEAR B SYLLABLE B091 TWO
    Interval{ 0x10050, 0x1005D }, // Lo  [14] LINEAR B SYMBOL B018..LINEAR B SYMBOL B089
    Interval{ 0x10080, 0x100FA }, // Lo [123] LINEAR B IDEOGRAM B100 MAN..LINEAR B IDEOGRAM VESSEL B305
    Interval{ 0x10140, 0x10174 }, // Nl  [53] GREEK ACROPHONIC ATTIC ONE QUARTER..GREEK ACROPHONIC STRATIAN FIFTY MNAS
    Interval{ 0x101FD, 0x101FD }, // Mn       PHAISTOS DISC SIGN COMBINING OBLIQUE STROKE
    Interval{ 0x10280, 0x1029C }, // Lo  [29] LYCIAN LETTER A..LYCIAN LETTER X
    Interval{ 0x102A0, 0x102D0 }, // Lo  [49] CARIAN LETTER A..CARIAN LETTER UUU3
    Interval{ 0x102E0, 0x102E0 }, // Mn       COPTIC EPACT THOUSANDS MARK
    Interval{ 0x10300, 0x1031F }, // Lo  [32] OLD ITALIC LETTER A..OLD ITALIC LETTER ESS
    Interval{ 0x1032D, 0x10340 }, // Lo  [20] OLD ITALIC LETTER YE..GOTHIC LETTER PAIRTHRA
    Interval{ 0x10341, 0x10341 }, // Nl       GOTHIC LETTER NINETY
    Interval{ 0x10342, 0x10349 }, // Lo   [8] GOTHIC LETTER RAIDA..GOTHIC LETTER OTHAL
    Interval{ 0x1034A, 0x1034A }, // Nl       GOTHIC LETTER NINE HUNDRED
    Interval{ 0x10350, 0x10375 }, // Lo  [38] OLD PERMIC LETTER AN..OLD PERMIC LETTER IA
    Interval{ 0x10376, 0x1037A }, // Mn   [5] COMBINING OLD PERMIC LETTER AN..COMBINING OLD PERMIC LETTER SII
    Interval{ 0x10380, 0x1039D }, // Lo  [30] UGARITIC LETTER ALPA..UGARITIC LETTER SSU
    Interval{ 0x103A0, 0x103C3 }, // Lo  [36] OLD PERSIAN SIGN A..OLD PERSIAN SIGN HA
    Interval{ 0x103C8, 0x103CF }, // Lo   [8] OLD PERSIAN SIGN AURAMAZDAA..OLD PERSIAN SIGN BUUMISH
    Interval{ 0x103D1, 0x103D5 }, // Nl   [5] OLD PERSIAN NUMBER ONE..OLD PERSIAN NUMBER HUNDRED
    Interval{ 0x10400, 0x1044F }, // L&  [80] DESERET CAPITAL LETTER LONG I..DESERET SMALL LETTER EW
    Interval{ 0x10450, 0x1049D }, // Lo  [78] SHAVIAN LETTER PEEP..OSMANYA LETTER OO
    Interval{ 0x104A0, 0x104A9 }, // Nd  [10] OSMANYA DIGIT ZERO..OSMANYA DIGIT NINE
    Interval{ 0x104B0, 0x104D3 }, // L&  [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Interval{ 0x104D8, 0x104FB }, // L&  [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Interval{ 0x10500, 0x10527 }, // Lo  [40] ELBASAN LETTER A..ELBASAN LETTER KHE
    Interval{ 0x10530, 0x10563 }, // Lo  [52] CAUCASIAN ALBANIAN LETTER ALT..CAUCASIAN ALBANIAN LETTER KIW
    Interval{ 0x10570, 0x1057A }, // L&  [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Interval{ 0x1057C, 0x1058A }, // L&  [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Interval{ 0x1058C, 0x10592 }, // L&   [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Interval{ 0x10594, 0x10595 }, // L&   [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Interval{ 0x10597, 0x105A1 }, // L&  [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Interval{ 0x105A3, 0x105B1 }, // L&  [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Interval{ 0x105B3, 0x105B9 }, // L&   [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Interval{ 0x105BB, 0x105BC }, // L&   [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Interval{ 0x105C0, 0x105F3 }, // Lo  [52] TODHRI LETTER A..TODHRI LETTER OO
    Interval{ 0x10600, 0x10736 }, // Lo [311] LINEAR A SIGN AB001..LINEAR A SIGN A664
    Interval{ 0x10740, 0x10755 }, // Lo  [22] LINEAR A SIGN A701 A..LINEAR A SIGN A732 JE
    Interval{ 0x10760, 0x10767 }, // Lo   [8] LINEAR A SIGN A800..LINEAR A SIGN A807
    Interval{ 0x10780, 0x10785 }, // Lm   [6] MODIFIER LETTER SMALL CAPITAL AA..MODIFIER LETTER SMALL B WITH HOOK
    Interval{ 0x10787, 0x107B0 }, // Lm  [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Interval{ 0x107B2, 0x107BA }, // Lm   [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Interval{ 0x10800, 0x10805 }, // Lo   [6] CYPRIOT SYLLABLE A..CYPRIOT SYLLABLE JA
    Interval{ 0x10808, 0x10808 }, // Lo       CYPRIOT SYLLABLE JO
    Interval{ 0x1080A, 0x10835 }, // Lo  [44] CYPRIOT SYLLABLE KA..CYPRIOT SYLLABLE WO
    Interval{ 0x10837, 0x10838 }, // Lo   [2] CYPRIOT SYLLABLE XA..CYPRIOT SYLLABLE XE
    Interval{ 0x1083C, 0x1083C }, // Lo       CYPRIOT SYLLABLE ZA
    Interval{ 0x1083F, 0x10855 }, // Lo  [23] CYPRIOT SYLLABLE ZO..IMPERIAL ARAMAIC LETTER TAW
    Interval{ 0x10860, 0x10876 }, // Lo  [23] PALMYRENE LETTER ALEPH..PALMYRENE LETTER TAW
    Interval{ 0x10880, 0x1089E }, // Lo  [31] NABATAEAN LETTER FINAL ALEPH..NABATAEAN LETTER TAW
    Interval{ 0x108E0, 0x108F2 }, // Lo  [19] HATRAN LETTER ALEPH..HATRAN LETTER QOPH
    Interval{ 0x108F4, 0x108F5 }, // Lo   [2] HATRAN LETTER SHIN..HATRAN LETTER TAW
    Interval{ 0x10900, 0x10915 }, // Lo  [22] PHOENICIAN LETTER ALF..PHOENICIAN LETTER TAU
    Interval{ 0x10920, 0x10939 }, // Lo  [26] LYDIAN LETTER A..LYDIAN LETTER C
    Interval{ 0x10940, 0x10959 }, // Lo  [26] SIDETIC LETTER N01..SIDETIC LETTER N26
    Interval{ 0x10980, 0x109B7 }, // Lo  [56] MEROITIC HIEROGLYPHIC LETTER A..MEROITIC CURSIVE LETTER DA
    Interval{ 0x109BE, 0x109BF }, // Lo   [2] MEROITIC CURSIVE LOGOGRAM RMT..MEROITIC CURSIVE LOGOGRAM IMN
    Interval{ 0x10A00, 0x10A00 }, // Lo       KHAROSHTHI LETTER A
    Interval{ 0x10A01, 0x10A03 }, // Mn   [3] KHAROSHTHI VOWEL SIGN I..KHAROSHTHI VOWEL SIGN VOCALIC R
    Interval{ 0x10A05, 0x10A06 }, // Mn   [2] KHAROSHTHI VOWEL SIGN E..KHAROSHTHI VOWEL SIGN O
    Interval{ 0x10A0C, 0x10A0F }, // Mn   [4] KHAROSHTHI VOWEL LENGTH MARK..KHAROSHTHI SIGN VISARGA
    Interval{ 0x10A10, 0x10A13 }, // Lo   [4] KHAROSHTHI LETTER KA..KHAROSHTHI LETTER GHA
    Interval{ 0x10A15, 0x10A17 }, // Lo   [3] KHAROSHTHI LETTER CA..KHAROSHTHI LETTER JA
    Interval{ 0x10A19, 0x10A35 }, // Lo  [29] KHAROSHTHI LETTER NYA..KHAROSHTHI LETTER VHA
    Interval{ 0x10A38, 0x10A3A }, // Mn   [3] KHAROSHTHI SIGN BAR ABOVE..KHAROSHTHI SIGN DOT BELOW
    Interval{ 0x10A3F, 0x10A3F }, // Mn       KHAROSHTHI VIRAMA
    Interval{ 0x10A60, 0x10A7C }, // Lo  [29] OLD SOUTH ARABIAN LETTER HE..OLD SOUTH ARABIAN LETTER THETH
    Interval{ 0x10A80, 0x10A9C }, // Lo  [29] OLD NORTH ARABIAN LETTER HEH..OLD NORTH ARABIAN LETTER ZAH
    Interval{ 0x10AC0, 0x10AC7 }, // Lo   [8] MANICHAEAN LETTER ALEPH..MANICHAEAN LETTER WAW
    Interval{ 0x10AC9, 0x10AE4 }, // Lo  [28] MANICHAEAN LETTER ZAYIN..MANICHAEAN LETTER TAW
    Interval{ 0x10AE5, 0x10AE6 }, // Mn   [2] MANICHAEAN ABBREVIATION MARK ABOVE..MANICHAEAN ABBREVIATION MARK BELOW
    Interval{ 0x10B00, 0x10B35 }, // Lo  [54] AVESTAN LETTER A..AVESTAN LETTER HE
    Interval{ 0x10B40, 0x10B55 }, // Lo  [22] INSCRIPTIONAL PARTHIAN LETTER ALEPH..INSCRIPTIONAL PARTHIAN LETTER TAW
    Interval{ 0x10B60, 0x10B72 }, // Lo  [19] INSCRIPTIONAL PAHLAVI LETTER ALEPH..INSCRIPTIONAL PAHLAVI LETTER TAW
    Interval{ 0x10B80, 0x10B91 }, // Lo  [18] PSALTER PAHLAVI LETTER ALEPH..PSALTER PAHLAVI LETTER TAW
    Interval{ 0x10C00, 0x10C48 }, // Lo  [73] OLD TURKIC LETTER ORKHON A..OLD TURKIC LETTER ORKHON BASH
    Interval{ 0x10C80, 0x10CB2 }, // L&  [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Interval{ 0x10CC0, 0x10CF2 }, // L&  [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Interval{ 0x10D00, 0x10D23 }, // Lo  [36] HANIFI ROHINGYA LETTER A..HANIFI ROHINGYA MARK NA KHONNA
    Interval{ 0x10D24, 0x10D27 }, // Mn   [4] HANIFI ROHINGYA SIGN HARBAHAY..HANIFI ROHINGYA SIGN TASSI
    Interval{ 0x10D30, 0x10D39 }, // Nd  [10] HANIFI ROHINGYA DIGIT ZERO..HANIFI ROHINGYA DIGIT NINE
    Interval{ 0x10D40, 0x10D49 }, // Nd  [10] GARAY DIGIT ZERO..GARAY DIGIT NINE
    Interval{ 0x10D4A, 0x10D4D }, // Lo   [4] GARAY VOWEL SIGN A..GARAY VOWEL SIGN EE
    Interval{ 0x10D4E, 0x10D4E }, // Lm       GARAY VOWEL LENGTH MARK
    Interval{ 0x10D4F, 0x10D4F }, // Lo       GARAY SUKUN
    Interval{ 0x10D50, 0x10D65 }, // L&  [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Interval{ 0x10D69, 0x10D6D }, // Mn   [5] GARAY VOWEL SIGN E..GARAY CONSONANT NASALIZATION MARK
    Interval{ 0x10D6F, 0x10D6F }, // Lm       GARAY REDUPLICATION MARK
    Interval{ 0x10D70, 0x10D85 }, // L&  [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Interval{ 0x10E80, 0x10EA9 }, // Lo  [42] YEZIDI LETTER ELIF..YEZIDI LETTER ET
    Interval{ 0x10EAB, 0x10EAC }, // Mn   [2] YEZIDI COMBINING HAMZA MARK..YEZIDI COMBINING MADDA MARK
    Interval{ 0x10EB0, 0x10EB1 }, // Lo   [2] YEZIDI LETTER LAM WITH DOT ABOVE..YEZIDI LETTER YOT WITH CIRCUMFLEX ABOVE
    Interval{ 0x10EC2, 0x10EC4 }, // Lo   [3] ARABIC LETTER DAL WITH TWO DOTS VERTICALLY BELOW..ARABIC LETTER KAF WITH TWO DOTS VERTICALLY BELOW
    Interval{ 0x10EC5, 0x10EC5 }, // Lm       ARABIC SMALL YEH BARREE WITH TWO DOTS BELOW
    Interval{ 0x10EC6, 0x10EC7 }, // Lo   [2] ARABIC LETTER THIN NOON..ARABIC LETTER YEH WITH FOUR DOTS BELOW
    Interval{ 0x10EFA, 0x10EFF }, // Mn   [6] ARABIC DOUBLE VERTICAL BAR BELOW..ARABIC SMALL LOW WORD MADDA
    Interval{ 0x10F00, 0x10F1C }, // Lo  [29] OLD SOGDIAN LETTER ALEPH..OLD SOGDIAN LETTER FINAL TAW WITH VERTICAL TAIL
    Interval{ 0x10F27, 0x10F27 }, // Lo       OLD SOGDIAN LIGATURE AYIN-DALETH
    Interval{ 0x10F30, 0x10F45 }, // Lo  [22] SOGDIAN LETTER ALEPH..SOGDIAN INDEPENDENT SHIN
    Interval{ 0x10F46, 0x10F50 }, // Mn  [11] SOGDIAN COMBINING DOT BELOW..SOGDIAN COMBINING STROKE BELOW
    Interval{ 0x10F70, 0x10F81 }, // Lo  [18] OLD UYGHUR LETTER ALEPH..OLD UYGHUR LETTER LESH
    Interval{ 0x10F82, 0x10F85 }, // Mn   [4] OLD UYGHUR COMBINING DOT ABOVE..OLD UYGHUR COMBINING TWO DOTS BELOW
    Interval{ 0x10FB0, 0x10FC4 }, // Lo  [21] CHORASMIAN LETTER ALEPH..CHORASMIAN LETTER TAW
    Interval{ 0x10FE0, 0x10FF6 }, // Lo  [23] ELYMAIC LETTER ALEPH..ELYMAIC LIGATURE ZAYIN-YODH
    Interval{ 0x11000, 0x11000 }, // Mc       BRAHMI SIGN CANDRABINDU
    Interval{ 0x11001, 0x11001 }, // Mn       BRAHMI SIGN ANUSVARA
    Interval{ 0x11002, 0x11002 }, // Mc       BRAHMI SIGN VISARGA
    Interval{ 0x11003, 0x11037 }, // Lo  [53] BRAHMI SIGN JIHVAMULIYA..BRAHMI LETTER OLD TAMIL NNNA
    Interval{ 0x11038, 0x11046 }, // Mn  [15] BRAHMI VOWEL SIGN AA..BRAHMI VIRAMA
    Interval{ 0x11066, 0x1106F }, // Nd  [10] BRAHMI DIGIT ZERO..BRAHMI DIGIT NINE
    Interval{ 0x11070, 0x11070 }, // Mn       BRAHMI SIGN OLD TAMIL VIRAMA
    Interval{ 0x11071, 0x11072 }, // Lo   [2] BRAHMI LETTER OLD TAMIL SHORT E..BRAHMI LETTER OLD TAMIL SHORT O
    Interval{ 0x11073, 0x11074 }, // Mn   [2] BRAHMI VOWEL SIGN OLD TAMIL SHORT E..BRAHMI VOWEL SIGN OLD TAMIL SHORT O
    Interval{ 0x11075, 0x11075 }, // Lo       BRAHMI LETTER OLD TAMIL LLA
    Interval{ 0x1107F, 0x11081 }, // Mn   [3] BRAHMI NUMBER JOINER..KAITHI SIGN ANUSVARA
    Interval{ 0x11082, 0x11082 }, // Mc       KAITHI SIGN VISARGA
    Interval{ 0x11083, 0x110AF }, // Lo  [45] KAITHI LETTER A..KAITHI LETTER HA
    Interval{ 0x110B0, 0x110B2 }, // Mc   [3] KAITHI VOWEL SIGN AA..KAITHI VOWEL SIGN II
    Interval{ 0x110B3, 0x110B6 }, // Mn   [4] KAITHI VOWEL SIGN U..KAITHI VOWEL SIGN AI
    Interval{ 0x110B7, 0x110B8 }, // Mc   [2] KAITHI VOWEL SIGN O..KAITHI VOWEL SIGN AU
    Interval{ 0x110B9, 0x110BA }, // Mn   [2] KAITHI SIGN VIRAMA..KAITHI SIGN NUKTA
    Interval{ 0x110C2, 0x110C2 }, // Mn       KAITHI VOWEL SIGN VOCALIC R
    Interval{ 0x110D0, 0x110E8 }, // Lo  [25] SORA SOMPENG LETTER SAH..SORA SOMPENG LETTER MAE
    Interval{ 0x110F0, 0x110F9 }, // Nd  [10] SORA SOMPENG DIGIT ZERO..SORA SOMPENG DIGIT NINE
    Interval{ 0x11100, 0x11102 }, // Mn   [3] CHAKMA SIGN CANDRABINDU..CHAKMA SIGN VISARGA
    Interval{ 0x11103, 0x11126 }, // Lo  [36] CHAKMA LETTER AA..CHAKMA LETTER HAA
    Interval{ 0x11127, 0x1112B }, // Mn   [5] CHAKMA VOWEL SIGN A..CHAKMA VOWEL SIGN UU
    Interval{ 0x1112C, 0x1112C }, // Mc       CHAKMA VOWEL SIGN E
    Interval{ 0x1112D, 0x11134 }, // Mn   [8] CHAKMA VOWEL SIGN AI..CHAKMA MAAYYAA
    Interval{ 0x11136, 0x1113F }, // Nd  [10] CHAKMA DIGIT ZERO..CHAKMA DIGIT NINE
    Interval{ 0x11144, 0x11144 }, // Lo       CHAKMA LETTER LHAA
    Interval{ 0x11145, 0x11146 }, // Mc   [2] CHAKMA VOWEL SIGN AA..CHAKMA VOWEL SIGN EI
    Interval{ 0x11147, 0x11147 }, // Lo       CHAKMA LETTER VAA
    Interval{ 0x11150, 0x11172 }, // Lo  [35] MAHAJANI LETTER A..MAHAJANI LETTER RRA
    Interval{ 0x11173, 0x11173 }, // Mn       MAHAJANI SIGN NUKTA
    Interval{ 0x11176, 0x11176 }, // Lo       MAHAJANI LIGATURE SHRI
    Interval{ 0x11180, 0x11181 }, // Mn   [2] SHARADA SIGN CANDRABINDU..SHARADA SIGN ANUSVARA
    Interval{ 0x11182, 0x11182 }, // Mc       SHARADA SIGN VISARGA
    Interval{ 0x11183, 0x111B2 }, // Lo  [48] SHARADA LETTER A..SHARADA LETTER HA
    Interval{ 0x111B3, 0x111B5 }, // Mc   [3] SHARADA VOWEL SIGN AA..SHARADA VOWEL SIGN II
    Interval{ 0x111B6, 0x111BE }, // Mn   [9] SHARADA VOWEL SIGN U..SHARADA VOWEL SIGN O
    Interval{ 0x111BF, 0x111C0 }, // Mc   [2] SHARADA VOWEL SIGN AU..SHARADA SIGN VIRAMA
    Interval{ 0x111C1, 0x111C4 }, // Lo   [4] SHARADA SIGN AVAGRAHA..SHARADA OM
    Interval{ 0x111C9, 0x111CC }, // Mn   [4] SHARADA SANDHI MARK..SHARADA EXTRA SHORT VOWEL MARK
    Interval{ 0x111CE, 0x111CE }, // Mc       SHARADA VOWEL SIGN PRISHTHAMATRA E
    Interval{ 0x111CF, 0x111CF }, // Mn       SHARADA SIGN INVERTED CANDRABINDU
    Interval{ 0x111D0, 0x111D9 }, // Nd  [10] SHARADA DIGIT ZERO..SHARADA DIGIT NINE
    Interval{ 0x111DA, 0x111DA }, // Lo       SHARADA EKAM
    Interval{ 0x111DC, 0x111DC }, // Lo       SHARADA HEADSTROKE
    Interval{ 0x11200, 0x11211 }, // Lo  [18] KHOJKI LETTER A..KHOJKI LETTER JJA
    Interval{ 0x11213, 0x1122B }, // Lo  [25] KHOJKI LETTER NYA..KHOJKI LETTER LLA
    Interval{ 0x1122C, 0x1122E }, // Mc   [3] KHOJKI VOWEL SIGN AA..KHOJKI VOWEL SIGN II
    Interval{ 0x1122F, 0x11231 }, // Mn   [3] KHOJKI VOWEL SIGN U..KHOJKI VOWEL SIGN AI
    Interval{ 0x11232, 0x11233 }, // Mc   [2] KHOJKI VOWEL SIGN O..KHOJKI VOWEL SIGN AU
    Interval{ 0x11234, 0x11234 }, // Mn       KHOJKI SIGN ANUSVARA
    Interval{ 0x11235, 0x11235 }, // Mc       KHOJKI SIGN VIRAMA
    Interval{ 0x11236, 0x11237 }, // Mn   [2] KHOJKI SIGN NUKTA..KHOJKI SIGN SHADDA
    Interval{ 0x1123E, 0x1123E }, // Mn       KHOJKI SIGN SUKUN
    Interval{ 0x1123F, 0x11240 }, // Lo   [2] KHOJKI LETTER QA..KHOJKI LETTER SHORT I
    Interval{ 0x11241, 0x11241 }, // Mn       KHOJKI VOWEL SIGN VOCALIC R
    Interval{ 0x11280, 0x11286 }, // Lo   [7] MULTANI LETTER A..MULTANI LETTER GA
    Interval{ 0x11288, 0x11288 }, // Lo       MULTANI LETTER GHA
    Interval{ 0x1128A, 0x1128D }, // Lo   [4] MULTANI LETTER CA..MULTANI LETTER JJA
    Interval{ 0x1128F, 0x1129D }, // Lo  [15] MULTANI LETTER NYA..MULTANI LETTER BA
    Interval{ 0x1129F, 0x112A8 }, // Lo  [10] MULTANI LETTER BHA..MULTANI LETTER RHA
    Interval{ 0x112B0, 0x112DE }, // Lo  [47] KHUDAWADI LETTER A..KHUDAWADI LETTER HA
    Interval{ 0x112DF, 0x112DF }, // Mn       KHUDAWADI SIGN ANUSVARA
    Interval{ 0x112E0, 0x112E2 }, // Mc   [3] KHUDAWADI VOWEL SIGN AA..KHUDAWADI VOWEL SIGN II
    Interval{ 0x112E3, 0x112EA }, // Mn   [8] KHUDAWADI VOWEL SIGN U..KHUDAWADI SIGN VIRAMA
    Interval{ 0x112F0, 0x112F9 }, // Nd  [10] KHUDAWADI DIGIT ZERO..KHUDAWADI DIGIT NINE
    Interval{ 0x11300, 0x11301 }, // Mn   [2] GRANTHA SIGN COMBINING ANUSVARA ABOVE..GRANTHA SIGN CANDRABINDU
    Interval{ 0x11302, 0x11303 }, // Mc   [2] GRANTHA SIGN ANUSVARA..GRANTHA SIGN VISARGA
    Interval{ 0x11305, 0x1130C }, // Lo   [8] GRANTHA LETTER A..GRANTHA LETTER VOCALIC L
    Interval{ 0x1130F, 0x11310 }, // Lo   [2] GRANTHA LETTER EE..GRANTHA LETTER AI
    Interval{ 0x11313, 0x11328 }, // Lo  [22] GRANTHA LETTER OO..GRANTHA LETTER NA
    Interval{ 0x1132A, 0x11330 }, // Lo   [7] GRANTHA LETTER PA..GRANTHA LETTER RA
    Interval{ 0x11332, 0x11333 }, // Lo   [2] GRANTHA LETTER LA..GRANTHA LETTER LLA
    Interval{ 0x11335, 0x11339 }, // Lo   [5] GRANTHA LETTER VA..GRANTHA LETTER HA
    Interval{ 0x1133B, 0x1133C }, // Mn   [2] COMBINING BINDU BELOW..GRANTHA SIGN NUKTA
    Interval{ 0x1133D, 0x1133D }, // Lo       GRANTHA SIGN AVAGRAHA
    Interval{ 0x1133E, 0x1133F }, // Mc   [2] GRANTHA VOWEL SIGN AA..GRANTHA VOWEL SIGN I
    Interval{ 0x11340, 0x11340 }, // Mn       GRANTHA VOWEL SIGN II
    Interval{ 0x11341, 0x11344 }, // Mc   [4] GRANTHA VOWEL SIGN U..GRANTHA VOWEL SIGN VOCALIC RR
    Interval{ 0x11347, 0x11348 }, // Mc   [2] GRANTHA VOWEL SIGN EE..GRANTHA VOWEL SIGN AI
    Interval{ 0x1134B, 0x1134D }, // Mc   [3] GRANTHA VOWEL SIGN OO..GRANTHA SIGN VIRAMA
    Interval{ 0x11350, 0x11350 }, // Lo       GRANTHA OM
    Interval{ 0x11357, 0x11357 }, // Mc       GRANTHA AU LENGTH MARK
    Interval{ 0x1135D, 0x11361 }, // Lo   [5] GRANTHA SIGN PLUTA..GRANTHA LETTER VOCALIC LL
    Interval{ 0x11362, 0x11363 }, // Mc   [2] GRANTHA VOWEL SIGN VOCALIC L..GRANTHA VOWEL SIGN VOCALIC LL
    Interval{ 0x11366, 0x1136C }, // Mn   [7] COMBINING GRANTHA DIGIT ZERO..COMBINING GRANTHA DIGIT SIX
    Interval{ 0x11370, 0x11374 }, // Mn   [5] COMBINING GRANTHA LETTER A..COMBINING GRANTHA LETTER PA
    Interval{ 0x11380, 0x11389 }, // Lo  [10] TULU-TIGALARI LETTER A..TULU-TIGALARI LETTER VOCALIC LL
    Interval{ 0x1138B, 0x1138B }, // Lo       TULU-TIGALARI LETTER EE
    Interval{ 0x1138E, 0x1138E }, // Lo       TULU-TIGALARI LETTER AI
    Interval{ 0x11390, 0x113B5 }, // Lo  [38] TULU-TIGALARI LETTER OO..TULU-TIGALARI LETTER LLLA
    Interval{ 0x113B7, 0x113B7 }, // Lo       TULU-TIGALARI SIGN AVAGRAHA
    Interval{ 0x113B8, 0x113BA }, // Mc   [3] TULU-TIGALARI VOWEL SIGN AA..TULU-TIGALARI VOWEL SIGN II
    Interval{ 0x113BB, 0x113C0 }, // Mn   [6] TULU-TIGALARI VOWEL SIGN U..TULU-TIGALARI VOWEL SIGN VOCALIC LL
    Interval{ 0x113C2, 0x113C2 }, // Mc       TULU-TIGALARI VOWEL SIGN EE
    Interval{ 0x113C5, 0x113C5 }, // Mc       TULU-TIGALARI VOWEL SIGN AI
    Interval{ 0x113C7, 0x113CA }, // Mc   [4] TULU-TIGALARI VOWEL SIGN OO..TULU-TIGALARI SIGN CANDRA ANUNASIKA
    Interval{ 0x113CC, 0x113CD }, // Mc   [2] TULU-TIGALARI SIGN ANUSVARA..TULU-TIGALARI SIGN VISARGA
    Interval{ 0x113CE, 0x113CE }, // Mn       TULU-TIGALARI SIGN VIRAMA
    Interval{ 0x113CF, 0x113CF }, // Mc       TULU-TIGALARI SIGN LOOPED VIRAMA
    Interval{ 0x113D0, 0x113D0 }, // Mn       TULU-TIGALARI CONJOINER
    Interval{ 0x113D1, 0x113D1 }, // Lo       TULU-TIGALARI REPHA
    Interval{ 0x113D2, 0x113D2 }, // Mn       TULU-TIGALARI GEMINATION MARK
    Interval{ 0x113D3, 0x113D3 }, // Lo       TULU-TIGALARI SIGN PLUTA
    Interval{ 0x113E1, 0x113E2 }, // Mn   [2] TULU-TIGALARI VEDIC TONE SVARITA..TULU-TIGALARI VEDIC TONE ANUDATTA
    Interval{ 0x11400, 0x11434 }, // Lo  [53] NEWA LETTER A..NEWA LETTER HA
    Interval{ 0x11435, 0x11437 }, // Mc   [3] NEWA VOWEL SIGN AA..NEWA VOWEL SIGN II
    Interval{ 0x11438, 0x1143F }, // Mn   [8] NEWA VOWEL SIGN U..NEWA VOWEL SIGN AI
    Interval{ 0x11440, 0x11441 }, // Mc   [2] NEWA VOWEL SIGN O..NEWA VOWEL SIGN AU
    Interval{ 0x11442, 0x11444 }, // Mn   [3] NEWA SIGN VIRAMA..NEWA SIGN ANUSVARA
    Interval{ 0x11445, 0x11445 }, // Mc       NEWA SIGN VISARGA
    Interval{ 0x11446, 0x11446 }, // Mn       NEWA SIGN NUKTA
    Interval{ 0x11447, 0x1144A }, // Lo   [4] NEWA SIGN AVAGRAHA..NEWA SIDDHI
    Interval{ 0x11450, 0x11459 }, // Nd  [10] NEWA DIGIT ZERO..NEWA DIGIT NINE
    Interval{ 0x1145E, 0x1145E }, // Mn       NEWA SANDHI MARK
    Interval{ 0x1145F, 0x11461 }, // Lo   [3] NEWA LETTER VEDIC ANUSVARA..NEWA SIGN UPADHMANIYA
    Interval{ 0x11480, 0x114AF }, // Lo  [48] TIRHUTA ANJI..TIRHUTA LETTER HA
    Interval{ 0x114B0, 0x114B2 }, // Mc   [3] TIRHUTA VOWEL SIGN AA..TIRHUTA VOWEL SIGN II
    Interval{ 0x114B3, 0x114B8 }, // Mn   [6] TIRHUTA VOWEL SIGN U..TIRHUTA VOWEL SIGN VOCALIC LL
    Interval{ 0x114B9, 0x114B9 }, // Mc       TIRHUTA VOWEL SIGN E
    Interval{ 0x114BA, 0x114BA }, // Mn       TIRHUTA VOWEL SIGN SHORT E
    Interval{ 0x114BB, 0x114BE }, // Mc   [4] TIRHUTA VOWEL SIGN AI..TIRHUTA VOWEL SIGN AU
    Interval{ 0x114BF, 0x114C0 }, // Mn   [2] TIRHUTA SIGN CANDRABINDU..TIRHUTA SIGN ANUSVARA
    Interval{ 0x114C1, 0x114C1 }, // Mc       TIRHUTA SIGN VISARGA
    Interval{ 0x114C2, 0x114C3 }, // Mn   [2] TIRHUTA SIGN VIRAMA..TIRHUTA SIGN NUKTA
    Interval{ 0x114C4, 0x114C5 }, // Lo   [2] TIRHUTA SIGN AVAGRAHA..TIRHUTA GVANG
    Interval{ 0x114C7, 0x114C7 }, // Lo       TIRHUTA OM
    Interval{ 0x114D0, 0x114D9 }, // Nd  [10] TIRHUTA DIGIT ZERO..TIRHUTA DIGIT NINE
    Interval{ 0x11580, 0x115AE }, // Lo  [47] SIDDHAM LETTER A..SIDDHAM LETTER HA
    Interval{ 0x115AF, 0x115B1 }, // Mc   [3] SIDDHAM VOWEL SIGN AA..SIDDHAM VOWEL SIGN II
    Interval{ 0x115B2, 0x115B5 }, // Mn   [4] SIDDHAM VOWEL SIGN U..SIDDHAM VOWEL SIGN VOCALIC RR
    Interval{ 0x115B8, 0x115BB }, // Mc   [4] SIDDHAM VOWEL SIGN E..SIDDHAM VOWEL SIGN AU
    Interval{ 0x115BC, 0x115BD }, // Mn   [2] SIDDHAM SIGN CANDRABINDU..SIDDHAM SIGN ANUSVARA
    Interval{ 0x115BE, 0x115BE }, // Mc       SIDDHAM SIGN VISARGA
    Interval{ 0x115BF, 0x115C0 }, // Mn   [2] SIDDHAM SIGN VIRAMA..SIDDHAM SIGN NUKTA
    Interval{ 0x115D8, 0x115DB }, // Lo   [4] SIDDHAM LETTER THREE-CIRCLE ALTERNATE I..SIDDHAM LETTER ALTERNATE U
    Interval{ 0x115DC, 0x115DD }, // Mn   [2] SIDDHAM VOWEL SIGN ALTERNATE U..SIDDHAM VOWEL SIGN ALTERNATE UU
    Interval{ 0x11600, 0x1162F }, // Lo  [48] MODI LETTER A..MODI LETTER LLA
    Interval{ 0x11630, 0x11632 }, // Mc   [3] MODI VOWEL SIGN AA..MODI VOWEL SIGN II
    Interval{ 0x11633, 0x1163A }, // Mn   [8] MODI VOWEL SIGN U..MODI VOWEL SIGN AI
    Interval{ 0x1163B, 0x1163C }, // Mc   [2] MODI VOWEL SIGN O..MODI VOWEL SIGN AU
    Interval{ 0x1163D, 0x1163D }, // Mn       MODI SIGN ANUSVARA
    Interval{ 0x1163E, 0x1163E }, // Mc       MODI SIGN VISARGA
    Interval{ 0x1163F, 0x11640 }, // Mn   [2] MODI SIGN VIRAMA..MODI SIGN ARDHACANDRA
    Interval{ 0x11644, 0x11644 }, // Lo       MODI SIGN HUVA
    Interval{ 0x11650, 0x11659 }, // Nd  [10] MODI DIGIT ZERO..MODI DIGIT NINE
    Interval{ 0x11680, 0x116AA }, // Lo  [43] TAKRI LETTER A..TAKRI LETTER RRA
    Interval{ 0x116AB, 0x116AB }, // Mn       TAKRI SIGN ANUSVARA
    Interval{ 0x116AC, 0x116AC }, // Mc       TAKRI SIGN VISARGA
    Interval{ 0x116AD, 0x116AD }, // Mn       TAKRI VOWEL SIGN AA
    Interval{ 0x116AE, 0x116AF }, // Mc   [2] TAKRI VOWEL SIGN I..TAKRI VOWEL SIGN II
    Interval{ 0x116B0, 0x116B5 }, // Mn   [6] TAKRI VOWEL SIGN U..TAKRI VOWEL SIGN AU
    Interval{ 0x116B6, 0x116B6 }, // Mc       TAKRI SIGN VIRAMA
    Interval{ 0x116B7, 0x116B7 }, // Mn       TAKRI SIGN NUKTA
    Interval{ 0x116B8, 0x116B8 }, // Lo       TAKRI LETTER ARCHAIC KHA
    Interval{ 0x116C0, 0x116C9 }, // Nd  [10] TAKRI DIGIT ZERO..TAKRI DIGIT NINE
    Interval{ 0x116D0, 0x116E3 }, // Nd  [20] MYANMAR PAO DIGIT ZERO..MYANMAR EASTERN PWO KAREN DIGIT NINE
    Interval{ 0x11700, 0x1171A }, // Lo  [27] AHOM LETTER KA..AHOM LETTER ALTERNATE BA
    Interval{ 0x1171D, 0x1171D }, // Mn       AHOM CONSONANT SIGN MEDIAL LA
    Interval{ 0x1171E, 0x1171E }, // Mc       AHOM CONSONANT SIGN MEDIAL RA
    Interval{ 0x1171F, 0x1171F }, // Mn       AHOM CONSONANT SIGN MEDIAL LIGATING RA
    Interval{ 0x11720, 0x11721 }, // Mc   [2] AHOM VOWEL SIGN A..AHOM VOWEL SIGN AA
    Interval{ 0x11722, 0x11725 }, // Mn   [4] AHOM VOWEL SIGN I..AHOM VOWEL SIGN UU
    Interval{ 0x11726, 0x11726 }, // Mc       AHOM VOWEL SIGN E
    Interval{ 0x11727, 0x1172B }, // Mn   [5] AHOM VOWEL SIGN AW..AHOM SIGN KILLER
    Interval{ 0x11730, 0x11739 }, // Nd  [10] AHOM DIGIT ZERO..AHOM DIGIT NINE
    Interval{ 0x11740, 0x11746 }, // Lo   [7] AHOM LETTER CA..AHOM LETTER LLA
    Interval{ 0x11800, 0x1182B }, // Lo  [44] DOGRA LETTER A..DOGRA LETTER RRA
    Interval{ 0x1182C, 0x1182E }, // Mc   [3] DOGRA VOWEL SIGN AA..DOGRA VOWEL SIGN II
    Interval{ 0x1182F, 0x11837 }, // Mn   [9] DOGRA VOWEL SIGN U..DOGRA SIGN ANUSVARA
    Interval{ 0x11838, 0x11838 }, // Mc       DOGRA SIGN VISARGA
    Interval{ 0x11839, 0x1183A }, // Mn   [2] DOGRA SIGN VIRAMA..DOGRA SIGN NUKTA
    Interval{ 0x118A0, 0x118DF }, // L&  [64] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Interval{ 0x118E0, 0x118E9 }, // Nd  [10] WARANG CITI DIGIT ZERO..WARANG CITI DIGIT NINE
    Interval{ 0x118FF, 0x11906 }, // Lo   [8] WARANG CITI OM..DIVES AKURU LETTER E
    Interval{ 0x11909, 0x11909 }, // Lo       DIVES AKURU LETTER O
    Interval{ 0x1190C, 0x11913 }, // Lo   [8] DIVES AKURU LETTER KA..DIVES AKURU LETTER JA
    Interval{ 0x11915, 0x11916 }, // Lo   [2] DIVES AKURU LETTER NYA..DIVES AKURU LETTER TTA
    Interval{ 0x11918, 0x1192F }, // Lo  [24] DIVES AKURU LETTER DDA..DIVES AKURU LETTER ZA
    Interval{ 0x11930, 0x11935 }, // Mc   [6] DIVES AKURU VOWEL SIGN AA..DIVES AKURU VOWEL SIGN E
    Interval{ 0x11937, 0x11938 }, // Mc   [2] DIVES AKURU VOWEL SIGN AI..DIVES AKURU VOWEL SIGN O
    Interval{ 0x1193B, 0x1193C }, // Mn   [2] DIVES AKURU SIGN ANUSVARA..DIVES AKURU SIGN CANDRABINDU
    Interval{ 0x1193D, 0x1193D }, // Mc       DIVES AKURU SIGN HALANTA
    Interval{ 0x1193E, 0x1193E }, // Mn       DIVES AKURU VIRAMA
    Interval{ 0x1193F, 0x1193F }, // Lo       DIVES AKURU PREFIXED NASAL SIGN
    Interval{ 0x11940, 0x11940 }, // Mc       DIVES AKURU MEDIAL YA
    Interval{ 0x11941, 0x11941 }, // Lo       DIVES AKURU INITIAL RA
    Interval{ 0x11942, 0x11942 }, // Mc       DIVES AKURU MEDIAL RA
    Interval{ 0x11943, 0x11943 }, // Mn       DIVES AKURU SIGN NUKTA
    Interval{ 0x11950, 0x11959 }, // Nd  [10] DIVES AKURU DIGIT ZERO..DIVES AKURU DIGIT NINE
    Interval{ 0x119A0, 0x119A7 }, // Lo   [8] NANDINAGARI LETTER A..NANDINAGARI LETTER VOCALIC RR
    Interval{ 0x119AA, 0x119D0 }, // Lo  [39] NANDINAGARI LETTER E..NANDINAGARI LETTER RRA
    Interval{ 0x119D1, 0x119D3 }, // Mc   [3] NANDINAGARI VOWEL SIGN AA..NANDINAGARI VOWEL SIGN II
    Interval{ 0x119D4, 0x119D7 }, // Mn   [4] NANDINAGARI VOWEL SIGN U..NANDINAGARI VOWEL SIGN VOCALIC RR
    Interval{ 0x119DA, 0x119DB }, // Mn   [2] NANDINAGARI VOWEL SIGN E..NANDINAGARI VOWEL SIGN AI
    Interval{ 0x119DC, 0x119DF }, // Mc   [4] NANDINAGARI VOWEL SIGN O..NANDINAGARI SIGN VISARGA
    Interval{ 0x119E0, 0x119E0 }, // Mn       NANDINAGARI SIGN VIRAMA
    Interval{ 0x119E1, 0x119E1 }, // Lo       NANDINAGARI SIGN AVAGRAHA
    Interval{ 0x119E3, 0x119E3 }, // Lo       NANDINAGARI HEADSTROKE
    Interval{ 0x119E4, 0x119E4 }, // Mc       NANDINAGARI VOWEL SIGN PRISHTHAMATRA E
    Interval{ 0x11A00, 0x11A00 }, // Lo       ZANABAZAR SQUARE LETTER A
    Interval{ 0x11A01, 0x11A0A }, // Mn  [10] ZANABAZAR SQUARE VOWEL SIGN I..ZANABAZAR SQUARE VOWEL LENGTH MARK
    Interval{ 0x11A0B, 0x11A32 }, // Lo  [40] ZANABAZAR SQUARE LETTER KA..ZANABAZAR SQUARE LETTER KSSA
    Interval{ 0x11A33, 0x11A38 }, // Mn   [6] ZANABAZAR SQUARE FINAL CONSONANT MARK..ZANABAZAR SQUARE SIGN ANUSVARA
    Interval{ 0x11A39, 0x11A39 }, // Mc       ZANABAZAR SQUARE SIGN VISARGA
    Interval{ 0x11A3A, 0x11A3A }, // Lo       ZANABAZAR SQUARE CLUSTER-INITIAL LETTER RA
    Interval{ 0x11A3B, 0x11A3E }, // Mn   [4] ZANABAZAR SQUARE CLUSTER-FINAL LETTER YA..ZANABAZAR SQUARE CLUSTER-FINAL LETTER VA
    Interval{ 0x11A47, 0x11A47 }, // Mn       ZANABAZAR SQUARE SUBJOINER
    Interval{ 0x11A50, 0x11A50 }, // Lo       SOYOMBO LETTER A
    Interval{ 0x11A51, 0x11A56 }, // Mn   [6] SOYOMBO VOWEL SIGN I..SOYOMBO VOWEL SIGN OE
    Interval{ 0x11A57, 0x11A58 }, // Mc   [2] SOYOMBO VOWEL SIGN AI..SOYOMBO VOWEL SIGN AU
    Interval{ 0x11A59, 0x11A5B }, // Mn   [3] SOYOMBO VOWEL SIGN VOCALIC R..SOYOMBO VOWEL LENGTH MARK
    Interval{ 0x11A5C, 0x11A89 }, // Lo  [46] SOYOMBO LETTER KA..SOYOMBO CLUSTER-INITIAL LETTER SA
    Interval{ 0x11A8A, 0x11A96 }, // Mn  [13] SOYOMBO FINAL CONSONANT SIGN G..SOYOMBO SIGN ANUSVARA
    Interval{ 0x11A97, 0x11A97 }, // Mc       SOYOMBO SIGN VISARGA
    Interval{ 0x11A98, 0x11A99 }, // Mn   [2] SOYOMBO GEMINATION MARK..SOYOMBO SUBJOINER
    Interval{ 0x11A9D, 0x11A9D }, // Lo       SOYOMBO MARK PLUTA
    Interval{ 0x11AB0, 0x11AF8 }, // Lo  [73] CANADIAN SYLLABICS NATTILIK HI..PAU CIN HAU GLOTTAL STOP FINAL
    Interval{ 0x11B60, 0x11B60 }, // Mn       SHARADA VOWEL SIGN OE
    Interval{ 0x11B61, 0x11B61 }, // Mc       SHARADA VOWEL SIGN OOE
    Interval{ 0x11B62, 0x11B64 }, // Mn   [3] SHARADA VOWEL SIGN UE..SHARADA VOWEL SIGN SHORT E
    Interval{ 0x11B65, 0x11B65 }, // Mc       SHARADA VOWEL SIGN SHORT O
    Interval{ 0x11B66, 0x11B66 }, // Mn       SHARADA VOWEL SIGN CANDRA E
    Interval{ 0x11B67, 0x11B67 }, // Mc       SHARADA VOWEL SIGN CANDRA O
    Interval{ 0x11BC0, 0x11BE0 }, // Lo  [33] SUNUWAR LETTER DEVI..SUNUWAR LETTER KLOKO
    Interval{ 0x11BF0, 0x11BF9 }, // Nd  [10] SUNUWAR DIGIT ZERO..SUNUWAR DIGIT NINE
    Interval{ 0x11C00, 0x11C08 }, // Lo   [9] BHAIKSUKI LETTER A..BHAIKSUKI LETTER VOCALIC L
    Interval{ 0x11C0A, 0x11C2E }, // Lo  [37] BHAIKSUKI LETTER E..BHAIKSUKI LETTER HA
    Interval{ 0x11C2F, 0x11C2F }, // Mc       BHAIKSUKI VOWEL SIGN AA
    Interval{ 0x11C30, 0x11C36 }, // Mn   [7] BHAIKSUKI VOWEL SIGN I..BHAIKSUKI VOWEL SIGN VOCALIC L
    Interval{ 0x11C38, 0x11C3D }, // Mn   [6] BHAIKSUKI VOWEL SIGN E..BHAIKSUKI SIGN ANUSVARA
    Interval{ 0x11C3E, 0x11C3E }, // Mc       BHAIKSUKI SIGN VISARGA
    Interval{ 0x11C3F, 0x11C3F }, // Mn       BHAIKSUKI SIGN VIRAMA
    Interval{ 0x11C40, 0x11C40 }, // Lo       BHAIKSUKI SIGN AVAGRAHA
    Interval{ 0x11C50, 0x11C59 }, // Nd  [10] BHAIKSUKI DIGIT ZERO..BHAIKSUKI DIGIT NINE
    Interval{ 0x11C72, 0x11C8F }, // Lo  [30] MARCHEN LETTER KA..MARCHEN LETTER A
    Interval{ 0x11C92, 0x11CA7 }, // Mn  [22] MARCHEN SUBJOINED LETTER KA..MARCHEN SUBJOINED LETTER ZA
    Interval{ 0x11CA9, 0x11CA9 }, // Mc       MARCHEN SUBJOINED LETTER YA
    Interval{ 0x11CAA, 0x11CB0 }, // Mn   [7] MARCHEN SUBJOINED LETTER RA..MARCHEN VOWEL SIGN AA
    Interval{ 0x11CB1, 0x11CB1 }, // Mc       MARCHEN VOWEL SIGN I
    Interval{ 0x11CB2, 0x11CB3 }, // Mn   [2] MARCHEN VOWEL SIGN U..MARCHEN VOWEL SIGN E
    Interval{ 0x11CB4, 0x11CB4 }, // Mc       MARCHEN VOWEL SIGN O
    Interval{ 0x11CB5, 0x11CB6 }, // Mn   [2] MARCHEN SIGN ANUSVARA..MARCHEN SIGN CANDRABINDU
    Interval{ 0x11D00, 0x11D06 }, // Lo   [7] MASARAM GONDI LETTER A..MASARAM GONDI LETTER E
    Interval{ 0x11D08, 0x11D09 }, // Lo   [2] MASARAM GONDI LETTER AI..MASARAM GONDI LETTER O
    Interval{ 0x11D0B, 0x11D30 }, // Lo  [38] MASARAM GONDI LETTER AU..MASARAM GONDI LETTER TRA
    Interval{ 0x11D31, 0x11D36 }, // Mn   [6] MASARAM GONDI VOWEL SIGN AA..MASARAM GONDI VOWEL SIGN VOCALIC R
    Interval{ 0x11D3A, 0x11D3A }, // Mn       MASARAM GONDI VOWEL SIGN E
    Interval{ 0x11D3C, 0x11D3D }, // Mn   [2] MASARAM GONDI VOWEL SIGN AI..MASARAM GONDI VOWEL SIGN O
    Interval{ 0x11D3F, 0x11D45 }, // Mn   [7] MASARAM GONDI VOWEL SIGN AU..MASARAM GONDI VIRAMA
    Interval{ 0x11D46, 0x11D46 }, // Lo       MASARAM GONDI REPHA
    Interval{ 0x11D47, 0x11D47 }, // Mn       MASARAM GONDI RA-KARA
    Interval{ 0x11D50, 0x11D59 }, // Nd  [10] MASARAM GONDI DIGIT ZERO..MASARAM GONDI DIGIT NINE
    Interval{ 0x11D60, 0x11D65 }, // Lo   [6] GUNJALA GONDI LETTER A..GUNJALA GONDI LETTER UU
    Interval{ 0x11D67, 0x11D68 }, // Lo   [2] GUNJALA GONDI LETTER EE..GUNJALA GONDI LETTER AI
    Interval{ 0x11D6A, 0x11D89 }, // Lo  [32] GUNJALA GONDI LETTER OO..GUNJALA GONDI LETTER SA
    Interval{ 0x11D8A, 0x11D8E }, // Mc   [5] GUNJALA GONDI VOWEL SIGN AA..GUNJALA GONDI VOWEL SIGN UU
    Interval{ 0x11D90, 0x11D91 }, // Mn   [2] GUNJALA GONDI VOWEL SIGN EE..GUNJALA GONDI VOWEL SIGN AI
    Interval{ 0x11D93, 0x11D94 }, // Mc   [2] GUNJALA GONDI VOWEL SIGN OO..GUNJALA GONDI VOWEL SIGN AU
    Interval{ 0x11D95, 0x11D95 }, // Mn       GUNJALA GONDI SIGN ANUSVARA
    Interval{ 0x11D96, 0x11D96 }, // Mc       GUNJALA GONDI SIGN VISARGA
    Interval{ 0x11D97, 0x11D97 }, // Mn       GUNJALA GONDI VIRAMA
    Interval{ 0x11D98, 0x11D98 }, // Lo       GUNJALA GONDI OM
    Interval{ 0x11DA0, 0x11DA9 }, // Nd  [10] GUNJALA GONDI DIGIT ZERO..GUNJALA GONDI DIGIT NINE
    Interval{ 0x11DB0, 0x11DD8 }, // Lo  [41] TOLONG SIKI LETTER I..TOLONG SIKI LETTER RRH
    Interval{ 0x11DD9, 0x11DD9 }, // Lm       TOLONG SIKI SIGN SELA
    Interval{ 0x11DDA, 0x11DDB }, // Lo   [2] TOLONG SIKI SIGN HECAKA..TOLONG SIKI UNGGA
    Interval{ 0x11DE0, 0x11DE9 }, // Nd  [10] TOLONG SIKI DIGIT ZERO..TOLONG SIKI DIGIT NINE
    Interval{ 0x11EE0, 0x11EF2 }, // Lo  [19] MAKASAR LETTER KA..MAKASAR ANGKA
    Interval{ 0x11EF3, 0x11EF4 }, // Mn   [2] MAKASAR VOWEL SIGN I..MAKASAR VOWEL SIGN U
    Interval{ 0x11EF5, 0x11EF6 }, // Mc   [2] MAKASAR VOWEL SIGN E..MAKASAR VOWEL SIGN O
    Interval{ 0x11F00, 0x11F01 }, // Mn   [2] KAWI SIGN CANDRABINDU..KAWI SIGN ANUSVARA
    Interval{ 0x11F02, 0x11F02 }, // Lo       KAWI SIGN REPHA
    Interval{ 0x11F03, 0x11F03 }, // Mc       KAWI SIGN VISARGA
    Interval{ 0x11F04, 0x11F10 }, // Lo  [13] KAWI LETTER A..KAWI LETTER O
    Interval{ 0x11F12, 0x11F33 }, // Lo  [34] KAWI LETTER KA..KAWI LETTER JNYA
    Interval{ 0x11F34, 0x11F35 }, // Mc   [2] KAWI VOWEL SIGN AA..KAWI VOWEL SIGN ALTERNATE AA
    Interval{ 0x11F36, 0x11F3A }, // Mn   [5] KAWI VOWEL SIGN I..KAWI VOWEL SIGN VOCALIC R
    Interval{ 0x11F3E, 0x11F3F }, // Mc   [2] KAWI VOWEL SIGN E..KAWI VOWEL SIGN AI
    Interval{ 0x11F40, 0x11F40 }, // Mn       KAWI VOWEL SIGN EU
    Interval{ 0x11F41, 0x11F41 }, // Mc       KAWI SIGN KILLER
    Interval{ 0x11F42, 0x11F42 }, // Mn       KAWI CONJOINER
    Interval{ 0x11F50, 0x11F59 }, // Nd  [10] KAWI DIGIT ZERO..KAWI DIGIT NINE
    Interval{ 0x11F5A, 0x11F5A }, // Mn       KAWI SIGN NUKTA
    Interval{ 0x11FB0, 0x11FB0 }, // Lo       LISU LETTER YHA
    Interval{ 0x12000, 0x12399 }, // Lo [922] CUNEIFORM SIGN A..CUNEIFORM SIGN U U
    Interval{ 0x12400, 0x1246E }, // Nl [111] CUNEIFORM NUMERIC SIGN TWO ASH..CUNEIFORM NUMERIC SIGN NINE U VARIANT FORM
    Interval{ 0x12480, 0x12543 }, // Lo [196] CUNEIFORM SIGN AB TIMES NUN TENU..CUNEIFORM SIGN ZU5 TIMES THREE DISH TENU
    Interval{ 0x12F90, 0x12FF0 }, // Lo  [97] CYPRO-MINOAN SIGN CM001..CYPRO-MINOAN SIGN CM114
    Interval{ 0x13000, 0x1342F }, // Lo [1072] EGYPTIAN HIEROGLYPH A001..EGYPTIAN HIEROGLYPH V011D
    Interval{ 0x13440, 0x13440 }, // Mn       EGYPTIAN HIEROGLYPH MIRROR HORIZONTALLY
    Interval{ 0x13441, 0x13446 }, // Lo   [6] EGYPTIAN HIEROGLYPH FULL BLANK..EGYPTIAN HIEROGLYPH WIDE LOST SIGN
    Interval{ 0x13447, 0x13455 }, // Mn  [15] EGYPTIAN HIEROGLYPH MODIFIER DAMAGED AT TOP START..EGYPTIAN HIEROGLYPH MODIFIER DAMAGED
    Interval{ 0x13460, 0x143FA }, // Lo [3995] EGYPTIAN HIEROGLYPH-13460..EGYPTIAN HIEROGLYPH-143FA
    Interval{ 0x14400, 0x14646 }, // Lo [583] ANATOLIAN HIEROGLYPH A001..ANATOLIAN HIEROGLYPH A530
    Interval{ 0x16100, 0x1611D }, // Lo  [30] GURUNG KHEMA LETTER A..GURUNG KHEMA LETTER SA
    Interval{ 0x1611E, 0x16129 }, // Mn  [12] GURUNG KHEMA VOWEL SIGN AA..GURUNG KHEMA VOWEL LENGTH MARK
    Interval{ 0x1612A, 0x1612C }, // Mc   [3] GURUNG KHEMA CONSONANT SIGN MEDIAL YA..GURUNG KHEMA CONSONANT SIGN MEDIAL HA
    Interval{ 0x1612D, 0x1612F }, // Mn   [3] GURUNG KHEMA SIGN ANUSVARA..GURUNG KHEMA SIGN THOLHOMA
    Interval{ 0x16130, 0x16139 }, // Nd  [10] GURUNG KHEMA DIGIT ZERO..GURUNG KHEMA DIGIT NINE
    Interval{ 0x16800, 0x16A38 }, // Lo [569] BAMUM LETTER PHASE-A NGKUE MFON..BAMUM LETTER PHASE-F VUEQ
    Interval{ 0x16A40, 0x16A5E }, // Lo  [31] MRO LETTER TA..MRO LETTER TEK
    Interval{ 0x16A60, 0x16A69 }, // Nd  [10] MRO DIGIT ZERO..MRO DIGIT NINE
    Interval{ 0x16A70, 0x16ABE }, // Lo  [79] TANGSA LETTER OZ..TANGSA LETTER ZA
    Interval{ 0x16AC0, 0x16AC9 }, // Nd  [10] TANGSA DIGIT ZERO..TANGSA DIGIT NINE
    Interval{ 0x16AD0, 0x16AED }, // Lo  [30] BASSA VAH LETTER ENNI..BASSA VAH LETTER I
    Interval{ 0x16AF0, 0x16AF4 }, // Mn   [5] BASSA VAH COMBINING HIGH TONE..BASSA VAH COMBINING HIGH-LOW TONE
    Interval{ 0x16B00, 0x16B2F }, // Lo  [48] PAHAWH HMONG VOWEL KEEB..PAHAWH HMONG CONSONANT CAU
    Interval{ 0x16B30, 0x16B36 }, // Mn   [7] PAHAWH HMONG MARK CIM TUB..PAHAWH HMONG MARK CIM TAUM
    Interval{ 0x16B40, 0x16B43 }, // Lm   [4] PAHAWH HMONG SIGN VOS SEEV..PAHAWH HMONG SIGN IB YAM
    Interval{ 0x16B50, 0x16B59 }, // Nd  [10] PAHAWH HMONG DIGIT ZERO..PAHAWH HMONG DIGIT NINE
    Interval{ 0x16B63, 0x16B77 }, // Lo  [21] PAHAWH HMONG SIGN VOS LUB..PAHAWH HMONG SIGN CIM NRES TOS
    Interval{ 0x16B7D, 0x16B8F }, // Lo  [19] PAHAWH HMONG CLAN SIGN TSHEEJ..PAHAWH HMONG CLAN SIGN VWJ
    Interval{ 0x16D40, 0x16D42 }, // Lm   [3] KIRAT RAI SIGN ANUSVARA..KIRAT RAI SIGN VISARGA
    Interval{ 0x16D43, 0x16D6A }, // Lo  [40] KIRAT RAI LETTER A..KIRAT RAI VOWEL SIGN AU
    Interval{ 0x16D6B, 0x16D6C }, // Lm   [2] KIRAT RAI SIGN VIRAMA..KIRAT RAI SIGN SAAT
    Interval{ 0x16D70, 0x16D79 }, // Nd  [10] KIRAT RAI DIGIT ZERO..KIRAT RAI DIGIT NINE
    Interval{ 0x16E40, 0x16E7F }, // L&  [64] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Interval{ 0x16EA0, 0x16EB8 }, // L&  [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Interval{ 0x16EBB, 0x16ED3 }, // L&  [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Interval{ 0x16F00, 0x16F4A }, // Lo  [75] MIAO LETTER PA..MIAO LETTER RTE
    Interval{ 0x16F4F, 0x16F4F }, // Mn       MIAO SIGN CONSONANT MODIFIER BAR
    Interval{ 0x16F50, 0x16F50 }, // Lo       MIAO LETTER NASALIZATION
    Interval{ 0x16F51, 0x16F87 }, // Mc  [55] MIAO SIGN ASPIRATION..MIAO VOWEL SIGN UI
    Interval{ 0x16F8F, 0x16F92 }, // Mn   [4] MIAO TONE RIGHT..MIAO TONE BELOW
    Interval{ 0x16F93, 0x16F9F }, // Lm  [13] MIAO LETTER TONE-2..MIAO LETTER REFORMED TONE-8
    Interval{ 0x16FE0, 0x16FE1 }, // Lm   [2] TANGUT ITERATION MARK..NUSHU ITERATION MARK
    Interval{ 0x16FE3, 0x16FE3 }, // Lm       OLD CHINESE ITERATION MARK
    Interval{ 0x16FE4, 0x16FE4 }, // Mn       KHITAN SMALL SCRIPT FILLER
    Interval{ 0x16FF0, 0x16FF1 }, // Mc   [2] VIETNAMESE ALTERNATE READING MARK CA..VIETNAMESE ALTERNATE READING MARK NHAY
    Interval{ 0x16FF2, 0x16FF3 }, // Lm   [2] CHINESE SMALL SIMPLIFIED ER..CHINESE SMALL TRADITIONAL ER
    Interval{ 0x16FF4, 0x16FF6 }, // Nl   [3] YANGQIN SIGN SLOW ONE BEAT..YANGQIN SIGN SLOW TWO BEATS
    Interval{ 0x17000, 0x18CD5 }, // Lo [7382] TANGUT IDEOGRAPH-17000..KHITAN SMALL SCRIPT CHARACTER-18CD5
    Interval{ 0x18CFF, 0x18D1E }, // Lo  [32] KHITAN SMALL SCRIPT CHARACTER-18CFF..TANGUT IDEOGRAPH-18D1E
    Interval{ 0x18D80, 0x18DF2 }, // Lo [115] TANGUT COMPONENT-769..TANGUT COMPONENT-883
    Interval{ 0x1AFF0, 0x1AFF3 }, // Lm   [4] KATAKANA LETTER MINNAN TONE-2..KATAKANA LETTER MINNAN TONE-5
    Interval{ 0x1AFF5, 0x1AFFB }, // Lm   [7] KATAKANA LETTER MINNAN TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-5
    Interval{ 0x1AFFD, 0x1AFFE }, // Lm   [2] KATAKANA LETTER MINNAN NASALIZED TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-8
    Interval{ 0x1B000, 0x1B122 }, // Lo [291] KATAKANA LETTER ARCHAIC E..KATAKANA LETTER ARCHAIC WU
    Interval{ 0x1B132, 0x1B132 }, // Lo       HIRAGANA LETTER SMALL KO
    Interval{ 0x1B150, 0x1B152 }, // Lo   [3] HIRAGANA LETTER SMALL WI..HIRAGANA LETTER SMALL WO
    Interval{ 0x1B155, 0x1B155 }, // Lo       KATAKANA LETTER SMALL KO
    Interval{ 0x1B164, 0x1B167 }, // Lo   [4] KATAKANA LETTER SMALL WI..KATAKANA LETTER SMALL N
    Interval{ 0x1B170, 0x1B2FB }, // Lo [396] NUSHU CHARACTER-1B170..NUSHU CHARACTER-1B2FB
    Interval{ 0x1BC00, 0x1BC6A }, // Lo [107] DUPLOYAN LETTER H..DUPLOYAN LETTER VOCALIC M
    Interval{ 0x1BC70, 0x1BC7C }, // Lo  [13] DUPLOYAN AFFIX LEFT HORIZONTAL SECANT..DUPLOYAN AFFIX ATTACHED TANGENT HOOK
    Interval{ 0x1BC80, 0x1BC88 }, // Lo   [9] DUPLOYAN AFFIX HIGH ACUTE..DUPLOYAN AFFIX HIGH VERTICAL
    Interval{ 0x1BC90, 0x1BC99 }, // Lo  [10] DUPLOYAN AFFIX LOW ACUTE..DUPLOYAN AFFIX LOW ARROW
    Interval{ 0x1BC9D, 0x1BC9E }, // Mn   [2] DUPLOYAN THICK LETTER SELECTOR..DUPLOYAN DOUBLE MARK
    Interval{ 0x1CCF0, 0x1CCF9 }, // Nd  [10] OUTLINED DIGIT ZERO..OUTLINED DIGIT NINE
    Interval{ 0x1CF00, 0x1CF2D }, // Mn  [46] ZNAMENNY COMBINING MARK GORAZDO NIZKO S KRYZHEM ON LEFT..ZNAMENNY COMBINING MARK KRYZH ON LEFT
    Interval{ 0x1CF30, 0x1CF46 }, // Mn  [23] ZNAMENNY COMBINING TONAL RANGE MARK MRACHNO..ZNAMENNY PRIZNAK MODIFIER ROG
    Interval{ 0x1D165, 0x1D166 }, // Mc   [2] MUSICAL SYMBOL COMBINING STEM..MUSICAL SYMBOL COMBINING SPRECHGESANG STEM
    Interval{ 0x1D167, 0x1D169 }, // Mn   [3] MUSICAL SYMBOL COMBINING TREMOLO-1..MUSICAL SYMBOL COMBINING TREMOLO-3
    Interval{ 0x1D16D, 0x1D172 }, // Mc   [6] MUSICAL SYMBOL COMBINING AUGMENTATION DOT..MUSICAL SYMBOL COMBINING FLAG-5
    Interval{ 0x1D17B, 0x1D182 }, // Mn   [8] MUSICAL SYMBOL COMBINING ACCENT..MUSICAL SYMBOL COMBINING LOURE
    Interval{ 0x1D185, 0x1D18B }, // Mn   [7] MUSICAL SYMBOL COMBINING DOIT..MUSICAL SYMBOL COMBINING TRIPLE TONGUE
    Interval{ 0x1D1AA, 0x1D1AD }, // Mn   [4] MUSICAL SYMBOL COMBINING DOWN BOW..MUSICAL SYMBOL COMBINING SNAP PIZZICATO
    Interval{ 0x1D242, 0x1D244 }, // Mn   [3] COMBINING GREEK MUSICAL TRISEME..COMBINING GREEK MUSICAL PENTASEME
    Interval{ 0x1D400, 0x1D454 }, // L&  [85] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL ITALIC SMALL G
    Interval{ 0x1D456, 0x1D49C }, // L&  [71] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL SCRIPT CAPITAL A
    Interval{ 0x1D49E, 0x1D49F }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Interval{ 0x1D4A2, 0x1D4A2 }, // L&       MATHEMATICAL SCRIPT CAPITAL G
    Interval{ 0x1D4A5, 0x1D4A6 }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Interval{ 0x1D4A9, 0x1D4AC }, // L&   [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Interval{ 0x1D4AE, 0x1D4B9 }, // L&  [12] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT SMALL D
    Interval{ 0x1D4BB, 0x1D4BB }, // L&       MATHEMATICAL SCRIPT SMALL F
    Interval{ 0x1D4BD, 0x1D4C3 }, // L&   [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Interval{ 0x1D4C5, 0x1D505 }, // L&  [65] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL FRAKTUR CAPITAL B
    Interval{ 0x1D507, 0x1D50A }, // L&   [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Interval{ 0x1D50D, 0x1D514 }, // L&   [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Interval{ 0x1D516, 0x1D51C }, // L&   [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Interval{ 0x1D51E, 0x1D539 }, // L&  [28] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Interval{ 0x1D53B, 0x1D53E }, // L&   [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Interval{ 0x1D540, 0x1D544 }, // L&   [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Interval{ 0x1D546, 0x1D546 }, // L&       MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Interval{ 0x1D54A, 0x1D550 }, // L&   [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Interval{ 0x1D552, 0x1D6A5 }, // L& [340] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    Interval{ 0x1D6A8, 0x1D6C0 }, // L&  [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    Interval{ 0x1D6C2, 0x1D6DA }, // L&  [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    Interval{ 0x1D6DC, 0x1D6FA }, // L&  [31] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL ITALIC CAPITAL OMEGA
    Interval{ 0x1D6FC, 0x1D714 }, // L&  [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    Interval{ 0x1D716, 0x1D734 }, // L&  [31] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D736, 0x1D74E }, // L&  [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D750, 0x1D76E }, // L&  [31] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    Interval{ 0x1D770, 0x1D788 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    Interval{ 0x1D78A, 0x1D7A8 }, // L&  [31] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D7AA, 0x1D7C2 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D7C4, 0x1D7CB }, // L&   [8] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD SMALL DIGAMMA
    Interval{ 0x1D7CE, 0x1D7FF }, // Nd  [50] MATHEMATICAL BOLD DIGIT ZERO..MATHEMATICAL MONOSPACE DIGIT NINE
    Interval{ 0x1DA00, 0x1DA36 }, // Mn  [55] SIGNWRITING HEAD RIM..SIGNWRITING AIR SUCKING IN
    Interval{ 0x1DA3B, 0x1DA6C }, // Mn  [50] SIGNWRITING MOUTH CLOSED NEUTRAL..SIGNWRITING EXCITEMENT
    Interval{ 0x1DA75, 0x1DA75 }, // Mn       SIGNWRITING UPPER BODY TILTING FROM HIP JOINTS
    Interval{ 0x1DA84, 0x1DA84 }, // Mn       SIGNWRITING LOCATION HEAD NECK
    Interval{ 0x1DA9B, 0x1DA9F }, // Mn   [5] SIGNWRITING FILL MODIFIER-2..SIGNWRITING FILL MODIFIER-6
    Interval{ 0x1DAA1, 0x1DAAF }, // Mn  [15] SIGNWRITING ROTATION MODIFIER-2..SIGNWRITING ROTATION MODIFIER-16
    Interval{ 0x1DF00, 0x1DF09 }, // L&  [10] LATIN SMALL LETTER FENG DIGRAPH WITH TRILL..LATIN SMALL LETTER T WITH HOOK AND RETROFLEX HOOK
    Interval{ 0x1DF0A, 0x1DF0A }, // Lo       LATIN LETTER RETROFLEX CLICK WITH RETROFLEX HOOK
    Interval{ 0x1DF0B, 0x1DF1E }, // L&  [20] LATIN SMALL LETTER ESH WITH DOUBLE BAR..LATIN SMALL LETTER S WITH CURL
    Interval{ 0x1DF25, 0x1DF2A }, // L&   [6] LATIN SMALL LETTER D WITH MID-HEIGHT LEFT HOOK..LATIN SMALL LETTER T WITH MID-HEIGHT LEFT HOOK
    Interval{ 0x1E000, 0x1E006 }, // Mn   [7] COMBINING GLAGOLITIC LETTER AZU..COMBINING GLAGOLITIC LETTER ZHIVETE
    Interval{ 0x1E008, 0x1E018 }, // Mn  [17] COMBINING GLAGOLITIC LETTER ZEMLJA..COMBINING GLAGOLITIC LETTER HERU
    Interval{ 0x1E01B, 0x1E021 }, // Mn   [7] COMBINING GLAGOLITIC LETTER SHTA..COMBINING GLAGOLITIC LETTER YATI
    Interval{ 0x1E023, 0x1E024 }, // Mn   [2] COMBINING GLAGOLITIC LETTER YU..COMBINING GLAGOLITIC LETTER SMALL YUS
    Interval{ 0x1E026, 0x1E02A }, // Mn   [5] COMBINING GLAGOLITIC LETTER YO..COMBINING GLAGOLITIC LETTER FITA
    Interval{ 0x1E030, 0x1E06D }, // Lm  [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Interval{ 0x1E08F, 0x1E08F }, // Mn       COMBINING CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
    Interval{ 0x1E100, 0x1E12C }, // Lo  [45] NYIAKENG PUACHUE HMONG LETTER MA..NYIAKENG PUACHUE HMONG LETTER W
    Interval{ 0x1E130, 0x1E136 }, // Mn   [7] NYIAKENG PUACHUE HMONG TONE-B..NYIAKENG PUACHUE HMONG TONE-D
    Interval{ 0x1E137, 0x1E13D }, // Lm   [7] NYIAKENG PUACHUE HMONG SIGN FOR PERSON..NYIAKENG PUACHUE HMONG SYLLABLE LENGTHENER
    Interval{ 0x1E140, 0x1E149 }, // Nd  [10] NYIAKENG PUACHUE HMONG DIGIT ZERO..NYIAKENG PUACHUE HMONG DIGIT NINE
    Interval{ 0x1E14E, 0x1E14E }, // Lo       NYIAKENG PUACHUE HMONG LOGOGRAM NYAJ
    Interval{ 0x1E290, 0x1E2AD }, // Lo  [30] TOTO LETTER PA..TOTO LETTER A
    Interval{ 0x1E2AE, 0x1E2AE }, // Mn       TOTO SIGN RISING TONE
    Interval{ 0x1E2C0, 0x1E2EB }, // Lo  [44] WANCHO LETTER AA..WANCHO LETTER YIH
    Interval{ 0x1E2EC, 0x1E2EF }, // Mn   [4] WANCHO TONE TUP..WANCHO TONE KOINI
    Interval{ 0x1E2F0, 0x1E2F9 }, // Nd  [10] WANCHO DIGIT ZERO..WANCHO DIGIT NINE
    Interval{ 0x1E4D0, 0x1E4EA }, // Lo  [27] NAG MUNDARI LETTER O..NAG MUNDARI LETTER ELL
    Interval{ 0x1E4EB, 0x1E4EB }, // Lm       NAG MUNDARI SIGN OJOD
    Interval{ 0x1E4EC, 0x1E4EF }, // Mn   [4] NAG MUNDARI SIGN MUHOR..NAG MUNDARI SIGN SUTUH
    Interval{ 0x1E4F0, 0x1E4F9 }, // Nd  [10] NAG MUNDARI DIGIT ZERO..NAG MUNDARI DIGIT NINE
    Interval{ 0x1E5D0, 0x1E5ED }, // Lo  [30] OL ONAL LETTER O..OL ONAL LETTER EG
    Interval{ 0x1E5EE, 0x1E5EF }, // Mn   [2] OL ONAL SIGN MU..OL ONAL SIGN IKIR
    Interval{ 0x1E5F0, 0x1E5F0 }, // Lo       OL ONAL SIGN HODDOND
    Interval{ 0x1E5F1, 0x1E5FA }, // Nd  [10] OL ONAL DIGIT ZERO..OL ONAL DIGIT NINE
    Interval{ 0x1E6C0, 0x1E6DE }, // Lo  [31] TAI YO LETTER LOW KO..TAI YO LETTER HIGH KVO
    Interval{ 0x1E6E0, 0x1E6E2 }, // Lo   [3] TAI YO LETTER AA..TAI YO LETTER UE
    Interval{ 0x1E6E3, 0x1E6E3 }, // Mn       TAI YO SIGN UE
    Interval{ 0x1E6E4, 0x1E6E5 }, // Lo   [2] TAI YO LETTER U..TAI YO LETTER AE
    Interval{ 0x1E6E6, 0x1E6E6 }, // Mn       TAI YO SIGN AU
    Interval{ 0x1E6E7, 0x1E6ED }, // Lo   [7] TAI YO LETTER O..TAI YO LETTER AUE
    Interval{ 0x1E6EE, 0x1E6EF }, // Mn   [2] TAI YO SIGN AY..TAI YO SIGN ANG
    Interval{ 0x1E6F0, 0x1E6F4 }, // Lo   [5] TAI YO LETTER AN..TAI YO LETTER AP
    Interval{ 0x1E6F5, 0x1E6F5 }, // Mn       TAI YO SIGN OM
    Interval{ 0x1E6FE, 0x1E6FE }, // Lo       TAI YO SYMBOL MUEANG
    Interval{ 0x1E6FF, 0x1E6FF }, // Lm       TAI YO XAM LAI
    Interval{ 0x1E7E0, 0x1E7E6 }, // Lo   [7] ETHIOPIC SYLLABLE HHYA..ETHIOPIC SYLLABLE HHYO
    Interval{ 0x1E7E8, 0x1E7EB }, // Lo   [4] ETHIOPIC SYLLABLE GURAGE HHWA..ETHIOPIC SYLLABLE HHWE
    Interval{ 0x1E7ED, 0x1E7EE }, // Lo   [2] ETHIOPIC SYLLABLE GURAGE MWI..ETHIOPIC SYLLABLE GURAGE MWEE
    Interval{ 0x1E7F0, 0x1E7FE }, // Lo  [15] ETHIOPIC SYLLABLE GURAGE QWI..ETHIOPIC SYLLABLE GURAGE PWEE
    Interval{ 0x1E800, 0x1E8C4 }, // Lo [197] MENDE KIKAKUI SYLLABLE M001 KI..MENDE KIKAKUI SYLLABLE M060 NYON
    Interval{ 0x1E8D0, 0x1E8D6 }, // Mn   [7] MENDE KIKAKUI COMBINING NUMBER TEENS..MENDE KIKAKUI COMBINING NUMBER MILLIONS
    Interval{ 0x1E900, 0x1E943 }, // L&  [68] ADLAM CAPITAL LETTER ALIF..ADLAM SMALL LETTER SHA
    Interval{ 0x1E944, 0x1E94A }, // Mn   [7] ADLAM ALIF LENGTHENER..ADLAM NUKTA
    Interval{ 0x1E94B, 0x1E94B }, // Lm       ADLAM NASALIZATION MARK
    Interval{ 0x1E950, 0x1E959 }, // Nd  [10] ADLAM DIGIT ZERO..ADLAM DIGIT NINE
    Interval{ 0x1EE00, 0x1EE03 }, // Lo   [4] ARABIC MATHEMATICAL ALEF..ARABIC MATHEMATICAL DAL
    Interval{ 0x1EE05, 0x1EE1F }, // Lo  [27] ARABIC MATHEMATICAL WAW..ARABIC MATHEMATICAL DOTLESS QAF
    Interval{ 0x1EE21, 0x1EE22 }, // Lo   [2] ARABIC MATHEMATICAL INITIAL BEH..ARABIC MATHEMATICAL INITIAL JEEM
    Interval{ 0x1EE24, 0x1EE24 }, // Lo       ARABIC MATHEMATICAL INITIAL HEH
    Interval{ 0x1EE27, 0x1EE27 }, // Lo       ARABIC MATHEMATICAL INITIAL HAH
    Interval{ 0x1EE29, 0x1EE32 }, // Lo  [10] ARABIC MATHEMATICAL INITIAL YEH..ARABIC MATHEMATICAL INITIAL QAF
    Interval{ 0x1EE34, 0x1EE37 }, // Lo   [4] ARABIC MATHEMATICAL INITIAL SHEEN..ARABIC MATHEMATICAL INITIAL KHAH
    Interval{ 0x1EE39, 0x1EE39 }, // Lo       ARABIC MATHEMATICAL INITIAL DAD
    Interval{ 0x1EE3B, 0x1EE3B }, // Lo       ARABIC MATHEMATICAL INITIAL GHAIN
    Interval{ 0x1EE42, 0x1EE42 }, // Lo       ARABIC MATHEMATICAL TAILED JEEM
    Interval{ 0x1EE47, 0x1EE47 }, // Lo       ARABIC MATHEMATICAL TAILED HAH
    Interval{ 0x1EE49, 0x1EE49 }, // Lo       ARABIC MATHEMATICAL TAILED YEH
    Interval{ 0x1EE4B, 0x1EE4B }, // Lo       ARABIC MATHEMATICAL TAILED LAM
    Interval{ 0x1EE4D, 0x1EE4F }, // Lo   [3] ARABIC MATHEMATICAL TAILED NOON..ARABIC MATHEMATICAL TAILED AIN
    Interval{ 0x1EE51, 0x1EE52 }, // Lo   [2] ARABIC MATHEMATICAL TAILED SAD..ARABIC MATHEMATICAL TAILED QAF
    Interval{ 0x1EE54, 0x1EE54 }, // Lo       ARABIC MATHEMATICAL TAILED SHEEN
    Interval{ 0x1EE57, 0x1EE57 }, // Lo       ARABIC MATHEMATICAL TAILED KHAH
    Interval{ 0x1EE59, 0x1EE59 }, // Lo       ARABIC MATHEMATICAL TAILED DAD
    Interval{ 0x1EE5B, 0x1EE5B }, // Lo       ARABIC MATHEMATICAL TAILED GHAIN
    Interval{ 0x1EE5D, 0x1EE5D }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS NOON
    Interval{ 0x1EE5F, 0x1EE5F }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS QAF
    Interval{ 0x1EE61, 0x1EE62 }, // Lo   [2] ARABIC MATHEMATICAL STRETCHED BEH..ARABIC MATHEMATICAL STRETCHED JEEM
    Interval{ 0x1EE64, 0x1EE64 }, // Lo       ARABIC MATHEMATICAL STRETCHED HEH
    Interval{ 0x1EE67, 0x1EE6A }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED HAH..ARABIC MATHEMATICAL STRETCHED KAF
    Interval{ 0x1EE6C, 0x1EE72 }, // Lo   [7] ARABIC MATHEMATICAL STRETCHED MEEM..ARABIC MATHEMATICAL STRETCHED QAF
    Interval{ 0x1EE74, 0x1EE77 }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED SHEEN..ARABIC MATHEMATICAL STRETCHED KHAH
    Interval{ 0x1EE79, 0x1EE7C }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED DAD..ARABIC MATHEMATICAL STRETCHED DOTLESS BEH
    Interval{ 0x1EE7E, 0x1EE7E }, // Lo       ARABIC MATHEMATICAL STRETCHED DOTLESS FEH
    Interval{ 0x1EE80, 0x1EE89 }, // Lo  [10] ARABIC MATHEMATICAL LOOPED ALEF..ARABIC MATHEMATICAL LOOPED YEH
    Interval{ 0x1EE8B, 0x1EE9B }, // Lo  [17] ARABIC MATHEMATICAL LOOPED LAM..ARABIC MATHEMATICAL LOOPED GHAIN
    Interval{ 0x1EEA1, 0x1EEA3 }, // Lo   [3] ARABIC MATHEMATICAL DOUBLE-STRUCK BEH..ARABIC MATHEMATICAL DOUBLE-STRUCK DAL
    Interval{ 0x1EEA5, 0x1EEA9 }, // Lo   [5] ARABIC MATHEMATICAL DOUBLE-STRUCK WAW..ARABIC MATHEMATICAL DOUBLE-STRUCK YEH
    Interval{ 0x1EEAB, 0x1EEBB }, // Lo  [17] ARABIC MATHEMATICAL DOUBLE-STRUCK LAM..ARABIC MATHEMATICAL DOUBLE-STRUCK GHAIN
    Interval{ 0x1FBF0, 0x1FBF9 }, // Nd  [10] SEGMENTED DIGIT ZERO..SEGMENTED DIGIT NINE
    Interval{ 0x20000, 0x2A6DF }, // Lo [42720] CJK UNIFIED IDEOGRAPH-20000..CJK UNIFIED IDEOGRAPH-2A6DF
    Interval{ 0x2A700, 0x2B81D }, // Lo [4382] CJK UNIFIED IDEOGRAPH-2A700..CJK UNIFIED IDEOGRAPH-2B81D
    Interval{ 0x2B820, 0x2CEAD }, // Lo [5774] CJK UNIFIED IDEOGRAPH-2B820..CJK UNIFIED IDEOGRAPH-2CEAD
    Interval{ 0x2CEB0, 0x2EBE0 }, // Lo [7473] CJK UNIFIED IDEOGRAPH-2CEB0..CJK UNIFIED IDEOGRAPH-2EBE0
    Interval{ 0x2EBF0, 0x2EE5D }, // Lo [622] CJK UNIFIED IDEOGRAPH-2EBF0..CJK UNIFIED IDEOGRAPH-2EE5D
    Interval{ 0x2F800, 0x2FA1D }, // Lo [542] CJK COMPATIBILITY IDEOGRAPH-2F800..CJK COMPATIBILITY IDEOGRAPH-2FA1D
    Interval{ 0x30000, 0x3134A }, // Lo [4939] CJK UNIFIED IDEOGRAPH-30000..CJK UNIFIED IDEOGRAPH-3134A
    Interval{ 0x31350, 0x33479 }, // Lo [8490] CJK UNIFIED IDEOGRAPH-31350..CJK UNIFIED IDEOGRAPH-33479
    Interval{ 0xE0100, 0xE01EF }, // Mn [240] VARIATION SELECTOR-17..VARIATION SELECTOR-256
}; // }}}
auto static const ID_Start = std::array<Interval, 776>{ // {{{
    Interval{ 0x0041, 0x005A }, // L&  [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Interval{ 0x0061, 0x007A }, // L&  [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Interval{ 0x00AA, 0x00AA }, // Lo       FEMININE ORDINAL INDICATOR
    Interval{ 0x00B5, 0x00B5 }, // L&       MICRO SIGN
    Interval{ 0x00BA, 0x00BA }, // Lo       MASCULINE ORDINAL INDICATOR
    Interval{ 0x00C0, 0x00D6 }, // L&  [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x00D8, 0x00F6 }, // L&  [31] LATIN CAPITAL LETTER O WITH STROKE..LATIN SMALL LETTER O WITH DIAERESIS
    Interval{ 0x00F8, 0x01BA }, // L& [195] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER EZH WITH TAIL
    Interval{ 0x01BB, 0x01BB }, // Lo       LATIN LETTER TWO WITH STROKE
    Interval{ 0x01BC, 0x01BF }, // L&   [4] LATIN CAPITAL LETTER TONE FIVE..LATIN LETTER WYNN
    Interval{ 0x01C0, 0x01C3 }, // Lo   [4] LATIN LETTER DENTAL CLICK..LATIN LETTER RETROFLEX CLICK
    Interval{ 0x01C4, 0x0293 }, // L& [208] LATIN CAPITAL LETTER DZ WITH CARON..LATIN SMALL LETTER EZH WITH CURL
    Interval{ 0x0294, 0x0295 }, // Lo   [2] LATIN LETTER GLOTTAL STOP..LATIN LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0x0296, 0x02AF }, // L&  [26] LATIN LETTER INVERTED GLOTTAL STOP..LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    Interval{ 0x02B0, 0x02C1 }, // Lm  [18] MODIFIER LETTER SMALL H..MODIFIER LETTER REVERSED GLOTTAL STOP
    Interval{ 0x02C6, 0x02D1 }, // Lm  [12] MODIFIER LETTER CIRCUMFLEX ACCENT..MODIFIER LETTER HALF TRIANGULAR COLON
    Interval{ 0x02E0, 0x02E4 }, // Lm   [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    Interval{ 0x02EC, 0x02EC }, // Lm       MODIFIER LETTER VOICING
    Interval{ 0x02EE, 0x02EE }, // Lm       MODIFIER LETTER DOUBLE APOSTROPHE
    Interval{ 0x0370, 0x0373 }, // L&   [4] GREEK CAPITAL LETTER HETA..GREEK SMALL LETTER ARCHAIC SAMPI
    Interval{ 0x0374, 0x0374 }, // Lm       GREEK NUMERAL SIGN
    Interval{ 0x0376, 0x0377 }, // L&   [2] GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA..GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037A, 0x037A }, // Lm       GREEK YPOGEGRAMMENI
    Interval{ 0x037B, 0x037D }, // L&   [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Interval{ 0x037F, 0x037F }, // L&       GREEK CAPITAL LETTER YOT
    Interval{ 0x0386, 0x0386 }, // L&       GREEK CAPITAL LETTER ALPHA WITH TONOS
    Interval{ 0x0388, 0x038A }, // L&   [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Interval{ 0x038C, 0x038C }, // L&       GREEK CAPITAL LETTER OMICRON WITH TONOS
    Interval{ 0x038E, 0x03A1 }, // L&  [20] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER RHO
    Interval{ 0x03A3, 0x03F5 }, // L&  [83] GREEK CAPITAL LETTER SIGMA..GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F7, 0x0481 }, // L& [139] GREEK CAPITAL LETTER SHO..CYRILLIC SMALL LETTER KOPPA
    Interval{ 0x048A, 0x052F }, // L& [166] CYRILLIC CAPITAL LETTER SHORT I WITH TAIL..CYRILLIC SMALL LETTER EL WITH DESCENDER
    Interval{ 0x0531, 0x0556 }, // L&  [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Interval{ 0x0559, 0x0559 }, // Lm       ARMENIAN MODIFIER LETTER LEFT HALF RING
    Interval{ 0x0560, 0x0588 }, // L&  [41] ARMENIAN SMALL LETTER TURNED AYB..ARMENIAN SMALL LETTER YI WITH STROKE
    Interval{ 0x05D0, 0x05EA }, // Lo  [27] HEBREW LETTER ALEF..HEBREW LETTER TAV
    Interval{ 0x05EF, 0x05F2 }, // Lo   [4] HEBREW YOD TRIANGLE..HEBREW LIGATURE YIDDISH DOUBLE YOD
    Interval{ 0x0620, 0x063F }, // Lo  [32] ARABIC LETTER KASHMIRI YEH..ARABIC LETTER FARSI YEH WITH THREE DOTS ABOVE
    Interval{ 0x0640, 0x0640 }, // Lm       ARABIC TATWEEL
    Interval{ 0x0641, 0x064A }, // Lo  [10] ARABIC LETTER FEH..ARABIC LETTER YEH
    Interval{ 0x066E, 0x066F }, // Lo   [2] ARABIC LETTER DOTLESS BEH..ARABIC LETTER DOTLESS QAF
    Interval{ 0x0671, 0x06D3 }, // Lo  [99] ARABIC LETTER ALEF WASLA..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE
    Interval{ 0x06D5, 0x06D5 }, // Lo       ARABIC LETTER AE
    Interval{ 0x06E5, 0x06E6 }, // Lm   [2] ARABIC SMALL WAW..ARABIC SMALL YEH
    Interval{ 0x06EE, 0x06EF }, // Lo   [2] ARABIC LETTER DAL WITH INVERTED V..ARABIC LETTER REH WITH INVERTED V
    Interval{ 0x06FA, 0x06FC }, // Lo   [3] ARABIC LETTER SHEEN WITH DOT BELOW..ARABIC LETTER GHAIN WITH DOT BELOW
    Interval{ 0x06FF, 0x06FF }, // Lo       ARABIC LETTER HEH WITH INVERTED V
    Interval{ 0x0710, 0x0710 }, // Lo       SYRIAC LETTER ALAPH
    Interval{ 0x0712, 0x072F }, // Lo  [30] SYRIAC LETTER BETH..SYRIAC LETTER PERSIAN DHALATH
    Interval{ 0x074D, 0x07A5 }, // Lo  [89] SYRIAC LETTER SOGDIAN ZHAIN..THAANA LETTER WAAVU
    Interval{ 0x07B1, 0x07B1 }, // Lo       THAANA LETTER NAA
    Interval{ 0x07CA, 0x07EA }, // Lo  [33] NKO LETTER A..NKO LETTER JONA RA
    Interval{ 0x07F4, 0x07F5 }, // Lm   [2] NKO HIGH TONE APOSTROPHE..NKO LOW TONE APOSTROPHE
    Interval{ 0x07FA, 0x07FA }, // Lm       NKO LAJANYALAN
    Interval{ 0x0800, 0x0815 }, // Lo  [22] SAMARITAN LETTER ALAF..SAMARITAN LETTER TAAF
    Interval{ 0x081A, 0x081A }, // Lm       SAMARITAN MODIFIER LETTER EPENTHETIC YUT
    Interval{ 0x0824, 0x0824 }, // Lm       SAMARITAN MODIFIER LETTER SHORT A
    Interval{ 0x0828, 0x0828 }, // Lm       SAMARITAN MODIFIER LETTER I
    Interval{ 0x0840, 0x0858 }, // Lo  [25] MANDAIC LETTER HALQA..MANDAIC LETTER AIN
    Interval{ 0x0860, 0x086A }, // Lo  [11] SYRIAC LETTER MALAYALAM NGA..SYRIAC LETTER MALAYALAM SSA
    Interval{ 0x0870, 0x0887 }, // Lo  [24] ARABIC LETTER ALEF WITH ATTACHED FATHA..ARABIC BASELINE ROUND DOT
    Interval{ 0x0889, 0x088F }, // Lo   [7] ARABIC LETTER NOON WITH INVERTED SMALL V..ARABIC LETTER NOON WITH RING ABOVE
    Interval{ 0x08A0, 0x08C8 }, // Lo  [41] ARABIC LETTER BEH WITH SMALL V BELOW..ARABIC LETTER GRAF
    Interval{ 0x08C9, 0x08C9 }, // Lm       ARABIC SMALL FARSI YEH
    Interval{ 0x0904, 0x0939 }, // Lo  [54] DEVANAGARI LETTER SHORT A..DEVANAGARI LETTER HA
    Interval{ 0x093D, 0x093D }, // Lo       DEVANAGARI SIGN AVAGRAHA
    Interval{ 0x0950, 0x0950 }, // Lo       DEVANAGARI OM
    Interval{ 0x0958, 0x0961 }, // Lo  [10] DEVANAGARI LETTER QA..DEVANAGARI LETTER VOCALIC LL
    Interval{ 0x0971, 0x0971 }, // Lm       DEVANAGARI SIGN HIGH SPACING DOT
    Interval{ 0x0972, 0x0980 }, // Lo  [15] DEVANAGARI LETTER CANDRA A..BENGALI ANJI
    Interval{ 0x0985, 0x098C }, // Lo   [8] BENGALI LETTER A..BENGALI LETTER VOCALIC L
    Interval{ 0x098F, 0x0990 }, // Lo   [2] BENGALI LETTER E..BENGALI LETTER AI
    Interval{ 0x0993, 0x09A8 }, // Lo  [22] BENGALI LETTER O..BENGALI LETTER NA
    Interval{ 0x09AA, 0x09B0 }, // Lo   [7] BENGALI LETTER PA..BENGALI LETTER RA
    Interval{ 0x09B2, 0x09B2 }, // Lo       BENGALI LETTER LA
    Interval{ 0x09B6, 0x09B9 }, // Lo   [4] BENGALI LETTER SHA..BENGALI LETTER HA
    Interval{ 0x09BD, 0x09BD }, // Lo       BENGALI SIGN AVAGRAHA
    Interval{ 0x09CE, 0x09CE }, // Lo       BENGALI LETTER KHANDA TA
    Interval{ 0x09DC, 0x09DD }, // Lo   [2] BENGALI LETTER RRA..BENGALI LETTER RHA
    Interval{ 0x09DF, 0x09E1 }, // Lo   [3] BENGALI LETTER YYA..BENGALI LETTER VOCALIC LL
    Interval{ 0x09F0, 0x09F1 }, // Lo   [2] BENGALI LETTER RA WITH MIDDLE DIAGONAL..BENGALI LETTER RA WITH LOWER DIAGONAL
    Interval{ 0x09FC, 0x09FC }, // Lo       BENGALI LETTER VEDIC ANUSVARA
    Interval{ 0x0A05, 0x0A0A }, // Lo   [6] GURMUKHI LETTER A..GURMUKHI LETTER UU
    Interval{ 0x0A0F, 0x0A10 }, // Lo   [2] GURMUKHI LETTER EE..GURMUKHI LETTER AI
    Interval{ 0x0A13, 0x0A28 }, // Lo  [22] GURMUKHI LETTER OO..GURMUKHI LETTER NA
    Interval{ 0x0A2A, 0x0A30 }, // Lo   [7] GURMUKHI LETTER PA..GURMUKHI LETTER RA
    Interval{ 0x0A32, 0x0A33 }, // Lo   [2] GURMUKHI LETTER LA..GURMUKHI LETTER LLA
    Interval{ 0x0A35, 0x0A36 }, // Lo   [2] GURMUKHI LETTER VA..GURMUKHI LETTER SHA
    Interval{ 0x0A38, 0x0A39 }, // Lo   [2] GURMUKHI LETTER SA..GURMUKHI LETTER HA
    Interval{ 0x0A59, 0x0A5C }, // Lo   [4] GURMUKHI LETTER KHHA..GURMUKHI LETTER RRA
    Interval{ 0x0A5E, 0x0A5E }, // Lo       GURMUKHI LETTER FA
    Interval{ 0x0A72, 0x0A74 }, // Lo   [3] GURMUKHI IRI..GURMUKHI EK ONKAR
    Interval{ 0x0A85, 0x0A8D }, // Lo   [9] GUJARATI LETTER A..GUJARATI VOWEL CANDRA E
    Interval{ 0x0A8F, 0x0A91 }, // Lo   [3] GUJARATI LETTER E..GUJARATI VOWEL CANDRA O
    Interval{ 0x0A93, 0x0AA8 }, // Lo  [22] GUJARATI LETTER O..GUJARATI LETTER NA
    Interval{ 0x0AAA, 0x0AB0 }, // Lo   [7] GUJARATI LETTER PA..GUJARATI LETTER RA
    Interval{ 0x0AB2, 0x0AB3 }, // Lo   [2] GUJARATI LETTER LA..GUJARATI LETTER LLA
    Interval{ 0x0AB5, 0x0AB9 }, // Lo   [5] GUJARATI LETTER VA..GUJARATI LETTER HA
    Interval{ 0x0ABD, 0x0ABD }, // Lo       GUJARATI SIGN AVAGRAHA
    Interval{ 0x0AD0, 0x0AD0 }, // Lo       GUJARATI OM
    Interval{ 0x0AE0, 0x0AE1 }, // Lo   [2] GUJARATI LETTER VOCALIC RR..GUJARATI LETTER VOCALIC LL
    Interval{ 0x0AF9, 0x0AF9 }, // Lo       GUJARATI LETTER ZHA
    Interval{ 0x0B05, 0x0B0C }, // Lo   [8] ORIYA LETTER A..ORIYA LETTER VOCALIC L
    Interval{ 0x0B0F, 0x0B10 }, // Lo   [2] ORIYA LETTER E..ORIYA LETTER AI
    Interval{ 0x0B13, 0x0B28 }, // Lo  [22] ORIYA LETTER O..ORIYA LETTER NA
    Interval{ 0x0B2A, 0x0B30 }, // Lo   [7] ORIYA LETTER PA..ORIYA LETTER RA
    Interval{ 0x0B32, 0x0B33 }, // Lo   [2] ORIYA LETTER LA..ORIYA LETTER LLA
    Interval{ 0x0B35, 0x0B39 }, // Lo   [5] ORIYA LETTER VA..ORIYA LETTER HA
    Interval{ 0x0B3D, 0x0B3D }, // Lo       ORIYA SIGN AVAGRAHA
    Interval{ 0x0B5C, 0x0B5D }, // Lo   [2] ORIYA LETTER RRA..ORIYA LETTER RHA
    Interval{ 0x0B5F, 0x0B61 }, // Lo   [3] ORIYA LETTER YYA..ORIYA LETTER VOCALIC LL
    Interval{ 0x0B71, 0x0B71 }, // Lo       ORIYA LETTER WA
    Interval{ 0x0B83, 0x0B83 }, // Lo       TAMIL SIGN VISARGA
    Interval{ 0x0B85, 0x0B8A }, // Lo   [6] TAMIL LETTER A..TAMIL LETTER UU
    Interval{ 0x0B8E, 0x0B90 }, // Lo   [3] TAMIL LETTER E..TAMIL LETTER AI
    Interval{ 0x0B92, 0x0B95 }, // Lo   [4] TAMIL LETTER O..TAMIL LETTER KA
    Interval{ 0x0B99, 0x0B9A }, // Lo   [2] TAMIL LETTER NGA..TAMIL LETTER CA
    Interval{ 0x0B9C, 0x0B9C }, // Lo       TAMIL LETTER JA
    Interval{ 0x0B9E, 0x0B9F }, // Lo   [2] TAMIL LETTER NYA..TAMIL LETTER TTA
    Interval{ 0x0BA3, 0x0BA4 }, // Lo   [2] TAMIL LETTER NNA..TAMIL LETTER TA
    Interval{ 0x0BA8, 0x0BAA }, // Lo   [3] TAMIL LETTER NA..TAMIL LETTER PA
    Interval{ 0x0BAE, 0x0BB9 }, // Lo  [12] TAMIL LETTER MA..TAMIL LETTER HA
    Interval{ 0x0BD0, 0x0BD0 }, // Lo       TAMIL OM
    Interval{ 0x0C05, 0x0C0C }, // Lo   [8] TELUGU LETTER A..TELUGU LETTER VOCALIC L
    Interval{ 0x0C0E, 0x0C10 }, // Lo   [3] TELUGU LETTER E..TELUGU LETTER AI
    Interval{ 0x0C12, 0x0C28 }, // Lo  [23] TELUGU LETTER O..TELUGU LETTER NA
    Interval{ 0x0C2A, 0x0C39 }, // Lo  [16] TELUGU LETTER PA..TELUGU LETTER HA
    Interval{ 0x0C3D, 0x0C3D }, // Lo       TELUGU SIGN AVAGRAHA
    Interval{ 0x0C58, 0x0C5A }, // Lo   [3] TELUGU LETTER TSA..TELUGU LETTER RRRA
    Interval{ 0x0C5C, 0x0C5D }, // Lo   [2] TELUGU ARCHAIC SHRII..TELUGU LETTER NAKAARA POLLU
    Interval{ 0x0C60, 0x0C61 }, // Lo   [2] TELUGU LETTER VOCALIC RR..TELUGU LETTER VOCALIC LL
    Interval{ 0x0C80, 0x0C80 }, // Lo       KANNADA SIGN SPACING CANDRABINDU
    Interval{ 0x0C85, 0x0C8C }, // Lo   [8] KANNADA LETTER A..KANNADA LETTER VOCALIC L
    Interval{ 0x0C8E, 0x0C90 }, // Lo   [3] KANNADA LETTER E..KANNADA LETTER AI
    Interval{ 0x0C92, 0x0CA8 }, // Lo  [23] KANNADA LETTER O..KANNADA LETTER NA
    Interval{ 0x0CAA, 0x0CB3 }, // Lo  [10] KANNADA LETTER PA..KANNADA LETTER LLA
    Interval{ 0x0CB5, 0x0CB9 }, // Lo   [5] KANNADA LETTER VA..KANNADA LETTER HA
    Interval{ 0x0CBD, 0x0CBD }, // Lo       KANNADA SIGN AVAGRAHA
    Interval{ 0x0CDC, 0x0CDE }, // Lo   [3] KANNADA ARCHAIC SHRII..KANNADA LETTER FA
    Interval{ 0x0CE0, 0x0CE1 }, // Lo   [2] KANNADA LETTER VOCALIC RR..KANNADA LETTER VOCALIC LL
    Interval{ 0x0CF1, 0x0CF2 }, // Lo   [2] KANNADA SIGN JIHVAMULIYA..KANNADA SIGN UPADHMANIYA
    Interval{ 0x0D04, 0x0D0C }, // Lo   [9] MALAYALAM LETTER VEDIC ANUSVARA..MALAYALAM LETTER VOCALIC L
    Interval{ 0x0D0E, 0x0D10 }, // Lo   [3] MALAYALAM LETTER E..MALAYALAM LETTER AI
    Interval{ 0x0D12, 0x0D3A }, // Lo  [41] MALAYALAM LETTER O..MALAYALAM LETTER TTTA
    Interval{ 0x0D3D, 0x0D3D }, // Lo       MALAYALAM SIGN AVAGRAHA
    Interval{ 0x0D4E, 0x0D4E }, // Lo       MALAYALAM LETTER DOT REPH
    Interval{ 0x0D54, 0x0D56 }, // Lo   [3] MALAYALAM LETTER CHILLU M..MALAYALAM LETTER CHILLU LLL
    Interval{ 0x0D5F, 0x0D61 }, // Lo   [3] MALAYALAM LETTER ARCHAIC II..MALAYALAM LETTER VOCALIC LL
    Interval{ 0x0D7A, 0x0D7F }, // Lo   [6] MALAYALAM LETTER CHILLU NN..MALAYALAM LETTER CHILLU K
    Interval{ 0x0D85, 0x0D96 }, // Lo  [18] SINHALA LETTER AYANNA..SINHALA LETTER AUYANNA
    Interval{ 0x0D9A, 0x0DB1 }, // Lo  [24] SINHALA LETTER ALPAPRAANA KAYANNA..SINHALA LETTER DANTAJA NAYANNA
    Interval{ 0x0DB3, 0x0DBB }, // Lo   [9] SINHALA LETTER SANYAKA DAYANNA..SINHALA LETTER RAYANNA
    Interval{ 0x0DBD, 0x0DBD }, // Lo       SINHALA LETTER DANTAJA LAYANNA
    Interval{ 0x0DC0, 0x0DC6 }, // Lo   [7] SINHALA LETTER VAYANNA..SINHALA LETTER FAYANNA
    Interval{ 0x0E01, 0x0E30 }, // Lo  [48] THAI CHARACTER KO KAI..THAI CHARACTER SARA A
    Interval{ 0x0E32, 0x0E33 }, // Lo   [2] THAI CHARACTER SARA AA..THAI CHARACTER SARA AM
    Interval{ 0x0E40, 0x0E45 }, // Lo   [6] THAI CHARACTER SARA E..THAI CHARACTER LAKKHANGYAO
    Interval{ 0x0E46, 0x0E46 }, // Lm       THAI CHARACTER MAIYAMOK
    Interval{ 0x0E81, 0x0E82 }, // Lo   [2] LAO LETTER KO..LAO LETTER KHO SUNG
    Interval{ 0x0E84, 0x0E84 }, // Lo       LAO LETTER KHO TAM
    Interval{ 0x0E86, 0x0E8A }, // Lo   [5] LAO LETTER PALI GHA..LAO LETTER SO TAM
    Interval{ 0x0E8C, 0x0EA3 }, // Lo  [24] LAO LETTER PALI JHA..LAO LETTER LO LING
    Interval{ 0x0EA5, 0x0EA5 }, // Lo       LAO LETTER LO LOOT
    Interval{ 0x0EA7, 0x0EB0 }, // Lo  [10] LAO LETTER WO..LAO VOWEL SIGN A
    Interval{ 0x0EB2, 0x0EB3 }, // Lo   [2] LAO VOWEL SIGN AA..LAO VOWEL SIGN AM
    Interval{ 0x0EBD, 0x0EBD }, // Lo       LAO SEMIVOWEL SIGN NYO
    Interval{ 0x0EC0, 0x0EC4 }, // Lo   [5] LAO VOWEL SIGN E..LAO VOWEL SIGN AI
    Interval{ 0x0EC6, 0x0EC6 }, // Lm       LAO KO LA
    Interval{ 0x0EDC, 0x0EDF }, // Lo   [4] LAO HO NO..LAO LETTER KHMU NYO
    Interval{ 0x0F00, 0x0F00 }, // Lo       TIBETAN SYLLABLE OM
    Interval{ 0x0F40, 0x0F47 }, // Lo   [8] TIBETAN LETTER KA..TIBETAN LETTER JA
    Interval{ 0x0F49, 0x0F6C }, // Lo  [36] TIBETAN LETTER NYA..TIBETAN LETTER RRA
    Interval{ 0x0F88, 0x0F8C }, // Lo   [5] TIBETAN SIGN LCE TSA CAN..TIBETAN SIGN INVERTED MCHU CAN
    Interval{ 0x1000, 0x102A }, // Lo  [43] MYANMAR LETTER KA..MYANMAR LETTER AU
    Interval{ 0x103F, 0x103F }, // Lo       MYANMAR LETTER GREAT SA
    Interval{ 0x1050, 0x1055 }, // Lo   [6] MYANMAR LETTER SHA..MYANMAR LETTER VOCALIC LL
    Interval{ 0x105A, 0x105D }, // Lo   [4] MYANMAR LETTER MON NGA..MYANMAR LETTER MON BBE
    Interval{ 0x1061, 0x1061 }, // Lo       MYANMAR LETTER SGAW KAREN SHA
    Interval{ 0x1065, 0x1066 }, // Lo   [2] MYANMAR LETTER WESTERN PWO KAREN THA..MYANMAR LETTER WESTERN PWO KAREN PWA
    Interval{ 0x106E, 0x1070 }, // Lo   [3] MYANMAR LETTER EASTERN PWO KAREN NNA..MYANMAR LETTER EASTERN PWO KAREN GHWA
    Interval{ 0x1075, 0x1081 }, // Lo  [13] MYANMAR LETTER SHAN KA..MYANMAR LETTER SHAN HA
    Interval{ 0x108E, 0x108E }, // Lo       MYANMAR LETTER RUMAI PALAUNG FA
    Interval{ 0x10A0, 0x10C5 }, // L&  [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Interval{ 0x10C7, 0x10C7 }, // L&       GEORGIAN CAPITAL LETTER YN
    Interval{ 0x10CD, 0x10CD }, // L&       GEORGIAN CAPITAL LETTER AEN
    Interval{ 0x10D0, 0x10FA }, // L&  [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Interval{ 0x10FC, 0x10FC }, // Lm       MODIFIER LETTER GEORGIAN NAR
    Interval{ 0x10FD, 0x10FF }, // L&   [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Interval{ 0x1100, 0x1248 }, // Lo [329] HANGUL CHOSEONG KIYEOK..ETHIOPIC SYLLABLE QWA
    Interval{ 0x124A, 0x124D }, // Lo   [4] ETHIOPIC SYLLABLE QWI..ETHIOPIC SYLLABLE QWE
    Interval{ 0x1250, 0x1256 }, // Lo   [7] ETHIOPIC SYLLABLE QHA..ETHIOPIC SYLLABLE QHO
    Interval{ 0x1258, 0x1258 }, // Lo       ETHIOPIC SYLLABLE QHWA
    Interval{ 0x125A, 0x125D }, // Lo   [4] ETHIOPIC SYLLABLE QHWI..ETHIOPIC SYLLABLE QHWE
    Interval{ 0x1260, 0x1288 }, // Lo  [41] ETHIOPIC SYLLABLE BA..ETHIOPIC SYLLABLE XWA
    Interval{ 0x128A, 0x128D }, // Lo   [4] ETHIOPIC SYLLABLE XWI..ETHIOPIC SYLLABLE XWE
    Interval{ 0x1290, 0x12B0 }, // Lo  [33] ETHIOPIC SYLLABLE NA..ETHIOPIC SYLLABLE KWA
    Interval{ 0x12B2, 0x12B5 }, // Lo   [4] ETHIOPIC SYLLABLE KWI..ETHIOPIC SYLLABLE KWE
    Interval{ 0x12B8, 0x12BE }, // Lo   [7] ETHIOPIC SYLLABLE KXA..ETHIOPIC SYLLABLE KXO
    Interval{ 0x12C0, 0x12C0 }, // Lo       ETHIOPIC SYLLABLE KXWA
    Interval{ 0x12C2, 0x12C5 }, // Lo   [4] ETHIOPIC SYLLABLE KXWI..ETHIOPIC SYLLABLE KXWE
    Interval{ 0x12C8, 0x12D6 }, // Lo  [15] ETHIOPIC SYLLABLE WA..ETHIOPIC SYLLABLE PHARYNGEAL O
    Interval{ 0x12D8, 0x1310 }, // Lo  [57] ETHIOPIC SYLLABLE ZA..ETHIOPIC SYLLABLE GWA
    Interval{ 0x1312, 0x1315 }, // Lo   [4] ETHIOPIC SYLLABLE GWI..ETHIOPIC SYLLABLE GWE
    Interval{ 0x1318, 0x135A }, // Lo  [67] ETHIOPIC SYLLABLE GGA..ETHIOPIC SYLLABLE FYA
    Interval{ 0x1380, 0x138F }, // Lo  [16] ETHIOPIC SYLLABLE SEBATBEIT MWA..ETHIOPIC SYLLABLE PWE
    Interval{ 0x13A0, 0x13F5 }, // L&  [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Interval{ 0x13F8, 0x13FD }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1401, 0x166C }, // Lo [620] CANADIAN SYLLABICS E..CANADIAN SYLLABICS CARRIER TTSA
    Interval{ 0x166F, 0x167F }, // Lo  [17] CANADIAN SYLLABICS QAI..CANADIAN SYLLABICS BLACKFOOT W
    Interval{ 0x1681, 0x169A }, // Lo  [26] OGHAM LETTER BEITH..OGHAM LETTER PEITH
    Interval{ 0x16A0, 0x16EA }, // Lo  [75] RUNIC LETTER FEHU FEOH FE F..RUNIC LETTER X
    Interval{ 0x16EE, 0x16F0 }, // Nl   [3] RUNIC ARLAUG SYMBOL..RUNIC BELGTHOR SYMBOL
    Interval{ 0x16F1, 0x16F8 }, // Lo   [8] RUNIC LETTER K..RUNIC LETTER FRANKS CASKET AESC
    Interval{ 0x1700, 0x1711 }, // Lo  [18] TAGALOG LETTER A..TAGALOG LETTER HA
    Interval{ 0x171F, 0x1731 }, // Lo  [19] TAGALOG LETTER ARCHAIC RA..HANUNOO LETTER HA
    Interval{ 0x1740, 0x1751 }, // Lo  [18] BUHID LETTER A..BUHID LETTER HA
    Interval{ 0x1760, 0x176C }, // Lo  [13] TAGBANWA LETTER A..TAGBANWA LETTER YA
    Interval{ 0x176E, 0x1770 }, // Lo   [3] TAGBANWA LETTER LA..TAGBANWA LETTER SA
    Interval{ 0x1780, 0x17B3 }, // Lo  [52] KHMER LETTER KA..KHMER INDEPENDENT VOWEL QAU
    Interval{ 0x17D7, 0x17D7 }, // Lm       KHMER SIGN LEK TOO
    Interval{ 0x17DC, 0x17DC }, // Lo       KHMER SIGN AVAKRAHASANYA
    Interval{ 0x1820, 0x1842 }, // Lo  [35] MONGOLIAN LETTER A..MONGOLIAN LETTER CHI
    Interval{ 0x1843, 0x1843 }, // Lm       MONGOLIAN LETTER TODO LONG VOWEL SIGN
    Interval{ 0x1844, 0x1878 }, // Lo  [53] MONGOLIAN LETTER TODO E..MONGOLIAN LETTER CHA WITH TWO DOTS
    Interval{ 0x1880, 0x1884 }, // Lo   [5] MONGOLIAN LETTER ALI GALI ANUSVARA ONE..MONGOLIAN LETTER ALI GALI INVERTED UBADAMA
    Interval{ 0x1885, 0x1886 }, // Mn   [2] MONGOLIAN LETTER ALI GALI BALUDA..MONGOLIAN LETTER ALI GALI THREE BALUDA
    Interval{ 0x1887, 0x18A8 }, // Lo  [34] MONGOLIAN LETTER ALI GALI A..MONGOLIAN LETTER MANCHU ALI GALI BHA
    Interval{ 0x18AA, 0x18AA }, // Lo       MONGOLIAN LETTER MANCHU ALI GALI LHA
    Interval{ 0x18B0, 0x18F5 }, // Lo  [70] CANADIAN SYLLABICS OY..CANADIAN SYLLABICS CARRIER DENTAL S
    Interval{ 0x1900, 0x191E }, // Lo  [31] LIMBU VOWEL-CARRIER LETTER..LIMBU LETTER TRA
    Interval{ 0x1950, 0x196D }, // Lo  [30] TAI LE LETTER KA..TAI LE LETTER AI
    Interval{ 0x1970, 0x1974 }, // Lo   [5] TAI LE LETTER TONE-2..TAI LE LETTER TONE-6
    Interval{ 0x1980, 0x19AB }, // Lo  [44] NEW TAI LUE LETTER HIGH QA..NEW TAI LUE LETTER LOW SUA
    Interval{ 0x19B0, 0x19C9 }, // Lo  [26] NEW TAI LUE VOWEL SIGN VOWEL SHORTENER..NEW TAI LUE TONE MARK-2
    Interval{ 0x1A00, 0x1A16 }, // Lo  [23] BUGINESE LETTER KA..BUGINESE LETTER HA
    Interval{ 0x1A20, 0x1A54 }, // Lo  [53] TAI THAM LETTER HIGH KA..TAI THAM LETTER GREAT SA
    Interval{ 0x1AA7, 0x1AA7 }, // Lm       TAI THAM SIGN MAI YAMOK
    Interval{ 0x1B05, 0x1B33 }, // Lo  [47] BALINESE LETTER AKARA..BALINESE LETTER HA
    Interval{ 0x1B45, 0x1B4C }, // Lo   [8] BALINESE LETTER KAF SASAK..BALINESE LETTER ARCHAIC JNYA
    Interval{ 0x1B83, 0x1BA0 }, // Lo  [30] SUNDANESE LETTER A..SUNDANESE LETTER HA
    Interval{ 0x1BAE, 0x1BAF }, // Lo   [2] SUNDANESE LETTER KHA..SUNDANESE LETTER SYA
    Interval{ 0x1BBA, 0x1BE5 }, // Lo  [44] SUNDANESE AVAGRAHA..BATAK LETTER U
    Interval{ 0x1C00, 0x1C23 }, // Lo  [36] LEPCHA LETTER KA..LEPCHA LETTER A
    Interval{ 0x1C4D, 0x1C4F }, // Lo   [3] LEPCHA LETTER TTA..LEPCHA LETTER DDA
    Interval{ 0x1C5A, 0x1C77 }, // Lo  [30] OL CHIKI LETTER LA..OL CHIKI LETTER OH
    Interval{ 0x1C78, 0x1C7D }, // Lm   [6] OL CHIKI MU TTUDDAG..OL CHIKI AHAD
    Interval{ 0x1C80, 0x1C8A }, // L&  [11] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER TJE
    Interval{ 0x1C90, 0x1CBA }, // L&  [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Interval{ 0x1CBD, 0x1CBF }, // L&   [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Interval{ 0x1CE9, 0x1CEC }, // Lo   [4] VEDIC SIGN ANUSVARA ANTARGOMUKHA..VEDIC SIGN ANUSVARA VAMAGOMUKHA WITH TAIL
    Interval{ 0x1CEE, 0x1CF3 }, // Lo   [6] VEDIC SIGN HEXIFORM LONG ANUSVARA..VEDIC SIGN ROTATED ARDHAVISARGA
    Interval{ 0x1CF5, 0x1CF6 }, // Lo   [2] VEDIC SIGN JIHVAMULIYA..VEDIC SIGN UPADHMANIYA
    Interval{ 0x1CFA, 0x1CFA }, // Lo       VEDIC SIGN DOUBLE ANUSVARA ANTARGOMUKHA
    Interval{ 0x1D00, 0x1D2B }, // L&  [44] LATIN LETTER SMALL CAPITAL A..CYRILLIC LETTER SMALL CAPITAL EL
    Interval{ 0x1D2C, 0x1D6A }, // Lm  [63] MODIFIER LETTER CAPITAL A..GREEK SUBSCRIPT SMALL LETTER CHI
    Interval{ 0x1D6B, 0x1D77 }, // L&  [13] LATIN SMALL LETTER UE..LATIN SMALL LETTER TURNED G
    Interval{ 0x1D78, 0x1D78 }, // Lm       MODIFIER LETTER CYRILLIC EN
    Interval{ 0x1D79, 0x1D9A }, // L&  [34] LATIN SMALL LETTER INSULAR G..LATIN SMALL LETTER EZH WITH RETROFLEX HOOK
    Interval{ 0x1D9B, 0x1DBF }, // Lm  [37] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL THETA
    Interval{ 0x1E00, 0x1F15 }, // L& [278] LATIN CAPITAL LETTER A WITH RING BELOW..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F18, 0x1F1D }, // L&   [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F20, 0x1F45 }, // L&  [38] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F48, 0x1F4D }, // L&   [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F50, 0x1F57 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F59, 0x1F59 }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA
    Interval{ 0x1F5B, 0x1F5B }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Interval{ 0x1F5D, 0x1F5D }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Interval{ 0x1F5F, 0x1F7D }, // L&  [31] GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI..GREEK SMALL LETTER OMEGA WITH OXIA
    Interval{ 0x1F80, 0x1FB4 }, // L&  [53] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB6, 0x1FBC }, // L&   [7] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Interval{ 0x1FBE, 0x1FBE }, // L&       GREEK PROSGEGRAMMENI
    Interval{ 0x1FC2, 0x1FC4 }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC6, 0x1FCC }, // L&   [7] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Interval{ 0x1FD0, 0x1FD3 }, // L&   [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Interval{ 0x1FD6, 0x1FDB }, // L&   [6] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK CAPITAL LETTER IOTA WITH OXIA
    Interval{ 0x1FE0, 0x1FEC }, // L&  [13] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Interval{ 0x1FF2, 0x1FF4 }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF6, 0x1FFC }, // L&   [7] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Interval{ 0x2071, 0x2071 }, // Lm       SUPERSCRIPT LATIN SMALL LETTER I
    Interval{ 0x207F, 0x207F }, // Lm       SUPERSCRIPT LATIN SMALL LETTER N
    Interval{ 0x2090, 0x209C }, // Lm  [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Interval{ 0x2102, 0x2102 }, // L&       DOUBLE-STRUCK CAPITAL C
    Interval{ 0x2107, 0x2107 }, // L&       EULER CONSTANT
    Interval{ 0x210A, 0x2113 }, // L&  [10] SCRIPT SMALL G..SCRIPT SMALL L
    Interval{ 0x2115, 0x2115 }, // L&       DOUBLE-STRUCK CAPITAL N
    Interval{ 0x2118, 0x2118 }, // Sm       SCRIPT CAPITAL P
    Interval{ 0x2119, 0x211D }, // L&   [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    Interval{ 0x2124, 0x2124 }, // L&       DOUBLE-STRUCK CAPITAL Z
    Interval{ 0x2126, 0x2126 }, // L&       OHM SIGN
    Interval{ 0x2128, 0x2128 }, // L&       BLACK-LETTER CAPITAL Z
    Interval{ 0x212A, 0x212D }, // L&   [4] KELVIN SIGN..BLACK-LETTER CAPITAL C
    Interval{ 0x212E, 0x212E }, // So       ESTIMATED SYMBOL
    Interval{ 0x212F, 0x2134 }, // L&   [6] SCRIPT SMALL E..SCRIPT SMALL O
    Interval{ 0x2135, 0x2138 }, // Lo   [4] ALEF SYMBOL..DALET SYMBOL
    Interval{ 0x2139, 0x2139 }, // L&       INFORMATION SOURCE
    Interval{ 0x213C, 0x213F }, // L&   [4] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK CAPITAL PI
    Interval{ 0x2145, 0x2149 }, // L&   [5] DOUBLE-STRUCK ITALIC CAPITAL D..DOUBLE-STRUCK ITALIC SMALL J
    Interval{ 0x214E, 0x214E }, // L&       TURNED SMALL F
    Interval{ 0x2160, 0x2182 }, // Nl  [35] ROMAN NUMERAL ONE..ROMAN NUMERAL TEN THOUSAND
    Interval{ 0x2183, 0x2184 }, // L&   [2] ROMAN NUMERAL REVERSED ONE HUNDRED..LATIN SMALL LETTER REVERSED C
    Interval{ 0x2185, 0x2188 }, // Nl   [4] ROMAN NUMERAL SIX LATE FORM..ROMAN NUMERAL ONE HUNDRED THOUSAND
    Interval{ 0x2C00, 0x2C7B }, // L& [124] GLAGOLITIC CAPITAL LETTER AZU..LATIN LETTER SMALL CAPITAL TURNED E
    Interval{ 0x2C7C, 0x2C7D }, // Lm   [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    Interval{ 0x2C7E, 0x2CE4 }, // L& [103] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC SYMBOL KAI
    Interval{ 0x2CEB, 0x2CEE }, // L&   [4] COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI..COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF2, 0x2CF3 }, // L&   [2] COPTIC CAPITAL LETTER BOHAIRIC KHEI..COPTIC SMALL LETTER BOHAIRIC KHEI
    Interval{ 0x2D00, 0x2D25 }, // L&  [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Interval{ 0x2D27, 0x2D27 }, // L&       GEORGIAN SMALL LETTER YN
    Interval{ 0x2D2D, 0x2D2D }, // L&       GEORGIAN SMALL LETTER AEN
    Interval{ 0x2D30, 0x2D67 }, // Lo  [56] TIFINAGH LETTER YA..TIFINAGH LETTER YO
    Interval{ 0x2D6F, 0x2D6F }, // Lm       TIFINAGH MODIFIER LETTER LABIALIZATION MARK
    Interval{ 0x2D80, 0x2D96 }, // Lo  [23] ETHIOPIC SYLLABLE LOA..ETHIOPIC SYLLABLE GGWE
    Interval{ 0x2DA0, 0x2DA6 }, // Lo   [7] ETHIOPIC SYLLABLE SSA..ETHIOPIC SYLLABLE SSO
    Interval{ 0x2DA8, 0x2DAE }, // Lo   [7] ETHIOPIC SYLLABLE CCA..ETHIOPIC SYLLABLE CCO
    Interval{ 0x2DB0, 0x2DB6 }, // Lo   [7] ETHIOPIC SYLLABLE ZZA..ETHIOPIC SYLLABLE ZZO
    Interval{ 0x2DB8, 0x2DBE }, // Lo   [7] ETHIOPIC SYLLABLE CCHA..ETHIOPIC SYLLABLE CCHO
    Interval{ 0x2DC0, 0x2DC6 }, // Lo   [7] ETHIOPIC SYLLABLE QYA..ETHIOPIC SYLLABLE QYO
    Interval{ 0x2DC8, 0x2DCE }, // Lo   [7] ETHIOPIC SYLLABLE KYA..ETHIOPIC SYLLABLE KYO
    Interval{ 0x2DD0, 0x2DD6 }, // Lo   [7] ETHIOPIC SYLLABLE XYA..ETHIOPIC SYLLABLE XYO
    Interval{ 0x2DD8, 0x2DDE }, // Lo   [7] ETHIOPIC SYLLABLE GYA..ETHIOPIC SYLLABLE GYO
    Interval{ 0x3005, 0x3005 }, // Lm       IDEOGRAPHIC ITERATION MARK
    Interval{ 0x3006, 0x3006 }, // Lo       IDEOGRAPHIC CLOSING MARK
    Interval{ 0x3007, 0x3007 }, // Nl       IDEOGRAPHIC NUMBER ZERO
    Interval{ 0x3021, 0x3029 }, // Nl   [9] HANGZHOU NUMERAL ONE..HANGZHOU NUMERAL NINE
    Interval{ 0x3031, 0x3035 }, // Lm   [5] VERTICAL KANA REPEAT MARK..VERTICAL KANA REPEAT MARK LOWER HALF
    Interval{ 0x3038, 0x303A }, // Nl   [3] HANGZHOU NUMERAL TEN..HANGZHOU NUMERAL THIRTY
    Interval{ 0x303B, 0x303B }, // Lm       VERTICAL IDEOGRAPHIC ITERATION MARK
    Interval{ 0x303C, 0x303C }, // Lo       MASU MARK
    Interval{ 0x3041, 0x3096 }, // Lo  [86] HIRAGANA LETTER SMALL A..HIRAGANA LETTER SMALL KE
    Interval{ 0x309B, 0x309C }, // Sk   [2] KATAKANA-HIRAGANA VOICED SOUND MARK..KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Interval{ 0x309D, 0x309E }, // Lm   [2] HIRAGANA ITERATION MARK..HIRAGANA VOICED ITERATION MARK
    Interval{ 0x309F, 0x309F }, // Lo       HIRAGANA DIGRAPH YORI
    Interval{ 0x30A1, 0x30FA }, // Lo  [90] KATAKANA LETTER SMALL A..KATAKANA LETTER VO
    Interval{ 0x30FC, 0x30FE }, // Lm   [3] KATAKANA-HIRAGANA PROLONGED SOUND MARK..KATAKANA VOICED ITERATION MARK
    Interval{ 0x30FF, 0x30FF }, // Lo       KATAKANA DIGRAPH KOTO
    Interval{ 0x3105, 0x312F }, // Lo  [43] BOPOMOFO LETTER B..BOPOMOFO LETTER NN
    Interval{ 0x3131, 0x318E }, // Lo  [94] HANGUL LETTER KIYEOK..HANGUL LETTER ARAEAE
    Interval{ 0x31A0, 0x31BF }, // Lo  [32] BOPOMOFO LETTER BU..BOPOMOFO LETTER AH
    Interval{ 0x31F0, 0x31FF }, // Lo  [16] KATAKANA LETTER SMALL KU..KATAKANA LETTER SMALL RO
    Interval{ 0x3400, 0x4DBF }, // Lo [6592] CJK UNIFIED IDEOGRAPH-3400..CJK UNIFIED IDEOGRAPH-4DBF
    Interval{ 0x4E00, 0xA014 }, // Lo [21013] CJK UNIFIED IDEOGRAPH-4E00..YI SYLLABLE E
    Interval{ 0xA015, 0xA015 }, // Lm       YI SYLLABLE WU
    Interval{ 0xA016, 0xA48C }, // Lo [1143] YI SYLLABLE BIT..YI SYLLABLE YYR
    Interval{ 0xA4D0, 0xA4F7 }, // Lo  [40] LISU LETTER BA..LISU LETTER OE
    Interval{ 0xA4F8, 0xA4FD }, // Lm   [6] LISU LETTER TONE MYA TI..LISU LETTER TONE MYA JEU
    Interval{ 0xA500, 0xA60B }, // Lo [268] VAI SYLLABLE EE..VAI SYLLABLE NG
    Interval{ 0xA60C, 0xA60C }, // Lm       VAI SYLLABLE LENGTHENER
    Interval{ 0xA610, 0xA61F }, // Lo  [16] VAI SYLLABLE NDOLE FA..VAI SYMBOL JONG
    Interval{ 0xA62A, 0xA62B }, // Lo   [2] VAI SYLLABLE NDOLE MA..VAI SYLLABLE NDOLE DO
    Interval{ 0xA640, 0xA66D }, // L&  [46] CYRILLIC CAPITAL LETTER ZEMLYA..CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA66E, 0xA66E }, // Lo       CYRILLIC LETTER MULTIOCULAR O
    Interval{ 0xA67F, 0xA67F }, // Lm       CYRILLIC PAYEROK
    Interval{ 0xA680, 0xA69B }, // L&  [28] CYRILLIC CAPITAL LETTER DWE..CYRILLIC SMALL LETTER CROSSED O
    Interval{ 0xA69C, 0xA69D }, // Lm   [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    Interval{ 0xA6A0, 0xA6E5 }, // Lo  [70] BAMUM LETTER A..BAMUM LETTER KI
    Interval{ 0xA6E6, 0xA6EF }, // Nl  [10] BAMUM LETTER MO..BAMUM LETTER KOGHOM
    Interval{ 0xA717, 0xA71F }, // Lm   [9] MODIFIER LETTER DOT VERTICAL BAR..MODIFIER LETTER LOW INVERTED EXCLAMATION MARK
    Interval{ 0xA722, 0xA76F }, // L&  [78] LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF..LATIN SMALL LETTER CON
    Interval{ 0xA770, 0xA770 }, // Lm       MODIFIER LETTER US
    Interval{ 0xA771, 0xA787 }, // L&  [23] LATIN SMALL LETTER DUM..LATIN SMALL LETTER INSULAR T
    Interval{ 0xA788, 0xA788 }, // Lm       MODIFIER LETTER LOW CIRCUMFLEX ACCENT
    Interval{ 0xA78B, 0xA78E }, // L&   [4] LATIN CAPITAL LETTER SALTILLO..LATIN SMALL LETTER L WITH RETROFLEX HOOK AND BELT
    Interval{ 0xA78F, 0xA78F }, // Lo       LATIN LETTER SINOLOGICAL DOT
    Interval{ 0xA790, 0xA7DC }, // L&  [77] LATIN CAPITAL LETTER N WITH DESCENDER..LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Interval{ 0xA7F1, 0xA7F4 }, // Lm   [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    Interval{ 0xA7F5, 0xA7F6 }, // L&   [2] LATIN CAPITAL LETTER REVERSED HALF H..LATIN SMALL LETTER REVERSED HALF H
    Interval{ 0xA7F7, 0xA7F7 }, // Lo       LATIN EPIGRAPHIC LETTER SIDEWAYS I
    Interval{ 0xA7F8, 0xA7F9 }, // Lm   [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    Interval{ 0xA7FA, 0xA7FA }, // L&       LATIN LETTER SMALL CAPITAL TURNED M
    Interval{ 0xA7FB, 0xA801 }, // Lo   [7] LATIN EPIGRAPHIC LETTER REVERSED F..SYLOTI NAGRI LETTER I
    Interval{ 0xA803, 0xA805 }, // Lo   [3] SYLOTI NAGRI LETTER U..SYLOTI NAGRI LETTER O
    Interval{ 0xA807, 0xA80A }, // Lo   [4] SYLOTI NAGRI LETTER KO..SYLOTI NAGRI LETTER GHO
    Interval{ 0xA80C, 0xA822 }, // Lo  [23] SYLOTI NAGRI LETTER CO..SYLOTI NAGRI LETTER HO
    Interval{ 0xA840, 0xA873 }, // Lo  [52] PHAGS-PA LETTER KA..PHAGS-PA LETTER CANDRABINDU
    Interval{ 0xA882, 0xA8B3 }, // Lo  [50] SAURASHTRA LETTER A..SAURASHTRA LETTER LLA
    Interval{ 0xA8F2, 0xA8F7 }, // Lo   [6] DEVANAGARI SIGN SPACING CANDRABINDU..DEVANAGARI SIGN CANDRABINDU AVAGRAHA
    Interval{ 0xA8FB, 0xA8FB }, // Lo       DEVANAGARI HEADSTROKE
    Interval{ 0xA8FD, 0xA8FE }, // Lo   [2] DEVANAGARI JAIN OM..DEVANAGARI LETTER AY
    Interval{ 0xA90A, 0xA925 }, // Lo  [28] KAYAH LI LETTER KA..KAYAH LI LETTER OO
    Interval{ 0xA930, 0xA946 }, // Lo  [23] REJANG LETTER KA..REJANG LETTER A
    Interval{ 0xA960, 0xA97C }, // Lo  [29] HANGUL CHOSEONG TIKEUT-MIEUM..HANGUL CHOSEONG SSANGYEORINHIEUH
    Interval{ 0xA984, 0xA9B2 }, // Lo  [47] JAVANESE LETTER A..JAVANESE LETTER HA
    Interval{ 0xA9CF, 0xA9CF }, // Lm       JAVANESE PANGRANGKEP
    Interval{ 0xA9E0, 0xA9E4 }, // Lo   [5] MYANMAR LETTER SHAN GHA..MYANMAR LETTER SHAN BHA
    Interval{ 0xA9E6, 0xA9E6 }, // Lm       MYANMAR MODIFIER LETTER SHAN REDUPLICATION
    Interval{ 0xA9E7, 0xA9EF }, // Lo   [9] MYANMAR LETTER TAI LAING NYA..MYANMAR LETTER TAI LAING NNA
    Interval{ 0xA9FA, 0xA9FE }, // Lo   [5] MYANMAR LETTER TAI LAING LLA..MYANMAR LETTER TAI LAING BHA
    Interval{ 0xAA00, 0xAA28 }, // Lo  [41] CHAM LETTER A..CHAM LETTER HA
    Interval{ 0xAA40, 0xAA42 }, // Lo   [3] CHAM LETTER FINAL K..CHAM LETTER FINAL NG
    Interval{ 0xAA44, 0xAA4B }, // Lo   [8] CHAM LETTER FINAL CH..CHAM LETTER FINAL SS
    Interval{ 0xAA60, 0xAA6F }, // Lo  [16] MYANMAR LETTER KHAMTI GA..MYANMAR LETTER KHAMTI FA
    Interval{ 0xAA70, 0xAA70 }, // Lm       MYANMAR MODIFIER LETTER KHAMTI REDUPLICATION
    Interval{ 0xAA71, 0xAA76 }, // Lo   [6] MYANMAR LETTER KHAMTI XA..MYANMAR LOGOGRAM KHAMTI HM
    Interval{ 0xAA7A, 0xAA7A }, // Lo       MYANMAR LETTER AITON RA
    Interval{ 0xAA7E, 0xAAAF }, // Lo  [50] MYANMAR LETTER SHWE PALAUNG CHA..TAI VIET LETTER HIGH O
    Interval{ 0xAAB1, 0xAAB1 }, // Lo       TAI VIET VOWEL AA
    Interval{ 0xAAB5, 0xAAB6 }, // Lo   [2] TAI VIET VOWEL E..TAI VIET VOWEL O
    Interval{ 0xAAB9, 0xAABD }, // Lo   [5] TAI VIET VOWEL UEA..TAI VIET VOWEL AN
    Interval{ 0xAAC0, 0xAAC0 }, // Lo       TAI VIET TONE MAI NUENG
    Interval{ 0xAAC2, 0xAAC2 }, // Lo       TAI VIET TONE MAI SONG
    Interval{ 0xAADB, 0xAADC }, // Lo   [2] TAI VIET SYMBOL KON..TAI VIET SYMBOL NUENG
    Interval{ 0xAADD, 0xAADD }, // Lm       TAI VIET SYMBOL SAM
    Interval{ 0xAAE0, 0xAAEA }, // Lo  [11] MEETEI MAYEK LETTER E..MEETEI MAYEK LETTER SSA
    Interval{ 0xAAF2, 0xAAF2 }, // Lo       MEETEI MAYEK ANJI
    Interval{ 0xAAF3, 0xAAF4 }, // Lm   [2] MEETEI MAYEK SYLLABLE REPETITION MARK..MEETEI MAYEK WORD REPETITION MARK
    Interval{ 0xAB01, 0xAB06 }, // Lo   [6] ETHIOPIC SYLLABLE TTHU..ETHIOPIC SYLLABLE TTHO
    Interval{ 0xAB09, 0xAB0E }, // Lo   [6] ETHIOPIC SYLLABLE DDHU..ETHIOPIC SYLLABLE DDHO
    Interval{ 0xAB11, 0xAB16 }, // Lo   [6] ETHIOPIC SYLLABLE DZU..ETHIOPIC SYLLABLE DZO
    Interval{ 0xAB20, 0xAB26 }, // Lo   [7] ETHIOPIC SYLLABLE CCHHA..ETHIOPIC SYLLABLE CCHHO
    Interval{ 0xAB28, 0xAB2E }, // Lo   [7] ETHIOPIC SYLLABLE BBA..ETHIOPIC SYLLABLE BBO
    Interval{ 0xAB30, 0xAB5A }, // L&  [43] LATIN SMALL LETTER BARRED ALPHA..LATIN SMALL LETTER Y WITH SHORT RIGHT LEG
    Interval{ 0xAB5C, 0xAB5F }, // Lm   [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    Interval{ 0xAB60, 0xAB68 }, // L&   [9] LATIN SMALL LETTER SAKHA YAT..LATIN SMALL LETTER TURNED R WITH MIDDLE TILDE
    Interval{ 0xAB69, 0xAB69 }, // Lm       MODIFIER LETTER SMALL TURNED W
    Interval{ 0xAB70, 0xABBF }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xABC0, 0xABE2 }, // Lo  [35] MEETEI MAYEK LETTER KOK..MEETEI MAYEK LETTER I LONSUM
    Interval{ 0xAC00, 0xD7A3 }, // Lo [11172] HANGUL SYLLABLE GA..HANGUL SYLLABLE HIH
    Interval{ 0xD7B0, 0xD7C6 }, // Lo  [23] HANGUL JUNGSEONG O-YEO..HANGUL JUNGSEONG ARAEA-E
    Interval{ 0xD7CB, 0xD7FB }, // Lo  [49] HANGUL JONGSEONG NIEUN-RIEUL..HANGUL JONGSEONG PHIEUPH-THIEUTH
    Interval{ 0xF900, 0xFA6D }, // Lo [366] CJK COMPATIBILITY IDEOGRAPH-F900..CJK COMPATIBILITY IDEOGRAPH-FA6D
    Interval{ 0xFA70, 0xFAD9 }, // Lo [106] CJK COMPATIBILITY IDEOGRAPH-FA70..CJK COMPATIBILITY IDEOGRAPH-FAD9
    Interval{ 0xFB00, 0xFB06 }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFB1D, 0xFB1D }, // Lo       HEBREW LETTER YOD WITH HIRIQ
    Interval{ 0xFB1F, 0xFB28 }, // Lo  [10] HEBREW LIGATURE YIDDISH YOD YOD PATAH..HEBREW LETTER WIDE TAV
    Interval{ 0xFB2A, 0xFB36 }, // Lo  [13] HEBREW LETTER SHIN WITH SHIN DOT..HEBREW LETTER ZAYIN WITH DAGESH
    Interval{ 0xFB38, 0xFB3C }, // Lo   [5] HEBREW LETTER TET WITH DAGESH..HEBREW LETTER LAMED WITH DAGESH
    Interval{ 0xFB3E, 0xFB3E }, // Lo       HEBREW LETTER MEM WITH DAGESH
    Interval{ 0xFB40, 0xFB41 }, // Lo   [2] HEBREW LETTER NUN WITH DAGESH..HEBREW LETTER SAMEKH WITH DAGESH
    Interval{ 0xFB43, 0xFB44 }, // Lo   [2] HEBREW LETTER FINAL PE WITH DAGESH..HEBREW LETTER PE WITH DAGESH
    Interval{ 0xFB46, 0xFBB1 }, // Lo [108] HEBREW LETTER TSADI WITH DAGESH..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE FINAL FORM
    Interval{ 0xFBD3, 0xFD3D }, // Lo [363] ARABIC LETTER NG ISOLATED FORM..ARABIC LIGATURE ALEF WITH FATHATAN ISOLATED FORM
    Interval{ 0xFD50, 0xFD8F }, // Lo  [64] ARABIC LIGATURE TEH WITH JEEM WITH MEEM INITIAL FORM..ARABIC LIGATURE MEEM WITH KHAH WITH MEEM INITIAL FORM
    Interval{ 0xFD92, 0xFDC7 }, // Lo  [54] ARABIC LIGATURE MEEM WITH JEEM WITH KHAH INITIAL FORM..ARABIC LIGATURE NOON WITH JEEM WITH YEH FINAL FORM
    Interval{ 0xFDF0, 0xFDFB }, // Lo  [12] ARABIC LIGATURE SALLA USED AS KORANIC STOP SIGN ISOLATED FORM..ARABIC LIGATURE JALLAJALALOUHOU
    Interval{ 0xFE70, 0xFE74 }, // Lo   [5] ARABIC FATHATAN ISOLATED FORM..ARABIC KASRATAN ISOLATED FORM
    Interval{ 0xFE76, 0xFEFC }, // Lo [135] ARABIC FATHA ISOLATED FORM..ARABIC LIGATURE LAM WITH ALEF FINAL FORM
    Interval{ 0xFF21, 0xFF3A }, // L&  [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Interval{ 0xFF41, 0xFF5A }, // L&  [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Interval{ 0xFF66, 0xFF6F }, // Lo  [10] HALFWIDTH KATAKANA LETTER WO..HALFWIDTH KATAKANA LETTER SMALL TU
    Interval{ 0xFF70, 0xFF70 }, // Lm       HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Interval{ 0xFF71, 0xFF9D }, // Lo  [45] HALFWIDTH KATAKANA LETTER A..HALFWIDTH KATAKANA LETTER N
    Interval{ 0xFF9E, 0xFF9F }, // Lm   [2] HALFWIDTH KATAKANA VOICED SOUND MARK..HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK
    Interval{ 0xFFA0, 0xFFBE }, // Lo  [31] HALFWIDTH HANGUL FILLER..HALFWIDTH HANGUL LETTER HIEUH
    Interval{ 0xFFC2, 0xFFC7 }, // Lo   [6] HALFWIDTH HANGUL LETTER A..HALFWIDTH HANGUL LETTER E
    Interval{ 0xFFCA, 0xFFCF }, // Lo   [6] HALFWIDTH HANGUL LETTER YEO..HALFWIDTH HANGUL LETTER OE
    Interval{ 0xFFD2, 0xFFD7 }, // Lo   [6] HALFWIDTH HANGUL LETTER YO..HALFWIDTH HANGUL LETTER YU
    Interval{ 0xFFDA, 0xFFDC }, // Lo   [3] HALFWIDTH HANGUL LETTER EU..HALFWIDTH HANGUL LETTER I
    Interval{ 0x10000, 0x1000B }, // Lo  [12] LINEAR B SYLLABLE B008 A..LINEAR B SYLLABLE B046 JE
    Interval{ 0x1000D, 0x10026 }, // Lo  [26] LINEAR B SYLLABLE B036 JO..LINEAR B SYLLABLE B032 QO
    Interval{ 0x10028, 0x1003A }, // Lo  [19] LINEAR B SYLLABLE B060 RA..LINEAR B SYLLABLE B042 WO
    Interval{ 0x1003C, 0x1003D }, // Lo   [2] LINEAR B SYLLABLE B017 ZA..LINEAR B SYLLABLE B074 ZE
    Interval{ 0x1003F, 0x1004D }, // Lo  [15] LINEAR B SYLLABLE B020 ZO..LINEAR B SYLLABLE B091 TWO
    Interval{ 0x10050, 0x1005D }, // Lo  [14] LINEAR B SYMBOL B018..LINEAR B SYMBOL B089
    Interval{ 0x10080, 0x100FA }, // Lo [123] LINEAR B IDEOGRAM B100 MAN..LINEAR B IDEOGRAM VESSEL B305
    Interval{ 0x10140, 0x10174 }, // Nl  [53] GREEK ACROPHONIC ATTIC ONE QUARTER..GREEK ACROPHONIC STRATIAN FIFTY MNAS
    Interval{ 0x10280, 0x1029C }, // Lo  [29] LYCIAN LETTER A..LYCIAN LETTER X
    Interval{ 0x102A0, 0x102D0 }, // Lo  [49] CARIAN LETTER A..CARIAN LETTER UUU3
    Interval{ 0x10300, 0x1031F }, // Lo  [32] OLD ITALIC LETTER A..OLD ITALIC LETTER ESS
    Interval{ 0x1032D, 0x10340 }, // Lo  [20] OLD ITALIC LETTER YE..GOTHIC LETTER PAIRTHRA
    Interval{ 0x10341, 0x10341 }, // Nl       GOTHIC LETTER NINETY
    Interval{ 0x10342, 0x10349 }, // Lo   [8] GOTHIC LETTER RAIDA..GOTHIC LETTER OTHAL
    Interval{ 0x1034A, 0x1034A }, // Nl       GOTHIC LETTER NINE HUNDRED
    Interval{ 0x10350, 0x10375 }, // Lo  [38] OLD PERMIC LETTER AN..OLD PERMIC LETTER IA
    Interval{ 0x10380, 0x1039D }, // Lo  [30] UGARITIC LETTER ALPA..UGARITIC LETTER SSU
    Interval{ 0x103A0, 0x103C3 }, // Lo  [36] OLD PERSIAN SIGN A..OLD PERSIAN SIGN HA
    Interval{ 0x103C8, 0x103CF }, // Lo   [8] OLD PERSIAN SIGN AURAMAZDAA..OLD PERSIAN SIGN BUUMISH
    Interval{ 0x103D1, 0x103D5 }, // Nl   [5] OLD PERSIAN NUMBER ONE..OLD PERSIAN NUMBER HUNDRED
    Interval{ 0x10400, 0x1044F }, // L&  [80] DESERET CAPITAL LETTER LONG I..DESERET SMALL LETTER EW
    Interval{ 0x10450, 0x1049D }, // Lo  [78] SHAVIAN LETTER PEEP..OSMANYA LETTER OO
    Interval{ 0x104B0, 0x104D3 }, // L&  [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Interval{ 0x104D8, 0x104FB }, // L&  [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Interval{ 0x10500, 0x10527 }, // Lo  [40] ELBASAN LETTER A..ELBASAN LETTER KHE
    Interval{ 0x10530, 0x10563 }, // Lo  [52] CAUCASIAN ALBANIAN LETTER ALT..CAUCASIAN ALBANIAN LETTER KIW
    Interval{ 0x10570, 0x1057A }, // L&  [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Interval{ 0x1057C, 0x1058A }, // L&  [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Interval{ 0x1058C, 0x10592 }, // L&   [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Interval{ 0x10594, 0x10595 }, // L&   [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Interval{ 0x10597, 0x105A1 }, // L&  [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Interval{ 0x105A3, 0x105B1 }, // L&  [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Interval{ 0x105B3, 0x105B9 }, // L&   [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Interval{ 0x105BB, 0x105BC }, // L&   [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Interval{ 0x105C0, 0x105F3 }, // Lo  [52] TODHRI LETTER A..TODHRI LETTER OO
    Interval{ 0x10600, 0x10736 }, // Lo [311] LINEAR A SIGN AB001..LINEAR A SIGN A664
    Interval{ 0x10740, 0x10755 }, // Lo  [22] LINEAR A SIGN A701 A..LINEAR A SIGN A732 JE
    Interval{ 0x10760, 0x10767 }, // Lo   [8] LINEAR A SIGN A800..LINEAR A SIGN A807
    Interval{ 0x10780, 0x10785 }, // Lm   [6] MODIFIER LETTER SMALL CAPITAL AA..MODIFIER LETTER SMALL B WITH HOOK
    Interval{ 0x10787, 0x107B0 }, // Lm  [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Interval{ 0x107B2, 0x107BA }, // Lm   [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Interval{ 0x10800, 0x10805 }, // Lo   [6] CYPRIOT SYLLABLE A..CYPRIOT SYLLABLE JA
    Interval{ 0x10808, 0x10808 }, // Lo       CYPRIOT SYLLABLE JO
    Interval{ 0x1080A, 0x10835 }, // Lo  [44] CYPRIOT SYLLABLE KA..CYPRIOT SYLLABLE WO
    Interval{ 0x10837, 0x10838 }, // Lo   [2] CYPRIOT SYLLABLE XA..CYPRIOT SYLLABLE XE
    Interval{ 0x1083C, 0x1083C }, // Lo       CYPRIOT SYLLABLE ZA
    Interval{ 0x1083F, 0x10855 }, // Lo  [23] CYPRIOT SYLLABLE ZO..IMPERIAL ARAMAIC LETTER TAW
    Interval{ 0x10860, 0x10876 }, // Lo  [23] PALMYRENE LETTER ALEPH..PALMYRENE LETTER TAW
    Interval{ 0x10880, 0x1089E }, // Lo  [31] NABATAEAN LETTER FINAL ALEPH..NABATAEAN LETTER TAW
    Interval{ 0x108E0, 0x108F2 }, // Lo  [19] HATRAN LETTER ALEPH..HATRAN LETTER QOPH
    Interval{ 0x108F4, 0x108F5 }, // Lo   [2] HATRAN LETTER SHIN..HATRAN LETTER TAW
    Interval{ 0x10900, 0x10915 }, // Lo  [22] PHOENICIAN LETTER ALF..PHOENICIAN LETTER TAU
    Interval{ 0x10920, 0x10939 }, // Lo  [26] LYDIAN LETTER A..LYDIAN LETTER C
    Interval{ 0x10940, 0x10959 }, // Lo  [26] SIDETIC LETTER N01..SIDETIC LETTER N26
    Interval{ 0x10980, 0x109B7 }, // Lo  [56] MEROITIC HIEROGLYPHIC LETTER A..MEROITIC CURSIVE LETTER DA
    Interval{ 0x109BE, 0x109BF }, // Lo   [2] MEROITIC CURSIVE LOGOGRAM RMT..MEROITIC CURSIVE LOGOGRAM IMN
    Interval{ 0x10A00, 0x10A00 }, // Lo       KHAROSHTHI LETTER A
    Interval{ 0x10A10, 0x10A13 }, // Lo   [4] KHAROSHTHI LETTER KA..KHAROSHTHI LETTER GHA
    Interval{ 0x10A15, 0x10A17 }, // Lo   [3] KHAROSHTHI LETTER CA..KHAROSHTHI LETTER JA
    Interval{ 0x10A19, 0x10A35 }, // Lo  [29] KHAROSHTHI LETTER NYA..KHAROSHTHI LETTER VHA
    Interval{ 0x10A60, 0x10A7C }, // Lo  [29] OLD SOUTH ARABIAN LETTER HE..OLD SOUTH ARABIAN LETTER THETH
    Interval{ 0x10A80, 0x10A9C }, // Lo  [29] OLD NORTH ARABIAN LETTER HEH..OLD NORTH ARABIAN LETTER ZAH
    Interval{ 0x10AC0, 0x10AC7 }, // Lo   [8] MANICHAEAN LETTER ALEPH..MANICHAEAN LETTER WAW
    Interval{ 0x10AC9, 0x10AE4 }, // Lo  [28] MANICHAEAN LETTER ZAYIN..MANICHAEAN LETTER TAW
    Interval{ 0x10B00, 0x10B35 }, // Lo  [54] AVESTAN LETTER A..AVESTAN LETTER HE
    Interval{ 0x10B40, 0x10B55 }, // Lo  [22] INSCRIPTIONAL PARTHIAN LETTER ALEPH..INSCRIPTIONAL PARTHIAN LETTER TAW
    Interval{ 0x10B60, 0x10B72 }, // Lo  [19] INSCRIPTIONAL PAHLAVI LETTER ALEPH..INSCRIPTIONAL PAHLAVI LETTER TAW
    Interval{ 0x10B80, 0x10B91 }, // Lo  [18] PSALTER PAHLAVI LETTER ALEPH..PSALTER PAHLAVI LETTER TAW
    Interval{ 0x10C00, 0x10C48 }, // Lo  [73] OLD TURKIC LETTER ORKHON A..OLD TURKIC LETTER ORKHON BASH
    Interval{ 0x10C80, 0x10CB2 }, // L&  [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Interval{ 0x10CC0, 0x10CF2 }, // L&  [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Interval{ 0x10D00, 0x10D23 }, // Lo  [36] HANIFI ROHINGYA LETTER A..HANIFI ROHINGYA MARK NA KHONNA
    Interval{ 0x10D4A, 0x10D4D }, // Lo   [4] GARAY VOWEL SIGN A..GARAY VOWEL SIGN EE
    Interval{ 0x10D4E, 0x10D4E }, // Lm       GARAY VOWEL LENGTH MARK
    Interval{ 0x10D4F, 0x10D4F }, // Lo       GARAY SUKUN
    Interval{ 0x10D50, 0x10D65 }, // L&  [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Interval{ 0x10D6F, 0x10D6F }, // Lm       GARAY REDUPLICATION MARK
    Interval{ 0x10D70, 0x10D85 }, // L&  [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Interval{ 0x10E80, 0x10EA9 }, // Lo  [42] YEZIDI LETTER ELIF..YEZIDI LETTER ET
    Interval{ 0x10EB0, 0x10EB1 }, // Lo   [2] YEZIDI LETTER LAM WITH DOT ABOVE..YEZIDI LETTER YOT WITH CIRCUMFLEX ABOVE
    Interval{ 0x10EC2, 0x10EC4 }, // Lo   [3] ARABIC LETTER DAL WITH TWO DOTS VERTICALLY BELOW..ARABIC LETTER KAF WITH TWO DOTS VERTICALLY BELOW
    Interval{ 0x10EC5, 0x10EC5 }, // Lm       ARABIC SMALL YEH BARREE WITH TWO DOTS BELOW
    Interval{ 0x10EC6, 0x10EC7 }, // Lo   [2] ARABIC LETTER THIN NOON..ARABIC LETTER YEH WITH FOUR DOTS BELOW
    Interval{ 0x10F00, 0x10F1C }, // Lo  [29] OLD SOGDIAN LETTER ALEPH..OLD SOGDIAN LETTER FINAL TAW WITH VERTICAL TAIL
    Interval{ 0x10F27, 0x10F27 }, // Lo       OLD SOGDIAN LIGATURE AYIN-DALETH
    Interval{ 0x10F30, 0x10F45 }, // Lo  [22] SOGDIAN LETTER ALEPH..SOGDIAN INDEPENDENT SHIN
    Interval{ 0x10F70, 0x10F81 }, // Lo  [18] OLD UYGHUR LETTER ALEPH..OLD UYGHUR LETTER LESH
    Interval{ 0x10FB0, 0x10FC4 }, // Lo  [21] CHORASMIAN LETTER ALEPH..CHORASMIAN LETTER TAW
    Interval{ 0x10FE0, 0x10FF6 }, // Lo  [23] ELYMAIC LETTER ALEPH..ELYMAIC LIGATURE ZAYIN-YODH
    Interval{ 0x11003, 0x11037 }, // Lo  [53] BRAHMI SIGN JIHVAMULIYA..BRAHMI LETTER OLD TAMIL NNNA
    Interval{ 0x11071, 0x11072 }, // Lo   [2] BRAHMI LETTER OLD TAMIL SHORT E..BRAHMI LETTER OLD TAMIL SHORT O
    Interval{ 0x11075, 0x11075 }, // Lo       BRAHMI LETTER OLD TAMIL LLA
    Interval{ 0x11083, 0x110AF }, // Lo  [45] KAITHI LETTER A..KAITHI LETTER HA
    Interval{ 0x110D0, 0x110E8 }, // Lo  [25] SORA SOMPENG LETTER SAH..SORA SOMPENG LETTER MAE
    Interval{ 0x11103, 0x11126 }, // Lo  [36] CHAKMA LETTER AA..CHAKMA LETTER HAA
    Interval{ 0x11144, 0x11144 }, // Lo       CHAKMA LETTER LHAA
    Interval{ 0x11147, 0x11147 }, // Lo       CHAKMA LETTER VAA
    Interval{ 0x11150, 0x11172 }, // Lo  [35] MAHAJANI LETTER A..MAHAJANI LETTER RRA
    Interval{ 0x11176, 0x11176 }, // Lo       MAHAJANI LIGATURE SHRI
    Interval{ 0x11183, 0x111B2 }, // Lo  [48] SHARADA LETTER A..SHARADA LETTER HA
    Interval{ 0x111C1, 0x111C4 }, // Lo   [4] SHARADA SIGN AVAGRAHA..SHARADA OM
    Interval{ 0x111DA, 0x111DA }, // Lo       SHARADA EKAM
    Interval{ 0x111DC, 0x111DC }, // Lo       SHARADA HEADSTROKE
    Interval{ 0x11200, 0x11211 }, // Lo  [18] KHOJKI LETTER A..KHOJKI LETTER JJA
    Interval{ 0x11213, 0x1122B }, // Lo  [25] KHOJKI LETTER NYA..KHOJKI LETTER LLA
    Interval{ 0x1123F, 0x11240 }, // Lo   [2] KHOJKI LETTER QA..KHOJKI LETTER SHORT I
    Interval{ 0x11280, 0x11286 }, // Lo   [7] MULTANI LETTER A..MULTANI LETTER GA
    Interval{ 0x11288, 0x11288 }, // Lo       MULTANI LETTER GHA
    Interval{ 0x1128A, 0x1128D }, // Lo   [4] MULTANI LETTER CA..MULTANI LETTER JJA
    Interval{ 0x1128F, 0x1129D }, // Lo  [15] MULTANI LETTER NYA..MULTANI LETTER BA
    Interval{ 0x1129F, 0x112A8 }, // Lo  [10] MULTANI LETTER BHA..MULTANI LETTER RHA
    Interval{ 0x112B0, 0x112DE }, // Lo  [47] KHUDAWADI LETTER A..KHUDAWADI LETTER HA
    Interval{ 0x11305, 0x1130C }, // Lo   [8] GRANTHA LETTER A..GRANTHA LETTER VOCALIC L
    Interval{ 0x1130F, 0x11310 }, // Lo   [2] GRANTHA LETTER EE..GRANTHA LETTER AI
    Interval{ 0x11313, 0x11328 }, // Lo  [22] GRANTHA LETTER OO..GRANTHA LETTER NA
    Interval{ 0x1132A, 0x11330 }, // Lo   [7] GRANTHA LETTER PA..GRANTHA LETTER RA
    Interval{ 0x11332, 0x11333 }, // Lo   [2] GRANTHA LETTER LA..GRANTHA LETTER LLA
    Interval{ 0x11335, 0x11339 }, // Lo   [5] GRANTHA LETTER VA..GRANTHA LETTER HA
    Interval{ 0x1133D, 0x1133D }, // Lo       GRANTHA SIGN AVAGRAHA
    Interval{ 0x11350, 0x11350 }, // Lo       GRANTHA OM
    Interval{ 0x1135D, 0x11361 }, // Lo   [5] GRANTHA SIGN PLUTA..GRANTHA LETTER VOCALIC LL
    Interval{ 0x11380, 0x11389 }, // Lo  [10] TULU-TIGALARI LETTER A..TULU-TIGALARI LETTER VOCALIC LL
    Interval{ 0x1138B, 0x1138B }, // Lo       TULU-TIGALARI LETTER EE
    Interval{ 0x1138E, 0x1138E }, // Lo       TULU-TIGALARI LETTER AI
    Interval{ 0x11390, 0x113B5 }, // Lo  [38] TULU-TIGALARI LETTER OO..TULU-TIGALARI LETTER LLLA
    Interval{ 0x113B7, 0x113B7 }, // Lo       TULU-TIGALARI SIGN AVAGRAHA
    Interval{ 0x113D1, 0x113D1 }, // Lo       TULU-TIGALARI REPHA
    Interval{ 0x113D3, 0x113D3 }, // Lo       TULU-TIGALARI SIGN PLUTA
    Interval{ 0x11400, 0x11434 }, // Lo  [53] NEWA LETTER A..NEWA LETTER HA
    Interval{ 0x11447, 0x1144A }, // Lo   [4] NEWA SIGN AVAGRAHA..NEWA SIDDHI
    Interval{ 0x1145F, 0x11461 }, // Lo   [3] NEWA LETTER VEDIC ANUSVARA..NEWA SIGN UPADHMANIYA
    Interval{ 0x11480, 0x114AF }, // Lo  [48] TIRHUTA ANJI..TIRHUTA LETTER HA
    Interval{ 0x114C4, 0x114C5 }, // Lo   [2] TIRHUTA SIGN AVAGRAHA..TIRHUTA GVANG
    Interval{ 0x114C7, 0x114C7 }, // Lo       TIRHUTA OM
    Interval{ 0x11580, 0x115AE }, // Lo  [47] SIDDHAM LETTER A..SIDDHAM LETTER HA
    Interval{ 0x115D8, 0x115DB }, // Lo   [4] SIDDHAM LETTER THREE-CIRCLE ALTERNATE I..SIDDHAM LETTER ALTERNATE U
    Interval{ 0x11600, 0x1162F }, // Lo  [48] MODI LETTER A..MODI LETTER LLA
    Interval{ 0x11644, 0x11644 }, // Lo       MODI SIGN HUVA
    Interval{ 0x11680, 0x116AA }, // Lo  [43] TAKRI LETTER A..TAKRI LETTER RRA
    Interval{ 0x116B8, 0x116B8 }, // Lo       TAKRI LETTER ARCHAIC KHA
    Interval{ 0x11700, 0x1171A }, // Lo  [27] AHOM LETTER KA..AHOM LETTER ALTERNATE BA
    Interval{ 0x11740, 0x11746 }, // Lo   [7] AHOM LETTER CA..AHOM LETTER LLA
    Interval{ 0x11800, 0x1182B }, // Lo  [44] DOGRA LETTER A..DOGRA LETTER RRA
    Interval{ 0x118A0, 0x118DF }, // L&  [64] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Interval{ 0x118FF, 0x11906 }, // Lo   [8] WARANG CITI OM..DIVES AKURU LETTER E
    Interval{ 0x11909, 0x11909 }, // Lo       DIVES AKURU LETTER O
    Interval{ 0x1190C, 0x11913 }, // Lo   [8] DIVES AKURU LETTER KA..DIVES AKURU LETTER JA
    Interval{ 0x11915, 0x11916 }, // Lo   [2] DIVES AKURU LETTER NYA..DIVES AKURU LETTER TTA
    Interval{ 0x11918, 0x1192F }, // Lo  [24] DIVES AKURU LETTER DDA..DIVES AKURU LETTER ZA
    Interval{ 0x1193F, 0x1193F }, // Lo       DIVES AKURU PREFIXED NASAL SIGN
    Interval{ 0x11941, 0x11941 }, // Lo       DIVES AKURU INITIAL RA
    Interval{ 0x119A0, 0x119A7 }, // Lo   [8] NANDINAGARI LETTER A..NANDINAGARI LETTER VOCALIC RR
    Interval{ 0x119AA, 0x119D0 }, // Lo  [39] NANDINAGARI LETTER E..NANDINAGARI LETTER RRA
    Interval{ 0x119E1, 0x119E1 }, // Lo       NANDINAGARI SIGN AVAGRAHA
    Interval{ 0x119E3, 0x119E3 }, // Lo       NANDINAGARI HEADSTROKE
    Interval{ 0x11A00, 0x11A00 }, // Lo       ZANABAZAR SQUARE LETTER A
    Interval{ 0x11A0B, 0x11A32 }, // Lo  [40] ZANABAZAR SQUARE LETTER KA..ZANABAZAR SQUARE LETTER KSSA
    Interval{ 0x11A3A, 0x11A3A }, // Lo       ZANABAZAR SQUARE CLUSTER-INITIAL LETTER RA
    Interval{ 0x11A50, 0x11A50 }, // Lo       SOYOMBO LETTER A
    Interval{ 0x11A5C, 0x11A89 }, // Lo  [46] SOYOMBO LETTER KA..SOYOMBO CLUSTER-INITIAL LETTER SA
    Interval{ 0x11A9D, 0x11A9D }, // Lo       SOYOMBO MARK PLUTA
    Interval{ 0x11AB0, 0x11AF8 }, // Lo  [73] CANADIAN SYLLABICS NATTILIK HI..PAU CIN HAU GLOTTAL STOP FINAL
    Interval{ 0x11BC0, 0x11BE0 }, // Lo  [33] SUNUWAR LETTER DEVI..SUNUWAR LETTER KLOKO
    Interval{ 0x11C00, 0x11C08 }, // Lo   [9] BHAIKSUKI LETTER A..BHAIKSUKI LETTER VOCALIC L
    Interval{ 0x11C0A, 0x11C2E }, // Lo  [37] BHAIKSUKI LETTER E..BHAIKSUKI LETTER HA
    Interval{ 0x11C40, 0x11C40 }, // Lo       BHAIKSUKI SIGN AVAGRAHA
    Interval{ 0x11C72, 0x11C8F }, // Lo  [30] MARCHEN LETTER KA..MARCHEN LETTER A
    Interval{ 0x11D00, 0x11D06 }, // Lo   [7] MASARAM GONDI LETTER A..MASARAM GONDI LETTER E
    Interval{ 0x11D08, 0x11D09 }, // Lo   [2] MASARAM GONDI LETTER AI..MASARAM GONDI LETTER O
    Interval{ 0x11D0B, 0x11D30 }, // Lo  [38] MASARAM GONDI LETTER AU..MASARAM GONDI LETTER TRA
    Interval{ 0x11D46, 0x11D46 }, // Lo       MASARAM GONDI REPHA
    Interval{ 0x11D60, 0x11D65 }, // Lo   [6] GUNJALA GONDI LETTER A..GUNJALA GONDI LETTER UU
    Interval{ 0x11D67, 0x11D68 }, // Lo   [2] GUNJALA GONDI LETTER EE..GUNJALA GONDI LETTER AI
    Interval{ 0x11D6A, 0x11D89 }, // Lo  [32] GUNJALA GONDI LETTER OO..GUNJALA GONDI LETTER SA
    Interval{ 0x11D98, 0x11D98 }, // Lo       GUNJALA GONDI OM
    Interval{ 0x11DB0, 0x11DD8 }, // Lo  [41] TOLONG SIKI LETTER I..TOLONG SIKI LETTER RRH
    Interval{ 0x11DD9, 0x11DD9 }, // Lm       TOLONG SIKI SIGN SELA
    Interval{ 0x11DDA, 0x11DDB }, // Lo   [2] TOLONG SIKI SIGN HECAKA..TOLONG SIKI UNGGA
    Interval{ 0x11EE0, 0x11EF2 }, // Lo  [19] MAKASAR LETTER KA..MAKASAR ANGKA
    Interval{ 0x11F02, 0x11F02 }, // Lo       KAWI SIGN REPHA
    Interval{ 0x11F04, 0x11F10 }, // Lo  [13] KAWI LETTER A..KAWI LETTER O
    Interval{ 0x11F12, 0x11F33 }, // Lo  [34] KAWI LETTER KA..KAWI LETTER JNYA
    Interval{ 0x11FB0, 0x11FB0 }, // Lo       LISU LETTER YHA
    Interval{ 0x12000, 0x12399 }, // Lo [922] CUNEIFORM SIGN A..CUNEIFORM SIGN U U
    Interval{ 0x12400, 0x1246E }, // Nl [111] CUNEIFORM NUMERIC SIGN TWO ASH..CUNEIFORM NUMERIC SIGN NINE U VARIANT FORM
    Interval{ 0x12480, 0x12543 }, // Lo [196] CUNEIFORM SIGN AB TIMES NUN TENU..CUNEIFORM SIGN ZU5 TIMES THREE DISH TENU
    Interval{ 0x12F90, 0x12FF0 }, // Lo  [97] CYPRO-MINOAN SIGN CM001..CYPRO-MINOAN SIGN CM114
    Interval{ 0x13000, 0x1342F }, // Lo [1072] EGYPTIAN HIEROGLYPH A001..EGYPTIAN HIEROGLYPH V011D
    Interval{ 0x13441, 0x13446 }, // Lo   [6] EGYPTIAN HIEROGLYPH FULL BLANK..EGYPTIAN HIEROGLYPH WIDE LOST SIGN
    Interval{ 0x13460, 0x143FA }, // Lo [3995] EGYPTIAN HIEROGLYPH-13460..EGYPTIAN HIEROGLYPH-143FA
    Interval{ 0x14400, 0x14646 }, // Lo [583] ANATOLIAN HIEROGLYPH A001..ANATOLIAN HIEROGLYPH A530
    Interval{ 0x16100, 0x1611D }, // Lo  [30] GURUNG KHEMA LETTER A..GURUNG KHEMA LETTER SA
    Interval{ 0x16800, 0x16A38 }, // Lo [569] BAMUM LETTER PHASE-A NGKUE MFON..BAMUM LETTER PHASE-F VUEQ
    Interval{ 0x16A40, 0x16A5E }, // Lo  [31] MRO LETTER TA..MRO LETTER TEK
    Interval{ 0x16A70, 0x16ABE }, // Lo  [79] TANGSA LETTER OZ..TANGSA LETTER ZA
    Interval{ 0x16AD0, 0x16AED }, // Lo  [30] BASSA VAH LETTER ENNI..BASSA VAH LETTER I
    Interval{ 0x16B00, 0x16B2F }, // Lo  [48] PAHAWH HMONG VOWEL KEEB..PAHAWH HMONG CONSONANT CAU
    Interval{ 0x16B40, 0x16B43 }, // Lm   [4] PAHAWH HMONG SIGN VOS SEEV..PAHAWH HMONG SIGN IB YAM
    Interval{ 0x16B63, 0x16B77 }, // Lo  [21] PAHAWH HMONG SIGN VOS LUB..PAHAWH HMONG SIGN CIM NRES TOS
    Interval{ 0x16B7D, 0x16B8F }, // Lo  [19] PAHAWH HMONG CLAN SIGN TSHEEJ..PAHAWH HMONG CLAN SIGN VWJ
    Interval{ 0x16D40, 0x16D42 }, // Lm   [3] KIRAT RAI SIGN ANUSVARA..KIRAT RAI SIGN VISARGA
    Interval{ 0x16D43, 0x16D6A }, // Lo  [40] KIRAT RAI LETTER A..KIRAT RAI VOWEL SIGN AU
    Interval{ 0x16D6B, 0x16D6C }, // Lm   [2] KIRAT RAI SIGN VIRAMA..KIRAT RAI SIGN SAAT
    Interval{ 0x16E40, 0x16E7F }, // L&  [64] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Interval{ 0x16EA0, 0x16EB8 }, // L&  [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Interval{ 0x16EBB, 0x16ED3 }, // L&  [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Interval{ 0x16F00, 0x16F4A }, // Lo  [75] MIAO LETTER PA..MIAO LETTER RTE
    Interval{ 0x16F50, 0x16F50 }, // Lo       MIAO LETTER NASALIZATION
    Interval{ 0x16F93, 0x16F9F }, // Lm  [13] MIAO LETTER TONE-2..MIAO LETTER REFORMED TONE-8
    Interval{ 0x16FE0, 0x16FE1 }, // Lm   [2] TANGUT ITERATION MARK..NUSHU ITERATION MARK
    Interval{ 0x16FE3, 0x16FE3 }, // Lm       OLD CHINESE ITERATION MARK
    Interval{ 0x16FF2, 0x16FF3 }, // Lm   [2] CHINESE SMALL SIMPLIFIED ER..CHINESE SMALL TRADITIONAL ER
    Interval{ 0x16FF4, 0x16FF6 }, // Nl   [3] YANGQIN SIGN SLOW ONE BEAT..YANGQIN SIGN SLOW TWO BEATS
    Interval{ 0x17000, 0x18CD5 }, // Lo [7382] TANGUT IDEOGRAPH-17000..KHITAN SMALL SCRIPT CHARACTER-18CD5
    Interval{ 0x18CFF, 0x18D1E }, // Lo  [32] KHITAN SMALL SCRIPT CHARACTER-18CFF..TANGUT IDEOGRAPH-18D1E
    Interval{ 0x18D80, 0x18DF2 }, // Lo [115] TANGUT COMPONENT-769..TANGUT COMPONENT-883
    Interval{ 0x1AFF0, 0x1AFF3 }, // Lm   [4] KATAKANA LETTER MINNAN TONE-2..KATAKANA LETTER MINNAN TONE-5
    Interval{ 0x1AFF5, 0x1AFFB }, // Lm   [7] KATAKANA LETTER MINNAN TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-5
    Interval{ 0x1AFFD, 0x1AFFE }, // Lm   [2] KATAKANA LETTER MINNAN NASALIZED TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-8
    Interval{ 0x1B000, 0x1B122 }, // Lo [291] KATAKANA LETTER ARCHAIC E..KATAKANA LETTER ARCHAIC WU
    Interval{ 0x1B132, 0x1B132 }, // Lo       HIRAGANA LETTER SMALL KO
    Interval{ 0x1B150, 0x1B152 }, // Lo   [3] HIRAGANA LETTER SMALL WI..HIRAGANA LETTER SMALL WO
    Interval{ 0x1B155, 0x1B155 }, // Lo       KATAKANA LETTER SMALL KO
    Interval{ 0x1B164, 0x1B167 }, // Lo   [4] KATAKANA LETTER SMALL WI..KATAKANA LETTER SMALL N
    Interval{ 0x1B170, 0x1B2FB }, // Lo [396] NUSHU CHARACTER-1B170..NUSHU CHARACTER-1B2FB
    Interval{ 0x1BC00, 0x1BC6A }, // Lo [107] DUPLOYAN LETTER H..DUPLOYAN LETTER VOCALIC M
    Interval{ 0x1BC70, 0x1BC7C }, // Lo  [13] DUPLOYAN AFFIX LEFT HORIZONTAL SECANT..DUPLOYAN AFFIX ATTACHED TANGENT HOOK
    Interval{ 0x1BC80, 0x1BC88 }, // Lo   [9] DUPLOYAN AFFIX HIGH ACUTE..DUPLOYAN AFFIX HIGH VERTICAL
    Interval{ 0x1BC90, 0x1BC99 }, // Lo  [10] DUPLOYAN AFFIX LOW ACUTE..DUPLOYAN AFFIX LOW ARROW
    Interval{ 0x1D400, 0x1D454 }, // L&  [85] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL ITALIC SMALL G
    Interval{ 0x1D456, 0x1D49C }, // L&  [71] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL SCRIPT CAPITAL A
    Interval{ 0x1D49E, 0x1D49F }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Interval{ 0x1D4A2, 0x1D4A2 }, // L&       MATHEMATICAL SCRIPT CAPITAL G
    Interval{ 0x1D4A5, 0x1D4A6 }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Interval{ 0x1D4A9, 0x1D4AC }, // L&   [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Interval{ 0x1D4AE, 0x1D4B9 }, // L&  [12] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT SMALL D
    Interval{ 0x1D4BB, 0x1D4BB }, // L&       MATHEMATICAL SCRIPT SMALL F
    Interval{ 0x1D4BD, 0x1D4C3 }, // L&   [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Interval{ 0x1D4C5, 0x1D505 }, // L&  [65] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL FRAKTUR CAPITAL B
    Interval{ 0x1D507, 0x1D50A }, // L&   [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Interval{ 0x1D50D, 0x1D514 }, // L&   [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Interval{ 0x1D516, 0x1D51C }, // L&   [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Interval{ 0x1D51E, 0x1D539 }, // L&  [28] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Interval{ 0x1D53B, 0x1D53E }, // L&   [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Interval{ 0x1D540, 0x1D544 }, // L&   [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Interval{ 0x1D546, 0x1D546 }, // L&       MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Interval{ 0x1D54A, 0x1D550 }, // L&   [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Interval{ 0x1D552, 0x1D6A5 }, // L& [340] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    Interval{ 0x1D6A8, 0x1D6C0 }, // L&  [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    Interval{ 0x1D6C2, 0x1D6DA }, // L&  [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    Interval{ 0x1D6DC, 0x1D6FA }, // L&  [31] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL ITALIC CAPITAL OMEGA
    Interval{ 0x1D6FC, 0x1D714 }, // L&  [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    Interval{ 0x1D716, 0x1D734 }, // L&  [31] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D736, 0x1D74E }, // L&  [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D750, 0x1D76E }, // L&  [31] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    Interval{ 0x1D770, 0x1D788 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    Interval{ 0x1D78A, 0x1D7A8 }, // L&  [31] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D7AA, 0x1D7C2 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D7C4, 0x1D7CB }, // L&   [8] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD SMALL DIGAMMA
    Interval{ 0x1DF00, 0x1DF09 }, // L&  [10] LATIN SMALL LETTER FENG DIGRAPH WITH TRILL..LATIN SMALL LETTER T WITH HOOK AND RETROFLEX HOOK
    Interval{ 0x1DF0A, 0x1DF0A }, // Lo       LATIN LETTER RETROFLEX CLICK WITH RETROFLEX HOOK
    Interval{ 0x1DF0B, 0x1DF1E }, // L&  [20] LATIN SMALL LETTER ESH WITH DOUBLE BAR..LATIN SMALL LETTER S WITH CURL
    Interval{ 0x1DF25, 0x1DF2A }, // L&   [6] LATIN SMALL LETTER D WITH MID-HEIGHT LEFT HOOK..LATIN SMALL LETTER T WITH MID-HEIGHT LEFT HOOK
    Interval{ 0x1E030, 0x1E06D }, // Lm  [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Interval{ 0x1E100, 0x1E12C }, // Lo  [45] NYIAKENG PUACHUE HMONG LETTER MA..NYIAKENG PUACHUE HMONG LETTER W
    Interval{ 0x1E137, 0x1E13D }, // Lm   [7] NYIAKENG PUACHUE HMONG SIGN FOR PERSON..NYIAKENG PUACHUE HMONG SYLLABLE LENGTHENER
    Interval{ 0x1E14E, 0x1E14E }, // Lo       NYIAKENG PUACHUE HMONG LOGOGRAM NYAJ
    Interval{ 0x1E290, 0x1E2AD }, // Lo  [30] TOTO LETTER PA..TOTO LETTER A
    Interval{ 0x1E2C0, 0x1E2EB }, // Lo  [44] WANCHO LETTER AA..WANCHO LETTER YIH
    Interval{ 0x1E4D0, 0x1E4EA }, // Lo  [27] NAG MUNDARI LETTER O..NAG MUNDARI LETTER ELL
    Interval{ 0x1E4EB, 0x1E4EB }, // Lm       NAG MUNDARI SIGN OJOD
    Interval{ 0x1E5D0, 0x1E5ED }, // Lo  [30] OL ONAL LETTER O..OL ONAL LETTER EG
    Interval{ 0x1E5F0, 0x1E5F0 }, // Lo       OL ONAL SIGN HODDOND
    Interval{ 0x1E6C0, 0x1E6DE }, // Lo  [31] TAI YO LETTER LOW KO..TAI YO LETTER HIGH KVO
    Interval{ 0x1E6E0, 0x1E6E2 }, // Lo   [3] TAI YO LETTER AA..TAI YO LETTER UE
    Interval{ 0x1E6E4, 0x1E6E5 }, // Lo   [2] TAI YO LETTER U..TAI YO LETTER AE
    Interval{ 0x1E6E7, 0x1E6ED }, // Lo   [7] TAI YO LETTER O..TAI YO LETTER AUE
    Interval{ 0x1E6F0, 0x1E6F4 }, // Lo   [5] TAI YO LETTER AN..TAI YO LETTER AP
    Interval{ 0x1E6FE, 0x1E6FE }, // Lo       TAI YO SYMBOL MUEANG
    Interval{ 0x1E6FF, 0x1E6FF }, // Lm       TAI YO XAM LAI
    Interval{ 0x1E7E0, 0x1E7E6 }, // Lo   [7] ETHIOPIC SYLLABLE HHYA..ETHIOPIC SYLLABLE HHYO
    Interval{ 0x1E7E8, 0x1E7EB }, // Lo   [4] ETHIOPIC SYLLABLE GURAGE HHWA..ETHIOPIC SYLLABLE HHWE
    Interval{ 0x1E7ED, 0x1E7EE }, // Lo   [2] ETHIOPIC SYLLABLE GURAGE MWI..ETHIOPIC SYLLABLE GURAGE MWEE
    Interval{ 0x1E7F0, 0x1E7FE }, // Lo  [15] ETHIOPIC SYLLABLE GURAGE QWI..ETHIOPIC SYLLABLE GURAGE PWEE
    Interval{ 0x1E800, 0x1E8C4 }, // Lo [197] MENDE KIKAKUI SYLLABLE M001 KI..MENDE KIKAKUI SYLLABLE M060 NYON
    Interval{ 0x1E900, 0x1E943 }, // L&  [68] ADLAM CAPITAL LETTER ALIF..ADLAM SMALL LETTER SHA
    Interval{ 0x1E94B, 0x1E94B }, // Lm       ADLAM NASALIZATION MARK
    Interval{ 0x1EE00, 0x1EE03 }, // Lo   [4] ARABIC MATHEMATICAL ALEF..ARABIC MATHEMATICAL DAL
    Interval{ 0x1EE05, 0x1EE1F }, // Lo  [27] ARABIC MATHEMATICAL WAW..ARABIC MATHEMATICAL DOTLESS QAF
    Interval{ 0x1EE21, 0x1EE22 }, // Lo   [2] ARABIC MATHEMATICAL INITIAL BEH..ARABIC MATHEMATICAL INITIAL JEEM
    Interval{ 0x1EE24, 0x1EE24 }, // Lo       ARABIC MATHEMATICAL INITIAL HEH
    Interval{ 0x1EE27, 0x1EE27 }, // Lo       ARABIC MATHEMATICAL INITIAL HAH
    Interval{ 0x1EE29, 0x1EE32 }, // Lo  [10] ARABIC MATHEMATICAL INITIAL YEH..ARABIC MATHEMATICAL INITIAL QAF
    Interval{ 0x1EE34, 0x1EE37 }, // Lo   [4] ARABIC MATHEMATICAL INITIAL SHEEN..ARABIC MATHEMATICAL INITIAL KHAH
    Interval{ 0x1EE39, 0x1EE39 }, // Lo       ARABIC MATHEMATICAL INITIAL DAD
    Interval{ 0x1EE3B, 0x1EE3B }, // Lo       ARABIC MATHEMATICAL INITIAL GHAIN
    Interval{ 0x1EE42, 0x1EE42 }, // Lo       ARABIC MATHEMATICAL TAILED JEEM
    Interval{ 0x1EE47, 0x1EE47 }, // Lo       ARABIC MATHEMATICAL TAILED HAH
    Interval{ 0x1EE49, 0x1EE49 }, // Lo       ARABIC MATHEMATICAL TAILED YEH
    Interval{ 0x1EE4B, 0x1EE4B }, // Lo       ARABIC MATHEMATICAL TAILED LAM
    Interval{ 0x1EE4D, 0x1EE4F }, // Lo   [3] ARABIC MATHEMATICAL TAILED NOON..ARABIC MATHEMATICAL TAILED AIN
    Interval{ 0x1EE51, 0x1EE52 }, // Lo   [2] ARABIC MATHEMATICAL TAILED SAD..ARABIC MATHEMATICAL TAILED QAF
    Interval{ 0x1EE54, 0x1EE54 }, // Lo       ARABIC MATHEMATICAL TAILED SHEEN
    Interval{ 0x1EE57, 0x1EE57 }, // Lo       ARABIC MATHEMATICAL TAILED KHAH
    Interval{ 0x1EE59, 0x1EE59 }, // Lo       ARABIC MATHEMATICAL TAILED DAD
    Interval{ 0x1EE5B, 0x1EE5B }, // Lo       ARABIC MATHEMATICAL TAILED GHAIN
    Interval{ 0x1EE5D, 0x1EE5D }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS NOON
    Interval{ 0x1EE5F, 0x1EE5F }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS QAF
    Interval{ 0x1EE61, 0x1EE62 }, // Lo   [2] ARABIC MATHEMATICAL STRETCHED BEH..ARABIC MATHEMATICAL STRETCHED JEEM
    Interval{ 0x1EE64, 0x1EE64 }, // Lo       ARABIC MATHEMATICAL STRETCHED HEH
    Interval{ 0x1EE67, 0x1EE6A }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED HAH..ARABIC MATHEMATICAL STRETCHED KAF
    Interval{ 0x1EE6C, 0x1EE72 }, // Lo   [7] ARABIC MATHEMATICAL STRETCHED MEEM..ARABIC MATHEMATICAL STRETCHED QAF
    Interval{ 0x1EE74, 0x1EE77 }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED SHEEN..ARABIC MATHEMATICAL STRETCHED KHAH
    Interval{ 0x1EE79, 0x1EE7C }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED DAD..ARABIC MATHEMATICAL STRETCHED DOTLESS BEH
    Interval{ 0x1EE7E, 0x1EE7E }, // Lo       ARABIC MATHEMATICAL STRETCHED DOTLESS FEH
    Interval{ 0x1EE80, 0x1EE89 }, // Lo  [10] ARABIC MATHEMATICAL LOOPED ALEF..ARABIC MATHEMATICAL LOOPED YEH
    Interval{ 0x1EE8B, 0x1EE9B }, // Lo  [17] ARABIC MATHEMATICAL LOOPED LAM..ARABIC MATHEMATICAL LOOPED GHAIN
    Interval{ 0x1EEA1, 0x1EEA3 }, // Lo   [3] ARABIC MATHEMATICAL DOUBLE-STRUCK BEH..ARABIC MATHEMATICAL DOUBLE-STRUCK DAL
    Interval{ 0x1EEA5, 0x1EEA9 }, // Lo   [5] ARABIC MATHEMATICAL DOUBLE-STRUCK WAW..ARABIC MATHEMATICAL DOUBLE-STRUCK YEH
    Interval{ 0x1EEAB, 0x1EEBB }, // Lo  [17] ARABIC MATHEMATICAL DOUBLE-STRUCK LAM..ARABIC MATHEMATICAL DOUBLE-STRUCK GHAIN
    Interval{ 0x20000, 0x2A6DF }, // Lo [42720] CJK UNIFIED IDEOGRAPH-20000..CJK UNIFIED IDEOGRAPH-2A6DF
    Interval{ 0x2A700, 0x2B81D }, // Lo [4382] CJK UNIFIED IDEOGRAPH-2A700..CJK UNIFIED IDEOGRAPH-2B81D
    Interval{ 0x2B820, 0x2CEAD }, // Lo [5774] CJK UNIFIED IDEOGRAPH-2B820..CJK UNIFIED IDEOGRAPH-2CEAD
    Interval{ 0x2CEB0, 0x2EBE0 }, // Lo [7473] CJK UNIFIED IDEOGRAPH-2CEB0..CJK UNIFIED IDEOGRAPH-2EBE0
    Interval{ 0x2EBF0, 0x2EE5D }, // Lo [622] CJK UNIFIED IDEOGRAPH-2EBF0..CJK UNIFIED IDEOGRAPH-2EE5D
    Interval{ 0x2F800, 0x2FA1D }, // Lo [542] CJK COMPATIBILITY IDEOGRAPH-2F800..CJK COMPATIBILITY IDEOGRAPH-2FA1D
    Interval{ 0x30000, 0x3134A }, // Lo [4939] CJK UNIFIED IDEOGRAPH-30000..CJK UNIFIED IDEOGRAPH-3134A
    Interval{ 0x31350, 0x33479 }, // Lo [8490] CJK UNIFIED IDEOGRAPH-31350..CJK UNIFIED IDEOGRAPH-33479
}; // }}}
auto static const Lowercase = std::array<Interval, 692>{ // {{{
    Interval{ 0x0061, 0x007A }, // L&  [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Interval{ 0x00AA, 0x00AA }, // Lo       FEMININE ORDINAL INDICATOR
    Interval{ 0x00B5, 0x00B5 }, // L&       MICRO SIGN
    Interval{ 0x00BA, 0x00BA }, // Lo       MASCULINE ORDINAL INDICATOR
    Interval{ 0x00DF, 0x00F6 }, // L&  [24] LATIN SMALL LETTER SHARP S..LATIN SMALL LETTER O WITH DIAERESIS
    Interval{ 0x00F8, 0x00FF }, // L&   [8] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER Y WITH DIAERESIS
    Interval{ 0x0101, 0x0101 }, // L&       LATIN SMALL LETTER A WITH MACRON
    Interval{ 0x0103, 0x0103 }, // L&       LATIN SMALL LETTER A WITH BREVE
    Interval{ 0x0105, 0x0105 }, // L&       LATIN SMALL LETTER A WITH OGONEK
    Interval{ 0x0107, 0x0107 }, // L&       LATIN SMALL LETTER C WITH ACUTE
    Interval{ 0x0109, 0x0109 }, // L&       LATIN SMALL LETTER C WITH CIRCUMFLEX
    Interval{ 0x010B, 0x010B }, // L&       LATIN SMALL LETTER C WITH DOT ABOVE
    Interval{ 0x010D, 0x010D }, // L&       LATIN SMALL LETTER C WITH CARON
    Interval{ 0x010F, 0x010F }, // L&       LATIN SMALL LETTER D WITH CARON
    Interval{ 0x0111, 0x0111 }, // L&       LATIN SMALL LETTER D WITH STROKE
    Interval{ 0x0113, 0x0113 }, // L&       LATIN SMALL LETTER E WITH MACRON
    Interval{ 0x0115, 0x0115 }, // L&       LATIN SMALL LETTER E WITH BREVE
    Interval{ 0x0117, 0x0117 }, // L&       LATIN SMALL LETTER E WITH DOT ABOVE
    Interval{ 0x0119, 0x0119 }, // L&       LATIN SMALL LETTER E WITH OGONEK
    Interval{ 0x011B, 0x011B }, // L&       LATIN SMALL LETTER E WITH CARON
    Interval{ 0x011D, 0x011D }, // L&       LATIN SMALL LETTER G WITH CIRCUMFLEX
    Interval{ 0x011F, 0x011F }, // L&       LATIN SMALL LETTER G WITH BREVE
    Interval{ 0x0121, 0x0121 }, // L&       LATIN SMALL LETTER G WITH DOT ABOVE
    Interval{ 0x0123, 0x0123 }, // L&       LATIN SMALL LETTER G WITH CEDILLA
    Interval{ 0x0125, 0x0125 }, // L&       LATIN SMALL LETTER H WITH CIRCUMFLEX
    Interval{ 0x0127, 0x0127 }, // L&       LATIN SMALL LETTER H WITH STROKE
    Interval{ 0x0129, 0x0129 }, // L&       LATIN SMALL LETTER I WITH TILDE
    Interval{ 0x012B, 0x012B }, // L&       LATIN SMALL LETTER I WITH MACRON
    Interval{ 0x012D, 0x012D }, // L&       LATIN SMALL LETTER I WITH BREVE
    Interval{ 0x012F, 0x012F }, // L&       LATIN SMALL LETTER I WITH OGONEK
    Interval{ 0x0131, 0x0131 }, // L&       LATIN SMALL LETTER DOTLESS I
    Interval{ 0x0133, 0x0133 }, // L&       LATIN SMALL LIGATURE IJ
    Interval{ 0x0135, 0x0135 }, // L&       LATIN SMALL LETTER J WITH CIRCUMFLEX
    Interval{ 0x0137, 0x0138 }, // L&   [2] LATIN SMALL LETTER K WITH CEDILLA..LATIN SMALL LETTER KRA
    Interval{ 0x013A, 0x013A }, // L&       LATIN SMALL LETTER L WITH ACUTE
    Interval{ 0x013C, 0x013C }, // L&       LATIN SMALL LETTER L WITH CEDILLA
    Interval{ 0x013E, 0x013E }, // L&       LATIN SMALL LETTER L WITH CARON
    Interval{ 0x0140, 0x0140 }, // L&       LATIN SMALL LETTER L WITH MIDDLE DOT
    Interval{ 0x0142, 0x0142 }, // L&       LATIN SMALL LETTER L WITH STROKE
    Interval{ 0x0144, 0x0144 }, // L&       LATIN SMALL LETTER N WITH ACUTE
    Interval{ 0x0146, 0x0146 }, // L&       LATIN SMALL LETTER N WITH CEDILLA
    Interval{ 0x0148, 0x0149 }, // L&   [2] LATIN SMALL LETTER N WITH CARON..LATIN SMALL LETTER N PRECEDED BY APOSTROPHE
    Interval{ 0x014B, 0x014B }, // L&       LATIN SMALL LETTER ENG
    Interval{ 0x014D, 0x014D }, // L&       LATIN SMALL LETTER O WITH MACRON
    Interval{ 0x014F, 0x014F }, // L&       LATIN SMALL LETTER O WITH BREVE
    Interval{ 0x0151, 0x0151 }, // L&       LATIN SMALL LETTER O WITH DOUBLE ACUTE
    Interval{ 0x0153, 0x0153 }, // L&       LATIN SMALL LIGATURE OE
    Interval{ 0x0155, 0x0155 }, // L&       LATIN SMALL LETTER R WITH ACUTE
    Interval{ 0x0157, 0x0157 }, // L&       LATIN SMALL LETTER R WITH CEDILLA
    Interval{ 0x0159, 0x0159 }, // L&       LATIN SMALL LETTER R WITH CARON
    Interval{ 0x015B, 0x015B }, // L&       LATIN SMALL LETTER S WITH ACUTE
    Interval{ 0x015D, 0x015D }, // L&       LATIN SMALL LETTER S WITH CIRCUMFLEX
    Interval{ 0x015F, 0x015F }, // L&       LATIN SMALL LETTER S WITH CEDILLA
    Interval{ 0x0161, 0x0161 }, // L&       LATIN SMALL LETTER S WITH CARON
    Interval{ 0x0163, 0x0163 }, // L&       LATIN SMALL LETTER T WITH CEDILLA
    Interval{ 0x0165, 0x0165 }, // L&       LATIN SMALL LETTER T WITH CARON
    Interval{ 0x0167, 0x0167 }, // L&       LATIN SMALL LETTER T WITH STROKE
    Interval{ 0x0169, 0x0169 }, // L&       LATIN SMALL LETTER U WITH TILDE
    Interval{ 0x016B, 0x016B }, // L&       LATIN SMALL LETTER U WITH MACRON
    Interval{ 0x016D, 0x016D }, // L&       LATIN SMALL LETTER U WITH BREVE
    Interval{ 0x016F, 0x016F }, // L&       LATIN SMALL LETTER U WITH RING ABOVE
    Interval{ 0x0171, 0x0171 }, // L&       LATIN SMALL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x0173, 0x0173 }, // L&       LATIN SMALL LETTER U WITH OGONEK
    Interval{ 0x0175, 0x0175 }, // L&       LATIN SMALL LETTER W WITH CIRCUMFLEX
    Interval{ 0x0177, 0x0177 }, // L&       LATIN SMALL LETTER Y WITH CIRCUMFLEX
    Interval{ 0x017A, 0x017A }, // L&       LATIN SMALL LETTER Z WITH ACUTE
    Interval{ 0x017C, 0x017C }, // L&       LATIN SMALL LETTER Z WITH DOT ABOVE
    Interval{ 0x017E, 0x0180 }, // L&   [3] LATIN SMALL LETTER Z WITH CARON..LATIN SMALL LETTER B WITH STROKE
    Interval{ 0x0183, 0x0183 }, // L&       LATIN SMALL LETTER B WITH TOPBAR
    Interval{ 0x0185, 0x0185 }, // L&       LATIN SMALL LETTER TONE SIX
    Interval{ 0x0188, 0x0188 }, // L&       LATIN SMALL LETTER C WITH HOOK
    Interval{ 0x018C, 0x018D }, // L&   [2] LATIN SMALL LETTER D WITH TOPBAR..LATIN SMALL LETTER TURNED DELTA
    Interval{ 0x0192, 0x0192 }, // L&       LATIN SMALL LETTER F WITH HOOK
    Interval{ 0x0195, 0x0195 }, // L&       LATIN SMALL LETTER HV
    Interval{ 0x0199, 0x019B }, // L&   [3] LATIN SMALL LETTER K WITH HOOK..LATIN SMALL LETTER LAMBDA WITH STROKE
    Interval{ 0x019E, 0x019E }, // L&       LATIN SMALL LETTER N WITH LONG RIGHT LEG
    Interval{ 0x01A1, 0x01A1 }, // L&       LATIN SMALL LETTER O WITH HORN
    Interval{ 0x01A3, 0x01A3 }, // L&       LATIN SMALL LETTER OI
    Interval{ 0x01A5, 0x01A5 }, // L&       LATIN SMALL LETTER P WITH HOOK
    Interval{ 0x01A8, 0x01A8 }, // L&       LATIN SMALL LETTER TONE TWO
    Interval{ 0x01AA, 0x01AB }, // L&   [2] LATIN LETTER REVERSED ESH LOOP..LATIN SMALL LETTER T WITH PALATAL HOOK
    Interval{ 0x01AD, 0x01AD }, // L&       LATIN SMALL LETTER T WITH HOOK
    Interval{ 0x01B0, 0x01B0 }, // L&       LATIN SMALL LETTER U WITH HORN
    Interval{ 0x01B4, 0x01B4 }, // L&       LATIN SMALL LETTER Y WITH HOOK
    Interval{ 0x01B6, 0x01B6 }, // L&       LATIN SMALL LETTER Z WITH STROKE
    Interval{ 0x01B9, 0x01BA }, // L&   [2] LATIN SMALL LETTER EZH REVERSED..LATIN SMALL LETTER EZH WITH TAIL
    Interval{ 0x01BD, 0x01BF }, // L&   [3] LATIN SMALL LETTER TONE FIVE..LATIN LETTER WYNN
    Interval{ 0x01C6, 0x01C6 }, // L&       LATIN SMALL LETTER DZ WITH CARON
    Interval{ 0x01C9, 0x01C9 }, // L&       LATIN SMALL LETTER LJ
    Interval{ 0x01CC, 0x01CC }, // L&       LATIN SMALL LETTER NJ
    Interval{ 0x01CE, 0x01CE }, // L&       LATIN SMALL LETTER A WITH CARON
    Interval{ 0x01D0, 0x01D0 }, // L&       LATIN SMALL LETTER I WITH CARON
    Interval{ 0x01D2, 0x01D2 }, // L&       LATIN SMALL LETTER O WITH CARON
    Interval{ 0x01D4, 0x01D4 }, // L&       LATIN SMALL LETTER U WITH CARON
    Interval{ 0x01D6, 0x01D6 }, // L&       LATIN SMALL LETTER U WITH DIAERESIS AND MACRON
    Interval{ 0x01D8, 0x01D8 }, // L&       LATIN SMALL LETTER U WITH DIAERESIS AND ACUTE
    Interval{ 0x01DA, 0x01DA }, // L&       LATIN SMALL LETTER U WITH DIAERESIS AND CARON
    Interval{ 0x01DC, 0x01DD }, // L&   [2] LATIN SMALL LETTER U WITH DIAERESIS AND GRAVE..LATIN SMALL LETTER TURNED E
    Interval{ 0x01DF, 0x01DF }, // L&       LATIN SMALL LETTER A WITH DIAERESIS AND MACRON
    Interval{ 0x01E1, 0x01E1 }, // L&       LATIN SMALL LETTER A WITH DOT ABOVE AND MACRON
    Interval{ 0x01E3, 0x01E3 }, // L&       LATIN SMALL LETTER AE WITH MACRON
    Interval{ 0x01E5, 0x01E5 }, // L&       LATIN SMALL LETTER G WITH STROKE
    Interval{ 0x01E7, 0x01E7 }, // L&       LATIN SMALL LETTER G WITH CARON
    Interval{ 0x01E9, 0x01E9 }, // L&       LATIN SMALL LETTER K WITH CARON
    Interval{ 0x01EB, 0x01EB }, // L&       LATIN SMALL LETTER O WITH OGONEK
    Interval{ 0x01ED, 0x01ED }, // L&       LATIN SMALL LETTER O WITH OGONEK AND MACRON
    Interval{ 0x01EF, 0x01F0 }, // L&   [2] LATIN SMALL LETTER EZH WITH CARON..LATIN SMALL LETTER J WITH CARON
    Interval{ 0x01F3, 0x01F3 }, // L&       LATIN SMALL LETTER DZ
    Interval{ 0x01F5, 0x01F5 }, // L&       LATIN SMALL LETTER G WITH ACUTE
    Interval{ 0x01F9, 0x01F9 }, // L&       LATIN SMALL LETTER N WITH GRAVE
    Interval{ 0x01FB, 0x01FB }, // L&       LATIN SMALL LETTER A WITH RING ABOVE AND ACUTE
    Interval{ 0x01FD, 0x01FD }, // L&       LATIN SMALL LETTER AE WITH ACUTE
    Interval{ 0x01FF, 0x01FF }, // L&       LATIN SMALL LETTER O WITH STROKE AND ACUTE
    Interval{ 0x0201, 0x0201 }, // L&       LATIN SMALL LETTER A WITH DOUBLE GRAVE
    Interval{ 0x0203, 0x0203 }, // L&       LATIN SMALL LETTER A WITH INVERTED BREVE
    Interval{ 0x0205, 0x0205 }, // L&       LATIN SMALL LETTER E WITH DOUBLE GRAVE
    Interval{ 0x0207, 0x0207 }, // L&       LATIN SMALL LETTER E WITH INVERTED BREVE
    Interval{ 0x0209, 0x0209 }, // L&       LATIN SMALL LETTER I WITH DOUBLE GRAVE
    Interval{ 0x020B, 0x020B }, // L&       LATIN SMALL LETTER I WITH INVERTED BREVE
    Interval{ 0x020D, 0x020D }, // L&       LATIN SMALL LETTER O WITH DOUBLE GRAVE
    Interval{ 0x020F, 0x020F }, // L&       LATIN SMALL LETTER O WITH INVERTED BREVE
    Interval{ 0x0211, 0x0211 }, // L&       LATIN SMALL LETTER R WITH DOUBLE GRAVE
    Interval{ 0x0213, 0x0213 }, // L&       LATIN SMALL LETTER R WITH INVERTED BREVE
    Interval{ 0x0215, 0x0215 }, // L&       LATIN SMALL LETTER U WITH DOUBLE GRAVE
    Interval{ 0x0217, 0x0217 }, // L&       LATIN SMALL LETTER U WITH INVERTED BREVE
    Interval{ 0x0219, 0x0219 }, // L&       LATIN SMALL LETTER S WITH COMMA BELOW
    Interval{ 0x021B, 0x021B }, // L&       LATIN SMALL LETTER T WITH COMMA BELOW
    Interval{ 0x021D, 0x021D }, // L&       LATIN SMALL LETTER YOGH
    Interval{ 0x021F, 0x021F }, // L&       LATIN SMALL LETTER H WITH CARON
    Interval{ 0x0221, 0x0221 }, // L&       LATIN SMALL LETTER D WITH CURL
    Interval{ 0x0223, 0x0223 }, // L&       LATIN SMALL LETTER OU
    Interval{ 0x0225, 0x0225 }, // L&       LATIN SMALL LETTER Z WITH HOOK
    Interval{ 0x0227, 0x0227 }, // L&       LATIN SMALL LETTER A WITH DOT ABOVE
    Interval{ 0x0229, 0x0229 }, // L&       LATIN SMALL LETTER E WITH CEDILLA
    Interval{ 0x022B, 0x022B }, // L&       LATIN SMALL LETTER O WITH DIAERESIS AND MACRON
    Interval{ 0x022D, 0x022D }, // L&       LATIN SMALL LETTER O WITH TILDE AND MACRON
    Interval{ 0x022F, 0x022F }, // L&       LATIN SMALL LETTER O WITH DOT ABOVE
    Interval{ 0x0231, 0x0231 }, // L&       LATIN SMALL LETTER O WITH DOT ABOVE AND MACRON
    Interval{ 0x0233, 0x0239 }, // L&   [7] LATIN SMALL LETTER Y WITH MACRON..LATIN SMALL LETTER QP DIGRAPH
    Interval{ 0x023C, 0x023C }, // L&       LATIN SMALL LETTER C WITH STROKE
    Interval{ 0x023F, 0x0240 }, // L&   [2] LATIN SMALL LETTER S WITH SWASH TAIL..LATIN SMALL LETTER Z WITH SWASH TAIL
    Interval{ 0x0242, 0x0242 }, // L&       LATIN SMALL LETTER GLOTTAL STOP
    Interval{ 0x0247, 0x0247 }, // L&       LATIN SMALL LETTER E WITH STROKE
    Interval{ 0x0249, 0x0249 }, // L&       LATIN SMALL LETTER J WITH STROKE
    Interval{ 0x024B, 0x024B }, // L&       LATIN SMALL LETTER Q WITH HOOK TAIL
    Interval{ 0x024D, 0x024D }, // L&       LATIN SMALL LETTER R WITH STROKE
    Interval{ 0x024F, 0x0293 }, // L&  [69] LATIN SMALL LETTER Y WITH STROKE..LATIN SMALL LETTER EZH WITH CURL
    Interval{ 0x0296, 0x02AF }, // L&  [26] LATIN LETTER INVERTED GLOTTAL STOP..LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    Interval{ 0x02B0, 0x02B8 }, // Lm   [9] MODIFIER LETTER SMALL H..MODIFIER LETTER SMALL Y
    Interval{ 0x02C0, 0x02C1 }, // Lm   [2] MODIFIER LETTER GLOTTAL STOP..MODIFIER LETTER REVERSED GLOTTAL STOP
    Interval{ 0x02E0, 0x02E4 }, // Lm   [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    Interval{ 0x0345, 0x0345 }, // Mn       COMBINING GREEK YPOGEGRAMMENI
    Interval{ 0x0371, 0x0371 }, // L&       GREEK SMALL LETTER HETA
    Interval{ 0x0373, 0x0373 }, // L&       GREEK SMALL LETTER ARCHAIC SAMPI
    Interval{ 0x0377, 0x0377 }, // L&       GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037A, 0x037A }, // Lm       GREEK YPOGEGRAMMENI
    Interval{ 0x037B, 0x037D }, // L&   [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Interval{ 0x0390, 0x0390 }, // L&       GREEK SMALL LETTER IOTA WITH DIALYTIKA AND TONOS
    Interval{ 0x03AC, 0x03CE }, // L&  [35] GREEK SMALL LETTER ALPHA WITH TONOS..GREEK SMALL LETTER OMEGA WITH TONOS
    Interval{ 0x03D0, 0x03D1 }, // L&   [2] GREEK BETA SYMBOL..GREEK THETA SYMBOL
    Interval{ 0x03D5, 0x03D7 }, // L&   [3] GREEK PHI SYMBOL..GREEK KAI SYMBOL
    Interval{ 0x03D9, 0x03D9 }, // L&       GREEK SMALL LETTER ARCHAIC KOPPA
    Interval{ 0x03DB, 0x03DB }, // L&       GREEK SMALL LETTER STIGMA
    Interval{ 0x03DD, 0x03DD }, // L&       GREEK SMALL LETTER DIGAMMA
    Interval{ 0x03DF, 0x03DF }, // L&       GREEK SMALL LETTER KOPPA
    Interval{ 0x03E1, 0x03E1 }, // L&       GREEK SMALL LETTER SAMPI
    Interval{ 0x03E3, 0x03E3 }, // L&       COPTIC SMALL LETTER SHEI
    Interval{ 0x03E5, 0x03E5 }, // L&       COPTIC SMALL LETTER FEI
    Interval{ 0x03E7, 0x03E7 }, // L&       COPTIC SMALL LETTER KHEI
    Interval{ 0x03E9, 0x03E9 }, // L&       COPTIC SMALL LETTER HORI
    Interval{ 0x03EB, 0x03EB }, // L&       COPTIC SMALL LETTER GANGIA
    Interval{ 0x03ED, 0x03ED }, // L&       COPTIC SMALL LETTER SHIMA
    Interval{ 0x03EF, 0x03F3 }, // L&   [5] COPTIC SMALL LETTER DEI..GREEK LETTER YOT
    Interval{ 0x03F5, 0x03F5 }, // L&       GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F8, 0x03F8 }, // L&       GREEK SMALL LETTER SHO
    Interval{ 0x03FB, 0x03FC }, // L&   [2] GREEK SMALL LETTER SAN..GREEK RHO WITH STROKE SYMBOL
    Interval{ 0x0430, 0x045F }, // L&  [48] CYRILLIC SMALL LETTER A..CYRILLIC SMALL LETTER DZHE
    Interval{ 0x0461, 0x0461 }, // L&       CYRILLIC SMALL LETTER OMEGA
    Interval{ 0x0463, 0x0463 }, // L&       CYRILLIC SMALL LETTER YAT
    Interval{ 0x0465, 0x0465 }, // L&       CYRILLIC SMALL LETTER IOTIFIED E
    Interval{ 0x0467, 0x0467 }, // L&       CYRILLIC SMALL LETTER LITTLE YUS
    Interval{ 0x0469, 0x0469 }, // L&       CYRILLIC SMALL LETTER IOTIFIED LITTLE YUS
    Interval{ 0x046B, 0x046B }, // L&       CYRILLIC SMALL LETTER BIG YUS
    Interval{ 0x046D, 0x046D }, // L&       CYRILLIC SMALL LETTER IOTIFIED BIG YUS
    Interval{ 0x046F, 0x046F }, // L&       CYRILLIC SMALL LETTER KSI
    Interval{ 0x0471, 0x0471 }, // L&       CYRILLIC SMALL LETTER PSI
    Interval{ 0x0473, 0x0473 }, // L&       CYRILLIC SMALL LETTER FITA
    Interval{ 0x0475, 0x0475 }, // L&       CYRILLIC SMALL LETTER IZHITSA
    Interval{ 0x0477, 0x0477 }, // L&       CYRILLIC SMALL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT
    Interval{ 0x0479, 0x0479 }, // L&       CYRILLIC SMALL LETTER UK
    Interval{ 0x047B, 0x047B }, // L&       CYRILLIC SMALL LETTER ROUND OMEGA
    Interval{ 0x047D, 0x047D }, // L&       CYRILLIC SMALL LETTER OMEGA WITH TITLO
    Interval{ 0x047F, 0x047F }, // L&       CYRILLIC SMALL LETTER OT
    Interval{ 0x0481, 0x0481 }, // L&       CYRILLIC SMALL LETTER KOPPA
    Interval{ 0x048B, 0x048B }, // L&       CYRILLIC SMALL LETTER SHORT I WITH TAIL
    Interval{ 0x048D, 0x048D }, // L&       CYRILLIC SMALL LETTER SEMISOFT SIGN
    Interval{ 0x048F, 0x048F }, // L&       CYRILLIC SMALL LETTER ER WITH TICK
    Interval{ 0x0491, 0x0491 }, // L&       CYRILLIC SMALL LETTER GHE WITH UPTURN
    Interval{ 0x0493, 0x0493 }, // L&       CYRILLIC SMALL LETTER GHE WITH STROKE
    Interval{ 0x0495, 0x0495 }, // L&       CYRILLIC SMALL LETTER GHE WITH MIDDLE HOOK
    Interval{ 0x0497, 0x0497 }, // L&       CYRILLIC SMALL LETTER ZHE WITH DESCENDER
    Interval{ 0x0499, 0x0499 }, // L&       CYRILLIC SMALL LETTER ZE WITH DESCENDER
    Interval{ 0x049B, 0x049B }, // L&       CYRILLIC SMALL LETTER KA WITH DESCENDER
    Interval{ 0x049D, 0x049D }, // L&       CYRILLIC SMALL LETTER KA WITH VERTICAL STROKE
    Interval{ 0x049F, 0x049F }, // L&       CYRILLIC SMALL LETTER KA WITH STROKE
    Interval{ 0x04A1, 0x04A1 }, // L&       CYRILLIC SMALL LETTER BASHKIR KA
    Interval{ 0x04A3, 0x04A3 }, // L&       CYRILLIC SMALL LETTER EN WITH DESCENDER
    Interval{ 0x04A5, 0x04A5 }, // L&       CYRILLIC SMALL LIGATURE EN GHE
    Interval{ 0x04A7, 0x04A7 }, // L&       CYRILLIC SMALL LETTER PE WITH MIDDLE HOOK
    Interval{ 0x04A9, 0x04A9 }, // L&       CYRILLIC SMALL LETTER ABKHASIAN HA
    Interval{ 0x04AB, 0x04AB }, // L&       CYRILLIC SMALL LETTER ES WITH DESCENDER
    Interval{ 0x04AD, 0x04AD }, // L&       CYRILLIC SMALL LETTER TE WITH DESCENDER
    Interval{ 0x04AF, 0x04AF }, // L&       CYRILLIC SMALL LETTER STRAIGHT U
    Interval{ 0x04B1, 0x04B1 }, // L&       CYRILLIC SMALL LETTER STRAIGHT U WITH STROKE
    Interval{ 0x04B3, 0x04B3 }, // L&       CYRILLIC SMALL LETTER HA WITH DESCENDER
    Interval{ 0x04B5, 0x04B5 }, // L&       CYRILLIC SMALL LIGATURE TE TSE
    Interval{ 0x04B7, 0x04B7 }, // L&       CYRILLIC SMALL LETTER CHE WITH DESCENDER
    Interval{ 0x04B9, 0x04B9 }, // L&       CYRILLIC SMALL LETTER CHE WITH VERTICAL STROKE
    Interval{ 0x04BB, 0x04BB }, // L&       CYRILLIC SMALL LETTER SHHA
    Interval{ 0x04BD, 0x04BD }, // L&       CYRILLIC SMALL LETTER ABKHASIAN CHE
    Interval{ 0x04BF, 0x04BF }, // L&       CYRILLIC SMALL LETTER ABKHASIAN CHE WITH DESCENDER
    Interval{ 0x04C2, 0x04C2 }, // L&       CYRILLIC SMALL LETTER ZHE WITH BREVE
    Interval{ 0x04C4, 0x04C4 }, // L&       CYRILLIC SMALL LETTER KA WITH HOOK
    Interval{ 0x04C6, 0x04C6 }, // L&       CYRILLIC SMALL LETTER EL WITH TAIL
    Interval{ 0x04C8, 0x04C8 }, // L&       CYRILLIC SMALL LETTER EN WITH HOOK
    Interval{ 0x04CA, 0x04CA }, // L&       CYRILLIC SMALL LETTER EN WITH TAIL
    Interval{ 0x04CC, 0x04CC }, // L&       CYRILLIC SMALL LETTER KHAKASSIAN CHE
    Interval{ 0x04CE, 0x04CF }, // L&   [2] CYRILLIC SMALL LETTER EM WITH TAIL..CYRILLIC SMALL LETTER PALOCHKA
    Interval{ 0x04D1, 0x04D1 }, // L&       CYRILLIC SMALL LETTER A WITH BREVE
    Interval{ 0x04D3, 0x04D3 }, // L&       CYRILLIC SMALL LETTER A WITH DIAERESIS
    Interval{ 0x04D5, 0x04D5 }, // L&       CYRILLIC SMALL LIGATURE A IE
    Interval{ 0x04D7, 0x04D7 }, // L&       CYRILLIC SMALL LETTER IE WITH BREVE
    Interval{ 0x04D9, 0x04D9 }, // L&       CYRILLIC SMALL LETTER SCHWA
    Interval{ 0x04DB, 0x04DB }, // L&       CYRILLIC SMALL LETTER SCHWA WITH DIAERESIS
    Interval{ 0x04DD, 0x04DD }, // L&       CYRILLIC SMALL LETTER ZHE WITH DIAERESIS
    Interval{ 0x04DF, 0x04DF }, // L&       CYRILLIC SMALL LETTER ZE WITH DIAERESIS
    Interval{ 0x04E1, 0x04E1 }, // L&       CYRILLIC SMALL LETTER ABKHASIAN DZE
    Interval{ 0x04E3, 0x04E3 }, // L&       CYRILLIC SMALL LETTER I WITH MACRON
    Interval{ 0x04E5, 0x04E5 }, // L&       CYRILLIC SMALL LETTER I WITH DIAERESIS
    Interval{ 0x04E7, 0x04E7 }, // L&       CYRILLIC SMALL LETTER O WITH DIAERESIS
    Interval{ 0x04E9, 0x04E9 }, // L&       CYRILLIC SMALL LETTER BARRED O
    Interval{ 0x04EB, 0x04EB }, // L&       CYRILLIC SMALL LETTER BARRED O WITH DIAERESIS
    Interval{ 0x04ED, 0x04ED }, // L&       CYRILLIC SMALL LETTER E WITH DIAERESIS
    Interval{ 0x04EF, 0x04EF }, // L&       CYRILLIC SMALL LETTER U WITH MACRON
    Interval{ 0x04F1, 0x04F1 }, // L&       CYRILLIC SMALL LETTER U WITH DIAERESIS
    Interval{ 0x04F3, 0x04F3 }, // L&       CYRILLIC SMALL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x04F5, 0x04F5 }, // L&       CYRILLIC SMALL LETTER CHE WITH DIAERESIS
    Interval{ 0x04F7, 0x04F7 }, // L&       CYRILLIC SMALL LETTER GHE WITH DESCENDER
    Interval{ 0x04F9, 0x04F9 }, // L&       CYRILLIC SMALL LETTER YERU WITH DIAERESIS
    Interval{ 0x04FB, 0x04FB }, // L&       CYRILLIC SMALL LETTER GHE WITH STROKE AND HOOK
    Interval{ 0x04FD, 0x04FD }, // L&       CYRILLIC SMALL LETTER HA WITH HOOK
    Interval{ 0x04FF, 0x04FF }, // L&       CYRILLIC SMALL LETTER HA WITH STROKE
    Interval{ 0x0501, 0x0501 }, // L&       CYRILLIC SMALL LETTER KOMI DE
    Interval{ 0x0503, 0x0503 }, // L&       CYRILLIC SMALL LETTER KOMI DJE
    Interval{ 0x0505, 0x0505 }, // L&       CYRILLIC SMALL LETTER KOMI ZJE
    Interval{ 0x0507, 0x0507 }, // L&       CYRILLIC SMALL LETTER KOMI DZJE
    Interval{ 0x0509, 0x0509 }, // L&       CYRILLIC SMALL LETTER KOMI LJE
    Interval{ 0x050B, 0x050B }, // L&       CYRILLIC SMALL LETTER KOMI NJE
    Interval{ 0x050D, 0x050D }, // L&       CYRILLIC SMALL LETTER KOMI SJE
    Interval{ 0x050F, 0x050F }, // L&       CYRILLIC SMALL LETTER KOMI TJE
    Interval{ 0x0511, 0x0511 }, // L&       CYRILLIC SMALL LETTER REVERSED ZE
    Interval{ 0x0513, 0x0513 }, // L&       CYRILLIC SMALL LETTER EL WITH HOOK
    Interval{ 0x0515, 0x0515 }, // L&       CYRILLIC SMALL LETTER LHA
    Interval{ 0x0517, 0x0517 }, // L&       CYRILLIC SMALL LETTER RHA
    Interval{ 0x0519, 0x0519 }, // L&       CYRILLIC SMALL LETTER YAE
    Interval{ 0x051B, 0x051B }, // L&       CYRILLIC SMALL LETTER QA
    Interval{ 0x051D, 0x051D }, // L&       CYRILLIC SMALL LETTER WE
    Interval{ 0x051F, 0x051F }, // L&       CYRILLIC SMALL LETTER ALEUT KA
    Interval{ 0x0521, 0x0521 }, // L&       CYRILLIC SMALL LETTER EL WITH MIDDLE HOOK
    Interval{ 0x0523, 0x0523 }, // L&       CYRILLIC SMALL LETTER EN WITH MIDDLE HOOK
    Interval{ 0x0525, 0x0525 }, // L&       CYRILLIC SMALL LETTER PE WITH DESCENDER
    Interval{ 0x0527, 0x0527 }, // L&       CYRILLIC SMALL LETTER SHHA WITH DESCENDER
    Interval{ 0x0529, 0x0529 }, // L&       CYRILLIC SMALL LETTER EN WITH LEFT HOOK
    Interval{ 0x052B, 0x052B }, // L&       CYRILLIC SMALL LETTER DZZHE
    Interval{ 0x052D, 0x052D }, // L&       CYRILLIC SMALL LETTER DCHE
    Interval{ 0x052F, 0x052F }, // L&       CYRILLIC SMALL LETTER EL WITH DESCENDER
    Interval{ 0x0560, 0x0588 }, // L&  [41] ARMENIAN SMALL LETTER TURNED AYB..ARMENIAN SMALL LETTER YI WITH STROKE
    Interval{ 0x10D0, 0x10FA }, // L&  [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Interval{ 0x10FC, 0x10FC }, // Lm       MODIFIER LETTER GEORGIAN NAR
    Interval{ 0x10FD, 0x10FF }, // L&   [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Interval{ 0x13F8, 0x13FD }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1C80, 0x1C88 }, // L&   [9] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER UNBLENDED UK
    Interval{ 0x1C8A, 0x1C8A }, // L&       CYRILLIC SMALL LETTER TJE
    Interval{ 0x1D00, 0x1D2B }, // L&  [44] LATIN LETTER SMALL CAPITAL A..CYRILLIC LETTER SMALL CAPITAL EL
    Interval{ 0x1D2C, 0x1D6A }, // Lm  [63] MODIFIER LETTER CAPITAL A..GREEK SUBSCRIPT SMALL LETTER CHI
    Interval{ 0x1D6B, 0x1D77 }, // L&  [13] LATIN SMALL LETTER UE..LATIN SMALL LETTER TURNED G
    Interval{ 0x1D78, 0x1D78 }, // Lm       MODIFIER LETTER CYRILLIC EN
    Interval{ 0x1D79, 0x1D9A }, // L&  [34] LATIN SMALL LETTER INSULAR G..LATIN SMALL LETTER EZH WITH RETROFLEX HOOK
    Interval{ 0x1D9B, 0x1DBF }, // Lm  [37] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL THETA
    Interval{ 0x1E01, 0x1E01 }, // L&       LATIN SMALL LETTER A WITH RING BELOW
    Interval{ 0x1E03, 0x1E03 }, // L&       LATIN SMALL LETTER B WITH DOT ABOVE
    Interval{ 0x1E05, 0x1E05 }, // L&       LATIN SMALL LETTER B WITH DOT BELOW
    Interval{ 0x1E07, 0x1E07 }, // L&       LATIN SMALL LETTER B WITH LINE BELOW
    Interval{ 0x1E09, 0x1E09 }, // L&       LATIN SMALL LETTER C WITH CEDILLA AND ACUTE
    Interval{ 0x1E0B, 0x1E0B }, // L&       LATIN SMALL LETTER D WITH DOT ABOVE
    Interval{ 0x1E0D, 0x1E0D }, // L&       LATIN SMALL LETTER D WITH DOT BELOW
    Interval{ 0x1E0F, 0x1E0F }, // L&       LATIN SMALL LETTER D WITH LINE BELOW
    Interval{ 0x1E11, 0x1E11 }, // L&       LATIN SMALL LETTER D WITH CEDILLA
    Interval{ 0x1E13, 0x1E13 }, // L&       LATIN SMALL LETTER D WITH CIRCUMFLEX BELOW
    Interval{ 0x1E15, 0x1E15 }, // L&       LATIN SMALL LETTER E WITH MACRON AND GRAVE
    Interval{ 0x1E17, 0x1E17 }, // L&       LATIN SMALL LETTER E WITH MACRON AND ACUTE
    Interval{ 0x1E19, 0x1E19 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX BELOW
    Interval{ 0x1E1B, 0x1E1B }, // L&       LATIN SMALL LETTER E WITH TILDE BELOW
    Interval{ 0x1E1D, 0x1E1D }, // L&       LATIN SMALL LETTER E WITH CEDILLA AND BREVE
    Interval{ 0x1E1F, 0x1E1F }, // L&       LATIN SMALL LETTER F WITH DOT ABOVE
    Interval{ 0x1E21, 0x1E21 }, // L&       LATIN SMALL LETTER G WITH MACRON
    Interval{ 0x1E23, 0x1E23 }, // L&       LATIN SMALL LETTER H WITH DOT ABOVE
    Interval{ 0x1E25, 0x1E25 }, // L&       LATIN SMALL LETTER H WITH DOT BELOW
    Interval{ 0x1E27, 0x1E27 }, // L&       LATIN SMALL LETTER H WITH DIAERESIS
    Interval{ 0x1E29, 0x1E29 }, // L&       LATIN SMALL LETTER H WITH CEDILLA
    Interval{ 0x1E2B, 0x1E2B }, // L&       LATIN SMALL LETTER H WITH BREVE BELOW
    Interval{ 0x1E2D, 0x1E2D }, // L&       LATIN SMALL LETTER I WITH TILDE BELOW
    Interval{ 0x1E2F, 0x1E2F }, // L&       LATIN SMALL LETTER I WITH DIAERESIS AND ACUTE
    Interval{ 0x1E31, 0x1E31 }, // L&       LATIN SMALL LETTER K WITH ACUTE
    Interval{ 0x1E33, 0x1E33 }, // L&       LATIN SMALL LETTER K WITH DOT BELOW
    Interval{ 0x1E35, 0x1E35 }, // L&       LATIN SMALL LETTER K WITH LINE BELOW
    Interval{ 0x1E37, 0x1E37 }, // L&       LATIN SMALL LETTER L WITH DOT BELOW
    Interval{ 0x1E39, 0x1E39 }, // L&       LATIN SMALL LETTER L WITH DOT BELOW AND MACRON
    Interval{ 0x1E3B, 0x1E3B }, // L&       LATIN SMALL LETTER L WITH LINE BELOW
    Interval{ 0x1E3D, 0x1E3D }, // L&       LATIN SMALL LETTER L WITH CIRCUMFLEX BELOW
    Interval{ 0x1E3F, 0x1E3F }, // L&       LATIN SMALL LETTER M WITH ACUTE
    Interval{ 0x1E41, 0x1E41 }, // L&       LATIN SMALL LETTER M WITH DOT ABOVE
    Interval{ 0x1E43, 0x1E43 }, // L&       LATIN SMALL LETTER M WITH DOT BELOW
    Interval{ 0x1E45, 0x1E45 }, // L&       LATIN SMALL LETTER N WITH DOT ABOVE
    Interval{ 0x1E47, 0x1E47 }, // L&       LATIN SMALL LETTER N WITH DOT BELOW
    Interval{ 0x1E49, 0x1E49 }, // L&       LATIN SMALL LETTER N WITH LINE BELOW
    Interval{ 0x1E4B, 0x1E4B }, // L&       LATIN SMALL LETTER N WITH CIRCUMFLEX BELOW
    Interval{ 0x1E4D, 0x1E4D }, // L&       LATIN SMALL LETTER O WITH TILDE AND ACUTE
    Interval{ 0x1E4F, 0x1E4F }, // L&       LATIN SMALL LETTER O WITH TILDE AND DIAERESIS
    Interval{ 0x1E51, 0x1E51 }, // L&       LATIN SMALL LETTER O WITH MACRON AND GRAVE
    Interval{ 0x1E53, 0x1E53 }, // L&       LATIN SMALL LETTER O WITH MACRON AND ACUTE
    Interval{ 0x1E55, 0x1E55 }, // L&       LATIN SMALL LETTER P WITH ACUTE
    Interval{ 0x1E57, 0x1E57 }, // L&       LATIN SMALL LETTER P WITH DOT ABOVE
    Interval{ 0x1E59, 0x1E59 }, // L&       LATIN SMALL LETTER R WITH DOT ABOVE
    Interval{ 0x1E5B, 0x1E5B }, // L&       LATIN SMALL LETTER R WITH DOT BELOW
    Interval{ 0x1E5D, 0x1E5D }, // L&       LATIN SMALL LETTER R WITH DOT BELOW AND MACRON
    Interval{ 0x1E5F, 0x1E5F }, // L&       LATIN SMALL LETTER R WITH LINE BELOW
    Interval{ 0x1E61, 0x1E61 }, // L&       LATIN SMALL LETTER S WITH DOT ABOVE
    Interval{ 0x1E63, 0x1E63 }, // L&       LATIN SMALL LETTER S WITH DOT BELOW
    Interval{ 0x1E65, 0x1E65 }, // L&       LATIN SMALL LETTER S WITH ACUTE AND DOT ABOVE
    Interval{ 0x1E67, 0x1E67 }, // L&       LATIN SMALL LETTER S WITH CARON AND DOT ABOVE
    Interval{ 0x1E69, 0x1E69 }, // L&       LATIN SMALL LETTER S WITH DOT BELOW AND DOT ABOVE
    Interval{ 0x1E6B, 0x1E6B }, // L&       LATIN SMALL LETTER T WITH DOT ABOVE
    Interval{ 0x1E6D, 0x1E6D }, // L&       LATIN SMALL LETTER T WITH DOT BELOW
    Interval{ 0x1E6F, 0x1E6F }, // L&       LATIN SMALL LETTER T WITH LINE BELOW
    Interval{ 0x1E71, 0x1E71 }, // L&       LATIN SMALL LETTER T WITH CIRCUMFLEX BELOW
    Interval{ 0x1E73, 0x1E73 }, // L&       LATIN SMALL LETTER U WITH DIAERESIS BELOW
    Interval{ 0x1E75, 0x1E75 }, // L&       LATIN SMALL LETTER U WITH TILDE BELOW
    Interval{ 0x1E77, 0x1E77 }, // L&       LATIN SMALL LETTER U WITH CIRCUMFLEX BELOW
    Interval{ 0x1E79, 0x1E79 }, // L&       LATIN SMALL LETTER U WITH TILDE AND ACUTE
    Interval{ 0x1E7B, 0x1E7B }, // L&       LATIN SMALL LETTER U WITH MACRON AND DIAERESIS
    Interval{ 0x1E7D, 0x1E7D }, // L&       LATIN SMALL LETTER V WITH TILDE
    Interval{ 0x1E7F, 0x1E7F }, // L&       LATIN SMALL LETTER V WITH DOT BELOW
    Interval{ 0x1E81, 0x1E81 }, // L&       LATIN SMALL LETTER W WITH GRAVE
    Interval{ 0x1E83, 0x1E83 }, // L&       LATIN SMALL LETTER W WITH ACUTE
    Interval{ 0x1E85, 0x1E85 }, // L&       LATIN SMALL LETTER W WITH DIAERESIS
    Interval{ 0x1E87, 0x1E87 }, // L&       LATIN SMALL LETTER W WITH DOT ABOVE
    Interval{ 0x1E89, 0x1E89 }, // L&       LATIN SMALL LETTER W WITH DOT BELOW
    Interval{ 0x1E8B, 0x1E8B }, // L&       LATIN SMALL LETTER X WITH DOT ABOVE
    Interval{ 0x1E8D, 0x1E8D }, // L&       LATIN SMALL LETTER X WITH DIAERESIS
    Interval{ 0x1E8F, 0x1E8F }, // L&       LATIN SMALL LETTER Y WITH DOT ABOVE
    Interval{ 0x1E91, 0x1E91 }, // L&       LATIN SMALL LETTER Z WITH CIRCUMFLEX
    Interval{ 0x1E93, 0x1E93 }, // L&       LATIN SMALL LETTER Z WITH DOT BELOW
    Interval{ 0x1E95, 0x1E9D }, // L&   [9] LATIN SMALL LETTER Z WITH LINE BELOW..LATIN SMALL LETTER LONG S WITH HIGH STROKE
    Interval{ 0x1E9F, 0x1E9F }, // L&       LATIN SMALL LETTER DELTA
    Interval{ 0x1EA1, 0x1EA1 }, // L&       LATIN SMALL LETTER A WITH DOT BELOW
    Interval{ 0x1EA3, 0x1EA3 }, // L&       LATIN SMALL LETTER A WITH HOOK ABOVE
    Interval{ 0x1EA5, 0x1EA5 }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EA7, 0x1EA7 }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EA9, 0x1EA9 }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EAB, 0x1EAB }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EAD, 0x1EAD }, // L&       LATIN SMALL LETTER A WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EAF, 0x1EAF }, // L&       LATIN SMALL LETTER A WITH BREVE AND ACUTE
    Interval{ 0x1EB1, 0x1EB1 }, // L&       LATIN SMALL LETTER A WITH BREVE AND GRAVE
    Interval{ 0x1EB3, 0x1EB3 }, // L&       LATIN SMALL LETTER A WITH BREVE AND HOOK ABOVE
    Interval{ 0x1EB5, 0x1EB5 }, // L&       LATIN SMALL LETTER A WITH BREVE AND TILDE
    Interval{ 0x1EB7, 0x1EB7 }, // L&       LATIN SMALL LETTER A WITH BREVE AND DOT BELOW
    Interval{ 0x1EB9, 0x1EB9 }, // L&       LATIN SMALL LETTER E WITH DOT BELOW
    Interval{ 0x1EBB, 0x1EBB }, // L&       LATIN SMALL LETTER E WITH HOOK ABOVE
    Interval{ 0x1EBD, 0x1EBD }, // L&       LATIN SMALL LETTER E WITH TILDE
    Interval{ 0x1EBF, 0x1EBF }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EC1, 0x1EC1 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EC3, 0x1EC3 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EC5, 0x1EC5 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EC7, 0x1EC7 }, // L&       LATIN SMALL LETTER E WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EC9, 0x1EC9 }, // L&       LATIN SMALL LETTER I WITH HOOK ABOVE
    Interval{ 0x1ECB, 0x1ECB }, // L&       LATIN SMALL LETTER I WITH DOT BELOW
    Interval{ 0x1ECD, 0x1ECD }, // L&       LATIN SMALL LETTER O WITH DOT BELOW
    Interval{ 0x1ECF, 0x1ECF }, // L&       LATIN SMALL LETTER O WITH HOOK ABOVE
    Interval{ 0x1ED1, 0x1ED1 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1ED3, 0x1ED3 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1ED5, 0x1ED5 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1ED7, 0x1ED7 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1ED9, 0x1ED9 }, // L&       LATIN SMALL LETTER O WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EDB, 0x1EDB }, // L&       LATIN SMALL LETTER O WITH HORN AND ACUTE
    Interval{ 0x1EDD, 0x1EDD }, // L&       LATIN SMALL LETTER O WITH HORN AND GRAVE
    Interval{ 0x1EDF, 0x1EDF }, // L&       LATIN SMALL LETTER O WITH HORN AND HOOK ABOVE
    Interval{ 0x1EE1, 0x1EE1 }, // L&       LATIN SMALL LETTER O WITH HORN AND TILDE
    Interval{ 0x1EE3, 0x1EE3 }, // L&       LATIN SMALL LETTER O WITH HORN AND DOT BELOW
    Interval{ 0x1EE5, 0x1EE5 }, // L&       LATIN SMALL LETTER U WITH DOT BELOW
    Interval{ 0x1EE7, 0x1EE7 }, // L&       LATIN SMALL LETTER U WITH HOOK ABOVE
    Interval{ 0x1EE9, 0x1EE9 }, // L&       LATIN SMALL LETTER U WITH HORN AND ACUTE
    Interval{ 0x1EEB, 0x1EEB }, // L&       LATIN SMALL LETTER U WITH HORN AND GRAVE
    Interval{ 0x1EED, 0x1EED }, // L&       LATIN SMALL LETTER U WITH HORN AND HOOK ABOVE
    Interval{ 0x1EEF, 0x1EEF }, // L&       LATIN SMALL LETTER U WITH HORN AND TILDE
    Interval{ 0x1EF1, 0x1EF1 }, // L&       LATIN SMALL LETTER U WITH HORN AND DOT BELOW
    Interval{ 0x1EF3, 0x1EF3 }, // L&       LATIN SMALL LETTER Y WITH GRAVE
    Interval{ 0x1EF5, 0x1EF5 }, // L&       LATIN SMALL LETTER Y WITH DOT BELOW
    Interval{ 0x1EF7, 0x1EF7 }, // L&       LATIN SMALL LETTER Y WITH HOOK ABOVE
    Interval{ 0x1EF9, 0x1EF9 }, // L&       LATIN SMALL LETTER Y WITH TILDE
    Interval{ 0x1EFB, 0x1EFB }, // L&       LATIN SMALL LETTER MIDDLE-WELSH LL
    Interval{ 0x1EFD, 0x1EFD }, // L&       LATIN SMALL LETTER MIDDLE-WELSH V
    Interval{ 0x1EFF, 0x1F07 }, // L&   [9] LATIN SMALL LETTER Y WITH LOOP..GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F10, 0x1F15 }, // L&   [6] GREEK SMALL LETTER EPSILON WITH PSILI..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F20, 0x1F27 }, // L&   [8] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F30, 0x1F37 }, // L&   [8] GREEK SMALL LETTER IOTA WITH PSILI..GREEK SMALL LETTER IOTA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F40, 0x1F45 }, // L&   [6] GREEK SMALL LETTER OMICRON WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F50, 0x1F57 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F60, 0x1F67 }, // L&   [8] GREEK SMALL LETTER OMEGA WITH PSILI..GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F70, 0x1F7D }, // L&  [14] GREEK SMALL LETTER ALPHA WITH VARIA..GREEK SMALL LETTER OMEGA WITH OXIA
    Interval{ 0x1F80, 0x1F87 }, // L&   [8] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1F90, 0x1F97 }, // L&   [8] GREEK SMALL LETTER ETA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FA0, 0x1FA7 }, // L&   [8] GREEK SMALL LETTER OMEGA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FB0, 0x1FB4 }, // L&   [5] GREEK SMALL LETTER ALPHA WITH VRACHY..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB6, 0x1FB7 }, // L&   [2] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK SMALL LETTER ALPHA WITH PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FBE, 0x1FBE }, // L&       GREEK PROSGEGRAMMENI
    Interval{ 0x1FC2, 0x1FC4 }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC6, 0x1FC7 }, // L&   [2] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK SMALL LETTER ETA WITH PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FD0, 0x1FD3 }, // L&   [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Interval{ 0x1FD6, 0x1FD7 }, // L&   [2] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND PERISPOMENI
    Interval{ 0x1FE0, 0x1FE7 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND PERISPOMENI
    Interval{ 0x1FF2, 0x1FF4 }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF6, 0x1FF7 }, // L&   [2] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK SMALL LETTER OMEGA WITH PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x2071, 0x2071 }, // Lm       SUPERSCRIPT LATIN SMALL LETTER I
    Interval{ 0x207F, 0x207F }, // Lm       SUPERSCRIPT LATIN SMALL LETTER N
    Interval{ 0x2090, 0x209C }, // Lm  [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Interval{ 0x210A, 0x210A }, // L&       SCRIPT SMALL G
    Interval{ 0x210E, 0x210F }, // L&   [2] PLANCK CONSTANT..PLANCK CONSTANT OVER TWO PI
    Interval{ 0x2113, 0x2113 }, // L&       SCRIPT SMALL L
    Interval{ 0x212F, 0x212F }, // L&       SCRIPT SMALL E
    Interval{ 0x2134, 0x2134 }, // L&       SCRIPT SMALL O
    Interval{ 0x2139, 0x2139 }, // L&       INFORMATION SOURCE
    Interval{ 0x213C, 0x213D }, // L&   [2] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK SMALL GAMMA
    Interval{ 0x2146, 0x2149 }, // L&   [4] DOUBLE-STRUCK ITALIC SMALL D..DOUBLE-STRUCK ITALIC SMALL J
    Interval{ 0x214E, 0x214E }, // L&       TURNED SMALL F
    Interval{ 0x2170, 0x217F }, // Nl  [16] SMALL ROMAN NUMERAL ONE..SMALL ROMAN NUMERAL ONE THOUSAND
    Interval{ 0x2184, 0x2184 }, // L&       LATIN SMALL LETTER REVERSED C
    Interval{ 0x24D0, 0x24E9 }, // So  [26] CIRCLED LATIN SMALL LETTER A..CIRCLED LATIN SMALL LETTER Z
    Interval{ 0x2C30, 0x2C5F }, // L&  [48] GLAGOLITIC SMALL LETTER AZU..GLAGOLITIC SMALL LETTER CAUDATE CHRIVI
    Interval{ 0x2C61, 0x2C61 }, // L&       LATIN SMALL LETTER L WITH DOUBLE BAR
    Interval{ 0x2C65, 0x2C66 }, // L&   [2] LATIN SMALL LETTER A WITH STROKE..LATIN SMALL LETTER T WITH DIAGONAL STROKE
    Interval{ 0x2C68, 0x2C68 }, // L&       LATIN SMALL LETTER H WITH DESCENDER
    Interval{ 0x2C6A, 0x2C6A }, // L&       LATIN SMALL LETTER K WITH DESCENDER
    Interval{ 0x2C6C, 0x2C6C }, // L&       LATIN SMALL LETTER Z WITH DESCENDER
    Interval{ 0x2C71, 0x2C71 }, // L&       LATIN SMALL LETTER V WITH RIGHT HOOK
    Interval{ 0x2C73, 0x2C74 }, // L&   [2] LATIN SMALL LETTER W WITH HOOK..LATIN SMALL LETTER V WITH CURL
    Interval{ 0x2C76, 0x2C7B }, // L&   [6] LATIN SMALL LETTER HALF H..LATIN LETTER SMALL CAPITAL TURNED E
    Interval{ 0x2C7C, 0x2C7D }, // Lm   [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    Interval{ 0x2C81, 0x2C81 }, // L&       COPTIC SMALL LETTER ALFA
    Interval{ 0x2C83, 0x2C83 }, // L&       COPTIC SMALL LETTER VIDA
    Interval{ 0x2C85, 0x2C85 }, // L&       COPTIC SMALL LETTER GAMMA
    Interval{ 0x2C87, 0x2C87 }, // L&       COPTIC SMALL LETTER DALDA
    Interval{ 0x2C89, 0x2C89 }, // L&       COPTIC SMALL LETTER EIE
    Interval{ 0x2C8B, 0x2C8B }, // L&       COPTIC SMALL LETTER SOU
    Interval{ 0x2C8D, 0x2C8D }, // L&       COPTIC SMALL LETTER ZATA
    Interval{ 0x2C8F, 0x2C8F }, // L&       COPTIC SMALL LETTER HATE
    Interval{ 0x2C91, 0x2C91 }, // L&       COPTIC SMALL LETTER THETHE
    Interval{ 0x2C93, 0x2C93 }, // L&       COPTIC SMALL LETTER IAUDA
    Interval{ 0x2C95, 0x2C95 }, // L&       COPTIC SMALL LETTER KAPA
    Interval{ 0x2C97, 0x2C97 }, // L&       COPTIC SMALL LETTER LAULA
    Interval{ 0x2C99, 0x2C99 }, // L&       COPTIC SMALL LETTER MI
    Interval{ 0x2C9B, 0x2C9B }, // L&       COPTIC SMALL LETTER NI
    Interval{ 0x2C9D, 0x2C9D }, // L&       COPTIC SMALL LETTER KSI
    Interval{ 0x2C9F, 0x2C9F }, // L&       COPTIC SMALL LETTER O
    Interval{ 0x2CA1, 0x2CA1 }, // L&       COPTIC SMALL LETTER PI
    Interval{ 0x2CA3, 0x2CA3 }, // L&       COPTIC SMALL LETTER RO
    Interval{ 0x2CA5, 0x2CA5 }, // L&       COPTIC SMALL LETTER SIMA
    Interval{ 0x2CA7, 0x2CA7 }, // L&       COPTIC SMALL LETTER TAU
    Interval{ 0x2CA9, 0x2CA9 }, // L&       COPTIC SMALL LETTER UA
    Interval{ 0x2CAB, 0x2CAB }, // L&       COPTIC SMALL LETTER FI
    Interval{ 0x2CAD, 0x2CAD }, // L&       COPTIC SMALL LETTER KHI
    Interval{ 0x2CAF, 0x2CAF }, // L&       COPTIC SMALL LETTER PSI
    Interval{ 0x2CB1, 0x2CB1 }, // L&       COPTIC SMALL LETTER OOU
    Interval{ 0x2CB3, 0x2CB3 }, // L&       COPTIC SMALL LETTER DIALECT-P ALEF
    Interval{ 0x2CB5, 0x2CB5 }, // L&       COPTIC SMALL LETTER OLD COPTIC AIN
    Interval{ 0x2CB7, 0x2CB7 }, // L&       COPTIC SMALL LETTER CRYPTOGRAMMIC EIE
    Interval{ 0x2CB9, 0x2CB9 }, // L&       COPTIC SMALL LETTER DIALECT-P KAPA
    Interval{ 0x2CBB, 0x2CBB }, // L&       COPTIC SMALL LETTER DIALECT-P NI
    Interval{ 0x2CBD, 0x2CBD }, // L&       COPTIC SMALL LETTER CRYPTOGRAMMIC NI
    Interval{ 0x2CBF, 0x2CBF }, // L&       COPTIC SMALL LETTER OLD COPTIC OOU
    Interval{ 0x2CC1, 0x2CC1 }, // L&       COPTIC SMALL LETTER SAMPI
    Interval{ 0x2CC3, 0x2CC3 }, // L&       COPTIC SMALL LETTER CROSSED SHEI
    Interval{ 0x2CC5, 0x2CC5 }, // L&       COPTIC SMALL LETTER OLD COPTIC SHEI
    Interval{ 0x2CC7, 0x2CC7 }, // L&       COPTIC SMALL LETTER OLD COPTIC ESH
    Interval{ 0x2CC9, 0x2CC9 }, // L&       COPTIC SMALL LETTER AKHMIMIC KHEI
    Interval{ 0x2CCB, 0x2CCB }, // L&       COPTIC SMALL LETTER DIALECT-P HORI
    Interval{ 0x2CCD, 0x2CCD }, // L&       COPTIC SMALL LETTER OLD COPTIC HORI
    Interval{ 0x2CCF, 0x2CCF }, // L&       COPTIC SMALL LETTER OLD COPTIC HA
    Interval{ 0x2CD1, 0x2CD1 }, // L&       COPTIC SMALL LETTER L-SHAPED HA
    Interval{ 0x2CD3, 0x2CD3 }, // L&       COPTIC SMALL LETTER OLD COPTIC HEI
    Interval{ 0x2CD5, 0x2CD5 }, // L&       COPTIC SMALL LETTER OLD COPTIC HAT
    Interval{ 0x2CD7, 0x2CD7 }, // L&       COPTIC SMALL LETTER OLD COPTIC GANGIA
    Interval{ 0x2CD9, 0x2CD9 }, // L&       COPTIC SMALL LETTER OLD COPTIC DJA
    Interval{ 0x2CDB, 0x2CDB }, // L&       COPTIC SMALL LETTER OLD COPTIC SHIMA
    Interval{ 0x2CDD, 0x2CDD }, // L&       COPTIC SMALL LETTER OLD NUBIAN SHIMA
    Interval{ 0x2CDF, 0x2CDF }, // L&       COPTIC SMALL LETTER OLD NUBIAN NGI
    Interval{ 0x2CE1, 0x2CE1 }, // L&       COPTIC SMALL LETTER OLD NUBIAN NYI
    Interval{ 0x2CE3, 0x2CE4 }, // L&   [2] COPTIC SMALL LETTER OLD NUBIAN WAU..COPTIC SYMBOL KAI
    Interval{ 0x2CEC, 0x2CEC }, // L&       COPTIC SMALL LETTER CRYPTOGRAMMIC SHEI
    Interval{ 0x2CEE, 0x2CEE }, // L&       COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF3, 0x2CF3 }, // L&       COPTIC SMALL LETTER BOHAIRIC KHEI
    Interval{ 0x2D00, 0x2D25 }, // L&  [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Interval{ 0x2D27, 0x2D27 }, // L&       GEORGIAN SMALL LETTER YN
    Interval{ 0x2D2D, 0x2D2D }, // L&       GEORGIAN SMALL LETTER AEN
    Interval{ 0xA641, 0xA641 }, // L&       CYRILLIC SMALL LETTER ZEMLYA
    Interval{ 0xA643, 0xA643 }, // L&       CYRILLIC SMALL LETTER DZELO
    Interval{ 0xA645, 0xA645 }, // L&       CYRILLIC SMALL LETTER REVERSED DZE
    Interval{ 0xA647, 0xA647 }, // L&       CYRILLIC SMALL LETTER IOTA
    Interval{ 0xA649, 0xA649 }, // L&       CYRILLIC SMALL LETTER DJERV
    Interval{ 0xA64B, 0xA64B }, // L&       CYRILLIC SMALL LETTER MONOGRAPH UK
    Interval{ 0xA64D, 0xA64D }, // L&       CYRILLIC SMALL LETTER BROAD OMEGA
    Interval{ 0xA64F, 0xA64F }, // L&       CYRILLIC SMALL LETTER NEUTRAL YER
    Interval{ 0xA651, 0xA651 }, // L&       CYRILLIC SMALL LETTER YERU WITH BACK YER
    Interval{ 0xA653, 0xA653 }, // L&       CYRILLIC SMALL LETTER IOTIFIED YAT
    Interval{ 0xA655, 0xA655 }, // L&       CYRILLIC SMALL LETTER REVERSED YU
    Interval{ 0xA657, 0xA657 }, // L&       CYRILLIC SMALL LETTER IOTIFIED A
    Interval{ 0xA659, 0xA659 }, // L&       CYRILLIC SMALL LETTER CLOSED LITTLE YUS
    Interval{ 0xA65B, 0xA65B }, // L&       CYRILLIC SMALL LETTER BLENDED YUS
    Interval{ 0xA65D, 0xA65D }, // L&       CYRILLIC SMALL LETTER IOTIFIED CLOSED LITTLE YUS
    Interval{ 0xA65F, 0xA65F }, // L&       CYRILLIC SMALL LETTER YN
    Interval{ 0xA661, 0xA661 }, // L&       CYRILLIC SMALL LETTER REVERSED TSE
    Interval{ 0xA663, 0xA663 }, // L&       CYRILLIC SMALL LETTER SOFT DE
    Interval{ 0xA665, 0xA665 }, // L&       CYRILLIC SMALL LETTER SOFT EL
    Interval{ 0xA667, 0xA667 }, // L&       CYRILLIC SMALL LETTER SOFT EM
    Interval{ 0xA669, 0xA669 }, // L&       CYRILLIC SMALL LETTER MONOCULAR O
    Interval{ 0xA66B, 0xA66B }, // L&       CYRILLIC SMALL LETTER BINOCULAR O
    Interval{ 0xA66D, 0xA66D }, // L&       CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA681, 0xA681 }, // L&       CYRILLIC SMALL LETTER DWE
    Interval{ 0xA683, 0xA683 }, // L&       CYRILLIC SMALL LETTER DZWE
    Interval{ 0xA685, 0xA685 }, // L&       CYRILLIC SMALL LETTER ZHWE
    Interval{ 0xA687, 0xA687 }, // L&       CYRILLIC SMALL LETTER CCHE
    Interval{ 0xA689, 0xA689 }, // L&       CYRILLIC SMALL LETTER DZZE
    Interval{ 0xA68B, 0xA68B }, // L&       CYRILLIC SMALL LETTER TE WITH MIDDLE HOOK
    Interval{ 0xA68D, 0xA68D }, // L&       CYRILLIC SMALL LETTER TWE
    Interval{ 0xA68F, 0xA68F }, // L&       CYRILLIC SMALL LETTER TSWE
    Interval{ 0xA691, 0xA691 }, // L&       CYRILLIC SMALL LETTER TSSE
    Interval{ 0xA693, 0xA693 }, // L&       CYRILLIC SMALL LETTER TCHE
    Interval{ 0xA695, 0xA695 }, // L&       CYRILLIC SMALL LETTER HWE
    Interval{ 0xA697, 0xA697 }, // L&       CYRILLIC SMALL LETTER SHWE
    Interval{ 0xA699, 0xA699 }, // L&       CYRILLIC SMALL LETTER DOUBLE O
    Interval{ 0xA69B, 0xA69B }, // L&       CYRILLIC SMALL LETTER CROSSED O
    Interval{ 0xA69C, 0xA69D }, // Lm   [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    Interval{ 0xA723, 0xA723 }, // L&       LATIN SMALL LETTER EGYPTOLOGICAL ALEF
    Interval{ 0xA725, 0xA725 }, // L&       LATIN SMALL LETTER EGYPTOLOGICAL AIN
    Interval{ 0xA727, 0xA727 }, // L&       LATIN SMALL LETTER HENG
    Interval{ 0xA729, 0xA729 }, // L&       LATIN SMALL LETTER TZ
    Interval{ 0xA72B, 0xA72B }, // L&       LATIN SMALL LETTER TRESILLO
    Interval{ 0xA72D, 0xA72D }, // L&       LATIN SMALL LETTER CUATRILLO
    Interval{ 0xA72F, 0xA731 }, // L&   [3] LATIN SMALL LETTER CUATRILLO WITH COMMA..LATIN LETTER SMALL CAPITAL S
    Interval{ 0xA733, 0xA733 }, // L&       LATIN SMALL LETTER AA
    Interval{ 0xA735, 0xA735 }, // L&       LATIN SMALL LETTER AO
    Interval{ 0xA737, 0xA737 }, // L&       LATIN SMALL LETTER AU
    Interval{ 0xA739, 0xA739 }, // L&       LATIN SMALL LETTER AV
    Interval{ 0xA73B, 0xA73B }, // L&       LATIN SMALL LETTER AV WITH HORIZONTAL BAR
    Interval{ 0xA73D, 0xA73D }, // L&       LATIN SMALL LETTER AY
    Interval{ 0xA73F, 0xA73F }, // L&       LATIN SMALL LETTER REVERSED C WITH DOT
    Interval{ 0xA741, 0xA741 }, // L&       LATIN SMALL LETTER K WITH STROKE
    Interval{ 0xA743, 0xA743 }, // L&       LATIN SMALL LETTER K WITH DIAGONAL STROKE
    Interval{ 0xA745, 0xA745 }, // L&       LATIN SMALL LETTER K WITH STROKE AND DIAGONAL STROKE
    Interval{ 0xA747, 0xA747 }, // L&       LATIN SMALL LETTER BROKEN L
    Interval{ 0xA749, 0xA749 }, // L&       LATIN SMALL LETTER L WITH HIGH STROKE
    Interval{ 0xA74B, 0xA74B }, // L&       LATIN SMALL LETTER O WITH LONG STROKE OVERLAY
    Interval{ 0xA74D, 0xA74D }, // L&       LATIN SMALL LETTER O WITH LOOP
    Interval{ 0xA74F, 0xA74F }, // L&       LATIN SMALL LETTER OO
    Interval{ 0xA751, 0xA751 }, // L&       LATIN SMALL LETTER P WITH STROKE THROUGH DESCENDER
    Interval{ 0xA753, 0xA753 }, // L&       LATIN SMALL LETTER P WITH FLOURISH
    Interval{ 0xA755, 0xA755 }, // L&       LATIN SMALL LETTER P WITH SQUIRREL TAIL
    Interval{ 0xA757, 0xA757 }, // L&       LATIN SMALL LETTER Q WITH STROKE THROUGH DESCENDER
    Interval{ 0xA759, 0xA759 }, // L&       LATIN SMALL LETTER Q WITH DIAGONAL STROKE
    Interval{ 0xA75B, 0xA75B }, // L&       LATIN SMALL LETTER R ROTUNDA
    Interval{ 0xA75D, 0xA75D }, // L&       LATIN SMALL LETTER RUM ROTUNDA
    Interval{ 0xA75F, 0xA75F }, // L&       LATIN SMALL LETTER V WITH DIAGONAL STROKE
    Interval{ 0xA761, 0xA761 }, // L&       LATIN SMALL LETTER VY
    Interval{ 0xA763, 0xA763 }, // L&       LATIN SMALL LETTER VISIGOTHIC Z
    Interval{ 0xA765, 0xA765 }, // L&       LATIN SMALL LETTER THORN WITH STROKE
    Interval{ 0xA767, 0xA767 }, // L&       LATIN SMALL LETTER THORN WITH STROKE THROUGH DESCENDER
    Interval{ 0xA769, 0xA769 }, // L&       LATIN SMALL LETTER VEND
    Interval{ 0xA76B, 0xA76B }, // L&       LATIN SMALL LETTER ET
    Interval{ 0xA76D, 0xA76D }, // L&       LATIN SMALL LETTER IS
    Interval{ 0xA76F, 0xA76F }, // L&       LATIN SMALL LETTER CON
    Interval{ 0xA770, 0xA770 }, // Lm       MODIFIER LETTER US
    Interval{ 0xA771, 0xA778 }, // L&   [8] LATIN SMALL LETTER DUM..LATIN SMALL LETTER UM
    Interval{ 0xA77A, 0xA77A }, // L&       LATIN SMALL LETTER INSULAR D
    Interval{ 0xA77C, 0xA77C }, // L&       LATIN SMALL LETTER INSULAR F
    Interval{ 0xA77F, 0xA77F }, // L&       LATIN SMALL LETTER TURNED INSULAR G
    Interval{ 0xA781, 0xA781 }, // L&       LATIN SMALL LETTER TURNED L
    Interval{ 0xA783, 0xA783 }, // L&       LATIN SMALL LETTER INSULAR R
    Interval{ 0xA785, 0xA785 }, // L&       LATIN SMALL LETTER INSULAR S
    Interval{ 0xA787, 0xA787 }, // L&       LATIN SMALL LETTER INSULAR T
    Interval{ 0xA78C, 0xA78C }, // L&       LATIN SMALL LETTER SALTILLO
    Interval{ 0xA78E, 0xA78E }, // L&       LATIN SMALL LETTER L WITH RETROFLEX HOOK AND BELT
    Interval{ 0xA791, 0xA791 }, // L&       LATIN SMALL LETTER N WITH DESCENDER
    Interval{ 0xA793, 0xA795 }, // L&   [3] LATIN SMALL LETTER C WITH BAR..LATIN SMALL LETTER H WITH PALATAL HOOK
    Interval{ 0xA797, 0xA797 }, // L&       LATIN SMALL LETTER B WITH FLOURISH
    Interval{ 0xA799, 0xA799 }, // L&       LATIN SMALL LETTER F WITH STROKE
    Interval{ 0xA79B, 0xA79B }, // L&       LATIN SMALL LETTER VOLAPUK AE
    Interval{ 0xA79D, 0xA79D }, // L&       LATIN SMALL LETTER VOLAPUK OE
    Interval{ 0xA79F, 0xA79F }, // L&       LATIN SMALL LETTER VOLAPUK UE
    Interval{ 0xA7A1, 0xA7A1 }, // L&       LATIN SMALL LETTER G WITH OBLIQUE STROKE
    Interval{ 0xA7A3, 0xA7A3 }, // L&       LATIN SMALL LETTER K WITH OBLIQUE STROKE
    Interval{ 0xA7A5, 0xA7A5 }, // L&       LATIN SMALL LETTER N WITH OBLIQUE STROKE
    Interval{ 0xA7A7, 0xA7A7 }, // L&       LATIN SMALL LETTER R WITH OBLIQUE STROKE
    Interval{ 0xA7A9, 0xA7A9 }, // L&       LATIN SMALL LETTER S WITH OBLIQUE STROKE
    Interval{ 0xA7AF, 0xA7AF }, // L&       LATIN LETTER SMALL CAPITAL Q
    Interval{ 0xA7B5, 0xA7B5 }, // L&       LATIN SMALL LETTER BETA
    Interval{ 0xA7B7, 0xA7B7 }, // L&       LATIN SMALL LETTER OMEGA
    Interval{ 0xA7B9, 0xA7B9 }, // L&       LATIN SMALL LETTER U WITH STROKE
    Interval{ 0xA7BB, 0xA7BB }, // L&       LATIN SMALL LETTER GLOTTAL A
    Interval{ 0xA7BD, 0xA7BD }, // L&       LATIN SMALL LETTER GLOTTAL I
    Interval{ 0xA7BF, 0xA7BF }, // L&       LATIN SMALL LETTER GLOTTAL U
    Interval{ 0xA7C1, 0xA7C1 }, // L&       LATIN SMALL LETTER OLD POLISH O
    Interval{ 0xA7C3, 0xA7C3 }, // L&       LATIN SMALL LETTER ANGLICANA W
    Interval{ 0xA7C8, 0xA7C8 }, // L&       LATIN SMALL LETTER D WITH SHORT STROKE OVERLAY
    Interval{ 0xA7CA, 0xA7CA }, // L&       LATIN SMALL LETTER S WITH SHORT STROKE OVERLAY
    Interval{ 0xA7CD, 0xA7CD }, // L&       LATIN SMALL LETTER S WITH DIAGONAL STROKE
    Interval{ 0xA7CF, 0xA7CF }, // L&       LATIN SMALL LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0xA7D1, 0xA7D1 }, // L&       LATIN SMALL LETTER CLOSED INSULAR G
    Interval{ 0xA7D3, 0xA7D3 }, // L&       LATIN SMALL LETTER DOUBLE THORN
    Interval{ 0xA7D5, 0xA7D5 }, // L&       LATIN SMALL LETTER DOUBLE WYNN
    Interval{ 0xA7D7, 0xA7D7 }, // L&       LATIN SMALL LETTER MIDDLE SCOTS S
    Interval{ 0xA7D9, 0xA7D9 }, // L&       LATIN SMALL LETTER SIGMOID S
    Interval{ 0xA7DB, 0xA7DB }, // L&       LATIN SMALL LETTER LAMBDA
    Interval{ 0xA7F1, 0xA7F4 }, // Lm   [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    Interval{ 0xA7F6, 0xA7F6 }, // L&       LATIN SMALL LETTER REVERSED HALF H
    Interval{ 0xA7F8, 0xA7F9 }, // Lm   [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    Interval{ 0xA7FA, 0xA7FA }, // L&       LATIN LETTER SMALL CAPITAL TURNED M
    Interval{ 0xAB30, 0xAB5A }, // L&  [43] LATIN SMALL LETTER BARRED ALPHA..LATIN SMALL LETTER Y WITH SHORT RIGHT LEG
    Interval{ 0xAB5C, 0xAB5F }, // Lm   [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    Interval{ 0xAB60, 0xAB68 }, // L&   [9] LATIN SMALL LETTER SAKHA YAT..LATIN SMALL LETTER TURNED R WITH MIDDLE TILDE
    Interval{ 0xAB69, 0xAB69 }, // Lm       MODIFIER LETTER SMALL TURNED W
    Interval{ 0xAB70, 0xABBF }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xFB00, 0xFB06 }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFF41, 0xFF5A }, // L&  [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Interval{ 0x10428, 0x1044F }, // L&  [40] DESERET SMALL LETTER LONG I..DESERET SMALL LETTER EW
    Interval{ 0x104D8, 0x104FB }, // L&  [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Interval{ 0x10597, 0x105A1 }, // L&  [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Interval{ 0x105A3, 0x105B1 }, // L&  [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Interval{ 0x105B3, 0x105B9 }, // L&   [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Interval{ 0x105BB, 0x105BC }, // L&   [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Interval{ 0x10780, 0x10780 }, // Lm       MODIFIER LETTER SMALL CAPITAL AA
    Interval{ 0x10783, 0x10785 }, // Lm   [3] MODIFIER LETTER SMALL AE..MODIFIER LETTER SMALL B WITH HOOK
    Interval{ 0x10787, 0x107B0 }, // Lm  [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Interval{ 0x107B2, 0x107BA }, // Lm   [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Interval{ 0x10CC0, 0x10CF2 }, // L&  [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Interval{ 0x10D70, 0x10D85 }, // L&  [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Interval{ 0x118C0, 0x118DF }, // L&  [32] WARANG CITI SMALL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Interval{ 0x16E60, 0x16E7F }, // L&  [32] MEDEFAIDRIN SMALL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Interval{ 0x16EBB, 0x16ED3 }, // L&  [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Interval{ 0x1D41A, 0x1D433 }, // L&  [26] MATHEMATICAL BOLD SMALL A..MATHEMATICAL BOLD SMALL Z
    Interval{ 0x1D44E, 0x1D454 }, // L&   [7] MATHEMATICAL ITALIC SMALL A..MATHEMATICAL ITALIC SMALL G
    Interval{ 0x1D456, 0x1D467 }, // L&  [18] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL ITALIC SMALL Z
    Interval{ 0x1D482, 0x1D49B }, // L&  [26] MATHEMATICAL BOLD ITALIC SMALL A..MATHEMATICAL BOLD ITALIC SMALL Z
    Interval{ 0x1D4B6, 0x1D4B9 }, // L&   [4] MATHEMATICAL SCRIPT SMALL A..MATHEMATICAL SCRIPT SMALL D
    Interval{ 0x1D4BB, 0x1D4BB }, // L&       MATHEMATICAL SCRIPT SMALL F
    Interval{ 0x1D4BD, 0x1D4C3 }, // L&   [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Interval{ 0x1D4C5, 0x1D4CF }, // L&  [11] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL SCRIPT SMALL Z
    Interval{ 0x1D4EA, 0x1D503 }, // L&  [26] MATHEMATICAL BOLD SCRIPT SMALL A..MATHEMATICAL BOLD SCRIPT SMALL Z
    Interval{ 0x1D51E, 0x1D537 }, // L&  [26] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL FRAKTUR SMALL Z
    Interval{ 0x1D552, 0x1D56B }, // L&  [26] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL DOUBLE-STRUCK SMALL Z
    Interval{ 0x1D586, 0x1D59F }, // L&  [26] MATHEMATICAL BOLD FRAKTUR SMALL A..MATHEMATICAL BOLD FRAKTUR SMALL Z
    Interval{ 0x1D5BA, 0x1D5D3 }, // L&  [26] MATHEMATICAL SANS-SERIF SMALL A..MATHEMATICAL SANS-SERIF SMALL Z
    Interval{ 0x1D5EE, 0x1D607 }, // L&  [26] MATHEMATICAL SANS-SERIF BOLD SMALL A..MATHEMATICAL SANS-SERIF BOLD SMALL Z
    Interval{ 0x1D622, 0x1D63B }, // L&  [26] MATHEMATICAL SANS-SERIF ITALIC SMALL A..MATHEMATICAL SANS-SERIF ITALIC SMALL Z
    Interval{ 0x1D656, 0x1D66F }, // L&  [26] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL A..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL Z
    Interval{ 0x1D68A, 0x1D6A5 }, // L&  [28] MATHEMATICAL MONOSPACE SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    Interval{ 0x1D6C2, 0x1D6DA }, // L&  [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    Interval{ 0x1D6DC, 0x1D6E1 }, // L&   [6] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL BOLD PI SYMBOL
    Interval{ 0x1D6FC, 0x1D714 }, // L&  [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    Interval{ 0x1D716, 0x1D71B }, // L&   [6] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL ITALIC PI SYMBOL
    Interval{ 0x1D736, 0x1D74E }, // L&  [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D750, 0x1D755 }, // L&   [6] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC PI SYMBOL
    Interval{ 0x1D770, 0x1D788 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    Interval{ 0x1D78A, 0x1D78F }, // L&   [6] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD PI SYMBOL
    Interval{ 0x1D7AA, 0x1D7C2 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D7C4, 0x1D7C9 }, // L&   [6] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC PI SYMBOL
    Interval{ 0x1D7CB, 0x1D7CB }, // L&       MATHEMATICAL BOLD SMALL DIGAMMA
    Interval{ 0x1DF00, 0x1DF09 }, // L&  [10] LATIN SMALL LETTER FENG DIGRAPH WITH TRILL..LATIN SMALL LETTER T WITH HOOK AND RETROFLEX HOOK
    Interval{ 0x1DF0B, 0x1DF1E }, // L&  [20] LATIN SMALL LETTER ESH WITH DOUBLE BAR..LATIN SMALL LETTER S WITH CURL
    Interval{ 0x1DF25, 0x1DF2A }, // L&   [6] LATIN SMALL LETTER D WITH MID-HEIGHT LEFT HOOK..LATIN SMALL LETTER T WITH MID-HEIGHT LEFT HOOK
    Interval{ 0x1E030, 0x1E06D }, // Lm  [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Interval{ 0x1E922, 0x1E943 }, // L&  [34] ADLAM SMALL LETTER ALIF..ADLAM SMALL LETTER SHA
}; // }}}
auto static const Math = std::array<Interval, 243>{ // {{{
    Interval{ 0x002B, 0x002B }, // Sm       PLUS SIGN
    Interval{ 0x003C, 0x003E }, // Sm   [3] LESS-THAN SIGN..GREATER-THAN SIGN
    Interval{ 0x005E, 0x005E }, // Sk       CIRCUMFLEX ACCENT
    Interval{ 0x007C, 0x007C }, // Sm       VERTICAL LINE
    Interval{ 0x007E, 0x007E }, // Sm       TILDE
    Interval{ 0x00AC, 0x00AC }, // Sm       NOT SIGN
    Interval{ 0x00B1, 0x00B1 }, // Sm       PLUS-MINUS SIGN
    Interval{ 0x00D7, 0x00D7 }, // Sm       MULTIPLICATION SIGN
    Interval{ 0x00F7, 0x00F7 }, // Sm       DIVISION SIGN
    Interval{ 0x03D0, 0x03D2 }, // L&   [3] GREEK BETA SYMBOL..GREEK UPSILON WITH HOOK SYMBOL
    Interval{ 0x03D5, 0x03D5 }, // L&       GREEK PHI SYMBOL
    Interval{ 0x03F0, 0x03F1 }, // L&   [2] GREEK KAPPA SYMBOL..GREEK RHO SYMBOL
    Interval{ 0x03F4, 0x03F5 }, // L&   [2] GREEK CAPITAL THETA SYMBOL..GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F6, 0x03F6 }, // Sm       GREEK REVERSED LUNATE EPSILON SYMBOL
    Interval{ 0x0606, 0x0608 }, // Sm   [3] ARABIC-INDIC CUBE ROOT..ARABIC RAY
    Interval{ 0x2016, 0x2016 }, // Po       DOUBLE VERTICAL LINE
    Interval{ 0x2032, 0x2034 }, // Po   [3] PRIME..TRIPLE PRIME
    Interval{ 0x2040, 0x2040 }, // Pc       CHARACTER TIE
    Interval{ 0x2044, 0x2044 }, // Sm       FRACTION SLASH
    Interval{ 0x2052, 0x2052 }, // Sm       COMMERCIAL MINUS SIGN
    Interval{ 0x2061, 0x2064 }, // Cf   [4] FUNCTION APPLICATION..INVISIBLE PLUS
    Interval{ 0x207A, 0x207C }, // Sm   [3] SUPERSCRIPT PLUS SIGN..SUPERSCRIPT EQUALS SIGN
    Interval{ 0x207D, 0x207D }, // Ps       SUPERSCRIPT LEFT PARENTHESIS
    Interval{ 0x207E, 0x207E }, // Pe       SUPERSCRIPT RIGHT PARENTHESIS
    Interval{ 0x208A, 0x208C }, // Sm   [3] SUBSCRIPT PLUS SIGN..SUBSCRIPT EQUALS SIGN
    Interval{ 0x208D, 0x208D }, // Ps       SUBSCRIPT LEFT PARENTHESIS
    Interval{ 0x208E, 0x208E }, // Pe       SUBSCRIPT RIGHT PARENTHESIS
    Interval{ 0x20D0, 0x20DC }, // Mn  [13] COMBINING LEFT HARPOON ABOVE..COMBINING FOUR DOTS ABOVE
    Interval{ 0x20E1, 0x20E1 }, // Mn       COMBINING LEFT RIGHT ARROW ABOVE
    Interval{ 0x20E5, 0x20E6 }, // Mn   [2] COMBINING REVERSE SOLIDUS OVERLAY..COMBINING DOUBLE VERTICAL STROKE OVERLAY
    Interval{ 0x20EB, 0x20EF }, // Mn   [5] COMBINING LONG DOUBLE SOLIDUS OVERLAY..COMBINING RIGHT ARROW BELOW
    Interval{ 0x2102, 0x2102 }, // L&       DOUBLE-STRUCK CAPITAL C
    Interval{ 0x2107, 0x2107 }, // L&       EULER CONSTANT
    Interval{ 0x210A, 0x2113 }, // L&  [10] SCRIPT SMALL G..SCRIPT SMALL L
    Interval{ 0x2115, 0x2115 }, // L&       DOUBLE-STRUCK CAPITAL N
    Interval{ 0x2118, 0x2118 }, // Sm       SCRIPT CAPITAL P
    Interval{ 0x2119, 0x211D }, // L&   [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    Interval{ 0x2124, 0x2124 }, // L&       DOUBLE-STRUCK CAPITAL Z
    Interval{ 0x2128, 0x2128 }, // L&       BLACK-LETTER CAPITAL Z
    Interval{ 0x2129, 0x2129 }, // So       TURNED GREEK SMALL LETTER IOTA
    Interval{ 0x212C, 0x212D }, // L&   [2] SCRIPT CAPITAL B..BLACK-LETTER CAPITAL C
    Interval{ 0x212F, 0x2131 }, // L&   [3] SCRIPT SMALL E..SCRIPT CAPITAL F
    Interval{ 0x2133, 0x2134 }, // L&   [2] SCRIPT CAPITAL M..SCRIPT SMALL O
    Interval{ 0x2135, 0x2138 }, // Lo   [4] ALEF SYMBOL..DALET SYMBOL
    Interval{ 0x213C, 0x213F }, // L&   [4] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK CAPITAL PI
    Interval{ 0x2140, 0x2144 }, // Sm   [5] DOUBLE-STRUCK N-ARY SUMMATION..TURNED SANS-SERIF CAPITAL Y
    Interval{ 0x2145, 0x2149 }, // L&   [5] DOUBLE-STRUCK ITALIC CAPITAL D..DOUBLE-STRUCK ITALIC SMALL J
    Interval{ 0x214B, 0x214B }, // Sm       TURNED AMPERSAND
    Interval{ 0x2190, 0x2194 }, // Sm   [5] LEFTWARDS ARROW..LEFT RIGHT ARROW
    Interval{ 0x2195, 0x2199 }, // So   [5] UP DOWN ARROW..SOUTH WEST ARROW
    Interval{ 0x219A, 0x219B }, // Sm   [2] LEFTWARDS ARROW WITH STROKE..RIGHTWARDS ARROW WITH STROKE
    Interval{ 0x219C, 0x219F }, // So   [4] LEFTWARDS WAVE ARROW..UPWARDS TWO HEADED ARROW
    Interval{ 0x21A0, 0x21A0 }, // Sm       RIGHTWARDS TWO HEADED ARROW
    Interval{ 0x21A1, 0x21A2 }, // So   [2] DOWNWARDS TWO HEADED ARROW..LEFTWARDS ARROW WITH TAIL
    Interval{ 0x21A3, 0x21A3 }, // Sm       RIGHTWARDS ARROW WITH TAIL
    Interval{ 0x21A4, 0x21A5 }, // So   [2] LEFTWARDS ARROW FROM BAR..UPWARDS ARROW FROM BAR
    Interval{ 0x21A6, 0x21A6 }, // Sm       RIGHTWARDS ARROW FROM BAR
    Interval{ 0x21A7, 0x21A7 }, // So       DOWNWARDS ARROW FROM BAR
    Interval{ 0x21A9, 0x21AD }, // So   [5] LEFTWARDS ARROW WITH HOOK..LEFT RIGHT WAVE ARROW
    Interval{ 0x21AE, 0x21AE }, // Sm       LEFT RIGHT ARROW WITH STROKE
    Interval{ 0x21B0, 0x21B1 }, // So   [2] UPWARDS ARROW WITH TIP LEFTWARDS..UPWARDS ARROW WITH TIP RIGHTWARDS
    Interval{ 0x21B6, 0x21B7 }, // So   [2] ANTICLOCKWISE TOP SEMICIRCLE ARROW..CLOCKWISE TOP SEMICIRCLE ARROW
    Interval{ 0x21BC, 0x21CD }, // So  [18] LEFTWARDS HARPOON WITH BARB UPWARDS..LEFTWARDS DOUBLE ARROW WITH STROKE
    Interval{ 0x21CE, 0x21CF }, // Sm   [2] LEFT RIGHT DOUBLE ARROW WITH STROKE..RIGHTWARDS DOUBLE ARROW WITH STROKE
    Interval{ 0x21D0, 0x21D1 }, // So   [2] LEFTWARDS DOUBLE ARROW..UPWARDS DOUBLE ARROW
    Interval{ 0x21D2, 0x21D2 }, // Sm       RIGHTWARDS DOUBLE ARROW
    Interval{ 0x21D3, 0x21D3 }, // So       DOWNWARDS DOUBLE ARROW
    Interval{ 0x21D4, 0x21D4 }, // Sm       LEFT RIGHT DOUBLE ARROW
    Interval{ 0x21D5, 0x21DB }, // So   [7] UP DOWN DOUBLE ARROW..RIGHTWARDS TRIPLE ARROW
    Interval{ 0x21DD, 0x21DD }, // So       RIGHTWARDS SQUIGGLE ARROW
    Interval{ 0x21E4, 0x21E5 }, // So   [2] LEFTWARDS ARROW TO BAR..RIGHTWARDS ARROW TO BAR
    Interval{ 0x21F4, 0x22FF }, // Sm [268] RIGHT ARROW WITH SMALL CIRCLE..Z NOTATION BAG MEMBERSHIP
    Interval{ 0x2308, 0x2308 }, // Ps       LEFT CEILING
    Interval{ 0x2309, 0x2309 }, // Pe       RIGHT CEILING
    Interval{ 0x230A, 0x230A }, // Ps       LEFT FLOOR
    Interval{ 0x230B, 0x230B }, // Pe       RIGHT FLOOR
    Interval{ 0x2320, 0x2321 }, // Sm   [2] TOP HALF INTEGRAL..BOTTOM HALF INTEGRAL
    Interval{ 0x237C, 0x237C }, // Sm       RIGHT ANGLE WITH DOWNWARDS ZIGZAG ARROW
    Interval{ 0x239B, 0x23B3 }, // Sm  [25] LEFT PARENTHESIS UPPER HOOK..SUMMATION BOTTOM
    Interval{ 0x23B4, 0x23B5 }, // So   [2] TOP SQUARE BRACKET..BOTTOM SQUARE BRACKET
    Interval{ 0x23B7, 0x23B7 }, // So       RADICAL SYMBOL BOTTOM
    Interval{ 0x23D0, 0x23D0 }, // So       VERTICAL LINE EXTENSION
    Interval{ 0x23DC, 0x23E1 }, // Sm   [6] TOP PARENTHESIS..BOTTOM TORTOISE SHELL BRACKET
    Interval{ 0x23E2, 0x23E2 }, // So       WHITE TRAPEZIUM
    Interval{ 0x25A0, 0x25A1 }, // So   [2] BLACK SQUARE..WHITE SQUARE
    Interval{ 0x25AE, 0x25B6 }, // So   [9] BLACK VERTICAL RECTANGLE..BLACK RIGHT-POINTING TRIANGLE
    Interval{ 0x25B7, 0x25B7 }, // Sm       WHITE RIGHT-POINTING TRIANGLE
    Interval{ 0x25BC, 0x25C0 }, // So   [5] BLACK DOWN-POINTING TRIANGLE..BLACK LEFT-POINTING TRIANGLE
    Interval{ 0x25C1, 0x25C1 }, // Sm       WHITE LEFT-POINTING TRIANGLE
    Interval{ 0x25C6, 0x25C7 }, // So   [2] BLACK DIAMOND..WHITE DIAMOND
    Interval{ 0x25CA, 0x25CB }, // So   [2] LOZENGE..WHITE CIRCLE
    Interval{ 0x25CF, 0x25D3 }, // So   [5] BLACK CIRCLE..CIRCLE WITH UPPER HALF BLACK
    Interval{ 0x25E2, 0x25E2 }, // So       BLACK LOWER RIGHT TRIANGLE
    Interval{ 0x25E4, 0x25E4 }, // So       BLACK UPPER LEFT TRIANGLE
    Interval{ 0x25E7, 0x25EC }, // So   [6] SQUARE WITH LEFT HALF BLACK..WHITE UP-POINTING TRIANGLE WITH DOT
    Interval{ 0x25F8, 0x25FF }, // Sm   [8] UPPER LEFT TRIANGLE..LOWER RIGHT TRIANGLE
    Interval{ 0x2605, 0x2606 }, // So   [2] BLACK STAR..WHITE STAR
    Interval{ 0x2640, 0x2640 }, // So       FEMALE SIGN
    Interval{ 0x2642, 0x2642 }, // So       MALE SIGN
    Interval{ 0x2660, 0x2663 }, // So   [4] BLACK SPADE SUIT..BLACK CLUB SUIT
    Interval{ 0x266D, 0x266E }, // So   [2] MUSIC FLAT SIGN..MUSIC NATURAL SIGN
    Interval{ 0x266F, 0x266F }, // Sm       MUSIC SHARP SIGN
    Interval{ 0x27C0, 0x27C4 }, // Sm   [5] THREE DIMENSIONAL ANGLE..OPEN SUPERSET
    Interval{ 0x27C5, 0x27C5 }, // Ps       LEFT S-SHAPED BAG DELIMITER
    Interval{ 0x27C6, 0x27C6 }, // Pe       RIGHT S-SHAPED BAG DELIMITER
    Interval{ 0x27C7, 0x27E5 }, // Sm  [31] OR WITH DOT INSIDE..WHITE SQUARE WITH RIGHTWARDS TICK
    Interval{ 0x27E6, 0x27E6 }, // Ps       MATHEMATICAL LEFT WHITE SQUARE BRACKET
    Interval{ 0x27E7, 0x27E7 }, // Pe       MATHEMATICAL RIGHT WHITE SQUARE BRACKET
    Interval{ 0x27E8, 0x27E8 }, // Ps       MATHEMATICAL LEFT ANGLE BRACKET
    Interval{ 0x27E9, 0x27E9 }, // Pe       MATHEMATICAL RIGHT ANGLE BRACKET
    Interval{ 0x27EA, 0x27EA }, // Ps       MATHEMATICAL LEFT DOUBLE ANGLE BRACKET
    Interval{ 0x27EB, 0x27EB }, // Pe       MATHEMATICAL RIGHT DOUBLE ANGLE BRACKET
    Interval{ 0x27EC, 0x27EC }, // Ps       MATHEMATICAL LEFT WHITE TORTOISE SHELL BRACKET
    Interval{ 0x27ED, 0x27ED }, // Pe       MATHEMATICAL RIGHT WHITE TORTOISE SHELL BRACKET
    Interval{ 0x27EE, 0x27EE }, // Ps       MATHEMATICAL LEFT FLATTENED PARENTHESIS
    Interval{ 0x27EF, 0x27EF }, // Pe       MATHEMATICAL RIGHT FLATTENED PARENTHESIS
    Interval{ 0x27F0, 0x27FF }, // Sm  [16] UPWARDS QUADRUPLE ARROW..LONG RIGHTWARDS SQUIGGLE ARROW
    Interval{ 0x2900, 0x2982 }, // Sm [131] RIGHTWARDS TWO-HEADED ARROW WITH VERTICAL STROKE..Z NOTATION TYPE COLON
    Interval{ 0x2983, 0x2983 }, // Ps       LEFT WHITE CURLY BRACKET
    Interval{ 0x2984, 0x2984 }, // Pe       RIGHT WHITE CURLY BRACKET
    Interval{ 0x2985, 0x2985 }, // Ps       LEFT WHITE PARENTHESIS
    Interval{ 0x2986, 0x2986 }, // Pe       RIGHT WHITE PARENTHESIS
    Interval{ 0x2987, 0x2987 }, // Ps       Z NOTATION LEFT IMAGE BRACKET
    Interval{ 0x2988, 0x2988 }, // Pe       Z NOTATION RIGHT IMAGE BRACKET
    Interval{ 0x2989, 0x2989 }, // Ps       Z NOTATION LEFT BINDING BRACKET
    Interval{ 0x298A, 0x298A }, // Pe       Z NOTATION RIGHT BINDING BRACKET
    Interval{ 0x298B, 0x298B }, // Ps       LEFT SQUARE BRACKET WITH UNDERBAR
    Interval{ 0x298C, 0x298C }, // Pe       RIGHT SQUARE BRACKET WITH UNDERBAR
    Interval{ 0x298D, 0x298D }, // Ps       LEFT SQUARE BRACKET WITH TICK IN TOP CORNER
    Interval{ 0x298E, 0x298E }, // Pe       RIGHT SQUARE BRACKET WITH TICK IN BOTTOM CORNER
    Interval{ 0x298F, 0x298F }, // Ps       LEFT SQUARE BRACKET WITH TICK IN BOTTOM CORNER
    Interval{ 0x2990, 0x2990 }, // Pe       RIGHT SQUARE BRACKET WITH TICK IN TOP CORNER
    Interval{ 0x2991, 0x2991 }, // Ps       LEFT ANGLE BRACKET WITH DOT
    Interval{ 0x2992, 0x2992 }, // Pe       RIGHT ANGLE BRACKET WITH DOT
    Interval{ 0x2993, 0x2993 }, // Ps       LEFT ARC LESS-THAN BRACKET
    Interval{ 0x2994, 0x2994 }, // Pe       RIGHT ARC GREATER-THAN BRACKET
    Interval{ 0x2995, 0x2995 }, // Ps       DOUBLE LEFT ARC GREATER-THAN BRACKET
    Interval{ 0x2996, 0x2996 }, // Pe       DOUBLE RIGHT ARC LESS-THAN BRACKET
    Interval{ 0x2997, 0x2997 }, // Ps       LEFT BLACK TORTOISE SHELL BRACKET
    Interval{ 0x2998, 0x2998 }, // Pe       RIGHT BLACK TORTOISE SHELL BRACKET
    Interval{ 0x2999, 0x29D7 }, // Sm  [63] DOTTED FENCE..BLACK HOURGLASS
    Interval{ 0x29D8, 0x29D8 }, // Ps       LEFT WIGGLY FENCE
    Interval{ 0x29D9, 0x29D9 }, // Pe       RIGHT WIGGLY FENCE
    Interval{ 0x29DA, 0x29DA }, // Ps       LEFT DOUBLE WIGGLY FENCE
    Interval{ 0x29DB, 0x29DB }, // Pe       RIGHT DOUBLE WIGGLY FENCE
    Interval{ 0x29DC, 0x29FB }, // Sm  [32] INCOMPLETE INFINITY..TRIPLE PLUS
    Interval{ 0x29FC, 0x29FC }, // Ps       LEFT-POINTING CURVED ANGLE BRACKET
    Interval{ 0x29FD, 0x29FD }, // Pe       RIGHT-POINTING CURVED ANGLE BRACKET
    Interval{ 0x29FE, 0x2AFF }, // Sm [258] TINY..N-ARY WHITE VERTICAL BAR
    Interval{ 0x2B30, 0x2B44 }, // Sm  [21] LEFT ARROW WITH SMALL CIRCLE..RIGHTWARDS ARROW THROUGH SUPERSET
    Interval{ 0x2B47, 0x2B4C }, // Sm   [6] REVERSE TILDE OPERATOR ABOVE RIGHTWARDS ARROW..RIGHTWARDS ARROW ABOVE REVERSE TILDE OPERATOR
    Interval{ 0xFB29, 0xFB29 }, // Sm       HEBREW LETTER ALTERNATIVE PLUS SIGN
    Interval{ 0xFE61, 0xFE61 }, // Po       SMALL ASTERISK
    Interval{ 0xFE62, 0xFE62 }, // Sm       SMALL PLUS SIGN
    Interval{ 0xFE63, 0xFE63 }, // Pd       SMALL HYPHEN-MINUS
    Interval{ 0xFE64, 0xFE66 }, // Sm   [3] SMALL LESS-THAN SIGN..SMALL EQUALS SIGN
    Interval{ 0xFE68, 0xFE68 }, // Po       SMALL REVERSE SOLIDUS
    Interval{ 0xFF0B, 0xFF0B }, // Sm       FULLWIDTH PLUS SIGN
    Interval{ 0xFF1C, 0xFF1E }, // Sm   [3] FULLWIDTH LESS-THAN SIGN..FULLWIDTH GREATER-THAN SIGN
    Interval{ 0xFF3C, 0xFF3C }, // Po       FULLWIDTH REVERSE SOLIDUS
    Interval{ 0xFF3E, 0xFF3E }, // Sk       FULLWIDTH CIRCUMFLEX ACCENT
    Interval{ 0xFF5C, 0xFF5C }, // Sm       FULLWIDTH VERTICAL LINE
    Interval{ 0xFF5E, 0xFF5E }, // Sm       FULLWIDTH TILDE
    Interval{ 0xFFE2, 0xFFE2 }, // Sm       FULLWIDTH NOT SIGN
    Interval{ 0xFFE9, 0xFFEC }, // Sm   [4] HALFWIDTH LEFTWARDS ARROW..HALFWIDTH DOWNWARDS ARROW
    Interval{ 0x10D8E, 0x10D8F }, // Sm   [2] GARAY PLUS SIGN..GARAY MINUS SIGN
    Interval{ 0x1CEF0, 0x1CEF0 }, // Sm       MEDIUM SMALL WHITE CIRCLE WITH HORIZONTAL BAR
    Interval{ 0x1D400, 0x1D454 }, // L&  [85] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL ITALIC SMALL G
    Interval{ 0x1D456, 0x1D49C }, // L&  [71] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL SCRIPT CAPITAL A
    Interval{ 0x1D49E, 0x1D49F }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Interval{ 0x1D4A2, 0x1D4A2 }, // L&       MATHEMATICAL SCRIPT CAPITAL G
    Interval{ 0x1D4A5, 0x1D4A6 }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Interval{ 0x1D4A9, 0x1D4AC }, // L&   [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Interval{ 0x1D4AE, 0x1D4B9 }, // L&  [12] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT SMALL D
    Interval{ 0x1D4BB, 0x1D4BB }, // L&       MATHEMATICAL SCRIPT SMALL F
    Interval{ 0x1D4BD, 0x1D4C3 }, // L&   [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Interval{ 0x1D4C5, 0x1D505 }, // L&  [65] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL FRAKTUR CAPITAL B
    Interval{ 0x1D507, 0x1D50A }, // L&   [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Interval{ 0x1D50D, 0x1D514 }, // L&   [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Interval{ 0x1D516, 0x1D51C }, // L&   [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Interval{ 0x1D51E, 0x1D539 }, // L&  [28] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Interval{ 0x1D53B, 0x1D53E }, // L&   [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Interval{ 0x1D540, 0x1D544 }, // L&   [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Interval{ 0x1D546, 0x1D546 }, // L&       MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Interval{ 0x1D54A, 0x1D550 }, // L&   [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Interval{ 0x1D552, 0x1D6A5 }, // L& [340] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    Interval{ 0x1D6A8, 0x1D6C0 }, // L&  [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    Interval{ 0x1D6C1, 0x1D6C1 }, // Sm       MATHEMATICAL BOLD NABLA
    Interval{ 0x1D6C2, 0x1D6DA }, // L&  [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    Interval{ 0x1D6DB, 0x1D6DB }, // Sm       MATHEMATICAL BOLD PARTIAL DIFFERENTIAL
    Interval{ 0x1D6DC, 0x1D6FA }, // L&  [31] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL ITALIC CAPITAL OMEGA
    Interval{ 0x1D6FB, 0x1D6FB }, // Sm       MATHEMATICAL ITALIC NABLA
    Interval{ 0x1D6FC, 0x1D714 }, // L&  [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    Interval{ 0x1D715, 0x1D715 }, // Sm       MATHEMATICAL ITALIC PARTIAL DIFFERENTIAL
    Interval{ 0x1D716, 0x1D734 }, // L&  [31] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D735, 0x1D735 }, // Sm       MATHEMATICAL BOLD ITALIC NABLA
    Interval{ 0x1D736, 0x1D74E }, // L&  [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D74F, 0x1D74F }, // Sm       MATHEMATICAL BOLD ITALIC PARTIAL DIFFERENTIAL
    Interval{ 0x1D750, 0x1D76E }, // L&  [31] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    Interval{ 0x1D76F, 0x1D76F }, // Sm       MATHEMATICAL SANS-SERIF BOLD NABLA
    Interval{ 0x1D770, 0x1D788 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    Interval{ 0x1D789, 0x1D789 }, // Sm       MATHEMATICAL SANS-SERIF BOLD PARTIAL DIFFERENTIAL
    Interval{ 0x1D78A, 0x1D7A8 }, // L&  [31] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D7A9, 0x1D7A9 }, // Sm       MATHEMATICAL SANS-SERIF BOLD ITALIC NABLA
    Interval{ 0x1D7AA, 0x1D7C2 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D7C3, 0x1D7C3 }, // Sm       MATHEMATICAL SANS-SERIF BOLD ITALIC PARTIAL DIFFERENTIAL
    Interval{ 0x1D7C4, 0x1D7CB }, // L&   [8] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD SMALL DIGAMMA
    Interval{ 0x1D7CE, 0x1D7FF }, // Nd  [50] MATHEMATICAL BOLD DIGIT ZERO..MATHEMATICAL MONOSPACE DIGIT NINE
    Interval{ 0x1EE00, 0x1EE03 }, // Lo   [4] ARABIC MATHEMATICAL ALEF..ARABIC MATHEMATICAL DAL
    Interval{ 0x1EE05, 0x1EE1F }, // Lo  [27] ARABIC MATHEMATICAL WAW..ARABIC MATHEMATICAL DOTLESS QAF
    Interval{ 0x1EE21, 0x1EE22 }, // Lo   [2] ARABIC MATHEMATICAL INITIAL BEH..ARABIC MATHEMATICAL INITIAL JEEM
    Interval{ 0x1EE24, 0x1EE24 }, // Lo       ARABIC MATHEMATICAL INITIAL HEH
    Interval{ 0x1EE27, 0x1EE27 }, // Lo       ARABIC MATHEMATICAL INITIAL HAH
    Interval{ 0x1EE29, 0x1EE32 }, // Lo  [10] ARABIC MATHEMATICAL INITIAL YEH..ARABIC MATHEMATICAL INITIAL QAF
    Interval{ 0x1EE34, 0x1EE37 }, // Lo   [4] ARABIC MATHEMATICAL INITIAL SHEEN..ARABIC MATHEMATICAL INITIAL KHAH
    Interval{ 0x1EE39, 0x1EE39 }, // Lo       ARABIC MATHEMATICAL INITIAL DAD
    Interval{ 0x1EE3B, 0x1EE3B }, // Lo       ARABIC MATHEMATICAL INITIAL GHAIN
    Interval{ 0x1EE42, 0x1EE42 }, // Lo       ARABIC MATHEMATICAL TAILED JEEM
    Interval{ 0x1EE47, 0x1EE47 }, // Lo       ARABIC MATHEMATICAL TAILED HAH
    Interval{ 0x1EE49, 0x1EE49 }, // Lo       ARABIC MATHEMATICAL TAILED YEH
    Interval{ 0x1EE4B, 0x1EE4B }, // Lo       ARABIC MATHEMATICAL TAILED LAM
    Interval{ 0x1EE4D, 0x1EE4F }, // Lo   [3] ARABIC MATHEMATICAL TAILED NOON..ARABIC MATHEMATICAL TAILED AIN
    Interval{ 0x1EE51, 0x1EE52 }, // Lo   [2] ARABIC MATHEMATICAL TAILED SAD..ARABIC MATHEMATICAL TAILED QAF
    Interval{ 0x1EE54, 0x1EE54 }, // Lo       ARABIC MATHEMATICAL TAILED SHEEN
    Interval{ 0x1EE57, 0x1EE57 }, // Lo       ARABIC MATHEMATICAL TAILED KHAH
    Interval{ 0x1EE59, 0x1EE59 }, // Lo       ARABIC MATHEMATICAL TAILED DAD
    Interval{ 0x1EE5B, 0x1EE5B }, // Lo       ARABIC MATHEMATICAL TAILED GHAIN
    Interval{ 0x1EE5D, 0x1EE5D }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS NOON
    Interval{ 0x1EE5F, 0x1EE5F }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS QAF
    Interval{ 0x1EE61, 0x1EE62 }, // Lo   [2] ARABIC MATHEMATICAL STRETCHED BEH..ARABIC MATHEMATICAL STRETCHED JEEM
    Interval{ 0x1EE64, 0x1EE64 }, // Lo       ARABIC MATHEMATICAL STRETCHED HEH
    Interval{ 0x1EE67, 0x1EE6A }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED HAH..ARABIC MATHEMATICAL STRETCHED KAF
    Interval{ 0x1EE6C, 0x1EE72 }, // Lo   [7] ARABIC MATHEMATICAL STRETCHED MEEM..ARABIC MATHEMATICAL STRETCHED QAF
    Interval{ 0x1EE74, 0x1EE77 }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED SHEEN..ARABIC MATHEMATICAL STRETCHED KHAH
    Interval{ 0x1EE79, 0x1EE7C }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED DAD..ARABIC MATHEMATICAL STRETCHED DOTLESS BEH
    Interval{ 0x1EE7E, 0x1EE7E }, // Lo       ARABIC MATHEMATICAL STRETCHED DOTLESS FEH
    Interval{ 0x1EE80, 0x1EE89 }, // Lo  [10] ARABIC MATHEMATICAL LOOPED ALEF..ARABIC MATHEMATICAL LOOPED YEH
    Interval{ 0x1EE8B, 0x1EE9B }, // Lo  [17] ARABIC MATHEMATICAL LOOPED LAM..ARABIC MATHEMATICAL LOOPED GHAIN
    Interval{ 0x1EEA1, 0x1EEA3 }, // Lo   [3] ARABIC MATHEMATICAL DOUBLE-STRUCK BEH..ARABIC MATHEMATICAL DOUBLE-STRUCK DAL
    Interval{ 0x1EEA5, 0x1EEA9 }, // Lo   [5] ARABIC MATHEMATICAL DOUBLE-STRUCK WAW..ARABIC MATHEMATICAL DOUBLE-STRUCK YEH
    Interval{ 0x1EEAB, 0x1EEBB }, // Lo  [17] ARABIC MATHEMATICAL DOUBLE-STRUCK LAM..ARABIC MATHEMATICAL DOUBLE-STRUCK GHAIN
    Interval{ 0x1EEF0, 0x1EEF1 }, // Sm   [2] ARABIC MATHEMATICAL OPERATOR MEEM WITH HAH WITH TATWEEL..ARABIC MATHEMATICAL OPERATOR HAH WITH DAL
    Interval{ 0x1F8D0, 0x1F8D8 }, // Sm   [9] LONG RIGHTWARDS ARROW OVER LONG LEFTWARDS ARROW..LONG LEFT RIGHT ARROW WITH DEPENDENT LOBE
}; // }}}
auto static const Uppercase = std::array<Interval, 660>{ // {{{
    Interval{ 0x0041, 0x005A }, // L&  [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Interval{ 0x00C0, 0x00D6 }, // L&  [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x00D8, 0x00DE }, // L&   [7] LATIN CAPITAL LETTER O WITH STROKE..LATIN CAPITAL LETTER THORN
    Interval{ 0x0100, 0x0100 }, // L&       LATIN CAPITAL LETTER A WITH MACRON
    Interval{ 0x0102, 0x0102 }, // L&       LATIN CAPITAL LETTER A WITH BREVE
    Interval{ 0x0104, 0x0104 }, // L&       LATIN CAPITAL LETTER A WITH OGONEK
    Interval{ 0x0106, 0x0106 }, // L&       LATIN CAPITAL LETTER C WITH ACUTE
    Interval{ 0x0108, 0x0108 }, // L&       LATIN CAPITAL LETTER C WITH CIRCUMFLEX
    Interval{ 0x010A, 0x010A }, // L&       LATIN CAPITAL LETTER C WITH DOT ABOVE
    Interval{ 0x010C, 0x010C }, // L&       LATIN CAPITAL LETTER C WITH CARON
    Interval{ 0x010E, 0x010E }, // L&       LATIN CAPITAL LETTER D WITH CARON
    Interval{ 0x0110, 0x0110 }, // L&       LATIN CAPITAL LETTER D WITH STROKE
    Interval{ 0x0112, 0x0112 }, // L&       LATIN CAPITAL LETTER E WITH MACRON
    Interval{ 0x0114, 0x0114 }, // L&       LATIN CAPITAL LETTER E WITH BREVE
    Interval{ 0x0116, 0x0116 }, // L&       LATIN CAPITAL LETTER E WITH DOT ABOVE
    Interval{ 0x0118, 0x0118 }, // L&       LATIN CAPITAL LETTER E WITH OGONEK
    Interval{ 0x011A, 0x011A }, // L&       LATIN CAPITAL LETTER E WITH CARON
    Interval{ 0x011C, 0x011C }, // L&       LATIN CAPITAL LETTER G WITH CIRCUMFLEX
    Interval{ 0x011E, 0x011E }, // L&       LATIN CAPITAL LETTER G WITH BREVE
    Interval{ 0x0120, 0x0120 }, // L&       LATIN CAPITAL LETTER G WITH DOT ABOVE
    Interval{ 0x0122, 0x0122 }, // L&       LATIN CAPITAL LETTER G WITH CEDILLA
    Interval{ 0x0124, 0x0124 }, // L&       LATIN CAPITAL LETTER H WITH CIRCUMFLEX
    Interval{ 0x0126, 0x0126 }, // L&       LATIN CAPITAL LETTER H WITH STROKE
    Interval{ 0x0128, 0x0128 }, // L&       LATIN CAPITAL LETTER I WITH TILDE
    Interval{ 0x012A, 0x012A }, // L&       LATIN CAPITAL LETTER I WITH MACRON
    Interval{ 0x012C, 0x012C }, // L&       LATIN CAPITAL LETTER I WITH BREVE
    Interval{ 0x012E, 0x012E }, // L&       LATIN CAPITAL LETTER I WITH OGONEK
    Interval{ 0x0130, 0x0130 }, // L&       LATIN CAPITAL LETTER I WITH DOT ABOVE
    Interval{ 0x0132, 0x0132 }, // L&       LATIN CAPITAL LIGATURE IJ
    Interval{ 0x0134, 0x0134 }, // L&       LATIN CAPITAL LETTER J WITH CIRCUMFLEX
    Interval{ 0x0136, 0x0136 }, // L&       LATIN CAPITAL LETTER K WITH CEDILLA
    Interval{ 0x0139, 0x0139 }, // L&       LATIN CAPITAL LETTER L WITH ACUTE
    Interval{ 0x013B, 0x013B }, // L&       LATIN CAPITAL LETTER L WITH CEDILLA
    Interval{ 0x013D, 0x013D }, // L&       LATIN CAPITAL LETTER L WITH CARON
    Interval{ 0x013F, 0x013F }, // L&       LATIN CAPITAL LETTER L WITH MIDDLE DOT
    Interval{ 0x0141, 0x0141 }, // L&       LATIN CAPITAL LETTER L WITH STROKE
    Interval{ 0x0143, 0x0143 }, // L&       LATIN CAPITAL LETTER N WITH ACUTE
    Interval{ 0x0145, 0x0145 }, // L&       LATIN CAPITAL LETTER N WITH CEDILLA
    Interval{ 0x0147, 0x0147 }, // L&       LATIN CAPITAL LETTER N WITH CARON
    Interval{ 0x014A, 0x014A }, // L&       LATIN CAPITAL LETTER ENG
    Interval{ 0x014C, 0x014C }, // L&       LATIN CAPITAL LETTER O WITH MACRON
    Interval{ 0x014E, 0x014E }, // L&       LATIN CAPITAL LETTER O WITH BREVE
    Interval{ 0x0150, 0x0150 }, // L&       LATIN CAPITAL LETTER O WITH DOUBLE ACUTE
    Interval{ 0x0152, 0x0152 }, // L&       LATIN CAPITAL LIGATURE OE
    Interval{ 0x0154, 0x0154 }, // L&       LATIN CAPITAL LETTER R WITH ACUTE
    Interval{ 0x0156, 0x0156 }, // L&       LATIN CAPITAL LETTER R WITH CEDILLA
    Interval{ 0x0158, 0x0158 }, // L&       LATIN CAPITAL LETTER R WITH CARON
    Interval{ 0x015A, 0x015A }, // L&       LATIN CAPITAL LETTER S WITH ACUTE
    Interval{ 0x015C, 0x015C }, // L&       LATIN CAPITAL LETTER S WITH CIRCUMFLEX
    Interval{ 0x015E, 0x015E }, // L&       LATIN CAPITAL LETTER S WITH CEDILLA
    Interval{ 0x0160, 0x0160 }, // L&       LATIN CAPITAL LETTER S WITH CARON
    Interval{ 0x0162, 0x0162 }, // L&       LATIN CAPITAL LETTER T WITH CEDILLA
    Interval{ 0x0164, 0x0164 }, // L&       LATIN CAPITAL LETTER T WITH CARON
    Interval{ 0x0166, 0x0166 }, // L&       LATIN CAPITAL LETTER T WITH STROKE
    Interval{ 0x0168, 0x0168 }, // L&       LATIN CAPITAL LETTER U WITH TILDE
    Interval{ 0x016A, 0x016A }, // L&       LATIN CAPITAL LETTER U WITH MACRON
    Interval{ 0x016C, 0x016C }, // L&       LATIN CAPITAL LETTER U WITH BREVE
    Interval{ 0x016E, 0x016E }, // L&       LATIN CAPITAL LETTER U WITH RING ABOVE
    Interval{ 0x0170, 0x0170 }, // L&       LATIN CAPITAL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x0172, 0x0172 }, // L&       LATIN CAPITAL LETTER U WITH OGONEK
    Interval{ 0x0174, 0x0174 }, // L&       LATIN CAPITAL LETTER W WITH CIRCUMFLEX
    Interval{ 0x0176, 0x0176 }, // L&       LATIN CAPITAL LETTER Y WITH CIRCUMFLEX
    Interval{ 0x0178, 0x0179 }, // L&   [2] LATIN CAPITAL LETTER Y WITH DIAERESIS..LATIN CAPITAL LETTER Z WITH ACUTE
    Interval{ 0x017B, 0x017B }, // L&       LATIN CAPITAL LETTER Z WITH DOT ABOVE
    Interval{ 0x017D, 0x017D }, // L&       LATIN CAPITAL LETTER Z WITH CARON
    Interval{ 0x0181, 0x0182 }, // L&   [2] LATIN CAPITAL LETTER B WITH HOOK..LATIN CAPITAL LETTER B WITH TOPBAR
    Interval{ 0x0184, 0x0184 }, // L&       LATIN CAPITAL LETTER TONE SIX
    Interval{ 0x0186, 0x0187 }, // L&   [2] LATIN CAPITAL LETTER OPEN O..LATIN CAPITAL LETTER C WITH HOOK
    Interval{ 0x0189, 0x018B }, // L&   [3] LATIN CAPITAL LETTER AFRICAN D..LATIN CAPITAL LETTER D WITH TOPBAR
    Interval{ 0x018E, 0x0191 }, // L&   [4] LATIN CAPITAL LETTER REVERSED E..LATIN CAPITAL LETTER F WITH HOOK
    Interval{ 0x0193, 0x0194 }, // L&   [2] LATIN CAPITAL LETTER G WITH HOOK..LATIN CAPITAL LETTER GAMMA
    Interval{ 0x0196, 0x0198 }, // L&   [3] LATIN CAPITAL LETTER IOTA..LATIN CAPITAL LETTER K WITH HOOK
    Interval{ 0x019C, 0x019D }, // L&   [2] LATIN CAPITAL LETTER TURNED M..LATIN CAPITAL LETTER N WITH LEFT HOOK
    Interval{ 0x019F, 0x01A0 }, // L&   [2] LATIN CAPITAL LETTER O WITH MIDDLE TILDE..LATIN CAPITAL LETTER O WITH HORN
    Interval{ 0x01A2, 0x01A2 }, // L&       LATIN CAPITAL LETTER OI
    Interval{ 0x01A4, 0x01A4 }, // L&       LATIN CAPITAL LETTER P WITH HOOK
    Interval{ 0x01A6, 0x01A7 }, // L&   [2] LATIN LETTER YR..LATIN CAPITAL LETTER TONE TWO
    Interval{ 0x01A9, 0x01A9 }, // L&       LATIN CAPITAL LETTER ESH
    Interval{ 0x01AC, 0x01AC }, // L&       LATIN CAPITAL LETTER T WITH HOOK
    Interval{ 0x01AE, 0x01AF }, // L&   [2] LATIN CAPITAL LETTER T WITH RETROFLEX HOOK..LATIN CAPITAL LETTER U WITH HORN
    Interval{ 0x01B1, 0x01B3 }, // L&   [3] LATIN CAPITAL LETTER UPSILON..LATIN CAPITAL LETTER Y WITH HOOK
    Interval{ 0x01B5, 0x01B5 }, // L&       LATIN CAPITAL LETTER Z WITH STROKE
    Interval{ 0x01B7, 0x01B8 }, // L&   [2] LATIN CAPITAL LETTER EZH..LATIN CAPITAL LETTER EZH REVERSED
    Interval{ 0x01BC, 0x01BC }, // L&       LATIN CAPITAL LETTER TONE FIVE
    Interval{ 0x01C4, 0x01C4 }, // L&       LATIN CAPITAL LETTER DZ WITH CARON
    Interval{ 0x01C7, 0x01C7 }, // L&       LATIN CAPITAL LETTER LJ
    Interval{ 0x01CA, 0x01CA }, // L&       LATIN CAPITAL LETTER NJ
    Interval{ 0x01CD, 0x01CD }, // L&       LATIN CAPITAL LETTER A WITH CARON
    Interval{ 0x01CF, 0x01CF }, // L&       LATIN CAPITAL LETTER I WITH CARON
    Interval{ 0x01D1, 0x01D1 }, // L&       LATIN CAPITAL LETTER O WITH CARON
    Interval{ 0x01D3, 0x01D3 }, // L&       LATIN CAPITAL LETTER U WITH CARON
    Interval{ 0x01D5, 0x01D5 }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS AND MACRON
    Interval{ 0x01D7, 0x01D7 }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS AND ACUTE
    Interval{ 0x01D9, 0x01D9 }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS AND CARON
    Interval{ 0x01DB, 0x01DB }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS AND GRAVE
    Interval{ 0x01DE, 0x01DE }, // L&       LATIN CAPITAL LETTER A WITH DIAERESIS AND MACRON
    Interval{ 0x01E0, 0x01E0 }, // L&       LATIN CAPITAL LETTER A WITH DOT ABOVE AND MACRON
    Interval{ 0x01E2, 0x01E2 }, // L&       LATIN CAPITAL LETTER AE WITH MACRON
    Interval{ 0x01E4, 0x01E4 }, // L&       LATIN CAPITAL LETTER G WITH STROKE
    Interval{ 0x01E6, 0x01E6 }, // L&       LATIN CAPITAL LETTER G WITH CARON
    Interval{ 0x01E8, 0x01E8 }, // L&       LATIN CAPITAL LETTER K WITH CARON
    Interval{ 0x01EA, 0x01EA }, // L&       LATIN CAPITAL LETTER O WITH OGONEK
    Interval{ 0x01EC, 0x01EC }, // L&       LATIN CAPITAL LETTER O WITH OGONEK AND MACRON
    Interval{ 0x01EE, 0x01EE }, // L&       LATIN CAPITAL LETTER EZH WITH CARON
    Interval{ 0x01F1, 0x01F1 }, // L&       LATIN CAPITAL LETTER DZ
    Interval{ 0x01F4, 0x01F4 }, // L&       LATIN CAPITAL LETTER G WITH ACUTE
    Interval{ 0x01F6, 0x01F8 }, // L&   [3] LATIN CAPITAL LETTER HWAIR..LATIN CAPITAL LETTER N WITH GRAVE
    Interval{ 0x01FA, 0x01FA }, // L&       LATIN CAPITAL LETTER A WITH RING ABOVE AND ACUTE
    Interval{ 0x01FC, 0x01FC }, // L&       LATIN CAPITAL LETTER AE WITH ACUTE
    Interval{ 0x01FE, 0x01FE }, // L&       LATIN CAPITAL LETTER O WITH STROKE AND ACUTE
    Interval{ 0x0200, 0x0200 }, // L&       LATIN CAPITAL LETTER A WITH DOUBLE GRAVE
    Interval{ 0x0202, 0x0202 }, // L&       LATIN CAPITAL LETTER A WITH INVERTED BREVE
    Interval{ 0x0204, 0x0204 }, // L&       LATIN CAPITAL LETTER E WITH DOUBLE GRAVE
    Interval{ 0x0206, 0x0206 }, // L&       LATIN CAPITAL LETTER E WITH INVERTED BREVE
    Interval{ 0x0208, 0x0208 }, // L&       LATIN CAPITAL LETTER I WITH DOUBLE GRAVE
    Interval{ 0x020A, 0x020A }, // L&       LATIN CAPITAL LETTER I WITH INVERTED BREVE
    Interval{ 0x020C, 0x020C }, // L&       LATIN CAPITAL LETTER O WITH DOUBLE GRAVE
    Interval{ 0x020E, 0x020E }, // L&       LATIN CAPITAL LETTER O WITH INVERTED BREVE
    Interval{ 0x0210, 0x0210 }, // L&       LATIN CAPITAL LETTER R WITH DOUBLE GRAVE
    Interval{ 0x0212, 0x0212 }, // L&       LATIN CAPITAL LETTER R WITH INVERTED BREVE
    Interval{ 0x0214, 0x0214 }, // L&       LATIN CAPITAL LETTER U WITH DOUBLE GRAVE
    Interval{ 0x0216, 0x0216 }, // L&       LATIN CAPITAL LETTER U WITH INVERTED BREVE
    Interval{ 0x0218, 0x0218 }, // L&       LATIN CAPITAL LETTER S WITH COMMA BELOW
    Interval{ 0x021A, 0x021A }, // L&       LATIN CAPITAL LETTER T WITH COMMA BELOW
    Interval{ 0x021C, 0x021C }, // L&       LATIN CAPITAL LETTER YOGH
    Interval{ 0x021E, 0x021E }, // L&       LATIN CAPITAL LETTER H WITH CARON
    Interval{ 0x0220, 0x0220 }, // L&       LATIN CAPITAL LETTER N WITH LONG RIGHT LEG
    Interval{ 0x0222, 0x0222 }, // L&       LATIN CAPITAL LETTER OU
    Interval{ 0x0224, 0x0224 }, // L&       LATIN CAPITAL LETTER Z WITH HOOK
    Interval{ 0x0226, 0x0226 }, // L&       LATIN CAPITAL LETTER A WITH DOT ABOVE
    Interval{ 0x0228, 0x0228 }, // L&       LATIN CAPITAL LETTER E WITH CEDILLA
    Interval{ 0x022A, 0x022A }, // L&       LATIN CAPITAL LETTER O WITH DIAERESIS AND MACRON
    Interval{ 0x022C, 0x022C }, // L&       LATIN CAPITAL LETTER O WITH TILDE AND MACRON
    Interval{ 0x022E, 0x022E }, // L&       LATIN CAPITAL LETTER O WITH DOT ABOVE
    Interval{ 0x0230, 0x0230 }, // L&       LATIN CAPITAL LETTER O WITH DOT ABOVE AND MACRON
    Interval{ 0x0232, 0x0232 }, // L&       LATIN CAPITAL LETTER Y WITH MACRON
    Interval{ 0x023A, 0x023B }, // L&   [2] LATIN CAPITAL LETTER A WITH STROKE..LATIN CAPITAL LETTER C WITH STROKE
    Interval{ 0x023D, 0x023E }, // L&   [2] LATIN CAPITAL LETTER L WITH BAR..LATIN CAPITAL LETTER T WITH DIAGONAL STROKE
    Interval{ 0x0241, 0x0241 }, // L&       LATIN CAPITAL LETTER GLOTTAL STOP
    Interval{ 0x0243, 0x0246 }, // L&   [4] LATIN CAPITAL LETTER B WITH STROKE..LATIN CAPITAL LETTER E WITH STROKE
    Interval{ 0x0248, 0x0248 }, // L&       LATIN CAPITAL LETTER J WITH STROKE
    Interval{ 0x024A, 0x024A }, // L&       LATIN CAPITAL LETTER SMALL Q WITH HOOK TAIL
    Interval{ 0x024C, 0x024C }, // L&       LATIN CAPITAL LETTER R WITH STROKE
    Interval{ 0x024E, 0x024E }, // L&       LATIN CAPITAL LETTER Y WITH STROKE
    Interval{ 0x0370, 0x0370 }, // L&       GREEK CAPITAL LETTER HETA
    Interval{ 0x0372, 0x0372 }, // L&       GREEK CAPITAL LETTER ARCHAIC SAMPI
    Interval{ 0x0376, 0x0376 }, // L&       GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037F, 0x037F }, // L&       GREEK CAPITAL LETTER YOT
    Interval{ 0x0386, 0x0386 }, // L&       GREEK CAPITAL LETTER ALPHA WITH TONOS
    Interval{ 0x0388, 0x038A }, // L&   [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Interval{ 0x038C, 0x038C }, // L&       GREEK CAPITAL LETTER OMICRON WITH TONOS
    Interval{ 0x038E, 0x038F }, // L&   [2] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER OMEGA WITH TONOS
    Interval{ 0x0391, 0x03A1 }, // L&  [17] GREEK CAPITAL LETTER ALPHA..GREEK CAPITAL LETTER RHO
    Interval{ 0x03A3, 0x03AB }, // L&   [9] GREEK CAPITAL LETTER SIGMA..GREEK CAPITAL LETTER UPSILON WITH DIALYTIKA
    Interval{ 0x03CF, 0x03CF }, // L&       GREEK CAPITAL KAI SYMBOL
    Interval{ 0x03D2, 0x03D4 }, // L&   [3] GREEK UPSILON WITH HOOK SYMBOL..GREEK UPSILON WITH DIAERESIS AND HOOK SYMBOL
    Interval{ 0x03D8, 0x03D8 }, // L&       GREEK LETTER ARCHAIC KOPPA
    Interval{ 0x03DA, 0x03DA }, // L&       GREEK LETTER STIGMA
    Interval{ 0x03DC, 0x03DC }, // L&       GREEK LETTER DIGAMMA
    Interval{ 0x03DE, 0x03DE }, // L&       GREEK LETTER KOPPA
    Interval{ 0x03E0, 0x03E0 }, // L&       GREEK LETTER SAMPI
    Interval{ 0x03E2, 0x03E2 }, // L&       COPTIC CAPITAL LETTER SHEI
    Interval{ 0x03E4, 0x03E4 }, // L&       COPTIC CAPITAL LETTER FEI
    Interval{ 0x03E6, 0x03E6 }, // L&       COPTIC CAPITAL LETTER KHEI
    Interval{ 0x03E8, 0x03E8 }, // L&       COPTIC CAPITAL LETTER HORI
    Interval{ 0x03EA, 0x03EA }, // L&       COPTIC CAPITAL LETTER GANGIA
    Interval{ 0x03EC, 0x03EC }, // L&       COPTIC CAPITAL LETTER SHIMA
    Interval{ 0x03EE, 0x03EE }, // L&       COPTIC CAPITAL LETTER DEI
    Interval{ 0x03F4, 0x03F4 }, // L&       GREEK CAPITAL THETA SYMBOL
    Interval{ 0x03F7, 0x03F7 }, // L&       GREEK CAPITAL LETTER SHO
    Interval{ 0x03F9, 0x03FA }, // L&   [2] GREEK CAPITAL LUNATE SIGMA SYMBOL..GREEK CAPITAL LETTER SAN
    Interval{ 0x03FD, 0x042F }, // L&  [51] GREEK CAPITAL REVERSED LUNATE SIGMA SYMBOL..CYRILLIC CAPITAL LETTER YA
    Interval{ 0x0460, 0x0460 }, // L&       CYRILLIC CAPITAL LETTER OMEGA
    Interval{ 0x0462, 0x0462 }, // L&       CYRILLIC CAPITAL LETTER YAT
    Interval{ 0x0464, 0x0464 }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED E
    Interval{ 0x0466, 0x0466 }, // L&       CYRILLIC CAPITAL LETTER LITTLE YUS
    Interval{ 0x0468, 0x0468 }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED LITTLE YUS
    Interval{ 0x046A, 0x046A }, // L&       CYRILLIC CAPITAL LETTER BIG YUS
    Interval{ 0x046C, 0x046C }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED BIG YUS
    Interval{ 0x046E, 0x046E }, // L&       CYRILLIC CAPITAL LETTER KSI
    Interval{ 0x0470, 0x0470 }, // L&       CYRILLIC CAPITAL LETTER PSI
    Interval{ 0x0472, 0x0472 }, // L&       CYRILLIC CAPITAL LETTER FITA
    Interval{ 0x0474, 0x0474 }, // L&       CYRILLIC CAPITAL LETTER IZHITSA
    Interval{ 0x0476, 0x0476 }, // L&       CYRILLIC CAPITAL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT
    Interval{ 0x0478, 0x0478 }, // L&       CYRILLIC CAPITAL LETTER UK
    Interval{ 0x047A, 0x047A }, // L&       CYRILLIC CAPITAL LETTER ROUND OMEGA
    Interval{ 0x047C, 0x047C }, // L&       CYRILLIC CAPITAL LETTER OMEGA WITH TITLO
    Interval{ 0x047E, 0x047E }, // L&       CYRILLIC CAPITAL LETTER OT
    Interval{ 0x0480, 0x0480 }, // L&       CYRILLIC CAPITAL LETTER KOPPA
    Interval{ 0x048A, 0x048A }, // L&       CYRILLIC CAPITAL LETTER SHORT I WITH TAIL
    Interval{ 0x048C, 0x048C }, // L&       CYRILLIC CAPITAL LETTER SEMISOFT SIGN
    Interval{ 0x048E, 0x048E }, // L&       CYRILLIC CAPITAL LETTER ER WITH TICK
    Interval{ 0x0490, 0x0490 }, // L&       CYRILLIC CAPITAL LETTER GHE WITH UPTURN
    Interval{ 0x0492, 0x0492 }, // L&       CYRILLIC CAPITAL LETTER GHE WITH STROKE
    Interval{ 0x0494, 0x0494 }, // L&       CYRILLIC CAPITAL LETTER GHE WITH MIDDLE HOOK
    Interval{ 0x0496, 0x0496 }, // L&       CYRILLIC CAPITAL LETTER ZHE WITH DESCENDER
    Interval{ 0x0498, 0x0498 }, // L&       CYRILLIC CAPITAL LETTER ZE WITH DESCENDER
    Interval{ 0x049A, 0x049A }, // L&       CYRILLIC CAPITAL LETTER KA WITH DESCENDER
    Interval{ 0x049C, 0x049C }, // L&       CYRILLIC CAPITAL LETTER KA WITH VERTICAL STROKE
    Interval{ 0x049E, 0x049E }, // L&       CYRILLIC CAPITAL LETTER KA WITH STROKE
    Interval{ 0x04A0, 0x04A0 }, // L&       CYRILLIC CAPITAL LETTER BASHKIR KA
    Interval{ 0x04A2, 0x04A2 }, // L&       CYRILLIC CAPITAL LETTER EN WITH DESCENDER
    Interval{ 0x04A4, 0x04A4 }, // L&       CYRILLIC CAPITAL LIGATURE EN GHE
    Interval{ 0x04A6, 0x04A6 }, // L&       CYRILLIC CAPITAL LETTER PE WITH MIDDLE HOOK
    Interval{ 0x04A8, 0x04A8 }, // L&       CYRILLIC CAPITAL LETTER ABKHASIAN HA
    Interval{ 0x04AA, 0x04AA }, // L&       CYRILLIC CAPITAL LETTER ES WITH DESCENDER
    Interval{ 0x04AC, 0x04AC }, // L&       CYRILLIC CAPITAL LETTER TE WITH DESCENDER
    Interval{ 0x04AE, 0x04AE }, // L&       CYRILLIC CAPITAL LETTER STRAIGHT U
    Interval{ 0x04B0, 0x04B0 }, // L&       CYRILLIC CAPITAL LETTER STRAIGHT U WITH STROKE
    Interval{ 0x04B2, 0x04B2 }, // L&       CYRILLIC CAPITAL LETTER HA WITH DESCENDER
    Interval{ 0x04B4, 0x04B4 }, // L&       CYRILLIC CAPITAL LIGATURE TE TSE
    Interval{ 0x04B6, 0x04B6 }, // L&       CYRILLIC CAPITAL LETTER CHE WITH DESCENDER
    Interval{ 0x04B8, 0x04B8 }, // L&       CYRILLIC CAPITAL LETTER CHE WITH VERTICAL STROKE
    Interval{ 0x04BA, 0x04BA }, // L&       CYRILLIC CAPITAL LETTER SHHA
    Interval{ 0x04BC, 0x04BC }, // L&       CYRILLIC CAPITAL LETTER ABKHASIAN CHE
    Interval{ 0x04BE, 0x04BE }, // L&       CYRILLIC CAPITAL LETTER ABKHASIAN CHE WITH DESCENDER
    Interval{ 0x04C0, 0x04C1 }, // L&   [2] CYRILLIC LETTER PALOCHKA..CYRILLIC CAPITAL LETTER ZHE WITH BREVE
    Interval{ 0x04C3, 0x04C3 }, // L&       CYRILLIC CAPITAL LETTER KA WITH HOOK
    Interval{ 0x04C5, 0x04C5 }, // L&       CYRILLIC CAPITAL LETTER EL WITH TAIL
    Interval{ 0x04C7, 0x04C7 }, // L&       CYRILLIC CAPITAL LETTER EN WITH HOOK
    Interval{ 0x04C9, 0x04C9 }, // L&       CYRILLIC CAPITAL LETTER EN WITH TAIL
    Interval{ 0x04CB, 0x04CB }, // L&       CYRILLIC CAPITAL LETTER KHAKASSIAN CHE
    Interval{ 0x04CD, 0x04CD }, // L&       CYRILLIC CAPITAL LETTER EM WITH TAIL
    Interval{ 0x04D0, 0x04D0 }, // L&       CYRILLIC CAPITAL LETTER A WITH BREVE
    Interval{ 0x04D2, 0x04D2 }, // L&       CYRILLIC CAPITAL LETTER A WITH DIAERESIS
    Interval{ 0x04D4, 0x04D4 }, // L&       CYRILLIC CAPITAL LIGATURE A IE
    Interval{ 0x04D6, 0x04D6 }, // L&       CYRILLIC CAPITAL LETTER IE WITH BREVE
    Interval{ 0x04D8, 0x04D8 }, // L&       CYRILLIC CAPITAL LETTER SCHWA
    Interval{ 0x04DA, 0x04DA }, // L&       CYRILLIC CAPITAL LETTER SCHWA WITH DIAERESIS
    Interval{ 0x04DC, 0x04DC }, // L&       CYRILLIC CAPITAL LETTER ZHE WITH DIAERESIS
    Interval{ 0x04DE, 0x04DE }, // L&       CYRILLIC CAPITAL LETTER ZE WITH DIAERESIS
    Interval{ 0x04E0, 0x04E0 }, // L&       CYRILLIC CAPITAL LETTER ABKHASIAN DZE
    Interval{ 0x04E2, 0x04E2 }, // L&       CYRILLIC CAPITAL LETTER I WITH MACRON
    Interval{ 0x04E4, 0x04E4 }, // L&       CYRILLIC CAPITAL LETTER I WITH DIAERESIS
    Interval{ 0x04E6, 0x04E6 }, // L&       CYRILLIC CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x04E8, 0x04E8 }, // L&       CYRILLIC CAPITAL LETTER BARRED O
    Interval{ 0x04EA, 0x04EA }, // L&       CYRILLIC CAPITAL LETTER BARRED O WITH DIAERESIS
    Interval{ 0x04EC, 0x04EC }, // L&       CYRILLIC CAPITAL LETTER E WITH DIAERESIS
    Interval{ 0x04EE, 0x04EE }, // L&       CYRILLIC CAPITAL LETTER U WITH MACRON
    Interval{ 0x04F0, 0x04F0 }, // L&       CYRILLIC CAPITAL LETTER U WITH DIAERESIS
    Interval{ 0x04F2, 0x04F2 }, // L&       CYRILLIC CAPITAL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x04F4, 0x04F4 }, // L&       CYRILLIC CAPITAL LETTER CHE WITH DIAERESIS
    Interval{ 0x04F6, 0x04F6 }, // L&       CYRILLIC CAPITAL LETTER GHE WITH DESCENDER
    Interval{ 0x04F8, 0x04F8 }, // L&       CYRILLIC CAPITAL LETTER YERU WITH DIAERESIS
    Interval{ 0x04FA, 0x04FA }, // L&       CYRILLIC CAPITAL LETTER GHE WITH STROKE AND HOOK
    Interval{ 0x04FC, 0x04FC }, // L&       CYRILLIC CAPITAL LETTER HA WITH HOOK
    Interval{ 0x04FE, 0x04FE }, // L&       CYRILLIC CAPITAL LETTER HA WITH STROKE
    Interval{ 0x0500, 0x0500 }, // L&       CYRILLIC CAPITAL LETTER KOMI DE
    Interval{ 0x0502, 0x0502 }, // L&       CYRILLIC CAPITAL LETTER KOMI DJE
    Interval{ 0x0504, 0x0504 }, // L&       CYRILLIC CAPITAL LETTER KOMI ZJE
    Interval{ 0x0506, 0x0506 }, // L&       CYRILLIC CAPITAL LETTER KOMI DZJE
    Interval{ 0x0508, 0x0508 }, // L&       CYRILLIC CAPITAL LETTER KOMI LJE
    Interval{ 0x050A, 0x050A }, // L&       CYRILLIC CAPITAL LETTER KOMI NJE
    Interval{ 0x050C, 0x050C }, // L&       CYRILLIC CAPITAL LETTER KOMI SJE
    Interval{ 0x050E, 0x050E }, // L&       CYRILLIC CAPITAL LETTER KOMI TJE
    Interval{ 0x0510, 0x0510 }, // L&       CYRILLIC CAPITAL LETTER REVERSED ZE
    Interval{ 0x0512, 0x0512 }, // L&       CYRILLIC CAPITAL LETTER EL WITH HOOK
    Interval{ 0x0514, 0x0514 }, // L&       CYRILLIC CAPITAL LETTER LHA
    Interval{ 0x0516, 0x0516 }, // L&       CYRILLIC CAPITAL LETTER RHA
    Interval{ 0x0518, 0x0518 }, // L&       CYRILLIC CAPITAL LETTER YAE
    Interval{ 0x051A, 0x051A }, // L&       CYRILLIC CAPITAL LETTER QA
    Interval{ 0x051C, 0x051C }, // L&       CYRILLIC CAPITAL LETTER WE
    Interval{ 0x051E, 0x051E }, // L&       CYRILLIC CAPITAL LETTER ALEUT KA
    Interval{ 0x0520, 0x0520 }, // L&       CYRILLIC CAPITAL LETTER EL WITH MIDDLE HOOK
    Interval{ 0x0522, 0x0522 }, // L&       CYRILLIC CAPITAL LETTER EN WITH MIDDLE HOOK
    Interval{ 0x0524, 0x0524 }, // L&       CYRILLIC CAPITAL LETTER PE WITH DESCENDER
    Interval{ 0x0526, 0x0526 }, // L&       CYRILLIC CAPITAL LETTER SHHA WITH DESCENDER
    Interval{ 0x0528, 0x0528 }, // L&       CYRILLIC CAPITAL LETTER EN WITH LEFT HOOK
    Interval{ 0x052A, 0x052A }, // L&       CYRILLIC CAPITAL LETTER DZZHE
    Interval{ 0x052C, 0x052C }, // L&       CYRILLIC CAPITAL LETTER DCHE
    Interval{ 0x052E, 0x052E }, // L&       CYRILLIC CAPITAL LETTER EL WITH DESCENDER
    Interval{ 0x0531, 0x0556 }, // L&  [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Interval{ 0x10A0, 0x10C5 }, // L&  [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Interval{ 0x10C7, 0x10C7 }, // L&       GEORGIAN CAPITAL LETTER YN
    Interval{ 0x10CD, 0x10CD }, // L&       GEORGIAN CAPITAL LETTER AEN
    Interval{ 0x13A0, 0x13F5 }, // L&  [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Interval{ 0x1C89, 0x1C89 }, // L&       CYRILLIC CAPITAL LETTER TJE
    Interval{ 0x1C90, 0x1CBA }, // L&  [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Interval{ 0x1CBD, 0x1CBF }, // L&   [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Interval{ 0x1E00, 0x1E00 }, // L&       LATIN CAPITAL LETTER A WITH RING BELOW
    Interval{ 0x1E02, 0x1E02 }, // L&       LATIN CAPITAL LETTER B WITH DOT ABOVE
    Interval{ 0x1E04, 0x1E04 }, // L&       LATIN CAPITAL LETTER B WITH DOT BELOW
    Interval{ 0x1E06, 0x1E06 }, // L&       LATIN CAPITAL LETTER B WITH LINE BELOW
    Interval{ 0x1E08, 0x1E08 }, // L&       LATIN CAPITAL LETTER C WITH CEDILLA AND ACUTE
    Interval{ 0x1E0A, 0x1E0A }, // L&       LATIN CAPITAL LETTER D WITH DOT ABOVE
    Interval{ 0x1E0C, 0x1E0C }, // L&       LATIN CAPITAL LETTER D WITH DOT BELOW
    Interval{ 0x1E0E, 0x1E0E }, // L&       LATIN CAPITAL LETTER D WITH LINE BELOW
    Interval{ 0x1E10, 0x1E10 }, // L&       LATIN CAPITAL LETTER D WITH CEDILLA
    Interval{ 0x1E12, 0x1E12 }, // L&       LATIN CAPITAL LETTER D WITH CIRCUMFLEX BELOW
    Interval{ 0x1E14, 0x1E14 }, // L&       LATIN CAPITAL LETTER E WITH MACRON AND GRAVE
    Interval{ 0x1E16, 0x1E16 }, // L&       LATIN CAPITAL LETTER E WITH MACRON AND ACUTE
    Interval{ 0x1E18, 0x1E18 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX BELOW
    Interval{ 0x1E1A, 0x1E1A }, // L&       LATIN CAPITAL LETTER E WITH TILDE BELOW
    Interval{ 0x1E1C, 0x1E1C }, // L&       LATIN CAPITAL LETTER E WITH CEDILLA AND BREVE
    Interval{ 0x1E1E, 0x1E1E }, // L&       LATIN CAPITAL LETTER F WITH DOT ABOVE
    Interval{ 0x1E20, 0x1E20 }, // L&       LATIN CAPITAL LETTER G WITH MACRON
    Interval{ 0x1E22, 0x1E22 }, // L&       LATIN CAPITAL LETTER H WITH DOT ABOVE
    Interval{ 0x1E24, 0x1E24 }, // L&       LATIN CAPITAL LETTER H WITH DOT BELOW
    Interval{ 0x1E26, 0x1E26 }, // L&       LATIN CAPITAL LETTER H WITH DIAERESIS
    Interval{ 0x1E28, 0x1E28 }, // L&       LATIN CAPITAL LETTER H WITH CEDILLA
    Interval{ 0x1E2A, 0x1E2A }, // L&       LATIN CAPITAL LETTER H WITH BREVE BELOW
    Interval{ 0x1E2C, 0x1E2C }, // L&       LATIN CAPITAL LETTER I WITH TILDE BELOW
    Interval{ 0x1E2E, 0x1E2E }, // L&       LATIN CAPITAL LETTER I WITH DIAERESIS AND ACUTE
    Interval{ 0x1E30, 0x1E30 }, // L&       LATIN CAPITAL LETTER K WITH ACUTE
    Interval{ 0x1E32, 0x1E32 }, // L&       LATIN CAPITAL LETTER K WITH DOT BELOW
    Interval{ 0x1E34, 0x1E34 }, // L&       LATIN CAPITAL LETTER K WITH LINE BELOW
    Interval{ 0x1E36, 0x1E36 }, // L&       LATIN CAPITAL LETTER L WITH DOT BELOW
    Interval{ 0x1E38, 0x1E38 }, // L&       LATIN CAPITAL LETTER L WITH DOT BELOW AND MACRON
    Interval{ 0x1E3A, 0x1E3A }, // L&       LATIN CAPITAL LETTER L WITH LINE BELOW
    Interval{ 0x1E3C, 0x1E3C }, // L&       LATIN CAPITAL LETTER L WITH CIRCUMFLEX BELOW
    Interval{ 0x1E3E, 0x1E3E }, // L&       LATIN CAPITAL LETTER M WITH ACUTE
    Interval{ 0x1E40, 0x1E40 }, // L&       LATIN CAPITAL LETTER M WITH DOT ABOVE
    Interval{ 0x1E42, 0x1E42 }, // L&       LATIN CAPITAL LETTER M WITH DOT BELOW
    Interval{ 0x1E44, 0x1E44 }, // L&       LATIN CAPITAL LETTER N WITH DOT ABOVE
    Interval{ 0x1E46, 0x1E46 }, // L&       LATIN CAPITAL LETTER N WITH DOT BELOW
    Interval{ 0x1E48, 0x1E48 }, // L&       LATIN CAPITAL LETTER N WITH LINE BELOW
    Interval{ 0x1E4A, 0x1E4A }, // L&       LATIN CAPITAL LETTER N WITH CIRCUMFLEX BELOW
    Interval{ 0x1E4C, 0x1E4C }, // L&       LATIN CAPITAL LETTER O WITH TILDE AND ACUTE
    Interval{ 0x1E4E, 0x1E4E }, // L&       LATIN CAPITAL LETTER O WITH TILDE AND DIAERESIS
    Interval{ 0x1E50, 0x1E50 }, // L&       LATIN CAPITAL LETTER O WITH MACRON AND GRAVE
    Interval{ 0x1E52, 0x1E52 }, // L&       LATIN CAPITAL LETTER O WITH MACRON AND ACUTE
    Interval{ 0x1E54, 0x1E54 }, // L&       LATIN CAPITAL LETTER P WITH ACUTE
    Interval{ 0x1E56, 0x1E56 }, // L&       LATIN CAPITAL LETTER P WITH DOT ABOVE
    Interval{ 0x1E58, 0x1E58 }, // L&       LATIN CAPITAL LETTER R WITH DOT ABOVE
    Interval{ 0x1E5A, 0x1E5A }, // L&       LATIN CAPITAL LETTER R WITH DOT BELOW
    Interval{ 0x1E5C, 0x1E5C }, // L&       LATIN CAPITAL LETTER R WITH DOT BELOW AND MACRON
    Interval{ 0x1E5E, 0x1E5E }, // L&       LATIN CAPITAL LETTER R WITH LINE BELOW
    Interval{ 0x1E60, 0x1E60 }, // L&       LATIN CAPITAL LETTER S WITH DOT ABOVE
    Interval{ 0x1E62, 0x1E62 }, // L&       LATIN CAPITAL LETTER S WITH DOT BELOW
    Interval{ 0x1E64, 0x1E64 }, // L&       LATIN CAPITAL LETTER S WITH ACUTE AND DOT ABOVE
    Interval{ 0x1E66, 0x1E66 }, // L&       LATIN CAPITAL LETTER S WITH CARON AND DOT ABOVE
    Interval{ 0x1E68, 0x1E68 }, // L&       LATIN CAPITAL LETTER S WITH DOT BELOW AND DOT ABOVE
    Interval{ 0x1E6A, 0x1E6A }, // L&       LATIN CAPITAL LETTER T WITH DOT ABOVE
    Interval{ 0x1E6C, 0x1E6C }, // L&       LATIN CAPITAL LETTER T WITH DOT BELOW
    Interval{ 0x1E6E, 0x1E6E }, // L&       LATIN CAPITAL LETTER T WITH LINE BELOW
    Interval{ 0x1E70, 0x1E70 }, // L&       LATIN CAPITAL LETTER T WITH CIRCUMFLEX BELOW
    Interval{ 0x1E72, 0x1E72 }, // L&       LATIN CAPITAL LETTER U WITH DIAERESIS BELOW
    Interval{ 0x1E74, 0x1E74 }, // L&       LATIN CAPITAL LETTER U WITH TILDE BELOW
    Interval{ 0x1E76, 0x1E76 }, // L&       LATIN CAPITAL LETTER U WITH CIRCUMFLEX BELOW
    Interval{ 0x1E78, 0x1E78 }, // L&       LATIN CAPITAL LETTER U WITH TILDE AND ACUTE
    Interval{ 0x1E7A, 0x1E7A }, // L&       LATIN CAPITAL LETTER U WITH MACRON AND DIAERESIS
    Interval{ 0x1E7C, 0x1E7C }, // L&       LATIN CAPITAL LETTER V WITH TILDE
    Interval{ 0x1E7E, 0x1E7E }, // L&       LATIN CAPITAL LETTER V WITH DOT BELOW
    Interval{ 0x1E80, 0x1E80 }, // L&       LATIN CAPITAL LETTER W WITH GRAVE
    Interval{ 0x1E82, 0x1E82 }, // L&       LATIN CAPITAL LETTER W WITH ACUTE
    Interval{ 0x1E84, 0x1E84 }, // L&       LATIN CAPITAL LETTER W WITH DIAERESIS
    Interval{ 0x1E86, 0x1E86 }, // L&       LATIN CAPITAL LETTER W WITH DOT ABOVE
    Interval{ 0x1E88, 0x1E88 }, // L&       LATIN CAPITAL LETTER W WITH DOT BELOW
    Interval{ 0x1E8A, 0x1E8A }, // L&       LATIN CAPITAL LETTER X WITH DOT ABOVE
    Interval{ 0x1E8C, 0x1E8C }, // L&       LATIN CAPITAL LETTER X WITH DIAERESIS
    Interval{ 0x1E8E, 0x1E8E }, // L&       LATIN CAPITAL LETTER Y WITH DOT ABOVE
    Interval{ 0x1E90, 0x1E90 }, // L&       LATIN CAPITAL LETTER Z WITH CIRCUMFLEX
    Interval{ 0x1E92, 0x1E92 }, // L&       LATIN CAPITAL LETTER Z WITH DOT BELOW
    Interval{ 0x1E94, 0x1E94 }, // L&       LATIN CAPITAL LETTER Z WITH LINE BELOW
    Interval{ 0x1E9E, 0x1E9E }, // L&       LATIN CAPITAL LETTER SHARP S
    Interval{ 0x1EA0, 0x1EA0 }, // L&       LATIN CAPITAL LETTER A WITH DOT BELOW
    Interval{ 0x1EA2, 0x1EA2 }, // L&       LATIN CAPITAL LETTER A WITH HOOK ABOVE
    Interval{ 0x1EA4, 0x1EA4 }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EA6, 0x1EA6 }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EA8, 0x1EA8 }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EAA, 0x1EAA }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EAC, 0x1EAC }, // L&       LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EAE, 0x1EAE }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND ACUTE
    Interval{ 0x1EB0, 0x1EB0 }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND GRAVE
    Interval{ 0x1EB2, 0x1EB2 }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND HOOK ABOVE
    Interval{ 0x1EB4, 0x1EB4 }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND TILDE
    Interval{ 0x1EB6, 0x1EB6 }, // L&       LATIN CAPITAL LETTER A WITH BREVE AND DOT BELOW
    Interval{ 0x1EB8, 0x1EB8 }, // L&       LATIN CAPITAL LETTER E WITH DOT BELOW
    Interval{ 0x1EBA, 0x1EBA }, // L&       LATIN CAPITAL LETTER E WITH HOOK ABOVE
    Interval{ 0x1EBC, 0x1EBC }, // L&       LATIN CAPITAL LETTER E WITH TILDE
    Interval{ 0x1EBE, 0x1EBE }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EC0, 0x1EC0 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EC2, 0x1EC2 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EC4, 0x1EC4 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EC6, 0x1EC6 }, // L&       LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EC8, 0x1EC8 }, // L&       LATIN CAPITAL LETTER I WITH HOOK ABOVE
    Interval{ 0x1ECA, 0x1ECA }, // L&       LATIN CAPITAL LETTER I WITH DOT BELOW
    Interval{ 0x1ECC, 0x1ECC }, // L&       LATIN CAPITAL LETTER O WITH DOT BELOW
    Interval{ 0x1ECE, 0x1ECE }, // L&       LATIN CAPITAL LETTER O WITH HOOK ABOVE
    Interval{ 0x1ED0, 0x1ED0 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1ED2, 0x1ED2 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1ED4, 0x1ED4 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1ED6, 0x1ED6 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1ED8, 0x1ED8 }, // L&       LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EDA, 0x1EDA }, // L&       LATIN CAPITAL LETTER O WITH HORN AND ACUTE
    Interval{ 0x1EDC, 0x1EDC }, // L&       LATIN CAPITAL LETTER O WITH HORN AND GRAVE
    Interval{ 0x1EDE, 0x1EDE }, // L&       LATIN CAPITAL LETTER O WITH HORN AND HOOK ABOVE
    Interval{ 0x1EE0, 0x1EE0 }, // L&       LATIN CAPITAL LETTER O WITH HORN AND TILDE
    Interval{ 0x1EE2, 0x1EE2 }, // L&       LATIN CAPITAL LETTER O WITH HORN AND DOT BELOW
    Interval{ 0x1EE4, 0x1EE4 }, // L&       LATIN CAPITAL LETTER U WITH DOT BELOW
    Interval{ 0x1EE6, 0x1EE6 }, // L&       LATIN CAPITAL LETTER U WITH HOOK ABOVE
    Interval{ 0x1EE8, 0x1EE8 }, // L&       LATIN CAPITAL LETTER U WITH HORN AND ACUTE
    Interval{ 0x1EEA, 0x1EEA }, // L&       LATIN CAPITAL LETTER U WITH HORN AND GRAVE
    Interval{ 0x1EEC, 0x1EEC }, // L&       LATIN CAPITAL LETTER U WITH HORN AND HOOK ABOVE
    Interval{ 0x1EEE, 0x1EEE }, // L&       LATIN CAPITAL LETTER U WITH HORN AND TILDE
    Interval{ 0x1EF0, 0x1EF0 }, // L&       LATIN CAPITAL LETTER U WITH HORN AND DOT BELOW
    Interval{ 0x1EF2, 0x1EF2 }, // L&       LATIN CAPITAL LETTER Y WITH GRAVE
    Interval{ 0x1EF4, 0x1EF4 }, // L&       LATIN CAPITAL LETTER Y WITH DOT BELOW
    Interval{ 0x1EF6, 0x1EF6 }, // L&       LATIN CAPITAL LETTER Y WITH HOOK ABOVE
    Interval{ 0x1EF8, 0x1EF8 }, // L&       LATIN CAPITAL LETTER Y WITH TILDE
    Interval{ 0x1EFA, 0x1EFA }, // L&       LATIN CAPITAL LETTER MIDDLE-WELSH LL
    Interval{ 0x1EFC, 0x1EFC }, // L&       LATIN CAPITAL LETTER MIDDLE-WELSH V
    Interval{ 0x1EFE, 0x1EFE }, // L&       LATIN CAPITAL LETTER Y WITH LOOP
    Interval{ 0x1F08, 0x1F0F }, // L&   [8] GREEK CAPITAL LETTER ALPHA WITH PSILI..GREEK CAPITAL LETTER ALPHA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F18, 0x1F1D }, // L&   [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F28, 0x1F2F }, // L&   [8] GREEK CAPITAL LETTER ETA WITH PSILI..GREEK CAPITAL LETTER ETA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F38, 0x1F3F }, // L&   [8] GREEK CAPITAL LETTER IOTA WITH PSILI..GREEK CAPITAL LETTER IOTA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F48, 0x1F4D }, // L&   [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F59, 0x1F59 }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA
    Interval{ 0x1F5B, 0x1F5B }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Interval{ 0x1F5D, 0x1F5D }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Interval{ 0x1F5F, 0x1F5F }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F68, 0x1F6F }, // L&   [8] GREEK CAPITAL LETTER OMEGA WITH PSILI..GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI
    Interval{ 0x1FB8, 0x1FBB }, // L&   [4] GREEK CAPITAL LETTER ALPHA WITH VRACHY..GREEK CAPITAL LETTER ALPHA WITH OXIA
    Interval{ 0x1FC8, 0x1FCB }, // L&   [4] GREEK CAPITAL LETTER EPSILON WITH VARIA..GREEK CAPITAL LETTER ETA WITH OXIA
    Interval{ 0x1FD8, 0x1FDB }, // L&   [4] GREEK CAPITAL LETTER IOTA WITH VRACHY..GREEK CAPITAL LETTER IOTA WITH OXIA
    Interval{ 0x1FE8, 0x1FEC }, // L&   [5] GREEK CAPITAL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Interval{ 0x1FF8, 0x1FFB }, // L&   [4] GREEK CAPITAL LETTER OMICRON WITH VARIA..GREEK CAPITAL LETTER OMEGA WITH OXIA
    Interval{ 0x2102, 0x2102 }, // L&       DOUBLE-STRUCK CAPITAL C
    Interval{ 0x2107, 0x2107 }, // L&       EULER CONSTANT
    Interval{ 0x210B, 0x210D }, // L&   [3] SCRIPT CAPITAL H..DOUBLE-STRUCK CAPITAL H
    Interval{ 0x2110, 0x2112 }, // L&   [3] SCRIPT CAPITAL I..SCRIPT CAPITAL L
    Interval{ 0x2115, 0x2115 }, // L&       DOUBLE-STRUCK CAPITAL N
    Interval{ 0x2119, 0x211D }, // L&   [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    Interval{ 0x2124, 0x2124 }, // L&       DOUBLE-STRUCK CAPITAL Z
    Interval{ 0x2126, 0x2126 }, // L&       OHM SIGN
    Interval{ 0x2128, 0x2128 }, // L&       BLACK-LETTER CAPITAL Z
    Interval{ 0x212A, 0x212D }, // L&   [4] KELVIN SIGN..BLACK-LETTER CAPITAL C
    Interval{ 0x2130, 0x2133 }, // L&   [4] SCRIPT CAPITAL E..SCRIPT CAPITAL M
    Interval{ 0x213E, 0x213F }, // L&   [2] DOUBLE-STRUCK CAPITAL GAMMA..DOUBLE-STRUCK CAPITAL PI
    Interval{ 0x2145, 0x2145 }, // L&       DOUBLE-STRUCK ITALIC CAPITAL D
    Interval{ 0x2160, 0x216F }, // Nl  [16] ROMAN NUMERAL ONE..ROMAN NUMERAL ONE THOUSAND
    Interval{ 0x2183, 0x2183 }, // L&       ROMAN NUMERAL REVERSED ONE HUNDRED
    Interval{ 0x24B6, 0x24CF }, // So  [26] CIRCLED LATIN CAPITAL LETTER A..CIRCLED LATIN CAPITAL LETTER Z
    Interval{ 0x2C00, 0x2C2F }, // L&  [48] GLAGOLITIC CAPITAL LETTER AZU..GLAGOLITIC CAPITAL LETTER CAUDATE CHRIVI
    Interval{ 0x2C60, 0x2C60 }, // L&       LATIN CAPITAL LETTER L WITH DOUBLE BAR
    Interval{ 0x2C62, 0x2C64 }, // L&   [3] LATIN CAPITAL LETTER L WITH MIDDLE TILDE..LATIN CAPITAL LETTER R WITH TAIL
    Interval{ 0x2C67, 0x2C67 }, // L&       LATIN CAPITAL LETTER H WITH DESCENDER
    Interval{ 0x2C69, 0x2C69 }, // L&       LATIN CAPITAL LETTER K WITH DESCENDER
    Interval{ 0x2C6B, 0x2C6B }, // L&       LATIN CAPITAL LETTER Z WITH DESCENDER
    Interval{ 0x2C6D, 0x2C70 }, // L&   [4] LATIN CAPITAL LETTER ALPHA..LATIN CAPITAL LETTER TURNED ALPHA
    Interval{ 0x2C72, 0x2C72 }, // L&       LATIN CAPITAL LETTER W WITH HOOK
    Interval{ 0x2C75, 0x2C75 }, // L&       LATIN CAPITAL LETTER HALF H
    Interval{ 0x2C7E, 0x2C80 }, // L&   [3] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC CAPITAL LETTER ALFA
    Interval{ 0x2C82, 0x2C82 }, // L&       COPTIC CAPITAL LETTER VIDA
    Interval{ 0x2C84, 0x2C84 }, // L&       COPTIC CAPITAL LETTER GAMMA
    Interval{ 0x2C86, 0x2C86 }, // L&       COPTIC CAPITAL LETTER DALDA
    Interval{ 0x2C88, 0x2C88 }, // L&       COPTIC CAPITAL LETTER EIE
    Interval{ 0x2C8A, 0x2C8A }, // L&       COPTIC CAPITAL LETTER SOU
    Interval{ 0x2C8C, 0x2C8C }, // L&       COPTIC CAPITAL LETTER ZATA
    Interval{ 0x2C8E, 0x2C8E }, // L&       COPTIC CAPITAL LETTER HATE
    Interval{ 0x2C90, 0x2C90 }, // L&       COPTIC CAPITAL LETTER THETHE
    Interval{ 0x2C92, 0x2C92 }, // L&       COPTIC CAPITAL LETTER IAUDA
    Interval{ 0x2C94, 0x2C94 }, // L&       COPTIC CAPITAL LETTER KAPA
    Interval{ 0x2C96, 0x2C96 }, // L&       COPTIC CAPITAL LETTER LAULA
    Interval{ 0x2C98, 0x2C98 }, // L&       COPTIC CAPITAL LETTER MI
    Interval{ 0x2C9A, 0x2C9A }, // L&       COPTIC CAPITAL LETTER NI
    Interval{ 0x2C9C, 0x2C9C }, // L&       COPTIC CAPITAL LETTER KSI
    Interval{ 0x2C9E, 0x2C9E }, // L&       COPTIC CAPITAL LETTER O
    Interval{ 0x2CA0, 0x2CA0 }, // L&       COPTIC CAPITAL LETTER PI
    Interval{ 0x2CA2, 0x2CA2 }, // L&       COPTIC CAPITAL LETTER RO
    Interval{ 0x2CA4, 0x2CA4 }, // L&       COPTIC CAPITAL LETTER SIMA
    Interval{ 0x2CA6, 0x2CA6 }, // L&       COPTIC CAPITAL LETTER TAU
    Interval{ 0x2CA8, 0x2CA8 }, // L&       COPTIC CAPITAL LETTER UA
    Interval{ 0x2CAA, 0x2CAA }, // L&       COPTIC CAPITAL LETTER FI
    Interval{ 0x2CAC, 0x2CAC }, // L&       COPTIC CAPITAL LETTER KHI
    Interval{ 0x2CAE, 0x2CAE }, // L&       COPTIC CAPITAL LETTER PSI
    Interval{ 0x2CB0, 0x2CB0 }, // L&       COPTIC CAPITAL LETTER OOU
    Interval{ 0x2CB2, 0x2CB2 }, // L&       COPTIC CAPITAL LETTER DIALECT-P ALEF
    Interval{ 0x2CB4, 0x2CB4 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC AIN
    Interval{ 0x2CB6, 0x2CB6 }, // L&       COPTIC CAPITAL LETTER CRYPTOGRAMMIC EIE
    Interval{ 0x2CB8, 0x2CB8 }, // L&       COPTIC CAPITAL LETTER DIALECT-P KAPA
    Interval{ 0x2CBA, 0x2CBA }, // L&       COPTIC CAPITAL LETTER DIALECT-P NI
    Interval{ 0x2CBC, 0x2CBC }, // L&       COPTIC CAPITAL LETTER CRYPTOGRAMMIC NI
    Interval{ 0x2CBE, 0x2CBE }, // L&       COPTIC CAPITAL LETTER OLD COPTIC OOU
    Interval{ 0x2CC0, 0x2CC0 }, // L&       COPTIC CAPITAL LETTER SAMPI
    Interval{ 0x2CC2, 0x2CC2 }, // L&       COPTIC CAPITAL LETTER CROSSED SHEI
    Interval{ 0x2CC4, 0x2CC4 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC SHEI
    Interval{ 0x2CC6, 0x2CC6 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC ESH
    Interval{ 0x2CC8, 0x2CC8 }, // L&       COPTIC CAPITAL LETTER AKHMIMIC KHEI
    Interval{ 0x2CCA, 0x2CCA }, // L&       COPTIC CAPITAL LETTER DIALECT-P HORI
    Interval{ 0x2CCC, 0x2CCC }, // L&       COPTIC CAPITAL LETTER OLD COPTIC HORI
    Interval{ 0x2CCE, 0x2CCE }, // L&       COPTIC CAPITAL LETTER OLD COPTIC HA
    Interval{ 0x2CD0, 0x2CD0 }, // L&       COPTIC CAPITAL LETTER L-SHAPED HA
    Interval{ 0x2CD2, 0x2CD2 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC HEI
    Interval{ 0x2CD4, 0x2CD4 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC HAT
    Interval{ 0x2CD6, 0x2CD6 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC GANGIA
    Interval{ 0x2CD8, 0x2CD8 }, // L&       COPTIC CAPITAL LETTER OLD COPTIC DJA
    Interval{ 0x2CDA, 0x2CDA }, // L&       COPTIC CAPITAL LETTER OLD COPTIC SHIMA
    Interval{ 0x2CDC, 0x2CDC }, // L&       COPTIC CAPITAL LETTER OLD NUBIAN SHIMA
    Interval{ 0x2CDE, 0x2CDE }, // L&       COPTIC CAPITAL LETTER OLD NUBIAN NGI
    Interval{ 0x2CE0, 0x2CE0 }, // L&       COPTIC CAPITAL LETTER OLD NUBIAN NYI
    Interval{ 0x2CE2, 0x2CE2 }, // L&       COPTIC CAPITAL LETTER OLD NUBIAN WAU
    Interval{ 0x2CEB, 0x2CEB }, // L&       COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI
    Interval{ 0x2CED, 0x2CED }, // L&       COPTIC CAPITAL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF2, 0x2CF2 }, // L&       COPTIC CAPITAL LETTER BOHAIRIC KHEI
    Interval{ 0xA640, 0xA640 }, // L&       CYRILLIC CAPITAL LETTER ZEMLYA
    Interval{ 0xA642, 0xA642 }, // L&       CYRILLIC CAPITAL LETTER DZELO
    Interval{ 0xA644, 0xA644 }, // L&       CYRILLIC CAPITAL LETTER REVERSED DZE
    Interval{ 0xA646, 0xA646 }, // L&       CYRILLIC CAPITAL LETTER IOTA
    Interval{ 0xA648, 0xA648 }, // L&       CYRILLIC CAPITAL LETTER DJERV
    Interval{ 0xA64A, 0xA64A }, // L&       CYRILLIC CAPITAL LETTER MONOGRAPH UK
    Interval{ 0xA64C, 0xA64C }, // L&       CYRILLIC CAPITAL LETTER BROAD OMEGA
    Interval{ 0xA64E, 0xA64E }, // L&       CYRILLIC CAPITAL LETTER NEUTRAL YER
    Interval{ 0xA650, 0xA650 }, // L&       CYRILLIC CAPITAL LETTER YERU WITH BACK YER
    Interval{ 0xA652, 0xA652 }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED YAT
    Interval{ 0xA654, 0xA654 }, // L&       CYRILLIC CAPITAL LETTER REVERSED YU
    Interval{ 0xA656, 0xA656 }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED A
    Interval{ 0xA658, 0xA658 }, // L&       CYRILLIC CAPITAL LETTER CLOSED LITTLE YUS
    Interval{ 0xA65A, 0xA65A }, // L&       CYRILLIC CAPITAL LETTER BLENDED YUS
    Interval{ 0xA65C, 0xA65C }, // L&       CYRILLIC CAPITAL LETTER IOTIFIED CLOSED LITTLE YUS
    Interval{ 0xA65E, 0xA65E }, // L&       CYRILLIC CAPITAL LETTER YN
    Interval{ 0xA660, 0xA660 }, // L&       CYRILLIC CAPITAL LETTER REVERSED TSE
    Interval{ 0xA662, 0xA662 }, // L&       CYRILLIC CAPITAL LETTER SOFT DE
    Interval{ 0xA664, 0xA664 }, // L&       CYRILLIC CAPITAL LETTER SOFT EL
    Interval{ 0xA666, 0xA666 }, // L&       CYRILLIC CAPITAL LETTER SOFT EM
    Interval{ 0xA668, 0xA668 }, // L&       CYRILLIC CAPITAL LETTER MONOCULAR O
    Interval{ 0xA66A, 0xA66A }, // L&       CYRILLIC CAPITAL LETTER BINOCULAR O
    Interval{ 0xA66C, 0xA66C }, // L&       CYRILLIC CAPITAL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA680, 0xA680 }, // L&       CYRILLIC CAPITAL LETTER DWE
    Interval{ 0xA682, 0xA682 }, // L&       CYRILLIC CAPITAL LETTER DZWE
    Interval{ 0xA684, 0xA684 }, // L&       CYRILLIC CAPITAL LETTER ZHWE
    Interval{ 0xA686, 0xA686 }, // L&       CYRILLIC CAPITAL LETTER CCHE
    Interval{ 0xA688, 0xA688 }, // L&       CYRILLIC CAPITAL LETTER DZZE
    Interval{ 0xA68A, 0xA68A }, // L&       CYRILLIC CAPITAL LETTER TE WITH MIDDLE HOOK
    Interval{ 0xA68C, 0xA68C }, // L&       CYRILLIC CAPITAL LETTER TWE
    Interval{ 0xA68E, 0xA68E }, // L&       CYRILLIC CAPITAL LETTER TSWE
    Interval{ 0xA690, 0xA690 }, // L&       CYRILLIC CAPITAL LETTER TSSE
    Interval{ 0xA692, 0xA692 }, // L&       CYRILLIC CAPITAL LETTER TCHE
    Interval{ 0xA694, 0xA694 }, // L&       CYRILLIC CAPITAL LETTER HWE
    Interval{ 0xA696, 0xA696 }, // L&       CYRILLIC CAPITAL LETTER SHWE
    Interval{ 0xA698, 0xA698 }, // L&       CYRILLIC CAPITAL LETTER DOUBLE O
    Interval{ 0xA69A, 0xA69A }, // L&       CYRILLIC CAPITAL LETTER CROSSED O
    Interval{ 0xA722, 0xA722 }, // L&       LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF
    Interval{ 0xA724, 0xA724 }, // L&       LATIN CAPITAL LETTER EGYPTOLOGICAL AIN
    Interval{ 0xA726, 0xA726 }, // L&       LATIN CAPITAL LETTER HENG
    Interval{ 0xA728, 0xA728 }, // L&       LATIN CAPITAL LETTER TZ
    Interval{ 0xA72A, 0xA72A }, // L&       LATIN CAPITAL LETTER TRESILLO
    Interval{ 0xA72C, 0xA72C }, // L&       LATIN CAPITAL LETTER CUATRILLO
    Interval{ 0xA72E, 0xA72E }, // L&       LATIN CAPITAL LETTER CUATRILLO WITH COMMA
    Interval{ 0xA732, 0xA732 }, // L&       LATIN CAPITAL LETTER AA
    Interval{ 0xA734, 0xA734 }, // L&       LATIN CAPITAL LETTER AO
    Interval{ 0xA736, 0xA736 }, // L&       LATIN CAPITAL LETTER AU
    Interval{ 0xA738, 0xA738 }, // L&       LATIN CAPITAL LETTER AV
    Interval{ 0xA73A, 0xA73A }, // L&       LATIN CAPITAL LETTER AV WITH HORIZONTAL BAR
    Interval{ 0xA73C, 0xA73C }, // L&       LATIN CAPITAL LETTER AY
    Interval{ 0xA73E, 0xA73E }, // L&       LATIN CAPITAL LETTER REVERSED C WITH DOT
    Interval{ 0xA740, 0xA740 }, // L&       LATIN CAPITAL LETTER K WITH STROKE
    Interval{ 0xA742, 0xA742 }, // L&       LATIN CAPITAL LETTER K WITH DIAGONAL STROKE
    Interval{ 0xA744, 0xA744 }, // L&       LATIN CAPITAL LETTER K WITH STROKE AND DIAGONAL STROKE
    Interval{ 0xA746, 0xA746 }, // L&       LATIN CAPITAL LETTER BROKEN L
    Interval{ 0xA748, 0xA748 }, // L&       LATIN CAPITAL LETTER L WITH HIGH STROKE
    Interval{ 0xA74A, 0xA74A }, // L&       LATIN CAPITAL LETTER O WITH LONG STROKE OVERLAY
    Interval{ 0xA74C, 0xA74C }, // L&       LATIN CAPITAL LETTER O WITH LOOP
    Interval{ 0xA74E, 0xA74E }, // L&       LATIN CAPITAL LETTER OO
    Interval{ 0xA750, 0xA750 }, // L&       LATIN CAPITAL LETTER P WITH STROKE THROUGH DESCENDER
    Interval{ 0xA752, 0xA752 }, // L&       LATIN CAPITAL LETTER P WITH FLOURISH
    Interval{ 0xA754, 0xA754 }, // L&       LATIN CAPITAL LETTER P WITH SQUIRREL TAIL
    Interval{ 0xA756, 0xA756 }, // L&       LATIN CAPITAL LETTER Q WITH STROKE THROUGH DESCENDER
    Interval{ 0xA758, 0xA758 }, // L&       LATIN CAPITAL LETTER Q WITH DIAGONAL STROKE
    Interval{ 0xA75A, 0xA75A }, // L&       LATIN CAPITAL LETTER R ROTUNDA
    Interval{ 0xA75C, 0xA75C }, // L&       LATIN CAPITAL LETTER RUM ROTUNDA
    Interval{ 0xA75E, 0xA75E }, // L&       LATIN CAPITAL LETTER V WITH DIAGONAL STROKE
    Interval{ 0xA760, 0xA760 }, // L&       LATIN CAPITAL LETTER VY
    Interval{ 0xA762, 0xA762 }, // L&       LATIN CAPITAL LETTER VISIGOTHIC Z
    Interval{ 0xA764, 0xA764 }, // L&       LATIN CAPITAL LETTER THORN WITH STROKE
    Interval{ 0xA766, 0xA766 }, // L&       LATIN CAPITAL LETTER THORN WITH STROKE THROUGH DESCENDER
    Interval{ 0xA768, 0xA768 }, // L&       LATIN CAPITAL LETTER VEND
    Interval{ 0xA76A, 0xA76A }, // L&       LATIN CAPITAL LETTER ET
    Interval{ 0xA76C, 0xA76C }, // L&       LATIN CAPITAL LETTER IS
    Interval{ 0xA76E, 0xA76E }, // L&       LATIN CAPITAL LETTER CON
    Interval{ 0xA779, 0xA779 }, // L&       LATIN CAPITAL LETTER INSULAR D
    Interval{ 0xA77B, 0xA77B }, // L&       LATIN CAPITAL LETTER INSULAR F
    Interval{ 0xA77D, 0xA77E }, // L&   [2] LATIN CAPITAL LETTER INSULAR G..LATIN CAPITAL LETTER TURNED INSULAR G
    Interval{ 0xA780, 0xA780 }, // L&       LATIN CAPITAL LETTER TURNED L
    Interval{ 0xA782, 0xA782 }, // L&       LATIN CAPITAL LETTER INSULAR R
    Interval{ 0xA784, 0xA784 }, // L&       LATIN CAPITAL LETTER INSULAR S
    Interval{ 0xA786, 0xA786 }, // L&       LATIN CAPITAL LETTER INSULAR T
    Interval{ 0xA78B, 0xA78B }, // L&       LATIN CAPITAL LETTER SALTILLO
    Interval{ 0xA78D, 0xA78D }, // L&       LATIN CAPITAL LETTER TURNED H
    Interval{ 0xA790, 0xA790 }, // L&       LATIN CAPITAL LETTER N WITH DESCENDER
    Interval{ 0xA792, 0xA792 }, // L&       LATIN CAPITAL LETTER C WITH BAR
    Interval{ 0xA796, 0xA796 }, // L&       LATIN CAPITAL LETTER B WITH FLOURISH
    Interval{ 0xA798, 0xA798 }, // L&       LATIN CAPITAL LETTER F WITH STROKE
    Interval{ 0xA79A, 0xA79A }, // L&       LATIN CAPITAL LETTER VOLAPUK AE
    Interval{ 0xA79C, 0xA79C }, // L&       LATIN CAPITAL LETTER VOLAPUK OE
    Interval{ 0xA79E, 0xA79E }, // L&       LATIN CAPITAL LETTER VOLAPUK UE
    Interval{ 0xA7A0, 0xA7A0 }, // L&       LATIN CAPITAL LETTER G WITH OBLIQUE STROKE
    Interval{ 0xA7A2, 0xA7A2 }, // L&       LATIN CAPITAL LETTER K WITH OBLIQUE STROKE
    Interval{ 0xA7A4, 0xA7A4 }, // L&       LATIN CAPITAL LETTER N WITH OBLIQUE STROKE
    Interval{ 0xA7A6, 0xA7A6 }, // L&       LATIN CAPITAL LETTER R WITH OBLIQUE STROKE
    Interval{ 0xA7A8, 0xA7A8 }, // L&       LATIN CAPITAL LETTER S WITH OBLIQUE STROKE
    Interval{ 0xA7AA, 0xA7AE }, // L&   [5] LATIN CAPITAL LETTER H WITH HOOK..LATIN CAPITAL LETTER SMALL CAPITAL I
    Interval{ 0xA7B0, 0xA7B4 }, // L&   [5] LATIN CAPITAL LETTER TURNED K..LATIN CAPITAL LETTER BETA
    Interval{ 0xA7B6, 0xA7B6 }, // L&       LATIN CAPITAL LETTER OMEGA
    Interval{ 0xA7B8, 0xA7B8 }, // L&       LATIN CAPITAL LETTER U WITH STROKE
    Interval{ 0xA7BA, 0xA7BA }, // L&       LATIN CAPITAL LETTER GLOTTAL A
    Interval{ 0xA7BC, 0xA7BC }, // L&       LATIN CAPITAL LETTER GLOTTAL I
    Interval{ 0xA7BE, 0xA7BE }, // L&       LATIN CAPITAL LETTER GLOTTAL U
    Interval{ 0xA7C0, 0xA7C0 }, // L&       LATIN CAPITAL LETTER OLD POLISH O
    Interval{ 0xA7C2, 0xA7C2 }, // L&       LATIN CAPITAL LETTER ANGLICANA W
    Interval{ 0xA7C4, 0xA7C7 }, // L&   [4] LATIN CAPITAL LETTER C WITH PALATAL HOOK..LATIN CAPITAL LETTER D WITH SHORT STROKE OVERLAY
    Interval{ 0xA7C9, 0xA7C9 }, // L&       LATIN CAPITAL LETTER S WITH SHORT STROKE OVERLAY
    Interval{ 0xA7CB, 0xA7CC }, // L&   [2] LATIN CAPITAL LETTER RAMS HORN..LATIN CAPITAL LETTER S WITH DIAGONAL STROKE
    Interval{ 0xA7CE, 0xA7CE }, // L&       LATIN CAPITAL LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0xA7D0, 0xA7D0 }, // L&       LATIN CAPITAL LETTER CLOSED INSULAR G
    Interval{ 0xA7D2, 0xA7D2 }, // L&       LATIN CAPITAL LETTER DOUBLE THORN
    Interval{ 0xA7D4, 0xA7D4 }, // L&       LATIN CAPITAL LETTER DOUBLE WYNN
    Interval{ 0xA7D6, 0xA7D6 }, // L&       LATIN CAPITAL LETTER MIDDLE SCOTS S
    Interval{ 0xA7D8, 0xA7D8 }, // L&       LATIN CAPITAL LETTER SIGMOID S
    Interval{ 0xA7DA, 0xA7DA }, // L&       LATIN CAPITAL LETTER LAMBDA
    Interval{ 0xA7DC, 0xA7DC }, // L&       LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Interval{ 0xA7F5, 0xA7F5 }, // L&       LATIN CAPITAL LETTER REVERSED HALF H
    Interval{ 0xFF21, 0xFF3A }, // L&  [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Interval{ 0x10400, 0x10427 }, // L&  [40] DESERET CAPITAL LETTER LONG I..DESERET CAPITAL LETTER EW
    Interval{ 0x104B0, 0x104D3 }, // L&  [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Interval{ 0x10570, 0x1057A }, // L&  [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Interval{ 0x1057C, 0x1058A }, // L&  [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Interval{ 0x1058C, 0x10592 }, // L&   [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Interval{ 0x10594, 0x10595 }, // L&   [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Interval{ 0x10C80, 0x10CB2 }, // L&  [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Interval{ 0x10D50, 0x10D65 }, // L&  [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Interval{ 0x118A0, 0x118BF }, // L&  [32] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI CAPITAL LETTER VIYO
    Interval{ 0x16E40, 0x16E5F }, // L&  [32] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN CAPITAL LETTER Y
    Interval{ 0x16EA0, 0x16EB8 }, // L&  [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Interval{ 0x1D400, 0x1D419 }, // L&  [26] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL BOLD CAPITAL Z
    Interval{ 0x1D434, 0x1D44D }, // L&  [26] MATHEMATICAL ITALIC CAPITAL A..MATHEMATICAL ITALIC CAPITAL Z
    Interval{ 0x1D468, 0x1D481 }, // L&  [26] MATHEMATICAL BOLD ITALIC CAPITAL A..MATHEMATICAL BOLD ITALIC CAPITAL Z
    Interval{ 0x1D49C, 0x1D49C }, // L&       MATHEMATICAL SCRIPT CAPITAL A
    Interval{ 0x1D49E, 0x1D49F }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Interval{ 0x1D4A2, 0x1D4A2 }, // L&       MATHEMATICAL SCRIPT CAPITAL G
    Interval{ 0x1D4A5, 0x1D4A6 }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Interval{ 0x1D4A9, 0x1D4AC }, // L&   [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Interval{ 0x1D4AE, 0x1D4B5 }, // L&   [8] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT CAPITAL Z
    Interval{ 0x1D4D0, 0x1D4E9 }, // L&  [26] MATHEMATICAL BOLD SCRIPT CAPITAL A..MATHEMATICAL BOLD SCRIPT CAPITAL Z
    Interval{ 0x1D504, 0x1D505 }, // L&   [2] MATHEMATICAL FRAKTUR CAPITAL A..MATHEMATICAL FRAKTUR CAPITAL B
    Interval{ 0x1D507, 0x1D50A }, // L&   [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Interval{ 0x1D50D, 0x1D514 }, // L&   [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Interval{ 0x1D516, 0x1D51C }, // L&   [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Interval{ 0x1D538, 0x1D539 }, // L&   [2] MATHEMATICAL DOUBLE-STRUCK CAPITAL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Interval{ 0x1D53B, 0x1D53E }, // L&   [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Interval{ 0x1D540, 0x1D544 }, // L&   [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Interval{ 0x1D546, 0x1D546 }, // L&       MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Interval{ 0x1D54A, 0x1D550 }, // L&   [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Interval{ 0x1D56C, 0x1D585 }, // L&  [26] MATHEMATICAL BOLD FRAKTUR CAPITAL A..MATHEMATICAL BOLD FRAKTUR CAPITAL Z
    Interval{ 0x1D5A0, 0x1D5B9 }, // L&  [26] MATHEMATICAL SANS-SERIF CAPITAL A..MATHEMATICAL SANS-SERIF CAPITAL Z
    Interval{ 0x1D5D4, 0x1D5ED }, // L&  [26] MATHEMATICAL SANS-SERIF BOLD CAPITAL A..MATHEMATICAL SANS-SERIF BOLD CAPITAL Z
    Interval{ 0x1D608, 0x1D621 }, // L&  [26] MATHEMATICAL SANS-SERIF ITALIC CAPITAL A..MATHEMATICAL SANS-SERIF ITALIC CAPITAL Z
    Interval{ 0x1D63C, 0x1D655 }, // L&  [26] MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL A..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL Z
    Interval{ 0x1D670, 0x1D689 }, // L&  [26] MATHEMATICAL MONOSPACE CAPITAL A..MATHEMATICAL MONOSPACE CAPITAL Z
    Interval{ 0x1D6A8, 0x1D6C0 }, // L&  [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    Interval{ 0x1D6E2, 0x1D6FA }, // L&  [25] MATHEMATICAL ITALIC CAPITAL ALPHA..MATHEMATICAL ITALIC CAPITAL OMEGA
    Interval{ 0x1D71C, 0x1D734 }, // L&  [25] MATHEMATICAL BOLD ITALIC CAPITAL ALPHA..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D756, 0x1D76E }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD CAPITAL ALPHA..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    Interval{ 0x1D790, 0x1D7A8 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D7CA, 0x1D7CA }, // L&       MATHEMATICAL BOLD CAPITAL DIGAMMA
    Interval{ 0x1E900, 0x1E921 }, // L&  [34] ADLAM CAPITAL LETTER ALIF..ADLAM CAPITAL LETTER SHA
    Interval{ 0x1F130, 0x1F149 }, // So  [26] SQUARED LATIN CAPITAL LETTER A..SQUARED LATIN CAPITAL LETTER Z
    Interval{ 0x1F150, 0x1F169 }, // So  [26] NEGATIVE CIRCLED LATIN CAPITAL LETTER A..NEGATIVE CIRCLED LATIN CAPITAL LETTER Z
    Interval{ 0x1F170, 0x1F189 }, // So  [26] NEGATIVE SQUARED LATIN CAPITAL LETTER A..NEGATIVE SQUARED LATIN CAPITAL LETTER Z
}; // }}}
auto static const XID_Continue = std::array<Interval, 1422>{ // {{{
    Interval{ 0x0030, 0x0039 }, // Nd  [10] DIGIT ZERO..DIGIT NINE
    Interval{ 0x0041, 0x005A }, // L&  [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Interval{ 0x005F, 0x005F }, // Pc       LOW LINE
    Interval{ 0x0061, 0x007A }, // L&  [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Interval{ 0x00AA, 0x00AA }, // Lo       FEMININE ORDINAL INDICATOR
    Interval{ 0x00B5, 0x00B5 }, // L&       MICRO SIGN
    Interval{ 0x00B7, 0x00B7 }, // Po       MIDDLE DOT
    Interval{ 0x00BA, 0x00BA }, // Lo       MASCULINE ORDINAL INDICATOR
    Interval{ 0x00C0, 0x00D6 }, // L&  [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x00D8, 0x00F6 }, // L&  [31] LATIN CAPITAL LETTER O WITH STROKE..LATIN SMALL LETTER O WITH DIAERESIS
    Interval{ 0x00F8, 0x01BA }, // L& [195] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER EZH WITH TAIL
    Interval{ 0x01BB, 0x01BB }, // Lo       LATIN LETTER TWO WITH STROKE
    Interval{ 0x01BC, 0x01BF }, // L&   [4] LATIN CAPITAL LETTER TONE FIVE..LATIN LETTER WYNN
    Interval{ 0x01C0, 0x01C3 }, // Lo   [4] LATIN LETTER DENTAL CLICK..LATIN LETTER RETROFLEX CLICK
    Interval{ 0x01C4, 0x0293 }, // L& [208] LATIN CAPITAL LETTER DZ WITH CARON..LATIN SMALL LETTER EZH WITH CURL
    Interval{ 0x0294, 0x0295 }, // Lo   [2] LATIN LETTER GLOTTAL STOP..LATIN LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0x0296, 0x02AF }, // L&  [26] LATIN LETTER INVERTED GLOTTAL STOP..LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    Interval{ 0x02B0, 0x02C1 }, // Lm  [18] MODIFIER LETTER SMALL H..MODIFIER LETTER REVERSED GLOTTAL STOP
    Interval{ 0x02C6, 0x02D1 }, // Lm  [12] MODIFIER LETTER CIRCUMFLEX ACCENT..MODIFIER LETTER HALF TRIANGULAR COLON
    Interval{ 0x02E0, 0x02E4 }, // Lm   [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    Interval{ 0x02EC, 0x02EC }, // Lm       MODIFIER LETTER VOICING
    Interval{ 0x02EE, 0x02EE }, // Lm       MODIFIER LETTER DOUBLE APOSTROPHE
    Interval{ 0x0300, 0x036F }, // Mn [112] COMBINING GRAVE ACCENT..COMBINING LATIN SMALL LETTER X
    Interval{ 0x0370, 0x0373 }, // L&   [4] GREEK CAPITAL LETTER HETA..GREEK SMALL LETTER ARCHAIC SAMPI
    Interval{ 0x0374, 0x0374 }, // Lm       GREEK NUMERAL SIGN
    Interval{ 0x0376, 0x0377 }, // L&   [2] GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA..GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037B, 0x037D }, // L&   [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Interval{ 0x037F, 0x037F }, // L&       GREEK CAPITAL LETTER YOT
    Interval{ 0x0386, 0x0386 }, // L&       GREEK CAPITAL LETTER ALPHA WITH TONOS
    Interval{ 0x0387, 0x0387 }, // Po       GREEK ANO TELEIA
    Interval{ 0x0388, 0x038A }, // L&   [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Interval{ 0x038C, 0x038C }, // L&       GREEK CAPITAL LETTER OMICRON WITH TONOS
    Interval{ 0x038E, 0x03A1 }, // L&  [20] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER RHO
    Interval{ 0x03A3, 0x03F5 }, // L&  [83] GREEK CAPITAL LETTER SIGMA..GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F7, 0x0481 }, // L& [139] GREEK CAPITAL LETTER SHO..CYRILLIC SMALL LETTER KOPPA
    Interval{ 0x0483, 0x0487 }, // Mn   [5] COMBINING CYRILLIC TITLO..COMBINING CYRILLIC POKRYTIE
    Interval{ 0x048A, 0x052F }, // L& [166] CYRILLIC CAPITAL LETTER SHORT I WITH TAIL..CYRILLIC SMALL LETTER EL WITH DESCENDER
    Interval{ 0x0531, 0x0556 }, // L&  [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Interval{ 0x0559, 0x0559 }, // Lm       ARMENIAN MODIFIER LETTER LEFT HALF RING
    Interval{ 0x0560, 0x0588 }, // L&  [41] ARMENIAN SMALL LETTER TURNED AYB..ARMENIAN SMALL LETTER YI WITH STROKE
    Interval{ 0x0591, 0x05BD }, // Mn  [45] HEBREW ACCENT ETNAHTA..HEBREW POINT METEG
    Interval{ 0x05BF, 0x05BF }, // Mn       HEBREW POINT RAFE
    Interval{ 0x05C1, 0x05C2 }, // Mn   [2] HEBREW POINT SHIN DOT..HEBREW POINT SIN DOT
    Interval{ 0x05C4, 0x05C5 }, // Mn   [2] HEBREW MARK UPPER DOT..HEBREW MARK LOWER DOT
    Interval{ 0x05C7, 0x05C7 }, // Mn       HEBREW POINT QAMATS QATAN
    Interval{ 0x05D0, 0x05EA }, // Lo  [27] HEBREW LETTER ALEF..HEBREW LETTER TAV
    Interval{ 0x05EF, 0x05F2 }, // Lo   [4] HEBREW YOD TRIANGLE..HEBREW LIGATURE YIDDISH DOUBLE YOD
    Interval{ 0x0610, 0x061A }, // Mn  [11] ARABIC SIGN SALLALLAHOU ALAYHE WASSALLAM..ARABIC SMALL KASRA
    Interval{ 0x0620, 0x063F }, // Lo  [32] ARABIC LETTER KASHMIRI YEH..ARABIC LETTER FARSI YEH WITH THREE DOTS ABOVE
    Interval{ 0x0640, 0x0640 }, // Lm       ARABIC TATWEEL
    Interval{ 0x0641, 0x064A }, // Lo  [10] ARABIC LETTER FEH..ARABIC LETTER YEH
    Interval{ 0x064B, 0x065F }, // Mn  [21] ARABIC FATHATAN..ARABIC WAVY HAMZA BELOW
    Interval{ 0x0660, 0x0669 }, // Nd  [10] ARABIC-INDIC DIGIT ZERO..ARABIC-INDIC DIGIT NINE
    Interval{ 0x066E, 0x066F }, // Lo   [2] ARABIC LETTER DOTLESS BEH..ARABIC LETTER DOTLESS QAF
    Interval{ 0x0670, 0x0670 }, // Mn       ARABIC LETTER SUPERSCRIPT ALEF
    Interval{ 0x0671, 0x06D3 }, // Lo  [99] ARABIC LETTER ALEF WASLA..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE
    Interval{ 0x06D5, 0x06D5 }, // Lo       ARABIC LETTER AE
    Interval{ 0x06D6, 0x06DC }, // Mn   [7] ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA..ARABIC SMALL HIGH SEEN
    Interval{ 0x06DF, 0x06E4 }, // Mn   [6] ARABIC SMALL HIGH ROUNDED ZERO..ARABIC SMALL HIGH MADDA
    Interval{ 0x06E5, 0x06E6 }, // Lm   [2] ARABIC SMALL WAW..ARABIC SMALL YEH
    Interval{ 0x06E7, 0x06E8 }, // Mn   [2] ARABIC SMALL HIGH YEH..ARABIC SMALL HIGH NOON
    Interval{ 0x06EA, 0x06ED }, // Mn   [4] ARABIC EMPTY CENTRE LOW STOP..ARABIC SMALL LOW MEEM
    Interval{ 0x06EE, 0x06EF }, // Lo   [2] ARABIC LETTER DAL WITH INVERTED V..ARABIC LETTER REH WITH INVERTED V
    Interval{ 0x06F0, 0x06F9 }, // Nd  [10] EXTENDED ARABIC-INDIC DIGIT ZERO..EXTENDED ARABIC-INDIC DIGIT NINE
    Interval{ 0x06FA, 0x06FC }, // Lo   [3] ARABIC LETTER SHEEN WITH DOT BELOW..ARABIC LETTER GHAIN WITH DOT BELOW
    Interval{ 0x06FF, 0x06FF }, // Lo       ARABIC LETTER HEH WITH INVERTED V
    Interval{ 0x0710, 0x0710 }, // Lo       SYRIAC LETTER ALAPH
    Interval{ 0x0711, 0x0711 }, // Mn       SYRIAC LETTER SUPERSCRIPT ALAPH
    Interval{ 0x0712, 0x072F }, // Lo  [30] SYRIAC LETTER BETH..SYRIAC LETTER PERSIAN DHALATH
    Interval{ 0x0730, 0x074A }, // Mn  [27] SYRIAC PTHAHA ABOVE..SYRIAC BARREKH
    Interval{ 0x074D, 0x07A5 }, // Lo  [89] SYRIAC LETTER SOGDIAN ZHAIN..THAANA LETTER WAAVU
    Interval{ 0x07A6, 0x07B0 }, // Mn  [11] THAANA ABAFILI..THAANA SUKUN
    Interval{ 0x07B1, 0x07B1 }, // Lo       THAANA LETTER NAA
    Interval{ 0x07C0, 0x07C9 }, // Nd  [10] NKO DIGIT ZERO..NKO DIGIT NINE
    Interval{ 0x07CA, 0x07EA }, // Lo  [33] NKO LETTER A..NKO LETTER JONA RA
    Interval{ 0x07EB, 0x07F3 }, // Mn   [9] NKO COMBINING SHORT HIGH TONE..NKO COMBINING DOUBLE DOT ABOVE
    Interval{ 0x07F4, 0x07F5 }, // Lm   [2] NKO HIGH TONE APOSTROPHE..NKO LOW TONE APOSTROPHE
    Interval{ 0x07FA, 0x07FA }, // Lm       NKO LAJANYALAN
    Interval{ 0x07FD, 0x07FD }, // Mn       NKO DANTAYALAN
    Interval{ 0x0800, 0x0815 }, // Lo  [22] SAMARITAN LETTER ALAF..SAMARITAN LETTER TAAF
    Interval{ 0x0816, 0x0819 }, // Mn   [4] SAMARITAN MARK IN..SAMARITAN MARK DAGESH
    Interval{ 0x081A, 0x081A }, // Lm       SAMARITAN MODIFIER LETTER EPENTHETIC YUT
    Interval{ 0x081B, 0x0823 }, // Mn   [9] SAMARITAN MARK EPENTHETIC YUT..SAMARITAN VOWEL SIGN A
    Interval{ 0x0824, 0x0824 }, // Lm       SAMARITAN MODIFIER LETTER SHORT A
    Interval{ 0x0825, 0x0827 }, // Mn   [3] SAMARITAN VOWEL SIGN SHORT A..SAMARITAN VOWEL SIGN U
    Interval{ 0x0828, 0x0828 }, // Lm       SAMARITAN MODIFIER LETTER I
    Interval{ 0x0829, 0x082D }, // Mn   [5] SAMARITAN VOWEL SIGN LONG I..SAMARITAN MARK NEQUDAA
    Interval{ 0x0840, 0x0858 }, // Lo  [25] MANDAIC LETTER HALQA..MANDAIC LETTER AIN
    Interval{ 0x0859, 0x085B }, // Mn   [3] MANDAIC AFFRICATION MARK..MANDAIC GEMINATION MARK
    Interval{ 0x0860, 0x086A }, // Lo  [11] SYRIAC LETTER MALAYALAM NGA..SYRIAC LETTER MALAYALAM SSA
    Interval{ 0x0870, 0x0887 }, // Lo  [24] ARABIC LETTER ALEF WITH ATTACHED FATHA..ARABIC BASELINE ROUND DOT
    Interval{ 0x0889, 0x088F }, // Lo   [7] ARABIC LETTER NOON WITH INVERTED SMALL V..ARABIC LETTER NOON WITH RING ABOVE
    Interval{ 0x0897, 0x089F }, // Mn   [9] ARABIC PEPET..ARABIC HALF MADDA OVER MADDA
    Interval{ 0x08A0, 0x08C8 }, // Lo  [41] ARABIC LETTER BEH WITH SMALL V BELOW..ARABIC LETTER GRAF
    Interval{ 0x08C9, 0x08C9 }, // Lm       ARABIC SMALL FARSI YEH
    Interval{ 0x08CA, 0x08E1 }, // Mn  [24] ARABIC SMALL HIGH FARSI YEH..ARABIC SMALL HIGH SIGN SAFHA
    Interval{ 0x08E3, 0x0902 }, // Mn  [32] ARABIC TURNED DAMMA BELOW..DEVANAGARI SIGN ANUSVARA
    Interval{ 0x0903, 0x0903 }, // Mc       DEVANAGARI SIGN VISARGA
    Interval{ 0x0904, 0x0939 }, // Lo  [54] DEVANAGARI LETTER SHORT A..DEVANAGARI LETTER HA
    Interval{ 0x093A, 0x093A }, // Mn       DEVANAGARI VOWEL SIGN OE
    Interval{ 0x093B, 0x093B }, // Mc       DEVANAGARI VOWEL SIGN OOE
    Interval{ 0x093C, 0x093C }, // Mn       DEVANAGARI SIGN NUKTA
    Interval{ 0x093D, 0x093D }, // Lo       DEVANAGARI SIGN AVAGRAHA
    Interval{ 0x093E, 0x0940 }, // Mc   [3] DEVANAGARI VOWEL SIGN AA..DEVANAGARI VOWEL SIGN II
    Interval{ 0x0941, 0x0948 }, // Mn   [8] DEVANAGARI VOWEL SIGN U..DEVANAGARI VOWEL SIGN AI
    Interval{ 0x0949, 0x094C }, // Mc   [4] DEVANAGARI VOWEL SIGN CANDRA O..DEVANAGARI VOWEL SIGN AU
    Interval{ 0x094D, 0x094D }, // Mn       DEVANAGARI SIGN VIRAMA
    Interval{ 0x094E, 0x094F }, // Mc   [2] DEVANAGARI VOWEL SIGN PRISHTHAMATRA E..DEVANAGARI VOWEL SIGN AW
    Interval{ 0x0950, 0x0950 }, // Lo       DEVANAGARI OM
    Interval{ 0x0951, 0x0957 }, // Mn   [7] DEVANAGARI STRESS SIGN UDATTA..DEVANAGARI VOWEL SIGN UUE
    Interval{ 0x0958, 0x0961 }, // Lo  [10] DEVANAGARI LETTER QA..DEVANAGARI LETTER VOCALIC LL
    Interval{ 0x0962, 0x0963 }, // Mn   [2] DEVANAGARI VOWEL SIGN VOCALIC L..DEVANAGARI VOWEL SIGN VOCALIC LL
    Interval{ 0x0966, 0x096F }, // Nd  [10] DEVANAGARI DIGIT ZERO..DEVANAGARI DIGIT NINE
    Interval{ 0x0971, 0x0971 }, // Lm       DEVANAGARI SIGN HIGH SPACING DOT
    Interval{ 0x0972, 0x0980 }, // Lo  [15] DEVANAGARI LETTER CANDRA A..BENGALI ANJI
    Interval{ 0x0981, 0x0981 }, // Mn       BENGALI SIGN CANDRABINDU
    Interval{ 0x0982, 0x0983 }, // Mc   [2] BENGALI SIGN ANUSVARA..BENGALI SIGN VISARGA
    Interval{ 0x0985, 0x098C }, // Lo   [8] BENGALI LETTER A..BENGALI LETTER VOCALIC L
    Interval{ 0x098F, 0x0990 }, // Lo   [2] BENGALI LETTER E..BENGALI LETTER AI
    Interval{ 0x0993, 0x09A8 }, // Lo  [22] BENGALI LETTER O..BENGALI LETTER NA
    Interval{ 0x09AA, 0x09B0 }, // Lo   [7] BENGALI LETTER PA..BENGALI LETTER RA
    Interval{ 0x09B2, 0x09B2 }, // Lo       BENGALI LETTER LA
    Interval{ 0x09B6, 0x09B9 }, // Lo   [4] BENGALI LETTER SHA..BENGALI LETTER HA
    Interval{ 0x09BC, 0x09BC }, // Mn       BENGALI SIGN NUKTA
    Interval{ 0x09BD, 0x09BD }, // Lo       BENGALI SIGN AVAGRAHA
    Interval{ 0x09BE, 0x09C0 }, // Mc   [3] BENGALI VOWEL SIGN AA..BENGALI VOWEL SIGN II
    Interval{ 0x09C1, 0x09C4 }, // Mn   [4] BENGALI VOWEL SIGN U..BENGALI VOWEL SIGN VOCALIC RR
    Interval{ 0x09C7, 0x09C8 }, // Mc   [2] BENGALI VOWEL SIGN E..BENGALI VOWEL SIGN AI
    Interval{ 0x09CB, 0x09CC }, // Mc   [2] BENGALI VOWEL SIGN O..BENGALI VOWEL SIGN AU
    Interval{ 0x09CD, 0x09CD }, // Mn       BENGALI SIGN VIRAMA
    Interval{ 0x09CE, 0x09CE }, // Lo       BENGALI LETTER KHANDA TA
    Interval{ 0x09D7, 0x09D7 }, // Mc       BENGALI AU LENGTH MARK
    Interval{ 0x09DC, 0x09DD }, // Lo   [2] BENGALI LETTER RRA..BENGALI LETTER RHA
    Interval{ 0x09DF, 0x09E1 }, // Lo   [3] BENGALI LETTER YYA..BENGALI LETTER VOCALIC LL
    Interval{ 0x09E2, 0x09E3 }, // Mn   [2] BENGALI VOWEL SIGN VOCALIC L..BENGALI VOWEL SIGN VOCALIC LL
    Interval{ 0x09E6, 0x09EF }, // Nd  [10] BENGALI DIGIT ZERO..BENGALI DIGIT NINE
    Interval{ 0x09F0, 0x09F1 }, // Lo   [2] BENGALI LETTER RA WITH MIDDLE DIAGONAL..BENGALI LETTER RA WITH LOWER DIAGONAL
    Interval{ 0x09FC, 0x09FC }, // Lo       BENGALI LETTER VEDIC ANUSVARA
    Interval{ 0x09FE, 0x09FE }, // Mn       BENGALI SANDHI MARK
    Interval{ 0x0A01, 0x0A02 }, // Mn   [2] GURMUKHI SIGN ADAK BINDI..GURMUKHI SIGN BINDI
    Interval{ 0x0A03, 0x0A03 }, // Mc       GURMUKHI SIGN VISARGA
    Interval{ 0x0A05, 0x0A0A }, // Lo   [6] GURMUKHI LETTER A..GURMUKHI LETTER UU
    Interval{ 0x0A0F, 0x0A10 }, // Lo   [2] GURMUKHI LETTER EE..GURMUKHI LETTER AI
    Interval{ 0x0A13, 0x0A28 }, // Lo  [22] GURMUKHI LETTER OO..GURMUKHI LETTER NA
    Interval{ 0x0A2A, 0x0A30 }, // Lo   [7] GURMUKHI LETTER PA..GURMUKHI LETTER RA
    Interval{ 0x0A32, 0x0A33 }, // Lo   [2] GURMUKHI LETTER LA..GURMUKHI LETTER LLA
    Interval{ 0x0A35, 0x0A36 }, // Lo   [2] GURMUKHI LETTER VA..GURMUKHI LETTER SHA
    Interval{ 0x0A38, 0x0A39 }, // Lo   [2] GURMUKHI LETTER SA..GURMUKHI LETTER HA
    Interval{ 0x0A3C, 0x0A3C }, // Mn       GURMUKHI SIGN NUKTA
    Interval{ 0x0A3E, 0x0A40 }, // Mc   [3] GURMUKHI VOWEL SIGN AA..GURMUKHI VOWEL SIGN II
    Interval{ 0x0A41, 0x0A42 }, // Mn   [2] GURMUKHI VOWEL SIGN U..GURMUKHI VOWEL SIGN UU
    Interval{ 0x0A47, 0x0A48 }, // Mn   [2] GURMUKHI VOWEL SIGN EE..GURMUKHI VOWEL SIGN AI
    Interval{ 0x0A4B, 0x0A4D }, // Mn   [3] GURMUKHI VOWEL SIGN OO..GURMUKHI SIGN VIRAMA
    Interval{ 0x0A51, 0x0A51 }, // Mn       GURMUKHI SIGN UDAAT
    Interval{ 0x0A59, 0x0A5C }, // Lo   [4] GURMUKHI LETTER KHHA..GURMUKHI LETTER RRA
    Interval{ 0x0A5E, 0x0A5E }, // Lo       GURMUKHI LETTER FA
    Interval{ 0x0A66, 0x0A6F }, // Nd  [10] GURMUKHI DIGIT ZERO..GURMUKHI DIGIT NINE
    Interval{ 0x0A70, 0x0A71 }, // Mn   [2] GURMUKHI TIPPI..GURMUKHI ADDAK
    Interval{ 0x0A72, 0x0A74 }, // Lo   [3] GURMUKHI IRI..GURMUKHI EK ONKAR
    Interval{ 0x0A75, 0x0A75 }, // Mn       GURMUKHI SIGN YAKASH
    Interval{ 0x0A81, 0x0A82 }, // Mn   [2] GUJARATI SIGN CANDRABINDU..GUJARATI SIGN ANUSVARA
    Interval{ 0x0A83, 0x0A83 }, // Mc       GUJARATI SIGN VISARGA
    Interval{ 0x0A85, 0x0A8D }, // Lo   [9] GUJARATI LETTER A..GUJARATI VOWEL CANDRA E
    Interval{ 0x0A8F, 0x0A91 }, // Lo   [3] GUJARATI LETTER E..GUJARATI VOWEL CANDRA O
    Interval{ 0x0A93, 0x0AA8 }, // Lo  [22] GUJARATI LETTER O..GUJARATI LETTER NA
    Interval{ 0x0AAA, 0x0AB0 }, // Lo   [7] GUJARATI LETTER PA..GUJARATI LETTER RA
    Interval{ 0x0AB2, 0x0AB3 }, // Lo   [2] GUJARATI LETTER LA..GUJARATI LETTER LLA
    Interval{ 0x0AB5, 0x0AB9 }, // Lo   [5] GUJARATI LETTER VA..GUJARATI LETTER HA
    Interval{ 0x0ABC, 0x0ABC }, // Mn       GUJARATI SIGN NUKTA
    Interval{ 0x0ABD, 0x0ABD }, // Lo       GUJARATI SIGN AVAGRAHA
    Interval{ 0x0ABE, 0x0AC0 }, // Mc   [3] GUJARATI VOWEL SIGN AA..GUJARATI VOWEL SIGN II
    Interval{ 0x0AC1, 0x0AC5 }, // Mn   [5] GUJARATI VOWEL SIGN U..GUJARATI VOWEL SIGN CANDRA E
    Interval{ 0x0AC7, 0x0AC8 }, // Mn   [2] GUJARATI VOWEL SIGN E..GUJARATI VOWEL SIGN AI
    Interval{ 0x0AC9, 0x0AC9 }, // Mc       GUJARATI VOWEL SIGN CANDRA O
    Interval{ 0x0ACB, 0x0ACC }, // Mc   [2] GUJARATI VOWEL SIGN O..GUJARATI VOWEL SIGN AU
    Interval{ 0x0ACD, 0x0ACD }, // Mn       GUJARATI SIGN VIRAMA
    Interval{ 0x0AD0, 0x0AD0 }, // Lo       GUJARATI OM
    Interval{ 0x0AE0, 0x0AE1 }, // Lo   [2] GUJARATI LETTER VOCALIC RR..GUJARATI LETTER VOCALIC LL
    Interval{ 0x0AE2, 0x0AE3 }, // Mn   [2] GUJARATI VOWEL SIGN VOCALIC L..GUJARATI VOWEL SIGN VOCALIC LL
    Interval{ 0x0AE6, 0x0AEF }, // Nd  [10] GUJARATI DIGIT ZERO..GUJARATI DIGIT NINE
    Interval{ 0x0AF9, 0x0AF9 }, // Lo       GUJARATI LETTER ZHA
    Interval{ 0x0AFA, 0x0AFF }, // Mn   [6] GUJARATI SIGN SUKUN..GUJARATI SIGN TWO-CIRCLE NUKTA ABOVE
    Interval{ 0x0B01, 0x0B01 }, // Mn       ORIYA SIGN CANDRABINDU
    Interval{ 0x0B02, 0x0B03 }, // Mc   [2] ORIYA SIGN ANUSVARA..ORIYA SIGN VISARGA
    Interval{ 0x0B05, 0x0B0C }, // Lo   [8] ORIYA LETTER A..ORIYA LETTER VOCALIC L
    Interval{ 0x0B0F, 0x0B10 }, // Lo   [2] ORIYA LETTER E..ORIYA LETTER AI
    Interval{ 0x0B13, 0x0B28 }, // Lo  [22] ORIYA LETTER O..ORIYA LETTER NA
    Interval{ 0x0B2A, 0x0B30 }, // Lo   [7] ORIYA LETTER PA..ORIYA LETTER RA
    Interval{ 0x0B32, 0x0B33 }, // Lo   [2] ORIYA LETTER LA..ORIYA LETTER LLA
    Interval{ 0x0B35, 0x0B39 }, // Lo   [5] ORIYA LETTER VA..ORIYA LETTER HA
    Interval{ 0x0B3C, 0x0B3C }, // Mn       ORIYA SIGN NUKTA
    Interval{ 0x0B3D, 0x0B3D }, // Lo       ORIYA SIGN AVAGRAHA
    Interval{ 0x0B3E, 0x0B3E }, // Mc       ORIYA VOWEL SIGN AA
    Interval{ 0x0B3F, 0x0B3F }, // Mn       ORIYA VOWEL SIGN I
    Interval{ 0x0B40, 0x0B40 }, // Mc       ORIYA VOWEL SIGN II
    Interval{ 0x0B41, 0x0B44 }, // Mn   [4] ORIYA VOWEL SIGN U..ORIYA VOWEL SIGN VOCALIC RR
    Interval{ 0x0B47, 0x0B48 }, // Mc   [2] ORIYA VOWEL SIGN E..ORIYA VOWEL SIGN AI
    Interval{ 0x0B4B, 0x0B4C }, // Mc   [2] ORIYA VOWEL SIGN O..ORIYA VOWEL SIGN AU
    Interval{ 0x0B4D, 0x0B4D }, // Mn       ORIYA SIGN VIRAMA
    Interval{ 0x0B55, 0x0B56 }, // Mn   [2] ORIYA SIGN OVERLINE..ORIYA AI LENGTH MARK
    Interval{ 0x0B57, 0x0B57 }, // Mc       ORIYA AU LENGTH MARK
    Interval{ 0x0B5C, 0x0B5D }, // Lo   [2] ORIYA LETTER RRA..ORIYA LETTER RHA
    Interval{ 0x0B5F, 0x0B61 }, // Lo   [3] ORIYA LETTER YYA..ORIYA LETTER VOCALIC LL
    Interval{ 0x0B62, 0x0B63 }, // Mn   [2] ORIYA VOWEL SIGN VOCALIC L..ORIYA VOWEL SIGN VOCALIC LL
    Interval{ 0x0B66, 0x0B6F }, // Nd  [10] ORIYA DIGIT ZERO..ORIYA DIGIT NINE
    Interval{ 0x0B71, 0x0B71 }, // Lo       ORIYA LETTER WA
    Interval{ 0x0B82, 0x0B82 }, // Mn       TAMIL SIGN ANUSVARA
    Interval{ 0x0B83, 0x0B83 }, // Lo       TAMIL SIGN VISARGA
    Interval{ 0x0B85, 0x0B8A }, // Lo   [6] TAMIL LETTER A..TAMIL LETTER UU
    Interval{ 0x0B8E, 0x0B90 }, // Lo   [3] TAMIL LETTER E..TAMIL LETTER AI
    Interval{ 0x0B92, 0x0B95 }, // Lo   [4] TAMIL LETTER O..TAMIL LETTER KA
    Interval{ 0x0B99, 0x0B9A }, // Lo   [2] TAMIL LETTER NGA..TAMIL LETTER CA
    Interval{ 0x0B9C, 0x0B9C }, // Lo       TAMIL LETTER JA
    Interval{ 0x0B9E, 0x0B9F }, // Lo   [2] TAMIL LETTER NYA..TAMIL LETTER TTA
    Interval{ 0x0BA3, 0x0BA4 }, // Lo   [2] TAMIL LETTER NNA..TAMIL LETTER TA
    Interval{ 0x0BA8, 0x0BAA }, // Lo   [3] TAMIL LETTER NA..TAMIL LETTER PA
    Interval{ 0x0BAE, 0x0BB9 }, // Lo  [12] TAMIL LETTER MA..TAMIL LETTER HA
    Interval{ 0x0BBE, 0x0BBF }, // Mc   [2] TAMIL VOWEL SIGN AA..TAMIL VOWEL SIGN I
    Interval{ 0x0BC0, 0x0BC0 }, // Mn       TAMIL VOWEL SIGN II
    Interval{ 0x0BC1, 0x0BC2 }, // Mc   [2] TAMIL VOWEL SIGN U..TAMIL VOWEL SIGN UU
    Interval{ 0x0BC6, 0x0BC8 }, // Mc   [3] TAMIL VOWEL SIGN E..TAMIL VOWEL SIGN AI
    Interval{ 0x0BCA, 0x0BCC }, // Mc   [3] TAMIL VOWEL SIGN O..TAMIL VOWEL SIGN AU
    Interval{ 0x0BCD, 0x0BCD }, // Mn       TAMIL SIGN VIRAMA
    Interval{ 0x0BD0, 0x0BD0 }, // Lo       TAMIL OM
    Interval{ 0x0BD7, 0x0BD7 }, // Mc       TAMIL AU LENGTH MARK
    Interval{ 0x0BE6, 0x0BEF }, // Nd  [10] TAMIL DIGIT ZERO..TAMIL DIGIT NINE
    Interval{ 0x0C00, 0x0C00 }, // Mn       TELUGU SIGN COMBINING CANDRABINDU ABOVE
    Interval{ 0x0C01, 0x0C03 }, // Mc   [3] TELUGU SIGN CANDRABINDU..TELUGU SIGN VISARGA
    Interval{ 0x0C04, 0x0C04 }, // Mn       TELUGU SIGN COMBINING ANUSVARA ABOVE
    Interval{ 0x0C05, 0x0C0C }, // Lo   [8] TELUGU LETTER A..TELUGU LETTER VOCALIC L
    Interval{ 0x0C0E, 0x0C10 }, // Lo   [3] TELUGU LETTER E..TELUGU LETTER AI
    Interval{ 0x0C12, 0x0C28 }, // Lo  [23] TELUGU LETTER O..TELUGU LETTER NA
    Interval{ 0x0C2A, 0x0C39 }, // Lo  [16] TELUGU LETTER PA..TELUGU LETTER HA
    Interval{ 0x0C3C, 0x0C3C }, // Mn       TELUGU SIGN NUKTA
    Interval{ 0x0C3D, 0x0C3D }, // Lo       TELUGU SIGN AVAGRAHA
    Interval{ 0x0C3E, 0x0C40 }, // Mn   [3] TELUGU VOWEL SIGN AA..TELUGU VOWEL SIGN II
    Interval{ 0x0C41, 0x0C44 }, // Mc   [4] TELUGU VOWEL SIGN U..TELUGU VOWEL SIGN VOCALIC RR
    Interval{ 0x0C46, 0x0C48 }, // Mn   [3] TELUGU VOWEL SIGN E..TELUGU VOWEL SIGN AI
    Interval{ 0x0C4A, 0x0C4D }, // Mn   [4] TELUGU VOWEL SIGN O..TELUGU SIGN VIRAMA
    Interval{ 0x0C55, 0x0C56 }, // Mn   [2] TELUGU LENGTH MARK..TELUGU AI LENGTH MARK
    Interval{ 0x0C58, 0x0C5A }, // Lo   [3] TELUGU LETTER TSA..TELUGU LETTER RRRA
    Interval{ 0x0C5C, 0x0C5D }, // Lo   [2] TELUGU ARCHAIC SHRII..TELUGU LETTER NAKAARA POLLU
    Interval{ 0x0C60, 0x0C61 }, // Lo   [2] TELUGU LETTER VOCALIC RR..TELUGU LETTER VOCALIC LL
    Interval{ 0x0C62, 0x0C63 }, // Mn   [2] TELUGU VOWEL SIGN VOCALIC L..TELUGU VOWEL SIGN VOCALIC LL
    Interval{ 0x0C66, 0x0C6F }, // Nd  [10] TELUGU DIGIT ZERO..TELUGU DIGIT NINE
    Interval{ 0x0C80, 0x0C80 }, // Lo       KANNADA SIGN SPACING CANDRABINDU
    Interval{ 0x0C81, 0x0C81 }, // Mn       KANNADA SIGN CANDRABINDU
    Interval{ 0x0C82, 0x0C83 }, // Mc   [2] KANNADA SIGN ANUSVARA..KANNADA SIGN VISARGA
    Interval{ 0x0C85, 0x0C8C }, // Lo   [8] KANNADA LETTER A..KANNADA LETTER VOCALIC L
    Interval{ 0x0C8E, 0x0C90 }, // Lo   [3] KANNADA LETTER E..KANNADA LETTER AI
    Interval{ 0x0C92, 0x0CA8 }, // Lo  [23] KANNADA LETTER O..KANNADA LETTER NA
    Interval{ 0x0CAA, 0x0CB3 }, // Lo  [10] KANNADA LETTER PA..KANNADA LETTER LLA
    Interval{ 0x0CB5, 0x0CB9 }, // Lo   [5] KANNADA LETTER VA..KANNADA LETTER HA
    Interval{ 0x0CBC, 0x0CBC }, // Mn       KANNADA SIGN NUKTA
    Interval{ 0x0CBD, 0x0CBD }, // Lo       KANNADA SIGN AVAGRAHA
    Interval{ 0x0CBE, 0x0CBE }, // Mc       KANNADA VOWEL SIGN AA
    Interval{ 0x0CBF, 0x0CBF }, // Mn       KANNADA VOWEL SIGN I
    Interval{ 0x0CC0, 0x0CC4 }, // Mc   [5] KANNADA VOWEL SIGN II..KANNADA VOWEL SIGN VOCALIC RR
    Interval{ 0x0CC6, 0x0CC6 }, // Mn       KANNADA VOWEL SIGN E
    Interval{ 0x0CC7, 0x0CC8 }, // Mc   [2] KANNADA VOWEL SIGN EE..KANNADA VOWEL SIGN AI
    Interval{ 0x0CCA, 0x0CCB }, // Mc   [2] KANNADA VOWEL SIGN O..KANNADA VOWEL SIGN OO
    Interval{ 0x0CCC, 0x0CCD }, // Mn   [2] KANNADA VOWEL SIGN AU..KANNADA SIGN VIRAMA
    Interval{ 0x0CD5, 0x0CD6 }, // Mc   [2] KANNADA LENGTH MARK..KANNADA AI LENGTH MARK
    Interval{ 0x0CDC, 0x0CDE }, // Lo   [3] KANNADA ARCHAIC SHRII..KANNADA LETTER FA
    Interval{ 0x0CE0, 0x0CE1 }, // Lo   [2] KANNADA LETTER VOCALIC RR..KANNADA LETTER VOCALIC LL
    Interval{ 0x0CE2, 0x0CE3 }, // Mn   [2] KANNADA VOWEL SIGN VOCALIC L..KANNADA VOWEL SIGN VOCALIC LL
    Interval{ 0x0CE6, 0x0CEF }, // Nd  [10] KANNADA DIGIT ZERO..KANNADA DIGIT NINE
    Interval{ 0x0CF1, 0x0CF2 }, // Lo   [2] KANNADA SIGN JIHVAMULIYA..KANNADA SIGN UPADHMANIYA
    Interval{ 0x0CF3, 0x0CF3 }, // Mc       KANNADA SIGN COMBINING ANUSVARA ABOVE RIGHT
    Interval{ 0x0D00, 0x0D01 }, // Mn   [2] MALAYALAM SIGN COMBINING ANUSVARA ABOVE..MALAYALAM SIGN CANDRABINDU
    Interval{ 0x0D02, 0x0D03 }, // Mc   [2] MALAYALAM SIGN ANUSVARA..MALAYALAM SIGN VISARGA
    Interval{ 0x0D04, 0x0D0C }, // Lo   [9] MALAYALAM LETTER VEDIC ANUSVARA..MALAYALAM LETTER VOCALIC L
    Interval{ 0x0D0E, 0x0D10 }, // Lo   [3] MALAYALAM LETTER E..MALAYALAM LETTER AI
    Interval{ 0x0D12, 0x0D3A }, // Lo  [41] MALAYALAM LETTER O..MALAYALAM LETTER TTTA
    Interval{ 0x0D3B, 0x0D3C }, // Mn   [2] MALAYALAM SIGN VERTICAL BAR VIRAMA..MALAYALAM SIGN CIRCULAR VIRAMA
    Interval{ 0x0D3D, 0x0D3D }, // Lo       MALAYALAM SIGN AVAGRAHA
    Interval{ 0x0D3E, 0x0D40 }, // Mc   [3] MALAYALAM VOWEL SIGN AA..MALAYALAM VOWEL SIGN II
    Interval{ 0x0D41, 0x0D44 }, // Mn   [4] MALAYALAM VOWEL SIGN U..MALAYALAM VOWEL SIGN VOCALIC RR
    Interval{ 0x0D46, 0x0D48 }, // Mc   [3] MALAYALAM VOWEL SIGN E..MALAYALAM VOWEL SIGN AI
    Interval{ 0x0D4A, 0x0D4C }, // Mc   [3] MALAYALAM VOWEL SIGN O..MALAYALAM VOWEL SIGN AU
    Interval{ 0x0D4D, 0x0D4D }, // Mn       MALAYALAM SIGN VIRAMA
    Interval{ 0x0D4E, 0x0D4E }, // Lo       MALAYALAM LETTER DOT REPH
    Interval{ 0x0D54, 0x0D56 }, // Lo   [3] MALAYALAM LETTER CHILLU M..MALAYALAM LETTER CHILLU LLL
    Interval{ 0x0D57, 0x0D57 }, // Mc       MALAYALAM AU LENGTH MARK
    Interval{ 0x0D5F, 0x0D61 }, // Lo   [3] MALAYALAM LETTER ARCHAIC II..MALAYALAM LETTER VOCALIC LL
    Interval{ 0x0D62, 0x0D63 }, // Mn   [2] MALAYALAM VOWEL SIGN VOCALIC L..MALAYALAM VOWEL SIGN VOCALIC LL
    Interval{ 0x0D66, 0x0D6F }, // Nd  [10] MALAYALAM DIGIT ZERO..MALAYALAM DIGIT NINE
    Interval{ 0x0D7A, 0x0D7F }, // Lo   [6] MALAYALAM LETTER CHILLU NN..MALAYALAM LETTER CHILLU K
    Interval{ 0x0D81, 0x0D81 }, // Mn       SINHALA SIGN CANDRABINDU
    Interval{ 0x0D82, 0x0D83 }, // Mc   [2] SINHALA SIGN ANUSVARAYA..SINHALA SIGN VISARGAYA
    Interval{ 0x0D85, 0x0D96 }, // Lo  [18] SINHALA LETTER AYANNA..SINHALA LETTER AUYANNA
    Interval{ 0x0D9A, 0x0DB1 }, // Lo  [24] SINHALA LETTER ALPAPRAANA KAYANNA..SINHALA LETTER DANTAJA NAYANNA
    Interval{ 0x0DB3, 0x0DBB }, // Lo   [9] SINHALA LETTER SANYAKA DAYANNA..SINHALA LETTER RAYANNA
    Interval{ 0x0DBD, 0x0DBD }, // Lo       SINHALA LETTER DANTAJA LAYANNA
    Interval{ 0x0DC0, 0x0DC6 }, // Lo   [7] SINHALA LETTER VAYANNA..SINHALA LETTER FAYANNA
    Interval{ 0x0DCA, 0x0DCA }, // Mn       SINHALA SIGN AL-LAKUNA
    Interval{ 0x0DCF, 0x0DD1 }, // Mc   [3] SINHALA VOWEL SIGN AELA-PILLA..SINHALA VOWEL SIGN DIGA AEDA-PILLA
    Interval{ 0x0DD2, 0x0DD4 }, // Mn   [3] SINHALA VOWEL SIGN KETTI IS-PILLA..SINHALA VOWEL SIGN KETTI PAA-PILLA
    Interval{ 0x0DD6, 0x0DD6 }, // Mn       SINHALA VOWEL SIGN DIGA PAA-PILLA
    Interval{ 0x0DD8, 0x0DDF }, // Mc   [8] SINHALA VOWEL SIGN GAETTA-PILLA..SINHALA VOWEL SIGN GAYANUKITTA
    Interval{ 0x0DE6, 0x0DEF }, // Nd  [10] SINHALA LITH DIGIT ZERO..SINHALA LITH DIGIT NINE
    Interval{ 0x0DF2, 0x0DF3 }, // Mc   [2] SINHALA VOWEL SIGN DIGA GAETTA-PILLA..SINHALA VOWEL SIGN DIGA GAYANUKITTA
    Interval{ 0x0E01, 0x0E30 }, // Lo  [48] THAI CHARACTER KO KAI..THAI CHARACTER SARA A
    Interval{ 0x0E31, 0x0E31 }, // Mn       THAI CHARACTER MAI HAN-AKAT
    Interval{ 0x0E32, 0x0E33 }, // Lo   [2] THAI CHARACTER SARA AA..THAI CHARACTER SARA AM
    Interval{ 0x0E34, 0x0E3A }, // Mn   [7] THAI CHARACTER SARA I..THAI CHARACTER PHINTHU
    Interval{ 0x0E40, 0x0E45 }, // Lo   [6] THAI CHARACTER SARA E..THAI CHARACTER LAKKHANGYAO
    Interval{ 0x0E46, 0x0E46 }, // Lm       THAI CHARACTER MAIYAMOK
    Interval{ 0x0E47, 0x0E4E }, // Mn   [8] THAI CHARACTER MAITAIKHU..THAI CHARACTER YAMAKKAN
    Interval{ 0x0E50, 0x0E59 }, // Nd  [10] THAI DIGIT ZERO..THAI DIGIT NINE
    Interval{ 0x0E81, 0x0E82 }, // Lo   [2] LAO LETTER KO..LAO LETTER KHO SUNG
    Interval{ 0x0E84, 0x0E84 }, // Lo       LAO LETTER KHO TAM
    Interval{ 0x0E86, 0x0E8A }, // Lo   [5] LAO LETTER PALI GHA..LAO LETTER SO TAM
    Interval{ 0x0E8C, 0x0EA3 }, // Lo  [24] LAO LETTER PALI JHA..LAO LETTER LO LING
    Interval{ 0x0EA5, 0x0EA5 }, // Lo       LAO LETTER LO LOOT
    Interval{ 0x0EA7, 0x0EB0 }, // Lo  [10] LAO LETTER WO..LAO VOWEL SIGN A
    Interval{ 0x0EB1, 0x0EB1 }, // Mn       LAO VOWEL SIGN MAI KAN
    Interval{ 0x0EB2, 0x0EB3 }, // Lo   [2] LAO VOWEL SIGN AA..LAO VOWEL SIGN AM
    Interval{ 0x0EB4, 0x0EBC }, // Mn   [9] LAO VOWEL SIGN I..LAO SEMIVOWEL SIGN LO
    Interval{ 0x0EBD, 0x0EBD }, // Lo       LAO SEMIVOWEL SIGN NYO
    Interval{ 0x0EC0, 0x0EC4 }, // Lo   [5] LAO VOWEL SIGN E..LAO VOWEL SIGN AI
    Interval{ 0x0EC6, 0x0EC6 }, // Lm       LAO KO LA
    Interval{ 0x0EC8, 0x0ECE }, // Mn   [7] LAO TONE MAI EK..LAO YAMAKKAN
    Interval{ 0x0ED0, 0x0ED9 }, // Nd  [10] LAO DIGIT ZERO..LAO DIGIT NINE
    Interval{ 0x0EDC, 0x0EDF }, // Lo   [4] LAO HO NO..LAO LETTER KHMU NYO
    Interval{ 0x0F00, 0x0F00 }, // Lo       TIBETAN SYLLABLE OM
    Interval{ 0x0F18, 0x0F19 }, // Mn   [2] TIBETAN ASTROLOGICAL SIGN -KHYUD PA..TIBETAN ASTROLOGICAL SIGN SDONG TSHUGS
    Interval{ 0x0F20, 0x0F29 }, // Nd  [10] TIBETAN DIGIT ZERO..TIBETAN DIGIT NINE
    Interval{ 0x0F35, 0x0F35 }, // Mn       TIBETAN MARK NGAS BZUNG NYI ZLA
    Interval{ 0x0F37, 0x0F37 }, // Mn       TIBETAN MARK NGAS BZUNG SGOR RTAGS
    Interval{ 0x0F39, 0x0F39 }, // Mn       TIBETAN MARK TSA -PHRU
    Interval{ 0x0F3E, 0x0F3F }, // Mc   [2] TIBETAN SIGN YAR TSHES..TIBETAN SIGN MAR TSHES
    Interval{ 0x0F40, 0x0F47 }, // Lo   [8] TIBETAN LETTER KA..TIBETAN LETTER JA
    Interval{ 0x0F49, 0x0F6C }, // Lo  [36] TIBETAN LETTER NYA..TIBETAN LETTER RRA
    Interval{ 0x0F71, 0x0F7E }, // Mn  [14] TIBETAN VOWEL SIGN AA..TIBETAN SIGN RJES SU NGA RO
    Interval{ 0x0F7F, 0x0F7F }, // Mc       TIBETAN SIGN RNAM BCAD
    Interval{ 0x0F80, 0x0F84 }, // Mn   [5] TIBETAN VOWEL SIGN REVERSED I..TIBETAN MARK HALANTA
    Interval{ 0x0F86, 0x0F87 }, // Mn   [2] TIBETAN SIGN LCI RTAGS..TIBETAN SIGN YANG RTAGS
    Interval{ 0x0F88, 0x0F8C }, // Lo   [5] TIBETAN SIGN LCE TSA CAN..TIBETAN SIGN INVERTED MCHU CAN
    Interval{ 0x0F8D, 0x0F97 }, // Mn  [11] TIBETAN SUBJOINED SIGN LCE TSA CAN..TIBETAN SUBJOINED LETTER JA
    Interval{ 0x0F99, 0x0FBC }, // Mn  [36] TIBETAN SUBJOINED LETTER NYA..TIBETAN SUBJOINED LETTER FIXED-FORM RA
    Interval{ 0x0FC6, 0x0FC6 }, // Mn       TIBETAN SYMBOL PADMA GDAN
    Interval{ 0x1000, 0x102A }, // Lo  [43] MYANMAR LETTER KA..MYANMAR LETTER AU
    Interval{ 0x102B, 0x102C }, // Mc   [2] MYANMAR VOWEL SIGN TALL AA..MYANMAR VOWEL SIGN AA
    Interval{ 0x102D, 0x1030 }, // Mn   [4] MYANMAR VOWEL SIGN I..MYANMAR VOWEL SIGN UU
    Interval{ 0x1031, 0x1031 }, // Mc       MYANMAR VOWEL SIGN E
    Interval{ 0x1032, 0x1037 }, // Mn   [6] MYANMAR VOWEL SIGN AI..MYANMAR SIGN DOT BELOW
    Interval{ 0x1038, 0x1038 }, // Mc       MYANMAR SIGN VISARGA
    Interval{ 0x1039, 0x103A }, // Mn   [2] MYANMAR SIGN VIRAMA..MYANMAR SIGN ASAT
    Interval{ 0x103B, 0x103C }, // Mc   [2] MYANMAR CONSONANT SIGN MEDIAL YA..MYANMAR CONSONANT SIGN MEDIAL RA
    Interval{ 0x103D, 0x103E }, // Mn   [2] MYANMAR CONSONANT SIGN MEDIAL WA..MYANMAR CONSONANT SIGN MEDIAL HA
    Interval{ 0x103F, 0x103F }, // Lo       MYANMAR LETTER GREAT SA
    Interval{ 0x1040, 0x1049 }, // Nd  [10] MYANMAR DIGIT ZERO..MYANMAR DIGIT NINE
    Interval{ 0x1050, 0x1055 }, // Lo   [6] MYANMAR LETTER SHA..MYANMAR LETTER VOCALIC LL
    Interval{ 0x1056, 0x1057 }, // Mc   [2] MYANMAR VOWEL SIGN VOCALIC R..MYANMAR VOWEL SIGN VOCALIC RR
    Interval{ 0x1058, 0x1059 }, // Mn   [2] MYANMAR VOWEL SIGN VOCALIC L..MYANMAR VOWEL SIGN VOCALIC LL
    Interval{ 0x105A, 0x105D }, // Lo   [4] MYANMAR LETTER MON NGA..MYANMAR LETTER MON BBE
    Interval{ 0x105E, 0x1060 }, // Mn   [3] MYANMAR CONSONANT SIGN MON MEDIAL NA..MYANMAR CONSONANT SIGN MON MEDIAL LA
    Interval{ 0x1061, 0x1061 }, // Lo       MYANMAR LETTER SGAW KAREN SHA
    Interval{ 0x1062, 0x1064 }, // Mc   [3] MYANMAR VOWEL SIGN SGAW KAREN EU..MYANMAR TONE MARK SGAW KAREN KE PHO
    Interval{ 0x1065, 0x1066 }, // Lo   [2] MYANMAR LETTER WESTERN PWO KAREN THA..MYANMAR LETTER WESTERN PWO KAREN PWA
    Interval{ 0x1067, 0x106D }, // Mc   [7] MYANMAR VOWEL SIGN WESTERN PWO KAREN EU..MYANMAR SIGN WESTERN PWO KAREN TONE-5
    Interval{ 0x106E, 0x1070 }, // Lo   [3] MYANMAR LETTER EASTERN PWO KAREN NNA..MYANMAR LETTER EASTERN PWO KAREN GHWA
    Interval{ 0x1071, 0x1074 }, // Mn   [4] MYANMAR VOWEL SIGN GEBA KAREN I..MYANMAR VOWEL SIGN KAYAH EE
    Interval{ 0x1075, 0x1081 }, // Lo  [13] MYANMAR LETTER SHAN KA..MYANMAR LETTER SHAN HA
    Interval{ 0x1082, 0x1082 }, // Mn       MYANMAR CONSONANT SIGN SHAN MEDIAL WA
    Interval{ 0x1083, 0x1084 }, // Mc   [2] MYANMAR VOWEL SIGN SHAN AA..MYANMAR VOWEL SIGN SHAN E
    Interval{ 0x1085, 0x1086 }, // Mn   [2] MYANMAR VOWEL SIGN SHAN E ABOVE..MYANMAR VOWEL SIGN SHAN FINAL Y
    Interval{ 0x1087, 0x108C }, // Mc   [6] MYANMAR SIGN SHAN TONE-2..MYANMAR SIGN SHAN COUNCIL TONE-3
    Interval{ 0x108D, 0x108D }, // Mn       MYANMAR SIGN SHAN COUNCIL EMPHATIC TONE
    Interval{ 0x108E, 0x108E }, // Lo       MYANMAR LETTER RUMAI PALAUNG FA
    Interval{ 0x108F, 0x108F }, // Mc       MYANMAR SIGN RUMAI PALAUNG TONE-5
    Interval{ 0x1090, 0x1099 }, // Nd  [10] MYANMAR SHAN DIGIT ZERO..MYANMAR SHAN DIGIT NINE
    Interval{ 0x109A, 0x109C }, // Mc   [3] MYANMAR SIGN KHAMTI TONE-1..MYANMAR VOWEL SIGN AITON A
    Interval{ 0x109D, 0x109D }, // Mn       MYANMAR VOWEL SIGN AITON AI
    Interval{ 0x10A0, 0x10C5 }, // L&  [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Interval{ 0x10C7, 0x10C7 }, // L&       GEORGIAN CAPITAL LETTER YN
    Interval{ 0x10CD, 0x10CD }, // L&       GEORGIAN CAPITAL LETTER AEN
    Interval{ 0x10D0, 0x10FA }, // L&  [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Interval{ 0x10FC, 0x10FC }, // Lm       MODIFIER LETTER GEORGIAN NAR
    Interval{ 0x10FD, 0x10FF }, // L&   [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Interval{ 0x1100, 0x1248 }, // Lo [329] HANGUL CHOSEONG KIYEOK..ETHIOPIC SYLLABLE QWA
    Interval{ 0x124A, 0x124D }, // Lo   [4] ETHIOPIC SYLLABLE QWI..ETHIOPIC SYLLABLE QWE
    Interval{ 0x1250, 0x1256 }, // Lo   [7] ETHIOPIC SYLLABLE QHA..ETHIOPIC SYLLABLE QHO
    Interval{ 0x1258, 0x1258 }, // Lo       ETHIOPIC SYLLABLE QHWA
    Interval{ 0x125A, 0x125D }, // Lo   [4] ETHIOPIC SYLLABLE QHWI..ETHIOPIC SYLLABLE QHWE
    Interval{ 0x1260, 0x1288 }, // Lo  [41] ETHIOPIC SYLLABLE BA..ETHIOPIC SYLLABLE XWA
    Interval{ 0x128A, 0x128D }, // Lo   [4] ETHIOPIC SYLLABLE XWI..ETHIOPIC SYLLABLE XWE
    Interval{ 0x1290, 0x12B0 }, // Lo  [33] ETHIOPIC SYLLABLE NA..ETHIOPIC SYLLABLE KWA
    Interval{ 0x12B2, 0x12B5 }, // Lo   [4] ETHIOPIC SYLLABLE KWI..ETHIOPIC SYLLABLE KWE
    Interval{ 0x12B8, 0x12BE }, // Lo   [7] ETHIOPIC SYLLABLE KXA..ETHIOPIC SYLLABLE KXO
    Interval{ 0x12C0, 0x12C0 }, // Lo       ETHIOPIC SYLLABLE KXWA
    Interval{ 0x12C2, 0x12C5 }, // Lo   [4] ETHIOPIC SYLLABLE KXWI..ETHIOPIC SYLLABLE KXWE
    Interval{ 0x12C8, 0x12D6 }, // Lo  [15] ETHIOPIC SYLLABLE WA..ETHIOPIC SYLLABLE PHARYNGEAL O
    Interval{ 0x12D8, 0x1310 }, // Lo  [57] ETHIOPIC SYLLABLE ZA..ETHIOPIC SYLLABLE GWA
    Interval{ 0x1312, 0x1315 }, // Lo   [4] ETHIOPIC SYLLABLE GWI..ETHIOPIC SYLLABLE GWE
    Interval{ 0x1318, 0x135A }, // Lo  [67] ETHIOPIC SYLLABLE GGA..ETHIOPIC SYLLABLE FYA
    Interval{ 0x135D, 0x135F }, // Mn   [3] ETHIOPIC COMBINING GEMINATION AND VOWEL LENGTH MARK..ETHIOPIC COMBINING GEMINATION MARK
    Interval{ 0x1369, 0x1371 }, // No   [9] ETHIOPIC DIGIT ONE..ETHIOPIC DIGIT NINE
    Interval{ 0x1380, 0x138F }, // Lo  [16] ETHIOPIC SYLLABLE SEBATBEIT MWA..ETHIOPIC SYLLABLE PWE
    Interval{ 0x13A0, 0x13F5 }, // L&  [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Interval{ 0x13F8, 0x13FD }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1401, 0x166C }, // Lo [620] CANADIAN SYLLABICS E..CANADIAN SYLLABICS CARRIER TTSA
    Interval{ 0x166F, 0x167F }, // Lo  [17] CANADIAN SYLLABICS QAI..CANADIAN SYLLABICS BLACKFOOT W
    Interval{ 0x1681, 0x169A }, // Lo  [26] OGHAM LETTER BEITH..OGHAM LETTER PEITH
    Interval{ 0x16A0, 0x16EA }, // Lo  [75] RUNIC LETTER FEHU FEOH FE F..RUNIC LETTER X
    Interval{ 0x16EE, 0x16F0 }, // Nl   [3] RUNIC ARLAUG SYMBOL..RUNIC BELGTHOR SYMBOL
    Interval{ 0x16F1, 0x16F8 }, // Lo   [8] RUNIC LETTER K..RUNIC LETTER FRANKS CASKET AESC
    Interval{ 0x1700, 0x1711 }, // Lo  [18] TAGALOG LETTER A..TAGALOG LETTER HA
    Interval{ 0x1712, 0x1714 }, // Mn   [3] TAGALOG VOWEL SIGN I..TAGALOG SIGN VIRAMA
    Interval{ 0x1715, 0x1715 }, // Mc       TAGALOG SIGN PAMUDPOD
    Interval{ 0x171F, 0x1731 }, // Lo  [19] TAGALOG LETTER ARCHAIC RA..HANUNOO LETTER HA
    Interval{ 0x1732, 0x1733 }, // Mn   [2] HANUNOO VOWEL SIGN I..HANUNOO VOWEL SIGN U
    Interval{ 0x1734, 0x1734 }, // Mc       HANUNOO SIGN PAMUDPOD
    Interval{ 0x1740, 0x1751 }, // Lo  [18] BUHID LETTER A..BUHID LETTER HA
    Interval{ 0x1752, 0x1753 }, // Mn   [2] BUHID VOWEL SIGN I..BUHID VOWEL SIGN U
    Interval{ 0x1760, 0x176C }, // Lo  [13] TAGBANWA LETTER A..TAGBANWA LETTER YA
    Interval{ 0x176E, 0x1770 }, // Lo   [3] TAGBANWA LETTER LA..TAGBANWA LETTER SA
    Interval{ 0x1772, 0x1773 }, // Mn   [2] TAGBANWA VOWEL SIGN I..TAGBANWA VOWEL SIGN U
    Interval{ 0x1780, 0x17B3 }, // Lo  [52] KHMER LETTER KA..KHMER INDEPENDENT VOWEL QAU
    Interval{ 0x17B4, 0x17B5 }, // Mn   [2] KHMER VOWEL INHERENT AQ..KHMER VOWEL INHERENT AA
    Interval{ 0x17B6, 0x17B6 }, // Mc       KHMER VOWEL SIGN AA
    Interval{ 0x17B7, 0x17BD }, // Mn   [7] KHMER VOWEL SIGN I..KHMER VOWEL SIGN UA
    Interval{ 0x17BE, 0x17C5 }, // Mc   [8] KHMER VOWEL SIGN OE..KHMER VOWEL SIGN AU
    Interval{ 0x17C6, 0x17C6 }, // Mn       KHMER SIGN NIKAHIT
    Interval{ 0x17C7, 0x17C8 }, // Mc   [2] KHMER SIGN REAHMUK..KHMER SIGN YUUKALEAPINTU
    Interval{ 0x17C9, 0x17D3 }, // Mn  [11] KHMER SIGN MUUSIKATOAN..KHMER SIGN BATHAMASAT
    Interval{ 0x17D7, 0x17D7 }, // Lm       KHMER SIGN LEK TOO
    Interval{ 0x17DC, 0x17DC }, // Lo       KHMER SIGN AVAKRAHASANYA
    Interval{ 0x17DD, 0x17DD }, // Mn       KHMER SIGN ATTHACAN
    Interval{ 0x17E0, 0x17E9 }, // Nd  [10] KHMER DIGIT ZERO..KHMER DIGIT NINE
    Interval{ 0x180B, 0x180D }, // Mn   [3] MONGOLIAN FREE VARIATION SELECTOR ONE..MONGOLIAN FREE VARIATION SELECTOR THREE
    Interval{ 0x180F, 0x180F }, // Mn       MONGOLIAN FREE VARIATION SELECTOR FOUR
    Interval{ 0x1810, 0x1819 }, // Nd  [10] MONGOLIAN DIGIT ZERO..MONGOLIAN DIGIT NINE
    Interval{ 0x1820, 0x1842 }, // Lo  [35] MONGOLIAN LETTER A..MONGOLIAN LETTER CHI
    Interval{ 0x1843, 0x1843 }, // Lm       MONGOLIAN LETTER TODO LONG VOWEL SIGN
    Interval{ 0x1844, 0x1878 }, // Lo  [53] MONGOLIAN LETTER TODO E..MONGOLIAN LETTER CHA WITH TWO DOTS
    Interval{ 0x1880, 0x1884 }, // Lo   [5] MONGOLIAN LETTER ALI GALI ANUSVARA ONE..MONGOLIAN LETTER ALI GALI INVERTED UBADAMA
    Interval{ 0x1885, 0x1886 }, // Mn   [2] MONGOLIAN LETTER ALI GALI BALUDA..MONGOLIAN LETTER ALI GALI THREE BALUDA
    Interval{ 0x1887, 0x18A8 }, // Lo  [34] MONGOLIAN LETTER ALI GALI A..MONGOLIAN LETTER MANCHU ALI GALI BHA
    Interval{ 0x18A9, 0x18A9 }, // Mn       MONGOLIAN LETTER ALI GALI DAGALGA
    Interval{ 0x18AA, 0x18AA }, // Lo       MONGOLIAN LETTER MANCHU ALI GALI LHA
    Interval{ 0x18B0, 0x18F5 }, // Lo  [70] CANADIAN SYLLABICS OY..CANADIAN SYLLABICS CARRIER DENTAL S
    Interval{ 0x1900, 0x191E }, // Lo  [31] LIMBU VOWEL-CARRIER LETTER..LIMBU LETTER TRA
    Interval{ 0x1920, 0x1922 }, // Mn   [3] LIMBU VOWEL SIGN A..LIMBU VOWEL SIGN U
    Interval{ 0x1923, 0x1926 }, // Mc   [4] LIMBU VOWEL SIGN EE..LIMBU VOWEL SIGN AU
    Interval{ 0x1927, 0x1928 }, // Mn   [2] LIMBU VOWEL SIGN E..LIMBU VOWEL SIGN O
    Interval{ 0x1929, 0x192B }, // Mc   [3] LIMBU SUBJOINED LETTER YA..LIMBU SUBJOINED LETTER WA
    Interval{ 0x1930, 0x1931 }, // Mc   [2] LIMBU SMALL LETTER KA..LIMBU SMALL LETTER NGA
    Interval{ 0x1932, 0x1932 }, // Mn       LIMBU SMALL LETTER ANUSVARA
    Interval{ 0x1933, 0x1938 }, // Mc   [6] LIMBU SMALL LETTER TA..LIMBU SMALL LETTER LA
    Interval{ 0x1939, 0x193B }, // Mn   [3] LIMBU SIGN MUKPHRENG..LIMBU SIGN SA-I
    Interval{ 0x1946, 0x194F }, // Nd  [10] LIMBU DIGIT ZERO..LIMBU DIGIT NINE
    Interval{ 0x1950, 0x196D }, // Lo  [30] TAI LE LETTER KA..TAI LE LETTER AI
    Interval{ 0x1970, 0x1974 }, // Lo   [5] TAI LE LETTER TONE-2..TAI LE LETTER TONE-6
    Interval{ 0x1980, 0x19AB }, // Lo  [44] NEW TAI LUE LETTER HIGH QA..NEW TAI LUE LETTER LOW SUA
    Interval{ 0x19B0, 0x19C9 }, // Lo  [26] NEW TAI LUE VOWEL SIGN VOWEL SHORTENER..NEW TAI LUE TONE MARK-2
    Interval{ 0x19D0, 0x19D9 }, // Nd  [10] NEW TAI LUE DIGIT ZERO..NEW TAI LUE DIGIT NINE
    Interval{ 0x19DA, 0x19DA }, // No       NEW TAI LUE THAM DIGIT ONE
    Interval{ 0x1A00, 0x1A16 }, // Lo  [23] BUGINESE LETTER KA..BUGINESE LETTER HA
    Interval{ 0x1A17, 0x1A18 }, // Mn   [2] BUGINESE VOWEL SIGN I..BUGINESE VOWEL SIGN U
    Interval{ 0x1A19, 0x1A1A }, // Mc   [2] BUGINESE VOWEL SIGN E..BUGINESE VOWEL SIGN O
    Interval{ 0x1A1B, 0x1A1B }, // Mn       BUGINESE VOWEL SIGN AE
    Interval{ 0x1A20, 0x1A54 }, // Lo  [53] TAI THAM LETTER HIGH KA..TAI THAM LETTER GREAT SA
    Interval{ 0x1A55, 0x1A55 }, // Mc       TAI THAM CONSONANT SIGN MEDIAL RA
    Interval{ 0x1A56, 0x1A56 }, // Mn       TAI THAM CONSONANT SIGN MEDIAL LA
    Interval{ 0x1A57, 0x1A57 }, // Mc       TAI THAM CONSONANT SIGN LA TANG LAI
    Interval{ 0x1A58, 0x1A5E }, // Mn   [7] TAI THAM SIGN MAI KANG LAI..TAI THAM CONSONANT SIGN SA
    Interval{ 0x1A60, 0x1A60 }, // Mn       TAI THAM SIGN SAKOT
    Interval{ 0x1A61, 0x1A61 }, // Mc       TAI THAM VOWEL SIGN A
    Interval{ 0x1A62, 0x1A62 }, // Mn       TAI THAM VOWEL SIGN MAI SAT
    Interval{ 0x1A63, 0x1A64 }, // Mc   [2] TAI THAM VOWEL SIGN AA..TAI THAM VOWEL SIGN TALL AA
    Interval{ 0x1A65, 0x1A6C }, // Mn   [8] TAI THAM VOWEL SIGN I..TAI THAM VOWEL SIGN OA BELOW
    Interval{ 0x1A6D, 0x1A72 }, // Mc   [6] TAI THAM VOWEL SIGN OY..TAI THAM VOWEL SIGN THAM AI
    Interval{ 0x1A73, 0x1A7C }, // Mn  [10] TAI THAM VOWEL SIGN OA ABOVE..TAI THAM SIGN KHUEN-LUE KARAN
    Interval{ 0x1A7F, 0x1A7F }, // Mn       TAI THAM COMBINING CRYPTOGRAMMIC DOT
    Interval{ 0x1A80, 0x1A89 }, // Nd  [10] TAI THAM HORA DIGIT ZERO..TAI THAM HORA DIGIT NINE
    Interval{ 0x1A90, 0x1A99 }, // Nd  [10] TAI THAM THAM DIGIT ZERO..TAI THAM THAM DIGIT NINE
    Interval{ 0x1AA7, 0x1AA7 }, // Lm       TAI THAM SIGN MAI YAMOK
    Interval{ 0x1AB0, 0x1ABD }, // Mn  [14] COMBINING DOUBLED CIRCUMFLEX ACCENT..COMBINING PARENTHESES BELOW
    Interval{ 0x1ABF, 0x1ADD }, // Mn  [31] COMBINING LATIN SMALL LETTER W BELOW..COMBINING DOT-AND-RING BELOW
    Interval{ 0x1AE0, 0x1AEB }, // Mn  [12] COMBINING LEFT TACK ABOVE..COMBINING DOUBLE RIGHTWARDS ARROW ABOVE
    Interval{ 0x1B00, 0x1B03 }, // Mn   [4] BALINESE SIGN ULU RICEM..BALINESE SIGN SURANG
    Interval{ 0x1B04, 0x1B04 }, // Mc       BALINESE SIGN BISAH
    Interval{ 0x1B05, 0x1B33 }, // Lo  [47] BALINESE LETTER AKARA..BALINESE LETTER HA
    Interval{ 0x1B34, 0x1B34 }, // Mn       BALINESE SIGN REREKAN
    Interval{ 0x1B35, 0x1B35 }, // Mc       BALINESE VOWEL SIGN TEDUNG
    Interval{ 0x1B36, 0x1B3A }, // Mn   [5] BALINESE VOWEL SIGN ULU..BALINESE VOWEL SIGN RA REPA
    Interval{ 0x1B3B, 0x1B3B }, // Mc       BALINESE VOWEL SIGN RA REPA TEDUNG
    Interval{ 0x1B3C, 0x1B3C }, // Mn       BALINESE VOWEL SIGN LA LENGA
    Interval{ 0x1B3D, 0x1B41 }, // Mc   [5] BALINESE VOWEL SIGN LA LENGA TEDUNG..BALINESE VOWEL SIGN TALING REPA TEDUNG
    Interval{ 0x1B42, 0x1B42 }, // Mn       BALINESE VOWEL SIGN PEPET
    Interval{ 0x1B43, 0x1B44 }, // Mc   [2] BALINESE VOWEL SIGN PEPET TEDUNG..BALINESE ADEG ADEG
    Interval{ 0x1B45, 0x1B4C }, // Lo   [8] BALINESE LETTER KAF SASAK..BALINESE LETTER ARCHAIC JNYA
    Interval{ 0x1B50, 0x1B59 }, // Nd  [10] BALINESE DIGIT ZERO..BALINESE DIGIT NINE
    Interval{ 0x1B6B, 0x1B73 }, // Mn   [9] BALINESE MUSICAL SYMBOL COMBINING TEGEH..BALINESE MUSICAL SYMBOL COMBINING GONG
    Interval{ 0x1B80, 0x1B81 }, // Mn   [2] SUNDANESE SIGN PANYECEK..SUNDANESE SIGN PANGLAYAR
    Interval{ 0x1B82, 0x1B82 }, // Mc       SUNDANESE SIGN PANGWISAD
    Interval{ 0x1B83, 0x1BA0 }, // Lo  [30] SUNDANESE LETTER A..SUNDANESE LETTER HA
    Interval{ 0x1BA1, 0x1BA1 }, // Mc       SUNDANESE CONSONANT SIGN PAMINGKAL
    Interval{ 0x1BA2, 0x1BA5 }, // Mn   [4] SUNDANESE CONSONANT SIGN PANYAKRA..SUNDANESE VOWEL SIGN PANYUKU
    Interval{ 0x1BA6, 0x1BA7 }, // Mc   [2] SUNDANESE VOWEL SIGN PANAELAENG..SUNDANESE VOWEL SIGN PANOLONG
    Interval{ 0x1BA8, 0x1BA9 }, // Mn   [2] SUNDANESE VOWEL SIGN PAMEPET..SUNDANESE VOWEL SIGN PANEULEUNG
    Interval{ 0x1BAA, 0x1BAA }, // Mc       SUNDANESE SIGN PAMAAEH
    Interval{ 0x1BAB, 0x1BAD }, // Mn   [3] SUNDANESE SIGN VIRAMA..SUNDANESE CONSONANT SIGN PASANGAN WA
    Interval{ 0x1BAE, 0x1BAF }, // Lo   [2] SUNDANESE LETTER KHA..SUNDANESE LETTER SYA
    Interval{ 0x1BB0, 0x1BB9 }, // Nd  [10] SUNDANESE DIGIT ZERO..SUNDANESE DIGIT NINE
    Interval{ 0x1BBA, 0x1BE5 }, // Lo  [44] SUNDANESE AVAGRAHA..BATAK LETTER U
    Interval{ 0x1BE6, 0x1BE6 }, // Mn       BATAK SIGN TOMPI
    Interval{ 0x1BE7, 0x1BE7 }, // Mc       BATAK VOWEL SIGN E
    Interval{ 0x1BE8, 0x1BE9 }, // Mn   [2] BATAK VOWEL SIGN PAKPAK E..BATAK VOWEL SIGN EE
    Interval{ 0x1BEA, 0x1BEC }, // Mc   [3] BATAK VOWEL SIGN I..BATAK VOWEL SIGN O
    Interval{ 0x1BED, 0x1BED }, // Mn       BATAK VOWEL SIGN KARO O
    Interval{ 0x1BEE, 0x1BEE }, // Mc       BATAK VOWEL SIGN U
    Interval{ 0x1BEF, 0x1BF1 }, // Mn   [3] BATAK VOWEL SIGN U FOR SIMALUNGUN SA..BATAK CONSONANT SIGN H
    Interval{ 0x1BF2, 0x1BF3 }, // Mc   [2] BATAK PANGOLAT..BATAK PANONGONAN
    Interval{ 0x1C00, 0x1C23 }, // Lo  [36] LEPCHA LETTER KA..LEPCHA LETTER A
    Interval{ 0x1C24, 0x1C2B }, // Mc   [8] LEPCHA SUBJOINED LETTER YA..LEPCHA VOWEL SIGN UU
    Interval{ 0x1C2C, 0x1C33 }, // Mn   [8] LEPCHA VOWEL SIGN E..LEPCHA CONSONANT SIGN T
    Interval{ 0x1C34, 0x1C35 }, // Mc   [2] LEPCHA CONSONANT SIGN NYIN-DO..LEPCHA CONSONANT SIGN KANG
    Interval{ 0x1C36, 0x1C37 }, // Mn   [2] LEPCHA SIGN RAN..LEPCHA SIGN NUKTA
    Interval{ 0x1C40, 0x1C49 }, // Nd  [10] LEPCHA DIGIT ZERO..LEPCHA DIGIT NINE
    Interval{ 0x1C4D, 0x1C4F }, // Lo   [3] LEPCHA LETTER TTA..LEPCHA LETTER DDA
    Interval{ 0x1C50, 0x1C59 }, // Nd  [10] OL CHIKI DIGIT ZERO..OL CHIKI DIGIT NINE
    Interval{ 0x1C5A, 0x1C77 }, // Lo  [30] OL CHIKI LETTER LA..OL CHIKI LETTER OH
    Interval{ 0x1C78, 0x1C7D }, // Lm   [6] OL CHIKI MU TTUDDAG..OL CHIKI AHAD
    Interval{ 0x1C80, 0x1C8A }, // L&  [11] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER TJE
    Interval{ 0x1C90, 0x1CBA }, // L&  [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Interval{ 0x1CBD, 0x1CBF }, // L&   [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Interval{ 0x1CD0, 0x1CD2 }, // Mn   [3] VEDIC TONE KARSHANA..VEDIC TONE PRENKHA
    Interval{ 0x1CD4, 0x1CE0 }, // Mn  [13] VEDIC SIGN YAJURVEDIC MIDLINE SVARITA..VEDIC TONE RIGVEDIC KASHMIRI INDEPENDENT SVARITA
    Interval{ 0x1CE1, 0x1CE1 }, // Mc       VEDIC TONE ATHARVAVEDIC INDEPENDENT SVARITA
    Interval{ 0x1CE2, 0x1CE8 }, // Mn   [7] VEDIC SIGN VISARGA SVARITA..VEDIC SIGN VISARGA ANUDATTA WITH TAIL
    Interval{ 0x1CE9, 0x1CEC }, // Lo   [4] VEDIC SIGN ANUSVARA ANTARGOMUKHA..VEDIC SIGN ANUSVARA VAMAGOMUKHA WITH TAIL
    Interval{ 0x1CED, 0x1CED }, // Mn       VEDIC SIGN TIRYAK
    Interval{ 0x1CEE, 0x1CF3 }, // Lo   [6] VEDIC SIGN HEXIFORM LONG ANUSVARA..VEDIC SIGN ROTATED ARDHAVISARGA
    Interval{ 0x1CF4, 0x1CF4 }, // Mn       VEDIC TONE CANDRA ABOVE
    Interval{ 0x1CF5, 0x1CF6 }, // Lo   [2] VEDIC SIGN JIHVAMULIYA..VEDIC SIGN UPADHMANIYA
    Interval{ 0x1CF7, 0x1CF7 }, // Mc       VEDIC SIGN ATIKRAMA
    Interval{ 0x1CF8, 0x1CF9 }, // Mn   [2] VEDIC TONE RING ABOVE..VEDIC TONE DOUBLE RING ABOVE
    Interval{ 0x1CFA, 0x1CFA }, // Lo       VEDIC SIGN DOUBLE ANUSVARA ANTARGOMUKHA
    Interval{ 0x1D00, 0x1D2B }, // L&  [44] LATIN LETTER SMALL CAPITAL A..CYRILLIC LETTER SMALL CAPITAL EL
    Interval{ 0x1D2C, 0x1D6A }, // Lm  [63] MODIFIER LETTER CAPITAL A..GREEK SUBSCRIPT SMALL LETTER CHI
    Interval{ 0x1D6B, 0x1D77 }, // L&  [13] LATIN SMALL LETTER UE..LATIN SMALL LETTER TURNED G
    Interval{ 0x1D78, 0x1D78 }, // Lm       MODIFIER LETTER CYRILLIC EN
    Interval{ 0x1D79, 0x1D9A }, // L&  [34] LATIN SMALL LETTER INSULAR G..LATIN SMALL LETTER EZH WITH RETROFLEX HOOK
    Interval{ 0x1D9B, 0x1DBF }, // Lm  [37] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL THETA
    Interval{ 0x1DC0, 0x1DFF }, // Mn  [64] COMBINING DOTTED GRAVE ACCENT..COMBINING RIGHT ARROWHEAD AND DOWN ARROWHEAD BELOW
    Interval{ 0x1E00, 0x1F15 }, // L& [278] LATIN CAPITAL LETTER A WITH RING BELOW..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F18, 0x1F1D }, // L&   [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F20, 0x1F45 }, // L&  [38] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F48, 0x1F4D }, // L&   [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F50, 0x1F57 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F59, 0x1F59 }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA
    Interval{ 0x1F5B, 0x1F5B }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Interval{ 0x1F5D, 0x1F5D }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Interval{ 0x1F5F, 0x1F7D }, // L&  [31] GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI..GREEK SMALL LETTER OMEGA WITH OXIA
    Interval{ 0x1F80, 0x1FB4 }, // L&  [53] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB6, 0x1FBC }, // L&   [7] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Interval{ 0x1FBE, 0x1FBE }, // L&       GREEK PROSGEGRAMMENI
    Interval{ 0x1FC2, 0x1FC4 }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC6, 0x1FCC }, // L&   [7] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Interval{ 0x1FD0, 0x1FD3 }, // L&   [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Interval{ 0x1FD6, 0x1FDB }, // L&   [6] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK CAPITAL LETTER IOTA WITH OXIA
    Interval{ 0x1FE0, 0x1FEC }, // L&  [13] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Interval{ 0x1FF2, 0x1FF4 }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF6, 0x1FFC }, // L&   [7] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Interval{ 0x200C, 0x200D }, // Cf   [2] ZERO WIDTH NON-JOINER..ZERO WIDTH JOINER
    Interval{ 0x203F, 0x2040 }, // Pc   [2] UNDERTIE..CHARACTER TIE
    Interval{ 0x2054, 0x2054 }, // Pc       INVERTED UNDERTIE
    Interval{ 0x2071, 0x2071 }, // Lm       SUPERSCRIPT LATIN SMALL LETTER I
    Interval{ 0x207F, 0x207F }, // Lm       SUPERSCRIPT LATIN SMALL LETTER N
    Interval{ 0x2090, 0x209C }, // Lm  [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Interval{ 0x20D0, 0x20DC }, // Mn  [13] COMBINING LEFT HARPOON ABOVE..COMBINING FOUR DOTS ABOVE
    Interval{ 0x20E1, 0x20E1 }, // Mn       COMBINING LEFT RIGHT ARROW ABOVE
    Interval{ 0x20E5, 0x20F0 }, // Mn  [12] COMBINING REVERSE SOLIDUS OVERLAY..COMBINING ASTERISK ABOVE
    Interval{ 0x2102, 0x2102 }, // L&       DOUBLE-STRUCK CAPITAL C
    Interval{ 0x2107, 0x2107 }, // L&       EULER CONSTANT
    Interval{ 0x210A, 0x2113 }, // L&  [10] SCRIPT SMALL G..SCRIPT SMALL L
    Interval{ 0x2115, 0x2115 }, // L&       DOUBLE-STRUCK CAPITAL N
    Interval{ 0x2118, 0x2118 }, // Sm       SCRIPT CAPITAL P
    Interval{ 0x2119, 0x211D }, // L&   [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    Interval{ 0x2124, 0x2124 }, // L&       DOUBLE-STRUCK CAPITAL Z
    Interval{ 0x2126, 0x2126 }, // L&       OHM SIGN
    Interval{ 0x2128, 0x2128 }, // L&       BLACK-LETTER CAPITAL Z
    Interval{ 0x212A, 0x212D }, // L&   [4] KELVIN SIGN..BLACK-LETTER CAPITAL C
    Interval{ 0x212E, 0x212E }, // So       ESTIMATED SYMBOL
    Interval{ 0x212F, 0x2134 }, // L&   [6] SCRIPT SMALL E..SCRIPT SMALL O
    Interval{ 0x2135, 0x2138 }, // Lo   [4] ALEF SYMBOL..DALET SYMBOL
    Interval{ 0x2139, 0x2139 }, // L&       INFORMATION SOURCE
    Interval{ 0x213C, 0x213F }, // L&   [4] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK CAPITAL PI
    Interval{ 0x2145, 0x2149 }, // L&   [5] DOUBLE-STRUCK ITALIC CAPITAL D..DOUBLE-STRUCK ITALIC SMALL J
    Interval{ 0x214E, 0x214E }, // L&       TURNED SMALL F
    Interval{ 0x2160, 0x2182 }, // Nl  [35] ROMAN NUMERAL ONE..ROMAN NUMERAL TEN THOUSAND
    Interval{ 0x2183, 0x2184 }, // L&   [2] ROMAN NUMERAL REVERSED ONE HUNDRED..LATIN SMALL LETTER REVERSED C
    Interval{ 0x2185, 0x2188 }, // Nl   [4] ROMAN NUMERAL SIX LATE FORM..ROMAN NUMERAL ONE HUNDRED THOUSAND
    Interval{ 0x2C00, 0x2C7B }, // L& [124] GLAGOLITIC CAPITAL LETTER AZU..LATIN LETTER SMALL CAPITAL TURNED E
    Interval{ 0x2C7C, 0x2C7D }, // Lm   [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    Interval{ 0x2C7E, 0x2CE4 }, // L& [103] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC SYMBOL KAI
    Interval{ 0x2CEB, 0x2CEE }, // L&   [4] COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI..COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CEF, 0x2CF1 }, // Mn   [3] COPTIC COMBINING NI ABOVE..COPTIC COMBINING SPIRITUS LENIS
    Interval{ 0x2CF2, 0x2CF3 }, // L&   [2] COPTIC CAPITAL LETTER BOHAIRIC KHEI..COPTIC SMALL LETTER BOHAIRIC KHEI
    Interval{ 0x2D00, 0x2D25 }, // L&  [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Interval{ 0x2D27, 0x2D27 }, // L&       GEORGIAN SMALL LETTER YN
    Interval{ 0x2D2D, 0x2D2D }, // L&       GEORGIAN SMALL LETTER AEN
    Interval{ 0x2D30, 0x2D67 }, // Lo  [56] TIFINAGH LETTER YA..TIFINAGH LETTER YO
    Interval{ 0x2D6F, 0x2D6F }, // Lm       TIFINAGH MODIFIER LETTER LABIALIZATION MARK
    Interval{ 0x2D7F, 0x2D7F }, // Mn       TIFINAGH CONSONANT JOINER
    Interval{ 0x2D80, 0x2D96 }, // Lo  [23] ETHIOPIC SYLLABLE LOA..ETHIOPIC SYLLABLE GGWE
    Interval{ 0x2DA0, 0x2DA6 }, // Lo   [7] ETHIOPIC SYLLABLE SSA..ETHIOPIC SYLLABLE SSO
    Interval{ 0x2DA8, 0x2DAE }, // Lo   [7] ETHIOPIC SYLLABLE CCA..ETHIOPIC SYLLABLE CCO
    Interval{ 0x2DB0, 0x2DB6 }, // Lo   [7] ETHIOPIC SYLLABLE ZZA..ETHIOPIC SYLLABLE ZZO
    Interval{ 0x2DB8, 0x2DBE }, // Lo   [7] ETHIOPIC SYLLABLE CCHA..ETHIOPIC SYLLABLE CCHO
    Interval{ 0x2DC0, 0x2DC6 }, // Lo   [7] ETHIOPIC SYLLABLE QYA..ETHIOPIC SYLLABLE QYO
    Interval{ 0x2DC8, 0x2DCE }, // Lo   [7] ETHIOPIC SYLLABLE KYA..ETHIOPIC SYLLABLE KYO
    Interval{ 0x2DD0, 0x2DD6 }, // Lo   [7] ETHIOPIC SYLLABLE XYA..ETHIOPIC SYLLABLE XYO
    Interval{ 0x2DD8, 0x2DDE }, // Lo   [7] ETHIOPIC SYLLABLE GYA..ETHIOPIC SYLLABLE GYO
    Interval{ 0x2DE0, 0x2DFF }, // Mn  [32] COMBINING CYRILLIC LETTER BE..COMBINING CYRILLIC LETTER IOTIFIED BIG YUS
    Interval{ 0x3005, 0x3005 }, // Lm       IDEOGRAPHIC ITERATION MARK
    Interval{ 0x3006, 0x3006 }, // Lo       IDEOGRAPHIC CLOSING MARK
    Interval{ 0x3007, 0x3007 }, // Nl       IDEOGRAPHIC NUMBER ZERO
    Interval{ 0x3021, 0x3029 }, // Nl   [9] HANGZHOU NUMERAL ONE..HANGZHOU NUMERAL NINE
    Interval{ 0x302A, 0x302D }, // Mn   [4] IDEOGRAPHIC LEVEL TONE MARK..IDEOGRAPHIC ENTERING TONE MARK
    Interval{ 0x302E, 0x302F }, // Mc   [2] HANGUL SINGLE DOT TONE MARK..HANGUL DOUBLE DOT TONE MARK
    Interval{ 0x3031, 0x3035 }, // Lm   [5] VERTICAL KANA REPEAT MARK..VERTICAL KANA REPEAT MARK LOWER HALF
    Interval{ 0x3038, 0x303A }, // Nl   [3] HANGZHOU NUMERAL TEN..HANGZHOU NUMERAL THIRTY
    Interval{ 0x303B, 0x303B }, // Lm       VERTICAL IDEOGRAPHIC ITERATION MARK
    Interval{ 0x303C, 0x303C }, // Lo       MASU MARK
    Interval{ 0x3041, 0x3096 }, // Lo  [86] HIRAGANA LETTER SMALL A..HIRAGANA LETTER SMALL KE
    Interval{ 0x3099, 0x309A }, // Mn   [2] COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK..COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Interval{ 0x309D, 0x309E }, // Lm   [2] HIRAGANA ITERATION MARK..HIRAGANA VOICED ITERATION MARK
    Interval{ 0x309F, 0x309F }, // Lo       HIRAGANA DIGRAPH YORI
    Interval{ 0x30A1, 0x30FA }, // Lo  [90] KATAKANA LETTER SMALL A..KATAKANA LETTER VO
    Interval{ 0x30FB, 0x30FB }, // Po       KATAKANA MIDDLE DOT
    Interval{ 0x30FC, 0x30FE }, // Lm   [3] KATAKANA-HIRAGANA PROLONGED SOUND MARK..KATAKANA VOICED ITERATION MARK
    Interval{ 0x30FF, 0x30FF }, // Lo       KATAKANA DIGRAPH KOTO
    Interval{ 0x3105, 0x312F }, // Lo  [43] BOPOMOFO LETTER B..BOPOMOFO LETTER NN
    Interval{ 0x3131, 0x318E }, // Lo  [94] HANGUL LETTER KIYEOK..HANGUL LETTER ARAEAE
    Interval{ 0x31A0, 0x31BF }, // Lo  [32] BOPOMOFO LETTER BU..BOPOMOFO LETTER AH
    Interval{ 0x31F0, 0x31FF }, // Lo  [16] KATAKANA LETTER SMALL KU..KATAKANA LETTER SMALL RO
    Interval{ 0x3400, 0x4DBF }, // Lo [6592] CJK UNIFIED IDEOGRAPH-3400..CJK UNIFIED IDEOGRAPH-4DBF
    Interval{ 0x4E00, 0xA014 }, // Lo [21013] CJK UNIFIED IDEOGRAPH-4E00..YI SYLLABLE E
    Interval{ 0xA015, 0xA015 }, // Lm       YI SYLLABLE WU
    Interval{ 0xA016, 0xA48C }, // Lo [1143] YI SYLLABLE BIT..YI SYLLABLE YYR
    Interval{ 0xA4D0, 0xA4F7 }, // Lo  [40] LISU LETTER BA..LISU LETTER OE
    Interval{ 0xA4F8, 0xA4FD }, // Lm   [6] LISU LETTER TONE MYA TI..LISU LETTER TONE MYA JEU
    Interval{ 0xA500, 0xA60B }, // Lo [268] VAI SYLLABLE EE..VAI SYLLABLE NG
    Interval{ 0xA60C, 0xA60C }, // Lm       VAI SYLLABLE LENGTHENER
    Interval{ 0xA610, 0xA61F }, // Lo  [16] VAI SYLLABLE NDOLE FA..VAI SYMBOL JONG
    Interval{ 0xA620, 0xA629 }, // Nd  [10] VAI DIGIT ZERO..VAI DIGIT NINE
    Interval{ 0xA62A, 0xA62B }, // Lo   [2] VAI SYLLABLE NDOLE MA..VAI SYLLABLE NDOLE DO
    Interval{ 0xA640, 0xA66D }, // L&  [46] CYRILLIC CAPITAL LETTER ZEMLYA..CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA66E, 0xA66E }, // Lo       CYRILLIC LETTER MULTIOCULAR O
    Interval{ 0xA66F, 0xA66F }, // Mn       COMBINING CYRILLIC VZMET
    Interval{ 0xA674, 0xA67D }, // Mn  [10] COMBINING CYRILLIC LETTER UKRAINIAN IE..COMBINING CYRILLIC PAYEROK
    Interval{ 0xA67F, 0xA67F }, // Lm       CYRILLIC PAYEROK
    Interval{ 0xA680, 0xA69B }, // L&  [28] CYRILLIC CAPITAL LETTER DWE..CYRILLIC SMALL LETTER CROSSED O
    Interval{ 0xA69C, 0xA69D }, // Lm   [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    Interval{ 0xA69E, 0xA69F }, // Mn   [2] COMBINING CYRILLIC LETTER EF..COMBINING CYRILLIC LETTER IOTIFIED E
    Interval{ 0xA6A0, 0xA6E5 }, // Lo  [70] BAMUM LETTER A..BAMUM LETTER KI
    Interval{ 0xA6E6, 0xA6EF }, // Nl  [10] BAMUM LETTER MO..BAMUM LETTER KOGHOM
    Interval{ 0xA6F0, 0xA6F1 }, // Mn   [2] BAMUM COMBINING MARK KOQNDON..BAMUM COMBINING MARK TUKWENTIS
    Interval{ 0xA717, 0xA71F }, // Lm   [9] MODIFIER LETTER DOT VERTICAL BAR..MODIFIER LETTER LOW INVERTED EXCLAMATION MARK
    Interval{ 0xA722, 0xA76F }, // L&  [78] LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF..LATIN SMALL LETTER CON
    Interval{ 0xA770, 0xA770 }, // Lm       MODIFIER LETTER US
    Interval{ 0xA771, 0xA787 }, // L&  [23] LATIN SMALL LETTER DUM..LATIN SMALL LETTER INSULAR T
    Interval{ 0xA788, 0xA788 }, // Lm       MODIFIER LETTER LOW CIRCUMFLEX ACCENT
    Interval{ 0xA78B, 0xA78E }, // L&   [4] LATIN CAPITAL LETTER SALTILLO..LATIN SMALL LETTER L WITH RETROFLEX HOOK AND BELT
    Interval{ 0xA78F, 0xA78F }, // Lo       LATIN LETTER SINOLOGICAL DOT
    Interval{ 0xA790, 0xA7DC }, // L&  [77] LATIN CAPITAL LETTER N WITH DESCENDER..LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Interval{ 0xA7F1, 0xA7F4 }, // Lm   [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    Interval{ 0xA7F5, 0xA7F6 }, // L&   [2] LATIN CAPITAL LETTER REVERSED HALF H..LATIN SMALL LETTER REVERSED HALF H
    Interval{ 0xA7F7, 0xA7F7 }, // Lo       LATIN EPIGRAPHIC LETTER SIDEWAYS I
    Interval{ 0xA7F8, 0xA7F9 }, // Lm   [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    Interval{ 0xA7FA, 0xA7FA }, // L&       LATIN LETTER SMALL CAPITAL TURNED M
    Interval{ 0xA7FB, 0xA801 }, // Lo   [7] LATIN EPIGRAPHIC LETTER REVERSED F..SYLOTI NAGRI LETTER I
    Interval{ 0xA802, 0xA802 }, // Mn       SYLOTI NAGRI SIGN DVISVARA
    Interval{ 0xA803, 0xA805 }, // Lo   [3] SYLOTI NAGRI LETTER U..SYLOTI NAGRI LETTER O
    Interval{ 0xA806, 0xA806 }, // Mn       SYLOTI NAGRI SIGN HASANTA
    Interval{ 0xA807, 0xA80A }, // Lo   [4] SYLOTI NAGRI LETTER KO..SYLOTI NAGRI LETTER GHO
    Interval{ 0xA80B, 0xA80B }, // Mn       SYLOTI NAGRI SIGN ANUSVARA
    Interval{ 0xA80C, 0xA822 }, // Lo  [23] SYLOTI NAGRI LETTER CO..SYLOTI NAGRI LETTER HO
    Interval{ 0xA823, 0xA824 }, // Mc   [2] SYLOTI NAGRI VOWEL SIGN A..SYLOTI NAGRI VOWEL SIGN I
    Interval{ 0xA825, 0xA826 }, // Mn   [2] SYLOTI NAGRI VOWEL SIGN U..SYLOTI NAGRI VOWEL SIGN E
    Interval{ 0xA827, 0xA827 }, // Mc       SYLOTI NAGRI VOWEL SIGN OO
    Interval{ 0xA82C, 0xA82C }, // Mn       SYLOTI NAGRI SIGN ALTERNATE HASANTA
    Interval{ 0xA840, 0xA873 }, // Lo  [52] PHAGS-PA LETTER KA..PHAGS-PA LETTER CANDRABINDU
    Interval{ 0xA880, 0xA881 }, // Mc   [2] SAURASHTRA SIGN ANUSVARA..SAURASHTRA SIGN VISARGA
    Interval{ 0xA882, 0xA8B3 }, // Lo  [50] SAURASHTRA LETTER A..SAURASHTRA LETTER LLA
    Interval{ 0xA8B4, 0xA8C3 }, // Mc  [16] SAURASHTRA CONSONANT SIGN HAARU..SAURASHTRA VOWEL SIGN AU
    Interval{ 0xA8C4, 0xA8C5 }, // Mn   [2] SAURASHTRA SIGN VIRAMA..SAURASHTRA SIGN CANDRABINDU
    Interval{ 0xA8D0, 0xA8D9 }, // Nd  [10] SAURASHTRA DIGIT ZERO..SAURASHTRA DIGIT NINE
    Interval{ 0xA8E0, 0xA8F1 }, // Mn  [18] COMBINING DEVANAGARI DIGIT ZERO..COMBINING DEVANAGARI SIGN AVAGRAHA
    Interval{ 0xA8F2, 0xA8F7 }, // Lo   [6] DEVANAGARI SIGN SPACING CANDRABINDU..DEVANAGARI SIGN CANDRABINDU AVAGRAHA
    Interval{ 0xA8FB, 0xA8FB }, // Lo       DEVANAGARI HEADSTROKE
    Interval{ 0xA8FD, 0xA8FE }, // Lo   [2] DEVANAGARI JAIN OM..DEVANAGARI LETTER AY
    Interval{ 0xA8FF, 0xA8FF }, // Mn       DEVANAGARI VOWEL SIGN AY
    Interval{ 0xA900, 0xA909 }, // Nd  [10] KAYAH LI DIGIT ZERO..KAYAH LI DIGIT NINE
    Interval{ 0xA90A, 0xA925 }, // Lo  [28] KAYAH LI LETTER KA..KAYAH LI LETTER OO
    Interval{ 0xA926, 0xA92D }, // Mn   [8] KAYAH LI VOWEL UE..KAYAH LI TONE CALYA PLOPHU
    Interval{ 0xA930, 0xA946 }, // Lo  [23] REJANG LETTER KA..REJANG LETTER A
    Interval{ 0xA947, 0xA951 }, // Mn  [11] REJANG VOWEL SIGN I..REJANG CONSONANT SIGN R
    Interval{ 0xA952, 0xA953 }, // Mc   [2] REJANG CONSONANT SIGN H..REJANG VIRAMA
    Interval{ 0xA960, 0xA97C }, // Lo  [29] HANGUL CHOSEONG TIKEUT-MIEUM..HANGUL CHOSEONG SSANGYEORINHIEUH
    Interval{ 0xA980, 0xA982 }, // Mn   [3] JAVANESE SIGN PANYANGGA..JAVANESE SIGN LAYAR
    Interval{ 0xA983, 0xA983 }, // Mc       JAVANESE SIGN WIGNYAN
    Interval{ 0xA984, 0xA9B2 }, // Lo  [47] JAVANESE LETTER A..JAVANESE LETTER HA
    Interval{ 0xA9B3, 0xA9B3 }, // Mn       JAVANESE SIGN CECAK TELU
    Interval{ 0xA9B4, 0xA9B5 }, // Mc   [2] JAVANESE VOWEL SIGN TARUNG..JAVANESE VOWEL SIGN TOLONG
    Interval{ 0xA9B6, 0xA9B9 }, // Mn   [4] JAVANESE VOWEL SIGN WULU..JAVANESE VOWEL SIGN SUKU MENDUT
    Interval{ 0xA9BA, 0xA9BB }, // Mc   [2] JAVANESE VOWEL SIGN TALING..JAVANESE VOWEL SIGN DIRGA MURE
    Interval{ 0xA9BC, 0xA9BD }, // Mn   [2] JAVANESE VOWEL SIGN PEPET..JAVANESE CONSONANT SIGN KERET
    Interval{ 0xA9BE, 0xA9C0 }, // Mc   [3] JAVANESE CONSONANT SIGN PENGKAL..JAVANESE PANGKON
    Interval{ 0xA9CF, 0xA9CF }, // Lm       JAVANESE PANGRANGKEP
    Interval{ 0xA9D0, 0xA9D9 }, // Nd  [10] JAVANESE DIGIT ZERO..JAVANESE DIGIT NINE
    Interval{ 0xA9E0, 0xA9E4 }, // Lo   [5] MYANMAR LETTER SHAN GHA..MYANMAR LETTER SHAN BHA
    Interval{ 0xA9E5, 0xA9E5 }, // Mn       MYANMAR SIGN SHAN SAW
    Interval{ 0xA9E6, 0xA9E6 }, // Lm       MYANMAR MODIFIER LETTER SHAN REDUPLICATION
    Interval{ 0xA9E7, 0xA9EF }, // Lo   [9] MYANMAR LETTER TAI LAING NYA..MYANMAR LETTER TAI LAING NNA
    Interval{ 0xA9F0, 0xA9F9 }, // Nd  [10] MYANMAR TAI LAING DIGIT ZERO..MYANMAR TAI LAING DIGIT NINE
    Interval{ 0xA9FA, 0xA9FE }, // Lo   [5] MYANMAR LETTER TAI LAING LLA..MYANMAR LETTER TAI LAING BHA
    Interval{ 0xAA00, 0xAA28 }, // Lo  [41] CHAM LETTER A..CHAM LETTER HA
    Interval{ 0xAA29, 0xAA2E }, // Mn   [6] CHAM VOWEL SIGN AA..CHAM VOWEL SIGN OE
    Interval{ 0xAA2F, 0xAA30 }, // Mc   [2] CHAM VOWEL SIGN O..CHAM VOWEL SIGN AI
    Interval{ 0xAA31, 0xAA32 }, // Mn   [2] CHAM VOWEL SIGN AU..CHAM VOWEL SIGN UE
    Interval{ 0xAA33, 0xAA34 }, // Mc   [2] CHAM CONSONANT SIGN YA..CHAM CONSONANT SIGN RA
    Interval{ 0xAA35, 0xAA36 }, // Mn   [2] CHAM CONSONANT SIGN LA..CHAM CONSONANT SIGN WA
    Interval{ 0xAA40, 0xAA42 }, // Lo   [3] CHAM LETTER FINAL K..CHAM LETTER FINAL NG
    Interval{ 0xAA43, 0xAA43 }, // Mn       CHAM CONSONANT SIGN FINAL NG
    Interval{ 0xAA44, 0xAA4B }, // Lo   [8] CHAM LETTER FINAL CH..CHAM LETTER FINAL SS
    Interval{ 0xAA4C, 0xAA4C }, // Mn       CHAM CONSONANT SIGN FINAL M
    Interval{ 0xAA4D, 0xAA4D }, // Mc       CHAM CONSONANT SIGN FINAL H
    Interval{ 0xAA50, 0xAA59 }, // Nd  [10] CHAM DIGIT ZERO..CHAM DIGIT NINE
    Interval{ 0xAA60, 0xAA6F }, // Lo  [16] MYANMAR LETTER KHAMTI GA..MYANMAR LETTER KHAMTI FA
    Interval{ 0xAA70, 0xAA70 }, // Lm       MYANMAR MODIFIER LETTER KHAMTI REDUPLICATION
    Interval{ 0xAA71, 0xAA76 }, // Lo   [6] MYANMAR LETTER KHAMTI XA..MYANMAR LOGOGRAM KHAMTI HM
    Interval{ 0xAA7A, 0xAA7A }, // Lo       MYANMAR LETTER AITON RA
    Interval{ 0xAA7B, 0xAA7B }, // Mc       MYANMAR SIGN PAO KAREN TONE
    Interval{ 0xAA7C, 0xAA7C }, // Mn       MYANMAR SIGN TAI LAING TONE-2
    Interval{ 0xAA7D, 0xAA7D }, // Mc       MYANMAR SIGN TAI LAING TONE-5
    Interval{ 0xAA7E, 0xAAAF }, // Lo  [50] MYANMAR LETTER SHWE PALAUNG CHA..TAI VIET LETTER HIGH O
    Interval{ 0xAAB0, 0xAAB0 }, // Mn       TAI VIET MAI KANG
    Interval{ 0xAAB1, 0xAAB1 }, // Lo       TAI VIET VOWEL AA
    Interval{ 0xAAB2, 0xAAB4 }, // Mn   [3] TAI VIET VOWEL I..TAI VIET VOWEL U
    Interval{ 0xAAB5, 0xAAB6 }, // Lo   [2] TAI VIET VOWEL E..TAI VIET VOWEL O
    Interval{ 0xAAB7, 0xAAB8 }, // Mn   [2] TAI VIET MAI KHIT..TAI VIET VOWEL IA
    Interval{ 0xAAB9, 0xAABD }, // Lo   [5] TAI VIET VOWEL UEA..TAI VIET VOWEL AN
    Interval{ 0xAABE, 0xAABF }, // Mn   [2] TAI VIET VOWEL AM..TAI VIET TONE MAI EK
    Interval{ 0xAAC0, 0xAAC0 }, // Lo       TAI VIET TONE MAI NUENG
    Interval{ 0xAAC1, 0xAAC1 }, // Mn       TAI VIET TONE MAI THO
    Interval{ 0xAAC2, 0xAAC2 }, // Lo       TAI VIET TONE MAI SONG
    Interval{ 0xAADB, 0xAADC }, // Lo   [2] TAI VIET SYMBOL KON..TAI VIET SYMBOL NUENG
    Interval{ 0xAADD, 0xAADD }, // Lm       TAI VIET SYMBOL SAM
    Interval{ 0xAAE0, 0xAAEA }, // Lo  [11] MEETEI MAYEK LETTER E..MEETEI MAYEK LETTER SSA
    Interval{ 0xAAEB, 0xAAEB }, // Mc       MEETEI MAYEK VOWEL SIGN II
    Interval{ 0xAAEC, 0xAAED }, // Mn   [2] MEETEI MAYEK VOWEL SIGN UU..MEETEI MAYEK VOWEL SIGN AAI
    Interval{ 0xAAEE, 0xAAEF }, // Mc   [2] MEETEI MAYEK VOWEL SIGN AU..MEETEI MAYEK VOWEL SIGN AAU
    Interval{ 0xAAF2, 0xAAF2 }, // Lo       MEETEI MAYEK ANJI
    Interval{ 0xAAF3, 0xAAF4 }, // Lm   [2] MEETEI MAYEK SYLLABLE REPETITION MARK..MEETEI MAYEK WORD REPETITION MARK
    Interval{ 0xAAF5, 0xAAF5 }, // Mc       MEETEI MAYEK VOWEL SIGN VISARGA
    Interval{ 0xAAF6, 0xAAF6 }, // Mn       MEETEI MAYEK VIRAMA
    Interval{ 0xAB01, 0xAB06 }, // Lo   [6] ETHIOPIC SYLLABLE TTHU..ETHIOPIC SYLLABLE TTHO
    Interval{ 0xAB09, 0xAB0E }, // Lo   [6] ETHIOPIC SYLLABLE DDHU..ETHIOPIC SYLLABLE DDHO
    Interval{ 0xAB11, 0xAB16 }, // Lo   [6] ETHIOPIC SYLLABLE DZU..ETHIOPIC SYLLABLE DZO
    Interval{ 0xAB20, 0xAB26 }, // Lo   [7] ETHIOPIC SYLLABLE CCHHA..ETHIOPIC SYLLABLE CCHHO
    Interval{ 0xAB28, 0xAB2E }, // Lo   [7] ETHIOPIC SYLLABLE BBA..ETHIOPIC SYLLABLE BBO
    Interval{ 0xAB30, 0xAB5A }, // L&  [43] LATIN SMALL LETTER BARRED ALPHA..LATIN SMALL LETTER Y WITH SHORT RIGHT LEG
    Interval{ 0xAB5C, 0xAB5F }, // Lm   [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    Interval{ 0xAB60, 0xAB68 }, // L&   [9] LATIN SMALL LETTER SAKHA YAT..LATIN SMALL LETTER TURNED R WITH MIDDLE TILDE
    Interval{ 0xAB69, 0xAB69 }, // Lm       MODIFIER LETTER SMALL TURNED W
    Interval{ 0xAB70, 0xABBF }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xABC0, 0xABE2 }, // Lo  [35] MEETEI MAYEK LETTER KOK..MEETEI MAYEK LETTER I LONSUM
    Interval{ 0xABE3, 0xABE4 }, // Mc   [2] MEETEI MAYEK VOWEL SIGN ONAP..MEETEI MAYEK VOWEL SIGN INAP
    Interval{ 0xABE5, 0xABE5 }, // Mn       MEETEI MAYEK VOWEL SIGN ANAP
    Interval{ 0xABE6, 0xABE7 }, // Mc   [2] MEETEI MAYEK VOWEL SIGN YENAP..MEETEI MAYEK VOWEL SIGN SOUNAP
    Interval{ 0xABE8, 0xABE8 }, // Mn       MEETEI MAYEK VOWEL SIGN UNAP
    Interval{ 0xABE9, 0xABEA }, // Mc   [2] MEETEI MAYEK VOWEL SIGN CHEINAP..MEETEI MAYEK VOWEL SIGN NUNG
    Interval{ 0xABEC, 0xABEC }, // Mc       MEETEI MAYEK LUM IYEK
    Interval{ 0xABED, 0xABED }, // Mn       MEETEI MAYEK APUN IYEK
    Interval{ 0xABF0, 0xABF9 }, // Nd  [10] MEETEI MAYEK DIGIT ZERO..MEETEI MAYEK DIGIT NINE
    Interval{ 0xAC00, 0xD7A3 }, // Lo [11172] HANGUL SYLLABLE GA..HANGUL SYLLABLE HIH
    Interval{ 0xD7B0, 0xD7C6 }, // Lo  [23] HANGUL JUNGSEONG O-YEO..HANGUL JUNGSEONG ARAEA-E
    Interval{ 0xD7CB, 0xD7FB }, // Lo  [49] HANGUL JONGSEONG NIEUN-RIEUL..HANGUL JONGSEONG PHIEUPH-THIEUTH
    Interval{ 0xF900, 0xFA6D }, // Lo [366] CJK COMPATIBILITY IDEOGRAPH-F900..CJK COMPATIBILITY IDEOGRAPH-FA6D
    Interval{ 0xFA70, 0xFAD9 }, // Lo [106] CJK COMPATIBILITY IDEOGRAPH-FA70..CJK COMPATIBILITY IDEOGRAPH-FAD9
    Interval{ 0xFB00, 0xFB06 }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFB1D, 0xFB1D }, // Lo       HEBREW LETTER YOD WITH HIRIQ
    Interval{ 0xFB1E, 0xFB1E }, // Mn       HEBREW POINT JUDEO-SPANISH VARIKA
    Interval{ 0xFB1F, 0xFB28 }, // Lo  [10] HEBREW LIGATURE YIDDISH YOD YOD PATAH..HEBREW LETTER WIDE TAV
    Interval{ 0xFB2A, 0xFB36 }, // Lo  [13] HEBREW LETTER SHIN WITH SHIN DOT..HEBREW LETTER ZAYIN WITH DAGESH
    Interval{ 0xFB38, 0xFB3C }, // Lo   [5] HEBREW LETTER TET WITH DAGESH..HEBREW LETTER LAMED WITH DAGESH
    Interval{ 0xFB3E, 0xFB3E }, // Lo       HEBREW LETTER MEM WITH DAGESH
    Interval{ 0xFB40, 0xFB41 }, // Lo   [2] HEBREW LETTER NUN WITH DAGESH..HEBREW LETTER SAMEKH WITH DAGESH
    Interval{ 0xFB43, 0xFB44 }, // Lo   [2] HEBREW LETTER FINAL PE WITH DAGESH..HEBREW LETTER PE WITH DAGESH
    Interval{ 0xFB46, 0xFBB1 }, // Lo [108] HEBREW LETTER TSADI WITH DAGESH..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE FINAL FORM
    Interval{ 0xFBD3, 0xFC5D }, // Lo [139] ARABIC LETTER NG ISOLATED FORM..ARABIC LIGATURE ALEF MAKSURA WITH SUPERSCRIPT ALEF ISOLATED FORM
    Interval{ 0xFC64, 0xFD3D }, // Lo [218] ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH REH FINAL FORM..ARABIC LIGATURE ALEF WITH FATHATAN ISOLATED FORM
    Interval{ 0xFD50, 0xFD8F }, // Lo  [64] ARABIC LIGATURE TEH WITH JEEM WITH MEEM INITIAL FORM..ARABIC LIGATURE MEEM WITH KHAH WITH MEEM INITIAL FORM
    Interval{ 0xFD92, 0xFDC7 }, // Lo  [54] ARABIC LIGATURE MEEM WITH JEEM WITH KHAH INITIAL FORM..ARABIC LIGATURE NOON WITH JEEM WITH YEH FINAL FORM
    Interval{ 0xFDF0, 0xFDF9 }, // Lo  [10] ARABIC LIGATURE SALLA USED AS KORANIC STOP SIGN ISOLATED FORM..ARABIC LIGATURE SALLA ISOLATED FORM
    Interval{ 0xFE00, 0xFE0F }, // Mn  [16] VARIATION SELECTOR-1..VARIATION SELECTOR-16
    Interval{ 0xFE20, 0xFE2F }, // Mn  [16] COMBINING LIGATURE LEFT HALF..COMBINING CYRILLIC TITLO RIGHT HALF
    Interval{ 0xFE33, 0xFE34 }, // Pc   [2] PRESENTATION FORM FOR VERTICAL LOW LINE..PRESENTATION FORM FOR VERTICAL WAVY LOW LINE
    Interval{ 0xFE4D, 0xFE4F }, // Pc   [3] DASHED LOW LINE..WAVY LOW LINE
    Interval{ 0xFE71, 0xFE71 }, // Lo       ARABIC TATWEEL WITH FATHATAN ABOVE
    Interval{ 0xFE73, 0xFE73 }, // Lo       ARABIC TAIL FRAGMENT
    Interval{ 0xFE77, 0xFE77 }, // Lo       ARABIC FATHA MEDIAL FORM
    Interval{ 0xFE79, 0xFE79 }, // Lo       ARABIC DAMMA MEDIAL FORM
    Interval{ 0xFE7B, 0xFE7B }, // Lo       ARABIC KASRA MEDIAL FORM
    Interval{ 0xFE7D, 0xFE7D }, // Lo       ARABIC SHADDA MEDIAL FORM
    Interval{ 0xFE7F, 0xFEFC }, // Lo [126] ARABIC SUKUN MEDIAL FORM..ARABIC LIGATURE LAM WITH ALEF FINAL FORM
    Interval{ 0xFF10, 0xFF19 }, // Nd  [10] FULLWIDTH DIGIT ZERO..FULLWIDTH DIGIT NINE
    Interval{ 0xFF21, 0xFF3A }, // L&  [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Interval{ 0xFF3F, 0xFF3F }, // Pc       FULLWIDTH LOW LINE
    Interval{ 0xFF41, 0xFF5A }, // L&  [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Interval{ 0xFF65, 0xFF65 }, // Po       HALFWIDTH KATAKANA MIDDLE DOT
    Interval{ 0xFF66, 0xFF6F }, // Lo  [10] HALFWIDTH KATAKANA LETTER WO..HALFWIDTH KATAKANA LETTER SMALL TU
    Interval{ 0xFF70, 0xFF70 }, // Lm       HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Interval{ 0xFF71, 0xFF9D }, // Lo  [45] HALFWIDTH KATAKANA LETTER A..HALFWIDTH KATAKANA LETTER N
    Interval{ 0xFF9E, 0xFF9F }, // Lm   [2] HALFWIDTH KATAKANA VOICED SOUND MARK..HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK
    Interval{ 0xFFA0, 0xFFBE }, // Lo  [31] HALFWIDTH HANGUL FILLER..HALFWIDTH HANGUL LETTER HIEUH
    Interval{ 0xFFC2, 0xFFC7 }, // Lo   [6] HALFWIDTH HANGUL LETTER A..HALFWIDTH HANGUL LETTER E
    Interval{ 0xFFCA, 0xFFCF }, // Lo   [6] HALFWIDTH HANGUL LETTER YEO..HALFWIDTH HANGUL LETTER OE
    Interval{ 0xFFD2, 0xFFD7 }, // Lo   [6] HALFWIDTH HANGUL LETTER YO..HALFWIDTH HANGUL LETTER YU
    Interval{ 0xFFDA, 0xFFDC }, // Lo   [3] HALFWIDTH HANGUL LETTER EU..HALFWIDTH HANGUL LETTER I
    Interval{ 0x10000, 0x1000B }, // Lo  [12] LINEAR B SYLLABLE B008 A..LINEAR B SYLLABLE B046 JE
    Interval{ 0x1000D, 0x10026 }, // Lo  [26] LINEAR B SYLLABLE B036 JO..LINEAR B SYLLABLE B032 QO
    Interval{ 0x10028, 0x1003A }, // Lo  [19] LINEAR B SYLLABLE B060 RA..LINEAR B SYLLABLE B042 WO
    Interval{ 0x1003C, 0x1003D }, // Lo   [2] LINEAR B SYLLABLE B017 ZA..LINEAR B SYLLABLE B074 ZE
    Interval{ 0x1003F, 0x1004D }, // Lo  [15] LINEAR B SYLLABLE B020 ZO..LINEAR B SYLLABLE B091 TWO
    Interval{ 0x10050, 0x1005D }, // Lo  [14] LINEAR B SYMBOL B018..LINEAR B SYMBOL B089
    Interval{ 0x10080, 0x100FA }, // Lo [123] LINEAR B IDEOGRAM B100 MAN..LINEAR B IDEOGRAM VESSEL B305
    Interval{ 0x10140, 0x10174 }, // Nl  [53] GREEK ACROPHONIC ATTIC ONE QUARTER..GREEK ACROPHONIC STRATIAN FIFTY MNAS
    Interval{ 0x101FD, 0x101FD }, // Mn       PHAISTOS DISC SIGN COMBINING OBLIQUE STROKE
    Interval{ 0x10280, 0x1029C }, // Lo  [29] LYCIAN LETTER A..LYCIAN LETTER X
    Interval{ 0x102A0, 0x102D0 }, // Lo  [49] CARIAN LETTER A..CARIAN LETTER UUU3
    Interval{ 0x102E0, 0x102E0 }, // Mn       COPTIC EPACT THOUSANDS MARK
    Interval{ 0x10300, 0x1031F }, // Lo  [32] OLD ITALIC LETTER A..OLD ITALIC LETTER ESS
    Interval{ 0x1032D, 0x10340 }, // Lo  [20] OLD ITALIC LETTER YE..GOTHIC LETTER PAIRTHRA
    Interval{ 0x10341, 0x10341 }, // Nl       GOTHIC LETTER NINETY
    Interval{ 0x10342, 0x10349 }, // Lo   [8] GOTHIC LETTER RAIDA..GOTHIC LETTER OTHAL
    Interval{ 0x1034A, 0x1034A }, // Nl       GOTHIC LETTER NINE HUNDRED
    Interval{ 0x10350, 0x10375 }, // Lo  [38] OLD PERMIC LETTER AN..OLD PERMIC LETTER IA
    Interval{ 0x10376, 0x1037A }, // Mn   [5] COMBINING OLD PERMIC LETTER AN..COMBINING OLD PERMIC LETTER SII
    Interval{ 0x10380, 0x1039D }, // Lo  [30] UGARITIC LETTER ALPA..UGARITIC LETTER SSU
    Interval{ 0x103A0, 0x103C3 }, // Lo  [36] OLD PERSIAN SIGN A..OLD PERSIAN SIGN HA
    Interval{ 0x103C8, 0x103CF }, // Lo   [8] OLD PERSIAN SIGN AURAMAZDAA..OLD PERSIAN SIGN BUUMISH
    Interval{ 0x103D1, 0x103D5 }, // Nl   [5] OLD PERSIAN NUMBER ONE..OLD PERSIAN NUMBER HUNDRED
    Interval{ 0x10400, 0x1044F }, // L&  [80] DESERET CAPITAL LETTER LONG I..DESERET SMALL LETTER EW
    Interval{ 0x10450, 0x1049D }, // Lo  [78] SHAVIAN LETTER PEEP..OSMANYA LETTER OO
    Interval{ 0x104A0, 0x104A9 }, // Nd  [10] OSMANYA DIGIT ZERO..OSMANYA DIGIT NINE
    Interval{ 0x104B0, 0x104D3 }, // L&  [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Interval{ 0x104D8, 0x104FB }, // L&  [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Interval{ 0x10500, 0x10527 }, // Lo  [40] ELBASAN LETTER A..ELBASAN LETTER KHE
    Interval{ 0x10530, 0x10563 }, // Lo  [52] CAUCASIAN ALBANIAN LETTER ALT..CAUCASIAN ALBANIAN LETTER KIW
    Interval{ 0x10570, 0x1057A }, // L&  [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Interval{ 0x1057C, 0x1058A }, // L&  [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Interval{ 0x1058C, 0x10592 }, // L&   [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Interval{ 0x10594, 0x10595 }, // L&   [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Interval{ 0x10597, 0x105A1 }, // L&  [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Interval{ 0x105A3, 0x105B1 }, // L&  [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Interval{ 0x105B3, 0x105B9 }, // L&   [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Interval{ 0x105BB, 0x105BC }, // L&   [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Interval{ 0x105C0, 0x105F3 }, // Lo  [52] TODHRI LETTER A..TODHRI LETTER OO
    Interval{ 0x10600, 0x10736 }, // Lo [311] LINEAR A SIGN AB001..LINEAR A SIGN A664
    Interval{ 0x10740, 0x10755 }, // Lo  [22] LINEAR A SIGN A701 A..LINEAR A SIGN A732 JE
    Interval{ 0x10760, 0x10767 }, // Lo   [8] LINEAR A SIGN A800..LINEAR A SIGN A807
    Interval{ 0x10780, 0x10785 }, // Lm   [6] MODIFIER LETTER SMALL CAPITAL AA..MODIFIER LETTER SMALL B WITH HOOK
    Interval{ 0x10787, 0x107B0 }, // Lm  [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Interval{ 0x107B2, 0x107BA }, // Lm   [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Interval{ 0x10800, 0x10805 }, // Lo   [6] CYPRIOT SYLLABLE A..CYPRIOT SYLLABLE JA
    Interval{ 0x10808, 0x10808 }, // Lo       CYPRIOT SYLLABLE JO
    Interval{ 0x1080A, 0x10835 }, // Lo  [44] CYPRIOT SYLLABLE KA..CYPRIOT SYLLABLE WO
    Interval{ 0x10837, 0x10838 }, // Lo   [2] CYPRIOT SYLLABLE XA..CYPRIOT SYLLABLE XE
    Interval{ 0x1083C, 0x1083C }, // Lo       CYPRIOT SYLLABLE ZA
    Interval{ 0x1083F, 0x10855 }, // Lo  [23] CYPRIOT SYLLABLE ZO..IMPERIAL ARAMAIC LETTER TAW
    Interval{ 0x10860, 0x10876 }, // Lo  [23] PALMYRENE LETTER ALEPH..PALMYRENE LETTER TAW
    Interval{ 0x10880, 0x1089E }, // Lo  [31] NABATAEAN LETTER FINAL ALEPH..NABATAEAN LETTER TAW
    Interval{ 0x108E0, 0x108F2 }, // Lo  [19] HATRAN LETTER ALEPH..HATRAN LETTER QOPH
    Interval{ 0x108F4, 0x108F5 }, // Lo   [2] HATRAN LETTER SHIN..HATRAN LETTER TAW
    Interval{ 0x10900, 0x10915 }, // Lo  [22] PHOENICIAN LETTER ALF..PHOENICIAN LETTER TAU
    Interval{ 0x10920, 0x10939 }, // Lo  [26] LYDIAN LETTER A..LYDIAN LETTER C
    Interval{ 0x10940, 0x10959 }, // Lo  [26] SIDETIC LETTER N01..SIDETIC LETTER N26
    Interval{ 0x10980, 0x109B7 }, // Lo  [56] MEROITIC HIEROGLYPHIC LETTER A..MEROITIC CURSIVE LETTER DA
    Interval{ 0x109BE, 0x109BF }, // Lo   [2] MEROITIC CURSIVE LOGOGRAM RMT..MEROITIC CURSIVE LOGOGRAM IMN
    Interval{ 0x10A00, 0x10A00 }, // Lo       KHAROSHTHI LETTER A
    Interval{ 0x10A01, 0x10A03 }, // Mn   [3] KHAROSHTHI VOWEL SIGN I..KHAROSHTHI VOWEL SIGN VOCALIC R
    Interval{ 0x10A05, 0x10A06 }, // Mn   [2] KHAROSHTHI VOWEL SIGN E..KHAROSHTHI VOWEL SIGN O
    Interval{ 0x10A0C, 0x10A0F }, // Mn   [4] KHAROSHTHI VOWEL LENGTH MARK..KHAROSHTHI SIGN VISARGA
    Interval{ 0x10A10, 0x10A13 }, // Lo   [4] KHAROSHTHI LETTER KA..KHAROSHTHI LETTER GHA
    Interval{ 0x10A15, 0x10A17 }, // Lo   [3] KHAROSHTHI LETTER CA..KHAROSHTHI LETTER JA
    Interval{ 0x10A19, 0x10A35 }, // Lo  [29] KHAROSHTHI LETTER NYA..KHAROSHTHI LETTER VHA
    Interval{ 0x10A38, 0x10A3A }, // Mn   [3] KHAROSHTHI SIGN BAR ABOVE..KHAROSHTHI SIGN DOT BELOW
    Interval{ 0x10A3F, 0x10A3F }, // Mn       KHAROSHTHI VIRAMA
    Interval{ 0x10A60, 0x10A7C }, // Lo  [29] OLD SOUTH ARABIAN LETTER HE..OLD SOUTH ARABIAN LETTER THETH
    Interval{ 0x10A80, 0x10A9C }, // Lo  [29] OLD NORTH ARABIAN LETTER HEH..OLD NORTH ARABIAN LETTER ZAH
    Interval{ 0x10AC0, 0x10AC7 }, // Lo   [8] MANICHAEAN LETTER ALEPH..MANICHAEAN LETTER WAW
    Interval{ 0x10AC9, 0x10AE4 }, // Lo  [28] MANICHAEAN LETTER ZAYIN..MANICHAEAN LETTER TAW
    Interval{ 0x10AE5, 0x10AE6 }, // Mn   [2] MANICHAEAN ABBREVIATION MARK ABOVE..MANICHAEAN ABBREVIATION MARK BELOW
    Interval{ 0x10B00, 0x10B35 }, // Lo  [54] AVESTAN LETTER A..AVESTAN LETTER HE
    Interval{ 0x10B40, 0x10B55 }, // Lo  [22] INSCRIPTIONAL PARTHIAN LETTER ALEPH..INSCRIPTIONAL PARTHIAN LETTER TAW
    Interval{ 0x10B60, 0x10B72 }, // Lo  [19] INSCRIPTIONAL PAHLAVI LETTER ALEPH..INSCRIPTIONAL PAHLAVI LETTER TAW
    Interval{ 0x10B80, 0x10B91 }, // Lo  [18] PSALTER PAHLAVI LETTER ALEPH..PSALTER PAHLAVI LETTER TAW
    Interval{ 0x10C00, 0x10C48 }, // Lo  [73] OLD TURKIC LETTER ORKHON A..OLD TURKIC LETTER ORKHON BASH
    Interval{ 0x10C80, 0x10CB2 }, // L&  [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Interval{ 0x10CC0, 0x10CF2 }, // L&  [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Interval{ 0x10D00, 0x10D23 }, // Lo  [36] HANIFI ROHINGYA LETTER A..HANIFI ROHINGYA MARK NA KHONNA
    Interval{ 0x10D24, 0x10D27 }, // Mn   [4] HANIFI ROHINGYA SIGN HARBAHAY..HANIFI ROHINGYA SIGN TASSI
    Interval{ 0x10D30, 0x10D39 }, // Nd  [10] HANIFI ROHINGYA DIGIT ZERO..HANIFI ROHINGYA DIGIT NINE
    Interval{ 0x10D40, 0x10D49 }, // Nd  [10] GARAY DIGIT ZERO..GARAY DIGIT NINE
    Interval{ 0x10D4A, 0x10D4D }, // Lo   [4] GARAY VOWEL SIGN A..GARAY VOWEL SIGN EE
    Interval{ 0x10D4E, 0x10D4E }, // Lm       GARAY VOWEL LENGTH MARK
    Interval{ 0x10D4F, 0x10D4F }, // Lo       GARAY SUKUN
    Interval{ 0x10D50, 0x10D65 }, // L&  [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Interval{ 0x10D69, 0x10D6D }, // Mn   [5] GARAY VOWEL SIGN E..GARAY CONSONANT NASALIZATION MARK
    Interval{ 0x10D6F, 0x10D6F }, // Lm       GARAY REDUPLICATION MARK
    Interval{ 0x10D70, 0x10D85 }, // L&  [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Interval{ 0x10E80, 0x10EA9 }, // Lo  [42] YEZIDI LETTER ELIF..YEZIDI LETTER ET
    Interval{ 0x10EAB, 0x10EAC }, // Mn   [2] YEZIDI COMBINING HAMZA MARK..YEZIDI COMBINING MADDA MARK
    Interval{ 0x10EB0, 0x10EB1 }, // Lo   [2] YEZIDI LETTER LAM WITH DOT ABOVE..YEZIDI LETTER YOT WITH CIRCUMFLEX ABOVE
    Interval{ 0x10EC2, 0x10EC4 }, // Lo   [3] ARABIC LETTER DAL WITH TWO DOTS VERTICALLY BELOW..ARABIC LETTER KAF WITH TWO DOTS VERTICALLY BELOW
    Interval{ 0x10EC5, 0x10EC5 }, // Lm       ARABIC SMALL YEH BARREE WITH TWO DOTS BELOW
    Interval{ 0x10EC6, 0x10EC7 }, // Lo   [2] ARABIC LETTER THIN NOON..ARABIC LETTER YEH WITH FOUR DOTS BELOW
    Interval{ 0x10EFA, 0x10EFF }, // Mn   [6] ARABIC DOUBLE VERTICAL BAR BELOW..ARABIC SMALL LOW WORD MADDA
    Interval{ 0x10F00, 0x10F1C }, // Lo  [29] OLD SOGDIAN LETTER ALEPH..OLD SOGDIAN LETTER FINAL TAW WITH VERTICAL TAIL
    Interval{ 0x10F27, 0x10F27 }, // Lo       OLD SOGDIAN LIGATURE AYIN-DALETH
    Interval{ 0x10F30, 0x10F45 }, // Lo  [22] SOGDIAN LETTER ALEPH..SOGDIAN INDEPENDENT SHIN
    Interval{ 0x10F46, 0x10F50 }, // Mn  [11] SOGDIAN COMBINING DOT BELOW..SOGDIAN COMBINING STROKE BELOW
    Interval{ 0x10F70, 0x10F81 }, // Lo  [18] OLD UYGHUR LETTER ALEPH..OLD UYGHUR LETTER LESH
    Interval{ 0x10F82, 0x10F85 }, // Mn   [4] OLD UYGHUR COMBINING DOT ABOVE..OLD UYGHUR COMBINING TWO DOTS BELOW
    Interval{ 0x10FB0, 0x10FC4 }, // Lo  [21] CHORASMIAN LETTER ALEPH..CHORASMIAN LETTER TAW
    Interval{ 0x10FE0, 0x10FF6 }, // Lo  [23] ELYMAIC LETTER ALEPH..ELYMAIC LIGATURE ZAYIN-YODH
    Interval{ 0x11000, 0x11000 }, // Mc       BRAHMI SIGN CANDRABINDU
    Interval{ 0x11001, 0x11001 }, // Mn       BRAHMI SIGN ANUSVARA
    Interval{ 0x11002, 0x11002 }, // Mc       BRAHMI SIGN VISARGA
    Interval{ 0x11003, 0x11037 }, // Lo  [53] BRAHMI SIGN JIHVAMULIYA..BRAHMI LETTER OLD TAMIL NNNA
    Interval{ 0x11038, 0x11046 }, // Mn  [15] BRAHMI VOWEL SIGN AA..BRAHMI VIRAMA
    Interval{ 0x11066, 0x1106F }, // Nd  [10] BRAHMI DIGIT ZERO..BRAHMI DIGIT NINE
    Interval{ 0x11070, 0x11070 }, // Mn       BRAHMI SIGN OLD TAMIL VIRAMA
    Interval{ 0x11071, 0x11072 }, // Lo   [2] BRAHMI LETTER OLD TAMIL SHORT E..BRAHMI LETTER OLD TAMIL SHORT O
    Interval{ 0x11073, 0x11074 }, // Mn   [2] BRAHMI VOWEL SIGN OLD TAMIL SHORT E..BRAHMI VOWEL SIGN OLD TAMIL SHORT O
    Interval{ 0x11075, 0x11075 }, // Lo       BRAHMI LETTER OLD TAMIL LLA
    Interval{ 0x1107F, 0x11081 }, // Mn   [3] BRAHMI NUMBER JOINER..KAITHI SIGN ANUSVARA
    Interval{ 0x11082, 0x11082 }, // Mc       KAITHI SIGN VISARGA
    Interval{ 0x11083, 0x110AF }, // Lo  [45] KAITHI LETTER A..KAITHI LETTER HA
    Interval{ 0x110B0, 0x110B2 }, // Mc   [3] KAITHI VOWEL SIGN AA..KAITHI VOWEL SIGN II
    Interval{ 0x110B3, 0x110B6 }, // Mn   [4] KAITHI VOWEL SIGN U..KAITHI VOWEL SIGN AI
    Interval{ 0x110B7, 0x110B8 }, // Mc   [2] KAITHI VOWEL SIGN O..KAITHI VOWEL SIGN AU
    Interval{ 0x110B9, 0x110BA }, // Mn   [2] KAITHI SIGN VIRAMA..KAITHI SIGN NUKTA
    Interval{ 0x110C2, 0x110C2 }, // Mn       KAITHI VOWEL SIGN VOCALIC R
    Interval{ 0x110D0, 0x110E8 }, // Lo  [25] SORA SOMPENG LETTER SAH..SORA SOMPENG LETTER MAE
    Interval{ 0x110F0, 0x110F9 }, // Nd  [10] SORA SOMPENG DIGIT ZERO..SORA SOMPENG DIGIT NINE
    Interval{ 0x11100, 0x11102 }, // Mn   [3] CHAKMA SIGN CANDRABINDU..CHAKMA SIGN VISARGA
    Interval{ 0x11103, 0x11126 }, // Lo  [36] CHAKMA LETTER AA..CHAKMA LETTER HAA
    Interval{ 0x11127, 0x1112B }, // Mn   [5] CHAKMA VOWEL SIGN A..CHAKMA VOWEL SIGN UU
    Interval{ 0x1112C, 0x1112C }, // Mc       CHAKMA VOWEL SIGN E
    Interval{ 0x1112D, 0x11134 }, // Mn   [8] CHAKMA VOWEL SIGN AI..CHAKMA MAAYYAA
    Interval{ 0x11136, 0x1113F }, // Nd  [10] CHAKMA DIGIT ZERO..CHAKMA DIGIT NINE
    Interval{ 0x11144, 0x11144 }, // Lo       CHAKMA LETTER LHAA
    Interval{ 0x11145, 0x11146 }, // Mc   [2] CHAKMA VOWEL SIGN AA..CHAKMA VOWEL SIGN EI
    Interval{ 0x11147, 0x11147 }, // Lo       CHAKMA LETTER VAA
    Interval{ 0x11150, 0x11172 }, // Lo  [35] MAHAJANI LETTER A..MAHAJANI LETTER RRA
    Interval{ 0x11173, 0x11173 }, // Mn       MAHAJANI SIGN NUKTA
    Interval{ 0x11176, 0x11176 }, // Lo       MAHAJANI LIGATURE SHRI
    Interval{ 0x11180, 0x11181 }, // Mn   [2] SHARADA SIGN CANDRABINDU..SHARADA SIGN ANUSVARA
    Interval{ 0x11182, 0x11182 }, // Mc       SHARADA SIGN VISARGA
    Interval{ 0x11183, 0x111B2 }, // Lo  [48] SHARADA LETTER A..SHARADA LETTER HA
    Interval{ 0x111B3, 0x111B5 }, // Mc   [3] SHARADA VOWEL SIGN AA..SHARADA VOWEL SIGN II
    Interval{ 0x111B6, 0x111BE }, // Mn   [9] SHARADA VOWEL SIGN U..SHARADA VOWEL SIGN O
    Interval{ 0x111BF, 0x111C0 }, // Mc   [2] SHARADA VOWEL SIGN AU..SHARADA SIGN VIRAMA
    Interval{ 0x111C1, 0x111C4 }, // Lo   [4] SHARADA SIGN AVAGRAHA..SHARADA OM
    Interval{ 0x111C9, 0x111CC }, // Mn   [4] SHARADA SANDHI MARK..SHARADA EXTRA SHORT VOWEL MARK
    Interval{ 0x111CE, 0x111CE }, // Mc       SHARADA VOWEL SIGN PRISHTHAMATRA E
    Interval{ 0x111CF, 0x111CF }, // Mn       SHARADA SIGN INVERTED CANDRABINDU
    Interval{ 0x111D0, 0x111D9 }, // Nd  [10] SHARADA DIGIT ZERO..SHARADA DIGIT NINE
    Interval{ 0x111DA, 0x111DA }, // Lo       SHARADA EKAM
    Interval{ 0x111DC, 0x111DC }, // Lo       SHARADA HEADSTROKE
    Interval{ 0x11200, 0x11211 }, // Lo  [18] KHOJKI LETTER A..KHOJKI LETTER JJA
    Interval{ 0x11213, 0x1122B }, // Lo  [25] KHOJKI LETTER NYA..KHOJKI LETTER LLA
    Interval{ 0x1122C, 0x1122E }, // Mc   [3] KHOJKI VOWEL SIGN AA..KHOJKI VOWEL SIGN II
    Interval{ 0x1122F, 0x11231 }, // Mn   [3] KHOJKI VOWEL SIGN U..KHOJKI VOWEL SIGN AI
    Interval{ 0x11232, 0x11233 }, // Mc   [2] KHOJKI VOWEL SIGN O..KHOJKI VOWEL SIGN AU
    Interval{ 0x11234, 0x11234 }, // Mn       KHOJKI SIGN ANUSVARA
    Interval{ 0x11235, 0x11235 }, // Mc       KHOJKI SIGN VIRAMA
    Interval{ 0x11236, 0x11237 }, // Mn   [2] KHOJKI SIGN NUKTA..KHOJKI SIGN SHADDA
    Interval{ 0x1123E, 0x1123E }, // Mn       KHOJKI SIGN SUKUN
    Interval{ 0x1123F, 0x11240 }, // Lo   [2] KHOJKI LETTER QA..KHOJKI LETTER SHORT I
    Interval{ 0x11241, 0x11241 }, // Mn       KHOJKI VOWEL SIGN VOCALIC R
    Interval{ 0x11280, 0x11286 }, // Lo   [7] MULTANI LETTER A..MULTANI LETTER GA
    Interval{ 0x11288, 0x11288 }, // Lo       MULTANI LETTER GHA
    Interval{ 0x1128A, 0x1128D }, // Lo   [4] MULTANI LETTER CA..MULTANI LETTER JJA
    Interval{ 0x1128F, 0x1129D }, // Lo  [15] MULTANI LETTER NYA..MULTANI LETTER BA
    Interval{ 0x1129F, 0x112A8 }, // Lo  [10] MULTANI LETTER BHA..MULTANI LETTER RHA
    Interval{ 0x112B0, 0x112DE }, // Lo  [47] KHUDAWADI LETTER A..KHUDAWADI LETTER HA
    Interval{ 0x112DF, 0x112DF }, // Mn       KHUDAWADI SIGN ANUSVARA
    Interval{ 0x112E0, 0x112E2 }, // Mc   [3] KHUDAWADI VOWEL SIGN AA..KHUDAWADI VOWEL SIGN II
    Interval{ 0x112E3, 0x112EA }, // Mn   [8] KHUDAWADI VOWEL SIGN U..KHUDAWADI SIGN VIRAMA
    Interval{ 0x112F0, 0x112F9 }, // Nd  [10] KHUDAWADI DIGIT ZERO..KHUDAWADI DIGIT NINE
    Interval{ 0x11300, 0x11301 }, // Mn   [2] GRANTHA SIGN COMBINING ANUSVARA ABOVE..GRANTHA SIGN CANDRABINDU
    Interval{ 0x11302, 0x11303 }, // Mc   [2] GRANTHA SIGN ANUSVARA..GRANTHA SIGN VISARGA
    Interval{ 0x11305, 0x1130C }, // Lo   [8] GRANTHA LETTER A..GRANTHA LETTER VOCALIC L
    Interval{ 0x1130F, 0x11310 }, // Lo   [2] GRANTHA LETTER EE..GRANTHA LETTER AI
    Interval{ 0x11313, 0x11328 }, // Lo  [22] GRANTHA LETTER OO..GRANTHA LETTER NA
    Interval{ 0x1132A, 0x11330 }, // Lo   [7] GRANTHA LETTER PA..GRANTHA LETTER RA
    Interval{ 0x11332, 0x11333 }, // Lo   [2] GRANTHA LETTER LA..GRANTHA LETTER LLA
    Interval{ 0x11335, 0x11339 }, // Lo   [5] GRANTHA LETTER VA..GRANTHA LETTER HA
    Interval{ 0x1133B, 0x1133C }, // Mn   [2] COMBINING BINDU BELOW..GRANTHA SIGN NUKTA
    Interval{ 0x1133D, 0x1133D }, // Lo       GRANTHA SIGN AVAGRAHA
    Interval{ 0x1133E, 0x1133F }, // Mc   [2] GRANTHA VOWEL SIGN AA..GRANTHA VOWEL SIGN I
    Interval{ 0x11340, 0x11340 }, // Mn       GRANTHA VOWEL SIGN II
    Interval{ 0x11341, 0x11344 }, // Mc   [4] GRANTHA VOWEL SIGN U..GRANTHA VOWEL SIGN VOCALIC RR
    Interval{ 0x11347, 0x11348 }, // Mc   [2] GRANTHA VOWEL SIGN EE..GRANTHA VOWEL SIGN AI
    Interval{ 0x1134B, 0x1134D }, // Mc   [3] GRANTHA VOWEL SIGN OO..GRANTHA SIGN VIRAMA
    Interval{ 0x11350, 0x11350 }, // Lo       GRANTHA OM
    Interval{ 0x11357, 0x11357 }, // Mc       GRANTHA AU LENGTH MARK
    Interval{ 0x1135D, 0x11361 }, // Lo   [5] GRANTHA SIGN PLUTA..GRANTHA LETTER VOCALIC LL
    Interval{ 0x11362, 0x11363 }, // Mc   [2] GRANTHA VOWEL SIGN VOCALIC L..GRANTHA VOWEL SIGN VOCALIC LL
    Interval{ 0x11366, 0x1136C }, // Mn   [7] COMBINING GRANTHA DIGIT ZERO..COMBINING GRANTHA DIGIT SIX
    Interval{ 0x11370, 0x11374 }, // Mn   [5] COMBINING GRANTHA LETTER A..COMBINING GRANTHA LETTER PA
    Interval{ 0x11380, 0x11389 }, // Lo  [10] TULU-TIGALARI LETTER A..TULU-TIGALARI LETTER VOCALIC LL
    Interval{ 0x1138B, 0x1138B }, // Lo       TULU-TIGALARI LETTER EE
    Interval{ 0x1138E, 0x1138E }, // Lo       TULU-TIGALARI LETTER AI
    Interval{ 0x11390, 0x113B5 }, // Lo  [38] TULU-TIGALARI LETTER OO..TULU-TIGALARI LETTER LLLA
    Interval{ 0x113B7, 0x113B7 }, // Lo       TULU-TIGALARI SIGN AVAGRAHA
    Interval{ 0x113B8, 0x113BA }, // Mc   [3] TULU-TIGALARI VOWEL SIGN AA..TULU-TIGALARI VOWEL SIGN II
    Interval{ 0x113BB, 0x113C0 }, // Mn   [6] TULU-TIGALARI VOWEL SIGN U..TULU-TIGALARI VOWEL SIGN VOCALIC LL
    Interval{ 0x113C2, 0x113C2 }, // Mc       TULU-TIGALARI VOWEL SIGN EE
    Interval{ 0x113C5, 0x113C5 }, // Mc       TULU-TIGALARI VOWEL SIGN AI
    Interval{ 0x113C7, 0x113CA }, // Mc   [4] TULU-TIGALARI VOWEL SIGN OO..TULU-TIGALARI SIGN CANDRA ANUNASIKA
    Interval{ 0x113CC, 0x113CD }, // Mc   [2] TULU-TIGALARI SIGN ANUSVARA..TULU-TIGALARI SIGN VISARGA
    Interval{ 0x113CE, 0x113CE }, // Mn       TULU-TIGALARI SIGN VIRAMA
    Interval{ 0x113CF, 0x113CF }, // Mc       TULU-TIGALARI SIGN LOOPED VIRAMA
    Interval{ 0x113D0, 0x113D0 }, // Mn       TULU-TIGALARI CONJOINER
    Interval{ 0x113D1, 0x113D1 }, // Lo       TULU-TIGALARI REPHA
    Interval{ 0x113D2, 0x113D2 }, // Mn       TULU-TIGALARI GEMINATION MARK
    Interval{ 0x113D3, 0x113D3 }, // Lo       TULU-TIGALARI SIGN PLUTA
    Interval{ 0x113E1, 0x113E2 }, // Mn   [2] TULU-TIGALARI VEDIC TONE SVARITA..TULU-TIGALARI VEDIC TONE ANUDATTA
    Interval{ 0x11400, 0x11434 }, // Lo  [53] NEWA LETTER A..NEWA LETTER HA
    Interval{ 0x11435, 0x11437 }, // Mc   [3] NEWA VOWEL SIGN AA..NEWA VOWEL SIGN II
    Interval{ 0x11438, 0x1143F }, // Mn   [8] NEWA VOWEL SIGN U..NEWA VOWEL SIGN AI
    Interval{ 0x11440, 0x11441 }, // Mc   [2] NEWA VOWEL SIGN O..NEWA VOWEL SIGN AU
    Interval{ 0x11442, 0x11444 }, // Mn   [3] NEWA SIGN VIRAMA..NEWA SIGN ANUSVARA
    Interval{ 0x11445, 0x11445 }, // Mc       NEWA SIGN VISARGA
    Interval{ 0x11446, 0x11446 }, // Mn       NEWA SIGN NUKTA
    Interval{ 0x11447, 0x1144A }, // Lo   [4] NEWA SIGN AVAGRAHA..NEWA SIDDHI
    Interval{ 0x11450, 0x11459 }, // Nd  [10] NEWA DIGIT ZERO..NEWA DIGIT NINE
    Interval{ 0x1145E, 0x1145E }, // Mn       NEWA SANDHI MARK
    Interval{ 0x1145F, 0x11461 }, // Lo   [3] NEWA LETTER VEDIC ANUSVARA..NEWA SIGN UPADHMANIYA
    Interval{ 0x11480, 0x114AF }, // Lo  [48] TIRHUTA ANJI..TIRHUTA LETTER HA
    Interval{ 0x114B0, 0x114B2 }, // Mc   [3] TIRHUTA VOWEL SIGN AA..TIRHUTA VOWEL SIGN II
    Interval{ 0x114B3, 0x114B8 }, // Mn   [6] TIRHUTA VOWEL SIGN U..TIRHUTA VOWEL SIGN VOCALIC LL
    Interval{ 0x114B9, 0x114B9 }, // Mc       TIRHUTA VOWEL SIGN E
    Interval{ 0x114BA, 0x114BA }, // Mn       TIRHUTA VOWEL SIGN SHORT E
    Interval{ 0x114BB, 0x114BE }, // Mc   [4] TIRHUTA VOWEL SIGN AI..TIRHUTA VOWEL SIGN AU
    Interval{ 0x114BF, 0x114C0 }, // Mn   [2] TIRHUTA SIGN CANDRABINDU..TIRHUTA SIGN ANUSVARA
    Interval{ 0x114C1, 0x114C1 }, // Mc       TIRHUTA SIGN VISARGA
    Interval{ 0x114C2, 0x114C3 }, // Mn   [2] TIRHUTA SIGN VIRAMA..TIRHUTA SIGN NUKTA
    Interval{ 0x114C4, 0x114C5 }, // Lo   [2] TIRHUTA SIGN AVAGRAHA..TIRHUTA GVANG
    Interval{ 0x114C7, 0x114C7 }, // Lo       TIRHUTA OM
    Interval{ 0x114D0, 0x114D9 }, // Nd  [10] TIRHUTA DIGIT ZERO..TIRHUTA DIGIT NINE
    Interval{ 0x11580, 0x115AE }, // Lo  [47] SIDDHAM LETTER A..SIDDHAM LETTER HA
    Interval{ 0x115AF, 0x115B1 }, // Mc   [3] SIDDHAM VOWEL SIGN AA..SIDDHAM VOWEL SIGN II
    Interval{ 0x115B2, 0x115B5 }, // Mn   [4] SIDDHAM VOWEL SIGN U..SIDDHAM VOWEL SIGN VOCALIC RR
    Interval{ 0x115B8, 0x115BB }, // Mc   [4] SIDDHAM VOWEL SIGN E..SIDDHAM VOWEL SIGN AU
    Interval{ 0x115BC, 0x115BD }, // Mn   [2] SIDDHAM SIGN CANDRABINDU..SIDDHAM SIGN ANUSVARA
    Interval{ 0x115BE, 0x115BE }, // Mc       SIDDHAM SIGN VISARGA
    Interval{ 0x115BF, 0x115C0 }, // Mn   [2] SIDDHAM SIGN VIRAMA..SIDDHAM SIGN NUKTA
    Interval{ 0x115D8, 0x115DB }, // Lo   [4] SIDDHAM LETTER THREE-CIRCLE ALTERNATE I..SIDDHAM LETTER ALTERNATE U
    Interval{ 0x115DC, 0x115DD }, // Mn   [2] SIDDHAM VOWEL SIGN ALTERNATE U..SIDDHAM VOWEL SIGN ALTERNATE UU
    Interval{ 0x11600, 0x1162F }, // Lo  [48] MODI LETTER A..MODI LETTER LLA
    Interval{ 0x11630, 0x11632 }, // Mc   [3] MODI VOWEL SIGN AA..MODI VOWEL SIGN II
    Interval{ 0x11633, 0x1163A }, // Mn   [8] MODI VOWEL SIGN U..MODI VOWEL SIGN AI
    Interval{ 0x1163B, 0x1163C }, // Mc   [2] MODI VOWEL SIGN O..MODI VOWEL SIGN AU
    Interval{ 0x1163D, 0x1163D }, // Mn       MODI SIGN ANUSVARA
    Interval{ 0x1163E, 0x1163E }, // Mc       MODI SIGN VISARGA
    Interval{ 0x1163F, 0x11640 }, // Mn   [2] MODI SIGN VIRAMA..MODI SIGN ARDHACANDRA
    Interval{ 0x11644, 0x11644 }, // Lo       MODI SIGN HUVA
    Interval{ 0x11650, 0x11659 }, // Nd  [10] MODI DIGIT ZERO..MODI DIGIT NINE
    Interval{ 0x11680, 0x116AA }, // Lo  [43] TAKRI LETTER A..TAKRI LETTER RRA
    Interval{ 0x116AB, 0x116AB }, // Mn       TAKRI SIGN ANUSVARA
    Interval{ 0x116AC, 0x116AC }, // Mc       TAKRI SIGN VISARGA
    Interval{ 0x116AD, 0x116AD }, // Mn       TAKRI VOWEL SIGN AA
    Interval{ 0x116AE, 0x116AF }, // Mc   [2] TAKRI VOWEL SIGN I..TAKRI VOWEL SIGN II
    Interval{ 0x116B0, 0x116B5 }, // Mn   [6] TAKRI VOWEL SIGN U..TAKRI VOWEL SIGN AU
    Interval{ 0x116B6, 0x116B6 }, // Mc       TAKRI SIGN VIRAMA
    Interval{ 0x116B7, 0x116B7 }, // Mn       TAKRI SIGN NUKTA
    Interval{ 0x116B8, 0x116B8 }, // Lo       TAKRI LETTER ARCHAIC KHA
    Interval{ 0x116C0, 0x116C9 }, // Nd  [10] TAKRI DIGIT ZERO..TAKRI DIGIT NINE
    Interval{ 0x116D0, 0x116E3 }, // Nd  [20] MYANMAR PAO DIGIT ZERO..MYANMAR EASTERN PWO KAREN DIGIT NINE
    Interval{ 0x11700, 0x1171A }, // Lo  [27] AHOM LETTER KA..AHOM LETTER ALTERNATE BA
    Interval{ 0x1171D, 0x1171D }, // Mn       AHOM CONSONANT SIGN MEDIAL LA
    Interval{ 0x1171E, 0x1171E }, // Mc       AHOM CONSONANT SIGN MEDIAL RA
    Interval{ 0x1171F, 0x1171F }, // Mn       AHOM CONSONANT SIGN MEDIAL LIGATING RA
    Interval{ 0x11720, 0x11721 }, // Mc   [2] AHOM VOWEL SIGN A..AHOM VOWEL SIGN AA
    Interval{ 0x11722, 0x11725 }, // Mn   [4] AHOM VOWEL SIGN I..AHOM VOWEL SIGN UU
    Interval{ 0x11726, 0x11726 }, // Mc       AHOM VOWEL SIGN E
    Interval{ 0x11727, 0x1172B }, // Mn   [5] AHOM VOWEL SIGN AW..AHOM SIGN KILLER
    Interval{ 0x11730, 0x11739 }, // Nd  [10] AHOM DIGIT ZERO..AHOM DIGIT NINE
    Interval{ 0x11740, 0x11746 }, // Lo   [7] AHOM LETTER CA..AHOM LETTER LLA
    Interval{ 0x11800, 0x1182B }, // Lo  [44] DOGRA LETTER A..DOGRA LETTER RRA
    Interval{ 0x1182C, 0x1182E }, // Mc   [3] DOGRA VOWEL SIGN AA..DOGRA VOWEL SIGN II
    Interval{ 0x1182F, 0x11837 }, // Mn   [9] DOGRA VOWEL SIGN U..DOGRA SIGN ANUSVARA
    Interval{ 0x11838, 0x11838 }, // Mc       DOGRA SIGN VISARGA
    Interval{ 0x11839, 0x1183A }, // Mn   [2] DOGRA SIGN VIRAMA..DOGRA SIGN NUKTA
    Interval{ 0x118A0, 0x118DF }, // L&  [64] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Interval{ 0x118E0, 0x118E9 }, // Nd  [10] WARANG CITI DIGIT ZERO..WARANG CITI DIGIT NINE
    Interval{ 0x118FF, 0x11906 }, // Lo   [8] WARANG CITI OM..DIVES AKURU LETTER E
    Interval{ 0x11909, 0x11909 }, // Lo       DIVES AKURU LETTER O
    Interval{ 0x1190C, 0x11913 }, // Lo   [8] DIVES AKURU LETTER KA..DIVES AKURU LETTER JA
    Interval{ 0x11915, 0x11916 }, // Lo   [2] DIVES AKURU LETTER NYA..DIVES AKURU LETTER TTA
    Interval{ 0x11918, 0x1192F }, // Lo  [24] DIVES AKURU LETTER DDA..DIVES AKURU LETTER ZA
    Interval{ 0x11930, 0x11935 }, // Mc   [6] DIVES AKURU VOWEL SIGN AA..DIVES AKURU VOWEL SIGN E
    Interval{ 0x11937, 0x11938 }, // Mc   [2] DIVES AKURU VOWEL SIGN AI..DIVES AKURU VOWEL SIGN O
    Interval{ 0x1193B, 0x1193C }, // Mn   [2] DIVES AKURU SIGN ANUSVARA..DIVES AKURU SIGN CANDRABINDU
    Interval{ 0x1193D, 0x1193D }, // Mc       DIVES AKURU SIGN HALANTA
    Interval{ 0x1193E, 0x1193E }, // Mn       DIVES AKURU VIRAMA
    Interval{ 0x1193F, 0x1193F }, // Lo       DIVES AKURU PREFIXED NASAL SIGN
    Interval{ 0x11940, 0x11940 }, // Mc       DIVES AKURU MEDIAL YA
    Interval{ 0x11941, 0x11941 }, // Lo       DIVES AKURU INITIAL RA
    Interval{ 0x11942, 0x11942 }, // Mc       DIVES AKURU MEDIAL RA
    Interval{ 0x11943, 0x11943 }, // Mn       DIVES AKURU SIGN NUKTA
    Interval{ 0x11950, 0x11959 }, // Nd  [10] DIVES AKURU DIGIT ZERO..DIVES AKURU DIGIT NINE
    Interval{ 0x119A0, 0x119A7 }, // Lo   [8] NANDINAGARI LETTER A..NANDINAGARI LETTER VOCALIC RR
    Interval{ 0x119AA, 0x119D0 }, // Lo  [39] NANDINAGARI LETTER E..NANDINAGARI LETTER RRA
    Interval{ 0x119D1, 0x119D3 }, // Mc   [3] NANDINAGARI VOWEL SIGN AA..NANDINAGARI VOWEL SIGN II
    Interval{ 0x119D4, 0x119D7 }, // Mn   [4] NANDINAGARI VOWEL SIGN U..NANDINAGARI VOWEL SIGN VOCALIC RR
    Interval{ 0x119DA, 0x119DB }, // Mn   [2] NANDINAGARI VOWEL SIGN E..NANDINAGARI VOWEL SIGN AI
    Interval{ 0x119DC, 0x119DF }, // Mc   [4] NANDINAGARI VOWEL SIGN O..NANDINAGARI SIGN VISARGA
    Interval{ 0x119E0, 0x119E0 }, // Mn       NANDINAGARI SIGN VIRAMA
    Interval{ 0x119E1, 0x119E1 }, // Lo       NANDINAGARI SIGN AVAGRAHA
    Interval{ 0x119E3, 0x119E3 }, // Lo       NANDINAGARI HEADSTROKE
    Interval{ 0x119E4, 0x119E4 }, // Mc       NANDINAGARI VOWEL SIGN PRISHTHAMATRA E
    Interval{ 0x11A00, 0x11A00 }, // Lo       ZANABAZAR SQUARE LETTER A
    Interval{ 0x11A01, 0x11A0A }, // Mn  [10] ZANABAZAR SQUARE VOWEL SIGN I..ZANABAZAR SQUARE VOWEL LENGTH MARK
    Interval{ 0x11A0B, 0x11A32 }, // Lo  [40] ZANABAZAR SQUARE LETTER KA..ZANABAZAR SQUARE LETTER KSSA
    Interval{ 0x11A33, 0x11A38 }, // Mn   [6] ZANABAZAR SQUARE FINAL CONSONANT MARK..ZANABAZAR SQUARE SIGN ANUSVARA
    Interval{ 0x11A39, 0x11A39 }, // Mc       ZANABAZAR SQUARE SIGN VISARGA
    Interval{ 0x11A3A, 0x11A3A }, // Lo       ZANABAZAR SQUARE CLUSTER-INITIAL LETTER RA
    Interval{ 0x11A3B, 0x11A3E }, // Mn   [4] ZANABAZAR SQUARE CLUSTER-FINAL LETTER YA..ZANABAZAR SQUARE CLUSTER-FINAL LETTER VA
    Interval{ 0x11A47, 0x11A47 }, // Mn       ZANABAZAR SQUARE SUBJOINER
    Interval{ 0x11A50, 0x11A50 }, // Lo       SOYOMBO LETTER A
    Interval{ 0x11A51, 0x11A56 }, // Mn   [6] SOYOMBO VOWEL SIGN I..SOYOMBO VOWEL SIGN OE
    Interval{ 0x11A57, 0x11A58 }, // Mc   [2] SOYOMBO VOWEL SIGN AI..SOYOMBO VOWEL SIGN AU
    Interval{ 0x11A59, 0x11A5B }, // Mn   [3] SOYOMBO VOWEL SIGN VOCALIC R..SOYOMBO VOWEL LENGTH MARK
    Interval{ 0x11A5C, 0x11A89 }, // Lo  [46] SOYOMBO LETTER KA..SOYOMBO CLUSTER-INITIAL LETTER SA
    Interval{ 0x11A8A, 0x11A96 }, // Mn  [13] SOYOMBO FINAL CONSONANT SIGN G..SOYOMBO SIGN ANUSVARA
    Interval{ 0x11A97, 0x11A97 }, // Mc       SOYOMBO SIGN VISARGA
    Interval{ 0x11A98, 0x11A99 }, // Mn   [2] SOYOMBO GEMINATION MARK..SOYOMBO SUBJOINER
    Interval{ 0x11A9D, 0x11A9D }, // Lo       SOYOMBO MARK PLUTA
    Interval{ 0x11AB0, 0x11AF8 }, // Lo  [73] CANADIAN SYLLABICS NATTILIK HI..PAU CIN HAU GLOTTAL STOP FINAL
    Interval{ 0x11B60, 0x11B60 }, // Mn       SHARADA VOWEL SIGN OE
    Interval{ 0x11B61, 0x11B61 }, // Mc       SHARADA VOWEL SIGN OOE
    Interval{ 0x11B62, 0x11B64 }, // Mn   [3] SHARADA VOWEL SIGN UE..SHARADA VOWEL SIGN SHORT E
    Interval{ 0x11B65, 0x11B65 }, // Mc       SHARADA VOWEL SIGN SHORT O
    Interval{ 0x11B66, 0x11B66 }, // Mn       SHARADA VOWEL SIGN CANDRA E
    Interval{ 0x11B67, 0x11B67 }, // Mc       SHARADA VOWEL SIGN CANDRA O
    Interval{ 0x11BC0, 0x11BE0 }, // Lo  [33] SUNUWAR LETTER DEVI..SUNUWAR LETTER KLOKO
    Interval{ 0x11BF0, 0x11BF9 }, // Nd  [10] SUNUWAR DIGIT ZERO..SUNUWAR DIGIT NINE
    Interval{ 0x11C00, 0x11C08 }, // Lo   [9] BHAIKSUKI LETTER A..BHAIKSUKI LETTER VOCALIC L
    Interval{ 0x11C0A, 0x11C2E }, // Lo  [37] BHAIKSUKI LETTER E..BHAIKSUKI LETTER HA
    Interval{ 0x11C2F, 0x11C2F }, // Mc       BHAIKSUKI VOWEL SIGN AA
    Interval{ 0x11C30, 0x11C36 }, // Mn   [7] BHAIKSUKI VOWEL SIGN I..BHAIKSUKI VOWEL SIGN VOCALIC L
    Interval{ 0x11C38, 0x11C3D }, // Mn   [6] BHAIKSUKI VOWEL SIGN E..BHAIKSUKI SIGN ANUSVARA
    Interval{ 0x11C3E, 0x11C3E }, // Mc       BHAIKSUKI SIGN VISARGA
    Interval{ 0x11C3F, 0x11C3F }, // Mn       BHAIKSUKI SIGN VIRAMA
    Interval{ 0x11C40, 0x11C40 }, // Lo       BHAIKSUKI SIGN AVAGRAHA
    Interval{ 0x11C50, 0x11C59 }, // Nd  [10] BHAIKSUKI DIGIT ZERO..BHAIKSUKI DIGIT NINE
    Interval{ 0x11C72, 0x11C8F }, // Lo  [30] MARCHEN LETTER KA..MARCHEN LETTER A
    Interval{ 0x11C92, 0x11CA7 }, // Mn  [22] MARCHEN SUBJOINED LETTER KA..MARCHEN SUBJOINED LETTER ZA
    Interval{ 0x11CA9, 0x11CA9 }, // Mc       MARCHEN SUBJOINED LETTER YA
    Interval{ 0x11CAA, 0x11CB0 }, // Mn   [7] MARCHEN SUBJOINED LETTER RA..MARCHEN VOWEL SIGN AA
    Interval{ 0x11CB1, 0x11CB1 }, // Mc       MARCHEN VOWEL SIGN I
    Interval{ 0x11CB2, 0x11CB3 }, // Mn   [2] MARCHEN VOWEL SIGN U..MARCHEN VOWEL SIGN E
    Interval{ 0x11CB4, 0x11CB4 }, // Mc       MARCHEN VOWEL SIGN O
    Interval{ 0x11CB5, 0x11CB6 }, // Mn   [2] MARCHEN SIGN ANUSVARA..MARCHEN SIGN CANDRABINDU
    Interval{ 0x11D00, 0x11D06 }, // Lo   [7] MASARAM GONDI LETTER A..MASARAM GONDI LETTER E
    Interval{ 0x11D08, 0x11D09 }, // Lo   [2] MASARAM GONDI LETTER AI..MASARAM GONDI LETTER O
    Interval{ 0x11D0B, 0x11D30 }, // Lo  [38] MASARAM GONDI LETTER AU..MASARAM GONDI LETTER TRA
    Interval{ 0x11D31, 0x11D36 }, // Mn   [6] MASARAM GONDI VOWEL SIGN AA..MASARAM GONDI VOWEL SIGN VOCALIC R
    Interval{ 0x11D3A, 0x11D3A }, // Mn       MASARAM GONDI VOWEL SIGN E
    Interval{ 0x11D3C, 0x11D3D }, // Mn   [2] MASARAM GONDI VOWEL SIGN AI..MASARAM GONDI VOWEL SIGN O
    Interval{ 0x11D3F, 0x11D45 }, // Mn   [7] MASARAM GONDI VOWEL SIGN AU..MASARAM GONDI VIRAMA
    Interval{ 0x11D46, 0x11D46 }, // Lo       MASARAM GONDI REPHA
    Interval{ 0x11D47, 0x11D47 }, // Mn       MASARAM GONDI RA-KARA
    Interval{ 0x11D50, 0x11D59 }, // Nd  [10] MASARAM GONDI DIGIT ZERO..MASARAM GONDI DIGIT NINE
    Interval{ 0x11D60, 0x11D65 }, // Lo   [6] GUNJALA GONDI LETTER A..GUNJALA GONDI LETTER UU
    Interval{ 0x11D67, 0x11D68 }, // Lo   [2] GUNJALA GONDI LETTER EE..GUNJALA GONDI LETTER AI
    Interval{ 0x11D6A, 0x11D89 }, // Lo  [32] GUNJALA GONDI LETTER OO..GUNJALA GONDI LETTER SA
    Interval{ 0x11D8A, 0x11D8E }, // Mc   [5] GUNJALA GONDI VOWEL SIGN AA..GUNJALA GONDI VOWEL SIGN UU
    Interval{ 0x11D90, 0x11D91 }, // Mn   [2] GUNJALA GONDI VOWEL SIGN EE..GUNJALA GONDI VOWEL SIGN AI
    Interval{ 0x11D93, 0x11D94 }, // Mc   [2] GUNJALA GONDI VOWEL SIGN OO..GUNJALA GONDI VOWEL SIGN AU
    Interval{ 0x11D95, 0x11D95 }, // Mn       GUNJALA GONDI SIGN ANUSVARA
    Interval{ 0x11D96, 0x11D96 }, // Mc       GUNJALA GONDI SIGN VISARGA
    Interval{ 0x11D97, 0x11D97 }, // Mn       GUNJALA GONDI VIRAMA
    Interval{ 0x11D98, 0x11D98 }, // Lo       GUNJALA GONDI OM
    Interval{ 0x11DA0, 0x11DA9 }, // Nd  [10] GUNJALA GONDI DIGIT ZERO..GUNJALA GONDI DIGIT NINE
    Interval{ 0x11DB0, 0x11DD8 }, // Lo  [41] TOLONG SIKI LETTER I..TOLONG SIKI LETTER RRH
    Interval{ 0x11DD9, 0x11DD9 }, // Lm       TOLONG SIKI SIGN SELA
    Interval{ 0x11DDA, 0x11DDB }, // Lo   [2] TOLONG SIKI SIGN HECAKA..TOLONG SIKI UNGGA
    Interval{ 0x11DE0, 0x11DE9 }, // Nd  [10] TOLONG SIKI DIGIT ZERO..TOLONG SIKI DIGIT NINE
    Interval{ 0x11EE0, 0x11EF2 }, // Lo  [19] MAKASAR LETTER KA..MAKASAR ANGKA
    Interval{ 0x11EF3, 0x11EF4 }, // Mn   [2] MAKASAR VOWEL SIGN I..MAKASAR VOWEL SIGN U
    Interval{ 0x11EF5, 0x11EF6 }, // Mc   [2] MAKASAR VOWEL SIGN E..MAKASAR VOWEL SIGN O
    Interval{ 0x11F00, 0x11F01 }, // Mn   [2] KAWI SIGN CANDRABINDU..KAWI SIGN ANUSVARA
    Interval{ 0x11F02, 0x11F02 }, // Lo       KAWI SIGN REPHA
    Interval{ 0x11F03, 0x11F03 }, // Mc       KAWI SIGN VISARGA
    Interval{ 0x11F04, 0x11F10 }, // Lo  [13] KAWI LETTER A..KAWI LETTER O
    Interval{ 0x11F12, 0x11F33 }, // Lo  [34] KAWI LETTER KA..KAWI LETTER JNYA
    Interval{ 0x11F34, 0x11F35 }, // Mc   [2] KAWI VOWEL SIGN AA..KAWI VOWEL SIGN ALTERNATE AA
    Interval{ 0x11F36, 0x11F3A }, // Mn   [5] KAWI VOWEL SIGN I..KAWI VOWEL SIGN VOCALIC R
    Interval{ 0x11F3E, 0x11F3F }, // Mc   [2] KAWI VOWEL SIGN E..KAWI VOWEL SIGN AI
    Interval{ 0x11F40, 0x11F40 }, // Mn       KAWI VOWEL SIGN EU
    Interval{ 0x11F41, 0x11F41 }, // Mc       KAWI SIGN KILLER
    Interval{ 0x11F42, 0x11F42 }, // Mn       KAWI CONJOINER
    Interval{ 0x11F50, 0x11F59 }, // Nd  [10] KAWI DIGIT ZERO..KAWI DIGIT NINE
    Interval{ 0x11F5A, 0x11F5A }, // Mn       KAWI SIGN NUKTA
    Interval{ 0x11FB0, 0x11FB0 }, // Lo       LISU LETTER YHA
    Interval{ 0x12000, 0x12399 }, // Lo [922] CUNEIFORM SIGN A..CUNEIFORM SIGN U U
    Interval{ 0x12400, 0x1246E }, // Nl [111] CUNEIFORM NUMERIC SIGN TWO ASH..CUNEIFORM NUMERIC SIGN NINE U VARIANT FORM
    Interval{ 0x12480, 0x12543 }, // Lo [196] CUNEIFORM SIGN AB TIMES NUN TENU..CUNEIFORM SIGN ZU5 TIMES THREE DISH TENU
    Interval{ 0x12F90, 0x12FF0 }, // Lo  [97] CYPRO-MINOAN SIGN CM001..CYPRO-MINOAN SIGN CM114
    Interval{ 0x13000, 0x1342F }, // Lo [1072] EGYPTIAN HIEROGLYPH A001..EGYPTIAN HIEROGLYPH V011D
    Interval{ 0x13440, 0x13440 }, // Mn       EGYPTIAN HIEROGLYPH MIRROR HORIZONTALLY
    Interval{ 0x13441, 0x13446 }, // Lo   [6] EGYPTIAN HIEROGLYPH FULL BLANK..EGYPTIAN HIEROGLYPH WIDE LOST SIGN
    Interval{ 0x13447, 0x13455 }, // Mn  [15] EGYPTIAN HIEROGLYPH MODIFIER DAMAGED AT TOP START..EGYPTIAN HIEROGLYPH MODIFIER DAMAGED
    Interval{ 0x13460, 0x143FA }, // Lo [3995] EGYPTIAN HIEROGLYPH-13460..EGYPTIAN HIEROGLYPH-143FA
    Interval{ 0x14400, 0x14646 }, // Lo [583] ANATOLIAN HIEROGLYPH A001..ANATOLIAN HIEROGLYPH A530
    Interval{ 0x16100, 0x1611D }, // Lo  [30] GURUNG KHEMA LETTER A..GURUNG KHEMA LETTER SA
    Interval{ 0x1611E, 0x16129 }, // Mn  [12] GURUNG KHEMA VOWEL SIGN AA..GURUNG KHEMA VOWEL LENGTH MARK
    Interval{ 0x1612A, 0x1612C }, // Mc   [3] GURUNG KHEMA CONSONANT SIGN MEDIAL YA..GURUNG KHEMA CONSONANT SIGN MEDIAL HA
    Interval{ 0x1612D, 0x1612F }, // Mn   [3] GURUNG KHEMA SIGN ANUSVARA..GURUNG KHEMA SIGN THOLHOMA
    Interval{ 0x16130, 0x16139 }, // Nd  [10] GURUNG KHEMA DIGIT ZERO..GURUNG KHEMA DIGIT NINE
    Interval{ 0x16800, 0x16A38 }, // Lo [569] BAMUM LETTER PHASE-A NGKUE MFON..BAMUM LETTER PHASE-F VUEQ
    Interval{ 0x16A40, 0x16A5E }, // Lo  [31] MRO LETTER TA..MRO LETTER TEK
    Interval{ 0x16A60, 0x16A69 }, // Nd  [10] MRO DIGIT ZERO..MRO DIGIT NINE
    Interval{ 0x16A70, 0x16ABE }, // Lo  [79] TANGSA LETTER OZ..TANGSA LETTER ZA
    Interval{ 0x16AC0, 0x16AC9 }, // Nd  [10] TANGSA DIGIT ZERO..TANGSA DIGIT NINE
    Interval{ 0x16AD0, 0x16AED }, // Lo  [30] BASSA VAH LETTER ENNI..BASSA VAH LETTER I
    Interval{ 0x16AF0, 0x16AF4 }, // Mn   [5] BASSA VAH COMBINING HIGH TONE..BASSA VAH COMBINING HIGH-LOW TONE
    Interval{ 0x16B00, 0x16B2F }, // Lo  [48] PAHAWH HMONG VOWEL KEEB..PAHAWH HMONG CONSONANT CAU
    Interval{ 0x16B30, 0x16B36 }, // Mn   [7] PAHAWH HMONG MARK CIM TUB..PAHAWH HMONG MARK CIM TAUM
    Interval{ 0x16B40, 0x16B43 }, // Lm   [4] PAHAWH HMONG SIGN VOS SEEV..PAHAWH HMONG SIGN IB YAM
    Interval{ 0x16B50, 0x16B59 }, // Nd  [10] PAHAWH HMONG DIGIT ZERO..PAHAWH HMONG DIGIT NINE
    Interval{ 0x16B63, 0x16B77 }, // Lo  [21] PAHAWH HMONG SIGN VOS LUB..PAHAWH HMONG SIGN CIM NRES TOS
    Interval{ 0x16B7D, 0x16B8F }, // Lo  [19] PAHAWH HMONG CLAN SIGN TSHEEJ..PAHAWH HMONG CLAN SIGN VWJ
    Interval{ 0x16D40, 0x16D42 }, // Lm   [3] KIRAT RAI SIGN ANUSVARA..KIRAT RAI SIGN VISARGA
    Interval{ 0x16D43, 0x16D6A }, // Lo  [40] KIRAT RAI LETTER A..KIRAT RAI VOWEL SIGN AU
    Interval{ 0x16D6B, 0x16D6C }, // Lm   [2] KIRAT RAI SIGN VIRAMA..KIRAT RAI SIGN SAAT
    Interval{ 0x16D70, 0x16D79 }, // Nd  [10] KIRAT RAI DIGIT ZERO..KIRAT RAI DIGIT NINE
    Interval{ 0x16E40, 0x16E7F }, // L&  [64] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Interval{ 0x16EA0, 0x16EB8 }, // L&  [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Interval{ 0x16EBB, 0x16ED3 }, // L&  [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Interval{ 0x16F00, 0x16F4A }, // Lo  [75] MIAO LETTER PA..MIAO LETTER RTE
    Interval{ 0x16F4F, 0x16F4F }, // Mn       MIAO SIGN CONSONANT MODIFIER BAR
    Interval{ 0x16F50, 0x16F50 }, // Lo       MIAO LETTER NASALIZATION
    Interval{ 0x16F51, 0x16F87 }, // Mc  [55] MIAO SIGN ASPIRATION..MIAO VOWEL SIGN UI
    Interval{ 0x16F8F, 0x16F92 }, // Mn   [4] MIAO TONE RIGHT..MIAO TONE BELOW
    Interval{ 0x16F93, 0x16F9F }, // Lm  [13] MIAO LETTER TONE-2..MIAO LETTER REFORMED TONE-8
    Interval{ 0x16FE0, 0x16FE1 }, // Lm   [2] TANGUT ITERATION MARK..NUSHU ITERATION MARK
    Interval{ 0x16FE3, 0x16FE3 }, // Lm       OLD CHINESE ITERATION MARK
    Interval{ 0x16FE4, 0x16FE4 }, // Mn       KHITAN SMALL SCRIPT FILLER
    Interval{ 0x16FF0, 0x16FF1 }, // Mc   [2] VIETNAMESE ALTERNATE READING MARK CA..VIETNAMESE ALTERNATE READING MARK NHAY
    Interval{ 0x16FF2, 0x16FF3 }, // Lm   [2] CHINESE SMALL SIMPLIFIED ER..CHINESE SMALL TRADITIONAL ER
    Interval{ 0x16FF4, 0x16FF6 }, // Nl   [3] YANGQIN SIGN SLOW ONE BEAT..YANGQIN SIGN SLOW TWO BEATS
    Interval{ 0x17000, 0x18CD5 }, // Lo [7382] TANGUT IDEOGRAPH-17000..KHITAN SMALL SCRIPT CHARACTER-18CD5
    Interval{ 0x18CFF, 0x18D1E }, // Lo  [32] KHITAN SMALL SCRIPT CHARACTER-18CFF..TANGUT IDEOGRAPH-18D1E
    Interval{ 0x18D80, 0x18DF2 }, // Lo [115] TANGUT COMPONENT-769..TANGUT COMPONENT-883
    Interval{ 0x1AFF0, 0x1AFF3 }, // Lm   [4] KATAKANA LETTER MINNAN TONE-2..KATAKANA LETTER MINNAN TONE-5
    Interval{ 0x1AFF5, 0x1AFFB }, // Lm   [7] KATAKANA LETTER MINNAN TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-5
    Interval{ 0x1AFFD, 0x1AFFE }, // Lm   [2] KATAKANA LETTER MINNAN NASALIZED TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-8
    Interval{ 0x1B000, 0x1B122 }, // Lo [291] KATAKANA LETTER ARCHAIC E..KATAKANA LETTER ARCHAIC WU
    Interval{ 0x1B132, 0x1B132 }, // Lo       HIRAGANA LETTER SMALL KO
    Interval{ 0x1B150, 0x1B152 }, // Lo   [3] HIRAGANA LETTER SMALL WI..HIRAGANA LETTER SMALL WO
    Interval{ 0x1B155, 0x1B155 }, // Lo       KATAKANA LETTER SMALL KO
    Interval{ 0x1B164, 0x1B167 }, // Lo   [4] KATAKANA LETTER SMALL WI..KATAKANA LETTER SMALL N
    Interval{ 0x1B170, 0x1B2FB }, // Lo [396] NUSHU CHARACTER-1B170..NUSHU CHARACTER-1B2FB
    Interval{ 0x1BC00, 0x1BC6A }, // Lo [107] DUPLOYAN LETTER H..DUPLOYAN LETTER VOCALIC M
    Interval{ 0x1BC70, 0x1BC7C }, // Lo  [13] DUPLOYAN AFFIX LEFT HORIZONTAL SECANT..DUPLOYAN AFFIX ATTACHED TANGENT HOOK
    Interval{ 0x1BC80, 0x1BC88 }, // Lo   [9] DUPLOYAN AFFIX HIGH ACUTE..DUPLOYAN AFFIX HIGH VERTICAL
    Interval{ 0x1BC90, 0x1BC99 }, // Lo  [10] DUPLOYAN AFFIX LOW ACUTE..DUPLOYAN AFFIX LOW ARROW
    Interval{ 0x1BC9D, 0x1BC9E }, // Mn   [2] DUPLOYAN THICK LETTER SELECTOR..DUPLOYAN DOUBLE MARK
    Interval{ 0x1CCF0, 0x1CCF9 }, // Nd  [10] OUTLINED DIGIT ZERO..OUTLINED DIGIT NINE
    Interval{ 0x1CF00, 0x1CF2D }, // Mn  [46] ZNAMENNY COMBINING MARK GORAZDO NIZKO S KRYZHEM ON LEFT..ZNAMENNY COMBINING MARK KRYZH ON LEFT
    Interval{ 0x1CF30, 0x1CF46 }, // Mn  [23] ZNAMENNY COMBINING TONAL RANGE MARK MRACHNO..ZNAMENNY PRIZNAK MODIFIER ROG
    Interval{ 0x1D165, 0x1D166 }, // Mc   [2] MUSICAL SYMBOL COMBINING STEM..MUSICAL SYMBOL COMBINING SPRECHGESANG STEM
    Interval{ 0x1D167, 0x1D169 }, // Mn   [3] MUSICAL SYMBOL COMBINING TREMOLO-1..MUSICAL SYMBOL COMBINING TREMOLO-3
    Interval{ 0x1D16D, 0x1D172 }, // Mc   [6] MUSICAL SYMBOL COMBINING AUGMENTATION DOT..MUSICAL SYMBOL COMBINING FLAG-5
    Interval{ 0x1D17B, 0x1D182 }, // Mn   [8] MUSICAL SYMBOL COMBINING ACCENT..MUSICAL SYMBOL COMBINING LOURE
    Interval{ 0x1D185, 0x1D18B }, // Mn   [7] MUSICAL SYMBOL COMBINING DOIT..MUSICAL SYMBOL COMBINING TRIPLE TONGUE
    Interval{ 0x1D1AA, 0x1D1AD }, // Mn   [4] MUSICAL SYMBOL COMBINING DOWN BOW..MUSICAL SYMBOL COMBINING SNAP PIZZICATO
    Interval{ 0x1D242, 0x1D244 }, // Mn   [3] COMBINING GREEK MUSICAL TRISEME..COMBINING GREEK MUSICAL PENTASEME
    Interval{ 0x1D400, 0x1D454 }, // L&  [85] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL ITALIC SMALL G
    Interval{ 0x1D456, 0x1D49C }, // L&  [71] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL SCRIPT CAPITAL A
    Interval{ 0x1D49E, 0x1D49F }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Interval{ 0x1D4A2, 0x1D4A2 }, // L&       MATHEMATICAL SCRIPT CAPITAL G
    Interval{ 0x1D4A5, 0x1D4A6 }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Interval{ 0x1D4A9, 0x1D4AC }, // L&   [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Interval{ 0x1D4AE, 0x1D4B9 }, // L&  [12] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT SMALL D
    Interval{ 0x1D4BB, 0x1D4BB }, // L&       MATHEMATICAL SCRIPT SMALL F
    Interval{ 0x1D4BD, 0x1D4C3 }, // L&   [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Interval{ 0x1D4C5, 0x1D505 }, // L&  [65] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL FRAKTUR CAPITAL B
    Interval{ 0x1D507, 0x1D50A }, // L&   [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Interval{ 0x1D50D, 0x1D514 }, // L&   [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Interval{ 0x1D516, 0x1D51C }, // L&   [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Interval{ 0x1D51E, 0x1D539 }, // L&  [28] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Interval{ 0x1D53B, 0x1D53E }, // L&   [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Interval{ 0x1D540, 0x1D544 }, // L&   [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Interval{ 0x1D546, 0x1D546 }, // L&       MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Interval{ 0x1D54A, 0x1D550 }, // L&   [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Interval{ 0x1D552, 0x1D6A5 }, // L& [340] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    Interval{ 0x1D6A8, 0x1D6C0 }, // L&  [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    Interval{ 0x1D6C2, 0x1D6DA }, // L&  [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    Interval{ 0x1D6DC, 0x1D6FA }, // L&  [31] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL ITALIC CAPITAL OMEGA
    Interval{ 0x1D6FC, 0x1D714 }, // L&  [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    Interval{ 0x1D716, 0x1D734 }, // L&  [31] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D736, 0x1D74E }, // L&  [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D750, 0x1D76E }, // L&  [31] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    Interval{ 0x1D770, 0x1D788 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    Interval{ 0x1D78A, 0x1D7A8 }, // L&  [31] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D7AA, 0x1D7C2 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D7C4, 0x1D7CB }, // L&   [8] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD SMALL DIGAMMA
    Interval{ 0x1D7CE, 0x1D7FF }, // Nd  [50] MATHEMATICAL BOLD DIGIT ZERO..MATHEMATICAL MONOSPACE DIGIT NINE
    Interval{ 0x1DA00, 0x1DA36 }, // Mn  [55] SIGNWRITING HEAD RIM..SIGNWRITING AIR SUCKING IN
    Interval{ 0x1DA3B, 0x1DA6C }, // Mn  [50] SIGNWRITING MOUTH CLOSED NEUTRAL..SIGNWRITING EXCITEMENT
    Interval{ 0x1DA75, 0x1DA75 }, // Mn       SIGNWRITING UPPER BODY TILTING FROM HIP JOINTS
    Interval{ 0x1DA84, 0x1DA84 }, // Mn       SIGNWRITING LOCATION HEAD NECK
    Interval{ 0x1DA9B, 0x1DA9F }, // Mn   [5] SIGNWRITING FILL MODIFIER-2..SIGNWRITING FILL MODIFIER-6
    Interval{ 0x1DAA1, 0x1DAAF }, // Mn  [15] SIGNWRITING ROTATION MODIFIER-2..SIGNWRITING ROTATION MODIFIER-16
    Interval{ 0x1DF00, 0x1DF09 }, // L&  [10] LATIN SMALL LETTER FENG DIGRAPH WITH TRILL..LATIN SMALL LETTER T WITH HOOK AND RETROFLEX HOOK
    Interval{ 0x1DF0A, 0x1DF0A }, // Lo       LATIN LETTER RETROFLEX CLICK WITH RETROFLEX HOOK
    Interval{ 0x1DF0B, 0x1DF1E }, // L&  [20] LATIN SMALL LETTER ESH WITH DOUBLE BAR..LATIN SMALL LETTER S WITH CURL
    Interval{ 0x1DF25, 0x1DF2A }, // L&   [6] LATIN SMALL LETTER D WITH MID-HEIGHT LEFT HOOK..LATIN SMALL LETTER T WITH MID-HEIGHT LEFT HOOK
    Interval{ 0x1E000, 0x1E006 }, // Mn   [7] COMBINING GLAGOLITIC LETTER AZU..COMBINING GLAGOLITIC LETTER ZHIVETE
    Interval{ 0x1E008, 0x1E018 }, // Mn  [17] COMBINING GLAGOLITIC LETTER ZEMLJA..COMBINING GLAGOLITIC LETTER HERU
    Interval{ 0x1E01B, 0x1E021 }, // Mn   [7] COMBINING GLAGOLITIC LETTER SHTA..COMBINING GLAGOLITIC LETTER YATI
    Interval{ 0x1E023, 0x1E024 }, // Mn   [2] COMBINING GLAGOLITIC LETTER YU..COMBINING GLAGOLITIC LETTER SMALL YUS
    Interval{ 0x1E026, 0x1E02A }, // Mn   [5] COMBINING GLAGOLITIC LETTER YO..COMBINING GLAGOLITIC LETTER FITA
    Interval{ 0x1E030, 0x1E06D }, // Lm  [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Interval{ 0x1E08F, 0x1E08F }, // Mn       COMBINING CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
    Interval{ 0x1E100, 0x1E12C }, // Lo  [45] NYIAKENG PUACHUE HMONG LETTER MA..NYIAKENG PUACHUE HMONG LETTER W
    Interval{ 0x1E130, 0x1E136 }, // Mn   [7] NYIAKENG PUACHUE HMONG TONE-B..NYIAKENG PUACHUE HMONG TONE-D
    Interval{ 0x1E137, 0x1E13D }, // Lm   [7] NYIAKENG PUACHUE HMONG SIGN FOR PERSON..NYIAKENG PUACHUE HMONG SYLLABLE LENGTHENER
    Interval{ 0x1E140, 0x1E149 }, // Nd  [10] NYIAKENG PUACHUE HMONG DIGIT ZERO..NYIAKENG PUACHUE HMONG DIGIT NINE
    Interval{ 0x1E14E, 0x1E14E }, // Lo       NYIAKENG PUACHUE HMONG LOGOGRAM NYAJ
    Interval{ 0x1E290, 0x1E2AD }, // Lo  [30] TOTO LETTER PA..TOTO LETTER A
    Interval{ 0x1E2AE, 0x1E2AE }, // Mn       TOTO SIGN RISING TONE
    Interval{ 0x1E2C0, 0x1E2EB }, // Lo  [44] WANCHO LETTER AA..WANCHO LETTER YIH
    Interval{ 0x1E2EC, 0x1E2EF }, // Mn   [4] WANCHO TONE TUP..WANCHO TONE KOINI
    Interval{ 0x1E2F0, 0x1E2F9 }, // Nd  [10] WANCHO DIGIT ZERO..WANCHO DIGIT NINE
    Interval{ 0x1E4D0, 0x1E4EA }, // Lo  [27] NAG MUNDARI LETTER O..NAG MUNDARI LETTER ELL
    Interval{ 0x1E4EB, 0x1E4EB }, // Lm       NAG MUNDARI SIGN OJOD
    Interval{ 0x1E4EC, 0x1E4EF }, // Mn   [4] NAG MUNDARI SIGN MUHOR..NAG MUNDARI SIGN SUTUH
    Interval{ 0x1E4F0, 0x1E4F9 }, // Nd  [10] NAG MUNDARI DIGIT ZERO..NAG MUNDARI DIGIT NINE
    Interval{ 0x1E5D0, 0x1E5ED }, // Lo  [30] OL ONAL LETTER O..OL ONAL LETTER EG
    Interval{ 0x1E5EE, 0x1E5EF }, // Mn   [2] OL ONAL SIGN MU..OL ONAL SIGN IKIR
    Interval{ 0x1E5F0, 0x1E5F0 }, // Lo       OL ONAL SIGN HODDOND
    Interval{ 0x1E5F1, 0x1E5FA }, // Nd  [10] OL ONAL DIGIT ZERO..OL ONAL DIGIT NINE
    Interval{ 0x1E6C0, 0x1E6DE }, // Lo  [31] TAI YO LETTER LOW KO..TAI YO LETTER HIGH KVO
    Interval{ 0x1E6E0, 0x1E6E2 }, // Lo   [3] TAI YO LETTER AA..TAI YO LETTER UE
    Interval{ 0x1E6E3, 0x1E6E3 }, // Mn       TAI YO SIGN UE
    Interval{ 0x1E6E4, 0x1E6E5 }, // Lo   [2] TAI YO LETTER U..TAI YO LETTER AE
    Interval{ 0x1E6E6, 0x1E6E6 }, // Mn       TAI YO SIGN AU
    Interval{ 0x1E6E7, 0x1E6ED }, // Lo   [7] TAI YO LETTER O..TAI YO LETTER AUE
    Interval{ 0x1E6EE, 0x1E6EF }, // Mn   [2] TAI YO SIGN AY..TAI YO SIGN ANG
    Interval{ 0x1E6F0, 0x1E6F4 }, // Lo   [5] TAI YO LETTER AN..TAI YO LETTER AP
    Interval{ 0x1E6F5, 0x1E6F5 }, // Mn       TAI YO SIGN OM
    Interval{ 0x1E6FE, 0x1E6FE }, // Lo       TAI YO SYMBOL MUEANG
    Interval{ 0x1E6FF, 0x1E6FF }, // Lm       TAI YO XAM LAI
    Interval{ 0x1E7E0, 0x1E7E6 }, // Lo   [7] ETHIOPIC SYLLABLE HHYA..ETHIOPIC SYLLABLE HHYO
    Interval{ 0x1E7E8, 0x1E7EB }, // Lo   [4] ETHIOPIC SYLLABLE GURAGE HHWA..ETHIOPIC SYLLABLE HHWE
    Interval{ 0x1E7ED, 0x1E7EE }, // Lo   [2] ETHIOPIC SYLLABLE GURAGE MWI..ETHIOPIC SYLLABLE GURAGE MWEE
    Interval{ 0x1E7F0, 0x1E7FE }, // Lo  [15] ETHIOPIC SYLLABLE GURAGE QWI..ETHIOPIC SYLLABLE GURAGE PWEE
    Interval{ 0x1E800, 0x1E8C4 }, // Lo [197] MENDE KIKAKUI SYLLABLE M001 KI..MENDE KIKAKUI SYLLABLE M060 NYON
    Interval{ 0x1E8D0, 0x1E8D6 }, // Mn   [7] MENDE KIKAKUI COMBINING NUMBER TEENS..MENDE KIKAKUI COMBINING NUMBER MILLIONS
    Interval{ 0x1E900, 0x1E943 }, // L&  [68] ADLAM CAPITAL LETTER ALIF..ADLAM SMALL LETTER SHA
    Interval{ 0x1E944, 0x1E94A }, // Mn   [7] ADLAM ALIF LENGTHENER..ADLAM NUKTA
    Interval{ 0x1E94B, 0x1E94B }, // Lm       ADLAM NASALIZATION MARK
    Interval{ 0x1E950, 0x1E959 }, // Nd  [10] ADLAM DIGIT ZERO..ADLAM DIGIT NINE
    Interval{ 0x1EE00, 0x1EE03 }, // Lo   [4] ARABIC MATHEMATICAL ALEF..ARABIC MATHEMATICAL DAL
    Interval{ 0x1EE05, 0x1EE1F }, // Lo  [27] ARABIC MATHEMATICAL WAW..ARABIC MATHEMATICAL DOTLESS QAF
    Interval{ 0x1EE21, 0x1EE22 }, // Lo   [2] ARABIC MATHEMATICAL INITIAL BEH..ARABIC MATHEMATICAL INITIAL JEEM
    Interval{ 0x1EE24, 0x1EE24 }, // Lo       ARABIC MATHEMATICAL INITIAL HEH
    Interval{ 0x1EE27, 0x1EE27 }, // Lo       ARABIC MATHEMATICAL INITIAL HAH
    Interval{ 0x1EE29, 0x1EE32 }, // Lo  [10] ARABIC MATHEMATICAL INITIAL YEH..ARABIC MATHEMATICAL INITIAL QAF
    Interval{ 0x1EE34, 0x1EE37 }, // Lo   [4] ARABIC MATHEMATICAL INITIAL SHEEN..ARABIC MATHEMATICAL INITIAL KHAH
    Interval{ 0x1EE39, 0x1EE39 }, // Lo       ARABIC MATHEMATICAL INITIAL DAD
    Interval{ 0x1EE3B, 0x1EE3B }, // Lo       ARABIC MATHEMATICAL INITIAL GHAIN
    Interval{ 0x1EE42, 0x1EE42 }, // Lo       ARABIC MATHEMATICAL TAILED JEEM
    Interval{ 0x1EE47, 0x1EE47 }, // Lo       ARABIC MATHEMATICAL TAILED HAH
    Interval{ 0x1EE49, 0x1EE49 }, // Lo       ARABIC MATHEMATICAL TAILED YEH
    Interval{ 0x1EE4B, 0x1EE4B }, // Lo       ARABIC MATHEMATICAL TAILED LAM
    Interval{ 0x1EE4D, 0x1EE4F }, // Lo   [3] ARABIC MATHEMATICAL TAILED NOON..ARABIC MATHEMATICAL TAILED AIN
    Interval{ 0x1EE51, 0x1EE52 }, // Lo   [2] ARABIC MATHEMATICAL TAILED SAD..ARABIC MATHEMATICAL TAILED QAF
    Interval{ 0x1EE54, 0x1EE54 }, // Lo       ARABIC MATHEMATICAL TAILED SHEEN
    Interval{ 0x1EE57, 0x1EE57 }, // Lo       ARABIC MATHEMATICAL TAILED KHAH
    Interval{ 0x1EE59, 0x1EE59 }, // Lo       ARABIC MATHEMATICAL TAILED DAD
    Interval{ 0x1EE5B, 0x1EE5B }, // Lo       ARABIC MATHEMATICAL TAILED GHAIN
    Interval{ 0x1EE5D, 0x1EE5D }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS NOON
    Interval{ 0x1EE5F, 0x1EE5F }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS QAF
    Interval{ 0x1EE61, 0x1EE62 }, // Lo   [2] ARABIC MATHEMATICAL STRETCHED BEH..ARABIC MATHEMATICAL STRETCHED JEEM
    Interval{ 0x1EE64, 0x1EE64 }, // Lo       ARABIC MATHEMATICAL STRETCHED HEH
    Interval{ 0x1EE67, 0x1EE6A }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED HAH..ARABIC MATHEMATICAL STRETCHED KAF
    Interval{ 0x1EE6C, 0x1EE72 }, // Lo   [7] ARABIC MATHEMATICAL STRETCHED MEEM..ARABIC MATHEMATICAL STRETCHED QAF
    Interval{ 0x1EE74, 0x1EE77 }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED SHEEN..ARABIC MATHEMATICAL STRETCHED KHAH
    Interval{ 0x1EE79, 0x1EE7C }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED DAD..ARABIC MATHEMATICAL STRETCHED DOTLESS BEH
    Interval{ 0x1EE7E, 0x1EE7E }, // Lo       ARABIC MATHEMATICAL STRETCHED DOTLESS FEH
    Interval{ 0x1EE80, 0x1EE89 }, // Lo  [10] ARABIC MATHEMATICAL LOOPED ALEF..ARABIC MATHEMATICAL LOOPED YEH
    Interval{ 0x1EE8B, 0x1EE9B }, // Lo  [17] ARABIC MATHEMATICAL LOOPED LAM..ARABIC MATHEMATICAL LOOPED GHAIN
    Interval{ 0x1EEA1, 0x1EEA3 }, // Lo   [3] ARABIC MATHEMATICAL DOUBLE-STRUCK BEH..ARABIC MATHEMATICAL DOUBLE-STRUCK DAL
    Interval{ 0x1EEA5, 0x1EEA9 }, // Lo   [5] ARABIC MATHEMATICAL DOUBLE-STRUCK WAW..ARABIC MATHEMATICAL DOUBLE-STRUCK YEH
    Interval{ 0x1EEAB, 0x1EEBB }, // Lo  [17] ARABIC MATHEMATICAL DOUBLE-STRUCK LAM..ARABIC MATHEMATICAL DOUBLE-STRUCK GHAIN
    Interval{ 0x1FBF0, 0x1FBF9 }, // Nd  [10] SEGMENTED DIGIT ZERO..SEGMENTED DIGIT NINE
    Interval{ 0x20000, 0x2A6DF }, // Lo [42720] CJK UNIFIED IDEOGRAPH-20000..CJK UNIFIED IDEOGRAPH-2A6DF
    Interval{ 0x2A700, 0x2B81D }, // Lo [4382] CJK UNIFIED IDEOGRAPH-2A700..CJK UNIFIED IDEOGRAPH-2B81D
    Interval{ 0x2B820, 0x2CEAD }, // Lo [5774] CJK UNIFIED IDEOGRAPH-2B820..CJK UNIFIED IDEOGRAPH-2CEAD
    Interval{ 0x2CEB0, 0x2EBE0 }, // Lo [7473] CJK UNIFIED IDEOGRAPH-2CEB0..CJK UNIFIED IDEOGRAPH-2EBE0
    Interval{ 0x2EBF0, 0x2EE5D }, // Lo [622] CJK UNIFIED IDEOGRAPH-2EBF0..CJK UNIFIED IDEOGRAPH-2EE5D
    Interval{ 0x2F800, 0x2FA1D }, // Lo [542] CJK COMPATIBILITY IDEOGRAPH-2F800..CJK COMPATIBILITY IDEOGRAPH-2FA1D
    Interval{ 0x30000, 0x3134A }, // Lo [4939] CJK UNIFIED IDEOGRAPH-30000..CJK UNIFIED IDEOGRAPH-3134A
    Interval{ 0x31350, 0x33479 }, // Lo [8490] CJK UNIFIED IDEOGRAPH-31350..CJK UNIFIED IDEOGRAPH-33479
    Interval{ 0xE0100, 0xE01EF }, // Mn [240] VARIATION SELECTOR-17..VARIATION SELECTOR-256
}; // }}}
auto static const XID_Start = std::array<Interval, 779>{ // {{{
    Interval{ 0x0041, 0x005A }, // L&  [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Interval{ 0x0061, 0x007A }, // L&  [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Interval{ 0x00AA, 0x00AA }, // Lo       FEMININE ORDINAL INDICATOR
    Interval{ 0x00B5, 0x00B5 }, // L&       MICRO SIGN
    Interval{ 0x00BA, 0x00BA }, // Lo       MASCULINE ORDINAL INDICATOR
    Interval{ 0x00C0, 0x00D6 }, // L&  [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x00D8, 0x00F6 }, // L&  [31] LATIN CAPITAL LETTER O WITH STROKE..LATIN SMALL LETTER O WITH DIAERESIS
    Interval{ 0x00F8, 0x01BA }, // L& [195] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER EZH WITH TAIL
    Interval{ 0x01BB, 0x01BB }, // Lo       LATIN LETTER TWO WITH STROKE
    Interval{ 0x01BC, 0x01BF }, // L&   [4] LATIN CAPITAL LETTER TONE FIVE..LATIN LETTER WYNN
    Interval{ 0x01C0, 0x01C3 }, // Lo   [4] LATIN LETTER DENTAL CLICK..LATIN LETTER RETROFLEX CLICK
    Interval{ 0x01C4, 0x0293 }, // L& [208] LATIN CAPITAL LETTER DZ WITH CARON..LATIN SMALL LETTER EZH WITH CURL
    Interval{ 0x0294, 0x0295 }, // Lo   [2] LATIN LETTER GLOTTAL STOP..LATIN LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0x0296, 0x02AF }, // L&  [26] LATIN LETTER INVERTED GLOTTAL STOP..LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    Interval{ 0x02B0, 0x02C1 }, // Lm  [18] MODIFIER LETTER SMALL H..MODIFIER LETTER REVERSED GLOTTAL STOP
    Interval{ 0x02C6, 0x02D1 }, // Lm  [12] MODIFIER LETTER CIRCUMFLEX ACCENT..MODIFIER LETTER HALF TRIANGULAR COLON
    Interval{ 0x02E0, 0x02E4 }, // Lm   [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    Interval{ 0x02EC, 0x02EC }, // Lm       MODIFIER LETTER VOICING
    Interval{ 0x02EE, 0x02EE }, // Lm       MODIFIER LETTER DOUBLE APOSTROPHE
    Interval{ 0x0370, 0x0373 }, // L&   [4] GREEK CAPITAL LETTER HETA..GREEK SMALL LETTER ARCHAIC SAMPI
    Interval{ 0x0374, 0x0374 }, // Lm       GREEK NUMERAL SIGN
    Interval{ 0x0376, 0x0377 }, // L&   [2] GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA..GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037B, 0x037D }, // L&   [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Interval{ 0x037F, 0x037F }, // L&       GREEK CAPITAL LETTER YOT
    Interval{ 0x0386, 0x0386 }, // L&       GREEK CAPITAL LETTER ALPHA WITH TONOS
    Interval{ 0x0388, 0x038A }, // L&   [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Interval{ 0x038C, 0x038C }, // L&       GREEK CAPITAL LETTER OMICRON WITH TONOS
    Interval{ 0x038E, 0x03A1 }, // L&  [20] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER RHO
    Interval{ 0x03A3, 0x03F5 }, // L&  [83] GREEK CAPITAL LETTER SIGMA..GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F7, 0x0481 }, // L& [139] GREEK CAPITAL LETTER SHO..CYRILLIC SMALL LETTER KOPPA
    Interval{ 0x048A, 0x052F }, // L& [166] CYRILLIC CAPITAL LETTER SHORT I WITH TAIL..CYRILLIC SMALL LETTER EL WITH DESCENDER
    Interval{ 0x0531, 0x0556 }, // L&  [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Interval{ 0x0559, 0x0559 }, // Lm       ARMENIAN MODIFIER LETTER LEFT HALF RING
    Interval{ 0x0560, 0x0588 }, // L&  [41] ARMENIAN SMALL LETTER TURNED AYB..ARMENIAN SMALL LETTER YI WITH STROKE
    Interval{ 0x05D0, 0x05EA }, // Lo  [27] HEBREW LETTER ALEF..HEBREW LETTER TAV
    Interval{ 0x05EF, 0x05F2 }, // Lo   [4] HEBREW YOD TRIANGLE..HEBREW LIGATURE YIDDISH DOUBLE YOD
    Interval{ 0x0620, 0x063F }, // Lo  [32] ARABIC LETTER KASHMIRI YEH..ARABIC LETTER FARSI YEH WITH THREE DOTS ABOVE
    Interval{ 0x0640, 0x0640 }, // Lm       ARABIC TATWEEL
    Interval{ 0x0641, 0x064A }, // Lo  [10] ARABIC LETTER FEH..ARABIC LETTER YEH
    Interval{ 0x066E, 0x066F }, // Lo   [2] ARABIC LETTER DOTLESS BEH..ARABIC LETTER DOTLESS QAF
    Interval{ 0x0671, 0x06D3 }, // Lo  [99] ARABIC LETTER ALEF WASLA..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE
    Interval{ 0x06D5, 0x06D5 }, // Lo       ARABIC LETTER AE
    Interval{ 0x06E5, 0x06E6 }, // Lm   [2] ARABIC SMALL WAW..ARABIC SMALL YEH
    Interval{ 0x06EE, 0x06EF }, // Lo   [2] ARABIC LETTER DAL WITH INVERTED V..ARABIC LETTER REH WITH INVERTED V
    Interval{ 0x06FA, 0x06FC }, // Lo   [3] ARABIC LETTER SHEEN WITH DOT BELOW..ARABIC LETTER GHAIN WITH DOT BELOW
    Interval{ 0x06FF, 0x06FF }, // Lo       ARABIC LETTER HEH WITH INVERTED V
    Interval{ 0x0710, 0x0710 }, // Lo       SYRIAC LETTER ALAPH
    Interval{ 0x0712, 0x072F }, // Lo  [30] SYRIAC LETTER BETH..SYRIAC LETTER PERSIAN DHALATH
    Interval{ 0x074D, 0x07A5 }, // Lo  [89] SYRIAC LETTER SOGDIAN ZHAIN..THAANA LETTER WAAVU
    Interval{ 0x07B1, 0x07B1 }, // Lo       THAANA LETTER NAA
    Interval{ 0x07CA, 0x07EA }, // Lo  [33] NKO LETTER A..NKO LETTER JONA RA
    Interval{ 0x07F4, 0x07F5 }, // Lm   [2] NKO HIGH TONE APOSTROPHE..NKO LOW TONE APOSTROPHE
    Interval{ 0x07FA, 0x07FA }, // Lm       NKO LAJANYALAN
    Interval{ 0x0800, 0x0815 }, // Lo  [22] SAMARITAN LETTER ALAF..SAMARITAN LETTER TAAF
    Interval{ 0x081A, 0x081A }, // Lm       SAMARITAN MODIFIER LETTER EPENTHETIC YUT
    Interval{ 0x0824, 0x0824 }, // Lm       SAMARITAN MODIFIER LETTER SHORT A
    Interval{ 0x0828, 0x0828 }, // Lm       SAMARITAN MODIFIER LETTER I
    Interval{ 0x0840, 0x0858 }, // Lo  [25] MANDAIC LETTER HALQA..MANDAIC LETTER AIN
    Interval{ 0x0860, 0x086A }, // Lo  [11] SYRIAC LETTER MALAYALAM NGA..SYRIAC LETTER MALAYALAM SSA
    Interval{ 0x0870, 0x0887 }, // Lo  [24] ARABIC LETTER ALEF WITH ATTACHED FATHA..ARABIC BASELINE ROUND DOT
    Interval{ 0x0889, 0x088F }, // Lo   [7] ARABIC LETTER NOON WITH INVERTED SMALL V..ARABIC LETTER NOON WITH RING ABOVE
    Interval{ 0x08A0, 0x08C8 }, // Lo  [41] ARABIC LETTER BEH WITH SMALL V BELOW..ARABIC LETTER GRAF
    Interval{ 0x08C9, 0x08C9 }, // Lm       ARABIC SMALL FARSI YEH
    Interval{ 0x0904, 0x0939 }, // Lo  [54] DEVANAGARI LETTER SHORT A..DEVANAGARI LETTER HA
    Interval{ 0x093D, 0x093D }, // Lo       DEVANAGARI SIGN AVAGRAHA
    Interval{ 0x0950, 0x0950 }, // Lo       DEVANAGARI OM
    Interval{ 0x0958, 0x0961 }, // Lo  [10] DEVANAGARI LETTER QA..DEVANAGARI LETTER VOCALIC LL
    Interval{ 0x0971, 0x0971 }, // Lm       DEVANAGARI SIGN HIGH SPACING DOT
    Interval{ 0x0972, 0x0980 }, // Lo  [15] DEVANAGARI LETTER CANDRA A..BENGALI ANJI
    Interval{ 0x0985, 0x098C }, // Lo   [8] BENGALI LETTER A..BENGALI LETTER VOCALIC L
    Interval{ 0x098F, 0x0990 }, // Lo   [2] BENGALI LETTER E..BENGALI LETTER AI
    Interval{ 0x0993, 0x09A8 }, // Lo  [22] BENGALI LETTER O..BENGALI LETTER NA
    Interval{ 0x09AA, 0x09B0 }, // Lo   [7] BENGALI LETTER PA..BENGALI LETTER RA
    Interval{ 0x09B2, 0x09B2 }, // Lo       BENGALI LETTER LA
    Interval{ 0x09B6, 0x09B9 }, // Lo   [4] BENGALI LETTER SHA..BENGALI LETTER HA
    Interval{ 0x09BD, 0x09BD }, // Lo       BENGALI SIGN AVAGRAHA
    Interval{ 0x09CE, 0x09CE }, // Lo       BENGALI LETTER KHANDA TA
    Interval{ 0x09DC, 0x09DD }, // Lo   [2] BENGALI LETTER RRA..BENGALI LETTER RHA
    Interval{ 0x09DF, 0x09E1 }, // Lo   [3] BENGALI LETTER YYA..BENGALI LETTER VOCALIC LL
    Interval{ 0x09F0, 0x09F1 }, // Lo   [2] BENGALI LETTER RA WITH MIDDLE DIAGONAL..BENGALI LETTER RA WITH LOWER DIAGONAL
    Interval{ 0x09FC, 0x09FC }, // Lo       BENGALI LETTER VEDIC ANUSVARA
    Interval{ 0x0A05, 0x0A0A }, // Lo   [6] GURMUKHI LETTER A..GURMUKHI LETTER UU
    Interval{ 0x0A0F, 0x0A10 }, // Lo   [2] GURMUKHI LETTER EE..GURMUKHI LETTER AI
    Interval{ 0x0A13, 0x0A28 }, // Lo  [22] GURMUKHI LETTER OO..GURMUKHI LETTER NA
    Interval{ 0x0A2A, 0x0A30 }, // Lo   [7] GURMUKHI LETTER PA..GURMUKHI LETTER RA
    Interval{ 0x0A32, 0x0A33 }, // Lo   [2] GURMUKHI LETTER LA..GURMUKHI LETTER LLA
    Interval{ 0x0A35, 0x0A36 }, // Lo   [2] GURMUKHI LETTER VA..GURMUKHI LETTER SHA
    Interval{ 0x0A38, 0x0A39 }, // Lo   [2] GURMUKHI LETTER SA..GURMUKHI LETTER HA
    Interval{ 0x0A59, 0x0A5C }, // Lo   [4] GURMUKHI LETTER KHHA..GURMUKHI LETTER RRA
    Interval{ 0x0A5E, 0x0A5E }, // Lo       GURMUKHI LETTER FA
    Interval{ 0x0A72, 0x0A74 }, // Lo   [3] GURMUKHI IRI..GURMUKHI EK ONKAR
    Interval{ 0x0A85, 0x0A8D }, // Lo   [9] GUJARATI LETTER A..GUJARATI VOWEL CANDRA E
    Interval{ 0x0A8F, 0x0A91 }, // Lo   [3] GUJARATI LETTER E..GUJARATI VOWEL CANDRA O
    Interval{ 0x0A93, 0x0AA8 }, // Lo  [22] GUJARATI LETTER O..GUJARATI LETTER NA
    Interval{ 0x0AAA, 0x0AB0 }, // Lo   [7] GUJARATI LETTER PA..GUJARATI LETTER RA
    Interval{ 0x0AB2, 0x0AB3 }, // Lo   [2] GUJARATI LETTER LA..GUJARATI LETTER LLA
    Interval{ 0x0AB5, 0x0AB9 }, // Lo   [5] GUJARATI LETTER VA..GUJARATI LETTER HA
    Interval{ 0x0ABD, 0x0ABD }, // Lo       GUJARATI SIGN AVAGRAHA
    Interval{ 0x0AD0, 0x0AD0 }, // Lo       GUJARATI OM
    Interval{ 0x0AE0, 0x0AE1 }, // Lo   [2] GUJARATI LETTER VOCALIC RR..GUJARATI LETTER VOCALIC LL
    Interval{ 0x0AF9, 0x0AF9 }, // Lo       GUJARATI LETTER ZHA
    Interval{ 0x0B05, 0x0B0C }, // Lo   [8] ORIYA LETTER A..ORIYA LETTER VOCALIC L
    Interval{ 0x0B0F, 0x0B10 }, // Lo   [2] ORIYA LETTER E..ORIYA LETTER AI
    Interval{ 0x0B13, 0x0B28 }, // Lo  [22] ORIYA LETTER O..ORIYA LETTER NA
    Interval{ 0x0B2A, 0x0B30 }, // Lo   [7] ORIYA LETTER PA..ORIYA LETTER RA
    Interval{ 0x0B32, 0x0B33 }, // Lo   [2] ORIYA LETTER LA..ORIYA LETTER LLA
    Interval{ 0x0B35, 0x0B39 }, // Lo   [5] ORIYA LETTER VA..ORIYA LETTER HA
    Interval{ 0x0B3D, 0x0B3D }, // Lo       ORIYA SIGN AVAGRAHA
    Interval{ 0x0B5C, 0x0B5D }, // Lo   [2] ORIYA LETTER RRA..ORIYA LETTER RHA
    Interval{ 0x0B5F, 0x0B61 }, // Lo   [3] ORIYA LETTER YYA..ORIYA LETTER VOCALIC LL
    Interval{ 0x0B71, 0x0B71 }, // Lo       ORIYA LETTER WA
    Interval{ 0x0B83, 0x0B83 }, // Lo       TAMIL SIGN VISARGA
    Interval{ 0x0B85, 0x0B8A }, // Lo   [6] TAMIL LETTER A..TAMIL LETTER UU
    Interval{ 0x0B8E, 0x0B90 }, // Lo   [3] TAMIL LETTER E..TAMIL LETTER AI
    Interval{ 0x0B92, 0x0B95 }, // Lo   [4] TAMIL LETTER O..TAMIL LETTER KA
    Interval{ 0x0B99, 0x0B9A }, // Lo   [2] TAMIL LETTER NGA..TAMIL LETTER CA
    Interval{ 0x0B9C, 0x0B9C }, // Lo       TAMIL LETTER JA
    Interval{ 0x0B9E, 0x0B9F }, // Lo   [2] TAMIL LETTER NYA..TAMIL LETTER TTA
    Interval{ 0x0BA3, 0x0BA4 }, // Lo   [2] TAMIL LETTER NNA..TAMIL LETTER TA
    Interval{ 0x0BA8, 0x0BAA }, // Lo   [3] TAMIL LETTER NA..TAMIL LETTER PA
    Interval{ 0x0BAE, 0x0BB9 }, // Lo  [12] TAMIL LETTER MA..TAMIL LETTER HA
    Interval{ 0x0BD0, 0x0BD0 }, // Lo       TAMIL OM
    Interval{ 0x0C05, 0x0C0C }, // Lo   [8] TELUGU LETTER A..TELUGU LETTER VOCALIC L
    Interval{ 0x0C0E, 0x0C10 }, // Lo   [3] TELUGU LETTER E..TELUGU LETTER AI
    Interval{ 0x0C12, 0x0C28 }, // Lo  [23] TELUGU LETTER O..TELUGU LETTER NA
    Interval{ 0x0C2A, 0x0C39 }, // Lo  [16] TELUGU LETTER PA..TELUGU LETTER HA
    Interval{ 0x0C3D, 0x0C3D }, // Lo       TELUGU SIGN AVAGRAHA
    Interval{ 0x0C58, 0x0C5A }, // Lo   [3] TELUGU LETTER TSA..TELUGU LETTER RRRA
    Interval{ 0x0C5C, 0x0C5D }, // Lo   [2] TELUGU ARCHAIC SHRII..TELUGU LETTER NAKAARA POLLU
    Interval{ 0x0C60, 0x0C61 }, // Lo   [2] TELUGU LETTER VOCALIC RR..TELUGU LETTER VOCALIC LL
    Interval{ 0x0C80, 0x0C80 }, // Lo       KANNADA SIGN SPACING CANDRABINDU
    Interval{ 0x0C85, 0x0C8C }, // Lo   [8] KANNADA LETTER A..KANNADA LETTER VOCALIC L
    Interval{ 0x0C8E, 0x0C90 }, // Lo   [3] KANNADA LETTER E..KANNADA LETTER AI
    Interval{ 0x0C92, 0x0CA8 }, // Lo  [23] KANNADA LETTER O..KANNADA LETTER NA
    Interval{ 0x0CAA, 0x0CB3 }, // Lo  [10] KANNADA LETTER PA..KANNADA LETTER LLA
    Interval{ 0x0CB5, 0x0CB9 }, // Lo   [5] KANNADA LETTER VA..KANNADA LETTER HA
    Interval{ 0x0CBD, 0x0CBD }, // Lo       KANNADA SIGN AVAGRAHA
    Interval{ 0x0CDC, 0x0CDE }, // Lo   [3] KANNADA ARCHAIC SHRII..KANNADA LETTER FA
    Interval{ 0x0CE0, 0x0CE1 }, // Lo   [2] KANNADA LETTER VOCALIC RR..KANNADA LETTER VOCALIC LL
    Interval{ 0x0CF1, 0x0CF2 }, // Lo   [2] KANNADA SIGN JIHVAMULIYA..KANNADA SIGN UPADHMANIYA
    Interval{ 0x0D04, 0x0D0C }, // Lo   [9] MALAYALAM LETTER VEDIC ANUSVARA..MALAYALAM LETTER VOCALIC L
    Interval{ 0x0D0E, 0x0D10 }, // Lo   [3] MALAYALAM LETTER E..MALAYALAM LETTER AI
    Interval{ 0x0D12, 0x0D3A }, // Lo  [41] MALAYALAM LETTER O..MALAYALAM LETTER TTTA
    Interval{ 0x0D3D, 0x0D3D }, // Lo       MALAYALAM SIGN AVAGRAHA
    Interval{ 0x0D4E, 0x0D4E }, // Lo       MALAYALAM LETTER DOT REPH
    Interval{ 0x0D54, 0x0D56 }, // Lo   [3] MALAYALAM LETTER CHILLU M..MALAYALAM LETTER CHILLU LLL
    Interval{ 0x0D5F, 0x0D61 }, // Lo   [3] MALAYALAM LETTER ARCHAIC II..MALAYALAM LETTER VOCALIC LL
    Interval{ 0x0D7A, 0x0D7F }, // Lo   [6] MALAYALAM LETTER CHILLU NN..MALAYALAM LETTER CHILLU K
    Interval{ 0x0D85, 0x0D96 }, // Lo  [18] SINHALA LETTER AYANNA..SINHALA LETTER AUYANNA
    Interval{ 0x0D9A, 0x0DB1 }, // Lo  [24] SINHALA LETTER ALPAPRAANA KAYANNA..SINHALA LETTER DANTAJA NAYANNA
    Interval{ 0x0DB3, 0x0DBB }, // Lo   [9] SINHALA LETTER SANYAKA DAYANNA..SINHALA LETTER RAYANNA
    Interval{ 0x0DBD, 0x0DBD }, // Lo       SINHALA LETTER DANTAJA LAYANNA
    Interval{ 0x0DC0, 0x0DC6 }, // Lo   [7] SINHALA LETTER VAYANNA..SINHALA LETTER FAYANNA
    Interval{ 0x0E01, 0x0E30 }, // Lo  [48] THAI CHARACTER KO KAI..THAI CHARACTER SARA A
    Interval{ 0x0E32, 0x0E32 }, // Lo       THAI CHARACTER SARA AA
    Interval{ 0x0E40, 0x0E45 }, // Lo   [6] THAI CHARACTER SARA E..THAI CHARACTER LAKKHANGYAO
    Interval{ 0x0E46, 0x0E46 }, // Lm       THAI CHARACTER MAIYAMOK
    Interval{ 0x0E81, 0x0E82 }, // Lo   [2] LAO LETTER KO..LAO LETTER KHO SUNG
    Interval{ 0x0E84, 0x0E84 }, // Lo       LAO LETTER KHO TAM
    Interval{ 0x0E86, 0x0E8A }, // Lo   [5] LAO LETTER PALI GHA..LAO LETTER SO TAM
    Interval{ 0x0E8C, 0x0EA3 }, // Lo  [24] LAO LETTER PALI JHA..LAO LETTER LO LING
    Interval{ 0x0EA5, 0x0EA5 }, // Lo       LAO LETTER LO LOOT
    Interval{ 0x0EA7, 0x0EB0 }, // Lo  [10] LAO LETTER WO..LAO VOWEL SIGN A
    Interval{ 0x0EB2, 0x0EB2 }, // Lo       LAO VOWEL SIGN AA
    Interval{ 0x0EBD, 0x0EBD }, // Lo       LAO SEMIVOWEL SIGN NYO
    Interval{ 0x0EC0, 0x0EC4 }, // Lo   [5] LAO VOWEL SIGN E..LAO VOWEL SIGN AI
    Interval{ 0x0EC6, 0x0EC6 }, // Lm       LAO KO LA
    Interval{ 0x0EDC, 0x0EDF }, // Lo   [4] LAO HO NO..LAO LETTER KHMU NYO
    Interval{ 0x0F00, 0x0F00 }, // Lo       TIBETAN SYLLABLE OM
    Interval{ 0x0F40, 0x0F47 }, // Lo   [8] TIBETAN LETTER KA..TIBETAN LETTER JA
    Interval{ 0x0F49, 0x0F6C }, // Lo  [36] TIBETAN LETTER NYA..TIBETAN LETTER RRA
    Interval{ 0x0F88, 0x0F8C }, // Lo   [5] TIBETAN SIGN LCE TSA CAN..TIBETAN SIGN INVERTED MCHU CAN
    Interval{ 0x1000, 0x102A }, // Lo  [43] MYANMAR LETTER KA..MYANMAR LETTER AU
    Interval{ 0x103F, 0x103F }, // Lo       MYANMAR LETTER GREAT SA
    Interval{ 0x1050, 0x1055 }, // Lo   [6] MYANMAR LETTER SHA..MYANMAR LETTER VOCALIC LL
    Interval{ 0x105A, 0x105D }, // Lo   [4] MYANMAR LETTER MON NGA..MYANMAR LETTER MON BBE
    Interval{ 0x1061, 0x1061 }, // Lo       MYANMAR LETTER SGAW KAREN SHA
    Interval{ 0x1065, 0x1066 }, // Lo   [2] MYANMAR LETTER WESTERN PWO KAREN THA..MYANMAR LETTER WESTERN PWO KAREN PWA
    Interval{ 0x106E, 0x1070 }, // Lo   [3] MYANMAR LETTER EASTERN PWO KAREN NNA..MYANMAR LETTER EASTERN PWO KAREN GHWA
    Interval{ 0x1075, 0x1081 }, // Lo  [13] MYANMAR LETTER SHAN KA..MYANMAR LETTER SHAN HA
    Interval{ 0x108E, 0x108E }, // Lo       MYANMAR LETTER RUMAI PALAUNG FA
    Interval{ 0x10A0, 0x10C5 }, // L&  [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Interval{ 0x10C7, 0x10C7 }, // L&       GEORGIAN CAPITAL LETTER YN
    Interval{ 0x10CD, 0x10CD }, // L&       GEORGIAN CAPITAL LETTER AEN
    Interval{ 0x10D0, 0x10FA }, // L&  [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Interval{ 0x10FC, 0x10FC }, // Lm       MODIFIER LETTER GEORGIAN NAR
    Interval{ 0x10FD, 0x10FF }, // L&   [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Interval{ 0x1100, 0x1248 }, // Lo [329] HANGUL CHOSEONG KIYEOK..ETHIOPIC SYLLABLE QWA
    Interval{ 0x124A, 0x124D }, // Lo   [4] ETHIOPIC SYLLABLE QWI..ETHIOPIC SYLLABLE QWE
    Interval{ 0x1250, 0x1256 }, // Lo   [7] ETHIOPIC SYLLABLE QHA..ETHIOPIC SYLLABLE QHO
    Interval{ 0x1258, 0x1258 }, // Lo       ETHIOPIC SYLLABLE QHWA
    Interval{ 0x125A, 0x125D }, // Lo   [4] ETHIOPIC SYLLABLE QHWI..ETHIOPIC SYLLABLE QHWE
    Interval{ 0x1260, 0x1288 }, // Lo  [41] ETHIOPIC SYLLABLE BA..ETHIOPIC SYLLABLE XWA
    Interval{ 0x128A, 0x128D }, // Lo   [4] ETHIOPIC SYLLABLE XWI..ETHIOPIC SYLLABLE XWE
    Interval{ 0x1290, 0x12B0 }, // Lo  [33] ETHIOPIC SYLLABLE NA..ETHIOPIC SYLLABLE KWA
    Interval{ 0x12B2, 0x12B5 }, // Lo   [4] ETHIOPIC SYLLABLE KWI..ETHIOPIC SYLLABLE KWE
    Interval{ 0x12B8, 0x12BE }, // Lo   [7] ETHIOPIC SYLLABLE KXA..ETHIOPIC SYLLABLE KXO
    Interval{ 0x12C0, 0x12C0 }, // Lo       ETHIOPIC SYLLABLE KXWA
    Interval{ 0x12C2, 0x12C5 }, // Lo   [4] ETHIOPIC SYLLABLE KXWI..ETHIOPIC SYLLABLE KXWE
    Interval{ 0x12C8, 0x12D6 }, // Lo  [15] ETHIOPIC SYLLABLE WA..ETHIOPIC SYLLABLE PHARYNGEAL O
    Interval{ 0x12D8, 0x1310 }, // Lo  [57] ETHIOPIC SYLLABLE ZA..ETHIOPIC SYLLABLE GWA
    Interval{ 0x1312, 0x1315 }, // Lo   [4] ETHIOPIC SYLLABLE GWI..ETHIOPIC SYLLABLE GWE
    Interval{ 0x1318, 0x135A }, // Lo  [67] ETHIOPIC SYLLABLE GGA..ETHIOPIC SYLLABLE FYA
    Interval{ 0x1380, 0x138F }, // Lo  [16] ETHIOPIC SYLLABLE SEBATBEIT MWA..ETHIOPIC SYLLABLE PWE
    Interval{ 0x13A0, 0x13F5 }, // L&  [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Interval{ 0x13F8, 0x13FD }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1401, 0x166C }, // Lo [620] CANADIAN SYLLABICS E..CANADIAN SYLLABICS CARRIER TTSA
    Interval{ 0x166F, 0x167F }, // Lo  [17] CANADIAN SYLLABICS QAI..CANADIAN SYLLABICS BLACKFOOT W
    Interval{ 0x1681, 0x169A }, // Lo  [26] OGHAM LETTER BEITH..OGHAM LETTER PEITH
    Interval{ 0x16A0, 0x16EA }, // Lo  [75] RUNIC LETTER FEHU FEOH FE F..RUNIC LETTER X
    Interval{ 0x16EE, 0x16F0 }, // Nl   [3] RUNIC ARLAUG SYMBOL..RUNIC BELGTHOR SYMBOL
    Interval{ 0x16F1, 0x16F8 }, // Lo   [8] RUNIC LETTER K..RUNIC LETTER FRANKS CASKET AESC
    Interval{ 0x1700, 0x1711 }, // Lo  [18] TAGALOG LETTER A..TAGALOG LETTER HA
    Interval{ 0x171F, 0x1731 }, // Lo  [19] TAGALOG LETTER ARCHAIC RA..HANUNOO LETTER HA
    Interval{ 0x1740, 0x1751 }, // Lo  [18] BUHID LETTER A..BUHID LETTER HA
    Interval{ 0x1760, 0x176C }, // Lo  [13] TAGBANWA LETTER A..TAGBANWA LETTER YA
    Interval{ 0x176E, 0x1770 }, // Lo   [3] TAGBANWA LETTER LA..TAGBANWA LETTER SA
    Interval{ 0x1780, 0x17B3 }, // Lo  [52] KHMER LETTER KA..KHMER INDEPENDENT VOWEL QAU
    Interval{ 0x17D7, 0x17D7 }, // Lm       KHMER SIGN LEK TOO
    Interval{ 0x17DC, 0x17DC }, // Lo       KHMER SIGN AVAKRAHASANYA
    Interval{ 0x1820, 0x1842 }, // Lo  [35] MONGOLIAN LETTER A..MONGOLIAN LETTER CHI
    Interval{ 0x1843, 0x1843 }, // Lm       MONGOLIAN LETTER TODO LONG VOWEL SIGN
    Interval{ 0x1844, 0x1878 }, // Lo  [53] MONGOLIAN LETTER TODO E..MONGOLIAN LETTER CHA WITH TWO DOTS
    Interval{ 0x1880, 0x1884 }, // Lo   [5] MONGOLIAN LETTER ALI GALI ANUSVARA ONE..MONGOLIAN LETTER ALI GALI INVERTED UBADAMA
    Interval{ 0x1885, 0x1886 }, // Mn   [2] MONGOLIAN LETTER ALI GALI BALUDA..MONGOLIAN LETTER ALI GALI THREE BALUDA
    Interval{ 0x1887, 0x18A8 }, // Lo  [34] MONGOLIAN LETTER ALI GALI A..MONGOLIAN LETTER MANCHU ALI GALI BHA
    Interval{ 0x18AA, 0x18AA }, // Lo       MONGOLIAN LETTER MANCHU ALI GALI LHA
    Interval{ 0x18B0, 0x18F5 }, // Lo  [70] CANADIAN SYLLABICS OY..CANADIAN SYLLABICS CARRIER DENTAL S
    Interval{ 0x1900, 0x191E }, // Lo  [31] LIMBU VOWEL-CARRIER LETTER..LIMBU LETTER TRA
    Interval{ 0x1950, 0x196D }, // Lo  [30] TAI LE LETTER KA..TAI LE LETTER AI
    Interval{ 0x1970, 0x1974 }, // Lo   [5] TAI LE LETTER TONE-2..TAI LE LETTER TONE-6
    Interval{ 0x1980, 0x19AB }, // Lo  [44] NEW TAI LUE LETTER HIGH QA..NEW TAI LUE LETTER LOW SUA
    Interval{ 0x19B0, 0x19C9 }, // Lo  [26] NEW TAI LUE VOWEL SIGN VOWEL SHORTENER..NEW TAI LUE TONE MARK-2
    Interval{ 0x1A00, 0x1A16 }, // Lo  [23] BUGINESE LETTER KA..BUGINESE LETTER HA
    Interval{ 0x1A20, 0x1A54 }, // Lo  [53] TAI THAM LETTER HIGH KA..TAI THAM LETTER GREAT SA
    Interval{ 0x1AA7, 0x1AA7 }, // Lm       TAI THAM SIGN MAI YAMOK
    Interval{ 0x1B05, 0x1B33 }, // Lo  [47] BALINESE LETTER AKARA..BALINESE LETTER HA
    Interval{ 0x1B45, 0x1B4C }, // Lo   [8] BALINESE LETTER KAF SASAK..BALINESE LETTER ARCHAIC JNYA
    Interval{ 0x1B83, 0x1BA0 }, // Lo  [30] SUNDANESE LETTER A..SUNDANESE LETTER HA
    Interval{ 0x1BAE, 0x1BAF }, // Lo   [2] SUNDANESE LETTER KHA..SUNDANESE LETTER SYA
    Interval{ 0x1BBA, 0x1BE5 }, // Lo  [44] SUNDANESE AVAGRAHA..BATAK LETTER U
    Interval{ 0x1C00, 0x1C23 }, // Lo  [36] LEPCHA LETTER KA..LEPCHA LETTER A
    Interval{ 0x1C4D, 0x1C4F }, // Lo   [3] LEPCHA LETTER TTA..LEPCHA LETTER DDA
    Interval{ 0x1C5A, 0x1C77 }, // Lo  [30] OL CHIKI LETTER LA..OL CHIKI LETTER OH
    Interval{ 0x1C78, 0x1C7D }, // Lm   [6] OL CHIKI MU TTUDDAG..OL CHIKI AHAD
    Interval{ 0x1C80, 0x1C8A }, // L&  [11] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER TJE
    Interval{ 0x1C90, 0x1CBA }, // L&  [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Interval{ 0x1CBD, 0x1CBF }, // L&   [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Interval{ 0x1CE9, 0x1CEC }, // Lo   [4] VEDIC SIGN ANUSVARA ANTARGOMUKHA..VEDIC SIGN ANUSVARA VAMAGOMUKHA WITH TAIL
    Interval{ 0x1CEE, 0x1CF3 }, // Lo   [6] VEDIC SIGN HEXIFORM LONG ANUSVARA..VEDIC SIGN ROTATED ARDHAVISARGA
    Interval{ 0x1CF5, 0x1CF6 }, // Lo   [2] VEDIC SIGN JIHVAMULIYA..VEDIC SIGN UPADHMANIYA
    Interval{ 0x1CFA, 0x1CFA }, // Lo       VEDIC SIGN DOUBLE ANUSVARA ANTARGOMUKHA
    Interval{ 0x1D00, 0x1D2B }, // L&  [44] LATIN LETTER SMALL CAPITAL A..CYRILLIC LETTER SMALL CAPITAL EL
    Interval{ 0x1D2C, 0x1D6A }, // Lm  [63] MODIFIER LETTER CAPITAL A..GREEK SUBSCRIPT SMALL LETTER CHI
    Interval{ 0x1D6B, 0x1D77 }, // L&  [13] LATIN SMALL LETTER UE..LATIN SMALL LETTER TURNED G
    Interval{ 0x1D78, 0x1D78 }, // Lm       MODIFIER LETTER CYRILLIC EN
    Interval{ 0x1D79, 0x1D9A }, // L&  [34] LATIN SMALL LETTER INSULAR G..LATIN SMALL LETTER EZH WITH RETROFLEX HOOK
    Interval{ 0x1D9B, 0x1DBF }, // Lm  [37] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL THETA
    Interval{ 0x1E00, 0x1F15 }, // L& [278] LATIN CAPITAL LETTER A WITH RING BELOW..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F18, 0x1F1D }, // L&   [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F20, 0x1F45 }, // L&  [38] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F48, 0x1F4D }, // L&   [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F50, 0x1F57 }, // L&   [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F59, 0x1F59 }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA
    Interval{ 0x1F5B, 0x1F5B }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Interval{ 0x1F5D, 0x1F5D }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Interval{ 0x1F5F, 0x1F7D }, // L&  [31] GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI..GREEK SMALL LETTER OMEGA WITH OXIA
    Interval{ 0x1F80, 0x1FB4 }, // L&  [53] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB6, 0x1FBC }, // L&   [7] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Interval{ 0x1FBE, 0x1FBE }, // L&       GREEK PROSGEGRAMMENI
    Interval{ 0x1FC2, 0x1FC4 }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC6, 0x1FCC }, // L&   [7] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Interval{ 0x1FD0, 0x1FD3 }, // L&   [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Interval{ 0x1FD6, 0x1FDB }, // L&   [6] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK CAPITAL LETTER IOTA WITH OXIA
    Interval{ 0x1FE0, 0x1FEC }, // L&  [13] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Interval{ 0x1FF2, 0x1FF4 }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF6, 0x1FFC }, // L&   [7] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Interval{ 0x2071, 0x2071 }, // Lm       SUPERSCRIPT LATIN SMALL LETTER I
    Interval{ 0x207F, 0x207F }, // Lm       SUPERSCRIPT LATIN SMALL LETTER N
    Interval{ 0x2090, 0x209C }, // Lm  [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Interval{ 0x2102, 0x2102 }, // L&       DOUBLE-STRUCK CAPITAL C
    Interval{ 0x2107, 0x2107 }, // L&       EULER CONSTANT
    Interval{ 0x210A, 0x2113 }, // L&  [10] SCRIPT SMALL G..SCRIPT SMALL L
    Interval{ 0x2115, 0x2115 }, // L&       DOUBLE-STRUCK CAPITAL N
    Interval{ 0x2118, 0x2118 }, // Sm       SCRIPT CAPITAL P
    Interval{ 0x2119, 0x211D }, // L&   [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    Interval{ 0x2124, 0x2124 }, // L&       DOUBLE-STRUCK CAPITAL Z
    Interval{ 0x2126, 0x2126 }, // L&       OHM SIGN
    Interval{ 0x2128, 0x2128 }, // L&       BLACK-LETTER CAPITAL Z
    Interval{ 0x212A, 0x212D }, // L&   [4] KELVIN SIGN..BLACK-LETTER CAPITAL C
    Interval{ 0x212E, 0x212E }, // So       ESTIMATED SYMBOL
    Interval{ 0x212F, 0x2134 }, // L&   [6] SCRIPT SMALL E..SCRIPT SMALL O
    Interval{ 0x2135, 0x2138 }, // Lo   [4] ALEF SYMBOL..DALET SYMBOL
    Interval{ 0x2139, 0x2139 }, // L&       INFORMATION SOURCE
    Interval{ 0x213C, 0x213F }, // L&   [4] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK CAPITAL PI
    Interval{ 0x2145, 0x2149 }, // L&   [5] DOUBLE-STRUCK ITALIC CAPITAL D..DOUBLE-STRUCK ITALIC SMALL J
    Interval{ 0x214E, 0x214E }, // L&       TURNED SMALL F
    Interval{ 0x2160, 0x2182 }, // Nl  [35] ROMAN NUMERAL ONE..ROMAN NUMERAL TEN THOUSAND
    Interval{ 0x2183, 0x2184 }, // L&   [2] ROMAN NUMERAL REVERSED ONE HUNDRED..LATIN SMALL LETTER REVERSED C
    Interval{ 0x2185, 0x2188 }, // Nl   [4] ROMAN NUMERAL SIX LATE FORM..ROMAN NUMERAL ONE HUNDRED THOUSAND
    Interval{ 0x2C00, 0x2C7B }, // L& [124] GLAGOLITIC CAPITAL LETTER AZU..LATIN LETTER SMALL CAPITAL TURNED E
    Interval{ 0x2C7C, 0x2C7D }, // Lm   [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    Interval{ 0x2C7E, 0x2CE4 }, // L& [103] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC SYMBOL KAI
    Interval{ 0x2CEB, 0x2CEE }, // L&   [4] COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI..COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF2, 0x2CF3 }, // L&   [2] COPTIC CAPITAL LETTER BOHAIRIC KHEI..COPTIC SMALL LETTER BOHAIRIC KHEI
    Interval{ 0x2D00, 0x2D25 }, // L&  [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Interval{ 0x2D27, 0x2D27 }, // L&       GEORGIAN SMALL LETTER YN
    Interval{ 0x2D2D, 0x2D2D }, // L&       GEORGIAN SMALL LETTER AEN
    Interval{ 0x2D30, 0x2D67 }, // Lo  [56] TIFINAGH LETTER YA..TIFINAGH LETTER YO
    Interval{ 0x2D6F, 0x2D6F }, // Lm       TIFINAGH MODIFIER LETTER LABIALIZATION MARK
    Interval{ 0x2D80, 0x2D96 }, // Lo  [23] ETHIOPIC SYLLABLE LOA..ETHIOPIC SYLLABLE GGWE
    Interval{ 0x2DA0, 0x2DA6 }, // Lo   [7] ETHIOPIC SYLLABLE SSA..ETHIOPIC SYLLABLE SSO
    Interval{ 0x2DA8, 0x2DAE }, // Lo   [7] ETHIOPIC SYLLABLE CCA..ETHIOPIC SYLLABLE CCO
    Interval{ 0x2DB0, 0x2DB6 }, // Lo   [7] ETHIOPIC SYLLABLE ZZA..ETHIOPIC SYLLABLE ZZO
    Interval{ 0x2DB8, 0x2DBE }, // Lo   [7] ETHIOPIC SYLLABLE CCHA..ETHIOPIC SYLLABLE CCHO
    Interval{ 0x2DC0, 0x2DC6 }, // Lo   [7] ETHIOPIC SYLLABLE QYA..ETHIOPIC SYLLABLE QYO
    Interval{ 0x2DC8, 0x2DCE }, // Lo   [7] ETHIOPIC SYLLABLE KYA..ETHIOPIC SYLLABLE KYO
    Interval{ 0x2DD0, 0x2DD6 }, // Lo   [7] ETHIOPIC SYLLABLE XYA..ETHIOPIC SYLLABLE XYO
    Interval{ 0x2DD8, 0x2DDE }, // Lo   [7] ETHIOPIC SYLLABLE GYA..ETHIOPIC SYLLABLE GYO
    Interval{ 0x3005, 0x3005 }, // Lm       IDEOGRAPHIC ITERATION MARK
    Interval{ 0x3006, 0x3006 }, // Lo       IDEOGRAPHIC CLOSING MARK
    Interval{ 0x3007, 0x3007 }, // Nl       IDEOGRAPHIC NUMBER ZERO
    Interval{ 0x3021, 0x3029 }, // Nl   [9] HANGZHOU NUMERAL ONE..HANGZHOU NUMERAL NINE
    Interval{ 0x3031, 0x3035 }, // Lm   [5] VERTICAL KANA REPEAT MARK..VERTICAL KANA REPEAT MARK LOWER HALF
    Interval{ 0x3038, 0x303A }, // Nl   [3] HANGZHOU NUMERAL TEN..HANGZHOU NUMERAL THIRTY
    Interval{ 0x303B, 0x303B }, // Lm       VERTICAL IDEOGRAPHIC ITERATION MARK
    Interval{ 0x303C, 0x303C }, // Lo       MASU MARK
    Interval{ 0x3041, 0x3096 }, // Lo  [86] HIRAGANA LETTER SMALL A..HIRAGANA LETTER SMALL KE
    Interval{ 0x309D, 0x309E }, // Lm   [2] HIRAGANA ITERATION MARK..HIRAGANA VOICED ITERATION MARK
    Interval{ 0x309F, 0x309F }, // Lo       HIRAGANA DIGRAPH YORI
    Interval{ 0x30A1, 0x30FA }, // Lo  [90] KATAKANA LETTER SMALL A..KATAKANA LETTER VO
    Interval{ 0x30FC, 0x30FE }, // Lm   [3] KATAKANA-HIRAGANA PROLONGED SOUND MARK..KATAKANA VOICED ITERATION MARK
    Interval{ 0x30FF, 0x30FF }, // Lo       KATAKANA DIGRAPH KOTO
    Interval{ 0x3105, 0x312F }, // Lo  [43] BOPOMOFO LETTER B..BOPOMOFO LETTER NN
    Interval{ 0x3131, 0x318E }, // Lo  [94] HANGUL LETTER KIYEOK..HANGUL LETTER ARAEAE
    Interval{ 0x31A0, 0x31BF }, // Lo  [32] BOPOMOFO LETTER BU..BOPOMOFO LETTER AH
    Interval{ 0x31F0, 0x31FF }, // Lo  [16] KATAKANA LETTER SMALL KU..KATAKANA LETTER SMALL RO
    Interval{ 0x3400, 0x4DBF }, // Lo [6592] CJK UNIFIED IDEOGRAPH-3400..CJK UNIFIED IDEOGRAPH-4DBF
    Interval{ 0x4E00, 0xA014 }, // Lo [21013] CJK UNIFIED IDEOGRAPH-4E00..YI SYLLABLE E
    Interval{ 0xA015, 0xA015 }, // Lm       YI SYLLABLE WU
    Interval{ 0xA016, 0xA48C }, // Lo [1143] YI SYLLABLE BIT..YI SYLLABLE YYR
    Interval{ 0xA4D0, 0xA4F7 }, // Lo  [40] LISU LETTER BA..LISU LETTER OE
    Interval{ 0xA4F8, 0xA4FD }, // Lm   [6] LISU LETTER TONE MYA TI..LISU LETTER TONE MYA JEU
    Interval{ 0xA500, 0xA60B }, // Lo [268] VAI SYLLABLE EE..VAI SYLLABLE NG
    Interval{ 0xA60C, 0xA60C }, // Lm       VAI SYLLABLE LENGTHENER
    Interval{ 0xA610, 0xA61F }, // Lo  [16] VAI SYLLABLE NDOLE FA..VAI SYMBOL JONG
    Interval{ 0xA62A, 0xA62B }, // Lo   [2] VAI SYLLABLE NDOLE MA..VAI SYLLABLE NDOLE DO
    Interval{ 0xA640, 0xA66D }, // L&  [46] CYRILLIC CAPITAL LETTER ZEMLYA..CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA66E, 0xA66E }, // Lo       CYRILLIC LETTER MULTIOCULAR O
    Interval{ 0xA67F, 0xA67F }, // Lm       CYRILLIC PAYEROK
    Interval{ 0xA680, 0xA69B }, // L&  [28] CYRILLIC CAPITAL LETTER DWE..CYRILLIC SMALL LETTER CROSSED O
    Interval{ 0xA69C, 0xA69D }, // Lm   [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    Interval{ 0xA6A0, 0xA6E5 }, // Lo  [70] BAMUM LETTER A..BAMUM LETTER KI
    Interval{ 0xA6E6, 0xA6EF }, // Nl  [10] BAMUM LETTER MO..BAMUM LETTER KOGHOM
    Interval{ 0xA717, 0xA71F }, // Lm   [9] MODIFIER LETTER DOT VERTICAL BAR..MODIFIER LETTER LOW INVERTED EXCLAMATION MARK
    Interval{ 0xA722, 0xA76F }, // L&  [78] LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF..LATIN SMALL LETTER CON
    Interval{ 0xA770, 0xA770 }, // Lm       MODIFIER LETTER US
    Interval{ 0xA771, 0xA787 }, // L&  [23] LATIN SMALL LETTER DUM..LATIN SMALL LETTER INSULAR T
    Interval{ 0xA788, 0xA788 }, // Lm       MODIFIER LETTER LOW CIRCUMFLEX ACCENT
    Interval{ 0xA78B, 0xA78E }, // L&   [4] LATIN CAPITAL LETTER SALTILLO..LATIN SMALL LETTER L WITH RETROFLEX HOOK AND BELT
    Interval{ 0xA78F, 0xA78F }, // Lo       LATIN LETTER SINOLOGICAL DOT
    Interval{ 0xA790, 0xA7DC }, // L&  [77] LATIN CAPITAL LETTER N WITH DESCENDER..LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Interval{ 0xA7F1, 0xA7F4 }, // Lm   [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    Interval{ 0xA7F5, 0xA7F6 }, // L&   [2] LATIN CAPITAL LETTER REVERSED HALF H..LATIN SMALL LETTER REVERSED HALF H
    Interval{ 0xA7F7, 0xA7F7 }, // Lo       LATIN EPIGRAPHIC LETTER SIDEWAYS I
    Interval{ 0xA7F8, 0xA7F9 }, // Lm   [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    Interval{ 0xA7FA, 0xA7FA }, // L&       LATIN LETTER SMALL CAPITAL TURNED M
    Interval{ 0xA7FB, 0xA801 }, // Lo   [7] LATIN EPIGRAPHIC LETTER REVERSED F..SYLOTI NAGRI LETTER I
    Interval{ 0xA803, 0xA805 }, // Lo   [3] SYLOTI NAGRI LETTER U..SYLOTI NAGRI LETTER O
    Interval{ 0xA807, 0xA80A }, // Lo   [4] SYLOTI NAGRI LETTER KO..SYLOTI NAGRI LETTER GHO
    Interval{ 0xA80C, 0xA822 }, // Lo  [23] SYLOTI NAGRI LETTER CO..SYLOTI NAGRI LETTER HO
    Interval{ 0xA840, 0xA873 }, // Lo  [52] PHAGS-PA LETTER KA..PHAGS-PA LETTER CANDRABINDU
    Interval{ 0xA882, 0xA8B3 }, // Lo  [50] SAURASHTRA LETTER A..SAURASHTRA LETTER LLA
    Interval{ 0xA8F2, 0xA8F7 }, // Lo   [6] DEVANAGARI SIGN SPACING CANDRABINDU..DEVANAGARI SIGN CANDRABINDU AVAGRAHA
    Interval{ 0xA8FB, 0xA8FB }, // Lo       DEVANAGARI HEADSTROKE
    Interval{ 0xA8FD, 0xA8FE }, // Lo   [2] DEVANAGARI JAIN OM..DEVANAGARI LETTER AY
    Interval{ 0xA90A, 0xA925 }, // Lo  [28] KAYAH LI LETTER KA..KAYAH LI LETTER OO
    Interval{ 0xA930, 0xA946 }, // Lo  [23] REJANG LETTER KA..REJANG LETTER A
    Interval{ 0xA960, 0xA97C }, // Lo  [29] HANGUL CHOSEONG TIKEUT-MIEUM..HANGUL CHOSEONG SSANGYEORINHIEUH
    Interval{ 0xA984, 0xA9B2 }, // Lo  [47] JAVANESE LETTER A..JAVANESE LETTER HA
    Interval{ 0xA9CF, 0xA9CF }, // Lm       JAVANESE PANGRANGKEP
    Interval{ 0xA9E0, 0xA9E4 }, // Lo   [5] MYANMAR LETTER SHAN GHA..MYANMAR LETTER SHAN BHA
    Interval{ 0xA9E6, 0xA9E6 }, // Lm       MYANMAR MODIFIER LETTER SHAN REDUPLICATION
    Interval{ 0xA9E7, 0xA9EF }, // Lo   [9] MYANMAR LETTER TAI LAING NYA..MYANMAR LETTER TAI LAING NNA
    Interval{ 0xA9FA, 0xA9FE }, // Lo   [5] MYANMAR LETTER TAI LAING LLA..MYANMAR LETTER TAI LAING BHA
    Interval{ 0xAA00, 0xAA28 }, // Lo  [41] CHAM LETTER A..CHAM LETTER HA
    Interval{ 0xAA40, 0xAA42 }, // Lo   [3] CHAM LETTER FINAL K..CHAM LETTER FINAL NG
    Interval{ 0xAA44, 0xAA4B }, // Lo   [8] CHAM LETTER FINAL CH..CHAM LETTER FINAL SS
    Interval{ 0xAA60, 0xAA6F }, // Lo  [16] MYANMAR LETTER KHAMTI GA..MYANMAR LETTER KHAMTI FA
    Interval{ 0xAA70, 0xAA70 }, // Lm       MYANMAR MODIFIER LETTER KHAMTI REDUPLICATION
    Interval{ 0xAA71, 0xAA76 }, // Lo   [6] MYANMAR LETTER KHAMTI XA..MYANMAR LOGOGRAM KHAMTI HM
    Interval{ 0xAA7A, 0xAA7A }, // Lo       MYANMAR LETTER AITON RA
    Interval{ 0xAA7E, 0xAAAF }, // Lo  [50] MYANMAR LETTER SHWE PALAUNG CHA..TAI VIET LETTER HIGH O
    Interval{ 0xAAB1, 0xAAB1 }, // Lo       TAI VIET VOWEL AA
    Interval{ 0xAAB5, 0xAAB6 }, // Lo   [2] TAI VIET VOWEL E..TAI VIET VOWEL O
    Interval{ 0xAAB9, 0xAABD }, // Lo   [5] TAI VIET VOWEL UEA..TAI VIET VOWEL AN
    Interval{ 0xAAC0, 0xAAC0 }, // Lo       TAI VIET TONE MAI NUENG
    Interval{ 0xAAC2, 0xAAC2 }, // Lo       TAI VIET TONE MAI SONG
    Interval{ 0xAADB, 0xAADC }, // Lo   [2] TAI VIET SYMBOL KON..TAI VIET SYMBOL NUENG
    Interval{ 0xAADD, 0xAADD }, // Lm       TAI VIET SYMBOL SAM
    Interval{ 0xAAE0, 0xAAEA }, // Lo  [11] MEETEI MAYEK LETTER E..MEETEI MAYEK LETTER SSA
    Interval{ 0xAAF2, 0xAAF2 }, // Lo       MEETEI MAYEK ANJI
    Interval{ 0xAAF3, 0xAAF4 }, // Lm   [2] MEETEI MAYEK SYLLABLE REPETITION MARK..MEETEI MAYEK WORD REPETITION MARK
    Interval{ 0xAB01, 0xAB06 }, // Lo   [6] ETHIOPIC SYLLABLE TTHU..ETHIOPIC SYLLABLE TTHO
    Interval{ 0xAB09, 0xAB0E }, // Lo   [6] ETHIOPIC SYLLABLE DDHU..ETHIOPIC SYLLABLE DDHO
    Interval{ 0xAB11, 0xAB16 }, // Lo   [6] ETHIOPIC SYLLABLE DZU..ETHIOPIC SYLLABLE DZO
    Interval{ 0xAB20, 0xAB26 }, // Lo   [7] ETHIOPIC SYLLABLE CCHHA..ETHIOPIC SYLLABLE CCHHO
    Interval{ 0xAB28, 0xAB2E }, // Lo   [7] ETHIOPIC SYLLABLE BBA..ETHIOPIC SYLLABLE BBO
    Interval{ 0xAB30, 0xAB5A }, // L&  [43] LATIN SMALL LETTER BARRED ALPHA..LATIN SMALL LETTER Y WITH SHORT RIGHT LEG
    Interval{ 0xAB5C, 0xAB5F }, // Lm   [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    Interval{ 0xAB60, 0xAB68 }, // L&   [9] LATIN SMALL LETTER SAKHA YAT..LATIN SMALL LETTER TURNED R WITH MIDDLE TILDE
    Interval{ 0xAB69, 0xAB69 }, // Lm       MODIFIER LETTER SMALL TURNED W
    Interval{ 0xAB70, 0xABBF }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xABC0, 0xABE2 }, // Lo  [35] MEETEI MAYEK LETTER KOK..MEETEI MAYEK LETTER I LONSUM
    Interval{ 0xAC00, 0xD7A3 }, // Lo [11172] HANGUL SYLLABLE GA..HANGUL SYLLABLE HIH
    Interval{ 0xD7B0, 0xD7C6 }, // Lo  [23] HANGUL JUNGSEONG O-YEO..HANGUL JUNGSEONG ARAEA-E
    Interval{ 0xD7CB, 0xD7FB }, // Lo  [49] HANGUL JONGSEONG NIEUN-RIEUL..HANGUL JONGSEONG PHIEUPH-THIEUTH
    Interval{ 0xF900, 0xFA6D }, // Lo [366] CJK COMPATIBILITY IDEOGRAPH-F900..CJK COMPATIBILITY IDEOGRAPH-FA6D
    Interval{ 0xFA70, 0xFAD9 }, // Lo [106] CJK COMPATIBILITY IDEOGRAPH-FA70..CJK COMPATIBILITY IDEOGRAPH-FAD9
    Interval{ 0xFB00, 0xFB06 }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFB1D, 0xFB1D }, // Lo       HEBREW LETTER YOD WITH HIRIQ
    Interval{ 0xFB1F, 0xFB28 }, // Lo  [10] HEBREW LIGATURE YIDDISH YOD YOD PATAH..HEBREW LETTER WIDE TAV
    Interval{ 0xFB2A, 0xFB36 }, // Lo  [13] HEBREW LETTER SHIN WITH SHIN DOT..HEBREW LETTER ZAYIN WITH DAGESH
    Interval{ 0xFB38, 0xFB3C }, // Lo   [5] HEBREW LETTER TET WITH DAGESH..HEBREW LETTER LAMED WITH DAGESH
    Interval{ 0xFB3E, 0xFB3E }, // Lo       HEBREW LETTER MEM WITH DAGESH
    Interval{ 0xFB40, 0xFB41 }, // Lo   [2] HEBREW LETTER NUN WITH DAGESH..HEBREW LETTER SAMEKH WITH DAGESH
    Interval{ 0xFB43, 0xFB44 }, // Lo   [2] HEBREW LETTER FINAL PE WITH DAGESH..HEBREW LETTER PE WITH DAGESH
    Interval{ 0xFB46, 0xFBB1 }, // Lo [108] HEBREW LETTER TSADI WITH DAGESH..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE FINAL FORM
    Interval{ 0xFBD3, 0xFC5D }, // Lo [139] ARABIC LETTER NG ISOLATED FORM..ARABIC LIGATURE ALEF MAKSURA WITH SUPERSCRIPT ALEF ISOLATED FORM
    Interval{ 0xFC64, 0xFD3D }, // Lo [218] ARABIC LIGATURE YEH WITH HAMZA ABOVE WITH REH FINAL FORM..ARABIC LIGATURE ALEF WITH FATHATAN ISOLATED FORM
    Interval{ 0xFD50, 0xFD8F }, // Lo  [64] ARABIC LIGATURE TEH WITH JEEM WITH MEEM INITIAL FORM..ARABIC LIGATURE MEEM WITH KHAH WITH MEEM INITIAL FORM
    Interval{ 0xFD92, 0xFDC7 }, // Lo  [54] ARABIC LIGATURE MEEM WITH JEEM WITH KHAH INITIAL FORM..ARABIC LIGATURE NOON WITH JEEM WITH YEH FINAL FORM
    Interval{ 0xFDF0, 0xFDF9 }, // Lo  [10] ARABIC LIGATURE SALLA USED AS KORANIC STOP SIGN ISOLATED FORM..ARABIC LIGATURE SALLA ISOLATED FORM
    Interval{ 0xFE71, 0xFE71 }, // Lo       ARABIC TATWEEL WITH FATHATAN ABOVE
    Interval{ 0xFE73, 0xFE73 }, // Lo       ARABIC TAIL FRAGMENT
    Interval{ 0xFE77, 0xFE77 }, // Lo       ARABIC FATHA MEDIAL FORM
    Interval{ 0xFE79, 0xFE79 }, // Lo       ARABIC DAMMA MEDIAL FORM
    Interval{ 0xFE7B, 0xFE7B }, // Lo       ARABIC KASRA MEDIAL FORM
    Interval{ 0xFE7D, 0xFE7D }, // Lo       ARABIC SHADDA MEDIAL FORM
    Interval{ 0xFE7F, 0xFEFC }, // Lo [126] ARABIC SUKUN MEDIAL FORM..ARABIC LIGATURE LAM WITH ALEF FINAL FORM
    Interval{ 0xFF21, 0xFF3A }, // L&  [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Interval{ 0xFF41, 0xFF5A }, // L&  [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Interval{ 0xFF66, 0xFF6F }, // Lo  [10] HALFWIDTH KATAKANA LETTER WO..HALFWIDTH KATAKANA LETTER SMALL TU
    Interval{ 0xFF70, 0xFF70 }, // Lm       HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Interval{ 0xFF71, 0xFF9D }, // Lo  [45] HALFWIDTH KATAKANA LETTER A..HALFWIDTH KATAKANA LETTER N
    Interval{ 0xFFA0, 0xFFBE }, // Lo  [31] HALFWIDTH HANGUL FILLER..HALFWIDTH HANGUL LETTER HIEUH
    Interval{ 0xFFC2, 0xFFC7 }, // Lo   [6] HALFWIDTH HANGUL LETTER A..HALFWIDTH HANGUL LETTER E
    Interval{ 0xFFCA, 0xFFCF }, // Lo   [6] HALFWIDTH HANGUL LETTER YEO..HALFWIDTH HANGUL LETTER OE
    Interval{ 0xFFD2, 0xFFD7 }, // Lo   [6] HALFWIDTH HANGUL LETTER YO..HALFWIDTH HANGUL LETTER YU
    Interval{ 0xFFDA, 0xFFDC }, // Lo   [3] HALFWIDTH HANGUL LETTER EU..HALFWIDTH HANGUL LETTER I
    Interval{ 0x10000, 0x1000B }, // Lo  [12] LINEAR B SYLLABLE B008 A..LINEAR B SYLLABLE B046 JE
    Interval{ 0x1000D, 0x10026 }, // Lo  [26] LINEAR B SYLLABLE B036 JO..LINEAR B SYLLABLE B032 QO
    Interval{ 0x10028, 0x1003A }, // Lo  [19] LINEAR B SYLLABLE B060 RA..LINEAR B SYLLABLE B042 WO
    Interval{ 0x1003C, 0x1003D }, // Lo   [2] LINEAR B SYLLABLE B017 ZA..LINEAR B SYLLABLE B074 ZE
    Interval{ 0x1003F, 0x1004D }, // Lo  [15] LINEAR B SYLLABLE B020 ZO..LINEAR B SYLLABLE B091 TWO
    Interval{ 0x10050, 0x1005D }, // Lo  [14] LINEAR B SYMBOL B018..LINEAR B SYMBOL B089
    Interval{ 0x10080, 0x100FA }, // Lo [123] LINEAR B IDEOGRAM B100 MAN..LINEAR B IDEOGRAM VESSEL B305
    Interval{ 0x10140, 0x10174 }, // Nl  [53] GREEK ACROPHONIC ATTIC ONE QUARTER..GREEK ACROPHONIC STRATIAN FIFTY MNAS
    Interval{ 0x10280, 0x1029C }, // Lo  [29] LYCIAN LETTER A..LYCIAN LETTER X
    Interval{ 0x102A0, 0x102D0 }, // Lo  [49] CARIAN LETTER A..CARIAN LETTER UUU3
    Interval{ 0x10300, 0x1031F }, // Lo  [32] OLD ITALIC LETTER A..OLD ITALIC LETTER ESS
    Interval{ 0x1032D, 0x10340 }, // Lo  [20] OLD ITALIC LETTER YE..GOTHIC LETTER PAIRTHRA
    Interval{ 0x10341, 0x10341 }, // Nl       GOTHIC LETTER NINETY
    Interval{ 0x10342, 0x10349 }, // Lo   [8] GOTHIC LETTER RAIDA..GOTHIC LETTER OTHAL
    Interval{ 0x1034A, 0x1034A }, // Nl       GOTHIC LETTER NINE HUNDRED
    Interval{ 0x10350, 0x10375 }, // Lo  [38] OLD PERMIC LETTER AN..OLD PERMIC LETTER IA
    Interval{ 0x10380, 0x1039D }, // Lo  [30] UGARITIC LETTER ALPA..UGARITIC LETTER SSU
    Interval{ 0x103A0, 0x103C3 }, // Lo  [36] OLD PERSIAN SIGN A..OLD PERSIAN SIGN HA
    Interval{ 0x103C8, 0x103CF }, // Lo   [8] OLD PERSIAN SIGN AURAMAZDAA..OLD PERSIAN SIGN BUUMISH
    Interval{ 0x103D1, 0x103D5 }, // Nl   [5] OLD PERSIAN NUMBER ONE..OLD PERSIAN NUMBER HUNDRED
    Interval{ 0x10400, 0x1044F }, // L&  [80] DESERET CAPITAL LETTER LONG I..DESERET SMALL LETTER EW
    Interval{ 0x10450, 0x1049D }, // Lo  [78] SHAVIAN LETTER PEEP..OSMANYA LETTER OO
    Interval{ 0x104B0, 0x104D3 }, // L&  [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Interval{ 0x104D8, 0x104FB }, // L&  [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Interval{ 0x10500, 0x10527 }, // Lo  [40] ELBASAN LETTER A..ELBASAN LETTER KHE
    Interval{ 0x10530, 0x10563 }, // Lo  [52] CAUCASIAN ALBANIAN LETTER ALT..CAUCASIAN ALBANIAN LETTER KIW
    Interval{ 0x10570, 0x1057A }, // L&  [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Interval{ 0x1057C, 0x1058A }, // L&  [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Interval{ 0x1058C, 0x10592 }, // L&   [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Interval{ 0x10594, 0x10595 }, // L&   [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Interval{ 0x10597, 0x105A1 }, // L&  [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Interval{ 0x105A3, 0x105B1 }, // L&  [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Interval{ 0x105B3, 0x105B9 }, // L&   [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Interval{ 0x105BB, 0x105BC }, // L&   [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Interval{ 0x105C0, 0x105F3 }, // Lo  [52] TODHRI LETTER A..TODHRI LETTER OO
    Interval{ 0x10600, 0x10736 }, // Lo [311] LINEAR A SIGN AB001..LINEAR A SIGN A664
    Interval{ 0x10740, 0x10755 }, // Lo  [22] LINEAR A SIGN A701 A..LINEAR A SIGN A732 JE
    Interval{ 0x10760, 0x10767 }, // Lo   [8] LINEAR A SIGN A800..LINEAR A SIGN A807
    Interval{ 0x10780, 0x10785 }, // Lm   [6] MODIFIER LETTER SMALL CAPITAL AA..MODIFIER LETTER SMALL B WITH HOOK
    Interval{ 0x10787, 0x107B0 }, // Lm  [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Interval{ 0x107B2, 0x107BA }, // Lm   [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Interval{ 0x10800, 0x10805 }, // Lo   [6] CYPRIOT SYLLABLE A..CYPRIOT SYLLABLE JA
    Interval{ 0x10808, 0x10808 }, // Lo       CYPRIOT SYLLABLE JO
    Interval{ 0x1080A, 0x10835 }, // Lo  [44] CYPRIOT SYLLABLE KA..CYPRIOT SYLLABLE WO
    Interval{ 0x10837, 0x10838 }, // Lo   [2] CYPRIOT SYLLABLE XA..CYPRIOT SYLLABLE XE
    Interval{ 0x1083C, 0x1083C }, // Lo       CYPRIOT SYLLABLE ZA
    Interval{ 0x1083F, 0x10855 }, // Lo  [23] CYPRIOT SYLLABLE ZO..IMPERIAL ARAMAIC LETTER TAW
    Interval{ 0x10860, 0x10876 }, // Lo  [23] PALMYRENE LETTER ALEPH..PALMYRENE LETTER TAW
    Interval{ 0x10880, 0x1089E }, // Lo  [31] NABATAEAN LETTER FINAL ALEPH..NABATAEAN LETTER TAW
    Interval{ 0x108E0, 0x108F2 }, // Lo  [19] HATRAN LETTER ALEPH..HATRAN LETTER QOPH
    Interval{ 0x108F4, 0x108F5 }, // Lo   [2] HATRAN LETTER SHIN..HATRAN LETTER TAW
    Interval{ 0x10900, 0x10915 }, // Lo  [22] PHOENICIAN LETTER ALF..PHOENICIAN LETTER TAU
    Interval{ 0x10920, 0x10939 }, // Lo  [26] LYDIAN LETTER A..LYDIAN LETTER C
    Interval{ 0x10940, 0x10959 }, // Lo  [26] SIDETIC LETTER N01..SIDETIC LETTER N26
    Interval{ 0x10980, 0x109B7 }, // Lo  [56] MEROITIC HIEROGLYPHIC LETTER A..MEROITIC CURSIVE LETTER DA
    Interval{ 0x109BE, 0x109BF }, // Lo   [2] MEROITIC CURSIVE LOGOGRAM RMT..MEROITIC CURSIVE LOGOGRAM IMN
    Interval{ 0x10A00, 0x10A00 }, // Lo       KHAROSHTHI LETTER A
    Interval{ 0x10A10, 0x10A13 }, // Lo   [4] KHAROSHTHI LETTER KA..KHAROSHTHI LETTER GHA
    Interval{ 0x10A15, 0x10A17 }, // Lo   [3] KHAROSHTHI LETTER CA..KHAROSHTHI LETTER JA
    Interval{ 0x10A19, 0x10A35 }, // Lo  [29] KHAROSHTHI LETTER NYA..KHAROSHTHI LETTER VHA
    Interval{ 0x10A60, 0x10A7C }, // Lo  [29] OLD SOUTH ARABIAN LETTER HE..OLD SOUTH ARABIAN LETTER THETH
    Interval{ 0x10A80, 0x10A9C }, // Lo  [29] OLD NORTH ARABIAN LETTER HEH..OLD NORTH ARABIAN LETTER ZAH
    Interval{ 0x10AC0, 0x10AC7 }, // Lo   [8] MANICHAEAN LETTER ALEPH..MANICHAEAN LETTER WAW
    Interval{ 0x10AC9, 0x10AE4 }, // Lo  [28] MANICHAEAN LETTER ZAYIN..MANICHAEAN LETTER TAW
    Interval{ 0x10B00, 0x10B35 }, // Lo  [54] AVESTAN LETTER A..AVESTAN LETTER HE
    Interval{ 0x10B40, 0x10B55 }, // Lo  [22] INSCRIPTIONAL PARTHIAN LETTER ALEPH..INSCRIPTIONAL PARTHIAN LETTER TAW
    Interval{ 0x10B60, 0x10B72 }, // Lo  [19] INSCRIPTIONAL PAHLAVI LETTER ALEPH..INSCRIPTIONAL PAHLAVI LETTER TAW
    Interval{ 0x10B80, 0x10B91 }, // Lo  [18] PSALTER PAHLAVI LETTER ALEPH..PSALTER PAHLAVI LETTER TAW
    Interval{ 0x10C00, 0x10C48 }, // Lo  [73] OLD TURKIC LETTER ORKHON A..OLD TURKIC LETTER ORKHON BASH
    Interval{ 0x10C80, 0x10CB2 }, // L&  [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Interval{ 0x10CC0, 0x10CF2 }, // L&  [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Interval{ 0x10D00, 0x10D23 }, // Lo  [36] HANIFI ROHINGYA LETTER A..HANIFI ROHINGYA MARK NA KHONNA
    Interval{ 0x10D4A, 0x10D4D }, // Lo   [4] GARAY VOWEL SIGN A..GARAY VOWEL SIGN EE
    Interval{ 0x10D4E, 0x10D4E }, // Lm       GARAY VOWEL LENGTH MARK
    Interval{ 0x10D4F, 0x10D4F }, // Lo       GARAY SUKUN
    Interval{ 0x10D50, 0x10D65 }, // L&  [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Interval{ 0x10D6F, 0x10D6F }, // Lm       GARAY REDUPLICATION MARK
    Interval{ 0x10D70, 0x10D85 }, // L&  [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Interval{ 0x10E80, 0x10EA9 }, // Lo  [42] YEZIDI LETTER ELIF..YEZIDI LETTER ET
    Interval{ 0x10EB0, 0x10EB1 }, // Lo   [2] YEZIDI LETTER LAM WITH DOT ABOVE..YEZIDI LETTER YOT WITH CIRCUMFLEX ABOVE
    Interval{ 0x10EC2, 0x10EC4 }, // Lo   [3] ARABIC LETTER DAL WITH TWO DOTS VERTICALLY BELOW..ARABIC LETTER KAF WITH TWO DOTS VERTICALLY BELOW
    Interval{ 0x10EC5, 0x10EC5 }, // Lm       ARABIC SMALL YEH BARREE WITH TWO DOTS BELOW
    Interval{ 0x10EC6, 0x10EC7 }, // Lo   [2] ARABIC LETTER THIN NOON..ARABIC LETTER YEH WITH FOUR DOTS BELOW
    Interval{ 0x10F00, 0x10F1C }, // Lo  [29] OLD SOGDIAN LETTER ALEPH..OLD SOGDIAN LETTER FINAL TAW WITH VERTICAL TAIL
    Interval{ 0x10F27, 0x10F27 }, // Lo       OLD SOGDIAN LIGATURE AYIN-DALETH
    Interval{ 0x10F30, 0x10F45 }, // Lo  [22] SOGDIAN LETTER ALEPH..SOGDIAN INDEPENDENT SHIN
    Interval{ 0x10F70, 0x10F81 }, // Lo  [18] OLD UYGHUR LETTER ALEPH..OLD UYGHUR LETTER LESH
    Interval{ 0x10FB0, 0x10FC4 }, // Lo  [21] CHORASMIAN LETTER ALEPH..CHORASMIAN LETTER TAW
    Interval{ 0x10FE0, 0x10FF6 }, // Lo  [23] ELYMAIC LETTER ALEPH..ELYMAIC LIGATURE ZAYIN-YODH
    Interval{ 0x11003, 0x11037 }, // Lo  [53] BRAHMI SIGN JIHVAMULIYA..BRAHMI LETTER OLD TAMIL NNNA
    Interval{ 0x11071, 0x11072 }, // Lo   [2] BRAHMI LETTER OLD TAMIL SHORT E..BRAHMI LETTER OLD TAMIL SHORT O
    Interval{ 0x11075, 0x11075 }, // Lo       BRAHMI LETTER OLD TAMIL LLA
    Interval{ 0x11083, 0x110AF }, // Lo  [45] KAITHI LETTER A..KAITHI LETTER HA
    Interval{ 0x110D0, 0x110E8 }, // Lo  [25] SORA SOMPENG LETTER SAH..SORA SOMPENG LETTER MAE
    Interval{ 0x11103, 0x11126 }, // Lo  [36] CHAKMA LETTER AA..CHAKMA LETTER HAA
    Interval{ 0x11144, 0x11144 }, // Lo       CHAKMA LETTER LHAA
    Interval{ 0x11147, 0x11147 }, // Lo       CHAKMA LETTER VAA
    Interval{ 0x11150, 0x11172 }, // Lo  [35] MAHAJANI LETTER A..MAHAJANI LETTER RRA
    Interval{ 0x11176, 0x11176 }, // Lo       MAHAJANI LIGATURE SHRI
    Interval{ 0x11183, 0x111B2 }, // Lo  [48] SHARADA LETTER A..SHARADA LETTER HA
    Interval{ 0x111C1, 0x111C4 }, // Lo   [4] SHARADA SIGN AVAGRAHA..SHARADA OM
    Interval{ 0x111DA, 0x111DA }, // Lo       SHARADA EKAM
    Interval{ 0x111DC, 0x111DC }, // Lo       SHARADA HEADSTROKE
    Interval{ 0x11200, 0x11211 }, // Lo  [18] KHOJKI LETTER A..KHOJKI LETTER JJA
    Interval{ 0x11213, 0x1122B }, // Lo  [25] KHOJKI LETTER NYA..KHOJKI LETTER LLA
    Interval{ 0x1123F, 0x11240 }, // Lo   [2] KHOJKI LETTER QA..KHOJKI LETTER SHORT I
    Interval{ 0x11280, 0x11286 }, // Lo   [7] MULTANI LETTER A..MULTANI LETTER GA
    Interval{ 0x11288, 0x11288 }, // Lo       MULTANI LETTER GHA
    Interval{ 0x1128A, 0x1128D }, // Lo   [4] MULTANI LETTER CA..MULTANI LETTER JJA
    Interval{ 0x1128F, 0x1129D }, // Lo  [15] MULTANI LETTER NYA..MULTANI LETTER BA
    Interval{ 0x1129F, 0x112A8 }, // Lo  [10] MULTANI LETTER BHA..MULTANI LETTER RHA
    Interval{ 0x112B0, 0x112DE }, // Lo  [47] KHUDAWADI LETTER A..KHUDAWADI LETTER HA
    Interval{ 0x11305, 0x1130C }, // Lo   [8] GRANTHA LETTER A..GRANTHA LETTER VOCALIC L
    Interval{ 0x1130F, 0x11310 }, // Lo   [2] GRANTHA LETTER EE..GRANTHA LETTER AI
    Interval{ 0x11313, 0x11328 }, // Lo  [22] GRANTHA LETTER OO..GRANTHA LETTER NA
    Interval{ 0x1132A, 0x11330 }, // Lo   [7] GRANTHA LETTER PA..GRANTHA LETTER RA
    Interval{ 0x11332, 0x11333 }, // Lo   [2] GRANTHA LETTER LA..GRANTHA LETTER LLA
    Interval{ 0x11335, 0x11339 }, // Lo   [5] GRANTHA LETTER VA..GRANTHA LETTER HA
    Interval{ 0x1133D, 0x1133D }, // Lo       GRANTHA SIGN AVAGRAHA
    Interval{ 0x11350, 0x11350 }, // Lo       GRANTHA OM
    Interval{ 0x1135D, 0x11361 }, // Lo   [5] GRANTHA SIGN PLUTA..GRANTHA LETTER VOCALIC LL
    Interval{ 0x11380, 0x11389 }, // Lo  [10] TULU-TIGALARI LETTER A..TULU-TIGALARI LETTER VOCALIC LL
    Interval{ 0x1138B, 0x1138B }, // Lo       TULU-TIGALARI LETTER EE
    Interval{ 0x1138E, 0x1138E }, // Lo       TULU-TIGALARI LETTER AI
    Interval{ 0x11390, 0x113B5 }, // Lo  [38] TULU-TIGALARI LETTER OO..TULU-TIGALARI LETTER LLLA
    Interval{ 0x113B7, 0x113B7 }, // Lo       TULU-TIGALARI SIGN AVAGRAHA
    Interval{ 0x113D1, 0x113D1 }, // Lo       TULU-TIGALARI REPHA
    Interval{ 0x113D3, 0x113D3 }, // Lo       TULU-TIGALARI SIGN PLUTA
    Interval{ 0x11400, 0x11434 }, // Lo  [53] NEWA LETTER A..NEWA LETTER HA
    Interval{ 0x11447, 0x1144A }, // Lo   [4] NEWA SIGN AVAGRAHA..NEWA SIDDHI
    Interval{ 0x1145F, 0x11461 }, // Lo   [3] NEWA LETTER VEDIC ANUSVARA..NEWA SIGN UPADHMANIYA
    Interval{ 0x11480, 0x114AF }, // Lo  [48] TIRHUTA ANJI..TIRHUTA LETTER HA
    Interval{ 0x114C4, 0x114C5 }, // Lo   [2] TIRHUTA SIGN AVAGRAHA..TIRHUTA GVANG
    Interval{ 0x114C7, 0x114C7 }, // Lo       TIRHUTA OM
    Interval{ 0x11580, 0x115AE }, // Lo  [47] SIDDHAM LETTER A..SIDDHAM LETTER HA
    Interval{ 0x115D8, 0x115DB }, // Lo   [4] SIDDHAM LETTER THREE-CIRCLE ALTERNATE I..SIDDHAM LETTER ALTERNATE U
    Interval{ 0x11600, 0x1162F }, // Lo  [48] MODI LETTER A..MODI LETTER LLA
    Interval{ 0x11644, 0x11644 }, // Lo       MODI SIGN HUVA
    Interval{ 0x11680, 0x116AA }, // Lo  [43] TAKRI LETTER A..TAKRI LETTER RRA
    Interval{ 0x116B8, 0x116B8 }, // Lo       TAKRI LETTER ARCHAIC KHA
    Interval{ 0x11700, 0x1171A }, // Lo  [27] AHOM LETTER KA..AHOM LETTER ALTERNATE BA
    Interval{ 0x11740, 0x11746 }, // Lo   [7] AHOM LETTER CA..AHOM LETTER LLA
    Interval{ 0x11800, 0x1182B }, // Lo  [44] DOGRA LETTER A..DOGRA LETTER RRA
    Interval{ 0x118A0, 0x118DF }, // L&  [64] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Interval{ 0x118FF, 0x11906 }, // Lo   [8] WARANG CITI OM..DIVES AKURU LETTER E
    Interval{ 0x11909, 0x11909 }, // Lo       DIVES AKURU LETTER O
    Interval{ 0x1190C, 0x11913 }, // Lo   [8] DIVES AKURU LETTER KA..DIVES AKURU LETTER JA
    Interval{ 0x11915, 0x11916 }, // Lo   [2] DIVES AKURU LETTER NYA..DIVES AKURU LETTER TTA
    Interval{ 0x11918, 0x1192F }, // Lo  [24] DIVES AKURU LETTER DDA..DIVES AKURU LETTER ZA
    Interval{ 0x1193F, 0x1193F }, // Lo       DIVES AKURU PREFIXED NASAL SIGN
    Interval{ 0x11941, 0x11941 }, // Lo       DIVES AKURU INITIAL RA
    Interval{ 0x119A0, 0x119A7 }, // Lo   [8] NANDINAGARI LETTER A..NANDINAGARI LETTER VOCALIC RR
    Interval{ 0x119AA, 0x119D0 }, // Lo  [39] NANDINAGARI LETTER E..NANDINAGARI LETTER RRA
    Interval{ 0x119E1, 0x119E1 }, // Lo       NANDINAGARI SIGN AVAGRAHA
    Interval{ 0x119E3, 0x119E3 }, // Lo       NANDINAGARI HEADSTROKE
    Interval{ 0x11A00, 0x11A00 }, // Lo       ZANABAZAR SQUARE LETTER A
    Interval{ 0x11A0B, 0x11A32 }, // Lo  [40] ZANABAZAR SQUARE LETTER KA..ZANABAZAR SQUARE LETTER KSSA
    Interval{ 0x11A3A, 0x11A3A }, // Lo       ZANABAZAR SQUARE CLUSTER-INITIAL LETTER RA
    Interval{ 0x11A50, 0x11A50 }, // Lo       SOYOMBO LETTER A
    Interval{ 0x11A5C, 0x11A89 }, // Lo  [46] SOYOMBO LETTER KA..SOYOMBO CLUSTER-INITIAL LETTER SA
    Interval{ 0x11A9D, 0x11A9D }, // Lo       SOYOMBO MARK PLUTA
    Interval{ 0x11AB0, 0x11AF8 }, // Lo  [73] CANADIAN SYLLABICS NATTILIK HI..PAU CIN HAU GLOTTAL STOP FINAL
    Interval{ 0x11BC0, 0x11BE0 }, // Lo  [33] SUNUWAR LETTER DEVI..SUNUWAR LETTER KLOKO
    Interval{ 0x11C00, 0x11C08 }, // Lo   [9] BHAIKSUKI LETTER A..BHAIKSUKI LETTER VOCALIC L
    Interval{ 0x11C0A, 0x11C2E }, // Lo  [37] BHAIKSUKI LETTER E..BHAIKSUKI LETTER HA
    Interval{ 0x11C40, 0x11C40 }, // Lo       BHAIKSUKI SIGN AVAGRAHA
    Interval{ 0x11C72, 0x11C8F }, // Lo  [30] MARCHEN LETTER KA..MARCHEN LETTER A
    Interval{ 0x11D00, 0x11D06 }, // Lo   [7] MASARAM GONDI LETTER A..MASARAM GONDI LETTER E
    Interval{ 0x11D08, 0x11D09 }, // Lo   [2] MASARAM GONDI LETTER AI..MASARAM GONDI LETTER O
    Interval{ 0x11D0B, 0x11D30 }, // Lo  [38] MASARAM GONDI LETTER AU..MASARAM GONDI LETTER TRA
    Interval{ 0x11D46, 0x11D46 }, // Lo       MASARAM GONDI REPHA
    Interval{ 0x11D60, 0x11D65 }, // Lo   [6] GUNJALA GONDI LETTER A..GUNJALA GONDI LETTER UU
    Interval{ 0x11D67, 0x11D68 }, // Lo   [2] GUNJALA GONDI LETTER EE..GUNJALA GONDI LETTER AI
    Interval{ 0x11D6A, 0x11D89 }, // Lo  [32] GUNJALA GONDI LETTER OO..GUNJALA GONDI LETTER SA
    Interval{ 0x11D98, 0x11D98 }, // Lo       GUNJALA GONDI OM
    Interval{ 0x11DB0, 0x11DD8 }, // Lo  [41] TOLONG SIKI LETTER I..TOLONG SIKI LETTER RRH
    Interval{ 0x11DD9, 0x11DD9 }, // Lm       TOLONG SIKI SIGN SELA
    Interval{ 0x11DDA, 0x11DDB }, // Lo   [2] TOLONG SIKI SIGN HECAKA..TOLONG SIKI UNGGA
    Interval{ 0x11EE0, 0x11EF2 }, // Lo  [19] MAKASAR LETTER KA..MAKASAR ANGKA
    Interval{ 0x11F02, 0x11F02 }, // Lo       KAWI SIGN REPHA
    Interval{ 0x11F04, 0x11F10 }, // Lo  [13] KAWI LETTER A..KAWI LETTER O
    Interval{ 0x11F12, 0x11F33 }, // Lo  [34] KAWI LETTER KA..KAWI LETTER JNYA
    Interval{ 0x11FB0, 0x11FB0 }, // Lo       LISU LETTER YHA
    Interval{ 0x12000, 0x12399 }, // Lo [922] CUNEIFORM SIGN A..CUNEIFORM SIGN U U
    Interval{ 0x12400, 0x1246E }, // Nl [111] CUNEIFORM NUMERIC SIGN TWO ASH..CUNEIFORM NUMERIC SIGN NINE U VARIANT FORM
    Interval{ 0x12480, 0x12543 }, // Lo [196] CUNEIFORM SIGN AB TIMES NUN TENU..CUNEIFORM SIGN ZU5 TIMES THREE DISH TENU
    Interval{ 0x12F90, 0x12FF0 }, // Lo  [97] CYPRO-MINOAN SIGN CM001..CYPRO-MINOAN SIGN CM114
    Interval{ 0x13000, 0x1342F }, // Lo [1072] EGYPTIAN HIEROGLYPH A001..EGYPTIAN HIEROGLYPH V011D
    Interval{ 0x13441, 0x13446 }, // Lo   [6] EGYPTIAN HIEROGLYPH FULL BLANK..EGYPTIAN HIEROGLYPH WIDE LOST SIGN
    Interval{ 0x13460, 0x143FA }, // Lo [3995] EGYPTIAN HIEROGLYPH-13460..EGYPTIAN HIEROGLYPH-143FA
    Interval{ 0x14400, 0x14646 }, // Lo [583] ANATOLIAN HIEROGLYPH A001..ANATOLIAN HIEROGLYPH A530
    Interval{ 0x16100, 0x1611D }, // Lo  [30] GURUNG KHEMA LETTER A..GURUNG KHEMA LETTER SA
    Interval{ 0x16800, 0x16A38 }, // Lo [569] BAMUM LETTER PHASE-A NGKUE MFON..BAMUM LETTER PHASE-F VUEQ
    Interval{ 0x16A40, 0x16A5E }, // Lo  [31] MRO LETTER TA..MRO LETTER TEK
    Interval{ 0x16A70, 0x16ABE }, // Lo  [79] TANGSA LETTER OZ..TANGSA LETTER ZA
    Interval{ 0x16AD0, 0x16AED }, // Lo  [30] BASSA VAH LETTER ENNI..BASSA VAH LETTER I
    Interval{ 0x16B00, 0x16B2F }, // Lo  [48] PAHAWH HMONG VOWEL KEEB..PAHAWH HMONG CONSONANT CAU
    Interval{ 0x16B40, 0x16B43 }, // Lm   [4] PAHAWH HMONG SIGN VOS SEEV..PAHAWH HMONG SIGN IB YAM
    Interval{ 0x16B63, 0x16B77 }, // Lo  [21] PAHAWH HMONG SIGN VOS LUB..PAHAWH HMONG SIGN CIM NRES TOS
    Interval{ 0x16B7D, 0x16B8F }, // Lo  [19] PAHAWH HMONG CLAN SIGN TSHEEJ..PAHAWH HMONG CLAN SIGN VWJ
    Interval{ 0x16D40, 0x16D42 }, // Lm   [3] KIRAT RAI SIGN ANUSVARA..KIRAT RAI SIGN VISARGA
    Interval{ 0x16D43, 0x16D6A }, // Lo  [40] KIRAT RAI LETTER A..KIRAT RAI VOWEL SIGN AU
    Interval{ 0x16D6B, 0x16D6C }, // Lm   [2] KIRAT RAI SIGN VIRAMA..KIRAT RAI SIGN SAAT
    Interval{ 0x16E40, 0x16E7F }, // L&  [64] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Interval{ 0x16EA0, 0x16EB8 }, // L&  [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Interval{ 0x16EBB, 0x16ED3 }, // L&  [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Interval{ 0x16F00, 0x16F4A }, // Lo  [75] MIAO LETTER PA..MIAO LETTER RTE
    Interval{ 0x16F50, 0x16F50 }, // Lo       MIAO LETTER NASALIZATION
    Interval{ 0x16F93, 0x16F9F }, // Lm  [13] MIAO LETTER TONE-2..MIAO LETTER REFORMED TONE-8
    Interval{ 0x16FE0, 0x16FE1 }, // Lm   [2] TANGUT ITERATION MARK..NUSHU ITERATION MARK
    Interval{ 0x16FE3, 0x16FE3 }, // Lm       OLD CHINESE ITERATION MARK
    Interval{ 0x16FF2, 0x16FF3 }, // Lm   [2] CHINESE SMALL SIMPLIFIED ER..CHINESE SMALL TRADITIONAL ER
    Interval{ 0x16FF4, 0x16FF6 }, // Nl   [3] YANGQIN SIGN SLOW ONE BEAT..YANGQIN SIGN SLOW TWO BEATS
    Interval{ 0x17000, 0x18CD5 }, // Lo [7382] TANGUT IDEOGRAPH-17000..KHITAN SMALL SCRIPT CHARACTER-18CD5
    Interval{ 0x18CFF, 0x18D1E }, // Lo  [32] KHITAN SMALL SCRIPT CHARACTER-18CFF..TANGUT IDEOGRAPH-18D1E
    Interval{ 0x18D80, 0x18DF2 }, // Lo [115] TANGUT COMPONENT-769..TANGUT COMPONENT-883
    Interval{ 0x1AFF0, 0x1AFF3 }, // Lm   [4] KATAKANA LETTER MINNAN TONE-2..KATAKANA LETTER MINNAN TONE-5
    Interval{ 0x1AFF5, 0x1AFFB }, // Lm   [7] KATAKANA LETTER MINNAN TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-5
    Interval{ 0x1AFFD, 0x1AFFE }, // Lm   [2] KATAKANA LETTER MINNAN NASALIZED TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-8
    Interval{ 0x1B000, 0x1B122 }, // Lo [291] KATAKANA LETTER ARCHAIC E..KATAKANA LETTER ARCHAIC WU
    Interval{ 0x1B132, 0x1B132 }, // Lo       HIRAGANA LETTER SMALL KO
    Interval{ 0x1B150, 0x1B152 }, // Lo   [3] HIRAGANA LETTER SMALL WI..HIRAGANA LETTER SMALL WO
    Interval{ 0x1B155, 0x1B155 }, // Lo       KATAKANA LETTER SMALL KO
    Interval{ 0x1B164, 0x1B167 }, // Lo   [4] KATAKANA LETTER SMALL WI..KATAKANA LETTER SMALL N
    Interval{ 0x1B170, 0x1B2FB }, // Lo [396] NUSHU CHARACTER-1B170..NUSHU CHARACTER-1B2FB
    Interval{ 0x1BC00, 0x1BC6A }, // Lo [107] DUPLOYAN LETTER H..DUPLOYAN LETTER VOCALIC M
    Interval{ 0x1BC70, 0x1BC7C }, // Lo  [13] DUPLOYAN AFFIX LEFT HORIZONTAL SECANT..DUPLOYAN AFFIX ATTACHED TANGENT HOOK
    Interval{ 0x1BC80, 0x1BC88 }, // Lo   [9] DUPLOYAN AFFIX HIGH ACUTE..DUPLOYAN AFFIX HIGH VERTICAL
    Interval{ 0x1BC90, 0x1BC99 }, // Lo  [10] DUPLOYAN AFFIX LOW ACUTE..DUPLOYAN AFFIX LOW ARROW
    Interval{ 0x1D400, 0x1D454 }, // L&  [85] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL ITALIC SMALL G
    Interval{ 0x1D456, 0x1D49C }, // L&  [71] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL SCRIPT CAPITAL A
    Interval{ 0x1D49E, 0x1D49F }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Interval{ 0x1D4A2, 0x1D4A2 }, // L&       MATHEMATICAL SCRIPT CAPITAL G
    Interval{ 0x1D4A5, 0x1D4A6 }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Interval{ 0x1D4A9, 0x1D4AC }, // L&   [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Interval{ 0x1D4AE, 0x1D4B9 }, // L&  [12] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT SMALL D
    Interval{ 0x1D4BB, 0x1D4BB }, // L&       MATHEMATICAL SCRIPT SMALL F
    Interval{ 0x1D4BD, 0x1D4C3 }, // L&   [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Interval{ 0x1D4C5, 0x1D505 }, // L&  [65] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL FRAKTUR CAPITAL B
    Interval{ 0x1D507, 0x1D50A }, // L&   [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Interval{ 0x1D50D, 0x1D514 }, // L&   [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Interval{ 0x1D516, 0x1D51C }, // L&   [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Interval{ 0x1D51E, 0x1D539 }, // L&  [28] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Interval{ 0x1D53B, 0x1D53E }, // L&   [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Interval{ 0x1D540, 0x1D544 }, // L&   [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Interval{ 0x1D546, 0x1D546 }, // L&       MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Interval{ 0x1D54A, 0x1D550 }, // L&   [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Interval{ 0x1D552, 0x1D6A5 }, // L& [340] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    Interval{ 0x1D6A8, 0x1D6C0 }, // L&  [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    Interval{ 0x1D6C2, 0x1D6DA }, // L&  [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    Interval{ 0x1D6DC, 0x1D6FA }, // L&  [31] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL ITALIC CAPITAL OMEGA
    Interval{ 0x1D6FC, 0x1D714 }, // L&  [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    Interval{ 0x1D716, 0x1D734 }, // L&  [31] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D736, 0x1D74E }, // L&  [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D750, 0x1D76E }, // L&  [31] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    Interval{ 0x1D770, 0x1D788 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    Interval{ 0x1D78A, 0x1D7A8 }, // L&  [31] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D7AA, 0x1D7C2 }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D7C4, 0x1D7CB }, // L&   [8] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD SMALL DIGAMMA
    Interval{ 0x1DF00, 0x1DF09 }, // L&  [10] LATIN SMALL LETTER FENG DIGRAPH WITH TRILL..LATIN SMALL LETTER T WITH HOOK AND RETROFLEX HOOK
    Interval{ 0x1DF0A, 0x1DF0A }, // Lo       LATIN LETTER RETROFLEX CLICK WITH RETROFLEX HOOK
    Interval{ 0x1DF0B, 0x1DF1E }, // L&  [20] LATIN SMALL LETTER ESH WITH DOUBLE BAR..LATIN SMALL LETTER S WITH CURL
    Interval{ 0x1DF25, 0x1DF2A }, // L&   [6] LATIN SMALL LETTER D WITH MID-HEIGHT LEFT HOOK..LATIN SMALL LETTER T WITH MID-HEIGHT LEFT HOOK
    Interval{ 0x1E030, 0x1E06D }, // Lm  [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Interval{ 0x1E100, 0x1E12C }, // Lo  [45] NYIAKENG PUACHUE HMONG LETTER MA..NYIAKENG PUACHUE HMONG LETTER W
    Interval{ 0x1E137, 0x1E13D }, // Lm   [7] NYIAKENG PUACHUE HMONG SIGN FOR PERSON..NYIAKENG PUACHUE HMONG SYLLABLE LENGTHENER
    Interval{ 0x1E14E, 0x1E14E }, // Lo       NYIAKENG PUACHUE HMONG LOGOGRAM NYAJ
    Interval{ 0x1E290, 0x1E2AD }, // Lo  [30] TOTO LETTER PA..TOTO LETTER A
    Interval{ 0x1E2C0, 0x1E2EB }, // Lo  [44] WANCHO LETTER AA..WANCHO LETTER YIH
    Interval{ 0x1E4D0, 0x1E4EA }, // Lo  [27] NAG MUNDARI LETTER O..NAG MUNDARI LETTER ELL
    Interval{ 0x1E4EB, 0x1E4EB }, // Lm       NAG MUNDARI SIGN OJOD
    Interval{ 0x1E5D0, 0x1E5ED }, // Lo  [30] OL ONAL LETTER O..OL ONAL LETTER EG
    Interval{ 0x1E5F0, 0x1E5F0 }, // Lo       OL ONAL SIGN HODDOND
    Interval{ 0x1E6C0, 0x1E6DE }, // Lo  [31] TAI YO LETTER LOW KO..TAI YO LETTER HIGH KVO
    Interval{ 0x1E6E0, 0x1E6E2 }, // Lo   [3] TAI YO LETTER AA..TAI YO LETTER UE
    Interval{ 0x1E6E4, 0x1E6E5 }, // Lo   [2] TAI YO LETTER U..TAI YO LETTER AE
    Interval{ 0x1E6E7, 0x1E6ED }, // Lo   [7] TAI YO LETTER O..TAI YO LETTER AUE
    Interval{ 0x1E6F0, 0x1E6F4 }, // Lo   [5] TAI YO LETTER AN..TAI YO LETTER AP
    Interval{ 0x1E6FE, 0x1E6FE }, // Lo       TAI YO SYMBOL MUEANG
    Interval{ 0x1E6FF, 0x1E6FF }, // Lm       TAI YO XAM LAI
    Interval{ 0x1E7E0, 0x1E7E6 }, // Lo   [7] ETHIOPIC SYLLABLE HHYA..ETHIOPIC SYLLABLE HHYO
    Interval{ 0x1E7E8, 0x1E7EB }, // Lo   [4] ETHIOPIC SYLLABLE GURAGE HHWA..ETHIOPIC SYLLABLE HHWE
    Interval{ 0x1E7ED, 0x1E7EE }, // Lo   [2] ETHIOPIC SYLLABLE GURAGE MWI..ETHIOPIC SYLLABLE GURAGE MWEE
    Interval{ 0x1E7F0, 0x1E7FE }, // Lo  [15] ETHIOPIC SYLLABLE GURAGE QWI..ETHIOPIC SYLLABLE GURAGE PWEE
    Interval{ 0x1E800, 0x1E8C4 }, // Lo [197] MENDE KIKAKUI SYLLABLE M001 KI..MENDE KIKAKUI SYLLABLE M060 NYON
    Interval{ 0x1E900, 0x1E943 }, // L&  [68] ADLAM CAPITAL LETTER ALIF..ADLAM SMALL LETTER SHA
    Interval{ 0x1E94B, 0x1E94B }, // Lm       ADLAM NASALIZATION MARK
    Interval{ 0x1EE00, 0x1EE03 }, // Lo   [4] ARABIC MATHEMATICAL ALEF..ARABIC MATHEMATICAL DAL
    Interval{ 0x1EE05, 0x1EE1F }, // Lo  [27] ARABIC MATHEMATICAL WAW..ARABIC MATHEMATICAL DOTLESS QAF
    Interval{ 0x1EE21, 0x1EE22 }, // Lo   [2] ARABIC MATHEMATICAL INITIAL BEH..ARABIC MATHEMATICAL INITIAL JEEM
    Interval{ 0x1EE24, 0x1EE24 }, // Lo       ARABIC MATHEMATICAL INITIAL HEH
    Interval{ 0x1EE27, 0x1EE27 }, // Lo       ARABIC MATHEMATICAL INITIAL HAH
    Interval{ 0x1EE29, 0x1EE32 }, // Lo  [10] ARABIC MATHEMATICAL INITIAL YEH..ARABIC MATHEMATICAL INITIAL QAF
    Interval{ 0x1EE34, 0x1EE37 }, // Lo   [4] ARABIC MATHEMATICAL INITIAL SHEEN..ARABIC MATHEMATICAL INITIAL KHAH
    Interval{ 0x1EE39, 0x1EE39 }, // Lo       ARABIC MATHEMATICAL INITIAL DAD
    Interval{ 0x1EE3B, 0x1EE3B }, // Lo       ARABIC MATHEMATICAL INITIAL GHAIN
    Interval{ 0x1EE42, 0x1EE42 }, // Lo       ARABIC MATHEMATICAL TAILED JEEM
    Interval{ 0x1EE47, 0x1EE47 }, // Lo       ARABIC MATHEMATICAL TAILED HAH
    Interval{ 0x1EE49, 0x1EE49 }, // Lo       ARABIC MATHEMATICAL TAILED YEH
    Interval{ 0x1EE4B, 0x1EE4B }, // Lo       ARABIC MATHEMATICAL TAILED LAM
    Interval{ 0x1EE4D, 0x1EE4F }, // Lo   [3] ARABIC MATHEMATICAL TAILED NOON..ARABIC MATHEMATICAL TAILED AIN
    Interval{ 0x1EE51, 0x1EE52 }, // Lo   [2] ARABIC MATHEMATICAL TAILED SAD..ARABIC MATHEMATICAL TAILED QAF
    Interval{ 0x1EE54, 0x1EE54 }, // Lo       ARABIC MATHEMATICAL TAILED SHEEN
    Interval{ 0x1EE57, 0x1EE57 }, // Lo       ARABIC MATHEMATICAL TAILED KHAH
    Interval{ 0x1EE59, 0x1EE59 }, // Lo       ARABIC MATHEMATICAL TAILED DAD
    Interval{ 0x1EE5B, 0x1EE5B }, // Lo       ARABIC MATHEMATICAL TAILED GHAIN
    Interval{ 0x1EE5D, 0x1EE5D }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS NOON
    Interval{ 0x1EE5F, 0x1EE5F }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS QAF
    Interval{ 0x1EE61, 0x1EE62 }, // Lo   [2] ARABIC MATHEMATICAL STRETCHED BEH..ARABIC MATHEMATICAL STRETCHED JEEM
    Interval{ 0x1EE64, 0x1EE64 }, // Lo       ARABIC MATHEMATICAL STRETCHED HEH
    Interval{ 0x1EE67, 0x1EE6A }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED HAH..ARABIC MATHEMATICAL STRETCHED KAF
    Interval{ 0x1EE6C, 0x1EE72 }, // Lo   [7] ARABIC MATHEMATICAL STRETCHED MEEM..ARABIC MATHEMATICAL STRETCHED QAF
    Interval{ 0x1EE74, 0x1EE77 }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED SHEEN..ARABIC MATHEMATICAL STRETCHED KHAH
    Interval{ 0x1EE79, 0x1EE7C }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED DAD..ARABIC MATHEMATICAL STRETCHED DOTLESS BEH
    Interval{ 0x1EE7E, 0x1EE7E }, // Lo       ARABIC MATHEMATICAL STRETCHED DOTLESS FEH
    Interval{ 0x1EE80, 0x1EE89 }, // Lo  [10] ARABIC MATHEMATICAL LOOPED ALEF..ARABIC MATHEMATICAL LOOPED YEH
    Interval{ 0x1EE8B, 0x1EE9B }, // Lo  [17] ARABIC MATHEMATICAL LOOPED LAM..ARABIC MATHEMATICAL LOOPED GHAIN
    Interval{ 0x1EEA1, 0x1EEA3 }, // Lo   [3] ARABIC MATHEMATICAL DOUBLE-STRUCK BEH..ARABIC MATHEMATICAL DOUBLE-STRUCK DAL
    Interval{ 0x1EEA5, 0x1EEA9 }, // Lo   [5] ARABIC MATHEMATICAL DOUBLE-STRUCK WAW..ARABIC MATHEMATICAL DOUBLE-STRUCK YEH
    Interval{ 0x1EEAB, 0x1EEBB }, // Lo  [17] ARABIC MATHEMATICAL DOUBLE-STRUCK LAM..ARABIC MATHEMATICAL DOUBLE-STRUCK GHAIN
    Interval{ 0x20000, 0x2A6DF }, // Lo [42720] CJK UNIFIED IDEOGRAPH-20000..CJK UNIFIED IDEOGRAPH-2A6DF
    Interval{ 0x2A700, 0x2B81D }, // Lo [4382] CJK UNIFIED IDEOGRAPH-2A700..CJK UNIFIED IDEOGRAPH-2B81D
    Interval{ 0x2B820, 0x2CEAD }, // Lo [5774] CJK UNIFIED IDEOGRAPH-2B820..CJK UNIFIED IDEOGRAPH-2CEAD
    Interval{ 0x2CEB0, 0x2EBE0 }, // Lo [7473] CJK UNIFIED IDEOGRAPH-2CEB0..CJK UNIFIED IDEOGRAPH-2EBE0
    Interval{ 0x2EBF0, 0x2EE5D }, // Lo [622] CJK UNIFIED IDEOGRAPH-2EBF0..CJK UNIFIED IDEOGRAPH-2EE5D
    Interval{ 0x2F800, 0x2FA1D }, // Lo [542] CJK COMPATIBILITY IDEOGRAPH-2F800..CJK COMPATIBILITY IDEOGRAPH-2FA1D
    Interval{ 0x30000, 0x3134A }, // Lo [4939] CJK UNIFIED IDEOGRAPH-30000..CJK UNIFIED IDEOGRAPH-3134A
    Interval{ 0x31350, 0x33479 }, // Lo [8490] CJK UNIFIED IDEOGRAPH-31350..CJK UNIFIED IDEOGRAPH-33479
}; // }}}
} // end namespace tables

bool contains(Core_Property prop, char32_t codepoint) noexcept {
    switch (prop)
    {
        case Core_Property::Alphabetic: return contains(tables::Alphabetic, codepoint);
        case Core_Property::Case_Ignorable: return contains(tables::Case_Ignorable, codepoint);
        case Core_Property::Cased: return contains(tables::Cased, codepoint);
        case Core_Property::Changes_When_Casefolded: return contains(tables::Changes_When_Casefolded, codepoint);
        case Core_Property::Changes_When_Casemapped: return contains(tables::Changes_When_Casemapped, codepoint);
        case Core_Property::Changes_When_Lowercased: return contains(tables::Changes_When_Lowercased, codepoint);
        case Core_Property::Changes_When_Titlecased: return contains(tables::Changes_When_Titlecased, codepoint);
        case Core_Property::Changes_When_Uppercased: return contains(tables::Changes_When_Uppercased, codepoint);
        case Core_Property::Default_Ignorable_Code_Point: return contains(tables::Default_Ignorable_Code_Point, codepoint);
        case Core_Property::Grapheme_Base: return contains(tables::Grapheme_Base, codepoint);
        case Core_Property::Grapheme_Extend: return contains(tables::Grapheme_Extend, codepoint);
        case Core_Property::Grapheme_Link: return contains(tables::Grapheme_Link, codepoint);
        case Core_Property::ID_Continue: return contains(tables::ID_Continue, codepoint);
        case Core_Property::ID_Start: return contains(tables::ID_Start, codepoint);
        case Core_Property::Lowercase: return contains(tables::Lowercase, codepoint);
        case Core_Property::Math: return contains(tables::Math, codepoint);
        case Core_Property::Uppercase: return contains(tables::Uppercase, codepoint);
        case Core_Property::XID_Continue: return contains(tables::XID_Continue, codepoint);
        case Core_Property::XID_Start: return contains(tables::XID_Start, codepoint);
    }
    return false;
}

namespace tables {
auto const General_Category = std::array<Prop<::unicode::General_Category>, 4144>{
    Prop<::unicode::General_Category> { { 0x0000, 0x001F }, ::unicode::General_Category::Control }, // [32] <control-0000>..<control-001F>
    Prop<::unicode::General_Category> { { 0x0020, 0x0020 }, ::unicode::General_Category::Space_Separator }, // SPACE
    Prop<::unicode::General_Category> { { 0x0021, 0x0023 }, ::unicode::General_Category::Other_Punctuation }, // [3] EXCLAMATION MARK..NUMBER SIGN
    Prop<::unicode::General_Category> { { 0x0024, 0x0024 }, ::unicode::General_Category::Currency_Symbol }, // DOLLAR SIGN
    Prop<::unicode::General_Category> { { 0x0025, 0x0027 }, ::unicode::General_Category::Other_Punctuation }, // [3] PERCENT SIGN..APOSTROPHE
    Prop<::unicode::General_Category> { { 0x0028, 0x0028 }, ::unicode::General_Category::Open_Punctuation }, // LEFT PARENTHESIS
    Prop<::unicode::General_Category> { { 0x0029, 0x0029 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT PARENTHESIS
    Prop<::unicode::General_Category> { { 0x002A, 0x002A }, ::unicode::General_Category::Other_Punctuation }, // ASTERISK
    Prop<::unicode::General_Category> { { 0x002B, 0x002B }, ::unicode::General_Category::Math_Symbol }, // PLUS SIGN
    Prop<::unicode::General_Category> { { 0x002C, 0x002C }, ::unicode::General_Category::Other_Punctuation }, // COMMA
    Prop<::unicode::General_Category> { { 0x002D, 0x002D }, ::unicode::General_Category::Dash_Punctuation }, // HYPHEN-MINUS
    Prop<::unicode::General_Category> { { 0x002E, 0x002F }, ::unicode::General_Category::Other_Punctuation }, // [2] FULL STOP..SOLIDUS
    Prop<::unicode::General_Category> { { 0x0030, 0x0039 }, ::unicode::General_Category::Decimal_Number }, // [10] DIGIT ZERO..DIGIT NINE
    Prop<::unicode::General_Category> { { 0x003A, 0x003B }, ::unicode::General_Category::Other_Punctuation }, // [2] COLON..SEMICOLON
    Prop<::unicode::General_Category> { { 0x003C, 0x003E }, ::unicode::General_Category::Math_Symbol }, // [3] LESS-THAN SIGN..GREATER-THAN SIGN
    Prop<::unicode::General_Category> { { 0x003F, 0x0040 }, ::unicode::General_Category::Other_Punctuation }, // [2] QUESTION MARK..COMMERCIAL AT
    Prop<::unicode::General_Category> { { 0x0041, 0x005A }, ::unicode::General_Category::Uppercase_Letter }, // [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Prop<::unicode::General_Category> { { 0x005B, 0x005B }, ::unicode::General_Category::Open_Punctuation }, // LEFT SQUARE BRACKET
    Prop<::unicode::General_Category> { { 0x005C, 0x005C }, ::unicode::General_Category::Other_Punctuation }, // REVERSE SOLIDUS
    Prop<::unicode::General_Category> { { 0x005D, 0x005D }, ::unicode::General_Category::Close_Punctuation }, // RIGHT SQUARE BRACKET
    Prop<::unicode::General_Category> { { 0x005E, 0x005E }, ::unicode::General_Category::Modifier_Symbol }, // CIRCUMFLEX ACCENT
    Prop<::unicode::General_Category> { { 0x005F, 0x005F }, ::unicode::General_Category::Connector_Punctuation }, // LOW LINE
    Prop<::unicode::General_Category> { { 0x0060, 0x0060 }, ::unicode::General_Category::Modifier_Symbol }, // GRAVE ACCENT
    Prop<::unicode::General_Category> { { 0x0061, 0x007A }, ::unicode::General_Category::Lowercase_Letter }, // [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Prop<::unicode::General_Category> { { 0x007B, 0x007B }, ::unicode::General_Category::Open_Punctuation }, // LEFT CURLY BRACKET
    Prop<::unicode::General_Category> { { 0x007C, 0x007C }, ::unicode::General_Category::Math_Symbol }, // VERTICAL LINE
    Prop<::unicode::General_Category> { { 0x007D, 0x007D }, ::unicode::General_Category::Close_Punctuation }, // RIGHT CURLY BRACKET
    Prop<::unicode::General_Category> { { 0x007E, 0x007E }, ::unicode::General_Category::Math_Symbol }, // TILDE
    Prop<::unicode::General_Category> { { 0x007F, 0x009F }, ::unicode::General_Category::Control }, // [33] <control-007F>..<control-009F>
    Prop<::unicode::General_Category> { { 0x00A0, 0x00A0 }, ::unicode::General_Category::Space_Separator }, // NO-BREAK SPACE
    Prop<::unicode::General_Category> { { 0x00A1, 0x00A1 }, ::unicode::General_Category::Other_Punctuation }, // INVERTED EXCLAMATION MARK
    Prop<::unicode::General_Category> { { 0x00A2, 0x00A5 }, ::unicode::General_Category::Currency_Symbol }, // [4] CENT SIGN..YEN SIGN
    Prop<::unicode::General_Category> { { 0x00A6, 0x00A6 }, ::unicode::General_Category::Other_Symbol }, // BROKEN BAR
    Prop<::unicode::General_Category> { { 0x00A7, 0x00A7 }, ::unicode::General_Category::Other_Punctuation }, // SECTION SIGN
    Prop<::unicode::General_Category> { { 0x00A8, 0x00A8 }, ::unicode::General_Category::Modifier_Symbol }, // DIAERESIS
    Prop<::unicode::General_Category> { { 0x00A9, 0x00A9 }, ::unicode::General_Category::Other_Symbol }, // COPYRIGHT SIGN
    Prop<::unicode::General_Category> { { 0x00AA, 0x00AA }, ::unicode::General_Category::Other_Letter }, // FEMININE ORDINAL INDICATOR
    Prop<::unicode::General_Category> { { 0x00AB, 0x00AB }, ::unicode::General_Category::Initial_Punctuation }, // LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x00AC, 0x00AC }, ::unicode::General_Category::Math_Symbol }, // NOT SIGN
    Prop<::unicode::General_Category> { { 0x00AD, 0x00AD }, ::unicode::General_Category::Format }, // SOFT HYPHEN
    Prop<::unicode::General_Category> { { 0x00AE, 0x00AE }, ::unicode::General_Category::Other_Symbol }, // REGISTERED SIGN
    Prop<::unicode::General_Category> { { 0x00AF, 0x00AF }, ::unicode::General_Category::Modifier_Symbol }, // MACRON
    Prop<::unicode::General_Category> { { 0x00B0, 0x00B0 }, ::unicode::General_Category::Other_Symbol }, // DEGREE SIGN
    Prop<::unicode::General_Category> { { 0x00B1, 0x00B1 }, ::unicode::General_Category::Math_Symbol }, // PLUS-MINUS SIGN
    Prop<::unicode::General_Category> { { 0x00B2, 0x00B3 }, ::unicode::General_Category::Other_Number }, // [2] SUPERSCRIPT TWO..SUPERSCRIPT THREE
    Prop<::unicode::General_Category> { { 0x00B4, 0x00B4 }, ::unicode::General_Category::Modifier_Symbol }, // ACUTE ACCENT
    Prop<::unicode::General_Category> { { 0x00B5, 0x00B5 }, ::unicode::General_Category::Lowercase_Letter }, // MICRO SIGN
    Prop<::unicode::General_Category> { { 0x00B6, 0x00B7 }, ::unicode::General_Category::Other_Punctuation }, // [2] PILCROW SIGN..MIDDLE DOT
    Prop<::unicode::General_Category> { { 0x00B8, 0x00B8 }, ::unicode::General_Category::Modifier_Symbol }, // CEDILLA
    Prop<::unicode::General_Category> { { 0x00B9, 0x00B9 }, ::unicode::General_Category::Other_Number }, // SUPERSCRIPT ONE
    Prop<::unicode::General_Category> { { 0x00BA, 0x00BA }, ::unicode::General_Category::Other_Letter }, // MASCULINE ORDINAL INDICATOR
    Prop<::unicode::General_Category> { { 0x00BB, 0x00BB }, ::unicode::General_Category::Final_Punctuation }, // RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x00BC, 0x00BE }, ::unicode::General_Category::Other_Number }, // [3] VULGAR FRACTION ONE QUARTER..VULGAR FRACTION THREE QUARTERS
    Prop<::unicode::General_Category> { { 0x00BF, 0x00BF }, ::unicode::General_Category::Other_Punctuation }, // INVERTED QUESTION MARK
    Prop<::unicode::General_Category> { { 0x00C0, 0x00D6 }, ::unicode::General_Category::Uppercase_Letter }, // [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x00D7, 0x00D7 }, ::unicode::General_Category::Math_Symbol }, // MULTIPLICATION SIGN
    Prop<::unicode::General_Category> { { 0x00D8, 0x00DE }, ::unicode::General_Category::Uppercase_Letter }, // [7] LATIN CAPITAL LETTER O WITH STROKE..LATIN CAPITAL LETTER THORN
    Prop<::unicode::General_Category> { { 0x00DF, 0x00F6 }, ::unicode::General_Category::Lowercase_Letter }, // [24] LATIN SMALL LETTER SHARP S..LATIN SMALL LETTER O WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x00F7, 0x00F7 }, ::unicode::General_Category::Math_Symbol }, // DIVISION SIGN
    Prop<::unicode::General_Category> { { 0x00F8, 0x00FF }, ::unicode::General_Category::Lowercase_Letter }, // [8] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER Y WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x0100, 0x0100 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH MACRON
    Prop<::unicode::General_Category> { { 0x0101, 0x0101 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH MACRON
    Prop<::unicode::General_Category> { { 0x0102, 0x0102 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH BREVE
    Prop<::unicode::General_Category> { { 0x0103, 0x0103 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH BREVE
    Prop<::unicode::General_Category> { { 0x0104, 0x0104 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH OGONEK
    Prop<::unicode::General_Category> { { 0x0105, 0x0105 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH OGONEK
    Prop<::unicode::General_Category> { { 0x0106, 0x0106 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER C WITH ACUTE
    Prop<::unicode::General_Category> { { 0x0107, 0x0107 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER C WITH ACUTE
    Prop<::unicode::General_Category> { { 0x0108, 0x0108 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER C WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x0109, 0x0109 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER C WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x010A, 0x010A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER C WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x010B, 0x010B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER C WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x010C, 0x010C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER C WITH CARON
    Prop<::unicode::General_Category> { { 0x010D, 0x010D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER C WITH CARON
    Prop<::unicode::General_Category> { { 0x010E, 0x010E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER D WITH CARON
    Prop<::unicode::General_Category> { { 0x010F, 0x010F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER D WITH CARON
    Prop<::unicode::General_Category> { { 0x0110, 0x0110 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER D WITH STROKE
    Prop<::unicode::General_Category> { { 0x0111, 0x0111 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER D WITH STROKE
    Prop<::unicode::General_Category> { { 0x0112, 0x0112 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH MACRON
    Prop<::unicode::General_Category> { { 0x0113, 0x0113 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH MACRON
    Prop<::unicode::General_Category> { { 0x0114, 0x0114 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH BREVE
    Prop<::unicode::General_Category> { { 0x0115, 0x0115 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH BREVE
    Prop<::unicode::General_Category> { { 0x0116, 0x0116 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x0117, 0x0117 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x0118, 0x0118 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH OGONEK
    Prop<::unicode::General_Category> { { 0x0119, 0x0119 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH OGONEK
    Prop<::unicode::General_Category> { { 0x011A, 0x011A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH CARON
    Prop<::unicode::General_Category> { { 0x011B, 0x011B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH CARON
    Prop<::unicode::General_Category> { { 0x011C, 0x011C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER G WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x011D, 0x011D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER G WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x011E, 0x011E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER G WITH BREVE
    Prop<::unicode::General_Category> { { 0x011F, 0x011F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER G WITH BREVE
    Prop<::unicode::General_Category> { { 0x0120, 0x0120 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER G WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x0121, 0x0121 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER G WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x0122, 0x0122 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER G WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x0123, 0x0123 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER G WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x0124, 0x0124 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER H WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x0125, 0x0125 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER H WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x0126, 0x0126 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER H WITH STROKE
    Prop<::unicode::General_Category> { { 0x0127, 0x0127 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER H WITH STROKE
    Prop<::unicode::General_Category> { { 0x0128, 0x0128 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER I WITH TILDE
    Prop<::unicode::General_Category> { { 0x0129, 0x0129 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER I WITH TILDE
    Prop<::unicode::General_Category> { { 0x012A, 0x012A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER I WITH MACRON
    Prop<::unicode::General_Category> { { 0x012B, 0x012B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER I WITH MACRON
    Prop<::unicode::General_Category> { { 0x012C, 0x012C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER I WITH BREVE
    Prop<::unicode::General_Category> { { 0x012D, 0x012D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER I WITH BREVE
    Prop<::unicode::General_Category> { { 0x012E, 0x012E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER I WITH OGONEK
    Prop<::unicode::General_Category> { { 0x012F, 0x012F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER I WITH OGONEK
    Prop<::unicode::General_Category> { { 0x0130, 0x0130 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER I WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x0131, 0x0131 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER DOTLESS I
    Prop<::unicode::General_Category> { { 0x0132, 0x0132 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LIGATURE IJ
    Prop<::unicode::General_Category> { { 0x0133, 0x0133 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LIGATURE IJ
    Prop<::unicode::General_Category> { { 0x0134, 0x0134 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER J WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x0135, 0x0135 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER J WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x0136, 0x0136 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER K WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x0137, 0x0138 }, ::unicode::General_Category::Lowercase_Letter }, // [2] LATIN SMALL LETTER K WITH CEDILLA..LATIN SMALL LETTER KRA
    Prop<::unicode::General_Category> { { 0x0139, 0x0139 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER L WITH ACUTE
    Prop<::unicode::General_Category> { { 0x013A, 0x013A }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER L WITH ACUTE
    Prop<::unicode::General_Category> { { 0x013B, 0x013B }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER L WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x013C, 0x013C }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER L WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x013D, 0x013D }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER L WITH CARON
    Prop<::unicode::General_Category> { { 0x013E, 0x013E }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER L WITH CARON
    Prop<::unicode::General_Category> { { 0x013F, 0x013F }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER L WITH MIDDLE DOT
    Prop<::unicode::General_Category> { { 0x0140, 0x0140 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER L WITH MIDDLE DOT
    Prop<::unicode::General_Category> { { 0x0141, 0x0141 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER L WITH STROKE
    Prop<::unicode::General_Category> { { 0x0142, 0x0142 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER L WITH STROKE
    Prop<::unicode::General_Category> { { 0x0143, 0x0143 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER N WITH ACUTE
    Prop<::unicode::General_Category> { { 0x0144, 0x0144 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER N WITH ACUTE
    Prop<::unicode::General_Category> { { 0x0145, 0x0145 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER N WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x0146, 0x0146 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER N WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x0147, 0x0147 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER N WITH CARON
    Prop<::unicode::General_Category> { { 0x0148, 0x0149 }, ::unicode::General_Category::Lowercase_Letter }, // [2] LATIN SMALL LETTER N WITH CARON..LATIN SMALL LETTER N PRECEDED BY APOSTROPHE
    Prop<::unicode::General_Category> { { 0x014A, 0x014A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER ENG
    Prop<::unicode::General_Category> { { 0x014B, 0x014B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER ENG
    Prop<::unicode::General_Category> { { 0x014C, 0x014C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH MACRON
    Prop<::unicode::General_Category> { { 0x014D, 0x014D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH MACRON
    Prop<::unicode::General_Category> { { 0x014E, 0x014E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH BREVE
    Prop<::unicode::General_Category> { { 0x014F, 0x014F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH BREVE
    Prop<::unicode::General_Category> { { 0x0150, 0x0150 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH DOUBLE ACUTE
    Prop<::unicode::General_Category> { { 0x0151, 0x0151 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH DOUBLE ACUTE
    Prop<::unicode::General_Category> { { 0x0152, 0x0152 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LIGATURE OE
    Prop<::unicode::General_Category> { { 0x0153, 0x0153 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LIGATURE OE
    Prop<::unicode::General_Category> { { 0x0154, 0x0154 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER R WITH ACUTE
    Prop<::unicode::General_Category> { { 0x0155, 0x0155 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER R WITH ACUTE
    Prop<::unicode::General_Category> { { 0x0156, 0x0156 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER R WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x0157, 0x0157 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER R WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x0158, 0x0158 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER R WITH CARON
    Prop<::unicode::General_Category> { { 0x0159, 0x0159 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER R WITH CARON
    Prop<::unicode::General_Category> { { 0x015A, 0x015A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER S WITH ACUTE
    Prop<::unicode::General_Category> { { 0x015B, 0x015B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH ACUTE
    Prop<::unicode::General_Category> { { 0x015C, 0x015C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER S WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x015D, 0x015D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x015E, 0x015E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER S WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x015F, 0x015F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x0160, 0x0160 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER S WITH CARON
    Prop<::unicode::General_Category> { { 0x0161, 0x0161 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH CARON
    Prop<::unicode::General_Category> { { 0x0162, 0x0162 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER T WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x0163, 0x0163 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER T WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x0164, 0x0164 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER T WITH CARON
    Prop<::unicode::General_Category> { { 0x0165, 0x0165 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER T WITH CARON
    Prop<::unicode::General_Category> { { 0x0166, 0x0166 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER T WITH STROKE
    Prop<::unicode::General_Category> { { 0x0167, 0x0167 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER T WITH STROKE
    Prop<::unicode::General_Category> { { 0x0168, 0x0168 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH TILDE
    Prop<::unicode::General_Category> { { 0x0169, 0x0169 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH TILDE
    Prop<::unicode::General_Category> { { 0x016A, 0x016A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH MACRON
    Prop<::unicode::General_Category> { { 0x016B, 0x016B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH MACRON
    Prop<::unicode::General_Category> { { 0x016C, 0x016C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH BREVE
    Prop<::unicode::General_Category> { { 0x016D, 0x016D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH BREVE
    Prop<::unicode::General_Category> { { 0x016E, 0x016E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH RING ABOVE
    Prop<::unicode::General_Category> { { 0x016F, 0x016F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH RING ABOVE
    Prop<::unicode::General_Category> { { 0x0170, 0x0170 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH DOUBLE ACUTE
    Prop<::unicode::General_Category> { { 0x0171, 0x0171 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH DOUBLE ACUTE
    Prop<::unicode::General_Category> { { 0x0172, 0x0172 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH OGONEK
    Prop<::unicode::General_Category> { { 0x0173, 0x0173 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH OGONEK
    Prop<::unicode::General_Category> { { 0x0174, 0x0174 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER W WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x0175, 0x0175 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER W WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x0176, 0x0176 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Y WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x0177, 0x0177 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Y WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x0178, 0x0179 }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN CAPITAL LETTER Y WITH DIAERESIS..LATIN CAPITAL LETTER Z WITH ACUTE
    Prop<::unicode::General_Category> { { 0x017A, 0x017A }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Z WITH ACUTE
    Prop<::unicode::General_Category> { { 0x017B, 0x017B }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Z WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x017C, 0x017C }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Z WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x017D, 0x017D }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Z WITH CARON
    Prop<::unicode::General_Category> { { 0x017E, 0x0180 }, ::unicode::General_Category::Lowercase_Letter }, // [3] LATIN SMALL LETTER Z WITH CARON..LATIN SMALL LETTER B WITH STROKE
    Prop<::unicode::General_Category> { { 0x0181, 0x0182 }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN CAPITAL LETTER B WITH HOOK..LATIN CAPITAL LETTER B WITH TOPBAR
    Prop<::unicode::General_Category> { { 0x0183, 0x0183 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER B WITH TOPBAR
    Prop<::unicode::General_Category> { { 0x0184, 0x0184 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER TONE SIX
    Prop<::unicode::General_Category> { { 0x0185, 0x0185 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER TONE SIX
    Prop<::unicode::General_Category> { { 0x0186, 0x0187 }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN CAPITAL LETTER OPEN O..LATIN CAPITAL LETTER C WITH HOOK
    Prop<::unicode::General_Category> { { 0x0188, 0x0188 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER C WITH HOOK
    Prop<::unicode::General_Category> { { 0x0189, 0x018B }, ::unicode::General_Category::Uppercase_Letter }, // [3] LATIN CAPITAL LETTER AFRICAN D..LATIN CAPITAL LETTER D WITH TOPBAR
    Prop<::unicode::General_Category> { { 0x018C, 0x018D }, ::unicode::General_Category::Lowercase_Letter }, // [2] LATIN SMALL LETTER D WITH TOPBAR..LATIN SMALL LETTER TURNED DELTA
    Prop<::unicode::General_Category> { { 0x018E, 0x0191 }, ::unicode::General_Category::Uppercase_Letter }, // [4] LATIN CAPITAL LETTER REVERSED E..LATIN CAPITAL LETTER F WITH HOOK
    Prop<::unicode::General_Category> { { 0x0192, 0x0192 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER F WITH HOOK
    Prop<::unicode::General_Category> { { 0x0193, 0x0194 }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN CAPITAL LETTER G WITH HOOK..LATIN CAPITAL LETTER GAMMA
    Prop<::unicode::General_Category> { { 0x0195, 0x0195 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER HV
    Prop<::unicode::General_Category> { { 0x0196, 0x0198 }, ::unicode::General_Category::Uppercase_Letter }, // [3] LATIN CAPITAL LETTER IOTA..LATIN CAPITAL LETTER K WITH HOOK
    Prop<::unicode::General_Category> { { 0x0199, 0x019B }, ::unicode::General_Category::Lowercase_Letter }, // [3] LATIN SMALL LETTER K WITH HOOK..LATIN SMALL LETTER LAMBDA WITH STROKE
    Prop<::unicode::General_Category> { { 0x019C, 0x019D }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN CAPITAL LETTER TURNED M..LATIN CAPITAL LETTER N WITH LEFT HOOK
    Prop<::unicode::General_Category> { { 0x019E, 0x019E }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER N WITH LONG RIGHT LEG
    Prop<::unicode::General_Category> { { 0x019F, 0x01A0 }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN CAPITAL LETTER O WITH MIDDLE TILDE..LATIN CAPITAL LETTER O WITH HORN
    Prop<::unicode::General_Category> { { 0x01A1, 0x01A1 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH HORN
    Prop<::unicode::General_Category> { { 0x01A2, 0x01A2 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER OI
    Prop<::unicode::General_Category> { { 0x01A3, 0x01A3 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER OI
    Prop<::unicode::General_Category> { { 0x01A4, 0x01A4 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER P WITH HOOK
    Prop<::unicode::General_Category> { { 0x01A5, 0x01A5 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER P WITH HOOK
    Prop<::unicode::General_Category> { { 0x01A6, 0x01A7 }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN LETTER YR..LATIN CAPITAL LETTER TONE TWO
    Prop<::unicode::General_Category> { { 0x01A8, 0x01A8 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER TONE TWO
    Prop<::unicode::General_Category> { { 0x01A9, 0x01A9 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER ESH
    Prop<::unicode::General_Category> { { 0x01AA, 0x01AB }, ::unicode::General_Category::Lowercase_Letter }, // [2] LATIN LETTER REVERSED ESH LOOP..LATIN SMALL LETTER T WITH PALATAL HOOK
    Prop<::unicode::General_Category> { { 0x01AC, 0x01AC }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER T WITH HOOK
    Prop<::unicode::General_Category> { { 0x01AD, 0x01AD }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER T WITH HOOK
    Prop<::unicode::General_Category> { { 0x01AE, 0x01AF }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN CAPITAL LETTER T WITH RETROFLEX HOOK..LATIN CAPITAL LETTER U WITH HORN
    Prop<::unicode::General_Category> { { 0x01B0, 0x01B0 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH HORN
    Prop<::unicode::General_Category> { { 0x01B1, 0x01B3 }, ::unicode::General_Category::Uppercase_Letter }, // [3] LATIN CAPITAL LETTER UPSILON..LATIN CAPITAL LETTER Y WITH HOOK
    Prop<::unicode::General_Category> { { 0x01B4, 0x01B4 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Y WITH HOOK
    Prop<::unicode::General_Category> { { 0x01B5, 0x01B5 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Z WITH STROKE
    Prop<::unicode::General_Category> { { 0x01B6, 0x01B6 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Z WITH STROKE
    Prop<::unicode::General_Category> { { 0x01B7, 0x01B8 }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN CAPITAL LETTER EZH..LATIN CAPITAL LETTER EZH REVERSED
    Prop<::unicode::General_Category> { { 0x01B9, 0x01BA }, ::unicode::General_Category::Lowercase_Letter }, // [2] LATIN SMALL LETTER EZH REVERSED..LATIN SMALL LETTER EZH WITH TAIL
    Prop<::unicode::General_Category> { { 0x01BB, 0x01BB }, ::unicode::General_Category::Other_Letter }, // LATIN LETTER TWO WITH STROKE
    Prop<::unicode::General_Category> { { 0x01BC, 0x01BC }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER TONE FIVE
    Prop<::unicode::General_Category> { { 0x01BD, 0x01BF }, ::unicode::General_Category::Lowercase_Letter }, // [3] LATIN SMALL LETTER TONE FIVE..LATIN LETTER WYNN
    Prop<::unicode::General_Category> { { 0x01C0, 0x01C3 }, ::unicode::General_Category::Other_Letter }, // [4] LATIN LETTER DENTAL CLICK..LATIN LETTER RETROFLEX CLICK
    Prop<::unicode::General_Category> { { 0x01C4, 0x01C4 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER DZ WITH CARON
    Prop<::unicode::General_Category> { { 0x01C5, 0x01C5 }, ::unicode::General_Category::Titlecase_Letter }, // LATIN CAPITAL LETTER D WITH SMALL LETTER Z WITH CARON
    Prop<::unicode::General_Category> { { 0x01C6, 0x01C6 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER DZ WITH CARON
    Prop<::unicode::General_Category> { { 0x01C7, 0x01C7 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER LJ
    Prop<::unicode::General_Category> { { 0x01C8, 0x01C8 }, ::unicode::General_Category::Titlecase_Letter }, // LATIN CAPITAL LETTER L WITH SMALL LETTER J
    Prop<::unicode::General_Category> { { 0x01C9, 0x01C9 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER LJ
    Prop<::unicode::General_Category> { { 0x01CA, 0x01CA }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER NJ
    Prop<::unicode::General_Category> { { 0x01CB, 0x01CB }, ::unicode::General_Category::Titlecase_Letter }, // LATIN CAPITAL LETTER N WITH SMALL LETTER J
    Prop<::unicode::General_Category> { { 0x01CC, 0x01CC }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER NJ
    Prop<::unicode::General_Category> { { 0x01CD, 0x01CD }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH CARON
    Prop<::unicode::General_Category> { { 0x01CE, 0x01CE }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH CARON
    Prop<::unicode::General_Category> { { 0x01CF, 0x01CF }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER I WITH CARON
    Prop<::unicode::General_Category> { { 0x01D0, 0x01D0 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER I WITH CARON
    Prop<::unicode::General_Category> { { 0x01D1, 0x01D1 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH CARON
    Prop<::unicode::General_Category> { { 0x01D2, 0x01D2 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH CARON
    Prop<::unicode::General_Category> { { 0x01D3, 0x01D3 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH CARON
    Prop<::unicode::General_Category> { { 0x01D4, 0x01D4 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH CARON
    Prop<::unicode::General_Category> { { 0x01D5, 0x01D5 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH DIAERESIS AND MACRON
    Prop<::unicode::General_Category> { { 0x01D6, 0x01D6 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH DIAERESIS AND MACRON
    Prop<::unicode::General_Category> { { 0x01D7, 0x01D7 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH DIAERESIS AND ACUTE
    Prop<::unicode::General_Category> { { 0x01D8, 0x01D8 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH DIAERESIS AND ACUTE
    Prop<::unicode::General_Category> { { 0x01D9, 0x01D9 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH DIAERESIS AND CARON
    Prop<::unicode::General_Category> { { 0x01DA, 0x01DA }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH DIAERESIS AND CARON
    Prop<::unicode::General_Category> { { 0x01DB, 0x01DB }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH DIAERESIS AND GRAVE
    Prop<::unicode::General_Category> { { 0x01DC, 0x01DD }, ::unicode::General_Category::Lowercase_Letter }, // [2] LATIN SMALL LETTER U WITH DIAERESIS AND GRAVE..LATIN SMALL LETTER TURNED E
    Prop<::unicode::General_Category> { { 0x01DE, 0x01DE }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH DIAERESIS AND MACRON
    Prop<::unicode::General_Category> { { 0x01DF, 0x01DF }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH DIAERESIS AND MACRON
    Prop<::unicode::General_Category> { { 0x01E0, 0x01E0 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH DOT ABOVE AND MACRON
    Prop<::unicode::General_Category> { { 0x01E1, 0x01E1 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH DOT ABOVE AND MACRON
    Prop<::unicode::General_Category> { { 0x01E2, 0x01E2 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER AE WITH MACRON
    Prop<::unicode::General_Category> { { 0x01E3, 0x01E3 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER AE WITH MACRON
    Prop<::unicode::General_Category> { { 0x01E4, 0x01E4 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER G WITH STROKE
    Prop<::unicode::General_Category> { { 0x01E5, 0x01E5 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER G WITH STROKE
    Prop<::unicode::General_Category> { { 0x01E6, 0x01E6 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER G WITH CARON
    Prop<::unicode::General_Category> { { 0x01E7, 0x01E7 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER G WITH CARON
    Prop<::unicode::General_Category> { { 0x01E8, 0x01E8 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER K WITH CARON
    Prop<::unicode::General_Category> { { 0x01E9, 0x01E9 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER K WITH CARON
    Prop<::unicode::General_Category> { { 0x01EA, 0x01EA }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH OGONEK
    Prop<::unicode::General_Category> { { 0x01EB, 0x01EB }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH OGONEK
    Prop<::unicode::General_Category> { { 0x01EC, 0x01EC }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH OGONEK AND MACRON
    Prop<::unicode::General_Category> { { 0x01ED, 0x01ED }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH OGONEK AND MACRON
    Prop<::unicode::General_Category> { { 0x01EE, 0x01EE }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER EZH WITH CARON
    Prop<::unicode::General_Category> { { 0x01EF, 0x01F0 }, ::unicode::General_Category::Lowercase_Letter }, // [2] LATIN SMALL LETTER EZH WITH CARON..LATIN SMALL LETTER J WITH CARON
    Prop<::unicode::General_Category> { { 0x01F1, 0x01F1 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER DZ
    Prop<::unicode::General_Category> { { 0x01F2, 0x01F2 }, ::unicode::General_Category::Titlecase_Letter }, // LATIN CAPITAL LETTER D WITH SMALL LETTER Z
    Prop<::unicode::General_Category> { { 0x01F3, 0x01F3 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER DZ
    Prop<::unicode::General_Category> { { 0x01F4, 0x01F4 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER G WITH ACUTE
    Prop<::unicode::General_Category> { { 0x01F5, 0x01F5 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER G WITH ACUTE
    Prop<::unicode::General_Category> { { 0x01F6, 0x01F8 }, ::unicode::General_Category::Uppercase_Letter }, // [3] LATIN CAPITAL LETTER HWAIR..LATIN CAPITAL LETTER N WITH GRAVE
    Prop<::unicode::General_Category> { { 0x01F9, 0x01F9 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER N WITH GRAVE
    Prop<::unicode::General_Category> { { 0x01FA, 0x01FA }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH RING ABOVE AND ACUTE
    Prop<::unicode::General_Category> { { 0x01FB, 0x01FB }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH RING ABOVE AND ACUTE
    Prop<::unicode::General_Category> { { 0x01FC, 0x01FC }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER AE WITH ACUTE
    Prop<::unicode::General_Category> { { 0x01FD, 0x01FD }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER AE WITH ACUTE
    Prop<::unicode::General_Category> { { 0x01FE, 0x01FE }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH STROKE AND ACUTE
    Prop<::unicode::General_Category> { { 0x01FF, 0x01FF }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH STROKE AND ACUTE
    Prop<::unicode::General_Category> { { 0x0200, 0x0200 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH DOUBLE GRAVE
    Prop<::unicode::General_Category> { { 0x0201, 0x0201 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH DOUBLE GRAVE
    Prop<::unicode::General_Category> { { 0x0202, 0x0202 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0x0203, 0x0203 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0x0204, 0x0204 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH DOUBLE GRAVE
    Prop<::unicode::General_Category> { { 0x0205, 0x0205 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH DOUBLE GRAVE
    Prop<::unicode::General_Category> { { 0x0206, 0x0206 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0x0207, 0x0207 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0x0208, 0x0208 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER I WITH DOUBLE GRAVE
    Prop<::unicode::General_Category> { { 0x0209, 0x0209 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER I WITH DOUBLE GRAVE
    Prop<::unicode::General_Category> { { 0x020A, 0x020A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER I WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0x020B, 0x020B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER I WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0x020C, 0x020C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH DOUBLE GRAVE
    Prop<::unicode::General_Category> { { 0x020D, 0x020D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH DOUBLE GRAVE
    Prop<::unicode::General_Category> { { 0x020E, 0x020E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0x020F, 0x020F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0x0210, 0x0210 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER R WITH DOUBLE GRAVE
    Prop<::unicode::General_Category> { { 0x0211, 0x0211 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER R WITH DOUBLE GRAVE
    Prop<::unicode::General_Category> { { 0x0212, 0x0212 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER R WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0x0213, 0x0213 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER R WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0x0214, 0x0214 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH DOUBLE GRAVE
    Prop<::unicode::General_Category> { { 0x0215, 0x0215 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH DOUBLE GRAVE
    Prop<::unicode::General_Category> { { 0x0216, 0x0216 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0x0217, 0x0217 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0x0218, 0x0218 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER S WITH COMMA BELOW
    Prop<::unicode::General_Category> { { 0x0219, 0x0219 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH COMMA BELOW
    Prop<::unicode::General_Category> { { 0x021A, 0x021A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER T WITH COMMA BELOW
    Prop<::unicode::General_Category> { { 0x021B, 0x021B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER T WITH COMMA BELOW
    Prop<::unicode::General_Category> { { 0x021C, 0x021C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER YOGH
    Prop<::unicode::General_Category> { { 0x021D, 0x021D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER YOGH
    Prop<::unicode::General_Category> { { 0x021E, 0x021E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER H WITH CARON
    Prop<::unicode::General_Category> { { 0x021F, 0x021F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER H WITH CARON
    Prop<::unicode::General_Category> { { 0x0220, 0x0220 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER N WITH LONG RIGHT LEG
    Prop<::unicode::General_Category> { { 0x0221, 0x0221 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER D WITH CURL
    Prop<::unicode::General_Category> { { 0x0222, 0x0222 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER OU
    Prop<::unicode::General_Category> { { 0x0223, 0x0223 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER OU
    Prop<::unicode::General_Category> { { 0x0224, 0x0224 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Z WITH HOOK
    Prop<::unicode::General_Category> { { 0x0225, 0x0225 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Z WITH HOOK
    Prop<::unicode::General_Category> { { 0x0226, 0x0226 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x0227, 0x0227 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x0228, 0x0228 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x0229, 0x0229 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x022A, 0x022A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH DIAERESIS AND MACRON
    Prop<::unicode::General_Category> { { 0x022B, 0x022B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH DIAERESIS AND MACRON
    Prop<::unicode::General_Category> { { 0x022C, 0x022C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH TILDE AND MACRON
    Prop<::unicode::General_Category> { { 0x022D, 0x022D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH TILDE AND MACRON
    Prop<::unicode::General_Category> { { 0x022E, 0x022E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x022F, 0x022F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x0230, 0x0230 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH DOT ABOVE AND MACRON
    Prop<::unicode::General_Category> { { 0x0231, 0x0231 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH DOT ABOVE AND MACRON
    Prop<::unicode::General_Category> { { 0x0232, 0x0232 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Y WITH MACRON
    Prop<::unicode::General_Category> { { 0x0233, 0x0239 }, ::unicode::General_Category::Lowercase_Letter }, // [7] LATIN SMALL LETTER Y WITH MACRON..LATIN SMALL LETTER QP DIGRAPH
    Prop<::unicode::General_Category> { { 0x023A, 0x023B }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN CAPITAL LETTER A WITH STROKE..LATIN CAPITAL LETTER C WITH STROKE
    Prop<::unicode::General_Category> { { 0x023C, 0x023C }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER C WITH STROKE
    Prop<::unicode::General_Category> { { 0x023D, 0x023E }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN CAPITAL LETTER L WITH BAR..LATIN CAPITAL LETTER T WITH DIAGONAL STROKE
    Prop<::unicode::General_Category> { { 0x023F, 0x0240 }, ::unicode::General_Category::Lowercase_Letter }, // [2] LATIN SMALL LETTER S WITH SWASH TAIL..LATIN SMALL LETTER Z WITH SWASH TAIL
    Prop<::unicode::General_Category> { { 0x0241, 0x0241 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER GLOTTAL STOP
    Prop<::unicode::General_Category> { { 0x0242, 0x0242 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER GLOTTAL STOP
    Prop<::unicode::General_Category> { { 0x0243, 0x0246 }, ::unicode::General_Category::Uppercase_Letter }, // [4] LATIN CAPITAL LETTER B WITH STROKE..LATIN CAPITAL LETTER E WITH STROKE
    Prop<::unicode::General_Category> { { 0x0247, 0x0247 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH STROKE
    Prop<::unicode::General_Category> { { 0x0248, 0x0248 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER J WITH STROKE
    Prop<::unicode::General_Category> { { 0x0249, 0x0249 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER J WITH STROKE
    Prop<::unicode::General_Category> { { 0x024A, 0x024A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER SMALL Q WITH HOOK TAIL
    Prop<::unicode::General_Category> { { 0x024B, 0x024B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Q WITH HOOK TAIL
    Prop<::unicode::General_Category> { { 0x024C, 0x024C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER R WITH STROKE
    Prop<::unicode::General_Category> { { 0x024D, 0x024D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER R WITH STROKE
    Prop<::unicode::General_Category> { { 0x024E, 0x024E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Y WITH STROKE
    Prop<::unicode::General_Category> { { 0x024F, 0x0293 }, ::unicode::General_Category::Lowercase_Letter }, // [69] LATIN SMALL LETTER Y WITH STROKE..LATIN SMALL LETTER EZH WITH CURL
    Prop<::unicode::General_Category> { { 0x0294, 0x0295 }, ::unicode::General_Category::Other_Letter }, // [2] LATIN LETTER GLOTTAL STOP..LATIN LETTER PHARYNGEAL VOICED FRICATIVE
    Prop<::unicode::General_Category> { { 0x0296, 0x02AF }, ::unicode::General_Category::Lowercase_Letter }, // [26] LATIN LETTER INVERTED GLOTTAL STOP..LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    Prop<::unicode::General_Category> { { 0x02B0, 0x02C1 }, ::unicode::General_Category::Modifier_Letter }, // [18] MODIFIER LETTER SMALL H..MODIFIER LETTER REVERSED GLOTTAL STOP
    Prop<::unicode::General_Category> { { 0x02C2, 0x02C5 }, ::unicode::General_Category::Modifier_Symbol }, // [4] MODIFIER LETTER LEFT ARROWHEAD..MODIFIER LETTER DOWN ARROWHEAD
    Prop<::unicode::General_Category> { { 0x02C6, 0x02D1 }, ::unicode::General_Category::Modifier_Letter }, // [12] MODIFIER LETTER CIRCUMFLEX ACCENT..MODIFIER LETTER HALF TRIANGULAR COLON
    Prop<::unicode::General_Category> { { 0x02D2, 0x02DF }, ::unicode::General_Category::Modifier_Symbol }, // [14] MODIFIER LETTER CENTRED RIGHT HALF RING..MODIFIER LETTER CROSS ACCENT
    Prop<::unicode::General_Category> { { 0x02E0, 0x02E4 }, ::unicode::General_Category::Modifier_Letter }, // [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    Prop<::unicode::General_Category> { { 0x02E5, 0x02EB }, ::unicode::General_Category::Modifier_Symbol }, // [7] MODIFIER LETTER EXTRA-HIGH TONE BAR..MODIFIER LETTER YANG DEPARTING TONE MARK
    Prop<::unicode::General_Category> { { 0x02EC, 0x02EC }, ::unicode::General_Category::Modifier_Letter }, // MODIFIER LETTER VOICING
    Prop<::unicode::General_Category> { { 0x02ED, 0x02ED }, ::unicode::General_Category::Modifier_Symbol }, // MODIFIER LETTER UNASPIRATED
    Prop<::unicode::General_Category> { { 0x02EE, 0x02EE }, ::unicode::General_Category::Modifier_Letter }, // MODIFIER LETTER DOUBLE APOSTROPHE
    Prop<::unicode::General_Category> { { 0x02EF, 0x02FF }, ::unicode::General_Category::Modifier_Symbol }, // [17] MODIFIER LETTER LOW DOWN ARROWHEAD..MODIFIER LETTER LOW LEFT ARROW
    Prop<::unicode::General_Category> { { 0x0300, 0x036F }, ::unicode::General_Category::Nonspacing_Mark }, // [112] COMBINING GRAVE ACCENT..COMBINING LATIN SMALL LETTER X
    Prop<::unicode::General_Category> { { 0x0370, 0x0370 }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL LETTER HETA
    Prop<::unicode::General_Category> { { 0x0371, 0x0371 }, ::unicode::General_Category::Lowercase_Letter }, // GREEK SMALL LETTER HETA
    Prop<::unicode::General_Category> { { 0x0372, 0x0372 }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL LETTER ARCHAIC SAMPI
    Prop<::unicode::General_Category> { { 0x0373, 0x0373 }, ::unicode::General_Category::Lowercase_Letter }, // GREEK SMALL LETTER ARCHAIC SAMPI
    Prop<::unicode::General_Category> { { 0x0374, 0x0374 }, ::unicode::General_Category::Modifier_Letter }, // GREEK NUMERAL SIGN
    Prop<::unicode::General_Category> { { 0x0375, 0x0375 }, ::unicode::General_Category::Modifier_Symbol }, // GREEK LOWER NUMERAL SIGN
    Prop<::unicode::General_Category> { { 0x0376, 0x0376 }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA
    Prop<::unicode::General_Category> { { 0x0377, 0x0377 }, ::unicode::General_Category::Lowercase_Letter }, // GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Prop<::unicode::General_Category> { { 0x0378, 0x0379 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0378>..<reserved-0379>
    Prop<::unicode::General_Category> { { 0x037A, 0x037A }, ::unicode::General_Category::Modifier_Letter }, // GREEK YPOGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x037B, 0x037D }, ::unicode::General_Category::Lowercase_Letter }, // [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Prop<::unicode::General_Category> { { 0x037E, 0x037E }, ::unicode::General_Category::Other_Punctuation }, // GREEK QUESTION MARK
    Prop<::unicode::General_Category> { { 0x037F, 0x037F }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL LETTER YOT
    Prop<::unicode::General_Category> { { 0x0380, 0x0383 }, ::unicode::General_Category::Unassigned }, // [4] <reserved-0380>..<reserved-0383>
    Prop<::unicode::General_Category> { { 0x0384, 0x0385 }, ::unicode::General_Category::Modifier_Symbol }, // [2] GREEK TONOS..GREEK DIALYTIKA TONOS
    Prop<::unicode::General_Category> { { 0x0386, 0x0386 }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL LETTER ALPHA WITH TONOS
    Prop<::unicode::General_Category> { { 0x0387, 0x0387 }, ::unicode::General_Category::Other_Punctuation }, // GREEK ANO TELEIA
    Prop<::unicode::General_Category> { { 0x0388, 0x038A }, ::unicode::General_Category::Uppercase_Letter }, // [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Prop<::unicode::General_Category> { { 0x038B, 0x038B }, ::unicode::General_Category::Unassigned }, // <reserved-038B>
    Prop<::unicode::General_Category> { { 0x038C, 0x038C }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL LETTER OMICRON WITH TONOS
    Prop<::unicode::General_Category> { { 0x038D, 0x038D }, ::unicode::General_Category::Unassigned }, // <reserved-038D>
    Prop<::unicode::General_Category> { { 0x038E, 0x038F }, ::unicode::General_Category::Uppercase_Letter }, // [2] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER OMEGA WITH TONOS
    Prop<::unicode::General_Category> { { 0x0390, 0x0390 }, ::unicode::General_Category::Lowercase_Letter }, // GREEK SMALL LETTER IOTA WITH DIALYTIKA AND TONOS
    Prop<::unicode::General_Category> { { 0x0391, 0x03A1 }, ::unicode::General_Category::Uppercase_Letter }, // [17] GREEK CAPITAL LETTER ALPHA..GREEK CAPITAL LETTER RHO
    Prop<::unicode::General_Category> { { 0x03A2, 0x03A2 }, ::unicode::General_Category::Unassigned }, // <reserved-03A2>
    Prop<::unicode::General_Category> { { 0x03A3, 0x03AB }, ::unicode::General_Category::Uppercase_Letter }, // [9] GREEK CAPITAL LETTER SIGMA..GREEK CAPITAL LETTER UPSILON WITH DIALYTIKA
    Prop<::unicode::General_Category> { { 0x03AC, 0x03CE }, ::unicode::General_Category::Lowercase_Letter }, // [35] GREEK SMALL LETTER ALPHA WITH TONOS..GREEK SMALL LETTER OMEGA WITH TONOS
    Prop<::unicode::General_Category> { { 0x03CF, 0x03CF }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL KAI SYMBOL
    Prop<::unicode::General_Category> { { 0x03D0, 0x03D1 }, ::unicode::General_Category::Lowercase_Letter }, // [2] GREEK BETA SYMBOL..GREEK THETA SYMBOL
    Prop<::unicode::General_Category> { { 0x03D2, 0x03D4 }, ::unicode::General_Category::Uppercase_Letter }, // [3] GREEK UPSILON WITH HOOK SYMBOL..GREEK UPSILON WITH DIAERESIS AND HOOK SYMBOL
    Prop<::unicode::General_Category> { { 0x03D5, 0x03D7 }, ::unicode::General_Category::Lowercase_Letter }, // [3] GREEK PHI SYMBOL..GREEK KAI SYMBOL
    Prop<::unicode::General_Category> { { 0x03D8, 0x03D8 }, ::unicode::General_Category::Uppercase_Letter }, // GREEK LETTER ARCHAIC KOPPA
    Prop<::unicode::General_Category> { { 0x03D9, 0x03D9 }, ::unicode::General_Category::Lowercase_Letter }, // GREEK SMALL LETTER ARCHAIC KOPPA
    Prop<::unicode::General_Category> { { 0x03DA, 0x03DA }, ::unicode::General_Category::Uppercase_Letter }, // GREEK LETTER STIGMA
    Prop<::unicode::General_Category> { { 0x03DB, 0x03DB }, ::unicode::General_Category::Lowercase_Letter }, // GREEK SMALL LETTER STIGMA
    Prop<::unicode::General_Category> { { 0x03DC, 0x03DC }, ::unicode::General_Category::Uppercase_Letter }, // GREEK LETTER DIGAMMA
    Prop<::unicode::General_Category> { { 0x03DD, 0x03DD }, ::unicode::General_Category::Lowercase_Letter }, // GREEK SMALL LETTER DIGAMMA
    Prop<::unicode::General_Category> { { 0x03DE, 0x03DE }, ::unicode::General_Category::Uppercase_Letter }, // GREEK LETTER KOPPA
    Prop<::unicode::General_Category> { { 0x03DF, 0x03DF }, ::unicode::General_Category::Lowercase_Letter }, // GREEK SMALL LETTER KOPPA
    Prop<::unicode::General_Category> { { 0x03E0, 0x03E0 }, ::unicode::General_Category::Uppercase_Letter }, // GREEK LETTER SAMPI
    Prop<::unicode::General_Category> { { 0x03E1, 0x03E1 }, ::unicode::General_Category::Lowercase_Letter }, // GREEK SMALL LETTER SAMPI
    Prop<::unicode::General_Category> { { 0x03E2, 0x03E2 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER SHEI
    Prop<::unicode::General_Category> { { 0x03E3, 0x03E3 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER SHEI
    Prop<::unicode::General_Category> { { 0x03E4, 0x03E4 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER FEI
    Prop<::unicode::General_Category> { { 0x03E5, 0x03E5 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER FEI
    Prop<::unicode::General_Category> { { 0x03E6, 0x03E6 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER KHEI
    Prop<::unicode::General_Category> { { 0x03E7, 0x03E7 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER KHEI
    Prop<::unicode::General_Category> { { 0x03E8, 0x03E8 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER HORI
    Prop<::unicode::General_Category> { { 0x03E9, 0x03E9 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER HORI
    Prop<::unicode::General_Category> { { 0x03EA, 0x03EA }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER GANGIA
    Prop<::unicode::General_Category> { { 0x03EB, 0x03EB }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER GANGIA
    Prop<::unicode::General_Category> { { 0x03EC, 0x03EC }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER SHIMA
    Prop<::unicode::General_Category> { { 0x03ED, 0x03ED }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER SHIMA
    Prop<::unicode::General_Category> { { 0x03EE, 0x03EE }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER DEI
    Prop<::unicode::General_Category> { { 0x03EF, 0x03F3 }, ::unicode::General_Category::Lowercase_Letter }, // [5] COPTIC SMALL LETTER DEI..GREEK LETTER YOT
    Prop<::unicode::General_Category> { { 0x03F4, 0x03F4 }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL THETA SYMBOL
    Prop<::unicode::General_Category> { { 0x03F5, 0x03F5 }, ::unicode::General_Category::Lowercase_Letter }, // GREEK LUNATE EPSILON SYMBOL
    Prop<::unicode::General_Category> { { 0x03F6, 0x03F6 }, ::unicode::General_Category::Math_Symbol }, // GREEK REVERSED LUNATE EPSILON SYMBOL
    Prop<::unicode::General_Category> { { 0x03F7, 0x03F7 }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL LETTER SHO
    Prop<::unicode::General_Category> { { 0x03F8, 0x03F8 }, ::unicode::General_Category::Lowercase_Letter }, // GREEK SMALL LETTER SHO
    Prop<::unicode::General_Category> { { 0x03F9, 0x03FA }, ::unicode::General_Category::Uppercase_Letter }, // [2] GREEK CAPITAL LUNATE SIGMA SYMBOL..GREEK CAPITAL LETTER SAN
    Prop<::unicode::General_Category> { { 0x03FB, 0x03FC }, ::unicode::General_Category::Lowercase_Letter }, // [2] GREEK SMALL LETTER SAN..GREEK RHO WITH STROKE SYMBOL
    Prop<::unicode::General_Category> { { 0x03FD, 0x042F }, ::unicode::General_Category::Uppercase_Letter }, // [51] GREEK CAPITAL REVERSED LUNATE SIGMA SYMBOL..CYRILLIC CAPITAL LETTER YA
    Prop<::unicode::General_Category> { { 0x0430, 0x045F }, ::unicode::General_Category::Lowercase_Letter }, // [48] CYRILLIC SMALL LETTER A..CYRILLIC SMALL LETTER DZHE
    Prop<::unicode::General_Category> { { 0x0460, 0x0460 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER OMEGA
    Prop<::unicode::General_Category> { { 0x0461, 0x0461 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER OMEGA
    Prop<::unicode::General_Category> { { 0x0462, 0x0462 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER YAT
    Prop<::unicode::General_Category> { { 0x0463, 0x0463 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER YAT
    Prop<::unicode::General_Category> { { 0x0464, 0x0464 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER IOTIFIED E
    Prop<::unicode::General_Category> { { 0x0465, 0x0465 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER IOTIFIED E
    Prop<::unicode::General_Category> { { 0x0466, 0x0466 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER LITTLE YUS
    Prop<::unicode::General_Category> { { 0x0467, 0x0467 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER LITTLE YUS
    Prop<::unicode::General_Category> { { 0x0468, 0x0468 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER IOTIFIED LITTLE YUS
    Prop<::unicode::General_Category> { { 0x0469, 0x0469 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER IOTIFIED LITTLE YUS
    Prop<::unicode::General_Category> { { 0x046A, 0x046A }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER BIG YUS
    Prop<::unicode::General_Category> { { 0x046B, 0x046B }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER BIG YUS
    Prop<::unicode::General_Category> { { 0x046C, 0x046C }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER IOTIFIED BIG YUS
    Prop<::unicode::General_Category> { { 0x046D, 0x046D }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER IOTIFIED BIG YUS
    Prop<::unicode::General_Category> { { 0x046E, 0x046E }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KSI
    Prop<::unicode::General_Category> { { 0x046F, 0x046F }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KSI
    Prop<::unicode::General_Category> { { 0x0470, 0x0470 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER PSI
    Prop<::unicode::General_Category> { { 0x0471, 0x0471 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER PSI
    Prop<::unicode::General_Category> { { 0x0472, 0x0472 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER FITA
    Prop<::unicode::General_Category> { { 0x0473, 0x0473 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER FITA
    Prop<::unicode::General_Category> { { 0x0474, 0x0474 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER IZHITSA
    Prop<::unicode::General_Category> { { 0x0475, 0x0475 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER IZHITSA
    Prop<::unicode::General_Category> { { 0x0476, 0x0476 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT
    Prop<::unicode::General_Category> { { 0x0477, 0x0477 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT
    Prop<::unicode::General_Category> { { 0x0478, 0x0478 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER UK
    Prop<::unicode::General_Category> { { 0x0479, 0x0479 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER UK
    Prop<::unicode::General_Category> { { 0x047A, 0x047A }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ROUND OMEGA
    Prop<::unicode::General_Category> { { 0x047B, 0x047B }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ROUND OMEGA
    Prop<::unicode::General_Category> { { 0x047C, 0x047C }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER OMEGA WITH TITLO
    Prop<::unicode::General_Category> { { 0x047D, 0x047D }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER OMEGA WITH TITLO
    Prop<::unicode::General_Category> { { 0x047E, 0x047E }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER OT
    Prop<::unicode::General_Category> { { 0x047F, 0x047F }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER OT
    Prop<::unicode::General_Category> { { 0x0480, 0x0480 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KOPPA
    Prop<::unicode::General_Category> { { 0x0481, 0x0481 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KOPPA
    Prop<::unicode::General_Category> { { 0x0482, 0x0482 }, ::unicode::General_Category::Other_Symbol }, // CYRILLIC THOUSANDS SIGN
    Prop<::unicode::General_Category> { { 0x0483, 0x0487 }, ::unicode::General_Category::Nonspacing_Mark }, // [5] COMBINING CYRILLIC TITLO..COMBINING CYRILLIC POKRYTIE
    Prop<::unicode::General_Category> { { 0x0488, 0x0489 }, ::unicode::General_Category::Enclosing_Mark }, // [2] COMBINING CYRILLIC HUNDRED THOUSANDS SIGN..COMBINING CYRILLIC MILLIONS SIGN
    Prop<::unicode::General_Category> { { 0x048A, 0x048A }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER SHORT I WITH TAIL
    Prop<::unicode::General_Category> { { 0x048B, 0x048B }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER SHORT I WITH TAIL
    Prop<::unicode::General_Category> { { 0x048C, 0x048C }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER SEMISOFT SIGN
    Prop<::unicode::General_Category> { { 0x048D, 0x048D }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER SEMISOFT SIGN
    Prop<::unicode::General_Category> { { 0x048E, 0x048E }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ER WITH TICK
    Prop<::unicode::General_Category> { { 0x048F, 0x048F }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ER WITH TICK
    Prop<::unicode::General_Category> { { 0x0490, 0x0490 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER GHE WITH UPTURN
    Prop<::unicode::General_Category> { { 0x0491, 0x0491 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER GHE WITH UPTURN
    Prop<::unicode::General_Category> { { 0x0492, 0x0492 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER GHE WITH STROKE
    Prop<::unicode::General_Category> { { 0x0493, 0x0493 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER GHE WITH STROKE
    Prop<::unicode::General_Category> { { 0x0494, 0x0494 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER GHE WITH MIDDLE HOOK
    Prop<::unicode::General_Category> { { 0x0495, 0x0495 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER GHE WITH MIDDLE HOOK
    Prop<::unicode::General_Category> { { 0x0496, 0x0496 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ZHE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x0497, 0x0497 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ZHE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x0498, 0x0498 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ZE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x0499, 0x0499 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ZE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x049A, 0x049A }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KA WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x049B, 0x049B }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KA WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x049C, 0x049C }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KA WITH VERTICAL STROKE
    Prop<::unicode::General_Category> { { 0x049D, 0x049D }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KA WITH VERTICAL STROKE
    Prop<::unicode::General_Category> { { 0x049E, 0x049E }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KA WITH STROKE
    Prop<::unicode::General_Category> { { 0x049F, 0x049F }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KA WITH STROKE
    Prop<::unicode::General_Category> { { 0x04A0, 0x04A0 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER BASHKIR KA
    Prop<::unicode::General_Category> { { 0x04A1, 0x04A1 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER BASHKIR KA
    Prop<::unicode::General_Category> { { 0x04A2, 0x04A2 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER EN WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04A3, 0x04A3 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER EN WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04A4, 0x04A4 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LIGATURE EN GHE
    Prop<::unicode::General_Category> { { 0x04A5, 0x04A5 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LIGATURE EN GHE
    Prop<::unicode::General_Category> { { 0x04A6, 0x04A6 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER PE WITH MIDDLE HOOK
    Prop<::unicode::General_Category> { { 0x04A7, 0x04A7 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER PE WITH MIDDLE HOOK
    Prop<::unicode::General_Category> { { 0x04A8, 0x04A8 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ABKHASIAN HA
    Prop<::unicode::General_Category> { { 0x04A9, 0x04A9 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ABKHASIAN HA
    Prop<::unicode::General_Category> { { 0x04AA, 0x04AA }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ES WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04AB, 0x04AB }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ES WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04AC, 0x04AC }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER TE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04AD, 0x04AD }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER TE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04AE, 0x04AE }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER STRAIGHT U
    Prop<::unicode::General_Category> { { 0x04AF, 0x04AF }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER STRAIGHT U
    Prop<::unicode::General_Category> { { 0x04B0, 0x04B0 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER STRAIGHT U WITH STROKE
    Prop<::unicode::General_Category> { { 0x04B1, 0x04B1 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER STRAIGHT U WITH STROKE
    Prop<::unicode::General_Category> { { 0x04B2, 0x04B2 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER HA WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04B3, 0x04B3 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER HA WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04B4, 0x04B4 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LIGATURE TE TSE
    Prop<::unicode::General_Category> { { 0x04B5, 0x04B5 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LIGATURE TE TSE
    Prop<::unicode::General_Category> { { 0x04B6, 0x04B6 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER CHE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04B7, 0x04B7 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER CHE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04B8, 0x04B8 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER CHE WITH VERTICAL STROKE
    Prop<::unicode::General_Category> { { 0x04B9, 0x04B9 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER CHE WITH VERTICAL STROKE
    Prop<::unicode::General_Category> { { 0x04BA, 0x04BA }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER SHHA
    Prop<::unicode::General_Category> { { 0x04BB, 0x04BB }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER SHHA
    Prop<::unicode::General_Category> { { 0x04BC, 0x04BC }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ABKHASIAN CHE
    Prop<::unicode::General_Category> { { 0x04BD, 0x04BD }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ABKHASIAN CHE
    Prop<::unicode::General_Category> { { 0x04BE, 0x04BE }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ABKHASIAN CHE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04BF, 0x04BF }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ABKHASIAN CHE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04C0, 0x04C1 }, ::unicode::General_Category::Uppercase_Letter }, // [2] CYRILLIC LETTER PALOCHKA..CYRILLIC CAPITAL LETTER ZHE WITH BREVE
    Prop<::unicode::General_Category> { { 0x04C2, 0x04C2 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ZHE WITH BREVE
    Prop<::unicode::General_Category> { { 0x04C3, 0x04C3 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KA WITH HOOK
    Prop<::unicode::General_Category> { { 0x04C4, 0x04C4 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KA WITH HOOK
    Prop<::unicode::General_Category> { { 0x04C5, 0x04C5 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER EL WITH TAIL
    Prop<::unicode::General_Category> { { 0x04C6, 0x04C6 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER EL WITH TAIL
    Prop<::unicode::General_Category> { { 0x04C7, 0x04C7 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER EN WITH HOOK
    Prop<::unicode::General_Category> { { 0x04C8, 0x04C8 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER EN WITH HOOK
    Prop<::unicode::General_Category> { { 0x04C9, 0x04C9 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER EN WITH TAIL
    Prop<::unicode::General_Category> { { 0x04CA, 0x04CA }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER EN WITH TAIL
    Prop<::unicode::General_Category> { { 0x04CB, 0x04CB }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KHAKASSIAN CHE
    Prop<::unicode::General_Category> { { 0x04CC, 0x04CC }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KHAKASSIAN CHE
    Prop<::unicode::General_Category> { { 0x04CD, 0x04CD }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER EM WITH TAIL
    Prop<::unicode::General_Category> { { 0x04CE, 0x04CF }, ::unicode::General_Category::Lowercase_Letter }, // [2] CYRILLIC SMALL LETTER EM WITH TAIL..CYRILLIC SMALL LETTER PALOCHKA
    Prop<::unicode::General_Category> { { 0x04D0, 0x04D0 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER A WITH BREVE
    Prop<::unicode::General_Category> { { 0x04D1, 0x04D1 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER A WITH BREVE
    Prop<::unicode::General_Category> { { 0x04D2, 0x04D2 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER A WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04D3, 0x04D3 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER A WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04D4, 0x04D4 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LIGATURE A IE
    Prop<::unicode::General_Category> { { 0x04D5, 0x04D5 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LIGATURE A IE
    Prop<::unicode::General_Category> { { 0x04D6, 0x04D6 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER IE WITH BREVE
    Prop<::unicode::General_Category> { { 0x04D7, 0x04D7 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER IE WITH BREVE
    Prop<::unicode::General_Category> { { 0x04D8, 0x04D8 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER SCHWA
    Prop<::unicode::General_Category> { { 0x04D9, 0x04D9 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER SCHWA
    Prop<::unicode::General_Category> { { 0x04DA, 0x04DA }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER SCHWA WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04DB, 0x04DB }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER SCHWA WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04DC, 0x04DC }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ZHE WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04DD, 0x04DD }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ZHE WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04DE, 0x04DE }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ZE WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04DF, 0x04DF }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ZE WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04E0, 0x04E0 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ABKHASIAN DZE
    Prop<::unicode::General_Category> { { 0x04E1, 0x04E1 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ABKHASIAN DZE
    Prop<::unicode::General_Category> { { 0x04E2, 0x04E2 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER I WITH MACRON
    Prop<::unicode::General_Category> { { 0x04E3, 0x04E3 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER I WITH MACRON
    Prop<::unicode::General_Category> { { 0x04E4, 0x04E4 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER I WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04E5, 0x04E5 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER I WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04E6, 0x04E6 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER O WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04E7, 0x04E7 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER O WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04E8, 0x04E8 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER BARRED O
    Prop<::unicode::General_Category> { { 0x04E9, 0x04E9 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER BARRED O
    Prop<::unicode::General_Category> { { 0x04EA, 0x04EA }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER BARRED O WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04EB, 0x04EB }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER BARRED O WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04EC, 0x04EC }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER E WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04ED, 0x04ED }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER E WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04EE, 0x04EE }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER U WITH MACRON
    Prop<::unicode::General_Category> { { 0x04EF, 0x04EF }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER U WITH MACRON
    Prop<::unicode::General_Category> { { 0x04F0, 0x04F0 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER U WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04F1, 0x04F1 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER U WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04F2, 0x04F2 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER U WITH DOUBLE ACUTE
    Prop<::unicode::General_Category> { { 0x04F3, 0x04F3 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER U WITH DOUBLE ACUTE
    Prop<::unicode::General_Category> { { 0x04F4, 0x04F4 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER CHE WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04F5, 0x04F5 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER CHE WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04F6, 0x04F6 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER GHE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04F7, 0x04F7 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER GHE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x04F8, 0x04F8 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER YERU WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04F9, 0x04F9 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER YERU WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x04FA, 0x04FA }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER GHE WITH STROKE AND HOOK
    Prop<::unicode::General_Category> { { 0x04FB, 0x04FB }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER GHE WITH STROKE AND HOOK
    Prop<::unicode::General_Category> { { 0x04FC, 0x04FC }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER HA WITH HOOK
    Prop<::unicode::General_Category> { { 0x04FD, 0x04FD }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER HA WITH HOOK
    Prop<::unicode::General_Category> { { 0x04FE, 0x04FE }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER HA WITH STROKE
    Prop<::unicode::General_Category> { { 0x04FF, 0x04FF }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER HA WITH STROKE
    Prop<::unicode::General_Category> { { 0x0500, 0x0500 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KOMI DE
    Prop<::unicode::General_Category> { { 0x0501, 0x0501 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KOMI DE
    Prop<::unicode::General_Category> { { 0x0502, 0x0502 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KOMI DJE
    Prop<::unicode::General_Category> { { 0x0503, 0x0503 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KOMI DJE
    Prop<::unicode::General_Category> { { 0x0504, 0x0504 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KOMI ZJE
    Prop<::unicode::General_Category> { { 0x0505, 0x0505 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KOMI ZJE
    Prop<::unicode::General_Category> { { 0x0506, 0x0506 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KOMI DZJE
    Prop<::unicode::General_Category> { { 0x0507, 0x0507 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KOMI DZJE
    Prop<::unicode::General_Category> { { 0x0508, 0x0508 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KOMI LJE
    Prop<::unicode::General_Category> { { 0x0509, 0x0509 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KOMI LJE
    Prop<::unicode::General_Category> { { 0x050A, 0x050A }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KOMI NJE
    Prop<::unicode::General_Category> { { 0x050B, 0x050B }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KOMI NJE
    Prop<::unicode::General_Category> { { 0x050C, 0x050C }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KOMI SJE
    Prop<::unicode::General_Category> { { 0x050D, 0x050D }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KOMI SJE
    Prop<::unicode::General_Category> { { 0x050E, 0x050E }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER KOMI TJE
    Prop<::unicode::General_Category> { { 0x050F, 0x050F }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER KOMI TJE
    Prop<::unicode::General_Category> { { 0x0510, 0x0510 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER REVERSED ZE
    Prop<::unicode::General_Category> { { 0x0511, 0x0511 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER REVERSED ZE
    Prop<::unicode::General_Category> { { 0x0512, 0x0512 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER EL WITH HOOK
    Prop<::unicode::General_Category> { { 0x0513, 0x0513 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER EL WITH HOOK
    Prop<::unicode::General_Category> { { 0x0514, 0x0514 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER LHA
    Prop<::unicode::General_Category> { { 0x0515, 0x0515 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER LHA
    Prop<::unicode::General_Category> { { 0x0516, 0x0516 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER RHA
    Prop<::unicode::General_Category> { { 0x0517, 0x0517 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER RHA
    Prop<::unicode::General_Category> { { 0x0518, 0x0518 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER YAE
    Prop<::unicode::General_Category> { { 0x0519, 0x0519 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER YAE
    Prop<::unicode::General_Category> { { 0x051A, 0x051A }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER QA
    Prop<::unicode::General_Category> { { 0x051B, 0x051B }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER QA
    Prop<::unicode::General_Category> { { 0x051C, 0x051C }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER WE
    Prop<::unicode::General_Category> { { 0x051D, 0x051D }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER WE
    Prop<::unicode::General_Category> { { 0x051E, 0x051E }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ALEUT KA
    Prop<::unicode::General_Category> { { 0x051F, 0x051F }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ALEUT KA
    Prop<::unicode::General_Category> { { 0x0520, 0x0520 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER EL WITH MIDDLE HOOK
    Prop<::unicode::General_Category> { { 0x0521, 0x0521 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER EL WITH MIDDLE HOOK
    Prop<::unicode::General_Category> { { 0x0522, 0x0522 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER EN WITH MIDDLE HOOK
    Prop<::unicode::General_Category> { { 0x0523, 0x0523 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER EN WITH MIDDLE HOOK
    Prop<::unicode::General_Category> { { 0x0524, 0x0524 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER PE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x0525, 0x0525 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER PE WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x0526, 0x0526 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER SHHA WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x0527, 0x0527 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER SHHA WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x0528, 0x0528 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER EN WITH LEFT HOOK
    Prop<::unicode::General_Category> { { 0x0529, 0x0529 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER EN WITH LEFT HOOK
    Prop<::unicode::General_Category> { { 0x052A, 0x052A }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER DZZHE
    Prop<::unicode::General_Category> { { 0x052B, 0x052B }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER DZZHE
    Prop<::unicode::General_Category> { { 0x052C, 0x052C }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER DCHE
    Prop<::unicode::General_Category> { { 0x052D, 0x052D }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER DCHE
    Prop<::unicode::General_Category> { { 0x052E, 0x052E }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER EL WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x052F, 0x052F }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER EL WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x0530, 0x0530 }, ::unicode::General_Category::Unassigned }, // <reserved-0530>
    Prop<::unicode::General_Category> { { 0x0531, 0x0556 }, ::unicode::General_Category::Uppercase_Letter }, // [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Prop<::unicode::General_Category> { { 0x0557, 0x0558 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0557>..<reserved-0558>
    Prop<::unicode::General_Category> { { 0x0559, 0x0559 }, ::unicode::General_Category::Modifier_Letter }, // ARMENIAN MODIFIER LETTER LEFT HALF RING
    Prop<::unicode::General_Category> { { 0x055A, 0x055F }, ::unicode::General_Category::Other_Punctuation }, // [6] ARMENIAN APOSTROPHE..ARMENIAN ABBREVIATION MARK
    Prop<::unicode::General_Category> { { 0x0560, 0x0588 }, ::unicode::General_Category::Lowercase_Letter }, // [41] ARMENIAN SMALL LETTER TURNED AYB..ARMENIAN SMALL LETTER YI WITH STROKE
    Prop<::unicode::General_Category> { { 0x0589, 0x0589 }, ::unicode::General_Category::Other_Punctuation }, // ARMENIAN FULL STOP
    Prop<::unicode::General_Category> { { 0x058A, 0x058A }, ::unicode::General_Category::Dash_Punctuation }, // ARMENIAN HYPHEN
    Prop<::unicode::General_Category> { { 0x058B, 0x058C }, ::unicode::General_Category::Unassigned }, // [2] <reserved-058B>..<reserved-058C>
    Prop<::unicode::General_Category> { { 0x058D, 0x058E }, ::unicode::General_Category::Other_Symbol }, // [2] RIGHT-FACING ARMENIAN ETERNITY SIGN..LEFT-FACING ARMENIAN ETERNITY SIGN
    Prop<::unicode::General_Category> { { 0x058F, 0x058F }, ::unicode::General_Category::Currency_Symbol }, // ARMENIAN DRAM SIGN
    Prop<::unicode::General_Category> { { 0x0590, 0x0590 }, ::unicode::General_Category::Unassigned }, // <reserved-0590>
    Prop<::unicode::General_Category> { { 0x0591, 0x05BD }, ::unicode::General_Category::Nonspacing_Mark }, // [45] HEBREW ACCENT ETNAHTA..HEBREW POINT METEG
    Prop<::unicode::General_Category> { { 0x05BE, 0x05BE }, ::unicode::General_Category::Dash_Punctuation }, // HEBREW PUNCTUATION MAQAF
    Prop<::unicode::General_Category> { { 0x05BF, 0x05BF }, ::unicode::General_Category::Nonspacing_Mark }, // HEBREW POINT RAFE
    Prop<::unicode::General_Category> { { 0x05C0, 0x05C0 }, ::unicode::General_Category::Other_Punctuation }, // HEBREW PUNCTUATION PASEQ
    Prop<::unicode::General_Category> { { 0x05C1, 0x05C2 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] HEBREW POINT SHIN DOT..HEBREW POINT SIN DOT
    Prop<::unicode::General_Category> { { 0x05C3, 0x05C3 }, ::unicode::General_Category::Other_Punctuation }, // HEBREW PUNCTUATION SOF PASUQ
    Prop<::unicode::General_Category> { { 0x05C4, 0x05C5 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] HEBREW MARK UPPER DOT..HEBREW MARK LOWER DOT
    Prop<::unicode::General_Category> { { 0x05C6, 0x05C6 }, ::unicode::General_Category::Other_Punctuation }, // HEBREW PUNCTUATION NUN HAFUKHA
    Prop<::unicode::General_Category> { { 0x05C7, 0x05C7 }, ::unicode::General_Category::Nonspacing_Mark }, // HEBREW POINT QAMATS QATAN
    Prop<::unicode::General_Category> { { 0x05C8, 0x05CF }, ::unicode::General_Category::Unassigned }, // [8] <reserved-05C8>..<reserved-05CF>
    Prop<::unicode::General_Category> { { 0x05D0, 0x05EA }, ::unicode::General_Category::Other_Letter }, // [27] HEBREW LETTER ALEF..HEBREW LETTER TAV
    Prop<::unicode::General_Category> { { 0x05EB, 0x05EE }, ::unicode::General_Category::Unassigned }, // [4] <reserved-05EB>..<reserved-05EE>
    Prop<::unicode::General_Category> { { 0x05EF, 0x05F2 }, ::unicode::General_Category::Other_Letter }, // [4] HEBREW YOD TRIANGLE..HEBREW LIGATURE YIDDISH DOUBLE YOD
    Prop<::unicode::General_Category> { { 0x05F3, 0x05F4 }, ::unicode::General_Category::Other_Punctuation }, // [2] HEBREW PUNCTUATION GERESH..HEBREW PUNCTUATION GERSHAYIM
    Prop<::unicode::General_Category> { { 0x05F5, 0x05FF }, ::unicode::General_Category::Unassigned }, // [11] <reserved-05F5>..<reserved-05FF>
    Prop<::unicode::General_Category> { { 0x0600, 0x0605 }, ::unicode::General_Category::Format }, // [6] ARABIC NUMBER SIGN..ARABIC NUMBER MARK ABOVE
    Prop<::unicode::General_Category> { { 0x0606, 0x0608 }, ::unicode::General_Category::Math_Symbol }, // [3] ARABIC-INDIC CUBE ROOT..ARABIC RAY
    Prop<::unicode::General_Category> { { 0x0609, 0x060A }, ::unicode::General_Category::Other_Punctuation }, // [2] ARABIC-INDIC PER MILLE SIGN..ARABIC-INDIC PER TEN THOUSAND SIGN
    Prop<::unicode::General_Category> { { 0x060B, 0x060B }, ::unicode::General_Category::Currency_Symbol }, // AFGHANI SIGN
    Prop<::unicode::General_Category> { { 0x060C, 0x060D }, ::unicode::General_Category::Other_Punctuation }, // [2] ARABIC COMMA..ARABIC DATE SEPARATOR
    Prop<::unicode::General_Category> { { 0x060E, 0x060F }, ::unicode::General_Category::Other_Symbol }, // [2] ARABIC POETIC VERSE SIGN..ARABIC SIGN MISRA
    Prop<::unicode::General_Category> { { 0x0610, 0x061A }, ::unicode::General_Category::Nonspacing_Mark }, // [11] ARABIC SIGN SALLALLAHOU ALAYHE WASSALLAM..ARABIC SMALL KASRA
    Prop<::unicode::General_Category> { { 0x061B, 0x061B }, ::unicode::General_Category::Other_Punctuation }, // ARABIC SEMICOLON
    Prop<::unicode::General_Category> { { 0x061C, 0x061C }, ::unicode::General_Category::Format }, // ARABIC LETTER MARK
    Prop<::unicode::General_Category> { { 0x061D, 0x061F }, ::unicode::General_Category::Other_Punctuation }, // [3] ARABIC END OF TEXT MARK..ARABIC QUESTION MARK
    Prop<::unicode::General_Category> { { 0x0620, 0x063F }, ::unicode::General_Category::Other_Letter }, // [32] ARABIC LETTER KASHMIRI YEH..ARABIC LETTER FARSI YEH WITH THREE DOTS ABOVE
    Prop<::unicode::General_Category> { { 0x0640, 0x0640 }, ::unicode::General_Category::Modifier_Letter }, // ARABIC TATWEEL
    Prop<::unicode::General_Category> { { 0x0641, 0x064A }, ::unicode::General_Category::Other_Letter }, // [10] ARABIC LETTER FEH..ARABIC LETTER YEH
    Prop<::unicode::General_Category> { { 0x064B, 0x065F }, ::unicode::General_Category::Nonspacing_Mark }, // [21] ARABIC FATHATAN..ARABIC WAVY HAMZA BELOW
    Prop<::unicode::General_Category> { { 0x0660, 0x0669 }, ::unicode::General_Category::Decimal_Number }, // [10] ARABIC-INDIC DIGIT ZERO..ARABIC-INDIC DIGIT NINE
    Prop<::unicode::General_Category> { { 0x066A, 0x066D }, ::unicode::General_Category::Other_Punctuation }, // [4] ARABIC PERCENT SIGN..ARABIC FIVE POINTED STAR
    Prop<::unicode::General_Category> { { 0x066E, 0x066F }, ::unicode::General_Category::Other_Letter }, // [2] ARABIC LETTER DOTLESS BEH..ARABIC LETTER DOTLESS QAF
    Prop<::unicode::General_Category> { { 0x0670, 0x0670 }, ::unicode::General_Category::Nonspacing_Mark }, // ARABIC LETTER SUPERSCRIPT ALEF
    Prop<::unicode::General_Category> { { 0x0671, 0x06D3 }, ::unicode::General_Category::Other_Letter }, // [99] ARABIC LETTER ALEF WASLA..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE
    Prop<::unicode::General_Category> { { 0x06D4, 0x06D4 }, ::unicode::General_Category::Other_Punctuation }, // ARABIC FULL STOP
    Prop<::unicode::General_Category> { { 0x06D5, 0x06D5 }, ::unicode::General_Category::Other_Letter }, // ARABIC LETTER AE
    Prop<::unicode::General_Category> { { 0x06D6, 0x06DC }, ::unicode::General_Category::Nonspacing_Mark }, // [7] ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA..ARABIC SMALL HIGH SEEN
    Prop<::unicode::General_Category> { { 0x06DD, 0x06DD }, ::unicode::General_Category::Format }, // ARABIC END OF AYAH
    Prop<::unicode::General_Category> { { 0x06DE, 0x06DE }, ::unicode::General_Category::Other_Symbol }, // ARABIC START OF RUB EL HIZB
    Prop<::unicode::General_Category> { { 0x06DF, 0x06E4 }, ::unicode::General_Category::Nonspacing_Mark }, // [6] ARABIC SMALL HIGH ROUNDED ZERO..ARABIC SMALL HIGH MADDA
    Prop<::unicode::General_Category> { { 0x06E5, 0x06E6 }, ::unicode::General_Category::Modifier_Letter }, // [2] ARABIC SMALL WAW..ARABIC SMALL YEH
    Prop<::unicode::General_Category> { { 0x06E7, 0x06E8 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] ARABIC SMALL HIGH YEH..ARABIC SMALL HIGH NOON
    Prop<::unicode::General_Category> { { 0x06E9, 0x06E9 }, ::unicode::General_Category::Other_Symbol }, // ARABIC PLACE OF SAJDAH
    Prop<::unicode::General_Category> { { 0x06EA, 0x06ED }, ::unicode::General_Category::Nonspacing_Mark }, // [4] ARABIC EMPTY CENTRE LOW STOP..ARABIC SMALL LOW MEEM
    Prop<::unicode::General_Category> { { 0x06EE, 0x06EF }, ::unicode::General_Category::Other_Letter }, // [2] ARABIC LETTER DAL WITH INVERTED V..ARABIC LETTER REH WITH INVERTED V
    Prop<::unicode::General_Category> { { 0x06F0, 0x06F9 }, ::unicode::General_Category::Decimal_Number }, // [10] EXTENDED ARABIC-INDIC DIGIT ZERO..EXTENDED ARABIC-INDIC DIGIT NINE
    Prop<::unicode::General_Category> { { 0x06FA, 0x06FC }, ::unicode::General_Category::Other_Letter }, // [3] ARABIC LETTER SHEEN WITH DOT BELOW..ARABIC LETTER GHAIN WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x06FD, 0x06FE }, ::unicode::General_Category::Other_Symbol }, // [2] ARABIC SIGN SINDHI AMPERSAND..ARABIC SIGN SINDHI POSTPOSITION MEN
    Prop<::unicode::General_Category> { { 0x06FF, 0x06FF }, ::unicode::General_Category::Other_Letter }, // ARABIC LETTER HEH WITH INVERTED V
    Prop<::unicode::General_Category> { { 0x0700, 0x070D }, ::unicode::General_Category::Other_Punctuation }, // [14] SYRIAC END OF PARAGRAPH..SYRIAC HARKLEAN ASTERISCUS
    Prop<::unicode::General_Category> { { 0x070E, 0x070E }, ::unicode::General_Category::Unassigned }, // <reserved-070E>
    Prop<::unicode::General_Category> { { 0x070F, 0x070F }, ::unicode::General_Category::Format }, // SYRIAC ABBREVIATION MARK
    Prop<::unicode::General_Category> { { 0x0710, 0x0710 }, ::unicode::General_Category::Other_Letter }, // SYRIAC LETTER ALAPH
    Prop<::unicode::General_Category> { { 0x0711, 0x0711 }, ::unicode::General_Category::Nonspacing_Mark }, // SYRIAC LETTER SUPERSCRIPT ALAPH
    Prop<::unicode::General_Category> { { 0x0712, 0x072F }, ::unicode::General_Category::Other_Letter }, // [30] SYRIAC LETTER BETH..SYRIAC LETTER PERSIAN DHALATH
    Prop<::unicode::General_Category> { { 0x0730, 0x074A }, ::unicode::General_Category::Nonspacing_Mark }, // [27] SYRIAC PTHAHA ABOVE..SYRIAC BARREKH
    Prop<::unicode::General_Category> { { 0x074B, 0x074C }, ::unicode::General_Category::Unassigned }, // [2] <reserved-074B>..<reserved-074C>
    Prop<::unicode::General_Category> { { 0x074D, 0x07A5 }, ::unicode::General_Category::Other_Letter }, // [89] SYRIAC LETTER SOGDIAN ZHAIN..THAANA LETTER WAAVU
    Prop<::unicode::General_Category> { { 0x07A6, 0x07B0 }, ::unicode::General_Category::Nonspacing_Mark }, // [11] THAANA ABAFILI..THAANA SUKUN
    Prop<::unicode::General_Category> { { 0x07B1, 0x07B1 }, ::unicode::General_Category::Other_Letter }, // THAANA LETTER NAA
    Prop<::unicode::General_Category> { { 0x07B2, 0x07BF }, ::unicode::General_Category::Unassigned }, // [14] <reserved-07B2>..<reserved-07BF>
    Prop<::unicode::General_Category> { { 0x07C0, 0x07C9 }, ::unicode::General_Category::Decimal_Number }, // [10] NKO DIGIT ZERO..NKO DIGIT NINE
    Prop<::unicode::General_Category> { { 0x07CA, 0x07EA }, ::unicode::General_Category::Other_Letter }, // [33] NKO LETTER A..NKO LETTER JONA RA
    Prop<::unicode::General_Category> { { 0x07EB, 0x07F3 }, ::unicode::General_Category::Nonspacing_Mark }, // [9] NKO COMBINING SHORT HIGH TONE..NKO COMBINING DOUBLE DOT ABOVE
    Prop<::unicode::General_Category> { { 0x07F4, 0x07F5 }, ::unicode::General_Category::Modifier_Letter }, // [2] NKO HIGH TONE APOSTROPHE..NKO LOW TONE APOSTROPHE
    Prop<::unicode::General_Category> { { 0x07F6, 0x07F6 }, ::unicode::General_Category::Other_Symbol }, // NKO SYMBOL OO DENNEN
    Prop<::unicode::General_Category> { { 0x07F7, 0x07F9 }, ::unicode::General_Category::Other_Punctuation }, // [3] NKO SYMBOL GBAKURUNEN..NKO EXCLAMATION MARK
    Prop<::unicode::General_Category> { { 0x07FA, 0x07FA }, ::unicode::General_Category::Modifier_Letter }, // NKO LAJANYALAN
    Prop<::unicode::General_Category> { { 0x07FB, 0x07FC }, ::unicode::General_Category::Unassigned }, // [2] <reserved-07FB>..<reserved-07FC>
    Prop<::unicode::General_Category> { { 0x07FD, 0x07FD }, ::unicode::General_Category::Nonspacing_Mark }, // NKO DANTAYALAN
    Prop<::unicode::General_Category> { { 0x07FE, 0x07FF }, ::unicode::General_Category::Currency_Symbol }, // [2] NKO DOROME SIGN..NKO TAMAN SIGN
    Prop<::unicode::General_Category> { { 0x0800, 0x0815 }, ::unicode::General_Category::Other_Letter }, // [22] SAMARITAN LETTER ALAF..SAMARITAN LETTER TAAF
    Prop<::unicode::General_Category> { { 0x0816, 0x0819 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] SAMARITAN MARK IN..SAMARITAN MARK DAGESH
    Prop<::unicode::General_Category> { { 0x081A, 0x081A }, ::unicode::General_Category::Modifier_Letter }, // SAMARITAN MODIFIER LETTER EPENTHETIC YUT
    Prop<::unicode::General_Category> { { 0x081B, 0x0823 }, ::unicode::General_Category::Nonspacing_Mark }, // [9] SAMARITAN MARK EPENTHETIC YUT..SAMARITAN VOWEL SIGN A
    Prop<::unicode::General_Category> { { 0x0824, 0x0824 }, ::unicode::General_Category::Modifier_Letter }, // SAMARITAN MODIFIER LETTER SHORT A
    Prop<::unicode::General_Category> { { 0x0825, 0x0827 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] SAMARITAN VOWEL SIGN SHORT A..SAMARITAN VOWEL SIGN U
    Prop<::unicode::General_Category> { { 0x0828, 0x0828 }, ::unicode::General_Category::Modifier_Letter }, // SAMARITAN MODIFIER LETTER I
    Prop<::unicode::General_Category> { { 0x0829, 0x082D }, ::unicode::General_Category::Nonspacing_Mark }, // [5] SAMARITAN VOWEL SIGN LONG I..SAMARITAN MARK NEQUDAA
    Prop<::unicode::General_Category> { { 0x082E, 0x082F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-082E>..<reserved-082F>
    Prop<::unicode::General_Category> { { 0x0830, 0x083E }, ::unicode::General_Category::Other_Punctuation }, // [15] SAMARITAN PUNCTUATION NEQUDAA..SAMARITAN PUNCTUATION ANNAAU
    Prop<::unicode::General_Category> { { 0x083F, 0x083F }, ::unicode::General_Category::Unassigned }, // <reserved-083F>
    Prop<::unicode::General_Category> { { 0x0840, 0x0858 }, ::unicode::General_Category::Other_Letter }, // [25] MANDAIC LETTER HALQA..MANDAIC LETTER AIN
    Prop<::unicode::General_Category> { { 0x0859, 0x085B }, ::unicode::General_Category::Nonspacing_Mark }, // [3] MANDAIC AFFRICATION MARK..MANDAIC GEMINATION MARK
    Prop<::unicode::General_Category> { { 0x085C, 0x085D }, ::unicode::General_Category::Unassigned }, // [2] <reserved-085C>..<reserved-085D>
    Prop<::unicode::General_Category> { { 0x085E, 0x085E }, ::unicode::General_Category::Other_Punctuation }, // MANDAIC PUNCTUATION
    Prop<::unicode::General_Category> { { 0x085F, 0x085F }, ::unicode::General_Category::Unassigned }, // <reserved-085F>
    Prop<::unicode::General_Category> { { 0x0860, 0x086A }, ::unicode::General_Category::Other_Letter }, // [11] SYRIAC LETTER MALAYALAM NGA..SYRIAC LETTER MALAYALAM SSA
    Prop<::unicode::General_Category> { { 0x086B, 0x086F }, ::unicode::General_Category::Unassigned }, // [5] <reserved-086B>..<reserved-086F>
    Prop<::unicode::General_Category> { { 0x0870, 0x0887 }, ::unicode::General_Category::Other_Letter }, // [24] ARABIC LETTER ALEF WITH ATTACHED FATHA..ARABIC BASELINE ROUND DOT
    Prop<::unicode::General_Category> { { 0x0888, 0x0888 }, ::unicode::General_Category::Modifier_Symbol }, // ARABIC RAISED ROUND DOT
    Prop<::unicode::General_Category> { { 0x0889, 0x088F }, ::unicode::General_Category::Other_Letter }, // [7] ARABIC LETTER NOON WITH INVERTED SMALL V..ARABIC LETTER NOON WITH RING ABOVE
    Prop<::unicode::General_Category> { { 0x0890, 0x0891 }, ::unicode::General_Category::Format }, // [2] ARABIC POUND MARK ABOVE..ARABIC PIASTRE MARK ABOVE
    Prop<::unicode::General_Category> { { 0x0892, 0x0896 }, ::unicode::General_Category::Unassigned }, // [5] <reserved-0892>..<reserved-0896>
    Prop<::unicode::General_Category> { { 0x0897, 0x089F }, ::unicode::General_Category::Nonspacing_Mark }, // [9] ARABIC PEPET..ARABIC HALF MADDA OVER MADDA
    Prop<::unicode::General_Category> { { 0x08A0, 0x08C8 }, ::unicode::General_Category::Other_Letter }, // [41] ARABIC LETTER BEH WITH SMALL V BELOW..ARABIC LETTER GRAF
    Prop<::unicode::General_Category> { { 0x08C9, 0x08C9 }, ::unicode::General_Category::Modifier_Letter }, // ARABIC SMALL FARSI YEH
    Prop<::unicode::General_Category> { { 0x08CA, 0x08E1 }, ::unicode::General_Category::Nonspacing_Mark }, // [24] ARABIC SMALL HIGH FARSI YEH..ARABIC SMALL HIGH SIGN SAFHA
    Prop<::unicode::General_Category> { { 0x08E2, 0x08E2 }, ::unicode::General_Category::Format }, // ARABIC DISPUTED END OF AYAH
    Prop<::unicode::General_Category> { { 0x08E3, 0x0902 }, ::unicode::General_Category::Nonspacing_Mark }, // [32] ARABIC TURNED DAMMA BELOW..DEVANAGARI SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x0903, 0x0903 }, ::unicode::General_Category::Spacing_Mark }, // DEVANAGARI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x0904, 0x0939 }, ::unicode::General_Category::Other_Letter }, // [54] DEVANAGARI LETTER SHORT A..DEVANAGARI LETTER HA
    Prop<::unicode::General_Category> { { 0x093A, 0x093A }, ::unicode::General_Category::Nonspacing_Mark }, // DEVANAGARI VOWEL SIGN OE
    Prop<::unicode::General_Category> { { 0x093B, 0x093B }, ::unicode::General_Category::Spacing_Mark }, // DEVANAGARI VOWEL SIGN OOE
    Prop<::unicode::General_Category> { { 0x093C, 0x093C }, ::unicode::General_Category::Nonspacing_Mark }, // DEVANAGARI SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x093D, 0x093D }, ::unicode::General_Category::Other_Letter }, // DEVANAGARI SIGN AVAGRAHA
    Prop<::unicode::General_Category> { { 0x093E, 0x0940 }, ::unicode::General_Category::Spacing_Mark }, // [3] DEVANAGARI VOWEL SIGN AA..DEVANAGARI VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x0941, 0x0948 }, ::unicode::General_Category::Nonspacing_Mark }, // [8] DEVANAGARI VOWEL SIGN U..DEVANAGARI VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x0949, 0x094C }, ::unicode::General_Category::Spacing_Mark }, // [4] DEVANAGARI VOWEL SIGN CANDRA O..DEVANAGARI VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x094D, 0x094D }, ::unicode::General_Category::Nonspacing_Mark }, // DEVANAGARI SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x094E, 0x094F }, ::unicode::General_Category::Spacing_Mark }, // [2] DEVANAGARI VOWEL SIGN PRISHTHAMATRA E..DEVANAGARI VOWEL SIGN AW
    Prop<::unicode::General_Category> { { 0x0950, 0x0950 }, ::unicode::General_Category::Other_Letter }, // DEVANAGARI OM
    Prop<::unicode::General_Category> { { 0x0951, 0x0957 }, ::unicode::General_Category::Nonspacing_Mark }, // [7] DEVANAGARI STRESS SIGN UDATTA..DEVANAGARI VOWEL SIGN UUE
    Prop<::unicode::General_Category> { { 0x0958, 0x0961 }, ::unicode::General_Category::Other_Letter }, // [10] DEVANAGARI LETTER QA..DEVANAGARI LETTER VOCALIC LL
    Prop<::unicode::General_Category> { { 0x0962, 0x0963 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] DEVANAGARI VOWEL SIGN VOCALIC L..DEVANAGARI VOWEL SIGN VOCALIC LL
    Prop<::unicode::General_Category> { { 0x0964, 0x0965 }, ::unicode::General_Category::Other_Punctuation }, // [2] DEVANAGARI DANDA..DEVANAGARI DOUBLE DANDA
    Prop<::unicode::General_Category> { { 0x0966, 0x096F }, ::unicode::General_Category::Decimal_Number }, // [10] DEVANAGARI DIGIT ZERO..DEVANAGARI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x0970, 0x0970 }, ::unicode::General_Category::Other_Punctuation }, // DEVANAGARI ABBREVIATION SIGN
    Prop<::unicode::General_Category> { { 0x0971, 0x0971 }, ::unicode::General_Category::Modifier_Letter }, // DEVANAGARI SIGN HIGH SPACING DOT
    Prop<::unicode::General_Category> { { 0x0972, 0x0980 }, ::unicode::General_Category::Other_Letter }, // [15] DEVANAGARI LETTER CANDRA A..BENGALI ANJI
    Prop<::unicode::General_Category> { { 0x0981, 0x0981 }, ::unicode::General_Category::Nonspacing_Mark }, // BENGALI SIGN CANDRABINDU
    Prop<::unicode::General_Category> { { 0x0982, 0x0983 }, ::unicode::General_Category::Spacing_Mark }, // [2] BENGALI SIGN ANUSVARA..BENGALI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x0984, 0x0984 }, ::unicode::General_Category::Unassigned }, // <reserved-0984>
    Prop<::unicode::General_Category> { { 0x0985, 0x098C }, ::unicode::General_Category::Other_Letter }, // [8] BENGALI LETTER A..BENGALI LETTER VOCALIC L
    Prop<::unicode::General_Category> { { 0x098D, 0x098E }, ::unicode::General_Category::Unassigned }, // [2] <reserved-098D>..<reserved-098E>
    Prop<::unicode::General_Category> { { 0x098F, 0x0990 }, ::unicode::General_Category::Other_Letter }, // [2] BENGALI LETTER E..BENGALI LETTER AI
    Prop<::unicode::General_Category> { { 0x0991, 0x0992 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0991>..<reserved-0992>
    Prop<::unicode::General_Category> { { 0x0993, 0x09A8 }, ::unicode::General_Category::Other_Letter }, // [22] BENGALI LETTER O..BENGALI LETTER NA
    Prop<::unicode::General_Category> { { 0x09A9, 0x09A9 }, ::unicode::General_Category::Unassigned }, // <reserved-09A9>
    Prop<::unicode::General_Category> { { 0x09AA, 0x09B0 }, ::unicode::General_Category::Other_Letter }, // [7] BENGALI LETTER PA..BENGALI LETTER RA
    Prop<::unicode::General_Category> { { 0x09B1, 0x09B1 }, ::unicode::General_Category::Unassigned }, // <reserved-09B1>
    Prop<::unicode::General_Category> { { 0x09B2, 0x09B2 }, ::unicode::General_Category::Other_Letter }, // BENGALI LETTER LA
    Prop<::unicode::General_Category> { { 0x09B3, 0x09B5 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-09B3>..<reserved-09B5>
    Prop<::unicode::General_Category> { { 0x09B6, 0x09B9 }, ::unicode::General_Category::Other_Letter }, // [4] BENGALI LETTER SHA..BENGALI LETTER HA
    Prop<::unicode::General_Category> { { 0x09BA, 0x09BB }, ::unicode::General_Category::Unassigned }, // [2] <reserved-09BA>..<reserved-09BB>
    Prop<::unicode::General_Category> { { 0x09BC, 0x09BC }, ::unicode::General_Category::Nonspacing_Mark }, // BENGALI SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x09BD, 0x09BD }, ::unicode::General_Category::Other_Letter }, // BENGALI SIGN AVAGRAHA
    Prop<::unicode::General_Category> { { 0x09BE, 0x09C0 }, ::unicode::General_Category::Spacing_Mark }, // [3] BENGALI VOWEL SIGN AA..BENGALI VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x09C1, 0x09C4 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] BENGALI VOWEL SIGN U..BENGALI VOWEL SIGN VOCALIC RR
    Prop<::unicode::General_Category> { { 0x09C5, 0x09C6 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-09C5>..<reserved-09C6>
    Prop<::unicode::General_Category> { { 0x09C7, 0x09C8 }, ::unicode::General_Category::Spacing_Mark }, // [2] BENGALI VOWEL SIGN E..BENGALI VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x09C9, 0x09CA }, ::unicode::General_Category::Unassigned }, // [2] <reserved-09C9>..<reserved-09CA>
    Prop<::unicode::General_Category> { { 0x09CB, 0x09CC }, ::unicode::General_Category::Spacing_Mark }, // [2] BENGALI VOWEL SIGN O..BENGALI VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x09CD, 0x09CD }, ::unicode::General_Category::Nonspacing_Mark }, // BENGALI SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x09CE, 0x09CE }, ::unicode::General_Category::Other_Letter }, // BENGALI LETTER KHANDA TA
    Prop<::unicode::General_Category> { { 0x09CF, 0x09D6 }, ::unicode::General_Category::Unassigned }, // [8] <reserved-09CF>..<reserved-09D6>
    Prop<::unicode::General_Category> { { 0x09D7, 0x09D7 }, ::unicode::General_Category::Spacing_Mark }, // BENGALI AU LENGTH MARK
    Prop<::unicode::General_Category> { { 0x09D8, 0x09DB }, ::unicode::General_Category::Unassigned }, // [4] <reserved-09D8>..<reserved-09DB>
    Prop<::unicode::General_Category> { { 0x09DC, 0x09DD }, ::unicode::General_Category::Other_Letter }, // [2] BENGALI LETTER RRA..BENGALI LETTER RHA
    Prop<::unicode::General_Category> { { 0x09DE, 0x09DE }, ::unicode::General_Category::Unassigned }, // <reserved-09DE>
    Prop<::unicode::General_Category> { { 0x09DF, 0x09E1 }, ::unicode::General_Category::Other_Letter }, // [3] BENGALI LETTER YYA..BENGALI LETTER VOCALIC LL
    Prop<::unicode::General_Category> { { 0x09E2, 0x09E3 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] BENGALI VOWEL SIGN VOCALIC L..BENGALI VOWEL SIGN VOCALIC LL
    Prop<::unicode::General_Category> { { 0x09E4, 0x09E5 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-09E4>..<reserved-09E5>
    Prop<::unicode::General_Category> { { 0x09E6, 0x09EF }, ::unicode::General_Category::Decimal_Number }, // [10] BENGALI DIGIT ZERO..BENGALI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x09F0, 0x09F1 }, ::unicode::General_Category::Other_Letter }, // [2] BENGALI LETTER RA WITH MIDDLE DIAGONAL..BENGALI LETTER RA WITH LOWER DIAGONAL
    Prop<::unicode::General_Category> { { 0x09F2, 0x09F3 }, ::unicode::General_Category::Currency_Symbol }, // [2] BENGALI RUPEE MARK..BENGALI RUPEE SIGN
    Prop<::unicode::General_Category> { { 0x09F4, 0x09F9 }, ::unicode::General_Category::Other_Number }, // [6] BENGALI CURRENCY NUMERATOR ONE..BENGALI CURRENCY DENOMINATOR SIXTEEN
    Prop<::unicode::General_Category> { { 0x09FA, 0x09FA }, ::unicode::General_Category::Other_Symbol }, // BENGALI ISSHAR
    Prop<::unicode::General_Category> { { 0x09FB, 0x09FB }, ::unicode::General_Category::Currency_Symbol }, // BENGALI GANDA MARK
    Prop<::unicode::General_Category> { { 0x09FC, 0x09FC }, ::unicode::General_Category::Other_Letter }, // BENGALI LETTER VEDIC ANUSVARA
    Prop<::unicode::General_Category> { { 0x09FD, 0x09FD }, ::unicode::General_Category::Other_Punctuation }, // BENGALI ABBREVIATION SIGN
    Prop<::unicode::General_Category> { { 0x09FE, 0x09FE }, ::unicode::General_Category::Nonspacing_Mark }, // BENGALI SANDHI MARK
    Prop<::unicode::General_Category> { { 0x09FF, 0x0A00 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-09FF>..<reserved-0A00>
    Prop<::unicode::General_Category> { { 0x0A01, 0x0A02 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] GURMUKHI SIGN ADAK BINDI..GURMUKHI SIGN BINDI
    Prop<::unicode::General_Category> { { 0x0A03, 0x0A03 }, ::unicode::General_Category::Spacing_Mark }, // GURMUKHI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x0A04, 0x0A04 }, ::unicode::General_Category::Unassigned }, // <reserved-0A04>
    Prop<::unicode::General_Category> { { 0x0A05, 0x0A0A }, ::unicode::General_Category::Other_Letter }, // [6] GURMUKHI LETTER A..GURMUKHI LETTER UU
    Prop<::unicode::General_Category> { { 0x0A0B, 0x0A0E }, ::unicode::General_Category::Unassigned }, // [4] <reserved-0A0B>..<reserved-0A0E>
    Prop<::unicode::General_Category> { { 0x0A0F, 0x0A10 }, ::unicode::General_Category::Other_Letter }, // [2] GURMUKHI LETTER EE..GURMUKHI LETTER AI
    Prop<::unicode::General_Category> { { 0x0A11, 0x0A12 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0A11>..<reserved-0A12>
    Prop<::unicode::General_Category> { { 0x0A13, 0x0A28 }, ::unicode::General_Category::Other_Letter }, // [22] GURMUKHI LETTER OO..GURMUKHI LETTER NA
    Prop<::unicode::General_Category> { { 0x0A29, 0x0A29 }, ::unicode::General_Category::Unassigned }, // <reserved-0A29>
    Prop<::unicode::General_Category> { { 0x0A2A, 0x0A30 }, ::unicode::General_Category::Other_Letter }, // [7] GURMUKHI LETTER PA..GURMUKHI LETTER RA
    Prop<::unicode::General_Category> { { 0x0A31, 0x0A31 }, ::unicode::General_Category::Unassigned }, // <reserved-0A31>
    Prop<::unicode::General_Category> { { 0x0A32, 0x0A33 }, ::unicode::General_Category::Other_Letter }, // [2] GURMUKHI LETTER LA..GURMUKHI LETTER LLA
    Prop<::unicode::General_Category> { { 0x0A34, 0x0A34 }, ::unicode::General_Category::Unassigned }, // <reserved-0A34>
    Prop<::unicode::General_Category> { { 0x0A35, 0x0A36 }, ::unicode::General_Category::Other_Letter }, // [2] GURMUKHI LETTER VA..GURMUKHI LETTER SHA
    Prop<::unicode::General_Category> { { 0x0A37, 0x0A37 }, ::unicode::General_Category::Unassigned }, // <reserved-0A37>
    Prop<::unicode::General_Category> { { 0x0A38, 0x0A39 }, ::unicode::General_Category::Other_Letter }, // [2] GURMUKHI LETTER SA..GURMUKHI LETTER HA
    Prop<::unicode::General_Category> { { 0x0A3A, 0x0A3B }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0A3A>..<reserved-0A3B>
    Prop<::unicode::General_Category> { { 0x0A3C, 0x0A3C }, ::unicode::General_Category::Nonspacing_Mark }, // GURMUKHI SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x0A3D, 0x0A3D }, ::unicode::General_Category::Unassigned }, // <reserved-0A3D>
    Prop<::unicode::General_Category> { { 0x0A3E, 0x0A40 }, ::unicode::General_Category::Spacing_Mark }, // [3] GURMUKHI VOWEL SIGN AA..GURMUKHI VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x0A41, 0x0A42 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] GURMUKHI VOWEL SIGN U..GURMUKHI VOWEL SIGN UU
    Prop<::unicode::General_Category> { { 0x0A43, 0x0A46 }, ::unicode::General_Category::Unassigned }, // [4] <reserved-0A43>..<reserved-0A46>
    Prop<::unicode::General_Category> { { 0x0A47, 0x0A48 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] GURMUKHI VOWEL SIGN EE..GURMUKHI VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x0A49, 0x0A4A }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0A49>..<reserved-0A4A>
    Prop<::unicode::General_Category> { { 0x0A4B, 0x0A4D }, ::unicode::General_Category::Nonspacing_Mark }, // [3] GURMUKHI VOWEL SIGN OO..GURMUKHI SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x0A4E, 0x0A50 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-0A4E>..<reserved-0A50>
    Prop<::unicode::General_Category> { { 0x0A51, 0x0A51 }, ::unicode::General_Category::Nonspacing_Mark }, // GURMUKHI SIGN UDAAT
    Prop<::unicode::General_Category> { { 0x0A52, 0x0A58 }, ::unicode::General_Category::Unassigned }, // [7] <reserved-0A52>..<reserved-0A58>
    Prop<::unicode::General_Category> { { 0x0A59, 0x0A5C }, ::unicode::General_Category::Other_Letter }, // [4] GURMUKHI LETTER KHHA..GURMUKHI LETTER RRA
    Prop<::unicode::General_Category> { { 0x0A5D, 0x0A5D }, ::unicode::General_Category::Unassigned }, // <reserved-0A5D>
    Prop<::unicode::General_Category> { { 0x0A5E, 0x0A5E }, ::unicode::General_Category::Other_Letter }, // GURMUKHI LETTER FA
    Prop<::unicode::General_Category> { { 0x0A5F, 0x0A65 }, ::unicode::General_Category::Unassigned }, // [7] <reserved-0A5F>..<reserved-0A65>
    Prop<::unicode::General_Category> { { 0x0A66, 0x0A6F }, ::unicode::General_Category::Decimal_Number }, // [10] GURMUKHI DIGIT ZERO..GURMUKHI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x0A70, 0x0A71 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] GURMUKHI TIPPI..GURMUKHI ADDAK
    Prop<::unicode::General_Category> { { 0x0A72, 0x0A74 }, ::unicode::General_Category::Other_Letter }, // [3] GURMUKHI IRI..GURMUKHI EK ONKAR
    Prop<::unicode::General_Category> { { 0x0A75, 0x0A75 }, ::unicode::General_Category::Nonspacing_Mark }, // GURMUKHI SIGN YAKASH
    Prop<::unicode::General_Category> { { 0x0A76, 0x0A76 }, ::unicode::General_Category::Other_Punctuation }, // GURMUKHI ABBREVIATION SIGN
    Prop<::unicode::General_Category> { { 0x0A77, 0x0A80 }, ::unicode::General_Category::Unassigned }, // [10] <reserved-0A77>..<reserved-0A80>
    Prop<::unicode::General_Category> { { 0x0A81, 0x0A82 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] GUJARATI SIGN CANDRABINDU..GUJARATI SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x0A83, 0x0A83 }, ::unicode::General_Category::Spacing_Mark }, // GUJARATI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x0A84, 0x0A84 }, ::unicode::General_Category::Unassigned }, // <reserved-0A84>
    Prop<::unicode::General_Category> { { 0x0A85, 0x0A8D }, ::unicode::General_Category::Other_Letter }, // [9] GUJARATI LETTER A..GUJARATI VOWEL CANDRA E
    Prop<::unicode::General_Category> { { 0x0A8E, 0x0A8E }, ::unicode::General_Category::Unassigned }, // <reserved-0A8E>
    Prop<::unicode::General_Category> { { 0x0A8F, 0x0A91 }, ::unicode::General_Category::Other_Letter }, // [3] GUJARATI LETTER E..GUJARATI VOWEL CANDRA O
    Prop<::unicode::General_Category> { { 0x0A92, 0x0A92 }, ::unicode::General_Category::Unassigned }, // <reserved-0A92>
    Prop<::unicode::General_Category> { { 0x0A93, 0x0AA8 }, ::unicode::General_Category::Other_Letter }, // [22] GUJARATI LETTER O..GUJARATI LETTER NA
    Prop<::unicode::General_Category> { { 0x0AA9, 0x0AA9 }, ::unicode::General_Category::Unassigned }, // <reserved-0AA9>
    Prop<::unicode::General_Category> { { 0x0AAA, 0x0AB0 }, ::unicode::General_Category::Other_Letter }, // [7] GUJARATI LETTER PA..GUJARATI LETTER RA
    Prop<::unicode::General_Category> { { 0x0AB1, 0x0AB1 }, ::unicode::General_Category::Unassigned }, // <reserved-0AB1>
    Prop<::unicode::General_Category> { { 0x0AB2, 0x0AB3 }, ::unicode::General_Category::Other_Letter }, // [2] GUJARATI LETTER LA..GUJARATI LETTER LLA
    Prop<::unicode::General_Category> { { 0x0AB4, 0x0AB4 }, ::unicode::General_Category::Unassigned }, // <reserved-0AB4>
    Prop<::unicode::General_Category> { { 0x0AB5, 0x0AB9 }, ::unicode::General_Category::Other_Letter }, // [5] GUJARATI LETTER VA..GUJARATI LETTER HA
    Prop<::unicode::General_Category> { { 0x0ABA, 0x0ABB }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0ABA>..<reserved-0ABB>
    Prop<::unicode::General_Category> { { 0x0ABC, 0x0ABC }, ::unicode::General_Category::Nonspacing_Mark }, // GUJARATI SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x0ABD, 0x0ABD }, ::unicode::General_Category::Other_Letter }, // GUJARATI SIGN AVAGRAHA
    Prop<::unicode::General_Category> { { 0x0ABE, 0x0AC0 }, ::unicode::General_Category::Spacing_Mark }, // [3] GUJARATI VOWEL SIGN AA..GUJARATI VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x0AC1, 0x0AC5 }, ::unicode::General_Category::Nonspacing_Mark }, // [5] GUJARATI VOWEL SIGN U..GUJARATI VOWEL SIGN CANDRA E
    Prop<::unicode::General_Category> { { 0x0AC6, 0x0AC6 }, ::unicode::General_Category::Unassigned }, // <reserved-0AC6>
    Prop<::unicode::General_Category> { { 0x0AC7, 0x0AC8 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] GUJARATI VOWEL SIGN E..GUJARATI VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x0AC9, 0x0AC9 }, ::unicode::General_Category::Spacing_Mark }, // GUJARATI VOWEL SIGN CANDRA O
    Prop<::unicode::General_Category> { { 0x0ACA, 0x0ACA }, ::unicode::General_Category::Unassigned }, // <reserved-0ACA>
    Prop<::unicode::General_Category> { { 0x0ACB, 0x0ACC }, ::unicode::General_Category::Spacing_Mark }, // [2] GUJARATI VOWEL SIGN O..GUJARATI VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x0ACD, 0x0ACD }, ::unicode::General_Category::Nonspacing_Mark }, // GUJARATI SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x0ACE, 0x0ACF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0ACE>..<reserved-0ACF>
    Prop<::unicode::General_Category> { { 0x0AD0, 0x0AD0 }, ::unicode::General_Category::Other_Letter }, // GUJARATI OM
    Prop<::unicode::General_Category> { { 0x0AD1, 0x0ADF }, ::unicode::General_Category::Unassigned }, // [15] <reserved-0AD1>..<reserved-0ADF>
    Prop<::unicode::General_Category> { { 0x0AE0, 0x0AE1 }, ::unicode::General_Category::Other_Letter }, // [2] GUJARATI LETTER VOCALIC RR..GUJARATI LETTER VOCALIC LL
    Prop<::unicode::General_Category> { { 0x0AE2, 0x0AE3 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] GUJARATI VOWEL SIGN VOCALIC L..GUJARATI VOWEL SIGN VOCALIC LL
    Prop<::unicode::General_Category> { { 0x0AE4, 0x0AE5 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0AE4>..<reserved-0AE5>
    Prop<::unicode::General_Category> { { 0x0AE6, 0x0AEF }, ::unicode::General_Category::Decimal_Number }, // [10] GUJARATI DIGIT ZERO..GUJARATI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x0AF0, 0x0AF0 }, ::unicode::General_Category::Other_Punctuation }, // GUJARATI ABBREVIATION SIGN
    Prop<::unicode::General_Category> { { 0x0AF1, 0x0AF1 }, ::unicode::General_Category::Currency_Symbol }, // GUJARATI RUPEE SIGN
    Prop<::unicode::General_Category> { { 0x0AF2, 0x0AF8 }, ::unicode::General_Category::Unassigned }, // [7] <reserved-0AF2>..<reserved-0AF8>
    Prop<::unicode::General_Category> { { 0x0AF9, 0x0AF9 }, ::unicode::General_Category::Other_Letter }, // GUJARATI LETTER ZHA
    Prop<::unicode::General_Category> { { 0x0AFA, 0x0AFF }, ::unicode::General_Category::Nonspacing_Mark }, // [6] GUJARATI SIGN SUKUN..GUJARATI SIGN TWO-CIRCLE NUKTA ABOVE
    Prop<::unicode::General_Category> { { 0x0B00, 0x0B00 }, ::unicode::General_Category::Unassigned }, // <reserved-0B00>
    Prop<::unicode::General_Category> { { 0x0B01, 0x0B01 }, ::unicode::General_Category::Nonspacing_Mark }, // ORIYA SIGN CANDRABINDU
    Prop<::unicode::General_Category> { { 0x0B02, 0x0B03 }, ::unicode::General_Category::Spacing_Mark }, // [2] ORIYA SIGN ANUSVARA..ORIYA SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x0B04, 0x0B04 }, ::unicode::General_Category::Unassigned }, // <reserved-0B04>
    Prop<::unicode::General_Category> { { 0x0B05, 0x0B0C }, ::unicode::General_Category::Other_Letter }, // [8] ORIYA LETTER A..ORIYA LETTER VOCALIC L
    Prop<::unicode::General_Category> { { 0x0B0D, 0x0B0E }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0B0D>..<reserved-0B0E>
    Prop<::unicode::General_Category> { { 0x0B0F, 0x0B10 }, ::unicode::General_Category::Other_Letter }, // [2] ORIYA LETTER E..ORIYA LETTER AI
    Prop<::unicode::General_Category> { { 0x0B11, 0x0B12 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0B11>..<reserved-0B12>
    Prop<::unicode::General_Category> { { 0x0B13, 0x0B28 }, ::unicode::General_Category::Other_Letter }, // [22] ORIYA LETTER O..ORIYA LETTER NA
    Prop<::unicode::General_Category> { { 0x0B29, 0x0B29 }, ::unicode::General_Category::Unassigned }, // <reserved-0B29>
    Prop<::unicode::General_Category> { { 0x0B2A, 0x0B30 }, ::unicode::General_Category::Other_Letter }, // [7] ORIYA LETTER PA..ORIYA LETTER RA
    Prop<::unicode::General_Category> { { 0x0B31, 0x0B31 }, ::unicode::General_Category::Unassigned }, // <reserved-0B31>
    Prop<::unicode::General_Category> { { 0x0B32, 0x0B33 }, ::unicode::General_Category::Other_Letter }, // [2] ORIYA LETTER LA..ORIYA LETTER LLA
    Prop<::unicode::General_Category> { { 0x0B34, 0x0B34 }, ::unicode::General_Category::Unassigned }, // <reserved-0B34>
    Prop<::unicode::General_Category> { { 0x0B35, 0x0B39 }, ::unicode::General_Category::Other_Letter }, // [5] ORIYA LETTER VA..ORIYA LETTER HA
    Prop<::unicode::General_Category> { { 0x0B3A, 0x0B3B }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0B3A>..<reserved-0B3B>
    Prop<::unicode::General_Category> { { 0x0B3C, 0x0B3C }, ::unicode::General_Category::Nonspacing_Mark }, // ORIYA SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x0B3D, 0x0B3D }, ::unicode::General_Category::Other_Letter }, // ORIYA SIGN AVAGRAHA
    Prop<::unicode::General_Category> { { 0x0B3E, 0x0B3E }, ::unicode::General_Category::Spacing_Mark }, // ORIYA VOWEL SIGN AA
    Prop<::unicode::General_Category> { { 0x0B3F, 0x0B3F }, ::unicode::General_Category::Nonspacing_Mark }, // ORIYA VOWEL SIGN I
    Prop<::unicode::General_Category> { { 0x0B40, 0x0B40 }, ::unicode::General_Category::Spacing_Mark }, // ORIYA VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x0B41, 0x0B44 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] ORIYA VOWEL SIGN U..ORIYA VOWEL SIGN VOCALIC RR
    Prop<::unicode::General_Category> { { 0x0B45, 0x0B46 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0B45>..<reserved-0B46>
    Prop<::unicode::General_Category> { { 0x0B47, 0x0B48 }, ::unicode::General_Category::Spacing_Mark }, // [2] ORIYA VOWEL SIGN E..ORIYA VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x0B49, 0x0B4A }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0B49>..<reserved-0B4A>
    Prop<::unicode::General_Category> { { 0x0B4B, 0x0B4C }, ::unicode::General_Category::Spacing_Mark }, // [2] ORIYA VOWEL SIGN O..ORIYA VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x0B4D, 0x0B4D }, ::unicode::General_Category::Nonspacing_Mark }, // ORIYA SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x0B4E, 0x0B54 }, ::unicode::General_Category::Unassigned }, // [7] <reserved-0B4E>..<reserved-0B54>
    Prop<::unicode::General_Category> { { 0x0B55, 0x0B56 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] ORIYA SIGN OVERLINE..ORIYA AI LENGTH MARK
    Prop<::unicode::General_Category> { { 0x0B57, 0x0B57 }, ::unicode::General_Category::Spacing_Mark }, // ORIYA AU LENGTH MARK
    Prop<::unicode::General_Category> { { 0x0B58, 0x0B5B }, ::unicode::General_Category::Unassigned }, // [4] <reserved-0B58>..<reserved-0B5B>
    Prop<::unicode::General_Category> { { 0x0B5C, 0x0B5D }, ::unicode::General_Category::Other_Letter }, // [2] ORIYA LETTER RRA..ORIYA LETTER RHA
    Prop<::unicode::General_Category> { { 0x0B5E, 0x0B5E }, ::unicode::General_Category::Unassigned }, // <reserved-0B5E>
    Prop<::unicode::General_Category> { { 0x0B5F, 0x0B61 }, ::unicode::General_Category::Other_Letter }, // [3] ORIYA LETTER YYA..ORIYA LETTER VOCALIC LL
    Prop<::unicode::General_Category> { { 0x0B62, 0x0B63 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] ORIYA VOWEL SIGN VOCALIC L..ORIYA VOWEL SIGN VOCALIC LL
    Prop<::unicode::General_Category> { { 0x0B64, 0x0B65 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0B64>..<reserved-0B65>
    Prop<::unicode::General_Category> { { 0x0B66, 0x0B6F }, ::unicode::General_Category::Decimal_Number }, // [10] ORIYA DIGIT ZERO..ORIYA DIGIT NINE
    Prop<::unicode::General_Category> { { 0x0B70, 0x0B70 }, ::unicode::General_Category::Other_Symbol }, // ORIYA ISSHAR
    Prop<::unicode::General_Category> { { 0x0B71, 0x0B71 }, ::unicode::General_Category::Other_Letter }, // ORIYA LETTER WA
    Prop<::unicode::General_Category> { { 0x0B72, 0x0B77 }, ::unicode::General_Category::Other_Number }, // [6] ORIYA FRACTION ONE QUARTER..ORIYA FRACTION THREE SIXTEENTHS
    Prop<::unicode::General_Category> { { 0x0B78, 0x0B81 }, ::unicode::General_Category::Unassigned }, // [10] <reserved-0B78>..<reserved-0B81>
    Prop<::unicode::General_Category> { { 0x0B82, 0x0B82 }, ::unicode::General_Category::Nonspacing_Mark }, // TAMIL SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x0B83, 0x0B83 }, ::unicode::General_Category::Other_Letter }, // TAMIL SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x0B84, 0x0B84 }, ::unicode::General_Category::Unassigned }, // <reserved-0B84>
    Prop<::unicode::General_Category> { { 0x0B85, 0x0B8A }, ::unicode::General_Category::Other_Letter }, // [6] TAMIL LETTER A..TAMIL LETTER UU
    Prop<::unicode::General_Category> { { 0x0B8B, 0x0B8D }, ::unicode::General_Category::Unassigned }, // [3] <reserved-0B8B>..<reserved-0B8D>
    Prop<::unicode::General_Category> { { 0x0B8E, 0x0B90 }, ::unicode::General_Category::Other_Letter }, // [3] TAMIL LETTER E..TAMIL LETTER AI
    Prop<::unicode::General_Category> { { 0x0B91, 0x0B91 }, ::unicode::General_Category::Unassigned }, // <reserved-0B91>
    Prop<::unicode::General_Category> { { 0x0B92, 0x0B95 }, ::unicode::General_Category::Other_Letter }, // [4] TAMIL LETTER O..TAMIL LETTER KA
    Prop<::unicode::General_Category> { { 0x0B96, 0x0B98 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-0B96>..<reserved-0B98>
    Prop<::unicode::General_Category> { { 0x0B99, 0x0B9A }, ::unicode::General_Category::Other_Letter }, // [2] TAMIL LETTER NGA..TAMIL LETTER CA
    Prop<::unicode::General_Category> { { 0x0B9B, 0x0B9B }, ::unicode::General_Category::Unassigned }, // <reserved-0B9B>
    Prop<::unicode::General_Category> { { 0x0B9C, 0x0B9C }, ::unicode::General_Category::Other_Letter }, // TAMIL LETTER JA
    Prop<::unicode::General_Category> { { 0x0B9D, 0x0B9D }, ::unicode::General_Category::Unassigned }, // <reserved-0B9D>
    Prop<::unicode::General_Category> { { 0x0B9E, 0x0B9F }, ::unicode::General_Category::Other_Letter }, // [2] TAMIL LETTER NYA..TAMIL LETTER TTA
    Prop<::unicode::General_Category> { { 0x0BA0, 0x0BA2 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-0BA0>..<reserved-0BA2>
    Prop<::unicode::General_Category> { { 0x0BA3, 0x0BA4 }, ::unicode::General_Category::Other_Letter }, // [2] TAMIL LETTER NNA..TAMIL LETTER TA
    Prop<::unicode::General_Category> { { 0x0BA5, 0x0BA7 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-0BA5>..<reserved-0BA7>
    Prop<::unicode::General_Category> { { 0x0BA8, 0x0BAA }, ::unicode::General_Category::Other_Letter }, // [3] TAMIL LETTER NA..TAMIL LETTER PA
    Prop<::unicode::General_Category> { { 0x0BAB, 0x0BAD }, ::unicode::General_Category::Unassigned }, // [3] <reserved-0BAB>..<reserved-0BAD>
    Prop<::unicode::General_Category> { { 0x0BAE, 0x0BB9 }, ::unicode::General_Category::Other_Letter }, // [12] TAMIL LETTER MA..TAMIL LETTER HA
    Prop<::unicode::General_Category> { { 0x0BBA, 0x0BBD }, ::unicode::General_Category::Unassigned }, // [4] <reserved-0BBA>..<reserved-0BBD>
    Prop<::unicode::General_Category> { { 0x0BBE, 0x0BBF }, ::unicode::General_Category::Spacing_Mark }, // [2] TAMIL VOWEL SIGN AA..TAMIL VOWEL SIGN I
    Prop<::unicode::General_Category> { { 0x0BC0, 0x0BC0 }, ::unicode::General_Category::Nonspacing_Mark }, // TAMIL VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x0BC1, 0x0BC2 }, ::unicode::General_Category::Spacing_Mark }, // [2] TAMIL VOWEL SIGN U..TAMIL VOWEL SIGN UU
    Prop<::unicode::General_Category> { { 0x0BC3, 0x0BC5 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-0BC3>..<reserved-0BC5>
    Prop<::unicode::General_Category> { { 0x0BC6, 0x0BC8 }, ::unicode::General_Category::Spacing_Mark }, // [3] TAMIL VOWEL SIGN E..TAMIL VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x0BC9, 0x0BC9 }, ::unicode::General_Category::Unassigned }, // <reserved-0BC9>
    Prop<::unicode::General_Category> { { 0x0BCA, 0x0BCC }, ::unicode::General_Category::Spacing_Mark }, // [3] TAMIL VOWEL SIGN O..TAMIL VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x0BCD, 0x0BCD }, ::unicode::General_Category::Nonspacing_Mark }, // TAMIL SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x0BCE, 0x0BCF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0BCE>..<reserved-0BCF>
    Prop<::unicode::General_Category> { { 0x0BD0, 0x0BD0 }, ::unicode::General_Category::Other_Letter }, // TAMIL OM
    Prop<::unicode::General_Category> { { 0x0BD1, 0x0BD6 }, ::unicode::General_Category::Unassigned }, // [6] <reserved-0BD1>..<reserved-0BD6>
    Prop<::unicode::General_Category> { { 0x0BD7, 0x0BD7 }, ::unicode::General_Category::Spacing_Mark }, // TAMIL AU LENGTH MARK
    Prop<::unicode::General_Category> { { 0x0BD8, 0x0BE5 }, ::unicode::General_Category::Unassigned }, // [14] <reserved-0BD8>..<reserved-0BE5>
    Prop<::unicode::General_Category> { { 0x0BE6, 0x0BEF }, ::unicode::General_Category::Decimal_Number }, // [10] TAMIL DIGIT ZERO..TAMIL DIGIT NINE
    Prop<::unicode::General_Category> { { 0x0BF0, 0x0BF2 }, ::unicode::General_Category::Other_Number }, // [3] TAMIL NUMBER TEN..TAMIL NUMBER ONE THOUSAND
    Prop<::unicode::General_Category> { { 0x0BF3, 0x0BF8 }, ::unicode::General_Category::Other_Symbol }, // [6] TAMIL DAY SIGN..TAMIL AS ABOVE SIGN
    Prop<::unicode::General_Category> { { 0x0BF9, 0x0BF9 }, ::unicode::General_Category::Currency_Symbol }, // TAMIL RUPEE SIGN
    Prop<::unicode::General_Category> { { 0x0BFA, 0x0BFA }, ::unicode::General_Category::Other_Symbol }, // TAMIL NUMBER SIGN
    Prop<::unicode::General_Category> { { 0x0BFB, 0x0BFF }, ::unicode::General_Category::Unassigned }, // [5] <reserved-0BFB>..<reserved-0BFF>
    Prop<::unicode::General_Category> { { 0x0C00, 0x0C00 }, ::unicode::General_Category::Nonspacing_Mark }, // TELUGU SIGN COMBINING CANDRABINDU ABOVE
    Prop<::unicode::General_Category> { { 0x0C01, 0x0C03 }, ::unicode::General_Category::Spacing_Mark }, // [3] TELUGU SIGN CANDRABINDU..TELUGU SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x0C04, 0x0C04 }, ::unicode::General_Category::Nonspacing_Mark }, // TELUGU SIGN COMBINING ANUSVARA ABOVE
    Prop<::unicode::General_Category> { { 0x0C05, 0x0C0C }, ::unicode::General_Category::Other_Letter }, // [8] TELUGU LETTER A..TELUGU LETTER VOCALIC L
    Prop<::unicode::General_Category> { { 0x0C0D, 0x0C0D }, ::unicode::General_Category::Unassigned }, // <reserved-0C0D>
    Prop<::unicode::General_Category> { { 0x0C0E, 0x0C10 }, ::unicode::General_Category::Other_Letter }, // [3] TELUGU LETTER E..TELUGU LETTER AI
    Prop<::unicode::General_Category> { { 0x0C11, 0x0C11 }, ::unicode::General_Category::Unassigned }, // <reserved-0C11>
    Prop<::unicode::General_Category> { { 0x0C12, 0x0C28 }, ::unicode::General_Category::Other_Letter }, // [23] TELUGU LETTER O..TELUGU LETTER NA
    Prop<::unicode::General_Category> { { 0x0C29, 0x0C29 }, ::unicode::General_Category::Unassigned }, // <reserved-0C29>
    Prop<::unicode::General_Category> { { 0x0C2A, 0x0C39 }, ::unicode::General_Category::Other_Letter }, // [16] TELUGU LETTER PA..TELUGU LETTER HA
    Prop<::unicode::General_Category> { { 0x0C3A, 0x0C3B }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0C3A>..<reserved-0C3B>
    Prop<::unicode::General_Category> { { 0x0C3C, 0x0C3C }, ::unicode::General_Category::Nonspacing_Mark }, // TELUGU SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x0C3D, 0x0C3D }, ::unicode::General_Category::Other_Letter }, // TELUGU SIGN AVAGRAHA
    Prop<::unicode::General_Category> { { 0x0C3E, 0x0C40 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] TELUGU VOWEL SIGN AA..TELUGU VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x0C41, 0x0C44 }, ::unicode::General_Category::Spacing_Mark }, // [4] TELUGU VOWEL SIGN U..TELUGU VOWEL SIGN VOCALIC RR
    Prop<::unicode::General_Category> { { 0x0C45, 0x0C45 }, ::unicode::General_Category::Unassigned }, // <reserved-0C45>
    Prop<::unicode::General_Category> { { 0x0C46, 0x0C48 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] TELUGU VOWEL SIGN E..TELUGU VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x0C49, 0x0C49 }, ::unicode::General_Category::Unassigned }, // <reserved-0C49>
    Prop<::unicode::General_Category> { { 0x0C4A, 0x0C4D }, ::unicode::General_Category::Nonspacing_Mark }, // [4] TELUGU VOWEL SIGN O..TELUGU SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x0C4E, 0x0C54 }, ::unicode::General_Category::Unassigned }, // [7] <reserved-0C4E>..<reserved-0C54>
    Prop<::unicode::General_Category> { { 0x0C55, 0x0C56 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] TELUGU LENGTH MARK..TELUGU AI LENGTH MARK
    Prop<::unicode::General_Category> { { 0x0C57, 0x0C57 }, ::unicode::General_Category::Unassigned }, // <reserved-0C57>
    Prop<::unicode::General_Category> { { 0x0C58, 0x0C5A }, ::unicode::General_Category::Other_Letter }, // [3] TELUGU LETTER TSA..TELUGU LETTER RRRA
    Prop<::unicode::General_Category> { { 0x0C5B, 0x0C5B }, ::unicode::General_Category::Unassigned }, // <reserved-0C5B>
    Prop<::unicode::General_Category> { { 0x0C5C, 0x0C5D }, ::unicode::General_Category::Other_Letter }, // [2] TELUGU ARCHAIC SHRII..TELUGU LETTER NAKAARA POLLU
    Prop<::unicode::General_Category> { { 0x0C5E, 0x0C5F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0C5E>..<reserved-0C5F>
    Prop<::unicode::General_Category> { { 0x0C60, 0x0C61 }, ::unicode::General_Category::Other_Letter }, // [2] TELUGU LETTER VOCALIC RR..TELUGU LETTER VOCALIC LL
    Prop<::unicode::General_Category> { { 0x0C62, 0x0C63 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] TELUGU VOWEL SIGN VOCALIC L..TELUGU VOWEL SIGN VOCALIC LL
    Prop<::unicode::General_Category> { { 0x0C64, 0x0C65 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0C64>..<reserved-0C65>
    Prop<::unicode::General_Category> { { 0x0C66, 0x0C6F }, ::unicode::General_Category::Decimal_Number }, // [10] TELUGU DIGIT ZERO..TELUGU DIGIT NINE
    Prop<::unicode::General_Category> { { 0x0C70, 0x0C76 }, ::unicode::General_Category::Unassigned }, // [7] <reserved-0C70>..<reserved-0C76>
    Prop<::unicode::General_Category> { { 0x0C77, 0x0C77 }, ::unicode::General_Category::Other_Punctuation }, // TELUGU SIGN SIDDHAM
    Prop<::unicode::General_Category> { { 0x0C78, 0x0C7E }, ::unicode::General_Category::Other_Number }, // [7] TELUGU FRACTION DIGIT ZERO FOR ODD POWERS OF FOUR..TELUGU FRACTION DIGIT THREE FOR EVEN POWERS OF FOUR
    Prop<::unicode::General_Category> { { 0x0C7F, 0x0C7F }, ::unicode::General_Category::Other_Symbol }, // TELUGU SIGN TUUMU
    Prop<::unicode::General_Category> { { 0x0C80, 0x0C80 }, ::unicode::General_Category::Other_Letter }, // KANNADA SIGN SPACING CANDRABINDU
    Prop<::unicode::General_Category> { { 0x0C81, 0x0C81 }, ::unicode::General_Category::Nonspacing_Mark }, // KANNADA SIGN CANDRABINDU
    Prop<::unicode::General_Category> { { 0x0C82, 0x0C83 }, ::unicode::General_Category::Spacing_Mark }, // [2] KANNADA SIGN ANUSVARA..KANNADA SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x0C84, 0x0C84 }, ::unicode::General_Category::Other_Punctuation }, // KANNADA SIGN SIDDHAM
    Prop<::unicode::General_Category> { { 0x0C85, 0x0C8C }, ::unicode::General_Category::Other_Letter }, // [8] KANNADA LETTER A..KANNADA LETTER VOCALIC L
    Prop<::unicode::General_Category> { { 0x0C8D, 0x0C8D }, ::unicode::General_Category::Unassigned }, // <reserved-0C8D>
    Prop<::unicode::General_Category> { { 0x0C8E, 0x0C90 }, ::unicode::General_Category::Other_Letter }, // [3] KANNADA LETTER E..KANNADA LETTER AI
    Prop<::unicode::General_Category> { { 0x0C91, 0x0C91 }, ::unicode::General_Category::Unassigned }, // <reserved-0C91>
    Prop<::unicode::General_Category> { { 0x0C92, 0x0CA8 }, ::unicode::General_Category::Other_Letter }, // [23] KANNADA LETTER O..KANNADA LETTER NA
    Prop<::unicode::General_Category> { { 0x0CA9, 0x0CA9 }, ::unicode::General_Category::Unassigned }, // <reserved-0CA9>
    Prop<::unicode::General_Category> { { 0x0CAA, 0x0CB3 }, ::unicode::General_Category::Other_Letter }, // [10] KANNADA LETTER PA..KANNADA LETTER LLA
    Prop<::unicode::General_Category> { { 0x0CB4, 0x0CB4 }, ::unicode::General_Category::Unassigned }, // <reserved-0CB4>
    Prop<::unicode::General_Category> { { 0x0CB5, 0x0CB9 }, ::unicode::General_Category::Other_Letter }, // [5] KANNADA LETTER VA..KANNADA LETTER HA
    Prop<::unicode::General_Category> { { 0x0CBA, 0x0CBB }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0CBA>..<reserved-0CBB>
    Prop<::unicode::General_Category> { { 0x0CBC, 0x0CBC }, ::unicode::General_Category::Nonspacing_Mark }, // KANNADA SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x0CBD, 0x0CBD }, ::unicode::General_Category::Other_Letter }, // KANNADA SIGN AVAGRAHA
    Prop<::unicode::General_Category> { { 0x0CBE, 0x0CBE }, ::unicode::General_Category::Spacing_Mark }, // KANNADA VOWEL SIGN AA
    Prop<::unicode::General_Category> { { 0x0CBF, 0x0CBF }, ::unicode::General_Category::Nonspacing_Mark }, // KANNADA VOWEL SIGN I
    Prop<::unicode::General_Category> { { 0x0CC0, 0x0CC4 }, ::unicode::General_Category::Spacing_Mark }, // [5] KANNADA VOWEL SIGN II..KANNADA VOWEL SIGN VOCALIC RR
    Prop<::unicode::General_Category> { { 0x0CC5, 0x0CC5 }, ::unicode::General_Category::Unassigned }, // <reserved-0CC5>
    Prop<::unicode::General_Category> { { 0x0CC6, 0x0CC6 }, ::unicode::General_Category::Nonspacing_Mark }, // KANNADA VOWEL SIGN E
    Prop<::unicode::General_Category> { { 0x0CC7, 0x0CC8 }, ::unicode::General_Category::Spacing_Mark }, // [2] KANNADA VOWEL SIGN EE..KANNADA VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x0CC9, 0x0CC9 }, ::unicode::General_Category::Unassigned }, // <reserved-0CC9>
    Prop<::unicode::General_Category> { { 0x0CCA, 0x0CCB }, ::unicode::General_Category::Spacing_Mark }, // [2] KANNADA VOWEL SIGN O..KANNADA VOWEL SIGN OO
    Prop<::unicode::General_Category> { { 0x0CCC, 0x0CCD }, ::unicode::General_Category::Nonspacing_Mark }, // [2] KANNADA VOWEL SIGN AU..KANNADA SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x0CCE, 0x0CD4 }, ::unicode::General_Category::Unassigned }, // [7] <reserved-0CCE>..<reserved-0CD4>
    Prop<::unicode::General_Category> { { 0x0CD5, 0x0CD6 }, ::unicode::General_Category::Spacing_Mark }, // [2] KANNADA LENGTH MARK..KANNADA AI LENGTH MARK
    Prop<::unicode::General_Category> { { 0x0CD7, 0x0CDB }, ::unicode::General_Category::Unassigned }, // [5] <reserved-0CD7>..<reserved-0CDB>
    Prop<::unicode::General_Category> { { 0x0CDC, 0x0CDE }, ::unicode::General_Category::Other_Letter }, // [3] KANNADA ARCHAIC SHRII..KANNADA LETTER FA
    Prop<::unicode::General_Category> { { 0x0CDF, 0x0CDF }, ::unicode::General_Category::Unassigned }, // <reserved-0CDF>
    Prop<::unicode::General_Category> { { 0x0CE0, 0x0CE1 }, ::unicode::General_Category::Other_Letter }, // [2] KANNADA LETTER VOCALIC RR..KANNADA LETTER VOCALIC LL
    Prop<::unicode::General_Category> { { 0x0CE2, 0x0CE3 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] KANNADA VOWEL SIGN VOCALIC L..KANNADA VOWEL SIGN VOCALIC LL
    Prop<::unicode::General_Category> { { 0x0CE4, 0x0CE5 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0CE4>..<reserved-0CE5>
    Prop<::unicode::General_Category> { { 0x0CE6, 0x0CEF }, ::unicode::General_Category::Decimal_Number }, // [10] KANNADA DIGIT ZERO..KANNADA DIGIT NINE
    Prop<::unicode::General_Category> { { 0x0CF0, 0x0CF0 }, ::unicode::General_Category::Unassigned }, // <reserved-0CF0>
    Prop<::unicode::General_Category> { { 0x0CF1, 0x0CF2 }, ::unicode::General_Category::Other_Letter }, // [2] KANNADA SIGN JIHVAMULIYA..KANNADA SIGN UPADHMANIYA
    Prop<::unicode::General_Category> { { 0x0CF3, 0x0CF3 }, ::unicode::General_Category::Spacing_Mark }, // KANNADA SIGN COMBINING ANUSVARA ABOVE RIGHT
    Prop<::unicode::General_Category> { { 0x0CF4, 0x0CFF }, ::unicode::General_Category::Unassigned }, // [12] <reserved-0CF4>..<reserved-0CFF>
    Prop<::unicode::General_Category> { { 0x0D00, 0x0D01 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MALAYALAM SIGN COMBINING ANUSVARA ABOVE..MALAYALAM SIGN CANDRABINDU
    Prop<::unicode::General_Category> { { 0x0D02, 0x0D03 }, ::unicode::General_Category::Spacing_Mark }, // [2] MALAYALAM SIGN ANUSVARA..MALAYALAM SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x0D04, 0x0D0C }, ::unicode::General_Category::Other_Letter }, // [9] MALAYALAM LETTER VEDIC ANUSVARA..MALAYALAM LETTER VOCALIC L
    Prop<::unicode::General_Category> { { 0x0D0D, 0x0D0D }, ::unicode::General_Category::Unassigned }, // <reserved-0D0D>
    Prop<::unicode::General_Category> { { 0x0D0E, 0x0D10 }, ::unicode::General_Category::Other_Letter }, // [3] MALAYALAM LETTER E..MALAYALAM LETTER AI
    Prop<::unicode::General_Category> { { 0x0D11, 0x0D11 }, ::unicode::General_Category::Unassigned }, // <reserved-0D11>
    Prop<::unicode::General_Category> { { 0x0D12, 0x0D3A }, ::unicode::General_Category::Other_Letter }, // [41] MALAYALAM LETTER O..MALAYALAM LETTER TTTA
    Prop<::unicode::General_Category> { { 0x0D3B, 0x0D3C }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MALAYALAM SIGN VERTICAL BAR VIRAMA..MALAYALAM SIGN CIRCULAR VIRAMA
    Prop<::unicode::General_Category> { { 0x0D3D, 0x0D3D }, ::unicode::General_Category::Other_Letter }, // MALAYALAM SIGN AVAGRAHA
    Prop<::unicode::General_Category> { { 0x0D3E, 0x0D40 }, ::unicode::General_Category::Spacing_Mark }, // [3] MALAYALAM VOWEL SIGN AA..MALAYALAM VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x0D41, 0x0D44 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] MALAYALAM VOWEL SIGN U..MALAYALAM VOWEL SIGN VOCALIC RR
    Prop<::unicode::General_Category> { { 0x0D45, 0x0D45 }, ::unicode::General_Category::Unassigned }, // <reserved-0D45>
    Prop<::unicode::General_Category> { { 0x0D46, 0x0D48 }, ::unicode::General_Category::Spacing_Mark }, // [3] MALAYALAM VOWEL SIGN E..MALAYALAM VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x0D49, 0x0D49 }, ::unicode::General_Category::Unassigned }, // <reserved-0D49>
    Prop<::unicode::General_Category> { { 0x0D4A, 0x0D4C }, ::unicode::General_Category::Spacing_Mark }, // [3] MALAYALAM VOWEL SIGN O..MALAYALAM VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x0D4D, 0x0D4D }, ::unicode::General_Category::Nonspacing_Mark }, // MALAYALAM SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x0D4E, 0x0D4E }, ::unicode::General_Category::Other_Letter }, // MALAYALAM LETTER DOT REPH
    Prop<::unicode::General_Category> { { 0x0D4F, 0x0D4F }, ::unicode::General_Category::Other_Symbol }, // MALAYALAM SIGN PARA
    Prop<::unicode::General_Category> { { 0x0D50, 0x0D53 }, ::unicode::General_Category::Unassigned }, // [4] <reserved-0D50>..<reserved-0D53>
    Prop<::unicode::General_Category> { { 0x0D54, 0x0D56 }, ::unicode::General_Category::Other_Letter }, // [3] MALAYALAM LETTER CHILLU M..MALAYALAM LETTER CHILLU LLL
    Prop<::unicode::General_Category> { { 0x0D57, 0x0D57 }, ::unicode::General_Category::Spacing_Mark }, // MALAYALAM AU LENGTH MARK
    Prop<::unicode::General_Category> { { 0x0D58, 0x0D5E }, ::unicode::General_Category::Other_Number }, // [7] MALAYALAM FRACTION ONE ONE-HUNDRED-AND-SIXTIETH..MALAYALAM FRACTION ONE FIFTH
    Prop<::unicode::General_Category> { { 0x0D5F, 0x0D61 }, ::unicode::General_Category::Other_Letter }, // [3] MALAYALAM LETTER ARCHAIC II..MALAYALAM LETTER VOCALIC LL
    Prop<::unicode::General_Category> { { 0x0D62, 0x0D63 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MALAYALAM VOWEL SIGN VOCALIC L..MALAYALAM VOWEL SIGN VOCALIC LL
    Prop<::unicode::General_Category> { { 0x0D64, 0x0D65 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0D64>..<reserved-0D65>
    Prop<::unicode::General_Category> { { 0x0D66, 0x0D6F }, ::unicode::General_Category::Decimal_Number }, // [10] MALAYALAM DIGIT ZERO..MALAYALAM DIGIT NINE
    Prop<::unicode::General_Category> { { 0x0D70, 0x0D78 }, ::unicode::General_Category::Other_Number }, // [9] MALAYALAM NUMBER TEN..MALAYALAM FRACTION THREE SIXTEENTHS
    Prop<::unicode::General_Category> { { 0x0D79, 0x0D79 }, ::unicode::General_Category::Other_Symbol }, // MALAYALAM DATE MARK
    Prop<::unicode::General_Category> { { 0x0D7A, 0x0D7F }, ::unicode::General_Category::Other_Letter }, // [6] MALAYALAM LETTER CHILLU NN..MALAYALAM LETTER CHILLU K
    Prop<::unicode::General_Category> { { 0x0D80, 0x0D80 }, ::unicode::General_Category::Unassigned }, // <reserved-0D80>
    Prop<::unicode::General_Category> { { 0x0D81, 0x0D81 }, ::unicode::General_Category::Nonspacing_Mark }, // SINHALA SIGN CANDRABINDU
    Prop<::unicode::General_Category> { { 0x0D82, 0x0D83 }, ::unicode::General_Category::Spacing_Mark }, // [2] SINHALA SIGN ANUSVARAYA..SINHALA SIGN VISARGAYA
    Prop<::unicode::General_Category> { { 0x0D84, 0x0D84 }, ::unicode::General_Category::Unassigned }, // <reserved-0D84>
    Prop<::unicode::General_Category> { { 0x0D85, 0x0D96 }, ::unicode::General_Category::Other_Letter }, // [18] SINHALA LETTER AYANNA..SINHALA LETTER AUYANNA
    Prop<::unicode::General_Category> { { 0x0D97, 0x0D99 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-0D97>..<reserved-0D99>
    Prop<::unicode::General_Category> { { 0x0D9A, 0x0DB1 }, ::unicode::General_Category::Other_Letter }, // [24] SINHALA LETTER ALPAPRAANA KAYANNA..SINHALA LETTER DANTAJA NAYANNA
    Prop<::unicode::General_Category> { { 0x0DB2, 0x0DB2 }, ::unicode::General_Category::Unassigned }, // <reserved-0DB2>
    Prop<::unicode::General_Category> { { 0x0DB3, 0x0DBB }, ::unicode::General_Category::Other_Letter }, // [9] SINHALA LETTER SANYAKA DAYANNA..SINHALA LETTER RAYANNA
    Prop<::unicode::General_Category> { { 0x0DBC, 0x0DBC }, ::unicode::General_Category::Unassigned }, // <reserved-0DBC>
    Prop<::unicode::General_Category> { { 0x0DBD, 0x0DBD }, ::unicode::General_Category::Other_Letter }, // SINHALA LETTER DANTAJA LAYANNA
    Prop<::unicode::General_Category> { { 0x0DBE, 0x0DBF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0DBE>..<reserved-0DBF>
    Prop<::unicode::General_Category> { { 0x0DC0, 0x0DC6 }, ::unicode::General_Category::Other_Letter }, // [7] SINHALA LETTER VAYANNA..SINHALA LETTER FAYANNA
    Prop<::unicode::General_Category> { { 0x0DC7, 0x0DC9 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-0DC7>..<reserved-0DC9>
    Prop<::unicode::General_Category> { { 0x0DCA, 0x0DCA }, ::unicode::General_Category::Nonspacing_Mark }, // SINHALA SIGN AL-LAKUNA
    Prop<::unicode::General_Category> { { 0x0DCB, 0x0DCE }, ::unicode::General_Category::Unassigned }, // [4] <reserved-0DCB>..<reserved-0DCE>
    Prop<::unicode::General_Category> { { 0x0DCF, 0x0DD1 }, ::unicode::General_Category::Spacing_Mark }, // [3] SINHALA VOWEL SIGN AELA-PILLA..SINHALA VOWEL SIGN DIGA AEDA-PILLA
    Prop<::unicode::General_Category> { { 0x0DD2, 0x0DD4 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] SINHALA VOWEL SIGN KETTI IS-PILLA..SINHALA VOWEL SIGN KETTI PAA-PILLA
    Prop<::unicode::General_Category> { { 0x0DD5, 0x0DD5 }, ::unicode::General_Category::Unassigned }, // <reserved-0DD5>
    Prop<::unicode::General_Category> { { 0x0DD6, 0x0DD6 }, ::unicode::General_Category::Nonspacing_Mark }, // SINHALA VOWEL SIGN DIGA PAA-PILLA
    Prop<::unicode::General_Category> { { 0x0DD7, 0x0DD7 }, ::unicode::General_Category::Unassigned }, // <reserved-0DD7>
    Prop<::unicode::General_Category> { { 0x0DD8, 0x0DDF }, ::unicode::General_Category::Spacing_Mark }, // [8] SINHALA VOWEL SIGN GAETTA-PILLA..SINHALA VOWEL SIGN GAYANUKITTA
    Prop<::unicode::General_Category> { { 0x0DE0, 0x0DE5 }, ::unicode::General_Category::Unassigned }, // [6] <reserved-0DE0>..<reserved-0DE5>
    Prop<::unicode::General_Category> { { 0x0DE6, 0x0DEF }, ::unicode::General_Category::Decimal_Number }, // [10] SINHALA LITH DIGIT ZERO..SINHALA LITH DIGIT NINE
    Prop<::unicode::General_Category> { { 0x0DF0, 0x0DF1 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0DF0>..<reserved-0DF1>
    Prop<::unicode::General_Category> { { 0x0DF2, 0x0DF3 }, ::unicode::General_Category::Spacing_Mark }, // [2] SINHALA VOWEL SIGN DIGA GAETTA-PILLA..SINHALA VOWEL SIGN DIGA GAYANUKITTA
    Prop<::unicode::General_Category> { { 0x0DF4, 0x0DF4 }, ::unicode::General_Category::Other_Punctuation }, // SINHALA PUNCTUATION KUNDDALIYA
    Prop<::unicode::General_Category> { { 0x0DF5, 0x0E00 }, ::unicode::General_Category::Unassigned }, // [12] <reserved-0DF5>..<reserved-0E00>
    Prop<::unicode::General_Category> { { 0x0E01, 0x0E30 }, ::unicode::General_Category::Other_Letter }, // [48] THAI CHARACTER KO KAI..THAI CHARACTER SARA A
    Prop<::unicode::General_Category> { { 0x0E31, 0x0E31 }, ::unicode::General_Category::Nonspacing_Mark }, // THAI CHARACTER MAI HAN-AKAT
    Prop<::unicode::General_Category> { { 0x0E32, 0x0E33 }, ::unicode::General_Category::Other_Letter }, // [2] THAI CHARACTER SARA AA..THAI CHARACTER SARA AM
    Prop<::unicode::General_Category> { { 0x0E34, 0x0E3A }, ::unicode::General_Category::Nonspacing_Mark }, // [7] THAI CHARACTER SARA I..THAI CHARACTER PHINTHU
    Prop<::unicode::General_Category> { { 0x0E3B, 0x0E3E }, ::unicode::General_Category::Unassigned }, // [4] <reserved-0E3B>..<reserved-0E3E>
    Prop<::unicode::General_Category> { { 0x0E3F, 0x0E3F }, ::unicode::General_Category::Currency_Symbol }, // THAI CURRENCY SYMBOL BAHT
    Prop<::unicode::General_Category> { { 0x0E40, 0x0E45 }, ::unicode::General_Category::Other_Letter }, // [6] THAI CHARACTER SARA E..THAI CHARACTER LAKKHANGYAO
    Prop<::unicode::General_Category> { { 0x0E46, 0x0E46 }, ::unicode::General_Category::Modifier_Letter }, // THAI CHARACTER MAIYAMOK
    Prop<::unicode::General_Category> { { 0x0E47, 0x0E4E }, ::unicode::General_Category::Nonspacing_Mark }, // [8] THAI CHARACTER MAITAIKHU..THAI CHARACTER YAMAKKAN
    Prop<::unicode::General_Category> { { 0x0E4F, 0x0E4F }, ::unicode::General_Category::Other_Punctuation }, // THAI CHARACTER FONGMAN
    Prop<::unicode::General_Category> { { 0x0E50, 0x0E59 }, ::unicode::General_Category::Decimal_Number }, // [10] THAI DIGIT ZERO..THAI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x0E5A, 0x0E5B }, ::unicode::General_Category::Other_Punctuation }, // [2] THAI CHARACTER ANGKHANKHU..THAI CHARACTER KHOMUT
    Prop<::unicode::General_Category> { { 0x0E5C, 0x0E80 }, ::unicode::General_Category::Unassigned }, // [37] <reserved-0E5C>..<reserved-0E80>
    Prop<::unicode::General_Category> { { 0x0E81, 0x0E82 }, ::unicode::General_Category::Other_Letter }, // [2] LAO LETTER KO..LAO LETTER KHO SUNG
    Prop<::unicode::General_Category> { { 0x0E83, 0x0E83 }, ::unicode::General_Category::Unassigned }, // <reserved-0E83>
    Prop<::unicode::General_Category> { { 0x0E84, 0x0E84 }, ::unicode::General_Category::Other_Letter }, // LAO LETTER KHO TAM
    Prop<::unicode::General_Category> { { 0x0E85, 0x0E85 }, ::unicode::General_Category::Unassigned }, // <reserved-0E85>
    Prop<::unicode::General_Category> { { 0x0E86, 0x0E8A }, ::unicode::General_Category::Other_Letter }, // [5] LAO LETTER PALI GHA..LAO LETTER SO TAM
    Prop<::unicode::General_Category> { { 0x0E8B, 0x0E8B }, ::unicode::General_Category::Unassigned }, // <reserved-0E8B>
    Prop<::unicode::General_Category> { { 0x0E8C, 0x0EA3 }, ::unicode::General_Category::Other_Letter }, // [24] LAO LETTER PALI JHA..LAO LETTER LO LING
    Prop<::unicode::General_Category> { { 0x0EA4, 0x0EA4 }, ::unicode::General_Category::Unassigned }, // <reserved-0EA4>
    Prop<::unicode::General_Category> { { 0x0EA5, 0x0EA5 }, ::unicode::General_Category::Other_Letter }, // LAO LETTER LO LOOT
    Prop<::unicode::General_Category> { { 0x0EA6, 0x0EA6 }, ::unicode::General_Category::Unassigned }, // <reserved-0EA6>
    Prop<::unicode::General_Category> { { 0x0EA7, 0x0EB0 }, ::unicode::General_Category::Other_Letter }, // [10] LAO LETTER WO..LAO VOWEL SIGN A
    Prop<::unicode::General_Category> { { 0x0EB1, 0x0EB1 }, ::unicode::General_Category::Nonspacing_Mark }, // LAO VOWEL SIGN MAI KAN
    Prop<::unicode::General_Category> { { 0x0EB2, 0x0EB3 }, ::unicode::General_Category::Other_Letter }, // [2] LAO VOWEL SIGN AA..LAO VOWEL SIGN AM
    Prop<::unicode::General_Category> { { 0x0EB4, 0x0EBC }, ::unicode::General_Category::Nonspacing_Mark }, // [9] LAO VOWEL SIGN I..LAO SEMIVOWEL SIGN LO
    Prop<::unicode::General_Category> { { 0x0EBD, 0x0EBD }, ::unicode::General_Category::Other_Letter }, // LAO SEMIVOWEL SIGN NYO
    Prop<::unicode::General_Category> { { 0x0EBE, 0x0EBF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0EBE>..<reserved-0EBF>
    Prop<::unicode::General_Category> { { 0x0EC0, 0x0EC4 }, ::unicode::General_Category::Other_Letter }, // [5] LAO VOWEL SIGN E..LAO VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x0EC5, 0x0EC5 }, ::unicode::General_Category::Unassigned }, // <reserved-0EC5>
    Prop<::unicode::General_Category> { { 0x0EC6, 0x0EC6 }, ::unicode::General_Category::Modifier_Letter }, // LAO KO LA
    Prop<::unicode::General_Category> { { 0x0EC7, 0x0EC7 }, ::unicode::General_Category::Unassigned }, // <reserved-0EC7>
    Prop<::unicode::General_Category> { { 0x0EC8, 0x0ECE }, ::unicode::General_Category::Nonspacing_Mark }, // [7] LAO TONE MAI EK..LAO YAMAKKAN
    Prop<::unicode::General_Category> { { 0x0ECF, 0x0ECF }, ::unicode::General_Category::Unassigned }, // <reserved-0ECF>
    Prop<::unicode::General_Category> { { 0x0ED0, 0x0ED9 }, ::unicode::General_Category::Decimal_Number }, // [10] LAO DIGIT ZERO..LAO DIGIT NINE
    Prop<::unicode::General_Category> { { 0x0EDA, 0x0EDB }, ::unicode::General_Category::Unassigned }, // [2] <reserved-0EDA>..<reserved-0EDB>
    Prop<::unicode::General_Category> { { 0x0EDC, 0x0EDF }, ::unicode::General_Category::Other_Letter }, // [4] LAO HO NO..LAO LETTER KHMU NYO
    Prop<::unicode::General_Category> { { 0x0EE0, 0x0EFF }, ::unicode::General_Category::Unassigned }, // [32] <reserved-0EE0>..<reserved-0EFF>
    Prop<::unicode::General_Category> { { 0x0F00, 0x0F00 }, ::unicode::General_Category::Other_Letter }, // TIBETAN SYLLABLE OM
    Prop<::unicode::General_Category> { { 0x0F01, 0x0F03 }, ::unicode::General_Category::Other_Symbol }, // [3] TIBETAN MARK GTER YIG MGO TRUNCATED A..TIBETAN MARK GTER YIG MGO -UM GTER TSHEG MA
    Prop<::unicode::General_Category> { { 0x0F04, 0x0F12 }, ::unicode::General_Category::Other_Punctuation }, // [15] TIBETAN MARK INITIAL YIG MGO MDUN MA..TIBETAN MARK RGYA GRAM SHAD
    Prop<::unicode::General_Category> { { 0x0F13, 0x0F13 }, ::unicode::General_Category::Other_Symbol }, // TIBETAN MARK CARET -DZUD RTAGS ME LONG CAN
    Prop<::unicode::General_Category> { { 0x0F14, 0x0F14 }, ::unicode::General_Category::Other_Punctuation }, // TIBETAN MARK GTER TSHEG
    Prop<::unicode::General_Category> { { 0x0F15, 0x0F17 }, ::unicode::General_Category::Other_Symbol }, // [3] TIBETAN LOGOTYPE SIGN CHAD RTAGS..TIBETAN ASTROLOGICAL SIGN SGRA GCAN -CHAR RTAGS
    Prop<::unicode::General_Category> { { 0x0F18, 0x0F19 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] TIBETAN ASTROLOGICAL SIGN -KHYUD PA..TIBETAN ASTROLOGICAL SIGN SDONG TSHUGS
    Prop<::unicode::General_Category> { { 0x0F1A, 0x0F1F }, ::unicode::General_Category::Other_Symbol }, // [6] TIBETAN SIGN RDEL DKAR GCIG..TIBETAN SIGN RDEL DKAR RDEL NAG
    Prop<::unicode::General_Category> { { 0x0F20, 0x0F29 }, ::unicode::General_Category::Decimal_Number }, // [10] TIBETAN DIGIT ZERO..TIBETAN DIGIT NINE
    Prop<::unicode::General_Category> { { 0x0F2A, 0x0F33 }, ::unicode::General_Category::Other_Number }, // [10] TIBETAN DIGIT HALF ONE..TIBETAN DIGIT HALF ZERO
    Prop<::unicode::General_Category> { { 0x0F34, 0x0F34 }, ::unicode::General_Category::Other_Symbol }, // TIBETAN MARK BSDUS RTAGS
    Prop<::unicode::General_Category> { { 0x0F35, 0x0F35 }, ::unicode::General_Category::Nonspacing_Mark }, // TIBETAN MARK NGAS BZUNG NYI ZLA
    Prop<::unicode::General_Category> { { 0x0F36, 0x0F36 }, ::unicode::General_Category::Other_Symbol }, // TIBETAN MARK CARET -DZUD RTAGS BZHI MIG CAN
    Prop<::unicode::General_Category> { { 0x0F37, 0x0F37 }, ::unicode::General_Category::Nonspacing_Mark }, // TIBETAN MARK NGAS BZUNG SGOR RTAGS
    Prop<::unicode::General_Category> { { 0x0F38, 0x0F38 }, ::unicode::General_Category::Other_Symbol }, // TIBETAN MARK CHE MGO
    Prop<::unicode::General_Category> { { 0x0F39, 0x0F39 }, ::unicode::General_Category::Nonspacing_Mark }, // TIBETAN MARK TSA -PHRU
    Prop<::unicode::General_Category> { { 0x0F3A, 0x0F3A }, ::unicode::General_Category::Open_Punctuation }, // TIBETAN MARK GUG RTAGS GYON
    Prop<::unicode::General_Category> { { 0x0F3B, 0x0F3B }, ::unicode::General_Category::Close_Punctuation }, // TIBETAN MARK GUG RTAGS GYAS
    Prop<::unicode::General_Category> { { 0x0F3C, 0x0F3C }, ::unicode::General_Category::Open_Punctuation }, // TIBETAN MARK ANG KHANG GYON
    Prop<::unicode::General_Category> { { 0x0F3D, 0x0F3D }, ::unicode::General_Category::Close_Punctuation }, // TIBETAN MARK ANG KHANG GYAS
    Prop<::unicode::General_Category> { { 0x0F3E, 0x0F3F }, ::unicode::General_Category::Spacing_Mark }, // [2] TIBETAN SIGN YAR TSHES..TIBETAN SIGN MAR TSHES
    Prop<::unicode::General_Category> { { 0x0F40, 0x0F47 }, ::unicode::General_Category::Other_Letter }, // [8] TIBETAN LETTER KA..TIBETAN LETTER JA
    Prop<::unicode::General_Category> { { 0x0F48, 0x0F48 }, ::unicode::General_Category::Unassigned }, // <reserved-0F48>
    Prop<::unicode::General_Category> { { 0x0F49, 0x0F6C }, ::unicode::General_Category::Other_Letter }, // [36] TIBETAN LETTER NYA..TIBETAN LETTER RRA
    Prop<::unicode::General_Category> { { 0x0F6D, 0x0F70 }, ::unicode::General_Category::Unassigned }, // [4] <reserved-0F6D>..<reserved-0F70>
    Prop<::unicode::General_Category> { { 0x0F71, 0x0F7E }, ::unicode::General_Category::Nonspacing_Mark }, // [14] TIBETAN VOWEL SIGN AA..TIBETAN SIGN RJES SU NGA RO
    Prop<::unicode::General_Category> { { 0x0F7F, 0x0F7F }, ::unicode::General_Category::Spacing_Mark }, // TIBETAN SIGN RNAM BCAD
    Prop<::unicode::General_Category> { { 0x0F80, 0x0F84 }, ::unicode::General_Category::Nonspacing_Mark }, // [5] TIBETAN VOWEL SIGN REVERSED I..TIBETAN MARK HALANTA
    Prop<::unicode::General_Category> { { 0x0F85, 0x0F85 }, ::unicode::General_Category::Other_Punctuation }, // TIBETAN MARK PALUTA
    Prop<::unicode::General_Category> { { 0x0F86, 0x0F87 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] TIBETAN SIGN LCI RTAGS..TIBETAN SIGN YANG RTAGS
    Prop<::unicode::General_Category> { { 0x0F88, 0x0F8C }, ::unicode::General_Category::Other_Letter }, // [5] TIBETAN SIGN LCE TSA CAN..TIBETAN SIGN INVERTED MCHU CAN
    Prop<::unicode::General_Category> { { 0x0F8D, 0x0F97 }, ::unicode::General_Category::Nonspacing_Mark }, // [11] TIBETAN SUBJOINED SIGN LCE TSA CAN..TIBETAN SUBJOINED LETTER JA
    Prop<::unicode::General_Category> { { 0x0F98, 0x0F98 }, ::unicode::General_Category::Unassigned }, // <reserved-0F98>
    Prop<::unicode::General_Category> { { 0x0F99, 0x0FBC }, ::unicode::General_Category::Nonspacing_Mark }, // [36] TIBETAN SUBJOINED LETTER NYA..TIBETAN SUBJOINED LETTER FIXED-FORM RA
    Prop<::unicode::General_Category> { { 0x0FBD, 0x0FBD }, ::unicode::General_Category::Unassigned }, // <reserved-0FBD>
    Prop<::unicode::General_Category> { { 0x0FBE, 0x0FC5 }, ::unicode::General_Category::Other_Symbol }, // [8] TIBETAN KU RU KHA..TIBETAN SYMBOL RDO RJE
    Prop<::unicode::General_Category> { { 0x0FC6, 0x0FC6 }, ::unicode::General_Category::Nonspacing_Mark }, // TIBETAN SYMBOL PADMA GDAN
    Prop<::unicode::General_Category> { { 0x0FC7, 0x0FCC }, ::unicode::General_Category::Other_Symbol }, // [6] TIBETAN SYMBOL RDO RJE RGYA GRAM..TIBETAN SYMBOL NOR BU BZHI -KHYIL
    Prop<::unicode::General_Category> { { 0x0FCD, 0x0FCD }, ::unicode::General_Category::Unassigned }, // <reserved-0FCD>
    Prop<::unicode::General_Category> { { 0x0FCE, 0x0FCF }, ::unicode::General_Category::Other_Symbol }, // [2] TIBETAN SIGN RDEL NAG RDEL DKAR..TIBETAN SIGN RDEL NAG GSUM
    Prop<::unicode::General_Category> { { 0x0FD0, 0x0FD4 }, ::unicode::General_Category::Other_Punctuation }, // [5] TIBETAN MARK BSKA- SHOG GI MGO RGYAN..TIBETAN MARK CLOSING BRDA RNYING YIG MGO SGAB MA
    Prop<::unicode::General_Category> { { 0x0FD5, 0x0FD8 }, ::unicode::General_Category::Other_Symbol }, // [4] RIGHT-FACING SVASTI SIGN..LEFT-FACING SVASTI SIGN WITH DOTS
    Prop<::unicode::General_Category> { { 0x0FD9, 0x0FDA }, ::unicode::General_Category::Other_Punctuation }, // [2] TIBETAN MARK LEADING MCHAN RTAGS..TIBETAN MARK TRAILING MCHAN RTAGS
    Prop<::unicode::General_Category> { { 0x0FDB, 0x0FFF }, ::unicode::General_Category::Unassigned }, // [37] <reserved-0FDB>..<reserved-0FFF>
    Prop<::unicode::General_Category> { { 0x1000, 0x102A }, ::unicode::General_Category::Other_Letter }, // [43] MYANMAR LETTER KA..MYANMAR LETTER AU
    Prop<::unicode::General_Category> { { 0x102B, 0x102C }, ::unicode::General_Category::Spacing_Mark }, // [2] MYANMAR VOWEL SIGN TALL AA..MYANMAR VOWEL SIGN AA
    Prop<::unicode::General_Category> { { 0x102D, 0x1030 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] MYANMAR VOWEL SIGN I..MYANMAR VOWEL SIGN UU
    Prop<::unicode::General_Category> { { 0x1031, 0x1031 }, ::unicode::General_Category::Spacing_Mark }, // MYANMAR VOWEL SIGN E
    Prop<::unicode::General_Category> { { 0x1032, 0x1037 }, ::unicode::General_Category::Nonspacing_Mark }, // [6] MYANMAR VOWEL SIGN AI..MYANMAR SIGN DOT BELOW
    Prop<::unicode::General_Category> { { 0x1038, 0x1038 }, ::unicode::General_Category::Spacing_Mark }, // MYANMAR SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x1039, 0x103A }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MYANMAR SIGN VIRAMA..MYANMAR SIGN ASAT
    Prop<::unicode::General_Category> { { 0x103B, 0x103C }, ::unicode::General_Category::Spacing_Mark }, // [2] MYANMAR CONSONANT SIGN MEDIAL YA..MYANMAR CONSONANT SIGN MEDIAL RA
    Prop<::unicode::General_Category> { { 0x103D, 0x103E }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MYANMAR CONSONANT SIGN MEDIAL WA..MYANMAR CONSONANT SIGN MEDIAL HA
    Prop<::unicode::General_Category> { { 0x103F, 0x103F }, ::unicode::General_Category::Other_Letter }, // MYANMAR LETTER GREAT SA
    Prop<::unicode::General_Category> { { 0x1040, 0x1049 }, ::unicode::General_Category::Decimal_Number }, // [10] MYANMAR DIGIT ZERO..MYANMAR DIGIT NINE
    Prop<::unicode::General_Category> { { 0x104A, 0x104F }, ::unicode::General_Category::Other_Punctuation }, // [6] MYANMAR SIGN LITTLE SECTION..MYANMAR SYMBOL GENITIVE
    Prop<::unicode::General_Category> { { 0x1050, 0x1055 }, ::unicode::General_Category::Other_Letter }, // [6] MYANMAR LETTER SHA..MYANMAR LETTER VOCALIC LL
    Prop<::unicode::General_Category> { { 0x1056, 0x1057 }, ::unicode::General_Category::Spacing_Mark }, // [2] MYANMAR VOWEL SIGN VOCALIC R..MYANMAR VOWEL SIGN VOCALIC RR
    Prop<::unicode::General_Category> { { 0x1058, 0x1059 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MYANMAR VOWEL SIGN VOCALIC L..MYANMAR VOWEL SIGN VOCALIC LL
    Prop<::unicode::General_Category> { { 0x105A, 0x105D }, ::unicode::General_Category::Other_Letter }, // [4] MYANMAR LETTER MON NGA..MYANMAR LETTER MON BBE
    Prop<::unicode::General_Category> { { 0x105E, 0x1060 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] MYANMAR CONSONANT SIGN MON MEDIAL NA..MYANMAR CONSONANT SIGN MON MEDIAL LA
    Prop<::unicode::General_Category> { { 0x1061, 0x1061 }, ::unicode::General_Category::Other_Letter }, // MYANMAR LETTER SGAW KAREN SHA
    Prop<::unicode::General_Category> { { 0x1062, 0x1064 }, ::unicode::General_Category::Spacing_Mark }, // [3] MYANMAR VOWEL SIGN SGAW KAREN EU..MYANMAR TONE MARK SGAW KAREN KE PHO
    Prop<::unicode::General_Category> { { 0x1065, 0x1066 }, ::unicode::General_Category::Other_Letter }, // [2] MYANMAR LETTER WESTERN PWO KAREN THA..MYANMAR LETTER WESTERN PWO KAREN PWA
    Prop<::unicode::General_Category> { { 0x1067, 0x106D }, ::unicode::General_Category::Spacing_Mark }, // [7] MYANMAR VOWEL SIGN WESTERN PWO KAREN EU..MYANMAR SIGN WESTERN PWO KAREN TONE-5
    Prop<::unicode::General_Category> { { 0x106E, 0x1070 }, ::unicode::General_Category::Other_Letter }, // [3] MYANMAR LETTER EASTERN PWO KAREN NNA..MYANMAR LETTER EASTERN PWO KAREN GHWA
    Prop<::unicode::General_Category> { { 0x1071, 0x1074 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] MYANMAR VOWEL SIGN GEBA KAREN I..MYANMAR VOWEL SIGN KAYAH EE
    Prop<::unicode::General_Category> { { 0x1075, 0x1081 }, ::unicode::General_Category::Other_Letter }, // [13] MYANMAR LETTER SHAN KA..MYANMAR LETTER SHAN HA
    Prop<::unicode::General_Category> { { 0x1082, 0x1082 }, ::unicode::General_Category::Nonspacing_Mark }, // MYANMAR CONSONANT SIGN SHAN MEDIAL WA
    Prop<::unicode::General_Category> { { 0x1083, 0x1084 }, ::unicode::General_Category::Spacing_Mark }, // [2] MYANMAR VOWEL SIGN SHAN AA..MYANMAR VOWEL SIGN SHAN E
    Prop<::unicode::General_Category> { { 0x1085, 0x1086 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MYANMAR VOWEL SIGN SHAN E ABOVE..MYANMAR VOWEL SIGN SHAN FINAL Y
    Prop<::unicode::General_Category> { { 0x1087, 0x108C }, ::unicode::General_Category::Spacing_Mark }, // [6] MYANMAR SIGN SHAN TONE-2..MYANMAR SIGN SHAN COUNCIL TONE-3
    Prop<::unicode::General_Category> { { 0x108D, 0x108D }, ::unicode::General_Category::Nonspacing_Mark }, // MYANMAR SIGN SHAN COUNCIL EMPHATIC TONE
    Prop<::unicode::General_Category> { { 0x108E, 0x108E }, ::unicode::General_Category::Other_Letter }, // MYANMAR LETTER RUMAI PALAUNG FA
    Prop<::unicode::General_Category> { { 0x108F, 0x108F }, ::unicode::General_Category::Spacing_Mark }, // MYANMAR SIGN RUMAI PALAUNG TONE-5
    Prop<::unicode::General_Category> { { 0x1090, 0x1099 }, ::unicode::General_Category::Decimal_Number }, // [10] MYANMAR SHAN DIGIT ZERO..MYANMAR SHAN DIGIT NINE
    Prop<::unicode::General_Category> { { 0x109A, 0x109C }, ::unicode::General_Category::Spacing_Mark }, // [3] MYANMAR SIGN KHAMTI TONE-1..MYANMAR VOWEL SIGN AITON A
    Prop<::unicode::General_Category> { { 0x109D, 0x109D }, ::unicode::General_Category::Nonspacing_Mark }, // MYANMAR VOWEL SIGN AITON AI
    Prop<::unicode::General_Category> { { 0x109E, 0x109F }, ::unicode::General_Category::Other_Symbol }, // [2] MYANMAR SYMBOL SHAN ONE..MYANMAR SYMBOL SHAN EXCLAMATION
    Prop<::unicode::General_Category> { { 0x10A0, 0x10C5 }, ::unicode::General_Category::Uppercase_Letter }, // [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Prop<::unicode::General_Category> { { 0x10C6, 0x10C6 }, ::unicode::General_Category::Unassigned }, // <reserved-10C6>
    Prop<::unicode::General_Category> { { 0x10C7, 0x10C7 }, ::unicode::General_Category::Uppercase_Letter }, // GEORGIAN CAPITAL LETTER YN
    Prop<::unicode::General_Category> { { 0x10C8, 0x10CC }, ::unicode::General_Category::Unassigned }, // [5] <reserved-10C8>..<reserved-10CC>
    Prop<::unicode::General_Category> { { 0x10CD, 0x10CD }, ::unicode::General_Category::Uppercase_Letter }, // GEORGIAN CAPITAL LETTER AEN
    Prop<::unicode::General_Category> { { 0x10CE, 0x10CF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-10CE>..<reserved-10CF>
    Prop<::unicode::General_Category> { { 0x10D0, 0x10FA }, ::unicode::General_Category::Lowercase_Letter }, // [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Prop<::unicode::General_Category> { { 0x10FB, 0x10FB }, ::unicode::General_Category::Other_Punctuation }, // GEORGIAN PARAGRAPH SEPARATOR
    Prop<::unicode::General_Category> { { 0x10FC, 0x10FC }, ::unicode::General_Category::Modifier_Letter }, // MODIFIER LETTER GEORGIAN NAR
    Prop<::unicode::General_Category> { { 0x10FD, 0x10FF }, ::unicode::General_Category::Lowercase_Letter }, // [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Prop<::unicode::General_Category> { { 0x1100, 0x1248 }, ::unicode::General_Category::Other_Letter }, // [329] HANGUL CHOSEONG KIYEOK..ETHIOPIC SYLLABLE QWA
    Prop<::unicode::General_Category> { { 0x1249, 0x1249 }, ::unicode::General_Category::Unassigned }, // <reserved-1249>
    Prop<::unicode::General_Category> { { 0x124A, 0x124D }, ::unicode::General_Category::Other_Letter }, // [4] ETHIOPIC SYLLABLE QWI..ETHIOPIC SYLLABLE QWE
    Prop<::unicode::General_Category> { { 0x124E, 0x124F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-124E>..<reserved-124F>
    Prop<::unicode::General_Category> { { 0x1250, 0x1256 }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE QHA..ETHIOPIC SYLLABLE QHO
    Prop<::unicode::General_Category> { { 0x1257, 0x1257 }, ::unicode::General_Category::Unassigned }, // <reserved-1257>
    Prop<::unicode::General_Category> { { 0x1258, 0x1258 }, ::unicode::General_Category::Other_Letter }, // ETHIOPIC SYLLABLE QHWA
    Prop<::unicode::General_Category> { { 0x1259, 0x1259 }, ::unicode::General_Category::Unassigned }, // <reserved-1259>
    Prop<::unicode::General_Category> { { 0x125A, 0x125D }, ::unicode::General_Category::Other_Letter }, // [4] ETHIOPIC SYLLABLE QHWI..ETHIOPIC SYLLABLE QHWE
    Prop<::unicode::General_Category> { { 0x125E, 0x125F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-125E>..<reserved-125F>
    Prop<::unicode::General_Category> { { 0x1260, 0x1288 }, ::unicode::General_Category::Other_Letter }, // [41] ETHIOPIC SYLLABLE BA..ETHIOPIC SYLLABLE XWA
    Prop<::unicode::General_Category> { { 0x1289, 0x1289 }, ::unicode::General_Category::Unassigned }, // <reserved-1289>
    Prop<::unicode::General_Category> { { 0x128A, 0x128D }, ::unicode::General_Category::Other_Letter }, // [4] ETHIOPIC SYLLABLE XWI..ETHIOPIC SYLLABLE XWE
    Prop<::unicode::General_Category> { { 0x128E, 0x128F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-128E>..<reserved-128F>
    Prop<::unicode::General_Category> { { 0x1290, 0x12B0 }, ::unicode::General_Category::Other_Letter }, // [33] ETHIOPIC SYLLABLE NA..ETHIOPIC SYLLABLE KWA
    Prop<::unicode::General_Category> { { 0x12B1, 0x12B1 }, ::unicode::General_Category::Unassigned }, // <reserved-12B1>
    Prop<::unicode::General_Category> { { 0x12B2, 0x12B5 }, ::unicode::General_Category::Other_Letter }, // [4] ETHIOPIC SYLLABLE KWI..ETHIOPIC SYLLABLE KWE
    Prop<::unicode::General_Category> { { 0x12B6, 0x12B7 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-12B6>..<reserved-12B7>
    Prop<::unicode::General_Category> { { 0x12B8, 0x12BE }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE KXA..ETHIOPIC SYLLABLE KXO
    Prop<::unicode::General_Category> { { 0x12BF, 0x12BF }, ::unicode::General_Category::Unassigned }, // <reserved-12BF>
    Prop<::unicode::General_Category> { { 0x12C0, 0x12C0 }, ::unicode::General_Category::Other_Letter }, // ETHIOPIC SYLLABLE KXWA
    Prop<::unicode::General_Category> { { 0x12C1, 0x12C1 }, ::unicode::General_Category::Unassigned }, // <reserved-12C1>
    Prop<::unicode::General_Category> { { 0x12C2, 0x12C5 }, ::unicode::General_Category::Other_Letter }, // [4] ETHIOPIC SYLLABLE KXWI..ETHIOPIC SYLLABLE KXWE
    Prop<::unicode::General_Category> { { 0x12C6, 0x12C7 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-12C6>..<reserved-12C7>
    Prop<::unicode::General_Category> { { 0x12C8, 0x12D6 }, ::unicode::General_Category::Other_Letter }, // [15] ETHIOPIC SYLLABLE WA..ETHIOPIC SYLLABLE PHARYNGEAL O
    Prop<::unicode::General_Category> { { 0x12D7, 0x12D7 }, ::unicode::General_Category::Unassigned }, // <reserved-12D7>
    Prop<::unicode::General_Category> { { 0x12D8, 0x1310 }, ::unicode::General_Category::Other_Letter }, // [57] ETHIOPIC SYLLABLE ZA..ETHIOPIC SYLLABLE GWA
    Prop<::unicode::General_Category> { { 0x1311, 0x1311 }, ::unicode::General_Category::Unassigned }, // <reserved-1311>
    Prop<::unicode::General_Category> { { 0x1312, 0x1315 }, ::unicode::General_Category::Other_Letter }, // [4] ETHIOPIC SYLLABLE GWI..ETHIOPIC SYLLABLE GWE
    Prop<::unicode::General_Category> { { 0x1316, 0x1317 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1316>..<reserved-1317>
    Prop<::unicode::General_Category> { { 0x1318, 0x135A }, ::unicode::General_Category::Other_Letter }, // [67] ETHIOPIC SYLLABLE GGA..ETHIOPIC SYLLABLE FYA
    Prop<::unicode::General_Category> { { 0x135B, 0x135C }, ::unicode::General_Category::Unassigned }, // [2] <reserved-135B>..<reserved-135C>
    Prop<::unicode::General_Category> { { 0x135D, 0x135F }, ::unicode::General_Category::Nonspacing_Mark }, // [3] ETHIOPIC COMBINING GEMINATION AND VOWEL LENGTH MARK..ETHIOPIC COMBINING GEMINATION MARK
    Prop<::unicode::General_Category> { { 0x1360, 0x1368 }, ::unicode::General_Category::Other_Punctuation }, // [9] ETHIOPIC SECTION MARK..ETHIOPIC PARAGRAPH SEPARATOR
    Prop<::unicode::General_Category> { { 0x1369, 0x137C }, ::unicode::General_Category::Other_Number }, // [20] ETHIOPIC DIGIT ONE..ETHIOPIC NUMBER TEN THOUSAND
    Prop<::unicode::General_Category> { { 0x137D, 0x137F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-137D>..<reserved-137F>
    Prop<::unicode::General_Category> { { 0x1380, 0x138F }, ::unicode::General_Category::Other_Letter }, // [16] ETHIOPIC SYLLABLE SEBATBEIT MWA..ETHIOPIC SYLLABLE PWE
    Prop<::unicode::General_Category> { { 0x1390, 0x1399 }, ::unicode::General_Category::Other_Symbol }, // [10] ETHIOPIC TONAL MARK YIZET..ETHIOPIC TONAL MARK KURT
    Prop<::unicode::General_Category> { { 0x139A, 0x139F }, ::unicode::General_Category::Unassigned }, // [6] <reserved-139A>..<reserved-139F>
    Prop<::unicode::General_Category> { { 0x13A0, 0x13F5 }, ::unicode::General_Category::Uppercase_Letter }, // [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Prop<::unicode::General_Category> { { 0x13F6, 0x13F7 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-13F6>..<reserved-13F7>
    Prop<::unicode::General_Category> { { 0x13F8, 0x13FD }, ::unicode::General_Category::Lowercase_Letter }, // [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Prop<::unicode::General_Category> { { 0x13FE, 0x13FF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-13FE>..<reserved-13FF>
    Prop<::unicode::General_Category> { { 0x1400, 0x1400 }, ::unicode::General_Category::Dash_Punctuation }, // CANADIAN SYLLABICS HYPHEN
    Prop<::unicode::General_Category> { { 0x1401, 0x166C }, ::unicode::General_Category::Other_Letter }, // [620] CANADIAN SYLLABICS E..CANADIAN SYLLABICS CARRIER TTSA
    Prop<::unicode::General_Category> { { 0x166D, 0x166D }, ::unicode::General_Category::Other_Symbol }, // CANADIAN SYLLABICS CHI SIGN
    Prop<::unicode::General_Category> { { 0x166E, 0x166E }, ::unicode::General_Category::Other_Punctuation }, // CANADIAN SYLLABICS FULL STOP
    Prop<::unicode::General_Category> { { 0x166F, 0x167F }, ::unicode::General_Category::Other_Letter }, // [17] CANADIAN SYLLABICS QAI..CANADIAN SYLLABICS BLACKFOOT W
    Prop<::unicode::General_Category> { { 0x1680, 0x1680 }, ::unicode::General_Category::Space_Separator }, // OGHAM SPACE MARK
    Prop<::unicode::General_Category> { { 0x1681, 0x169A }, ::unicode::General_Category::Other_Letter }, // [26] OGHAM LETTER BEITH..OGHAM LETTER PEITH
    Prop<::unicode::General_Category> { { 0x169B, 0x169B }, ::unicode::General_Category::Open_Punctuation }, // OGHAM FEATHER MARK
    Prop<::unicode::General_Category> { { 0x169C, 0x169C }, ::unicode::General_Category::Close_Punctuation }, // OGHAM REVERSED FEATHER MARK
    Prop<::unicode::General_Category> { { 0x169D, 0x169F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-169D>..<reserved-169F>
    Prop<::unicode::General_Category> { { 0x16A0, 0x16EA }, ::unicode::General_Category::Other_Letter }, // [75] RUNIC LETTER FEHU FEOH FE F..RUNIC LETTER X
    Prop<::unicode::General_Category> { { 0x16EB, 0x16ED }, ::unicode::General_Category::Other_Punctuation }, // [3] RUNIC SINGLE PUNCTUATION..RUNIC CROSS PUNCTUATION
    Prop<::unicode::General_Category> { { 0x16EE, 0x16F0 }, ::unicode::General_Category::Letter_Number }, // [3] RUNIC ARLAUG SYMBOL..RUNIC BELGTHOR SYMBOL
    Prop<::unicode::General_Category> { { 0x16F1, 0x16F8 }, ::unicode::General_Category::Other_Letter }, // [8] RUNIC LETTER K..RUNIC LETTER FRANKS CASKET AESC
    Prop<::unicode::General_Category> { { 0x16F9, 0x16FF }, ::unicode::General_Category::Unassigned }, // [7] <reserved-16F9>..<reserved-16FF>
    Prop<::unicode::General_Category> { { 0x1700, 0x1711 }, ::unicode::General_Category::Other_Letter }, // [18] TAGALOG LETTER A..TAGALOG LETTER HA
    Prop<::unicode::General_Category> { { 0x1712, 0x1714 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] TAGALOG VOWEL SIGN I..TAGALOG SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x1715, 0x1715 }, ::unicode::General_Category::Spacing_Mark }, // TAGALOG SIGN PAMUDPOD
    Prop<::unicode::General_Category> { { 0x1716, 0x171E }, ::unicode::General_Category::Unassigned }, // [9] <reserved-1716>..<reserved-171E>
    Prop<::unicode::General_Category> { { 0x171F, 0x1731 }, ::unicode::General_Category::Other_Letter }, // [19] TAGALOG LETTER ARCHAIC RA..HANUNOO LETTER HA
    Prop<::unicode::General_Category> { { 0x1732, 0x1733 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] HANUNOO VOWEL SIGN I..HANUNOO VOWEL SIGN U
    Prop<::unicode::General_Category> { { 0x1734, 0x1734 }, ::unicode::General_Category::Spacing_Mark }, // HANUNOO SIGN PAMUDPOD
    Prop<::unicode::General_Category> { { 0x1735, 0x1736 }, ::unicode::General_Category::Other_Punctuation }, // [2] PHILIPPINE SINGLE PUNCTUATION..PHILIPPINE DOUBLE PUNCTUATION
    Prop<::unicode::General_Category> { { 0x1737, 0x173F }, ::unicode::General_Category::Unassigned }, // [9] <reserved-1737>..<reserved-173F>
    Prop<::unicode::General_Category> { { 0x1740, 0x1751 }, ::unicode::General_Category::Other_Letter }, // [18] BUHID LETTER A..BUHID LETTER HA
    Prop<::unicode::General_Category> { { 0x1752, 0x1753 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] BUHID VOWEL SIGN I..BUHID VOWEL SIGN U
    Prop<::unicode::General_Category> { { 0x1754, 0x175F }, ::unicode::General_Category::Unassigned }, // [12] <reserved-1754>..<reserved-175F>
    Prop<::unicode::General_Category> { { 0x1760, 0x176C }, ::unicode::General_Category::Other_Letter }, // [13] TAGBANWA LETTER A..TAGBANWA LETTER YA
    Prop<::unicode::General_Category> { { 0x176D, 0x176D }, ::unicode::General_Category::Unassigned }, // <reserved-176D>
    Prop<::unicode::General_Category> { { 0x176E, 0x1770 }, ::unicode::General_Category::Other_Letter }, // [3] TAGBANWA LETTER LA..TAGBANWA LETTER SA
    Prop<::unicode::General_Category> { { 0x1771, 0x1771 }, ::unicode::General_Category::Unassigned }, // <reserved-1771>
    Prop<::unicode::General_Category> { { 0x1772, 0x1773 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] TAGBANWA VOWEL SIGN I..TAGBANWA VOWEL SIGN U
    Prop<::unicode::General_Category> { { 0x1774, 0x177F }, ::unicode::General_Category::Unassigned }, // [12] <reserved-1774>..<reserved-177F>
    Prop<::unicode::General_Category> { { 0x1780, 0x17B3 }, ::unicode::General_Category::Other_Letter }, // [52] KHMER LETTER KA..KHMER INDEPENDENT VOWEL QAU
    Prop<::unicode::General_Category> { { 0x17B4, 0x17B5 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] KHMER VOWEL INHERENT AQ..KHMER VOWEL INHERENT AA
    Prop<::unicode::General_Category> { { 0x17B6, 0x17B6 }, ::unicode::General_Category::Spacing_Mark }, // KHMER VOWEL SIGN AA
    Prop<::unicode::General_Category> { { 0x17B7, 0x17BD }, ::unicode::General_Category::Nonspacing_Mark }, // [7] KHMER VOWEL SIGN I..KHMER VOWEL SIGN UA
    Prop<::unicode::General_Category> { { 0x17BE, 0x17C5 }, ::unicode::General_Category::Spacing_Mark }, // [8] KHMER VOWEL SIGN OE..KHMER VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x17C6, 0x17C6 }, ::unicode::General_Category::Nonspacing_Mark }, // KHMER SIGN NIKAHIT
    Prop<::unicode::General_Category> { { 0x17C7, 0x17C8 }, ::unicode::General_Category::Spacing_Mark }, // [2] KHMER SIGN REAHMUK..KHMER SIGN YUUKALEAPINTU
    Prop<::unicode::General_Category> { { 0x17C9, 0x17D3 }, ::unicode::General_Category::Nonspacing_Mark }, // [11] KHMER SIGN MUUSIKATOAN..KHMER SIGN BATHAMASAT
    Prop<::unicode::General_Category> { { 0x17D4, 0x17D6 }, ::unicode::General_Category::Other_Punctuation }, // [3] KHMER SIGN KHAN..KHMER SIGN CAMNUC PII KUUH
    Prop<::unicode::General_Category> { { 0x17D7, 0x17D7 }, ::unicode::General_Category::Modifier_Letter }, // KHMER SIGN LEK TOO
    Prop<::unicode::General_Category> { { 0x17D8, 0x17DA }, ::unicode::General_Category::Other_Punctuation }, // [3] KHMER SIGN BEYYAL..KHMER SIGN KOOMUUT
    Prop<::unicode::General_Category> { { 0x17DB, 0x17DB }, ::unicode::General_Category::Currency_Symbol }, // KHMER CURRENCY SYMBOL RIEL
    Prop<::unicode::General_Category> { { 0x17DC, 0x17DC }, ::unicode::General_Category::Other_Letter }, // KHMER SIGN AVAKRAHASANYA
    Prop<::unicode::General_Category> { { 0x17DD, 0x17DD }, ::unicode::General_Category::Nonspacing_Mark }, // KHMER SIGN ATTHACAN
    Prop<::unicode::General_Category> { { 0x17DE, 0x17DF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-17DE>..<reserved-17DF>
    Prop<::unicode::General_Category> { { 0x17E0, 0x17E9 }, ::unicode::General_Category::Decimal_Number }, // [10] KHMER DIGIT ZERO..KHMER DIGIT NINE
    Prop<::unicode::General_Category> { { 0x17EA, 0x17EF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-17EA>..<reserved-17EF>
    Prop<::unicode::General_Category> { { 0x17F0, 0x17F9 }, ::unicode::General_Category::Other_Number }, // [10] KHMER SYMBOL LEK ATTAK SON..KHMER SYMBOL LEK ATTAK PRAM-BUON
    Prop<::unicode::General_Category> { { 0x17FA, 0x17FF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-17FA>..<reserved-17FF>
    Prop<::unicode::General_Category> { { 0x1800, 0x1805 }, ::unicode::General_Category::Other_Punctuation }, // [6] MONGOLIAN BIRGA..MONGOLIAN FOUR DOTS
    Prop<::unicode::General_Category> { { 0x1806, 0x1806 }, ::unicode::General_Category::Dash_Punctuation }, // MONGOLIAN TODO SOFT HYPHEN
    Prop<::unicode::General_Category> { { 0x1807, 0x180A }, ::unicode::General_Category::Other_Punctuation }, // [4] MONGOLIAN SIBE SYLLABLE BOUNDARY MARKER..MONGOLIAN NIRUGU
    Prop<::unicode::General_Category> { { 0x180B, 0x180D }, ::unicode::General_Category::Nonspacing_Mark }, // [3] MONGOLIAN FREE VARIATION SELECTOR ONE..MONGOLIAN FREE VARIATION SELECTOR THREE
    Prop<::unicode::General_Category> { { 0x180E, 0x180E }, ::unicode::General_Category::Format }, // MONGOLIAN VOWEL SEPARATOR
    Prop<::unicode::General_Category> { { 0x180F, 0x180F }, ::unicode::General_Category::Nonspacing_Mark }, // MONGOLIAN FREE VARIATION SELECTOR FOUR
    Prop<::unicode::General_Category> { { 0x1810, 0x1819 }, ::unicode::General_Category::Decimal_Number }, // [10] MONGOLIAN DIGIT ZERO..MONGOLIAN DIGIT NINE
    Prop<::unicode::General_Category> { { 0x181A, 0x181F }, ::unicode::General_Category::Unassigned }, // [6] <reserved-181A>..<reserved-181F>
    Prop<::unicode::General_Category> { { 0x1820, 0x1842 }, ::unicode::General_Category::Other_Letter }, // [35] MONGOLIAN LETTER A..MONGOLIAN LETTER CHI
    Prop<::unicode::General_Category> { { 0x1843, 0x1843 }, ::unicode::General_Category::Modifier_Letter }, // MONGOLIAN LETTER TODO LONG VOWEL SIGN
    Prop<::unicode::General_Category> { { 0x1844, 0x1878 }, ::unicode::General_Category::Other_Letter }, // [53] MONGOLIAN LETTER TODO E..MONGOLIAN LETTER CHA WITH TWO DOTS
    Prop<::unicode::General_Category> { { 0x1879, 0x187F }, ::unicode::General_Category::Unassigned }, // [7] <reserved-1879>..<reserved-187F>
    Prop<::unicode::General_Category> { { 0x1880, 0x1884 }, ::unicode::General_Category::Other_Letter }, // [5] MONGOLIAN LETTER ALI GALI ANUSVARA ONE..MONGOLIAN LETTER ALI GALI INVERTED UBADAMA
    Prop<::unicode::General_Category> { { 0x1885, 0x1886 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MONGOLIAN LETTER ALI GALI BALUDA..MONGOLIAN LETTER ALI GALI THREE BALUDA
    Prop<::unicode::General_Category> { { 0x1887, 0x18A8 }, ::unicode::General_Category::Other_Letter }, // [34] MONGOLIAN LETTER ALI GALI A..MONGOLIAN LETTER MANCHU ALI GALI BHA
    Prop<::unicode::General_Category> { { 0x18A9, 0x18A9 }, ::unicode::General_Category::Nonspacing_Mark }, // MONGOLIAN LETTER ALI GALI DAGALGA
    Prop<::unicode::General_Category> { { 0x18AA, 0x18AA }, ::unicode::General_Category::Other_Letter }, // MONGOLIAN LETTER MANCHU ALI GALI LHA
    Prop<::unicode::General_Category> { { 0x18AB, 0x18AF }, ::unicode::General_Category::Unassigned }, // [5] <reserved-18AB>..<reserved-18AF>
    Prop<::unicode::General_Category> { { 0x18B0, 0x18F5 }, ::unicode::General_Category::Other_Letter }, // [70] CANADIAN SYLLABICS OY..CANADIAN SYLLABICS CARRIER DENTAL S
    Prop<::unicode::General_Category> { { 0x18F6, 0x18FF }, ::unicode::General_Category::Unassigned }, // [10] <reserved-18F6>..<reserved-18FF>
    Prop<::unicode::General_Category> { { 0x1900, 0x191E }, ::unicode::General_Category::Other_Letter }, // [31] LIMBU VOWEL-CARRIER LETTER..LIMBU LETTER TRA
    Prop<::unicode::General_Category> { { 0x191F, 0x191F }, ::unicode::General_Category::Unassigned }, // <reserved-191F>
    Prop<::unicode::General_Category> { { 0x1920, 0x1922 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] LIMBU VOWEL SIGN A..LIMBU VOWEL SIGN U
    Prop<::unicode::General_Category> { { 0x1923, 0x1926 }, ::unicode::General_Category::Spacing_Mark }, // [4] LIMBU VOWEL SIGN EE..LIMBU VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x1927, 0x1928 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] LIMBU VOWEL SIGN E..LIMBU VOWEL SIGN O
    Prop<::unicode::General_Category> { { 0x1929, 0x192B }, ::unicode::General_Category::Spacing_Mark }, // [3] LIMBU SUBJOINED LETTER YA..LIMBU SUBJOINED LETTER WA
    Prop<::unicode::General_Category> { { 0x192C, 0x192F }, ::unicode::General_Category::Unassigned }, // [4] <reserved-192C>..<reserved-192F>
    Prop<::unicode::General_Category> { { 0x1930, 0x1931 }, ::unicode::General_Category::Spacing_Mark }, // [2] LIMBU SMALL LETTER KA..LIMBU SMALL LETTER NGA
    Prop<::unicode::General_Category> { { 0x1932, 0x1932 }, ::unicode::General_Category::Nonspacing_Mark }, // LIMBU SMALL LETTER ANUSVARA
    Prop<::unicode::General_Category> { { 0x1933, 0x1938 }, ::unicode::General_Category::Spacing_Mark }, // [6] LIMBU SMALL LETTER TA..LIMBU SMALL LETTER LA
    Prop<::unicode::General_Category> { { 0x1939, 0x193B }, ::unicode::General_Category::Nonspacing_Mark }, // [3] LIMBU SIGN MUKPHRENG..LIMBU SIGN SA-I
    Prop<::unicode::General_Category> { { 0x193C, 0x193F }, ::unicode::General_Category::Unassigned }, // [4] <reserved-193C>..<reserved-193F>
    Prop<::unicode::General_Category> { { 0x1940, 0x1940 }, ::unicode::General_Category::Other_Symbol }, // LIMBU SIGN LOO
    Prop<::unicode::General_Category> { { 0x1941, 0x1943 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1941>..<reserved-1943>
    Prop<::unicode::General_Category> { { 0x1944, 0x1945 }, ::unicode::General_Category::Other_Punctuation }, // [2] LIMBU EXCLAMATION MARK..LIMBU QUESTION MARK
    Prop<::unicode::General_Category> { { 0x1946, 0x194F }, ::unicode::General_Category::Decimal_Number }, // [10] LIMBU DIGIT ZERO..LIMBU DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1950, 0x196D }, ::unicode::General_Category::Other_Letter }, // [30] TAI LE LETTER KA..TAI LE LETTER AI
    Prop<::unicode::General_Category> { { 0x196E, 0x196F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-196E>..<reserved-196F>
    Prop<::unicode::General_Category> { { 0x1970, 0x1974 }, ::unicode::General_Category::Other_Letter }, // [5] TAI LE LETTER TONE-2..TAI LE LETTER TONE-6
    Prop<::unicode::General_Category> { { 0x1975, 0x197F }, ::unicode::General_Category::Unassigned }, // [11] <reserved-1975>..<reserved-197F>
    Prop<::unicode::General_Category> { { 0x1980, 0x19AB }, ::unicode::General_Category::Other_Letter }, // [44] NEW TAI LUE LETTER HIGH QA..NEW TAI LUE LETTER LOW SUA
    Prop<::unicode::General_Category> { { 0x19AC, 0x19AF }, ::unicode::General_Category::Unassigned }, // [4] <reserved-19AC>..<reserved-19AF>
    Prop<::unicode::General_Category> { { 0x19B0, 0x19C9 }, ::unicode::General_Category::Other_Letter }, // [26] NEW TAI LUE VOWEL SIGN VOWEL SHORTENER..NEW TAI LUE TONE MARK-2
    Prop<::unicode::General_Category> { { 0x19CA, 0x19CF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-19CA>..<reserved-19CF>
    Prop<::unicode::General_Category> { { 0x19D0, 0x19D9 }, ::unicode::General_Category::Decimal_Number }, // [10] NEW TAI LUE DIGIT ZERO..NEW TAI LUE DIGIT NINE
    Prop<::unicode::General_Category> { { 0x19DA, 0x19DA }, ::unicode::General_Category::Other_Number }, // NEW TAI LUE THAM DIGIT ONE
    Prop<::unicode::General_Category> { { 0x19DB, 0x19DD }, ::unicode::General_Category::Unassigned }, // [3] <reserved-19DB>..<reserved-19DD>
    Prop<::unicode::General_Category> { { 0x19DE, 0x19FF }, ::unicode::General_Category::Other_Symbol }, // [34] NEW TAI LUE SIGN LAE..KHMER SYMBOL DAP-PRAM ROC
    Prop<::unicode::General_Category> { { 0x1A00, 0x1A16 }, ::unicode::General_Category::Other_Letter }, // [23] BUGINESE LETTER KA..BUGINESE LETTER HA
    Prop<::unicode::General_Category> { { 0x1A17, 0x1A18 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] BUGINESE VOWEL SIGN I..BUGINESE VOWEL SIGN U
    Prop<::unicode::General_Category> { { 0x1A19, 0x1A1A }, ::unicode::General_Category::Spacing_Mark }, // [2] BUGINESE VOWEL SIGN E..BUGINESE VOWEL SIGN O
    Prop<::unicode::General_Category> { { 0x1A1B, 0x1A1B }, ::unicode::General_Category::Nonspacing_Mark }, // BUGINESE VOWEL SIGN AE
    Prop<::unicode::General_Category> { { 0x1A1C, 0x1A1D }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1A1C>..<reserved-1A1D>
    Prop<::unicode::General_Category> { { 0x1A1E, 0x1A1F }, ::unicode::General_Category::Other_Punctuation }, // [2] BUGINESE PALLAWA..BUGINESE END OF SECTION
    Prop<::unicode::General_Category> { { 0x1A20, 0x1A54 }, ::unicode::General_Category::Other_Letter }, // [53] TAI THAM LETTER HIGH KA..TAI THAM LETTER GREAT SA
    Prop<::unicode::General_Category> { { 0x1A55, 0x1A55 }, ::unicode::General_Category::Spacing_Mark }, // TAI THAM CONSONANT SIGN MEDIAL RA
    Prop<::unicode::General_Category> { { 0x1A56, 0x1A56 }, ::unicode::General_Category::Nonspacing_Mark }, // TAI THAM CONSONANT SIGN MEDIAL LA
    Prop<::unicode::General_Category> { { 0x1A57, 0x1A57 }, ::unicode::General_Category::Spacing_Mark }, // TAI THAM CONSONANT SIGN LA TANG LAI
    Prop<::unicode::General_Category> { { 0x1A58, 0x1A5E }, ::unicode::General_Category::Nonspacing_Mark }, // [7] TAI THAM SIGN MAI KANG LAI..TAI THAM CONSONANT SIGN SA
    Prop<::unicode::General_Category> { { 0x1A5F, 0x1A5F }, ::unicode::General_Category::Unassigned }, // <reserved-1A5F>
    Prop<::unicode::General_Category> { { 0x1A60, 0x1A60 }, ::unicode::General_Category::Nonspacing_Mark }, // TAI THAM SIGN SAKOT
    Prop<::unicode::General_Category> { { 0x1A61, 0x1A61 }, ::unicode::General_Category::Spacing_Mark }, // TAI THAM VOWEL SIGN A
    Prop<::unicode::General_Category> { { 0x1A62, 0x1A62 }, ::unicode::General_Category::Nonspacing_Mark }, // TAI THAM VOWEL SIGN MAI SAT
    Prop<::unicode::General_Category> { { 0x1A63, 0x1A64 }, ::unicode::General_Category::Spacing_Mark }, // [2] TAI THAM VOWEL SIGN AA..TAI THAM VOWEL SIGN TALL AA
    Prop<::unicode::General_Category> { { 0x1A65, 0x1A6C }, ::unicode::General_Category::Nonspacing_Mark }, // [8] TAI THAM VOWEL SIGN I..TAI THAM VOWEL SIGN OA BELOW
    Prop<::unicode::General_Category> { { 0x1A6D, 0x1A72 }, ::unicode::General_Category::Spacing_Mark }, // [6] TAI THAM VOWEL SIGN OY..TAI THAM VOWEL SIGN THAM AI
    Prop<::unicode::General_Category> { { 0x1A73, 0x1A7C }, ::unicode::General_Category::Nonspacing_Mark }, // [10] TAI THAM VOWEL SIGN OA ABOVE..TAI THAM SIGN KHUEN-LUE KARAN
    Prop<::unicode::General_Category> { { 0x1A7D, 0x1A7E }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1A7D>..<reserved-1A7E>
    Prop<::unicode::General_Category> { { 0x1A7F, 0x1A7F }, ::unicode::General_Category::Nonspacing_Mark }, // TAI THAM COMBINING CRYPTOGRAMMIC DOT
    Prop<::unicode::General_Category> { { 0x1A80, 0x1A89 }, ::unicode::General_Category::Decimal_Number }, // [10] TAI THAM HORA DIGIT ZERO..TAI THAM HORA DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1A8A, 0x1A8F }, ::unicode::General_Category::Unassigned }, // [6] <reserved-1A8A>..<reserved-1A8F>
    Prop<::unicode::General_Category> { { 0x1A90, 0x1A99 }, ::unicode::General_Category::Decimal_Number }, // [10] TAI THAM THAM DIGIT ZERO..TAI THAM THAM DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1A9A, 0x1A9F }, ::unicode::General_Category::Unassigned }, // [6] <reserved-1A9A>..<reserved-1A9F>
    Prop<::unicode::General_Category> { { 0x1AA0, 0x1AA6 }, ::unicode::General_Category::Other_Punctuation }, // [7] TAI THAM SIGN WIANG..TAI THAM SIGN REVERSED ROTATED RANA
    Prop<::unicode::General_Category> { { 0x1AA7, 0x1AA7 }, ::unicode::General_Category::Modifier_Letter }, // TAI THAM SIGN MAI YAMOK
    Prop<::unicode::General_Category> { { 0x1AA8, 0x1AAD }, ::unicode::General_Category::Other_Punctuation }, // [6] TAI THAM SIGN KAAN..TAI THAM SIGN CAANG
    Prop<::unicode::General_Category> { { 0x1AAE, 0x1AAF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1AAE>..<reserved-1AAF>
    Prop<::unicode::General_Category> { { 0x1AB0, 0x1ABD }, ::unicode::General_Category::Nonspacing_Mark }, // [14] COMBINING DOUBLED CIRCUMFLEX ACCENT..COMBINING PARENTHESES BELOW
    Prop<::unicode::General_Category> { { 0x1ABE, 0x1ABE }, ::unicode::General_Category::Enclosing_Mark }, // COMBINING PARENTHESES OVERLAY
    Prop<::unicode::General_Category> { { 0x1ABF, 0x1ADD }, ::unicode::General_Category::Nonspacing_Mark }, // [31] COMBINING LATIN SMALL LETTER W BELOW..COMBINING DOT-AND-RING BELOW
    Prop<::unicode::General_Category> { { 0x1ADE, 0x1ADF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1ADE>..<reserved-1ADF>
    Prop<::unicode::General_Category> { { 0x1AE0, 0x1AEB }, ::unicode::General_Category::Nonspacing_Mark }, // [12] COMBINING LEFT TACK ABOVE..COMBINING DOUBLE RIGHTWARDS ARROW ABOVE
    Prop<::unicode::General_Category> { { 0x1AEC, 0x1AFF }, ::unicode::General_Category::Unassigned }, // [20] <reserved-1AEC>..<reserved-1AFF>
    Prop<::unicode::General_Category> { { 0x1B00, 0x1B03 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] BALINESE SIGN ULU RICEM..BALINESE SIGN SURANG
    Prop<::unicode::General_Category> { { 0x1B04, 0x1B04 }, ::unicode::General_Category::Spacing_Mark }, // BALINESE SIGN BISAH
    Prop<::unicode::General_Category> { { 0x1B05, 0x1B33 }, ::unicode::General_Category::Other_Letter }, // [47] BALINESE LETTER AKARA..BALINESE LETTER HA
    Prop<::unicode::General_Category> { { 0x1B34, 0x1B34 }, ::unicode::General_Category::Nonspacing_Mark }, // BALINESE SIGN REREKAN
    Prop<::unicode::General_Category> { { 0x1B35, 0x1B35 }, ::unicode::General_Category::Spacing_Mark }, // BALINESE VOWEL SIGN TEDUNG
    Prop<::unicode::General_Category> { { 0x1B36, 0x1B3A }, ::unicode::General_Category::Nonspacing_Mark }, // [5] BALINESE VOWEL SIGN ULU..BALINESE VOWEL SIGN RA REPA
    Prop<::unicode::General_Category> { { 0x1B3B, 0x1B3B }, ::unicode::General_Category::Spacing_Mark }, // BALINESE VOWEL SIGN RA REPA TEDUNG
    Prop<::unicode::General_Category> { { 0x1B3C, 0x1B3C }, ::unicode::General_Category::Nonspacing_Mark }, // BALINESE VOWEL SIGN LA LENGA
    Prop<::unicode::General_Category> { { 0x1B3D, 0x1B41 }, ::unicode::General_Category::Spacing_Mark }, // [5] BALINESE VOWEL SIGN LA LENGA TEDUNG..BALINESE VOWEL SIGN TALING REPA TEDUNG
    Prop<::unicode::General_Category> { { 0x1B42, 0x1B42 }, ::unicode::General_Category::Nonspacing_Mark }, // BALINESE VOWEL SIGN PEPET
    Prop<::unicode::General_Category> { { 0x1B43, 0x1B44 }, ::unicode::General_Category::Spacing_Mark }, // [2] BALINESE VOWEL SIGN PEPET TEDUNG..BALINESE ADEG ADEG
    Prop<::unicode::General_Category> { { 0x1B45, 0x1B4C }, ::unicode::General_Category::Other_Letter }, // [8] BALINESE LETTER KAF SASAK..BALINESE LETTER ARCHAIC JNYA
    Prop<::unicode::General_Category> { { 0x1B4D, 0x1B4D }, ::unicode::General_Category::Unassigned }, // <reserved-1B4D>
    Prop<::unicode::General_Category> { { 0x1B4E, 0x1B4F }, ::unicode::General_Category::Other_Punctuation }, // [2] BALINESE INVERTED CARIK SIKI..BALINESE INVERTED CARIK PAREREN
    Prop<::unicode::General_Category> { { 0x1B50, 0x1B59 }, ::unicode::General_Category::Decimal_Number }, // [10] BALINESE DIGIT ZERO..BALINESE DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1B5A, 0x1B60 }, ::unicode::General_Category::Other_Punctuation }, // [7] BALINESE PANTI..BALINESE PAMENENG
    Prop<::unicode::General_Category> { { 0x1B61, 0x1B6A }, ::unicode::General_Category::Other_Symbol }, // [10] BALINESE MUSICAL SYMBOL DONG..BALINESE MUSICAL SYMBOL DANG GEDE
    Prop<::unicode::General_Category> { { 0x1B6B, 0x1B73 }, ::unicode::General_Category::Nonspacing_Mark }, // [9] BALINESE MUSICAL SYMBOL COMBINING TEGEH..BALINESE MUSICAL SYMBOL COMBINING GONG
    Prop<::unicode::General_Category> { { 0x1B74, 0x1B7C }, ::unicode::General_Category::Other_Symbol }, // [9] BALINESE MUSICAL SYMBOL RIGHT-HAND OPEN DUG..BALINESE MUSICAL SYMBOL LEFT-HAND OPEN PING
    Prop<::unicode::General_Category> { { 0x1B7D, 0x1B7F }, ::unicode::General_Category::Other_Punctuation }, // [3] BALINESE PANTI LANTANG..BALINESE PANTI BAWAK
    Prop<::unicode::General_Category> { { 0x1B80, 0x1B81 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] SUNDANESE SIGN PANYECEK..SUNDANESE SIGN PANGLAYAR
    Prop<::unicode::General_Category> { { 0x1B82, 0x1B82 }, ::unicode::General_Category::Spacing_Mark }, // SUNDANESE SIGN PANGWISAD
    Prop<::unicode::General_Category> { { 0x1B83, 0x1BA0 }, ::unicode::General_Category::Other_Letter }, // [30] SUNDANESE LETTER A..SUNDANESE LETTER HA
    Prop<::unicode::General_Category> { { 0x1BA1, 0x1BA1 }, ::unicode::General_Category::Spacing_Mark }, // SUNDANESE CONSONANT SIGN PAMINGKAL
    Prop<::unicode::General_Category> { { 0x1BA2, 0x1BA5 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] SUNDANESE CONSONANT SIGN PANYAKRA..SUNDANESE VOWEL SIGN PANYUKU
    Prop<::unicode::General_Category> { { 0x1BA6, 0x1BA7 }, ::unicode::General_Category::Spacing_Mark }, // [2] SUNDANESE VOWEL SIGN PANAELAENG..SUNDANESE VOWEL SIGN PANOLONG
    Prop<::unicode::General_Category> { { 0x1BA8, 0x1BA9 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] SUNDANESE VOWEL SIGN PAMEPET..SUNDANESE VOWEL SIGN PANEULEUNG
    Prop<::unicode::General_Category> { { 0x1BAA, 0x1BAA }, ::unicode::General_Category::Spacing_Mark }, // SUNDANESE SIGN PAMAAEH
    Prop<::unicode::General_Category> { { 0x1BAB, 0x1BAD }, ::unicode::General_Category::Nonspacing_Mark }, // [3] SUNDANESE SIGN VIRAMA..SUNDANESE CONSONANT SIGN PASANGAN WA
    Prop<::unicode::General_Category> { { 0x1BAE, 0x1BAF }, ::unicode::General_Category::Other_Letter }, // [2] SUNDANESE LETTER KHA..SUNDANESE LETTER SYA
    Prop<::unicode::General_Category> { { 0x1BB0, 0x1BB9 }, ::unicode::General_Category::Decimal_Number }, // [10] SUNDANESE DIGIT ZERO..SUNDANESE DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1BBA, 0x1BE5 }, ::unicode::General_Category::Other_Letter }, // [44] SUNDANESE AVAGRAHA..BATAK LETTER U
    Prop<::unicode::General_Category> { { 0x1BE6, 0x1BE6 }, ::unicode::General_Category::Nonspacing_Mark }, // BATAK SIGN TOMPI
    Prop<::unicode::General_Category> { { 0x1BE7, 0x1BE7 }, ::unicode::General_Category::Spacing_Mark }, // BATAK VOWEL SIGN E
    Prop<::unicode::General_Category> { { 0x1BE8, 0x1BE9 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] BATAK VOWEL SIGN PAKPAK E..BATAK VOWEL SIGN EE
    Prop<::unicode::General_Category> { { 0x1BEA, 0x1BEC }, ::unicode::General_Category::Spacing_Mark }, // [3] BATAK VOWEL SIGN I..BATAK VOWEL SIGN O
    Prop<::unicode::General_Category> { { 0x1BED, 0x1BED }, ::unicode::General_Category::Nonspacing_Mark }, // BATAK VOWEL SIGN KARO O
    Prop<::unicode::General_Category> { { 0x1BEE, 0x1BEE }, ::unicode::General_Category::Spacing_Mark }, // BATAK VOWEL SIGN U
    Prop<::unicode::General_Category> { { 0x1BEF, 0x1BF1 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] BATAK VOWEL SIGN U FOR SIMALUNGUN SA..BATAK CONSONANT SIGN H
    Prop<::unicode::General_Category> { { 0x1BF2, 0x1BF3 }, ::unicode::General_Category::Spacing_Mark }, // [2] BATAK PANGOLAT..BATAK PANONGONAN
    Prop<::unicode::General_Category> { { 0x1BF4, 0x1BFB }, ::unicode::General_Category::Unassigned }, // [8] <reserved-1BF4>..<reserved-1BFB>
    Prop<::unicode::General_Category> { { 0x1BFC, 0x1BFF }, ::unicode::General_Category::Other_Punctuation }, // [4] BATAK SYMBOL BINDU NA METEK..BATAK SYMBOL BINDU PANGOLAT
    Prop<::unicode::General_Category> { { 0x1C00, 0x1C23 }, ::unicode::General_Category::Other_Letter }, // [36] LEPCHA LETTER KA..LEPCHA LETTER A
    Prop<::unicode::General_Category> { { 0x1C24, 0x1C2B }, ::unicode::General_Category::Spacing_Mark }, // [8] LEPCHA SUBJOINED LETTER YA..LEPCHA VOWEL SIGN UU
    Prop<::unicode::General_Category> { { 0x1C2C, 0x1C33 }, ::unicode::General_Category::Nonspacing_Mark }, // [8] LEPCHA VOWEL SIGN E..LEPCHA CONSONANT SIGN T
    Prop<::unicode::General_Category> { { 0x1C34, 0x1C35 }, ::unicode::General_Category::Spacing_Mark }, // [2] LEPCHA CONSONANT SIGN NYIN-DO..LEPCHA CONSONANT SIGN KANG
    Prop<::unicode::General_Category> { { 0x1C36, 0x1C37 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] LEPCHA SIGN RAN..LEPCHA SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x1C38, 0x1C3A }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1C38>..<reserved-1C3A>
    Prop<::unicode::General_Category> { { 0x1C3B, 0x1C3F }, ::unicode::General_Category::Other_Punctuation }, // [5] LEPCHA PUNCTUATION TA-ROL..LEPCHA PUNCTUATION TSHOOK
    Prop<::unicode::General_Category> { { 0x1C40, 0x1C49 }, ::unicode::General_Category::Decimal_Number }, // [10] LEPCHA DIGIT ZERO..LEPCHA DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1C4A, 0x1C4C }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1C4A>..<reserved-1C4C>
    Prop<::unicode::General_Category> { { 0x1C4D, 0x1C4F }, ::unicode::General_Category::Other_Letter }, // [3] LEPCHA LETTER TTA..LEPCHA LETTER DDA
    Prop<::unicode::General_Category> { { 0x1C50, 0x1C59 }, ::unicode::General_Category::Decimal_Number }, // [10] OL CHIKI DIGIT ZERO..OL CHIKI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1C5A, 0x1C77 }, ::unicode::General_Category::Other_Letter }, // [30] OL CHIKI LETTER LA..OL CHIKI LETTER OH
    Prop<::unicode::General_Category> { { 0x1C78, 0x1C7D }, ::unicode::General_Category::Modifier_Letter }, // [6] OL CHIKI MU TTUDDAG..OL CHIKI AHAD
    Prop<::unicode::General_Category> { { 0x1C7E, 0x1C7F }, ::unicode::General_Category::Other_Punctuation }, // [2] OL CHIKI PUNCTUATION MUCAAD..OL CHIKI PUNCTUATION DOUBLE MUCAAD
    Prop<::unicode::General_Category> { { 0x1C80, 0x1C88 }, ::unicode::General_Category::Lowercase_Letter }, // [9] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER UNBLENDED UK
    Prop<::unicode::General_Category> { { 0x1C89, 0x1C89 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER TJE
    Prop<::unicode::General_Category> { { 0x1C8A, 0x1C8A }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER TJE
    Prop<::unicode::General_Category> { { 0x1C8B, 0x1C8F }, ::unicode::General_Category::Unassigned }, // [5] <reserved-1C8B>..<reserved-1C8F>
    Prop<::unicode::General_Category> { { 0x1C90, 0x1CBA }, ::unicode::General_Category::Uppercase_Letter }, // [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Prop<::unicode::General_Category> { { 0x1CBB, 0x1CBC }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1CBB>..<reserved-1CBC>
    Prop<::unicode::General_Category> { { 0x1CBD, 0x1CBF }, ::unicode::General_Category::Uppercase_Letter }, // [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Prop<::unicode::General_Category> { { 0x1CC0, 0x1CC7 }, ::unicode::General_Category::Other_Punctuation }, // [8] SUNDANESE PUNCTUATION BINDU SURYA..SUNDANESE PUNCTUATION BINDU BA SATANGA
    Prop<::unicode::General_Category> { { 0x1CC8, 0x1CCF }, ::unicode::General_Category::Unassigned }, // [8] <reserved-1CC8>..<reserved-1CCF>
    Prop<::unicode::General_Category> { { 0x1CD0, 0x1CD2 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] VEDIC TONE KARSHANA..VEDIC TONE PRENKHA
    Prop<::unicode::General_Category> { { 0x1CD3, 0x1CD3 }, ::unicode::General_Category::Other_Punctuation }, // VEDIC SIGN NIHSHVASA
    Prop<::unicode::General_Category> { { 0x1CD4, 0x1CE0 }, ::unicode::General_Category::Nonspacing_Mark }, // [13] VEDIC SIGN YAJURVEDIC MIDLINE SVARITA..VEDIC TONE RIGVEDIC KASHMIRI INDEPENDENT SVARITA
    Prop<::unicode::General_Category> { { 0x1CE1, 0x1CE1 }, ::unicode::General_Category::Spacing_Mark }, // VEDIC TONE ATHARVAVEDIC INDEPENDENT SVARITA
    Prop<::unicode::General_Category> { { 0x1CE2, 0x1CE8 }, ::unicode::General_Category::Nonspacing_Mark }, // [7] VEDIC SIGN VISARGA SVARITA..VEDIC SIGN VISARGA ANUDATTA WITH TAIL
    Prop<::unicode::General_Category> { { 0x1CE9, 0x1CEC }, ::unicode::General_Category::Other_Letter }, // [4] VEDIC SIGN ANUSVARA ANTARGOMUKHA..VEDIC SIGN ANUSVARA VAMAGOMUKHA WITH TAIL
    Prop<::unicode::General_Category> { { 0x1CED, 0x1CED }, ::unicode::General_Category::Nonspacing_Mark }, // VEDIC SIGN TIRYAK
    Prop<::unicode::General_Category> { { 0x1CEE, 0x1CF3 }, ::unicode::General_Category::Other_Letter }, // [6] VEDIC SIGN HEXIFORM LONG ANUSVARA..VEDIC SIGN ROTATED ARDHAVISARGA
    Prop<::unicode::General_Category> { { 0x1CF4, 0x1CF4 }, ::unicode::General_Category::Nonspacing_Mark }, // VEDIC TONE CANDRA ABOVE
    Prop<::unicode::General_Category> { { 0x1CF5, 0x1CF6 }, ::unicode::General_Category::Other_Letter }, // [2] VEDIC SIGN JIHVAMULIYA..VEDIC SIGN UPADHMANIYA
    Prop<::unicode::General_Category> { { 0x1CF7, 0x1CF7 }, ::unicode::General_Category::Spacing_Mark }, // VEDIC SIGN ATIKRAMA
    Prop<::unicode::General_Category> { { 0x1CF8, 0x1CF9 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] VEDIC TONE RING ABOVE..VEDIC TONE DOUBLE RING ABOVE
    Prop<::unicode::General_Category> { { 0x1CFA, 0x1CFA }, ::unicode::General_Category::Other_Letter }, // VEDIC SIGN DOUBLE ANUSVARA ANTARGOMUKHA
    Prop<::unicode::General_Category> { { 0x1CFB, 0x1CFF }, ::unicode::General_Category::Unassigned }, // [5] <reserved-1CFB>..<reserved-1CFF>
    Prop<::unicode::General_Category> { { 0x1D00, 0x1D2B }, ::unicode::General_Category::Lowercase_Letter }, // [44] LATIN LETTER SMALL CAPITAL A..CYRILLIC LETTER SMALL CAPITAL EL
    Prop<::unicode::General_Category> { { 0x1D2C, 0x1D6A }, ::unicode::General_Category::Modifier_Letter }, // [63] MODIFIER LETTER CAPITAL A..GREEK SUBSCRIPT SMALL LETTER CHI
    Prop<::unicode::General_Category> { { 0x1D6B, 0x1D77 }, ::unicode::General_Category::Lowercase_Letter }, // [13] LATIN SMALL LETTER UE..LATIN SMALL LETTER TURNED G
    Prop<::unicode::General_Category> { { 0x1D78, 0x1D78 }, ::unicode::General_Category::Modifier_Letter }, // MODIFIER LETTER CYRILLIC EN
    Prop<::unicode::General_Category> { { 0x1D79, 0x1D9A }, ::unicode::General_Category::Lowercase_Letter }, // [34] LATIN SMALL LETTER INSULAR G..LATIN SMALL LETTER EZH WITH RETROFLEX HOOK
    Prop<::unicode::General_Category> { { 0x1D9B, 0x1DBF }, ::unicode::General_Category::Modifier_Letter }, // [37] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL THETA
    Prop<::unicode::General_Category> { { 0x1DC0, 0x1DFF }, ::unicode::General_Category::Nonspacing_Mark }, // [64] COMBINING DOTTED GRAVE ACCENT..COMBINING RIGHT ARROWHEAD AND DOWN ARROWHEAD BELOW
    Prop<::unicode::General_Category> { { 0x1E00, 0x1E00 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH RING BELOW
    Prop<::unicode::General_Category> { { 0x1E01, 0x1E01 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH RING BELOW
    Prop<::unicode::General_Category> { { 0x1E02, 0x1E02 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER B WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E03, 0x1E03 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER B WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E04, 0x1E04 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER B WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E05, 0x1E05 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER B WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E06, 0x1E06 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER B WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E07, 0x1E07 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER B WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E08, 0x1E08 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER C WITH CEDILLA AND ACUTE
    Prop<::unicode::General_Category> { { 0x1E09, 0x1E09 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER C WITH CEDILLA AND ACUTE
    Prop<::unicode::General_Category> { { 0x1E0A, 0x1E0A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER D WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E0B, 0x1E0B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER D WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E0C, 0x1E0C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER D WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E0D, 0x1E0D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER D WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E0E, 0x1E0E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER D WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E0F, 0x1E0F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER D WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E10, 0x1E10 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER D WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x1E11, 0x1E11 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER D WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x1E12, 0x1E12 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER D WITH CIRCUMFLEX BELOW
    Prop<::unicode::General_Category> { { 0x1E13, 0x1E13 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER D WITH CIRCUMFLEX BELOW
    Prop<::unicode::General_Category> { { 0x1E14, 0x1E14 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH MACRON AND GRAVE
    Prop<::unicode::General_Category> { { 0x1E15, 0x1E15 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH MACRON AND GRAVE
    Prop<::unicode::General_Category> { { 0x1E16, 0x1E16 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH MACRON AND ACUTE
    Prop<::unicode::General_Category> { { 0x1E17, 0x1E17 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH MACRON AND ACUTE
    Prop<::unicode::General_Category> { { 0x1E18, 0x1E18 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH CIRCUMFLEX BELOW
    Prop<::unicode::General_Category> { { 0x1E19, 0x1E19 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH CIRCUMFLEX BELOW
    Prop<::unicode::General_Category> { { 0x1E1A, 0x1E1A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH TILDE BELOW
    Prop<::unicode::General_Category> { { 0x1E1B, 0x1E1B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH TILDE BELOW
    Prop<::unicode::General_Category> { { 0x1E1C, 0x1E1C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH CEDILLA AND BREVE
    Prop<::unicode::General_Category> { { 0x1E1D, 0x1E1D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH CEDILLA AND BREVE
    Prop<::unicode::General_Category> { { 0x1E1E, 0x1E1E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER F WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E1F, 0x1E1F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER F WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E20, 0x1E20 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER G WITH MACRON
    Prop<::unicode::General_Category> { { 0x1E21, 0x1E21 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER G WITH MACRON
    Prop<::unicode::General_Category> { { 0x1E22, 0x1E22 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER H WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E23, 0x1E23 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER H WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E24, 0x1E24 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER H WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E25, 0x1E25 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER H WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E26, 0x1E26 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER H WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x1E27, 0x1E27 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER H WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x1E28, 0x1E28 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER H WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x1E29, 0x1E29 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER H WITH CEDILLA
    Prop<::unicode::General_Category> { { 0x1E2A, 0x1E2A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER H WITH BREVE BELOW
    Prop<::unicode::General_Category> { { 0x1E2B, 0x1E2B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER H WITH BREVE BELOW
    Prop<::unicode::General_Category> { { 0x1E2C, 0x1E2C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER I WITH TILDE BELOW
    Prop<::unicode::General_Category> { { 0x1E2D, 0x1E2D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER I WITH TILDE BELOW
    Prop<::unicode::General_Category> { { 0x1E2E, 0x1E2E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER I WITH DIAERESIS AND ACUTE
    Prop<::unicode::General_Category> { { 0x1E2F, 0x1E2F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER I WITH DIAERESIS AND ACUTE
    Prop<::unicode::General_Category> { { 0x1E30, 0x1E30 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER K WITH ACUTE
    Prop<::unicode::General_Category> { { 0x1E31, 0x1E31 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER K WITH ACUTE
    Prop<::unicode::General_Category> { { 0x1E32, 0x1E32 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER K WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E33, 0x1E33 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER K WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E34, 0x1E34 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER K WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E35, 0x1E35 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER K WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E36, 0x1E36 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER L WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E37, 0x1E37 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER L WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E38, 0x1E38 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER L WITH DOT BELOW AND MACRON
    Prop<::unicode::General_Category> { { 0x1E39, 0x1E39 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER L WITH DOT BELOW AND MACRON
    Prop<::unicode::General_Category> { { 0x1E3A, 0x1E3A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER L WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E3B, 0x1E3B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER L WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E3C, 0x1E3C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER L WITH CIRCUMFLEX BELOW
    Prop<::unicode::General_Category> { { 0x1E3D, 0x1E3D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER L WITH CIRCUMFLEX BELOW
    Prop<::unicode::General_Category> { { 0x1E3E, 0x1E3E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER M WITH ACUTE
    Prop<::unicode::General_Category> { { 0x1E3F, 0x1E3F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER M WITH ACUTE
    Prop<::unicode::General_Category> { { 0x1E40, 0x1E40 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER M WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E41, 0x1E41 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER M WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E42, 0x1E42 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER M WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E43, 0x1E43 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER M WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E44, 0x1E44 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER N WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E45, 0x1E45 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER N WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E46, 0x1E46 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER N WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E47, 0x1E47 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER N WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E48, 0x1E48 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER N WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E49, 0x1E49 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER N WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E4A, 0x1E4A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER N WITH CIRCUMFLEX BELOW
    Prop<::unicode::General_Category> { { 0x1E4B, 0x1E4B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER N WITH CIRCUMFLEX BELOW
    Prop<::unicode::General_Category> { { 0x1E4C, 0x1E4C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH TILDE AND ACUTE
    Prop<::unicode::General_Category> { { 0x1E4D, 0x1E4D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH TILDE AND ACUTE
    Prop<::unicode::General_Category> { { 0x1E4E, 0x1E4E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH TILDE AND DIAERESIS
    Prop<::unicode::General_Category> { { 0x1E4F, 0x1E4F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH TILDE AND DIAERESIS
    Prop<::unicode::General_Category> { { 0x1E50, 0x1E50 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH MACRON AND GRAVE
    Prop<::unicode::General_Category> { { 0x1E51, 0x1E51 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH MACRON AND GRAVE
    Prop<::unicode::General_Category> { { 0x1E52, 0x1E52 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH MACRON AND ACUTE
    Prop<::unicode::General_Category> { { 0x1E53, 0x1E53 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH MACRON AND ACUTE
    Prop<::unicode::General_Category> { { 0x1E54, 0x1E54 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER P WITH ACUTE
    Prop<::unicode::General_Category> { { 0x1E55, 0x1E55 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER P WITH ACUTE
    Prop<::unicode::General_Category> { { 0x1E56, 0x1E56 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER P WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E57, 0x1E57 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER P WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E58, 0x1E58 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER R WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E59, 0x1E59 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER R WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E5A, 0x1E5A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER R WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E5B, 0x1E5B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER R WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E5C, 0x1E5C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER R WITH DOT BELOW AND MACRON
    Prop<::unicode::General_Category> { { 0x1E5D, 0x1E5D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER R WITH DOT BELOW AND MACRON
    Prop<::unicode::General_Category> { { 0x1E5E, 0x1E5E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER R WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E5F, 0x1E5F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER R WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E60, 0x1E60 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER S WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E61, 0x1E61 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E62, 0x1E62 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER S WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E63, 0x1E63 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E64, 0x1E64 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER S WITH ACUTE AND DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E65, 0x1E65 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH ACUTE AND DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E66, 0x1E66 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER S WITH CARON AND DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E67, 0x1E67 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH CARON AND DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E68, 0x1E68 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER S WITH DOT BELOW AND DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E69, 0x1E69 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH DOT BELOW AND DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E6A, 0x1E6A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER T WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E6B, 0x1E6B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER T WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E6C, 0x1E6C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER T WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E6D, 0x1E6D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER T WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E6E, 0x1E6E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER T WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E6F, 0x1E6F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER T WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E70, 0x1E70 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER T WITH CIRCUMFLEX BELOW
    Prop<::unicode::General_Category> { { 0x1E71, 0x1E71 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER T WITH CIRCUMFLEX BELOW
    Prop<::unicode::General_Category> { { 0x1E72, 0x1E72 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH DIAERESIS BELOW
    Prop<::unicode::General_Category> { { 0x1E73, 0x1E73 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH DIAERESIS BELOW
    Prop<::unicode::General_Category> { { 0x1E74, 0x1E74 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH TILDE BELOW
    Prop<::unicode::General_Category> { { 0x1E75, 0x1E75 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH TILDE BELOW
    Prop<::unicode::General_Category> { { 0x1E76, 0x1E76 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH CIRCUMFLEX BELOW
    Prop<::unicode::General_Category> { { 0x1E77, 0x1E77 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH CIRCUMFLEX BELOW
    Prop<::unicode::General_Category> { { 0x1E78, 0x1E78 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH TILDE AND ACUTE
    Prop<::unicode::General_Category> { { 0x1E79, 0x1E79 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH TILDE AND ACUTE
    Prop<::unicode::General_Category> { { 0x1E7A, 0x1E7A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH MACRON AND DIAERESIS
    Prop<::unicode::General_Category> { { 0x1E7B, 0x1E7B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH MACRON AND DIAERESIS
    Prop<::unicode::General_Category> { { 0x1E7C, 0x1E7C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER V WITH TILDE
    Prop<::unicode::General_Category> { { 0x1E7D, 0x1E7D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER V WITH TILDE
    Prop<::unicode::General_Category> { { 0x1E7E, 0x1E7E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER V WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E7F, 0x1E7F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER V WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E80, 0x1E80 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER W WITH GRAVE
    Prop<::unicode::General_Category> { { 0x1E81, 0x1E81 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER W WITH GRAVE
    Prop<::unicode::General_Category> { { 0x1E82, 0x1E82 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER W WITH ACUTE
    Prop<::unicode::General_Category> { { 0x1E83, 0x1E83 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER W WITH ACUTE
    Prop<::unicode::General_Category> { { 0x1E84, 0x1E84 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER W WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x1E85, 0x1E85 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER W WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x1E86, 0x1E86 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER W WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E87, 0x1E87 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER W WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E88, 0x1E88 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER W WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E89, 0x1E89 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER W WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E8A, 0x1E8A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER X WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E8B, 0x1E8B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER X WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E8C, 0x1E8C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER X WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x1E8D, 0x1E8D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER X WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x1E8E, 0x1E8E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Y WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E8F, 0x1E8F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Y WITH DOT ABOVE
    Prop<::unicode::General_Category> { { 0x1E90, 0x1E90 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Z WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x1E91, 0x1E91 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Z WITH CIRCUMFLEX
    Prop<::unicode::General_Category> { { 0x1E92, 0x1E92 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Z WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E93, 0x1E93 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Z WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1E94, 0x1E94 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Z WITH LINE BELOW
    Prop<::unicode::General_Category> { { 0x1E95, 0x1E9D }, ::unicode::General_Category::Lowercase_Letter }, // [9] LATIN SMALL LETTER Z WITH LINE BELOW..LATIN SMALL LETTER LONG S WITH HIGH STROKE
    Prop<::unicode::General_Category> { { 0x1E9E, 0x1E9E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER SHARP S
    Prop<::unicode::General_Category> { { 0x1E9F, 0x1E9F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER DELTA
    Prop<::unicode::General_Category> { { 0x1EA0, 0x1EA0 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EA1, 0x1EA1 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EA2, 0x1EA2 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EA3, 0x1EA3 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EA4, 0x1EA4 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND ACUTE
    Prop<::unicode::General_Category> { { 0x1EA5, 0x1EA5 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH CIRCUMFLEX AND ACUTE
    Prop<::unicode::General_Category> { { 0x1EA6, 0x1EA6 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND GRAVE
    Prop<::unicode::General_Category> { { 0x1EA7, 0x1EA7 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH CIRCUMFLEX AND GRAVE
    Prop<::unicode::General_Category> { { 0x1EA8, 0x1EA8 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EA9, 0x1EA9 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EAA, 0x1EAA }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND TILDE
    Prop<::unicode::General_Category> { { 0x1EAB, 0x1EAB }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH CIRCUMFLEX AND TILDE
    Prop<::unicode::General_Category> { { 0x1EAC, 0x1EAC }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EAD, 0x1EAD }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH CIRCUMFLEX AND DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EAE, 0x1EAE }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH BREVE AND ACUTE
    Prop<::unicode::General_Category> { { 0x1EAF, 0x1EAF }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH BREVE AND ACUTE
    Prop<::unicode::General_Category> { { 0x1EB0, 0x1EB0 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH BREVE AND GRAVE
    Prop<::unicode::General_Category> { { 0x1EB1, 0x1EB1 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH BREVE AND GRAVE
    Prop<::unicode::General_Category> { { 0x1EB2, 0x1EB2 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH BREVE AND HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EB3, 0x1EB3 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH BREVE AND HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EB4, 0x1EB4 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH BREVE AND TILDE
    Prop<::unicode::General_Category> { { 0x1EB5, 0x1EB5 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH BREVE AND TILDE
    Prop<::unicode::General_Category> { { 0x1EB6, 0x1EB6 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER A WITH BREVE AND DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EB7, 0x1EB7 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER A WITH BREVE AND DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EB8, 0x1EB8 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EB9, 0x1EB9 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EBA, 0x1EBA }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EBB, 0x1EBB }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EBC, 0x1EBC }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH TILDE
    Prop<::unicode::General_Category> { { 0x1EBD, 0x1EBD }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH TILDE
    Prop<::unicode::General_Category> { { 0x1EBE, 0x1EBE }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND ACUTE
    Prop<::unicode::General_Category> { { 0x1EBF, 0x1EBF }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH CIRCUMFLEX AND ACUTE
    Prop<::unicode::General_Category> { { 0x1EC0, 0x1EC0 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND GRAVE
    Prop<::unicode::General_Category> { { 0x1EC1, 0x1EC1 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH CIRCUMFLEX AND GRAVE
    Prop<::unicode::General_Category> { { 0x1EC2, 0x1EC2 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EC3, 0x1EC3 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EC4, 0x1EC4 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND TILDE
    Prop<::unicode::General_Category> { { 0x1EC5, 0x1EC5 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH CIRCUMFLEX AND TILDE
    Prop<::unicode::General_Category> { { 0x1EC6, 0x1EC6 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EC7, 0x1EC7 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER E WITH CIRCUMFLEX AND DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EC8, 0x1EC8 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER I WITH HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EC9, 0x1EC9 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER I WITH HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1ECA, 0x1ECA }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER I WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1ECB, 0x1ECB }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER I WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1ECC, 0x1ECC }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1ECD, 0x1ECD }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1ECE, 0x1ECE }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1ECF, 0x1ECF }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1ED0, 0x1ED0 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND ACUTE
    Prop<::unicode::General_Category> { { 0x1ED1, 0x1ED1 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH CIRCUMFLEX AND ACUTE
    Prop<::unicode::General_Category> { { 0x1ED2, 0x1ED2 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND GRAVE
    Prop<::unicode::General_Category> { { 0x1ED3, 0x1ED3 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH CIRCUMFLEX AND GRAVE
    Prop<::unicode::General_Category> { { 0x1ED4, 0x1ED4 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1ED5, 0x1ED5 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1ED6, 0x1ED6 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND TILDE
    Prop<::unicode::General_Category> { { 0x1ED7, 0x1ED7 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH CIRCUMFLEX AND TILDE
    Prop<::unicode::General_Category> { { 0x1ED8, 0x1ED8 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND DOT BELOW
    Prop<::unicode::General_Category> { { 0x1ED9, 0x1ED9 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH CIRCUMFLEX AND DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EDA, 0x1EDA }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH HORN AND ACUTE
    Prop<::unicode::General_Category> { { 0x1EDB, 0x1EDB }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH HORN AND ACUTE
    Prop<::unicode::General_Category> { { 0x1EDC, 0x1EDC }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH HORN AND GRAVE
    Prop<::unicode::General_Category> { { 0x1EDD, 0x1EDD }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH HORN AND GRAVE
    Prop<::unicode::General_Category> { { 0x1EDE, 0x1EDE }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH HORN AND HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EDF, 0x1EDF }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH HORN AND HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EE0, 0x1EE0 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH HORN AND TILDE
    Prop<::unicode::General_Category> { { 0x1EE1, 0x1EE1 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH HORN AND TILDE
    Prop<::unicode::General_Category> { { 0x1EE2, 0x1EE2 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH HORN AND DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EE3, 0x1EE3 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH HORN AND DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EE4, 0x1EE4 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EE5, 0x1EE5 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EE6, 0x1EE6 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EE7, 0x1EE7 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EE8, 0x1EE8 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH HORN AND ACUTE
    Prop<::unicode::General_Category> { { 0x1EE9, 0x1EE9 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH HORN AND ACUTE
    Prop<::unicode::General_Category> { { 0x1EEA, 0x1EEA }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH HORN AND GRAVE
    Prop<::unicode::General_Category> { { 0x1EEB, 0x1EEB }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH HORN AND GRAVE
    Prop<::unicode::General_Category> { { 0x1EEC, 0x1EEC }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH HORN AND HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EED, 0x1EED }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH HORN AND HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EEE, 0x1EEE }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH HORN AND TILDE
    Prop<::unicode::General_Category> { { 0x1EEF, 0x1EEF }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH HORN AND TILDE
    Prop<::unicode::General_Category> { { 0x1EF0, 0x1EF0 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH HORN AND DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EF1, 0x1EF1 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH HORN AND DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EF2, 0x1EF2 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Y WITH GRAVE
    Prop<::unicode::General_Category> { { 0x1EF3, 0x1EF3 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Y WITH GRAVE
    Prop<::unicode::General_Category> { { 0x1EF4, 0x1EF4 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Y WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EF5, 0x1EF5 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Y WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x1EF6, 0x1EF6 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Y WITH HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EF7, 0x1EF7 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Y WITH HOOK ABOVE
    Prop<::unicode::General_Category> { { 0x1EF8, 0x1EF8 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Y WITH TILDE
    Prop<::unicode::General_Category> { { 0x1EF9, 0x1EF9 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Y WITH TILDE
    Prop<::unicode::General_Category> { { 0x1EFA, 0x1EFA }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER MIDDLE-WELSH LL
    Prop<::unicode::General_Category> { { 0x1EFB, 0x1EFB }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER MIDDLE-WELSH LL
    Prop<::unicode::General_Category> { { 0x1EFC, 0x1EFC }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER MIDDLE-WELSH V
    Prop<::unicode::General_Category> { { 0x1EFD, 0x1EFD }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER MIDDLE-WELSH V
    Prop<::unicode::General_Category> { { 0x1EFE, 0x1EFE }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Y WITH LOOP
    Prop<::unicode::General_Category> { { 0x1EFF, 0x1F07 }, ::unicode::General_Category::Lowercase_Letter }, // [9] LATIN SMALL LETTER Y WITH LOOP..GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1F08, 0x1F0F }, ::unicode::General_Category::Uppercase_Letter }, // [8] GREEK CAPITAL LETTER ALPHA WITH PSILI..GREEK CAPITAL LETTER ALPHA WITH DASIA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1F10, 0x1F15 }, ::unicode::General_Category::Lowercase_Letter }, // [6] GREEK SMALL LETTER EPSILON WITH PSILI..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Prop<::unicode::General_Category> { { 0x1F16, 0x1F17 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1F16>..<reserved-1F17>
    Prop<::unicode::General_Category> { { 0x1F18, 0x1F1D }, ::unicode::General_Category::Uppercase_Letter }, // [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Prop<::unicode::General_Category> { { 0x1F1E, 0x1F1F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1F1E>..<reserved-1F1F>
    Prop<::unicode::General_Category> { { 0x1F20, 0x1F27 }, ::unicode::General_Category::Lowercase_Letter }, // [8] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1F28, 0x1F2F }, ::unicode::General_Category::Uppercase_Letter }, // [8] GREEK CAPITAL LETTER ETA WITH PSILI..GREEK CAPITAL LETTER ETA WITH DASIA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1F30, 0x1F37 }, ::unicode::General_Category::Lowercase_Letter }, // [8] GREEK SMALL LETTER IOTA WITH PSILI..GREEK SMALL LETTER IOTA WITH DASIA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1F38, 0x1F3F }, ::unicode::General_Category::Uppercase_Letter }, // [8] GREEK CAPITAL LETTER IOTA WITH PSILI..GREEK CAPITAL LETTER IOTA WITH DASIA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1F40, 0x1F45 }, ::unicode::General_Category::Lowercase_Letter }, // [6] GREEK SMALL LETTER OMICRON WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Prop<::unicode::General_Category> { { 0x1F46, 0x1F47 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1F46>..<reserved-1F47>
    Prop<::unicode::General_Category> { { 0x1F48, 0x1F4D }, ::unicode::General_Category::Uppercase_Letter }, // [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Prop<::unicode::General_Category> { { 0x1F4E, 0x1F4F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1F4E>..<reserved-1F4F>
    Prop<::unicode::General_Category> { { 0x1F50, 0x1F57 }, ::unicode::General_Category::Lowercase_Letter }, // [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1F58, 0x1F58 }, ::unicode::General_Category::Unassigned }, // <reserved-1F58>
    Prop<::unicode::General_Category> { { 0x1F59, 0x1F59 }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL LETTER UPSILON WITH DASIA
    Prop<::unicode::General_Category> { { 0x1F5A, 0x1F5A }, ::unicode::General_Category::Unassigned }, // <reserved-1F5A>
    Prop<::unicode::General_Category> { { 0x1F5B, 0x1F5B }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Prop<::unicode::General_Category> { { 0x1F5C, 0x1F5C }, ::unicode::General_Category::Unassigned }, // <reserved-1F5C>
    Prop<::unicode::General_Category> { { 0x1F5D, 0x1F5D }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Prop<::unicode::General_Category> { { 0x1F5E, 0x1F5E }, ::unicode::General_Category::Unassigned }, // <reserved-1F5E>
    Prop<::unicode::General_Category> { { 0x1F5F, 0x1F5F }, ::unicode::General_Category::Uppercase_Letter }, // GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1F60, 0x1F67 }, ::unicode::General_Category::Lowercase_Letter }, // [8] GREEK SMALL LETTER OMEGA WITH PSILI..GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1F68, 0x1F6F }, ::unicode::General_Category::Uppercase_Letter }, // [8] GREEK CAPITAL LETTER OMEGA WITH PSILI..GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1F70, 0x1F7D }, ::unicode::General_Category::Lowercase_Letter }, // [14] GREEK SMALL LETTER ALPHA WITH VARIA..GREEK SMALL LETTER OMEGA WITH OXIA
    Prop<::unicode::General_Category> { { 0x1F7E, 0x1F7F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1F7E>..<reserved-1F7F>
    Prop<::unicode::General_Category> { { 0x1F80, 0x1F87 }, ::unicode::General_Category::Lowercase_Letter }, // [8] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1F88, 0x1F8F }, ::unicode::General_Category::Titlecase_Letter }, // [8] GREEK CAPITAL LETTER ALPHA WITH PSILI AND PROSGEGRAMMENI..GREEK CAPITAL LETTER ALPHA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1F90, 0x1F97 }, ::unicode::General_Category::Lowercase_Letter }, // [8] GREEK SMALL LETTER ETA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1F98, 0x1F9F }, ::unicode::General_Category::Titlecase_Letter }, // [8] GREEK CAPITAL LETTER ETA WITH PSILI AND PROSGEGRAMMENI..GREEK CAPITAL LETTER ETA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FA0, 0x1FA7 }, ::unicode::General_Category::Lowercase_Letter }, // [8] GREEK SMALL LETTER OMEGA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FA8, 0x1FAF }, ::unicode::General_Category::Titlecase_Letter }, // [8] GREEK CAPITAL LETTER OMEGA WITH PSILI AND PROSGEGRAMMENI..GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FB0, 0x1FB4 }, ::unicode::General_Category::Lowercase_Letter }, // [5] GREEK SMALL LETTER ALPHA WITH VRACHY..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FB5, 0x1FB5 }, ::unicode::General_Category::Unassigned }, // <reserved-1FB5>
    Prop<::unicode::General_Category> { { 0x1FB6, 0x1FB7 }, ::unicode::General_Category::Lowercase_Letter }, // [2] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK SMALL LETTER ALPHA WITH PERISPOMENI AND YPOGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FB8, 0x1FBB }, ::unicode::General_Category::Uppercase_Letter }, // [4] GREEK CAPITAL LETTER ALPHA WITH VRACHY..GREEK CAPITAL LETTER ALPHA WITH OXIA
    Prop<::unicode::General_Category> { { 0x1FBC, 0x1FBC }, ::unicode::General_Category::Titlecase_Letter }, // GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FBD, 0x1FBD }, ::unicode::General_Category::Modifier_Symbol }, // GREEK KORONIS
    Prop<::unicode::General_Category> { { 0x1FBE, 0x1FBE }, ::unicode::General_Category::Lowercase_Letter }, // GREEK PROSGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FBF, 0x1FC1 }, ::unicode::General_Category::Modifier_Symbol }, // [3] GREEK PSILI..GREEK DIALYTIKA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1FC2, 0x1FC4 }, ::unicode::General_Category::Lowercase_Letter }, // [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FC5, 0x1FC5 }, ::unicode::General_Category::Unassigned }, // <reserved-1FC5>
    Prop<::unicode::General_Category> { { 0x1FC6, 0x1FC7 }, ::unicode::General_Category::Lowercase_Letter }, // [2] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK SMALL LETTER ETA WITH PERISPOMENI AND YPOGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FC8, 0x1FCB }, ::unicode::General_Category::Uppercase_Letter }, // [4] GREEK CAPITAL LETTER EPSILON WITH VARIA..GREEK CAPITAL LETTER ETA WITH OXIA
    Prop<::unicode::General_Category> { { 0x1FCC, 0x1FCC }, ::unicode::General_Category::Titlecase_Letter }, // GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FCD, 0x1FCF }, ::unicode::General_Category::Modifier_Symbol }, // [3] GREEK PSILI AND VARIA..GREEK PSILI AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1FD0, 0x1FD3 }, ::unicode::General_Category::Lowercase_Letter }, // [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Prop<::unicode::General_Category> { { 0x1FD4, 0x1FD5 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1FD4>..<reserved-1FD5>
    Prop<::unicode::General_Category> { { 0x1FD6, 0x1FD7 }, ::unicode::General_Category::Lowercase_Letter }, // [2] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1FD8, 0x1FDB }, ::unicode::General_Category::Uppercase_Letter }, // [4] GREEK CAPITAL LETTER IOTA WITH VRACHY..GREEK CAPITAL LETTER IOTA WITH OXIA
    Prop<::unicode::General_Category> { { 0x1FDC, 0x1FDC }, ::unicode::General_Category::Unassigned }, // <reserved-1FDC>
    Prop<::unicode::General_Category> { { 0x1FDD, 0x1FDF }, ::unicode::General_Category::Modifier_Symbol }, // [3] GREEK DASIA AND VARIA..GREEK DASIA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1FE0, 0x1FE7 }, ::unicode::General_Category::Lowercase_Letter }, // [8] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND PERISPOMENI
    Prop<::unicode::General_Category> { { 0x1FE8, 0x1FEC }, ::unicode::General_Category::Uppercase_Letter }, // [5] GREEK CAPITAL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Prop<::unicode::General_Category> { { 0x1FED, 0x1FEF }, ::unicode::General_Category::Modifier_Symbol }, // [3] GREEK DIALYTIKA AND VARIA..GREEK VARIA
    Prop<::unicode::General_Category> { { 0x1FF0, 0x1FF1 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1FF0>..<reserved-1FF1>
    Prop<::unicode::General_Category> { { 0x1FF2, 0x1FF4 }, ::unicode::General_Category::Lowercase_Letter }, // [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FF5, 0x1FF5 }, ::unicode::General_Category::Unassigned }, // <reserved-1FF5>
    Prop<::unicode::General_Category> { { 0x1FF6, 0x1FF7 }, ::unicode::General_Category::Lowercase_Letter }, // [2] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK SMALL LETTER OMEGA WITH PERISPOMENI AND YPOGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FF8, 0x1FFB }, ::unicode::General_Category::Uppercase_Letter }, // [4] GREEK CAPITAL LETTER OMICRON WITH VARIA..GREEK CAPITAL LETTER OMEGA WITH OXIA
    Prop<::unicode::General_Category> { { 0x1FFC, 0x1FFC }, ::unicode::General_Category::Titlecase_Letter }, // GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Prop<::unicode::General_Category> { { 0x1FFD, 0x1FFE }, ::unicode::General_Category::Modifier_Symbol }, // [2] GREEK OXIA..GREEK DASIA
    Prop<::unicode::General_Category> { { 0x1FFF, 0x1FFF }, ::unicode::General_Category::Unassigned }, // <reserved-1FFF>
    Prop<::unicode::General_Category> { { 0x2000, 0x200A }, ::unicode::General_Category::Space_Separator }, // [11] EN QUAD..HAIR SPACE
    Prop<::unicode::General_Category> { { 0x200B, 0x200F }, ::unicode::General_Category::Format }, // [5] ZERO WIDTH SPACE..RIGHT-TO-LEFT MARK
    Prop<::unicode::General_Category> { { 0x2010, 0x2015 }, ::unicode::General_Category::Dash_Punctuation }, // [6] HYPHEN..HORIZONTAL BAR
    Prop<::unicode::General_Category> { { 0x2016, 0x2017 }, ::unicode::General_Category::Other_Punctuation }, // [2] DOUBLE VERTICAL LINE..DOUBLE LOW LINE
    Prop<::unicode::General_Category> { { 0x2018, 0x2018 }, ::unicode::General_Category::Initial_Punctuation }, // LEFT SINGLE QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x2019, 0x2019 }, ::unicode::General_Category::Final_Punctuation }, // RIGHT SINGLE QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x201A, 0x201A }, ::unicode::General_Category::Open_Punctuation }, // SINGLE LOW-9 QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x201B, 0x201C }, ::unicode::General_Category::Initial_Punctuation }, // [2] SINGLE HIGH-REVERSED-9 QUOTATION MARK..LEFT DOUBLE QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x201D, 0x201D }, ::unicode::General_Category::Final_Punctuation }, // RIGHT DOUBLE QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x201E, 0x201E }, ::unicode::General_Category::Open_Punctuation }, // DOUBLE LOW-9 QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x201F, 0x201F }, ::unicode::General_Category::Initial_Punctuation }, // DOUBLE HIGH-REVERSED-9 QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x2020, 0x2027 }, ::unicode::General_Category::Other_Punctuation }, // [8] DAGGER..HYPHENATION POINT
    Prop<::unicode::General_Category> { { 0x2028, 0x2028 }, ::unicode::General_Category::Line_Separator }, // LINE SEPARATOR
    Prop<::unicode::General_Category> { { 0x2029, 0x2029 }, ::unicode::General_Category::Paragraph_Separator }, // PARAGRAPH SEPARATOR
    Prop<::unicode::General_Category> { { 0x202A, 0x202E }, ::unicode::General_Category::Format }, // [5] LEFT-TO-RIGHT EMBEDDING..RIGHT-TO-LEFT OVERRIDE
    Prop<::unicode::General_Category> { { 0x202F, 0x202F }, ::unicode::General_Category::Space_Separator }, // NARROW NO-BREAK SPACE
    Prop<::unicode::General_Category> { { 0x2030, 0x2038 }, ::unicode::General_Category::Other_Punctuation }, // [9] PER MILLE SIGN..CARET
    Prop<::unicode::General_Category> { { 0x2039, 0x2039 }, ::unicode::General_Category::Initial_Punctuation }, // SINGLE LEFT-POINTING ANGLE QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x203A, 0x203A }, ::unicode::General_Category::Final_Punctuation }, // SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x203B, 0x203E }, ::unicode::General_Category::Other_Punctuation }, // [4] REFERENCE MARK..OVERLINE
    Prop<::unicode::General_Category> { { 0x203F, 0x2040 }, ::unicode::General_Category::Connector_Punctuation }, // [2] UNDERTIE..CHARACTER TIE
    Prop<::unicode::General_Category> { { 0x2041, 0x2043 }, ::unicode::General_Category::Other_Punctuation }, // [3] CARET INSERTION POINT..HYPHEN BULLET
    Prop<::unicode::General_Category> { { 0x2044, 0x2044 }, ::unicode::General_Category::Math_Symbol }, // FRACTION SLASH
    Prop<::unicode::General_Category> { { 0x2045, 0x2045 }, ::unicode::General_Category::Open_Punctuation }, // LEFT SQUARE BRACKET WITH QUILL
    Prop<::unicode::General_Category> { { 0x2046, 0x2046 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT SQUARE BRACKET WITH QUILL
    Prop<::unicode::General_Category> { { 0x2047, 0x2051 }, ::unicode::General_Category::Other_Punctuation }, // [11] DOUBLE QUESTION MARK..TWO ASTERISKS ALIGNED VERTICALLY
    Prop<::unicode::General_Category> { { 0x2052, 0x2052 }, ::unicode::General_Category::Math_Symbol }, // COMMERCIAL MINUS SIGN
    Prop<::unicode::General_Category> { { 0x2053, 0x2053 }, ::unicode::General_Category::Other_Punctuation }, // SWUNG DASH
    Prop<::unicode::General_Category> { { 0x2054, 0x2054 }, ::unicode::General_Category::Connector_Punctuation }, // INVERTED UNDERTIE
    Prop<::unicode::General_Category> { { 0x2055, 0x205E }, ::unicode::General_Category::Other_Punctuation }, // [10] FLOWER PUNCTUATION MARK..VERTICAL FOUR DOTS
    Prop<::unicode::General_Category> { { 0x205F, 0x205F }, ::unicode::General_Category::Space_Separator }, // MEDIUM MATHEMATICAL SPACE
    Prop<::unicode::General_Category> { { 0x2060, 0x2064 }, ::unicode::General_Category::Format }, // [5] WORD JOINER..INVISIBLE PLUS
    Prop<::unicode::General_Category> { { 0x2065, 0x2065 }, ::unicode::General_Category::Unassigned }, // <reserved-2065>
    Prop<::unicode::General_Category> { { 0x2066, 0x206F }, ::unicode::General_Category::Format }, // [10] LEFT-TO-RIGHT ISOLATE..NOMINAL DIGIT SHAPES
    Prop<::unicode::General_Category> { { 0x2070, 0x2070 }, ::unicode::General_Category::Other_Number }, // SUPERSCRIPT ZERO
    Prop<::unicode::General_Category> { { 0x2071, 0x2071 }, ::unicode::General_Category::Modifier_Letter }, // SUPERSCRIPT LATIN SMALL LETTER I
    Prop<::unicode::General_Category> { { 0x2072, 0x2073 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-2072>..<reserved-2073>
    Prop<::unicode::General_Category> { { 0x2074, 0x2079 }, ::unicode::General_Category::Other_Number }, // [6] SUPERSCRIPT FOUR..SUPERSCRIPT NINE
    Prop<::unicode::General_Category> { { 0x207A, 0x207C }, ::unicode::General_Category::Math_Symbol }, // [3] SUPERSCRIPT PLUS SIGN..SUPERSCRIPT EQUALS SIGN
    Prop<::unicode::General_Category> { { 0x207D, 0x207D }, ::unicode::General_Category::Open_Punctuation }, // SUPERSCRIPT LEFT PARENTHESIS
    Prop<::unicode::General_Category> { { 0x207E, 0x207E }, ::unicode::General_Category::Close_Punctuation }, // SUPERSCRIPT RIGHT PARENTHESIS
    Prop<::unicode::General_Category> { { 0x207F, 0x207F }, ::unicode::General_Category::Modifier_Letter }, // SUPERSCRIPT LATIN SMALL LETTER N
    Prop<::unicode::General_Category> { { 0x2080, 0x2089 }, ::unicode::General_Category::Other_Number }, // [10] SUBSCRIPT ZERO..SUBSCRIPT NINE
    Prop<::unicode::General_Category> { { 0x208A, 0x208C }, ::unicode::General_Category::Math_Symbol }, // [3] SUBSCRIPT PLUS SIGN..SUBSCRIPT EQUALS SIGN
    Prop<::unicode::General_Category> { { 0x208D, 0x208D }, ::unicode::General_Category::Open_Punctuation }, // SUBSCRIPT LEFT PARENTHESIS
    Prop<::unicode::General_Category> { { 0x208E, 0x208E }, ::unicode::General_Category::Close_Punctuation }, // SUBSCRIPT RIGHT PARENTHESIS
    Prop<::unicode::General_Category> { { 0x208F, 0x208F }, ::unicode::General_Category::Unassigned }, // <reserved-208F>
    Prop<::unicode::General_Category> { { 0x2090, 0x209C }, ::unicode::General_Category::Modifier_Letter }, // [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Prop<::unicode::General_Category> { { 0x209D, 0x209F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-209D>..<reserved-209F>
    Prop<::unicode::General_Category> { { 0x20A0, 0x20C1 }, ::unicode::General_Category::Currency_Symbol }, // [34] EURO-CURRENCY SIGN..SAUDI RIYAL SIGN
    Prop<::unicode::General_Category> { { 0x20C2, 0x20CF }, ::unicode::General_Category::Unassigned }, // [14] <reserved-20C2>..<reserved-20CF>
    Prop<::unicode::General_Category> { { 0x20D0, 0x20DC }, ::unicode::General_Category::Nonspacing_Mark }, // [13] COMBINING LEFT HARPOON ABOVE..COMBINING FOUR DOTS ABOVE
    Prop<::unicode::General_Category> { { 0x20DD, 0x20E0 }, ::unicode::General_Category::Enclosing_Mark }, // [4] COMBINING ENCLOSING CIRCLE..COMBINING ENCLOSING CIRCLE BACKSLASH
    Prop<::unicode::General_Category> { { 0x20E1, 0x20E1 }, ::unicode::General_Category::Nonspacing_Mark }, // COMBINING LEFT RIGHT ARROW ABOVE
    Prop<::unicode::General_Category> { { 0x20E2, 0x20E4 }, ::unicode::General_Category::Enclosing_Mark }, // [3] COMBINING ENCLOSING SCREEN..COMBINING ENCLOSING UPWARD POINTING TRIANGLE
    Prop<::unicode::General_Category> { { 0x20E5, 0x20F0 }, ::unicode::General_Category::Nonspacing_Mark }, // [12] COMBINING REVERSE SOLIDUS OVERLAY..COMBINING ASTERISK ABOVE
    Prop<::unicode::General_Category> { { 0x20F1, 0x20FF }, ::unicode::General_Category::Unassigned }, // [15] <reserved-20F1>..<reserved-20FF>
    Prop<::unicode::General_Category> { { 0x2100, 0x2101 }, ::unicode::General_Category::Other_Symbol }, // [2] ACCOUNT OF..ADDRESSED TO THE SUBJECT
    Prop<::unicode::General_Category> { { 0x2102, 0x2102 }, ::unicode::General_Category::Uppercase_Letter }, // DOUBLE-STRUCK CAPITAL C
    Prop<::unicode::General_Category> { { 0x2103, 0x2106 }, ::unicode::General_Category::Other_Symbol }, // [4] DEGREE CELSIUS..CADA UNA
    Prop<::unicode::General_Category> { { 0x2107, 0x2107 }, ::unicode::General_Category::Uppercase_Letter }, // EULER CONSTANT
    Prop<::unicode::General_Category> { { 0x2108, 0x2109 }, ::unicode::General_Category::Other_Symbol }, // [2] SCRUPLE..DEGREE FAHRENHEIT
    Prop<::unicode::General_Category> { { 0x210A, 0x210A }, ::unicode::General_Category::Lowercase_Letter }, // SCRIPT SMALL G
    Prop<::unicode::General_Category> { { 0x210B, 0x210D }, ::unicode::General_Category::Uppercase_Letter }, // [3] SCRIPT CAPITAL H..DOUBLE-STRUCK CAPITAL H
    Prop<::unicode::General_Category> { { 0x210E, 0x210F }, ::unicode::General_Category::Lowercase_Letter }, // [2] PLANCK CONSTANT..PLANCK CONSTANT OVER TWO PI
    Prop<::unicode::General_Category> { { 0x2110, 0x2112 }, ::unicode::General_Category::Uppercase_Letter }, // [3] SCRIPT CAPITAL I..SCRIPT CAPITAL L
    Prop<::unicode::General_Category> { { 0x2113, 0x2113 }, ::unicode::General_Category::Lowercase_Letter }, // SCRIPT SMALL L
    Prop<::unicode::General_Category> { { 0x2114, 0x2114 }, ::unicode::General_Category::Other_Symbol }, // L B BAR SYMBOL
    Prop<::unicode::General_Category> { { 0x2115, 0x2115 }, ::unicode::General_Category::Uppercase_Letter }, // DOUBLE-STRUCK CAPITAL N
    Prop<::unicode::General_Category> { { 0x2116, 0x2117 }, ::unicode::General_Category::Other_Symbol }, // [2] NUMERO SIGN..SOUND RECORDING COPYRIGHT
    Prop<::unicode::General_Category> { { 0x2118, 0x2118 }, ::unicode::General_Category::Math_Symbol }, // SCRIPT CAPITAL P
    Prop<::unicode::General_Category> { { 0x2119, 0x211D }, ::unicode::General_Category::Uppercase_Letter }, // [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    Prop<::unicode::General_Category> { { 0x211E, 0x2123 }, ::unicode::General_Category::Other_Symbol }, // [6] PRESCRIPTION TAKE..VERSICLE
    Prop<::unicode::General_Category> { { 0x2124, 0x2124 }, ::unicode::General_Category::Uppercase_Letter }, // DOUBLE-STRUCK CAPITAL Z
    Prop<::unicode::General_Category> { { 0x2125, 0x2125 }, ::unicode::General_Category::Other_Symbol }, // OUNCE SIGN
    Prop<::unicode::General_Category> { { 0x2126, 0x2126 }, ::unicode::General_Category::Uppercase_Letter }, // OHM SIGN
    Prop<::unicode::General_Category> { { 0x2127, 0x2127 }, ::unicode::General_Category::Other_Symbol }, // INVERTED OHM SIGN
    Prop<::unicode::General_Category> { { 0x2128, 0x2128 }, ::unicode::General_Category::Uppercase_Letter }, // BLACK-LETTER CAPITAL Z
    Prop<::unicode::General_Category> { { 0x2129, 0x2129 }, ::unicode::General_Category::Other_Symbol }, // TURNED GREEK SMALL LETTER IOTA
    Prop<::unicode::General_Category> { { 0x212A, 0x212D }, ::unicode::General_Category::Uppercase_Letter }, // [4] KELVIN SIGN..BLACK-LETTER CAPITAL C
    Prop<::unicode::General_Category> { { 0x212E, 0x212E }, ::unicode::General_Category::Other_Symbol }, // ESTIMATED SYMBOL
    Prop<::unicode::General_Category> { { 0x212F, 0x212F }, ::unicode::General_Category::Lowercase_Letter }, // SCRIPT SMALL E
    Prop<::unicode::General_Category> { { 0x2130, 0x2133 }, ::unicode::General_Category::Uppercase_Letter }, // [4] SCRIPT CAPITAL E..SCRIPT CAPITAL M
    Prop<::unicode::General_Category> { { 0x2134, 0x2134 }, ::unicode::General_Category::Lowercase_Letter }, // SCRIPT SMALL O
    Prop<::unicode::General_Category> { { 0x2135, 0x2138 }, ::unicode::General_Category::Other_Letter }, // [4] ALEF SYMBOL..DALET SYMBOL
    Prop<::unicode::General_Category> { { 0x2139, 0x2139 }, ::unicode::General_Category::Lowercase_Letter }, // INFORMATION SOURCE
    Prop<::unicode::General_Category> { { 0x213A, 0x213B }, ::unicode::General_Category::Other_Symbol }, // [2] ROTATED CAPITAL Q..FACSIMILE SIGN
    Prop<::unicode::General_Category> { { 0x213C, 0x213D }, ::unicode::General_Category::Lowercase_Letter }, // [2] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK SMALL GAMMA
    Prop<::unicode::General_Category> { { 0x213E, 0x213F }, ::unicode::General_Category::Uppercase_Letter }, // [2] DOUBLE-STRUCK CAPITAL GAMMA..DOUBLE-STRUCK CAPITAL PI
    Prop<::unicode::General_Category> { { 0x2140, 0x2144 }, ::unicode::General_Category::Math_Symbol }, // [5] DOUBLE-STRUCK N-ARY SUMMATION..TURNED SANS-SERIF CAPITAL Y
    Prop<::unicode::General_Category> { { 0x2145, 0x2145 }, ::unicode::General_Category::Uppercase_Letter }, // DOUBLE-STRUCK ITALIC CAPITAL D
    Prop<::unicode::General_Category> { { 0x2146, 0x2149 }, ::unicode::General_Category::Lowercase_Letter }, // [4] DOUBLE-STRUCK ITALIC SMALL D..DOUBLE-STRUCK ITALIC SMALL J
    Prop<::unicode::General_Category> { { 0x214A, 0x214A }, ::unicode::General_Category::Other_Symbol }, // PROPERTY LINE
    Prop<::unicode::General_Category> { { 0x214B, 0x214B }, ::unicode::General_Category::Math_Symbol }, // TURNED AMPERSAND
    Prop<::unicode::General_Category> { { 0x214C, 0x214D }, ::unicode::General_Category::Other_Symbol }, // [2] PER SIGN..AKTIESELSKAB
    Prop<::unicode::General_Category> { { 0x214E, 0x214E }, ::unicode::General_Category::Lowercase_Letter }, // TURNED SMALL F
    Prop<::unicode::General_Category> { { 0x214F, 0x214F }, ::unicode::General_Category::Other_Symbol }, // SYMBOL FOR SAMARITAN SOURCE
    Prop<::unicode::General_Category> { { 0x2150, 0x215F }, ::unicode::General_Category::Other_Number }, // [16] VULGAR FRACTION ONE SEVENTH..FRACTION NUMERATOR ONE
    Prop<::unicode::General_Category> { { 0x2160, 0x2182 }, ::unicode::General_Category::Letter_Number }, // [35] ROMAN NUMERAL ONE..ROMAN NUMERAL TEN THOUSAND
    Prop<::unicode::General_Category> { { 0x2183, 0x2183 }, ::unicode::General_Category::Uppercase_Letter }, // ROMAN NUMERAL REVERSED ONE HUNDRED
    Prop<::unicode::General_Category> { { 0x2184, 0x2184 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER REVERSED C
    Prop<::unicode::General_Category> { { 0x2185, 0x2188 }, ::unicode::General_Category::Letter_Number }, // [4] ROMAN NUMERAL SIX LATE FORM..ROMAN NUMERAL ONE HUNDRED THOUSAND
    Prop<::unicode::General_Category> { { 0x2189, 0x2189 }, ::unicode::General_Category::Other_Number }, // VULGAR FRACTION ZERO THIRDS
    Prop<::unicode::General_Category> { { 0x218A, 0x218B }, ::unicode::General_Category::Other_Symbol }, // [2] TURNED DIGIT TWO..TURNED DIGIT THREE
    Prop<::unicode::General_Category> { { 0x218C, 0x218F }, ::unicode::General_Category::Unassigned }, // [4] <reserved-218C>..<reserved-218F>
    Prop<::unicode::General_Category> { { 0x2190, 0x2194 }, ::unicode::General_Category::Math_Symbol }, // [5] LEFTWARDS ARROW..LEFT RIGHT ARROW
    Prop<::unicode::General_Category> { { 0x2195, 0x2199 }, ::unicode::General_Category::Other_Symbol }, // [5] UP DOWN ARROW..SOUTH WEST ARROW
    Prop<::unicode::General_Category> { { 0x219A, 0x219B }, ::unicode::General_Category::Math_Symbol }, // [2] LEFTWARDS ARROW WITH STROKE..RIGHTWARDS ARROW WITH STROKE
    Prop<::unicode::General_Category> { { 0x219C, 0x219F }, ::unicode::General_Category::Other_Symbol }, // [4] LEFTWARDS WAVE ARROW..UPWARDS TWO HEADED ARROW
    Prop<::unicode::General_Category> { { 0x21A0, 0x21A0 }, ::unicode::General_Category::Math_Symbol }, // RIGHTWARDS TWO HEADED ARROW
    Prop<::unicode::General_Category> { { 0x21A1, 0x21A2 }, ::unicode::General_Category::Other_Symbol }, // [2] DOWNWARDS TWO HEADED ARROW..LEFTWARDS ARROW WITH TAIL
    Prop<::unicode::General_Category> { { 0x21A3, 0x21A3 }, ::unicode::General_Category::Math_Symbol }, // RIGHTWARDS ARROW WITH TAIL
    Prop<::unicode::General_Category> { { 0x21A4, 0x21A5 }, ::unicode::General_Category::Other_Symbol }, // [2] LEFTWARDS ARROW FROM BAR..UPWARDS ARROW FROM BAR
    Prop<::unicode::General_Category> { { 0x21A6, 0x21A6 }, ::unicode::General_Category::Math_Symbol }, // RIGHTWARDS ARROW FROM BAR
    Prop<::unicode::General_Category> { { 0x21A7, 0x21AD }, ::unicode::General_Category::Other_Symbol }, // [7] DOWNWARDS ARROW FROM BAR..LEFT RIGHT WAVE ARROW
    Prop<::unicode::General_Category> { { 0x21AE, 0x21AE }, ::unicode::General_Category::Math_Symbol }, // LEFT RIGHT ARROW WITH STROKE
    Prop<::unicode::General_Category> { { 0x21AF, 0x21CD }, ::unicode::General_Category::Other_Symbol }, // [31] DOWNWARDS ZIGZAG ARROW..LEFTWARDS DOUBLE ARROW WITH STROKE
    Prop<::unicode::General_Category> { { 0x21CE, 0x21CF }, ::unicode::General_Category::Math_Symbol }, // [2] LEFT RIGHT DOUBLE ARROW WITH STROKE..RIGHTWARDS DOUBLE ARROW WITH STROKE
    Prop<::unicode::General_Category> { { 0x21D0, 0x21D1 }, ::unicode::General_Category::Other_Symbol }, // [2] LEFTWARDS DOUBLE ARROW..UPWARDS DOUBLE ARROW
    Prop<::unicode::General_Category> { { 0x21D2, 0x21D2 }, ::unicode::General_Category::Math_Symbol }, // RIGHTWARDS DOUBLE ARROW
    Prop<::unicode::General_Category> { { 0x21D3, 0x21D3 }, ::unicode::General_Category::Other_Symbol }, // DOWNWARDS DOUBLE ARROW
    Prop<::unicode::General_Category> { { 0x21D4, 0x21D4 }, ::unicode::General_Category::Math_Symbol }, // LEFT RIGHT DOUBLE ARROW
    Prop<::unicode::General_Category> { { 0x21D5, 0x21F3 }, ::unicode::General_Category::Other_Symbol }, // [31] UP DOWN DOUBLE ARROW..UP DOWN WHITE ARROW
    Prop<::unicode::General_Category> { { 0x21F4, 0x22FF }, ::unicode::General_Category::Math_Symbol }, // [268] RIGHT ARROW WITH SMALL CIRCLE..Z NOTATION BAG MEMBERSHIP
    Prop<::unicode::General_Category> { { 0x2300, 0x2307 }, ::unicode::General_Category::Other_Symbol }, // [8] DIAMETER SIGN..WAVY LINE
    Prop<::unicode::General_Category> { { 0x2308, 0x2308 }, ::unicode::General_Category::Open_Punctuation }, // LEFT CEILING
    Prop<::unicode::General_Category> { { 0x2309, 0x2309 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT CEILING
    Prop<::unicode::General_Category> { { 0x230A, 0x230A }, ::unicode::General_Category::Open_Punctuation }, // LEFT FLOOR
    Prop<::unicode::General_Category> { { 0x230B, 0x230B }, ::unicode::General_Category::Close_Punctuation }, // RIGHT FLOOR
    Prop<::unicode::General_Category> { { 0x230C, 0x231F }, ::unicode::General_Category::Other_Symbol }, // [20] BOTTOM RIGHT CROP..BOTTOM RIGHT CORNER
    Prop<::unicode::General_Category> { { 0x2320, 0x2321 }, ::unicode::General_Category::Math_Symbol }, // [2] TOP HALF INTEGRAL..BOTTOM HALF INTEGRAL
    Prop<::unicode::General_Category> { { 0x2322, 0x2328 }, ::unicode::General_Category::Other_Symbol }, // [7] FROWN..KEYBOARD
    Prop<::unicode::General_Category> { { 0x2329, 0x2329 }, ::unicode::General_Category::Open_Punctuation }, // LEFT-POINTING ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0x232A, 0x232A }, ::unicode::General_Category::Close_Punctuation }, // RIGHT-POINTING ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0x232B, 0x237B }, ::unicode::General_Category::Other_Symbol }, // [81] ERASE TO THE LEFT..NOT CHECK MARK
    Prop<::unicode::General_Category> { { 0x237C, 0x237C }, ::unicode::General_Category::Math_Symbol }, // RIGHT ANGLE WITH DOWNWARDS ZIGZAG ARROW
    Prop<::unicode::General_Category> { { 0x237D, 0x239A }, ::unicode::General_Category::Other_Symbol }, // [30] SHOULDERED OPEN BOX..CLEAR SCREEN SYMBOL
    Prop<::unicode::General_Category> { { 0x239B, 0x23B3 }, ::unicode::General_Category::Math_Symbol }, // [25] LEFT PARENTHESIS UPPER HOOK..SUMMATION BOTTOM
    Prop<::unicode::General_Category> { { 0x23B4, 0x23DB }, ::unicode::General_Category::Other_Symbol }, // [40] TOP SQUARE BRACKET..FUSE
    Prop<::unicode::General_Category> { { 0x23DC, 0x23E1 }, ::unicode::General_Category::Math_Symbol }, // [6] TOP PARENTHESIS..BOTTOM TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0x23E2, 0x2429 }, ::unicode::General_Category::Other_Symbol }, // [72] WHITE TRAPEZIUM..SYMBOL FOR DELETE MEDIUM SHADE FORM
    Prop<::unicode::General_Category> { { 0x242A, 0x243F }, ::unicode::General_Category::Unassigned }, // [22] <reserved-242A>..<reserved-243F>
    Prop<::unicode::General_Category> { { 0x2440, 0x244A }, ::unicode::General_Category::Other_Symbol }, // [11] OCR HOOK..OCR DOUBLE BACKSLASH
    Prop<::unicode::General_Category> { { 0x244B, 0x245F }, ::unicode::General_Category::Unassigned }, // [21] <reserved-244B>..<reserved-245F>
    Prop<::unicode::General_Category> { { 0x2460, 0x249B }, ::unicode::General_Category::Other_Number }, // [60] CIRCLED DIGIT ONE..NUMBER TWENTY FULL STOP
    Prop<::unicode::General_Category> { { 0x249C, 0x24E9 }, ::unicode::General_Category::Other_Symbol }, // [78] PARENTHESIZED LATIN SMALL LETTER A..CIRCLED LATIN SMALL LETTER Z
    Prop<::unicode::General_Category> { { 0x24EA, 0x24FF }, ::unicode::General_Category::Other_Number }, // [22] CIRCLED DIGIT ZERO..NEGATIVE CIRCLED DIGIT ZERO
    Prop<::unicode::General_Category> { { 0x2500, 0x25B6 }, ::unicode::General_Category::Other_Symbol }, // [183] BOX DRAWINGS LIGHT HORIZONTAL..BLACK RIGHT-POINTING TRIANGLE
    Prop<::unicode::General_Category> { { 0x25B7, 0x25B7 }, ::unicode::General_Category::Math_Symbol }, // WHITE RIGHT-POINTING TRIANGLE
    Prop<::unicode::General_Category> { { 0x25B8, 0x25C0 }, ::unicode::General_Category::Other_Symbol }, // [9] BLACK RIGHT-POINTING SMALL TRIANGLE..BLACK LEFT-POINTING TRIANGLE
    Prop<::unicode::General_Category> { { 0x25C1, 0x25C1 }, ::unicode::General_Category::Math_Symbol }, // WHITE LEFT-POINTING TRIANGLE
    Prop<::unicode::General_Category> { { 0x25C2, 0x25F7 }, ::unicode::General_Category::Other_Symbol }, // [54] BLACK LEFT-POINTING SMALL TRIANGLE..WHITE CIRCLE WITH UPPER RIGHT QUADRANT
    Prop<::unicode::General_Category> { { 0x25F8, 0x25FF }, ::unicode::General_Category::Math_Symbol }, // [8] UPPER LEFT TRIANGLE..LOWER RIGHT TRIANGLE
    Prop<::unicode::General_Category> { { 0x2600, 0x266E }, ::unicode::General_Category::Other_Symbol }, // [111] BLACK SUN WITH RAYS..MUSIC NATURAL SIGN
    Prop<::unicode::General_Category> { { 0x266F, 0x266F }, ::unicode::General_Category::Math_Symbol }, // MUSIC SHARP SIGN
    Prop<::unicode::General_Category> { { 0x2670, 0x2767 }, ::unicode::General_Category::Other_Symbol }, // [248] WEST SYRIAC CROSS..ROTATED FLORAL HEART BULLET
    Prop<::unicode::General_Category> { { 0x2768, 0x2768 }, ::unicode::General_Category::Open_Punctuation }, // MEDIUM LEFT PARENTHESIS ORNAMENT
    Prop<::unicode::General_Category> { { 0x2769, 0x2769 }, ::unicode::General_Category::Close_Punctuation }, // MEDIUM RIGHT PARENTHESIS ORNAMENT
    Prop<::unicode::General_Category> { { 0x276A, 0x276A }, ::unicode::General_Category::Open_Punctuation }, // MEDIUM FLATTENED LEFT PARENTHESIS ORNAMENT
    Prop<::unicode::General_Category> { { 0x276B, 0x276B }, ::unicode::General_Category::Close_Punctuation }, // MEDIUM FLATTENED RIGHT PARENTHESIS ORNAMENT
    Prop<::unicode::General_Category> { { 0x276C, 0x276C }, ::unicode::General_Category::Open_Punctuation }, // MEDIUM LEFT-POINTING ANGLE BRACKET ORNAMENT
    Prop<::unicode::General_Category> { { 0x276D, 0x276D }, ::unicode::General_Category::Close_Punctuation }, // MEDIUM RIGHT-POINTING ANGLE BRACKET ORNAMENT
    Prop<::unicode::General_Category> { { 0x276E, 0x276E }, ::unicode::General_Category::Open_Punctuation }, // HEAVY LEFT-POINTING ANGLE QUOTATION MARK ORNAMENT
    Prop<::unicode::General_Category> { { 0x276F, 0x276F }, ::unicode::General_Category::Close_Punctuation }, // HEAVY RIGHT-POINTING ANGLE QUOTATION MARK ORNAMENT
    Prop<::unicode::General_Category> { { 0x2770, 0x2770 }, ::unicode::General_Category::Open_Punctuation }, // HEAVY LEFT-POINTING ANGLE BRACKET ORNAMENT
    Prop<::unicode::General_Category> { { 0x2771, 0x2771 }, ::unicode::General_Category::Close_Punctuation }, // HEAVY RIGHT-POINTING ANGLE BRACKET ORNAMENT
    Prop<::unicode::General_Category> { { 0x2772, 0x2772 }, ::unicode::General_Category::Open_Punctuation }, // LIGHT LEFT TORTOISE SHELL BRACKET ORNAMENT
    Prop<::unicode::General_Category> { { 0x2773, 0x2773 }, ::unicode::General_Category::Close_Punctuation }, // LIGHT RIGHT TORTOISE SHELL BRACKET ORNAMENT
    Prop<::unicode::General_Category> { { 0x2774, 0x2774 }, ::unicode::General_Category::Open_Punctuation }, // MEDIUM LEFT CURLY BRACKET ORNAMENT
    Prop<::unicode::General_Category> { { 0x2775, 0x2775 }, ::unicode::General_Category::Close_Punctuation }, // MEDIUM RIGHT CURLY BRACKET ORNAMENT
    Prop<::unicode::General_Category> { { 0x2776, 0x2793 }, ::unicode::General_Category::Other_Number }, // [30] DINGBAT NEGATIVE CIRCLED DIGIT ONE..DINGBAT NEGATIVE CIRCLED SANS-SERIF NUMBER TEN
    Prop<::unicode::General_Category> { { 0x2794, 0x27BF }, ::unicode::General_Category::Other_Symbol }, // [44] HEAVY WIDE-HEADED RIGHTWARDS ARROW..DOUBLE CURLY LOOP
    Prop<::unicode::General_Category> { { 0x27C0, 0x27C4 }, ::unicode::General_Category::Math_Symbol }, // [5] THREE DIMENSIONAL ANGLE..OPEN SUPERSET
    Prop<::unicode::General_Category> { { 0x27C5, 0x27C5 }, ::unicode::General_Category::Open_Punctuation }, // LEFT S-SHAPED BAG DELIMITER
    Prop<::unicode::General_Category> { { 0x27C6, 0x27C6 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT S-SHAPED BAG DELIMITER
    Prop<::unicode::General_Category> { { 0x27C7, 0x27E5 }, ::unicode::General_Category::Math_Symbol }, // [31] OR WITH DOT INSIDE..WHITE SQUARE WITH RIGHTWARDS TICK
    Prop<::unicode::General_Category> { { 0x27E6, 0x27E6 }, ::unicode::General_Category::Open_Punctuation }, // MATHEMATICAL LEFT WHITE SQUARE BRACKET
    Prop<::unicode::General_Category> { { 0x27E7, 0x27E7 }, ::unicode::General_Category::Close_Punctuation }, // MATHEMATICAL RIGHT WHITE SQUARE BRACKET
    Prop<::unicode::General_Category> { { 0x27E8, 0x27E8 }, ::unicode::General_Category::Open_Punctuation }, // MATHEMATICAL LEFT ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0x27E9, 0x27E9 }, ::unicode::General_Category::Close_Punctuation }, // MATHEMATICAL RIGHT ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0x27EA, 0x27EA }, ::unicode::General_Category::Open_Punctuation }, // MATHEMATICAL LEFT DOUBLE ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0x27EB, 0x27EB }, ::unicode::General_Category::Close_Punctuation }, // MATHEMATICAL RIGHT DOUBLE ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0x27EC, 0x27EC }, ::unicode::General_Category::Open_Punctuation }, // MATHEMATICAL LEFT WHITE TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0x27ED, 0x27ED }, ::unicode::General_Category::Close_Punctuation }, // MATHEMATICAL RIGHT WHITE TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0x27EE, 0x27EE }, ::unicode::General_Category::Open_Punctuation }, // MATHEMATICAL LEFT FLATTENED PARENTHESIS
    Prop<::unicode::General_Category> { { 0x27EF, 0x27EF }, ::unicode::General_Category::Close_Punctuation }, // MATHEMATICAL RIGHT FLATTENED PARENTHESIS
    Prop<::unicode::General_Category> { { 0x27F0, 0x27FF }, ::unicode::General_Category::Math_Symbol }, // [16] UPWARDS QUADRUPLE ARROW..LONG RIGHTWARDS SQUIGGLE ARROW
    Prop<::unicode::General_Category> { { 0x2800, 0x28FF }, ::unicode::General_Category::Other_Symbol }, // [256] BRAILLE PATTERN BLANK..BRAILLE PATTERN DOTS-12345678
    Prop<::unicode::General_Category> { { 0x2900, 0x2982 }, ::unicode::General_Category::Math_Symbol }, // [131] RIGHTWARDS TWO-HEADED ARROW WITH VERTICAL STROKE..Z NOTATION TYPE COLON
    Prop<::unicode::General_Category> { { 0x2983, 0x2983 }, ::unicode::General_Category::Open_Punctuation }, // LEFT WHITE CURLY BRACKET
    Prop<::unicode::General_Category> { { 0x2984, 0x2984 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT WHITE CURLY BRACKET
    Prop<::unicode::General_Category> { { 0x2985, 0x2985 }, ::unicode::General_Category::Open_Punctuation }, // LEFT WHITE PARENTHESIS
    Prop<::unicode::General_Category> { { 0x2986, 0x2986 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT WHITE PARENTHESIS
    Prop<::unicode::General_Category> { { 0x2987, 0x2987 }, ::unicode::General_Category::Open_Punctuation }, // Z NOTATION LEFT IMAGE BRACKET
    Prop<::unicode::General_Category> { { 0x2988, 0x2988 }, ::unicode::General_Category::Close_Punctuation }, // Z NOTATION RIGHT IMAGE BRACKET
    Prop<::unicode::General_Category> { { 0x2989, 0x2989 }, ::unicode::General_Category::Open_Punctuation }, // Z NOTATION LEFT BINDING BRACKET
    Prop<::unicode::General_Category> { { 0x298A, 0x298A }, ::unicode::General_Category::Close_Punctuation }, // Z NOTATION RIGHT BINDING BRACKET
    Prop<::unicode::General_Category> { { 0x298B, 0x298B }, ::unicode::General_Category::Open_Punctuation }, // LEFT SQUARE BRACKET WITH UNDERBAR
    Prop<::unicode::General_Category> { { 0x298C, 0x298C }, ::unicode::General_Category::Close_Punctuation }, // RIGHT SQUARE BRACKET WITH UNDERBAR
    Prop<::unicode::General_Category> { { 0x298D, 0x298D }, ::unicode::General_Category::Open_Punctuation }, // LEFT SQUARE BRACKET WITH TICK IN TOP CORNER
    Prop<::unicode::General_Category> { { 0x298E, 0x298E }, ::unicode::General_Category::Close_Punctuation }, // RIGHT SQUARE BRACKET WITH TICK IN BOTTOM CORNER
    Prop<::unicode::General_Category> { { 0x298F, 0x298F }, ::unicode::General_Category::Open_Punctuation }, // LEFT SQUARE BRACKET WITH TICK IN BOTTOM CORNER
    Prop<::unicode::General_Category> { { 0x2990, 0x2990 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT SQUARE BRACKET WITH TICK IN TOP CORNER
    Prop<::unicode::General_Category> { { 0x2991, 0x2991 }, ::unicode::General_Category::Open_Punctuation }, // LEFT ANGLE BRACKET WITH DOT
    Prop<::unicode::General_Category> { { 0x2992, 0x2992 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT ANGLE BRACKET WITH DOT
    Prop<::unicode::General_Category> { { 0x2993, 0x2993 }, ::unicode::General_Category::Open_Punctuation }, // LEFT ARC LESS-THAN BRACKET
    Prop<::unicode::General_Category> { { 0x2994, 0x2994 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT ARC GREATER-THAN BRACKET
    Prop<::unicode::General_Category> { { 0x2995, 0x2995 }, ::unicode::General_Category::Open_Punctuation }, // DOUBLE LEFT ARC GREATER-THAN BRACKET
    Prop<::unicode::General_Category> { { 0x2996, 0x2996 }, ::unicode::General_Category::Close_Punctuation }, // DOUBLE RIGHT ARC LESS-THAN BRACKET
    Prop<::unicode::General_Category> { { 0x2997, 0x2997 }, ::unicode::General_Category::Open_Punctuation }, // LEFT BLACK TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0x2998, 0x2998 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT BLACK TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0x2999, 0x29D7 }, ::unicode::General_Category::Math_Symbol }, // [63] DOTTED FENCE..BLACK HOURGLASS
    Prop<::unicode::General_Category> { { 0x29D8, 0x29D8 }, ::unicode::General_Category::Open_Punctuation }, // LEFT WIGGLY FENCE
    Prop<::unicode::General_Category> { { 0x29D9, 0x29D9 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT WIGGLY FENCE
    Prop<::unicode::General_Category> { { 0x29DA, 0x29DA }, ::unicode::General_Category::Open_Punctuation }, // LEFT DOUBLE WIGGLY FENCE
    Prop<::unicode::General_Category> { { 0x29DB, 0x29DB }, ::unicode::General_Category::Close_Punctuation }, // RIGHT DOUBLE WIGGLY FENCE
    Prop<::unicode::General_Category> { { 0x29DC, 0x29FB }, ::unicode::General_Category::Math_Symbol }, // [32] INCOMPLETE INFINITY..TRIPLE PLUS
    Prop<::unicode::General_Category> { { 0x29FC, 0x29FC }, ::unicode::General_Category::Open_Punctuation }, // LEFT-POINTING CURVED ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0x29FD, 0x29FD }, ::unicode::General_Category::Close_Punctuation }, // RIGHT-POINTING CURVED ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0x29FE, 0x2AFF }, ::unicode::General_Category::Math_Symbol }, // [258] TINY..N-ARY WHITE VERTICAL BAR
    Prop<::unicode::General_Category> { { 0x2B00, 0x2B2F }, ::unicode::General_Category::Other_Symbol }, // [48] NORTH EAST WHITE ARROW..WHITE VERTICAL ELLIPSE
    Prop<::unicode::General_Category> { { 0x2B30, 0x2B44 }, ::unicode::General_Category::Math_Symbol }, // [21] LEFT ARROW WITH SMALL CIRCLE..RIGHTWARDS ARROW THROUGH SUPERSET
    Prop<::unicode::General_Category> { { 0x2B45, 0x2B46 }, ::unicode::General_Category::Other_Symbol }, // [2] LEFTWARDS QUADRUPLE ARROW..RIGHTWARDS QUADRUPLE ARROW
    Prop<::unicode::General_Category> { { 0x2B47, 0x2B4C }, ::unicode::General_Category::Math_Symbol }, // [6] REVERSE TILDE OPERATOR ABOVE RIGHTWARDS ARROW..RIGHTWARDS ARROW ABOVE REVERSE TILDE OPERATOR
    Prop<::unicode::General_Category> { { 0x2B4D, 0x2B73 }, ::unicode::General_Category::Other_Symbol }, // [39] DOWNWARDS TRIANGLE-HEADED ZIGZAG ARROW..DOWNWARDS TRIANGLE-HEADED ARROW TO BAR
    Prop<::unicode::General_Category> { { 0x2B74, 0x2B75 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-2B74>..<reserved-2B75>
    Prop<::unicode::General_Category> { { 0x2B76, 0x2BFF }, ::unicode::General_Category::Other_Symbol }, // [138] NORTH WEST TRIANGLE-HEADED ARROW TO BAR..HELLSCHREIBER PAUSE SYMBOL
    Prop<::unicode::General_Category> { { 0x2C00, 0x2C2F }, ::unicode::General_Category::Uppercase_Letter }, // [48] GLAGOLITIC CAPITAL LETTER AZU..GLAGOLITIC CAPITAL LETTER CAUDATE CHRIVI
    Prop<::unicode::General_Category> { { 0x2C30, 0x2C5F }, ::unicode::General_Category::Lowercase_Letter }, // [48] GLAGOLITIC SMALL LETTER AZU..GLAGOLITIC SMALL LETTER CAUDATE CHRIVI
    Prop<::unicode::General_Category> { { 0x2C60, 0x2C60 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER L WITH DOUBLE BAR
    Prop<::unicode::General_Category> { { 0x2C61, 0x2C61 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER L WITH DOUBLE BAR
    Prop<::unicode::General_Category> { { 0x2C62, 0x2C64 }, ::unicode::General_Category::Uppercase_Letter }, // [3] LATIN CAPITAL LETTER L WITH MIDDLE TILDE..LATIN CAPITAL LETTER R WITH TAIL
    Prop<::unicode::General_Category> { { 0x2C65, 0x2C66 }, ::unicode::General_Category::Lowercase_Letter }, // [2] LATIN SMALL LETTER A WITH STROKE..LATIN SMALL LETTER T WITH DIAGONAL STROKE
    Prop<::unicode::General_Category> { { 0x2C67, 0x2C67 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER H WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x2C68, 0x2C68 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER H WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x2C69, 0x2C69 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER K WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x2C6A, 0x2C6A }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER K WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x2C6B, 0x2C6B }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Z WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x2C6C, 0x2C6C }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Z WITH DESCENDER
    Prop<::unicode::General_Category> { { 0x2C6D, 0x2C70 }, ::unicode::General_Category::Uppercase_Letter }, // [4] LATIN CAPITAL LETTER ALPHA..LATIN CAPITAL LETTER TURNED ALPHA
    Prop<::unicode::General_Category> { { 0x2C71, 0x2C71 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER V WITH RIGHT HOOK
    Prop<::unicode::General_Category> { { 0x2C72, 0x2C72 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER W WITH HOOK
    Prop<::unicode::General_Category> { { 0x2C73, 0x2C74 }, ::unicode::General_Category::Lowercase_Letter }, // [2] LATIN SMALL LETTER W WITH HOOK..LATIN SMALL LETTER V WITH CURL
    Prop<::unicode::General_Category> { { 0x2C75, 0x2C75 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER HALF H
    Prop<::unicode::General_Category> { { 0x2C76, 0x2C7B }, ::unicode::General_Category::Lowercase_Letter }, // [6] LATIN SMALL LETTER HALF H..LATIN LETTER SMALL CAPITAL TURNED E
    Prop<::unicode::General_Category> { { 0x2C7C, 0x2C7D }, ::unicode::General_Category::Modifier_Letter }, // [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    Prop<::unicode::General_Category> { { 0x2C7E, 0x2C80 }, ::unicode::General_Category::Uppercase_Letter }, // [3] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC CAPITAL LETTER ALFA
    Prop<::unicode::General_Category> { { 0x2C81, 0x2C81 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER ALFA
    Prop<::unicode::General_Category> { { 0x2C82, 0x2C82 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER VIDA
    Prop<::unicode::General_Category> { { 0x2C83, 0x2C83 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER VIDA
    Prop<::unicode::General_Category> { { 0x2C84, 0x2C84 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER GAMMA
    Prop<::unicode::General_Category> { { 0x2C85, 0x2C85 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER GAMMA
    Prop<::unicode::General_Category> { { 0x2C86, 0x2C86 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER DALDA
    Prop<::unicode::General_Category> { { 0x2C87, 0x2C87 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER DALDA
    Prop<::unicode::General_Category> { { 0x2C88, 0x2C88 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER EIE
    Prop<::unicode::General_Category> { { 0x2C89, 0x2C89 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER EIE
    Prop<::unicode::General_Category> { { 0x2C8A, 0x2C8A }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER SOU
    Prop<::unicode::General_Category> { { 0x2C8B, 0x2C8B }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER SOU
    Prop<::unicode::General_Category> { { 0x2C8C, 0x2C8C }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER ZATA
    Prop<::unicode::General_Category> { { 0x2C8D, 0x2C8D }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER ZATA
    Prop<::unicode::General_Category> { { 0x2C8E, 0x2C8E }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER HATE
    Prop<::unicode::General_Category> { { 0x2C8F, 0x2C8F }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER HATE
    Prop<::unicode::General_Category> { { 0x2C90, 0x2C90 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER THETHE
    Prop<::unicode::General_Category> { { 0x2C91, 0x2C91 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER THETHE
    Prop<::unicode::General_Category> { { 0x2C92, 0x2C92 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER IAUDA
    Prop<::unicode::General_Category> { { 0x2C93, 0x2C93 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER IAUDA
    Prop<::unicode::General_Category> { { 0x2C94, 0x2C94 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER KAPA
    Prop<::unicode::General_Category> { { 0x2C95, 0x2C95 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER KAPA
    Prop<::unicode::General_Category> { { 0x2C96, 0x2C96 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER LAULA
    Prop<::unicode::General_Category> { { 0x2C97, 0x2C97 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER LAULA
    Prop<::unicode::General_Category> { { 0x2C98, 0x2C98 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER MI
    Prop<::unicode::General_Category> { { 0x2C99, 0x2C99 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER MI
    Prop<::unicode::General_Category> { { 0x2C9A, 0x2C9A }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER NI
    Prop<::unicode::General_Category> { { 0x2C9B, 0x2C9B }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER NI
    Prop<::unicode::General_Category> { { 0x2C9C, 0x2C9C }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER KSI
    Prop<::unicode::General_Category> { { 0x2C9D, 0x2C9D }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER KSI
    Prop<::unicode::General_Category> { { 0x2C9E, 0x2C9E }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER O
    Prop<::unicode::General_Category> { { 0x2C9F, 0x2C9F }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER O
    Prop<::unicode::General_Category> { { 0x2CA0, 0x2CA0 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER PI
    Prop<::unicode::General_Category> { { 0x2CA1, 0x2CA1 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER PI
    Prop<::unicode::General_Category> { { 0x2CA2, 0x2CA2 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER RO
    Prop<::unicode::General_Category> { { 0x2CA3, 0x2CA3 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER RO
    Prop<::unicode::General_Category> { { 0x2CA4, 0x2CA4 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER SIMA
    Prop<::unicode::General_Category> { { 0x2CA5, 0x2CA5 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER SIMA
    Prop<::unicode::General_Category> { { 0x2CA6, 0x2CA6 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER TAU
    Prop<::unicode::General_Category> { { 0x2CA7, 0x2CA7 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER TAU
    Prop<::unicode::General_Category> { { 0x2CA8, 0x2CA8 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER UA
    Prop<::unicode::General_Category> { { 0x2CA9, 0x2CA9 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER UA
    Prop<::unicode::General_Category> { { 0x2CAA, 0x2CAA }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER FI
    Prop<::unicode::General_Category> { { 0x2CAB, 0x2CAB }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER FI
    Prop<::unicode::General_Category> { { 0x2CAC, 0x2CAC }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER KHI
    Prop<::unicode::General_Category> { { 0x2CAD, 0x2CAD }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER KHI
    Prop<::unicode::General_Category> { { 0x2CAE, 0x2CAE }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER PSI
    Prop<::unicode::General_Category> { { 0x2CAF, 0x2CAF }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER PSI
    Prop<::unicode::General_Category> { { 0x2CB0, 0x2CB0 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OOU
    Prop<::unicode::General_Category> { { 0x2CB1, 0x2CB1 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OOU
    Prop<::unicode::General_Category> { { 0x2CB2, 0x2CB2 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER DIALECT-P ALEF
    Prop<::unicode::General_Category> { { 0x2CB3, 0x2CB3 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER DIALECT-P ALEF
    Prop<::unicode::General_Category> { { 0x2CB4, 0x2CB4 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD COPTIC AIN
    Prop<::unicode::General_Category> { { 0x2CB5, 0x2CB5 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD COPTIC AIN
    Prop<::unicode::General_Category> { { 0x2CB6, 0x2CB6 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER CRYPTOGRAMMIC EIE
    Prop<::unicode::General_Category> { { 0x2CB7, 0x2CB7 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER CRYPTOGRAMMIC EIE
    Prop<::unicode::General_Category> { { 0x2CB8, 0x2CB8 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER DIALECT-P KAPA
    Prop<::unicode::General_Category> { { 0x2CB9, 0x2CB9 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER DIALECT-P KAPA
    Prop<::unicode::General_Category> { { 0x2CBA, 0x2CBA }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER DIALECT-P NI
    Prop<::unicode::General_Category> { { 0x2CBB, 0x2CBB }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER DIALECT-P NI
    Prop<::unicode::General_Category> { { 0x2CBC, 0x2CBC }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER CRYPTOGRAMMIC NI
    Prop<::unicode::General_Category> { { 0x2CBD, 0x2CBD }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER CRYPTOGRAMMIC NI
    Prop<::unicode::General_Category> { { 0x2CBE, 0x2CBE }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD COPTIC OOU
    Prop<::unicode::General_Category> { { 0x2CBF, 0x2CBF }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD COPTIC OOU
    Prop<::unicode::General_Category> { { 0x2CC0, 0x2CC0 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER SAMPI
    Prop<::unicode::General_Category> { { 0x2CC1, 0x2CC1 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER SAMPI
    Prop<::unicode::General_Category> { { 0x2CC2, 0x2CC2 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER CROSSED SHEI
    Prop<::unicode::General_Category> { { 0x2CC3, 0x2CC3 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER CROSSED SHEI
    Prop<::unicode::General_Category> { { 0x2CC4, 0x2CC4 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD COPTIC SHEI
    Prop<::unicode::General_Category> { { 0x2CC5, 0x2CC5 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD COPTIC SHEI
    Prop<::unicode::General_Category> { { 0x2CC6, 0x2CC6 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD COPTIC ESH
    Prop<::unicode::General_Category> { { 0x2CC7, 0x2CC7 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD COPTIC ESH
    Prop<::unicode::General_Category> { { 0x2CC8, 0x2CC8 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER AKHMIMIC KHEI
    Prop<::unicode::General_Category> { { 0x2CC9, 0x2CC9 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER AKHMIMIC KHEI
    Prop<::unicode::General_Category> { { 0x2CCA, 0x2CCA }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER DIALECT-P HORI
    Prop<::unicode::General_Category> { { 0x2CCB, 0x2CCB }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER DIALECT-P HORI
    Prop<::unicode::General_Category> { { 0x2CCC, 0x2CCC }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD COPTIC HORI
    Prop<::unicode::General_Category> { { 0x2CCD, 0x2CCD }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD COPTIC HORI
    Prop<::unicode::General_Category> { { 0x2CCE, 0x2CCE }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD COPTIC HA
    Prop<::unicode::General_Category> { { 0x2CCF, 0x2CCF }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD COPTIC HA
    Prop<::unicode::General_Category> { { 0x2CD0, 0x2CD0 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER L-SHAPED HA
    Prop<::unicode::General_Category> { { 0x2CD1, 0x2CD1 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER L-SHAPED HA
    Prop<::unicode::General_Category> { { 0x2CD2, 0x2CD2 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD COPTIC HEI
    Prop<::unicode::General_Category> { { 0x2CD3, 0x2CD3 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD COPTIC HEI
    Prop<::unicode::General_Category> { { 0x2CD4, 0x2CD4 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD COPTIC HAT
    Prop<::unicode::General_Category> { { 0x2CD5, 0x2CD5 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD COPTIC HAT
    Prop<::unicode::General_Category> { { 0x2CD6, 0x2CD6 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD COPTIC GANGIA
    Prop<::unicode::General_Category> { { 0x2CD7, 0x2CD7 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD COPTIC GANGIA
    Prop<::unicode::General_Category> { { 0x2CD8, 0x2CD8 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD COPTIC DJA
    Prop<::unicode::General_Category> { { 0x2CD9, 0x2CD9 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD COPTIC DJA
    Prop<::unicode::General_Category> { { 0x2CDA, 0x2CDA }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD COPTIC SHIMA
    Prop<::unicode::General_Category> { { 0x2CDB, 0x2CDB }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD COPTIC SHIMA
    Prop<::unicode::General_Category> { { 0x2CDC, 0x2CDC }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD NUBIAN SHIMA
    Prop<::unicode::General_Category> { { 0x2CDD, 0x2CDD }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD NUBIAN SHIMA
    Prop<::unicode::General_Category> { { 0x2CDE, 0x2CDE }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD NUBIAN NGI
    Prop<::unicode::General_Category> { { 0x2CDF, 0x2CDF }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD NUBIAN NGI
    Prop<::unicode::General_Category> { { 0x2CE0, 0x2CE0 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD NUBIAN NYI
    Prop<::unicode::General_Category> { { 0x2CE1, 0x2CE1 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER OLD NUBIAN NYI
    Prop<::unicode::General_Category> { { 0x2CE2, 0x2CE2 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER OLD NUBIAN WAU
    Prop<::unicode::General_Category> { { 0x2CE3, 0x2CE4 }, ::unicode::General_Category::Lowercase_Letter }, // [2] COPTIC SMALL LETTER OLD NUBIAN WAU..COPTIC SYMBOL KAI
    Prop<::unicode::General_Category> { { 0x2CE5, 0x2CEA }, ::unicode::General_Category::Other_Symbol }, // [6] COPTIC SYMBOL MI RO..COPTIC SYMBOL SHIMA SIMA
    Prop<::unicode::General_Category> { { 0x2CEB, 0x2CEB }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI
    Prop<::unicode::General_Category> { { 0x2CEC, 0x2CEC }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER CRYPTOGRAMMIC SHEI
    Prop<::unicode::General_Category> { { 0x2CED, 0x2CED }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER CRYPTOGRAMMIC GANGIA
    Prop<::unicode::General_Category> { { 0x2CEE, 0x2CEE }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Prop<::unicode::General_Category> { { 0x2CEF, 0x2CF1 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] COPTIC COMBINING NI ABOVE..COPTIC COMBINING SPIRITUS LENIS
    Prop<::unicode::General_Category> { { 0x2CF2, 0x2CF2 }, ::unicode::General_Category::Uppercase_Letter }, // COPTIC CAPITAL LETTER BOHAIRIC KHEI
    Prop<::unicode::General_Category> { { 0x2CF3, 0x2CF3 }, ::unicode::General_Category::Lowercase_Letter }, // COPTIC SMALL LETTER BOHAIRIC KHEI
    Prop<::unicode::General_Category> { { 0x2CF4, 0x2CF8 }, ::unicode::General_Category::Unassigned }, // [5] <reserved-2CF4>..<reserved-2CF8>
    Prop<::unicode::General_Category> { { 0x2CF9, 0x2CFC }, ::unicode::General_Category::Other_Punctuation }, // [4] COPTIC OLD NUBIAN FULL STOP..COPTIC OLD NUBIAN VERSE DIVIDER
    Prop<::unicode::General_Category> { { 0x2CFD, 0x2CFD }, ::unicode::General_Category::Other_Number }, // COPTIC FRACTION ONE HALF
    Prop<::unicode::General_Category> { { 0x2CFE, 0x2CFF }, ::unicode::General_Category::Other_Punctuation }, // [2] COPTIC FULL STOP..COPTIC MORPHOLOGICAL DIVIDER
    Prop<::unicode::General_Category> { { 0x2D00, 0x2D25 }, ::unicode::General_Category::Lowercase_Letter }, // [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Prop<::unicode::General_Category> { { 0x2D26, 0x2D26 }, ::unicode::General_Category::Unassigned }, // <reserved-2D26>
    Prop<::unicode::General_Category> { { 0x2D27, 0x2D27 }, ::unicode::General_Category::Lowercase_Letter }, // GEORGIAN SMALL LETTER YN
    Prop<::unicode::General_Category> { { 0x2D28, 0x2D2C }, ::unicode::General_Category::Unassigned }, // [5] <reserved-2D28>..<reserved-2D2C>
    Prop<::unicode::General_Category> { { 0x2D2D, 0x2D2D }, ::unicode::General_Category::Lowercase_Letter }, // GEORGIAN SMALL LETTER AEN
    Prop<::unicode::General_Category> { { 0x2D2E, 0x2D2F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-2D2E>..<reserved-2D2F>
    Prop<::unicode::General_Category> { { 0x2D30, 0x2D67 }, ::unicode::General_Category::Other_Letter }, // [56] TIFINAGH LETTER YA..TIFINAGH LETTER YO
    Prop<::unicode::General_Category> { { 0x2D68, 0x2D6E }, ::unicode::General_Category::Unassigned }, // [7] <reserved-2D68>..<reserved-2D6E>
    Prop<::unicode::General_Category> { { 0x2D6F, 0x2D6F }, ::unicode::General_Category::Modifier_Letter }, // TIFINAGH MODIFIER LETTER LABIALIZATION MARK
    Prop<::unicode::General_Category> { { 0x2D70, 0x2D70 }, ::unicode::General_Category::Other_Punctuation }, // TIFINAGH SEPARATOR MARK
    Prop<::unicode::General_Category> { { 0x2D71, 0x2D7E }, ::unicode::General_Category::Unassigned }, // [14] <reserved-2D71>..<reserved-2D7E>
    Prop<::unicode::General_Category> { { 0x2D7F, 0x2D7F }, ::unicode::General_Category::Nonspacing_Mark }, // TIFINAGH CONSONANT JOINER
    Prop<::unicode::General_Category> { { 0x2D80, 0x2D96 }, ::unicode::General_Category::Other_Letter }, // [23] ETHIOPIC SYLLABLE LOA..ETHIOPIC SYLLABLE GGWE
    Prop<::unicode::General_Category> { { 0x2D97, 0x2D9F }, ::unicode::General_Category::Unassigned }, // [9] <reserved-2D97>..<reserved-2D9F>
    Prop<::unicode::General_Category> { { 0x2DA0, 0x2DA6 }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE SSA..ETHIOPIC SYLLABLE SSO
    Prop<::unicode::General_Category> { { 0x2DA7, 0x2DA7 }, ::unicode::General_Category::Unassigned }, // <reserved-2DA7>
    Prop<::unicode::General_Category> { { 0x2DA8, 0x2DAE }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE CCA..ETHIOPIC SYLLABLE CCO
    Prop<::unicode::General_Category> { { 0x2DAF, 0x2DAF }, ::unicode::General_Category::Unassigned }, // <reserved-2DAF>
    Prop<::unicode::General_Category> { { 0x2DB0, 0x2DB6 }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE ZZA..ETHIOPIC SYLLABLE ZZO
    Prop<::unicode::General_Category> { { 0x2DB7, 0x2DB7 }, ::unicode::General_Category::Unassigned }, // <reserved-2DB7>
    Prop<::unicode::General_Category> { { 0x2DB8, 0x2DBE }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE CCHA..ETHIOPIC SYLLABLE CCHO
    Prop<::unicode::General_Category> { { 0x2DBF, 0x2DBF }, ::unicode::General_Category::Unassigned }, // <reserved-2DBF>
    Prop<::unicode::General_Category> { { 0x2DC0, 0x2DC6 }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE QYA..ETHIOPIC SYLLABLE QYO
    Prop<::unicode::General_Category> { { 0x2DC7, 0x2DC7 }, ::unicode::General_Category::Unassigned }, // <reserved-2DC7>
    Prop<::unicode::General_Category> { { 0x2DC8, 0x2DCE }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE KYA..ETHIOPIC SYLLABLE KYO
    Prop<::unicode::General_Category> { { 0x2DCF, 0x2DCF }, ::unicode::General_Category::Unassigned }, // <reserved-2DCF>
    Prop<::unicode::General_Category> { { 0x2DD0, 0x2DD6 }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE XYA..ETHIOPIC SYLLABLE XYO
    Prop<::unicode::General_Category> { { 0x2DD7, 0x2DD7 }, ::unicode::General_Category::Unassigned }, // <reserved-2DD7>
    Prop<::unicode::General_Category> { { 0x2DD8, 0x2DDE }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE GYA..ETHIOPIC SYLLABLE GYO
    Prop<::unicode::General_Category> { { 0x2DDF, 0x2DDF }, ::unicode::General_Category::Unassigned }, // <reserved-2DDF>
    Prop<::unicode::General_Category> { { 0x2DE0, 0x2DFF }, ::unicode::General_Category::Nonspacing_Mark }, // [32] COMBINING CYRILLIC LETTER BE..COMBINING CYRILLIC LETTER IOTIFIED BIG YUS
    Prop<::unicode::General_Category> { { 0x2E00, 0x2E01 }, ::unicode::General_Category::Other_Punctuation }, // [2] RIGHT ANGLE SUBSTITUTION MARKER..RIGHT ANGLE DOTTED SUBSTITUTION MARKER
    Prop<::unicode::General_Category> { { 0x2E02, 0x2E02 }, ::unicode::General_Category::Initial_Punctuation }, // LEFT SUBSTITUTION BRACKET
    Prop<::unicode::General_Category> { { 0x2E03, 0x2E03 }, ::unicode::General_Category::Final_Punctuation }, // RIGHT SUBSTITUTION BRACKET
    Prop<::unicode::General_Category> { { 0x2E04, 0x2E04 }, ::unicode::General_Category::Initial_Punctuation }, // LEFT DOTTED SUBSTITUTION BRACKET
    Prop<::unicode::General_Category> { { 0x2E05, 0x2E05 }, ::unicode::General_Category::Final_Punctuation }, // RIGHT DOTTED SUBSTITUTION BRACKET
    Prop<::unicode::General_Category> { { 0x2E06, 0x2E08 }, ::unicode::General_Category::Other_Punctuation }, // [3] RAISED INTERPOLATION MARKER..DOTTED TRANSPOSITION MARKER
    Prop<::unicode::General_Category> { { 0x2E09, 0x2E09 }, ::unicode::General_Category::Initial_Punctuation }, // LEFT TRANSPOSITION BRACKET
    Prop<::unicode::General_Category> { { 0x2E0A, 0x2E0A }, ::unicode::General_Category::Final_Punctuation }, // RIGHT TRANSPOSITION BRACKET
    Prop<::unicode::General_Category> { { 0x2E0B, 0x2E0B }, ::unicode::General_Category::Other_Punctuation }, // RAISED SQUARE
    Prop<::unicode::General_Category> { { 0x2E0C, 0x2E0C }, ::unicode::General_Category::Initial_Punctuation }, // LEFT RAISED OMISSION BRACKET
    Prop<::unicode::General_Category> { { 0x2E0D, 0x2E0D }, ::unicode::General_Category::Final_Punctuation }, // RIGHT RAISED OMISSION BRACKET
    Prop<::unicode::General_Category> { { 0x2E0E, 0x2E16 }, ::unicode::General_Category::Other_Punctuation }, // [9] EDITORIAL CORONIS..DOTTED RIGHT-POINTING ANGLE
    Prop<::unicode::General_Category> { { 0x2E17, 0x2E17 }, ::unicode::General_Category::Dash_Punctuation }, // DOUBLE OBLIQUE HYPHEN
    Prop<::unicode::General_Category> { { 0x2E18, 0x2E19 }, ::unicode::General_Category::Other_Punctuation }, // [2] INVERTED INTERROBANG..PALM BRANCH
    Prop<::unicode::General_Category> { { 0x2E1A, 0x2E1A }, ::unicode::General_Category::Dash_Punctuation }, // HYPHEN WITH DIAERESIS
    Prop<::unicode::General_Category> { { 0x2E1B, 0x2E1B }, ::unicode::General_Category::Other_Punctuation }, // TILDE WITH RING ABOVE
    Prop<::unicode::General_Category> { { 0x2E1C, 0x2E1C }, ::unicode::General_Category::Initial_Punctuation }, // LEFT LOW PARAPHRASE BRACKET
    Prop<::unicode::General_Category> { { 0x2E1D, 0x2E1D }, ::unicode::General_Category::Final_Punctuation }, // RIGHT LOW PARAPHRASE BRACKET
    Prop<::unicode::General_Category> { { 0x2E1E, 0x2E1F }, ::unicode::General_Category::Other_Punctuation }, // [2] TILDE WITH DOT ABOVE..TILDE WITH DOT BELOW
    Prop<::unicode::General_Category> { { 0x2E20, 0x2E20 }, ::unicode::General_Category::Initial_Punctuation }, // LEFT VERTICAL BAR WITH QUILL
    Prop<::unicode::General_Category> { { 0x2E21, 0x2E21 }, ::unicode::General_Category::Final_Punctuation }, // RIGHT VERTICAL BAR WITH QUILL
    Prop<::unicode::General_Category> { { 0x2E22, 0x2E22 }, ::unicode::General_Category::Open_Punctuation }, // TOP LEFT HALF BRACKET
    Prop<::unicode::General_Category> { { 0x2E23, 0x2E23 }, ::unicode::General_Category::Close_Punctuation }, // TOP RIGHT HALF BRACKET
    Prop<::unicode::General_Category> { { 0x2E24, 0x2E24 }, ::unicode::General_Category::Open_Punctuation }, // BOTTOM LEFT HALF BRACKET
    Prop<::unicode::General_Category> { { 0x2E25, 0x2E25 }, ::unicode::General_Category::Close_Punctuation }, // BOTTOM RIGHT HALF BRACKET
    Prop<::unicode::General_Category> { { 0x2E26, 0x2E26 }, ::unicode::General_Category::Open_Punctuation }, // LEFT SIDEWAYS U BRACKET
    Prop<::unicode::General_Category> { { 0x2E27, 0x2E27 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT SIDEWAYS U BRACKET
    Prop<::unicode::General_Category> { { 0x2E28, 0x2E28 }, ::unicode::General_Category::Open_Punctuation }, // LEFT DOUBLE PARENTHESIS
    Prop<::unicode::General_Category> { { 0x2E29, 0x2E29 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT DOUBLE PARENTHESIS
    Prop<::unicode::General_Category> { { 0x2E2A, 0x2E2E }, ::unicode::General_Category::Other_Punctuation }, // [5] TWO DOTS OVER ONE DOT PUNCTUATION..REVERSED QUESTION MARK
    Prop<::unicode::General_Category> { { 0x2E2F, 0x2E2F }, ::unicode::General_Category::Modifier_Letter }, // VERTICAL TILDE
    Prop<::unicode::General_Category> { { 0x2E30, 0x2E39 }, ::unicode::General_Category::Other_Punctuation }, // [10] RING POINT..TOP HALF SECTION SIGN
    Prop<::unicode::General_Category> { { 0x2E3A, 0x2E3B }, ::unicode::General_Category::Dash_Punctuation }, // [2] TWO-EM DASH..THREE-EM DASH
    Prop<::unicode::General_Category> { { 0x2E3C, 0x2E3F }, ::unicode::General_Category::Other_Punctuation }, // [4] STENOGRAPHIC FULL STOP..CAPITULUM
    Prop<::unicode::General_Category> { { 0x2E40, 0x2E40 }, ::unicode::General_Category::Dash_Punctuation }, // DOUBLE HYPHEN
    Prop<::unicode::General_Category> { { 0x2E41, 0x2E41 }, ::unicode::General_Category::Other_Punctuation }, // REVERSED COMMA
    Prop<::unicode::General_Category> { { 0x2E42, 0x2E42 }, ::unicode::General_Category::Open_Punctuation }, // DOUBLE LOW-REVERSED-9 QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x2E43, 0x2E4F }, ::unicode::General_Category::Other_Punctuation }, // [13] DASH WITH LEFT UPTURN..CORNISH VERSE DIVIDER
    Prop<::unicode::General_Category> { { 0x2E50, 0x2E51 }, ::unicode::General_Category::Other_Symbol }, // [2] CROSS PATTY WITH RIGHT CROSSBAR..CROSS PATTY WITH LEFT CROSSBAR
    Prop<::unicode::General_Category> { { 0x2E52, 0x2E54 }, ::unicode::General_Category::Other_Punctuation }, // [3] TIRONIAN SIGN CAPITAL ET..MEDIEVAL QUESTION MARK
    Prop<::unicode::General_Category> { { 0x2E55, 0x2E55 }, ::unicode::General_Category::Open_Punctuation }, // LEFT SQUARE BRACKET WITH STROKE
    Prop<::unicode::General_Category> { { 0x2E56, 0x2E56 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT SQUARE BRACKET WITH STROKE
    Prop<::unicode::General_Category> { { 0x2E57, 0x2E57 }, ::unicode::General_Category::Open_Punctuation }, // LEFT SQUARE BRACKET WITH DOUBLE STROKE
    Prop<::unicode::General_Category> { { 0x2E58, 0x2E58 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT SQUARE BRACKET WITH DOUBLE STROKE
    Prop<::unicode::General_Category> { { 0x2E59, 0x2E59 }, ::unicode::General_Category::Open_Punctuation }, // TOP HALF LEFT PARENTHESIS
    Prop<::unicode::General_Category> { { 0x2E5A, 0x2E5A }, ::unicode::General_Category::Close_Punctuation }, // TOP HALF RIGHT PARENTHESIS
    Prop<::unicode::General_Category> { { 0x2E5B, 0x2E5B }, ::unicode::General_Category::Open_Punctuation }, // BOTTOM HALF LEFT PARENTHESIS
    Prop<::unicode::General_Category> { { 0x2E5C, 0x2E5C }, ::unicode::General_Category::Close_Punctuation }, // BOTTOM HALF RIGHT PARENTHESIS
    Prop<::unicode::General_Category> { { 0x2E5D, 0x2E5D }, ::unicode::General_Category::Dash_Punctuation }, // OBLIQUE HYPHEN
    Prop<::unicode::General_Category> { { 0x2E5E, 0x2E7F }, ::unicode::General_Category::Unassigned }, // [34] <reserved-2E5E>..<reserved-2E7F>
    Prop<::unicode::General_Category> { { 0x2E80, 0x2E99 }, ::unicode::General_Category::Other_Symbol }, // [26] CJK RADICAL REPEAT..CJK RADICAL RAP
    Prop<::unicode::General_Category> { { 0x2E9A, 0x2E9A }, ::unicode::General_Category::Unassigned }, // <reserved-2E9A>
    Prop<::unicode::General_Category> { { 0x2E9B, 0x2EF3 }, ::unicode::General_Category::Other_Symbol }, // [89] CJK RADICAL CHOKE..CJK RADICAL C-SIMPLIFIED TURTLE
    Prop<::unicode::General_Category> { { 0x2EF4, 0x2EFF }, ::unicode::General_Category::Unassigned }, // [12] <reserved-2EF4>..<reserved-2EFF>
    Prop<::unicode::General_Category> { { 0x2F00, 0x2FD5 }, ::unicode::General_Category::Other_Symbol }, // [214] KANGXI RADICAL ONE..KANGXI RADICAL FLUTE
    Prop<::unicode::General_Category> { { 0x2FD6, 0x2FEF }, ::unicode::General_Category::Unassigned }, // [26] <reserved-2FD6>..<reserved-2FEF>
    Prop<::unicode::General_Category> { { 0x2FF0, 0x2FFF }, ::unicode::General_Category::Other_Symbol }, // [16] IDEOGRAPHIC DESCRIPTION CHARACTER LEFT TO RIGHT..IDEOGRAPHIC DESCRIPTION CHARACTER ROTATION
    Prop<::unicode::General_Category> { { 0x3000, 0x3000 }, ::unicode::General_Category::Space_Separator }, // IDEOGRAPHIC SPACE
    Prop<::unicode::General_Category> { { 0x3001, 0x3003 }, ::unicode::General_Category::Other_Punctuation }, // [3] IDEOGRAPHIC COMMA..DITTO MARK
    Prop<::unicode::General_Category> { { 0x3004, 0x3004 }, ::unicode::General_Category::Other_Symbol }, // JAPANESE INDUSTRIAL STANDARD SYMBOL
    Prop<::unicode::General_Category> { { 0x3005, 0x3005 }, ::unicode::General_Category::Modifier_Letter }, // IDEOGRAPHIC ITERATION MARK
    Prop<::unicode::General_Category> { { 0x3006, 0x3006 }, ::unicode::General_Category::Other_Letter }, // IDEOGRAPHIC CLOSING MARK
    Prop<::unicode::General_Category> { { 0x3007, 0x3007 }, ::unicode::General_Category::Letter_Number }, // IDEOGRAPHIC NUMBER ZERO
    Prop<::unicode::General_Category> { { 0x3008, 0x3008 }, ::unicode::General_Category::Open_Punctuation }, // LEFT ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0x3009, 0x3009 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0x300A, 0x300A }, ::unicode::General_Category::Open_Punctuation }, // LEFT DOUBLE ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0x300B, 0x300B }, ::unicode::General_Category::Close_Punctuation }, // RIGHT DOUBLE ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0x300C, 0x300C }, ::unicode::General_Category::Open_Punctuation }, // LEFT CORNER BRACKET
    Prop<::unicode::General_Category> { { 0x300D, 0x300D }, ::unicode::General_Category::Close_Punctuation }, // RIGHT CORNER BRACKET
    Prop<::unicode::General_Category> { { 0x300E, 0x300E }, ::unicode::General_Category::Open_Punctuation }, // LEFT WHITE CORNER BRACKET
    Prop<::unicode::General_Category> { { 0x300F, 0x300F }, ::unicode::General_Category::Close_Punctuation }, // RIGHT WHITE CORNER BRACKET
    Prop<::unicode::General_Category> { { 0x3010, 0x3010 }, ::unicode::General_Category::Open_Punctuation }, // LEFT BLACK LENTICULAR BRACKET
    Prop<::unicode::General_Category> { { 0x3011, 0x3011 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT BLACK LENTICULAR BRACKET
    Prop<::unicode::General_Category> { { 0x3012, 0x3013 }, ::unicode::General_Category::Other_Symbol }, // [2] POSTAL MARK..GETA MARK
    Prop<::unicode::General_Category> { { 0x3014, 0x3014 }, ::unicode::General_Category::Open_Punctuation }, // LEFT TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0x3015, 0x3015 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0x3016, 0x3016 }, ::unicode::General_Category::Open_Punctuation }, // LEFT WHITE LENTICULAR BRACKET
    Prop<::unicode::General_Category> { { 0x3017, 0x3017 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT WHITE LENTICULAR BRACKET
    Prop<::unicode::General_Category> { { 0x3018, 0x3018 }, ::unicode::General_Category::Open_Punctuation }, // LEFT WHITE TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0x3019, 0x3019 }, ::unicode::General_Category::Close_Punctuation }, // RIGHT WHITE TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0x301A, 0x301A }, ::unicode::General_Category::Open_Punctuation }, // LEFT WHITE SQUARE BRACKET
    Prop<::unicode::General_Category> { { 0x301B, 0x301B }, ::unicode::General_Category::Close_Punctuation }, // RIGHT WHITE SQUARE BRACKET
    Prop<::unicode::General_Category> { { 0x301C, 0x301C }, ::unicode::General_Category::Dash_Punctuation }, // WAVE DASH
    Prop<::unicode::General_Category> { { 0x301D, 0x301D }, ::unicode::General_Category::Open_Punctuation }, // REVERSED DOUBLE PRIME QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x301E, 0x301F }, ::unicode::General_Category::Close_Punctuation }, // [2] DOUBLE PRIME QUOTATION MARK..LOW DOUBLE PRIME QUOTATION MARK
    Prop<::unicode::General_Category> { { 0x3020, 0x3020 }, ::unicode::General_Category::Other_Symbol }, // POSTAL MARK FACE
    Prop<::unicode::General_Category> { { 0x3021, 0x3029 }, ::unicode::General_Category::Letter_Number }, // [9] HANGZHOU NUMERAL ONE..HANGZHOU NUMERAL NINE
    Prop<::unicode::General_Category> { { 0x302A, 0x302D }, ::unicode::General_Category::Nonspacing_Mark }, // [4] IDEOGRAPHIC LEVEL TONE MARK..IDEOGRAPHIC ENTERING TONE MARK
    Prop<::unicode::General_Category> { { 0x302E, 0x302F }, ::unicode::General_Category::Spacing_Mark }, // [2] HANGUL SINGLE DOT TONE MARK..HANGUL DOUBLE DOT TONE MARK
    Prop<::unicode::General_Category> { { 0x3030, 0x3030 }, ::unicode::General_Category::Dash_Punctuation }, // WAVY DASH
    Prop<::unicode::General_Category> { { 0x3031, 0x3035 }, ::unicode::General_Category::Modifier_Letter }, // [5] VERTICAL KANA REPEAT MARK..VERTICAL KANA REPEAT MARK LOWER HALF
    Prop<::unicode::General_Category> { { 0x3036, 0x3037 }, ::unicode::General_Category::Other_Symbol }, // [2] CIRCLED POSTAL MARK..IDEOGRAPHIC TELEGRAPH LINE FEED SEPARATOR SYMBOL
    Prop<::unicode::General_Category> { { 0x3038, 0x303A }, ::unicode::General_Category::Letter_Number }, // [3] HANGZHOU NUMERAL TEN..HANGZHOU NUMERAL THIRTY
    Prop<::unicode::General_Category> { { 0x303B, 0x303B }, ::unicode::General_Category::Modifier_Letter }, // VERTICAL IDEOGRAPHIC ITERATION MARK
    Prop<::unicode::General_Category> { { 0x303C, 0x303C }, ::unicode::General_Category::Other_Letter }, // MASU MARK
    Prop<::unicode::General_Category> { { 0x303D, 0x303D }, ::unicode::General_Category::Other_Punctuation }, // PART ALTERNATION MARK
    Prop<::unicode::General_Category> { { 0x303E, 0x303F }, ::unicode::General_Category::Other_Symbol }, // [2] IDEOGRAPHIC VARIATION INDICATOR..IDEOGRAPHIC HALF FILL SPACE
    Prop<::unicode::General_Category> { { 0x3040, 0x3040 }, ::unicode::General_Category::Unassigned }, // <reserved-3040>
    Prop<::unicode::General_Category> { { 0x3041, 0x3096 }, ::unicode::General_Category::Other_Letter }, // [86] HIRAGANA LETTER SMALL A..HIRAGANA LETTER SMALL KE
    Prop<::unicode::General_Category> { { 0x3097, 0x3098 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-3097>..<reserved-3098>
    Prop<::unicode::General_Category> { { 0x3099, 0x309A }, ::unicode::General_Category::Nonspacing_Mark }, // [2] COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK..COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Prop<::unicode::General_Category> { { 0x309B, 0x309C }, ::unicode::General_Category::Modifier_Symbol }, // [2] KATAKANA-HIRAGANA VOICED SOUND MARK..KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Prop<::unicode::General_Category> { { 0x309D, 0x309E }, ::unicode::General_Category::Modifier_Letter }, // [2] HIRAGANA ITERATION MARK..HIRAGANA VOICED ITERATION MARK
    Prop<::unicode::General_Category> { { 0x309F, 0x309F }, ::unicode::General_Category::Other_Letter }, // HIRAGANA DIGRAPH YORI
    Prop<::unicode::General_Category> { { 0x30A0, 0x30A0 }, ::unicode::General_Category::Dash_Punctuation }, // KATAKANA-HIRAGANA DOUBLE HYPHEN
    Prop<::unicode::General_Category> { { 0x30A1, 0x30FA }, ::unicode::General_Category::Other_Letter }, // [90] KATAKANA LETTER SMALL A..KATAKANA LETTER VO
    Prop<::unicode::General_Category> { { 0x30FB, 0x30FB }, ::unicode::General_Category::Other_Punctuation }, // KATAKANA MIDDLE DOT
    Prop<::unicode::General_Category> { { 0x30FC, 0x30FE }, ::unicode::General_Category::Modifier_Letter }, // [3] KATAKANA-HIRAGANA PROLONGED SOUND MARK..KATAKANA VOICED ITERATION MARK
    Prop<::unicode::General_Category> { { 0x30FF, 0x30FF }, ::unicode::General_Category::Other_Letter }, // KATAKANA DIGRAPH KOTO
    Prop<::unicode::General_Category> { { 0x3100, 0x3104 }, ::unicode::General_Category::Unassigned }, // [5] <reserved-3100>..<reserved-3104>
    Prop<::unicode::General_Category> { { 0x3105, 0x312F }, ::unicode::General_Category::Other_Letter }, // [43] BOPOMOFO LETTER B..BOPOMOFO LETTER NN
    Prop<::unicode::General_Category> { { 0x3130, 0x3130 }, ::unicode::General_Category::Unassigned }, // <reserved-3130>
    Prop<::unicode::General_Category> { { 0x3131, 0x318E }, ::unicode::General_Category::Other_Letter }, // [94] HANGUL LETTER KIYEOK..HANGUL LETTER ARAEAE
    Prop<::unicode::General_Category> { { 0x318F, 0x318F }, ::unicode::General_Category::Unassigned }, // <reserved-318F>
    Prop<::unicode::General_Category> { { 0x3190, 0x3191 }, ::unicode::General_Category::Other_Symbol }, // [2] IDEOGRAPHIC ANNOTATION LINKING MARK..IDEOGRAPHIC ANNOTATION REVERSE MARK
    Prop<::unicode::General_Category> { { 0x3192, 0x3195 }, ::unicode::General_Category::Other_Number }, // [4] IDEOGRAPHIC ANNOTATION ONE MARK..IDEOGRAPHIC ANNOTATION FOUR MARK
    Prop<::unicode::General_Category> { { 0x3196, 0x319F }, ::unicode::General_Category::Other_Symbol }, // [10] IDEOGRAPHIC ANNOTATION TOP MARK..IDEOGRAPHIC ANNOTATION MAN MARK
    Prop<::unicode::General_Category> { { 0x31A0, 0x31BF }, ::unicode::General_Category::Other_Letter }, // [32] BOPOMOFO LETTER BU..BOPOMOFO LETTER AH
    Prop<::unicode::General_Category> { { 0x31C0, 0x31E5 }, ::unicode::General_Category::Other_Symbol }, // [38] CJK STROKE T..CJK STROKE SZP
    Prop<::unicode::General_Category> { { 0x31E6, 0x31EE }, ::unicode::General_Category::Unassigned }, // [9] <reserved-31E6>..<reserved-31EE>
    Prop<::unicode::General_Category> { { 0x31EF, 0x31EF }, ::unicode::General_Category::Other_Symbol }, // IDEOGRAPHIC DESCRIPTION CHARACTER SUBTRACTION
    Prop<::unicode::General_Category> { { 0x31F0, 0x31FF }, ::unicode::General_Category::Other_Letter }, // [16] KATAKANA LETTER SMALL KU..KATAKANA LETTER SMALL RO
    Prop<::unicode::General_Category> { { 0x3200, 0x321E }, ::unicode::General_Category::Other_Symbol }, // [31] PARENTHESIZED HANGUL KIYEOK..PARENTHESIZED KOREAN CHARACTER O HU
    Prop<::unicode::General_Category> { { 0x321F, 0x321F }, ::unicode::General_Category::Unassigned }, // <reserved-321F>
    Prop<::unicode::General_Category> { { 0x3220, 0x3229 }, ::unicode::General_Category::Other_Number }, // [10] PARENTHESIZED IDEOGRAPH ONE..PARENTHESIZED IDEOGRAPH TEN
    Prop<::unicode::General_Category> { { 0x322A, 0x3247 }, ::unicode::General_Category::Other_Symbol }, // [30] PARENTHESIZED IDEOGRAPH MOON..CIRCLED IDEOGRAPH KOTO
    Prop<::unicode::General_Category> { { 0x3248, 0x324F }, ::unicode::General_Category::Other_Number }, // [8] CIRCLED NUMBER TEN ON BLACK SQUARE..CIRCLED NUMBER EIGHTY ON BLACK SQUARE
    Prop<::unicode::General_Category> { { 0x3250, 0x3250 }, ::unicode::General_Category::Other_Symbol }, // PARTNERSHIP SIGN
    Prop<::unicode::General_Category> { { 0x3251, 0x325F }, ::unicode::General_Category::Other_Number }, // [15] CIRCLED NUMBER TWENTY ONE..CIRCLED NUMBER THIRTY FIVE
    Prop<::unicode::General_Category> { { 0x3260, 0x327F }, ::unicode::General_Category::Other_Symbol }, // [32] CIRCLED HANGUL KIYEOK..KOREAN STANDARD SYMBOL
    Prop<::unicode::General_Category> { { 0x3280, 0x3289 }, ::unicode::General_Category::Other_Number }, // [10] CIRCLED IDEOGRAPH ONE..CIRCLED IDEOGRAPH TEN
    Prop<::unicode::General_Category> { { 0x328A, 0x32B0 }, ::unicode::General_Category::Other_Symbol }, // [39] CIRCLED IDEOGRAPH MOON..CIRCLED IDEOGRAPH NIGHT
    Prop<::unicode::General_Category> { { 0x32B1, 0x32BF }, ::unicode::General_Category::Other_Number }, // [15] CIRCLED NUMBER THIRTY SIX..CIRCLED NUMBER FIFTY
    Prop<::unicode::General_Category> { { 0x32C0, 0x33FF }, ::unicode::General_Category::Other_Symbol }, // [320] IDEOGRAPHIC TELEGRAPH SYMBOL FOR JANUARY..SQUARE GAL
    Prop<::unicode::General_Category> { { 0x3400, 0x4DBF }, ::unicode::General_Category::Other_Letter }, // [6592] CJK UNIFIED IDEOGRAPH-3400..CJK UNIFIED IDEOGRAPH-4DBF
    Prop<::unicode::General_Category> { { 0x4DC0, 0x4DFF }, ::unicode::General_Category::Other_Symbol }, // [64] HEXAGRAM FOR THE CREATIVE HEAVEN..HEXAGRAM FOR BEFORE COMPLETION
    Prop<::unicode::General_Category> { { 0x4E00, 0xA014 }, ::unicode::General_Category::Other_Letter }, // [21013] CJK UNIFIED IDEOGRAPH-4E00..YI SYLLABLE E
    Prop<::unicode::General_Category> { { 0xA015, 0xA015 }, ::unicode::General_Category::Modifier_Letter }, // YI SYLLABLE WU
    Prop<::unicode::General_Category> { { 0xA016, 0xA48C }, ::unicode::General_Category::Other_Letter }, // [1143] YI SYLLABLE BIT..YI SYLLABLE YYR
    Prop<::unicode::General_Category> { { 0xA48D, 0xA48F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-A48D>..<reserved-A48F>
    Prop<::unicode::General_Category> { { 0xA490, 0xA4C6 }, ::unicode::General_Category::Other_Symbol }, // [55] YI RADICAL QOT..YI RADICAL KE
    Prop<::unicode::General_Category> { { 0xA4C7, 0xA4CF }, ::unicode::General_Category::Unassigned }, // [9] <reserved-A4C7>..<reserved-A4CF>
    Prop<::unicode::General_Category> { { 0xA4D0, 0xA4F7 }, ::unicode::General_Category::Other_Letter }, // [40] LISU LETTER BA..LISU LETTER OE
    Prop<::unicode::General_Category> { { 0xA4F8, 0xA4FD }, ::unicode::General_Category::Modifier_Letter }, // [6] LISU LETTER TONE MYA TI..LISU LETTER TONE MYA JEU
    Prop<::unicode::General_Category> { { 0xA4FE, 0xA4FF }, ::unicode::General_Category::Other_Punctuation }, // [2] LISU PUNCTUATION COMMA..LISU PUNCTUATION FULL STOP
    Prop<::unicode::General_Category> { { 0xA500, 0xA60B }, ::unicode::General_Category::Other_Letter }, // [268] VAI SYLLABLE EE..VAI SYLLABLE NG
    Prop<::unicode::General_Category> { { 0xA60C, 0xA60C }, ::unicode::General_Category::Modifier_Letter }, // VAI SYLLABLE LENGTHENER
    Prop<::unicode::General_Category> { { 0xA60D, 0xA60F }, ::unicode::General_Category::Other_Punctuation }, // [3] VAI COMMA..VAI QUESTION MARK
    Prop<::unicode::General_Category> { { 0xA610, 0xA61F }, ::unicode::General_Category::Other_Letter }, // [16] VAI SYLLABLE NDOLE FA..VAI SYMBOL JONG
    Prop<::unicode::General_Category> { { 0xA620, 0xA629 }, ::unicode::General_Category::Decimal_Number }, // [10] VAI DIGIT ZERO..VAI DIGIT NINE
    Prop<::unicode::General_Category> { { 0xA62A, 0xA62B }, ::unicode::General_Category::Other_Letter }, // [2] VAI SYLLABLE NDOLE MA..VAI SYLLABLE NDOLE DO
    Prop<::unicode::General_Category> { { 0xA62C, 0xA63F }, ::unicode::General_Category::Unassigned }, // [20] <reserved-A62C>..<reserved-A63F>
    Prop<::unicode::General_Category> { { 0xA640, 0xA640 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ZEMLYA
    Prop<::unicode::General_Category> { { 0xA641, 0xA641 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ZEMLYA
    Prop<::unicode::General_Category> { { 0xA642, 0xA642 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER DZELO
    Prop<::unicode::General_Category> { { 0xA643, 0xA643 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER DZELO
    Prop<::unicode::General_Category> { { 0xA644, 0xA644 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER REVERSED DZE
    Prop<::unicode::General_Category> { { 0xA645, 0xA645 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER REVERSED DZE
    Prop<::unicode::General_Category> { { 0xA646, 0xA646 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER IOTA
    Prop<::unicode::General_Category> { { 0xA647, 0xA647 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER IOTA
    Prop<::unicode::General_Category> { { 0xA648, 0xA648 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER DJERV
    Prop<::unicode::General_Category> { { 0xA649, 0xA649 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER DJERV
    Prop<::unicode::General_Category> { { 0xA64A, 0xA64A }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER MONOGRAPH UK
    Prop<::unicode::General_Category> { { 0xA64B, 0xA64B }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER MONOGRAPH UK
    Prop<::unicode::General_Category> { { 0xA64C, 0xA64C }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER BROAD OMEGA
    Prop<::unicode::General_Category> { { 0xA64D, 0xA64D }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER BROAD OMEGA
    Prop<::unicode::General_Category> { { 0xA64E, 0xA64E }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER NEUTRAL YER
    Prop<::unicode::General_Category> { { 0xA64F, 0xA64F }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER NEUTRAL YER
    Prop<::unicode::General_Category> { { 0xA650, 0xA650 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER YERU WITH BACK YER
    Prop<::unicode::General_Category> { { 0xA651, 0xA651 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER YERU WITH BACK YER
    Prop<::unicode::General_Category> { { 0xA652, 0xA652 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER IOTIFIED YAT
    Prop<::unicode::General_Category> { { 0xA653, 0xA653 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER IOTIFIED YAT
    Prop<::unicode::General_Category> { { 0xA654, 0xA654 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER REVERSED YU
    Prop<::unicode::General_Category> { { 0xA655, 0xA655 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER REVERSED YU
    Prop<::unicode::General_Category> { { 0xA656, 0xA656 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER IOTIFIED A
    Prop<::unicode::General_Category> { { 0xA657, 0xA657 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER IOTIFIED A
    Prop<::unicode::General_Category> { { 0xA658, 0xA658 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER CLOSED LITTLE YUS
    Prop<::unicode::General_Category> { { 0xA659, 0xA659 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER CLOSED LITTLE YUS
    Prop<::unicode::General_Category> { { 0xA65A, 0xA65A }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER BLENDED YUS
    Prop<::unicode::General_Category> { { 0xA65B, 0xA65B }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER BLENDED YUS
    Prop<::unicode::General_Category> { { 0xA65C, 0xA65C }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER IOTIFIED CLOSED LITTLE YUS
    Prop<::unicode::General_Category> { { 0xA65D, 0xA65D }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER IOTIFIED CLOSED LITTLE YUS
    Prop<::unicode::General_Category> { { 0xA65E, 0xA65E }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER YN
    Prop<::unicode::General_Category> { { 0xA65F, 0xA65F }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER YN
    Prop<::unicode::General_Category> { { 0xA660, 0xA660 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER REVERSED TSE
    Prop<::unicode::General_Category> { { 0xA661, 0xA661 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER REVERSED TSE
    Prop<::unicode::General_Category> { { 0xA662, 0xA662 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER SOFT DE
    Prop<::unicode::General_Category> { { 0xA663, 0xA663 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER SOFT DE
    Prop<::unicode::General_Category> { { 0xA664, 0xA664 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER SOFT EL
    Prop<::unicode::General_Category> { { 0xA665, 0xA665 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER SOFT EL
    Prop<::unicode::General_Category> { { 0xA666, 0xA666 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER SOFT EM
    Prop<::unicode::General_Category> { { 0xA667, 0xA667 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER SOFT EM
    Prop<::unicode::General_Category> { { 0xA668, 0xA668 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER MONOCULAR O
    Prop<::unicode::General_Category> { { 0xA669, 0xA669 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER MONOCULAR O
    Prop<::unicode::General_Category> { { 0xA66A, 0xA66A }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER BINOCULAR O
    Prop<::unicode::General_Category> { { 0xA66B, 0xA66B }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER BINOCULAR O
    Prop<::unicode::General_Category> { { 0xA66C, 0xA66C }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER DOUBLE MONOCULAR O
    Prop<::unicode::General_Category> { { 0xA66D, 0xA66D }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Prop<::unicode::General_Category> { { 0xA66E, 0xA66E }, ::unicode::General_Category::Other_Letter }, // CYRILLIC LETTER MULTIOCULAR O
    Prop<::unicode::General_Category> { { 0xA66F, 0xA66F }, ::unicode::General_Category::Nonspacing_Mark }, // COMBINING CYRILLIC VZMET
    Prop<::unicode::General_Category> { { 0xA670, 0xA672 }, ::unicode::General_Category::Enclosing_Mark }, // [3] COMBINING CYRILLIC TEN MILLIONS SIGN..COMBINING CYRILLIC THOUSAND MILLIONS SIGN
    Prop<::unicode::General_Category> { { 0xA673, 0xA673 }, ::unicode::General_Category::Other_Punctuation }, // SLAVONIC ASTERISK
    Prop<::unicode::General_Category> { { 0xA674, 0xA67D }, ::unicode::General_Category::Nonspacing_Mark }, // [10] COMBINING CYRILLIC LETTER UKRAINIAN IE..COMBINING CYRILLIC PAYEROK
    Prop<::unicode::General_Category> { { 0xA67E, 0xA67E }, ::unicode::General_Category::Other_Punctuation }, // CYRILLIC KAVYKA
    Prop<::unicode::General_Category> { { 0xA67F, 0xA67F }, ::unicode::General_Category::Modifier_Letter }, // CYRILLIC PAYEROK
    Prop<::unicode::General_Category> { { 0xA680, 0xA680 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER DWE
    Prop<::unicode::General_Category> { { 0xA681, 0xA681 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER DWE
    Prop<::unicode::General_Category> { { 0xA682, 0xA682 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER DZWE
    Prop<::unicode::General_Category> { { 0xA683, 0xA683 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER DZWE
    Prop<::unicode::General_Category> { { 0xA684, 0xA684 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER ZHWE
    Prop<::unicode::General_Category> { { 0xA685, 0xA685 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER ZHWE
    Prop<::unicode::General_Category> { { 0xA686, 0xA686 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER CCHE
    Prop<::unicode::General_Category> { { 0xA687, 0xA687 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER CCHE
    Prop<::unicode::General_Category> { { 0xA688, 0xA688 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER DZZE
    Prop<::unicode::General_Category> { { 0xA689, 0xA689 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER DZZE
    Prop<::unicode::General_Category> { { 0xA68A, 0xA68A }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER TE WITH MIDDLE HOOK
    Prop<::unicode::General_Category> { { 0xA68B, 0xA68B }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER TE WITH MIDDLE HOOK
    Prop<::unicode::General_Category> { { 0xA68C, 0xA68C }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER TWE
    Prop<::unicode::General_Category> { { 0xA68D, 0xA68D }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER TWE
    Prop<::unicode::General_Category> { { 0xA68E, 0xA68E }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER TSWE
    Prop<::unicode::General_Category> { { 0xA68F, 0xA68F }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER TSWE
    Prop<::unicode::General_Category> { { 0xA690, 0xA690 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER TSSE
    Prop<::unicode::General_Category> { { 0xA691, 0xA691 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER TSSE
    Prop<::unicode::General_Category> { { 0xA692, 0xA692 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER TCHE
    Prop<::unicode::General_Category> { { 0xA693, 0xA693 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER TCHE
    Prop<::unicode::General_Category> { { 0xA694, 0xA694 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER HWE
    Prop<::unicode::General_Category> { { 0xA695, 0xA695 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER HWE
    Prop<::unicode::General_Category> { { 0xA696, 0xA696 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER SHWE
    Prop<::unicode::General_Category> { { 0xA697, 0xA697 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER SHWE
    Prop<::unicode::General_Category> { { 0xA698, 0xA698 }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER DOUBLE O
    Prop<::unicode::General_Category> { { 0xA699, 0xA699 }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER DOUBLE O
    Prop<::unicode::General_Category> { { 0xA69A, 0xA69A }, ::unicode::General_Category::Uppercase_Letter }, // CYRILLIC CAPITAL LETTER CROSSED O
    Prop<::unicode::General_Category> { { 0xA69B, 0xA69B }, ::unicode::General_Category::Lowercase_Letter }, // CYRILLIC SMALL LETTER CROSSED O
    Prop<::unicode::General_Category> { { 0xA69C, 0xA69D }, ::unicode::General_Category::Modifier_Letter }, // [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    Prop<::unicode::General_Category> { { 0xA69E, 0xA69F }, ::unicode::General_Category::Nonspacing_Mark }, // [2] COMBINING CYRILLIC LETTER EF..COMBINING CYRILLIC LETTER IOTIFIED E
    Prop<::unicode::General_Category> { { 0xA6A0, 0xA6E5 }, ::unicode::General_Category::Other_Letter }, // [70] BAMUM LETTER A..BAMUM LETTER KI
    Prop<::unicode::General_Category> { { 0xA6E6, 0xA6EF }, ::unicode::General_Category::Letter_Number }, // [10] BAMUM LETTER MO..BAMUM LETTER KOGHOM
    Prop<::unicode::General_Category> { { 0xA6F0, 0xA6F1 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] BAMUM COMBINING MARK KOQNDON..BAMUM COMBINING MARK TUKWENTIS
    Prop<::unicode::General_Category> { { 0xA6F2, 0xA6F7 }, ::unicode::General_Category::Other_Punctuation }, // [6] BAMUM NJAEMLI..BAMUM QUESTION MARK
    Prop<::unicode::General_Category> { { 0xA6F8, 0xA6FF }, ::unicode::General_Category::Unassigned }, // [8] <reserved-A6F8>..<reserved-A6FF>
    Prop<::unicode::General_Category> { { 0xA700, 0xA716 }, ::unicode::General_Category::Modifier_Symbol }, // [23] MODIFIER LETTER CHINESE TONE YIN PING..MODIFIER LETTER EXTRA-LOW LEFT-STEM TONE BAR
    Prop<::unicode::General_Category> { { 0xA717, 0xA71F }, ::unicode::General_Category::Modifier_Letter }, // [9] MODIFIER LETTER DOT VERTICAL BAR..MODIFIER LETTER LOW INVERTED EXCLAMATION MARK
    Prop<::unicode::General_Category> { { 0xA720, 0xA721 }, ::unicode::General_Category::Modifier_Symbol }, // [2] MODIFIER LETTER STRESS AND HIGH TONE..MODIFIER LETTER STRESS AND LOW TONE
    Prop<::unicode::General_Category> { { 0xA722, 0xA722 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF
    Prop<::unicode::General_Category> { { 0xA723, 0xA723 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER EGYPTOLOGICAL ALEF
    Prop<::unicode::General_Category> { { 0xA724, 0xA724 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER EGYPTOLOGICAL AIN
    Prop<::unicode::General_Category> { { 0xA725, 0xA725 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER EGYPTOLOGICAL AIN
    Prop<::unicode::General_Category> { { 0xA726, 0xA726 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER HENG
    Prop<::unicode::General_Category> { { 0xA727, 0xA727 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER HENG
    Prop<::unicode::General_Category> { { 0xA728, 0xA728 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER TZ
    Prop<::unicode::General_Category> { { 0xA729, 0xA729 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER TZ
    Prop<::unicode::General_Category> { { 0xA72A, 0xA72A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER TRESILLO
    Prop<::unicode::General_Category> { { 0xA72B, 0xA72B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER TRESILLO
    Prop<::unicode::General_Category> { { 0xA72C, 0xA72C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER CUATRILLO
    Prop<::unicode::General_Category> { { 0xA72D, 0xA72D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER CUATRILLO
    Prop<::unicode::General_Category> { { 0xA72E, 0xA72E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER CUATRILLO WITH COMMA
    Prop<::unicode::General_Category> { { 0xA72F, 0xA731 }, ::unicode::General_Category::Lowercase_Letter }, // [3] LATIN SMALL LETTER CUATRILLO WITH COMMA..LATIN LETTER SMALL CAPITAL S
    Prop<::unicode::General_Category> { { 0xA732, 0xA732 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER AA
    Prop<::unicode::General_Category> { { 0xA733, 0xA733 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER AA
    Prop<::unicode::General_Category> { { 0xA734, 0xA734 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER AO
    Prop<::unicode::General_Category> { { 0xA735, 0xA735 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER AO
    Prop<::unicode::General_Category> { { 0xA736, 0xA736 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER AU
    Prop<::unicode::General_Category> { { 0xA737, 0xA737 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER AU
    Prop<::unicode::General_Category> { { 0xA738, 0xA738 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER AV
    Prop<::unicode::General_Category> { { 0xA739, 0xA739 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER AV
    Prop<::unicode::General_Category> { { 0xA73A, 0xA73A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER AV WITH HORIZONTAL BAR
    Prop<::unicode::General_Category> { { 0xA73B, 0xA73B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER AV WITH HORIZONTAL BAR
    Prop<::unicode::General_Category> { { 0xA73C, 0xA73C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER AY
    Prop<::unicode::General_Category> { { 0xA73D, 0xA73D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER AY
    Prop<::unicode::General_Category> { { 0xA73E, 0xA73E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER REVERSED C WITH DOT
    Prop<::unicode::General_Category> { { 0xA73F, 0xA73F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER REVERSED C WITH DOT
    Prop<::unicode::General_Category> { { 0xA740, 0xA740 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER K WITH STROKE
    Prop<::unicode::General_Category> { { 0xA741, 0xA741 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER K WITH STROKE
    Prop<::unicode::General_Category> { { 0xA742, 0xA742 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER K WITH DIAGONAL STROKE
    Prop<::unicode::General_Category> { { 0xA743, 0xA743 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER K WITH DIAGONAL STROKE
    Prop<::unicode::General_Category> { { 0xA744, 0xA744 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER K WITH STROKE AND DIAGONAL STROKE
    Prop<::unicode::General_Category> { { 0xA745, 0xA745 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER K WITH STROKE AND DIAGONAL STROKE
    Prop<::unicode::General_Category> { { 0xA746, 0xA746 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER BROKEN L
    Prop<::unicode::General_Category> { { 0xA747, 0xA747 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER BROKEN L
    Prop<::unicode::General_Category> { { 0xA748, 0xA748 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER L WITH HIGH STROKE
    Prop<::unicode::General_Category> { { 0xA749, 0xA749 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER L WITH HIGH STROKE
    Prop<::unicode::General_Category> { { 0xA74A, 0xA74A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH LONG STROKE OVERLAY
    Prop<::unicode::General_Category> { { 0xA74B, 0xA74B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH LONG STROKE OVERLAY
    Prop<::unicode::General_Category> { { 0xA74C, 0xA74C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER O WITH LOOP
    Prop<::unicode::General_Category> { { 0xA74D, 0xA74D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER O WITH LOOP
    Prop<::unicode::General_Category> { { 0xA74E, 0xA74E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER OO
    Prop<::unicode::General_Category> { { 0xA74F, 0xA74F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER OO
    Prop<::unicode::General_Category> { { 0xA750, 0xA750 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER P WITH STROKE THROUGH DESCENDER
    Prop<::unicode::General_Category> { { 0xA751, 0xA751 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER P WITH STROKE THROUGH DESCENDER
    Prop<::unicode::General_Category> { { 0xA752, 0xA752 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER P WITH FLOURISH
    Prop<::unicode::General_Category> { { 0xA753, 0xA753 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER P WITH FLOURISH
    Prop<::unicode::General_Category> { { 0xA754, 0xA754 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER P WITH SQUIRREL TAIL
    Prop<::unicode::General_Category> { { 0xA755, 0xA755 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER P WITH SQUIRREL TAIL
    Prop<::unicode::General_Category> { { 0xA756, 0xA756 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Q WITH STROKE THROUGH DESCENDER
    Prop<::unicode::General_Category> { { 0xA757, 0xA757 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Q WITH STROKE THROUGH DESCENDER
    Prop<::unicode::General_Category> { { 0xA758, 0xA758 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER Q WITH DIAGONAL STROKE
    Prop<::unicode::General_Category> { { 0xA759, 0xA759 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER Q WITH DIAGONAL STROKE
    Prop<::unicode::General_Category> { { 0xA75A, 0xA75A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER R ROTUNDA
    Prop<::unicode::General_Category> { { 0xA75B, 0xA75B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER R ROTUNDA
    Prop<::unicode::General_Category> { { 0xA75C, 0xA75C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER RUM ROTUNDA
    Prop<::unicode::General_Category> { { 0xA75D, 0xA75D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER RUM ROTUNDA
    Prop<::unicode::General_Category> { { 0xA75E, 0xA75E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER V WITH DIAGONAL STROKE
    Prop<::unicode::General_Category> { { 0xA75F, 0xA75F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER V WITH DIAGONAL STROKE
    Prop<::unicode::General_Category> { { 0xA760, 0xA760 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER VY
    Prop<::unicode::General_Category> { { 0xA761, 0xA761 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER VY
    Prop<::unicode::General_Category> { { 0xA762, 0xA762 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER VISIGOTHIC Z
    Prop<::unicode::General_Category> { { 0xA763, 0xA763 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER VISIGOTHIC Z
    Prop<::unicode::General_Category> { { 0xA764, 0xA764 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER THORN WITH STROKE
    Prop<::unicode::General_Category> { { 0xA765, 0xA765 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER THORN WITH STROKE
    Prop<::unicode::General_Category> { { 0xA766, 0xA766 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER THORN WITH STROKE THROUGH DESCENDER
    Prop<::unicode::General_Category> { { 0xA767, 0xA767 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER THORN WITH STROKE THROUGH DESCENDER
    Prop<::unicode::General_Category> { { 0xA768, 0xA768 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER VEND
    Prop<::unicode::General_Category> { { 0xA769, 0xA769 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER VEND
    Prop<::unicode::General_Category> { { 0xA76A, 0xA76A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER ET
    Prop<::unicode::General_Category> { { 0xA76B, 0xA76B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER ET
    Prop<::unicode::General_Category> { { 0xA76C, 0xA76C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER IS
    Prop<::unicode::General_Category> { { 0xA76D, 0xA76D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER IS
    Prop<::unicode::General_Category> { { 0xA76E, 0xA76E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER CON
    Prop<::unicode::General_Category> { { 0xA76F, 0xA76F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER CON
    Prop<::unicode::General_Category> { { 0xA770, 0xA770 }, ::unicode::General_Category::Modifier_Letter }, // MODIFIER LETTER US
    Prop<::unicode::General_Category> { { 0xA771, 0xA778 }, ::unicode::General_Category::Lowercase_Letter }, // [8] LATIN SMALL LETTER DUM..LATIN SMALL LETTER UM
    Prop<::unicode::General_Category> { { 0xA779, 0xA779 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER INSULAR D
    Prop<::unicode::General_Category> { { 0xA77A, 0xA77A }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER INSULAR D
    Prop<::unicode::General_Category> { { 0xA77B, 0xA77B }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER INSULAR F
    Prop<::unicode::General_Category> { { 0xA77C, 0xA77C }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER INSULAR F
    Prop<::unicode::General_Category> { { 0xA77D, 0xA77E }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN CAPITAL LETTER INSULAR G..LATIN CAPITAL LETTER TURNED INSULAR G
    Prop<::unicode::General_Category> { { 0xA77F, 0xA77F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER TURNED INSULAR G
    Prop<::unicode::General_Category> { { 0xA780, 0xA780 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER TURNED L
    Prop<::unicode::General_Category> { { 0xA781, 0xA781 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER TURNED L
    Prop<::unicode::General_Category> { { 0xA782, 0xA782 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER INSULAR R
    Prop<::unicode::General_Category> { { 0xA783, 0xA783 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER INSULAR R
    Prop<::unicode::General_Category> { { 0xA784, 0xA784 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER INSULAR S
    Prop<::unicode::General_Category> { { 0xA785, 0xA785 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER INSULAR S
    Prop<::unicode::General_Category> { { 0xA786, 0xA786 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER INSULAR T
    Prop<::unicode::General_Category> { { 0xA787, 0xA787 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER INSULAR T
    Prop<::unicode::General_Category> { { 0xA788, 0xA788 }, ::unicode::General_Category::Modifier_Letter }, // MODIFIER LETTER LOW CIRCUMFLEX ACCENT
    Prop<::unicode::General_Category> { { 0xA789, 0xA78A }, ::unicode::General_Category::Modifier_Symbol }, // [2] MODIFIER LETTER COLON..MODIFIER LETTER SHORT EQUALS SIGN
    Prop<::unicode::General_Category> { { 0xA78B, 0xA78B }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER SALTILLO
    Prop<::unicode::General_Category> { { 0xA78C, 0xA78C }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER SALTILLO
    Prop<::unicode::General_Category> { { 0xA78D, 0xA78D }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER TURNED H
    Prop<::unicode::General_Category> { { 0xA78E, 0xA78E }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER L WITH RETROFLEX HOOK AND BELT
    Prop<::unicode::General_Category> { { 0xA78F, 0xA78F }, ::unicode::General_Category::Other_Letter }, // LATIN LETTER SINOLOGICAL DOT
    Prop<::unicode::General_Category> { { 0xA790, 0xA790 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER N WITH DESCENDER
    Prop<::unicode::General_Category> { { 0xA791, 0xA791 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER N WITH DESCENDER
    Prop<::unicode::General_Category> { { 0xA792, 0xA792 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER C WITH BAR
    Prop<::unicode::General_Category> { { 0xA793, 0xA795 }, ::unicode::General_Category::Lowercase_Letter }, // [3] LATIN SMALL LETTER C WITH BAR..LATIN SMALL LETTER H WITH PALATAL HOOK
    Prop<::unicode::General_Category> { { 0xA796, 0xA796 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER B WITH FLOURISH
    Prop<::unicode::General_Category> { { 0xA797, 0xA797 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER B WITH FLOURISH
    Prop<::unicode::General_Category> { { 0xA798, 0xA798 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER F WITH STROKE
    Prop<::unicode::General_Category> { { 0xA799, 0xA799 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER F WITH STROKE
    Prop<::unicode::General_Category> { { 0xA79A, 0xA79A }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER VOLAPUK AE
    Prop<::unicode::General_Category> { { 0xA79B, 0xA79B }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER VOLAPUK AE
    Prop<::unicode::General_Category> { { 0xA79C, 0xA79C }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER VOLAPUK OE
    Prop<::unicode::General_Category> { { 0xA79D, 0xA79D }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER VOLAPUK OE
    Prop<::unicode::General_Category> { { 0xA79E, 0xA79E }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER VOLAPUK UE
    Prop<::unicode::General_Category> { { 0xA79F, 0xA79F }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER VOLAPUK UE
    Prop<::unicode::General_Category> { { 0xA7A0, 0xA7A0 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER G WITH OBLIQUE STROKE
    Prop<::unicode::General_Category> { { 0xA7A1, 0xA7A1 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER G WITH OBLIQUE STROKE
    Prop<::unicode::General_Category> { { 0xA7A2, 0xA7A2 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER K WITH OBLIQUE STROKE
    Prop<::unicode::General_Category> { { 0xA7A3, 0xA7A3 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER K WITH OBLIQUE STROKE
    Prop<::unicode::General_Category> { { 0xA7A4, 0xA7A4 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER N WITH OBLIQUE STROKE
    Prop<::unicode::General_Category> { { 0xA7A5, 0xA7A5 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER N WITH OBLIQUE STROKE
    Prop<::unicode::General_Category> { { 0xA7A6, 0xA7A6 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER R WITH OBLIQUE STROKE
    Prop<::unicode::General_Category> { { 0xA7A7, 0xA7A7 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER R WITH OBLIQUE STROKE
    Prop<::unicode::General_Category> { { 0xA7A8, 0xA7A8 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER S WITH OBLIQUE STROKE
    Prop<::unicode::General_Category> { { 0xA7A9, 0xA7A9 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH OBLIQUE STROKE
    Prop<::unicode::General_Category> { { 0xA7AA, 0xA7AE }, ::unicode::General_Category::Uppercase_Letter }, // [5] LATIN CAPITAL LETTER H WITH HOOK..LATIN CAPITAL LETTER SMALL CAPITAL I
    Prop<::unicode::General_Category> { { 0xA7AF, 0xA7AF }, ::unicode::General_Category::Lowercase_Letter }, // LATIN LETTER SMALL CAPITAL Q
    Prop<::unicode::General_Category> { { 0xA7B0, 0xA7B4 }, ::unicode::General_Category::Uppercase_Letter }, // [5] LATIN CAPITAL LETTER TURNED K..LATIN CAPITAL LETTER BETA
    Prop<::unicode::General_Category> { { 0xA7B5, 0xA7B5 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER BETA
    Prop<::unicode::General_Category> { { 0xA7B6, 0xA7B6 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER OMEGA
    Prop<::unicode::General_Category> { { 0xA7B7, 0xA7B7 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER OMEGA
    Prop<::unicode::General_Category> { { 0xA7B8, 0xA7B8 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER U WITH STROKE
    Prop<::unicode::General_Category> { { 0xA7B9, 0xA7B9 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER U WITH STROKE
    Prop<::unicode::General_Category> { { 0xA7BA, 0xA7BA }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER GLOTTAL A
    Prop<::unicode::General_Category> { { 0xA7BB, 0xA7BB }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER GLOTTAL A
    Prop<::unicode::General_Category> { { 0xA7BC, 0xA7BC }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER GLOTTAL I
    Prop<::unicode::General_Category> { { 0xA7BD, 0xA7BD }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER GLOTTAL I
    Prop<::unicode::General_Category> { { 0xA7BE, 0xA7BE }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER GLOTTAL U
    Prop<::unicode::General_Category> { { 0xA7BF, 0xA7BF }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER GLOTTAL U
    Prop<::unicode::General_Category> { { 0xA7C0, 0xA7C0 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER OLD POLISH O
    Prop<::unicode::General_Category> { { 0xA7C1, 0xA7C1 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER OLD POLISH O
    Prop<::unicode::General_Category> { { 0xA7C2, 0xA7C2 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER ANGLICANA W
    Prop<::unicode::General_Category> { { 0xA7C3, 0xA7C3 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER ANGLICANA W
    Prop<::unicode::General_Category> { { 0xA7C4, 0xA7C7 }, ::unicode::General_Category::Uppercase_Letter }, // [4] LATIN CAPITAL LETTER C WITH PALATAL HOOK..LATIN CAPITAL LETTER D WITH SHORT STROKE OVERLAY
    Prop<::unicode::General_Category> { { 0xA7C8, 0xA7C8 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER D WITH SHORT STROKE OVERLAY
    Prop<::unicode::General_Category> { { 0xA7C9, 0xA7C9 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER S WITH SHORT STROKE OVERLAY
    Prop<::unicode::General_Category> { { 0xA7CA, 0xA7CA }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH SHORT STROKE OVERLAY
    Prop<::unicode::General_Category> { { 0xA7CB, 0xA7CC }, ::unicode::General_Category::Uppercase_Letter }, // [2] LATIN CAPITAL LETTER RAMS HORN..LATIN CAPITAL LETTER S WITH DIAGONAL STROKE
    Prop<::unicode::General_Category> { { 0xA7CD, 0xA7CD }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER S WITH DIAGONAL STROKE
    Prop<::unicode::General_Category> { { 0xA7CE, 0xA7CE }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER PHARYNGEAL VOICED FRICATIVE
    Prop<::unicode::General_Category> { { 0xA7CF, 0xA7CF }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER PHARYNGEAL VOICED FRICATIVE
    Prop<::unicode::General_Category> { { 0xA7D0, 0xA7D0 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER CLOSED INSULAR G
    Prop<::unicode::General_Category> { { 0xA7D1, 0xA7D1 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER CLOSED INSULAR G
    Prop<::unicode::General_Category> { { 0xA7D2, 0xA7D2 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER DOUBLE THORN
    Prop<::unicode::General_Category> { { 0xA7D3, 0xA7D3 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER DOUBLE THORN
    Prop<::unicode::General_Category> { { 0xA7D4, 0xA7D4 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER DOUBLE WYNN
    Prop<::unicode::General_Category> { { 0xA7D5, 0xA7D5 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER DOUBLE WYNN
    Prop<::unicode::General_Category> { { 0xA7D6, 0xA7D6 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER MIDDLE SCOTS S
    Prop<::unicode::General_Category> { { 0xA7D7, 0xA7D7 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER MIDDLE SCOTS S
    Prop<::unicode::General_Category> { { 0xA7D8, 0xA7D8 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER SIGMOID S
    Prop<::unicode::General_Category> { { 0xA7D9, 0xA7D9 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER SIGMOID S
    Prop<::unicode::General_Category> { { 0xA7DA, 0xA7DA }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER LAMBDA
    Prop<::unicode::General_Category> { { 0xA7DB, 0xA7DB }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER LAMBDA
    Prop<::unicode::General_Category> { { 0xA7DC, 0xA7DC }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Prop<::unicode::General_Category> { { 0xA7DD, 0xA7F0 }, ::unicode::General_Category::Unassigned }, // [20] <reserved-A7DD>..<reserved-A7F0>
    Prop<::unicode::General_Category> { { 0xA7F1, 0xA7F4 }, ::unicode::General_Category::Modifier_Letter }, // [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    Prop<::unicode::General_Category> { { 0xA7F5, 0xA7F5 }, ::unicode::General_Category::Uppercase_Letter }, // LATIN CAPITAL LETTER REVERSED HALF H
    Prop<::unicode::General_Category> { { 0xA7F6, 0xA7F6 }, ::unicode::General_Category::Lowercase_Letter }, // LATIN SMALL LETTER REVERSED HALF H
    Prop<::unicode::General_Category> { { 0xA7F7, 0xA7F7 }, ::unicode::General_Category::Other_Letter }, // LATIN EPIGRAPHIC LETTER SIDEWAYS I
    Prop<::unicode::General_Category> { { 0xA7F8, 0xA7F9 }, ::unicode::General_Category::Modifier_Letter }, // [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    Prop<::unicode::General_Category> { { 0xA7FA, 0xA7FA }, ::unicode::General_Category::Lowercase_Letter }, // LATIN LETTER SMALL CAPITAL TURNED M
    Prop<::unicode::General_Category> { { 0xA7FB, 0xA801 }, ::unicode::General_Category::Other_Letter }, // [7] LATIN EPIGRAPHIC LETTER REVERSED F..SYLOTI NAGRI LETTER I
    Prop<::unicode::General_Category> { { 0xA802, 0xA802 }, ::unicode::General_Category::Nonspacing_Mark }, // SYLOTI NAGRI SIGN DVISVARA
    Prop<::unicode::General_Category> { { 0xA803, 0xA805 }, ::unicode::General_Category::Other_Letter }, // [3] SYLOTI NAGRI LETTER U..SYLOTI NAGRI LETTER O
    Prop<::unicode::General_Category> { { 0xA806, 0xA806 }, ::unicode::General_Category::Nonspacing_Mark }, // SYLOTI NAGRI SIGN HASANTA
    Prop<::unicode::General_Category> { { 0xA807, 0xA80A }, ::unicode::General_Category::Other_Letter }, // [4] SYLOTI NAGRI LETTER KO..SYLOTI NAGRI LETTER GHO
    Prop<::unicode::General_Category> { { 0xA80B, 0xA80B }, ::unicode::General_Category::Nonspacing_Mark }, // SYLOTI NAGRI SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0xA80C, 0xA822 }, ::unicode::General_Category::Other_Letter }, // [23] SYLOTI NAGRI LETTER CO..SYLOTI NAGRI LETTER HO
    Prop<::unicode::General_Category> { { 0xA823, 0xA824 }, ::unicode::General_Category::Spacing_Mark }, // [2] SYLOTI NAGRI VOWEL SIGN A..SYLOTI NAGRI VOWEL SIGN I
    Prop<::unicode::General_Category> { { 0xA825, 0xA826 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] SYLOTI NAGRI VOWEL SIGN U..SYLOTI NAGRI VOWEL SIGN E
    Prop<::unicode::General_Category> { { 0xA827, 0xA827 }, ::unicode::General_Category::Spacing_Mark }, // SYLOTI NAGRI VOWEL SIGN OO
    Prop<::unicode::General_Category> { { 0xA828, 0xA82B }, ::unicode::General_Category::Other_Symbol }, // [4] SYLOTI NAGRI POETRY MARK-1..SYLOTI NAGRI POETRY MARK-4
    Prop<::unicode::General_Category> { { 0xA82C, 0xA82C }, ::unicode::General_Category::Nonspacing_Mark }, // SYLOTI NAGRI SIGN ALTERNATE HASANTA
    Prop<::unicode::General_Category> { { 0xA82D, 0xA82F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-A82D>..<reserved-A82F>
    Prop<::unicode::General_Category> { { 0xA830, 0xA835 }, ::unicode::General_Category::Other_Number }, // [6] NORTH INDIC FRACTION ONE QUARTER..NORTH INDIC FRACTION THREE SIXTEENTHS
    Prop<::unicode::General_Category> { { 0xA836, 0xA837 }, ::unicode::General_Category::Other_Symbol }, // [2] NORTH INDIC QUARTER MARK..NORTH INDIC PLACEHOLDER MARK
    Prop<::unicode::General_Category> { { 0xA838, 0xA838 }, ::unicode::General_Category::Currency_Symbol }, // NORTH INDIC RUPEE MARK
    Prop<::unicode::General_Category> { { 0xA839, 0xA839 }, ::unicode::General_Category::Other_Symbol }, // NORTH INDIC QUANTITY MARK
    Prop<::unicode::General_Category> { { 0xA83A, 0xA83F }, ::unicode::General_Category::Unassigned }, // [6] <reserved-A83A>..<reserved-A83F>
    Prop<::unicode::General_Category> { { 0xA840, 0xA873 }, ::unicode::General_Category::Other_Letter }, // [52] PHAGS-PA LETTER KA..PHAGS-PA LETTER CANDRABINDU
    Prop<::unicode::General_Category> { { 0xA874, 0xA877 }, ::unicode::General_Category::Other_Punctuation }, // [4] PHAGS-PA SINGLE HEAD MARK..PHAGS-PA MARK DOUBLE SHAD
    Prop<::unicode::General_Category> { { 0xA878, 0xA87F }, ::unicode::General_Category::Unassigned }, // [8] <reserved-A878>..<reserved-A87F>
    Prop<::unicode::General_Category> { { 0xA880, 0xA881 }, ::unicode::General_Category::Spacing_Mark }, // [2] SAURASHTRA SIGN ANUSVARA..SAURASHTRA SIGN VISARGA
    Prop<::unicode::General_Category> { { 0xA882, 0xA8B3 }, ::unicode::General_Category::Other_Letter }, // [50] SAURASHTRA LETTER A..SAURASHTRA LETTER LLA
    Prop<::unicode::General_Category> { { 0xA8B4, 0xA8C3 }, ::unicode::General_Category::Spacing_Mark }, // [16] SAURASHTRA CONSONANT SIGN HAARU..SAURASHTRA VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0xA8C4, 0xA8C5 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] SAURASHTRA SIGN VIRAMA..SAURASHTRA SIGN CANDRABINDU
    Prop<::unicode::General_Category> { { 0xA8C6, 0xA8CD }, ::unicode::General_Category::Unassigned }, // [8] <reserved-A8C6>..<reserved-A8CD>
    Prop<::unicode::General_Category> { { 0xA8CE, 0xA8CF }, ::unicode::General_Category::Other_Punctuation }, // [2] SAURASHTRA DANDA..SAURASHTRA DOUBLE DANDA
    Prop<::unicode::General_Category> { { 0xA8D0, 0xA8D9 }, ::unicode::General_Category::Decimal_Number }, // [10] SAURASHTRA DIGIT ZERO..SAURASHTRA DIGIT NINE
    Prop<::unicode::General_Category> { { 0xA8DA, 0xA8DF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-A8DA>..<reserved-A8DF>
    Prop<::unicode::General_Category> { { 0xA8E0, 0xA8F1 }, ::unicode::General_Category::Nonspacing_Mark }, // [18] COMBINING DEVANAGARI DIGIT ZERO..COMBINING DEVANAGARI SIGN AVAGRAHA
    Prop<::unicode::General_Category> { { 0xA8F2, 0xA8F7 }, ::unicode::General_Category::Other_Letter }, // [6] DEVANAGARI SIGN SPACING CANDRABINDU..DEVANAGARI SIGN CANDRABINDU AVAGRAHA
    Prop<::unicode::General_Category> { { 0xA8F8, 0xA8FA }, ::unicode::General_Category::Other_Punctuation }, // [3] DEVANAGARI SIGN PUSHPIKA..DEVANAGARI CARET
    Prop<::unicode::General_Category> { { 0xA8FB, 0xA8FB }, ::unicode::General_Category::Other_Letter }, // DEVANAGARI HEADSTROKE
    Prop<::unicode::General_Category> { { 0xA8FC, 0xA8FC }, ::unicode::General_Category::Other_Punctuation }, // DEVANAGARI SIGN SIDDHAM
    Prop<::unicode::General_Category> { { 0xA8FD, 0xA8FE }, ::unicode::General_Category::Other_Letter }, // [2] DEVANAGARI JAIN OM..DEVANAGARI LETTER AY
    Prop<::unicode::General_Category> { { 0xA8FF, 0xA8FF }, ::unicode::General_Category::Nonspacing_Mark }, // DEVANAGARI VOWEL SIGN AY
    Prop<::unicode::General_Category> { { 0xA900, 0xA909 }, ::unicode::General_Category::Decimal_Number }, // [10] KAYAH LI DIGIT ZERO..KAYAH LI DIGIT NINE
    Prop<::unicode::General_Category> { { 0xA90A, 0xA925 }, ::unicode::General_Category::Other_Letter }, // [28] KAYAH LI LETTER KA..KAYAH LI LETTER OO
    Prop<::unicode::General_Category> { { 0xA926, 0xA92D }, ::unicode::General_Category::Nonspacing_Mark }, // [8] KAYAH LI VOWEL UE..KAYAH LI TONE CALYA PLOPHU
    Prop<::unicode::General_Category> { { 0xA92E, 0xA92F }, ::unicode::General_Category::Other_Punctuation }, // [2] KAYAH LI SIGN CWI..KAYAH LI SIGN SHYA
    Prop<::unicode::General_Category> { { 0xA930, 0xA946 }, ::unicode::General_Category::Other_Letter }, // [23] REJANG LETTER KA..REJANG LETTER A
    Prop<::unicode::General_Category> { { 0xA947, 0xA951 }, ::unicode::General_Category::Nonspacing_Mark }, // [11] REJANG VOWEL SIGN I..REJANG CONSONANT SIGN R
    Prop<::unicode::General_Category> { { 0xA952, 0xA953 }, ::unicode::General_Category::Spacing_Mark }, // [2] REJANG CONSONANT SIGN H..REJANG VIRAMA
    Prop<::unicode::General_Category> { { 0xA954, 0xA95E }, ::unicode::General_Category::Unassigned }, // [11] <reserved-A954>..<reserved-A95E>
    Prop<::unicode::General_Category> { { 0xA95F, 0xA95F }, ::unicode::General_Category::Other_Punctuation }, // REJANG SECTION MARK
    Prop<::unicode::General_Category> { { 0xA960, 0xA97C }, ::unicode::General_Category::Other_Letter }, // [29] HANGUL CHOSEONG TIKEUT-MIEUM..HANGUL CHOSEONG SSANGYEORINHIEUH
    Prop<::unicode::General_Category> { { 0xA97D, 0xA97F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-A97D>..<reserved-A97F>
    Prop<::unicode::General_Category> { { 0xA980, 0xA982 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] JAVANESE SIGN PANYANGGA..JAVANESE SIGN LAYAR
    Prop<::unicode::General_Category> { { 0xA983, 0xA983 }, ::unicode::General_Category::Spacing_Mark }, // JAVANESE SIGN WIGNYAN
    Prop<::unicode::General_Category> { { 0xA984, 0xA9B2 }, ::unicode::General_Category::Other_Letter }, // [47] JAVANESE LETTER A..JAVANESE LETTER HA
    Prop<::unicode::General_Category> { { 0xA9B3, 0xA9B3 }, ::unicode::General_Category::Nonspacing_Mark }, // JAVANESE SIGN CECAK TELU
    Prop<::unicode::General_Category> { { 0xA9B4, 0xA9B5 }, ::unicode::General_Category::Spacing_Mark }, // [2] JAVANESE VOWEL SIGN TARUNG..JAVANESE VOWEL SIGN TOLONG
    Prop<::unicode::General_Category> { { 0xA9B6, 0xA9B9 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] JAVANESE VOWEL SIGN WULU..JAVANESE VOWEL SIGN SUKU MENDUT
    Prop<::unicode::General_Category> { { 0xA9BA, 0xA9BB }, ::unicode::General_Category::Spacing_Mark }, // [2] JAVANESE VOWEL SIGN TALING..JAVANESE VOWEL SIGN DIRGA MURE
    Prop<::unicode::General_Category> { { 0xA9BC, 0xA9BD }, ::unicode::General_Category::Nonspacing_Mark }, // [2] JAVANESE VOWEL SIGN PEPET..JAVANESE CONSONANT SIGN KERET
    Prop<::unicode::General_Category> { { 0xA9BE, 0xA9C0 }, ::unicode::General_Category::Spacing_Mark }, // [3] JAVANESE CONSONANT SIGN PENGKAL..JAVANESE PANGKON
    Prop<::unicode::General_Category> { { 0xA9C1, 0xA9CD }, ::unicode::General_Category::Other_Punctuation }, // [13] JAVANESE LEFT RERENGGAN..JAVANESE TURNED PADA PISELEH
    Prop<::unicode::General_Category> { { 0xA9CE, 0xA9CE }, ::unicode::General_Category::Unassigned }, // <reserved-A9CE>
    Prop<::unicode::General_Category> { { 0xA9CF, 0xA9CF }, ::unicode::General_Category::Modifier_Letter }, // JAVANESE PANGRANGKEP
    Prop<::unicode::General_Category> { { 0xA9D0, 0xA9D9 }, ::unicode::General_Category::Decimal_Number }, // [10] JAVANESE DIGIT ZERO..JAVANESE DIGIT NINE
    Prop<::unicode::General_Category> { { 0xA9DA, 0xA9DD }, ::unicode::General_Category::Unassigned }, // [4] <reserved-A9DA>..<reserved-A9DD>
    Prop<::unicode::General_Category> { { 0xA9DE, 0xA9DF }, ::unicode::General_Category::Other_Punctuation }, // [2] JAVANESE PADA TIRTA TUMETES..JAVANESE PADA ISEN-ISEN
    Prop<::unicode::General_Category> { { 0xA9E0, 0xA9E4 }, ::unicode::General_Category::Other_Letter }, // [5] MYANMAR LETTER SHAN GHA..MYANMAR LETTER SHAN BHA
    Prop<::unicode::General_Category> { { 0xA9E5, 0xA9E5 }, ::unicode::General_Category::Nonspacing_Mark }, // MYANMAR SIGN SHAN SAW
    Prop<::unicode::General_Category> { { 0xA9E6, 0xA9E6 }, ::unicode::General_Category::Modifier_Letter }, // MYANMAR MODIFIER LETTER SHAN REDUPLICATION
    Prop<::unicode::General_Category> { { 0xA9E7, 0xA9EF }, ::unicode::General_Category::Other_Letter }, // [9] MYANMAR LETTER TAI LAING NYA..MYANMAR LETTER TAI LAING NNA
    Prop<::unicode::General_Category> { { 0xA9F0, 0xA9F9 }, ::unicode::General_Category::Decimal_Number }, // [10] MYANMAR TAI LAING DIGIT ZERO..MYANMAR TAI LAING DIGIT NINE
    Prop<::unicode::General_Category> { { 0xA9FA, 0xA9FE }, ::unicode::General_Category::Other_Letter }, // [5] MYANMAR LETTER TAI LAING LLA..MYANMAR LETTER TAI LAING BHA
    Prop<::unicode::General_Category> { { 0xA9FF, 0xA9FF }, ::unicode::General_Category::Unassigned }, // <reserved-A9FF>
    Prop<::unicode::General_Category> { { 0xAA00, 0xAA28 }, ::unicode::General_Category::Other_Letter }, // [41] CHAM LETTER A..CHAM LETTER HA
    Prop<::unicode::General_Category> { { 0xAA29, 0xAA2E }, ::unicode::General_Category::Nonspacing_Mark }, // [6] CHAM VOWEL SIGN AA..CHAM VOWEL SIGN OE
    Prop<::unicode::General_Category> { { 0xAA2F, 0xAA30 }, ::unicode::General_Category::Spacing_Mark }, // [2] CHAM VOWEL SIGN O..CHAM VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0xAA31, 0xAA32 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] CHAM VOWEL SIGN AU..CHAM VOWEL SIGN UE
    Prop<::unicode::General_Category> { { 0xAA33, 0xAA34 }, ::unicode::General_Category::Spacing_Mark }, // [2] CHAM CONSONANT SIGN YA..CHAM CONSONANT SIGN RA
    Prop<::unicode::General_Category> { { 0xAA35, 0xAA36 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] CHAM CONSONANT SIGN LA..CHAM CONSONANT SIGN WA
    Prop<::unicode::General_Category> { { 0xAA37, 0xAA3F }, ::unicode::General_Category::Unassigned }, // [9] <reserved-AA37>..<reserved-AA3F>
    Prop<::unicode::General_Category> { { 0xAA40, 0xAA42 }, ::unicode::General_Category::Other_Letter }, // [3] CHAM LETTER FINAL K..CHAM LETTER FINAL NG
    Prop<::unicode::General_Category> { { 0xAA43, 0xAA43 }, ::unicode::General_Category::Nonspacing_Mark }, // CHAM CONSONANT SIGN FINAL NG
    Prop<::unicode::General_Category> { { 0xAA44, 0xAA4B }, ::unicode::General_Category::Other_Letter }, // [8] CHAM LETTER FINAL CH..CHAM LETTER FINAL SS
    Prop<::unicode::General_Category> { { 0xAA4C, 0xAA4C }, ::unicode::General_Category::Nonspacing_Mark }, // CHAM CONSONANT SIGN FINAL M
    Prop<::unicode::General_Category> { { 0xAA4D, 0xAA4D }, ::unicode::General_Category::Spacing_Mark }, // CHAM CONSONANT SIGN FINAL H
    Prop<::unicode::General_Category> { { 0xAA4E, 0xAA4F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-AA4E>..<reserved-AA4F>
    Prop<::unicode::General_Category> { { 0xAA50, 0xAA59 }, ::unicode::General_Category::Decimal_Number }, // [10] CHAM DIGIT ZERO..CHAM DIGIT NINE
    Prop<::unicode::General_Category> { { 0xAA5A, 0xAA5B }, ::unicode::General_Category::Unassigned }, // [2] <reserved-AA5A>..<reserved-AA5B>
    Prop<::unicode::General_Category> { { 0xAA5C, 0xAA5F }, ::unicode::General_Category::Other_Punctuation }, // [4] CHAM PUNCTUATION SPIRAL..CHAM PUNCTUATION TRIPLE DANDA
    Prop<::unicode::General_Category> { { 0xAA60, 0xAA6F }, ::unicode::General_Category::Other_Letter }, // [16] MYANMAR LETTER KHAMTI GA..MYANMAR LETTER KHAMTI FA
    Prop<::unicode::General_Category> { { 0xAA70, 0xAA70 }, ::unicode::General_Category::Modifier_Letter }, // MYANMAR MODIFIER LETTER KHAMTI REDUPLICATION
    Prop<::unicode::General_Category> { { 0xAA71, 0xAA76 }, ::unicode::General_Category::Other_Letter }, // [6] MYANMAR LETTER KHAMTI XA..MYANMAR LOGOGRAM KHAMTI HM
    Prop<::unicode::General_Category> { { 0xAA77, 0xAA79 }, ::unicode::General_Category::Other_Symbol }, // [3] MYANMAR SYMBOL AITON EXCLAMATION..MYANMAR SYMBOL AITON TWO
    Prop<::unicode::General_Category> { { 0xAA7A, 0xAA7A }, ::unicode::General_Category::Other_Letter }, // MYANMAR LETTER AITON RA
    Prop<::unicode::General_Category> { { 0xAA7B, 0xAA7B }, ::unicode::General_Category::Spacing_Mark }, // MYANMAR SIGN PAO KAREN TONE
    Prop<::unicode::General_Category> { { 0xAA7C, 0xAA7C }, ::unicode::General_Category::Nonspacing_Mark }, // MYANMAR SIGN TAI LAING TONE-2
    Prop<::unicode::General_Category> { { 0xAA7D, 0xAA7D }, ::unicode::General_Category::Spacing_Mark }, // MYANMAR SIGN TAI LAING TONE-5
    Prop<::unicode::General_Category> { { 0xAA7E, 0xAAAF }, ::unicode::General_Category::Other_Letter }, // [50] MYANMAR LETTER SHWE PALAUNG CHA..TAI VIET LETTER HIGH O
    Prop<::unicode::General_Category> { { 0xAAB0, 0xAAB0 }, ::unicode::General_Category::Nonspacing_Mark }, // TAI VIET MAI KANG
    Prop<::unicode::General_Category> { { 0xAAB1, 0xAAB1 }, ::unicode::General_Category::Other_Letter }, // TAI VIET VOWEL AA
    Prop<::unicode::General_Category> { { 0xAAB2, 0xAAB4 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] TAI VIET VOWEL I..TAI VIET VOWEL U
    Prop<::unicode::General_Category> { { 0xAAB5, 0xAAB6 }, ::unicode::General_Category::Other_Letter }, // [2] TAI VIET VOWEL E..TAI VIET VOWEL O
    Prop<::unicode::General_Category> { { 0xAAB7, 0xAAB8 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] TAI VIET MAI KHIT..TAI VIET VOWEL IA
    Prop<::unicode::General_Category> { { 0xAAB9, 0xAABD }, ::unicode::General_Category::Other_Letter }, // [5] TAI VIET VOWEL UEA..TAI VIET VOWEL AN
    Prop<::unicode::General_Category> { { 0xAABE, 0xAABF }, ::unicode::General_Category::Nonspacing_Mark }, // [2] TAI VIET VOWEL AM..TAI VIET TONE MAI EK
    Prop<::unicode::General_Category> { { 0xAAC0, 0xAAC0 }, ::unicode::General_Category::Other_Letter }, // TAI VIET TONE MAI NUENG
    Prop<::unicode::General_Category> { { 0xAAC1, 0xAAC1 }, ::unicode::General_Category::Nonspacing_Mark }, // TAI VIET TONE MAI THO
    Prop<::unicode::General_Category> { { 0xAAC2, 0xAAC2 }, ::unicode::General_Category::Other_Letter }, // TAI VIET TONE MAI SONG
    Prop<::unicode::General_Category> { { 0xAAC3, 0xAADA }, ::unicode::General_Category::Unassigned }, // [24] <reserved-AAC3>..<reserved-AADA>
    Prop<::unicode::General_Category> { { 0xAADB, 0xAADC }, ::unicode::General_Category::Other_Letter }, // [2] TAI VIET SYMBOL KON..TAI VIET SYMBOL NUENG
    Prop<::unicode::General_Category> { { 0xAADD, 0xAADD }, ::unicode::General_Category::Modifier_Letter }, // TAI VIET SYMBOL SAM
    Prop<::unicode::General_Category> { { 0xAADE, 0xAADF }, ::unicode::General_Category::Other_Punctuation }, // [2] TAI VIET SYMBOL HO HOI..TAI VIET SYMBOL KOI KOI
    Prop<::unicode::General_Category> { { 0xAAE0, 0xAAEA }, ::unicode::General_Category::Other_Letter }, // [11] MEETEI MAYEK LETTER E..MEETEI MAYEK LETTER SSA
    Prop<::unicode::General_Category> { { 0xAAEB, 0xAAEB }, ::unicode::General_Category::Spacing_Mark }, // MEETEI MAYEK VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0xAAEC, 0xAAED }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MEETEI MAYEK VOWEL SIGN UU..MEETEI MAYEK VOWEL SIGN AAI
    Prop<::unicode::General_Category> { { 0xAAEE, 0xAAEF }, ::unicode::General_Category::Spacing_Mark }, // [2] MEETEI MAYEK VOWEL SIGN AU..MEETEI MAYEK VOWEL SIGN AAU
    Prop<::unicode::General_Category> { { 0xAAF0, 0xAAF1 }, ::unicode::General_Category::Other_Punctuation }, // [2] MEETEI MAYEK CHEIKHAN..MEETEI MAYEK AHANG KHUDAM
    Prop<::unicode::General_Category> { { 0xAAF2, 0xAAF2 }, ::unicode::General_Category::Other_Letter }, // MEETEI MAYEK ANJI
    Prop<::unicode::General_Category> { { 0xAAF3, 0xAAF4 }, ::unicode::General_Category::Modifier_Letter }, // [2] MEETEI MAYEK SYLLABLE REPETITION MARK..MEETEI MAYEK WORD REPETITION MARK
    Prop<::unicode::General_Category> { { 0xAAF5, 0xAAF5 }, ::unicode::General_Category::Spacing_Mark }, // MEETEI MAYEK VOWEL SIGN VISARGA
    Prop<::unicode::General_Category> { { 0xAAF6, 0xAAF6 }, ::unicode::General_Category::Nonspacing_Mark }, // MEETEI MAYEK VIRAMA
    Prop<::unicode::General_Category> { { 0xAAF7, 0xAB00 }, ::unicode::General_Category::Unassigned }, // [10] <reserved-AAF7>..<reserved-AB00>
    Prop<::unicode::General_Category> { { 0xAB01, 0xAB06 }, ::unicode::General_Category::Other_Letter }, // [6] ETHIOPIC SYLLABLE TTHU..ETHIOPIC SYLLABLE TTHO
    Prop<::unicode::General_Category> { { 0xAB07, 0xAB08 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-AB07>..<reserved-AB08>
    Prop<::unicode::General_Category> { { 0xAB09, 0xAB0E }, ::unicode::General_Category::Other_Letter }, // [6] ETHIOPIC SYLLABLE DDHU..ETHIOPIC SYLLABLE DDHO
    Prop<::unicode::General_Category> { { 0xAB0F, 0xAB10 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-AB0F>..<reserved-AB10>
    Prop<::unicode::General_Category> { { 0xAB11, 0xAB16 }, ::unicode::General_Category::Other_Letter }, // [6] ETHIOPIC SYLLABLE DZU..ETHIOPIC SYLLABLE DZO
    Prop<::unicode::General_Category> { { 0xAB17, 0xAB1F }, ::unicode::General_Category::Unassigned }, // [9] <reserved-AB17>..<reserved-AB1F>
    Prop<::unicode::General_Category> { { 0xAB20, 0xAB26 }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE CCHHA..ETHIOPIC SYLLABLE CCHHO
    Prop<::unicode::General_Category> { { 0xAB27, 0xAB27 }, ::unicode::General_Category::Unassigned }, // <reserved-AB27>
    Prop<::unicode::General_Category> { { 0xAB28, 0xAB2E }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE BBA..ETHIOPIC SYLLABLE BBO
    Prop<::unicode::General_Category> { { 0xAB2F, 0xAB2F }, ::unicode::General_Category::Unassigned }, // <reserved-AB2F>
    Prop<::unicode::General_Category> { { 0xAB30, 0xAB5A }, ::unicode::General_Category::Lowercase_Letter }, // [43] LATIN SMALL LETTER BARRED ALPHA..LATIN SMALL LETTER Y WITH SHORT RIGHT LEG
    Prop<::unicode::General_Category> { { 0xAB5B, 0xAB5B }, ::unicode::General_Category::Modifier_Symbol }, // MODIFIER BREVE WITH INVERTED BREVE
    Prop<::unicode::General_Category> { { 0xAB5C, 0xAB5F }, ::unicode::General_Category::Modifier_Letter }, // [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    Prop<::unicode::General_Category> { { 0xAB60, 0xAB68 }, ::unicode::General_Category::Lowercase_Letter }, // [9] LATIN SMALL LETTER SAKHA YAT..LATIN SMALL LETTER TURNED R WITH MIDDLE TILDE
    Prop<::unicode::General_Category> { { 0xAB69, 0xAB69 }, ::unicode::General_Category::Modifier_Letter }, // MODIFIER LETTER SMALL TURNED W
    Prop<::unicode::General_Category> { { 0xAB6A, 0xAB6B }, ::unicode::General_Category::Modifier_Symbol }, // [2] MODIFIER LETTER LEFT TACK..MODIFIER LETTER RIGHT TACK
    Prop<::unicode::General_Category> { { 0xAB6C, 0xAB6F }, ::unicode::General_Category::Unassigned }, // [4] <reserved-AB6C>..<reserved-AB6F>
    Prop<::unicode::General_Category> { { 0xAB70, 0xABBF }, ::unicode::General_Category::Lowercase_Letter }, // [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Prop<::unicode::General_Category> { { 0xABC0, 0xABE2 }, ::unicode::General_Category::Other_Letter }, // [35] MEETEI MAYEK LETTER KOK..MEETEI MAYEK LETTER I LONSUM
    Prop<::unicode::General_Category> { { 0xABE3, 0xABE4 }, ::unicode::General_Category::Spacing_Mark }, // [2] MEETEI MAYEK VOWEL SIGN ONAP..MEETEI MAYEK VOWEL SIGN INAP
    Prop<::unicode::General_Category> { { 0xABE5, 0xABE5 }, ::unicode::General_Category::Nonspacing_Mark }, // MEETEI MAYEK VOWEL SIGN ANAP
    Prop<::unicode::General_Category> { { 0xABE6, 0xABE7 }, ::unicode::General_Category::Spacing_Mark }, // [2] MEETEI MAYEK VOWEL SIGN YENAP..MEETEI MAYEK VOWEL SIGN SOUNAP
    Prop<::unicode::General_Category> { { 0xABE8, 0xABE8 }, ::unicode::General_Category::Nonspacing_Mark }, // MEETEI MAYEK VOWEL SIGN UNAP
    Prop<::unicode::General_Category> { { 0xABE9, 0xABEA }, ::unicode::General_Category::Spacing_Mark }, // [2] MEETEI MAYEK VOWEL SIGN CHEINAP..MEETEI MAYEK VOWEL SIGN NUNG
    Prop<::unicode::General_Category> { { 0xABEB, 0xABEB }, ::unicode::General_Category::Other_Punctuation }, // MEETEI MAYEK CHEIKHEI
    Prop<::unicode::General_Category> { { 0xABEC, 0xABEC }, ::unicode::General_Category::Spacing_Mark }, // MEETEI MAYEK LUM IYEK
    Prop<::unicode::General_Category> { { 0xABED, 0xABED }, ::unicode::General_Category::Nonspacing_Mark }, // MEETEI MAYEK APUN IYEK
    Prop<::unicode::General_Category> { { 0xABEE, 0xABEF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-ABEE>..<reserved-ABEF>
    Prop<::unicode::General_Category> { { 0xABF0, 0xABF9 }, ::unicode::General_Category::Decimal_Number }, // [10] MEETEI MAYEK DIGIT ZERO..MEETEI MAYEK DIGIT NINE
    Prop<::unicode::General_Category> { { 0xABFA, 0xABFF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-ABFA>..<reserved-ABFF>
    Prop<::unicode::General_Category> { { 0xAC00, 0xD7A3 }, ::unicode::General_Category::Other_Letter }, // [11172] HANGUL SYLLABLE GA..HANGUL SYLLABLE HIH
    Prop<::unicode::General_Category> { { 0xD7A4, 0xD7AF }, ::unicode::General_Category::Unassigned }, // [12] <reserved-D7A4>..<reserved-D7AF>
    Prop<::unicode::General_Category> { { 0xD7B0, 0xD7C6 }, ::unicode::General_Category::Other_Letter }, // [23] HANGUL JUNGSEONG O-YEO..HANGUL JUNGSEONG ARAEA-E
    Prop<::unicode::General_Category> { { 0xD7C7, 0xD7CA }, ::unicode::General_Category::Unassigned }, // [4] <reserved-D7C7>..<reserved-D7CA>
    Prop<::unicode::General_Category> { { 0xD7CB, 0xD7FB }, ::unicode::General_Category::Other_Letter }, // [49] HANGUL JONGSEONG NIEUN-RIEUL..HANGUL JONGSEONG PHIEUPH-THIEUTH
    Prop<::unicode::General_Category> { { 0xD7FC, 0xD7FF }, ::unicode::General_Category::Unassigned }, // [4] <reserved-D7FC>..<reserved-D7FF>
    Prop<::unicode::General_Category> { { 0xD800, 0xDFFF }, ::unicode::General_Category::Surrogate }, // [2048] <surrogate-D800>..<surrogate-DFFF>
    Prop<::unicode::General_Category> { { 0xE000, 0xF8FF }, ::unicode::General_Category::Private_Use }, // [6400] <private-use-E000>..<private-use-F8FF>
    Prop<::unicode::General_Category> { { 0xF900, 0xFA6D }, ::unicode::General_Category::Other_Letter }, // [366] CJK COMPATIBILITY IDEOGRAPH-F900..CJK COMPATIBILITY IDEOGRAPH-FA6D
    Prop<::unicode::General_Category> { { 0xFA6E, 0xFA6F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-FA6E>..<reserved-FA6F>
    Prop<::unicode::General_Category> { { 0xFA70, 0xFAD9 }, ::unicode::General_Category::Other_Letter }, // [106] CJK COMPATIBILITY IDEOGRAPH-FA70..CJK COMPATIBILITY IDEOGRAPH-FAD9
    Prop<::unicode::General_Category> { { 0xFADA, 0xFAFF }, ::unicode::General_Category::Unassigned }, // [38] <reserved-FADA>..<reserved-FAFF>
    Prop<::unicode::General_Category> { { 0xFB00, 0xFB06 }, ::unicode::General_Category::Lowercase_Letter }, // [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Prop<::unicode::General_Category> { { 0xFB07, 0xFB12 }, ::unicode::General_Category::Unassigned }, // [12] <reserved-FB07>..<reserved-FB12>
    Prop<::unicode::General_Category> { { 0xFB13, 0xFB17 }, ::unicode::General_Category::Lowercase_Letter }, // [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Prop<::unicode::General_Category> { { 0xFB18, 0xFB1C }, ::unicode::General_Category::Unassigned }, // [5] <reserved-FB18>..<reserved-FB1C>
    Prop<::unicode::General_Category> { { 0xFB1D, 0xFB1D }, ::unicode::General_Category::Other_Letter }, // HEBREW LETTER YOD WITH HIRIQ
    Prop<::unicode::General_Category> { { 0xFB1E, 0xFB1E }, ::unicode::General_Category::Nonspacing_Mark }, // HEBREW POINT JUDEO-SPANISH VARIKA
    Prop<::unicode::General_Category> { { 0xFB1F, 0xFB28 }, ::unicode::General_Category::Other_Letter }, // [10] HEBREW LIGATURE YIDDISH YOD YOD PATAH..HEBREW LETTER WIDE TAV
    Prop<::unicode::General_Category> { { 0xFB29, 0xFB29 }, ::unicode::General_Category::Math_Symbol }, // HEBREW LETTER ALTERNATIVE PLUS SIGN
    Prop<::unicode::General_Category> { { 0xFB2A, 0xFB36 }, ::unicode::General_Category::Other_Letter }, // [13] HEBREW LETTER SHIN WITH SHIN DOT..HEBREW LETTER ZAYIN WITH DAGESH
    Prop<::unicode::General_Category> { { 0xFB37, 0xFB37 }, ::unicode::General_Category::Unassigned }, // <reserved-FB37>
    Prop<::unicode::General_Category> { { 0xFB38, 0xFB3C }, ::unicode::General_Category::Other_Letter }, // [5] HEBREW LETTER TET WITH DAGESH..HEBREW LETTER LAMED WITH DAGESH
    Prop<::unicode::General_Category> { { 0xFB3D, 0xFB3D }, ::unicode::General_Category::Unassigned }, // <reserved-FB3D>
    Prop<::unicode::General_Category> { { 0xFB3E, 0xFB3E }, ::unicode::General_Category::Other_Letter }, // HEBREW LETTER MEM WITH DAGESH
    Prop<::unicode::General_Category> { { 0xFB3F, 0xFB3F }, ::unicode::General_Category::Unassigned }, // <reserved-FB3F>
    Prop<::unicode::General_Category> { { 0xFB40, 0xFB41 }, ::unicode::General_Category::Other_Letter }, // [2] HEBREW LETTER NUN WITH DAGESH..HEBREW LETTER SAMEKH WITH DAGESH
    Prop<::unicode::General_Category> { { 0xFB42, 0xFB42 }, ::unicode::General_Category::Unassigned }, // <reserved-FB42>
    Prop<::unicode::General_Category> { { 0xFB43, 0xFB44 }, ::unicode::General_Category::Other_Letter }, // [2] HEBREW LETTER FINAL PE WITH DAGESH..HEBREW LETTER PE WITH DAGESH
    Prop<::unicode::General_Category> { { 0xFB45, 0xFB45 }, ::unicode::General_Category::Unassigned }, // <reserved-FB45>
    Prop<::unicode::General_Category> { { 0xFB46, 0xFBB1 }, ::unicode::General_Category::Other_Letter }, // [108] HEBREW LETTER TSADI WITH DAGESH..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE FINAL FORM
    Prop<::unicode::General_Category> { { 0xFBB2, 0xFBC2 }, ::unicode::General_Category::Modifier_Symbol }, // [17] ARABIC SYMBOL DOT ABOVE..ARABIC SYMBOL WASLA ABOVE
    Prop<::unicode::General_Category> { { 0xFBC3, 0xFBD2 }, ::unicode::General_Category::Other_Symbol }, // [16] ARABIC LIGATURE JALLA WA-ALAA..ARABIC LIGATURE ALAYHI AR-RAHMAH
    Prop<::unicode::General_Category> { { 0xFBD3, 0xFD3D }, ::unicode::General_Category::Other_Letter }, // [363] ARABIC LETTER NG ISOLATED FORM..ARABIC LIGATURE ALEF WITH FATHATAN ISOLATED FORM
    Prop<::unicode::General_Category> { { 0xFD3E, 0xFD3E }, ::unicode::General_Category::Close_Punctuation }, // ORNATE LEFT PARENTHESIS
    Prop<::unicode::General_Category> { { 0xFD3F, 0xFD3F }, ::unicode::General_Category::Open_Punctuation }, // ORNATE RIGHT PARENTHESIS
    Prop<::unicode::General_Category> { { 0xFD40, 0xFD4F }, ::unicode::General_Category::Other_Symbol }, // [16] ARABIC LIGATURE RAHIMAHU ALLAAH..ARABIC LIGATURE RAHIMAHUM ALLAAH
    Prop<::unicode::General_Category> { { 0xFD50, 0xFD8F }, ::unicode::General_Category::Other_Letter }, // [64] ARABIC LIGATURE TEH WITH JEEM WITH MEEM INITIAL FORM..ARABIC LIGATURE MEEM WITH KHAH WITH MEEM INITIAL FORM
    Prop<::unicode::General_Category> { { 0xFD90, 0xFD91 }, ::unicode::General_Category::Other_Symbol }, // [2] ARABIC LIGATURE RAHMATU ALLAAHI ALAYH..ARABIC LIGATURE RAHMATU ALLAAHI ALAYHAA
    Prop<::unicode::General_Category> { { 0xFD92, 0xFDC7 }, ::unicode::General_Category::Other_Letter }, // [54] ARABIC LIGATURE MEEM WITH JEEM WITH KHAH INITIAL FORM..ARABIC LIGATURE NOON WITH JEEM WITH YEH FINAL FORM
    Prop<::unicode::General_Category> { { 0xFDC8, 0xFDCF }, ::unicode::General_Category::Other_Symbol }, // [8] ARABIC LIGATURE RAHIMAHU ALLAAH TAAALAA..ARABIC LIGATURE SALAAMUHU ALAYNAA
    Prop<::unicode::General_Category> { { 0xFDD0, 0xFDEF }, ::unicode::General_Category::Unassigned }, // [32] <noncharacter-FDD0>..<noncharacter-FDEF>
    Prop<::unicode::General_Category> { { 0xFDF0, 0xFDFB }, ::unicode::General_Category::Other_Letter }, // [12] ARABIC LIGATURE SALLA USED AS KORANIC STOP SIGN ISOLATED FORM..ARABIC LIGATURE JALLAJALALOUHOU
    Prop<::unicode::General_Category> { { 0xFDFC, 0xFDFC }, ::unicode::General_Category::Currency_Symbol }, // RIAL SIGN
    Prop<::unicode::General_Category> { { 0xFDFD, 0xFDFF }, ::unicode::General_Category::Other_Symbol }, // [3] ARABIC LIGATURE BISMILLAH AR-RAHMAN AR-RAHEEM..ARABIC LIGATURE AZZA WA JALL
    Prop<::unicode::General_Category> { { 0xFE00, 0xFE0F }, ::unicode::General_Category::Nonspacing_Mark }, // [16] VARIATION SELECTOR-1..VARIATION SELECTOR-16
    Prop<::unicode::General_Category> { { 0xFE10, 0xFE16 }, ::unicode::General_Category::Other_Punctuation }, // [7] PRESENTATION FORM FOR VERTICAL COMMA..PRESENTATION FORM FOR VERTICAL QUESTION MARK
    Prop<::unicode::General_Category> { { 0xFE17, 0xFE17 }, ::unicode::General_Category::Open_Punctuation }, // PRESENTATION FORM FOR VERTICAL LEFT WHITE LENTICULAR BRACKET
    Prop<::unicode::General_Category> { { 0xFE18, 0xFE18 }, ::unicode::General_Category::Close_Punctuation }, // PRESENTATION FORM FOR VERTICAL RIGHT WHITE LENTICULAR BRAKCET
    Prop<::unicode::General_Category> { { 0xFE19, 0xFE19 }, ::unicode::General_Category::Other_Punctuation }, // PRESENTATION FORM FOR VERTICAL HORIZONTAL ELLIPSIS
    Prop<::unicode::General_Category> { { 0xFE1A, 0xFE1F }, ::unicode::General_Category::Unassigned }, // [6] <reserved-FE1A>..<reserved-FE1F>
    Prop<::unicode::General_Category> { { 0xFE20, 0xFE2F }, ::unicode::General_Category::Nonspacing_Mark }, // [16] COMBINING LIGATURE LEFT HALF..COMBINING CYRILLIC TITLO RIGHT HALF
    Prop<::unicode::General_Category> { { 0xFE30, 0xFE30 }, ::unicode::General_Category::Other_Punctuation }, // PRESENTATION FORM FOR VERTICAL TWO DOT LEADER
    Prop<::unicode::General_Category> { { 0xFE31, 0xFE32 }, ::unicode::General_Category::Dash_Punctuation }, // [2] PRESENTATION FORM FOR VERTICAL EM DASH..PRESENTATION FORM FOR VERTICAL EN DASH
    Prop<::unicode::General_Category> { { 0xFE33, 0xFE34 }, ::unicode::General_Category::Connector_Punctuation }, // [2] PRESENTATION FORM FOR VERTICAL LOW LINE..PRESENTATION FORM FOR VERTICAL WAVY LOW LINE
    Prop<::unicode::General_Category> { { 0xFE35, 0xFE35 }, ::unicode::General_Category::Open_Punctuation }, // PRESENTATION FORM FOR VERTICAL LEFT PARENTHESIS
    Prop<::unicode::General_Category> { { 0xFE36, 0xFE36 }, ::unicode::General_Category::Close_Punctuation }, // PRESENTATION FORM FOR VERTICAL RIGHT PARENTHESIS
    Prop<::unicode::General_Category> { { 0xFE37, 0xFE37 }, ::unicode::General_Category::Open_Punctuation }, // PRESENTATION FORM FOR VERTICAL LEFT CURLY BRACKET
    Prop<::unicode::General_Category> { { 0xFE38, 0xFE38 }, ::unicode::General_Category::Close_Punctuation }, // PRESENTATION FORM FOR VERTICAL RIGHT CURLY BRACKET
    Prop<::unicode::General_Category> { { 0xFE39, 0xFE39 }, ::unicode::General_Category::Open_Punctuation }, // PRESENTATION FORM FOR VERTICAL LEFT TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0xFE3A, 0xFE3A }, ::unicode::General_Category::Close_Punctuation }, // PRESENTATION FORM FOR VERTICAL RIGHT TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0xFE3B, 0xFE3B }, ::unicode::General_Category::Open_Punctuation }, // PRESENTATION FORM FOR VERTICAL LEFT BLACK LENTICULAR BRACKET
    Prop<::unicode::General_Category> { { 0xFE3C, 0xFE3C }, ::unicode::General_Category::Close_Punctuation }, // PRESENTATION FORM FOR VERTICAL RIGHT BLACK LENTICULAR BRACKET
    Prop<::unicode::General_Category> { { 0xFE3D, 0xFE3D }, ::unicode::General_Category::Open_Punctuation }, // PRESENTATION FORM FOR VERTICAL LEFT DOUBLE ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0xFE3E, 0xFE3E }, ::unicode::General_Category::Close_Punctuation }, // PRESENTATION FORM FOR VERTICAL RIGHT DOUBLE ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0xFE3F, 0xFE3F }, ::unicode::General_Category::Open_Punctuation }, // PRESENTATION FORM FOR VERTICAL LEFT ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0xFE40, 0xFE40 }, ::unicode::General_Category::Close_Punctuation }, // PRESENTATION FORM FOR VERTICAL RIGHT ANGLE BRACKET
    Prop<::unicode::General_Category> { { 0xFE41, 0xFE41 }, ::unicode::General_Category::Open_Punctuation }, // PRESENTATION FORM FOR VERTICAL LEFT CORNER BRACKET
    Prop<::unicode::General_Category> { { 0xFE42, 0xFE42 }, ::unicode::General_Category::Close_Punctuation }, // PRESENTATION FORM FOR VERTICAL RIGHT CORNER BRACKET
    Prop<::unicode::General_Category> { { 0xFE43, 0xFE43 }, ::unicode::General_Category::Open_Punctuation }, // PRESENTATION FORM FOR VERTICAL LEFT WHITE CORNER BRACKET
    Prop<::unicode::General_Category> { { 0xFE44, 0xFE44 }, ::unicode::General_Category::Close_Punctuation }, // PRESENTATION FORM FOR VERTICAL RIGHT WHITE CORNER BRACKET
    Prop<::unicode::General_Category> { { 0xFE45, 0xFE46 }, ::unicode::General_Category::Other_Punctuation }, // [2] SESAME DOT..WHITE SESAME DOT
    Prop<::unicode::General_Category> { { 0xFE47, 0xFE47 }, ::unicode::General_Category::Open_Punctuation }, // PRESENTATION FORM FOR VERTICAL LEFT SQUARE BRACKET
    Prop<::unicode::General_Category> { { 0xFE48, 0xFE48 }, ::unicode::General_Category::Close_Punctuation }, // PRESENTATION FORM FOR VERTICAL RIGHT SQUARE BRACKET
    Prop<::unicode::General_Category> { { 0xFE49, 0xFE4C }, ::unicode::General_Category::Other_Punctuation }, // [4] DASHED OVERLINE..DOUBLE WAVY OVERLINE
    Prop<::unicode::General_Category> { { 0xFE4D, 0xFE4F }, ::unicode::General_Category::Connector_Punctuation }, // [3] DASHED LOW LINE..WAVY LOW LINE
    Prop<::unicode::General_Category> { { 0xFE50, 0xFE52 }, ::unicode::General_Category::Other_Punctuation }, // [3] SMALL COMMA..SMALL FULL STOP
    Prop<::unicode::General_Category> { { 0xFE53, 0xFE53 }, ::unicode::General_Category::Unassigned }, // <reserved-FE53>
    Prop<::unicode::General_Category> { { 0xFE54, 0xFE57 }, ::unicode::General_Category::Other_Punctuation }, // [4] SMALL SEMICOLON..SMALL EXCLAMATION MARK
    Prop<::unicode::General_Category> { { 0xFE58, 0xFE58 }, ::unicode::General_Category::Dash_Punctuation }, // SMALL EM DASH
    Prop<::unicode::General_Category> { { 0xFE59, 0xFE59 }, ::unicode::General_Category::Open_Punctuation }, // SMALL LEFT PARENTHESIS
    Prop<::unicode::General_Category> { { 0xFE5A, 0xFE5A }, ::unicode::General_Category::Close_Punctuation }, // SMALL RIGHT PARENTHESIS
    Prop<::unicode::General_Category> { { 0xFE5B, 0xFE5B }, ::unicode::General_Category::Open_Punctuation }, // SMALL LEFT CURLY BRACKET
    Prop<::unicode::General_Category> { { 0xFE5C, 0xFE5C }, ::unicode::General_Category::Close_Punctuation }, // SMALL RIGHT CURLY BRACKET
    Prop<::unicode::General_Category> { { 0xFE5D, 0xFE5D }, ::unicode::General_Category::Open_Punctuation }, // SMALL LEFT TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0xFE5E, 0xFE5E }, ::unicode::General_Category::Close_Punctuation }, // SMALL RIGHT TORTOISE SHELL BRACKET
    Prop<::unicode::General_Category> { { 0xFE5F, 0xFE61 }, ::unicode::General_Category::Other_Punctuation }, // [3] SMALL NUMBER SIGN..SMALL ASTERISK
    Prop<::unicode::General_Category> { { 0xFE62, 0xFE62 }, ::unicode::General_Category::Math_Symbol }, // SMALL PLUS SIGN
    Prop<::unicode::General_Category> { { 0xFE63, 0xFE63 }, ::unicode::General_Category::Dash_Punctuation }, // SMALL HYPHEN-MINUS
    Prop<::unicode::General_Category> { { 0xFE64, 0xFE66 }, ::unicode::General_Category::Math_Symbol }, // [3] SMALL LESS-THAN SIGN..SMALL EQUALS SIGN
    Prop<::unicode::General_Category> { { 0xFE67, 0xFE67 }, ::unicode::General_Category::Unassigned }, // <reserved-FE67>
    Prop<::unicode::General_Category> { { 0xFE68, 0xFE68 }, ::unicode::General_Category::Other_Punctuation }, // SMALL REVERSE SOLIDUS
    Prop<::unicode::General_Category> { { 0xFE69, 0xFE69 }, ::unicode::General_Category::Currency_Symbol }, // SMALL DOLLAR SIGN
    Prop<::unicode::General_Category> { { 0xFE6A, 0xFE6B }, ::unicode::General_Category::Other_Punctuation }, // [2] SMALL PERCENT SIGN..SMALL COMMERCIAL AT
    Prop<::unicode::General_Category> { { 0xFE6C, 0xFE6F }, ::unicode::General_Category::Unassigned }, // [4] <reserved-FE6C>..<reserved-FE6F>
    Prop<::unicode::General_Category> { { 0xFE70, 0xFE74 }, ::unicode::General_Category::Other_Letter }, // [5] ARABIC FATHATAN ISOLATED FORM..ARABIC KASRATAN ISOLATED FORM
    Prop<::unicode::General_Category> { { 0xFE75, 0xFE75 }, ::unicode::General_Category::Unassigned }, // <reserved-FE75>
    Prop<::unicode::General_Category> { { 0xFE76, 0xFEFC }, ::unicode::General_Category::Other_Letter }, // [135] ARABIC FATHA ISOLATED FORM..ARABIC LIGATURE LAM WITH ALEF FINAL FORM
    Prop<::unicode::General_Category> { { 0xFEFD, 0xFEFE }, ::unicode::General_Category::Unassigned }, // [2] <reserved-FEFD>..<reserved-FEFE>
    Prop<::unicode::General_Category> { { 0xFEFF, 0xFEFF }, ::unicode::General_Category::Format }, // ZERO WIDTH NO-BREAK SPACE
    Prop<::unicode::General_Category> { { 0xFF00, 0xFF00 }, ::unicode::General_Category::Unassigned }, // <reserved-FF00>
    Prop<::unicode::General_Category> { { 0xFF01, 0xFF03 }, ::unicode::General_Category::Other_Punctuation }, // [3] FULLWIDTH EXCLAMATION MARK..FULLWIDTH NUMBER SIGN
    Prop<::unicode::General_Category> { { 0xFF04, 0xFF04 }, ::unicode::General_Category::Currency_Symbol }, // FULLWIDTH DOLLAR SIGN
    Prop<::unicode::General_Category> { { 0xFF05, 0xFF07 }, ::unicode::General_Category::Other_Punctuation }, // [3] FULLWIDTH PERCENT SIGN..FULLWIDTH APOSTROPHE
    Prop<::unicode::General_Category> { { 0xFF08, 0xFF08 }, ::unicode::General_Category::Open_Punctuation }, // FULLWIDTH LEFT PARENTHESIS
    Prop<::unicode::General_Category> { { 0xFF09, 0xFF09 }, ::unicode::General_Category::Close_Punctuation }, // FULLWIDTH RIGHT PARENTHESIS
    Prop<::unicode::General_Category> { { 0xFF0A, 0xFF0A }, ::unicode::General_Category::Other_Punctuation }, // FULLWIDTH ASTERISK
    Prop<::unicode::General_Category> { { 0xFF0B, 0xFF0B }, ::unicode::General_Category::Math_Symbol }, // FULLWIDTH PLUS SIGN
    Prop<::unicode::General_Category> { { 0xFF0C, 0xFF0C }, ::unicode::General_Category::Other_Punctuation }, // FULLWIDTH COMMA
    Prop<::unicode::General_Category> { { 0xFF0D, 0xFF0D }, ::unicode::General_Category::Dash_Punctuation }, // FULLWIDTH HYPHEN-MINUS
    Prop<::unicode::General_Category> { { 0xFF0E, 0xFF0F }, ::unicode::General_Category::Other_Punctuation }, // [2] FULLWIDTH FULL STOP..FULLWIDTH SOLIDUS
    Prop<::unicode::General_Category> { { 0xFF10, 0xFF19 }, ::unicode::General_Category::Decimal_Number }, // [10] FULLWIDTH DIGIT ZERO..FULLWIDTH DIGIT NINE
    Prop<::unicode::General_Category> { { 0xFF1A, 0xFF1B }, ::unicode::General_Category::Other_Punctuation }, // [2] FULLWIDTH COLON..FULLWIDTH SEMICOLON
    Prop<::unicode::General_Category> { { 0xFF1C, 0xFF1E }, ::unicode::General_Category::Math_Symbol }, // [3] FULLWIDTH LESS-THAN SIGN..FULLWIDTH GREATER-THAN SIGN
    Prop<::unicode::General_Category> { { 0xFF1F, 0xFF20 }, ::unicode::General_Category::Other_Punctuation }, // [2] FULLWIDTH QUESTION MARK..FULLWIDTH COMMERCIAL AT
    Prop<::unicode::General_Category> { { 0xFF21, 0xFF3A }, ::unicode::General_Category::Uppercase_Letter }, // [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Prop<::unicode::General_Category> { { 0xFF3B, 0xFF3B }, ::unicode::General_Category::Open_Punctuation }, // FULLWIDTH LEFT SQUARE BRACKET
    Prop<::unicode::General_Category> { { 0xFF3C, 0xFF3C }, ::unicode::General_Category::Other_Punctuation }, // FULLWIDTH REVERSE SOLIDUS
    Prop<::unicode::General_Category> { { 0xFF3D, 0xFF3D }, ::unicode::General_Category::Close_Punctuation }, // FULLWIDTH RIGHT SQUARE BRACKET
    Prop<::unicode::General_Category> { { 0xFF3E, 0xFF3E }, ::unicode::General_Category::Modifier_Symbol }, // FULLWIDTH CIRCUMFLEX ACCENT
    Prop<::unicode::General_Category> { { 0xFF3F, 0xFF3F }, ::unicode::General_Category::Connector_Punctuation }, // FULLWIDTH LOW LINE
    Prop<::unicode::General_Category> { { 0xFF40, 0xFF40 }, ::unicode::General_Category::Modifier_Symbol }, // FULLWIDTH GRAVE ACCENT
    Prop<::unicode::General_Category> { { 0xFF41, 0xFF5A }, ::unicode::General_Category::Lowercase_Letter }, // [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Prop<::unicode::General_Category> { { 0xFF5B, 0xFF5B }, ::unicode::General_Category::Open_Punctuation }, // FULLWIDTH LEFT CURLY BRACKET
    Prop<::unicode::General_Category> { { 0xFF5C, 0xFF5C }, ::unicode::General_Category::Math_Symbol }, // FULLWIDTH VERTICAL LINE
    Prop<::unicode::General_Category> { { 0xFF5D, 0xFF5D }, ::unicode::General_Category::Close_Punctuation }, // FULLWIDTH RIGHT CURLY BRACKET
    Prop<::unicode::General_Category> { { 0xFF5E, 0xFF5E }, ::unicode::General_Category::Math_Symbol }, // FULLWIDTH TILDE
    Prop<::unicode::General_Category> { { 0xFF5F, 0xFF5F }, ::unicode::General_Category::Open_Punctuation }, // FULLWIDTH LEFT WHITE PARENTHESIS
    Prop<::unicode::General_Category> { { 0xFF60, 0xFF60 }, ::unicode::General_Category::Close_Punctuation }, // FULLWIDTH RIGHT WHITE PARENTHESIS
    Prop<::unicode::General_Category> { { 0xFF61, 0xFF61 }, ::unicode::General_Category::Other_Punctuation }, // HALFWIDTH IDEOGRAPHIC FULL STOP
    Prop<::unicode::General_Category> { { 0xFF62, 0xFF62 }, ::unicode::General_Category::Open_Punctuation }, // HALFWIDTH LEFT CORNER BRACKET
    Prop<::unicode::General_Category> { { 0xFF63, 0xFF63 }, ::unicode::General_Category::Close_Punctuation }, // HALFWIDTH RIGHT CORNER BRACKET
    Prop<::unicode::General_Category> { { 0xFF64, 0xFF65 }, ::unicode::General_Category::Other_Punctuation }, // [2] HALFWIDTH IDEOGRAPHIC COMMA..HALFWIDTH KATAKANA MIDDLE DOT
    Prop<::unicode::General_Category> { { 0xFF66, 0xFF6F }, ::unicode::General_Category::Other_Letter }, // [10] HALFWIDTH KATAKANA LETTER WO..HALFWIDTH KATAKANA LETTER SMALL TU
    Prop<::unicode::General_Category> { { 0xFF70, 0xFF70 }, ::unicode::General_Category::Modifier_Letter }, // HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Prop<::unicode::General_Category> { { 0xFF71, 0xFF9D }, ::unicode::General_Category::Other_Letter }, // [45] HALFWIDTH KATAKANA LETTER A..HALFWIDTH KATAKANA LETTER N
    Prop<::unicode::General_Category> { { 0xFF9E, 0xFF9F }, ::unicode::General_Category::Modifier_Letter }, // [2] HALFWIDTH KATAKANA VOICED SOUND MARK..HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK
    Prop<::unicode::General_Category> { { 0xFFA0, 0xFFBE }, ::unicode::General_Category::Other_Letter }, // [31] HALFWIDTH HANGUL FILLER..HALFWIDTH HANGUL LETTER HIEUH
    Prop<::unicode::General_Category> { { 0xFFBF, 0xFFC1 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-FFBF>..<reserved-FFC1>
    Prop<::unicode::General_Category> { { 0xFFC2, 0xFFC7 }, ::unicode::General_Category::Other_Letter }, // [6] HALFWIDTH HANGUL LETTER A..HALFWIDTH HANGUL LETTER E
    Prop<::unicode::General_Category> { { 0xFFC8, 0xFFC9 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-FFC8>..<reserved-FFC9>
    Prop<::unicode::General_Category> { { 0xFFCA, 0xFFCF }, ::unicode::General_Category::Other_Letter }, // [6] HALFWIDTH HANGUL LETTER YEO..HALFWIDTH HANGUL LETTER OE
    Prop<::unicode::General_Category> { { 0xFFD0, 0xFFD1 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-FFD0>..<reserved-FFD1>
    Prop<::unicode::General_Category> { { 0xFFD2, 0xFFD7 }, ::unicode::General_Category::Other_Letter }, // [6] HALFWIDTH HANGUL LETTER YO..HALFWIDTH HANGUL LETTER YU
    Prop<::unicode::General_Category> { { 0xFFD8, 0xFFD9 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-FFD8>..<reserved-FFD9>
    Prop<::unicode::General_Category> { { 0xFFDA, 0xFFDC }, ::unicode::General_Category::Other_Letter }, // [3] HALFWIDTH HANGUL LETTER EU..HALFWIDTH HANGUL LETTER I
    Prop<::unicode::General_Category> { { 0xFFDD, 0xFFDF }, ::unicode::General_Category::Unassigned }, // [3] <reserved-FFDD>..<reserved-FFDF>
    Prop<::unicode::General_Category> { { 0xFFE0, 0xFFE1 }, ::unicode::General_Category::Currency_Symbol }, // [2] FULLWIDTH CENT SIGN..FULLWIDTH POUND SIGN
    Prop<::unicode::General_Category> { { 0xFFE2, 0xFFE2 }, ::unicode::General_Category::Math_Symbol }, // FULLWIDTH NOT SIGN
    Prop<::unicode::General_Category> { { 0xFFE3, 0xFFE3 }, ::unicode::General_Category::Modifier_Symbol }, // FULLWIDTH MACRON
    Prop<::unicode::General_Category> { { 0xFFE4, 0xFFE4 }, ::unicode::General_Category::Other_Symbol }, // FULLWIDTH BROKEN BAR
    Prop<::unicode::General_Category> { { 0xFFE5, 0xFFE6 }, ::unicode::General_Category::Currency_Symbol }, // [2] FULLWIDTH YEN SIGN..FULLWIDTH WON SIGN
    Prop<::unicode::General_Category> { { 0xFFE7, 0xFFE7 }, ::unicode::General_Category::Unassigned }, // <reserved-FFE7>
    Prop<::unicode::General_Category> { { 0xFFE8, 0xFFE8 }, ::unicode::General_Category::Other_Symbol }, // HALFWIDTH FORMS LIGHT VERTICAL
    Prop<::unicode::General_Category> { { 0xFFE9, 0xFFEC }, ::unicode::General_Category::Math_Symbol }, // [4] HALFWIDTH LEFTWARDS ARROW..HALFWIDTH DOWNWARDS ARROW
    Prop<::unicode::General_Category> { { 0xFFED, 0xFFEE }, ::unicode::General_Category::Other_Symbol }, // [2] HALFWIDTH BLACK SQUARE..HALFWIDTH WHITE CIRCLE
    Prop<::unicode::General_Category> { { 0xFFEF, 0xFFF8 }, ::unicode::General_Category::Unassigned }, // [10] <reserved-FFEF>..<reserved-FFF8>
    Prop<::unicode::General_Category> { { 0xFFF9, 0xFFFB }, ::unicode::General_Category::Format }, // [3] INTERLINEAR ANNOTATION ANCHOR..INTERLINEAR ANNOTATION TERMINATOR
    Prop<::unicode::General_Category> { { 0xFFFC, 0xFFFD }, ::unicode::General_Category::Other_Symbol }, // [2] OBJECT REPLACEMENT CHARACTER..REPLACEMENT CHARACTER
    Prop<::unicode::General_Category> { { 0xFFFE, 0xFFFF }, ::unicode::General_Category::Unassigned }, // [2] <noncharacter-FFFE>..<noncharacter-FFFF>
    Prop<::unicode::General_Category> { { 0x10000, 0x1000B }, ::unicode::General_Category::Other_Letter }, // [12] LINEAR B SYLLABLE B008 A..LINEAR B SYLLABLE B046 JE
    Prop<::unicode::General_Category> { { 0x1000C, 0x1000C }, ::unicode::General_Category::Unassigned }, // <reserved-1000C>
    Prop<::unicode::General_Category> { { 0x1000D, 0x10026 }, ::unicode::General_Category::Other_Letter }, // [26] LINEAR B SYLLABLE B036 JO..LINEAR B SYLLABLE B032 QO
    Prop<::unicode::General_Category> { { 0x10027, 0x10027 }, ::unicode::General_Category::Unassigned }, // <reserved-10027>
    Prop<::unicode::General_Category> { { 0x10028, 0x1003A }, ::unicode::General_Category::Other_Letter }, // [19] LINEAR B SYLLABLE B060 RA..LINEAR B SYLLABLE B042 WO
    Prop<::unicode::General_Category> { { 0x1003B, 0x1003B }, ::unicode::General_Category::Unassigned }, // <reserved-1003B>
    Prop<::unicode::General_Category> { { 0x1003C, 0x1003D }, ::unicode::General_Category::Other_Letter }, // [2] LINEAR B SYLLABLE B017 ZA..LINEAR B SYLLABLE B074 ZE
    Prop<::unicode::General_Category> { { 0x1003E, 0x1003E }, ::unicode::General_Category::Unassigned }, // <reserved-1003E>
    Prop<::unicode::General_Category> { { 0x1003F, 0x1004D }, ::unicode::General_Category::Other_Letter }, // [15] LINEAR B SYLLABLE B020 ZO..LINEAR B SYLLABLE B091 TWO
    Prop<::unicode::General_Category> { { 0x1004E, 0x1004F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1004E>..<reserved-1004F>
    Prop<::unicode::General_Category> { { 0x10050, 0x1005D }, ::unicode::General_Category::Other_Letter }, // [14] LINEAR B SYMBOL B018..LINEAR B SYMBOL B089
    Prop<::unicode::General_Category> { { 0x1005E, 0x1007F }, ::unicode::General_Category::Unassigned }, // [34] <reserved-1005E>..<reserved-1007F>
    Prop<::unicode::General_Category> { { 0x10080, 0x100FA }, ::unicode::General_Category::Other_Letter }, // [123] LINEAR B IDEOGRAM B100 MAN..LINEAR B IDEOGRAM VESSEL B305
    Prop<::unicode::General_Category> { { 0x100FB, 0x100FF }, ::unicode::General_Category::Unassigned }, // [5] <reserved-100FB>..<reserved-100FF>
    Prop<::unicode::General_Category> { { 0x10100, 0x10102 }, ::unicode::General_Category::Other_Punctuation }, // [3] AEGEAN WORD SEPARATOR LINE..AEGEAN CHECK MARK
    Prop<::unicode::General_Category> { { 0x10103, 0x10106 }, ::unicode::General_Category::Unassigned }, // [4] <reserved-10103>..<reserved-10106>
    Prop<::unicode::General_Category> { { 0x10107, 0x10133 }, ::unicode::General_Category::Other_Number }, // [45] AEGEAN NUMBER ONE..AEGEAN NUMBER NINETY THOUSAND
    Prop<::unicode::General_Category> { { 0x10134, 0x10136 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-10134>..<reserved-10136>
    Prop<::unicode::General_Category> { { 0x10137, 0x1013F }, ::unicode::General_Category::Other_Symbol }, // [9] AEGEAN WEIGHT BASE UNIT..AEGEAN MEASURE THIRD SUBUNIT
    Prop<::unicode::General_Category> { { 0x10140, 0x10174 }, ::unicode::General_Category::Letter_Number }, // [53] GREEK ACROPHONIC ATTIC ONE QUARTER..GREEK ACROPHONIC STRATIAN FIFTY MNAS
    Prop<::unicode::General_Category> { { 0x10175, 0x10178 }, ::unicode::General_Category::Other_Number }, // [4] GREEK ONE HALF SIGN..GREEK THREE QUARTERS SIGN
    Prop<::unicode::General_Category> { { 0x10179, 0x10189 }, ::unicode::General_Category::Other_Symbol }, // [17] GREEK YEAR SIGN..GREEK TRYBLION BASE SIGN
    Prop<::unicode::General_Category> { { 0x1018A, 0x1018B }, ::unicode::General_Category::Other_Number }, // [2] GREEK ZERO SIGN..GREEK ONE QUARTER SIGN
    Prop<::unicode::General_Category> { { 0x1018C, 0x1018E }, ::unicode::General_Category::Other_Symbol }, // [3] GREEK SINUSOID SIGN..NOMISMA SIGN
    Prop<::unicode::General_Category> { { 0x1018F, 0x1018F }, ::unicode::General_Category::Unassigned }, // <reserved-1018F>
    Prop<::unicode::General_Category> { { 0x10190, 0x1019C }, ::unicode::General_Category::Other_Symbol }, // [13] ROMAN SEXTANS SIGN..ASCIA SYMBOL
    Prop<::unicode::General_Category> { { 0x1019D, 0x1019F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1019D>..<reserved-1019F>
    Prop<::unicode::General_Category> { { 0x101A0, 0x101A0 }, ::unicode::General_Category::Other_Symbol }, // GREEK SYMBOL TAU RHO
    Prop<::unicode::General_Category> { { 0x101A1, 0x101CF }, ::unicode::General_Category::Unassigned }, // [47] <reserved-101A1>..<reserved-101CF>
    Prop<::unicode::General_Category> { { 0x101D0, 0x101FC }, ::unicode::General_Category::Other_Symbol }, // [45] PHAISTOS DISC SIGN PEDESTRIAN..PHAISTOS DISC SIGN WAVY BAND
    Prop<::unicode::General_Category> { { 0x101FD, 0x101FD }, ::unicode::General_Category::Nonspacing_Mark }, // PHAISTOS DISC SIGN COMBINING OBLIQUE STROKE
    Prop<::unicode::General_Category> { { 0x101FE, 0x1027F }, ::unicode::General_Category::Unassigned }, // [130] <reserved-101FE>..<reserved-1027F>
    Prop<::unicode::General_Category> { { 0x10280, 0x1029C }, ::unicode::General_Category::Other_Letter }, // [29] LYCIAN LETTER A..LYCIAN LETTER X
    Prop<::unicode::General_Category> { { 0x1029D, 0x1029F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1029D>..<reserved-1029F>
    Prop<::unicode::General_Category> { { 0x102A0, 0x102D0 }, ::unicode::General_Category::Other_Letter }, // [49] CARIAN LETTER A..CARIAN LETTER UUU3
    Prop<::unicode::General_Category> { { 0x102D1, 0x102DF }, ::unicode::General_Category::Unassigned }, // [15] <reserved-102D1>..<reserved-102DF>
    Prop<::unicode::General_Category> { { 0x102E0, 0x102E0 }, ::unicode::General_Category::Nonspacing_Mark }, // COPTIC EPACT THOUSANDS MARK
    Prop<::unicode::General_Category> { { 0x102E1, 0x102FB }, ::unicode::General_Category::Other_Number }, // [27] COPTIC EPACT DIGIT ONE..COPTIC EPACT NUMBER NINE HUNDRED
    Prop<::unicode::General_Category> { { 0x102FC, 0x102FF }, ::unicode::General_Category::Unassigned }, // [4] <reserved-102FC>..<reserved-102FF>
    Prop<::unicode::General_Category> { { 0x10300, 0x1031F }, ::unicode::General_Category::Other_Letter }, // [32] OLD ITALIC LETTER A..OLD ITALIC LETTER ESS
    Prop<::unicode::General_Category> { { 0x10320, 0x10323 }, ::unicode::General_Category::Other_Number }, // [4] OLD ITALIC NUMERAL ONE..OLD ITALIC NUMERAL FIFTY
    Prop<::unicode::General_Category> { { 0x10324, 0x1032C }, ::unicode::General_Category::Unassigned }, // [9] <reserved-10324>..<reserved-1032C>
    Prop<::unicode::General_Category> { { 0x1032D, 0x10340 }, ::unicode::General_Category::Other_Letter }, // [20] OLD ITALIC LETTER YE..GOTHIC LETTER PAIRTHRA
    Prop<::unicode::General_Category> { { 0x10341, 0x10341 }, ::unicode::General_Category::Letter_Number }, // GOTHIC LETTER NINETY
    Prop<::unicode::General_Category> { { 0x10342, 0x10349 }, ::unicode::General_Category::Other_Letter }, // [8] GOTHIC LETTER RAIDA..GOTHIC LETTER OTHAL
    Prop<::unicode::General_Category> { { 0x1034A, 0x1034A }, ::unicode::General_Category::Letter_Number }, // GOTHIC LETTER NINE HUNDRED
    Prop<::unicode::General_Category> { { 0x1034B, 0x1034F }, ::unicode::General_Category::Unassigned }, // [5] <reserved-1034B>..<reserved-1034F>
    Prop<::unicode::General_Category> { { 0x10350, 0x10375 }, ::unicode::General_Category::Other_Letter }, // [38] OLD PERMIC LETTER AN..OLD PERMIC LETTER IA
    Prop<::unicode::General_Category> { { 0x10376, 0x1037A }, ::unicode::General_Category::Nonspacing_Mark }, // [5] COMBINING OLD PERMIC LETTER AN..COMBINING OLD PERMIC LETTER SII
    Prop<::unicode::General_Category> { { 0x1037B, 0x1037F }, ::unicode::General_Category::Unassigned }, // [5] <reserved-1037B>..<reserved-1037F>
    Prop<::unicode::General_Category> { { 0x10380, 0x1039D }, ::unicode::General_Category::Other_Letter }, // [30] UGARITIC LETTER ALPA..UGARITIC LETTER SSU
    Prop<::unicode::General_Category> { { 0x1039E, 0x1039E }, ::unicode::General_Category::Unassigned }, // <reserved-1039E>
    Prop<::unicode::General_Category> { { 0x1039F, 0x1039F }, ::unicode::General_Category::Other_Punctuation }, // UGARITIC WORD DIVIDER
    Prop<::unicode::General_Category> { { 0x103A0, 0x103C3 }, ::unicode::General_Category::Other_Letter }, // [36] OLD PERSIAN SIGN A..OLD PERSIAN SIGN HA
    Prop<::unicode::General_Category> { { 0x103C4, 0x103C7 }, ::unicode::General_Category::Unassigned }, // [4] <reserved-103C4>..<reserved-103C7>
    Prop<::unicode::General_Category> { { 0x103C8, 0x103CF }, ::unicode::General_Category::Other_Letter }, // [8] OLD PERSIAN SIGN AURAMAZDAA..OLD PERSIAN SIGN BUUMISH
    Prop<::unicode::General_Category> { { 0x103D0, 0x103D0 }, ::unicode::General_Category::Other_Punctuation }, // OLD PERSIAN WORD DIVIDER
    Prop<::unicode::General_Category> { { 0x103D1, 0x103D5 }, ::unicode::General_Category::Letter_Number }, // [5] OLD PERSIAN NUMBER ONE..OLD PERSIAN NUMBER HUNDRED
    Prop<::unicode::General_Category> { { 0x103D6, 0x103FF }, ::unicode::General_Category::Unassigned }, // [42] <reserved-103D6>..<reserved-103FF>
    Prop<::unicode::General_Category> { { 0x10400, 0x10427 }, ::unicode::General_Category::Uppercase_Letter }, // [40] DESERET CAPITAL LETTER LONG I..DESERET CAPITAL LETTER EW
    Prop<::unicode::General_Category> { { 0x10428, 0x1044F }, ::unicode::General_Category::Lowercase_Letter }, // [40] DESERET SMALL LETTER LONG I..DESERET SMALL LETTER EW
    Prop<::unicode::General_Category> { { 0x10450, 0x1049D }, ::unicode::General_Category::Other_Letter }, // [78] SHAVIAN LETTER PEEP..OSMANYA LETTER OO
    Prop<::unicode::General_Category> { { 0x1049E, 0x1049F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1049E>..<reserved-1049F>
    Prop<::unicode::General_Category> { { 0x104A0, 0x104A9 }, ::unicode::General_Category::Decimal_Number }, // [10] OSMANYA DIGIT ZERO..OSMANYA DIGIT NINE
    Prop<::unicode::General_Category> { { 0x104AA, 0x104AF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-104AA>..<reserved-104AF>
    Prop<::unicode::General_Category> { { 0x104B0, 0x104D3 }, ::unicode::General_Category::Uppercase_Letter }, // [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Prop<::unicode::General_Category> { { 0x104D4, 0x104D7 }, ::unicode::General_Category::Unassigned }, // [4] <reserved-104D4>..<reserved-104D7>
    Prop<::unicode::General_Category> { { 0x104D8, 0x104FB }, ::unicode::General_Category::Lowercase_Letter }, // [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Prop<::unicode::General_Category> { { 0x104FC, 0x104FF }, ::unicode::General_Category::Unassigned }, // [4] <reserved-104FC>..<reserved-104FF>
    Prop<::unicode::General_Category> { { 0x10500, 0x10527 }, ::unicode::General_Category::Other_Letter }, // [40] ELBASAN LETTER A..ELBASAN LETTER KHE
    Prop<::unicode::General_Category> { { 0x10528, 0x1052F }, ::unicode::General_Category::Unassigned }, // [8] <reserved-10528>..<reserved-1052F>
    Prop<::unicode::General_Category> { { 0x10530, 0x10563 }, ::unicode::General_Category::Other_Letter }, // [52] CAUCASIAN ALBANIAN LETTER ALT..CAUCASIAN ALBANIAN LETTER KIW
    Prop<::unicode::General_Category> { { 0x10564, 0x1056E }, ::unicode::General_Category::Unassigned }, // [11] <reserved-10564>..<reserved-1056E>
    Prop<::unicode::General_Category> { { 0x1056F, 0x1056F }, ::unicode::General_Category::Other_Punctuation }, // CAUCASIAN ALBANIAN CITATION MARK
    Prop<::unicode::General_Category> { { 0x10570, 0x1057A }, ::unicode::General_Category::Uppercase_Letter }, // [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Prop<::unicode::General_Category> { { 0x1057B, 0x1057B }, ::unicode::General_Category::Unassigned }, // <reserved-1057B>
    Prop<::unicode::General_Category> { { 0x1057C, 0x1058A }, ::unicode::General_Category::Uppercase_Letter }, // [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Prop<::unicode::General_Category> { { 0x1058B, 0x1058B }, ::unicode::General_Category::Unassigned }, // <reserved-1058B>
    Prop<::unicode::General_Category> { { 0x1058C, 0x10592 }, ::unicode::General_Category::Uppercase_Letter }, // [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Prop<::unicode::General_Category> { { 0x10593, 0x10593 }, ::unicode::General_Category::Unassigned }, // <reserved-10593>
    Prop<::unicode::General_Category> { { 0x10594, 0x10595 }, ::unicode::General_Category::Uppercase_Letter }, // [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Prop<::unicode::General_Category> { { 0x10596, 0x10596 }, ::unicode::General_Category::Unassigned }, // <reserved-10596>
    Prop<::unicode::General_Category> { { 0x10597, 0x105A1 }, ::unicode::General_Category::Lowercase_Letter }, // [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Prop<::unicode::General_Category> { { 0x105A2, 0x105A2 }, ::unicode::General_Category::Unassigned }, // <reserved-105A2>
    Prop<::unicode::General_Category> { { 0x105A3, 0x105B1 }, ::unicode::General_Category::Lowercase_Letter }, // [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Prop<::unicode::General_Category> { { 0x105B2, 0x105B2 }, ::unicode::General_Category::Unassigned }, // <reserved-105B2>
    Prop<::unicode::General_Category> { { 0x105B3, 0x105B9 }, ::unicode::General_Category::Lowercase_Letter }, // [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Prop<::unicode::General_Category> { { 0x105BA, 0x105BA }, ::unicode::General_Category::Unassigned }, // <reserved-105BA>
    Prop<::unicode::General_Category> { { 0x105BB, 0x105BC }, ::unicode::General_Category::Lowercase_Letter }, // [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Prop<::unicode::General_Category> { { 0x105BD, 0x105BF }, ::unicode::General_Category::Unassigned }, // [3] <reserved-105BD>..<reserved-105BF>
    Prop<::unicode::General_Category> { { 0x105C0, 0x105F3 }, ::unicode::General_Category::Other_Letter }, // [52] TODHRI LETTER A..TODHRI LETTER OO
    Prop<::unicode::General_Category> { { 0x105F4, 0x105FF }, ::unicode::General_Category::Unassigned }, // [12] <reserved-105F4>..<reserved-105FF>
    Prop<::unicode::General_Category> { { 0x10600, 0x10736 }, ::unicode::General_Category::Other_Letter }, // [311] LINEAR A SIGN AB001..LINEAR A SIGN A664
    Prop<::unicode::General_Category> { { 0x10737, 0x1073F }, ::unicode::General_Category::Unassigned }, // [9] <reserved-10737>..<reserved-1073F>
    Prop<::unicode::General_Category> { { 0x10740, 0x10755 }, ::unicode::General_Category::Other_Letter }, // [22] LINEAR A SIGN A701 A..LINEAR A SIGN A732 JE
    Prop<::unicode::General_Category> { { 0x10756, 0x1075F }, ::unicode::General_Category::Unassigned }, // [10] <reserved-10756>..<reserved-1075F>
    Prop<::unicode::General_Category> { { 0x10760, 0x10767 }, ::unicode::General_Category::Other_Letter }, // [8] LINEAR A SIGN A800..LINEAR A SIGN A807
    Prop<::unicode::General_Category> { { 0x10768, 0x1077F }, ::unicode::General_Category::Unassigned }, // [24] <reserved-10768>..<reserved-1077F>
    Prop<::unicode::General_Category> { { 0x10780, 0x10785 }, ::unicode::General_Category::Modifier_Letter }, // [6] MODIFIER LETTER SMALL CAPITAL AA..MODIFIER LETTER SMALL B WITH HOOK
    Prop<::unicode::General_Category> { { 0x10786, 0x10786 }, ::unicode::General_Category::Unassigned }, // <reserved-10786>
    Prop<::unicode::General_Category> { { 0x10787, 0x107B0 }, ::unicode::General_Category::Modifier_Letter }, // [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Prop<::unicode::General_Category> { { 0x107B1, 0x107B1 }, ::unicode::General_Category::Unassigned }, // <reserved-107B1>
    Prop<::unicode::General_Category> { { 0x107B2, 0x107BA }, ::unicode::General_Category::Modifier_Letter }, // [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Prop<::unicode::General_Category> { { 0x107BB, 0x107FF }, ::unicode::General_Category::Unassigned }, // [69] <reserved-107BB>..<reserved-107FF>
    Prop<::unicode::General_Category> { { 0x10800, 0x10805 }, ::unicode::General_Category::Other_Letter }, // [6] CYPRIOT SYLLABLE A..CYPRIOT SYLLABLE JA
    Prop<::unicode::General_Category> { { 0x10806, 0x10807 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-10806>..<reserved-10807>
    Prop<::unicode::General_Category> { { 0x10808, 0x10808 }, ::unicode::General_Category::Other_Letter }, // CYPRIOT SYLLABLE JO
    Prop<::unicode::General_Category> { { 0x10809, 0x10809 }, ::unicode::General_Category::Unassigned }, // <reserved-10809>
    Prop<::unicode::General_Category> { { 0x1080A, 0x10835 }, ::unicode::General_Category::Other_Letter }, // [44] CYPRIOT SYLLABLE KA..CYPRIOT SYLLABLE WO
    Prop<::unicode::General_Category> { { 0x10836, 0x10836 }, ::unicode::General_Category::Unassigned }, // <reserved-10836>
    Prop<::unicode::General_Category> { { 0x10837, 0x10838 }, ::unicode::General_Category::Other_Letter }, // [2] CYPRIOT SYLLABLE XA..CYPRIOT SYLLABLE XE
    Prop<::unicode::General_Category> { { 0x10839, 0x1083B }, ::unicode::General_Category::Unassigned }, // [3] <reserved-10839>..<reserved-1083B>
    Prop<::unicode::General_Category> { { 0x1083C, 0x1083C }, ::unicode::General_Category::Other_Letter }, // CYPRIOT SYLLABLE ZA
    Prop<::unicode::General_Category> { { 0x1083D, 0x1083E }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1083D>..<reserved-1083E>
    Prop<::unicode::General_Category> { { 0x1083F, 0x10855 }, ::unicode::General_Category::Other_Letter }, // [23] CYPRIOT SYLLABLE ZO..IMPERIAL ARAMAIC LETTER TAW
    Prop<::unicode::General_Category> { { 0x10856, 0x10856 }, ::unicode::General_Category::Unassigned }, // <reserved-10856>
    Prop<::unicode::General_Category> { { 0x10857, 0x10857 }, ::unicode::General_Category::Other_Punctuation }, // IMPERIAL ARAMAIC SECTION SIGN
    Prop<::unicode::General_Category> { { 0x10858, 0x1085F }, ::unicode::General_Category::Other_Number }, // [8] IMPERIAL ARAMAIC NUMBER ONE..IMPERIAL ARAMAIC NUMBER TEN THOUSAND
    Prop<::unicode::General_Category> { { 0x10860, 0x10876 }, ::unicode::General_Category::Other_Letter }, // [23] PALMYRENE LETTER ALEPH..PALMYRENE LETTER TAW
    Prop<::unicode::General_Category> { { 0x10877, 0x10878 }, ::unicode::General_Category::Other_Symbol }, // [2] PALMYRENE LEFT-POINTING FLEURON..PALMYRENE RIGHT-POINTING FLEURON
    Prop<::unicode::General_Category> { { 0x10879, 0x1087F }, ::unicode::General_Category::Other_Number }, // [7] PALMYRENE NUMBER ONE..PALMYRENE NUMBER TWENTY
    Prop<::unicode::General_Category> { { 0x10880, 0x1089E }, ::unicode::General_Category::Other_Letter }, // [31] NABATAEAN LETTER FINAL ALEPH..NABATAEAN LETTER TAW
    Prop<::unicode::General_Category> { { 0x1089F, 0x108A6 }, ::unicode::General_Category::Unassigned }, // [8] <reserved-1089F>..<reserved-108A6>
    Prop<::unicode::General_Category> { { 0x108A7, 0x108AF }, ::unicode::General_Category::Other_Number }, // [9] NABATAEAN NUMBER ONE..NABATAEAN NUMBER ONE HUNDRED
    Prop<::unicode::General_Category> { { 0x108B0, 0x108DF }, ::unicode::General_Category::Unassigned }, // [48] <reserved-108B0>..<reserved-108DF>
    Prop<::unicode::General_Category> { { 0x108E0, 0x108F2 }, ::unicode::General_Category::Other_Letter }, // [19] HATRAN LETTER ALEPH..HATRAN LETTER QOPH
    Prop<::unicode::General_Category> { { 0x108F3, 0x108F3 }, ::unicode::General_Category::Unassigned }, // <reserved-108F3>
    Prop<::unicode::General_Category> { { 0x108F4, 0x108F5 }, ::unicode::General_Category::Other_Letter }, // [2] HATRAN LETTER SHIN..HATRAN LETTER TAW
    Prop<::unicode::General_Category> { { 0x108F6, 0x108FA }, ::unicode::General_Category::Unassigned }, // [5] <reserved-108F6>..<reserved-108FA>
    Prop<::unicode::General_Category> { { 0x108FB, 0x108FF }, ::unicode::General_Category::Other_Number }, // [5] HATRAN NUMBER ONE..HATRAN NUMBER ONE HUNDRED
    Prop<::unicode::General_Category> { { 0x10900, 0x10915 }, ::unicode::General_Category::Other_Letter }, // [22] PHOENICIAN LETTER ALF..PHOENICIAN LETTER TAU
    Prop<::unicode::General_Category> { { 0x10916, 0x1091B }, ::unicode::General_Category::Other_Number }, // [6] PHOENICIAN NUMBER ONE..PHOENICIAN NUMBER THREE
    Prop<::unicode::General_Category> { { 0x1091C, 0x1091E }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1091C>..<reserved-1091E>
    Prop<::unicode::General_Category> { { 0x1091F, 0x1091F }, ::unicode::General_Category::Other_Punctuation }, // PHOENICIAN WORD SEPARATOR
    Prop<::unicode::General_Category> { { 0x10920, 0x10939 }, ::unicode::General_Category::Other_Letter }, // [26] LYDIAN LETTER A..LYDIAN LETTER C
    Prop<::unicode::General_Category> { { 0x1093A, 0x1093E }, ::unicode::General_Category::Unassigned }, // [5] <reserved-1093A>..<reserved-1093E>
    Prop<::unicode::General_Category> { { 0x1093F, 0x1093F }, ::unicode::General_Category::Other_Punctuation }, // LYDIAN TRIANGULAR MARK
    Prop<::unicode::General_Category> { { 0x10940, 0x10959 }, ::unicode::General_Category::Other_Letter }, // [26] SIDETIC LETTER N01..SIDETIC LETTER N26
    Prop<::unicode::General_Category> { { 0x1095A, 0x1097F }, ::unicode::General_Category::Unassigned }, // [38] <reserved-1095A>..<reserved-1097F>
    Prop<::unicode::General_Category> { { 0x10980, 0x109B7 }, ::unicode::General_Category::Other_Letter }, // [56] MEROITIC HIEROGLYPHIC LETTER A..MEROITIC CURSIVE LETTER DA
    Prop<::unicode::General_Category> { { 0x109B8, 0x109BB }, ::unicode::General_Category::Unassigned }, // [4] <reserved-109B8>..<reserved-109BB>
    Prop<::unicode::General_Category> { { 0x109BC, 0x109BD }, ::unicode::General_Category::Other_Number }, // [2] MEROITIC CURSIVE FRACTION ELEVEN TWELFTHS..MEROITIC CURSIVE FRACTION ONE HALF
    Prop<::unicode::General_Category> { { 0x109BE, 0x109BF }, ::unicode::General_Category::Other_Letter }, // [2] MEROITIC CURSIVE LOGOGRAM RMT..MEROITIC CURSIVE LOGOGRAM IMN
    Prop<::unicode::General_Category> { { 0x109C0, 0x109CF }, ::unicode::General_Category::Other_Number }, // [16] MEROITIC CURSIVE NUMBER ONE..MEROITIC CURSIVE NUMBER SEVENTY
    Prop<::unicode::General_Category> { { 0x109D0, 0x109D1 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-109D0>..<reserved-109D1>
    Prop<::unicode::General_Category> { { 0x109D2, 0x109FF }, ::unicode::General_Category::Other_Number }, // [46] MEROITIC CURSIVE NUMBER ONE HUNDRED..MEROITIC CURSIVE FRACTION TEN TWELFTHS
    Prop<::unicode::General_Category> { { 0x10A00, 0x10A00 }, ::unicode::General_Category::Other_Letter }, // KHAROSHTHI LETTER A
    Prop<::unicode::General_Category> { { 0x10A01, 0x10A03 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] KHAROSHTHI VOWEL SIGN I..KHAROSHTHI VOWEL SIGN VOCALIC R
    Prop<::unicode::General_Category> { { 0x10A04, 0x10A04 }, ::unicode::General_Category::Unassigned }, // <reserved-10A04>
    Prop<::unicode::General_Category> { { 0x10A05, 0x10A06 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] KHAROSHTHI VOWEL SIGN E..KHAROSHTHI VOWEL SIGN O
    Prop<::unicode::General_Category> { { 0x10A07, 0x10A0B }, ::unicode::General_Category::Unassigned }, // [5] <reserved-10A07>..<reserved-10A0B>
    Prop<::unicode::General_Category> { { 0x10A0C, 0x10A0F }, ::unicode::General_Category::Nonspacing_Mark }, // [4] KHAROSHTHI VOWEL LENGTH MARK..KHAROSHTHI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x10A10, 0x10A13 }, ::unicode::General_Category::Other_Letter }, // [4] KHAROSHTHI LETTER KA..KHAROSHTHI LETTER GHA
    Prop<::unicode::General_Category> { { 0x10A14, 0x10A14 }, ::unicode::General_Category::Unassigned }, // <reserved-10A14>
    Prop<::unicode::General_Category> { { 0x10A15, 0x10A17 }, ::unicode::General_Category::Other_Letter }, // [3] KHAROSHTHI LETTER CA..KHAROSHTHI LETTER JA
    Prop<::unicode::General_Category> { { 0x10A18, 0x10A18 }, ::unicode::General_Category::Unassigned }, // <reserved-10A18>
    Prop<::unicode::General_Category> { { 0x10A19, 0x10A35 }, ::unicode::General_Category::Other_Letter }, // [29] KHAROSHTHI LETTER NYA..KHAROSHTHI LETTER VHA
    Prop<::unicode::General_Category> { { 0x10A36, 0x10A37 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-10A36>..<reserved-10A37>
    Prop<::unicode::General_Category> { { 0x10A38, 0x10A3A }, ::unicode::General_Category::Nonspacing_Mark }, // [3] KHAROSHTHI SIGN BAR ABOVE..KHAROSHTHI SIGN DOT BELOW
    Prop<::unicode::General_Category> { { 0x10A3B, 0x10A3E }, ::unicode::General_Category::Unassigned }, // [4] <reserved-10A3B>..<reserved-10A3E>
    Prop<::unicode::General_Category> { { 0x10A3F, 0x10A3F }, ::unicode::General_Category::Nonspacing_Mark }, // KHAROSHTHI VIRAMA
    Prop<::unicode::General_Category> { { 0x10A40, 0x10A48 }, ::unicode::General_Category::Other_Number }, // [9] KHAROSHTHI DIGIT ONE..KHAROSHTHI FRACTION ONE HALF
    Prop<::unicode::General_Category> { { 0x10A49, 0x10A4F }, ::unicode::General_Category::Unassigned }, // [7] <reserved-10A49>..<reserved-10A4F>
    Prop<::unicode::General_Category> { { 0x10A50, 0x10A58 }, ::unicode::General_Category::Other_Punctuation }, // [9] KHAROSHTHI PUNCTUATION DOT..KHAROSHTHI PUNCTUATION LINES
    Prop<::unicode::General_Category> { { 0x10A59, 0x10A5F }, ::unicode::General_Category::Unassigned }, // [7] <reserved-10A59>..<reserved-10A5F>
    Prop<::unicode::General_Category> { { 0x10A60, 0x10A7C }, ::unicode::General_Category::Other_Letter }, // [29] OLD SOUTH ARABIAN LETTER HE..OLD SOUTH ARABIAN LETTER THETH
    Prop<::unicode::General_Category> { { 0x10A7D, 0x10A7E }, ::unicode::General_Category::Other_Number }, // [2] OLD SOUTH ARABIAN NUMBER ONE..OLD SOUTH ARABIAN NUMBER FIFTY
    Prop<::unicode::General_Category> { { 0x10A7F, 0x10A7F }, ::unicode::General_Category::Other_Punctuation }, // OLD SOUTH ARABIAN NUMERIC INDICATOR
    Prop<::unicode::General_Category> { { 0x10A80, 0x10A9C }, ::unicode::General_Category::Other_Letter }, // [29] OLD NORTH ARABIAN LETTER HEH..OLD NORTH ARABIAN LETTER ZAH
    Prop<::unicode::General_Category> { { 0x10A9D, 0x10A9F }, ::unicode::General_Category::Other_Number }, // [3] OLD NORTH ARABIAN NUMBER ONE..OLD NORTH ARABIAN NUMBER TWENTY
    Prop<::unicode::General_Category> { { 0x10AA0, 0x10ABF }, ::unicode::General_Category::Unassigned }, // [32] <reserved-10AA0>..<reserved-10ABF>
    Prop<::unicode::General_Category> { { 0x10AC0, 0x10AC7 }, ::unicode::General_Category::Other_Letter }, // [8] MANICHAEAN LETTER ALEPH..MANICHAEAN LETTER WAW
    Prop<::unicode::General_Category> { { 0x10AC8, 0x10AC8 }, ::unicode::General_Category::Other_Symbol }, // MANICHAEAN SIGN UD
    Prop<::unicode::General_Category> { { 0x10AC9, 0x10AE4 }, ::unicode::General_Category::Other_Letter }, // [28] MANICHAEAN LETTER ZAYIN..MANICHAEAN LETTER TAW
    Prop<::unicode::General_Category> { { 0x10AE5, 0x10AE6 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MANICHAEAN ABBREVIATION MARK ABOVE..MANICHAEAN ABBREVIATION MARK BELOW
    Prop<::unicode::General_Category> { { 0x10AE7, 0x10AEA }, ::unicode::General_Category::Unassigned }, // [4] <reserved-10AE7>..<reserved-10AEA>
    Prop<::unicode::General_Category> { { 0x10AEB, 0x10AEF }, ::unicode::General_Category::Other_Number }, // [5] MANICHAEAN NUMBER ONE..MANICHAEAN NUMBER ONE HUNDRED
    Prop<::unicode::General_Category> { { 0x10AF0, 0x10AF6 }, ::unicode::General_Category::Other_Punctuation }, // [7] MANICHAEAN PUNCTUATION STAR..MANICHAEAN PUNCTUATION LINE FILLER
    Prop<::unicode::General_Category> { { 0x10AF7, 0x10AFF }, ::unicode::General_Category::Unassigned }, // [9] <reserved-10AF7>..<reserved-10AFF>
    Prop<::unicode::General_Category> { { 0x10B00, 0x10B35 }, ::unicode::General_Category::Other_Letter }, // [54] AVESTAN LETTER A..AVESTAN LETTER HE
    Prop<::unicode::General_Category> { { 0x10B36, 0x10B38 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-10B36>..<reserved-10B38>
    Prop<::unicode::General_Category> { { 0x10B39, 0x10B3F }, ::unicode::General_Category::Other_Punctuation }, // [7] AVESTAN ABBREVIATION MARK..LARGE ONE RING OVER TWO RINGS PUNCTUATION
    Prop<::unicode::General_Category> { { 0x10B40, 0x10B55 }, ::unicode::General_Category::Other_Letter }, // [22] INSCRIPTIONAL PARTHIAN LETTER ALEPH..INSCRIPTIONAL PARTHIAN LETTER TAW
    Prop<::unicode::General_Category> { { 0x10B56, 0x10B57 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-10B56>..<reserved-10B57>
    Prop<::unicode::General_Category> { { 0x10B58, 0x10B5F }, ::unicode::General_Category::Other_Number }, // [8] INSCRIPTIONAL PARTHIAN NUMBER ONE..INSCRIPTIONAL PARTHIAN NUMBER ONE THOUSAND
    Prop<::unicode::General_Category> { { 0x10B60, 0x10B72 }, ::unicode::General_Category::Other_Letter }, // [19] INSCRIPTIONAL PAHLAVI LETTER ALEPH..INSCRIPTIONAL PAHLAVI LETTER TAW
    Prop<::unicode::General_Category> { { 0x10B73, 0x10B77 }, ::unicode::General_Category::Unassigned }, // [5] <reserved-10B73>..<reserved-10B77>
    Prop<::unicode::General_Category> { { 0x10B78, 0x10B7F }, ::unicode::General_Category::Other_Number }, // [8] INSCRIPTIONAL PAHLAVI NUMBER ONE..INSCRIPTIONAL PAHLAVI NUMBER ONE THOUSAND
    Prop<::unicode::General_Category> { { 0x10B80, 0x10B91 }, ::unicode::General_Category::Other_Letter }, // [18] PSALTER PAHLAVI LETTER ALEPH..PSALTER PAHLAVI LETTER TAW
    Prop<::unicode::General_Category> { { 0x10B92, 0x10B98 }, ::unicode::General_Category::Unassigned }, // [7] <reserved-10B92>..<reserved-10B98>
    Prop<::unicode::General_Category> { { 0x10B99, 0x10B9C }, ::unicode::General_Category::Other_Punctuation }, // [4] PSALTER PAHLAVI SECTION MARK..PSALTER PAHLAVI FOUR DOTS WITH DOT
    Prop<::unicode::General_Category> { { 0x10B9D, 0x10BA8 }, ::unicode::General_Category::Unassigned }, // [12] <reserved-10B9D>..<reserved-10BA8>
    Prop<::unicode::General_Category> { { 0x10BA9, 0x10BAF }, ::unicode::General_Category::Other_Number }, // [7] PSALTER PAHLAVI NUMBER ONE..PSALTER PAHLAVI NUMBER ONE HUNDRED
    Prop<::unicode::General_Category> { { 0x10BB0, 0x10BFF }, ::unicode::General_Category::Unassigned }, // [80] <reserved-10BB0>..<reserved-10BFF>
    Prop<::unicode::General_Category> { { 0x10C00, 0x10C48 }, ::unicode::General_Category::Other_Letter }, // [73] OLD TURKIC LETTER ORKHON A..OLD TURKIC LETTER ORKHON BASH
    Prop<::unicode::General_Category> { { 0x10C49, 0x10C7F }, ::unicode::General_Category::Unassigned }, // [55] <reserved-10C49>..<reserved-10C7F>
    Prop<::unicode::General_Category> { { 0x10C80, 0x10CB2 }, ::unicode::General_Category::Uppercase_Letter }, // [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Prop<::unicode::General_Category> { { 0x10CB3, 0x10CBF }, ::unicode::General_Category::Unassigned }, // [13] <reserved-10CB3>..<reserved-10CBF>
    Prop<::unicode::General_Category> { { 0x10CC0, 0x10CF2 }, ::unicode::General_Category::Lowercase_Letter }, // [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Prop<::unicode::General_Category> { { 0x10CF3, 0x10CF9 }, ::unicode::General_Category::Unassigned }, // [7] <reserved-10CF3>..<reserved-10CF9>
    Prop<::unicode::General_Category> { { 0x10CFA, 0x10CFF }, ::unicode::General_Category::Other_Number }, // [6] OLD HUNGARIAN NUMBER ONE..OLD HUNGARIAN NUMBER ONE THOUSAND
    Prop<::unicode::General_Category> { { 0x10D00, 0x10D23 }, ::unicode::General_Category::Other_Letter }, // [36] HANIFI ROHINGYA LETTER A..HANIFI ROHINGYA MARK NA KHONNA
    Prop<::unicode::General_Category> { { 0x10D24, 0x10D27 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] HANIFI ROHINGYA SIGN HARBAHAY..HANIFI ROHINGYA SIGN TASSI
    Prop<::unicode::General_Category> { { 0x10D28, 0x10D2F }, ::unicode::General_Category::Unassigned }, // [8] <reserved-10D28>..<reserved-10D2F>
    Prop<::unicode::General_Category> { { 0x10D30, 0x10D39 }, ::unicode::General_Category::Decimal_Number }, // [10] HANIFI ROHINGYA DIGIT ZERO..HANIFI ROHINGYA DIGIT NINE
    Prop<::unicode::General_Category> { { 0x10D3A, 0x10D3F }, ::unicode::General_Category::Unassigned }, // [6] <reserved-10D3A>..<reserved-10D3F>
    Prop<::unicode::General_Category> { { 0x10D40, 0x10D49 }, ::unicode::General_Category::Decimal_Number }, // [10] GARAY DIGIT ZERO..GARAY DIGIT NINE
    Prop<::unicode::General_Category> { { 0x10D4A, 0x10D4D }, ::unicode::General_Category::Other_Letter }, // [4] GARAY VOWEL SIGN A..GARAY VOWEL SIGN EE
    Prop<::unicode::General_Category> { { 0x10D4E, 0x10D4E }, ::unicode::General_Category::Modifier_Letter }, // GARAY VOWEL LENGTH MARK
    Prop<::unicode::General_Category> { { 0x10D4F, 0x10D4F }, ::unicode::General_Category::Other_Letter }, // GARAY SUKUN
    Prop<::unicode::General_Category> { { 0x10D50, 0x10D65 }, ::unicode::General_Category::Uppercase_Letter }, // [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Prop<::unicode::General_Category> { { 0x10D66, 0x10D68 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-10D66>..<reserved-10D68>
    Prop<::unicode::General_Category> { { 0x10D69, 0x10D6D }, ::unicode::General_Category::Nonspacing_Mark }, // [5] GARAY VOWEL SIGN E..GARAY CONSONANT NASALIZATION MARK
    Prop<::unicode::General_Category> { { 0x10D6E, 0x10D6E }, ::unicode::General_Category::Dash_Punctuation }, // GARAY HYPHEN
    Prop<::unicode::General_Category> { { 0x10D6F, 0x10D6F }, ::unicode::General_Category::Modifier_Letter }, // GARAY REDUPLICATION MARK
    Prop<::unicode::General_Category> { { 0x10D70, 0x10D85 }, ::unicode::General_Category::Lowercase_Letter }, // [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Prop<::unicode::General_Category> { { 0x10D86, 0x10D8D }, ::unicode::General_Category::Unassigned }, // [8] <reserved-10D86>..<reserved-10D8D>
    Prop<::unicode::General_Category> { { 0x10D8E, 0x10D8F }, ::unicode::General_Category::Math_Symbol }, // [2] GARAY PLUS SIGN..GARAY MINUS SIGN
    Prop<::unicode::General_Category> { { 0x10D90, 0x10E5F }, ::unicode::General_Category::Unassigned }, // [208] <reserved-10D90>..<reserved-10E5F>
    Prop<::unicode::General_Category> { { 0x10E60, 0x10E7E }, ::unicode::General_Category::Other_Number }, // [31] RUMI DIGIT ONE..RUMI FRACTION TWO THIRDS
    Prop<::unicode::General_Category> { { 0x10E7F, 0x10E7F }, ::unicode::General_Category::Unassigned }, // <reserved-10E7F>
    Prop<::unicode::General_Category> { { 0x10E80, 0x10EA9 }, ::unicode::General_Category::Other_Letter }, // [42] YEZIDI LETTER ELIF..YEZIDI LETTER ET
    Prop<::unicode::General_Category> { { 0x10EAA, 0x10EAA }, ::unicode::General_Category::Unassigned }, // <reserved-10EAA>
    Prop<::unicode::General_Category> { { 0x10EAB, 0x10EAC }, ::unicode::General_Category::Nonspacing_Mark }, // [2] YEZIDI COMBINING HAMZA MARK..YEZIDI COMBINING MADDA MARK
    Prop<::unicode::General_Category> { { 0x10EAD, 0x10EAD }, ::unicode::General_Category::Dash_Punctuation }, // YEZIDI HYPHENATION MARK
    Prop<::unicode::General_Category> { { 0x10EAE, 0x10EAF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-10EAE>..<reserved-10EAF>
    Prop<::unicode::General_Category> { { 0x10EB0, 0x10EB1 }, ::unicode::General_Category::Other_Letter }, // [2] YEZIDI LETTER LAM WITH DOT ABOVE..YEZIDI LETTER YOT WITH CIRCUMFLEX ABOVE
    Prop<::unicode::General_Category> { { 0x10EB2, 0x10EC1 }, ::unicode::General_Category::Unassigned }, // [16] <reserved-10EB2>..<reserved-10EC1>
    Prop<::unicode::General_Category> { { 0x10EC2, 0x10EC4 }, ::unicode::General_Category::Other_Letter }, // [3] ARABIC LETTER DAL WITH TWO DOTS VERTICALLY BELOW..ARABIC LETTER KAF WITH TWO DOTS VERTICALLY BELOW
    Prop<::unicode::General_Category> { { 0x10EC5, 0x10EC5 }, ::unicode::General_Category::Modifier_Letter }, // ARABIC SMALL YEH BARREE WITH TWO DOTS BELOW
    Prop<::unicode::General_Category> { { 0x10EC6, 0x10EC7 }, ::unicode::General_Category::Other_Letter }, // [2] ARABIC LETTER THIN NOON..ARABIC LETTER YEH WITH FOUR DOTS BELOW
    Prop<::unicode::General_Category> { { 0x10EC8, 0x10ECF }, ::unicode::General_Category::Unassigned }, // [8] <reserved-10EC8>..<reserved-10ECF>
    Prop<::unicode::General_Category> { { 0x10ED0, 0x10ED0 }, ::unicode::General_Category::Other_Punctuation }, // ARABIC BIBLICAL END OF VERSE
    Prop<::unicode::General_Category> { { 0x10ED1, 0x10ED8 }, ::unicode::General_Category::Other_Symbol }, // [8] ARABIC LIGATURE ALAYHAA AS-SALAATU WAS-SALAAM..ARABIC LIGATURE NAWWARA ALLAAHU MARQADAH
    Prop<::unicode::General_Category> { { 0x10ED9, 0x10EF9 }, ::unicode::General_Category::Unassigned }, // [33] <reserved-10ED9>..<reserved-10EF9>
    Prop<::unicode::General_Category> { { 0x10EFA, 0x10EFF }, ::unicode::General_Category::Nonspacing_Mark }, // [6] ARABIC DOUBLE VERTICAL BAR BELOW..ARABIC SMALL LOW WORD MADDA
    Prop<::unicode::General_Category> { { 0x10F00, 0x10F1C }, ::unicode::General_Category::Other_Letter }, // [29] OLD SOGDIAN LETTER ALEPH..OLD SOGDIAN LETTER FINAL TAW WITH VERTICAL TAIL
    Prop<::unicode::General_Category> { { 0x10F1D, 0x10F26 }, ::unicode::General_Category::Other_Number }, // [10] OLD SOGDIAN NUMBER ONE..OLD SOGDIAN FRACTION ONE HALF
    Prop<::unicode::General_Category> { { 0x10F27, 0x10F27 }, ::unicode::General_Category::Other_Letter }, // OLD SOGDIAN LIGATURE AYIN-DALETH
    Prop<::unicode::General_Category> { { 0x10F28, 0x10F2F }, ::unicode::General_Category::Unassigned }, // [8] <reserved-10F28>..<reserved-10F2F>
    Prop<::unicode::General_Category> { { 0x10F30, 0x10F45 }, ::unicode::General_Category::Other_Letter }, // [22] SOGDIAN LETTER ALEPH..SOGDIAN INDEPENDENT SHIN
    Prop<::unicode::General_Category> { { 0x10F46, 0x10F50 }, ::unicode::General_Category::Nonspacing_Mark }, // [11] SOGDIAN COMBINING DOT BELOW..SOGDIAN COMBINING STROKE BELOW
    Prop<::unicode::General_Category> { { 0x10F51, 0x10F54 }, ::unicode::General_Category::Other_Number }, // [4] SOGDIAN NUMBER ONE..SOGDIAN NUMBER ONE HUNDRED
    Prop<::unicode::General_Category> { { 0x10F55, 0x10F59 }, ::unicode::General_Category::Other_Punctuation }, // [5] SOGDIAN PUNCTUATION TWO VERTICAL BARS..SOGDIAN PUNCTUATION HALF CIRCLE WITH DOT
    Prop<::unicode::General_Category> { { 0x10F5A, 0x10F6F }, ::unicode::General_Category::Unassigned }, // [22] <reserved-10F5A>..<reserved-10F6F>
    Prop<::unicode::General_Category> { { 0x10F70, 0x10F81 }, ::unicode::General_Category::Other_Letter }, // [18] OLD UYGHUR LETTER ALEPH..OLD UYGHUR LETTER LESH
    Prop<::unicode::General_Category> { { 0x10F82, 0x10F85 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] OLD UYGHUR COMBINING DOT ABOVE..OLD UYGHUR COMBINING TWO DOTS BELOW
    Prop<::unicode::General_Category> { { 0x10F86, 0x10F89 }, ::unicode::General_Category::Other_Punctuation }, // [4] OLD UYGHUR PUNCTUATION BAR..OLD UYGHUR PUNCTUATION FOUR DOTS
    Prop<::unicode::General_Category> { { 0x10F8A, 0x10FAF }, ::unicode::General_Category::Unassigned }, // [38] <reserved-10F8A>..<reserved-10FAF>
    Prop<::unicode::General_Category> { { 0x10FB0, 0x10FC4 }, ::unicode::General_Category::Other_Letter }, // [21] CHORASMIAN LETTER ALEPH..CHORASMIAN LETTER TAW
    Prop<::unicode::General_Category> { { 0x10FC5, 0x10FCB }, ::unicode::General_Category::Other_Number }, // [7] CHORASMIAN NUMBER ONE..CHORASMIAN NUMBER ONE HUNDRED
    Prop<::unicode::General_Category> { { 0x10FCC, 0x10FDF }, ::unicode::General_Category::Unassigned }, // [20] <reserved-10FCC>..<reserved-10FDF>
    Prop<::unicode::General_Category> { { 0x10FE0, 0x10FF6 }, ::unicode::General_Category::Other_Letter }, // [23] ELYMAIC LETTER ALEPH..ELYMAIC LIGATURE ZAYIN-YODH
    Prop<::unicode::General_Category> { { 0x10FF7, 0x10FFF }, ::unicode::General_Category::Unassigned }, // [9] <reserved-10FF7>..<reserved-10FFF>
    Prop<::unicode::General_Category> { { 0x11000, 0x11000 }, ::unicode::General_Category::Spacing_Mark }, // BRAHMI SIGN CANDRABINDU
    Prop<::unicode::General_Category> { { 0x11001, 0x11001 }, ::unicode::General_Category::Nonspacing_Mark }, // BRAHMI SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x11002, 0x11002 }, ::unicode::General_Category::Spacing_Mark }, // BRAHMI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x11003, 0x11037 }, ::unicode::General_Category::Other_Letter }, // [53] BRAHMI SIGN JIHVAMULIYA..BRAHMI LETTER OLD TAMIL NNNA
    Prop<::unicode::General_Category> { { 0x11038, 0x11046 }, ::unicode::General_Category::Nonspacing_Mark }, // [15] BRAHMI VOWEL SIGN AA..BRAHMI VIRAMA
    Prop<::unicode::General_Category> { { 0x11047, 0x1104D }, ::unicode::General_Category::Other_Punctuation }, // [7] BRAHMI DANDA..BRAHMI PUNCTUATION LOTUS
    Prop<::unicode::General_Category> { { 0x1104E, 0x11051 }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1104E>..<reserved-11051>
    Prop<::unicode::General_Category> { { 0x11052, 0x11065 }, ::unicode::General_Category::Other_Number }, // [20] BRAHMI NUMBER ONE..BRAHMI NUMBER ONE THOUSAND
    Prop<::unicode::General_Category> { { 0x11066, 0x1106F }, ::unicode::General_Category::Decimal_Number }, // [10] BRAHMI DIGIT ZERO..BRAHMI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x11070, 0x11070 }, ::unicode::General_Category::Nonspacing_Mark }, // BRAHMI SIGN OLD TAMIL VIRAMA
    Prop<::unicode::General_Category> { { 0x11071, 0x11072 }, ::unicode::General_Category::Other_Letter }, // [2] BRAHMI LETTER OLD TAMIL SHORT E..BRAHMI LETTER OLD TAMIL SHORT O
    Prop<::unicode::General_Category> { { 0x11073, 0x11074 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] BRAHMI VOWEL SIGN OLD TAMIL SHORT E..BRAHMI VOWEL SIGN OLD TAMIL SHORT O
    Prop<::unicode::General_Category> { { 0x11075, 0x11075 }, ::unicode::General_Category::Other_Letter }, // BRAHMI LETTER OLD TAMIL LLA
    Prop<::unicode::General_Category> { { 0x11076, 0x1107E }, ::unicode::General_Category::Unassigned }, // [9] <reserved-11076>..<reserved-1107E>
    Prop<::unicode::General_Category> { { 0x1107F, 0x11081 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] BRAHMI NUMBER JOINER..KAITHI SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x11082, 0x11082 }, ::unicode::General_Category::Spacing_Mark }, // KAITHI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x11083, 0x110AF }, ::unicode::General_Category::Other_Letter }, // [45] KAITHI LETTER A..KAITHI LETTER HA
    Prop<::unicode::General_Category> { { 0x110B0, 0x110B2 }, ::unicode::General_Category::Spacing_Mark }, // [3] KAITHI VOWEL SIGN AA..KAITHI VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x110B3, 0x110B6 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] KAITHI VOWEL SIGN U..KAITHI VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x110B7, 0x110B8 }, ::unicode::General_Category::Spacing_Mark }, // [2] KAITHI VOWEL SIGN O..KAITHI VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x110B9, 0x110BA }, ::unicode::General_Category::Nonspacing_Mark }, // [2] KAITHI SIGN VIRAMA..KAITHI SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x110BB, 0x110BC }, ::unicode::General_Category::Other_Punctuation }, // [2] KAITHI ABBREVIATION SIGN..KAITHI ENUMERATION SIGN
    Prop<::unicode::General_Category> { { 0x110BD, 0x110BD }, ::unicode::General_Category::Format }, // KAITHI NUMBER SIGN
    Prop<::unicode::General_Category> { { 0x110BE, 0x110C1 }, ::unicode::General_Category::Other_Punctuation }, // [4] KAITHI SECTION MARK..KAITHI DOUBLE DANDA
    Prop<::unicode::General_Category> { { 0x110C2, 0x110C2 }, ::unicode::General_Category::Nonspacing_Mark }, // KAITHI VOWEL SIGN VOCALIC R
    Prop<::unicode::General_Category> { { 0x110C3, 0x110CC }, ::unicode::General_Category::Unassigned }, // [10] <reserved-110C3>..<reserved-110CC>
    Prop<::unicode::General_Category> { { 0x110CD, 0x110CD }, ::unicode::General_Category::Format }, // KAITHI NUMBER SIGN ABOVE
    Prop<::unicode::General_Category> { { 0x110CE, 0x110CF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-110CE>..<reserved-110CF>
    Prop<::unicode::General_Category> { { 0x110D0, 0x110E8 }, ::unicode::General_Category::Other_Letter }, // [25] SORA SOMPENG LETTER SAH..SORA SOMPENG LETTER MAE
    Prop<::unicode::General_Category> { { 0x110E9, 0x110EF }, ::unicode::General_Category::Unassigned }, // [7] <reserved-110E9>..<reserved-110EF>
    Prop<::unicode::General_Category> { { 0x110F0, 0x110F9 }, ::unicode::General_Category::Decimal_Number }, // [10] SORA SOMPENG DIGIT ZERO..SORA SOMPENG DIGIT NINE
    Prop<::unicode::General_Category> { { 0x110FA, 0x110FF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-110FA>..<reserved-110FF>
    Prop<::unicode::General_Category> { { 0x11100, 0x11102 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] CHAKMA SIGN CANDRABINDU..CHAKMA SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x11103, 0x11126 }, ::unicode::General_Category::Other_Letter }, // [36] CHAKMA LETTER AA..CHAKMA LETTER HAA
    Prop<::unicode::General_Category> { { 0x11127, 0x1112B }, ::unicode::General_Category::Nonspacing_Mark }, // [5] CHAKMA VOWEL SIGN A..CHAKMA VOWEL SIGN UU
    Prop<::unicode::General_Category> { { 0x1112C, 0x1112C }, ::unicode::General_Category::Spacing_Mark }, // CHAKMA VOWEL SIGN E
    Prop<::unicode::General_Category> { { 0x1112D, 0x11134 }, ::unicode::General_Category::Nonspacing_Mark }, // [8] CHAKMA VOWEL SIGN AI..CHAKMA MAAYYAA
    Prop<::unicode::General_Category> { { 0x11135, 0x11135 }, ::unicode::General_Category::Unassigned }, // <reserved-11135>
    Prop<::unicode::General_Category> { { 0x11136, 0x1113F }, ::unicode::General_Category::Decimal_Number }, // [10] CHAKMA DIGIT ZERO..CHAKMA DIGIT NINE
    Prop<::unicode::General_Category> { { 0x11140, 0x11143 }, ::unicode::General_Category::Other_Punctuation }, // [4] CHAKMA SECTION MARK..CHAKMA QUESTION MARK
    Prop<::unicode::General_Category> { { 0x11144, 0x11144 }, ::unicode::General_Category::Other_Letter }, // CHAKMA LETTER LHAA
    Prop<::unicode::General_Category> { { 0x11145, 0x11146 }, ::unicode::General_Category::Spacing_Mark }, // [2] CHAKMA VOWEL SIGN AA..CHAKMA VOWEL SIGN EI
    Prop<::unicode::General_Category> { { 0x11147, 0x11147 }, ::unicode::General_Category::Other_Letter }, // CHAKMA LETTER VAA
    Prop<::unicode::General_Category> { { 0x11148, 0x1114F }, ::unicode::General_Category::Unassigned }, // [8] <reserved-11148>..<reserved-1114F>
    Prop<::unicode::General_Category> { { 0x11150, 0x11172 }, ::unicode::General_Category::Other_Letter }, // [35] MAHAJANI LETTER A..MAHAJANI LETTER RRA
    Prop<::unicode::General_Category> { { 0x11173, 0x11173 }, ::unicode::General_Category::Nonspacing_Mark }, // MAHAJANI SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x11174, 0x11175 }, ::unicode::General_Category::Other_Punctuation }, // [2] MAHAJANI ABBREVIATION SIGN..MAHAJANI SECTION MARK
    Prop<::unicode::General_Category> { { 0x11176, 0x11176 }, ::unicode::General_Category::Other_Letter }, // MAHAJANI LIGATURE SHRI
    Prop<::unicode::General_Category> { { 0x11177, 0x1117F }, ::unicode::General_Category::Unassigned }, // [9] <reserved-11177>..<reserved-1117F>
    Prop<::unicode::General_Category> { { 0x11180, 0x11181 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] SHARADA SIGN CANDRABINDU..SHARADA SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x11182, 0x11182 }, ::unicode::General_Category::Spacing_Mark }, // SHARADA SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x11183, 0x111B2 }, ::unicode::General_Category::Other_Letter }, // [48] SHARADA LETTER A..SHARADA LETTER HA
    Prop<::unicode::General_Category> { { 0x111B3, 0x111B5 }, ::unicode::General_Category::Spacing_Mark }, // [3] SHARADA VOWEL SIGN AA..SHARADA VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x111B6, 0x111BE }, ::unicode::General_Category::Nonspacing_Mark }, // [9] SHARADA VOWEL SIGN U..SHARADA VOWEL SIGN O
    Prop<::unicode::General_Category> { { 0x111BF, 0x111C0 }, ::unicode::General_Category::Spacing_Mark }, // [2] SHARADA VOWEL SIGN AU..SHARADA SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x111C1, 0x111C4 }, ::unicode::General_Category::Other_Letter }, // [4] SHARADA SIGN AVAGRAHA..SHARADA OM
    Prop<::unicode::General_Category> { { 0x111C5, 0x111C8 }, ::unicode::General_Category::Other_Punctuation }, // [4] SHARADA DANDA..SHARADA SEPARATOR
    Prop<::unicode::General_Category> { { 0x111C9, 0x111CC }, ::unicode::General_Category::Nonspacing_Mark }, // [4] SHARADA SANDHI MARK..SHARADA EXTRA SHORT VOWEL MARK
    Prop<::unicode::General_Category> { { 0x111CD, 0x111CD }, ::unicode::General_Category::Other_Punctuation }, // SHARADA SUTRA MARK
    Prop<::unicode::General_Category> { { 0x111CE, 0x111CE }, ::unicode::General_Category::Spacing_Mark }, // SHARADA VOWEL SIGN PRISHTHAMATRA E
    Prop<::unicode::General_Category> { { 0x111CF, 0x111CF }, ::unicode::General_Category::Nonspacing_Mark }, // SHARADA SIGN INVERTED CANDRABINDU
    Prop<::unicode::General_Category> { { 0x111D0, 0x111D9 }, ::unicode::General_Category::Decimal_Number }, // [10] SHARADA DIGIT ZERO..SHARADA DIGIT NINE
    Prop<::unicode::General_Category> { { 0x111DA, 0x111DA }, ::unicode::General_Category::Other_Letter }, // SHARADA EKAM
    Prop<::unicode::General_Category> { { 0x111DB, 0x111DB }, ::unicode::General_Category::Other_Punctuation }, // SHARADA SIGN SIDDHAM
    Prop<::unicode::General_Category> { { 0x111DC, 0x111DC }, ::unicode::General_Category::Other_Letter }, // SHARADA HEADSTROKE
    Prop<::unicode::General_Category> { { 0x111DD, 0x111DF }, ::unicode::General_Category::Other_Punctuation }, // [3] SHARADA CONTINUATION SIGN..SHARADA SECTION MARK-2
    Prop<::unicode::General_Category> { { 0x111E0, 0x111E0 }, ::unicode::General_Category::Unassigned }, // <reserved-111E0>
    Prop<::unicode::General_Category> { { 0x111E1, 0x111F4 }, ::unicode::General_Category::Other_Number }, // [20] SINHALA ARCHAIC DIGIT ONE..SINHALA ARCHAIC NUMBER ONE THOUSAND
    Prop<::unicode::General_Category> { { 0x111F5, 0x111FF }, ::unicode::General_Category::Unassigned }, // [11] <reserved-111F5>..<reserved-111FF>
    Prop<::unicode::General_Category> { { 0x11200, 0x11211 }, ::unicode::General_Category::Other_Letter }, // [18] KHOJKI LETTER A..KHOJKI LETTER JJA
    Prop<::unicode::General_Category> { { 0x11212, 0x11212 }, ::unicode::General_Category::Unassigned }, // <reserved-11212>
    Prop<::unicode::General_Category> { { 0x11213, 0x1122B }, ::unicode::General_Category::Other_Letter }, // [25] KHOJKI LETTER NYA..KHOJKI LETTER LLA
    Prop<::unicode::General_Category> { { 0x1122C, 0x1122E }, ::unicode::General_Category::Spacing_Mark }, // [3] KHOJKI VOWEL SIGN AA..KHOJKI VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x1122F, 0x11231 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] KHOJKI VOWEL SIGN U..KHOJKI VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x11232, 0x11233 }, ::unicode::General_Category::Spacing_Mark }, // [2] KHOJKI VOWEL SIGN O..KHOJKI VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x11234, 0x11234 }, ::unicode::General_Category::Nonspacing_Mark }, // KHOJKI SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x11235, 0x11235 }, ::unicode::General_Category::Spacing_Mark }, // KHOJKI SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x11236, 0x11237 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] KHOJKI SIGN NUKTA..KHOJKI SIGN SHADDA
    Prop<::unicode::General_Category> { { 0x11238, 0x1123D }, ::unicode::General_Category::Other_Punctuation }, // [6] KHOJKI DANDA..KHOJKI ABBREVIATION SIGN
    Prop<::unicode::General_Category> { { 0x1123E, 0x1123E }, ::unicode::General_Category::Nonspacing_Mark }, // KHOJKI SIGN SUKUN
    Prop<::unicode::General_Category> { { 0x1123F, 0x11240 }, ::unicode::General_Category::Other_Letter }, // [2] KHOJKI LETTER QA..KHOJKI LETTER SHORT I
    Prop<::unicode::General_Category> { { 0x11241, 0x11241 }, ::unicode::General_Category::Nonspacing_Mark }, // KHOJKI VOWEL SIGN VOCALIC R
    Prop<::unicode::General_Category> { { 0x11242, 0x1127F }, ::unicode::General_Category::Unassigned }, // [62] <reserved-11242>..<reserved-1127F>
    Prop<::unicode::General_Category> { { 0x11280, 0x11286 }, ::unicode::General_Category::Other_Letter }, // [7] MULTANI LETTER A..MULTANI LETTER GA
    Prop<::unicode::General_Category> { { 0x11287, 0x11287 }, ::unicode::General_Category::Unassigned }, // <reserved-11287>
    Prop<::unicode::General_Category> { { 0x11288, 0x11288 }, ::unicode::General_Category::Other_Letter }, // MULTANI LETTER GHA
    Prop<::unicode::General_Category> { { 0x11289, 0x11289 }, ::unicode::General_Category::Unassigned }, // <reserved-11289>
    Prop<::unicode::General_Category> { { 0x1128A, 0x1128D }, ::unicode::General_Category::Other_Letter }, // [4] MULTANI LETTER CA..MULTANI LETTER JJA
    Prop<::unicode::General_Category> { { 0x1128E, 0x1128E }, ::unicode::General_Category::Unassigned }, // <reserved-1128E>
    Prop<::unicode::General_Category> { { 0x1128F, 0x1129D }, ::unicode::General_Category::Other_Letter }, // [15] MULTANI LETTER NYA..MULTANI LETTER BA
    Prop<::unicode::General_Category> { { 0x1129E, 0x1129E }, ::unicode::General_Category::Unassigned }, // <reserved-1129E>
    Prop<::unicode::General_Category> { { 0x1129F, 0x112A8 }, ::unicode::General_Category::Other_Letter }, // [10] MULTANI LETTER BHA..MULTANI LETTER RHA
    Prop<::unicode::General_Category> { { 0x112A9, 0x112A9 }, ::unicode::General_Category::Other_Punctuation }, // MULTANI SECTION MARK
    Prop<::unicode::General_Category> { { 0x112AA, 0x112AF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-112AA>..<reserved-112AF>
    Prop<::unicode::General_Category> { { 0x112B0, 0x112DE }, ::unicode::General_Category::Other_Letter }, // [47] KHUDAWADI LETTER A..KHUDAWADI LETTER HA
    Prop<::unicode::General_Category> { { 0x112DF, 0x112DF }, ::unicode::General_Category::Nonspacing_Mark }, // KHUDAWADI SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x112E0, 0x112E2 }, ::unicode::General_Category::Spacing_Mark }, // [3] KHUDAWADI VOWEL SIGN AA..KHUDAWADI VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x112E3, 0x112EA }, ::unicode::General_Category::Nonspacing_Mark }, // [8] KHUDAWADI VOWEL SIGN U..KHUDAWADI SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x112EB, 0x112EF }, ::unicode::General_Category::Unassigned }, // [5] <reserved-112EB>..<reserved-112EF>
    Prop<::unicode::General_Category> { { 0x112F0, 0x112F9 }, ::unicode::General_Category::Decimal_Number }, // [10] KHUDAWADI DIGIT ZERO..KHUDAWADI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x112FA, 0x112FF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-112FA>..<reserved-112FF>
    Prop<::unicode::General_Category> { { 0x11300, 0x11301 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] GRANTHA SIGN COMBINING ANUSVARA ABOVE..GRANTHA SIGN CANDRABINDU
    Prop<::unicode::General_Category> { { 0x11302, 0x11303 }, ::unicode::General_Category::Spacing_Mark }, // [2] GRANTHA SIGN ANUSVARA..GRANTHA SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x11304, 0x11304 }, ::unicode::General_Category::Unassigned }, // <reserved-11304>
    Prop<::unicode::General_Category> { { 0x11305, 0x1130C }, ::unicode::General_Category::Other_Letter }, // [8] GRANTHA LETTER A..GRANTHA LETTER VOCALIC L
    Prop<::unicode::General_Category> { { 0x1130D, 0x1130E }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1130D>..<reserved-1130E>
    Prop<::unicode::General_Category> { { 0x1130F, 0x11310 }, ::unicode::General_Category::Other_Letter }, // [2] GRANTHA LETTER EE..GRANTHA LETTER AI
    Prop<::unicode::General_Category> { { 0x11311, 0x11312 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-11311>..<reserved-11312>
    Prop<::unicode::General_Category> { { 0x11313, 0x11328 }, ::unicode::General_Category::Other_Letter }, // [22] GRANTHA LETTER OO..GRANTHA LETTER NA
    Prop<::unicode::General_Category> { { 0x11329, 0x11329 }, ::unicode::General_Category::Unassigned }, // <reserved-11329>
    Prop<::unicode::General_Category> { { 0x1132A, 0x11330 }, ::unicode::General_Category::Other_Letter }, // [7] GRANTHA LETTER PA..GRANTHA LETTER RA
    Prop<::unicode::General_Category> { { 0x11331, 0x11331 }, ::unicode::General_Category::Unassigned }, // <reserved-11331>
    Prop<::unicode::General_Category> { { 0x11332, 0x11333 }, ::unicode::General_Category::Other_Letter }, // [2] GRANTHA LETTER LA..GRANTHA LETTER LLA
    Prop<::unicode::General_Category> { { 0x11334, 0x11334 }, ::unicode::General_Category::Unassigned }, // <reserved-11334>
    Prop<::unicode::General_Category> { { 0x11335, 0x11339 }, ::unicode::General_Category::Other_Letter }, // [5] GRANTHA LETTER VA..GRANTHA LETTER HA
    Prop<::unicode::General_Category> { { 0x1133A, 0x1133A }, ::unicode::General_Category::Unassigned }, // <reserved-1133A>
    Prop<::unicode::General_Category> { { 0x1133B, 0x1133C }, ::unicode::General_Category::Nonspacing_Mark }, // [2] COMBINING BINDU BELOW..GRANTHA SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x1133D, 0x1133D }, ::unicode::General_Category::Other_Letter }, // GRANTHA SIGN AVAGRAHA
    Prop<::unicode::General_Category> { { 0x1133E, 0x1133F }, ::unicode::General_Category::Spacing_Mark }, // [2] GRANTHA VOWEL SIGN AA..GRANTHA VOWEL SIGN I
    Prop<::unicode::General_Category> { { 0x11340, 0x11340 }, ::unicode::General_Category::Nonspacing_Mark }, // GRANTHA VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x11341, 0x11344 }, ::unicode::General_Category::Spacing_Mark }, // [4] GRANTHA VOWEL SIGN U..GRANTHA VOWEL SIGN VOCALIC RR
    Prop<::unicode::General_Category> { { 0x11345, 0x11346 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-11345>..<reserved-11346>
    Prop<::unicode::General_Category> { { 0x11347, 0x11348 }, ::unicode::General_Category::Spacing_Mark }, // [2] GRANTHA VOWEL SIGN EE..GRANTHA VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x11349, 0x1134A }, ::unicode::General_Category::Unassigned }, // [2] <reserved-11349>..<reserved-1134A>
    Prop<::unicode::General_Category> { { 0x1134B, 0x1134D }, ::unicode::General_Category::Spacing_Mark }, // [3] GRANTHA VOWEL SIGN OO..GRANTHA SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x1134E, 0x1134F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1134E>..<reserved-1134F>
    Prop<::unicode::General_Category> { { 0x11350, 0x11350 }, ::unicode::General_Category::Other_Letter }, // GRANTHA OM
    Prop<::unicode::General_Category> { { 0x11351, 0x11356 }, ::unicode::General_Category::Unassigned }, // [6] <reserved-11351>..<reserved-11356>
    Prop<::unicode::General_Category> { { 0x11357, 0x11357 }, ::unicode::General_Category::Spacing_Mark }, // GRANTHA AU LENGTH MARK
    Prop<::unicode::General_Category> { { 0x11358, 0x1135C }, ::unicode::General_Category::Unassigned }, // [5] <reserved-11358>..<reserved-1135C>
    Prop<::unicode::General_Category> { { 0x1135D, 0x11361 }, ::unicode::General_Category::Other_Letter }, // [5] GRANTHA SIGN PLUTA..GRANTHA LETTER VOCALIC LL
    Prop<::unicode::General_Category> { { 0x11362, 0x11363 }, ::unicode::General_Category::Spacing_Mark }, // [2] GRANTHA VOWEL SIGN VOCALIC L..GRANTHA VOWEL SIGN VOCALIC LL
    Prop<::unicode::General_Category> { { 0x11364, 0x11365 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-11364>..<reserved-11365>
    Prop<::unicode::General_Category> { { 0x11366, 0x1136C }, ::unicode::General_Category::Nonspacing_Mark }, // [7] COMBINING GRANTHA DIGIT ZERO..COMBINING GRANTHA DIGIT SIX
    Prop<::unicode::General_Category> { { 0x1136D, 0x1136F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1136D>..<reserved-1136F>
    Prop<::unicode::General_Category> { { 0x11370, 0x11374 }, ::unicode::General_Category::Nonspacing_Mark }, // [5] COMBINING GRANTHA LETTER A..COMBINING GRANTHA LETTER PA
    Prop<::unicode::General_Category> { { 0x11375, 0x1137F }, ::unicode::General_Category::Unassigned }, // [11] <reserved-11375>..<reserved-1137F>
    Prop<::unicode::General_Category> { { 0x11380, 0x11389 }, ::unicode::General_Category::Other_Letter }, // [10] TULU-TIGALARI LETTER A..TULU-TIGALARI LETTER VOCALIC LL
    Prop<::unicode::General_Category> { { 0x1138A, 0x1138A }, ::unicode::General_Category::Unassigned }, // <reserved-1138A>
    Prop<::unicode::General_Category> { { 0x1138B, 0x1138B }, ::unicode::General_Category::Other_Letter }, // TULU-TIGALARI LETTER EE
    Prop<::unicode::General_Category> { { 0x1138C, 0x1138D }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1138C>..<reserved-1138D>
    Prop<::unicode::General_Category> { { 0x1138E, 0x1138E }, ::unicode::General_Category::Other_Letter }, // TULU-TIGALARI LETTER AI
    Prop<::unicode::General_Category> { { 0x1138F, 0x1138F }, ::unicode::General_Category::Unassigned }, // <reserved-1138F>
    Prop<::unicode::General_Category> { { 0x11390, 0x113B5 }, ::unicode::General_Category::Other_Letter }, // [38] TULU-TIGALARI LETTER OO..TULU-TIGALARI LETTER LLLA
    Prop<::unicode::General_Category> { { 0x113B6, 0x113B6 }, ::unicode::General_Category::Unassigned }, // <reserved-113B6>
    Prop<::unicode::General_Category> { { 0x113B7, 0x113B7 }, ::unicode::General_Category::Other_Letter }, // TULU-TIGALARI SIGN AVAGRAHA
    Prop<::unicode::General_Category> { { 0x113B8, 0x113BA }, ::unicode::General_Category::Spacing_Mark }, // [3] TULU-TIGALARI VOWEL SIGN AA..TULU-TIGALARI VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x113BB, 0x113C0 }, ::unicode::General_Category::Nonspacing_Mark }, // [6] TULU-TIGALARI VOWEL SIGN U..TULU-TIGALARI VOWEL SIGN VOCALIC LL
    Prop<::unicode::General_Category> { { 0x113C1, 0x113C1 }, ::unicode::General_Category::Unassigned }, // <reserved-113C1>
    Prop<::unicode::General_Category> { { 0x113C2, 0x113C2 }, ::unicode::General_Category::Spacing_Mark }, // TULU-TIGALARI VOWEL SIGN EE
    Prop<::unicode::General_Category> { { 0x113C3, 0x113C4 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-113C3>..<reserved-113C4>
    Prop<::unicode::General_Category> { { 0x113C5, 0x113C5 }, ::unicode::General_Category::Spacing_Mark }, // TULU-TIGALARI VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x113C6, 0x113C6 }, ::unicode::General_Category::Unassigned }, // <reserved-113C6>
    Prop<::unicode::General_Category> { { 0x113C7, 0x113CA }, ::unicode::General_Category::Spacing_Mark }, // [4] TULU-TIGALARI VOWEL SIGN OO..TULU-TIGALARI SIGN CANDRA ANUNASIKA
    Prop<::unicode::General_Category> { { 0x113CB, 0x113CB }, ::unicode::General_Category::Unassigned }, // <reserved-113CB>
    Prop<::unicode::General_Category> { { 0x113CC, 0x113CD }, ::unicode::General_Category::Spacing_Mark }, // [2] TULU-TIGALARI SIGN ANUSVARA..TULU-TIGALARI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x113CE, 0x113CE }, ::unicode::General_Category::Nonspacing_Mark }, // TULU-TIGALARI SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x113CF, 0x113CF }, ::unicode::General_Category::Spacing_Mark }, // TULU-TIGALARI SIGN LOOPED VIRAMA
    Prop<::unicode::General_Category> { { 0x113D0, 0x113D0 }, ::unicode::General_Category::Nonspacing_Mark }, // TULU-TIGALARI CONJOINER
    Prop<::unicode::General_Category> { { 0x113D1, 0x113D1 }, ::unicode::General_Category::Other_Letter }, // TULU-TIGALARI REPHA
    Prop<::unicode::General_Category> { { 0x113D2, 0x113D2 }, ::unicode::General_Category::Nonspacing_Mark }, // TULU-TIGALARI GEMINATION MARK
    Prop<::unicode::General_Category> { { 0x113D3, 0x113D3 }, ::unicode::General_Category::Other_Letter }, // TULU-TIGALARI SIGN PLUTA
    Prop<::unicode::General_Category> { { 0x113D4, 0x113D5 }, ::unicode::General_Category::Other_Punctuation }, // [2] TULU-TIGALARI DANDA..TULU-TIGALARI DOUBLE DANDA
    Prop<::unicode::General_Category> { { 0x113D6, 0x113D6 }, ::unicode::General_Category::Unassigned }, // <reserved-113D6>
    Prop<::unicode::General_Category> { { 0x113D7, 0x113D8 }, ::unicode::General_Category::Other_Punctuation }, // [2] TULU-TIGALARI SIGN OM PUSHPIKA..TULU-TIGALARI SIGN SHRII PUSHPIKA
    Prop<::unicode::General_Category> { { 0x113D9, 0x113E0 }, ::unicode::General_Category::Unassigned }, // [8] <reserved-113D9>..<reserved-113E0>
    Prop<::unicode::General_Category> { { 0x113E1, 0x113E2 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] TULU-TIGALARI VEDIC TONE SVARITA..TULU-TIGALARI VEDIC TONE ANUDATTA
    Prop<::unicode::General_Category> { { 0x113E3, 0x113FF }, ::unicode::General_Category::Unassigned }, // [29] <reserved-113E3>..<reserved-113FF>
    Prop<::unicode::General_Category> { { 0x11400, 0x11434 }, ::unicode::General_Category::Other_Letter }, // [53] NEWA LETTER A..NEWA LETTER HA
    Prop<::unicode::General_Category> { { 0x11435, 0x11437 }, ::unicode::General_Category::Spacing_Mark }, // [3] NEWA VOWEL SIGN AA..NEWA VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x11438, 0x1143F }, ::unicode::General_Category::Nonspacing_Mark }, // [8] NEWA VOWEL SIGN U..NEWA VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x11440, 0x11441 }, ::unicode::General_Category::Spacing_Mark }, // [2] NEWA VOWEL SIGN O..NEWA VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x11442, 0x11444 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] NEWA SIGN VIRAMA..NEWA SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x11445, 0x11445 }, ::unicode::General_Category::Spacing_Mark }, // NEWA SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x11446, 0x11446 }, ::unicode::General_Category::Nonspacing_Mark }, // NEWA SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x11447, 0x1144A }, ::unicode::General_Category::Other_Letter }, // [4] NEWA SIGN AVAGRAHA..NEWA SIDDHI
    Prop<::unicode::General_Category> { { 0x1144B, 0x1144F }, ::unicode::General_Category::Other_Punctuation }, // [5] NEWA DANDA..NEWA ABBREVIATION SIGN
    Prop<::unicode::General_Category> { { 0x11450, 0x11459 }, ::unicode::General_Category::Decimal_Number }, // [10] NEWA DIGIT ZERO..NEWA DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1145A, 0x1145B }, ::unicode::General_Category::Other_Punctuation }, // [2] NEWA DOUBLE COMMA..NEWA PLACEHOLDER MARK
    Prop<::unicode::General_Category> { { 0x1145C, 0x1145C }, ::unicode::General_Category::Unassigned }, // <reserved-1145C>
    Prop<::unicode::General_Category> { { 0x1145D, 0x1145D }, ::unicode::General_Category::Other_Punctuation }, // NEWA INSERTION SIGN
    Prop<::unicode::General_Category> { { 0x1145E, 0x1145E }, ::unicode::General_Category::Nonspacing_Mark }, // NEWA SANDHI MARK
    Prop<::unicode::General_Category> { { 0x1145F, 0x11461 }, ::unicode::General_Category::Other_Letter }, // [3] NEWA LETTER VEDIC ANUSVARA..NEWA SIGN UPADHMANIYA
    Prop<::unicode::General_Category> { { 0x11462, 0x1147F }, ::unicode::General_Category::Unassigned }, // [30] <reserved-11462>..<reserved-1147F>
    Prop<::unicode::General_Category> { { 0x11480, 0x114AF }, ::unicode::General_Category::Other_Letter }, // [48] TIRHUTA ANJI..TIRHUTA LETTER HA
    Prop<::unicode::General_Category> { { 0x114B0, 0x114B2 }, ::unicode::General_Category::Spacing_Mark }, // [3] TIRHUTA VOWEL SIGN AA..TIRHUTA VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x114B3, 0x114B8 }, ::unicode::General_Category::Nonspacing_Mark }, // [6] TIRHUTA VOWEL SIGN U..TIRHUTA VOWEL SIGN VOCALIC LL
    Prop<::unicode::General_Category> { { 0x114B9, 0x114B9 }, ::unicode::General_Category::Spacing_Mark }, // TIRHUTA VOWEL SIGN E
    Prop<::unicode::General_Category> { { 0x114BA, 0x114BA }, ::unicode::General_Category::Nonspacing_Mark }, // TIRHUTA VOWEL SIGN SHORT E
    Prop<::unicode::General_Category> { { 0x114BB, 0x114BE }, ::unicode::General_Category::Spacing_Mark }, // [4] TIRHUTA VOWEL SIGN AI..TIRHUTA VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x114BF, 0x114C0 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] TIRHUTA SIGN CANDRABINDU..TIRHUTA SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x114C1, 0x114C1 }, ::unicode::General_Category::Spacing_Mark }, // TIRHUTA SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x114C2, 0x114C3 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] TIRHUTA SIGN VIRAMA..TIRHUTA SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x114C4, 0x114C5 }, ::unicode::General_Category::Other_Letter }, // [2] TIRHUTA SIGN AVAGRAHA..TIRHUTA GVANG
    Prop<::unicode::General_Category> { { 0x114C6, 0x114C6 }, ::unicode::General_Category::Other_Punctuation }, // TIRHUTA ABBREVIATION SIGN
    Prop<::unicode::General_Category> { { 0x114C7, 0x114C7 }, ::unicode::General_Category::Other_Letter }, // TIRHUTA OM
    Prop<::unicode::General_Category> { { 0x114C8, 0x114CF }, ::unicode::General_Category::Unassigned }, // [8] <reserved-114C8>..<reserved-114CF>
    Prop<::unicode::General_Category> { { 0x114D0, 0x114D9 }, ::unicode::General_Category::Decimal_Number }, // [10] TIRHUTA DIGIT ZERO..TIRHUTA DIGIT NINE
    Prop<::unicode::General_Category> { { 0x114DA, 0x1157F }, ::unicode::General_Category::Unassigned }, // [166] <reserved-114DA>..<reserved-1157F>
    Prop<::unicode::General_Category> { { 0x11580, 0x115AE }, ::unicode::General_Category::Other_Letter }, // [47] SIDDHAM LETTER A..SIDDHAM LETTER HA
    Prop<::unicode::General_Category> { { 0x115AF, 0x115B1 }, ::unicode::General_Category::Spacing_Mark }, // [3] SIDDHAM VOWEL SIGN AA..SIDDHAM VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x115B2, 0x115B5 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] SIDDHAM VOWEL SIGN U..SIDDHAM VOWEL SIGN VOCALIC RR
    Prop<::unicode::General_Category> { { 0x115B6, 0x115B7 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-115B6>..<reserved-115B7>
    Prop<::unicode::General_Category> { { 0x115B8, 0x115BB }, ::unicode::General_Category::Spacing_Mark }, // [4] SIDDHAM VOWEL SIGN E..SIDDHAM VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x115BC, 0x115BD }, ::unicode::General_Category::Nonspacing_Mark }, // [2] SIDDHAM SIGN CANDRABINDU..SIDDHAM SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x115BE, 0x115BE }, ::unicode::General_Category::Spacing_Mark }, // SIDDHAM SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x115BF, 0x115C0 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] SIDDHAM SIGN VIRAMA..SIDDHAM SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x115C1, 0x115D7 }, ::unicode::General_Category::Other_Punctuation }, // [23] SIDDHAM SIGN SIDDHAM..SIDDHAM SECTION MARK WITH CIRCLES AND FOUR ENCLOSURES
    Prop<::unicode::General_Category> { { 0x115D8, 0x115DB }, ::unicode::General_Category::Other_Letter }, // [4] SIDDHAM LETTER THREE-CIRCLE ALTERNATE I..SIDDHAM LETTER ALTERNATE U
    Prop<::unicode::General_Category> { { 0x115DC, 0x115DD }, ::unicode::General_Category::Nonspacing_Mark }, // [2] SIDDHAM VOWEL SIGN ALTERNATE U..SIDDHAM VOWEL SIGN ALTERNATE UU
    Prop<::unicode::General_Category> { { 0x115DE, 0x115FF }, ::unicode::General_Category::Unassigned }, // [34] <reserved-115DE>..<reserved-115FF>
    Prop<::unicode::General_Category> { { 0x11600, 0x1162F }, ::unicode::General_Category::Other_Letter }, // [48] MODI LETTER A..MODI LETTER LLA
    Prop<::unicode::General_Category> { { 0x11630, 0x11632 }, ::unicode::General_Category::Spacing_Mark }, // [3] MODI VOWEL SIGN AA..MODI VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x11633, 0x1163A }, ::unicode::General_Category::Nonspacing_Mark }, // [8] MODI VOWEL SIGN U..MODI VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x1163B, 0x1163C }, ::unicode::General_Category::Spacing_Mark }, // [2] MODI VOWEL SIGN O..MODI VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x1163D, 0x1163D }, ::unicode::General_Category::Nonspacing_Mark }, // MODI SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x1163E, 0x1163E }, ::unicode::General_Category::Spacing_Mark }, // MODI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x1163F, 0x11640 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MODI SIGN VIRAMA..MODI SIGN ARDHACANDRA
    Prop<::unicode::General_Category> { { 0x11641, 0x11643 }, ::unicode::General_Category::Other_Punctuation }, // [3] MODI DANDA..MODI ABBREVIATION SIGN
    Prop<::unicode::General_Category> { { 0x11644, 0x11644 }, ::unicode::General_Category::Other_Letter }, // MODI SIGN HUVA
    Prop<::unicode::General_Category> { { 0x11645, 0x1164F }, ::unicode::General_Category::Unassigned }, // [11] <reserved-11645>..<reserved-1164F>
    Prop<::unicode::General_Category> { { 0x11650, 0x11659 }, ::unicode::General_Category::Decimal_Number }, // [10] MODI DIGIT ZERO..MODI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1165A, 0x1165F }, ::unicode::General_Category::Unassigned }, // [6] <reserved-1165A>..<reserved-1165F>
    Prop<::unicode::General_Category> { { 0x11660, 0x1166C }, ::unicode::General_Category::Other_Punctuation }, // [13] MONGOLIAN BIRGA WITH ORNAMENT..MONGOLIAN TURNED SWIRL BIRGA WITH DOUBLE ORNAMENT
    Prop<::unicode::General_Category> { { 0x1166D, 0x1167F }, ::unicode::General_Category::Unassigned }, // [19] <reserved-1166D>..<reserved-1167F>
    Prop<::unicode::General_Category> { { 0x11680, 0x116AA }, ::unicode::General_Category::Other_Letter }, // [43] TAKRI LETTER A..TAKRI LETTER RRA
    Prop<::unicode::General_Category> { { 0x116AB, 0x116AB }, ::unicode::General_Category::Nonspacing_Mark }, // TAKRI SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x116AC, 0x116AC }, ::unicode::General_Category::Spacing_Mark }, // TAKRI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x116AD, 0x116AD }, ::unicode::General_Category::Nonspacing_Mark }, // TAKRI VOWEL SIGN AA
    Prop<::unicode::General_Category> { { 0x116AE, 0x116AF }, ::unicode::General_Category::Spacing_Mark }, // [2] TAKRI VOWEL SIGN I..TAKRI VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x116B0, 0x116B5 }, ::unicode::General_Category::Nonspacing_Mark }, // [6] TAKRI VOWEL SIGN U..TAKRI VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x116B6, 0x116B6 }, ::unicode::General_Category::Spacing_Mark }, // TAKRI SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x116B7, 0x116B7 }, ::unicode::General_Category::Nonspacing_Mark }, // TAKRI SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x116B8, 0x116B8 }, ::unicode::General_Category::Other_Letter }, // TAKRI LETTER ARCHAIC KHA
    Prop<::unicode::General_Category> { { 0x116B9, 0x116B9 }, ::unicode::General_Category::Other_Punctuation }, // TAKRI ABBREVIATION SIGN
    Prop<::unicode::General_Category> { { 0x116BA, 0x116BF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-116BA>..<reserved-116BF>
    Prop<::unicode::General_Category> { { 0x116C0, 0x116C9 }, ::unicode::General_Category::Decimal_Number }, // [10] TAKRI DIGIT ZERO..TAKRI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x116CA, 0x116CF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-116CA>..<reserved-116CF>
    Prop<::unicode::General_Category> { { 0x116D0, 0x116E3 }, ::unicode::General_Category::Decimal_Number }, // [20] MYANMAR PAO DIGIT ZERO..MYANMAR EASTERN PWO KAREN DIGIT NINE
    Prop<::unicode::General_Category> { { 0x116E4, 0x116FF }, ::unicode::General_Category::Unassigned }, // [28] <reserved-116E4>..<reserved-116FF>
    Prop<::unicode::General_Category> { { 0x11700, 0x1171A }, ::unicode::General_Category::Other_Letter }, // [27] AHOM LETTER KA..AHOM LETTER ALTERNATE BA
    Prop<::unicode::General_Category> { { 0x1171B, 0x1171C }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1171B>..<reserved-1171C>
    Prop<::unicode::General_Category> { { 0x1171D, 0x1171D }, ::unicode::General_Category::Nonspacing_Mark }, // AHOM CONSONANT SIGN MEDIAL LA
    Prop<::unicode::General_Category> { { 0x1171E, 0x1171E }, ::unicode::General_Category::Spacing_Mark }, // AHOM CONSONANT SIGN MEDIAL RA
    Prop<::unicode::General_Category> { { 0x1171F, 0x1171F }, ::unicode::General_Category::Nonspacing_Mark }, // AHOM CONSONANT SIGN MEDIAL LIGATING RA
    Prop<::unicode::General_Category> { { 0x11720, 0x11721 }, ::unicode::General_Category::Spacing_Mark }, // [2] AHOM VOWEL SIGN A..AHOM VOWEL SIGN AA
    Prop<::unicode::General_Category> { { 0x11722, 0x11725 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] AHOM VOWEL SIGN I..AHOM VOWEL SIGN UU
    Prop<::unicode::General_Category> { { 0x11726, 0x11726 }, ::unicode::General_Category::Spacing_Mark }, // AHOM VOWEL SIGN E
    Prop<::unicode::General_Category> { { 0x11727, 0x1172B }, ::unicode::General_Category::Nonspacing_Mark }, // [5] AHOM VOWEL SIGN AW..AHOM SIGN KILLER
    Prop<::unicode::General_Category> { { 0x1172C, 0x1172F }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1172C>..<reserved-1172F>
    Prop<::unicode::General_Category> { { 0x11730, 0x11739 }, ::unicode::General_Category::Decimal_Number }, // [10] AHOM DIGIT ZERO..AHOM DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1173A, 0x1173B }, ::unicode::General_Category::Other_Number }, // [2] AHOM NUMBER TEN..AHOM NUMBER TWENTY
    Prop<::unicode::General_Category> { { 0x1173C, 0x1173E }, ::unicode::General_Category::Other_Punctuation }, // [3] AHOM SIGN SMALL SECTION..AHOM SIGN RULAI
    Prop<::unicode::General_Category> { { 0x1173F, 0x1173F }, ::unicode::General_Category::Other_Symbol }, // AHOM SYMBOL VI
    Prop<::unicode::General_Category> { { 0x11740, 0x11746 }, ::unicode::General_Category::Other_Letter }, // [7] AHOM LETTER CA..AHOM LETTER LLA
    Prop<::unicode::General_Category> { { 0x11747, 0x117FF }, ::unicode::General_Category::Unassigned }, // [185] <reserved-11747>..<reserved-117FF>
    Prop<::unicode::General_Category> { { 0x11800, 0x1182B }, ::unicode::General_Category::Other_Letter }, // [44] DOGRA LETTER A..DOGRA LETTER RRA
    Prop<::unicode::General_Category> { { 0x1182C, 0x1182E }, ::unicode::General_Category::Spacing_Mark }, // [3] DOGRA VOWEL SIGN AA..DOGRA VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x1182F, 0x11837 }, ::unicode::General_Category::Nonspacing_Mark }, // [9] DOGRA VOWEL SIGN U..DOGRA SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x11838, 0x11838 }, ::unicode::General_Category::Spacing_Mark }, // DOGRA SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x11839, 0x1183A }, ::unicode::General_Category::Nonspacing_Mark }, // [2] DOGRA SIGN VIRAMA..DOGRA SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x1183B, 0x1183B }, ::unicode::General_Category::Other_Punctuation }, // DOGRA ABBREVIATION SIGN
    Prop<::unicode::General_Category> { { 0x1183C, 0x1189F }, ::unicode::General_Category::Unassigned }, // [100] <reserved-1183C>..<reserved-1189F>
    Prop<::unicode::General_Category> { { 0x118A0, 0x118BF }, ::unicode::General_Category::Uppercase_Letter }, // [32] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI CAPITAL LETTER VIYO
    Prop<::unicode::General_Category> { { 0x118C0, 0x118DF }, ::unicode::General_Category::Lowercase_Letter }, // [32] WARANG CITI SMALL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Prop<::unicode::General_Category> { { 0x118E0, 0x118E9 }, ::unicode::General_Category::Decimal_Number }, // [10] WARANG CITI DIGIT ZERO..WARANG CITI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x118EA, 0x118F2 }, ::unicode::General_Category::Other_Number }, // [9] WARANG CITI NUMBER TEN..WARANG CITI NUMBER NINETY
    Prop<::unicode::General_Category> { { 0x118F3, 0x118FE }, ::unicode::General_Category::Unassigned }, // [12] <reserved-118F3>..<reserved-118FE>
    Prop<::unicode::General_Category> { { 0x118FF, 0x11906 }, ::unicode::General_Category::Other_Letter }, // [8] WARANG CITI OM..DIVES AKURU LETTER E
    Prop<::unicode::General_Category> { { 0x11907, 0x11908 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-11907>..<reserved-11908>
    Prop<::unicode::General_Category> { { 0x11909, 0x11909 }, ::unicode::General_Category::Other_Letter }, // DIVES AKURU LETTER O
    Prop<::unicode::General_Category> { { 0x1190A, 0x1190B }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1190A>..<reserved-1190B>
    Prop<::unicode::General_Category> { { 0x1190C, 0x11913 }, ::unicode::General_Category::Other_Letter }, // [8] DIVES AKURU LETTER KA..DIVES AKURU LETTER JA
    Prop<::unicode::General_Category> { { 0x11914, 0x11914 }, ::unicode::General_Category::Unassigned }, // <reserved-11914>
    Prop<::unicode::General_Category> { { 0x11915, 0x11916 }, ::unicode::General_Category::Other_Letter }, // [2] DIVES AKURU LETTER NYA..DIVES AKURU LETTER TTA
    Prop<::unicode::General_Category> { { 0x11917, 0x11917 }, ::unicode::General_Category::Unassigned }, // <reserved-11917>
    Prop<::unicode::General_Category> { { 0x11918, 0x1192F }, ::unicode::General_Category::Other_Letter }, // [24] DIVES AKURU LETTER DDA..DIVES AKURU LETTER ZA
    Prop<::unicode::General_Category> { { 0x11930, 0x11935 }, ::unicode::General_Category::Spacing_Mark }, // [6] DIVES AKURU VOWEL SIGN AA..DIVES AKURU VOWEL SIGN E
    Prop<::unicode::General_Category> { { 0x11936, 0x11936 }, ::unicode::General_Category::Unassigned }, // <reserved-11936>
    Prop<::unicode::General_Category> { { 0x11937, 0x11938 }, ::unicode::General_Category::Spacing_Mark }, // [2] DIVES AKURU VOWEL SIGN AI..DIVES AKURU VOWEL SIGN O
    Prop<::unicode::General_Category> { { 0x11939, 0x1193A }, ::unicode::General_Category::Unassigned }, // [2] <reserved-11939>..<reserved-1193A>
    Prop<::unicode::General_Category> { { 0x1193B, 0x1193C }, ::unicode::General_Category::Nonspacing_Mark }, // [2] DIVES AKURU SIGN ANUSVARA..DIVES AKURU SIGN CANDRABINDU
    Prop<::unicode::General_Category> { { 0x1193D, 0x1193D }, ::unicode::General_Category::Spacing_Mark }, // DIVES AKURU SIGN HALANTA
    Prop<::unicode::General_Category> { { 0x1193E, 0x1193E }, ::unicode::General_Category::Nonspacing_Mark }, // DIVES AKURU VIRAMA
    Prop<::unicode::General_Category> { { 0x1193F, 0x1193F }, ::unicode::General_Category::Other_Letter }, // DIVES AKURU PREFIXED NASAL SIGN
    Prop<::unicode::General_Category> { { 0x11940, 0x11940 }, ::unicode::General_Category::Spacing_Mark }, // DIVES AKURU MEDIAL YA
    Prop<::unicode::General_Category> { { 0x11941, 0x11941 }, ::unicode::General_Category::Other_Letter }, // DIVES AKURU INITIAL RA
    Prop<::unicode::General_Category> { { 0x11942, 0x11942 }, ::unicode::General_Category::Spacing_Mark }, // DIVES AKURU MEDIAL RA
    Prop<::unicode::General_Category> { { 0x11943, 0x11943 }, ::unicode::General_Category::Nonspacing_Mark }, // DIVES AKURU SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x11944, 0x11946 }, ::unicode::General_Category::Other_Punctuation }, // [3] DIVES AKURU DOUBLE DANDA..DIVES AKURU END OF TEXT MARK
    Prop<::unicode::General_Category> { { 0x11947, 0x1194F }, ::unicode::General_Category::Unassigned }, // [9] <reserved-11947>..<reserved-1194F>
    Prop<::unicode::General_Category> { { 0x11950, 0x11959 }, ::unicode::General_Category::Decimal_Number }, // [10] DIVES AKURU DIGIT ZERO..DIVES AKURU DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1195A, 0x1199F }, ::unicode::General_Category::Unassigned }, // [70] <reserved-1195A>..<reserved-1199F>
    Prop<::unicode::General_Category> { { 0x119A0, 0x119A7 }, ::unicode::General_Category::Other_Letter }, // [8] NANDINAGARI LETTER A..NANDINAGARI LETTER VOCALIC RR
    Prop<::unicode::General_Category> { { 0x119A8, 0x119A9 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-119A8>..<reserved-119A9>
    Prop<::unicode::General_Category> { { 0x119AA, 0x119D0 }, ::unicode::General_Category::Other_Letter }, // [39] NANDINAGARI LETTER E..NANDINAGARI LETTER RRA
    Prop<::unicode::General_Category> { { 0x119D1, 0x119D3 }, ::unicode::General_Category::Spacing_Mark }, // [3] NANDINAGARI VOWEL SIGN AA..NANDINAGARI VOWEL SIGN II
    Prop<::unicode::General_Category> { { 0x119D4, 0x119D7 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] NANDINAGARI VOWEL SIGN U..NANDINAGARI VOWEL SIGN VOCALIC RR
    Prop<::unicode::General_Category> { { 0x119D8, 0x119D9 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-119D8>..<reserved-119D9>
    Prop<::unicode::General_Category> { { 0x119DA, 0x119DB }, ::unicode::General_Category::Nonspacing_Mark }, // [2] NANDINAGARI VOWEL SIGN E..NANDINAGARI VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x119DC, 0x119DF }, ::unicode::General_Category::Spacing_Mark }, // [4] NANDINAGARI VOWEL SIGN O..NANDINAGARI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x119E0, 0x119E0 }, ::unicode::General_Category::Nonspacing_Mark }, // NANDINAGARI SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x119E1, 0x119E1 }, ::unicode::General_Category::Other_Letter }, // NANDINAGARI SIGN AVAGRAHA
    Prop<::unicode::General_Category> { { 0x119E2, 0x119E2 }, ::unicode::General_Category::Other_Punctuation }, // NANDINAGARI SIGN SIDDHAM
    Prop<::unicode::General_Category> { { 0x119E3, 0x119E3 }, ::unicode::General_Category::Other_Letter }, // NANDINAGARI HEADSTROKE
    Prop<::unicode::General_Category> { { 0x119E4, 0x119E4 }, ::unicode::General_Category::Spacing_Mark }, // NANDINAGARI VOWEL SIGN PRISHTHAMATRA E
    Prop<::unicode::General_Category> { { 0x119E5, 0x119FF }, ::unicode::General_Category::Unassigned }, // [27] <reserved-119E5>..<reserved-119FF>
    Prop<::unicode::General_Category> { { 0x11A00, 0x11A00 }, ::unicode::General_Category::Other_Letter }, // ZANABAZAR SQUARE LETTER A
    Prop<::unicode::General_Category> { { 0x11A01, 0x11A0A }, ::unicode::General_Category::Nonspacing_Mark }, // [10] ZANABAZAR SQUARE VOWEL SIGN I..ZANABAZAR SQUARE VOWEL LENGTH MARK
    Prop<::unicode::General_Category> { { 0x11A0B, 0x11A32 }, ::unicode::General_Category::Other_Letter }, // [40] ZANABAZAR SQUARE LETTER KA..ZANABAZAR SQUARE LETTER KSSA
    Prop<::unicode::General_Category> { { 0x11A33, 0x11A38 }, ::unicode::General_Category::Nonspacing_Mark }, // [6] ZANABAZAR SQUARE FINAL CONSONANT MARK..ZANABAZAR SQUARE SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x11A39, 0x11A39 }, ::unicode::General_Category::Spacing_Mark }, // ZANABAZAR SQUARE SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x11A3A, 0x11A3A }, ::unicode::General_Category::Other_Letter }, // ZANABAZAR SQUARE CLUSTER-INITIAL LETTER RA
    Prop<::unicode::General_Category> { { 0x11A3B, 0x11A3E }, ::unicode::General_Category::Nonspacing_Mark }, // [4] ZANABAZAR SQUARE CLUSTER-FINAL LETTER YA..ZANABAZAR SQUARE CLUSTER-FINAL LETTER VA
    Prop<::unicode::General_Category> { { 0x11A3F, 0x11A46 }, ::unicode::General_Category::Other_Punctuation }, // [8] ZANABAZAR SQUARE INITIAL HEAD MARK..ZANABAZAR SQUARE CLOSING DOUBLE-LINED HEAD MARK
    Prop<::unicode::General_Category> { { 0x11A47, 0x11A47 }, ::unicode::General_Category::Nonspacing_Mark }, // ZANABAZAR SQUARE SUBJOINER
    Prop<::unicode::General_Category> { { 0x11A48, 0x11A4F }, ::unicode::General_Category::Unassigned }, // [8] <reserved-11A48>..<reserved-11A4F>
    Prop<::unicode::General_Category> { { 0x11A50, 0x11A50 }, ::unicode::General_Category::Other_Letter }, // SOYOMBO LETTER A
    Prop<::unicode::General_Category> { { 0x11A51, 0x11A56 }, ::unicode::General_Category::Nonspacing_Mark }, // [6] SOYOMBO VOWEL SIGN I..SOYOMBO VOWEL SIGN OE
    Prop<::unicode::General_Category> { { 0x11A57, 0x11A58 }, ::unicode::General_Category::Spacing_Mark }, // [2] SOYOMBO VOWEL SIGN AI..SOYOMBO VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x11A59, 0x11A5B }, ::unicode::General_Category::Nonspacing_Mark }, // [3] SOYOMBO VOWEL SIGN VOCALIC R..SOYOMBO VOWEL LENGTH MARK
    Prop<::unicode::General_Category> { { 0x11A5C, 0x11A89 }, ::unicode::General_Category::Other_Letter }, // [46] SOYOMBO LETTER KA..SOYOMBO CLUSTER-INITIAL LETTER SA
    Prop<::unicode::General_Category> { { 0x11A8A, 0x11A96 }, ::unicode::General_Category::Nonspacing_Mark }, // [13] SOYOMBO FINAL CONSONANT SIGN G..SOYOMBO SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x11A97, 0x11A97 }, ::unicode::General_Category::Spacing_Mark }, // SOYOMBO SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x11A98, 0x11A99 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] SOYOMBO GEMINATION MARK..SOYOMBO SUBJOINER
    Prop<::unicode::General_Category> { { 0x11A9A, 0x11A9C }, ::unicode::General_Category::Other_Punctuation }, // [3] SOYOMBO MARK TSHEG..SOYOMBO MARK DOUBLE SHAD
    Prop<::unicode::General_Category> { { 0x11A9D, 0x11A9D }, ::unicode::General_Category::Other_Letter }, // SOYOMBO MARK PLUTA
    Prop<::unicode::General_Category> { { 0x11A9E, 0x11AA2 }, ::unicode::General_Category::Other_Punctuation }, // [5] SOYOMBO HEAD MARK WITH MOON AND SUN AND TRIPLE FLAME..SOYOMBO TERMINAL MARK-2
    Prop<::unicode::General_Category> { { 0x11AA3, 0x11AAF }, ::unicode::General_Category::Unassigned }, // [13] <reserved-11AA3>..<reserved-11AAF>
    Prop<::unicode::General_Category> { { 0x11AB0, 0x11AF8 }, ::unicode::General_Category::Other_Letter }, // [73] CANADIAN SYLLABICS NATTILIK HI..PAU CIN HAU GLOTTAL STOP FINAL
    Prop<::unicode::General_Category> { { 0x11AF9, 0x11AFF }, ::unicode::General_Category::Unassigned }, // [7] <reserved-11AF9>..<reserved-11AFF>
    Prop<::unicode::General_Category> { { 0x11B00, 0x11B09 }, ::unicode::General_Category::Other_Punctuation }, // [10] DEVANAGARI HEAD MARK..DEVANAGARI SIGN MINDU
    Prop<::unicode::General_Category> { { 0x11B0A, 0x11B5F }, ::unicode::General_Category::Unassigned }, // [86] <reserved-11B0A>..<reserved-11B5F>
    Prop<::unicode::General_Category> { { 0x11B60, 0x11B60 }, ::unicode::General_Category::Nonspacing_Mark }, // SHARADA VOWEL SIGN OE
    Prop<::unicode::General_Category> { { 0x11B61, 0x11B61 }, ::unicode::General_Category::Spacing_Mark }, // SHARADA VOWEL SIGN OOE
    Prop<::unicode::General_Category> { { 0x11B62, 0x11B64 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] SHARADA VOWEL SIGN UE..SHARADA VOWEL SIGN SHORT E
    Prop<::unicode::General_Category> { { 0x11B65, 0x11B65 }, ::unicode::General_Category::Spacing_Mark }, // SHARADA VOWEL SIGN SHORT O
    Prop<::unicode::General_Category> { { 0x11B66, 0x11B66 }, ::unicode::General_Category::Nonspacing_Mark }, // SHARADA VOWEL SIGN CANDRA E
    Prop<::unicode::General_Category> { { 0x11B67, 0x11B67 }, ::unicode::General_Category::Spacing_Mark }, // SHARADA VOWEL SIGN CANDRA O
    Prop<::unicode::General_Category> { { 0x11B68, 0x11BBF }, ::unicode::General_Category::Unassigned }, // [88] <reserved-11B68>..<reserved-11BBF>
    Prop<::unicode::General_Category> { { 0x11BC0, 0x11BE0 }, ::unicode::General_Category::Other_Letter }, // [33] SUNUWAR LETTER DEVI..SUNUWAR LETTER KLOKO
    Prop<::unicode::General_Category> { { 0x11BE1, 0x11BE1 }, ::unicode::General_Category::Other_Punctuation }, // SUNUWAR SIGN PVO
    Prop<::unicode::General_Category> { { 0x11BE2, 0x11BEF }, ::unicode::General_Category::Unassigned }, // [14] <reserved-11BE2>..<reserved-11BEF>
    Prop<::unicode::General_Category> { { 0x11BF0, 0x11BF9 }, ::unicode::General_Category::Decimal_Number }, // [10] SUNUWAR DIGIT ZERO..SUNUWAR DIGIT NINE
    Prop<::unicode::General_Category> { { 0x11BFA, 0x11BFF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-11BFA>..<reserved-11BFF>
    Prop<::unicode::General_Category> { { 0x11C00, 0x11C08 }, ::unicode::General_Category::Other_Letter }, // [9] BHAIKSUKI LETTER A..BHAIKSUKI LETTER VOCALIC L
    Prop<::unicode::General_Category> { { 0x11C09, 0x11C09 }, ::unicode::General_Category::Unassigned }, // <reserved-11C09>
    Prop<::unicode::General_Category> { { 0x11C0A, 0x11C2E }, ::unicode::General_Category::Other_Letter }, // [37] BHAIKSUKI LETTER E..BHAIKSUKI LETTER HA
    Prop<::unicode::General_Category> { { 0x11C2F, 0x11C2F }, ::unicode::General_Category::Spacing_Mark }, // BHAIKSUKI VOWEL SIGN AA
    Prop<::unicode::General_Category> { { 0x11C30, 0x11C36 }, ::unicode::General_Category::Nonspacing_Mark }, // [7] BHAIKSUKI VOWEL SIGN I..BHAIKSUKI VOWEL SIGN VOCALIC L
    Prop<::unicode::General_Category> { { 0x11C37, 0x11C37 }, ::unicode::General_Category::Unassigned }, // <reserved-11C37>
    Prop<::unicode::General_Category> { { 0x11C38, 0x11C3D }, ::unicode::General_Category::Nonspacing_Mark }, // [6] BHAIKSUKI VOWEL SIGN E..BHAIKSUKI SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x11C3E, 0x11C3E }, ::unicode::General_Category::Spacing_Mark }, // BHAIKSUKI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x11C3F, 0x11C3F }, ::unicode::General_Category::Nonspacing_Mark }, // BHAIKSUKI SIGN VIRAMA
    Prop<::unicode::General_Category> { { 0x11C40, 0x11C40 }, ::unicode::General_Category::Other_Letter }, // BHAIKSUKI SIGN AVAGRAHA
    Prop<::unicode::General_Category> { { 0x11C41, 0x11C45 }, ::unicode::General_Category::Other_Punctuation }, // [5] BHAIKSUKI DANDA..BHAIKSUKI GAP FILLER-2
    Prop<::unicode::General_Category> { { 0x11C46, 0x11C4F }, ::unicode::General_Category::Unassigned }, // [10] <reserved-11C46>..<reserved-11C4F>
    Prop<::unicode::General_Category> { { 0x11C50, 0x11C59 }, ::unicode::General_Category::Decimal_Number }, // [10] BHAIKSUKI DIGIT ZERO..BHAIKSUKI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x11C5A, 0x11C6C }, ::unicode::General_Category::Other_Number }, // [19] BHAIKSUKI NUMBER ONE..BHAIKSUKI HUNDREDS UNIT MARK
    Prop<::unicode::General_Category> { { 0x11C6D, 0x11C6F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-11C6D>..<reserved-11C6F>
    Prop<::unicode::General_Category> { { 0x11C70, 0x11C71 }, ::unicode::General_Category::Other_Punctuation }, // [2] MARCHEN HEAD MARK..MARCHEN MARK SHAD
    Prop<::unicode::General_Category> { { 0x11C72, 0x11C8F }, ::unicode::General_Category::Other_Letter }, // [30] MARCHEN LETTER KA..MARCHEN LETTER A
    Prop<::unicode::General_Category> { { 0x11C90, 0x11C91 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-11C90>..<reserved-11C91>
    Prop<::unicode::General_Category> { { 0x11C92, 0x11CA7 }, ::unicode::General_Category::Nonspacing_Mark }, // [22] MARCHEN SUBJOINED LETTER KA..MARCHEN SUBJOINED LETTER ZA
    Prop<::unicode::General_Category> { { 0x11CA8, 0x11CA8 }, ::unicode::General_Category::Unassigned }, // <reserved-11CA8>
    Prop<::unicode::General_Category> { { 0x11CA9, 0x11CA9 }, ::unicode::General_Category::Spacing_Mark }, // MARCHEN SUBJOINED LETTER YA
    Prop<::unicode::General_Category> { { 0x11CAA, 0x11CB0 }, ::unicode::General_Category::Nonspacing_Mark }, // [7] MARCHEN SUBJOINED LETTER RA..MARCHEN VOWEL SIGN AA
    Prop<::unicode::General_Category> { { 0x11CB1, 0x11CB1 }, ::unicode::General_Category::Spacing_Mark }, // MARCHEN VOWEL SIGN I
    Prop<::unicode::General_Category> { { 0x11CB2, 0x11CB3 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MARCHEN VOWEL SIGN U..MARCHEN VOWEL SIGN E
    Prop<::unicode::General_Category> { { 0x11CB4, 0x11CB4 }, ::unicode::General_Category::Spacing_Mark }, // MARCHEN VOWEL SIGN O
    Prop<::unicode::General_Category> { { 0x11CB5, 0x11CB6 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MARCHEN SIGN ANUSVARA..MARCHEN SIGN CANDRABINDU
    Prop<::unicode::General_Category> { { 0x11CB7, 0x11CFF }, ::unicode::General_Category::Unassigned }, // [73] <reserved-11CB7>..<reserved-11CFF>
    Prop<::unicode::General_Category> { { 0x11D00, 0x11D06 }, ::unicode::General_Category::Other_Letter }, // [7] MASARAM GONDI LETTER A..MASARAM GONDI LETTER E
    Prop<::unicode::General_Category> { { 0x11D07, 0x11D07 }, ::unicode::General_Category::Unassigned }, // <reserved-11D07>
    Prop<::unicode::General_Category> { { 0x11D08, 0x11D09 }, ::unicode::General_Category::Other_Letter }, // [2] MASARAM GONDI LETTER AI..MASARAM GONDI LETTER O
    Prop<::unicode::General_Category> { { 0x11D0A, 0x11D0A }, ::unicode::General_Category::Unassigned }, // <reserved-11D0A>
    Prop<::unicode::General_Category> { { 0x11D0B, 0x11D30 }, ::unicode::General_Category::Other_Letter }, // [38] MASARAM GONDI LETTER AU..MASARAM GONDI LETTER TRA
    Prop<::unicode::General_Category> { { 0x11D31, 0x11D36 }, ::unicode::General_Category::Nonspacing_Mark }, // [6] MASARAM GONDI VOWEL SIGN AA..MASARAM GONDI VOWEL SIGN VOCALIC R
    Prop<::unicode::General_Category> { { 0x11D37, 0x11D39 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-11D37>..<reserved-11D39>
    Prop<::unicode::General_Category> { { 0x11D3A, 0x11D3A }, ::unicode::General_Category::Nonspacing_Mark }, // MASARAM GONDI VOWEL SIGN E
    Prop<::unicode::General_Category> { { 0x11D3B, 0x11D3B }, ::unicode::General_Category::Unassigned }, // <reserved-11D3B>
    Prop<::unicode::General_Category> { { 0x11D3C, 0x11D3D }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MASARAM GONDI VOWEL SIGN AI..MASARAM GONDI VOWEL SIGN O
    Prop<::unicode::General_Category> { { 0x11D3E, 0x11D3E }, ::unicode::General_Category::Unassigned }, // <reserved-11D3E>
    Prop<::unicode::General_Category> { { 0x11D3F, 0x11D45 }, ::unicode::General_Category::Nonspacing_Mark }, // [7] MASARAM GONDI VOWEL SIGN AU..MASARAM GONDI VIRAMA
    Prop<::unicode::General_Category> { { 0x11D46, 0x11D46 }, ::unicode::General_Category::Other_Letter }, // MASARAM GONDI REPHA
    Prop<::unicode::General_Category> { { 0x11D47, 0x11D47 }, ::unicode::General_Category::Nonspacing_Mark }, // MASARAM GONDI RA-KARA
    Prop<::unicode::General_Category> { { 0x11D48, 0x11D4F }, ::unicode::General_Category::Unassigned }, // [8] <reserved-11D48>..<reserved-11D4F>
    Prop<::unicode::General_Category> { { 0x11D50, 0x11D59 }, ::unicode::General_Category::Decimal_Number }, // [10] MASARAM GONDI DIGIT ZERO..MASARAM GONDI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x11D5A, 0x11D5F }, ::unicode::General_Category::Unassigned }, // [6] <reserved-11D5A>..<reserved-11D5F>
    Prop<::unicode::General_Category> { { 0x11D60, 0x11D65 }, ::unicode::General_Category::Other_Letter }, // [6] GUNJALA GONDI LETTER A..GUNJALA GONDI LETTER UU
    Prop<::unicode::General_Category> { { 0x11D66, 0x11D66 }, ::unicode::General_Category::Unassigned }, // <reserved-11D66>
    Prop<::unicode::General_Category> { { 0x11D67, 0x11D68 }, ::unicode::General_Category::Other_Letter }, // [2] GUNJALA GONDI LETTER EE..GUNJALA GONDI LETTER AI
    Prop<::unicode::General_Category> { { 0x11D69, 0x11D69 }, ::unicode::General_Category::Unassigned }, // <reserved-11D69>
    Prop<::unicode::General_Category> { { 0x11D6A, 0x11D89 }, ::unicode::General_Category::Other_Letter }, // [32] GUNJALA GONDI LETTER OO..GUNJALA GONDI LETTER SA
    Prop<::unicode::General_Category> { { 0x11D8A, 0x11D8E }, ::unicode::General_Category::Spacing_Mark }, // [5] GUNJALA GONDI VOWEL SIGN AA..GUNJALA GONDI VOWEL SIGN UU
    Prop<::unicode::General_Category> { { 0x11D8F, 0x11D8F }, ::unicode::General_Category::Unassigned }, // <reserved-11D8F>
    Prop<::unicode::General_Category> { { 0x11D90, 0x11D91 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] GUNJALA GONDI VOWEL SIGN EE..GUNJALA GONDI VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x11D92, 0x11D92 }, ::unicode::General_Category::Unassigned }, // <reserved-11D92>
    Prop<::unicode::General_Category> { { 0x11D93, 0x11D94 }, ::unicode::General_Category::Spacing_Mark }, // [2] GUNJALA GONDI VOWEL SIGN OO..GUNJALA GONDI VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x11D95, 0x11D95 }, ::unicode::General_Category::Nonspacing_Mark }, // GUNJALA GONDI SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x11D96, 0x11D96 }, ::unicode::General_Category::Spacing_Mark }, // GUNJALA GONDI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x11D97, 0x11D97 }, ::unicode::General_Category::Nonspacing_Mark }, // GUNJALA GONDI VIRAMA
    Prop<::unicode::General_Category> { { 0x11D98, 0x11D98 }, ::unicode::General_Category::Other_Letter }, // GUNJALA GONDI OM
    Prop<::unicode::General_Category> { { 0x11D99, 0x11D9F }, ::unicode::General_Category::Unassigned }, // [7] <reserved-11D99>..<reserved-11D9F>
    Prop<::unicode::General_Category> { { 0x11DA0, 0x11DA9 }, ::unicode::General_Category::Decimal_Number }, // [10] GUNJALA GONDI DIGIT ZERO..GUNJALA GONDI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x11DAA, 0x11DAF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-11DAA>..<reserved-11DAF>
    Prop<::unicode::General_Category> { { 0x11DB0, 0x11DD8 }, ::unicode::General_Category::Other_Letter }, // [41] TOLONG SIKI LETTER I..TOLONG SIKI LETTER RRH
    Prop<::unicode::General_Category> { { 0x11DD9, 0x11DD9 }, ::unicode::General_Category::Modifier_Letter }, // TOLONG SIKI SIGN SELA
    Prop<::unicode::General_Category> { { 0x11DDA, 0x11DDB }, ::unicode::General_Category::Other_Letter }, // [2] TOLONG SIKI SIGN HECAKA..TOLONG SIKI UNGGA
    Prop<::unicode::General_Category> { { 0x11DDC, 0x11DDF }, ::unicode::General_Category::Unassigned }, // [4] <reserved-11DDC>..<reserved-11DDF>
    Prop<::unicode::General_Category> { { 0x11DE0, 0x11DE9 }, ::unicode::General_Category::Decimal_Number }, // [10] TOLONG SIKI DIGIT ZERO..TOLONG SIKI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x11DEA, 0x11EDF }, ::unicode::General_Category::Unassigned }, // [246] <reserved-11DEA>..<reserved-11EDF>
    Prop<::unicode::General_Category> { { 0x11EE0, 0x11EF2 }, ::unicode::General_Category::Other_Letter }, // [19] MAKASAR LETTER KA..MAKASAR ANGKA
    Prop<::unicode::General_Category> { { 0x11EF3, 0x11EF4 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] MAKASAR VOWEL SIGN I..MAKASAR VOWEL SIGN U
    Prop<::unicode::General_Category> { { 0x11EF5, 0x11EF6 }, ::unicode::General_Category::Spacing_Mark }, // [2] MAKASAR VOWEL SIGN E..MAKASAR VOWEL SIGN O
    Prop<::unicode::General_Category> { { 0x11EF7, 0x11EF8 }, ::unicode::General_Category::Other_Punctuation }, // [2] MAKASAR PASSIMBANG..MAKASAR END OF SECTION
    Prop<::unicode::General_Category> { { 0x11EF9, 0x11EFF }, ::unicode::General_Category::Unassigned }, // [7] <reserved-11EF9>..<reserved-11EFF>
    Prop<::unicode::General_Category> { { 0x11F00, 0x11F01 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] KAWI SIGN CANDRABINDU..KAWI SIGN ANUSVARA
    Prop<::unicode::General_Category> { { 0x11F02, 0x11F02 }, ::unicode::General_Category::Other_Letter }, // KAWI SIGN REPHA
    Prop<::unicode::General_Category> { { 0x11F03, 0x11F03 }, ::unicode::General_Category::Spacing_Mark }, // KAWI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x11F04, 0x11F10 }, ::unicode::General_Category::Other_Letter }, // [13] KAWI LETTER A..KAWI LETTER O
    Prop<::unicode::General_Category> { { 0x11F11, 0x11F11 }, ::unicode::General_Category::Unassigned }, // <reserved-11F11>
    Prop<::unicode::General_Category> { { 0x11F12, 0x11F33 }, ::unicode::General_Category::Other_Letter }, // [34] KAWI LETTER KA..KAWI LETTER JNYA
    Prop<::unicode::General_Category> { { 0x11F34, 0x11F35 }, ::unicode::General_Category::Spacing_Mark }, // [2] KAWI VOWEL SIGN AA..KAWI VOWEL SIGN ALTERNATE AA
    Prop<::unicode::General_Category> { { 0x11F36, 0x11F3A }, ::unicode::General_Category::Nonspacing_Mark }, // [5] KAWI VOWEL SIGN I..KAWI VOWEL SIGN VOCALIC R
    Prop<::unicode::General_Category> { { 0x11F3B, 0x11F3D }, ::unicode::General_Category::Unassigned }, // [3] <reserved-11F3B>..<reserved-11F3D>
    Prop<::unicode::General_Category> { { 0x11F3E, 0x11F3F }, ::unicode::General_Category::Spacing_Mark }, // [2] KAWI VOWEL SIGN E..KAWI VOWEL SIGN AI
    Prop<::unicode::General_Category> { { 0x11F40, 0x11F40 }, ::unicode::General_Category::Nonspacing_Mark }, // KAWI VOWEL SIGN EU
    Prop<::unicode::General_Category> { { 0x11F41, 0x11F41 }, ::unicode::General_Category::Spacing_Mark }, // KAWI SIGN KILLER
    Prop<::unicode::General_Category> { { 0x11F42, 0x11F42 }, ::unicode::General_Category::Nonspacing_Mark }, // KAWI CONJOINER
    Prop<::unicode::General_Category> { { 0x11F43, 0x11F4F }, ::unicode::General_Category::Other_Punctuation }, // [13] KAWI DANDA..KAWI PUNCTUATION CLOSING SPIRAL
    Prop<::unicode::General_Category> { { 0x11F50, 0x11F59 }, ::unicode::General_Category::Decimal_Number }, // [10] KAWI DIGIT ZERO..KAWI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x11F5A, 0x11F5A }, ::unicode::General_Category::Nonspacing_Mark }, // KAWI SIGN NUKTA
    Prop<::unicode::General_Category> { { 0x11F5B, 0x11FAF }, ::unicode::General_Category::Unassigned }, // [85] <reserved-11F5B>..<reserved-11FAF>
    Prop<::unicode::General_Category> { { 0x11FB0, 0x11FB0 }, ::unicode::General_Category::Other_Letter }, // LISU LETTER YHA
    Prop<::unicode::General_Category> { { 0x11FB1, 0x11FBF }, ::unicode::General_Category::Unassigned }, // [15] <reserved-11FB1>..<reserved-11FBF>
    Prop<::unicode::General_Category> { { 0x11FC0, 0x11FD4 }, ::unicode::General_Category::Other_Number }, // [21] TAMIL FRACTION ONE THREE-HUNDRED-AND-TWENTIETH..TAMIL FRACTION DOWNSCALING FACTOR KIIZH
    Prop<::unicode::General_Category> { { 0x11FD5, 0x11FDC }, ::unicode::General_Category::Other_Symbol }, // [8] TAMIL SIGN NEL..TAMIL SIGN MUKKURUNI
    Prop<::unicode::General_Category> { { 0x11FDD, 0x11FE0 }, ::unicode::General_Category::Currency_Symbol }, // [4] TAMIL SIGN KAACU..TAMIL SIGN VARAAKAN
    Prop<::unicode::General_Category> { { 0x11FE1, 0x11FF1 }, ::unicode::General_Category::Other_Symbol }, // [17] TAMIL SIGN PAARAM..TAMIL SIGN VAKAIYARAA
    Prop<::unicode::General_Category> { { 0x11FF2, 0x11FFE }, ::unicode::General_Category::Unassigned }, // [13] <reserved-11FF2>..<reserved-11FFE>
    Prop<::unicode::General_Category> { { 0x11FFF, 0x11FFF }, ::unicode::General_Category::Other_Punctuation }, // TAMIL PUNCTUATION END OF TEXT
    Prop<::unicode::General_Category> { { 0x12000, 0x12399 }, ::unicode::General_Category::Other_Letter }, // [922] CUNEIFORM SIGN A..CUNEIFORM SIGN U U
    Prop<::unicode::General_Category> { { 0x1239A, 0x123FF }, ::unicode::General_Category::Unassigned }, // [102] <reserved-1239A>..<reserved-123FF>
    Prop<::unicode::General_Category> { { 0x12400, 0x1246E }, ::unicode::General_Category::Letter_Number }, // [111] CUNEIFORM NUMERIC SIGN TWO ASH..CUNEIFORM NUMERIC SIGN NINE U VARIANT FORM
    Prop<::unicode::General_Category> { { 0x1246F, 0x1246F }, ::unicode::General_Category::Unassigned }, // <reserved-1246F>
    Prop<::unicode::General_Category> { { 0x12470, 0x12474 }, ::unicode::General_Category::Other_Punctuation }, // [5] CUNEIFORM PUNCTUATION SIGN OLD ASSYRIAN WORD DIVIDER..CUNEIFORM PUNCTUATION SIGN DIAGONAL QUADCOLON
    Prop<::unicode::General_Category> { { 0x12475, 0x1247F }, ::unicode::General_Category::Unassigned }, // [11] <reserved-12475>..<reserved-1247F>
    Prop<::unicode::General_Category> { { 0x12480, 0x12543 }, ::unicode::General_Category::Other_Letter }, // [196] CUNEIFORM SIGN AB TIMES NUN TENU..CUNEIFORM SIGN ZU5 TIMES THREE DISH TENU
    Prop<::unicode::General_Category> { { 0x12544, 0x12F8F }, ::unicode::General_Category::Unassigned }, // [2636] <reserved-12544>..<reserved-12F8F>
    Prop<::unicode::General_Category> { { 0x12F90, 0x12FF0 }, ::unicode::General_Category::Other_Letter }, // [97] CYPRO-MINOAN SIGN CM001..CYPRO-MINOAN SIGN CM114
    Prop<::unicode::General_Category> { { 0x12FF1, 0x12FF2 }, ::unicode::General_Category::Other_Punctuation }, // [2] CYPRO-MINOAN SIGN CM301..CYPRO-MINOAN SIGN CM302
    Prop<::unicode::General_Category> { { 0x12FF3, 0x12FFF }, ::unicode::General_Category::Unassigned }, // [13] <reserved-12FF3>..<reserved-12FFF>
    Prop<::unicode::General_Category> { { 0x13000, 0x1342F }, ::unicode::General_Category::Other_Letter }, // [1072] EGYPTIAN HIEROGLYPH A001..EGYPTIAN HIEROGLYPH V011D
    Prop<::unicode::General_Category> { { 0x13430, 0x1343F }, ::unicode::General_Category::Format }, // [16] EGYPTIAN HIEROGLYPH VERTICAL JOINER..EGYPTIAN HIEROGLYPH END WALLED ENCLOSURE
    Prop<::unicode::General_Category> { { 0x13440, 0x13440 }, ::unicode::General_Category::Nonspacing_Mark }, // EGYPTIAN HIEROGLYPH MIRROR HORIZONTALLY
    Prop<::unicode::General_Category> { { 0x13441, 0x13446 }, ::unicode::General_Category::Other_Letter }, // [6] EGYPTIAN HIEROGLYPH FULL BLANK..EGYPTIAN HIEROGLYPH WIDE LOST SIGN
    Prop<::unicode::General_Category> { { 0x13447, 0x13455 }, ::unicode::General_Category::Nonspacing_Mark }, // [15] EGYPTIAN HIEROGLYPH MODIFIER DAMAGED AT TOP START..EGYPTIAN HIEROGLYPH MODIFIER DAMAGED
    Prop<::unicode::General_Category> { { 0x13456, 0x1345F }, ::unicode::General_Category::Unassigned }, // [10] <reserved-13456>..<reserved-1345F>
    Prop<::unicode::General_Category> { { 0x13460, 0x143FA }, ::unicode::General_Category::Other_Letter }, // [3995] EGYPTIAN HIEROGLYPH-13460..EGYPTIAN HIEROGLYPH-143FA
    Prop<::unicode::General_Category> { { 0x143FB, 0x143FF }, ::unicode::General_Category::Unassigned }, // [5] <reserved-143FB>..<reserved-143FF>
    Prop<::unicode::General_Category> { { 0x14400, 0x14646 }, ::unicode::General_Category::Other_Letter }, // [583] ANATOLIAN HIEROGLYPH A001..ANATOLIAN HIEROGLYPH A530
    Prop<::unicode::General_Category> { { 0x14647, 0x160FF }, ::unicode::General_Category::Unassigned }, // [6841] <reserved-14647>..<reserved-160FF>
    Prop<::unicode::General_Category> { { 0x16100, 0x1611D }, ::unicode::General_Category::Other_Letter }, // [30] GURUNG KHEMA LETTER A..GURUNG KHEMA LETTER SA
    Prop<::unicode::General_Category> { { 0x1611E, 0x16129 }, ::unicode::General_Category::Nonspacing_Mark }, // [12] GURUNG KHEMA VOWEL SIGN AA..GURUNG KHEMA VOWEL LENGTH MARK
    Prop<::unicode::General_Category> { { 0x1612A, 0x1612C }, ::unicode::General_Category::Spacing_Mark }, // [3] GURUNG KHEMA CONSONANT SIGN MEDIAL YA..GURUNG KHEMA CONSONANT SIGN MEDIAL HA
    Prop<::unicode::General_Category> { { 0x1612D, 0x1612F }, ::unicode::General_Category::Nonspacing_Mark }, // [3] GURUNG KHEMA SIGN ANUSVARA..GURUNG KHEMA SIGN THOLHOMA
    Prop<::unicode::General_Category> { { 0x16130, 0x16139 }, ::unicode::General_Category::Decimal_Number }, // [10] GURUNG KHEMA DIGIT ZERO..GURUNG KHEMA DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1613A, 0x167FF }, ::unicode::General_Category::Unassigned }, // [1734] <reserved-1613A>..<reserved-167FF>
    Prop<::unicode::General_Category> { { 0x16800, 0x16A38 }, ::unicode::General_Category::Other_Letter }, // [569] BAMUM LETTER PHASE-A NGKUE MFON..BAMUM LETTER PHASE-F VUEQ
    Prop<::unicode::General_Category> { { 0x16A39, 0x16A3F }, ::unicode::General_Category::Unassigned }, // [7] <reserved-16A39>..<reserved-16A3F>
    Prop<::unicode::General_Category> { { 0x16A40, 0x16A5E }, ::unicode::General_Category::Other_Letter }, // [31] MRO LETTER TA..MRO LETTER TEK
    Prop<::unicode::General_Category> { { 0x16A5F, 0x16A5F }, ::unicode::General_Category::Unassigned }, // <reserved-16A5F>
    Prop<::unicode::General_Category> { { 0x16A60, 0x16A69 }, ::unicode::General_Category::Decimal_Number }, // [10] MRO DIGIT ZERO..MRO DIGIT NINE
    Prop<::unicode::General_Category> { { 0x16A6A, 0x16A6D }, ::unicode::General_Category::Unassigned }, // [4] <reserved-16A6A>..<reserved-16A6D>
    Prop<::unicode::General_Category> { { 0x16A6E, 0x16A6F }, ::unicode::General_Category::Other_Punctuation }, // [2] MRO DANDA..MRO DOUBLE DANDA
    Prop<::unicode::General_Category> { { 0x16A70, 0x16ABE }, ::unicode::General_Category::Other_Letter }, // [79] TANGSA LETTER OZ..TANGSA LETTER ZA
    Prop<::unicode::General_Category> { { 0x16ABF, 0x16ABF }, ::unicode::General_Category::Unassigned }, // <reserved-16ABF>
    Prop<::unicode::General_Category> { { 0x16AC0, 0x16AC9 }, ::unicode::General_Category::Decimal_Number }, // [10] TANGSA DIGIT ZERO..TANGSA DIGIT NINE
    Prop<::unicode::General_Category> { { 0x16ACA, 0x16ACF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-16ACA>..<reserved-16ACF>
    Prop<::unicode::General_Category> { { 0x16AD0, 0x16AED }, ::unicode::General_Category::Other_Letter }, // [30] BASSA VAH LETTER ENNI..BASSA VAH LETTER I
    Prop<::unicode::General_Category> { { 0x16AEE, 0x16AEF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-16AEE>..<reserved-16AEF>
    Prop<::unicode::General_Category> { { 0x16AF0, 0x16AF4 }, ::unicode::General_Category::Nonspacing_Mark }, // [5] BASSA VAH COMBINING HIGH TONE..BASSA VAH COMBINING HIGH-LOW TONE
    Prop<::unicode::General_Category> { { 0x16AF5, 0x16AF5 }, ::unicode::General_Category::Other_Punctuation }, // BASSA VAH FULL STOP
    Prop<::unicode::General_Category> { { 0x16AF6, 0x16AFF }, ::unicode::General_Category::Unassigned }, // [10] <reserved-16AF6>..<reserved-16AFF>
    Prop<::unicode::General_Category> { { 0x16B00, 0x16B2F }, ::unicode::General_Category::Other_Letter }, // [48] PAHAWH HMONG VOWEL KEEB..PAHAWH HMONG CONSONANT CAU
    Prop<::unicode::General_Category> { { 0x16B30, 0x16B36 }, ::unicode::General_Category::Nonspacing_Mark }, // [7] PAHAWH HMONG MARK CIM TUB..PAHAWH HMONG MARK CIM TAUM
    Prop<::unicode::General_Category> { { 0x16B37, 0x16B3B }, ::unicode::General_Category::Other_Punctuation }, // [5] PAHAWH HMONG SIGN VOS THOM..PAHAWH HMONG SIGN VOS FEEM
    Prop<::unicode::General_Category> { { 0x16B3C, 0x16B3F }, ::unicode::General_Category::Other_Symbol }, // [4] PAHAWH HMONG SIGN XYEEM NTXIV..PAHAWH HMONG SIGN XYEEM FAIB
    Prop<::unicode::General_Category> { { 0x16B40, 0x16B43 }, ::unicode::General_Category::Modifier_Letter }, // [4] PAHAWH HMONG SIGN VOS SEEV..PAHAWH HMONG SIGN IB YAM
    Prop<::unicode::General_Category> { { 0x16B44, 0x16B44 }, ::unicode::General_Category::Other_Punctuation }, // PAHAWH HMONG SIGN XAUS
    Prop<::unicode::General_Category> { { 0x16B45, 0x16B45 }, ::unicode::General_Category::Other_Symbol }, // PAHAWH HMONG SIGN CIM TSOV ROG
    Prop<::unicode::General_Category> { { 0x16B46, 0x16B4F }, ::unicode::General_Category::Unassigned }, // [10] <reserved-16B46>..<reserved-16B4F>
    Prop<::unicode::General_Category> { { 0x16B50, 0x16B59 }, ::unicode::General_Category::Decimal_Number }, // [10] PAHAWH HMONG DIGIT ZERO..PAHAWH HMONG DIGIT NINE
    Prop<::unicode::General_Category> { { 0x16B5A, 0x16B5A }, ::unicode::General_Category::Unassigned }, // <reserved-16B5A>
    Prop<::unicode::General_Category> { { 0x16B5B, 0x16B61 }, ::unicode::General_Category::Other_Number }, // [7] PAHAWH HMONG NUMBER TENS..PAHAWH HMONG NUMBER TRILLIONS
    Prop<::unicode::General_Category> { { 0x16B62, 0x16B62 }, ::unicode::General_Category::Unassigned }, // <reserved-16B62>
    Prop<::unicode::General_Category> { { 0x16B63, 0x16B77 }, ::unicode::General_Category::Other_Letter }, // [21] PAHAWH HMONG SIGN VOS LUB..PAHAWH HMONG SIGN CIM NRES TOS
    Prop<::unicode::General_Category> { { 0x16B78, 0x16B7C }, ::unicode::General_Category::Unassigned }, // [5] <reserved-16B78>..<reserved-16B7C>
    Prop<::unicode::General_Category> { { 0x16B7D, 0x16B8F }, ::unicode::General_Category::Other_Letter }, // [19] PAHAWH HMONG CLAN SIGN TSHEEJ..PAHAWH HMONG CLAN SIGN VWJ
    Prop<::unicode::General_Category> { { 0x16B90, 0x16D3F }, ::unicode::General_Category::Unassigned }, // [432] <reserved-16B90>..<reserved-16D3F>
    Prop<::unicode::General_Category> { { 0x16D40, 0x16D42 }, ::unicode::General_Category::Modifier_Letter }, // [3] KIRAT RAI SIGN ANUSVARA..KIRAT RAI SIGN VISARGA
    Prop<::unicode::General_Category> { { 0x16D43, 0x16D6A }, ::unicode::General_Category::Other_Letter }, // [40] KIRAT RAI LETTER A..KIRAT RAI VOWEL SIGN AU
    Prop<::unicode::General_Category> { { 0x16D6B, 0x16D6C }, ::unicode::General_Category::Modifier_Letter }, // [2] KIRAT RAI SIGN VIRAMA..KIRAT RAI SIGN SAAT
    Prop<::unicode::General_Category> { { 0x16D6D, 0x16D6F }, ::unicode::General_Category::Other_Punctuation }, // [3] KIRAT RAI SIGN YUPI..KIRAT RAI DOUBLE DANDA
    Prop<::unicode::General_Category> { { 0x16D70, 0x16D79 }, ::unicode::General_Category::Decimal_Number }, // [10] KIRAT RAI DIGIT ZERO..KIRAT RAI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x16D7A, 0x16E3F }, ::unicode::General_Category::Unassigned }, // [198] <reserved-16D7A>..<reserved-16E3F>
    Prop<::unicode::General_Category> { { 0x16E40, 0x16E5F }, ::unicode::General_Category::Uppercase_Letter }, // [32] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN CAPITAL LETTER Y
    Prop<::unicode::General_Category> { { 0x16E60, 0x16E7F }, ::unicode::General_Category::Lowercase_Letter }, // [32] MEDEFAIDRIN SMALL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Prop<::unicode::General_Category> { { 0x16E80, 0x16E96 }, ::unicode::General_Category::Other_Number }, // [23] MEDEFAIDRIN DIGIT ZERO..MEDEFAIDRIN DIGIT THREE ALTERNATE FORM
    Prop<::unicode::General_Category> { { 0x16E97, 0x16E9A }, ::unicode::General_Category::Other_Punctuation }, // [4] MEDEFAIDRIN COMMA..MEDEFAIDRIN EXCLAMATION OH
    Prop<::unicode::General_Category> { { 0x16E9B, 0x16E9F }, ::unicode::General_Category::Unassigned }, // [5] <reserved-16E9B>..<reserved-16E9F>
    Prop<::unicode::General_Category> { { 0x16EA0, 0x16EB8 }, ::unicode::General_Category::Uppercase_Letter }, // [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Prop<::unicode::General_Category> { { 0x16EB9, 0x16EBA }, ::unicode::General_Category::Unassigned }, // [2] <reserved-16EB9>..<reserved-16EBA>
    Prop<::unicode::General_Category> { { 0x16EBB, 0x16ED3 }, ::unicode::General_Category::Lowercase_Letter }, // [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Prop<::unicode::General_Category> { { 0x16ED4, 0x16EFF }, ::unicode::General_Category::Unassigned }, // [44] <reserved-16ED4>..<reserved-16EFF>
    Prop<::unicode::General_Category> { { 0x16F00, 0x16F4A }, ::unicode::General_Category::Other_Letter }, // [75] MIAO LETTER PA..MIAO LETTER RTE
    Prop<::unicode::General_Category> { { 0x16F4B, 0x16F4E }, ::unicode::General_Category::Unassigned }, // [4] <reserved-16F4B>..<reserved-16F4E>
    Prop<::unicode::General_Category> { { 0x16F4F, 0x16F4F }, ::unicode::General_Category::Nonspacing_Mark }, // MIAO SIGN CONSONANT MODIFIER BAR
    Prop<::unicode::General_Category> { { 0x16F50, 0x16F50 }, ::unicode::General_Category::Other_Letter }, // MIAO LETTER NASALIZATION
    Prop<::unicode::General_Category> { { 0x16F51, 0x16F87 }, ::unicode::General_Category::Spacing_Mark }, // [55] MIAO SIGN ASPIRATION..MIAO VOWEL SIGN UI
    Prop<::unicode::General_Category> { { 0x16F88, 0x16F8E }, ::unicode::General_Category::Unassigned }, // [7] <reserved-16F88>..<reserved-16F8E>
    Prop<::unicode::General_Category> { { 0x16F8F, 0x16F92 }, ::unicode::General_Category::Nonspacing_Mark }, // [4] MIAO TONE RIGHT..MIAO TONE BELOW
    Prop<::unicode::General_Category> { { 0x16F93, 0x16F9F }, ::unicode::General_Category::Modifier_Letter }, // [13] MIAO LETTER TONE-2..MIAO LETTER REFORMED TONE-8
    Prop<::unicode::General_Category> { { 0x16FA0, 0x16FDF }, ::unicode::General_Category::Unassigned }, // [64] <reserved-16FA0>..<reserved-16FDF>
    Prop<::unicode::General_Category> { { 0x16FE0, 0x16FE1 }, ::unicode::General_Category::Modifier_Letter }, // [2] TANGUT ITERATION MARK..NUSHU ITERATION MARK
    Prop<::unicode::General_Category> { { 0x16FE2, 0x16FE2 }, ::unicode::General_Category::Other_Punctuation }, // OLD CHINESE HOOK MARK
    Prop<::unicode::General_Category> { { 0x16FE3, 0x16FE3 }, ::unicode::General_Category::Modifier_Letter }, // OLD CHINESE ITERATION MARK
    Prop<::unicode::General_Category> { { 0x16FE4, 0x16FE4 }, ::unicode::General_Category::Nonspacing_Mark }, // KHITAN SMALL SCRIPT FILLER
    Prop<::unicode::General_Category> { { 0x16FE5, 0x16FEF }, ::unicode::General_Category::Unassigned }, // [11] <reserved-16FE5>..<reserved-16FEF>
    Prop<::unicode::General_Category> { { 0x16FF0, 0x16FF1 }, ::unicode::General_Category::Spacing_Mark }, // [2] VIETNAMESE ALTERNATE READING MARK CA..VIETNAMESE ALTERNATE READING MARK NHAY
    Prop<::unicode::General_Category> { { 0x16FF2, 0x16FF3 }, ::unicode::General_Category::Modifier_Letter }, // [2] CHINESE SMALL SIMPLIFIED ER..CHINESE SMALL TRADITIONAL ER
    Prop<::unicode::General_Category> { { 0x16FF4, 0x16FF6 }, ::unicode::General_Category::Letter_Number }, // [3] YANGQIN SIGN SLOW ONE BEAT..YANGQIN SIGN SLOW TWO BEATS
    Prop<::unicode::General_Category> { { 0x16FF7, 0x16FFF }, ::unicode::General_Category::Unassigned }, // [9] <reserved-16FF7>..<reserved-16FFF>
    Prop<::unicode::General_Category> { { 0x17000, 0x18CD5 }, ::unicode::General_Category::Other_Letter }, // [7382] TANGUT IDEOGRAPH-17000..KHITAN SMALL SCRIPT CHARACTER-18CD5
    Prop<::unicode::General_Category> { { 0x18CD6, 0x18CFE }, ::unicode::General_Category::Unassigned }, // [41] <reserved-18CD6>..<reserved-18CFE>
    Prop<::unicode::General_Category> { { 0x18CFF, 0x18D1E }, ::unicode::General_Category::Other_Letter }, // [32] KHITAN SMALL SCRIPT CHARACTER-18CFF..TANGUT IDEOGRAPH-18D1E
    Prop<::unicode::General_Category> { { 0x18D1F, 0x18D7F }, ::unicode::General_Category::Unassigned }, // [97] <reserved-18D1F>..<reserved-18D7F>
    Prop<::unicode::General_Category> { { 0x18D80, 0x18DF2 }, ::unicode::General_Category::Other_Letter }, // [115] TANGUT COMPONENT-769..TANGUT COMPONENT-883
    Prop<::unicode::General_Category> { { 0x18DF3, 0x1AFEF }, ::unicode::General_Category::Unassigned }, // [8701] <reserved-18DF3>..<reserved-1AFEF>
    Prop<::unicode::General_Category> { { 0x1AFF0, 0x1AFF3 }, ::unicode::General_Category::Modifier_Letter }, // [4] KATAKANA LETTER MINNAN TONE-2..KATAKANA LETTER MINNAN TONE-5
    Prop<::unicode::General_Category> { { 0x1AFF4, 0x1AFF4 }, ::unicode::General_Category::Unassigned }, // <reserved-1AFF4>
    Prop<::unicode::General_Category> { { 0x1AFF5, 0x1AFFB }, ::unicode::General_Category::Modifier_Letter }, // [7] KATAKANA LETTER MINNAN TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-5
    Prop<::unicode::General_Category> { { 0x1AFFC, 0x1AFFC }, ::unicode::General_Category::Unassigned }, // <reserved-1AFFC>
    Prop<::unicode::General_Category> { { 0x1AFFD, 0x1AFFE }, ::unicode::General_Category::Modifier_Letter }, // [2] KATAKANA LETTER MINNAN NASALIZED TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-8
    Prop<::unicode::General_Category> { { 0x1AFFF, 0x1AFFF }, ::unicode::General_Category::Unassigned }, // <reserved-1AFFF>
    Prop<::unicode::General_Category> { { 0x1B000, 0x1B122 }, ::unicode::General_Category::Other_Letter }, // [291] KATAKANA LETTER ARCHAIC E..KATAKANA LETTER ARCHAIC WU
    Prop<::unicode::General_Category> { { 0x1B123, 0x1B131 }, ::unicode::General_Category::Unassigned }, // [15] <reserved-1B123>..<reserved-1B131>
    Prop<::unicode::General_Category> { { 0x1B132, 0x1B132 }, ::unicode::General_Category::Other_Letter }, // HIRAGANA LETTER SMALL KO
    Prop<::unicode::General_Category> { { 0x1B133, 0x1B14F }, ::unicode::General_Category::Unassigned }, // [29] <reserved-1B133>..<reserved-1B14F>
    Prop<::unicode::General_Category> { { 0x1B150, 0x1B152 }, ::unicode::General_Category::Other_Letter }, // [3] HIRAGANA LETTER SMALL WI..HIRAGANA LETTER SMALL WO
    Prop<::unicode::General_Category> { { 0x1B153, 0x1B154 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1B153>..<reserved-1B154>
    Prop<::unicode::General_Category> { { 0x1B155, 0x1B155 }, ::unicode::General_Category::Other_Letter }, // KATAKANA LETTER SMALL KO
    Prop<::unicode::General_Category> { { 0x1B156, 0x1B163 }, ::unicode::General_Category::Unassigned }, // [14] <reserved-1B156>..<reserved-1B163>
    Prop<::unicode::General_Category> { { 0x1B164, 0x1B167 }, ::unicode::General_Category::Other_Letter }, // [4] KATAKANA LETTER SMALL WI..KATAKANA LETTER SMALL N
    Prop<::unicode::General_Category> { { 0x1B168, 0x1B16F }, ::unicode::General_Category::Unassigned }, // [8] <reserved-1B168>..<reserved-1B16F>
    Prop<::unicode::General_Category> { { 0x1B170, 0x1B2FB }, ::unicode::General_Category::Other_Letter }, // [396] NUSHU CHARACTER-1B170..NUSHU CHARACTER-1B2FB
    Prop<::unicode::General_Category> { { 0x1B2FC, 0x1BBFF }, ::unicode::General_Category::Unassigned }, // [2308] <reserved-1B2FC>..<reserved-1BBFF>
    Prop<::unicode::General_Category> { { 0x1BC00, 0x1BC6A }, ::unicode::General_Category::Other_Letter }, // [107] DUPLOYAN LETTER H..DUPLOYAN LETTER VOCALIC M
    Prop<::unicode::General_Category> { { 0x1BC6B, 0x1BC6F }, ::unicode::General_Category::Unassigned }, // [5] <reserved-1BC6B>..<reserved-1BC6F>
    Prop<::unicode::General_Category> { { 0x1BC70, 0x1BC7C }, ::unicode::General_Category::Other_Letter }, // [13] DUPLOYAN AFFIX LEFT HORIZONTAL SECANT..DUPLOYAN AFFIX ATTACHED TANGENT HOOK
    Prop<::unicode::General_Category> { { 0x1BC7D, 0x1BC7F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1BC7D>..<reserved-1BC7F>
    Prop<::unicode::General_Category> { { 0x1BC80, 0x1BC88 }, ::unicode::General_Category::Other_Letter }, // [9] DUPLOYAN AFFIX HIGH ACUTE..DUPLOYAN AFFIX HIGH VERTICAL
    Prop<::unicode::General_Category> { { 0x1BC89, 0x1BC8F }, ::unicode::General_Category::Unassigned }, // [7] <reserved-1BC89>..<reserved-1BC8F>
    Prop<::unicode::General_Category> { { 0x1BC90, 0x1BC99 }, ::unicode::General_Category::Other_Letter }, // [10] DUPLOYAN AFFIX LOW ACUTE..DUPLOYAN AFFIX LOW ARROW
    Prop<::unicode::General_Category> { { 0x1BC9A, 0x1BC9B }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1BC9A>..<reserved-1BC9B>
    Prop<::unicode::General_Category> { { 0x1BC9C, 0x1BC9C }, ::unicode::General_Category::Other_Symbol }, // DUPLOYAN SIGN O WITH CROSS
    Prop<::unicode::General_Category> { { 0x1BC9D, 0x1BC9E }, ::unicode::General_Category::Nonspacing_Mark }, // [2] DUPLOYAN THICK LETTER SELECTOR..DUPLOYAN DOUBLE MARK
    Prop<::unicode::General_Category> { { 0x1BC9F, 0x1BC9F }, ::unicode::General_Category::Other_Punctuation }, // DUPLOYAN PUNCTUATION CHINOOK FULL STOP
    Prop<::unicode::General_Category> { { 0x1BCA0, 0x1BCA3 }, ::unicode::General_Category::Format }, // [4] SHORTHAND FORMAT LETTER OVERLAP..SHORTHAND FORMAT UP STEP
    Prop<::unicode::General_Category> { { 0x1BCA4, 0x1CBFF }, ::unicode::General_Category::Unassigned }, // [3932] <reserved-1BCA4>..<reserved-1CBFF>
    Prop<::unicode::General_Category> { { 0x1CC00, 0x1CCEF }, ::unicode::General_Category::Other_Symbol }, // [240] UP-POINTING GO-KART..OUTLINED LATIN CAPITAL LETTER Z
    Prop<::unicode::General_Category> { { 0x1CCF0, 0x1CCF9 }, ::unicode::General_Category::Decimal_Number }, // [10] OUTLINED DIGIT ZERO..OUTLINED DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1CCFA, 0x1CCFC }, ::unicode::General_Category::Other_Symbol }, // [3] SNAKE SYMBOL..NOSE SYMBOL
    Prop<::unicode::General_Category> { { 0x1CCFD, 0x1CCFF }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1CCFD>..<reserved-1CCFF>
    Prop<::unicode::General_Category> { { 0x1CD00, 0x1CEB3 }, ::unicode::General_Category::Other_Symbol }, // [436] BLOCK OCTANT-3..BLACK RIGHT TRIANGLE CARET
    Prop<::unicode::General_Category> { { 0x1CEB4, 0x1CEB9 }, ::unicode::General_Category::Unassigned }, // [6] <reserved-1CEB4>..<reserved-1CEB9>
    Prop<::unicode::General_Category> { { 0x1CEBA, 0x1CED0 }, ::unicode::General_Category::Other_Symbol }, // [23] FRAGILE SYMBOL..LEUKOTHEA
    Prop<::unicode::General_Category> { { 0x1CED1, 0x1CEDF }, ::unicode::General_Category::Unassigned }, // [15] <reserved-1CED1>..<reserved-1CEDF>
    Prop<::unicode::General_Category> { { 0x1CEE0, 0x1CEEF }, ::unicode::General_Category::Other_Symbol }, // [16] GEOMANTIC FIGURE POPULUS..GEOMANTIC FIGURE VIA
    Prop<::unicode::General_Category> { { 0x1CEF0, 0x1CEF0 }, ::unicode::General_Category::Math_Symbol }, // MEDIUM SMALL WHITE CIRCLE WITH HORIZONTAL BAR
    Prop<::unicode::General_Category> { { 0x1CEF1, 0x1CEFF }, ::unicode::General_Category::Unassigned }, // [15] <reserved-1CEF1>..<reserved-1CEFF>
    Prop<::unicode::General_Category> { { 0x1CF00, 0x1CF2D }, ::unicode::General_Category::Nonspacing_Mark }, // [46] ZNAMENNY COMBINING MARK GORAZDO NIZKO S KRYZHEM ON LEFT..ZNAMENNY COMBINING MARK KRYZH ON LEFT
    Prop<::unicode::General_Category> { { 0x1CF2E, 0x1CF2F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1CF2E>..<reserved-1CF2F>
    Prop<::unicode::General_Category> { { 0x1CF30, 0x1CF46 }, ::unicode::General_Category::Nonspacing_Mark }, // [23] ZNAMENNY COMBINING TONAL RANGE MARK MRACHNO..ZNAMENNY PRIZNAK MODIFIER ROG
    Prop<::unicode::General_Category> { { 0x1CF47, 0x1CF4F }, ::unicode::General_Category::Unassigned }, // [9] <reserved-1CF47>..<reserved-1CF4F>
    Prop<::unicode::General_Category> { { 0x1CF50, 0x1CFC3 }, ::unicode::General_Category::Other_Symbol }, // [116] ZNAMENNY NEUME KRYUK..ZNAMENNY NEUME PAUK
    Prop<::unicode::General_Category> { { 0x1CFC4, 0x1CFFF }, ::unicode::General_Category::Unassigned }, // [60] <reserved-1CFC4>..<reserved-1CFFF>
    Prop<::unicode::General_Category> { { 0x1D000, 0x1D0F5 }, ::unicode::General_Category::Other_Symbol }, // [246] BYZANTINE MUSICAL SYMBOL PSILI..BYZANTINE MUSICAL SYMBOL GORGON NEO KATO
    Prop<::unicode::General_Category> { { 0x1D0F6, 0x1D0FF }, ::unicode::General_Category::Unassigned }, // [10] <reserved-1D0F6>..<reserved-1D0FF>
    Prop<::unicode::General_Category> { { 0x1D100, 0x1D126 }, ::unicode::General_Category::Other_Symbol }, // [39] MUSICAL SYMBOL SINGLE BARLINE..MUSICAL SYMBOL DRUM CLEF-2
    Prop<::unicode::General_Category> { { 0x1D127, 0x1D128 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1D127>..<reserved-1D128>
    Prop<::unicode::General_Category> { { 0x1D129, 0x1D164 }, ::unicode::General_Category::Other_Symbol }, // [60] MUSICAL SYMBOL MULTIPLE MEASURE REST..MUSICAL SYMBOL ONE HUNDRED TWENTY-EIGHTH NOTE
    Prop<::unicode::General_Category> { { 0x1D165, 0x1D166 }, ::unicode::General_Category::Spacing_Mark }, // [2] MUSICAL SYMBOL COMBINING STEM..MUSICAL SYMBOL COMBINING SPRECHGESANG STEM
    Prop<::unicode::General_Category> { { 0x1D167, 0x1D169 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] MUSICAL SYMBOL COMBINING TREMOLO-1..MUSICAL SYMBOL COMBINING TREMOLO-3
    Prop<::unicode::General_Category> { { 0x1D16A, 0x1D16C }, ::unicode::General_Category::Other_Symbol }, // [3] MUSICAL SYMBOL FINGERED TREMOLO-1..MUSICAL SYMBOL FINGERED TREMOLO-3
    Prop<::unicode::General_Category> { { 0x1D16D, 0x1D172 }, ::unicode::General_Category::Spacing_Mark }, // [6] MUSICAL SYMBOL COMBINING AUGMENTATION DOT..MUSICAL SYMBOL COMBINING FLAG-5
    Prop<::unicode::General_Category> { { 0x1D173, 0x1D17A }, ::unicode::General_Category::Format }, // [8] MUSICAL SYMBOL BEGIN BEAM..MUSICAL SYMBOL END PHRASE
    Prop<::unicode::General_Category> { { 0x1D17B, 0x1D182 }, ::unicode::General_Category::Nonspacing_Mark }, // [8] MUSICAL SYMBOL COMBINING ACCENT..MUSICAL SYMBOL COMBINING LOURE
    Prop<::unicode::General_Category> { { 0x1D183, 0x1D184 }, ::unicode::General_Category::Other_Symbol }, // [2] MUSICAL SYMBOL ARPEGGIATO UP..MUSICAL SYMBOL ARPEGGIATO DOWN
    Prop<::unicode::General_Category> { { 0x1D185, 0x1D18B }, ::unicode::General_Category::Nonspacing_Mark }, // [7] MUSICAL SYMBOL COMBINING DOIT..MUSICAL SYMBOL COMBINING TRIPLE TONGUE
    Prop<::unicode::General_Category> { { 0x1D18C, 0x1D1A9 }, ::unicode::General_Category::Other_Symbol }, // [30] MUSICAL SYMBOL RINFORZANDO..MUSICAL SYMBOL DEGREE SLASH
    Prop<::unicode::General_Category> { { 0x1D1AA, 0x1D1AD }, ::unicode::General_Category::Nonspacing_Mark }, // [4] MUSICAL SYMBOL COMBINING DOWN BOW..MUSICAL SYMBOL COMBINING SNAP PIZZICATO
    Prop<::unicode::General_Category> { { 0x1D1AE, 0x1D1EA }, ::unicode::General_Category::Other_Symbol }, // [61] MUSICAL SYMBOL PEDAL MARK..MUSICAL SYMBOL KORON
    Prop<::unicode::General_Category> { { 0x1D1EB, 0x1D1FF }, ::unicode::General_Category::Unassigned }, // [21] <reserved-1D1EB>..<reserved-1D1FF>
    Prop<::unicode::General_Category> { { 0x1D200, 0x1D241 }, ::unicode::General_Category::Other_Symbol }, // [66] GREEK VOCAL NOTATION SYMBOL-1..GREEK INSTRUMENTAL NOTATION SYMBOL-54
    Prop<::unicode::General_Category> { { 0x1D242, 0x1D244 }, ::unicode::General_Category::Nonspacing_Mark }, // [3] COMBINING GREEK MUSICAL TRISEME..COMBINING GREEK MUSICAL PENTASEME
    Prop<::unicode::General_Category> { { 0x1D245, 0x1D245 }, ::unicode::General_Category::Other_Symbol }, // GREEK MUSICAL LEIMMA
    Prop<::unicode::General_Category> { { 0x1D246, 0x1D2BF }, ::unicode::General_Category::Unassigned }, // [122] <reserved-1D246>..<reserved-1D2BF>
    Prop<::unicode::General_Category> { { 0x1D2C0, 0x1D2D3 }, ::unicode::General_Category::Other_Number }, // [20] KAKTOVIK NUMERAL ZERO..KAKTOVIK NUMERAL NINETEEN
    Prop<::unicode::General_Category> { { 0x1D2D4, 0x1D2DF }, ::unicode::General_Category::Unassigned }, // [12] <reserved-1D2D4>..<reserved-1D2DF>
    Prop<::unicode::General_Category> { { 0x1D2E0, 0x1D2F3 }, ::unicode::General_Category::Other_Number }, // [20] MAYAN NUMERAL ZERO..MAYAN NUMERAL NINETEEN
    Prop<::unicode::General_Category> { { 0x1D2F4, 0x1D2FF }, ::unicode::General_Category::Unassigned }, // [12] <reserved-1D2F4>..<reserved-1D2FF>
    Prop<::unicode::General_Category> { { 0x1D300, 0x1D356 }, ::unicode::General_Category::Other_Symbol }, // [87] MONOGRAM FOR EARTH..TETRAGRAM FOR FOSTERING
    Prop<::unicode::General_Category> { { 0x1D357, 0x1D35F }, ::unicode::General_Category::Unassigned }, // [9] <reserved-1D357>..<reserved-1D35F>
    Prop<::unicode::General_Category> { { 0x1D360, 0x1D378 }, ::unicode::General_Category::Other_Number }, // [25] COUNTING ROD UNIT DIGIT ONE..TALLY MARK FIVE
    Prop<::unicode::General_Category> { { 0x1D379, 0x1D3FF }, ::unicode::General_Category::Unassigned }, // [135] <reserved-1D379>..<reserved-1D3FF>
    Prop<::unicode::General_Category> { { 0x1D400, 0x1D419 }, ::unicode::General_Category::Uppercase_Letter }, // [26] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL BOLD CAPITAL Z
    Prop<::unicode::General_Category> { { 0x1D41A, 0x1D433 }, ::unicode::General_Category::Lowercase_Letter }, // [26] MATHEMATICAL BOLD SMALL A..MATHEMATICAL BOLD SMALL Z
    Prop<::unicode::General_Category> { { 0x1D434, 0x1D44D }, ::unicode::General_Category::Uppercase_Letter }, // [26] MATHEMATICAL ITALIC CAPITAL A..MATHEMATICAL ITALIC CAPITAL Z
    Prop<::unicode::General_Category> { { 0x1D44E, 0x1D454 }, ::unicode::General_Category::Lowercase_Letter }, // [7] MATHEMATICAL ITALIC SMALL A..MATHEMATICAL ITALIC SMALL G
    Prop<::unicode::General_Category> { { 0x1D455, 0x1D455 }, ::unicode::General_Category::Unassigned }, // <reserved-1D455>
    Prop<::unicode::General_Category> { { 0x1D456, 0x1D467 }, ::unicode::General_Category::Lowercase_Letter }, // [18] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL ITALIC SMALL Z
    Prop<::unicode::General_Category> { { 0x1D468, 0x1D481 }, ::unicode::General_Category::Uppercase_Letter }, // [26] MATHEMATICAL BOLD ITALIC CAPITAL A..MATHEMATICAL BOLD ITALIC CAPITAL Z
    Prop<::unicode::General_Category> { { 0x1D482, 0x1D49B }, ::unicode::General_Category::Lowercase_Letter }, // [26] MATHEMATICAL BOLD ITALIC SMALL A..MATHEMATICAL BOLD ITALIC SMALL Z
    Prop<::unicode::General_Category> { { 0x1D49C, 0x1D49C }, ::unicode::General_Category::Uppercase_Letter }, // MATHEMATICAL SCRIPT CAPITAL A
    Prop<::unicode::General_Category> { { 0x1D49D, 0x1D49D }, ::unicode::General_Category::Unassigned }, // <reserved-1D49D>
    Prop<::unicode::General_Category> { { 0x1D49E, 0x1D49F }, ::unicode::General_Category::Uppercase_Letter }, // [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Prop<::unicode::General_Category> { { 0x1D4A0, 0x1D4A1 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1D4A0>..<reserved-1D4A1>
    Prop<::unicode::General_Category> { { 0x1D4A2, 0x1D4A2 }, ::unicode::General_Category::Uppercase_Letter }, // MATHEMATICAL SCRIPT CAPITAL G
    Prop<::unicode::General_Category> { { 0x1D4A3, 0x1D4A4 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1D4A3>..<reserved-1D4A4>
    Prop<::unicode::General_Category> { { 0x1D4A5, 0x1D4A6 }, ::unicode::General_Category::Uppercase_Letter }, // [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Prop<::unicode::General_Category> { { 0x1D4A7, 0x1D4A8 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1D4A7>..<reserved-1D4A8>
    Prop<::unicode::General_Category> { { 0x1D4A9, 0x1D4AC }, ::unicode::General_Category::Uppercase_Letter }, // [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Prop<::unicode::General_Category> { { 0x1D4AD, 0x1D4AD }, ::unicode::General_Category::Unassigned }, // <reserved-1D4AD>
    Prop<::unicode::General_Category> { { 0x1D4AE, 0x1D4B5 }, ::unicode::General_Category::Uppercase_Letter }, // [8] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT CAPITAL Z
    Prop<::unicode::General_Category> { { 0x1D4B6, 0x1D4B9 }, ::unicode::General_Category::Lowercase_Letter }, // [4] MATHEMATICAL SCRIPT SMALL A..MATHEMATICAL SCRIPT SMALL D
    Prop<::unicode::General_Category> { { 0x1D4BA, 0x1D4BA }, ::unicode::General_Category::Unassigned }, // <reserved-1D4BA>
    Prop<::unicode::General_Category> { { 0x1D4BB, 0x1D4BB }, ::unicode::General_Category::Lowercase_Letter }, // MATHEMATICAL SCRIPT SMALL F
    Prop<::unicode::General_Category> { { 0x1D4BC, 0x1D4BC }, ::unicode::General_Category::Unassigned }, // <reserved-1D4BC>
    Prop<::unicode::General_Category> { { 0x1D4BD, 0x1D4C3 }, ::unicode::General_Category::Lowercase_Letter }, // [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Prop<::unicode::General_Category> { { 0x1D4C4, 0x1D4C4 }, ::unicode::General_Category::Unassigned }, // <reserved-1D4C4>
    Prop<::unicode::General_Category> { { 0x1D4C5, 0x1D4CF }, ::unicode::General_Category::Lowercase_Letter }, // [11] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL SCRIPT SMALL Z
    Prop<::unicode::General_Category> { { 0x1D4D0, 0x1D4E9 }, ::unicode::General_Category::Uppercase_Letter }, // [26] MATHEMATICAL BOLD SCRIPT CAPITAL A..MATHEMATICAL BOLD SCRIPT CAPITAL Z
    Prop<::unicode::General_Category> { { 0x1D4EA, 0x1D503 }, ::unicode::General_Category::Lowercase_Letter }, // [26] MATHEMATICAL BOLD SCRIPT SMALL A..MATHEMATICAL BOLD SCRIPT SMALL Z
    Prop<::unicode::General_Category> { { 0x1D504, 0x1D505 }, ::unicode::General_Category::Uppercase_Letter }, // [2] MATHEMATICAL FRAKTUR CAPITAL A..MATHEMATICAL FRAKTUR CAPITAL B
    Prop<::unicode::General_Category> { { 0x1D506, 0x1D506 }, ::unicode::General_Category::Unassigned }, // <reserved-1D506>
    Prop<::unicode::General_Category> { { 0x1D507, 0x1D50A }, ::unicode::General_Category::Uppercase_Letter }, // [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Prop<::unicode::General_Category> { { 0x1D50B, 0x1D50C }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1D50B>..<reserved-1D50C>
    Prop<::unicode::General_Category> { { 0x1D50D, 0x1D514 }, ::unicode::General_Category::Uppercase_Letter }, // [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Prop<::unicode::General_Category> { { 0x1D515, 0x1D515 }, ::unicode::General_Category::Unassigned }, // <reserved-1D515>
    Prop<::unicode::General_Category> { { 0x1D516, 0x1D51C }, ::unicode::General_Category::Uppercase_Letter }, // [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Prop<::unicode::General_Category> { { 0x1D51D, 0x1D51D }, ::unicode::General_Category::Unassigned }, // <reserved-1D51D>
    Prop<::unicode::General_Category> { { 0x1D51E, 0x1D537 }, ::unicode::General_Category::Lowercase_Letter }, // [26] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL FRAKTUR SMALL Z
    Prop<::unicode::General_Category> { { 0x1D538, 0x1D539 }, ::unicode::General_Category::Uppercase_Letter }, // [2] MATHEMATICAL DOUBLE-STRUCK CAPITAL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Prop<::unicode::General_Category> { { 0x1D53A, 0x1D53A }, ::unicode::General_Category::Unassigned }, // <reserved-1D53A>
    Prop<::unicode::General_Category> { { 0x1D53B, 0x1D53E }, ::unicode::General_Category::Uppercase_Letter }, // [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Prop<::unicode::General_Category> { { 0x1D53F, 0x1D53F }, ::unicode::General_Category::Unassigned }, // <reserved-1D53F>
    Prop<::unicode::General_Category> { { 0x1D540, 0x1D544 }, ::unicode::General_Category::Uppercase_Letter }, // [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Prop<::unicode::General_Category> { { 0x1D545, 0x1D545 }, ::unicode::General_Category::Unassigned }, // <reserved-1D545>
    Prop<::unicode::General_Category> { { 0x1D546, 0x1D546 }, ::unicode::General_Category::Uppercase_Letter }, // MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Prop<::unicode::General_Category> { { 0x1D547, 0x1D549 }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1D547>..<reserved-1D549>
    Prop<::unicode::General_Category> { { 0x1D54A, 0x1D550 }, ::unicode::General_Category::Uppercase_Letter }, // [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Prop<::unicode::General_Category> { { 0x1D551, 0x1D551 }, ::unicode::General_Category::Unassigned }, // <reserved-1D551>
    Prop<::unicode::General_Category> { { 0x1D552, 0x1D56B }, ::unicode::General_Category::Lowercase_Letter }, // [26] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL DOUBLE-STRUCK SMALL Z
    Prop<::unicode::General_Category> { { 0x1D56C, 0x1D585 }, ::unicode::General_Category::Uppercase_Letter }, // [26] MATHEMATICAL BOLD FRAKTUR CAPITAL A..MATHEMATICAL BOLD FRAKTUR CAPITAL Z
    Prop<::unicode::General_Category> { { 0x1D586, 0x1D59F }, ::unicode::General_Category::Lowercase_Letter }, // [26] MATHEMATICAL BOLD FRAKTUR SMALL A..MATHEMATICAL BOLD FRAKTUR SMALL Z
    Prop<::unicode::General_Category> { { 0x1D5A0, 0x1D5B9 }, ::unicode::General_Category::Uppercase_Letter }, // [26] MATHEMATICAL SANS-SERIF CAPITAL A..MATHEMATICAL SANS-SERIF CAPITAL Z
    Prop<::unicode::General_Category> { { 0x1D5BA, 0x1D5D3 }, ::unicode::General_Category::Lowercase_Letter }, // [26] MATHEMATICAL SANS-SERIF SMALL A..MATHEMATICAL SANS-SERIF SMALL Z
    Prop<::unicode::General_Category> { { 0x1D5D4, 0x1D5ED }, ::unicode::General_Category::Uppercase_Letter }, // [26] MATHEMATICAL SANS-SERIF BOLD CAPITAL A..MATHEMATICAL SANS-SERIF BOLD CAPITAL Z
    Prop<::unicode::General_Category> { { 0x1D5EE, 0x1D607 }, ::unicode::General_Category::Lowercase_Letter }, // [26] MATHEMATICAL SANS-SERIF BOLD SMALL A..MATHEMATICAL SANS-SERIF BOLD SMALL Z
    Prop<::unicode::General_Category> { { 0x1D608, 0x1D621 }, ::unicode::General_Category::Uppercase_Letter }, // [26] MATHEMATICAL SANS-SERIF ITALIC CAPITAL A..MATHEMATICAL SANS-SERIF ITALIC CAPITAL Z
    Prop<::unicode::General_Category> { { 0x1D622, 0x1D63B }, ::unicode::General_Category::Lowercase_Letter }, // [26] MATHEMATICAL SANS-SERIF ITALIC SMALL A..MATHEMATICAL SANS-SERIF ITALIC SMALL Z
    Prop<::unicode::General_Category> { { 0x1D63C, 0x1D655 }, ::unicode::General_Category::Uppercase_Letter }, // [26] MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL A..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL Z
    Prop<::unicode::General_Category> { { 0x1D656, 0x1D66F }, ::unicode::General_Category::Lowercase_Letter }, // [26] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL A..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL Z
    Prop<::unicode::General_Category> { { 0x1D670, 0x1D689 }, ::unicode::General_Category::Uppercase_Letter }, // [26] MATHEMATICAL MONOSPACE CAPITAL A..MATHEMATICAL MONOSPACE CAPITAL Z
    Prop<::unicode::General_Category> { { 0x1D68A, 0x1D6A5 }, ::unicode::General_Category::Lowercase_Letter }, // [28] MATHEMATICAL MONOSPACE SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    Prop<::unicode::General_Category> { { 0x1D6A6, 0x1D6A7 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1D6A6>..<reserved-1D6A7>
    Prop<::unicode::General_Category> { { 0x1D6A8, 0x1D6C0 }, ::unicode::General_Category::Uppercase_Letter }, // [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    Prop<::unicode::General_Category> { { 0x1D6C1, 0x1D6C1 }, ::unicode::General_Category::Math_Symbol }, // MATHEMATICAL BOLD NABLA
    Prop<::unicode::General_Category> { { 0x1D6C2, 0x1D6DA }, ::unicode::General_Category::Lowercase_Letter }, // [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    Prop<::unicode::General_Category> { { 0x1D6DB, 0x1D6DB }, ::unicode::General_Category::Math_Symbol }, // MATHEMATICAL BOLD PARTIAL DIFFERENTIAL
    Prop<::unicode::General_Category> { { 0x1D6DC, 0x1D6E1 }, ::unicode::General_Category::Lowercase_Letter }, // [6] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL BOLD PI SYMBOL
    Prop<::unicode::General_Category> { { 0x1D6E2, 0x1D6FA }, ::unicode::General_Category::Uppercase_Letter }, // [25] MATHEMATICAL ITALIC CAPITAL ALPHA..MATHEMATICAL ITALIC CAPITAL OMEGA
    Prop<::unicode::General_Category> { { 0x1D6FB, 0x1D6FB }, ::unicode::General_Category::Math_Symbol }, // MATHEMATICAL ITALIC NABLA
    Prop<::unicode::General_Category> { { 0x1D6FC, 0x1D714 }, ::unicode::General_Category::Lowercase_Letter }, // [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    Prop<::unicode::General_Category> { { 0x1D715, 0x1D715 }, ::unicode::General_Category::Math_Symbol }, // MATHEMATICAL ITALIC PARTIAL DIFFERENTIAL
    Prop<::unicode::General_Category> { { 0x1D716, 0x1D71B }, ::unicode::General_Category::Lowercase_Letter }, // [6] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL ITALIC PI SYMBOL
    Prop<::unicode::General_Category> { { 0x1D71C, 0x1D734 }, ::unicode::General_Category::Uppercase_Letter }, // [25] MATHEMATICAL BOLD ITALIC CAPITAL ALPHA..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    Prop<::unicode::General_Category> { { 0x1D735, 0x1D735 }, ::unicode::General_Category::Math_Symbol }, // MATHEMATICAL BOLD ITALIC NABLA
    Prop<::unicode::General_Category> { { 0x1D736, 0x1D74E }, ::unicode::General_Category::Lowercase_Letter }, // [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    Prop<::unicode::General_Category> { { 0x1D74F, 0x1D74F }, ::unicode::General_Category::Math_Symbol }, // MATHEMATICAL BOLD ITALIC PARTIAL DIFFERENTIAL
    Prop<::unicode::General_Category> { { 0x1D750, 0x1D755 }, ::unicode::General_Category::Lowercase_Letter }, // [6] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC PI SYMBOL
    Prop<::unicode::General_Category> { { 0x1D756, 0x1D76E }, ::unicode::General_Category::Uppercase_Letter }, // [25] MATHEMATICAL SANS-SERIF BOLD CAPITAL ALPHA..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    Prop<::unicode::General_Category> { { 0x1D76F, 0x1D76F }, ::unicode::General_Category::Math_Symbol }, // MATHEMATICAL SANS-SERIF BOLD NABLA
    Prop<::unicode::General_Category> { { 0x1D770, 0x1D788 }, ::unicode::General_Category::Lowercase_Letter }, // [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    Prop<::unicode::General_Category> { { 0x1D789, 0x1D789 }, ::unicode::General_Category::Math_Symbol }, // MATHEMATICAL SANS-SERIF BOLD PARTIAL DIFFERENTIAL
    Prop<::unicode::General_Category> { { 0x1D78A, 0x1D78F }, ::unicode::General_Category::Lowercase_Letter }, // [6] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD PI SYMBOL
    Prop<::unicode::General_Category> { { 0x1D790, 0x1D7A8 }, ::unicode::General_Category::Uppercase_Letter }, // [25] MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    Prop<::unicode::General_Category> { { 0x1D7A9, 0x1D7A9 }, ::unicode::General_Category::Math_Symbol }, // MATHEMATICAL SANS-SERIF BOLD ITALIC NABLA
    Prop<::unicode::General_Category> { { 0x1D7AA, 0x1D7C2 }, ::unicode::General_Category::Lowercase_Letter }, // [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    Prop<::unicode::General_Category> { { 0x1D7C3, 0x1D7C3 }, ::unicode::General_Category::Math_Symbol }, // MATHEMATICAL SANS-SERIF BOLD ITALIC PARTIAL DIFFERENTIAL
    Prop<::unicode::General_Category> { { 0x1D7C4, 0x1D7C9 }, ::unicode::General_Category::Lowercase_Letter }, // [6] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC PI SYMBOL
    Prop<::unicode::General_Category> { { 0x1D7CA, 0x1D7CA }, ::unicode::General_Category::Uppercase_Letter }, // MATHEMATICAL BOLD CAPITAL DIGAMMA
    Prop<::unicode::General_Category> { { 0x1D7CB, 0x1D7CB }, ::unicode::General_Category::Lowercase_Letter }, // MATHEMATICAL BOLD SMALL DIGAMMA
    Prop<::unicode::General_Category> { { 0x1D7CC, 0x1D7CD }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1D7CC>..<reserved-1D7CD>
    Prop<::unicode::General_Category> { { 0x1D7CE, 0x1D7FF }, ::unicode::General_Category::Decimal_Number }, // [50] MATHEMATICAL BOLD DIGIT ZERO..MATHEMATICAL MONOSPACE DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1D800, 0x1D9FF }, ::unicode::General_Category::Other_Symbol }, // [512] SIGNWRITING HAND-FIST INDEX..SIGNWRITING HEAD
    Prop<::unicode::General_Category> { { 0x1DA00, 0x1DA36 }, ::unicode::General_Category::Nonspacing_Mark }, // [55] SIGNWRITING HEAD RIM..SIGNWRITING AIR SUCKING IN
    Prop<::unicode::General_Category> { { 0x1DA37, 0x1DA3A }, ::unicode::General_Category::Other_Symbol }, // [4] SIGNWRITING AIR BLOW SMALL ROTATIONS..SIGNWRITING BREATH EXHALE
    Prop<::unicode::General_Category> { { 0x1DA3B, 0x1DA6C }, ::unicode::General_Category::Nonspacing_Mark }, // [50] SIGNWRITING MOUTH CLOSED NEUTRAL..SIGNWRITING EXCITEMENT
    Prop<::unicode::General_Category> { { 0x1DA6D, 0x1DA74 }, ::unicode::General_Category::Other_Symbol }, // [8] SIGNWRITING SHOULDER HIP SPINE..SIGNWRITING TORSO-FLOORPLANE TWISTING
    Prop<::unicode::General_Category> { { 0x1DA75, 0x1DA75 }, ::unicode::General_Category::Nonspacing_Mark }, // SIGNWRITING UPPER BODY TILTING FROM HIP JOINTS
    Prop<::unicode::General_Category> { { 0x1DA76, 0x1DA83 }, ::unicode::General_Category::Other_Symbol }, // [14] SIGNWRITING LIMB COMBINATION..SIGNWRITING LOCATION DEPTH
    Prop<::unicode::General_Category> { { 0x1DA84, 0x1DA84 }, ::unicode::General_Category::Nonspacing_Mark }, // SIGNWRITING LOCATION HEAD NECK
    Prop<::unicode::General_Category> { { 0x1DA85, 0x1DA86 }, ::unicode::General_Category::Other_Symbol }, // [2] SIGNWRITING LOCATION TORSO..SIGNWRITING LOCATION LIMBS DIGITS
    Prop<::unicode::General_Category> { { 0x1DA87, 0x1DA8B }, ::unicode::General_Category::Other_Punctuation }, // [5] SIGNWRITING COMMA..SIGNWRITING PARENTHESIS
    Prop<::unicode::General_Category> { { 0x1DA8C, 0x1DA9A }, ::unicode::General_Category::Unassigned }, // [15] <reserved-1DA8C>..<reserved-1DA9A>
    Prop<::unicode::General_Category> { { 0x1DA9B, 0x1DA9F }, ::unicode::General_Category::Nonspacing_Mark }, // [5] SIGNWRITING FILL MODIFIER-2..SIGNWRITING FILL MODIFIER-6
    Prop<::unicode::General_Category> { { 0x1DAA0, 0x1DAA0 }, ::unicode::General_Category::Unassigned }, // <reserved-1DAA0>
    Prop<::unicode::General_Category> { { 0x1DAA1, 0x1DAAF }, ::unicode::General_Category::Nonspacing_Mark }, // [15] SIGNWRITING ROTATION MODIFIER-2..SIGNWRITING ROTATION MODIFIER-16
    Prop<::unicode::General_Category> { { 0x1DAB0, 0x1DEFF }, ::unicode::General_Category::Unassigned }, // [1104] <reserved-1DAB0>..<reserved-1DEFF>
    Prop<::unicode::General_Category> { { 0x1DF00, 0x1DF09 }, ::unicode::General_Category::Lowercase_Letter }, // [10] LATIN SMALL LETTER FENG DIGRAPH WITH TRILL..LATIN SMALL LETTER T WITH HOOK AND RETROFLEX HOOK
    Prop<::unicode::General_Category> { { 0x1DF0A, 0x1DF0A }, ::unicode::General_Category::Other_Letter }, // LATIN LETTER RETROFLEX CLICK WITH RETROFLEX HOOK
    Prop<::unicode::General_Category> { { 0x1DF0B, 0x1DF1E }, ::unicode::General_Category::Lowercase_Letter }, // [20] LATIN SMALL LETTER ESH WITH DOUBLE BAR..LATIN SMALL LETTER S WITH CURL
    Prop<::unicode::General_Category> { { 0x1DF1F, 0x1DF24 }, ::unicode::General_Category::Unassigned }, // [6] <reserved-1DF1F>..<reserved-1DF24>
    Prop<::unicode::General_Category> { { 0x1DF25, 0x1DF2A }, ::unicode::General_Category::Lowercase_Letter }, // [6] LATIN SMALL LETTER D WITH MID-HEIGHT LEFT HOOK..LATIN SMALL LETTER T WITH MID-HEIGHT LEFT HOOK
    Prop<::unicode::General_Category> { { 0x1DF2B, 0x1DFFF }, ::unicode::General_Category::Unassigned }, // [213] <reserved-1DF2B>..<reserved-1DFFF>
    Prop<::unicode::General_Category> { { 0x1E000, 0x1E006 }, ::unicode::General_Category::Nonspacing_Mark }, // [7] COMBINING GLAGOLITIC LETTER AZU..COMBINING GLAGOLITIC LETTER ZHIVETE
    Prop<::unicode::General_Category> { { 0x1E007, 0x1E007 }, ::unicode::General_Category::Unassigned }, // <reserved-1E007>
    Prop<::unicode::General_Category> { { 0x1E008, 0x1E018 }, ::unicode::General_Category::Nonspacing_Mark }, // [17] COMBINING GLAGOLITIC LETTER ZEMLJA..COMBINING GLAGOLITIC LETTER HERU
    Prop<::unicode::General_Category> { { 0x1E019, 0x1E01A }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1E019>..<reserved-1E01A>
    Prop<::unicode::General_Category> { { 0x1E01B, 0x1E021 }, ::unicode::General_Category::Nonspacing_Mark }, // [7] COMBINING GLAGOLITIC LETTER SHTA..COMBINING GLAGOLITIC LETTER YATI
    Prop<::unicode::General_Category> { { 0x1E022, 0x1E022 }, ::unicode::General_Category::Unassigned }, // <reserved-1E022>
    Prop<::unicode::General_Category> { { 0x1E023, 0x1E024 }, ::unicode::General_Category::Nonspacing_Mark }, // [2] COMBINING GLAGOLITIC LETTER YU..COMBINING GLAGOLITIC LETTER SMALL YUS
    Prop<::unicode::General_Category> { { 0x1E025, 0x1E025 }, ::unicode::General_Category::Unassigned }, // <reserved-1E025>
    Prop<::unicode::General_Category> { { 0x1E026, 0x1E02A }, ::unicode::General_Category::Nonspacing_Mark }, // [5] COMBINING GLAGOLITIC LETTER YO..COMBINING GLAGOLITIC LETTER FITA
    Prop<::unicode::General_Category> { { 0x1E02B, 0x1E02F }, ::unicode::General_Category::Unassigned }, // [5] <reserved-1E02B>..<reserved-1E02F>
    Prop<::unicode::General_Category> { { 0x1E030, 0x1E06D }, ::unicode::General_Category::Modifier_Letter }, // [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Prop<::unicode::General_Category> { { 0x1E06E, 0x1E08E }, ::unicode::General_Category::Unassigned }, // [33] <reserved-1E06E>..<reserved-1E08E>
    Prop<::unicode::General_Category> { { 0x1E08F, 0x1E08F }, ::unicode::General_Category::Nonspacing_Mark }, // COMBINING CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
    Prop<::unicode::General_Category> { { 0x1E090, 0x1E0FF }, ::unicode::General_Category::Unassigned }, // [112] <reserved-1E090>..<reserved-1E0FF>
    Prop<::unicode::General_Category> { { 0x1E100, 0x1E12C }, ::unicode::General_Category::Other_Letter }, // [45] NYIAKENG PUACHUE HMONG LETTER MA..NYIAKENG PUACHUE HMONG LETTER W
    Prop<::unicode::General_Category> { { 0x1E12D, 0x1E12F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1E12D>..<reserved-1E12F>
    Prop<::unicode::General_Category> { { 0x1E130, 0x1E136 }, ::unicode::General_Category::Nonspacing_Mark }, // [7] NYIAKENG PUACHUE HMONG TONE-B..NYIAKENG PUACHUE HMONG TONE-D
    Prop<::unicode::General_Category> { { 0x1E137, 0x1E13D }, ::unicode::General_Category::Modifier_Letter }, // [7] NYIAKENG PUACHUE HMONG SIGN FOR PERSON..NYIAKENG PUACHUE HMONG SYLLABLE LENGTHENER
    Prop<::unicode::General_Category> { { 0x1E13E, 0x1E13F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1E13E>..<reserved-1E13F>
    Prop<::unicode::General_Category> { { 0x1E140, 0x1E149 }, ::unicode::General_Category::Decimal_Number }, // [10] NYIAKENG PUACHUE HMONG DIGIT ZERO..NYIAKENG PUACHUE HMONG DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1E14A, 0x1E14D }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1E14A>..<reserved-1E14D>
    Prop<::unicode::General_Category> { { 0x1E14E, 0x1E14E }, ::unicode::General_Category::Other_Letter }, // NYIAKENG PUACHUE HMONG LOGOGRAM NYAJ
    Prop<::unicode::General_Category> { { 0x1E14F, 0x1E14F }, ::unicode::General_Category::Other_Symbol }, // NYIAKENG PUACHUE HMONG CIRCLED CA
    Prop<::unicode::General_Category> { { 0x1E150, 0x1E28F }, ::unicode::General_Category::Unassigned }, // [320] <reserved-1E150>..<reserved-1E28F>
    Prop<::unicode::General_Category> { { 0x1E290, 0x1E2AD }, ::unicode::General_Category::Other_Letter }, // [30] TOTO LETTER PA..TOTO LETTER A
    Prop<::unicode::General_Category> { { 0x1E2AE, 0x1E2AE }, ::unicode::General_Category::Nonspacing_Mark }, // TOTO SIGN RISING TONE
    Prop<::unicode::General_Category> { { 0x1E2AF, 0x1E2BF }, ::unicode::General_Category::Unassigned }, // [17] <reserved-1E2AF>..<reserved-1E2BF>
    Prop<::unicode::General_Category> { { 0x1E2C0, 0x1E2EB }, ::unicode::General_Category::Other_Letter }, // [44] WANCHO LETTER AA..WANCHO LETTER YIH
    Prop<::unicode::General_Category> { { 0x1E2EC, 0x1E2EF }, ::unicode::General_Category::Nonspacing_Mark }, // [4] WANCHO TONE TUP..WANCHO TONE KOINI
    Prop<::unicode::General_Category> { { 0x1E2F0, 0x1E2F9 }, ::unicode::General_Category::Decimal_Number }, // [10] WANCHO DIGIT ZERO..WANCHO DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1E2FA, 0x1E2FE }, ::unicode::General_Category::Unassigned }, // [5] <reserved-1E2FA>..<reserved-1E2FE>
    Prop<::unicode::General_Category> { { 0x1E2FF, 0x1E2FF }, ::unicode::General_Category::Currency_Symbol }, // WANCHO NGUN SIGN
    Prop<::unicode::General_Category> { { 0x1E300, 0x1E4CF }, ::unicode::General_Category::Unassigned }, // [464] <reserved-1E300>..<reserved-1E4CF>
    Prop<::unicode::General_Category> { { 0x1E4D0, 0x1E4EA }, ::unicode::General_Category::Other_Letter }, // [27] NAG MUNDARI LETTER O..NAG MUNDARI LETTER ELL
    Prop<::unicode::General_Category> { { 0x1E4EB, 0x1E4EB }, ::unicode::General_Category::Modifier_Letter }, // NAG MUNDARI SIGN OJOD
    Prop<::unicode::General_Category> { { 0x1E4EC, 0x1E4EF }, ::unicode::General_Category::Nonspacing_Mark }, // [4] NAG MUNDARI SIGN MUHOR..NAG MUNDARI SIGN SUTUH
    Prop<::unicode::General_Category> { { 0x1E4F0, 0x1E4F9 }, ::unicode::General_Category::Decimal_Number }, // [10] NAG MUNDARI DIGIT ZERO..NAG MUNDARI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1E4FA, 0x1E5CF }, ::unicode::General_Category::Unassigned }, // [214] <reserved-1E4FA>..<reserved-1E5CF>
    Prop<::unicode::General_Category> { { 0x1E5D0, 0x1E5ED }, ::unicode::General_Category::Other_Letter }, // [30] OL ONAL LETTER O..OL ONAL LETTER EG
    Prop<::unicode::General_Category> { { 0x1E5EE, 0x1E5EF }, ::unicode::General_Category::Nonspacing_Mark }, // [2] OL ONAL SIGN MU..OL ONAL SIGN IKIR
    Prop<::unicode::General_Category> { { 0x1E5F0, 0x1E5F0 }, ::unicode::General_Category::Other_Letter }, // OL ONAL SIGN HODDOND
    Prop<::unicode::General_Category> { { 0x1E5F1, 0x1E5FA }, ::unicode::General_Category::Decimal_Number }, // [10] OL ONAL DIGIT ZERO..OL ONAL DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1E5FB, 0x1E5FE }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1E5FB>..<reserved-1E5FE>
    Prop<::unicode::General_Category> { { 0x1E5FF, 0x1E5FF }, ::unicode::General_Category::Other_Punctuation }, // OL ONAL ABBREVIATION SIGN
    Prop<::unicode::General_Category> { { 0x1E600, 0x1E6BF }, ::unicode::General_Category::Unassigned }, // [192] <reserved-1E600>..<reserved-1E6BF>
    Prop<::unicode::General_Category> { { 0x1E6C0, 0x1E6DE }, ::unicode::General_Category::Other_Letter }, // [31] TAI YO LETTER LOW KO..TAI YO LETTER HIGH KVO
    Prop<::unicode::General_Category> { { 0x1E6DF, 0x1E6DF }, ::unicode::General_Category::Unassigned }, // <reserved-1E6DF>
    Prop<::unicode::General_Category> { { 0x1E6E0, 0x1E6E2 }, ::unicode::General_Category::Other_Letter }, // [3] TAI YO LETTER AA..TAI YO LETTER UE
    Prop<::unicode::General_Category> { { 0x1E6E3, 0x1E6E3 }, ::unicode::General_Category::Nonspacing_Mark }, // TAI YO SIGN UE
    Prop<::unicode::General_Category> { { 0x1E6E4, 0x1E6E5 }, ::unicode::General_Category::Other_Letter }, // [2] TAI YO LETTER U..TAI YO LETTER AE
    Prop<::unicode::General_Category> { { 0x1E6E6, 0x1E6E6 }, ::unicode::General_Category::Nonspacing_Mark }, // TAI YO SIGN AU
    Prop<::unicode::General_Category> { { 0x1E6E7, 0x1E6ED }, ::unicode::General_Category::Other_Letter }, // [7] TAI YO LETTER O..TAI YO LETTER AUE
    Prop<::unicode::General_Category> { { 0x1E6EE, 0x1E6EF }, ::unicode::General_Category::Nonspacing_Mark }, // [2] TAI YO SIGN AY..TAI YO SIGN ANG
    Prop<::unicode::General_Category> { { 0x1E6F0, 0x1E6F4 }, ::unicode::General_Category::Other_Letter }, // [5] TAI YO LETTER AN..TAI YO LETTER AP
    Prop<::unicode::General_Category> { { 0x1E6F5, 0x1E6F5 }, ::unicode::General_Category::Nonspacing_Mark }, // TAI YO SIGN OM
    Prop<::unicode::General_Category> { { 0x1E6F6, 0x1E6FD }, ::unicode::General_Category::Unassigned }, // [8] <reserved-1E6F6>..<reserved-1E6FD>
    Prop<::unicode::General_Category> { { 0x1E6FE, 0x1E6FE }, ::unicode::General_Category::Other_Letter }, // TAI YO SYMBOL MUEANG
    Prop<::unicode::General_Category> { { 0x1E6FF, 0x1E6FF }, ::unicode::General_Category::Modifier_Letter }, // TAI YO XAM LAI
    Prop<::unicode::General_Category> { { 0x1E700, 0x1E7DF }, ::unicode::General_Category::Unassigned }, // [224] <reserved-1E700>..<reserved-1E7DF>
    Prop<::unicode::General_Category> { { 0x1E7E0, 0x1E7E6 }, ::unicode::General_Category::Other_Letter }, // [7] ETHIOPIC SYLLABLE HHYA..ETHIOPIC SYLLABLE HHYO
    Prop<::unicode::General_Category> { { 0x1E7E7, 0x1E7E7 }, ::unicode::General_Category::Unassigned }, // <reserved-1E7E7>
    Prop<::unicode::General_Category> { { 0x1E7E8, 0x1E7EB }, ::unicode::General_Category::Other_Letter }, // [4] ETHIOPIC SYLLABLE GURAGE HHWA..ETHIOPIC SYLLABLE HHWE
    Prop<::unicode::General_Category> { { 0x1E7EC, 0x1E7EC }, ::unicode::General_Category::Unassigned }, // <reserved-1E7EC>
    Prop<::unicode::General_Category> { { 0x1E7ED, 0x1E7EE }, ::unicode::General_Category::Other_Letter }, // [2] ETHIOPIC SYLLABLE GURAGE MWI..ETHIOPIC SYLLABLE GURAGE MWEE
    Prop<::unicode::General_Category> { { 0x1E7EF, 0x1E7EF }, ::unicode::General_Category::Unassigned }, // <reserved-1E7EF>
    Prop<::unicode::General_Category> { { 0x1E7F0, 0x1E7FE }, ::unicode::General_Category::Other_Letter }, // [15] ETHIOPIC SYLLABLE GURAGE QWI..ETHIOPIC SYLLABLE GURAGE PWEE
    Prop<::unicode::General_Category> { { 0x1E7FF, 0x1E7FF }, ::unicode::General_Category::Unassigned }, // <reserved-1E7FF>
    Prop<::unicode::General_Category> { { 0x1E800, 0x1E8C4 }, ::unicode::General_Category::Other_Letter }, // [197] MENDE KIKAKUI SYLLABLE M001 KI..MENDE KIKAKUI SYLLABLE M060 NYON
    Prop<::unicode::General_Category> { { 0x1E8C5, 0x1E8C6 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1E8C5>..<reserved-1E8C6>
    Prop<::unicode::General_Category> { { 0x1E8C7, 0x1E8CF }, ::unicode::General_Category::Other_Number }, // [9] MENDE KIKAKUI DIGIT ONE..MENDE KIKAKUI DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1E8D0, 0x1E8D6 }, ::unicode::General_Category::Nonspacing_Mark }, // [7] MENDE KIKAKUI COMBINING NUMBER TEENS..MENDE KIKAKUI COMBINING NUMBER MILLIONS
    Prop<::unicode::General_Category> { { 0x1E8D7, 0x1E8FF }, ::unicode::General_Category::Unassigned }, // [41] <reserved-1E8D7>..<reserved-1E8FF>
    Prop<::unicode::General_Category> { { 0x1E900, 0x1E921 }, ::unicode::General_Category::Uppercase_Letter }, // [34] ADLAM CAPITAL LETTER ALIF..ADLAM CAPITAL LETTER SHA
    Prop<::unicode::General_Category> { { 0x1E922, 0x1E943 }, ::unicode::General_Category::Lowercase_Letter }, // [34] ADLAM SMALL LETTER ALIF..ADLAM SMALL LETTER SHA
    Prop<::unicode::General_Category> { { 0x1E944, 0x1E94A }, ::unicode::General_Category::Nonspacing_Mark }, // [7] ADLAM ALIF LENGTHENER..ADLAM NUKTA
    Prop<::unicode::General_Category> { { 0x1E94B, 0x1E94B }, ::unicode::General_Category::Modifier_Letter }, // ADLAM NASALIZATION MARK
    Prop<::unicode::General_Category> { { 0x1E94C, 0x1E94F }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1E94C>..<reserved-1E94F>
    Prop<::unicode::General_Category> { { 0x1E950, 0x1E959 }, ::unicode::General_Category::Decimal_Number }, // [10] ADLAM DIGIT ZERO..ADLAM DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1E95A, 0x1E95D }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1E95A>..<reserved-1E95D>
    Prop<::unicode::General_Category> { { 0x1E95E, 0x1E95F }, ::unicode::General_Category::Other_Punctuation }, // [2] ADLAM INITIAL EXCLAMATION MARK..ADLAM INITIAL QUESTION MARK
    Prop<::unicode::General_Category> { { 0x1E960, 0x1EC70 }, ::unicode::General_Category::Unassigned }, // [785] <reserved-1E960>..<reserved-1EC70>
    Prop<::unicode::General_Category> { { 0x1EC71, 0x1ECAB }, ::unicode::General_Category::Other_Number }, // [59] INDIC SIYAQ NUMBER ONE..INDIC SIYAQ NUMBER PREFIXED NINE
    Prop<::unicode::General_Category> { { 0x1ECAC, 0x1ECAC }, ::unicode::General_Category::Other_Symbol }, // INDIC SIYAQ PLACEHOLDER
    Prop<::unicode::General_Category> { { 0x1ECAD, 0x1ECAF }, ::unicode::General_Category::Other_Number }, // [3] INDIC SIYAQ FRACTION ONE QUARTER..INDIC SIYAQ FRACTION THREE QUARTERS
    Prop<::unicode::General_Category> { { 0x1ECB0, 0x1ECB0 }, ::unicode::General_Category::Currency_Symbol }, // INDIC SIYAQ RUPEE MARK
    Prop<::unicode::General_Category> { { 0x1ECB1, 0x1ECB4 }, ::unicode::General_Category::Other_Number }, // [4] INDIC SIYAQ NUMBER ALTERNATE ONE..INDIC SIYAQ ALTERNATE LAKH MARK
    Prop<::unicode::General_Category> { { 0x1ECB5, 0x1ED00 }, ::unicode::General_Category::Unassigned }, // [76] <reserved-1ECB5>..<reserved-1ED00>
    Prop<::unicode::General_Category> { { 0x1ED01, 0x1ED2D }, ::unicode::General_Category::Other_Number }, // [45] OTTOMAN SIYAQ NUMBER ONE..OTTOMAN SIYAQ NUMBER NINETY THOUSAND
    Prop<::unicode::General_Category> { { 0x1ED2E, 0x1ED2E }, ::unicode::General_Category::Other_Symbol }, // OTTOMAN SIYAQ MARRATAN
    Prop<::unicode::General_Category> { { 0x1ED2F, 0x1ED3D }, ::unicode::General_Category::Other_Number }, // [15] OTTOMAN SIYAQ ALTERNATE NUMBER TWO..OTTOMAN SIYAQ FRACTION ONE SIXTH
    Prop<::unicode::General_Category> { { 0x1ED3E, 0x1EDFF }, ::unicode::General_Category::Unassigned }, // [194] <reserved-1ED3E>..<reserved-1EDFF>
    Prop<::unicode::General_Category> { { 0x1EE00, 0x1EE03 }, ::unicode::General_Category::Other_Letter }, // [4] ARABIC MATHEMATICAL ALEF..ARABIC MATHEMATICAL DAL
    Prop<::unicode::General_Category> { { 0x1EE04, 0x1EE04 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE04>
    Prop<::unicode::General_Category> { { 0x1EE05, 0x1EE1F }, ::unicode::General_Category::Other_Letter }, // [27] ARABIC MATHEMATICAL WAW..ARABIC MATHEMATICAL DOTLESS QAF
    Prop<::unicode::General_Category> { { 0x1EE20, 0x1EE20 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE20>
    Prop<::unicode::General_Category> { { 0x1EE21, 0x1EE22 }, ::unicode::General_Category::Other_Letter }, // [2] ARABIC MATHEMATICAL INITIAL BEH..ARABIC MATHEMATICAL INITIAL JEEM
    Prop<::unicode::General_Category> { { 0x1EE23, 0x1EE23 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE23>
    Prop<::unicode::General_Category> { { 0x1EE24, 0x1EE24 }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL INITIAL HEH
    Prop<::unicode::General_Category> { { 0x1EE25, 0x1EE26 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1EE25>..<reserved-1EE26>
    Prop<::unicode::General_Category> { { 0x1EE27, 0x1EE27 }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL INITIAL HAH
    Prop<::unicode::General_Category> { { 0x1EE28, 0x1EE28 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE28>
    Prop<::unicode::General_Category> { { 0x1EE29, 0x1EE32 }, ::unicode::General_Category::Other_Letter }, // [10] ARABIC MATHEMATICAL INITIAL YEH..ARABIC MATHEMATICAL INITIAL QAF
    Prop<::unicode::General_Category> { { 0x1EE33, 0x1EE33 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE33>
    Prop<::unicode::General_Category> { { 0x1EE34, 0x1EE37 }, ::unicode::General_Category::Other_Letter }, // [4] ARABIC MATHEMATICAL INITIAL SHEEN..ARABIC MATHEMATICAL INITIAL KHAH
    Prop<::unicode::General_Category> { { 0x1EE38, 0x1EE38 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE38>
    Prop<::unicode::General_Category> { { 0x1EE39, 0x1EE39 }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL INITIAL DAD
    Prop<::unicode::General_Category> { { 0x1EE3A, 0x1EE3A }, ::unicode::General_Category::Unassigned }, // <reserved-1EE3A>
    Prop<::unicode::General_Category> { { 0x1EE3B, 0x1EE3B }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL INITIAL GHAIN
    Prop<::unicode::General_Category> { { 0x1EE3C, 0x1EE41 }, ::unicode::General_Category::Unassigned }, // [6] <reserved-1EE3C>..<reserved-1EE41>
    Prop<::unicode::General_Category> { { 0x1EE42, 0x1EE42 }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL TAILED JEEM
    Prop<::unicode::General_Category> { { 0x1EE43, 0x1EE46 }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1EE43>..<reserved-1EE46>
    Prop<::unicode::General_Category> { { 0x1EE47, 0x1EE47 }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL TAILED HAH
    Prop<::unicode::General_Category> { { 0x1EE48, 0x1EE48 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE48>
    Prop<::unicode::General_Category> { { 0x1EE49, 0x1EE49 }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL TAILED YEH
    Prop<::unicode::General_Category> { { 0x1EE4A, 0x1EE4A }, ::unicode::General_Category::Unassigned }, // <reserved-1EE4A>
    Prop<::unicode::General_Category> { { 0x1EE4B, 0x1EE4B }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL TAILED LAM
    Prop<::unicode::General_Category> { { 0x1EE4C, 0x1EE4C }, ::unicode::General_Category::Unassigned }, // <reserved-1EE4C>
    Prop<::unicode::General_Category> { { 0x1EE4D, 0x1EE4F }, ::unicode::General_Category::Other_Letter }, // [3] ARABIC MATHEMATICAL TAILED NOON..ARABIC MATHEMATICAL TAILED AIN
    Prop<::unicode::General_Category> { { 0x1EE50, 0x1EE50 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE50>
    Prop<::unicode::General_Category> { { 0x1EE51, 0x1EE52 }, ::unicode::General_Category::Other_Letter }, // [2] ARABIC MATHEMATICAL TAILED SAD..ARABIC MATHEMATICAL TAILED QAF
    Prop<::unicode::General_Category> { { 0x1EE53, 0x1EE53 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE53>
    Prop<::unicode::General_Category> { { 0x1EE54, 0x1EE54 }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL TAILED SHEEN
    Prop<::unicode::General_Category> { { 0x1EE55, 0x1EE56 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1EE55>..<reserved-1EE56>
    Prop<::unicode::General_Category> { { 0x1EE57, 0x1EE57 }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL TAILED KHAH
    Prop<::unicode::General_Category> { { 0x1EE58, 0x1EE58 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE58>
    Prop<::unicode::General_Category> { { 0x1EE59, 0x1EE59 }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL TAILED DAD
    Prop<::unicode::General_Category> { { 0x1EE5A, 0x1EE5A }, ::unicode::General_Category::Unassigned }, // <reserved-1EE5A>
    Prop<::unicode::General_Category> { { 0x1EE5B, 0x1EE5B }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL TAILED GHAIN
    Prop<::unicode::General_Category> { { 0x1EE5C, 0x1EE5C }, ::unicode::General_Category::Unassigned }, // <reserved-1EE5C>
    Prop<::unicode::General_Category> { { 0x1EE5D, 0x1EE5D }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL TAILED DOTLESS NOON
    Prop<::unicode::General_Category> { { 0x1EE5E, 0x1EE5E }, ::unicode::General_Category::Unassigned }, // <reserved-1EE5E>
    Prop<::unicode::General_Category> { { 0x1EE5F, 0x1EE5F }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL TAILED DOTLESS QAF
    Prop<::unicode::General_Category> { { 0x1EE60, 0x1EE60 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE60>
    Prop<::unicode::General_Category> { { 0x1EE61, 0x1EE62 }, ::unicode::General_Category::Other_Letter }, // [2] ARABIC MATHEMATICAL STRETCHED BEH..ARABIC MATHEMATICAL STRETCHED JEEM
    Prop<::unicode::General_Category> { { 0x1EE63, 0x1EE63 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE63>
    Prop<::unicode::General_Category> { { 0x1EE64, 0x1EE64 }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL STRETCHED HEH
    Prop<::unicode::General_Category> { { 0x1EE65, 0x1EE66 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1EE65>..<reserved-1EE66>
    Prop<::unicode::General_Category> { { 0x1EE67, 0x1EE6A }, ::unicode::General_Category::Other_Letter }, // [4] ARABIC MATHEMATICAL STRETCHED HAH..ARABIC MATHEMATICAL STRETCHED KAF
    Prop<::unicode::General_Category> { { 0x1EE6B, 0x1EE6B }, ::unicode::General_Category::Unassigned }, // <reserved-1EE6B>
    Prop<::unicode::General_Category> { { 0x1EE6C, 0x1EE72 }, ::unicode::General_Category::Other_Letter }, // [7] ARABIC MATHEMATICAL STRETCHED MEEM..ARABIC MATHEMATICAL STRETCHED QAF
    Prop<::unicode::General_Category> { { 0x1EE73, 0x1EE73 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE73>
    Prop<::unicode::General_Category> { { 0x1EE74, 0x1EE77 }, ::unicode::General_Category::Other_Letter }, // [4] ARABIC MATHEMATICAL STRETCHED SHEEN..ARABIC MATHEMATICAL STRETCHED KHAH
    Prop<::unicode::General_Category> { { 0x1EE78, 0x1EE78 }, ::unicode::General_Category::Unassigned }, // <reserved-1EE78>
    Prop<::unicode::General_Category> { { 0x1EE79, 0x1EE7C }, ::unicode::General_Category::Other_Letter }, // [4] ARABIC MATHEMATICAL STRETCHED DAD..ARABIC MATHEMATICAL STRETCHED DOTLESS BEH
    Prop<::unicode::General_Category> { { 0x1EE7D, 0x1EE7D }, ::unicode::General_Category::Unassigned }, // <reserved-1EE7D>
    Prop<::unicode::General_Category> { { 0x1EE7E, 0x1EE7E }, ::unicode::General_Category::Other_Letter }, // ARABIC MATHEMATICAL STRETCHED DOTLESS FEH
    Prop<::unicode::General_Category> { { 0x1EE7F, 0x1EE7F }, ::unicode::General_Category::Unassigned }, // <reserved-1EE7F>
    Prop<::unicode::General_Category> { { 0x1EE80, 0x1EE89 }, ::unicode::General_Category::Other_Letter }, // [10] ARABIC MATHEMATICAL LOOPED ALEF..ARABIC MATHEMATICAL LOOPED YEH
    Prop<::unicode::General_Category> { { 0x1EE8A, 0x1EE8A }, ::unicode::General_Category::Unassigned }, // <reserved-1EE8A>
    Prop<::unicode::General_Category> { { 0x1EE8B, 0x1EE9B }, ::unicode::General_Category::Other_Letter }, // [17] ARABIC MATHEMATICAL LOOPED LAM..ARABIC MATHEMATICAL LOOPED GHAIN
    Prop<::unicode::General_Category> { { 0x1EE9C, 0x1EEA0 }, ::unicode::General_Category::Unassigned }, // [5] <reserved-1EE9C>..<reserved-1EEA0>
    Prop<::unicode::General_Category> { { 0x1EEA1, 0x1EEA3 }, ::unicode::General_Category::Other_Letter }, // [3] ARABIC MATHEMATICAL DOUBLE-STRUCK BEH..ARABIC MATHEMATICAL DOUBLE-STRUCK DAL
    Prop<::unicode::General_Category> { { 0x1EEA4, 0x1EEA4 }, ::unicode::General_Category::Unassigned }, // <reserved-1EEA4>
    Prop<::unicode::General_Category> { { 0x1EEA5, 0x1EEA9 }, ::unicode::General_Category::Other_Letter }, // [5] ARABIC MATHEMATICAL DOUBLE-STRUCK WAW..ARABIC MATHEMATICAL DOUBLE-STRUCK YEH
    Prop<::unicode::General_Category> { { 0x1EEAA, 0x1EEAA }, ::unicode::General_Category::Unassigned }, // <reserved-1EEAA>
    Prop<::unicode::General_Category> { { 0x1EEAB, 0x1EEBB }, ::unicode::General_Category::Other_Letter }, // [17] ARABIC MATHEMATICAL DOUBLE-STRUCK LAM..ARABIC MATHEMATICAL DOUBLE-STRUCK GHAIN
    Prop<::unicode::General_Category> { { 0x1EEBC, 0x1EEEF }, ::unicode::General_Category::Unassigned }, // [52] <reserved-1EEBC>..<reserved-1EEEF>
    Prop<::unicode::General_Category> { { 0x1EEF0, 0x1EEF1 }, ::unicode::General_Category::Math_Symbol }, // [2] ARABIC MATHEMATICAL OPERATOR MEEM WITH HAH WITH TATWEEL..ARABIC MATHEMATICAL OPERATOR HAH WITH DAL
    Prop<::unicode::General_Category> { { 0x1EEF2, 0x1EFFF }, ::unicode::General_Category::Unassigned }, // [270] <reserved-1EEF2>..<reserved-1EFFF>
    Prop<::unicode::General_Category> { { 0x1F000, 0x1F02B }, ::unicode::General_Category::Other_Symbol }, // [44] MAHJONG TILE EAST WIND..MAHJONG TILE BACK
    Prop<::unicode::General_Category> { { 0x1F02C, 0x1F02F }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1F02C>..<reserved-1F02F>
    Prop<::unicode::General_Category> { { 0x1F030, 0x1F093 }, ::unicode::General_Category::Other_Symbol }, // [100] DOMINO TILE HORIZONTAL BACK..DOMINO TILE VERTICAL-06-06
    Prop<::unicode::General_Category> { { 0x1F094, 0x1F09F }, ::unicode::General_Category::Unassigned }, // [12] <reserved-1F094>..<reserved-1F09F>
    Prop<::unicode::General_Category> { { 0x1F0A0, 0x1F0AE }, ::unicode::General_Category::Other_Symbol }, // [15] PLAYING CARD BACK..PLAYING CARD KING OF SPADES
    Prop<::unicode::General_Category> { { 0x1F0AF, 0x1F0B0 }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1F0AF>..<reserved-1F0B0>
    Prop<::unicode::General_Category> { { 0x1F0B1, 0x1F0BF }, ::unicode::General_Category::Other_Symbol }, // [15] PLAYING CARD ACE OF HEARTS..PLAYING CARD RED JOKER
    Prop<::unicode::General_Category> { { 0x1F0C0, 0x1F0C0 }, ::unicode::General_Category::Unassigned }, // <reserved-1F0C0>
    Prop<::unicode::General_Category> { { 0x1F0C1, 0x1F0CF }, ::unicode::General_Category::Other_Symbol }, // [15] PLAYING CARD ACE OF DIAMONDS..PLAYING CARD BLACK JOKER
    Prop<::unicode::General_Category> { { 0x1F0D0, 0x1F0D0 }, ::unicode::General_Category::Unassigned }, // <reserved-1F0D0>
    Prop<::unicode::General_Category> { { 0x1F0D1, 0x1F0F5 }, ::unicode::General_Category::Other_Symbol }, // [37] PLAYING CARD ACE OF CLUBS..PLAYING CARD TRUMP-21
    Prop<::unicode::General_Category> { { 0x1F0F6, 0x1F0FF }, ::unicode::General_Category::Unassigned }, // [10] <reserved-1F0F6>..<reserved-1F0FF>
    Prop<::unicode::General_Category> { { 0x1F100, 0x1F10C }, ::unicode::General_Category::Other_Number }, // [13] DIGIT ZERO FULL STOP..DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT ZERO
    Prop<::unicode::General_Category> { { 0x1F10D, 0x1F1AD }, ::unicode::General_Category::Other_Symbol }, // [161] CIRCLED ZERO WITH SLASH..MASK WORK SYMBOL
    Prop<::unicode::General_Category> { { 0x1F1AE, 0x1F1E5 }, ::unicode::General_Category::Unassigned }, // [56] <reserved-1F1AE>..<reserved-1F1E5>
    Prop<::unicode::General_Category> { { 0x1F1E6, 0x1F202 }, ::unicode::General_Category::Other_Symbol }, // [29] REGIONAL INDICATOR SYMBOL LETTER A..SQUARED KATAKANA SA
    Prop<::unicode::General_Category> { { 0x1F203, 0x1F20F }, ::unicode::General_Category::Unassigned }, // [13] <reserved-1F203>..<reserved-1F20F>
    Prop<::unicode::General_Category> { { 0x1F210, 0x1F23B }, ::unicode::General_Category::Other_Symbol }, // [44] SQUARED CJK UNIFIED IDEOGRAPH-624B..SQUARED CJK UNIFIED IDEOGRAPH-914D
    Prop<::unicode::General_Category> { { 0x1F23C, 0x1F23F }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1F23C>..<reserved-1F23F>
    Prop<::unicode::General_Category> { { 0x1F240, 0x1F248 }, ::unicode::General_Category::Other_Symbol }, // [9] TORTOISE SHELL BRACKETED CJK UNIFIED IDEOGRAPH-672C..TORTOISE SHELL BRACKETED CJK UNIFIED IDEOGRAPH-6557
    Prop<::unicode::General_Category> { { 0x1F249, 0x1F24F }, ::unicode::General_Category::Unassigned }, // [7] <reserved-1F249>..<reserved-1F24F>
    Prop<::unicode::General_Category> { { 0x1F250, 0x1F251 }, ::unicode::General_Category::Other_Symbol }, // [2] CIRCLED IDEOGRAPH ADVANTAGE..CIRCLED IDEOGRAPH ACCEPT
    Prop<::unicode::General_Category> { { 0x1F252, 0x1F25F }, ::unicode::General_Category::Unassigned }, // [14] <reserved-1F252>..<reserved-1F25F>
    Prop<::unicode::General_Category> { { 0x1F260, 0x1F265 }, ::unicode::General_Category::Other_Symbol }, // [6] ROUNDED SYMBOL FOR FU..ROUNDED SYMBOL FOR CAI
    Prop<::unicode::General_Category> { { 0x1F266, 0x1F2FF }, ::unicode::General_Category::Unassigned }, // [154] <reserved-1F266>..<reserved-1F2FF>
    Prop<::unicode::General_Category> { { 0x1F300, 0x1F3FA }, ::unicode::General_Category::Other_Symbol }, // [251] CYCLONE..AMPHORA
    Prop<::unicode::General_Category> { { 0x1F3FB, 0x1F3FF }, ::unicode::General_Category::Modifier_Symbol }, // [5] EMOJI MODIFIER FITZPATRICK TYPE-1-2..EMOJI MODIFIER FITZPATRICK TYPE-6
    Prop<::unicode::General_Category> { { 0x1F400, 0x1F6D8 }, ::unicode::General_Category::Other_Symbol }, // [729] RAT..LANDSLIDE
    Prop<::unicode::General_Category> { { 0x1F6D9, 0x1F6DB }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1F6D9>..<reserved-1F6DB>
    Prop<::unicode::General_Category> { { 0x1F6DC, 0x1F6EC }, ::unicode::General_Category::Other_Symbol }, // [17] WIRELESS..AIRPLANE ARRIVING
    Prop<::unicode::General_Category> { { 0x1F6ED, 0x1F6EF }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1F6ED>..<reserved-1F6EF>
    Prop<::unicode::General_Category> { { 0x1F6F0, 0x1F6FC }, ::unicode::General_Category::Other_Symbol }, // [13] SATELLITE..ROLLER SKATE
    Prop<::unicode::General_Category> { { 0x1F6FD, 0x1F6FF }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1F6FD>..<reserved-1F6FF>
    Prop<::unicode::General_Category> { { 0x1F700, 0x1F7D9 }, ::unicode::General_Category::Other_Symbol }, // [218] ALCHEMICAL SYMBOL FOR QUINTESSENCE..NINE POINTED WHITE STAR
    Prop<::unicode::General_Category> { { 0x1F7DA, 0x1F7DF }, ::unicode::General_Category::Unassigned }, // [6] <reserved-1F7DA>..<reserved-1F7DF>
    Prop<::unicode::General_Category> { { 0x1F7E0, 0x1F7EB }, ::unicode::General_Category::Other_Symbol }, // [12] LARGE ORANGE CIRCLE..LARGE BROWN SQUARE
    Prop<::unicode::General_Category> { { 0x1F7EC, 0x1F7EF }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1F7EC>..<reserved-1F7EF>
    Prop<::unicode::General_Category> { { 0x1F7F0, 0x1F7F0 }, ::unicode::General_Category::Other_Symbol }, // HEAVY EQUALS SIGN
    Prop<::unicode::General_Category> { { 0x1F7F1, 0x1F7FF }, ::unicode::General_Category::Unassigned }, // [15] <reserved-1F7F1>..<reserved-1F7FF>
    Prop<::unicode::General_Category> { { 0x1F800, 0x1F80B }, ::unicode::General_Category::Other_Symbol }, // [12] LEFTWARDS ARROW WITH SMALL TRIANGLE ARROWHEAD..DOWNWARDS ARROW WITH LARGE TRIANGLE ARROWHEAD
    Prop<::unicode::General_Category> { { 0x1F80C, 0x1F80F }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1F80C>..<reserved-1F80F>
    Prop<::unicode::General_Category> { { 0x1F810, 0x1F847 }, ::unicode::General_Category::Other_Symbol }, // [56] LEFTWARDS ARROW WITH SMALL EQUILATERAL ARROWHEAD..DOWNWARDS HEAVY ARROW
    Prop<::unicode::General_Category> { { 0x1F848, 0x1F84F }, ::unicode::General_Category::Unassigned }, // [8] <reserved-1F848>..<reserved-1F84F>
    Prop<::unicode::General_Category> { { 0x1F850, 0x1F859 }, ::unicode::General_Category::Other_Symbol }, // [10] LEFTWARDS SANS-SERIF ARROW..UP DOWN SANS-SERIF ARROW
    Prop<::unicode::General_Category> { { 0x1F85A, 0x1F85F }, ::unicode::General_Category::Unassigned }, // [6] <reserved-1F85A>..<reserved-1F85F>
    Prop<::unicode::General_Category> { { 0x1F860, 0x1F887 }, ::unicode::General_Category::Other_Symbol }, // [40] WIDE-HEADED LEFTWARDS LIGHT BARB ARROW..WIDE-HEADED SOUTH WEST VERY HEAVY BARB ARROW
    Prop<::unicode::General_Category> { { 0x1F888, 0x1F88F }, ::unicode::General_Category::Unassigned }, // [8] <reserved-1F888>..<reserved-1F88F>
    Prop<::unicode::General_Category> { { 0x1F890, 0x1F8AD }, ::unicode::General_Category::Other_Symbol }, // [30] LEFTWARDS TRIANGLE ARROWHEAD..WHITE ARROW SHAFT WIDTH TWO THIRDS
    Prop<::unicode::General_Category> { { 0x1F8AE, 0x1F8AF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1F8AE>..<reserved-1F8AF>
    Prop<::unicode::General_Category> { { 0x1F8B0, 0x1F8BB }, ::unicode::General_Category::Other_Symbol }, // [12] ARROW POINTING UPWARDS THEN NORTH WEST..SOUTH WEST ARROW FROM BAR
    Prop<::unicode::General_Category> { { 0x1F8BC, 0x1F8BF }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1F8BC>..<reserved-1F8BF>
    Prop<::unicode::General_Category> { { 0x1F8C0, 0x1F8C1 }, ::unicode::General_Category::Other_Symbol }, // [2] LEFTWARDS ARROW FROM DOWNWARDS ARROW..RIGHTWARDS ARROW FROM DOWNWARDS ARROW
    Prop<::unicode::General_Category> { { 0x1F8C2, 0x1F8CF }, ::unicode::General_Category::Unassigned }, // [14] <reserved-1F8C2>..<reserved-1F8CF>
    Prop<::unicode::General_Category> { { 0x1F8D0, 0x1F8D8 }, ::unicode::General_Category::Math_Symbol }, // [9] LONG RIGHTWARDS ARROW OVER LONG LEFTWARDS ARROW..LONG LEFT RIGHT ARROW WITH DEPENDENT LOBE
    Prop<::unicode::General_Category> { { 0x1F8D9, 0x1F8FF }, ::unicode::General_Category::Unassigned }, // [39] <reserved-1F8D9>..<reserved-1F8FF>
    Prop<::unicode::General_Category> { { 0x1F900, 0x1FA57 }, ::unicode::General_Category::Other_Symbol }, // [344] CIRCLED CROSS FORMEE WITH FOUR DOTS..BLACK CHESS ALFIL
    Prop<::unicode::General_Category> { { 0x1FA58, 0x1FA5F }, ::unicode::General_Category::Unassigned }, // [8] <reserved-1FA58>..<reserved-1FA5F>
    Prop<::unicode::General_Category> { { 0x1FA60, 0x1FA6D }, ::unicode::General_Category::Other_Symbol }, // [14] XIANGQI RED GENERAL..XIANGQI BLACK SOLDIER
    Prop<::unicode::General_Category> { { 0x1FA6E, 0x1FA6F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1FA6E>..<reserved-1FA6F>
    Prop<::unicode::General_Category> { { 0x1FA70, 0x1FA7C }, ::unicode::General_Category::Other_Symbol }, // [13] BALLET SHOES..CRUTCH
    Prop<::unicode::General_Category> { { 0x1FA7D, 0x1FA7F }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1FA7D>..<reserved-1FA7F>
    Prop<::unicode::General_Category> { { 0x1FA80, 0x1FA8A }, ::unicode::General_Category::Other_Symbol }, // [11] YO-YO..TROMBONE
    Prop<::unicode::General_Category> { { 0x1FA8B, 0x1FA8D }, ::unicode::General_Category::Unassigned }, // [3] <reserved-1FA8B>..<reserved-1FA8D>
    Prop<::unicode::General_Category> { { 0x1FA8E, 0x1FAC6 }, ::unicode::General_Category::Other_Symbol }, // [57] TREASURE CHEST..FINGERPRINT
    Prop<::unicode::General_Category> { { 0x1FAC7, 0x1FAC7 }, ::unicode::General_Category::Unassigned }, // <reserved-1FAC7>
    Prop<::unicode::General_Category> { { 0x1FAC8, 0x1FAC8 }, ::unicode::General_Category::Other_Symbol }, // HAIRY CREATURE
    Prop<::unicode::General_Category> { { 0x1FAC9, 0x1FACC }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1FAC9>..<reserved-1FACC>
    Prop<::unicode::General_Category> { { 0x1FACD, 0x1FADC }, ::unicode::General_Category::Other_Symbol }, // [16] ORCA..ROOT VEGETABLE
    Prop<::unicode::General_Category> { { 0x1FADD, 0x1FADE }, ::unicode::General_Category::Unassigned }, // [2] <reserved-1FADD>..<reserved-1FADE>
    Prop<::unicode::General_Category> { { 0x1FADF, 0x1FAEA }, ::unicode::General_Category::Other_Symbol }, // [12] SPLATTER..DISTORTED FACE
    Prop<::unicode::General_Category> { { 0x1FAEB, 0x1FAEE }, ::unicode::General_Category::Unassigned }, // [4] <reserved-1FAEB>..<reserved-1FAEE>
    Prop<::unicode::General_Category> { { 0x1FAEF, 0x1FAF8 }, ::unicode::General_Category::Other_Symbol }, // [10] FIGHT CLOUD..RIGHTWARDS PUSHING HAND
    Prop<::unicode::General_Category> { { 0x1FAF9, 0x1FAFF }, ::unicode::General_Category::Unassigned }, // [7] <reserved-1FAF9>..<reserved-1FAFF>
    Prop<::unicode::General_Category> { { 0x1FB00, 0x1FB92 }, ::unicode::General_Category::Other_Symbol }, // [147] BLOCK SEXTANT-1..UPPER HALF INVERSE MEDIUM SHADE AND LOWER HALF BLOCK
    Prop<::unicode::General_Category> { { 0x1FB93, 0x1FB93 }, ::unicode::General_Category::Unassigned }, // <reserved-1FB93>
    Prop<::unicode::General_Category> { { 0x1FB94, 0x1FBEF }, ::unicode::General_Category::Other_Symbol }, // [92] LEFT HALF INVERSE MEDIUM SHADE AND RIGHT HALF BLOCK..TOP LEFT JUSTIFIED LOWER RIGHT QUARTER BLACK CIRCLE
    Prop<::unicode::General_Category> { { 0x1FBF0, 0x1FBF9 }, ::unicode::General_Category::Decimal_Number }, // [10] SEGMENTED DIGIT ZERO..SEGMENTED DIGIT NINE
    Prop<::unicode::General_Category> { { 0x1FBFA, 0x1FBFA }, ::unicode::General_Category::Other_Symbol }, // ALARM BELL SYMBOL
    Prop<::unicode::General_Category> { { 0x1FBFB, 0x1FFFF }, ::unicode::General_Category::Unassigned }, // [1029] <reserved-1FBFB>..<noncharacter-1FFFF>
    Prop<::unicode::General_Category> { { 0x20000, 0x2A6DF }, ::unicode::General_Category::Other_Letter }, // [42720] CJK UNIFIED IDEOGRAPH-20000..CJK UNIFIED IDEOGRAPH-2A6DF
    Prop<::unicode::General_Category> { { 0x2A6E0, 0x2A6FF }, ::unicode::General_Category::Unassigned }, // [32] <reserved-2A6E0>..<reserved-2A6FF>
    Prop<::unicode::General_Category> { { 0x2A700, 0x2B81D }, ::unicode::General_Category::Other_Letter }, // [4382] CJK UNIFIED IDEOGRAPH-2A700..CJK UNIFIED IDEOGRAPH-2B81D
    Prop<::unicode::General_Category> { { 0x2B81E, 0x2B81F }, ::unicode::General_Category::Unassigned }, // [2] <reserved-2B81E>..<reserved-2B81F>
    Prop<::unicode::General_Category> { { 0x2B820, 0x2CEAD }, ::unicode::General_Category::Other_Letter }, // [5774] CJK UNIFIED IDEOGRAPH-2B820..CJK UNIFIED IDEOGRAPH-2CEAD
    Prop<::unicode::General_Category> { { 0x2CEAE, 0x2CEAF }, ::unicode::General_Category::Unassigned }, // [2] <reserved-2CEAE>..<reserved-2CEAF>
    Prop<::unicode::General_Category> { { 0x2CEB0, 0x2EBE0 }, ::unicode::General_Category::Other_Letter }, // [7473] CJK UNIFIED IDEOGRAPH-2CEB0..CJK UNIFIED IDEOGRAPH-2EBE0
    Prop<::unicode::General_Category> { { 0x2EBE1, 0x2EBEF }, ::unicode::General_Category::Unassigned }, // [15] <reserved-2EBE1>..<reserved-2EBEF>
    Prop<::unicode::General_Category> { { 0x2EBF0, 0x2EE5D }, ::unicode::General_Category::Other_Letter }, // [622] CJK UNIFIED IDEOGRAPH-2EBF0..CJK UNIFIED IDEOGRAPH-2EE5D
    Prop<::unicode::General_Category> { { 0x2EE5E, 0x2F7FF }, ::unicode::General_Category::Unassigned }, // [2466] <reserved-2EE5E>..<reserved-2F7FF>
    Prop<::unicode::General_Category> { { 0x2F800, 0x2FA1D }, ::unicode::General_Category::Other_Letter }, // [542] CJK COMPATIBILITY IDEOGRAPH-2F800..CJK COMPATIBILITY IDEOGRAPH-2FA1D
    Prop<::unicode::General_Category> { { 0x2FA1E, 0x2FFFF }, ::unicode::General_Category::Unassigned }, // [1506] <reserved-2FA1E>..<noncharacter-2FFFF>
    Prop<::unicode::General_Category> { { 0x30000, 0x3134A }, ::unicode::General_Category::Other_Letter }, // [4939] CJK UNIFIED IDEOGRAPH-30000..CJK UNIFIED IDEOGRAPH-3134A
    Prop<::unicode::General_Category> { { 0x3134B, 0x3134F }, ::unicode::General_Category::Unassigned }, // [5] <reserved-3134B>..<reserved-3134F>
    Prop<::unicode::General_Category> { { 0x31350, 0x33479 }, ::unicode::General_Category::Other_Letter }, // [8490] CJK UNIFIED IDEOGRAPH-31350..CJK UNIFIED IDEOGRAPH-33479
    Prop<::unicode::General_Category> { { 0x3347A, 0xE0000 }, ::unicode::General_Category::Unassigned }, // [707463] <reserved-3347A>..<reserved-E0000>
    Prop<::unicode::General_Category> { { 0xE0001, 0xE0001 }, ::unicode::General_Category::Format }, // LANGUAGE TAG
    Prop<::unicode::General_Category> { { 0xE0002, 0xE001F }, ::unicode::General_Category::Unassigned }, // [30] <reserved-E0002>..<reserved-E001F>
    Prop<::unicode::General_Category> { { 0xE0020, 0xE007F }, ::unicode::General_Category::Format }, // [96] TAG SPACE..CANCEL TAG
    Prop<::unicode::General_Category> { { 0xE0080, 0xE00FF }, ::unicode::General_Category::Unassigned }, // [128] <reserved-E0080>..<reserved-E00FF>
    Prop<::unicode::General_Category> { { 0xE0100, 0xE01EF }, ::unicode::General_Category::Nonspacing_Mark }, // [240] VARIATION SELECTOR-17..VARIATION SELECTOR-256
    Prop<::unicode::General_Category> { { 0xE01F0, 0xEFFFF }, ::unicode::General_Category::Unassigned }, // [65040] <reserved-E01F0>..<noncharacter-EFFFF>
    Prop<::unicode::General_Category> { { 0xF0000, 0xFFFFD }, ::unicode::General_Category::Private_Use }, // [65534] <private-use-F0000>..<private-use-FFFFD>
    Prop<::unicode::General_Category> { { 0xFFFFE, 0xFFFFF }, ::unicode::General_Category::Unassigned }, // [2] <noncharacter-FFFFE>..<noncharacter-FFFFF>
    Prop<::unicode::General_Category> { { 0x100000, 0x10FFFD }, ::unicode::General_Category::Private_Use }, // [65534] <private-use-100000>..<private-use-10FFFD>
    Prop<::unicode::General_Category> { { 0x10FFFE, 0x10FFFF }, ::unicode::General_Category::Unassigned }, // [2] <noncharacter-10FFFE>..<noncharacter-10FFFF>
};
} // end namespace tables

namespace general_category
{
    General_Category get(char32_t value) noexcept {
        return search(tables::General_Category, value).value_or(General_Category::Unspecified);
    }
}

namespace tables {
auto static const Close_Punctuation = std::array<Interval, 76>{ // {{{
    Interval{ 0x0029, 0x0029 }, // RIGHT PARENTHESIS
    Interval{ 0x005D, 0x005D }, // RIGHT SQUARE BRACKET
    Interval{ 0x007D, 0x007D }, // RIGHT CURLY BRACKET
    Interval{ 0x0F3B, 0x0F3B }, // TIBETAN MARK GUG RTAGS GYAS
    Interval{ 0x0F3D, 0x0F3D }, // TIBETAN MARK ANG KHANG GYAS
    Interval{ 0x169C, 0x169C }, // OGHAM REVERSED FEATHER MARK
    Interval{ 0x2046, 0x2046 }, // RIGHT SQUARE BRACKET WITH QUILL
    Interval{ 0x207E, 0x207E }, // SUPERSCRIPT RIGHT PARENTHESIS
    Interval{ 0x208E, 0x208E }, // SUBSCRIPT RIGHT PARENTHESIS
    Interval{ 0x2309, 0x2309 }, // RIGHT CEILING
    Interval{ 0x230B, 0x230B }, // RIGHT FLOOR
    Interval{ 0x232A, 0x232A }, // RIGHT-POINTING ANGLE BRACKET
    Interval{ 0x2769, 0x2769 }, // MEDIUM RIGHT PARENTHESIS ORNAMENT
    Interval{ 0x276B, 0x276B }, // MEDIUM FLATTENED RIGHT PARENTHESIS ORNAMENT
    Interval{ 0x276D, 0x276D }, // MEDIUM RIGHT-POINTING ANGLE BRACKET ORNAMENT
    Interval{ 0x276F, 0x276F }, // HEAVY RIGHT-POINTING ANGLE QUOTATION MARK ORNAMENT
    Interval{ 0x2771, 0x2771 }, // HEAVY RIGHT-POINTING ANGLE BRACKET ORNAMENT
    Interval{ 0x2773, 0x2773 }, // LIGHT RIGHT TORTOISE SHELL BRACKET ORNAMENT
    Interval{ 0x2775, 0x2775 }, // MEDIUM RIGHT CURLY BRACKET ORNAMENT
    Interval{ 0x27C6, 0x27C6 }, // RIGHT S-SHAPED BAG DELIMITER
    Interval{ 0x27E7, 0x27E7 }, // MATHEMATICAL RIGHT WHITE SQUARE BRACKET
    Interval{ 0x27E9, 0x27E9 }, // MATHEMATICAL RIGHT ANGLE BRACKET
    Interval{ 0x27EB, 0x27EB }, // MATHEMATICAL RIGHT DOUBLE ANGLE BRACKET
    Interval{ 0x27ED, 0x27ED }, // MATHEMATICAL RIGHT WHITE TORTOISE SHELL BRACKET
    Interval{ 0x27EF, 0x27EF }, // MATHEMATICAL RIGHT FLATTENED PARENTHESIS
    Interval{ 0x2984, 0x2984 }, // RIGHT WHITE CURLY BRACKET
    Interval{ 0x2986, 0x2986 }, // RIGHT WHITE PARENTHESIS
    Interval{ 0x2988, 0x2988 }, // Z NOTATION RIGHT IMAGE BRACKET
    Interval{ 0x298A, 0x298A }, // Z NOTATION RIGHT BINDING BRACKET
    Interval{ 0x298C, 0x298C }, // RIGHT SQUARE BRACKET WITH UNDERBAR
    Interval{ 0x298E, 0x298E }, // RIGHT SQUARE BRACKET WITH TICK IN BOTTOM CORNER
    Interval{ 0x2990, 0x2990 }, // RIGHT SQUARE BRACKET WITH TICK IN TOP CORNER
    Interval{ 0x2992, 0x2992 }, // RIGHT ANGLE BRACKET WITH DOT
    Interval{ 0x2994, 0x2994 }, // RIGHT ARC GREATER-THAN BRACKET
    Interval{ 0x2996, 0x2996 }, // DOUBLE RIGHT ARC LESS-THAN BRACKET
    Interval{ 0x2998, 0x2998 }, // RIGHT BLACK TORTOISE SHELL BRACKET
    Interval{ 0x29D9, 0x29D9 }, // RIGHT WIGGLY FENCE
    Interval{ 0x29DB, 0x29DB }, // RIGHT DOUBLE WIGGLY FENCE
    Interval{ 0x29FD, 0x29FD }, // RIGHT-POINTING CURVED ANGLE BRACKET
    Interval{ 0x2E23, 0x2E23 }, // TOP RIGHT HALF BRACKET
    Interval{ 0x2E25, 0x2E25 }, // BOTTOM RIGHT HALF BRACKET
    Interval{ 0x2E27, 0x2E27 }, // RIGHT SIDEWAYS U BRACKET
    Interval{ 0x2E29, 0x2E29 }, // RIGHT DOUBLE PARENTHESIS
    Interval{ 0x2E56, 0x2E56 }, // RIGHT SQUARE BRACKET WITH STROKE
    Interval{ 0x2E58, 0x2E58 }, // RIGHT SQUARE BRACKET WITH DOUBLE STROKE
    Interval{ 0x2E5A, 0x2E5A }, // TOP HALF RIGHT PARENTHESIS
    Interval{ 0x2E5C, 0x2E5C }, // BOTTOM HALF RIGHT PARENTHESIS
    Interval{ 0x3009, 0x3009 }, // RIGHT ANGLE BRACKET
    Interval{ 0x300B, 0x300B }, // RIGHT DOUBLE ANGLE BRACKET
    Interval{ 0x300D, 0x300D }, // RIGHT CORNER BRACKET
    Interval{ 0x300F, 0x300F }, // RIGHT WHITE CORNER BRACKET
    Interval{ 0x3011, 0x3011 }, // RIGHT BLACK LENTICULAR BRACKET
    Interval{ 0x3015, 0x3015 }, // RIGHT TORTOISE SHELL BRACKET
    Interval{ 0x3017, 0x3017 }, // RIGHT WHITE LENTICULAR BRACKET
    Interval{ 0x3019, 0x3019 }, // RIGHT WHITE TORTOISE SHELL BRACKET
    Interval{ 0x301B, 0x301B }, // RIGHT WHITE SQUARE BRACKET
    Interval{ 0x301E, 0x301F }, // [2] DOUBLE PRIME QUOTATION MARK..LOW DOUBLE PRIME QUOTATION MARK
    Interval{ 0xFD3E, 0xFD3E }, // ORNATE LEFT PARENTHESIS
    Interval{ 0xFE18, 0xFE18 }, // PRESENTATION FORM FOR VERTICAL RIGHT WHITE LENTICULAR BRAKCET
    Interval{ 0xFE36, 0xFE36 }, // PRESENTATION FORM FOR VERTICAL RIGHT PARENTHESIS
    Interval{ 0xFE38, 0xFE38 }, // PRESENTATION FORM FOR VERTICAL RIGHT CURLY BRACKET
    Interval{ 0xFE3A, 0xFE3A }, // PRESENTATION FORM FOR VERTICAL RIGHT TORTOISE SHELL BRACKET
    Interval{ 0xFE3C, 0xFE3C }, // PRESENTATION FORM FOR VERTICAL RIGHT BLACK LENTICULAR BRACKET
    Interval{ 0xFE3E, 0xFE3E }, // PRESENTATION FORM FOR VERTICAL RIGHT DOUBLE ANGLE BRACKET
    Interval{ 0xFE40, 0xFE40 }, // PRESENTATION FORM FOR VERTICAL RIGHT ANGLE BRACKET
    Interval{ 0xFE42, 0xFE42 }, // PRESENTATION FORM FOR VERTICAL RIGHT CORNER BRACKET
    Interval{ 0xFE44, 0xFE44 }, // PRESENTATION FORM FOR VERTICAL RIGHT WHITE CORNER BRACKET
    Interval{ 0xFE48, 0xFE48 }, // PRESENTATION FORM FOR VERTICAL RIGHT SQUARE BRACKET
    Interval{ 0xFE5A, 0xFE5A }, // SMALL RIGHT PARENTHESIS
    Interval{ 0xFE5C, 0xFE5C }, // SMALL RIGHT CURLY BRACKET
    Interval{ 0xFE5E, 0xFE5E }, // SMALL RIGHT TORTOISE SHELL BRACKET
    Interval{ 0xFF09, 0xFF09 }, // FULLWIDTH RIGHT PARENTHESIS
    Interval{ 0xFF3D, 0xFF3D }, // FULLWIDTH RIGHT SQUARE BRACKET
    Interval{ 0xFF5D, 0xFF5D }, // FULLWIDTH RIGHT CURLY BRACKET
    Interval{ 0xFF60, 0xFF60 }, // FULLWIDTH RIGHT WHITE PARENTHESIS
    Interval{ 0xFF63, 0xFF63 }, // HALFWIDTH RIGHT CORNER BRACKET
}; // }}}
auto static const Connector_Punctuation = std::array<Interval, 6>{ // {{{
    Interval{ 0x005F, 0x005F }, // LOW LINE
    Interval{ 0x203F, 0x2040 }, // [2] UNDERTIE..CHARACTER TIE
    Interval{ 0x2054, 0x2054 }, // INVERTED UNDERTIE
    Interval{ 0xFE33, 0xFE34 }, // [2] PRESENTATION FORM FOR VERTICAL LOW LINE..PRESENTATION FORM FOR VERTICAL WAVY LOW LINE
    Interval{ 0xFE4D, 0xFE4F }, // [3] DASHED LOW LINE..WAVY LOW LINE
    Interval{ 0xFF3F, 0xFF3F }, // FULLWIDTH LOW LINE
}; // }}}
auto static const Control = std::array<Interval, 2>{ // {{{
    Interval{ 0x0000, 0x001F }, // [32] <control-0000>..<control-001F>
    Interval{ 0x007F, 0x009F }, // [33] <control-007F>..<control-009F>
}; // }}}
auto static const Currency_Symbol = std::array<Interval, 21>{ // {{{
    Interval{ 0x0024, 0x0024 }, // DOLLAR SIGN
    Interval{ 0x00A2, 0x00A5 }, // [4] CENT SIGN..YEN SIGN
    Interval{ 0x058F, 0x058F }, // ARMENIAN DRAM SIGN
    Interval{ 0x060B, 0x060B }, // AFGHANI SIGN
    Interval{ 0x07FE, 0x07FF }, // [2] NKO DOROME SIGN..NKO TAMAN SIGN
    Interval{ 0x09F2, 0x09F3 }, // [2] BENGALI RUPEE MARK..BENGALI RUPEE SIGN
    Interval{ 0x09FB, 0x09FB }, // BENGALI GANDA MARK
    Interval{ 0x0AF1, 0x0AF1 }, // GUJARATI RUPEE SIGN
    Interval{ 0x0BF9, 0x0BF9 }, // TAMIL RUPEE SIGN
    Interval{ 0x0E3F, 0x0E3F }, // THAI CURRENCY SYMBOL BAHT
    Interval{ 0x17DB, 0x17DB }, // KHMER CURRENCY SYMBOL RIEL
    Interval{ 0x20A0, 0x20C1 }, // [34] EURO-CURRENCY SIGN..SAUDI RIYAL SIGN
    Interval{ 0xA838, 0xA838 }, // NORTH INDIC RUPEE MARK
    Interval{ 0xFDFC, 0xFDFC }, // RIAL SIGN
    Interval{ 0xFE69, 0xFE69 }, // SMALL DOLLAR SIGN
    Interval{ 0xFF04, 0xFF04 }, // FULLWIDTH DOLLAR SIGN
    Interval{ 0xFFE0, 0xFFE1 }, // [2] FULLWIDTH CENT SIGN..FULLWIDTH POUND SIGN
    Interval{ 0xFFE5, 0xFFE6 }, // [2] FULLWIDTH YEN SIGN..FULLWIDTH WON SIGN
    Interval{ 0x11FDD, 0x11FE0 }, // [4] TAMIL SIGN KAACU..TAMIL SIGN VARAAKAN
    Interval{ 0x1E2FF, 0x1E2FF }, // WANCHO NGUN SIGN
    Interval{ 0x1ECB0, 0x1ECB0 }, // INDIC SIYAQ RUPEE MARK
}; // }}}
auto static const Dash_Punctuation = std::array<Interval, 20>{ // {{{
    Interval{ 0x002D, 0x002D }, // HYPHEN-MINUS
    Interval{ 0x058A, 0x058A }, // ARMENIAN HYPHEN
    Interval{ 0x05BE, 0x05BE }, // HEBREW PUNCTUATION MAQAF
    Interval{ 0x1400, 0x1400 }, // CANADIAN SYLLABICS HYPHEN
    Interval{ 0x1806, 0x1806 }, // MONGOLIAN TODO SOFT HYPHEN
    Interval{ 0x2010, 0x2015 }, // [6] HYPHEN..HORIZONTAL BAR
    Interval{ 0x2E17, 0x2E17 }, // DOUBLE OBLIQUE HYPHEN
    Interval{ 0x2E1A, 0x2E1A }, // HYPHEN WITH DIAERESIS
    Interval{ 0x2E3A, 0x2E3B }, // [2] TWO-EM DASH..THREE-EM DASH
    Interval{ 0x2E40, 0x2E40 }, // DOUBLE HYPHEN
    Interval{ 0x2E5D, 0x2E5D }, // OBLIQUE HYPHEN
    Interval{ 0x301C, 0x301C }, // WAVE DASH
    Interval{ 0x3030, 0x3030 }, // WAVY DASH
    Interval{ 0x30A0, 0x30A0 }, // KATAKANA-HIRAGANA DOUBLE HYPHEN
    Interval{ 0xFE31, 0xFE32 }, // [2] PRESENTATION FORM FOR VERTICAL EM DASH..PRESENTATION FORM FOR VERTICAL EN DASH
    Interval{ 0xFE58, 0xFE58 }, // SMALL EM DASH
    Interval{ 0xFE63, 0xFE63 }, // SMALL HYPHEN-MINUS
    Interval{ 0xFF0D, 0xFF0D }, // FULLWIDTH HYPHEN-MINUS
    Interval{ 0x10D6E, 0x10D6E }, // GARAY HYPHEN
    Interval{ 0x10EAD, 0x10EAD }, // YEZIDI HYPHENATION MARK
}; // }}}
auto static const Decimal_Number = std::array<Interval, 72>{ // {{{
    Interval{ 0x0030, 0x0039 }, // [10] DIGIT ZERO..DIGIT NINE
    Interval{ 0x0660, 0x0669 }, // [10] ARABIC-INDIC DIGIT ZERO..ARABIC-INDIC DIGIT NINE
    Interval{ 0x06F0, 0x06F9 }, // [10] EXTENDED ARABIC-INDIC DIGIT ZERO..EXTENDED ARABIC-INDIC DIGIT NINE
    Interval{ 0x07C0, 0x07C9 }, // [10] NKO DIGIT ZERO..NKO DIGIT NINE
    Interval{ 0x0966, 0x096F }, // [10] DEVANAGARI DIGIT ZERO..DEVANAGARI DIGIT NINE
    Interval{ 0x09E6, 0x09EF }, // [10] BENGALI DIGIT ZERO..BENGALI DIGIT NINE
    Interval{ 0x0A66, 0x0A6F }, // [10] GURMUKHI DIGIT ZERO..GURMUKHI DIGIT NINE
    Interval{ 0x0AE6, 0x0AEF }, // [10] GUJARATI DIGIT ZERO..GUJARATI DIGIT NINE
    Interval{ 0x0B66, 0x0B6F }, // [10] ORIYA DIGIT ZERO..ORIYA DIGIT NINE
    Interval{ 0x0BE6, 0x0BEF }, // [10] TAMIL DIGIT ZERO..TAMIL DIGIT NINE
    Interval{ 0x0C66, 0x0C6F }, // [10] TELUGU DIGIT ZERO..TELUGU DIGIT NINE
    Interval{ 0x0CE6, 0x0CEF }, // [10] KANNADA DIGIT ZERO..KANNADA DIGIT NINE
    Interval{ 0x0D66, 0x0D6F }, // [10] MALAYALAM DIGIT ZERO..MALAYALAM DIGIT NINE
    Interval{ 0x0DE6, 0x0DEF }, // [10] SINHALA LITH DIGIT ZERO..SINHALA LITH DIGIT NINE
    Interval{ 0x0E50, 0x0E59 }, // [10] THAI DIGIT ZERO..THAI DIGIT NINE
    Interval{ 0x0ED0, 0x0ED9 }, // [10] LAO DIGIT ZERO..LAO DIGIT NINE
    Interval{ 0x0F20, 0x0F29 }, // [10] TIBETAN DIGIT ZERO..TIBETAN DIGIT NINE
    Interval{ 0x1040, 0x1049 }, // [10] MYANMAR DIGIT ZERO..MYANMAR DIGIT NINE
    Interval{ 0x1090, 0x1099 }, // [10] MYANMAR SHAN DIGIT ZERO..MYANMAR SHAN DIGIT NINE
    Interval{ 0x17E0, 0x17E9 }, // [10] KHMER DIGIT ZERO..KHMER DIGIT NINE
    Interval{ 0x1810, 0x1819 }, // [10] MONGOLIAN DIGIT ZERO..MONGOLIAN DIGIT NINE
    Interval{ 0x1946, 0x194F }, // [10] LIMBU DIGIT ZERO..LIMBU DIGIT NINE
    Interval{ 0x19D0, 0x19D9 }, // [10] NEW TAI LUE DIGIT ZERO..NEW TAI LUE DIGIT NINE
    Interval{ 0x1A80, 0x1A89 }, // [10] TAI THAM HORA DIGIT ZERO..TAI THAM HORA DIGIT NINE
    Interval{ 0x1A90, 0x1A99 }, // [10] TAI THAM THAM DIGIT ZERO..TAI THAM THAM DIGIT NINE
    Interval{ 0x1B50, 0x1B59 }, // [10] BALINESE DIGIT ZERO..BALINESE DIGIT NINE
    Interval{ 0x1BB0, 0x1BB9 }, // [10] SUNDANESE DIGIT ZERO..SUNDANESE DIGIT NINE
    Interval{ 0x1C40, 0x1C49 }, // [10] LEPCHA DIGIT ZERO..LEPCHA DIGIT NINE
    Interval{ 0x1C50, 0x1C59 }, // [10] OL CHIKI DIGIT ZERO..OL CHIKI DIGIT NINE
    Interval{ 0xA620, 0xA629 }, // [10] VAI DIGIT ZERO..VAI DIGIT NINE
    Interval{ 0xA8D0, 0xA8D9 }, // [10] SAURASHTRA DIGIT ZERO..SAURASHTRA DIGIT NINE
    Interval{ 0xA900, 0xA909 }, // [10] KAYAH LI DIGIT ZERO..KAYAH LI DIGIT NINE
    Interval{ 0xA9D0, 0xA9D9 }, // [10] JAVANESE DIGIT ZERO..JAVANESE DIGIT NINE
    Interval{ 0xA9F0, 0xA9F9 }, // [10] MYANMAR TAI LAING DIGIT ZERO..MYANMAR TAI LAING DIGIT NINE
    Interval{ 0xAA50, 0xAA59 }, // [10] CHAM DIGIT ZERO..CHAM DIGIT NINE
    Interval{ 0xABF0, 0xABF9 }, // [10] MEETEI MAYEK DIGIT ZERO..MEETEI MAYEK DIGIT NINE
    Interval{ 0xFF10, 0xFF19 }, // [10] FULLWIDTH DIGIT ZERO..FULLWIDTH DIGIT NINE
    Interval{ 0x104A0, 0x104A9 }, // [10] OSMANYA DIGIT ZERO..OSMANYA DIGIT NINE
    Interval{ 0x10D30, 0x10D39 }, // [10] HANIFI ROHINGYA DIGIT ZERO..HANIFI ROHINGYA DIGIT NINE
    Interval{ 0x10D40, 0x10D49 }, // [10] GARAY DIGIT ZERO..GARAY DIGIT NINE
    Interval{ 0x11066, 0x1106F }, // [10] BRAHMI DIGIT ZERO..BRAHMI DIGIT NINE
    Interval{ 0x110F0, 0x110F9 }, // [10] SORA SOMPENG DIGIT ZERO..SORA SOMPENG DIGIT NINE
    Interval{ 0x11136, 0x1113F }, // [10] CHAKMA DIGIT ZERO..CHAKMA DIGIT NINE
    Interval{ 0x111D0, 0x111D9 }, // [10] SHARADA DIGIT ZERO..SHARADA DIGIT NINE
    Interval{ 0x112F0, 0x112F9 }, // [10] KHUDAWADI DIGIT ZERO..KHUDAWADI DIGIT NINE
    Interval{ 0x11450, 0x11459 }, // [10] NEWA DIGIT ZERO..NEWA DIGIT NINE
    Interval{ 0x114D0, 0x114D9 }, // [10] TIRHUTA DIGIT ZERO..TIRHUTA DIGIT NINE
    Interval{ 0x11650, 0x11659 }, // [10] MODI DIGIT ZERO..MODI DIGIT NINE
    Interval{ 0x116C0, 0x116C9 }, // [10] TAKRI DIGIT ZERO..TAKRI DIGIT NINE
    Interval{ 0x116D0, 0x116E3 }, // [20] MYANMAR PAO DIGIT ZERO..MYANMAR EASTERN PWO KAREN DIGIT NINE
    Interval{ 0x11730, 0x11739 }, // [10] AHOM DIGIT ZERO..AHOM DIGIT NINE
    Interval{ 0x118E0, 0x118E9 }, // [10] WARANG CITI DIGIT ZERO..WARANG CITI DIGIT NINE
    Interval{ 0x11950, 0x11959 }, // [10] DIVES AKURU DIGIT ZERO..DIVES AKURU DIGIT NINE
    Interval{ 0x11BF0, 0x11BF9 }, // [10] SUNUWAR DIGIT ZERO..SUNUWAR DIGIT NINE
    Interval{ 0x11C50, 0x11C59 }, // [10] BHAIKSUKI DIGIT ZERO..BHAIKSUKI DIGIT NINE
    Interval{ 0x11D50, 0x11D59 }, // [10] MASARAM GONDI DIGIT ZERO..MASARAM GONDI DIGIT NINE
    Interval{ 0x11DA0, 0x11DA9 }, // [10] GUNJALA GONDI DIGIT ZERO..GUNJALA GONDI DIGIT NINE
    Interval{ 0x11DE0, 0x11DE9 }, // [10] TOLONG SIKI DIGIT ZERO..TOLONG SIKI DIGIT NINE
    Interval{ 0x11F50, 0x11F59 }, // [10] KAWI DIGIT ZERO..KAWI DIGIT NINE
    Interval{ 0x16130, 0x16139 }, // [10] GURUNG KHEMA DIGIT ZERO..GURUNG KHEMA DIGIT NINE
    Interval{ 0x16A60, 0x16A69 }, // [10] MRO DIGIT ZERO..MRO DIGIT NINE
    Interval{ 0x16AC0, 0x16AC9 }, // [10] TANGSA DIGIT ZERO..TANGSA DIGIT NINE
    Interval{ 0x16B50, 0x16B59 }, // [10] PAHAWH HMONG DIGIT ZERO..PAHAWH HMONG DIGIT NINE
    Interval{ 0x16D70, 0x16D79 }, // [10] KIRAT RAI DIGIT ZERO..KIRAT RAI DIGIT NINE
    Interval{ 0x1CCF0, 0x1CCF9 }, // [10] OUTLINED DIGIT ZERO..OUTLINED DIGIT NINE
    Interval{ 0x1D7CE, 0x1D7FF }, // [50] MATHEMATICAL BOLD DIGIT ZERO..MATHEMATICAL MONOSPACE DIGIT NINE
    Interval{ 0x1E140, 0x1E149 }, // [10] NYIAKENG PUACHUE HMONG DIGIT ZERO..NYIAKENG PUACHUE HMONG DIGIT NINE
    Interval{ 0x1E2F0, 0x1E2F9 }, // [10] WANCHO DIGIT ZERO..WANCHO DIGIT NINE
    Interval{ 0x1E4F0, 0x1E4F9 }, // [10] NAG MUNDARI DIGIT ZERO..NAG MUNDARI DIGIT NINE
    Interval{ 0x1E5F1, 0x1E5FA }, // [10] OL ONAL DIGIT ZERO..OL ONAL DIGIT NINE
    Interval{ 0x1E950, 0x1E959 }, // [10] ADLAM DIGIT ZERO..ADLAM DIGIT NINE
    Interval{ 0x1FBF0, 0x1FBF9 }, // [10] SEGMENTED DIGIT ZERO..SEGMENTED DIGIT NINE
}; // }}}
auto static const Enclosing_Mark = std::array<Interval, 5>{ // {{{
    Interval{ 0x0488, 0x0489 }, // [2] COMBINING CYRILLIC HUNDRED THOUSANDS SIGN..COMBINING CYRILLIC MILLIONS SIGN
    Interval{ 0x1ABE, 0x1ABE }, // COMBINING PARENTHESES OVERLAY
    Interval{ 0x20DD, 0x20E0 }, // [4] COMBINING ENCLOSING CIRCLE..COMBINING ENCLOSING CIRCLE BACKSLASH
    Interval{ 0x20E2, 0x20E4 }, // [3] COMBINING ENCLOSING SCREEN..COMBINING ENCLOSING UPWARD POINTING TRIANGLE
    Interval{ 0xA670, 0xA672 }, // [3] COMBINING CYRILLIC TEN MILLIONS SIGN..COMBINING CYRILLIC THOUSAND MILLIONS SIGN
}; // }}}
auto static const Final_Punctuation = std::array<Interval, 10>{ // {{{
    Interval{ 0x00BB, 0x00BB }, // RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    Interval{ 0x2019, 0x2019 }, // RIGHT SINGLE QUOTATION MARK
    Interval{ 0x201D, 0x201D }, // RIGHT DOUBLE QUOTATION MARK
    Interval{ 0x203A, 0x203A }, // SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
    Interval{ 0x2E03, 0x2E03 }, // RIGHT SUBSTITUTION BRACKET
    Interval{ 0x2E05, 0x2E05 }, // RIGHT DOTTED SUBSTITUTION BRACKET
    Interval{ 0x2E0A, 0x2E0A }, // RIGHT TRANSPOSITION BRACKET
    Interval{ 0x2E0D, 0x2E0D }, // RIGHT RAISED OMISSION BRACKET
    Interval{ 0x2E1D, 0x2E1D }, // RIGHT LOW PARAPHRASE BRACKET
    Interval{ 0x2E21, 0x2E21 }, // RIGHT VERTICAL BAR WITH QUILL
}; // }}}
auto static const Format = std::array<Interval, 21>{ // {{{
    Interval{ 0x00AD, 0x00AD }, // SOFT HYPHEN
    Interval{ 0x0600, 0x0605 }, // [6] ARABIC NUMBER SIGN..ARABIC NUMBER MARK ABOVE
    Interval{ 0x061C, 0x061C }, // ARABIC LETTER MARK
    Interval{ 0x06DD, 0x06DD }, // ARABIC END OF AYAH
    Interval{ 0x070F, 0x070F }, // SYRIAC ABBREVIATION MARK
    Interval{ 0x0890, 0x0891 }, // [2] ARABIC POUND MARK ABOVE..ARABIC PIASTRE MARK ABOVE
    Interval{ 0x08E2, 0x08E2 }, // ARABIC DISPUTED END OF AYAH
    Interval{ 0x180E, 0x180E }, // MONGOLIAN VOWEL SEPARATOR
    Interval{ 0x200B, 0x200F }, // [5] ZERO WIDTH SPACE..RIGHT-TO-LEFT MARK
    Interval{ 0x202A, 0x202E }, // [5] LEFT-TO-RIGHT EMBEDDING..RIGHT-TO-LEFT OVERRIDE
    Interval{ 0x2060, 0x2064 }, // [5] WORD JOINER..INVISIBLE PLUS
    Interval{ 0x2066, 0x206F }, // [10] LEFT-TO-RIGHT ISOLATE..NOMINAL DIGIT SHAPES
    Interval{ 0xFEFF, 0xFEFF }, // ZERO WIDTH NO-BREAK SPACE
    Interval{ 0xFFF9, 0xFFFB }, // [3] INTERLINEAR ANNOTATION ANCHOR..INTERLINEAR ANNOTATION TERMINATOR
    Interval{ 0x110BD, 0x110BD }, // KAITHI NUMBER SIGN
    Interval{ 0x110CD, 0x110CD }, // KAITHI NUMBER SIGN ABOVE
    Interval{ 0x13430, 0x1343F }, // [16] EGYPTIAN HIEROGLYPH VERTICAL JOINER..EGYPTIAN HIEROGLYPH END WALLED ENCLOSURE
    Interval{ 0x1BCA0, 0x1BCA3 }, // [4] SHORTHAND FORMAT LETTER OVERLAP..SHORTHAND FORMAT UP STEP
    Interval{ 0x1D173, 0x1D17A }, // [8] MUSICAL SYMBOL BEGIN BEAM..MUSICAL SYMBOL END PHRASE
    Interval{ 0xE0001, 0xE0001 }, // LANGUAGE TAG
    Interval{ 0xE0020, 0xE007F }, // [96] TAG SPACE..CANCEL TAG
}; // }}}
auto static const Initial_Punctuation = std::array<Interval, 11>{ // {{{
    Interval{ 0x00AB, 0x00AB }, // LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    Interval{ 0x2018, 0x2018 }, // LEFT SINGLE QUOTATION MARK
    Interval{ 0x201B, 0x201C }, // [2] SINGLE HIGH-REVERSED-9 QUOTATION MARK..LEFT DOUBLE QUOTATION MARK
    Interval{ 0x201F, 0x201F }, // DOUBLE HIGH-REVERSED-9 QUOTATION MARK
    Interval{ 0x2039, 0x2039 }, // SINGLE LEFT-POINTING ANGLE QUOTATION MARK
    Interval{ 0x2E02, 0x2E02 }, // LEFT SUBSTITUTION BRACKET
    Interval{ 0x2E04, 0x2E04 }, // LEFT DOTTED SUBSTITUTION BRACKET
    Interval{ 0x2E09, 0x2E09 }, // LEFT TRANSPOSITION BRACKET
    Interval{ 0x2E0C, 0x2E0C }, // LEFT RAISED OMISSION BRACKET
    Interval{ 0x2E1C, 0x2E1C }, // LEFT LOW PARAPHRASE BRACKET
    Interval{ 0x2E20, 0x2E20 }, // LEFT VERTICAL BAR WITH QUILL
}; // }}}
auto static const Letter_Number = std::array<Interval, 13>{ // {{{
    Interval{ 0x16EE, 0x16F0 }, // [3] RUNIC ARLAUG SYMBOL..RUNIC BELGTHOR SYMBOL
    Interval{ 0x2160, 0x2182 }, // [35] ROMAN NUMERAL ONE..ROMAN NUMERAL TEN THOUSAND
    Interval{ 0x2185, 0x2188 }, // [4] ROMAN NUMERAL SIX LATE FORM..ROMAN NUMERAL ONE HUNDRED THOUSAND
    Interval{ 0x3007, 0x3007 }, // IDEOGRAPHIC NUMBER ZERO
    Interval{ 0x3021, 0x3029 }, // [9] HANGZHOU NUMERAL ONE..HANGZHOU NUMERAL NINE
    Interval{ 0x3038, 0x303A }, // [3] HANGZHOU NUMERAL TEN..HANGZHOU NUMERAL THIRTY
    Interval{ 0xA6E6, 0xA6EF }, // [10] BAMUM LETTER MO..BAMUM LETTER KOGHOM
    Interval{ 0x10140, 0x10174 }, // [53] GREEK ACROPHONIC ATTIC ONE QUARTER..GREEK ACROPHONIC STRATIAN FIFTY MNAS
    Interval{ 0x10341, 0x10341 }, // GOTHIC LETTER NINETY
    Interval{ 0x1034A, 0x1034A }, // GOTHIC LETTER NINE HUNDRED
    Interval{ 0x103D1, 0x103D5 }, // [5] OLD PERSIAN NUMBER ONE..OLD PERSIAN NUMBER HUNDRED
    Interval{ 0x12400, 0x1246E }, // [111] CUNEIFORM NUMERIC SIGN TWO ASH..CUNEIFORM NUMERIC SIGN NINE U VARIANT FORM
    Interval{ 0x16FF4, 0x16FF6 }, // [3] YANGQIN SIGN SLOW ONE BEAT..YANGQIN SIGN SLOW TWO BEATS
}; // }}}
auto static const Line_Separator = std::array<Interval, 1>{ // {{{
    Interval{ 0x2028, 0x2028 }, // LINE SEPARATOR
}; // }}}
auto static const Lowercase_Letter = std::array<Interval, 664>{ // {{{
    Interval{ 0x0061, 0x007A }, // [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Interval{ 0x00B5, 0x00B5 }, // MICRO SIGN
    Interval{ 0x00DF, 0x00F6 }, // [24] LATIN SMALL LETTER SHARP S..LATIN SMALL LETTER O WITH DIAERESIS
    Interval{ 0x00F8, 0x00FF }, // [8] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER Y WITH DIAERESIS
    Interval{ 0x0101, 0x0101 }, // LATIN SMALL LETTER A WITH MACRON
    Interval{ 0x0103, 0x0103 }, // LATIN SMALL LETTER A WITH BREVE
    Interval{ 0x0105, 0x0105 }, // LATIN SMALL LETTER A WITH OGONEK
    Interval{ 0x0107, 0x0107 }, // LATIN SMALL LETTER C WITH ACUTE
    Interval{ 0x0109, 0x0109 }, // LATIN SMALL LETTER C WITH CIRCUMFLEX
    Interval{ 0x010B, 0x010B }, // LATIN SMALL LETTER C WITH DOT ABOVE
    Interval{ 0x010D, 0x010D }, // LATIN SMALL LETTER C WITH CARON
    Interval{ 0x010F, 0x010F }, // LATIN SMALL LETTER D WITH CARON
    Interval{ 0x0111, 0x0111 }, // LATIN SMALL LETTER D WITH STROKE
    Interval{ 0x0113, 0x0113 }, // LATIN SMALL LETTER E WITH MACRON
    Interval{ 0x0115, 0x0115 }, // LATIN SMALL LETTER E WITH BREVE
    Interval{ 0x0117, 0x0117 }, // LATIN SMALL LETTER E WITH DOT ABOVE
    Interval{ 0x0119, 0x0119 }, // LATIN SMALL LETTER E WITH OGONEK
    Interval{ 0x011B, 0x011B }, // LATIN SMALL LETTER E WITH CARON
    Interval{ 0x011D, 0x011D }, // LATIN SMALL LETTER G WITH CIRCUMFLEX
    Interval{ 0x011F, 0x011F }, // LATIN SMALL LETTER G WITH BREVE
    Interval{ 0x0121, 0x0121 }, // LATIN SMALL LETTER G WITH DOT ABOVE
    Interval{ 0x0123, 0x0123 }, // LATIN SMALL LETTER G WITH CEDILLA
    Interval{ 0x0125, 0x0125 }, // LATIN SMALL LETTER H WITH CIRCUMFLEX
    Interval{ 0x0127, 0x0127 }, // LATIN SMALL LETTER H WITH STROKE
    Interval{ 0x0129, 0x0129 }, // LATIN SMALL LETTER I WITH TILDE
    Interval{ 0x012B, 0x012B }, // LATIN SMALL LETTER I WITH MACRON
    Interval{ 0x012D, 0x012D }, // LATIN SMALL LETTER I WITH BREVE
    Interval{ 0x012F, 0x012F }, // LATIN SMALL LETTER I WITH OGONEK
    Interval{ 0x0131, 0x0131 }, // LATIN SMALL LETTER DOTLESS I
    Interval{ 0x0133, 0x0133 }, // LATIN SMALL LIGATURE IJ
    Interval{ 0x0135, 0x0135 }, // LATIN SMALL LETTER J WITH CIRCUMFLEX
    Interval{ 0x0137, 0x0138 }, // [2] LATIN SMALL LETTER K WITH CEDILLA..LATIN SMALL LETTER KRA
    Interval{ 0x013A, 0x013A }, // LATIN SMALL LETTER L WITH ACUTE
    Interval{ 0x013C, 0x013C }, // LATIN SMALL LETTER L WITH CEDILLA
    Interval{ 0x013E, 0x013E }, // LATIN SMALL LETTER L WITH CARON
    Interval{ 0x0140, 0x0140 }, // LATIN SMALL LETTER L WITH MIDDLE DOT
    Interval{ 0x0142, 0x0142 }, // LATIN SMALL LETTER L WITH STROKE
    Interval{ 0x0144, 0x0144 }, // LATIN SMALL LETTER N WITH ACUTE
    Interval{ 0x0146, 0x0146 }, // LATIN SMALL LETTER N WITH CEDILLA
    Interval{ 0x0148, 0x0149 }, // [2] LATIN SMALL LETTER N WITH CARON..LATIN SMALL LETTER N PRECEDED BY APOSTROPHE
    Interval{ 0x014B, 0x014B }, // LATIN SMALL LETTER ENG
    Interval{ 0x014D, 0x014D }, // LATIN SMALL LETTER O WITH MACRON
    Interval{ 0x014F, 0x014F }, // LATIN SMALL LETTER O WITH BREVE
    Interval{ 0x0151, 0x0151 }, // LATIN SMALL LETTER O WITH DOUBLE ACUTE
    Interval{ 0x0153, 0x0153 }, // LATIN SMALL LIGATURE OE
    Interval{ 0x0155, 0x0155 }, // LATIN SMALL LETTER R WITH ACUTE
    Interval{ 0x0157, 0x0157 }, // LATIN SMALL LETTER R WITH CEDILLA
    Interval{ 0x0159, 0x0159 }, // LATIN SMALL LETTER R WITH CARON
    Interval{ 0x015B, 0x015B }, // LATIN SMALL LETTER S WITH ACUTE
    Interval{ 0x015D, 0x015D }, // LATIN SMALL LETTER S WITH CIRCUMFLEX
    Interval{ 0x015F, 0x015F }, // LATIN SMALL LETTER S WITH CEDILLA
    Interval{ 0x0161, 0x0161 }, // LATIN SMALL LETTER S WITH CARON
    Interval{ 0x0163, 0x0163 }, // LATIN SMALL LETTER T WITH CEDILLA
    Interval{ 0x0165, 0x0165 }, // LATIN SMALL LETTER T WITH CARON
    Interval{ 0x0167, 0x0167 }, // LATIN SMALL LETTER T WITH STROKE
    Interval{ 0x0169, 0x0169 }, // LATIN SMALL LETTER U WITH TILDE
    Interval{ 0x016B, 0x016B }, // LATIN SMALL LETTER U WITH MACRON
    Interval{ 0x016D, 0x016D }, // LATIN SMALL LETTER U WITH BREVE
    Interval{ 0x016F, 0x016F }, // LATIN SMALL LETTER U WITH RING ABOVE
    Interval{ 0x0171, 0x0171 }, // LATIN SMALL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x0173, 0x0173 }, // LATIN SMALL LETTER U WITH OGONEK
    Interval{ 0x0175, 0x0175 }, // LATIN SMALL LETTER W WITH CIRCUMFLEX
    Interval{ 0x0177, 0x0177 }, // LATIN SMALL LETTER Y WITH CIRCUMFLEX
    Interval{ 0x017A, 0x017A }, // LATIN SMALL LETTER Z WITH ACUTE
    Interval{ 0x017C, 0x017C }, // LATIN SMALL LETTER Z WITH DOT ABOVE
    Interval{ 0x017E, 0x0180 }, // [3] LATIN SMALL LETTER Z WITH CARON..LATIN SMALL LETTER B WITH STROKE
    Interval{ 0x0183, 0x0183 }, // LATIN SMALL LETTER B WITH TOPBAR
    Interval{ 0x0185, 0x0185 }, // LATIN SMALL LETTER TONE SIX
    Interval{ 0x0188, 0x0188 }, // LATIN SMALL LETTER C WITH HOOK
    Interval{ 0x018C, 0x018D }, // [2] LATIN SMALL LETTER D WITH TOPBAR..LATIN SMALL LETTER TURNED DELTA
    Interval{ 0x0192, 0x0192 }, // LATIN SMALL LETTER F WITH HOOK
    Interval{ 0x0195, 0x0195 }, // LATIN SMALL LETTER HV
    Interval{ 0x0199, 0x019B }, // [3] LATIN SMALL LETTER K WITH HOOK..LATIN SMALL LETTER LAMBDA WITH STROKE
    Interval{ 0x019E, 0x019E }, // LATIN SMALL LETTER N WITH LONG RIGHT LEG
    Interval{ 0x01A1, 0x01A1 }, // LATIN SMALL LETTER O WITH HORN
    Interval{ 0x01A3, 0x01A3 }, // LATIN SMALL LETTER OI
    Interval{ 0x01A5, 0x01A5 }, // LATIN SMALL LETTER P WITH HOOK
    Interval{ 0x01A8, 0x01A8 }, // LATIN SMALL LETTER TONE TWO
    Interval{ 0x01AA, 0x01AB }, // [2] LATIN LETTER REVERSED ESH LOOP..LATIN SMALL LETTER T WITH PALATAL HOOK
    Interval{ 0x01AD, 0x01AD }, // LATIN SMALL LETTER T WITH HOOK
    Interval{ 0x01B0, 0x01B0 }, // LATIN SMALL LETTER U WITH HORN
    Interval{ 0x01B4, 0x01B4 }, // LATIN SMALL LETTER Y WITH HOOK
    Interval{ 0x01B6, 0x01B6 }, // LATIN SMALL LETTER Z WITH STROKE
    Interval{ 0x01B9, 0x01BA }, // [2] LATIN SMALL LETTER EZH REVERSED..LATIN SMALL LETTER EZH WITH TAIL
    Interval{ 0x01BD, 0x01BF }, // [3] LATIN SMALL LETTER TONE FIVE..LATIN LETTER WYNN
    Interval{ 0x01C6, 0x01C6 }, // LATIN SMALL LETTER DZ WITH CARON
    Interval{ 0x01C9, 0x01C9 }, // LATIN SMALL LETTER LJ
    Interval{ 0x01CC, 0x01CC }, // LATIN SMALL LETTER NJ
    Interval{ 0x01CE, 0x01CE }, // LATIN SMALL LETTER A WITH CARON
    Interval{ 0x01D0, 0x01D0 }, // LATIN SMALL LETTER I WITH CARON
    Interval{ 0x01D2, 0x01D2 }, // LATIN SMALL LETTER O WITH CARON
    Interval{ 0x01D4, 0x01D4 }, // LATIN SMALL LETTER U WITH CARON
    Interval{ 0x01D6, 0x01D6 }, // LATIN SMALL LETTER U WITH DIAERESIS AND MACRON
    Interval{ 0x01D8, 0x01D8 }, // LATIN SMALL LETTER U WITH DIAERESIS AND ACUTE
    Interval{ 0x01DA, 0x01DA }, // LATIN SMALL LETTER U WITH DIAERESIS AND CARON
    Interval{ 0x01DC, 0x01DD }, // [2] LATIN SMALL LETTER U WITH DIAERESIS AND GRAVE..LATIN SMALL LETTER TURNED E
    Interval{ 0x01DF, 0x01DF }, // LATIN SMALL LETTER A WITH DIAERESIS AND MACRON
    Interval{ 0x01E1, 0x01E1 }, // LATIN SMALL LETTER A WITH DOT ABOVE AND MACRON
    Interval{ 0x01E3, 0x01E3 }, // LATIN SMALL LETTER AE WITH MACRON
    Interval{ 0x01E5, 0x01E5 }, // LATIN SMALL LETTER G WITH STROKE
    Interval{ 0x01E7, 0x01E7 }, // LATIN SMALL LETTER G WITH CARON
    Interval{ 0x01E9, 0x01E9 }, // LATIN SMALL LETTER K WITH CARON
    Interval{ 0x01EB, 0x01EB }, // LATIN SMALL LETTER O WITH OGONEK
    Interval{ 0x01ED, 0x01ED }, // LATIN SMALL LETTER O WITH OGONEK AND MACRON
    Interval{ 0x01EF, 0x01F0 }, // [2] LATIN SMALL LETTER EZH WITH CARON..LATIN SMALL LETTER J WITH CARON
    Interval{ 0x01F3, 0x01F3 }, // LATIN SMALL LETTER DZ
    Interval{ 0x01F5, 0x01F5 }, // LATIN SMALL LETTER G WITH ACUTE
    Interval{ 0x01F9, 0x01F9 }, // LATIN SMALL LETTER N WITH GRAVE
    Interval{ 0x01FB, 0x01FB }, // LATIN SMALL LETTER A WITH RING ABOVE AND ACUTE
    Interval{ 0x01FD, 0x01FD }, // LATIN SMALL LETTER AE WITH ACUTE
    Interval{ 0x01FF, 0x01FF }, // LATIN SMALL LETTER O WITH STROKE AND ACUTE
    Interval{ 0x0201, 0x0201 }, // LATIN SMALL LETTER A WITH DOUBLE GRAVE
    Interval{ 0x0203, 0x0203 }, // LATIN SMALL LETTER A WITH INVERTED BREVE
    Interval{ 0x0205, 0x0205 }, // LATIN SMALL LETTER E WITH DOUBLE GRAVE
    Interval{ 0x0207, 0x0207 }, // LATIN SMALL LETTER E WITH INVERTED BREVE
    Interval{ 0x0209, 0x0209 }, // LATIN SMALL LETTER I WITH DOUBLE GRAVE
    Interval{ 0x020B, 0x020B }, // LATIN SMALL LETTER I WITH INVERTED BREVE
    Interval{ 0x020D, 0x020D }, // LATIN SMALL LETTER O WITH DOUBLE GRAVE
    Interval{ 0x020F, 0x020F }, // LATIN SMALL LETTER O WITH INVERTED BREVE
    Interval{ 0x0211, 0x0211 }, // LATIN SMALL LETTER R WITH DOUBLE GRAVE
    Interval{ 0x0213, 0x0213 }, // LATIN SMALL LETTER R WITH INVERTED BREVE
    Interval{ 0x0215, 0x0215 }, // LATIN SMALL LETTER U WITH DOUBLE GRAVE
    Interval{ 0x0217, 0x0217 }, // LATIN SMALL LETTER U WITH INVERTED BREVE
    Interval{ 0x0219, 0x0219 }, // LATIN SMALL LETTER S WITH COMMA BELOW
    Interval{ 0x021B, 0x021B }, // LATIN SMALL LETTER T WITH COMMA BELOW
    Interval{ 0x021D, 0x021D }, // LATIN SMALL LETTER YOGH
    Interval{ 0x021F, 0x021F }, // LATIN SMALL LETTER H WITH CARON
    Interval{ 0x0221, 0x0221 }, // LATIN SMALL LETTER D WITH CURL
    Interval{ 0x0223, 0x0223 }, // LATIN SMALL LETTER OU
    Interval{ 0x0225, 0x0225 }, // LATIN SMALL LETTER Z WITH HOOK
    Interval{ 0x0227, 0x0227 }, // LATIN SMALL LETTER A WITH DOT ABOVE
    Interval{ 0x0229, 0x0229 }, // LATIN SMALL LETTER E WITH CEDILLA
    Interval{ 0x022B, 0x022B }, // LATIN SMALL LETTER O WITH DIAERESIS AND MACRON
    Interval{ 0x022D, 0x022D }, // LATIN SMALL LETTER O WITH TILDE AND MACRON
    Interval{ 0x022F, 0x022F }, // LATIN SMALL LETTER O WITH DOT ABOVE
    Interval{ 0x0231, 0x0231 }, // LATIN SMALL LETTER O WITH DOT ABOVE AND MACRON
    Interval{ 0x0233, 0x0239 }, // [7] LATIN SMALL LETTER Y WITH MACRON..LATIN SMALL LETTER QP DIGRAPH
    Interval{ 0x023C, 0x023C }, // LATIN SMALL LETTER C WITH STROKE
    Interval{ 0x023F, 0x0240 }, // [2] LATIN SMALL LETTER S WITH SWASH TAIL..LATIN SMALL LETTER Z WITH SWASH TAIL
    Interval{ 0x0242, 0x0242 }, // LATIN SMALL LETTER GLOTTAL STOP
    Interval{ 0x0247, 0x0247 }, // LATIN SMALL LETTER E WITH STROKE
    Interval{ 0x0249, 0x0249 }, // LATIN SMALL LETTER J WITH STROKE
    Interval{ 0x024B, 0x024B }, // LATIN SMALL LETTER Q WITH HOOK TAIL
    Interval{ 0x024D, 0x024D }, // LATIN SMALL LETTER R WITH STROKE
    Interval{ 0x024F, 0x0293 }, // [69] LATIN SMALL LETTER Y WITH STROKE..LATIN SMALL LETTER EZH WITH CURL
    Interval{ 0x0296, 0x02AF }, // [26] LATIN LETTER INVERTED GLOTTAL STOP..LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    Interval{ 0x0371, 0x0371 }, // GREEK SMALL LETTER HETA
    Interval{ 0x0373, 0x0373 }, // GREEK SMALL LETTER ARCHAIC SAMPI
    Interval{ 0x0377, 0x0377 }, // GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037B, 0x037D }, // [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Interval{ 0x0390, 0x0390 }, // GREEK SMALL LETTER IOTA WITH DIALYTIKA AND TONOS
    Interval{ 0x03AC, 0x03CE }, // [35] GREEK SMALL LETTER ALPHA WITH TONOS..GREEK SMALL LETTER OMEGA WITH TONOS
    Interval{ 0x03D0, 0x03D1 }, // [2] GREEK BETA SYMBOL..GREEK THETA SYMBOL
    Interval{ 0x03D5, 0x03D7 }, // [3] GREEK PHI SYMBOL..GREEK KAI SYMBOL
    Interval{ 0x03D9, 0x03D9 }, // GREEK SMALL LETTER ARCHAIC KOPPA
    Interval{ 0x03DB, 0x03DB }, // GREEK SMALL LETTER STIGMA
    Interval{ 0x03DD, 0x03DD }, // GREEK SMALL LETTER DIGAMMA
    Interval{ 0x03DF, 0x03DF }, // GREEK SMALL LETTER KOPPA
    Interval{ 0x03E1, 0x03E1 }, // GREEK SMALL LETTER SAMPI
    Interval{ 0x03E3, 0x03E3 }, // COPTIC SMALL LETTER SHEI
    Interval{ 0x03E5, 0x03E5 }, // COPTIC SMALL LETTER FEI
    Interval{ 0x03E7, 0x03E7 }, // COPTIC SMALL LETTER KHEI
    Interval{ 0x03E9, 0x03E9 }, // COPTIC SMALL LETTER HORI
    Interval{ 0x03EB, 0x03EB }, // COPTIC SMALL LETTER GANGIA
    Interval{ 0x03ED, 0x03ED }, // COPTIC SMALL LETTER SHIMA
    Interval{ 0x03EF, 0x03F3 }, // [5] COPTIC SMALL LETTER DEI..GREEK LETTER YOT
    Interval{ 0x03F5, 0x03F5 }, // GREEK LUNATE EPSILON SYMBOL
    Interval{ 0x03F8, 0x03F8 }, // GREEK SMALL LETTER SHO
    Interval{ 0x03FB, 0x03FC }, // [2] GREEK SMALL LETTER SAN..GREEK RHO WITH STROKE SYMBOL
    Interval{ 0x0430, 0x045F }, // [48] CYRILLIC SMALL LETTER A..CYRILLIC SMALL LETTER DZHE
    Interval{ 0x0461, 0x0461 }, // CYRILLIC SMALL LETTER OMEGA
    Interval{ 0x0463, 0x0463 }, // CYRILLIC SMALL LETTER YAT
    Interval{ 0x0465, 0x0465 }, // CYRILLIC SMALL LETTER IOTIFIED E
    Interval{ 0x0467, 0x0467 }, // CYRILLIC SMALL LETTER LITTLE YUS
    Interval{ 0x0469, 0x0469 }, // CYRILLIC SMALL LETTER IOTIFIED LITTLE YUS
    Interval{ 0x046B, 0x046B }, // CYRILLIC SMALL LETTER BIG YUS
    Interval{ 0x046D, 0x046D }, // CYRILLIC SMALL LETTER IOTIFIED BIG YUS
    Interval{ 0x046F, 0x046F }, // CYRILLIC SMALL LETTER KSI
    Interval{ 0x0471, 0x0471 }, // CYRILLIC SMALL LETTER PSI
    Interval{ 0x0473, 0x0473 }, // CYRILLIC SMALL LETTER FITA
    Interval{ 0x0475, 0x0475 }, // CYRILLIC SMALL LETTER IZHITSA
    Interval{ 0x0477, 0x0477 }, // CYRILLIC SMALL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT
    Interval{ 0x0479, 0x0479 }, // CYRILLIC SMALL LETTER UK
    Interval{ 0x047B, 0x047B }, // CYRILLIC SMALL LETTER ROUND OMEGA
    Interval{ 0x047D, 0x047D }, // CYRILLIC SMALL LETTER OMEGA WITH TITLO
    Interval{ 0x047F, 0x047F }, // CYRILLIC SMALL LETTER OT
    Interval{ 0x0481, 0x0481 }, // CYRILLIC SMALL LETTER KOPPA
    Interval{ 0x048B, 0x048B }, // CYRILLIC SMALL LETTER SHORT I WITH TAIL
    Interval{ 0x048D, 0x048D }, // CYRILLIC SMALL LETTER SEMISOFT SIGN
    Interval{ 0x048F, 0x048F }, // CYRILLIC SMALL LETTER ER WITH TICK
    Interval{ 0x0491, 0x0491 }, // CYRILLIC SMALL LETTER GHE WITH UPTURN
    Interval{ 0x0493, 0x0493 }, // CYRILLIC SMALL LETTER GHE WITH STROKE
    Interval{ 0x0495, 0x0495 }, // CYRILLIC SMALL LETTER GHE WITH MIDDLE HOOK
    Interval{ 0x0497, 0x0497 }, // CYRILLIC SMALL LETTER ZHE WITH DESCENDER
    Interval{ 0x0499, 0x0499 }, // CYRILLIC SMALL LETTER ZE WITH DESCENDER
    Interval{ 0x049B, 0x049B }, // CYRILLIC SMALL LETTER KA WITH DESCENDER
    Interval{ 0x049D, 0x049D }, // CYRILLIC SMALL LETTER KA WITH VERTICAL STROKE
    Interval{ 0x049F, 0x049F }, // CYRILLIC SMALL LETTER KA WITH STROKE
    Interval{ 0x04A1, 0x04A1 }, // CYRILLIC SMALL LETTER BASHKIR KA
    Interval{ 0x04A3, 0x04A3 }, // CYRILLIC SMALL LETTER EN WITH DESCENDER
    Interval{ 0x04A5, 0x04A5 }, // CYRILLIC SMALL LIGATURE EN GHE
    Interval{ 0x04A7, 0x04A7 }, // CYRILLIC SMALL LETTER PE WITH MIDDLE HOOK
    Interval{ 0x04A9, 0x04A9 }, // CYRILLIC SMALL LETTER ABKHASIAN HA
    Interval{ 0x04AB, 0x04AB }, // CYRILLIC SMALL LETTER ES WITH DESCENDER
    Interval{ 0x04AD, 0x04AD }, // CYRILLIC SMALL LETTER TE WITH DESCENDER
    Interval{ 0x04AF, 0x04AF }, // CYRILLIC SMALL LETTER STRAIGHT U
    Interval{ 0x04B1, 0x04B1 }, // CYRILLIC SMALL LETTER STRAIGHT U WITH STROKE
    Interval{ 0x04B3, 0x04B3 }, // CYRILLIC SMALL LETTER HA WITH DESCENDER
    Interval{ 0x04B5, 0x04B5 }, // CYRILLIC SMALL LIGATURE TE TSE
    Interval{ 0x04B7, 0x04B7 }, // CYRILLIC SMALL LETTER CHE WITH DESCENDER
    Interval{ 0x04B9, 0x04B9 }, // CYRILLIC SMALL LETTER CHE WITH VERTICAL STROKE
    Interval{ 0x04BB, 0x04BB }, // CYRILLIC SMALL LETTER SHHA
    Interval{ 0x04BD, 0x04BD }, // CYRILLIC SMALL LETTER ABKHASIAN CHE
    Interval{ 0x04BF, 0x04BF }, // CYRILLIC SMALL LETTER ABKHASIAN CHE WITH DESCENDER
    Interval{ 0x04C2, 0x04C2 }, // CYRILLIC SMALL LETTER ZHE WITH BREVE
    Interval{ 0x04C4, 0x04C4 }, // CYRILLIC SMALL LETTER KA WITH HOOK
    Interval{ 0x04C6, 0x04C6 }, // CYRILLIC SMALL LETTER EL WITH TAIL
    Interval{ 0x04C8, 0x04C8 }, // CYRILLIC SMALL LETTER EN WITH HOOK
    Interval{ 0x04CA, 0x04CA }, // CYRILLIC SMALL LETTER EN WITH TAIL
    Interval{ 0x04CC, 0x04CC }, // CYRILLIC SMALL LETTER KHAKASSIAN CHE
    Interval{ 0x04CE, 0x04CF }, // [2] CYRILLIC SMALL LETTER EM WITH TAIL..CYRILLIC SMALL LETTER PALOCHKA
    Interval{ 0x04D1, 0x04D1 }, // CYRILLIC SMALL LETTER A WITH BREVE
    Interval{ 0x04D3, 0x04D3 }, // CYRILLIC SMALL LETTER A WITH DIAERESIS
    Interval{ 0x04D5, 0x04D5 }, // CYRILLIC SMALL LIGATURE A IE
    Interval{ 0x04D7, 0x04D7 }, // CYRILLIC SMALL LETTER IE WITH BREVE
    Interval{ 0x04D9, 0x04D9 }, // CYRILLIC SMALL LETTER SCHWA
    Interval{ 0x04DB, 0x04DB }, // CYRILLIC SMALL LETTER SCHWA WITH DIAERESIS
    Interval{ 0x04DD, 0x04DD }, // CYRILLIC SMALL LETTER ZHE WITH DIAERESIS
    Interval{ 0x04DF, 0x04DF }, // CYRILLIC SMALL LETTER ZE WITH DIAERESIS
    Interval{ 0x04E1, 0x04E1 }, // CYRILLIC SMALL LETTER ABKHASIAN DZE
    Interval{ 0x04E3, 0x04E3 }, // CYRILLIC SMALL LETTER I WITH MACRON
    Interval{ 0x04E5, 0x04E5 }, // CYRILLIC SMALL LETTER I WITH DIAERESIS
    Interval{ 0x04E7, 0x04E7 }, // CYRILLIC SMALL LETTER O WITH DIAERESIS
    Interval{ 0x04E9, 0x04E9 }, // CYRILLIC SMALL LETTER BARRED O
    Interval{ 0x04EB, 0x04EB }, // CYRILLIC SMALL LETTER BARRED O WITH DIAERESIS
    Interval{ 0x04ED, 0x04ED }, // CYRILLIC SMALL LETTER E WITH DIAERESIS
    Interval{ 0x04EF, 0x04EF }, // CYRILLIC SMALL LETTER U WITH MACRON
    Interval{ 0x04F1, 0x04F1 }, // CYRILLIC SMALL LETTER U WITH DIAERESIS
    Interval{ 0x04F3, 0x04F3 }, // CYRILLIC SMALL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x04F5, 0x04F5 }, // CYRILLIC SMALL LETTER CHE WITH DIAERESIS
    Interval{ 0x04F7, 0x04F7 }, // CYRILLIC SMALL LETTER GHE WITH DESCENDER
    Interval{ 0x04F9, 0x04F9 }, // CYRILLIC SMALL LETTER YERU WITH DIAERESIS
    Interval{ 0x04FB, 0x04FB }, // CYRILLIC SMALL LETTER GHE WITH STROKE AND HOOK
    Interval{ 0x04FD, 0x04FD }, // CYRILLIC SMALL LETTER HA WITH HOOK
    Interval{ 0x04FF, 0x04FF }, // CYRILLIC SMALL LETTER HA WITH STROKE
    Interval{ 0x0501, 0x0501 }, // CYRILLIC SMALL LETTER KOMI DE
    Interval{ 0x0503, 0x0503 }, // CYRILLIC SMALL LETTER KOMI DJE
    Interval{ 0x0505, 0x0505 }, // CYRILLIC SMALL LETTER KOMI ZJE
    Interval{ 0x0507, 0x0507 }, // CYRILLIC SMALL LETTER KOMI DZJE
    Interval{ 0x0509, 0x0509 }, // CYRILLIC SMALL LETTER KOMI LJE
    Interval{ 0x050B, 0x050B }, // CYRILLIC SMALL LETTER KOMI NJE
    Interval{ 0x050D, 0x050D }, // CYRILLIC SMALL LETTER KOMI SJE
    Interval{ 0x050F, 0x050F }, // CYRILLIC SMALL LETTER KOMI TJE
    Interval{ 0x0511, 0x0511 }, // CYRILLIC SMALL LETTER REVERSED ZE
    Interval{ 0x0513, 0x0513 }, // CYRILLIC SMALL LETTER EL WITH HOOK
    Interval{ 0x0515, 0x0515 }, // CYRILLIC SMALL LETTER LHA
    Interval{ 0x0517, 0x0517 }, // CYRILLIC SMALL LETTER RHA
    Interval{ 0x0519, 0x0519 }, // CYRILLIC SMALL LETTER YAE
    Interval{ 0x051B, 0x051B }, // CYRILLIC SMALL LETTER QA
    Interval{ 0x051D, 0x051D }, // CYRILLIC SMALL LETTER WE
    Interval{ 0x051F, 0x051F }, // CYRILLIC SMALL LETTER ALEUT KA
    Interval{ 0x0521, 0x0521 }, // CYRILLIC SMALL LETTER EL WITH MIDDLE HOOK
    Interval{ 0x0523, 0x0523 }, // CYRILLIC SMALL LETTER EN WITH MIDDLE HOOK
    Interval{ 0x0525, 0x0525 }, // CYRILLIC SMALL LETTER PE WITH DESCENDER
    Interval{ 0x0527, 0x0527 }, // CYRILLIC SMALL LETTER SHHA WITH DESCENDER
    Interval{ 0x0529, 0x0529 }, // CYRILLIC SMALL LETTER EN WITH LEFT HOOK
    Interval{ 0x052B, 0x052B }, // CYRILLIC SMALL LETTER DZZHE
    Interval{ 0x052D, 0x052D }, // CYRILLIC SMALL LETTER DCHE
    Interval{ 0x052F, 0x052F }, // CYRILLIC SMALL LETTER EL WITH DESCENDER
    Interval{ 0x0560, 0x0588 }, // [41] ARMENIAN SMALL LETTER TURNED AYB..ARMENIAN SMALL LETTER YI WITH STROKE
    Interval{ 0x10D0, 0x10FA }, // [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Interval{ 0x10FD, 0x10FF }, // [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Interval{ 0x13F8, 0x13FD }, // [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Interval{ 0x1C80, 0x1C88 }, // [9] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER UNBLENDED UK
    Interval{ 0x1C8A, 0x1C8A }, // CYRILLIC SMALL LETTER TJE
    Interval{ 0x1D00, 0x1D2B }, // [44] LATIN LETTER SMALL CAPITAL A..CYRILLIC LETTER SMALL CAPITAL EL
    Interval{ 0x1D6B, 0x1D77 }, // [13] LATIN SMALL LETTER UE..LATIN SMALL LETTER TURNED G
    Interval{ 0x1D79, 0x1D9A }, // [34] LATIN SMALL LETTER INSULAR G..LATIN SMALL LETTER EZH WITH RETROFLEX HOOK
    Interval{ 0x1E01, 0x1E01 }, // LATIN SMALL LETTER A WITH RING BELOW
    Interval{ 0x1E03, 0x1E03 }, // LATIN SMALL LETTER B WITH DOT ABOVE
    Interval{ 0x1E05, 0x1E05 }, // LATIN SMALL LETTER B WITH DOT BELOW
    Interval{ 0x1E07, 0x1E07 }, // LATIN SMALL LETTER B WITH LINE BELOW
    Interval{ 0x1E09, 0x1E09 }, // LATIN SMALL LETTER C WITH CEDILLA AND ACUTE
    Interval{ 0x1E0B, 0x1E0B }, // LATIN SMALL LETTER D WITH DOT ABOVE
    Interval{ 0x1E0D, 0x1E0D }, // LATIN SMALL LETTER D WITH DOT BELOW
    Interval{ 0x1E0F, 0x1E0F }, // LATIN SMALL LETTER D WITH LINE BELOW
    Interval{ 0x1E11, 0x1E11 }, // LATIN SMALL LETTER D WITH CEDILLA
    Interval{ 0x1E13, 0x1E13 }, // LATIN SMALL LETTER D WITH CIRCUMFLEX BELOW
    Interval{ 0x1E15, 0x1E15 }, // LATIN SMALL LETTER E WITH MACRON AND GRAVE
    Interval{ 0x1E17, 0x1E17 }, // LATIN SMALL LETTER E WITH MACRON AND ACUTE
    Interval{ 0x1E19, 0x1E19 }, // LATIN SMALL LETTER E WITH CIRCUMFLEX BELOW
    Interval{ 0x1E1B, 0x1E1B }, // LATIN SMALL LETTER E WITH TILDE BELOW
    Interval{ 0x1E1D, 0x1E1D }, // LATIN SMALL LETTER E WITH CEDILLA AND BREVE
    Interval{ 0x1E1F, 0x1E1F }, // LATIN SMALL LETTER F WITH DOT ABOVE
    Interval{ 0x1E21, 0x1E21 }, // LATIN SMALL LETTER G WITH MACRON
    Interval{ 0x1E23, 0x1E23 }, // LATIN SMALL LETTER H WITH DOT ABOVE
    Interval{ 0x1E25, 0x1E25 }, // LATIN SMALL LETTER H WITH DOT BELOW
    Interval{ 0x1E27, 0x1E27 }, // LATIN SMALL LETTER H WITH DIAERESIS
    Interval{ 0x1E29, 0x1E29 }, // LATIN SMALL LETTER H WITH CEDILLA
    Interval{ 0x1E2B, 0x1E2B }, // LATIN SMALL LETTER H WITH BREVE BELOW
    Interval{ 0x1E2D, 0x1E2D }, // LATIN SMALL LETTER I WITH TILDE BELOW
    Interval{ 0x1E2F, 0x1E2F }, // LATIN SMALL LETTER I WITH DIAERESIS AND ACUTE
    Interval{ 0x1E31, 0x1E31 }, // LATIN SMALL LETTER K WITH ACUTE
    Interval{ 0x1E33, 0x1E33 }, // LATIN SMALL LETTER K WITH DOT BELOW
    Interval{ 0x1E35, 0x1E35 }, // LATIN SMALL LETTER K WITH LINE BELOW
    Interval{ 0x1E37, 0x1E37 }, // LATIN SMALL LETTER L WITH DOT BELOW
    Interval{ 0x1E39, 0x1E39 }, // LATIN SMALL LETTER L WITH DOT BELOW AND MACRON
    Interval{ 0x1E3B, 0x1E3B }, // LATIN SMALL LETTER L WITH LINE BELOW
    Interval{ 0x1E3D, 0x1E3D }, // LATIN SMALL LETTER L WITH CIRCUMFLEX BELOW
    Interval{ 0x1E3F, 0x1E3F }, // LATIN SMALL LETTER M WITH ACUTE
    Interval{ 0x1E41, 0x1E41 }, // LATIN SMALL LETTER M WITH DOT ABOVE
    Interval{ 0x1E43, 0x1E43 }, // LATIN SMALL LETTER M WITH DOT BELOW
    Interval{ 0x1E45, 0x1E45 }, // LATIN SMALL LETTER N WITH DOT ABOVE
    Interval{ 0x1E47, 0x1E47 }, // LATIN SMALL LETTER N WITH DOT BELOW
    Interval{ 0x1E49, 0x1E49 }, // LATIN SMALL LETTER N WITH LINE BELOW
    Interval{ 0x1E4B, 0x1E4B }, // LATIN SMALL LETTER N WITH CIRCUMFLEX BELOW
    Interval{ 0x1E4D, 0x1E4D }, // LATIN SMALL LETTER O WITH TILDE AND ACUTE
    Interval{ 0x1E4F, 0x1E4F }, // LATIN SMALL LETTER O WITH TILDE AND DIAERESIS
    Interval{ 0x1E51, 0x1E51 }, // LATIN SMALL LETTER O WITH MACRON AND GRAVE
    Interval{ 0x1E53, 0x1E53 }, // LATIN SMALL LETTER O WITH MACRON AND ACUTE
    Interval{ 0x1E55, 0x1E55 }, // LATIN SMALL LETTER P WITH ACUTE
    Interval{ 0x1E57, 0x1E57 }, // LATIN SMALL LETTER P WITH DOT ABOVE
    Interval{ 0x1E59, 0x1E59 }, // LATIN SMALL LETTER R WITH DOT ABOVE
    Interval{ 0x1E5B, 0x1E5B }, // LATIN SMALL LETTER R WITH DOT BELOW
    Interval{ 0x1E5D, 0x1E5D }, // LATIN SMALL LETTER R WITH DOT BELOW AND MACRON
    Interval{ 0x1E5F, 0x1E5F }, // LATIN SMALL LETTER R WITH LINE BELOW
    Interval{ 0x1E61, 0x1E61 }, // LATIN SMALL LETTER S WITH DOT ABOVE
    Interval{ 0x1E63, 0x1E63 }, // LATIN SMALL LETTER S WITH DOT BELOW
    Interval{ 0x1E65, 0x1E65 }, // LATIN SMALL LETTER S WITH ACUTE AND DOT ABOVE
    Interval{ 0x1E67, 0x1E67 }, // LATIN SMALL LETTER S WITH CARON AND DOT ABOVE
    Interval{ 0x1E69, 0x1E69 }, // LATIN SMALL LETTER S WITH DOT BELOW AND DOT ABOVE
    Interval{ 0x1E6B, 0x1E6B }, // LATIN SMALL LETTER T WITH DOT ABOVE
    Interval{ 0x1E6D, 0x1E6D }, // LATIN SMALL LETTER T WITH DOT BELOW
    Interval{ 0x1E6F, 0x1E6F }, // LATIN SMALL LETTER T WITH LINE BELOW
    Interval{ 0x1E71, 0x1E71 }, // LATIN SMALL LETTER T WITH CIRCUMFLEX BELOW
    Interval{ 0x1E73, 0x1E73 }, // LATIN SMALL LETTER U WITH DIAERESIS BELOW
    Interval{ 0x1E75, 0x1E75 }, // LATIN SMALL LETTER U WITH TILDE BELOW
    Interval{ 0x1E77, 0x1E77 }, // LATIN SMALL LETTER U WITH CIRCUMFLEX BELOW
    Interval{ 0x1E79, 0x1E79 }, // LATIN SMALL LETTER U WITH TILDE AND ACUTE
    Interval{ 0x1E7B, 0x1E7B }, // LATIN SMALL LETTER U WITH MACRON AND DIAERESIS
    Interval{ 0x1E7D, 0x1E7D }, // LATIN SMALL LETTER V WITH TILDE
    Interval{ 0x1E7F, 0x1E7F }, // LATIN SMALL LETTER V WITH DOT BELOW
    Interval{ 0x1E81, 0x1E81 }, // LATIN SMALL LETTER W WITH GRAVE
    Interval{ 0x1E83, 0x1E83 }, // LATIN SMALL LETTER W WITH ACUTE
    Interval{ 0x1E85, 0x1E85 }, // LATIN SMALL LETTER W WITH DIAERESIS
    Interval{ 0x1E87, 0x1E87 }, // LATIN SMALL LETTER W WITH DOT ABOVE
    Interval{ 0x1E89, 0x1E89 }, // LATIN SMALL LETTER W WITH DOT BELOW
    Interval{ 0x1E8B, 0x1E8B }, // LATIN SMALL LETTER X WITH DOT ABOVE
    Interval{ 0x1E8D, 0x1E8D }, // LATIN SMALL LETTER X WITH DIAERESIS
    Interval{ 0x1E8F, 0x1E8F }, // LATIN SMALL LETTER Y WITH DOT ABOVE
    Interval{ 0x1E91, 0x1E91 }, // LATIN SMALL LETTER Z WITH CIRCUMFLEX
    Interval{ 0x1E93, 0x1E93 }, // LATIN SMALL LETTER Z WITH DOT BELOW
    Interval{ 0x1E95, 0x1E9D }, // [9] LATIN SMALL LETTER Z WITH LINE BELOW..LATIN SMALL LETTER LONG S WITH HIGH STROKE
    Interval{ 0x1E9F, 0x1E9F }, // LATIN SMALL LETTER DELTA
    Interval{ 0x1EA1, 0x1EA1 }, // LATIN SMALL LETTER A WITH DOT BELOW
    Interval{ 0x1EA3, 0x1EA3 }, // LATIN SMALL LETTER A WITH HOOK ABOVE
    Interval{ 0x1EA5, 0x1EA5 }, // LATIN SMALL LETTER A WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EA7, 0x1EA7 }, // LATIN SMALL LETTER A WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EA9, 0x1EA9 }, // LATIN SMALL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EAB, 0x1EAB }, // LATIN SMALL LETTER A WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EAD, 0x1EAD }, // LATIN SMALL LETTER A WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EAF, 0x1EAF }, // LATIN SMALL LETTER A WITH BREVE AND ACUTE
    Interval{ 0x1EB1, 0x1EB1 }, // LATIN SMALL LETTER A WITH BREVE AND GRAVE
    Interval{ 0x1EB3, 0x1EB3 }, // LATIN SMALL LETTER A WITH BREVE AND HOOK ABOVE
    Interval{ 0x1EB5, 0x1EB5 }, // LATIN SMALL LETTER A WITH BREVE AND TILDE
    Interval{ 0x1EB7, 0x1EB7 }, // LATIN SMALL LETTER A WITH BREVE AND DOT BELOW
    Interval{ 0x1EB9, 0x1EB9 }, // LATIN SMALL LETTER E WITH DOT BELOW
    Interval{ 0x1EBB, 0x1EBB }, // LATIN SMALL LETTER E WITH HOOK ABOVE
    Interval{ 0x1EBD, 0x1EBD }, // LATIN SMALL LETTER E WITH TILDE
    Interval{ 0x1EBF, 0x1EBF }, // LATIN SMALL LETTER E WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EC1, 0x1EC1 }, // LATIN SMALL LETTER E WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EC3, 0x1EC3 }, // LATIN SMALL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EC5, 0x1EC5 }, // LATIN SMALL LETTER E WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EC7, 0x1EC7 }, // LATIN SMALL LETTER E WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EC9, 0x1EC9 }, // LATIN SMALL LETTER I WITH HOOK ABOVE
    Interval{ 0x1ECB, 0x1ECB }, // LATIN SMALL LETTER I WITH DOT BELOW
    Interval{ 0x1ECD, 0x1ECD }, // LATIN SMALL LETTER O WITH DOT BELOW
    Interval{ 0x1ECF, 0x1ECF }, // LATIN SMALL LETTER O WITH HOOK ABOVE
    Interval{ 0x1ED1, 0x1ED1 }, // LATIN SMALL LETTER O WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1ED3, 0x1ED3 }, // LATIN SMALL LETTER O WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1ED5, 0x1ED5 }, // LATIN SMALL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1ED7, 0x1ED7 }, // LATIN SMALL LETTER O WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1ED9, 0x1ED9 }, // LATIN SMALL LETTER O WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EDB, 0x1EDB }, // LATIN SMALL LETTER O WITH HORN AND ACUTE
    Interval{ 0x1EDD, 0x1EDD }, // LATIN SMALL LETTER O WITH HORN AND GRAVE
    Interval{ 0x1EDF, 0x1EDF }, // LATIN SMALL LETTER O WITH HORN AND HOOK ABOVE
    Interval{ 0x1EE1, 0x1EE1 }, // LATIN SMALL LETTER O WITH HORN AND TILDE
    Interval{ 0x1EE3, 0x1EE3 }, // LATIN SMALL LETTER O WITH HORN AND DOT BELOW
    Interval{ 0x1EE5, 0x1EE5 }, // LATIN SMALL LETTER U WITH DOT BELOW
    Interval{ 0x1EE7, 0x1EE7 }, // LATIN SMALL LETTER U WITH HOOK ABOVE
    Interval{ 0x1EE9, 0x1EE9 }, // LATIN SMALL LETTER U WITH HORN AND ACUTE
    Interval{ 0x1EEB, 0x1EEB }, // LATIN SMALL LETTER U WITH HORN AND GRAVE
    Interval{ 0x1EED, 0x1EED }, // LATIN SMALL LETTER U WITH HORN AND HOOK ABOVE
    Interval{ 0x1EEF, 0x1EEF }, // LATIN SMALL LETTER U WITH HORN AND TILDE
    Interval{ 0x1EF1, 0x1EF1 }, // LATIN SMALL LETTER U WITH HORN AND DOT BELOW
    Interval{ 0x1EF3, 0x1EF3 }, // LATIN SMALL LETTER Y WITH GRAVE
    Interval{ 0x1EF5, 0x1EF5 }, // LATIN SMALL LETTER Y WITH DOT BELOW
    Interval{ 0x1EF7, 0x1EF7 }, // LATIN SMALL LETTER Y WITH HOOK ABOVE
    Interval{ 0x1EF9, 0x1EF9 }, // LATIN SMALL LETTER Y WITH TILDE
    Interval{ 0x1EFB, 0x1EFB }, // LATIN SMALL LETTER MIDDLE-WELSH LL
    Interval{ 0x1EFD, 0x1EFD }, // LATIN SMALL LETTER MIDDLE-WELSH V
    Interval{ 0x1EFF, 0x1F07 }, // [9] LATIN SMALL LETTER Y WITH LOOP..GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F10, 0x1F15 }, // [6] GREEK SMALL LETTER EPSILON WITH PSILI..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F20, 0x1F27 }, // [8] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F30, 0x1F37 }, // [8] GREEK SMALL LETTER IOTA WITH PSILI..GREEK SMALL LETTER IOTA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F40, 0x1F45 }, // [6] GREEK SMALL LETTER OMICRON WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F50, 0x1F57 }, // [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F60, 0x1F67 }, // [8] GREEK SMALL LETTER OMEGA WITH PSILI..GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F70, 0x1F7D }, // [14] GREEK SMALL LETTER ALPHA WITH VARIA..GREEK SMALL LETTER OMEGA WITH OXIA
    Interval{ 0x1F80, 0x1F87 }, // [8] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1F90, 0x1F97 }, // [8] GREEK SMALL LETTER ETA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FA0, 0x1FA7 }, // [8] GREEK SMALL LETTER OMEGA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH DASIA AND PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FB0, 0x1FB4 }, // [5] GREEK SMALL LETTER ALPHA WITH VRACHY..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FB6, 0x1FB7 }, // [2] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK SMALL LETTER ALPHA WITH PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FBE, 0x1FBE }, // GREEK PROSGEGRAMMENI
    Interval{ 0x1FC2, 0x1FC4 }, // [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FC6, 0x1FC7 }, // [2] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK SMALL LETTER ETA WITH PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x1FD0, 0x1FD3 }, // [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Interval{ 0x1FD6, 0x1FD7 }, // [2] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND PERISPOMENI
    Interval{ 0x1FE0, 0x1FE7 }, // [8] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND PERISPOMENI
    Interval{ 0x1FF2, 0x1FF4 }, // [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Interval{ 0x1FF6, 0x1FF7 }, // [2] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK SMALL LETTER OMEGA WITH PERISPOMENI AND YPOGEGRAMMENI
    Interval{ 0x210A, 0x210A }, // SCRIPT SMALL G
    Interval{ 0x210E, 0x210F }, // [2] PLANCK CONSTANT..PLANCK CONSTANT OVER TWO PI
    Interval{ 0x2113, 0x2113 }, // SCRIPT SMALL L
    Interval{ 0x212F, 0x212F }, // SCRIPT SMALL E
    Interval{ 0x2134, 0x2134 }, // SCRIPT SMALL O
    Interval{ 0x2139, 0x2139 }, // INFORMATION SOURCE
    Interval{ 0x213C, 0x213D }, // [2] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK SMALL GAMMA
    Interval{ 0x2146, 0x2149 }, // [4] DOUBLE-STRUCK ITALIC SMALL D..DOUBLE-STRUCK ITALIC SMALL J
    Interval{ 0x214E, 0x214E }, // TURNED SMALL F
    Interval{ 0x2184, 0x2184 }, // LATIN SMALL LETTER REVERSED C
    Interval{ 0x2C30, 0x2C5F }, // [48] GLAGOLITIC SMALL LETTER AZU..GLAGOLITIC SMALL LETTER CAUDATE CHRIVI
    Interval{ 0x2C61, 0x2C61 }, // LATIN SMALL LETTER L WITH DOUBLE BAR
    Interval{ 0x2C65, 0x2C66 }, // [2] LATIN SMALL LETTER A WITH STROKE..LATIN SMALL LETTER T WITH DIAGONAL STROKE
    Interval{ 0x2C68, 0x2C68 }, // LATIN SMALL LETTER H WITH DESCENDER
    Interval{ 0x2C6A, 0x2C6A }, // LATIN SMALL LETTER K WITH DESCENDER
    Interval{ 0x2C6C, 0x2C6C }, // LATIN SMALL LETTER Z WITH DESCENDER
    Interval{ 0x2C71, 0x2C71 }, // LATIN SMALL LETTER V WITH RIGHT HOOK
    Interval{ 0x2C73, 0x2C74 }, // [2] LATIN SMALL LETTER W WITH HOOK..LATIN SMALL LETTER V WITH CURL
    Interval{ 0x2C76, 0x2C7B }, // [6] LATIN SMALL LETTER HALF H..LATIN LETTER SMALL CAPITAL TURNED E
    Interval{ 0x2C81, 0x2C81 }, // COPTIC SMALL LETTER ALFA
    Interval{ 0x2C83, 0x2C83 }, // COPTIC SMALL LETTER VIDA
    Interval{ 0x2C85, 0x2C85 }, // COPTIC SMALL LETTER GAMMA
    Interval{ 0x2C87, 0x2C87 }, // COPTIC SMALL LETTER DALDA
    Interval{ 0x2C89, 0x2C89 }, // COPTIC SMALL LETTER EIE
    Interval{ 0x2C8B, 0x2C8B }, // COPTIC SMALL LETTER SOU
    Interval{ 0x2C8D, 0x2C8D }, // COPTIC SMALL LETTER ZATA
    Interval{ 0x2C8F, 0x2C8F }, // COPTIC SMALL LETTER HATE
    Interval{ 0x2C91, 0x2C91 }, // COPTIC SMALL LETTER THETHE
    Interval{ 0x2C93, 0x2C93 }, // COPTIC SMALL LETTER IAUDA
    Interval{ 0x2C95, 0x2C95 }, // COPTIC SMALL LETTER KAPA
    Interval{ 0x2C97, 0x2C97 }, // COPTIC SMALL LETTER LAULA
    Interval{ 0x2C99, 0x2C99 }, // COPTIC SMALL LETTER MI
    Interval{ 0x2C9B, 0x2C9B }, // COPTIC SMALL LETTER NI
    Interval{ 0x2C9D, 0x2C9D }, // COPTIC SMALL LETTER KSI
    Interval{ 0x2C9F, 0x2C9F }, // COPTIC SMALL LETTER O
    Interval{ 0x2CA1, 0x2CA1 }, // COPTIC SMALL LETTER PI
    Interval{ 0x2CA3, 0x2CA3 }, // COPTIC SMALL LETTER RO
    Interval{ 0x2CA5, 0x2CA5 }, // COPTIC SMALL LETTER SIMA
    Interval{ 0x2CA7, 0x2CA7 }, // COPTIC SMALL LETTER TAU
    Interval{ 0x2CA9, 0x2CA9 }, // COPTIC SMALL LETTER UA
    Interval{ 0x2CAB, 0x2CAB }, // COPTIC SMALL LETTER FI
    Interval{ 0x2CAD, 0x2CAD }, // COPTIC SMALL LETTER KHI
    Interval{ 0x2CAF, 0x2CAF }, // COPTIC SMALL LETTER PSI
    Interval{ 0x2CB1, 0x2CB1 }, // COPTIC SMALL LETTER OOU
    Interval{ 0x2CB3, 0x2CB3 }, // COPTIC SMALL LETTER DIALECT-P ALEF
    Interval{ 0x2CB5, 0x2CB5 }, // COPTIC SMALL LETTER OLD COPTIC AIN
    Interval{ 0x2CB7, 0x2CB7 }, // COPTIC SMALL LETTER CRYPTOGRAMMIC EIE
    Interval{ 0x2CB9, 0x2CB9 }, // COPTIC SMALL LETTER DIALECT-P KAPA
    Interval{ 0x2CBB, 0x2CBB }, // COPTIC SMALL LETTER DIALECT-P NI
    Interval{ 0x2CBD, 0x2CBD }, // COPTIC SMALL LETTER CRYPTOGRAMMIC NI
    Interval{ 0x2CBF, 0x2CBF }, // COPTIC SMALL LETTER OLD COPTIC OOU
    Interval{ 0x2CC1, 0x2CC1 }, // COPTIC SMALL LETTER SAMPI
    Interval{ 0x2CC3, 0x2CC3 }, // COPTIC SMALL LETTER CROSSED SHEI
    Interval{ 0x2CC5, 0x2CC5 }, // COPTIC SMALL LETTER OLD COPTIC SHEI
    Interval{ 0x2CC7, 0x2CC7 }, // COPTIC SMALL LETTER OLD COPTIC ESH
    Interval{ 0x2CC9, 0x2CC9 }, // COPTIC SMALL LETTER AKHMIMIC KHEI
    Interval{ 0x2CCB, 0x2CCB }, // COPTIC SMALL LETTER DIALECT-P HORI
    Interval{ 0x2CCD, 0x2CCD }, // COPTIC SMALL LETTER OLD COPTIC HORI
    Interval{ 0x2CCF, 0x2CCF }, // COPTIC SMALL LETTER OLD COPTIC HA
    Interval{ 0x2CD1, 0x2CD1 }, // COPTIC SMALL LETTER L-SHAPED HA
    Interval{ 0x2CD3, 0x2CD3 }, // COPTIC SMALL LETTER OLD COPTIC HEI
    Interval{ 0x2CD5, 0x2CD5 }, // COPTIC SMALL LETTER OLD COPTIC HAT
    Interval{ 0x2CD7, 0x2CD7 }, // COPTIC SMALL LETTER OLD COPTIC GANGIA
    Interval{ 0x2CD9, 0x2CD9 }, // COPTIC SMALL LETTER OLD COPTIC DJA
    Interval{ 0x2CDB, 0x2CDB }, // COPTIC SMALL LETTER OLD COPTIC SHIMA
    Interval{ 0x2CDD, 0x2CDD }, // COPTIC SMALL LETTER OLD NUBIAN SHIMA
    Interval{ 0x2CDF, 0x2CDF }, // COPTIC SMALL LETTER OLD NUBIAN NGI
    Interval{ 0x2CE1, 0x2CE1 }, // COPTIC SMALL LETTER OLD NUBIAN NYI
    Interval{ 0x2CE3, 0x2CE4 }, // [2] COPTIC SMALL LETTER OLD NUBIAN WAU..COPTIC SYMBOL KAI
    Interval{ 0x2CEC, 0x2CEC }, // COPTIC SMALL LETTER CRYPTOGRAMMIC SHEI
    Interval{ 0x2CEE, 0x2CEE }, // COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF3, 0x2CF3 }, // COPTIC SMALL LETTER BOHAIRIC KHEI
    Interval{ 0x2D00, 0x2D25 }, // [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Interval{ 0x2D27, 0x2D27 }, // GEORGIAN SMALL LETTER YN
    Interval{ 0x2D2D, 0x2D2D }, // GEORGIAN SMALL LETTER AEN
    Interval{ 0xA641, 0xA641 }, // CYRILLIC SMALL LETTER ZEMLYA
    Interval{ 0xA643, 0xA643 }, // CYRILLIC SMALL LETTER DZELO
    Interval{ 0xA645, 0xA645 }, // CYRILLIC SMALL LETTER REVERSED DZE
    Interval{ 0xA647, 0xA647 }, // CYRILLIC SMALL LETTER IOTA
    Interval{ 0xA649, 0xA649 }, // CYRILLIC SMALL LETTER DJERV
    Interval{ 0xA64B, 0xA64B }, // CYRILLIC SMALL LETTER MONOGRAPH UK
    Interval{ 0xA64D, 0xA64D }, // CYRILLIC SMALL LETTER BROAD OMEGA
    Interval{ 0xA64F, 0xA64F }, // CYRILLIC SMALL LETTER NEUTRAL YER
    Interval{ 0xA651, 0xA651 }, // CYRILLIC SMALL LETTER YERU WITH BACK YER
    Interval{ 0xA653, 0xA653 }, // CYRILLIC SMALL LETTER IOTIFIED YAT
    Interval{ 0xA655, 0xA655 }, // CYRILLIC SMALL LETTER REVERSED YU
    Interval{ 0xA657, 0xA657 }, // CYRILLIC SMALL LETTER IOTIFIED A
    Interval{ 0xA659, 0xA659 }, // CYRILLIC SMALL LETTER CLOSED LITTLE YUS
    Interval{ 0xA65B, 0xA65B }, // CYRILLIC SMALL LETTER BLENDED YUS
    Interval{ 0xA65D, 0xA65D }, // CYRILLIC SMALL LETTER IOTIFIED CLOSED LITTLE YUS
    Interval{ 0xA65F, 0xA65F }, // CYRILLIC SMALL LETTER YN
    Interval{ 0xA661, 0xA661 }, // CYRILLIC SMALL LETTER REVERSED TSE
    Interval{ 0xA663, 0xA663 }, // CYRILLIC SMALL LETTER SOFT DE
    Interval{ 0xA665, 0xA665 }, // CYRILLIC SMALL LETTER SOFT EL
    Interval{ 0xA667, 0xA667 }, // CYRILLIC SMALL LETTER SOFT EM
    Interval{ 0xA669, 0xA669 }, // CYRILLIC SMALL LETTER MONOCULAR O
    Interval{ 0xA66B, 0xA66B }, // CYRILLIC SMALL LETTER BINOCULAR O
    Interval{ 0xA66D, 0xA66D }, // CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA681, 0xA681 }, // CYRILLIC SMALL LETTER DWE
    Interval{ 0xA683, 0xA683 }, // CYRILLIC SMALL LETTER DZWE
    Interval{ 0xA685, 0xA685 }, // CYRILLIC SMALL LETTER ZHWE
    Interval{ 0xA687, 0xA687 }, // CYRILLIC SMALL LETTER CCHE
    Interval{ 0xA689, 0xA689 }, // CYRILLIC SMALL LETTER DZZE
    Interval{ 0xA68B, 0xA68B }, // CYRILLIC SMALL LETTER TE WITH MIDDLE HOOK
    Interval{ 0xA68D, 0xA68D }, // CYRILLIC SMALL LETTER TWE
    Interval{ 0xA68F, 0xA68F }, // CYRILLIC SMALL LETTER TSWE
    Interval{ 0xA691, 0xA691 }, // CYRILLIC SMALL LETTER TSSE
    Interval{ 0xA693, 0xA693 }, // CYRILLIC SMALL LETTER TCHE
    Interval{ 0xA695, 0xA695 }, // CYRILLIC SMALL LETTER HWE
    Interval{ 0xA697, 0xA697 }, // CYRILLIC SMALL LETTER SHWE
    Interval{ 0xA699, 0xA699 }, // CYRILLIC SMALL LETTER DOUBLE O
    Interval{ 0xA69B, 0xA69B }, // CYRILLIC SMALL LETTER CROSSED O
    Interval{ 0xA723, 0xA723 }, // LATIN SMALL LETTER EGYPTOLOGICAL ALEF
    Interval{ 0xA725, 0xA725 }, // LATIN SMALL LETTER EGYPTOLOGICAL AIN
    Interval{ 0xA727, 0xA727 }, // LATIN SMALL LETTER HENG
    Interval{ 0xA729, 0xA729 }, // LATIN SMALL LETTER TZ
    Interval{ 0xA72B, 0xA72B }, // LATIN SMALL LETTER TRESILLO
    Interval{ 0xA72D, 0xA72D }, // LATIN SMALL LETTER CUATRILLO
    Interval{ 0xA72F, 0xA731 }, // [3] LATIN SMALL LETTER CUATRILLO WITH COMMA..LATIN LETTER SMALL CAPITAL S
    Interval{ 0xA733, 0xA733 }, // LATIN SMALL LETTER AA
    Interval{ 0xA735, 0xA735 }, // LATIN SMALL LETTER AO
    Interval{ 0xA737, 0xA737 }, // LATIN SMALL LETTER AU
    Interval{ 0xA739, 0xA739 }, // LATIN SMALL LETTER AV
    Interval{ 0xA73B, 0xA73B }, // LATIN SMALL LETTER AV WITH HORIZONTAL BAR
    Interval{ 0xA73D, 0xA73D }, // LATIN SMALL LETTER AY
    Interval{ 0xA73F, 0xA73F }, // LATIN SMALL LETTER REVERSED C WITH DOT
    Interval{ 0xA741, 0xA741 }, // LATIN SMALL LETTER K WITH STROKE
    Interval{ 0xA743, 0xA743 }, // LATIN SMALL LETTER K WITH DIAGONAL STROKE
    Interval{ 0xA745, 0xA745 }, // LATIN SMALL LETTER K WITH STROKE AND DIAGONAL STROKE
    Interval{ 0xA747, 0xA747 }, // LATIN SMALL LETTER BROKEN L
    Interval{ 0xA749, 0xA749 }, // LATIN SMALL LETTER L WITH HIGH STROKE
    Interval{ 0xA74B, 0xA74B }, // LATIN SMALL LETTER O WITH LONG STROKE OVERLAY
    Interval{ 0xA74D, 0xA74D }, // LATIN SMALL LETTER O WITH LOOP
    Interval{ 0xA74F, 0xA74F }, // LATIN SMALL LETTER OO
    Interval{ 0xA751, 0xA751 }, // LATIN SMALL LETTER P WITH STROKE THROUGH DESCENDER
    Interval{ 0xA753, 0xA753 }, // LATIN SMALL LETTER P WITH FLOURISH
    Interval{ 0xA755, 0xA755 }, // LATIN SMALL LETTER P WITH SQUIRREL TAIL
    Interval{ 0xA757, 0xA757 }, // LATIN SMALL LETTER Q WITH STROKE THROUGH DESCENDER
    Interval{ 0xA759, 0xA759 }, // LATIN SMALL LETTER Q WITH DIAGONAL STROKE
    Interval{ 0xA75B, 0xA75B }, // LATIN SMALL LETTER R ROTUNDA
    Interval{ 0xA75D, 0xA75D }, // LATIN SMALL LETTER RUM ROTUNDA
    Interval{ 0xA75F, 0xA75F }, // LATIN SMALL LETTER V WITH DIAGONAL STROKE
    Interval{ 0xA761, 0xA761 }, // LATIN SMALL LETTER VY
    Interval{ 0xA763, 0xA763 }, // LATIN SMALL LETTER VISIGOTHIC Z
    Interval{ 0xA765, 0xA765 }, // LATIN SMALL LETTER THORN WITH STROKE
    Interval{ 0xA767, 0xA767 }, // LATIN SMALL LETTER THORN WITH STROKE THROUGH DESCENDER
    Interval{ 0xA769, 0xA769 }, // LATIN SMALL LETTER VEND
    Interval{ 0xA76B, 0xA76B }, // LATIN SMALL LETTER ET
    Interval{ 0xA76D, 0xA76D }, // LATIN SMALL LETTER IS
    Interval{ 0xA76F, 0xA76F }, // LATIN SMALL LETTER CON
    Interval{ 0xA771, 0xA778 }, // [8] LATIN SMALL LETTER DUM..LATIN SMALL LETTER UM
    Interval{ 0xA77A, 0xA77A }, // LATIN SMALL LETTER INSULAR D
    Interval{ 0xA77C, 0xA77C }, // LATIN SMALL LETTER INSULAR F
    Interval{ 0xA77F, 0xA77F }, // LATIN SMALL LETTER TURNED INSULAR G
    Interval{ 0xA781, 0xA781 }, // LATIN SMALL LETTER TURNED L
    Interval{ 0xA783, 0xA783 }, // LATIN SMALL LETTER INSULAR R
    Interval{ 0xA785, 0xA785 }, // LATIN SMALL LETTER INSULAR S
    Interval{ 0xA787, 0xA787 }, // LATIN SMALL LETTER INSULAR T
    Interval{ 0xA78C, 0xA78C }, // LATIN SMALL LETTER SALTILLO
    Interval{ 0xA78E, 0xA78E }, // LATIN SMALL LETTER L WITH RETROFLEX HOOK AND BELT
    Interval{ 0xA791, 0xA791 }, // LATIN SMALL LETTER N WITH DESCENDER
    Interval{ 0xA793, 0xA795 }, // [3] LATIN SMALL LETTER C WITH BAR..LATIN SMALL LETTER H WITH PALATAL HOOK
    Interval{ 0xA797, 0xA797 }, // LATIN SMALL LETTER B WITH FLOURISH
    Interval{ 0xA799, 0xA799 }, // LATIN SMALL LETTER F WITH STROKE
    Interval{ 0xA79B, 0xA79B }, // LATIN SMALL LETTER VOLAPUK AE
    Interval{ 0xA79D, 0xA79D }, // LATIN SMALL LETTER VOLAPUK OE
    Interval{ 0xA79F, 0xA79F }, // LATIN SMALL LETTER VOLAPUK UE
    Interval{ 0xA7A1, 0xA7A1 }, // LATIN SMALL LETTER G WITH OBLIQUE STROKE
    Interval{ 0xA7A3, 0xA7A3 }, // LATIN SMALL LETTER K WITH OBLIQUE STROKE
    Interval{ 0xA7A5, 0xA7A5 }, // LATIN SMALL LETTER N WITH OBLIQUE STROKE
    Interval{ 0xA7A7, 0xA7A7 }, // LATIN SMALL LETTER R WITH OBLIQUE STROKE
    Interval{ 0xA7A9, 0xA7A9 }, // LATIN SMALL LETTER S WITH OBLIQUE STROKE
    Interval{ 0xA7AF, 0xA7AF }, // LATIN LETTER SMALL CAPITAL Q
    Interval{ 0xA7B5, 0xA7B5 }, // LATIN SMALL LETTER BETA
    Interval{ 0xA7B7, 0xA7B7 }, // LATIN SMALL LETTER OMEGA
    Interval{ 0xA7B9, 0xA7B9 }, // LATIN SMALL LETTER U WITH STROKE
    Interval{ 0xA7BB, 0xA7BB }, // LATIN SMALL LETTER GLOTTAL A
    Interval{ 0xA7BD, 0xA7BD }, // LATIN SMALL LETTER GLOTTAL I
    Interval{ 0xA7BF, 0xA7BF }, // LATIN SMALL LETTER GLOTTAL U
    Interval{ 0xA7C1, 0xA7C1 }, // LATIN SMALL LETTER OLD POLISH O
    Interval{ 0xA7C3, 0xA7C3 }, // LATIN SMALL LETTER ANGLICANA W
    Interval{ 0xA7C8, 0xA7C8 }, // LATIN SMALL LETTER D WITH SHORT STROKE OVERLAY
    Interval{ 0xA7CA, 0xA7CA }, // LATIN SMALL LETTER S WITH SHORT STROKE OVERLAY
    Interval{ 0xA7CD, 0xA7CD }, // LATIN SMALL LETTER S WITH DIAGONAL STROKE
    Interval{ 0xA7CF, 0xA7CF }, // LATIN SMALL LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0xA7D1, 0xA7D1 }, // LATIN SMALL LETTER CLOSED INSULAR G
    Interval{ 0xA7D3, 0xA7D3 }, // LATIN SMALL LETTER DOUBLE THORN
    Interval{ 0xA7D5, 0xA7D5 }, // LATIN SMALL LETTER DOUBLE WYNN
    Interval{ 0xA7D7, 0xA7D7 }, // LATIN SMALL LETTER MIDDLE SCOTS S
    Interval{ 0xA7D9, 0xA7D9 }, // LATIN SMALL LETTER SIGMOID S
    Interval{ 0xA7DB, 0xA7DB }, // LATIN SMALL LETTER LAMBDA
    Interval{ 0xA7F6, 0xA7F6 }, // LATIN SMALL LETTER REVERSED HALF H
    Interval{ 0xA7FA, 0xA7FA }, // LATIN LETTER SMALL CAPITAL TURNED M
    Interval{ 0xAB30, 0xAB5A }, // [43] LATIN SMALL LETTER BARRED ALPHA..LATIN SMALL LETTER Y WITH SHORT RIGHT LEG
    Interval{ 0xAB60, 0xAB68 }, // [9] LATIN SMALL LETTER SAKHA YAT..LATIN SMALL LETTER TURNED R WITH MIDDLE TILDE
    Interval{ 0xAB70, 0xABBF }, // [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Interval{ 0xFB00, 0xFB06 }, // [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Interval{ 0xFB13, 0xFB17 }, // [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Interval{ 0xFF41, 0xFF5A }, // [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Interval{ 0x10428, 0x1044F }, // [40] DESERET SMALL LETTER LONG I..DESERET SMALL LETTER EW
    Interval{ 0x104D8, 0x104FB }, // [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Interval{ 0x10597, 0x105A1 }, // [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Interval{ 0x105A3, 0x105B1 }, // [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Interval{ 0x105B3, 0x105B9 }, // [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Interval{ 0x105BB, 0x105BC }, // [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Interval{ 0x10CC0, 0x10CF2 }, // [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Interval{ 0x10D70, 0x10D85 }, // [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Interval{ 0x118C0, 0x118DF }, // [32] WARANG CITI SMALL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Interval{ 0x16E60, 0x16E7F }, // [32] MEDEFAIDRIN SMALL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Interval{ 0x16EBB, 0x16ED3 }, // [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Interval{ 0x1D41A, 0x1D433 }, // [26] MATHEMATICAL BOLD SMALL A..MATHEMATICAL BOLD SMALL Z
    Interval{ 0x1D44E, 0x1D454 }, // [7] MATHEMATICAL ITALIC SMALL A..MATHEMATICAL ITALIC SMALL G
    Interval{ 0x1D456, 0x1D467 }, // [18] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL ITALIC SMALL Z
    Interval{ 0x1D482, 0x1D49B }, // [26] MATHEMATICAL BOLD ITALIC SMALL A..MATHEMATICAL BOLD ITALIC SMALL Z
    Interval{ 0x1D4B6, 0x1D4B9 }, // [4] MATHEMATICAL SCRIPT SMALL A..MATHEMATICAL SCRIPT SMALL D
    Interval{ 0x1D4BB, 0x1D4BB }, // MATHEMATICAL SCRIPT SMALL F
    Interval{ 0x1D4BD, 0x1D4C3 }, // [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Interval{ 0x1D4C5, 0x1D4CF }, // [11] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL SCRIPT SMALL Z
    Interval{ 0x1D4EA, 0x1D503 }, // [26] MATHEMATICAL BOLD SCRIPT SMALL A..MATHEMATICAL BOLD SCRIPT SMALL Z
    Interval{ 0x1D51E, 0x1D537 }, // [26] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL FRAKTUR SMALL Z
    Interval{ 0x1D552, 0x1D56B }, // [26] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL DOUBLE-STRUCK SMALL Z
    Interval{ 0x1D586, 0x1D59F }, // [26] MATHEMATICAL BOLD FRAKTUR SMALL A..MATHEMATICAL BOLD FRAKTUR SMALL Z
    Interval{ 0x1D5BA, 0x1D5D3 }, // [26] MATHEMATICAL SANS-SERIF SMALL A..MATHEMATICAL SANS-SERIF SMALL Z
    Interval{ 0x1D5EE, 0x1D607 }, // [26] MATHEMATICAL SANS-SERIF BOLD SMALL A..MATHEMATICAL SANS-SERIF BOLD SMALL Z
    Interval{ 0x1D622, 0x1D63B }, // [26] MATHEMATICAL SANS-SERIF ITALIC SMALL A..MATHEMATICAL SANS-SERIF ITALIC SMALL Z
    Interval{ 0x1D656, 0x1D66F }, // [26] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL A..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL Z
    Interval{ 0x1D68A, 0x1D6A5 }, // [28] MATHEMATICAL MONOSPACE SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    Interval{ 0x1D6C2, 0x1D6DA }, // [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    Interval{ 0x1D6DC, 0x1D6E1 }, // [6] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL BOLD PI SYMBOL
    Interval{ 0x1D6FC, 0x1D714 }, // [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    Interval{ 0x1D716, 0x1D71B }, // [6] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL ITALIC PI SYMBOL
    Interval{ 0x1D736, 0x1D74E }, // [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D750, 0x1D755 }, // [6] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC PI SYMBOL
    Interval{ 0x1D770, 0x1D788 }, // [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    Interval{ 0x1D78A, 0x1D78F }, // [6] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD PI SYMBOL
    Interval{ 0x1D7AA, 0x1D7C2 }, // [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    Interval{ 0x1D7C4, 0x1D7C9 }, // [6] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC PI SYMBOL
    Interval{ 0x1D7CB, 0x1D7CB }, // MATHEMATICAL BOLD SMALL DIGAMMA
    Interval{ 0x1DF00, 0x1DF09 }, // [10] LATIN SMALL LETTER FENG DIGRAPH WITH TRILL..LATIN SMALL LETTER T WITH HOOK AND RETROFLEX HOOK
    Interval{ 0x1DF0B, 0x1DF1E }, // [20] LATIN SMALL LETTER ESH WITH DOUBLE BAR..LATIN SMALL LETTER S WITH CURL
    Interval{ 0x1DF25, 0x1DF2A }, // [6] LATIN SMALL LETTER D WITH MID-HEIGHT LEFT HOOK..LATIN SMALL LETTER T WITH MID-HEIGHT LEFT HOOK
    Interval{ 0x1E922, 0x1E943 }, // [34] ADLAM SMALL LETTER ALIF..ADLAM SMALL LETTER SHA
}; // }}}
auto static const Math_Symbol = std::array<Interval, 67>{ // {{{
    Interval{ 0x002B, 0x002B }, // PLUS SIGN
    Interval{ 0x003C, 0x003E }, // [3] LESS-THAN SIGN..GREATER-THAN SIGN
    Interval{ 0x007C, 0x007C }, // VERTICAL LINE
    Interval{ 0x007E, 0x007E }, // TILDE
    Interval{ 0x00AC, 0x00AC }, // NOT SIGN
    Interval{ 0x00B1, 0x00B1 }, // PLUS-MINUS SIGN
    Interval{ 0x00D7, 0x00D7 }, // MULTIPLICATION SIGN
    Interval{ 0x00F7, 0x00F7 }, // DIVISION SIGN
    Interval{ 0x03F6, 0x03F6 }, // GREEK REVERSED LUNATE EPSILON SYMBOL
    Interval{ 0x0606, 0x0608 }, // [3] ARABIC-INDIC CUBE ROOT..ARABIC RAY
    Interval{ 0x2044, 0x2044 }, // FRACTION SLASH
    Interval{ 0x2052, 0x2052 }, // COMMERCIAL MINUS SIGN
    Interval{ 0x207A, 0x207C }, // [3] SUPERSCRIPT PLUS SIGN..SUPERSCRIPT EQUALS SIGN
    Interval{ 0x208A, 0x208C }, // [3] SUBSCRIPT PLUS SIGN..SUBSCRIPT EQUALS SIGN
    Interval{ 0x2118, 0x2118 }, // SCRIPT CAPITAL P
    Interval{ 0x2140, 0x2144 }, // [5] DOUBLE-STRUCK N-ARY SUMMATION..TURNED SANS-SERIF CAPITAL Y
    Interval{ 0x214B, 0x214B }, // TURNED AMPERSAND
    Interval{ 0x2190, 0x2194 }, // [5] LEFTWARDS ARROW..LEFT RIGHT ARROW
    Interval{ 0x219A, 0x219B }, // [2] LEFTWARDS ARROW WITH STROKE..RIGHTWARDS ARROW WITH STROKE
    Interval{ 0x21A0, 0x21A0 }, // RIGHTWARDS TWO HEADED ARROW
    Interval{ 0x21A3, 0x21A3 }, // RIGHTWARDS ARROW WITH TAIL
    Interval{ 0x21A6, 0x21A6 }, // RIGHTWARDS ARROW FROM BAR
    Interval{ 0x21AE, 0x21AE }, // LEFT RIGHT ARROW WITH STROKE
    Interval{ 0x21CE, 0x21CF }, // [2] LEFT RIGHT DOUBLE ARROW WITH STROKE..RIGHTWARDS DOUBLE ARROW WITH STROKE
    Interval{ 0x21D2, 0x21D2 }, // RIGHTWARDS DOUBLE ARROW
    Interval{ 0x21D4, 0x21D4 }, // LEFT RIGHT DOUBLE ARROW
    Interval{ 0x21F4, 0x22FF }, // [268] RIGHT ARROW WITH SMALL CIRCLE..Z NOTATION BAG MEMBERSHIP
    Interval{ 0x2320, 0x2321 }, // [2] TOP HALF INTEGRAL..BOTTOM HALF INTEGRAL
    Interval{ 0x237C, 0x237C }, // RIGHT ANGLE WITH DOWNWARDS ZIGZAG ARROW
    Interval{ 0x239B, 0x23B3 }, // [25] LEFT PARENTHESIS UPPER HOOK..SUMMATION BOTTOM
    Interval{ 0x23DC, 0x23E1 }, // [6] TOP PARENTHESIS..BOTTOM TORTOISE SHELL BRACKET
    Interval{ 0x25B7, 0x25B7 }, // WHITE RIGHT-POINTING TRIANGLE
    Interval{ 0x25C1, 0x25C1 }, // WHITE LEFT-POINTING TRIANGLE
    Interval{ 0x25F8, 0x25FF }, // [8] UPPER LEFT TRIANGLE..LOWER RIGHT TRIANGLE
    Interval{ 0x266F, 0x266F }, // MUSIC SHARP SIGN
    Interval{ 0x27C0, 0x27C4 }, // [5] THREE DIMENSIONAL ANGLE..OPEN SUPERSET
    Interval{ 0x27C7, 0x27E5 }, // [31] OR WITH DOT INSIDE..WHITE SQUARE WITH RIGHTWARDS TICK
    Interval{ 0x27F0, 0x27FF }, // [16] UPWARDS QUADRUPLE ARROW..LONG RIGHTWARDS SQUIGGLE ARROW
    Interval{ 0x2900, 0x2982 }, // [131] RIGHTWARDS TWO-HEADED ARROW WITH VERTICAL STROKE..Z NOTATION TYPE COLON
    Interval{ 0x2999, 0x29D7 }, // [63] DOTTED FENCE..BLACK HOURGLASS
    Interval{ 0x29DC, 0x29FB }, // [32] INCOMPLETE INFINITY..TRIPLE PLUS
    Interval{ 0x29FE, 0x2AFF }, // [258] TINY..N-ARY WHITE VERTICAL BAR
    Interval{ 0x2B30, 0x2B44 }, // [21] LEFT ARROW WITH SMALL CIRCLE..RIGHTWARDS ARROW THROUGH SUPERSET
    Interval{ 0x2B47, 0x2B4C }, // [6] REVERSE TILDE OPERATOR ABOVE RIGHTWARDS ARROW..RIGHTWARDS ARROW ABOVE REVERSE TILDE OPERATOR
    Interval{ 0xFB29, 0xFB29 }, // HEBREW LETTER ALTERNATIVE PLUS SIGN
    Interval{ 0xFE62, 0xFE62 }, // SMALL PLUS SIGN
    Interval{ 0xFE64, 0xFE66 }, // [3] SMALL LESS-THAN SIGN..SMALL EQUALS SIGN
    Interval{ 0xFF0B, 0xFF0B }, // FULLWIDTH PLUS SIGN
    Interval{ 0xFF1C, 0xFF1E }, // [3] FULLWIDTH LESS-THAN SIGN..FULLWIDTH GREATER-THAN SIGN
    Interval{ 0xFF5C, 0xFF5C }, // FULLWIDTH VERTICAL LINE
    Interval{ 0xFF5E, 0xFF5E }, // FULLWIDTH TILDE
    Interval{ 0xFFE2, 0xFFE2 }, // FULLWIDTH NOT SIGN
    Interval{ 0xFFE9, 0xFFEC }, // [4] HALFWIDTH LEFTWARDS ARROW..HALFWIDTH DOWNWARDS ARROW
    Interval{ 0x10D8E, 0x10D8F }, // [2] GARAY PLUS SIGN..GARAY MINUS SIGN
    Interval{ 0x1CEF0, 0x1CEF0 }, // MEDIUM SMALL WHITE CIRCLE WITH HORIZONTAL BAR
    Interval{ 0x1D6C1, 0x1D6C1 }, // MATHEMATICAL BOLD NABLA
    Interval{ 0x1D6DB, 0x1D6DB }, // MATHEMATICAL BOLD PARTIAL DIFFERENTIAL
    Interval{ 0x1D6FB, 0x1D6FB }, // MATHEMATICAL ITALIC NABLA
    Interval{ 0x1D715, 0x1D715 }, // MATHEMATICAL ITALIC PARTIAL DIFFERENTIAL
    Interval{ 0x1D735, 0x1D735 }, // MATHEMATICAL BOLD ITALIC NABLA
    Interval{ 0x1D74F, 0x1D74F }, // MATHEMATICAL BOLD ITALIC PARTIAL DIFFERENTIAL
    Interval{ 0x1D76F, 0x1D76F }, // MATHEMATICAL SANS-SERIF BOLD NABLA
    Interval{ 0x1D789, 0x1D789 }, // MATHEMATICAL SANS-SERIF BOLD PARTIAL DIFFERENTIAL
    Interval{ 0x1D7A9, 0x1D7A9 }, // MATHEMATICAL SANS-SERIF BOLD ITALIC NABLA
    Interval{ 0x1D7C3, 0x1D7C3 }, // MATHEMATICAL SANS-SERIF BOLD ITALIC PARTIAL DIFFERENTIAL
    Interval{ 0x1EEF0, 0x1EEF1 }, // [2] ARABIC MATHEMATICAL OPERATOR MEEM WITH HAH WITH TATWEEL..ARABIC MATHEMATICAL OPERATOR HAH WITH DAL
    Interval{ 0x1F8D0, 0x1F8D8 }, // [9] LONG RIGHTWARDS ARROW OVER LONG LEFTWARDS ARROW..LONG LEFT RIGHT ARROW WITH DEPENDENT LOBE
}; // }}}
auto static const Modifier_Letter = std::array<Interval, 79>{ // {{{
    Interval{ 0x02B0, 0x02C1 }, // [18] MODIFIER LETTER SMALL H..MODIFIER LETTER REVERSED GLOTTAL STOP
    Interval{ 0x02C6, 0x02D1 }, // [12] MODIFIER LETTER CIRCUMFLEX ACCENT..MODIFIER LETTER HALF TRIANGULAR COLON
    Interval{ 0x02E0, 0x02E4 }, // [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    Interval{ 0x02EC, 0x02EC }, // MODIFIER LETTER VOICING
    Interval{ 0x02EE, 0x02EE }, // MODIFIER LETTER DOUBLE APOSTROPHE
    Interval{ 0x0374, 0x0374 }, // GREEK NUMERAL SIGN
    Interval{ 0x037A, 0x037A }, // GREEK YPOGEGRAMMENI
    Interval{ 0x0559, 0x0559 }, // ARMENIAN MODIFIER LETTER LEFT HALF RING
    Interval{ 0x0640, 0x0640 }, // ARABIC TATWEEL
    Interval{ 0x06E5, 0x06E6 }, // [2] ARABIC SMALL WAW..ARABIC SMALL YEH
    Interval{ 0x07F4, 0x07F5 }, // [2] NKO HIGH TONE APOSTROPHE..NKO LOW TONE APOSTROPHE
    Interval{ 0x07FA, 0x07FA }, // NKO LAJANYALAN
    Interval{ 0x081A, 0x081A }, // SAMARITAN MODIFIER LETTER EPENTHETIC YUT
    Interval{ 0x0824, 0x0824 }, // SAMARITAN MODIFIER LETTER SHORT A
    Interval{ 0x0828, 0x0828 }, // SAMARITAN MODIFIER LETTER I
    Interval{ 0x08C9, 0x08C9 }, // ARABIC SMALL FARSI YEH
    Interval{ 0x0971, 0x0971 }, // DEVANAGARI SIGN HIGH SPACING DOT
    Interval{ 0x0E46, 0x0E46 }, // THAI CHARACTER MAIYAMOK
    Interval{ 0x0EC6, 0x0EC6 }, // LAO KO LA
    Interval{ 0x10FC, 0x10FC }, // MODIFIER LETTER GEORGIAN NAR
    Interval{ 0x17D7, 0x17D7 }, // KHMER SIGN LEK TOO
    Interval{ 0x1843, 0x1843 }, // MONGOLIAN LETTER TODO LONG VOWEL SIGN
    Interval{ 0x1AA7, 0x1AA7 }, // TAI THAM SIGN MAI YAMOK
    Interval{ 0x1C78, 0x1C7D }, // [6] OL CHIKI MU TTUDDAG..OL CHIKI AHAD
    Interval{ 0x1D2C, 0x1D6A }, // [63] MODIFIER LETTER CAPITAL A..GREEK SUBSCRIPT SMALL LETTER CHI
    Interval{ 0x1D78, 0x1D78 }, // MODIFIER LETTER CYRILLIC EN
    Interval{ 0x1D9B, 0x1DBF }, // [37] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL THETA
    Interval{ 0x2071, 0x2071 }, // SUPERSCRIPT LATIN SMALL LETTER I
    Interval{ 0x207F, 0x207F }, // SUPERSCRIPT LATIN SMALL LETTER N
    Interval{ 0x2090, 0x209C }, // [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Interval{ 0x2C7C, 0x2C7D }, // [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    Interval{ 0x2D6F, 0x2D6F }, // TIFINAGH MODIFIER LETTER LABIALIZATION MARK
    Interval{ 0x2E2F, 0x2E2F }, // VERTICAL TILDE
    Interval{ 0x3005, 0x3005 }, // IDEOGRAPHIC ITERATION MARK
    Interval{ 0x3031, 0x3035 }, // [5] VERTICAL KANA REPEAT MARK..VERTICAL KANA REPEAT MARK LOWER HALF
    Interval{ 0x303B, 0x303B }, // VERTICAL IDEOGRAPHIC ITERATION MARK
    Interval{ 0x309D, 0x309E }, // [2] HIRAGANA ITERATION MARK..HIRAGANA VOICED ITERATION MARK
    Interval{ 0x30FC, 0x30FE }, // [3] KATAKANA-HIRAGANA PROLONGED SOUND MARK..KATAKANA VOICED ITERATION MARK
    Interval{ 0xA015, 0xA015 }, // YI SYLLABLE WU
    Interval{ 0xA4F8, 0xA4FD }, // [6] LISU LETTER TONE MYA TI..LISU LETTER TONE MYA JEU
    Interval{ 0xA60C, 0xA60C }, // VAI SYLLABLE LENGTHENER
    Interval{ 0xA67F, 0xA67F }, // CYRILLIC PAYEROK
    Interval{ 0xA69C, 0xA69D }, // [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    Interval{ 0xA717, 0xA71F }, // [9] MODIFIER LETTER DOT VERTICAL BAR..MODIFIER LETTER LOW INVERTED EXCLAMATION MARK
    Interval{ 0xA770, 0xA770 }, // MODIFIER LETTER US
    Interval{ 0xA788, 0xA788 }, // MODIFIER LETTER LOW CIRCUMFLEX ACCENT
    Interval{ 0xA7F1, 0xA7F4 }, // [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    Interval{ 0xA7F8, 0xA7F9 }, // [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    Interval{ 0xA9CF, 0xA9CF }, // JAVANESE PANGRANGKEP
    Interval{ 0xA9E6, 0xA9E6 }, // MYANMAR MODIFIER LETTER SHAN REDUPLICATION
    Interval{ 0xAA70, 0xAA70 }, // MYANMAR MODIFIER LETTER KHAMTI REDUPLICATION
    Interval{ 0xAADD, 0xAADD }, // TAI VIET SYMBOL SAM
    Interval{ 0xAAF3, 0xAAF4 }, // [2] MEETEI MAYEK SYLLABLE REPETITION MARK..MEETEI MAYEK WORD REPETITION MARK
    Interval{ 0xAB5C, 0xAB5F }, // [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    Interval{ 0xAB69, 0xAB69 }, // MODIFIER LETTER SMALL TURNED W
    Interval{ 0xFF70, 0xFF70 }, // HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Interval{ 0xFF9E, 0xFF9F }, // [2] HALFWIDTH KATAKANA VOICED SOUND MARK..HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK
    Interval{ 0x10780, 0x10785 }, // [6] MODIFIER LETTER SMALL CAPITAL AA..MODIFIER LETTER SMALL B WITH HOOK
    Interval{ 0x10787, 0x107B0 }, // [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Interval{ 0x107B2, 0x107BA }, // [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Interval{ 0x10D4E, 0x10D4E }, // GARAY VOWEL LENGTH MARK
    Interval{ 0x10D6F, 0x10D6F }, // GARAY REDUPLICATION MARK
    Interval{ 0x10EC5, 0x10EC5 }, // ARABIC SMALL YEH BARREE WITH TWO DOTS BELOW
    Interval{ 0x11DD9, 0x11DD9 }, // TOLONG SIKI SIGN SELA
    Interval{ 0x16B40, 0x16B43 }, // [4] PAHAWH HMONG SIGN VOS SEEV..PAHAWH HMONG SIGN IB YAM
    Interval{ 0x16D40, 0x16D42 }, // [3] KIRAT RAI SIGN ANUSVARA..KIRAT RAI SIGN VISARGA
    Interval{ 0x16D6B, 0x16D6C }, // [2] KIRAT RAI SIGN VIRAMA..KIRAT RAI SIGN SAAT
    Interval{ 0x16F93, 0x16F9F }, // [13] MIAO LETTER TONE-2..MIAO LETTER REFORMED TONE-8
    Interval{ 0x16FE0, 0x16FE1 }, // [2] TANGUT ITERATION MARK..NUSHU ITERATION MARK
    Interval{ 0x16FE3, 0x16FE3 }, // OLD CHINESE ITERATION MARK
    Interval{ 0x16FF2, 0x16FF3 }, // [2] CHINESE SMALL SIMPLIFIED ER..CHINESE SMALL TRADITIONAL ER
    Interval{ 0x1AFF0, 0x1AFF3 }, // [4] KATAKANA LETTER MINNAN TONE-2..KATAKANA LETTER MINNAN TONE-5
    Interval{ 0x1AFF5, 0x1AFFB }, // [7] KATAKANA LETTER MINNAN TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-5
    Interval{ 0x1AFFD, 0x1AFFE }, // [2] KATAKANA LETTER MINNAN NASALIZED TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-8
    Interval{ 0x1E030, 0x1E06D }, // [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Interval{ 0x1E137, 0x1E13D }, // [7] NYIAKENG PUACHUE HMONG SIGN FOR PERSON..NYIAKENG PUACHUE HMONG SYLLABLE LENGTHENER
    Interval{ 0x1E4EB, 0x1E4EB }, // NAG MUNDARI SIGN OJOD
    Interval{ 0x1E6FF, 0x1E6FF }, // TAI YO XAM LAI
    Interval{ 0x1E94B, 0x1E94B }, // ADLAM NASALIZATION MARK
}; // }}}
auto static const Modifier_Symbol = std::array<Interval, 31>{ // {{{
    Interval{ 0x005E, 0x005E }, // CIRCUMFLEX ACCENT
    Interval{ 0x0060, 0x0060 }, // GRAVE ACCENT
    Interval{ 0x00A8, 0x00A8 }, // DIAERESIS
    Interval{ 0x00AF, 0x00AF }, // MACRON
    Interval{ 0x00B4, 0x00B4 }, // ACUTE ACCENT
    Interval{ 0x00B8, 0x00B8 }, // CEDILLA
    Interval{ 0x02C2, 0x02C5 }, // [4] MODIFIER LETTER LEFT ARROWHEAD..MODIFIER LETTER DOWN ARROWHEAD
    Interval{ 0x02D2, 0x02DF }, // [14] MODIFIER LETTER CENTRED RIGHT HALF RING..MODIFIER LETTER CROSS ACCENT
    Interval{ 0x02E5, 0x02EB }, // [7] MODIFIER LETTER EXTRA-HIGH TONE BAR..MODIFIER LETTER YANG DEPARTING TONE MARK
    Interval{ 0x02ED, 0x02ED }, // MODIFIER LETTER UNASPIRATED
    Interval{ 0x02EF, 0x02FF }, // [17] MODIFIER LETTER LOW DOWN ARROWHEAD..MODIFIER LETTER LOW LEFT ARROW
    Interval{ 0x0375, 0x0375 }, // GREEK LOWER NUMERAL SIGN
    Interval{ 0x0384, 0x0385 }, // [2] GREEK TONOS..GREEK DIALYTIKA TONOS
    Interval{ 0x0888, 0x0888 }, // ARABIC RAISED ROUND DOT
    Interval{ 0x1FBD, 0x1FBD }, // GREEK KORONIS
    Interval{ 0x1FBF, 0x1FC1 }, // [3] GREEK PSILI..GREEK DIALYTIKA AND PERISPOMENI
    Interval{ 0x1FCD, 0x1FCF }, // [3] GREEK PSILI AND VARIA..GREEK PSILI AND PERISPOMENI
    Interval{ 0x1FDD, 0x1FDF }, // [3] GREEK DASIA AND VARIA..GREEK DASIA AND PERISPOMENI
    Interval{ 0x1FED, 0x1FEF }, // [3] GREEK DIALYTIKA AND VARIA..GREEK VARIA
    Interval{ 0x1FFD, 0x1FFE }, // [2] GREEK OXIA..GREEK DASIA
    Interval{ 0x309B, 0x309C }, // [2] KATAKANA-HIRAGANA VOICED SOUND MARK..KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Interval{ 0xA700, 0xA716 }, // [23] MODIFIER LETTER CHINESE TONE YIN PING..MODIFIER LETTER EXTRA-LOW LEFT-STEM TONE BAR
    Interval{ 0xA720, 0xA721 }, // [2] MODIFIER LETTER STRESS AND HIGH TONE..MODIFIER LETTER STRESS AND LOW TONE
    Interval{ 0xA789, 0xA78A }, // [2] MODIFIER LETTER COLON..MODIFIER LETTER SHORT EQUALS SIGN
    Interval{ 0xAB5B, 0xAB5B }, // MODIFIER BREVE WITH INVERTED BREVE
    Interval{ 0xAB6A, 0xAB6B }, // [2] MODIFIER LETTER LEFT TACK..MODIFIER LETTER RIGHT TACK
    Interval{ 0xFBB2, 0xFBC2 }, // [17] ARABIC SYMBOL DOT ABOVE..ARABIC SYMBOL WASLA ABOVE
    Interval{ 0xFF3E, 0xFF3E }, // FULLWIDTH CIRCUMFLEX ACCENT
    Interval{ 0xFF40, 0xFF40 }, // FULLWIDTH GRAVE ACCENT
    Interval{ 0xFFE3, 0xFFE3 }, // FULLWIDTH MACRON
    Interval{ 0x1F3FB, 0x1F3FF }, // [5] EMOJI MODIFIER FITZPATRICK TYPE-1-2..EMOJI MODIFIER FITZPATRICK TYPE-6
}; // }}}
auto static const Nonspacing_Mark = std::array<Interval, 365>{ // {{{
    Interval{ 0x0300, 0x036F }, // [112] COMBINING GRAVE ACCENT..COMBINING LATIN SMALL LETTER X
    Interval{ 0x0483, 0x0487 }, // [5] COMBINING CYRILLIC TITLO..COMBINING CYRILLIC POKRYTIE
    Interval{ 0x0591, 0x05BD }, // [45] HEBREW ACCENT ETNAHTA..HEBREW POINT METEG
    Interval{ 0x05BF, 0x05BF }, // HEBREW POINT RAFE
    Interval{ 0x05C1, 0x05C2 }, // [2] HEBREW POINT SHIN DOT..HEBREW POINT SIN DOT
    Interval{ 0x05C4, 0x05C5 }, // [2] HEBREW MARK UPPER DOT..HEBREW MARK LOWER DOT
    Interval{ 0x05C7, 0x05C7 }, // HEBREW POINT QAMATS QATAN
    Interval{ 0x0610, 0x061A }, // [11] ARABIC SIGN SALLALLAHOU ALAYHE WASSALLAM..ARABIC SMALL KASRA
    Interval{ 0x064B, 0x065F }, // [21] ARABIC FATHATAN..ARABIC WAVY HAMZA BELOW
    Interval{ 0x0670, 0x0670 }, // ARABIC LETTER SUPERSCRIPT ALEF
    Interval{ 0x06D6, 0x06DC }, // [7] ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA..ARABIC SMALL HIGH SEEN
    Interval{ 0x06DF, 0x06E4 }, // [6] ARABIC SMALL HIGH ROUNDED ZERO..ARABIC SMALL HIGH MADDA
    Interval{ 0x06E7, 0x06E8 }, // [2] ARABIC SMALL HIGH YEH..ARABIC SMALL HIGH NOON
    Interval{ 0x06EA, 0x06ED }, // [4] ARABIC EMPTY CENTRE LOW STOP..ARABIC SMALL LOW MEEM
    Interval{ 0x0711, 0x0711 }, // SYRIAC LETTER SUPERSCRIPT ALAPH
    Interval{ 0x0730, 0x074A }, // [27] SYRIAC PTHAHA ABOVE..SYRIAC BARREKH
    Interval{ 0x07A6, 0x07B0 }, // [11] THAANA ABAFILI..THAANA SUKUN
    Interval{ 0x07EB, 0x07F3 }, // [9] NKO COMBINING SHORT HIGH TONE..NKO COMBINING DOUBLE DOT ABOVE
    Interval{ 0x07FD, 0x07FD }, // NKO DANTAYALAN
    Interval{ 0x0816, 0x0819 }, // [4] SAMARITAN MARK IN..SAMARITAN MARK DAGESH
    Interval{ 0x081B, 0x0823 }, // [9] SAMARITAN MARK EPENTHETIC YUT..SAMARITAN VOWEL SIGN A
    Interval{ 0x0825, 0x0827 }, // [3] SAMARITAN VOWEL SIGN SHORT A..SAMARITAN VOWEL SIGN U
    Interval{ 0x0829, 0x082D }, // [5] SAMARITAN VOWEL SIGN LONG I..SAMARITAN MARK NEQUDAA
    Interval{ 0x0859, 0x085B }, // [3] MANDAIC AFFRICATION MARK..MANDAIC GEMINATION MARK
    Interval{ 0x0897, 0x089F }, // [9] ARABIC PEPET..ARABIC HALF MADDA OVER MADDA
    Interval{ 0x08CA, 0x08E1 }, // [24] ARABIC SMALL HIGH FARSI YEH..ARABIC SMALL HIGH SIGN SAFHA
    Interval{ 0x08E3, 0x0902 }, // [32] ARABIC TURNED DAMMA BELOW..DEVANAGARI SIGN ANUSVARA
    Interval{ 0x093A, 0x093A }, // DEVANAGARI VOWEL SIGN OE
    Interval{ 0x093C, 0x093C }, // DEVANAGARI SIGN NUKTA
    Interval{ 0x0941, 0x0948 }, // [8] DEVANAGARI VOWEL SIGN U..DEVANAGARI VOWEL SIGN AI
    Interval{ 0x094D, 0x094D }, // DEVANAGARI SIGN VIRAMA
    Interval{ 0x0951, 0x0957 }, // [7] DEVANAGARI STRESS SIGN UDATTA..DEVANAGARI VOWEL SIGN UUE
    Interval{ 0x0962, 0x0963 }, // [2] DEVANAGARI VOWEL SIGN VOCALIC L..DEVANAGARI VOWEL SIGN VOCALIC LL
    Interval{ 0x0981, 0x0981 }, // BENGALI SIGN CANDRABINDU
    Interval{ 0x09BC, 0x09BC }, // BENGALI SIGN NUKTA
    Interval{ 0x09C1, 0x09C4 }, // [4] BENGALI VOWEL SIGN U..BENGALI VOWEL SIGN VOCALIC RR
    Interval{ 0x09CD, 0x09CD }, // BENGALI SIGN VIRAMA
    Interval{ 0x09E2, 0x09E3 }, // [2] BENGALI VOWEL SIGN VOCALIC L..BENGALI VOWEL SIGN VOCALIC LL
    Interval{ 0x09FE, 0x09FE }, // BENGALI SANDHI MARK
    Interval{ 0x0A01, 0x0A02 }, // [2] GURMUKHI SIGN ADAK BINDI..GURMUKHI SIGN BINDI
    Interval{ 0x0A3C, 0x0A3C }, // GURMUKHI SIGN NUKTA
    Interval{ 0x0A41, 0x0A42 }, // [2] GURMUKHI VOWEL SIGN U..GURMUKHI VOWEL SIGN UU
    Interval{ 0x0A47, 0x0A48 }, // [2] GURMUKHI VOWEL SIGN EE..GURMUKHI VOWEL SIGN AI
    Interval{ 0x0A4B, 0x0A4D }, // [3] GURMUKHI VOWEL SIGN OO..GURMUKHI SIGN VIRAMA
    Interval{ 0x0A51, 0x0A51 }, // GURMUKHI SIGN UDAAT
    Interval{ 0x0A70, 0x0A71 }, // [2] GURMUKHI TIPPI..GURMUKHI ADDAK
    Interval{ 0x0A75, 0x0A75 }, // GURMUKHI SIGN YAKASH
    Interval{ 0x0A81, 0x0A82 }, // [2] GUJARATI SIGN CANDRABINDU..GUJARATI SIGN ANUSVARA
    Interval{ 0x0ABC, 0x0ABC }, // GUJARATI SIGN NUKTA
    Interval{ 0x0AC1, 0x0AC5 }, // [5] GUJARATI VOWEL SIGN U..GUJARATI VOWEL SIGN CANDRA E
    Interval{ 0x0AC7, 0x0AC8 }, // [2] GUJARATI VOWEL SIGN E..GUJARATI VOWEL SIGN AI
    Interval{ 0x0ACD, 0x0ACD }, // GUJARATI SIGN VIRAMA
    Interval{ 0x0AE2, 0x0AE3 }, // [2] GUJARATI VOWEL SIGN VOCALIC L..GUJARATI VOWEL SIGN VOCALIC LL
    Interval{ 0x0AFA, 0x0AFF }, // [6] GUJARATI SIGN SUKUN..GUJARATI SIGN TWO-CIRCLE NUKTA ABOVE
    Interval{ 0x0B01, 0x0B01 }, // ORIYA SIGN CANDRABINDU
    Interval{ 0x0B3C, 0x0B3C }, // ORIYA SIGN NUKTA
    Interval{ 0x0B3F, 0x0B3F }, // ORIYA VOWEL SIGN I
    Interval{ 0x0B41, 0x0B44 }, // [4] ORIYA VOWEL SIGN U..ORIYA VOWEL SIGN VOCALIC RR
    Interval{ 0x0B4D, 0x0B4D }, // ORIYA SIGN VIRAMA
    Interval{ 0x0B55, 0x0B56 }, // [2] ORIYA SIGN OVERLINE..ORIYA AI LENGTH MARK
    Interval{ 0x0B62, 0x0B63 }, // [2] ORIYA VOWEL SIGN VOCALIC L..ORIYA VOWEL SIGN VOCALIC LL
    Interval{ 0x0B82, 0x0B82 }, // TAMIL SIGN ANUSVARA
    Interval{ 0x0BC0, 0x0BC0 }, // TAMIL VOWEL SIGN II
    Interval{ 0x0BCD, 0x0BCD }, // TAMIL SIGN VIRAMA
    Interval{ 0x0C00, 0x0C00 }, // TELUGU SIGN COMBINING CANDRABINDU ABOVE
    Interval{ 0x0C04, 0x0C04 }, // TELUGU SIGN COMBINING ANUSVARA ABOVE
    Interval{ 0x0C3C, 0x0C3C }, // TELUGU SIGN NUKTA
    Interval{ 0x0C3E, 0x0C40 }, // [3] TELUGU VOWEL SIGN AA..TELUGU VOWEL SIGN II
    Interval{ 0x0C46, 0x0C48 }, // [3] TELUGU VOWEL SIGN E..TELUGU VOWEL SIGN AI
    Interval{ 0x0C4A, 0x0C4D }, // [4] TELUGU VOWEL SIGN O..TELUGU SIGN VIRAMA
    Interval{ 0x0C55, 0x0C56 }, // [2] TELUGU LENGTH MARK..TELUGU AI LENGTH MARK
    Interval{ 0x0C62, 0x0C63 }, // [2] TELUGU VOWEL SIGN VOCALIC L..TELUGU VOWEL SIGN VOCALIC LL
    Interval{ 0x0C81, 0x0C81 }, // KANNADA SIGN CANDRABINDU
    Interval{ 0x0CBC, 0x0CBC }, // KANNADA SIGN NUKTA
    Interval{ 0x0CBF, 0x0CBF }, // KANNADA VOWEL SIGN I
    Interval{ 0x0CC6, 0x0CC6 }, // KANNADA VOWEL SIGN E
    Interval{ 0x0CCC, 0x0CCD }, // [2] KANNADA VOWEL SIGN AU..KANNADA SIGN VIRAMA
    Interval{ 0x0CE2, 0x0CE3 }, // [2] KANNADA VOWEL SIGN VOCALIC L..KANNADA VOWEL SIGN VOCALIC LL
    Interval{ 0x0D00, 0x0D01 }, // [2] MALAYALAM SIGN COMBINING ANUSVARA ABOVE..MALAYALAM SIGN CANDRABINDU
    Interval{ 0x0D3B, 0x0D3C }, // [2] MALAYALAM SIGN VERTICAL BAR VIRAMA..MALAYALAM SIGN CIRCULAR VIRAMA
    Interval{ 0x0D41, 0x0D44 }, // [4] MALAYALAM VOWEL SIGN U..MALAYALAM VOWEL SIGN VOCALIC RR
    Interval{ 0x0D4D, 0x0D4D }, // MALAYALAM SIGN VIRAMA
    Interval{ 0x0D62, 0x0D63 }, // [2] MALAYALAM VOWEL SIGN VOCALIC L..MALAYALAM VOWEL SIGN VOCALIC LL
    Interval{ 0x0D81, 0x0D81 }, // SINHALA SIGN CANDRABINDU
    Interval{ 0x0DCA, 0x0DCA }, // SINHALA SIGN AL-LAKUNA
    Interval{ 0x0DD2, 0x0DD4 }, // [3] SINHALA VOWEL SIGN KETTI IS-PILLA..SINHALA VOWEL SIGN KETTI PAA-PILLA
    Interval{ 0x0DD6, 0x0DD6 }, // SINHALA VOWEL SIGN DIGA PAA-PILLA
    Interval{ 0x0E31, 0x0E31 }, // THAI CHARACTER MAI HAN-AKAT
    Interval{ 0x0E34, 0x0E3A }, // [7] THAI CHARACTER SARA I..THAI CHARACTER PHINTHU
    Interval{ 0x0E47, 0x0E4E }, // [8] THAI CHARACTER MAITAIKHU..THAI CHARACTER YAMAKKAN
    Interval{ 0x0EB1, 0x0EB1 }, // LAO VOWEL SIGN MAI KAN
    Interval{ 0x0EB4, 0x0EBC }, // [9] LAO VOWEL SIGN I..LAO SEMIVOWEL SIGN LO
    Interval{ 0x0EC8, 0x0ECE }, // [7] LAO TONE MAI EK..LAO YAMAKKAN
    Interval{ 0x0F18, 0x0F19 }, // [2] TIBETAN ASTROLOGICAL SIGN -KHYUD PA..TIBETAN ASTROLOGICAL SIGN SDONG TSHUGS
    Interval{ 0x0F35, 0x0F35 }, // TIBETAN MARK NGAS BZUNG NYI ZLA
    Interval{ 0x0F37, 0x0F37 }, // TIBETAN MARK NGAS BZUNG SGOR RTAGS
    Interval{ 0x0F39, 0x0F39 }, // TIBETAN MARK TSA -PHRU
    Interval{ 0x0F71, 0x0F7E }, // [14] TIBETAN VOWEL SIGN AA..TIBETAN SIGN RJES SU NGA RO
    Interval{ 0x0F80, 0x0F84 }, // [5] TIBETAN VOWEL SIGN REVERSED I..TIBETAN MARK HALANTA
    Interval{ 0x0F86, 0x0F87 }, // [2] TIBETAN SIGN LCI RTAGS..TIBETAN SIGN YANG RTAGS
    Interval{ 0x0F8D, 0x0F97 }, // [11] TIBETAN SUBJOINED SIGN LCE TSA CAN..TIBETAN SUBJOINED LETTER JA
    Interval{ 0x0F99, 0x0FBC }, // [36] TIBETAN SUBJOINED LETTER NYA..TIBETAN SUBJOINED LETTER FIXED-FORM RA
    Interval{ 0x0FC6, 0x0FC6 }, // TIBETAN SYMBOL PADMA GDAN
    Interval{ 0x102D, 0x1030 }, // [4] MYANMAR VOWEL SIGN I..MYANMAR VOWEL SIGN UU
    Interval{ 0x1032, 0x1037 }, // [6] MYANMAR VOWEL SIGN AI..MYANMAR SIGN DOT BELOW
    Interval{ 0x1039, 0x103A }, // [2] MYANMAR SIGN VIRAMA..MYANMAR SIGN ASAT
    Interval{ 0x103D, 0x103E }, // [2] MYANMAR CONSONANT SIGN MEDIAL WA..MYANMAR CONSONANT SIGN MEDIAL HA
    Interval{ 0x1058, 0x1059 }, // [2] MYANMAR VOWEL SIGN VOCALIC L..MYANMAR VOWEL SIGN VOCALIC LL
    Interval{ 0x105E, 0x1060 }, // [3] MYANMAR CONSONANT SIGN MON MEDIAL NA..MYANMAR CONSONANT SIGN MON MEDIAL LA
    Interval{ 0x1071, 0x1074 }, // [4] MYANMAR VOWEL SIGN GEBA KAREN I..MYANMAR VOWEL SIGN KAYAH EE
    Interval{ 0x1082, 0x1082 }, // MYANMAR CONSONANT SIGN SHAN MEDIAL WA
    Interval{ 0x1085, 0x1086 }, // [2] MYANMAR VOWEL SIGN SHAN E ABOVE..MYANMAR VOWEL SIGN SHAN FINAL Y
    Interval{ 0x108D, 0x108D }, // MYANMAR SIGN SHAN COUNCIL EMPHATIC TONE
    Interval{ 0x109D, 0x109D }, // MYANMAR VOWEL SIGN AITON AI
    Interval{ 0x135D, 0x135F }, // [3] ETHIOPIC COMBINING GEMINATION AND VOWEL LENGTH MARK..ETHIOPIC COMBINING GEMINATION MARK
    Interval{ 0x1712, 0x1714 }, // [3] TAGALOG VOWEL SIGN I..TAGALOG SIGN VIRAMA
    Interval{ 0x1732, 0x1733 }, // [2] HANUNOO VOWEL SIGN I..HANUNOO VOWEL SIGN U
    Interval{ 0x1752, 0x1753 }, // [2] BUHID VOWEL SIGN I..BUHID VOWEL SIGN U
    Interval{ 0x1772, 0x1773 }, // [2] TAGBANWA VOWEL SIGN I..TAGBANWA VOWEL SIGN U
    Interval{ 0x17B4, 0x17B5 }, // [2] KHMER VOWEL INHERENT AQ..KHMER VOWEL INHERENT AA
    Interval{ 0x17B7, 0x17BD }, // [7] KHMER VOWEL SIGN I..KHMER VOWEL SIGN UA
    Interval{ 0x17C6, 0x17C6 }, // KHMER SIGN NIKAHIT
    Interval{ 0x17C9, 0x17D3 }, // [11] KHMER SIGN MUUSIKATOAN..KHMER SIGN BATHAMASAT
    Interval{ 0x17DD, 0x17DD }, // KHMER SIGN ATTHACAN
    Interval{ 0x180B, 0x180D }, // [3] MONGOLIAN FREE VARIATION SELECTOR ONE..MONGOLIAN FREE VARIATION SELECTOR THREE
    Interval{ 0x180F, 0x180F }, // MONGOLIAN FREE VARIATION SELECTOR FOUR
    Interval{ 0x1885, 0x1886 }, // [2] MONGOLIAN LETTER ALI GALI BALUDA..MONGOLIAN LETTER ALI GALI THREE BALUDA
    Interval{ 0x18A9, 0x18A9 }, // MONGOLIAN LETTER ALI GALI DAGALGA
    Interval{ 0x1920, 0x1922 }, // [3] LIMBU VOWEL SIGN A..LIMBU VOWEL SIGN U
    Interval{ 0x1927, 0x1928 }, // [2] LIMBU VOWEL SIGN E..LIMBU VOWEL SIGN O
    Interval{ 0x1932, 0x1932 }, // LIMBU SMALL LETTER ANUSVARA
    Interval{ 0x1939, 0x193B }, // [3] LIMBU SIGN MUKPHRENG..LIMBU SIGN SA-I
    Interval{ 0x1A17, 0x1A18 }, // [2] BUGINESE VOWEL SIGN I..BUGINESE VOWEL SIGN U
    Interval{ 0x1A1B, 0x1A1B }, // BUGINESE VOWEL SIGN AE
    Interval{ 0x1A56, 0x1A56 }, // TAI THAM CONSONANT SIGN MEDIAL LA
    Interval{ 0x1A58, 0x1A5E }, // [7] TAI THAM SIGN MAI KANG LAI..TAI THAM CONSONANT SIGN SA
    Interval{ 0x1A60, 0x1A60 }, // TAI THAM SIGN SAKOT
    Interval{ 0x1A62, 0x1A62 }, // TAI THAM VOWEL SIGN MAI SAT
    Interval{ 0x1A65, 0x1A6C }, // [8] TAI THAM VOWEL SIGN I..TAI THAM VOWEL SIGN OA BELOW
    Interval{ 0x1A73, 0x1A7C }, // [10] TAI THAM VOWEL SIGN OA ABOVE..TAI THAM SIGN KHUEN-LUE KARAN
    Interval{ 0x1A7F, 0x1A7F }, // TAI THAM COMBINING CRYPTOGRAMMIC DOT
    Interval{ 0x1AB0, 0x1ABD }, // [14] COMBINING DOUBLED CIRCUMFLEX ACCENT..COMBINING PARENTHESES BELOW
    Interval{ 0x1ABF, 0x1ADD }, // [31] COMBINING LATIN SMALL LETTER W BELOW..COMBINING DOT-AND-RING BELOW
    Interval{ 0x1AE0, 0x1AEB }, // [12] COMBINING LEFT TACK ABOVE..COMBINING DOUBLE RIGHTWARDS ARROW ABOVE
    Interval{ 0x1B00, 0x1B03 }, // [4] BALINESE SIGN ULU RICEM..BALINESE SIGN SURANG
    Interval{ 0x1B34, 0x1B34 }, // BALINESE SIGN REREKAN
    Interval{ 0x1B36, 0x1B3A }, // [5] BALINESE VOWEL SIGN ULU..BALINESE VOWEL SIGN RA REPA
    Interval{ 0x1B3C, 0x1B3C }, // BALINESE VOWEL SIGN LA LENGA
    Interval{ 0x1B42, 0x1B42 }, // BALINESE VOWEL SIGN PEPET
    Interval{ 0x1B6B, 0x1B73 }, // [9] BALINESE MUSICAL SYMBOL COMBINING TEGEH..BALINESE MUSICAL SYMBOL COMBINING GONG
    Interval{ 0x1B80, 0x1B81 }, // [2] SUNDANESE SIGN PANYECEK..SUNDANESE SIGN PANGLAYAR
    Interval{ 0x1BA2, 0x1BA5 }, // [4] SUNDANESE CONSONANT SIGN PANYAKRA..SUNDANESE VOWEL SIGN PANYUKU
    Interval{ 0x1BA8, 0x1BA9 }, // [2] SUNDANESE VOWEL SIGN PAMEPET..SUNDANESE VOWEL SIGN PANEULEUNG
    Interval{ 0x1BAB, 0x1BAD }, // [3] SUNDANESE SIGN VIRAMA..SUNDANESE CONSONANT SIGN PASANGAN WA
    Interval{ 0x1BE6, 0x1BE6 }, // BATAK SIGN TOMPI
    Interval{ 0x1BE8, 0x1BE9 }, // [2] BATAK VOWEL SIGN PAKPAK E..BATAK VOWEL SIGN EE
    Interval{ 0x1BED, 0x1BED }, // BATAK VOWEL SIGN KARO O
    Interval{ 0x1BEF, 0x1BF1 }, // [3] BATAK VOWEL SIGN U FOR SIMALUNGUN SA..BATAK CONSONANT SIGN H
    Interval{ 0x1C2C, 0x1C33 }, // [8] LEPCHA VOWEL SIGN E..LEPCHA CONSONANT SIGN T
    Interval{ 0x1C36, 0x1C37 }, // [2] LEPCHA SIGN RAN..LEPCHA SIGN NUKTA
    Interval{ 0x1CD0, 0x1CD2 }, // [3] VEDIC TONE KARSHANA..VEDIC TONE PRENKHA
    Interval{ 0x1CD4, 0x1CE0 }, // [13] VEDIC SIGN YAJURVEDIC MIDLINE SVARITA..VEDIC TONE RIGVEDIC KASHMIRI INDEPENDENT SVARITA
    Interval{ 0x1CE2, 0x1CE8 }, // [7] VEDIC SIGN VISARGA SVARITA..VEDIC SIGN VISARGA ANUDATTA WITH TAIL
    Interval{ 0x1CED, 0x1CED }, // VEDIC SIGN TIRYAK
    Interval{ 0x1CF4, 0x1CF4 }, // VEDIC TONE CANDRA ABOVE
    Interval{ 0x1CF8, 0x1CF9 }, // [2] VEDIC TONE RING ABOVE..VEDIC TONE DOUBLE RING ABOVE
    Interval{ 0x1DC0, 0x1DFF }, // [64] COMBINING DOTTED GRAVE ACCENT..COMBINING RIGHT ARROWHEAD AND DOWN ARROWHEAD BELOW
    Interval{ 0x20D0, 0x20DC }, // [13] COMBINING LEFT HARPOON ABOVE..COMBINING FOUR DOTS ABOVE
    Interval{ 0x20E1, 0x20E1 }, // COMBINING LEFT RIGHT ARROW ABOVE
    Interval{ 0x20E5, 0x20F0 }, // [12] COMBINING REVERSE SOLIDUS OVERLAY..COMBINING ASTERISK ABOVE
    Interval{ 0x2CEF, 0x2CF1 }, // [3] COPTIC COMBINING NI ABOVE..COPTIC COMBINING SPIRITUS LENIS
    Interval{ 0x2D7F, 0x2D7F }, // TIFINAGH CONSONANT JOINER
    Interval{ 0x2DE0, 0x2DFF }, // [32] COMBINING CYRILLIC LETTER BE..COMBINING CYRILLIC LETTER IOTIFIED BIG YUS
    Interval{ 0x302A, 0x302D }, // [4] IDEOGRAPHIC LEVEL TONE MARK..IDEOGRAPHIC ENTERING TONE MARK
    Interval{ 0x3099, 0x309A }, // [2] COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK..COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Interval{ 0xA66F, 0xA66F }, // COMBINING CYRILLIC VZMET
    Interval{ 0xA674, 0xA67D }, // [10] COMBINING CYRILLIC LETTER UKRAINIAN IE..COMBINING CYRILLIC PAYEROK
    Interval{ 0xA69E, 0xA69F }, // [2] COMBINING CYRILLIC LETTER EF..COMBINING CYRILLIC LETTER IOTIFIED E
    Interval{ 0xA6F0, 0xA6F1 }, // [2] BAMUM COMBINING MARK KOQNDON..BAMUM COMBINING MARK TUKWENTIS
    Interval{ 0xA802, 0xA802 }, // SYLOTI NAGRI SIGN DVISVARA
    Interval{ 0xA806, 0xA806 }, // SYLOTI NAGRI SIGN HASANTA
    Interval{ 0xA80B, 0xA80B }, // SYLOTI NAGRI SIGN ANUSVARA
    Interval{ 0xA825, 0xA826 }, // [2] SYLOTI NAGRI VOWEL SIGN U..SYLOTI NAGRI VOWEL SIGN E
    Interval{ 0xA82C, 0xA82C }, // SYLOTI NAGRI SIGN ALTERNATE HASANTA
    Interval{ 0xA8C4, 0xA8C5 }, // [2] SAURASHTRA SIGN VIRAMA..SAURASHTRA SIGN CANDRABINDU
    Interval{ 0xA8E0, 0xA8F1 }, // [18] COMBINING DEVANAGARI DIGIT ZERO..COMBINING DEVANAGARI SIGN AVAGRAHA
    Interval{ 0xA8FF, 0xA8FF }, // DEVANAGARI VOWEL SIGN AY
    Interval{ 0xA926, 0xA92D }, // [8] KAYAH LI VOWEL UE..KAYAH LI TONE CALYA PLOPHU
    Interval{ 0xA947, 0xA951 }, // [11] REJANG VOWEL SIGN I..REJANG CONSONANT SIGN R
    Interval{ 0xA980, 0xA982 }, // [3] JAVANESE SIGN PANYANGGA..JAVANESE SIGN LAYAR
    Interval{ 0xA9B3, 0xA9B3 }, // JAVANESE SIGN CECAK TELU
    Interval{ 0xA9B6, 0xA9B9 }, // [4] JAVANESE VOWEL SIGN WULU..JAVANESE VOWEL SIGN SUKU MENDUT
    Interval{ 0xA9BC, 0xA9BD }, // [2] JAVANESE VOWEL SIGN PEPET..JAVANESE CONSONANT SIGN KERET
    Interval{ 0xA9E5, 0xA9E5 }, // MYANMAR SIGN SHAN SAW
    Interval{ 0xAA29, 0xAA2E }, // [6] CHAM VOWEL SIGN AA..CHAM VOWEL SIGN OE
    Interval{ 0xAA31, 0xAA32 }, // [2] CHAM VOWEL SIGN AU..CHAM VOWEL SIGN UE
    Interval{ 0xAA35, 0xAA36 }, // [2] CHAM CONSONANT SIGN LA..CHAM CONSONANT SIGN WA
    Interval{ 0xAA43, 0xAA43 }, // CHAM CONSONANT SIGN FINAL NG
    Interval{ 0xAA4C, 0xAA4C }, // CHAM CONSONANT SIGN FINAL M
    Interval{ 0xAA7C, 0xAA7C }, // MYANMAR SIGN TAI LAING TONE-2
    Interval{ 0xAAB0, 0xAAB0 }, // TAI VIET MAI KANG
    Interval{ 0xAAB2, 0xAAB4 }, // [3] TAI VIET VOWEL I..TAI VIET VOWEL U
    Interval{ 0xAAB7, 0xAAB8 }, // [2] TAI VIET MAI KHIT..TAI VIET VOWEL IA
    Interval{ 0xAABE, 0xAABF }, // [2] TAI VIET VOWEL AM..TAI VIET TONE MAI EK
    Interval{ 0xAAC1, 0xAAC1 }, // TAI VIET TONE MAI THO
    Interval{ 0xAAEC, 0xAAED }, // [2] MEETEI MAYEK VOWEL SIGN UU..MEETEI MAYEK VOWEL SIGN AAI
    Interval{ 0xAAF6, 0xAAF6 }, // MEETEI MAYEK VIRAMA
    Interval{ 0xABE5, 0xABE5 }, // MEETEI MAYEK VOWEL SIGN ANAP
    Interval{ 0xABE8, 0xABE8 }, // MEETEI MAYEK VOWEL SIGN UNAP
    Interval{ 0xABED, 0xABED }, // MEETEI MAYEK APUN IYEK
    Interval{ 0xFB1E, 0xFB1E }, // HEBREW POINT JUDEO-SPANISH VARIKA
    Interval{ 0xFE00, 0xFE0F }, // [16] VARIATION SELECTOR-1..VARIATION SELECTOR-16
    Interval{ 0xFE20, 0xFE2F }, // [16] COMBINING LIGATURE LEFT HALF..COMBINING CYRILLIC TITLO RIGHT HALF
    Interval{ 0x101FD, 0x101FD }, // PHAISTOS DISC SIGN COMBINING OBLIQUE STROKE
    Interval{ 0x102E0, 0x102E0 }, // COPTIC EPACT THOUSANDS MARK
    Interval{ 0x10376, 0x1037A }, // [5] COMBINING OLD PERMIC LETTER AN..COMBINING OLD PERMIC LETTER SII
    Interval{ 0x10A01, 0x10A03 }, // [3] KHAROSHTHI VOWEL SIGN I..KHAROSHTHI VOWEL SIGN VOCALIC R
    Interval{ 0x10A05, 0x10A06 }, // [2] KHAROSHTHI VOWEL SIGN E..KHAROSHTHI VOWEL SIGN O
    Interval{ 0x10A0C, 0x10A0F }, // [4] KHAROSHTHI VOWEL LENGTH MARK..KHAROSHTHI SIGN VISARGA
    Interval{ 0x10A38, 0x10A3A }, // [3] KHAROSHTHI SIGN BAR ABOVE..KHAROSHTHI SIGN DOT BELOW
    Interval{ 0x10A3F, 0x10A3F }, // KHAROSHTHI VIRAMA
    Interval{ 0x10AE5, 0x10AE6 }, // [2] MANICHAEAN ABBREVIATION MARK ABOVE..MANICHAEAN ABBREVIATION MARK BELOW
    Interval{ 0x10D24, 0x10D27 }, // [4] HANIFI ROHINGYA SIGN HARBAHAY..HANIFI ROHINGYA SIGN TASSI
    Interval{ 0x10D69, 0x10D6D }, // [5] GARAY VOWEL SIGN E..GARAY CONSONANT NASALIZATION MARK
    Interval{ 0x10EAB, 0x10EAC }, // [2] YEZIDI COMBINING HAMZA MARK..YEZIDI COMBINING MADDA MARK
    Interval{ 0x10EFA, 0x10EFF }, // [6] ARABIC DOUBLE VERTICAL BAR BELOW..ARABIC SMALL LOW WORD MADDA
    Interval{ 0x10F46, 0x10F50 }, // [11] SOGDIAN COMBINING DOT BELOW..SOGDIAN COMBINING STROKE BELOW
    Interval{ 0x10F82, 0x10F85 }, // [4] OLD UYGHUR COMBINING DOT ABOVE..OLD UYGHUR COMBINING TWO DOTS BELOW
    Interval{ 0x11001, 0x11001 }, // BRAHMI SIGN ANUSVARA
    Interval{ 0x11038, 0x11046 }, // [15] BRAHMI VOWEL SIGN AA..BRAHMI VIRAMA
    Interval{ 0x11070, 0x11070 }, // BRAHMI SIGN OLD TAMIL VIRAMA
    Interval{ 0x11073, 0x11074 }, // [2] BRAHMI VOWEL SIGN OLD TAMIL SHORT E..BRAHMI VOWEL SIGN OLD TAMIL SHORT O
    Interval{ 0x1107F, 0x11081 }, // [3] BRAHMI NUMBER JOINER..KAITHI SIGN ANUSVARA
    Interval{ 0x110B3, 0x110B6 }, // [4] KAITHI VOWEL SIGN U..KAITHI VOWEL SIGN AI
    Interval{ 0x110B9, 0x110BA }, // [2] KAITHI SIGN VIRAMA..KAITHI SIGN NUKTA
    Interval{ 0x110C2, 0x110C2 }, // KAITHI VOWEL SIGN VOCALIC R
    Interval{ 0x11100, 0x11102 }, // [3] CHAKMA SIGN CANDRABINDU..CHAKMA SIGN VISARGA
    Interval{ 0x11127, 0x1112B }, // [5] CHAKMA VOWEL SIGN A..CHAKMA VOWEL SIGN UU
    Interval{ 0x1112D, 0x11134 }, // [8] CHAKMA VOWEL SIGN AI..CHAKMA MAAYYAA
    Interval{ 0x11173, 0x11173 }, // MAHAJANI SIGN NUKTA
    Interval{ 0x11180, 0x11181 }, // [2] SHARADA SIGN CANDRABINDU..SHARADA SIGN ANUSVARA
    Interval{ 0x111B6, 0x111BE }, // [9] SHARADA VOWEL SIGN U..SHARADA VOWEL SIGN O
    Interval{ 0x111C9, 0x111CC }, // [4] SHARADA SANDHI MARK..SHARADA EXTRA SHORT VOWEL MARK
    Interval{ 0x111CF, 0x111CF }, // SHARADA SIGN INVERTED CANDRABINDU
    Interval{ 0x1122F, 0x11231 }, // [3] KHOJKI VOWEL SIGN U..KHOJKI VOWEL SIGN AI
    Interval{ 0x11234, 0x11234 }, // KHOJKI SIGN ANUSVARA
    Interval{ 0x11236, 0x11237 }, // [2] KHOJKI SIGN NUKTA..KHOJKI SIGN SHADDA
    Interval{ 0x1123E, 0x1123E }, // KHOJKI SIGN SUKUN
    Interval{ 0x11241, 0x11241 }, // KHOJKI VOWEL SIGN VOCALIC R
    Interval{ 0x112DF, 0x112DF }, // KHUDAWADI SIGN ANUSVARA
    Interval{ 0x112E3, 0x112EA }, // [8] KHUDAWADI VOWEL SIGN U..KHUDAWADI SIGN VIRAMA
    Interval{ 0x11300, 0x11301 }, // [2] GRANTHA SIGN COMBINING ANUSVARA ABOVE..GRANTHA SIGN CANDRABINDU
    Interval{ 0x1133B, 0x1133C }, // [2] COMBINING BINDU BELOW..GRANTHA SIGN NUKTA
    Interval{ 0x11340, 0x11340 }, // GRANTHA VOWEL SIGN II
    Interval{ 0x11366, 0x1136C }, // [7] COMBINING GRANTHA DIGIT ZERO..COMBINING GRANTHA DIGIT SIX
    Interval{ 0x11370, 0x11374 }, // [5] COMBINING GRANTHA LETTER A..COMBINING GRANTHA LETTER PA
    Interval{ 0x113BB, 0x113C0 }, // [6] TULU-TIGALARI VOWEL SIGN U..TULU-TIGALARI VOWEL SIGN VOCALIC LL
    Interval{ 0x113CE, 0x113CE }, // TULU-TIGALARI SIGN VIRAMA
    Interval{ 0x113D0, 0x113D0 }, // TULU-TIGALARI CONJOINER
    Interval{ 0x113D2, 0x113D2 }, // TULU-TIGALARI GEMINATION MARK
    Interval{ 0x113E1, 0x113E2 }, // [2] TULU-TIGALARI VEDIC TONE SVARITA..TULU-TIGALARI VEDIC TONE ANUDATTA
    Interval{ 0x11438, 0x1143F }, // [8] NEWA VOWEL SIGN U..NEWA VOWEL SIGN AI
    Interval{ 0x11442, 0x11444 }, // [3] NEWA SIGN VIRAMA..NEWA SIGN ANUSVARA
    Interval{ 0x11446, 0x11446 }, // NEWA SIGN NUKTA
    Interval{ 0x1145E, 0x1145E }, // NEWA SANDHI MARK
    Interval{ 0x114B3, 0x114B8 }, // [6] TIRHUTA VOWEL SIGN U..TIRHUTA VOWEL SIGN VOCALIC LL
    Interval{ 0x114BA, 0x114BA }, // TIRHUTA VOWEL SIGN SHORT E
    Interval{ 0x114BF, 0x114C0 }, // [2] TIRHUTA SIGN CANDRABINDU..TIRHUTA SIGN ANUSVARA
    Interval{ 0x114C2, 0x114C3 }, // [2] TIRHUTA SIGN VIRAMA..TIRHUTA SIGN NUKTA
    Interval{ 0x115B2, 0x115B5 }, // [4] SIDDHAM VOWEL SIGN U..SIDDHAM VOWEL SIGN VOCALIC RR
    Interval{ 0x115BC, 0x115BD }, // [2] SIDDHAM SIGN CANDRABINDU..SIDDHAM SIGN ANUSVARA
    Interval{ 0x115BF, 0x115C0 }, // [2] SIDDHAM SIGN VIRAMA..SIDDHAM SIGN NUKTA
    Interval{ 0x115DC, 0x115DD }, // [2] SIDDHAM VOWEL SIGN ALTERNATE U..SIDDHAM VOWEL SIGN ALTERNATE UU
    Interval{ 0x11633, 0x1163A }, // [8] MODI VOWEL SIGN U..MODI VOWEL SIGN AI
    Interval{ 0x1163D, 0x1163D }, // MODI SIGN ANUSVARA
    Interval{ 0x1163F, 0x11640 }, // [2] MODI SIGN VIRAMA..MODI SIGN ARDHACANDRA
    Interval{ 0x116AB, 0x116AB }, // TAKRI SIGN ANUSVARA
    Interval{ 0x116AD, 0x116AD }, // TAKRI VOWEL SIGN AA
    Interval{ 0x116B0, 0x116B5 }, // [6] TAKRI VOWEL SIGN U..TAKRI VOWEL SIGN AU
    Interval{ 0x116B7, 0x116B7 }, // TAKRI SIGN NUKTA
    Interval{ 0x1171D, 0x1171D }, // AHOM CONSONANT SIGN MEDIAL LA
    Interval{ 0x1171F, 0x1171F }, // AHOM CONSONANT SIGN MEDIAL LIGATING RA
    Interval{ 0x11722, 0x11725 }, // [4] AHOM VOWEL SIGN I..AHOM VOWEL SIGN UU
    Interval{ 0x11727, 0x1172B }, // [5] AHOM VOWEL SIGN AW..AHOM SIGN KILLER
    Interval{ 0x1182F, 0x11837 }, // [9] DOGRA VOWEL SIGN U..DOGRA SIGN ANUSVARA
    Interval{ 0x11839, 0x1183A }, // [2] DOGRA SIGN VIRAMA..DOGRA SIGN NUKTA
    Interval{ 0x1193B, 0x1193C }, // [2] DIVES AKURU SIGN ANUSVARA..DIVES AKURU SIGN CANDRABINDU
    Interval{ 0x1193E, 0x1193E }, // DIVES AKURU VIRAMA
    Interval{ 0x11943, 0x11943 }, // DIVES AKURU SIGN NUKTA
    Interval{ 0x119D4, 0x119D7 }, // [4] NANDINAGARI VOWEL SIGN U..NANDINAGARI VOWEL SIGN VOCALIC RR
    Interval{ 0x119DA, 0x119DB }, // [2] NANDINAGARI VOWEL SIGN E..NANDINAGARI VOWEL SIGN AI
    Interval{ 0x119E0, 0x119E0 }, // NANDINAGARI SIGN VIRAMA
    Interval{ 0x11A01, 0x11A0A }, // [10] ZANABAZAR SQUARE VOWEL SIGN I..ZANABAZAR SQUARE VOWEL LENGTH MARK
    Interval{ 0x11A33, 0x11A38 }, // [6] ZANABAZAR SQUARE FINAL CONSONANT MARK..ZANABAZAR SQUARE SIGN ANUSVARA
    Interval{ 0x11A3B, 0x11A3E }, // [4] ZANABAZAR SQUARE CLUSTER-FINAL LETTER YA..ZANABAZAR SQUARE CLUSTER-FINAL LETTER VA
    Interval{ 0x11A47, 0x11A47 }, // ZANABAZAR SQUARE SUBJOINER
    Interval{ 0x11A51, 0x11A56 }, // [6] SOYOMBO VOWEL SIGN I..SOYOMBO VOWEL SIGN OE
    Interval{ 0x11A59, 0x11A5B }, // [3] SOYOMBO VOWEL SIGN VOCALIC R..SOYOMBO VOWEL LENGTH MARK
    Interval{ 0x11A8A, 0x11A96 }, // [13] SOYOMBO FINAL CONSONANT SIGN G..SOYOMBO SIGN ANUSVARA
    Interval{ 0x11A98, 0x11A99 }, // [2] SOYOMBO GEMINATION MARK..SOYOMBO SUBJOINER
    Interval{ 0x11B60, 0x11B60 }, // SHARADA VOWEL SIGN OE
    Interval{ 0x11B62, 0x11B64 }, // [3] SHARADA VOWEL SIGN UE..SHARADA VOWEL SIGN SHORT E
    Interval{ 0x11B66, 0x11B66 }, // SHARADA VOWEL SIGN CANDRA E
    Interval{ 0x11C30, 0x11C36 }, // [7] BHAIKSUKI VOWEL SIGN I..BHAIKSUKI VOWEL SIGN VOCALIC L
    Interval{ 0x11C38, 0x11C3D }, // [6] BHAIKSUKI VOWEL SIGN E..BHAIKSUKI SIGN ANUSVARA
    Interval{ 0x11C3F, 0x11C3F }, // BHAIKSUKI SIGN VIRAMA
    Interval{ 0x11C92, 0x11CA7 }, // [22] MARCHEN SUBJOINED LETTER KA..MARCHEN SUBJOINED LETTER ZA
    Interval{ 0x11CAA, 0x11CB0 }, // [7] MARCHEN SUBJOINED LETTER RA..MARCHEN VOWEL SIGN AA
    Interval{ 0x11CB2, 0x11CB3 }, // [2] MARCHEN VOWEL SIGN U..MARCHEN VOWEL SIGN E
    Interval{ 0x11CB5, 0x11CB6 }, // [2] MARCHEN SIGN ANUSVARA..MARCHEN SIGN CANDRABINDU
    Interval{ 0x11D31, 0x11D36 }, // [6] MASARAM GONDI VOWEL SIGN AA..MASARAM GONDI VOWEL SIGN VOCALIC R
    Interval{ 0x11D3A, 0x11D3A }, // MASARAM GONDI VOWEL SIGN E
    Interval{ 0x11D3C, 0x11D3D }, // [2] MASARAM GONDI VOWEL SIGN AI..MASARAM GONDI VOWEL SIGN O
    Interval{ 0x11D3F, 0x11D45 }, // [7] MASARAM GONDI VOWEL SIGN AU..MASARAM GONDI VIRAMA
    Interval{ 0x11D47, 0x11D47 }, // MASARAM GONDI RA-KARA
    Interval{ 0x11D90, 0x11D91 }, // [2] GUNJALA GONDI VOWEL SIGN EE..GUNJALA GONDI VOWEL SIGN AI
    Interval{ 0x11D95, 0x11D95 }, // GUNJALA GONDI SIGN ANUSVARA
    Interval{ 0x11D97, 0x11D97 }, // GUNJALA GONDI VIRAMA
    Interval{ 0x11EF3, 0x11EF4 }, // [2] MAKASAR VOWEL SIGN I..MAKASAR VOWEL SIGN U
    Interval{ 0x11F00, 0x11F01 }, // [2] KAWI SIGN CANDRABINDU..KAWI SIGN ANUSVARA
    Interval{ 0x11F36, 0x11F3A }, // [5] KAWI VOWEL SIGN I..KAWI VOWEL SIGN VOCALIC R
    Interval{ 0x11F40, 0x11F40 }, // KAWI VOWEL SIGN EU
    Interval{ 0x11F42, 0x11F42 }, // KAWI CONJOINER
    Interval{ 0x11F5A, 0x11F5A }, // KAWI SIGN NUKTA
    Interval{ 0x13440, 0x13440 }, // EGYPTIAN HIEROGLYPH MIRROR HORIZONTALLY
    Interval{ 0x13447, 0x13455 }, // [15] EGYPTIAN HIEROGLYPH MODIFIER DAMAGED AT TOP START..EGYPTIAN HIEROGLYPH MODIFIER DAMAGED
    Interval{ 0x1611E, 0x16129 }, // [12] GURUNG KHEMA VOWEL SIGN AA..GURUNG KHEMA VOWEL LENGTH MARK
    Interval{ 0x1612D, 0x1612F }, // [3] GURUNG KHEMA SIGN ANUSVARA..GURUNG KHEMA SIGN THOLHOMA
    Interval{ 0x16AF0, 0x16AF4 }, // [5] BASSA VAH COMBINING HIGH TONE..BASSA VAH COMBINING HIGH-LOW TONE
    Interval{ 0x16B30, 0x16B36 }, // [7] PAHAWH HMONG MARK CIM TUB..PAHAWH HMONG MARK CIM TAUM
    Interval{ 0x16F4F, 0x16F4F }, // MIAO SIGN CONSONANT MODIFIER BAR
    Interval{ 0x16F8F, 0x16F92 }, // [4] MIAO TONE RIGHT..MIAO TONE BELOW
    Interval{ 0x16FE4, 0x16FE4 }, // KHITAN SMALL SCRIPT FILLER
    Interval{ 0x1BC9D, 0x1BC9E }, // [2] DUPLOYAN THICK LETTER SELECTOR..DUPLOYAN DOUBLE MARK
    Interval{ 0x1CF00, 0x1CF2D }, // [46] ZNAMENNY COMBINING MARK GORAZDO NIZKO S KRYZHEM ON LEFT..ZNAMENNY COMBINING MARK KRYZH ON LEFT
    Interval{ 0x1CF30, 0x1CF46 }, // [23] ZNAMENNY COMBINING TONAL RANGE MARK MRACHNO..ZNAMENNY PRIZNAK MODIFIER ROG
    Interval{ 0x1D167, 0x1D169 }, // [3] MUSICAL SYMBOL COMBINING TREMOLO-1..MUSICAL SYMBOL COMBINING TREMOLO-3
    Interval{ 0x1D17B, 0x1D182 }, // [8] MUSICAL SYMBOL COMBINING ACCENT..MUSICAL SYMBOL COMBINING LOURE
    Interval{ 0x1D185, 0x1D18B }, // [7] MUSICAL SYMBOL COMBINING DOIT..MUSICAL SYMBOL COMBINING TRIPLE TONGUE
    Interval{ 0x1D1AA, 0x1D1AD }, // [4] MUSICAL SYMBOL COMBINING DOWN BOW..MUSICAL SYMBOL COMBINING SNAP PIZZICATO
    Interval{ 0x1D242, 0x1D244 }, // [3] COMBINING GREEK MUSICAL TRISEME..COMBINING GREEK MUSICAL PENTASEME
    Interval{ 0x1DA00, 0x1DA36 }, // [55] SIGNWRITING HEAD RIM..SIGNWRITING AIR SUCKING IN
    Interval{ 0x1DA3B, 0x1DA6C }, // [50] SIGNWRITING MOUTH CLOSED NEUTRAL..SIGNWRITING EXCITEMENT
    Interval{ 0x1DA75, 0x1DA75 }, // SIGNWRITING UPPER BODY TILTING FROM HIP JOINTS
    Interval{ 0x1DA84, 0x1DA84 }, // SIGNWRITING LOCATION HEAD NECK
    Interval{ 0x1DA9B, 0x1DA9F }, // [5] SIGNWRITING FILL MODIFIER-2..SIGNWRITING FILL MODIFIER-6
    Interval{ 0x1DAA1, 0x1DAAF }, // [15] SIGNWRITING ROTATION MODIFIER-2..SIGNWRITING ROTATION MODIFIER-16
    Interval{ 0x1E000, 0x1E006 }, // [7] COMBINING GLAGOLITIC LETTER AZU..COMBINING GLAGOLITIC LETTER ZHIVETE
    Interval{ 0x1E008, 0x1E018 }, // [17] COMBINING GLAGOLITIC LETTER ZEMLJA..COMBINING GLAGOLITIC LETTER HERU
    Interval{ 0x1E01B, 0x1E021 }, // [7] COMBINING GLAGOLITIC LETTER SHTA..COMBINING GLAGOLITIC LETTER YATI
    Interval{ 0x1E023, 0x1E024 }, // [2] COMBINING GLAGOLITIC LETTER YU..COMBINING GLAGOLITIC LETTER SMALL YUS
    Interval{ 0x1E026, 0x1E02A }, // [5] COMBINING GLAGOLITIC LETTER YO..COMBINING GLAGOLITIC LETTER FITA
    Interval{ 0x1E08F, 0x1E08F }, // COMBINING CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
    Interval{ 0x1E130, 0x1E136 }, // [7] NYIAKENG PUACHUE HMONG TONE-B..NYIAKENG PUACHUE HMONG TONE-D
    Interval{ 0x1E2AE, 0x1E2AE }, // TOTO SIGN RISING TONE
    Interval{ 0x1E2EC, 0x1E2EF }, // [4] WANCHO TONE TUP..WANCHO TONE KOINI
    Interval{ 0x1E4EC, 0x1E4EF }, // [4] NAG MUNDARI SIGN MUHOR..NAG MUNDARI SIGN SUTUH
    Interval{ 0x1E5EE, 0x1E5EF }, // [2] OL ONAL SIGN MU..OL ONAL SIGN IKIR
    Interval{ 0x1E6E3, 0x1E6E3 }, // TAI YO SIGN UE
    Interval{ 0x1E6E6, 0x1E6E6 }, // TAI YO SIGN AU
    Interval{ 0x1E6EE, 0x1E6EF }, // [2] TAI YO SIGN AY..TAI YO SIGN ANG
    Interval{ 0x1E6F5, 0x1E6F5 }, // TAI YO SIGN OM
    Interval{ 0x1E8D0, 0x1E8D6 }, // [7] MENDE KIKAKUI COMBINING NUMBER TEENS..MENDE KIKAKUI COMBINING NUMBER MILLIONS
    Interval{ 0x1E944, 0x1E94A }, // [7] ADLAM ALIF LENGTHENER..ADLAM NUKTA
    Interval{ 0xE0100, 0xE01EF }, // [240] VARIATION SELECTOR-17..VARIATION SELECTOR-256
}; // }}}
auto static const Open_Punctuation = std::array<Interval, 79>{ // {{{
    Interval{ 0x0028, 0x0028 }, // LEFT PARENTHESIS
    Interval{ 0x005B, 0x005B }, // LEFT SQUARE BRACKET
    Interval{ 0x007B, 0x007B }, // LEFT CURLY BRACKET
    Interval{ 0x0F3A, 0x0F3A }, // TIBETAN MARK GUG RTAGS GYON
    Interval{ 0x0F3C, 0x0F3C }, // TIBETAN MARK ANG KHANG GYON
    Interval{ 0x169B, 0x169B }, // OGHAM FEATHER MARK
    Interval{ 0x201A, 0x201A }, // SINGLE LOW-9 QUOTATION MARK
    Interval{ 0x201E, 0x201E }, // DOUBLE LOW-9 QUOTATION MARK
    Interval{ 0x2045, 0x2045 }, // LEFT SQUARE BRACKET WITH QUILL
    Interval{ 0x207D, 0x207D }, // SUPERSCRIPT LEFT PARENTHESIS
    Interval{ 0x208D, 0x208D }, // SUBSCRIPT LEFT PARENTHESIS
    Interval{ 0x2308, 0x2308 }, // LEFT CEILING
    Interval{ 0x230A, 0x230A }, // LEFT FLOOR
    Interval{ 0x2329, 0x2329 }, // LEFT-POINTING ANGLE BRACKET
    Interval{ 0x2768, 0x2768 }, // MEDIUM LEFT PARENTHESIS ORNAMENT
    Interval{ 0x276A, 0x276A }, // MEDIUM FLATTENED LEFT PARENTHESIS ORNAMENT
    Interval{ 0x276C, 0x276C }, // MEDIUM LEFT-POINTING ANGLE BRACKET ORNAMENT
    Interval{ 0x276E, 0x276E }, // HEAVY LEFT-POINTING ANGLE QUOTATION MARK ORNAMENT
    Interval{ 0x2770, 0x2770 }, // HEAVY LEFT-POINTING ANGLE BRACKET ORNAMENT
    Interval{ 0x2772, 0x2772 }, // LIGHT LEFT TORTOISE SHELL BRACKET ORNAMENT
    Interval{ 0x2774, 0x2774 }, // MEDIUM LEFT CURLY BRACKET ORNAMENT
    Interval{ 0x27C5, 0x27C5 }, // LEFT S-SHAPED BAG DELIMITER
    Interval{ 0x27E6, 0x27E6 }, // MATHEMATICAL LEFT WHITE SQUARE BRACKET
    Interval{ 0x27E8, 0x27E8 }, // MATHEMATICAL LEFT ANGLE BRACKET
    Interval{ 0x27EA, 0x27EA }, // MATHEMATICAL LEFT DOUBLE ANGLE BRACKET
    Interval{ 0x27EC, 0x27EC }, // MATHEMATICAL LEFT WHITE TORTOISE SHELL BRACKET
    Interval{ 0x27EE, 0x27EE }, // MATHEMATICAL LEFT FLATTENED PARENTHESIS
    Interval{ 0x2983, 0x2983 }, // LEFT WHITE CURLY BRACKET
    Interval{ 0x2985, 0x2985 }, // LEFT WHITE PARENTHESIS
    Interval{ 0x2987, 0x2987 }, // Z NOTATION LEFT IMAGE BRACKET
    Interval{ 0x2989, 0x2989 }, // Z NOTATION LEFT BINDING BRACKET
    Interval{ 0x298B, 0x298B }, // LEFT SQUARE BRACKET WITH UNDERBAR
    Interval{ 0x298D, 0x298D }, // LEFT SQUARE BRACKET WITH TICK IN TOP CORNER
    Interval{ 0x298F, 0x298F }, // LEFT SQUARE BRACKET WITH TICK IN BOTTOM CORNER
    Interval{ 0x2991, 0x2991 }, // LEFT ANGLE BRACKET WITH DOT
    Interval{ 0x2993, 0x2993 }, // LEFT ARC LESS-THAN BRACKET
    Interval{ 0x2995, 0x2995 }, // DOUBLE LEFT ARC GREATER-THAN BRACKET
    Interval{ 0x2997, 0x2997 }, // LEFT BLACK TORTOISE SHELL BRACKET
    Interval{ 0x29D8, 0x29D8 }, // LEFT WIGGLY FENCE
    Interval{ 0x29DA, 0x29DA }, // LEFT DOUBLE WIGGLY FENCE
    Interval{ 0x29FC, 0x29FC }, // LEFT-POINTING CURVED ANGLE BRACKET
    Interval{ 0x2E22, 0x2E22 }, // TOP LEFT HALF BRACKET
    Interval{ 0x2E24, 0x2E24 }, // BOTTOM LEFT HALF BRACKET
    Interval{ 0x2E26, 0x2E26 }, // LEFT SIDEWAYS U BRACKET
    Interval{ 0x2E28, 0x2E28 }, // LEFT DOUBLE PARENTHESIS
    Interval{ 0x2E42, 0x2E42 }, // DOUBLE LOW-REVERSED-9 QUOTATION MARK
    Interval{ 0x2E55, 0x2E55 }, // LEFT SQUARE BRACKET WITH STROKE
    Interval{ 0x2E57, 0x2E57 }, // LEFT SQUARE BRACKET WITH DOUBLE STROKE
    Interval{ 0x2E59, 0x2E59 }, // TOP HALF LEFT PARENTHESIS
    Interval{ 0x2E5B, 0x2E5B }, // BOTTOM HALF LEFT PARENTHESIS
    Interval{ 0x3008, 0x3008 }, // LEFT ANGLE BRACKET
    Interval{ 0x300A, 0x300A }, // LEFT DOUBLE ANGLE BRACKET
    Interval{ 0x300C, 0x300C }, // LEFT CORNER BRACKET
    Interval{ 0x300E, 0x300E }, // LEFT WHITE CORNER BRACKET
    Interval{ 0x3010, 0x3010 }, // LEFT BLACK LENTICULAR BRACKET
    Interval{ 0x3014, 0x3014 }, // LEFT TORTOISE SHELL BRACKET
    Interval{ 0x3016, 0x3016 }, // LEFT WHITE LENTICULAR BRACKET
    Interval{ 0x3018, 0x3018 }, // LEFT WHITE TORTOISE SHELL BRACKET
    Interval{ 0x301A, 0x301A }, // LEFT WHITE SQUARE BRACKET
    Interval{ 0x301D, 0x301D }, // REVERSED DOUBLE PRIME QUOTATION MARK
    Interval{ 0xFD3F, 0xFD3F }, // ORNATE RIGHT PARENTHESIS
    Interval{ 0xFE17, 0xFE17 }, // PRESENTATION FORM FOR VERTICAL LEFT WHITE LENTICULAR BRACKET
    Interval{ 0xFE35, 0xFE35 }, // PRESENTATION FORM FOR VERTICAL LEFT PARENTHESIS
    Interval{ 0xFE37, 0xFE37 }, // PRESENTATION FORM FOR VERTICAL LEFT CURLY BRACKET
    Interval{ 0xFE39, 0xFE39 }, // PRESENTATION FORM FOR VERTICAL LEFT TORTOISE SHELL BRACKET
    Interval{ 0xFE3B, 0xFE3B }, // PRESENTATION FORM FOR VERTICAL LEFT BLACK LENTICULAR BRACKET
    Interval{ 0xFE3D, 0xFE3D }, // PRESENTATION FORM FOR VERTICAL LEFT DOUBLE ANGLE BRACKET
    Interval{ 0xFE3F, 0xFE3F }, // PRESENTATION FORM FOR VERTICAL LEFT ANGLE BRACKET
    Interval{ 0xFE41, 0xFE41 }, // PRESENTATION FORM FOR VERTICAL LEFT CORNER BRACKET
    Interval{ 0xFE43, 0xFE43 }, // PRESENTATION FORM FOR VERTICAL LEFT WHITE CORNER BRACKET
    Interval{ 0xFE47, 0xFE47 }, // PRESENTATION FORM FOR VERTICAL LEFT SQUARE BRACKET
    Interval{ 0xFE59, 0xFE59 }, // SMALL LEFT PARENTHESIS
    Interval{ 0xFE5B, 0xFE5B }, // SMALL LEFT CURLY BRACKET
    Interval{ 0xFE5D, 0xFE5D }, // SMALL LEFT TORTOISE SHELL BRACKET
    Interval{ 0xFF08, 0xFF08 }, // FULLWIDTH LEFT PARENTHESIS
    Interval{ 0xFF3B, 0xFF3B }, // FULLWIDTH LEFT SQUARE BRACKET
    Interval{ 0xFF5B, 0xFF5B }, // FULLWIDTH LEFT CURLY BRACKET
    Interval{ 0xFF5F, 0xFF5F }, // FULLWIDTH LEFT WHITE PARENTHESIS
    Interval{ 0xFF62, 0xFF62 }, // HALFWIDTH LEFT CORNER BRACKET
}; // }}}
auto static const Other_Letter = std::array<Interval, 537>{ // {{{
    Interval{ 0x00AA, 0x00AA }, // FEMININE ORDINAL INDICATOR
    Interval{ 0x00BA, 0x00BA }, // MASCULINE ORDINAL INDICATOR
    Interval{ 0x01BB, 0x01BB }, // LATIN LETTER TWO WITH STROKE
    Interval{ 0x01C0, 0x01C3 }, // [4] LATIN LETTER DENTAL CLICK..LATIN LETTER RETROFLEX CLICK
    Interval{ 0x0294, 0x0295 }, // [2] LATIN LETTER GLOTTAL STOP..LATIN LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0x05D0, 0x05EA }, // [27] HEBREW LETTER ALEF..HEBREW LETTER TAV
    Interval{ 0x05EF, 0x05F2 }, // [4] HEBREW YOD TRIANGLE..HEBREW LIGATURE YIDDISH DOUBLE YOD
    Interval{ 0x0620, 0x063F }, // [32] ARABIC LETTER KASHMIRI YEH..ARABIC LETTER FARSI YEH WITH THREE DOTS ABOVE
    Interval{ 0x0641, 0x064A }, // [10] ARABIC LETTER FEH..ARABIC LETTER YEH
    Interval{ 0x066E, 0x066F }, // [2] ARABIC LETTER DOTLESS BEH..ARABIC LETTER DOTLESS QAF
    Interval{ 0x0671, 0x06D3 }, // [99] ARABIC LETTER ALEF WASLA..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE
    Interval{ 0x06D5, 0x06D5 }, // ARABIC LETTER AE
    Interval{ 0x06EE, 0x06EF }, // [2] ARABIC LETTER DAL WITH INVERTED V..ARABIC LETTER REH WITH INVERTED V
    Interval{ 0x06FA, 0x06FC }, // [3] ARABIC LETTER SHEEN WITH DOT BELOW..ARABIC LETTER GHAIN WITH DOT BELOW
    Interval{ 0x06FF, 0x06FF }, // ARABIC LETTER HEH WITH INVERTED V
    Interval{ 0x0710, 0x0710 }, // SYRIAC LETTER ALAPH
    Interval{ 0x0712, 0x072F }, // [30] SYRIAC LETTER BETH..SYRIAC LETTER PERSIAN DHALATH
    Interval{ 0x074D, 0x07A5 }, // [89] SYRIAC LETTER SOGDIAN ZHAIN..THAANA LETTER WAAVU
    Interval{ 0x07B1, 0x07B1 }, // THAANA LETTER NAA
    Interval{ 0x07CA, 0x07EA }, // [33] NKO LETTER A..NKO LETTER JONA RA
    Interval{ 0x0800, 0x0815 }, // [22] SAMARITAN LETTER ALAF..SAMARITAN LETTER TAAF
    Interval{ 0x0840, 0x0858 }, // [25] MANDAIC LETTER HALQA..MANDAIC LETTER AIN
    Interval{ 0x0860, 0x086A }, // [11] SYRIAC LETTER MALAYALAM NGA..SYRIAC LETTER MALAYALAM SSA
    Interval{ 0x0870, 0x0887 }, // [24] ARABIC LETTER ALEF WITH ATTACHED FATHA..ARABIC BASELINE ROUND DOT
    Interval{ 0x0889, 0x088F }, // [7] ARABIC LETTER NOON WITH INVERTED SMALL V..ARABIC LETTER NOON WITH RING ABOVE
    Interval{ 0x08A0, 0x08C8 }, // [41] ARABIC LETTER BEH WITH SMALL V BELOW..ARABIC LETTER GRAF
    Interval{ 0x0904, 0x0939 }, // [54] DEVANAGARI LETTER SHORT A..DEVANAGARI LETTER HA
    Interval{ 0x093D, 0x093D }, // DEVANAGARI SIGN AVAGRAHA
    Interval{ 0x0950, 0x0950 }, // DEVANAGARI OM
    Interval{ 0x0958, 0x0961 }, // [10] DEVANAGARI LETTER QA..DEVANAGARI LETTER VOCALIC LL
    Interval{ 0x0972, 0x0980 }, // [15] DEVANAGARI LETTER CANDRA A..BENGALI ANJI
    Interval{ 0x0985, 0x098C }, // [8] BENGALI LETTER A..BENGALI LETTER VOCALIC L
    Interval{ 0x098F, 0x0990 }, // [2] BENGALI LETTER E..BENGALI LETTER AI
    Interval{ 0x0993, 0x09A8 }, // [22] BENGALI LETTER O..BENGALI LETTER NA
    Interval{ 0x09AA, 0x09B0 }, // [7] BENGALI LETTER PA..BENGALI LETTER RA
    Interval{ 0x09B2, 0x09B2 }, // BENGALI LETTER LA
    Interval{ 0x09B6, 0x09B9 }, // [4] BENGALI LETTER SHA..BENGALI LETTER HA
    Interval{ 0x09BD, 0x09BD }, // BENGALI SIGN AVAGRAHA
    Interval{ 0x09CE, 0x09CE }, // BENGALI LETTER KHANDA TA
    Interval{ 0x09DC, 0x09DD }, // [2] BENGALI LETTER RRA..BENGALI LETTER RHA
    Interval{ 0x09DF, 0x09E1 }, // [3] BENGALI LETTER YYA..BENGALI LETTER VOCALIC LL
    Interval{ 0x09F0, 0x09F1 }, // [2] BENGALI LETTER RA WITH MIDDLE DIAGONAL..BENGALI LETTER RA WITH LOWER DIAGONAL
    Interval{ 0x09FC, 0x09FC }, // BENGALI LETTER VEDIC ANUSVARA
    Interval{ 0x0A05, 0x0A0A }, // [6] GURMUKHI LETTER A..GURMUKHI LETTER UU
    Interval{ 0x0A0F, 0x0A10 }, // [2] GURMUKHI LETTER EE..GURMUKHI LETTER AI
    Interval{ 0x0A13, 0x0A28 }, // [22] GURMUKHI LETTER OO..GURMUKHI LETTER NA
    Interval{ 0x0A2A, 0x0A30 }, // [7] GURMUKHI LETTER PA..GURMUKHI LETTER RA
    Interval{ 0x0A32, 0x0A33 }, // [2] GURMUKHI LETTER LA..GURMUKHI LETTER LLA
    Interval{ 0x0A35, 0x0A36 }, // [2] GURMUKHI LETTER VA..GURMUKHI LETTER SHA
    Interval{ 0x0A38, 0x0A39 }, // [2] GURMUKHI LETTER SA..GURMUKHI LETTER HA
    Interval{ 0x0A59, 0x0A5C }, // [4] GURMUKHI LETTER KHHA..GURMUKHI LETTER RRA
    Interval{ 0x0A5E, 0x0A5E }, // GURMUKHI LETTER FA
    Interval{ 0x0A72, 0x0A74 }, // [3] GURMUKHI IRI..GURMUKHI EK ONKAR
    Interval{ 0x0A85, 0x0A8D }, // [9] GUJARATI LETTER A..GUJARATI VOWEL CANDRA E
    Interval{ 0x0A8F, 0x0A91 }, // [3] GUJARATI LETTER E..GUJARATI VOWEL CANDRA O
    Interval{ 0x0A93, 0x0AA8 }, // [22] GUJARATI LETTER O..GUJARATI LETTER NA
    Interval{ 0x0AAA, 0x0AB0 }, // [7] GUJARATI LETTER PA..GUJARATI LETTER RA
    Interval{ 0x0AB2, 0x0AB3 }, // [2] GUJARATI LETTER LA..GUJARATI LETTER LLA
    Interval{ 0x0AB5, 0x0AB9 }, // [5] GUJARATI LETTER VA..GUJARATI LETTER HA
    Interval{ 0x0ABD, 0x0ABD }, // GUJARATI SIGN AVAGRAHA
    Interval{ 0x0AD0, 0x0AD0 }, // GUJARATI OM
    Interval{ 0x0AE0, 0x0AE1 }, // [2] GUJARATI LETTER VOCALIC RR..GUJARATI LETTER VOCALIC LL
    Interval{ 0x0AF9, 0x0AF9 }, // GUJARATI LETTER ZHA
    Interval{ 0x0B05, 0x0B0C }, // [8] ORIYA LETTER A..ORIYA LETTER VOCALIC L
    Interval{ 0x0B0F, 0x0B10 }, // [2] ORIYA LETTER E..ORIYA LETTER AI
    Interval{ 0x0B13, 0x0B28 }, // [22] ORIYA LETTER O..ORIYA LETTER NA
    Interval{ 0x0B2A, 0x0B30 }, // [7] ORIYA LETTER PA..ORIYA LETTER RA
    Interval{ 0x0B32, 0x0B33 }, // [2] ORIYA LETTER LA..ORIYA LETTER LLA
    Interval{ 0x0B35, 0x0B39 }, // [5] ORIYA LETTER VA..ORIYA LETTER HA
    Interval{ 0x0B3D, 0x0B3D }, // ORIYA SIGN AVAGRAHA
    Interval{ 0x0B5C, 0x0B5D }, // [2] ORIYA LETTER RRA..ORIYA LETTER RHA
    Interval{ 0x0B5F, 0x0B61 }, // [3] ORIYA LETTER YYA..ORIYA LETTER VOCALIC LL
    Interval{ 0x0B71, 0x0B71 }, // ORIYA LETTER WA
    Interval{ 0x0B83, 0x0B83 }, // TAMIL SIGN VISARGA
    Interval{ 0x0B85, 0x0B8A }, // [6] TAMIL LETTER A..TAMIL LETTER UU
    Interval{ 0x0B8E, 0x0B90 }, // [3] TAMIL LETTER E..TAMIL LETTER AI
    Interval{ 0x0B92, 0x0B95 }, // [4] TAMIL LETTER O..TAMIL LETTER KA
    Interval{ 0x0B99, 0x0B9A }, // [2] TAMIL LETTER NGA..TAMIL LETTER CA
    Interval{ 0x0B9C, 0x0B9C }, // TAMIL LETTER JA
    Interval{ 0x0B9E, 0x0B9F }, // [2] TAMIL LETTER NYA..TAMIL LETTER TTA
    Interval{ 0x0BA3, 0x0BA4 }, // [2] TAMIL LETTER NNA..TAMIL LETTER TA
    Interval{ 0x0BA8, 0x0BAA }, // [3] TAMIL LETTER NA..TAMIL LETTER PA
    Interval{ 0x0BAE, 0x0BB9 }, // [12] TAMIL LETTER MA..TAMIL LETTER HA
    Interval{ 0x0BD0, 0x0BD0 }, // TAMIL OM
    Interval{ 0x0C05, 0x0C0C }, // [8] TELUGU LETTER A..TELUGU LETTER VOCALIC L
    Interval{ 0x0C0E, 0x0C10 }, // [3] TELUGU LETTER E..TELUGU LETTER AI
    Interval{ 0x0C12, 0x0C28 }, // [23] TELUGU LETTER O..TELUGU LETTER NA
    Interval{ 0x0C2A, 0x0C39 }, // [16] TELUGU LETTER PA..TELUGU LETTER HA
    Interval{ 0x0C3D, 0x0C3D }, // TELUGU SIGN AVAGRAHA
    Interval{ 0x0C58, 0x0C5A }, // [3] TELUGU LETTER TSA..TELUGU LETTER RRRA
    Interval{ 0x0C5C, 0x0C5D }, // [2] TELUGU ARCHAIC SHRII..TELUGU LETTER NAKAARA POLLU
    Interval{ 0x0C60, 0x0C61 }, // [2] TELUGU LETTER VOCALIC RR..TELUGU LETTER VOCALIC LL
    Interval{ 0x0C80, 0x0C80 }, // KANNADA SIGN SPACING CANDRABINDU
    Interval{ 0x0C85, 0x0C8C }, // [8] KANNADA LETTER A..KANNADA LETTER VOCALIC L
    Interval{ 0x0C8E, 0x0C90 }, // [3] KANNADA LETTER E..KANNADA LETTER AI
    Interval{ 0x0C92, 0x0CA8 }, // [23] KANNADA LETTER O..KANNADA LETTER NA
    Interval{ 0x0CAA, 0x0CB3 }, // [10] KANNADA LETTER PA..KANNADA LETTER LLA
    Interval{ 0x0CB5, 0x0CB9 }, // [5] KANNADA LETTER VA..KANNADA LETTER HA
    Interval{ 0x0CBD, 0x0CBD }, // KANNADA SIGN AVAGRAHA
    Interval{ 0x0CDC, 0x0CDE }, // [3] KANNADA ARCHAIC SHRII..KANNADA LETTER FA
    Interval{ 0x0CE0, 0x0CE1 }, // [2] KANNADA LETTER VOCALIC RR..KANNADA LETTER VOCALIC LL
    Interval{ 0x0CF1, 0x0CF2 }, // [2] KANNADA SIGN JIHVAMULIYA..KANNADA SIGN UPADHMANIYA
    Interval{ 0x0D04, 0x0D0C }, // [9] MALAYALAM LETTER VEDIC ANUSVARA..MALAYALAM LETTER VOCALIC L
    Interval{ 0x0D0E, 0x0D10 }, // [3] MALAYALAM LETTER E..MALAYALAM LETTER AI
    Interval{ 0x0D12, 0x0D3A }, // [41] MALAYALAM LETTER O..MALAYALAM LETTER TTTA
    Interval{ 0x0D3D, 0x0D3D }, // MALAYALAM SIGN AVAGRAHA
    Interval{ 0x0D4E, 0x0D4E }, // MALAYALAM LETTER DOT REPH
    Interval{ 0x0D54, 0x0D56 }, // [3] MALAYALAM LETTER CHILLU M..MALAYALAM LETTER CHILLU LLL
    Interval{ 0x0D5F, 0x0D61 }, // [3] MALAYALAM LETTER ARCHAIC II..MALAYALAM LETTER VOCALIC LL
    Interval{ 0x0D7A, 0x0D7F }, // [6] MALAYALAM LETTER CHILLU NN..MALAYALAM LETTER CHILLU K
    Interval{ 0x0D85, 0x0D96 }, // [18] SINHALA LETTER AYANNA..SINHALA LETTER AUYANNA
    Interval{ 0x0D9A, 0x0DB1 }, // [24] SINHALA LETTER ALPAPRAANA KAYANNA..SINHALA LETTER DANTAJA NAYANNA
    Interval{ 0x0DB3, 0x0DBB }, // [9] SINHALA LETTER SANYAKA DAYANNA..SINHALA LETTER RAYANNA
    Interval{ 0x0DBD, 0x0DBD }, // SINHALA LETTER DANTAJA LAYANNA
    Interval{ 0x0DC0, 0x0DC6 }, // [7] SINHALA LETTER VAYANNA..SINHALA LETTER FAYANNA
    Interval{ 0x0E01, 0x0E30 }, // [48] THAI CHARACTER KO KAI..THAI CHARACTER SARA A
    Interval{ 0x0E32, 0x0E33 }, // [2] THAI CHARACTER SARA AA..THAI CHARACTER SARA AM
    Interval{ 0x0E40, 0x0E45 }, // [6] THAI CHARACTER SARA E..THAI CHARACTER LAKKHANGYAO
    Interval{ 0x0E81, 0x0E82 }, // [2] LAO LETTER KO..LAO LETTER KHO SUNG
    Interval{ 0x0E84, 0x0E84 }, // LAO LETTER KHO TAM
    Interval{ 0x0E86, 0x0E8A }, // [5] LAO LETTER PALI GHA..LAO LETTER SO TAM
    Interval{ 0x0E8C, 0x0EA3 }, // [24] LAO LETTER PALI JHA..LAO LETTER LO LING
    Interval{ 0x0EA5, 0x0EA5 }, // LAO LETTER LO LOOT
    Interval{ 0x0EA7, 0x0EB0 }, // [10] LAO LETTER WO..LAO VOWEL SIGN A
    Interval{ 0x0EB2, 0x0EB3 }, // [2] LAO VOWEL SIGN AA..LAO VOWEL SIGN AM
    Interval{ 0x0EBD, 0x0EBD }, // LAO SEMIVOWEL SIGN NYO
    Interval{ 0x0EC0, 0x0EC4 }, // [5] LAO VOWEL SIGN E..LAO VOWEL SIGN AI
    Interval{ 0x0EDC, 0x0EDF }, // [4] LAO HO NO..LAO LETTER KHMU NYO
    Interval{ 0x0F00, 0x0F00 }, // TIBETAN SYLLABLE OM
    Interval{ 0x0F40, 0x0F47 }, // [8] TIBETAN LETTER KA..TIBETAN LETTER JA
    Interval{ 0x0F49, 0x0F6C }, // [36] TIBETAN LETTER NYA..TIBETAN LETTER RRA
    Interval{ 0x0F88, 0x0F8C }, // [5] TIBETAN SIGN LCE TSA CAN..TIBETAN SIGN INVERTED MCHU CAN
    Interval{ 0x1000, 0x102A }, // [43] MYANMAR LETTER KA..MYANMAR LETTER AU
    Interval{ 0x103F, 0x103F }, // MYANMAR LETTER GREAT SA
    Interval{ 0x1050, 0x1055 }, // [6] MYANMAR LETTER SHA..MYANMAR LETTER VOCALIC LL
    Interval{ 0x105A, 0x105D }, // [4] MYANMAR LETTER MON NGA..MYANMAR LETTER MON BBE
    Interval{ 0x1061, 0x1061 }, // MYANMAR LETTER SGAW KAREN SHA
    Interval{ 0x1065, 0x1066 }, // [2] MYANMAR LETTER WESTERN PWO KAREN THA..MYANMAR LETTER WESTERN PWO KAREN PWA
    Interval{ 0x106E, 0x1070 }, // [3] MYANMAR LETTER EASTERN PWO KAREN NNA..MYANMAR LETTER EASTERN PWO KAREN GHWA
    Interval{ 0x1075, 0x1081 }, // [13] MYANMAR LETTER SHAN KA..MYANMAR LETTER SHAN HA
    Interval{ 0x108E, 0x108E }, // MYANMAR LETTER RUMAI PALAUNG FA
    Interval{ 0x1100, 0x1248 }, // [329] HANGUL CHOSEONG KIYEOK..ETHIOPIC SYLLABLE QWA
    Interval{ 0x124A, 0x124D }, // [4] ETHIOPIC SYLLABLE QWI..ETHIOPIC SYLLABLE QWE
    Interval{ 0x1250, 0x1256 }, // [7] ETHIOPIC SYLLABLE QHA..ETHIOPIC SYLLABLE QHO
    Interval{ 0x1258, 0x1258 }, // ETHIOPIC SYLLABLE QHWA
    Interval{ 0x125A, 0x125D }, // [4] ETHIOPIC SYLLABLE QHWI..ETHIOPIC SYLLABLE QHWE
    Interval{ 0x1260, 0x1288 }, // [41] ETHIOPIC SYLLABLE BA..ETHIOPIC SYLLABLE XWA
    Interval{ 0x128A, 0x128D }, // [4] ETHIOPIC SYLLABLE XWI..ETHIOPIC SYLLABLE XWE
    Interval{ 0x1290, 0x12B0 }, // [33] ETHIOPIC SYLLABLE NA..ETHIOPIC SYLLABLE KWA
    Interval{ 0x12B2, 0x12B5 }, // [4] ETHIOPIC SYLLABLE KWI..ETHIOPIC SYLLABLE KWE
    Interval{ 0x12B8, 0x12BE }, // [7] ETHIOPIC SYLLABLE KXA..ETHIOPIC SYLLABLE KXO
    Interval{ 0x12C0, 0x12C0 }, // ETHIOPIC SYLLABLE KXWA
    Interval{ 0x12C2, 0x12C5 }, // [4] ETHIOPIC SYLLABLE KXWI..ETHIOPIC SYLLABLE KXWE
    Interval{ 0x12C8, 0x12D6 }, // [15] ETHIOPIC SYLLABLE WA..ETHIOPIC SYLLABLE PHARYNGEAL O
    Interval{ 0x12D8, 0x1310 }, // [57] ETHIOPIC SYLLABLE ZA..ETHIOPIC SYLLABLE GWA
    Interval{ 0x1312, 0x1315 }, // [4] ETHIOPIC SYLLABLE GWI..ETHIOPIC SYLLABLE GWE
    Interval{ 0x1318, 0x135A }, // [67] ETHIOPIC SYLLABLE GGA..ETHIOPIC SYLLABLE FYA
    Interval{ 0x1380, 0x138F }, // [16] ETHIOPIC SYLLABLE SEBATBEIT MWA..ETHIOPIC SYLLABLE PWE
    Interval{ 0x1401, 0x166C }, // [620] CANADIAN SYLLABICS E..CANADIAN SYLLABICS CARRIER TTSA
    Interval{ 0x166F, 0x167F }, // [17] CANADIAN SYLLABICS QAI..CANADIAN SYLLABICS BLACKFOOT W
    Interval{ 0x1681, 0x169A }, // [26] OGHAM LETTER BEITH..OGHAM LETTER PEITH
    Interval{ 0x16A0, 0x16EA }, // [75] RUNIC LETTER FEHU FEOH FE F..RUNIC LETTER X
    Interval{ 0x16F1, 0x16F8 }, // [8] RUNIC LETTER K..RUNIC LETTER FRANKS CASKET AESC
    Interval{ 0x1700, 0x1711 }, // [18] TAGALOG LETTER A..TAGALOG LETTER HA
    Interval{ 0x171F, 0x1731 }, // [19] TAGALOG LETTER ARCHAIC RA..HANUNOO LETTER HA
    Interval{ 0x1740, 0x1751 }, // [18] BUHID LETTER A..BUHID LETTER HA
    Interval{ 0x1760, 0x176C }, // [13] TAGBANWA LETTER A..TAGBANWA LETTER YA
    Interval{ 0x176E, 0x1770 }, // [3] TAGBANWA LETTER LA..TAGBANWA LETTER SA
    Interval{ 0x1780, 0x17B3 }, // [52] KHMER LETTER KA..KHMER INDEPENDENT VOWEL QAU
    Interval{ 0x17DC, 0x17DC }, // KHMER SIGN AVAKRAHASANYA
    Interval{ 0x1820, 0x1842 }, // [35] MONGOLIAN LETTER A..MONGOLIAN LETTER CHI
    Interval{ 0x1844, 0x1878 }, // [53] MONGOLIAN LETTER TODO E..MONGOLIAN LETTER CHA WITH TWO DOTS
    Interval{ 0x1880, 0x1884 }, // [5] MONGOLIAN LETTER ALI GALI ANUSVARA ONE..MONGOLIAN LETTER ALI GALI INVERTED UBADAMA
    Interval{ 0x1887, 0x18A8 }, // [34] MONGOLIAN LETTER ALI GALI A..MONGOLIAN LETTER MANCHU ALI GALI BHA
    Interval{ 0x18AA, 0x18AA }, // MONGOLIAN LETTER MANCHU ALI GALI LHA
    Interval{ 0x18B0, 0x18F5 }, // [70] CANADIAN SYLLABICS OY..CANADIAN SYLLABICS CARRIER DENTAL S
    Interval{ 0x1900, 0x191E }, // [31] LIMBU VOWEL-CARRIER LETTER..LIMBU LETTER TRA
    Interval{ 0x1950, 0x196D }, // [30] TAI LE LETTER KA..TAI LE LETTER AI
    Interval{ 0x1970, 0x1974 }, // [5] TAI LE LETTER TONE-2..TAI LE LETTER TONE-6
    Interval{ 0x1980, 0x19AB }, // [44] NEW TAI LUE LETTER HIGH QA..NEW TAI LUE LETTER LOW SUA
    Interval{ 0x19B0, 0x19C9 }, // [26] NEW TAI LUE VOWEL SIGN VOWEL SHORTENER..NEW TAI LUE TONE MARK-2
    Interval{ 0x1A00, 0x1A16 }, // [23] BUGINESE LETTER KA..BUGINESE LETTER HA
    Interval{ 0x1A20, 0x1A54 }, // [53] TAI THAM LETTER HIGH KA..TAI THAM LETTER GREAT SA
    Interval{ 0x1B05, 0x1B33 }, // [47] BALINESE LETTER AKARA..BALINESE LETTER HA
    Interval{ 0x1B45, 0x1B4C }, // [8] BALINESE LETTER KAF SASAK..BALINESE LETTER ARCHAIC JNYA
    Interval{ 0x1B83, 0x1BA0 }, // [30] SUNDANESE LETTER A..SUNDANESE LETTER HA
    Interval{ 0x1BAE, 0x1BAF }, // [2] SUNDANESE LETTER KHA..SUNDANESE LETTER SYA
    Interval{ 0x1BBA, 0x1BE5 }, // [44] SUNDANESE AVAGRAHA..BATAK LETTER U
    Interval{ 0x1C00, 0x1C23 }, // [36] LEPCHA LETTER KA..LEPCHA LETTER A
    Interval{ 0x1C4D, 0x1C4F }, // [3] LEPCHA LETTER TTA..LEPCHA LETTER DDA
    Interval{ 0x1C5A, 0x1C77 }, // [30] OL CHIKI LETTER LA..OL CHIKI LETTER OH
    Interval{ 0x1CE9, 0x1CEC }, // [4] VEDIC SIGN ANUSVARA ANTARGOMUKHA..VEDIC SIGN ANUSVARA VAMAGOMUKHA WITH TAIL
    Interval{ 0x1CEE, 0x1CF3 }, // [6] VEDIC SIGN HEXIFORM LONG ANUSVARA..VEDIC SIGN ROTATED ARDHAVISARGA
    Interval{ 0x1CF5, 0x1CF6 }, // [2] VEDIC SIGN JIHVAMULIYA..VEDIC SIGN UPADHMANIYA
    Interval{ 0x1CFA, 0x1CFA }, // VEDIC SIGN DOUBLE ANUSVARA ANTARGOMUKHA
    Interval{ 0x2135, 0x2138 }, // [4] ALEF SYMBOL..DALET SYMBOL
    Interval{ 0x2D30, 0x2D67 }, // [56] TIFINAGH LETTER YA..TIFINAGH LETTER YO
    Interval{ 0x2D80, 0x2D96 }, // [23] ETHIOPIC SYLLABLE LOA..ETHIOPIC SYLLABLE GGWE
    Interval{ 0x2DA0, 0x2DA6 }, // [7] ETHIOPIC SYLLABLE SSA..ETHIOPIC SYLLABLE SSO
    Interval{ 0x2DA8, 0x2DAE }, // [7] ETHIOPIC SYLLABLE CCA..ETHIOPIC SYLLABLE CCO
    Interval{ 0x2DB0, 0x2DB6 }, // [7] ETHIOPIC SYLLABLE ZZA..ETHIOPIC SYLLABLE ZZO
    Interval{ 0x2DB8, 0x2DBE }, // [7] ETHIOPIC SYLLABLE CCHA..ETHIOPIC SYLLABLE CCHO
    Interval{ 0x2DC0, 0x2DC6 }, // [7] ETHIOPIC SYLLABLE QYA..ETHIOPIC SYLLABLE QYO
    Interval{ 0x2DC8, 0x2DCE }, // [7] ETHIOPIC SYLLABLE KYA..ETHIOPIC SYLLABLE KYO
    Interval{ 0x2DD0, 0x2DD6 }, // [7] ETHIOPIC SYLLABLE XYA..ETHIOPIC SYLLABLE XYO
    Interval{ 0x2DD8, 0x2DDE }, // [7] ETHIOPIC SYLLABLE GYA..ETHIOPIC SYLLABLE GYO
    Interval{ 0x3006, 0x3006 }, // IDEOGRAPHIC CLOSING MARK
    Interval{ 0x303C, 0x303C }, // MASU MARK
    Interval{ 0x3041, 0x3096 }, // [86] HIRAGANA LETTER SMALL A..HIRAGANA LETTER SMALL KE
    Interval{ 0x309F, 0x309F }, // HIRAGANA DIGRAPH YORI
    Interval{ 0x30A1, 0x30FA }, // [90] KATAKANA LETTER SMALL A..KATAKANA LETTER VO
    Interval{ 0x30FF, 0x30FF }, // KATAKANA DIGRAPH KOTO
    Interval{ 0x3105, 0x312F }, // [43] BOPOMOFO LETTER B..BOPOMOFO LETTER NN
    Interval{ 0x3131, 0x318E }, // [94] HANGUL LETTER KIYEOK..HANGUL LETTER ARAEAE
    Interval{ 0x31A0, 0x31BF }, // [32] BOPOMOFO LETTER BU..BOPOMOFO LETTER AH
    Interval{ 0x31F0, 0x31FF }, // [16] KATAKANA LETTER SMALL KU..KATAKANA LETTER SMALL RO
    Interval{ 0x3400, 0x4DBF }, // [6592] CJK UNIFIED IDEOGRAPH-3400..CJK UNIFIED IDEOGRAPH-4DBF
    Interval{ 0x4E00, 0xA014 }, // [21013] CJK UNIFIED IDEOGRAPH-4E00..YI SYLLABLE E
    Interval{ 0xA016, 0xA48C }, // [1143] YI SYLLABLE BIT..YI SYLLABLE YYR
    Interval{ 0xA4D0, 0xA4F7 }, // [40] LISU LETTER BA..LISU LETTER OE
    Interval{ 0xA500, 0xA60B }, // [268] VAI SYLLABLE EE..VAI SYLLABLE NG
    Interval{ 0xA610, 0xA61F }, // [16] VAI SYLLABLE NDOLE FA..VAI SYMBOL JONG
    Interval{ 0xA62A, 0xA62B }, // [2] VAI SYLLABLE NDOLE MA..VAI SYLLABLE NDOLE DO
    Interval{ 0xA66E, 0xA66E }, // CYRILLIC LETTER MULTIOCULAR O
    Interval{ 0xA6A0, 0xA6E5 }, // [70] BAMUM LETTER A..BAMUM LETTER KI
    Interval{ 0xA78F, 0xA78F }, // LATIN LETTER SINOLOGICAL DOT
    Interval{ 0xA7F7, 0xA7F7 }, // LATIN EPIGRAPHIC LETTER SIDEWAYS I
    Interval{ 0xA7FB, 0xA801 }, // [7] LATIN EPIGRAPHIC LETTER REVERSED F..SYLOTI NAGRI LETTER I
    Interval{ 0xA803, 0xA805 }, // [3] SYLOTI NAGRI LETTER U..SYLOTI NAGRI LETTER O
    Interval{ 0xA807, 0xA80A }, // [4] SYLOTI NAGRI LETTER KO..SYLOTI NAGRI LETTER GHO
    Interval{ 0xA80C, 0xA822 }, // [23] SYLOTI NAGRI LETTER CO..SYLOTI NAGRI LETTER HO
    Interval{ 0xA840, 0xA873 }, // [52] PHAGS-PA LETTER KA..PHAGS-PA LETTER CANDRABINDU
    Interval{ 0xA882, 0xA8B3 }, // [50] SAURASHTRA LETTER A..SAURASHTRA LETTER LLA
    Interval{ 0xA8F2, 0xA8F7 }, // [6] DEVANAGARI SIGN SPACING CANDRABINDU..DEVANAGARI SIGN CANDRABINDU AVAGRAHA
    Interval{ 0xA8FB, 0xA8FB }, // DEVANAGARI HEADSTROKE
    Interval{ 0xA8FD, 0xA8FE }, // [2] DEVANAGARI JAIN OM..DEVANAGARI LETTER AY
    Interval{ 0xA90A, 0xA925 }, // [28] KAYAH LI LETTER KA..KAYAH LI LETTER OO
    Interval{ 0xA930, 0xA946 }, // [23] REJANG LETTER KA..REJANG LETTER A
    Interval{ 0xA960, 0xA97C }, // [29] HANGUL CHOSEONG TIKEUT-MIEUM..HANGUL CHOSEONG SSANGYEORINHIEUH
    Interval{ 0xA984, 0xA9B2 }, // [47] JAVANESE LETTER A..JAVANESE LETTER HA
    Interval{ 0xA9E0, 0xA9E4 }, // [5] MYANMAR LETTER SHAN GHA..MYANMAR LETTER SHAN BHA
    Interval{ 0xA9E7, 0xA9EF }, // [9] MYANMAR LETTER TAI LAING NYA..MYANMAR LETTER TAI LAING NNA
    Interval{ 0xA9FA, 0xA9FE }, // [5] MYANMAR LETTER TAI LAING LLA..MYANMAR LETTER TAI LAING BHA
    Interval{ 0xAA00, 0xAA28 }, // [41] CHAM LETTER A..CHAM LETTER HA
    Interval{ 0xAA40, 0xAA42 }, // [3] CHAM LETTER FINAL K..CHAM LETTER FINAL NG
    Interval{ 0xAA44, 0xAA4B }, // [8] CHAM LETTER FINAL CH..CHAM LETTER FINAL SS
    Interval{ 0xAA60, 0xAA6F }, // [16] MYANMAR LETTER KHAMTI GA..MYANMAR LETTER KHAMTI FA
    Interval{ 0xAA71, 0xAA76 }, // [6] MYANMAR LETTER KHAMTI XA..MYANMAR LOGOGRAM KHAMTI HM
    Interval{ 0xAA7A, 0xAA7A }, // MYANMAR LETTER AITON RA
    Interval{ 0xAA7E, 0xAAAF }, // [50] MYANMAR LETTER SHWE PALAUNG CHA..TAI VIET LETTER HIGH O
    Interval{ 0xAAB1, 0xAAB1 }, // TAI VIET VOWEL AA
    Interval{ 0xAAB5, 0xAAB6 }, // [2] TAI VIET VOWEL E..TAI VIET VOWEL O
    Interval{ 0xAAB9, 0xAABD }, // [5] TAI VIET VOWEL UEA..TAI VIET VOWEL AN
    Interval{ 0xAAC0, 0xAAC0 }, // TAI VIET TONE MAI NUENG
    Interval{ 0xAAC2, 0xAAC2 }, // TAI VIET TONE MAI SONG
    Interval{ 0xAADB, 0xAADC }, // [2] TAI VIET SYMBOL KON..TAI VIET SYMBOL NUENG
    Interval{ 0xAAE0, 0xAAEA }, // [11] MEETEI MAYEK LETTER E..MEETEI MAYEK LETTER SSA
    Interval{ 0xAAF2, 0xAAF2 }, // MEETEI MAYEK ANJI
    Interval{ 0xAB01, 0xAB06 }, // [6] ETHIOPIC SYLLABLE TTHU..ETHIOPIC SYLLABLE TTHO
    Interval{ 0xAB09, 0xAB0E }, // [6] ETHIOPIC SYLLABLE DDHU..ETHIOPIC SYLLABLE DDHO
    Interval{ 0xAB11, 0xAB16 }, // [6] ETHIOPIC SYLLABLE DZU..ETHIOPIC SYLLABLE DZO
    Interval{ 0xAB20, 0xAB26 }, // [7] ETHIOPIC SYLLABLE CCHHA..ETHIOPIC SYLLABLE CCHHO
    Interval{ 0xAB28, 0xAB2E }, // [7] ETHIOPIC SYLLABLE BBA..ETHIOPIC SYLLABLE BBO
    Interval{ 0xABC0, 0xABE2 }, // [35] MEETEI MAYEK LETTER KOK..MEETEI MAYEK LETTER I LONSUM
    Interval{ 0xAC00, 0xD7A3 }, // [11172] HANGUL SYLLABLE GA..HANGUL SYLLABLE HIH
    Interval{ 0xD7B0, 0xD7C6 }, // [23] HANGUL JUNGSEONG O-YEO..HANGUL JUNGSEONG ARAEA-E
    Interval{ 0xD7CB, 0xD7FB }, // [49] HANGUL JONGSEONG NIEUN-RIEUL..HANGUL JONGSEONG PHIEUPH-THIEUTH
    Interval{ 0xF900, 0xFA6D }, // [366] CJK COMPATIBILITY IDEOGRAPH-F900..CJK COMPATIBILITY IDEOGRAPH-FA6D
    Interval{ 0xFA70, 0xFAD9 }, // [106] CJK COMPATIBILITY IDEOGRAPH-FA70..CJK COMPATIBILITY IDEOGRAPH-FAD9
    Interval{ 0xFB1D, 0xFB1D }, // HEBREW LETTER YOD WITH HIRIQ
    Interval{ 0xFB1F, 0xFB28 }, // [10] HEBREW LIGATURE YIDDISH YOD YOD PATAH..HEBREW LETTER WIDE TAV
    Interval{ 0xFB2A, 0xFB36 }, // [13] HEBREW LETTER SHIN WITH SHIN DOT..HEBREW LETTER ZAYIN WITH DAGESH
    Interval{ 0xFB38, 0xFB3C }, // [5] HEBREW LETTER TET WITH DAGESH..HEBREW LETTER LAMED WITH DAGESH
    Interval{ 0xFB3E, 0xFB3E }, // HEBREW LETTER MEM WITH DAGESH
    Interval{ 0xFB40, 0xFB41 }, // [2] HEBREW LETTER NUN WITH DAGESH..HEBREW LETTER SAMEKH WITH DAGESH
    Interval{ 0xFB43, 0xFB44 }, // [2] HEBREW LETTER FINAL PE WITH DAGESH..HEBREW LETTER PE WITH DAGESH
    Interval{ 0xFB46, 0xFBB1 }, // [108] HEBREW LETTER TSADI WITH DAGESH..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE FINAL FORM
    Interval{ 0xFBD3, 0xFD3D }, // [363] ARABIC LETTER NG ISOLATED FORM..ARABIC LIGATURE ALEF WITH FATHATAN ISOLATED FORM
    Interval{ 0xFD50, 0xFD8F }, // [64] ARABIC LIGATURE TEH WITH JEEM WITH MEEM INITIAL FORM..ARABIC LIGATURE MEEM WITH KHAH WITH MEEM INITIAL FORM
    Interval{ 0xFD92, 0xFDC7 }, // [54] ARABIC LIGATURE MEEM WITH JEEM WITH KHAH INITIAL FORM..ARABIC LIGATURE NOON WITH JEEM WITH YEH FINAL FORM
    Interval{ 0xFDF0, 0xFDFB }, // [12] ARABIC LIGATURE SALLA USED AS KORANIC STOP SIGN ISOLATED FORM..ARABIC LIGATURE JALLAJALALOUHOU
    Interval{ 0xFE70, 0xFE74 }, // [5] ARABIC FATHATAN ISOLATED FORM..ARABIC KASRATAN ISOLATED FORM
    Interval{ 0xFE76, 0xFEFC }, // [135] ARABIC FATHA ISOLATED FORM..ARABIC LIGATURE LAM WITH ALEF FINAL FORM
    Interval{ 0xFF66, 0xFF6F }, // [10] HALFWIDTH KATAKANA LETTER WO..HALFWIDTH KATAKANA LETTER SMALL TU
    Interval{ 0xFF71, 0xFF9D }, // [45] HALFWIDTH KATAKANA LETTER A..HALFWIDTH KATAKANA LETTER N
    Interval{ 0xFFA0, 0xFFBE }, // [31] HALFWIDTH HANGUL FILLER..HALFWIDTH HANGUL LETTER HIEUH
    Interval{ 0xFFC2, 0xFFC7 }, // [6] HALFWIDTH HANGUL LETTER A..HALFWIDTH HANGUL LETTER E
    Interval{ 0xFFCA, 0xFFCF }, // [6] HALFWIDTH HANGUL LETTER YEO..HALFWIDTH HANGUL LETTER OE
    Interval{ 0xFFD2, 0xFFD7 }, // [6] HALFWIDTH HANGUL LETTER YO..HALFWIDTH HANGUL LETTER YU
    Interval{ 0xFFDA, 0xFFDC }, // [3] HALFWIDTH HANGUL LETTER EU..HALFWIDTH HANGUL LETTER I
    Interval{ 0x10000, 0x1000B }, // [12] LINEAR B SYLLABLE B008 A..LINEAR B SYLLABLE B046 JE
    Interval{ 0x1000D, 0x10026 }, // [26] LINEAR B SYLLABLE B036 JO..LINEAR B SYLLABLE B032 QO
    Interval{ 0x10028, 0x1003A }, // [19] LINEAR B SYLLABLE B060 RA..LINEAR B SYLLABLE B042 WO
    Interval{ 0x1003C, 0x1003D }, // [2] LINEAR B SYLLABLE B017 ZA..LINEAR B SYLLABLE B074 ZE
    Interval{ 0x1003F, 0x1004D }, // [15] LINEAR B SYLLABLE B020 ZO..LINEAR B SYLLABLE B091 TWO
    Interval{ 0x10050, 0x1005D }, // [14] LINEAR B SYMBOL B018..LINEAR B SYMBOL B089
    Interval{ 0x10080, 0x100FA }, // [123] LINEAR B IDEOGRAM B100 MAN..LINEAR B IDEOGRAM VESSEL B305
    Interval{ 0x10280, 0x1029C }, // [29] LYCIAN LETTER A..LYCIAN LETTER X
    Interval{ 0x102A0, 0x102D0 }, // [49] CARIAN LETTER A..CARIAN LETTER UUU3
    Interval{ 0x10300, 0x1031F }, // [32] OLD ITALIC LETTER A..OLD ITALIC LETTER ESS
    Interval{ 0x1032D, 0x10340 }, // [20] OLD ITALIC LETTER YE..GOTHIC LETTER PAIRTHRA
    Interval{ 0x10342, 0x10349 }, // [8] GOTHIC LETTER RAIDA..GOTHIC LETTER OTHAL
    Interval{ 0x10350, 0x10375 }, // [38] OLD PERMIC LETTER AN..OLD PERMIC LETTER IA
    Interval{ 0x10380, 0x1039D }, // [30] UGARITIC LETTER ALPA..UGARITIC LETTER SSU
    Interval{ 0x103A0, 0x103C3 }, // [36] OLD PERSIAN SIGN A..OLD PERSIAN SIGN HA
    Interval{ 0x103C8, 0x103CF }, // [8] OLD PERSIAN SIGN AURAMAZDAA..OLD PERSIAN SIGN BUUMISH
    Interval{ 0x10450, 0x1049D }, // [78] SHAVIAN LETTER PEEP..OSMANYA LETTER OO
    Interval{ 0x10500, 0x10527 }, // [40] ELBASAN LETTER A..ELBASAN LETTER KHE
    Interval{ 0x10530, 0x10563 }, // [52] CAUCASIAN ALBANIAN LETTER ALT..CAUCASIAN ALBANIAN LETTER KIW
    Interval{ 0x105C0, 0x105F3 }, // [52] TODHRI LETTER A..TODHRI LETTER OO
    Interval{ 0x10600, 0x10736 }, // [311] LINEAR A SIGN AB001..LINEAR A SIGN A664
    Interval{ 0x10740, 0x10755 }, // [22] LINEAR A SIGN A701 A..LINEAR A SIGN A732 JE
    Interval{ 0x10760, 0x10767 }, // [8] LINEAR A SIGN A800..LINEAR A SIGN A807
    Interval{ 0x10800, 0x10805 }, // [6] CYPRIOT SYLLABLE A..CYPRIOT SYLLABLE JA
    Interval{ 0x10808, 0x10808 }, // CYPRIOT SYLLABLE JO
    Interval{ 0x1080A, 0x10835 }, // [44] CYPRIOT SYLLABLE KA..CYPRIOT SYLLABLE WO
    Interval{ 0x10837, 0x10838 }, // [2] CYPRIOT SYLLABLE XA..CYPRIOT SYLLABLE XE
    Interval{ 0x1083C, 0x1083C }, // CYPRIOT SYLLABLE ZA
    Interval{ 0x1083F, 0x10855 }, // [23] CYPRIOT SYLLABLE ZO..IMPERIAL ARAMAIC LETTER TAW
    Interval{ 0x10860, 0x10876 }, // [23] PALMYRENE LETTER ALEPH..PALMYRENE LETTER TAW
    Interval{ 0x10880, 0x1089E }, // [31] NABATAEAN LETTER FINAL ALEPH..NABATAEAN LETTER TAW
    Interval{ 0x108E0, 0x108F2 }, // [19] HATRAN LETTER ALEPH..HATRAN LETTER QOPH
    Interval{ 0x108F4, 0x108F5 }, // [2] HATRAN LETTER SHIN..HATRAN LETTER TAW
    Interval{ 0x10900, 0x10915 }, // [22] PHOENICIAN LETTER ALF..PHOENICIAN LETTER TAU
    Interval{ 0x10920, 0x10939 }, // [26] LYDIAN LETTER A..LYDIAN LETTER C
    Interval{ 0x10940, 0x10959 }, // [26] SIDETIC LETTER N01..SIDETIC LETTER N26
    Interval{ 0x10980, 0x109B7 }, // [56] MEROITIC HIEROGLYPHIC LETTER A..MEROITIC CURSIVE LETTER DA
    Interval{ 0x109BE, 0x109BF }, // [2] MEROITIC CURSIVE LOGOGRAM RMT..MEROITIC CURSIVE LOGOGRAM IMN
    Interval{ 0x10A00, 0x10A00 }, // KHAROSHTHI LETTER A
    Interval{ 0x10A10, 0x10A13 }, // [4] KHAROSHTHI LETTER KA..KHAROSHTHI LETTER GHA
    Interval{ 0x10A15, 0x10A17 }, // [3] KHAROSHTHI LETTER CA..KHAROSHTHI LETTER JA
    Interval{ 0x10A19, 0x10A35 }, // [29] KHAROSHTHI LETTER NYA..KHAROSHTHI LETTER VHA
    Interval{ 0x10A60, 0x10A7C }, // [29] OLD SOUTH ARABIAN LETTER HE..OLD SOUTH ARABIAN LETTER THETH
    Interval{ 0x10A80, 0x10A9C }, // [29] OLD NORTH ARABIAN LETTER HEH..OLD NORTH ARABIAN LETTER ZAH
    Interval{ 0x10AC0, 0x10AC7 }, // [8] MANICHAEAN LETTER ALEPH..MANICHAEAN LETTER WAW
    Interval{ 0x10AC9, 0x10AE4 }, // [28] MANICHAEAN LETTER ZAYIN..MANICHAEAN LETTER TAW
    Interval{ 0x10B00, 0x10B35 }, // [54] AVESTAN LETTER A..AVESTAN LETTER HE
    Interval{ 0x10B40, 0x10B55 }, // [22] INSCRIPTIONAL PARTHIAN LETTER ALEPH..INSCRIPTIONAL PARTHIAN LETTER TAW
    Interval{ 0x10B60, 0x10B72 }, // [19] INSCRIPTIONAL PAHLAVI LETTER ALEPH..INSCRIPTIONAL PAHLAVI LETTER TAW
    Interval{ 0x10B80, 0x10B91 }, // [18] PSALTER PAHLAVI LETTER ALEPH..PSALTER PAHLAVI LETTER TAW
    Interval{ 0x10C00, 0x10C48 }, // [73] OLD TURKIC LETTER ORKHON A..OLD TURKIC LETTER ORKHON BASH
    Interval{ 0x10D00, 0x10D23 }, // [36] HANIFI ROHINGYA LETTER A..HANIFI ROHINGYA MARK NA KHONNA
    Interval{ 0x10D4A, 0x10D4D }, // [4] GARAY VOWEL SIGN A..GARAY VOWEL SIGN EE
    Interval{ 0x10D4F, 0x10D4F }, // GARAY SUKUN
    Interval{ 0x10E80, 0x10EA9 }, // [42] YEZIDI LETTER ELIF..YEZIDI LETTER ET
    Interval{ 0x10EB0, 0x10EB1 }, // [2] YEZIDI LETTER LAM WITH DOT ABOVE..YEZIDI LETTER YOT WITH CIRCUMFLEX ABOVE
    Interval{ 0x10EC2, 0x10EC4 }, // [3] ARABIC LETTER DAL WITH TWO DOTS VERTICALLY BELOW..ARABIC LETTER KAF WITH TWO DOTS VERTICALLY BELOW
    Interval{ 0x10EC6, 0x10EC7 }, // [2] ARABIC LETTER THIN NOON..ARABIC LETTER YEH WITH FOUR DOTS BELOW
    Interval{ 0x10F00, 0x10F1C }, // [29] OLD SOGDIAN LETTER ALEPH..OLD SOGDIAN LETTER FINAL TAW WITH VERTICAL TAIL
    Interval{ 0x10F27, 0x10F27 }, // OLD SOGDIAN LIGATURE AYIN-DALETH
    Interval{ 0x10F30, 0x10F45 }, // [22] SOGDIAN LETTER ALEPH..SOGDIAN INDEPENDENT SHIN
    Interval{ 0x10F70, 0x10F81 }, // [18] OLD UYGHUR LETTER ALEPH..OLD UYGHUR LETTER LESH
    Interval{ 0x10FB0, 0x10FC4 }, // [21] CHORASMIAN LETTER ALEPH..CHORASMIAN LETTER TAW
    Interval{ 0x10FE0, 0x10FF6 }, // [23] ELYMAIC LETTER ALEPH..ELYMAIC LIGATURE ZAYIN-YODH
    Interval{ 0x11003, 0x11037 }, // [53] BRAHMI SIGN JIHVAMULIYA..BRAHMI LETTER OLD TAMIL NNNA
    Interval{ 0x11071, 0x11072 }, // [2] BRAHMI LETTER OLD TAMIL SHORT E..BRAHMI LETTER OLD TAMIL SHORT O
    Interval{ 0x11075, 0x11075 }, // BRAHMI LETTER OLD TAMIL LLA
    Interval{ 0x11083, 0x110AF }, // [45] KAITHI LETTER A..KAITHI LETTER HA
    Interval{ 0x110D0, 0x110E8 }, // [25] SORA SOMPENG LETTER SAH..SORA SOMPENG LETTER MAE
    Interval{ 0x11103, 0x11126 }, // [36] CHAKMA LETTER AA..CHAKMA LETTER HAA
    Interval{ 0x11144, 0x11144 }, // CHAKMA LETTER LHAA
    Interval{ 0x11147, 0x11147 }, // CHAKMA LETTER VAA
    Interval{ 0x11150, 0x11172 }, // [35] MAHAJANI LETTER A..MAHAJANI LETTER RRA
    Interval{ 0x11176, 0x11176 }, // MAHAJANI LIGATURE SHRI
    Interval{ 0x11183, 0x111B2 }, // [48] SHARADA LETTER A..SHARADA LETTER HA
    Interval{ 0x111C1, 0x111C4 }, // [4] SHARADA SIGN AVAGRAHA..SHARADA OM
    Interval{ 0x111DA, 0x111DA }, // SHARADA EKAM
    Interval{ 0x111DC, 0x111DC }, // SHARADA HEADSTROKE
    Interval{ 0x11200, 0x11211 }, // [18] KHOJKI LETTER A..KHOJKI LETTER JJA
    Interval{ 0x11213, 0x1122B }, // [25] KHOJKI LETTER NYA..KHOJKI LETTER LLA
    Interval{ 0x1123F, 0x11240 }, // [2] KHOJKI LETTER QA..KHOJKI LETTER SHORT I
    Interval{ 0x11280, 0x11286 }, // [7] MULTANI LETTER A..MULTANI LETTER GA
    Interval{ 0x11288, 0x11288 }, // MULTANI LETTER GHA
    Interval{ 0x1128A, 0x1128D }, // [4] MULTANI LETTER CA..MULTANI LETTER JJA
    Interval{ 0x1128F, 0x1129D }, // [15] MULTANI LETTER NYA..MULTANI LETTER BA
    Interval{ 0x1129F, 0x112A8 }, // [10] MULTANI LETTER BHA..MULTANI LETTER RHA
    Interval{ 0x112B0, 0x112DE }, // [47] KHUDAWADI LETTER A..KHUDAWADI LETTER HA
    Interval{ 0x11305, 0x1130C }, // [8] GRANTHA LETTER A..GRANTHA LETTER VOCALIC L
    Interval{ 0x1130F, 0x11310 }, // [2] GRANTHA LETTER EE..GRANTHA LETTER AI
    Interval{ 0x11313, 0x11328 }, // [22] GRANTHA LETTER OO..GRANTHA LETTER NA
    Interval{ 0x1132A, 0x11330 }, // [7] GRANTHA LETTER PA..GRANTHA LETTER RA
    Interval{ 0x11332, 0x11333 }, // [2] GRANTHA LETTER LA..GRANTHA LETTER LLA
    Interval{ 0x11335, 0x11339 }, // [5] GRANTHA LETTER VA..GRANTHA LETTER HA
    Interval{ 0x1133D, 0x1133D }, // GRANTHA SIGN AVAGRAHA
    Interval{ 0x11350, 0x11350 }, // GRANTHA OM
    Interval{ 0x1135D, 0x11361 }, // [5] GRANTHA SIGN PLUTA..GRANTHA LETTER VOCALIC LL
    Interval{ 0x11380, 0x11389 }, // [10] TULU-TIGALARI LETTER A..TULU-TIGALARI LETTER VOCALIC LL
    Interval{ 0x1138B, 0x1138B }, // TULU-TIGALARI LETTER EE
    Interval{ 0x1138E, 0x1138E }, // TULU-TIGALARI LETTER AI
    Interval{ 0x11390, 0x113B5 }, // [38] TULU-TIGALARI LETTER OO..TULU-TIGALARI LETTER LLLA
    Interval{ 0x113B7, 0x113B7 }, // TULU-TIGALARI SIGN AVAGRAHA
    Interval{ 0x113D1, 0x113D1 }, // TULU-TIGALARI REPHA
    Interval{ 0x113D3, 0x113D3 }, // TULU-TIGALARI SIGN PLUTA
    Interval{ 0x11400, 0x11434 }, // [53] NEWA LETTER A..NEWA LETTER HA
    Interval{ 0x11447, 0x1144A }, // [4] NEWA SIGN AVAGRAHA..NEWA SIDDHI
    Interval{ 0x1145F, 0x11461 }, // [3] NEWA LETTER VEDIC ANUSVARA..NEWA SIGN UPADHMANIYA
    Interval{ 0x11480, 0x114AF }, // [48] TIRHUTA ANJI..TIRHUTA LETTER HA
    Interval{ 0x114C4, 0x114C5 }, // [2] TIRHUTA SIGN AVAGRAHA..TIRHUTA GVANG
    Interval{ 0x114C7, 0x114C7 }, // TIRHUTA OM
    Interval{ 0x11580, 0x115AE }, // [47] SIDDHAM LETTER A..SIDDHAM LETTER HA
    Interval{ 0x115D8, 0x115DB }, // [4] SIDDHAM LETTER THREE-CIRCLE ALTERNATE I..SIDDHAM LETTER ALTERNATE U
    Interval{ 0x11600, 0x1162F }, // [48] MODI LETTER A..MODI LETTER LLA
    Interval{ 0x11644, 0x11644 }, // MODI SIGN HUVA
    Interval{ 0x11680, 0x116AA }, // [43] TAKRI LETTER A..TAKRI LETTER RRA
    Interval{ 0x116B8, 0x116B8 }, // TAKRI LETTER ARCHAIC KHA
    Interval{ 0x11700, 0x1171A }, // [27] AHOM LETTER KA..AHOM LETTER ALTERNATE BA
    Interval{ 0x11740, 0x11746 }, // [7] AHOM LETTER CA..AHOM LETTER LLA
    Interval{ 0x11800, 0x1182B }, // [44] DOGRA LETTER A..DOGRA LETTER RRA
    Interval{ 0x118FF, 0x11906 }, // [8] WARANG CITI OM..DIVES AKURU LETTER E
    Interval{ 0x11909, 0x11909 }, // DIVES AKURU LETTER O
    Interval{ 0x1190C, 0x11913 }, // [8] DIVES AKURU LETTER KA..DIVES AKURU LETTER JA
    Interval{ 0x11915, 0x11916 }, // [2] DIVES AKURU LETTER NYA..DIVES AKURU LETTER TTA
    Interval{ 0x11918, 0x1192F }, // [24] DIVES AKURU LETTER DDA..DIVES AKURU LETTER ZA
    Interval{ 0x1193F, 0x1193F }, // DIVES AKURU PREFIXED NASAL SIGN
    Interval{ 0x11941, 0x11941 }, // DIVES AKURU INITIAL RA
    Interval{ 0x119A0, 0x119A7 }, // [8] NANDINAGARI LETTER A..NANDINAGARI LETTER VOCALIC RR
    Interval{ 0x119AA, 0x119D0 }, // [39] NANDINAGARI LETTER E..NANDINAGARI LETTER RRA
    Interval{ 0x119E1, 0x119E1 }, // NANDINAGARI SIGN AVAGRAHA
    Interval{ 0x119E3, 0x119E3 }, // NANDINAGARI HEADSTROKE
    Interval{ 0x11A00, 0x11A00 }, // ZANABAZAR SQUARE LETTER A
    Interval{ 0x11A0B, 0x11A32 }, // [40] ZANABAZAR SQUARE LETTER KA..ZANABAZAR SQUARE LETTER KSSA
    Interval{ 0x11A3A, 0x11A3A }, // ZANABAZAR SQUARE CLUSTER-INITIAL LETTER RA
    Interval{ 0x11A50, 0x11A50 }, // SOYOMBO LETTER A
    Interval{ 0x11A5C, 0x11A89 }, // [46] SOYOMBO LETTER KA..SOYOMBO CLUSTER-INITIAL LETTER SA
    Interval{ 0x11A9D, 0x11A9D }, // SOYOMBO MARK PLUTA
    Interval{ 0x11AB0, 0x11AF8 }, // [73] CANADIAN SYLLABICS NATTILIK HI..PAU CIN HAU GLOTTAL STOP FINAL
    Interval{ 0x11BC0, 0x11BE0 }, // [33] SUNUWAR LETTER DEVI..SUNUWAR LETTER KLOKO
    Interval{ 0x11C00, 0x11C08 }, // [9] BHAIKSUKI LETTER A..BHAIKSUKI LETTER VOCALIC L
    Interval{ 0x11C0A, 0x11C2E }, // [37] BHAIKSUKI LETTER E..BHAIKSUKI LETTER HA
    Interval{ 0x11C40, 0x11C40 }, // BHAIKSUKI SIGN AVAGRAHA
    Interval{ 0x11C72, 0x11C8F }, // [30] MARCHEN LETTER KA..MARCHEN LETTER A
    Interval{ 0x11D00, 0x11D06 }, // [7] MASARAM GONDI LETTER A..MASARAM GONDI LETTER E
    Interval{ 0x11D08, 0x11D09 }, // [2] MASARAM GONDI LETTER AI..MASARAM GONDI LETTER O
    Interval{ 0x11D0B, 0x11D30 }, // [38] MASARAM GONDI LETTER AU..MASARAM GONDI LETTER TRA
    Interval{ 0x11D46, 0x11D46 }, // MASARAM GONDI REPHA
    Interval{ 0x11D60, 0x11D65 }, // [6] GUNJALA GONDI LETTER A..GUNJALA GONDI LETTER UU
    Interval{ 0x11D67, 0x11D68 }, // [2] GUNJALA GONDI LETTER EE..GUNJALA GONDI LETTER AI
    Interval{ 0x11D6A, 0x11D89 }, // [32] GUNJALA GONDI LETTER OO..GUNJALA GONDI LETTER SA
    Interval{ 0x11D98, 0x11D98 }, // GUNJALA GONDI OM
    Interval{ 0x11DB0, 0x11DD8 }, // [41] TOLONG SIKI LETTER I..TOLONG SIKI LETTER RRH
    Interval{ 0x11DDA, 0x11DDB }, // [2] TOLONG SIKI SIGN HECAKA..TOLONG SIKI UNGGA
    Interval{ 0x11EE0, 0x11EF2 }, // [19] MAKASAR LETTER KA..MAKASAR ANGKA
    Interval{ 0x11F02, 0x11F02 }, // KAWI SIGN REPHA
    Interval{ 0x11F04, 0x11F10 }, // [13] KAWI LETTER A..KAWI LETTER O
    Interval{ 0x11F12, 0x11F33 }, // [34] KAWI LETTER KA..KAWI LETTER JNYA
    Interval{ 0x11FB0, 0x11FB0 }, // LISU LETTER YHA
    Interval{ 0x12000, 0x12399 }, // [922] CUNEIFORM SIGN A..CUNEIFORM SIGN U U
    Interval{ 0x12480, 0x12543 }, // [196] CUNEIFORM SIGN AB TIMES NUN TENU..CUNEIFORM SIGN ZU5 TIMES THREE DISH TENU
    Interval{ 0x12F90, 0x12FF0 }, // [97] CYPRO-MINOAN SIGN CM001..CYPRO-MINOAN SIGN CM114
    Interval{ 0x13000, 0x1342F }, // [1072] EGYPTIAN HIEROGLYPH A001..EGYPTIAN HIEROGLYPH V011D
    Interval{ 0x13441, 0x13446 }, // [6] EGYPTIAN HIEROGLYPH FULL BLANK..EGYPTIAN HIEROGLYPH WIDE LOST SIGN
    Interval{ 0x13460, 0x143FA }, // [3995] EGYPTIAN HIEROGLYPH-13460..EGYPTIAN HIEROGLYPH-143FA
    Interval{ 0x14400, 0x14646 }, // [583] ANATOLIAN HIEROGLYPH A001..ANATOLIAN HIEROGLYPH A530
    Interval{ 0x16100, 0x1611D }, // [30] GURUNG KHEMA LETTER A..GURUNG KHEMA LETTER SA
    Interval{ 0x16800, 0x16A38 }, // [569] BAMUM LETTER PHASE-A NGKUE MFON..BAMUM LETTER PHASE-F VUEQ
    Interval{ 0x16A40, 0x16A5E }, // [31] MRO LETTER TA..MRO LETTER TEK
    Interval{ 0x16A70, 0x16ABE }, // [79] TANGSA LETTER OZ..TANGSA LETTER ZA
    Interval{ 0x16AD0, 0x16AED }, // [30] BASSA VAH LETTER ENNI..BASSA VAH LETTER I
    Interval{ 0x16B00, 0x16B2F }, // [48] PAHAWH HMONG VOWEL KEEB..PAHAWH HMONG CONSONANT CAU
    Interval{ 0x16B63, 0x16B77 }, // [21] PAHAWH HMONG SIGN VOS LUB..PAHAWH HMONG SIGN CIM NRES TOS
    Interval{ 0x16B7D, 0x16B8F }, // [19] PAHAWH HMONG CLAN SIGN TSHEEJ..PAHAWH HMONG CLAN SIGN VWJ
    Interval{ 0x16D43, 0x16D6A }, // [40] KIRAT RAI LETTER A..KIRAT RAI VOWEL SIGN AU
    Interval{ 0x16F00, 0x16F4A }, // [75] MIAO LETTER PA..MIAO LETTER RTE
    Interval{ 0x16F50, 0x16F50 }, // MIAO LETTER NASALIZATION
    Interval{ 0x17000, 0x18CD5 }, // [7382] TANGUT IDEOGRAPH-17000..KHITAN SMALL SCRIPT CHARACTER-18CD5
    Interval{ 0x18CFF, 0x18D1E }, // [32] KHITAN SMALL SCRIPT CHARACTER-18CFF..TANGUT IDEOGRAPH-18D1E
    Interval{ 0x18D80, 0x18DF2 }, // [115] TANGUT COMPONENT-769..TANGUT COMPONENT-883
    Interval{ 0x1B000, 0x1B122 }, // [291] KATAKANA LETTER ARCHAIC E..KATAKANA LETTER ARCHAIC WU
    Interval{ 0x1B132, 0x1B132 }, // HIRAGANA LETTER SMALL KO
    Interval{ 0x1B150, 0x1B152 }, // [3] HIRAGANA LETTER SMALL WI..HIRAGANA LETTER SMALL WO
    Interval{ 0x1B155, 0x1B155 }, // KATAKANA LETTER SMALL KO
    Interval{ 0x1B164, 0x1B167 }, // [4] KATAKANA LETTER SMALL WI..KATAKANA LETTER SMALL N
    Interval{ 0x1B170, 0x1B2FB }, // [396] NUSHU CHARACTER-1B170..NUSHU CHARACTER-1B2FB
    Interval{ 0x1BC00, 0x1BC6A }, // [107] DUPLOYAN LETTER H..DUPLOYAN LETTER VOCALIC M
    Interval{ 0x1BC70, 0x1BC7C }, // [13] DUPLOYAN AFFIX LEFT HORIZONTAL SECANT..DUPLOYAN AFFIX ATTACHED TANGENT HOOK
    Interval{ 0x1BC80, 0x1BC88 }, // [9] DUPLOYAN AFFIX HIGH ACUTE..DUPLOYAN AFFIX HIGH VERTICAL
    Interval{ 0x1BC90, 0x1BC99 }, // [10] DUPLOYAN AFFIX LOW ACUTE..DUPLOYAN AFFIX LOW ARROW
    Interval{ 0x1DF0A, 0x1DF0A }, // LATIN LETTER RETROFLEX CLICK WITH RETROFLEX HOOK
    Interval{ 0x1E100, 0x1E12C }, // [45] NYIAKENG PUACHUE HMONG LETTER MA..NYIAKENG PUACHUE HMONG LETTER W
    Interval{ 0x1E14E, 0x1E14E }, // NYIAKENG PUACHUE HMONG LOGOGRAM NYAJ
    Interval{ 0x1E290, 0x1E2AD }, // [30] TOTO LETTER PA..TOTO LETTER A
    Interval{ 0x1E2C0, 0x1E2EB }, // [44] WANCHO LETTER AA..WANCHO LETTER YIH
    Interval{ 0x1E4D0, 0x1E4EA }, // [27] NAG MUNDARI LETTER O..NAG MUNDARI LETTER ELL
    Interval{ 0x1E5D0, 0x1E5ED }, // [30] OL ONAL LETTER O..OL ONAL LETTER EG
    Interval{ 0x1E5F0, 0x1E5F0 }, // OL ONAL SIGN HODDOND
    Interval{ 0x1E6C0, 0x1E6DE }, // [31] TAI YO LETTER LOW KO..TAI YO LETTER HIGH KVO
    Interval{ 0x1E6E0, 0x1E6E2 }, // [3] TAI YO LETTER AA..TAI YO LETTER UE
    Interval{ 0x1E6E4, 0x1E6E5 }, // [2] TAI YO LETTER U..TAI YO LETTER AE
    Interval{ 0x1E6E7, 0x1E6ED }, // [7] TAI YO LETTER O..TAI YO LETTER AUE
    Interval{ 0x1E6F0, 0x1E6F4 }, // [5] TAI YO LETTER AN..TAI YO LETTER AP
    Interval{ 0x1E6FE, 0x1E6FE }, // TAI YO SYMBOL MUEANG
    Interval{ 0x1E7E0, 0x1E7E6 }, // [7] ETHIOPIC SYLLABLE HHYA..ETHIOPIC SYLLABLE HHYO
    Interval{ 0x1E7E8, 0x1E7EB }, // [4] ETHIOPIC SYLLABLE GURAGE HHWA..ETHIOPIC SYLLABLE HHWE
    Interval{ 0x1E7ED, 0x1E7EE }, // [2] ETHIOPIC SYLLABLE GURAGE MWI..ETHIOPIC SYLLABLE GURAGE MWEE
    Interval{ 0x1E7F0, 0x1E7FE }, // [15] ETHIOPIC SYLLABLE GURAGE QWI..ETHIOPIC SYLLABLE GURAGE PWEE
    Interval{ 0x1E800, 0x1E8C4 }, // [197] MENDE KIKAKUI SYLLABLE M001 KI..MENDE KIKAKUI SYLLABLE M060 NYON
    Interval{ 0x1EE00, 0x1EE03 }, // [4] ARABIC MATHEMATICAL ALEF..ARABIC MATHEMATICAL DAL
    Interval{ 0x1EE05, 0x1EE1F }, // [27] ARABIC MATHEMATICAL WAW..ARABIC MATHEMATICAL DOTLESS QAF
    Interval{ 0x1EE21, 0x1EE22 }, // [2] ARABIC MATHEMATICAL INITIAL BEH..ARABIC MATHEMATICAL INITIAL JEEM
    Interval{ 0x1EE24, 0x1EE24 }, // ARABIC MATHEMATICAL INITIAL HEH
    Interval{ 0x1EE27, 0x1EE27 }, // ARABIC MATHEMATICAL INITIAL HAH
    Interval{ 0x1EE29, 0x1EE32 }, // [10] ARABIC MATHEMATICAL INITIAL YEH..ARABIC MATHEMATICAL INITIAL QAF
    Interval{ 0x1EE34, 0x1EE37 }, // [4] ARABIC MATHEMATICAL INITIAL SHEEN..ARABIC MATHEMATICAL INITIAL KHAH
    Interval{ 0x1EE39, 0x1EE39 }, // ARABIC MATHEMATICAL INITIAL DAD
    Interval{ 0x1EE3B, 0x1EE3B }, // ARABIC MATHEMATICAL INITIAL GHAIN
    Interval{ 0x1EE42, 0x1EE42 }, // ARABIC MATHEMATICAL TAILED JEEM
    Interval{ 0x1EE47, 0x1EE47 }, // ARABIC MATHEMATICAL TAILED HAH
    Interval{ 0x1EE49, 0x1EE49 }, // ARABIC MATHEMATICAL TAILED YEH
    Interval{ 0x1EE4B, 0x1EE4B }, // ARABIC MATHEMATICAL TAILED LAM
    Interval{ 0x1EE4D, 0x1EE4F }, // [3] ARABIC MATHEMATICAL TAILED NOON..ARABIC MATHEMATICAL TAILED AIN
    Interval{ 0x1EE51, 0x1EE52 }, // [2] ARABIC MATHEMATICAL TAILED SAD..ARABIC MATHEMATICAL TAILED QAF
    Interval{ 0x1EE54, 0x1EE54 }, // ARABIC MATHEMATICAL TAILED SHEEN
    Interval{ 0x1EE57, 0x1EE57 }, // ARABIC MATHEMATICAL TAILED KHAH
    Interval{ 0x1EE59, 0x1EE59 }, // ARABIC MATHEMATICAL TAILED DAD
    Interval{ 0x1EE5B, 0x1EE5B }, // ARABIC MATHEMATICAL TAILED GHAIN
    Interval{ 0x1EE5D, 0x1EE5D }, // ARABIC MATHEMATICAL TAILED DOTLESS NOON
    Interval{ 0x1EE5F, 0x1EE5F }, // ARABIC MATHEMATICAL TAILED DOTLESS QAF
    Interval{ 0x1EE61, 0x1EE62 }, // [2] ARABIC MATHEMATICAL STRETCHED BEH..ARABIC MATHEMATICAL STRETCHED JEEM
    Interval{ 0x1EE64, 0x1EE64 }, // ARABIC MATHEMATICAL STRETCHED HEH
    Interval{ 0x1EE67, 0x1EE6A }, // [4] ARABIC MATHEMATICAL STRETCHED HAH..ARABIC MATHEMATICAL STRETCHED KAF
    Interval{ 0x1EE6C, 0x1EE72 }, // [7] ARABIC MATHEMATICAL STRETCHED MEEM..ARABIC MATHEMATICAL STRETCHED QAF
    Interval{ 0x1EE74, 0x1EE77 }, // [4] ARABIC MATHEMATICAL STRETCHED SHEEN..ARABIC MATHEMATICAL STRETCHED KHAH
    Interval{ 0x1EE79, 0x1EE7C }, // [4] ARABIC MATHEMATICAL STRETCHED DAD..ARABIC MATHEMATICAL STRETCHED DOTLESS BEH
    Interval{ 0x1EE7E, 0x1EE7E }, // ARABIC MATHEMATICAL STRETCHED DOTLESS FEH
    Interval{ 0x1EE80, 0x1EE89 }, // [10] ARABIC MATHEMATICAL LOOPED ALEF..ARABIC MATHEMATICAL LOOPED YEH
    Interval{ 0x1EE8B, 0x1EE9B }, // [17] ARABIC MATHEMATICAL LOOPED LAM..ARABIC MATHEMATICAL LOOPED GHAIN
    Interval{ 0x1EEA1, 0x1EEA3 }, // [3] ARABIC MATHEMATICAL DOUBLE-STRUCK BEH..ARABIC MATHEMATICAL DOUBLE-STRUCK DAL
    Interval{ 0x1EEA5, 0x1EEA9 }, // [5] ARABIC MATHEMATICAL DOUBLE-STRUCK WAW..ARABIC MATHEMATICAL DOUBLE-STRUCK YEH
    Interval{ 0x1EEAB, 0x1EEBB }, // [17] ARABIC MATHEMATICAL DOUBLE-STRUCK LAM..ARABIC MATHEMATICAL DOUBLE-STRUCK GHAIN
    Interval{ 0x20000, 0x2A6DF }, // [42720] CJK UNIFIED IDEOGRAPH-20000..CJK UNIFIED IDEOGRAPH-2A6DF
    Interval{ 0x2A700, 0x2B81D }, // [4382] CJK UNIFIED IDEOGRAPH-2A700..CJK UNIFIED IDEOGRAPH-2B81D
    Interval{ 0x2B820, 0x2CEAD }, // [5774] CJK UNIFIED IDEOGRAPH-2B820..CJK UNIFIED IDEOGRAPH-2CEAD
    Interval{ 0x2CEB0, 0x2EBE0 }, // [7473] CJK UNIFIED IDEOGRAPH-2CEB0..CJK UNIFIED IDEOGRAPH-2EBE0
    Interval{ 0x2EBF0, 0x2EE5D }, // [622] CJK UNIFIED IDEOGRAPH-2EBF0..CJK UNIFIED IDEOGRAPH-2EE5D
    Interval{ 0x2F800, 0x2FA1D }, // [542] CJK COMPATIBILITY IDEOGRAPH-2F800..CJK COMPATIBILITY IDEOGRAPH-2FA1D
    Interval{ 0x30000, 0x3134A }, // [4939] CJK UNIFIED IDEOGRAPH-30000..CJK UNIFIED IDEOGRAPH-3134A
    Interval{ 0x31350, 0x33479 }, // [8490] CJK UNIFIED IDEOGRAPH-31350..CJK UNIFIED IDEOGRAPH-33479
}; // }}}
auto static const Other_Number = std::array<Interval, 72>{ // {{{
    Interval{ 0x00B2, 0x00B3 }, // [2] SUPERSCRIPT TWO..SUPERSCRIPT THREE
    Interval{ 0x00B9, 0x00B9 }, // SUPERSCRIPT ONE
    Interval{ 0x00BC, 0x00BE }, // [3] VULGAR FRACTION ONE QUARTER..VULGAR FRACTION THREE QUARTERS
    Interval{ 0x09F4, 0x09F9 }, // [6] BENGALI CURRENCY NUMERATOR ONE..BENGALI CURRENCY DENOMINATOR SIXTEEN
    Interval{ 0x0B72, 0x0B77 }, // [6] ORIYA FRACTION ONE QUARTER..ORIYA FRACTION THREE SIXTEENTHS
    Interval{ 0x0BF0, 0x0BF2 }, // [3] TAMIL NUMBER TEN..TAMIL NUMBER ONE THOUSAND
    Interval{ 0x0C78, 0x0C7E }, // [7] TELUGU FRACTION DIGIT ZERO FOR ODD POWERS OF FOUR..TELUGU FRACTION DIGIT THREE FOR EVEN POWERS OF FOUR
    Interval{ 0x0D58, 0x0D5E }, // [7] MALAYALAM FRACTION ONE ONE-HUNDRED-AND-SIXTIETH..MALAYALAM FRACTION ONE FIFTH
    Interval{ 0x0D70, 0x0D78 }, // [9] MALAYALAM NUMBER TEN..MALAYALAM FRACTION THREE SIXTEENTHS
    Interval{ 0x0F2A, 0x0F33 }, // [10] TIBETAN DIGIT HALF ONE..TIBETAN DIGIT HALF ZERO
    Interval{ 0x1369, 0x137C }, // [20] ETHIOPIC DIGIT ONE..ETHIOPIC NUMBER TEN THOUSAND
    Interval{ 0x17F0, 0x17F9 }, // [10] KHMER SYMBOL LEK ATTAK SON..KHMER SYMBOL LEK ATTAK PRAM-BUON
    Interval{ 0x19DA, 0x19DA }, // NEW TAI LUE THAM DIGIT ONE
    Interval{ 0x2070, 0x2070 }, // SUPERSCRIPT ZERO
    Interval{ 0x2074, 0x2079 }, // [6] SUPERSCRIPT FOUR..SUPERSCRIPT NINE
    Interval{ 0x2080, 0x2089 }, // [10] SUBSCRIPT ZERO..SUBSCRIPT NINE
    Interval{ 0x2150, 0x215F }, // [16] VULGAR FRACTION ONE SEVENTH..FRACTION NUMERATOR ONE
    Interval{ 0x2189, 0x2189 }, // VULGAR FRACTION ZERO THIRDS
    Interval{ 0x2460, 0x249B }, // [60] CIRCLED DIGIT ONE..NUMBER TWENTY FULL STOP
    Interval{ 0x24EA, 0x24FF }, // [22] CIRCLED DIGIT ZERO..NEGATIVE CIRCLED DIGIT ZERO
    Interval{ 0x2776, 0x2793 }, // [30] DINGBAT NEGATIVE CIRCLED DIGIT ONE..DINGBAT NEGATIVE CIRCLED SANS-SERIF NUMBER TEN
    Interval{ 0x2CFD, 0x2CFD }, // COPTIC FRACTION ONE HALF
    Interval{ 0x3192, 0x3195 }, // [4] IDEOGRAPHIC ANNOTATION ONE MARK..IDEOGRAPHIC ANNOTATION FOUR MARK
    Interval{ 0x3220, 0x3229 }, // [10] PARENTHESIZED IDEOGRAPH ONE..PARENTHESIZED IDEOGRAPH TEN
    Interval{ 0x3248, 0x324F }, // [8] CIRCLED NUMBER TEN ON BLACK SQUARE..CIRCLED NUMBER EIGHTY ON BLACK SQUARE
    Interval{ 0x3251, 0x325F }, // [15] CIRCLED NUMBER TWENTY ONE..CIRCLED NUMBER THIRTY FIVE
    Interval{ 0x3280, 0x3289 }, // [10] CIRCLED IDEOGRAPH ONE..CIRCLED IDEOGRAPH TEN
    Interval{ 0x32B1, 0x32BF }, // [15] CIRCLED NUMBER THIRTY SIX..CIRCLED NUMBER FIFTY
    Interval{ 0xA830, 0xA835 }, // [6] NORTH INDIC FRACTION ONE QUARTER..NORTH INDIC FRACTION THREE SIXTEENTHS
    Interval{ 0x10107, 0x10133 }, // [45] AEGEAN NUMBER ONE..AEGEAN NUMBER NINETY THOUSAND
    Interval{ 0x10175, 0x10178 }, // [4] GREEK ONE HALF SIGN..GREEK THREE QUARTERS SIGN
    Interval{ 0x1018A, 0x1018B }, // [2] GREEK ZERO SIGN..GREEK ONE QUARTER SIGN
    Interval{ 0x102E1, 0x102FB }, // [27] COPTIC EPACT DIGIT ONE..COPTIC EPACT NUMBER NINE HUNDRED
    Interval{ 0x10320, 0x10323 }, // [4] OLD ITALIC NUMERAL ONE..OLD ITALIC NUMERAL FIFTY
    Interval{ 0x10858, 0x1085F }, // [8] IMPERIAL ARAMAIC NUMBER ONE..IMPERIAL ARAMAIC NUMBER TEN THOUSAND
    Interval{ 0x10879, 0x1087F }, // [7] PALMYRENE NUMBER ONE..PALMYRENE NUMBER TWENTY
    Interval{ 0x108A7, 0x108AF }, // [9] NABATAEAN NUMBER ONE..NABATAEAN NUMBER ONE HUNDRED
    Interval{ 0x108FB, 0x108FF }, // [5] HATRAN NUMBER ONE..HATRAN NUMBER ONE HUNDRED
    Interval{ 0x10916, 0x1091B }, // [6] PHOENICIAN NUMBER ONE..PHOENICIAN NUMBER THREE
    Interval{ 0x109BC, 0x109BD }, // [2] MEROITIC CURSIVE FRACTION ELEVEN TWELFTHS..MEROITIC CURSIVE FRACTION ONE HALF
    Interval{ 0x109C0, 0x109CF }, // [16] MEROITIC CURSIVE NUMBER ONE..MEROITIC CURSIVE NUMBER SEVENTY
    Interval{ 0x109D2, 0x109FF }, // [46] MEROITIC CURSIVE NUMBER ONE HUNDRED..MEROITIC CURSIVE FRACTION TEN TWELFTHS
    Interval{ 0x10A40, 0x10A48 }, // [9] KHAROSHTHI DIGIT ONE..KHAROSHTHI FRACTION ONE HALF
    Interval{ 0x10A7D, 0x10A7E }, // [2] OLD SOUTH ARABIAN NUMBER ONE..OLD SOUTH ARABIAN NUMBER FIFTY
    Interval{ 0x10A9D, 0x10A9F }, // [3] OLD NORTH ARABIAN NUMBER ONE..OLD NORTH ARABIAN NUMBER TWENTY
    Interval{ 0x10AEB, 0x10AEF }, // [5] MANICHAEAN NUMBER ONE..MANICHAEAN NUMBER ONE HUNDRED
    Interval{ 0x10B58, 0x10B5F }, // [8] INSCRIPTIONAL PARTHIAN NUMBER ONE..INSCRIPTIONAL PARTHIAN NUMBER ONE THOUSAND
    Interval{ 0x10B78, 0x10B7F }, // [8] INSCRIPTIONAL PAHLAVI NUMBER ONE..INSCRIPTIONAL PAHLAVI NUMBER ONE THOUSAND
    Interval{ 0x10BA9, 0x10BAF }, // [7] PSALTER PAHLAVI NUMBER ONE..PSALTER PAHLAVI NUMBER ONE HUNDRED
    Interval{ 0x10CFA, 0x10CFF }, // [6] OLD HUNGARIAN NUMBER ONE..OLD HUNGARIAN NUMBER ONE THOUSAND
    Interval{ 0x10E60, 0x10E7E }, // [31] RUMI DIGIT ONE..RUMI FRACTION TWO THIRDS
    Interval{ 0x10F1D, 0x10F26 }, // [10] OLD SOGDIAN NUMBER ONE..OLD SOGDIAN FRACTION ONE HALF
    Interval{ 0x10F51, 0x10F54 }, // [4] SOGDIAN NUMBER ONE..SOGDIAN NUMBER ONE HUNDRED
    Interval{ 0x10FC5, 0x10FCB }, // [7] CHORASMIAN NUMBER ONE..CHORASMIAN NUMBER ONE HUNDRED
    Interval{ 0x11052, 0x11065 }, // [20] BRAHMI NUMBER ONE..BRAHMI NUMBER ONE THOUSAND
    Interval{ 0x111E1, 0x111F4 }, // [20] SINHALA ARCHAIC DIGIT ONE..SINHALA ARCHAIC NUMBER ONE THOUSAND
    Interval{ 0x1173A, 0x1173B }, // [2] AHOM NUMBER TEN..AHOM NUMBER TWENTY
    Interval{ 0x118EA, 0x118F2 }, // [9] WARANG CITI NUMBER TEN..WARANG CITI NUMBER NINETY
    Interval{ 0x11C5A, 0x11C6C }, // [19] BHAIKSUKI NUMBER ONE..BHAIKSUKI HUNDREDS UNIT MARK
    Interval{ 0x11FC0, 0x11FD4 }, // [21] TAMIL FRACTION ONE THREE-HUNDRED-AND-TWENTIETH..TAMIL FRACTION DOWNSCALING FACTOR KIIZH
    Interval{ 0x16B5B, 0x16B61 }, // [7] PAHAWH HMONG NUMBER TENS..PAHAWH HMONG NUMBER TRILLIONS
    Interval{ 0x16E80, 0x16E96 }, // [23] MEDEFAIDRIN DIGIT ZERO..MEDEFAIDRIN DIGIT THREE ALTERNATE FORM
    Interval{ 0x1D2C0, 0x1D2D3 }, // [20] KAKTOVIK NUMERAL ZERO..KAKTOVIK NUMERAL NINETEEN
    Interval{ 0x1D2E0, 0x1D2F3 }, // [20] MAYAN NUMERAL ZERO..MAYAN NUMERAL NINETEEN
    Interval{ 0x1D360, 0x1D378 }, // [25] COUNTING ROD UNIT DIGIT ONE..TALLY MARK FIVE
    Interval{ 0x1E8C7, 0x1E8CF }, // [9] MENDE KIKAKUI DIGIT ONE..MENDE KIKAKUI DIGIT NINE
    Interval{ 0x1EC71, 0x1ECAB }, // [59] INDIC SIYAQ NUMBER ONE..INDIC SIYAQ NUMBER PREFIXED NINE
    Interval{ 0x1ECAD, 0x1ECAF }, // [3] INDIC SIYAQ FRACTION ONE QUARTER..INDIC SIYAQ FRACTION THREE QUARTERS
    Interval{ 0x1ECB1, 0x1ECB4 }, // [4] INDIC SIYAQ NUMBER ALTERNATE ONE..INDIC SIYAQ ALTERNATE LAKH MARK
    Interval{ 0x1ED01, 0x1ED2D }, // [45] OTTOMAN SIYAQ NUMBER ONE..OTTOMAN SIYAQ NUMBER NINETY THOUSAND
    Interval{ 0x1ED2F, 0x1ED3D }, // [15] OTTOMAN SIYAQ ALTERNATE NUMBER TWO..OTTOMAN SIYAQ FRACTION ONE SIXTH
    Interval{ 0x1F100, 0x1F10C }, // [13] DIGIT ZERO FULL STOP..DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT ZERO
}; // }}}
auto static const Other_Punctuation = std::array<Interval, 194>{ // {{{
    Interval{ 0x0021, 0x0023 }, // [3] EXCLAMATION MARK..NUMBER SIGN
    Interval{ 0x0025, 0x0027 }, // [3] PERCENT SIGN..APOSTROPHE
    Interval{ 0x002A, 0x002A }, // ASTERISK
    Interval{ 0x002C, 0x002C }, // COMMA
    Interval{ 0x002E, 0x002F }, // [2] FULL STOP..SOLIDUS
    Interval{ 0x003A, 0x003B }, // [2] COLON..SEMICOLON
    Interval{ 0x003F, 0x0040 }, // [2] QUESTION MARK..COMMERCIAL AT
    Interval{ 0x005C, 0x005C }, // REVERSE SOLIDUS
    Interval{ 0x00A1, 0x00A1 }, // INVERTED EXCLAMATION MARK
    Interval{ 0x00A7, 0x00A7 }, // SECTION SIGN
    Interval{ 0x00B6, 0x00B7 }, // [2] PILCROW SIGN..MIDDLE DOT
    Interval{ 0x00BF, 0x00BF }, // INVERTED QUESTION MARK
    Interval{ 0x037E, 0x037E }, // GREEK QUESTION MARK
    Interval{ 0x0387, 0x0387 }, // GREEK ANO TELEIA
    Interval{ 0x055A, 0x055F }, // [6] ARMENIAN APOSTROPHE..ARMENIAN ABBREVIATION MARK
    Interval{ 0x0589, 0x0589 }, // ARMENIAN FULL STOP
    Interval{ 0x05C0, 0x05C0 }, // HEBREW PUNCTUATION PASEQ
    Interval{ 0x05C3, 0x05C3 }, // HEBREW PUNCTUATION SOF PASUQ
    Interval{ 0x05C6, 0x05C6 }, // HEBREW PUNCTUATION NUN HAFUKHA
    Interval{ 0x05F3, 0x05F4 }, // [2] HEBREW PUNCTUATION GERESH..HEBREW PUNCTUATION GERSHAYIM
    Interval{ 0x0609, 0x060A }, // [2] ARABIC-INDIC PER MILLE SIGN..ARABIC-INDIC PER TEN THOUSAND SIGN
    Interval{ 0x060C, 0x060D }, // [2] ARABIC COMMA..ARABIC DATE SEPARATOR
    Interval{ 0x061B, 0x061B }, // ARABIC SEMICOLON
    Interval{ 0x061D, 0x061F }, // [3] ARABIC END OF TEXT MARK..ARABIC QUESTION MARK
    Interval{ 0x066A, 0x066D }, // [4] ARABIC PERCENT SIGN..ARABIC FIVE POINTED STAR
    Interval{ 0x06D4, 0x06D4 }, // ARABIC FULL STOP
    Interval{ 0x0700, 0x070D }, // [14] SYRIAC END OF PARAGRAPH..SYRIAC HARKLEAN ASTERISCUS
    Interval{ 0x07F7, 0x07F9 }, // [3] NKO SYMBOL GBAKURUNEN..NKO EXCLAMATION MARK
    Interval{ 0x0830, 0x083E }, // [15] SAMARITAN PUNCTUATION NEQUDAA..SAMARITAN PUNCTUATION ANNAAU
    Interval{ 0x085E, 0x085E }, // MANDAIC PUNCTUATION
    Interval{ 0x0964, 0x0965 }, // [2] DEVANAGARI DANDA..DEVANAGARI DOUBLE DANDA
    Interval{ 0x0970, 0x0970 }, // DEVANAGARI ABBREVIATION SIGN
    Interval{ 0x09FD, 0x09FD }, // BENGALI ABBREVIATION SIGN
    Interval{ 0x0A76, 0x0A76 }, // GURMUKHI ABBREVIATION SIGN
    Interval{ 0x0AF0, 0x0AF0 }, // GUJARATI ABBREVIATION SIGN
    Interval{ 0x0C77, 0x0C77 }, // TELUGU SIGN SIDDHAM
    Interval{ 0x0C84, 0x0C84 }, // KANNADA SIGN SIDDHAM
    Interval{ 0x0DF4, 0x0DF4 }, // SINHALA PUNCTUATION KUNDDALIYA
    Interval{ 0x0E4F, 0x0E4F }, // THAI CHARACTER FONGMAN
    Interval{ 0x0E5A, 0x0E5B }, // [2] THAI CHARACTER ANGKHANKHU..THAI CHARACTER KHOMUT
    Interval{ 0x0F04, 0x0F12 }, // [15] TIBETAN MARK INITIAL YIG MGO MDUN MA..TIBETAN MARK RGYA GRAM SHAD
    Interval{ 0x0F14, 0x0F14 }, // TIBETAN MARK GTER TSHEG
    Interval{ 0x0F85, 0x0F85 }, // TIBETAN MARK PALUTA
    Interval{ 0x0FD0, 0x0FD4 }, // [5] TIBETAN MARK BSKA- SHOG GI MGO RGYAN..TIBETAN MARK CLOSING BRDA RNYING YIG MGO SGAB MA
    Interval{ 0x0FD9, 0x0FDA }, // [2] TIBETAN MARK LEADING MCHAN RTAGS..TIBETAN MARK TRAILING MCHAN RTAGS
    Interval{ 0x104A, 0x104F }, // [6] MYANMAR SIGN LITTLE SECTION..MYANMAR SYMBOL GENITIVE
    Interval{ 0x10FB, 0x10FB }, // GEORGIAN PARAGRAPH SEPARATOR
    Interval{ 0x1360, 0x1368 }, // [9] ETHIOPIC SECTION MARK..ETHIOPIC PARAGRAPH SEPARATOR
    Interval{ 0x166E, 0x166E }, // CANADIAN SYLLABICS FULL STOP
    Interval{ 0x16EB, 0x16ED }, // [3] RUNIC SINGLE PUNCTUATION..RUNIC CROSS PUNCTUATION
    Interval{ 0x1735, 0x1736 }, // [2] PHILIPPINE SINGLE PUNCTUATION..PHILIPPINE DOUBLE PUNCTUATION
    Interval{ 0x17D4, 0x17D6 }, // [3] KHMER SIGN KHAN..KHMER SIGN CAMNUC PII KUUH
    Interval{ 0x17D8, 0x17DA }, // [3] KHMER SIGN BEYYAL..KHMER SIGN KOOMUUT
    Interval{ 0x1800, 0x1805 }, // [6] MONGOLIAN BIRGA..MONGOLIAN FOUR DOTS
    Interval{ 0x1807, 0x180A }, // [4] MONGOLIAN SIBE SYLLABLE BOUNDARY MARKER..MONGOLIAN NIRUGU
    Interval{ 0x1944, 0x1945 }, // [2] LIMBU EXCLAMATION MARK..LIMBU QUESTION MARK
    Interval{ 0x1A1E, 0x1A1F }, // [2] BUGINESE PALLAWA..BUGINESE END OF SECTION
    Interval{ 0x1AA0, 0x1AA6 }, // [7] TAI THAM SIGN WIANG..TAI THAM SIGN REVERSED ROTATED RANA
    Interval{ 0x1AA8, 0x1AAD }, // [6] TAI THAM SIGN KAAN..TAI THAM SIGN CAANG
    Interval{ 0x1B4E, 0x1B4F }, // [2] BALINESE INVERTED CARIK SIKI..BALINESE INVERTED CARIK PAREREN
    Interval{ 0x1B5A, 0x1B60 }, // [7] BALINESE PANTI..BALINESE PAMENENG
    Interval{ 0x1B7D, 0x1B7F }, // [3] BALINESE PANTI LANTANG..BALINESE PANTI BAWAK
    Interval{ 0x1BFC, 0x1BFF }, // [4] BATAK SYMBOL BINDU NA METEK..BATAK SYMBOL BINDU PANGOLAT
    Interval{ 0x1C3B, 0x1C3F }, // [5] LEPCHA PUNCTUATION TA-ROL..LEPCHA PUNCTUATION TSHOOK
    Interval{ 0x1C7E, 0x1C7F }, // [2] OL CHIKI PUNCTUATION MUCAAD..OL CHIKI PUNCTUATION DOUBLE MUCAAD
    Interval{ 0x1CC0, 0x1CC7 }, // [8] SUNDANESE PUNCTUATION BINDU SURYA..SUNDANESE PUNCTUATION BINDU BA SATANGA
    Interval{ 0x1CD3, 0x1CD3 }, // VEDIC SIGN NIHSHVASA
    Interval{ 0x2016, 0x2017 }, // [2] DOUBLE VERTICAL LINE..DOUBLE LOW LINE
    Interval{ 0x2020, 0x2027 }, // [8] DAGGER..HYPHENATION POINT
    Interval{ 0x2030, 0x2038 }, // [9] PER MILLE SIGN..CARET
    Interval{ 0x203B, 0x203E }, // [4] REFERENCE MARK..OVERLINE
    Interval{ 0x2041, 0x2043 }, // [3] CARET INSERTION POINT..HYPHEN BULLET
    Interval{ 0x2047, 0x2051 }, // [11] DOUBLE QUESTION MARK..TWO ASTERISKS ALIGNED VERTICALLY
    Interval{ 0x2053, 0x2053 }, // SWUNG DASH
    Interval{ 0x2055, 0x205E }, // [10] FLOWER PUNCTUATION MARK..VERTICAL FOUR DOTS
    Interval{ 0x2CF9, 0x2CFC }, // [4] COPTIC OLD NUBIAN FULL STOP..COPTIC OLD NUBIAN VERSE DIVIDER
    Interval{ 0x2CFE, 0x2CFF }, // [2] COPTIC FULL STOP..COPTIC MORPHOLOGICAL DIVIDER
    Interval{ 0x2D70, 0x2D70 }, // TIFINAGH SEPARATOR MARK
    Interval{ 0x2E00, 0x2E01 }, // [2] RIGHT ANGLE SUBSTITUTION MARKER..RIGHT ANGLE DOTTED SUBSTITUTION MARKER
    Interval{ 0x2E06, 0x2E08 }, // [3] RAISED INTERPOLATION MARKER..DOTTED TRANSPOSITION MARKER
    Interval{ 0x2E0B, 0x2E0B }, // RAISED SQUARE
    Interval{ 0x2E0E, 0x2E16 }, // [9] EDITORIAL CORONIS..DOTTED RIGHT-POINTING ANGLE
    Interval{ 0x2E18, 0x2E19 }, // [2] INVERTED INTERROBANG..PALM BRANCH
    Interval{ 0x2E1B, 0x2E1B }, // TILDE WITH RING ABOVE
    Interval{ 0x2E1E, 0x2E1F }, // [2] TILDE WITH DOT ABOVE..TILDE WITH DOT BELOW
    Interval{ 0x2E2A, 0x2E2E }, // [5] TWO DOTS OVER ONE DOT PUNCTUATION..REVERSED QUESTION MARK
    Interval{ 0x2E30, 0x2E39 }, // [10] RING POINT..TOP HALF SECTION SIGN
    Interval{ 0x2E3C, 0x2E3F }, // [4] STENOGRAPHIC FULL STOP..CAPITULUM
    Interval{ 0x2E41, 0x2E41 }, // REVERSED COMMA
    Interval{ 0x2E43, 0x2E4F }, // [13] DASH WITH LEFT UPTURN..CORNISH VERSE DIVIDER
    Interval{ 0x2E52, 0x2E54 }, // [3] TIRONIAN SIGN CAPITAL ET..MEDIEVAL QUESTION MARK
    Interval{ 0x3001, 0x3003 }, // [3] IDEOGRAPHIC COMMA..DITTO MARK
    Interval{ 0x303D, 0x303D }, // PART ALTERNATION MARK
    Interval{ 0x30FB, 0x30FB }, // KATAKANA MIDDLE DOT
    Interval{ 0xA4FE, 0xA4FF }, // [2] LISU PUNCTUATION COMMA..LISU PUNCTUATION FULL STOP
    Interval{ 0xA60D, 0xA60F }, // [3] VAI COMMA..VAI QUESTION MARK
    Interval{ 0xA673, 0xA673 }, // SLAVONIC ASTERISK
    Interval{ 0xA67E, 0xA67E }, // CYRILLIC KAVYKA
    Interval{ 0xA6F2, 0xA6F7 }, // [6] BAMUM NJAEMLI..BAMUM QUESTION MARK
    Interval{ 0xA874, 0xA877 }, // [4] PHAGS-PA SINGLE HEAD MARK..PHAGS-PA MARK DOUBLE SHAD
    Interval{ 0xA8CE, 0xA8CF }, // [2] SAURASHTRA DANDA..SAURASHTRA DOUBLE DANDA
    Interval{ 0xA8F8, 0xA8FA }, // [3] DEVANAGARI SIGN PUSHPIKA..DEVANAGARI CARET
    Interval{ 0xA8FC, 0xA8FC }, // DEVANAGARI SIGN SIDDHAM
    Interval{ 0xA92E, 0xA92F }, // [2] KAYAH LI SIGN CWI..KAYAH LI SIGN SHYA
    Interval{ 0xA95F, 0xA95F }, // REJANG SECTION MARK
    Interval{ 0xA9C1, 0xA9CD }, // [13] JAVANESE LEFT RERENGGAN..JAVANESE TURNED PADA PISELEH
    Interval{ 0xA9DE, 0xA9DF }, // [2] JAVANESE PADA TIRTA TUMETES..JAVANESE PADA ISEN-ISEN
    Interval{ 0xAA5C, 0xAA5F }, // [4] CHAM PUNCTUATION SPIRAL..CHAM PUNCTUATION TRIPLE DANDA
    Interval{ 0xAADE, 0xAADF }, // [2] TAI VIET SYMBOL HO HOI..TAI VIET SYMBOL KOI KOI
    Interval{ 0xAAF0, 0xAAF1 }, // [2] MEETEI MAYEK CHEIKHAN..MEETEI MAYEK AHANG KHUDAM
    Interval{ 0xABEB, 0xABEB }, // MEETEI MAYEK CHEIKHEI
    Interval{ 0xFE10, 0xFE16 }, // [7] PRESENTATION FORM FOR VERTICAL COMMA..PRESENTATION FORM FOR VERTICAL QUESTION MARK
    Interval{ 0xFE19, 0xFE19 }, // PRESENTATION FORM FOR VERTICAL HORIZONTAL ELLIPSIS
    Interval{ 0xFE30, 0xFE30 }, // PRESENTATION FORM FOR VERTICAL TWO DOT LEADER
    Interval{ 0xFE45, 0xFE46 }, // [2] SESAME DOT..WHITE SESAME DOT
    Interval{ 0xFE49, 0xFE4C }, // [4] DASHED OVERLINE..DOUBLE WAVY OVERLINE
    Interval{ 0xFE50, 0xFE52 }, // [3] SMALL COMMA..SMALL FULL STOP
    Interval{ 0xFE54, 0xFE57 }, // [4] SMALL SEMICOLON..SMALL EXCLAMATION MARK
    Interval{ 0xFE5F, 0xFE61 }, // [3] SMALL NUMBER SIGN..SMALL ASTERISK
    Interval{ 0xFE68, 0xFE68 }, // SMALL REVERSE SOLIDUS
    Interval{ 0xFE6A, 0xFE6B }, // [2] SMALL PERCENT SIGN..SMALL COMMERCIAL AT
    Interval{ 0xFF01, 0xFF03 }, // [3] FULLWIDTH EXCLAMATION MARK..FULLWIDTH NUMBER SIGN
    Interval{ 0xFF05, 0xFF07 }, // [3] FULLWIDTH PERCENT SIGN..FULLWIDTH APOSTROPHE
    Interval{ 0xFF0A, 0xFF0A }, // FULLWIDTH ASTERISK
    Interval{ 0xFF0C, 0xFF0C }, // FULLWIDTH COMMA
    Interval{ 0xFF0E, 0xFF0F }, // [2] FULLWIDTH FULL STOP..FULLWIDTH SOLIDUS
    Interval{ 0xFF1A, 0xFF1B }, // [2] FULLWIDTH COLON..FULLWIDTH SEMICOLON
    Interval{ 0xFF1F, 0xFF20 }, // [2] FULLWIDTH QUESTION MARK..FULLWIDTH COMMERCIAL AT
    Interval{ 0xFF3C, 0xFF3C }, // FULLWIDTH REVERSE SOLIDUS
    Interval{ 0xFF61, 0xFF61 }, // HALFWIDTH IDEOGRAPHIC FULL STOP
    Interval{ 0xFF64, 0xFF65 }, // [2] HALFWIDTH IDEOGRAPHIC COMMA..HALFWIDTH KATAKANA MIDDLE DOT
    Interval{ 0x10100, 0x10102 }, // [3] AEGEAN WORD SEPARATOR LINE..AEGEAN CHECK MARK
    Interval{ 0x1039F, 0x1039F }, // UGARITIC WORD DIVIDER
    Interval{ 0x103D0, 0x103D0 }, // OLD PERSIAN WORD DIVIDER
    Interval{ 0x1056F, 0x1056F }, // CAUCASIAN ALBANIAN CITATION MARK
    Interval{ 0x10857, 0x10857 }, // IMPERIAL ARAMAIC SECTION SIGN
    Interval{ 0x1091F, 0x1091F }, // PHOENICIAN WORD SEPARATOR
    Interval{ 0x1093F, 0x1093F }, // LYDIAN TRIANGULAR MARK
    Interval{ 0x10A50, 0x10A58 }, // [9] KHAROSHTHI PUNCTUATION DOT..KHAROSHTHI PUNCTUATION LINES
    Interval{ 0x10A7F, 0x10A7F }, // OLD SOUTH ARABIAN NUMERIC INDICATOR
    Interval{ 0x10AF0, 0x10AF6 }, // [7] MANICHAEAN PUNCTUATION STAR..MANICHAEAN PUNCTUATION LINE FILLER
    Interval{ 0x10B39, 0x10B3F }, // [7] AVESTAN ABBREVIATION MARK..LARGE ONE RING OVER TWO RINGS PUNCTUATION
    Interval{ 0x10B99, 0x10B9C }, // [4] PSALTER PAHLAVI SECTION MARK..PSALTER PAHLAVI FOUR DOTS WITH DOT
    Interval{ 0x10ED0, 0x10ED0 }, // ARABIC BIBLICAL END OF VERSE
    Interval{ 0x10F55, 0x10F59 }, // [5] SOGDIAN PUNCTUATION TWO VERTICAL BARS..SOGDIAN PUNCTUATION HALF CIRCLE WITH DOT
    Interval{ 0x10F86, 0x10F89 }, // [4] OLD UYGHUR PUNCTUATION BAR..OLD UYGHUR PUNCTUATION FOUR DOTS
    Interval{ 0x11047, 0x1104D }, // [7] BRAHMI DANDA..BRAHMI PUNCTUATION LOTUS
    Interval{ 0x110BB, 0x110BC }, // [2] KAITHI ABBREVIATION SIGN..KAITHI ENUMERATION SIGN
    Interval{ 0x110BE, 0x110C1 }, // [4] KAITHI SECTION MARK..KAITHI DOUBLE DANDA
    Interval{ 0x11140, 0x11143 }, // [4] CHAKMA SECTION MARK..CHAKMA QUESTION MARK
    Interval{ 0x11174, 0x11175 }, // [2] MAHAJANI ABBREVIATION SIGN..MAHAJANI SECTION MARK
    Interval{ 0x111C5, 0x111C8 }, // [4] SHARADA DANDA..SHARADA SEPARATOR
    Interval{ 0x111CD, 0x111CD }, // SHARADA SUTRA MARK
    Interval{ 0x111DB, 0x111DB }, // SHARADA SIGN SIDDHAM
    Interval{ 0x111DD, 0x111DF }, // [3] SHARADA CONTINUATION SIGN..SHARADA SECTION MARK-2
    Interval{ 0x11238, 0x1123D }, // [6] KHOJKI DANDA..KHOJKI ABBREVIATION SIGN
    Interval{ 0x112A9, 0x112A9 }, // MULTANI SECTION MARK
    Interval{ 0x113D4, 0x113D5 }, // [2] TULU-TIGALARI DANDA..TULU-TIGALARI DOUBLE DANDA
    Interval{ 0x113D7, 0x113D8 }, // [2] TULU-TIGALARI SIGN OM PUSHPIKA..TULU-TIGALARI SIGN SHRII PUSHPIKA
    Interval{ 0x1144B, 0x1144F }, // [5] NEWA DANDA..NEWA ABBREVIATION SIGN
    Interval{ 0x1145A, 0x1145B }, // [2] NEWA DOUBLE COMMA..NEWA PLACEHOLDER MARK
    Interval{ 0x1145D, 0x1145D }, // NEWA INSERTION SIGN
    Interval{ 0x114C6, 0x114C6 }, // TIRHUTA ABBREVIATION SIGN
    Interval{ 0x115C1, 0x115D7 }, // [23] SIDDHAM SIGN SIDDHAM..SIDDHAM SECTION MARK WITH CIRCLES AND FOUR ENCLOSURES
    Interval{ 0x11641, 0x11643 }, // [3] MODI DANDA..MODI ABBREVIATION SIGN
    Interval{ 0x11660, 0x1166C }, // [13] MONGOLIAN BIRGA WITH ORNAMENT..MONGOLIAN TURNED SWIRL BIRGA WITH DOUBLE ORNAMENT
    Interval{ 0x116B9, 0x116B9 }, // TAKRI ABBREVIATION SIGN
    Interval{ 0x1173C, 0x1173E }, // [3] AHOM SIGN SMALL SECTION..AHOM SIGN RULAI
    Interval{ 0x1183B, 0x1183B }, // DOGRA ABBREVIATION SIGN
    Interval{ 0x11944, 0x11946 }, // [3] DIVES AKURU DOUBLE DANDA..DIVES AKURU END OF TEXT MARK
    Interval{ 0x119E2, 0x119E2 }, // NANDINAGARI SIGN SIDDHAM
    Interval{ 0x11A3F, 0x11A46 }, // [8] ZANABAZAR SQUARE INITIAL HEAD MARK..ZANABAZAR SQUARE CLOSING DOUBLE-LINED HEAD MARK
    Interval{ 0x11A9A, 0x11A9C }, // [3] SOYOMBO MARK TSHEG..SOYOMBO MARK DOUBLE SHAD
    Interval{ 0x11A9E, 0x11AA2 }, // [5] SOYOMBO HEAD MARK WITH MOON AND SUN AND TRIPLE FLAME..SOYOMBO TERMINAL MARK-2
    Interval{ 0x11B00, 0x11B09 }, // [10] DEVANAGARI HEAD MARK..DEVANAGARI SIGN MINDU
    Interval{ 0x11BE1, 0x11BE1 }, // SUNUWAR SIGN PVO
    Interval{ 0x11C41, 0x11C45 }, // [5] BHAIKSUKI DANDA..BHAIKSUKI GAP FILLER-2
    Interval{ 0x11C70, 0x11C71 }, // [2] MARCHEN HEAD MARK..MARCHEN MARK SHAD
    Interval{ 0x11EF7, 0x11EF8 }, // [2] MAKASAR PASSIMBANG..MAKASAR END OF SECTION
    Interval{ 0x11F43, 0x11F4F }, // [13] KAWI DANDA..KAWI PUNCTUATION CLOSING SPIRAL
    Interval{ 0x11FFF, 0x11FFF }, // TAMIL PUNCTUATION END OF TEXT
    Interval{ 0x12470, 0x12474 }, // [5] CUNEIFORM PUNCTUATION SIGN OLD ASSYRIAN WORD DIVIDER..CUNEIFORM PUNCTUATION SIGN DIAGONAL QUADCOLON
    Interval{ 0x12FF1, 0x12FF2 }, // [2] CYPRO-MINOAN SIGN CM301..CYPRO-MINOAN SIGN CM302
    Interval{ 0x16A6E, 0x16A6F }, // [2] MRO DANDA..MRO DOUBLE DANDA
    Interval{ 0x16AF5, 0x16AF5 }, // BASSA VAH FULL STOP
    Interval{ 0x16B37, 0x16B3B }, // [5] PAHAWH HMONG SIGN VOS THOM..PAHAWH HMONG SIGN VOS FEEM
    Interval{ 0x16B44, 0x16B44 }, // PAHAWH HMONG SIGN XAUS
    Interval{ 0x16D6D, 0x16D6F }, // [3] KIRAT RAI SIGN YUPI..KIRAT RAI DOUBLE DANDA
    Interval{ 0x16E97, 0x16E9A }, // [4] MEDEFAIDRIN COMMA..MEDEFAIDRIN EXCLAMATION OH
    Interval{ 0x16FE2, 0x16FE2 }, // OLD CHINESE HOOK MARK
    Interval{ 0x1BC9F, 0x1BC9F }, // DUPLOYAN PUNCTUATION CHINOOK FULL STOP
    Interval{ 0x1DA87, 0x1DA8B }, // [5] SIGNWRITING COMMA..SIGNWRITING PARENTHESIS
    Interval{ 0x1E5FF, 0x1E5FF }, // OL ONAL ABBREVIATION SIGN
    Interval{ 0x1E95E, 0x1E95F }, // [2] ADLAM INITIAL EXCLAMATION MARK..ADLAM INITIAL QUESTION MARK
}; // }}}
auto static const Other_Symbol = std::array<Interval, 193>{ // {{{
    Interval{ 0x00A6, 0x00A6 }, // BROKEN BAR
    Interval{ 0x00A9, 0x00A9 }, // COPYRIGHT SIGN
    Interval{ 0x00AE, 0x00AE }, // REGISTERED SIGN
    Interval{ 0x00B0, 0x00B0 }, // DEGREE SIGN
    Interval{ 0x0482, 0x0482 }, // CYRILLIC THOUSANDS SIGN
    Interval{ 0x058D, 0x058E }, // [2] RIGHT-FACING ARMENIAN ETERNITY SIGN..LEFT-FACING ARMENIAN ETERNITY SIGN
    Interval{ 0x060E, 0x060F }, // [2] ARABIC POETIC VERSE SIGN..ARABIC SIGN MISRA
    Interval{ 0x06DE, 0x06DE }, // ARABIC START OF RUB EL HIZB
    Interval{ 0x06E9, 0x06E9 }, // ARABIC PLACE OF SAJDAH
    Interval{ 0x06FD, 0x06FE }, // [2] ARABIC SIGN SINDHI AMPERSAND..ARABIC SIGN SINDHI POSTPOSITION MEN
    Interval{ 0x07F6, 0x07F6 }, // NKO SYMBOL OO DENNEN
    Interval{ 0x09FA, 0x09FA }, // BENGALI ISSHAR
    Interval{ 0x0B70, 0x0B70 }, // ORIYA ISSHAR
    Interval{ 0x0BF3, 0x0BF8 }, // [6] TAMIL DAY SIGN..TAMIL AS ABOVE SIGN
    Interval{ 0x0BFA, 0x0BFA }, // TAMIL NUMBER SIGN
    Interval{ 0x0C7F, 0x0C7F }, // TELUGU SIGN TUUMU
    Interval{ 0x0D4F, 0x0D4F }, // MALAYALAM SIGN PARA
    Interval{ 0x0D79, 0x0D79 }, // MALAYALAM DATE MARK
    Interval{ 0x0F01, 0x0F03 }, // [3] TIBETAN MARK GTER YIG MGO TRUNCATED A..TIBETAN MARK GTER YIG MGO -UM GTER TSHEG MA
    Interval{ 0x0F13, 0x0F13 }, // TIBETAN MARK CARET -DZUD RTAGS ME LONG CAN
    Interval{ 0x0F15, 0x0F17 }, // [3] TIBETAN LOGOTYPE SIGN CHAD RTAGS..TIBETAN ASTROLOGICAL SIGN SGRA GCAN -CHAR RTAGS
    Interval{ 0x0F1A, 0x0F1F }, // [6] TIBETAN SIGN RDEL DKAR GCIG..TIBETAN SIGN RDEL DKAR RDEL NAG
    Interval{ 0x0F34, 0x0F34 }, // TIBETAN MARK BSDUS RTAGS
    Interval{ 0x0F36, 0x0F36 }, // TIBETAN MARK CARET -DZUD RTAGS BZHI MIG CAN
    Interval{ 0x0F38, 0x0F38 }, // TIBETAN MARK CHE MGO
    Interval{ 0x0FBE, 0x0FC5 }, // [8] TIBETAN KU RU KHA..TIBETAN SYMBOL RDO RJE
    Interval{ 0x0FC7, 0x0FCC }, // [6] TIBETAN SYMBOL RDO RJE RGYA GRAM..TIBETAN SYMBOL NOR BU BZHI -KHYIL
    Interval{ 0x0FCE, 0x0FCF }, // [2] TIBETAN SIGN RDEL NAG RDEL DKAR..TIBETAN SIGN RDEL NAG GSUM
    Interval{ 0x0FD5, 0x0FD8 }, // [4] RIGHT-FACING SVASTI SIGN..LEFT-FACING SVASTI SIGN WITH DOTS
    Interval{ 0x109E, 0x109F }, // [2] MYANMAR SYMBOL SHAN ONE..MYANMAR SYMBOL SHAN EXCLAMATION
    Interval{ 0x1390, 0x1399 }, // [10] ETHIOPIC TONAL MARK YIZET..ETHIOPIC TONAL MARK KURT
    Interval{ 0x166D, 0x166D }, // CANADIAN SYLLABICS CHI SIGN
    Interval{ 0x1940, 0x1940 }, // LIMBU SIGN LOO
    Interval{ 0x19DE, 0x19FF }, // [34] NEW TAI LUE SIGN LAE..KHMER SYMBOL DAP-PRAM ROC
    Interval{ 0x1B61, 0x1B6A }, // [10] BALINESE MUSICAL SYMBOL DONG..BALINESE MUSICAL SYMBOL DANG GEDE
    Interval{ 0x1B74, 0x1B7C }, // [9] BALINESE MUSICAL SYMBOL RIGHT-HAND OPEN DUG..BALINESE MUSICAL SYMBOL LEFT-HAND OPEN PING
    Interval{ 0x2100, 0x2101 }, // [2] ACCOUNT OF..ADDRESSED TO THE SUBJECT
    Interval{ 0x2103, 0x2106 }, // [4] DEGREE CELSIUS..CADA UNA
    Interval{ 0x2108, 0x2109 }, // [2] SCRUPLE..DEGREE FAHRENHEIT
    Interval{ 0x2114, 0x2114 }, // L B BAR SYMBOL
    Interval{ 0x2116, 0x2117 }, // [2] NUMERO SIGN..SOUND RECORDING COPYRIGHT
    Interval{ 0x211E, 0x2123 }, // [6] PRESCRIPTION TAKE..VERSICLE
    Interval{ 0x2125, 0x2125 }, // OUNCE SIGN
    Interval{ 0x2127, 0x2127 }, // INVERTED OHM SIGN
    Interval{ 0x2129, 0x2129 }, // TURNED GREEK SMALL LETTER IOTA
    Interval{ 0x212E, 0x212E }, // ESTIMATED SYMBOL
    Interval{ 0x213A, 0x213B }, // [2] ROTATED CAPITAL Q..FACSIMILE SIGN
    Interval{ 0x214A, 0x214A }, // PROPERTY LINE
    Interval{ 0x214C, 0x214D }, // [2] PER SIGN..AKTIESELSKAB
    Interval{ 0x214F, 0x214F }, // SYMBOL FOR SAMARITAN SOURCE
    Interval{ 0x218A, 0x218B }, // [2] TURNED DIGIT TWO..TURNED DIGIT THREE
    Interval{ 0x2195, 0x2199 }, // [5] UP DOWN ARROW..SOUTH WEST ARROW
    Interval{ 0x219C, 0x219F }, // [4] LEFTWARDS WAVE ARROW..UPWARDS TWO HEADED ARROW
    Interval{ 0x21A1, 0x21A2 }, // [2] DOWNWARDS TWO HEADED ARROW..LEFTWARDS ARROW WITH TAIL
    Interval{ 0x21A4, 0x21A5 }, // [2] LEFTWARDS ARROW FROM BAR..UPWARDS ARROW FROM BAR
    Interval{ 0x21A7, 0x21AD }, // [7] DOWNWARDS ARROW FROM BAR..LEFT RIGHT WAVE ARROW
    Interval{ 0x21AF, 0x21CD }, // [31] DOWNWARDS ZIGZAG ARROW..LEFTWARDS DOUBLE ARROW WITH STROKE
    Interval{ 0x21D0, 0x21D1 }, // [2] LEFTWARDS DOUBLE ARROW..UPWARDS DOUBLE ARROW
    Interval{ 0x21D3, 0x21D3 }, // DOWNWARDS DOUBLE ARROW
    Interval{ 0x21D5, 0x21F3 }, // [31] UP DOWN DOUBLE ARROW..UP DOWN WHITE ARROW
    Interval{ 0x2300, 0x2307 }, // [8] DIAMETER SIGN..WAVY LINE
    Interval{ 0x230C, 0x231F }, // [20] BOTTOM RIGHT CROP..BOTTOM RIGHT CORNER
    Interval{ 0x2322, 0x2328 }, // [7] FROWN..KEYBOARD
    Interval{ 0x232B, 0x237B }, // [81] ERASE TO THE LEFT..NOT CHECK MARK
    Interval{ 0x237D, 0x239A }, // [30] SHOULDERED OPEN BOX..CLEAR SCREEN SYMBOL
    Interval{ 0x23B4, 0x23DB }, // [40] TOP SQUARE BRACKET..FUSE
    Interval{ 0x23E2, 0x2429 }, // [72] WHITE TRAPEZIUM..SYMBOL FOR DELETE MEDIUM SHADE FORM
    Interval{ 0x2440, 0x244A }, // [11] OCR HOOK..OCR DOUBLE BACKSLASH
    Interval{ 0x249C, 0x24E9 }, // [78] PARENTHESIZED LATIN SMALL LETTER A..CIRCLED LATIN SMALL LETTER Z
    Interval{ 0x2500, 0x25B6 }, // [183] BOX DRAWINGS LIGHT HORIZONTAL..BLACK RIGHT-POINTING TRIANGLE
    Interval{ 0x25B8, 0x25C0 }, // [9] BLACK RIGHT-POINTING SMALL TRIANGLE..BLACK LEFT-POINTING TRIANGLE
    Interval{ 0x25C2, 0x25F7 }, // [54] BLACK LEFT-POINTING SMALL TRIANGLE..WHITE CIRCLE WITH UPPER RIGHT QUADRANT
    Interval{ 0x2600, 0x266E }, // [111] BLACK SUN WITH RAYS..MUSIC NATURAL SIGN
    Interval{ 0x2670, 0x2767 }, // [248] WEST SYRIAC CROSS..ROTATED FLORAL HEART BULLET
    Interval{ 0x2794, 0x27BF }, // [44] HEAVY WIDE-HEADED RIGHTWARDS ARROW..DOUBLE CURLY LOOP
    Interval{ 0x2800, 0x28FF }, // [256] BRAILLE PATTERN BLANK..BRAILLE PATTERN DOTS-12345678
    Interval{ 0x2B00, 0x2B2F }, // [48] NORTH EAST WHITE ARROW..WHITE VERTICAL ELLIPSE
    Interval{ 0x2B45, 0x2B46 }, // [2] LEFTWARDS QUADRUPLE ARROW..RIGHTWARDS QUADRUPLE ARROW
    Interval{ 0x2B4D, 0x2B73 }, // [39] DOWNWARDS TRIANGLE-HEADED ZIGZAG ARROW..DOWNWARDS TRIANGLE-HEADED ARROW TO BAR
    Interval{ 0x2B76, 0x2BFF }, // [138] NORTH WEST TRIANGLE-HEADED ARROW TO BAR..HELLSCHREIBER PAUSE SYMBOL
    Interval{ 0x2CE5, 0x2CEA }, // [6] COPTIC SYMBOL MI RO..COPTIC SYMBOL SHIMA SIMA
    Interval{ 0x2E50, 0x2E51 }, // [2] CROSS PATTY WITH RIGHT CROSSBAR..CROSS PATTY WITH LEFT CROSSBAR
    Interval{ 0x2E80, 0x2E99 }, // [26] CJK RADICAL REPEAT..CJK RADICAL RAP
    Interval{ 0x2E9B, 0x2EF3 }, // [89] CJK RADICAL CHOKE..CJK RADICAL C-SIMPLIFIED TURTLE
    Interval{ 0x2F00, 0x2FD5 }, // [214] KANGXI RADICAL ONE..KANGXI RADICAL FLUTE
    Interval{ 0x2FF0, 0x2FFF }, // [16] IDEOGRAPHIC DESCRIPTION CHARACTER LEFT TO RIGHT..IDEOGRAPHIC DESCRIPTION CHARACTER ROTATION
    Interval{ 0x3004, 0x3004 }, // JAPANESE INDUSTRIAL STANDARD SYMBOL
    Interval{ 0x3012, 0x3013 }, // [2] POSTAL MARK..GETA MARK
    Interval{ 0x3020, 0x3020 }, // POSTAL MARK FACE
    Interval{ 0x3036, 0x3037 }, // [2] CIRCLED POSTAL MARK..IDEOGRAPHIC TELEGRAPH LINE FEED SEPARATOR SYMBOL
    Interval{ 0x303E, 0x303F }, // [2] IDEOGRAPHIC VARIATION INDICATOR..IDEOGRAPHIC HALF FILL SPACE
    Interval{ 0x3190, 0x3191 }, // [2] IDEOGRAPHIC ANNOTATION LINKING MARK..IDEOGRAPHIC ANNOTATION REVERSE MARK
    Interval{ 0x3196, 0x319F }, // [10] IDEOGRAPHIC ANNOTATION TOP MARK..IDEOGRAPHIC ANNOTATION MAN MARK
    Interval{ 0x31C0, 0x31E5 }, // [38] CJK STROKE T..CJK STROKE SZP
    Interval{ 0x31EF, 0x31EF }, // IDEOGRAPHIC DESCRIPTION CHARACTER SUBTRACTION
    Interval{ 0x3200, 0x321E }, // [31] PARENTHESIZED HANGUL KIYEOK..PARENTHESIZED KOREAN CHARACTER O HU
    Interval{ 0x322A, 0x3247 }, // [30] PARENTHESIZED IDEOGRAPH MOON..CIRCLED IDEOGRAPH KOTO
    Interval{ 0x3250, 0x3250 }, // PARTNERSHIP SIGN
    Interval{ 0x3260, 0x327F }, // [32] CIRCLED HANGUL KIYEOK..KOREAN STANDARD SYMBOL
    Interval{ 0x328A, 0x32B0 }, // [39] CIRCLED IDEOGRAPH MOON..CIRCLED IDEOGRAPH NIGHT
    Interval{ 0x32C0, 0x33FF }, // [320] IDEOGRAPHIC TELEGRAPH SYMBOL FOR JANUARY..SQUARE GAL
    Interval{ 0x4DC0, 0x4DFF }, // [64] HEXAGRAM FOR THE CREATIVE HEAVEN..HEXAGRAM FOR BEFORE COMPLETION
    Interval{ 0xA490, 0xA4C6 }, // [55] YI RADICAL QOT..YI RADICAL KE
    Interval{ 0xA828, 0xA82B }, // [4] SYLOTI NAGRI POETRY MARK-1..SYLOTI NAGRI POETRY MARK-4
    Interval{ 0xA836, 0xA837 }, // [2] NORTH INDIC QUARTER MARK..NORTH INDIC PLACEHOLDER MARK
    Interval{ 0xA839, 0xA839 }, // NORTH INDIC QUANTITY MARK
    Interval{ 0xAA77, 0xAA79 }, // [3] MYANMAR SYMBOL AITON EXCLAMATION..MYANMAR SYMBOL AITON TWO
    Interval{ 0xFBC3, 0xFBD2 }, // [16] ARABIC LIGATURE JALLA WA-ALAA..ARABIC LIGATURE ALAYHI AR-RAHMAH
    Interval{ 0xFD40, 0xFD4F }, // [16] ARABIC LIGATURE RAHIMAHU ALLAAH..ARABIC LIGATURE RAHIMAHUM ALLAAH
    Interval{ 0xFD90, 0xFD91 }, // [2] ARABIC LIGATURE RAHMATU ALLAAHI ALAYH..ARABIC LIGATURE RAHMATU ALLAAHI ALAYHAA
    Interval{ 0xFDC8, 0xFDCF }, // [8] ARABIC LIGATURE RAHIMAHU ALLAAH TAAALAA..ARABIC LIGATURE SALAAMUHU ALAYNAA
    Interval{ 0xFDFD, 0xFDFF }, // [3] ARABIC LIGATURE BISMILLAH AR-RAHMAN AR-RAHEEM..ARABIC LIGATURE AZZA WA JALL
    Interval{ 0xFFE4, 0xFFE4 }, // FULLWIDTH BROKEN BAR
    Interval{ 0xFFE8, 0xFFE8 }, // HALFWIDTH FORMS LIGHT VERTICAL
    Interval{ 0xFFED, 0xFFEE }, // [2] HALFWIDTH BLACK SQUARE..HALFWIDTH WHITE CIRCLE
    Interval{ 0xFFFC, 0xFFFD }, // [2] OBJECT REPLACEMENT CHARACTER..REPLACEMENT CHARACTER
    Interval{ 0x10137, 0x1013F }, // [9] AEGEAN WEIGHT BASE UNIT..AEGEAN MEASURE THIRD SUBUNIT
    Interval{ 0x10179, 0x10189 }, // [17] GREEK YEAR SIGN..GREEK TRYBLION BASE SIGN
    Interval{ 0x1018C, 0x1018E }, // [3] GREEK SINUSOID SIGN..NOMISMA SIGN
    Interval{ 0x10190, 0x1019C }, // [13] ROMAN SEXTANS SIGN..ASCIA SYMBOL
    Interval{ 0x101A0, 0x101A0 }, // GREEK SYMBOL TAU RHO
    Interval{ 0x101D0, 0x101FC }, // [45] PHAISTOS DISC SIGN PEDESTRIAN..PHAISTOS DISC SIGN WAVY BAND
    Interval{ 0x10877, 0x10878 }, // [2] PALMYRENE LEFT-POINTING FLEURON..PALMYRENE RIGHT-POINTING FLEURON
    Interval{ 0x10AC8, 0x10AC8 }, // MANICHAEAN SIGN UD
    Interval{ 0x10ED1, 0x10ED8 }, // [8] ARABIC LIGATURE ALAYHAA AS-SALAATU WAS-SALAAM..ARABIC LIGATURE NAWWARA ALLAAHU MARQADAH
    Interval{ 0x1173F, 0x1173F }, // AHOM SYMBOL VI
    Interval{ 0x11FD5, 0x11FDC }, // [8] TAMIL SIGN NEL..TAMIL SIGN MUKKURUNI
    Interval{ 0x11FE1, 0x11FF1 }, // [17] TAMIL SIGN PAARAM..TAMIL SIGN VAKAIYARAA
    Interval{ 0x16B3C, 0x16B3F }, // [4] PAHAWH HMONG SIGN XYEEM NTXIV..PAHAWH HMONG SIGN XYEEM FAIB
    Interval{ 0x16B45, 0x16B45 }, // PAHAWH HMONG SIGN CIM TSOV ROG
    Interval{ 0x1BC9C, 0x1BC9C }, // DUPLOYAN SIGN O WITH CROSS
    Interval{ 0x1CC00, 0x1CCEF }, // [240] UP-POINTING GO-KART..OUTLINED LATIN CAPITAL LETTER Z
    Interval{ 0x1CCFA, 0x1CCFC }, // [3] SNAKE SYMBOL..NOSE SYMBOL
    Interval{ 0x1CD00, 0x1CEB3 }, // [436] BLOCK OCTANT-3..BLACK RIGHT TRIANGLE CARET
    Interval{ 0x1CEBA, 0x1CED0 }, // [23] FRAGILE SYMBOL..LEUKOTHEA
    Interval{ 0x1CEE0, 0x1CEEF }, // [16] GEOMANTIC FIGURE POPULUS..GEOMANTIC FIGURE VIA
    Interval{ 0x1CF50, 0x1CFC3 }, // [116] ZNAMENNY NEUME KRYUK..ZNAMENNY NEUME PAUK
    Interval{ 0x1D000, 0x1D0F5 }, // [246] BYZANTINE MUSICAL SYMBOL PSILI..BYZANTINE MUSICAL SYMBOL GORGON NEO KATO
    Interval{ 0x1D100, 0x1D126 }, // [39] MUSICAL SYMBOL SINGLE BARLINE..MUSICAL SYMBOL DRUM CLEF-2
    Interval{ 0x1D129, 0x1D164 }, // [60] MUSICAL SYMBOL MULTIPLE MEASURE REST..MUSICAL SYMBOL ONE HUNDRED TWENTY-EIGHTH NOTE
    Interval{ 0x1D16A, 0x1D16C }, // [3] MUSICAL SYMBOL FINGERED TREMOLO-1..MUSICAL SYMBOL FINGERED TREMOLO-3
    Interval{ 0x1D183, 0x1D184 }, // [2] MUSICAL SYMBOL ARPEGGIATO UP..MUSICAL SYMBOL ARPEGGIATO DOWN
    Interval{ 0x1D18C, 0x1D1A9 }, // [30] MUSICAL SYMBOL RINFORZANDO..MUSICAL SYMBOL DEGREE SLASH
    Interval{ 0x1D1AE, 0x1D1EA }, // [61] MUSICAL SYMBOL PEDAL MARK..MUSICAL SYMBOL KORON
    Interval{ 0x1D200, 0x1D241 }, // [66] GREEK VOCAL NOTATION SYMBOL-1..GREEK INSTRUMENTAL NOTATION SYMBOL-54
    Interval{ 0x1D245, 0x1D245 }, // GREEK MUSICAL LEIMMA
    Interval{ 0x1D300, 0x1D356 }, // [87] MONOGRAM FOR EARTH..TETRAGRAM FOR FOSTERING
    Interval{ 0x1D800, 0x1D9FF }, // [512] SIGNWRITING HAND-FIST INDEX..SIGNWRITING HEAD
    Interval{ 0x1DA37, 0x1DA3A }, // [4] SIGNWRITING AIR BLOW SMALL ROTATIONS..SIGNWRITING BREATH EXHALE
    Interval{ 0x1DA6D, 0x1DA74 }, // [8] SIGNWRITING SHOULDER HIP SPINE..SIGNWRITING TORSO-FLOORPLANE TWISTING
    Interval{ 0x1DA76, 0x1DA83 }, // [14] SIGNWRITING LIMB COMBINATION..SIGNWRITING LOCATION DEPTH
    Interval{ 0x1DA85, 0x1DA86 }, // [2] SIGNWRITING LOCATION TORSO..SIGNWRITING LOCATION LIMBS DIGITS
    Interval{ 0x1E14F, 0x1E14F }, // NYIAKENG PUACHUE HMONG CIRCLED CA
    Interval{ 0x1ECAC, 0x1ECAC }, // INDIC SIYAQ PLACEHOLDER
    Interval{ 0x1ED2E, 0x1ED2E }, // OTTOMAN SIYAQ MARRATAN
    Interval{ 0x1F000, 0x1F02B }, // [44] MAHJONG TILE EAST WIND..MAHJONG TILE BACK
    Interval{ 0x1F030, 0x1F093 }, // [100] DOMINO TILE HORIZONTAL BACK..DOMINO TILE VERTICAL-06-06
    Interval{ 0x1F0A0, 0x1F0AE }, // [15] PLAYING CARD BACK..PLAYING CARD KING OF SPADES
    Interval{ 0x1F0B1, 0x1F0BF }, // [15] PLAYING CARD ACE OF HEARTS..PLAYING CARD RED JOKER
    Interval{ 0x1F0C1, 0x1F0CF }, // [15] PLAYING CARD ACE OF DIAMONDS..PLAYING CARD BLACK JOKER
    Interval{ 0x1F0D1, 0x1F0F5 }, // [37] PLAYING CARD ACE OF CLUBS..PLAYING CARD TRUMP-21
    Interval{ 0x1F10D, 0x1F1AD }, // [161] CIRCLED ZERO WITH SLASH..MASK WORK SYMBOL
    Interval{ 0x1F1E6, 0x1F202 }, // [29] REGIONAL INDICATOR SYMBOL LETTER A..SQUARED KATAKANA SA
    Interval{ 0x1F210, 0x1F23B }, // [44] SQUARED CJK UNIFIED IDEOGRAPH-624B..SQUARED CJK UNIFIED IDEOGRAPH-914D
    Interval{ 0x1F240, 0x1F248 }, // [9] TORTOISE SHELL BRACKETED CJK UNIFIED IDEOGRAPH-672C..TORTOISE SHELL BRACKETED CJK UNIFIED IDEOGRAPH-6557
    Interval{ 0x1F250, 0x1F251 }, // [2] CIRCLED IDEOGRAPH ADVANTAGE..CIRCLED IDEOGRAPH ACCEPT
    Interval{ 0x1F260, 0x1F265 }, // [6] ROUNDED SYMBOL FOR FU..ROUNDED SYMBOL FOR CAI
    Interval{ 0x1F300, 0x1F3FA }, // [251] CYCLONE..AMPHORA
    Interval{ 0x1F400, 0x1F6D8 }, // [729] RAT..LANDSLIDE
    Interval{ 0x1F6DC, 0x1F6EC }, // [17] WIRELESS..AIRPLANE ARRIVING
    Interval{ 0x1F6F0, 0x1F6FC }, // [13] SATELLITE..ROLLER SKATE
    Interval{ 0x1F700, 0x1F7D9 }, // [218] ALCHEMICAL SYMBOL FOR QUINTESSENCE..NINE POINTED WHITE STAR
    Interval{ 0x1F7E0, 0x1F7EB }, // [12] LARGE ORANGE CIRCLE..LARGE BROWN SQUARE
    Interval{ 0x1F7F0, 0x1F7F0 }, // HEAVY EQUALS SIGN
    Interval{ 0x1F800, 0x1F80B }, // [12] LEFTWARDS ARROW WITH SMALL TRIANGLE ARROWHEAD..DOWNWARDS ARROW WITH LARGE TRIANGLE ARROWHEAD
    Interval{ 0x1F810, 0x1F847 }, // [56] LEFTWARDS ARROW WITH SMALL EQUILATERAL ARROWHEAD..DOWNWARDS HEAVY ARROW
    Interval{ 0x1F850, 0x1F859 }, // [10] LEFTWARDS SANS-SERIF ARROW..UP DOWN SANS-SERIF ARROW
    Interval{ 0x1F860, 0x1F887 }, // [40] WIDE-HEADED LEFTWARDS LIGHT BARB ARROW..WIDE-HEADED SOUTH WEST VERY HEAVY BARB ARROW
    Interval{ 0x1F890, 0x1F8AD }, // [30] LEFTWARDS TRIANGLE ARROWHEAD..WHITE ARROW SHAFT WIDTH TWO THIRDS
    Interval{ 0x1F8B0, 0x1F8BB }, // [12] ARROW POINTING UPWARDS THEN NORTH WEST..SOUTH WEST ARROW FROM BAR
    Interval{ 0x1F8C0, 0x1F8C1 }, // [2] LEFTWARDS ARROW FROM DOWNWARDS ARROW..RIGHTWARDS ARROW FROM DOWNWARDS ARROW
    Interval{ 0x1F900, 0x1FA57 }, // [344] CIRCLED CROSS FORMEE WITH FOUR DOTS..BLACK CHESS ALFIL
    Interval{ 0x1FA60, 0x1FA6D }, // [14] XIANGQI RED GENERAL..XIANGQI BLACK SOLDIER
    Interval{ 0x1FA70, 0x1FA7C }, // [13] BALLET SHOES..CRUTCH
    Interval{ 0x1FA80, 0x1FA8A }, // [11] YO-YO..TROMBONE
    Interval{ 0x1FA8E, 0x1FAC6 }, // [57] TREASURE CHEST..FINGERPRINT
    Interval{ 0x1FAC8, 0x1FAC8 }, // HAIRY CREATURE
    Interval{ 0x1FACD, 0x1FADC }, // [16] ORCA..ROOT VEGETABLE
    Interval{ 0x1FADF, 0x1FAEA }, // [12] SPLATTER..DISTORTED FACE
    Interval{ 0x1FAEF, 0x1FAF8 }, // [10] FIGHT CLOUD..RIGHTWARDS PUSHING HAND
    Interval{ 0x1FB00, 0x1FB92 }, // [147] BLOCK SEXTANT-1..UPPER HALF INVERSE MEDIUM SHADE AND LOWER HALF BLOCK
    Interval{ 0x1FB94, 0x1FBEF }, // [92] LEFT HALF INVERSE MEDIUM SHADE AND RIGHT HALF BLOCK..TOP LEFT JUSTIFIED LOWER RIGHT QUARTER BLACK CIRCLE
    Interval{ 0x1FBFA, 0x1FBFA }, // ALARM BELL SYMBOL
}; // }}}
auto static const Paragraph_Separator = std::array<Interval, 1>{ // {{{
    Interval{ 0x2029, 0x2029 }, // PARAGRAPH SEPARATOR
}; // }}}
auto static const Private_Use = std::array<Interval, 3>{ // {{{
    Interval{ 0xE000, 0xF8FF }, // [6400] <private-use-E000>..<private-use-F8FF>
    Interval{ 0xF0000, 0xFFFFD }, // [65534] <private-use-F0000>..<private-use-FFFFD>
    Interval{ 0x100000, 0x10FFFD }, // [65534] <private-use-100000>..<private-use-10FFFD>
}; // }}}
auto static const Space_Separator = std::array<Interval, 7>{ // {{{
    Interval{ 0x0020, 0x0020 }, // SPACE
    Interval{ 0x00A0, 0x00A0 }, // NO-BREAK SPACE
    Interval{ 0x1680, 0x1680 }, // OGHAM SPACE MARK
    Interval{ 0x2000, 0x200A }, // [11] EN QUAD..HAIR SPACE
    Interval{ 0x202F, 0x202F }, // NARROW NO-BREAK SPACE
    Interval{ 0x205F, 0x205F }, // MEDIUM MATHEMATICAL SPACE
    Interval{ 0x3000, 0x3000 }, // IDEOGRAPHIC SPACE
}; // }}}
auto static const Spacing_Mark = std::array<Interval, 193>{ // {{{
    Interval{ 0x0903, 0x0903 }, // DEVANAGARI SIGN VISARGA
    Interval{ 0x093B, 0x093B }, // DEVANAGARI VOWEL SIGN OOE
    Interval{ 0x093E, 0x0940 }, // [3] DEVANAGARI VOWEL SIGN AA..DEVANAGARI VOWEL SIGN II
    Interval{ 0x0949, 0x094C }, // [4] DEVANAGARI VOWEL SIGN CANDRA O..DEVANAGARI VOWEL SIGN AU
    Interval{ 0x094E, 0x094F }, // [2] DEVANAGARI VOWEL SIGN PRISHTHAMATRA E..DEVANAGARI VOWEL SIGN AW
    Interval{ 0x0982, 0x0983 }, // [2] BENGALI SIGN ANUSVARA..BENGALI SIGN VISARGA
    Interval{ 0x09BE, 0x09C0 }, // [3] BENGALI VOWEL SIGN AA..BENGALI VOWEL SIGN II
    Interval{ 0x09C7, 0x09C8 }, // [2] BENGALI VOWEL SIGN E..BENGALI VOWEL SIGN AI
    Interval{ 0x09CB, 0x09CC }, // [2] BENGALI VOWEL SIGN O..BENGALI VOWEL SIGN AU
    Interval{ 0x09D7, 0x09D7 }, // BENGALI AU LENGTH MARK
    Interval{ 0x0A03, 0x0A03 }, // GURMUKHI SIGN VISARGA
    Interval{ 0x0A3E, 0x0A40 }, // [3] GURMUKHI VOWEL SIGN AA..GURMUKHI VOWEL SIGN II
    Interval{ 0x0A83, 0x0A83 }, // GUJARATI SIGN VISARGA
    Interval{ 0x0ABE, 0x0AC0 }, // [3] GUJARATI VOWEL SIGN AA..GUJARATI VOWEL SIGN II
    Interval{ 0x0AC9, 0x0AC9 }, // GUJARATI VOWEL SIGN CANDRA O
    Interval{ 0x0ACB, 0x0ACC }, // [2] GUJARATI VOWEL SIGN O..GUJARATI VOWEL SIGN AU
    Interval{ 0x0B02, 0x0B03 }, // [2] ORIYA SIGN ANUSVARA..ORIYA SIGN VISARGA
    Interval{ 0x0B3E, 0x0B3E }, // ORIYA VOWEL SIGN AA
    Interval{ 0x0B40, 0x0B40 }, // ORIYA VOWEL SIGN II
    Interval{ 0x0B47, 0x0B48 }, // [2] ORIYA VOWEL SIGN E..ORIYA VOWEL SIGN AI
    Interval{ 0x0B4B, 0x0B4C }, // [2] ORIYA VOWEL SIGN O..ORIYA VOWEL SIGN AU
    Interval{ 0x0B57, 0x0B57 }, // ORIYA AU LENGTH MARK
    Interval{ 0x0BBE, 0x0BBF }, // [2] TAMIL VOWEL SIGN AA..TAMIL VOWEL SIGN I
    Interval{ 0x0BC1, 0x0BC2 }, // [2] TAMIL VOWEL SIGN U..TAMIL VOWEL SIGN UU
    Interval{ 0x0BC6, 0x0BC8 }, // [3] TAMIL VOWEL SIGN E..TAMIL VOWEL SIGN AI
    Interval{ 0x0BCA, 0x0BCC }, // [3] TAMIL VOWEL SIGN O..TAMIL VOWEL SIGN AU
    Interval{ 0x0BD7, 0x0BD7 }, // TAMIL AU LENGTH MARK
    Interval{ 0x0C01, 0x0C03 }, // [3] TELUGU SIGN CANDRABINDU..TELUGU SIGN VISARGA
    Interval{ 0x0C41, 0x0C44 }, // [4] TELUGU VOWEL SIGN U..TELUGU VOWEL SIGN VOCALIC RR
    Interval{ 0x0C82, 0x0C83 }, // [2] KANNADA SIGN ANUSVARA..KANNADA SIGN VISARGA
    Interval{ 0x0CBE, 0x0CBE }, // KANNADA VOWEL SIGN AA
    Interval{ 0x0CC0, 0x0CC4 }, // [5] KANNADA VOWEL SIGN II..KANNADA VOWEL SIGN VOCALIC RR
    Interval{ 0x0CC7, 0x0CC8 }, // [2] KANNADA VOWEL SIGN EE..KANNADA VOWEL SIGN AI
    Interval{ 0x0CCA, 0x0CCB }, // [2] KANNADA VOWEL SIGN O..KANNADA VOWEL SIGN OO
    Interval{ 0x0CD5, 0x0CD6 }, // [2] KANNADA LENGTH MARK..KANNADA AI LENGTH MARK
    Interval{ 0x0CF3, 0x0CF3 }, // KANNADA SIGN COMBINING ANUSVARA ABOVE RIGHT
    Interval{ 0x0D02, 0x0D03 }, // [2] MALAYALAM SIGN ANUSVARA..MALAYALAM SIGN VISARGA
    Interval{ 0x0D3E, 0x0D40 }, // [3] MALAYALAM VOWEL SIGN AA..MALAYALAM VOWEL SIGN II
    Interval{ 0x0D46, 0x0D48 }, // [3] MALAYALAM VOWEL SIGN E..MALAYALAM VOWEL SIGN AI
    Interval{ 0x0D4A, 0x0D4C }, // [3] MALAYALAM VOWEL SIGN O..MALAYALAM VOWEL SIGN AU
    Interval{ 0x0D57, 0x0D57 }, // MALAYALAM AU LENGTH MARK
    Interval{ 0x0D82, 0x0D83 }, // [2] SINHALA SIGN ANUSVARAYA..SINHALA SIGN VISARGAYA
    Interval{ 0x0DCF, 0x0DD1 }, // [3] SINHALA VOWEL SIGN AELA-PILLA..SINHALA VOWEL SIGN DIGA AEDA-PILLA
    Interval{ 0x0DD8, 0x0DDF }, // [8] SINHALA VOWEL SIGN GAETTA-PILLA..SINHALA VOWEL SIGN GAYANUKITTA
    Interval{ 0x0DF2, 0x0DF3 }, // [2] SINHALA VOWEL SIGN DIGA GAETTA-PILLA..SINHALA VOWEL SIGN DIGA GAYANUKITTA
    Interval{ 0x0F3E, 0x0F3F }, // [2] TIBETAN SIGN YAR TSHES..TIBETAN SIGN MAR TSHES
    Interval{ 0x0F7F, 0x0F7F }, // TIBETAN SIGN RNAM BCAD
    Interval{ 0x102B, 0x102C }, // [2] MYANMAR VOWEL SIGN TALL AA..MYANMAR VOWEL SIGN AA
    Interval{ 0x1031, 0x1031 }, // MYANMAR VOWEL SIGN E
    Interval{ 0x1038, 0x1038 }, // MYANMAR SIGN VISARGA
    Interval{ 0x103B, 0x103C }, // [2] MYANMAR CONSONANT SIGN MEDIAL YA..MYANMAR CONSONANT SIGN MEDIAL RA
    Interval{ 0x1056, 0x1057 }, // [2] MYANMAR VOWEL SIGN VOCALIC R..MYANMAR VOWEL SIGN VOCALIC RR
    Interval{ 0x1062, 0x1064 }, // [3] MYANMAR VOWEL SIGN SGAW KAREN EU..MYANMAR TONE MARK SGAW KAREN KE PHO
    Interval{ 0x1067, 0x106D }, // [7] MYANMAR VOWEL SIGN WESTERN PWO KAREN EU..MYANMAR SIGN WESTERN PWO KAREN TONE-5
    Interval{ 0x1083, 0x1084 }, // [2] MYANMAR VOWEL SIGN SHAN AA..MYANMAR VOWEL SIGN SHAN E
    Interval{ 0x1087, 0x108C }, // [6] MYANMAR SIGN SHAN TONE-2..MYANMAR SIGN SHAN COUNCIL TONE-3
    Interval{ 0x108F, 0x108F }, // MYANMAR SIGN RUMAI PALAUNG TONE-5
    Interval{ 0x109A, 0x109C }, // [3] MYANMAR SIGN KHAMTI TONE-1..MYANMAR VOWEL SIGN AITON A
    Interval{ 0x1715, 0x1715 }, // TAGALOG SIGN PAMUDPOD
    Interval{ 0x1734, 0x1734 }, // HANUNOO SIGN PAMUDPOD
    Interval{ 0x17B6, 0x17B6 }, // KHMER VOWEL SIGN AA
    Interval{ 0x17BE, 0x17C5 }, // [8] KHMER VOWEL SIGN OE..KHMER VOWEL SIGN AU
    Interval{ 0x17C7, 0x17C8 }, // [2] KHMER SIGN REAHMUK..KHMER SIGN YUUKALEAPINTU
    Interval{ 0x1923, 0x1926 }, // [4] LIMBU VOWEL SIGN EE..LIMBU VOWEL SIGN AU
    Interval{ 0x1929, 0x192B }, // [3] LIMBU SUBJOINED LETTER YA..LIMBU SUBJOINED LETTER WA
    Interval{ 0x1930, 0x1931 }, // [2] LIMBU SMALL LETTER KA..LIMBU SMALL LETTER NGA
    Interval{ 0x1933, 0x1938 }, // [6] LIMBU SMALL LETTER TA..LIMBU SMALL LETTER LA
    Interval{ 0x1A19, 0x1A1A }, // [2] BUGINESE VOWEL SIGN E..BUGINESE VOWEL SIGN O
    Interval{ 0x1A55, 0x1A55 }, // TAI THAM CONSONANT SIGN MEDIAL RA
    Interval{ 0x1A57, 0x1A57 }, // TAI THAM CONSONANT SIGN LA TANG LAI
    Interval{ 0x1A61, 0x1A61 }, // TAI THAM VOWEL SIGN A
    Interval{ 0x1A63, 0x1A64 }, // [2] TAI THAM VOWEL SIGN AA..TAI THAM VOWEL SIGN TALL AA
    Interval{ 0x1A6D, 0x1A72 }, // [6] TAI THAM VOWEL SIGN OY..TAI THAM VOWEL SIGN THAM AI
    Interval{ 0x1B04, 0x1B04 }, // BALINESE SIGN BISAH
    Interval{ 0x1B35, 0x1B35 }, // BALINESE VOWEL SIGN TEDUNG
    Interval{ 0x1B3B, 0x1B3B }, // BALINESE VOWEL SIGN RA REPA TEDUNG
    Interval{ 0x1B3D, 0x1B41 }, // [5] BALINESE VOWEL SIGN LA LENGA TEDUNG..BALINESE VOWEL SIGN TALING REPA TEDUNG
    Interval{ 0x1B43, 0x1B44 }, // [2] BALINESE VOWEL SIGN PEPET TEDUNG..BALINESE ADEG ADEG
    Interval{ 0x1B82, 0x1B82 }, // SUNDANESE SIGN PANGWISAD
    Interval{ 0x1BA1, 0x1BA1 }, // SUNDANESE CONSONANT SIGN PAMINGKAL
    Interval{ 0x1BA6, 0x1BA7 }, // [2] SUNDANESE VOWEL SIGN PANAELAENG..SUNDANESE VOWEL SIGN PANOLONG
    Interval{ 0x1BAA, 0x1BAA }, // SUNDANESE SIGN PAMAAEH
    Interval{ 0x1BE7, 0x1BE7 }, // BATAK VOWEL SIGN E
    Interval{ 0x1BEA, 0x1BEC }, // [3] BATAK VOWEL SIGN I..BATAK VOWEL SIGN O
    Interval{ 0x1BEE, 0x1BEE }, // BATAK VOWEL SIGN U
    Interval{ 0x1BF2, 0x1BF3 }, // [2] BATAK PANGOLAT..BATAK PANONGONAN
    Interval{ 0x1C24, 0x1C2B }, // [8] LEPCHA SUBJOINED LETTER YA..LEPCHA VOWEL SIGN UU
    Interval{ 0x1C34, 0x1C35 }, // [2] LEPCHA CONSONANT SIGN NYIN-DO..LEPCHA CONSONANT SIGN KANG
    Interval{ 0x1CE1, 0x1CE1 }, // VEDIC TONE ATHARVAVEDIC INDEPENDENT SVARITA
    Interval{ 0x1CF7, 0x1CF7 }, // VEDIC SIGN ATIKRAMA
    Interval{ 0x302E, 0x302F }, // [2] HANGUL SINGLE DOT TONE MARK..HANGUL DOUBLE DOT TONE MARK
    Interval{ 0xA823, 0xA824 }, // [2] SYLOTI NAGRI VOWEL SIGN A..SYLOTI NAGRI VOWEL SIGN I
    Interval{ 0xA827, 0xA827 }, // SYLOTI NAGRI VOWEL SIGN OO
    Interval{ 0xA880, 0xA881 }, // [2] SAURASHTRA SIGN ANUSVARA..SAURASHTRA SIGN VISARGA
    Interval{ 0xA8B4, 0xA8C3 }, // [16] SAURASHTRA CONSONANT SIGN HAARU..SAURASHTRA VOWEL SIGN AU
    Interval{ 0xA952, 0xA953 }, // [2] REJANG CONSONANT SIGN H..REJANG VIRAMA
    Interval{ 0xA983, 0xA983 }, // JAVANESE SIGN WIGNYAN
    Interval{ 0xA9B4, 0xA9B5 }, // [2] JAVANESE VOWEL SIGN TARUNG..JAVANESE VOWEL SIGN TOLONG
    Interval{ 0xA9BA, 0xA9BB }, // [2] JAVANESE VOWEL SIGN TALING..JAVANESE VOWEL SIGN DIRGA MURE
    Interval{ 0xA9BE, 0xA9C0 }, // [3] JAVANESE CONSONANT SIGN PENGKAL..JAVANESE PANGKON
    Interval{ 0xAA2F, 0xAA30 }, // [2] CHAM VOWEL SIGN O..CHAM VOWEL SIGN AI
    Interval{ 0xAA33, 0xAA34 }, // [2] CHAM CONSONANT SIGN YA..CHAM CONSONANT SIGN RA
    Interval{ 0xAA4D, 0xAA4D }, // CHAM CONSONANT SIGN FINAL H
    Interval{ 0xAA7B, 0xAA7B }, // MYANMAR SIGN PAO KAREN TONE
    Interval{ 0xAA7D, 0xAA7D }, // MYANMAR SIGN TAI LAING TONE-5
    Interval{ 0xAAEB, 0xAAEB }, // MEETEI MAYEK VOWEL SIGN II
    Interval{ 0xAAEE, 0xAAEF }, // [2] MEETEI MAYEK VOWEL SIGN AU..MEETEI MAYEK VOWEL SIGN AAU
    Interval{ 0xAAF5, 0xAAF5 }, // MEETEI MAYEK VOWEL SIGN VISARGA
    Interval{ 0xABE3, 0xABE4 }, // [2] MEETEI MAYEK VOWEL SIGN ONAP..MEETEI MAYEK VOWEL SIGN INAP
    Interval{ 0xABE6, 0xABE7 }, // [2] MEETEI MAYEK VOWEL SIGN YENAP..MEETEI MAYEK VOWEL SIGN SOUNAP
    Interval{ 0xABE9, 0xABEA }, // [2] MEETEI MAYEK VOWEL SIGN CHEINAP..MEETEI MAYEK VOWEL SIGN NUNG
    Interval{ 0xABEC, 0xABEC }, // MEETEI MAYEK LUM IYEK
    Interval{ 0x11000, 0x11000 }, // BRAHMI SIGN CANDRABINDU
    Interval{ 0x11002, 0x11002 }, // BRAHMI SIGN VISARGA
    Interval{ 0x11082, 0x11082 }, // KAITHI SIGN VISARGA
    Interval{ 0x110B0, 0x110B2 }, // [3] KAITHI VOWEL SIGN AA..KAITHI VOWEL SIGN II
    Interval{ 0x110B7, 0x110B8 }, // [2] KAITHI VOWEL SIGN O..KAITHI VOWEL SIGN AU
    Interval{ 0x1112C, 0x1112C }, // CHAKMA VOWEL SIGN E
    Interval{ 0x11145, 0x11146 }, // [2] CHAKMA VOWEL SIGN AA..CHAKMA VOWEL SIGN EI
    Interval{ 0x11182, 0x11182 }, // SHARADA SIGN VISARGA
    Interval{ 0x111B3, 0x111B5 }, // [3] SHARADA VOWEL SIGN AA..SHARADA VOWEL SIGN II
    Interval{ 0x111BF, 0x111C0 }, // [2] SHARADA VOWEL SIGN AU..SHARADA SIGN VIRAMA
    Interval{ 0x111CE, 0x111CE }, // SHARADA VOWEL SIGN PRISHTHAMATRA E
    Interval{ 0x1122C, 0x1122E }, // [3] KHOJKI VOWEL SIGN AA..KHOJKI VOWEL SIGN II
    Interval{ 0x11232, 0x11233 }, // [2] KHOJKI VOWEL SIGN O..KHOJKI VOWEL SIGN AU
    Interval{ 0x11235, 0x11235 }, // KHOJKI SIGN VIRAMA
    Interval{ 0x112E0, 0x112E2 }, // [3] KHUDAWADI VOWEL SIGN AA..KHUDAWADI VOWEL SIGN II
    Interval{ 0x11302, 0x11303 }, // [2] GRANTHA SIGN ANUSVARA..GRANTHA SIGN VISARGA
    Interval{ 0x1133E, 0x1133F }, // [2] GRANTHA VOWEL SIGN AA..GRANTHA VOWEL SIGN I
    Interval{ 0x11341, 0x11344 }, // [4] GRANTHA VOWEL SIGN U..GRANTHA VOWEL SIGN VOCALIC RR
    Interval{ 0x11347, 0x11348 }, // [2] GRANTHA VOWEL SIGN EE..GRANTHA VOWEL SIGN AI
    Interval{ 0x1134B, 0x1134D }, // [3] GRANTHA VOWEL SIGN OO..GRANTHA SIGN VIRAMA
    Interval{ 0x11357, 0x11357 }, // GRANTHA AU LENGTH MARK
    Interval{ 0x11362, 0x11363 }, // [2] GRANTHA VOWEL SIGN VOCALIC L..GRANTHA VOWEL SIGN VOCALIC LL
    Interval{ 0x113B8, 0x113BA }, // [3] TULU-TIGALARI VOWEL SIGN AA..TULU-TIGALARI VOWEL SIGN II
    Interval{ 0x113C2, 0x113C2 }, // TULU-TIGALARI VOWEL SIGN EE
    Interval{ 0x113C5, 0x113C5 }, // TULU-TIGALARI VOWEL SIGN AI
    Interval{ 0x113C7, 0x113CA }, // [4] TULU-TIGALARI VOWEL SIGN OO..TULU-TIGALARI SIGN CANDRA ANUNASIKA
    Interval{ 0x113CC, 0x113CD }, // [2] TULU-TIGALARI SIGN ANUSVARA..TULU-TIGALARI SIGN VISARGA
    Interval{ 0x113CF, 0x113CF }, // TULU-TIGALARI SIGN LOOPED VIRAMA
    Interval{ 0x11435, 0x11437 }, // [3] NEWA VOWEL SIGN AA..NEWA VOWEL SIGN II
    Interval{ 0x11440, 0x11441 }, // [2] NEWA VOWEL SIGN O..NEWA VOWEL SIGN AU
    Interval{ 0x11445, 0x11445 }, // NEWA SIGN VISARGA
    Interval{ 0x114B0, 0x114B2 }, // [3] TIRHUTA VOWEL SIGN AA..TIRHUTA VOWEL SIGN II
    Interval{ 0x114B9, 0x114B9 }, // TIRHUTA VOWEL SIGN E
    Interval{ 0x114BB, 0x114BE }, // [4] TIRHUTA VOWEL SIGN AI..TIRHUTA VOWEL SIGN AU
    Interval{ 0x114C1, 0x114C1 }, // TIRHUTA SIGN VISARGA
    Interval{ 0x115AF, 0x115B1 }, // [3] SIDDHAM VOWEL SIGN AA..SIDDHAM VOWEL SIGN II
    Interval{ 0x115B8, 0x115BB }, // [4] SIDDHAM VOWEL SIGN E..SIDDHAM VOWEL SIGN AU
    Interval{ 0x115BE, 0x115BE }, // SIDDHAM SIGN VISARGA
    Interval{ 0x11630, 0x11632 }, // [3] MODI VOWEL SIGN AA..MODI VOWEL SIGN II
    Interval{ 0x1163B, 0x1163C }, // [2] MODI VOWEL SIGN O..MODI VOWEL SIGN AU
    Interval{ 0x1163E, 0x1163E }, // MODI SIGN VISARGA
    Interval{ 0x116AC, 0x116AC }, // TAKRI SIGN VISARGA
    Interval{ 0x116AE, 0x116AF }, // [2] TAKRI VOWEL SIGN I..TAKRI VOWEL SIGN II
    Interval{ 0x116B6, 0x116B6 }, // TAKRI SIGN VIRAMA
    Interval{ 0x1171E, 0x1171E }, // AHOM CONSONANT SIGN MEDIAL RA
    Interval{ 0x11720, 0x11721 }, // [2] AHOM VOWEL SIGN A..AHOM VOWEL SIGN AA
    Interval{ 0x11726, 0x11726 }, // AHOM VOWEL SIGN E
    Interval{ 0x1182C, 0x1182E }, // [3] DOGRA VOWEL SIGN AA..DOGRA VOWEL SIGN II
    Interval{ 0x11838, 0x11838 }, // DOGRA SIGN VISARGA
    Interval{ 0x11930, 0x11935 }, // [6] DIVES AKURU VOWEL SIGN AA..DIVES AKURU VOWEL SIGN E
    Interval{ 0x11937, 0x11938 }, // [2] DIVES AKURU VOWEL SIGN AI..DIVES AKURU VOWEL SIGN O
    Interval{ 0x1193D, 0x1193D }, // DIVES AKURU SIGN HALANTA
    Interval{ 0x11940, 0x11940 }, // DIVES AKURU MEDIAL YA
    Interval{ 0x11942, 0x11942 }, // DIVES AKURU MEDIAL RA
    Interval{ 0x119D1, 0x119D3 }, // [3] NANDINAGARI VOWEL SIGN AA..NANDINAGARI VOWEL SIGN II
    Interval{ 0x119DC, 0x119DF }, // [4] NANDINAGARI VOWEL SIGN O..NANDINAGARI SIGN VISARGA
    Interval{ 0x119E4, 0x119E4 }, // NANDINAGARI VOWEL SIGN PRISHTHAMATRA E
    Interval{ 0x11A39, 0x11A39 }, // ZANABAZAR SQUARE SIGN VISARGA
    Interval{ 0x11A57, 0x11A58 }, // [2] SOYOMBO VOWEL SIGN AI..SOYOMBO VOWEL SIGN AU
    Interval{ 0x11A97, 0x11A97 }, // SOYOMBO SIGN VISARGA
    Interval{ 0x11B61, 0x11B61 }, // SHARADA VOWEL SIGN OOE
    Interval{ 0x11B65, 0x11B65 }, // SHARADA VOWEL SIGN SHORT O
    Interval{ 0x11B67, 0x11B67 }, // SHARADA VOWEL SIGN CANDRA O
    Interval{ 0x11C2F, 0x11C2F }, // BHAIKSUKI VOWEL SIGN AA
    Interval{ 0x11C3E, 0x11C3E }, // BHAIKSUKI SIGN VISARGA
    Interval{ 0x11CA9, 0x11CA9 }, // MARCHEN SUBJOINED LETTER YA
    Interval{ 0x11CB1, 0x11CB1 }, // MARCHEN VOWEL SIGN I
    Interval{ 0x11CB4, 0x11CB4 }, // MARCHEN VOWEL SIGN O
    Interval{ 0x11D8A, 0x11D8E }, // [5] GUNJALA GONDI VOWEL SIGN AA..GUNJALA GONDI VOWEL SIGN UU
    Interval{ 0x11D93, 0x11D94 }, // [2] GUNJALA GONDI VOWEL SIGN OO..GUNJALA GONDI VOWEL SIGN AU
    Interval{ 0x11D96, 0x11D96 }, // GUNJALA GONDI SIGN VISARGA
    Interval{ 0x11EF5, 0x11EF6 }, // [2] MAKASAR VOWEL SIGN E..MAKASAR VOWEL SIGN O
    Interval{ 0x11F03, 0x11F03 }, // KAWI SIGN VISARGA
    Interval{ 0x11F34, 0x11F35 }, // [2] KAWI VOWEL SIGN AA..KAWI VOWEL SIGN ALTERNATE AA
    Interval{ 0x11F3E, 0x11F3F }, // [2] KAWI VOWEL SIGN E..KAWI VOWEL SIGN AI
    Interval{ 0x11F41, 0x11F41 }, // KAWI SIGN KILLER
    Interval{ 0x1612A, 0x1612C }, // [3] GURUNG KHEMA CONSONANT SIGN MEDIAL YA..GURUNG KHEMA CONSONANT SIGN MEDIAL HA
    Interval{ 0x16F51, 0x16F87 }, // [55] MIAO SIGN ASPIRATION..MIAO VOWEL SIGN UI
    Interval{ 0x16FF0, 0x16FF1 }, // [2] VIETNAMESE ALTERNATE READING MARK CA..VIETNAMESE ALTERNATE READING MARK NHAY
    Interval{ 0x1D165, 0x1D166 }, // [2] MUSICAL SYMBOL COMBINING STEM..MUSICAL SYMBOL COMBINING SPRECHGESANG STEM
    Interval{ 0x1D16D, 0x1D172 }, // [6] MUSICAL SYMBOL COMBINING AUGMENTATION DOT..MUSICAL SYMBOL COMBINING FLAG-5
}; // }}}
auto static const Surrogate = std::array<Interval, 1>{ // {{{
    Interval{ 0xD800, 0xDFFF }, // [2048] <surrogate-D800>..<surrogate-DFFF>
}; // }}}
auto static const Titlecase_Letter = std::array<Interval, 10>{ // {{{
    Interval{ 0x01C5, 0x01C5 }, // LATIN CAPITAL LETTER D WITH SMALL LETTER Z WITH CARON
    Interval{ 0x01C8, 0x01C8 }, // LATIN CAPITAL LETTER L WITH SMALL LETTER J
    Interval{ 0x01CB, 0x01CB }, // LATIN CAPITAL LETTER N WITH SMALL LETTER J
    Interval{ 0x01F2, 0x01F2 }, // LATIN CAPITAL LETTER D WITH SMALL LETTER Z
    Interval{ 0x1F88, 0x1F8F }, // [8] GREEK CAPITAL LETTER ALPHA WITH PSILI AND PROSGEGRAMMENI..GREEK CAPITAL LETTER ALPHA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI
    Interval{ 0x1F98, 0x1F9F }, // [8] GREEK CAPITAL LETTER ETA WITH PSILI AND PROSGEGRAMMENI..GREEK CAPITAL LETTER ETA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI
    Interval{ 0x1FA8, 0x1FAF }, // [8] GREEK CAPITAL LETTER OMEGA WITH PSILI AND PROSGEGRAMMENI..GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI AND PROSGEGRAMMENI
    Interval{ 0x1FBC, 0x1FBC }, // GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Interval{ 0x1FCC, 0x1FCC }, // GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Interval{ 0x1FFC, 0x1FFC }, // GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
}; // }}}
auto static const Unassigned = std::array<Interval, 735>{ // {{{
    Interval{ 0x0378, 0x0379 }, // [2] <reserved-0378>..<reserved-0379>
    Interval{ 0x0380, 0x0383 }, // [4] <reserved-0380>..<reserved-0383>
    Interval{ 0x038B, 0x038B }, // <reserved-038B>
    Interval{ 0x038D, 0x038D }, // <reserved-038D>
    Interval{ 0x03A2, 0x03A2 }, // <reserved-03A2>
    Interval{ 0x0530, 0x0530 }, // <reserved-0530>
    Interval{ 0x0557, 0x0558 }, // [2] <reserved-0557>..<reserved-0558>
    Interval{ 0x058B, 0x058C }, // [2] <reserved-058B>..<reserved-058C>
    Interval{ 0x0590, 0x0590 }, // <reserved-0590>
    Interval{ 0x05C8, 0x05CF }, // [8] <reserved-05C8>..<reserved-05CF>
    Interval{ 0x05EB, 0x05EE }, // [4] <reserved-05EB>..<reserved-05EE>
    Interval{ 0x05F5, 0x05FF }, // [11] <reserved-05F5>..<reserved-05FF>
    Interval{ 0x070E, 0x070E }, // <reserved-070E>
    Interval{ 0x074B, 0x074C }, // [2] <reserved-074B>..<reserved-074C>
    Interval{ 0x07B2, 0x07BF }, // [14] <reserved-07B2>..<reserved-07BF>
    Interval{ 0x07FB, 0x07FC }, // [2] <reserved-07FB>..<reserved-07FC>
    Interval{ 0x082E, 0x082F }, // [2] <reserved-082E>..<reserved-082F>
    Interval{ 0x083F, 0x083F }, // <reserved-083F>
    Interval{ 0x085C, 0x085D }, // [2] <reserved-085C>..<reserved-085D>
    Interval{ 0x085F, 0x085F }, // <reserved-085F>
    Interval{ 0x086B, 0x086F }, // [5] <reserved-086B>..<reserved-086F>
    Interval{ 0x0892, 0x0896 }, // [5] <reserved-0892>..<reserved-0896>
    Interval{ 0x0984, 0x0984 }, // <reserved-0984>
    Interval{ 0x098D, 0x098E }, // [2] <reserved-098D>..<reserved-098E>
    Interval{ 0x0991, 0x0992 }, // [2] <reserved-0991>..<reserved-0992>
    Interval{ 0x09A9, 0x09A9 }, // <reserved-09A9>
    Interval{ 0x09B1, 0x09B1 }, // <reserved-09B1>
    Interval{ 0x09B3, 0x09B5 }, // [3] <reserved-09B3>..<reserved-09B5>
    Interval{ 0x09BA, 0x09BB }, // [2] <reserved-09BA>..<reserved-09BB>
    Interval{ 0x09C5, 0x09C6 }, // [2] <reserved-09C5>..<reserved-09C6>
    Interval{ 0x09C9, 0x09CA }, // [2] <reserved-09C9>..<reserved-09CA>
    Interval{ 0x09CF, 0x09D6 }, // [8] <reserved-09CF>..<reserved-09D6>
    Interval{ 0x09D8, 0x09DB }, // [4] <reserved-09D8>..<reserved-09DB>
    Interval{ 0x09DE, 0x09DE }, // <reserved-09DE>
    Interval{ 0x09E4, 0x09E5 }, // [2] <reserved-09E4>..<reserved-09E5>
    Interval{ 0x09FF, 0x0A00 }, // [2] <reserved-09FF>..<reserved-0A00>
    Interval{ 0x0A04, 0x0A04 }, // <reserved-0A04>
    Interval{ 0x0A0B, 0x0A0E }, // [4] <reserved-0A0B>..<reserved-0A0E>
    Interval{ 0x0A11, 0x0A12 }, // [2] <reserved-0A11>..<reserved-0A12>
    Interval{ 0x0A29, 0x0A29 }, // <reserved-0A29>
    Interval{ 0x0A31, 0x0A31 }, // <reserved-0A31>
    Interval{ 0x0A34, 0x0A34 }, // <reserved-0A34>
    Interval{ 0x0A37, 0x0A37 }, // <reserved-0A37>
    Interval{ 0x0A3A, 0x0A3B }, // [2] <reserved-0A3A>..<reserved-0A3B>
    Interval{ 0x0A3D, 0x0A3D }, // <reserved-0A3D>
    Interval{ 0x0A43, 0x0A46 }, // [4] <reserved-0A43>..<reserved-0A46>
    Interval{ 0x0A49, 0x0A4A }, // [2] <reserved-0A49>..<reserved-0A4A>
    Interval{ 0x0A4E, 0x0A50 }, // [3] <reserved-0A4E>..<reserved-0A50>
    Interval{ 0x0A52, 0x0A58 }, // [7] <reserved-0A52>..<reserved-0A58>
    Interval{ 0x0A5D, 0x0A5D }, // <reserved-0A5D>
    Interval{ 0x0A5F, 0x0A65 }, // [7] <reserved-0A5F>..<reserved-0A65>
    Interval{ 0x0A77, 0x0A80 }, // [10] <reserved-0A77>..<reserved-0A80>
    Interval{ 0x0A84, 0x0A84 }, // <reserved-0A84>
    Interval{ 0x0A8E, 0x0A8E }, // <reserved-0A8E>
    Interval{ 0x0A92, 0x0A92 }, // <reserved-0A92>
    Interval{ 0x0AA9, 0x0AA9 }, // <reserved-0AA9>
    Interval{ 0x0AB1, 0x0AB1 }, // <reserved-0AB1>
    Interval{ 0x0AB4, 0x0AB4 }, // <reserved-0AB4>
    Interval{ 0x0ABA, 0x0ABB }, // [2] <reserved-0ABA>..<reserved-0ABB>
    Interval{ 0x0AC6, 0x0AC6 }, // <reserved-0AC6>
    Interval{ 0x0ACA, 0x0ACA }, // <reserved-0ACA>
    Interval{ 0x0ACE, 0x0ACF }, // [2] <reserved-0ACE>..<reserved-0ACF>
    Interval{ 0x0AD1, 0x0ADF }, // [15] <reserved-0AD1>..<reserved-0ADF>
    Interval{ 0x0AE4, 0x0AE5 }, // [2] <reserved-0AE4>..<reserved-0AE5>
    Interval{ 0x0AF2, 0x0AF8 }, // [7] <reserved-0AF2>..<reserved-0AF8>
    Interval{ 0x0B00, 0x0B00 }, // <reserved-0B00>
    Interval{ 0x0B04, 0x0B04 }, // <reserved-0B04>
    Interval{ 0x0B0D, 0x0B0E }, // [2] <reserved-0B0D>..<reserved-0B0E>
    Interval{ 0x0B11, 0x0B12 }, // [2] <reserved-0B11>..<reserved-0B12>
    Interval{ 0x0B29, 0x0B29 }, // <reserved-0B29>
    Interval{ 0x0B31, 0x0B31 }, // <reserved-0B31>
    Interval{ 0x0B34, 0x0B34 }, // <reserved-0B34>
    Interval{ 0x0B3A, 0x0B3B }, // [2] <reserved-0B3A>..<reserved-0B3B>
    Interval{ 0x0B45, 0x0B46 }, // [2] <reserved-0B45>..<reserved-0B46>
    Interval{ 0x0B49, 0x0B4A }, // [2] <reserved-0B49>..<reserved-0B4A>
    Interval{ 0x0B4E, 0x0B54 }, // [7] <reserved-0B4E>..<reserved-0B54>
    Interval{ 0x0B58, 0x0B5B }, // [4] <reserved-0B58>..<reserved-0B5B>
    Interval{ 0x0B5E, 0x0B5E }, // <reserved-0B5E>
    Interval{ 0x0B64, 0x0B65 }, // [2] <reserved-0B64>..<reserved-0B65>
    Interval{ 0x0B78, 0x0B81 }, // [10] <reserved-0B78>..<reserved-0B81>
    Interval{ 0x0B84, 0x0B84 }, // <reserved-0B84>
    Interval{ 0x0B8B, 0x0B8D }, // [3] <reserved-0B8B>..<reserved-0B8D>
    Interval{ 0x0B91, 0x0B91 }, // <reserved-0B91>
    Interval{ 0x0B96, 0x0B98 }, // [3] <reserved-0B96>..<reserved-0B98>
    Interval{ 0x0B9B, 0x0B9B }, // <reserved-0B9B>
    Interval{ 0x0B9D, 0x0B9D }, // <reserved-0B9D>
    Interval{ 0x0BA0, 0x0BA2 }, // [3] <reserved-0BA0>..<reserved-0BA2>
    Interval{ 0x0BA5, 0x0BA7 }, // [3] <reserved-0BA5>..<reserved-0BA7>
    Interval{ 0x0BAB, 0x0BAD }, // [3] <reserved-0BAB>..<reserved-0BAD>
    Interval{ 0x0BBA, 0x0BBD }, // [4] <reserved-0BBA>..<reserved-0BBD>
    Interval{ 0x0BC3, 0x0BC5 }, // [3] <reserved-0BC3>..<reserved-0BC5>
    Interval{ 0x0BC9, 0x0BC9 }, // <reserved-0BC9>
    Interval{ 0x0BCE, 0x0BCF }, // [2] <reserved-0BCE>..<reserved-0BCF>
    Interval{ 0x0BD1, 0x0BD6 }, // [6] <reserved-0BD1>..<reserved-0BD6>
    Interval{ 0x0BD8, 0x0BE5 }, // [14] <reserved-0BD8>..<reserved-0BE5>
    Interval{ 0x0BFB, 0x0BFF }, // [5] <reserved-0BFB>..<reserved-0BFF>
    Interval{ 0x0C0D, 0x0C0D }, // <reserved-0C0D>
    Interval{ 0x0C11, 0x0C11 }, // <reserved-0C11>
    Interval{ 0x0C29, 0x0C29 }, // <reserved-0C29>
    Interval{ 0x0C3A, 0x0C3B }, // [2] <reserved-0C3A>..<reserved-0C3B>
    Interval{ 0x0C45, 0x0C45 }, // <reserved-0C45>
    Interval{ 0x0C49, 0x0C49 }, // <reserved-0C49>
    Interval{ 0x0C4E, 0x0C54 }, // [7] <reserved-0C4E>..<reserved-0C54>
    Interval{ 0x0C57, 0x0C57 }, // <reserved-0C57>
    Interval{ 0x0C5B, 0x0C5B }, // <reserved-0C5B>
    Interval{ 0x0C5E, 0x0C5F }, // [2] <reserved-0C5E>..<reserved-0C5F>
    Interval{ 0x0C64, 0x0C65 }, // [2] <reserved-0C64>..<reserved-0C65>
    Interval{ 0x0C70, 0x0C76 }, // [7] <reserved-0C70>..<reserved-0C76>
    Interval{ 0x0C8D, 0x0C8D }, // <reserved-0C8D>
    Interval{ 0x0C91, 0x0C91 }, // <reserved-0C91>
    Interval{ 0x0CA9, 0x0CA9 }, // <reserved-0CA9>
    Interval{ 0x0CB4, 0x0CB4 }, // <reserved-0CB4>
    Interval{ 0x0CBA, 0x0CBB }, // [2] <reserved-0CBA>..<reserved-0CBB>
    Interval{ 0x0CC5, 0x0CC5 }, // <reserved-0CC5>
    Interval{ 0x0CC9, 0x0CC9 }, // <reserved-0CC9>
    Interval{ 0x0CCE, 0x0CD4 }, // [7] <reserved-0CCE>..<reserved-0CD4>
    Interval{ 0x0CD7, 0x0CDB }, // [5] <reserved-0CD7>..<reserved-0CDB>
    Interval{ 0x0CDF, 0x0CDF }, // <reserved-0CDF>
    Interval{ 0x0CE4, 0x0CE5 }, // [2] <reserved-0CE4>..<reserved-0CE5>
    Interval{ 0x0CF0, 0x0CF0 }, // <reserved-0CF0>
    Interval{ 0x0CF4, 0x0CFF }, // [12] <reserved-0CF4>..<reserved-0CFF>
    Interval{ 0x0D0D, 0x0D0D }, // <reserved-0D0D>
    Interval{ 0x0D11, 0x0D11 }, // <reserved-0D11>
    Interval{ 0x0D45, 0x0D45 }, // <reserved-0D45>
    Interval{ 0x0D49, 0x0D49 }, // <reserved-0D49>
    Interval{ 0x0D50, 0x0D53 }, // [4] <reserved-0D50>..<reserved-0D53>
    Interval{ 0x0D64, 0x0D65 }, // [2] <reserved-0D64>..<reserved-0D65>
    Interval{ 0x0D80, 0x0D80 }, // <reserved-0D80>
    Interval{ 0x0D84, 0x0D84 }, // <reserved-0D84>
    Interval{ 0x0D97, 0x0D99 }, // [3] <reserved-0D97>..<reserved-0D99>
    Interval{ 0x0DB2, 0x0DB2 }, // <reserved-0DB2>
    Interval{ 0x0DBC, 0x0DBC }, // <reserved-0DBC>
    Interval{ 0x0DBE, 0x0DBF }, // [2] <reserved-0DBE>..<reserved-0DBF>
    Interval{ 0x0DC7, 0x0DC9 }, // [3] <reserved-0DC7>..<reserved-0DC9>
    Interval{ 0x0DCB, 0x0DCE }, // [4] <reserved-0DCB>..<reserved-0DCE>
    Interval{ 0x0DD5, 0x0DD5 }, // <reserved-0DD5>
    Interval{ 0x0DD7, 0x0DD7 }, // <reserved-0DD7>
    Interval{ 0x0DE0, 0x0DE5 }, // [6] <reserved-0DE0>..<reserved-0DE5>
    Interval{ 0x0DF0, 0x0DF1 }, // [2] <reserved-0DF0>..<reserved-0DF1>
    Interval{ 0x0DF5, 0x0E00 }, // [12] <reserved-0DF5>..<reserved-0E00>
    Interval{ 0x0E3B, 0x0E3E }, // [4] <reserved-0E3B>..<reserved-0E3E>
    Interval{ 0x0E5C, 0x0E80 }, // [37] <reserved-0E5C>..<reserved-0E80>
    Interval{ 0x0E83, 0x0E83 }, // <reserved-0E83>
    Interval{ 0x0E85, 0x0E85 }, // <reserved-0E85>
    Interval{ 0x0E8B, 0x0E8B }, // <reserved-0E8B>
    Interval{ 0x0EA4, 0x0EA4 }, // <reserved-0EA4>
    Interval{ 0x0EA6, 0x0EA6 }, // <reserved-0EA6>
    Interval{ 0x0EBE, 0x0EBF }, // [2] <reserved-0EBE>..<reserved-0EBF>
    Interval{ 0x0EC5, 0x0EC5 }, // <reserved-0EC5>
    Interval{ 0x0EC7, 0x0EC7 }, // <reserved-0EC7>
    Interval{ 0x0ECF, 0x0ECF }, // <reserved-0ECF>
    Interval{ 0x0EDA, 0x0EDB }, // [2] <reserved-0EDA>..<reserved-0EDB>
    Interval{ 0x0EE0, 0x0EFF }, // [32] <reserved-0EE0>..<reserved-0EFF>
    Interval{ 0x0F48, 0x0F48 }, // <reserved-0F48>
    Interval{ 0x0F6D, 0x0F70 }, // [4] <reserved-0F6D>..<reserved-0F70>
    Interval{ 0x0F98, 0x0F98 }, // <reserved-0F98>
    Interval{ 0x0FBD, 0x0FBD }, // <reserved-0FBD>
    Interval{ 0x0FCD, 0x0FCD }, // <reserved-0FCD>
    Interval{ 0x0FDB, 0x0FFF }, // [37] <reserved-0FDB>..<reserved-0FFF>
    Interval{ 0x10C6, 0x10C6 }, // <reserved-10C6>
    Interval{ 0x10C8, 0x10CC }, // [5] <reserved-10C8>..<reserved-10CC>
    Interval{ 0x10CE, 0x10CF }, // [2] <reserved-10CE>..<reserved-10CF>
    Interval{ 0x1249, 0x1249 }, // <reserved-1249>
    Interval{ 0x124E, 0x124F }, // [2] <reserved-124E>..<reserved-124F>
    Interval{ 0x1257, 0x1257 }, // <reserved-1257>
    Interval{ 0x1259, 0x1259 }, // <reserved-1259>
    Interval{ 0x125E, 0x125F }, // [2] <reserved-125E>..<reserved-125F>
    Interval{ 0x1289, 0x1289 }, // <reserved-1289>
    Interval{ 0x128E, 0x128F }, // [2] <reserved-128E>..<reserved-128F>
    Interval{ 0x12B1, 0x12B1 }, // <reserved-12B1>
    Interval{ 0x12B6, 0x12B7 }, // [2] <reserved-12B6>..<reserved-12B7>
    Interval{ 0x12BF, 0x12BF }, // <reserved-12BF>
    Interval{ 0x12C1, 0x12C1 }, // <reserved-12C1>
    Interval{ 0x12C6, 0x12C7 }, // [2] <reserved-12C6>..<reserved-12C7>
    Interval{ 0x12D7, 0x12D7 }, // <reserved-12D7>
    Interval{ 0x1311, 0x1311 }, // <reserved-1311>
    Interval{ 0x1316, 0x1317 }, // [2] <reserved-1316>..<reserved-1317>
    Interval{ 0x135B, 0x135C }, // [2] <reserved-135B>..<reserved-135C>
    Interval{ 0x137D, 0x137F }, // [3] <reserved-137D>..<reserved-137F>
    Interval{ 0x139A, 0x139F }, // [6] <reserved-139A>..<reserved-139F>
    Interval{ 0x13F6, 0x13F7 }, // [2] <reserved-13F6>..<reserved-13F7>
    Interval{ 0x13FE, 0x13FF }, // [2] <reserved-13FE>..<reserved-13FF>
    Interval{ 0x169D, 0x169F }, // [3] <reserved-169D>..<reserved-169F>
    Interval{ 0x16F9, 0x16FF }, // [7] <reserved-16F9>..<reserved-16FF>
    Interval{ 0x1716, 0x171E }, // [9] <reserved-1716>..<reserved-171E>
    Interval{ 0x1737, 0x173F }, // [9] <reserved-1737>..<reserved-173F>
    Interval{ 0x1754, 0x175F }, // [12] <reserved-1754>..<reserved-175F>
    Interval{ 0x176D, 0x176D }, // <reserved-176D>
    Interval{ 0x1771, 0x1771 }, // <reserved-1771>
    Interval{ 0x1774, 0x177F }, // [12] <reserved-1774>..<reserved-177F>
    Interval{ 0x17DE, 0x17DF }, // [2] <reserved-17DE>..<reserved-17DF>
    Interval{ 0x17EA, 0x17EF }, // [6] <reserved-17EA>..<reserved-17EF>
    Interval{ 0x17FA, 0x17FF }, // [6] <reserved-17FA>..<reserved-17FF>
    Interval{ 0x181A, 0x181F }, // [6] <reserved-181A>..<reserved-181F>
    Interval{ 0x1879, 0x187F }, // [7] <reserved-1879>..<reserved-187F>
    Interval{ 0x18AB, 0x18AF }, // [5] <reserved-18AB>..<reserved-18AF>
    Interval{ 0x18F6, 0x18FF }, // [10] <reserved-18F6>..<reserved-18FF>
    Interval{ 0x191F, 0x191F }, // <reserved-191F>
    Interval{ 0x192C, 0x192F }, // [4] <reserved-192C>..<reserved-192F>
    Interval{ 0x193C, 0x193F }, // [4] <reserved-193C>..<reserved-193F>
    Interval{ 0x1941, 0x1943 }, // [3] <reserved-1941>..<reserved-1943>
    Interval{ 0x196E, 0x196F }, // [2] <reserved-196E>..<reserved-196F>
    Interval{ 0x1975, 0x197F }, // [11] <reserved-1975>..<reserved-197F>
    Interval{ 0x19AC, 0x19AF }, // [4] <reserved-19AC>..<reserved-19AF>
    Interval{ 0x19CA, 0x19CF }, // [6] <reserved-19CA>..<reserved-19CF>
    Interval{ 0x19DB, 0x19DD }, // [3] <reserved-19DB>..<reserved-19DD>
    Interval{ 0x1A1C, 0x1A1D }, // [2] <reserved-1A1C>..<reserved-1A1D>
    Interval{ 0x1A5F, 0x1A5F }, // <reserved-1A5F>
    Interval{ 0x1A7D, 0x1A7E }, // [2] <reserved-1A7D>..<reserved-1A7E>
    Interval{ 0x1A8A, 0x1A8F }, // [6] <reserved-1A8A>..<reserved-1A8F>
    Interval{ 0x1A9A, 0x1A9F }, // [6] <reserved-1A9A>..<reserved-1A9F>
    Interval{ 0x1AAE, 0x1AAF }, // [2] <reserved-1AAE>..<reserved-1AAF>
    Interval{ 0x1ADE, 0x1ADF }, // [2] <reserved-1ADE>..<reserved-1ADF>
    Interval{ 0x1AEC, 0x1AFF }, // [20] <reserved-1AEC>..<reserved-1AFF>
    Interval{ 0x1B4D, 0x1B4D }, // <reserved-1B4D>
    Interval{ 0x1BF4, 0x1BFB }, // [8] <reserved-1BF4>..<reserved-1BFB>
    Interval{ 0x1C38, 0x1C3A }, // [3] <reserved-1C38>..<reserved-1C3A>
    Interval{ 0x1C4A, 0x1C4C }, // [3] <reserved-1C4A>..<reserved-1C4C>
    Interval{ 0x1C8B, 0x1C8F }, // [5] <reserved-1C8B>..<reserved-1C8F>
    Interval{ 0x1CBB, 0x1CBC }, // [2] <reserved-1CBB>..<reserved-1CBC>
    Interval{ 0x1CC8, 0x1CCF }, // [8] <reserved-1CC8>..<reserved-1CCF>
    Interval{ 0x1CFB, 0x1CFF }, // [5] <reserved-1CFB>..<reserved-1CFF>
    Interval{ 0x1F16, 0x1F17 }, // [2] <reserved-1F16>..<reserved-1F17>
    Interval{ 0x1F1E, 0x1F1F }, // [2] <reserved-1F1E>..<reserved-1F1F>
    Interval{ 0x1F46, 0x1F47 }, // [2] <reserved-1F46>..<reserved-1F47>
    Interval{ 0x1F4E, 0x1F4F }, // [2] <reserved-1F4E>..<reserved-1F4F>
    Interval{ 0x1F58, 0x1F58 }, // <reserved-1F58>
    Interval{ 0x1F5A, 0x1F5A }, // <reserved-1F5A>
    Interval{ 0x1F5C, 0x1F5C }, // <reserved-1F5C>
    Interval{ 0x1F5E, 0x1F5E }, // <reserved-1F5E>
    Interval{ 0x1F7E, 0x1F7F }, // [2] <reserved-1F7E>..<reserved-1F7F>
    Interval{ 0x1FB5, 0x1FB5 }, // <reserved-1FB5>
    Interval{ 0x1FC5, 0x1FC5 }, // <reserved-1FC5>
    Interval{ 0x1FD4, 0x1FD5 }, // [2] <reserved-1FD4>..<reserved-1FD5>
    Interval{ 0x1FDC, 0x1FDC }, // <reserved-1FDC>
    Interval{ 0x1FF0, 0x1FF1 }, // [2] <reserved-1FF0>..<reserved-1FF1>
    Interval{ 0x1FF5, 0x1FF5 }, // <reserved-1FF5>
    Interval{ 0x1FFF, 0x1FFF }, // <reserved-1FFF>
    Interval{ 0x2065, 0x2065 }, // <reserved-2065>
    Interval{ 0x2072, 0x2073 }, // [2] <reserved-2072>..<reserved-2073>
    Interval{ 0x208F, 0x208F }, // <reserved-208F>
    Interval{ 0x209D, 0x209F }, // [3] <reserved-209D>..<reserved-209F>
    Interval{ 0x20C2, 0x20CF }, // [14] <reserved-20C2>..<reserved-20CF>
    Interval{ 0x20F1, 0x20FF }, // [15] <reserved-20F1>..<reserved-20FF>
    Interval{ 0x218C, 0x218F }, // [4] <reserved-218C>..<reserved-218F>
    Interval{ 0x242A, 0x243F }, // [22] <reserved-242A>..<reserved-243F>
    Interval{ 0x244B, 0x245F }, // [21] <reserved-244B>..<reserved-245F>
    Interval{ 0x2B74, 0x2B75 }, // [2] <reserved-2B74>..<reserved-2B75>
    Interval{ 0x2CF4, 0x2CF8 }, // [5] <reserved-2CF4>..<reserved-2CF8>
    Interval{ 0x2D26, 0x2D26 }, // <reserved-2D26>
    Interval{ 0x2D28, 0x2D2C }, // [5] <reserved-2D28>..<reserved-2D2C>
    Interval{ 0x2D2E, 0x2D2F }, // [2] <reserved-2D2E>..<reserved-2D2F>
    Interval{ 0x2D68, 0x2D6E }, // [7] <reserved-2D68>..<reserved-2D6E>
    Interval{ 0x2D71, 0x2D7E }, // [14] <reserved-2D71>..<reserved-2D7E>
    Interval{ 0x2D97, 0x2D9F }, // [9] <reserved-2D97>..<reserved-2D9F>
    Interval{ 0x2DA7, 0x2DA7 }, // <reserved-2DA7>
    Interval{ 0x2DAF, 0x2DAF }, // <reserved-2DAF>
    Interval{ 0x2DB7, 0x2DB7 }, // <reserved-2DB7>
    Interval{ 0x2DBF, 0x2DBF }, // <reserved-2DBF>
    Interval{ 0x2DC7, 0x2DC7 }, // <reserved-2DC7>
    Interval{ 0x2DCF, 0x2DCF }, // <reserved-2DCF>
    Interval{ 0x2DD7, 0x2DD7 }, // <reserved-2DD7>
    Interval{ 0x2DDF, 0x2DDF }, // <reserved-2DDF>
    Interval{ 0x2E5E, 0x2E7F }, // [34] <reserved-2E5E>..<reserved-2E7F>
    Interval{ 0x2E9A, 0x2E9A }, // <reserved-2E9A>
    Interval{ 0x2EF4, 0x2EFF }, // [12] <reserved-2EF4>..<reserved-2EFF>
    Interval{ 0x2FD6, 0x2FEF }, // [26] <reserved-2FD6>..<reserved-2FEF>
    Interval{ 0x3040, 0x3040 }, // <reserved-3040>
    Interval{ 0x3097, 0x3098 }, // [2] <reserved-3097>..<reserved-3098>
    Interval{ 0x3100, 0x3104 }, // [5] <reserved-3100>..<reserved-3104>
    Interval{ 0x3130, 0x3130 }, // <reserved-3130>
    Interval{ 0x318F, 0x318F }, // <reserved-318F>
    Interval{ 0x31E6, 0x31EE }, // [9] <reserved-31E6>..<reserved-31EE>
    Interval{ 0x321F, 0x321F }, // <reserved-321F>
    Interval{ 0xA48D, 0xA48F }, // [3] <reserved-A48D>..<reserved-A48F>
    Interval{ 0xA4C7, 0xA4CF }, // [9] <reserved-A4C7>..<reserved-A4CF>
    Interval{ 0xA62C, 0xA63F }, // [20] <reserved-A62C>..<reserved-A63F>
    Interval{ 0xA6F8, 0xA6FF }, // [8] <reserved-A6F8>..<reserved-A6FF>
    Interval{ 0xA7DD, 0xA7F0 }, // [20] <reserved-A7DD>..<reserved-A7F0>
    Interval{ 0xA82D, 0xA82F }, // [3] <reserved-A82D>..<reserved-A82F>
    Interval{ 0xA83A, 0xA83F }, // [6] <reserved-A83A>..<reserved-A83F>
    Interval{ 0xA878, 0xA87F }, // [8] <reserved-A878>..<reserved-A87F>
    Interval{ 0xA8C6, 0xA8CD }, // [8] <reserved-A8C6>..<reserved-A8CD>
    Interval{ 0xA8DA, 0xA8DF }, // [6] <reserved-A8DA>..<reserved-A8DF>
    Interval{ 0xA954, 0xA95E }, // [11] <reserved-A954>..<reserved-A95E>
    Interval{ 0xA97D, 0xA97F }, // [3] <reserved-A97D>..<reserved-A97F>
    Interval{ 0xA9CE, 0xA9CE }, // <reserved-A9CE>
    Interval{ 0xA9DA, 0xA9DD }, // [4] <reserved-A9DA>..<reserved-A9DD>
    Interval{ 0xA9FF, 0xA9FF }, // <reserved-A9FF>
    Interval{ 0xAA37, 0xAA3F }, // [9] <reserved-AA37>..<reserved-AA3F>
    Interval{ 0xAA4E, 0xAA4F }, // [2] <reserved-AA4E>..<reserved-AA4F>
    Interval{ 0xAA5A, 0xAA5B }, // [2] <reserved-AA5A>..<reserved-AA5B>
    Interval{ 0xAAC3, 0xAADA }, // [24] <reserved-AAC3>..<reserved-AADA>
    Interval{ 0xAAF7, 0xAB00 }, // [10] <reserved-AAF7>..<reserved-AB00>
    Interval{ 0xAB07, 0xAB08 }, // [2] <reserved-AB07>..<reserved-AB08>
    Interval{ 0xAB0F, 0xAB10 }, // [2] <reserved-AB0F>..<reserved-AB10>
    Interval{ 0xAB17, 0xAB1F }, // [9] <reserved-AB17>..<reserved-AB1F>
    Interval{ 0xAB27, 0xAB27 }, // <reserved-AB27>
    Interval{ 0xAB2F, 0xAB2F }, // <reserved-AB2F>
    Interval{ 0xAB6C, 0xAB6F }, // [4] <reserved-AB6C>..<reserved-AB6F>
    Interval{ 0xABEE, 0xABEF }, // [2] <reserved-ABEE>..<reserved-ABEF>
    Interval{ 0xABFA, 0xABFF }, // [6] <reserved-ABFA>..<reserved-ABFF>
    Interval{ 0xD7A4, 0xD7AF }, // [12] <reserved-D7A4>..<reserved-D7AF>
    Interval{ 0xD7C7, 0xD7CA }, // [4] <reserved-D7C7>..<reserved-D7CA>
    Interval{ 0xD7FC, 0xD7FF }, // [4] <reserved-D7FC>..<reserved-D7FF>
    Interval{ 0xFA6E, 0xFA6F }, // [2] <reserved-FA6E>..<reserved-FA6F>
    Interval{ 0xFADA, 0xFAFF }, // [38] <reserved-FADA>..<reserved-FAFF>
    Interval{ 0xFB07, 0xFB12 }, // [12] <reserved-FB07>..<reserved-FB12>
    Interval{ 0xFB18, 0xFB1C }, // [5] <reserved-FB18>..<reserved-FB1C>
    Interval{ 0xFB37, 0xFB37 }, // <reserved-FB37>
    Interval{ 0xFB3D, 0xFB3D }, // <reserved-FB3D>
    Interval{ 0xFB3F, 0xFB3F }, // <reserved-FB3F>
    Interval{ 0xFB42, 0xFB42 }, // <reserved-FB42>
    Interval{ 0xFB45, 0xFB45 }, // <reserved-FB45>
    Interval{ 0xFDD0, 0xFDEF }, // [32] <noncharacter-FDD0>..<noncharacter-FDEF>
    Interval{ 0xFE1A, 0xFE1F }, // [6] <reserved-FE1A>..<reserved-FE1F>
    Interval{ 0xFE53, 0xFE53 }, // <reserved-FE53>
    Interval{ 0xFE67, 0xFE67 }, // <reserved-FE67>
    Interval{ 0xFE6C, 0xFE6F }, // [4] <reserved-FE6C>..<reserved-FE6F>
    Interval{ 0xFE75, 0xFE75 }, // <reserved-FE75>
    Interval{ 0xFEFD, 0xFEFE }, // [2] <reserved-FEFD>..<reserved-FEFE>
    Interval{ 0xFF00, 0xFF00 }, // <reserved-FF00>
    Interval{ 0xFFBF, 0xFFC1 }, // [3] <reserved-FFBF>..<reserved-FFC1>
    Interval{ 0xFFC8, 0xFFC9 }, // [2] <reserved-FFC8>..<reserved-FFC9>
    Interval{ 0xFFD0, 0xFFD1 }, // [2] <reserved-FFD0>..<reserved-FFD1>
    Interval{ 0xFFD8, 0xFFD9 }, // [2] <reserved-FFD8>..<reserved-FFD9>
    Interval{ 0xFFDD, 0xFFDF }, // [3] <reserved-FFDD>..<reserved-FFDF>
    Interval{ 0xFFE7, 0xFFE7 }, // <reserved-FFE7>
    Interval{ 0xFFEF, 0xFFF8 }, // [10] <reserved-FFEF>..<reserved-FFF8>
    Interval{ 0xFFFE, 0xFFFF }, // [2] <noncharacter-FFFE>..<noncharacter-FFFF>
    Interval{ 0x1000C, 0x1000C }, // <reserved-1000C>
    Interval{ 0x10027, 0x10027 }, // <reserved-10027>
    Interval{ 0x1003B, 0x1003B }, // <reserved-1003B>
    Interval{ 0x1003E, 0x1003E }, // <reserved-1003E>
    Interval{ 0x1004E, 0x1004F }, // [2] <reserved-1004E>..<reserved-1004F>
    Interval{ 0x1005E, 0x1007F }, // [34] <reserved-1005E>..<reserved-1007F>
    Interval{ 0x100FB, 0x100FF }, // [5] <reserved-100FB>..<reserved-100FF>
    Interval{ 0x10103, 0x10106 }, // [4] <reserved-10103>..<reserved-10106>
    Interval{ 0x10134, 0x10136 }, // [3] <reserved-10134>..<reserved-10136>
    Interval{ 0x1018F, 0x1018F }, // <reserved-1018F>
    Interval{ 0x1019D, 0x1019F }, // [3] <reserved-1019D>..<reserved-1019F>
    Interval{ 0x101A1, 0x101CF }, // [47] <reserved-101A1>..<reserved-101CF>
    Interval{ 0x101FE, 0x1027F }, // [130] <reserved-101FE>..<reserved-1027F>
    Interval{ 0x1029D, 0x1029F }, // [3] <reserved-1029D>..<reserved-1029F>
    Interval{ 0x102D1, 0x102DF }, // [15] <reserved-102D1>..<reserved-102DF>
    Interval{ 0x102FC, 0x102FF }, // [4] <reserved-102FC>..<reserved-102FF>
    Interval{ 0x10324, 0x1032C }, // [9] <reserved-10324>..<reserved-1032C>
    Interval{ 0x1034B, 0x1034F }, // [5] <reserved-1034B>..<reserved-1034F>
    Interval{ 0x1037B, 0x1037F }, // [5] <reserved-1037B>..<reserved-1037F>
    Interval{ 0x1039E, 0x1039E }, // <reserved-1039E>
    Interval{ 0x103C4, 0x103C7 }, // [4] <reserved-103C4>..<reserved-103C7>
    Interval{ 0x103D6, 0x103FF }, // [42] <reserved-103D6>..<reserved-103FF>
    Interval{ 0x1049E, 0x1049F }, // [2] <reserved-1049E>..<reserved-1049F>
    Interval{ 0x104AA, 0x104AF }, // [6] <reserved-104AA>..<reserved-104AF>
    Interval{ 0x104D4, 0x104D7 }, // [4] <reserved-104D4>..<reserved-104D7>
    Interval{ 0x104FC, 0x104FF }, // [4] <reserved-104FC>..<reserved-104FF>
    Interval{ 0x10528, 0x1052F }, // [8] <reserved-10528>..<reserved-1052F>
    Interval{ 0x10564, 0x1056E }, // [11] <reserved-10564>..<reserved-1056E>
    Interval{ 0x1057B, 0x1057B }, // <reserved-1057B>
    Interval{ 0x1058B, 0x1058B }, // <reserved-1058B>
    Interval{ 0x10593, 0x10593 }, // <reserved-10593>
    Interval{ 0x10596, 0x10596 }, // <reserved-10596>
    Interval{ 0x105A2, 0x105A2 }, // <reserved-105A2>
    Interval{ 0x105B2, 0x105B2 }, // <reserved-105B2>
    Interval{ 0x105BA, 0x105BA }, // <reserved-105BA>
    Interval{ 0x105BD, 0x105BF }, // [3] <reserved-105BD>..<reserved-105BF>
    Interval{ 0x105F4, 0x105FF }, // [12] <reserved-105F4>..<reserved-105FF>
    Interval{ 0x10737, 0x1073F }, // [9] <reserved-10737>..<reserved-1073F>
    Interval{ 0x10756, 0x1075F }, // [10] <reserved-10756>..<reserved-1075F>
    Interval{ 0x10768, 0x1077F }, // [24] <reserved-10768>..<reserved-1077F>
    Interval{ 0x10786, 0x10786 }, // <reserved-10786>
    Interval{ 0x107B1, 0x107B1 }, // <reserved-107B1>
    Interval{ 0x107BB, 0x107FF }, // [69] <reserved-107BB>..<reserved-107FF>
    Interval{ 0x10806, 0x10807 }, // [2] <reserved-10806>..<reserved-10807>
    Interval{ 0x10809, 0x10809 }, // <reserved-10809>
    Interval{ 0x10836, 0x10836 }, // <reserved-10836>
    Interval{ 0x10839, 0x1083B }, // [3] <reserved-10839>..<reserved-1083B>
    Interval{ 0x1083D, 0x1083E }, // [2] <reserved-1083D>..<reserved-1083E>
    Interval{ 0x10856, 0x10856 }, // <reserved-10856>
    Interval{ 0x1089F, 0x108A6 }, // [8] <reserved-1089F>..<reserved-108A6>
    Interval{ 0x108B0, 0x108DF }, // [48] <reserved-108B0>..<reserved-108DF>
    Interval{ 0x108F3, 0x108F3 }, // <reserved-108F3>
    Interval{ 0x108F6, 0x108FA }, // [5] <reserved-108F6>..<reserved-108FA>
    Interval{ 0x1091C, 0x1091E }, // [3] <reserved-1091C>..<reserved-1091E>
    Interval{ 0x1093A, 0x1093E }, // [5] <reserved-1093A>..<reserved-1093E>
    Interval{ 0x1095A, 0x1097F }, // [38] <reserved-1095A>..<reserved-1097F>
    Interval{ 0x109B8, 0x109BB }, // [4] <reserved-109B8>..<reserved-109BB>
    Interval{ 0x109D0, 0x109D1 }, // [2] <reserved-109D0>..<reserved-109D1>
    Interval{ 0x10A04, 0x10A04 }, // <reserved-10A04>
    Interval{ 0x10A07, 0x10A0B }, // [5] <reserved-10A07>..<reserved-10A0B>
    Interval{ 0x10A14, 0x10A14 }, // <reserved-10A14>
    Interval{ 0x10A18, 0x10A18 }, // <reserved-10A18>
    Interval{ 0x10A36, 0x10A37 }, // [2] <reserved-10A36>..<reserved-10A37>
    Interval{ 0x10A3B, 0x10A3E }, // [4] <reserved-10A3B>..<reserved-10A3E>
    Interval{ 0x10A49, 0x10A4F }, // [7] <reserved-10A49>..<reserved-10A4F>
    Interval{ 0x10A59, 0x10A5F }, // [7] <reserved-10A59>..<reserved-10A5F>
    Interval{ 0x10AA0, 0x10ABF }, // [32] <reserved-10AA0>..<reserved-10ABF>
    Interval{ 0x10AE7, 0x10AEA }, // [4] <reserved-10AE7>..<reserved-10AEA>
    Interval{ 0x10AF7, 0x10AFF }, // [9] <reserved-10AF7>..<reserved-10AFF>
    Interval{ 0x10B36, 0x10B38 }, // [3] <reserved-10B36>..<reserved-10B38>
    Interval{ 0x10B56, 0x10B57 }, // [2] <reserved-10B56>..<reserved-10B57>
    Interval{ 0x10B73, 0x10B77 }, // [5] <reserved-10B73>..<reserved-10B77>
    Interval{ 0x10B92, 0x10B98 }, // [7] <reserved-10B92>..<reserved-10B98>
    Interval{ 0x10B9D, 0x10BA8 }, // [12] <reserved-10B9D>..<reserved-10BA8>
    Interval{ 0x10BB0, 0x10BFF }, // [80] <reserved-10BB0>..<reserved-10BFF>
    Interval{ 0x10C49, 0x10C7F }, // [55] <reserved-10C49>..<reserved-10C7F>
    Interval{ 0x10CB3, 0x10CBF }, // [13] <reserved-10CB3>..<reserved-10CBF>
    Interval{ 0x10CF3, 0x10CF9 }, // [7] <reserved-10CF3>..<reserved-10CF9>
    Interval{ 0x10D28, 0x10D2F }, // [8] <reserved-10D28>..<reserved-10D2F>
    Interval{ 0x10D3A, 0x10D3F }, // [6] <reserved-10D3A>..<reserved-10D3F>
    Interval{ 0x10D66, 0x10D68 }, // [3] <reserved-10D66>..<reserved-10D68>
    Interval{ 0x10D86, 0x10D8D }, // [8] <reserved-10D86>..<reserved-10D8D>
    Interval{ 0x10D90, 0x10E5F }, // [208] <reserved-10D90>..<reserved-10E5F>
    Interval{ 0x10E7F, 0x10E7F }, // <reserved-10E7F>
    Interval{ 0x10EAA, 0x10EAA }, // <reserved-10EAA>
    Interval{ 0x10EAE, 0x10EAF }, // [2] <reserved-10EAE>..<reserved-10EAF>
    Interval{ 0x10EB2, 0x10EC1 }, // [16] <reserved-10EB2>..<reserved-10EC1>
    Interval{ 0x10EC8, 0x10ECF }, // [8] <reserved-10EC8>..<reserved-10ECF>
    Interval{ 0x10ED9, 0x10EF9 }, // [33] <reserved-10ED9>..<reserved-10EF9>
    Interval{ 0x10F28, 0x10F2F }, // [8] <reserved-10F28>..<reserved-10F2F>
    Interval{ 0x10F5A, 0x10F6F }, // [22] <reserved-10F5A>..<reserved-10F6F>
    Interval{ 0x10F8A, 0x10FAF }, // [38] <reserved-10F8A>..<reserved-10FAF>
    Interval{ 0x10FCC, 0x10FDF }, // [20] <reserved-10FCC>..<reserved-10FDF>
    Interval{ 0x10FF7, 0x10FFF }, // [9] <reserved-10FF7>..<reserved-10FFF>
    Interval{ 0x1104E, 0x11051 }, // [4] <reserved-1104E>..<reserved-11051>
    Interval{ 0x11076, 0x1107E }, // [9] <reserved-11076>..<reserved-1107E>
    Interval{ 0x110C3, 0x110CC }, // [10] <reserved-110C3>..<reserved-110CC>
    Interval{ 0x110CE, 0x110CF }, // [2] <reserved-110CE>..<reserved-110CF>
    Interval{ 0x110E9, 0x110EF }, // [7] <reserved-110E9>..<reserved-110EF>
    Interval{ 0x110FA, 0x110FF }, // [6] <reserved-110FA>..<reserved-110FF>
    Interval{ 0x11135, 0x11135 }, // <reserved-11135>
    Interval{ 0x11148, 0x1114F }, // [8] <reserved-11148>..<reserved-1114F>
    Interval{ 0x11177, 0x1117F }, // [9] <reserved-11177>..<reserved-1117F>
    Interval{ 0x111E0, 0x111E0 }, // <reserved-111E0>
    Interval{ 0x111F5, 0x111FF }, // [11] <reserved-111F5>..<reserved-111FF>
    Interval{ 0x11212, 0x11212 }, // <reserved-11212>
    Interval{ 0x11242, 0x1127F }, // [62] <reserved-11242>..<reserved-1127F>
    Interval{ 0x11287, 0x11287 }, // <reserved-11287>
    Interval{ 0x11289, 0x11289 }, // <reserved-11289>
    Interval{ 0x1128E, 0x1128E }, // <reserved-1128E>
    Interval{ 0x1129E, 0x1129E }, // <reserved-1129E>
    Interval{ 0x112AA, 0x112AF }, // [6] <reserved-112AA>..<reserved-112AF>
    Interval{ 0x112EB, 0x112EF }, // [5] <reserved-112EB>..<reserved-112EF>
    Interval{ 0x112FA, 0x112FF }, // [6] <reserved-112FA>..<reserved-112FF>
    Interval{ 0x11304, 0x11304 }, // <reserved-11304>
    Interval{ 0x1130D, 0x1130E }, // [2] <reserved-1130D>..<reserved-1130E>
    Interval{ 0x11311, 0x11312 }, // [2] <reserved-11311>..<reserved-11312>
    Interval{ 0x11329, 0x11329 }, // <reserved-11329>
    Interval{ 0x11331, 0x11331 }, // <reserved-11331>
    Interval{ 0x11334, 0x11334 }, // <reserved-11334>
    Interval{ 0x1133A, 0x1133A }, // <reserved-1133A>
    Interval{ 0x11345, 0x11346 }, // [2] <reserved-11345>..<reserved-11346>
    Interval{ 0x11349, 0x1134A }, // [2] <reserved-11349>..<reserved-1134A>
    Interval{ 0x1134E, 0x1134F }, // [2] <reserved-1134E>..<reserved-1134F>
    Interval{ 0x11351, 0x11356 }, // [6] <reserved-11351>..<reserved-11356>
    Interval{ 0x11358, 0x1135C }, // [5] <reserved-11358>..<reserved-1135C>
    Interval{ 0x11364, 0x11365 }, // [2] <reserved-11364>..<reserved-11365>
    Interval{ 0x1136D, 0x1136F }, // [3] <reserved-1136D>..<reserved-1136F>
    Interval{ 0x11375, 0x1137F }, // [11] <reserved-11375>..<reserved-1137F>
    Interval{ 0x1138A, 0x1138A }, // <reserved-1138A>
    Interval{ 0x1138C, 0x1138D }, // [2] <reserved-1138C>..<reserved-1138D>
    Interval{ 0x1138F, 0x1138F }, // <reserved-1138F>
    Interval{ 0x113B6, 0x113B6 }, // <reserved-113B6>
    Interval{ 0x113C1, 0x113C1 }, // <reserved-113C1>
    Interval{ 0x113C3, 0x113C4 }, // [2] <reserved-113C3>..<reserved-113C4>
    Interval{ 0x113C6, 0x113C6 }, // <reserved-113C6>
    Interval{ 0x113CB, 0x113CB }, // <reserved-113CB>
    Interval{ 0x113D6, 0x113D6 }, // <reserved-113D6>
    Interval{ 0x113D9, 0x113E0 }, // [8] <reserved-113D9>..<reserved-113E0>
    Interval{ 0x113E3, 0x113FF }, // [29] <reserved-113E3>..<reserved-113FF>
    Interval{ 0x1145C, 0x1145C }, // <reserved-1145C>
    Interval{ 0x11462, 0x1147F }, // [30] <reserved-11462>..<reserved-1147F>
    Interval{ 0x114C8, 0x114CF }, // [8] <reserved-114C8>..<reserved-114CF>
    Interval{ 0x114DA, 0x1157F }, // [166] <reserved-114DA>..<reserved-1157F>
    Interval{ 0x115B6, 0x115B7 }, // [2] <reserved-115B6>..<reserved-115B7>
    Interval{ 0x115DE, 0x115FF }, // [34] <reserved-115DE>..<reserved-115FF>
    Interval{ 0x11645, 0x1164F }, // [11] <reserved-11645>..<reserved-1164F>
    Interval{ 0x1165A, 0x1165F }, // [6] <reserved-1165A>..<reserved-1165F>
    Interval{ 0x1166D, 0x1167F }, // [19] <reserved-1166D>..<reserved-1167F>
    Interval{ 0x116BA, 0x116BF }, // [6] <reserved-116BA>..<reserved-116BF>
    Interval{ 0x116CA, 0x116CF }, // [6] <reserved-116CA>..<reserved-116CF>
    Interval{ 0x116E4, 0x116FF }, // [28] <reserved-116E4>..<reserved-116FF>
    Interval{ 0x1171B, 0x1171C }, // [2] <reserved-1171B>..<reserved-1171C>
    Interval{ 0x1172C, 0x1172F }, // [4] <reserved-1172C>..<reserved-1172F>
    Interval{ 0x11747, 0x117FF }, // [185] <reserved-11747>..<reserved-117FF>
    Interval{ 0x1183C, 0x1189F }, // [100] <reserved-1183C>..<reserved-1189F>
    Interval{ 0x118F3, 0x118FE }, // [12] <reserved-118F3>..<reserved-118FE>
    Interval{ 0x11907, 0x11908 }, // [2] <reserved-11907>..<reserved-11908>
    Interval{ 0x1190A, 0x1190B }, // [2] <reserved-1190A>..<reserved-1190B>
    Interval{ 0x11914, 0x11914 }, // <reserved-11914>
    Interval{ 0x11917, 0x11917 }, // <reserved-11917>
    Interval{ 0x11936, 0x11936 }, // <reserved-11936>
    Interval{ 0x11939, 0x1193A }, // [2] <reserved-11939>..<reserved-1193A>
    Interval{ 0x11947, 0x1194F }, // [9] <reserved-11947>..<reserved-1194F>
    Interval{ 0x1195A, 0x1199F }, // [70] <reserved-1195A>..<reserved-1199F>
    Interval{ 0x119A8, 0x119A9 }, // [2] <reserved-119A8>..<reserved-119A9>
    Interval{ 0x119D8, 0x119D9 }, // [2] <reserved-119D8>..<reserved-119D9>
    Interval{ 0x119E5, 0x119FF }, // [27] <reserved-119E5>..<reserved-119FF>
    Interval{ 0x11A48, 0x11A4F }, // [8] <reserved-11A48>..<reserved-11A4F>
    Interval{ 0x11AA3, 0x11AAF }, // [13] <reserved-11AA3>..<reserved-11AAF>
    Interval{ 0x11AF9, 0x11AFF }, // [7] <reserved-11AF9>..<reserved-11AFF>
    Interval{ 0x11B0A, 0x11B5F }, // [86] <reserved-11B0A>..<reserved-11B5F>
    Interval{ 0x11B68, 0x11BBF }, // [88] <reserved-11B68>..<reserved-11BBF>
    Interval{ 0x11BE2, 0x11BEF }, // [14] <reserved-11BE2>..<reserved-11BEF>
    Interval{ 0x11BFA, 0x11BFF }, // [6] <reserved-11BFA>..<reserved-11BFF>
    Interval{ 0x11C09, 0x11C09 }, // <reserved-11C09>
    Interval{ 0x11C37, 0x11C37 }, // <reserved-11C37>
    Interval{ 0x11C46, 0x11C4F }, // [10] <reserved-11C46>..<reserved-11C4F>
    Interval{ 0x11C6D, 0x11C6F }, // [3] <reserved-11C6D>..<reserved-11C6F>
    Interval{ 0x11C90, 0x11C91 }, // [2] <reserved-11C90>..<reserved-11C91>
    Interval{ 0x11CA8, 0x11CA8 }, // <reserved-11CA8>
    Interval{ 0x11CB7, 0x11CFF }, // [73] <reserved-11CB7>..<reserved-11CFF>
    Interval{ 0x11D07, 0x11D07 }, // <reserved-11D07>
    Interval{ 0x11D0A, 0x11D0A }, // <reserved-11D0A>
    Interval{ 0x11D37, 0x11D39 }, // [3] <reserved-11D37>..<reserved-11D39>
    Interval{ 0x11D3B, 0x11D3B }, // <reserved-11D3B>
    Interval{ 0x11D3E, 0x11D3E }, // <reserved-11D3E>
    Interval{ 0x11D48, 0x11D4F }, // [8] <reserved-11D48>..<reserved-11D4F>
    Interval{ 0x11D5A, 0x11D5F }, // [6] <reserved-11D5A>..<reserved-11D5F>
    Interval{ 0x11D66, 0x11D66 }, // <reserved-11D66>
    Interval{ 0x11D69, 0x11D69 }, // <reserved-11D69>
    Interval{ 0x11D8F, 0x11D8F }, // <reserved-11D8F>
    Interval{ 0x11D92, 0x11D92 }, // <reserved-11D92>
    Interval{ 0x11D99, 0x11D9F }, // [7] <reserved-11D99>..<reserved-11D9F>
    Interval{ 0x11DAA, 0x11DAF }, // [6] <reserved-11DAA>..<reserved-11DAF>
    Interval{ 0x11DDC, 0x11DDF }, // [4] <reserved-11DDC>..<reserved-11DDF>
    Interval{ 0x11DEA, 0x11EDF }, // [246] <reserved-11DEA>..<reserved-11EDF>
    Interval{ 0x11EF9, 0x11EFF }, // [7] <reserved-11EF9>..<reserved-11EFF>
    Interval{ 0x11F11, 0x11F11 }, // <reserved-11F11>
    Interval{ 0x11F3B, 0x11F3D }, // [3] <reserved-11F3B>..<reserved-11F3D>
    Interval{ 0x11F5B, 0x11FAF }, // [85] <reserved-11F5B>..<reserved-11FAF>
    Interval{ 0x11FB1, 0x11FBF }, // [15] <reserved-11FB1>..<reserved-11FBF>
    Interval{ 0x11FF2, 0x11FFE }, // [13] <reserved-11FF2>..<reserved-11FFE>
    Interval{ 0x1239A, 0x123FF }, // [102] <reserved-1239A>..<reserved-123FF>
    Interval{ 0x1246F, 0x1246F }, // <reserved-1246F>
    Interval{ 0x12475, 0x1247F }, // [11] <reserved-12475>..<reserved-1247F>
    Interval{ 0x12544, 0x12F8F }, // [2636] <reserved-12544>..<reserved-12F8F>
    Interval{ 0x12FF3, 0x12FFF }, // [13] <reserved-12FF3>..<reserved-12FFF>
    Interval{ 0x13456, 0x1345F }, // [10] <reserved-13456>..<reserved-1345F>
    Interval{ 0x143FB, 0x143FF }, // [5] <reserved-143FB>..<reserved-143FF>
    Interval{ 0x14647, 0x160FF }, // [6841] <reserved-14647>..<reserved-160FF>
    Interval{ 0x1613A, 0x167FF }, // [1734] <reserved-1613A>..<reserved-167FF>
    Interval{ 0x16A39, 0x16A3F }, // [7] <reserved-16A39>..<reserved-16A3F>
    Interval{ 0x16A5F, 0x16A5F }, // <reserved-16A5F>
    Interval{ 0x16A6A, 0x16A6D }, // [4] <reserved-16A6A>..<reserved-16A6D>
    Interval{ 0x16ABF, 0x16ABF }, // <reserved-16ABF>
    Interval{ 0x16ACA, 0x16ACF }, // [6] <reserved-16ACA>..<reserved-16ACF>
    Interval{ 0x16AEE, 0x16AEF }, // [2] <reserved-16AEE>..<reserved-16AEF>
    Interval{ 0x16AF6, 0x16AFF }, // [10] <reserved-16AF6>..<reserved-16AFF>
    Interval{ 0x16B46, 0x16B4F }, // [10] <reserved-16B46>..<reserved-16B4F>
    Interval{ 0x16B5A, 0x16B5A }, // <reserved-16B5A>
    Interval{ 0x16B62, 0x16B62 }, // <reserved-16B62>
    Interval{ 0x16B78, 0x16B7C }, // [5] <reserved-16B78>..<reserved-16B7C>
    Interval{ 0x16B90, 0x16D3F }, // [432] <reserved-16B90>..<reserved-16D3F>
    Interval{ 0x16D7A, 0x16E3F }, // [198] <reserved-16D7A>..<reserved-16E3F>
    Interval{ 0x16E9B, 0x16E9F }, // [5] <reserved-16E9B>..<reserved-16E9F>
    Interval{ 0x16EB9, 0x16EBA }, // [2] <reserved-16EB9>..<reserved-16EBA>
    Interval{ 0x16ED4, 0x16EFF }, // [44] <reserved-16ED4>..<reserved-16EFF>
    Interval{ 0x16F4B, 0x16F4E }, // [4] <reserved-16F4B>..<reserved-16F4E>
    Interval{ 0x16F88, 0x16F8E }, // [7] <reserved-16F88>..<reserved-16F8E>
    Interval{ 0x16FA0, 0x16FDF }, // [64] <reserved-16FA0>..<reserved-16FDF>
    Interval{ 0x16FE5, 0x16FEF }, // [11] <reserved-16FE5>..<reserved-16FEF>
    Interval{ 0x16FF7, 0x16FFF }, // [9] <reserved-16FF7>..<reserved-16FFF>
    Interval{ 0x18CD6, 0x18CFE }, // [41] <reserved-18CD6>..<reserved-18CFE>
    Interval{ 0x18D1F, 0x18D7F }, // [97] <reserved-18D1F>..<reserved-18D7F>
    Interval{ 0x18DF3, 0x1AFEF }, // [8701] <reserved-18DF3>..<reserved-1AFEF>
    Interval{ 0x1AFF4, 0x1AFF4 }, // <reserved-1AFF4>
    Interval{ 0x1AFFC, 0x1AFFC }, // <reserved-1AFFC>
    Interval{ 0x1AFFF, 0x1AFFF }, // <reserved-1AFFF>
    Interval{ 0x1B123, 0x1B131 }, // [15] <reserved-1B123>..<reserved-1B131>
    Interval{ 0x1B133, 0x1B14F }, // [29] <reserved-1B133>..<reserved-1B14F>
    Interval{ 0x1B153, 0x1B154 }, // [2] <reserved-1B153>..<reserved-1B154>
    Interval{ 0x1B156, 0x1B163 }, // [14] <reserved-1B156>..<reserved-1B163>
    Interval{ 0x1B168, 0x1B16F }, // [8] <reserved-1B168>..<reserved-1B16F>
    Interval{ 0x1B2FC, 0x1BBFF }, // [2308] <reserved-1B2FC>..<reserved-1BBFF>
    Interval{ 0x1BC6B, 0x1BC6F }, // [5] <reserved-1BC6B>..<reserved-1BC6F>
    Interval{ 0x1BC7D, 0x1BC7F }, // [3] <reserved-1BC7D>..<reserved-1BC7F>
    Interval{ 0x1BC89, 0x1BC8F }, // [7] <reserved-1BC89>..<reserved-1BC8F>
    Interval{ 0x1BC9A, 0x1BC9B }, // [2] <reserved-1BC9A>..<reserved-1BC9B>
    Interval{ 0x1BCA4, 0x1CBFF }, // [3932] <reserved-1BCA4>..<reserved-1CBFF>
    Interval{ 0x1CCFD, 0x1CCFF }, // [3] <reserved-1CCFD>..<reserved-1CCFF>
    Interval{ 0x1CEB4, 0x1CEB9 }, // [6] <reserved-1CEB4>..<reserved-1CEB9>
    Interval{ 0x1CED1, 0x1CEDF }, // [15] <reserved-1CED1>..<reserved-1CEDF>
    Interval{ 0x1CEF1, 0x1CEFF }, // [15] <reserved-1CEF1>..<reserved-1CEFF>
    Interval{ 0x1CF2E, 0x1CF2F }, // [2] <reserved-1CF2E>..<reserved-1CF2F>
    Interval{ 0x1CF47, 0x1CF4F }, // [9] <reserved-1CF47>..<reserved-1CF4F>
    Interval{ 0x1CFC4, 0x1CFFF }, // [60] <reserved-1CFC4>..<reserved-1CFFF>
    Interval{ 0x1D0F6, 0x1D0FF }, // [10] <reserved-1D0F6>..<reserved-1D0FF>
    Interval{ 0x1D127, 0x1D128 }, // [2] <reserved-1D127>..<reserved-1D128>
    Interval{ 0x1D1EB, 0x1D1FF }, // [21] <reserved-1D1EB>..<reserved-1D1FF>
    Interval{ 0x1D246, 0x1D2BF }, // [122] <reserved-1D246>..<reserved-1D2BF>
    Interval{ 0x1D2D4, 0x1D2DF }, // [12] <reserved-1D2D4>..<reserved-1D2DF>
    Interval{ 0x1D2F4, 0x1D2FF }, // [12] <reserved-1D2F4>..<reserved-1D2FF>
    Interval{ 0x1D357, 0x1D35F }, // [9] <reserved-1D357>..<reserved-1D35F>
    Interval{ 0x1D379, 0x1D3FF }, // [135] <reserved-1D379>..<reserved-1D3FF>
    Interval{ 0x1D455, 0x1D455 }, // <reserved-1D455>
    Interval{ 0x1D49D, 0x1D49D }, // <reserved-1D49D>
    Interval{ 0x1D4A0, 0x1D4A1 }, // [2] <reserved-1D4A0>..<reserved-1D4A1>
    Interval{ 0x1D4A3, 0x1D4A4 }, // [2] <reserved-1D4A3>..<reserved-1D4A4>
    Interval{ 0x1D4A7, 0x1D4A8 }, // [2] <reserved-1D4A7>..<reserved-1D4A8>
    Interval{ 0x1D4AD, 0x1D4AD }, // <reserved-1D4AD>
    Interval{ 0x1D4BA, 0x1D4BA }, // <reserved-1D4BA>
    Interval{ 0x1D4BC, 0x1D4BC }, // <reserved-1D4BC>
    Interval{ 0x1D4C4, 0x1D4C4 }, // <reserved-1D4C4>
    Interval{ 0x1D506, 0x1D506 }, // <reserved-1D506>
    Interval{ 0x1D50B, 0x1D50C }, // [2] <reserved-1D50B>..<reserved-1D50C>
    Interval{ 0x1D515, 0x1D515 }, // <reserved-1D515>
    Interval{ 0x1D51D, 0x1D51D }, // <reserved-1D51D>
    Interval{ 0x1D53A, 0x1D53A }, // <reserved-1D53A>
    Interval{ 0x1D53F, 0x1D53F }, // <reserved-1D53F>
    Interval{ 0x1D545, 0x1D545 }, // <reserved-1D545>
    Interval{ 0x1D547, 0x1D549 }, // [3] <reserved-1D547>..<reserved-1D549>
    Interval{ 0x1D551, 0x1D551 }, // <reserved-1D551>
    Interval{ 0x1D6A6, 0x1D6A7 }, // [2] <reserved-1D6A6>..<reserved-1D6A7>
    Interval{ 0x1D7CC, 0x1D7CD }, // [2] <reserved-1D7CC>..<reserved-1D7CD>
    Interval{ 0x1DA8C, 0x1DA9A }, // [15] <reserved-1DA8C>..<reserved-1DA9A>
    Interval{ 0x1DAA0, 0x1DAA0 }, // <reserved-1DAA0>
    Interval{ 0x1DAB0, 0x1DEFF }, // [1104] <reserved-1DAB0>..<reserved-1DEFF>
    Interval{ 0x1DF1F, 0x1DF24 }, // [6] <reserved-1DF1F>..<reserved-1DF24>
    Interval{ 0x1DF2B, 0x1DFFF }, // [213] <reserved-1DF2B>..<reserved-1DFFF>
    Interval{ 0x1E007, 0x1E007 }, // <reserved-1E007>
    Interval{ 0x1E019, 0x1E01A }, // [2] <reserved-1E019>..<reserved-1E01A>
    Interval{ 0x1E022, 0x1E022 }, // <reserved-1E022>
    Interval{ 0x1E025, 0x1E025 }, // <reserved-1E025>
    Interval{ 0x1E02B, 0x1E02F }, // [5] <reserved-1E02B>..<reserved-1E02F>
    Interval{ 0x1E06E, 0x1E08E }, // [33] <reserved-1E06E>..<reserved-1E08E>
    Interval{ 0x1E090, 0x1E0FF }, // [112] <reserved-1E090>..<reserved-1E0FF>
    Interval{ 0x1E12D, 0x1E12F }, // [3] <reserved-1E12D>..<reserved-1E12F>
    Interval{ 0x1E13E, 0x1E13F }, // [2] <reserved-1E13E>..<reserved-1E13F>
    Interval{ 0x1E14A, 0x1E14D }, // [4] <reserved-1E14A>..<reserved-1E14D>
    Interval{ 0x1E150, 0x1E28F }, // [320] <reserved-1E150>..<reserved-1E28F>
    Interval{ 0x1E2AF, 0x1E2BF }, // [17] <reserved-1E2AF>..<reserved-1E2BF>
    Interval{ 0x1E2FA, 0x1E2FE }, // [5] <reserved-1E2FA>..<reserved-1E2FE>
    Interval{ 0x1E300, 0x1E4CF }, // [464] <reserved-1E300>..<reserved-1E4CF>
    Interval{ 0x1E4FA, 0x1E5CF }, // [214] <reserved-1E4FA>..<reserved-1E5CF>
    Interval{ 0x1E5FB, 0x1E5FE }, // [4] <reserved-1E5FB>..<reserved-1E5FE>
    Interval{ 0x1E600, 0x1E6BF }, // [192] <reserved-1E600>..<reserved-1E6BF>
    Interval{ 0x1E6DF, 0x1E6DF }, // <reserved-1E6DF>
    Interval{ 0x1E6F6, 0x1E6FD }, // [8] <reserved-1E6F6>..<reserved-1E6FD>
    Interval{ 0x1E700, 0x1E7DF }, // [224] <reserved-1E700>..<reserved-1E7DF>
    Interval{ 0x1E7E7, 0x1E7E7 }, // <reserved-1E7E7>
    Interval{ 0x1E7EC, 0x1E7EC }, // <reserved-1E7EC>
    Interval{ 0x1E7EF, 0x1E7EF }, // <reserved-1E7EF>
    Interval{ 0x1E7FF, 0x1E7FF }, // <reserved-1E7FF>
    Interval{ 0x1E8C5, 0x1E8C6 }, // [2] <reserved-1E8C5>..<reserved-1E8C6>
    Interval{ 0x1E8D7, 0x1E8FF }, // [41] <reserved-1E8D7>..<reserved-1E8FF>
    Interval{ 0x1E94C, 0x1E94F }, // [4] <reserved-1E94C>..<reserved-1E94F>
    Interval{ 0x1E95A, 0x1E95D }, // [4] <reserved-1E95A>..<reserved-1E95D>
    Interval{ 0x1E960, 0x1EC70 }, // [785] <reserved-1E960>..<reserved-1EC70>
    Interval{ 0x1ECB5, 0x1ED00 }, // [76] <reserved-1ECB5>..<reserved-1ED00>
    Interval{ 0x1ED3E, 0x1EDFF }, // [194] <reserved-1ED3E>..<reserved-1EDFF>
    Interval{ 0x1EE04, 0x1EE04 }, // <reserved-1EE04>
    Interval{ 0x1EE20, 0x1EE20 }, // <reserved-1EE20>
    Interval{ 0x1EE23, 0x1EE23 }, // <reserved-1EE23>
    Interval{ 0x1EE25, 0x1EE26 }, // [2] <reserved-1EE25>..<reserved-1EE26>
    Interval{ 0x1EE28, 0x1EE28 }, // <reserved-1EE28>
    Interval{ 0x1EE33, 0x1EE33 }, // <reserved-1EE33>
    Interval{ 0x1EE38, 0x1EE38 }, // <reserved-1EE38>
    Interval{ 0x1EE3A, 0x1EE3A }, // <reserved-1EE3A>
    Interval{ 0x1EE3C, 0x1EE41 }, // [6] <reserved-1EE3C>..<reserved-1EE41>
    Interval{ 0x1EE43, 0x1EE46 }, // [4] <reserved-1EE43>..<reserved-1EE46>
    Interval{ 0x1EE48, 0x1EE48 }, // <reserved-1EE48>
    Interval{ 0x1EE4A, 0x1EE4A }, // <reserved-1EE4A>
    Interval{ 0x1EE4C, 0x1EE4C }, // <reserved-1EE4C>
    Interval{ 0x1EE50, 0x1EE50 }, // <reserved-1EE50>
    Interval{ 0x1EE53, 0x1EE53 }, // <reserved-1EE53>
    Interval{ 0x1EE55, 0x1EE56 }, // [2] <reserved-1EE55>..<reserved-1EE56>
    Interval{ 0x1EE58, 0x1EE58 }, // <reserved-1EE58>
    Interval{ 0x1EE5A, 0x1EE5A }, // <reserved-1EE5A>
    Interval{ 0x1EE5C, 0x1EE5C }, // <reserved-1EE5C>
    Interval{ 0x1EE5E, 0x1EE5E }, // <reserved-1EE5E>
    Interval{ 0x1EE60, 0x1EE60 }, // <reserved-1EE60>
    Interval{ 0x1EE63, 0x1EE63 }, // <reserved-1EE63>
    Interval{ 0x1EE65, 0x1EE66 }, // [2] <reserved-1EE65>..<reserved-1EE66>
    Interval{ 0x1EE6B, 0x1EE6B }, // <reserved-1EE6B>
    Interval{ 0x1EE73, 0x1EE73 }, // <reserved-1EE73>
    Interval{ 0x1EE78, 0x1EE78 }, // <reserved-1EE78>
    Interval{ 0x1EE7D, 0x1EE7D }, // <reserved-1EE7D>
    Interval{ 0x1EE7F, 0x1EE7F }, // <reserved-1EE7F>
    Interval{ 0x1EE8A, 0x1EE8A }, // <reserved-1EE8A>
    Interval{ 0x1EE9C, 0x1EEA0 }, // [5] <reserved-1EE9C>..<reserved-1EEA0>
    Interval{ 0x1EEA4, 0x1EEA4 }, // <reserved-1EEA4>
    Interval{ 0x1EEAA, 0x1EEAA }, // <reserved-1EEAA>
    Interval{ 0x1EEBC, 0x1EEEF }, // [52] <reserved-1EEBC>..<reserved-1EEEF>
    Interval{ 0x1EEF2, 0x1EFFF }, // [270] <reserved-1EEF2>..<reserved-1EFFF>
    Interval{ 0x1F02C, 0x1F02F }, // [4] <reserved-1F02C>..<reserved-1F02F>
    Interval{ 0x1F094, 0x1F09F }, // [12] <reserved-1F094>..<reserved-1F09F>
    Interval{ 0x1F0AF, 0x1F0B0 }, // [2] <reserved-1F0AF>..<reserved-1F0B0>
    Interval{ 0x1F0C0, 0x1F0C0 }, // <reserved-1F0C0>
    Interval{ 0x1F0D0, 0x1F0D0 }, // <reserved-1F0D0>
    Interval{ 0x1F0F6, 0x1F0FF }, // [10] <reserved-1F0F6>..<reserved-1F0FF>
    Interval{ 0x1F1AE, 0x1F1E5 }, // [56] <reserved-1F1AE>..<reserved-1F1E5>
    Interval{ 0x1F203, 0x1F20F }, // [13] <reserved-1F203>..<reserved-1F20F>
    Interval{ 0x1F23C, 0x1F23F }, // [4] <reserved-1F23C>..<reserved-1F23F>
    Interval{ 0x1F249, 0x1F24F }, // [7] <reserved-1F249>..<reserved-1F24F>
    Interval{ 0x1F252, 0x1F25F }, // [14] <reserved-1F252>..<reserved-1F25F>
    Interval{ 0x1F266, 0x1F2FF }, // [154] <reserved-1F266>..<reserved-1F2FF>
    Interval{ 0x1F6D9, 0x1F6DB }, // [3] <reserved-1F6D9>..<reserved-1F6DB>
    Interval{ 0x1F6ED, 0x1F6EF }, // [3] <reserved-1F6ED>..<reserved-1F6EF>
    Interval{ 0x1F6FD, 0x1F6FF }, // [3] <reserved-1F6FD>..<reserved-1F6FF>
    Interval{ 0x1F7DA, 0x1F7DF }, // [6] <reserved-1F7DA>..<reserved-1F7DF>
    Interval{ 0x1F7EC, 0x1F7EF }, // [4] <reserved-1F7EC>..<reserved-1F7EF>
    Interval{ 0x1F7F1, 0x1F7FF }, // [15] <reserved-1F7F1>..<reserved-1F7FF>
    Interval{ 0x1F80C, 0x1F80F }, // [4] <reserved-1F80C>..<reserved-1F80F>
    Interval{ 0x1F848, 0x1F84F }, // [8] <reserved-1F848>..<reserved-1F84F>
    Interval{ 0x1F85A, 0x1F85F }, // [6] <reserved-1F85A>..<reserved-1F85F>
    Interval{ 0x1F888, 0x1F88F }, // [8] <reserved-1F888>..<reserved-1F88F>
    Interval{ 0x1F8AE, 0x1F8AF }, // [2] <reserved-1F8AE>..<reserved-1F8AF>
    Interval{ 0x1F8BC, 0x1F8BF }, // [4] <reserved-1F8BC>..<reserved-1F8BF>
    Interval{ 0x1F8C2, 0x1F8CF }, // [14] <reserved-1F8C2>..<reserved-1F8CF>
    Interval{ 0x1F8D9, 0x1F8FF }, // [39] <reserved-1F8D9>..<reserved-1F8FF>
    Interval{ 0x1FA58, 0x1FA5F }, // [8] <reserved-1FA58>..<reserved-1FA5F>
    Interval{ 0x1FA6E, 0x1FA6F }, // [2] <reserved-1FA6E>..<reserved-1FA6F>
    Interval{ 0x1FA7D, 0x1FA7F }, // [3] <reserved-1FA7D>..<reserved-1FA7F>
    Interval{ 0x1FA8B, 0x1FA8D }, // [3] <reserved-1FA8B>..<reserved-1FA8D>
    Interval{ 0x1FAC7, 0x1FAC7 }, // <reserved-1FAC7>
    Interval{ 0x1FAC9, 0x1FACC }, // [4] <reserved-1FAC9>..<reserved-1FACC>
    Interval{ 0x1FADD, 0x1FADE }, // [2] <reserved-1FADD>..<reserved-1FADE>
    Interval{ 0x1FAEB, 0x1FAEE }, // [4] <reserved-1FAEB>..<reserved-1FAEE>
    Interval{ 0x1FAF9, 0x1FAFF }, // [7] <reserved-1FAF9>..<reserved-1FAFF>
    Interval{ 0x1FB93, 0x1FB93 }, // <reserved-1FB93>
    Interval{ 0x1FBFB, 0x1FFFF }, // [1029] <reserved-1FBFB>..<noncharacter-1FFFF>
    Interval{ 0x2A6E0, 0x2A6FF }, // [32] <reserved-2A6E0>..<reserved-2A6FF>
    Interval{ 0x2B81E, 0x2B81F }, // [2] <reserved-2B81E>..<reserved-2B81F>
    Interval{ 0x2CEAE, 0x2CEAF }, // [2] <reserved-2CEAE>..<reserved-2CEAF>
    Interval{ 0x2EBE1, 0x2EBEF }, // [15] <reserved-2EBE1>..<reserved-2EBEF>
    Interval{ 0x2EE5E, 0x2F7FF }, // [2466] <reserved-2EE5E>..<reserved-2F7FF>
    Interval{ 0x2FA1E, 0x2FFFF }, // [1506] <reserved-2FA1E>..<noncharacter-2FFFF>
    Interval{ 0x3134B, 0x3134F }, // [5] <reserved-3134B>..<reserved-3134F>
    Interval{ 0x3347A, 0xE0000 }, // [707463] <reserved-3347A>..<reserved-E0000>
    Interval{ 0xE0002, 0xE001F }, // [30] <reserved-E0002>..<reserved-E001F>
    Interval{ 0xE0080, 0xE00FF }, // [128] <reserved-E0080>..<reserved-E00FF>
    Interval{ 0xE01F0, 0xEFFFF }, // [65040] <reserved-E01F0>..<noncharacter-EFFFF>
    Interval{ 0xFFFFE, 0xFFFFF }, // [2] <noncharacter-FFFFE>..<noncharacter-FFFFF>
    Interval{ 0x10FFFE, 0x10FFFF }, // [2] <noncharacter-10FFFE>..<noncharacter-10FFFF>
}; // }}}
auto static const Uppercase_Letter = std::array<Interval, 655>{ // {{{
    Interval{ 0x0041, 0x005A }, // [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Interval{ 0x00C0, 0x00D6 }, // [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x00D8, 0x00DE }, // [7] LATIN CAPITAL LETTER O WITH STROKE..LATIN CAPITAL LETTER THORN
    Interval{ 0x0100, 0x0100 }, // LATIN CAPITAL LETTER A WITH MACRON
    Interval{ 0x0102, 0x0102 }, // LATIN CAPITAL LETTER A WITH BREVE
    Interval{ 0x0104, 0x0104 }, // LATIN CAPITAL LETTER A WITH OGONEK
    Interval{ 0x0106, 0x0106 }, // LATIN CAPITAL LETTER C WITH ACUTE
    Interval{ 0x0108, 0x0108 }, // LATIN CAPITAL LETTER C WITH CIRCUMFLEX
    Interval{ 0x010A, 0x010A }, // LATIN CAPITAL LETTER C WITH DOT ABOVE
    Interval{ 0x010C, 0x010C }, // LATIN CAPITAL LETTER C WITH CARON
    Interval{ 0x010E, 0x010E }, // LATIN CAPITAL LETTER D WITH CARON
    Interval{ 0x0110, 0x0110 }, // LATIN CAPITAL LETTER D WITH STROKE
    Interval{ 0x0112, 0x0112 }, // LATIN CAPITAL LETTER E WITH MACRON
    Interval{ 0x0114, 0x0114 }, // LATIN CAPITAL LETTER E WITH BREVE
    Interval{ 0x0116, 0x0116 }, // LATIN CAPITAL LETTER E WITH DOT ABOVE
    Interval{ 0x0118, 0x0118 }, // LATIN CAPITAL LETTER E WITH OGONEK
    Interval{ 0x011A, 0x011A }, // LATIN CAPITAL LETTER E WITH CARON
    Interval{ 0x011C, 0x011C }, // LATIN CAPITAL LETTER G WITH CIRCUMFLEX
    Interval{ 0x011E, 0x011E }, // LATIN CAPITAL LETTER G WITH BREVE
    Interval{ 0x0120, 0x0120 }, // LATIN CAPITAL LETTER G WITH DOT ABOVE
    Interval{ 0x0122, 0x0122 }, // LATIN CAPITAL LETTER G WITH CEDILLA
    Interval{ 0x0124, 0x0124 }, // LATIN CAPITAL LETTER H WITH CIRCUMFLEX
    Interval{ 0x0126, 0x0126 }, // LATIN CAPITAL LETTER H WITH STROKE
    Interval{ 0x0128, 0x0128 }, // LATIN CAPITAL LETTER I WITH TILDE
    Interval{ 0x012A, 0x012A }, // LATIN CAPITAL LETTER I WITH MACRON
    Interval{ 0x012C, 0x012C }, // LATIN CAPITAL LETTER I WITH BREVE
    Interval{ 0x012E, 0x012E }, // LATIN CAPITAL LETTER I WITH OGONEK
    Interval{ 0x0130, 0x0130 }, // LATIN CAPITAL LETTER I WITH DOT ABOVE
    Interval{ 0x0132, 0x0132 }, // LATIN CAPITAL LIGATURE IJ
    Interval{ 0x0134, 0x0134 }, // LATIN CAPITAL LETTER J WITH CIRCUMFLEX
    Interval{ 0x0136, 0x0136 }, // LATIN CAPITAL LETTER K WITH CEDILLA
    Interval{ 0x0139, 0x0139 }, // LATIN CAPITAL LETTER L WITH ACUTE
    Interval{ 0x013B, 0x013B }, // LATIN CAPITAL LETTER L WITH CEDILLA
    Interval{ 0x013D, 0x013D }, // LATIN CAPITAL LETTER L WITH CARON
    Interval{ 0x013F, 0x013F }, // LATIN CAPITAL LETTER L WITH MIDDLE DOT
    Interval{ 0x0141, 0x0141 }, // LATIN CAPITAL LETTER L WITH STROKE
    Interval{ 0x0143, 0x0143 }, // LATIN CAPITAL LETTER N WITH ACUTE
    Interval{ 0x0145, 0x0145 }, // LATIN CAPITAL LETTER N WITH CEDILLA
    Interval{ 0x0147, 0x0147 }, // LATIN CAPITAL LETTER N WITH CARON
    Interval{ 0x014A, 0x014A }, // LATIN CAPITAL LETTER ENG
    Interval{ 0x014C, 0x014C }, // LATIN CAPITAL LETTER O WITH MACRON
    Interval{ 0x014E, 0x014E }, // LATIN CAPITAL LETTER O WITH BREVE
    Interval{ 0x0150, 0x0150 }, // LATIN CAPITAL LETTER O WITH DOUBLE ACUTE
    Interval{ 0x0152, 0x0152 }, // LATIN CAPITAL LIGATURE OE
    Interval{ 0x0154, 0x0154 }, // LATIN CAPITAL LETTER R WITH ACUTE
    Interval{ 0x0156, 0x0156 }, // LATIN CAPITAL LETTER R WITH CEDILLA
    Interval{ 0x0158, 0x0158 }, // LATIN CAPITAL LETTER R WITH CARON
    Interval{ 0x015A, 0x015A }, // LATIN CAPITAL LETTER S WITH ACUTE
    Interval{ 0x015C, 0x015C }, // LATIN CAPITAL LETTER S WITH CIRCUMFLEX
    Interval{ 0x015E, 0x015E }, // LATIN CAPITAL LETTER S WITH CEDILLA
    Interval{ 0x0160, 0x0160 }, // LATIN CAPITAL LETTER S WITH CARON
    Interval{ 0x0162, 0x0162 }, // LATIN CAPITAL LETTER T WITH CEDILLA
    Interval{ 0x0164, 0x0164 }, // LATIN CAPITAL LETTER T WITH CARON
    Interval{ 0x0166, 0x0166 }, // LATIN CAPITAL LETTER T WITH STROKE
    Interval{ 0x0168, 0x0168 }, // LATIN CAPITAL LETTER U WITH TILDE
    Interval{ 0x016A, 0x016A }, // LATIN CAPITAL LETTER U WITH MACRON
    Interval{ 0x016C, 0x016C }, // LATIN CAPITAL LETTER U WITH BREVE
    Interval{ 0x016E, 0x016E }, // LATIN CAPITAL LETTER U WITH RING ABOVE
    Interval{ 0x0170, 0x0170 }, // LATIN CAPITAL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x0172, 0x0172 }, // LATIN CAPITAL LETTER U WITH OGONEK
    Interval{ 0x0174, 0x0174 }, // LATIN CAPITAL LETTER W WITH CIRCUMFLEX
    Interval{ 0x0176, 0x0176 }, // LATIN CAPITAL LETTER Y WITH CIRCUMFLEX
    Interval{ 0x0178, 0x0179 }, // [2] LATIN CAPITAL LETTER Y WITH DIAERESIS..LATIN CAPITAL LETTER Z WITH ACUTE
    Interval{ 0x017B, 0x017B }, // LATIN CAPITAL LETTER Z WITH DOT ABOVE
    Interval{ 0x017D, 0x017D }, // LATIN CAPITAL LETTER Z WITH CARON
    Interval{ 0x0181, 0x0182 }, // [2] LATIN CAPITAL LETTER B WITH HOOK..LATIN CAPITAL LETTER B WITH TOPBAR
    Interval{ 0x0184, 0x0184 }, // LATIN CAPITAL LETTER TONE SIX
    Interval{ 0x0186, 0x0187 }, // [2] LATIN CAPITAL LETTER OPEN O..LATIN CAPITAL LETTER C WITH HOOK
    Interval{ 0x0189, 0x018B }, // [3] LATIN CAPITAL LETTER AFRICAN D..LATIN CAPITAL LETTER D WITH TOPBAR
    Interval{ 0x018E, 0x0191 }, // [4] LATIN CAPITAL LETTER REVERSED E..LATIN CAPITAL LETTER F WITH HOOK
    Interval{ 0x0193, 0x0194 }, // [2] LATIN CAPITAL LETTER G WITH HOOK..LATIN CAPITAL LETTER GAMMA
    Interval{ 0x0196, 0x0198 }, // [3] LATIN CAPITAL LETTER IOTA..LATIN CAPITAL LETTER K WITH HOOK
    Interval{ 0x019C, 0x019D }, // [2] LATIN CAPITAL LETTER TURNED M..LATIN CAPITAL LETTER N WITH LEFT HOOK
    Interval{ 0x019F, 0x01A0 }, // [2] LATIN CAPITAL LETTER O WITH MIDDLE TILDE..LATIN CAPITAL LETTER O WITH HORN
    Interval{ 0x01A2, 0x01A2 }, // LATIN CAPITAL LETTER OI
    Interval{ 0x01A4, 0x01A4 }, // LATIN CAPITAL LETTER P WITH HOOK
    Interval{ 0x01A6, 0x01A7 }, // [2] LATIN LETTER YR..LATIN CAPITAL LETTER TONE TWO
    Interval{ 0x01A9, 0x01A9 }, // LATIN CAPITAL LETTER ESH
    Interval{ 0x01AC, 0x01AC }, // LATIN CAPITAL LETTER T WITH HOOK
    Interval{ 0x01AE, 0x01AF }, // [2] LATIN CAPITAL LETTER T WITH RETROFLEX HOOK..LATIN CAPITAL LETTER U WITH HORN
    Interval{ 0x01B1, 0x01B3 }, // [3] LATIN CAPITAL LETTER UPSILON..LATIN CAPITAL LETTER Y WITH HOOK
    Interval{ 0x01B5, 0x01B5 }, // LATIN CAPITAL LETTER Z WITH STROKE
    Interval{ 0x01B7, 0x01B8 }, // [2] LATIN CAPITAL LETTER EZH..LATIN CAPITAL LETTER EZH REVERSED
    Interval{ 0x01BC, 0x01BC }, // LATIN CAPITAL LETTER TONE FIVE
    Interval{ 0x01C4, 0x01C4 }, // LATIN CAPITAL LETTER DZ WITH CARON
    Interval{ 0x01C7, 0x01C7 }, // LATIN CAPITAL LETTER LJ
    Interval{ 0x01CA, 0x01CA }, // LATIN CAPITAL LETTER NJ
    Interval{ 0x01CD, 0x01CD }, // LATIN CAPITAL LETTER A WITH CARON
    Interval{ 0x01CF, 0x01CF }, // LATIN CAPITAL LETTER I WITH CARON
    Interval{ 0x01D1, 0x01D1 }, // LATIN CAPITAL LETTER O WITH CARON
    Interval{ 0x01D3, 0x01D3 }, // LATIN CAPITAL LETTER U WITH CARON
    Interval{ 0x01D5, 0x01D5 }, // LATIN CAPITAL LETTER U WITH DIAERESIS AND MACRON
    Interval{ 0x01D7, 0x01D7 }, // LATIN CAPITAL LETTER U WITH DIAERESIS AND ACUTE
    Interval{ 0x01D9, 0x01D9 }, // LATIN CAPITAL LETTER U WITH DIAERESIS AND CARON
    Interval{ 0x01DB, 0x01DB }, // LATIN CAPITAL LETTER U WITH DIAERESIS AND GRAVE
    Interval{ 0x01DE, 0x01DE }, // LATIN CAPITAL LETTER A WITH DIAERESIS AND MACRON
    Interval{ 0x01E0, 0x01E0 }, // LATIN CAPITAL LETTER A WITH DOT ABOVE AND MACRON
    Interval{ 0x01E2, 0x01E2 }, // LATIN CAPITAL LETTER AE WITH MACRON
    Interval{ 0x01E4, 0x01E4 }, // LATIN CAPITAL LETTER G WITH STROKE
    Interval{ 0x01E6, 0x01E6 }, // LATIN CAPITAL LETTER G WITH CARON
    Interval{ 0x01E8, 0x01E8 }, // LATIN CAPITAL LETTER K WITH CARON
    Interval{ 0x01EA, 0x01EA }, // LATIN CAPITAL LETTER O WITH OGONEK
    Interval{ 0x01EC, 0x01EC }, // LATIN CAPITAL LETTER O WITH OGONEK AND MACRON
    Interval{ 0x01EE, 0x01EE }, // LATIN CAPITAL LETTER EZH WITH CARON
    Interval{ 0x01F1, 0x01F1 }, // LATIN CAPITAL LETTER DZ
    Interval{ 0x01F4, 0x01F4 }, // LATIN CAPITAL LETTER G WITH ACUTE
    Interval{ 0x01F6, 0x01F8 }, // [3] LATIN CAPITAL LETTER HWAIR..LATIN CAPITAL LETTER N WITH GRAVE
    Interval{ 0x01FA, 0x01FA }, // LATIN CAPITAL LETTER A WITH RING ABOVE AND ACUTE
    Interval{ 0x01FC, 0x01FC }, // LATIN CAPITAL LETTER AE WITH ACUTE
    Interval{ 0x01FE, 0x01FE }, // LATIN CAPITAL LETTER O WITH STROKE AND ACUTE
    Interval{ 0x0200, 0x0200 }, // LATIN CAPITAL LETTER A WITH DOUBLE GRAVE
    Interval{ 0x0202, 0x0202 }, // LATIN CAPITAL LETTER A WITH INVERTED BREVE
    Interval{ 0x0204, 0x0204 }, // LATIN CAPITAL LETTER E WITH DOUBLE GRAVE
    Interval{ 0x0206, 0x0206 }, // LATIN CAPITAL LETTER E WITH INVERTED BREVE
    Interval{ 0x0208, 0x0208 }, // LATIN CAPITAL LETTER I WITH DOUBLE GRAVE
    Interval{ 0x020A, 0x020A }, // LATIN CAPITAL LETTER I WITH INVERTED BREVE
    Interval{ 0x020C, 0x020C }, // LATIN CAPITAL LETTER O WITH DOUBLE GRAVE
    Interval{ 0x020E, 0x020E }, // LATIN CAPITAL LETTER O WITH INVERTED BREVE
    Interval{ 0x0210, 0x0210 }, // LATIN CAPITAL LETTER R WITH DOUBLE GRAVE
    Interval{ 0x0212, 0x0212 }, // LATIN CAPITAL LETTER R WITH INVERTED BREVE
    Interval{ 0x0214, 0x0214 }, // LATIN CAPITAL LETTER U WITH DOUBLE GRAVE
    Interval{ 0x0216, 0x0216 }, // LATIN CAPITAL LETTER U WITH INVERTED BREVE
    Interval{ 0x0218, 0x0218 }, // LATIN CAPITAL LETTER S WITH COMMA BELOW
    Interval{ 0x021A, 0x021A }, // LATIN CAPITAL LETTER T WITH COMMA BELOW
    Interval{ 0x021C, 0x021C }, // LATIN CAPITAL LETTER YOGH
    Interval{ 0x021E, 0x021E }, // LATIN CAPITAL LETTER H WITH CARON
    Interval{ 0x0220, 0x0220 }, // LATIN CAPITAL LETTER N WITH LONG RIGHT LEG
    Interval{ 0x0222, 0x0222 }, // LATIN CAPITAL LETTER OU
    Interval{ 0x0224, 0x0224 }, // LATIN CAPITAL LETTER Z WITH HOOK
    Interval{ 0x0226, 0x0226 }, // LATIN CAPITAL LETTER A WITH DOT ABOVE
    Interval{ 0x0228, 0x0228 }, // LATIN CAPITAL LETTER E WITH CEDILLA
    Interval{ 0x022A, 0x022A }, // LATIN CAPITAL LETTER O WITH DIAERESIS AND MACRON
    Interval{ 0x022C, 0x022C }, // LATIN CAPITAL LETTER O WITH TILDE AND MACRON
    Interval{ 0x022E, 0x022E }, // LATIN CAPITAL LETTER O WITH DOT ABOVE
    Interval{ 0x0230, 0x0230 }, // LATIN CAPITAL LETTER O WITH DOT ABOVE AND MACRON
    Interval{ 0x0232, 0x0232 }, // LATIN CAPITAL LETTER Y WITH MACRON
    Interval{ 0x023A, 0x023B }, // [2] LATIN CAPITAL LETTER A WITH STROKE..LATIN CAPITAL LETTER C WITH STROKE
    Interval{ 0x023D, 0x023E }, // [2] LATIN CAPITAL LETTER L WITH BAR..LATIN CAPITAL LETTER T WITH DIAGONAL STROKE
    Interval{ 0x0241, 0x0241 }, // LATIN CAPITAL LETTER GLOTTAL STOP
    Interval{ 0x0243, 0x0246 }, // [4] LATIN CAPITAL LETTER B WITH STROKE..LATIN CAPITAL LETTER E WITH STROKE
    Interval{ 0x0248, 0x0248 }, // LATIN CAPITAL LETTER J WITH STROKE
    Interval{ 0x024A, 0x024A }, // LATIN CAPITAL LETTER SMALL Q WITH HOOK TAIL
    Interval{ 0x024C, 0x024C }, // LATIN CAPITAL LETTER R WITH STROKE
    Interval{ 0x024E, 0x024E }, // LATIN CAPITAL LETTER Y WITH STROKE
    Interval{ 0x0370, 0x0370 }, // GREEK CAPITAL LETTER HETA
    Interval{ 0x0372, 0x0372 }, // GREEK CAPITAL LETTER ARCHAIC SAMPI
    Interval{ 0x0376, 0x0376 }, // GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA
    Interval{ 0x037F, 0x037F }, // GREEK CAPITAL LETTER YOT
    Interval{ 0x0386, 0x0386 }, // GREEK CAPITAL LETTER ALPHA WITH TONOS
    Interval{ 0x0388, 0x038A }, // [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Interval{ 0x038C, 0x038C }, // GREEK CAPITAL LETTER OMICRON WITH TONOS
    Interval{ 0x038E, 0x038F }, // [2] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER OMEGA WITH TONOS
    Interval{ 0x0391, 0x03A1 }, // [17] GREEK CAPITAL LETTER ALPHA..GREEK CAPITAL LETTER RHO
    Interval{ 0x03A3, 0x03AB }, // [9] GREEK CAPITAL LETTER SIGMA..GREEK CAPITAL LETTER UPSILON WITH DIALYTIKA
    Interval{ 0x03CF, 0x03CF }, // GREEK CAPITAL KAI SYMBOL
    Interval{ 0x03D2, 0x03D4 }, // [3] GREEK UPSILON WITH HOOK SYMBOL..GREEK UPSILON WITH DIAERESIS AND HOOK SYMBOL
    Interval{ 0x03D8, 0x03D8 }, // GREEK LETTER ARCHAIC KOPPA
    Interval{ 0x03DA, 0x03DA }, // GREEK LETTER STIGMA
    Interval{ 0x03DC, 0x03DC }, // GREEK LETTER DIGAMMA
    Interval{ 0x03DE, 0x03DE }, // GREEK LETTER KOPPA
    Interval{ 0x03E0, 0x03E0 }, // GREEK LETTER SAMPI
    Interval{ 0x03E2, 0x03E2 }, // COPTIC CAPITAL LETTER SHEI
    Interval{ 0x03E4, 0x03E4 }, // COPTIC CAPITAL LETTER FEI
    Interval{ 0x03E6, 0x03E6 }, // COPTIC CAPITAL LETTER KHEI
    Interval{ 0x03E8, 0x03E8 }, // COPTIC CAPITAL LETTER HORI
    Interval{ 0x03EA, 0x03EA }, // COPTIC CAPITAL LETTER GANGIA
    Interval{ 0x03EC, 0x03EC }, // COPTIC CAPITAL LETTER SHIMA
    Interval{ 0x03EE, 0x03EE }, // COPTIC CAPITAL LETTER DEI
    Interval{ 0x03F4, 0x03F4 }, // GREEK CAPITAL THETA SYMBOL
    Interval{ 0x03F7, 0x03F7 }, // GREEK CAPITAL LETTER SHO
    Interval{ 0x03F9, 0x03FA }, // [2] GREEK CAPITAL LUNATE SIGMA SYMBOL..GREEK CAPITAL LETTER SAN
    Interval{ 0x03FD, 0x042F }, // [51] GREEK CAPITAL REVERSED LUNATE SIGMA SYMBOL..CYRILLIC CAPITAL LETTER YA
    Interval{ 0x0460, 0x0460 }, // CYRILLIC CAPITAL LETTER OMEGA
    Interval{ 0x0462, 0x0462 }, // CYRILLIC CAPITAL LETTER YAT
    Interval{ 0x0464, 0x0464 }, // CYRILLIC CAPITAL LETTER IOTIFIED E
    Interval{ 0x0466, 0x0466 }, // CYRILLIC CAPITAL LETTER LITTLE YUS
    Interval{ 0x0468, 0x0468 }, // CYRILLIC CAPITAL LETTER IOTIFIED LITTLE YUS
    Interval{ 0x046A, 0x046A }, // CYRILLIC CAPITAL LETTER BIG YUS
    Interval{ 0x046C, 0x046C }, // CYRILLIC CAPITAL LETTER IOTIFIED BIG YUS
    Interval{ 0x046E, 0x046E }, // CYRILLIC CAPITAL LETTER KSI
    Interval{ 0x0470, 0x0470 }, // CYRILLIC CAPITAL LETTER PSI
    Interval{ 0x0472, 0x0472 }, // CYRILLIC CAPITAL LETTER FITA
    Interval{ 0x0474, 0x0474 }, // CYRILLIC CAPITAL LETTER IZHITSA
    Interval{ 0x0476, 0x0476 }, // CYRILLIC CAPITAL LETTER IZHITSA WITH DOUBLE GRAVE ACCENT
    Interval{ 0x0478, 0x0478 }, // CYRILLIC CAPITAL LETTER UK
    Interval{ 0x047A, 0x047A }, // CYRILLIC CAPITAL LETTER ROUND OMEGA
    Interval{ 0x047C, 0x047C }, // CYRILLIC CAPITAL LETTER OMEGA WITH TITLO
    Interval{ 0x047E, 0x047E }, // CYRILLIC CAPITAL LETTER OT
    Interval{ 0x0480, 0x0480 }, // CYRILLIC CAPITAL LETTER KOPPA
    Interval{ 0x048A, 0x048A }, // CYRILLIC CAPITAL LETTER SHORT I WITH TAIL
    Interval{ 0x048C, 0x048C }, // CYRILLIC CAPITAL LETTER SEMISOFT SIGN
    Interval{ 0x048E, 0x048E }, // CYRILLIC CAPITAL LETTER ER WITH TICK
    Interval{ 0x0490, 0x0490 }, // CYRILLIC CAPITAL LETTER GHE WITH UPTURN
    Interval{ 0x0492, 0x0492 }, // CYRILLIC CAPITAL LETTER GHE WITH STROKE
    Interval{ 0x0494, 0x0494 }, // CYRILLIC CAPITAL LETTER GHE WITH MIDDLE HOOK
    Interval{ 0x0496, 0x0496 }, // CYRILLIC CAPITAL LETTER ZHE WITH DESCENDER
    Interval{ 0x0498, 0x0498 }, // CYRILLIC CAPITAL LETTER ZE WITH DESCENDER
    Interval{ 0x049A, 0x049A }, // CYRILLIC CAPITAL LETTER KA WITH DESCENDER
    Interval{ 0x049C, 0x049C }, // CYRILLIC CAPITAL LETTER KA WITH VERTICAL STROKE
    Interval{ 0x049E, 0x049E }, // CYRILLIC CAPITAL LETTER KA WITH STROKE
    Interval{ 0x04A0, 0x04A0 }, // CYRILLIC CAPITAL LETTER BASHKIR KA
    Interval{ 0x04A2, 0x04A2 }, // CYRILLIC CAPITAL LETTER EN WITH DESCENDER
    Interval{ 0x04A4, 0x04A4 }, // CYRILLIC CAPITAL LIGATURE EN GHE
    Interval{ 0x04A6, 0x04A6 }, // CYRILLIC CAPITAL LETTER PE WITH MIDDLE HOOK
    Interval{ 0x04A8, 0x04A8 }, // CYRILLIC CAPITAL LETTER ABKHASIAN HA
    Interval{ 0x04AA, 0x04AA }, // CYRILLIC CAPITAL LETTER ES WITH DESCENDER
    Interval{ 0x04AC, 0x04AC }, // CYRILLIC CAPITAL LETTER TE WITH DESCENDER
    Interval{ 0x04AE, 0x04AE }, // CYRILLIC CAPITAL LETTER STRAIGHT U
    Interval{ 0x04B0, 0x04B0 }, // CYRILLIC CAPITAL LETTER STRAIGHT U WITH STROKE
    Interval{ 0x04B2, 0x04B2 }, // CYRILLIC CAPITAL LETTER HA WITH DESCENDER
    Interval{ 0x04B4, 0x04B4 }, // CYRILLIC CAPITAL LIGATURE TE TSE
    Interval{ 0x04B6, 0x04B6 }, // CYRILLIC CAPITAL LETTER CHE WITH DESCENDER
    Interval{ 0x04B8, 0x04B8 }, // CYRILLIC CAPITAL LETTER CHE WITH VERTICAL STROKE
    Interval{ 0x04BA, 0x04BA }, // CYRILLIC CAPITAL LETTER SHHA
    Interval{ 0x04BC, 0x04BC }, // CYRILLIC CAPITAL LETTER ABKHASIAN CHE
    Interval{ 0x04BE, 0x04BE }, // CYRILLIC CAPITAL LETTER ABKHASIAN CHE WITH DESCENDER
    Interval{ 0x04C0, 0x04C1 }, // [2] CYRILLIC LETTER PALOCHKA..CYRILLIC CAPITAL LETTER ZHE WITH BREVE
    Interval{ 0x04C3, 0x04C3 }, // CYRILLIC CAPITAL LETTER KA WITH HOOK
    Interval{ 0x04C5, 0x04C5 }, // CYRILLIC CAPITAL LETTER EL WITH TAIL
    Interval{ 0x04C7, 0x04C7 }, // CYRILLIC CAPITAL LETTER EN WITH HOOK
    Interval{ 0x04C9, 0x04C9 }, // CYRILLIC CAPITAL LETTER EN WITH TAIL
    Interval{ 0x04CB, 0x04CB }, // CYRILLIC CAPITAL LETTER KHAKASSIAN CHE
    Interval{ 0x04CD, 0x04CD }, // CYRILLIC CAPITAL LETTER EM WITH TAIL
    Interval{ 0x04D0, 0x04D0 }, // CYRILLIC CAPITAL LETTER A WITH BREVE
    Interval{ 0x04D2, 0x04D2 }, // CYRILLIC CAPITAL LETTER A WITH DIAERESIS
    Interval{ 0x04D4, 0x04D4 }, // CYRILLIC CAPITAL LIGATURE A IE
    Interval{ 0x04D6, 0x04D6 }, // CYRILLIC CAPITAL LETTER IE WITH BREVE
    Interval{ 0x04D8, 0x04D8 }, // CYRILLIC CAPITAL LETTER SCHWA
    Interval{ 0x04DA, 0x04DA }, // CYRILLIC CAPITAL LETTER SCHWA WITH DIAERESIS
    Interval{ 0x04DC, 0x04DC }, // CYRILLIC CAPITAL LETTER ZHE WITH DIAERESIS
    Interval{ 0x04DE, 0x04DE }, // CYRILLIC CAPITAL LETTER ZE WITH DIAERESIS
    Interval{ 0x04E0, 0x04E0 }, // CYRILLIC CAPITAL LETTER ABKHASIAN DZE
    Interval{ 0x04E2, 0x04E2 }, // CYRILLIC CAPITAL LETTER I WITH MACRON
    Interval{ 0x04E4, 0x04E4 }, // CYRILLIC CAPITAL LETTER I WITH DIAERESIS
    Interval{ 0x04E6, 0x04E6 }, // CYRILLIC CAPITAL LETTER O WITH DIAERESIS
    Interval{ 0x04E8, 0x04E8 }, // CYRILLIC CAPITAL LETTER BARRED O
    Interval{ 0x04EA, 0x04EA }, // CYRILLIC CAPITAL LETTER BARRED O WITH DIAERESIS
    Interval{ 0x04EC, 0x04EC }, // CYRILLIC CAPITAL LETTER E WITH DIAERESIS
    Interval{ 0x04EE, 0x04EE }, // CYRILLIC CAPITAL LETTER U WITH MACRON
    Interval{ 0x04F0, 0x04F0 }, // CYRILLIC CAPITAL LETTER U WITH DIAERESIS
    Interval{ 0x04F2, 0x04F2 }, // CYRILLIC CAPITAL LETTER U WITH DOUBLE ACUTE
    Interval{ 0x04F4, 0x04F4 }, // CYRILLIC CAPITAL LETTER CHE WITH DIAERESIS
    Interval{ 0x04F6, 0x04F6 }, // CYRILLIC CAPITAL LETTER GHE WITH DESCENDER
    Interval{ 0x04F8, 0x04F8 }, // CYRILLIC CAPITAL LETTER YERU WITH DIAERESIS
    Interval{ 0x04FA, 0x04FA }, // CYRILLIC CAPITAL LETTER GHE WITH STROKE AND HOOK
    Interval{ 0x04FC, 0x04FC }, // CYRILLIC CAPITAL LETTER HA WITH HOOK
    Interval{ 0x04FE, 0x04FE }, // CYRILLIC CAPITAL LETTER HA WITH STROKE
    Interval{ 0x0500, 0x0500 }, // CYRILLIC CAPITAL LETTER KOMI DE
    Interval{ 0x0502, 0x0502 }, // CYRILLIC CAPITAL LETTER KOMI DJE
    Interval{ 0x0504, 0x0504 }, // CYRILLIC CAPITAL LETTER KOMI ZJE
    Interval{ 0x0506, 0x0506 }, // CYRILLIC CAPITAL LETTER KOMI DZJE
    Interval{ 0x0508, 0x0508 }, // CYRILLIC CAPITAL LETTER KOMI LJE
    Interval{ 0x050A, 0x050A }, // CYRILLIC CAPITAL LETTER KOMI NJE
    Interval{ 0x050C, 0x050C }, // CYRILLIC CAPITAL LETTER KOMI SJE
    Interval{ 0x050E, 0x050E }, // CYRILLIC CAPITAL LETTER KOMI TJE
    Interval{ 0x0510, 0x0510 }, // CYRILLIC CAPITAL LETTER REVERSED ZE
    Interval{ 0x0512, 0x0512 }, // CYRILLIC CAPITAL LETTER EL WITH HOOK
    Interval{ 0x0514, 0x0514 }, // CYRILLIC CAPITAL LETTER LHA
    Interval{ 0x0516, 0x0516 }, // CYRILLIC CAPITAL LETTER RHA
    Interval{ 0x0518, 0x0518 }, // CYRILLIC CAPITAL LETTER YAE
    Interval{ 0x051A, 0x051A }, // CYRILLIC CAPITAL LETTER QA
    Interval{ 0x051C, 0x051C }, // CYRILLIC CAPITAL LETTER WE
    Interval{ 0x051E, 0x051E }, // CYRILLIC CAPITAL LETTER ALEUT KA
    Interval{ 0x0520, 0x0520 }, // CYRILLIC CAPITAL LETTER EL WITH MIDDLE HOOK
    Interval{ 0x0522, 0x0522 }, // CYRILLIC CAPITAL LETTER EN WITH MIDDLE HOOK
    Interval{ 0x0524, 0x0524 }, // CYRILLIC CAPITAL LETTER PE WITH DESCENDER
    Interval{ 0x0526, 0x0526 }, // CYRILLIC CAPITAL LETTER SHHA WITH DESCENDER
    Interval{ 0x0528, 0x0528 }, // CYRILLIC CAPITAL LETTER EN WITH LEFT HOOK
    Interval{ 0x052A, 0x052A }, // CYRILLIC CAPITAL LETTER DZZHE
    Interval{ 0x052C, 0x052C }, // CYRILLIC CAPITAL LETTER DCHE
    Interval{ 0x052E, 0x052E }, // CYRILLIC CAPITAL LETTER EL WITH DESCENDER
    Interval{ 0x0531, 0x0556 }, // [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Interval{ 0x10A0, 0x10C5 }, // [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Interval{ 0x10C7, 0x10C7 }, // GEORGIAN CAPITAL LETTER YN
    Interval{ 0x10CD, 0x10CD }, // GEORGIAN CAPITAL LETTER AEN
    Interval{ 0x13A0, 0x13F5 }, // [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Interval{ 0x1C89, 0x1C89 }, // CYRILLIC CAPITAL LETTER TJE
    Interval{ 0x1C90, 0x1CBA }, // [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Interval{ 0x1CBD, 0x1CBF }, // [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Interval{ 0x1E00, 0x1E00 }, // LATIN CAPITAL LETTER A WITH RING BELOW
    Interval{ 0x1E02, 0x1E02 }, // LATIN CAPITAL LETTER B WITH DOT ABOVE
    Interval{ 0x1E04, 0x1E04 }, // LATIN CAPITAL LETTER B WITH DOT BELOW
    Interval{ 0x1E06, 0x1E06 }, // LATIN CAPITAL LETTER B WITH LINE BELOW
    Interval{ 0x1E08, 0x1E08 }, // LATIN CAPITAL LETTER C WITH CEDILLA AND ACUTE
    Interval{ 0x1E0A, 0x1E0A }, // LATIN CAPITAL LETTER D WITH DOT ABOVE
    Interval{ 0x1E0C, 0x1E0C }, // LATIN CAPITAL LETTER D WITH DOT BELOW
    Interval{ 0x1E0E, 0x1E0E }, // LATIN CAPITAL LETTER D WITH LINE BELOW
    Interval{ 0x1E10, 0x1E10 }, // LATIN CAPITAL LETTER D WITH CEDILLA
    Interval{ 0x1E12, 0x1E12 }, // LATIN CAPITAL LETTER D WITH CIRCUMFLEX BELOW
    Interval{ 0x1E14, 0x1E14 }, // LATIN CAPITAL LETTER E WITH MACRON AND GRAVE
    Interval{ 0x1E16, 0x1E16 }, // LATIN CAPITAL LETTER E WITH MACRON AND ACUTE
    Interval{ 0x1E18, 0x1E18 }, // LATIN CAPITAL LETTER E WITH CIRCUMFLEX BELOW
    Interval{ 0x1E1A, 0x1E1A }, // LATIN CAPITAL LETTER E WITH TILDE BELOW
    Interval{ 0x1E1C, 0x1E1C }, // LATIN CAPITAL LETTER E WITH CEDILLA AND BREVE
    Interval{ 0x1E1E, 0x1E1E }, // LATIN CAPITAL LETTER F WITH DOT ABOVE
    Interval{ 0x1E20, 0x1E20 }, // LATIN CAPITAL LETTER G WITH MACRON
    Interval{ 0x1E22, 0x1E22 }, // LATIN CAPITAL LETTER H WITH DOT ABOVE
    Interval{ 0x1E24, 0x1E24 }, // LATIN CAPITAL LETTER H WITH DOT BELOW
    Interval{ 0x1E26, 0x1E26 }, // LATIN CAPITAL LETTER H WITH DIAERESIS
    Interval{ 0x1E28, 0x1E28 }, // LATIN CAPITAL LETTER H WITH CEDILLA
    Interval{ 0x1E2A, 0x1E2A }, // LATIN CAPITAL LETTER H WITH BREVE BELOW
    Interval{ 0x1E2C, 0x1E2C }, // LATIN CAPITAL LETTER I WITH TILDE BELOW
    Interval{ 0x1E2E, 0x1E2E }, // LATIN CAPITAL LETTER I WITH DIAERESIS AND ACUTE
    Interval{ 0x1E30, 0x1E30 }, // LATIN CAPITAL LETTER K WITH ACUTE
    Interval{ 0x1E32, 0x1E32 }, // LATIN CAPITAL LETTER K WITH DOT BELOW
    Interval{ 0x1E34, 0x1E34 }, // LATIN CAPITAL LETTER K WITH LINE BELOW
    Interval{ 0x1E36, 0x1E36 }, // LATIN CAPITAL LETTER L WITH DOT BELOW
    Interval{ 0x1E38, 0x1E38 }, // LATIN CAPITAL LETTER L WITH DOT BELOW AND MACRON
    Interval{ 0x1E3A, 0x1E3A }, // LATIN CAPITAL LETTER L WITH LINE BELOW
    Interval{ 0x1E3C, 0x1E3C }, // LATIN CAPITAL LETTER L WITH CIRCUMFLEX BELOW
    Interval{ 0x1E3E, 0x1E3E }, // LATIN CAPITAL LETTER M WITH ACUTE
    Interval{ 0x1E40, 0x1E40 }, // LATIN CAPITAL LETTER M WITH DOT ABOVE
    Interval{ 0x1E42, 0x1E42 }, // LATIN CAPITAL LETTER M WITH DOT BELOW
    Interval{ 0x1E44, 0x1E44 }, // LATIN CAPITAL LETTER N WITH DOT ABOVE
    Interval{ 0x1E46, 0x1E46 }, // LATIN CAPITAL LETTER N WITH DOT BELOW
    Interval{ 0x1E48, 0x1E48 }, // LATIN CAPITAL LETTER N WITH LINE BELOW
    Interval{ 0x1E4A, 0x1E4A }, // LATIN CAPITAL LETTER N WITH CIRCUMFLEX BELOW
    Interval{ 0x1E4C, 0x1E4C }, // LATIN CAPITAL LETTER O WITH TILDE AND ACUTE
    Interval{ 0x1E4E, 0x1E4E }, // LATIN CAPITAL LETTER O WITH TILDE AND DIAERESIS
    Interval{ 0x1E50, 0x1E50 }, // LATIN CAPITAL LETTER O WITH MACRON AND GRAVE
    Interval{ 0x1E52, 0x1E52 }, // LATIN CAPITAL LETTER O WITH MACRON AND ACUTE
    Interval{ 0x1E54, 0x1E54 }, // LATIN CAPITAL LETTER P WITH ACUTE
    Interval{ 0x1E56, 0x1E56 }, // LATIN CAPITAL LETTER P WITH DOT ABOVE
    Interval{ 0x1E58, 0x1E58 }, // LATIN CAPITAL LETTER R WITH DOT ABOVE
    Interval{ 0x1E5A, 0x1E5A }, // LATIN CAPITAL LETTER R WITH DOT BELOW
    Interval{ 0x1E5C, 0x1E5C }, // LATIN CAPITAL LETTER R WITH DOT BELOW AND MACRON
    Interval{ 0x1E5E, 0x1E5E }, // LATIN CAPITAL LETTER R WITH LINE BELOW
    Interval{ 0x1E60, 0x1E60 }, // LATIN CAPITAL LETTER S WITH DOT ABOVE
    Interval{ 0x1E62, 0x1E62 }, // LATIN CAPITAL LETTER S WITH DOT BELOW
    Interval{ 0x1E64, 0x1E64 }, // LATIN CAPITAL LETTER S WITH ACUTE AND DOT ABOVE
    Interval{ 0x1E66, 0x1E66 }, // LATIN CAPITAL LETTER S WITH CARON AND DOT ABOVE
    Interval{ 0x1E68, 0x1E68 }, // LATIN CAPITAL LETTER S WITH DOT BELOW AND DOT ABOVE
    Interval{ 0x1E6A, 0x1E6A }, // LATIN CAPITAL LETTER T WITH DOT ABOVE
    Interval{ 0x1E6C, 0x1E6C }, // LATIN CAPITAL LETTER T WITH DOT BELOW
    Interval{ 0x1E6E, 0x1E6E }, // LATIN CAPITAL LETTER T WITH LINE BELOW
    Interval{ 0x1E70, 0x1E70 }, // LATIN CAPITAL LETTER T WITH CIRCUMFLEX BELOW
    Interval{ 0x1E72, 0x1E72 }, // LATIN CAPITAL LETTER U WITH DIAERESIS BELOW
    Interval{ 0x1E74, 0x1E74 }, // LATIN CAPITAL LETTER U WITH TILDE BELOW
    Interval{ 0x1E76, 0x1E76 }, // LATIN CAPITAL LETTER U WITH CIRCUMFLEX BELOW
    Interval{ 0x1E78, 0x1E78 }, // LATIN CAPITAL LETTER U WITH TILDE AND ACUTE
    Interval{ 0x1E7A, 0x1E7A }, // LATIN CAPITAL LETTER U WITH MACRON AND DIAERESIS
    Interval{ 0x1E7C, 0x1E7C }, // LATIN CAPITAL LETTER V WITH TILDE
    Interval{ 0x1E7E, 0x1E7E }, // LATIN CAPITAL LETTER V WITH DOT BELOW
    Interval{ 0x1E80, 0x1E80 }, // LATIN CAPITAL LETTER W WITH GRAVE
    Interval{ 0x1E82, 0x1E82 }, // LATIN CAPITAL LETTER W WITH ACUTE
    Interval{ 0x1E84, 0x1E84 }, // LATIN CAPITAL LETTER W WITH DIAERESIS
    Interval{ 0x1E86, 0x1E86 }, // LATIN CAPITAL LETTER W WITH DOT ABOVE
    Interval{ 0x1E88, 0x1E88 }, // LATIN CAPITAL LETTER W WITH DOT BELOW
    Interval{ 0x1E8A, 0x1E8A }, // LATIN CAPITAL LETTER X WITH DOT ABOVE
    Interval{ 0x1E8C, 0x1E8C }, // LATIN CAPITAL LETTER X WITH DIAERESIS
    Interval{ 0x1E8E, 0x1E8E }, // LATIN CAPITAL LETTER Y WITH DOT ABOVE
    Interval{ 0x1E90, 0x1E90 }, // LATIN CAPITAL LETTER Z WITH CIRCUMFLEX
    Interval{ 0x1E92, 0x1E92 }, // LATIN CAPITAL LETTER Z WITH DOT BELOW
    Interval{ 0x1E94, 0x1E94 }, // LATIN CAPITAL LETTER Z WITH LINE BELOW
    Interval{ 0x1E9E, 0x1E9E }, // LATIN CAPITAL LETTER SHARP S
    Interval{ 0x1EA0, 0x1EA0 }, // LATIN CAPITAL LETTER A WITH DOT BELOW
    Interval{ 0x1EA2, 0x1EA2 }, // LATIN CAPITAL LETTER A WITH HOOK ABOVE
    Interval{ 0x1EA4, 0x1EA4 }, // LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EA6, 0x1EA6 }, // LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EA8, 0x1EA8 }, // LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EAA, 0x1EAA }, // LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EAC, 0x1EAC }, // LATIN CAPITAL LETTER A WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EAE, 0x1EAE }, // LATIN CAPITAL LETTER A WITH BREVE AND ACUTE
    Interval{ 0x1EB0, 0x1EB0 }, // LATIN CAPITAL LETTER A WITH BREVE AND GRAVE
    Interval{ 0x1EB2, 0x1EB2 }, // LATIN CAPITAL LETTER A WITH BREVE AND HOOK ABOVE
    Interval{ 0x1EB4, 0x1EB4 }, // LATIN CAPITAL LETTER A WITH BREVE AND TILDE
    Interval{ 0x1EB6, 0x1EB6 }, // LATIN CAPITAL LETTER A WITH BREVE AND DOT BELOW
    Interval{ 0x1EB8, 0x1EB8 }, // LATIN CAPITAL LETTER E WITH DOT BELOW
    Interval{ 0x1EBA, 0x1EBA }, // LATIN CAPITAL LETTER E WITH HOOK ABOVE
    Interval{ 0x1EBC, 0x1EBC }, // LATIN CAPITAL LETTER E WITH TILDE
    Interval{ 0x1EBE, 0x1EBE }, // LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1EC0, 0x1EC0 }, // LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1EC2, 0x1EC2 }, // LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1EC4, 0x1EC4 }, // LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1EC6, 0x1EC6 }, // LATIN CAPITAL LETTER E WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EC8, 0x1EC8 }, // LATIN CAPITAL LETTER I WITH HOOK ABOVE
    Interval{ 0x1ECA, 0x1ECA }, // LATIN CAPITAL LETTER I WITH DOT BELOW
    Interval{ 0x1ECC, 0x1ECC }, // LATIN CAPITAL LETTER O WITH DOT BELOW
    Interval{ 0x1ECE, 0x1ECE }, // LATIN CAPITAL LETTER O WITH HOOK ABOVE
    Interval{ 0x1ED0, 0x1ED0 }, // LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND ACUTE
    Interval{ 0x1ED2, 0x1ED2 }, // LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND GRAVE
    Interval{ 0x1ED4, 0x1ED4 }, // LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND HOOK ABOVE
    Interval{ 0x1ED6, 0x1ED6 }, // LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND TILDE
    Interval{ 0x1ED8, 0x1ED8 }, // LATIN CAPITAL LETTER O WITH CIRCUMFLEX AND DOT BELOW
    Interval{ 0x1EDA, 0x1EDA }, // LATIN CAPITAL LETTER O WITH HORN AND ACUTE
    Interval{ 0x1EDC, 0x1EDC }, // LATIN CAPITAL LETTER O WITH HORN AND GRAVE
    Interval{ 0x1EDE, 0x1EDE }, // LATIN CAPITAL LETTER O WITH HORN AND HOOK ABOVE
    Interval{ 0x1EE0, 0x1EE0 }, // LATIN CAPITAL LETTER O WITH HORN AND TILDE
    Interval{ 0x1EE2, 0x1EE2 }, // LATIN CAPITAL LETTER O WITH HORN AND DOT BELOW
    Interval{ 0x1EE4, 0x1EE4 }, // LATIN CAPITAL LETTER U WITH DOT BELOW
    Interval{ 0x1EE6, 0x1EE6 }, // LATIN CAPITAL LETTER U WITH HOOK ABOVE
    Interval{ 0x1EE8, 0x1EE8 }, // LATIN CAPITAL LETTER U WITH HORN AND ACUTE
    Interval{ 0x1EEA, 0x1EEA }, // LATIN CAPITAL LETTER U WITH HORN AND GRAVE
    Interval{ 0x1EEC, 0x1EEC }, // LATIN CAPITAL LETTER U WITH HORN AND HOOK ABOVE
    Interval{ 0x1EEE, 0x1EEE }, // LATIN CAPITAL LETTER U WITH HORN AND TILDE
    Interval{ 0x1EF0, 0x1EF0 }, // LATIN CAPITAL LETTER U WITH HORN AND DOT BELOW
    Interval{ 0x1EF2, 0x1EF2 }, // LATIN CAPITAL LETTER Y WITH GRAVE
    Interval{ 0x1EF4, 0x1EF4 }, // LATIN CAPITAL LETTER Y WITH DOT BELOW
    Interval{ 0x1EF6, 0x1EF6 }, // LATIN CAPITAL LETTER Y WITH HOOK ABOVE
    Interval{ 0x1EF8, 0x1EF8 }, // LATIN CAPITAL LETTER Y WITH TILDE
    Interval{ 0x1EFA, 0x1EFA }, // LATIN CAPITAL LETTER MIDDLE-WELSH LL
    Interval{ 0x1EFC, 0x1EFC }, // LATIN CAPITAL LETTER MIDDLE-WELSH V
    Interval{ 0x1EFE, 0x1EFE }, // LATIN CAPITAL LETTER Y WITH LOOP
    Interval{ 0x1F08, 0x1F0F }, // [8] GREEK CAPITAL LETTER ALPHA WITH PSILI..GREEK CAPITAL LETTER ALPHA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F18, 0x1F1D }, // [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Interval{ 0x1F28, 0x1F2F }, // [8] GREEK CAPITAL LETTER ETA WITH PSILI..GREEK CAPITAL LETTER ETA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F38, 0x1F3F }, // [8] GREEK CAPITAL LETTER IOTA WITH PSILI..GREEK CAPITAL LETTER IOTA WITH DASIA AND PERISPOMENI
    Interval{ 0x1F48, 0x1F4D }, // [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Interval{ 0x1F59, 0x1F59 }, // GREEK CAPITAL LETTER UPSILON WITH DASIA
    Interval{ 0x1F5B, 0x1F5B }, // GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Interval{ 0x1F5D, 0x1F5D }, // GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Interval{ 0x1F5F, 0x1F5F }, // GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Interval{ 0x1F68, 0x1F6F }, // [8] GREEK CAPITAL LETTER OMEGA WITH PSILI..GREEK CAPITAL LETTER OMEGA WITH DASIA AND PERISPOMENI
    Interval{ 0x1FB8, 0x1FBB }, // [4] GREEK CAPITAL LETTER ALPHA WITH VRACHY..GREEK CAPITAL LETTER ALPHA WITH OXIA
    Interval{ 0x1FC8, 0x1FCB }, // [4] GREEK CAPITAL LETTER EPSILON WITH VARIA..GREEK CAPITAL LETTER ETA WITH OXIA
    Interval{ 0x1FD8, 0x1FDB }, // [4] GREEK CAPITAL LETTER IOTA WITH VRACHY..GREEK CAPITAL LETTER IOTA WITH OXIA
    Interval{ 0x1FE8, 0x1FEC }, // [5] GREEK CAPITAL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Interval{ 0x1FF8, 0x1FFB }, // [4] GREEK CAPITAL LETTER OMICRON WITH VARIA..GREEK CAPITAL LETTER OMEGA WITH OXIA
    Interval{ 0x2102, 0x2102 }, // DOUBLE-STRUCK CAPITAL C
    Interval{ 0x2107, 0x2107 }, // EULER CONSTANT
    Interval{ 0x210B, 0x210D }, // [3] SCRIPT CAPITAL H..DOUBLE-STRUCK CAPITAL H
    Interval{ 0x2110, 0x2112 }, // [3] SCRIPT CAPITAL I..SCRIPT CAPITAL L
    Interval{ 0x2115, 0x2115 }, // DOUBLE-STRUCK CAPITAL N
    Interval{ 0x2119, 0x211D }, // [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    Interval{ 0x2124, 0x2124 }, // DOUBLE-STRUCK CAPITAL Z
    Interval{ 0x2126, 0x2126 }, // OHM SIGN
    Interval{ 0x2128, 0x2128 }, // BLACK-LETTER CAPITAL Z
    Interval{ 0x212A, 0x212D }, // [4] KELVIN SIGN..BLACK-LETTER CAPITAL C
    Interval{ 0x2130, 0x2133 }, // [4] SCRIPT CAPITAL E..SCRIPT CAPITAL M
    Interval{ 0x213E, 0x213F }, // [2] DOUBLE-STRUCK CAPITAL GAMMA..DOUBLE-STRUCK CAPITAL PI
    Interval{ 0x2145, 0x2145 }, // DOUBLE-STRUCK ITALIC CAPITAL D
    Interval{ 0x2183, 0x2183 }, // ROMAN NUMERAL REVERSED ONE HUNDRED
    Interval{ 0x2C00, 0x2C2F }, // [48] GLAGOLITIC CAPITAL LETTER AZU..GLAGOLITIC CAPITAL LETTER CAUDATE CHRIVI
    Interval{ 0x2C60, 0x2C60 }, // LATIN CAPITAL LETTER L WITH DOUBLE BAR
    Interval{ 0x2C62, 0x2C64 }, // [3] LATIN CAPITAL LETTER L WITH MIDDLE TILDE..LATIN CAPITAL LETTER R WITH TAIL
    Interval{ 0x2C67, 0x2C67 }, // LATIN CAPITAL LETTER H WITH DESCENDER
    Interval{ 0x2C69, 0x2C69 }, // LATIN CAPITAL LETTER K WITH DESCENDER
    Interval{ 0x2C6B, 0x2C6B }, // LATIN CAPITAL LETTER Z WITH DESCENDER
    Interval{ 0x2C6D, 0x2C70 }, // [4] LATIN CAPITAL LETTER ALPHA..LATIN CAPITAL LETTER TURNED ALPHA
    Interval{ 0x2C72, 0x2C72 }, // LATIN CAPITAL LETTER W WITH HOOK
    Interval{ 0x2C75, 0x2C75 }, // LATIN CAPITAL LETTER HALF H
    Interval{ 0x2C7E, 0x2C80 }, // [3] LATIN CAPITAL LETTER S WITH SWASH TAIL..COPTIC CAPITAL LETTER ALFA
    Interval{ 0x2C82, 0x2C82 }, // COPTIC CAPITAL LETTER VIDA
    Interval{ 0x2C84, 0x2C84 }, // COPTIC CAPITAL LETTER GAMMA
    Interval{ 0x2C86, 0x2C86 }, // COPTIC CAPITAL LETTER DALDA
    Interval{ 0x2C88, 0x2C88 }, // COPTIC CAPITAL LETTER EIE
    Interval{ 0x2C8A, 0x2C8A }, // COPTIC CAPITAL LETTER SOU
    Interval{ 0x2C8C, 0x2C8C }, // COPTIC CAPITAL LETTER ZATA
    Interval{ 0x2C8E, 0x2C8E }, // COPTIC CAPITAL LETTER HATE
    Interval{ 0x2C90, 0x2C90 }, // COPTIC CAPITAL LETTER THETHE
    Interval{ 0x2C92, 0x2C92 }, // COPTIC CAPITAL LETTER IAUDA
    Interval{ 0x2C94, 0x2C94 }, // COPTIC CAPITAL LETTER KAPA
    Interval{ 0x2C96, 0x2C96 }, // COPTIC CAPITAL LETTER LAULA
    Interval{ 0x2C98, 0x2C98 }, // COPTIC CAPITAL LETTER MI
    Interval{ 0x2C9A, 0x2C9A }, // COPTIC CAPITAL LETTER NI
    Interval{ 0x2C9C, 0x2C9C }, // COPTIC CAPITAL LETTER KSI
    Interval{ 0x2C9E, 0x2C9E }, // COPTIC CAPITAL LETTER O
    Interval{ 0x2CA0, 0x2CA0 }, // COPTIC CAPITAL LETTER PI
    Interval{ 0x2CA2, 0x2CA2 }, // COPTIC CAPITAL LETTER RO
    Interval{ 0x2CA4, 0x2CA4 }, // COPTIC CAPITAL LETTER SIMA
    Interval{ 0x2CA6, 0x2CA6 }, // COPTIC CAPITAL LETTER TAU
    Interval{ 0x2CA8, 0x2CA8 }, // COPTIC CAPITAL LETTER UA
    Interval{ 0x2CAA, 0x2CAA }, // COPTIC CAPITAL LETTER FI
    Interval{ 0x2CAC, 0x2CAC }, // COPTIC CAPITAL LETTER KHI
    Interval{ 0x2CAE, 0x2CAE }, // COPTIC CAPITAL LETTER PSI
    Interval{ 0x2CB0, 0x2CB0 }, // COPTIC CAPITAL LETTER OOU
    Interval{ 0x2CB2, 0x2CB2 }, // COPTIC CAPITAL LETTER DIALECT-P ALEF
    Interval{ 0x2CB4, 0x2CB4 }, // COPTIC CAPITAL LETTER OLD COPTIC AIN
    Interval{ 0x2CB6, 0x2CB6 }, // COPTIC CAPITAL LETTER CRYPTOGRAMMIC EIE
    Interval{ 0x2CB8, 0x2CB8 }, // COPTIC CAPITAL LETTER DIALECT-P KAPA
    Interval{ 0x2CBA, 0x2CBA }, // COPTIC CAPITAL LETTER DIALECT-P NI
    Interval{ 0x2CBC, 0x2CBC }, // COPTIC CAPITAL LETTER CRYPTOGRAMMIC NI
    Interval{ 0x2CBE, 0x2CBE }, // COPTIC CAPITAL LETTER OLD COPTIC OOU
    Interval{ 0x2CC0, 0x2CC0 }, // COPTIC CAPITAL LETTER SAMPI
    Interval{ 0x2CC2, 0x2CC2 }, // COPTIC CAPITAL LETTER CROSSED SHEI
    Interval{ 0x2CC4, 0x2CC4 }, // COPTIC CAPITAL LETTER OLD COPTIC SHEI
    Interval{ 0x2CC6, 0x2CC6 }, // COPTIC CAPITAL LETTER OLD COPTIC ESH
    Interval{ 0x2CC8, 0x2CC8 }, // COPTIC CAPITAL LETTER AKHMIMIC KHEI
    Interval{ 0x2CCA, 0x2CCA }, // COPTIC CAPITAL LETTER DIALECT-P HORI
    Interval{ 0x2CCC, 0x2CCC }, // COPTIC CAPITAL LETTER OLD COPTIC HORI
    Interval{ 0x2CCE, 0x2CCE }, // COPTIC CAPITAL LETTER OLD COPTIC HA
    Interval{ 0x2CD0, 0x2CD0 }, // COPTIC CAPITAL LETTER L-SHAPED HA
    Interval{ 0x2CD2, 0x2CD2 }, // COPTIC CAPITAL LETTER OLD COPTIC HEI
    Interval{ 0x2CD4, 0x2CD4 }, // COPTIC CAPITAL LETTER OLD COPTIC HAT
    Interval{ 0x2CD6, 0x2CD6 }, // COPTIC CAPITAL LETTER OLD COPTIC GANGIA
    Interval{ 0x2CD8, 0x2CD8 }, // COPTIC CAPITAL LETTER OLD COPTIC DJA
    Interval{ 0x2CDA, 0x2CDA }, // COPTIC CAPITAL LETTER OLD COPTIC SHIMA
    Interval{ 0x2CDC, 0x2CDC }, // COPTIC CAPITAL LETTER OLD NUBIAN SHIMA
    Interval{ 0x2CDE, 0x2CDE }, // COPTIC CAPITAL LETTER OLD NUBIAN NGI
    Interval{ 0x2CE0, 0x2CE0 }, // COPTIC CAPITAL LETTER OLD NUBIAN NYI
    Interval{ 0x2CE2, 0x2CE2 }, // COPTIC CAPITAL LETTER OLD NUBIAN WAU
    Interval{ 0x2CEB, 0x2CEB }, // COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI
    Interval{ 0x2CED, 0x2CED }, // COPTIC CAPITAL LETTER CRYPTOGRAMMIC GANGIA
    Interval{ 0x2CF2, 0x2CF2 }, // COPTIC CAPITAL LETTER BOHAIRIC KHEI
    Interval{ 0xA640, 0xA640 }, // CYRILLIC CAPITAL LETTER ZEMLYA
    Interval{ 0xA642, 0xA642 }, // CYRILLIC CAPITAL LETTER DZELO
    Interval{ 0xA644, 0xA644 }, // CYRILLIC CAPITAL LETTER REVERSED DZE
    Interval{ 0xA646, 0xA646 }, // CYRILLIC CAPITAL LETTER IOTA
    Interval{ 0xA648, 0xA648 }, // CYRILLIC CAPITAL LETTER DJERV
    Interval{ 0xA64A, 0xA64A }, // CYRILLIC CAPITAL LETTER MONOGRAPH UK
    Interval{ 0xA64C, 0xA64C }, // CYRILLIC CAPITAL LETTER BROAD OMEGA
    Interval{ 0xA64E, 0xA64E }, // CYRILLIC CAPITAL LETTER NEUTRAL YER
    Interval{ 0xA650, 0xA650 }, // CYRILLIC CAPITAL LETTER YERU WITH BACK YER
    Interval{ 0xA652, 0xA652 }, // CYRILLIC CAPITAL LETTER IOTIFIED YAT
    Interval{ 0xA654, 0xA654 }, // CYRILLIC CAPITAL LETTER REVERSED YU
    Interval{ 0xA656, 0xA656 }, // CYRILLIC CAPITAL LETTER IOTIFIED A
    Interval{ 0xA658, 0xA658 }, // CYRILLIC CAPITAL LETTER CLOSED LITTLE YUS
    Interval{ 0xA65A, 0xA65A }, // CYRILLIC CAPITAL LETTER BLENDED YUS
    Interval{ 0xA65C, 0xA65C }, // CYRILLIC CAPITAL LETTER IOTIFIED CLOSED LITTLE YUS
    Interval{ 0xA65E, 0xA65E }, // CYRILLIC CAPITAL LETTER YN
    Interval{ 0xA660, 0xA660 }, // CYRILLIC CAPITAL LETTER REVERSED TSE
    Interval{ 0xA662, 0xA662 }, // CYRILLIC CAPITAL LETTER SOFT DE
    Interval{ 0xA664, 0xA664 }, // CYRILLIC CAPITAL LETTER SOFT EL
    Interval{ 0xA666, 0xA666 }, // CYRILLIC CAPITAL LETTER SOFT EM
    Interval{ 0xA668, 0xA668 }, // CYRILLIC CAPITAL LETTER MONOCULAR O
    Interval{ 0xA66A, 0xA66A }, // CYRILLIC CAPITAL LETTER BINOCULAR O
    Interval{ 0xA66C, 0xA66C }, // CYRILLIC CAPITAL LETTER DOUBLE MONOCULAR O
    Interval{ 0xA680, 0xA680 }, // CYRILLIC CAPITAL LETTER DWE
    Interval{ 0xA682, 0xA682 }, // CYRILLIC CAPITAL LETTER DZWE
    Interval{ 0xA684, 0xA684 }, // CYRILLIC CAPITAL LETTER ZHWE
    Interval{ 0xA686, 0xA686 }, // CYRILLIC CAPITAL LETTER CCHE
    Interval{ 0xA688, 0xA688 }, // CYRILLIC CAPITAL LETTER DZZE
    Interval{ 0xA68A, 0xA68A }, // CYRILLIC CAPITAL LETTER TE WITH MIDDLE HOOK
    Interval{ 0xA68C, 0xA68C }, // CYRILLIC CAPITAL LETTER TWE
    Interval{ 0xA68E, 0xA68E }, // CYRILLIC CAPITAL LETTER TSWE
    Interval{ 0xA690, 0xA690 }, // CYRILLIC CAPITAL LETTER TSSE
    Interval{ 0xA692, 0xA692 }, // CYRILLIC CAPITAL LETTER TCHE
    Interval{ 0xA694, 0xA694 }, // CYRILLIC CAPITAL LETTER HWE
    Interval{ 0xA696, 0xA696 }, // CYRILLIC CAPITAL LETTER SHWE
    Interval{ 0xA698, 0xA698 }, // CYRILLIC CAPITAL LETTER DOUBLE O
    Interval{ 0xA69A, 0xA69A }, // CYRILLIC CAPITAL LETTER CROSSED O
    Interval{ 0xA722, 0xA722 }, // LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF
    Interval{ 0xA724, 0xA724 }, // LATIN CAPITAL LETTER EGYPTOLOGICAL AIN
    Interval{ 0xA726, 0xA726 }, // LATIN CAPITAL LETTER HENG
    Interval{ 0xA728, 0xA728 }, // LATIN CAPITAL LETTER TZ
    Interval{ 0xA72A, 0xA72A }, // LATIN CAPITAL LETTER TRESILLO
    Interval{ 0xA72C, 0xA72C }, // LATIN CAPITAL LETTER CUATRILLO
    Interval{ 0xA72E, 0xA72E }, // LATIN CAPITAL LETTER CUATRILLO WITH COMMA
    Interval{ 0xA732, 0xA732 }, // LATIN CAPITAL LETTER AA
    Interval{ 0xA734, 0xA734 }, // LATIN CAPITAL LETTER AO
    Interval{ 0xA736, 0xA736 }, // LATIN CAPITAL LETTER AU
    Interval{ 0xA738, 0xA738 }, // LATIN CAPITAL LETTER AV
    Interval{ 0xA73A, 0xA73A }, // LATIN CAPITAL LETTER AV WITH HORIZONTAL BAR
    Interval{ 0xA73C, 0xA73C }, // LATIN CAPITAL LETTER AY
    Interval{ 0xA73E, 0xA73E }, // LATIN CAPITAL LETTER REVERSED C WITH DOT
    Interval{ 0xA740, 0xA740 }, // LATIN CAPITAL LETTER K WITH STROKE
    Interval{ 0xA742, 0xA742 }, // LATIN CAPITAL LETTER K WITH DIAGONAL STROKE
    Interval{ 0xA744, 0xA744 }, // LATIN CAPITAL LETTER K WITH STROKE AND DIAGONAL STROKE
    Interval{ 0xA746, 0xA746 }, // LATIN CAPITAL LETTER BROKEN L
    Interval{ 0xA748, 0xA748 }, // LATIN CAPITAL LETTER L WITH HIGH STROKE
    Interval{ 0xA74A, 0xA74A }, // LATIN CAPITAL LETTER O WITH LONG STROKE OVERLAY
    Interval{ 0xA74C, 0xA74C }, // LATIN CAPITAL LETTER O WITH LOOP
    Interval{ 0xA74E, 0xA74E }, // LATIN CAPITAL LETTER OO
    Interval{ 0xA750, 0xA750 }, // LATIN CAPITAL LETTER P WITH STROKE THROUGH DESCENDER
    Interval{ 0xA752, 0xA752 }, // LATIN CAPITAL LETTER P WITH FLOURISH
    Interval{ 0xA754, 0xA754 }, // LATIN CAPITAL LETTER P WITH SQUIRREL TAIL
    Interval{ 0xA756, 0xA756 }, // LATIN CAPITAL LETTER Q WITH STROKE THROUGH DESCENDER
    Interval{ 0xA758, 0xA758 }, // LATIN CAPITAL LETTER Q WITH DIAGONAL STROKE
    Interval{ 0xA75A, 0xA75A }, // LATIN CAPITAL LETTER R ROTUNDA
    Interval{ 0xA75C, 0xA75C }, // LATIN CAPITAL LETTER RUM ROTUNDA
    Interval{ 0xA75E, 0xA75E }, // LATIN CAPITAL LETTER V WITH DIAGONAL STROKE
    Interval{ 0xA760, 0xA760 }, // LATIN CAPITAL LETTER VY
    Interval{ 0xA762, 0xA762 }, // LATIN CAPITAL LETTER VISIGOTHIC Z
    Interval{ 0xA764, 0xA764 }, // LATIN CAPITAL LETTER THORN WITH STROKE
    Interval{ 0xA766, 0xA766 }, // LATIN CAPITAL LETTER THORN WITH STROKE THROUGH DESCENDER
    Interval{ 0xA768, 0xA768 }, // LATIN CAPITAL LETTER VEND
    Interval{ 0xA76A, 0xA76A }, // LATIN CAPITAL LETTER ET
    Interval{ 0xA76C, 0xA76C }, // LATIN CAPITAL LETTER IS
    Interval{ 0xA76E, 0xA76E }, // LATIN CAPITAL LETTER CON
    Interval{ 0xA779, 0xA779 }, // LATIN CAPITAL LETTER INSULAR D
    Interval{ 0xA77B, 0xA77B }, // LATIN CAPITAL LETTER INSULAR F
    Interval{ 0xA77D, 0xA77E }, // [2] LATIN CAPITAL LETTER INSULAR G..LATIN CAPITAL LETTER TURNED INSULAR G
    Interval{ 0xA780, 0xA780 }, // LATIN CAPITAL LETTER TURNED L
    Interval{ 0xA782, 0xA782 }, // LATIN CAPITAL LETTER INSULAR R
    Interval{ 0xA784, 0xA784 }, // LATIN CAPITAL LETTER INSULAR S
    Interval{ 0xA786, 0xA786 }, // LATIN CAPITAL LETTER INSULAR T
    Interval{ 0xA78B, 0xA78B }, // LATIN CAPITAL LETTER SALTILLO
    Interval{ 0xA78D, 0xA78D }, // LATIN CAPITAL LETTER TURNED H
    Interval{ 0xA790, 0xA790 }, // LATIN CAPITAL LETTER N WITH DESCENDER
    Interval{ 0xA792, 0xA792 }, // LATIN CAPITAL LETTER C WITH BAR
    Interval{ 0xA796, 0xA796 }, // LATIN CAPITAL LETTER B WITH FLOURISH
    Interval{ 0xA798, 0xA798 }, // LATIN CAPITAL LETTER F WITH STROKE
    Interval{ 0xA79A, 0xA79A }, // LATIN CAPITAL LETTER VOLAPUK AE
    Interval{ 0xA79C, 0xA79C }, // LATIN CAPITAL LETTER VOLAPUK OE
    Interval{ 0xA79E, 0xA79E }, // LATIN CAPITAL LETTER VOLAPUK UE
    Interval{ 0xA7A0, 0xA7A0 }, // LATIN CAPITAL LETTER G WITH OBLIQUE STROKE
    Interval{ 0xA7A2, 0xA7A2 }, // LATIN CAPITAL LETTER K WITH OBLIQUE STROKE
    Interval{ 0xA7A4, 0xA7A4 }, // LATIN CAPITAL LETTER N WITH OBLIQUE STROKE
    Interval{ 0xA7A6, 0xA7A6 }, // LATIN CAPITAL LETTER R WITH OBLIQUE STROKE
    Interval{ 0xA7A8, 0xA7A8 }, // LATIN CAPITAL LETTER S WITH OBLIQUE STROKE
    Interval{ 0xA7AA, 0xA7AE }, // [5] LATIN CAPITAL LETTER H WITH HOOK..LATIN CAPITAL LETTER SMALL CAPITAL I
    Interval{ 0xA7B0, 0xA7B4 }, // [5] LATIN CAPITAL LETTER TURNED K..LATIN CAPITAL LETTER BETA
    Interval{ 0xA7B6, 0xA7B6 }, // LATIN CAPITAL LETTER OMEGA
    Interval{ 0xA7B8, 0xA7B8 }, // LATIN CAPITAL LETTER U WITH STROKE
    Interval{ 0xA7BA, 0xA7BA }, // LATIN CAPITAL LETTER GLOTTAL A
    Interval{ 0xA7BC, 0xA7BC }, // LATIN CAPITAL LETTER GLOTTAL I
    Interval{ 0xA7BE, 0xA7BE }, // LATIN CAPITAL LETTER GLOTTAL U
    Interval{ 0xA7C0, 0xA7C0 }, // LATIN CAPITAL LETTER OLD POLISH O
    Interval{ 0xA7C2, 0xA7C2 }, // LATIN CAPITAL LETTER ANGLICANA W
    Interval{ 0xA7C4, 0xA7C7 }, // [4] LATIN CAPITAL LETTER C WITH PALATAL HOOK..LATIN CAPITAL LETTER D WITH SHORT STROKE OVERLAY
    Interval{ 0xA7C9, 0xA7C9 }, // LATIN CAPITAL LETTER S WITH SHORT STROKE OVERLAY
    Interval{ 0xA7CB, 0xA7CC }, // [2] LATIN CAPITAL LETTER RAMS HORN..LATIN CAPITAL LETTER S WITH DIAGONAL STROKE
    Interval{ 0xA7CE, 0xA7CE }, // LATIN CAPITAL LETTER PHARYNGEAL VOICED FRICATIVE
    Interval{ 0xA7D0, 0xA7D0 }, // LATIN CAPITAL LETTER CLOSED INSULAR G
    Interval{ 0xA7D2, 0xA7D2 }, // LATIN CAPITAL LETTER DOUBLE THORN
    Interval{ 0xA7D4, 0xA7D4 }, // LATIN CAPITAL LETTER DOUBLE WYNN
    Interval{ 0xA7D6, 0xA7D6 }, // LATIN CAPITAL LETTER MIDDLE SCOTS S
    Interval{ 0xA7D8, 0xA7D8 }, // LATIN CAPITAL LETTER SIGMOID S
    Interval{ 0xA7DA, 0xA7DA }, // LATIN CAPITAL LETTER LAMBDA
    Interval{ 0xA7DC, 0xA7DC }, // LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Interval{ 0xA7F5, 0xA7F5 }, // LATIN CAPITAL LETTER REVERSED HALF H
    Interval{ 0xFF21, 0xFF3A }, // [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Interval{ 0x10400, 0x10427 }, // [40] DESERET CAPITAL LETTER LONG I..DESERET CAPITAL LETTER EW
    Interval{ 0x104B0, 0x104D3 }, // [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Interval{ 0x10570, 0x1057A }, // [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Interval{ 0x1057C, 0x1058A }, // [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Interval{ 0x1058C, 0x10592 }, // [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Interval{ 0x10594, 0x10595 }, // [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Interval{ 0x10C80, 0x10CB2 }, // [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Interval{ 0x10D50, 0x10D65 }, // [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Interval{ 0x118A0, 0x118BF }, // [32] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI CAPITAL LETTER VIYO
    Interval{ 0x16E40, 0x16E5F }, // [32] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN CAPITAL LETTER Y
    Interval{ 0x16EA0, 0x16EB8 }, // [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Interval{ 0x1D400, 0x1D419 }, // [26] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL BOLD CAPITAL Z
    Interval{ 0x1D434, 0x1D44D }, // [26] MATHEMATICAL ITALIC CAPITAL A..MATHEMATICAL ITALIC CAPITAL Z
    Interval{ 0x1D468, 0x1D481 }, // [26] MATHEMATICAL BOLD ITALIC CAPITAL A..MATHEMATICAL BOLD ITALIC CAPITAL Z
    Interval{ 0x1D49C, 0x1D49C }, // MATHEMATICAL SCRIPT CAPITAL A
    Interval{ 0x1D49E, 0x1D49F }, // [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Interval{ 0x1D4A2, 0x1D4A2 }, // MATHEMATICAL SCRIPT CAPITAL G
    Interval{ 0x1D4A5, 0x1D4A6 }, // [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Interval{ 0x1D4A9, 0x1D4AC }, // [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Interval{ 0x1D4AE, 0x1D4B5 }, // [8] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT CAPITAL Z
    Interval{ 0x1D4D0, 0x1D4E9 }, // [26] MATHEMATICAL BOLD SCRIPT CAPITAL A..MATHEMATICAL BOLD SCRIPT CAPITAL Z
    Interval{ 0x1D504, 0x1D505 }, // [2] MATHEMATICAL FRAKTUR CAPITAL A..MATHEMATICAL FRAKTUR CAPITAL B
    Interval{ 0x1D507, 0x1D50A }, // [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Interval{ 0x1D50D, 0x1D514 }, // [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Interval{ 0x1D516, 0x1D51C }, // [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Interval{ 0x1D538, 0x1D539 }, // [2] MATHEMATICAL DOUBLE-STRUCK CAPITAL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Interval{ 0x1D53B, 0x1D53E }, // [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Interval{ 0x1D540, 0x1D544 }, // [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Interval{ 0x1D546, 0x1D546 }, // MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Interval{ 0x1D54A, 0x1D550 }, // [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Interval{ 0x1D56C, 0x1D585 }, // [26] MATHEMATICAL BOLD FRAKTUR CAPITAL A..MATHEMATICAL BOLD FRAKTUR CAPITAL Z
    Interval{ 0x1D5A0, 0x1D5B9 }, // [26] MATHEMATICAL SANS-SERIF CAPITAL A..MATHEMATICAL SANS-SERIF CAPITAL Z
    Interval{ 0x1D5D4, 0x1D5ED }, // [26] MATHEMATICAL SANS-SERIF BOLD CAPITAL A..MATHEMATICAL SANS-SERIF BOLD CAPITAL Z
    Interval{ 0x1D608, 0x1D621 }, // [26] MATHEMATICAL SANS-SERIF ITALIC CAPITAL A..MATHEMATICAL SANS-SERIF ITALIC CAPITAL Z
    Interval{ 0x1D63C, 0x1D655 }, // [26] MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL A..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL Z
    Interval{ 0x1D670, 0x1D689 }, // [26] MATHEMATICAL MONOSPACE CAPITAL A..MATHEMATICAL MONOSPACE CAPITAL Z
    Interval{ 0x1D6A8, 0x1D6C0 }, // [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    Interval{ 0x1D6E2, 0x1D6FA }, // [25] MATHEMATICAL ITALIC CAPITAL ALPHA..MATHEMATICAL ITALIC CAPITAL OMEGA
    Interval{ 0x1D71C, 0x1D734 }, // [25] MATHEMATICAL BOLD ITALIC CAPITAL ALPHA..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D756, 0x1D76E }, // [25] MATHEMATICAL SANS-SERIF BOLD CAPITAL ALPHA..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    Interval{ 0x1D790, 0x1D7A8 }, // [25] MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    Interval{ 0x1D7CA, 0x1D7CA }, // MATHEMATICAL BOLD CAPITAL DIGAMMA
    Interval{ 0x1E900, 0x1E921 }, // [34] ADLAM CAPITAL LETTER ALIF..ADLAM CAPITAL LETTER SHA
}; // }}}
} // end namespace tables

bool contains(General_Category generalCategory, char32_t codepoint) noexcept {
    switch (generalCategory)
    {
        case General_Category::Close_Punctuation: return contains(tables::Close_Punctuation, codepoint);
        case General_Category::Connector_Punctuation: return contains(tables::Connector_Punctuation, codepoint);
        case General_Category::Control: return contains(tables::Control, codepoint);
        case General_Category::Currency_Symbol: return contains(tables::Currency_Symbol, codepoint);
        case General_Category::Dash_Punctuation: return contains(tables::Dash_Punctuation, codepoint);
        case General_Category::Decimal_Number: return contains(tables::Decimal_Number, codepoint);
        case General_Category::Enclosing_Mark: return contains(tables::Enclosing_Mark, codepoint);
        case General_Category::Final_Punctuation: return contains(tables::Final_Punctuation, codepoint);
        case General_Category::Format: return contains(tables::Format, codepoint);
        case General_Category::Initial_Punctuation: return contains(tables::Initial_Punctuation, codepoint);
        case General_Category::Letter_Number: return contains(tables::Letter_Number, codepoint);
        case General_Category::Line_Separator: return contains(tables::Line_Separator, codepoint);
        case General_Category::Lowercase_Letter: return contains(tables::Lowercase_Letter, codepoint);
        case General_Category::Math_Symbol: return contains(tables::Math_Symbol, codepoint);
        case General_Category::Modifier_Letter: return contains(tables::Modifier_Letter, codepoint);
        case General_Category::Modifier_Symbol: return contains(tables::Modifier_Symbol, codepoint);
        case General_Category::Nonspacing_Mark: return contains(tables::Nonspacing_Mark, codepoint);
        case General_Category::Open_Punctuation: return contains(tables::Open_Punctuation, codepoint);
        case General_Category::Other_Letter: return contains(tables::Other_Letter, codepoint);
        case General_Category::Other_Number: return contains(tables::Other_Number, codepoint);
        case General_Category::Other_Punctuation: return contains(tables::Other_Punctuation, codepoint);
        case General_Category::Other_Symbol: return contains(tables::Other_Symbol, codepoint);
        case General_Category::Paragraph_Separator: return contains(tables::Paragraph_Separator, codepoint);
        case General_Category::Private_Use: return contains(tables::Private_Use, codepoint);
        case General_Category::Space_Separator: return contains(tables::Space_Separator, codepoint);
        case General_Category::Spacing_Mark: return contains(tables::Spacing_Mark, codepoint);
        case General_Category::Surrogate: return contains(tables::Surrogate, codepoint);
        case General_Category::Titlecase_Letter: return contains(tables::Titlecase_Letter, codepoint);
        case General_Category::Unassigned: return contains(tables::Unassigned, codepoint);
        case General_Category::Uppercase_Letter: return contains(tables::Uppercase_Letter, codepoint);
        case General_Category::Unspecified: return false;
    }
    return false;
}

namespace tables {
auto static const Script = std::array<Prop<unicode::Script>, 2287>{ // {{{
    Prop<unicode::Script> { { 0x0000, 0x001F }, unicode::Script::Common }, // Cc  [32] <control-0000>..<control-001F>
    Prop<unicode::Script> { { 0x0020, 0x0020 }, unicode::Script::Common }, // Zs       SPACE
    Prop<unicode::Script> { { 0x0021, 0x0023 }, unicode::Script::Common }, // Po   [3] EXCLAMATION MARK..NUMBER SIGN
    Prop<unicode::Script> { { 0x0024, 0x0024 }, unicode::Script::Common }, // Sc       DOLLAR SIGN
    Prop<unicode::Script> { { 0x0025, 0x0027 }, unicode::Script::Common }, // Po   [3] PERCENT SIGN..APOSTROPHE
    Prop<unicode::Script> { { 0x0028, 0x0028 }, unicode::Script::Common }, // Ps       LEFT PARENTHESIS
    Prop<unicode::Script> { { 0x0029, 0x0029 }, unicode::Script::Common }, // Pe       RIGHT PARENTHESIS
    Prop<unicode::Script> { { 0x002A, 0x002A }, unicode::Script::Common }, // Po       ASTERISK
    Prop<unicode::Script> { { 0x002B, 0x002B }, unicode::Script::Common }, // Sm       PLUS SIGN
    Prop<unicode::Script> { { 0x002C, 0x002C }, unicode::Script::Common }, // Po       COMMA
    Prop<unicode::Script> { { 0x002D, 0x002D }, unicode::Script::Common }, // Pd       HYPHEN-MINUS
    Prop<unicode::Script> { { 0x002E, 0x002F }, unicode::Script::Common }, // Po   [2] FULL STOP..SOLIDUS
    Prop<unicode::Script> { { 0x0030, 0x0039 }, unicode::Script::Common }, // Nd  [10] DIGIT ZERO..DIGIT NINE
    Prop<unicode::Script> { { 0x003A, 0x003B }, unicode::Script::Common }, // Po   [2] COLON..SEMICOLON
    Prop<unicode::Script> { { 0x003C, 0x003E }, unicode::Script::Common }, // Sm   [3] LESS-THAN SIGN..GREATER-THAN SIGN
    Prop<unicode::Script> { { 0x003F, 0x0040 }, unicode::Script::Common }, // Po   [2] QUESTION MARK..COMMERCIAL AT
    Prop<unicode::Script> { { 0x0041, 0x005A }, unicode::Script::Latin }, // L&  [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    Prop<unicode::Script> { { 0x005B, 0x005B }, unicode::Script::Common }, // Ps       LEFT SQUARE BRACKET
    Prop<unicode::Script> { { 0x005C, 0x005C }, unicode::Script::Common }, // Po       REVERSE SOLIDUS
    Prop<unicode::Script> { { 0x005D, 0x005D }, unicode::Script::Common }, // Pe       RIGHT SQUARE BRACKET
    Prop<unicode::Script> { { 0x005E, 0x005E }, unicode::Script::Common }, // Sk       CIRCUMFLEX ACCENT
    Prop<unicode::Script> { { 0x005F, 0x005F }, unicode::Script::Common }, // Pc       LOW LINE
    Prop<unicode::Script> { { 0x0060, 0x0060 }, unicode::Script::Common }, // Sk       GRAVE ACCENT
    Prop<unicode::Script> { { 0x0061, 0x007A }, unicode::Script::Latin }, // L&  [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    Prop<unicode::Script> { { 0x007B, 0x007B }, unicode::Script::Common }, // Ps       LEFT CURLY BRACKET
    Prop<unicode::Script> { { 0x007C, 0x007C }, unicode::Script::Common }, // Sm       VERTICAL LINE
    Prop<unicode::Script> { { 0x007D, 0x007D }, unicode::Script::Common }, // Pe       RIGHT CURLY BRACKET
    Prop<unicode::Script> { { 0x007E, 0x007E }, unicode::Script::Common }, // Sm       TILDE
    Prop<unicode::Script> { { 0x007F, 0x009F }, unicode::Script::Common }, // Cc  [33] <control-007F>..<control-009F>
    Prop<unicode::Script> { { 0x00A0, 0x00A0 }, unicode::Script::Common }, // Zs       NO-BREAK SPACE
    Prop<unicode::Script> { { 0x00A1, 0x00A1 }, unicode::Script::Common }, // Po       INVERTED EXCLAMATION MARK
    Prop<unicode::Script> { { 0x00A2, 0x00A5 }, unicode::Script::Common }, // Sc   [4] CENT SIGN..YEN SIGN
    Prop<unicode::Script> { { 0x00A6, 0x00A6 }, unicode::Script::Common }, // So       BROKEN BAR
    Prop<unicode::Script> { { 0x00A7, 0x00A7 }, unicode::Script::Common }, // Po       SECTION SIGN
    Prop<unicode::Script> { { 0x00A8, 0x00A8 }, unicode::Script::Common }, // Sk       DIAERESIS
    Prop<unicode::Script> { { 0x00A9, 0x00A9 }, unicode::Script::Common }, // So       COPYRIGHT SIGN
    Prop<unicode::Script> { { 0x00AA, 0x00AA }, unicode::Script::Latin }, // Lo       FEMININE ORDINAL INDICATOR
    Prop<unicode::Script> { { 0x00AB, 0x00AB }, unicode::Script::Common }, // Pi       LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    Prop<unicode::Script> { { 0x00AC, 0x00AC }, unicode::Script::Common }, // Sm       NOT SIGN
    Prop<unicode::Script> { { 0x00AD, 0x00AD }, unicode::Script::Common }, // Cf       SOFT HYPHEN
    Prop<unicode::Script> { { 0x00AE, 0x00AE }, unicode::Script::Common }, // So       REGISTERED SIGN
    Prop<unicode::Script> { { 0x00AF, 0x00AF }, unicode::Script::Common }, // Sk       MACRON
    Prop<unicode::Script> { { 0x00B0, 0x00B0 }, unicode::Script::Common }, // So       DEGREE SIGN
    Prop<unicode::Script> { { 0x00B1, 0x00B1 }, unicode::Script::Common }, // Sm       PLUS-MINUS SIGN
    Prop<unicode::Script> { { 0x00B2, 0x00B3 }, unicode::Script::Common }, // No   [2] SUPERSCRIPT TWO..SUPERSCRIPT THREE
    Prop<unicode::Script> { { 0x00B4, 0x00B4 }, unicode::Script::Common }, // Sk       ACUTE ACCENT
    Prop<unicode::Script> { { 0x00B5, 0x00B5 }, unicode::Script::Common }, // L&       MICRO SIGN
    Prop<unicode::Script> { { 0x00B6, 0x00B7 }, unicode::Script::Common }, // Po   [2] PILCROW SIGN..MIDDLE DOT
    Prop<unicode::Script> { { 0x00B8, 0x00B8 }, unicode::Script::Common }, // Sk       CEDILLA
    Prop<unicode::Script> { { 0x00B9, 0x00B9 }, unicode::Script::Common }, // No       SUPERSCRIPT ONE
    Prop<unicode::Script> { { 0x00BA, 0x00BA }, unicode::Script::Latin }, // Lo       MASCULINE ORDINAL INDICATOR
    Prop<unicode::Script> { { 0x00BB, 0x00BB }, unicode::Script::Common }, // Pf       RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    Prop<unicode::Script> { { 0x00BC, 0x00BE }, unicode::Script::Common }, // No   [3] VULGAR FRACTION ONE QUARTER..VULGAR FRACTION THREE QUARTERS
    Prop<unicode::Script> { { 0x00BF, 0x00BF }, unicode::Script::Common }, // Po       INVERTED QUESTION MARK
    Prop<unicode::Script> { { 0x00C0, 0x00D6 }, unicode::Script::Latin }, // L&  [23] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER O WITH DIAERESIS
    Prop<unicode::Script> { { 0x00D7, 0x00D7 }, unicode::Script::Common }, // Sm       MULTIPLICATION SIGN
    Prop<unicode::Script> { { 0x00D8, 0x00F6 }, unicode::Script::Latin }, // L&  [31] LATIN CAPITAL LETTER O WITH STROKE..LATIN SMALL LETTER O WITH DIAERESIS
    Prop<unicode::Script> { { 0x00F7, 0x00F7 }, unicode::Script::Common }, // Sm       DIVISION SIGN
    Prop<unicode::Script> { { 0x00F8, 0x01BA }, unicode::Script::Latin }, // L& [195] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER EZH WITH TAIL
    Prop<unicode::Script> { { 0x01BB, 0x01BB }, unicode::Script::Latin }, // Lo       LATIN LETTER TWO WITH STROKE
    Prop<unicode::Script> { { 0x01BC, 0x01BF }, unicode::Script::Latin }, // L&   [4] LATIN CAPITAL LETTER TONE FIVE..LATIN LETTER WYNN
    Prop<unicode::Script> { { 0x01C0, 0x01C3 }, unicode::Script::Latin }, // Lo   [4] LATIN LETTER DENTAL CLICK..LATIN LETTER RETROFLEX CLICK
    Prop<unicode::Script> { { 0x01C4, 0x0293 }, unicode::Script::Latin }, // L& [208] LATIN CAPITAL LETTER DZ WITH CARON..LATIN SMALL LETTER EZH WITH CURL
    Prop<unicode::Script> { { 0x0294, 0x0295 }, unicode::Script::Latin }, // Lo   [2] LATIN LETTER GLOTTAL STOP..LATIN LETTER PHARYNGEAL VOICED FRICATIVE
    Prop<unicode::Script> { { 0x0296, 0x02AF }, unicode::Script::Latin }, // L&  [26] LATIN LETTER INVERTED GLOTTAL STOP..LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    Prop<unicode::Script> { { 0x02B0, 0x02B8 }, unicode::Script::Latin }, // Lm   [9] MODIFIER LETTER SMALL H..MODIFIER LETTER SMALL Y
    Prop<unicode::Script> { { 0x02B9, 0x02C1 }, unicode::Script::Common }, // Lm   [9] MODIFIER LETTER PRIME..MODIFIER LETTER REVERSED GLOTTAL STOP
    Prop<unicode::Script> { { 0x02C2, 0x02C5 }, unicode::Script::Common }, // Sk   [4] MODIFIER LETTER LEFT ARROWHEAD..MODIFIER LETTER DOWN ARROWHEAD
    Prop<unicode::Script> { { 0x02C6, 0x02D1 }, unicode::Script::Common }, // Lm  [12] MODIFIER LETTER CIRCUMFLEX ACCENT..MODIFIER LETTER HALF TRIANGULAR COLON
    Prop<unicode::Script> { { 0x02D2, 0x02DF }, unicode::Script::Common }, // Sk  [14] MODIFIER LETTER CENTRED RIGHT HALF RING..MODIFIER LETTER CROSS ACCENT
    Prop<unicode::Script> { { 0x02E0, 0x02E4 }, unicode::Script::Latin }, // Lm   [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    Prop<unicode::Script> { { 0x02E5, 0x02E9 }, unicode::Script::Common }, // Sk   [5] MODIFIER LETTER EXTRA-HIGH TONE BAR..MODIFIER LETTER EXTRA-LOW TONE BAR
    Prop<unicode::Script> { { 0x02EA, 0x02EB }, unicode::Script::Bopomofo }, // Sk   [2] MODIFIER LETTER YIN DEPARTING TONE MARK..MODIFIER LETTER YANG DEPARTING TONE MARK
    Prop<unicode::Script> { { 0x02EC, 0x02EC }, unicode::Script::Common }, // Lm       MODIFIER LETTER VOICING
    Prop<unicode::Script> { { 0x02ED, 0x02ED }, unicode::Script::Common }, // Sk       MODIFIER LETTER UNASPIRATED
    Prop<unicode::Script> { { 0x02EE, 0x02EE }, unicode::Script::Common }, // Lm       MODIFIER LETTER DOUBLE APOSTROPHE
    Prop<unicode::Script> { { 0x02EF, 0x02FF }, unicode::Script::Common }, // Sk  [17] MODIFIER LETTER LOW DOWN ARROWHEAD..MODIFIER LETTER LOW LEFT ARROW
    Prop<unicode::Script> { { 0x0300, 0x036F }, unicode::Script::Inherited }, // Mn [112] COMBINING GRAVE ACCENT..COMBINING LATIN SMALL LETTER X
    Prop<unicode::Script> { { 0x0370, 0x0373 }, unicode::Script::Greek }, // L&   [4] GREEK CAPITAL LETTER HETA..GREEK SMALL LETTER ARCHAIC SAMPI
    Prop<unicode::Script> { { 0x0374, 0x0374 }, unicode::Script::Common }, // Lm       GREEK NUMERAL SIGN
    Prop<unicode::Script> { { 0x0375, 0x0375 }, unicode::Script::Greek }, // Sk       GREEK LOWER NUMERAL SIGN
    Prop<unicode::Script> { { 0x0376, 0x0377 }, unicode::Script::Greek }, // L&   [2] GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA..GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Prop<unicode::Script> { { 0x037A, 0x037A }, unicode::Script::Greek }, // Lm       GREEK YPOGEGRAMMENI
    Prop<unicode::Script> { { 0x037B, 0x037D }, unicode::Script::Greek }, // L&   [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Prop<unicode::Script> { { 0x037E, 0x037E }, unicode::Script::Common }, // Po       GREEK QUESTION MARK
    Prop<unicode::Script> { { 0x037F, 0x037F }, unicode::Script::Greek }, // L&       GREEK CAPITAL LETTER YOT
    Prop<unicode::Script> { { 0x0384, 0x0384 }, unicode::Script::Greek }, // Sk       GREEK TONOS
    Prop<unicode::Script> { { 0x0385, 0x0385 }, unicode::Script::Common }, // Sk       GREEK DIALYTIKA TONOS
    Prop<unicode::Script> { { 0x0386, 0x0386 }, unicode::Script::Greek }, // L&       GREEK CAPITAL LETTER ALPHA WITH TONOS
    Prop<unicode::Script> { { 0x0387, 0x0387 }, unicode::Script::Common }, // Po       GREEK ANO TELEIA
    Prop<unicode::Script> { { 0x0388, 0x038A }, unicode::Script::Greek }, // L&   [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Prop<unicode::Script> { { 0x038C, 0x038C }, unicode::Script::Greek }, // L&       GREEK CAPITAL LETTER OMICRON WITH TONOS
    Prop<unicode::Script> { { 0x038E, 0x03A1 }, unicode::Script::Greek }, // L&  [20] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK CAPITAL LETTER RHO
    Prop<unicode::Script> { { 0x03A3, 0x03E1 }, unicode::Script::Greek }, // L&  [63] GREEK CAPITAL LETTER SIGMA..GREEK SMALL LETTER SAMPI
    Prop<unicode::Script> { { 0x03E2, 0x03EF }, unicode::Script::Coptic }, // L&  [14] COPTIC CAPITAL LETTER SHEI..COPTIC SMALL LETTER DEI
    Prop<unicode::Script> { { 0x03F0, 0x03F5 }, unicode::Script::Greek }, // L&   [6] GREEK KAPPA SYMBOL..GREEK LUNATE EPSILON SYMBOL
    Prop<unicode::Script> { { 0x03F6, 0x03F6 }, unicode::Script::Greek }, // Sm       GREEK REVERSED LUNATE EPSILON SYMBOL
    Prop<unicode::Script> { { 0x03F7, 0x03FF }, unicode::Script::Greek }, // L&   [9] GREEK CAPITAL LETTER SHO..GREEK CAPITAL REVERSED DOTTED LUNATE SIGMA SYMBOL
    Prop<unicode::Script> { { 0x0400, 0x0481 }, unicode::Script::Cyrillic }, // L& [130] CYRILLIC CAPITAL LETTER IE WITH GRAVE..CYRILLIC SMALL LETTER KOPPA
    Prop<unicode::Script> { { 0x0482, 0x0482 }, unicode::Script::Cyrillic }, // So       CYRILLIC THOUSANDS SIGN
    Prop<unicode::Script> { { 0x0483, 0x0484 }, unicode::Script::Cyrillic }, // Mn   [2] COMBINING CYRILLIC TITLO..COMBINING CYRILLIC PALATALIZATION
    Prop<unicode::Script> { { 0x0485, 0x0486 }, unicode::Script::Inherited }, // Mn   [2] COMBINING CYRILLIC DASIA PNEUMATA..COMBINING CYRILLIC PSILI PNEUMATA
    Prop<unicode::Script> { { 0x0487, 0x0487 }, unicode::Script::Cyrillic }, // Mn       COMBINING CYRILLIC POKRYTIE
    Prop<unicode::Script> { { 0x0488, 0x0489 }, unicode::Script::Cyrillic }, // Me   [2] COMBINING CYRILLIC HUNDRED THOUSANDS SIGN..COMBINING CYRILLIC MILLIONS SIGN
    Prop<unicode::Script> { { 0x048A, 0x052F }, unicode::Script::Cyrillic }, // L& [166] CYRILLIC CAPITAL LETTER SHORT I WITH TAIL..CYRILLIC SMALL LETTER EL WITH DESCENDER
    Prop<unicode::Script> { { 0x0531, 0x0556 }, unicode::Script::Armenian }, // L&  [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    Prop<unicode::Script> { { 0x0559, 0x0559 }, unicode::Script::Armenian }, // Lm       ARMENIAN MODIFIER LETTER LEFT HALF RING
    Prop<unicode::Script> { { 0x055A, 0x055F }, unicode::Script::Armenian }, // Po   [6] ARMENIAN APOSTROPHE..ARMENIAN ABBREVIATION MARK
    Prop<unicode::Script> { { 0x0560, 0x0588 }, unicode::Script::Armenian }, // L&  [41] ARMENIAN SMALL LETTER TURNED AYB..ARMENIAN SMALL LETTER YI WITH STROKE
    Prop<unicode::Script> { { 0x0589, 0x0589 }, unicode::Script::Armenian }, // Po       ARMENIAN FULL STOP
    Prop<unicode::Script> { { 0x058A, 0x058A }, unicode::Script::Armenian }, // Pd       ARMENIAN HYPHEN
    Prop<unicode::Script> { { 0x058D, 0x058E }, unicode::Script::Armenian }, // So   [2] RIGHT-FACING ARMENIAN ETERNITY SIGN..LEFT-FACING ARMENIAN ETERNITY SIGN
    Prop<unicode::Script> { { 0x058F, 0x058F }, unicode::Script::Armenian }, // Sc       ARMENIAN DRAM SIGN
    Prop<unicode::Script> { { 0x0591, 0x05BD }, unicode::Script::Hebrew }, // Mn  [45] HEBREW ACCENT ETNAHTA..HEBREW POINT METEG
    Prop<unicode::Script> { { 0x05BE, 0x05BE }, unicode::Script::Hebrew }, // Pd       HEBREW PUNCTUATION MAQAF
    Prop<unicode::Script> { { 0x05BF, 0x05BF }, unicode::Script::Hebrew }, // Mn       HEBREW POINT RAFE
    Prop<unicode::Script> { { 0x05C0, 0x05C0 }, unicode::Script::Hebrew }, // Po       HEBREW PUNCTUATION PASEQ
    Prop<unicode::Script> { { 0x05C1, 0x05C2 }, unicode::Script::Hebrew }, // Mn   [2] HEBREW POINT SHIN DOT..HEBREW POINT SIN DOT
    Prop<unicode::Script> { { 0x05C3, 0x05C3 }, unicode::Script::Hebrew }, // Po       HEBREW PUNCTUATION SOF PASUQ
    Prop<unicode::Script> { { 0x05C4, 0x05C5 }, unicode::Script::Hebrew }, // Mn   [2] HEBREW MARK UPPER DOT..HEBREW MARK LOWER DOT
    Prop<unicode::Script> { { 0x05C6, 0x05C6 }, unicode::Script::Hebrew }, // Po       HEBREW PUNCTUATION NUN HAFUKHA
    Prop<unicode::Script> { { 0x05C7, 0x05C7 }, unicode::Script::Hebrew }, // Mn       HEBREW POINT QAMATS QATAN
    Prop<unicode::Script> { { 0x05D0, 0x05EA }, unicode::Script::Hebrew }, // Lo  [27] HEBREW LETTER ALEF..HEBREW LETTER TAV
    Prop<unicode::Script> { { 0x05EF, 0x05F2 }, unicode::Script::Hebrew }, // Lo   [4] HEBREW YOD TRIANGLE..HEBREW LIGATURE YIDDISH DOUBLE YOD
    Prop<unicode::Script> { { 0x05F3, 0x05F4 }, unicode::Script::Hebrew }, // Po   [2] HEBREW PUNCTUATION GERESH..HEBREW PUNCTUATION GERSHAYIM
    Prop<unicode::Script> { { 0x0600, 0x0604 }, unicode::Script::Arabic }, // Cf   [5] ARABIC NUMBER SIGN..ARABIC SIGN SAMVAT
    Prop<unicode::Script> { { 0x0605, 0x0605 }, unicode::Script::Common }, // Cf       ARABIC NUMBER MARK ABOVE
    Prop<unicode::Script> { { 0x0606, 0x0608 }, unicode::Script::Arabic }, // Sm   [3] ARABIC-INDIC CUBE ROOT..ARABIC RAY
    Prop<unicode::Script> { { 0x0609, 0x060A }, unicode::Script::Arabic }, // Po   [2] ARABIC-INDIC PER MILLE SIGN..ARABIC-INDIC PER TEN THOUSAND SIGN
    Prop<unicode::Script> { { 0x060B, 0x060B }, unicode::Script::Arabic }, // Sc       AFGHANI SIGN
    Prop<unicode::Script> { { 0x060C, 0x060C }, unicode::Script::Common }, // Po       ARABIC COMMA
    Prop<unicode::Script> { { 0x060D, 0x060D }, unicode::Script::Arabic }, // Po       ARABIC DATE SEPARATOR
    Prop<unicode::Script> { { 0x060E, 0x060F }, unicode::Script::Arabic }, // So   [2] ARABIC POETIC VERSE SIGN..ARABIC SIGN MISRA
    Prop<unicode::Script> { { 0x0610, 0x061A }, unicode::Script::Arabic }, // Mn  [11] ARABIC SIGN SALLALLAHOU ALAYHE WASSALLAM..ARABIC SMALL KASRA
    Prop<unicode::Script> { { 0x061B, 0x061B }, unicode::Script::Common }, // Po       ARABIC SEMICOLON
    Prop<unicode::Script> { { 0x061C, 0x061C }, unicode::Script::Arabic }, // Cf       ARABIC LETTER MARK
    Prop<unicode::Script> { { 0x061D, 0x061E }, unicode::Script::Arabic }, // Po   [2] ARABIC END OF TEXT MARK..ARABIC TRIPLE DOT PUNCTUATION MARK
    Prop<unicode::Script> { { 0x061F, 0x061F }, unicode::Script::Common }, // Po       ARABIC QUESTION MARK
    Prop<unicode::Script> { { 0x0620, 0x063F }, unicode::Script::Arabic }, // Lo  [32] ARABIC LETTER KASHMIRI YEH..ARABIC LETTER FARSI YEH WITH THREE DOTS ABOVE
    Prop<unicode::Script> { { 0x0640, 0x0640 }, unicode::Script::Common }, // Lm       ARABIC TATWEEL
    Prop<unicode::Script> { { 0x0641, 0x064A }, unicode::Script::Arabic }, // Lo  [10] ARABIC LETTER FEH..ARABIC LETTER YEH
    Prop<unicode::Script> { { 0x064B, 0x0655 }, unicode::Script::Inherited }, // Mn  [11] ARABIC FATHATAN..ARABIC HAMZA BELOW
    Prop<unicode::Script> { { 0x0656, 0x065F }, unicode::Script::Arabic }, // Mn  [10] ARABIC SUBSCRIPT ALEF..ARABIC WAVY HAMZA BELOW
    Prop<unicode::Script> { { 0x0660, 0x0669 }, unicode::Script::Arabic }, // Nd  [10] ARABIC-INDIC DIGIT ZERO..ARABIC-INDIC DIGIT NINE
    Prop<unicode::Script> { { 0x066A, 0x066D }, unicode::Script::Arabic }, // Po   [4] ARABIC PERCENT SIGN..ARABIC FIVE POINTED STAR
    Prop<unicode::Script> { { 0x066E, 0x066F }, unicode::Script::Arabic }, // Lo   [2] ARABIC LETTER DOTLESS BEH..ARABIC LETTER DOTLESS QAF
    Prop<unicode::Script> { { 0x0670, 0x0670 }, unicode::Script::Inherited }, // Mn       ARABIC LETTER SUPERSCRIPT ALEF
    Prop<unicode::Script> { { 0x0671, 0x06D3 }, unicode::Script::Arabic }, // Lo  [99] ARABIC LETTER ALEF WASLA..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE
    Prop<unicode::Script> { { 0x06D4, 0x06D4 }, unicode::Script::Arabic }, // Po       ARABIC FULL STOP
    Prop<unicode::Script> { { 0x06D5, 0x06D5 }, unicode::Script::Arabic }, // Lo       ARABIC LETTER AE
    Prop<unicode::Script> { { 0x06D6, 0x06DC }, unicode::Script::Arabic }, // Mn   [7] ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA..ARABIC SMALL HIGH SEEN
    Prop<unicode::Script> { { 0x06DD, 0x06DD }, unicode::Script::Common }, // Cf       ARABIC END OF AYAH
    Prop<unicode::Script> { { 0x06DE, 0x06DE }, unicode::Script::Arabic }, // So       ARABIC START OF RUB EL HIZB
    Prop<unicode::Script> { { 0x06DF, 0x06E4 }, unicode::Script::Arabic }, // Mn   [6] ARABIC SMALL HIGH ROUNDED ZERO..ARABIC SMALL HIGH MADDA
    Prop<unicode::Script> { { 0x06E5, 0x06E6 }, unicode::Script::Arabic }, // Lm   [2] ARABIC SMALL WAW..ARABIC SMALL YEH
    Prop<unicode::Script> { { 0x06E7, 0x06E8 }, unicode::Script::Arabic }, // Mn   [2] ARABIC SMALL HIGH YEH..ARABIC SMALL HIGH NOON
    Prop<unicode::Script> { { 0x06E9, 0x06E9 }, unicode::Script::Arabic }, // So       ARABIC PLACE OF SAJDAH
    Prop<unicode::Script> { { 0x06EA, 0x06ED }, unicode::Script::Arabic }, // Mn   [4] ARABIC EMPTY CENTRE LOW STOP..ARABIC SMALL LOW MEEM
    Prop<unicode::Script> { { 0x06EE, 0x06EF }, unicode::Script::Arabic }, // Lo   [2] ARABIC LETTER DAL WITH INVERTED V..ARABIC LETTER REH WITH INVERTED V
    Prop<unicode::Script> { { 0x06F0, 0x06F9 }, unicode::Script::Arabic }, // Nd  [10] EXTENDED ARABIC-INDIC DIGIT ZERO..EXTENDED ARABIC-INDIC DIGIT NINE
    Prop<unicode::Script> { { 0x06FA, 0x06FC }, unicode::Script::Arabic }, // Lo   [3] ARABIC LETTER SHEEN WITH DOT BELOW..ARABIC LETTER GHAIN WITH DOT BELOW
    Prop<unicode::Script> { { 0x06FD, 0x06FE }, unicode::Script::Arabic }, // So   [2] ARABIC SIGN SINDHI AMPERSAND..ARABIC SIGN SINDHI POSTPOSITION MEN
    Prop<unicode::Script> { { 0x06FF, 0x06FF }, unicode::Script::Arabic }, // Lo       ARABIC LETTER HEH WITH INVERTED V
    Prop<unicode::Script> { { 0x0700, 0x070D }, unicode::Script::Syriac }, // Po  [14] SYRIAC END OF PARAGRAPH..SYRIAC HARKLEAN ASTERISCUS
    Prop<unicode::Script> { { 0x070F, 0x070F }, unicode::Script::Syriac }, // Cf       SYRIAC ABBREVIATION MARK
    Prop<unicode::Script> { { 0x0710, 0x0710 }, unicode::Script::Syriac }, // Lo       SYRIAC LETTER ALAPH
    Prop<unicode::Script> { { 0x0711, 0x0711 }, unicode::Script::Syriac }, // Mn       SYRIAC LETTER SUPERSCRIPT ALAPH
    Prop<unicode::Script> { { 0x0712, 0x072F }, unicode::Script::Syriac }, // Lo  [30] SYRIAC LETTER BETH..SYRIAC LETTER PERSIAN DHALATH
    Prop<unicode::Script> { { 0x0730, 0x074A }, unicode::Script::Syriac }, // Mn  [27] SYRIAC PTHAHA ABOVE..SYRIAC BARREKH
    Prop<unicode::Script> { { 0x074D, 0x074F }, unicode::Script::Syriac }, // Lo   [3] SYRIAC LETTER SOGDIAN ZHAIN..SYRIAC LETTER SOGDIAN FE
    Prop<unicode::Script> { { 0x0750, 0x077F }, unicode::Script::Arabic }, // Lo  [48] ARABIC LETTER BEH WITH THREE DOTS HORIZONTALLY BELOW..ARABIC LETTER KAF WITH TWO DOTS ABOVE
    Prop<unicode::Script> { { 0x0780, 0x07A5 }, unicode::Script::Thaana }, // Lo  [38] THAANA LETTER HAA..THAANA LETTER WAAVU
    Prop<unicode::Script> { { 0x07A6, 0x07B0 }, unicode::Script::Thaana }, // Mn  [11] THAANA ABAFILI..THAANA SUKUN
    Prop<unicode::Script> { { 0x07B1, 0x07B1 }, unicode::Script::Thaana }, // Lo       THAANA LETTER NAA
    Prop<unicode::Script> { { 0x07C0, 0x07C9 }, unicode::Script::Nko }, // Nd  [10] NKO DIGIT ZERO..NKO DIGIT NINE
    Prop<unicode::Script> { { 0x07CA, 0x07EA }, unicode::Script::Nko }, // Lo  [33] NKO LETTER A..NKO LETTER JONA RA
    Prop<unicode::Script> { { 0x07EB, 0x07F3 }, unicode::Script::Nko }, // Mn   [9] NKO COMBINING SHORT HIGH TONE..NKO COMBINING DOUBLE DOT ABOVE
    Prop<unicode::Script> { { 0x07F4, 0x07F5 }, unicode::Script::Nko }, // Lm   [2] NKO HIGH TONE APOSTROPHE..NKO LOW TONE APOSTROPHE
    Prop<unicode::Script> { { 0x07F6, 0x07F6 }, unicode::Script::Nko }, // So       NKO SYMBOL OO DENNEN
    Prop<unicode::Script> { { 0x07F7, 0x07F9 }, unicode::Script::Nko }, // Po   [3] NKO SYMBOL GBAKURUNEN..NKO EXCLAMATION MARK
    Prop<unicode::Script> { { 0x07FA, 0x07FA }, unicode::Script::Nko }, // Lm       NKO LAJANYALAN
    Prop<unicode::Script> { { 0x07FD, 0x07FD }, unicode::Script::Nko }, // Mn       NKO DANTAYALAN
    Prop<unicode::Script> { { 0x07FE, 0x07FF }, unicode::Script::Nko }, // Sc   [2] NKO DOROME SIGN..NKO TAMAN SIGN
    Prop<unicode::Script> { { 0x0800, 0x0815 }, unicode::Script::Samaritan }, // Lo  [22] SAMARITAN LETTER ALAF..SAMARITAN LETTER TAAF
    Prop<unicode::Script> { { 0x0816, 0x0819 }, unicode::Script::Samaritan }, // Mn   [4] SAMARITAN MARK IN..SAMARITAN MARK DAGESH
    Prop<unicode::Script> { { 0x081A, 0x081A }, unicode::Script::Samaritan }, // Lm       SAMARITAN MODIFIER LETTER EPENTHETIC YUT
    Prop<unicode::Script> { { 0x081B, 0x0823 }, unicode::Script::Samaritan }, // Mn   [9] SAMARITAN MARK EPENTHETIC YUT..SAMARITAN VOWEL SIGN A
    Prop<unicode::Script> { { 0x0824, 0x0824 }, unicode::Script::Samaritan }, // Lm       SAMARITAN MODIFIER LETTER SHORT A
    Prop<unicode::Script> { { 0x0825, 0x0827 }, unicode::Script::Samaritan }, // Mn   [3] SAMARITAN VOWEL SIGN SHORT A..SAMARITAN VOWEL SIGN U
    Prop<unicode::Script> { { 0x0828, 0x0828 }, unicode::Script::Samaritan }, // Lm       SAMARITAN MODIFIER LETTER I
    Prop<unicode::Script> { { 0x0829, 0x082D }, unicode::Script::Samaritan }, // Mn   [5] SAMARITAN VOWEL SIGN LONG I..SAMARITAN MARK NEQUDAA
    Prop<unicode::Script> { { 0x0830, 0x083E }, unicode::Script::Samaritan }, // Po  [15] SAMARITAN PUNCTUATION NEQUDAA..SAMARITAN PUNCTUATION ANNAAU
    Prop<unicode::Script> { { 0x0840, 0x0858 }, unicode::Script::Mandaic }, // Lo  [25] MANDAIC LETTER HALQA..MANDAIC LETTER AIN
    Prop<unicode::Script> { { 0x0859, 0x085B }, unicode::Script::Mandaic }, // Mn   [3] MANDAIC AFFRICATION MARK..MANDAIC GEMINATION MARK
    Prop<unicode::Script> { { 0x085E, 0x085E }, unicode::Script::Mandaic }, // Po       MANDAIC PUNCTUATION
    Prop<unicode::Script> { { 0x0860, 0x086A }, unicode::Script::Syriac }, // Lo  [11] SYRIAC LETTER MALAYALAM NGA..SYRIAC LETTER MALAYALAM SSA
    Prop<unicode::Script> { { 0x0870, 0x0887 }, unicode::Script::Arabic }, // Lo  [24] ARABIC LETTER ALEF WITH ATTACHED FATHA..ARABIC BASELINE ROUND DOT
    Prop<unicode::Script> { { 0x0888, 0x0888 }, unicode::Script::Arabic }, // Sk       ARABIC RAISED ROUND DOT
    Prop<unicode::Script> { { 0x0889, 0x088F }, unicode::Script::Arabic }, // Lo   [7] ARABIC LETTER NOON WITH INVERTED SMALL V..ARABIC LETTER NOON WITH RING ABOVE
    Prop<unicode::Script> { { 0x0890, 0x0891 }, unicode::Script::Arabic }, // Cf   [2] ARABIC POUND MARK ABOVE..ARABIC PIASTRE MARK ABOVE
    Prop<unicode::Script> { { 0x0897, 0x089F }, unicode::Script::Arabic }, // Mn   [9] ARABIC PEPET..ARABIC HALF MADDA OVER MADDA
    Prop<unicode::Script> { { 0x08A0, 0x08C8 }, unicode::Script::Arabic }, // Lo  [41] ARABIC LETTER BEH WITH SMALL V BELOW..ARABIC LETTER GRAF
    Prop<unicode::Script> { { 0x08C9, 0x08C9 }, unicode::Script::Arabic }, // Lm       ARABIC SMALL FARSI YEH
    Prop<unicode::Script> { { 0x08CA, 0x08E1 }, unicode::Script::Arabic }, // Mn  [24] ARABIC SMALL HIGH FARSI YEH..ARABIC SMALL HIGH SIGN SAFHA
    Prop<unicode::Script> { { 0x08E2, 0x08E2 }, unicode::Script::Common }, // Cf       ARABIC DISPUTED END OF AYAH
    Prop<unicode::Script> { { 0x08E3, 0x08FF }, unicode::Script::Arabic }, // Mn  [29] ARABIC TURNED DAMMA BELOW..ARABIC MARK SIDEWAYS NOON GHUNNA
    Prop<unicode::Script> { { 0x0900, 0x0902 }, unicode::Script::Devanagari }, // Mn   [3] DEVANAGARI SIGN INVERTED CANDRABINDU..DEVANAGARI SIGN ANUSVARA
    Prop<unicode::Script> { { 0x0903, 0x0903 }, unicode::Script::Devanagari }, // Mc       DEVANAGARI SIGN VISARGA
    Prop<unicode::Script> { { 0x0904, 0x0939 }, unicode::Script::Devanagari }, // Lo  [54] DEVANAGARI LETTER SHORT A..DEVANAGARI LETTER HA
    Prop<unicode::Script> { { 0x093A, 0x093A }, unicode::Script::Devanagari }, // Mn       DEVANAGARI VOWEL SIGN OE
    Prop<unicode::Script> { { 0x093B, 0x093B }, unicode::Script::Devanagari }, // Mc       DEVANAGARI VOWEL SIGN OOE
    Prop<unicode::Script> { { 0x093C, 0x093C }, unicode::Script::Devanagari }, // Mn       DEVANAGARI SIGN NUKTA
    Prop<unicode::Script> { { 0x093D, 0x093D }, unicode::Script::Devanagari }, // Lo       DEVANAGARI SIGN AVAGRAHA
    Prop<unicode::Script> { { 0x093E, 0x0940 }, unicode::Script::Devanagari }, // Mc   [3] DEVANAGARI VOWEL SIGN AA..DEVANAGARI VOWEL SIGN II
    Prop<unicode::Script> { { 0x0941, 0x0948 }, unicode::Script::Devanagari }, // Mn   [8] DEVANAGARI VOWEL SIGN U..DEVANAGARI VOWEL SIGN AI
    Prop<unicode::Script> { { 0x0949, 0x094C }, unicode::Script::Devanagari }, // Mc   [4] DEVANAGARI VOWEL SIGN CANDRA O..DEVANAGARI VOWEL SIGN AU
    Prop<unicode::Script> { { 0x094D, 0x094D }, unicode::Script::Devanagari }, // Mn       DEVANAGARI SIGN VIRAMA
    Prop<unicode::Script> { { 0x094E, 0x094F }, unicode::Script::Devanagari }, // Mc   [2] DEVANAGARI VOWEL SIGN PRISHTHAMATRA E..DEVANAGARI VOWEL SIGN AW
    Prop<unicode::Script> { { 0x0950, 0x0950 }, unicode::Script::Devanagari }, // Lo       DEVANAGARI OM
    Prop<unicode::Script> { { 0x0951, 0x0954 }, unicode::Script::Inherited }, // Mn   [4] DEVANAGARI STRESS SIGN UDATTA..DEVANAGARI ACUTE ACCENT
    Prop<unicode::Script> { { 0x0955, 0x0957 }, unicode::Script::Devanagari }, // Mn   [3] DEVANAGARI VOWEL SIGN CANDRA LONG E..DEVANAGARI VOWEL SIGN UUE
    Prop<unicode::Script> { { 0x0958, 0x0961 }, unicode::Script::Devanagari }, // Lo  [10] DEVANAGARI LETTER QA..DEVANAGARI LETTER VOCALIC LL
    Prop<unicode::Script> { { 0x0962, 0x0963 }, unicode::Script::Devanagari }, // Mn   [2] DEVANAGARI VOWEL SIGN VOCALIC L..DEVANAGARI VOWEL SIGN VOCALIC LL
    Prop<unicode::Script> { { 0x0964, 0x0965 }, unicode::Script::Common }, // Po   [2] DEVANAGARI DANDA..DEVANAGARI DOUBLE DANDA
    Prop<unicode::Script> { { 0x0966, 0x096F }, unicode::Script::Devanagari }, // Nd  [10] DEVANAGARI DIGIT ZERO..DEVANAGARI DIGIT NINE
    Prop<unicode::Script> { { 0x0970, 0x0970 }, unicode::Script::Devanagari }, // Po       DEVANAGARI ABBREVIATION SIGN
    Prop<unicode::Script> { { 0x0971, 0x0971 }, unicode::Script::Devanagari }, // Lm       DEVANAGARI SIGN HIGH SPACING DOT
    Prop<unicode::Script> { { 0x0972, 0x097F }, unicode::Script::Devanagari }, // Lo  [14] DEVANAGARI LETTER CANDRA A..DEVANAGARI LETTER BBA
    Prop<unicode::Script> { { 0x0980, 0x0980 }, unicode::Script::Bengali }, // Lo       BENGALI ANJI
    Prop<unicode::Script> { { 0x0981, 0x0981 }, unicode::Script::Bengali }, // Mn       BENGALI SIGN CANDRABINDU
    Prop<unicode::Script> { { 0x0982, 0x0983 }, unicode::Script::Bengali }, // Mc   [2] BENGALI SIGN ANUSVARA..BENGALI SIGN VISARGA
    Prop<unicode::Script> { { 0x0985, 0x098C }, unicode::Script::Bengali }, // Lo   [8] BENGALI LETTER A..BENGALI LETTER VOCALIC L
    Prop<unicode::Script> { { 0x098F, 0x0990 }, unicode::Script::Bengali }, // Lo   [2] BENGALI LETTER E..BENGALI LETTER AI
    Prop<unicode::Script> { { 0x0993, 0x09A8 }, unicode::Script::Bengali }, // Lo  [22] BENGALI LETTER O..BENGALI LETTER NA
    Prop<unicode::Script> { { 0x09AA, 0x09B0 }, unicode::Script::Bengali }, // Lo   [7] BENGALI LETTER PA..BENGALI LETTER RA
    Prop<unicode::Script> { { 0x09B2, 0x09B2 }, unicode::Script::Bengali }, // Lo       BENGALI LETTER LA
    Prop<unicode::Script> { { 0x09B6, 0x09B9 }, unicode::Script::Bengali }, // Lo   [4] BENGALI LETTER SHA..BENGALI LETTER HA
    Prop<unicode::Script> { { 0x09BC, 0x09BC }, unicode::Script::Bengali }, // Mn       BENGALI SIGN NUKTA
    Prop<unicode::Script> { { 0x09BD, 0x09BD }, unicode::Script::Bengali }, // Lo       BENGALI SIGN AVAGRAHA
    Prop<unicode::Script> { { 0x09BE, 0x09C0 }, unicode::Script::Bengali }, // Mc   [3] BENGALI VOWEL SIGN AA..BENGALI VOWEL SIGN II
    Prop<unicode::Script> { { 0x09C1, 0x09C4 }, unicode::Script::Bengali }, // Mn   [4] BENGALI VOWEL SIGN U..BENGALI VOWEL SIGN VOCALIC RR
    Prop<unicode::Script> { { 0x09C7, 0x09C8 }, unicode::Script::Bengali }, // Mc   [2] BENGALI VOWEL SIGN E..BENGALI VOWEL SIGN AI
    Prop<unicode::Script> { { 0x09CB, 0x09CC }, unicode::Script::Bengali }, // Mc   [2] BENGALI VOWEL SIGN O..BENGALI VOWEL SIGN AU
    Prop<unicode::Script> { { 0x09CD, 0x09CD }, unicode::Script::Bengali }, // Mn       BENGALI SIGN VIRAMA
    Prop<unicode::Script> { { 0x09CE, 0x09CE }, unicode::Script::Bengali }, // Lo       BENGALI LETTER KHANDA TA
    Prop<unicode::Script> { { 0x09D7, 0x09D7 }, unicode::Script::Bengali }, // Mc       BENGALI AU LENGTH MARK
    Prop<unicode::Script> { { 0x09DC, 0x09DD }, unicode::Script::Bengali }, // Lo   [2] BENGALI LETTER RRA..BENGALI LETTER RHA
    Prop<unicode::Script> { { 0x09DF, 0x09E1 }, unicode::Script::Bengali }, // Lo   [3] BENGALI LETTER YYA..BENGALI LETTER VOCALIC LL
    Prop<unicode::Script> { { 0x09E2, 0x09E3 }, unicode::Script::Bengali }, // Mn   [2] BENGALI VOWEL SIGN VOCALIC L..BENGALI VOWEL SIGN VOCALIC LL
    Prop<unicode::Script> { { 0x09E6, 0x09EF }, unicode::Script::Bengali }, // Nd  [10] BENGALI DIGIT ZERO..BENGALI DIGIT NINE
    Prop<unicode::Script> { { 0x09F0, 0x09F1 }, unicode::Script::Bengali }, // Lo   [2] BENGALI LETTER RA WITH MIDDLE DIAGONAL..BENGALI LETTER RA WITH LOWER DIAGONAL
    Prop<unicode::Script> { { 0x09F2, 0x09F3 }, unicode::Script::Bengali }, // Sc   [2] BENGALI RUPEE MARK..BENGALI RUPEE SIGN
    Prop<unicode::Script> { { 0x09F4, 0x09F9 }, unicode::Script::Bengali }, // No   [6] BENGALI CURRENCY NUMERATOR ONE..BENGALI CURRENCY DENOMINATOR SIXTEEN
    Prop<unicode::Script> { { 0x09FA, 0x09FA }, unicode::Script::Bengali }, // So       BENGALI ISSHAR
    Prop<unicode::Script> { { 0x09FB, 0x09FB }, unicode::Script::Bengali }, // Sc       BENGALI GANDA MARK
    Prop<unicode::Script> { { 0x09FC, 0x09FC }, unicode::Script::Bengali }, // Lo       BENGALI LETTER VEDIC ANUSVARA
    Prop<unicode::Script> { { 0x09FD, 0x09FD }, unicode::Script::Bengali }, // Po       BENGALI ABBREVIATION SIGN
    Prop<unicode::Script> { { 0x09FE, 0x09FE }, unicode::Script::Bengali }, // Mn       BENGALI SANDHI MARK
    Prop<unicode::Script> { { 0x0A01, 0x0A02 }, unicode::Script::Gurmukhi }, // Mn   [2] GURMUKHI SIGN ADAK BINDI..GURMUKHI SIGN BINDI
    Prop<unicode::Script> { { 0x0A03, 0x0A03 }, unicode::Script::Gurmukhi }, // Mc       GURMUKHI SIGN VISARGA
    Prop<unicode::Script> { { 0x0A05, 0x0A0A }, unicode::Script::Gurmukhi }, // Lo   [6] GURMUKHI LETTER A..GURMUKHI LETTER UU
    Prop<unicode::Script> { { 0x0A0F, 0x0A10 }, unicode::Script::Gurmukhi }, // Lo   [2] GURMUKHI LETTER EE..GURMUKHI LETTER AI
    Prop<unicode::Script> { { 0x0A13, 0x0A28 }, unicode::Script::Gurmukhi }, // Lo  [22] GURMUKHI LETTER OO..GURMUKHI LETTER NA
    Prop<unicode::Script> { { 0x0A2A, 0x0A30 }, unicode::Script::Gurmukhi }, // Lo   [7] GURMUKHI LETTER PA..GURMUKHI LETTER RA
    Prop<unicode::Script> { { 0x0A32, 0x0A33 }, unicode::Script::Gurmukhi }, // Lo   [2] GURMUKHI LETTER LA..GURMUKHI LETTER LLA
    Prop<unicode::Script> { { 0x0A35, 0x0A36 }, unicode::Script::Gurmukhi }, // Lo   [2] GURMUKHI LETTER VA..GURMUKHI LETTER SHA
    Prop<unicode::Script> { { 0x0A38, 0x0A39 }, unicode::Script::Gurmukhi }, // Lo   [2] GURMUKHI LETTER SA..GURMUKHI LETTER HA
    Prop<unicode::Script> { { 0x0A3C, 0x0A3C }, unicode::Script::Gurmukhi }, // Mn       GURMUKHI SIGN NUKTA
    Prop<unicode::Script> { { 0x0A3E, 0x0A40 }, unicode::Script::Gurmukhi }, // Mc   [3] GURMUKHI VOWEL SIGN AA..GURMUKHI VOWEL SIGN II
    Prop<unicode::Script> { { 0x0A41, 0x0A42 }, unicode::Script::Gurmukhi }, // Mn   [2] GURMUKHI VOWEL SIGN U..GURMUKHI VOWEL SIGN UU
    Prop<unicode::Script> { { 0x0A47, 0x0A48 }, unicode::Script::Gurmukhi }, // Mn   [2] GURMUKHI VOWEL SIGN EE..GURMUKHI VOWEL SIGN AI
    Prop<unicode::Script> { { 0x0A4B, 0x0A4D }, unicode::Script::Gurmukhi }, // Mn   [3] GURMUKHI VOWEL SIGN OO..GURMUKHI SIGN VIRAMA
    Prop<unicode::Script> { { 0x0A51, 0x0A51 }, unicode::Script::Gurmukhi }, // Mn       GURMUKHI SIGN UDAAT
    Prop<unicode::Script> { { 0x0A59, 0x0A5C }, unicode::Script::Gurmukhi }, // Lo   [4] GURMUKHI LETTER KHHA..GURMUKHI LETTER RRA
    Prop<unicode::Script> { { 0x0A5E, 0x0A5E }, unicode::Script::Gurmukhi }, // Lo       GURMUKHI LETTER FA
    Prop<unicode::Script> { { 0x0A66, 0x0A6F }, unicode::Script::Gurmukhi }, // Nd  [10] GURMUKHI DIGIT ZERO..GURMUKHI DIGIT NINE
    Prop<unicode::Script> { { 0x0A70, 0x0A71 }, unicode::Script::Gurmukhi }, // Mn   [2] GURMUKHI TIPPI..GURMUKHI ADDAK
    Prop<unicode::Script> { { 0x0A72, 0x0A74 }, unicode::Script::Gurmukhi }, // Lo   [3] GURMUKHI IRI..GURMUKHI EK ONKAR
    Prop<unicode::Script> { { 0x0A75, 0x0A75 }, unicode::Script::Gurmukhi }, // Mn       GURMUKHI SIGN YAKASH
    Prop<unicode::Script> { { 0x0A76, 0x0A76 }, unicode::Script::Gurmukhi }, // Po       GURMUKHI ABBREVIATION SIGN
    Prop<unicode::Script> { { 0x0A81, 0x0A82 }, unicode::Script::Gujarati }, // Mn   [2] GUJARATI SIGN CANDRABINDU..GUJARATI SIGN ANUSVARA
    Prop<unicode::Script> { { 0x0A83, 0x0A83 }, unicode::Script::Gujarati }, // Mc       GUJARATI SIGN VISARGA
    Prop<unicode::Script> { { 0x0A85, 0x0A8D }, unicode::Script::Gujarati }, // Lo   [9] GUJARATI LETTER A..GUJARATI VOWEL CANDRA E
    Prop<unicode::Script> { { 0x0A8F, 0x0A91 }, unicode::Script::Gujarati }, // Lo   [3] GUJARATI LETTER E..GUJARATI VOWEL CANDRA O
    Prop<unicode::Script> { { 0x0A93, 0x0AA8 }, unicode::Script::Gujarati }, // Lo  [22] GUJARATI LETTER O..GUJARATI LETTER NA
    Prop<unicode::Script> { { 0x0AAA, 0x0AB0 }, unicode::Script::Gujarati }, // Lo   [7] GUJARATI LETTER PA..GUJARATI LETTER RA
    Prop<unicode::Script> { { 0x0AB2, 0x0AB3 }, unicode::Script::Gujarati }, // Lo   [2] GUJARATI LETTER LA..GUJARATI LETTER LLA
    Prop<unicode::Script> { { 0x0AB5, 0x0AB9 }, unicode::Script::Gujarati }, // Lo   [5] GUJARATI LETTER VA..GUJARATI LETTER HA
    Prop<unicode::Script> { { 0x0ABC, 0x0ABC }, unicode::Script::Gujarati }, // Mn       GUJARATI SIGN NUKTA
    Prop<unicode::Script> { { 0x0ABD, 0x0ABD }, unicode::Script::Gujarati }, // Lo       GUJARATI SIGN AVAGRAHA
    Prop<unicode::Script> { { 0x0ABE, 0x0AC0 }, unicode::Script::Gujarati }, // Mc   [3] GUJARATI VOWEL SIGN AA..GUJARATI VOWEL SIGN II
    Prop<unicode::Script> { { 0x0AC1, 0x0AC5 }, unicode::Script::Gujarati }, // Mn   [5] GUJARATI VOWEL SIGN U..GUJARATI VOWEL SIGN CANDRA E
    Prop<unicode::Script> { { 0x0AC7, 0x0AC8 }, unicode::Script::Gujarati }, // Mn   [2] GUJARATI VOWEL SIGN E..GUJARATI VOWEL SIGN AI
    Prop<unicode::Script> { { 0x0AC9, 0x0AC9 }, unicode::Script::Gujarati }, // Mc       GUJARATI VOWEL SIGN CANDRA O
    Prop<unicode::Script> { { 0x0ACB, 0x0ACC }, unicode::Script::Gujarati }, // Mc   [2] GUJARATI VOWEL SIGN O..GUJARATI VOWEL SIGN AU
    Prop<unicode::Script> { { 0x0ACD, 0x0ACD }, unicode::Script::Gujarati }, // Mn       GUJARATI SIGN VIRAMA
    Prop<unicode::Script> { { 0x0AD0, 0x0AD0 }, unicode::Script::Gujarati }, // Lo       GUJARATI OM
    Prop<unicode::Script> { { 0x0AE0, 0x0AE1 }, unicode::Script::Gujarati }, // Lo   [2] GUJARATI LETTER VOCALIC RR..GUJARATI LETTER VOCALIC LL
    Prop<unicode::Script> { { 0x0AE2, 0x0AE3 }, unicode::Script::Gujarati }, // Mn   [2] GUJARATI VOWEL SIGN VOCALIC L..GUJARATI VOWEL SIGN VOCALIC LL
    Prop<unicode::Script> { { 0x0AE6, 0x0AEF }, unicode::Script::Gujarati }, // Nd  [10] GUJARATI DIGIT ZERO..GUJARATI DIGIT NINE
    Prop<unicode::Script> { { 0x0AF0, 0x0AF0 }, unicode::Script::Gujarati }, // Po       GUJARATI ABBREVIATION SIGN
    Prop<unicode::Script> { { 0x0AF1, 0x0AF1 }, unicode::Script::Gujarati }, // Sc       GUJARATI RUPEE SIGN
    Prop<unicode::Script> { { 0x0AF9, 0x0AF9 }, unicode::Script::Gujarati }, // Lo       GUJARATI LETTER ZHA
    Prop<unicode::Script> { { 0x0AFA, 0x0AFF }, unicode::Script::Gujarati }, // Mn   [6] GUJARATI SIGN SUKUN..GUJARATI SIGN TWO-CIRCLE NUKTA ABOVE
    Prop<unicode::Script> { { 0x0B01, 0x0B01 }, unicode::Script::Oriya }, // Mn       ORIYA SIGN CANDRABINDU
    Prop<unicode::Script> { { 0x0B02, 0x0B03 }, unicode::Script::Oriya }, // Mc   [2] ORIYA SIGN ANUSVARA..ORIYA SIGN VISARGA
    Prop<unicode::Script> { { 0x0B05, 0x0B0C }, unicode::Script::Oriya }, // Lo   [8] ORIYA LETTER A..ORIYA LETTER VOCALIC L
    Prop<unicode::Script> { { 0x0B0F, 0x0B10 }, unicode::Script::Oriya }, // Lo   [2] ORIYA LETTER E..ORIYA LETTER AI
    Prop<unicode::Script> { { 0x0B13, 0x0B28 }, unicode::Script::Oriya }, // Lo  [22] ORIYA LETTER O..ORIYA LETTER NA
    Prop<unicode::Script> { { 0x0B2A, 0x0B30 }, unicode::Script::Oriya }, // Lo   [7] ORIYA LETTER PA..ORIYA LETTER RA
    Prop<unicode::Script> { { 0x0B32, 0x0B33 }, unicode::Script::Oriya }, // Lo   [2] ORIYA LETTER LA..ORIYA LETTER LLA
    Prop<unicode::Script> { { 0x0B35, 0x0B39 }, unicode::Script::Oriya }, // Lo   [5] ORIYA LETTER VA..ORIYA LETTER HA
    Prop<unicode::Script> { { 0x0B3C, 0x0B3C }, unicode::Script::Oriya }, // Mn       ORIYA SIGN NUKTA
    Prop<unicode::Script> { { 0x0B3D, 0x0B3D }, unicode::Script::Oriya }, // Lo       ORIYA SIGN AVAGRAHA
    Prop<unicode::Script> { { 0x0B3E, 0x0B3E }, unicode::Script::Oriya }, // Mc       ORIYA VOWEL SIGN AA
    Prop<unicode::Script> { { 0x0B3F, 0x0B3F }, unicode::Script::Oriya }, // Mn       ORIYA VOWEL SIGN I
    Prop<unicode::Script> { { 0x0B40, 0x0B40 }, unicode::Script::Oriya }, // Mc       ORIYA VOWEL SIGN II
    Prop<unicode::Script> { { 0x0B41, 0x0B44 }, unicode::Script::Oriya }, // Mn   [4] ORIYA VOWEL SIGN U..ORIYA VOWEL SIGN VOCALIC RR
    Prop<unicode::Script> { { 0x0B47, 0x0B48 }, unicode::Script::Oriya }, // Mc   [2] ORIYA VOWEL SIGN E..ORIYA VOWEL SIGN AI
    Prop<unicode::Script> { { 0x0B4B, 0x0B4C }, unicode::Script::Oriya }, // Mc   [2] ORIYA VOWEL SIGN O..ORIYA VOWEL SIGN AU
    Prop<unicode::Script> { { 0x0B4D, 0x0B4D }, unicode::Script::Oriya }, // Mn       ORIYA SIGN VIRAMA
    Prop<unicode::Script> { { 0x0B55, 0x0B56 }, unicode::Script::Oriya }, // Mn   [2] ORIYA SIGN OVERLINE..ORIYA AI LENGTH MARK
    Prop<unicode::Script> { { 0x0B57, 0x0B57 }, unicode::Script::Oriya }, // Mc       ORIYA AU LENGTH MARK
    Prop<unicode::Script> { { 0x0B5C, 0x0B5D }, unicode::Script::Oriya }, // Lo   [2] ORIYA LETTER RRA..ORIYA LETTER RHA
    Prop<unicode::Script> { { 0x0B5F, 0x0B61 }, unicode::Script::Oriya }, // Lo   [3] ORIYA LETTER YYA..ORIYA LETTER VOCALIC LL
    Prop<unicode::Script> { { 0x0B62, 0x0B63 }, unicode::Script::Oriya }, // Mn   [2] ORIYA VOWEL SIGN VOCALIC L..ORIYA VOWEL SIGN VOCALIC LL
    Prop<unicode::Script> { { 0x0B66, 0x0B6F }, unicode::Script::Oriya }, // Nd  [10] ORIYA DIGIT ZERO..ORIYA DIGIT NINE
    Prop<unicode::Script> { { 0x0B70, 0x0B70 }, unicode::Script::Oriya }, // So       ORIYA ISSHAR
    Prop<unicode::Script> { { 0x0B71, 0x0B71 }, unicode::Script::Oriya }, // Lo       ORIYA LETTER WA
    Prop<unicode::Script> { { 0x0B72, 0x0B77 }, unicode::Script::Oriya }, // No   [6] ORIYA FRACTION ONE QUARTER..ORIYA FRACTION THREE SIXTEENTHS
    Prop<unicode::Script> { { 0x0B82, 0x0B82 }, unicode::Script::Tamil }, // Mn       TAMIL SIGN ANUSVARA
    Prop<unicode::Script> { { 0x0B83, 0x0B83 }, unicode::Script::Tamil }, // Lo       TAMIL SIGN VISARGA
    Prop<unicode::Script> { { 0x0B85, 0x0B8A }, unicode::Script::Tamil }, // Lo   [6] TAMIL LETTER A..TAMIL LETTER UU
    Prop<unicode::Script> { { 0x0B8E, 0x0B90 }, unicode::Script::Tamil }, // Lo   [3] TAMIL LETTER E..TAMIL LETTER AI
    Prop<unicode::Script> { { 0x0B92, 0x0B95 }, unicode::Script::Tamil }, // Lo   [4] TAMIL LETTER O..TAMIL LETTER KA
    Prop<unicode::Script> { { 0x0B99, 0x0B9A }, unicode::Script::Tamil }, // Lo   [2] TAMIL LETTER NGA..TAMIL LETTER CA
    Prop<unicode::Script> { { 0x0B9C, 0x0B9C }, unicode::Script::Tamil }, // Lo       TAMIL LETTER JA
    Prop<unicode::Script> { { 0x0B9E, 0x0B9F }, unicode::Script::Tamil }, // Lo   [2] TAMIL LETTER NYA..TAMIL LETTER TTA
    Prop<unicode::Script> { { 0x0BA3, 0x0BA4 }, unicode::Script::Tamil }, // Lo   [2] TAMIL LETTER NNA..TAMIL LETTER TA
    Prop<unicode::Script> { { 0x0BA8, 0x0BAA }, unicode::Script::Tamil }, // Lo   [3] TAMIL LETTER NA..TAMIL LETTER PA
    Prop<unicode::Script> { { 0x0BAE, 0x0BB9 }, unicode::Script::Tamil }, // Lo  [12] TAMIL LETTER MA..TAMIL LETTER HA
    Prop<unicode::Script> { { 0x0BBE, 0x0BBF }, unicode::Script::Tamil }, // Mc   [2] TAMIL VOWEL SIGN AA..TAMIL VOWEL SIGN I
    Prop<unicode::Script> { { 0x0BC0, 0x0BC0 }, unicode::Script::Tamil }, // Mn       TAMIL VOWEL SIGN II
    Prop<unicode::Script> { { 0x0BC1, 0x0BC2 }, unicode::Script::Tamil }, // Mc   [2] TAMIL VOWEL SIGN U..TAMIL VOWEL SIGN UU
    Prop<unicode::Script> { { 0x0BC6, 0x0BC8 }, unicode::Script::Tamil }, // Mc   [3] TAMIL VOWEL SIGN E..TAMIL VOWEL SIGN AI
    Prop<unicode::Script> { { 0x0BCA, 0x0BCC }, unicode::Script::Tamil }, // Mc   [3] TAMIL VOWEL SIGN O..TAMIL VOWEL SIGN AU
    Prop<unicode::Script> { { 0x0BCD, 0x0BCD }, unicode::Script::Tamil }, // Mn       TAMIL SIGN VIRAMA
    Prop<unicode::Script> { { 0x0BD0, 0x0BD0 }, unicode::Script::Tamil }, // Lo       TAMIL OM
    Prop<unicode::Script> { { 0x0BD7, 0x0BD7 }, unicode::Script::Tamil }, // Mc       TAMIL AU LENGTH MARK
    Prop<unicode::Script> { { 0x0BE6, 0x0BEF }, unicode::Script::Tamil }, // Nd  [10] TAMIL DIGIT ZERO..TAMIL DIGIT NINE
    Prop<unicode::Script> { { 0x0BF0, 0x0BF2 }, unicode::Script::Tamil }, // No   [3] TAMIL NUMBER TEN..TAMIL NUMBER ONE THOUSAND
    Prop<unicode::Script> { { 0x0BF3, 0x0BF8 }, unicode::Script::Tamil }, // So   [6] TAMIL DAY SIGN..TAMIL AS ABOVE SIGN
    Prop<unicode::Script> { { 0x0BF9, 0x0BF9 }, unicode::Script::Tamil }, // Sc       TAMIL RUPEE SIGN
    Prop<unicode::Script> { { 0x0BFA, 0x0BFA }, unicode::Script::Tamil }, // So       TAMIL NUMBER SIGN
    Prop<unicode::Script> { { 0x0C00, 0x0C00 }, unicode::Script::Telugu }, // Mn       TELUGU SIGN COMBINING CANDRABINDU ABOVE
    Prop<unicode::Script> { { 0x0C01, 0x0C03 }, unicode::Script::Telugu }, // Mc   [3] TELUGU SIGN CANDRABINDU..TELUGU SIGN VISARGA
    Prop<unicode::Script> { { 0x0C04, 0x0C04 }, unicode::Script::Telugu }, // Mn       TELUGU SIGN COMBINING ANUSVARA ABOVE
    Prop<unicode::Script> { { 0x0C05, 0x0C0C }, unicode::Script::Telugu }, // Lo   [8] TELUGU LETTER A..TELUGU LETTER VOCALIC L
    Prop<unicode::Script> { { 0x0C0E, 0x0C10 }, unicode::Script::Telugu }, // Lo   [3] TELUGU LETTER E..TELUGU LETTER AI
    Prop<unicode::Script> { { 0x0C12, 0x0C28 }, unicode::Script::Telugu }, // Lo  [23] TELUGU LETTER O..TELUGU LETTER NA
    Prop<unicode::Script> { { 0x0C2A, 0x0C39 }, unicode::Script::Telugu }, // Lo  [16] TELUGU LETTER PA..TELUGU LETTER HA
    Prop<unicode::Script> { { 0x0C3C, 0x0C3C }, unicode::Script::Telugu }, // Mn       TELUGU SIGN NUKTA
    Prop<unicode::Script> { { 0x0C3D, 0x0C3D }, unicode::Script::Telugu }, // Lo       TELUGU SIGN AVAGRAHA
    Prop<unicode::Script> { { 0x0C3E, 0x0C40 }, unicode::Script::Telugu }, // Mn   [3] TELUGU VOWEL SIGN AA..TELUGU VOWEL SIGN II
    Prop<unicode::Script> { { 0x0C41, 0x0C44 }, unicode::Script::Telugu }, // Mc   [4] TELUGU VOWEL SIGN U..TELUGU VOWEL SIGN VOCALIC RR
    Prop<unicode::Script> { { 0x0C46, 0x0C48 }, unicode::Script::Telugu }, // Mn   [3] TELUGU VOWEL SIGN E..TELUGU VOWEL SIGN AI
    Prop<unicode::Script> { { 0x0C4A, 0x0C4D }, unicode::Script::Telugu }, // Mn   [4] TELUGU VOWEL SIGN O..TELUGU SIGN VIRAMA
    Prop<unicode::Script> { { 0x0C55, 0x0C56 }, unicode::Script::Telugu }, // Mn   [2] TELUGU LENGTH MARK..TELUGU AI LENGTH MARK
    Prop<unicode::Script> { { 0x0C58, 0x0C5A }, unicode::Script::Telugu }, // Lo   [3] TELUGU LETTER TSA..TELUGU LETTER RRRA
    Prop<unicode::Script> { { 0x0C5C, 0x0C5D }, unicode::Script::Telugu }, // Lo   [2] TELUGU ARCHAIC SHRII..TELUGU LETTER NAKAARA POLLU
    Prop<unicode::Script> { { 0x0C60, 0x0C61 }, unicode::Script::Telugu }, // Lo   [2] TELUGU LETTER VOCALIC RR..TELUGU LETTER VOCALIC LL
    Prop<unicode::Script> { { 0x0C62, 0x0C63 }, unicode::Script::Telugu }, // Mn   [2] TELUGU VOWEL SIGN VOCALIC L..TELUGU VOWEL SIGN VOCALIC LL
    Prop<unicode::Script> { { 0x0C66, 0x0C6F }, unicode::Script::Telugu }, // Nd  [10] TELUGU DIGIT ZERO..TELUGU DIGIT NINE
    Prop<unicode::Script> { { 0x0C77, 0x0C77 }, unicode::Script::Telugu }, // Po       TELUGU SIGN SIDDHAM
    Prop<unicode::Script> { { 0x0C78, 0x0C7E }, unicode::Script::Telugu }, // No   [7] TELUGU FRACTION DIGIT ZERO FOR ODD POWERS OF FOUR..TELUGU FRACTION DIGIT THREE FOR EVEN POWERS OF FOUR
    Prop<unicode::Script> { { 0x0C7F, 0x0C7F }, unicode::Script::Telugu }, // So       TELUGU SIGN TUUMU
    Prop<unicode::Script> { { 0x0C80, 0x0C80 }, unicode::Script::Kannada }, // Lo       KANNADA SIGN SPACING CANDRABINDU
    Prop<unicode::Script> { { 0x0C81, 0x0C81 }, unicode::Script::Kannada }, // Mn       KANNADA SIGN CANDRABINDU
    Prop<unicode::Script> { { 0x0C82, 0x0C83 }, unicode::Script::Kannada }, // Mc   [2] KANNADA SIGN ANUSVARA..KANNADA SIGN VISARGA
    Prop<unicode::Script> { { 0x0C84, 0x0C84 }, unicode::Script::Kannada }, // Po       KANNADA SIGN SIDDHAM
    Prop<unicode::Script> { { 0x0C85, 0x0C8C }, unicode::Script::Kannada }, // Lo   [8] KANNADA LETTER A..KANNADA LETTER VOCALIC L
    Prop<unicode::Script> { { 0x0C8E, 0x0C90 }, unicode::Script::Kannada }, // Lo   [3] KANNADA LETTER E..KANNADA LETTER AI
    Prop<unicode::Script> { { 0x0C92, 0x0CA8 }, unicode::Script::Kannada }, // Lo  [23] KANNADA LETTER O..KANNADA LETTER NA
    Prop<unicode::Script> { { 0x0CAA, 0x0CB3 }, unicode::Script::Kannada }, // Lo  [10] KANNADA LETTER PA..KANNADA LETTER LLA
    Prop<unicode::Script> { { 0x0CB5, 0x0CB9 }, unicode::Script::Kannada }, // Lo   [5] KANNADA LETTER VA..KANNADA LETTER HA
    Prop<unicode::Script> { { 0x0CBC, 0x0CBC }, unicode::Script::Kannada }, // Mn       KANNADA SIGN NUKTA
    Prop<unicode::Script> { { 0x0CBD, 0x0CBD }, unicode::Script::Kannada }, // Lo       KANNADA SIGN AVAGRAHA
    Prop<unicode::Script> { { 0x0CBE, 0x0CBE }, unicode::Script::Kannada }, // Mc       KANNADA VOWEL SIGN AA
    Prop<unicode::Script> { { 0x0CBF, 0x0CBF }, unicode::Script::Kannada }, // Mn       KANNADA VOWEL SIGN I
    Prop<unicode::Script> { { 0x0CC0, 0x0CC4 }, unicode::Script::Kannada }, // Mc   [5] KANNADA VOWEL SIGN II..KANNADA VOWEL SIGN VOCALIC RR
    Prop<unicode::Script> { { 0x0CC6, 0x0CC6 }, unicode::Script::Kannada }, // Mn       KANNADA VOWEL SIGN E
    Prop<unicode::Script> { { 0x0CC7, 0x0CC8 }, unicode::Script::Kannada }, // Mc   [2] KANNADA VOWEL SIGN EE..KANNADA VOWEL SIGN AI
    Prop<unicode::Script> { { 0x0CCA, 0x0CCB }, unicode::Script::Kannada }, // Mc   [2] KANNADA VOWEL SIGN O..KANNADA VOWEL SIGN OO
    Prop<unicode::Script> { { 0x0CCC, 0x0CCD }, unicode::Script::Kannada }, // Mn   [2] KANNADA VOWEL SIGN AU..KANNADA SIGN VIRAMA
    Prop<unicode::Script> { { 0x0CD5, 0x0CD6 }, unicode::Script::Kannada }, // Mc   [2] KANNADA LENGTH MARK..KANNADA AI LENGTH MARK
    Prop<unicode::Script> { { 0x0CDC, 0x0CDE }, unicode::Script::Kannada }, // Lo   [3] KANNADA ARCHAIC SHRII..KANNADA LETTER FA
    Prop<unicode::Script> { { 0x0CE0, 0x0CE1 }, unicode::Script::Kannada }, // Lo   [2] KANNADA LETTER VOCALIC RR..KANNADA LETTER VOCALIC LL
    Prop<unicode::Script> { { 0x0CE2, 0x0CE3 }, unicode::Script::Kannada }, // Mn   [2] KANNADA VOWEL SIGN VOCALIC L..KANNADA VOWEL SIGN VOCALIC LL
    Prop<unicode::Script> { { 0x0CE6, 0x0CEF }, unicode::Script::Kannada }, // Nd  [10] KANNADA DIGIT ZERO..KANNADA DIGIT NINE
    Prop<unicode::Script> { { 0x0CF1, 0x0CF2 }, unicode::Script::Kannada }, // Lo   [2] KANNADA SIGN JIHVAMULIYA..KANNADA SIGN UPADHMANIYA
    Prop<unicode::Script> { { 0x0CF3, 0x0CF3 }, unicode::Script::Kannada }, // Mc       KANNADA SIGN COMBINING ANUSVARA ABOVE RIGHT
    Prop<unicode::Script> { { 0x0D00, 0x0D01 }, unicode::Script::Malayalam }, // Mn   [2] MALAYALAM SIGN COMBINING ANUSVARA ABOVE..MALAYALAM SIGN CANDRABINDU
    Prop<unicode::Script> { { 0x0D02, 0x0D03 }, unicode::Script::Malayalam }, // Mc   [2] MALAYALAM SIGN ANUSVARA..MALAYALAM SIGN VISARGA
    Prop<unicode::Script> { { 0x0D04, 0x0D0C }, unicode::Script::Malayalam }, // Lo   [9] MALAYALAM LETTER VEDIC ANUSVARA..MALAYALAM LETTER VOCALIC L
    Prop<unicode::Script> { { 0x0D0E, 0x0D10 }, unicode::Script::Malayalam }, // Lo   [3] MALAYALAM LETTER E..MALAYALAM LETTER AI
    Prop<unicode::Script> { { 0x0D12, 0x0D3A }, unicode::Script::Malayalam }, // Lo  [41] MALAYALAM LETTER O..MALAYALAM LETTER TTTA
    Prop<unicode::Script> { { 0x0D3B, 0x0D3C }, unicode::Script::Malayalam }, // Mn   [2] MALAYALAM SIGN VERTICAL BAR VIRAMA..MALAYALAM SIGN CIRCULAR VIRAMA
    Prop<unicode::Script> { { 0x0D3D, 0x0D3D }, unicode::Script::Malayalam }, // Lo       MALAYALAM SIGN AVAGRAHA
    Prop<unicode::Script> { { 0x0D3E, 0x0D40 }, unicode::Script::Malayalam }, // Mc   [3] MALAYALAM VOWEL SIGN AA..MALAYALAM VOWEL SIGN II
    Prop<unicode::Script> { { 0x0D41, 0x0D44 }, unicode::Script::Malayalam }, // Mn   [4] MALAYALAM VOWEL SIGN U..MALAYALAM VOWEL SIGN VOCALIC RR
    Prop<unicode::Script> { { 0x0D46, 0x0D48 }, unicode::Script::Malayalam }, // Mc   [3] MALAYALAM VOWEL SIGN E..MALAYALAM VOWEL SIGN AI
    Prop<unicode::Script> { { 0x0D4A, 0x0D4C }, unicode::Script::Malayalam }, // Mc   [3] MALAYALAM VOWEL SIGN O..MALAYALAM VOWEL SIGN AU
    Prop<unicode::Script> { { 0x0D4D, 0x0D4D }, unicode::Script::Malayalam }, // Mn       MALAYALAM SIGN VIRAMA
    Prop<unicode::Script> { { 0x0D4E, 0x0D4E }, unicode::Script::Malayalam }, // Lo       MALAYALAM LETTER DOT REPH
    Prop<unicode::Script> { { 0x0D4F, 0x0D4F }, unicode::Script::Malayalam }, // So       MALAYALAM SIGN PARA
    Prop<unicode::Script> { { 0x0D54, 0x0D56 }, unicode::Script::Malayalam }, // Lo   [3] MALAYALAM LETTER CHILLU M..MALAYALAM LETTER CHILLU LLL
    Prop<unicode::Script> { { 0x0D57, 0x0D57 }, unicode::Script::Malayalam }, // Mc       MALAYALAM AU LENGTH MARK
    Prop<unicode::Script> { { 0x0D58, 0x0D5E }, unicode::Script::Malayalam }, // No   [7] MALAYALAM FRACTION ONE ONE-HUNDRED-AND-SIXTIETH..MALAYALAM FRACTION ONE FIFTH
    Prop<unicode::Script> { { 0x0D5F, 0x0D61 }, unicode::Script::Malayalam }, // Lo   [3] MALAYALAM LETTER ARCHAIC II..MALAYALAM LETTER VOCALIC LL
    Prop<unicode::Script> { { 0x0D62, 0x0D63 }, unicode::Script::Malayalam }, // Mn   [2] MALAYALAM VOWEL SIGN VOCALIC L..MALAYALAM VOWEL SIGN VOCALIC LL
    Prop<unicode::Script> { { 0x0D66, 0x0D6F }, unicode::Script::Malayalam }, // Nd  [10] MALAYALAM DIGIT ZERO..MALAYALAM DIGIT NINE
    Prop<unicode::Script> { { 0x0D70, 0x0D78 }, unicode::Script::Malayalam }, // No   [9] MALAYALAM NUMBER TEN..MALAYALAM FRACTION THREE SIXTEENTHS
    Prop<unicode::Script> { { 0x0D79, 0x0D79 }, unicode::Script::Malayalam }, // So       MALAYALAM DATE MARK
    Prop<unicode::Script> { { 0x0D7A, 0x0D7F }, unicode::Script::Malayalam }, // Lo   [6] MALAYALAM LETTER CHILLU NN..MALAYALAM LETTER CHILLU K
    Prop<unicode::Script> { { 0x0D81, 0x0D81 }, unicode::Script::Sinhala }, // Mn       SINHALA SIGN CANDRABINDU
    Prop<unicode::Script> { { 0x0D82, 0x0D83 }, unicode::Script::Sinhala }, // Mc   [2] SINHALA SIGN ANUSVARAYA..SINHALA SIGN VISARGAYA
    Prop<unicode::Script> { { 0x0D85, 0x0D96 }, unicode::Script::Sinhala }, // Lo  [18] SINHALA LETTER AYANNA..SINHALA LETTER AUYANNA
    Prop<unicode::Script> { { 0x0D9A, 0x0DB1 }, unicode::Script::Sinhala }, // Lo  [24] SINHALA LETTER ALPAPRAANA KAYANNA..SINHALA LETTER DANTAJA NAYANNA
    Prop<unicode::Script> { { 0x0DB3, 0x0DBB }, unicode::Script::Sinhala }, // Lo   [9] SINHALA LETTER SANYAKA DAYANNA..SINHALA LETTER RAYANNA
    Prop<unicode::Script> { { 0x0DBD, 0x0DBD }, unicode::Script::Sinhala }, // Lo       SINHALA LETTER DANTAJA LAYANNA
    Prop<unicode::Script> { { 0x0DC0, 0x0DC6 }, unicode::Script::Sinhala }, // Lo   [7] SINHALA LETTER VAYANNA..SINHALA LETTER FAYANNA
    Prop<unicode::Script> { { 0x0DCA, 0x0DCA }, unicode::Script::Sinhala }, // Mn       SINHALA SIGN AL-LAKUNA
    Prop<unicode::Script> { { 0x0DCF, 0x0DD1 }, unicode::Script::Sinhala }, // Mc   [3] SINHALA VOWEL SIGN AELA-PILLA..SINHALA VOWEL SIGN DIGA AEDA-PILLA
    Prop<unicode::Script> { { 0x0DD2, 0x0DD4 }, unicode::Script::Sinhala }, // Mn   [3] SINHALA VOWEL SIGN KETTI IS-PILLA..SINHALA VOWEL SIGN KETTI PAA-PILLA
    Prop<unicode::Script> { { 0x0DD6, 0x0DD6 }, unicode::Script::Sinhala }, // Mn       SINHALA VOWEL SIGN DIGA PAA-PILLA
    Prop<unicode::Script> { { 0x0DD8, 0x0DDF }, unicode::Script::Sinhala }, // Mc   [8] SINHALA VOWEL SIGN GAETTA-PILLA..SINHALA VOWEL SIGN GAYANUKITTA
    Prop<unicode::Script> { { 0x0DE6, 0x0DEF }, unicode::Script::Sinhala }, // Nd  [10] SINHALA LITH DIGIT ZERO..SINHALA LITH DIGIT NINE
    Prop<unicode::Script> { { 0x0DF2, 0x0DF3 }, unicode::Script::Sinhala }, // Mc   [2] SINHALA VOWEL SIGN DIGA GAETTA-PILLA..SINHALA VOWEL SIGN DIGA GAYANUKITTA
    Prop<unicode::Script> { { 0x0DF4, 0x0DF4 }, unicode::Script::Sinhala }, // Po       SINHALA PUNCTUATION KUNDDALIYA
    Prop<unicode::Script> { { 0x0E01, 0x0E30 }, unicode::Script::Thai }, // Lo  [48] THAI CHARACTER KO KAI..THAI CHARACTER SARA A
    Prop<unicode::Script> { { 0x0E31, 0x0E31 }, unicode::Script::Thai }, // Mn       THAI CHARACTER MAI HAN-AKAT
    Prop<unicode::Script> { { 0x0E32, 0x0E33 }, unicode::Script::Thai }, // Lo   [2] THAI CHARACTER SARA AA..THAI CHARACTER SARA AM
    Prop<unicode::Script> { { 0x0E34, 0x0E3A }, unicode::Script::Thai }, // Mn   [7] THAI CHARACTER SARA I..THAI CHARACTER PHINTHU
    Prop<unicode::Script> { { 0x0E3F, 0x0E3F }, unicode::Script::Common }, // Sc       THAI CURRENCY SYMBOL BAHT
    Prop<unicode::Script> { { 0x0E40, 0x0E45 }, unicode::Script::Thai }, // Lo   [6] THAI CHARACTER SARA E..THAI CHARACTER LAKKHANGYAO
    Prop<unicode::Script> { { 0x0E46, 0x0E46 }, unicode::Script::Thai }, // Lm       THAI CHARACTER MAIYAMOK
    Prop<unicode::Script> { { 0x0E47, 0x0E4E }, unicode::Script::Thai }, // Mn   [8] THAI CHARACTER MAITAIKHU..THAI CHARACTER YAMAKKAN
    Prop<unicode::Script> { { 0x0E4F, 0x0E4F }, unicode::Script::Thai }, // Po       THAI CHARACTER FONGMAN
    Prop<unicode::Script> { { 0x0E50, 0x0E59 }, unicode::Script::Thai }, // Nd  [10] THAI DIGIT ZERO..THAI DIGIT NINE
    Prop<unicode::Script> { { 0x0E5A, 0x0E5B }, unicode::Script::Thai }, // Po   [2] THAI CHARACTER ANGKHANKHU..THAI CHARACTER KHOMUT
    Prop<unicode::Script> { { 0x0E81, 0x0E82 }, unicode::Script::Lao }, // Lo   [2] LAO LETTER KO..LAO LETTER KHO SUNG
    Prop<unicode::Script> { { 0x0E84, 0x0E84 }, unicode::Script::Lao }, // Lo       LAO LETTER KHO TAM
    Prop<unicode::Script> { { 0x0E86, 0x0E8A }, unicode::Script::Lao }, // Lo   [5] LAO LETTER PALI GHA..LAO LETTER SO TAM
    Prop<unicode::Script> { { 0x0E8C, 0x0EA3 }, unicode::Script::Lao }, // Lo  [24] LAO LETTER PALI JHA..LAO LETTER LO LING
    Prop<unicode::Script> { { 0x0EA5, 0x0EA5 }, unicode::Script::Lao }, // Lo       LAO LETTER LO LOOT
    Prop<unicode::Script> { { 0x0EA7, 0x0EB0 }, unicode::Script::Lao }, // Lo  [10] LAO LETTER WO..LAO VOWEL SIGN A
    Prop<unicode::Script> { { 0x0EB1, 0x0EB1 }, unicode::Script::Lao }, // Mn       LAO VOWEL SIGN MAI KAN
    Prop<unicode::Script> { { 0x0EB2, 0x0EB3 }, unicode::Script::Lao }, // Lo   [2] LAO VOWEL SIGN AA..LAO VOWEL SIGN AM
    Prop<unicode::Script> { { 0x0EB4, 0x0EBC }, unicode::Script::Lao }, // Mn   [9] LAO VOWEL SIGN I..LAO SEMIVOWEL SIGN LO
    Prop<unicode::Script> { { 0x0EBD, 0x0EBD }, unicode::Script::Lao }, // Lo       LAO SEMIVOWEL SIGN NYO
    Prop<unicode::Script> { { 0x0EC0, 0x0EC4 }, unicode::Script::Lao }, // Lo   [5] LAO VOWEL SIGN E..LAO VOWEL SIGN AI
    Prop<unicode::Script> { { 0x0EC6, 0x0EC6 }, unicode::Script::Lao }, // Lm       LAO KO LA
    Prop<unicode::Script> { { 0x0EC8, 0x0ECE }, unicode::Script::Lao }, // Mn   [7] LAO TONE MAI EK..LAO YAMAKKAN
    Prop<unicode::Script> { { 0x0ED0, 0x0ED9 }, unicode::Script::Lao }, // Nd  [10] LAO DIGIT ZERO..LAO DIGIT NINE
    Prop<unicode::Script> { { 0x0EDC, 0x0EDF }, unicode::Script::Lao }, // Lo   [4] LAO HO NO..LAO LETTER KHMU NYO
    Prop<unicode::Script> { { 0x0F00, 0x0F00 }, unicode::Script::Tibetan }, // Lo       TIBETAN SYLLABLE OM
    Prop<unicode::Script> { { 0x0F01, 0x0F03 }, unicode::Script::Tibetan }, // So   [3] TIBETAN MARK GTER YIG MGO TRUNCATED A..TIBETAN MARK GTER YIG MGO -UM GTER TSHEG MA
    Prop<unicode::Script> { { 0x0F04, 0x0F12 }, unicode::Script::Tibetan }, // Po  [15] TIBETAN MARK INITIAL YIG MGO MDUN MA..TIBETAN MARK RGYA GRAM SHAD
    Prop<unicode::Script> { { 0x0F13, 0x0F13 }, unicode::Script::Tibetan }, // So       TIBETAN MARK CARET -DZUD RTAGS ME LONG CAN
    Prop<unicode::Script> { { 0x0F14, 0x0F14 }, unicode::Script::Tibetan }, // Po       TIBETAN MARK GTER TSHEG
    Prop<unicode::Script> { { 0x0F15, 0x0F17 }, unicode::Script::Tibetan }, // So   [3] TIBETAN LOGOTYPE SIGN CHAD RTAGS..TIBETAN ASTROLOGICAL SIGN SGRA GCAN -CHAR RTAGS
    Prop<unicode::Script> { { 0x0F18, 0x0F19 }, unicode::Script::Tibetan }, // Mn   [2] TIBETAN ASTROLOGICAL SIGN -KHYUD PA..TIBETAN ASTROLOGICAL SIGN SDONG TSHUGS
    Prop<unicode::Script> { { 0x0F1A, 0x0F1F }, unicode::Script::Tibetan }, // So   [6] TIBETAN SIGN RDEL DKAR GCIG..TIBETAN SIGN RDEL DKAR RDEL NAG
    Prop<unicode::Script> { { 0x0F20, 0x0F29 }, unicode::Script::Tibetan }, // Nd  [10] TIBETAN DIGIT ZERO..TIBETAN DIGIT NINE
    Prop<unicode::Script> { { 0x0F2A, 0x0F33 }, unicode::Script::Tibetan }, // No  [10] TIBETAN DIGIT HALF ONE..TIBETAN DIGIT HALF ZERO
    Prop<unicode::Script> { { 0x0F34, 0x0F34 }, unicode::Script::Tibetan }, // So       TIBETAN MARK BSDUS RTAGS
    Prop<unicode::Script> { { 0x0F35, 0x0F35 }, unicode::Script::Tibetan }, // Mn       TIBETAN MARK NGAS BZUNG NYI ZLA
    Prop<unicode::Script> { { 0x0F36, 0x0F36 }, unicode::Script::Tibetan }, // So       TIBETAN MARK CARET -DZUD RTAGS BZHI MIG CAN
    Prop<unicode::Script> { { 0x0F37, 0x0F37 }, unicode::Script::Tibetan }, // Mn       TIBETAN MARK NGAS BZUNG SGOR RTAGS
    Prop<unicode::Script> { { 0x0F38, 0x0F38 }, unicode::Script::Tibetan }, // So       TIBETAN MARK CHE MGO
    Prop<unicode::Script> { { 0x0F39, 0x0F39 }, unicode::Script::Tibetan }, // Mn       TIBETAN MARK TSA -PHRU
    Prop<unicode::Script> { { 0x0F3A, 0x0F3A }, unicode::Script::Tibetan }, // Ps       TIBETAN MARK GUG RTAGS GYON
    Prop<unicode::Script> { { 0x0F3B, 0x0F3B }, unicode::Script::Tibetan }, // Pe       TIBETAN MARK GUG RTAGS GYAS
    Prop<unicode::Script> { { 0x0F3C, 0x0F3C }, unicode::Script::Tibetan }, // Ps       TIBETAN MARK ANG KHANG GYON
    Prop<unicode::Script> { { 0x0F3D, 0x0F3D }, unicode::Script::Tibetan }, // Pe       TIBETAN MARK ANG KHANG GYAS
    Prop<unicode::Script> { { 0x0F3E, 0x0F3F }, unicode::Script::Tibetan }, // Mc   [2] TIBETAN SIGN YAR TSHES..TIBETAN SIGN MAR TSHES
    Prop<unicode::Script> { { 0x0F40, 0x0F47 }, unicode::Script::Tibetan }, // Lo   [8] TIBETAN LETTER KA..TIBETAN LETTER JA
    Prop<unicode::Script> { { 0x0F49, 0x0F6C }, unicode::Script::Tibetan }, // Lo  [36] TIBETAN LETTER NYA..TIBETAN LETTER RRA
    Prop<unicode::Script> { { 0x0F71, 0x0F7E }, unicode::Script::Tibetan }, // Mn  [14] TIBETAN VOWEL SIGN AA..TIBETAN SIGN RJES SU NGA RO
    Prop<unicode::Script> { { 0x0F7F, 0x0F7F }, unicode::Script::Tibetan }, // Mc       TIBETAN SIGN RNAM BCAD
    Prop<unicode::Script> { { 0x0F80, 0x0F84 }, unicode::Script::Tibetan }, // Mn   [5] TIBETAN VOWEL SIGN REVERSED I..TIBETAN MARK HALANTA
    Prop<unicode::Script> { { 0x0F85, 0x0F85 }, unicode::Script::Tibetan }, // Po       TIBETAN MARK PALUTA
    Prop<unicode::Script> { { 0x0F86, 0x0F87 }, unicode::Script::Tibetan }, // Mn   [2] TIBETAN SIGN LCI RTAGS..TIBETAN SIGN YANG RTAGS
    Prop<unicode::Script> { { 0x0F88, 0x0F8C }, unicode::Script::Tibetan }, // Lo   [5] TIBETAN SIGN LCE TSA CAN..TIBETAN SIGN INVERTED MCHU CAN
    Prop<unicode::Script> { { 0x0F8D, 0x0F97 }, unicode::Script::Tibetan }, // Mn  [11] TIBETAN SUBJOINED SIGN LCE TSA CAN..TIBETAN SUBJOINED LETTER JA
    Prop<unicode::Script> { { 0x0F99, 0x0FBC }, unicode::Script::Tibetan }, // Mn  [36] TIBETAN SUBJOINED LETTER NYA..TIBETAN SUBJOINED LETTER FIXED-FORM RA
    Prop<unicode::Script> { { 0x0FBE, 0x0FC5 }, unicode::Script::Tibetan }, // So   [8] TIBETAN KU RU KHA..TIBETAN SYMBOL RDO RJE
    Prop<unicode::Script> { { 0x0FC6, 0x0FC6 }, unicode::Script::Tibetan }, // Mn       TIBETAN SYMBOL PADMA GDAN
    Prop<unicode::Script> { { 0x0FC7, 0x0FCC }, unicode::Script::Tibetan }, // So   [6] TIBETAN SYMBOL RDO RJE RGYA GRAM..TIBETAN SYMBOL NOR BU BZHI -KHYIL
    Prop<unicode::Script> { { 0x0FCE, 0x0FCF }, unicode::Script::Tibetan }, // So   [2] TIBETAN SIGN RDEL NAG RDEL DKAR..TIBETAN SIGN RDEL NAG GSUM
    Prop<unicode::Script> { { 0x0FD0, 0x0FD4 }, unicode::Script::Tibetan }, // Po   [5] TIBETAN MARK BSKA- SHOG GI MGO RGYAN..TIBETAN MARK CLOSING BRDA RNYING YIG MGO SGAB MA
    Prop<unicode::Script> { { 0x0FD5, 0x0FD8 }, unicode::Script::Common }, // So   [4] RIGHT-FACING SVASTI SIGN..LEFT-FACING SVASTI SIGN WITH DOTS
    Prop<unicode::Script> { { 0x0FD9, 0x0FDA }, unicode::Script::Tibetan }, // Po   [2] TIBETAN MARK LEADING MCHAN RTAGS..TIBETAN MARK TRAILING MCHAN RTAGS
    Prop<unicode::Script> { { 0x1000, 0x102A }, unicode::Script::Myanmar }, // Lo  [43] MYANMAR LETTER KA..MYANMAR LETTER AU
    Prop<unicode::Script> { { 0x102B, 0x102C }, unicode::Script::Myanmar }, // Mc   [2] MYANMAR VOWEL SIGN TALL AA..MYANMAR VOWEL SIGN AA
    Prop<unicode::Script> { { 0x102D, 0x1030 }, unicode::Script::Myanmar }, // Mn   [4] MYANMAR VOWEL SIGN I..MYANMAR VOWEL SIGN UU
    Prop<unicode::Script> { { 0x1031, 0x1031 }, unicode::Script::Myanmar }, // Mc       MYANMAR VOWEL SIGN E
    Prop<unicode::Script> { { 0x1032, 0x1037 }, unicode::Script::Myanmar }, // Mn   [6] MYANMAR VOWEL SIGN AI..MYANMAR SIGN DOT BELOW
    Prop<unicode::Script> { { 0x1038, 0x1038 }, unicode::Script::Myanmar }, // Mc       MYANMAR SIGN VISARGA
    Prop<unicode::Script> { { 0x1039, 0x103A }, unicode::Script::Myanmar }, // Mn   [2] MYANMAR SIGN VIRAMA..MYANMAR SIGN ASAT
    Prop<unicode::Script> { { 0x103B, 0x103C }, unicode::Script::Myanmar }, // Mc   [2] MYANMAR CONSONANT SIGN MEDIAL YA..MYANMAR CONSONANT SIGN MEDIAL RA
    Prop<unicode::Script> { { 0x103D, 0x103E }, unicode::Script::Myanmar }, // Mn   [2] MYANMAR CONSONANT SIGN MEDIAL WA..MYANMAR CONSONANT SIGN MEDIAL HA
    Prop<unicode::Script> { { 0x103F, 0x103F }, unicode::Script::Myanmar }, // Lo       MYANMAR LETTER GREAT SA
    Prop<unicode::Script> { { 0x1040, 0x1049 }, unicode::Script::Myanmar }, // Nd  [10] MYANMAR DIGIT ZERO..MYANMAR DIGIT NINE
    Prop<unicode::Script> { { 0x104A, 0x104F }, unicode::Script::Myanmar }, // Po   [6] MYANMAR SIGN LITTLE SECTION..MYANMAR SYMBOL GENITIVE
    Prop<unicode::Script> { { 0x1050, 0x1055 }, unicode::Script::Myanmar }, // Lo   [6] MYANMAR LETTER SHA..MYANMAR LETTER VOCALIC LL
    Prop<unicode::Script> { { 0x1056, 0x1057 }, unicode::Script::Myanmar }, // Mc   [2] MYANMAR VOWEL SIGN VOCALIC R..MYANMAR VOWEL SIGN VOCALIC RR
    Prop<unicode::Script> { { 0x1058, 0x1059 }, unicode::Script::Myanmar }, // Mn   [2] MYANMAR VOWEL SIGN VOCALIC L..MYANMAR VOWEL SIGN VOCALIC LL
    Prop<unicode::Script> { { 0x105A, 0x105D }, unicode::Script::Myanmar }, // Lo   [4] MYANMAR LETTER MON NGA..MYANMAR LETTER MON BBE
    Prop<unicode::Script> { { 0x105E, 0x1060 }, unicode::Script::Myanmar }, // Mn   [3] MYANMAR CONSONANT SIGN MON MEDIAL NA..MYANMAR CONSONANT SIGN MON MEDIAL LA
    Prop<unicode::Script> { { 0x1061, 0x1061 }, unicode::Script::Myanmar }, // Lo       MYANMAR LETTER SGAW KAREN SHA
    Prop<unicode::Script> { { 0x1062, 0x1064 }, unicode::Script::Myanmar }, // Mc   [3] MYANMAR VOWEL SIGN SGAW KAREN EU..MYANMAR TONE MARK SGAW KAREN KE PHO
    Prop<unicode::Script> { { 0x1065, 0x1066 }, unicode::Script::Myanmar }, // Lo   [2] MYANMAR LETTER WESTERN PWO KAREN THA..MYANMAR LETTER WESTERN PWO KAREN PWA
    Prop<unicode::Script> { { 0x1067, 0x106D }, unicode::Script::Myanmar }, // Mc   [7] MYANMAR VOWEL SIGN WESTERN PWO KAREN EU..MYANMAR SIGN WESTERN PWO KAREN TONE-5
    Prop<unicode::Script> { { 0x106E, 0x1070 }, unicode::Script::Myanmar }, // Lo   [3] MYANMAR LETTER EASTERN PWO KAREN NNA..MYANMAR LETTER EASTERN PWO KAREN GHWA
    Prop<unicode::Script> { { 0x1071, 0x1074 }, unicode::Script::Myanmar }, // Mn   [4] MYANMAR VOWEL SIGN GEBA KAREN I..MYANMAR VOWEL SIGN KAYAH EE
    Prop<unicode::Script> { { 0x1075, 0x1081 }, unicode::Script::Myanmar }, // Lo  [13] MYANMAR LETTER SHAN KA..MYANMAR LETTER SHAN HA
    Prop<unicode::Script> { { 0x1082, 0x1082 }, unicode::Script::Myanmar }, // Mn       MYANMAR CONSONANT SIGN SHAN MEDIAL WA
    Prop<unicode::Script> { { 0x1083, 0x1084 }, unicode::Script::Myanmar }, // Mc   [2] MYANMAR VOWEL SIGN SHAN AA..MYANMAR VOWEL SIGN SHAN E
    Prop<unicode::Script> { { 0x1085, 0x1086 }, unicode::Script::Myanmar }, // Mn   [2] MYANMAR VOWEL SIGN SHAN E ABOVE..MYANMAR VOWEL SIGN SHAN FINAL Y
    Prop<unicode::Script> { { 0x1087, 0x108C }, unicode::Script::Myanmar }, // Mc   [6] MYANMAR SIGN SHAN TONE-2..MYANMAR SIGN SHAN COUNCIL TONE-3
    Prop<unicode::Script> { { 0x108D, 0x108D }, unicode::Script::Myanmar }, // Mn       MYANMAR SIGN SHAN COUNCIL EMPHATIC TONE
    Prop<unicode::Script> { { 0x108E, 0x108E }, unicode::Script::Myanmar }, // Lo       MYANMAR LETTER RUMAI PALAUNG FA
    Prop<unicode::Script> { { 0x108F, 0x108F }, unicode::Script::Myanmar }, // Mc       MYANMAR SIGN RUMAI PALAUNG TONE-5
    Prop<unicode::Script> { { 0x1090, 0x1099 }, unicode::Script::Myanmar }, // Nd  [10] MYANMAR SHAN DIGIT ZERO..MYANMAR SHAN DIGIT NINE
    Prop<unicode::Script> { { 0x109A, 0x109C }, unicode::Script::Myanmar }, // Mc   [3] MYANMAR SIGN KHAMTI TONE-1..MYANMAR VOWEL SIGN AITON A
    Prop<unicode::Script> { { 0x109D, 0x109D }, unicode::Script::Myanmar }, // Mn       MYANMAR VOWEL SIGN AITON AI
    Prop<unicode::Script> { { 0x109E, 0x109F }, unicode::Script::Myanmar }, // So   [2] MYANMAR SYMBOL SHAN ONE..MYANMAR SYMBOL SHAN EXCLAMATION
    Prop<unicode::Script> { { 0x10A0, 0x10C5 }, unicode::Script::Georgian }, // L&  [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Prop<unicode::Script> { { 0x10C7, 0x10C7 }, unicode::Script::Georgian }, // L&       GEORGIAN CAPITAL LETTER YN
    Prop<unicode::Script> { { 0x10CD, 0x10CD }, unicode::Script::Georgian }, // L&       GEORGIAN CAPITAL LETTER AEN
    Prop<unicode::Script> { { 0x10D0, 0x10FA }, unicode::Script::Georgian }, // L&  [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    Prop<unicode::Script> { { 0x10FB, 0x10FB }, unicode::Script::Common }, // Po       GEORGIAN PARAGRAPH SEPARATOR
    Prop<unicode::Script> { { 0x10FC, 0x10FC }, unicode::Script::Georgian }, // Lm       MODIFIER LETTER GEORGIAN NAR
    Prop<unicode::Script> { { 0x10FD, 0x10FF }, unicode::Script::Georgian }, // L&   [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Prop<unicode::Script> { { 0x1100, 0x11FF }, unicode::Script::Hangul }, // Lo [256] HANGUL CHOSEONG KIYEOK..HANGUL JONGSEONG SSANGNIEUN
    Prop<unicode::Script> { { 0x1200, 0x1248 }, unicode::Script::Ethiopic }, // Lo  [73] ETHIOPIC SYLLABLE HA..ETHIOPIC SYLLABLE QWA
    Prop<unicode::Script> { { 0x124A, 0x124D }, unicode::Script::Ethiopic }, // Lo   [4] ETHIOPIC SYLLABLE QWI..ETHIOPIC SYLLABLE QWE
    Prop<unicode::Script> { { 0x1250, 0x1256 }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE QHA..ETHIOPIC SYLLABLE QHO
    Prop<unicode::Script> { { 0x1258, 0x1258 }, unicode::Script::Ethiopic }, // Lo       ETHIOPIC SYLLABLE QHWA
    Prop<unicode::Script> { { 0x125A, 0x125D }, unicode::Script::Ethiopic }, // Lo   [4] ETHIOPIC SYLLABLE QHWI..ETHIOPIC SYLLABLE QHWE
    Prop<unicode::Script> { { 0x1260, 0x1288 }, unicode::Script::Ethiopic }, // Lo  [41] ETHIOPIC SYLLABLE BA..ETHIOPIC SYLLABLE XWA
    Prop<unicode::Script> { { 0x128A, 0x128D }, unicode::Script::Ethiopic }, // Lo   [4] ETHIOPIC SYLLABLE XWI..ETHIOPIC SYLLABLE XWE
    Prop<unicode::Script> { { 0x1290, 0x12B0 }, unicode::Script::Ethiopic }, // Lo  [33] ETHIOPIC SYLLABLE NA..ETHIOPIC SYLLABLE KWA
    Prop<unicode::Script> { { 0x12B2, 0x12B5 }, unicode::Script::Ethiopic }, // Lo   [4] ETHIOPIC SYLLABLE KWI..ETHIOPIC SYLLABLE KWE
    Prop<unicode::Script> { { 0x12B8, 0x12BE }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE KXA..ETHIOPIC SYLLABLE KXO
    Prop<unicode::Script> { { 0x12C0, 0x12C0 }, unicode::Script::Ethiopic }, // Lo       ETHIOPIC SYLLABLE KXWA
    Prop<unicode::Script> { { 0x12C2, 0x12C5 }, unicode::Script::Ethiopic }, // Lo   [4] ETHIOPIC SYLLABLE KXWI..ETHIOPIC SYLLABLE KXWE
    Prop<unicode::Script> { { 0x12C8, 0x12D6 }, unicode::Script::Ethiopic }, // Lo  [15] ETHIOPIC SYLLABLE WA..ETHIOPIC SYLLABLE PHARYNGEAL O
    Prop<unicode::Script> { { 0x12D8, 0x1310 }, unicode::Script::Ethiopic }, // Lo  [57] ETHIOPIC SYLLABLE ZA..ETHIOPIC SYLLABLE GWA
    Prop<unicode::Script> { { 0x1312, 0x1315 }, unicode::Script::Ethiopic }, // Lo   [4] ETHIOPIC SYLLABLE GWI..ETHIOPIC SYLLABLE GWE
    Prop<unicode::Script> { { 0x1318, 0x135A }, unicode::Script::Ethiopic }, // Lo  [67] ETHIOPIC SYLLABLE GGA..ETHIOPIC SYLLABLE FYA
    Prop<unicode::Script> { { 0x135D, 0x135F }, unicode::Script::Ethiopic }, // Mn   [3] ETHIOPIC COMBINING GEMINATION AND VOWEL LENGTH MARK..ETHIOPIC COMBINING GEMINATION MARK
    Prop<unicode::Script> { { 0x1360, 0x1368 }, unicode::Script::Ethiopic }, // Po   [9] ETHIOPIC SECTION MARK..ETHIOPIC PARAGRAPH SEPARATOR
    Prop<unicode::Script> { { 0x1369, 0x137C }, unicode::Script::Ethiopic }, // No  [20] ETHIOPIC DIGIT ONE..ETHIOPIC NUMBER TEN THOUSAND
    Prop<unicode::Script> { { 0x1380, 0x138F }, unicode::Script::Ethiopic }, // Lo  [16] ETHIOPIC SYLLABLE SEBATBEIT MWA..ETHIOPIC SYLLABLE PWE
    Prop<unicode::Script> { { 0x1390, 0x1399 }, unicode::Script::Ethiopic }, // So  [10] ETHIOPIC TONAL MARK YIZET..ETHIOPIC TONAL MARK KURT
    Prop<unicode::Script> { { 0x13A0, 0x13F5 }, unicode::Script::Cherokee }, // L&  [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Prop<unicode::Script> { { 0x13F8, 0x13FD }, unicode::Script::Cherokee }, // L&   [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    Prop<unicode::Script> { { 0x1400, 0x1400 }, unicode::Script::Canadian_Aboriginal }, // Pd       CANADIAN SYLLABICS HYPHEN
    Prop<unicode::Script> { { 0x1401, 0x166C }, unicode::Script::Canadian_Aboriginal }, // Lo [620] CANADIAN SYLLABICS E..CANADIAN SYLLABICS CARRIER TTSA
    Prop<unicode::Script> { { 0x166D, 0x166D }, unicode::Script::Canadian_Aboriginal }, // So       CANADIAN SYLLABICS CHI SIGN
    Prop<unicode::Script> { { 0x166E, 0x166E }, unicode::Script::Canadian_Aboriginal }, // Po       CANADIAN SYLLABICS FULL STOP
    Prop<unicode::Script> { { 0x166F, 0x167F }, unicode::Script::Canadian_Aboriginal }, // Lo  [17] CANADIAN SYLLABICS QAI..CANADIAN SYLLABICS BLACKFOOT W
    Prop<unicode::Script> { { 0x1680, 0x1680 }, unicode::Script::Ogham }, // Zs       OGHAM SPACE MARK
    Prop<unicode::Script> { { 0x1681, 0x169A }, unicode::Script::Ogham }, // Lo  [26] OGHAM LETTER BEITH..OGHAM LETTER PEITH
    Prop<unicode::Script> { { 0x169B, 0x169B }, unicode::Script::Ogham }, // Ps       OGHAM FEATHER MARK
    Prop<unicode::Script> { { 0x169C, 0x169C }, unicode::Script::Ogham }, // Pe       OGHAM REVERSED FEATHER MARK
    Prop<unicode::Script> { { 0x16A0, 0x16EA }, unicode::Script::Runic }, // Lo  [75] RUNIC LETTER FEHU FEOH FE F..RUNIC LETTER X
    Prop<unicode::Script> { { 0x16EB, 0x16ED }, unicode::Script::Common }, // Po   [3] RUNIC SINGLE PUNCTUATION..RUNIC CROSS PUNCTUATION
    Prop<unicode::Script> { { 0x16EE, 0x16F0 }, unicode::Script::Runic }, // Nl   [3] RUNIC ARLAUG SYMBOL..RUNIC BELGTHOR SYMBOL
    Prop<unicode::Script> { { 0x16F1, 0x16F8 }, unicode::Script::Runic }, // Lo   [8] RUNIC LETTER K..RUNIC LETTER FRANKS CASKET AESC
    Prop<unicode::Script> { { 0x1700, 0x1711 }, unicode::Script::Tagalog }, // Lo  [18] TAGALOG LETTER A..TAGALOG LETTER HA
    Prop<unicode::Script> { { 0x1712, 0x1714 }, unicode::Script::Tagalog }, // Mn   [3] TAGALOG VOWEL SIGN I..TAGALOG SIGN VIRAMA
    Prop<unicode::Script> { { 0x1715, 0x1715 }, unicode::Script::Tagalog }, // Mc       TAGALOG SIGN PAMUDPOD
    Prop<unicode::Script> { { 0x171F, 0x171F }, unicode::Script::Tagalog }, // Lo       TAGALOG LETTER ARCHAIC RA
    Prop<unicode::Script> { { 0x1720, 0x1731 }, unicode::Script::Hanunoo }, // Lo  [18] HANUNOO LETTER A..HANUNOO LETTER HA
    Prop<unicode::Script> { { 0x1732, 0x1733 }, unicode::Script::Hanunoo }, // Mn   [2] HANUNOO VOWEL SIGN I..HANUNOO VOWEL SIGN U
    Prop<unicode::Script> { { 0x1734, 0x1734 }, unicode::Script::Hanunoo }, // Mc       HANUNOO SIGN PAMUDPOD
    Prop<unicode::Script> { { 0x1735, 0x1736 }, unicode::Script::Common }, // Po   [2] PHILIPPINE SINGLE PUNCTUATION..PHILIPPINE DOUBLE PUNCTUATION
    Prop<unicode::Script> { { 0x1740, 0x1751 }, unicode::Script::Buhid }, // Lo  [18] BUHID LETTER A..BUHID LETTER HA
    Prop<unicode::Script> { { 0x1752, 0x1753 }, unicode::Script::Buhid }, // Mn   [2] BUHID VOWEL SIGN I..BUHID VOWEL SIGN U
    Prop<unicode::Script> { { 0x1760, 0x176C }, unicode::Script::Tagbanwa }, // Lo  [13] TAGBANWA LETTER A..TAGBANWA LETTER YA
    Prop<unicode::Script> { { 0x176E, 0x1770 }, unicode::Script::Tagbanwa }, // Lo   [3] TAGBANWA LETTER LA..TAGBANWA LETTER SA
    Prop<unicode::Script> { { 0x1772, 0x1773 }, unicode::Script::Tagbanwa }, // Mn   [2] TAGBANWA VOWEL SIGN I..TAGBANWA VOWEL SIGN U
    Prop<unicode::Script> { { 0x1780, 0x17B3 }, unicode::Script::Khmer }, // Lo  [52] KHMER LETTER KA..KHMER INDEPENDENT VOWEL QAU
    Prop<unicode::Script> { { 0x17B4, 0x17B5 }, unicode::Script::Khmer }, // Mn   [2] KHMER VOWEL INHERENT AQ..KHMER VOWEL INHERENT AA
    Prop<unicode::Script> { { 0x17B6, 0x17B6 }, unicode::Script::Khmer }, // Mc       KHMER VOWEL SIGN AA
    Prop<unicode::Script> { { 0x17B7, 0x17BD }, unicode::Script::Khmer }, // Mn   [7] KHMER VOWEL SIGN I..KHMER VOWEL SIGN UA
    Prop<unicode::Script> { { 0x17BE, 0x17C5 }, unicode::Script::Khmer }, // Mc   [8] KHMER VOWEL SIGN OE..KHMER VOWEL SIGN AU
    Prop<unicode::Script> { { 0x17C6, 0x17C6 }, unicode::Script::Khmer }, // Mn       KHMER SIGN NIKAHIT
    Prop<unicode::Script> { { 0x17C7, 0x17C8 }, unicode::Script::Khmer }, // Mc   [2] KHMER SIGN REAHMUK..KHMER SIGN YUUKALEAPINTU
    Prop<unicode::Script> { { 0x17C9, 0x17D3 }, unicode::Script::Khmer }, // Mn  [11] KHMER SIGN MUUSIKATOAN..KHMER SIGN BATHAMASAT
    Prop<unicode::Script> { { 0x17D4, 0x17D6 }, unicode::Script::Khmer }, // Po   [3] KHMER SIGN KHAN..KHMER SIGN CAMNUC PII KUUH
    Prop<unicode::Script> { { 0x17D7, 0x17D7 }, unicode::Script::Khmer }, // Lm       KHMER SIGN LEK TOO
    Prop<unicode::Script> { { 0x17D8, 0x17DA }, unicode::Script::Khmer }, // Po   [3] KHMER SIGN BEYYAL..KHMER SIGN KOOMUUT
    Prop<unicode::Script> { { 0x17DB, 0x17DB }, unicode::Script::Khmer }, // Sc       KHMER CURRENCY SYMBOL RIEL
    Prop<unicode::Script> { { 0x17DC, 0x17DC }, unicode::Script::Khmer }, // Lo       KHMER SIGN AVAKRAHASANYA
    Prop<unicode::Script> { { 0x17DD, 0x17DD }, unicode::Script::Khmer }, // Mn       KHMER SIGN ATTHACAN
    Prop<unicode::Script> { { 0x17E0, 0x17E9 }, unicode::Script::Khmer }, // Nd  [10] KHMER DIGIT ZERO..KHMER DIGIT NINE
    Prop<unicode::Script> { { 0x17F0, 0x17F9 }, unicode::Script::Khmer }, // No  [10] KHMER SYMBOL LEK ATTAK SON..KHMER SYMBOL LEK ATTAK PRAM-BUON
    Prop<unicode::Script> { { 0x1800, 0x1801 }, unicode::Script::Mongolian }, // Po   [2] MONGOLIAN BIRGA..MONGOLIAN ELLIPSIS
    Prop<unicode::Script> { { 0x1802, 0x1803 }, unicode::Script::Common }, // Po   [2] MONGOLIAN COMMA..MONGOLIAN FULL STOP
    Prop<unicode::Script> { { 0x1804, 0x1804 }, unicode::Script::Mongolian }, // Po       MONGOLIAN COLON
    Prop<unicode::Script> { { 0x1805, 0x1805 }, unicode::Script::Common }, // Po       MONGOLIAN FOUR DOTS
    Prop<unicode::Script> { { 0x1806, 0x1806 }, unicode::Script::Mongolian }, // Pd       MONGOLIAN TODO SOFT HYPHEN
    Prop<unicode::Script> { { 0x1807, 0x180A }, unicode::Script::Mongolian }, // Po   [4] MONGOLIAN SIBE SYLLABLE BOUNDARY MARKER..MONGOLIAN NIRUGU
    Prop<unicode::Script> { { 0x180B, 0x180D }, unicode::Script::Mongolian }, // Mn   [3] MONGOLIAN FREE VARIATION SELECTOR ONE..MONGOLIAN FREE VARIATION SELECTOR THREE
    Prop<unicode::Script> { { 0x180E, 0x180E }, unicode::Script::Mongolian }, // Cf       MONGOLIAN VOWEL SEPARATOR
    Prop<unicode::Script> { { 0x180F, 0x180F }, unicode::Script::Mongolian }, // Mn       MONGOLIAN FREE VARIATION SELECTOR FOUR
    Prop<unicode::Script> { { 0x1810, 0x1819 }, unicode::Script::Mongolian }, // Nd  [10] MONGOLIAN DIGIT ZERO..MONGOLIAN DIGIT NINE
    Prop<unicode::Script> { { 0x1820, 0x1842 }, unicode::Script::Mongolian }, // Lo  [35] MONGOLIAN LETTER A..MONGOLIAN LETTER CHI
    Prop<unicode::Script> { { 0x1843, 0x1843 }, unicode::Script::Mongolian }, // Lm       MONGOLIAN LETTER TODO LONG VOWEL SIGN
    Prop<unicode::Script> { { 0x1844, 0x1878 }, unicode::Script::Mongolian }, // Lo  [53] MONGOLIAN LETTER TODO E..MONGOLIAN LETTER CHA WITH TWO DOTS
    Prop<unicode::Script> { { 0x1880, 0x1884 }, unicode::Script::Mongolian }, // Lo   [5] MONGOLIAN LETTER ALI GALI ANUSVARA ONE..MONGOLIAN LETTER ALI GALI INVERTED UBADAMA
    Prop<unicode::Script> { { 0x1885, 0x1886 }, unicode::Script::Mongolian }, // Mn   [2] MONGOLIAN LETTER ALI GALI BALUDA..MONGOLIAN LETTER ALI GALI THREE BALUDA
    Prop<unicode::Script> { { 0x1887, 0x18A8 }, unicode::Script::Mongolian }, // Lo  [34] MONGOLIAN LETTER ALI GALI A..MONGOLIAN LETTER MANCHU ALI GALI BHA
    Prop<unicode::Script> { { 0x18A9, 0x18A9 }, unicode::Script::Mongolian }, // Mn       MONGOLIAN LETTER ALI GALI DAGALGA
    Prop<unicode::Script> { { 0x18AA, 0x18AA }, unicode::Script::Mongolian }, // Lo       MONGOLIAN LETTER MANCHU ALI GALI LHA
    Prop<unicode::Script> { { 0x18B0, 0x18F5 }, unicode::Script::Canadian_Aboriginal }, // Lo  [70] CANADIAN SYLLABICS OY..CANADIAN SYLLABICS CARRIER DENTAL S
    Prop<unicode::Script> { { 0x1900, 0x191E }, unicode::Script::Limbu }, // Lo  [31] LIMBU VOWEL-CARRIER LETTER..LIMBU LETTER TRA
    Prop<unicode::Script> { { 0x1920, 0x1922 }, unicode::Script::Limbu }, // Mn   [3] LIMBU VOWEL SIGN A..LIMBU VOWEL SIGN U
    Prop<unicode::Script> { { 0x1923, 0x1926 }, unicode::Script::Limbu }, // Mc   [4] LIMBU VOWEL SIGN EE..LIMBU VOWEL SIGN AU
    Prop<unicode::Script> { { 0x1927, 0x1928 }, unicode::Script::Limbu }, // Mn   [2] LIMBU VOWEL SIGN E..LIMBU VOWEL SIGN O
    Prop<unicode::Script> { { 0x1929, 0x192B }, unicode::Script::Limbu }, // Mc   [3] LIMBU SUBJOINED LETTER YA..LIMBU SUBJOINED LETTER WA
    Prop<unicode::Script> { { 0x1930, 0x1931 }, unicode::Script::Limbu }, // Mc   [2] LIMBU SMALL LETTER KA..LIMBU SMALL LETTER NGA
    Prop<unicode::Script> { { 0x1932, 0x1932 }, unicode::Script::Limbu }, // Mn       LIMBU SMALL LETTER ANUSVARA
    Prop<unicode::Script> { { 0x1933, 0x1938 }, unicode::Script::Limbu }, // Mc   [6] LIMBU SMALL LETTER TA..LIMBU SMALL LETTER LA
    Prop<unicode::Script> { { 0x1939, 0x193B }, unicode::Script::Limbu }, // Mn   [3] LIMBU SIGN MUKPHRENG..LIMBU SIGN SA-I
    Prop<unicode::Script> { { 0x1940, 0x1940 }, unicode::Script::Limbu }, // So       LIMBU SIGN LOO
    Prop<unicode::Script> { { 0x1944, 0x1945 }, unicode::Script::Limbu }, // Po   [2] LIMBU EXCLAMATION MARK..LIMBU QUESTION MARK
    Prop<unicode::Script> { { 0x1946, 0x194F }, unicode::Script::Limbu }, // Nd  [10] LIMBU DIGIT ZERO..LIMBU DIGIT NINE
    Prop<unicode::Script> { { 0x1950, 0x196D }, unicode::Script::Tai_Le }, // Lo  [30] TAI LE LETTER KA..TAI LE LETTER AI
    Prop<unicode::Script> { { 0x1970, 0x1974 }, unicode::Script::Tai_Le }, // Lo   [5] TAI LE LETTER TONE-2..TAI LE LETTER TONE-6
    Prop<unicode::Script> { { 0x1980, 0x19AB }, unicode::Script::New_Tai_Lue }, // Lo  [44] NEW TAI LUE LETTER HIGH QA..NEW TAI LUE LETTER LOW SUA
    Prop<unicode::Script> { { 0x19B0, 0x19C9 }, unicode::Script::New_Tai_Lue }, // Lo  [26] NEW TAI LUE VOWEL SIGN VOWEL SHORTENER..NEW TAI LUE TONE MARK-2
    Prop<unicode::Script> { { 0x19D0, 0x19D9 }, unicode::Script::New_Tai_Lue }, // Nd  [10] NEW TAI LUE DIGIT ZERO..NEW TAI LUE DIGIT NINE
    Prop<unicode::Script> { { 0x19DA, 0x19DA }, unicode::Script::New_Tai_Lue }, // No       NEW TAI LUE THAM DIGIT ONE
    Prop<unicode::Script> { { 0x19DE, 0x19DF }, unicode::Script::New_Tai_Lue }, // So   [2] NEW TAI LUE SIGN LAE..NEW TAI LUE SIGN LAEV
    Prop<unicode::Script> { { 0x19E0, 0x19FF }, unicode::Script::Khmer }, // So  [32] KHMER SYMBOL PATHAMASAT..KHMER SYMBOL DAP-PRAM ROC
    Prop<unicode::Script> { { 0x1A00, 0x1A16 }, unicode::Script::Buginese }, // Lo  [23] BUGINESE LETTER KA..BUGINESE LETTER HA
    Prop<unicode::Script> { { 0x1A17, 0x1A18 }, unicode::Script::Buginese }, // Mn   [2] BUGINESE VOWEL SIGN I..BUGINESE VOWEL SIGN U
    Prop<unicode::Script> { { 0x1A19, 0x1A1A }, unicode::Script::Buginese }, // Mc   [2] BUGINESE VOWEL SIGN E..BUGINESE VOWEL SIGN O
    Prop<unicode::Script> { { 0x1A1B, 0x1A1B }, unicode::Script::Buginese }, // Mn       BUGINESE VOWEL SIGN AE
    Prop<unicode::Script> { { 0x1A1E, 0x1A1F }, unicode::Script::Buginese }, // Po   [2] BUGINESE PALLAWA..BUGINESE END OF SECTION
    Prop<unicode::Script> { { 0x1A20, 0x1A54 }, unicode::Script::Tai_Tham }, // Lo  [53] TAI THAM LETTER HIGH KA..TAI THAM LETTER GREAT SA
    Prop<unicode::Script> { { 0x1A55, 0x1A55 }, unicode::Script::Tai_Tham }, // Mc       TAI THAM CONSONANT SIGN MEDIAL RA
    Prop<unicode::Script> { { 0x1A56, 0x1A56 }, unicode::Script::Tai_Tham }, // Mn       TAI THAM CONSONANT SIGN MEDIAL LA
    Prop<unicode::Script> { { 0x1A57, 0x1A57 }, unicode::Script::Tai_Tham }, // Mc       TAI THAM CONSONANT SIGN LA TANG LAI
    Prop<unicode::Script> { { 0x1A58, 0x1A5E }, unicode::Script::Tai_Tham }, // Mn   [7] TAI THAM SIGN MAI KANG LAI..TAI THAM CONSONANT SIGN SA
    Prop<unicode::Script> { { 0x1A60, 0x1A60 }, unicode::Script::Tai_Tham }, // Mn       TAI THAM SIGN SAKOT
    Prop<unicode::Script> { { 0x1A61, 0x1A61 }, unicode::Script::Tai_Tham }, // Mc       TAI THAM VOWEL SIGN A
    Prop<unicode::Script> { { 0x1A62, 0x1A62 }, unicode::Script::Tai_Tham }, // Mn       TAI THAM VOWEL SIGN MAI SAT
    Prop<unicode::Script> { { 0x1A63, 0x1A64 }, unicode::Script::Tai_Tham }, // Mc   [2] TAI THAM VOWEL SIGN AA..TAI THAM VOWEL SIGN TALL AA
    Prop<unicode::Script> { { 0x1A65, 0x1A6C }, unicode::Script::Tai_Tham }, // Mn   [8] TAI THAM VOWEL SIGN I..TAI THAM VOWEL SIGN OA BELOW
    Prop<unicode::Script> { { 0x1A6D, 0x1A72 }, unicode::Script::Tai_Tham }, // Mc   [6] TAI THAM VOWEL SIGN OY..TAI THAM VOWEL SIGN THAM AI
    Prop<unicode::Script> { { 0x1A73, 0x1A7C }, unicode::Script::Tai_Tham }, // Mn  [10] TAI THAM VOWEL SIGN OA ABOVE..TAI THAM SIGN KHUEN-LUE KARAN
    Prop<unicode::Script> { { 0x1A7F, 0x1A7F }, unicode::Script::Tai_Tham }, // Mn       TAI THAM COMBINING CRYPTOGRAMMIC DOT
    Prop<unicode::Script> { { 0x1A80, 0x1A89 }, unicode::Script::Tai_Tham }, // Nd  [10] TAI THAM HORA DIGIT ZERO..TAI THAM HORA DIGIT NINE
    Prop<unicode::Script> { { 0x1A90, 0x1A99 }, unicode::Script::Tai_Tham }, // Nd  [10] TAI THAM THAM DIGIT ZERO..TAI THAM THAM DIGIT NINE
    Prop<unicode::Script> { { 0x1AA0, 0x1AA6 }, unicode::Script::Tai_Tham }, // Po   [7] TAI THAM SIGN WIANG..TAI THAM SIGN REVERSED ROTATED RANA
    Prop<unicode::Script> { { 0x1AA7, 0x1AA7 }, unicode::Script::Tai_Tham }, // Lm       TAI THAM SIGN MAI YAMOK
    Prop<unicode::Script> { { 0x1AA8, 0x1AAD }, unicode::Script::Tai_Tham }, // Po   [6] TAI THAM SIGN KAAN..TAI THAM SIGN CAANG
    Prop<unicode::Script> { { 0x1AB0, 0x1ABD }, unicode::Script::Inherited }, // Mn  [14] COMBINING DOUBLED CIRCUMFLEX ACCENT..COMBINING PARENTHESES BELOW
    Prop<unicode::Script> { { 0x1ABE, 0x1ABE }, unicode::Script::Inherited }, // Me       COMBINING PARENTHESES OVERLAY
    Prop<unicode::Script> { { 0x1ABF, 0x1ADD }, unicode::Script::Inherited }, // Mn  [31] COMBINING LATIN SMALL LETTER W BELOW..COMBINING DOT-AND-RING BELOW
    Prop<unicode::Script> { { 0x1AE0, 0x1AEB }, unicode::Script::Inherited }, // Mn  [12] COMBINING LEFT TACK ABOVE..COMBINING DOUBLE RIGHTWARDS ARROW ABOVE
    Prop<unicode::Script> { { 0x1B00, 0x1B03 }, unicode::Script::Balinese }, // Mn   [4] BALINESE SIGN ULU RICEM..BALINESE SIGN SURANG
    Prop<unicode::Script> { { 0x1B04, 0x1B04 }, unicode::Script::Balinese }, // Mc       BALINESE SIGN BISAH
    Prop<unicode::Script> { { 0x1B05, 0x1B33 }, unicode::Script::Balinese }, // Lo  [47] BALINESE LETTER AKARA..BALINESE LETTER HA
    Prop<unicode::Script> { { 0x1B34, 0x1B34 }, unicode::Script::Balinese }, // Mn       BALINESE SIGN REREKAN
    Prop<unicode::Script> { { 0x1B35, 0x1B35 }, unicode::Script::Balinese }, // Mc       BALINESE VOWEL SIGN TEDUNG
    Prop<unicode::Script> { { 0x1B36, 0x1B3A }, unicode::Script::Balinese }, // Mn   [5] BALINESE VOWEL SIGN ULU..BALINESE VOWEL SIGN RA REPA
    Prop<unicode::Script> { { 0x1B3B, 0x1B3B }, unicode::Script::Balinese }, // Mc       BALINESE VOWEL SIGN RA REPA TEDUNG
    Prop<unicode::Script> { { 0x1B3C, 0x1B3C }, unicode::Script::Balinese }, // Mn       BALINESE VOWEL SIGN LA LENGA
    Prop<unicode::Script> { { 0x1B3D, 0x1B41 }, unicode::Script::Balinese }, // Mc   [5] BALINESE VOWEL SIGN LA LENGA TEDUNG..BALINESE VOWEL SIGN TALING REPA TEDUNG
    Prop<unicode::Script> { { 0x1B42, 0x1B42 }, unicode::Script::Balinese }, // Mn       BALINESE VOWEL SIGN PEPET
    Prop<unicode::Script> { { 0x1B43, 0x1B44 }, unicode::Script::Balinese }, // Mc   [2] BALINESE VOWEL SIGN PEPET TEDUNG..BALINESE ADEG ADEG
    Prop<unicode::Script> { { 0x1B45, 0x1B4C }, unicode::Script::Balinese }, // Lo   [8] BALINESE LETTER KAF SASAK..BALINESE LETTER ARCHAIC JNYA
    Prop<unicode::Script> { { 0x1B4E, 0x1B4F }, unicode::Script::Balinese }, // Po   [2] BALINESE INVERTED CARIK SIKI..BALINESE INVERTED CARIK PAREREN
    Prop<unicode::Script> { { 0x1B50, 0x1B59 }, unicode::Script::Balinese }, // Nd  [10] BALINESE DIGIT ZERO..BALINESE DIGIT NINE
    Prop<unicode::Script> { { 0x1B5A, 0x1B60 }, unicode::Script::Balinese }, // Po   [7] BALINESE PANTI..BALINESE PAMENENG
    Prop<unicode::Script> { { 0x1B61, 0x1B6A }, unicode::Script::Balinese }, // So  [10] BALINESE MUSICAL SYMBOL DONG..BALINESE MUSICAL SYMBOL DANG GEDE
    Prop<unicode::Script> { { 0x1B6B, 0x1B73 }, unicode::Script::Balinese }, // Mn   [9] BALINESE MUSICAL SYMBOL COMBINING TEGEH..BALINESE MUSICAL SYMBOL COMBINING GONG
    Prop<unicode::Script> { { 0x1B74, 0x1B7C }, unicode::Script::Balinese }, // So   [9] BALINESE MUSICAL SYMBOL RIGHT-HAND OPEN DUG..BALINESE MUSICAL SYMBOL LEFT-HAND OPEN PING
    Prop<unicode::Script> { { 0x1B7D, 0x1B7F }, unicode::Script::Balinese }, // Po   [3] BALINESE PANTI LANTANG..BALINESE PANTI BAWAK
    Prop<unicode::Script> { { 0x1B80, 0x1B81 }, unicode::Script::Sundanese }, // Mn   [2] SUNDANESE SIGN PANYECEK..SUNDANESE SIGN PANGLAYAR
    Prop<unicode::Script> { { 0x1B82, 0x1B82 }, unicode::Script::Sundanese }, // Mc       SUNDANESE SIGN PANGWISAD
    Prop<unicode::Script> { { 0x1B83, 0x1BA0 }, unicode::Script::Sundanese }, // Lo  [30] SUNDANESE LETTER A..SUNDANESE LETTER HA
    Prop<unicode::Script> { { 0x1BA1, 0x1BA1 }, unicode::Script::Sundanese }, // Mc       SUNDANESE CONSONANT SIGN PAMINGKAL
    Prop<unicode::Script> { { 0x1BA2, 0x1BA5 }, unicode::Script::Sundanese }, // Mn   [4] SUNDANESE CONSONANT SIGN PANYAKRA..SUNDANESE VOWEL SIGN PANYUKU
    Prop<unicode::Script> { { 0x1BA6, 0x1BA7 }, unicode::Script::Sundanese }, // Mc   [2] SUNDANESE VOWEL SIGN PANAELAENG..SUNDANESE VOWEL SIGN PANOLONG
    Prop<unicode::Script> { { 0x1BA8, 0x1BA9 }, unicode::Script::Sundanese }, // Mn   [2] SUNDANESE VOWEL SIGN PAMEPET..SUNDANESE VOWEL SIGN PANEULEUNG
    Prop<unicode::Script> { { 0x1BAA, 0x1BAA }, unicode::Script::Sundanese }, // Mc       SUNDANESE SIGN PAMAAEH
    Prop<unicode::Script> { { 0x1BAB, 0x1BAD }, unicode::Script::Sundanese }, // Mn   [3] SUNDANESE SIGN VIRAMA..SUNDANESE CONSONANT SIGN PASANGAN WA
    Prop<unicode::Script> { { 0x1BAE, 0x1BAF }, unicode::Script::Sundanese }, // Lo   [2] SUNDANESE LETTER KHA..SUNDANESE LETTER SYA
    Prop<unicode::Script> { { 0x1BB0, 0x1BB9 }, unicode::Script::Sundanese }, // Nd  [10] SUNDANESE DIGIT ZERO..SUNDANESE DIGIT NINE
    Prop<unicode::Script> { { 0x1BBA, 0x1BBF }, unicode::Script::Sundanese }, // Lo   [6] SUNDANESE AVAGRAHA..SUNDANESE LETTER FINAL M
    Prop<unicode::Script> { { 0x1BC0, 0x1BE5 }, unicode::Script::Batak }, // Lo  [38] BATAK LETTER A..BATAK LETTER U
    Prop<unicode::Script> { { 0x1BE6, 0x1BE6 }, unicode::Script::Batak }, // Mn       BATAK SIGN TOMPI
    Prop<unicode::Script> { { 0x1BE7, 0x1BE7 }, unicode::Script::Batak }, // Mc       BATAK VOWEL SIGN E
    Prop<unicode::Script> { { 0x1BE8, 0x1BE9 }, unicode::Script::Batak }, // Mn   [2] BATAK VOWEL SIGN PAKPAK E..BATAK VOWEL SIGN EE
    Prop<unicode::Script> { { 0x1BEA, 0x1BEC }, unicode::Script::Batak }, // Mc   [3] BATAK VOWEL SIGN I..BATAK VOWEL SIGN O
    Prop<unicode::Script> { { 0x1BED, 0x1BED }, unicode::Script::Batak }, // Mn       BATAK VOWEL SIGN KARO O
    Prop<unicode::Script> { { 0x1BEE, 0x1BEE }, unicode::Script::Batak }, // Mc       BATAK VOWEL SIGN U
    Prop<unicode::Script> { { 0x1BEF, 0x1BF1 }, unicode::Script::Batak }, // Mn   [3] BATAK VOWEL SIGN U FOR SIMALUNGUN SA..BATAK CONSONANT SIGN H
    Prop<unicode::Script> { { 0x1BF2, 0x1BF3 }, unicode::Script::Batak }, // Mc   [2] BATAK PANGOLAT..BATAK PANONGONAN
    Prop<unicode::Script> { { 0x1BFC, 0x1BFF }, unicode::Script::Batak }, // Po   [4] BATAK SYMBOL BINDU NA METEK..BATAK SYMBOL BINDU PANGOLAT
    Prop<unicode::Script> { { 0x1C00, 0x1C23 }, unicode::Script::Lepcha }, // Lo  [36] LEPCHA LETTER KA..LEPCHA LETTER A
    Prop<unicode::Script> { { 0x1C24, 0x1C2B }, unicode::Script::Lepcha }, // Mc   [8] LEPCHA SUBJOINED LETTER YA..LEPCHA VOWEL SIGN UU
    Prop<unicode::Script> { { 0x1C2C, 0x1C33 }, unicode::Script::Lepcha }, // Mn   [8] LEPCHA VOWEL SIGN E..LEPCHA CONSONANT SIGN T
    Prop<unicode::Script> { { 0x1C34, 0x1C35 }, unicode::Script::Lepcha }, // Mc   [2] LEPCHA CONSONANT SIGN NYIN-DO..LEPCHA CONSONANT SIGN KANG
    Prop<unicode::Script> { { 0x1C36, 0x1C37 }, unicode::Script::Lepcha }, // Mn   [2] LEPCHA SIGN RAN..LEPCHA SIGN NUKTA
    Prop<unicode::Script> { { 0x1C3B, 0x1C3F }, unicode::Script::Lepcha }, // Po   [5] LEPCHA PUNCTUATION TA-ROL..LEPCHA PUNCTUATION TSHOOK
    Prop<unicode::Script> { { 0x1C40, 0x1C49 }, unicode::Script::Lepcha }, // Nd  [10] LEPCHA DIGIT ZERO..LEPCHA DIGIT NINE
    Prop<unicode::Script> { { 0x1C4D, 0x1C4F }, unicode::Script::Lepcha }, // Lo   [3] LEPCHA LETTER TTA..LEPCHA LETTER DDA
    Prop<unicode::Script> { { 0x1C50, 0x1C59 }, unicode::Script::Ol_Chiki }, // Nd  [10] OL CHIKI DIGIT ZERO..OL CHIKI DIGIT NINE
    Prop<unicode::Script> { { 0x1C5A, 0x1C77 }, unicode::Script::Ol_Chiki }, // Lo  [30] OL CHIKI LETTER LA..OL CHIKI LETTER OH
    Prop<unicode::Script> { { 0x1C78, 0x1C7D }, unicode::Script::Ol_Chiki }, // Lm   [6] OL CHIKI MU TTUDDAG..OL CHIKI AHAD
    Prop<unicode::Script> { { 0x1C7E, 0x1C7F }, unicode::Script::Ol_Chiki }, // Po   [2] OL CHIKI PUNCTUATION MUCAAD..OL CHIKI PUNCTUATION DOUBLE MUCAAD
    Prop<unicode::Script> { { 0x1C80, 0x1C8A }, unicode::Script::Cyrillic }, // L&  [11] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER TJE
    Prop<unicode::Script> { { 0x1C90, 0x1CBA }, unicode::Script::Georgian }, // L&  [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    Prop<unicode::Script> { { 0x1CBD, 0x1CBF }, unicode::Script::Georgian }, // L&   [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    Prop<unicode::Script> { { 0x1CC0, 0x1CC7 }, unicode::Script::Sundanese }, // Po   [8] SUNDANESE PUNCTUATION BINDU SURYA..SUNDANESE PUNCTUATION BINDU BA SATANGA
    Prop<unicode::Script> { { 0x1CD0, 0x1CD2 }, unicode::Script::Inherited }, // Mn   [3] VEDIC TONE KARSHANA..VEDIC TONE PRENKHA
    Prop<unicode::Script> { { 0x1CD3, 0x1CD3 }, unicode::Script::Common }, // Po       VEDIC SIGN NIHSHVASA
    Prop<unicode::Script> { { 0x1CD4, 0x1CE0 }, unicode::Script::Inherited }, // Mn  [13] VEDIC SIGN YAJURVEDIC MIDLINE SVARITA..VEDIC TONE RIGVEDIC KASHMIRI INDEPENDENT SVARITA
    Prop<unicode::Script> { { 0x1CE1, 0x1CE1 }, unicode::Script::Common }, // Mc       VEDIC TONE ATHARVAVEDIC INDEPENDENT SVARITA
    Prop<unicode::Script> { { 0x1CE2, 0x1CE8 }, unicode::Script::Inherited }, // Mn   [7] VEDIC SIGN VISARGA SVARITA..VEDIC SIGN VISARGA ANUDATTA WITH TAIL
    Prop<unicode::Script> { { 0x1CE9, 0x1CEC }, unicode::Script::Common }, // Lo   [4] VEDIC SIGN ANUSVARA ANTARGOMUKHA..VEDIC SIGN ANUSVARA VAMAGOMUKHA WITH TAIL
    Prop<unicode::Script> { { 0x1CED, 0x1CED }, unicode::Script::Inherited }, // Mn       VEDIC SIGN TIRYAK
    Prop<unicode::Script> { { 0x1CEE, 0x1CF3 }, unicode::Script::Common }, // Lo   [6] VEDIC SIGN HEXIFORM LONG ANUSVARA..VEDIC SIGN ROTATED ARDHAVISARGA
    Prop<unicode::Script> { { 0x1CF4, 0x1CF4 }, unicode::Script::Inherited }, // Mn       VEDIC TONE CANDRA ABOVE
    Prop<unicode::Script> { { 0x1CF5, 0x1CF6 }, unicode::Script::Common }, // Lo   [2] VEDIC SIGN JIHVAMULIYA..VEDIC SIGN UPADHMANIYA
    Prop<unicode::Script> { { 0x1CF7, 0x1CF7 }, unicode::Script::Common }, // Mc       VEDIC SIGN ATIKRAMA
    Prop<unicode::Script> { { 0x1CF8, 0x1CF9 }, unicode::Script::Inherited }, // Mn   [2] VEDIC TONE RING ABOVE..VEDIC TONE DOUBLE RING ABOVE
    Prop<unicode::Script> { { 0x1CFA, 0x1CFA }, unicode::Script::Common }, // Lo       VEDIC SIGN DOUBLE ANUSVARA ANTARGOMUKHA
    Prop<unicode::Script> { { 0x1D00, 0x1D25 }, unicode::Script::Latin }, // L&  [38] LATIN LETTER SMALL CAPITAL A..LATIN LETTER AIN
    Prop<unicode::Script> { { 0x1D26, 0x1D2A }, unicode::Script::Greek }, // L&   [5] GREEK LETTER SMALL CAPITAL GAMMA..GREEK LETTER SMALL CAPITAL PSI
    Prop<unicode::Script> { { 0x1D2B, 0x1D2B }, unicode::Script::Cyrillic }, // L&       CYRILLIC LETTER SMALL CAPITAL EL
    Prop<unicode::Script> { { 0x1D2C, 0x1D5C }, unicode::Script::Latin }, // Lm  [49] MODIFIER LETTER CAPITAL A..MODIFIER LETTER SMALL AIN
    Prop<unicode::Script> { { 0x1D5D, 0x1D61 }, unicode::Script::Greek }, // Lm   [5] MODIFIER LETTER SMALL BETA..MODIFIER LETTER SMALL CHI
    Prop<unicode::Script> { { 0x1D62, 0x1D65 }, unicode::Script::Latin }, // Lm   [4] LATIN SUBSCRIPT SMALL LETTER I..LATIN SUBSCRIPT SMALL LETTER V
    Prop<unicode::Script> { { 0x1D66, 0x1D6A }, unicode::Script::Greek }, // Lm   [5] GREEK SUBSCRIPT SMALL LETTER BETA..GREEK SUBSCRIPT SMALL LETTER CHI
    Prop<unicode::Script> { { 0x1D6B, 0x1D77 }, unicode::Script::Latin }, // L&  [13] LATIN SMALL LETTER UE..LATIN SMALL LETTER TURNED G
    Prop<unicode::Script> { { 0x1D78, 0x1D78 }, unicode::Script::Cyrillic }, // Lm       MODIFIER LETTER CYRILLIC EN
    Prop<unicode::Script> { { 0x1D79, 0x1D9A }, unicode::Script::Latin }, // L&  [34] LATIN SMALL LETTER INSULAR G..LATIN SMALL LETTER EZH WITH RETROFLEX HOOK
    Prop<unicode::Script> { { 0x1D9B, 0x1DBE }, unicode::Script::Latin }, // Lm  [36] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL EZH
    Prop<unicode::Script> { { 0x1DBF, 0x1DBF }, unicode::Script::Greek }, // Lm       MODIFIER LETTER SMALL THETA
    Prop<unicode::Script> { { 0x1DC0, 0x1DFF }, unicode::Script::Inherited }, // Mn  [64] COMBINING DOTTED GRAVE ACCENT..COMBINING RIGHT ARROWHEAD AND DOWN ARROWHEAD BELOW
    Prop<unicode::Script> { { 0x1E00, 0x1EFF }, unicode::Script::Latin }, // L& [256] LATIN CAPITAL LETTER A WITH RING BELOW..LATIN SMALL LETTER Y WITH LOOP
    Prop<unicode::Script> { { 0x1F00, 0x1F15 }, unicode::Script::Greek }, // L&  [22] GREEK SMALL LETTER ALPHA WITH PSILI..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Prop<unicode::Script> { { 0x1F18, 0x1F1D }, unicode::Script::Greek }, // L&   [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Prop<unicode::Script> { { 0x1F20, 0x1F45 }, unicode::Script::Greek }, // L&  [38] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Prop<unicode::Script> { { 0x1F48, 0x1F4D }, unicode::Script::Greek }, // L&   [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Prop<unicode::Script> { { 0x1F50, 0x1F57 }, unicode::Script::Greek }, // L&   [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Prop<unicode::Script> { { 0x1F59, 0x1F59 }, unicode::Script::Greek }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA
    Prop<unicode::Script> { { 0x1F5B, 0x1F5B }, unicode::Script::Greek }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Prop<unicode::Script> { { 0x1F5D, 0x1F5D }, unicode::Script::Greek }, // L&       GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Prop<unicode::Script> { { 0x1F5F, 0x1F7D }, unicode::Script::Greek }, // L&  [31] GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI..GREEK SMALL LETTER OMEGA WITH OXIA
    Prop<unicode::Script> { { 0x1F80, 0x1FB4 }, unicode::Script::Greek }, // L&  [53] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    Prop<unicode::Script> { { 0x1FB6, 0x1FBC }, unicode::Script::Greek }, // L&   [7] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    Prop<unicode::Script> { { 0x1FBD, 0x1FBD }, unicode::Script::Greek }, // Sk       GREEK KORONIS
    Prop<unicode::Script> { { 0x1FBE, 0x1FBE }, unicode::Script::Greek }, // L&       GREEK PROSGEGRAMMENI
    Prop<unicode::Script> { { 0x1FBF, 0x1FC1 }, unicode::Script::Greek }, // Sk   [3] GREEK PSILI..GREEK DIALYTIKA AND PERISPOMENI
    Prop<unicode::Script> { { 0x1FC2, 0x1FC4 }, unicode::Script::Greek }, // L&   [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Prop<unicode::Script> { { 0x1FC6, 0x1FCC }, unicode::Script::Greek }, // L&   [7] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    Prop<unicode::Script> { { 0x1FCD, 0x1FCF }, unicode::Script::Greek }, // Sk   [3] GREEK PSILI AND VARIA..GREEK PSILI AND PERISPOMENI
    Prop<unicode::Script> { { 0x1FD0, 0x1FD3 }, unicode::Script::Greek }, // L&   [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Prop<unicode::Script> { { 0x1FD6, 0x1FDB }, unicode::Script::Greek }, // L&   [6] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK CAPITAL LETTER IOTA WITH OXIA
    Prop<unicode::Script> { { 0x1FDD, 0x1FDF }, unicode::Script::Greek }, // Sk   [3] GREEK DASIA AND VARIA..GREEK DASIA AND PERISPOMENI
    Prop<unicode::Script> { { 0x1FE0, 0x1FEC }, unicode::Script::Greek }, // L&  [13] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    Prop<unicode::Script> { { 0x1FED, 0x1FEF }, unicode::Script::Greek }, // Sk   [3] GREEK DIALYTIKA AND VARIA..GREEK VARIA
    Prop<unicode::Script> { { 0x1FF2, 0x1FF4 }, unicode::Script::Greek }, // L&   [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    Prop<unicode::Script> { { 0x1FF6, 0x1FFC }, unicode::Script::Greek }, // L&   [7] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    Prop<unicode::Script> { { 0x1FFD, 0x1FFE }, unicode::Script::Greek }, // Sk   [2] GREEK OXIA..GREEK DASIA
    Prop<unicode::Script> { { 0x2000, 0x200A }, unicode::Script::Common }, // Zs  [11] EN QUAD..HAIR SPACE
    Prop<unicode::Script> { { 0x200B, 0x200B }, unicode::Script::Common }, // Cf       ZERO WIDTH SPACE
    Prop<unicode::Script> { { 0x200C, 0x200D }, unicode::Script::Inherited }, // Cf   [2] ZERO WIDTH NON-JOINER..ZERO WIDTH JOINER
    Prop<unicode::Script> { { 0x200E, 0x200F }, unicode::Script::Common }, // Cf   [2] LEFT-TO-RIGHT MARK..RIGHT-TO-LEFT MARK
    Prop<unicode::Script> { { 0x2010, 0x2015 }, unicode::Script::Common }, // Pd   [6] HYPHEN..HORIZONTAL BAR
    Prop<unicode::Script> { { 0x2016, 0x2017 }, unicode::Script::Common }, // Po   [2] DOUBLE VERTICAL LINE..DOUBLE LOW LINE
    Prop<unicode::Script> { { 0x2018, 0x2018 }, unicode::Script::Common }, // Pi       LEFT SINGLE QUOTATION MARK
    Prop<unicode::Script> { { 0x2019, 0x2019 }, unicode::Script::Common }, // Pf       RIGHT SINGLE QUOTATION MARK
    Prop<unicode::Script> { { 0x201A, 0x201A }, unicode::Script::Common }, // Ps       SINGLE LOW-9 QUOTATION MARK
    Prop<unicode::Script> { { 0x201B, 0x201C }, unicode::Script::Common }, // Pi   [2] SINGLE HIGH-REVERSED-9 QUOTATION MARK..LEFT DOUBLE QUOTATION MARK
    Prop<unicode::Script> { { 0x201D, 0x201D }, unicode::Script::Common }, // Pf       RIGHT DOUBLE QUOTATION MARK
    Prop<unicode::Script> { { 0x201E, 0x201E }, unicode::Script::Common }, // Ps       DOUBLE LOW-9 QUOTATION MARK
    Prop<unicode::Script> { { 0x201F, 0x201F }, unicode::Script::Common }, // Pi       DOUBLE HIGH-REVERSED-9 QUOTATION MARK
    Prop<unicode::Script> { { 0x2020, 0x2027 }, unicode::Script::Common }, // Po   [8] DAGGER..HYPHENATION POINT
    Prop<unicode::Script> { { 0x2028, 0x2028 }, unicode::Script::Common }, // Zl       LINE SEPARATOR
    Prop<unicode::Script> { { 0x2029, 0x2029 }, unicode::Script::Common }, // Zp       PARAGRAPH SEPARATOR
    Prop<unicode::Script> { { 0x202A, 0x202E }, unicode::Script::Common }, // Cf   [5] LEFT-TO-RIGHT EMBEDDING..RIGHT-TO-LEFT OVERRIDE
    Prop<unicode::Script> { { 0x202F, 0x202F }, unicode::Script::Common }, // Zs       NARROW NO-BREAK SPACE
    Prop<unicode::Script> { { 0x2030, 0x2038 }, unicode::Script::Common }, // Po   [9] PER MILLE SIGN..CARET
    Prop<unicode::Script> { { 0x2039, 0x2039 }, unicode::Script::Common }, // Pi       SINGLE LEFT-POINTING ANGLE QUOTATION MARK
    Prop<unicode::Script> { { 0x203A, 0x203A }, unicode::Script::Common }, // Pf       SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
    Prop<unicode::Script> { { 0x203B, 0x203E }, unicode::Script::Common }, // Po   [4] REFERENCE MARK..OVERLINE
    Prop<unicode::Script> { { 0x203F, 0x2040 }, unicode::Script::Common }, // Pc   [2] UNDERTIE..CHARACTER TIE
    Prop<unicode::Script> { { 0x2041, 0x2043 }, unicode::Script::Common }, // Po   [3] CARET INSERTION POINT..HYPHEN BULLET
    Prop<unicode::Script> { { 0x2044, 0x2044 }, unicode::Script::Common }, // Sm       FRACTION SLASH
    Prop<unicode::Script> { { 0x2045, 0x2045 }, unicode::Script::Common }, // Ps       LEFT SQUARE BRACKET WITH QUILL
    Prop<unicode::Script> { { 0x2046, 0x2046 }, unicode::Script::Common }, // Pe       RIGHT SQUARE BRACKET WITH QUILL
    Prop<unicode::Script> { { 0x2047, 0x2051 }, unicode::Script::Common }, // Po  [11] DOUBLE QUESTION MARK..TWO ASTERISKS ALIGNED VERTICALLY
    Prop<unicode::Script> { { 0x2052, 0x2052 }, unicode::Script::Common }, // Sm       COMMERCIAL MINUS SIGN
    Prop<unicode::Script> { { 0x2053, 0x2053 }, unicode::Script::Common }, // Po       SWUNG DASH
    Prop<unicode::Script> { { 0x2054, 0x2054 }, unicode::Script::Common }, // Pc       INVERTED UNDERTIE
    Prop<unicode::Script> { { 0x2055, 0x205E }, unicode::Script::Common }, // Po  [10] FLOWER PUNCTUATION MARK..VERTICAL FOUR DOTS
    Prop<unicode::Script> { { 0x205F, 0x205F }, unicode::Script::Common }, // Zs       MEDIUM MATHEMATICAL SPACE
    Prop<unicode::Script> { { 0x2060, 0x2064 }, unicode::Script::Common }, // Cf   [5] WORD JOINER..INVISIBLE PLUS
    Prop<unicode::Script> { { 0x2066, 0x206F }, unicode::Script::Common }, // Cf  [10] LEFT-TO-RIGHT ISOLATE..NOMINAL DIGIT SHAPES
    Prop<unicode::Script> { { 0x2070, 0x2070 }, unicode::Script::Common }, // No       SUPERSCRIPT ZERO
    Prop<unicode::Script> { { 0x2071, 0x2071 }, unicode::Script::Latin }, // Lm       SUPERSCRIPT LATIN SMALL LETTER I
    Prop<unicode::Script> { { 0x2074, 0x2079 }, unicode::Script::Common }, // No   [6] SUPERSCRIPT FOUR..SUPERSCRIPT NINE
    Prop<unicode::Script> { { 0x207A, 0x207C }, unicode::Script::Common }, // Sm   [3] SUPERSCRIPT PLUS SIGN..SUPERSCRIPT EQUALS SIGN
    Prop<unicode::Script> { { 0x207D, 0x207D }, unicode::Script::Common }, // Ps       SUPERSCRIPT LEFT PARENTHESIS
    Prop<unicode::Script> { { 0x207E, 0x207E }, unicode::Script::Common }, // Pe       SUPERSCRIPT RIGHT PARENTHESIS
    Prop<unicode::Script> { { 0x207F, 0x207F }, unicode::Script::Latin }, // Lm       SUPERSCRIPT LATIN SMALL LETTER N
    Prop<unicode::Script> { { 0x2080, 0x2089 }, unicode::Script::Common }, // No  [10] SUBSCRIPT ZERO..SUBSCRIPT NINE
    Prop<unicode::Script> { { 0x208A, 0x208C }, unicode::Script::Common }, // Sm   [3] SUBSCRIPT PLUS SIGN..SUBSCRIPT EQUALS SIGN
    Prop<unicode::Script> { { 0x208D, 0x208D }, unicode::Script::Common }, // Ps       SUBSCRIPT LEFT PARENTHESIS
    Prop<unicode::Script> { { 0x208E, 0x208E }, unicode::Script::Common }, // Pe       SUBSCRIPT RIGHT PARENTHESIS
    Prop<unicode::Script> { { 0x2090, 0x209C }, unicode::Script::Latin }, // Lm  [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Prop<unicode::Script> { { 0x20A0, 0x20C1 }, unicode::Script::Common }, // Sc  [34] EURO-CURRENCY SIGN..SAUDI RIYAL SIGN
    Prop<unicode::Script> { { 0x20D0, 0x20DC }, unicode::Script::Inherited }, // Mn  [13] COMBINING LEFT HARPOON ABOVE..COMBINING FOUR DOTS ABOVE
    Prop<unicode::Script> { { 0x20DD, 0x20E0 }, unicode::Script::Inherited }, // Me   [4] COMBINING ENCLOSING CIRCLE..COMBINING ENCLOSING CIRCLE BACKSLASH
    Prop<unicode::Script> { { 0x20E1, 0x20E1 }, unicode::Script::Inherited }, // Mn       COMBINING LEFT RIGHT ARROW ABOVE
    Prop<unicode::Script> { { 0x20E2, 0x20E4 }, unicode::Script::Inherited }, // Me   [3] COMBINING ENCLOSING SCREEN..COMBINING ENCLOSING UPWARD POINTING TRIANGLE
    Prop<unicode::Script> { { 0x20E5, 0x20F0 }, unicode::Script::Inherited }, // Mn  [12] COMBINING REVERSE SOLIDUS OVERLAY..COMBINING ASTERISK ABOVE
    Prop<unicode::Script> { { 0x2100, 0x2101 }, unicode::Script::Common }, // So   [2] ACCOUNT OF..ADDRESSED TO THE SUBJECT
    Prop<unicode::Script> { { 0x2102, 0x2102 }, unicode::Script::Common }, // L&       DOUBLE-STRUCK CAPITAL C
    Prop<unicode::Script> { { 0x2103, 0x2106 }, unicode::Script::Common }, // So   [4] DEGREE CELSIUS..CADA UNA
    Prop<unicode::Script> { { 0x2107, 0x2107 }, unicode::Script::Common }, // L&       EULER CONSTANT
    Prop<unicode::Script> { { 0x2108, 0x2109 }, unicode::Script::Common }, // So   [2] SCRUPLE..DEGREE FAHRENHEIT
    Prop<unicode::Script> { { 0x210A, 0x2113 }, unicode::Script::Common }, // L&  [10] SCRIPT SMALL G..SCRIPT SMALL L
    Prop<unicode::Script> { { 0x2114, 0x2114 }, unicode::Script::Common }, // So       L B BAR SYMBOL
    Prop<unicode::Script> { { 0x2115, 0x2115 }, unicode::Script::Common }, // L&       DOUBLE-STRUCK CAPITAL N
    Prop<unicode::Script> { { 0x2116, 0x2117 }, unicode::Script::Common }, // So   [2] NUMERO SIGN..SOUND RECORDING COPYRIGHT
    Prop<unicode::Script> { { 0x2118, 0x2118 }, unicode::Script::Common }, // Sm       SCRIPT CAPITAL P
    Prop<unicode::Script> { { 0x2119, 0x211D }, unicode::Script::Common }, // L&   [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    Prop<unicode::Script> { { 0x211E, 0x2123 }, unicode::Script::Common }, // So   [6] PRESCRIPTION TAKE..VERSICLE
    Prop<unicode::Script> { { 0x2124, 0x2124 }, unicode::Script::Common }, // L&       DOUBLE-STRUCK CAPITAL Z
    Prop<unicode::Script> { { 0x2125, 0x2125 }, unicode::Script::Common }, // So       OUNCE SIGN
    Prop<unicode::Script> { { 0x2126, 0x2126 }, unicode::Script::Greek }, // L&       OHM SIGN
    Prop<unicode::Script> { { 0x2127, 0x2127 }, unicode::Script::Common }, // So       INVERTED OHM SIGN
    Prop<unicode::Script> { { 0x2128, 0x2128 }, unicode::Script::Common }, // L&       BLACK-LETTER CAPITAL Z
    Prop<unicode::Script> { { 0x2129, 0x2129 }, unicode::Script::Common }, // So       TURNED GREEK SMALL LETTER IOTA
    Prop<unicode::Script> { { 0x212A, 0x212B }, unicode::Script::Latin }, // L&   [2] KELVIN SIGN..ANGSTROM SIGN
    Prop<unicode::Script> { { 0x212C, 0x212D }, unicode::Script::Common }, // L&   [2] SCRIPT CAPITAL B..BLACK-LETTER CAPITAL C
    Prop<unicode::Script> { { 0x212E, 0x212E }, unicode::Script::Common }, // So       ESTIMATED SYMBOL
    Prop<unicode::Script> { { 0x212F, 0x2131 }, unicode::Script::Common }, // L&   [3] SCRIPT SMALL E..SCRIPT CAPITAL F
    Prop<unicode::Script> { { 0x2132, 0x2132 }, unicode::Script::Latin }, // L&       TURNED CAPITAL F
    Prop<unicode::Script> { { 0x2133, 0x2134 }, unicode::Script::Common }, // L&   [2] SCRIPT CAPITAL M..SCRIPT SMALL O
    Prop<unicode::Script> { { 0x2135, 0x2138 }, unicode::Script::Common }, // Lo   [4] ALEF SYMBOL..DALET SYMBOL
    Prop<unicode::Script> { { 0x2139, 0x2139 }, unicode::Script::Common }, // L&       INFORMATION SOURCE
    Prop<unicode::Script> { { 0x213A, 0x213B }, unicode::Script::Common }, // So   [2] ROTATED CAPITAL Q..FACSIMILE SIGN
    Prop<unicode::Script> { { 0x213C, 0x213F }, unicode::Script::Common }, // L&   [4] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK CAPITAL PI
    Prop<unicode::Script> { { 0x2140, 0x2144 }, unicode::Script::Common }, // Sm   [5] DOUBLE-STRUCK N-ARY SUMMATION..TURNED SANS-SERIF CAPITAL Y
    Prop<unicode::Script> { { 0x2145, 0x2149 }, unicode::Script::Common }, // L&   [5] DOUBLE-STRUCK ITALIC CAPITAL D..DOUBLE-STRUCK ITALIC SMALL J
    Prop<unicode::Script> { { 0x214A, 0x214A }, unicode::Script::Common }, // So       PROPERTY LINE
    Prop<unicode::Script> { { 0x214B, 0x214B }, unicode::Script::Common }, // Sm       TURNED AMPERSAND
    Prop<unicode::Script> { { 0x214C, 0x214D }, unicode::Script::Common }, // So   [2] PER SIGN..AKTIESELSKAB
    Prop<unicode::Script> { { 0x214E, 0x214E }, unicode::Script::Latin }, // L&       TURNED SMALL F
    Prop<unicode::Script> { { 0x214F, 0x214F }, unicode::Script::Common }, // So       SYMBOL FOR SAMARITAN SOURCE
    Prop<unicode::Script> { { 0x2150, 0x215F }, unicode::Script::Common }, // No  [16] VULGAR FRACTION ONE SEVENTH..FRACTION NUMERATOR ONE
    Prop<unicode::Script> { { 0x2160, 0x2182 }, unicode::Script::Latin }, // Nl  [35] ROMAN NUMERAL ONE..ROMAN NUMERAL TEN THOUSAND
    Prop<unicode::Script> { { 0x2183, 0x2184 }, unicode::Script::Latin }, // L&   [2] ROMAN NUMERAL REVERSED ONE HUNDRED..LATIN SMALL LETTER REVERSED C
    Prop<unicode::Script> { { 0x2185, 0x2188 }, unicode::Script::Latin }, // Nl   [4] ROMAN NUMERAL SIX LATE FORM..ROMAN NUMERAL ONE HUNDRED THOUSAND
    Prop<unicode::Script> { { 0x2189, 0x2189 }, unicode::Script::Common }, // No       VULGAR FRACTION ZERO THIRDS
    Prop<unicode::Script> { { 0x218A, 0x218B }, unicode::Script::Common }, // So   [2] TURNED DIGIT TWO..TURNED DIGIT THREE
    Prop<unicode::Script> { { 0x2190, 0x2194 }, unicode::Script::Common }, // Sm   [5] LEFTWARDS ARROW..LEFT RIGHT ARROW
    Prop<unicode::Script> { { 0x2195, 0x2199 }, unicode::Script::Common }, // So   [5] UP DOWN ARROW..SOUTH WEST ARROW
    Prop<unicode::Script> { { 0x219A, 0x219B }, unicode::Script::Common }, // Sm   [2] LEFTWARDS ARROW WITH STROKE..RIGHTWARDS ARROW WITH STROKE
    Prop<unicode::Script> { { 0x219C, 0x219F }, unicode::Script::Common }, // So   [4] LEFTWARDS WAVE ARROW..UPWARDS TWO HEADED ARROW
    Prop<unicode::Script> { { 0x21A0, 0x21A0 }, unicode::Script::Common }, // Sm       RIGHTWARDS TWO HEADED ARROW
    Prop<unicode::Script> { { 0x21A1, 0x21A2 }, unicode::Script::Common }, // So   [2] DOWNWARDS TWO HEADED ARROW..LEFTWARDS ARROW WITH TAIL
    Prop<unicode::Script> { { 0x21A3, 0x21A3 }, unicode::Script::Common }, // Sm       RIGHTWARDS ARROW WITH TAIL
    Prop<unicode::Script> { { 0x21A4, 0x21A5 }, unicode::Script::Common }, // So   [2] LEFTWARDS ARROW FROM BAR..UPWARDS ARROW FROM BAR
    Prop<unicode::Script> { { 0x21A6, 0x21A6 }, unicode::Script::Common }, // Sm       RIGHTWARDS ARROW FROM BAR
    Prop<unicode::Script> { { 0x21A7, 0x21AD }, unicode::Script::Common }, // So   [7] DOWNWARDS ARROW FROM BAR..LEFT RIGHT WAVE ARROW
    Prop<unicode::Script> { { 0x21AE, 0x21AE }, unicode::Script::Common }, // Sm       LEFT RIGHT ARROW WITH STROKE
    Prop<unicode::Script> { { 0x21AF, 0x21CD }, unicode::Script::Common }, // So  [31] DOWNWARDS ZIGZAG ARROW..LEFTWARDS DOUBLE ARROW WITH STROKE
    Prop<unicode::Script> { { 0x21CE, 0x21CF }, unicode::Script::Common }, // Sm   [2] LEFT RIGHT DOUBLE ARROW WITH STROKE..RIGHTWARDS DOUBLE ARROW WITH STROKE
    Prop<unicode::Script> { { 0x21D0, 0x21D1 }, unicode::Script::Common }, // So   [2] LEFTWARDS DOUBLE ARROW..UPWARDS DOUBLE ARROW
    Prop<unicode::Script> { { 0x21D2, 0x21D2 }, unicode::Script::Common }, // Sm       RIGHTWARDS DOUBLE ARROW
    Prop<unicode::Script> { { 0x21D3, 0x21D3 }, unicode::Script::Common }, // So       DOWNWARDS DOUBLE ARROW
    Prop<unicode::Script> { { 0x21D4, 0x21D4 }, unicode::Script::Common }, // Sm       LEFT RIGHT DOUBLE ARROW
    Prop<unicode::Script> { { 0x21D5, 0x21F3 }, unicode::Script::Common }, // So  [31] UP DOWN DOUBLE ARROW..UP DOWN WHITE ARROW
    Prop<unicode::Script> { { 0x21F4, 0x22FF }, unicode::Script::Common }, // Sm [268] RIGHT ARROW WITH SMALL CIRCLE..Z NOTATION BAG MEMBERSHIP
    Prop<unicode::Script> { { 0x2300, 0x2307 }, unicode::Script::Common }, // So   [8] DIAMETER SIGN..WAVY LINE
    Prop<unicode::Script> { { 0x2308, 0x2308 }, unicode::Script::Common }, // Ps       LEFT CEILING
    Prop<unicode::Script> { { 0x2309, 0x2309 }, unicode::Script::Common }, // Pe       RIGHT CEILING
    Prop<unicode::Script> { { 0x230A, 0x230A }, unicode::Script::Common }, // Ps       LEFT FLOOR
    Prop<unicode::Script> { { 0x230B, 0x230B }, unicode::Script::Common }, // Pe       RIGHT FLOOR
    Prop<unicode::Script> { { 0x230C, 0x231F }, unicode::Script::Common }, // So  [20] BOTTOM RIGHT CROP..BOTTOM RIGHT CORNER
    Prop<unicode::Script> { { 0x2320, 0x2321 }, unicode::Script::Common }, // Sm   [2] TOP HALF INTEGRAL..BOTTOM HALF INTEGRAL
    Prop<unicode::Script> { { 0x2322, 0x2328 }, unicode::Script::Common }, // So   [7] FROWN..KEYBOARD
    Prop<unicode::Script> { { 0x2329, 0x2329 }, unicode::Script::Common }, // Ps       LEFT-POINTING ANGLE BRACKET
    Prop<unicode::Script> { { 0x232A, 0x232A }, unicode::Script::Common }, // Pe       RIGHT-POINTING ANGLE BRACKET
    Prop<unicode::Script> { { 0x232B, 0x237B }, unicode::Script::Common }, // So  [81] ERASE TO THE LEFT..NOT CHECK MARK
    Prop<unicode::Script> { { 0x237C, 0x237C }, unicode::Script::Common }, // Sm       RIGHT ANGLE WITH DOWNWARDS ZIGZAG ARROW
    Prop<unicode::Script> { { 0x237D, 0x239A }, unicode::Script::Common }, // So  [30] SHOULDERED OPEN BOX..CLEAR SCREEN SYMBOL
    Prop<unicode::Script> { { 0x239B, 0x23B3 }, unicode::Script::Common }, // Sm  [25] LEFT PARENTHESIS UPPER HOOK..SUMMATION BOTTOM
    Prop<unicode::Script> { { 0x23B4, 0x23DB }, unicode::Script::Common }, // So  [40] TOP SQUARE BRACKET..FUSE
    Prop<unicode::Script> { { 0x23DC, 0x23E1 }, unicode::Script::Common }, // Sm   [6] TOP PARENTHESIS..BOTTOM TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0x23E2, 0x2429 }, unicode::Script::Common }, // So  [72] WHITE TRAPEZIUM..SYMBOL FOR DELETE MEDIUM SHADE FORM
    Prop<unicode::Script> { { 0x2440, 0x244A }, unicode::Script::Common }, // So  [11] OCR HOOK..OCR DOUBLE BACKSLASH
    Prop<unicode::Script> { { 0x2460, 0x249B }, unicode::Script::Common }, // No  [60] CIRCLED DIGIT ONE..NUMBER TWENTY FULL STOP
    Prop<unicode::Script> { { 0x249C, 0x24E9 }, unicode::Script::Common }, // So  [78] PARENTHESIZED LATIN SMALL LETTER A..CIRCLED LATIN SMALL LETTER Z
    Prop<unicode::Script> { { 0x24EA, 0x24FF }, unicode::Script::Common }, // No  [22] CIRCLED DIGIT ZERO..NEGATIVE CIRCLED DIGIT ZERO
    Prop<unicode::Script> { { 0x2500, 0x25B6 }, unicode::Script::Common }, // So [183] BOX DRAWINGS LIGHT HORIZONTAL..BLACK RIGHT-POINTING TRIANGLE
    Prop<unicode::Script> { { 0x25B7, 0x25B7 }, unicode::Script::Common }, // Sm       WHITE RIGHT-POINTING TRIANGLE
    Prop<unicode::Script> { { 0x25B8, 0x25C0 }, unicode::Script::Common }, // So   [9] BLACK RIGHT-POINTING SMALL TRIANGLE..BLACK LEFT-POINTING TRIANGLE
    Prop<unicode::Script> { { 0x25C1, 0x25C1 }, unicode::Script::Common }, // Sm       WHITE LEFT-POINTING TRIANGLE
    Prop<unicode::Script> { { 0x25C2, 0x25F7 }, unicode::Script::Common }, // So  [54] BLACK LEFT-POINTING SMALL TRIANGLE..WHITE CIRCLE WITH UPPER RIGHT QUADRANT
    Prop<unicode::Script> { { 0x25F8, 0x25FF }, unicode::Script::Common }, // Sm   [8] UPPER LEFT TRIANGLE..LOWER RIGHT TRIANGLE
    Prop<unicode::Script> { { 0x2600, 0x266E }, unicode::Script::Common }, // So [111] BLACK SUN WITH RAYS..MUSIC NATURAL SIGN
    Prop<unicode::Script> { { 0x266F, 0x266F }, unicode::Script::Common }, // Sm       MUSIC SHARP SIGN
    Prop<unicode::Script> { { 0x2670, 0x2767 }, unicode::Script::Common }, // So [248] WEST SYRIAC CROSS..ROTATED FLORAL HEART BULLET
    Prop<unicode::Script> { { 0x2768, 0x2768 }, unicode::Script::Common }, // Ps       MEDIUM LEFT PARENTHESIS ORNAMENT
    Prop<unicode::Script> { { 0x2769, 0x2769 }, unicode::Script::Common }, // Pe       MEDIUM RIGHT PARENTHESIS ORNAMENT
    Prop<unicode::Script> { { 0x276A, 0x276A }, unicode::Script::Common }, // Ps       MEDIUM FLATTENED LEFT PARENTHESIS ORNAMENT
    Prop<unicode::Script> { { 0x276B, 0x276B }, unicode::Script::Common }, // Pe       MEDIUM FLATTENED RIGHT PARENTHESIS ORNAMENT
    Prop<unicode::Script> { { 0x276C, 0x276C }, unicode::Script::Common }, // Ps       MEDIUM LEFT-POINTING ANGLE BRACKET ORNAMENT
    Prop<unicode::Script> { { 0x276D, 0x276D }, unicode::Script::Common }, // Pe       MEDIUM RIGHT-POINTING ANGLE BRACKET ORNAMENT
    Prop<unicode::Script> { { 0x276E, 0x276E }, unicode::Script::Common }, // Ps       HEAVY LEFT-POINTING ANGLE QUOTATION MARK ORNAMENT
    Prop<unicode::Script> { { 0x276F, 0x276F }, unicode::Script::Common }, // Pe       HEAVY RIGHT-POINTING ANGLE QUOTATION MARK ORNAMENT
    Prop<unicode::Script> { { 0x2770, 0x2770 }, unicode::Script::Common }, // Ps       HEAVY LEFT-POINTING ANGLE BRACKET ORNAMENT
    Prop<unicode::Script> { { 0x2771, 0x2771 }, unicode::Script::Common }, // Pe       HEAVY RIGHT-POINTING ANGLE BRACKET ORNAMENT
    Prop<unicode::Script> { { 0x2772, 0x2772 }, unicode::Script::Common }, // Ps       LIGHT LEFT TORTOISE SHELL BRACKET ORNAMENT
    Prop<unicode::Script> { { 0x2773, 0x2773 }, unicode::Script::Common }, // Pe       LIGHT RIGHT TORTOISE SHELL BRACKET ORNAMENT
    Prop<unicode::Script> { { 0x2774, 0x2774 }, unicode::Script::Common }, // Ps       MEDIUM LEFT CURLY BRACKET ORNAMENT
    Prop<unicode::Script> { { 0x2775, 0x2775 }, unicode::Script::Common }, // Pe       MEDIUM RIGHT CURLY BRACKET ORNAMENT
    Prop<unicode::Script> { { 0x2776, 0x2793 }, unicode::Script::Common }, // No  [30] DINGBAT NEGATIVE CIRCLED DIGIT ONE..DINGBAT NEGATIVE CIRCLED SANS-SERIF NUMBER TEN
    Prop<unicode::Script> { { 0x2794, 0x27BF }, unicode::Script::Common }, // So  [44] HEAVY WIDE-HEADED RIGHTWARDS ARROW..DOUBLE CURLY LOOP
    Prop<unicode::Script> { { 0x27C0, 0x27C4 }, unicode::Script::Common }, // Sm   [5] THREE DIMENSIONAL ANGLE..OPEN SUPERSET
    Prop<unicode::Script> { { 0x27C5, 0x27C5 }, unicode::Script::Common }, // Ps       LEFT S-SHAPED BAG DELIMITER
    Prop<unicode::Script> { { 0x27C6, 0x27C6 }, unicode::Script::Common }, // Pe       RIGHT S-SHAPED BAG DELIMITER
    Prop<unicode::Script> { { 0x27C7, 0x27E5 }, unicode::Script::Common }, // Sm  [31] OR WITH DOT INSIDE..WHITE SQUARE WITH RIGHTWARDS TICK
    Prop<unicode::Script> { { 0x27E6, 0x27E6 }, unicode::Script::Common }, // Ps       MATHEMATICAL LEFT WHITE SQUARE BRACKET
    Prop<unicode::Script> { { 0x27E7, 0x27E7 }, unicode::Script::Common }, // Pe       MATHEMATICAL RIGHT WHITE SQUARE BRACKET
    Prop<unicode::Script> { { 0x27E8, 0x27E8 }, unicode::Script::Common }, // Ps       MATHEMATICAL LEFT ANGLE BRACKET
    Prop<unicode::Script> { { 0x27E9, 0x27E9 }, unicode::Script::Common }, // Pe       MATHEMATICAL RIGHT ANGLE BRACKET
    Prop<unicode::Script> { { 0x27EA, 0x27EA }, unicode::Script::Common }, // Ps       MATHEMATICAL LEFT DOUBLE ANGLE BRACKET
    Prop<unicode::Script> { { 0x27EB, 0x27EB }, unicode::Script::Common }, // Pe       MATHEMATICAL RIGHT DOUBLE ANGLE BRACKET
    Prop<unicode::Script> { { 0x27EC, 0x27EC }, unicode::Script::Common }, // Ps       MATHEMATICAL LEFT WHITE TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0x27ED, 0x27ED }, unicode::Script::Common }, // Pe       MATHEMATICAL RIGHT WHITE TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0x27EE, 0x27EE }, unicode::Script::Common }, // Ps       MATHEMATICAL LEFT FLATTENED PARENTHESIS
    Prop<unicode::Script> { { 0x27EF, 0x27EF }, unicode::Script::Common }, // Pe       MATHEMATICAL RIGHT FLATTENED PARENTHESIS
    Prop<unicode::Script> { { 0x27F0, 0x27FF }, unicode::Script::Common }, // Sm  [16] UPWARDS QUADRUPLE ARROW..LONG RIGHTWARDS SQUIGGLE ARROW
    Prop<unicode::Script> { { 0x2800, 0x28FF }, unicode::Script::Braille }, // So [256] BRAILLE PATTERN BLANK..BRAILLE PATTERN DOTS-12345678
    Prop<unicode::Script> { { 0x2900, 0x2982 }, unicode::Script::Common }, // Sm [131] RIGHTWARDS TWO-HEADED ARROW WITH VERTICAL STROKE..Z NOTATION TYPE COLON
    Prop<unicode::Script> { { 0x2983, 0x2983 }, unicode::Script::Common }, // Ps       LEFT WHITE CURLY BRACKET
    Prop<unicode::Script> { { 0x2984, 0x2984 }, unicode::Script::Common }, // Pe       RIGHT WHITE CURLY BRACKET
    Prop<unicode::Script> { { 0x2985, 0x2985 }, unicode::Script::Common }, // Ps       LEFT WHITE PARENTHESIS
    Prop<unicode::Script> { { 0x2986, 0x2986 }, unicode::Script::Common }, // Pe       RIGHT WHITE PARENTHESIS
    Prop<unicode::Script> { { 0x2987, 0x2987 }, unicode::Script::Common }, // Ps       Z NOTATION LEFT IMAGE BRACKET
    Prop<unicode::Script> { { 0x2988, 0x2988 }, unicode::Script::Common }, // Pe       Z NOTATION RIGHT IMAGE BRACKET
    Prop<unicode::Script> { { 0x2989, 0x2989 }, unicode::Script::Common }, // Ps       Z NOTATION LEFT BINDING BRACKET
    Prop<unicode::Script> { { 0x298A, 0x298A }, unicode::Script::Common }, // Pe       Z NOTATION RIGHT BINDING BRACKET
    Prop<unicode::Script> { { 0x298B, 0x298B }, unicode::Script::Common }, // Ps       LEFT SQUARE BRACKET WITH UNDERBAR
    Prop<unicode::Script> { { 0x298C, 0x298C }, unicode::Script::Common }, // Pe       RIGHT SQUARE BRACKET WITH UNDERBAR
    Prop<unicode::Script> { { 0x298D, 0x298D }, unicode::Script::Common }, // Ps       LEFT SQUARE BRACKET WITH TICK IN TOP CORNER
    Prop<unicode::Script> { { 0x298E, 0x298E }, unicode::Script::Common }, // Pe       RIGHT SQUARE BRACKET WITH TICK IN BOTTOM CORNER
    Prop<unicode::Script> { { 0x298F, 0x298F }, unicode::Script::Common }, // Ps       LEFT SQUARE BRACKET WITH TICK IN BOTTOM CORNER
    Prop<unicode::Script> { { 0x2990, 0x2990 }, unicode::Script::Common }, // Pe       RIGHT SQUARE BRACKET WITH TICK IN TOP CORNER
    Prop<unicode::Script> { { 0x2991, 0x2991 }, unicode::Script::Common }, // Ps       LEFT ANGLE BRACKET WITH DOT
    Prop<unicode::Script> { { 0x2992, 0x2992 }, unicode::Script::Common }, // Pe       RIGHT ANGLE BRACKET WITH DOT
    Prop<unicode::Script> { { 0x2993, 0x2993 }, unicode::Script::Common }, // Ps       LEFT ARC LESS-THAN BRACKET
    Prop<unicode::Script> { { 0x2994, 0x2994 }, unicode::Script::Common }, // Pe       RIGHT ARC GREATER-THAN BRACKET
    Prop<unicode::Script> { { 0x2995, 0x2995 }, unicode::Script::Common }, // Ps       DOUBLE LEFT ARC GREATER-THAN BRACKET
    Prop<unicode::Script> { { 0x2996, 0x2996 }, unicode::Script::Common }, // Pe       DOUBLE RIGHT ARC LESS-THAN BRACKET
    Prop<unicode::Script> { { 0x2997, 0x2997 }, unicode::Script::Common }, // Ps       LEFT BLACK TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0x2998, 0x2998 }, unicode::Script::Common }, // Pe       RIGHT BLACK TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0x2999, 0x29D7 }, unicode::Script::Common }, // Sm  [63] DOTTED FENCE..BLACK HOURGLASS
    Prop<unicode::Script> { { 0x29D8, 0x29D8 }, unicode::Script::Common }, // Ps       LEFT WIGGLY FENCE
    Prop<unicode::Script> { { 0x29D9, 0x29D9 }, unicode::Script::Common }, // Pe       RIGHT WIGGLY FENCE
    Prop<unicode::Script> { { 0x29DA, 0x29DA }, unicode::Script::Common }, // Ps       LEFT DOUBLE WIGGLY FENCE
    Prop<unicode::Script> { { 0x29DB, 0x29DB }, unicode::Script::Common }, // Pe       RIGHT DOUBLE WIGGLY FENCE
    Prop<unicode::Script> { { 0x29DC, 0x29FB }, unicode::Script::Common }, // Sm  [32] INCOMPLETE INFINITY..TRIPLE PLUS
    Prop<unicode::Script> { { 0x29FC, 0x29FC }, unicode::Script::Common }, // Ps       LEFT-POINTING CURVED ANGLE BRACKET
    Prop<unicode::Script> { { 0x29FD, 0x29FD }, unicode::Script::Common }, // Pe       RIGHT-POINTING CURVED ANGLE BRACKET
    Prop<unicode::Script> { { 0x29FE, 0x2AFF }, unicode::Script::Common }, // Sm [258] TINY..N-ARY WHITE VERTICAL BAR
    Prop<unicode::Script> { { 0x2B00, 0x2B2F }, unicode::Script::Common }, // So  [48] NORTH EAST WHITE ARROW..WHITE VERTICAL ELLIPSE
    Prop<unicode::Script> { { 0x2B30, 0x2B44 }, unicode::Script::Common }, // Sm  [21] LEFT ARROW WITH SMALL CIRCLE..RIGHTWARDS ARROW THROUGH SUPERSET
    Prop<unicode::Script> { { 0x2B45, 0x2B46 }, unicode::Script::Common }, // So   [2] LEFTWARDS QUADRUPLE ARROW..RIGHTWARDS QUADRUPLE ARROW
    Prop<unicode::Script> { { 0x2B47, 0x2B4C }, unicode::Script::Common }, // Sm   [6] REVERSE TILDE OPERATOR ABOVE RIGHTWARDS ARROW..RIGHTWARDS ARROW ABOVE REVERSE TILDE OPERATOR
    Prop<unicode::Script> { { 0x2B4D, 0x2B73 }, unicode::Script::Common }, // So  [39] DOWNWARDS TRIANGLE-HEADED ZIGZAG ARROW..DOWNWARDS TRIANGLE-HEADED ARROW TO BAR
    Prop<unicode::Script> { { 0x2B76, 0x2BFF }, unicode::Script::Common }, // So [138] NORTH WEST TRIANGLE-HEADED ARROW TO BAR..HELLSCHREIBER PAUSE SYMBOL
    Prop<unicode::Script> { { 0x2C00, 0x2C5F }, unicode::Script::Glagolitic }, // L&  [96] GLAGOLITIC CAPITAL LETTER AZU..GLAGOLITIC SMALL LETTER CAUDATE CHRIVI
    Prop<unicode::Script> { { 0x2C60, 0x2C7B }, unicode::Script::Latin }, // L&  [28] LATIN CAPITAL LETTER L WITH DOUBLE BAR..LATIN LETTER SMALL CAPITAL TURNED E
    Prop<unicode::Script> { { 0x2C7C, 0x2C7D }, unicode::Script::Latin }, // Lm   [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    Prop<unicode::Script> { { 0x2C7E, 0x2C7F }, unicode::Script::Latin }, // L&   [2] LATIN CAPITAL LETTER S WITH SWASH TAIL..LATIN CAPITAL LETTER Z WITH SWASH TAIL
    Prop<unicode::Script> { { 0x2C80, 0x2CE4 }, unicode::Script::Coptic }, // L& [101] COPTIC CAPITAL LETTER ALFA..COPTIC SYMBOL KAI
    Prop<unicode::Script> { { 0x2CE5, 0x2CEA }, unicode::Script::Coptic }, // So   [6] COPTIC SYMBOL MI RO..COPTIC SYMBOL SHIMA SIMA
    Prop<unicode::Script> { { 0x2CEB, 0x2CEE }, unicode::Script::Coptic }, // L&   [4] COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI..COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    Prop<unicode::Script> { { 0x2CEF, 0x2CF1 }, unicode::Script::Coptic }, // Mn   [3] COPTIC COMBINING NI ABOVE..COPTIC COMBINING SPIRITUS LENIS
    Prop<unicode::Script> { { 0x2CF2, 0x2CF3 }, unicode::Script::Coptic }, // L&   [2] COPTIC CAPITAL LETTER BOHAIRIC KHEI..COPTIC SMALL LETTER BOHAIRIC KHEI
    Prop<unicode::Script> { { 0x2CF9, 0x2CFC }, unicode::Script::Coptic }, // Po   [4] COPTIC OLD NUBIAN FULL STOP..COPTIC OLD NUBIAN VERSE DIVIDER
    Prop<unicode::Script> { { 0x2CFD, 0x2CFD }, unicode::Script::Coptic }, // No       COPTIC FRACTION ONE HALF
    Prop<unicode::Script> { { 0x2CFE, 0x2CFF }, unicode::Script::Coptic }, // Po   [2] COPTIC FULL STOP..COPTIC MORPHOLOGICAL DIVIDER
    Prop<unicode::Script> { { 0x2D00, 0x2D25 }, unicode::Script::Georgian }, // L&  [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Prop<unicode::Script> { { 0x2D27, 0x2D27 }, unicode::Script::Georgian }, // L&       GEORGIAN SMALL LETTER YN
    Prop<unicode::Script> { { 0x2D2D, 0x2D2D }, unicode::Script::Georgian }, // L&       GEORGIAN SMALL LETTER AEN
    Prop<unicode::Script> { { 0x2D30, 0x2D67 }, unicode::Script::Tifinagh }, // Lo  [56] TIFINAGH LETTER YA..TIFINAGH LETTER YO
    Prop<unicode::Script> { { 0x2D6F, 0x2D6F }, unicode::Script::Tifinagh }, // Lm       TIFINAGH MODIFIER LETTER LABIALIZATION MARK
    Prop<unicode::Script> { { 0x2D70, 0x2D70 }, unicode::Script::Tifinagh }, // Po       TIFINAGH SEPARATOR MARK
    Prop<unicode::Script> { { 0x2D7F, 0x2D7F }, unicode::Script::Tifinagh }, // Mn       TIFINAGH CONSONANT JOINER
    Prop<unicode::Script> { { 0x2D80, 0x2D96 }, unicode::Script::Ethiopic }, // Lo  [23] ETHIOPIC SYLLABLE LOA..ETHIOPIC SYLLABLE GGWE
    Prop<unicode::Script> { { 0x2DA0, 0x2DA6 }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE SSA..ETHIOPIC SYLLABLE SSO
    Prop<unicode::Script> { { 0x2DA8, 0x2DAE }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE CCA..ETHIOPIC SYLLABLE CCO
    Prop<unicode::Script> { { 0x2DB0, 0x2DB6 }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE ZZA..ETHIOPIC SYLLABLE ZZO
    Prop<unicode::Script> { { 0x2DB8, 0x2DBE }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE CCHA..ETHIOPIC SYLLABLE CCHO
    Prop<unicode::Script> { { 0x2DC0, 0x2DC6 }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE QYA..ETHIOPIC SYLLABLE QYO
    Prop<unicode::Script> { { 0x2DC8, 0x2DCE }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE KYA..ETHIOPIC SYLLABLE KYO
    Prop<unicode::Script> { { 0x2DD0, 0x2DD6 }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE XYA..ETHIOPIC SYLLABLE XYO
    Prop<unicode::Script> { { 0x2DD8, 0x2DDE }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE GYA..ETHIOPIC SYLLABLE GYO
    Prop<unicode::Script> { { 0x2DE0, 0x2DFF }, unicode::Script::Cyrillic }, // Mn  [32] COMBINING CYRILLIC LETTER BE..COMBINING CYRILLIC LETTER IOTIFIED BIG YUS
    Prop<unicode::Script> { { 0x2E00, 0x2E01 }, unicode::Script::Common }, // Po   [2] RIGHT ANGLE SUBSTITUTION MARKER..RIGHT ANGLE DOTTED SUBSTITUTION MARKER
    Prop<unicode::Script> { { 0x2E02, 0x2E02 }, unicode::Script::Common }, // Pi       LEFT SUBSTITUTION BRACKET
    Prop<unicode::Script> { { 0x2E03, 0x2E03 }, unicode::Script::Common }, // Pf       RIGHT SUBSTITUTION BRACKET
    Prop<unicode::Script> { { 0x2E04, 0x2E04 }, unicode::Script::Common }, // Pi       LEFT DOTTED SUBSTITUTION BRACKET
    Prop<unicode::Script> { { 0x2E05, 0x2E05 }, unicode::Script::Common }, // Pf       RIGHT DOTTED SUBSTITUTION BRACKET
    Prop<unicode::Script> { { 0x2E06, 0x2E08 }, unicode::Script::Common }, // Po   [3] RAISED INTERPOLATION MARKER..DOTTED TRANSPOSITION MARKER
    Prop<unicode::Script> { { 0x2E09, 0x2E09 }, unicode::Script::Common }, // Pi       LEFT TRANSPOSITION BRACKET
    Prop<unicode::Script> { { 0x2E0A, 0x2E0A }, unicode::Script::Common }, // Pf       RIGHT TRANSPOSITION BRACKET
    Prop<unicode::Script> { { 0x2E0B, 0x2E0B }, unicode::Script::Common }, // Po       RAISED SQUARE
    Prop<unicode::Script> { { 0x2E0C, 0x2E0C }, unicode::Script::Common }, // Pi       LEFT RAISED OMISSION BRACKET
    Prop<unicode::Script> { { 0x2E0D, 0x2E0D }, unicode::Script::Common }, // Pf       RIGHT RAISED OMISSION BRACKET
    Prop<unicode::Script> { { 0x2E0E, 0x2E16 }, unicode::Script::Common }, // Po   [9] EDITORIAL CORONIS..DOTTED RIGHT-POINTING ANGLE
    Prop<unicode::Script> { { 0x2E17, 0x2E17 }, unicode::Script::Common }, // Pd       DOUBLE OBLIQUE HYPHEN
    Prop<unicode::Script> { { 0x2E18, 0x2E19 }, unicode::Script::Common }, // Po   [2] INVERTED INTERROBANG..PALM BRANCH
    Prop<unicode::Script> { { 0x2E1A, 0x2E1A }, unicode::Script::Common }, // Pd       HYPHEN WITH DIAERESIS
    Prop<unicode::Script> { { 0x2E1B, 0x2E1B }, unicode::Script::Common }, // Po       TILDE WITH RING ABOVE
    Prop<unicode::Script> { { 0x2E1C, 0x2E1C }, unicode::Script::Common }, // Pi       LEFT LOW PARAPHRASE BRACKET
    Prop<unicode::Script> { { 0x2E1D, 0x2E1D }, unicode::Script::Common }, // Pf       RIGHT LOW PARAPHRASE BRACKET
    Prop<unicode::Script> { { 0x2E1E, 0x2E1F }, unicode::Script::Common }, // Po   [2] TILDE WITH DOT ABOVE..TILDE WITH DOT BELOW
    Prop<unicode::Script> { { 0x2E20, 0x2E20 }, unicode::Script::Common }, // Pi       LEFT VERTICAL BAR WITH QUILL
    Prop<unicode::Script> { { 0x2E21, 0x2E21 }, unicode::Script::Common }, // Pf       RIGHT VERTICAL BAR WITH QUILL
    Prop<unicode::Script> { { 0x2E22, 0x2E22 }, unicode::Script::Common }, // Ps       TOP LEFT HALF BRACKET
    Prop<unicode::Script> { { 0x2E23, 0x2E23 }, unicode::Script::Common }, // Pe       TOP RIGHT HALF BRACKET
    Prop<unicode::Script> { { 0x2E24, 0x2E24 }, unicode::Script::Common }, // Ps       BOTTOM LEFT HALF BRACKET
    Prop<unicode::Script> { { 0x2E25, 0x2E25 }, unicode::Script::Common }, // Pe       BOTTOM RIGHT HALF BRACKET
    Prop<unicode::Script> { { 0x2E26, 0x2E26 }, unicode::Script::Common }, // Ps       LEFT SIDEWAYS U BRACKET
    Prop<unicode::Script> { { 0x2E27, 0x2E27 }, unicode::Script::Common }, // Pe       RIGHT SIDEWAYS U BRACKET
    Prop<unicode::Script> { { 0x2E28, 0x2E28 }, unicode::Script::Common }, // Ps       LEFT DOUBLE PARENTHESIS
    Prop<unicode::Script> { { 0x2E29, 0x2E29 }, unicode::Script::Common }, // Pe       RIGHT DOUBLE PARENTHESIS
    Prop<unicode::Script> { { 0x2E2A, 0x2E2E }, unicode::Script::Common }, // Po   [5] TWO DOTS OVER ONE DOT PUNCTUATION..REVERSED QUESTION MARK
    Prop<unicode::Script> { { 0x2E2F, 0x2E2F }, unicode::Script::Common }, // Lm       VERTICAL TILDE
    Prop<unicode::Script> { { 0x2E30, 0x2E39 }, unicode::Script::Common }, // Po  [10] RING POINT..TOP HALF SECTION SIGN
    Prop<unicode::Script> { { 0x2E3A, 0x2E3B }, unicode::Script::Common }, // Pd   [2] TWO-EM DASH..THREE-EM DASH
    Prop<unicode::Script> { { 0x2E3C, 0x2E3F }, unicode::Script::Common }, // Po   [4] STENOGRAPHIC FULL STOP..CAPITULUM
    Prop<unicode::Script> { { 0x2E40, 0x2E40 }, unicode::Script::Common }, // Pd       DOUBLE HYPHEN
    Prop<unicode::Script> { { 0x2E41, 0x2E41 }, unicode::Script::Common }, // Po       REVERSED COMMA
    Prop<unicode::Script> { { 0x2E42, 0x2E42 }, unicode::Script::Common }, // Ps       DOUBLE LOW-REVERSED-9 QUOTATION MARK
    Prop<unicode::Script> { { 0x2E43, 0x2E4F }, unicode::Script::Common }, // Po  [13] DASH WITH LEFT UPTURN..CORNISH VERSE DIVIDER
    Prop<unicode::Script> { { 0x2E50, 0x2E51 }, unicode::Script::Common }, // So   [2] CROSS PATTY WITH RIGHT CROSSBAR..CROSS PATTY WITH LEFT CROSSBAR
    Prop<unicode::Script> { { 0x2E52, 0x2E54 }, unicode::Script::Common }, // Po   [3] TIRONIAN SIGN CAPITAL ET..MEDIEVAL QUESTION MARK
    Prop<unicode::Script> { { 0x2E55, 0x2E55 }, unicode::Script::Common }, // Ps       LEFT SQUARE BRACKET WITH STROKE
    Prop<unicode::Script> { { 0x2E56, 0x2E56 }, unicode::Script::Common }, // Pe       RIGHT SQUARE BRACKET WITH STROKE
    Prop<unicode::Script> { { 0x2E57, 0x2E57 }, unicode::Script::Common }, // Ps       LEFT SQUARE BRACKET WITH DOUBLE STROKE
    Prop<unicode::Script> { { 0x2E58, 0x2E58 }, unicode::Script::Common }, // Pe       RIGHT SQUARE BRACKET WITH DOUBLE STROKE
    Prop<unicode::Script> { { 0x2E59, 0x2E59 }, unicode::Script::Common }, // Ps       TOP HALF LEFT PARENTHESIS
    Prop<unicode::Script> { { 0x2E5A, 0x2E5A }, unicode::Script::Common }, // Pe       TOP HALF RIGHT PARENTHESIS
    Prop<unicode::Script> { { 0x2E5B, 0x2E5B }, unicode::Script::Common }, // Ps       BOTTOM HALF LEFT PARENTHESIS
    Prop<unicode::Script> { { 0x2E5C, 0x2E5C }, unicode::Script::Common }, // Pe       BOTTOM HALF RIGHT PARENTHESIS
    Prop<unicode::Script> { { 0x2E5D, 0x2E5D }, unicode::Script::Common }, // Pd       OBLIQUE HYPHEN
    Prop<unicode::Script> { { 0x2E80, 0x2E99 }, unicode::Script::Han }, // So  [26] CJK RADICAL REPEAT..CJK RADICAL RAP
    Prop<unicode::Script> { { 0x2E9B, 0x2EF3 }, unicode::Script::Han }, // So  [89] CJK RADICAL CHOKE..CJK RADICAL C-SIMPLIFIED TURTLE
    Prop<unicode::Script> { { 0x2F00, 0x2FD5 }, unicode::Script::Han }, // So [214] KANGXI RADICAL ONE..KANGXI RADICAL FLUTE
    Prop<unicode::Script> { { 0x2FF0, 0x2FFF }, unicode::Script::Common }, // So  [16] IDEOGRAPHIC DESCRIPTION CHARACTER LEFT TO RIGHT..IDEOGRAPHIC DESCRIPTION CHARACTER ROTATION
    Prop<unicode::Script> { { 0x3000, 0x3000 }, unicode::Script::Common }, // Zs       IDEOGRAPHIC SPACE
    Prop<unicode::Script> { { 0x3001, 0x3003 }, unicode::Script::Common }, // Po   [3] IDEOGRAPHIC COMMA..DITTO MARK
    Prop<unicode::Script> { { 0x3004, 0x3004 }, unicode::Script::Common }, // So       JAPANESE INDUSTRIAL STANDARD SYMBOL
    Prop<unicode::Script> { { 0x3005, 0x3005 }, unicode::Script::Han }, // Lm       IDEOGRAPHIC ITERATION MARK
    Prop<unicode::Script> { { 0x3006, 0x3006 }, unicode::Script::Common }, // Lo       IDEOGRAPHIC CLOSING MARK
    Prop<unicode::Script> { { 0x3007, 0x3007 }, unicode::Script::Han }, // Nl       IDEOGRAPHIC NUMBER ZERO
    Prop<unicode::Script> { { 0x3008, 0x3008 }, unicode::Script::Common }, // Ps       LEFT ANGLE BRACKET
    Prop<unicode::Script> { { 0x3009, 0x3009 }, unicode::Script::Common }, // Pe       RIGHT ANGLE BRACKET
    Prop<unicode::Script> { { 0x300A, 0x300A }, unicode::Script::Common }, // Ps       LEFT DOUBLE ANGLE BRACKET
    Prop<unicode::Script> { { 0x300B, 0x300B }, unicode::Script::Common }, // Pe       RIGHT DOUBLE ANGLE BRACKET
    Prop<unicode::Script> { { 0x300C, 0x300C }, unicode::Script::Common }, // Ps       LEFT CORNER BRACKET
    Prop<unicode::Script> { { 0x300D, 0x300D }, unicode::Script::Common }, // Pe       RIGHT CORNER BRACKET
    Prop<unicode::Script> { { 0x300E, 0x300E }, unicode::Script::Common }, // Ps       LEFT WHITE CORNER BRACKET
    Prop<unicode::Script> { { 0x300F, 0x300F }, unicode::Script::Common }, // Pe       RIGHT WHITE CORNER BRACKET
    Prop<unicode::Script> { { 0x3010, 0x3010 }, unicode::Script::Common }, // Ps       LEFT BLACK LENTICULAR BRACKET
    Prop<unicode::Script> { { 0x3011, 0x3011 }, unicode::Script::Common }, // Pe       RIGHT BLACK LENTICULAR BRACKET
    Prop<unicode::Script> { { 0x3012, 0x3013 }, unicode::Script::Common }, // So   [2] POSTAL MARK..GETA MARK
    Prop<unicode::Script> { { 0x3014, 0x3014 }, unicode::Script::Common }, // Ps       LEFT TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0x3015, 0x3015 }, unicode::Script::Common }, // Pe       RIGHT TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0x3016, 0x3016 }, unicode::Script::Common }, // Ps       LEFT WHITE LENTICULAR BRACKET
    Prop<unicode::Script> { { 0x3017, 0x3017 }, unicode::Script::Common }, // Pe       RIGHT WHITE LENTICULAR BRACKET
    Prop<unicode::Script> { { 0x3018, 0x3018 }, unicode::Script::Common }, // Ps       LEFT WHITE TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0x3019, 0x3019 }, unicode::Script::Common }, // Pe       RIGHT WHITE TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0x301A, 0x301A }, unicode::Script::Common }, // Ps       LEFT WHITE SQUARE BRACKET
    Prop<unicode::Script> { { 0x301B, 0x301B }, unicode::Script::Common }, // Pe       RIGHT WHITE SQUARE BRACKET
    Prop<unicode::Script> { { 0x301C, 0x301C }, unicode::Script::Common }, // Pd       WAVE DASH
    Prop<unicode::Script> { { 0x301D, 0x301D }, unicode::Script::Common }, // Ps       REVERSED DOUBLE PRIME QUOTATION MARK
    Prop<unicode::Script> { { 0x301E, 0x301F }, unicode::Script::Common }, // Pe   [2] DOUBLE PRIME QUOTATION MARK..LOW DOUBLE PRIME QUOTATION MARK
    Prop<unicode::Script> { { 0x3020, 0x3020 }, unicode::Script::Common }, // So       POSTAL MARK FACE
    Prop<unicode::Script> { { 0x3021, 0x3029 }, unicode::Script::Han }, // Nl   [9] HANGZHOU NUMERAL ONE..HANGZHOU NUMERAL NINE
    Prop<unicode::Script> { { 0x302A, 0x302D }, unicode::Script::Inherited }, // Mn   [4] IDEOGRAPHIC LEVEL TONE MARK..IDEOGRAPHIC ENTERING TONE MARK
    Prop<unicode::Script> { { 0x302E, 0x302F }, unicode::Script::Hangul }, // Mc   [2] HANGUL SINGLE DOT TONE MARK..HANGUL DOUBLE DOT TONE MARK
    Prop<unicode::Script> { { 0x3030, 0x3030 }, unicode::Script::Common }, // Pd       WAVY DASH
    Prop<unicode::Script> { { 0x3031, 0x3035 }, unicode::Script::Common }, // Lm   [5] VERTICAL KANA REPEAT MARK..VERTICAL KANA REPEAT MARK LOWER HALF
    Prop<unicode::Script> { { 0x3036, 0x3037 }, unicode::Script::Common }, // So   [2] CIRCLED POSTAL MARK..IDEOGRAPHIC TELEGRAPH LINE FEED SEPARATOR SYMBOL
    Prop<unicode::Script> { { 0x3038, 0x303A }, unicode::Script::Han }, // Nl   [3] HANGZHOU NUMERAL TEN..HANGZHOU NUMERAL THIRTY
    Prop<unicode::Script> { { 0x303B, 0x303B }, unicode::Script::Han }, // Lm       VERTICAL IDEOGRAPHIC ITERATION MARK
    Prop<unicode::Script> { { 0x303C, 0x303C }, unicode::Script::Common }, // Lo       MASU MARK
    Prop<unicode::Script> { { 0x303D, 0x303D }, unicode::Script::Common }, // Po       PART ALTERNATION MARK
    Prop<unicode::Script> { { 0x303E, 0x303F }, unicode::Script::Common }, // So   [2] IDEOGRAPHIC VARIATION INDICATOR..IDEOGRAPHIC HALF FILL SPACE
    Prop<unicode::Script> { { 0x3041, 0x3096 }, unicode::Script::Hiragana }, // Lo  [86] HIRAGANA LETTER SMALL A..HIRAGANA LETTER SMALL KE
    Prop<unicode::Script> { { 0x3099, 0x309A }, unicode::Script::Inherited }, // Mn   [2] COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK..COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Prop<unicode::Script> { { 0x309B, 0x309C }, unicode::Script::Common }, // Sk   [2] KATAKANA-HIRAGANA VOICED SOUND MARK..KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Prop<unicode::Script> { { 0x309D, 0x309E }, unicode::Script::Hiragana }, // Lm   [2] HIRAGANA ITERATION MARK..HIRAGANA VOICED ITERATION MARK
    Prop<unicode::Script> { { 0x309F, 0x309F }, unicode::Script::Hiragana }, // Lo       HIRAGANA DIGRAPH YORI
    Prop<unicode::Script> { { 0x30A0, 0x30A0 }, unicode::Script::Common }, // Pd       KATAKANA-HIRAGANA DOUBLE HYPHEN
    Prop<unicode::Script> { { 0x30A1, 0x30FA }, unicode::Script::Katakana }, // Lo  [90] KATAKANA LETTER SMALL A..KATAKANA LETTER VO
    Prop<unicode::Script> { { 0x30FB, 0x30FB }, unicode::Script::Common }, // Po       KATAKANA MIDDLE DOT
    Prop<unicode::Script> { { 0x30FC, 0x30FC }, unicode::Script::Common }, // Lm       KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Prop<unicode::Script> { { 0x30FD, 0x30FE }, unicode::Script::Katakana }, // Lm   [2] KATAKANA ITERATION MARK..KATAKANA VOICED ITERATION MARK
    Prop<unicode::Script> { { 0x30FF, 0x30FF }, unicode::Script::Katakana }, // Lo       KATAKANA DIGRAPH KOTO
    Prop<unicode::Script> { { 0x3105, 0x312F }, unicode::Script::Bopomofo }, // Lo  [43] BOPOMOFO LETTER B..BOPOMOFO LETTER NN
    Prop<unicode::Script> { { 0x3131, 0x318E }, unicode::Script::Hangul }, // Lo  [94] HANGUL LETTER KIYEOK..HANGUL LETTER ARAEAE
    Prop<unicode::Script> { { 0x3190, 0x3191 }, unicode::Script::Common }, // So   [2] IDEOGRAPHIC ANNOTATION LINKING MARK..IDEOGRAPHIC ANNOTATION REVERSE MARK
    Prop<unicode::Script> { { 0x3192, 0x3195 }, unicode::Script::Common }, // No   [4] IDEOGRAPHIC ANNOTATION ONE MARK..IDEOGRAPHIC ANNOTATION FOUR MARK
    Prop<unicode::Script> { { 0x3196, 0x319F }, unicode::Script::Common }, // So  [10] IDEOGRAPHIC ANNOTATION TOP MARK..IDEOGRAPHIC ANNOTATION MAN MARK
    Prop<unicode::Script> { { 0x31A0, 0x31BF }, unicode::Script::Bopomofo }, // Lo  [32] BOPOMOFO LETTER BU..BOPOMOFO LETTER AH
    Prop<unicode::Script> { { 0x31C0, 0x31E5 }, unicode::Script::Common }, // So  [38] CJK STROKE T..CJK STROKE SZP
    Prop<unicode::Script> { { 0x31EF, 0x31EF }, unicode::Script::Common }, // So       IDEOGRAPHIC DESCRIPTION CHARACTER SUBTRACTION
    Prop<unicode::Script> { { 0x31F0, 0x31FF }, unicode::Script::Katakana }, // Lo  [16] KATAKANA LETTER SMALL KU..KATAKANA LETTER SMALL RO
    Prop<unicode::Script> { { 0x3200, 0x321E }, unicode::Script::Hangul }, // So  [31] PARENTHESIZED HANGUL KIYEOK..PARENTHESIZED KOREAN CHARACTER O HU
    Prop<unicode::Script> { { 0x3220, 0x3229 }, unicode::Script::Common }, // No  [10] PARENTHESIZED IDEOGRAPH ONE..PARENTHESIZED IDEOGRAPH TEN
    Prop<unicode::Script> { { 0x322A, 0x3247 }, unicode::Script::Common }, // So  [30] PARENTHESIZED IDEOGRAPH MOON..CIRCLED IDEOGRAPH KOTO
    Prop<unicode::Script> { { 0x3248, 0x324F }, unicode::Script::Common }, // No   [8] CIRCLED NUMBER TEN ON BLACK SQUARE..CIRCLED NUMBER EIGHTY ON BLACK SQUARE
    Prop<unicode::Script> { { 0x3250, 0x3250 }, unicode::Script::Common }, // So       PARTNERSHIP SIGN
    Prop<unicode::Script> { { 0x3251, 0x325F }, unicode::Script::Common }, // No  [15] CIRCLED NUMBER TWENTY ONE..CIRCLED NUMBER THIRTY FIVE
    Prop<unicode::Script> { { 0x3260, 0x327E }, unicode::Script::Hangul }, // So  [31] CIRCLED HANGUL KIYEOK..CIRCLED HANGUL IEUNG U
    Prop<unicode::Script> { { 0x327F, 0x327F }, unicode::Script::Common }, // So       KOREAN STANDARD SYMBOL
    Prop<unicode::Script> { { 0x3280, 0x3289 }, unicode::Script::Common }, // No  [10] CIRCLED IDEOGRAPH ONE..CIRCLED IDEOGRAPH TEN
    Prop<unicode::Script> { { 0x328A, 0x32B0 }, unicode::Script::Common }, // So  [39] CIRCLED IDEOGRAPH MOON..CIRCLED IDEOGRAPH NIGHT
    Prop<unicode::Script> { { 0x32B1, 0x32BF }, unicode::Script::Common }, // No  [15] CIRCLED NUMBER THIRTY SIX..CIRCLED NUMBER FIFTY
    Prop<unicode::Script> { { 0x32C0, 0x32CF }, unicode::Script::Common }, // So  [16] IDEOGRAPHIC TELEGRAPH SYMBOL FOR JANUARY..LIMITED LIABILITY SIGN
    Prop<unicode::Script> { { 0x32D0, 0x32FE }, unicode::Script::Katakana }, // So  [47] CIRCLED KATAKANA A..CIRCLED KATAKANA WO
    Prop<unicode::Script> { { 0x32FF, 0x32FF }, unicode::Script::Common }, // So       SQUARE ERA NAME REIWA
    Prop<unicode::Script> { { 0x3300, 0x3357 }, unicode::Script::Katakana }, // So  [88] SQUARE APAATO..SQUARE WATTO
    Prop<unicode::Script> { { 0x3358, 0x33FF }, unicode::Script::Common }, // So [168] IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR ZERO..SQUARE GAL
    Prop<unicode::Script> { { 0x3400, 0x4DBF }, unicode::Script::Han }, // Lo [6592] CJK UNIFIED IDEOGRAPH-3400..CJK UNIFIED IDEOGRAPH-4DBF
    Prop<unicode::Script> { { 0x4DC0, 0x4DFF }, unicode::Script::Common }, // So  [64] HEXAGRAM FOR THE CREATIVE HEAVEN..HEXAGRAM FOR BEFORE COMPLETION
    Prop<unicode::Script> { { 0x4E00, 0x9FFF }, unicode::Script::Han }, // Lo [20992] CJK UNIFIED IDEOGRAPH-4E00..CJK UNIFIED IDEOGRAPH-9FFF
    Prop<unicode::Script> { { 0xA000, 0xA014 }, unicode::Script::Yi }, // Lo  [21] YI SYLLABLE IT..YI SYLLABLE E
    Prop<unicode::Script> { { 0xA015, 0xA015 }, unicode::Script::Yi }, // Lm       YI SYLLABLE WU
    Prop<unicode::Script> { { 0xA016, 0xA48C }, unicode::Script::Yi }, // Lo [1143] YI SYLLABLE BIT..YI SYLLABLE YYR
    Prop<unicode::Script> { { 0xA490, 0xA4C6 }, unicode::Script::Yi }, // So  [55] YI RADICAL QOT..YI RADICAL KE
    Prop<unicode::Script> { { 0xA4D0, 0xA4F7 }, unicode::Script::Lisu }, // Lo  [40] LISU LETTER BA..LISU LETTER OE
    Prop<unicode::Script> { { 0xA4F8, 0xA4FD }, unicode::Script::Lisu }, // Lm   [6] LISU LETTER TONE MYA TI..LISU LETTER TONE MYA JEU
    Prop<unicode::Script> { { 0xA4FE, 0xA4FF }, unicode::Script::Lisu }, // Po   [2] LISU PUNCTUATION COMMA..LISU PUNCTUATION FULL STOP
    Prop<unicode::Script> { { 0xA500, 0xA60B }, unicode::Script::Vai }, // Lo [268] VAI SYLLABLE EE..VAI SYLLABLE NG
    Prop<unicode::Script> { { 0xA60C, 0xA60C }, unicode::Script::Vai }, // Lm       VAI SYLLABLE LENGTHENER
    Prop<unicode::Script> { { 0xA60D, 0xA60F }, unicode::Script::Vai }, // Po   [3] VAI COMMA..VAI QUESTION MARK
    Prop<unicode::Script> { { 0xA610, 0xA61F }, unicode::Script::Vai }, // Lo  [16] VAI SYLLABLE NDOLE FA..VAI SYMBOL JONG
    Prop<unicode::Script> { { 0xA620, 0xA629 }, unicode::Script::Vai }, // Nd  [10] VAI DIGIT ZERO..VAI DIGIT NINE
    Prop<unicode::Script> { { 0xA62A, 0xA62B }, unicode::Script::Vai }, // Lo   [2] VAI SYLLABLE NDOLE MA..VAI SYLLABLE NDOLE DO
    Prop<unicode::Script> { { 0xA640, 0xA66D }, unicode::Script::Cyrillic }, // L&  [46] CYRILLIC CAPITAL LETTER ZEMLYA..CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    Prop<unicode::Script> { { 0xA66E, 0xA66E }, unicode::Script::Cyrillic }, // Lo       CYRILLIC LETTER MULTIOCULAR O
    Prop<unicode::Script> { { 0xA66F, 0xA66F }, unicode::Script::Cyrillic }, // Mn       COMBINING CYRILLIC VZMET
    Prop<unicode::Script> { { 0xA670, 0xA672 }, unicode::Script::Cyrillic }, // Me   [3] COMBINING CYRILLIC TEN MILLIONS SIGN..COMBINING CYRILLIC THOUSAND MILLIONS SIGN
    Prop<unicode::Script> { { 0xA673, 0xA673 }, unicode::Script::Cyrillic }, // Po       SLAVONIC ASTERISK
    Prop<unicode::Script> { { 0xA674, 0xA67D }, unicode::Script::Cyrillic }, // Mn  [10] COMBINING CYRILLIC LETTER UKRAINIAN IE..COMBINING CYRILLIC PAYEROK
    Prop<unicode::Script> { { 0xA67E, 0xA67E }, unicode::Script::Cyrillic }, // Po       CYRILLIC KAVYKA
    Prop<unicode::Script> { { 0xA67F, 0xA67F }, unicode::Script::Cyrillic }, // Lm       CYRILLIC PAYEROK
    Prop<unicode::Script> { { 0xA680, 0xA69B }, unicode::Script::Cyrillic }, // L&  [28] CYRILLIC CAPITAL LETTER DWE..CYRILLIC SMALL LETTER CROSSED O
    Prop<unicode::Script> { { 0xA69C, 0xA69D }, unicode::Script::Cyrillic }, // Lm   [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    Prop<unicode::Script> { { 0xA69E, 0xA69F }, unicode::Script::Cyrillic }, // Mn   [2] COMBINING CYRILLIC LETTER EF..COMBINING CYRILLIC LETTER IOTIFIED E
    Prop<unicode::Script> { { 0xA6A0, 0xA6E5 }, unicode::Script::Bamum }, // Lo  [70] BAMUM LETTER A..BAMUM LETTER KI
    Prop<unicode::Script> { { 0xA6E6, 0xA6EF }, unicode::Script::Bamum }, // Nl  [10] BAMUM LETTER MO..BAMUM LETTER KOGHOM
    Prop<unicode::Script> { { 0xA6F0, 0xA6F1 }, unicode::Script::Bamum }, // Mn   [2] BAMUM COMBINING MARK KOQNDON..BAMUM COMBINING MARK TUKWENTIS
    Prop<unicode::Script> { { 0xA6F2, 0xA6F7 }, unicode::Script::Bamum }, // Po   [6] BAMUM NJAEMLI..BAMUM QUESTION MARK
    Prop<unicode::Script> { { 0xA700, 0xA716 }, unicode::Script::Common }, // Sk  [23] MODIFIER LETTER CHINESE TONE YIN PING..MODIFIER LETTER EXTRA-LOW LEFT-STEM TONE BAR
    Prop<unicode::Script> { { 0xA717, 0xA71F }, unicode::Script::Common }, // Lm   [9] MODIFIER LETTER DOT VERTICAL BAR..MODIFIER LETTER LOW INVERTED EXCLAMATION MARK
    Prop<unicode::Script> { { 0xA720, 0xA721 }, unicode::Script::Common }, // Sk   [2] MODIFIER LETTER STRESS AND HIGH TONE..MODIFIER LETTER STRESS AND LOW TONE
    Prop<unicode::Script> { { 0xA722, 0xA76F }, unicode::Script::Latin }, // L&  [78] LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF..LATIN SMALL LETTER CON
    Prop<unicode::Script> { { 0xA770, 0xA770 }, unicode::Script::Latin }, // Lm       MODIFIER LETTER US
    Prop<unicode::Script> { { 0xA771, 0xA787 }, unicode::Script::Latin }, // L&  [23] LATIN SMALL LETTER DUM..LATIN SMALL LETTER INSULAR T
    Prop<unicode::Script> { { 0xA788, 0xA788 }, unicode::Script::Common }, // Lm       MODIFIER LETTER LOW CIRCUMFLEX ACCENT
    Prop<unicode::Script> { { 0xA789, 0xA78A }, unicode::Script::Common }, // Sk   [2] MODIFIER LETTER COLON..MODIFIER LETTER SHORT EQUALS SIGN
    Prop<unicode::Script> { { 0xA78B, 0xA78E }, unicode::Script::Latin }, // L&   [4] LATIN CAPITAL LETTER SALTILLO..LATIN SMALL LETTER L WITH RETROFLEX HOOK AND BELT
    Prop<unicode::Script> { { 0xA78F, 0xA78F }, unicode::Script::Latin }, // Lo       LATIN LETTER SINOLOGICAL DOT
    Prop<unicode::Script> { { 0xA790, 0xA7DC }, unicode::Script::Latin }, // L&  [77] LATIN CAPITAL LETTER N WITH DESCENDER..LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Prop<unicode::Script> { { 0xA7F1, 0xA7F4 }, unicode::Script::Latin }, // Lm   [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    Prop<unicode::Script> { { 0xA7F5, 0xA7F6 }, unicode::Script::Latin }, // L&   [2] LATIN CAPITAL LETTER REVERSED HALF H..LATIN SMALL LETTER REVERSED HALF H
    Prop<unicode::Script> { { 0xA7F7, 0xA7F7 }, unicode::Script::Latin }, // Lo       LATIN EPIGRAPHIC LETTER SIDEWAYS I
    Prop<unicode::Script> { { 0xA7F8, 0xA7F9 }, unicode::Script::Latin }, // Lm   [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    Prop<unicode::Script> { { 0xA7FA, 0xA7FA }, unicode::Script::Latin }, // L&       LATIN LETTER SMALL CAPITAL TURNED M
    Prop<unicode::Script> { { 0xA7FB, 0xA7FF }, unicode::Script::Latin }, // Lo   [5] LATIN EPIGRAPHIC LETTER REVERSED F..LATIN EPIGRAPHIC LETTER ARCHAIC M
    Prop<unicode::Script> { { 0xA800, 0xA801 }, unicode::Script::Syloti_Nagri }, // Lo   [2] SYLOTI NAGRI LETTER A..SYLOTI NAGRI LETTER I
    Prop<unicode::Script> { { 0xA802, 0xA802 }, unicode::Script::Syloti_Nagri }, // Mn       SYLOTI NAGRI SIGN DVISVARA
    Prop<unicode::Script> { { 0xA803, 0xA805 }, unicode::Script::Syloti_Nagri }, // Lo   [3] SYLOTI NAGRI LETTER U..SYLOTI NAGRI LETTER O
    Prop<unicode::Script> { { 0xA806, 0xA806 }, unicode::Script::Syloti_Nagri }, // Mn       SYLOTI NAGRI SIGN HASANTA
    Prop<unicode::Script> { { 0xA807, 0xA80A }, unicode::Script::Syloti_Nagri }, // Lo   [4] SYLOTI NAGRI LETTER KO..SYLOTI NAGRI LETTER GHO
    Prop<unicode::Script> { { 0xA80B, 0xA80B }, unicode::Script::Syloti_Nagri }, // Mn       SYLOTI NAGRI SIGN ANUSVARA
    Prop<unicode::Script> { { 0xA80C, 0xA822 }, unicode::Script::Syloti_Nagri }, // Lo  [23] SYLOTI NAGRI LETTER CO..SYLOTI NAGRI LETTER HO
    Prop<unicode::Script> { { 0xA823, 0xA824 }, unicode::Script::Syloti_Nagri }, // Mc   [2] SYLOTI NAGRI VOWEL SIGN A..SYLOTI NAGRI VOWEL SIGN I
    Prop<unicode::Script> { { 0xA825, 0xA826 }, unicode::Script::Syloti_Nagri }, // Mn   [2] SYLOTI NAGRI VOWEL SIGN U..SYLOTI NAGRI VOWEL SIGN E
    Prop<unicode::Script> { { 0xA827, 0xA827 }, unicode::Script::Syloti_Nagri }, // Mc       SYLOTI NAGRI VOWEL SIGN OO
    Prop<unicode::Script> { { 0xA828, 0xA82B }, unicode::Script::Syloti_Nagri }, // So   [4] SYLOTI NAGRI POETRY MARK-1..SYLOTI NAGRI POETRY MARK-4
    Prop<unicode::Script> { { 0xA82C, 0xA82C }, unicode::Script::Syloti_Nagri }, // Mn       SYLOTI NAGRI SIGN ALTERNATE HASANTA
    Prop<unicode::Script> { { 0xA830, 0xA835 }, unicode::Script::Common }, // No   [6] NORTH INDIC FRACTION ONE QUARTER..NORTH INDIC FRACTION THREE SIXTEENTHS
    Prop<unicode::Script> { { 0xA836, 0xA837 }, unicode::Script::Common }, // So   [2] NORTH INDIC QUARTER MARK..NORTH INDIC PLACEHOLDER MARK
    Prop<unicode::Script> { { 0xA838, 0xA838 }, unicode::Script::Common }, // Sc       NORTH INDIC RUPEE MARK
    Prop<unicode::Script> { { 0xA839, 0xA839 }, unicode::Script::Common }, // So       NORTH INDIC QUANTITY MARK
    Prop<unicode::Script> { { 0xA840, 0xA873 }, unicode::Script::Phags_Pa }, // Lo  [52] PHAGS-PA LETTER KA..PHAGS-PA LETTER CANDRABINDU
    Prop<unicode::Script> { { 0xA874, 0xA877 }, unicode::Script::Phags_Pa }, // Po   [4] PHAGS-PA SINGLE HEAD MARK..PHAGS-PA MARK DOUBLE SHAD
    Prop<unicode::Script> { { 0xA880, 0xA881 }, unicode::Script::Saurashtra }, // Mc   [2] SAURASHTRA SIGN ANUSVARA..SAURASHTRA SIGN VISARGA
    Prop<unicode::Script> { { 0xA882, 0xA8B3 }, unicode::Script::Saurashtra }, // Lo  [50] SAURASHTRA LETTER A..SAURASHTRA LETTER LLA
    Prop<unicode::Script> { { 0xA8B4, 0xA8C3 }, unicode::Script::Saurashtra }, // Mc  [16] SAURASHTRA CONSONANT SIGN HAARU..SAURASHTRA VOWEL SIGN AU
    Prop<unicode::Script> { { 0xA8C4, 0xA8C5 }, unicode::Script::Saurashtra }, // Mn   [2] SAURASHTRA SIGN VIRAMA..SAURASHTRA SIGN CANDRABINDU
    Prop<unicode::Script> { { 0xA8CE, 0xA8CF }, unicode::Script::Saurashtra }, // Po   [2] SAURASHTRA DANDA..SAURASHTRA DOUBLE DANDA
    Prop<unicode::Script> { { 0xA8D0, 0xA8D9 }, unicode::Script::Saurashtra }, // Nd  [10] SAURASHTRA DIGIT ZERO..SAURASHTRA DIGIT NINE
    Prop<unicode::Script> { { 0xA8E0, 0xA8F1 }, unicode::Script::Devanagari }, // Mn  [18] COMBINING DEVANAGARI DIGIT ZERO..COMBINING DEVANAGARI SIGN AVAGRAHA
    Prop<unicode::Script> { { 0xA8F2, 0xA8F7 }, unicode::Script::Devanagari }, // Lo   [6] DEVANAGARI SIGN SPACING CANDRABINDU..DEVANAGARI SIGN CANDRABINDU AVAGRAHA
    Prop<unicode::Script> { { 0xA8F8, 0xA8FA }, unicode::Script::Devanagari }, // Po   [3] DEVANAGARI SIGN PUSHPIKA..DEVANAGARI CARET
    Prop<unicode::Script> { { 0xA8FB, 0xA8FB }, unicode::Script::Devanagari }, // Lo       DEVANAGARI HEADSTROKE
    Prop<unicode::Script> { { 0xA8FC, 0xA8FC }, unicode::Script::Devanagari }, // Po       DEVANAGARI SIGN SIDDHAM
    Prop<unicode::Script> { { 0xA8FD, 0xA8FE }, unicode::Script::Devanagari }, // Lo   [2] DEVANAGARI JAIN OM..DEVANAGARI LETTER AY
    Prop<unicode::Script> { { 0xA8FF, 0xA8FF }, unicode::Script::Devanagari }, // Mn       DEVANAGARI VOWEL SIGN AY
    Prop<unicode::Script> { { 0xA900, 0xA909 }, unicode::Script::Kayah_Li }, // Nd  [10] KAYAH LI DIGIT ZERO..KAYAH LI DIGIT NINE
    Prop<unicode::Script> { { 0xA90A, 0xA925 }, unicode::Script::Kayah_Li }, // Lo  [28] KAYAH LI LETTER KA..KAYAH LI LETTER OO
    Prop<unicode::Script> { { 0xA926, 0xA92D }, unicode::Script::Kayah_Li }, // Mn   [8] KAYAH LI VOWEL UE..KAYAH LI TONE CALYA PLOPHU
    Prop<unicode::Script> { { 0xA92E, 0xA92E }, unicode::Script::Common }, // Po       KAYAH LI SIGN CWI
    Prop<unicode::Script> { { 0xA92F, 0xA92F }, unicode::Script::Kayah_Li }, // Po       KAYAH LI SIGN SHYA
    Prop<unicode::Script> { { 0xA930, 0xA946 }, unicode::Script::Rejang }, // Lo  [23] REJANG LETTER KA..REJANG LETTER A
    Prop<unicode::Script> { { 0xA947, 0xA951 }, unicode::Script::Rejang }, // Mn  [11] REJANG VOWEL SIGN I..REJANG CONSONANT SIGN R
    Prop<unicode::Script> { { 0xA952, 0xA953 }, unicode::Script::Rejang }, // Mc   [2] REJANG CONSONANT SIGN H..REJANG VIRAMA
    Prop<unicode::Script> { { 0xA95F, 0xA95F }, unicode::Script::Rejang }, // Po       REJANG SECTION MARK
    Prop<unicode::Script> { { 0xA960, 0xA97C }, unicode::Script::Hangul }, // Lo  [29] HANGUL CHOSEONG TIKEUT-MIEUM..HANGUL CHOSEONG SSANGYEORINHIEUH
    Prop<unicode::Script> { { 0xA980, 0xA982 }, unicode::Script::Javanese }, // Mn   [3] JAVANESE SIGN PANYANGGA..JAVANESE SIGN LAYAR
    Prop<unicode::Script> { { 0xA983, 0xA983 }, unicode::Script::Javanese }, // Mc       JAVANESE SIGN WIGNYAN
    Prop<unicode::Script> { { 0xA984, 0xA9B2 }, unicode::Script::Javanese }, // Lo  [47] JAVANESE LETTER A..JAVANESE LETTER HA
    Prop<unicode::Script> { { 0xA9B3, 0xA9B3 }, unicode::Script::Javanese }, // Mn       JAVANESE SIGN CECAK TELU
    Prop<unicode::Script> { { 0xA9B4, 0xA9B5 }, unicode::Script::Javanese }, // Mc   [2] JAVANESE VOWEL SIGN TARUNG..JAVANESE VOWEL SIGN TOLONG
    Prop<unicode::Script> { { 0xA9B6, 0xA9B9 }, unicode::Script::Javanese }, // Mn   [4] JAVANESE VOWEL SIGN WULU..JAVANESE VOWEL SIGN SUKU MENDUT
    Prop<unicode::Script> { { 0xA9BA, 0xA9BB }, unicode::Script::Javanese }, // Mc   [2] JAVANESE VOWEL SIGN TALING..JAVANESE VOWEL SIGN DIRGA MURE
    Prop<unicode::Script> { { 0xA9BC, 0xA9BD }, unicode::Script::Javanese }, // Mn   [2] JAVANESE VOWEL SIGN PEPET..JAVANESE CONSONANT SIGN KERET
    Prop<unicode::Script> { { 0xA9BE, 0xA9C0 }, unicode::Script::Javanese }, // Mc   [3] JAVANESE CONSONANT SIGN PENGKAL..JAVANESE PANGKON
    Prop<unicode::Script> { { 0xA9C1, 0xA9CD }, unicode::Script::Javanese }, // Po  [13] JAVANESE LEFT RERENGGAN..JAVANESE TURNED PADA PISELEH
    Prop<unicode::Script> { { 0xA9CF, 0xA9CF }, unicode::Script::Common }, // Lm       JAVANESE PANGRANGKEP
    Prop<unicode::Script> { { 0xA9D0, 0xA9D9 }, unicode::Script::Javanese }, // Nd  [10] JAVANESE DIGIT ZERO..JAVANESE DIGIT NINE
    Prop<unicode::Script> { { 0xA9DE, 0xA9DF }, unicode::Script::Javanese }, // Po   [2] JAVANESE PADA TIRTA TUMETES..JAVANESE PADA ISEN-ISEN
    Prop<unicode::Script> { { 0xA9E0, 0xA9E4 }, unicode::Script::Myanmar }, // Lo   [5] MYANMAR LETTER SHAN GHA..MYANMAR LETTER SHAN BHA
    Prop<unicode::Script> { { 0xA9E5, 0xA9E5 }, unicode::Script::Myanmar }, // Mn       MYANMAR SIGN SHAN SAW
    Prop<unicode::Script> { { 0xA9E6, 0xA9E6 }, unicode::Script::Myanmar }, // Lm       MYANMAR MODIFIER LETTER SHAN REDUPLICATION
    Prop<unicode::Script> { { 0xA9E7, 0xA9EF }, unicode::Script::Myanmar }, // Lo   [9] MYANMAR LETTER TAI LAING NYA..MYANMAR LETTER TAI LAING NNA
    Prop<unicode::Script> { { 0xA9F0, 0xA9F9 }, unicode::Script::Myanmar }, // Nd  [10] MYANMAR TAI LAING DIGIT ZERO..MYANMAR TAI LAING DIGIT NINE
    Prop<unicode::Script> { { 0xA9FA, 0xA9FE }, unicode::Script::Myanmar }, // Lo   [5] MYANMAR LETTER TAI LAING LLA..MYANMAR LETTER TAI LAING BHA
    Prop<unicode::Script> { { 0xAA00, 0xAA28 }, unicode::Script::Cham }, // Lo  [41] CHAM LETTER A..CHAM LETTER HA
    Prop<unicode::Script> { { 0xAA29, 0xAA2E }, unicode::Script::Cham }, // Mn   [6] CHAM VOWEL SIGN AA..CHAM VOWEL SIGN OE
    Prop<unicode::Script> { { 0xAA2F, 0xAA30 }, unicode::Script::Cham }, // Mc   [2] CHAM VOWEL SIGN O..CHAM VOWEL SIGN AI
    Prop<unicode::Script> { { 0xAA31, 0xAA32 }, unicode::Script::Cham }, // Mn   [2] CHAM VOWEL SIGN AU..CHAM VOWEL SIGN UE
    Prop<unicode::Script> { { 0xAA33, 0xAA34 }, unicode::Script::Cham }, // Mc   [2] CHAM CONSONANT SIGN YA..CHAM CONSONANT SIGN RA
    Prop<unicode::Script> { { 0xAA35, 0xAA36 }, unicode::Script::Cham }, // Mn   [2] CHAM CONSONANT SIGN LA..CHAM CONSONANT SIGN WA
    Prop<unicode::Script> { { 0xAA40, 0xAA42 }, unicode::Script::Cham }, // Lo   [3] CHAM LETTER FINAL K..CHAM LETTER FINAL NG
    Prop<unicode::Script> { { 0xAA43, 0xAA43 }, unicode::Script::Cham }, // Mn       CHAM CONSONANT SIGN FINAL NG
    Prop<unicode::Script> { { 0xAA44, 0xAA4B }, unicode::Script::Cham }, // Lo   [8] CHAM LETTER FINAL CH..CHAM LETTER FINAL SS
    Prop<unicode::Script> { { 0xAA4C, 0xAA4C }, unicode::Script::Cham }, // Mn       CHAM CONSONANT SIGN FINAL M
    Prop<unicode::Script> { { 0xAA4D, 0xAA4D }, unicode::Script::Cham }, // Mc       CHAM CONSONANT SIGN FINAL H
    Prop<unicode::Script> { { 0xAA50, 0xAA59 }, unicode::Script::Cham }, // Nd  [10] CHAM DIGIT ZERO..CHAM DIGIT NINE
    Prop<unicode::Script> { { 0xAA5C, 0xAA5F }, unicode::Script::Cham }, // Po   [4] CHAM PUNCTUATION SPIRAL..CHAM PUNCTUATION TRIPLE DANDA
    Prop<unicode::Script> { { 0xAA60, 0xAA6F }, unicode::Script::Myanmar }, // Lo  [16] MYANMAR LETTER KHAMTI GA..MYANMAR LETTER KHAMTI FA
    Prop<unicode::Script> { { 0xAA70, 0xAA70 }, unicode::Script::Myanmar }, // Lm       MYANMAR MODIFIER LETTER KHAMTI REDUPLICATION
    Prop<unicode::Script> { { 0xAA71, 0xAA76 }, unicode::Script::Myanmar }, // Lo   [6] MYANMAR LETTER KHAMTI XA..MYANMAR LOGOGRAM KHAMTI HM
    Prop<unicode::Script> { { 0xAA77, 0xAA79 }, unicode::Script::Myanmar }, // So   [3] MYANMAR SYMBOL AITON EXCLAMATION..MYANMAR SYMBOL AITON TWO
    Prop<unicode::Script> { { 0xAA7A, 0xAA7A }, unicode::Script::Myanmar }, // Lo       MYANMAR LETTER AITON RA
    Prop<unicode::Script> { { 0xAA7B, 0xAA7B }, unicode::Script::Myanmar }, // Mc       MYANMAR SIGN PAO KAREN TONE
    Prop<unicode::Script> { { 0xAA7C, 0xAA7C }, unicode::Script::Myanmar }, // Mn       MYANMAR SIGN TAI LAING TONE-2
    Prop<unicode::Script> { { 0xAA7D, 0xAA7D }, unicode::Script::Myanmar }, // Mc       MYANMAR SIGN TAI LAING TONE-5
    Prop<unicode::Script> { { 0xAA7E, 0xAA7F }, unicode::Script::Myanmar }, // Lo   [2] MYANMAR LETTER SHWE PALAUNG CHA..MYANMAR LETTER SHWE PALAUNG SHA
    Prop<unicode::Script> { { 0xAA80, 0xAAAF }, unicode::Script::Tai_Viet }, // Lo  [48] TAI VIET LETTER LOW KO..TAI VIET LETTER HIGH O
    Prop<unicode::Script> { { 0xAAB0, 0xAAB0 }, unicode::Script::Tai_Viet }, // Mn       TAI VIET MAI KANG
    Prop<unicode::Script> { { 0xAAB1, 0xAAB1 }, unicode::Script::Tai_Viet }, // Lo       TAI VIET VOWEL AA
    Prop<unicode::Script> { { 0xAAB2, 0xAAB4 }, unicode::Script::Tai_Viet }, // Mn   [3] TAI VIET VOWEL I..TAI VIET VOWEL U
    Prop<unicode::Script> { { 0xAAB5, 0xAAB6 }, unicode::Script::Tai_Viet }, // Lo   [2] TAI VIET VOWEL E..TAI VIET VOWEL O
    Prop<unicode::Script> { { 0xAAB7, 0xAAB8 }, unicode::Script::Tai_Viet }, // Mn   [2] TAI VIET MAI KHIT..TAI VIET VOWEL IA
    Prop<unicode::Script> { { 0xAAB9, 0xAABD }, unicode::Script::Tai_Viet }, // Lo   [5] TAI VIET VOWEL UEA..TAI VIET VOWEL AN
    Prop<unicode::Script> { { 0xAABE, 0xAABF }, unicode::Script::Tai_Viet }, // Mn   [2] TAI VIET VOWEL AM..TAI VIET TONE MAI EK
    Prop<unicode::Script> { { 0xAAC0, 0xAAC0 }, unicode::Script::Tai_Viet }, // Lo       TAI VIET TONE MAI NUENG
    Prop<unicode::Script> { { 0xAAC1, 0xAAC1 }, unicode::Script::Tai_Viet }, // Mn       TAI VIET TONE MAI THO
    Prop<unicode::Script> { { 0xAAC2, 0xAAC2 }, unicode::Script::Tai_Viet }, // Lo       TAI VIET TONE MAI SONG
    Prop<unicode::Script> { { 0xAADB, 0xAADC }, unicode::Script::Tai_Viet }, // Lo   [2] TAI VIET SYMBOL KON..TAI VIET SYMBOL NUENG
    Prop<unicode::Script> { { 0xAADD, 0xAADD }, unicode::Script::Tai_Viet }, // Lm       TAI VIET SYMBOL SAM
    Prop<unicode::Script> { { 0xAADE, 0xAADF }, unicode::Script::Tai_Viet }, // Po   [2] TAI VIET SYMBOL HO HOI..TAI VIET SYMBOL KOI KOI
    Prop<unicode::Script> { { 0xAAE0, 0xAAEA }, unicode::Script::Meetei_Mayek }, // Lo  [11] MEETEI MAYEK LETTER E..MEETEI MAYEK LETTER SSA
    Prop<unicode::Script> { { 0xAAEB, 0xAAEB }, unicode::Script::Meetei_Mayek }, // Mc       MEETEI MAYEK VOWEL SIGN II
    Prop<unicode::Script> { { 0xAAEC, 0xAAED }, unicode::Script::Meetei_Mayek }, // Mn   [2] MEETEI MAYEK VOWEL SIGN UU..MEETEI MAYEK VOWEL SIGN AAI
    Prop<unicode::Script> { { 0xAAEE, 0xAAEF }, unicode::Script::Meetei_Mayek }, // Mc   [2] MEETEI MAYEK VOWEL SIGN AU..MEETEI MAYEK VOWEL SIGN AAU
    Prop<unicode::Script> { { 0xAAF0, 0xAAF1 }, unicode::Script::Meetei_Mayek }, // Po   [2] MEETEI MAYEK CHEIKHAN..MEETEI MAYEK AHANG KHUDAM
    Prop<unicode::Script> { { 0xAAF2, 0xAAF2 }, unicode::Script::Meetei_Mayek }, // Lo       MEETEI MAYEK ANJI
    Prop<unicode::Script> { { 0xAAF3, 0xAAF4 }, unicode::Script::Meetei_Mayek }, // Lm   [2] MEETEI MAYEK SYLLABLE REPETITION MARK..MEETEI MAYEK WORD REPETITION MARK
    Prop<unicode::Script> { { 0xAAF5, 0xAAF5 }, unicode::Script::Meetei_Mayek }, // Mc       MEETEI MAYEK VOWEL SIGN VISARGA
    Prop<unicode::Script> { { 0xAAF6, 0xAAF6 }, unicode::Script::Meetei_Mayek }, // Mn       MEETEI MAYEK VIRAMA
    Prop<unicode::Script> { { 0xAB01, 0xAB06 }, unicode::Script::Ethiopic }, // Lo   [6] ETHIOPIC SYLLABLE TTHU..ETHIOPIC SYLLABLE TTHO
    Prop<unicode::Script> { { 0xAB09, 0xAB0E }, unicode::Script::Ethiopic }, // Lo   [6] ETHIOPIC SYLLABLE DDHU..ETHIOPIC SYLLABLE DDHO
    Prop<unicode::Script> { { 0xAB11, 0xAB16 }, unicode::Script::Ethiopic }, // Lo   [6] ETHIOPIC SYLLABLE DZU..ETHIOPIC SYLLABLE DZO
    Prop<unicode::Script> { { 0xAB20, 0xAB26 }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE CCHHA..ETHIOPIC SYLLABLE CCHHO
    Prop<unicode::Script> { { 0xAB28, 0xAB2E }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE BBA..ETHIOPIC SYLLABLE BBO
    Prop<unicode::Script> { { 0xAB30, 0xAB5A }, unicode::Script::Latin }, // L&  [43] LATIN SMALL LETTER BARRED ALPHA..LATIN SMALL LETTER Y WITH SHORT RIGHT LEG
    Prop<unicode::Script> { { 0xAB5B, 0xAB5B }, unicode::Script::Common }, // Sk       MODIFIER BREVE WITH INVERTED BREVE
    Prop<unicode::Script> { { 0xAB5C, 0xAB5F }, unicode::Script::Latin }, // Lm   [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    Prop<unicode::Script> { { 0xAB60, 0xAB64 }, unicode::Script::Latin }, // L&   [5] LATIN SMALL LETTER SAKHA YAT..LATIN SMALL LETTER INVERTED ALPHA
    Prop<unicode::Script> { { 0xAB65, 0xAB65 }, unicode::Script::Greek }, // L&       GREEK LETTER SMALL CAPITAL OMEGA
    Prop<unicode::Script> { { 0xAB66, 0xAB68 }, unicode::Script::Latin }, // L&   [3] LATIN SMALL LETTER DZ DIGRAPH WITH RETROFLEX HOOK..LATIN SMALL LETTER TURNED R WITH MIDDLE TILDE
    Prop<unicode::Script> { { 0xAB69, 0xAB69 }, unicode::Script::Latin }, // Lm       MODIFIER LETTER SMALL TURNED W
    Prop<unicode::Script> { { 0xAB6A, 0xAB6B }, unicode::Script::Common }, // Sk   [2] MODIFIER LETTER LEFT TACK..MODIFIER LETTER RIGHT TACK
    Prop<unicode::Script> { { 0xAB70, 0xABBF }, unicode::Script::Cherokee }, // L&  [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    Prop<unicode::Script> { { 0xABC0, 0xABE2 }, unicode::Script::Meetei_Mayek }, // Lo  [35] MEETEI MAYEK LETTER KOK..MEETEI MAYEK LETTER I LONSUM
    Prop<unicode::Script> { { 0xABE3, 0xABE4 }, unicode::Script::Meetei_Mayek }, // Mc   [2] MEETEI MAYEK VOWEL SIGN ONAP..MEETEI MAYEK VOWEL SIGN INAP
    Prop<unicode::Script> { { 0xABE5, 0xABE5 }, unicode::Script::Meetei_Mayek }, // Mn       MEETEI MAYEK VOWEL SIGN ANAP
    Prop<unicode::Script> { { 0xABE6, 0xABE7 }, unicode::Script::Meetei_Mayek }, // Mc   [2] MEETEI MAYEK VOWEL SIGN YENAP..MEETEI MAYEK VOWEL SIGN SOUNAP
    Prop<unicode::Script> { { 0xABE8, 0xABE8 }, unicode::Script::Meetei_Mayek }, // Mn       MEETEI MAYEK VOWEL SIGN UNAP
    Prop<unicode::Script> { { 0xABE9, 0xABEA }, unicode::Script::Meetei_Mayek }, // Mc   [2] MEETEI MAYEK VOWEL SIGN CHEINAP..MEETEI MAYEK VOWEL SIGN NUNG
    Prop<unicode::Script> { { 0xABEB, 0xABEB }, unicode::Script::Meetei_Mayek }, // Po       MEETEI MAYEK CHEIKHEI
    Prop<unicode::Script> { { 0xABEC, 0xABEC }, unicode::Script::Meetei_Mayek }, // Mc       MEETEI MAYEK LUM IYEK
    Prop<unicode::Script> { { 0xABED, 0xABED }, unicode::Script::Meetei_Mayek }, // Mn       MEETEI MAYEK APUN IYEK
    Prop<unicode::Script> { { 0xABF0, 0xABF9 }, unicode::Script::Meetei_Mayek }, // Nd  [10] MEETEI MAYEK DIGIT ZERO..MEETEI MAYEK DIGIT NINE
    Prop<unicode::Script> { { 0xAC00, 0xD7A3 }, unicode::Script::Hangul }, // Lo [11172] HANGUL SYLLABLE GA..HANGUL SYLLABLE HIH
    Prop<unicode::Script> { { 0xD7B0, 0xD7C6 }, unicode::Script::Hangul }, // Lo  [23] HANGUL JUNGSEONG O-YEO..HANGUL JUNGSEONG ARAEA-E
    Prop<unicode::Script> { { 0xD7CB, 0xD7FB }, unicode::Script::Hangul }, // Lo  [49] HANGUL JONGSEONG NIEUN-RIEUL..HANGUL JONGSEONG PHIEUPH-THIEUTH
    Prop<unicode::Script> { { 0xF900, 0xFA6D }, unicode::Script::Han }, // Lo [366] CJK COMPATIBILITY IDEOGRAPH-F900..CJK COMPATIBILITY IDEOGRAPH-FA6D
    Prop<unicode::Script> { { 0xFA70, 0xFAD9 }, unicode::Script::Han }, // Lo [106] CJK COMPATIBILITY IDEOGRAPH-FA70..CJK COMPATIBILITY IDEOGRAPH-FAD9
    Prop<unicode::Script> { { 0xFB00, 0xFB06 }, unicode::Script::Latin }, // L&   [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Prop<unicode::Script> { { 0xFB13, 0xFB17 }, unicode::Script::Armenian }, // L&   [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    Prop<unicode::Script> { { 0xFB1D, 0xFB1D }, unicode::Script::Hebrew }, // Lo       HEBREW LETTER YOD WITH HIRIQ
    Prop<unicode::Script> { { 0xFB1E, 0xFB1E }, unicode::Script::Hebrew }, // Mn       HEBREW POINT JUDEO-SPANISH VARIKA
    Prop<unicode::Script> { { 0xFB1F, 0xFB28 }, unicode::Script::Hebrew }, // Lo  [10] HEBREW LIGATURE YIDDISH YOD YOD PATAH..HEBREW LETTER WIDE TAV
    Prop<unicode::Script> { { 0xFB29, 0xFB29 }, unicode::Script::Hebrew }, // Sm       HEBREW LETTER ALTERNATIVE PLUS SIGN
    Prop<unicode::Script> { { 0xFB2A, 0xFB36 }, unicode::Script::Hebrew }, // Lo  [13] HEBREW LETTER SHIN WITH SHIN DOT..HEBREW LETTER ZAYIN WITH DAGESH
    Prop<unicode::Script> { { 0xFB38, 0xFB3C }, unicode::Script::Hebrew }, // Lo   [5] HEBREW LETTER TET WITH DAGESH..HEBREW LETTER LAMED WITH DAGESH
    Prop<unicode::Script> { { 0xFB3E, 0xFB3E }, unicode::Script::Hebrew }, // Lo       HEBREW LETTER MEM WITH DAGESH
    Prop<unicode::Script> { { 0xFB40, 0xFB41 }, unicode::Script::Hebrew }, // Lo   [2] HEBREW LETTER NUN WITH DAGESH..HEBREW LETTER SAMEKH WITH DAGESH
    Prop<unicode::Script> { { 0xFB43, 0xFB44 }, unicode::Script::Hebrew }, // Lo   [2] HEBREW LETTER FINAL PE WITH DAGESH..HEBREW LETTER PE WITH DAGESH
    Prop<unicode::Script> { { 0xFB46, 0xFB4F }, unicode::Script::Hebrew }, // Lo  [10] HEBREW LETTER TSADI WITH DAGESH..HEBREW LIGATURE ALEF LAMED
    Prop<unicode::Script> { { 0xFB50, 0xFBB1 }, unicode::Script::Arabic }, // Lo  [98] ARABIC LETTER ALEF WASLA ISOLATED FORM..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE FINAL FORM
    Prop<unicode::Script> { { 0xFBB2, 0xFBC2 }, unicode::Script::Arabic }, // Sk  [17] ARABIC SYMBOL DOT ABOVE..ARABIC SYMBOL WASLA ABOVE
    Prop<unicode::Script> { { 0xFBC3, 0xFBD2 }, unicode::Script::Arabic }, // So  [16] ARABIC LIGATURE JALLA WA-ALAA..ARABIC LIGATURE ALAYHI AR-RAHMAH
    Prop<unicode::Script> { { 0xFBD3, 0xFD3D }, unicode::Script::Arabic }, // Lo [363] ARABIC LETTER NG ISOLATED FORM..ARABIC LIGATURE ALEF WITH FATHATAN ISOLATED FORM
    Prop<unicode::Script> { { 0xFD3E, 0xFD3E }, unicode::Script::Common }, // Pe       ORNATE LEFT PARENTHESIS
    Prop<unicode::Script> { { 0xFD3F, 0xFD3F }, unicode::Script::Common }, // Ps       ORNATE RIGHT PARENTHESIS
    Prop<unicode::Script> { { 0xFD40, 0xFD4F }, unicode::Script::Arabic }, // So  [16] ARABIC LIGATURE RAHIMAHU ALLAAH..ARABIC LIGATURE RAHIMAHUM ALLAAH
    Prop<unicode::Script> { { 0xFD50, 0xFD8F }, unicode::Script::Arabic }, // Lo  [64] ARABIC LIGATURE TEH WITH JEEM WITH MEEM INITIAL FORM..ARABIC LIGATURE MEEM WITH KHAH WITH MEEM INITIAL FORM
    Prop<unicode::Script> { { 0xFD90, 0xFD91 }, unicode::Script::Arabic }, // So   [2] ARABIC LIGATURE RAHMATU ALLAAHI ALAYH..ARABIC LIGATURE RAHMATU ALLAAHI ALAYHAA
    Prop<unicode::Script> { { 0xFD92, 0xFDC7 }, unicode::Script::Arabic }, // Lo  [54] ARABIC LIGATURE MEEM WITH JEEM WITH KHAH INITIAL FORM..ARABIC LIGATURE NOON WITH JEEM WITH YEH FINAL FORM
    Prop<unicode::Script> { { 0xFDC8, 0xFDCF }, unicode::Script::Arabic }, // So   [8] ARABIC LIGATURE RAHIMAHU ALLAAH TAAALAA..ARABIC LIGATURE SALAAMUHU ALAYNAA
    Prop<unicode::Script> { { 0xFDF0, 0xFDFB }, unicode::Script::Arabic }, // Lo  [12] ARABIC LIGATURE SALLA USED AS KORANIC STOP SIGN ISOLATED FORM..ARABIC LIGATURE JALLAJALALOUHOU
    Prop<unicode::Script> { { 0xFDFC, 0xFDFC }, unicode::Script::Arabic }, // Sc       RIAL SIGN
    Prop<unicode::Script> { { 0xFDFD, 0xFDFF }, unicode::Script::Arabic }, // So   [3] ARABIC LIGATURE BISMILLAH AR-RAHMAN AR-RAHEEM..ARABIC LIGATURE AZZA WA JALL
    Prop<unicode::Script> { { 0xFE00, 0xFE0F }, unicode::Script::Inherited }, // Mn  [16] VARIATION SELECTOR-1..VARIATION SELECTOR-16
    Prop<unicode::Script> { { 0xFE10, 0xFE16 }, unicode::Script::Common }, // Po   [7] PRESENTATION FORM FOR VERTICAL COMMA..PRESENTATION FORM FOR VERTICAL QUESTION MARK
    Prop<unicode::Script> { { 0xFE17, 0xFE17 }, unicode::Script::Common }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT WHITE LENTICULAR BRACKET
    Prop<unicode::Script> { { 0xFE18, 0xFE18 }, unicode::Script::Common }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT WHITE LENTICULAR BRAKCET
    Prop<unicode::Script> { { 0xFE19, 0xFE19 }, unicode::Script::Common }, // Po       PRESENTATION FORM FOR VERTICAL HORIZONTAL ELLIPSIS
    Prop<unicode::Script> { { 0xFE20, 0xFE2D }, unicode::Script::Inherited }, // Mn  [14] COMBINING LIGATURE LEFT HALF..COMBINING CONJOINING MACRON BELOW
    Prop<unicode::Script> { { 0xFE2E, 0xFE2F }, unicode::Script::Cyrillic }, // Mn   [2] COMBINING CYRILLIC TITLO LEFT HALF..COMBINING CYRILLIC TITLO RIGHT HALF
    Prop<unicode::Script> { { 0xFE30, 0xFE30 }, unicode::Script::Common }, // Po       PRESENTATION FORM FOR VERTICAL TWO DOT LEADER
    Prop<unicode::Script> { { 0xFE31, 0xFE32 }, unicode::Script::Common }, // Pd   [2] PRESENTATION FORM FOR VERTICAL EM DASH..PRESENTATION FORM FOR VERTICAL EN DASH
    Prop<unicode::Script> { { 0xFE33, 0xFE34 }, unicode::Script::Common }, // Pc   [2] PRESENTATION FORM FOR VERTICAL LOW LINE..PRESENTATION FORM FOR VERTICAL WAVY LOW LINE
    Prop<unicode::Script> { { 0xFE35, 0xFE35 }, unicode::Script::Common }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT PARENTHESIS
    Prop<unicode::Script> { { 0xFE36, 0xFE36 }, unicode::Script::Common }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT PARENTHESIS
    Prop<unicode::Script> { { 0xFE37, 0xFE37 }, unicode::Script::Common }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT CURLY BRACKET
    Prop<unicode::Script> { { 0xFE38, 0xFE38 }, unicode::Script::Common }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT CURLY BRACKET
    Prop<unicode::Script> { { 0xFE39, 0xFE39 }, unicode::Script::Common }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0xFE3A, 0xFE3A }, unicode::Script::Common }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0xFE3B, 0xFE3B }, unicode::Script::Common }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT BLACK LENTICULAR BRACKET
    Prop<unicode::Script> { { 0xFE3C, 0xFE3C }, unicode::Script::Common }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT BLACK LENTICULAR BRACKET
    Prop<unicode::Script> { { 0xFE3D, 0xFE3D }, unicode::Script::Common }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT DOUBLE ANGLE BRACKET
    Prop<unicode::Script> { { 0xFE3E, 0xFE3E }, unicode::Script::Common }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT DOUBLE ANGLE BRACKET
    Prop<unicode::Script> { { 0xFE3F, 0xFE3F }, unicode::Script::Common }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT ANGLE BRACKET
    Prop<unicode::Script> { { 0xFE40, 0xFE40 }, unicode::Script::Common }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT ANGLE BRACKET
    Prop<unicode::Script> { { 0xFE41, 0xFE41 }, unicode::Script::Common }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT CORNER BRACKET
    Prop<unicode::Script> { { 0xFE42, 0xFE42 }, unicode::Script::Common }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT CORNER BRACKET
    Prop<unicode::Script> { { 0xFE43, 0xFE43 }, unicode::Script::Common }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT WHITE CORNER BRACKET
    Prop<unicode::Script> { { 0xFE44, 0xFE44 }, unicode::Script::Common }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT WHITE CORNER BRACKET
    Prop<unicode::Script> { { 0xFE45, 0xFE46 }, unicode::Script::Common }, // Po   [2] SESAME DOT..WHITE SESAME DOT
    Prop<unicode::Script> { { 0xFE47, 0xFE47 }, unicode::Script::Common }, // Ps       PRESENTATION FORM FOR VERTICAL LEFT SQUARE BRACKET
    Prop<unicode::Script> { { 0xFE48, 0xFE48 }, unicode::Script::Common }, // Pe       PRESENTATION FORM FOR VERTICAL RIGHT SQUARE BRACKET
    Prop<unicode::Script> { { 0xFE49, 0xFE4C }, unicode::Script::Common }, // Po   [4] DASHED OVERLINE..DOUBLE WAVY OVERLINE
    Prop<unicode::Script> { { 0xFE4D, 0xFE4F }, unicode::Script::Common }, // Pc   [3] DASHED LOW LINE..WAVY LOW LINE
    Prop<unicode::Script> { { 0xFE50, 0xFE52 }, unicode::Script::Common }, // Po   [3] SMALL COMMA..SMALL FULL STOP
    Prop<unicode::Script> { { 0xFE54, 0xFE57 }, unicode::Script::Common }, // Po   [4] SMALL SEMICOLON..SMALL EXCLAMATION MARK
    Prop<unicode::Script> { { 0xFE58, 0xFE58 }, unicode::Script::Common }, // Pd       SMALL EM DASH
    Prop<unicode::Script> { { 0xFE59, 0xFE59 }, unicode::Script::Common }, // Ps       SMALL LEFT PARENTHESIS
    Prop<unicode::Script> { { 0xFE5A, 0xFE5A }, unicode::Script::Common }, // Pe       SMALL RIGHT PARENTHESIS
    Prop<unicode::Script> { { 0xFE5B, 0xFE5B }, unicode::Script::Common }, // Ps       SMALL LEFT CURLY BRACKET
    Prop<unicode::Script> { { 0xFE5C, 0xFE5C }, unicode::Script::Common }, // Pe       SMALL RIGHT CURLY BRACKET
    Prop<unicode::Script> { { 0xFE5D, 0xFE5D }, unicode::Script::Common }, // Ps       SMALL LEFT TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0xFE5E, 0xFE5E }, unicode::Script::Common }, // Pe       SMALL RIGHT TORTOISE SHELL BRACKET
    Prop<unicode::Script> { { 0xFE5F, 0xFE61 }, unicode::Script::Common }, // Po   [3] SMALL NUMBER SIGN..SMALL ASTERISK
    Prop<unicode::Script> { { 0xFE62, 0xFE62 }, unicode::Script::Common }, // Sm       SMALL PLUS SIGN
    Prop<unicode::Script> { { 0xFE63, 0xFE63 }, unicode::Script::Common }, // Pd       SMALL HYPHEN-MINUS
    Prop<unicode::Script> { { 0xFE64, 0xFE66 }, unicode::Script::Common }, // Sm   [3] SMALL LESS-THAN SIGN..SMALL EQUALS SIGN
    Prop<unicode::Script> { { 0xFE68, 0xFE68 }, unicode::Script::Common }, // Po       SMALL REVERSE SOLIDUS
    Prop<unicode::Script> { { 0xFE69, 0xFE69 }, unicode::Script::Common }, // Sc       SMALL DOLLAR SIGN
    Prop<unicode::Script> { { 0xFE6A, 0xFE6B }, unicode::Script::Common }, // Po   [2] SMALL PERCENT SIGN..SMALL COMMERCIAL AT
    Prop<unicode::Script> { { 0xFE70, 0xFE74 }, unicode::Script::Arabic }, // Lo   [5] ARABIC FATHATAN ISOLATED FORM..ARABIC KASRATAN ISOLATED FORM
    Prop<unicode::Script> { { 0xFE76, 0xFEFC }, unicode::Script::Arabic }, // Lo [135] ARABIC FATHA ISOLATED FORM..ARABIC LIGATURE LAM WITH ALEF FINAL FORM
    Prop<unicode::Script> { { 0xFEFF, 0xFEFF }, unicode::Script::Common }, // Cf       ZERO WIDTH NO-BREAK SPACE
    Prop<unicode::Script> { { 0xFF01, 0xFF03 }, unicode::Script::Common }, // Po   [3] FULLWIDTH EXCLAMATION MARK..FULLWIDTH NUMBER SIGN
    Prop<unicode::Script> { { 0xFF04, 0xFF04 }, unicode::Script::Common }, // Sc       FULLWIDTH DOLLAR SIGN
    Prop<unicode::Script> { { 0xFF05, 0xFF07 }, unicode::Script::Common }, // Po   [3] FULLWIDTH PERCENT SIGN..FULLWIDTH APOSTROPHE
    Prop<unicode::Script> { { 0xFF08, 0xFF08 }, unicode::Script::Common }, // Ps       FULLWIDTH LEFT PARENTHESIS
    Prop<unicode::Script> { { 0xFF09, 0xFF09 }, unicode::Script::Common }, // Pe       FULLWIDTH RIGHT PARENTHESIS
    Prop<unicode::Script> { { 0xFF0A, 0xFF0A }, unicode::Script::Common }, // Po       FULLWIDTH ASTERISK
    Prop<unicode::Script> { { 0xFF0B, 0xFF0B }, unicode::Script::Common }, // Sm       FULLWIDTH PLUS SIGN
    Prop<unicode::Script> { { 0xFF0C, 0xFF0C }, unicode::Script::Common }, // Po       FULLWIDTH COMMA
    Prop<unicode::Script> { { 0xFF0D, 0xFF0D }, unicode::Script::Common }, // Pd       FULLWIDTH HYPHEN-MINUS
    Prop<unicode::Script> { { 0xFF0E, 0xFF0F }, unicode::Script::Common }, // Po   [2] FULLWIDTH FULL STOP..FULLWIDTH SOLIDUS
    Prop<unicode::Script> { { 0xFF10, 0xFF19 }, unicode::Script::Common }, // Nd  [10] FULLWIDTH DIGIT ZERO..FULLWIDTH DIGIT NINE
    Prop<unicode::Script> { { 0xFF1A, 0xFF1B }, unicode::Script::Common }, // Po   [2] FULLWIDTH COLON..FULLWIDTH SEMICOLON
    Prop<unicode::Script> { { 0xFF1C, 0xFF1E }, unicode::Script::Common }, // Sm   [3] FULLWIDTH LESS-THAN SIGN..FULLWIDTH GREATER-THAN SIGN
    Prop<unicode::Script> { { 0xFF1F, 0xFF20 }, unicode::Script::Common }, // Po   [2] FULLWIDTH QUESTION MARK..FULLWIDTH COMMERCIAL AT
    Prop<unicode::Script> { { 0xFF21, 0xFF3A }, unicode::Script::Latin }, // L&  [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    Prop<unicode::Script> { { 0xFF3B, 0xFF3B }, unicode::Script::Common }, // Ps       FULLWIDTH LEFT SQUARE BRACKET
    Prop<unicode::Script> { { 0xFF3C, 0xFF3C }, unicode::Script::Common }, // Po       FULLWIDTH REVERSE SOLIDUS
    Prop<unicode::Script> { { 0xFF3D, 0xFF3D }, unicode::Script::Common }, // Pe       FULLWIDTH RIGHT SQUARE BRACKET
    Prop<unicode::Script> { { 0xFF3E, 0xFF3E }, unicode::Script::Common }, // Sk       FULLWIDTH CIRCUMFLEX ACCENT
    Prop<unicode::Script> { { 0xFF3F, 0xFF3F }, unicode::Script::Common }, // Pc       FULLWIDTH LOW LINE
    Prop<unicode::Script> { { 0xFF40, 0xFF40 }, unicode::Script::Common }, // Sk       FULLWIDTH GRAVE ACCENT
    Prop<unicode::Script> { { 0xFF41, 0xFF5A }, unicode::Script::Latin }, // L&  [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    Prop<unicode::Script> { { 0xFF5B, 0xFF5B }, unicode::Script::Common }, // Ps       FULLWIDTH LEFT CURLY BRACKET
    Prop<unicode::Script> { { 0xFF5C, 0xFF5C }, unicode::Script::Common }, // Sm       FULLWIDTH VERTICAL LINE
    Prop<unicode::Script> { { 0xFF5D, 0xFF5D }, unicode::Script::Common }, // Pe       FULLWIDTH RIGHT CURLY BRACKET
    Prop<unicode::Script> { { 0xFF5E, 0xFF5E }, unicode::Script::Common }, // Sm       FULLWIDTH TILDE
    Prop<unicode::Script> { { 0xFF5F, 0xFF5F }, unicode::Script::Common }, // Ps       FULLWIDTH LEFT WHITE PARENTHESIS
    Prop<unicode::Script> { { 0xFF60, 0xFF60 }, unicode::Script::Common }, // Pe       FULLWIDTH RIGHT WHITE PARENTHESIS
    Prop<unicode::Script> { { 0xFF61, 0xFF61 }, unicode::Script::Common }, // Po       HALFWIDTH IDEOGRAPHIC FULL STOP
    Prop<unicode::Script> { { 0xFF62, 0xFF62 }, unicode::Script::Common }, // Ps       HALFWIDTH LEFT CORNER BRACKET
    Prop<unicode::Script> { { 0xFF63, 0xFF63 }, unicode::Script::Common }, // Pe       HALFWIDTH RIGHT CORNER BRACKET
    Prop<unicode::Script> { { 0xFF64, 0xFF65 }, unicode::Script::Common }, // Po   [2] HALFWIDTH IDEOGRAPHIC COMMA..HALFWIDTH KATAKANA MIDDLE DOT
    Prop<unicode::Script> { { 0xFF66, 0xFF6F }, unicode::Script::Katakana }, // Lo  [10] HALFWIDTH KATAKANA LETTER WO..HALFWIDTH KATAKANA LETTER SMALL TU
    Prop<unicode::Script> { { 0xFF70, 0xFF70 }, unicode::Script::Common }, // Lm       HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Prop<unicode::Script> { { 0xFF71, 0xFF9D }, unicode::Script::Katakana }, // Lo  [45] HALFWIDTH KATAKANA LETTER A..HALFWIDTH KATAKANA LETTER N
    Prop<unicode::Script> { { 0xFF9E, 0xFF9F }, unicode::Script::Common }, // Lm   [2] HALFWIDTH KATAKANA VOICED SOUND MARK..HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK
    Prop<unicode::Script> { { 0xFFA0, 0xFFBE }, unicode::Script::Hangul }, // Lo  [31] HALFWIDTH HANGUL FILLER..HALFWIDTH HANGUL LETTER HIEUH
    Prop<unicode::Script> { { 0xFFC2, 0xFFC7 }, unicode::Script::Hangul }, // Lo   [6] HALFWIDTH HANGUL LETTER A..HALFWIDTH HANGUL LETTER E
    Prop<unicode::Script> { { 0xFFCA, 0xFFCF }, unicode::Script::Hangul }, // Lo   [6] HALFWIDTH HANGUL LETTER YEO..HALFWIDTH HANGUL LETTER OE
    Prop<unicode::Script> { { 0xFFD2, 0xFFD7 }, unicode::Script::Hangul }, // Lo   [6] HALFWIDTH HANGUL LETTER YO..HALFWIDTH HANGUL LETTER YU
    Prop<unicode::Script> { { 0xFFDA, 0xFFDC }, unicode::Script::Hangul }, // Lo   [3] HALFWIDTH HANGUL LETTER EU..HALFWIDTH HANGUL LETTER I
    Prop<unicode::Script> { { 0xFFE0, 0xFFE1 }, unicode::Script::Common }, // Sc   [2] FULLWIDTH CENT SIGN..FULLWIDTH POUND SIGN
    Prop<unicode::Script> { { 0xFFE2, 0xFFE2 }, unicode::Script::Common }, // Sm       FULLWIDTH NOT SIGN
    Prop<unicode::Script> { { 0xFFE3, 0xFFE3 }, unicode::Script::Common }, // Sk       FULLWIDTH MACRON
    Prop<unicode::Script> { { 0xFFE4, 0xFFE4 }, unicode::Script::Common }, // So       FULLWIDTH BROKEN BAR
    Prop<unicode::Script> { { 0xFFE5, 0xFFE6 }, unicode::Script::Common }, // Sc   [2] FULLWIDTH YEN SIGN..FULLWIDTH WON SIGN
    Prop<unicode::Script> { { 0xFFE8, 0xFFE8 }, unicode::Script::Common }, // So       HALFWIDTH FORMS LIGHT VERTICAL
    Prop<unicode::Script> { { 0xFFE9, 0xFFEC }, unicode::Script::Common }, // Sm   [4] HALFWIDTH LEFTWARDS ARROW..HALFWIDTH DOWNWARDS ARROW
    Prop<unicode::Script> { { 0xFFED, 0xFFEE }, unicode::Script::Common }, // So   [2] HALFWIDTH BLACK SQUARE..HALFWIDTH WHITE CIRCLE
    Prop<unicode::Script> { { 0xFFF9, 0xFFFB }, unicode::Script::Common }, // Cf   [3] INTERLINEAR ANNOTATION ANCHOR..INTERLINEAR ANNOTATION TERMINATOR
    Prop<unicode::Script> { { 0xFFFC, 0xFFFD }, unicode::Script::Common }, // So   [2] OBJECT REPLACEMENT CHARACTER..REPLACEMENT CHARACTER
    Prop<unicode::Script> { { 0x10000, 0x1000B }, unicode::Script::Linear_B }, // Lo  [12] LINEAR B SYLLABLE B008 A..LINEAR B SYLLABLE B046 JE
    Prop<unicode::Script> { { 0x1000D, 0x10026 }, unicode::Script::Linear_B }, // Lo  [26] LINEAR B SYLLABLE B036 JO..LINEAR B SYLLABLE B032 QO
    Prop<unicode::Script> { { 0x10028, 0x1003A }, unicode::Script::Linear_B }, // Lo  [19] LINEAR B SYLLABLE B060 RA..LINEAR B SYLLABLE B042 WO
    Prop<unicode::Script> { { 0x1003C, 0x1003D }, unicode::Script::Linear_B }, // Lo   [2] LINEAR B SYLLABLE B017 ZA..LINEAR B SYLLABLE B074 ZE
    Prop<unicode::Script> { { 0x1003F, 0x1004D }, unicode::Script::Linear_B }, // Lo  [15] LINEAR B SYLLABLE B020 ZO..LINEAR B SYLLABLE B091 TWO
    Prop<unicode::Script> { { 0x10050, 0x1005D }, unicode::Script::Linear_B }, // Lo  [14] LINEAR B SYMBOL B018..LINEAR B SYMBOL B089
    Prop<unicode::Script> { { 0x10080, 0x100FA }, unicode::Script::Linear_B }, // Lo [123] LINEAR B IDEOGRAM B100 MAN..LINEAR B IDEOGRAM VESSEL B305
    Prop<unicode::Script> { { 0x10100, 0x10102 }, unicode::Script::Common }, // Po   [3] AEGEAN WORD SEPARATOR LINE..AEGEAN CHECK MARK
    Prop<unicode::Script> { { 0x10107, 0x10133 }, unicode::Script::Common }, // No  [45] AEGEAN NUMBER ONE..AEGEAN NUMBER NINETY THOUSAND
    Prop<unicode::Script> { { 0x10137, 0x1013F }, unicode::Script::Common }, // So   [9] AEGEAN WEIGHT BASE UNIT..AEGEAN MEASURE THIRD SUBUNIT
    Prop<unicode::Script> { { 0x10140, 0x10174 }, unicode::Script::Greek }, // Nl  [53] GREEK ACROPHONIC ATTIC ONE QUARTER..GREEK ACROPHONIC STRATIAN FIFTY MNAS
    Prop<unicode::Script> { { 0x10175, 0x10178 }, unicode::Script::Greek }, // No   [4] GREEK ONE HALF SIGN..GREEK THREE QUARTERS SIGN
    Prop<unicode::Script> { { 0x10179, 0x10189 }, unicode::Script::Greek }, // So  [17] GREEK YEAR SIGN..GREEK TRYBLION BASE SIGN
    Prop<unicode::Script> { { 0x1018A, 0x1018B }, unicode::Script::Greek }, // No   [2] GREEK ZERO SIGN..GREEK ONE QUARTER SIGN
    Prop<unicode::Script> { { 0x1018C, 0x1018E }, unicode::Script::Greek }, // So   [3] GREEK SINUSOID SIGN..NOMISMA SIGN
    Prop<unicode::Script> { { 0x10190, 0x1019C }, unicode::Script::Common }, // So  [13] ROMAN SEXTANS SIGN..ASCIA SYMBOL
    Prop<unicode::Script> { { 0x101A0, 0x101A0 }, unicode::Script::Greek }, // So       GREEK SYMBOL TAU RHO
    Prop<unicode::Script> { { 0x101D0, 0x101FC }, unicode::Script::Common }, // So  [45] PHAISTOS DISC SIGN PEDESTRIAN..PHAISTOS DISC SIGN WAVY BAND
    Prop<unicode::Script> { { 0x101FD, 0x101FD }, unicode::Script::Inherited }, // Mn       PHAISTOS DISC SIGN COMBINING OBLIQUE STROKE
    Prop<unicode::Script> { { 0x10280, 0x1029C }, unicode::Script::Lycian }, // Lo  [29] LYCIAN LETTER A..LYCIAN LETTER X
    Prop<unicode::Script> { { 0x102A0, 0x102D0 }, unicode::Script::Carian }, // Lo  [49] CARIAN LETTER A..CARIAN LETTER UUU3
    Prop<unicode::Script> { { 0x102E0, 0x102E0 }, unicode::Script::Inherited }, // Mn       COPTIC EPACT THOUSANDS MARK
    Prop<unicode::Script> { { 0x102E1, 0x102FB }, unicode::Script::Common }, // No  [27] COPTIC EPACT DIGIT ONE..COPTIC EPACT NUMBER NINE HUNDRED
    Prop<unicode::Script> { { 0x10300, 0x1031F }, unicode::Script::Old_Italic }, // Lo  [32] OLD ITALIC LETTER A..OLD ITALIC LETTER ESS
    Prop<unicode::Script> { { 0x10320, 0x10323 }, unicode::Script::Old_Italic }, // No   [4] OLD ITALIC NUMERAL ONE..OLD ITALIC NUMERAL FIFTY
    Prop<unicode::Script> { { 0x1032D, 0x1032F }, unicode::Script::Old_Italic }, // Lo   [3] OLD ITALIC LETTER YE..OLD ITALIC LETTER SOUTHERN TSE
    Prop<unicode::Script> { { 0x10330, 0x10340 }, unicode::Script::Gothic }, // Lo  [17] GOTHIC LETTER AHSA..GOTHIC LETTER PAIRTHRA
    Prop<unicode::Script> { { 0x10341, 0x10341 }, unicode::Script::Gothic }, // Nl       GOTHIC LETTER NINETY
    Prop<unicode::Script> { { 0x10342, 0x10349 }, unicode::Script::Gothic }, // Lo   [8] GOTHIC LETTER RAIDA..GOTHIC LETTER OTHAL
    Prop<unicode::Script> { { 0x1034A, 0x1034A }, unicode::Script::Gothic }, // Nl       GOTHIC LETTER NINE HUNDRED
    Prop<unicode::Script> { { 0x10350, 0x10375 }, unicode::Script::Old_Permic }, // Lo  [38] OLD PERMIC LETTER AN..OLD PERMIC LETTER IA
    Prop<unicode::Script> { { 0x10376, 0x1037A }, unicode::Script::Old_Permic }, // Mn   [5] COMBINING OLD PERMIC LETTER AN..COMBINING OLD PERMIC LETTER SII
    Prop<unicode::Script> { { 0x10380, 0x1039D }, unicode::Script::Ugaritic }, // Lo  [30] UGARITIC LETTER ALPA..UGARITIC LETTER SSU
    Prop<unicode::Script> { { 0x1039F, 0x1039F }, unicode::Script::Ugaritic }, // Po       UGARITIC WORD DIVIDER
    Prop<unicode::Script> { { 0x103A0, 0x103C3 }, unicode::Script::Old_Persian }, // Lo  [36] OLD PERSIAN SIGN A..OLD PERSIAN SIGN HA
    Prop<unicode::Script> { { 0x103C8, 0x103CF }, unicode::Script::Old_Persian }, // Lo   [8] OLD PERSIAN SIGN AURAMAZDAA..OLD PERSIAN SIGN BUUMISH
    Prop<unicode::Script> { { 0x103D0, 0x103D0 }, unicode::Script::Old_Persian }, // Po       OLD PERSIAN WORD DIVIDER
    Prop<unicode::Script> { { 0x103D1, 0x103D5 }, unicode::Script::Old_Persian }, // Nl   [5] OLD PERSIAN NUMBER ONE..OLD PERSIAN NUMBER HUNDRED
    Prop<unicode::Script> { { 0x10400, 0x1044F }, unicode::Script::Deseret }, // L&  [80] DESERET CAPITAL LETTER LONG I..DESERET SMALL LETTER EW
    Prop<unicode::Script> { { 0x10450, 0x1047F }, unicode::Script::Shavian }, // Lo  [48] SHAVIAN LETTER PEEP..SHAVIAN LETTER YEW
    Prop<unicode::Script> { { 0x10480, 0x1049D }, unicode::Script::Osmanya }, // Lo  [30] OSMANYA LETTER ALEF..OSMANYA LETTER OO
    Prop<unicode::Script> { { 0x104A0, 0x104A9 }, unicode::Script::Osmanya }, // Nd  [10] OSMANYA DIGIT ZERO..OSMANYA DIGIT NINE
    Prop<unicode::Script> { { 0x104B0, 0x104D3 }, unicode::Script::Osage }, // L&  [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Prop<unicode::Script> { { 0x104D8, 0x104FB }, unicode::Script::Osage }, // L&  [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Prop<unicode::Script> { { 0x10500, 0x10527 }, unicode::Script::Elbasan }, // Lo  [40] ELBASAN LETTER A..ELBASAN LETTER KHE
    Prop<unicode::Script> { { 0x10530, 0x10563 }, unicode::Script::Caucasian_Albanian }, // Lo  [52] CAUCASIAN ALBANIAN LETTER ALT..CAUCASIAN ALBANIAN LETTER KIW
    Prop<unicode::Script> { { 0x1056F, 0x1056F }, unicode::Script::Caucasian_Albanian }, // Po       CAUCASIAN ALBANIAN CITATION MARK
    Prop<unicode::Script> { { 0x10570, 0x1057A }, unicode::Script::Vithkuqi }, // L&  [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Prop<unicode::Script> { { 0x1057C, 0x1058A }, unicode::Script::Vithkuqi }, // L&  [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Prop<unicode::Script> { { 0x1058C, 0x10592 }, unicode::Script::Vithkuqi }, // L&   [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Prop<unicode::Script> { { 0x10594, 0x10595 }, unicode::Script::Vithkuqi }, // L&   [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Prop<unicode::Script> { { 0x10597, 0x105A1 }, unicode::Script::Vithkuqi }, // L&  [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Prop<unicode::Script> { { 0x105A3, 0x105B1 }, unicode::Script::Vithkuqi }, // L&  [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Prop<unicode::Script> { { 0x105B3, 0x105B9 }, unicode::Script::Vithkuqi }, // L&   [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Prop<unicode::Script> { { 0x105BB, 0x105BC }, unicode::Script::Vithkuqi }, // L&   [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Prop<unicode::Script> { { 0x105C0, 0x105F3 }, unicode::Script::Todhri }, // Lo  [52] TODHRI LETTER A..TODHRI LETTER OO
    Prop<unicode::Script> { { 0x10600, 0x10736 }, unicode::Script::Linear_A }, // Lo [311] LINEAR A SIGN AB001..LINEAR A SIGN A664
    Prop<unicode::Script> { { 0x10740, 0x10755 }, unicode::Script::Linear_A }, // Lo  [22] LINEAR A SIGN A701 A..LINEAR A SIGN A732 JE
    Prop<unicode::Script> { { 0x10760, 0x10767 }, unicode::Script::Linear_A }, // Lo   [8] LINEAR A SIGN A800..LINEAR A SIGN A807
    Prop<unicode::Script> { { 0x10780, 0x10785 }, unicode::Script::Latin }, // Lm   [6] MODIFIER LETTER SMALL CAPITAL AA..MODIFIER LETTER SMALL B WITH HOOK
    Prop<unicode::Script> { { 0x10787, 0x107B0 }, unicode::Script::Latin }, // Lm  [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Prop<unicode::Script> { { 0x107B2, 0x107BA }, unicode::Script::Latin }, // Lm   [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Prop<unicode::Script> { { 0x10800, 0x10805 }, unicode::Script::Cypriot }, // Lo   [6] CYPRIOT SYLLABLE A..CYPRIOT SYLLABLE JA
    Prop<unicode::Script> { { 0x10808, 0x10808 }, unicode::Script::Cypriot }, // Lo       CYPRIOT SYLLABLE JO
    Prop<unicode::Script> { { 0x1080A, 0x10835 }, unicode::Script::Cypriot }, // Lo  [44] CYPRIOT SYLLABLE KA..CYPRIOT SYLLABLE WO
    Prop<unicode::Script> { { 0x10837, 0x10838 }, unicode::Script::Cypriot }, // Lo   [2] CYPRIOT SYLLABLE XA..CYPRIOT SYLLABLE XE
    Prop<unicode::Script> { { 0x1083C, 0x1083C }, unicode::Script::Cypriot }, // Lo       CYPRIOT SYLLABLE ZA
    Prop<unicode::Script> { { 0x1083F, 0x1083F }, unicode::Script::Cypriot }, // Lo       CYPRIOT SYLLABLE ZO
    Prop<unicode::Script> { { 0x10840, 0x10855 }, unicode::Script::Imperial_Aramaic }, // Lo  [22] IMPERIAL ARAMAIC LETTER ALEPH..IMPERIAL ARAMAIC LETTER TAW
    Prop<unicode::Script> { { 0x10857, 0x10857 }, unicode::Script::Imperial_Aramaic }, // Po       IMPERIAL ARAMAIC SECTION SIGN
    Prop<unicode::Script> { { 0x10858, 0x1085F }, unicode::Script::Imperial_Aramaic }, // No   [8] IMPERIAL ARAMAIC NUMBER ONE..IMPERIAL ARAMAIC NUMBER TEN THOUSAND
    Prop<unicode::Script> { { 0x10860, 0x10876 }, unicode::Script::Palmyrene }, // Lo  [23] PALMYRENE LETTER ALEPH..PALMYRENE LETTER TAW
    Prop<unicode::Script> { { 0x10877, 0x10878 }, unicode::Script::Palmyrene }, // So   [2] PALMYRENE LEFT-POINTING FLEURON..PALMYRENE RIGHT-POINTING FLEURON
    Prop<unicode::Script> { { 0x10879, 0x1087F }, unicode::Script::Palmyrene }, // No   [7] PALMYRENE NUMBER ONE..PALMYRENE NUMBER TWENTY
    Prop<unicode::Script> { { 0x10880, 0x1089E }, unicode::Script::Nabataean }, // Lo  [31] NABATAEAN LETTER FINAL ALEPH..NABATAEAN LETTER TAW
    Prop<unicode::Script> { { 0x108A7, 0x108AF }, unicode::Script::Nabataean }, // No   [9] NABATAEAN NUMBER ONE..NABATAEAN NUMBER ONE HUNDRED
    Prop<unicode::Script> { { 0x108E0, 0x108F2 }, unicode::Script::Hatran }, // Lo  [19] HATRAN LETTER ALEPH..HATRAN LETTER QOPH
    Prop<unicode::Script> { { 0x108F4, 0x108F5 }, unicode::Script::Hatran }, // Lo   [2] HATRAN LETTER SHIN..HATRAN LETTER TAW
    Prop<unicode::Script> { { 0x108FB, 0x108FF }, unicode::Script::Hatran }, // No   [5] HATRAN NUMBER ONE..HATRAN NUMBER ONE HUNDRED
    Prop<unicode::Script> { { 0x10900, 0x10915 }, unicode::Script::Phoenician }, // Lo  [22] PHOENICIAN LETTER ALF..PHOENICIAN LETTER TAU
    Prop<unicode::Script> { { 0x10916, 0x1091B }, unicode::Script::Phoenician }, // No   [6] PHOENICIAN NUMBER ONE..PHOENICIAN NUMBER THREE
    Prop<unicode::Script> { { 0x1091F, 0x1091F }, unicode::Script::Phoenician }, // Po       PHOENICIAN WORD SEPARATOR
    Prop<unicode::Script> { { 0x10920, 0x10939 }, unicode::Script::Lydian }, // Lo  [26] LYDIAN LETTER A..LYDIAN LETTER C
    Prop<unicode::Script> { { 0x1093F, 0x1093F }, unicode::Script::Lydian }, // Po       LYDIAN TRIANGULAR MARK
    Prop<unicode::Script> { { 0x10940, 0x10959 }, unicode::Script::Sidetic }, // Lo  [26] SIDETIC LETTER N01..SIDETIC LETTER N26
    Prop<unicode::Script> { { 0x10980, 0x1099F }, unicode::Script::Meroitic_Hieroglyphs }, // Lo  [32] MEROITIC HIEROGLYPHIC LETTER A..MEROITIC HIEROGLYPHIC SYMBOL VIDJ-2
    Prop<unicode::Script> { { 0x109A0, 0x109B7 }, unicode::Script::Meroitic_Cursive }, // Lo  [24] MEROITIC CURSIVE LETTER A..MEROITIC CURSIVE LETTER DA
    Prop<unicode::Script> { { 0x109BC, 0x109BD }, unicode::Script::Meroitic_Cursive }, // No   [2] MEROITIC CURSIVE FRACTION ELEVEN TWELFTHS..MEROITIC CURSIVE FRACTION ONE HALF
    Prop<unicode::Script> { { 0x109BE, 0x109BF }, unicode::Script::Meroitic_Cursive }, // Lo   [2] MEROITIC CURSIVE LOGOGRAM RMT..MEROITIC CURSIVE LOGOGRAM IMN
    Prop<unicode::Script> { { 0x109C0, 0x109CF }, unicode::Script::Meroitic_Cursive }, // No  [16] MEROITIC CURSIVE NUMBER ONE..MEROITIC CURSIVE NUMBER SEVENTY
    Prop<unicode::Script> { { 0x109D2, 0x109FF }, unicode::Script::Meroitic_Cursive }, // No  [46] MEROITIC CURSIVE NUMBER ONE HUNDRED..MEROITIC CURSIVE FRACTION TEN TWELFTHS
    Prop<unicode::Script> { { 0x10A00, 0x10A00 }, unicode::Script::Kharoshthi }, // Lo       KHAROSHTHI LETTER A
    Prop<unicode::Script> { { 0x10A01, 0x10A03 }, unicode::Script::Kharoshthi }, // Mn   [3] KHAROSHTHI VOWEL SIGN I..KHAROSHTHI VOWEL SIGN VOCALIC R
    Prop<unicode::Script> { { 0x10A05, 0x10A06 }, unicode::Script::Kharoshthi }, // Mn   [2] KHAROSHTHI VOWEL SIGN E..KHAROSHTHI VOWEL SIGN O
    Prop<unicode::Script> { { 0x10A0C, 0x10A0F }, unicode::Script::Kharoshthi }, // Mn   [4] KHAROSHTHI VOWEL LENGTH MARK..KHAROSHTHI SIGN VISARGA
    Prop<unicode::Script> { { 0x10A10, 0x10A13 }, unicode::Script::Kharoshthi }, // Lo   [4] KHAROSHTHI LETTER KA..KHAROSHTHI LETTER GHA
    Prop<unicode::Script> { { 0x10A15, 0x10A17 }, unicode::Script::Kharoshthi }, // Lo   [3] KHAROSHTHI LETTER CA..KHAROSHTHI LETTER JA
    Prop<unicode::Script> { { 0x10A19, 0x10A35 }, unicode::Script::Kharoshthi }, // Lo  [29] KHAROSHTHI LETTER NYA..KHAROSHTHI LETTER VHA
    Prop<unicode::Script> { { 0x10A38, 0x10A3A }, unicode::Script::Kharoshthi }, // Mn   [3] KHAROSHTHI SIGN BAR ABOVE..KHAROSHTHI SIGN DOT BELOW
    Prop<unicode::Script> { { 0x10A3F, 0x10A3F }, unicode::Script::Kharoshthi }, // Mn       KHAROSHTHI VIRAMA
    Prop<unicode::Script> { { 0x10A40, 0x10A48 }, unicode::Script::Kharoshthi }, // No   [9] KHAROSHTHI DIGIT ONE..KHAROSHTHI FRACTION ONE HALF
    Prop<unicode::Script> { { 0x10A50, 0x10A58 }, unicode::Script::Kharoshthi }, // Po   [9] KHAROSHTHI PUNCTUATION DOT..KHAROSHTHI PUNCTUATION LINES
    Prop<unicode::Script> { { 0x10A60, 0x10A7C }, unicode::Script::Old_South_Arabian }, // Lo  [29] OLD SOUTH ARABIAN LETTER HE..OLD SOUTH ARABIAN LETTER THETH
    Prop<unicode::Script> { { 0x10A7D, 0x10A7E }, unicode::Script::Old_South_Arabian }, // No   [2] OLD SOUTH ARABIAN NUMBER ONE..OLD SOUTH ARABIAN NUMBER FIFTY
    Prop<unicode::Script> { { 0x10A7F, 0x10A7F }, unicode::Script::Old_South_Arabian }, // Po       OLD SOUTH ARABIAN NUMERIC INDICATOR
    Prop<unicode::Script> { { 0x10A80, 0x10A9C }, unicode::Script::Old_North_Arabian }, // Lo  [29] OLD NORTH ARABIAN LETTER HEH..OLD NORTH ARABIAN LETTER ZAH
    Prop<unicode::Script> { { 0x10A9D, 0x10A9F }, unicode::Script::Old_North_Arabian }, // No   [3] OLD NORTH ARABIAN NUMBER ONE..OLD NORTH ARABIAN NUMBER TWENTY
    Prop<unicode::Script> { { 0x10AC0, 0x10AC7 }, unicode::Script::Manichaean }, // Lo   [8] MANICHAEAN LETTER ALEPH..MANICHAEAN LETTER WAW
    Prop<unicode::Script> { { 0x10AC8, 0x10AC8 }, unicode::Script::Manichaean }, // So       MANICHAEAN SIGN UD
    Prop<unicode::Script> { { 0x10AC9, 0x10AE4 }, unicode::Script::Manichaean }, // Lo  [28] MANICHAEAN LETTER ZAYIN..MANICHAEAN LETTER TAW
    Prop<unicode::Script> { { 0x10AE5, 0x10AE6 }, unicode::Script::Manichaean }, // Mn   [2] MANICHAEAN ABBREVIATION MARK ABOVE..MANICHAEAN ABBREVIATION MARK BELOW
    Prop<unicode::Script> { { 0x10AEB, 0x10AEF }, unicode::Script::Manichaean }, // No   [5] MANICHAEAN NUMBER ONE..MANICHAEAN NUMBER ONE HUNDRED
    Prop<unicode::Script> { { 0x10AF0, 0x10AF6 }, unicode::Script::Manichaean }, // Po   [7] MANICHAEAN PUNCTUATION STAR..MANICHAEAN PUNCTUATION LINE FILLER
    Prop<unicode::Script> { { 0x10B00, 0x10B35 }, unicode::Script::Avestan }, // Lo  [54] AVESTAN LETTER A..AVESTAN LETTER HE
    Prop<unicode::Script> { { 0x10B39, 0x10B3F }, unicode::Script::Avestan }, // Po   [7] AVESTAN ABBREVIATION MARK..LARGE ONE RING OVER TWO RINGS PUNCTUATION
    Prop<unicode::Script> { { 0x10B40, 0x10B55 }, unicode::Script::Inscriptional_Parthian }, // Lo  [22] INSCRIPTIONAL PARTHIAN LETTER ALEPH..INSCRIPTIONAL PARTHIAN LETTER TAW
    Prop<unicode::Script> { { 0x10B58, 0x10B5F }, unicode::Script::Inscriptional_Parthian }, // No   [8] INSCRIPTIONAL PARTHIAN NUMBER ONE..INSCRIPTIONAL PARTHIAN NUMBER ONE THOUSAND
    Prop<unicode::Script> { { 0x10B60, 0x10B72 }, unicode::Script::Inscriptional_Pahlavi }, // Lo  [19] INSCRIPTIONAL PAHLAVI LETTER ALEPH..INSCRIPTIONAL PAHLAVI LETTER TAW
    Prop<unicode::Script> { { 0x10B78, 0x10B7F }, unicode::Script::Inscriptional_Pahlavi }, // No   [8] INSCRIPTIONAL PAHLAVI NUMBER ONE..INSCRIPTIONAL PAHLAVI NUMBER ONE THOUSAND
    Prop<unicode::Script> { { 0x10B80, 0x10B91 }, unicode::Script::Psalter_Pahlavi }, // Lo  [18] PSALTER PAHLAVI LETTER ALEPH..PSALTER PAHLAVI LETTER TAW
    Prop<unicode::Script> { { 0x10B99, 0x10B9C }, unicode::Script::Psalter_Pahlavi }, // Po   [4] PSALTER PAHLAVI SECTION MARK..PSALTER PAHLAVI FOUR DOTS WITH DOT
    Prop<unicode::Script> { { 0x10BA9, 0x10BAF }, unicode::Script::Psalter_Pahlavi }, // No   [7] PSALTER PAHLAVI NUMBER ONE..PSALTER PAHLAVI NUMBER ONE HUNDRED
    Prop<unicode::Script> { { 0x10C00, 0x10C48 }, unicode::Script::Old_Turkic }, // Lo  [73] OLD TURKIC LETTER ORKHON A..OLD TURKIC LETTER ORKHON BASH
    Prop<unicode::Script> { { 0x10C80, 0x10CB2 }, unicode::Script::Old_Hungarian }, // L&  [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Prop<unicode::Script> { { 0x10CC0, 0x10CF2 }, unicode::Script::Old_Hungarian }, // L&  [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    Prop<unicode::Script> { { 0x10CFA, 0x10CFF }, unicode::Script::Old_Hungarian }, // No   [6] OLD HUNGARIAN NUMBER ONE..OLD HUNGARIAN NUMBER ONE THOUSAND
    Prop<unicode::Script> { { 0x10D00, 0x10D23 }, unicode::Script::Hanifi_Rohingya }, // Lo  [36] HANIFI ROHINGYA LETTER A..HANIFI ROHINGYA MARK NA KHONNA
    Prop<unicode::Script> { { 0x10D24, 0x10D27 }, unicode::Script::Hanifi_Rohingya }, // Mn   [4] HANIFI ROHINGYA SIGN HARBAHAY..HANIFI ROHINGYA SIGN TASSI
    Prop<unicode::Script> { { 0x10D30, 0x10D39 }, unicode::Script::Hanifi_Rohingya }, // Nd  [10] HANIFI ROHINGYA DIGIT ZERO..HANIFI ROHINGYA DIGIT NINE
    Prop<unicode::Script> { { 0x10D40, 0x10D49 }, unicode::Script::Garay }, // Nd  [10] GARAY DIGIT ZERO..GARAY DIGIT NINE
    Prop<unicode::Script> { { 0x10D4A, 0x10D4D }, unicode::Script::Garay }, // Lo   [4] GARAY VOWEL SIGN A..GARAY VOWEL SIGN EE
    Prop<unicode::Script> { { 0x10D4E, 0x10D4E }, unicode::Script::Garay }, // Lm       GARAY VOWEL LENGTH MARK
    Prop<unicode::Script> { { 0x10D4F, 0x10D4F }, unicode::Script::Garay }, // Lo       GARAY SUKUN
    Prop<unicode::Script> { { 0x10D50, 0x10D65 }, unicode::Script::Garay }, // L&  [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Prop<unicode::Script> { { 0x10D69, 0x10D6D }, unicode::Script::Garay }, // Mn   [5] GARAY VOWEL SIGN E..GARAY CONSONANT NASALIZATION MARK
    Prop<unicode::Script> { { 0x10D6E, 0x10D6E }, unicode::Script::Garay }, // Pd       GARAY HYPHEN
    Prop<unicode::Script> { { 0x10D6F, 0x10D6F }, unicode::Script::Garay }, // Lm       GARAY REDUPLICATION MARK
    Prop<unicode::Script> { { 0x10D70, 0x10D85 }, unicode::Script::Garay }, // L&  [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Prop<unicode::Script> { { 0x10D8E, 0x10D8F }, unicode::Script::Garay }, // Sm   [2] GARAY PLUS SIGN..GARAY MINUS SIGN
    Prop<unicode::Script> { { 0x10E60, 0x10E7E }, unicode::Script::Arabic }, // No  [31] RUMI DIGIT ONE..RUMI FRACTION TWO THIRDS
    Prop<unicode::Script> { { 0x10E80, 0x10EA9 }, unicode::Script::Yezidi }, // Lo  [42] YEZIDI LETTER ELIF..YEZIDI LETTER ET
    Prop<unicode::Script> { { 0x10EAB, 0x10EAC }, unicode::Script::Yezidi }, // Mn   [2] YEZIDI COMBINING HAMZA MARK..YEZIDI COMBINING MADDA MARK
    Prop<unicode::Script> { { 0x10EAD, 0x10EAD }, unicode::Script::Yezidi }, // Pd       YEZIDI HYPHENATION MARK
    Prop<unicode::Script> { { 0x10EB0, 0x10EB1 }, unicode::Script::Yezidi }, // Lo   [2] YEZIDI LETTER LAM WITH DOT ABOVE..YEZIDI LETTER YOT WITH CIRCUMFLEX ABOVE
    Prop<unicode::Script> { { 0x10EC2, 0x10EC4 }, unicode::Script::Arabic }, // Lo   [3] ARABIC LETTER DAL WITH TWO DOTS VERTICALLY BELOW..ARABIC LETTER KAF WITH TWO DOTS VERTICALLY BELOW
    Prop<unicode::Script> { { 0x10EC5, 0x10EC5 }, unicode::Script::Arabic }, // Lm       ARABIC SMALL YEH BARREE WITH TWO DOTS BELOW
    Prop<unicode::Script> { { 0x10EC6, 0x10EC7 }, unicode::Script::Arabic }, // Lo   [2] ARABIC LETTER THIN NOON..ARABIC LETTER YEH WITH FOUR DOTS BELOW
    Prop<unicode::Script> { { 0x10ED0, 0x10ED0 }, unicode::Script::Arabic }, // Po       ARABIC BIBLICAL END OF VERSE
    Prop<unicode::Script> { { 0x10ED1, 0x10ED8 }, unicode::Script::Arabic }, // So   [8] ARABIC LIGATURE ALAYHAA AS-SALAATU WAS-SALAAM..ARABIC LIGATURE NAWWARA ALLAAHU MARQADAH
    Prop<unicode::Script> { { 0x10EFA, 0x10EFF }, unicode::Script::Arabic }, // Mn   [6] ARABIC DOUBLE VERTICAL BAR BELOW..ARABIC SMALL LOW WORD MADDA
    Prop<unicode::Script> { { 0x10F00, 0x10F1C }, unicode::Script::Old_Sogdian }, // Lo  [29] OLD SOGDIAN LETTER ALEPH..OLD SOGDIAN LETTER FINAL TAW WITH VERTICAL TAIL
    Prop<unicode::Script> { { 0x10F1D, 0x10F26 }, unicode::Script::Old_Sogdian }, // No  [10] OLD SOGDIAN NUMBER ONE..OLD SOGDIAN FRACTION ONE HALF
    Prop<unicode::Script> { { 0x10F27, 0x10F27 }, unicode::Script::Old_Sogdian }, // Lo       OLD SOGDIAN LIGATURE AYIN-DALETH
    Prop<unicode::Script> { { 0x10F30, 0x10F45 }, unicode::Script::Sogdian }, // Lo  [22] SOGDIAN LETTER ALEPH..SOGDIAN INDEPENDENT SHIN
    Prop<unicode::Script> { { 0x10F46, 0x10F50 }, unicode::Script::Sogdian }, // Mn  [11] SOGDIAN COMBINING DOT BELOW..SOGDIAN COMBINING STROKE BELOW
    Prop<unicode::Script> { { 0x10F51, 0x10F54 }, unicode::Script::Sogdian }, // No   [4] SOGDIAN NUMBER ONE..SOGDIAN NUMBER ONE HUNDRED
    Prop<unicode::Script> { { 0x10F55, 0x10F59 }, unicode::Script::Sogdian }, // Po   [5] SOGDIAN PUNCTUATION TWO VERTICAL BARS..SOGDIAN PUNCTUATION HALF CIRCLE WITH DOT
    Prop<unicode::Script> { { 0x10F70, 0x10F81 }, unicode::Script::Old_Uyghur }, // Lo  [18] OLD UYGHUR LETTER ALEPH..OLD UYGHUR LETTER LESH
    Prop<unicode::Script> { { 0x10F82, 0x10F85 }, unicode::Script::Old_Uyghur }, // Mn   [4] OLD UYGHUR COMBINING DOT ABOVE..OLD UYGHUR COMBINING TWO DOTS BELOW
    Prop<unicode::Script> { { 0x10F86, 0x10F89 }, unicode::Script::Old_Uyghur }, // Po   [4] OLD UYGHUR PUNCTUATION BAR..OLD UYGHUR PUNCTUATION FOUR DOTS
    Prop<unicode::Script> { { 0x10FB0, 0x10FC4 }, unicode::Script::Chorasmian }, // Lo  [21] CHORASMIAN LETTER ALEPH..CHORASMIAN LETTER TAW
    Prop<unicode::Script> { { 0x10FC5, 0x10FCB }, unicode::Script::Chorasmian }, // No   [7] CHORASMIAN NUMBER ONE..CHORASMIAN NUMBER ONE HUNDRED
    Prop<unicode::Script> { { 0x10FE0, 0x10FF6 }, unicode::Script::Elymaic }, // Lo  [23] ELYMAIC LETTER ALEPH..ELYMAIC LIGATURE ZAYIN-YODH
    Prop<unicode::Script> { { 0x11000, 0x11000 }, unicode::Script::Brahmi }, // Mc       BRAHMI SIGN CANDRABINDU
    Prop<unicode::Script> { { 0x11001, 0x11001 }, unicode::Script::Brahmi }, // Mn       BRAHMI SIGN ANUSVARA
    Prop<unicode::Script> { { 0x11002, 0x11002 }, unicode::Script::Brahmi }, // Mc       BRAHMI SIGN VISARGA
    Prop<unicode::Script> { { 0x11003, 0x11037 }, unicode::Script::Brahmi }, // Lo  [53] BRAHMI SIGN JIHVAMULIYA..BRAHMI LETTER OLD TAMIL NNNA
    Prop<unicode::Script> { { 0x11038, 0x11046 }, unicode::Script::Brahmi }, // Mn  [15] BRAHMI VOWEL SIGN AA..BRAHMI VIRAMA
    Prop<unicode::Script> { { 0x11047, 0x1104D }, unicode::Script::Brahmi }, // Po   [7] BRAHMI DANDA..BRAHMI PUNCTUATION LOTUS
    Prop<unicode::Script> { { 0x11052, 0x11065 }, unicode::Script::Brahmi }, // No  [20] BRAHMI NUMBER ONE..BRAHMI NUMBER ONE THOUSAND
    Prop<unicode::Script> { { 0x11066, 0x1106F }, unicode::Script::Brahmi }, // Nd  [10] BRAHMI DIGIT ZERO..BRAHMI DIGIT NINE
    Prop<unicode::Script> { { 0x11070, 0x11070 }, unicode::Script::Brahmi }, // Mn       BRAHMI SIGN OLD TAMIL VIRAMA
    Prop<unicode::Script> { { 0x11071, 0x11072 }, unicode::Script::Brahmi }, // Lo   [2] BRAHMI LETTER OLD TAMIL SHORT E..BRAHMI LETTER OLD TAMIL SHORT O
    Prop<unicode::Script> { { 0x11073, 0x11074 }, unicode::Script::Brahmi }, // Mn   [2] BRAHMI VOWEL SIGN OLD TAMIL SHORT E..BRAHMI VOWEL SIGN OLD TAMIL SHORT O
    Prop<unicode::Script> { { 0x11075, 0x11075 }, unicode::Script::Brahmi }, // Lo       BRAHMI LETTER OLD TAMIL LLA
    Prop<unicode::Script> { { 0x1107F, 0x1107F }, unicode::Script::Brahmi }, // Mn       BRAHMI NUMBER JOINER
    Prop<unicode::Script> { { 0x11080, 0x11081 }, unicode::Script::Kaithi }, // Mn   [2] KAITHI SIGN CANDRABINDU..KAITHI SIGN ANUSVARA
    Prop<unicode::Script> { { 0x11082, 0x11082 }, unicode::Script::Kaithi }, // Mc       KAITHI SIGN VISARGA
    Prop<unicode::Script> { { 0x11083, 0x110AF }, unicode::Script::Kaithi }, // Lo  [45] KAITHI LETTER A..KAITHI LETTER HA
    Prop<unicode::Script> { { 0x110B0, 0x110B2 }, unicode::Script::Kaithi }, // Mc   [3] KAITHI VOWEL SIGN AA..KAITHI VOWEL SIGN II
    Prop<unicode::Script> { { 0x110B3, 0x110B6 }, unicode::Script::Kaithi }, // Mn   [4] KAITHI VOWEL SIGN U..KAITHI VOWEL SIGN AI
    Prop<unicode::Script> { { 0x110B7, 0x110B8 }, unicode::Script::Kaithi }, // Mc   [2] KAITHI VOWEL SIGN O..KAITHI VOWEL SIGN AU
    Prop<unicode::Script> { { 0x110B9, 0x110BA }, unicode::Script::Kaithi }, // Mn   [2] KAITHI SIGN VIRAMA..KAITHI SIGN NUKTA
    Prop<unicode::Script> { { 0x110BB, 0x110BC }, unicode::Script::Kaithi }, // Po   [2] KAITHI ABBREVIATION SIGN..KAITHI ENUMERATION SIGN
    Prop<unicode::Script> { { 0x110BD, 0x110BD }, unicode::Script::Kaithi }, // Cf       KAITHI NUMBER SIGN
    Prop<unicode::Script> { { 0x110BE, 0x110C1 }, unicode::Script::Kaithi }, // Po   [4] KAITHI SECTION MARK..KAITHI DOUBLE DANDA
    Prop<unicode::Script> { { 0x110C2, 0x110C2 }, unicode::Script::Kaithi }, // Mn       KAITHI VOWEL SIGN VOCALIC R
    Prop<unicode::Script> { { 0x110CD, 0x110CD }, unicode::Script::Kaithi }, // Cf       KAITHI NUMBER SIGN ABOVE
    Prop<unicode::Script> { { 0x110D0, 0x110E8 }, unicode::Script::Sora_Sompeng }, // Lo  [25] SORA SOMPENG LETTER SAH..SORA SOMPENG LETTER MAE
    Prop<unicode::Script> { { 0x110F0, 0x110F9 }, unicode::Script::Sora_Sompeng }, // Nd  [10] SORA SOMPENG DIGIT ZERO..SORA SOMPENG DIGIT NINE
    Prop<unicode::Script> { { 0x11100, 0x11102 }, unicode::Script::Chakma }, // Mn   [3] CHAKMA SIGN CANDRABINDU..CHAKMA SIGN VISARGA
    Prop<unicode::Script> { { 0x11103, 0x11126 }, unicode::Script::Chakma }, // Lo  [36] CHAKMA LETTER AA..CHAKMA LETTER HAA
    Prop<unicode::Script> { { 0x11127, 0x1112B }, unicode::Script::Chakma }, // Mn   [5] CHAKMA VOWEL SIGN A..CHAKMA VOWEL SIGN UU
    Prop<unicode::Script> { { 0x1112C, 0x1112C }, unicode::Script::Chakma }, // Mc       CHAKMA VOWEL SIGN E
    Prop<unicode::Script> { { 0x1112D, 0x11134 }, unicode::Script::Chakma }, // Mn   [8] CHAKMA VOWEL SIGN AI..CHAKMA MAAYYAA
    Prop<unicode::Script> { { 0x11136, 0x1113F }, unicode::Script::Chakma }, // Nd  [10] CHAKMA DIGIT ZERO..CHAKMA DIGIT NINE
    Prop<unicode::Script> { { 0x11140, 0x11143 }, unicode::Script::Chakma }, // Po   [4] CHAKMA SECTION MARK..CHAKMA QUESTION MARK
    Prop<unicode::Script> { { 0x11144, 0x11144 }, unicode::Script::Chakma }, // Lo       CHAKMA LETTER LHAA
    Prop<unicode::Script> { { 0x11145, 0x11146 }, unicode::Script::Chakma }, // Mc   [2] CHAKMA VOWEL SIGN AA..CHAKMA VOWEL SIGN EI
    Prop<unicode::Script> { { 0x11147, 0x11147 }, unicode::Script::Chakma }, // Lo       CHAKMA LETTER VAA
    Prop<unicode::Script> { { 0x11150, 0x11172 }, unicode::Script::Mahajani }, // Lo  [35] MAHAJANI LETTER A..MAHAJANI LETTER RRA
    Prop<unicode::Script> { { 0x11173, 0x11173 }, unicode::Script::Mahajani }, // Mn       MAHAJANI SIGN NUKTA
    Prop<unicode::Script> { { 0x11174, 0x11175 }, unicode::Script::Mahajani }, // Po   [2] MAHAJANI ABBREVIATION SIGN..MAHAJANI SECTION MARK
    Prop<unicode::Script> { { 0x11176, 0x11176 }, unicode::Script::Mahajani }, // Lo       MAHAJANI LIGATURE SHRI
    Prop<unicode::Script> { { 0x11180, 0x11181 }, unicode::Script::Sharada }, // Mn   [2] SHARADA SIGN CANDRABINDU..SHARADA SIGN ANUSVARA
    Prop<unicode::Script> { { 0x11182, 0x11182 }, unicode::Script::Sharada }, // Mc       SHARADA SIGN VISARGA
    Prop<unicode::Script> { { 0x11183, 0x111B2 }, unicode::Script::Sharada }, // Lo  [48] SHARADA LETTER A..SHARADA LETTER HA
    Prop<unicode::Script> { { 0x111B3, 0x111B5 }, unicode::Script::Sharada }, // Mc   [3] SHARADA VOWEL SIGN AA..SHARADA VOWEL SIGN II
    Prop<unicode::Script> { { 0x111B6, 0x111BE }, unicode::Script::Sharada }, // Mn   [9] SHARADA VOWEL SIGN U..SHARADA VOWEL SIGN O
    Prop<unicode::Script> { { 0x111BF, 0x111C0 }, unicode::Script::Sharada }, // Mc   [2] SHARADA VOWEL SIGN AU..SHARADA SIGN VIRAMA
    Prop<unicode::Script> { { 0x111C1, 0x111C4 }, unicode::Script::Sharada }, // Lo   [4] SHARADA SIGN AVAGRAHA..SHARADA OM
    Prop<unicode::Script> { { 0x111C5, 0x111C8 }, unicode::Script::Sharada }, // Po   [4] SHARADA DANDA..SHARADA SEPARATOR
    Prop<unicode::Script> { { 0x111C9, 0x111CC }, unicode::Script::Sharada }, // Mn   [4] SHARADA SANDHI MARK..SHARADA EXTRA SHORT VOWEL MARK
    Prop<unicode::Script> { { 0x111CD, 0x111CD }, unicode::Script::Sharada }, // Po       SHARADA SUTRA MARK
    Prop<unicode::Script> { { 0x111CE, 0x111CE }, unicode::Script::Sharada }, // Mc       SHARADA VOWEL SIGN PRISHTHAMATRA E
    Prop<unicode::Script> { { 0x111CF, 0x111CF }, unicode::Script::Sharada }, // Mn       SHARADA SIGN INVERTED CANDRABINDU
    Prop<unicode::Script> { { 0x111D0, 0x111D9 }, unicode::Script::Sharada }, // Nd  [10] SHARADA DIGIT ZERO..SHARADA DIGIT NINE
    Prop<unicode::Script> { { 0x111DA, 0x111DA }, unicode::Script::Sharada }, // Lo       SHARADA EKAM
    Prop<unicode::Script> { { 0x111DB, 0x111DB }, unicode::Script::Sharada }, // Po       SHARADA SIGN SIDDHAM
    Prop<unicode::Script> { { 0x111DC, 0x111DC }, unicode::Script::Sharada }, // Lo       SHARADA HEADSTROKE
    Prop<unicode::Script> { { 0x111DD, 0x111DF }, unicode::Script::Sharada }, // Po   [3] SHARADA CONTINUATION SIGN..SHARADA SECTION MARK-2
    Prop<unicode::Script> { { 0x111E1, 0x111F4 }, unicode::Script::Sinhala }, // No  [20] SINHALA ARCHAIC DIGIT ONE..SINHALA ARCHAIC NUMBER ONE THOUSAND
    Prop<unicode::Script> { { 0x11200, 0x11211 }, unicode::Script::Khojki }, // Lo  [18] KHOJKI LETTER A..KHOJKI LETTER JJA
    Prop<unicode::Script> { { 0x11213, 0x1122B }, unicode::Script::Khojki }, // Lo  [25] KHOJKI LETTER NYA..KHOJKI LETTER LLA
    Prop<unicode::Script> { { 0x1122C, 0x1122E }, unicode::Script::Khojki }, // Mc   [3] KHOJKI VOWEL SIGN AA..KHOJKI VOWEL SIGN II
    Prop<unicode::Script> { { 0x1122F, 0x11231 }, unicode::Script::Khojki }, // Mn   [3] KHOJKI VOWEL SIGN U..KHOJKI VOWEL SIGN AI
    Prop<unicode::Script> { { 0x11232, 0x11233 }, unicode::Script::Khojki }, // Mc   [2] KHOJKI VOWEL SIGN O..KHOJKI VOWEL SIGN AU
    Prop<unicode::Script> { { 0x11234, 0x11234 }, unicode::Script::Khojki }, // Mn       KHOJKI SIGN ANUSVARA
    Prop<unicode::Script> { { 0x11235, 0x11235 }, unicode::Script::Khojki }, // Mc       KHOJKI SIGN VIRAMA
    Prop<unicode::Script> { { 0x11236, 0x11237 }, unicode::Script::Khojki }, // Mn   [2] KHOJKI SIGN NUKTA..KHOJKI SIGN SHADDA
    Prop<unicode::Script> { { 0x11238, 0x1123D }, unicode::Script::Khojki }, // Po   [6] KHOJKI DANDA..KHOJKI ABBREVIATION SIGN
    Prop<unicode::Script> { { 0x1123E, 0x1123E }, unicode::Script::Khojki }, // Mn       KHOJKI SIGN SUKUN
    Prop<unicode::Script> { { 0x1123F, 0x11240 }, unicode::Script::Khojki }, // Lo   [2] KHOJKI LETTER QA..KHOJKI LETTER SHORT I
    Prop<unicode::Script> { { 0x11241, 0x11241 }, unicode::Script::Khojki }, // Mn       KHOJKI VOWEL SIGN VOCALIC R
    Prop<unicode::Script> { { 0x11280, 0x11286 }, unicode::Script::Multani }, // Lo   [7] MULTANI LETTER A..MULTANI LETTER GA
    Prop<unicode::Script> { { 0x11288, 0x11288 }, unicode::Script::Multani }, // Lo       MULTANI LETTER GHA
    Prop<unicode::Script> { { 0x1128A, 0x1128D }, unicode::Script::Multani }, // Lo   [4] MULTANI LETTER CA..MULTANI LETTER JJA
    Prop<unicode::Script> { { 0x1128F, 0x1129D }, unicode::Script::Multani }, // Lo  [15] MULTANI LETTER NYA..MULTANI LETTER BA
    Prop<unicode::Script> { { 0x1129F, 0x112A8 }, unicode::Script::Multani }, // Lo  [10] MULTANI LETTER BHA..MULTANI LETTER RHA
    Prop<unicode::Script> { { 0x112A9, 0x112A9 }, unicode::Script::Multani }, // Po       MULTANI SECTION MARK
    Prop<unicode::Script> { { 0x112B0, 0x112DE }, unicode::Script::Khudawadi }, // Lo  [47] KHUDAWADI LETTER A..KHUDAWADI LETTER HA
    Prop<unicode::Script> { { 0x112DF, 0x112DF }, unicode::Script::Khudawadi }, // Mn       KHUDAWADI SIGN ANUSVARA
    Prop<unicode::Script> { { 0x112E0, 0x112E2 }, unicode::Script::Khudawadi }, // Mc   [3] KHUDAWADI VOWEL SIGN AA..KHUDAWADI VOWEL SIGN II
    Prop<unicode::Script> { { 0x112E3, 0x112EA }, unicode::Script::Khudawadi }, // Mn   [8] KHUDAWADI VOWEL SIGN U..KHUDAWADI SIGN VIRAMA
    Prop<unicode::Script> { { 0x112F0, 0x112F9 }, unicode::Script::Khudawadi }, // Nd  [10] KHUDAWADI DIGIT ZERO..KHUDAWADI DIGIT NINE
    Prop<unicode::Script> { { 0x11300, 0x11301 }, unicode::Script::Grantha }, // Mn   [2] GRANTHA SIGN COMBINING ANUSVARA ABOVE..GRANTHA SIGN CANDRABINDU
    Prop<unicode::Script> { { 0x11302, 0x11303 }, unicode::Script::Grantha }, // Mc   [2] GRANTHA SIGN ANUSVARA..GRANTHA SIGN VISARGA
    Prop<unicode::Script> { { 0x11305, 0x1130C }, unicode::Script::Grantha }, // Lo   [8] GRANTHA LETTER A..GRANTHA LETTER VOCALIC L
    Prop<unicode::Script> { { 0x1130F, 0x11310 }, unicode::Script::Grantha }, // Lo   [2] GRANTHA LETTER EE..GRANTHA LETTER AI
    Prop<unicode::Script> { { 0x11313, 0x11328 }, unicode::Script::Grantha }, // Lo  [22] GRANTHA LETTER OO..GRANTHA LETTER NA
    Prop<unicode::Script> { { 0x1132A, 0x11330 }, unicode::Script::Grantha }, // Lo   [7] GRANTHA LETTER PA..GRANTHA LETTER RA
    Prop<unicode::Script> { { 0x11332, 0x11333 }, unicode::Script::Grantha }, // Lo   [2] GRANTHA LETTER LA..GRANTHA LETTER LLA
    Prop<unicode::Script> { { 0x11335, 0x11339 }, unicode::Script::Grantha }, // Lo   [5] GRANTHA LETTER VA..GRANTHA LETTER HA
    Prop<unicode::Script> { { 0x1133B, 0x1133B }, unicode::Script::Inherited }, // Mn       COMBINING BINDU BELOW
    Prop<unicode::Script> { { 0x1133C, 0x1133C }, unicode::Script::Grantha }, // Mn       GRANTHA SIGN NUKTA
    Prop<unicode::Script> { { 0x1133D, 0x1133D }, unicode::Script::Grantha }, // Lo       GRANTHA SIGN AVAGRAHA
    Prop<unicode::Script> { { 0x1133E, 0x1133F }, unicode::Script::Grantha }, // Mc   [2] GRANTHA VOWEL SIGN AA..GRANTHA VOWEL SIGN I
    Prop<unicode::Script> { { 0x11340, 0x11340 }, unicode::Script::Grantha }, // Mn       GRANTHA VOWEL SIGN II
    Prop<unicode::Script> { { 0x11341, 0x11344 }, unicode::Script::Grantha }, // Mc   [4] GRANTHA VOWEL SIGN U..GRANTHA VOWEL SIGN VOCALIC RR
    Prop<unicode::Script> { { 0x11347, 0x11348 }, unicode::Script::Grantha }, // Mc   [2] GRANTHA VOWEL SIGN EE..GRANTHA VOWEL SIGN AI
    Prop<unicode::Script> { { 0x1134B, 0x1134D }, unicode::Script::Grantha }, // Mc   [3] GRANTHA VOWEL SIGN OO..GRANTHA SIGN VIRAMA
    Prop<unicode::Script> { { 0x11350, 0x11350 }, unicode::Script::Grantha }, // Lo       GRANTHA OM
    Prop<unicode::Script> { { 0x11357, 0x11357 }, unicode::Script::Grantha }, // Mc       GRANTHA AU LENGTH MARK
    Prop<unicode::Script> { { 0x1135D, 0x11361 }, unicode::Script::Grantha }, // Lo   [5] GRANTHA SIGN PLUTA..GRANTHA LETTER VOCALIC LL
    Prop<unicode::Script> { { 0x11362, 0x11363 }, unicode::Script::Grantha }, // Mc   [2] GRANTHA VOWEL SIGN VOCALIC L..GRANTHA VOWEL SIGN VOCALIC LL
    Prop<unicode::Script> { { 0x11366, 0x1136C }, unicode::Script::Grantha }, // Mn   [7] COMBINING GRANTHA DIGIT ZERO..COMBINING GRANTHA DIGIT SIX
    Prop<unicode::Script> { { 0x11370, 0x11374 }, unicode::Script::Grantha }, // Mn   [5] COMBINING GRANTHA LETTER A..COMBINING GRANTHA LETTER PA
    Prop<unicode::Script> { { 0x11380, 0x11389 }, unicode::Script::Tulu_Tigalari }, // Lo  [10] TULU-TIGALARI LETTER A..TULU-TIGALARI LETTER VOCALIC LL
    Prop<unicode::Script> { { 0x1138B, 0x1138B }, unicode::Script::Tulu_Tigalari }, // Lo       TULU-TIGALARI LETTER EE
    Prop<unicode::Script> { { 0x1138E, 0x1138E }, unicode::Script::Tulu_Tigalari }, // Lo       TULU-TIGALARI LETTER AI
    Prop<unicode::Script> { { 0x11390, 0x113B5 }, unicode::Script::Tulu_Tigalari }, // Lo  [38] TULU-TIGALARI LETTER OO..TULU-TIGALARI LETTER LLLA
    Prop<unicode::Script> { { 0x113B7, 0x113B7 }, unicode::Script::Tulu_Tigalari }, // Lo       TULU-TIGALARI SIGN AVAGRAHA
    Prop<unicode::Script> { { 0x113B8, 0x113BA }, unicode::Script::Tulu_Tigalari }, // Mc   [3] TULU-TIGALARI VOWEL SIGN AA..TULU-TIGALARI VOWEL SIGN II
    Prop<unicode::Script> { { 0x113BB, 0x113C0 }, unicode::Script::Tulu_Tigalari }, // Mn   [6] TULU-TIGALARI VOWEL SIGN U..TULU-TIGALARI VOWEL SIGN VOCALIC LL
    Prop<unicode::Script> { { 0x113C2, 0x113C2 }, unicode::Script::Tulu_Tigalari }, // Mc       TULU-TIGALARI VOWEL SIGN EE
    Prop<unicode::Script> { { 0x113C5, 0x113C5 }, unicode::Script::Tulu_Tigalari }, // Mc       TULU-TIGALARI VOWEL SIGN AI
    Prop<unicode::Script> { { 0x113C7, 0x113CA }, unicode::Script::Tulu_Tigalari }, // Mc   [4] TULU-TIGALARI VOWEL SIGN OO..TULU-TIGALARI SIGN CANDRA ANUNASIKA
    Prop<unicode::Script> { { 0x113CC, 0x113CD }, unicode::Script::Tulu_Tigalari }, // Mc   [2] TULU-TIGALARI SIGN ANUSVARA..TULU-TIGALARI SIGN VISARGA
    Prop<unicode::Script> { { 0x113CE, 0x113CE }, unicode::Script::Tulu_Tigalari }, // Mn       TULU-TIGALARI SIGN VIRAMA
    Prop<unicode::Script> { { 0x113CF, 0x113CF }, unicode::Script::Tulu_Tigalari }, // Mc       TULU-TIGALARI SIGN LOOPED VIRAMA
    Prop<unicode::Script> { { 0x113D0, 0x113D0 }, unicode::Script::Tulu_Tigalari }, // Mn       TULU-TIGALARI CONJOINER
    Prop<unicode::Script> { { 0x113D1, 0x113D1 }, unicode::Script::Tulu_Tigalari }, // Lo       TULU-TIGALARI REPHA
    Prop<unicode::Script> { { 0x113D2, 0x113D2 }, unicode::Script::Tulu_Tigalari }, // Mn       TULU-TIGALARI GEMINATION MARK
    Prop<unicode::Script> { { 0x113D3, 0x113D3 }, unicode::Script::Tulu_Tigalari }, // Lo       TULU-TIGALARI SIGN PLUTA
    Prop<unicode::Script> { { 0x113D4, 0x113D5 }, unicode::Script::Tulu_Tigalari }, // Po   [2] TULU-TIGALARI DANDA..TULU-TIGALARI DOUBLE DANDA
    Prop<unicode::Script> { { 0x113D7, 0x113D8 }, unicode::Script::Tulu_Tigalari }, // Po   [2] TULU-TIGALARI SIGN OM PUSHPIKA..TULU-TIGALARI SIGN SHRII PUSHPIKA
    Prop<unicode::Script> { { 0x113E1, 0x113E2 }, unicode::Script::Tulu_Tigalari }, // Mn   [2] TULU-TIGALARI VEDIC TONE SVARITA..TULU-TIGALARI VEDIC TONE ANUDATTA
    Prop<unicode::Script> { { 0x11400, 0x11434 }, unicode::Script::Newa }, // Lo  [53] NEWA LETTER A..NEWA LETTER HA
    Prop<unicode::Script> { { 0x11435, 0x11437 }, unicode::Script::Newa }, // Mc   [3] NEWA VOWEL SIGN AA..NEWA VOWEL SIGN II
    Prop<unicode::Script> { { 0x11438, 0x1143F }, unicode::Script::Newa }, // Mn   [8] NEWA VOWEL SIGN U..NEWA VOWEL SIGN AI
    Prop<unicode::Script> { { 0x11440, 0x11441 }, unicode::Script::Newa }, // Mc   [2] NEWA VOWEL SIGN O..NEWA VOWEL SIGN AU
    Prop<unicode::Script> { { 0x11442, 0x11444 }, unicode::Script::Newa }, // Mn   [3] NEWA SIGN VIRAMA..NEWA SIGN ANUSVARA
    Prop<unicode::Script> { { 0x11445, 0x11445 }, unicode::Script::Newa }, // Mc       NEWA SIGN VISARGA
    Prop<unicode::Script> { { 0x11446, 0x11446 }, unicode::Script::Newa }, // Mn       NEWA SIGN NUKTA
    Prop<unicode::Script> { { 0x11447, 0x1144A }, unicode::Script::Newa }, // Lo   [4] NEWA SIGN AVAGRAHA..NEWA SIDDHI
    Prop<unicode::Script> { { 0x1144B, 0x1144F }, unicode::Script::Newa }, // Po   [5] NEWA DANDA..NEWA ABBREVIATION SIGN
    Prop<unicode::Script> { { 0x11450, 0x11459 }, unicode::Script::Newa }, // Nd  [10] NEWA DIGIT ZERO..NEWA DIGIT NINE
    Prop<unicode::Script> { { 0x1145A, 0x1145B }, unicode::Script::Newa }, // Po   [2] NEWA DOUBLE COMMA..NEWA PLACEHOLDER MARK
    Prop<unicode::Script> { { 0x1145D, 0x1145D }, unicode::Script::Newa }, // Po       NEWA INSERTION SIGN
    Prop<unicode::Script> { { 0x1145E, 0x1145E }, unicode::Script::Newa }, // Mn       NEWA SANDHI MARK
    Prop<unicode::Script> { { 0x1145F, 0x11461 }, unicode::Script::Newa }, // Lo   [3] NEWA LETTER VEDIC ANUSVARA..NEWA SIGN UPADHMANIYA
    Prop<unicode::Script> { { 0x11480, 0x114AF }, unicode::Script::Tirhuta }, // Lo  [48] TIRHUTA ANJI..TIRHUTA LETTER HA
    Prop<unicode::Script> { { 0x114B0, 0x114B2 }, unicode::Script::Tirhuta }, // Mc   [3] TIRHUTA VOWEL SIGN AA..TIRHUTA VOWEL SIGN II
    Prop<unicode::Script> { { 0x114B3, 0x114B8 }, unicode::Script::Tirhuta }, // Mn   [6] TIRHUTA VOWEL SIGN U..TIRHUTA VOWEL SIGN VOCALIC LL
    Prop<unicode::Script> { { 0x114B9, 0x114B9 }, unicode::Script::Tirhuta }, // Mc       TIRHUTA VOWEL SIGN E
    Prop<unicode::Script> { { 0x114BA, 0x114BA }, unicode::Script::Tirhuta }, // Mn       TIRHUTA VOWEL SIGN SHORT E
    Prop<unicode::Script> { { 0x114BB, 0x114BE }, unicode::Script::Tirhuta }, // Mc   [4] TIRHUTA VOWEL SIGN AI..TIRHUTA VOWEL SIGN AU
    Prop<unicode::Script> { { 0x114BF, 0x114C0 }, unicode::Script::Tirhuta }, // Mn   [2] TIRHUTA SIGN CANDRABINDU..TIRHUTA SIGN ANUSVARA
    Prop<unicode::Script> { { 0x114C1, 0x114C1 }, unicode::Script::Tirhuta }, // Mc       TIRHUTA SIGN VISARGA
    Prop<unicode::Script> { { 0x114C2, 0x114C3 }, unicode::Script::Tirhuta }, // Mn   [2] TIRHUTA SIGN VIRAMA..TIRHUTA SIGN NUKTA
    Prop<unicode::Script> { { 0x114C4, 0x114C5 }, unicode::Script::Tirhuta }, // Lo   [2] TIRHUTA SIGN AVAGRAHA..TIRHUTA GVANG
    Prop<unicode::Script> { { 0x114C6, 0x114C6 }, unicode::Script::Tirhuta }, // Po       TIRHUTA ABBREVIATION SIGN
    Prop<unicode::Script> { { 0x114C7, 0x114C7 }, unicode::Script::Tirhuta }, // Lo       TIRHUTA OM
    Prop<unicode::Script> { { 0x114D0, 0x114D9 }, unicode::Script::Tirhuta }, // Nd  [10] TIRHUTA DIGIT ZERO..TIRHUTA DIGIT NINE
    Prop<unicode::Script> { { 0x11580, 0x115AE }, unicode::Script::Siddham }, // Lo  [47] SIDDHAM LETTER A..SIDDHAM LETTER HA
    Prop<unicode::Script> { { 0x115AF, 0x115B1 }, unicode::Script::Siddham }, // Mc   [3] SIDDHAM VOWEL SIGN AA..SIDDHAM VOWEL SIGN II
    Prop<unicode::Script> { { 0x115B2, 0x115B5 }, unicode::Script::Siddham }, // Mn   [4] SIDDHAM VOWEL SIGN U..SIDDHAM VOWEL SIGN VOCALIC RR
    Prop<unicode::Script> { { 0x115B8, 0x115BB }, unicode::Script::Siddham }, // Mc   [4] SIDDHAM VOWEL SIGN E..SIDDHAM VOWEL SIGN AU
    Prop<unicode::Script> { { 0x115BC, 0x115BD }, unicode::Script::Siddham }, // Mn   [2] SIDDHAM SIGN CANDRABINDU..SIDDHAM SIGN ANUSVARA
    Prop<unicode::Script> { { 0x115BE, 0x115BE }, unicode::Script::Siddham }, // Mc       SIDDHAM SIGN VISARGA
    Prop<unicode::Script> { { 0x115BF, 0x115C0 }, unicode::Script::Siddham }, // Mn   [2] SIDDHAM SIGN VIRAMA..SIDDHAM SIGN NUKTA
    Prop<unicode::Script> { { 0x115C1, 0x115D7 }, unicode::Script::Siddham }, // Po  [23] SIDDHAM SIGN SIDDHAM..SIDDHAM SECTION MARK WITH CIRCLES AND FOUR ENCLOSURES
    Prop<unicode::Script> { { 0x115D8, 0x115DB }, unicode::Script::Siddham }, // Lo   [4] SIDDHAM LETTER THREE-CIRCLE ALTERNATE I..SIDDHAM LETTER ALTERNATE U
    Prop<unicode::Script> { { 0x115DC, 0x115DD }, unicode::Script::Siddham }, // Mn   [2] SIDDHAM VOWEL SIGN ALTERNATE U..SIDDHAM VOWEL SIGN ALTERNATE UU
    Prop<unicode::Script> { { 0x11600, 0x1162F }, unicode::Script::Modi }, // Lo  [48] MODI LETTER A..MODI LETTER LLA
    Prop<unicode::Script> { { 0x11630, 0x11632 }, unicode::Script::Modi }, // Mc   [3] MODI VOWEL SIGN AA..MODI VOWEL SIGN II
    Prop<unicode::Script> { { 0x11633, 0x1163A }, unicode::Script::Modi }, // Mn   [8] MODI VOWEL SIGN U..MODI VOWEL SIGN AI
    Prop<unicode::Script> { { 0x1163B, 0x1163C }, unicode::Script::Modi }, // Mc   [2] MODI VOWEL SIGN O..MODI VOWEL SIGN AU
    Prop<unicode::Script> { { 0x1163D, 0x1163D }, unicode::Script::Modi }, // Mn       MODI SIGN ANUSVARA
    Prop<unicode::Script> { { 0x1163E, 0x1163E }, unicode::Script::Modi }, // Mc       MODI SIGN VISARGA
    Prop<unicode::Script> { { 0x1163F, 0x11640 }, unicode::Script::Modi }, // Mn   [2] MODI SIGN VIRAMA..MODI SIGN ARDHACANDRA
    Prop<unicode::Script> { { 0x11641, 0x11643 }, unicode::Script::Modi }, // Po   [3] MODI DANDA..MODI ABBREVIATION SIGN
    Prop<unicode::Script> { { 0x11644, 0x11644 }, unicode::Script::Modi }, // Lo       MODI SIGN HUVA
    Prop<unicode::Script> { { 0x11650, 0x11659 }, unicode::Script::Modi }, // Nd  [10] MODI DIGIT ZERO..MODI DIGIT NINE
    Prop<unicode::Script> { { 0x11660, 0x1166C }, unicode::Script::Mongolian }, // Po  [13] MONGOLIAN BIRGA WITH ORNAMENT..MONGOLIAN TURNED SWIRL BIRGA WITH DOUBLE ORNAMENT
    Prop<unicode::Script> { { 0x11680, 0x116AA }, unicode::Script::Takri }, // Lo  [43] TAKRI LETTER A..TAKRI LETTER RRA
    Prop<unicode::Script> { { 0x116AB, 0x116AB }, unicode::Script::Takri }, // Mn       TAKRI SIGN ANUSVARA
    Prop<unicode::Script> { { 0x116AC, 0x116AC }, unicode::Script::Takri }, // Mc       TAKRI SIGN VISARGA
    Prop<unicode::Script> { { 0x116AD, 0x116AD }, unicode::Script::Takri }, // Mn       TAKRI VOWEL SIGN AA
    Prop<unicode::Script> { { 0x116AE, 0x116AF }, unicode::Script::Takri }, // Mc   [2] TAKRI VOWEL SIGN I..TAKRI VOWEL SIGN II
    Prop<unicode::Script> { { 0x116B0, 0x116B5 }, unicode::Script::Takri }, // Mn   [6] TAKRI VOWEL SIGN U..TAKRI VOWEL SIGN AU
    Prop<unicode::Script> { { 0x116B6, 0x116B6 }, unicode::Script::Takri }, // Mc       TAKRI SIGN VIRAMA
    Prop<unicode::Script> { { 0x116B7, 0x116B7 }, unicode::Script::Takri }, // Mn       TAKRI SIGN NUKTA
    Prop<unicode::Script> { { 0x116B8, 0x116B8 }, unicode::Script::Takri }, // Lo       TAKRI LETTER ARCHAIC KHA
    Prop<unicode::Script> { { 0x116B9, 0x116B9 }, unicode::Script::Takri }, // Po       TAKRI ABBREVIATION SIGN
    Prop<unicode::Script> { { 0x116C0, 0x116C9 }, unicode::Script::Takri }, // Nd  [10] TAKRI DIGIT ZERO..TAKRI DIGIT NINE
    Prop<unicode::Script> { { 0x116D0, 0x116E3 }, unicode::Script::Myanmar }, // Nd  [20] MYANMAR PAO DIGIT ZERO..MYANMAR EASTERN PWO KAREN DIGIT NINE
    Prop<unicode::Script> { { 0x11700, 0x1171A }, unicode::Script::Ahom }, // Lo  [27] AHOM LETTER KA..AHOM LETTER ALTERNATE BA
    Prop<unicode::Script> { { 0x1171D, 0x1171D }, unicode::Script::Ahom }, // Mn       AHOM CONSONANT SIGN MEDIAL LA
    Prop<unicode::Script> { { 0x1171E, 0x1171E }, unicode::Script::Ahom }, // Mc       AHOM CONSONANT SIGN MEDIAL RA
    Prop<unicode::Script> { { 0x1171F, 0x1171F }, unicode::Script::Ahom }, // Mn       AHOM CONSONANT SIGN MEDIAL LIGATING RA
    Prop<unicode::Script> { { 0x11720, 0x11721 }, unicode::Script::Ahom }, // Mc   [2] AHOM VOWEL SIGN A..AHOM VOWEL SIGN AA
    Prop<unicode::Script> { { 0x11722, 0x11725 }, unicode::Script::Ahom }, // Mn   [4] AHOM VOWEL SIGN I..AHOM VOWEL SIGN UU
    Prop<unicode::Script> { { 0x11726, 0x11726 }, unicode::Script::Ahom }, // Mc       AHOM VOWEL SIGN E
    Prop<unicode::Script> { { 0x11727, 0x1172B }, unicode::Script::Ahom }, // Mn   [5] AHOM VOWEL SIGN AW..AHOM SIGN KILLER
    Prop<unicode::Script> { { 0x11730, 0x11739 }, unicode::Script::Ahom }, // Nd  [10] AHOM DIGIT ZERO..AHOM DIGIT NINE
    Prop<unicode::Script> { { 0x1173A, 0x1173B }, unicode::Script::Ahom }, // No   [2] AHOM NUMBER TEN..AHOM NUMBER TWENTY
    Prop<unicode::Script> { { 0x1173C, 0x1173E }, unicode::Script::Ahom }, // Po   [3] AHOM SIGN SMALL SECTION..AHOM SIGN RULAI
    Prop<unicode::Script> { { 0x1173F, 0x1173F }, unicode::Script::Ahom }, // So       AHOM SYMBOL VI
    Prop<unicode::Script> { { 0x11740, 0x11746 }, unicode::Script::Ahom }, // Lo   [7] AHOM LETTER CA..AHOM LETTER LLA
    Prop<unicode::Script> { { 0x11800, 0x1182B }, unicode::Script::Dogra }, // Lo  [44] DOGRA LETTER A..DOGRA LETTER RRA
    Prop<unicode::Script> { { 0x1182C, 0x1182E }, unicode::Script::Dogra }, // Mc   [3] DOGRA VOWEL SIGN AA..DOGRA VOWEL SIGN II
    Prop<unicode::Script> { { 0x1182F, 0x11837 }, unicode::Script::Dogra }, // Mn   [9] DOGRA VOWEL SIGN U..DOGRA SIGN ANUSVARA
    Prop<unicode::Script> { { 0x11838, 0x11838 }, unicode::Script::Dogra }, // Mc       DOGRA SIGN VISARGA
    Prop<unicode::Script> { { 0x11839, 0x1183A }, unicode::Script::Dogra }, // Mn   [2] DOGRA SIGN VIRAMA..DOGRA SIGN NUKTA
    Prop<unicode::Script> { { 0x1183B, 0x1183B }, unicode::Script::Dogra }, // Po       DOGRA ABBREVIATION SIGN
    Prop<unicode::Script> { { 0x118A0, 0x118DF }, unicode::Script::Warang_Citi }, // L&  [64] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    Prop<unicode::Script> { { 0x118E0, 0x118E9 }, unicode::Script::Warang_Citi }, // Nd  [10] WARANG CITI DIGIT ZERO..WARANG CITI DIGIT NINE
    Prop<unicode::Script> { { 0x118EA, 0x118F2 }, unicode::Script::Warang_Citi }, // No   [9] WARANG CITI NUMBER TEN..WARANG CITI NUMBER NINETY
    Prop<unicode::Script> { { 0x118FF, 0x118FF }, unicode::Script::Warang_Citi }, // Lo       WARANG CITI OM
    Prop<unicode::Script> { { 0x11900, 0x11906 }, unicode::Script::Dives_Akuru }, // Lo   [7] DIVES AKURU LETTER A..DIVES AKURU LETTER E
    Prop<unicode::Script> { { 0x11909, 0x11909 }, unicode::Script::Dives_Akuru }, // Lo       DIVES AKURU LETTER O
    Prop<unicode::Script> { { 0x1190C, 0x11913 }, unicode::Script::Dives_Akuru }, // Lo   [8] DIVES AKURU LETTER KA..DIVES AKURU LETTER JA
    Prop<unicode::Script> { { 0x11915, 0x11916 }, unicode::Script::Dives_Akuru }, // Lo   [2] DIVES AKURU LETTER NYA..DIVES AKURU LETTER TTA
    Prop<unicode::Script> { { 0x11918, 0x1192F }, unicode::Script::Dives_Akuru }, // Lo  [24] DIVES AKURU LETTER DDA..DIVES AKURU LETTER ZA
    Prop<unicode::Script> { { 0x11930, 0x11935 }, unicode::Script::Dives_Akuru }, // Mc   [6] DIVES AKURU VOWEL SIGN AA..DIVES AKURU VOWEL SIGN E
    Prop<unicode::Script> { { 0x11937, 0x11938 }, unicode::Script::Dives_Akuru }, // Mc   [2] DIVES AKURU VOWEL SIGN AI..DIVES AKURU VOWEL SIGN O
    Prop<unicode::Script> { { 0x1193B, 0x1193C }, unicode::Script::Dives_Akuru }, // Mn   [2] DIVES AKURU SIGN ANUSVARA..DIVES AKURU SIGN CANDRABINDU
    Prop<unicode::Script> { { 0x1193D, 0x1193D }, unicode::Script::Dives_Akuru }, // Mc       DIVES AKURU SIGN HALANTA
    Prop<unicode::Script> { { 0x1193E, 0x1193E }, unicode::Script::Dives_Akuru }, // Mn       DIVES AKURU VIRAMA
    Prop<unicode::Script> { { 0x1193F, 0x1193F }, unicode::Script::Dives_Akuru }, // Lo       DIVES AKURU PREFIXED NASAL SIGN
    Prop<unicode::Script> { { 0x11940, 0x11940 }, unicode::Script::Dives_Akuru }, // Mc       DIVES AKURU MEDIAL YA
    Prop<unicode::Script> { { 0x11941, 0x11941 }, unicode::Script::Dives_Akuru }, // Lo       DIVES AKURU INITIAL RA
    Prop<unicode::Script> { { 0x11942, 0x11942 }, unicode::Script::Dives_Akuru }, // Mc       DIVES AKURU MEDIAL RA
    Prop<unicode::Script> { { 0x11943, 0x11943 }, unicode::Script::Dives_Akuru }, // Mn       DIVES AKURU SIGN NUKTA
    Prop<unicode::Script> { { 0x11944, 0x11946 }, unicode::Script::Dives_Akuru }, // Po   [3] DIVES AKURU DOUBLE DANDA..DIVES AKURU END OF TEXT MARK
    Prop<unicode::Script> { { 0x11950, 0x11959 }, unicode::Script::Dives_Akuru }, // Nd  [10] DIVES AKURU DIGIT ZERO..DIVES AKURU DIGIT NINE
    Prop<unicode::Script> { { 0x119A0, 0x119A7 }, unicode::Script::Nandinagari }, // Lo   [8] NANDINAGARI LETTER A..NANDINAGARI LETTER VOCALIC RR
    Prop<unicode::Script> { { 0x119AA, 0x119D0 }, unicode::Script::Nandinagari }, // Lo  [39] NANDINAGARI LETTER E..NANDINAGARI LETTER RRA
    Prop<unicode::Script> { { 0x119D1, 0x119D3 }, unicode::Script::Nandinagari }, // Mc   [3] NANDINAGARI VOWEL SIGN AA..NANDINAGARI VOWEL SIGN II
    Prop<unicode::Script> { { 0x119D4, 0x119D7 }, unicode::Script::Nandinagari }, // Mn   [4] NANDINAGARI VOWEL SIGN U..NANDINAGARI VOWEL SIGN VOCALIC RR
    Prop<unicode::Script> { { 0x119DA, 0x119DB }, unicode::Script::Nandinagari }, // Mn   [2] NANDINAGARI VOWEL SIGN E..NANDINAGARI VOWEL SIGN AI
    Prop<unicode::Script> { { 0x119DC, 0x119DF }, unicode::Script::Nandinagari }, // Mc   [4] NANDINAGARI VOWEL SIGN O..NANDINAGARI SIGN VISARGA
    Prop<unicode::Script> { { 0x119E0, 0x119E0 }, unicode::Script::Nandinagari }, // Mn       NANDINAGARI SIGN VIRAMA
    Prop<unicode::Script> { { 0x119E1, 0x119E1 }, unicode::Script::Nandinagari }, // Lo       NANDINAGARI SIGN AVAGRAHA
    Prop<unicode::Script> { { 0x119E2, 0x119E2 }, unicode::Script::Nandinagari }, // Po       NANDINAGARI SIGN SIDDHAM
    Prop<unicode::Script> { { 0x119E3, 0x119E3 }, unicode::Script::Nandinagari }, // Lo       NANDINAGARI HEADSTROKE
    Prop<unicode::Script> { { 0x119E4, 0x119E4 }, unicode::Script::Nandinagari }, // Mc       NANDINAGARI VOWEL SIGN PRISHTHAMATRA E
    Prop<unicode::Script> { { 0x11A00, 0x11A00 }, unicode::Script::Zanabazar_Square }, // Lo       ZANABAZAR SQUARE LETTER A
    Prop<unicode::Script> { { 0x11A01, 0x11A0A }, unicode::Script::Zanabazar_Square }, // Mn  [10] ZANABAZAR SQUARE VOWEL SIGN I..ZANABAZAR SQUARE VOWEL LENGTH MARK
    Prop<unicode::Script> { { 0x11A0B, 0x11A32 }, unicode::Script::Zanabazar_Square }, // Lo  [40] ZANABAZAR SQUARE LETTER KA..ZANABAZAR SQUARE LETTER KSSA
    Prop<unicode::Script> { { 0x11A33, 0x11A38 }, unicode::Script::Zanabazar_Square }, // Mn   [6] ZANABAZAR SQUARE FINAL CONSONANT MARK..ZANABAZAR SQUARE SIGN ANUSVARA
    Prop<unicode::Script> { { 0x11A39, 0x11A39 }, unicode::Script::Zanabazar_Square }, // Mc       ZANABAZAR SQUARE SIGN VISARGA
    Prop<unicode::Script> { { 0x11A3A, 0x11A3A }, unicode::Script::Zanabazar_Square }, // Lo       ZANABAZAR SQUARE CLUSTER-INITIAL LETTER RA
    Prop<unicode::Script> { { 0x11A3B, 0x11A3E }, unicode::Script::Zanabazar_Square }, // Mn   [4] ZANABAZAR SQUARE CLUSTER-FINAL LETTER YA..ZANABAZAR SQUARE CLUSTER-FINAL LETTER VA
    Prop<unicode::Script> { { 0x11A3F, 0x11A46 }, unicode::Script::Zanabazar_Square }, // Po   [8] ZANABAZAR SQUARE INITIAL HEAD MARK..ZANABAZAR SQUARE CLOSING DOUBLE-LINED HEAD MARK
    Prop<unicode::Script> { { 0x11A47, 0x11A47 }, unicode::Script::Zanabazar_Square }, // Mn       ZANABAZAR SQUARE SUBJOINER
    Prop<unicode::Script> { { 0x11A50, 0x11A50 }, unicode::Script::Soyombo }, // Lo       SOYOMBO LETTER A
    Prop<unicode::Script> { { 0x11A51, 0x11A56 }, unicode::Script::Soyombo }, // Mn   [6] SOYOMBO VOWEL SIGN I..SOYOMBO VOWEL SIGN OE
    Prop<unicode::Script> { { 0x11A57, 0x11A58 }, unicode::Script::Soyombo }, // Mc   [2] SOYOMBO VOWEL SIGN AI..SOYOMBO VOWEL SIGN AU
    Prop<unicode::Script> { { 0x11A59, 0x11A5B }, unicode::Script::Soyombo }, // Mn   [3] SOYOMBO VOWEL SIGN VOCALIC R..SOYOMBO VOWEL LENGTH MARK
    Prop<unicode::Script> { { 0x11A5C, 0x11A89 }, unicode::Script::Soyombo }, // Lo  [46] SOYOMBO LETTER KA..SOYOMBO CLUSTER-INITIAL LETTER SA
    Prop<unicode::Script> { { 0x11A8A, 0x11A96 }, unicode::Script::Soyombo }, // Mn  [13] SOYOMBO FINAL CONSONANT SIGN G..SOYOMBO SIGN ANUSVARA
    Prop<unicode::Script> { { 0x11A97, 0x11A97 }, unicode::Script::Soyombo }, // Mc       SOYOMBO SIGN VISARGA
    Prop<unicode::Script> { { 0x11A98, 0x11A99 }, unicode::Script::Soyombo }, // Mn   [2] SOYOMBO GEMINATION MARK..SOYOMBO SUBJOINER
    Prop<unicode::Script> { { 0x11A9A, 0x11A9C }, unicode::Script::Soyombo }, // Po   [3] SOYOMBO MARK TSHEG..SOYOMBO MARK DOUBLE SHAD
    Prop<unicode::Script> { { 0x11A9D, 0x11A9D }, unicode::Script::Soyombo }, // Lo       SOYOMBO MARK PLUTA
    Prop<unicode::Script> { { 0x11A9E, 0x11AA2 }, unicode::Script::Soyombo }, // Po   [5] SOYOMBO HEAD MARK WITH MOON AND SUN AND TRIPLE FLAME..SOYOMBO TERMINAL MARK-2
    Prop<unicode::Script> { { 0x11AB0, 0x11ABF }, unicode::Script::Canadian_Aboriginal }, // Lo  [16] CANADIAN SYLLABICS NATTILIK HI..CANADIAN SYLLABICS SPA
    Prop<unicode::Script> { { 0x11AC0, 0x11AF8 }, unicode::Script::Pau_Cin_Hau }, // Lo  [57] PAU CIN HAU LETTER PA..PAU CIN HAU GLOTTAL STOP FINAL
    Prop<unicode::Script> { { 0x11B00, 0x11B09 }, unicode::Script::Devanagari }, // Po  [10] DEVANAGARI HEAD MARK..DEVANAGARI SIGN MINDU
    Prop<unicode::Script> { { 0x11B60, 0x11B60 }, unicode::Script::Sharada }, // Mn       SHARADA VOWEL SIGN OE
    Prop<unicode::Script> { { 0x11B61, 0x11B61 }, unicode::Script::Sharada }, // Mc       SHARADA VOWEL SIGN OOE
    Prop<unicode::Script> { { 0x11B62, 0x11B64 }, unicode::Script::Sharada }, // Mn   [3] SHARADA VOWEL SIGN UE..SHARADA VOWEL SIGN SHORT E
    Prop<unicode::Script> { { 0x11B65, 0x11B65 }, unicode::Script::Sharada }, // Mc       SHARADA VOWEL SIGN SHORT O
    Prop<unicode::Script> { { 0x11B66, 0x11B66 }, unicode::Script::Sharada }, // Mn       SHARADA VOWEL SIGN CANDRA E
    Prop<unicode::Script> { { 0x11B67, 0x11B67 }, unicode::Script::Sharada }, // Mc       SHARADA VOWEL SIGN CANDRA O
    Prop<unicode::Script> { { 0x11BC0, 0x11BE0 }, unicode::Script::Sunuwar }, // Lo  [33] SUNUWAR LETTER DEVI..SUNUWAR LETTER KLOKO
    Prop<unicode::Script> { { 0x11BE1, 0x11BE1 }, unicode::Script::Sunuwar }, // Po       SUNUWAR SIGN PVO
    Prop<unicode::Script> { { 0x11BF0, 0x11BF9 }, unicode::Script::Sunuwar }, // Nd  [10] SUNUWAR DIGIT ZERO..SUNUWAR DIGIT NINE
    Prop<unicode::Script> { { 0x11C00, 0x11C08 }, unicode::Script::Bhaiksuki }, // Lo   [9] BHAIKSUKI LETTER A..BHAIKSUKI LETTER VOCALIC L
    Prop<unicode::Script> { { 0x11C0A, 0x11C2E }, unicode::Script::Bhaiksuki }, // Lo  [37] BHAIKSUKI LETTER E..BHAIKSUKI LETTER HA
    Prop<unicode::Script> { { 0x11C2F, 0x11C2F }, unicode::Script::Bhaiksuki }, // Mc       BHAIKSUKI VOWEL SIGN AA
    Prop<unicode::Script> { { 0x11C30, 0x11C36 }, unicode::Script::Bhaiksuki }, // Mn   [7] BHAIKSUKI VOWEL SIGN I..BHAIKSUKI VOWEL SIGN VOCALIC L
    Prop<unicode::Script> { { 0x11C38, 0x11C3D }, unicode::Script::Bhaiksuki }, // Mn   [6] BHAIKSUKI VOWEL SIGN E..BHAIKSUKI SIGN ANUSVARA
    Prop<unicode::Script> { { 0x11C3E, 0x11C3E }, unicode::Script::Bhaiksuki }, // Mc       BHAIKSUKI SIGN VISARGA
    Prop<unicode::Script> { { 0x11C3F, 0x11C3F }, unicode::Script::Bhaiksuki }, // Mn       BHAIKSUKI SIGN VIRAMA
    Prop<unicode::Script> { { 0x11C40, 0x11C40 }, unicode::Script::Bhaiksuki }, // Lo       BHAIKSUKI SIGN AVAGRAHA
    Prop<unicode::Script> { { 0x11C41, 0x11C45 }, unicode::Script::Bhaiksuki }, // Po   [5] BHAIKSUKI DANDA..BHAIKSUKI GAP FILLER-2
    Prop<unicode::Script> { { 0x11C50, 0x11C59 }, unicode::Script::Bhaiksuki }, // Nd  [10] BHAIKSUKI DIGIT ZERO..BHAIKSUKI DIGIT NINE
    Prop<unicode::Script> { { 0x11C5A, 0x11C6C }, unicode::Script::Bhaiksuki }, // No  [19] BHAIKSUKI NUMBER ONE..BHAIKSUKI HUNDREDS UNIT MARK
    Prop<unicode::Script> { { 0x11C70, 0x11C71 }, unicode::Script::Marchen }, // Po   [2] MARCHEN HEAD MARK..MARCHEN MARK SHAD
    Prop<unicode::Script> { { 0x11C72, 0x11C8F }, unicode::Script::Marchen }, // Lo  [30] MARCHEN LETTER KA..MARCHEN LETTER A
    Prop<unicode::Script> { { 0x11C92, 0x11CA7 }, unicode::Script::Marchen }, // Mn  [22] MARCHEN SUBJOINED LETTER KA..MARCHEN SUBJOINED LETTER ZA
    Prop<unicode::Script> { { 0x11CA9, 0x11CA9 }, unicode::Script::Marchen }, // Mc       MARCHEN SUBJOINED LETTER YA
    Prop<unicode::Script> { { 0x11CAA, 0x11CB0 }, unicode::Script::Marchen }, // Mn   [7] MARCHEN SUBJOINED LETTER RA..MARCHEN VOWEL SIGN AA
    Prop<unicode::Script> { { 0x11CB1, 0x11CB1 }, unicode::Script::Marchen }, // Mc       MARCHEN VOWEL SIGN I
    Prop<unicode::Script> { { 0x11CB2, 0x11CB3 }, unicode::Script::Marchen }, // Mn   [2] MARCHEN VOWEL SIGN U..MARCHEN VOWEL SIGN E
    Prop<unicode::Script> { { 0x11CB4, 0x11CB4 }, unicode::Script::Marchen }, // Mc       MARCHEN VOWEL SIGN O
    Prop<unicode::Script> { { 0x11CB5, 0x11CB6 }, unicode::Script::Marchen }, // Mn   [2] MARCHEN SIGN ANUSVARA..MARCHEN SIGN CANDRABINDU
    Prop<unicode::Script> { { 0x11D00, 0x11D06 }, unicode::Script::Masaram_Gondi }, // Lo   [7] MASARAM GONDI LETTER A..MASARAM GONDI LETTER E
    Prop<unicode::Script> { { 0x11D08, 0x11D09 }, unicode::Script::Masaram_Gondi }, // Lo   [2] MASARAM GONDI LETTER AI..MASARAM GONDI LETTER O
    Prop<unicode::Script> { { 0x11D0B, 0x11D30 }, unicode::Script::Masaram_Gondi }, // Lo  [38] MASARAM GONDI LETTER AU..MASARAM GONDI LETTER TRA
    Prop<unicode::Script> { { 0x11D31, 0x11D36 }, unicode::Script::Masaram_Gondi }, // Mn   [6] MASARAM GONDI VOWEL SIGN AA..MASARAM GONDI VOWEL SIGN VOCALIC R
    Prop<unicode::Script> { { 0x11D3A, 0x11D3A }, unicode::Script::Masaram_Gondi }, // Mn       MASARAM GONDI VOWEL SIGN E
    Prop<unicode::Script> { { 0x11D3C, 0x11D3D }, unicode::Script::Masaram_Gondi }, // Mn   [2] MASARAM GONDI VOWEL SIGN AI..MASARAM GONDI VOWEL SIGN O
    Prop<unicode::Script> { { 0x11D3F, 0x11D45 }, unicode::Script::Masaram_Gondi }, // Mn   [7] MASARAM GONDI VOWEL SIGN AU..MASARAM GONDI VIRAMA
    Prop<unicode::Script> { { 0x11D46, 0x11D46 }, unicode::Script::Masaram_Gondi }, // Lo       MASARAM GONDI REPHA
    Prop<unicode::Script> { { 0x11D47, 0x11D47 }, unicode::Script::Masaram_Gondi }, // Mn       MASARAM GONDI RA-KARA
    Prop<unicode::Script> { { 0x11D50, 0x11D59 }, unicode::Script::Masaram_Gondi }, // Nd  [10] MASARAM GONDI DIGIT ZERO..MASARAM GONDI DIGIT NINE
    Prop<unicode::Script> { { 0x11D60, 0x11D65 }, unicode::Script::Gunjala_Gondi }, // Lo   [6] GUNJALA GONDI LETTER A..GUNJALA GONDI LETTER UU
    Prop<unicode::Script> { { 0x11D67, 0x11D68 }, unicode::Script::Gunjala_Gondi }, // Lo   [2] GUNJALA GONDI LETTER EE..GUNJALA GONDI LETTER AI
    Prop<unicode::Script> { { 0x11D6A, 0x11D89 }, unicode::Script::Gunjala_Gondi }, // Lo  [32] GUNJALA GONDI LETTER OO..GUNJALA GONDI LETTER SA
    Prop<unicode::Script> { { 0x11D8A, 0x11D8E }, unicode::Script::Gunjala_Gondi }, // Mc   [5] GUNJALA GONDI VOWEL SIGN AA..GUNJALA GONDI VOWEL SIGN UU
    Prop<unicode::Script> { { 0x11D90, 0x11D91 }, unicode::Script::Gunjala_Gondi }, // Mn   [2] GUNJALA GONDI VOWEL SIGN EE..GUNJALA GONDI VOWEL SIGN AI
    Prop<unicode::Script> { { 0x11D93, 0x11D94 }, unicode::Script::Gunjala_Gondi }, // Mc   [2] GUNJALA GONDI VOWEL SIGN OO..GUNJALA GONDI VOWEL SIGN AU
    Prop<unicode::Script> { { 0x11D95, 0x11D95 }, unicode::Script::Gunjala_Gondi }, // Mn       GUNJALA GONDI SIGN ANUSVARA
    Prop<unicode::Script> { { 0x11D96, 0x11D96 }, unicode::Script::Gunjala_Gondi }, // Mc       GUNJALA GONDI SIGN VISARGA
    Prop<unicode::Script> { { 0x11D97, 0x11D97 }, unicode::Script::Gunjala_Gondi }, // Mn       GUNJALA GONDI VIRAMA
    Prop<unicode::Script> { { 0x11D98, 0x11D98 }, unicode::Script::Gunjala_Gondi }, // Lo       GUNJALA GONDI OM
    Prop<unicode::Script> { { 0x11DA0, 0x11DA9 }, unicode::Script::Gunjala_Gondi }, // Nd  [10] GUNJALA GONDI DIGIT ZERO..GUNJALA GONDI DIGIT NINE
    Prop<unicode::Script> { { 0x11DB0, 0x11DD8 }, unicode::Script::Tolong_Siki }, // Lo  [41] TOLONG SIKI LETTER I..TOLONG SIKI LETTER RRH
    Prop<unicode::Script> { { 0x11DD9, 0x11DD9 }, unicode::Script::Tolong_Siki }, // Lm       TOLONG SIKI SIGN SELA
    Prop<unicode::Script> { { 0x11DDA, 0x11DDB }, unicode::Script::Tolong_Siki }, // Lo   [2] TOLONG SIKI SIGN HECAKA..TOLONG SIKI UNGGA
    Prop<unicode::Script> { { 0x11DE0, 0x11DE9 }, unicode::Script::Tolong_Siki }, // Nd  [10] TOLONG SIKI DIGIT ZERO..TOLONG SIKI DIGIT NINE
    Prop<unicode::Script> { { 0x11EE0, 0x11EF2 }, unicode::Script::Makasar }, // Lo  [19] MAKASAR LETTER KA..MAKASAR ANGKA
    Prop<unicode::Script> { { 0x11EF3, 0x11EF4 }, unicode::Script::Makasar }, // Mn   [2] MAKASAR VOWEL SIGN I..MAKASAR VOWEL SIGN U
    Prop<unicode::Script> { { 0x11EF5, 0x11EF6 }, unicode::Script::Makasar }, // Mc   [2] MAKASAR VOWEL SIGN E..MAKASAR VOWEL SIGN O
    Prop<unicode::Script> { { 0x11EF7, 0x11EF8 }, unicode::Script::Makasar }, // Po   [2] MAKASAR PASSIMBANG..MAKASAR END OF SECTION
    Prop<unicode::Script> { { 0x11F00, 0x11F01 }, unicode::Script::Kawi }, // Mn   [2] KAWI SIGN CANDRABINDU..KAWI SIGN ANUSVARA
    Prop<unicode::Script> { { 0x11F02, 0x11F02 }, unicode::Script::Kawi }, // Lo       KAWI SIGN REPHA
    Prop<unicode::Script> { { 0x11F03, 0x11F03 }, unicode::Script::Kawi }, // Mc       KAWI SIGN VISARGA
    Prop<unicode::Script> { { 0x11F04, 0x11F10 }, unicode::Script::Kawi }, // Lo  [13] KAWI LETTER A..KAWI LETTER O
    Prop<unicode::Script> { { 0x11F12, 0x11F33 }, unicode::Script::Kawi }, // Lo  [34] KAWI LETTER KA..KAWI LETTER JNYA
    Prop<unicode::Script> { { 0x11F34, 0x11F35 }, unicode::Script::Kawi }, // Mc   [2] KAWI VOWEL SIGN AA..KAWI VOWEL SIGN ALTERNATE AA
    Prop<unicode::Script> { { 0x11F36, 0x11F3A }, unicode::Script::Kawi }, // Mn   [5] KAWI VOWEL SIGN I..KAWI VOWEL SIGN VOCALIC R
    Prop<unicode::Script> { { 0x11F3E, 0x11F3F }, unicode::Script::Kawi }, // Mc   [2] KAWI VOWEL SIGN E..KAWI VOWEL SIGN AI
    Prop<unicode::Script> { { 0x11F40, 0x11F40 }, unicode::Script::Kawi }, // Mn       KAWI VOWEL SIGN EU
    Prop<unicode::Script> { { 0x11F41, 0x11F41 }, unicode::Script::Kawi }, // Mc       KAWI SIGN KILLER
    Prop<unicode::Script> { { 0x11F42, 0x11F42 }, unicode::Script::Kawi }, // Mn       KAWI CONJOINER
    Prop<unicode::Script> { { 0x11F43, 0x11F4F }, unicode::Script::Kawi }, // Po  [13] KAWI DANDA..KAWI PUNCTUATION CLOSING SPIRAL
    Prop<unicode::Script> { { 0x11F50, 0x11F59 }, unicode::Script::Kawi }, // Nd  [10] KAWI DIGIT ZERO..KAWI DIGIT NINE
    Prop<unicode::Script> { { 0x11F5A, 0x11F5A }, unicode::Script::Kawi }, // Mn       KAWI SIGN NUKTA
    Prop<unicode::Script> { { 0x11FB0, 0x11FB0 }, unicode::Script::Lisu }, // Lo       LISU LETTER YHA
    Prop<unicode::Script> { { 0x11FC0, 0x11FD4 }, unicode::Script::Tamil }, // No  [21] TAMIL FRACTION ONE THREE-HUNDRED-AND-TWENTIETH..TAMIL FRACTION DOWNSCALING FACTOR KIIZH
    Prop<unicode::Script> { { 0x11FD5, 0x11FDC }, unicode::Script::Tamil }, // So   [8] TAMIL SIGN NEL..TAMIL SIGN MUKKURUNI
    Prop<unicode::Script> { { 0x11FDD, 0x11FE0 }, unicode::Script::Tamil }, // Sc   [4] TAMIL SIGN KAACU..TAMIL SIGN VARAAKAN
    Prop<unicode::Script> { { 0x11FE1, 0x11FF1 }, unicode::Script::Tamil }, // So  [17] TAMIL SIGN PAARAM..TAMIL SIGN VAKAIYARAA
    Prop<unicode::Script> { { 0x11FFF, 0x11FFF }, unicode::Script::Tamil }, // Po       TAMIL PUNCTUATION END OF TEXT
    Prop<unicode::Script> { { 0x12000, 0x12399 }, unicode::Script::Cuneiform }, // Lo [922] CUNEIFORM SIGN A..CUNEIFORM SIGN U U
    Prop<unicode::Script> { { 0x12400, 0x1246E }, unicode::Script::Cuneiform }, // Nl [111] CUNEIFORM NUMERIC SIGN TWO ASH..CUNEIFORM NUMERIC SIGN NINE U VARIANT FORM
    Prop<unicode::Script> { { 0x12470, 0x12474 }, unicode::Script::Cuneiform }, // Po   [5] CUNEIFORM PUNCTUATION SIGN OLD ASSYRIAN WORD DIVIDER..CUNEIFORM PUNCTUATION SIGN DIAGONAL QUADCOLON
    Prop<unicode::Script> { { 0x12480, 0x12543 }, unicode::Script::Cuneiform }, // Lo [196] CUNEIFORM SIGN AB TIMES NUN TENU..CUNEIFORM SIGN ZU5 TIMES THREE DISH TENU
    Prop<unicode::Script> { { 0x12F90, 0x12FF0 }, unicode::Script::Cypro_Minoan }, // Lo  [97] CYPRO-MINOAN SIGN CM001..CYPRO-MINOAN SIGN CM114
    Prop<unicode::Script> { { 0x12FF1, 0x12FF2 }, unicode::Script::Cypro_Minoan }, // Po   [2] CYPRO-MINOAN SIGN CM301..CYPRO-MINOAN SIGN CM302
    Prop<unicode::Script> { { 0x13000, 0x1342F }, unicode::Script::Egyptian_Hieroglyphs }, // Lo [1072] EGYPTIAN HIEROGLYPH A001..EGYPTIAN HIEROGLYPH V011D
    Prop<unicode::Script> { { 0x13430, 0x1343F }, unicode::Script::Egyptian_Hieroglyphs }, // Cf  [16] EGYPTIAN HIEROGLYPH VERTICAL JOINER..EGYPTIAN HIEROGLYPH END WALLED ENCLOSURE
    Prop<unicode::Script> { { 0x13440, 0x13440 }, unicode::Script::Egyptian_Hieroglyphs }, // Mn       EGYPTIAN HIEROGLYPH MIRROR HORIZONTALLY
    Prop<unicode::Script> { { 0x13441, 0x13446 }, unicode::Script::Egyptian_Hieroglyphs }, // Lo   [6] EGYPTIAN HIEROGLYPH FULL BLANK..EGYPTIAN HIEROGLYPH WIDE LOST SIGN
    Prop<unicode::Script> { { 0x13447, 0x13455 }, unicode::Script::Egyptian_Hieroglyphs }, // Mn  [15] EGYPTIAN HIEROGLYPH MODIFIER DAMAGED AT TOP START..EGYPTIAN HIEROGLYPH MODIFIER DAMAGED
    Prop<unicode::Script> { { 0x13460, 0x143FA }, unicode::Script::Egyptian_Hieroglyphs }, // Lo [3995] EGYPTIAN HIEROGLYPH-13460..EGYPTIAN HIEROGLYPH-143FA
    Prop<unicode::Script> { { 0x14400, 0x14646 }, unicode::Script::Anatolian_Hieroglyphs }, // Lo [583] ANATOLIAN HIEROGLYPH A001..ANATOLIAN HIEROGLYPH A530
    Prop<unicode::Script> { { 0x16100, 0x1611D }, unicode::Script::Gurung_Khema }, // Lo  [30] GURUNG KHEMA LETTER A..GURUNG KHEMA LETTER SA
    Prop<unicode::Script> { { 0x1611E, 0x16129 }, unicode::Script::Gurung_Khema }, // Mn  [12] GURUNG KHEMA VOWEL SIGN AA..GURUNG KHEMA VOWEL LENGTH MARK
    Prop<unicode::Script> { { 0x1612A, 0x1612C }, unicode::Script::Gurung_Khema }, // Mc   [3] GURUNG KHEMA CONSONANT SIGN MEDIAL YA..GURUNG KHEMA CONSONANT SIGN MEDIAL HA
    Prop<unicode::Script> { { 0x1612D, 0x1612F }, unicode::Script::Gurung_Khema }, // Mn   [3] GURUNG KHEMA SIGN ANUSVARA..GURUNG KHEMA SIGN THOLHOMA
    Prop<unicode::Script> { { 0x16130, 0x16139 }, unicode::Script::Gurung_Khema }, // Nd  [10] GURUNG KHEMA DIGIT ZERO..GURUNG KHEMA DIGIT NINE
    Prop<unicode::Script> { { 0x16800, 0x16A38 }, unicode::Script::Bamum }, // Lo [569] BAMUM LETTER PHASE-A NGKUE MFON..BAMUM LETTER PHASE-F VUEQ
    Prop<unicode::Script> { { 0x16A40, 0x16A5E }, unicode::Script::Mro }, // Lo  [31] MRO LETTER TA..MRO LETTER TEK
    Prop<unicode::Script> { { 0x16A60, 0x16A69 }, unicode::Script::Mro }, // Nd  [10] MRO DIGIT ZERO..MRO DIGIT NINE
    Prop<unicode::Script> { { 0x16A6E, 0x16A6F }, unicode::Script::Mro }, // Po   [2] MRO DANDA..MRO DOUBLE DANDA
    Prop<unicode::Script> { { 0x16A70, 0x16ABE }, unicode::Script::Tangsa }, // Lo  [79] TANGSA LETTER OZ..TANGSA LETTER ZA
    Prop<unicode::Script> { { 0x16AC0, 0x16AC9 }, unicode::Script::Tangsa }, // Nd  [10] TANGSA DIGIT ZERO..TANGSA DIGIT NINE
    Prop<unicode::Script> { { 0x16AD0, 0x16AED }, unicode::Script::Bassa_Vah }, // Lo  [30] BASSA VAH LETTER ENNI..BASSA VAH LETTER I
    Prop<unicode::Script> { { 0x16AF0, 0x16AF4 }, unicode::Script::Bassa_Vah }, // Mn   [5] BASSA VAH COMBINING HIGH TONE..BASSA VAH COMBINING HIGH-LOW TONE
    Prop<unicode::Script> { { 0x16AF5, 0x16AF5 }, unicode::Script::Bassa_Vah }, // Po       BASSA VAH FULL STOP
    Prop<unicode::Script> { { 0x16B00, 0x16B2F }, unicode::Script::Pahawh_Hmong }, // Lo  [48] PAHAWH HMONG VOWEL KEEB..PAHAWH HMONG CONSONANT CAU
    Prop<unicode::Script> { { 0x16B30, 0x16B36 }, unicode::Script::Pahawh_Hmong }, // Mn   [7] PAHAWH HMONG MARK CIM TUB..PAHAWH HMONG MARK CIM TAUM
    Prop<unicode::Script> { { 0x16B37, 0x16B3B }, unicode::Script::Pahawh_Hmong }, // Po   [5] PAHAWH HMONG SIGN VOS THOM..PAHAWH HMONG SIGN VOS FEEM
    Prop<unicode::Script> { { 0x16B3C, 0x16B3F }, unicode::Script::Pahawh_Hmong }, // So   [4] PAHAWH HMONG SIGN XYEEM NTXIV..PAHAWH HMONG SIGN XYEEM FAIB
    Prop<unicode::Script> { { 0x16B40, 0x16B43 }, unicode::Script::Pahawh_Hmong }, // Lm   [4] PAHAWH HMONG SIGN VOS SEEV..PAHAWH HMONG SIGN IB YAM
    Prop<unicode::Script> { { 0x16B44, 0x16B44 }, unicode::Script::Pahawh_Hmong }, // Po       PAHAWH HMONG SIGN XAUS
    Prop<unicode::Script> { { 0x16B45, 0x16B45 }, unicode::Script::Pahawh_Hmong }, // So       PAHAWH HMONG SIGN CIM TSOV ROG
    Prop<unicode::Script> { { 0x16B50, 0x16B59 }, unicode::Script::Pahawh_Hmong }, // Nd  [10] PAHAWH HMONG DIGIT ZERO..PAHAWH HMONG DIGIT NINE
    Prop<unicode::Script> { { 0x16B5B, 0x16B61 }, unicode::Script::Pahawh_Hmong }, // No   [7] PAHAWH HMONG NUMBER TENS..PAHAWH HMONG NUMBER TRILLIONS
    Prop<unicode::Script> { { 0x16B63, 0x16B77 }, unicode::Script::Pahawh_Hmong }, // Lo  [21] PAHAWH HMONG SIGN VOS LUB..PAHAWH HMONG SIGN CIM NRES TOS
    Prop<unicode::Script> { { 0x16B7D, 0x16B8F }, unicode::Script::Pahawh_Hmong }, // Lo  [19] PAHAWH HMONG CLAN SIGN TSHEEJ..PAHAWH HMONG CLAN SIGN VWJ
    Prop<unicode::Script> { { 0x16D40, 0x16D42 }, unicode::Script::Kirat_Rai }, // Lm   [3] KIRAT RAI SIGN ANUSVARA..KIRAT RAI SIGN VISARGA
    Prop<unicode::Script> { { 0x16D43, 0x16D6A }, unicode::Script::Kirat_Rai }, // Lo  [40] KIRAT RAI LETTER A..KIRAT RAI VOWEL SIGN AU
    Prop<unicode::Script> { { 0x16D6B, 0x16D6C }, unicode::Script::Kirat_Rai }, // Lm   [2] KIRAT RAI SIGN VIRAMA..KIRAT RAI SIGN SAAT
    Prop<unicode::Script> { { 0x16D6D, 0x16D6F }, unicode::Script::Kirat_Rai }, // Po   [3] KIRAT RAI SIGN YUPI..KIRAT RAI DOUBLE DANDA
    Prop<unicode::Script> { { 0x16D70, 0x16D79 }, unicode::Script::Kirat_Rai }, // Nd  [10] KIRAT RAI DIGIT ZERO..KIRAT RAI DIGIT NINE
    Prop<unicode::Script> { { 0x16E40, 0x16E7F }, unicode::Script::Medefaidrin }, // L&  [64] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    Prop<unicode::Script> { { 0x16E80, 0x16E96 }, unicode::Script::Medefaidrin }, // No  [23] MEDEFAIDRIN DIGIT ZERO..MEDEFAIDRIN DIGIT THREE ALTERNATE FORM
    Prop<unicode::Script> { { 0x16E97, 0x16E9A }, unicode::Script::Medefaidrin }, // Po   [4] MEDEFAIDRIN COMMA..MEDEFAIDRIN EXCLAMATION OH
    Prop<unicode::Script> { { 0x16EA0, 0x16EB8 }, unicode::Script::Beria_Erfe }, // L&  [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Prop<unicode::Script> { { 0x16EBB, 0x16ED3 }, unicode::Script::Beria_Erfe }, // L&  [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Prop<unicode::Script> { { 0x16F00, 0x16F4A }, unicode::Script::Miao }, // Lo  [75] MIAO LETTER PA..MIAO LETTER RTE
    Prop<unicode::Script> { { 0x16F4F, 0x16F4F }, unicode::Script::Miao }, // Mn       MIAO SIGN CONSONANT MODIFIER BAR
    Prop<unicode::Script> { { 0x16F50, 0x16F50 }, unicode::Script::Miao }, // Lo       MIAO LETTER NASALIZATION
    Prop<unicode::Script> { { 0x16F51, 0x16F87 }, unicode::Script::Miao }, // Mc  [55] MIAO SIGN ASPIRATION..MIAO VOWEL SIGN UI
    Prop<unicode::Script> { { 0x16F8F, 0x16F92 }, unicode::Script::Miao }, // Mn   [4] MIAO TONE RIGHT..MIAO TONE BELOW
    Prop<unicode::Script> { { 0x16F93, 0x16F9F }, unicode::Script::Miao }, // Lm  [13] MIAO LETTER TONE-2..MIAO LETTER REFORMED TONE-8
    Prop<unicode::Script> { { 0x16FE0, 0x16FE0 }, unicode::Script::Tangut }, // Lm       TANGUT ITERATION MARK
    Prop<unicode::Script> { { 0x16FE1, 0x16FE1 }, unicode::Script::Nushu }, // Lm       NUSHU ITERATION MARK
    Prop<unicode::Script> { { 0x16FE2, 0x16FE2 }, unicode::Script::Han }, // Po       OLD CHINESE HOOK MARK
    Prop<unicode::Script> { { 0x16FE3, 0x16FE3 }, unicode::Script::Han }, // Lm       OLD CHINESE ITERATION MARK
    Prop<unicode::Script> { { 0x16FE4, 0x16FE4 }, unicode::Script::Khitan_Small_Script }, // Mn       KHITAN SMALL SCRIPT FILLER
    Prop<unicode::Script> { { 0x16FF0, 0x16FF1 }, unicode::Script::Han }, // Mc   [2] VIETNAMESE ALTERNATE READING MARK CA..VIETNAMESE ALTERNATE READING MARK NHAY
    Prop<unicode::Script> { { 0x16FF2, 0x16FF3 }, unicode::Script::Han }, // Lm   [2] CHINESE SMALL SIMPLIFIED ER..CHINESE SMALL TRADITIONAL ER
    Prop<unicode::Script> { { 0x16FF4, 0x16FF6 }, unicode::Script::Han }, // Nl   [3] YANGQIN SIGN SLOW ONE BEAT..YANGQIN SIGN SLOW TWO BEATS
    Prop<unicode::Script> { { 0x17000, 0x18AFF }, unicode::Script::Tangut }, // Lo [6912] TANGUT IDEOGRAPH-17000..TANGUT COMPONENT-768
    Prop<unicode::Script> { { 0x18B00, 0x18CD5 }, unicode::Script::Khitan_Small_Script }, // Lo [470] KHITAN SMALL SCRIPT CHARACTER-18B00..KHITAN SMALL SCRIPT CHARACTER-18CD5
    Prop<unicode::Script> { { 0x18CFF, 0x18CFF }, unicode::Script::Khitan_Small_Script }, // Lo       KHITAN SMALL SCRIPT CHARACTER-18CFF
    Prop<unicode::Script> { { 0x18D00, 0x18D1E }, unicode::Script::Tangut }, // Lo  [31] TANGUT IDEOGRAPH-18D00..TANGUT IDEOGRAPH-18D1E
    Prop<unicode::Script> { { 0x18D80, 0x18DF2 }, unicode::Script::Tangut }, // Lo [115] TANGUT COMPONENT-769..TANGUT COMPONENT-883
    Prop<unicode::Script> { { 0x1AFF0, 0x1AFF3 }, unicode::Script::Katakana }, // Lm   [4] KATAKANA LETTER MINNAN TONE-2..KATAKANA LETTER MINNAN TONE-5
    Prop<unicode::Script> { { 0x1AFF5, 0x1AFFB }, unicode::Script::Katakana }, // Lm   [7] KATAKANA LETTER MINNAN TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-5
    Prop<unicode::Script> { { 0x1AFFD, 0x1AFFE }, unicode::Script::Katakana }, // Lm   [2] KATAKANA LETTER MINNAN NASALIZED TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-8
    Prop<unicode::Script> { { 0x1B000, 0x1B000 }, unicode::Script::Katakana }, // Lo       KATAKANA LETTER ARCHAIC E
    Prop<unicode::Script> { { 0x1B001, 0x1B11F }, unicode::Script::Hiragana }, // Lo [287] HIRAGANA LETTER ARCHAIC YE..HIRAGANA LETTER ARCHAIC WU
    Prop<unicode::Script> { { 0x1B120, 0x1B122 }, unicode::Script::Katakana }, // Lo   [3] KATAKANA LETTER ARCHAIC YI..KATAKANA LETTER ARCHAIC WU
    Prop<unicode::Script> { { 0x1B132, 0x1B132 }, unicode::Script::Hiragana }, // Lo       HIRAGANA LETTER SMALL KO
    Prop<unicode::Script> { { 0x1B150, 0x1B152 }, unicode::Script::Hiragana }, // Lo   [3] HIRAGANA LETTER SMALL WI..HIRAGANA LETTER SMALL WO
    Prop<unicode::Script> { { 0x1B155, 0x1B155 }, unicode::Script::Katakana }, // Lo       KATAKANA LETTER SMALL KO
    Prop<unicode::Script> { { 0x1B164, 0x1B167 }, unicode::Script::Katakana }, // Lo   [4] KATAKANA LETTER SMALL WI..KATAKANA LETTER SMALL N
    Prop<unicode::Script> { { 0x1B170, 0x1B2FB }, unicode::Script::Nushu }, // Lo [396] NUSHU CHARACTER-1B170..NUSHU CHARACTER-1B2FB
    Prop<unicode::Script> { { 0x1BC00, 0x1BC6A }, unicode::Script::Duployan }, // Lo [107] DUPLOYAN LETTER H..DUPLOYAN LETTER VOCALIC M
    Prop<unicode::Script> { { 0x1BC70, 0x1BC7C }, unicode::Script::Duployan }, // Lo  [13] DUPLOYAN AFFIX LEFT HORIZONTAL SECANT..DUPLOYAN AFFIX ATTACHED TANGENT HOOK
    Prop<unicode::Script> { { 0x1BC80, 0x1BC88 }, unicode::Script::Duployan }, // Lo   [9] DUPLOYAN AFFIX HIGH ACUTE..DUPLOYAN AFFIX HIGH VERTICAL
    Prop<unicode::Script> { { 0x1BC90, 0x1BC99 }, unicode::Script::Duployan }, // Lo  [10] DUPLOYAN AFFIX LOW ACUTE..DUPLOYAN AFFIX LOW ARROW
    Prop<unicode::Script> { { 0x1BC9C, 0x1BC9C }, unicode::Script::Duployan }, // So       DUPLOYAN SIGN O WITH CROSS
    Prop<unicode::Script> { { 0x1BC9D, 0x1BC9E }, unicode::Script::Duployan }, // Mn   [2] DUPLOYAN THICK LETTER SELECTOR..DUPLOYAN DOUBLE MARK
    Prop<unicode::Script> { { 0x1BC9F, 0x1BC9F }, unicode::Script::Duployan }, // Po       DUPLOYAN PUNCTUATION CHINOOK FULL STOP
    Prop<unicode::Script> { { 0x1BCA0, 0x1BCA3 }, unicode::Script::Common }, // Cf   [4] SHORTHAND FORMAT LETTER OVERLAP..SHORTHAND FORMAT UP STEP
    Prop<unicode::Script> { { 0x1CC00, 0x1CCEF }, unicode::Script::Common }, // So [240] UP-POINTING GO-KART..OUTLINED LATIN CAPITAL LETTER Z
    Prop<unicode::Script> { { 0x1CCF0, 0x1CCF9 }, unicode::Script::Common }, // Nd  [10] OUTLINED DIGIT ZERO..OUTLINED DIGIT NINE
    Prop<unicode::Script> { { 0x1CCFA, 0x1CCFC }, unicode::Script::Common }, // So   [3] SNAKE SYMBOL..NOSE SYMBOL
    Prop<unicode::Script> { { 0x1CD00, 0x1CEB3 }, unicode::Script::Common }, // So [436] BLOCK OCTANT-3..BLACK RIGHT TRIANGLE CARET
    Prop<unicode::Script> { { 0x1CEBA, 0x1CED0 }, unicode::Script::Common }, // So  [23] FRAGILE SYMBOL..LEUKOTHEA
    Prop<unicode::Script> { { 0x1CEE0, 0x1CEEF }, unicode::Script::Common }, // So  [16] GEOMANTIC FIGURE POPULUS..GEOMANTIC FIGURE VIA
    Prop<unicode::Script> { { 0x1CEF0, 0x1CEF0 }, unicode::Script::Common }, // Sm       MEDIUM SMALL WHITE CIRCLE WITH HORIZONTAL BAR
    Prop<unicode::Script> { { 0x1CF00, 0x1CF2D }, unicode::Script::Inherited }, // Mn  [46] ZNAMENNY COMBINING MARK GORAZDO NIZKO S KRYZHEM ON LEFT..ZNAMENNY COMBINING MARK KRYZH ON LEFT
    Prop<unicode::Script> { { 0x1CF30, 0x1CF46 }, unicode::Script::Inherited }, // Mn  [23] ZNAMENNY COMBINING TONAL RANGE MARK MRACHNO..ZNAMENNY PRIZNAK MODIFIER ROG
    Prop<unicode::Script> { { 0x1CF50, 0x1CFC3 }, unicode::Script::Common }, // So [116] ZNAMENNY NEUME KRYUK..ZNAMENNY NEUME PAUK
    Prop<unicode::Script> { { 0x1D000, 0x1D0F5 }, unicode::Script::Common }, // So [246] BYZANTINE MUSICAL SYMBOL PSILI..BYZANTINE MUSICAL SYMBOL GORGON NEO KATO
    Prop<unicode::Script> { { 0x1D100, 0x1D126 }, unicode::Script::Common }, // So  [39] MUSICAL SYMBOL SINGLE BARLINE..MUSICAL SYMBOL DRUM CLEF-2
    Prop<unicode::Script> { { 0x1D129, 0x1D164 }, unicode::Script::Common }, // So  [60] MUSICAL SYMBOL MULTIPLE MEASURE REST..MUSICAL SYMBOL ONE HUNDRED TWENTY-EIGHTH NOTE
    Prop<unicode::Script> { { 0x1D165, 0x1D166 }, unicode::Script::Common }, // Mc   [2] MUSICAL SYMBOL COMBINING STEM..MUSICAL SYMBOL COMBINING SPRECHGESANG STEM
    Prop<unicode::Script> { { 0x1D167, 0x1D169 }, unicode::Script::Inherited }, // Mn   [3] MUSICAL SYMBOL COMBINING TREMOLO-1..MUSICAL SYMBOL COMBINING TREMOLO-3
    Prop<unicode::Script> { { 0x1D16A, 0x1D16C }, unicode::Script::Common }, // So   [3] MUSICAL SYMBOL FINGERED TREMOLO-1..MUSICAL SYMBOL FINGERED TREMOLO-3
    Prop<unicode::Script> { { 0x1D16D, 0x1D172 }, unicode::Script::Common }, // Mc   [6] MUSICAL SYMBOL COMBINING AUGMENTATION DOT..MUSICAL SYMBOL COMBINING FLAG-5
    Prop<unicode::Script> { { 0x1D173, 0x1D17A }, unicode::Script::Common }, // Cf   [8] MUSICAL SYMBOL BEGIN BEAM..MUSICAL SYMBOL END PHRASE
    Prop<unicode::Script> { { 0x1D17B, 0x1D182 }, unicode::Script::Inherited }, // Mn   [8] MUSICAL SYMBOL COMBINING ACCENT..MUSICAL SYMBOL COMBINING LOURE
    Prop<unicode::Script> { { 0x1D183, 0x1D184 }, unicode::Script::Common }, // So   [2] MUSICAL SYMBOL ARPEGGIATO UP..MUSICAL SYMBOL ARPEGGIATO DOWN
    Prop<unicode::Script> { { 0x1D185, 0x1D18B }, unicode::Script::Inherited }, // Mn   [7] MUSICAL SYMBOL COMBINING DOIT..MUSICAL SYMBOL COMBINING TRIPLE TONGUE
    Prop<unicode::Script> { { 0x1D18C, 0x1D1A9 }, unicode::Script::Common }, // So  [30] MUSICAL SYMBOL RINFORZANDO..MUSICAL SYMBOL DEGREE SLASH
    Prop<unicode::Script> { { 0x1D1AA, 0x1D1AD }, unicode::Script::Inherited }, // Mn   [4] MUSICAL SYMBOL COMBINING DOWN BOW..MUSICAL SYMBOL COMBINING SNAP PIZZICATO
    Prop<unicode::Script> { { 0x1D1AE, 0x1D1EA }, unicode::Script::Common }, // So  [61] MUSICAL SYMBOL PEDAL MARK..MUSICAL SYMBOL KORON
    Prop<unicode::Script> { { 0x1D200, 0x1D241 }, unicode::Script::Greek }, // So  [66] GREEK VOCAL NOTATION SYMBOL-1..GREEK INSTRUMENTAL NOTATION SYMBOL-54
    Prop<unicode::Script> { { 0x1D242, 0x1D244 }, unicode::Script::Greek }, // Mn   [3] COMBINING GREEK MUSICAL TRISEME..COMBINING GREEK MUSICAL PENTASEME
    Prop<unicode::Script> { { 0x1D245, 0x1D245 }, unicode::Script::Greek }, // So       GREEK MUSICAL LEIMMA
    Prop<unicode::Script> { { 0x1D2C0, 0x1D2D3 }, unicode::Script::Common }, // No  [20] KAKTOVIK NUMERAL ZERO..KAKTOVIK NUMERAL NINETEEN
    Prop<unicode::Script> { { 0x1D2E0, 0x1D2F3 }, unicode::Script::Common }, // No  [20] MAYAN NUMERAL ZERO..MAYAN NUMERAL NINETEEN
    Prop<unicode::Script> { { 0x1D300, 0x1D356 }, unicode::Script::Common }, // So  [87] MONOGRAM FOR EARTH..TETRAGRAM FOR FOSTERING
    Prop<unicode::Script> { { 0x1D360, 0x1D378 }, unicode::Script::Common }, // No  [25] COUNTING ROD UNIT DIGIT ONE..TALLY MARK FIVE
    Prop<unicode::Script> { { 0x1D400, 0x1D454 }, unicode::Script::Common }, // L&  [85] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL ITALIC SMALL G
    Prop<unicode::Script> { { 0x1D456, 0x1D49C }, unicode::Script::Common }, // L&  [71] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL SCRIPT CAPITAL A
    Prop<unicode::Script> { { 0x1D49E, 0x1D49F }, unicode::Script::Common }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Prop<unicode::Script> { { 0x1D4A2, 0x1D4A2 }, unicode::Script::Common }, // L&       MATHEMATICAL SCRIPT CAPITAL G
    Prop<unicode::Script> { { 0x1D4A5, 0x1D4A6 }, unicode::Script::Common }, // L&   [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Prop<unicode::Script> { { 0x1D4A9, 0x1D4AC }, unicode::Script::Common }, // L&   [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Prop<unicode::Script> { { 0x1D4AE, 0x1D4B9 }, unicode::Script::Common }, // L&  [12] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT SMALL D
    Prop<unicode::Script> { { 0x1D4BB, 0x1D4BB }, unicode::Script::Common }, // L&       MATHEMATICAL SCRIPT SMALL F
    Prop<unicode::Script> { { 0x1D4BD, 0x1D4C3 }, unicode::Script::Common }, // L&   [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Prop<unicode::Script> { { 0x1D4C5, 0x1D505 }, unicode::Script::Common }, // L&  [65] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL FRAKTUR CAPITAL B
    Prop<unicode::Script> { { 0x1D507, 0x1D50A }, unicode::Script::Common }, // L&   [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Prop<unicode::Script> { { 0x1D50D, 0x1D514 }, unicode::Script::Common }, // L&   [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Prop<unicode::Script> { { 0x1D516, 0x1D51C }, unicode::Script::Common }, // L&   [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Prop<unicode::Script> { { 0x1D51E, 0x1D539 }, unicode::Script::Common }, // L&  [28] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Prop<unicode::Script> { { 0x1D53B, 0x1D53E }, unicode::Script::Common }, // L&   [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Prop<unicode::Script> { { 0x1D540, 0x1D544 }, unicode::Script::Common }, // L&   [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Prop<unicode::Script> { { 0x1D546, 0x1D546 }, unicode::Script::Common }, // L&       MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Prop<unicode::Script> { { 0x1D54A, 0x1D550 }, unicode::Script::Common }, // L&   [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Prop<unicode::Script> { { 0x1D552, 0x1D6A5 }, unicode::Script::Common }, // L& [340] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    Prop<unicode::Script> { { 0x1D6A8, 0x1D6C0 }, unicode::Script::Common }, // L&  [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    Prop<unicode::Script> { { 0x1D6C1, 0x1D6C1 }, unicode::Script::Common }, // Sm       MATHEMATICAL BOLD NABLA
    Prop<unicode::Script> { { 0x1D6C2, 0x1D6DA }, unicode::Script::Common }, // L&  [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    Prop<unicode::Script> { { 0x1D6DB, 0x1D6DB }, unicode::Script::Common }, // Sm       MATHEMATICAL BOLD PARTIAL DIFFERENTIAL
    Prop<unicode::Script> { { 0x1D6DC, 0x1D6FA }, unicode::Script::Common }, // L&  [31] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL ITALIC CAPITAL OMEGA
    Prop<unicode::Script> { { 0x1D6FB, 0x1D6FB }, unicode::Script::Common }, // Sm       MATHEMATICAL ITALIC NABLA
    Prop<unicode::Script> { { 0x1D6FC, 0x1D714 }, unicode::Script::Common }, // L&  [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    Prop<unicode::Script> { { 0x1D715, 0x1D715 }, unicode::Script::Common }, // Sm       MATHEMATICAL ITALIC PARTIAL DIFFERENTIAL
    Prop<unicode::Script> { { 0x1D716, 0x1D734 }, unicode::Script::Common }, // L&  [31] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    Prop<unicode::Script> { { 0x1D735, 0x1D735 }, unicode::Script::Common }, // Sm       MATHEMATICAL BOLD ITALIC NABLA
    Prop<unicode::Script> { { 0x1D736, 0x1D74E }, unicode::Script::Common }, // L&  [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    Prop<unicode::Script> { { 0x1D74F, 0x1D74F }, unicode::Script::Common }, // Sm       MATHEMATICAL BOLD ITALIC PARTIAL DIFFERENTIAL
    Prop<unicode::Script> { { 0x1D750, 0x1D76E }, unicode::Script::Common }, // L&  [31] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    Prop<unicode::Script> { { 0x1D76F, 0x1D76F }, unicode::Script::Common }, // Sm       MATHEMATICAL SANS-SERIF BOLD NABLA
    Prop<unicode::Script> { { 0x1D770, 0x1D788 }, unicode::Script::Common }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    Prop<unicode::Script> { { 0x1D789, 0x1D789 }, unicode::Script::Common }, // Sm       MATHEMATICAL SANS-SERIF BOLD PARTIAL DIFFERENTIAL
    Prop<unicode::Script> { { 0x1D78A, 0x1D7A8 }, unicode::Script::Common }, // L&  [31] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    Prop<unicode::Script> { { 0x1D7A9, 0x1D7A9 }, unicode::Script::Common }, // Sm       MATHEMATICAL SANS-SERIF BOLD ITALIC NABLA
    Prop<unicode::Script> { { 0x1D7AA, 0x1D7C2 }, unicode::Script::Common }, // L&  [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    Prop<unicode::Script> { { 0x1D7C3, 0x1D7C3 }, unicode::Script::Common }, // Sm       MATHEMATICAL SANS-SERIF BOLD ITALIC PARTIAL DIFFERENTIAL
    Prop<unicode::Script> { { 0x1D7C4, 0x1D7CB }, unicode::Script::Common }, // L&   [8] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD SMALL DIGAMMA
    Prop<unicode::Script> { { 0x1D7CE, 0x1D7FF }, unicode::Script::Common }, // Nd  [50] MATHEMATICAL BOLD DIGIT ZERO..MATHEMATICAL MONOSPACE DIGIT NINE
    Prop<unicode::Script> { { 0x1D800, 0x1D9FF }, unicode::Script::SignWriting }, // So [512] SIGNWRITING HAND-FIST INDEX..SIGNWRITING HEAD
    Prop<unicode::Script> { { 0x1DA00, 0x1DA36 }, unicode::Script::SignWriting }, // Mn  [55] SIGNWRITING HEAD RIM..SIGNWRITING AIR SUCKING IN
    Prop<unicode::Script> { { 0x1DA37, 0x1DA3A }, unicode::Script::SignWriting }, // So   [4] SIGNWRITING AIR BLOW SMALL ROTATIONS..SIGNWRITING BREATH EXHALE
    Prop<unicode::Script> { { 0x1DA3B, 0x1DA6C }, unicode::Script::SignWriting }, // Mn  [50] SIGNWRITING MOUTH CLOSED NEUTRAL..SIGNWRITING EXCITEMENT
    Prop<unicode::Script> { { 0x1DA6D, 0x1DA74 }, unicode::Script::SignWriting }, // So   [8] SIGNWRITING SHOULDER HIP SPINE..SIGNWRITING TORSO-FLOORPLANE TWISTING
    Prop<unicode::Script> { { 0x1DA75, 0x1DA75 }, unicode::Script::SignWriting }, // Mn       SIGNWRITING UPPER BODY TILTING FROM HIP JOINTS
    Prop<unicode::Script> { { 0x1DA76, 0x1DA83 }, unicode::Script::SignWriting }, // So  [14] SIGNWRITING LIMB COMBINATION..SIGNWRITING LOCATION DEPTH
    Prop<unicode::Script> { { 0x1DA84, 0x1DA84 }, unicode::Script::SignWriting }, // Mn       SIGNWRITING LOCATION HEAD NECK
    Prop<unicode::Script> { { 0x1DA85, 0x1DA86 }, unicode::Script::SignWriting }, // So   [2] SIGNWRITING LOCATION TORSO..SIGNWRITING LOCATION LIMBS DIGITS
    Prop<unicode::Script> { { 0x1DA87, 0x1DA8B }, unicode::Script::SignWriting }, // Po   [5] SIGNWRITING COMMA..SIGNWRITING PARENTHESIS
    Prop<unicode::Script> { { 0x1DA9B, 0x1DA9F }, unicode::Script::SignWriting }, // Mn   [5] SIGNWRITING FILL MODIFIER-2..SIGNWRITING FILL MODIFIER-6
    Prop<unicode::Script> { { 0x1DAA1, 0x1DAAF }, unicode::Script::SignWriting }, // Mn  [15] SIGNWRITING ROTATION MODIFIER-2..SIGNWRITING ROTATION MODIFIER-16
    Prop<unicode::Script> { { 0x1DF00, 0x1DF09 }, unicode::Script::Latin }, // L&  [10] LATIN SMALL LETTER FENG DIGRAPH WITH TRILL..LATIN SMALL LETTER T WITH HOOK AND RETROFLEX HOOK
    Prop<unicode::Script> { { 0x1DF0A, 0x1DF0A }, unicode::Script::Latin }, // Lo       LATIN LETTER RETROFLEX CLICK WITH RETROFLEX HOOK
    Prop<unicode::Script> { { 0x1DF0B, 0x1DF1E }, unicode::Script::Latin }, // L&  [20] LATIN SMALL LETTER ESH WITH DOUBLE BAR..LATIN SMALL LETTER S WITH CURL
    Prop<unicode::Script> { { 0x1DF25, 0x1DF2A }, unicode::Script::Latin }, // L&   [6] LATIN SMALL LETTER D WITH MID-HEIGHT LEFT HOOK..LATIN SMALL LETTER T WITH MID-HEIGHT LEFT HOOK
    Prop<unicode::Script> { { 0x1E000, 0x1E006 }, unicode::Script::Glagolitic }, // Mn   [7] COMBINING GLAGOLITIC LETTER AZU..COMBINING GLAGOLITIC LETTER ZHIVETE
    Prop<unicode::Script> { { 0x1E008, 0x1E018 }, unicode::Script::Glagolitic }, // Mn  [17] COMBINING GLAGOLITIC LETTER ZEMLJA..COMBINING GLAGOLITIC LETTER HERU
    Prop<unicode::Script> { { 0x1E01B, 0x1E021 }, unicode::Script::Glagolitic }, // Mn   [7] COMBINING GLAGOLITIC LETTER SHTA..COMBINING GLAGOLITIC LETTER YATI
    Prop<unicode::Script> { { 0x1E023, 0x1E024 }, unicode::Script::Glagolitic }, // Mn   [2] COMBINING GLAGOLITIC LETTER YU..COMBINING GLAGOLITIC LETTER SMALL YUS
    Prop<unicode::Script> { { 0x1E026, 0x1E02A }, unicode::Script::Glagolitic }, // Mn   [5] COMBINING GLAGOLITIC LETTER YO..COMBINING GLAGOLITIC LETTER FITA
    Prop<unicode::Script> { { 0x1E030, 0x1E06D }, unicode::Script::Cyrillic }, // Lm  [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Prop<unicode::Script> { { 0x1E08F, 0x1E08F }, unicode::Script::Cyrillic }, // Mn       COMBINING CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
    Prop<unicode::Script> { { 0x1E100, 0x1E12C }, unicode::Script::Nyiakeng_Puachue_Hmong }, // Lo  [45] NYIAKENG PUACHUE HMONG LETTER MA..NYIAKENG PUACHUE HMONG LETTER W
    Prop<unicode::Script> { { 0x1E130, 0x1E136 }, unicode::Script::Nyiakeng_Puachue_Hmong }, // Mn   [7] NYIAKENG PUACHUE HMONG TONE-B..NYIAKENG PUACHUE HMONG TONE-D
    Prop<unicode::Script> { { 0x1E137, 0x1E13D }, unicode::Script::Nyiakeng_Puachue_Hmong }, // Lm   [7] NYIAKENG PUACHUE HMONG SIGN FOR PERSON..NYIAKENG PUACHUE HMONG SYLLABLE LENGTHENER
    Prop<unicode::Script> { { 0x1E140, 0x1E149 }, unicode::Script::Nyiakeng_Puachue_Hmong }, // Nd  [10] NYIAKENG PUACHUE HMONG DIGIT ZERO..NYIAKENG PUACHUE HMONG DIGIT NINE
    Prop<unicode::Script> { { 0x1E14E, 0x1E14E }, unicode::Script::Nyiakeng_Puachue_Hmong }, // Lo       NYIAKENG PUACHUE HMONG LOGOGRAM NYAJ
    Prop<unicode::Script> { { 0x1E14F, 0x1E14F }, unicode::Script::Nyiakeng_Puachue_Hmong }, // So       NYIAKENG PUACHUE HMONG CIRCLED CA
    Prop<unicode::Script> { { 0x1E290, 0x1E2AD }, unicode::Script::Toto }, // Lo  [30] TOTO LETTER PA..TOTO LETTER A
    Prop<unicode::Script> { { 0x1E2AE, 0x1E2AE }, unicode::Script::Toto }, // Mn       TOTO SIGN RISING TONE
    Prop<unicode::Script> { { 0x1E2C0, 0x1E2EB }, unicode::Script::Wancho }, // Lo  [44] WANCHO LETTER AA..WANCHO LETTER YIH
    Prop<unicode::Script> { { 0x1E2EC, 0x1E2EF }, unicode::Script::Wancho }, // Mn   [4] WANCHO TONE TUP..WANCHO TONE KOINI
    Prop<unicode::Script> { { 0x1E2F0, 0x1E2F9 }, unicode::Script::Wancho }, // Nd  [10] WANCHO DIGIT ZERO..WANCHO DIGIT NINE
    Prop<unicode::Script> { { 0x1E2FF, 0x1E2FF }, unicode::Script::Wancho }, // Sc       WANCHO NGUN SIGN
    Prop<unicode::Script> { { 0x1E4D0, 0x1E4EA }, unicode::Script::Nag_Mundari }, // Lo  [27] NAG MUNDARI LETTER O..NAG MUNDARI LETTER ELL
    Prop<unicode::Script> { { 0x1E4EB, 0x1E4EB }, unicode::Script::Nag_Mundari }, // Lm       NAG MUNDARI SIGN OJOD
    Prop<unicode::Script> { { 0x1E4EC, 0x1E4EF }, unicode::Script::Nag_Mundari }, // Mn   [4] NAG MUNDARI SIGN MUHOR..NAG MUNDARI SIGN SUTUH
    Prop<unicode::Script> { { 0x1E4F0, 0x1E4F9 }, unicode::Script::Nag_Mundari }, // Nd  [10] NAG MUNDARI DIGIT ZERO..NAG MUNDARI DIGIT NINE
    Prop<unicode::Script> { { 0x1E5D0, 0x1E5ED }, unicode::Script::Ol_Onal }, // Lo  [30] OL ONAL LETTER O..OL ONAL LETTER EG
    Prop<unicode::Script> { { 0x1E5EE, 0x1E5EF }, unicode::Script::Ol_Onal }, // Mn   [2] OL ONAL SIGN MU..OL ONAL SIGN IKIR
    Prop<unicode::Script> { { 0x1E5F0, 0x1E5F0 }, unicode::Script::Ol_Onal }, // Lo       OL ONAL SIGN HODDOND
    Prop<unicode::Script> { { 0x1E5F1, 0x1E5FA }, unicode::Script::Ol_Onal }, // Nd  [10] OL ONAL DIGIT ZERO..OL ONAL DIGIT NINE
    Prop<unicode::Script> { { 0x1E5FF, 0x1E5FF }, unicode::Script::Ol_Onal }, // Po       OL ONAL ABBREVIATION SIGN
    Prop<unicode::Script> { { 0x1E6C0, 0x1E6DE }, unicode::Script::Tai_Yo }, // Lo  [31] TAI YO LETTER LOW KO..TAI YO LETTER HIGH KVO
    Prop<unicode::Script> { { 0x1E6E0, 0x1E6E2 }, unicode::Script::Tai_Yo }, // Lo   [3] TAI YO LETTER AA..TAI YO LETTER UE
    Prop<unicode::Script> { { 0x1E6E3, 0x1E6E3 }, unicode::Script::Tai_Yo }, // Mn       TAI YO SIGN UE
    Prop<unicode::Script> { { 0x1E6E4, 0x1E6E5 }, unicode::Script::Tai_Yo }, // Lo   [2] TAI YO LETTER U..TAI YO LETTER AE
    Prop<unicode::Script> { { 0x1E6E6, 0x1E6E6 }, unicode::Script::Tai_Yo }, // Mn       TAI YO SIGN AU
    Prop<unicode::Script> { { 0x1E6E7, 0x1E6ED }, unicode::Script::Tai_Yo }, // Lo   [7] TAI YO LETTER O..TAI YO LETTER AUE
    Prop<unicode::Script> { { 0x1E6EE, 0x1E6EF }, unicode::Script::Tai_Yo }, // Mn   [2] TAI YO SIGN AY..TAI YO SIGN ANG
    Prop<unicode::Script> { { 0x1E6F0, 0x1E6F4 }, unicode::Script::Tai_Yo }, // Lo   [5] TAI YO LETTER AN..TAI YO LETTER AP
    Prop<unicode::Script> { { 0x1E6F5, 0x1E6F5 }, unicode::Script::Tai_Yo }, // Mn       TAI YO SIGN OM
    Prop<unicode::Script> { { 0x1E6FE, 0x1E6FE }, unicode::Script::Tai_Yo }, // Lo       TAI YO SYMBOL MUEANG
    Prop<unicode::Script> { { 0x1E6FF, 0x1E6FF }, unicode::Script::Tai_Yo }, // Lm       TAI YO XAM LAI
    Prop<unicode::Script> { { 0x1E7E0, 0x1E7E6 }, unicode::Script::Ethiopic }, // Lo   [7] ETHIOPIC SYLLABLE HHYA..ETHIOPIC SYLLABLE HHYO
    Prop<unicode::Script> { { 0x1E7E8, 0x1E7EB }, unicode::Script::Ethiopic }, // Lo   [4] ETHIOPIC SYLLABLE GURAGE HHWA..ETHIOPIC SYLLABLE HHWE
    Prop<unicode::Script> { { 0x1E7ED, 0x1E7EE }, unicode::Script::Ethiopic }, // Lo   [2] ETHIOPIC SYLLABLE GURAGE MWI..ETHIOPIC SYLLABLE GURAGE MWEE
    Prop<unicode::Script> { { 0x1E7F0, 0x1E7FE }, unicode::Script::Ethiopic }, // Lo  [15] ETHIOPIC SYLLABLE GURAGE QWI..ETHIOPIC SYLLABLE GURAGE PWEE
    Prop<unicode::Script> { { 0x1E800, 0x1E8C4 }, unicode::Script::Mende_Kikakui }, // Lo [197] MENDE KIKAKUI SYLLABLE M001 KI..MENDE KIKAKUI SYLLABLE M060 NYON
    Prop<unicode::Script> { { 0x1E8C7, 0x1E8CF }, unicode::Script::Mende_Kikakui }, // No   [9] MENDE KIKAKUI DIGIT ONE..MENDE KIKAKUI DIGIT NINE
    Prop<unicode::Script> { { 0x1E8D0, 0x1E8D6 }, unicode::Script::Mende_Kikakui }, // Mn   [7] MENDE KIKAKUI COMBINING NUMBER TEENS..MENDE KIKAKUI COMBINING NUMBER MILLIONS
    Prop<unicode::Script> { { 0x1E900, 0x1E943 }, unicode::Script::Adlam }, // L&  [68] ADLAM CAPITAL LETTER ALIF..ADLAM SMALL LETTER SHA
    Prop<unicode::Script> { { 0x1E944, 0x1E94A }, unicode::Script::Adlam }, // Mn   [7] ADLAM ALIF LENGTHENER..ADLAM NUKTA
    Prop<unicode::Script> { { 0x1E94B, 0x1E94B }, unicode::Script::Adlam }, // Lm       ADLAM NASALIZATION MARK
    Prop<unicode::Script> { { 0x1E950, 0x1E959 }, unicode::Script::Adlam }, // Nd  [10] ADLAM DIGIT ZERO..ADLAM DIGIT NINE
    Prop<unicode::Script> { { 0x1E95E, 0x1E95F }, unicode::Script::Adlam }, // Po   [2] ADLAM INITIAL EXCLAMATION MARK..ADLAM INITIAL QUESTION MARK
    Prop<unicode::Script> { { 0x1EC71, 0x1ECAB }, unicode::Script::Common }, // No  [59] INDIC SIYAQ NUMBER ONE..INDIC SIYAQ NUMBER PREFIXED NINE
    Prop<unicode::Script> { { 0x1ECAC, 0x1ECAC }, unicode::Script::Common }, // So       INDIC SIYAQ PLACEHOLDER
    Prop<unicode::Script> { { 0x1ECAD, 0x1ECAF }, unicode::Script::Common }, // No   [3] INDIC SIYAQ FRACTION ONE QUARTER..INDIC SIYAQ FRACTION THREE QUARTERS
    Prop<unicode::Script> { { 0x1ECB0, 0x1ECB0 }, unicode::Script::Common }, // Sc       INDIC SIYAQ RUPEE MARK
    Prop<unicode::Script> { { 0x1ECB1, 0x1ECB4 }, unicode::Script::Common }, // No   [4] INDIC SIYAQ NUMBER ALTERNATE ONE..INDIC SIYAQ ALTERNATE LAKH MARK
    Prop<unicode::Script> { { 0x1ED01, 0x1ED2D }, unicode::Script::Common }, // No  [45] OTTOMAN SIYAQ NUMBER ONE..OTTOMAN SIYAQ NUMBER NINETY THOUSAND
    Prop<unicode::Script> { { 0x1ED2E, 0x1ED2E }, unicode::Script::Common }, // So       OTTOMAN SIYAQ MARRATAN
    Prop<unicode::Script> { { 0x1ED2F, 0x1ED3D }, unicode::Script::Common }, // No  [15] OTTOMAN SIYAQ ALTERNATE NUMBER TWO..OTTOMAN SIYAQ FRACTION ONE SIXTH
    Prop<unicode::Script> { { 0x1EE00, 0x1EE03 }, unicode::Script::Arabic }, // Lo   [4] ARABIC MATHEMATICAL ALEF..ARABIC MATHEMATICAL DAL
    Prop<unicode::Script> { { 0x1EE05, 0x1EE1F }, unicode::Script::Arabic }, // Lo  [27] ARABIC MATHEMATICAL WAW..ARABIC MATHEMATICAL DOTLESS QAF
    Prop<unicode::Script> { { 0x1EE21, 0x1EE22 }, unicode::Script::Arabic }, // Lo   [2] ARABIC MATHEMATICAL INITIAL BEH..ARABIC MATHEMATICAL INITIAL JEEM
    Prop<unicode::Script> { { 0x1EE24, 0x1EE24 }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL INITIAL HEH
    Prop<unicode::Script> { { 0x1EE27, 0x1EE27 }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL INITIAL HAH
    Prop<unicode::Script> { { 0x1EE29, 0x1EE32 }, unicode::Script::Arabic }, // Lo  [10] ARABIC MATHEMATICAL INITIAL YEH..ARABIC MATHEMATICAL INITIAL QAF
    Prop<unicode::Script> { { 0x1EE34, 0x1EE37 }, unicode::Script::Arabic }, // Lo   [4] ARABIC MATHEMATICAL INITIAL SHEEN..ARABIC MATHEMATICAL INITIAL KHAH
    Prop<unicode::Script> { { 0x1EE39, 0x1EE39 }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL INITIAL DAD
    Prop<unicode::Script> { { 0x1EE3B, 0x1EE3B }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL INITIAL GHAIN
    Prop<unicode::Script> { { 0x1EE42, 0x1EE42 }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL TAILED JEEM
    Prop<unicode::Script> { { 0x1EE47, 0x1EE47 }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL TAILED HAH
    Prop<unicode::Script> { { 0x1EE49, 0x1EE49 }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL TAILED YEH
    Prop<unicode::Script> { { 0x1EE4B, 0x1EE4B }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL TAILED LAM
    Prop<unicode::Script> { { 0x1EE4D, 0x1EE4F }, unicode::Script::Arabic }, // Lo   [3] ARABIC MATHEMATICAL TAILED NOON..ARABIC MATHEMATICAL TAILED AIN
    Prop<unicode::Script> { { 0x1EE51, 0x1EE52 }, unicode::Script::Arabic }, // Lo   [2] ARABIC MATHEMATICAL TAILED SAD..ARABIC MATHEMATICAL TAILED QAF
    Prop<unicode::Script> { { 0x1EE54, 0x1EE54 }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL TAILED SHEEN
    Prop<unicode::Script> { { 0x1EE57, 0x1EE57 }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL TAILED KHAH
    Prop<unicode::Script> { { 0x1EE59, 0x1EE59 }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL TAILED DAD
    Prop<unicode::Script> { { 0x1EE5B, 0x1EE5B }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL TAILED GHAIN
    Prop<unicode::Script> { { 0x1EE5D, 0x1EE5D }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS NOON
    Prop<unicode::Script> { { 0x1EE5F, 0x1EE5F }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL TAILED DOTLESS QAF
    Prop<unicode::Script> { { 0x1EE61, 0x1EE62 }, unicode::Script::Arabic }, // Lo   [2] ARABIC MATHEMATICAL STRETCHED BEH..ARABIC MATHEMATICAL STRETCHED JEEM
    Prop<unicode::Script> { { 0x1EE64, 0x1EE64 }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL STRETCHED HEH
    Prop<unicode::Script> { { 0x1EE67, 0x1EE6A }, unicode::Script::Arabic }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED HAH..ARABIC MATHEMATICAL STRETCHED KAF
    Prop<unicode::Script> { { 0x1EE6C, 0x1EE72 }, unicode::Script::Arabic }, // Lo   [7] ARABIC MATHEMATICAL STRETCHED MEEM..ARABIC MATHEMATICAL STRETCHED QAF
    Prop<unicode::Script> { { 0x1EE74, 0x1EE77 }, unicode::Script::Arabic }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED SHEEN..ARABIC MATHEMATICAL STRETCHED KHAH
    Prop<unicode::Script> { { 0x1EE79, 0x1EE7C }, unicode::Script::Arabic }, // Lo   [4] ARABIC MATHEMATICAL STRETCHED DAD..ARABIC MATHEMATICAL STRETCHED DOTLESS BEH
    Prop<unicode::Script> { { 0x1EE7E, 0x1EE7E }, unicode::Script::Arabic }, // Lo       ARABIC MATHEMATICAL STRETCHED DOTLESS FEH
    Prop<unicode::Script> { { 0x1EE80, 0x1EE89 }, unicode::Script::Arabic }, // Lo  [10] ARABIC MATHEMATICAL LOOPED ALEF..ARABIC MATHEMATICAL LOOPED YEH
    Prop<unicode::Script> { { 0x1EE8B, 0x1EE9B }, unicode::Script::Arabic }, // Lo  [17] ARABIC MATHEMATICAL LOOPED LAM..ARABIC MATHEMATICAL LOOPED GHAIN
    Prop<unicode::Script> { { 0x1EEA1, 0x1EEA3 }, unicode::Script::Arabic }, // Lo   [3] ARABIC MATHEMATICAL DOUBLE-STRUCK BEH..ARABIC MATHEMATICAL DOUBLE-STRUCK DAL
    Prop<unicode::Script> { { 0x1EEA5, 0x1EEA9 }, unicode::Script::Arabic }, // Lo   [5] ARABIC MATHEMATICAL DOUBLE-STRUCK WAW..ARABIC MATHEMATICAL DOUBLE-STRUCK YEH
    Prop<unicode::Script> { { 0x1EEAB, 0x1EEBB }, unicode::Script::Arabic }, // Lo  [17] ARABIC MATHEMATICAL DOUBLE-STRUCK LAM..ARABIC MATHEMATICAL DOUBLE-STRUCK GHAIN
    Prop<unicode::Script> { { 0x1EEF0, 0x1EEF1 }, unicode::Script::Arabic }, // Sm   [2] ARABIC MATHEMATICAL OPERATOR MEEM WITH HAH WITH TATWEEL..ARABIC MATHEMATICAL OPERATOR HAH WITH DAL
    Prop<unicode::Script> { { 0x1F000, 0x1F02B }, unicode::Script::Common }, // So  [44] MAHJONG TILE EAST WIND..MAHJONG TILE BACK
    Prop<unicode::Script> { { 0x1F030, 0x1F093 }, unicode::Script::Common }, // So [100] DOMINO TILE HORIZONTAL BACK..DOMINO TILE VERTICAL-06-06
    Prop<unicode::Script> { { 0x1F0A0, 0x1F0AE }, unicode::Script::Common }, // So  [15] PLAYING CARD BACK..PLAYING CARD KING OF SPADES
    Prop<unicode::Script> { { 0x1F0B1, 0x1F0BF }, unicode::Script::Common }, // So  [15] PLAYING CARD ACE OF HEARTS..PLAYING CARD RED JOKER
    Prop<unicode::Script> { { 0x1F0C1, 0x1F0CF }, unicode::Script::Common }, // So  [15] PLAYING CARD ACE OF DIAMONDS..PLAYING CARD BLACK JOKER
    Prop<unicode::Script> { { 0x1F0D1, 0x1F0F5 }, unicode::Script::Common }, // So  [37] PLAYING CARD ACE OF CLUBS..PLAYING CARD TRUMP-21
    Prop<unicode::Script> { { 0x1F100, 0x1F10C }, unicode::Script::Common }, // No  [13] DIGIT ZERO FULL STOP..DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT ZERO
    Prop<unicode::Script> { { 0x1F10D, 0x1F1AD }, unicode::Script::Common }, // So [161] CIRCLED ZERO WITH SLASH..MASK WORK SYMBOL
    Prop<unicode::Script> { { 0x1F1E6, 0x1F1FF }, unicode::Script::Common }, // So  [26] REGIONAL INDICATOR SYMBOL LETTER A..REGIONAL INDICATOR SYMBOL LETTER Z
    Prop<unicode::Script> { { 0x1F200, 0x1F200 }, unicode::Script::Hiragana }, // So       SQUARE HIRAGANA HOKA
    Prop<unicode::Script> { { 0x1F201, 0x1F202 }, unicode::Script::Common }, // So   [2] SQUARED KATAKANA KOKO..SQUARED KATAKANA SA
    Prop<unicode::Script> { { 0x1F210, 0x1F23B }, unicode::Script::Common }, // So  [44] SQUARED CJK UNIFIED IDEOGRAPH-624B..SQUARED CJK UNIFIED IDEOGRAPH-914D
    Prop<unicode::Script> { { 0x1F240, 0x1F248 }, unicode::Script::Common }, // So   [9] TORTOISE SHELL BRACKETED CJK UNIFIED IDEOGRAPH-672C..TORTOISE SHELL BRACKETED CJK UNIFIED IDEOGRAPH-6557
    Prop<unicode::Script> { { 0x1F250, 0x1F251 }, unicode::Script::Common }, // So   [2] CIRCLED IDEOGRAPH ADVANTAGE..CIRCLED IDEOGRAPH ACCEPT
    Prop<unicode::Script> { { 0x1F260, 0x1F265 }, unicode::Script::Common }, // So   [6] ROUNDED SYMBOL FOR FU..ROUNDED SYMBOL FOR CAI
    Prop<unicode::Script> { { 0x1F300, 0x1F3FA }, unicode::Script::Common }, // So [251] CYCLONE..AMPHORA
    Prop<unicode::Script> { { 0x1F3FB, 0x1F3FF }, unicode::Script::Common }, // Sk   [5] EMOJI MODIFIER FITZPATRICK TYPE-1-2..EMOJI MODIFIER FITZPATRICK TYPE-6
    Prop<unicode::Script> { { 0x1F400, 0x1F6D8 }, unicode::Script::Common }, // So [729] RAT..LANDSLIDE
    Prop<unicode::Script> { { 0x1F6DC, 0x1F6EC }, unicode::Script::Common }, // So  [17] WIRELESS..AIRPLANE ARRIVING
    Prop<unicode::Script> { { 0x1F6F0, 0x1F6FC }, unicode::Script::Common }, // So  [13] SATELLITE..ROLLER SKATE
    Prop<unicode::Script> { { 0x1F700, 0x1F7D9 }, unicode::Script::Common }, // So [218] ALCHEMICAL SYMBOL FOR QUINTESSENCE..NINE POINTED WHITE STAR
    Prop<unicode::Script> { { 0x1F7E0, 0x1F7EB }, unicode::Script::Common }, // So  [12] LARGE ORANGE CIRCLE..LARGE BROWN SQUARE
    Prop<unicode::Script> { { 0x1F7F0, 0x1F7F0 }, unicode::Script::Common }, // So       HEAVY EQUALS SIGN
    Prop<unicode::Script> { { 0x1F800, 0x1F80B }, unicode::Script::Common }, // So  [12] LEFTWARDS ARROW WITH SMALL TRIANGLE ARROWHEAD..DOWNWARDS ARROW WITH LARGE TRIANGLE ARROWHEAD
    Prop<unicode::Script> { { 0x1F810, 0x1F847 }, unicode::Script::Common }, // So  [56] LEFTWARDS ARROW WITH SMALL EQUILATERAL ARROWHEAD..DOWNWARDS HEAVY ARROW
    Prop<unicode::Script> { { 0x1F850, 0x1F859 }, unicode::Script::Common }, // So  [10] LEFTWARDS SANS-SERIF ARROW..UP DOWN SANS-SERIF ARROW
    Prop<unicode::Script> { { 0x1F860, 0x1F887 }, unicode::Script::Common }, // So  [40] WIDE-HEADED LEFTWARDS LIGHT BARB ARROW..WIDE-HEADED SOUTH WEST VERY HEAVY BARB ARROW
    Prop<unicode::Script> { { 0x1F890, 0x1F8AD }, unicode::Script::Common }, // So  [30] LEFTWARDS TRIANGLE ARROWHEAD..WHITE ARROW SHAFT WIDTH TWO THIRDS
    Prop<unicode::Script> { { 0x1F8B0, 0x1F8BB }, unicode::Script::Common }, // So  [12] ARROW POINTING UPWARDS THEN NORTH WEST..SOUTH WEST ARROW FROM BAR
    Prop<unicode::Script> { { 0x1F8C0, 0x1F8C1 }, unicode::Script::Common }, // So   [2] LEFTWARDS ARROW FROM DOWNWARDS ARROW..RIGHTWARDS ARROW FROM DOWNWARDS ARROW
    Prop<unicode::Script> { { 0x1F8D0, 0x1F8D8 }, unicode::Script::Common }, // Sm   [9] LONG RIGHTWARDS ARROW OVER LONG LEFTWARDS ARROW..LONG LEFT RIGHT ARROW WITH DEPENDENT LOBE
    Prop<unicode::Script> { { 0x1F900, 0x1FA57 }, unicode::Script::Common }, // So [344] CIRCLED CROSS FORMEE WITH FOUR DOTS..BLACK CHESS ALFIL
    Prop<unicode::Script> { { 0x1FA60, 0x1FA6D }, unicode::Script::Common }, // So  [14] XIANGQI RED GENERAL..XIANGQI BLACK SOLDIER
    Prop<unicode::Script> { { 0x1FA70, 0x1FA7C }, unicode::Script::Common }, // So  [13] BALLET SHOES..CRUTCH
    Prop<unicode::Script> { { 0x1FA80, 0x1FA8A }, unicode::Script::Common }, // So  [11] YO-YO..TROMBONE
    Prop<unicode::Script> { { 0x1FA8E, 0x1FAC6 }, unicode::Script::Common }, // So  [57] TREASURE CHEST..FINGERPRINT
    Prop<unicode::Script> { { 0x1FAC8, 0x1FAC8 }, unicode::Script::Common }, // So       HAIRY CREATURE
    Prop<unicode::Script> { { 0x1FACD, 0x1FADC }, unicode::Script::Common }, // So  [16] ORCA..ROOT VEGETABLE
    Prop<unicode::Script> { { 0x1FADF, 0x1FAEA }, unicode::Script::Common }, // So  [12] SPLATTER..DISTORTED FACE
    Prop<unicode::Script> { { 0x1FAEF, 0x1FAF8 }, unicode::Script::Common }, // So  [10] FIGHT CLOUD..RIGHTWARDS PUSHING HAND
    Prop<unicode::Script> { { 0x1FB00, 0x1FB92 }, unicode::Script::Common }, // So [147] BLOCK SEXTANT-1..UPPER HALF INVERSE MEDIUM SHADE AND LOWER HALF BLOCK
    Prop<unicode::Script> { { 0x1FB94, 0x1FBEF }, unicode::Script::Common }, // So  [92] LEFT HALF INVERSE MEDIUM SHADE AND RIGHT HALF BLOCK..TOP LEFT JUSTIFIED LOWER RIGHT QUARTER BLACK CIRCLE
    Prop<unicode::Script> { { 0x1FBF0, 0x1FBF9 }, unicode::Script::Common }, // Nd  [10] SEGMENTED DIGIT ZERO..SEGMENTED DIGIT NINE
    Prop<unicode::Script> { { 0x1FBFA, 0x1FBFA }, unicode::Script::Common }, // So       ALARM BELL SYMBOL
    Prop<unicode::Script> { { 0x20000, 0x2A6DF }, unicode::Script::Han }, // Lo [42720] CJK UNIFIED IDEOGRAPH-20000..CJK UNIFIED IDEOGRAPH-2A6DF
    Prop<unicode::Script> { { 0x2A700, 0x2B81D }, unicode::Script::Han }, // Lo [4382] CJK UNIFIED IDEOGRAPH-2A700..CJK UNIFIED IDEOGRAPH-2B81D
    Prop<unicode::Script> { { 0x2B820, 0x2CEAD }, unicode::Script::Han }, // Lo [5774] CJK UNIFIED IDEOGRAPH-2B820..CJK UNIFIED IDEOGRAPH-2CEAD
    Prop<unicode::Script> { { 0x2CEB0, 0x2EBE0 }, unicode::Script::Han }, // Lo [7473] CJK UNIFIED IDEOGRAPH-2CEB0..CJK UNIFIED IDEOGRAPH-2EBE0
    Prop<unicode::Script> { { 0x2EBF0, 0x2EE5D }, unicode::Script::Han }, // Lo [622] CJK UNIFIED IDEOGRAPH-2EBF0..CJK UNIFIED IDEOGRAPH-2EE5D
    Prop<unicode::Script> { { 0x2F800, 0x2FA1D }, unicode::Script::Han }, // Lo [542] CJK COMPATIBILITY IDEOGRAPH-2F800..CJK COMPATIBILITY IDEOGRAPH-2FA1D
    Prop<unicode::Script> { { 0x30000, 0x3134A }, unicode::Script::Han }, // Lo [4939] CJK UNIFIED IDEOGRAPH-30000..CJK UNIFIED IDEOGRAPH-3134A
    Prop<unicode::Script> { { 0x31350, 0x33479 }, unicode::Script::Han }, // Lo [8490] CJK UNIFIED IDEOGRAPH-31350..CJK UNIFIED IDEOGRAPH-33479
    Prop<unicode::Script> { { 0xE0001, 0xE0001 }, unicode::Script::Common }, // Cf       LANGUAGE TAG
    Prop<unicode::Script> { { 0xE0020, 0xE007F }, unicode::Script::Common }, // Cf  [96] TAG SPACE..CANCEL TAG
    Prop<unicode::Script> { { 0xE0100, 0xE01EF }, unicode::Script::Inherited }, // Mn [240] VARIATION SELECTOR-17..VARIATION SELECTOR-256
}; // }}}
} // end namespace tables

Script script(char32_t codepoint) noexcept {
    return search(tables::Script, codepoint).value_or(Script::Unknown);
}

namespace tables { // {{{ ScriptExtensions
auto static const sce_Avst_Cari_Copt_Dupl_Elba_Geor_Glag_Gong_Goth_Grek_Hani_Latn_Lydi_Mahj_Perm_Shaw = std::array<unicode::Script, 16>{
    unicode::Script::Avestan,
    unicode::Script::Carian,
    unicode::Script::Coptic,
    unicode::Script::Duployan,
    unicode::Script::Elbasan,
    unicode::Script::Georgian,
    unicode::Script::Glagolitic,
    unicode::Script::Gunjala_Gondi,
    unicode::Script::Gothic,
    unicode::Script::Greek,
    unicode::Script::Han,
    unicode::Script::Latin,
    unicode::Script::Lydian,
    unicode::Script::Mahajani,
    unicode::Script::Old_Permic,
    unicode::Script::Shavian,
};

auto static const sce_Beng_Cyrl_Deva_Latn_Lisu_Thai_Toto = std::array<unicode::Script, 7>{
    unicode::Script::Bengali,
    unicode::Script::Cyrillic,
    unicode::Script::Devanagari,
    unicode::Script::Latin,
    unicode::Script::Lisu,
    unicode::Script::Thai,
    unicode::Script::Toto,
};

auto static const sce_Bopo_Latn = std::array<unicode::Script, 2>{
    unicode::Script::Bopomofo,
    unicode::Script::Latin,
};

auto static const sce_Latn_Lisu = std::array<unicode::Script, 2>{
    unicode::Script::Latin,
    unicode::Script::Lisu,
};

auto static const sce_Latn_Thai = std::array<unicode::Script, 2>{
    unicode::Script::Latin,
    unicode::Script::Thai,
};

auto static const sce_Cher_Copt_Cyrl_Grek_Latn_Perm_Sunu_Tale = std::array<unicode::Script, 8>{
    unicode::Script::Cherokee,
    unicode::Script::Coptic,
    unicode::Script::Cyrillic,
    unicode::Script::Greek,
    unicode::Script::Latin,
    unicode::Script::Old_Permic,
    unicode::Script::Sunuwar,
    unicode::Script::Tai_Le,
};

auto static const sce_Cher_Cyrl_Grek_Latn_Osge_Sunu_Tale_Todr = std::array<unicode::Script, 8>{
    unicode::Script::Cherokee,
    unicode::Script::Cyrillic,
    unicode::Script::Greek,
    unicode::Script::Latin,
    unicode::Script::Osage,
    unicode::Script::Sunuwar,
    unicode::Script::Tai_Le,
    unicode::Script::Todhri,
};

auto static const sce_Cher_Cyrl_Latn_Tfng = std::array<unicode::Script, 4>{
    unicode::Script::Cherokee,
    unicode::Script::Cyrillic,
    unicode::Script::Latin,
    unicode::Script::Tifinagh,
};

auto static const sce_Glag_Latn_Sunu_Syrc_Thai = std::array<unicode::Script, 5>{
    unicode::Script::Glagolitic,
    unicode::Script::Latin,
    unicode::Script::Sunuwar,
    unicode::Script::Syriac,
    unicode::Script::Thai,
};

auto static const sce_Aghb_Cher_Copt_Cyrl_Goth_Grek_Latn_Osge_Syrc_Tfng_Todr = std::array<unicode::Script, 11>{
    unicode::Script::Caucasian_Albanian,
    unicode::Script::Cherokee,
    unicode::Script::Coptic,
    unicode::Script::Cyrillic,
    unicode::Script::Gothic,
    unicode::Script::Greek,
    unicode::Script::Latin,
    unicode::Script::Osage,
    unicode::Script::Syriac,
    unicode::Script::Tifinagh,
    unicode::Script::Todhri,
};

auto static const sce_Copt_Elba_Glag_Goth_Kana_Latn = std::array<unicode::Script, 6>{
    unicode::Script::Coptic,
    unicode::Script::Elbasan,
    unicode::Script::Glagolitic,
    unicode::Script::Gothic,
    unicode::Script::Katakana,
    unicode::Script::Latin,
};

auto static const sce_Cyrl_Grek_Latn_Perm_Tfng = std::array<unicode::Script, 5>{
    unicode::Script::Cyrillic,
    unicode::Script::Greek,
    unicode::Script::Latin,
    unicode::Script::Old_Permic,
    unicode::Script::Tifinagh,
};

auto static const sce_Copt_Dupl_Hebr_Latn_Perm_Syrc_Tale_Tfng_Todr = std::array<unicode::Script, 9>{
    unicode::Script::Coptic,
    unicode::Script::Duployan,
    unicode::Script::Hebrew,
    unicode::Script::Latin,
    unicode::Script::Old_Permic,
    unicode::Script::Syriac,
    unicode::Script::Tai_Le,
    unicode::Script::Tifinagh,
    unicode::Script::Todhri,
};

auto static const sce_Armn_Cyrl_Dupl_Goth_Grek_Hebr_Latn_Perm_Syrc_Tale_Tfng = std::array<unicode::Script, 11>{
    unicode::Script::Armenian,
    unicode::Script::Cyrillic,
    unicode::Script::Duployan,
    unicode::Script::Gothic,
    unicode::Script::Greek,
    unicode::Script::Hebrew,
    unicode::Script::Latin,
    unicode::Script::Old_Permic,
    unicode::Script::Syriac,
    unicode::Script::Tai_Le,
    unicode::Script::Tifinagh,
};

auto static const sce_Latn_Tfng = std::array<unicode::Script, 2>{
    unicode::Script::Latin,
    unicode::Script::Tifinagh,
};

auto static const sce_Dupl_Latn_Syrc = std::array<unicode::Script, 3>{
    unicode::Script::Duployan,
    unicode::Script::Latin,
    unicode::Script::Syriac,
};

auto static const sce_Cher_Cyrl_Latn_Osge = std::array<unicode::Script, 4>{
    unicode::Script::Cherokee,
    unicode::Script::Cyrillic,
    unicode::Script::Latin,
    unicode::Script::Osage,
};

auto static const sce_Cher_Latn_Tale = std::array<unicode::Script, 3>{
    unicode::Script::Cherokee,
    unicode::Script::Latin,
    unicode::Script::Tai_Le,
};

auto static const sce_Latn_Sunu = std::array<unicode::Script, 2>{
    unicode::Script::Latin,
    unicode::Script::Sunuwar,
};

auto static const sce_Ethi_Latn = std::array<unicode::Script, 2>{
    unicode::Script::Ethiopic,
    unicode::Script::Latin,
};

auto static const sce_Cyrl_Latn_Todr = std::array<unicode::Script, 3>{
    unicode::Script::Cyrillic,
    unicode::Script::Latin,
    unicode::Script::Todhri,
};

auto static const sce_Grek_Latn_Perm_Todr = std::array<unicode::Script, 4>{
    unicode::Script::Greek,
    unicode::Script::Latin,
    unicode::Script::Old_Permic,
    unicode::Script::Todhri,
};

auto static const sce_Cher_Dupl_Kana_Latn_Syrc_Tfng = std::array<unicode::Script, 6>{
    unicode::Script::Cherokee,
    unicode::Script::Duployan,
    unicode::Script::Katakana,
    unicode::Script::Latin,
    unicode::Script::Syriac,
    unicode::Script::Tifinagh,
};

auto static const sce_Cher_Dupl_Latn_Syrc = std::array<unicode::Script, 4>{
    unicode::Script::Cherokee,
    unicode::Script::Duployan,
    unicode::Script::Latin,
    unicode::Script::Syriac,
};

auto static const sce_Latn_Syrc = std::array<unicode::Script, 2>{
    unicode::Script::Latin,
    unicode::Script::Syriac,
};

auto static const sce_Latn_Sunu_Syrc = std::array<unicode::Script, 3>{
    unicode::Script::Latin,
    unicode::Script::Sunuwar,
    unicode::Script::Syriac,
};

auto static const sce_Cher_Latn_Syrc = std::array<unicode::Script, 3>{
    unicode::Script::Cherokee,
    unicode::Script::Latin,
    unicode::Script::Syriac,
};

auto static const sce_Aghb_Cher_Goth_Latn_Sunu_Syrc_Thai = std::array<unicode::Script, 7>{
    unicode::Script::Caucasian_Albanian,
    unicode::Script::Cherokee,
    unicode::Script::Gothic,
    unicode::Script::Latin,
    unicode::Script::Sunuwar,
    unicode::Script::Syriac,
    unicode::Script::Thai,
};

auto static const sce_Grek = std::array<unicode::Script, 1>{
    unicode::Script::Greek,
};

auto static const sce_Latn_Osge = std::array<unicode::Script, 2>{
    unicode::Script::Latin,
    unicode::Script::Osage,
};

auto static const sce_Aghb_Latn_Todr = std::array<unicode::Script, 3>{
    unicode::Script::Caucasian_Albanian,
    unicode::Script::Latin,
    unicode::Script::Todhri,
};

auto static const sce_Latn = std::array<unicode::Script, 1>{
    unicode::Script::Latin,
};

auto static const sce_Copt_Grek = std::array<unicode::Script, 2>{
    unicode::Script::Coptic,
    unicode::Script::Greek,
};

auto static const sce_Cyrl_Perm = std::array<unicode::Script, 2>{
    unicode::Script::Cyrillic,
    unicode::Script::Old_Permic,
};

auto static const sce_Cyrl_Glag = std::array<unicode::Script, 2>{
    unicode::Script::Cyrillic,
    unicode::Script::Glagolitic,
};

auto static const sce_Cyrl_Latn = std::array<unicode::Script, 2>{
    unicode::Script::Cyrillic,
    unicode::Script::Latin,
};

auto static const sce_Armn_Geor_Glag = std::array<unicode::Script, 3>{
    unicode::Script::Armenian,
    unicode::Script::Georgian,
    unicode::Script::Glagolitic,
};

auto static const sce_Arab_Gara_Nkoo_Rohg_Syrc_Thaa_Yezi = std::array<unicode::Script, 7>{
    unicode::Script::Arabic,
    unicode::Script::Garay,
    unicode::Script::Nko,
    unicode::Script::Hanifi_Rohingya,
    unicode::Script::Syriac,
    unicode::Script::Thaana,
    unicode::Script::Yezidi,
};

auto static const sce_Arab_Syrc_Thaa = std::array<unicode::Script, 3>{
    unicode::Script::Arabic,
    unicode::Script::Syriac,
    unicode::Script::Thaana,
};

auto static const sce_Adlm_Arab_Gara_Nkoo_Rohg_Syrc_Thaa_Yezi = std::array<unicode::Script, 8>{
    unicode::Script::Adlam,
    unicode::Script::Arabic,
    unicode::Script::Garay,
    unicode::Script::Nko,
    unicode::Script::Hanifi_Rohingya,
    unicode::Script::Syriac,
    unicode::Script::Thaana,
    unicode::Script::Yezidi,
};

auto static const sce_Adlm_Arab_Mand_Mani_Ougr_Phlp_Rohg_Sogd_Syrc = std::array<unicode::Script, 9>{
    unicode::Script::Adlam,
    unicode::Script::Arabic,
    unicode::Script::Mandaic,
    unicode::Script::Manichaean,
    unicode::Script::Old_Uyghur,
    unicode::Script::Psalter_Pahlavi,
    unicode::Script::Hanifi_Rohingya,
    unicode::Script::Sogdian,
    unicode::Script::Syriac,
};

auto static const sce_Arab_Syrc = std::array<unicode::Script, 2>{
    unicode::Script::Arabic,
    unicode::Script::Syriac,
};

auto static const sce_Arab_Thaa_Yezi = std::array<unicode::Script, 3>{
    unicode::Script::Arabic,
    unicode::Script::Thaana,
    unicode::Script::Yezidi,
};

auto static const sce_Arab_Rohg = std::array<unicode::Script, 2>{
    unicode::Script::Arabic,
    unicode::Script::Hanifi_Rohingya,
};

auto static const sce_Beng_Deva_Gran_Gujr_Guru_Knda_Latn_Mlym_Nand_Newa_Orya_Shrd_Taml_Telu_Tirh = std::array<unicode::Script, 15>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
    unicode::Script::Grantha,
    unicode::Script::Gujarati,
    unicode::Script::Gurmukhi,
    unicode::Script::Kannada,
    unicode::Script::Latin,
    unicode::Script::Malayalam,
    unicode::Script::Nandinagari,
    unicode::Script::Newa,
    unicode::Script::Oriya,
    unicode::Script::Sharada,
    unicode::Script::Tamil,
    unicode::Script::Telugu,
    unicode::Script::Tirhuta,
};

auto static const sce_Beng_Deva_Gran_Gujr_Guru_Knda_Latn_Mlym_Newa_Orya_Taml_Telu_Tirh = std::array<unicode::Script, 13>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
    unicode::Script::Grantha,
    unicode::Script::Gujarati,
    unicode::Script::Gurmukhi,
    unicode::Script::Kannada,
    unicode::Script::Latin,
    unicode::Script::Malayalam,
    unicode::Script::Newa,
    unicode::Script::Oriya,
    unicode::Script::Tamil,
    unicode::Script::Telugu,
    unicode::Script::Tirhuta,
};

auto static const sce_Beng_Deva_Dogr_Gong_Gonm_Gran_Gujr_Guru_Knda_Mahj_Mlym_Nand_Onao_Orya_Sind_Sinh_Sylo_Takr_Taml_Telu_Tirh = std::array<unicode::Script, 21>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
    unicode::Script::Dogra,
    unicode::Script::Gunjala_Gondi,
    unicode::Script::Masaram_Gondi,
    unicode::Script::Grantha,
    unicode::Script::Gujarati,
    unicode::Script::Gurmukhi,
    unicode::Script::Kannada,
    unicode::Script::Mahajani,
    unicode::Script::Malayalam,
    unicode::Script::Nandinagari,
    unicode::Script::Ol_Onal,
    unicode::Script::Oriya,
    unicode::Script::Khudawadi,
    unicode::Script::Sinhala,
    unicode::Script::Syloti_Nagri,
    unicode::Script::Takri,
    unicode::Script::Tamil,
    unicode::Script::Telugu,
    unicode::Script::Tirhuta,
};

auto static const sce_Beng_Deva_Dogr_Gong_Gonm_Gran_Gujr_Gukh_Guru_Knda_Limb_Mahj_Mlym_Nand_Onao_Orya_Sind_Sinh_Sylo_Takr_Taml_Telu_Tirh = std::array<unicode::Script, 23>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
    unicode::Script::Dogra,
    unicode::Script::Gunjala_Gondi,
    unicode::Script::Masaram_Gondi,
    unicode::Script::Grantha,
    unicode::Script::Gujarati,
    unicode::Script::Gurung_Khema,
    unicode::Script::Gurmukhi,
    unicode::Script::Kannada,
    unicode::Script::Limbu,
    unicode::Script::Mahajani,
    unicode::Script::Malayalam,
    unicode::Script::Nandinagari,
    unicode::Script::Ol_Onal,
    unicode::Script::Oriya,
    unicode::Script::Khudawadi,
    unicode::Script::Sinhala,
    unicode::Script::Syloti_Nagri,
    unicode::Script::Takri,
    unicode::Script::Tamil,
    unicode::Script::Telugu,
    unicode::Script::Tirhuta,
};

auto static const sce_Deva_Dogr_Kthi_Mahj = std::array<unicode::Script, 4>{
    unicode::Script::Devanagari,
    unicode::Script::Dogra,
    unicode::Script::Kaithi,
    unicode::Script::Mahajani,
};

auto static const sce_Beng_Cakm_Sylo = std::array<unicode::Script, 3>{
    unicode::Script::Bengali,
    unicode::Script::Chakma,
    unicode::Script::Syloti_Nagri,
};

auto static const sce_Guru_Mult = std::array<unicode::Script, 2>{
    unicode::Script::Gurmukhi,
    unicode::Script::Multani,
};

auto static const sce_Gujr_Khoj = std::array<unicode::Script, 2>{
    unicode::Script::Gujarati,
    unicode::Script::Khojki,
};

auto static const sce_Gran_Taml = std::array<unicode::Script, 2>{
    unicode::Script::Grantha,
    unicode::Script::Tamil,
};

auto static const sce_Knda_Nand_Tutg = std::array<unicode::Script, 3>{
    unicode::Script::Kannada,
    unicode::Script::Nandinagari,
    unicode::Script::Tulu_Tigalari,
};

auto static const sce_Cakm_Mymr_Tale = std::array<unicode::Script, 3>{
    unicode::Script::Chakma,
    unicode::Script::Myanmar,
    unicode::Script::Tai_Le,
};

auto static const sce_Geor_Glag_Latn = std::array<unicode::Script, 3>{
    unicode::Script::Georgian,
    unicode::Script::Glagolitic,
    unicode::Script::Latin,
};

auto static const sce_Runr = std::array<unicode::Script, 1>{
    unicode::Script::Runic,
};

auto static const sce_Buhd_Hano_Tagb_Tglg = std::array<unicode::Script, 4>{
    unicode::Script::Buhid,
    unicode::Script::Hanunoo,
    unicode::Script::Tagbanwa,
    unicode::Script::Tagalog,
};

auto static const sce_Mong_Phag = std::array<unicode::Script, 2>{
    unicode::Script::Mongolian,
    unicode::Script::Phags_Pa,
};

auto static const sce_Beng_Deva_Gran_Knda = std::array<unicode::Script, 4>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
    unicode::Script::Grantha,
    unicode::Script::Kannada,
};

auto static const sce_Deva = std::array<unicode::Script, 1>{
    unicode::Script::Devanagari,
};

auto static const sce_Deva_Gran_Knda = std::array<unicode::Script, 3>{
    unicode::Script::Devanagari,
    unicode::Script::Grantha,
    unicode::Script::Kannada,
};

auto static const sce_Beng_Deva_Newa_Telu_Tirh = std::array<unicode::Script, 5>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
    unicode::Script::Newa,
    unicode::Script::Telugu,
    unicode::Script::Tirhuta,
};

auto static const sce_Beng_Deva_Telu = std::array<unicode::Script, 3>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
    unicode::Script::Telugu,
};

auto static const sce_Deva_Newa_Shrd = std::array<unicode::Script, 3>{
    unicode::Script::Devanagari,
    unicode::Script::Newa,
    unicode::Script::Sharada,
};

auto static const sce_Beng_Deva_Newa_Telu = std::array<unicode::Script, 4>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
    unicode::Script::Newa,
    unicode::Script::Telugu,
};

auto static const sce_Deva_Shrd = std::array<unicode::Script, 2>{
    unicode::Script::Devanagari,
    unicode::Script::Sharada,
};

auto static const sce_Deva_Knda_Mlym_Orya_Taml_Telu = std::array<unicode::Script, 6>{
    unicode::Script::Devanagari,
    unicode::Script::Kannada,
    unicode::Script::Malayalam,
    unicode::Script::Oriya,
    unicode::Script::Tamil,
    unicode::Script::Telugu,
};

auto static const sce_Beng_Deva = std::array<unicode::Script, 2>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
};

auto static const sce_Deva_Newa_Tirh = std::array<unicode::Script, 3>{
    unicode::Script::Devanagari,
    unicode::Script::Newa,
    unicode::Script::Tirhuta,
};

auto static const sce_Deva_Nand_Newa = std::array<unicode::Script, 3>{
    unicode::Script::Devanagari,
    unicode::Script::Nandinagari,
    unicode::Script::Newa,
};

auto static const sce_Beng_Deva_Shrd = std::array<unicode::Script, 3>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
    unicode::Script::Sharada,
};

auto static const sce_Deva_Newa = std::array<unicode::Script, 2>{
    unicode::Script::Devanagari,
    unicode::Script::Newa,
};

auto static const sce_Beng_Deva_Newa_Shrd = std::array<unicode::Script, 4>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
    unicode::Script::Newa,
    unicode::Script::Sharada,
};

auto static const sce_Beng_Deva_Gran_Knda_Mlym_Nand_Orya_Sinh_Telu_Tirh_Tutg = std::array<unicode::Script, 11>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
    unicode::Script::Grantha,
    unicode::Script::Kannada,
    unicode::Script::Malayalam,
    unicode::Script::Nandinagari,
    unicode::Script::Oriya,
    unicode::Script::Sinhala,
    unicode::Script::Telugu,
    unicode::Script::Tirhuta,
    unicode::Script::Tulu_Tigalari,
};

auto static const sce_Deva_Gran = std::array<unicode::Script, 2>{
    unicode::Script::Devanagari,
    unicode::Script::Grantha,
};

auto static const sce_Deva_Gran_Knda_Tutg = std::array<unicode::Script, 4>{
    unicode::Script::Devanagari,
    unicode::Script::Grantha,
    unicode::Script::Kannada,
    unicode::Script::Tulu_Tigalari,
};

auto static const sce_Beng = std::array<unicode::Script, 1>{
    unicode::Script::Bengali,
};

auto static const sce_Nand = std::array<unicode::Script, 1>{
    unicode::Script::Nandinagari,
};

auto static const sce_Cyrl_Latn_Syrc = std::array<unicode::Script, 3>{
    unicode::Script::Cyrillic,
    unicode::Script::Latin,
    unicode::Script::Syriac,
};

auto static const sce_Syrc = std::array<unicode::Script, 1>{
    unicode::Script::Syriac,
};

auto static const sce_Latn_Mong_Phag = std::array<unicode::Script, 3>{
    unicode::Script::Latin,
    unicode::Script::Mongolian,
    unicode::Script::Phags_Pa,
};

auto static const sce_Adlm_Arab = std::array<unicode::Script, 2>{
    unicode::Script::Adlam,
    unicode::Script::Arabic,
};

auto static const sce_Cari_Geor_Glag_Hung_Lyci_Orkh = std::array<unicode::Script, 6>{
    unicode::Script::Carian,
    unicode::Script::Georgian,
    unicode::Script::Glagolitic,
    unicode::Script::Old_Hungarian,
    unicode::Script::Lycian,
    unicode::Script::Old_Turkic,
};

auto static const sce_Cari_Grek_Hung_Mero = std::array<unicode::Script, 4>{
    unicode::Script::Carian,
    unicode::Script::Greek,
    unicode::Script::Old_Hungarian,
    unicode::Script::Meroitic_Hieroglyphs,
};

auto static const sce_Deva_Gran_Latn = std::array<unicode::Script, 3>{
    unicode::Script::Devanagari,
    unicode::Script::Grantha,
    unicode::Script::Latin,
};

auto static const sce_Copt_Latn = std::array<unicode::Script, 2>{
    unicode::Script::Coptic,
    unicode::Script::Latin,
};

auto static const sce_Avst_Orkh = std::array<unicode::Script, 2>{
    unicode::Script::Avestan,
    unicode::Script::Old_Turkic,
};

auto static const sce_Avst_Cari_Geor_Hung_Kthi_Lydi_Samr = std::array<unicode::Script, 7>{
    unicode::Script::Avestan,
    unicode::Script::Carian,
    unicode::Script::Georgian,
    unicode::Script::Old_Hungarian,
    unicode::Script::Kaithi,
    unicode::Script::Lydian,
    unicode::Script::Samaritan,
};

auto static const sce_Dupl = std::array<unicode::Script, 1>{
    unicode::Script::Duployan,
};

auto static const sce_Adlm_Arab_Hung = std::array<unicode::Script, 3>{
    unicode::Script::Adlam,
    unicode::Script::Arabic,
    unicode::Script::Old_Hungarian,
};

auto static const sce_Hani_Tang = std::array<unicode::Script, 2>{
    unicode::Script::Han,
    unicode::Script::Tangut,
};

auto static const sce_Bopo_Hang_Hani_Hira_Kana_Mong_Yiii = std::array<unicode::Script, 7>{
    unicode::Script::Bopomofo,
    unicode::Script::Hangul,
    unicode::Script::Han,
    unicode::Script::Hiragana,
    unicode::Script::Katakana,
    unicode::Script::Mongolian,
    unicode::Script::Yi,
};

auto static const sce_Bopo_Hang_Hani_Hira_Kana_Mong_Phag_Yiii = std::array<unicode::Script, 8>{
    unicode::Script::Bopomofo,
    unicode::Script::Hangul,
    unicode::Script::Han,
    unicode::Script::Hiragana,
    unicode::Script::Katakana,
    unicode::Script::Mongolian,
    unicode::Script::Phags_Pa,
    unicode::Script::Yi,
};

auto static const sce_Bopo_Hang_Hani_Hira_Kana = std::array<unicode::Script, 5>{
    unicode::Script::Bopomofo,
    unicode::Script::Hangul,
    unicode::Script::Han,
    unicode::Script::Hiragana,
    unicode::Script::Katakana,
};

auto static const sce_Hani = std::array<unicode::Script, 1>{
    unicode::Script::Han,
};

auto static const sce_Bopo_Hang_Hani_Hira_Kana_Mong_Tibt_Yiii = std::array<unicode::Script, 8>{
    unicode::Script::Bopomofo,
    unicode::Script::Hangul,
    unicode::Script::Han,
    unicode::Script::Hiragana,
    unicode::Script::Katakana,
    unicode::Script::Mongolian,
    unicode::Script::Tibetan,
    unicode::Script::Yi,
};

auto static const sce_Bopo_Hang_Hani_Hira_Kana_Lisu_Mong_Tibt_Yiii = std::array<unicode::Script, 9>{
    unicode::Script::Bopomofo,
    unicode::Script::Hangul,
    unicode::Script::Han,
    unicode::Script::Hiragana,
    unicode::Script::Katakana,
    unicode::Script::Lisu,
    unicode::Script::Mongolian,
    unicode::Script::Tibetan,
    unicode::Script::Yi,
};

auto static const sce_Bopo_Hang_Hani_Hira_Kana_Yiii = std::array<unicode::Script, 6>{
    unicode::Script::Bopomofo,
    unicode::Script::Hangul,
    unicode::Script::Han,
    unicode::Script::Hiragana,
    unicode::Script::Katakana,
    unicode::Script::Yi,
};

auto static const sce_Bopo_Hani = std::array<unicode::Script, 2>{
    unicode::Script::Bopomofo,
    unicode::Script::Han,
};

auto static const sce_Hira_Kana = std::array<unicode::Script, 2>{
    unicode::Script::Hiragana,
    unicode::Script::Katakana,
};

auto static const sce_Hani_Hira_Kana = std::array<unicode::Script, 3>{
    unicode::Script::Han,
    unicode::Script::Hiragana,
    unicode::Script::Katakana,
};

auto static const sce_Hani_Latn = std::array<unicode::Script, 2>{
    unicode::Script::Han,
    unicode::Script::Latin,
};

auto static const sce_Deva_Dogr_Gujr_Guru_Khoj_Knda_Kthi_Mahj_Mlym_Modi_Nand_Shrd_Sind_Takr_Tirh_Tutg = std::array<unicode::Script, 16>{
    unicode::Script::Devanagari,
    unicode::Script::Dogra,
    unicode::Script::Gujarati,
    unicode::Script::Gurmukhi,
    unicode::Script::Khojki,
    unicode::Script::Kannada,
    unicode::Script::Kaithi,
    unicode::Script::Mahajani,
    unicode::Script::Malayalam,
    unicode::Script::Modi,
    unicode::Script::Nandinagari,
    unicode::Script::Sharada,
    unicode::Script::Khudawadi,
    unicode::Script::Takri,
    unicode::Script::Tirhuta,
    unicode::Script::Tulu_Tigalari,
};

auto static const sce_Deva_Dogr_Gujr_Guru_Khoj_Knda_Kthi_Mahj_Modi_Nand_Shrd_Sind_Takr_Tirh_Tutg = std::array<unicode::Script, 15>{
    unicode::Script::Devanagari,
    unicode::Script::Dogra,
    unicode::Script::Gujarati,
    unicode::Script::Gurmukhi,
    unicode::Script::Khojki,
    unicode::Script::Kannada,
    unicode::Script::Kaithi,
    unicode::Script::Mahajani,
    unicode::Script::Modi,
    unicode::Script::Nandinagari,
    unicode::Script::Sharada,
    unicode::Script::Khudawadi,
    unicode::Script::Takri,
    unicode::Script::Tirhuta,
    unicode::Script::Tulu_Tigalari,
};

auto static const sce_Deva_Dogr_Gujr_Guru_Khoj_Kthi_Mahj_Modi_Sind_Takr_Tirh = std::array<unicode::Script, 11>{
    unicode::Script::Devanagari,
    unicode::Script::Dogra,
    unicode::Script::Gujarati,
    unicode::Script::Gurmukhi,
    unicode::Script::Khojki,
    unicode::Script::Kaithi,
    unicode::Script::Mahajani,
    unicode::Script::Modi,
    unicode::Script::Khudawadi,
    unicode::Script::Takri,
    unicode::Script::Tirhuta,
};

auto static const sce_Deva_Dogr_Gujr_Guru_Khoj_Kthi_Mahj_Modi_Shrd_Sind_Takr_Tirh = std::array<unicode::Script, 12>{
    unicode::Script::Devanagari,
    unicode::Script::Dogra,
    unicode::Script::Gujarati,
    unicode::Script::Gurmukhi,
    unicode::Script::Khojki,
    unicode::Script::Kaithi,
    unicode::Script::Mahajani,
    unicode::Script::Modi,
    unicode::Script::Sharada,
    unicode::Script::Khudawadi,
    unicode::Script::Takri,
    unicode::Script::Tirhuta,
};

auto static const sce_Beng_Deva_Tutg = std::array<unicode::Script, 3>{
    unicode::Script::Bengali,
    unicode::Script::Devanagari,
    unicode::Script::Tulu_Tigalari,
};

auto static const sce_Deva_Taml = std::array<unicode::Script, 2>{
    unicode::Script::Devanagari,
    unicode::Script::Tamil,
};

auto static const sce_Kali_Latn_Mymr = std::array<unicode::Script, 3>{
    unicode::Script::Kayah_Li,
    unicode::Script::Latin,
    unicode::Script::Myanmar,
};

auto static const sce_Bugi_Java = std::array<unicode::Script, 2>{
    unicode::Script::Buginese,
    unicode::Script::Javanese,
};

auto static const sce_Arab_Nkoo = std::array<unicode::Script, 2>{
    unicode::Script::Arabic,
    unicode::Script::Nko,
};

auto static const sce_Arab_Thaa = std::array<unicode::Script, 2>{
    unicode::Script::Arabic,
    unicode::Script::Thaana,
};

auto static const sce_Cpmn_Cprt_Linb = std::array<unicode::Script, 3>{
    unicode::Script::Cypro_Minoan,
    unicode::Script::Cypriot,
    unicode::Script::Linear_B,
};

auto static const sce_Cprt_Linb = std::array<unicode::Script, 2>{
    unicode::Script::Cypriot,
    unicode::Script::Linear_B,
};

auto static const sce_Cprt_Lina_Linb = std::array<unicode::Script, 3>{
    unicode::Script::Cypriot,
    unicode::Script::Linear_A,
    unicode::Script::Linear_B,
};

auto static const sce_Arab_Copt = std::array<unicode::Script, 2>{
    unicode::Script::Arabic,
    unicode::Script::Coptic,
};

auto static const sce_Mani_Ougr = std::array<unicode::Script, 2>{
    unicode::Script::Manichaean,
    unicode::Script::Old_Uyghur,
};

static const std::array<Prop<std::pair<unicode::Script const*, std::size_t>>, 206> sce { {
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x00B7, 0x00B7 }, { sce_Avst_Cari_Copt_Dupl_Elba_Geor_Glag_Gong_Goth_Grek_Hani_Latn_Lydi_Mahj_Perm_Shaw.data(), sce_Avst_Cari_Copt_Dupl_Elba_Geor_Glag_Gong_Goth_Grek_Hani_Latn_Lydi_Mahj_Perm_Shaw.size() } }, // Po MIDDLE DOT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x02BC, 0x02BC }, { sce_Beng_Cyrl_Deva_Latn_Lisu_Thai_Toto.data(), sce_Beng_Cyrl_Deva_Latn_Lisu_Thai_Toto.size() } }, // Lm   MODIFIER LETTER APOSTROPHE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x02C7, 0x02C7 }, { sce_Bopo_Latn.data(), sce_Bopo_Latn.size() } }, // Lm      CARON
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x02C9, 0x02CB }, { sce_Bopo_Latn.data(), sce_Bopo_Latn.size() } }, // Lm  [3] MODIFIER LETTER MACRON..MODIFIER LETTER GRAVE ACCENT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x02CD, 0x02CD }, { sce_Latn_Lisu.data(), sce_Latn_Lisu.size() } }, // Lm      MODIFIER LETTER LOW MACRON
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x02D7, 0x02D7 }, { sce_Latn_Thai.data(), sce_Latn_Thai.size() } }, // Sk      MODIFIER LETTER MINUS SIGN
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x02D9, 0x02D9 }, { sce_Bopo_Latn.data(), sce_Bopo_Latn.size() } }, // Sk      DOT ABOVE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0300, 0x0300 }, { sce_Cher_Copt_Cyrl_Grek_Latn_Perm_Sunu_Tale.data(), sce_Cher_Copt_Cyrl_Grek_Latn_Perm_Sunu_Tale.size() } }, // Mn COMBINING GRAVE ACCENT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0301, 0x0301 }, { sce_Cher_Cyrl_Grek_Latn_Osge_Sunu_Tale_Todr.data(), sce_Cher_Cyrl_Grek_Latn_Osge_Sunu_Tale_Todr.size() } }, // Mn COMBINING ACUTE ACCENT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0302, 0x0302 }, { sce_Cher_Cyrl_Latn_Tfng.data(), sce_Cher_Cyrl_Latn_Tfng.size() } }, // Mn      COMBINING CIRCUMFLEX ACCENT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0303, 0x0303 }, { sce_Glag_Latn_Sunu_Syrc_Thai.data(), sce_Glag_Latn_Sunu_Syrc_Thai.size() } }, // Mn      COMBINING TILDE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0304, 0x0304 }, { sce_Aghb_Cher_Copt_Cyrl_Goth_Grek_Latn_Osge_Syrc_Tfng_Todr.data(), sce_Aghb_Cher_Copt_Cyrl_Goth_Grek_Latn_Osge_Syrc_Tfng_Todr.size() } }, // Mn COMBINING MACRON
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0305, 0x0305 }, { sce_Copt_Elba_Glag_Goth_Kana_Latn.data(), sce_Copt_Elba_Glag_Goth_Kana_Latn.size() } }, // Mn      COMBINING OVERLINE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0306, 0x0306 }, { sce_Cyrl_Grek_Latn_Perm_Tfng.data(), sce_Cyrl_Grek_Latn_Perm_Tfng.size() } }, // Mn      COMBINING BREVE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0307, 0x0307 }, { sce_Copt_Dupl_Hebr_Latn_Perm_Syrc_Tale_Tfng_Todr.data(), sce_Copt_Dupl_Hebr_Latn_Perm_Syrc_Tale_Tfng_Todr.size() } }, // Mn COMBINING DOT ABOVE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0308, 0x0308 }, { sce_Armn_Cyrl_Dupl_Goth_Grek_Hebr_Latn_Perm_Syrc_Tale_Tfng.data(), sce_Armn_Cyrl_Dupl_Goth_Grek_Hebr_Latn_Perm_Syrc_Tale_Tfng.size() } }, // Mn COMBINING DIAERESIS
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0309, 0x0309 }, { sce_Latn_Tfng.data(), sce_Latn_Tfng.size() } }, // Mn      COMBINING HOOK ABOVE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x030A, 0x030A }, { sce_Dupl_Latn_Syrc.data(), sce_Dupl_Latn_Syrc.size() } }, // Mn      COMBINING RING ABOVE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x030B, 0x030B }, { sce_Cher_Cyrl_Latn_Osge.data(), sce_Cher_Cyrl_Latn_Osge.size() } }, // Mn      COMBINING DOUBLE ACUTE ACCENT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x030C, 0x030C }, { sce_Cher_Latn_Tale.data(), sce_Cher_Latn_Tale.size() } }, // Mn      COMBINING CARON
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x030D, 0x030D }, { sce_Latn_Sunu.data(), sce_Latn_Sunu.size() } }, // Mn      COMBINING VERTICAL LINE ABOVE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x030E, 0x030E }, { sce_Ethi_Latn.data(), sce_Ethi_Latn.size() } }, // Mn      COMBINING DOUBLE VERTICAL LINE ABOVE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0310, 0x0310 }, { sce_Latn_Sunu.data(), sce_Latn_Sunu.size() } }, // Mn      COMBINING CANDRABINDU
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0311, 0x0311 }, { sce_Cyrl_Latn_Todr.data(), sce_Cyrl_Latn_Todr.size() } }, // Mn      COMBINING INVERTED BREVE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0313, 0x0313 }, { sce_Grek_Latn_Perm_Todr.data(), sce_Grek_Latn_Perm_Todr.size() } }, // Mn      COMBINING COMMA ABOVE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0323, 0x0323 }, { sce_Cher_Dupl_Kana_Latn_Syrc_Tfng.data(), sce_Cher_Dupl_Kana_Latn_Syrc_Tfng.size() } }, // Mn      COMBINING DOT BELOW
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0324, 0x0324 }, { sce_Cher_Dupl_Latn_Syrc.data(), sce_Cher_Dupl_Latn_Syrc.size() } }, // Mn      COMBINING DIAERESIS BELOW
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0325, 0x0325 }, { sce_Latn_Syrc.data(), sce_Latn_Syrc.size() } }, // Mn      COMBINING RING BELOW
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x032D, 0x032D }, { sce_Latn_Sunu_Syrc.data(), sce_Latn_Sunu_Syrc.size() } }, // Mn      COMBINING CIRCUMFLEX ACCENT BELOW
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x032E, 0x032E }, { sce_Latn_Syrc.data(), sce_Latn_Syrc.size() } }, // Mn      COMBINING BREVE BELOW
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0330, 0x0330 }, { sce_Cher_Latn_Syrc.data(), sce_Cher_Latn_Syrc.size() } }, // Mn      COMBINING TILDE BELOW
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0331, 0x0331 }, { sce_Aghb_Cher_Goth_Latn_Sunu_Syrc_Thai.data(), sce_Aghb_Cher_Goth_Latn_Sunu_Syrc_Thai.size() } }, // Mn   COMBINING MACRON BELOW
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0342, 0x0342 }, { sce_Grek.data(), sce_Grek.size() } }, // Mn      COMBINING GREEK PERISPOMENI
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0345, 0x0345 }, { sce_Grek.data(), sce_Grek.size() } }, // Mn      COMBINING GREEK YPOGEGRAMMENI
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0358, 0x0358 }, { sce_Latn_Osge.data(), sce_Latn_Osge.size() } }, // Mn      COMBINING DOT ABOVE RIGHT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x035E, 0x035E }, { sce_Aghb_Latn_Todr.data(), sce_Aghb_Latn_Todr.size() } }, // Mn      COMBINING DOUBLE MACRON
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0363, 0x036F }, { sce_Latn.data(), sce_Latn.size() } }, // Mn [13] COMBINING LATIN SMALL LETTER A..COMBINING LATIN SMALL LETTER X
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0374, 0x0374 }, { sce_Copt_Grek.data(), sce_Copt_Grek.size() } }, // Lm      GREEK NUMERAL SIGN
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0375, 0x0375 }, { sce_Copt_Grek.data(), sce_Copt_Grek.size() } }, // Sk      GREEK LOWER NUMERAL SIGN
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0483, 0x0483 }, { sce_Cyrl_Perm.data(), sce_Cyrl_Perm.size() } }, // Mn      COMBINING CYRILLIC TITLO
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0484, 0x0484 }, { sce_Cyrl_Glag.data(), sce_Cyrl_Glag.size() } }, // Mn      COMBINING CYRILLIC PALATALIZATION
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0485, 0x0486 }, { sce_Cyrl_Latn.data(), sce_Cyrl_Latn.size() } }, // Mn  [2] COMBINING CYRILLIC DASIA PNEUMATA..COMBINING CYRILLIC PSILI PNEUMATA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0487, 0x0487 }, { sce_Cyrl_Glag.data(), sce_Cyrl_Glag.size() } }, // Mn      COMBINING CYRILLIC POKRYTIE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0589, 0x0589 }, { sce_Armn_Geor_Glag.data(), sce_Armn_Geor_Glag.size() } }, // Po      ARMENIAN FULL STOP
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x060C, 0x060C }, { sce_Arab_Gara_Nkoo_Rohg_Syrc_Thaa_Yezi.data(), sce_Arab_Gara_Nkoo_Rohg_Syrc_Thaa_Yezi.size() } }, // Po   ARABIC COMMA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x061B, 0x061B }, { sce_Arab_Gara_Nkoo_Rohg_Syrc_Thaa_Yezi.data(), sce_Arab_Gara_Nkoo_Rohg_Syrc_Thaa_Yezi.size() } }, // Po   ARABIC SEMICOLON
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x061C, 0x061C }, { sce_Arab_Syrc_Thaa.data(), sce_Arab_Syrc_Thaa.size() } }, // Cf      ARABIC LETTER MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x061F, 0x061F }, { sce_Adlm_Arab_Gara_Nkoo_Rohg_Syrc_Thaa_Yezi.data(), sce_Adlm_Arab_Gara_Nkoo_Rohg_Syrc_Thaa_Yezi.size() } }, // Po ARABIC QUESTION MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0640, 0x0640 }, { sce_Adlm_Arab_Mand_Mani_Ougr_Phlp_Rohg_Sogd_Syrc.data(), sce_Adlm_Arab_Mand_Mani_Ougr_Phlp_Rohg_Sogd_Syrc.size() } }, // Lm ARABIC TATWEEL
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x064B, 0x0655 }, { sce_Arab_Syrc.data(), sce_Arab_Syrc.size() } }, // Mn [11] ARABIC FATHATAN..ARABIC HAMZA BELOW
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0660, 0x0669 }, { sce_Arab_Thaa_Yezi.data(), sce_Arab_Thaa_Yezi.size() } }, // Nd [10] ARABIC-INDIC DIGIT ZERO..ARABIC-INDIC DIGIT NINE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0670, 0x0670 }, { sce_Arab_Syrc.data(), sce_Arab_Syrc.size() } }, // Mn      ARABIC LETTER SUPERSCRIPT ALEF
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x06D4, 0x06D4 }, { sce_Arab_Rohg.data(), sce_Arab_Rohg.size() } }, // Po      ARABIC FULL STOP
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0951, 0x0951 }, { sce_Beng_Deva_Gran_Gujr_Guru_Knda_Latn_Mlym_Nand_Newa_Orya_Shrd_Taml_Telu_Tirh.data(), sce_Beng_Deva_Gran_Gujr_Guru_Knda_Latn_Mlym_Nand_Newa_Orya_Shrd_Taml_Telu_Tirh.size() } }, // Mn DEVANAGARI STRESS SIGN UDATTA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0952, 0x0952 }, { sce_Beng_Deva_Gran_Gujr_Guru_Knda_Latn_Mlym_Newa_Orya_Taml_Telu_Tirh.data(), sce_Beng_Deva_Gran_Gujr_Guru_Knda_Latn_Mlym_Newa_Orya_Taml_Telu_Tirh.size() } }, // Mn DEVANAGARI STRESS SIGN ANUDATTA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0964, 0x0964 }, { sce_Beng_Deva_Dogr_Gong_Gonm_Gran_Gujr_Guru_Knda_Mahj_Mlym_Nand_Onao_Orya_Sind_Sinh_Sylo_Takr_Taml_Telu_Tirh.data(), sce_Beng_Deva_Dogr_Gong_Gonm_Gran_Gujr_Guru_Knda_Mahj_Mlym_Nand_Onao_Orya_Sind_Sinh_Sylo_Takr_Taml_Telu_Tirh.size() } }, // Po DEVANAGARI DANDA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0965, 0x0965 }, { sce_Beng_Deva_Dogr_Gong_Gonm_Gran_Gujr_Gukh_Guru_Knda_Limb_Mahj_Mlym_Nand_Onao_Orya_Sind_Sinh_Sylo_Takr_Taml_Telu_Tirh.data(), sce_Beng_Deva_Dogr_Gong_Gonm_Gran_Gujr_Gukh_Guru_Knda_Limb_Mahj_Mlym_Nand_Onao_Orya_Sind_Sinh_Sylo_Takr_Taml_Telu_Tirh.size() } }, // Po DEVANAGARI DOUBLE DANDA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0966, 0x096F }, { sce_Deva_Dogr_Kthi_Mahj.data(), sce_Deva_Dogr_Kthi_Mahj.size() } }, // Nd [10] DEVANAGARI DIGIT ZERO..DEVANAGARI DIGIT NINE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x09E6, 0x09EF }, { sce_Beng_Cakm_Sylo.data(), sce_Beng_Cakm_Sylo.size() } }, // Nd [10] BENGALI DIGIT ZERO..BENGALI DIGIT NINE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0A66, 0x0A6F }, { sce_Guru_Mult.data(), sce_Guru_Mult.size() } }, // Nd [10] GURMUKHI DIGIT ZERO..GURMUKHI DIGIT NINE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0AE6, 0x0AEF }, { sce_Gujr_Khoj.data(), sce_Gujr_Khoj.size() } }, // Nd [10] GUJARATI DIGIT ZERO..GUJARATI DIGIT NINE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0BE6, 0x0BEF }, { sce_Gran_Taml.data(), sce_Gran_Taml.size() } }, // Nd [10] TAMIL DIGIT ZERO..TAMIL DIGIT NINE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0BF0, 0x0BF2 }, { sce_Gran_Taml.data(), sce_Gran_Taml.size() } }, // No  [3] TAMIL NUMBER TEN..TAMIL NUMBER ONE THOUSAND
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0BF3, 0x0BF3 }, { sce_Gran_Taml.data(), sce_Gran_Taml.size() } }, // So      TAMIL DAY SIGN
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x0CE6, 0x0CEF }, { sce_Knda_Nand_Tutg.data(), sce_Knda_Nand_Tutg.size() } }, // Nd [10] KANNADA DIGIT ZERO..KANNADA DIGIT NINE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1040, 0x1049 }, { sce_Cakm_Mymr_Tale.data(), sce_Cakm_Mymr_Tale.size() } }, // Nd [10] MYANMAR DIGIT ZERO..MYANMAR DIGIT NINE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x10FB, 0x10FB }, { sce_Geor_Glag_Latn.data(), sce_Geor_Glag_Latn.size() } }, // Po      GEORGIAN PARAGRAPH SEPARATOR
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x16EB, 0x16ED }, { sce_Runr.data(), sce_Runr.size() } }, // Po  [3] RUNIC SINGLE PUNCTUATION..RUNIC CROSS PUNCTUATION
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1735, 0x1736 }, { sce_Buhd_Hano_Tagb_Tglg.data(), sce_Buhd_Hano_Tagb_Tglg.size() } }, // Po  [2] PHILIPPINE SINGLE PUNCTUATION..PHILIPPINE DOUBLE PUNCTUATION
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1802, 0x1803 }, { sce_Mong_Phag.data(), sce_Mong_Phag.size() } }, // Po  [2] MONGOLIAN COMMA..MONGOLIAN FULL STOP
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1805, 0x1805 }, { sce_Mong_Phag.data(), sce_Mong_Phag.size() } }, // Po      MONGOLIAN FOUR DOTS
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CD0, 0x1CD0 }, { sce_Beng_Deva_Gran_Knda.data(), sce_Beng_Deva_Gran_Knda.size() } }, // Mn      VEDIC TONE KARSHANA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CD1, 0x1CD1 }, { sce_Deva.data(), sce_Deva.size() } }, // Mn      VEDIC TONE SHARA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CD2, 0x1CD2 }, { sce_Beng_Deva_Gran_Knda.data(), sce_Beng_Deva_Gran_Knda.size() } }, // Mn      VEDIC TONE PRENKHA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CD3, 0x1CD3 }, { sce_Deva_Gran_Knda.data(), sce_Deva_Gran_Knda.size() } }, // Po      VEDIC SIGN NIHSHVASA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CD4, 0x1CD4 }, { sce_Deva.data(), sce_Deva.size() } }, // Mn      VEDIC SIGN YAJURVEDIC MIDLINE SVARITA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CD5, 0x1CD5 }, { sce_Beng_Deva_Newa_Telu_Tirh.data(), sce_Beng_Deva_Newa_Telu_Tirh.size() } }, // Mn      VEDIC TONE YAJURVEDIC AGGRAVATED INDEPENDENT SVARITA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CD6, 0x1CD6 }, { sce_Beng_Deva_Telu.data(), sce_Beng_Deva_Telu.size() } }, // Mn      VEDIC TONE YAJURVEDIC INDEPENDENT SVARITA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CD7, 0x1CD7 }, { sce_Deva_Newa_Shrd.data(), sce_Deva_Newa_Shrd.size() } }, // Mn      VEDIC TONE YAJURVEDIC KATHAKA INDEPENDENT SVARITA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CD8, 0x1CD8 }, { sce_Beng_Deva_Newa_Telu.data(), sce_Beng_Deva_Newa_Telu.size() } }, // Mn      VEDIC TONE CANDRA BELOW
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CD9, 0x1CD9 }, { sce_Deva_Shrd.data(), sce_Deva_Shrd.size() } }, // Mn      VEDIC TONE YAJURVEDIC KATHAKA INDEPENDENT SVARITA SCHROEDER
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CDA, 0x1CDA }, { sce_Deva_Knda_Mlym_Orya_Taml_Telu.data(), sce_Deva_Knda_Mlym_Orya_Taml_Telu.size() } }, // Mn      VEDIC TONE DOUBLE SVARITA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CDB, 0x1CDB }, { sce_Deva.data(), sce_Deva.size() } }, // Mn      VEDIC TONE TRIPLE SVARITA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CDC, 0x1CDD }, { sce_Deva_Shrd.data(), sce_Deva_Shrd.size() } }, // Mn  [2] VEDIC TONE KATHAKA ANUDATTA..VEDIC TONE DOT BELOW
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CDE, 0x1CDF }, { sce_Deva.data(), sce_Deva.size() } }, // Mn  [2] VEDIC TONE TWO DOTS BELOW..VEDIC TONE THREE DOTS BELOW
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CE0, 0x1CE0 }, { sce_Deva_Shrd.data(), sce_Deva_Shrd.size() } }, // Mn      VEDIC TONE RIGVEDIC KASHMIRI INDEPENDENT SVARITA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CE1, 0x1CE1 }, { sce_Beng_Deva.data(), sce_Beng_Deva.size() } }, // Mc      VEDIC TONE ATHARVAVEDIC INDEPENDENT SVARITA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CE2, 0x1CE2 }, { sce_Deva_Newa_Tirh.data(), sce_Deva_Newa_Tirh.size() } }, // Mn      VEDIC SIGN VISARGA SVARITA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CE3, 0x1CE8 }, { sce_Deva.data(), sce_Deva.size() } }, // Mn  [6] VEDIC SIGN VISARGA UDATTA..VEDIC SIGN VISARGA ANUDATTA WITH TAIL
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CE9, 0x1CE9 }, { sce_Deva_Nand_Newa.data(), sce_Deva_Nand_Newa.size() } }, // Lo      VEDIC SIGN ANUSVARA ANTARGOMUKHA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CEA, 0x1CEA }, { sce_Beng_Deva_Shrd.data(), sce_Beng_Deva_Shrd.size() } }, // Lo      VEDIC SIGN ANUSVARA BAHIRGOMUKHA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CEB, 0x1CEB }, { sce_Deva_Newa.data(), sce_Deva_Newa.size() } }, // Lo      VEDIC SIGN ANUSVARA VAMAGOMUKHA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CEC, 0x1CEC }, { sce_Deva.data(), sce_Deva.size() } }, // Lo      VEDIC SIGN ANUSVARA VAMAGOMUKHA WITH TAIL
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CED, 0x1CED }, { sce_Beng_Deva_Newa_Shrd.data(), sce_Beng_Deva_Newa_Shrd.size() } }, // Mn      VEDIC SIGN TIRYAK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CEE, 0x1CF1 }, { sce_Deva.data(), sce_Deva.size() } }, // Lo  [4] VEDIC SIGN HEXIFORM LONG ANUSVARA..VEDIC SIGN ANUSVARA UBHAYATO MUKHA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CF2, 0x1CF2 }, { sce_Beng_Deva_Gran_Knda_Mlym_Nand_Orya_Sinh_Telu_Tirh_Tutg.data(), sce_Beng_Deva_Gran_Knda_Mlym_Nand_Orya_Sinh_Telu_Tirh_Tutg.size() } }, // Lo VEDIC SIGN ARDHAVISARGA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CF3, 0x1CF3 }, { sce_Deva_Gran.data(), sce_Deva_Gran.size() } }, // Lo      VEDIC SIGN ROTATED ARDHAVISARGA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CF4, 0x1CF4 }, { sce_Deva_Gran_Knda_Tutg.data(), sce_Deva_Gran_Knda_Tutg.size() } }, // Mn      VEDIC TONE CANDRA ABOVE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CF5, 0x1CF6 }, { sce_Beng_Deva.data(), sce_Beng_Deva.size() } }, // Lo  [2] VEDIC SIGN JIHVAMULIYA..VEDIC SIGN UPADHMANIYA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CF7, 0x1CF7 }, { sce_Beng.data(), sce_Beng.size() } }, // Mc      VEDIC SIGN ATIKRAMA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CF8, 0x1CF9 }, { sce_Deva_Gran.data(), sce_Deva_Gran.size() } }, // Mn  [2] VEDIC TONE RING ABOVE..VEDIC TONE DOUBLE RING ABOVE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1CFA, 0x1CFA }, { sce_Nand.data(), sce_Nand.size() } }, // Lo      VEDIC SIGN DOUBLE ANUSVARA ANTARGOMUKHA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1DC0, 0x1DC1 }, { sce_Grek.data(), sce_Grek.size() } }, // Mn  [2] COMBINING DOTTED GRAVE ACCENT..COMBINING DOTTED ACUTE ACCENT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1DF8, 0x1DF8 }, { sce_Cyrl_Latn_Syrc.data(), sce_Cyrl_Latn_Syrc.size() } }, // Mn      COMBINING DOT ABOVE LEFT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1DFA, 0x1DFA }, { sce_Syrc.data(), sce_Syrc.size() } }, // Mn      COMBINING DOT BELOW LEFT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x202F, 0x202F }, { sce_Latn_Mong_Phag.data(), sce_Latn_Mong_Phag.size() } }, // Zs      NARROW NO-BREAK SPACE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x204F, 0x204F }, { sce_Adlm_Arab.data(), sce_Adlm_Arab.size() } }, // Po      REVERSED SEMICOLON
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x205A, 0x205A }, { sce_Cari_Geor_Glag_Hung_Lyci_Orkh.data(), sce_Cari_Geor_Glag_Hung_Lyci_Orkh.size() } }, // Po      TWO DOT PUNCTUATION
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x205D, 0x205D }, { sce_Cari_Grek_Hung_Mero.data(), sce_Cari_Grek_Hung_Mero.size() } }, // Po      TRICOLON
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x20F0, 0x20F0 }, { sce_Deva_Gran_Latn.data(), sce_Deva_Gran_Latn.size() } }, // Mn      COMBINING ASTERISK ABOVE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x2E17, 0x2E17 }, { sce_Copt_Latn.data(), sce_Copt_Latn.size() } }, // Pd      DOUBLE OBLIQUE HYPHEN
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x2E30, 0x2E30 }, { sce_Avst_Orkh.data(), sce_Avst_Orkh.size() } }, // Po      RING POINT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x2E31, 0x2E31 }, { sce_Avst_Cari_Geor_Hung_Kthi_Lydi_Samr.data(), sce_Avst_Cari_Geor_Hung_Kthi_Lydi_Samr.size() } }, // Po   WORD SEPARATOR MIDDLE DOT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x2E3C, 0x2E3C }, { sce_Dupl.data(), sce_Dupl.size() } }, // Po      STENOGRAPHIC FULL STOP
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x2E41, 0x2E41 }, { sce_Adlm_Arab_Hung.data(), sce_Adlm_Arab_Hung.size() } }, // Po      REVERSED COMMA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x2E43, 0x2E43 }, { sce_Cyrl_Glag.data(), sce_Cyrl_Glag.size() } }, // Po      DASH WITH LEFT UPTURN
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x2FF0, 0x2FFF }, { sce_Hani_Tang.data(), sce_Hani_Tang.size() } }, // So [16] IDEOGRAPHIC DESCRIPTION CHARACTER LEFT TO RIGHT..IDEOGRAPHIC DESCRIPTION CHARACTER ROTATION
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3001, 0x3001 }, { sce_Bopo_Hang_Hani_Hira_Kana_Mong_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Mong_Yiii.size() } }, // Po   IDEOGRAPHIC COMMA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3002, 0x3002 }, { sce_Bopo_Hang_Hani_Hira_Kana_Mong_Phag_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Mong_Phag_Yiii.size() } }, // Po IDEOGRAPHIC FULL STOP
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3003, 0x3003 }, { sce_Bopo_Hang_Hani_Hira_Kana.data(), sce_Bopo_Hang_Hani_Hira_Kana.size() } }, // Po      DITTO MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3006, 0x3006 }, { sce_Hani.data(), sce_Hani.size() } }, // Lo      IDEOGRAPHIC CLOSING MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3008, 0x3008 }, { sce_Bopo_Hang_Hani_Hira_Kana_Mong_Tibt_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Mong_Tibt_Yiii.size() } }, // Ps LEFT ANGLE BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3009, 0x3009 }, { sce_Bopo_Hang_Hani_Hira_Kana_Mong_Tibt_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Mong_Tibt_Yiii.size() } }, // Pe RIGHT ANGLE BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x300A, 0x300A }, { sce_Bopo_Hang_Hani_Hira_Kana_Lisu_Mong_Tibt_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Lisu_Mong_Tibt_Yiii.size() } }, // Ps LEFT DOUBLE ANGLE BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x300B, 0x300B }, { sce_Bopo_Hang_Hani_Hira_Kana_Lisu_Mong_Tibt_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Lisu_Mong_Tibt_Yiii.size() } }, // Pe RIGHT DOUBLE ANGLE BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x300C, 0x300C }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Ps      LEFT CORNER BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x300D, 0x300D }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Pe      RIGHT CORNER BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x300E, 0x300E }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Ps      LEFT WHITE CORNER BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x300F, 0x300F }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Pe      RIGHT WHITE CORNER BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3010, 0x3010 }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Ps      LEFT BLACK LENTICULAR BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3011, 0x3011 }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Pe      RIGHT BLACK LENTICULAR BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3013, 0x3013 }, { sce_Bopo_Hang_Hani_Hira_Kana.data(), sce_Bopo_Hang_Hani_Hira_Kana.size() } }, // So      GETA MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3014, 0x3014 }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Ps      LEFT TORTOISE SHELL BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3015, 0x3015 }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Pe      RIGHT TORTOISE SHELL BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3016, 0x3016 }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Ps      LEFT WHITE LENTICULAR BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3017, 0x3017 }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Pe      RIGHT WHITE LENTICULAR BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3018, 0x3018 }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Ps      LEFT WHITE TORTOISE SHELL BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3019, 0x3019 }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Pe      RIGHT WHITE TORTOISE SHELL BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x301A, 0x301A }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Ps      LEFT WHITE SQUARE BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x301B, 0x301B }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Pe      RIGHT WHITE SQUARE BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x301C, 0x301C }, { sce_Bopo_Hang_Hani_Hira_Kana.data(), sce_Bopo_Hang_Hani_Hira_Kana.size() } }, // Pd      WAVE DASH
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x301D, 0x301D }, { sce_Bopo_Hang_Hani_Hira_Kana.data(), sce_Bopo_Hang_Hani_Hira_Kana.size() } }, // Ps      REVERSED DOUBLE PRIME QUOTATION MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x301E, 0x301F }, { sce_Bopo_Hang_Hani_Hira_Kana.data(), sce_Bopo_Hang_Hani_Hira_Kana.size() } }, // Pe  [2] DOUBLE PRIME QUOTATION MARK..LOW DOUBLE PRIME QUOTATION MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x302A, 0x302D }, { sce_Bopo_Hani.data(), sce_Bopo_Hani.size() } }, // Mn  [4] IDEOGRAPHIC LEVEL TONE MARK..IDEOGRAPHIC ENTERING TONE MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3030, 0x3030 }, { sce_Bopo_Hang_Hani_Hira_Kana.data(), sce_Bopo_Hang_Hani_Hira_Kana.size() } }, // Pd      WAVY DASH
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3031, 0x3035 }, { sce_Hira_Kana.data(), sce_Hira_Kana.size() } }, // Lm  [5] VERTICAL KANA REPEAT MARK..VERTICAL KANA REPEAT MARK LOWER HALF
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3037, 0x3037 }, { sce_Bopo_Hang_Hani_Hira_Kana.data(), sce_Bopo_Hang_Hani_Hira_Kana.size() } }, // So      IDEOGRAPHIC TELEGRAPH LINE FEED SEPARATOR SYMBOL
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x303C, 0x303C }, { sce_Hani_Hira_Kana.data(), sce_Hani_Hira_Kana.size() } }, // Lo      MASU MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x303D, 0x303D }, { sce_Hani_Hira_Kana.data(), sce_Hani_Hira_Kana.size() } }, // Po      PART ALTERNATION MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x303E, 0x303F }, { sce_Hani.data(), sce_Hani.size() } }, // So  [2] IDEOGRAPHIC VARIATION INDICATOR..IDEOGRAPHIC HALF FILL SPACE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3099, 0x309A }, { sce_Hira_Kana.data(), sce_Hira_Kana.size() } }, // Mn  [2] COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK..COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x309B, 0x309C }, { sce_Hira_Kana.data(), sce_Hira_Kana.size() } }, // Sk  [2] KATAKANA-HIRAGANA VOICED SOUND MARK..KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x30A0, 0x30A0 }, { sce_Hira_Kana.data(), sce_Hira_Kana.size() } }, // Pd      KATAKANA-HIRAGANA DOUBLE HYPHEN
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x30FB, 0x30FB }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Po      KATAKANA MIDDLE DOT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x30FC, 0x30FC }, { sce_Hira_Kana.data(), sce_Hira_Kana.size() } }, // Lm      KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3190, 0x3191 }, { sce_Hani.data(), sce_Hani.size() } }, // So  [2] IDEOGRAPHIC ANNOTATION LINKING MARK..IDEOGRAPHIC ANNOTATION REVERSE MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3192, 0x3195 }, { sce_Hani.data(), sce_Hani.size() } }, // No  [4] IDEOGRAPHIC ANNOTATION ONE MARK..IDEOGRAPHIC ANNOTATION FOUR MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3196, 0x319F }, { sce_Hani.data(), sce_Hani.size() } }, // So [10] IDEOGRAPHIC ANNOTATION TOP MARK..IDEOGRAPHIC ANNOTATION MAN MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x31C0, 0x31E5 }, { sce_Hani.data(), sce_Hani.size() } }, // So [38] CJK STROKE T..CJK STROKE SZP
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x31EF, 0x31EF }, { sce_Hani_Tang.data(), sce_Hani_Tang.size() } }, // So      IDEOGRAPHIC DESCRIPTION CHARACTER SUBTRACTION
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3220, 0x3229 }, { sce_Hani.data(), sce_Hani.size() } }, // No [10] PARENTHESIZED IDEOGRAPH ONE..PARENTHESIZED IDEOGRAPH TEN
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x322A, 0x3247 }, { sce_Hani.data(), sce_Hani.size() } }, // So [30] PARENTHESIZED IDEOGRAPH MOON..CIRCLED IDEOGRAPH KOTO
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3280, 0x3289 }, { sce_Hani.data(), sce_Hani.size() } }, // No [10] CIRCLED IDEOGRAPH ONE..CIRCLED IDEOGRAPH TEN
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x328A, 0x32B0 }, { sce_Hani.data(), sce_Hani.size() } }, // So [39] CIRCLED IDEOGRAPH MOON..CIRCLED IDEOGRAPH NIGHT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x32C0, 0x32CB }, { sce_Hani.data(), sce_Hani.size() } }, // So [12] IDEOGRAPHIC TELEGRAPH SYMBOL FOR JANUARY..IDEOGRAPHIC TELEGRAPH SYMBOL FOR DECEMBER
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x32FF, 0x32FF }, { sce_Hani.data(), sce_Hani.size() } }, // So      SQUARE ERA NAME REIWA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x3358, 0x3370 }, { sce_Hani.data(), sce_Hani.size() } }, // So [25] IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR ZERO..IDEOGRAPHIC TELEGRAPH SYMBOL FOR HOUR TWENTY-FOUR
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x337B, 0x337F }, { sce_Hani.data(), sce_Hani.size() } }, // So  [5] SQUARE ERA NAME HEISEI..SQUARE CORPORATION
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x33E0, 0x33FE }, { sce_Hani.data(), sce_Hani.size() } }, // So [31] IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY ONE..IDEOGRAPHIC TELEGRAPH SYMBOL FOR DAY THIRTY-ONE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xA66F, 0xA66F }, { sce_Cyrl_Glag.data(), sce_Cyrl_Glag.size() } }, // Mn      COMBINING CYRILLIC VZMET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xA700, 0xA707 }, { sce_Hani_Latn.data(), sce_Hani_Latn.size() } }, // Sk  [8] MODIFIER LETTER CHINESE TONE YIN PING..MODIFIER LETTER CHINESE TONE YANG RU
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xA830, 0xA832 }, { sce_Deva_Dogr_Gujr_Guru_Khoj_Knda_Kthi_Mahj_Mlym_Modi_Nand_Shrd_Sind_Takr_Tirh_Tutg.data(), sce_Deva_Dogr_Gujr_Guru_Khoj_Knda_Kthi_Mahj_Mlym_Modi_Nand_Shrd_Sind_Takr_Tirh_Tutg.size() } }, // No [3] NORTH INDIC FRACTION ONE QUARTER..NORTH INDIC FRACTION THREE QUARTERS
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xA833, 0xA835 }, { sce_Deva_Dogr_Gujr_Guru_Khoj_Knda_Kthi_Mahj_Modi_Nand_Shrd_Sind_Takr_Tirh_Tutg.data(), sce_Deva_Dogr_Gujr_Guru_Khoj_Knda_Kthi_Mahj_Modi_Nand_Shrd_Sind_Takr_Tirh_Tutg.size() } }, // No [3] NORTH INDIC FRACTION ONE SIXTEENTH..NORTH INDIC FRACTION THREE SIXTEENTHS
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xA836, 0xA837 }, { sce_Deva_Dogr_Gujr_Guru_Khoj_Kthi_Mahj_Modi_Sind_Takr_Tirh.data(), sce_Deva_Dogr_Gujr_Guru_Khoj_Kthi_Mahj_Modi_Sind_Takr_Tirh.size() } }, // So [2] NORTH INDIC QUARTER MARK..NORTH INDIC PLACEHOLDER MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xA838, 0xA838 }, { sce_Deva_Dogr_Gujr_Guru_Khoj_Kthi_Mahj_Modi_Shrd_Sind_Takr_Tirh.data(), sce_Deva_Dogr_Gujr_Guru_Khoj_Kthi_Mahj_Modi_Shrd_Sind_Takr_Tirh.size() } }, // Sc NORTH INDIC RUPEE MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xA839, 0xA839 }, { sce_Deva_Dogr_Gujr_Guru_Khoj_Kthi_Mahj_Modi_Sind_Takr_Tirh.data(), sce_Deva_Dogr_Gujr_Guru_Khoj_Kthi_Mahj_Modi_Sind_Takr_Tirh.size() } }, // So NORTH INDIC QUANTITY MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xA8F1, 0xA8F1 }, { sce_Beng_Deva_Tutg.data(), sce_Beng_Deva_Tutg.size() } }, // Mn      COMBINING DEVANAGARI SIGN AVAGRAHA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xA8F3, 0xA8F3 }, { sce_Deva_Taml.data(), sce_Deva_Taml.size() } }, // Lo      DEVANAGARI SIGN CANDRABINDU VIRAMA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xA92E, 0xA92E }, { sce_Kali_Latn_Mymr.data(), sce_Kali_Latn_Mymr.size() } }, // Po      KAYAH LI SIGN CWI
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xA9CF, 0xA9CF }, { sce_Bugi_Java.data(), sce_Bugi_Java.size() } }, // Lm      JAVANESE PANGRANGKEP
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xFD3E, 0xFD3E }, { sce_Arab_Nkoo.data(), sce_Arab_Nkoo.size() } }, // Pe      ORNATE LEFT PARENTHESIS
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xFD3F, 0xFD3F }, { sce_Arab_Nkoo.data(), sce_Arab_Nkoo.size() } }, // Ps      ORNATE RIGHT PARENTHESIS
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xFDF2, 0xFDF2 }, { sce_Arab_Thaa.data(), sce_Arab_Thaa.size() } }, // Lo      ARABIC LIGATURE ALLAH ISOLATED FORM
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xFDFD, 0xFDFD }, { sce_Arab_Thaa.data(), sce_Arab_Thaa.size() } }, // So      ARABIC LIGATURE BISMILLAH AR-RAHMAN AR-RAHEEM
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xFE45, 0xFE46 }, { sce_Bopo_Hang_Hani_Hira_Kana.data(), sce_Bopo_Hang_Hani_Hira_Kana.size() } }, // Po  [2] SESAME DOT..WHITE SESAME DOT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xFF61, 0xFF61 }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Po      HALFWIDTH IDEOGRAPHIC FULL STOP
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xFF62, 0xFF62 }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Ps      HALFWIDTH LEFT CORNER BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xFF63, 0xFF63 }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Pe      HALFWIDTH RIGHT CORNER BRACKET
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xFF64, 0xFF65 }, { sce_Bopo_Hang_Hani_Hira_Kana_Yiii.data(), sce_Bopo_Hang_Hani_Hira_Kana_Yiii.size() } }, // Po  [2] HALFWIDTH IDEOGRAPHIC COMMA..HALFWIDTH KATAKANA MIDDLE DOT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xFF70, 0xFF70 }, { sce_Hira_Kana.data(), sce_Hira_Kana.size() } }, // Lm      HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0xFF9E, 0xFF9F }, { sce_Hira_Kana.data(), sce_Hira_Kana.size() } }, // Lm  [2] HALFWIDTH KATAKANA VOICED SOUND MARK..HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x10100, 0x10101 }, { sce_Cpmn_Cprt_Linb.data(), sce_Cpmn_Cprt_Linb.size() } }, // Po  [2] AEGEAN WORD SEPARATOR LINE..AEGEAN WORD SEPARATOR DOT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x10102, 0x10102 }, { sce_Cprt_Linb.data(), sce_Cprt_Linb.size() } }, // Po      AEGEAN CHECK MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x10107, 0x10133 }, { sce_Cprt_Lina_Linb.data(), sce_Cprt_Lina_Linb.size() } }, // No [45] AEGEAN NUMBER ONE..AEGEAN NUMBER NINETY THOUSAND
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x10137, 0x1013F }, { sce_Cprt_Linb.data(), sce_Cprt_Linb.size() } }, // So  [9] AEGEAN WEIGHT BASE UNIT..AEGEAN MEASURE THIRD SUBUNIT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x102E0, 0x102E0 }, { sce_Arab_Copt.data(), sce_Arab_Copt.size() } }, // Mn      COPTIC EPACT THOUSANDS MARK
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x102E1, 0x102FB }, { sce_Arab_Copt.data(), sce_Arab_Copt.size() } }, // No [27] COPTIC EPACT DIGIT ONE..COPTIC EPACT NUMBER NINE HUNDRED
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x10AF2, 0x10AF2 }, { sce_Mani_Ougr.data(), sce_Mani_Ougr.size() } }, // Po      MANICHAEAN PUNCTUATION DOUBLE DOT WITHIN DOT
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x11301, 0x11301 }, { sce_Gran_Taml.data(), sce_Gran_Taml.size() } }, // Mn      GRANTHA SIGN CANDRABINDU
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x11303, 0x11303 }, { sce_Gran_Taml.data(), sce_Gran_Taml.size() } }, // Mc      GRANTHA SIGN VISARGA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1133B, 0x1133C }, { sce_Gran_Taml.data(), sce_Gran_Taml.size() } }, // Mn  [2] COMBINING BINDU BELOW..GRANTHA SIGN NUKTA
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x11FD0, 0x11FD1 }, { sce_Gran_Taml.data(), sce_Gran_Taml.size() } }, // No  [2] TAMIL FRACTION ONE QUARTER..TAMIL FRACTION ONE HALF-1
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x11FD3, 0x11FD3 }, { sce_Gran_Taml.data(), sce_Gran_Taml.size() } }, // No      TAMIL FRACTION THREE QUARTERS
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1BCA0, 0x1BCA3 }, { sce_Dupl.data(), sce_Dupl.size() } }, // Cf  [4] SHORTHAND FORMAT LETTER OVERLAP..SHORTHAND FORMAT UP STEP
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1D360, 0x1D371 }, { sce_Hani.data(), sce_Hani.size() } }, // No [18] COUNTING ROD UNIT DIGIT ONE..COUNTING ROD TENS DIGIT NINE
    Prop<std::pair<unicode::Script const*, std::size_t>> { { 0x1F250, 0x1F251 }, { sce_Hani.data(), sce_Hani.size() } }, // So  [2] CIRCLED IDEOGRAPH ADVANTAGE..CIRCLED IDEOGRAPH ACCEPT
} };
} // }}}

std::optional<std::span<Script const>> script_extensions(char32_t codepoint) noexcept {
    auto const p = search(tables::sce, codepoint);
    if (!p.has_value())
        return std::nullopt;
    return std::span<Script const>{ p->first, p->second };
}

namespace tables {
auto static const Block = std::array<Prop<::unicode::Block>, 346>{ // {{{
    Prop<::unicode::Block> { { 0x0000, 0x007F }, ::unicode::Block::Basic_Latin },
    Prop<::unicode::Block> { { 0x0080, 0x00FF }, ::unicode::Block::Latin_1_Supplement },
    Prop<::unicode::Block> { { 0x0100, 0x017F }, ::unicode::Block::Latin_Extended_A },
    Prop<::unicode::Block> { { 0x0180, 0x024F }, ::unicode::Block::Latin_Extended_B },
    Prop<::unicode::Block> { { 0x0250, 0x02AF }, ::unicode::Block::IPA_Extensions },
    Prop<::unicode::Block> { { 0x02B0, 0x02FF }, ::unicode::Block::Spacing_Modifier_Letters },
    Prop<::unicode::Block> { { 0x0300, 0x036F }, ::unicode::Block::Combining_Diacritical_Marks },
    Prop<::unicode::Block> { { 0x0370, 0x03FF }, ::unicode::Block::Greek_and_Coptic },
    Prop<::unicode::Block> { { 0x0400, 0x04FF }, ::unicode::Block::Cyrillic },
    Prop<::unicode::Block> { { 0x0500, 0x052F }, ::unicode::Block::Cyrillic_Supplement },
    Prop<::unicode::Block> { { 0x0530, 0x058F }, ::unicode::Block::Armenian },
    Prop<::unicode::Block> { { 0x0590, 0x05FF }, ::unicode::Block::Hebrew },
    Prop<::unicode::Block> { { 0x0600, 0x06FF }, ::unicode::Block::Arabic },
    Prop<::unicode::Block> { { 0x0700, 0x074F }, ::unicode::Block::Syriac },
    Prop<::unicode::Block> { { 0x0750, 0x077F }, ::unicode::Block::Arabic_Supplement },
    Prop<::unicode::Block> { { 0x0780, 0x07BF }, ::unicode::Block::Thaana },
    Prop<::unicode::Block> { { 0x07C0, 0x07FF }, ::unicode::Block::NKo },
    Prop<::unicode::Block> { { 0x0800, 0x083F }, ::unicode::Block::Samaritan },
    Prop<::unicode::Block> { { 0x0840, 0x085F }, ::unicode::Block::Mandaic },
    Prop<::unicode::Block> { { 0x0860, 0x086F }, ::unicode::Block::Syriac_Supplement },
    Prop<::unicode::Block> { { 0x0870, 0x089F }, ::unicode::Block::Arabic_Extended_B },
    Prop<::unicode::Block> { { 0x08A0, 0x08FF }, ::unicode::Block::Arabic_Extended_A },
    Prop<::unicode::Block> { { 0x0900, 0x097F }, ::unicode::Block::Devanagari },
    Prop<::unicode::Block> { { 0x0980, 0x09FF }, ::unicode::Block::Bengali },
    Prop<::unicode::Block> { { 0x0A00, 0x0A7F }, ::unicode::Block::Gurmukhi },
    Prop<::unicode::Block> { { 0x0A80, 0x0AFF }, ::unicode::Block::Gujarati },
    Prop<::unicode::Block> { { 0x0B00, 0x0B7F }, ::unicode::Block::Oriya },
    Prop<::unicode::Block> { { 0x0B80, 0x0BFF }, ::unicode::Block::Tamil },
    Prop<::unicode::Block> { { 0x0C00, 0x0C7F }, ::unicode::Block::Telugu },
    Prop<::unicode::Block> { { 0x0C80, 0x0CFF }, ::unicode::Block::Kannada },
    Prop<::unicode::Block> { { 0x0D00, 0x0D7F }, ::unicode::Block::Malayalam },
    Prop<::unicode::Block> { { 0x0D80, 0x0DFF }, ::unicode::Block::Sinhala },
    Prop<::unicode::Block> { { 0x0E00, 0x0E7F }, ::unicode::Block::Thai },
    Prop<::unicode::Block> { { 0x0E80, 0x0EFF }, ::unicode::Block::Lao },
    Prop<::unicode::Block> { { 0x0F00, 0x0FFF }, ::unicode::Block::Tibetan },
    Prop<::unicode::Block> { { 0x1000, 0x109F }, ::unicode::Block::Myanmar },
    Prop<::unicode::Block> { { 0x10A0, 0x10FF }, ::unicode::Block::Georgian },
    Prop<::unicode::Block> { { 0x1100, 0x11FF }, ::unicode::Block::Hangul_Jamo },
    Prop<::unicode::Block> { { 0x1200, 0x137F }, ::unicode::Block::Ethiopic },
    Prop<::unicode::Block> { { 0x1380, 0x139F }, ::unicode::Block::Ethiopic_Supplement },
    Prop<::unicode::Block> { { 0x13A0, 0x13FF }, ::unicode::Block::Cherokee },
    Prop<::unicode::Block> { { 0x1400, 0x167F }, ::unicode::Block::Unified_Canadian_Aboriginal_Syllabics },
    Prop<::unicode::Block> { { 0x1680, 0x169F }, ::unicode::Block::Ogham },
    Prop<::unicode::Block> { { 0x16A0, 0x16FF }, ::unicode::Block::Runic },
    Prop<::unicode::Block> { { 0x1700, 0x171F }, ::unicode::Block::Tagalog },
    Prop<::unicode::Block> { { 0x1720, 0x173F }, ::unicode::Block::Hanunoo },
    Prop<::unicode::Block> { { 0x1740, 0x175F }, ::unicode::Block::Buhid },
    Prop<::unicode::Block> { { 0x1760, 0x177F }, ::unicode::Block::Tagbanwa },
    Prop<::unicode::Block> { { 0x1780, 0x17FF }, ::unicode::Block::Khmer },
    Prop<::unicode::Block> { { 0x1800, 0x18AF }, ::unicode::Block::Mongolian },
    Prop<::unicode::Block> { { 0x18B0, 0x18FF }, ::unicode::Block::Unified_Canadian_Aboriginal_Syllabics_Extended },
    Prop<::unicode::Block> { { 0x1900, 0x194F }, ::unicode::Block::Limbu },
    Prop<::unicode::Block> { { 0x1950, 0x197F }, ::unicode::Block::Tai_Le },
    Prop<::unicode::Block> { { 0x1980, 0x19DF }, ::unicode::Block::New_Tai_Lue },
    Prop<::unicode::Block> { { 0x19E0, 0x19FF }, ::unicode::Block::Khmer_Symbols },
    Prop<::unicode::Block> { { 0x1A00, 0x1A1F }, ::unicode::Block::Buginese },
    Prop<::unicode::Block> { { 0x1A20, 0x1AAF }, ::unicode::Block::Tai_Tham },
    Prop<::unicode::Block> { { 0x1AB0, 0x1AFF }, ::unicode::Block::Combining_Diacritical_Marks_Extended },
    Prop<::unicode::Block> { { 0x1B00, 0x1B7F }, ::unicode::Block::Balinese },
    Prop<::unicode::Block> { { 0x1B80, 0x1BBF }, ::unicode::Block::Sundanese },
    Prop<::unicode::Block> { { 0x1BC0, 0x1BFF }, ::unicode::Block::Batak },
    Prop<::unicode::Block> { { 0x1C00, 0x1C4F }, ::unicode::Block::Lepcha },
    Prop<::unicode::Block> { { 0x1C50, 0x1C7F }, ::unicode::Block::Ol_Chiki },
    Prop<::unicode::Block> { { 0x1C80, 0x1C8F }, ::unicode::Block::Cyrillic_Extended_C },
    Prop<::unicode::Block> { { 0x1C90, 0x1CBF }, ::unicode::Block::Georgian_Extended },
    Prop<::unicode::Block> { { 0x1CC0, 0x1CCF }, ::unicode::Block::Sundanese_Supplement },
    Prop<::unicode::Block> { { 0x1CD0, 0x1CFF }, ::unicode::Block::Vedic_Extensions },
    Prop<::unicode::Block> { { 0x1D00, 0x1D7F }, ::unicode::Block::Phonetic_Extensions },
    Prop<::unicode::Block> { { 0x1D80, 0x1DBF }, ::unicode::Block::Phonetic_Extensions_Supplement },
    Prop<::unicode::Block> { { 0x1DC0, 0x1DFF }, ::unicode::Block::Combining_Diacritical_Marks_Supplement },
    Prop<::unicode::Block> { { 0x1E00, 0x1EFF }, ::unicode::Block::Latin_Extended_Additional },
    Prop<::unicode::Block> { { 0x1F00, 0x1FFF }, ::unicode::Block::Greek_Extended },
    Prop<::unicode::Block> { { 0x2000, 0x206F }, ::unicode::Block::General_Punctuation },
    Prop<::unicode::Block> { { 0x2070, 0x209F }, ::unicode::Block::Superscripts_and_Subscripts },
    Prop<::unicode::Block> { { 0x20A0, 0x20CF }, ::unicode::Block::Currency_Symbols },
    Prop<::unicode::Block> { { 0x20D0, 0x20FF }, ::unicode::Block::Combining_Diacritical_Marks_for_Symbols },
    Prop<::unicode::Block> { { 0x2100, 0x214F }, ::unicode::Block::Letterlike_Symbols },
    Prop<::unicode::Block> { { 0x2150, 0x218F }, ::unicode::Block::Number_Forms },
    Prop<::unicode::Block> { { 0x2190, 0x21FF }, ::unicode::Block::Arrows },
    Prop<::unicode::Block> { { 0x2200, 0x22FF }, ::unicode::Block::Mathematical_Operators },
    Prop<::unicode::Block> { { 0x2300, 0x23FF }, ::unicode::Block::Miscellaneous_Technical },
    Prop<::unicode::Block> { { 0x2400, 0x243F }, ::unicode::Block::Control_Pictures },
    Prop<::unicode::Block> { { 0x2440, 0x245F }, ::unicode::Block::Optical_Character_Recognition },
    Prop<::unicode::Block> { { 0x2460, 0x24FF }, ::unicode::Block::Enclosed_Alphanumerics },
    Prop<::unicode::Block> { { 0x2500, 0x257F }, ::unicode::Block::Box_Drawing },
    Prop<::unicode::Block> { { 0x2580, 0x259F }, ::unicode::Block::Block_Elements },
    Prop<::unicode::Block> { { 0x25A0, 0x25FF }, ::unicode::Block::Geometric_Shapes },
    Prop<::unicode::Block> { { 0x2600, 0x26FF }, ::unicode::Block::Miscellaneous_Symbols },
    Prop<::unicode::Block> { { 0x2700, 0x27BF }, ::unicode::Block::Dingbats },
    Prop<::unicode::Block> { { 0x27C0, 0x27EF }, ::unicode::Block::Miscellaneous_Mathematical_Symbols_A },
    Prop<::unicode::Block> { { 0x27F0, 0x27FF }, ::unicode::Block::Supplemental_Arrows_A },
    Prop<::unicode::Block> { { 0x2800, 0x28FF }, ::unicode::Block::Braille_Patterns },
    Prop<::unicode::Block> { { 0x2900, 0x297F }, ::unicode::Block::Supplemental_Arrows_B },
    Prop<::unicode::Block> { { 0x2980, 0x29FF }, ::unicode::Block::Miscellaneous_Mathematical_Symbols_B },
    Prop<::unicode::Block> { { 0x2A00, 0x2AFF }, ::unicode::Block::Supplemental_Mathematical_Operators },
    Prop<::unicode::Block> { { 0x2B00, 0x2BFF }, ::unicode::Block::Miscellaneous_Symbols_and_Arrows },
    Prop<::unicode::Block> { { 0x2C00, 0x2C5F }, ::unicode::Block::Glagolitic },
    Prop<::unicode::Block> { { 0x2C60, 0x2C7F }, ::unicode::Block::Latin_Extended_C },
    Prop<::unicode::Block> { { 0x2C80, 0x2CFF }, ::unicode::Block::Coptic },
    Prop<::unicode::Block> { { 0x2D00, 0x2D2F }, ::unicode::Block::Georgian_Supplement },
    Prop<::unicode::Block> { { 0x2D30, 0x2D7F }, ::unicode::Block::Tifinagh },
    Prop<::unicode::Block> { { 0x2D80, 0x2DDF }, ::unicode::Block::Ethiopic_Extended },
    Prop<::unicode::Block> { { 0x2DE0, 0x2DFF }, ::unicode::Block::Cyrillic_Extended_A },
    Prop<::unicode::Block> { { 0x2E00, 0x2E7F }, ::unicode::Block::Supplemental_Punctuation },
    Prop<::unicode::Block> { { 0x2E80, 0x2EFF }, ::unicode::Block::CJK_Radicals_Supplement },
    Prop<::unicode::Block> { { 0x2F00, 0x2FDF }, ::unicode::Block::Kangxi_Radicals },
    Prop<::unicode::Block> { { 0x2FF0, 0x2FFF }, ::unicode::Block::Ideographic_Description_Characters },
    Prop<::unicode::Block> { { 0x3000, 0x303F }, ::unicode::Block::CJK_Symbols_and_Punctuation },
    Prop<::unicode::Block> { { 0x3040, 0x309F }, ::unicode::Block::Hiragana },
    Prop<::unicode::Block> { { 0x30A0, 0x30FF }, ::unicode::Block::Katakana },
    Prop<::unicode::Block> { { 0x3100, 0x312F }, ::unicode::Block::Bopomofo },
    Prop<::unicode::Block> { { 0x3130, 0x318F }, ::unicode::Block::Hangul_Compatibility_Jamo },
    Prop<::unicode::Block> { { 0x3190, 0x319F }, ::unicode::Block::Kanbun },
    Prop<::unicode::Block> { { 0x31A0, 0x31BF }, ::unicode::Block::Bopomofo_Extended },
    Prop<::unicode::Block> { { 0x31C0, 0x31EF }, ::unicode::Block::CJK_Strokes },
    Prop<::unicode::Block> { { 0x31F0, 0x31FF }, ::unicode::Block::Katakana_Phonetic_Extensions },
    Prop<::unicode::Block> { { 0x3200, 0x32FF }, ::unicode::Block::Enclosed_CJK_Letters_and_Months },
    Prop<::unicode::Block> { { 0x3300, 0x33FF }, ::unicode::Block::CJK_Compatibility },
    Prop<::unicode::Block> { { 0x3400, 0x4DBF }, ::unicode::Block::CJK_Unified_Ideographs_Extension_A },
    Prop<::unicode::Block> { { 0x4DC0, 0x4DFF }, ::unicode::Block::Yijing_Hexagram_Symbols },
    Prop<::unicode::Block> { { 0x4E00, 0x9FFF }, ::unicode::Block::CJK_Unified_Ideographs },
    Prop<::unicode::Block> { { 0xA000, 0xA48F }, ::unicode::Block::Yi_Syllables },
    Prop<::unicode::Block> { { 0xA490, 0xA4CF }, ::unicode::Block::Yi_Radicals },
    Prop<::unicode::Block> { { 0xA4D0, 0xA4FF }, ::unicode::Block::Lisu },
    Prop<::unicode::Block> { { 0xA500, 0xA63F }, ::unicode::Block::Vai },
    Prop<::unicode::Block> { { 0xA640, 0xA69F }, ::unicode::Block::Cyrillic_Extended_B },
    Prop<::unicode::Block> { { 0xA6A0, 0xA6FF }, ::unicode::Block::Bamum },
    Prop<::unicode::Block> { { 0xA700, 0xA71F }, ::unicode::Block::Modifier_Tone_Letters },
    Prop<::unicode::Block> { { 0xA720, 0xA7FF }, ::unicode::Block::Latin_Extended_D },
    Prop<::unicode::Block> { { 0xA800, 0xA82F }, ::unicode::Block::Syloti_Nagri },
    Prop<::unicode::Block> { { 0xA830, 0xA83F }, ::unicode::Block::Common_Indic_Number_Forms },
    Prop<::unicode::Block> { { 0xA840, 0xA87F }, ::unicode::Block::Phags_pa },
    Prop<::unicode::Block> { { 0xA880, 0xA8DF }, ::unicode::Block::Saurashtra },
    Prop<::unicode::Block> { { 0xA8E0, 0xA8FF }, ::unicode::Block::Devanagari_Extended },
    Prop<::unicode::Block> { { 0xA900, 0xA92F }, ::unicode::Block::Kayah_Li },
    Prop<::unicode::Block> { { 0xA930, 0xA95F }, ::unicode::Block::Rejang },
    Prop<::unicode::Block> { { 0xA960, 0xA97F }, ::unicode::Block::Hangul_Jamo_Extended_A },
    Prop<::unicode::Block> { { 0xA980, 0xA9DF }, ::unicode::Block::Javanese },
    Prop<::unicode::Block> { { 0xA9E0, 0xA9FF }, ::unicode::Block::Myanmar_Extended_B },
    Prop<::unicode::Block> { { 0xAA00, 0xAA5F }, ::unicode::Block::Cham },
    Prop<::unicode::Block> { { 0xAA60, 0xAA7F }, ::unicode::Block::Myanmar_Extended_A },
    Prop<::unicode::Block> { { 0xAA80, 0xAADF }, ::unicode::Block::Tai_Viet },
    Prop<::unicode::Block> { { 0xAAE0, 0xAAFF }, ::unicode::Block::Meetei_Mayek_Extensions },
    Prop<::unicode::Block> { { 0xAB00, 0xAB2F }, ::unicode::Block::Ethiopic_Extended_A },
    Prop<::unicode::Block> { { 0xAB30, 0xAB6F }, ::unicode::Block::Latin_Extended_E },
    Prop<::unicode::Block> { { 0xAB70, 0xABBF }, ::unicode::Block::Cherokee_Supplement },
    Prop<::unicode::Block> { { 0xABC0, 0xABFF }, ::unicode::Block::Meetei_Mayek },
    Prop<::unicode::Block> { { 0xAC00, 0xD7AF }, ::unicode::Block::Hangul_Syllables },
    Prop<::unicode::Block> { { 0xD7B0, 0xD7FF }, ::unicode::Block::Hangul_Jamo_Extended_B },
    Prop<::unicode::Block> { { 0xD800, 0xDB7F }, ::unicode::Block::High_Surrogates },
    Prop<::unicode::Block> { { 0xDB80, 0xDBFF }, ::unicode::Block::High_Private_Use_Surrogates },
    Prop<::unicode::Block> { { 0xDC00, 0xDFFF }, ::unicode::Block::Low_Surrogates },
    Prop<::unicode::Block> { { 0xE000, 0xF8FF }, ::unicode::Block::Private_Use_Area },
    Prop<::unicode::Block> { { 0xF900, 0xFAFF }, ::unicode::Block::CJK_Compatibility_Ideographs },
    Prop<::unicode::Block> { { 0xFB00, 0xFB4F }, ::unicode::Block::Alphabetic_Presentation_Forms },
    Prop<::unicode::Block> { { 0xFB50, 0xFDFF }, ::unicode::Block::Arabic_Presentation_Forms_A },
    Prop<::unicode::Block> { { 0xFE00, 0xFE0F }, ::unicode::Block::Variation_Selectors },
    Prop<::unicode::Block> { { 0xFE10, 0xFE1F }, ::unicode::Block::Vertical_Forms },
    Prop<::unicode::Block> { { 0xFE20, 0xFE2F }, ::unicode::Block::Combining_Half_Marks },
    Prop<::unicode::Block> { { 0xFE30, 0xFE4F }, ::unicode::Block::CJK_Compatibility_Forms },
    Prop<::unicode::Block> { { 0xFE50, 0xFE6F }, ::unicode::Block::Small_Form_Variants },
    Prop<::unicode::Block> { { 0xFE70, 0xFEFF }, ::unicode::Block::Arabic_Presentation_Forms_B },
    Prop<::unicode::Block> { { 0xFF00, 0xFFEF }, ::unicode::Block::Halfwidth_and_Fullwidth_Forms },
    Prop<::unicode::Block> { { 0xFFF0, 0xFFFF }, ::unicode::Block::Specials },
    Prop<::unicode::Block> { { 0x10000, 0x1007F }, ::unicode::Block::Linear_B_Syllabary },
    Prop<::unicode::Block> { { 0x10080, 0x100FF }, ::unicode::Block::Linear_B_Ideograms },
    Prop<::unicode::Block> { { 0x10100, 0x1013F }, ::unicode::Block::Aegean_Numbers },
    Prop<::unicode::Block> { { 0x10140, 0x1018F }, ::unicode::Block::Ancient_Greek_Numbers },
    Prop<::unicode::Block> { { 0x10190, 0x101CF }, ::unicode::Block::Ancient_Symbols },
    Prop<::unicode::Block> { { 0x101D0, 0x101FF }, ::unicode::Block::Phaistos_Disc },
    Prop<::unicode::Block> { { 0x10280, 0x1029F }, ::unicode::Block::Lycian },
    Prop<::unicode::Block> { { 0x102A0, 0x102DF }, ::unicode::Block::Carian },
    Prop<::unicode::Block> { { 0x102E0, 0x102FF }, ::unicode::Block::Coptic_Epact_Numbers },
    Prop<::unicode::Block> { { 0x10300, 0x1032F }, ::unicode::Block::Old_Italic },
    Prop<::unicode::Block> { { 0x10330, 0x1034F }, ::unicode::Block::Gothic },
    Prop<::unicode::Block> { { 0x10350, 0x1037F }, ::unicode::Block::Old_Permic },
    Prop<::unicode::Block> { { 0x10380, 0x1039F }, ::unicode::Block::Ugaritic },
    Prop<::unicode::Block> { { 0x103A0, 0x103DF }, ::unicode::Block::Old_Persian },
    Prop<::unicode::Block> { { 0x10400, 0x1044F }, ::unicode::Block::Deseret },
    Prop<::unicode::Block> { { 0x10450, 0x1047F }, ::unicode::Block::Shavian },
    Prop<::unicode::Block> { { 0x10480, 0x104AF }, ::unicode::Block::Osmanya },
    Prop<::unicode::Block> { { 0x104B0, 0x104FF }, ::unicode::Block::Osage },
    Prop<::unicode::Block> { { 0x10500, 0x1052F }, ::unicode::Block::Elbasan },
    Prop<::unicode::Block> { { 0x10530, 0x1056F }, ::unicode::Block::Caucasian_Albanian },
    Prop<::unicode::Block> { { 0x10570, 0x105BF }, ::unicode::Block::Vithkuqi },
    Prop<::unicode::Block> { { 0x105C0, 0x105FF }, ::unicode::Block::Todhri },
    Prop<::unicode::Block> { { 0x10600, 0x1077F }, ::unicode::Block::Linear_A },
    Prop<::unicode::Block> { { 0x10780, 0x107BF }, ::unicode::Block::Latin_Extended_F },
    Prop<::unicode::Block> { { 0x10800, 0x1083F }, ::unicode::Block::Cypriot_Syllabary },
    Prop<::unicode::Block> { { 0x10840, 0x1085F }, ::unicode::Block::Imperial_Aramaic },
    Prop<::unicode::Block> { { 0x10860, 0x1087F }, ::unicode::Block::Palmyrene },
    Prop<::unicode::Block> { { 0x10880, 0x108AF }, ::unicode::Block::Nabataean },
    Prop<::unicode::Block> { { 0x108E0, 0x108FF }, ::unicode::Block::Hatran },
    Prop<::unicode::Block> { { 0x10900, 0x1091F }, ::unicode::Block::Phoenician },
    Prop<::unicode::Block> { { 0x10920, 0x1093F }, ::unicode::Block::Lydian },
    Prop<::unicode::Block> { { 0x10940, 0x1095F }, ::unicode::Block::Sidetic },
    Prop<::unicode::Block> { { 0x10980, 0x1099F }, ::unicode::Block::Meroitic_Hieroglyphs },
    Prop<::unicode::Block> { { 0x109A0, 0x109FF }, ::unicode::Block::Meroitic_Cursive },
    Prop<::unicode::Block> { { 0x10A00, 0x10A5F }, ::unicode::Block::Kharoshthi },
    Prop<::unicode::Block> { { 0x10A60, 0x10A7F }, ::unicode::Block::Old_South_Arabian },
    Prop<::unicode::Block> { { 0x10A80, 0x10A9F }, ::unicode::Block::Old_North_Arabian },
    Prop<::unicode::Block> { { 0x10AC0, 0x10AFF }, ::unicode::Block::Manichaean },
    Prop<::unicode::Block> { { 0x10B00, 0x10B3F }, ::unicode::Block::Avestan },
    Prop<::unicode::Block> { { 0x10B40, 0x10B5F }, ::unicode::Block::Inscriptional_Parthian },
    Prop<::unicode::Block> { { 0x10B60, 0x10B7F }, ::unicode::Block::Inscriptional_Pahlavi },
    Prop<::unicode::Block> { { 0x10B80, 0x10BAF }, ::unicode::Block::Psalter_Pahlavi },
    Prop<::unicode::Block> { { 0x10C00, 0x10C4F }, ::unicode::Block::Old_Turkic },
    Prop<::unicode::Block> { { 0x10C80, 0x10CFF }, ::unicode::Block::Old_Hungarian },
    Prop<::unicode::Block> { { 0x10D00, 0x10D3F }, ::unicode::Block::Hanifi_Rohingya },
    Prop<::unicode::Block> { { 0x10D40, 0x10D8F }, ::unicode::Block::Garay },
    Prop<::unicode::Block> { { 0x10E60, 0x10E7F }, ::unicode::Block::Rumi_Numeral_Symbols },
    Prop<::unicode::Block> { { 0x10E80, 0x10EBF }, ::unicode::Block::Yezidi },
    Prop<::unicode::Block> { { 0x10EC0, 0x10EFF }, ::unicode::Block::Arabic_Extended_C },
    Prop<::unicode::Block> { { 0x10F00, 0x10F2F }, ::unicode::Block::Old_Sogdian },
    Prop<::unicode::Block> { { 0x10F30, 0x10F6F }, ::unicode::Block::Sogdian },
    Prop<::unicode::Block> { { 0x10F70, 0x10FAF }, ::unicode::Block::Old_Uyghur },
    Prop<::unicode::Block> { { 0x10FB0, 0x10FDF }, ::unicode::Block::Chorasmian },
    Prop<::unicode::Block> { { 0x10FE0, 0x10FFF }, ::unicode::Block::Elymaic },
    Prop<::unicode::Block> { { 0x11000, 0x1107F }, ::unicode::Block::Brahmi },
    Prop<::unicode::Block> { { 0x11080, 0x110CF }, ::unicode::Block::Kaithi },
    Prop<::unicode::Block> { { 0x110D0, 0x110FF }, ::unicode::Block::Sora_Sompeng },
    Prop<::unicode::Block> { { 0x11100, 0x1114F }, ::unicode::Block::Chakma },
    Prop<::unicode::Block> { { 0x11150, 0x1117F }, ::unicode::Block::Mahajani },
    Prop<::unicode::Block> { { 0x11180, 0x111DF }, ::unicode::Block::Sharada },
    Prop<::unicode::Block> { { 0x111E0, 0x111FF }, ::unicode::Block::Sinhala_Archaic_Numbers },
    Prop<::unicode::Block> { { 0x11200, 0x1124F }, ::unicode::Block::Khojki },
    Prop<::unicode::Block> { { 0x11280, 0x112AF }, ::unicode::Block::Multani },
    Prop<::unicode::Block> { { 0x112B0, 0x112FF }, ::unicode::Block::Khudawadi },
    Prop<::unicode::Block> { { 0x11300, 0x1137F }, ::unicode::Block::Grantha },
    Prop<::unicode::Block> { { 0x11380, 0x113FF }, ::unicode::Block::Tulu_Tigalari },
    Prop<::unicode::Block> { { 0x11400, 0x1147F }, ::unicode::Block::Newa },
    Prop<::unicode::Block> { { 0x11480, 0x114DF }, ::unicode::Block::Tirhuta },
    Prop<::unicode::Block> { { 0x11580, 0x115FF }, ::unicode::Block::Siddham },
    Prop<::unicode::Block> { { 0x11600, 0x1165F }, ::unicode::Block::Modi },
    Prop<::unicode::Block> { { 0x11660, 0x1167F }, ::unicode::Block::Mongolian_Supplement },
    Prop<::unicode::Block> { { 0x11680, 0x116CF }, ::unicode::Block::Takri },
    Prop<::unicode::Block> { { 0x116D0, 0x116FF }, ::unicode::Block::Myanmar_Extended_C },
    Prop<::unicode::Block> { { 0x11700, 0x1174F }, ::unicode::Block::Ahom },
    Prop<::unicode::Block> { { 0x11800, 0x1184F }, ::unicode::Block::Dogra },
    Prop<::unicode::Block> { { 0x118A0, 0x118FF }, ::unicode::Block::Warang_Citi },
    Prop<::unicode::Block> { { 0x11900, 0x1195F }, ::unicode::Block::Dives_Akuru },
    Prop<::unicode::Block> { { 0x119A0, 0x119FF }, ::unicode::Block::Nandinagari },
    Prop<::unicode::Block> { { 0x11A00, 0x11A4F }, ::unicode::Block::Zanabazar_Square },
    Prop<::unicode::Block> { { 0x11A50, 0x11AAF }, ::unicode::Block::Soyombo },
    Prop<::unicode::Block> { { 0x11AB0, 0x11ABF }, ::unicode::Block::Unified_Canadian_Aboriginal_Syllabics_Extended_A },
    Prop<::unicode::Block> { { 0x11AC0, 0x11AFF }, ::unicode::Block::Pau_Cin_Hau },
    Prop<::unicode::Block> { { 0x11B00, 0x11B5F }, ::unicode::Block::Devanagari_Extended_A },
    Prop<::unicode::Block> { { 0x11B60, 0x11B7F }, ::unicode::Block::Sharada_Supplement },
    Prop<::unicode::Block> { { 0x11BC0, 0x11BFF }, ::unicode::Block::Sunuwar },
    Prop<::unicode::Block> { { 0x11C00, 0x11C6F }, ::unicode::Block::Bhaiksuki },
    Prop<::unicode::Block> { { 0x11C70, 0x11CBF }, ::unicode::Block::Marchen },
    Prop<::unicode::Block> { { 0x11D00, 0x11D5F }, ::unicode::Block::Masaram_Gondi },
    Prop<::unicode::Block> { { 0x11D60, 0x11DAF }, ::unicode::Block::Gunjala_Gondi },
    Prop<::unicode::Block> { { 0x11DB0, 0x11DEF }, ::unicode::Block::Tolong_Siki },
    Prop<::unicode::Block> { { 0x11EE0, 0x11EFF }, ::unicode::Block::Makasar },
    Prop<::unicode::Block> { { 0x11F00, 0x11F5F }, ::unicode::Block::Kawi },
    Prop<::unicode::Block> { { 0x11FB0, 0x11FBF }, ::unicode::Block::Lisu_Supplement },
    Prop<::unicode::Block> { { 0x11FC0, 0x11FFF }, ::unicode::Block::Tamil_Supplement },
    Prop<::unicode::Block> { { 0x12000, 0x123FF }, ::unicode::Block::Cuneiform },
    Prop<::unicode::Block> { { 0x12400, 0x1247F }, ::unicode::Block::Cuneiform_Numbers_and_Punctuation },
    Prop<::unicode::Block> { { 0x12480, 0x1254F }, ::unicode::Block::Early_Dynastic_Cuneiform },
    Prop<::unicode::Block> { { 0x12F90, 0x12FFF }, ::unicode::Block::Cypro_Minoan },
    Prop<::unicode::Block> { { 0x13000, 0x1342F }, ::unicode::Block::Egyptian_Hieroglyphs },
    Prop<::unicode::Block> { { 0x13430, 0x1345F }, ::unicode::Block::Egyptian_Hieroglyph_Format_Controls },
    Prop<::unicode::Block> { { 0x13460, 0x143FF }, ::unicode::Block::Egyptian_Hieroglyphs_Extended_A },
    Prop<::unicode::Block> { { 0x14400, 0x1467F }, ::unicode::Block::Anatolian_Hieroglyphs },
    Prop<::unicode::Block> { { 0x16100, 0x1613F }, ::unicode::Block::Gurung_Khema },
    Prop<::unicode::Block> { { 0x16800, 0x16A3F }, ::unicode::Block::Bamum_Supplement },
    Prop<::unicode::Block> { { 0x16A40, 0x16A6F }, ::unicode::Block::Mro },
    Prop<::unicode::Block> { { 0x16A70, 0x16ACF }, ::unicode::Block::Tangsa },
    Prop<::unicode::Block> { { 0x16AD0, 0x16AFF }, ::unicode::Block::Bassa_Vah },
    Prop<::unicode::Block> { { 0x16B00, 0x16B8F }, ::unicode::Block::Pahawh_Hmong },
    Prop<::unicode::Block> { { 0x16D40, 0x16D7F }, ::unicode::Block::Kirat_Rai },
    Prop<::unicode::Block> { { 0x16E40, 0x16E9F }, ::unicode::Block::Medefaidrin },
    Prop<::unicode::Block> { { 0x16EA0, 0x16EDF }, ::unicode::Block::Beria_Erfe },
    Prop<::unicode::Block> { { 0x16F00, 0x16F9F }, ::unicode::Block::Miao },
    Prop<::unicode::Block> { { 0x16FE0, 0x16FFF }, ::unicode::Block::Ideographic_Symbols_and_Punctuation },
    Prop<::unicode::Block> { { 0x17000, 0x187FF }, ::unicode::Block::Tangut },
    Prop<::unicode::Block> { { 0x18800, 0x18AFF }, ::unicode::Block::Tangut_Components },
    Prop<::unicode::Block> { { 0x18B00, 0x18CFF }, ::unicode::Block::Khitan_Small_Script },
    Prop<::unicode::Block> { { 0x18D00, 0x18D7F }, ::unicode::Block::Tangut_Supplement },
    Prop<::unicode::Block> { { 0x18D80, 0x18DFF }, ::unicode::Block::Tangut_Components_Supplement },
    Prop<::unicode::Block> { { 0x1AFF0, 0x1AFFF }, ::unicode::Block::Kana_Extended_B },
    Prop<::unicode::Block> { { 0x1B000, 0x1B0FF }, ::unicode::Block::Kana_Supplement },
    Prop<::unicode::Block> { { 0x1B100, 0x1B12F }, ::unicode::Block::Kana_Extended_A },
    Prop<::unicode::Block> { { 0x1B130, 0x1B16F }, ::unicode::Block::Small_Kana_Extension },
    Prop<::unicode::Block> { { 0x1B170, 0x1B2FF }, ::unicode::Block::Nushu },
    Prop<::unicode::Block> { { 0x1BC00, 0x1BC9F }, ::unicode::Block::Duployan },
    Prop<::unicode::Block> { { 0x1BCA0, 0x1BCAF }, ::unicode::Block::Shorthand_Format_Controls },
    Prop<::unicode::Block> { { 0x1CC00, 0x1CEBF }, ::unicode::Block::Symbols_for_Legacy_Computing_Supplement },
    Prop<::unicode::Block> { { 0x1CEC0, 0x1CEFF }, ::unicode::Block::Miscellaneous_Symbols_Supplement },
    Prop<::unicode::Block> { { 0x1CF00, 0x1CFCF }, ::unicode::Block::Znamenny_Musical_Notation },
    Prop<::unicode::Block> { { 0x1D000, 0x1D0FF }, ::unicode::Block::Byzantine_Musical_Symbols },
    Prop<::unicode::Block> { { 0x1D100, 0x1D1FF }, ::unicode::Block::Musical_Symbols },
    Prop<::unicode::Block> { { 0x1D200, 0x1D24F }, ::unicode::Block::Ancient_Greek_Musical_Notation },
    Prop<::unicode::Block> { { 0x1D2C0, 0x1D2DF }, ::unicode::Block::Kaktovik_Numerals },
    Prop<::unicode::Block> { { 0x1D2E0, 0x1D2FF }, ::unicode::Block::Mayan_Numerals },
    Prop<::unicode::Block> { { 0x1D300, 0x1D35F }, ::unicode::Block::Tai_Xuan_Jing_Symbols },
    Prop<::unicode::Block> { { 0x1D360, 0x1D37F }, ::unicode::Block::Counting_Rod_Numerals },
    Prop<::unicode::Block> { { 0x1D400, 0x1D7FF }, ::unicode::Block::Mathematical_Alphanumeric_Symbols },
    Prop<::unicode::Block> { { 0x1D800, 0x1DAAF }, ::unicode::Block::Sutton_SignWriting },
    Prop<::unicode::Block> { { 0x1DF00, 0x1DFFF }, ::unicode::Block::Latin_Extended_G },
    Prop<::unicode::Block> { { 0x1E000, 0x1E02F }, ::unicode::Block::Glagolitic_Supplement },
    Prop<::unicode::Block> { { 0x1E030, 0x1E08F }, ::unicode::Block::Cyrillic_Extended_D },
    Prop<::unicode::Block> { { 0x1E100, 0x1E14F }, ::unicode::Block::Nyiakeng_Puachue_Hmong },
    Prop<::unicode::Block> { { 0x1E290, 0x1E2BF }, ::unicode::Block::Toto },
    Prop<::unicode::Block> { { 0x1E2C0, 0x1E2FF }, ::unicode::Block::Wancho },
    Prop<::unicode::Block> { { 0x1E4D0, 0x1E4FF }, ::unicode::Block::Nag_Mundari },
    Prop<::unicode::Block> { { 0x1E5D0, 0x1E5FF }, ::unicode::Block::Ol_Onal },
    Prop<::unicode::Block> { { 0x1E6C0, 0x1E6FF }, ::unicode::Block::Tai_Yo },
    Prop<::unicode::Block> { { 0x1E7E0, 0x1E7FF }, ::unicode::Block::Ethiopic_Extended_B },
    Prop<::unicode::Block> { { 0x1E800, 0x1E8DF }, ::unicode::Block::Mende_Kikakui },
    Prop<::unicode::Block> { { 0x1E900, 0x1E95F }, ::unicode::Block::Adlam },
    Prop<::unicode::Block> { { 0x1EC70, 0x1ECBF }, ::unicode::Block::Indic_Siyaq_Numbers },
    Prop<::unicode::Block> { { 0x1ED00, 0x1ED4F }, ::unicode::Block::Ottoman_Siyaq_Numbers },
    Prop<::unicode::Block> { { 0x1EE00, 0x1EEFF }, ::unicode::Block::Arabic_Mathematical_Alphabetic_Symbols },
    Prop<::unicode::Block> { { 0x1F000, 0x1F02F }, ::unicode::Block::Mahjong_Tiles },
    Prop<::unicode::Block> { { 0x1F030, 0x1F09F }, ::unicode::Block::Domino_Tiles },
    Prop<::unicode::Block> { { 0x1F0A0, 0x1F0FF }, ::unicode::Block::Playing_Cards },
    Prop<::unicode::Block> { { 0x1F100, 0x1F1FF }, ::unicode::Block::Enclosed_Alphanumeric_Supplement },
    Prop<::unicode::Block> { { 0x1F200, 0x1F2FF }, ::unicode::Block::Enclosed_Ideographic_Supplement },
    Prop<::unicode::Block> { { 0x1F300, 0x1F5FF }, ::unicode::Block::Miscellaneous_Symbols_and_Pictographs },
    Prop<::unicode::Block> { { 0x1F600, 0x1F64F }, ::unicode::Block::Emoticons },
    Prop<::unicode::Block> { { 0x1F650, 0x1F67F }, ::unicode::Block::Ornamental_Dingbats },
    Prop<::unicode::Block> { { 0x1F680, 0x1F6FF }, ::unicode::Block::Transport_and_Map_Symbols },
    Prop<::unicode::Block> { { 0x1F700, 0x1F77F }, ::unicode::Block::Alchemical_Symbols },
    Prop<::unicode::Block> { { 0x1F780, 0x1F7FF }, ::unicode::Block::Geometric_Shapes_Extended },
    Prop<::unicode::Block> { { 0x1F800, 0x1F8FF }, ::unicode::Block::Supplemental_Arrows_C },
    Prop<::unicode::Block> { { 0x1F900, 0x1F9FF }, ::unicode::Block::Supplemental_Symbols_and_Pictographs },
    Prop<::unicode::Block> { { 0x1FA00, 0x1FA6F }, ::unicode::Block::Chess_Symbols },
    Prop<::unicode::Block> { { 0x1FA70, 0x1FAFF }, ::unicode::Block::Symbols_and_Pictographs_Extended_A },
    Prop<::unicode::Block> { { 0x1FB00, 0x1FBFF }, ::unicode::Block::Symbols_for_Legacy_Computing },
    Prop<::unicode::Block> { { 0x20000, 0x2A6DF }, ::unicode::Block::CJK_Unified_Ideographs_Extension_B },
    Prop<::unicode::Block> { { 0x2A700, 0x2B73F }, ::unicode::Block::CJK_Unified_Ideographs_Extension_C },
    Prop<::unicode::Block> { { 0x2B740, 0x2B81F }, ::unicode::Block::CJK_Unified_Ideographs_Extension_D },
    Prop<::unicode::Block> { { 0x2B820, 0x2CEAF }, ::unicode::Block::CJK_Unified_Ideographs_Extension_E },
    Prop<::unicode::Block> { { 0x2CEB0, 0x2EBEF }, ::unicode::Block::CJK_Unified_Ideographs_Extension_F },
    Prop<::unicode::Block> { { 0x2EBF0, 0x2EE5F }, ::unicode::Block::CJK_Unified_Ideographs_Extension_I },
    Prop<::unicode::Block> { { 0x2F800, 0x2FA1F }, ::unicode::Block::CJK_Compatibility_Ideographs_Supplement },
    Prop<::unicode::Block> { { 0x30000, 0x3134F }, ::unicode::Block::CJK_Unified_Ideographs_Extension_G },
    Prop<::unicode::Block> { { 0x31350, 0x323AF }, ::unicode::Block::CJK_Unified_Ideographs_Extension_H },
    Prop<::unicode::Block> { { 0x323B0, 0x3347F }, ::unicode::Block::CJK_Unified_Ideographs_Extension_J },
    Prop<::unicode::Block> { { 0xE0000, 0xE007F }, ::unicode::Block::Tags },
    Prop<::unicode::Block> { { 0xE0100, 0xE01EF }, ::unicode::Block::Variation_Selectors_Supplement },
    Prop<::unicode::Block> { { 0xF0000, 0xFFFFF }, ::unicode::Block::Supplementary_Private_Use_Area_A },
    Prop<::unicode::Block> { { 0x100000, 0x10FFFF }, ::unicode::Block::Supplementary_Private_Use_Area_B },
};
} // end namespace tables {}

Block block(char32_t codepoint) noexcept {
    return search(tables::Block, codepoint).value_or(::unicode::Block::Unspecified);
}

namespace tables {
auto static const Grapheme_Cluster_Break = std::array<Prop<::unicode::Grapheme_Cluster_Break>, 1429>{ // {{{
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0000, 0x0009 }, ::unicode::Grapheme_Cluster_Break::Control }, // Cc  [10] <control-0000>..<control-0009>
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x000A, 0x000A }, ::unicode::Grapheme_Cluster_Break::LF }, // Cc       <control-000A>
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x000B, 0x000C }, ::unicode::Grapheme_Cluster_Break::Control }, // Cc   [2] <control-000B>..<control-000C>
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x000D, 0x000D }, ::unicode::Grapheme_Cluster_Break::CR }, // Cc       <control-000D>
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x000E, 0x001F }, ::unicode::Grapheme_Cluster_Break::Control }, // Cc  [18] <control-000E>..<control-001F>
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x007F, 0x009F }, ::unicode::Grapheme_Cluster_Break::Control }, // Cc  [33] <control-007F>..<control-009F>
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x00AD, 0x00AD }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf       SOFT HYPHEN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0300, 0x036F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn [112] COMBINING GRAVE ACCENT..COMBINING LATIN SMALL LETTER X
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0483, 0x0487 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] COMBINING CYRILLIC TITLO..COMBINING CYRILLIC POKRYTIE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0488, 0x0489 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Me   [2] COMBINING CYRILLIC HUNDRED THOUSANDS SIGN..COMBINING CYRILLIC MILLIONS SIGN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0591, 0x05BD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [45] HEBREW ACCENT ETNAHTA..HEBREW POINT METEG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x05BF, 0x05BF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       HEBREW POINT RAFE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x05C1, 0x05C2 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] HEBREW POINT SHIN DOT..HEBREW POINT SIN DOT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x05C4, 0x05C5 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] HEBREW MARK UPPER DOT..HEBREW MARK LOWER DOT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x05C7, 0x05C7 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       HEBREW POINT QAMATS QATAN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0600, 0x0605 }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Cf   [6] ARABIC NUMBER SIGN..ARABIC NUMBER MARK ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0610, 0x061A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [11] ARABIC SIGN SALLALLAHOU ALAYHE WASSALLAM..ARABIC SMALL KASRA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x061C, 0x061C }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf       ARABIC LETTER MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x064B, 0x065F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [21] ARABIC FATHATAN..ARABIC WAVY HAMZA BELOW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0670, 0x0670 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       ARABIC LETTER SUPERSCRIPT ALEF
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x06D6, 0x06DC }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA..ARABIC SMALL HIGH SEEN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x06DD, 0x06DD }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Cf       ARABIC END OF AYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x06DF, 0x06E4 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [6] ARABIC SMALL HIGH ROUNDED ZERO..ARABIC SMALL HIGH MADDA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x06E7, 0x06E8 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] ARABIC SMALL HIGH YEH..ARABIC SMALL HIGH NOON
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x06EA, 0x06ED }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] ARABIC EMPTY CENTRE LOW STOP..ARABIC SMALL LOW MEEM
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x070F, 0x070F }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Cf       SYRIAC ABBREVIATION MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0711, 0x0711 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SYRIAC LETTER SUPERSCRIPT ALAPH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0730, 0x074A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [27] SYRIAC PTHAHA ABOVE..SYRIAC BARREKH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x07A6, 0x07B0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [11] THAANA ABAFILI..THAANA SUKUN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x07EB, 0x07F3 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [9] NKO COMBINING SHORT HIGH TONE..NKO COMBINING DOUBLE DOT ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x07FD, 0x07FD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       NKO DANTAYALAN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0816, 0x0819 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] SAMARITAN MARK IN..SAMARITAN MARK DAGESH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x081B, 0x0823 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [9] SAMARITAN MARK EPENTHETIC YUT..SAMARITAN VOWEL SIGN A
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0825, 0x0827 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] SAMARITAN VOWEL SIGN SHORT A..SAMARITAN VOWEL SIGN U
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0829, 0x082D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] SAMARITAN VOWEL SIGN LONG I..SAMARITAN MARK NEQUDAA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0859, 0x085B }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] MANDAIC AFFRICATION MARK..MANDAIC GEMINATION MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0890, 0x0891 }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Cf   [2] ARABIC POUND MARK ABOVE..ARABIC PIASTRE MARK ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0897, 0x089F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [9] ARABIC PEPET..ARABIC HALF MADDA OVER MADDA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x08CA, 0x08E1 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [24] ARABIC SMALL HIGH FARSI YEH..ARABIC SMALL HIGH SIGN SAFHA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x08E2, 0x08E2 }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Cf       ARABIC DISPUTED END OF AYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x08E3, 0x0902 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [32] ARABIC TURNED DAMMA BELOW..DEVANAGARI SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0903, 0x0903 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       DEVANAGARI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x093A, 0x093A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       DEVANAGARI VOWEL SIGN OE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x093B, 0x093B }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       DEVANAGARI VOWEL SIGN OOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x093C, 0x093C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       DEVANAGARI SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x093E, 0x0940 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] DEVANAGARI VOWEL SIGN AA..DEVANAGARI VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0941, 0x0948 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [8] DEVANAGARI VOWEL SIGN U..DEVANAGARI VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0949, 0x094C }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [4] DEVANAGARI VOWEL SIGN CANDRA O..DEVANAGARI VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x094D, 0x094D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       DEVANAGARI SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x094E, 0x094F }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] DEVANAGARI VOWEL SIGN PRISHTHAMATRA E..DEVANAGARI VOWEL SIGN AW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0951, 0x0957 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] DEVANAGARI STRESS SIGN UDATTA..DEVANAGARI VOWEL SIGN UUE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0962, 0x0963 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] DEVANAGARI VOWEL SIGN VOCALIC L..DEVANAGARI VOWEL SIGN VOCALIC LL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0981, 0x0981 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BENGALI SIGN CANDRABINDU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0982, 0x0983 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] BENGALI SIGN ANUSVARA..BENGALI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x09BC, 0x09BC }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BENGALI SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x09BE, 0x09BE }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       BENGALI VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x09BF, 0x09C0 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] BENGALI VOWEL SIGN I..BENGALI VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x09C1, 0x09C4 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] BENGALI VOWEL SIGN U..BENGALI VOWEL SIGN VOCALIC RR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x09C7, 0x09C8 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] BENGALI VOWEL SIGN E..BENGALI VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x09CB, 0x09CC }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] BENGALI VOWEL SIGN O..BENGALI VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x09CD, 0x09CD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BENGALI SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x09D7, 0x09D7 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       BENGALI AU LENGTH MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x09E2, 0x09E3 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] BENGALI VOWEL SIGN VOCALIC L..BENGALI VOWEL SIGN VOCALIC LL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x09FE, 0x09FE }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BENGALI SANDHI MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0A01, 0x0A02 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] GURMUKHI SIGN ADAK BINDI..GURMUKHI SIGN BINDI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0A03, 0x0A03 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       GURMUKHI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0A3C, 0x0A3C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       GURMUKHI SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0A3E, 0x0A40 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] GURMUKHI VOWEL SIGN AA..GURMUKHI VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0A41, 0x0A42 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] GURMUKHI VOWEL SIGN U..GURMUKHI VOWEL SIGN UU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0A47, 0x0A48 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] GURMUKHI VOWEL SIGN EE..GURMUKHI VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0A4B, 0x0A4D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] GURMUKHI VOWEL SIGN OO..GURMUKHI SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0A51, 0x0A51 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       GURMUKHI SIGN UDAAT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0A70, 0x0A71 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] GURMUKHI TIPPI..GURMUKHI ADDAK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0A75, 0x0A75 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       GURMUKHI SIGN YAKASH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0A81, 0x0A82 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] GUJARATI SIGN CANDRABINDU..GUJARATI SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0A83, 0x0A83 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       GUJARATI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0ABC, 0x0ABC }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       GUJARATI SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0ABE, 0x0AC0 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] GUJARATI VOWEL SIGN AA..GUJARATI VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0AC1, 0x0AC5 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] GUJARATI VOWEL SIGN U..GUJARATI VOWEL SIGN CANDRA E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0AC7, 0x0AC8 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] GUJARATI VOWEL SIGN E..GUJARATI VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0AC9, 0x0AC9 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       GUJARATI VOWEL SIGN CANDRA O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0ACB, 0x0ACC }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] GUJARATI VOWEL SIGN O..GUJARATI VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0ACD, 0x0ACD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       GUJARATI SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0AE2, 0x0AE3 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] GUJARATI VOWEL SIGN VOCALIC L..GUJARATI VOWEL SIGN VOCALIC LL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0AFA, 0x0AFF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [6] GUJARATI SIGN SUKUN..GUJARATI SIGN TWO-CIRCLE NUKTA ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B01, 0x0B01 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       ORIYA SIGN CANDRABINDU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B02, 0x0B03 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] ORIYA SIGN ANUSVARA..ORIYA SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B3C, 0x0B3C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       ORIYA SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B3E, 0x0B3E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       ORIYA VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B3F, 0x0B3F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       ORIYA VOWEL SIGN I
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B40, 0x0B40 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       ORIYA VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B41, 0x0B44 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] ORIYA VOWEL SIGN U..ORIYA VOWEL SIGN VOCALIC RR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B47, 0x0B48 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] ORIYA VOWEL SIGN E..ORIYA VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B4B, 0x0B4C }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] ORIYA VOWEL SIGN O..ORIYA VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B4D, 0x0B4D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       ORIYA SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B55, 0x0B56 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] ORIYA SIGN OVERLINE..ORIYA AI LENGTH MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B57, 0x0B57 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       ORIYA AU LENGTH MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B62, 0x0B63 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] ORIYA VOWEL SIGN VOCALIC L..ORIYA VOWEL SIGN VOCALIC LL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0B82, 0x0B82 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAMIL SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0BBE, 0x0BBE }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       TAMIL VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0BBF, 0x0BBF }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       TAMIL VOWEL SIGN I
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0BC0, 0x0BC0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAMIL VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0BC1, 0x0BC2 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] TAMIL VOWEL SIGN U..TAMIL VOWEL SIGN UU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0BC6, 0x0BC8 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] TAMIL VOWEL SIGN E..TAMIL VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0BCA, 0x0BCC }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] TAMIL VOWEL SIGN O..TAMIL VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0BCD, 0x0BCD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAMIL SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0BD7, 0x0BD7 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       TAMIL AU LENGTH MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0C00, 0x0C00 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TELUGU SIGN COMBINING CANDRABINDU ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0C01, 0x0C03 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] TELUGU SIGN CANDRABINDU..TELUGU SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0C04, 0x0C04 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TELUGU SIGN COMBINING ANUSVARA ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0C3C, 0x0C3C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TELUGU SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0C3E, 0x0C40 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] TELUGU VOWEL SIGN AA..TELUGU VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0C41, 0x0C44 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [4] TELUGU VOWEL SIGN U..TELUGU VOWEL SIGN VOCALIC RR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0C46, 0x0C48 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] TELUGU VOWEL SIGN E..TELUGU VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0C4A, 0x0C4D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] TELUGU VOWEL SIGN O..TELUGU SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0C55, 0x0C56 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] TELUGU LENGTH MARK..TELUGU AI LENGTH MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0C62, 0x0C63 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] TELUGU VOWEL SIGN VOCALIC L..TELUGU VOWEL SIGN VOCALIC LL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0C81, 0x0C81 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KANNADA SIGN CANDRABINDU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0C82, 0x0C83 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] KANNADA SIGN ANUSVARA..KANNADA SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CBC, 0x0CBC }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KANNADA SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CBE, 0x0CBE }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       KANNADA VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CBF, 0x0CBF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KANNADA VOWEL SIGN I
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CC0, 0x0CC0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       KANNADA VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CC1, 0x0CC1 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       KANNADA VOWEL SIGN U
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CC2, 0x0CC2 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       KANNADA VOWEL SIGN UU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CC3, 0x0CC4 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] KANNADA VOWEL SIGN VOCALIC R..KANNADA VOWEL SIGN VOCALIC RR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CC6, 0x0CC6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KANNADA VOWEL SIGN E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CC7, 0x0CC8 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc   [2] KANNADA VOWEL SIGN EE..KANNADA VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CCA, 0x0CCB }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc   [2] KANNADA VOWEL SIGN O..KANNADA VOWEL SIGN OO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CCC, 0x0CCD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] KANNADA VOWEL SIGN AU..KANNADA SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CD5, 0x0CD6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc   [2] KANNADA LENGTH MARK..KANNADA AI LENGTH MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CE2, 0x0CE3 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] KANNADA VOWEL SIGN VOCALIC L..KANNADA VOWEL SIGN VOCALIC LL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0CF3, 0x0CF3 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       KANNADA SIGN COMBINING ANUSVARA ABOVE RIGHT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D00, 0x0D01 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MALAYALAM SIGN COMBINING ANUSVARA ABOVE..MALAYALAM SIGN CANDRABINDU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D02, 0x0D03 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] MALAYALAM SIGN ANUSVARA..MALAYALAM SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D3B, 0x0D3C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MALAYALAM SIGN VERTICAL BAR VIRAMA..MALAYALAM SIGN CIRCULAR VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D3E, 0x0D3E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       MALAYALAM VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D3F, 0x0D40 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] MALAYALAM VOWEL SIGN I..MALAYALAM VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D41, 0x0D44 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] MALAYALAM VOWEL SIGN U..MALAYALAM VOWEL SIGN VOCALIC RR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D46, 0x0D48 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] MALAYALAM VOWEL SIGN E..MALAYALAM VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D4A, 0x0D4C }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] MALAYALAM VOWEL SIGN O..MALAYALAM VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D4D, 0x0D4D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MALAYALAM SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D4E, 0x0D4E }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Lo       MALAYALAM LETTER DOT REPH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D57, 0x0D57 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       MALAYALAM AU LENGTH MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D62, 0x0D63 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MALAYALAM VOWEL SIGN VOCALIC L..MALAYALAM VOWEL SIGN VOCALIC LL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D81, 0x0D81 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SINHALA SIGN CANDRABINDU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0D82, 0x0D83 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] SINHALA SIGN ANUSVARAYA..SINHALA SIGN VISARGAYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0DCA, 0x0DCA }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SINHALA SIGN AL-LAKUNA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0DCF, 0x0DCF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       SINHALA VOWEL SIGN AELA-PILLA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0DD0, 0x0DD1 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] SINHALA VOWEL SIGN KETTI AEDA-PILLA..SINHALA VOWEL SIGN DIGA AEDA-PILLA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0DD2, 0x0DD4 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] SINHALA VOWEL SIGN KETTI IS-PILLA..SINHALA VOWEL SIGN KETTI PAA-PILLA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0DD6, 0x0DD6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SINHALA VOWEL SIGN DIGA PAA-PILLA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0DD8, 0x0DDE }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [7] SINHALA VOWEL SIGN GAETTA-PILLA..SINHALA VOWEL SIGN KOMBUVA HAA GAYANUKITTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0DDF, 0x0DDF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       SINHALA VOWEL SIGN GAYANUKITTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0DF2, 0x0DF3 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] SINHALA VOWEL SIGN DIGA GAETTA-PILLA..SINHALA VOWEL SIGN DIGA GAYANUKITTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0E31, 0x0E31 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       THAI CHARACTER MAI HAN-AKAT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0E33, 0x0E33 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Lo       THAI CHARACTER SARA AM
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0E34, 0x0E3A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] THAI CHARACTER SARA I..THAI CHARACTER PHINTHU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0E47, 0x0E4E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [8] THAI CHARACTER MAITAIKHU..THAI CHARACTER YAMAKKAN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0EB1, 0x0EB1 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       LAO VOWEL SIGN MAI KAN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0EB3, 0x0EB3 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Lo       LAO VOWEL SIGN AM
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0EB4, 0x0EBC }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [9] LAO VOWEL SIGN I..LAO SEMIVOWEL SIGN LO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0EC8, 0x0ECE }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] LAO TONE MAI EK..LAO YAMAKKAN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0F18, 0x0F19 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] TIBETAN ASTROLOGICAL SIGN -KHYUD PA..TIBETAN ASTROLOGICAL SIGN SDONG TSHUGS
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0F35, 0x0F35 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TIBETAN MARK NGAS BZUNG NYI ZLA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0F37, 0x0F37 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TIBETAN MARK NGAS BZUNG SGOR RTAGS
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0F39, 0x0F39 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TIBETAN MARK TSA -PHRU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0F3E, 0x0F3F }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] TIBETAN SIGN YAR TSHES..TIBETAN SIGN MAR TSHES
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0F71, 0x0F7E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [14] TIBETAN VOWEL SIGN AA..TIBETAN SIGN RJES SU NGA RO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0F7F, 0x0F7F }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       TIBETAN SIGN RNAM BCAD
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0F80, 0x0F84 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] TIBETAN VOWEL SIGN REVERSED I..TIBETAN MARK HALANTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0F86, 0x0F87 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] TIBETAN SIGN LCI RTAGS..TIBETAN SIGN YANG RTAGS
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0F8D, 0x0F97 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [11] TIBETAN SUBJOINED SIGN LCE TSA CAN..TIBETAN SUBJOINED LETTER JA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0F99, 0x0FBC }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [36] TIBETAN SUBJOINED LETTER NYA..TIBETAN SUBJOINED LETTER FIXED-FORM RA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x0FC6, 0x0FC6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TIBETAN SYMBOL PADMA GDAN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x102D, 0x1030 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] MYANMAR VOWEL SIGN I..MYANMAR VOWEL SIGN UU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1031, 0x1031 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       MYANMAR VOWEL SIGN E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1032, 0x1037 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [6] MYANMAR VOWEL SIGN AI..MYANMAR SIGN DOT BELOW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1039, 0x103A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MYANMAR SIGN VIRAMA..MYANMAR SIGN ASAT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x103B, 0x103C }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] MYANMAR CONSONANT SIGN MEDIAL YA..MYANMAR CONSONANT SIGN MEDIAL RA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x103D, 0x103E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MYANMAR CONSONANT SIGN MEDIAL WA..MYANMAR CONSONANT SIGN MEDIAL HA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1056, 0x1057 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] MYANMAR VOWEL SIGN VOCALIC R..MYANMAR VOWEL SIGN VOCALIC RR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1058, 0x1059 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MYANMAR VOWEL SIGN VOCALIC L..MYANMAR VOWEL SIGN VOCALIC LL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x105E, 0x1060 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] MYANMAR CONSONANT SIGN MON MEDIAL NA..MYANMAR CONSONANT SIGN MON MEDIAL LA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1071, 0x1074 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] MYANMAR VOWEL SIGN GEBA KAREN I..MYANMAR VOWEL SIGN KAYAH EE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1082, 0x1082 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MYANMAR CONSONANT SIGN SHAN MEDIAL WA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1084, 0x1084 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       MYANMAR VOWEL SIGN SHAN E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1085, 0x1086 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MYANMAR VOWEL SIGN SHAN E ABOVE..MYANMAR VOWEL SIGN SHAN FINAL Y
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x108D, 0x108D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MYANMAR SIGN SHAN COUNCIL EMPHATIC TONE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x109D, 0x109D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MYANMAR VOWEL SIGN AITON AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1100, 0x115F }, ::unicode::Grapheme_Cluster_Break::L }, // Lo  [96] HANGUL CHOSEONG KIYEOK..HANGUL CHOSEONG FILLER
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1160, 0x11A7 }, ::unicode::Grapheme_Cluster_Break::V }, // Lo  [72] HANGUL JUNGSEONG FILLER..HANGUL JUNGSEONG O-YAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A8, 0x11FF }, ::unicode::Grapheme_Cluster_Break::T }, // Lo  [88] HANGUL JONGSEONG KIYEOK..HANGUL JONGSEONG SSANGNIEUN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x135D, 0x135F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] ETHIOPIC COMBINING GEMINATION AND VOWEL LENGTH MARK..ETHIOPIC COMBINING GEMINATION MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1712, 0x1714 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] TAGALOG VOWEL SIGN I..TAGALOG SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1715, 0x1715 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       TAGALOG SIGN PAMUDPOD
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1732, 0x1733 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] HANUNOO VOWEL SIGN I..HANUNOO VOWEL SIGN U
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1734, 0x1734 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       HANUNOO SIGN PAMUDPOD
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1752, 0x1753 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] BUHID VOWEL SIGN I..BUHID VOWEL SIGN U
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1772, 0x1773 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] TAGBANWA VOWEL SIGN I..TAGBANWA VOWEL SIGN U
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x17B4, 0x17B5 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] KHMER VOWEL INHERENT AQ..KHMER VOWEL INHERENT AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x17B6, 0x17B6 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       KHMER VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x17B7, 0x17BD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] KHMER VOWEL SIGN I..KHMER VOWEL SIGN UA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x17BE, 0x17C5 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [8] KHMER VOWEL SIGN OE..KHMER VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x17C6, 0x17C6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KHMER SIGN NIKAHIT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x17C7, 0x17C8 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] KHMER SIGN REAHMUK..KHMER SIGN YUUKALEAPINTU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x17C9, 0x17D3 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [11] KHMER SIGN MUUSIKATOAN..KHMER SIGN BATHAMASAT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x17DD, 0x17DD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KHMER SIGN ATTHACAN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x180B, 0x180D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] MONGOLIAN FREE VARIATION SELECTOR ONE..MONGOLIAN FREE VARIATION SELECTOR THREE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x180E, 0x180E }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf       MONGOLIAN VOWEL SEPARATOR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x180F, 0x180F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MONGOLIAN FREE VARIATION SELECTOR FOUR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1885, 0x1886 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MONGOLIAN LETTER ALI GALI BALUDA..MONGOLIAN LETTER ALI GALI THREE BALUDA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x18A9, 0x18A9 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MONGOLIAN LETTER ALI GALI DAGALGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1920, 0x1922 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] LIMBU VOWEL SIGN A..LIMBU VOWEL SIGN U
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1923, 0x1926 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [4] LIMBU VOWEL SIGN EE..LIMBU VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1927, 0x1928 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] LIMBU VOWEL SIGN E..LIMBU VOWEL SIGN O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1929, 0x192B }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] LIMBU SUBJOINED LETTER YA..LIMBU SUBJOINED LETTER WA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1930, 0x1931 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] LIMBU SMALL LETTER KA..LIMBU SMALL LETTER NGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1932, 0x1932 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       LIMBU SMALL LETTER ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1933, 0x1938 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [6] LIMBU SMALL LETTER TA..LIMBU SMALL LETTER LA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1939, 0x193B }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] LIMBU SIGN MUKPHRENG..LIMBU SIGN SA-I
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A17, 0x1A18 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] BUGINESE VOWEL SIGN I..BUGINESE VOWEL SIGN U
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A19, 0x1A1A }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] BUGINESE VOWEL SIGN E..BUGINESE VOWEL SIGN O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A1B, 0x1A1B }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BUGINESE VOWEL SIGN AE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A55, 0x1A55 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       TAI THAM CONSONANT SIGN MEDIAL RA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A56, 0x1A56 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAI THAM CONSONANT SIGN MEDIAL LA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A57, 0x1A57 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       TAI THAM CONSONANT SIGN LA TANG LAI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A58, 0x1A5E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] TAI THAM SIGN MAI KANG LAI..TAI THAM CONSONANT SIGN SA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A60, 0x1A60 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAI THAM SIGN SAKOT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A62, 0x1A62 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAI THAM VOWEL SIGN MAI SAT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A65, 0x1A6C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [8] TAI THAM VOWEL SIGN I..TAI THAM VOWEL SIGN OA BELOW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A6D, 0x1A72 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [6] TAI THAM VOWEL SIGN OY..TAI THAM VOWEL SIGN THAM AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A73, 0x1A7C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [10] TAI THAM VOWEL SIGN OA ABOVE..TAI THAM SIGN KHUEN-LUE KARAN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1A7F, 0x1A7F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAI THAM COMBINING CRYPTOGRAMMIC DOT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1AB0, 0x1ABD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [14] COMBINING DOUBLED CIRCUMFLEX ACCENT..COMBINING PARENTHESES BELOW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1ABE, 0x1ABE }, ::unicode::Grapheme_Cluster_Break::Extend }, // Me       COMBINING PARENTHESES OVERLAY
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1ABF, 0x1ADD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [31] COMBINING LATIN SMALL LETTER W BELOW..COMBINING DOT-AND-RING BELOW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1AE0, 0x1AEB }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [12] COMBINING LEFT TACK ABOVE..COMBINING DOUBLE RIGHTWARDS ARROW ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B00, 0x1B03 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] BALINESE SIGN ULU RICEM..BALINESE SIGN SURANG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B04, 0x1B04 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       BALINESE SIGN BISAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B34, 0x1B34 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BALINESE SIGN REREKAN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B35, 0x1B35 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       BALINESE VOWEL SIGN TEDUNG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B36, 0x1B3A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] BALINESE VOWEL SIGN ULU..BALINESE VOWEL SIGN RA REPA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B3B, 0x1B3B }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       BALINESE VOWEL SIGN RA REPA TEDUNG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B3C, 0x1B3C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BALINESE VOWEL SIGN LA LENGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B3D, 0x1B3D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       BALINESE VOWEL SIGN LA LENGA TEDUNG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B3E, 0x1B41 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [4] BALINESE VOWEL SIGN TALING..BALINESE VOWEL SIGN TALING REPA TEDUNG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B42, 0x1B42 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BALINESE VOWEL SIGN PEPET
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B43, 0x1B44 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc   [2] BALINESE VOWEL SIGN PEPET TEDUNG..BALINESE ADEG ADEG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B6B, 0x1B73 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [9] BALINESE MUSICAL SYMBOL COMBINING TEGEH..BALINESE MUSICAL SYMBOL COMBINING GONG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B80, 0x1B81 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] SUNDANESE SIGN PANYECEK..SUNDANESE SIGN PANGLAYAR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1B82, 0x1B82 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       SUNDANESE SIGN PANGWISAD
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BA1, 0x1BA1 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       SUNDANESE CONSONANT SIGN PAMINGKAL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BA2, 0x1BA5 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] SUNDANESE CONSONANT SIGN PANYAKRA..SUNDANESE VOWEL SIGN PANYUKU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BA6, 0x1BA7 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] SUNDANESE VOWEL SIGN PANAELAENG..SUNDANESE VOWEL SIGN PANOLONG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BA8, 0x1BA9 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] SUNDANESE VOWEL SIGN PAMEPET..SUNDANESE VOWEL SIGN PANEULEUNG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BAA, 0x1BAA }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       SUNDANESE SIGN PAMAAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BAB, 0x1BAD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] SUNDANESE SIGN VIRAMA..SUNDANESE CONSONANT SIGN PASANGAN WA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BE6, 0x1BE6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BATAK SIGN TOMPI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BE7, 0x1BE7 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       BATAK VOWEL SIGN E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BE8, 0x1BE9 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] BATAK VOWEL SIGN PAKPAK E..BATAK VOWEL SIGN EE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BEA, 0x1BEC }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] BATAK VOWEL SIGN I..BATAK VOWEL SIGN O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BED, 0x1BED }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BATAK VOWEL SIGN KARO O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BEE, 0x1BEE }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       BATAK VOWEL SIGN U
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BEF, 0x1BF1 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] BATAK VOWEL SIGN U FOR SIMALUNGUN SA..BATAK CONSONANT SIGN H
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BF2, 0x1BF3 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc   [2] BATAK PANGOLAT..BATAK PANONGONAN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1C24, 0x1C2B }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [8] LEPCHA SUBJOINED LETTER YA..LEPCHA VOWEL SIGN UU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1C2C, 0x1C33 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [8] LEPCHA VOWEL SIGN E..LEPCHA CONSONANT SIGN T
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1C34, 0x1C35 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] LEPCHA CONSONANT SIGN NYIN-DO..LEPCHA CONSONANT SIGN KANG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1C36, 0x1C37 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] LEPCHA SIGN RAN..LEPCHA SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1CD0, 0x1CD2 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] VEDIC TONE KARSHANA..VEDIC TONE PRENKHA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1CD4, 0x1CE0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [13] VEDIC SIGN YAJURVEDIC MIDLINE SVARITA..VEDIC TONE RIGVEDIC KASHMIRI INDEPENDENT SVARITA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1CE1, 0x1CE1 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       VEDIC TONE ATHARVAVEDIC INDEPENDENT SVARITA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1CE2, 0x1CE8 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] VEDIC SIGN VISARGA SVARITA..VEDIC SIGN VISARGA ANUDATTA WITH TAIL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1CED, 0x1CED }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       VEDIC SIGN TIRYAK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1CF4, 0x1CF4 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       VEDIC TONE CANDRA ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1CF7, 0x1CF7 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       VEDIC SIGN ATIKRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1CF8, 0x1CF9 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] VEDIC TONE RING ABOVE..VEDIC TONE DOUBLE RING ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1DC0, 0x1DFF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [64] COMBINING DOTTED GRAVE ACCENT..COMBINING RIGHT ARROWHEAD AND DOWN ARROWHEAD BELOW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x200B, 0x200B }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf       ZERO WIDTH SPACE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x200C, 0x200C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Cf       ZERO WIDTH NON-JOINER
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x200D, 0x200D }, ::unicode::Grapheme_Cluster_Break::ZWJ }, // Cf       ZERO WIDTH JOINER
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x200E, 0x200F }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf   [2] LEFT-TO-RIGHT MARK..RIGHT-TO-LEFT MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x2028, 0x2028 }, ::unicode::Grapheme_Cluster_Break::Control }, // Zl       LINE SEPARATOR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x2029, 0x2029 }, ::unicode::Grapheme_Cluster_Break::Control }, // Zp       PARAGRAPH SEPARATOR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x202A, 0x202E }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf   [5] LEFT-TO-RIGHT EMBEDDING..RIGHT-TO-LEFT OVERRIDE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x2060, 0x2064 }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf   [5] WORD JOINER..INVISIBLE PLUS
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x2065, 0x2065 }, ::unicode::Grapheme_Cluster_Break::Control }, // Cn       <reserved-2065>
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x2066, 0x206F }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf  [10] LEFT-TO-RIGHT ISOLATE..NOMINAL DIGIT SHAPES
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x20D0, 0x20DC }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [13] COMBINING LEFT HARPOON ABOVE..COMBINING FOUR DOTS ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x20DD, 0x20E0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Me   [4] COMBINING ENCLOSING CIRCLE..COMBINING ENCLOSING CIRCLE BACKSLASH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x20E1, 0x20E1 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       COMBINING LEFT RIGHT ARROW ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x20E2, 0x20E4 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Me   [3] COMBINING ENCLOSING SCREEN..COMBINING ENCLOSING UPWARD POINTING TRIANGLE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x20E5, 0x20F0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [12] COMBINING REVERSE SOLIDUS OVERLAY..COMBINING ASTERISK ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x2CEF, 0x2CF1 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] COPTIC COMBINING NI ABOVE..COPTIC COMBINING SPIRITUS LENIS
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x2D7F, 0x2D7F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TIFINAGH CONSONANT JOINER
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x2DE0, 0x2DFF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [32] COMBINING CYRILLIC LETTER BE..COMBINING CYRILLIC LETTER IOTIFIED BIG YUS
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x302A, 0x302D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] IDEOGRAPHIC LEVEL TONE MARK..IDEOGRAPHIC ENTERING TONE MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x302E, 0x302F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc   [2] HANGUL SINGLE DOT TONE MARK..HANGUL DOUBLE DOT TONE MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x3099, 0x309A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK..COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA66F, 0xA66F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       COMBINING CYRILLIC VZMET
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA670, 0xA672 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Me   [3] COMBINING CYRILLIC TEN MILLIONS SIGN..COMBINING CYRILLIC THOUSAND MILLIONS SIGN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA674, 0xA67D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [10] COMBINING CYRILLIC LETTER UKRAINIAN IE..COMBINING CYRILLIC PAYEROK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA69E, 0xA69F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] COMBINING CYRILLIC LETTER EF..COMBINING CYRILLIC LETTER IOTIFIED E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA6F0, 0xA6F1 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] BAMUM COMBINING MARK KOQNDON..BAMUM COMBINING MARK TUKWENTIS
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA802, 0xA802 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SYLOTI NAGRI SIGN DVISVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA806, 0xA806 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SYLOTI NAGRI SIGN HASANTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA80B, 0xA80B }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SYLOTI NAGRI SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA823, 0xA824 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] SYLOTI NAGRI VOWEL SIGN A..SYLOTI NAGRI VOWEL SIGN I
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA825, 0xA826 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] SYLOTI NAGRI VOWEL SIGN U..SYLOTI NAGRI VOWEL SIGN E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA827, 0xA827 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       SYLOTI NAGRI VOWEL SIGN OO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA82C, 0xA82C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SYLOTI NAGRI SIGN ALTERNATE HASANTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA880, 0xA881 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] SAURASHTRA SIGN ANUSVARA..SAURASHTRA SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA8B4, 0xA8C3 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc  [16] SAURASHTRA CONSONANT SIGN HAARU..SAURASHTRA VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA8C4, 0xA8C5 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] SAURASHTRA SIGN VIRAMA..SAURASHTRA SIGN CANDRABINDU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA8E0, 0xA8F1 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [18] COMBINING DEVANAGARI DIGIT ZERO..COMBINING DEVANAGARI SIGN AVAGRAHA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA8FF, 0xA8FF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       DEVANAGARI VOWEL SIGN AY
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA926, 0xA92D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [8] KAYAH LI VOWEL UE..KAYAH LI TONE CALYA PLOPHU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA947, 0xA951 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [11] REJANG VOWEL SIGN I..REJANG CONSONANT SIGN R
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA952, 0xA952 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       REJANG CONSONANT SIGN H
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA953, 0xA953 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       REJANG VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA960, 0xA97C }, ::unicode::Grapheme_Cluster_Break::L }, // Lo  [29] HANGUL CHOSEONG TIKEUT-MIEUM..HANGUL CHOSEONG SSANGYEORINHIEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA980, 0xA982 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] JAVANESE SIGN PANYANGGA..JAVANESE SIGN LAYAR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA983, 0xA983 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       JAVANESE SIGN WIGNYAN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA9B3, 0xA9B3 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       JAVANESE SIGN CECAK TELU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA9B4, 0xA9B5 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] JAVANESE VOWEL SIGN TARUNG..JAVANESE VOWEL SIGN TOLONG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA9B6, 0xA9B9 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] JAVANESE VOWEL SIGN WULU..JAVANESE VOWEL SIGN SUKU MENDUT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA9BA, 0xA9BB }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] JAVANESE VOWEL SIGN TALING..JAVANESE VOWEL SIGN DIRGA MURE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA9BC, 0xA9BD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] JAVANESE VOWEL SIGN PEPET..JAVANESE CONSONANT SIGN KERET
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA9BE, 0xA9BF }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] JAVANESE CONSONANT SIGN PENGKAL..JAVANESE CONSONANT SIGN CAKRA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA9C0, 0xA9C0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       JAVANESE PANGKON
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xA9E5, 0xA9E5 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MYANMAR SIGN SHAN SAW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAA29, 0xAA2E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [6] CHAM VOWEL SIGN AA..CHAM VOWEL SIGN OE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAA2F, 0xAA30 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] CHAM VOWEL SIGN O..CHAM VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAA31, 0xAA32 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] CHAM VOWEL SIGN AU..CHAM VOWEL SIGN UE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAA33, 0xAA34 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] CHAM CONSONANT SIGN YA..CHAM CONSONANT SIGN RA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAA35, 0xAA36 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] CHAM CONSONANT SIGN LA..CHAM CONSONANT SIGN WA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAA43, 0xAA43 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       CHAM CONSONANT SIGN FINAL NG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAA4C, 0xAA4C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       CHAM CONSONANT SIGN FINAL M
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAA4D, 0xAA4D }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       CHAM CONSONANT SIGN FINAL H
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAA7C, 0xAA7C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MYANMAR SIGN TAI LAING TONE-2
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAAB0, 0xAAB0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAI VIET MAI KANG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAAB2, 0xAAB4 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] TAI VIET VOWEL I..TAI VIET VOWEL U
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAAB7, 0xAAB8 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] TAI VIET MAI KHIT..TAI VIET VOWEL IA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAABE, 0xAABF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] TAI VIET VOWEL AM..TAI VIET TONE MAI EK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAAC1, 0xAAC1 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAI VIET TONE MAI THO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAAEB, 0xAAEB }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       MEETEI MAYEK VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAAEC, 0xAAED }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MEETEI MAYEK VOWEL SIGN UU..MEETEI MAYEK VOWEL SIGN AAI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAAEE, 0xAAEF }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] MEETEI MAYEK VOWEL SIGN AU..MEETEI MAYEK VOWEL SIGN AAU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAAF5, 0xAAF5 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       MEETEI MAYEK VOWEL SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAAF6, 0xAAF6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MEETEI MAYEK VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xABE3, 0xABE4 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] MEETEI MAYEK VOWEL SIGN ONAP..MEETEI MAYEK VOWEL SIGN INAP
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xABE5, 0xABE5 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MEETEI MAYEK VOWEL SIGN ANAP
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xABE6, 0xABE7 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] MEETEI MAYEK VOWEL SIGN YENAP..MEETEI MAYEK VOWEL SIGN SOUNAP
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xABE8, 0xABE8 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MEETEI MAYEK VOWEL SIGN UNAP
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xABE9, 0xABEA }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] MEETEI MAYEK VOWEL SIGN CHEINAP..MEETEI MAYEK VOWEL SIGN NUNG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xABEC, 0xABEC }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       MEETEI MAYEK LUM IYEK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xABED, 0xABED }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MEETEI MAYEK APUN IYEK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAC00, 0xAC00 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAC01, 0xAC1B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GAG..HANGUL SYLLABLE GAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAC1C, 0xAC1C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAC1D, 0xAC37 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GAEG..HANGUL SYLLABLE GAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAC38, 0xAC38 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAC39, 0xAC53 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GYAG..HANGUL SYLLABLE GYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAC54, 0xAC54 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAC55, 0xAC6F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GYAEG..HANGUL SYLLABLE GYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAC70, 0xAC70 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAC71, 0xAC8B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GEOG..HANGUL SYLLABLE GEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAC8C, 0xAC8C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAC8D, 0xACA7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GEG..HANGUL SYLLABLE GEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xACA8, 0xACA8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xACA9, 0xACC3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GYEOG..HANGUL SYLLABLE GYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xACC4, 0xACC4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xACC5, 0xACDF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GYEG..HANGUL SYLLABLE GYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xACE0, 0xACE0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xACE1, 0xACFB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GOG..HANGUL SYLLABLE GOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xACFC, 0xACFC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xACFD, 0xAD17 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GWAG..HANGUL SYLLABLE GWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAD18, 0xAD18 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAD19, 0xAD33 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GWAEG..HANGUL SYLLABLE GWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAD34, 0xAD34 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAD35, 0xAD4F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GOEG..HANGUL SYLLABLE GOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAD50, 0xAD50 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAD51, 0xAD6B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GYOG..HANGUL SYLLABLE GYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAD6C, 0xAD6C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAD6D, 0xAD87 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GUG..HANGUL SYLLABLE GUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAD88, 0xAD88 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAD89, 0xADA3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GWEOG..HANGUL SYLLABLE GWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xADA4, 0xADA4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xADA5, 0xADBF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GWEG..HANGUL SYLLABLE GWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xADC0, 0xADC0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xADC1, 0xADDB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GWIG..HANGUL SYLLABLE GWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xADDC, 0xADDC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xADDD, 0xADF7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GYUG..HANGUL SYLLABLE GYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xADF8, 0xADF8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xADF9, 0xAE13 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GEUG..HANGUL SYLLABLE GEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAE14, 0xAE14 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAE15, 0xAE2F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GYIG..HANGUL SYLLABLE GYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAE30, 0xAE30 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAE31, 0xAE4B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GIG..HANGUL SYLLABLE GIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAE4C, 0xAE4C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAE4D, 0xAE67 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGAG..HANGUL SYLLABLE GGAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAE68, 0xAE68 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAE69, 0xAE83 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGAEG..HANGUL SYLLABLE GGAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAE84, 0xAE84 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAE85, 0xAE9F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGYAG..HANGUL SYLLABLE GGYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAEA0, 0xAEA0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAEA1, 0xAEBB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGYAEG..HANGUL SYLLABLE GGYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAEBC, 0xAEBC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAEBD, 0xAED7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGEOG..HANGUL SYLLABLE GGEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAED8, 0xAED8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAED9, 0xAEF3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGEG..HANGUL SYLLABLE GGEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAEF4, 0xAEF4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAEF5, 0xAF0F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGYEOG..HANGUL SYLLABLE GGYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAF10, 0xAF10 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAF11, 0xAF2B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGYEG..HANGUL SYLLABLE GGYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAF2C, 0xAF2C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAF2D, 0xAF47 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGOG..HANGUL SYLLABLE GGOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAF48, 0xAF48 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAF49, 0xAF63 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGWAG..HANGUL SYLLABLE GGWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAF64, 0xAF64 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAF65, 0xAF7F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGWAEG..HANGUL SYLLABLE GGWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAF80, 0xAF80 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAF81, 0xAF9B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGOEG..HANGUL SYLLABLE GGOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAF9C, 0xAF9C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAF9D, 0xAFB7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGYOG..HANGUL SYLLABLE GGYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAFB8, 0xAFB8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAFB9, 0xAFD3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGUG..HANGUL SYLLABLE GGUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAFD4, 0xAFD4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAFD5, 0xAFEF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGWEOG..HANGUL SYLLABLE GGWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAFF0, 0xAFF0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xAFF1, 0xB00B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGWEG..HANGUL SYLLABLE GGWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB00C, 0xB00C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB00D, 0xB027 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGWIG..HANGUL SYLLABLE GGWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB028, 0xB028 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB029, 0xB043 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGYUG..HANGUL SYLLABLE GGYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB044, 0xB044 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB045, 0xB05F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGEUG..HANGUL SYLLABLE GGEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB060, 0xB060 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB061, 0xB07B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGYIG..HANGUL SYLLABLE GGYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB07C, 0xB07C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE GGI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB07D, 0xB097 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE GGIG..HANGUL SYLLABLE GGIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB098, 0xB098 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB099, 0xB0B3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NAG..HANGUL SYLLABLE NAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB0B4, 0xB0B4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB0B5, 0xB0CF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NAEG..HANGUL SYLLABLE NAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB0D0, 0xB0D0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB0D1, 0xB0EB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NYAG..HANGUL SYLLABLE NYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB0EC, 0xB0EC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB0ED, 0xB107 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NYAEG..HANGUL SYLLABLE NYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB108, 0xB108 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB109, 0xB123 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NEOG..HANGUL SYLLABLE NEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB124, 0xB124 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB125, 0xB13F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NEG..HANGUL SYLLABLE NEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB140, 0xB140 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB141, 0xB15B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NYEOG..HANGUL SYLLABLE NYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB15C, 0xB15C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB15D, 0xB177 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NYEG..HANGUL SYLLABLE NYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB178, 0xB178 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB179, 0xB193 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NOG..HANGUL SYLLABLE NOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB194, 0xB194 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB195, 0xB1AF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NWAG..HANGUL SYLLABLE NWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB1B0, 0xB1B0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB1B1, 0xB1CB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NWAEG..HANGUL SYLLABLE NWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB1CC, 0xB1CC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB1CD, 0xB1E7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NOEG..HANGUL SYLLABLE NOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB1E8, 0xB1E8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB1E9, 0xB203 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NYOG..HANGUL SYLLABLE NYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB204, 0xB204 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB205, 0xB21F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NUG..HANGUL SYLLABLE NUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB220, 0xB220 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB221, 0xB23B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NWEOG..HANGUL SYLLABLE NWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB23C, 0xB23C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB23D, 0xB257 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NWEG..HANGUL SYLLABLE NWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB258, 0xB258 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB259, 0xB273 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NWIG..HANGUL SYLLABLE NWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB274, 0xB274 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB275, 0xB28F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NYUG..HANGUL SYLLABLE NYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB290, 0xB290 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB291, 0xB2AB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NEUG..HANGUL SYLLABLE NEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB2AC, 0xB2AC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB2AD, 0xB2C7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NYIG..HANGUL SYLLABLE NYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB2C8, 0xB2C8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE NI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB2C9, 0xB2E3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE NIG..HANGUL SYLLABLE NIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB2E4, 0xB2E4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB2E5, 0xB2FF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DAG..HANGUL SYLLABLE DAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB300, 0xB300 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB301, 0xB31B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DAEG..HANGUL SYLLABLE DAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB31C, 0xB31C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB31D, 0xB337 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DYAG..HANGUL SYLLABLE DYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB338, 0xB338 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB339, 0xB353 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DYAEG..HANGUL SYLLABLE DYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB354, 0xB354 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB355, 0xB36F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DEOG..HANGUL SYLLABLE DEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB370, 0xB370 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB371, 0xB38B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DEG..HANGUL SYLLABLE DEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB38C, 0xB38C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB38D, 0xB3A7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DYEOG..HANGUL SYLLABLE DYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB3A8, 0xB3A8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB3A9, 0xB3C3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DYEG..HANGUL SYLLABLE DYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB3C4, 0xB3C4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB3C5, 0xB3DF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DOG..HANGUL SYLLABLE DOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB3E0, 0xB3E0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB3E1, 0xB3FB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DWAG..HANGUL SYLLABLE DWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB3FC, 0xB3FC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB3FD, 0xB417 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DWAEG..HANGUL SYLLABLE DWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB418, 0xB418 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB419, 0xB433 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DOEG..HANGUL SYLLABLE DOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB434, 0xB434 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB435, 0xB44F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DYOG..HANGUL SYLLABLE DYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB450, 0xB450 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB451, 0xB46B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DUG..HANGUL SYLLABLE DUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB46C, 0xB46C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB46D, 0xB487 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DWEOG..HANGUL SYLLABLE DWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB488, 0xB488 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB489, 0xB4A3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DWEG..HANGUL SYLLABLE DWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB4A4, 0xB4A4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB4A5, 0xB4BF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DWIG..HANGUL SYLLABLE DWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB4C0, 0xB4C0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB4C1, 0xB4DB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DYUG..HANGUL SYLLABLE DYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB4DC, 0xB4DC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB4DD, 0xB4F7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DEUG..HANGUL SYLLABLE DEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB4F8, 0xB4F8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB4F9, 0xB513 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DYIG..HANGUL SYLLABLE DYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB514, 0xB514 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB515, 0xB52F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DIG..HANGUL SYLLABLE DIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB530, 0xB530 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB531, 0xB54B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDAG..HANGUL SYLLABLE DDAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB54C, 0xB54C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB54D, 0xB567 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDAEG..HANGUL SYLLABLE DDAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB568, 0xB568 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB569, 0xB583 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDYAG..HANGUL SYLLABLE DDYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB584, 0xB584 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB585, 0xB59F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDYAEG..HANGUL SYLLABLE DDYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB5A0, 0xB5A0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB5A1, 0xB5BB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDEOG..HANGUL SYLLABLE DDEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB5BC, 0xB5BC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB5BD, 0xB5D7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDEG..HANGUL SYLLABLE DDEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB5D8, 0xB5D8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB5D9, 0xB5F3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDYEOG..HANGUL SYLLABLE DDYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB5F4, 0xB5F4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB5F5, 0xB60F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDYEG..HANGUL SYLLABLE DDYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB610, 0xB610 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB611, 0xB62B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDOG..HANGUL SYLLABLE DDOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB62C, 0xB62C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB62D, 0xB647 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDWAG..HANGUL SYLLABLE DDWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB648, 0xB648 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB649, 0xB663 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDWAEG..HANGUL SYLLABLE DDWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB664, 0xB664 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB665, 0xB67F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDOEG..HANGUL SYLLABLE DDOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB680, 0xB680 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB681, 0xB69B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDYOG..HANGUL SYLLABLE DDYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB69C, 0xB69C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB69D, 0xB6B7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDUG..HANGUL SYLLABLE DDUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB6B8, 0xB6B8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB6B9, 0xB6D3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDWEOG..HANGUL SYLLABLE DDWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB6D4, 0xB6D4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB6D5, 0xB6EF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDWEG..HANGUL SYLLABLE DDWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB6F0, 0xB6F0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB6F1, 0xB70B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDWIG..HANGUL SYLLABLE DDWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB70C, 0xB70C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB70D, 0xB727 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDYUG..HANGUL SYLLABLE DDYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB728, 0xB728 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB729, 0xB743 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDEUG..HANGUL SYLLABLE DDEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB744, 0xB744 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB745, 0xB75F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDYIG..HANGUL SYLLABLE DDYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB760, 0xB760 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE DDI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB761, 0xB77B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE DDIG..HANGUL SYLLABLE DDIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB77C, 0xB77C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB77D, 0xB797 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RAG..HANGUL SYLLABLE RAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB798, 0xB798 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB799, 0xB7B3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RAEG..HANGUL SYLLABLE RAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB7B4, 0xB7B4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB7B5, 0xB7CF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RYAG..HANGUL SYLLABLE RYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB7D0, 0xB7D0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB7D1, 0xB7EB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RYAEG..HANGUL SYLLABLE RYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB7EC, 0xB7EC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE REO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB7ED, 0xB807 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE REOG..HANGUL SYLLABLE REOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB808, 0xB808 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB809, 0xB823 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE REG..HANGUL SYLLABLE REH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB824, 0xB824 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB825, 0xB83F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RYEOG..HANGUL SYLLABLE RYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB840, 0xB840 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB841, 0xB85B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RYEG..HANGUL SYLLABLE RYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB85C, 0xB85C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB85D, 0xB877 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE ROG..HANGUL SYLLABLE ROH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB878, 0xB878 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB879, 0xB893 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RWAG..HANGUL SYLLABLE RWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB894, 0xB894 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB895, 0xB8AF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RWAEG..HANGUL SYLLABLE RWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB8B0, 0xB8B0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE ROE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB8B1, 0xB8CB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE ROEG..HANGUL SYLLABLE ROEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB8CC, 0xB8CC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB8CD, 0xB8E7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RYOG..HANGUL SYLLABLE RYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB8E8, 0xB8E8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB8E9, 0xB903 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RUG..HANGUL SYLLABLE RUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB904, 0xB904 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB905, 0xB91F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RWEOG..HANGUL SYLLABLE RWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB920, 0xB920 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB921, 0xB93B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RWEG..HANGUL SYLLABLE RWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB93C, 0xB93C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB93D, 0xB957 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RWIG..HANGUL SYLLABLE RWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB958, 0xB958 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB959, 0xB973 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RYUG..HANGUL SYLLABLE RYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB974, 0xB974 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE REU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB975, 0xB98F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE REUG..HANGUL SYLLABLE REUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB990, 0xB990 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB991, 0xB9AB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RYIG..HANGUL SYLLABLE RYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB9AC, 0xB9AC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE RI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB9AD, 0xB9C7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE RIG..HANGUL SYLLABLE RIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB9C8, 0xB9C8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB9C9, 0xB9E3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MAG..HANGUL SYLLABLE MAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB9E4, 0xB9E4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xB9E5, 0xB9FF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MAEG..HANGUL SYLLABLE MAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBA00, 0xBA00 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBA01, 0xBA1B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MYAG..HANGUL SYLLABLE MYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBA1C, 0xBA1C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBA1D, 0xBA37 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MYAEG..HANGUL SYLLABLE MYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBA38, 0xBA38 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBA39, 0xBA53 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MEOG..HANGUL SYLLABLE MEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBA54, 0xBA54 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE ME
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBA55, 0xBA6F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MEG..HANGUL SYLLABLE MEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBA70, 0xBA70 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBA71, 0xBA8B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MYEOG..HANGUL SYLLABLE MYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBA8C, 0xBA8C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBA8D, 0xBAA7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MYEG..HANGUL SYLLABLE MYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBAA8, 0xBAA8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBAA9, 0xBAC3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MOG..HANGUL SYLLABLE MOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBAC4, 0xBAC4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBAC5, 0xBADF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MWAG..HANGUL SYLLABLE MWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBAE0, 0xBAE0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBAE1, 0xBAFB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MWAEG..HANGUL SYLLABLE MWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBAFC, 0xBAFC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBAFD, 0xBB17 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MOEG..HANGUL SYLLABLE MOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBB18, 0xBB18 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBB19, 0xBB33 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MYOG..HANGUL SYLLABLE MYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBB34, 0xBB34 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBB35, 0xBB4F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MUG..HANGUL SYLLABLE MUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBB50, 0xBB50 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBB51, 0xBB6B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MWEOG..HANGUL SYLLABLE MWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBB6C, 0xBB6C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBB6D, 0xBB87 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MWEG..HANGUL SYLLABLE MWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBB88, 0xBB88 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBB89, 0xBBA3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MWIG..HANGUL SYLLABLE MWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBBA4, 0xBBA4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBBA5, 0xBBBF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MYUG..HANGUL SYLLABLE MYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBBC0, 0xBBC0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBBC1, 0xBBDB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MEUG..HANGUL SYLLABLE MEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBBDC, 0xBBDC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBBDD, 0xBBF7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MYIG..HANGUL SYLLABLE MYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBBF8, 0xBBF8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE MI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBBF9, 0xBC13 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE MIG..HANGUL SYLLABLE MIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBC14, 0xBC14 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBC15, 0xBC2F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BAG..HANGUL SYLLABLE BAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBC30, 0xBC30 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBC31, 0xBC4B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BAEG..HANGUL SYLLABLE BAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBC4C, 0xBC4C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBC4D, 0xBC67 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BYAG..HANGUL SYLLABLE BYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBC68, 0xBC68 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBC69, 0xBC83 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BYAEG..HANGUL SYLLABLE BYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBC84, 0xBC84 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBC85, 0xBC9F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BEOG..HANGUL SYLLABLE BEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBCA0, 0xBCA0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBCA1, 0xBCBB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BEG..HANGUL SYLLABLE BEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBCBC, 0xBCBC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBCBD, 0xBCD7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BYEOG..HANGUL SYLLABLE BYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBCD8, 0xBCD8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBCD9, 0xBCF3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BYEG..HANGUL SYLLABLE BYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBCF4, 0xBCF4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBCF5, 0xBD0F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BOG..HANGUL SYLLABLE BOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBD10, 0xBD10 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBD11, 0xBD2B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BWAG..HANGUL SYLLABLE BWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBD2C, 0xBD2C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBD2D, 0xBD47 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BWAEG..HANGUL SYLLABLE BWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBD48, 0xBD48 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBD49, 0xBD63 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BOEG..HANGUL SYLLABLE BOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBD64, 0xBD64 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBD65, 0xBD7F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BYOG..HANGUL SYLLABLE BYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBD80, 0xBD80 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBD81, 0xBD9B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BUG..HANGUL SYLLABLE BUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBD9C, 0xBD9C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBD9D, 0xBDB7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BWEOG..HANGUL SYLLABLE BWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBDB8, 0xBDB8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBDB9, 0xBDD3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BWEG..HANGUL SYLLABLE BWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBDD4, 0xBDD4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBDD5, 0xBDEF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BWIG..HANGUL SYLLABLE BWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBDF0, 0xBDF0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBDF1, 0xBE0B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BYUG..HANGUL SYLLABLE BYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBE0C, 0xBE0C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBE0D, 0xBE27 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BEUG..HANGUL SYLLABLE BEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBE28, 0xBE28 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBE29, 0xBE43 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BYIG..HANGUL SYLLABLE BYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBE44, 0xBE44 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBE45, 0xBE5F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BIG..HANGUL SYLLABLE BIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBE60, 0xBE60 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBE61, 0xBE7B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBAG..HANGUL SYLLABLE BBAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBE7C, 0xBE7C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBE7D, 0xBE97 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBAEG..HANGUL SYLLABLE BBAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBE98, 0xBE98 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBE99, 0xBEB3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBYAG..HANGUL SYLLABLE BBYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBEB4, 0xBEB4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBEB5, 0xBECF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBYAEG..HANGUL SYLLABLE BBYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBED0, 0xBED0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBED1, 0xBEEB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBEOG..HANGUL SYLLABLE BBEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBEEC, 0xBEEC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBEED, 0xBF07 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBEG..HANGUL SYLLABLE BBEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBF08, 0xBF08 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBF09, 0xBF23 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBYEOG..HANGUL SYLLABLE BBYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBF24, 0xBF24 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBF25, 0xBF3F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBYEG..HANGUL SYLLABLE BBYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBF40, 0xBF40 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBF41, 0xBF5B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBOG..HANGUL SYLLABLE BBOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBF5C, 0xBF5C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBF5D, 0xBF77 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBWAG..HANGUL SYLLABLE BBWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBF78, 0xBF78 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBF79, 0xBF93 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBWAEG..HANGUL SYLLABLE BBWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBF94, 0xBF94 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBF95, 0xBFAF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBOEG..HANGUL SYLLABLE BBOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBFB0, 0xBFB0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBFB1, 0xBFCB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBYOG..HANGUL SYLLABLE BBYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBFCC, 0xBFCC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBFCD, 0xBFE7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBUG..HANGUL SYLLABLE BBUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBFE8, 0xBFE8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xBFE9, 0xC003 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBWEOG..HANGUL SYLLABLE BBWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC004, 0xC004 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC005, 0xC01F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBWEG..HANGUL SYLLABLE BBWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC020, 0xC020 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC021, 0xC03B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBWIG..HANGUL SYLLABLE BBWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC03C, 0xC03C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC03D, 0xC057 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBYUG..HANGUL SYLLABLE BBYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC058, 0xC058 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC059, 0xC073 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBEUG..HANGUL SYLLABLE BBEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC074, 0xC074 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC075, 0xC08F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBYIG..HANGUL SYLLABLE BBYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC090, 0xC090 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE BBI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC091, 0xC0AB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE BBIG..HANGUL SYLLABLE BBIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC0AC, 0xC0AC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC0AD, 0xC0C7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SAG..HANGUL SYLLABLE SAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC0C8, 0xC0C8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC0C9, 0xC0E3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SAEG..HANGUL SYLLABLE SAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC0E4, 0xC0E4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC0E5, 0xC0FF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SYAG..HANGUL SYLLABLE SYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC100, 0xC100 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC101, 0xC11B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SYAEG..HANGUL SYLLABLE SYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC11C, 0xC11C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC11D, 0xC137 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SEOG..HANGUL SYLLABLE SEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC138, 0xC138 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC139, 0xC153 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SEG..HANGUL SYLLABLE SEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC154, 0xC154 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC155, 0xC16F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SYEOG..HANGUL SYLLABLE SYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC170, 0xC170 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC171, 0xC18B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SYEG..HANGUL SYLLABLE SYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC18C, 0xC18C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC18D, 0xC1A7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SOG..HANGUL SYLLABLE SOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC1A8, 0xC1A8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC1A9, 0xC1C3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SWAG..HANGUL SYLLABLE SWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC1C4, 0xC1C4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC1C5, 0xC1DF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SWAEG..HANGUL SYLLABLE SWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC1E0, 0xC1E0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC1E1, 0xC1FB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SOEG..HANGUL SYLLABLE SOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC1FC, 0xC1FC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC1FD, 0xC217 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SYOG..HANGUL SYLLABLE SYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC218, 0xC218 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC219, 0xC233 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SUG..HANGUL SYLLABLE SUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC234, 0xC234 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC235, 0xC24F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SWEOG..HANGUL SYLLABLE SWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC250, 0xC250 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC251, 0xC26B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SWEG..HANGUL SYLLABLE SWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC26C, 0xC26C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC26D, 0xC287 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SWIG..HANGUL SYLLABLE SWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC288, 0xC288 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC289, 0xC2A3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SYUG..HANGUL SYLLABLE SYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC2A4, 0xC2A4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC2A5, 0xC2BF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SEUG..HANGUL SYLLABLE SEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC2C0, 0xC2C0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC2C1, 0xC2DB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SYIG..HANGUL SYLLABLE SYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC2DC, 0xC2DC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC2DD, 0xC2F7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SIG..HANGUL SYLLABLE SIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC2F8, 0xC2F8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC2F9, 0xC313 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSAG..HANGUL SYLLABLE SSAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC314, 0xC314 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC315, 0xC32F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSAEG..HANGUL SYLLABLE SSAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC330, 0xC330 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC331, 0xC34B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSYAG..HANGUL SYLLABLE SSYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC34C, 0xC34C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC34D, 0xC367 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSYAEG..HANGUL SYLLABLE SSYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC368, 0xC368 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC369, 0xC383 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSEOG..HANGUL SYLLABLE SSEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC384, 0xC384 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC385, 0xC39F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSEG..HANGUL SYLLABLE SSEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC3A0, 0xC3A0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC3A1, 0xC3BB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSYEOG..HANGUL SYLLABLE SSYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC3BC, 0xC3BC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC3BD, 0xC3D7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSYEG..HANGUL SYLLABLE SSYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC3D8, 0xC3D8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC3D9, 0xC3F3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSOG..HANGUL SYLLABLE SSOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC3F4, 0xC3F4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC3F5, 0xC40F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSWAG..HANGUL SYLLABLE SSWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC410, 0xC410 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC411, 0xC42B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSWAEG..HANGUL SYLLABLE SSWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC42C, 0xC42C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC42D, 0xC447 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSOEG..HANGUL SYLLABLE SSOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC448, 0xC448 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC449, 0xC463 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSYOG..HANGUL SYLLABLE SSYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC464, 0xC464 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC465, 0xC47F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSUG..HANGUL SYLLABLE SSUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC480, 0xC480 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC481, 0xC49B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSWEOG..HANGUL SYLLABLE SSWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC49C, 0xC49C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC49D, 0xC4B7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSWEG..HANGUL SYLLABLE SSWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC4B8, 0xC4B8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC4B9, 0xC4D3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSWIG..HANGUL SYLLABLE SSWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC4D4, 0xC4D4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC4D5, 0xC4EF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSYUG..HANGUL SYLLABLE SSYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC4F0, 0xC4F0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC4F1, 0xC50B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSEUG..HANGUL SYLLABLE SSEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC50C, 0xC50C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC50D, 0xC527 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSYIG..HANGUL SYLLABLE SSYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC528, 0xC528 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE SSI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC529, 0xC543 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE SSIG..HANGUL SYLLABLE SSIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC544, 0xC544 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE A
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC545, 0xC55F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE AG..HANGUL SYLLABLE AH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC560, 0xC560 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE AE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC561, 0xC57B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE AEG..HANGUL SYLLABLE AEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC57C, 0xC57C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE YA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC57D, 0xC597 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE YAG..HANGUL SYLLABLE YAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC598, 0xC598 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE YAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC599, 0xC5B3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE YAEG..HANGUL SYLLABLE YAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC5B4, 0xC5B4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE EO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC5B5, 0xC5CF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE EOG..HANGUL SYLLABLE EOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC5D0, 0xC5D0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC5D1, 0xC5EB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE EG..HANGUL SYLLABLE EH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC5EC, 0xC5EC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE YEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC5ED, 0xC607 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE YEOG..HANGUL SYLLABLE YEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC608, 0xC608 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE YE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC609, 0xC623 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE YEG..HANGUL SYLLABLE YEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC624, 0xC624 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC625, 0xC63F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE OG..HANGUL SYLLABLE OH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC640, 0xC640 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE WA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC641, 0xC65B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE WAG..HANGUL SYLLABLE WAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC65C, 0xC65C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE WAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC65D, 0xC677 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE WAEG..HANGUL SYLLABLE WAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC678, 0xC678 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE OE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC679, 0xC693 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE OEG..HANGUL SYLLABLE OEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC694, 0xC694 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE YO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC695, 0xC6AF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE YOG..HANGUL SYLLABLE YOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC6B0, 0xC6B0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE U
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC6B1, 0xC6CB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE UG..HANGUL SYLLABLE UH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC6CC, 0xC6CC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE WEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC6CD, 0xC6E7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE WEOG..HANGUL SYLLABLE WEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC6E8, 0xC6E8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE WE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC6E9, 0xC703 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE WEG..HANGUL SYLLABLE WEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC704, 0xC704 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE WI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC705, 0xC71F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE WIG..HANGUL SYLLABLE WIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC720, 0xC720 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE YU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC721, 0xC73B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE YUG..HANGUL SYLLABLE YUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC73C, 0xC73C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE EU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC73D, 0xC757 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE EUG..HANGUL SYLLABLE EUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC758, 0xC758 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE YI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC759, 0xC773 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE YIG..HANGUL SYLLABLE YIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC774, 0xC774 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE I
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC775, 0xC78F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE IG..HANGUL SYLLABLE IH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC790, 0xC790 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC791, 0xC7AB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JAG..HANGUL SYLLABLE JAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC7AC, 0xC7AC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC7AD, 0xC7C7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JAEG..HANGUL SYLLABLE JAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC7C8, 0xC7C8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC7C9, 0xC7E3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JYAG..HANGUL SYLLABLE JYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC7E4, 0xC7E4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC7E5, 0xC7FF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JYAEG..HANGUL SYLLABLE JYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC800, 0xC800 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC801, 0xC81B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JEOG..HANGUL SYLLABLE JEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC81C, 0xC81C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC81D, 0xC837 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JEG..HANGUL SYLLABLE JEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC838, 0xC838 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC839, 0xC853 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JYEOG..HANGUL SYLLABLE JYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC854, 0xC854 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC855, 0xC86F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JYEG..HANGUL SYLLABLE JYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC870, 0xC870 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC871, 0xC88B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JOG..HANGUL SYLLABLE JOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC88C, 0xC88C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC88D, 0xC8A7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JWAG..HANGUL SYLLABLE JWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC8A8, 0xC8A8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC8A9, 0xC8C3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JWAEG..HANGUL SYLLABLE JWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC8C4, 0xC8C4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC8C5, 0xC8DF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JOEG..HANGUL SYLLABLE JOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC8E0, 0xC8E0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC8E1, 0xC8FB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JYOG..HANGUL SYLLABLE JYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC8FC, 0xC8FC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC8FD, 0xC917 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JUG..HANGUL SYLLABLE JUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC918, 0xC918 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC919, 0xC933 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JWEOG..HANGUL SYLLABLE JWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC934, 0xC934 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC935, 0xC94F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JWEG..HANGUL SYLLABLE JWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC950, 0xC950 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC951, 0xC96B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JWIG..HANGUL SYLLABLE JWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC96C, 0xC96C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC96D, 0xC987 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JYUG..HANGUL SYLLABLE JYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC988, 0xC988 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC989, 0xC9A3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JEUG..HANGUL SYLLABLE JEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC9A4, 0xC9A4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC9A5, 0xC9BF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JYIG..HANGUL SYLLABLE JYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC9C0, 0xC9C0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC9C1, 0xC9DB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JIG..HANGUL SYLLABLE JIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC9DC, 0xC9DC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC9DD, 0xC9F7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJAG..HANGUL SYLLABLE JJAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC9F8, 0xC9F8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xC9F9, 0xCA13 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJAEG..HANGUL SYLLABLE JJAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCA14, 0xCA14 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCA15, 0xCA2F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJYAG..HANGUL SYLLABLE JJYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCA30, 0xCA30 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCA31, 0xCA4B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJYAEG..HANGUL SYLLABLE JJYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCA4C, 0xCA4C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCA4D, 0xCA67 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJEOG..HANGUL SYLLABLE JJEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCA68, 0xCA68 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCA69, 0xCA83 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJEG..HANGUL SYLLABLE JJEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCA84, 0xCA84 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCA85, 0xCA9F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJYEOG..HANGUL SYLLABLE JJYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCAA0, 0xCAA0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCAA1, 0xCABB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJYEG..HANGUL SYLLABLE JJYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCABC, 0xCABC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCABD, 0xCAD7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJOG..HANGUL SYLLABLE JJOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCAD8, 0xCAD8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCAD9, 0xCAF3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJWAG..HANGUL SYLLABLE JJWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCAF4, 0xCAF4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCAF5, 0xCB0F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJWAEG..HANGUL SYLLABLE JJWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCB10, 0xCB10 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCB11, 0xCB2B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJOEG..HANGUL SYLLABLE JJOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCB2C, 0xCB2C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCB2D, 0xCB47 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJYOG..HANGUL SYLLABLE JJYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCB48, 0xCB48 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCB49, 0xCB63 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJUG..HANGUL SYLLABLE JJUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCB64, 0xCB64 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCB65, 0xCB7F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJWEOG..HANGUL SYLLABLE JJWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCB80, 0xCB80 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCB81, 0xCB9B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJWEG..HANGUL SYLLABLE JJWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCB9C, 0xCB9C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCB9D, 0xCBB7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJWIG..HANGUL SYLLABLE JJWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCBB8, 0xCBB8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCBB9, 0xCBD3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJYUG..HANGUL SYLLABLE JJYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCBD4, 0xCBD4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCBD5, 0xCBEF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJEUG..HANGUL SYLLABLE JJEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCBF0, 0xCBF0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCBF1, 0xCC0B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJYIG..HANGUL SYLLABLE JJYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCC0C, 0xCC0C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE JJI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCC0D, 0xCC27 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE JJIG..HANGUL SYLLABLE JJIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCC28, 0xCC28 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCC29, 0xCC43 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CAG..HANGUL SYLLABLE CAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCC44, 0xCC44 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCC45, 0xCC5F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CAEG..HANGUL SYLLABLE CAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCC60, 0xCC60 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCC61, 0xCC7B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CYAG..HANGUL SYLLABLE CYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCC7C, 0xCC7C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCC7D, 0xCC97 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CYAEG..HANGUL SYLLABLE CYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCC98, 0xCC98 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCC99, 0xCCB3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CEOG..HANGUL SYLLABLE CEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCCB4, 0xCCB4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCCB5, 0xCCCF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CEG..HANGUL SYLLABLE CEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCCD0, 0xCCD0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCCD1, 0xCCEB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CYEOG..HANGUL SYLLABLE CYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCCEC, 0xCCEC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCCED, 0xCD07 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CYEG..HANGUL SYLLABLE CYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCD08, 0xCD08 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCD09, 0xCD23 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE COG..HANGUL SYLLABLE COH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCD24, 0xCD24 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCD25, 0xCD3F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CWAG..HANGUL SYLLABLE CWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCD40, 0xCD40 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCD41, 0xCD5B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CWAEG..HANGUL SYLLABLE CWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCD5C, 0xCD5C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE COE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCD5D, 0xCD77 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE COEG..HANGUL SYLLABLE COEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCD78, 0xCD78 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCD79, 0xCD93 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CYOG..HANGUL SYLLABLE CYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCD94, 0xCD94 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCD95, 0xCDAF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CUG..HANGUL SYLLABLE CUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCDB0, 0xCDB0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCDB1, 0xCDCB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CWEOG..HANGUL SYLLABLE CWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCDCC, 0xCDCC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCDCD, 0xCDE7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CWEG..HANGUL SYLLABLE CWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCDE8, 0xCDE8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCDE9, 0xCE03 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CWIG..HANGUL SYLLABLE CWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCE04, 0xCE04 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCE05, 0xCE1F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CYUG..HANGUL SYLLABLE CYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCE20, 0xCE20 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCE21, 0xCE3B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CEUG..HANGUL SYLLABLE CEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCE3C, 0xCE3C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCE3D, 0xCE57 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CYIG..HANGUL SYLLABLE CYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCE58, 0xCE58 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE CI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCE59, 0xCE73 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE CIG..HANGUL SYLLABLE CIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCE74, 0xCE74 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCE75, 0xCE8F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KAG..HANGUL SYLLABLE KAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCE90, 0xCE90 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCE91, 0xCEAB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KAEG..HANGUL SYLLABLE KAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCEAC, 0xCEAC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCEAD, 0xCEC7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KYAG..HANGUL SYLLABLE KYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCEC8, 0xCEC8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCEC9, 0xCEE3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KYAEG..HANGUL SYLLABLE KYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCEE4, 0xCEE4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCEE5, 0xCEFF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KEOG..HANGUL SYLLABLE KEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCF00, 0xCF00 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCF01, 0xCF1B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KEG..HANGUL SYLLABLE KEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCF1C, 0xCF1C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCF1D, 0xCF37 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KYEOG..HANGUL SYLLABLE KYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCF38, 0xCF38 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCF39, 0xCF53 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KYEG..HANGUL SYLLABLE KYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCF54, 0xCF54 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCF55, 0xCF6F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KOG..HANGUL SYLLABLE KOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCF70, 0xCF70 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCF71, 0xCF8B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KWAG..HANGUL SYLLABLE KWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCF8C, 0xCF8C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCF8D, 0xCFA7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KWAEG..HANGUL SYLLABLE KWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCFA8, 0xCFA8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCFA9, 0xCFC3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KOEG..HANGUL SYLLABLE KOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCFC4, 0xCFC4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCFC5, 0xCFDF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KYOG..HANGUL SYLLABLE KYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCFE0, 0xCFE0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCFE1, 0xCFFB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KUG..HANGUL SYLLABLE KUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCFFC, 0xCFFC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xCFFD, 0xD017 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KWEOG..HANGUL SYLLABLE KWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD018, 0xD018 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD019, 0xD033 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KWEG..HANGUL SYLLABLE KWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD034, 0xD034 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD035, 0xD04F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KWIG..HANGUL SYLLABLE KWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD050, 0xD050 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD051, 0xD06B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KYUG..HANGUL SYLLABLE KYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD06C, 0xD06C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD06D, 0xD087 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KEUG..HANGUL SYLLABLE KEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD088, 0xD088 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD089, 0xD0A3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KYIG..HANGUL SYLLABLE KYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD0A4, 0xD0A4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE KI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD0A5, 0xD0BF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE KIG..HANGUL SYLLABLE KIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD0C0, 0xD0C0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD0C1, 0xD0DB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TAG..HANGUL SYLLABLE TAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD0DC, 0xD0DC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD0DD, 0xD0F7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TAEG..HANGUL SYLLABLE TAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD0F8, 0xD0F8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD0F9, 0xD113 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TYAG..HANGUL SYLLABLE TYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD114, 0xD114 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD115, 0xD12F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TYAEG..HANGUL SYLLABLE TYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD130, 0xD130 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD131, 0xD14B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TEOG..HANGUL SYLLABLE TEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD14C, 0xD14C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD14D, 0xD167 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TEG..HANGUL SYLLABLE TEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD168, 0xD168 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD169, 0xD183 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TYEOG..HANGUL SYLLABLE TYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD184, 0xD184 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD185, 0xD19F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TYEG..HANGUL SYLLABLE TYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD1A0, 0xD1A0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD1A1, 0xD1BB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TOG..HANGUL SYLLABLE TOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD1BC, 0xD1BC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD1BD, 0xD1D7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TWAG..HANGUL SYLLABLE TWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD1D8, 0xD1D8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD1D9, 0xD1F3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TWAEG..HANGUL SYLLABLE TWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD1F4, 0xD1F4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD1F5, 0xD20F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TOEG..HANGUL SYLLABLE TOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD210, 0xD210 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD211, 0xD22B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TYOG..HANGUL SYLLABLE TYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD22C, 0xD22C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD22D, 0xD247 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TUG..HANGUL SYLLABLE TUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD248, 0xD248 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD249, 0xD263 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TWEOG..HANGUL SYLLABLE TWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD264, 0xD264 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD265, 0xD27F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TWEG..HANGUL SYLLABLE TWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD280, 0xD280 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD281, 0xD29B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TWIG..HANGUL SYLLABLE TWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD29C, 0xD29C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD29D, 0xD2B7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TYUG..HANGUL SYLLABLE TYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD2B8, 0xD2B8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD2B9, 0xD2D3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TEUG..HANGUL SYLLABLE TEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD2D4, 0xD2D4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD2D5, 0xD2EF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TYIG..HANGUL SYLLABLE TYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD2F0, 0xD2F0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE TI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD2F1, 0xD30B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE TIG..HANGUL SYLLABLE TIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD30C, 0xD30C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD30D, 0xD327 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PAG..HANGUL SYLLABLE PAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD328, 0xD328 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD329, 0xD343 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PAEG..HANGUL SYLLABLE PAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD344, 0xD344 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD345, 0xD35F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PYAG..HANGUL SYLLABLE PYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD360, 0xD360 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD361, 0xD37B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PYAEG..HANGUL SYLLABLE PYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD37C, 0xD37C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD37D, 0xD397 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PEOG..HANGUL SYLLABLE PEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD398, 0xD398 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD399, 0xD3B3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PEG..HANGUL SYLLABLE PEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD3B4, 0xD3B4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD3B5, 0xD3CF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PYEOG..HANGUL SYLLABLE PYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD3D0, 0xD3D0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD3D1, 0xD3EB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PYEG..HANGUL SYLLABLE PYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD3EC, 0xD3EC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD3ED, 0xD407 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE POG..HANGUL SYLLABLE POH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD408, 0xD408 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD409, 0xD423 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PWAG..HANGUL SYLLABLE PWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD424, 0xD424 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD425, 0xD43F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PWAEG..HANGUL SYLLABLE PWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD440, 0xD440 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE POE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD441, 0xD45B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE POEG..HANGUL SYLLABLE POEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD45C, 0xD45C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD45D, 0xD477 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PYOG..HANGUL SYLLABLE PYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD478, 0xD478 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD479, 0xD493 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PUG..HANGUL SYLLABLE PUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD494, 0xD494 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD495, 0xD4AF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PWEOG..HANGUL SYLLABLE PWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD4B0, 0xD4B0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD4B1, 0xD4CB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PWEG..HANGUL SYLLABLE PWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD4CC, 0xD4CC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD4CD, 0xD4E7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PWIG..HANGUL SYLLABLE PWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD4E8, 0xD4E8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD4E9, 0xD503 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PYUG..HANGUL SYLLABLE PYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD504, 0xD504 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD505, 0xD51F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PEUG..HANGUL SYLLABLE PEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD520, 0xD520 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD521, 0xD53B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PYIG..HANGUL SYLLABLE PYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD53C, 0xD53C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE PI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD53D, 0xD557 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE PIG..HANGUL SYLLABLE PIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD558, 0xD558 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD559, 0xD573 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HAG..HANGUL SYLLABLE HAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD574, 0xD574 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD575, 0xD58F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HAEG..HANGUL SYLLABLE HAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD590, 0xD590 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD591, 0xD5AB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HYAG..HANGUL SYLLABLE HYAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD5AC, 0xD5AC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HYAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD5AD, 0xD5C7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HYAEG..HANGUL SYLLABLE HYAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD5C8, 0xD5C8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD5C9, 0xD5E3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HEOG..HANGUL SYLLABLE HEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD5E4, 0xD5E4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD5E5, 0xD5FF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HEG..HANGUL SYLLABLE HEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD600, 0xD600 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HYEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD601, 0xD61B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HYEOG..HANGUL SYLLABLE HYEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD61C, 0xD61C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HYE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD61D, 0xD637 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HYEG..HANGUL SYLLABLE HYEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD638, 0xD638 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD639, 0xD653 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HOG..HANGUL SYLLABLE HOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD654, 0xD654 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HWA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD655, 0xD66F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HWAG..HANGUL SYLLABLE HWAH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD670, 0xD670 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HWAE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD671, 0xD68B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HWAEG..HANGUL SYLLABLE HWAEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD68C, 0xD68C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD68D, 0xD6A7 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HOEG..HANGUL SYLLABLE HOEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD6A8, 0xD6A8 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HYO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD6A9, 0xD6C3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HYOG..HANGUL SYLLABLE HYOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD6C4, 0xD6C4 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD6C5, 0xD6DF }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HUG..HANGUL SYLLABLE HUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD6E0, 0xD6E0 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HWEO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD6E1, 0xD6FB }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HWEOG..HANGUL SYLLABLE HWEOH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD6FC, 0xD6FC }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HWE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD6FD, 0xD717 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HWEG..HANGUL SYLLABLE HWEH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD718, 0xD718 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HWI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD719, 0xD733 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HWIG..HANGUL SYLLABLE HWIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD734, 0xD734 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HYU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD735, 0xD74F }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HYUG..HANGUL SYLLABLE HYUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD750, 0xD750 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HEU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD751, 0xD76B }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HEUG..HANGUL SYLLABLE HEUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD76C, 0xD76C }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HYI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD76D, 0xD787 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HYIG..HANGUL SYLLABLE HYIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD788, 0xD788 }, ::unicode::Grapheme_Cluster_Break::LV }, // Lo       HANGUL SYLLABLE HI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD789, 0xD7A3 }, ::unicode::Grapheme_Cluster_Break::LVT }, // Lo  [27] HANGUL SYLLABLE HIG..HANGUL SYLLABLE HIH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD7B0, 0xD7C6 }, ::unicode::Grapheme_Cluster_Break::V }, // Lo  [23] HANGUL JUNGSEONG O-YEO..HANGUL JUNGSEONG ARAEA-E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xD7CB, 0xD7FB }, ::unicode::Grapheme_Cluster_Break::T }, // Lo  [49] HANGUL JONGSEONG NIEUN-RIEUL..HANGUL JONGSEONG PHIEUPH-THIEUTH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xFB1E, 0xFB1E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       HEBREW POINT JUDEO-SPANISH VARIKA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xFE00, 0xFE0F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [16] VARIATION SELECTOR-1..VARIATION SELECTOR-16
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xFE20, 0xFE2F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [16] COMBINING LIGATURE LEFT HALF..COMBINING CYRILLIC TITLO RIGHT HALF
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xFEFF, 0xFEFF }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf       ZERO WIDTH NO-BREAK SPACE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xFF9E, 0xFF9F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Lm   [2] HALFWIDTH KATAKANA VOICED SOUND MARK..HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xFFF0, 0xFFF8 }, ::unicode::Grapheme_Cluster_Break::Control }, // Cn   [9] <reserved-FFF0>..<reserved-FFF8>
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xFFF9, 0xFFFB }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf   [3] INTERLINEAR ANNOTATION ANCHOR..INTERLINEAR ANNOTATION TERMINATOR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x101FD, 0x101FD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       PHAISTOS DISC SIGN COMBINING OBLIQUE STROKE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x102E0, 0x102E0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       COPTIC EPACT THOUSANDS MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10376, 0x1037A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] COMBINING OLD PERMIC LETTER AN..COMBINING OLD PERMIC LETTER SII
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10A01, 0x10A03 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] KHAROSHTHI VOWEL SIGN I..KHAROSHTHI VOWEL SIGN VOCALIC R
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10A05, 0x10A06 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] KHAROSHTHI VOWEL SIGN E..KHAROSHTHI VOWEL SIGN O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10A0C, 0x10A0F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] KHAROSHTHI VOWEL LENGTH MARK..KHAROSHTHI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10A38, 0x10A3A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] KHAROSHTHI SIGN BAR ABOVE..KHAROSHTHI SIGN DOT BELOW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10A3F, 0x10A3F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KHAROSHTHI VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10AE5, 0x10AE6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MANICHAEAN ABBREVIATION MARK ABOVE..MANICHAEAN ABBREVIATION MARK BELOW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10D24, 0x10D27 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] HANIFI ROHINGYA SIGN HARBAHAY..HANIFI ROHINGYA SIGN TASSI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10D69, 0x10D6D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] GARAY VOWEL SIGN E..GARAY CONSONANT NASALIZATION MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10EAB, 0x10EAC }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] YEZIDI COMBINING HAMZA MARK..YEZIDI COMBINING MADDA MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10EFA, 0x10EFF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [6] ARABIC DOUBLE VERTICAL BAR BELOW..ARABIC SMALL LOW WORD MADDA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10F46, 0x10F50 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [11] SOGDIAN COMBINING DOT BELOW..SOGDIAN COMBINING STROKE BELOW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x10F82, 0x10F85 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] OLD UYGHUR COMBINING DOT ABOVE..OLD UYGHUR COMBINING TWO DOTS BELOW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11000, 0x11000 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       BRAHMI SIGN CANDRABINDU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11001, 0x11001 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BRAHMI SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11002, 0x11002 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       BRAHMI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11038, 0x11046 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [15] BRAHMI VOWEL SIGN AA..BRAHMI VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11070, 0x11070 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BRAHMI SIGN OLD TAMIL VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11073, 0x11074 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] BRAHMI VOWEL SIGN OLD TAMIL SHORT E..BRAHMI VOWEL SIGN OLD TAMIL SHORT O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1107F, 0x11081 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] BRAHMI NUMBER JOINER..KAITHI SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11082, 0x11082 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       KAITHI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x110B0, 0x110B2 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] KAITHI VOWEL SIGN AA..KAITHI VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x110B3, 0x110B6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] KAITHI VOWEL SIGN U..KAITHI VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x110B7, 0x110B8 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] KAITHI VOWEL SIGN O..KAITHI VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x110B9, 0x110BA }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] KAITHI SIGN VIRAMA..KAITHI SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x110BD, 0x110BD }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Cf       KAITHI NUMBER SIGN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x110C2, 0x110C2 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KAITHI VOWEL SIGN VOCALIC R
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x110CD, 0x110CD }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Cf       KAITHI NUMBER SIGN ABOVE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11100, 0x11102 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] CHAKMA SIGN CANDRABINDU..CHAKMA SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11127, 0x1112B }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] CHAKMA VOWEL SIGN A..CHAKMA VOWEL SIGN UU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1112C, 0x1112C }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       CHAKMA VOWEL SIGN E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1112D, 0x11134 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [8] CHAKMA VOWEL SIGN AI..CHAKMA MAAYYAA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11145, 0x11146 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] CHAKMA VOWEL SIGN AA..CHAKMA VOWEL SIGN EI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11173, 0x11173 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MAHAJANI SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11180, 0x11181 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] SHARADA SIGN CANDRABINDU..SHARADA SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11182, 0x11182 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       SHARADA SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x111B3, 0x111B5 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] SHARADA VOWEL SIGN AA..SHARADA VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x111B6, 0x111BE }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [9] SHARADA VOWEL SIGN U..SHARADA VOWEL SIGN O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x111BF, 0x111BF }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       SHARADA VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x111C0, 0x111C0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       SHARADA SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x111C2, 0x111C3 }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Lo   [2] SHARADA SIGN JIHVAMULIYA..SHARADA SIGN UPADHMANIYA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x111C9, 0x111CC }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] SHARADA SANDHI MARK..SHARADA EXTRA SHORT VOWEL MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x111CE, 0x111CE }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       SHARADA VOWEL SIGN PRISHTHAMATRA E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x111CF, 0x111CF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SHARADA SIGN INVERTED CANDRABINDU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1122C, 0x1122E }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] KHOJKI VOWEL SIGN AA..KHOJKI VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1122F, 0x11231 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] KHOJKI VOWEL SIGN U..KHOJKI VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11232, 0x11233 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] KHOJKI VOWEL SIGN O..KHOJKI VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11234, 0x11234 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KHOJKI SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11235, 0x11235 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       KHOJKI SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11236, 0x11237 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] KHOJKI SIGN NUKTA..KHOJKI SIGN SHADDA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1123E, 0x1123E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KHOJKI SIGN SUKUN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11241, 0x11241 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KHOJKI VOWEL SIGN VOCALIC R
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x112DF, 0x112DF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KHUDAWADI SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x112E0, 0x112E2 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] KHUDAWADI VOWEL SIGN AA..KHUDAWADI VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x112E3, 0x112EA }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [8] KHUDAWADI VOWEL SIGN U..KHUDAWADI SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11300, 0x11301 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] GRANTHA SIGN COMBINING ANUSVARA ABOVE..GRANTHA SIGN CANDRABINDU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11302, 0x11303 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] GRANTHA SIGN ANUSVARA..GRANTHA SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1133B, 0x1133C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] COMBINING BINDU BELOW..GRANTHA SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1133E, 0x1133E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       GRANTHA VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1133F, 0x1133F }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       GRANTHA VOWEL SIGN I
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11340, 0x11340 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       GRANTHA VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11341, 0x11344 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [4] GRANTHA VOWEL SIGN U..GRANTHA VOWEL SIGN VOCALIC RR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11347, 0x11348 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] GRANTHA VOWEL SIGN EE..GRANTHA VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1134B, 0x1134C }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] GRANTHA VOWEL SIGN OO..GRANTHA VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1134D, 0x1134D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       GRANTHA SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11357, 0x11357 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       GRANTHA AU LENGTH MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11362, 0x11363 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] GRANTHA VOWEL SIGN VOCALIC L..GRANTHA VOWEL SIGN VOCALIC LL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11366, 0x1136C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] COMBINING GRANTHA DIGIT ZERO..COMBINING GRANTHA DIGIT SIX
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11370, 0x11374 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] COMBINING GRANTHA LETTER A..COMBINING GRANTHA LETTER PA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113B8, 0x113B8 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       TULU-TIGALARI VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113B9, 0x113BA }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] TULU-TIGALARI VOWEL SIGN I..TULU-TIGALARI VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113BB, 0x113C0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [6] TULU-TIGALARI VOWEL SIGN U..TULU-TIGALARI VOWEL SIGN VOCALIC LL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113C2, 0x113C2 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       TULU-TIGALARI VOWEL SIGN EE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113C5, 0x113C5 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       TULU-TIGALARI VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113C7, 0x113C9 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc   [3] TULU-TIGALARI VOWEL SIGN OO..TULU-TIGALARI AU LENGTH MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113CA, 0x113CA }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       TULU-TIGALARI SIGN CANDRA ANUNASIKA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113CC, 0x113CD }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] TULU-TIGALARI SIGN ANUSVARA..TULU-TIGALARI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113CE, 0x113CE }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TULU-TIGALARI SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113CF, 0x113CF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       TULU-TIGALARI SIGN LOOPED VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113D0, 0x113D0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TULU-TIGALARI CONJOINER
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113D1, 0x113D1 }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Lo       TULU-TIGALARI REPHA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113D2, 0x113D2 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TULU-TIGALARI GEMINATION MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x113E1, 0x113E2 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] TULU-TIGALARI VEDIC TONE SVARITA..TULU-TIGALARI VEDIC TONE ANUDATTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11435, 0x11437 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] NEWA VOWEL SIGN AA..NEWA VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11438, 0x1143F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [8] NEWA VOWEL SIGN U..NEWA VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11440, 0x11441 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] NEWA VOWEL SIGN O..NEWA VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11442, 0x11444 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] NEWA SIGN VIRAMA..NEWA SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11445, 0x11445 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       NEWA SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11446, 0x11446 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       NEWA SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1145E, 0x1145E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       NEWA SANDHI MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x114B0, 0x114B0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       TIRHUTA VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x114B1, 0x114B2 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] TIRHUTA VOWEL SIGN I..TIRHUTA VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x114B3, 0x114B8 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [6] TIRHUTA VOWEL SIGN U..TIRHUTA VOWEL SIGN VOCALIC LL
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x114B9, 0x114B9 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       TIRHUTA VOWEL SIGN E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x114BA, 0x114BA }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TIRHUTA VOWEL SIGN SHORT E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x114BB, 0x114BC }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] TIRHUTA VOWEL SIGN AI..TIRHUTA VOWEL SIGN O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x114BD, 0x114BD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       TIRHUTA VOWEL SIGN SHORT O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x114BE, 0x114BE }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       TIRHUTA VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x114BF, 0x114C0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] TIRHUTA SIGN CANDRABINDU..TIRHUTA SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x114C1, 0x114C1 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       TIRHUTA SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x114C2, 0x114C3 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] TIRHUTA SIGN VIRAMA..TIRHUTA SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x115AF, 0x115AF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       SIDDHAM VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x115B0, 0x115B1 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] SIDDHAM VOWEL SIGN I..SIDDHAM VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x115B2, 0x115B5 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] SIDDHAM VOWEL SIGN U..SIDDHAM VOWEL SIGN VOCALIC RR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x115B8, 0x115BB }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [4] SIDDHAM VOWEL SIGN E..SIDDHAM VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x115BC, 0x115BD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] SIDDHAM SIGN CANDRABINDU..SIDDHAM SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x115BE, 0x115BE }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       SIDDHAM SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x115BF, 0x115C0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] SIDDHAM SIGN VIRAMA..SIDDHAM SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x115DC, 0x115DD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] SIDDHAM VOWEL SIGN ALTERNATE U..SIDDHAM VOWEL SIGN ALTERNATE UU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11630, 0x11632 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] MODI VOWEL SIGN AA..MODI VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11633, 0x1163A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [8] MODI VOWEL SIGN U..MODI VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1163B, 0x1163C }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] MODI VOWEL SIGN O..MODI VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1163D, 0x1163D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MODI SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1163E, 0x1163E }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       MODI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1163F, 0x11640 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MODI SIGN VIRAMA..MODI SIGN ARDHACANDRA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x116AB, 0x116AB }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAKRI SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x116AC, 0x116AC }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       TAKRI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x116AD, 0x116AD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAKRI VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x116AE, 0x116AF }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] TAKRI VOWEL SIGN I..TAKRI VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x116B0, 0x116B5 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [6] TAKRI VOWEL SIGN U..TAKRI VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x116B6, 0x116B6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       TAKRI SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x116B7, 0x116B7 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAKRI SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1171D, 0x1171D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       AHOM CONSONANT SIGN MEDIAL LA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1171E, 0x1171E }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       AHOM CONSONANT SIGN MEDIAL RA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1171F, 0x1171F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       AHOM CONSONANT SIGN MEDIAL LIGATING RA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11722, 0x11725 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] AHOM VOWEL SIGN I..AHOM VOWEL SIGN UU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11726, 0x11726 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       AHOM VOWEL SIGN E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11727, 0x1172B }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] AHOM VOWEL SIGN AW..AHOM SIGN KILLER
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1182C, 0x1182E }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] DOGRA VOWEL SIGN AA..DOGRA VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1182F, 0x11837 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [9] DOGRA VOWEL SIGN U..DOGRA SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11838, 0x11838 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       DOGRA SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11839, 0x1183A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] DOGRA SIGN VIRAMA..DOGRA SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11930, 0x11930 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       DIVES AKURU VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11931, 0x11935 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [5] DIVES AKURU VOWEL SIGN I..DIVES AKURU VOWEL SIGN E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11937, 0x11938 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] DIVES AKURU VOWEL SIGN AI..DIVES AKURU VOWEL SIGN O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1193B, 0x1193C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] DIVES AKURU SIGN ANUSVARA..DIVES AKURU SIGN CANDRABINDU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1193D, 0x1193D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       DIVES AKURU SIGN HALANTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1193E, 0x1193E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       DIVES AKURU VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1193F, 0x1193F }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Lo       DIVES AKURU PREFIXED NASAL SIGN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11940, 0x11940 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       DIVES AKURU MEDIAL YA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11941, 0x11941 }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Lo       DIVES AKURU INITIAL RA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11942, 0x11942 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       DIVES AKURU MEDIAL RA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11943, 0x11943 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       DIVES AKURU SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x119D1, 0x119D3 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] NANDINAGARI VOWEL SIGN AA..NANDINAGARI VOWEL SIGN II
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x119D4, 0x119D7 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] NANDINAGARI VOWEL SIGN U..NANDINAGARI VOWEL SIGN VOCALIC RR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x119DA, 0x119DB }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] NANDINAGARI VOWEL SIGN E..NANDINAGARI VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x119DC, 0x119DF }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [4] NANDINAGARI VOWEL SIGN O..NANDINAGARI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x119E0, 0x119E0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       NANDINAGARI SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x119E4, 0x119E4 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       NANDINAGARI VOWEL SIGN PRISHTHAMATRA E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A01, 0x11A0A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [10] ZANABAZAR SQUARE VOWEL SIGN I..ZANABAZAR SQUARE VOWEL LENGTH MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A33, 0x11A38 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [6] ZANABAZAR SQUARE FINAL CONSONANT MARK..ZANABAZAR SQUARE SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A39, 0x11A39 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       ZANABAZAR SQUARE SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A3B, 0x11A3E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] ZANABAZAR SQUARE CLUSTER-FINAL LETTER YA..ZANABAZAR SQUARE CLUSTER-FINAL LETTER VA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A47, 0x11A47 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       ZANABAZAR SQUARE SUBJOINER
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A51, 0x11A56 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [6] SOYOMBO VOWEL SIGN I..SOYOMBO VOWEL SIGN OE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A57, 0x11A58 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] SOYOMBO VOWEL SIGN AI..SOYOMBO VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A59, 0x11A5B }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] SOYOMBO VOWEL SIGN VOCALIC R..SOYOMBO VOWEL LENGTH MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A84, 0x11A89 }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Lo   [6] SOYOMBO SIGN JIHVAMULIYA..SOYOMBO CLUSTER-INITIAL LETTER SA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A8A, 0x11A96 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [13] SOYOMBO FINAL CONSONANT SIGN G..SOYOMBO SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A97, 0x11A97 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       SOYOMBO SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11A98, 0x11A99 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] SOYOMBO GEMINATION MARK..SOYOMBO SUBJOINER
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11B60, 0x11B60 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SHARADA VOWEL SIGN OE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11B61, 0x11B61 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       SHARADA VOWEL SIGN OOE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11B62, 0x11B64 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] SHARADA VOWEL SIGN UE..SHARADA VOWEL SIGN SHORT E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11B65, 0x11B65 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       SHARADA VOWEL SIGN SHORT O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11B66, 0x11B66 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SHARADA VOWEL SIGN CANDRA E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11B67, 0x11B67 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       SHARADA VOWEL SIGN CANDRA O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11C2F, 0x11C2F }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       BHAIKSUKI VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11C30, 0x11C36 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] BHAIKSUKI VOWEL SIGN I..BHAIKSUKI VOWEL SIGN VOCALIC L
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11C38, 0x11C3D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [6] BHAIKSUKI VOWEL SIGN E..BHAIKSUKI SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11C3E, 0x11C3E }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       BHAIKSUKI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11C3F, 0x11C3F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       BHAIKSUKI SIGN VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11C92, 0x11CA7 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [22] MARCHEN SUBJOINED LETTER KA..MARCHEN SUBJOINED LETTER ZA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11CA9, 0x11CA9 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       MARCHEN SUBJOINED LETTER YA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11CAA, 0x11CB0 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] MARCHEN SUBJOINED LETTER RA..MARCHEN VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11CB1, 0x11CB1 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       MARCHEN VOWEL SIGN I
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11CB2, 0x11CB3 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MARCHEN VOWEL SIGN U..MARCHEN VOWEL SIGN E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11CB4, 0x11CB4 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       MARCHEN VOWEL SIGN O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11CB5, 0x11CB6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MARCHEN SIGN ANUSVARA..MARCHEN SIGN CANDRABINDU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11D31, 0x11D36 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [6] MASARAM GONDI VOWEL SIGN AA..MASARAM GONDI VOWEL SIGN VOCALIC R
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11D3A, 0x11D3A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MASARAM GONDI VOWEL SIGN E
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11D3C, 0x11D3D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MASARAM GONDI VOWEL SIGN AI..MASARAM GONDI VOWEL SIGN O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11D3F, 0x11D45 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] MASARAM GONDI VOWEL SIGN AU..MASARAM GONDI VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11D46, 0x11D46 }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Lo       MASARAM GONDI REPHA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11D47, 0x11D47 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MASARAM GONDI RA-KARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11D8A, 0x11D8E }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [5] GUNJALA GONDI VOWEL SIGN AA..GUNJALA GONDI VOWEL SIGN UU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11D90, 0x11D91 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] GUNJALA GONDI VOWEL SIGN EE..GUNJALA GONDI VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11D93, 0x11D94 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] GUNJALA GONDI VOWEL SIGN OO..GUNJALA GONDI VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11D95, 0x11D95 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       GUNJALA GONDI SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11D96, 0x11D96 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       GUNJALA GONDI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11D97, 0x11D97 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       GUNJALA GONDI VIRAMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11EF3, 0x11EF4 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] MAKASAR VOWEL SIGN I..MAKASAR VOWEL SIGN U
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11EF5, 0x11EF6 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] MAKASAR VOWEL SIGN E..MAKASAR VOWEL SIGN O
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11F00, 0x11F01 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] KAWI SIGN CANDRABINDU..KAWI SIGN ANUSVARA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11F02, 0x11F02 }, ::unicode::Grapheme_Cluster_Break::Prepend }, // Lo       KAWI SIGN REPHA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11F03, 0x11F03 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc       KAWI SIGN VISARGA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11F34, 0x11F35 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] KAWI VOWEL SIGN AA..KAWI VOWEL SIGN ALTERNATE AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11F36, 0x11F3A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] KAWI VOWEL SIGN I..KAWI VOWEL SIGN VOCALIC R
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11F3E, 0x11F3F }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [2] KAWI VOWEL SIGN E..KAWI VOWEL SIGN AI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11F40, 0x11F40 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KAWI VOWEL SIGN EU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11F41, 0x11F41 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc       KAWI SIGN KILLER
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11F42, 0x11F42 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KAWI CONJOINER
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x11F5A, 0x11F5A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KAWI SIGN NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x13430, 0x1343F }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf  [16] EGYPTIAN HIEROGLYPH VERTICAL JOINER..EGYPTIAN HIEROGLYPH END WALLED ENCLOSURE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x13440, 0x13440 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       EGYPTIAN HIEROGLYPH MIRROR HORIZONTALLY
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x13447, 0x13455 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [15] EGYPTIAN HIEROGLYPH MODIFIER DAMAGED AT TOP START..EGYPTIAN HIEROGLYPH MODIFIER DAMAGED
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1611E, 0x16129 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [12] GURUNG KHEMA VOWEL SIGN AA..GURUNG KHEMA VOWEL LENGTH MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1612A, 0x1612C }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc   [3] GURUNG KHEMA CONSONANT SIGN MEDIAL YA..GURUNG KHEMA CONSONANT SIGN MEDIAL HA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1612D, 0x1612F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] GURUNG KHEMA SIGN ANUSVARA..GURUNG KHEMA SIGN THOLHOMA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x16AF0, 0x16AF4 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] BASSA VAH COMBINING HIGH TONE..BASSA VAH COMBINING HIGH-LOW TONE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x16B30, 0x16B36 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] PAHAWH HMONG MARK CIM TUB..PAHAWH HMONG MARK CIM TAUM
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x16D63, 0x16D63 }, ::unicode::Grapheme_Cluster_Break::V }, // Lo       KIRAT RAI VOWEL SIGN AA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x16D67, 0x16D6A }, ::unicode::Grapheme_Cluster_Break::V }, // Lo   [4] KIRAT RAI VOWEL SIGN E..KIRAT RAI VOWEL SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x16F4F, 0x16F4F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       MIAO SIGN CONSONANT MODIFIER BAR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x16F51, 0x16F87 }, ::unicode::Grapheme_Cluster_Break::SpacingMark }, // Mc  [55] MIAO SIGN ASPIRATION..MIAO VOWEL SIGN UI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x16F8F, 0x16F92 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] MIAO TONE RIGHT..MIAO TONE BELOW
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x16FE4, 0x16FE4 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       KHITAN SMALL SCRIPT FILLER
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x16FF0, 0x16FF1 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc   [2] VIETNAMESE ALTERNATE READING MARK CA..VIETNAMESE ALTERNATE READING MARK NHAY
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BC9D, 0x1BC9E }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] DUPLOYAN THICK LETTER SELECTOR..DUPLOYAN DOUBLE MARK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1BCA0, 0x1BCA3 }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf   [4] SHORTHAND FORMAT LETTER OVERLAP..SHORTHAND FORMAT UP STEP
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1CF00, 0x1CF2D }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [46] ZNAMENNY COMBINING MARK GORAZDO NIZKO S KRYZHEM ON LEFT..ZNAMENNY COMBINING MARK KRYZH ON LEFT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1CF30, 0x1CF46 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [23] ZNAMENNY COMBINING TONAL RANGE MARK MRACHNO..ZNAMENNY PRIZNAK MODIFIER ROG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1D165, 0x1D166 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc   [2] MUSICAL SYMBOL COMBINING STEM..MUSICAL SYMBOL COMBINING SPRECHGESANG STEM
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1D167, 0x1D169 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] MUSICAL SYMBOL COMBINING TREMOLO-1..MUSICAL SYMBOL COMBINING TREMOLO-3
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1D16D, 0x1D172 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mc   [6] MUSICAL SYMBOL COMBINING AUGMENTATION DOT..MUSICAL SYMBOL COMBINING FLAG-5
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1D173, 0x1D17A }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf   [8] MUSICAL SYMBOL BEGIN BEAM..MUSICAL SYMBOL END PHRASE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1D17B, 0x1D182 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [8] MUSICAL SYMBOL COMBINING ACCENT..MUSICAL SYMBOL COMBINING LOURE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1D185, 0x1D18B }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] MUSICAL SYMBOL COMBINING DOIT..MUSICAL SYMBOL COMBINING TRIPLE TONGUE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1D1AA, 0x1D1AD }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] MUSICAL SYMBOL COMBINING DOWN BOW..MUSICAL SYMBOL COMBINING SNAP PIZZICATO
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1D242, 0x1D244 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [3] COMBINING GREEK MUSICAL TRISEME..COMBINING GREEK MUSICAL PENTASEME
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1DA00, 0x1DA36 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [55] SIGNWRITING HEAD RIM..SIGNWRITING AIR SUCKING IN
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1DA3B, 0x1DA6C }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [50] SIGNWRITING MOUTH CLOSED NEUTRAL..SIGNWRITING EXCITEMENT
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1DA75, 0x1DA75 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SIGNWRITING UPPER BODY TILTING FROM HIP JOINTS
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1DA84, 0x1DA84 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       SIGNWRITING LOCATION HEAD NECK
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1DA9B, 0x1DA9F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] SIGNWRITING FILL MODIFIER-2..SIGNWRITING FILL MODIFIER-6
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1DAA1, 0x1DAAF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [15] SIGNWRITING ROTATION MODIFIER-2..SIGNWRITING ROTATION MODIFIER-16
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E000, 0x1E006 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] COMBINING GLAGOLITIC LETTER AZU..COMBINING GLAGOLITIC LETTER ZHIVETE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E008, 0x1E018 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn  [17] COMBINING GLAGOLITIC LETTER ZEMLJA..COMBINING GLAGOLITIC LETTER HERU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E01B, 0x1E021 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] COMBINING GLAGOLITIC LETTER SHTA..COMBINING GLAGOLITIC LETTER YATI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E023, 0x1E024 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] COMBINING GLAGOLITIC LETTER YU..COMBINING GLAGOLITIC LETTER SMALL YUS
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E026, 0x1E02A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [5] COMBINING GLAGOLITIC LETTER YO..COMBINING GLAGOLITIC LETTER FITA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E08F, 0x1E08F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       COMBINING CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E130, 0x1E136 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] NYIAKENG PUACHUE HMONG TONE-B..NYIAKENG PUACHUE HMONG TONE-D
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E2AE, 0x1E2AE }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TOTO SIGN RISING TONE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E2EC, 0x1E2EF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] WANCHO TONE TUP..WANCHO TONE KOINI
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E4EC, 0x1E4EF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [4] NAG MUNDARI SIGN MUHOR..NAG MUNDARI SIGN SUTUH
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E5EE, 0x1E5EF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] OL ONAL SIGN MU..OL ONAL SIGN IKIR
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E6E3, 0x1E6E3 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAI YO SIGN UE
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E6E6, 0x1E6E6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAI YO SIGN AU
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E6EE, 0x1E6EF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [2] TAI YO SIGN AY..TAI YO SIGN ANG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E6F5, 0x1E6F5 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn       TAI YO SIGN OM
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E8D0, 0x1E8D6 }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] MENDE KIKAKUI COMBINING NUMBER TEENS..MENDE KIKAKUI COMBINING NUMBER MILLIONS
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1E944, 0x1E94A }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn   [7] ADLAM ALIF LENGTHENER..ADLAM NUKTA
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1F1E6, 0x1F1FF }, ::unicode::Grapheme_Cluster_Break::Regional_Indicator }, // So  [26] REGIONAL INDICATOR SYMBOL LETTER A..REGIONAL INDICATOR SYMBOL LETTER Z
    Prop<::unicode::Grapheme_Cluster_Break> { { 0x1F3FB, 0x1F3FF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Sk   [5] EMOJI MODIFIER FITZPATRICK TYPE-1-2..EMOJI MODIFIER FITZPATRICK TYPE-6
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xE0000, 0xE0000 }, ::unicode::Grapheme_Cluster_Break::Control }, // Cn       <reserved-E0000>
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xE0001, 0xE0001 }, ::unicode::Grapheme_Cluster_Break::Control }, // Cf       LANGUAGE TAG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xE0002, 0xE001F }, ::unicode::Grapheme_Cluster_Break::Control }, // Cn  [30] <reserved-E0002>..<reserved-E001F>
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xE0020, 0xE007F }, ::unicode::Grapheme_Cluster_Break::Extend }, // Cf  [96] TAG SPACE..CANCEL TAG
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xE0080, 0xE00FF }, ::unicode::Grapheme_Cluster_Break::Control }, // Cn [128] <reserved-E0080>..<reserved-E00FF>
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xE0100, 0xE01EF }, ::unicode::Grapheme_Cluster_Break::Extend }, // Mn [240] VARIATION SELECTOR-17..VARIATION SELECTOR-256
    Prop<::unicode::Grapheme_Cluster_Break> { { 0xE01F0, 0xE0FFF }, ::unicode::Grapheme_Cluster_Break::Control }, // Cn [3600] <reserved-E01F0>..<reserved-E0FFF>
}; // }}}
} // end namespace tables

Grapheme_Cluster_Break grapheme_cluster_break(char32_t codepoint) noexcept {
    return search(tables::Grapheme_Cluster_Break, codepoint).value_or(Grapheme_Cluster_Break::Undefined);
}


namespace tables {
auto static const EastAsianWidth = std::array<Prop<::unicode::EastAsianWidth>, 1201>{ // {{{
    Prop<::unicode::EastAsianWidth> { { 0x0000, 0x001F }, ::unicode::EastAsianWidth::Neutral }, // Cc    [32] <control-0000>..<control-001F>
    // 0x0020 .. 0x0020: Zs         SPACE
    // 0x0021 .. 0x0023: Po     [3] EXCLAMATION MARK..NUMBER SIGN
    // 0x0024 .. 0x0024: Sc         DOLLAR SIGN
    // 0x0025 .. 0x0027: Po     [3] PERCENT SIGN..APOSTROPHE
    // 0x0028 .. 0x0028: Ps         LEFT PARENTHESIS
    // 0x0029 .. 0x0029: Pe         RIGHT PARENTHESIS
    // 0x002A .. 0x002A: Po         ASTERISK
    // 0x002B .. 0x002B: Sm         PLUS SIGN
    // 0x002C .. 0x002C: Po         COMMA
    // 0x002D .. 0x002D: Pd         HYPHEN-MINUS
    // 0x002E .. 0x002F: Po     [2] FULL STOP..SOLIDUS
    // 0x0030 .. 0x0039: Nd    [10] DIGIT ZERO..DIGIT NINE
    // 0x003A .. 0x003B: Po     [2] COLON..SEMICOLON
    // 0x003C .. 0x003E: Sm     [3] LESS-THAN SIGN..GREATER-THAN SIGN
    // 0x003F .. 0x0040: Po     [2] QUESTION MARK..COMMERCIAL AT
    // 0x0041 .. 0x005A: Lu    [26] LATIN CAPITAL LETTER A..LATIN CAPITAL LETTER Z
    // 0x005B .. 0x005B: Ps         LEFT SQUARE BRACKET
    // 0x005C .. 0x005C: Po         REVERSE SOLIDUS
    // 0x005D .. 0x005D: Pe         RIGHT SQUARE BRACKET
    // 0x005E .. 0x005E: Sk         CIRCUMFLEX ACCENT
    // 0x005F .. 0x005F: Pc         LOW LINE
    // 0x0060 .. 0x0060: Sk         GRAVE ACCENT
    // 0x0061 .. 0x007A: Ll    [26] LATIN SMALL LETTER A..LATIN SMALL LETTER Z
    // 0x007B .. 0x007B: Ps         LEFT CURLY BRACKET
    // 0x007C .. 0x007C: Sm         VERTICAL LINE
    // 0x007D .. 0x007D: Pe         RIGHT CURLY BRACKET
    // 0x007E .. 0x007E: Sm         TILDE
    Prop<::unicode::EastAsianWidth> { { 0x0020, 0x007E }, ::unicode::EastAsianWidth::Narrow }, // #27
    // 0x007F .. 0x007F: Cc         <control-007F>
    // 0x0080 .. 0x009F: Cc    [32] <control-0080>..<control-009F>
    // 0x00A0 .. 0x00A0: Zs         NO-BREAK SPACE
    Prop<::unicode::EastAsianWidth> { { 0x007F, 0x00A0 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x00A1, 0x00A1 }, ::unicode::EastAsianWidth::Ambiguous }, // Po         INVERTED EXCLAMATION MARK
    Prop<::unicode::EastAsianWidth> { { 0x00A2, 0x00A3 }, ::unicode::EastAsianWidth::Narrow }, // Sc     [2] CENT SIGN..POUND SIGN
    Prop<::unicode::EastAsianWidth> { { 0x00A4, 0x00A4 }, ::unicode::EastAsianWidth::Ambiguous }, // Sc         CURRENCY SIGN
    // 0x00A5 .. 0x00A5: Sc         YEN SIGN
    // 0x00A6 .. 0x00A6: So         BROKEN BAR
    Prop<::unicode::EastAsianWidth> { { 0x00A5, 0x00A6 }, ::unicode::EastAsianWidth::Narrow }, // #2
    // 0x00A7 .. 0x00A7: Po         SECTION SIGN
    // 0x00A8 .. 0x00A8: Sk         DIAERESIS
    Prop<::unicode::EastAsianWidth> { { 0x00A7, 0x00A8 }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x00A9, 0x00A9 }, ::unicode::EastAsianWidth::Neutral }, // So         COPYRIGHT SIGN
    Prop<::unicode::EastAsianWidth> { { 0x00AA, 0x00AA }, ::unicode::EastAsianWidth::Ambiguous }, // Lo         FEMININE ORDINAL INDICATOR
    Prop<::unicode::EastAsianWidth> { { 0x00AB, 0x00AB }, ::unicode::EastAsianWidth::Neutral }, // Pi         LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
    Prop<::unicode::EastAsianWidth> { { 0x00AC, 0x00AC }, ::unicode::EastAsianWidth::Narrow }, // Sm         NOT SIGN
    // 0x00AD .. 0x00AD: Cf         SOFT HYPHEN
    // 0x00AE .. 0x00AE: So         REGISTERED SIGN
    Prop<::unicode::EastAsianWidth> { { 0x00AD, 0x00AE }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x00AF, 0x00AF }, ::unicode::EastAsianWidth::Narrow }, // Sk         MACRON
    // 0x00B0 .. 0x00B0: So         DEGREE SIGN
    // 0x00B1 .. 0x00B1: Sm         PLUS-MINUS SIGN
    // 0x00B2 .. 0x00B3: No     [2] SUPERSCRIPT TWO..SUPERSCRIPT THREE
    // 0x00B4 .. 0x00B4: Sk         ACUTE ACCENT
    Prop<::unicode::EastAsianWidth> { { 0x00B0, 0x00B4 }, ::unicode::EastAsianWidth::Ambiguous }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x00B5, 0x00B5 }, ::unicode::EastAsianWidth::Neutral }, // Ll         MICRO SIGN
    // 0x00B6 .. 0x00B7: Po     [2] PILCROW SIGN..MIDDLE DOT
    // 0x00B8 .. 0x00B8: Sk         CEDILLA
    // 0x00B9 .. 0x00B9: No         SUPERSCRIPT ONE
    // 0x00BA .. 0x00BA: Lo         MASCULINE ORDINAL INDICATOR
    Prop<::unicode::EastAsianWidth> { { 0x00B6, 0x00BA }, ::unicode::EastAsianWidth::Ambiguous }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x00BB, 0x00BB }, ::unicode::EastAsianWidth::Neutral }, // Pf         RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
    // 0x00BC .. 0x00BE: No     [3] VULGAR FRACTION ONE QUARTER..VULGAR FRACTION THREE QUARTERS
    // 0x00BF .. 0x00BF: Po         INVERTED QUESTION MARK
    Prop<::unicode::EastAsianWidth> { { 0x00BC, 0x00BF }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x00C0, 0x00C5 }, ::unicode::EastAsianWidth::Neutral }, // Lu     [6] LATIN CAPITAL LETTER A WITH GRAVE..LATIN CAPITAL LETTER A WITH RING ABOVE
    Prop<::unicode::EastAsianWidth> { { 0x00C6, 0x00C6 }, ::unicode::EastAsianWidth::Ambiguous }, // Lu         LATIN CAPITAL LETTER AE
    Prop<::unicode::EastAsianWidth> { { 0x00C7, 0x00CF }, ::unicode::EastAsianWidth::Neutral }, // Lu     [9] LATIN CAPITAL LETTER C WITH CEDILLA..LATIN CAPITAL LETTER I WITH DIAERESIS
    Prop<::unicode::EastAsianWidth> { { 0x00D0, 0x00D0 }, ::unicode::EastAsianWidth::Ambiguous }, // Lu         LATIN CAPITAL LETTER ETH
    Prop<::unicode::EastAsianWidth> { { 0x00D1, 0x00D6 }, ::unicode::EastAsianWidth::Neutral }, // Lu     [6] LATIN CAPITAL LETTER N WITH TILDE..LATIN CAPITAL LETTER O WITH DIAERESIS
    // 0x00D7 .. 0x00D7: Sm         MULTIPLICATION SIGN
    // 0x00D8 .. 0x00D8: Lu         LATIN CAPITAL LETTER O WITH STROKE
    Prop<::unicode::EastAsianWidth> { { 0x00D7, 0x00D8 }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x00D9, 0x00DD }, ::unicode::EastAsianWidth::Neutral }, // Lu     [5] LATIN CAPITAL LETTER U WITH GRAVE..LATIN CAPITAL LETTER Y WITH ACUTE
    Prop<::unicode::EastAsianWidth> { { 0x00DE, 0x00E1 }, ::unicode::EastAsianWidth::Ambiguous }, // L&     [4] LATIN CAPITAL LETTER THORN..LATIN SMALL LETTER A WITH ACUTE
    Prop<::unicode::EastAsianWidth> { { 0x00E2, 0x00E5 }, ::unicode::EastAsianWidth::Neutral }, // Ll     [4] LATIN SMALL LETTER A WITH CIRCUMFLEX..LATIN SMALL LETTER A WITH RING ABOVE
    Prop<::unicode::EastAsianWidth> { { 0x00E6, 0x00E6 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER AE
    Prop<::unicode::EastAsianWidth> { { 0x00E7, 0x00E7 }, ::unicode::EastAsianWidth::Neutral }, // Ll         LATIN SMALL LETTER C WITH CEDILLA
    Prop<::unicode::EastAsianWidth> { { 0x00E8, 0x00EA }, ::unicode::EastAsianWidth::Ambiguous }, // Ll     [3] LATIN SMALL LETTER E WITH GRAVE..LATIN SMALL LETTER E WITH CIRCUMFLEX
    Prop<::unicode::EastAsianWidth> { { 0x00EB, 0x00EB }, ::unicode::EastAsianWidth::Neutral }, // Ll         LATIN SMALL LETTER E WITH DIAERESIS
    Prop<::unicode::EastAsianWidth> { { 0x00EC, 0x00ED }, ::unicode::EastAsianWidth::Ambiguous }, // Ll     [2] LATIN SMALL LETTER I WITH GRAVE..LATIN SMALL LETTER I WITH ACUTE
    Prop<::unicode::EastAsianWidth> { { 0x00EE, 0x00EF }, ::unicode::EastAsianWidth::Neutral }, // Ll     [2] LATIN SMALL LETTER I WITH CIRCUMFLEX..LATIN SMALL LETTER I WITH DIAERESIS
    Prop<::unicode::EastAsianWidth> { { 0x00F0, 0x00F0 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER ETH
    Prop<::unicode::EastAsianWidth> { { 0x00F1, 0x00F1 }, ::unicode::EastAsianWidth::Neutral }, // Ll         LATIN SMALL LETTER N WITH TILDE
    Prop<::unicode::EastAsianWidth> { { 0x00F2, 0x00F3 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll     [2] LATIN SMALL LETTER O WITH GRAVE..LATIN SMALL LETTER O WITH ACUTE
    Prop<::unicode::EastAsianWidth> { { 0x00F4, 0x00F6 }, ::unicode::EastAsianWidth::Neutral }, // Ll     [3] LATIN SMALL LETTER O WITH CIRCUMFLEX..LATIN SMALL LETTER O WITH DIAERESIS
    // 0x00F7 .. 0x00F7: Sm         DIVISION SIGN
    // 0x00F8 .. 0x00FA: Ll     [3] LATIN SMALL LETTER O WITH STROKE..LATIN SMALL LETTER U WITH ACUTE
    Prop<::unicode::EastAsianWidth> { { 0x00F7, 0x00FA }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x00FB, 0x00FB }, ::unicode::EastAsianWidth::Neutral }, // Ll         LATIN SMALL LETTER U WITH CIRCUMFLEX
    Prop<::unicode::EastAsianWidth> { { 0x00FC, 0x00FC }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER U WITH DIAERESIS
    Prop<::unicode::EastAsianWidth> { { 0x00FD, 0x00FD }, ::unicode::EastAsianWidth::Neutral }, // Ll         LATIN SMALL LETTER Y WITH ACUTE
    Prop<::unicode::EastAsianWidth> { { 0x00FE, 0x00FE }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER THORN
    // 0x00FF .. 0x00FF: Ll         LATIN SMALL LETTER Y WITH DIAERESIS
    // 0x0100 .. 0x0100: Lu         LATIN CAPITAL LETTER A WITH MACRON
    Prop<::unicode::EastAsianWidth> { { 0x00FF, 0x0100 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0101, 0x0101 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER A WITH MACRON
    Prop<::unicode::EastAsianWidth> { { 0x0102, 0x0110 }, ::unicode::EastAsianWidth::Neutral }, // L&    [15] LATIN CAPITAL LETTER A WITH BREVE..LATIN CAPITAL LETTER D WITH STROKE
    Prop<::unicode::EastAsianWidth> { { 0x0111, 0x0111 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER D WITH STROKE
    Prop<::unicode::EastAsianWidth> { { 0x0112, 0x0112 }, ::unicode::EastAsianWidth::Neutral }, // Lu         LATIN CAPITAL LETTER E WITH MACRON
    Prop<::unicode::EastAsianWidth> { { 0x0113, 0x0113 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER E WITH MACRON
    Prop<::unicode::EastAsianWidth> { { 0x0114, 0x011A }, ::unicode::EastAsianWidth::Neutral }, // L&     [7] LATIN CAPITAL LETTER E WITH BREVE..LATIN CAPITAL LETTER E WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x011B, 0x011B }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER E WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x011C, 0x0125 }, ::unicode::EastAsianWidth::Neutral }, // L&    [10] LATIN CAPITAL LETTER G WITH CIRCUMFLEX..LATIN SMALL LETTER H WITH CIRCUMFLEX
    Prop<::unicode::EastAsianWidth> { { 0x0126, 0x0127 }, ::unicode::EastAsianWidth::Ambiguous }, // L&     [2] LATIN CAPITAL LETTER H WITH STROKE..LATIN SMALL LETTER H WITH STROKE
    Prop<::unicode::EastAsianWidth> { { 0x0128, 0x012A }, ::unicode::EastAsianWidth::Neutral }, // L&     [3] LATIN CAPITAL LETTER I WITH TILDE..LATIN CAPITAL LETTER I WITH MACRON
    Prop<::unicode::EastAsianWidth> { { 0x012B, 0x012B }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER I WITH MACRON
    Prop<::unicode::EastAsianWidth> { { 0x012C, 0x0130 }, ::unicode::EastAsianWidth::Neutral }, // L&     [5] LATIN CAPITAL LETTER I WITH BREVE..LATIN CAPITAL LETTER I WITH DOT ABOVE
    Prop<::unicode::EastAsianWidth> { { 0x0131, 0x0133 }, ::unicode::EastAsianWidth::Ambiguous }, // L&     [3] LATIN SMALL LETTER DOTLESS I..LATIN SMALL LIGATURE IJ
    Prop<::unicode::EastAsianWidth> { { 0x0134, 0x0137 }, ::unicode::EastAsianWidth::Neutral }, // L&     [4] LATIN CAPITAL LETTER J WITH CIRCUMFLEX..LATIN SMALL LETTER K WITH CEDILLA
    Prop<::unicode::EastAsianWidth> { { 0x0138, 0x0138 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER KRA
    Prop<::unicode::EastAsianWidth> { { 0x0139, 0x013E }, ::unicode::EastAsianWidth::Neutral }, // L&     [6] LATIN CAPITAL LETTER L WITH ACUTE..LATIN SMALL LETTER L WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x013F, 0x0142 }, ::unicode::EastAsianWidth::Ambiguous }, // L&     [4] LATIN CAPITAL LETTER L WITH MIDDLE DOT..LATIN SMALL LETTER L WITH STROKE
    Prop<::unicode::EastAsianWidth> { { 0x0143, 0x0143 }, ::unicode::EastAsianWidth::Neutral }, // Lu         LATIN CAPITAL LETTER N WITH ACUTE
    Prop<::unicode::EastAsianWidth> { { 0x0144, 0x0144 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER N WITH ACUTE
    Prop<::unicode::EastAsianWidth> { { 0x0145, 0x0147 }, ::unicode::EastAsianWidth::Neutral }, // L&     [3] LATIN CAPITAL LETTER N WITH CEDILLA..LATIN CAPITAL LETTER N WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x0148, 0x014B }, ::unicode::EastAsianWidth::Ambiguous }, // L&     [4] LATIN SMALL LETTER N WITH CARON..LATIN SMALL LETTER ENG
    Prop<::unicode::EastAsianWidth> { { 0x014C, 0x014C }, ::unicode::EastAsianWidth::Neutral }, // Lu         LATIN CAPITAL LETTER O WITH MACRON
    Prop<::unicode::EastAsianWidth> { { 0x014D, 0x014D }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER O WITH MACRON
    Prop<::unicode::EastAsianWidth> { { 0x014E, 0x0151 }, ::unicode::EastAsianWidth::Neutral }, // L&     [4] LATIN CAPITAL LETTER O WITH BREVE..LATIN SMALL LETTER O WITH DOUBLE ACUTE
    Prop<::unicode::EastAsianWidth> { { 0x0152, 0x0153 }, ::unicode::EastAsianWidth::Ambiguous }, // L&     [2] LATIN CAPITAL LIGATURE OE..LATIN SMALL LIGATURE OE
    Prop<::unicode::EastAsianWidth> { { 0x0154, 0x0165 }, ::unicode::EastAsianWidth::Neutral }, // L&    [18] LATIN CAPITAL LETTER R WITH ACUTE..LATIN SMALL LETTER T WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x0166, 0x0167 }, ::unicode::EastAsianWidth::Ambiguous }, // L&     [2] LATIN CAPITAL LETTER T WITH STROKE..LATIN SMALL LETTER T WITH STROKE
    Prop<::unicode::EastAsianWidth> { { 0x0168, 0x016A }, ::unicode::EastAsianWidth::Neutral }, // L&     [3] LATIN CAPITAL LETTER U WITH TILDE..LATIN CAPITAL LETTER U WITH MACRON
    Prop<::unicode::EastAsianWidth> { { 0x016B, 0x016B }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER U WITH MACRON
    // 0x016C .. 0x017F: L&    [20] LATIN CAPITAL LETTER U WITH BREVE..LATIN SMALL LETTER LONG S
    // 0x0180 .. 0x01BA: L&    [59] LATIN SMALL LETTER B WITH STROKE..LATIN SMALL LETTER EZH WITH TAIL
    // 0x01BB .. 0x01BB: Lo         LATIN LETTER TWO WITH STROKE
    // 0x01BC .. 0x01BF: L&     [4] LATIN CAPITAL LETTER TONE FIVE..LATIN LETTER WYNN
    // 0x01C0 .. 0x01C3: Lo     [4] LATIN LETTER DENTAL CLICK..LATIN LETTER RETROFLEX CLICK
    // 0x01C4 .. 0x01CD: L&    [10] LATIN CAPITAL LETTER DZ WITH CARON..LATIN CAPITAL LETTER A WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x016C, 0x01CD }, ::unicode::EastAsianWidth::Neutral }, // #6
    Prop<::unicode::EastAsianWidth> { { 0x01CE, 0x01CE }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER A WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x01CF, 0x01CF }, ::unicode::EastAsianWidth::Neutral }, // Lu         LATIN CAPITAL LETTER I WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x01D0, 0x01D0 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER I WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x01D1, 0x01D1 }, ::unicode::EastAsianWidth::Neutral }, // Lu         LATIN CAPITAL LETTER O WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x01D2, 0x01D2 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER O WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x01D3, 0x01D3 }, ::unicode::EastAsianWidth::Neutral }, // Lu         LATIN CAPITAL LETTER U WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x01D4, 0x01D4 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER U WITH CARON
    Prop<::unicode::EastAsianWidth> { { 0x01D5, 0x01D5 }, ::unicode::EastAsianWidth::Neutral }, // Lu         LATIN CAPITAL LETTER U WITH DIAERESIS AND MACRON
    Prop<::unicode::EastAsianWidth> { { 0x01D6, 0x01D6 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER U WITH DIAERESIS AND MACRON
    Prop<::unicode::EastAsianWidth> { { 0x01D7, 0x01D7 }, ::unicode::EastAsianWidth::Neutral }, // Lu         LATIN CAPITAL LETTER U WITH DIAERESIS AND ACUTE
    Prop<::unicode::EastAsianWidth> { { 0x01D8, 0x01D8 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER U WITH DIAERESIS AND ACUTE
    Prop<::unicode::EastAsianWidth> { { 0x01D9, 0x01D9 }, ::unicode::EastAsianWidth::Neutral }, // Lu         LATIN CAPITAL LETTER U WITH DIAERESIS AND CARON
    Prop<::unicode::EastAsianWidth> { { 0x01DA, 0x01DA }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER U WITH DIAERESIS AND CARON
    Prop<::unicode::EastAsianWidth> { { 0x01DB, 0x01DB }, ::unicode::EastAsianWidth::Neutral }, // Lu         LATIN CAPITAL LETTER U WITH DIAERESIS AND GRAVE
    Prop<::unicode::EastAsianWidth> { { 0x01DC, 0x01DC }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER U WITH DIAERESIS AND GRAVE
    // 0x01DD .. 0x024F: L&   [115] LATIN SMALL LETTER TURNED E..LATIN SMALL LETTER Y WITH STROKE
    // 0x0250 .. 0x0250: Ll         LATIN SMALL LETTER TURNED A
    Prop<::unicode::EastAsianWidth> { { 0x01DD, 0x0250 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0251, 0x0251 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER ALPHA
    Prop<::unicode::EastAsianWidth> { { 0x0252, 0x0260 }, ::unicode::EastAsianWidth::Neutral }, // Ll    [15] LATIN SMALL LETTER TURNED ALPHA..LATIN SMALL LETTER G WITH HOOK
    Prop<::unicode::EastAsianWidth> { { 0x0261, 0x0261 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         LATIN SMALL LETTER SCRIPT G
    // 0x0262 .. 0x0293: Ll    [50] LATIN LETTER SMALL CAPITAL G..LATIN SMALL LETTER EZH WITH CURL
    // 0x0294 .. 0x0295: Lo     [2] LATIN LETTER GLOTTAL STOP..LATIN LETTER PHARYNGEAL VOICED FRICATIVE
    // 0x0296 .. 0x02AF: Ll    [26] LATIN LETTER INVERTED GLOTTAL STOP..LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
    // 0x02B0 .. 0x02C1: Lm    [18] MODIFIER LETTER SMALL H..MODIFIER LETTER REVERSED GLOTTAL STOP
    // 0x02C2 .. 0x02C3: Sk     [2] MODIFIER LETTER LEFT ARROWHEAD..MODIFIER LETTER RIGHT ARROWHEAD
    Prop<::unicode::EastAsianWidth> { { 0x0262, 0x02C3 }, ::unicode::EastAsianWidth::Neutral }, // #5
    Prop<::unicode::EastAsianWidth> { { 0x02C4, 0x02C4 }, ::unicode::EastAsianWidth::Ambiguous }, // Sk         MODIFIER LETTER UP ARROWHEAD
    // 0x02C5 .. 0x02C5: Sk         MODIFIER LETTER DOWN ARROWHEAD
    // 0x02C6 .. 0x02C6: Lm         MODIFIER LETTER CIRCUMFLEX ACCENT
    Prop<::unicode::EastAsianWidth> { { 0x02C5, 0x02C6 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x02C7, 0x02C7 }, ::unicode::EastAsianWidth::Ambiguous }, // Lm         CARON
    Prop<::unicode::EastAsianWidth> { { 0x02C8, 0x02C8 }, ::unicode::EastAsianWidth::Neutral }, // Lm         MODIFIER LETTER VERTICAL LINE
    Prop<::unicode::EastAsianWidth> { { 0x02C9, 0x02CB }, ::unicode::EastAsianWidth::Ambiguous }, // Lm     [3] MODIFIER LETTER MACRON..MODIFIER LETTER GRAVE ACCENT
    Prop<::unicode::EastAsianWidth> { { 0x02CC, 0x02CC }, ::unicode::EastAsianWidth::Neutral }, // Lm         MODIFIER LETTER LOW VERTICAL LINE
    Prop<::unicode::EastAsianWidth> { { 0x02CD, 0x02CD }, ::unicode::EastAsianWidth::Ambiguous }, // Lm         MODIFIER LETTER LOW MACRON
    Prop<::unicode::EastAsianWidth> { { 0x02CE, 0x02CF }, ::unicode::EastAsianWidth::Neutral }, // Lm     [2] MODIFIER LETTER LOW GRAVE ACCENT..MODIFIER LETTER LOW ACUTE ACCENT
    Prop<::unicode::EastAsianWidth> { { 0x02D0, 0x02D0 }, ::unicode::EastAsianWidth::Ambiguous }, // Lm         MODIFIER LETTER TRIANGULAR COLON
    // 0x02D1 .. 0x02D1: Lm         MODIFIER LETTER HALF TRIANGULAR COLON
    // 0x02D2 .. 0x02D7: Sk     [6] MODIFIER LETTER CENTRED RIGHT HALF RING..MODIFIER LETTER MINUS SIGN
    Prop<::unicode::EastAsianWidth> { { 0x02D1, 0x02D7 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x02D8, 0x02DB }, ::unicode::EastAsianWidth::Ambiguous }, // Sk     [4] BREVE..OGONEK
    Prop<::unicode::EastAsianWidth> { { 0x02DC, 0x02DC }, ::unicode::EastAsianWidth::Neutral }, // Sk         SMALL TILDE
    Prop<::unicode::EastAsianWidth> { { 0x02DD, 0x02DD }, ::unicode::EastAsianWidth::Ambiguous }, // Sk         DOUBLE ACUTE ACCENT
    Prop<::unicode::EastAsianWidth> { { 0x02DE, 0x02DE }, ::unicode::EastAsianWidth::Neutral }, // Sk         MODIFIER LETTER RHOTIC HOOK
    Prop<::unicode::EastAsianWidth> { { 0x02DF, 0x02DF }, ::unicode::EastAsianWidth::Ambiguous }, // Sk         MODIFIER LETTER CROSS ACCENT
    // 0x02E0 .. 0x02E4: Lm     [5] MODIFIER LETTER SMALL GAMMA..MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
    // 0x02E5 .. 0x02EB: Sk     [7] MODIFIER LETTER EXTRA-HIGH TONE BAR..MODIFIER LETTER YANG DEPARTING TONE MARK
    // 0x02EC .. 0x02EC: Lm         MODIFIER LETTER VOICING
    // 0x02ED .. 0x02ED: Sk         MODIFIER LETTER UNASPIRATED
    // 0x02EE .. 0x02EE: Lm         MODIFIER LETTER DOUBLE APOSTROPHE
    // 0x02EF .. 0x02FF: Sk    [17] MODIFIER LETTER LOW DOWN ARROWHEAD..MODIFIER LETTER LOW LEFT ARROW
    Prop<::unicode::EastAsianWidth> { { 0x02E0, 0x02FF }, ::unicode::EastAsianWidth::Neutral }, // #6
    Prop<::unicode::EastAsianWidth> { { 0x0300, 0x036F }, ::unicode::EastAsianWidth::Ambiguous }, // Mn   [112] COMBINING GRAVE ACCENT..COMBINING LATIN SMALL LETTER X
    // 0x0370 .. 0x0373: L&     [4] GREEK CAPITAL LETTER HETA..GREEK SMALL LETTER ARCHAIC SAMPI
    // 0x0374 .. 0x0374: Lm         GREEK NUMERAL SIGN
    // 0x0375 .. 0x0375: Sk         GREEK LOWER NUMERAL SIGN
    // 0x0376 .. 0x0377: L&     [2] GREEK CAPITAL LETTER PAMPHYLIAN DIGAMMA..GREEK SMALL LETTER PAMPHYLIAN DIGAMMA
    Prop<::unicode::EastAsianWidth> { { 0x0370, 0x0377 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x037A .. 0x037A: Lm         GREEK YPOGEGRAMMENI
    // 0x037B .. 0x037D: Ll     [3] GREEK SMALL REVERSED LUNATE SIGMA SYMBOL..GREEK SMALL REVERSED DOTTED LUNATE SIGMA SYMBOL
    // 0x037E .. 0x037E: Po         GREEK QUESTION MARK
    // 0x037F .. 0x037F: Lu         GREEK CAPITAL LETTER YOT
    Prop<::unicode::EastAsianWidth> { { 0x037A, 0x037F }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x0384 .. 0x0385: Sk     [2] GREEK TONOS..GREEK DIALYTIKA TONOS
    // 0x0386 .. 0x0386: Lu         GREEK CAPITAL LETTER ALPHA WITH TONOS
    // 0x0387 .. 0x0387: Po         GREEK ANO TELEIA
    // 0x0388 .. 0x038A: Lu     [3] GREEK CAPITAL LETTER EPSILON WITH TONOS..GREEK CAPITAL LETTER IOTA WITH TONOS
    Prop<::unicode::EastAsianWidth> { { 0x0384, 0x038A }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x038C, 0x038C }, ::unicode::EastAsianWidth::Neutral }, // Lu         GREEK CAPITAL LETTER OMICRON WITH TONOS
    Prop<::unicode::EastAsianWidth> { { 0x038E, 0x0390 }, ::unicode::EastAsianWidth::Neutral }, // L&     [3] GREEK CAPITAL LETTER UPSILON WITH TONOS..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND TONOS
    Prop<::unicode::EastAsianWidth> { { 0x0391, 0x03A1 }, ::unicode::EastAsianWidth::Ambiguous }, // Lu    [17] GREEK CAPITAL LETTER ALPHA..GREEK CAPITAL LETTER RHO
    Prop<::unicode::EastAsianWidth> { { 0x03A3, 0x03A9 }, ::unicode::EastAsianWidth::Ambiguous }, // Lu     [7] GREEK CAPITAL LETTER SIGMA..GREEK CAPITAL LETTER OMEGA
    Prop<::unicode::EastAsianWidth> { { 0x03AA, 0x03B0 }, ::unicode::EastAsianWidth::Neutral }, // L&     [7] GREEK CAPITAL LETTER IOTA WITH DIALYTIKA..GREEK SMALL LETTER UPSILON WITH DIALYTIKA AND TONOS
    Prop<::unicode::EastAsianWidth> { { 0x03B1, 0x03C1 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll    [17] GREEK SMALL LETTER ALPHA..GREEK SMALL LETTER RHO
    Prop<::unicode::EastAsianWidth> { { 0x03C2, 0x03C2 }, ::unicode::EastAsianWidth::Neutral }, // Ll         GREEK SMALL LETTER FINAL SIGMA
    Prop<::unicode::EastAsianWidth> { { 0x03C3, 0x03C9 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll     [7] GREEK SMALL LETTER SIGMA..GREEK SMALL LETTER OMEGA
    // 0x03CA .. 0x03F5: L&    [44] GREEK SMALL LETTER IOTA WITH DIALYTIKA..GREEK LUNATE EPSILON SYMBOL
    // 0x03F6 .. 0x03F6: Sm         GREEK REVERSED LUNATE EPSILON SYMBOL
    // 0x03F7 .. 0x03FF: L&     [9] GREEK CAPITAL LETTER SHO..GREEK CAPITAL REVERSED DOTTED LUNATE SIGMA SYMBOL
    // 0x0400 .. 0x0400: Lu         CYRILLIC CAPITAL LETTER IE WITH GRAVE
    Prop<::unicode::EastAsianWidth> { { 0x03CA, 0x0400 }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x0401, 0x0401 }, ::unicode::EastAsianWidth::Ambiguous }, // Lu         CYRILLIC CAPITAL LETTER IO
    Prop<::unicode::EastAsianWidth> { { 0x0402, 0x040F }, ::unicode::EastAsianWidth::Neutral }, // Lu    [14] CYRILLIC CAPITAL LETTER DJE..CYRILLIC CAPITAL LETTER DZHE
    Prop<::unicode::EastAsianWidth> { { 0x0410, 0x044F }, ::unicode::EastAsianWidth::Ambiguous }, // L&    [64] CYRILLIC CAPITAL LETTER A..CYRILLIC SMALL LETTER YA
    Prop<::unicode::EastAsianWidth> { { 0x0450, 0x0450 }, ::unicode::EastAsianWidth::Neutral }, // Ll         CYRILLIC SMALL LETTER IE WITH GRAVE
    Prop<::unicode::EastAsianWidth> { { 0x0451, 0x0451 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         CYRILLIC SMALL LETTER IO
    // 0x0452 .. 0x0481: L&    [48] CYRILLIC SMALL LETTER DJE..CYRILLIC SMALL LETTER KOPPA
    // 0x0482 .. 0x0482: So         CYRILLIC THOUSANDS SIGN
    // 0x0483 .. 0x0487: Mn     [5] COMBINING CYRILLIC TITLO..COMBINING CYRILLIC POKRYTIE
    // 0x0488 .. 0x0489: Me     [2] COMBINING CYRILLIC HUNDRED THOUSANDS SIGN..COMBINING CYRILLIC MILLIONS SIGN
    // 0x048A .. 0x04FF: L&   [118] CYRILLIC CAPITAL LETTER SHORT I WITH TAIL..CYRILLIC SMALL LETTER HA WITH STROKE
    // 0x0500 .. 0x052F: L&    [48] CYRILLIC CAPITAL LETTER KOMI DE..CYRILLIC SMALL LETTER EL WITH DESCENDER
    Prop<::unicode::EastAsianWidth> { { 0x0452, 0x052F }, ::unicode::EastAsianWidth::Neutral }, // #6
    Prop<::unicode::EastAsianWidth> { { 0x0531, 0x0556 }, ::unicode::EastAsianWidth::Neutral }, // Lu    [38] ARMENIAN CAPITAL LETTER AYB..ARMENIAN CAPITAL LETTER FEH
    // 0x0559 .. 0x0559: Lm         ARMENIAN MODIFIER LETTER LEFT HALF RING
    // 0x055A .. 0x055F: Po     [6] ARMENIAN APOSTROPHE..ARMENIAN ABBREVIATION MARK
    // 0x0560 .. 0x0588: Ll    [41] ARMENIAN SMALL LETTER TURNED AYB..ARMENIAN SMALL LETTER YI WITH STROKE
    // 0x0589 .. 0x0589: Po         ARMENIAN FULL STOP
    // 0x058A .. 0x058A: Pd         ARMENIAN HYPHEN
    Prop<::unicode::EastAsianWidth> { { 0x0559, 0x058A }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x058D .. 0x058E: So     [2] RIGHT-FACING ARMENIAN ETERNITY SIGN..LEFT-FACING ARMENIAN ETERNITY SIGN
    // 0x058F .. 0x058F: Sc         ARMENIAN DRAM SIGN
    Prop<::unicode::EastAsianWidth> { { 0x058D, 0x058F }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x0591 .. 0x05BD: Mn    [45] HEBREW ACCENT ETNAHTA..HEBREW POINT METEG
    // 0x05BE .. 0x05BE: Pd         HEBREW PUNCTUATION MAQAF
    // 0x05BF .. 0x05BF: Mn         HEBREW POINT RAFE
    // 0x05C0 .. 0x05C0: Po         HEBREW PUNCTUATION PASEQ
    // 0x05C1 .. 0x05C2: Mn     [2] HEBREW POINT SHIN DOT..HEBREW POINT SIN DOT
    // 0x05C3 .. 0x05C3: Po         HEBREW PUNCTUATION SOF PASUQ
    // 0x05C4 .. 0x05C5: Mn     [2] HEBREW MARK UPPER DOT..HEBREW MARK LOWER DOT
    // 0x05C6 .. 0x05C6: Po         HEBREW PUNCTUATION NUN HAFUKHA
    // 0x05C7 .. 0x05C7: Mn         HEBREW POINT QAMATS QATAN
    Prop<::unicode::EastAsianWidth> { { 0x0591, 0x05C7 }, ::unicode::EastAsianWidth::Neutral }, // #9
    Prop<::unicode::EastAsianWidth> { { 0x05D0, 0x05EA }, ::unicode::EastAsianWidth::Neutral }, // Lo    [27] HEBREW LETTER ALEF..HEBREW LETTER TAV
    // 0x05EF .. 0x05F2: Lo     [4] HEBREW YOD TRIANGLE..HEBREW LIGATURE YIDDISH DOUBLE YOD
    // 0x05F3 .. 0x05F4: Po     [2] HEBREW PUNCTUATION GERESH..HEBREW PUNCTUATION GERSHAYIM
    Prop<::unicode::EastAsianWidth> { { 0x05EF, 0x05F4 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x0600 .. 0x0605: Cf     [6] ARABIC NUMBER SIGN..ARABIC NUMBER MARK ABOVE
    // 0x0606 .. 0x0608: Sm     [3] ARABIC-INDIC CUBE ROOT..ARABIC RAY
    // 0x0609 .. 0x060A: Po     [2] ARABIC-INDIC PER MILLE SIGN..ARABIC-INDIC PER TEN THOUSAND SIGN
    // 0x060B .. 0x060B: Sc         AFGHANI SIGN
    // 0x060C .. 0x060D: Po     [2] ARABIC COMMA..ARABIC DATE SEPARATOR
    // 0x060E .. 0x060F: So     [2] ARABIC POETIC VERSE SIGN..ARABIC SIGN MISRA
    // 0x0610 .. 0x061A: Mn    [11] ARABIC SIGN SALLALLAHOU ALAYHE WASSALLAM..ARABIC SMALL KASRA
    // 0x061B .. 0x061B: Po         ARABIC SEMICOLON
    // 0x061C .. 0x061C: Cf         ARABIC LETTER MARK
    // 0x061D .. 0x061F: Po     [3] ARABIC END OF TEXT MARK..ARABIC QUESTION MARK
    // 0x0620 .. 0x063F: Lo    [32] ARABIC LETTER KASHMIRI YEH..ARABIC LETTER FARSI YEH WITH THREE DOTS ABOVE
    // 0x0640 .. 0x0640: Lm         ARABIC TATWEEL
    // 0x0641 .. 0x064A: Lo    [10] ARABIC LETTER FEH..ARABIC LETTER YEH
    // 0x064B .. 0x065F: Mn    [21] ARABIC FATHATAN..ARABIC WAVY HAMZA BELOW
    // 0x0660 .. 0x0669: Nd    [10] ARABIC-INDIC DIGIT ZERO..ARABIC-INDIC DIGIT NINE
    // 0x066A .. 0x066D: Po     [4] ARABIC PERCENT SIGN..ARABIC FIVE POINTED STAR
    // 0x066E .. 0x066F: Lo     [2] ARABIC LETTER DOTLESS BEH..ARABIC LETTER DOTLESS QAF
    // 0x0670 .. 0x0670: Mn         ARABIC LETTER SUPERSCRIPT ALEF
    // 0x0671 .. 0x06D3: Lo    [99] ARABIC LETTER ALEF WASLA..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE
    // 0x06D4 .. 0x06D4: Po         ARABIC FULL STOP
    // 0x06D5 .. 0x06D5: Lo         ARABIC LETTER AE
    // 0x06D6 .. 0x06DC: Mn     [7] ARABIC SMALL HIGH LIGATURE SAD WITH LAM WITH ALEF MAKSURA..ARABIC SMALL HIGH SEEN
    // 0x06DD .. 0x06DD: Cf         ARABIC END OF AYAH
    // 0x06DE .. 0x06DE: So         ARABIC START OF RUB EL HIZB
    // 0x06DF .. 0x06E4: Mn     [6] ARABIC SMALL HIGH ROUNDED ZERO..ARABIC SMALL HIGH MADDA
    // 0x06E5 .. 0x06E6: Lm     [2] ARABIC SMALL WAW..ARABIC SMALL YEH
    // 0x06E7 .. 0x06E8: Mn     [2] ARABIC SMALL HIGH YEH..ARABIC SMALL HIGH NOON
    // 0x06E9 .. 0x06E9: So         ARABIC PLACE OF SAJDAH
    // 0x06EA .. 0x06ED: Mn     [4] ARABIC EMPTY CENTRE LOW STOP..ARABIC SMALL LOW MEEM
    // 0x06EE .. 0x06EF: Lo     [2] ARABIC LETTER DAL WITH INVERTED V..ARABIC LETTER REH WITH INVERTED V
    // 0x06F0 .. 0x06F9: Nd    [10] EXTENDED ARABIC-INDIC DIGIT ZERO..EXTENDED ARABIC-INDIC DIGIT NINE
    // 0x06FA .. 0x06FC: Lo     [3] ARABIC LETTER SHEEN WITH DOT BELOW..ARABIC LETTER GHAIN WITH DOT BELOW
    // 0x06FD .. 0x06FE: So     [2] ARABIC SIGN SINDHI AMPERSAND..ARABIC SIGN SINDHI POSTPOSITION MEN
    // 0x06FF .. 0x06FF: Lo         ARABIC LETTER HEH WITH INVERTED V
    // 0x0700 .. 0x070D: Po    [14] SYRIAC END OF PARAGRAPH..SYRIAC HARKLEAN ASTERISCUS
    Prop<::unicode::EastAsianWidth> { { 0x0600, 0x070D }, ::unicode::EastAsianWidth::Neutral }, // #35
    // 0x070F .. 0x070F: Cf         SYRIAC ABBREVIATION MARK
    // 0x0710 .. 0x0710: Lo         SYRIAC LETTER ALAPH
    // 0x0711 .. 0x0711: Mn         SYRIAC LETTER SUPERSCRIPT ALAPH
    // 0x0712 .. 0x072F: Lo    [30] SYRIAC LETTER BETH..SYRIAC LETTER PERSIAN DHALATH
    // 0x0730 .. 0x074A: Mn    [27] SYRIAC PTHAHA ABOVE..SYRIAC BARREKH
    Prop<::unicode::EastAsianWidth> { { 0x070F, 0x074A }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x074D .. 0x074F: Lo     [3] SYRIAC LETTER SOGDIAN ZHAIN..SYRIAC LETTER SOGDIAN FE
    // 0x0750 .. 0x077F: Lo    [48] ARABIC LETTER BEH WITH THREE DOTS HORIZONTALLY BELOW..ARABIC LETTER KAF WITH TWO DOTS ABOVE
    // 0x0780 .. 0x07A5: Lo    [38] THAANA LETTER HAA..THAANA LETTER WAAVU
    // 0x07A6 .. 0x07B0: Mn    [11] THAANA ABAFILI..THAANA SUKUN
    // 0x07B1 .. 0x07B1: Lo         THAANA LETTER NAA
    Prop<::unicode::EastAsianWidth> { { 0x074D, 0x07B1 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x07C0 .. 0x07C9: Nd    [10] NKO DIGIT ZERO..NKO DIGIT NINE
    // 0x07CA .. 0x07EA: Lo    [33] NKO LETTER A..NKO LETTER JONA RA
    // 0x07EB .. 0x07F3: Mn     [9] NKO COMBINING SHORT HIGH TONE..NKO COMBINING DOUBLE DOT ABOVE
    // 0x07F4 .. 0x07F5: Lm     [2] NKO HIGH TONE APOSTROPHE..NKO LOW TONE APOSTROPHE
    // 0x07F6 .. 0x07F6: So         NKO SYMBOL OO DENNEN
    // 0x07F7 .. 0x07F9: Po     [3] NKO SYMBOL GBAKURUNEN..NKO EXCLAMATION MARK
    // 0x07FA .. 0x07FA: Lm         NKO LAJANYALAN
    Prop<::unicode::EastAsianWidth> { { 0x07C0, 0x07FA }, ::unicode::EastAsianWidth::Neutral }, // #7
    // 0x07FD .. 0x07FD: Mn         NKO DANTAYALAN
    // 0x07FE .. 0x07FF: Sc     [2] NKO DOROME SIGN..NKO TAMAN SIGN
    // 0x0800 .. 0x0815: Lo    [22] SAMARITAN LETTER ALAF..SAMARITAN LETTER TAAF
    // 0x0816 .. 0x0819: Mn     [4] SAMARITAN MARK IN..SAMARITAN MARK DAGESH
    // 0x081A .. 0x081A: Lm         SAMARITAN MODIFIER LETTER EPENTHETIC YUT
    // 0x081B .. 0x0823: Mn     [9] SAMARITAN MARK EPENTHETIC YUT..SAMARITAN VOWEL SIGN A
    // 0x0824 .. 0x0824: Lm         SAMARITAN MODIFIER LETTER SHORT A
    // 0x0825 .. 0x0827: Mn     [3] SAMARITAN VOWEL SIGN SHORT A..SAMARITAN VOWEL SIGN U
    // 0x0828 .. 0x0828: Lm         SAMARITAN MODIFIER LETTER I
    // 0x0829 .. 0x082D: Mn     [5] SAMARITAN VOWEL SIGN LONG I..SAMARITAN MARK NEQUDAA
    Prop<::unicode::EastAsianWidth> { { 0x07FD, 0x082D }, ::unicode::EastAsianWidth::Neutral }, // #10
    Prop<::unicode::EastAsianWidth> { { 0x0830, 0x083E }, ::unicode::EastAsianWidth::Neutral }, // Po    [15] SAMARITAN PUNCTUATION NEQUDAA..SAMARITAN PUNCTUATION ANNAAU
    // 0x0840 .. 0x0858: Lo    [25] MANDAIC LETTER HALQA..MANDAIC LETTER AIN
    // 0x0859 .. 0x085B: Mn     [3] MANDAIC AFFRICATION MARK..MANDAIC GEMINATION MARK
    Prop<::unicode::EastAsianWidth> { { 0x0840, 0x085B }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x085E, 0x085E }, ::unicode::EastAsianWidth::Neutral }, // Po         MANDAIC PUNCTUATION
    Prop<::unicode::EastAsianWidth> { { 0x0860, 0x086A }, ::unicode::EastAsianWidth::Neutral }, // Lo    [11] SYRIAC LETTER MALAYALAM NGA..SYRIAC LETTER MALAYALAM SSA
    // 0x0870 .. 0x0887: Lo    [24] ARABIC LETTER ALEF WITH ATTACHED FATHA..ARABIC BASELINE ROUND DOT
    // 0x0888 .. 0x0888: Sk         ARABIC RAISED ROUND DOT
    // 0x0889 .. 0x088F: Lo     [7] ARABIC LETTER NOON WITH INVERTED SMALL V..ARABIC LETTER NOON WITH RING ABOVE
    // 0x0890 .. 0x0891: Cf     [2] ARABIC POUND MARK ABOVE..ARABIC PIASTRE MARK ABOVE
    Prop<::unicode::EastAsianWidth> { { 0x0870, 0x0891 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x0897 .. 0x089F: Mn     [9] ARABIC PEPET..ARABIC HALF MADDA OVER MADDA
    // 0x08A0 .. 0x08C8: Lo    [41] ARABIC LETTER BEH WITH SMALL V BELOW..ARABIC LETTER GRAF
    // 0x08C9 .. 0x08C9: Lm         ARABIC SMALL FARSI YEH
    // 0x08CA .. 0x08E1: Mn    [24] ARABIC SMALL HIGH FARSI YEH..ARABIC SMALL HIGH SIGN SAFHA
    // 0x08E2 .. 0x08E2: Cf         ARABIC DISPUTED END OF AYAH
    // 0x08E3 .. 0x08FF: Mn    [29] ARABIC TURNED DAMMA BELOW..ARABIC MARK SIDEWAYS NOON GHUNNA
    // 0x0900 .. 0x0902: Mn     [3] DEVANAGARI SIGN INVERTED CANDRABINDU..DEVANAGARI SIGN ANUSVARA
    // 0x0903 .. 0x0903: Mc         DEVANAGARI SIGN VISARGA
    // 0x0904 .. 0x0939: Lo    [54] DEVANAGARI LETTER SHORT A..DEVANAGARI LETTER HA
    // 0x093A .. 0x093A: Mn         DEVANAGARI VOWEL SIGN OE
    // 0x093B .. 0x093B: Mc         DEVANAGARI VOWEL SIGN OOE
    // 0x093C .. 0x093C: Mn         DEVANAGARI SIGN NUKTA
    // 0x093D .. 0x093D: Lo         DEVANAGARI SIGN AVAGRAHA
    // 0x093E .. 0x0940: Mc     [3] DEVANAGARI VOWEL SIGN AA..DEVANAGARI VOWEL SIGN II
    // 0x0941 .. 0x0948: Mn     [8] DEVANAGARI VOWEL SIGN U..DEVANAGARI VOWEL SIGN AI
    // 0x0949 .. 0x094C: Mc     [4] DEVANAGARI VOWEL SIGN CANDRA O..DEVANAGARI VOWEL SIGN AU
    // 0x094D .. 0x094D: Mn         DEVANAGARI SIGN VIRAMA
    // 0x094E .. 0x094F: Mc     [2] DEVANAGARI VOWEL SIGN PRISHTHAMATRA E..DEVANAGARI VOWEL SIGN AW
    // 0x0950 .. 0x0950: Lo         DEVANAGARI OM
    // 0x0951 .. 0x0957: Mn     [7] DEVANAGARI STRESS SIGN UDATTA..DEVANAGARI VOWEL SIGN UUE
    // 0x0958 .. 0x0961: Lo    [10] DEVANAGARI LETTER QA..DEVANAGARI LETTER VOCALIC LL
    // 0x0962 .. 0x0963: Mn     [2] DEVANAGARI VOWEL SIGN VOCALIC L..DEVANAGARI VOWEL SIGN VOCALIC LL
    // 0x0964 .. 0x0965: Po     [2] DEVANAGARI DANDA..DEVANAGARI DOUBLE DANDA
    // 0x0966 .. 0x096F: Nd    [10] DEVANAGARI DIGIT ZERO..DEVANAGARI DIGIT NINE
    // 0x0970 .. 0x0970: Po         DEVANAGARI ABBREVIATION SIGN
    // 0x0971 .. 0x0971: Lm         DEVANAGARI SIGN HIGH SPACING DOT
    // 0x0972 .. 0x097F: Lo    [14] DEVANAGARI LETTER CANDRA A..DEVANAGARI LETTER BBA
    // 0x0980 .. 0x0980: Lo         BENGALI ANJI
    // 0x0981 .. 0x0981: Mn         BENGALI SIGN CANDRABINDU
    // 0x0982 .. 0x0983: Mc     [2] BENGALI SIGN ANUSVARA..BENGALI SIGN VISARGA
    Prop<::unicode::EastAsianWidth> { { 0x0897, 0x0983 }, ::unicode::EastAsianWidth::Neutral }, // #30
    Prop<::unicode::EastAsianWidth> { { 0x0985, 0x098C }, ::unicode::EastAsianWidth::Neutral }, // Lo     [8] BENGALI LETTER A..BENGALI LETTER VOCALIC L
    Prop<::unicode::EastAsianWidth> { { 0x098F, 0x0990 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] BENGALI LETTER E..BENGALI LETTER AI
    Prop<::unicode::EastAsianWidth> { { 0x0993, 0x09A8 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [22] BENGALI LETTER O..BENGALI LETTER NA
    Prop<::unicode::EastAsianWidth> { { 0x09AA, 0x09B0 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] BENGALI LETTER PA..BENGALI LETTER RA
    Prop<::unicode::EastAsianWidth> { { 0x09B2, 0x09B2 }, ::unicode::EastAsianWidth::Neutral }, // Lo         BENGALI LETTER LA
    Prop<::unicode::EastAsianWidth> { { 0x09B6, 0x09B9 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] BENGALI LETTER SHA..BENGALI LETTER HA
    // 0x09BC .. 0x09BC: Mn         BENGALI SIGN NUKTA
    // 0x09BD .. 0x09BD: Lo         BENGALI SIGN AVAGRAHA
    // 0x09BE .. 0x09C0: Mc     [3] BENGALI VOWEL SIGN AA..BENGALI VOWEL SIGN II
    // 0x09C1 .. 0x09C4: Mn     [4] BENGALI VOWEL SIGN U..BENGALI VOWEL SIGN VOCALIC RR
    Prop<::unicode::EastAsianWidth> { { 0x09BC, 0x09C4 }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x09C7, 0x09C8 }, ::unicode::EastAsianWidth::Neutral }, // Mc     [2] BENGALI VOWEL SIGN E..BENGALI VOWEL SIGN AI
    // 0x09CB .. 0x09CC: Mc     [2] BENGALI VOWEL SIGN O..BENGALI VOWEL SIGN AU
    // 0x09CD .. 0x09CD: Mn         BENGALI SIGN VIRAMA
    // 0x09CE .. 0x09CE: Lo         BENGALI LETTER KHANDA TA
    Prop<::unicode::EastAsianWidth> { { 0x09CB, 0x09CE }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x09D7, 0x09D7 }, ::unicode::EastAsianWidth::Neutral }, // Mc         BENGALI AU LENGTH MARK
    Prop<::unicode::EastAsianWidth> { { 0x09DC, 0x09DD }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] BENGALI LETTER RRA..BENGALI LETTER RHA
    // 0x09DF .. 0x09E1: Lo     [3] BENGALI LETTER YYA..BENGALI LETTER VOCALIC LL
    // 0x09E2 .. 0x09E3: Mn     [2] BENGALI VOWEL SIGN VOCALIC L..BENGALI VOWEL SIGN VOCALIC LL
    Prop<::unicode::EastAsianWidth> { { 0x09DF, 0x09E3 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x09E6 .. 0x09EF: Nd    [10] BENGALI DIGIT ZERO..BENGALI DIGIT NINE
    // 0x09F0 .. 0x09F1: Lo     [2] BENGALI LETTER RA WITH MIDDLE DIAGONAL..BENGALI LETTER RA WITH LOWER DIAGONAL
    // 0x09F2 .. 0x09F3: Sc     [2] BENGALI RUPEE MARK..BENGALI RUPEE SIGN
    // 0x09F4 .. 0x09F9: No     [6] BENGALI CURRENCY NUMERATOR ONE..BENGALI CURRENCY DENOMINATOR SIXTEEN
    // 0x09FA .. 0x09FA: So         BENGALI ISSHAR
    // 0x09FB .. 0x09FB: Sc         BENGALI GANDA MARK
    // 0x09FC .. 0x09FC: Lo         BENGALI LETTER VEDIC ANUSVARA
    // 0x09FD .. 0x09FD: Po         BENGALI ABBREVIATION SIGN
    // 0x09FE .. 0x09FE: Mn         BENGALI SANDHI MARK
    Prop<::unicode::EastAsianWidth> { { 0x09E6, 0x09FE }, ::unicode::EastAsianWidth::Neutral }, // #9
    // 0x0A01 .. 0x0A02: Mn     [2] GURMUKHI SIGN ADAK BINDI..GURMUKHI SIGN BINDI
    // 0x0A03 .. 0x0A03: Mc         GURMUKHI SIGN VISARGA
    Prop<::unicode::EastAsianWidth> { { 0x0A01, 0x0A03 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0A05, 0x0A0A }, ::unicode::EastAsianWidth::Neutral }, // Lo     [6] GURMUKHI LETTER A..GURMUKHI LETTER UU
    Prop<::unicode::EastAsianWidth> { { 0x0A0F, 0x0A10 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] GURMUKHI LETTER EE..GURMUKHI LETTER AI
    Prop<::unicode::EastAsianWidth> { { 0x0A13, 0x0A28 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [22] GURMUKHI LETTER OO..GURMUKHI LETTER NA
    Prop<::unicode::EastAsianWidth> { { 0x0A2A, 0x0A30 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] GURMUKHI LETTER PA..GURMUKHI LETTER RA
    Prop<::unicode::EastAsianWidth> { { 0x0A32, 0x0A33 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] GURMUKHI LETTER LA..GURMUKHI LETTER LLA
    Prop<::unicode::EastAsianWidth> { { 0x0A35, 0x0A36 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] GURMUKHI LETTER VA..GURMUKHI LETTER SHA
    Prop<::unicode::EastAsianWidth> { { 0x0A38, 0x0A39 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] GURMUKHI LETTER SA..GURMUKHI LETTER HA
    Prop<::unicode::EastAsianWidth> { { 0x0A3C, 0x0A3C }, ::unicode::EastAsianWidth::Neutral }, // Mn         GURMUKHI SIGN NUKTA
    // 0x0A3E .. 0x0A40: Mc     [3] GURMUKHI VOWEL SIGN AA..GURMUKHI VOWEL SIGN II
    // 0x0A41 .. 0x0A42: Mn     [2] GURMUKHI VOWEL SIGN U..GURMUKHI VOWEL SIGN UU
    Prop<::unicode::EastAsianWidth> { { 0x0A3E, 0x0A42 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0A47, 0x0A48 }, ::unicode::EastAsianWidth::Neutral }, // Mn     [2] GURMUKHI VOWEL SIGN EE..GURMUKHI VOWEL SIGN AI
    Prop<::unicode::EastAsianWidth> { { 0x0A4B, 0x0A4D }, ::unicode::EastAsianWidth::Neutral }, // Mn     [3] GURMUKHI VOWEL SIGN OO..GURMUKHI SIGN VIRAMA
    Prop<::unicode::EastAsianWidth> { { 0x0A51, 0x0A51 }, ::unicode::EastAsianWidth::Neutral }, // Mn         GURMUKHI SIGN UDAAT
    Prop<::unicode::EastAsianWidth> { { 0x0A59, 0x0A5C }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] GURMUKHI LETTER KHHA..GURMUKHI LETTER RRA
    Prop<::unicode::EastAsianWidth> { { 0x0A5E, 0x0A5E }, ::unicode::EastAsianWidth::Neutral }, // Lo         GURMUKHI LETTER FA
    // 0x0A66 .. 0x0A6F: Nd    [10] GURMUKHI DIGIT ZERO..GURMUKHI DIGIT NINE
    // 0x0A70 .. 0x0A71: Mn     [2] GURMUKHI TIPPI..GURMUKHI ADDAK
    // 0x0A72 .. 0x0A74: Lo     [3] GURMUKHI IRI..GURMUKHI EK ONKAR
    // 0x0A75 .. 0x0A75: Mn         GURMUKHI SIGN YAKASH
    // 0x0A76 .. 0x0A76: Po         GURMUKHI ABBREVIATION SIGN
    Prop<::unicode::EastAsianWidth> { { 0x0A66, 0x0A76 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x0A81 .. 0x0A82: Mn     [2] GUJARATI SIGN CANDRABINDU..GUJARATI SIGN ANUSVARA
    // 0x0A83 .. 0x0A83: Mc         GUJARATI SIGN VISARGA
    Prop<::unicode::EastAsianWidth> { { 0x0A81, 0x0A83 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0A85, 0x0A8D }, ::unicode::EastAsianWidth::Neutral }, // Lo     [9] GUJARATI LETTER A..GUJARATI VOWEL CANDRA E
    Prop<::unicode::EastAsianWidth> { { 0x0A8F, 0x0A91 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [3] GUJARATI LETTER E..GUJARATI VOWEL CANDRA O
    Prop<::unicode::EastAsianWidth> { { 0x0A93, 0x0AA8 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [22] GUJARATI LETTER O..GUJARATI LETTER NA
    Prop<::unicode::EastAsianWidth> { { 0x0AAA, 0x0AB0 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] GUJARATI LETTER PA..GUJARATI LETTER RA
    Prop<::unicode::EastAsianWidth> { { 0x0AB2, 0x0AB3 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] GUJARATI LETTER LA..GUJARATI LETTER LLA
    Prop<::unicode::EastAsianWidth> { { 0x0AB5, 0x0AB9 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [5] GUJARATI LETTER VA..GUJARATI LETTER HA
    // 0x0ABC .. 0x0ABC: Mn         GUJARATI SIGN NUKTA
    // 0x0ABD .. 0x0ABD: Lo         GUJARATI SIGN AVAGRAHA
    // 0x0ABE .. 0x0AC0: Mc     [3] GUJARATI VOWEL SIGN AA..GUJARATI VOWEL SIGN II
    // 0x0AC1 .. 0x0AC5: Mn     [5] GUJARATI VOWEL SIGN U..GUJARATI VOWEL SIGN CANDRA E
    Prop<::unicode::EastAsianWidth> { { 0x0ABC, 0x0AC5 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x0AC7 .. 0x0AC8: Mn     [2] GUJARATI VOWEL SIGN E..GUJARATI VOWEL SIGN AI
    // 0x0AC9 .. 0x0AC9: Mc         GUJARATI VOWEL SIGN CANDRA O
    Prop<::unicode::EastAsianWidth> { { 0x0AC7, 0x0AC9 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x0ACB .. 0x0ACC: Mc     [2] GUJARATI VOWEL SIGN O..GUJARATI VOWEL SIGN AU
    // 0x0ACD .. 0x0ACD: Mn         GUJARATI SIGN VIRAMA
    Prop<::unicode::EastAsianWidth> { { 0x0ACB, 0x0ACD }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0AD0, 0x0AD0 }, ::unicode::EastAsianWidth::Neutral }, // Lo         GUJARATI OM
    // 0x0AE0 .. 0x0AE1: Lo     [2] GUJARATI LETTER VOCALIC RR..GUJARATI LETTER VOCALIC LL
    // 0x0AE2 .. 0x0AE3: Mn     [2] GUJARATI VOWEL SIGN VOCALIC L..GUJARATI VOWEL SIGN VOCALIC LL
    Prop<::unicode::EastAsianWidth> { { 0x0AE0, 0x0AE3 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x0AE6 .. 0x0AEF: Nd    [10] GUJARATI DIGIT ZERO..GUJARATI DIGIT NINE
    // 0x0AF0 .. 0x0AF0: Po         GUJARATI ABBREVIATION SIGN
    // 0x0AF1 .. 0x0AF1: Sc         GUJARATI RUPEE SIGN
    Prop<::unicode::EastAsianWidth> { { 0x0AE6, 0x0AF1 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x0AF9 .. 0x0AF9: Lo         GUJARATI LETTER ZHA
    // 0x0AFA .. 0x0AFF: Mn     [6] GUJARATI SIGN SUKUN..GUJARATI SIGN TWO-CIRCLE NUKTA ABOVE
    Prop<::unicode::EastAsianWidth> { { 0x0AF9, 0x0AFF }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x0B01 .. 0x0B01: Mn         ORIYA SIGN CANDRABINDU
    // 0x0B02 .. 0x0B03: Mc     [2] ORIYA SIGN ANUSVARA..ORIYA SIGN VISARGA
    Prop<::unicode::EastAsianWidth> { { 0x0B01, 0x0B03 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0B05, 0x0B0C }, ::unicode::EastAsianWidth::Neutral }, // Lo     [8] ORIYA LETTER A..ORIYA LETTER VOCALIC L
    Prop<::unicode::EastAsianWidth> { { 0x0B0F, 0x0B10 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] ORIYA LETTER E..ORIYA LETTER AI
    Prop<::unicode::EastAsianWidth> { { 0x0B13, 0x0B28 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [22] ORIYA LETTER O..ORIYA LETTER NA
    Prop<::unicode::EastAsianWidth> { { 0x0B2A, 0x0B30 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ORIYA LETTER PA..ORIYA LETTER RA
    Prop<::unicode::EastAsianWidth> { { 0x0B32, 0x0B33 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] ORIYA LETTER LA..ORIYA LETTER LLA
    Prop<::unicode::EastAsianWidth> { { 0x0B35, 0x0B39 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [5] ORIYA LETTER VA..ORIYA LETTER HA
    // 0x0B3C .. 0x0B3C: Mn         ORIYA SIGN NUKTA
    // 0x0B3D .. 0x0B3D: Lo         ORIYA SIGN AVAGRAHA
    // 0x0B3E .. 0x0B3E: Mc         ORIYA VOWEL SIGN AA
    // 0x0B3F .. 0x0B3F: Mn         ORIYA VOWEL SIGN I
    // 0x0B40 .. 0x0B40: Mc         ORIYA VOWEL SIGN II
    // 0x0B41 .. 0x0B44: Mn     [4] ORIYA VOWEL SIGN U..ORIYA VOWEL SIGN VOCALIC RR
    Prop<::unicode::EastAsianWidth> { { 0x0B3C, 0x0B44 }, ::unicode::EastAsianWidth::Neutral }, // #6
    Prop<::unicode::EastAsianWidth> { { 0x0B47, 0x0B48 }, ::unicode::EastAsianWidth::Neutral }, // Mc     [2] ORIYA VOWEL SIGN E..ORIYA VOWEL SIGN AI
    // 0x0B4B .. 0x0B4C: Mc     [2] ORIYA VOWEL SIGN O..ORIYA VOWEL SIGN AU
    // 0x0B4D .. 0x0B4D: Mn         ORIYA SIGN VIRAMA
    Prop<::unicode::EastAsianWidth> { { 0x0B4B, 0x0B4D }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x0B55 .. 0x0B56: Mn     [2] ORIYA SIGN OVERLINE..ORIYA AI LENGTH MARK
    // 0x0B57 .. 0x0B57: Mc         ORIYA AU LENGTH MARK
    Prop<::unicode::EastAsianWidth> { { 0x0B55, 0x0B57 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0B5C, 0x0B5D }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] ORIYA LETTER RRA..ORIYA LETTER RHA
    // 0x0B5F .. 0x0B61: Lo     [3] ORIYA LETTER YYA..ORIYA LETTER VOCALIC LL
    // 0x0B62 .. 0x0B63: Mn     [2] ORIYA VOWEL SIGN VOCALIC L..ORIYA VOWEL SIGN VOCALIC LL
    Prop<::unicode::EastAsianWidth> { { 0x0B5F, 0x0B63 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x0B66 .. 0x0B6F: Nd    [10] ORIYA DIGIT ZERO..ORIYA DIGIT NINE
    // 0x0B70 .. 0x0B70: So         ORIYA ISSHAR
    // 0x0B71 .. 0x0B71: Lo         ORIYA LETTER WA
    // 0x0B72 .. 0x0B77: No     [6] ORIYA FRACTION ONE QUARTER..ORIYA FRACTION THREE SIXTEENTHS
    Prop<::unicode::EastAsianWidth> { { 0x0B66, 0x0B77 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x0B82 .. 0x0B82: Mn         TAMIL SIGN ANUSVARA
    // 0x0B83 .. 0x0B83: Lo         TAMIL SIGN VISARGA
    Prop<::unicode::EastAsianWidth> { { 0x0B82, 0x0B83 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0B85, 0x0B8A }, ::unicode::EastAsianWidth::Neutral }, // Lo     [6] TAMIL LETTER A..TAMIL LETTER UU
    Prop<::unicode::EastAsianWidth> { { 0x0B8E, 0x0B90 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [3] TAMIL LETTER E..TAMIL LETTER AI
    Prop<::unicode::EastAsianWidth> { { 0x0B92, 0x0B95 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] TAMIL LETTER O..TAMIL LETTER KA
    Prop<::unicode::EastAsianWidth> { { 0x0B99, 0x0B9A }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] TAMIL LETTER NGA..TAMIL LETTER CA
    Prop<::unicode::EastAsianWidth> { { 0x0B9C, 0x0B9C }, ::unicode::EastAsianWidth::Neutral }, // Lo         TAMIL LETTER JA
    Prop<::unicode::EastAsianWidth> { { 0x0B9E, 0x0B9F }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] TAMIL LETTER NYA..TAMIL LETTER TTA
    Prop<::unicode::EastAsianWidth> { { 0x0BA3, 0x0BA4 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] TAMIL LETTER NNA..TAMIL LETTER TA
    Prop<::unicode::EastAsianWidth> { { 0x0BA8, 0x0BAA }, ::unicode::EastAsianWidth::Neutral }, // Lo     [3] TAMIL LETTER NA..TAMIL LETTER PA
    Prop<::unicode::EastAsianWidth> { { 0x0BAE, 0x0BB9 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [12] TAMIL LETTER MA..TAMIL LETTER HA
    // 0x0BBE .. 0x0BBF: Mc     [2] TAMIL VOWEL SIGN AA..TAMIL VOWEL SIGN I
    // 0x0BC0 .. 0x0BC0: Mn         TAMIL VOWEL SIGN II
    // 0x0BC1 .. 0x0BC2: Mc     [2] TAMIL VOWEL SIGN U..TAMIL VOWEL SIGN UU
    Prop<::unicode::EastAsianWidth> { { 0x0BBE, 0x0BC2 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x0BC6, 0x0BC8 }, ::unicode::EastAsianWidth::Neutral }, // Mc     [3] TAMIL VOWEL SIGN E..TAMIL VOWEL SIGN AI
    // 0x0BCA .. 0x0BCC: Mc     [3] TAMIL VOWEL SIGN O..TAMIL VOWEL SIGN AU
    // 0x0BCD .. 0x0BCD: Mn         TAMIL SIGN VIRAMA
    Prop<::unicode::EastAsianWidth> { { 0x0BCA, 0x0BCD }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0BD0, 0x0BD0 }, ::unicode::EastAsianWidth::Neutral }, // Lo         TAMIL OM
    Prop<::unicode::EastAsianWidth> { { 0x0BD7, 0x0BD7 }, ::unicode::EastAsianWidth::Neutral }, // Mc         TAMIL AU LENGTH MARK
    // 0x0BE6 .. 0x0BEF: Nd    [10] TAMIL DIGIT ZERO..TAMIL DIGIT NINE
    // 0x0BF0 .. 0x0BF2: No     [3] TAMIL NUMBER TEN..TAMIL NUMBER ONE THOUSAND
    // 0x0BF3 .. 0x0BF8: So     [6] TAMIL DAY SIGN..TAMIL AS ABOVE SIGN
    // 0x0BF9 .. 0x0BF9: Sc         TAMIL RUPEE SIGN
    // 0x0BFA .. 0x0BFA: So         TAMIL NUMBER SIGN
    Prop<::unicode::EastAsianWidth> { { 0x0BE6, 0x0BFA }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x0C00 .. 0x0C00: Mn         TELUGU SIGN COMBINING CANDRABINDU ABOVE
    // 0x0C01 .. 0x0C03: Mc     [3] TELUGU SIGN CANDRABINDU..TELUGU SIGN VISARGA
    // 0x0C04 .. 0x0C04: Mn         TELUGU SIGN COMBINING ANUSVARA ABOVE
    // 0x0C05 .. 0x0C0C: Lo     [8] TELUGU LETTER A..TELUGU LETTER VOCALIC L
    Prop<::unicode::EastAsianWidth> { { 0x0C00, 0x0C0C }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x0C0E, 0x0C10 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [3] TELUGU LETTER E..TELUGU LETTER AI
    Prop<::unicode::EastAsianWidth> { { 0x0C12, 0x0C28 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [23] TELUGU LETTER O..TELUGU LETTER NA
    Prop<::unicode::EastAsianWidth> { { 0x0C2A, 0x0C39 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [16] TELUGU LETTER PA..TELUGU LETTER HA
    // 0x0C3C .. 0x0C3C: Mn         TELUGU SIGN NUKTA
    // 0x0C3D .. 0x0C3D: Lo         TELUGU SIGN AVAGRAHA
    // 0x0C3E .. 0x0C40: Mn     [3] TELUGU VOWEL SIGN AA..TELUGU VOWEL SIGN II
    // 0x0C41 .. 0x0C44: Mc     [4] TELUGU VOWEL SIGN U..TELUGU VOWEL SIGN VOCALIC RR
    Prop<::unicode::EastAsianWidth> { { 0x0C3C, 0x0C44 }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x0C46, 0x0C48 }, ::unicode::EastAsianWidth::Neutral }, // Mn     [3] TELUGU VOWEL SIGN E..TELUGU VOWEL SIGN AI
    Prop<::unicode::EastAsianWidth> { { 0x0C4A, 0x0C4D }, ::unicode::EastAsianWidth::Neutral }, // Mn     [4] TELUGU VOWEL SIGN O..TELUGU SIGN VIRAMA
    Prop<::unicode::EastAsianWidth> { { 0x0C55, 0x0C56 }, ::unicode::EastAsianWidth::Neutral }, // Mn     [2] TELUGU LENGTH MARK..TELUGU AI LENGTH MARK
    Prop<::unicode::EastAsianWidth> { { 0x0C58, 0x0C5A }, ::unicode::EastAsianWidth::Neutral }, // Lo     [3] TELUGU LETTER TSA..TELUGU LETTER RRRA
    Prop<::unicode::EastAsianWidth> { { 0x0C5C, 0x0C5D }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] TELUGU ARCHAIC SHRII..TELUGU LETTER NAKAARA POLLU
    // 0x0C60 .. 0x0C61: Lo     [2] TELUGU LETTER VOCALIC RR..TELUGU LETTER VOCALIC LL
    // 0x0C62 .. 0x0C63: Mn     [2] TELUGU VOWEL SIGN VOCALIC L..TELUGU VOWEL SIGN VOCALIC LL
    Prop<::unicode::EastAsianWidth> { { 0x0C60, 0x0C63 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0C66, 0x0C6F }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] TELUGU DIGIT ZERO..TELUGU DIGIT NINE
    // 0x0C77 .. 0x0C77: Po         TELUGU SIGN SIDDHAM
    // 0x0C78 .. 0x0C7E: No     [7] TELUGU FRACTION DIGIT ZERO FOR ODD POWERS OF FOUR..TELUGU FRACTION DIGIT THREE FOR EVEN POWERS OF FOUR
    // 0x0C7F .. 0x0C7F: So         TELUGU SIGN TUUMU
    // 0x0C80 .. 0x0C80: Lo         KANNADA SIGN SPACING CANDRABINDU
    // 0x0C81 .. 0x0C81: Mn         KANNADA SIGN CANDRABINDU
    // 0x0C82 .. 0x0C83: Mc     [2] KANNADA SIGN ANUSVARA..KANNADA SIGN VISARGA
    // 0x0C84 .. 0x0C84: Po         KANNADA SIGN SIDDHAM
    // 0x0C85 .. 0x0C8C: Lo     [8] KANNADA LETTER A..KANNADA LETTER VOCALIC L
    Prop<::unicode::EastAsianWidth> { { 0x0C77, 0x0C8C }, ::unicode::EastAsianWidth::Neutral }, // #8
    Prop<::unicode::EastAsianWidth> { { 0x0C8E, 0x0C90 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [3] KANNADA LETTER E..KANNADA LETTER AI
    Prop<::unicode::EastAsianWidth> { { 0x0C92, 0x0CA8 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [23] KANNADA LETTER O..KANNADA LETTER NA
    Prop<::unicode::EastAsianWidth> { { 0x0CAA, 0x0CB3 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [10] KANNADA LETTER PA..KANNADA LETTER LLA
    Prop<::unicode::EastAsianWidth> { { 0x0CB5, 0x0CB9 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [5] KANNADA LETTER VA..KANNADA LETTER HA
    // 0x0CBC .. 0x0CBC: Mn         KANNADA SIGN NUKTA
    // 0x0CBD .. 0x0CBD: Lo         KANNADA SIGN AVAGRAHA
    // 0x0CBE .. 0x0CBE: Mc         KANNADA VOWEL SIGN AA
    // 0x0CBF .. 0x0CBF: Mn         KANNADA VOWEL SIGN I
    // 0x0CC0 .. 0x0CC4: Mc     [5] KANNADA VOWEL SIGN II..KANNADA VOWEL SIGN VOCALIC RR
    Prop<::unicode::EastAsianWidth> { { 0x0CBC, 0x0CC4 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x0CC6 .. 0x0CC6: Mn         KANNADA VOWEL SIGN E
    // 0x0CC7 .. 0x0CC8: Mc     [2] KANNADA VOWEL SIGN EE..KANNADA VOWEL SIGN AI
    Prop<::unicode::EastAsianWidth> { { 0x0CC6, 0x0CC8 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x0CCA .. 0x0CCB: Mc     [2] KANNADA VOWEL SIGN O..KANNADA VOWEL SIGN OO
    // 0x0CCC .. 0x0CCD: Mn     [2] KANNADA VOWEL SIGN AU..KANNADA SIGN VIRAMA
    Prop<::unicode::EastAsianWidth> { { 0x0CCA, 0x0CCD }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0CD5, 0x0CD6 }, ::unicode::EastAsianWidth::Neutral }, // Mc     [2] KANNADA LENGTH MARK..KANNADA AI LENGTH MARK
    Prop<::unicode::EastAsianWidth> { { 0x0CDC, 0x0CDE }, ::unicode::EastAsianWidth::Neutral }, // Lo     [3] KANNADA ARCHAIC SHRII..KANNADA LETTER FA
    // 0x0CE0 .. 0x0CE1: Lo     [2] KANNADA LETTER VOCALIC RR..KANNADA LETTER VOCALIC LL
    // 0x0CE2 .. 0x0CE3: Mn     [2] KANNADA VOWEL SIGN VOCALIC L..KANNADA VOWEL SIGN VOCALIC LL
    Prop<::unicode::EastAsianWidth> { { 0x0CE0, 0x0CE3 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0CE6, 0x0CEF }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] KANNADA DIGIT ZERO..KANNADA DIGIT NINE
    // 0x0CF1 .. 0x0CF2: Lo     [2] KANNADA SIGN JIHVAMULIYA..KANNADA SIGN UPADHMANIYA
    // 0x0CF3 .. 0x0CF3: Mc         KANNADA SIGN COMBINING ANUSVARA ABOVE RIGHT
    Prop<::unicode::EastAsianWidth> { { 0x0CF1, 0x0CF3 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x0D00 .. 0x0D01: Mn     [2] MALAYALAM SIGN COMBINING ANUSVARA ABOVE..MALAYALAM SIGN CANDRABINDU
    // 0x0D02 .. 0x0D03: Mc     [2] MALAYALAM SIGN ANUSVARA..MALAYALAM SIGN VISARGA
    // 0x0D04 .. 0x0D0C: Lo     [9] MALAYALAM LETTER VEDIC ANUSVARA..MALAYALAM LETTER VOCALIC L
    Prop<::unicode::EastAsianWidth> { { 0x0D00, 0x0D0C }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x0D0E, 0x0D10 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [3] MALAYALAM LETTER E..MALAYALAM LETTER AI
    // 0x0D12 .. 0x0D3A: Lo    [41] MALAYALAM LETTER O..MALAYALAM LETTER TTTA
    // 0x0D3B .. 0x0D3C: Mn     [2] MALAYALAM SIGN VERTICAL BAR VIRAMA..MALAYALAM SIGN CIRCULAR VIRAMA
    // 0x0D3D .. 0x0D3D: Lo         MALAYALAM SIGN AVAGRAHA
    // 0x0D3E .. 0x0D40: Mc     [3] MALAYALAM VOWEL SIGN AA..MALAYALAM VOWEL SIGN II
    // 0x0D41 .. 0x0D44: Mn     [4] MALAYALAM VOWEL SIGN U..MALAYALAM VOWEL SIGN VOCALIC RR
    Prop<::unicode::EastAsianWidth> { { 0x0D12, 0x0D44 }, ::unicode::EastAsianWidth::Neutral }, // #5
    Prop<::unicode::EastAsianWidth> { { 0x0D46, 0x0D48 }, ::unicode::EastAsianWidth::Neutral }, // Mc     [3] MALAYALAM VOWEL SIGN E..MALAYALAM VOWEL SIGN AI
    // 0x0D4A .. 0x0D4C: Mc     [3] MALAYALAM VOWEL SIGN O..MALAYALAM VOWEL SIGN AU
    // 0x0D4D .. 0x0D4D: Mn         MALAYALAM SIGN VIRAMA
    // 0x0D4E .. 0x0D4E: Lo         MALAYALAM LETTER DOT REPH
    // 0x0D4F .. 0x0D4F: So         MALAYALAM SIGN PARA
    Prop<::unicode::EastAsianWidth> { { 0x0D4A, 0x0D4F }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x0D54 .. 0x0D56: Lo     [3] MALAYALAM LETTER CHILLU M..MALAYALAM LETTER CHILLU LLL
    // 0x0D57 .. 0x0D57: Mc         MALAYALAM AU LENGTH MARK
    // 0x0D58 .. 0x0D5E: No     [7] MALAYALAM FRACTION ONE ONE-HUNDRED-AND-SIXTIETH..MALAYALAM FRACTION ONE FIFTH
    // 0x0D5F .. 0x0D61: Lo     [3] MALAYALAM LETTER ARCHAIC II..MALAYALAM LETTER VOCALIC LL
    // 0x0D62 .. 0x0D63: Mn     [2] MALAYALAM VOWEL SIGN VOCALIC L..MALAYALAM VOWEL SIGN VOCALIC LL
    Prop<::unicode::EastAsianWidth> { { 0x0D54, 0x0D63 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x0D66 .. 0x0D6F: Nd    [10] MALAYALAM DIGIT ZERO..MALAYALAM DIGIT NINE
    // 0x0D70 .. 0x0D78: No     [9] MALAYALAM NUMBER TEN..MALAYALAM FRACTION THREE SIXTEENTHS
    // 0x0D79 .. 0x0D79: So         MALAYALAM DATE MARK
    // 0x0D7A .. 0x0D7F: Lo     [6] MALAYALAM LETTER CHILLU NN..MALAYALAM LETTER CHILLU K
    Prop<::unicode::EastAsianWidth> { { 0x0D66, 0x0D7F }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x0D81 .. 0x0D81: Mn         SINHALA SIGN CANDRABINDU
    // 0x0D82 .. 0x0D83: Mc     [2] SINHALA SIGN ANUSVARAYA..SINHALA SIGN VISARGAYA
    Prop<::unicode::EastAsianWidth> { { 0x0D81, 0x0D83 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0D85, 0x0D96 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [18] SINHALA LETTER AYANNA..SINHALA LETTER AUYANNA
    Prop<::unicode::EastAsianWidth> { { 0x0D9A, 0x0DB1 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [24] SINHALA LETTER ALPAPRAANA KAYANNA..SINHALA LETTER DANTAJA NAYANNA
    Prop<::unicode::EastAsianWidth> { { 0x0DB3, 0x0DBB }, ::unicode::EastAsianWidth::Neutral }, // Lo     [9] SINHALA LETTER SANYAKA DAYANNA..SINHALA LETTER RAYANNA
    Prop<::unicode::EastAsianWidth> { { 0x0DBD, 0x0DBD }, ::unicode::EastAsianWidth::Neutral }, // Lo         SINHALA LETTER DANTAJA LAYANNA
    Prop<::unicode::EastAsianWidth> { { 0x0DC0, 0x0DC6 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] SINHALA LETTER VAYANNA..SINHALA LETTER FAYANNA
    Prop<::unicode::EastAsianWidth> { { 0x0DCA, 0x0DCA }, ::unicode::EastAsianWidth::Neutral }, // Mn         SINHALA SIGN AL-LAKUNA
    // 0x0DCF .. 0x0DD1: Mc     [3] SINHALA VOWEL SIGN AELA-PILLA..SINHALA VOWEL SIGN DIGA AEDA-PILLA
    // 0x0DD2 .. 0x0DD4: Mn     [3] SINHALA VOWEL SIGN KETTI IS-PILLA..SINHALA VOWEL SIGN KETTI PAA-PILLA
    Prop<::unicode::EastAsianWidth> { { 0x0DCF, 0x0DD4 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x0DD6, 0x0DD6 }, ::unicode::EastAsianWidth::Neutral }, // Mn         SINHALA VOWEL SIGN DIGA PAA-PILLA
    Prop<::unicode::EastAsianWidth> { { 0x0DD8, 0x0DDF }, ::unicode::EastAsianWidth::Neutral }, // Mc     [8] SINHALA VOWEL SIGN GAETTA-PILLA..SINHALA VOWEL SIGN GAYANUKITTA
    Prop<::unicode::EastAsianWidth> { { 0x0DE6, 0x0DEF }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] SINHALA LITH DIGIT ZERO..SINHALA LITH DIGIT NINE
    // 0x0DF2 .. 0x0DF3: Mc     [2] SINHALA VOWEL SIGN DIGA GAETTA-PILLA..SINHALA VOWEL SIGN DIGA GAYANUKITTA
    // 0x0DF4 .. 0x0DF4: Po         SINHALA PUNCTUATION KUNDDALIYA
    Prop<::unicode::EastAsianWidth> { { 0x0DF2, 0x0DF4 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x0E01 .. 0x0E30: Lo    [48] THAI CHARACTER KO KAI..THAI CHARACTER SARA A
    // 0x0E31 .. 0x0E31: Mn         THAI CHARACTER MAI HAN-AKAT
    // 0x0E32 .. 0x0E33: Lo     [2] THAI CHARACTER SARA AA..THAI CHARACTER SARA AM
    // 0x0E34 .. 0x0E3A: Mn     [7] THAI CHARACTER SARA I..THAI CHARACTER PHINTHU
    Prop<::unicode::EastAsianWidth> { { 0x0E01, 0x0E3A }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x0E3F .. 0x0E3F: Sc         THAI CURRENCY SYMBOL BAHT
    // 0x0E40 .. 0x0E45: Lo     [6] THAI CHARACTER SARA E..THAI CHARACTER LAKKHANGYAO
    // 0x0E46 .. 0x0E46: Lm         THAI CHARACTER MAIYAMOK
    // 0x0E47 .. 0x0E4E: Mn     [8] THAI CHARACTER MAITAIKHU..THAI CHARACTER YAMAKKAN
    // 0x0E4F .. 0x0E4F: Po         THAI CHARACTER FONGMAN
    // 0x0E50 .. 0x0E59: Nd    [10] THAI DIGIT ZERO..THAI DIGIT NINE
    // 0x0E5A .. 0x0E5B: Po     [2] THAI CHARACTER ANGKHANKHU..THAI CHARACTER KHOMUT
    Prop<::unicode::EastAsianWidth> { { 0x0E3F, 0x0E5B }, ::unicode::EastAsianWidth::Neutral }, // #7
    Prop<::unicode::EastAsianWidth> { { 0x0E81, 0x0E82 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] LAO LETTER KO..LAO LETTER KHO SUNG
    Prop<::unicode::EastAsianWidth> { { 0x0E84, 0x0E84 }, ::unicode::EastAsianWidth::Neutral }, // Lo         LAO LETTER KHO TAM
    Prop<::unicode::EastAsianWidth> { { 0x0E86, 0x0E8A }, ::unicode::EastAsianWidth::Neutral }, // Lo     [5] LAO LETTER PALI GHA..LAO LETTER SO TAM
    Prop<::unicode::EastAsianWidth> { { 0x0E8C, 0x0EA3 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [24] LAO LETTER PALI JHA..LAO LETTER LO LING
    Prop<::unicode::EastAsianWidth> { { 0x0EA5, 0x0EA5 }, ::unicode::EastAsianWidth::Neutral }, // Lo         LAO LETTER LO LOOT
    // 0x0EA7 .. 0x0EB0: Lo    [10] LAO LETTER WO..LAO VOWEL SIGN A
    // 0x0EB1 .. 0x0EB1: Mn         LAO VOWEL SIGN MAI KAN
    // 0x0EB2 .. 0x0EB3: Lo     [2] LAO VOWEL SIGN AA..LAO VOWEL SIGN AM
    // 0x0EB4 .. 0x0EBC: Mn     [9] LAO VOWEL SIGN I..LAO SEMIVOWEL SIGN LO
    // 0x0EBD .. 0x0EBD: Lo         LAO SEMIVOWEL SIGN NYO
    Prop<::unicode::EastAsianWidth> { { 0x0EA7, 0x0EBD }, ::unicode::EastAsianWidth::Neutral }, // #5
    Prop<::unicode::EastAsianWidth> { { 0x0EC0, 0x0EC4 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [5] LAO VOWEL SIGN E..LAO VOWEL SIGN AI
    Prop<::unicode::EastAsianWidth> { { 0x0EC6, 0x0EC6 }, ::unicode::EastAsianWidth::Neutral }, // Lm         LAO KO LA
    Prop<::unicode::EastAsianWidth> { { 0x0EC8, 0x0ECE }, ::unicode::EastAsianWidth::Neutral }, // Mn     [7] LAO TONE MAI EK..LAO YAMAKKAN
    Prop<::unicode::EastAsianWidth> { { 0x0ED0, 0x0ED9 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] LAO DIGIT ZERO..LAO DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x0EDC, 0x0EDF }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] LAO HO NO..LAO LETTER KHMU NYO
    // 0x0F00 .. 0x0F00: Lo         TIBETAN SYLLABLE OM
    // 0x0F01 .. 0x0F03: So     [3] TIBETAN MARK GTER YIG MGO TRUNCATED A..TIBETAN MARK GTER YIG MGO -UM GTER TSHEG MA
    // 0x0F04 .. 0x0F12: Po    [15] TIBETAN MARK INITIAL YIG MGO MDUN MA..TIBETAN MARK RGYA GRAM SHAD
    // 0x0F13 .. 0x0F13: So         TIBETAN MARK CARET -DZUD RTAGS ME LONG CAN
    // 0x0F14 .. 0x0F14: Po         TIBETAN MARK GTER TSHEG
    // 0x0F15 .. 0x0F17: So     [3] TIBETAN LOGOTYPE SIGN CHAD RTAGS..TIBETAN ASTROLOGICAL SIGN SGRA GCAN -CHAR RTAGS
    // 0x0F18 .. 0x0F19: Mn     [2] TIBETAN ASTROLOGICAL SIGN -KHYUD PA..TIBETAN ASTROLOGICAL SIGN SDONG TSHUGS
    // 0x0F1A .. 0x0F1F: So     [6] TIBETAN SIGN RDEL DKAR GCIG..TIBETAN SIGN RDEL DKAR RDEL NAG
    // 0x0F20 .. 0x0F29: Nd    [10] TIBETAN DIGIT ZERO..TIBETAN DIGIT NINE
    // 0x0F2A .. 0x0F33: No    [10] TIBETAN DIGIT HALF ONE..TIBETAN DIGIT HALF ZERO
    // 0x0F34 .. 0x0F34: So         TIBETAN MARK BSDUS RTAGS
    // 0x0F35 .. 0x0F35: Mn         TIBETAN MARK NGAS BZUNG NYI ZLA
    // 0x0F36 .. 0x0F36: So         TIBETAN MARK CARET -DZUD RTAGS BZHI MIG CAN
    // 0x0F37 .. 0x0F37: Mn         TIBETAN MARK NGAS BZUNG SGOR RTAGS
    // 0x0F38 .. 0x0F38: So         TIBETAN MARK CHE MGO
    // 0x0F39 .. 0x0F39: Mn         TIBETAN MARK TSA -PHRU
    // 0x0F3A .. 0x0F3A: Ps         TIBETAN MARK GUG RTAGS GYON
    // 0x0F3B .. 0x0F3B: Pe         TIBETAN MARK GUG RTAGS GYAS
    // 0x0F3C .. 0x0F3C: Ps         TIBETAN MARK ANG KHANG GYON
    // 0x0F3D .. 0x0F3D: Pe         TIBETAN MARK ANG KHANG GYAS
    // 0x0F3E .. 0x0F3F: Mc     [2] TIBETAN SIGN YAR TSHES..TIBETAN SIGN MAR TSHES
    // 0x0F40 .. 0x0F47: Lo     [8] TIBETAN LETTER KA..TIBETAN LETTER JA
    Prop<::unicode::EastAsianWidth> { { 0x0F00, 0x0F47 }, ::unicode::EastAsianWidth::Neutral }, // #22
    Prop<::unicode::EastAsianWidth> { { 0x0F49, 0x0F6C }, ::unicode::EastAsianWidth::Neutral }, // Lo    [36] TIBETAN LETTER NYA..TIBETAN LETTER RRA
    // 0x0F71 .. 0x0F7E: Mn    [14] TIBETAN VOWEL SIGN AA..TIBETAN SIGN RJES SU NGA RO
    // 0x0F7F .. 0x0F7F: Mc         TIBETAN SIGN RNAM BCAD
    // 0x0F80 .. 0x0F84: Mn     [5] TIBETAN VOWEL SIGN REVERSED I..TIBETAN MARK HALANTA
    // 0x0F85 .. 0x0F85: Po         TIBETAN MARK PALUTA
    // 0x0F86 .. 0x0F87: Mn     [2] TIBETAN SIGN LCI RTAGS..TIBETAN SIGN YANG RTAGS
    // 0x0F88 .. 0x0F8C: Lo     [5] TIBETAN SIGN LCE TSA CAN..TIBETAN SIGN INVERTED MCHU CAN
    // 0x0F8D .. 0x0F97: Mn    [11] TIBETAN SUBJOINED SIGN LCE TSA CAN..TIBETAN SUBJOINED LETTER JA
    Prop<::unicode::EastAsianWidth> { { 0x0F71, 0x0F97 }, ::unicode::EastAsianWidth::Neutral }, // #7
    Prop<::unicode::EastAsianWidth> { { 0x0F99, 0x0FBC }, ::unicode::EastAsianWidth::Neutral }, // Mn    [36] TIBETAN SUBJOINED LETTER NYA..TIBETAN SUBJOINED LETTER FIXED-FORM RA
    // 0x0FBE .. 0x0FC5: So     [8] TIBETAN KU RU KHA..TIBETAN SYMBOL RDO RJE
    // 0x0FC6 .. 0x0FC6: Mn         TIBETAN SYMBOL PADMA GDAN
    // 0x0FC7 .. 0x0FCC: So     [6] TIBETAN SYMBOL RDO RJE RGYA GRAM..TIBETAN SYMBOL NOR BU BZHI -KHYIL
    Prop<::unicode::EastAsianWidth> { { 0x0FBE, 0x0FCC }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x0FCE .. 0x0FCF: So     [2] TIBETAN SIGN RDEL NAG RDEL DKAR..TIBETAN SIGN RDEL NAG GSUM
    // 0x0FD0 .. 0x0FD4: Po     [5] TIBETAN MARK BSKA- SHOG GI MGO RGYAN..TIBETAN MARK CLOSING BRDA RNYING YIG MGO SGAB MA
    // 0x0FD5 .. 0x0FD8: So     [4] RIGHT-FACING SVASTI SIGN..LEFT-FACING SVASTI SIGN WITH DOTS
    // 0x0FD9 .. 0x0FDA: Po     [2] TIBETAN MARK LEADING MCHAN RTAGS..TIBETAN MARK TRAILING MCHAN RTAGS
    Prop<::unicode::EastAsianWidth> { { 0x0FCE, 0x0FDA }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x1000 .. 0x102A: Lo    [43] MYANMAR LETTER KA..MYANMAR LETTER AU
    // 0x102B .. 0x102C: Mc     [2] MYANMAR VOWEL SIGN TALL AA..MYANMAR VOWEL SIGN AA
    // 0x102D .. 0x1030: Mn     [4] MYANMAR VOWEL SIGN I..MYANMAR VOWEL SIGN UU
    // 0x1031 .. 0x1031: Mc         MYANMAR VOWEL SIGN E
    // 0x1032 .. 0x1037: Mn     [6] MYANMAR VOWEL SIGN AI..MYANMAR SIGN DOT BELOW
    // 0x1038 .. 0x1038: Mc         MYANMAR SIGN VISARGA
    // 0x1039 .. 0x103A: Mn     [2] MYANMAR SIGN VIRAMA..MYANMAR SIGN ASAT
    // 0x103B .. 0x103C: Mc     [2] MYANMAR CONSONANT SIGN MEDIAL YA..MYANMAR CONSONANT SIGN MEDIAL RA
    // 0x103D .. 0x103E: Mn     [2] MYANMAR CONSONANT SIGN MEDIAL WA..MYANMAR CONSONANT SIGN MEDIAL HA
    // 0x103F .. 0x103F: Lo         MYANMAR LETTER GREAT SA
    // 0x1040 .. 0x1049: Nd    [10] MYANMAR DIGIT ZERO..MYANMAR DIGIT NINE
    // 0x104A .. 0x104F: Po     [6] MYANMAR SIGN LITTLE SECTION..MYANMAR SYMBOL GENITIVE
    // 0x1050 .. 0x1055: Lo     [6] MYANMAR LETTER SHA..MYANMAR LETTER VOCALIC LL
    // 0x1056 .. 0x1057: Mc     [2] MYANMAR VOWEL SIGN VOCALIC R..MYANMAR VOWEL SIGN VOCALIC RR
    // 0x1058 .. 0x1059: Mn     [2] MYANMAR VOWEL SIGN VOCALIC L..MYANMAR VOWEL SIGN VOCALIC LL
    // 0x105A .. 0x105D: Lo     [4] MYANMAR LETTER MON NGA..MYANMAR LETTER MON BBE
    // 0x105E .. 0x1060: Mn     [3] MYANMAR CONSONANT SIGN MON MEDIAL NA..MYANMAR CONSONANT SIGN MON MEDIAL LA
    // 0x1061 .. 0x1061: Lo         MYANMAR LETTER SGAW KAREN SHA
    // 0x1062 .. 0x1064: Mc     [3] MYANMAR VOWEL SIGN SGAW KAREN EU..MYANMAR TONE MARK SGAW KAREN KE PHO
    // 0x1065 .. 0x1066: Lo     [2] MYANMAR LETTER WESTERN PWO KAREN THA..MYANMAR LETTER WESTERN PWO KAREN PWA
    // 0x1067 .. 0x106D: Mc     [7] MYANMAR VOWEL SIGN WESTERN PWO KAREN EU..MYANMAR SIGN WESTERN PWO KAREN TONE-5
    // 0x106E .. 0x1070: Lo     [3] MYANMAR LETTER EASTERN PWO KAREN NNA..MYANMAR LETTER EASTERN PWO KAREN GHWA
    // 0x1071 .. 0x1074: Mn     [4] MYANMAR VOWEL SIGN GEBA KAREN I..MYANMAR VOWEL SIGN KAYAH EE
    // 0x1075 .. 0x1081: Lo    [13] MYANMAR LETTER SHAN KA..MYANMAR LETTER SHAN HA
    // 0x1082 .. 0x1082: Mn         MYANMAR CONSONANT SIGN SHAN MEDIAL WA
    // 0x1083 .. 0x1084: Mc     [2] MYANMAR VOWEL SIGN SHAN AA..MYANMAR VOWEL SIGN SHAN E
    // 0x1085 .. 0x1086: Mn     [2] MYANMAR VOWEL SIGN SHAN E ABOVE..MYANMAR VOWEL SIGN SHAN FINAL Y
    // 0x1087 .. 0x108C: Mc     [6] MYANMAR SIGN SHAN TONE-2..MYANMAR SIGN SHAN COUNCIL TONE-3
    // 0x108D .. 0x108D: Mn         MYANMAR SIGN SHAN COUNCIL EMPHATIC TONE
    // 0x108E .. 0x108E: Lo         MYANMAR LETTER RUMAI PALAUNG FA
    // 0x108F .. 0x108F: Mc         MYANMAR SIGN RUMAI PALAUNG TONE-5
    // 0x1090 .. 0x1099: Nd    [10] MYANMAR SHAN DIGIT ZERO..MYANMAR SHAN DIGIT NINE
    // 0x109A .. 0x109C: Mc     [3] MYANMAR SIGN KHAMTI TONE-1..MYANMAR VOWEL SIGN AITON A
    // 0x109D .. 0x109D: Mn         MYANMAR VOWEL SIGN AITON AI
    // 0x109E .. 0x109F: So     [2] MYANMAR SYMBOL SHAN ONE..MYANMAR SYMBOL SHAN EXCLAMATION
    // 0x10A0 .. 0x10C5: Lu    [38] GEORGIAN CAPITAL LETTER AN..GEORGIAN CAPITAL LETTER HOE
    Prop<::unicode::EastAsianWidth> { { 0x1000, 0x10C5 }, ::unicode::EastAsianWidth::Neutral }, // #36
    Prop<::unicode::EastAsianWidth> { { 0x10C7, 0x10C7 }, ::unicode::EastAsianWidth::Neutral }, // Lu         GEORGIAN CAPITAL LETTER YN
    Prop<::unicode::EastAsianWidth> { { 0x10CD, 0x10CD }, ::unicode::EastAsianWidth::Neutral }, // Lu         GEORGIAN CAPITAL LETTER AEN
    // 0x10D0 .. 0x10FA: Ll    [43] GEORGIAN LETTER AN..GEORGIAN LETTER AIN
    // 0x10FB .. 0x10FB: Po         GEORGIAN PARAGRAPH SEPARATOR
    // 0x10FC .. 0x10FC: Lm         MODIFIER LETTER GEORGIAN NAR
    // 0x10FD .. 0x10FF: Ll     [3] GEORGIAN LETTER AEN..GEORGIAN LETTER LABIAL SIGN
    Prop<::unicode::EastAsianWidth> { { 0x10D0, 0x10FF }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x1100, 0x115F }, ::unicode::EastAsianWidth::Wide }, // Lo    [96] HANGUL CHOSEONG KIYEOK..HANGUL CHOSEONG FILLER
    // 0x1160 .. 0x11FF: Lo   [160] HANGUL JUNGSEONG FILLER..HANGUL JONGSEONG SSANGNIEUN
    // 0x1200 .. 0x1248: Lo    [73] ETHIOPIC SYLLABLE HA..ETHIOPIC SYLLABLE QWA
    Prop<::unicode::EastAsianWidth> { { 0x1160, 0x1248 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x124A, 0x124D }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] ETHIOPIC SYLLABLE QWI..ETHIOPIC SYLLABLE QWE
    Prop<::unicode::EastAsianWidth> { { 0x1250, 0x1256 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE QHA..ETHIOPIC SYLLABLE QHO
    Prop<::unicode::EastAsianWidth> { { 0x1258, 0x1258 }, ::unicode::EastAsianWidth::Neutral }, // Lo         ETHIOPIC SYLLABLE QHWA
    Prop<::unicode::EastAsianWidth> { { 0x125A, 0x125D }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] ETHIOPIC SYLLABLE QHWI..ETHIOPIC SYLLABLE QHWE
    Prop<::unicode::EastAsianWidth> { { 0x1260, 0x1288 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [41] ETHIOPIC SYLLABLE BA..ETHIOPIC SYLLABLE XWA
    Prop<::unicode::EastAsianWidth> { { 0x128A, 0x128D }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] ETHIOPIC SYLLABLE XWI..ETHIOPIC SYLLABLE XWE
    Prop<::unicode::EastAsianWidth> { { 0x1290, 0x12B0 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [33] ETHIOPIC SYLLABLE NA..ETHIOPIC SYLLABLE KWA
    Prop<::unicode::EastAsianWidth> { { 0x12B2, 0x12B5 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] ETHIOPIC SYLLABLE KWI..ETHIOPIC SYLLABLE KWE
    Prop<::unicode::EastAsianWidth> { { 0x12B8, 0x12BE }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE KXA..ETHIOPIC SYLLABLE KXO
    Prop<::unicode::EastAsianWidth> { { 0x12C0, 0x12C0 }, ::unicode::EastAsianWidth::Neutral }, // Lo         ETHIOPIC SYLLABLE KXWA
    Prop<::unicode::EastAsianWidth> { { 0x12C2, 0x12C5 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] ETHIOPIC SYLLABLE KXWI..ETHIOPIC SYLLABLE KXWE
    Prop<::unicode::EastAsianWidth> { { 0x12C8, 0x12D6 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [15] ETHIOPIC SYLLABLE WA..ETHIOPIC SYLLABLE PHARYNGEAL O
    Prop<::unicode::EastAsianWidth> { { 0x12D8, 0x1310 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [57] ETHIOPIC SYLLABLE ZA..ETHIOPIC SYLLABLE GWA
    Prop<::unicode::EastAsianWidth> { { 0x1312, 0x1315 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] ETHIOPIC SYLLABLE GWI..ETHIOPIC SYLLABLE GWE
    Prop<::unicode::EastAsianWidth> { { 0x1318, 0x135A }, ::unicode::EastAsianWidth::Neutral }, // Lo    [67] ETHIOPIC SYLLABLE GGA..ETHIOPIC SYLLABLE FYA
    // 0x135D .. 0x135F: Mn     [3] ETHIOPIC COMBINING GEMINATION AND VOWEL LENGTH MARK..ETHIOPIC COMBINING GEMINATION MARK
    // 0x1360 .. 0x1368: Po     [9] ETHIOPIC SECTION MARK..ETHIOPIC PARAGRAPH SEPARATOR
    // 0x1369 .. 0x137C: No    [20] ETHIOPIC DIGIT ONE..ETHIOPIC NUMBER TEN THOUSAND
    Prop<::unicode::EastAsianWidth> { { 0x135D, 0x137C }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x1380 .. 0x138F: Lo    [16] ETHIOPIC SYLLABLE SEBATBEIT MWA..ETHIOPIC SYLLABLE PWE
    // 0x1390 .. 0x1399: So    [10] ETHIOPIC TONAL MARK YIZET..ETHIOPIC TONAL MARK KURT
    Prop<::unicode::EastAsianWidth> { { 0x1380, 0x1399 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x13A0, 0x13F5 }, ::unicode::EastAsianWidth::Neutral }, // Lu    [86] CHEROKEE LETTER A..CHEROKEE LETTER MV
    Prop<::unicode::EastAsianWidth> { { 0x13F8, 0x13FD }, ::unicode::EastAsianWidth::Neutral }, // Ll     [6] CHEROKEE SMALL LETTER YE..CHEROKEE SMALL LETTER MV
    // 0x1400 .. 0x1400: Pd         CANADIAN SYLLABICS HYPHEN
    // 0x1401 .. 0x166C: Lo   [620] CANADIAN SYLLABICS E..CANADIAN SYLLABICS CARRIER TTSA
    // 0x166D .. 0x166D: So         CANADIAN SYLLABICS CHI SIGN
    // 0x166E .. 0x166E: Po         CANADIAN SYLLABICS FULL STOP
    // 0x166F .. 0x167F: Lo    [17] CANADIAN SYLLABICS QAI..CANADIAN SYLLABICS BLACKFOOT W
    // 0x1680 .. 0x1680: Zs         OGHAM SPACE MARK
    // 0x1681 .. 0x169A: Lo    [26] OGHAM LETTER BEITH..OGHAM LETTER PEITH
    // 0x169B .. 0x169B: Ps         OGHAM FEATHER MARK
    // 0x169C .. 0x169C: Pe         OGHAM REVERSED FEATHER MARK
    Prop<::unicode::EastAsianWidth> { { 0x1400, 0x169C }, ::unicode::EastAsianWidth::Neutral }, // #9
    // 0x16A0 .. 0x16EA: Lo    [75] RUNIC LETTER FEHU FEOH FE F..RUNIC LETTER X
    // 0x16EB .. 0x16ED: Po     [3] RUNIC SINGLE PUNCTUATION..RUNIC CROSS PUNCTUATION
    // 0x16EE .. 0x16F0: Nl     [3] RUNIC ARLAUG SYMBOL..RUNIC BELGTHOR SYMBOL
    // 0x16F1 .. 0x16F8: Lo     [8] RUNIC LETTER K..RUNIC LETTER FRANKS CASKET AESC
    Prop<::unicode::EastAsianWidth> { { 0x16A0, 0x16F8 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x1700 .. 0x1711: Lo    [18] TAGALOG LETTER A..TAGALOG LETTER HA
    // 0x1712 .. 0x1714: Mn     [3] TAGALOG VOWEL SIGN I..TAGALOG SIGN VIRAMA
    // 0x1715 .. 0x1715: Mc         TAGALOG SIGN PAMUDPOD
    Prop<::unicode::EastAsianWidth> { { 0x1700, 0x1715 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x171F .. 0x171F: Lo         TAGALOG LETTER ARCHAIC RA
    // 0x1720 .. 0x1731: Lo    [18] HANUNOO LETTER A..HANUNOO LETTER HA
    // 0x1732 .. 0x1733: Mn     [2] HANUNOO VOWEL SIGN I..HANUNOO VOWEL SIGN U
    // 0x1734 .. 0x1734: Mc         HANUNOO SIGN PAMUDPOD
    // 0x1735 .. 0x1736: Po     [2] PHILIPPINE SINGLE PUNCTUATION..PHILIPPINE DOUBLE PUNCTUATION
    Prop<::unicode::EastAsianWidth> { { 0x171F, 0x1736 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x1740 .. 0x1751: Lo    [18] BUHID LETTER A..BUHID LETTER HA
    // 0x1752 .. 0x1753: Mn     [2] BUHID VOWEL SIGN I..BUHID VOWEL SIGN U
    Prop<::unicode::EastAsianWidth> { { 0x1740, 0x1753 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x1760, 0x176C }, ::unicode::EastAsianWidth::Neutral }, // Lo    [13] TAGBANWA LETTER A..TAGBANWA LETTER YA
    Prop<::unicode::EastAsianWidth> { { 0x176E, 0x1770 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [3] TAGBANWA LETTER LA..TAGBANWA LETTER SA
    Prop<::unicode::EastAsianWidth> { { 0x1772, 0x1773 }, ::unicode::EastAsianWidth::Neutral }, // Mn     [2] TAGBANWA VOWEL SIGN I..TAGBANWA VOWEL SIGN U
    // 0x1780 .. 0x17B3: Lo    [52] KHMER LETTER KA..KHMER INDEPENDENT VOWEL QAU
    // 0x17B4 .. 0x17B5: Mn     [2] KHMER VOWEL INHERENT AQ..KHMER VOWEL INHERENT AA
    // 0x17B6 .. 0x17B6: Mc         KHMER VOWEL SIGN AA
    // 0x17B7 .. 0x17BD: Mn     [7] KHMER VOWEL SIGN I..KHMER VOWEL SIGN UA
    // 0x17BE .. 0x17C5: Mc     [8] KHMER VOWEL SIGN OE..KHMER VOWEL SIGN AU
    // 0x17C6 .. 0x17C6: Mn         KHMER SIGN NIKAHIT
    // 0x17C7 .. 0x17C8: Mc     [2] KHMER SIGN REAHMUK..KHMER SIGN YUUKALEAPINTU
    // 0x17C9 .. 0x17D3: Mn    [11] KHMER SIGN MUUSIKATOAN..KHMER SIGN BATHAMASAT
    // 0x17D4 .. 0x17D6: Po     [3] KHMER SIGN KHAN..KHMER SIGN CAMNUC PII KUUH
    // 0x17D7 .. 0x17D7: Lm         KHMER SIGN LEK TOO
    // 0x17D8 .. 0x17DA: Po     [3] KHMER SIGN BEYYAL..KHMER SIGN KOOMUUT
    // 0x17DB .. 0x17DB: Sc         KHMER CURRENCY SYMBOL RIEL
    // 0x17DC .. 0x17DC: Lo         KHMER SIGN AVAKRAHASANYA
    // 0x17DD .. 0x17DD: Mn         KHMER SIGN ATTHACAN
    Prop<::unicode::EastAsianWidth> { { 0x1780, 0x17DD }, ::unicode::EastAsianWidth::Neutral }, // #14
    Prop<::unicode::EastAsianWidth> { { 0x17E0, 0x17E9 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] KHMER DIGIT ZERO..KHMER DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x17F0, 0x17F9 }, ::unicode::EastAsianWidth::Neutral }, // No    [10] KHMER SYMBOL LEK ATTAK SON..KHMER SYMBOL LEK ATTAK PRAM-BUON
    // 0x1800 .. 0x1805: Po     [6] MONGOLIAN BIRGA..MONGOLIAN FOUR DOTS
    // 0x1806 .. 0x1806: Pd         MONGOLIAN TODO SOFT HYPHEN
    // 0x1807 .. 0x180A: Po     [4] MONGOLIAN SIBE SYLLABLE BOUNDARY MARKER..MONGOLIAN NIRUGU
    // 0x180B .. 0x180D: Mn     [3] MONGOLIAN FREE VARIATION SELECTOR ONE..MONGOLIAN FREE VARIATION SELECTOR THREE
    // 0x180E .. 0x180E: Cf         MONGOLIAN VOWEL SEPARATOR
    // 0x180F .. 0x180F: Mn         MONGOLIAN FREE VARIATION SELECTOR FOUR
    // 0x1810 .. 0x1819: Nd    [10] MONGOLIAN DIGIT ZERO..MONGOLIAN DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x1800, 0x1819 }, ::unicode::EastAsianWidth::Neutral }, // #7
    // 0x1820 .. 0x1842: Lo    [35] MONGOLIAN LETTER A..MONGOLIAN LETTER CHI
    // 0x1843 .. 0x1843: Lm         MONGOLIAN LETTER TODO LONG VOWEL SIGN
    // 0x1844 .. 0x1878: Lo    [53] MONGOLIAN LETTER TODO E..MONGOLIAN LETTER CHA WITH TWO DOTS
    Prop<::unicode::EastAsianWidth> { { 0x1820, 0x1878 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x1880 .. 0x1884: Lo     [5] MONGOLIAN LETTER ALI GALI ANUSVARA ONE..MONGOLIAN LETTER ALI GALI INVERTED UBADAMA
    // 0x1885 .. 0x1886: Mn     [2] MONGOLIAN LETTER ALI GALI BALUDA..MONGOLIAN LETTER ALI GALI THREE BALUDA
    // 0x1887 .. 0x18A8: Lo    [34] MONGOLIAN LETTER ALI GALI A..MONGOLIAN LETTER MANCHU ALI GALI BHA
    // 0x18A9 .. 0x18A9: Mn         MONGOLIAN LETTER ALI GALI DAGALGA
    // 0x18AA .. 0x18AA: Lo         MONGOLIAN LETTER MANCHU ALI GALI LHA
    Prop<::unicode::EastAsianWidth> { { 0x1880, 0x18AA }, ::unicode::EastAsianWidth::Neutral }, // #5
    Prop<::unicode::EastAsianWidth> { { 0x18B0, 0x18F5 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [70] CANADIAN SYLLABICS OY..CANADIAN SYLLABICS CARRIER DENTAL S
    Prop<::unicode::EastAsianWidth> { { 0x1900, 0x191E }, ::unicode::EastAsianWidth::Neutral }, // Lo    [31] LIMBU VOWEL-CARRIER LETTER..LIMBU LETTER TRA
    // 0x1920 .. 0x1922: Mn     [3] LIMBU VOWEL SIGN A..LIMBU VOWEL SIGN U
    // 0x1923 .. 0x1926: Mc     [4] LIMBU VOWEL SIGN EE..LIMBU VOWEL SIGN AU
    // 0x1927 .. 0x1928: Mn     [2] LIMBU VOWEL SIGN E..LIMBU VOWEL SIGN O
    // 0x1929 .. 0x192B: Mc     [3] LIMBU SUBJOINED LETTER YA..LIMBU SUBJOINED LETTER WA
    Prop<::unicode::EastAsianWidth> { { 0x1920, 0x192B }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x1930 .. 0x1931: Mc     [2] LIMBU SMALL LETTER KA..LIMBU SMALL LETTER NGA
    // 0x1932 .. 0x1932: Mn         LIMBU SMALL LETTER ANUSVARA
    // 0x1933 .. 0x1938: Mc     [6] LIMBU SMALL LETTER TA..LIMBU SMALL LETTER LA
    // 0x1939 .. 0x193B: Mn     [3] LIMBU SIGN MUKPHRENG..LIMBU SIGN SA-I
    Prop<::unicode::EastAsianWidth> { { 0x1930, 0x193B }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x1940, 0x1940 }, ::unicode::EastAsianWidth::Neutral }, // So         LIMBU SIGN LOO
    // 0x1944 .. 0x1945: Po     [2] LIMBU EXCLAMATION MARK..LIMBU QUESTION MARK
    // 0x1946 .. 0x194F: Nd    [10] LIMBU DIGIT ZERO..LIMBU DIGIT NINE
    // 0x1950 .. 0x196D: Lo    [30] TAI LE LETTER KA..TAI LE LETTER AI
    Prop<::unicode::EastAsianWidth> { { 0x1944, 0x196D }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x1970, 0x1974 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [5] TAI LE LETTER TONE-2..TAI LE LETTER TONE-6
    Prop<::unicode::EastAsianWidth> { { 0x1980, 0x19AB }, ::unicode::EastAsianWidth::Neutral }, // Lo    [44] NEW TAI LUE LETTER HIGH QA..NEW TAI LUE LETTER LOW SUA
    Prop<::unicode::EastAsianWidth> { { 0x19B0, 0x19C9 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [26] NEW TAI LUE VOWEL SIGN VOWEL SHORTENER..NEW TAI LUE TONE MARK-2
    // 0x19D0 .. 0x19D9: Nd    [10] NEW TAI LUE DIGIT ZERO..NEW TAI LUE DIGIT NINE
    // 0x19DA .. 0x19DA: No         NEW TAI LUE THAM DIGIT ONE
    Prop<::unicode::EastAsianWidth> { { 0x19D0, 0x19DA }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x19DE .. 0x19DF: So     [2] NEW TAI LUE SIGN LAE..NEW TAI LUE SIGN LAEV
    // 0x19E0 .. 0x19FF: So    [32] KHMER SYMBOL PATHAMASAT..KHMER SYMBOL DAP-PRAM ROC
    // 0x1A00 .. 0x1A16: Lo    [23] BUGINESE LETTER KA..BUGINESE LETTER HA
    // 0x1A17 .. 0x1A18: Mn     [2] BUGINESE VOWEL SIGN I..BUGINESE VOWEL SIGN U
    // 0x1A19 .. 0x1A1A: Mc     [2] BUGINESE VOWEL SIGN E..BUGINESE VOWEL SIGN O
    // 0x1A1B .. 0x1A1B: Mn         BUGINESE VOWEL SIGN AE
    Prop<::unicode::EastAsianWidth> { { 0x19DE, 0x1A1B }, ::unicode::EastAsianWidth::Neutral }, // #6
    // 0x1A1E .. 0x1A1F: Po     [2] BUGINESE PALLAWA..BUGINESE END OF SECTION
    // 0x1A20 .. 0x1A54: Lo    [53] TAI THAM LETTER HIGH KA..TAI THAM LETTER GREAT SA
    // 0x1A55 .. 0x1A55: Mc         TAI THAM CONSONANT SIGN MEDIAL RA
    // 0x1A56 .. 0x1A56: Mn         TAI THAM CONSONANT SIGN MEDIAL LA
    // 0x1A57 .. 0x1A57: Mc         TAI THAM CONSONANT SIGN LA TANG LAI
    // 0x1A58 .. 0x1A5E: Mn     [7] TAI THAM SIGN MAI KANG LAI..TAI THAM CONSONANT SIGN SA
    Prop<::unicode::EastAsianWidth> { { 0x1A1E, 0x1A5E }, ::unicode::EastAsianWidth::Neutral }, // #6
    // 0x1A60 .. 0x1A60: Mn         TAI THAM SIGN SAKOT
    // 0x1A61 .. 0x1A61: Mc         TAI THAM VOWEL SIGN A
    // 0x1A62 .. 0x1A62: Mn         TAI THAM VOWEL SIGN MAI SAT
    // 0x1A63 .. 0x1A64: Mc     [2] TAI THAM VOWEL SIGN AA..TAI THAM VOWEL SIGN TALL AA
    // 0x1A65 .. 0x1A6C: Mn     [8] TAI THAM VOWEL SIGN I..TAI THAM VOWEL SIGN OA BELOW
    // 0x1A6D .. 0x1A72: Mc     [6] TAI THAM VOWEL SIGN OY..TAI THAM VOWEL SIGN THAM AI
    // 0x1A73 .. 0x1A7C: Mn    [10] TAI THAM VOWEL SIGN OA ABOVE..TAI THAM SIGN KHUEN-LUE KARAN
    Prop<::unicode::EastAsianWidth> { { 0x1A60, 0x1A7C }, ::unicode::EastAsianWidth::Neutral }, // #7
    // 0x1A7F .. 0x1A7F: Mn         TAI THAM COMBINING CRYPTOGRAMMIC DOT
    // 0x1A80 .. 0x1A89: Nd    [10] TAI THAM HORA DIGIT ZERO..TAI THAM HORA DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x1A7F, 0x1A89 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x1A90, 0x1A99 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] TAI THAM THAM DIGIT ZERO..TAI THAM THAM DIGIT NINE
    // 0x1AA0 .. 0x1AA6: Po     [7] TAI THAM SIGN WIANG..TAI THAM SIGN REVERSED ROTATED RANA
    // 0x1AA7 .. 0x1AA7: Lm         TAI THAM SIGN MAI YAMOK
    // 0x1AA8 .. 0x1AAD: Po     [6] TAI THAM SIGN KAAN..TAI THAM SIGN CAANG
    Prop<::unicode::EastAsianWidth> { { 0x1AA0, 0x1AAD }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x1AB0 .. 0x1ABD: Mn    [14] COMBINING DOUBLED CIRCUMFLEX ACCENT..COMBINING PARENTHESES BELOW
    // 0x1ABE .. 0x1ABE: Me         COMBINING PARENTHESES OVERLAY
    // 0x1ABF .. 0x1ADD: Mn    [31] COMBINING LATIN SMALL LETTER W BELOW..COMBINING DOT-AND-RING BELOW
    Prop<::unicode::EastAsianWidth> { { 0x1AB0, 0x1ADD }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x1AE0, 0x1AEB }, ::unicode::EastAsianWidth::Neutral }, // Mn    [12] COMBINING LEFT TACK ABOVE..COMBINING DOUBLE RIGHTWARDS ARROW ABOVE
    // 0x1B00 .. 0x1B03: Mn     [4] BALINESE SIGN ULU RICEM..BALINESE SIGN SURANG
    // 0x1B04 .. 0x1B04: Mc         BALINESE SIGN BISAH
    // 0x1B05 .. 0x1B33: Lo    [47] BALINESE LETTER AKARA..BALINESE LETTER HA
    // 0x1B34 .. 0x1B34: Mn         BALINESE SIGN REREKAN
    // 0x1B35 .. 0x1B35: Mc         BALINESE VOWEL SIGN TEDUNG
    // 0x1B36 .. 0x1B3A: Mn     [5] BALINESE VOWEL SIGN ULU..BALINESE VOWEL SIGN RA REPA
    // 0x1B3B .. 0x1B3B: Mc         BALINESE VOWEL SIGN RA REPA TEDUNG
    // 0x1B3C .. 0x1B3C: Mn         BALINESE VOWEL SIGN LA LENGA
    // 0x1B3D .. 0x1B41: Mc     [5] BALINESE VOWEL SIGN LA LENGA TEDUNG..BALINESE VOWEL SIGN TALING REPA TEDUNG
    // 0x1B42 .. 0x1B42: Mn         BALINESE VOWEL SIGN PEPET
    // 0x1B43 .. 0x1B44: Mc     [2] BALINESE VOWEL SIGN PEPET TEDUNG..BALINESE ADEG ADEG
    // 0x1B45 .. 0x1B4C: Lo     [8] BALINESE LETTER KAF SASAK..BALINESE LETTER ARCHAIC JNYA
    Prop<::unicode::EastAsianWidth> { { 0x1B00, 0x1B4C }, ::unicode::EastAsianWidth::Neutral }, // #12
    // 0x1B4E .. 0x1B4F: Po     [2] BALINESE INVERTED CARIK SIKI..BALINESE INVERTED CARIK PAREREN
    // 0x1B50 .. 0x1B59: Nd    [10] BALINESE DIGIT ZERO..BALINESE DIGIT NINE
    // 0x1B5A .. 0x1B60: Po     [7] BALINESE PANTI..BALINESE PAMENENG
    // 0x1B61 .. 0x1B6A: So    [10] BALINESE MUSICAL SYMBOL DONG..BALINESE MUSICAL SYMBOL DANG GEDE
    // 0x1B6B .. 0x1B73: Mn     [9] BALINESE MUSICAL SYMBOL COMBINING TEGEH..BALINESE MUSICAL SYMBOL COMBINING GONG
    // 0x1B74 .. 0x1B7C: So     [9] BALINESE MUSICAL SYMBOL RIGHT-HAND OPEN DUG..BALINESE MUSICAL SYMBOL LEFT-HAND OPEN PING
    // 0x1B7D .. 0x1B7F: Po     [3] BALINESE PANTI LANTANG..BALINESE PANTI BAWAK
    // 0x1B80 .. 0x1B81: Mn     [2] SUNDANESE SIGN PANYECEK..SUNDANESE SIGN PANGLAYAR
    // 0x1B82 .. 0x1B82: Mc         SUNDANESE SIGN PANGWISAD
    // 0x1B83 .. 0x1BA0: Lo    [30] SUNDANESE LETTER A..SUNDANESE LETTER HA
    // 0x1BA1 .. 0x1BA1: Mc         SUNDANESE CONSONANT SIGN PAMINGKAL
    // 0x1BA2 .. 0x1BA5: Mn     [4] SUNDANESE CONSONANT SIGN PANYAKRA..SUNDANESE VOWEL SIGN PANYUKU
    // 0x1BA6 .. 0x1BA7: Mc     [2] SUNDANESE VOWEL SIGN PANAELAENG..SUNDANESE VOWEL SIGN PANOLONG
    // 0x1BA8 .. 0x1BA9: Mn     [2] SUNDANESE VOWEL SIGN PAMEPET..SUNDANESE VOWEL SIGN PANEULEUNG
    // 0x1BAA .. 0x1BAA: Mc         SUNDANESE SIGN PAMAAEH
    // 0x1BAB .. 0x1BAD: Mn     [3] SUNDANESE SIGN VIRAMA..SUNDANESE CONSONANT SIGN PASANGAN WA
    // 0x1BAE .. 0x1BAF: Lo     [2] SUNDANESE LETTER KHA..SUNDANESE LETTER SYA
    // 0x1BB0 .. 0x1BB9: Nd    [10] SUNDANESE DIGIT ZERO..SUNDANESE DIGIT NINE
    // 0x1BBA .. 0x1BBF: Lo     [6] SUNDANESE AVAGRAHA..SUNDANESE LETTER FINAL M
    // 0x1BC0 .. 0x1BE5: Lo    [38] BATAK LETTER A..BATAK LETTER U
    // 0x1BE6 .. 0x1BE6: Mn         BATAK SIGN TOMPI
    // 0x1BE7 .. 0x1BE7: Mc         BATAK VOWEL SIGN E
    // 0x1BE8 .. 0x1BE9: Mn     [2] BATAK VOWEL SIGN PAKPAK E..BATAK VOWEL SIGN EE
    // 0x1BEA .. 0x1BEC: Mc     [3] BATAK VOWEL SIGN I..BATAK VOWEL SIGN O
    // 0x1BED .. 0x1BED: Mn         BATAK VOWEL SIGN KARO O
    // 0x1BEE .. 0x1BEE: Mc         BATAK VOWEL SIGN U
    // 0x1BEF .. 0x1BF1: Mn     [3] BATAK VOWEL SIGN U FOR SIMALUNGUN SA..BATAK CONSONANT SIGN H
    // 0x1BF2 .. 0x1BF3: Mc     [2] BATAK PANGOLAT..BATAK PANONGONAN
    Prop<::unicode::EastAsianWidth> { { 0x1B4E, 0x1BF3 }, ::unicode::EastAsianWidth::Neutral }, // #28
    // 0x1BFC .. 0x1BFF: Po     [4] BATAK SYMBOL BINDU NA METEK..BATAK SYMBOL BINDU PANGOLAT
    // 0x1C00 .. 0x1C23: Lo    [36] LEPCHA LETTER KA..LEPCHA LETTER A
    // 0x1C24 .. 0x1C2B: Mc     [8] LEPCHA SUBJOINED LETTER YA..LEPCHA VOWEL SIGN UU
    // 0x1C2C .. 0x1C33: Mn     [8] LEPCHA VOWEL SIGN E..LEPCHA CONSONANT SIGN T
    // 0x1C34 .. 0x1C35: Mc     [2] LEPCHA CONSONANT SIGN NYIN-DO..LEPCHA CONSONANT SIGN KANG
    // 0x1C36 .. 0x1C37: Mn     [2] LEPCHA SIGN RAN..LEPCHA SIGN NUKTA
    Prop<::unicode::EastAsianWidth> { { 0x1BFC, 0x1C37 }, ::unicode::EastAsianWidth::Neutral }, // #6
    // 0x1C3B .. 0x1C3F: Po     [5] LEPCHA PUNCTUATION TA-ROL..LEPCHA PUNCTUATION TSHOOK
    // 0x1C40 .. 0x1C49: Nd    [10] LEPCHA DIGIT ZERO..LEPCHA DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x1C3B, 0x1C49 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x1C4D .. 0x1C4F: Lo     [3] LEPCHA LETTER TTA..LEPCHA LETTER DDA
    // 0x1C50 .. 0x1C59: Nd    [10] OL CHIKI DIGIT ZERO..OL CHIKI DIGIT NINE
    // 0x1C5A .. 0x1C77: Lo    [30] OL CHIKI LETTER LA..OL CHIKI LETTER OH
    // 0x1C78 .. 0x1C7D: Lm     [6] OL CHIKI MU TTUDDAG..OL CHIKI AHAD
    // 0x1C7E .. 0x1C7F: Po     [2] OL CHIKI PUNCTUATION MUCAAD..OL CHIKI PUNCTUATION DOUBLE MUCAAD
    // 0x1C80 .. 0x1C8A: L&    [11] CYRILLIC SMALL LETTER ROUNDED VE..CYRILLIC SMALL LETTER TJE
    Prop<::unicode::EastAsianWidth> { { 0x1C4D, 0x1C8A }, ::unicode::EastAsianWidth::Neutral }, // #6
    Prop<::unicode::EastAsianWidth> { { 0x1C90, 0x1CBA }, ::unicode::EastAsianWidth::Neutral }, // Lu    [43] GEORGIAN MTAVRULI CAPITAL LETTER AN..GEORGIAN MTAVRULI CAPITAL LETTER AIN
    // 0x1CBD .. 0x1CBF: Lu     [3] GEORGIAN MTAVRULI CAPITAL LETTER AEN..GEORGIAN MTAVRULI CAPITAL LETTER LABIAL SIGN
    // 0x1CC0 .. 0x1CC7: Po     [8] SUNDANESE PUNCTUATION BINDU SURYA..SUNDANESE PUNCTUATION BINDU BA SATANGA
    Prop<::unicode::EastAsianWidth> { { 0x1CBD, 0x1CC7 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x1CD0 .. 0x1CD2: Mn     [3] VEDIC TONE KARSHANA..VEDIC TONE PRENKHA
    // 0x1CD3 .. 0x1CD3: Po         VEDIC SIGN NIHSHVASA
    // 0x1CD4 .. 0x1CE0: Mn    [13] VEDIC SIGN YAJURVEDIC MIDLINE SVARITA..VEDIC TONE RIGVEDIC KASHMIRI INDEPENDENT SVARITA
    // 0x1CE1 .. 0x1CE1: Mc         VEDIC TONE ATHARVAVEDIC INDEPENDENT SVARITA
    // 0x1CE2 .. 0x1CE8: Mn     [7] VEDIC SIGN VISARGA SVARITA..VEDIC SIGN VISARGA ANUDATTA WITH TAIL
    // 0x1CE9 .. 0x1CEC: Lo     [4] VEDIC SIGN ANUSVARA ANTARGOMUKHA..VEDIC SIGN ANUSVARA VAMAGOMUKHA WITH TAIL
    // 0x1CED .. 0x1CED: Mn         VEDIC SIGN TIRYAK
    // 0x1CEE .. 0x1CF3: Lo     [6] VEDIC SIGN HEXIFORM LONG ANUSVARA..VEDIC SIGN ROTATED ARDHAVISARGA
    // 0x1CF4 .. 0x1CF4: Mn         VEDIC TONE CANDRA ABOVE
    // 0x1CF5 .. 0x1CF6: Lo     [2] VEDIC SIGN JIHVAMULIYA..VEDIC SIGN UPADHMANIYA
    // 0x1CF7 .. 0x1CF7: Mc         VEDIC SIGN ATIKRAMA
    // 0x1CF8 .. 0x1CF9: Mn     [2] VEDIC TONE RING ABOVE..VEDIC TONE DOUBLE RING ABOVE
    // 0x1CFA .. 0x1CFA: Lo         VEDIC SIGN DOUBLE ANUSVARA ANTARGOMUKHA
    Prop<::unicode::EastAsianWidth> { { 0x1CD0, 0x1CFA }, ::unicode::EastAsianWidth::Neutral }, // #13
    // 0x1D00 .. 0x1D2B: Ll    [44] LATIN LETTER SMALL CAPITAL A..CYRILLIC LETTER SMALL CAPITAL EL
    // 0x1D2C .. 0x1D6A: Lm    [63] MODIFIER LETTER CAPITAL A..GREEK SUBSCRIPT SMALL LETTER CHI
    // 0x1D6B .. 0x1D77: Ll    [13] LATIN SMALL LETTER UE..LATIN SMALL LETTER TURNED G
    // 0x1D78 .. 0x1D78: Lm         MODIFIER LETTER CYRILLIC EN
    // 0x1D79 .. 0x1D7F: Ll     [7] LATIN SMALL LETTER INSULAR G..LATIN SMALL LETTER UPSILON WITH STROKE
    // 0x1D80 .. 0x1D9A: Ll    [27] LATIN SMALL LETTER B WITH PALATAL HOOK..LATIN SMALL LETTER EZH WITH RETROFLEX HOOK
    // 0x1D9B .. 0x1DBF: Lm    [37] MODIFIER LETTER SMALL TURNED ALPHA..MODIFIER LETTER SMALL THETA
    // 0x1DC0 .. 0x1DFF: Mn    [64] COMBINING DOTTED GRAVE ACCENT..COMBINING RIGHT ARROWHEAD AND DOWN ARROWHEAD BELOW
    // 0x1E00 .. 0x1EFF: L&   [256] LATIN CAPITAL LETTER A WITH RING BELOW..LATIN SMALL LETTER Y WITH LOOP
    // 0x1F00 .. 0x1F15: L&    [22] GREEK SMALL LETTER ALPHA WITH PSILI..GREEK SMALL LETTER EPSILON WITH DASIA AND OXIA
    Prop<::unicode::EastAsianWidth> { { 0x1D00, 0x1F15 }, ::unicode::EastAsianWidth::Neutral }, // #10
    Prop<::unicode::EastAsianWidth> { { 0x1F18, 0x1F1D }, ::unicode::EastAsianWidth::Neutral }, // Lu     [6] GREEK CAPITAL LETTER EPSILON WITH PSILI..GREEK CAPITAL LETTER EPSILON WITH DASIA AND OXIA
    Prop<::unicode::EastAsianWidth> { { 0x1F20, 0x1F45 }, ::unicode::EastAsianWidth::Neutral }, // L&    [38] GREEK SMALL LETTER ETA WITH PSILI..GREEK SMALL LETTER OMICRON WITH DASIA AND OXIA
    Prop<::unicode::EastAsianWidth> { { 0x1F48, 0x1F4D }, ::unicode::EastAsianWidth::Neutral }, // Lu     [6] GREEK CAPITAL LETTER OMICRON WITH PSILI..GREEK CAPITAL LETTER OMICRON WITH DASIA AND OXIA
    Prop<::unicode::EastAsianWidth> { { 0x1F50, 0x1F57 }, ::unicode::EastAsianWidth::Neutral }, // Ll     [8] GREEK SMALL LETTER UPSILON WITH PSILI..GREEK SMALL LETTER UPSILON WITH DASIA AND PERISPOMENI
    Prop<::unicode::EastAsianWidth> { { 0x1F59, 0x1F59 }, ::unicode::EastAsianWidth::Neutral }, // Lu         GREEK CAPITAL LETTER UPSILON WITH DASIA
    Prop<::unicode::EastAsianWidth> { { 0x1F5B, 0x1F5B }, ::unicode::EastAsianWidth::Neutral }, // Lu         GREEK CAPITAL LETTER UPSILON WITH DASIA AND VARIA
    Prop<::unicode::EastAsianWidth> { { 0x1F5D, 0x1F5D }, ::unicode::EastAsianWidth::Neutral }, // Lu         GREEK CAPITAL LETTER UPSILON WITH DASIA AND OXIA
    Prop<::unicode::EastAsianWidth> { { 0x1F5F, 0x1F7D }, ::unicode::EastAsianWidth::Neutral }, // L&    [31] GREEK CAPITAL LETTER UPSILON WITH DASIA AND PERISPOMENI..GREEK SMALL LETTER OMEGA WITH OXIA
    Prop<::unicode::EastAsianWidth> { { 0x1F80, 0x1FB4 }, ::unicode::EastAsianWidth::Neutral }, // L&    [53] GREEK SMALL LETTER ALPHA WITH PSILI AND YPOGEGRAMMENI..GREEK SMALL LETTER ALPHA WITH OXIA AND YPOGEGRAMMENI
    // 0x1FB6 .. 0x1FBC: L&     [7] GREEK SMALL LETTER ALPHA WITH PERISPOMENI..GREEK CAPITAL LETTER ALPHA WITH PROSGEGRAMMENI
    // 0x1FBD .. 0x1FBD: Sk         GREEK KORONIS
    // 0x1FBE .. 0x1FBE: Ll         GREEK PROSGEGRAMMENI
    // 0x1FBF .. 0x1FC1: Sk     [3] GREEK PSILI..GREEK DIALYTIKA AND PERISPOMENI
    // 0x1FC2 .. 0x1FC4: Ll     [3] GREEK SMALL LETTER ETA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER ETA WITH OXIA AND YPOGEGRAMMENI
    Prop<::unicode::EastAsianWidth> { { 0x1FB6, 0x1FC4 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x1FC6 .. 0x1FCC: L&     [7] GREEK SMALL LETTER ETA WITH PERISPOMENI..GREEK CAPITAL LETTER ETA WITH PROSGEGRAMMENI
    // 0x1FCD .. 0x1FCF: Sk     [3] GREEK PSILI AND VARIA..GREEK PSILI AND PERISPOMENI
    // 0x1FD0 .. 0x1FD3: Ll     [4] GREEK SMALL LETTER IOTA WITH VRACHY..GREEK SMALL LETTER IOTA WITH DIALYTIKA AND OXIA
    Prop<::unicode::EastAsianWidth> { { 0x1FC6, 0x1FD3 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x1FD6, 0x1FDB }, ::unicode::EastAsianWidth::Neutral }, // L&     [6] GREEK SMALL LETTER IOTA WITH PERISPOMENI..GREEK CAPITAL LETTER IOTA WITH OXIA
    // 0x1FDD .. 0x1FDF: Sk     [3] GREEK DASIA AND VARIA..GREEK DASIA AND PERISPOMENI
    // 0x1FE0 .. 0x1FEC: L&    [13] GREEK SMALL LETTER UPSILON WITH VRACHY..GREEK CAPITAL LETTER RHO WITH DASIA
    // 0x1FED .. 0x1FEF: Sk     [3] GREEK DIALYTIKA AND VARIA..GREEK VARIA
    Prop<::unicode::EastAsianWidth> { { 0x1FDD, 0x1FEF }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x1FF2, 0x1FF4 }, ::unicode::EastAsianWidth::Neutral }, // Ll     [3] GREEK SMALL LETTER OMEGA WITH VARIA AND YPOGEGRAMMENI..GREEK SMALL LETTER OMEGA WITH OXIA AND YPOGEGRAMMENI
    // 0x1FF6 .. 0x1FFC: L&     [7] GREEK SMALL LETTER OMEGA WITH PERISPOMENI..GREEK CAPITAL LETTER OMEGA WITH PROSGEGRAMMENI
    // 0x1FFD .. 0x1FFE: Sk     [2] GREEK OXIA..GREEK DASIA
    Prop<::unicode::EastAsianWidth> { { 0x1FF6, 0x1FFE }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x2000 .. 0x200A: Zs    [11] EN QUAD..HAIR SPACE
    // 0x200B .. 0x200F: Cf     [5] ZERO WIDTH SPACE..RIGHT-TO-LEFT MARK
    Prop<::unicode::EastAsianWidth> { { 0x2000, 0x200F }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x2010, 0x2010 }, ::unicode::EastAsianWidth::Ambiguous }, // Pd         HYPHEN
    Prop<::unicode::EastAsianWidth> { { 0x2011, 0x2012 }, ::unicode::EastAsianWidth::Neutral }, // Pd     [2] NON-BREAKING HYPHEN..FIGURE DASH
    // 0x2013 .. 0x2015: Pd     [3] EN DASH..HORIZONTAL BAR
    // 0x2016 .. 0x2016: Po         DOUBLE VERTICAL LINE
    Prop<::unicode::EastAsianWidth> { { 0x2013, 0x2016 }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x2017, 0x2017 }, ::unicode::EastAsianWidth::Neutral }, // Po         DOUBLE LOW LINE
    // 0x2018 .. 0x2018: Pi         LEFT SINGLE QUOTATION MARK
    // 0x2019 .. 0x2019: Pf         RIGHT SINGLE QUOTATION MARK
    Prop<::unicode::EastAsianWidth> { { 0x2018, 0x2019 }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    // 0x201A .. 0x201A: Ps         SINGLE LOW-9 QUOTATION MARK
    // 0x201B .. 0x201B: Pi         SINGLE HIGH-REVERSED-9 QUOTATION MARK
    Prop<::unicode::EastAsianWidth> { { 0x201A, 0x201B }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x201C .. 0x201C: Pi         LEFT DOUBLE QUOTATION MARK
    // 0x201D .. 0x201D: Pf         RIGHT DOUBLE QUOTATION MARK
    Prop<::unicode::EastAsianWidth> { { 0x201C, 0x201D }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    // 0x201E .. 0x201E: Ps         DOUBLE LOW-9 QUOTATION MARK
    // 0x201F .. 0x201F: Pi         DOUBLE HIGH-REVERSED-9 QUOTATION MARK
    Prop<::unicode::EastAsianWidth> { { 0x201E, 0x201F }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x2020, 0x2022 }, ::unicode::EastAsianWidth::Ambiguous }, // Po     [3] DAGGER..BULLET
    Prop<::unicode::EastAsianWidth> { { 0x2023, 0x2023 }, ::unicode::EastAsianWidth::Neutral }, // Po         TRIANGULAR BULLET
    Prop<::unicode::EastAsianWidth> { { 0x2024, 0x2027 }, ::unicode::EastAsianWidth::Ambiguous }, // Po     [4] ONE DOT LEADER..HYPHENATION POINT
    // 0x2028 .. 0x2028: Zl         LINE SEPARATOR
    // 0x2029 .. 0x2029: Zp         PARAGRAPH SEPARATOR
    // 0x202A .. 0x202E: Cf     [5] LEFT-TO-RIGHT EMBEDDING..RIGHT-TO-LEFT OVERRIDE
    // 0x202F .. 0x202F: Zs         NARROW NO-BREAK SPACE
    Prop<::unicode::EastAsianWidth> { { 0x2028, 0x202F }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x2030, 0x2030 }, ::unicode::EastAsianWidth::Ambiguous }, // Po         PER MILLE SIGN
    Prop<::unicode::EastAsianWidth> { { 0x2031, 0x2031 }, ::unicode::EastAsianWidth::Neutral }, // Po         PER TEN THOUSAND SIGN
    Prop<::unicode::EastAsianWidth> { { 0x2032, 0x2033 }, ::unicode::EastAsianWidth::Ambiguous }, // Po     [2] PRIME..DOUBLE PRIME
    Prop<::unicode::EastAsianWidth> { { 0x2034, 0x2034 }, ::unicode::EastAsianWidth::Neutral }, // Po         TRIPLE PRIME
    Prop<::unicode::EastAsianWidth> { { 0x2035, 0x2035 }, ::unicode::EastAsianWidth::Ambiguous }, // Po         REVERSED PRIME
    // 0x2036 .. 0x2038: Po     [3] REVERSED DOUBLE PRIME..CARET
    // 0x2039 .. 0x2039: Pi         SINGLE LEFT-POINTING ANGLE QUOTATION MARK
    // 0x203A .. 0x203A: Pf         SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
    Prop<::unicode::EastAsianWidth> { { 0x2036, 0x203A }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x203B, 0x203B }, ::unicode::EastAsianWidth::Ambiguous }, // Po         REFERENCE MARK
    Prop<::unicode::EastAsianWidth> { { 0x203C, 0x203D }, ::unicode::EastAsianWidth::Neutral }, // Po     [2] DOUBLE EXCLAMATION MARK..INTERROBANG
    Prop<::unicode::EastAsianWidth> { { 0x203E, 0x203E }, ::unicode::EastAsianWidth::Ambiguous }, // Po         OVERLINE
    // 0x203F .. 0x2040: Pc     [2] UNDERTIE..CHARACTER TIE
    // 0x2041 .. 0x2043: Po     [3] CARET INSERTION POINT..HYPHEN BULLET
    // 0x2044 .. 0x2044: Sm         FRACTION SLASH
    // 0x2045 .. 0x2045: Ps         LEFT SQUARE BRACKET WITH QUILL
    // 0x2046 .. 0x2046: Pe         RIGHT SQUARE BRACKET WITH QUILL
    // 0x2047 .. 0x2051: Po    [11] DOUBLE QUESTION MARK..TWO ASTERISKS ALIGNED VERTICALLY
    // 0x2052 .. 0x2052: Sm         COMMERCIAL MINUS SIGN
    // 0x2053 .. 0x2053: Po         SWUNG DASH
    // 0x2054 .. 0x2054: Pc         INVERTED UNDERTIE
    // 0x2055 .. 0x205E: Po    [10] FLOWER PUNCTUATION MARK..VERTICAL FOUR DOTS
    // 0x205F .. 0x205F: Zs         MEDIUM MATHEMATICAL SPACE
    // 0x2060 .. 0x2064: Cf     [5] WORD JOINER..INVISIBLE PLUS
    Prop<::unicode::EastAsianWidth> { { 0x203F, 0x2064 }, ::unicode::EastAsianWidth::Neutral }, // #12
    // 0x2066 .. 0x206F: Cf    [10] LEFT-TO-RIGHT ISOLATE..NOMINAL DIGIT SHAPES
    // 0x2070 .. 0x2070: No         SUPERSCRIPT ZERO
    // 0x2071 .. 0x2071: Lm         SUPERSCRIPT LATIN SMALL LETTER I
    Prop<::unicode::EastAsianWidth> { { 0x2066, 0x2071 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x2074, 0x2074 }, ::unicode::EastAsianWidth::Ambiguous }, // No         SUPERSCRIPT FOUR
    // 0x2075 .. 0x2079: No     [5] SUPERSCRIPT FIVE..SUPERSCRIPT NINE
    // 0x207A .. 0x207C: Sm     [3] SUPERSCRIPT PLUS SIGN..SUPERSCRIPT EQUALS SIGN
    // 0x207D .. 0x207D: Ps         SUPERSCRIPT LEFT PARENTHESIS
    // 0x207E .. 0x207E: Pe         SUPERSCRIPT RIGHT PARENTHESIS
    Prop<::unicode::EastAsianWidth> { { 0x2075, 0x207E }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x207F, 0x207F }, ::unicode::EastAsianWidth::Ambiguous }, // Lm         SUPERSCRIPT LATIN SMALL LETTER N
    Prop<::unicode::EastAsianWidth> { { 0x2080, 0x2080 }, ::unicode::EastAsianWidth::Neutral }, // No         SUBSCRIPT ZERO
    Prop<::unicode::EastAsianWidth> { { 0x2081, 0x2084 }, ::unicode::EastAsianWidth::Ambiguous }, // No     [4] SUBSCRIPT ONE..SUBSCRIPT FOUR
    // 0x2085 .. 0x2089: No     [5] SUBSCRIPT FIVE..SUBSCRIPT NINE
    // 0x208A .. 0x208C: Sm     [3] SUBSCRIPT PLUS SIGN..SUBSCRIPT EQUALS SIGN
    // 0x208D .. 0x208D: Ps         SUBSCRIPT LEFT PARENTHESIS
    // 0x208E .. 0x208E: Pe         SUBSCRIPT RIGHT PARENTHESIS
    Prop<::unicode::EastAsianWidth> { { 0x2085, 0x208E }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x2090, 0x209C }, ::unicode::EastAsianWidth::Neutral }, // Lm    [13] LATIN SUBSCRIPT SMALL LETTER A..LATIN SUBSCRIPT SMALL LETTER T
    Prop<::unicode::EastAsianWidth> { { 0x20A0, 0x20A8 }, ::unicode::EastAsianWidth::Neutral }, // Sc     [9] EURO-CURRENCY SIGN..RUPEE SIGN
    Prop<::unicode::EastAsianWidth> { { 0x20A9, 0x20A9 }, ::unicode::EastAsianWidth::HalfWidth }, // Sc         WON SIGN
    Prop<::unicode::EastAsianWidth> { { 0x20AA, 0x20AB }, ::unicode::EastAsianWidth::Neutral }, // Sc     [2] NEW SHEQEL SIGN..DONG SIGN
    Prop<::unicode::EastAsianWidth> { { 0x20AC, 0x20AC }, ::unicode::EastAsianWidth::Ambiguous }, // Sc         EURO SIGN
    Prop<::unicode::EastAsianWidth> { { 0x20AD, 0x20C1 }, ::unicode::EastAsianWidth::Neutral }, // Sc    [21] KIP SIGN..SAUDI RIYAL SIGN
    // 0x20D0 .. 0x20DC: Mn    [13] COMBINING LEFT HARPOON ABOVE..COMBINING FOUR DOTS ABOVE
    // 0x20DD .. 0x20E0: Me     [4] COMBINING ENCLOSING CIRCLE..COMBINING ENCLOSING CIRCLE BACKSLASH
    // 0x20E1 .. 0x20E1: Mn         COMBINING LEFT RIGHT ARROW ABOVE
    // 0x20E2 .. 0x20E4: Me     [3] COMBINING ENCLOSING SCREEN..COMBINING ENCLOSING UPWARD POINTING TRIANGLE
    // 0x20E5 .. 0x20F0: Mn    [12] COMBINING REVERSE SOLIDUS OVERLAY..COMBINING ASTERISK ABOVE
    Prop<::unicode::EastAsianWidth> { { 0x20D0, 0x20F0 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x2100 .. 0x2101: So     [2] ACCOUNT OF..ADDRESSED TO THE SUBJECT
    // 0x2102 .. 0x2102: Lu         DOUBLE-STRUCK CAPITAL C
    Prop<::unicode::EastAsianWidth> { { 0x2100, 0x2102 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x2103, 0x2103 }, ::unicode::EastAsianWidth::Ambiguous }, // So         DEGREE CELSIUS
    Prop<::unicode::EastAsianWidth> { { 0x2104, 0x2104 }, ::unicode::EastAsianWidth::Neutral }, // So         CENTRE LINE SYMBOL
    Prop<::unicode::EastAsianWidth> { { 0x2105, 0x2105 }, ::unicode::EastAsianWidth::Ambiguous }, // So         CARE OF
    // 0x2106 .. 0x2106: So         CADA UNA
    // 0x2107 .. 0x2107: Lu         EULER CONSTANT
    // 0x2108 .. 0x2108: So         SCRUPLE
    Prop<::unicode::EastAsianWidth> { { 0x2106, 0x2108 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x2109, 0x2109 }, ::unicode::EastAsianWidth::Ambiguous }, // So         DEGREE FAHRENHEIT
    Prop<::unicode::EastAsianWidth> { { 0x210A, 0x2112 }, ::unicode::EastAsianWidth::Neutral }, // L&     [9] SCRIPT SMALL G..SCRIPT CAPITAL L
    Prop<::unicode::EastAsianWidth> { { 0x2113, 0x2113 }, ::unicode::EastAsianWidth::Ambiguous }, // Ll         SCRIPT SMALL L
    // 0x2114 .. 0x2114: So         L B BAR SYMBOL
    // 0x2115 .. 0x2115: Lu         DOUBLE-STRUCK CAPITAL N
    Prop<::unicode::EastAsianWidth> { { 0x2114, 0x2115 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x2116, 0x2116 }, ::unicode::EastAsianWidth::Ambiguous }, // So         NUMERO SIGN
    // 0x2117 .. 0x2117: So         SOUND RECORDING COPYRIGHT
    // 0x2118 .. 0x2118: Sm         SCRIPT CAPITAL P
    // 0x2119 .. 0x211D: Lu     [5] DOUBLE-STRUCK CAPITAL P..DOUBLE-STRUCK CAPITAL R
    // 0x211E .. 0x2120: So     [3] PRESCRIPTION TAKE..SERVICE MARK
    Prop<::unicode::EastAsianWidth> { { 0x2117, 0x2120 }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x2121, 0x2122 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] TELEPHONE SIGN..TRADE MARK SIGN
    // 0x2123 .. 0x2123: So         VERSICLE
    // 0x2124 .. 0x2124: Lu         DOUBLE-STRUCK CAPITAL Z
    // 0x2125 .. 0x2125: So         OUNCE SIGN
    Prop<::unicode::EastAsianWidth> { { 0x2123, 0x2125 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x2126, 0x2126 }, ::unicode::EastAsianWidth::Ambiguous }, // Lu         OHM SIGN
    // 0x2127 .. 0x2127: So         INVERTED OHM SIGN
    // 0x2128 .. 0x2128: Lu         BLACK-LETTER CAPITAL Z
    // 0x2129 .. 0x2129: So         TURNED GREEK SMALL LETTER IOTA
    // 0x212A .. 0x212A: Lu         KELVIN SIGN
    Prop<::unicode::EastAsianWidth> { { 0x2127, 0x212A }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x212B, 0x212B }, ::unicode::EastAsianWidth::Ambiguous }, // Lu         ANGSTROM SIGN
    // 0x212C .. 0x212D: Lu     [2] SCRIPT CAPITAL B..BLACK-LETTER CAPITAL C
    // 0x212E .. 0x212E: So         ESTIMATED SYMBOL
    // 0x212F .. 0x2134: L&     [6] SCRIPT SMALL E..SCRIPT SMALL O
    // 0x2135 .. 0x2138: Lo     [4] ALEF SYMBOL..DALET SYMBOL
    // 0x2139 .. 0x2139: Ll         INFORMATION SOURCE
    // 0x213A .. 0x213B: So     [2] ROTATED CAPITAL Q..FACSIMILE SIGN
    // 0x213C .. 0x213F: L&     [4] DOUBLE-STRUCK SMALL PI..DOUBLE-STRUCK CAPITAL PI
    // 0x2140 .. 0x2144: Sm     [5] DOUBLE-STRUCK N-ARY SUMMATION..TURNED SANS-SERIF CAPITAL Y
    // 0x2145 .. 0x2149: L&     [5] DOUBLE-STRUCK ITALIC CAPITAL D..DOUBLE-STRUCK ITALIC SMALL J
    // 0x214A .. 0x214A: So         PROPERTY LINE
    // 0x214B .. 0x214B: Sm         TURNED AMPERSAND
    // 0x214C .. 0x214D: So     [2] PER SIGN..AKTIESELSKAB
    // 0x214E .. 0x214E: Ll         TURNED SMALL F
    // 0x214F .. 0x214F: So         SYMBOL FOR SAMARITAN SOURCE
    // 0x2150 .. 0x2152: No     [3] VULGAR FRACTION ONE SEVENTH..VULGAR FRACTION ONE TENTH
    Prop<::unicode::EastAsianWidth> { { 0x212C, 0x2152 }, ::unicode::EastAsianWidth::Neutral }, // #15
    Prop<::unicode::EastAsianWidth> { { 0x2153, 0x2154 }, ::unicode::EastAsianWidth::Ambiguous }, // No     [2] VULGAR FRACTION ONE THIRD..VULGAR FRACTION TWO THIRDS
    Prop<::unicode::EastAsianWidth> { { 0x2155, 0x215A }, ::unicode::EastAsianWidth::Neutral }, // No     [6] VULGAR FRACTION ONE FIFTH..VULGAR FRACTION FIVE SIXTHS
    Prop<::unicode::EastAsianWidth> { { 0x215B, 0x215E }, ::unicode::EastAsianWidth::Ambiguous }, // No     [4] VULGAR FRACTION ONE EIGHTH..VULGAR FRACTION SEVEN EIGHTHS
    Prop<::unicode::EastAsianWidth> { { 0x215F, 0x215F }, ::unicode::EastAsianWidth::Neutral }, // No         FRACTION NUMERATOR ONE
    Prop<::unicode::EastAsianWidth> { { 0x2160, 0x216B }, ::unicode::EastAsianWidth::Ambiguous }, // Nl    [12] ROMAN NUMERAL ONE..ROMAN NUMERAL TWELVE
    Prop<::unicode::EastAsianWidth> { { 0x216C, 0x216F }, ::unicode::EastAsianWidth::Neutral }, // Nl     [4] ROMAN NUMERAL FIFTY..ROMAN NUMERAL ONE THOUSAND
    Prop<::unicode::EastAsianWidth> { { 0x2170, 0x2179 }, ::unicode::EastAsianWidth::Ambiguous }, // Nl    [10] SMALL ROMAN NUMERAL ONE..SMALL ROMAN NUMERAL TEN
    // 0x217A .. 0x2182: Nl     [9] SMALL ROMAN NUMERAL ELEVEN..ROMAN NUMERAL TEN THOUSAND
    // 0x2183 .. 0x2184: L&     [2] ROMAN NUMERAL REVERSED ONE HUNDRED..LATIN SMALL LETTER REVERSED C
    // 0x2185 .. 0x2188: Nl     [4] ROMAN NUMERAL SIX LATE FORM..ROMAN NUMERAL ONE HUNDRED THOUSAND
    Prop<::unicode::EastAsianWidth> { { 0x217A, 0x2188 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x2189, 0x2189 }, ::unicode::EastAsianWidth::Ambiguous }, // No         VULGAR FRACTION ZERO THIRDS
    Prop<::unicode::EastAsianWidth> { { 0x218A, 0x218B }, ::unicode::EastAsianWidth::Neutral }, // So     [2] TURNED DIGIT TWO..TURNED DIGIT THREE
    // 0x2190 .. 0x2194: Sm     [5] LEFTWARDS ARROW..LEFT RIGHT ARROW
    // 0x2195 .. 0x2199: So     [5] UP DOWN ARROW..SOUTH WEST ARROW
    Prop<::unicode::EastAsianWidth> { { 0x2190, 0x2199 }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    // 0x219A .. 0x219B: Sm     [2] LEFTWARDS ARROW WITH STROKE..RIGHTWARDS ARROW WITH STROKE
    // 0x219C .. 0x219F: So     [4] LEFTWARDS WAVE ARROW..UPWARDS TWO HEADED ARROW
    // 0x21A0 .. 0x21A0: Sm         RIGHTWARDS TWO HEADED ARROW
    // 0x21A1 .. 0x21A2: So     [2] DOWNWARDS TWO HEADED ARROW..LEFTWARDS ARROW WITH TAIL
    // 0x21A3 .. 0x21A3: Sm         RIGHTWARDS ARROW WITH TAIL
    // 0x21A4 .. 0x21A5: So     [2] LEFTWARDS ARROW FROM BAR..UPWARDS ARROW FROM BAR
    // 0x21A6 .. 0x21A6: Sm         RIGHTWARDS ARROW FROM BAR
    // 0x21A7 .. 0x21AD: So     [7] DOWNWARDS ARROW FROM BAR..LEFT RIGHT WAVE ARROW
    // 0x21AE .. 0x21AE: Sm         LEFT RIGHT ARROW WITH STROKE
    // 0x21AF .. 0x21B7: So     [9] DOWNWARDS ZIGZAG ARROW..CLOCKWISE TOP SEMICIRCLE ARROW
    Prop<::unicode::EastAsianWidth> { { 0x219A, 0x21B7 }, ::unicode::EastAsianWidth::Neutral }, // #10
    Prop<::unicode::EastAsianWidth> { { 0x21B8, 0x21B9 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] NORTH WEST ARROW TO LONG BAR..LEFTWARDS ARROW TO BAR OVER RIGHTWARDS ARROW TO BAR
    // 0x21BA .. 0x21CD: So    [20] ANTICLOCKWISE OPEN CIRCLE ARROW..LEFTWARDS DOUBLE ARROW WITH STROKE
    // 0x21CE .. 0x21CF: Sm     [2] LEFT RIGHT DOUBLE ARROW WITH STROKE..RIGHTWARDS DOUBLE ARROW WITH STROKE
    // 0x21D0 .. 0x21D1: So     [2] LEFTWARDS DOUBLE ARROW..UPWARDS DOUBLE ARROW
    Prop<::unicode::EastAsianWidth> { { 0x21BA, 0x21D1 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x21D2, 0x21D2 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         RIGHTWARDS DOUBLE ARROW
    Prop<::unicode::EastAsianWidth> { { 0x21D3, 0x21D3 }, ::unicode::EastAsianWidth::Neutral }, // So         DOWNWARDS DOUBLE ARROW
    Prop<::unicode::EastAsianWidth> { { 0x21D4, 0x21D4 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         LEFT RIGHT DOUBLE ARROW
    Prop<::unicode::EastAsianWidth> { { 0x21D5, 0x21E6 }, ::unicode::EastAsianWidth::Neutral }, // So    [18] UP DOWN DOUBLE ARROW..LEFTWARDS WHITE ARROW
    Prop<::unicode::EastAsianWidth> { { 0x21E7, 0x21E7 }, ::unicode::EastAsianWidth::Ambiguous }, // So         UPWARDS WHITE ARROW
    // 0x21E8 .. 0x21F3: So    [12] RIGHTWARDS WHITE ARROW..UP DOWN WHITE ARROW
    // 0x21F4 .. 0x21FF: Sm    [12] RIGHT ARROW WITH SMALL CIRCLE..LEFT RIGHT OPEN-HEADED ARROW
    Prop<::unicode::EastAsianWidth> { { 0x21E8, 0x21FF }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x2200, 0x2200 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         FOR ALL
    Prop<::unicode::EastAsianWidth> { { 0x2201, 0x2201 }, ::unicode::EastAsianWidth::Neutral }, // Sm         COMPLEMENT
    Prop<::unicode::EastAsianWidth> { { 0x2202, 0x2203 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm     [2] PARTIAL DIFFERENTIAL..THERE EXISTS
    Prop<::unicode::EastAsianWidth> { { 0x2204, 0x2206 }, ::unicode::EastAsianWidth::Neutral }, // Sm     [3] THERE DOES NOT EXIST..INCREMENT
    Prop<::unicode::EastAsianWidth> { { 0x2207, 0x2208 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm     [2] NABLA..ELEMENT OF
    Prop<::unicode::EastAsianWidth> { { 0x2209, 0x220A }, ::unicode::EastAsianWidth::Neutral }, // Sm     [2] NOT AN ELEMENT OF..SMALL ELEMENT OF
    Prop<::unicode::EastAsianWidth> { { 0x220B, 0x220B }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         CONTAINS AS MEMBER
    Prop<::unicode::EastAsianWidth> { { 0x220C, 0x220E }, ::unicode::EastAsianWidth::Neutral }, // Sm     [3] DOES NOT CONTAIN AS MEMBER..END OF PROOF
    Prop<::unicode::EastAsianWidth> { { 0x220F, 0x220F }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         N-ARY PRODUCT
    Prop<::unicode::EastAsianWidth> { { 0x2210, 0x2210 }, ::unicode::EastAsianWidth::Neutral }, // Sm         N-ARY COPRODUCT
    Prop<::unicode::EastAsianWidth> { { 0x2211, 0x2211 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         N-ARY SUMMATION
    Prop<::unicode::EastAsianWidth> { { 0x2212, 0x2214 }, ::unicode::EastAsianWidth::Neutral }, // Sm     [3] MINUS SIGN..DOT PLUS
    Prop<::unicode::EastAsianWidth> { { 0x2215, 0x2215 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         DIVISION SLASH
    Prop<::unicode::EastAsianWidth> { { 0x2216, 0x2219 }, ::unicode::EastAsianWidth::Neutral }, // Sm     [4] SET MINUS..BULLET OPERATOR
    Prop<::unicode::EastAsianWidth> { { 0x221A, 0x221A }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         SQUARE ROOT
    Prop<::unicode::EastAsianWidth> { { 0x221B, 0x221C }, ::unicode::EastAsianWidth::Neutral }, // Sm     [2] CUBE ROOT..FOURTH ROOT
    Prop<::unicode::EastAsianWidth> { { 0x221D, 0x2220 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm     [4] PROPORTIONAL TO..ANGLE
    Prop<::unicode::EastAsianWidth> { { 0x2221, 0x2222 }, ::unicode::EastAsianWidth::Neutral }, // Sm     [2] MEASURED ANGLE..SPHERICAL ANGLE
    Prop<::unicode::EastAsianWidth> { { 0x2223, 0x2223 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         DIVIDES
    Prop<::unicode::EastAsianWidth> { { 0x2224, 0x2224 }, ::unicode::EastAsianWidth::Neutral }, // Sm         DOES NOT DIVIDE
    Prop<::unicode::EastAsianWidth> { { 0x2225, 0x2225 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         PARALLEL TO
    Prop<::unicode::EastAsianWidth> { { 0x2226, 0x2226 }, ::unicode::EastAsianWidth::Neutral }, // Sm         NOT PARALLEL TO
    Prop<::unicode::EastAsianWidth> { { 0x2227, 0x222C }, ::unicode::EastAsianWidth::Ambiguous }, // Sm     [6] LOGICAL AND..DOUBLE INTEGRAL
    Prop<::unicode::EastAsianWidth> { { 0x222D, 0x222D }, ::unicode::EastAsianWidth::Neutral }, // Sm         TRIPLE INTEGRAL
    Prop<::unicode::EastAsianWidth> { { 0x222E, 0x222E }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         CONTOUR INTEGRAL
    Prop<::unicode::EastAsianWidth> { { 0x222F, 0x2233 }, ::unicode::EastAsianWidth::Neutral }, // Sm     [5] SURFACE INTEGRAL..ANTICLOCKWISE CONTOUR INTEGRAL
    Prop<::unicode::EastAsianWidth> { { 0x2234, 0x2237 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm     [4] THEREFORE..PROPORTION
    Prop<::unicode::EastAsianWidth> { { 0x2238, 0x223B }, ::unicode::EastAsianWidth::Neutral }, // Sm     [4] DOT MINUS..HOMOTHETIC
    Prop<::unicode::EastAsianWidth> { { 0x223C, 0x223D }, ::unicode::EastAsianWidth::Ambiguous }, // Sm     [2] TILDE OPERATOR..REVERSED TILDE
    Prop<::unicode::EastAsianWidth> { { 0x223E, 0x2247 }, ::unicode::EastAsianWidth::Neutral }, // Sm    [10] INVERTED LAZY S..NEITHER APPROXIMATELY NOR ACTUALLY EQUAL TO
    Prop<::unicode::EastAsianWidth> { { 0x2248, 0x2248 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         ALMOST EQUAL TO
    Prop<::unicode::EastAsianWidth> { { 0x2249, 0x224B }, ::unicode::EastAsianWidth::Neutral }, // Sm     [3] NOT ALMOST EQUAL TO..TRIPLE TILDE
    Prop<::unicode::EastAsianWidth> { { 0x224C, 0x224C }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         ALL EQUAL TO
    Prop<::unicode::EastAsianWidth> { { 0x224D, 0x2251 }, ::unicode::EastAsianWidth::Neutral }, // Sm     [5] EQUIVALENT TO..GEOMETRICALLY EQUAL TO
    Prop<::unicode::EastAsianWidth> { { 0x2252, 0x2252 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         APPROXIMATELY EQUAL TO OR THE IMAGE OF
    Prop<::unicode::EastAsianWidth> { { 0x2253, 0x225F }, ::unicode::EastAsianWidth::Neutral }, // Sm    [13] IMAGE OF OR APPROXIMATELY EQUAL TO..QUESTIONED EQUAL TO
    Prop<::unicode::EastAsianWidth> { { 0x2260, 0x2261 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm     [2] NOT EQUAL TO..IDENTICAL TO
    Prop<::unicode::EastAsianWidth> { { 0x2262, 0x2263 }, ::unicode::EastAsianWidth::Neutral }, // Sm     [2] NOT IDENTICAL TO..STRICTLY EQUIVALENT TO
    Prop<::unicode::EastAsianWidth> { { 0x2264, 0x2267 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm     [4] LESS-THAN OR EQUAL TO..GREATER-THAN OVER EQUAL TO
    Prop<::unicode::EastAsianWidth> { { 0x2268, 0x2269 }, ::unicode::EastAsianWidth::Neutral }, // Sm     [2] LESS-THAN BUT NOT EQUAL TO..GREATER-THAN BUT NOT EQUAL TO
    Prop<::unicode::EastAsianWidth> { { 0x226A, 0x226B }, ::unicode::EastAsianWidth::Ambiguous }, // Sm     [2] MUCH LESS-THAN..MUCH GREATER-THAN
    Prop<::unicode::EastAsianWidth> { { 0x226C, 0x226D }, ::unicode::EastAsianWidth::Neutral }, // Sm     [2] BETWEEN..NOT EQUIVALENT TO
    Prop<::unicode::EastAsianWidth> { { 0x226E, 0x226F }, ::unicode::EastAsianWidth::Ambiguous }, // Sm     [2] NOT LESS-THAN..NOT GREATER-THAN
    Prop<::unicode::EastAsianWidth> { { 0x2270, 0x2281 }, ::unicode::EastAsianWidth::Neutral }, // Sm    [18] NEITHER LESS-THAN NOR EQUAL TO..DOES NOT SUCCEED
    Prop<::unicode::EastAsianWidth> { { 0x2282, 0x2283 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm     [2] SUBSET OF..SUPERSET OF
    Prop<::unicode::EastAsianWidth> { { 0x2284, 0x2285 }, ::unicode::EastAsianWidth::Neutral }, // Sm     [2] NOT A SUBSET OF..NOT A SUPERSET OF
    Prop<::unicode::EastAsianWidth> { { 0x2286, 0x2287 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm     [2] SUBSET OF OR EQUAL TO..SUPERSET OF OR EQUAL TO
    Prop<::unicode::EastAsianWidth> { { 0x2288, 0x2294 }, ::unicode::EastAsianWidth::Neutral }, // Sm    [13] NEITHER A SUBSET OF NOR EQUAL TO..SQUARE CUP
    Prop<::unicode::EastAsianWidth> { { 0x2295, 0x2295 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         CIRCLED PLUS
    Prop<::unicode::EastAsianWidth> { { 0x2296, 0x2298 }, ::unicode::EastAsianWidth::Neutral }, // Sm     [3] CIRCLED MINUS..CIRCLED DIVISION SLASH
    Prop<::unicode::EastAsianWidth> { { 0x2299, 0x2299 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         CIRCLED DOT OPERATOR
    Prop<::unicode::EastAsianWidth> { { 0x229A, 0x22A4 }, ::unicode::EastAsianWidth::Neutral }, // Sm    [11] CIRCLED RING OPERATOR..DOWN TACK
    Prop<::unicode::EastAsianWidth> { { 0x22A5, 0x22A5 }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         UP TACK
    Prop<::unicode::EastAsianWidth> { { 0x22A6, 0x22BE }, ::unicode::EastAsianWidth::Neutral }, // Sm    [25] ASSERTION..RIGHT ANGLE WITH ARC
    Prop<::unicode::EastAsianWidth> { { 0x22BF, 0x22BF }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         RIGHT TRIANGLE
    // 0x22C0 .. 0x22FF: Sm    [64] N-ARY LOGICAL AND..Z NOTATION BAG MEMBERSHIP
    // 0x2300 .. 0x2307: So     [8] DIAMETER SIGN..WAVY LINE
    // 0x2308 .. 0x2308: Ps         LEFT CEILING
    // 0x2309 .. 0x2309: Pe         RIGHT CEILING
    // 0x230A .. 0x230A: Ps         LEFT FLOOR
    // 0x230B .. 0x230B: Pe         RIGHT FLOOR
    // 0x230C .. 0x2311: So     [6] BOTTOM RIGHT CROP..SQUARE LOZENGE
    Prop<::unicode::EastAsianWidth> { { 0x22C0, 0x2311 }, ::unicode::EastAsianWidth::Neutral }, // #7
    Prop<::unicode::EastAsianWidth> { { 0x2312, 0x2312 }, ::unicode::EastAsianWidth::Ambiguous }, // So         ARC
    Prop<::unicode::EastAsianWidth> { { 0x2313, 0x2319 }, ::unicode::EastAsianWidth::Neutral }, // So     [7] SEGMENT..TURNED NOT SIGN
    Prop<::unicode::EastAsianWidth> { { 0x231A, 0x231B }, ::unicode::EastAsianWidth::Wide }, // So     [2] WATCH..HOURGLASS
    // 0x231C .. 0x231F: So     [4] TOP LEFT CORNER..BOTTOM RIGHT CORNER
    // 0x2320 .. 0x2321: Sm     [2] TOP HALF INTEGRAL..BOTTOM HALF INTEGRAL
    // 0x2322 .. 0x2328: So     [7] FROWN..KEYBOARD
    Prop<::unicode::EastAsianWidth> { { 0x231C, 0x2328 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x2329 .. 0x2329: Ps         LEFT-POINTING ANGLE BRACKET
    // 0x232A .. 0x232A: Pe         RIGHT-POINTING ANGLE BRACKET
    Prop<::unicode::EastAsianWidth> { { 0x2329, 0x232A }, ::unicode::EastAsianWidth::Wide }, // #2
    // 0x232B .. 0x237B: So    [81] ERASE TO THE LEFT..NOT CHECK MARK
    // 0x237C .. 0x237C: Sm         RIGHT ANGLE WITH DOWNWARDS ZIGZAG ARROW
    // 0x237D .. 0x239A: So    [30] SHOULDERED OPEN BOX..CLEAR SCREEN SYMBOL
    // 0x239B .. 0x23B3: Sm    [25] LEFT PARENTHESIS UPPER HOOK..SUMMATION BOTTOM
    // 0x23B4 .. 0x23DB: So    [40] TOP SQUARE BRACKET..FUSE
    // 0x23DC .. 0x23E1: Sm     [6] TOP PARENTHESIS..BOTTOM TORTOISE SHELL BRACKET
    // 0x23E2 .. 0x23E8: So     [7] WHITE TRAPEZIUM..DECIMAL EXPONENT SYMBOL
    Prop<::unicode::EastAsianWidth> { { 0x232B, 0x23E8 }, ::unicode::EastAsianWidth::Neutral }, // #7
    Prop<::unicode::EastAsianWidth> { { 0x23E9, 0x23EC }, ::unicode::EastAsianWidth::Wide }, // So     [4] BLACK RIGHT-POINTING DOUBLE TRIANGLE..BLACK DOWN-POINTING DOUBLE TRIANGLE
    Prop<::unicode::EastAsianWidth> { { 0x23ED, 0x23EF }, ::unicode::EastAsianWidth::Neutral }, // So     [3] BLACK RIGHT-POINTING DOUBLE TRIANGLE WITH VERTICAL BAR..BLACK RIGHT-POINTING TRIANGLE WITH DOUBLE VERTICAL BAR
    Prop<::unicode::EastAsianWidth> { { 0x23F0, 0x23F0 }, ::unicode::EastAsianWidth::Wide }, // So         ALARM CLOCK
    Prop<::unicode::EastAsianWidth> { { 0x23F1, 0x23F2 }, ::unicode::EastAsianWidth::Neutral }, // So     [2] STOPWATCH..TIMER CLOCK
    Prop<::unicode::EastAsianWidth> { { 0x23F3, 0x23F3 }, ::unicode::EastAsianWidth::Wide }, // So         HOURGLASS WITH FLOWING SAND
    // 0x23F4 .. 0x23FF: So    [12] BLACK MEDIUM LEFT-POINTING TRIANGLE..OBSERVER EYE SYMBOL
    // 0x2400 .. 0x2429: So    [42] SYMBOL FOR NULL..SYMBOL FOR DELETE MEDIUM SHADE FORM
    Prop<::unicode::EastAsianWidth> { { 0x23F4, 0x2429 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x2440, 0x244A }, ::unicode::EastAsianWidth::Neutral }, // So    [11] OCR HOOK..OCR DOUBLE BACKSLASH
    // 0x2460 .. 0x249B: No    [60] CIRCLED DIGIT ONE..NUMBER TWENTY FULL STOP
    // 0x249C .. 0x24E9: So    [78] PARENTHESIZED LATIN SMALL LETTER A..CIRCLED LATIN SMALL LETTER Z
    Prop<::unicode::EastAsianWidth> { { 0x2460, 0x24E9 }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x24EA, 0x24EA }, ::unicode::EastAsianWidth::Neutral }, // No         CIRCLED DIGIT ZERO
    // 0x24EB .. 0x24FF: No    [21] NEGATIVE CIRCLED NUMBER ELEVEN..NEGATIVE CIRCLED DIGIT ZERO
    // 0x2500 .. 0x254B: So    [76] BOX DRAWINGS LIGHT HORIZONTAL..BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL
    Prop<::unicode::EastAsianWidth> { { 0x24EB, 0x254B }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x254C, 0x254F }, ::unicode::EastAsianWidth::Neutral }, // So     [4] BOX DRAWINGS LIGHT DOUBLE DASH HORIZONTAL..BOX DRAWINGS HEAVY DOUBLE DASH VERTICAL
    Prop<::unicode::EastAsianWidth> { { 0x2550, 0x2573 }, ::unicode::EastAsianWidth::Ambiguous }, // So    [36] BOX DRAWINGS DOUBLE HORIZONTAL..BOX DRAWINGS LIGHT DIAGONAL CROSS
    Prop<::unicode::EastAsianWidth> { { 0x2574, 0x257F }, ::unicode::EastAsianWidth::Neutral }, // So    [12] BOX DRAWINGS LIGHT LEFT..BOX DRAWINGS HEAVY UP AND LIGHT DOWN
    Prop<::unicode::EastAsianWidth> { { 0x2580, 0x258F }, ::unicode::EastAsianWidth::Ambiguous }, // So    [16] UPPER HALF BLOCK..LEFT ONE EIGHTH BLOCK
    Prop<::unicode::EastAsianWidth> { { 0x2590, 0x2591 }, ::unicode::EastAsianWidth::Neutral }, // So     [2] RIGHT HALF BLOCK..LIGHT SHADE
    Prop<::unicode::EastAsianWidth> { { 0x2592, 0x2595 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [4] MEDIUM SHADE..RIGHT ONE EIGHTH BLOCK
    Prop<::unicode::EastAsianWidth> { { 0x2596, 0x259F }, ::unicode::EastAsianWidth::Neutral }, // So    [10] QUADRANT LOWER LEFT..QUADRANT UPPER RIGHT AND LOWER LEFT AND LOWER RIGHT
    Prop<::unicode::EastAsianWidth> { { 0x25A0, 0x25A1 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] BLACK SQUARE..WHITE SQUARE
    Prop<::unicode::EastAsianWidth> { { 0x25A2, 0x25A2 }, ::unicode::EastAsianWidth::Neutral }, // So         WHITE SQUARE WITH ROUNDED CORNERS
    Prop<::unicode::EastAsianWidth> { { 0x25A3, 0x25A9 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [7] WHITE SQUARE CONTAINING BLACK SMALL SQUARE..SQUARE WITH DIAGONAL CROSSHATCH FILL
    Prop<::unicode::EastAsianWidth> { { 0x25AA, 0x25B1 }, ::unicode::EastAsianWidth::Neutral }, // So     [8] BLACK SMALL SQUARE..WHITE PARALLELOGRAM
    Prop<::unicode::EastAsianWidth> { { 0x25B2, 0x25B3 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] BLACK UP-POINTING TRIANGLE..WHITE UP-POINTING TRIANGLE
    Prop<::unicode::EastAsianWidth> { { 0x25B4, 0x25B5 }, ::unicode::EastAsianWidth::Neutral }, // So     [2] BLACK UP-POINTING SMALL TRIANGLE..WHITE UP-POINTING SMALL TRIANGLE
    // 0x25B6 .. 0x25B6: So         BLACK RIGHT-POINTING TRIANGLE
    // 0x25B7 .. 0x25B7: Sm         WHITE RIGHT-POINTING TRIANGLE
    Prop<::unicode::EastAsianWidth> { { 0x25B6, 0x25B7 }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x25B8, 0x25BB }, ::unicode::EastAsianWidth::Neutral }, // So     [4] BLACK RIGHT-POINTING SMALL TRIANGLE..WHITE RIGHT-POINTING POINTER
    Prop<::unicode::EastAsianWidth> { { 0x25BC, 0x25BD }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] BLACK DOWN-POINTING TRIANGLE..WHITE DOWN-POINTING TRIANGLE
    Prop<::unicode::EastAsianWidth> { { 0x25BE, 0x25BF }, ::unicode::EastAsianWidth::Neutral }, // So     [2] BLACK DOWN-POINTING SMALL TRIANGLE..WHITE DOWN-POINTING SMALL TRIANGLE
    // 0x25C0 .. 0x25C0: So         BLACK LEFT-POINTING TRIANGLE
    // 0x25C1 .. 0x25C1: Sm         WHITE LEFT-POINTING TRIANGLE
    Prop<::unicode::EastAsianWidth> { { 0x25C0, 0x25C1 }, ::unicode::EastAsianWidth::Ambiguous }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x25C2, 0x25C5 }, ::unicode::EastAsianWidth::Neutral }, // So     [4] BLACK LEFT-POINTING SMALL TRIANGLE..WHITE LEFT-POINTING POINTER
    Prop<::unicode::EastAsianWidth> { { 0x25C6, 0x25C8 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [3] BLACK DIAMOND..WHITE DIAMOND CONTAINING BLACK SMALL DIAMOND
    Prop<::unicode::EastAsianWidth> { { 0x25C9, 0x25CA }, ::unicode::EastAsianWidth::Neutral }, // So     [2] FISHEYE..LOZENGE
    Prop<::unicode::EastAsianWidth> { { 0x25CB, 0x25CB }, ::unicode::EastAsianWidth::Ambiguous }, // So         WHITE CIRCLE
    Prop<::unicode::EastAsianWidth> { { 0x25CC, 0x25CD }, ::unicode::EastAsianWidth::Neutral }, // So     [2] DOTTED CIRCLE..CIRCLE WITH VERTICAL FILL
    Prop<::unicode::EastAsianWidth> { { 0x25CE, 0x25D1 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [4] BULLSEYE..CIRCLE WITH RIGHT HALF BLACK
    Prop<::unicode::EastAsianWidth> { { 0x25D2, 0x25E1 }, ::unicode::EastAsianWidth::Neutral }, // So    [16] CIRCLE WITH LOWER HALF BLACK..LOWER HALF CIRCLE
    Prop<::unicode::EastAsianWidth> { { 0x25E2, 0x25E5 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [4] BLACK LOWER RIGHT TRIANGLE..BLACK UPPER RIGHT TRIANGLE
    Prop<::unicode::EastAsianWidth> { { 0x25E6, 0x25EE }, ::unicode::EastAsianWidth::Neutral }, // So     [9] WHITE BULLET..UP-POINTING TRIANGLE WITH RIGHT HALF BLACK
    Prop<::unicode::EastAsianWidth> { { 0x25EF, 0x25EF }, ::unicode::EastAsianWidth::Ambiguous }, // So         LARGE CIRCLE
    // 0x25F0 .. 0x25F7: So     [8] WHITE SQUARE WITH UPPER LEFT QUADRANT..WHITE CIRCLE WITH UPPER RIGHT QUADRANT
    // 0x25F8 .. 0x25FC: Sm     [5] UPPER LEFT TRIANGLE..BLACK MEDIUM SQUARE
    Prop<::unicode::EastAsianWidth> { { 0x25F0, 0x25FC }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x25FD, 0x25FE }, ::unicode::EastAsianWidth::Wide }, // Sm     [2] WHITE MEDIUM SMALL SQUARE..BLACK MEDIUM SMALL SQUARE
    // 0x25FF .. 0x25FF: Sm         LOWER RIGHT TRIANGLE
    // 0x2600 .. 0x2604: So     [5] BLACK SUN WITH RAYS..COMET
    Prop<::unicode::EastAsianWidth> { { 0x25FF, 0x2604 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x2605, 0x2606 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] BLACK STAR..WHITE STAR
    Prop<::unicode::EastAsianWidth> { { 0x2607, 0x2608 }, ::unicode::EastAsianWidth::Neutral }, // So     [2] LIGHTNING..THUNDERSTORM
    Prop<::unicode::EastAsianWidth> { { 0x2609, 0x2609 }, ::unicode::EastAsianWidth::Ambiguous }, // So         SUN
    Prop<::unicode::EastAsianWidth> { { 0x260A, 0x260D }, ::unicode::EastAsianWidth::Neutral }, // So     [4] ASCENDING NODE..OPPOSITION
    Prop<::unicode::EastAsianWidth> { { 0x260E, 0x260F }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] BLACK TELEPHONE..WHITE TELEPHONE
    Prop<::unicode::EastAsianWidth> { { 0x2610, 0x2613 }, ::unicode::EastAsianWidth::Neutral }, // So     [4] BALLOT BOX..SALTIRE
    Prop<::unicode::EastAsianWidth> { { 0x2614, 0x2615 }, ::unicode::EastAsianWidth::Wide }, // So     [2] UMBRELLA WITH RAIN DROPS..HOT BEVERAGE
    Prop<::unicode::EastAsianWidth> { { 0x2616, 0x261B }, ::unicode::EastAsianWidth::Neutral }, // So     [6] WHITE SHOGI PIECE..BLACK RIGHT POINTING INDEX
    Prop<::unicode::EastAsianWidth> { { 0x261C, 0x261C }, ::unicode::EastAsianWidth::Ambiguous }, // So         WHITE LEFT POINTING INDEX
    Prop<::unicode::EastAsianWidth> { { 0x261D, 0x261D }, ::unicode::EastAsianWidth::Neutral }, // So         WHITE UP POINTING INDEX
    Prop<::unicode::EastAsianWidth> { { 0x261E, 0x261E }, ::unicode::EastAsianWidth::Ambiguous }, // So         WHITE RIGHT POINTING INDEX
    Prop<::unicode::EastAsianWidth> { { 0x261F, 0x262F }, ::unicode::EastAsianWidth::Neutral }, // So    [17] WHITE DOWN POINTING INDEX..YIN YANG
    Prop<::unicode::EastAsianWidth> { { 0x2630, 0x2637 }, ::unicode::EastAsianWidth::Wide }, // So     [8] TRIGRAM FOR HEAVEN..TRIGRAM FOR EARTH
    Prop<::unicode::EastAsianWidth> { { 0x2638, 0x263F }, ::unicode::EastAsianWidth::Neutral }, // So     [8] WHEEL OF DHARMA..MERCURY
    Prop<::unicode::EastAsianWidth> { { 0x2640, 0x2640 }, ::unicode::EastAsianWidth::Ambiguous }, // So         FEMALE SIGN
    Prop<::unicode::EastAsianWidth> { { 0x2641, 0x2641 }, ::unicode::EastAsianWidth::Neutral }, // So         EARTH
    Prop<::unicode::EastAsianWidth> { { 0x2642, 0x2642 }, ::unicode::EastAsianWidth::Ambiguous }, // So         MALE SIGN
    Prop<::unicode::EastAsianWidth> { { 0x2643, 0x2647 }, ::unicode::EastAsianWidth::Neutral }, // So     [5] JUPITER..PLUTO
    Prop<::unicode::EastAsianWidth> { { 0x2648, 0x2653 }, ::unicode::EastAsianWidth::Wide }, // So    [12] ARIES..PISCES
    Prop<::unicode::EastAsianWidth> { { 0x2654, 0x265F }, ::unicode::EastAsianWidth::Neutral }, // So    [12] WHITE CHESS KING..BLACK CHESS PAWN
    Prop<::unicode::EastAsianWidth> { { 0x2660, 0x2661 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] BLACK SPADE SUIT..WHITE HEART SUIT
    Prop<::unicode::EastAsianWidth> { { 0x2662, 0x2662 }, ::unicode::EastAsianWidth::Neutral }, // So         WHITE DIAMOND SUIT
    Prop<::unicode::EastAsianWidth> { { 0x2663, 0x2665 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [3] BLACK CLUB SUIT..BLACK HEART SUIT
    Prop<::unicode::EastAsianWidth> { { 0x2666, 0x2666 }, ::unicode::EastAsianWidth::Neutral }, // So         BLACK DIAMOND SUIT
    Prop<::unicode::EastAsianWidth> { { 0x2667, 0x266A }, ::unicode::EastAsianWidth::Ambiguous }, // So     [4] WHITE CLUB SUIT..EIGHTH NOTE
    Prop<::unicode::EastAsianWidth> { { 0x266B, 0x266B }, ::unicode::EastAsianWidth::Neutral }, // So         BEAMED EIGHTH NOTES
    Prop<::unicode::EastAsianWidth> { { 0x266C, 0x266D }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] BEAMED SIXTEENTH NOTES..MUSIC FLAT SIGN
    Prop<::unicode::EastAsianWidth> { { 0x266E, 0x266E }, ::unicode::EastAsianWidth::Neutral }, // So         MUSIC NATURAL SIGN
    Prop<::unicode::EastAsianWidth> { { 0x266F, 0x266F }, ::unicode::EastAsianWidth::Ambiguous }, // Sm         MUSIC SHARP SIGN
    Prop<::unicode::EastAsianWidth> { { 0x2670, 0x267E }, ::unicode::EastAsianWidth::Neutral }, // So    [15] WEST SYRIAC CROSS..PERMANENT PAPER SIGN
    Prop<::unicode::EastAsianWidth> { { 0x267F, 0x267F }, ::unicode::EastAsianWidth::Wide }, // So         WHEELCHAIR SYMBOL
    Prop<::unicode::EastAsianWidth> { { 0x2680, 0x2689 }, ::unicode::EastAsianWidth::Neutral }, // So    [10] DIE FACE-1..BLACK CIRCLE WITH TWO WHITE DOTS
    Prop<::unicode::EastAsianWidth> { { 0x268A, 0x268F }, ::unicode::EastAsianWidth::Wide }, // So     [6] MONOGRAM FOR YANG..DIGRAM FOR GREATER YIN
    Prop<::unicode::EastAsianWidth> { { 0x2690, 0x2692 }, ::unicode::EastAsianWidth::Neutral }, // So     [3] WHITE FLAG..HAMMER AND PICK
    Prop<::unicode::EastAsianWidth> { { 0x2693, 0x2693 }, ::unicode::EastAsianWidth::Wide }, // So         ANCHOR
    Prop<::unicode::EastAsianWidth> { { 0x2694, 0x269D }, ::unicode::EastAsianWidth::Neutral }, // So    [10] CROSSED SWORDS..OUTLINED WHITE STAR
    Prop<::unicode::EastAsianWidth> { { 0x269E, 0x269F }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] THREE LINES CONVERGING RIGHT..THREE LINES CONVERGING LEFT
    Prop<::unicode::EastAsianWidth> { { 0x26A0, 0x26A0 }, ::unicode::EastAsianWidth::Neutral }, // So         WARNING SIGN
    Prop<::unicode::EastAsianWidth> { { 0x26A1, 0x26A1 }, ::unicode::EastAsianWidth::Wide }, // So         HIGH VOLTAGE SIGN
    Prop<::unicode::EastAsianWidth> { { 0x26A2, 0x26A9 }, ::unicode::EastAsianWidth::Neutral }, // So     [8] DOUBLED FEMALE SIGN..HORIZONTAL MALE WITH STROKE SIGN
    Prop<::unicode::EastAsianWidth> { { 0x26AA, 0x26AB }, ::unicode::EastAsianWidth::Wide }, // So     [2] MEDIUM WHITE CIRCLE..MEDIUM BLACK CIRCLE
    Prop<::unicode::EastAsianWidth> { { 0x26AC, 0x26BC }, ::unicode::EastAsianWidth::Neutral }, // So    [17] MEDIUM SMALL WHITE CIRCLE..SESQUIQUADRATE
    Prop<::unicode::EastAsianWidth> { { 0x26BD, 0x26BE }, ::unicode::EastAsianWidth::Wide }, // So     [2] SOCCER BALL..BASEBALL
    Prop<::unicode::EastAsianWidth> { { 0x26BF, 0x26BF }, ::unicode::EastAsianWidth::Ambiguous }, // So         SQUARED KEY
    Prop<::unicode::EastAsianWidth> { { 0x26C0, 0x26C3 }, ::unicode::EastAsianWidth::Neutral }, // So     [4] WHITE DRAUGHTS MAN..BLACK DRAUGHTS KING
    Prop<::unicode::EastAsianWidth> { { 0x26C4, 0x26C5 }, ::unicode::EastAsianWidth::Wide }, // So     [2] SNOWMAN WITHOUT SNOW..SUN BEHIND CLOUD
    Prop<::unicode::EastAsianWidth> { { 0x26C6, 0x26CD }, ::unicode::EastAsianWidth::Ambiguous }, // So     [8] RAIN..DISABLED CAR
    Prop<::unicode::EastAsianWidth> { { 0x26CE, 0x26CE }, ::unicode::EastAsianWidth::Wide }, // So         OPHIUCHUS
    Prop<::unicode::EastAsianWidth> { { 0x26CF, 0x26D3 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [5] PICK..CHAINS
    Prop<::unicode::EastAsianWidth> { { 0x26D4, 0x26D4 }, ::unicode::EastAsianWidth::Wide }, // So         NO ENTRY
    Prop<::unicode::EastAsianWidth> { { 0x26D5, 0x26E1 }, ::unicode::EastAsianWidth::Ambiguous }, // So    [13] ALTERNATE ONE-WAY LEFT WAY TRAFFIC..RESTRICTED LEFT ENTRY-2
    Prop<::unicode::EastAsianWidth> { { 0x26E2, 0x26E2 }, ::unicode::EastAsianWidth::Neutral }, // So         ASTRONOMICAL SYMBOL FOR URANUS
    Prop<::unicode::EastAsianWidth> { { 0x26E3, 0x26E3 }, ::unicode::EastAsianWidth::Ambiguous }, // So         HEAVY CIRCLE WITH STROKE AND TWO DOTS ABOVE
    Prop<::unicode::EastAsianWidth> { { 0x26E4, 0x26E7 }, ::unicode::EastAsianWidth::Neutral }, // So     [4] PENTAGRAM..INVERTED PENTAGRAM
    Prop<::unicode::EastAsianWidth> { { 0x26E8, 0x26E9 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] BLACK CROSS ON SHIELD..SHINTO SHRINE
    Prop<::unicode::EastAsianWidth> { { 0x26EA, 0x26EA }, ::unicode::EastAsianWidth::Wide }, // So         CHURCH
    Prop<::unicode::EastAsianWidth> { { 0x26EB, 0x26F1 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [7] CASTLE..UMBRELLA ON GROUND
    Prop<::unicode::EastAsianWidth> { { 0x26F2, 0x26F3 }, ::unicode::EastAsianWidth::Wide }, // So     [2] FOUNTAIN..FLAG IN HOLE
    Prop<::unicode::EastAsianWidth> { { 0x26F4, 0x26F4 }, ::unicode::EastAsianWidth::Ambiguous }, // So         FERRY
    Prop<::unicode::EastAsianWidth> { { 0x26F5, 0x26F5 }, ::unicode::EastAsianWidth::Wide }, // So         SAILBOAT
    Prop<::unicode::EastAsianWidth> { { 0x26F6, 0x26F9 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [4] SQUARE FOUR CORNERS..PERSON WITH BALL
    Prop<::unicode::EastAsianWidth> { { 0x26FA, 0x26FA }, ::unicode::EastAsianWidth::Wide }, // So         TENT
    Prop<::unicode::EastAsianWidth> { { 0x26FB, 0x26FC }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] JAPANESE BANK SYMBOL..HEADSTONE GRAVEYARD SYMBOL
    Prop<::unicode::EastAsianWidth> { { 0x26FD, 0x26FD }, ::unicode::EastAsianWidth::Wide }, // So         FUEL PUMP
    Prop<::unicode::EastAsianWidth> { { 0x26FE, 0x26FF }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] CUP ON BLACK SQUARE..WHITE FLAG WITH HORIZONTAL MIDDLE BLACK STRIPE
    Prop<::unicode::EastAsianWidth> { { 0x2700, 0x2704 }, ::unicode::EastAsianWidth::Neutral }, // So     [5] BLACK SAFETY SCISSORS..WHITE SCISSORS
    Prop<::unicode::EastAsianWidth> { { 0x2705, 0x2705 }, ::unicode::EastAsianWidth::Wide }, // So         WHITE HEAVY CHECK MARK
    Prop<::unicode::EastAsianWidth> { { 0x2706, 0x2709 }, ::unicode::EastAsianWidth::Neutral }, // So     [4] TELEPHONE LOCATION SIGN..ENVELOPE
    Prop<::unicode::EastAsianWidth> { { 0x270A, 0x270B }, ::unicode::EastAsianWidth::Wide }, // So     [2] RAISED FIST..RAISED HAND
    Prop<::unicode::EastAsianWidth> { { 0x270C, 0x2727 }, ::unicode::EastAsianWidth::Neutral }, // So    [28] VICTORY HAND..WHITE FOUR POINTED STAR
    Prop<::unicode::EastAsianWidth> { { 0x2728, 0x2728 }, ::unicode::EastAsianWidth::Wide }, // So         SPARKLES
    Prop<::unicode::EastAsianWidth> { { 0x2729, 0x273C }, ::unicode::EastAsianWidth::Neutral }, // So    [20] STRESS OUTLINED WHITE STAR..OPEN CENTRE TEARDROP-SPOKED ASTERISK
    Prop<::unicode::EastAsianWidth> { { 0x273D, 0x273D }, ::unicode::EastAsianWidth::Ambiguous }, // So         HEAVY TEARDROP-SPOKED ASTERISK
    Prop<::unicode::EastAsianWidth> { { 0x273E, 0x274B }, ::unicode::EastAsianWidth::Neutral }, // So    [14] SIX PETALLED BLACK AND WHITE FLORETTE..HEAVY EIGHT TEARDROP-SPOKED PROPELLER ASTERISK
    Prop<::unicode::EastAsianWidth> { { 0x274C, 0x274C }, ::unicode::EastAsianWidth::Wide }, // So         CROSS MARK
    Prop<::unicode::EastAsianWidth> { { 0x274D, 0x274D }, ::unicode::EastAsianWidth::Neutral }, // So         SHADOWED WHITE CIRCLE
    Prop<::unicode::EastAsianWidth> { { 0x274E, 0x274E }, ::unicode::EastAsianWidth::Wide }, // So         NEGATIVE SQUARED CROSS MARK
    Prop<::unicode::EastAsianWidth> { { 0x274F, 0x2752 }, ::unicode::EastAsianWidth::Neutral }, // So     [4] LOWER RIGHT DROP-SHADOWED WHITE SQUARE..UPPER RIGHT SHADOWED WHITE SQUARE
    Prop<::unicode::EastAsianWidth> { { 0x2753, 0x2755 }, ::unicode::EastAsianWidth::Wide }, // So     [3] BLACK QUESTION MARK ORNAMENT..WHITE EXCLAMATION MARK ORNAMENT
    Prop<::unicode::EastAsianWidth> { { 0x2756, 0x2756 }, ::unicode::EastAsianWidth::Neutral }, // So         BLACK DIAMOND MINUS WHITE X
    Prop<::unicode::EastAsianWidth> { { 0x2757, 0x2757 }, ::unicode::EastAsianWidth::Wide }, // So         HEAVY EXCLAMATION MARK SYMBOL
    // 0x2758 .. 0x2767: So    [16] LIGHT VERTICAL BAR..ROTATED FLORAL HEART BULLET
    // 0x2768 .. 0x2768: Ps         MEDIUM LEFT PARENTHESIS ORNAMENT
    // 0x2769 .. 0x2769: Pe         MEDIUM RIGHT PARENTHESIS ORNAMENT
    // 0x276A .. 0x276A: Ps         MEDIUM FLATTENED LEFT PARENTHESIS ORNAMENT
    // 0x276B .. 0x276B: Pe         MEDIUM FLATTENED RIGHT PARENTHESIS ORNAMENT
    // 0x276C .. 0x276C: Ps         MEDIUM LEFT-POINTING ANGLE BRACKET ORNAMENT
    // 0x276D .. 0x276D: Pe         MEDIUM RIGHT-POINTING ANGLE BRACKET ORNAMENT
    // 0x276E .. 0x276E: Ps         HEAVY LEFT-POINTING ANGLE QUOTATION MARK ORNAMENT
    // 0x276F .. 0x276F: Pe         HEAVY RIGHT-POINTING ANGLE QUOTATION MARK ORNAMENT
    // 0x2770 .. 0x2770: Ps         HEAVY LEFT-POINTING ANGLE BRACKET ORNAMENT
    // 0x2771 .. 0x2771: Pe         HEAVY RIGHT-POINTING ANGLE BRACKET ORNAMENT
    // 0x2772 .. 0x2772: Ps         LIGHT LEFT TORTOISE SHELL BRACKET ORNAMENT
    // 0x2773 .. 0x2773: Pe         LIGHT RIGHT TORTOISE SHELL BRACKET ORNAMENT
    // 0x2774 .. 0x2774: Ps         MEDIUM LEFT CURLY BRACKET ORNAMENT
    // 0x2775 .. 0x2775: Pe         MEDIUM RIGHT CURLY BRACKET ORNAMENT
    Prop<::unicode::EastAsianWidth> { { 0x2758, 0x2775 }, ::unicode::EastAsianWidth::Neutral }, // #15
    Prop<::unicode::EastAsianWidth> { { 0x2776, 0x277F }, ::unicode::EastAsianWidth::Ambiguous }, // No    [10] DINGBAT NEGATIVE CIRCLED DIGIT ONE..DINGBAT NEGATIVE CIRCLED NUMBER TEN
    // 0x2780 .. 0x2793: No    [20] DINGBAT CIRCLED SANS-SERIF DIGIT ONE..DINGBAT NEGATIVE CIRCLED SANS-SERIF NUMBER TEN
    // 0x2794 .. 0x2794: So         HEAVY WIDE-HEADED RIGHTWARDS ARROW
    Prop<::unicode::EastAsianWidth> { { 0x2780, 0x2794 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x2795, 0x2797 }, ::unicode::EastAsianWidth::Wide }, // So     [3] HEAVY PLUS SIGN..HEAVY DIVISION SIGN
    Prop<::unicode::EastAsianWidth> { { 0x2798, 0x27AF }, ::unicode::EastAsianWidth::Neutral }, // So    [24] HEAVY SOUTH EAST ARROW..NOTCHED LOWER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW
    Prop<::unicode::EastAsianWidth> { { 0x27B0, 0x27B0 }, ::unicode::EastAsianWidth::Wide }, // So         CURLY LOOP
    Prop<::unicode::EastAsianWidth> { { 0x27B1, 0x27BE }, ::unicode::EastAsianWidth::Neutral }, // So    [14] NOTCHED UPPER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW..OPEN-OUTLINED RIGHTWARDS ARROW
    Prop<::unicode::EastAsianWidth> { { 0x27BF, 0x27BF }, ::unicode::EastAsianWidth::Wide }, // So         DOUBLE CURLY LOOP
    // 0x27C0 .. 0x27C4: Sm     [5] THREE DIMENSIONAL ANGLE..OPEN SUPERSET
    // 0x27C5 .. 0x27C5: Ps         LEFT S-SHAPED BAG DELIMITER
    // 0x27C6 .. 0x27C6: Pe         RIGHT S-SHAPED BAG DELIMITER
    // 0x27C7 .. 0x27E5: Sm    [31] OR WITH DOT INSIDE..WHITE SQUARE WITH RIGHTWARDS TICK
    Prop<::unicode::EastAsianWidth> { { 0x27C0, 0x27E5 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x27E6 .. 0x27E6: Ps         MATHEMATICAL LEFT WHITE SQUARE BRACKET
    // 0x27E7 .. 0x27E7: Pe         MATHEMATICAL RIGHT WHITE SQUARE BRACKET
    // 0x27E8 .. 0x27E8: Ps         MATHEMATICAL LEFT ANGLE BRACKET
    // 0x27E9 .. 0x27E9: Pe         MATHEMATICAL RIGHT ANGLE BRACKET
    // 0x27EA .. 0x27EA: Ps         MATHEMATICAL LEFT DOUBLE ANGLE BRACKET
    // 0x27EB .. 0x27EB: Pe         MATHEMATICAL RIGHT DOUBLE ANGLE BRACKET
    // 0x27EC .. 0x27EC: Ps         MATHEMATICAL LEFT WHITE TORTOISE SHELL BRACKET
    // 0x27ED .. 0x27ED: Pe         MATHEMATICAL RIGHT WHITE TORTOISE SHELL BRACKET
    Prop<::unicode::EastAsianWidth> { { 0x27E6, 0x27ED }, ::unicode::EastAsianWidth::Narrow }, // #8
    // 0x27EE .. 0x27EE: Ps         MATHEMATICAL LEFT FLATTENED PARENTHESIS
    // 0x27EF .. 0x27EF: Pe         MATHEMATICAL RIGHT FLATTENED PARENTHESIS
    // 0x27F0 .. 0x27FF: Sm    [16] UPWARDS QUADRUPLE ARROW..LONG RIGHTWARDS SQUIGGLE ARROW
    // 0x2800 .. 0x28FF: So   [256] BRAILLE PATTERN BLANK..BRAILLE PATTERN DOTS-12345678
    // 0x2900 .. 0x297F: Sm   [128] RIGHTWARDS TWO-HEADED ARROW WITH VERTICAL STROKE..DOWN FISH TAIL
    // 0x2980 .. 0x2982: Sm     [3] TRIPLE VERTICAL BAR DELIMITER..Z NOTATION TYPE COLON
    // 0x2983 .. 0x2983: Ps         LEFT WHITE CURLY BRACKET
    // 0x2984 .. 0x2984: Pe         RIGHT WHITE CURLY BRACKET
    Prop<::unicode::EastAsianWidth> { { 0x27EE, 0x2984 }, ::unicode::EastAsianWidth::Neutral }, // #8
    // 0x2985 .. 0x2985: Ps         LEFT WHITE PARENTHESIS
    // 0x2986 .. 0x2986: Pe         RIGHT WHITE PARENTHESIS
    Prop<::unicode::EastAsianWidth> { { 0x2985, 0x2986 }, ::unicode::EastAsianWidth::Narrow }, // #2
    // 0x2987 .. 0x2987: Ps         Z NOTATION LEFT IMAGE BRACKET
    // 0x2988 .. 0x2988: Pe         Z NOTATION RIGHT IMAGE BRACKET
    // 0x2989 .. 0x2989: Ps         Z NOTATION LEFT BINDING BRACKET
    // 0x298A .. 0x298A: Pe         Z NOTATION RIGHT BINDING BRACKET
    // 0x298B .. 0x298B: Ps         LEFT SQUARE BRACKET WITH UNDERBAR
    // 0x298C .. 0x298C: Pe         RIGHT SQUARE BRACKET WITH UNDERBAR
    // 0x298D .. 0x298D: Ps         LEFT SQUARE BRACKET WITH TICK IN TOP CORNER
    // 0x298E .. 0x298E: Pe         RIGHT SQUARE BRACKET WITH TICK IN BOTTOM CORNER
    // 0x298F .. 0x298F: Ps         LEFT SQUARE BRACKET WITH TICK IN BOTTOM CORNER
    // 0x2990 .. 0x2990: Pe         RIGHT SQUARE BRACKET WITH TICK IN TOP CORNER
    // 0x2991 .. 0x2991: Ps         LEFT ANGLE BRACKET WITH DOT
    // 0x2992 .. 0x2992: Pe         RIGHT ANGLE BRACKET WITH DOT
    // 0x2993 .. 0x2993: Ps         LEFT ARC LESS-THAN BRACKET
    // 0x2994 .. 0x2994: Pe         RIGHT ARC GREATER-THAN BRACKET
    // 0x2995 .. 0x2995: Ps         DOUBLE LEFT ARC GREATER-THAN BRACKET
    // 0x2996 .. 0x2996: Pe         DOUBLE RIGHT ARC LESS-THAN BRACKET
    // 0x2997 .. 0x2997: Ps         LEFT BLACK TORTOISE SHELL BRACKET
    // 0x2998 .. 0x2998: Pe         RIGHT BLACK TORTOISE SHELL BRACKET
    // 0x2999 .. 0x29D7: Sm    [63] DOTTED FENCE..BLACK HOURGLASS
    // 0x29D8 .. 0x29D8: Ps         LEFT WIGGLY FENCE
    // 0x29D9 .. 0x29D9: Pe         RIGHT WIGGLY FENCE
    // 0x29DA .. 0x29DA: Ps         LEFT DOUBLE WIGGLY FENCE
    // 0x29DB .. 0x29DB: Pe         RIGHT DOUBLE WIGGLY FENCE
    // 0x29DC .. 0x29FB: Sm    [32] INCOMPLETE INFINITY..TRIPLE PLUS
    // 0x29FC .. 0x29FC: Ps         LEFT-POINTING CURVED ANGLE BRACKET
    // 0x29FD .. 0x29FD: Pe         RIGHT-POINTING CURVED ANGLE BRACKET
    // 0x29FE .. 0x29FF: Sm     [2] TINY..MINY
    // 0x2A00 .. 0x2AFF: Sm   [256] N-ARY CIRCLED DOT OPERATOR..N-ARY WHITE VERTICAL BAR
    // 0x2B00 .. 0x2B1A: So    [27] NORTH EAST WHITE ARROW..DOTTED SQUARE
    Prop<::unicode::EastAsianWidth> { { 0x2987, 0x2B1A }, ::unicode::EastAsianWidth::Neutral }, // #29
    Prop<::unicode::EastAsianWidth> { { 0x2B1B, 0x2B1C }, ::unicode::EastAsianWidth::Wide }, // So     [2] BLACK LARGE SQUARE..WHITE LARGE SQUARE
    // 0x2B1D .. 0x2B2F: So    [19] BLACK VERY SMALL SQUARE..WHITE VERTICAL ELLIPSE
    // 0x2B30 .. 0x2B44: Sm    [21] LEFT ARROW WITH SMALL CIRCLE..RIGHTWARDS ARROW THROUGH SUPERSET
    // 0x2B45 .. 0x2B46: So     [2] LEFTWARDS QUADRUPLE ARROW..RIGHTWARDS QUADRUPLE ARROW
    // 0x2B47 .. 0x2B4C: Sm     [6] REVERSE TILDE OPERATOR ABOVE RIGHTWARDS ARROW..RIGHTWARDS ARROW ABOVE REVERSE TILDE OPERATOR
    // 0x2B4D .. 0x2B4F: So     [3] DOWNWARDS TRIANGLE-HEADED ZIGZAG ARROW..SHORT BACKSLANTED SOUTH ARROW
    Prop<::unicode::EastAsianWidth> { { 0x2B1D, 0x2B4F }, ::unicode::EastAsianWidth::Neutral }, // #5
    Prop<::unicode::EastAsianWidth> { { 0x2B50, 0x2B50 }, ::unicode::EastAsianWidth::Wide }, // So         WHITE MEDIUM STAR
    Prop<::unicode::EastAsianWidth> { { 0x2B51, 0x2B54 }, ::unicode::EastAsianWidth::Neutral }, // So     [4] BLACK SMALL STAR..WHITE RIGHT-POINTING PENTAGON
    Prop<::unicode::EastAsianWidth> { { 0x2B55, 0x2B55 }, ::unicode::EastAsianWidth::Wide }, // So         HEAVY LARGE CIRCLE
    Prop<::unicode::EastAsianWidth> { { 0x2B56, 0x2B59 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [4] HEAVY OVAL WITH OVAL INSIDE..HEAVY CIRCLED SALTIRE
    Prop<::unicode::EastAsianWidth> { { 0x2B5A, 0x2B73 }, ::unicode::EastAsianWidth::Neutral }, // So    [26] SLANTED NORTH ARROW WITH HOOKED HEAD..DOWNWARDS TRIANGLE-HEADED ARROW TO BAR
    // 0x2B76 .. 0x2BFF: So   [138] NORTH WEST TRIANGLE-HEADED ARROW TO BAR..HELLSCHREIBER PAUSE SYMBOL
    // 0x2C00 .. 0x2C5F: L&    [96] GLAGOLITIC CAPITAL LETTER AZU..GLAGOLITIC SMALL LETTER CAUDATE CHRIVI
    // 0x2C60 .. 0x2C7B: L&    [28] LATIN CAPITAL LETTER L WITH DOUBLE BAR..LATIN LETTER SMALL CAPITAL TURNED E
    // 0x2C7C .. 0x2C7D: Lm     [2] LATIN SUBSCRIPT SMALL LETTER J..MODIFIER LETTER CAPITAL V
    // 0x2C7E .. 0x2C7F: Lu     [2] LATIN CAPITAL LETTER S WITH SWASH TAIL..LATIN CAPITAL LETTER Z WITH SWASH TAIL
    // 0x2C80 .. 0x2CE4: L&   [101] COPTIC CAPITAL LETTER ALFA..COPTIC SYMBOL KAI
    // 0x2CE5 .. 0x2CEA: So     [6] COPTIC SYMBOL MI RO..COPTIC SYMBOL SHIMA SIMA
    // 0x2CEB .. 0x2CEE: L&     [4] COPTIC CAPITAL LETTER CRYPTOGRAMMIC SHEI..COPTIC SMALL LETTER CRYPTOGRAMMIC GANGIA
    // 0x2CEF .. 0x2CF1: Mn     [3] COPTIC COMBINING NI ABOVE..COPTIC COMBINING SPIRITUS LENIS
    // 0x2CF2 .. 0x2CF3: L&     [2] COPTIC CAPITAL LETTER BOHAIRIC KHEI..COPTIC SMALL LETTER BOHAIRIC KHEI
    Prop<::unicode::EastAsianWidth> { { 0x2B76, 0x2CF3 }, ::unicode::EastAsianWidth::Neutral }, // #10
    // 0x2CF9 .. 0x2CFC: Po     [4] COPTIC OLD NUBIAN FULL STOP..COPTIC OLD NUBIAN VERSE DIVIDER
    // 0x2CFD .. 0x2CFD: No         COPTIC FRACTION ONE HALF
    // 0x2CFE .. 0x2CFF: Po     [2] COPTIC FULL STOP..COPTIC MORPHOLOGICAL DIVIDER
    // 0x2D00 .. 0x2D25: Ll    [38] GEORGIAN SMALL LETTER AN..GEORGIAN SMALL LETTER HOE
    Prop<::unicode::EastAsianWidth> { { 0x2CF9, 0x2D25 }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x2D27, 0x2D27 }, ::unicode::EastAsianWidth::Neutral }, // Ll         GEORGIAN SMALL LETTER YN
    Prop<::unicode::EastAsianWidth> { { 0x2D2D, 0x2D2D }, ::unicode::EastAsianWidth::Neutral }, // Ll         GEORGIAN SMALL LETTER AEN
    Prop<::unicode::EastAsianWidth> { { 0x2D30, 0x2D67 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [56] TIFINAGH LETTER YA..TIFINAGH LETTER YO
    // 0x2D6F .. 0x2D6F: Lm         TIFINAGH MODIFIER LETTER LABIALIZATION MARK
    // 0x2D70 .. 0x2D70: Po         TIFINAGH SEPARATOR MARK
    Prop<::unicode::EastAsianWidth> { { 0x2D6F, 0x2D70 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x2D7F .. 0x2D7F: Mn         TIFINAGH CONSONANT JOINER
    // 0x2D80 .. 0x2D96: Lo    [23] ETHIOPIC SYLLABLE LOA..ETHIOPIC SYLLABLE GGWE
    Prop<::unicode::EastAsianWidth> { { 0x2D7F, 0x2D96 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x2DA0, 0x2DA6 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE SSA..ETHIOPIC SYLLABLE SSO
    Prop<::unicode::EastAsianWidth> { { 0x2DA8, 0x2DAE }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE CCA..ETHIOPIC SYLLABLE CCO
    Prop<::unicode::EastAsianWidth> { { 0x2DB0, 0x2DB6 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE ZZA..ETHIOPIC SYLLABLE ZZO
    Prop<::unicode::EastAsianWidth> { { 0x2DB8, 0x2DBE }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE CCHA..ETHIOPIC SYLLABLE CCHO
    Prop<::unicode::EastAsianWidth> { { 0x2DC0, 0x2DC6 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE QYA..ETHIOPIC SYLLABLE QYO
    Prop<::unicode::EastAsianWidth> { { 0x2DC8, 0x2DCE }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE KYA..ETHIOPIC SYLLABLE KYO
    Prop<::unicode::EastAsianWidth> { { 0x2DD0, 0x2DD6 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE XYA..ETHIOPIC SYLLABLE XYO
    Prop<::unicode::EastAsianWidth> { { 0x2DD8, 0x2DDE }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE GYA..ETHIOPIC SYLLABLE GYO
    // 0x2DE0 .. 0x2DFF: Mn    [32] COMBINING CYRILLIC LETTER BE..COMBINING CYRILLIC LETTER IOTIFIED BIG YUS
    // 0x2E00 .. 0x2E01: Po     [2] RIGHT ANGLE SUBSTITUTION MARKER..RIGHT ANGLE DOTTED SUBSTITUTION MARKER
    // 0x2E02 .. 0x2E02: Pi         LEFT SUBSTITUTION BRACKET
    // 0x2E03 .. 0x2E03: Pf         RIGHT SUBSTITUTION BRACKET
    // 0x2E04 .. 0x2E04: Pi         LEFT DOTTED SUBSTITUTION BRACKET
    // 0x2E05 .. 0x2E05: Pf         RIGHT DOTTED SUBSTITUTION BRACKET
    // 0x2E06 .. 0x2E08: Po     [3] RAISED INTERPOLATION MARKER..DOTTED TRANSPOSITION MARKER
    // 0x2E09 .. 0x2E09: Pi         LEFT TRANSPOSITION BRACKET
    // 0x2E0A .. 0x2E0A: Pf         RIGHT TRANSPOSITION BRACKET
    // 0x2E0B .. 0x2E0B: Po         RAISED SQUARE
    // 0x2E0C .. 0x2E0C: Pi         LEFT RAISED OMISSION BRACKET
    // 0x2E0D .. 0x2E0D: Pf         RIGHT RAISED OMISSION BRACKET
    // 0x2E0E .. 0x2E16: Po     [9] EDITORIAL CORONIS..DOTTED RIGHT-POINTING ANGLE
    // 0x2E17 .. 0x2E17: Pd         DOUBLE OBLIQUE HYPHEN
    // 0x2E18 .. 0x2E19: Po     [2] INVERTED INTERROBANG..PALM BRANCH
    // 0x2E1A .. 0x2E1A: Pd         HYPHEN WITH DIAERESIS
    // 0x2E1B .. 0x2E1B: Po         TILDE WITH RING ABOVE
    // 0x2E1C .. 0x2E1C: Pi         LEFT LOW PARAPHRASE BRACKET
    // 0x2E1D .. 0x2E1D: Pf         RIGHT LOW PARAPHRASE BRACKET
    // 0x2E1E .. 0x2E1F: Po     [2] TILDE WITH DOT ABOVE..TILDE WITH DOT BELOW
    // 0x2E20 .. 0x2E20: Pi         LEFT VERTICAL BAR WITH QUILL
    // 0x2E21 .. 0x2E21: Pf         RIGHT VERTICAL BAR WITH QUILL
    // 0x2E22 .. 0x2E22: Ps         TOP LEFT HALF BRACKET
    // 0x2E23 .. 0x2E23: Pe         TOP RIGHT HALF BRACKET
    // 0x2E24 .. 0x2E24: Ps         BOTTOM LEFT HALF BRACKET
    // 0x2E25 .. 0x2E25: Pe         BOTTOM RIGHT HALF BRACKET
    // 0x2E26 .. 0x2E26: Ps         LEFT SIDEWAYS U BRACKET
    // 0x2E27 .. 0x2E27: Pe         RIGHT SIDEWAYS U BRACKET
    // 0x2E28 .. 0x2E28: Ps         LEFT DOUBLE PARENTHESIS
    // 0x2E29 .. 0x2E29: Pe         RIGHT DOUBLE PARENTHESIS
    // 0x2E2A .. 0x2E2E: Po     [5] TWO DOTS OVER ONE DOT PUNCTUATION..REVERSED QUESTION MARK
    // 0x2E2F .. 0x2E2F: Lm         VERTICAL TILDE
    // 0x2E30 .. 0x2E39: Po    [10] RING POINT..TOP HALF SECTION SIGN
    // 0x2E3A .. 0x2E3B: Pd     [2] TWO-EM DASH..THREE-EM DASH
    // 0x2E3C .. 0x2E3F: Po     [4] STENOGRAPHIC FULL STOP..CAPITULUM
    // 0x2E40 .. 0x2E40: Pd         DOUBLE HYPHEN
    // 0x2E41 .. 0x2E41: Po         REVERSED COMMA
    // 0x2E42 .. 0x2E42: Ps         DOUBLE LOW-REVERSED-9 QUOTATION MARK
    // 0x2E43 .. 0x2E4F: Po    [13] DASH WITH LEFT UPTURN..CORNISH VERSE DIVIDER
    // 0x2E50 .. 0x2E51: So     [2] CROSS PATTY WITH RIGHT CROSSBAR..CROSS PATTY WITH LEFT CROSSBAR
    // 0x2E52 .. 0x2E54: Po     [3] TIRONIAN SIGN CAPITAL ET..MEDIEVAL QUESTION MARK
    // 0x2E55 .. 0x2E55: Ps         LEFT SQUARE BRACKET WITH STROKE
    // 0x2E56 .. 0x2E56: Pe         RIGHT SQUARE BRACKET WITH STROKE
    // 0x2E57 .. 0x2E57: Ps         LEFT SQUARE BRACKET WITH DOUBLE STROKE
    // 0x2E58 .. 0x2E58: Pe         RIGHT SQUARE BRACKET WITH DOUBLE STROKE
    // 0x2E59 .. 0x2E59: Ps         TOP HALF LEFT PARENTHESIS
    // 0x2E5A .. 0x2E5A: Pe         TOP HALF RIGHT PARENTHESIS
    // 0x2E5B .. 0x2E5B: Ps         BOTTOM HALF LEFT PARENTHESIS
    // 0x2E5C .. 0x2E5C: Pe         BOTTOM HALF RIGHT PARENTHESIS
    // 0x2E5D .. 0x2E5D: Pd         OBLIQUE HYPHEN
    Prop<::unicode::EastAsianWidth> { { 0x2DE0, 0x2E5D }, ::unicode::EastAsianWidth::Neutral }, // #50
    Prop<::unicode::EastAsianWidth> { { 0x2E80, 0x2E99 }, ::unicode::EastAsianWidth::Wide }, // So    [26] CJK RADICAL REPEAT..CJK RADICAL RAP
    Prop<::unicode::EastAsianWidth> { { 0x2E9B, 0x2EF3 }, ::unicode::EastAsianWidth::Wide }, // So    [89] CJK RADICAL CHOKE..CJK RADICAL C-SIMPLIFIED TURTLE
    Prop<::unicode::EastAsianWidth> { { 0x2F00, 0x2FD5 }, ::unicode::EastAsianWidth::Wide }, // So   [214] KANGXI RADICAL ONE..KANGXI RADICAL FLUTE
    Prop<::unicode::EastAsianWidth> { { 0x2FF0, 0x2FFF }, ::unicode::EastAsianWidth::Wide }, // So    [16] IDEOGRAPHIC DESCRIPTION CHARACTER LEFT TO RIGHT..IDEOGRAPHIC DESCRIPTION CHARACTER ROTATION
    Prop<::unicode::EastAsianWidth> { { 0x3000, 0x3000 }, ::unicode::EastAsianWidth::FullWidth }, // Zs         IDEOGRAPHIC SPACE
    // 0x3001 .. 0x3003: Po     [3] IDEOGRAPHIC COMMA..DITTO MARK
    // 0x3004 .. 0x3004: So         JAPANESE INDUSTRIAL STANDARD SYMBOL
    // 0x3005 .. 0x3005: Lm         IDEOGRAPHIC ITERATION MARK
    // 0x3006 .. 0x3006: Lo         IDEOGRAPHIC CLOSING MARK
    // 0x3007 .. 0x3007: Nl         IDEOGRAPHIC NUMBER ZERO
    // 0x3008 .. 0x3008: Ps         LEFT ANGLE BRACKET
    // 0x3009 .. 0x3009: Pe         RIGHT ANGLE BRACKET
    // 0x300A .. 0x300A: Ps         LEFT DOUBLE ANGLE BRACKET
    // 0x300B .. 0x300B: Pe         RIGHT DOUBLE ANGLE BRACKET
    // 0x300C .. 0x300C: Ps         LEFT CORNER BRACKET
    // 0x300D .. 0x300D: Pe         RIGHT CORNER BRACKET
    // 0x300E .. 0x300E: Ps         LEFT WHITE CORNER BRACKET
    // 0x300F .. 0x300F: Pe         RIGHT WHITE CORNER BRACKET
    // 0x3010 .. 0x3010: Ps         LEFT BLACK LENTICULAR BRACKET
    // 0x3011 .. 0x3011: Pe         RIGHT BLACK LENTICULAR BRACKET
    // 0x3012 .. 0x3013: So     [2] POSTAL MARK..GETA MARK
    // 0x3014 .. 0x3014: Ps         LEFT TORTOISE SHELL BRACKET
    // 0x3015 .. 0x3015: Pe         RIGHT TORTOISE SHELL BRACKET
    // 0x3016 .. 0x3016: Ps         LEFT WHITE LENTICULAR BRACKET
    // 0x3017 .. 0x3017: Pe         RIGHT WHITE LENTICULAR BRACKET
    // 0x3018 .. 0x3018: Ps         LEFT WHITE TORTOISE SHELL BRACKET
    // 0x3019 .. 0x3019: Pe         RIGHT WHITE TORTOISE SHELL BRACKET
    // 0x301A .. 0x301A: Ps         LEFT WHITE SQUARE BRACKET
    // 0x301B .. 0x301B: Pe         RIGHT WHITE SQUARE BRACKET
    // 0x301C .. 0x301C: Pd         WAVE DASH
    // 0x301D .. 0x301D: Ps         REVERSED DOUBLE PRIME QUOTATION MARK
    // 0x301E .. 0x301F: Pe     [2] DOUBLE PRIME QUOTATION MARK..LOW DOUBLE PRIME QUOTATION MARK
    // 0x3020 .. 0x3020: So         POSTAL MARK FACE
    // 0x3021 .. 0x3029: Nl     [9] HANGZHOU NUMERAL ONE..HANGZHOU NUMERAL NINE
    // 0x302A .. 0x302D: Mn     [4] IDEOGRAPHIC LEVEL TONE MARK..IDEOGRAPHIC ENTERING TONE MARK
    // 0x302E .. 0x302F: Mc     [2] HANGUL SINGLE DOT TONE MARK..HANGUL DOUBLE DOT TONE MARK
    // 0x3030 .. 0x3030: Pd         WAVY DASH
    // 0x3031 .. 0x3035: Lm     [5] VERTICAL KANA REPEAT MARK..VERTICAL KANA REPEAT MARK LOWER HALF
    // 0x3036 .. 0x3037: So     [2] CIRCLED POSTAL MARK..IDEOGRAPHIC TELEGRAPH LINE FEED SEPARATOR SYMBOL
    // 0x3038 .. 0x303A: Nl     [3] HANGZHOU NUMERAL TEN..HANGZHOU NUMERAL THIRTY
    // 0x303B .. 0x303B: Lm         VERTICAL IDEOGRAPHIC ITERATION MARK
    // 0x303C .. 0x303C: Lo         MASU MARK
    // 0x303D .. 0x303D: Po         PART ALTERNATION MARK
    // 0x303E .. 0x303E: So         IDEOGRAPHIC VARIATION INDICATOR
    Prop<::unicode::EastAsianWidth> { { 0x3001, 0x303E }, ::unicode::EastAsianWidth::Wide }, // #39
    Prop<::unicode::EastAsianWidth> { { 0x303F, 0x303F }, ::unicode::EastAsianWidth::Neutral }, // So         IDEOGRAPHIC HALF FILL SPACE
    Prop<::unicode::EastAsianWidth> { { 0x3041, 0x3096 }, ::unicode::EastAsianWidth::Wide }, // Lo    [86] HIRAGANA LETTER SMALL A..HIRAGANA LETTER SMALL KE
    // 0x3099 .. 0x309A: Mn     [2] COMBINING KATAKANA-HIRAGANA VOICED SOUND MARK..COMBINING KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    // 0x309B .. 0x309C: Sk     [2] KATAKANA-HIRAGANA VOICED SOUND MARK..KATAKANA-HIRAGANA SEMI-VOICED SOUND MARK
    // 0x309D .. 0x309E: Lm     [2] HIRAGANA ITERATION MARK..HIRAGANA VOICED ITERATION MARK
    // 0x309F .. 0x309F: Lo         HIRAGANA DIGRAPH YORI
    // 0x30A0 .. 0x30A0: Pd         KATAKANA-HIRAGANA DOUBLE HYPHEN
    // 0x30A1 .. 0x30FA: Lo    [90] KATAKANA LETTER SMALL A..KATAKANA LETTER VO
    // 0x30FB .. 0x30FB: Po         KATAKANA MIDDLE DOT
    // 0x30FC .. 0x30FE: Lm     [3] KATAKANA-HIRAGANA PROLONGED SOUND MARK..KATAKANA VOICED ITERATION MARK
    // 0x30FF .. 0x30FF: Lo         KATAKANA DIGRAPH KOTO
    Prop<::unicode::EastAsianWidth> { { 0x3099, 0x30FF }, ::unicode::EastAsianWidth::Wide }, // #9
    Prop<::unicode::EastAsianWidth> { { 0x3105, 0x312F }, ::unicode::EastAsianWidth::Wide }, // Lo    [43] BOPOMOFO LETTER B..BOPOMOFO LETTER NN
    Prop<::unicode::EastAsianWidth> { { 0x3131, 0x318E }, ::unicode::EastAsianWidth::Wide }, // Lo    [94] HANGUL LETTER KIYEOK..HANGUL LETTER ARAEAE
    // 0x3190 .. 0x3191: So     [2] IDEOGRAPHIC ANNOTATION LINKING MARK..IDEOGRAPHIC ANNOTATION REVERSE MARK
    // 0x3192 .. 0x3195: No     [4] IDEOGRAPHIC ANNOTATION ONE MARK..IDEOGRAPHIC ANNOTATION FOUR MARK
    // 0x3196 .. 0x319F: So    [10] IDEOGRAPHIC ANNOTATION TOP MARK..IDEOGRAPHIC ANNOTATION MAN MARK
    // 0x31A0 .. 0x31BF: Lo    [32] BOPOMOFO LETTER BU..BOPOMOFO LETTER AH
    // 0x31C0 .. 0x31E5: So    [38] CJK STROKE T..CJK STROKE SZP
    Prop<::unicode::EastAsianWidth> { { 0x3190, 0x31E5 }, ::unicode::EastAsianWidth::Wide }, // #5
    // 0x31EF .. 0x31EF: So         IDEOGRAPHIC DESCRIPTION CHARACTER SUBTRACTION
    // 0x31F0 .. 0x31FF: Lo    [16] KATAKANA LETTER SMALL KU..KATAKANA LETTER SMALL RO
    // 0x3200 .. 0x321E: So    [31] PARENTHESIZED HANGUL KIYEOK..PARENTHESIZED KOREAN CHARACTER O HU
    Prop<::unicode::EastAsianWidth> { { 0x31EF, 0x321E }, ::unicode::EastAsianWidth::Wide }, // #3
    // 0x3220 .. 0x3229: No    [10] PARENTHESIZED IDEOGRAPH ONE..PARENTHESIZED IDEOGRAPH TEN
    // 0x322A .. 0x3247: So    [30] PARENTHESIZED IDEOGRAPH MOON..CIRCLED IDEOGRAPH KOTO
    Prop<::unicode::EastAsianWidth> { { 0x3220, 0x3247 }, ::unicode::EastAsianWidth::Wide }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x3248, 0x324F }, ::unicode::EastAsianWidth::Ambiguous }, // No     [8] CIRCLED NUMBER TEN ON BLACK SQUARE..CIRCLED NUMBER EIGHTY ON BLACK SQUARE
    // 0x3250 .. 0x3250: So         PARTNERSHIP SIGN
    // 0x3251 .. 0x325F: No    [15] CIRCLED NUMBER TWENTY ONE..CIRCLED NUMBER THIRTY FIVE
    // 0x3260 .. 0x327F: So    [32] CIRCLED HANGUL KIYEOK..KOREAN STANDARD SYMBOL
    // 0x3280 .. 0x3289: No    [10] CIRCLED IDEOGRAPH ONE..CIRCLED IDEOGRAPH TEN
    // 0x328A .. 0x32B0: So    [39] CIRCLED IDEOGRAPH MOON..CIRCLED IDEOGRAPH NIGHT
    // 0x32B1 .. 0x32BF: No    [15] CIRCLED NUMBER THIRTY SIX..CIRCLED NUMBER FIFTY
    // 0x32C0 .. 0x32FF: So    [64] IDEOGRAPHIC TELEGRAPH SYMBOL FOR JANUARY..SQUARE ERA NAME REIWA
    // 0x3300 .. 0x33FF: So   [256] SQUARE APAATO..SQUARE GAL
    // 0x3400 .. 0x4DBF: Lo  [6592] CJK UNIFIED IDEOGRAPH-3400..CJK UNIFIED IDEOGRAPH-4DBF
    // 0x4DC0 .. 0x4DFF: So    [64] HEXAGRAM FOR THE CREATIVE HEAVEN..HEXAGRAM FOR BEFORE COMPLETION
    // 0x4E00 .. 0x9FFF: Lo [20992] CJK UNIFIED IDEOGRAPH-4E00..CJK UNIFIED IDEOGRAPH-9FFF
    // 0xA000 .. 0xA014: Lo    [21] YI SYLLABLE IT..YI SYLLABLE E
    // 0xA015 .. 0xA015: Lm         YI SYLLABLE WU
    // 0xA016 .. 0xA48C: Lo  [1143] YI SYLLABLE BIT..YI SYLLABLE YYR
    Prop<::unicode::EastAsianWidth> { { 0x3250, 0xA48C }, ::unicode::EastAsianWidth::Wide }, // #14
    Prop<::unicode::EastAsianWidth> { { 0xA490, 0xA4C6 }, ::unicode::EastAsianWidth::Wide }, // So    [55] YI RADICAL QOT..YI RADICAL KE
    // 0xA4D0 .. 0xA4F7: Lo    [40] LISU LETTER BA..LISU LETTER OE
    // 0xA4F8 .. 0xA4FD: Lm     [6] LISU LETTER TONE MYA TI..LISU LETTER TONE MYA JEU
    // 0xA4FE .. 0xA4FF: Po     [2] LISU PUNCTUATION COMMA..LISU PUNCTUATION FULL STOP
    // 0xA500 .. 0xA60B: Lo   [268] VAI SYLLABLE EE..VAI SYLLABLE NG
    // 0xA60C .. 0xA60C: Lm         VAI SYLLABLE LENGTHENER
    // 0xA60D .. 0xA60F: Po     [3] VAI COMMA..VAI QUESTION MARK
    // 0xA610 .. 0xA61F: Lo    [16] VAI SYLLABLE NDOLE FA..VAI SYMBOL JONG
    // 0xA620 .. 0xA629: Nd    [10] VAI DIGIT ZERO..VAI DIGIT NINE
    // 0xA62A .. 0xA62B: Lo     [2] VAI SYLLABLE NDOLE MA..VAI SYLLABLE NDOLE DO
    Prop<::unicode::EastAsianWidth> { { 0xA4D0, 0xA62B }, ::unicode::EastAsianWidth::Neutral }, // #9
    // 0xA640 .. 0xA66D: L&    [46] CYRILLIC CAPITAL LETTER ZEMLYA..CYRILLIC SMALL LETTER DOUBLE MONOCULAR O
    // 0xA66E .. 0xA66E: Lo         CYRILLIC LETTER MULTIOCULAR O
    // 0xA66F .. 0xA66F: Mn         COMBINING CYRILLIC VZMET
    // 0xA670 .. 0xA672: Me     [3] COMBINING CYRILLIC TEN MILLIONS SIGN..COMBINING CYRILLIC THOUSAND MILLIONS SIGN
    // 0xA673 .. 0xA673: Po         SLAVONIC ASTERISK
    // 0xA674 .. 0xA67D: Mn    [10] COMBINING CYRILLIC LETTER UKRAINIAN IE..COMBINING CYRILLIC PAYEROK
    // 0xA67E .. 0xA67E: Po         CYRILLIC KAVYKA
    // 0xA67F .. 0xA67F: Lm         CYRILLIC PAYEROK
    // 0xA680 .. 0xA69B: L&    [28] CYRILLIC CAPITAL LETTER DWE..CYRILLIC SMALL LETTER CROSSED O
    // 0xA69C .. 0xA69D: Lm     [2] MODIFIER LETTER CYRILLIC HARD SIGN..MODIFIER LETTER CYRILLIC SOFT SIGN
    // 0xA69E .. 0xA69F: Mn     [2] COMBINING CYRILLIC LETTER EF..COMBINING CYRILLIC LETTER IOTIFIED E
    // 0xA6A0 .. 0xA6E5: Lo    [70] BAMUM LETTER A..BAMUM LETTER KI
    // 0xA6E6 .. 0xA6EF: Nl    [10] BAMUM LETTER MO..BAMUM LETTER KOGHOM
    // 0xA6F0 .. 0xA6F1: Mn     [2] BAMUM COMBINING MARK KOQNDON..BAMUM COMBINING MARK TUKWENTIS
    // 0xA6F2 .. 0xA6F7: Po     [6] BAMUM NJAEMLI..BAMUM QUESTION MARK
    Prop<::unicode::EastAsianWidth> { { 0xA640, 0xA6F7 }, ::unicode::EastAsianWidth::Neutral }, // #15
    // 0xA700 .. 0xA716: Sk    [23] MODIFIER LETTER CHINESE TONE YIN PING..MODIFIER LETTER EXTRA-LOW LEFT-STEM TONE BAR
    // 0xA717 .. 0xA71F: Lm     [9] MODIFIER LETTER DOT VERTICAL BAR..MODIFIER LETTER LOW INVERTED EXCLAMATION MARK
    // 0xA720 .. 0xA721: Sk     [2] MODIFIER LETTER STRESS AND HIGH TONE..MODIFIER LETTER STRESS AND LOW TONE
    // 0xA722 .. 0xA76F: L&    [78] LATIN CAPITAL LETTER EGYPTOLOGICAL ALEF..LATIN SMALL LETTER CON
    // 0xA770 .. 0xA770: Lm         MODIFIER LETTER US
    // 0xA771 .. 0xA787: L&    [23] LATIN SMALL LETTER DUM..LATIN SMALL LETTER INSULAR T
    // 0xA788 .. 0xA788: Lm         MODIFIER LETTER LOW CIRCUMFLEX ACCENT
    // 0xA789 .. 0xA78A: Sk     [2] MODIFIER LETTER COLON..MODIFIER LETTER SHORT EQUALS SIGN
    // 0xA78B .. 0xA78E: L&     [4] LATIN CAPITAL LETTER SALTILLO..LATIN SMALL LETTER L WITH RETROFLEX HOOK AND BELT
    // 0xA78F .. 0xA78F: Lo         LATIN LETTER SINOLOGICAL DOT
    // 0xA790 .. 0xA7DC: L&    [77] LATIN CAPITAL LETTER N WITH DESCENDER..LATIN CAPITAL LETTER LAMBDA WITH STROKE
    Prop<::unicode::EastAsianWidth> { { 0xA700, 0xA7DC }, ::unicode::EastAsianWidth::Neutral }, // #11
    // 0xA7F1 .. 0xA7F4: Lm     [4] MODIFIER LETTER CAPITAL S..MODIFIER LETTER CAPITAL Q
    // 0xA7F5 .. 0xA7F6: L&     [2] LATIN CAPITAL LETTER REVERSED HALF H..LATIN SMALL LETTER REVERSED HALF H
    // 0xA7F7 .. 0xA7F7: Lo         LATIN EPIGRAPHIC LETTER SIDEWAYS I
    // 0xA7F8 .. 0xA7F9: Lm     [2] MODIFIER LETTER CAPITAL H WITH STROKE..MODIFIER LETTER SMALL LIGATURE OE
    // 0xA7FA .. 0xA7FA: Ll         LATIN LETTER SMALL CAPITAL TURNED M
    // 0xA7FB .. 0xA7FF: Lo     [5] LATIN EPIGRAPHIC LETTER REVERSED F..LATIN EPIGRAPHIC LETTER ARCHAIC M
    // 0xA800 .. 0xA801: Lo     [2] SYLOTI NAGRI LETTER A..SYLOTI NAGRI LETTER I
    // 0xA802 .. 0xA802: Mn         SYLOTI NAGRI SIGN DVISVARA
    // 0xA803 .. 0xA805: Lo     [3] SYLOTI NAGRI LETTER U..SYLOTI NAGRI LETTER O
    // 0xA806 .. 0xA806: Mn         SYLOTI NAGRI SIGN HASANTA
    // 0xA807 .. 0xA80A: Lo     [4] SYLOTI NAGRI LETTER KO..SYLOTI NAGRI LETTER GHO
    // 0xA80B .. 0xA80B: Mn         SYLOTI NAGRI SIGN ANUSVARA
    // 0xA80C .. 0xA822: Lo    [23] SYLOTI NAGRI LETTER CO..SYLOTI NAGRI LETTER HO
    // 0xA823 .. 0xA824: Mc     [2] SYLOTI NAGRI VOWEL SIGN A..SYLOTI NAGRI VOWEL SIGN I
    // 0xA825 .. 0xA826: Mn     [2] SYLOTI NAGRI VOWEL SIGN U..SYLOTI NAGRI VOWEL SIGN E
    // 0xA827 .. 0xA827: Mc         SYLOTI NAGRI VOWEL SIGN OO
    // 0xA828 .. 0xA82B: So     [4] SYLOTI NAGRI POETRY MARK-1..SYLOTI NAGRI POETRY MARK-4
    // 0xA82C .. 0xA82C: Mn         SYLOTI NAGRI SIGN ALTERNATE HASANTA
    Prop<::unicode::EastAsianWidth> { { 0xA7F1, 0xA82C }, ::unicode::EastAsianWidth::Neutral }, // #18
    // 0xA830 .. 0xA835: No     [6] NORTH INDIC FRACTION ONE QUARTER..NORTH INDIC FRACTION THREE SIXTEENTHS
    // 0xA836 .. 0xA837: So     [2] NORTH INDIC QUARTER MARK..NORTH INDIC PLACEHOLDER MARK
    // 0xA838 .. 0xA838: Sc         NORTH INDIC RUPEE MARK
    // 0xA839 .. 0xA839: So         NORTH INDIC QUANTITY MARK
    Prop<::unicode::EastAsianWidth> { { 0xA830, 0xA839 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0xA840 .. 0xA873: Lo    [52] PHAGS-PA LETTER KA..PHAGS-PA LETTER CANDRABINDU
    // 0xA874 .. 0xA877: Po     [4] PHAGS-PA SINGLE HEAD MARK..PHAGS-PA MARK DOUBLE SHAD
    Prop<::unicode::EastAsianWidth> { { 0xA840, 0xA877 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0xA880 .. 0xA881: Mc     [2] SAURASHTRA SIGN ANUSVARA..SAURASHTRA SIGN VISARGA
    // 0xA882 .. 0xA8B3: Lo    [50] SAURASHTRA LETTER A..SAURASHTRA LETTER LLA
    // 0xA8B4 .. 0xA8C3: Mc    [16] SAURASHTRA CONSONANT SIGN HAARU..SAURASHTRA VOWEL SIGN AU
    // 0xA8C4 .. 0xA8C5: Mn     [2] SAURASHTRA SIGN VIRAMA..SAURASHTRA SIGN CANDRABINDU
    Prop<::unicode::EastAsianWidth> { { 0xA880, 0xA8C5 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0xA8CE .. 0xA8CF: Po     [2] SAURASHTRA DANDA..SAURASHTRA DOUBLE DANDA
    // 0xA8D0 .. 0xA8D9: Nd    [10] SAURASHTRA DIGIT ZERO..SAURASHTRA DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0xA8CE, 0xA8D9 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0xA8E0 .. 0xA8F1: Mn    [18] COMBINING DEVANAGARI DIGIT ZERO..COMBINING DEVANAGARI SIGN AVAGRAHA
    // 0xA8F2 .. 0xA8F7: Lo     [6] DEVANAGARI SIGN SPACING CANDRABINDU..DEVANAGARI SIGN CANDRABINDU AVAGRAHA
    // 0xA8F8 .. 0xA8FA: Po     [3] DEVANAGARI SIGN PUSHPIKA..DEVANAGARI CARET
    // 0xA8FB .. 0xA8FB: Lo         DEVANAGARI HEADSTROKE
    // 0xA8FC .. 0xA8FC: Po         DEVANAGARI SIGN SIDDHAM
    // 0xA8FD .. 0xA8FE: Lo     [2] DEVANAGARI JAIN OM..DEVANAGARI LETTER AY
    // 0xA8FF .. 0xA8FF: Mn         DEVANAGARI VOWEL SIGN AY
    // 0xA900 .. 0xA909: Nd    [10] KAYAH LI DIGIT ZERO..KAYAH LI DIGIT NINE
    // 0xA90A .. 0xA925: Lo    [28] KAYAH LI LETTER KA..KAYAH LI LETTER OO
    // 0xA926 .. 0xA92D: Mn     [8] KAYAH LI VOWEL UE..KAYAH LI TONE CALYA PLOPHU
    // 0xA92E .. 0xA92F: Po     [2] KAYAH LI SIGN CWI..KAYAH LI SIGN SHYA
    // 0xA930 .. 0xA946: Lo    [23] REJANG LETTER KA..REJANG LETTER A
    // 0xA947 .. 0xA951: Mn    [11] REJANG VOWEL SIGN I..REJANG CONSONANT SIGN R
    // 0xA952 .. 0xA953: Mc     [2] REJANG CONSONANT SIGN H..REJANG VIRAMA
    Prop<::unicode::EastAsianWidth> { { 0xA8E0, 0xA953 }, ::unicode::EastAsianWidth::Neutral }, // #14
    Prop<::unicode::EastAsianWidth> { { 0xA95F, 0xA95F }, ::unicode::EastAsianWidth::Neutral }, // Po         REJANG SECTION MARK
    Prop<::unicode::EastAsianWidth> { { 0xA960, 0xA97C }, ::unicode::EastAsianWidth::Wide }, // Lo    [29] HANGUL CHOSEONG TIKEUT-MIEUM..HANGUL CHOSEONG SSANGYEORINHIEUH
    // 0xA980 .. 0xA982: Mn     [3] JAVANESE SIGN PANYANGGA..JAVANESE SIGN LAYAR
    // 0xA983 .. 0xA983: Mc         JAVANESE SIGN WIGNYAN
    // 0xA984 .. 0xA9B2: Lo    [47] JAVANESE LETTER A..JAVANESE LETTER HA
    // 0xA9B3 .. 0xA9B3: Mn         JAVANESE SIGN CECAK TELU
    // 0xA9B4 .. 0xA9B5: Mc     [2] JAVANESE VOWEL SIGN TARUNG..JAVANESE VOWEL SIGN TOLONG
    // 0xA9B6 .. 0xA9B9: Mn     [4] JAVANESE VOWEL SIGN WULU..JAVANESE VOWEL SIGN SUKU MENDUT
    // 0xA9BA .. 0xA9BB: Mc     [2] JAVANESE VOWEL SIGN TALING..JAVANESE VOWEL SIGN DIRGA MURE
    // 0xA9BC .. 0xA9BD: Mn     [2] JAVANESE VOWEL SIGN PEPET..JAVANESE CONSONANT SIGN KERET
    // 0xA9BE .. 0xA9C0: Mc     [3] JAVANESE CONSONANT SIGN PENGKAL..JAVANESE PANGKON
    // 0xA9C1 .. 0xA9CD: Po    [13] JAVANESE LEFT RERENGGAN..JAVANESE TURNED PADA PISELEH
    Prop<::unicode::EastAsianWidth> { { 0xA980, 0xA9CD }, ::unicode::EastAsianWidth::Neutral }, // #10
    // 0xA9CF .. 0xA9CF: Lm         JAVANESE PANGRANGKEP
    // 0xA9D0 .. 0xA9D9: Nd    [10] JAVANESE DIGIT ZERO..JAVANESE DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0xA9CF, 0xA9D9 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0xA9DE .. 0xA9DF: Po     [2] JAVANESE PADA TIRTA TUMETES..JAVANESE PADA ISEN-ISEN
    // 0xA9E0 .. 0xA9E4: Lo     [5] MYANMAR LETTER SHAN GHA..MYANMAR LETTER SHAN BHA
    // 0xA9E5 .. 0xA9E5: Mn         MYANMAR SIGN SHAN SAW
    // 0xA9E6 .. 0xA9E6: Lm         MYANMAR MODIFIER LETTER SHAN REDUPLICATION
    // 0xA9E7 .. 0xA9EF: Lo     [9] MYANMAR LETTER TAI LAING NYA..MYANMAR LETTER TAI LAING NNA
    // 0xA9F0 .. 0xA9F9: Nd    [10] MYANMAR TAI LAING DIGIT ZERO..MYANMAR TAI LAING DIGIT NINE
    // 0xA9FA .. 0xA9FE: Lo     [5] MYANMAR LETTER TAI LAING LLA..MYANMAR LETTER TAI LAING BHA
    Prop<::unicode::EastAsianWidth> { { 0xA9DE, 0xA9FE }, ::unicode::EastAsianWidth::Neutral }, // #7
    // 0xAA00 .. 0xAA28: Lo    [41] CHAM LETTER A..CHAM LETTER HA
    // 0xAA29 .. 0xAA2E: Mn     [6] CHAM VOWEL SIGN AA..CHAM VOWEL SIGN OE
    // 0xAA2F .. 0xAA30: Mc     [2] CHAM VOWEL SIGN O..CHAM VOWEL SIGN AI
    // 0xAA31 .. 0xAA32: Mn     [2] CHAM VOWEL SIGN AU..CHAM VOWEL SIGN UE
    // 0xAA33 .. 0xAA34: Mc     [2] CHAM CONSONANT SIGN YA..CHAM CONSONANT SIGN RA
    // 0xAA35 .. 0xAA36: Mn     [2] CHAM CONSONANT SIGN LA..CHAM CONSONANT SIGN WA
    Prop<::unicode::EastAsianWidth> { { 0xAA00, 0xAA36 }, ::unicode::EastAsianWidth::Neutral }, // #6
    // 0xAA40 .. 0xAA42: Lo     [3] CHAM LETTER FINAL K..CHAM LETTER FINAL NG
    // 0xAA43 .. 0xAA43: Mn         CHAM CONSONANT SIGN FINAL NG
    // 0xAA44 .. 0xAA4B: Lo     [8] CHAM LETTER FINAL CH..CHAM LETTER FINAL SS
    // 0xAA4C .. 0xAA4C: Mn         CHAM CONSONANT SIGN FINAL M
    // 0xAA4D .. 0xAA4D: Mc         CHAM CONSONANT SIGN FINAL H
    Prop<::unicode::EastAsianWidth> { { 0xAA40, 0xAA4D }, ::unicode::EastAsianWidth::Neutral }, // #5
    Prop<::unicode::EastAsianWidth> { { 0xAA50, 0xAA59 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] CHAM DIGIT ZERO..CHAM DIGIT NINE
    // 0xAA5C .. 0xAA5F: Po     [4] CHAM PUNCTUATION SPIRAL..CHAM PUNCTUATION TRIPLE DANDA
    // 0xAA60 .. 0xAA6F: Lo    [16] MYANMAR LETTER KHAMTI GA..MYANMAR LETTER KHAMTI FA
    // 0xAA70 .. 0xAA70: Lm         MYANMAR MODIFIER LETTER KHAMTI REDUPLICATION
    // 0xAA71 .. 0xAA76: Lo     [6] MYANMAR LETTER KHAMTI XA..MYANMAR LOGOGRAM KHAMTI HM
    // 0xAA77 .. 0xAA79: So     [3] MYANMAR SYMBOL AITON EXCLAMATION..MYANMAR SYMBOL AITON TWO
    // 0xAA7A .. 0xAA7A: Lo         MYANMAR LETTER AITON RA
    // 0xAA7B .. 0xAA7B: Mc         MYANMAR SIGN PAO KAREN TONE
    // 0xAA7C .. 0xAA7C: Mn         MYANMAR SIGN TAI LAING TONE-2
    // 0xAA7D .. 0xAA7D: Mc         MYANMAR SIGN TAI LAING TONE-5
    // 0xAA7E .. 0xAA7F: Lo     [2] MYANMAR LETTER SHWE PALAUNG CHA..MYANMAR LETTER SHWE PALAUNG SHA
    // 0xAA80 .. 0xAAAF: Lo    [48] TAI VIET LETTER LOW KO..TAI VIET LETTER HIGH O
    // 0xAAB0 .. 0xAAB0: Mn         TAI VIET MAI KANG
    // 0xAAB1 .. 0xAAB1: Lo         TAI VIET VOWEL AA
    // 0xAAB2 .. 0xAAB4: Mn     [3] TAI VIET VOWEL I..TAI VIET VOWEL U
    // 0xAAB5 .. 0xAAB6: Lo     [2] TAI VIET VOWEL E..TAI VIET VOWEL O
    // 0xAAB7 .. 0xAAB8: Mn     [2] TAI VIET MAI KHIT..TAI VIET VOWEL IA
    // 0xAAB9 .. 0xAABD: Lo     [5] TAI VIET VOWEL UEA..TAI VIET VOWEL AN
    // 0xAABE .. 0xAABF: Mn     [2] TAI VIET VOWEL AM..TAI VIET TONE MAI EK
    // 0xAAC0 .. 0xAAC0: Lo         TAI VIET TONE MAI NUENG
    // 0xAAC1 .. 0xAAC1: Mn         TAI VIET TONE MAI THO
    // 0xAAC2 .. 0xAAC2: Lo         TAI VIET TONE MAI SONG
    Prop<::unicode::EastAsianWidth> { { 0xAA5C, 0xAAC2 }, ::unicode::EastAsianWidth::Neutral }, // #21
    // 0xAADB .. 0xAADC: Lo     [2] TAI VIET SYMBOL KON..TAI VIET SYMBOL NUENG
    // 0xAADD .. 0xAADD: Lm         TAI VIET SYMBOL SAM
    // 0xAADE .. 0xAADF: Po     [2] TAI VIET SYMBOL HO HOI..TAI VIET SYMBOL KOI KOI
    // 0xAAE0 .. 0xAAEA: Lo    [11] MEETEI MAYEK LETTER E..MEETEI MAYEK LETTER SSA
    // 0xAAEB .. 0xAAEB: Mc         MEETEI MAYEK VOWEL SIGN II
    // 0xAAEC .. 0xAAED: Mn     [2] MEETEI MAYEK VOWEL SIGN UU..MEETEI MAYEK VOWEL SIGN AAI
    // 0xAAEE .. 0xAAEF: Mc     [2] MEETEI MAYEK VOWEL SIGN AU..MEETEI MAYEK VOWEL SIGN AAU
    // 0xAAF0 .. 0xAAF1: Po     [2] MEETEI MAYEK CHEIKHAN..MEETEI MAYEK AHANG KHUDAM
    // 0xAAF2 .. 0xAAF2: Lo         MEETEI MAYEK ANJI
    // 0xAAF3 .. 0xAAF4: Lm     [2] MEETEI MAYEK SYLLABLE REPETITION MARK..MEETEI MAYEK WORD REPETITION MARK
    // 0xAAF5 .. 0xAAF5: Mc         MEETEI MAYEK VOWEL SIGN VISARGA
    // 0xAAF6 .. 0xAAF6: Mn         MEETEI MAYEK VIRAMA
    Prop<::unicode::EastAsianWidth> { { 0xAADB, 0xAAF6 }, ::unicode::EastAsianWidth::Neutral }, // #12
    Prop<::unicode::EastAsianWidth> { { 0xAB01, 0xAB06 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [6] ETHIOPIC SYLLABLE TTHU..ETHIOPIC SYLLABLE TTHO
    Prop<::unicode::EastAsianWidth> { { 0xAB09, 0xAB0E }, ::unicode::EastAsianWidth::Neutral }, // Lo     [6] ETHIOPIC SYLLABLE DDHU..ETHIOPIC SYLLABLE DDHO
    Prop<::unicode::EastAsianWidth> { { 0xAB11, 0xAB16 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [6] ETHIOPIC SYLLABLE DZU..ETHIOPIC SYLLABLE DZO
    Prop<::unicode::EastAsianWidth> { { 0xAB20, 0xAB26 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE CCHHA..ETHIOPIC SYLLABLE CCHHO
    Prop<::unicode::EastAsianWidth> { { 0xAB28, 0xAB2E }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE BBA..ETHIOPIC SYLLABLE BBO
    // 0xAB30 .. 0xAB5A: Ll    [43] LATIN SMALL LETTER BARRED ALPHA..LATIN SMALL LETTER Y WITH SHORT RIGHT LEG
    // 0xAB5B .. 0xAB5B: Sk         MODIFIER BREVE WITH INVERTED BREVE
    // 0xAB5C .. 0xAB5F: Lm     [4] MODIFIER LETTER SMALL HENG..MODIFIER LETTER SMALL U WITH LEFT HOOK
    // 0xAB60 .. 0xAB68: Ll     [9] LATIN SMALL LETTER SAKHA YAT..LATIN SMALL LETTER TURNED R WITH MIDDLE TILDE
    // 0xAB69 .. 0xAB69: Lm         MODIFIER LETTER SMALL TURNED W
    // 0xAB6A .. 0xAB6B: Sk     [2] MODIFIER LETTER LEFT TACK..MODIFIER LETTER RIGHT TACK
    Prop<::unicode::EastAsianWidth> { { 0xAB30, 0xAB6B }, ::unicode::EastAsianWidth::Neutral }, // #6
    // 0xAB70 .. 0xABBF: Ll    [80] CHEROKEE SMALL LETTER A..CHEROKEE SMALL LETTER YA
    // 0xABC0 .. 0xABE2: Lo    [35] MEETEI MAYEK LETTER KOK..MEETEI MAYEK LETTER I LONSUM
    // 0xABE3 .. 0xABE4: Mc     [2] MEETEI MAYEK VOWEL SIGN ONAP..MEETEI MAYEK VOWEL SIGN INAP
    // 0xABE5 .. 0xABE5: Mn         MEETEI MAYEK VOWEL SIGN ANAP
    // 0xABE6 .. 0xABE7: Mc     [2] MEETEI MAYEK VOWEL SIGN YENAP..MEETEI MAYEK VOWEL SIGN SOUNAP
    // 0xABE8 .. 0xABE8: Mn         MEETEI MAYEK VOWEL SIGN UNAP
    // 0xABE9 .. 0xABEA: Mc     [2] MEETEI MAYEK VOWEL SIGN CHEINAP..MEETEI MAYEK VOWEL SIGN NUNG
    // 0xABEB .. 0xABEB: Po         MEETEI MAYEK CHEIKHEI
    // 0xABEC .. 0xABEC: Mc         MEETEI MAYEK LUM IYEK
    // 0xABED .. 0xABED: Mn         MEETEI MAYEK APUN IYEK
    Prop<::unicode::EastAsianWidth> { { 0xAB70, 0xABED }, ::unicode::EastAsianWidth::Neutral }, // #10
    Prop<::unicode::EastAsianWidth> { { 0xABF0, 0xABF9 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] MEETEI MAYEK DIGIT ZERO..MEETEI MAYEK DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0xAC00, 0xD7A3 }, ::unicode::EastAsianWidth::Wide }, // Lo [11172] HANGUL SYLLABLE GA..HANGUL SYLLABLE HIH
    Prop<::unicode::EastAsianWidth> { { 0xD7B0, 0xD7C6 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [23] HANGUL JUNGSEONG O-YEO..HANGUL JUNGSEONG ARAEA-E
    Prop<::unicode::EastAsianWidth> { { 0xD7CB, 0xD7FB }, ::unicode::EastAsianWidth::Neutral }, // Lo    [49] HANGUL JONGSEONG NIEUN-RIEUL..HANGUL JONGSEONG PHIEUPH-THIEUTH
    // 0xD800 .. 0xDB7F: Cs   [896] <surrogate-D800>..<surrogate-DB7F>
    // 0xDB80 .. 0xDBFF: Cs   [128] <surrogate-DB80>..<surrogate-DBFF>
    // 0xDC00 .. 0xDFFF: Cs  [1024] <surrogate-DC00>..<surrogate-DFFF>
    Prop<::unicode::EastAsianWidth> { { 0xD800, 0xDFFF }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0xE000, 0xF8FF }, ::unicode::EastAsianWidth::Ambiguous }, // Co  [6400] <private-use-E000>..<private-use-F8FF>
    // 0xF900 .. 0xFA6D: Lo   [366] CJK COMPATIBILITY IDEOGRAPH-F900..CJK COMPATIBILITY IDEOGRAPH-FA6D
    // 0xFA6E .. 0xFA6F: Cn     [2] <reserved-FA6E>..<reserved-FA6F>
    // 0xFA70 .. 0xFAD9: Lo   [106] CJK COMPATIBILITY IDEOGRAPH-FA70..CJK COMPATIBILITY IDEOGRAPH-FAD9
    // 0xFADA .. 0xFAFF: Cn    [38] <reserved-FADA>..<reserved-FAFF>
    Prop<::unicode::EastAsianWidth> { { 0xF900, 0xFAFF }, ::unicode::EastAsianWidth::Wide }, // #4
    Prop<::unicode::EastAsianWidth> { { 0xFB00, 0xFB06 }, ::unicode::EastAsianWidth::Neutral }, // Ll     [7] LATIN SMALL LIGATURE FF..LATIN SMALL LIGATURE ST
    Prop<::unicode::EastAsianWidth> { { 0xFB13, 0xFB17 }, ::unicode::EastAsianWidth::Neutral }, // Ll     [5] ARMENIAN SMALL LIGATURE MEN NOW..ARMENIAN SMALL LIGATURE MEN XEH
    // 0xFB1D .. 0xFB1D: Lo         HEBREW LETTER YOD WITH HIRIQ
    // 0xFB1E .. 0xFB1E: Mn         HEBREW POINT JUDEO-SPANISH VARIKA
    // 0xFB1F .. 0xFB28: Lo    [10] HEBREW LIGATURE YIDDISH YOD YOD PATAH..HEBREW LETTER WIDE TAV
    // 0xFB29 .. 0xFB29: Sm         HEBREW LETTER ALTERNATIVE PLUS SIGN
    // 0xFB2A .. 0xFB36: Lo    [13] HEBREW LETTER SHIN WITH SHIN DOT..HEBREW LETTER ZAYIN WITH DAGESH
    Prop<::unicode::EastAsianWidth> { { 0xFB1D, 0xFB36 }, ::unicode::EastAsianWidth::Neutral }, // #5
    Prop<::unicode::EastAsianWidth> { { 0xFB38, 0xFB3C }, ::unicode::EastAsianWidth::Neutral }, // Lo     [5] HEBREW LETTER TET WITH DAGESH..HEBREW LETTER LAMED WITH DAGESH
    Prop<::unicode::EastAsianWidth> { { 0xFB3E, 0xFB3E }, ::unicode::EastAsianWidth::Neutral }, // Lo         HEBREW LETTER MEM WITH DAGESH
    Prop<::unicode::EastAsianWidth> { { 0xFB40, 0xFB41 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] HEBREW LETTER NUN WITH DAGESH..HEBREW LETTER SAMEKH WITH DAGESH
    Prop<::unicode::EastAsianWidth> { { 0xFB43, 0xFB44 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] HEBREW LETTER FINAL PE WITH DAGESH..HEBREW LETTER PE WITH DAGESH
    // 0xFB46 .. 0xFB4F: Lo    [10] HEBREW LETTER TSADI WITH DAGESH..HEBREW LIGATURE ALEF LAMED
    // 0xFB50 .. 0xFBB1: Lo    [98] ARABIC LETTER ALEF WASLA ISOLATED FORM..ARABIC LETTER YEH BARREE WITH HAMZA ABOVE FINAL FORM
    // 0xFBB2 .. 0xFBC2: Sk    [17] ARABIC SYMBOL DOT ABOVE..ARABIC SYMBOL WASLA ABOVE
    // 0xFBC3 .. 0xFBD2: So    [16] ARABIC LIGATURE JALLA WA-ALAA..ARABIC LIGATURE ALAYHI AR-RAHMAH
    // 0xFBD3 .. 0xFD3D: Lo   [363] ARABIC LETTER NG ISOLATED FORM..ARABIC LIGATURE ALEF WITH FATHATAN ISOLATED FORM
    // 0xFD3E .. 0xFD3E: Pe         ORNATE LEFT PARENTHESIS
    // 0xFD3F .. 0xFD3F: Ps         ORNATE RIGHT PARENTHESIS
    // 0xFD40 .. 0xFD4F: So    [16] ARABIC LIGATURE RAHIMAHU ALLAAH..ARABIC LIGATURE RAHIMAHUM ALLAAH
    // 0xFD50 .. 0xFD8F: Lo    [64] ARABIC LIGATURE TEH WITH JEEM WITH MEEM INITIAL FORM..ARABIC LIGATURE MEEM WITH KHAH WITH MEEM INITIAL FORM
    // 0xFD90 .. 0xFD91: So     [2] ARABIC LIGATURE RAHMATU ALLAAHI ALAYH..ARABIC LIGATURE RAHMATU ALLAAHI ALAYHAA
    // 0xFD92 .. 0xFDC7: Lo    [54] ARABIC LIGATURE MEEM WITH JEEM WITH KHAH INITIAL FORM..ARABIC LIGATURE NOON WITH JEEM WITH YEH FINAL FORM
    // 0xFDC8 .. 0xFDCF: So     [8] ARABIC LIGATURE RAHIMAHU ALLAAH TAAALAA..ARABIC LIGATURE SALAAMUHU ALAYNAA
    Prop<::unicode::EastAsianWidth> { { 0xFB46, 0xFDCF }, ::unicode::EastAsianWidth::Neutral }, // #12
    // 0xFDF0 .. 0xFDFB: Lo    [12] ARABIC LIGATURE SALLA USED AS KORANIC STOP SIGN ISOLATED FORM..ARABIC LIGATURE JALLAJALALOUHOU
    // 0xFDFC .. 0xFDFC: Sc         RIAL SIGN
    // 0xFDFD .. 0xFDFF: So     [3] ARABIC LIGATURE BISMILLAH AR-RAHMAN AR-RAHEEM..ARABIC LIGATURE AZZA WA JALL
    Prop<::unicode::EastAsianWidth> { { 0xFDF0, 0xFDFF }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0xFE00, 0xFE0F }, ::unicode::EastAsianWidth::Ambiguous }, // Mn    [16] VARIATION SELECTOR-1..VARIATION SELECTOR-16
    // 0xFE10 .. 0xFE16: Po     [7] PRESENTATION FORM FOR VERTICAL COMMA..PRESENTATION FORM FOR VERTICAL QUESTION MARK
    // 0xFE17 .. 0xFE17: Ps         PRESENTATION FORM FOR VERTICAL LEFT WHITE LENTICULAR BRACKET
    // 0xFE18 .. 0xFE18: Pe         PRESENTATION FORM FOR VERTICAL RIGHT WHITE LENTICULAR BRAKCET
    // 0xFE19 .. 0xFE19: Po         PRESENTATION FORM FOR VERTICAL HORIZONTAL ELLIPSIS
    Prop<::unicode::EastAsianWidth> { { 0xFE10, 0xFE19 }, ::unicode::EastAsianWidth::Wide }, // #4
    Prop<::unicode::EastAsianWidth> { { 0xFE20, 0xFE2F }, ::unicode::EastAsianWidth::Neutral }, // Mn    [16] COMBINING LIGATURE LEFT HALF..COMBINING CYRILLIC TITLO RIGHT HALF
    // 0xFE30 .. 0xFE30: Po         PRESENTATION FORM FOR VERTICAL TWO DOT LEADER
    // 0xFE31 .. 0xFE32: Pd     [2] PRESENTATION FORM FOR VERTICAL EM DASH..PRESENTATION FORM FOR VERTICAL EN DASH
    // 0xFE33 .. 0xFE34: Pc     [2] PRESENTATION FORM FOR VERTICAL LOW LINE..PRESENTATION FORM FOR VERTICAL WAVY LOW LINE
    // 0xFE35 .. 0xFE35: Ps         PRESENTATION FORM FOR VERTICAL LEFT PARENTHESIS
    // 0xFE36 .. 0xFE36: Pe         PRESENTATION FORM FOR VERTICAL RIGHT PARENTHESIS
    // 0xFE37 .. 0xFE37: Ps         PRESENTATION FORM FOR VERTICAL LEFT CURLY BRACKET
    // 0xFE38 .. 0xFE38: Pe         PRESENTATION FORM FOR VERTICAL RIGHT CURLY BRACKET
    // 0xFE39 .. 0xFE39: Ps         PRESENTATION FORM FOR VERTICAL LEFT TORTOISE SHELL BRACKET
    // 0xFE3A .. 0xFE3A: Pe         PRESENTATION FORM FOR VERTICAL RIGHT TORTOISE SHELL BRACKET
    // 0xFE3B .. 0xFE3B: Ps         PRESENTATION FORM FOR VERTICAL LEFT BLACK LENTICULAR BRACKET
    // 0xFE3C .. 0xFE3C: Pe         PRESENTATION FORM FOR VERTICAL RIGHT BLACK LENTICULAR BRACKET
    // 0xFE3D .. 0xFE3D: Ps         PRESENTATION FORM FOR VERTICAL LEFT DOUBLE ANGLE BRACKET
    // 0xFE3E .. 0xFE3E: Pe         PRESENTATION FORM FOR VERTICAL RIGHT DOUBLE ANGLE BRACKET
    // 0xFE3F .. 0xFE3F: Ps         PRESENTATION FORM FOR VERTICAL LEFT ANGLE BRACKET
    // 0xFE40 .. 0xFE40: Pe         PRESENTATION FORM FOR VERTICAL RIGHT ANGLE BRACKET
    // 0xFE41 .. 0xFE41: Ps         PRESENTATION FORM FOR VERTICAL LEFT CORNER BRACKET
    // 0xFE42 .. 0xFE42: Pe         PRESENTATION FORM FOR VERTICAL RIGHT CORNER BRACKET
    // 0xFE43 .. 0xFE43: Ps         PRESENTATION FORM FOR VERTICAL LEFT WHITE CORNER BRACKET
    // 0xFE44 .. 0xFE44: Pe         PRESENTATION FORM FOR VERTICAL RIGHT WHITE CORNER BRACKET
    // 0xFE45 .. 0xFE46: Po     [2] SESAME DOT..WHITE SESAME DOT
    // 0xFE47 .. 0xFE47: Ps         PRESENTATION FORM FOR VERTICAL LEFT SQUARE BRACKET
    // 0xFE48 .. 0xFE48: Pe         PRESENTATION FORM FOR VERTICAL RIGHT SQUARE BRACKET
    // 0xFE49 .. 0xFE4C: Po     [4] DASHED OVERLINE..DOUBLE WAVY OVERLINE
    // 0xFE4D .. 0xFE4F: Pc     [3] DASHED LOW LINE..WAVY LOW LINE
    // 0xFE50 .. 0xFE52: Po     [3] SMALL COMMA..SMALL FULL STOP
    Prop<::unicode::EastAsianWidth> { { 0xFE30, 0xFE52 }, ::unicode::EastAsianWidth::Wide }, // #25
    // 0xFE54 .. 0xFE57: Po     [4] SMALL SEMICOLON..SMALL EXCLAMATION MARK
    // 0xFE58 .. 0xFE58: Pd         SMALL EM DASH
    // 0xFE59 .. 0xFE59: Ps         SMALL LEFT PARENTHESIS
    // 0xFE5A .. 0xFE5A: Pe         SMALL RIGHT PARENTHESIS
    // 0xFE5B .. 0xFE5B: Ps         SMALL LEFT CURLY BRACKET
    // 0xFE5C .. 0xFE5C: Pe         SMALL RIGHT CURLY BRACKET
    // 0xFE5D .. 0xFE5D: Ps         SMALL LEFT TORTOISE SHELL BRACKET
    // 0xFE5E .. 0xFE5E: Pe         SMALL RIGHT TORTOISE SHELL BRACKET
    // 0xFE5F .. 0xFE61: Po     [3] SMALL NUMBER SIGN..SMALL ASTERISK
    // 0xFE62 .. 0xFE62: Sm         SMALL PLUS SIGN
    // 0xFE63 .. 0xFE63: Pd         SMALL HYPHEN-MINUS
    // 0xFE64 .. 0xFE66: Sm     [3] SMALL LESS-THAN SIGN..SMALL EQUALS SIGN
    Prop<::unicode::EastAsianWidth> { { 0xFE54, 0xFE66 }, ::unicode::EastAsianWidth::Wide }, // #12
    // 0xFE68 .. 0xFE68: Po         SMALL REVERSE SOLIDUS
    // 0xFE69 .. 0xFE69: Sc         SMALL DOLLAR SIGN
    // 0xFE6A .. 0xFE6B: Po     [2] SMALL PERCENT SIGN..SMALL COMMERCIAL AT
    Prop<::unicode::EastAsianWidth> { { 0xFE68, 0xFE6B }, ::unicode::EastAsianWidth::Wide }, // #3
    Prop<::unicode::EastAsianWidth> { { 0xFE70, 0xFE74 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [5] ARABIC FATHATAN ISOLATED FORM..ARABIC KASRATAN ISOLATED FORM
    Prop<::unicode::EastAsianWidth> { { 0xFE76, 0xFEFC }, ::unicode::EastAsianWidth::Neutral }, // Lo   [135] ARABIC FATHA ISOLATED FORM..ARABIC LIGATURE LAM WITH ALEF FINAL FORM
    Prop<::unicode::EastAsianWidth> { { 0xFEFF, 0xFEFF }, ::unicode::EastAsianWidth::Neutral }, // Cf         ZERO WIDTH NO-BREAK SPACE
    // 0xFF01 .. 0xFF03: Po     [3] FULLWIDTH EXCLAMATION MARK..FULLWIDTH NUMBER SIGN
    // 0xFF04 .. 0xFF04: Sc         FULLWIDTH DOLLAR SIGN
    // 0xFF05 .. 0xFF07: Po     [3] FULLWIDTH PERCENT SIGN..FULLWIDTH APOSTROPHE
    // 0xFF08 .. 0xFF08: Ps         FULLWIDTH LEFT PARENTHESIS
    // 0xFF09 .. 0xFF09: Pe         FULLWIDTH RIGHT PARENTHESIS
    // 0xFF0A .. 0xFF0A: Po         FULLWIDTH ASTERISK
    // 0xFF0B .. 0xFF0B: Sm         FULLWIDTH PLUS SIGN
    // 0xFF0C .. 0xFF0C: Po         FULLWIDTH COMMA
    // 0xFF0D .. 0xFF0D: Pd         FULLWIDTH HYPHEN-MINUS
    // 0xFF0E .. 0xFF0F: Po     [2] FULLWIDTH FULL STOP..FULLWIDTH SOLIDUS
    // 0xFF10 .. 0xFF19: Nd    [10] FULLWIDTH DIGIT ZERO..FULLWIDTH DIGIT NINE
    // 0xFF1A .. 0xFF1B: Po     [2] FULLWIDTH COLON..FULLWIDTH SEMICOLON
    // 0xFF1C .. 0xFF1E: Sm     [3] FULLWIDTH LESS-THAN SIGN..FULLWIDTH GREATER-THAN SIGN
    // 0xFF1F .. 0xFF20: Po     [2] FULLWIDTH QUESTION MARK..FULLWIDTH COMMERCIAL AT
    // 0xFF21 .. 0xFF3A: Lu    [26] FULLWIDTH LATIN CAPITAL LETTER A..FULLWIDTH LATIN CAPITAL LETTER Z
    // 0xFF3B .. 0xFF3B: Ps         FULLWIDTH LEFT SQUARE BRACKET
    // 0xFF3C .. 0xFF3C: Po         FULLWIDTH REVERSE SOLIDUS
    // 0xFF3D .. 0xFF3D: Pe         FULLWIDTH RIGHT SQUARE BRACKET
    // 0xFF3E .. 0xFF3E: Sk         FULLWIDTH CIRCUMFLEX ACCENT
    // 0xFF3F .. 0xFF3F: Pc         FULLWIDTH LOW LINE
    // 0xFF40 .. 0xFF40: Sk         FULLWIDTH GRAVE ACCENT
    // 0xFF41 .. 0xFF5A: Ll    [26] FULLWIDTH LATIN SMALL LETTER A..FULLWIDTH LATIN SMALL LETTER Z
    // 0xFF5B .. 0xFF5B: Ps         FULLWIDTH LEFT CURLY BRACKET
    // 0xFF5C .. 0xFF5C: Sm         FULLWIDTH VERTICAL LINE
    // 0xFF5D .. 0xFF5D: Pe         FULLWIDTH RIGHT CURLY BRACKET
    // 0xFF5E .. 0xFF5E: Sm         FULLWIDTH TILDE
    // 0xFF5F .. 0xFF5F: Ps         FULLWIDTH LEFT WHITE PARENTHESIS
    // 0xFF60 .. 0xFF60: Pe         FULLWIDTH RIGHT WHITE PARENTHESIS
    Prop<::unicode::EastAsianWidth> { { 0xFF01, 0xFF60 }, ::unicode::EastAsianWidth::FullWidth }, // #28
    // 0xFF61 .. 0xFF61: Po         HALFWIDTH IDEOGRAPHIC FULL STOP
    // 0xFF62 .. 0xFF62: Ps         HALFWIDTH LEFT CORNER BRACKET
    // 0xFF63 .. 0xFF63: Pe         HALFWIDTH RIGHT CORNER BRACKET
    // 0xFF64 .. 0xFF65: Po     [2] HALFWIDTH IDEOGRAPHIC COMMA..HALFWIDTH KATAKANA MIDDLE DOT
    // 0xFF66 .. 0xFF6F: Lo    [10] HALFWIDTH KATAKANA LETTER WO..HALFWIDTH KATAKANA LETTER SMALL TU
    // 0xFF70 .. 0xFF70: Lm         HALFWIDTH KATAKANA-HIRAGANA PROLONGED SOUND MARK
    // 0xFF71 .. 0xFF9D: Lo    [45] HALFWIDTH KATAKANA LETTER A..HALFWIDTH KATAKANA LETTER N
    // 0xFF9E .. 0xFF9F: Lm     [2] HALFWIDTH KATAKANA VOICED SOUND MARK..HALFWIDTH KATAKANA SEMI-VOICED SOUND MARK
    // 0xFFA0 .. 0xFFBE: Lo    [31] HALFWIDTH HANGUL FILLER..HALFWIDTH HANGUL LETTER HIEUH
    Prop<::unicode::EastAsianWidth> { { 0xFF61, 0xFFBE }, ::unicode::EastAsianWidth::HalfWidth }, // #9
    Prop<::unicode::EastAsianWidth> { { 0xFFC2, 0xFFC7 }, ::unicode::EastAsianWidth::HalfWidth }, // Lo     [6] HALFWIDTH HANGUL LETTER A..HALFWIDTH HANGUL LETTER E
    Prop<::unicode::EastAsianWidth> { { 0xFFCA, 0xFFCF }, ::unicode::EastAsianWidth::HalfWidth }, // Lo     [6] HALFWIDTH HANGUL LETTER YEO..HALFWIDTH HANGUL LETTER OE
    Prop<::unicode::EastAsianWidth> { { 0xFFD2, 0xFFD7 }, ::unicode::EastAsianWidth::HalfWidth }, // Lo     [6] HALFWIDTH HANGUL LETTER YO..HALFWIDTH HANGUL LETTER YU
    Prop<::unicode::EastAsianWidth> { { 0xFFDA, 0xFFDC }, ::unicode::EastAsianWidth::HalfWidth }, // Lo     [3] HALFWIDTH HANGUL LETTER EU..HALFWIDTH HANGUL LETTER I
    // 0xFFE0 .. 0xFFE1: Sc     [2] FULLWIDTH CENT SIGN..FULLWIDTH POUND SIGN
    // 0xFFE2 .. 0xFFE2: Sm         FULLWIDTH NOT SIGN
    // 0xFFE3 .. 0xFFE3: Sk         FULLWIDTH MACRON
    // 0xFFE4 .. 0xFFE4: So         FULLWIDTH BROKEN BAR
    // 0xFFE5 .. 0xFFE6: Sc     [2] FULLWIDTH YEN SIGN..FULLWIDTH WON SIGN
    Prop<::unicode::EastAsianWidth> { { 0xFFE0, 0xFFE6 }, ::unicode::EastAsianWidth::FullWidth }, // #5
    // 0xFFE8 .. 0xFFE8: So         HALFWIDTH FORMS LIGHT VERTICAL
    // 0xFFE9 .. 0xFFEC: Sm     [4] HALFWIDTH LEFTWARDS ARROW..HALFWIDTH DOWNWARDS ARROW
    // 0xFFED .. 0xFFEE: So     [2] HALFWIDTH BLACK SQUARE..HALFWIDTH WHITE CIRCLE
    Prop<::unicode::EastAsianWidth> { { 0xFFE8, 0xFFEE }, ::unicode::EastAsianWidth::HalfWidth }, // #3
    // 0xFFF9 .. 0xFFFB: Cf     [3] INTERLINEAR ANNOTATION ANCHOR..INTERLINEAR ANNOTATION TERMINATOR
    // 0xFFFC .. 0xFFFC: So         OBJECT REPLACEMENT CHARACTER
    Prop<::unicode::EastAsianWidth> { { 0xFFF9, 0xFFFC }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0xFFFD, 0xFFFD }, ::unicode::EastAsianWidth::Ambiguous }, // So         REPLACEMENT CHARACTER
    Prop<::unicode::EastAsianWidth> { { 0x10000, 0x1000B }, ::unicode::EastAsianWidth::Neutral }, // Lo    [12] LINEAR B SYLLABLE B008 A..LINEAR B SYLLABLE B046 JE
    Prop<::unicode::EastAsianWidth> { { 0x1000D, 0x10026 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [26] LINEAR B SYLLABLE B036 JO..LINEAR B SYLLABLE B032 QO
    Prop<::unicode::EastAsianWidth> { { 0x10028, 0x1003A }, ::unicode::EastAsianWidth::Neutral }, // Lo    [19] LINEAR B SYLLABLE B060 RA..LINEAR B SYLLABLE B042 WO
    Prop<::unicode::EastAsianWidth> { { 0x1003C, 0x1003D }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] LINEAR B SYLLABLE B017 ZA..LINEAR B SYLLABLE B074 ZE
    Prop<::unicode::EastAsianWidth> { { 0x1003F, 0x1004D }, ::unicode::EastAsianWidth::Neutral }, // Lo    [15] LINEAR B SYLLABLE B020 ZO..LINEAR B SYLLABLE B091 TWO
    Prop<::unicode::EastAsianWidth> { { 0x10050, 0x1005D }, ::unicode::EastAsianWidth::Neutral }, // Lo    [14] LINEAR B SYMBOL B018..LINEAR B SYMBOL B089
    Prop<::unicode::EastAsianWidth> { { 0x10080, 0x100FA }, ::unicode::EastAsianWidth::Neutral }, // Lo   [123] LINEAR B IDEOGRAM B100 MAN..LINEAR B IDEOGRAM VESSEL B305
    Prop<::unicode::EastAsianWidth> { { 0x10100, 0x10102 }, ::unicode::EastAsianWidth::Neutral }, // Po     [3] AEGEAN WORD SEPARATOR LINE..AEGEAN CHECK MARK
    Prop<::unicode::EastAsianWidth> { { 0x10107, 0x10133 }, ::unicode::EastAsianWidth::Neutral }, // No    [45] AEGEAN NUMBER ONE..AEGEAN NUMBER NINETY THOUSAND
    // 0x10137 .. 0x1013F: So     [9] AEGEAN WEIGHT BASE UNIT..AEGEAN MEASURE THIRD SUBUNIT
    // 0x10140 .. 0x10174: Nl    [53] GREEK ACROPHONIC ATTIC ONE QUARTER..GREEK ACROPHONIC STRATIAN FIFTY MNAS
    // 0x10175 .. 0x10178: No     [4] GREEK ONE HALF SIGN..GREEK THREE QUARTERS SIGN
    // 0x10179 .. 0x10189: So    [17] GREEK YEAR SIGN..GREEK TRYBLION BASE SIGN
    // 0x1018A .. 0x1018B: No     [2] GREEK ZERO SIGN..GREEK ONE QUARTER SIGN
    // 0x1018C .. 0x1018E: So     [3] GREEK SINUSOID SIGN..NOMISMA SIGN
    Prop<::unicode::EastAsianWidth> { { 0x10137, 0x1018E }, ::unicode::EastAsianWidth::Neutral }, // #6
    Prop<::unicode::EastAsianWidth> { { 0x10190, 0x1019C }, ::unicode::EastAsianWidth::Neutral }, // So    [13] ROMAN SEXTANS SIGN..ASCIA SYMBOL
    Prop<::unicode::EastAsianWidth> { { 0x101A0, 0x101A0 }, ::unicode::EastAsianWidth::Neutral }, // So         GREEK SYMBOL TAU RHO
    // 0x101D0 .. 0x101FC: So    [45] PHAISTOS DISC SIGN PEDESTRIAN..PHAISTOS DISC SIGN WAVY BAND
    // 0x101FD .. 0x101FD: Mn         PHAISTOS DISC SIGN COMBINING OBLIQUE STROKE
    Prop<::unicode::EastAsianWidth> { { 0x101D0, 0x101FD }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x10280, 0x1029C }, ::unicode::EastAsianWidth::Neutral }, // Lo    [29] LYCIAN LETTER A..LYCIAN LETTER X
    Prop<::unicode::EastAsianWidth> { { 0x102A0, 0x102D0 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [49] CARIAN LETTER A..CARIAN LETTER UUU3
    // 0x102E0 .. 0x102E0: Mn         COPTIC EPACT THOUSANDS MARK
    // 0x102E1 .. 0x102FB: No    [27] COPTIC EPACT DIGIT ONE..COPTIC EPACT NUMBER NINE HUNDRED
    Prop<::unicode::EastAsianWidth> { { 0x102E0, 0x102FB }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x10300 .. 0x1031F: Lo    [32] OLD ITALIC LETTER A..OLD ITALIC LETTER ESS
    // 0x10320 .. 0x10323: No     [4] OLD ITALIC NUMERAL ONE..OLD ITALIC NUMERAL FIFTY
    Prop<::unicode::EastAsianWidth> { { 0x10300, 0x10323 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x1032D .. 0x1032F: Lo     [3] OLD ITALIC LETTER YE..OLD ITALIC LETTER SOUTHERN TSE
    // 0x10330 .. 0x10340: Lo    [17] GOTHIC LETTER AHSA..GOTHIC LETTER PAIRTHRA
    // 0x10341 .. 0x10341: Nl         GOTHIC LETTER NINETY
    // 0x10342 .. 0x10349: Lo     [8] GOTHIC LETTER RAIDA..GOTHIC LETTER OTHAL
    // 0x1034A .. 0x1034A: Nl         GOTHIC LETTER NINE HUNDRED
    Prop<::unicode::EastAsianWidth> { { 0x1032D, 0x1034A }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x10350 .. 0x10375: Lo    [38] OLD PERMIC LETTER AN..OLD PERMIC LETTER IA
    // 0x10376 .. 0x1037A: Mn     [5] COMBINING OLD PERMIC LETTER AN..COMBINING OLD PERMIC LETTER SII
    Prop<::unicode::EastAsianWidth> { { 0x10350, 0x1037A }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x10380, 0x1039D }, ::unicode::EastAsianWidth::Neutral }, // Lo    [30] UGARITIC LETTER ALPA..UGARITIC LETTER SSU
    // 0x1039F .. 0x1039F: Po         UGARITIC WORD DIVIDER
    // 0x103A0 .. 0x103C3: Lo    [36] OLD PERSIAN SIGN A..OLD PERSIAN SIGN HA
    Prop<::unicode::EastAsianWidth> { { 0x1039F, 0x103C3 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x103C8 .. 0x103CF: Lo     [8] OLD PERSIAN SIGN AURAMAZDAA..OLD PERSIAN SIGN BUUMISH
    // 0x103D0 .. 0x103D0: Po         OLD PERSIAN WORD DIVIDER
    // 0x103D1 .. 0x103D5: Nl     [5] OLD PERSIAN NUMBER ONE..OLD PERSIAN NUMBER HUNDRED
    Prop<::unicode::EastAsianWidth> { { 0x103C8, 0x103D5 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x10400 .. 0x1044F: L&    [80] DESERET CAPITAL LETTER LONG I..DESERET SMALL LETTER EW
    // 0x10450 .. 0x1047F: Lo    [48] SHAVIAN LETTER PEEP..SHAVIAN LETTER YEW
    // 0x10480 .. 0x1049D: Lo    [30] OSMANYA LETTER ALEF..OSMANYA LETTER OO
    Prop<::unicode::EastAsianWidth> { { 0x10400, 0x1049D }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x104A0, 0x104A9 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] OSMANYA DIGIT ZERO..OSMANYA DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x104B0, 0x104D3 }, ::unicode::EastAsianWidth::Neutral }, // Lu    [36] OSAGE CAPITAL LETTER A..OSAGE CAPITAL LETTER ZHA
    Prop<::unicode::EastAsianWidth> { { 0x104D8, 0x104FB }, ::unicode::EastAsianWidth::Neutral }, // Ll    [36] OSAGE SMALL LETTER A..OSAGE SMALL LETTER ZHA
    Prop<::unicode::EastAsianWidth> { { 0x10500, 0x10527 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [40] ELBASAN LETTER A..ELBASAN LETTER KHE
    Prop<::unicode::EastAsianWidth> { { 0x10530, 0x10563 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [52] CAUCASIAN ALBANIAN LETTER ALT..CAUCASIAN ALBANIAN LETTER KIW
    // 0x1056F .. 0x1056F: Po         CAUCASIAN ALBANIAN CITATION MARK
    // 0x10570 .. 0x1057A: Lu    [11] VITHKUQI CAPITAL LETTER A..VITHKUQI CAPITAL LETTER GA
    Prop<::unicode::EastAsianWidth> { { 0x1056F, 0x1057A }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x1057C, 0x1058A }, ::unicode::EastAsianWidth::Neutral }, // Lu    [15] VITHKUQI CAPITAL LETTER HA..VITHKUQI CAPITAL LETTER RE
    Prop<::unicode::EastAsianWidth> { { 0x1058C, 0x10592 }, ::unicode::EastAsianWidth::Neutral }, // Lu     [7] VITHKUQI CAPITAL LETTER SE..VITHKUQI CAPITAL LETTER XE
    Prop<::unicode::EastAsianWidth> { { 0x10594, 0x10595 }, ::unicode::EastAsianWidth::Neutral }, // Lu     [2] VITHKUQI CAPITAL LETTER Y..VITHKUQI CAPITAL LETTER ZE
    Prop<::unicode::EastAsianWidth> { { 0x10597, 0x105A1 }, ::unicode::EastAsianWidth::Neutral }, // Ll    [11] VITHKUQI SMALL LETTER A..VITHKUQI SMALL LETTER GA
    Prop<::unicode::EastAsianWidth> { { 0x105A3, 0x105B1 }, ::unicode::EastAsianWidth::Neutral }, // Ll    [15] VITHKUQI SMALL LETTER HA..VITHKUQI SMALL LETTER RE
    Prop<::unicode::EastAsianWidth> { { 0x105B3, 0x105B9 }, ::unicode::EastAsianWidth::Neutral }, // Ll     [7] VITHKUQI SMALL LETTER SE..VITHKUQI SMALL LETTER XE
    Prop<::unicode::EastAsianWidth> { { 0x105BB, 0x105BC }, ::unicode::EastAsianWidth::Neutral }, // Ll     [2] VITHKUQI SMALL LETTER Y..VITHKUQI SMALL LETTER ZE
    Prop<::unicode::EastAsianWidth> { { 0x105C0, 0x105F3 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [52] TODHRI LETTER A..TODHRI LETTER OO
    Prop<::unicode::EastAsianWidth> { { 0x10600, 0x10736 }, ::unicode::EastAsianWidth::Neutral }, // Lo   [311] LINEAR A SIGN AB001..LINEAR A SIGN A664
    Prop<::unicode::EastAsianWidth> { { 0x10740, 0x10755 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [22] LINEAR A SIGN A701 A..LINEAR A SIGN A732 JE
    Prop<::unicode::EastAsianWidth> { { 0x10760, 0x10767 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [8] LINEAR A SIGN A800..LINEAR A SIGN A807
    Prop<::unicode::EastAsianWidth> { { 0x10780, 0x10785 }, ::unicode::EastAsianWidth::Neutral }, // Lm     [6] MODIFIER LETTER SMALL CAPITAL AA..MODIFIER LETTER SMALL B WITH HOOK
    Prop<::unicode::EastAsianWidth> { { 0x10787, 0x107B0 }, ::unicode::EastAsianWidth::Neutral }, // Lm    [42] MODIFIER LETTER SMALL DZ DIGRAPH..MODIFIER LETTER SMALL V WITH RIGHT HOOK
    Prop<::unicode::EastAsianWidth> { { 0x107B2, 0x107BA }, ::unicode::EastAsianWidth::Neutral }, // Lm     [9] MODIFIER LETTER SMALL CAPITAL Y..MODIFIER LETTER SMALL S WITH CURL
    Prop<::unicode::EastAsianWidth> { { 0x10800, 0x10805 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [6] CYPRIOT SYLLABLE A..CYPRIOT SYLLABLE JA
    Prop<::unicode::EastAsianWidth> { { 0x10808, 0x10808 }, ::unicode::EastAsianWidth::Neutral }, // Lo         CYPRIOT SYLLABLE JO
    Prop<::unicode::EastAsianWidth> { { 0x1080A, 0x10835 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [44] CYPRIOT SYLLABLE KA..CYPRIOT SYLLABLE WO
    Prop<::unicode::EastAsianWidth> { { 0x10837, 0x10838 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] CYPRIOT SYLLABLE XA..CYPRIOT SYLLABLE XE
    Prop<::unicode::EastAsianWidth> { { 0x1083C, 0x1083C }, ::unicode::EastAsianWidth::Neutral }, // Lo         CYPRIOT SYLLABLE ZA
    // 0x1083F .. 0x1083F: Lo         CYPRIOT SYLLABLE ZO
    // 0x10840 .. 0x10855: Lo    [22] IMPERIAL ARAMAIC LETTER ALEPH..IMPERIAL ARAMAIC LETTER TAW
    Prop<::unicode::EastAsianWidth> { { 0x1083F, 0x10855 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x10857 .. 0x10857: Po         IMPERIAL ARAMAIC SECTION SIGN
    // 0x10858 .. 0x1085F: No     [8] IMPERIAL ARAMAIC NUMBER ONE..IMPERIAL ARAMAIC NUMBER TEN THOUSAND
    // 0x10860 .. 0x10876: Lo    [23] PALMYRENE LETTER ALEPH..PALMYRENE LETTER TAW
    // 0x10877 .. 0x10878: So     [2] PALMYRENE LEFT-POINTING FLEURON..PALMYRENE RIGHT-POINTING FLEURON
    // 0x10879 .. 0x1087F: No     [7] PALMYRENE NUMBER ONE..PALMYRENE NUMBER TWENTY
    // 0x10880 .. 0x1089E: Lo    [31] NABATAEAN LETTER FINAL ALEPH..NABATAEAN LETTER TAW
    Prop<::unicode::EastAsianWidth> { { 0x10857, 0x1089E }, ::unicode::EastAsianWidth::Neutral }, // #6
    Prop<::unicode::EastAsianWidth> { { 0x108A7, 0x108AF }, ::unicode::EastAsianWidth::Neutral }, // No     [9] NABATAEAN NUMBER ONE..NABATAEAN NUMBER ONE HUNDRED
    Prop<::unicode::EastAsianWidth> { { 0x108E0, 0x108F2 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [19] HATRAN LETTER ALEPH..HATRAN LETTER QOPH
    Prop<::unicode::EastAsianWidth> { { 0x108F4, 0x108F5 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] HATRAN LETTER SHIN..HATRAN LETTER TAW
    // 0x108FB .. 0x108FF: No     [5] HATRAN NUMBER ONE..HATRAN NUMBER ONE HUNDRED
    // 0x10900 .. 0x10915: Lo    [22] PHOENICIAN LETTER ALF..PHOENICIAN LETTER TAU
    // 0x10916 .. 0x1091B: No     [6] PHOENICIAN NUMBER ONE..PHOENICIAN NUMBER THREE
    Prop<::unicode::EastAsianWidth> { { 0x108FB, 0x1091B }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x1091F .. 0x1091F: Po         PHOENICIAN WORD SEPARATOR
    // 0x10920 .. 0x10939: Lo    [26] LYDIAN LETTER A..LYDIAN LETTER C
    Prop<::unicode::EastAsianWidth> { { 0x1091F, 0x10939 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x1093F .. 0x1093F: Po         LYDIAN TRIANGULAR MARK
    // 0x10940 .. 0x10959: Lo    [26] SIDETIC LETTER N01..SIDETIC LETTER N26
    Prop<::unicode::EastAsianWidth> { { 0x1093F, 0x10959 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x10980 .. 0x1099F: Lo    [32] MEROITIC HIEROGLYPHIC LETTER A..MEROITIC HIEROGLYPHIC SYMBOL VIDJ-2
    // 0x109A0 .. 0x109B7: Lo    [24] MEROITIC CURSIVE LETTER A..MEROITIC CURSIVE LETTER DA
    Prop<::unicode::EastAsianWidth> { { 0x10980, 0x109B7 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x109BC .. 0x109BD: No     [2] MEROITIC CURSIVE FRACTION ELEVEN TWELFTHS..MEROITIC CURSIVE FRACTION ONE HALF
    // 0x109BE .. 0x109BF: Lo     [2] MEROITIC CURSIVE LOGOGRAM RMT..MEROITIC CURSIVE LOGOGRAM IMN
    // 0x109C0 .. 0x109CF: No    [16] MEROITIC CURSIVE NUMBER ONE..MEROITIC CURSIVE NUMBER SEVENTY
    Prop<::unicode::EastAsianWidth> { { 0x109BC, 0x109CF }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x109D2 .. 0x109FF: No    [46] MEROITIC CURSIVE NUMBER ONE HUNDRED..MEROITIC CURSIVE FRACTION TEN TWELFTHS
    // 0x10A00 .. 0x10A00: Lo         KHAROSHTHI LETTER A
    // 0x10A01 .. 0x10A03: Mn     [3] KHAROSHTHI VOWEL SIGN I..KHAROSHTHI VOWEL SIGN VOCALIC R
    Prop<::unicode::EastAsianWidth> { { 0x109D2, 0x10A03 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x10A05, 0x10A06 }, ::unicode::EastAsianWidth::Neutral }, // Mn     [2] KHAROSHTHI VOWEL SIGN E..KHAROSHTHI VOWEL SIGN O
    // 0x10A0C .. 0x10A0F: Mn     [4] KHAROSHTHI VOWEL LENGTH MARK..KHAROSHTHI SIGN VISARGA
    // 0x10A10 .. 0x10A13: Lo     [4] KHAROSHTHI LETTER KA..KHAROSHTHI LETTER GHA
    Prop<::unicode::EastAsianWidth> { { 0x10A0C, 0x10A13 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x10A15, 0x10A17 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [3] KHAROSHTHI LETTER CA..KHAROSHTHI LETTER JA
    Prop<::unicode::EastAsianWidth> { { 0x10A19, 0x10A35 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [29] KHAROSHTHI LETTER NYA..KHAROSHTHI LETTER VHA
    Prop<::unicode::EastAsianWidth> { { 0x10A38, 0x10A3A }, ::unicode::EastAsianWidth::Neutral }, // Mn     [3] KHAROSHTHI SIGN BAR ABOVE..KHAROSHTHI SIGN DOT BELOW
    // 0x10A3F .. 0x10A3F: Mn         KHAROSHTHI VIRAMA
    // 0x10A40 .. 0x10A48: No     [9] KHAROSHTHI DIGIT ONE..KHAROSHTHI FRACTION ONE HALF
    Prop<::unicode::EastAsianWidth> { { 0x10A3F, 0x10A48 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x10A50, 0x10A58 }, ::unicode::EastAsianWidth::Neutral }, // Po     [9] KHAROSHTHI PUNCTUATION DOT..KHAROSHTHI PUNCTUATION LINES
    // 0x10A60 .. 0x10A7C: Lo    [29] OLD SOUTH ARABIAN LETTER HE..OLD SOUTH ARABIAN LETTER THETH
    // 0x10A7D .. 0x10A7E: No     [2] OLD SOUTH ARABIAN NUMBER ONE..OLD SOUTH ARABIAN NUMBER FIFTY
    // 0x10A7F .. 0x10A7F: Po         OLD SOUTH ARABIAN NUMERIC INDICATOR
    // 0x10A80 .. 0x10A9C: Lo    [29] OLD NORTH ARABIAN LETTER HEH..OLD NORTH ARABIAN LETTER ZAH
    // 0x10A9D .. 0x10A9F: No     [3] OLD NORTH ARABIAN NUMBER ONE..OLD NORTH ARABIAN NUMBER TWENTY
    Prop<::unicode::EastAsianWidth> { { 0x10A60, 0x10A9F }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x10AC0 .. 0x10AC7: Lo     [8] MANICHAEAN LETTER ALEPH..MANICHAEAN LETTER WAW
    // 0x10AC8 .. 0x10AC8: So         MANICHAEAN SIGN UD
    // 0x10AC9 .. 0x10AE4: Lo    [28] MANICHAEAN LETTER ZAYIN..MANICHAEAN LETTER TAW
    // 0x10AE5 .. 0x10AE6: Mn     [2] MANICHAEAN ABBREVIATION MARK ABOVE..MANICHAEAN ABBREVIATION MARK BELOW
    Prop<::unicode::EastAsianWidth> { { 0x10AC0, 0x10AE6 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x10AEB .. 0x10AEF: No     [5] MANICHAEAN NUMBER ONE..MANICHAEAN NUMBER ONE HUNDRED
    // 0x10AF0 .. 0x10AF6: Po     [7] MANICHAEAN PUNCTUATION STAR..MANICHAEAN PUNCTUATION LINE FILLER
    Prop<::unicode::EastAsianWidth> { { 0x10AEB, 0x10AF6 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x10B00, 0x10B35 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [54] AVESTAN LETTER A..AVESTAN LETTER HE
    // 0x10B39 .. 0x10B3F: Po     [7] AVESTAN ABBREVIATION MARK..LARGE ONE RING OVER TWO RINGS PUNCTUATION
    // 0x10B40 .. 0x10B55: Lo    [22] INSCRIPTIONAL PARTHIAN LETTER ALEPH..INSCRIPTIONAL PARTHIAN LETTER TAW
    Prop<::unicode::EastAsianWidth> { { 0x10B39, 0x10B55 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x10B58 .. 0x10B5F: No     [8] INSCRIPTIONAL PARTHIAN NUMBER ONE..INSCRIPTIONAL PARTHIAN NUMBER ONE THOUSAND
    // 0x10B60 .. 0x10B72: Lo    [19] INSCRIPTIONAL PAHLAVI LETTER ALEPH..INSCRIPTIONAL PAHLAVI LETTER TAW
    Prop<::unicode::EastAsianWidth> { { 0x10B58, 0x10B72 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x10B78 .. 0x10B7F: No     [8] INSCRIPTIONAL PAHLAVI NUMBER ONE..INSCRIPTIONAL PAHLAVI NUMBER ONE THOUSAND
    // 0x10B80 .. 0x10B91: Lo    [18] PSALTER PAHLAVI LETTER ALEPH..PSALTER PAHLAVI LETTER TAW
    Prop<::unicode::EastAsianWidth> { { 0x10B78, 0x10B91 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x10B99, 0x10B9C }, ::unicode::EastAsianWidth::Neutral }, // Po     [4] PSALTER PAHLAVI SECTION MARK..PSALTER PAHLAVI FOUR DOTS WITH DOT
    Prop<::unicode::EastAsianWidth> { { 0x10BA9, 0x10BAF }, ::unicode::EastAsianWidth::Neutral }, // No     [7] PSALTER PAHLAVI NUMBER ONE..PSALTER PAHLAVI NUMBER ONE HUNDRED
    Prop<::unicode::EastAsianWidth> { { 0x10C00, 0x10C48 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [73] OLD TURKIC LETTER ORKHON A..OLD TURKIC LETTER ORKHON BASH
    Prop<::unicode::EastAsianWidth> { { 0x10C80, 0x10CB2 }, ::unicode::EastAsianWidth::Neutral }, // Lu    [51] OLD HUNGARIAN CAPITAL LETTER A..OLD HUNGARIAN CAPITAL LETTER US
    Prop<::unicode::EastAsianWidth> { { 0x10CC0, 0x10CF2 }, ::unicode::EastAsianWidth::Neutral }, // Ll    [51] OLD HUNGARIAN SMALL LETTER A..OLD HUNGARIAN SMALL LETTER US
    // 0x10CFA .. 0x10CFF: No     [6] OLD HUNGARIAN NUMBER ONE..OLD HUNGARIAN NUMBER ONE THOUSAND
    // 0x10D00 .. 0x10D23: Lo    [36] HANIFI ROHINGYA LETTER A..HANIFI ROHINGYA MARK NA KHONNA
    // 0x10D24 .. 0x10D27: Mn     [4] HANIFI ROHINGYA SIGN HARBAHAY..HANIFI ROHINGYA SIGN TASSI
    Prop<::unicode::EastAsianWidth> { { 0x10CFA, 0x10D27 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x10D30, 0x10D39 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] HANIFI ROHINGYA DIGIT ZERO..HANIFI ROHINGYA DIGIT NINE
    // 0x10D40 .. 0x10D49: Nd    [10] GARAY DIGIT ZERO..GARAY DIGIT NINE
    // 0x10D4A .. 0x10D4D: Lo     [4] GARAY VOWEL SIGN A..GARAY VOWEL SIGN EE
    // 0x10D4E .. 0x10D4E: Lm         GARAY VOWEL LENGTH MARK
    // 0x10D4F .. 0x10D4F: Lo         GARAY SUKUN
    // 0x10D50 .. 0x10D65: Lu    [22] GARAY CAPITAL LETTER A..GARAY CAPITAL LETTER OLD NA
    Prop<::unicode::EastAsianWidth> { { 0x10D40, 0x10D65 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x10D69 .. 0x10D6D: Mn     [5] GARAY VOWEL SIGN E..GARAY CONSONANT NASALIZATION MARK
    // 0x10D6E .. 0x10D6E: Pd         GARAY HYPHEN
    // 0x10D6F .. 0x10D6F: Lm         GARAY REDUPLICATION MARK
    // 0x10D70 .. 0x10D85: Ll    [22] GARAY SMALL LETTER A..GARAY SMALL LETTER OLD NA
    Prop<::unicode::EastAsianWidth> { { 0x10D69, 0x10D85 }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x10D8E, 0x10D8F }, ::unicode::EastAsianWidth::Neutral }, // Sm     [2] GARAY PLUS SIGN..GARAY MINUS SIGN
    Prop<::unicode::EastAsianWidth> { { 0x10E60, 0x10E7E }, ::unicode::EastAsianWidth::Neutral }, // No    [31] RUMI DIGIT ONE..RUMI FRACTION TWO THIRDS
    Prop<::unicode::EastAsianWidth> { { 0x10E80, 0x10EA9 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [42] YEZIDI LETTER ELIF..YEZIDI LETTER ET
    // 0x10EAB .. 0x10EAC: Mn     [2] YEZIDI COMBINING HAMZA MARK..YEZIDI COMBINING MADDA MARK
    // 0x10EAD .. 0x10EAD: Pd         YEZIDI HYPHENATION MARK
    Prop<::unicode::EastAsianWidth> { { 0x10EAB, 0x10EAD }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x10EB0, 0x10EB1 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] YEZIDI LETTER LAM WITH DOT ABOVE..YEZIDI LETTER YOT WITH CIRCUMFLEX ABOVE
    // 0x10EC2 .. 0x10EC4: Lo     [3] ARABIC LETTER DAL WITH TWO DOTS VERTICALLY BELOW..ARABIC LETTER KAF WITH TWO DOTS VERTICALLY BELOW
    // 0x10EC5 .. 0x10EC5: Lm         ARABIC SMALL YEH BARREE WITH TWO DOTS BELOW
    // 0x10EC6 .. 0x10EC7: Lo     [2] ARABIC LETTER THIN NOON..ARABIC LETTER YEH WITH FOUR DOTS BELOW
    Prop<::unicode::EastAsianWidth> { { 0x10EC2, 0x10EC7 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x10ED0 .. 0x10ED0: Po         ARABIC BIBLICAL END OF VERSE
    // 0x10ED1 .. 0x10ED8: So     [8] ARABIC LIGATURE ALAYHAA AS-SALAATU WAS-SALAAM..ARABIC LIGATURE NAWWARA ALLAAHU MARQADAH
    Prop<::unicode::EastAsianWidth> { { 0x10ED0, 0x10ED8 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x10EFA .. 0x10EFF: Mn     [6] ARABIC DOUBLE VERTICAL BAR BELOW..ARABIC SMALL LOW WORD MADDA
    // 0x10F00 .. 0x10F1C: Lo    [29] OLD SOGDIAN LETTER ALEPH..OLD SOGDIAN LETTER FINAL TAW WITH VERTICAL TAIL
    // 0x10F1D .. 0x10F26: No    [10] OLD SOGDIAN NUMBER ONE..OLD SOGDIAN FRACTION ONE HALF
    // 0x10F27 .. 0x10F27: Lo         OLD SOGDIAN LIGATURE AYIN-DALETH
    Prop<::unicode::EastAsianWidth> { { 0x10EFA, 0x10F27 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x10F30 .. 0x10F45: Lo    [22] SOGDIAN LETTER ALEPH..SOGDIAN INDEPENDENT SHIN
    // 0x10F46 .. 0x10F50: Mn    [11] SOGDIAN COMBINING DOT BELOW..SOGDIAN COMBINING STROKE BELOW
    // 0x10F51 .. 0x10F54: No     [4] SOGDIAN NUMBER ONE..SOGDIAN NUMBER ONE HUNDRED
    // 0x10F55 .. 0x10F59: Po     [5] SOGDIAN PUNCTUATION TWO VERTICAL BARS..SOGDIAN PUNCTUATION HALF CIRCLE WITH DOT
    Prop<::unicode::EastAsianWidth> { { 0x10F30, 0x10F59 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x10F70 .. 0x10F81: Lo    [18] OLD UYGHUR LETTER ALEPH..OLD UYGHUR LETTER LESH
    // 0x10F82 .. 0x10F85: Mn     [4] OLD UYGHUR COMBINING DOT ABOVE..OLD UYGHUR COMBINING TWO DOTS BELOW
    // 0x10F86 .. 0x10F89: Po     [4] OLD UYGHUR PUNCTUATION BAR..OLD UYGHUR PUNCTUATION FOUR DOTS
    Prop<::unicode::EastAsianWidth> { { 0x10F70, 0x10F89 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x10FB0 .. 0x10FC4: Lo    [21] CHORASMIAN LETTER ALEPH..CHORASMIAN LETTER TAW
    // 0x10FC5 .. 0x10FCB: No     [7] CHORASMIAN NUMBER ONE..CHORASMIAN NUMBER ONE HUNDRED
    Prop<::unicode::EastAsianWidth> { { 0x10FB0, 0x10FCB }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x10FE0, 0x10FF6 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [23] ELYMAIC LETTER ALEPH..ELYMAIC LIGATURE ZAYIN-YODH
    // 0x11000 .. 0x11000: Mc         BRAHMI SIGN CANDRABINDU
    // 0x11001 .. 0x11001: Mn         BRAHMI SIGN ANUSVARA
    // 0x11002 .. 0x11002: Mc         BRAHMI SIGN VISARGA
    // 0x11003 .. 0x11037: Lo    [53] BRAHMI SIGN JIHVAMULIYA..BRAHMI LETTER OLD TAMIL NNNA
    // 0x11038 .. 0x11046: Mn    [15] BRAHMI VOWEL SIGN AA..BRAHMI VIRAMA
    // 0x11047 .. 0x1104D: Po     [7] BRAHMI DANDA..BRAHMI PUNCTUATION LOTUS
    Prop<::unicode::EastAsianWidth> { { 0x11000, 0x1104D }, ::unicode::EastAsianWidth::Neutral }, // #6
    // 0x11052 .. 0x11065: No    [20] BRAHMI NUMBER ONE..BRAHMI NUMBER ONE THOUSAND
    // 0x11066 .. 0x1106F: Nd    [10] BRAHMI DIGIT ZERO..BRAHMI DIGIT NINE
    // 0x11070 .. 0x11070: Mn         BRAHMI SIGN OLD TAMIL VIRAMA
    // 0x11071 .. 0x11072: Lo     [2] BRAHMI LETTER OLD TAMIL SHORT E..BRAHMI LETTER OLD TAMIL SHORT O
    // 0x11073 .. 0x11074: Mn     [2] BRAHMI VOWEL SIGN OLD TAMIL SHORT E..BRAHMI VOWEL SIGN OLD TAMIL SHORT O
    // 0x11075 .. 0x11075: Lo         BRAHMI LETTER OLD TAMIL LLA
    Prop<::unicode::EastAsianWidth> { { 0x11052, 0x11075 }, ::unicode::EastAsianWidth::Neutral }, // #6
    // 0x1107F .. 0x1107F: Mn         BRAHMI NUMBER JOINER
    // 0x11080 .. 0x11081: Mn     [2] KAITHI SIGN CANDRABINDU..KAITHI SIGN ANUSVARA
    // 0x11082 .. 0x11082: Mc         KAITHI SIGN VISARGA
    // 0x11083 .. 0x110AF: Lo    [45] KAITHI LETTER A..KAITHI LETTER HA
    // 0x110B0 .. 0x110B2: Mc     [3] KAITHI VOWEL SIGN AA..KAITHI VOWEL SIGN II
    // 0x110B3 .. 0x110B6: Mn     [4] KAITHI VOWEL SIGN U..KAITHI VOWEL SIGN AI
    // 0x110B7 .. 0x110B8: Mc     [2] KAITHI VOWEL SIGN O..KAITHI VOWEL SIGN AU
    // 0x110B9 .. 0x110BA: Mn     [2] KAITHI SIGN VIRAMA..KAITHI SIGN NUKTA
    // 0x110BB .. 0x110BC: Po     [2] KAITHI ABBREVIATION SIGN..KAITHI ENUMERATION SIGN
    // 0x110BD .. 0x110BD: Cf         KAITHI NUMBER SIGN
    // 0x110BE .. 0x110C1: Po     [4] KAITHI SECTION MARK..KAITHI DOUBLE DANDA
    // 0x110C2 .. 0x110C2: Mn         KAITHI VOWEL SIGN VOCALIC R
    Prop<::unicode::EastAsianWidth> { { 0x1107F, 0x110C2 }, ::unicode::EastAsianWidth::Neutral }, // #12
    Prop<::unicode::EastAsianWidth> { { 0x110CD, 0x110CD }, ::unicode::EastAsianWidth::Neutral }, // Cf         KAITHI NUMBER SIGN ABOVE
    Prop<::unicode::EastAsianWidth> { { 0x110D0, 0x110E8 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [25] SORA SOMPENG LETTER SAH..SORA SOMPENG LETTER MAE
    Prop<::unicode::EastAsianWidth> { { 0x110F0, 0x110F9 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] SORA SOMPENG DIGIT ZERO..SORA SOMPENG DIGIT NINE
    // 0x11100 .. 0x11102: Mn     [3] CHAKMA SIGN CANDRABINDU..CHAKMA SIGN VISARGA
    // 0x11103 .. 0x11126: Lo    [36] CHAKMA LETTER AA..CHAKMA LETTER HAA
    // 0x11127 .. 0x1112B: Mn     [5] CHAKMA VOWEL SIGN A..CHAKMA VOWEL SIGN UU
    // 0x1112C .. 0x1112C: Mc         CHAKMA VOWEL SIGN E
    // 0x1112D .. 0x11134: Mn     [8] CHAKMA VOWEL SIGN AI..CHAKMA MAAYYAA
    Prop<::unicode::EastAsianWidth> { { 0x11100, 0x11134 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x11136 .. 0x1113F: Nd    [10] CHAKMA DIGIT ZERO..CHAKMA DIGIT NINE
    // 0x11140 .. 0x11143: Po     [4] CHAKMA SECTION MARK..CHAKMA QUESTION MARK
    // 0x11144 .. 0x11144: Lo         CHAKMA LETTER LHAA
    // 0x11145 .. 0x11146: Mc     [2] CHAKMA VOWEL SIGN AA..CHAKMA VOWEL SIGN EI
    // 0x11147 .. 0x11147: Lo         CHAKMA LETTER VAA
    Prop<::unicode::EastAsianWidth> { { 0x11136, 0x11147 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x11150 .. 0x11172: Lo    [35] MAHAJANI LETTER A..MAHAJANI LETTER RRA
    // 0x11173 .. 0x11173: Mn         MAHAJANI SIGN NUKTA
    // 0x11174 .. 0x11175: Po     [2] MAHAJANI ABBREVIATION SIGN..MAHAJANI SECTION MARK
    // 0x11176 .. 0x11176: Lo         MAHAJANI LIGATURE SHRI
    Prop<::unicode::EastAsianWidth> { { 0x11150, 0x11176 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x11180 .. 0x11181: Mn     [2] SHARADA SIGN CANDRABINDU..SHARADA SIGN ANUSVARA
    // 0x11182 .. 0x11182: Mc         SHARADA SIGN VISARGA
    // 0x11183 .. 0x111B2: Lo    [48] SHARADA LETTER A..SHARADA LETTER HA
    // 0x111B3 .. 0x111B5: Mc     [3] SHARADA VOWEL SIGN AA..SHARADA VOWEL SIGN II
    // 0x111B6 .. 0x111BE: Mn     [9] SHARADA VOWEL SIGN U..SHARADA VOWEL SIGN O
    // 0x111BF .. 0x111C0: Mc     [2] SHARADA VOWEL SIGN AU..SHARADA SIGN VIRAMA
    // 0x111C1 .. 0x111C4: Lo     [4] SHARADA SIGN AVAGRAHA..SHARADA OM
    // 0x111C5 .. 0x111C8: Po     [4] SHARADA DANDA..SHARADA SEPARATOR
    // 0x111C9 .. 0x111CC: Mn     [4] SHARADA SANDHI MARK..SHARADA EXTRA SHORT VOWEL MARK
    // 0x111CD .. 0x111CD: Po         SHARADA SUTRA MARK
    // 0x111CE .. 0x111CE: Mc         SHARADA VOWEL SIGN PRISHTHAMATRA E
    // 0x111CF .. 0x111CF: Mn         SHARADA SIGN INVERTED CANDRABINDU
    // 0x111D0 .. 0x111D9: Nd    [10] SHARADA DIGIT ZERO..SHARADA DIGIT NINE
    // 0x111DA .. 0x111DA: Lo         SHARADA EKAM
    // 0x111DB .. 0x111DB: Po         SHARADA SIGN SIDDHAM
    // 0x111DC .. 0x111DC: Lo         SHARADA HEADSTROKE
    // 0x111DD .. 0x111DF: Po     [3] SHARADA CONTINUATION SIGN..SHARADA SECTION MARK-2
    Prop<::unicode::EastAsianWidth> { { 0x11180, 0x111DF }, ::unicode::EastAsianWidth::Neutral }, // #17
    Prop<::unicode::EastAsianWidth> { { 0x111E1, 0x111F4 }, ::unicode::EastAsianWidth::Neutral }, // No    [20] SINHALA ARCHAIC DIGIT ONE..SINHALA ARCHAIC NUMBER ONE THOUSAND
    Prop<::unicode::EastAsianWidth> { { 0x11200, 0x11211 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [18] KHOJKI LETTER A..KHOJKI LETTER JJA
    // 0x11213 .. 0x1122B: Lo    [25] KHOJKI LETTER NYA..KHOJKI LETTER LLA
    // 0x1122C .. 0x1122E: Mc     [3] KHOJKI VOWEL SIGN AA..KHOJKI VOWEL SIGN II
    // 0x1122F .. 0x11231: Mn     [3] KHOJKI VOWEL SIGN U..KHOJKI VOWEL SIGN AI
    // 0x11232 .. 0x11233: Mc     [2] KHOJKI VOWEL SIGN O..KHOJKI VOWEL SIGN AU
    // 0x11234 .. 0x11234: Mn         KHOJKI SIGN ANUSVARA
    // 0x11235 .. 0x11235: Mc         KHOJKI SIGN VIRAMA
    // 0x11236 .. 0x11237: Mn     [2] KHOJKI SIGN NUKTA..KHOJKI SIGN SHADDA
    // 0x11238 .. 0x1123D: Po     [6] KHOJKI DANDA..KHOJKI ABBREVIATION SIGN
    // 0x1123E .. 0x1123E: Mn         KHOJKI SIGN SUKUN
    // 0x1123F .. 0x11240: Lo     [2] KHOJKI LETTER QA..KHOJKI LETTER SHORT I
    // 0x11241 .. 0x11241: Mn         KHOJKI VOWEL SIGN VOCALIC R
    Prop<::unicode::EastAsianWidth> { { 0x11213, 0x11241 }, ::unicode::EastAsianWidth::Neutral }, // #11
    Prop<::unicode::EastAsianWidth> { { 0x11280, 0x11286 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] MULTANI LETTER A..MULTANI LETTER GA
    Prop<::unicode::EastAsianWidth> { { 0x11288, 0x11288 }, ::unicode::EastAsianWidth::Neutral }, // Lo         MULTANI LETTER GHA
    Prop<::unicode::EastAsianWidth> { { 0x1128A, 0x1128D }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] MULTANI LETTER CA..MULTANI LETTER JJA
    Prop<::unicode::EastAsianWidth> { { 0x1128F, 0x1129D }, ::unicode::EastAsianWidth::Neutral }, // Lo    [15] MULTANI LETTER NYA..MULTANI LETTER BA
    // 0x1129F .. 0x112A8: Lo    [10] MULTANI LETTER BHA..MULTANI LETTER RHA
    // 0x112A9 .. 0x112A9: Po         MULTANI SECTION MARK
    Prop<::unicode::EastAsianWidth> { { 0x1129F, 0x112A9 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x112B0 .. 0x112DE: Lo    [47] KHUDAWADI LETTER A..KHUDAWADI LETTER HA
    // 0x112DF .. 0x112DF: Mn         KHUDAWADI SIGN ANUSVARA
    // 0x112E0 .. 0x112E2: Mc     [3] KHUDAWADI VOWEL SIGN AA..KHUDAWADI VOWEL SIGN II
    // 0x112E3 .. 0x112EA: Mn     [8] KHUDAWADI VOWEL SIGN U..KHUDAWADI SIGN VIRAMA
    Prop<::unicode::EastAsianWidth> { { 0x112B0, 0x112EA }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x112F0, 0x112F9 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] KHUDAWADI DIGIT ZERO..KHUDAWADI DIGIT NINE
    // 0x11300 .. 0x11301: Mn     [2] GRANTHA SIGN COMBINING ANUSVARA ABOVE..GRANTHA SIGN CANDRABINDU
    // 0x11302 .. 0x11303: Mc     [2] GRANTHA SIGN ANUSVARA..GRANTHA SIGN VISARGA
    Prop<::unicode::EastAsianWidth> { { 0x11300, 0x11303 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x11305, 0x1130C }, ::unicode::EastAsianWidth::Neutral }, // Lo     [8] GRANTHA LETTER A..GRANTHA LETTER VOCALIC L
    Prop<::unicode::EastAsianWidth> { { 0x1130F, 0x11310 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] GRANTHA LETTER EE..GRANTHA LETTER AI
    Prop<::unicode::EastAsianWidth> { { 0x11313, 0x11328 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [22] GRANTHA LETTER OO..GRANTHA LETTER NA
    Prop<::unicode::EastAsianWidth> { { 0x1132A, 0x11330 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] GRANTHA LETTER PA..GRANTHA LETTER RA
    Prop<::unicode::EastAsianWidth> { { 0x11332, 0x11333 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] GRANTHA LETTER LA..GRANTHA LETTER LLA
    Prop<::unicode::EastAsianWidth> { { 0x11335, 0x11339 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [5] GRANTHA LETTER VA..GRANTHA LETTER HA
    // 0x1133B .. 0x1133C: Mn     [2] COMBINING BINDU BELOW..GRANTHA SIGN NUKTA
    // 0x1133D .. 0x1133D: Lo         GRANTHA SIGN AVAGRAHA
    // 0x1133E .. 0x1133F: Mc     [2] GRANTHA VOWEL SIGN AA..GRANTHA VOWEL SIGN I
    // 0x11340 .. 0x11340: Mn         GRANTHA VOWEL SIGN II
    // 0x11341 .. 0x11344: Mc     [4] GRANTHA VOWEL SIGN U..GRANTHA VOWEL SIGN VOCALIC RR
    Prop<::unicode::EastAsianWidth> { { 0x1133B, 0x11344 }, ::unicode::EastAsianWidth::Neutral }, // #5
    Prop<::unicode::EastAsianWidth> { { 0x11347, 0x11348 }, ::unicode::EastAsianWidth::Neutral }, // Mc     [2] GRANTHA VOWEL SIGN EE..GRANTHA VOWEL SIGN AI
    Prop<::unicode::EastAsianWidth> { { 0x1134B, 0x1134D }, ::unicode::EastAsianWidth::Neutral }, // Mc     [3] GRANTHA VOWEL SIGN OO..GRANTHA SIGN VIRAMA
    Prop<::unicode::EastAsianWidth> { { 0x11350, 0x11350 }, ::unicode::EastAsianWidth::Neutral }, // Lo         GRANTHA OM
    Prop<::unicode::EastAsianWidth> { { 0x11357, 0x11357 }, ::unicode::EastAsianWidth::Neutral }, // Mc         GRANTHA AU LENGTH MARK
    // 0x1135D .. 0x11361: Lo     [5] GRANTHA SIGN PLUTA..GRANTHA LETTER VOCALIC LL
    // 0x11362 .. 0x11363: Mc     [2] GRANTHA VOWEL SIGN VOCALIC L..GRANTHA VOWEL SIGN VOCALIC LL
    Prop<::unicode::EastAsianWidth> { { 0x1135D, 0x11363 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x11366, 0x1136C }, ::unicode::EastAsianWidth::Neutral }, // Mn     [7] COMBINING GRANTHA DIGIT ZERO..COMBINING GRANTHA DIGIT SIX
    Prop<::unicode::EastAsianWidth> { { 0x11370, 0x11374 }, ::unicode::EastAsianWidth::Neutral }, // Mn     [5] COMBINING GRANTHA LETTER A..COMBINING GRANTHA LETTER PA
    Prop<::unicode::EastAsianWidth> { { 0x11380, 0x11389 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [10] TULU-TIGALARI LETTER A..TULU-TIGALARI LETTER VOCALIC LL
    Prop<::unicode::EastAsianWidth> { { 0x1138B, 0x1138B }, ::unicode::EastAsianWidth::Neutral }, // Lo         TULU-TIGALARI LETTER EE
    Prop<::unicode::EastAsianWidth> { { 0x1138E, 0x1138E }, ::unicode::EastAsianWidth::Neutral }, // Lo         TULU-TIGALARI LETTER AI
    Prop<::unicode::EastAsianWidth> { { 0x11390, 0x113B5 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [38] TULU-TIGALARI LETTER OO..TULU-TIGALARI LETTER LLLA
    // 0x113B7 .. 0x113B7: Lo         TULU-TIGALARI SIGN AVAGRAHA
    // 0x113B8 .. 0x113BA: Mc     [3] TULU-TIGALARI VOWEL SIGN AA..TULU-TIGALARI VOWEL SIGN II
    // 0x113BB .. 0x113C0: Mn     [6] TULU-TIGALARI VOWEL SIGN U..TULU-TIGALARI VOWEL SIGN VOCALIC LL
    Prop<::unicode::EastAsianWidth> { { 0x113B7, 0x113C0 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x113C2, 0x113C2 }, ::unicode::EastAsianWidth::Neutral }, // Mc         TULU-TIGALARI VOWEL SIGN EE
    Prop<::unicode::EastAsianWidth> { { 0x113C5, 0x113C5 }, ::unicode::EastAsianWidth::Neutral }, // Mc         TULU-TIGALARI VOWEL SIGN AI
    Prop<::unicode::EastAsianWidth> { { 0x113C7, 0x113CA }, ::unicode::EastAsianWidth::Neutral }, // Mc     [4] TULU-TIGALARI VOWEL SIGN OO..TULU-TIGALARI SIGN CANDRA ANUNASIKA
    // 0x113CC .. 0x113CD: Mc     [2] TULU-TIGALARI SIGN ANUSVARA..TULU-TIGALARI SIGN VISARGA
    // 0x113CE .. 0x113CE: Mn         TULU-TIGALARI SIGN VIRAMA
    // 0x113CF .. 0x113CF: Mc         TULU-TIGALARI SIGN LOOPED VIRAMA
    // 0x113D0 .. 0x113D0: Mn         TULU-TIGALARI CONJOINER
    // 0x113D1 .. 0x113D1: Lo         TULU-TIGALARI REPHA
    // 0x113D2 .. 0x113D2: Mn         TULU-TIGALARI GEMINATION MARK
    // 0x113D3 .. 0x113D3: Lo         TULU-TIGALARI SIGN PLUTA
    // 0x113D4 .. 0x113D5: Po     [2] TULU-TIGALARI DANDA..TULU-TIGALARI DOUBLE DANDA
    Prop<::unicode::EastAsianWidth> { { 0x113CC, 0x113D5 }, ::unicode::EastAsianWidth::Neutral }, // #8
    Prop<::unicode::EastAsianWidth> { { 0x113D7, 0x113D8 }, ::unicode::EastAsianWidth::Neutral }, // Po     [2] TULU-TIGALARI SIGN OM PUSHPIKA..TULU-TIGALARI SIGN SHRII PUSHPIKA
    Prop<::unicode::EastAsianWidth> { { 0x113E1, 0x113E2 }, ::unicode::EastAsianWidth::Neutral }, // Mn     [2] TULU-TIGALARI VEDIC TONE SVARITA..TULU-TIGALARI VEDIC TONE ANUDATTA
    // 0x11400 .. 0x11434: Lo    [53] NEWA LETTER A..NEWA LETTER HA
    // 0x11435 .. 0x11437: Mc     [3] NEWA VOWEL SIGN AA..NEWA VOWEL SIGN II
    // 0x11438 .. 0x1143F: Mn     [8] NEWA VOWEL SIGN U..NEWA VOWEL SIGN AI
    // 0x11440 .. 0x11441: Mc     [2] NEWA VOWEL SIGN O..NEWA VOWEL SIGN AU
    // 0x11442 .. 0x11444: Mn     [3] NEWA SIGN VIRAMA..NEWA SIGN ANUSVARA
    // 0x11445 .. 0x11445: Mc         NEWA SIGN VISARGA
    // 0x11446 .. 0x11446: Mn         NEWA SIGN NUKTA
    // 0x11447 .. 0x1144A: Lo     [4] NEWA SIGN AVAGRAHA..NEWA SIDDHI
    // 0x1144B .. 0x1144F: Po     [5] NEWA DANDA..NEWA ABBREVIATION SIGN
    // 0x11450 .. 0x11459: Nd    [10] NEWA DIGIT ZERO..NEWA DIGIT NINE
    // 0x1145A .. 0x1145B: Po     [2] NEWA DOUBLE COMMA..NEWA PLACEHOLDER MARK
    Prop<::unicode::EastAsianWidth> { { 0x11400, 0x1145B }, ::unicode::EastAsianWidth::Neutral }, // #11
    // 0x1145D .. 0x1145D: Po         NEWA INSERTION SIGN
    // 0x1145E .. 0x1145E: Mn         NEWA SANDHI MARK
    // 0x1145F .. 0x11461: Lo     [3] NEWA LETTER VEDIC ANUSVARA..NEWA SIGN UPADHMANIYA
    Prop<::unicode::EastAsianWidth> { { 0x1145D, 0x11461 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x11480 .. 0x114AF: Lo    [48] TIRHUTA ANJI..TIRHUTA LETTER HA
    // 0x114B0 .. 0x114B2: Mc     [3] TIRHUTA VOWEL SIGN AA..TIRHUTA VOWEL SIGN II
    // 0x114B3 .. 0x114B8: Mn     [6] TIRHUTA VOWEL SIGN U..TIRHUTA VOWEL SIGN VOCALIC LL
    // 0x114B9 .. 0x114B9: Mc         TIRHUTA VOWEL SIGN E
    // 0x114BA .. 0x114BA: Mn         TIRHUTA VOWEL SIGN SHORT E
    // 0x114BB .. 0x114BE: Mc     [4] TIRHUTA VOWEL SIGN AI..TIRHUTA VOWEL SIGN AU
    // 0x114BF .. 0x114C0: Mn     [2] TIRHUTA SIGN CANDRABINDU..TIRHUTA SIGN ANUSVARA
    // 0x114C1 .. 0x114C1: Mc         TIRHUTA SIGN VISARGA
    // 0x114C2 .. 0x114C3: Mn     [2] TIRHUTA SIGN VIRAMA..TIRHUTA SIGN NUKTA
    // 0x114C4 .. 0x114C5: Lo     [2] TIRHUTA SIGN AVAGRAHA..TIRHUTA GVANG
    // 0x114C6 .. 0x114C6: Po         TIRHUTA ABBREVIATION SIGN
    // 0x114C7 .. 0x114C7: Lo         TIRHUTA OM
    Prop<::unicode::EastAsianWidth> { { 0x11480, 0x114C7 }, ::unicode::EastAsianWidth::Neutral }, // #12
    Prop<::unicode::EastAsianWidth> { { 0x114D0, 0x114D9 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] TIRHUTA DIGIT ZERO..TIRHUTA DIGIT NINE
    // 0x11580 .. 0x115AE: Lo    [47] SIDDHAM LETTER A..SIDDHAM LETTER HA
    // 0x115AF .. 0x115B1: Mc     [3] SIDDHAM VOWEL SIGN AA..SIDDHAM VOWEL SIGN II
    // 0x115B2 .. 0x115B5: Mn     [4] SIDDHAM VOWEL SIGN U..SIDDHAM VOWEL SIGN VOCALIC RR
    Prop<::unicode::EastAsianWidth> { { 0x11580, 0x115B5 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x115B8 .. 0x115BB: Mc     [4] SIDDHAM VOWEL SIGN E..SIDDHAM VOWEL SIGN AU
    // 0x115BC .. 0x115BD: Mn     [2] SIDDHAM SIGN CANDRABINDU..SIDDHAM SIGN ANUSVARA
    // 0x115BE .. 0x115BE: Mc         SIDDHAM SIGN VISARGA
    // 0x115BF .. 0x115C0: Mn     [2] SIDDHAM SIGN VIRAMA..SIDDHAM SIGN NUKTA
    // 0x115C1 .. 0x115D7: Po    [23] SIDDHAM SIGN SIDDHAM..SIDDHAM SECTION MARK WITH CIRCLES AND FOUR ENCLOSURES
    // 0x115D8 .. 0x115DB: Lo     [4] SIDDHAM LETTER THREE-CIRCLE ALTERNATE I..SIDDHAM LETTER ALTERNATE U
    // 0x115DC .. 0x115DD: Mn     [2] SIDDHAM VOWEL SIGN ALTERNATE U..SIDDHAM VOWEL SIGN ALTERNATE UU
    Prop<::unicode::EastAsianWidth> { { 0x115B8, 0x115DD }, ::unicode::EastAsianWidth::Neutral }, // #7
    // 0x11600 .. 0x1162F: Lo    [48] MODI LETTER A..MODI LETTER LLA
    // 0x11630 .. 0x11632: Mc     [3] MODI VOWEL SIGN AA..MODI VOWEL SIGN II
    // 0x11633 .. 0x1163A: Mn     [8] MODI VOWEL SIGN U..MODI VOWEL SIGN AI
    // 0x1163B .. 0x1163C: Mc     [2] MODI VOWEL SIGN O..MODI VOWEL SIGN AU
    // 0x1163D .. 0x1163D: Mn         MODI SIGN ANUSVARA
    // 0x1163E .. 0x1163E: Mc         MODI SIGN VISARGA
    // 0x1163F .. 0x11640: Mn     [2] MODI SIGN VIRAMA..MODI SIGN ARDHACANDRA
    // 0x11641 .. 0x11643: Po     [3] MODI DANDA..MODI ABBREVIATION SIGN
    // 0x11644 .. 0x11644: Lo         MODI SIGN HUVA
    Prop<::unicode::EastAsianWidth> { { 0x11600, 0x11644 }, ::unicode::EastAsianWidth::Neutral }, // #9
    Prop<::unicode::EastAsianWidth> { { 0x11650, 0x11659 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] MODI DIGIT ZERO..MODI DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x11660, 0x1166C }, ::unicode::EastAsianWidth::Neutral }, // Po    [13] MONGOLIAN BIRGA WITH ORNAMENT..MONGOLIAN TURNED SWIRL BIRGA WITH DOUBLE ORNAMENT
    // 0x11680 .. 0x116AA: Lo    [43] TAKRI LETTER A..TAKRI LETTER RRA
    // 0x116AB .. 0x116AB: Mn         TAKRI SIGN ANUSVARA
    // 0x116AC .. 0x116AC: Mc         TAKRI SIGN VISARGA
    // 0x116AD .. 0x116AD: Mn         TAKRI VOWEL SIGN AA
    // 0x116AE .. 0x116AF: Mc     [2] TAKRI VOWEL SIGN I..TAKRI VOWEL SIGN II
    // 0x116B0 .. 0x116B5: Mn     [6] TAKRI VOWEL SIGN U..TAKRI VOWEL SIGN AU
    // 0x116B6 .. 0x116B6: Mc         TAKRI SIGN VIRAMA
    // 0x116B7 .. 0x116B7: Mn         TAKRI SIGN NUKTA
    // 0x116B8 .. 0x116B8: Lo         TAKRI LETTER ARCHAIC KHA
    // 0x116B9 .. 0x116B9: Po         TAKRI ABBREVIATION SIGN
    Prop<::unicode::EastAsianWidth> { { 0x11680, 0x116B9 }, ::unicode::EastAsianWidth::Neutral }, // #10
    Prop<::unicode::EastAsianWidth> { { 0x116C0, 0x116C9 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] TAKRI DIGIT ZERO..TAKRI DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x116D0, 0x116E3 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [20] MYANMAR PAO DIGIT ZERO..MYANMAR EASTERN PWO KAREN DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x11700, 0x1171A }, ::unicode::EastAsianWidth::Neutral }, // Lo    [27] AHOM LETTER KA..AHOM LETTER ALTERNATE BA
    // 0x1171D .. 0x1171D: Mn         AHOM CONSONANT SIGN MEDIAL LA
    // 0x1171E .. 0x1171E: Mc         AHOM CONSONANT SIGN MEDIAL RA
    // 0x1171F .. 0x1171F: Mn         AHOM CONSONANT SIGN MEDIAL LIGATING RA
    // 0x11720 .. 0x11721: Mc     [2] AHOM VOWEL SIGN A..AHOM VOWEL SIGN AA
    // 0x11722 .. 0x11725: Mn     [4] AHOM VOWEL SIGN I..AHOM VOWEL SIGN UU
    // 0x11726 .. 0x11726: Mc         AHOM VOWEL SIGN E
    // 0x11727 .. 0x1172B: Mn     [5] AHOM VOWEL SIGN AW..AHOM SIGN KILLER
    Prop<::unicode::EastAsianWidth> { { 0x1171D, 0x1172B }, ::unicode::EastAsianWidth::Neutral }, // #7
    // 0x11730 .. 0x11739: Nd    [10] AHOM DIGIT ZERO..AHOM DIGIT NINE
    // 0x1173A .. 0x1173B: No     [2] AHOM NUMBER TEN..AHOM NUMBER TWENTY
    // 0x1173C .. 0x1173E: Po     [3] AHOM SIGN SMALL SECTION..AHOM SIGN RULAI
    // 0x1173F .. 0x1173F: So         AHOM SYMBOL VI
    // 0x11740 .. 0x11746: Lo     [7] AHOM LETTER CA..AHOM LETTER LLA
    Prop<::unicode::EastAsianWidth> { { 0x11730, 0x11746 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x11800 .. 0x1182B: Lo    [44] DOGRA LETTER A..DOGRA LETTER RRA
    // 0x1182C .. 0x1182E: Mc     [3] DOGRA VOWEL SIGN AA..DOGRA VOWEL SIGN II
    // 0x1182F .. 0x11837: Mn     [9] DOGRA VOWEL SIGN U..DOGRA SIGN ANUSVARA
    // 0x11838 .. 0x11838: Mc         DOGRA SIGN VISARGA
    // 0x11839 .. 0x1183A: Mn     [2] DOGRA SIGN VIRAMA..DOGRA SIGN NUKTA
    // 0x1183B .. 0x1183B: Po         DOGRA ABBREVIATION SIGN
    Prop<::unicode::EastAsianWidth> { { 0x11800, 0x1183B }, ::unicode::EastAsianWidth::Neutral }, // #6
    // 0x118A0 .. 0x118DF: L&    [64] WARANG CITI CAPITAL LETTER NGAA..WARANG CITI SMALL LETTER VIYO
    // 0x118E0 .. 0x118E9: Nd    [10] WARANG CITI DIGIT ZERO..WARANG CITI DIGIT NINE
    // 0x118EA .. 0x118F2: No     [9] WARANG CITI NUMBER TEN..WARANG CITI NUMBER NINETY
    Prop<::unicode::EastAsianWidth> { { 0x118A0, 0x118F2 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x118FF .. 0x118FF: Lo         WARANG CITI OM
    // 0x11900 .. 0x11906: Lo     [7] DIVES AKURU LETTER A..DIVES AKURU LETTER E
    Prop<::unicode::EastAsianWidth> { { 0x118FF, 0x11906 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x11909, 0x11909 }, ::unicode::EastAsianWidth::Neutral }, // Lo         DIVES AKURU LETTER O
    Prop<::unicode::EastAsianWidth> { { 0x1190C, 0x11913 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [8] DIVES AKURU LETTER KA..DIVES AKURU LETTER JA
    Prop<::unicode::EastAsianWidth> { { 0x11915, 0x11916 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] DIVES AKURU LETTER NYA..DIVES AKURU LETTER TTA
    // 0x11918 .. 0x1192F: Lo    [24] DIVES AKURU LETTER DDA..DIVES AKURU LETTER ZA
    // 0x11930 .. 0x11935: Mc     [6] DIVES AKURU VOWEL SIGN AA..DIVES AKURU VOWEL SIGN E
    Prop<::unicode::EastAsianWidth> { { 0x11918, 0x11935 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x11937, 0x11938 }, ::unicode::EastAsianWidth::Neutral }, // Mc     [2] DIVES AKURU VOWEL SIGN AI..DIVES AKURU VOWEL SIGN O
    // 0x1193B .. 0x1193C: Mn     [2] DIVES AKURU SIGN ANUSVARA..DIVES AKURU SIGN CANDRABINDU
    // 0x1193D .. 0x1193D: Mc         DIVES AKURU SIGN HALANTA
    // 0x1193E .. 0x1193E: Mn         DIVES AKURU VIRAMA
    // 0x1193F .. 0x1193F: Lo         DIVES AKURU PREFIXED NASAL SIGN
    // 0x11940 .. 0x11940: Mc         DIVES AKURU MEDIAL YA
    // 0x11941 .. 0x11941: Lo         DIVES AKURU INITIAL RA
    // 0x11942 .. 0x11942: Mc         DIVES AKURU MEDIAL RA
    // 0x11943 .. 0x11943: Mn         DIVES AKURU SIGN NUKTA
    // 0x11944 .. 0x11946: Po     [3] DIVES AKURU DOUBLE DANDA..DIVES AKURU END OF TEXT MARK
    Prop<::unicode::EastAsianWidth> { { 0x1193B, 0x11946 }, ::unicode::EastAsianWidth::Neutral }, // #9
    Prop<::unicode::EastAsianWidth> { { 0x11950, 0x11959 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] DIVES AKURU DIGIT ZERO..DIVES AKURU DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x119A0, 0x119A7 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [8] NANDINAGARI LETTER A..NANDINAGARI LETTER VOCALIC RR
    // 0x119AA .. 0x119D0: Lo    [39] NANDINAGARI LETTER E..NANDINAGARI LETTER RRA
    // 0x119D1 .. 0x119D3: Mc     [3] NANDINAGARI VOWEL SIGN AA..NANDINAGARI VOWEL SIGN II
    // 0x119D4 .. 0x119D7: Mn     [4] NANDINAGARI VOWEL SIGN U..NANDINAGARI VOWEL SIGN VOCALIC RR
    Prop<::unicode::EastAsianWidth> { { 0x119AA, 0x119D7 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x119DA .. 0x119DB: Mn     [2] NANDINAGARI VOWEL SIGN E..NANDINAGARI VOWEL SIGN AI
    // 0x119DC .. 0x119DF: Mc     [4] NANDINAGARI VOWEL SIGN O..NANDINAGARI SIGN VISARGA
    // 0x119E0 .. 0x119E0: Mn         NANDINAGARI SIGN VIRAMA
    // 0x119E1 .. 0x119E1: Lo         NANDINAGARI SIGN AVAGRAHA
    // 0x119E2 .. 0x119E2: Po         NANDINAGARI SIGN SIDDHAM
    // 0x119E3 .. 0x119E3: Lo         NANDINAGARI HEADSTROKE
    // 0x119E4 .. 0x119E4: Mc         NANDINAGARI VOWEL SIGN PRISHTHAMATRA E
    Prop<::unicode::EastAsianWidth> { { 0x119DA, 0x119E4 }, ::unicode::EastAsianWidth::Neutral }, // #7
    // 0x11A00 .. 0x11A00: Lo         ZANABAZAR SQUARE LETTER A
    // 0x11A01 .. 0x11A0A: Mn    [10] ZANABAZAR SQUARE VOWEL SIGN I..ZANABAZAR SQUARE VOWEL LENGTH MARK
    // 0x11A0B .. 0x11A32: Lo    [40] ZANABAZAR SQUARE LETTER KA..ZANABAZAR SQUARE LETTER KSSA
    // 0x11A33 .. 0x11A38: Mn     [6] ZANABAZAR SQUARE FINAL CONSONANT MARK..ZANABAZAR SQUARE SIGN ANUSVARA
    // 0x11A39 .. 0x11A39: Mc         ZANABAZAR SQUARE SIGN VISARGA
    // 0x11A3A .. 0x11A3A: Lo         ZANABAZAR SQUARE CLUSTER-INITIAL LETTER RA
    // 0x11A3B .. 0x11A3E: Mn     [4] ZANABAZAR SQUARE CLUSTER-FINAL LETTER YA..ZANABAZAR SQUARE CLUSTER-FINAL LETTER VA
    // 0x11A3F .. 0x11A46: Po     [8] ZANABAZAR SQUARE INITIAL HEAD MARK..ZANABAZAR SQUARE CLOSING DOUBLE-LINED HEAD MARK
    // 0x11A47 .. 0x11A47: Mn         ZANABAZAR SQUARE SUBJOINER
    Prop<::unicode::EastAsianWidth> { { 0x11A00, 0x11A47 }, ::unicode::EastAsianWidth::Neutral }, // #9
    // 0x11A50 .. 0x11A50: Lo         SOYOMBO LETTER A
    // 0x11A51 .. 0x11A56: Mn     [6] SOYOMBO VOWEL SIGN I..SOYOMBO VOWEL SIGN OE
    // 0x11A57 .. 0x11A58: Mc     [2] SOYOMBO VOWEL SIGN AI..SOYOMBO VOWEL SIGN AU
    // 0x11A59 .. 0x11A5B: Mn     [3] SOYOMBO VOWEL SIGN VOCALIC R..SOYOMBO VOWEL LENGTH MARK
    // 0x11A5C .. 0x11A89: Lo    [46] SOYOMBO LETTER KA..SOYOMBO CLUSTER-INITIAL LETTER SA
    // 0x11A8A .. 0x11A96: Mn    [13] SOYOMBO FINAL CONSONANT SIGN G..SOYOMBO SIGN ANUSVARA
    // 0x11A97 .. 0x11A97: Mc         SOYOMBO SIGN VISARGA
    // 0x11A98 .. 0x11A99: Mn     [2] SOYOMBO GEMINATION MARK..SOYOMBO SUBJOINER
    // 0x11A9A .. 0x11A9C: Po     [3] SOYOMBO MARK TSHEG..SOYOMBO MARK DOUBLE SHAD
    // 0x11A9D .. 0x11A9D: Lo         SOYOMBO MARK PLUTA
    // 0x11A9E .. 0x11AA2: Po     [5] SOYOMBO HEAD MARK WITH MOON AND SUN AND TRIPLE FLAME..SOYOMBO TERMINAL MARK-2
    Prop<::unicode::EastAsianWidth> { { 0x11A50, 0x11AA2 }, ::unicode::EastAsianWidth::Neutral }, // #11
    // 0x11AB0 .. 0x11ABF: Lo    [16] CANADIAN SYLLABICS NATTILIK HI..CANADIAN SYLLABICS SPA
    // 0x11AC0 .. 0x11AF8: Lo    [57] PAU CIN HAU LETTER PA..PAU CIN HAU GLOTTAL STOP FINAL
    Prop<::unicode::EastAsianWidth> { { 0x11AB0, 0x11AF8 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x11B00, 0x11B09 }, ::unicode::EastAsianWidth::Neutral }, // Po    [10] DEVANAGARI HEAD MARK..DEVANAGARI SIGN MINDU
    // 0x11B60 .. 0x11B60: Mn         SHARADA VOWEL SIGN OE
    // 0x11B61 .. 0x11B61: Mc         SHARADA VOWEL SIGN OOE
    // 0x11B62 .. 0x11B64: Mn     [3] SHARADA VOWEL SIGN UE..SHARADA VOWEL SIGN SHORT E
    // 0x11B65 .. 0x11B65: Mc         SHARADA VOWEL SIGN SHORT O
    // 0x11B66 .. 0x11B66: Mn         SHARADA VOWEL SIGN CANDRA E
    // 0x11B67 .. 0x11B67: Mc         SHARADA VOWEL SIGN CANDRA O
    Prop<::unicode::EastAsianWidth> { { 0x11B60, 0x11B67 }, ::unicode::EastAsianWidth::Neutral }, // #6
    // 0x11BC0 .. 0x11BE0: Lo    [33] SUNUWAR LETTER DEVI..SUNUWAR LETTER KLOKO
    // 0x11BE1 .. 0x11BE1: Po         SUNUWAR SIGN PVO
    Prop<::unicode::EastAsianWidth> { { 0x11BC0, 0x11BE1 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x11BF0, 0x11BF9 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] SUNUWAR DIGIT ZERO..SUNUWAR DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x11C00, 0x11C08 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [9] BHAIKSUKI LETTER A..BHAIKSUKI LETTER VOCALIC L
    // 0x11C0A .. 0x11C2E: Lo    [37] BHAIKSUKI LETTER E..BHAIKSUKI LETTER HA
    // 0x11C2F .. 0x11C2F: Mc         BHAIKSUKI VOWEL SIGN AA
    // 0x11C30 .. 0x11C36: Mn     [7] BHAIKSUKI VOWEL SIGN I..BHAIKSUKI VOWEL SIGN VOCALIC L
    Prop<::unicode::EastAsianWidth> { { 0x11C0A, 0x11C36 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x11C38 .. 0x11C3D: Mn     [6] BHAIKSUKI VOWEL SIGN E..BHAIKSUKI SIGN ANUSVARA
    // 0x11C3E .. 0x11C3E: Mc         BHAIKSUKI SIGN VISARGA
    // 0x11C3F .. 0x11C3F: Mn         BHAIKSUKI SIGN VIRAMA
    // 0x11C40 .. 0x11C40: Lo         BHAIKSUKI SIGN AVAGRAHA
    // 0x11C41 .. 0x11C45: Po     [5] BHAIKSUKI DANDA..BHAIKSUKI GAP FILLER-2
    Prop<::unicode::EastAsianWidth> { { 0x11C38, 0x11C45 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x11C50 .. 0x11C59: Nd    [10] BHAIKSUKI DIGIT ZERO..BHAIKSUKI DIGIT NINE
    // 0x11C5A .. 0x11C6C: No    [19] BHAIKSUKI NUMBER ONE..BHAIKSUKI HUNDREDS UNIT MARK
    Prop<::unicode::EastAsianWidth> { { 0x11C50, 0x11C6C }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x11C70 .. 0x11C71: Po     [2] MARCHEN HEAD MARK..MARCHEN MARK SHAD
    // 0x11C72 .. 0x11C8F: Lo    [30] MARCHEN LETTER KA..MARCHEN LETTER A
    Prop<::unicode::EastAsianWidth> { { 0x11C70, 0x11C8F }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x11C92, 0x11CA7 }, ::unicode::EastAsianWidth::Neutral }, // Mn    [22] MARCHEN SUBJOINED LETTER KA..MARCHEN SUBJOINED LETTER ZA
    // 0x11CA9 .. 0x11CA9: Mc         MARCHEN SUBJOINED LETTER YA
    // 0x11CAA .. 0x11CB0: Mn     [7] MARCHEN SUBJOINED LETTER RA..MARCHEN VOWEL SIGN AA
    // 0x11CB1 .. 0x11CB1: Mc         MARCHEN VOWEL SIGN I
    // 0x11CB2 .. 0x11CB3: Mn     [2] MARCHEN VOWEL SIGN U..MARCHEN VOWEL SIGN E
    // 0x11CB4 .. 0x11CB4: Mc         MARCHEN VOWEL SIGN O
    // 0x11CB5 .. 0x11CB6: Mn     [2] MARCHEN SIGN ANUSVARA..MARCHEN SIGN CANDRABINDU
    Prop<::unicode::EastAsianWidth> { { 0x11CA9, 0x11CB6 }, ::unicode::EastAsianWidth::Neutral }, // #6
    Prop<::unicode::EastAsianWidth> { { 0x11D00, 0x11D06 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] MASARAM GONDI LETTER A..MASARAM GONDI LETTER E
    Prop<::unicode::EastAsianWidth> { { 0x11D08, 0x11D09 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] MASARAM GONDI LETTER AI..MASARAM GONDI LETTER O
    // 0x11D0B .. 0x11D30: Lo    [38] MASARAM GONDI LETTER AU..MASARAM GONDI LETTER TRA
    // 0x11D31 .. 0x11D36: Mn     [6] MASARAM GONDI VOWEL SIGN AA..MASARAM GONDI VOWEL SIGN VOCALIC R
    Prop<::unicode::EastAsianWidth> { { 0x11D0B, 0x11D36 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x11D3A, 0x11D3A }, ::unicode::EastAsianWidth::Neutral }, // Mn         MASARAM GONDI VOWEL SIGN E
    Prop<::unicode::EastAsianWidth> { { 0x11D3C, 0x11D3D }, ::unicode::EastAsianWidth::Neutral }, // Mn     [2] MASARAM GONDI VOWEL SIGN AI..MASARAM GONDI VOWEL SIGN O
    // 0x11D3F .. 0x11D45: Mn     [7] MASARAM GONDI VOWEL SIGN AU..MASARAM GONDI VIRAMA
    // 0x11D46 .. 0x11D46: Lo         MASARAM GONDI REPHA
    // 0x11D47 .. 0x11D47: Mn         MASARAM GONDI RA-KARA
    Prop<::unicode::EastAsianWidth> { { 0x11D3F, 0x11D47 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x11D50, 0x11D59 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] MASARAM GONDI DIGIT ZERO..MASARAM GONDI DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x11D60, 0x11D65 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [6] GUNJALA GONDI LETTER A..GUNJALA GONDI LETTER UU
    Prop<::unicode::EastAsianWidth> { { 0x11D67, 0x11D68 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] GUNJALA GONDI LETTER EE..GUNJALA GONDI LETTER AI
    // 0x11D6A .. 0x11D89: Lo    [32] GUNJALA GONDI LETTER OO..GUNJALA GONDI LETTER SA
    // 0x11D8A .. 0x11D8E: Mc     [5] GUNJALA GONDI VOWEL SIGN AA..GUNJALA GONDI VOWEL SIGN UU
    Prop<::unicode::EastAsianWidth> { { 0x11D6A, 0x11D8E }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x11D90, 0x11D91 }, ::unicode::EastAsianWidth::Neutral }, // Mn     [2] GUNJALA GONDI VOWEL SIGN EE..GUNJALA GONDI VOWEL SIGN AI
    // 0x11D93 .. 0x11D94: Mc     [2] GUNJALA GONDI VOWEL SIGN OO..GUNJALA GONDI VOWEL SIGN AU
    // 0x11D95 .. 0x11D95: Mn         GUNJALA GONDI SIGN ANUSVARA
    // 0x11D96 .. 0x11D96: Mc         GUNJALA GONDI SIGN VISARGA
    // 0x11D97 .. 0x11D97: Mn         GUNJALA GONDI VIRAMA
    // 0x11D98 .. 0x11D98: Lo         GUNJALA GONDI OM
    Prop<::unicode::EastAsianWidth> { { 0x11D93, 0x11D98 }, ::unicode::EastAsianWidth::Neutral }, // #5
    Prop<::unicode::EastAsianWidth> { { 0x11DA0, 0x11DA9 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] GUNJALA GONDI DIGIT ZERO..GUNJALA GONDI DIGIT NINE
    // 0x11DB0 .. 0x11DD8: Lo    [41] TOLONG SIKI LETTER I..TOLONG SIKI LETTER RRH
    // 0x11DD9 .. 0x11DD9: Lm         TOLONG SIKI SIGN SELA
    // 0x11DDA .. 0x11DDB: Lo     [2] TOLONG SIKI SIGN HECAKA..TOLONG SIKI UNGGA
    Prop<::unicode::EastAsianWidth> { { 0x11DB0, 0x11DDB }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x11DE0, 0x11DE9 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] TOLONG SIKI DIGIT ZERO..TOLONG SIKI DIGIT NINE
    // 0x11EE0 .. 0x11EF2: Lo    [19] MAKASAR LETTER KA..MAKASAR ANGKA
    // 0x11EF3 .. 0x11EF4: Mn     [2] MAKASAR VOWEL SIGN I..MAKASAR VOWEL SIGN U
    // 0x11EF5 .. 0x11EF6: Mc     [2] MAKASAR VOWEL SIGN E..MAKASAR VOWEL SIGN O
    // 0x11EF7 .. 0x11EF8: Po     [2] MAKASAR PASSIMBANG..MAKASAR END OF SECTION
    Prop<::unicode::EastAsianWidth> { { 0x11EE0, 0x11EF8 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x11F00 .. 0x11F01: Mn     [2] KAWI SIGN CANDRABINDU..KAWI SIGN ANUSVARA
    // 0x11F02 .. 0x11F02: Lo         KAWI SIGN REPHA
    // 0x11F03 .. 0x11F03: Mc         KAWI SIGN VISARGA
    // 0x11F04 .. 0x11F10: Lo    [13] KAWI LETTER A..KAWI LETTER O
    Prop<::unicode::EastAsianWidth> { { 0x11F00, 0x11F10 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x11F12 .. 0x11F33: Lo    [34] KAWI LETTER KA..KAWI LETTER JNYA
    // 0x11F34 .. 0x11F35: Mc     [2] KAWI VOWEL SIGN AA..KAWI VOWEL SIGN ALTERNATE AA
    // 0x11F36 .. 0x11F3A: Mn     [5] KAWI VOWEL SIGN I..KAWI VOWEL SIGN VOCALIC R
    Prop<::unicode::EastAsianWidth> { { 0x11F12, 0x11F3A }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x11F3E .. 0x11F3F: Mc     [2] KAWI VOWEL SIGN E..KAWI VOWEL SIGN AI
    // 0x11F40 .. 0x11F40: Mn         KAWI VOWEL SIGN EU
    // 0x11F41 .. 0x11F41: Mc         KAWI SIGN KILLER
    // 0x11F42 .. 0x11F42: Mn         KAWI CONJOINER
    // 0x11F43 .. 0x11F4F: Po    [13] KAWI DANDA..KAWI PUNCTUATION CLOSING SPIRAL
    // 0x11F50 .. 0x11F59: Nd    [10] KAWI DIGIT ZERO..KAWI DIGIT NINE
    // 0x11F5A .. 0x11F5A: Mn         KAWI SIGN NUKTA
    Prop<::unicode::EastAsianWidth> { { 0x11F3E, 0x11F5A }, ::unicode::EastAsianWidth::Neutral }, // #7
    Prop<::unicode::EastAsianWidth> { { 0x11FB0, 0x11FB0 }, ::unicode::EastAsianWidth::Neutral }, // Lo         LISU LETTER YHA
    // 0x11FC0 .. 0x11FD4: No    [21] TAMIL FRACTION ONE THREE-HUNDRED-AND-TWENTIETH..TAMIL FRACTION DOWNSCALING FACTOR KIIZH
    // 0x11FD5 .. 0x11FDC: So     [8] TAMIL SIGN NEL..TAMIL SIGN MUKKURUNI
    // 0x11FDD .. 0x11FE0: Sc     [4] TAMIL SIGN KAACU..TAMIL SIGN VARAAKAN
    // 0x11FE1 .. 0x11FF1: So    [17] TAMIL SIGN PAARAM..TAMIL SIGN VAKAIYARAA
    Prop<::unicode::EastAsianWidth> { { 0x11FC0, 0x11FF1 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x11FFF .. 0x11FFF: Po         TAMIL PUNCTUATION END OF TEXT
    // 0x12000 .. 0x12399: Lo   [922] CUNEIFORM SIGN A..CUNEIFORM SIGN U U
    Prop<::unicode::EastAsianWidth> { { 0x11FFF, 0x12399 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x12400, 0x1246E }, ::unicode::EastAsianWidth::Neutral }, // Nl   [111] CUNEIFORM NUMERIC SIGN TWO ASH..CUNEIFORM NUMERIC SIGN NINE U VARIANT FORM
    Prop<::unicode::EastAsianWidth> { { 0x12470, 0x12474 }, ::unicode::EastAsianWidth::Neutral }, // Po     [5] CUNEIFORM PUNCTUATION SIGN OLD ASSYRIAN WORD DIVIDER..CUNEIFORM PUNCTUATION SIGN DIAGONAL QUADCOLON
    Prop<::unicode::EastAsianWidth> { { 0x12480, 0x12543 }, ::unicode::EastAsianWidth::Neutral }, // Lo   [196] CUNEIFORM SIGN AB TIMES NUN TENU..CUNEIFORM SIGN ZU5 TIMES THREE DISH TENU
    // 0x12F90 .. 0x12FF0: Lo    [97] CYPRO-MINOAN SIGN CM001..CYPRO-MINOAN SIGN CM114
    // 0x12FF1 .. 0x12FF2: Po     [2] CYPRO-MINOAN SIGN CM301..CYPRO-MINOAN SIGN CM302
    Prop<::unicode::EastAsianWidth> { { 0x12F90, 0x12FF2 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x13000 .. 0x1342F: Lo  [1072] EGYPTIAN HIEROGLYPH A001..EGYPTIAN HIEROGLYPH V011D
    // 0x13430 .. 0x1343F: Cf    [16] EGYPTIAN HIEROGLYPH VERTICAL JOINER..EGYPTIAN HIEROGLYPH END WALLED ENCLOSURE
    // 0x13440 .. 0x13440: Mn         EGYPTIAN HIEROGLYPH MIRROR HORIZONTALLY
    // 0x13441 .. 0x13446: Lo     [6] EGYPTIAN HIEROGLYPH FULL BLANK..EGYPTIAN HIEROGLYPH WIDE LOST SIGN
    // 0x13447 .. 0x13455: Mn    [15] EGYPTIAN HIEROGLYPH MODIFIER DAMAGED AT TOP START..EGYPTIAN HIEROGLYPH MODIFIER DAMAGED
    Prop<::unicode::EastAsianWidth> { { 0x13000, 0x13455 }, ::unicode::EastAsianWidth::Neutral }, // #5
    Prop<::unicode::EastAsianWidth> { { 0x13460, 0x143FA }, ::unicode::EastAsianWidth::Neutral }, // Lo  [3995] EGYPTIAN HIEROGLYPH-13460..EGYPTIAN HIEROGLYPH-143FA
    Prop<::unicode::EastAsianWidth> { { 0x14400, 0x14646 }, ::unicode::EastAsianWidth::Neutral }, // Lo   [583] ANATOLIAN HIEROGLYPH A001..ANATOLIAN HIEROGLYPH A530
    // 0x16100 .. 0x1611D: Lo    [30] GURUNG KHEMA LETTER A..GURUNG KHEMA LETTER SA
    // 0x1611E .. 0x16129: Mn    [12] GURUNG KHEMA VOWEL SIGN AA..GURUNG KHEMA VOWEL LENGTH MARK
    // 0x1612A .. 0x1612C: Mc     [3] GURUNG KHEMA CONSONANT SIGN MEDIAL YA..GURUNG KHEMA CONSONANT SIGN MEDIAL HA
    // 0x1612D .. 0x1612F: Mn     [3] GURUNG KHEMA SIGN ANUSVARA..GURUNG KHEMA SIGN THOLHOMA
    // 0x16130 .. 0x16139: Nd    [10] GURUNG KHEMA DIGIT ZERO..GURUNG KHEMA DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x16100, 0x16139 }, ::unicode::EastAsianWidth::Neutral }, // #5
    Prop<::unicode::EastAsianWidth> { { 0x16800, 0x16A38 }, ::unicode::EastAsianWidth::Neutral }, // Lo   [569] BAMUM LETTER PHASE-A NGKUE MFON..BAMUM LETTER PHASE-F VUEQ
    Prop<::unicode::EastAsianWidth> { { 0x16A40, 0x16A5E }, ::unicode::EastAsianWidth::Neutral }, // Lo    [31] MRO LETTER TA..MRO LETTER TEK
    Prop<::unicode::EastAsianWidth> { { 0x16A60, 0x16A69 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] MRO DIGIT ZERO..MRO DIGIT NINE
    // 0x16A6E .. 0x16A6F: Po     [2] MRO DANDA..MRO DOUBLE DANDA
    // 0x16A70 .. 0x16ABE: Lo    [79] TANGSA LETTER OZ..TANGSA LETTER ZA
    Prop<::unicode::EastAsianWidth> { { 0x16A6E, 0x16ABE }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x16AC0, 0x16AC9 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] TANGSA DIGIT ZERO..TANGSA DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x16AD0, 0x16AED }, ::unicode::EastAsianWidth::Neutral }, // Lo    [30] BASSA VAH LETTER ENNI..BASSA VAH LETTER I
    // 0x16AF0 .. 0x16AF4: Mn     [5] BASSA VAH COMBINING HIGH TONE..BASSA VAH COMBINING HIGH-LOW TONE
    // 0x16AF5 .. 0x16AF5: Po         BASSA VAH FULL STOP
    Prop<::unicode::EastAsianWidth> { { 0x16AF0, 0x16AF5 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x16B00 .. 0x16B2F: Lo    [48] PAHAWH HMONG VOWEL KEEB..PAHAWH HMONG CONSONANT CAU
    // 0x16B30 .. 0x16B36: Mn     [7] PAHAWH HMONG MARK CIM TUB..PAHAWH HMONG MARK CIM TAUM
    // 0x16B37 .. 0x16B3B: Po     [5] PAHAWH HMONG SIGN VOS THOM..PAHAWH HMONG SIGN VOS FEEM
    // 0x16B3C .. 0x16B3F: So     [4] PAHAWH HMONG SIGN XYEEM NTXIV..PAHAWH HMONG SIGN XYEEM FAIB
    // 0x16B40 .. 0x16B43: Lm     [4] PAHAWH HMONG SIGN VOS SEEV..PAHAWH HMONG SIGN IB YAM
    // 0x16B44 .. 0x16B44: Po         PAHAWH HMONG SIGN XAUS
    // 0x16B45 .. 0x16B45: So         PAHAWH HMONG SIGN CIM TSOV ROG
    Prop<::unicode::EastAsianWidth> { { 0x16B00, 0x16B45 }, ::unicode::EastAsianWidth::Neutral }, // #7
    Prop<::unicode::EastAsianWidth> { { 0x16B50, 0x16B59 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] PAHAWH HMONG DIGIT ZERO..PAHAWH HMONG DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x16B5B, 0x16B61 }, ::unicode::EastAsianWidth::Neutral }, // No     [7] PAHAWH HMONG NUMBER TENS..PAHAWH HMONG NUMBER TRILLIONS
    Prop<::unicode::EastAsianWidth> { { 0x16B63, 0x16B77 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [21] PAHAWH HMONG SIGN VOS LUB..PAHAWH HMONG SIGN CIM NRES TOS
    Prop<::unicode::EastAsianWidth> { { 0x16B7D, 0x16B8F }, ::unicode::EastAsianWidth::Neutral }, // Lo    [19] PAHAWH HMONG CLAN SIGN TSHEEJ..PAHAWH HMONG CLAN SIGN VWJ
    // 0x16D40 .. 0x16D42: Lm     [3] KIRAT RAI SIGN ANUSVARA..KIRAT RAI SIGN VISARGA
    // 0x16D43 .. 0x16D6A: Lo    [40] KIRAT RAI LETTER A..KIRAT RAI VOWEL SIGN AU
    // 0x16D6B .. 0x16D6C: Lm     [2] KIRAT RAI SIGN VIRAMA..KIRAT RAI SIGN SAAT
    // 0x16D6D .. 0x16D6F: Po     [3] KIRAT RAI SIGN YUPI..KIRAT RAI DOUBLE DANDA
    // 0x16D70 .. 0x16D79: Nd    [10] KIRAT RAI DIGIT ZERO..KIRAT RAI DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x16D40, 0x16D79 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x16E40 .. 0x16E7F: L&    [64] MEDEFAIDRIN CAPITAL LETTER M..MEDEFAIDRIN SMALL LETTER Y
    // 0x16E80 .. 0x16E96: No    [23] MEDEFAIDRIN DIGIT ZERO..MEDEFAIDRIN DIGIT THREE ALTERNATE FORM
    // 0x16E97 .. 0x16E9A: Po     [4] MEDEFAIDRIN COMMA..MEDEFAIDRIN EXCLAMATION OH
    Prop<::unicode::EastAsianWidth> { { 0x16E40, 0x16E9A }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x16EA0, 0x16EB8 }, ::unicode::EastAsianWidth::Neutral }, // Lu    [25] BERIA ERFE CAPITAL LETTER ARKAB..BERIA ERFE CAPITAL LETTER AY
    Prop<::unicode::EastAsianWidth> { { 0x16EBB, 0x16ED3 }, ::unicode::EastAsianWidth::Neutral }, // Ll    [25] BERIA ERFE SMALL LETTER ARKAB..BERIA ERFE SMALL LETTER AY
    Prop<::unicode::EastAsianWidth> { { 0x16F00, 0x16F4A }, ::unicode::EastAsianWidth::Neutral }, // Lo    [75] MIAO LETTER PA..MIAO LETTER RTE
    // 0x16F4F .. 0x16F4F: Mn         MIAO SIGN CONSONANT MODIFIER BAR
    // 0x16F50 .. 0x16F50: Lo         MIAO LETTER NASALIZATION
    // 0x16F51 .. 0x16F87: Mc    [55] MIAO SIGN ASPIRATION..MIAO VOWEL SIGN UI
    Prop<::unicode::EastAsianWidth> { { 0x16F4F, 0x16F87 }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x16F8F .. 0x16F92: Mn     [4] MIAO TONE RIGHT..MIAO TONE BELOW
    // 0x16F93 .. 0x16F9F: Lm    [13] MIAO LETTER TONE-2..MIAO LETTER REFORMED TONE-8
    Prop<::unicode::EastAsianWidth> { { 0x16F8F, 0x16F9F }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x16FE0 .. 0x16FE1: Lm     [2] TANGUT ITERATION MARK..NUSHU ITERATION MARK
    // 0x16FE2 .. 0x16FE2: Po         OLD CHINESE HOOK MARK
    // 0x16FE3 .. 0x16FE3: Lm         OLD CHINESE ITERATION MARK
    // 0x16FE4 .. 0x16FE4: Mn         KHITAN SMALL SCRIPT FILLER
    Prop<::unicode::EastAsianWidth> { { 0x16FE0, 0x16FE4 }, ::unicode::EastAsianWidth::Wide }, // #4
    // 0x16FF0 .. 0x16FF1: Mc     [2] VIETNAMESE ALTERNATE READING MARK CA..VIETNAMESE ALTERNATE READING MARK NHAY
    // 0x16FF2 .. 0x16FF3: Lm     [2] CHINESE SMALL SIMPLIFIED ER..CHINESE SMALL TRADITIONAL ER
    // 0x16FF4 .. 0x16FF6: Nl     [3] YANGQIN SIGN SLOW ONE BEAT..YANGQIN SIGN SLOW TWO BEATS
    Prop<::unicode::EastAsianWidth> { { 0x16FF0, 0x16FF6 }, ::unicode::EastAsianWidth::Wide }, // #3
    // 0x17000 .. 0x187FF: Lo  [6144] TANGUT IDEOGRAPH-17000..TANGUT IDEOGRAPH-187FF
    // 0x18800 .. 0x18AFF: Lo   [768] TANGUT COMPONENT-001..TANGUT COMPONENT-768
    // 0x18B00 .. 0x18CD5: Lo   [470] KHITAN SMALL SCRIPT CHARACTER-18B00..KHITAN SMALL SCRIPT CHARACTER-18CD5
    Prop<::unicode::EastAsianWidth> { { 0x17000, 0x18CD5 }, ::unicode::EastAsianWidth::Wide }, // #3
    // 0x18CFF .. 0x18CFF: Lo         KHITAN SMALL SCRIPT CHARACTER-18CFF
    // 0x18D00 .. 0x18D1E: Lo    [31] TANGUT IDEOGRAPH-18D00..TANGUT IDEOGRAPH-18D1E
    Prop<::unicode::EastAsianWidth> { { 0x18CFF, 0x18D1E }, ::unicode::EastAsianWidth::Wide }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x18D80, 0x18DF2 }, ::unicode::EastAsianWidth::Wide }, // Lo   [115] TANGUT COMPONENT-769..TANGUT COMPONENT-883
    Prop<::unicode::EastAsianWidth> { { 0x1AFF0, 0x1AFF3 }, ::unicode::EastAsianWidth::Wide }, // Lm     [4] KATAKANA LETTER MINNAN TONE-2..KATAKANA LETTER MINNAN TONE-5
    Prop<::unicode::EastAsianWidth> { { 0x1AFF5, 0x1AFFB }, ::unicode::EastAsianWidth::Wide }, // Lm     [7] KATAKANA LETTER MINNAN TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-5
    Prop<::unicode::EastAsianWidth> { { 0x1AFFD, 0x1AFFE }, ::unicode::EastAsianWidth::Wide }, // Lm     [2] KATAKANA LETTER MINNAN NASALIZED TONE-7..KATAKANA LETTER MINNAN NASALIZED TONE-8
    // 0x1B000 .. 0x1B0FF: Lo   [256] KATAKANA LETTER ARCHAIC E..HENTAIGANA LETTER RE-2
    // 0x1B100 .. 0x1B122: Lo    [35] HENTAIGANA LETTER RE-3..KATAKANA LETTER ARCHAIC WU
    Prop<::unicode::EastAsianWidth> { { 0x1B000, 0x1B122 }, ::unicode::EastAsianWidth::Wide }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x1B132, 0x1B132 }, ::unicode::EastAsianWidth::Wide }, // Lo         HIRAGANA LETTER SMALL KO
    Prop<::unicode::EastAsianWidth> { { 0x1B150, 0x1B152 }, ::unicode::EastAsianWidth::Wide }, // Lo     [3] HIRAGANA LETTER SMALL WI..HIRAGANA LETTER SMALL WO
    Prop<::unicode::EastAsianWidth> { { 0x1B155, 0x1B155 }, ::unicode::EastAsianWidth::Wide }, // Lo         KATAKANA LETTER SMALL KO
    Prop<::unicode::EastAsianWidth> { { 0x1B164, 0x1B167 }, ::unicode::EastAsianWidth::Wide }, // Lo     [4] KATAKANA LETTER SMALL WI..KATAKANA LETTER SMALL N
    Prop<::unicode::EastAsianWidth> { { 0x1B170, 0x1B2FB }, ::unicode::EastAsianWidth::Wide }, // Lo   [396] NUSHU CHARACTER-1B170..NUSHU CHARACTER-1B2FB
    Prop<::unicode::EastAsianWidth> { { 0x1BC00, 0x1BC6A }, ::unicode::EastAsianWidth::Neutral }, // Lo   [107] DUPLOYAN LETTER H..DUPLOYAN LETTER VOCALIC M
    Prop<::unicode::EastAsianWidth> { { 0x1BC70, 0x1BC7C }, ::unicode::EastAsianWidth::Neutral }, // Lo    [13] DUPLOYAN AFFIX LEFT HORIZONTAL SECANT..DUPLOYAN AFFIX ATTACHED TANGENT HOOK
    Prop<::unicode::EastAsianWidth> { { 0x1BC80, 0x1BC88 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [9] DUPLOYAN AFFIX HIGH ACUTE..DUPLOYAN AFFIX HIGH VERTICAL
    Prop<::unicode::EastAsianWidth> { { 0x1BC90, 0x1BC99 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [10] DUPLOYAN AFFIX LOW ACUTE..DUPLOYAN AFFIX LOW ARROW
    // 0x1BC9C .. 0x1BC9C: So         DUPLOYAN SIGN O WITH CROSS
    // 0x1BC9D .. 0x1BC9E: Mn     [2] DUPLOYAN THICK LETTER SELECTOR..DUPLOYAN DOUBLE MARK
    // 0x1BC9F .. 0x1BC9F: Po         DUPLOYAN PUNCTUATION CHINOOK FULL STOP
    // 0x1BCA0 .. 0x1BCA3: Cf     [4] SHORTHAND FORMAT LETTER OVERLAP..SHORTHAND FORMAT UP STEP
    Prop<::unicode::EastAsianWidth> { { 0x1BC9C, 0x1BCA3 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x1CC00 .. 0x1CCEF: So   [240] UP-POINTING GO-KART..OUTLINED LATIN CAPITAL LETTER Z
    // 0x1CCF0 .. 0x1CCF9: Nd    [10] OUTLINED DIGIT ZERO..OUTLINED DIGIT NINE
    // 0x1CCFA .. 0x1CCFC: So     [3] SNAKE SYMBOL..NOSE SYMBOL
    Prop<::unicode::EastAsianWidth> { { 0x1CC00, 0x1CCFC }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x1CD00, 0x1CEB3 }, ::unicode::EastAsianWidth::Neutral }, // So   [436] BLOCK OCTANT-3..BLACK RIGHT TRIANGLE CARET
    // 0x1CEBA .. 0x1CEBF: So     [6] FRAGILE SYMBOL..STRAWBERRY SYMBOL
    // 0x1CEC0 .. 0x1CED0: So    [17] HEBE..LEUKOTHEA
    Prop<::unicode::EastAsianWidth> { { 0x1CEBA, 0x1CED0 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x1CEE0 .. 0x1CEEF: So    [16] GEOMANTIC FIGURE POPULUS..GEOMANTIC FIGURE VIA
    // 0x1CEF0 .. 0x1CEF0: Sm         MEDIUM SMALL WHITE CIRCLE WITH HORIZONTAL BAR
    Prop<::unicode::EastAsianWidth> { { 0x1CEE0, 0x1CEF0 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x1CF00, 0x1CF2D }, ::unicode::EastAsianWidth::Neutral }, // Mn    [46] ZNAMENNY COMBINING MARK GORAZDO NIZKO S KRYZHEM ON LEFT..ZNAMENNY COMBINING MARK KRYZH ON LEFT
    Prop<::unicode::EastAsianWidth> { { 0x1CF30, 0x1CF46 }, ::unicode::EastAsianWidth::Neutral }, // Mn    [23] ZNAMENNY COMBINING TONAL RANGE MARK MRACHNO..ZNAMENNY PRIZNAK MODIFIER ROG
    Prop<::unicode::EastAsianWidth> { { 0x1CF50, 0x1CFC3 }, ::unicode::EastAsianWidth::Neutral }, // So   [116] ZNAMENNY NEUME KRYUK..ZNAMENNY NEUME PAUK
    Prop<::unicode::EastAsianWidth> { { 0x1D000, 0x1D0F5 }, ::unicode::EastAsianWidth::Neutral }, // So   [246] BYZANTINE MUSICAL SYMBOL PSILI..BYZANTINE MUSICAL SYMBOL GORGON NEO KATO
    Prop<::unicode::EastAsianWidth> { { 0x1D100, 0x1D126 }, ::unicode::EastAsianWidth::Neutral }, // So    [39] MUSICAL SYMBOL SINGLE BARLINE..MUSICAL SYMBOL DRUM CLEF-2
    // 0x1D129 .. 0x1D164: So    [60] MUSICAL SYMBOL MULTIPLE MEASURE REST..MUSICAL SYMBOL ONE HUNDRED TWENTY-EIGHTH NOTE
    // 0x1D165 .. 0x1D166: Mc     [2] MUSICAL SYMBOL COMBINING STEM..MUSICAL SYMBOL COMBINING SPRECHGESANG STEM
    // 0x1D167 .. 0x1D169: Mn     [3] MUSICAL SYMBOL COMBINING TREMOLO-1..MUSICAL SYMBOL COMBINING TREMOLO-3
    // 0x1D16A .. 0x1D16C: So     [3] MUSICAL SYMBOL FINGERED TREMOLO-1..MUSICAL SYMBOL FINGERED TREMOLO-3
    // 0x1D16D .. 0x1D172: Mc     [6] MUSICAL SYMBOL COMBINING AUGMENTATION DOT..MUSICAL SYMBOL COMBINING FLAG-5
    // 0x1D173 .. 0x1D17A: Cf     [8] MUSICAL SYMBOL BEGIN BEAM..MUSICAL SYMBOL END PHRASE
    // 0x1D17B .. 0x1D182: Mn     [8] MUSICAL SYMBOL COMBINING ACCENT..MUSICAL SYMBOL COMBINING LOURE
    // 0x1D183 .. 0x1D184: So     [2] MUSICAL SYMBOL ARPEGGIATO UP..MUSICAL SYMBOL ARPEGGIATO DOWN
    // 0x1D185 .. 0x1D18B: Mn     [7] MUSICAL SYMBOL COMBINING DOIT..MUSICAL SYMBOL COMBINING TRIPLE TONGUE
    // 0x1D18C .. 0x1D1A9: So    [30] MUSICAL SYMBOL RINFORZANDO..MUSICAL SYMBOL DEGREE SLASH
    // 0x1D1AA .. 0x1D1AD: Mn     [4] MUSICAL SYMBOL COMBINING DOWN BOW..MUSICAL SYMBOL COMBINING SNAP PIZZICATO
    // 0x1D1AE .. 0x1D1EA: So    [61] MUSICAL SYMBOL PEDAL MARK..MUSICAL SYMBOL KORON
    Prop<::unicode::EastAsianWidth> { { 0x1D129, 0x1D1EA }, ::unicode::EastAsianWidth::Neutral }, // #12
    // 0x1D200 .. 0x1D241: So    [66] GREEK VOCAL NOTATION SYMBOL-1..GREEK INSTRUMENTAL NOTATION SYMBOL-54
    // 0x1D242 .. 0x1D244: Mn     [3] COMBINING GREEK MUSICAL TRISEME..COMBINING GREEK MUSICAL PENTASEME
    // 0x1D245 .. 0x1D245: So         GREEK MUSICAL LEIMMA
    Prop<::unicode::EastAsianWidth> { { 0x1D200, 0x1D245 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x1D2C0, 0x1D2D3 }, ::unicode::EastAsianWidth::Neutral }, // No    [20] KAKTOVIK NUMERAL ZERO..KAKTOVIK NUMERAL NINETEEN
    Prop<::unicode::EastAsianWidth> { { 0x1D2E0, 0x1D2F3 }, ::unicode::EastAsianWidth::Neutral }, // No    [20] MAYAN NUMERAL ZERO..MAYAN NUMERAL NINETEEN
    Prop<::unicode::EastAsianWidth> { { 0x1D300, 0x1D356 }, ::unicode::EastAsianWidth::Wide }, // So    [87] MONOGRAM FOR EARTH..TETRAGRAM FOR FOSTERING
    Prop<::unicode::EastAsianWidth> { { 0x1D360, 0x1D376 }, ::unicode::EastAsianWidth::Wide }, // No    [23] COUNTING ROD UNIT DIGIT ONE..IDEOGRAPHIC TALLY MARK FIVE
    Prop<::unicode::EastAsianWidth> { { 0x1D377, 0x1D378 }, ::unicode::EastAsianWidth::Neutral }, // No     [2] TALLY MARK ONE..TALLY MARK FIVE
    Prop<::unicode::EastAsianWidth> { { 0x1D400, 0x1D454 }, ::unicode::EastAsianWidth::Neutral }, // L&    [85] MATHEMATICAL BOLD CAPITAL A..MATHEMATICAL ITALIC SMALL G
    Prop<::unicode::EastAsianWidth> { { 0x1D456, 0x1D49C }, ::unicode::EastAsianWidth::Neutral }, // L&    [71] MATHEMATICAL ITALIC SMALL I..MATHEMATICAL SCRIPT CAPITAL A
    Prop<::unicode::EastAsianWidth> { { 0x1D49E, 0x1D49F }, ::unicode::EastAsianWidth::Neutral }, // Lu     [2] MATHEMATICAL SCRIPT CAPITAL C..MATHEMATICAL SCRIPT CAPITAL D
    Prop<::unicode::EastAsianWidth> { { 0x1D4A2, 0x1D4A2 }, ::unicode::EastAsianWidth::Neutral }, // Lu         MATHEMATICAL SCRIPT CAPITAL G
    Prop<::unicode::EastAsianWidth> { { 0x1D4A5, 0x1D4A6 }, ::unicode::EastAsianWidth::Neutral }, // Lu     [2] MATHEMATICAL SCRIPT CAPITAL J..MATHEMATICAL SCRIPT CAPITAL K
    Prop<::unicode::EastAsianWidth> { { 0x1D4A9, 0x1D4AC }, ::unicode::EastAsianWidth::Neutral }, // Lu     [4] MATHEMATICAL SCRIPT CAPITAL N..MATHEMATICAL SCRIPT CAPITAL Q
    Prop<::unicode::EastAsianWidth> { { 0x1D4AE, 0x1D4B9 }, ::unicode::EastAsianWidth::Neutral }, // L&    [12] MATHEMATICAL SCRIPT CAPITAL S..MATHEMATICAL SCRIPT SMALL D
    Prop<::unicode::EastAsianWidth> { { 0x1D4BB, 0x1D4BB }, ::unicode::EastAsianWidth::Neutral }, // Ll         MATHEMATICAL SCRIPT SMALL F
    Prop<::unicode::EastAsianWidth> { { 0x1D4BD, 0x1D4C3 }, ::unicode::EastAsianWidth::Neutral }, // Ll     [7] MATHEMATICAL SCRIPT SMALL H..MATHEMATICAL SCRIPT SMALL N
    Prop<::unicode::EastAsianWidth> { { 0x1D4C5, 0x1D505 }, ::unicode::EastAsianWidth::Neutral }, // L&    [65] MATHEMATICAL SCRIPT SMALL P..MATHEMATICAL FRAKTUR CAPITAL B
    Prop<::unicode::EastAsianWidth> { { 0x1D507, 0x1D50A }, ::unicode::EastAsianWidth::Neutral }, // Lu     [4] MATHEMATICAL FRAKTUR CAPITAL D..MATHEMATICAL FRAKTUR CAPITAL G
    Prop<::unicode::EastAsianWidth> { { 0x1D50D, 0x1D514 }, ::unicode::EastAsianWidth::Neutral }, // Lu     [8] MATHEMATICAL FRAKTUR CAPITAL J..MATHEMATICAL FRAKTUR CAPITAL Q
    Prop<::unicode::EastAsianWidth> { { 0x1D516, 0x1D51C }, ::unicode::EastAsianWidth::Neutral }, // Lu     [7] MATHEMATICAL FRAKTUR CAPITAL S..MATHEMATICAL FRAKTUR CAPITAL Y
    Prop<::unicode::EastAsianWidth> { { 0x1D51E, 0x1D539 }, ::unicode::EastAsianWidth::Neutral }, // L&    [28] MATHEMATICAL FRAKTUR SMALL A..MATHEMATICAL DOUBLE-STRUCK CAPITAL B
    Prop<::unicode::EastAsianWidth> { { 0x1D53B, 0x1D53E }, ::unicode::EastAsianWidth::Neutral }, // Lu     [4] MATHEMATICAL DOUBLE-STRUCK CAPITAL D..MATHEMATICAL DOUBLE-STRUCK CAPITAL G
    Prop<::unicode::EastAsianWidth> { { 0x1D540, 0x1D544 }, ::unicode::EastAsianWidth::Neutral }, // Lu     [5] MATHEMATICAL DOUBLE-STRUCK CAPITAL I..MATHEMATICAL DOUBLE-STRUCK CAPITAL M
    Prop<::unicode::EastAsianWidth> { { 0x1D546, 0x1D546 }, ::unicode::EastAsianWidth::Neutral }, // Lu         MATHEMATICAL DOUBLE-STRUCK CAPITAL O
    Prop<::unicode::EastAsianWidth> { { 0x1D54A, 0x1D550 }, ::unicode::EastAsianWidth::Neutral }, // Lu     [7] MATHEMATICAL DOUBLE-STRUCK CAPITAL S..MATHEMATICAL DOUBLE-STRUCK CAPITAL Y
    Prop<::unicode::EastAsianWidth> { { 0x1D552, 0x1D6A5 }, ::unicode::EastAsianWidth::Neutral }, // L&   [340] MATHEMATICAL DOUBLE-STRUCK SMALL A..MATHEMATICAL ITALIC SMALL DOTLESS J
    // 0x1D6A8 .. 0x1D6C0: Lu    [25] MATHEMATICAL BOLD CAPITAL ALPHA..MATHEMATICAL BOLD CAPITAL OMEGA
    // 0x1D6C1 .. 0x1D6C1: Sm         MATHEMATICAL BOLD NABLA
    // 0x1D6C2 .. 0x1D6DA: Ll    [25] MATHEMATICAL BOLD SMALL ALPHA..MATHEMATICAL BOLD SMALL OMEGA
    // 0x1D6DB .. 0x1D6DB: Sm         MATHEMATICAL BOLD PARTIAL DIFFERENTIAL
    // 0x1D6DC .. 0x1D6FA: L&    [31] MATHEMATICAL BOLD EPSILON SYMBOL..MATHEMATICAL ITALIC CAPITAL OMEGA
    // 0x1D6FB .. 0x1D6FB: Sm         MATHEMATICAL ITALIC NABLA
    // 0x1D6FC .. 0x1D714: Ll    [25] MATHEMATICAL ITALIC SMALL ALPHA..MATHEMATICAL ITALIC SMALL OMEGA
    // 0x1D715 .. 0x1D715: Sm         MATHEMATICAL ITALIC PARTIAL DIFFERENTIAL
    // 0x1D716 .. 0x1D734: L&    [31] MATHEMATICAL ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD ITALIC CAPITAL OMEGA
    // 0x1D735 .. 0x1D735: Sm         MATHEMATICAL BOLD ITALIC NABLA
    // 0x1D736 .. 0x1D74E: Ll    [25] MATHEMATICAL BOLD ITALIC SMALL ALPHA..MATHEMATICAL BOLD ITALIC SMALL OMEGA
    // 0x1D74F .. 0x1D74F: Sm         MATHEMATICAL BOLD ITALIC PARTIAL DIFFERENTIAL
    // 0x1D750 .. 0x1D76E: L&    [31] MATHEMATICAL BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD CAPITAL OMEGA
    // 0x1D76F .. 0x1D76F: Sm         MATHEMATICAL SANS-SERIF BOLD NABLA
    // 0x1D770 .. 0x1D788: Ll    [25] MATHEMATICAL SANS-SERIF BOLD SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD SMALL OMEGA
    // 0x1D789 .. 0x1D789: Sm         MATHEMATICAL SANS-SERIF BOLD PARTIAL DIFFERENTIAL
    // 0x1D78A .. 0x1D7A8: L&    [31] MATHEMATICAL SANS-SERIF BOLD EPSILON SYMBOL..MATHEMATICAL SANS-SERIF BOLD ITALIC CAPITAL OMEGA
    // 0x1D7A9 .. 0x1D7A9: Sm         MATHEMATICAL SANS-SERIF BOLD ITALIC NABLA
    // 0x1D7AA .. 0x1D7C2: Ll    [25] MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL ALPHA..MATHEMATICAL SANS-SERIF BOLD ITALIC SMALL OMEGA
    // 0x1D7C3 .. 0x1D7C3: Sm         MATHEMATICAL SANS-SERIF BOLD ITALIC PARTIAL DIFFERENTIAL
    // 0x1D7C4 .. 0x1D7CB: L&     [8] MATHEMATICAL SANS-SERIF BOLD ITALIC EPSILON SYMBOL..MATHEMATICAL BOLD SMALL DIGAMMA
    Prop<::unicode::EastAsianWidth> { { 0x1D6A8, 0x1D7CB }, ::unicode::EastAsianWidth::Neutral }, // #21
    // 0x1D7CE .. 0x1D7FF: Nd    [50] MATHEMATICAL BOLD DIGIT ZERO..MATHEMATICAL MONOSPACE DIGIT NINE
    // 0x1D800 .. 0x1D9FF: So   [512] SIGNWRITING HAND-FIST INDEX..SIGNWRITING HEAD
    // 0x1DA00 .. 0x1DA36: Mn    [55] SIGNWRITING HEAD RIM..SIGNWRITING AIR SUCKING IN
    // 0x1DA37 .. 0x1DA3A: So     [4] SIGNWRITING AIR BLOW SMALL ROTATIONS..SIGNWRITING BREATH EXHALE
    // 0x1DA3B .. 0x1DA6C: Mn    [50] SIGNWRITING MOUTH CLOSED NEUTRAL..SIGNWRITING EXCITEMENT
    // 0x1DA6D .. 0x1DA74: So     [8] SIGNWRITING SHOULDER HIP SPINE..SIGNWRITING TORSO-FLOORPLANE TWISTING
    // 0x1DA75 .. 0x1DA75: Mn         SIGNWRITING UPPER BODY TILTING FROM HIP JOINTS
    // 0x1DA76 .. 0x1DA83: So    [14] SIGNWRITING LIMB COMBINATION..SIGNWRITING LOCATION DEPTH
    // 0x1DA84 .. 0x1DA84: Mn         SIGNWRITING LOCATION HEAD NECK
    // 0x1DA85 .. 0x1DA86: So     [2] SIGNWRITING LOCATION TORSO..SIGNWRITING LOCATION LIMBS DIGITS
    // 0x1DA87 .. 0x1DA8B: Po     [5] SIGNWRITING COMMA..SIGNWRITING PARENTHESIS
    Prop<::unicode::EastAsianWidth> { { 0x1D7CE, 0x1DA8B }, ::unicode::EastAsianWidth::Neutral }, // #11
    Prop<::unicode::EastAsianWidth> { { 0x1DA9B, 0x1DA9F }, ::unicode::EastAsianWidth::Neutral }, // Mn     [5] SIGNWRITING FILL MODIFIER-2..SIGNWRITING FILL MODIFIER-6
    Prop<::unicode::EastAsianWidth> { { 0x1DAA1, 0x1DAAF }, ::unicode::EastAsianWidth::Neutral }, // Mn    [15] SIGNWRITING ROTATION MODIFIER-2..SIGNWRITING ROTATION MODIFIER-16
    // 0x1DF00 .. 0x1DF09: Ll    [10] LATIN SMALL LETTER FENG DIGRAPH WITH TRILL..LATIN SMALL LETTER T WITH HOOK AND RETROFLEX HOOK
    // 0x1DF0A .. 0x1DF0A: Lo         LATIN LETTER RETROFLEX CLICK WITH RETROFLEX HOOK
    // 0x1DF0B .. 0x1DF1E: Ll    [20] LATIN SMALL LETTER ESH WITH DOUBLE BAR..LATIN SMALL LETTER S WITH CURL
    Prop<::unicode::EastAsianWidth> { { 0x1DF00, 0x1DF1E }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x1DF25, 0x1DF2A }, ::unicode::EastAsianWidth::Neutral }, // Ll     [6] LATIN SMALL LETTER D WITH MID-HEIGHT LEFT HOOK..LATIN SMALL LETTER T WITH MID-HEIGHT LEFT HOOK
    Prop<::unicode::EastAsianWidth> { { 0x1E000, 0x1E006 }, ::unicode::EastAsianWidth::Neutral }, // Mn     [7] COMBINING GLAGOLITIC LETTER AZU..COMBINING GLAGOLITIC LETTER ZHIVETE
    Prop<::unicode::EastAsianWidth> { { 0x1E008, 0x1E018 }, ::unicode::EastAsianWidth::Neutral }, // Mn    [17] COMBINING GLAGOLITIC LETTER ZEMLJA..COMBINING GLAGOLITIC LETTER HERU
    Prop<::unicode::EastAsianWidth> { { 0x1E01B, 0x1E021 }, ::unicode::EastAsianWidth::Neutral }, // Mn     [7] COMBINING GLAGOLITIC LETTER SHTA..COMBINING GLAGOLITIC LETTER YATI
    Prop<::unicode::EastAsianWidth> { { 0x1E023, 0x1E024 }, ::unicode::EastAsianWidth::Neutral }, // Mn     [2] COMBINING GLAGOLITIC LETTER YU..COMBINING GLAGOLITIC LETTER SMALL YUS
    Prop<::unicode::EastAsianWidth> { { 0x1E026, 0x1E02A }, ::unicode::EastAsianWidth::Neutral }, // Mn     [5] COMBINING GLAGOLITIC LETTER YO..COMBINING GLAGOLITIC LETTER FITA
    Prop<::unicode::EastAsianWidth> { { 0x1E030, 0x1E06D }, ::unicode::EastAsianWidth::Neutral }, // Lm    [62] MODIFIER LETTER CYRILLIC SMALL A..MODIFIER LETTER CYRILLIC SMALL STRAIGHT U WITH STROKE
    Prop<::unicode::EastAsianWidth> { { 0x1E08F, 0x1E08F }, ::unicode::EastAsianWidth::Neutral }, // Mn         COMBINING CYRILLIC SMALL LETTER BYELORUSSIAN-UKRAINIAN I
    Prop<::unicode::EastAsianWidth> { { 0x1E100, 0x1E12C }, ::unicode::EastAsianWidth::Neutral }, // Lo    [45] NYIAKENG PUACHUE HMONG LETTER MA..NYIAKENG PUACHUE HMONG LETTER W
    // 0x1E130 .. 0x1E136: Mn     [7] NYIAKENG PUACHUE HMONG TONE-B..NYIAKENG PUACHUE HMONG TONE-D
    // 0x1E137 .. 0x1E13D: Lm     [7] NYIAKENG PUACHUE HMONG SIGN FOR PERSON..NYIAKENG PUACHUE HMONG SYLLABLE LENGTHENER
    Prop<::unicode::EastAsianWidth> { { 0x1E130, 0x1E13D }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x1E140, 0x1E149 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] NYIAKENG PUACHUE HMONG DIGIT ZERO..NYIAKENG PUACHUE HMONG DIGIT NINE
    // 0x1E14E .. 0x1E14E: Lo         NYIAKENG PUACHUE HMONG LOGOGRAM NYAJ
    // 0x1E14F .. 0x1E14F: So         NYIAKENG PUACHUE HMONG CIRCLED CA
    Prop<::unicode::EastAsianWidth> { { 0x1E14E, 0x1E14F }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x1E290 .. 0x1E2AD: Lo    [30] TOTO LETTER PA..TOTO LETTER A
    // 0x1E2AE .. 0x1E2AE: Mn         TOTO SIGN RISING TONE
    Prop<::unicode::EastAsianWidth> { { 0x1E290, 0x1E2AE }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x1E2C0 .. 0x1E2EB: Lo    [44] WANCHO LETTER AA..WANCHO LETTER YIH
    // 0x1E2EC .. 0x1E2EF: Mn     [4] WANCHO TONE TUP..WANCHO TONE KOINI
    // 0x1E2F0 .. 0x1E2F9: Nd    [10] WANCHO DIGIT ZERO..WANCHO DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x1E2C0, 0x1E2F9 }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x1E2FF, 0x1E2FF }, ::unicode::EastAsianWidth::Neutral }, // Sc         WANCHO NGUN SIGN
    // 0x1E4D0 .. 0x1E4EA: Lo    [27] NAG MUNDARI LETTER O..NAG MUNDARI LETTER ELL
    // 0x1E4EB .. 0x1E4EB: Lm         NAG MUNDARI SIGN OJOD
    // 0x1E4EC .. 0x1E4EF: Mn     [4] NAG MUNDARI SIGN MUHOR..NAG MUNDARI SIGN SUTUH
    // 0x1E4F0 .. 0x1E4F9: Nd    [10] NAG MUNDARI DIGIT ZERO..NAG MUNDARI DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x1E4D0, 0x1E4F9 }, ::unicode::EastAsianWidth::Neutral }, // #4
    // 0x1E5D0 .. 0x1E5ED: Lo    [30] OL ONAL LETTER O..OL ONAL LETTER EG
    // 0x1E5EE .. 0x1E5EF: Mn     [2] OL ONAL SIGN MU..OL ONAL SIGN IKIR
    // 0x1E5F0 .. 0x1E5F0: Lo         OL ONAL SIGN HODDOND
    // 0x1E5F1 .. 0x1E5FA: Nd    [10] OL ONAL DIGIT ZERO..OL ONAL DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x1E5D0, 0x1E5FA }, ::unicode::EastAsianWidth::Neutral }, // #4
    Prop<::unicode::EastAsianWidth> { { 0x1E5FF, 0x1E5FF }, ::unicode::EastAsianWidth::Neutral }, // Po         OL ONAL ABBREVIATION SIGN
    Prop<::unicode::EastAsianWidth> { { 0x1E6C0, 0x1E6DE }, ::unicode::EastAsianWidth::Neutral }, // Lo    [31] TAI YO LETTER LOW KO..TAI YO LETTER HIGH KVO
    // 0x1E6E0 .. 0x1E6E2: Lo     [3] TAI YO LETTER AA..TAI YO LETTER UE
    // 0x1E6E3 .. 0x1E6E3: Mn         TAI YO SIGN UE
    // 0x1E6E4 .. 0x1E6E5: Lo     [2] TAI YO LETTER U..TAI YO LETTER AE
    // 0x1E6E6 .. 0x1E6E6: Mn         TAI YO SIGN AU
    // 0x1E6E7 .. 0x1E6ED: Lo     [7] TAI YO LETTER O..TAI YO LETTER AUE
    // 0x1E6EE .. 0x1E6EF: Mn     [2] TAI YO SIGN AY..TAI YO SIGN ANG
    // 0x1E6F0 .. 0x1E6F4: Lo     [5] TAI YO LETTER AN..TAI YO LETTER AP
    // 0x1E6F5 .. 0x1E6F5: Mn         TAI YO SIGN OM
    Prop<::unicode::EastAsianWidth> { { 0x1E6E0, 0x1E6F5 }, ::unicode::EastAsianWidth::Neutral }, // #8
    // 0x1E6FE .. 0x1E6FE: Lo         TAI YO SYMBOL MUEANG
    // 0x1E6FF .. 0x1E6FF: Lm         TAI YO XAM LAI
    Prop<::unicode::EastAsianWidth> { { 0x1E6FE, 0x1E6FF }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x1E7E0, 0x1E7E6 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ETHIOPIC SYLLABLE HHYA..ETHIOPIC SYLLABLE HHYO
    Prop<::unicode::EastAsianWidth> { { 0x1E7E8, 0x1E7EB }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] ETHIOPIC SYLLABLE GURAGE HHWA..ETHIOPIC SYLLABLE HHWE
    Prop<::unicode::EastAsianWidth> { { 0x1E7ED, 0x1E7EE }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] ETHIOPIC SYLLABLE GURAGE MWI..ETHIOPIC SYLLABLE GURAGE MWEE
    Prop<::unicode::EastAsianWidth> { { 0x1E7F0, 0x1E7FE }, ::unicode::EastAsianWidth::Neutral }, // Lo    [15] ETHIOPIC SYLLABLE GURAGE QWI..ETHIOPIC SYLLABLE GURAGE PWEE
    Prop<::unicode::EastAsianWidth> { { 0x1E800, 0x1E8C4 }, ::unicode::EastAsianWidth::Neutral }, // Lo   [197] MENDE KIKAKUI SYLLABLE M001 KI..MENDE KIKAKUI SYLLABLE M060 NYON
    // 0x1E8C7 .. 0x1E8CF: No     [9] MENDE KIKAKUI DIGIT ONE..MENDE KIKAKUI DIGIT NINE
    // 0x1E8D0 .. 0x1E8D6: Mn     [7] MENDE KIKAKUI COMBINING NUMBER TEENS..MENDE KIKAKUI COMBINING NUMBER MILLIONS
    Prop<::unicode::EastAsianWidth> { { 0x1E8C7, 0x1E8D6 }, ::unicode::EastAsianWidth::Neutral }, // #2
    // 0x1E900 .. 0x1E943: L&    [68] ADLAM CAPITAL LETTER ALIF..ADLAM SMALL LETTER SHA
    // 0x1E944 .. 0x1E94A: Mn     [7] ADLAM ALIF LENGTHENER..ADLAM NUKTA
    // 0x1E94B .. 0x1E94B: Lm         ADLAM NASALIZATION MARK
    Prop<::unicode::EastAsianWidth> { { 0x1E900, 0x1E94B }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x1E950, 0x1E959 }, ::unicode::EastAsianWidth::Neutral }, // Nd    [10] ADLAM DIGIT ZERO..ADLAM DIGIT NINE
    Prop<::unicode::EastAsianWidth> { { 0x1E95E, 0x1E95F }, ::unicode::EastAsianWidth::Neutral }, // Po     [2] ADLAM INITIAL EXCLAMATION MARK..ADLAM INITIAL QUESTION MARK
    // 0x1EC71 .. 0x1ECAB: No    [59] INDIC SIYAQ NUMBER ONE..INDIC SIYAQ NUMBER PREFIXED NINE
    // 0x1ECAC .. 0x1ECAC: So         INDIC SIYAQ PLACEHOLDER
    // 0x1ECAD .. 0x1ECAF: No     [3] INDIC SIYAQ FRACTION ONE QUARTER..INDIC SIYAQ FRACTION THREE QUARTERS
    // 0x1ECB0 .. 0x1ECB0: Sc         INDIC SIYAQ RUPEE MARK
    // 0x1ECB1 .. 0x1ECB4: No     [4] INDIC SIYAQ NUMBER ALTERNATE ONE..INDIC SIYAQ ALTERNATE LAKH MARK
    Prop<::unicode::EastAsianWidth> { { 0x1EC71, 0x1ECB4 }, ::unicode::EastAsianWidth::Neutral }, // #5
    // 0x1ED01 .. 0x1ED2D: No    [45] OTTOMAN SIYAQ NUMBER ONE..OTTOMAN SIYAQ NUMBER NINETY THOUSAND
    // 0x1ED2E .. 0x1ED2E: So         OTTOMAN SIYAQ MARRATAN
    // 0x1ED2F .. 0x1ED3D: No    [15] OTTOMAN SIYAQ ALTERNATE NUMBER TWO..OTTOMAN SIYAQ FRACTION ONE SIXTH
    Prop<::unicode::EastAsianWidth> { { 0x1ED01, 0x1ED3D }, ::unicode::EastAsianWidth::Neutral }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x1EE00, 0x1EE03 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] ARABIC MATHEMATICAL ALEF..ARABIC MATHEMATICAL DAL
    Prop<::unicode::EastAsianWidth> { { 0x1EE05, 0x1EE1F }, ::unicode::EastAsianWidth::Neutral }, // Lo    [27] ARABIC MATHEMATICAL WAW..ARABIC MATHEMATICAL DOTLESS QAF
    Prop<::unicode::EastAsianWidth> { { 0x1EE21, 0x1EE22 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] ARABIC MATHEMATICAL INITIAL BEH..ARABIC MATHEMATICAL INITIAL JEEM
    Prop<::unicode::EastAsianWidth> { { 0x1EE24, 0x1EE24 }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL INITIAL HEH
    Prop<::unicode::EastAsianWidth> { { 0x1EE27, 0x1EE27 }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL INITIAL HAH
    Prop<::unicode::EastAsianWidth> { { 0x1EE29, 0x1EE32 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [10] ARABIC MATHEMATICAL INITIAL YEH..ARABIC MATHEMATICAL INITIAL QAF
    Prop<::unicode::EastAsianWidth> { { 0x1EE34, 0x1EE37 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] ARABIC MATHEMATICAL INITIAL SHEEN..ARABIC MATHEMATICAL INITIAL KHAH
    Prop<::unicode::EastAsianWidth> { { 0x1EE39, 0x1EE39 }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL INITIAL DAD
    Prop<::unicode::EastAsianWidth> { { 0x1EE3B, 0x1EE3B }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL INITIAL GHAIN
    Prop<::unicode::EastAsianWidth> { { 0x1EE42, 0x1EE42 }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL TAILED JEEM
    Prop<::unicode::EastAsianWidth> { { 0x1EE47, 0x1EE47 }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL TAILED HAH
    Prop<::unicode::EastAsianWidth> { { 0x1EE49, 0x1EE49 }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL TAILED YEH
    Prop<::unicode::EastAsianWidth> { { 0x1EE4B, 0x1EE4B }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL TAILED LAM
    Prop<::unicode::EastAsianWidth> { { 0x1EE4D, 0x1EE4F }, ::unicode::EastAsianWidth::Neutral }, // Lo     [3] ARABIC MATHEMATICAL TAILED NOON..ARABIC MATHEMATICAL TAILED AIN
    Prop<::unicode::EastAsianWidth> { { 0x1EE51, 0x1EE52 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] ARABIC MATHEMATICAL TAILED SAD..ARABIC MATHEMATICAL TAILED QAF
    Prop<::unicode::EastAsianWidth> { { 0x1EE54, 0x1EE54 }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL TAILED SHEEN
    Prop<::unicode::EastAsianWidth> { { 0x1EE57, 0x1EE57 }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL TAILED KHAH
    Prop<::unicode::EastAsianWidth> { { 0x1EE59, 0x1EE59 }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL TAILED DAD
    Prop<::unicode::EastAsianWidth> { { 0x1EE5B, 0x1EE5B }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL TAILED GHAIN
    Prop<::unicode::EastAsianWidth> { { 0x1EE5D, 0x1EE5D }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL TAILED DOTLESS NOON
    Prop<::unicode::EastAsianWidth> { { 0x1EE5F, 0x1EE5F }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL TAILED DOTLESS QAF
    Prop<::unicode::EastAsianWidth> { { 0x1EE61, 0x1EE62 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [2] ARABIC MATHEMATICAL STRETCHED BEH..ARABIC MATHEMATICAL STRETCHED JEEM
    Prop<::unicode::EastAsianWidth> { { 0x1EE64, 0x1EE64 }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL STRETCHED HEH
    Prop<::unicode::EastAsianWidth> { { 0x1EE67, 0x1EE6A }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] ARABIC MATHEMATICAL STRETCHED HAH..ARABIC MATHEMATICAL STRETCHED KAF
    Prop<::unicode::EastAsianWidth> { { 0x1EE6C, 0x1EE72 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [7] ARABIC MATHEMATICAL STRETCHED MEEM..ARABIC MATHEMATICAL STRETCHED QAF
    Prop<::unicode::EastAsianWidth> { { 0x1EE74, 0x1EE77 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] ARABIC MATHEMATICAL STRETCHED SHEEN..ARABIC MATHEMATICAL STRETCHED KHAH
    Prop<::unicode::EastAsianWidth> { { 0x1EE79, 0x1EE7C }, ::unicode::EastAsianWidth::Neutral }, // Lo     [4] ARABIC MATHEMATICAL STRETCHED DAD..ARABIC MATHEMATICAL STRETCHED DOTLESS BEH
    Prop<::unicode::EastAsianWidth> { { 0x1EE7E, 0x1EE7E }, ::unicode::EastAsianWidth::Neutral }, // Lo         ARABIC MATHEMATICAL STRETCHED DOTLESS FEH
    Prop<::unicode::EastAsianWidth> { { 0x1EE80, 0x1EE89 }, ::unicode::EastAsianWidth::Neutral }, // Lo    [10] ARABIC MATHEMATICAL LOOPED ALEF..ARABIC MATHEMATICAL LOOPED YEH
    Prop<::unicode::EastAsianWidth> { { 0x1EE8B, 0x1EE9B }, ::unicode::EastAsianWidth::Neutral }, // Lo    [17] ARABIC MATHEMATICAL LOOPED LAM..ARABIC MATHEMATICAL LOOPED GHAIN
    Prop<::unicode::EastAsianWidth> { { 0x1EEA1, 0x1EEA3 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [3] ARABIC MATHEMATICAL DOUBLE-STRUCK BEH..ARABIC MATHEMATICAL DOUBLE-STRUCK DAL
    Prop<::unicode::EastAsianWidth> { { 0x1EEA5, 0x1EEA9 }, ::unicode::EastAsianWidth::Neutral }, // Lo     [5] ARABIC MATHEMATICAL DOUBLE-STRUCK WAW..ARABIC MATHEMATICAL DOUBLE-STRUCK YEH
    Prop<::unicode::EastAsianWidth> { { 0x1EEAB, 0x1EEBB }, ::unicode::EastAsianWidth::Neutral }, // Lo    [17] ARABIC MATHEMATICAL DOUBLE-STRUCK LAM..ARABIC MATHEMATICAL DOUBLE-STRUCK GHAIN
    Prop<::unicode::EastAsianWidth> { { 0x1EEF0, 0x1EEF1 }, ::unicode::EastAsianWidth::Neutral }, // Sm     [2] ARABIC MATHEMATICAL OPERATOR MEEM WITH HAH WITH TATWEEL..ARABIC MATHEMATICAL OPERATOR HAH WITH DAL
    Prop<::unicode::EastAsianWidth> { { 0x1F000, 0x1F003 }, ::unicode::EastAsianWidth::Neutral }, // So     [4] MAHJONG TILE EAST WIND..MAHJONG TILE NORTH WIND
    Prop<::unicode::EastAsianWidth> { { 0x1F004, 0x1F004 }, ::unicode::EastAsianWidth::Wide }, // So         MAHJONG TILE RED DRAGON
    Prop<::unicode::EastAsianWidth> { { 0x1F005, 0x1F02B }, ::unicode::EastAsianWidth::Neutral }, // So    [39] MAHJONG TILE GREEN DRAGON..MAHJONG TILE BACK
    Prop<::unicode::EastAsianWidth> { { 0x1F030, 0x1F093 }, ::unicode::EastAsianWidth::Neutral }, // So   [100] DOMINO TILE HORIZONTAL BACK..DOMINO TILE VERTICAL-06-06
    Prop<::unicode::EastAsianWidth> { { 0x1F0A0, 0x1F0AE }, ::unicode::EastAsianWidth::Neutral }, // So    [15] PLAYING CARD BACK..PLAYING CARD KING OF SPADES
    Prop<::unicode::EastAsianWidth> { { 0x1F0B1, 0x1F0BF }, ::unicode::EastAsianWidth::Neutral }, // So    [15] PLAYING CARD ACE OF HEARTS..PLAYING CARD RED JOKER
    Prop<::unicode::EastAsianWidth> { { 0x1F0C1, 0x1F0CE }, ::unicode::EastAsianWidth::Neutral }, // So    [14] PLAYING CARD ACE OF DIAMONDS..PLAYING CARD KING OF DIAMONDS
    Prop<::unicode::EastAsianWidth> { { 0x1F0CF, 0x1F0CF }, ::unicode::EastAsianWidth::Wide }, // So         PLAYING CARD BLACK JOKER
    Prop<::unicode::EastAsianWidth> { { 0x1F0D1, 0x1F0F5 }, ::unicode::EastAsianWidth::Neutral }, // So    [37] PLAYING CARD ACE OF CLUBS..PLAYING CARD TRUMP-21
    Prop<::unicode::EastAsianWidth> { { 0x1F100, 0x1F10A }, ::unicode::EastAsianWidth::Ambiguous }, // No    [11] DIGIT ZERO FULL STOP..DIGIT NINE COMMA
    // 0x1F10B .. 0x1F10C: No     [2] DINGBAT CIRCLED SANS-SERIF DIGIT ZERO..DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT ZERO
    // 0x1F10D .. 0x1F10F: So     [3] CIRCLED ZERO WITH SLASH..CIRCLED DOLLAR SIGN WITH OVERLAID BACKSLASH
    Prop<::unicode::EastAsianWidth> { { 0x1F10B, 0x1F10F }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x1F110, 0x1F12D }, ::unicode::EastAsianWidth::Ambiguous }, // So    [30] PARENTHESIZED LATIN CAPITAL LETTER A..CIRCLED CD
    Prop<::unicode::EastAsianWidth> { { 0x1F12E, 0x1F12F }, ::unicode::EastAsianWidth::Neutral }, // So     [2] CIRCLED WZ..COPYLEFT SYMBOL
    Prop<::unicode::EastAsianWidth> { { 0x1F130, 0x1F169 }, ::unicode::EastAsianWidth::Ambiguous }, // So    [58] SQUARED LATIN CAPITAL LETTER A..NEGATIVE CIRCLED LATIN CAPITAL LETTER Z
    Prop<::unicode::EastAsianWidth> { { 0x1F16A, 0x1F16F }, ::unicode::EastAsianWidth::Neutral }, // So     [6] RAISED MC SIGN..CIRCLED HUMAN FIGURE
    Prop<::unicode::EastAsianWidth> { { 0x1F170, 0x1F18D }, ::unicode::EastAsianWidth::Ambiguous }, // So    [30] NEGATIVE SQUARED LATIN CAPITAL LETTER A..NEGATIVE SQUARED SA
    Prop<::unicode::EastAsianWidth> { { 0x1F18E, 0x1F18E }, ::unicode::EastAsianWidth::Wide }, // So         NEGATIVE SQUARED AB
    Prop<::unicode::EastAsianWidth> { { 0x1F18F, 0x1F190 }, ::unicode::EastAsianWidth::Ambiguous }, // So     [2] NEGATIVE SQUARED WC..SQUARE DJ
    Prop<::unicode::EastAsianWidth> { { 0x1F191, 0x1F19A }, ::unicode::EastAsianWidth::Wide }, // So    [10] SQUARED CL..SQUARED VS
    Prop<::unicode::EastAsianWidth> { { 0x1F19B, 0x1F1AC }, ::unicode::EastAsianWidth::Ambiguous }, // So    [18] SQUARED THREE D..SQUARED VOD
    Prop<::unicode::EastAsianWidth> { { 0x1F1AD, 0x1F1AD }, ::unicode::EastAsianWidth::Neutral }, // So         MASK WORK SYMBOL
    Prop<::unicode::EastAsianWidth> { { 0x1F1E6, 0x1F1FF }, ::unicode::EastAsianWidth::Neutral }, // So    [26] REGIONAL INDICATOR SYMBOL LETTER A..REGIONAL INDICATOR SYMBOL LETTER Z
    Prop<::unicode::EastAsianWidth> { { 0x1F200, 0x1F202 }, ::unicode::EastAsianWidth::Wide }, // So     [3] SQUARE HIRAGANA HOKA..SQUARED KATAKANA SA
    Prop<::unicode::EastAsianWidth> { { 0x1F210, 0x1F23B }, ::unicode::EastAsianWidth::Wide }, // So    [44] SQUARED CJK UNIFIED IDEOGRAPH-624B..SQUARED CJK UNIFIED IDEOGRAPH-914D
    Prop<::unicode::EastAsianWidth> { { 0x1F240, 0x1F248 }, ::unicode::EastAsianWidth::Wide }, // So     [9] TORTOISE SHELL BRACKETED CJK UNIFIED IDEOGRAPH-672C..TORTOISE SHELL BRACKETED CJK UNIFIED IDEOGRAPH-6557
    Prop<::unicode::EastAsianWidth> { { 0x1F250, 0x1F251 }, ::unicode::EastAsianWidth::Wide }, // So     [2] CIRCLED IDEOGRAPH ADVANTAGE..CIRCLED IDEOGRAPH ACCEPT
    Prop<::unicode::EastAsianWidth> { { 0x1F260, 0x1F265 }, ::unicode::EastAsianWidth::Wide }, // So     [6] ROUNDED SYMBOL FOR FU..ROUNDED SYMBOL FOR CAI
    Prop<::unicode::EastAsianWidth> { { 0x1F300, 0x1F320 }, ::unicode::EastAsianWidth::Wide }, // So    [33] CYCLONE..SHOOTING STAR
    Prop<::unicode::EastAsianWidth> { { 0x1F321, 0x1F32C }, ::unicode::EastAsianWidth::Neutral }, // So    [12] THERMOMETER..WIND BLOWING FACE
    Prop<::unicode::EastAsianWidth> { { 0x1F32D, 0x1F335 }, ::unicode::EastAsianWidth::Wide }, // So     [9] HOT DOG..CACTUS
    Prop<::unicode::EastAsianWidth> { { 0x1F336, 0x1F336 }, ::unicode::EastAsianWidth::Neutral }, // So         HOT PEPPER
    Prop<::unicode::EastAsianWidth> { { 0x1F337, 0x1F37C }, ::unicode::EastAsianWidth::Wide }, // So    [70] TULIP..BABY BOTTLE
    Prop<::unicode::EastAsianWidth> { { 0x1F37D, 0x1F37D }, ::unicode::EastAsianWidth::Neutral }, // So         FORK AND KNIFE WITH PLATE
    Prop<::unicode::EastAsianWidth> { { 0x1F37E, 0x1F393 }, ::unicode::EastAsianWidth::Wide }, // So    [22] BOTTLE WITH POPPING CORK..GRADUATION CAP
    Prop<::unicode::EastAsianWidth> { { 0x1F394, 0x1F39F }, ::unicode::EastAsianWidth::Neutral }, // So    [12] HEART WITH TIP ON THE LEFT..ADMISSION TICKETS
    Prop<::unicode::EastAsianWidth> { { 0x1F3A0, 0x1F3CA }, ::unicode::EastAsianWidth::Wide }, // So    [43] CAROUSEL HORSE..SWIMMER
    Prop<::unicode::EastAsianWidth> { { 0x1F3CB, 0x1F3CE }, ::unicode::EastAsianWidth::Neutral }, // So     [4] WEIGHT LIFTER..RACING CAR
    Prop<::unicode::EastAsianWidth> { { 0x1F3CF, 0x1F3D3 }, ::unicode::EastAsianWidth::Wide }, // So     [5] CRICKET BAT AND BALL..TABLE TENNIS PADDLE AND BALL
    Prop<::unicode::EastAsianWidth> { { 0x1F3D4, 0x1F3DF }, ::unicode::EastAsianWidth::Neutral }, // So    [12] SNOW CAPPED MOUNTAIN..STADIUM
    Prop<::unicode::EastAsianWidth> { { 0x1F3E0, 0x1F3F0 }, ::unicode::EastAsianWidth::Wide }, // So    [17] HOUSE BUILDING..EUROPEAN CASTLE
    Prop<::unicode::EastAsianWidth> { { 0x1F3F1, 0x1F3F3 }, ::unicode::EastAsianWidth::Neutral }, // So     [3] WHITE PENNANT..WAVING WHITE FLAG
    Prop<::unicode::EastAsianWidth> { { 0x1F3F4, 0x1F3F4 }, ::unicode::EastAsianWidth::Wide }, // So         WAVING BLACK FLAG
    Prop<::unicode::EastAsianWidth> { { 0x1F3F5, 0x1F3F7 }, ::unicode::EastAsianWidth::Neutral }, // So     [3] ROSETTE..LABEL
    // 0x1F3F8 .. 0x1F3FA: So     [3] BADMINTON RACQUET AND SHUTTLECOCK..AMPHORA
    // 0x1F3FB .. 0x1F3FF: Sk     [5] EMOJI MODIFIER FITZPATRICK TYPE-1-2..EMOJI MODIFIER FITZPATRICK TYPE-6
    // 0x1F400 .. 0x1F43E: So    [63] RAT..PAW PRINTS
    Prop<::unicode::EastAsianWidth> { { 0x1F3F8, 0x1F43E }, ::unicode::EastAsianWidth::Wide }, // #3
    Prop<::unicode::EastAsianWidth> { { 0x1F43F, 0x1F43F }, ::unicode::EastAsianWidth::Neutral }, // So         CHIPMUNK
    Prop<::unicode::EastAsianWidth> { { 0x1F440, 0x1F440 }, ::unicode::EastAsianWidth::Wide }, // So         EYES
    Prop<::unicode::EastAsianWidth> { { 0x1F441, 0x1F441 }, ::unicode::EastAsianWidth::Neutral }, // So         EYE
    Prop<::unicode::EastAsianWidth> { { 0x1F442, 0x1F4FC }, ::unicode::EastAsianWidth::Wide }, // So   [187] EAR..VIDEOCASSETTE
    Prop<::unicode::EastAsianWidth> { { 0x1F4FD, 0x1F4FE }, ::unicode::EastAsianWidth::Neutral }, // So     [2] FILM PROJECTOR..PORTABLE STEREO
    Prop<::unicode::EastAsianWidth> { { 0x1F4FF, 0x1F53D }, ::unicode::EastAsianWidth::Wide }, // So    [63] PRAYER BEADS..DOWN-POINTING SMALL RED TRIANGLE
    Prop<::unicode::EastAsianWidth> { { 0x1F53E, 0x1F54A }, ::unicode::EastAsianWidth::Neutral }, // So    [13] LOWER RIGHT SHADOWED WHITE CIRCLE..DOVE OF PEACE
    Prop<::unicode::EastAsianWidth> { { 0x1F54B, 0x1F54E }, ::unicode::EastAsianWidth::Wide }, // So     [4] KAABA..MENORAH WITH NINE BRANCHES
    Prop<::unicode::EastAsianWidth> { { 0x1F54F, 0x1F54F }, ::unicode::EastAsianWidth::Neutral }, // So         BOWL OF HYGIEIA
    Prop<::unicode::EastAsianWidth> { { 0x1F550, 0x1F567 }, ::unicode::EastAsianWidth::Wide }, // So    [24] CLOCK FACE ONE OCLOCK..CLOCK FACE TWELVE-THIRTY
    Prop<::unicode::EastAsianWidth> { { 0x1F568, 0x1F579 }, ::unicode::EastAsianWidth::Neutral }, // So    [18] RIGHT SPEAKER..JOYSTICK
    Prop<::unicode::EastAsianWidth> { { 0x1F57A, 0x1F57A }, ::unicode::EastAsianWidth::Wide }, // So         MAN DANCING
    Prop<::unicode::EastAsianWidth> { { 0x1F57B, 0x1F594 }, ::unicode::EastAsianWidth::Neutral }, // So    [26] LEFT HAND TELEPHONE RECEIVER..REVERSED VICTORY HAND
    Prop<::unicode::EastAsianWidth> { { 0x1F595, 0x1F596 }, ::unicode::EastAsianWidth::Wide }, // So     [2] REVERSED HAND WITH MIDDLE FINGER EXTENDED..RAISED HAND WITH PART BETWEEN MIDDLE AND RING FINGERS
    Prop<::unicode::EastAsianWidth> { { 0x1F597, 0x1F5A3 }, ::unicode::EastAsianWidth::Neutral }, // So    [13] WHITE DOWN POINTING LEFT HAND INDEX..BLACK DOWN POINTING BACKHAND INDEX
    Prop<::unicode::EastAsianWidth> { { 0x1F5A4, 0x1F5A4 }, ::unicode::EastAsianWidth::Wide }, // So         BLACK HEART
    Prop<::unicode::EastAsianWidth> { { 0x1F5A5, 0x1F5FA }, ::unicode::EastAsianWidth::Neutral }, // So    [86] DESKTOP COMPUTER..WORLD MAP
    // 0x1F5FB .. 0x1F5FF: So     [5] MOUNT FUJI..MOYAI
    // 0x1F600 .. 0x1F64F: So    [80] GRINNING FACE..PERSON WITH FOLDED HANDS
    Prop<::unicode::EastAsianWidth> { { 0x1F5FB, 0x1F64F }, ::unicode::EastAsianWidth::Wide }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x1F650, 0x1F67F }, ::unicode::EastAsianWidth::Neutral }, // So    [48] NORTH WEST POINTING LEAF..REVERSE CHECKER BOARD
    Prop<::unicode::EastAsianWidth> { { 0x1F680, 0x1F6C5 }, ::unicode::EastAsianWidth::Wide }, // So    [70] ROCKET..LEFT LUGGAGE
    Prop<::unicode::EastAsianWidth> { { 0x1F6C6, 0x1F6CB }, ::unicode::EastAsianWidth::Neutral }, // So     [6] TRIANGLE WITH ROUNDED CORNERS..COUCH AND LAMP
    Prop<::unicode::EastAsianWidth> { { 0x1F6CC, 0x1F6CC }, ::unicode::EastAsianWidth::Wide }, // So         SLEEPING ACCOMMODATION
    Prop<::unicode::EastAsianWidth> { { 0x1F6CD, 0x1F6CF }, ::unicode::EastAsianWidth::Neutral }, // So     [3] SHOPPING BAGS..BED
    Prop<::unicode::EastAsianWidth> { { 0x1F6D0, 0x1F6D2 }, ::unicode::EastAsianWidth::Wide }, // So     [3] PLACE OF WORSHIP..SHOPPING TROLLEY
    Prop<::unicode::EastAsianWidth> { { 0x1F6D3, 0x1F6D4 }, ::unicode::EastAsianWidth::Neutral }, // So     [2] STUPA..PAGODA
    Prop<::unicode::EastAsianWidth> { { 0x1F6D5, 0x1F6D8 }, ::unicode::EastAsianWidth::Wide }, // So     [4] HINDU TEMPLE..LANDSLIDE
    Prop<::unicode::EastAsianWidth> { { 0x1F6DC, 0x1F6DF }, ::unicode::EastAsianWidth::Wide }, // So     [4] WIRELESS..RING BUOY
    Prop<::unicode::EastAsianWidth> { { 0x1F6E0, 0x1F6EA }, ::unicode::EastAsianWidth::Neutral }, // So    [11] HAMMER AND WRENCH..NORTHEAST-POINTING AIRPLANE
    Prop<::unicode::EastAsianWidth> { { 0x1F6EB, 0x1F6EC }, ::unicode::EastAsianWidth::Wide }, // So     [2] AIRPLANE DEPARTURE..AIRPLANE ARRIVING
    Prop<::unicode::EastAsianWidth> { { 0x1F6F0, 0x1F6F3 }, ::unicode::EastAsianWidth::Neutral }, // So     [4] SATELLITE..PASSENGER SHIP
    Prop<::unicode::EastAsianWidth> { { 0x1F6F4, 0x1F6FC }, ::unicode::EastAsianWidth::Wide }, // So     [9] SCOOTER..ROLLER SKATE
    // 0x1F700 .. 0x1F77F: So   [128] ALCHEMICAL SYMBOL FOR QUINTESSENCE..ORCUS
    // 0x1F780 .. 0x1F7D9: So    [90] BLACK LEFT-POINTING ISOSCELES RIGHT TRIANGLE..NINE POINTED WHITE STAR
    Prop<::unicode::EastAsianWidth> { { 0x1F700, 0x1F7D9 }, ::unicode::EastAsianWidth::Neutral }, // #2
    Prop<::unicode::EastAsianWidth> { { 0x1F7E0, 0x1F7EB }, ::unicode::EastAsianWidth::Wide }, // So    [12] LARGE ORANGE CIRCLE..LARGE BROWN SQUARE
    Prop<::unicode::EastAsianWidth> { { 0x1F7F0, 0x1F7F0 }, ::unicode::EastAsianWidth::Wide }, // So         HEAVY EQUALS SIGN
    Prop<::unicode::EastAsianWidth> { { 0x1F800, 0x1F80B }, ::unicode::EastAsianWidth::Neutral }, // So    [12] LEFTWARDS ARROW WITH SMALL TRIANGLE ARROWHEAD..DOWNWARDS ARROW WITH LARGE TRIANGLE ARROWHEAD
    Prop<::unicode::EastAsianWidth> { { 0x1F810, 0x1F847 }, ::unicode::EastAsianWidth::Neutral }, // So    [56] LEFTWARDS ARROW WITH SMALL EQUILATERAL ARROWHEAD..DOWNWARDS HEAVY ARROW
    Prop<::unicode::EastAsianWidth> { { 0x1F850, 0x1F859 }, ::unicode::EastAsianWidth::Neutral }, // So    [10] LEFTWARDS SANS-SERIF ARROW..UP DOWN SANS-SERIF ARROW
    Prop<::unicode::EastAsianWidth> { { 0x1F860, 0x1F887 }, ::unicode::EastAsianWidth::Neutral }, // So    [40] WIDE-HEADED LEFTWARDS LIGHT BARB ARROW..WIDE-HEADED SOUTH WEST VERY HEAVY BARB ARROW
    Prop<::unicode::EastAsianWidth> { { 0x1F890, 0x1F8AD }, ::unicode::EastAsianWidth::Neutral }, // So    [30] LEFTWARDS TRIANGLE ARROWHEAD..WHITE ARROW SHAFT WIDTH TWO THIRDS
    Prop<::unicode::EastAsianWidth> { { 0x1F8B0, 0x1F8BB }, ::unicode::EastAsianWidth::Neutral }, // So    [12] ARROW POINTING UPWARDS THEN NORTH WEST..SOUTH WEST ARROW FROM BAR
    Prop<::unicode::EastAsianWidth> { { 0x1F8C0, 0x1F8C1 }, ::unicode::EastAsianWidth::Neutral }, // So     [2] LEFTWARDS ARROW FROM DOWNWARDS ARROW..RIGHTWARDS ARROW FROM DOWNWARDS ARROW
    Prop<::unicode::EastAsianWidth> { { 0x1F8D0, 0x1F8D8 }, ::unicode::EastAsianWidth::Neutral }, // Sm     [9] LONG RIGHTWARDS ARROW OVER LONG LEFTWARDS ARROW..LONG LEFT RIGHT ARROW WITH DEPENDENT LOBE
    Prop<::unicode::EastAsianWidth> { { 0x1F900, 0x1F90B }, ::unicode::EastAsianWidth::Neutral }, // So    [12] CIRCLED CROSS FORMEE WITH FOUR DOTS..DOWNWARD FACING NOTCHED HOOK WITH DOT
    Prop<::unicode::EastAsianWidth> { { 0x1F90C, 0x1F93A }, ::unicode::EastAsianWidth::Wide }, // So    [47] PINCHED FINGERS..FENCER
    Prop<::unicode::EastAsianWidth> { { 0x1F93B, 0x1F93B }, ::unicode::EastAsianWidth::Neutral }, // So         MODERN PENTATHLON
    Prop<::unicode::EastAsianWidth> { { 0x1F93C, 0x1F945 }, ::unicode::EastAsianWidth::Wide }, // So    [10] WRESTLERS..GOAL NET
    Prop<::unicode::EastAsianWidth> { { 0x1F946, 0x1F946 }, ::unicode::EastAsianWidth::Neutral }, // So         RIFLE
    Prop<::unicode::EastAsianWidth> { { 0x1F947, 0x1F9FF }, ::unicode::EastAsianWidth::Wide }, // So   [185] FIRST PLACE MEDAL..NAZAR AMULET
    Prop<::unicode::EastAsianWidth> { { 0x1FA00, 0x1FA57 }, ::unicode::EastAsianWidth::Neutral }, // So    [88] NEUTRAL CHESS KING..BLACK CHESS ALFIL
    Prop<::unicode::EastAsianWidth> { { 0x1FA60, 0x1FA6D }, ::unicode::EastAsianWidth::Neutral }, // So    [14] XIANGQI RED GENERAL..XIANGQI BLACK SOLDIER
    Prop<::unicode::EastAsianWidth> { { 0x1FA70, 0x1FA7C }, ::unicode::EastAsianWidth::Wide }, // So    [13] BALLET SHOES..CRUTCH
    Prop<::unicode::EastAsianWidth> { { 0x1FA80, 0x1FA8A }, ::unicode::EastAsianWidth::Wide }, // So    [11] YO-YO..TROMBONE
    Prop<::unicode::EastAsianWidth> { { 0x1FA8E, 0x1FAC6 }, ::unicode::EastAsianWidth::Wide }, // So    [57] TREASURE CHEST..FINGERPRINT
    Prop<::unicode::EastAsianWidth> { { 0x1FAC8, 0x1FAC8 }, ::unicode::EastAsianWidth::Wide }, // So         HAIRY CREATURE
    Prop<::unicode::EastAsianWidth> { { 0x1FACD, 0x1FADC }, ::unicode::EastAsianWidth::Wide }, // So    [16] ORCA..ROOT VEGETABLE
    Prop<::unicode::EastAsianWidth> { { 0x1FADF, 0x1FAEA }, ::unicode::EastAsianWidth::Wide }, // So    [12] SPLATTER..DISTORTED FACE
    Prop<::unicode::EastAsianWidth> { { 0x1FAEF, 0x1FAF8 }, ::unicode::EastAsianWidth::Wide }, // So    [10] FIGHT CLOUD..RIGHTWARDS PUSHING HAND
    Prop<::unicode::EastAsianWidth> { { 0x1FB00, 0x1FB92 }, ::unicode::EastAsianWidth::Neutral }, // So   [147] BLOCK SEXTANT-1..UPPER HALF INVERSE MEDIUM SHADE AND LOWER HALF BLOCK
    // 0x1FB94 .. 0x1FBEF: So    [92] LEFT HALF INVERSE MEDIUM SHADE AND RIGHT HALF BLOCK..TOP LEFT JUSTIFIED LOWER RIGHT QUARTER BLACK CIRCLE
    // 0x1FBF0 .. 0x1FBF9: Nd    [10] SEGMENTED DIGIT ZERO..SEGMENTED DIGIT NINE
    // 0x1FBFA .. 0x1FBFA: So         ALARM BELL SYMBOL
    Prop<::unicode::EastAsianWidth> { { 0x1FB94, 0x1FBFA }, ::unicode::EastAsianWidth::Neutral }, // #3
    // 0x20000 .. 0x2A6DF: Lo [42720] CJK UNIFIED IDEOGRAPH-20000..CJK UNIFIED IDEOGRAPH-2A6DF
    // 0x2A6E0 .. 0x2A6FF: Cn    [32] <reserved-2A6E0>..<reserved-2A6FF>
    // 0x2A700 .. 0x2B81D: Lo  [4382] CJK UNIFIED IDEOGRAPH-2A700..CJK UNIFIED IDEOGRAPH-2B81D
    // 0x2B81E .. 0x2B81F: Cn     [2] <reserved-2B81E>..<reserved-2B81F>
    // 0x2B820 .. 0x2CEAD: Lo  [5774] CJK UNIFIED IDEOGRAPH-2B820..CJK UNIFIED IDEOGRAPH-2CEAD
    // 0x2CEAE .. 0x2CEAF: Cn     [2] <reserved-2CEAE>..<reserved-2CEAF>
    // 0x2CEB0 .. 0x2EBE0: Lo  [7473] CJK UNIFIED IDEOGRAPH-2CEB0..CJK UNIFIED IDEOGRAPH-2EBE0
    // 0x2EBE1 .. 0x2EBEF: Cn    [15] <reserved-2EBE1>..<reserved-2EBEF>
    // 0x2EBF0 .. 0x2EE5D: Lo   [622] CJK UNIFIED IDEOGRAPH-2EBF0..CJK UNIFIED IDEOGRAPH-2EE5D
    // 0x2EE5E .. 0x2F7FF: Cn  [2466] <reserved-2EE5E>..<reserved-2F7FF>
    // 0x2F800 .. 0x2FA1D: Lo   [542] CJK COMPATIBILITY IDEOGRAPH-2F800..CJK COMPATIBILITY IDEOGRAPH-2FA1D
    // 0x2FA1E .. 0x2FA1F: Cn     [2] <reserved-2FA1E>..<reserved-2FA1F>
    // 0x2FA20 .. 0x2FFFD: Cn  [1502] <reserved-2FA20>..<reserved-2FFFD>
    Prop<::unicode::EastAsianWidth> { { 0x20000, 0x2FFFD }, ::unicode::EastAsianWidth::Wide }, // #13
    // 0x30000 .. 0x3134A: Lo  [4939] CJK UNIFIED IDEOGRAPH-30000..CJK UNIFIED IDEOGRAPH-3134A
    // 0x3134B .. 0x3134F: Cn     [5] <reserved-3134B>..<reserved-3134F>
    // 0x31350 .. 0x33479: Lo  [8490] CJK UNIFIED IDEOGRAPH-31350..CJK UNIFIED IDEOGRAPH-33479
    // 0x3347A .. 0x3FFFD: Cn [52100] <reserved-3347A>..<reserved-3FFFD>
    Prop<::unicode::EastAsianWidth> { { 0x30000, 0x3FFFD }, ::unicode::EastAsianWidth::Wide }, // #4
    Prop<::unicode::EastAsianWidth> { { 0xE0001, 0xE0001 }, ::unicode::EastAsianWidth::Neutral }, // Cf         LANGUAGE TAG
    Prop<::unicode::EastAsianWidth> { { 0xE0020, 0xE007F }, ::unicode::EastAsianWidth::Neutral }, // Cf    [96] TAG SPACE..CANCEL TAG
    Prop<::unicode::EastAsianWidth> { { 0xE0100, 0xE01EF }, ::unicode::EastAsianWidth::Ambiguous }, // Mn   [240] VARIATION SELECTOR-17..VARIATION SELECTOR-256
    Prop<::unicode::EastAsianWidth> { { 0xF0000, 0xFFFFD }, ::unicode::EastAsianWidth::Ambiguous }, // Co [65534] <private-use-F0000>..<private-use-FFFFD>
    Prop<::unicode::EastAsianWidth> { { 0x100000, 0x10FFFD }, ::unicode::EastAsianWidth::Ambiguous }, // Co [65534] <private-use-100000>..<private-use-10FFFD>
}; // }}}
} // end namespace tables

EastAsianWidth east_asian_width(char32_t codepoint) noexcept {
    return search(tables::EastAsianWidth, codepoint).value_or(EastAsianWidth::Unspecified);
}

namespace tables {
auto static const Emoji = std::array<Interval, 418>{ // {{{
    Interval{ 0x0023, 0x0023 }, // E0.0   [1] (#️)       hash sign
    Interval{ 0x002A, 0x002A }, // E0.0   [1] (*️)       asterisk
    Interval{ 0x0030, 0x0039 }, // E0.0  [10] (0️..9️)    digit zero..digit nine
    Interval{ 0x00A9, 0x00A9 }, // E0.6   [1] (©️)       copyright
    Interval{ 0x00AE, 0x00AE }, // E0.6   [1] (®️)       registered
    Interval{ 0x203C, 0x203C }, // E0.6   [1] (‼️)       double exclamation mark
    Interval{ 0x2049, 0x2049 }, // E0.6   [1] (⁉️)       exclamation question mark
    Interval{ 0x2122, 0x2122 }, // E0.6   [1] (™️)       trade mark
    Interval{ 0x2139, 0x2139 }, // E0.6   [1] (ℹ️)       information
    Interval{ 0x2194, 0x2199 }, // E0.6   [6] (↔️..↙️)    left-right arrow..down-left arrow
    Interval{ 0x21A9, 0x21AA }, // E0.6   [2] (↩️..↪️)    right arrow curving left..left arrow curving right
    Interval{ 0x231A, 0x231B }, // E0.6   [2] (⌚..⌛)    watch..hourglass done
    Interval{ 0x2328, 0x2328 }, // E1.0   [1] (⌨️)       keyboard
    Interval{ 0x23CF, 0x23CF }, // E1.0   [1] (⏏️)       eject button
    Interval{ 0x23E9, 0x23EC }, // E0.6   [4] (⏩..⏬)    fast-forward button..fast down button
    Interval{ 0x23ED, 0x23EE }, // E0.7   [2] (⏭️..⏮️)    next track button..last track button
    Interval{ 0x23EF, 0x23EF }, // E1.0   [1] (⏯️)       play or pause button
    Interval{ 0x23F0, 0x23F0 }, // E0.6   [1] (⏰)       alarm clock
    Interval{ 0x23F1, 0x23F2 }, // E1.0   [2] (⏱️..⏲️)    stopwatch..timer clock
    Interval{ 0x23F3, 0x23F3 }, // E0.6   [1] (⏳)       hourglass not done
    Interval{ 0x23F8, 0x23FA }, // E0.7   [3] (⏸️..⏺️)    pause button..record button
    Interval{ 0x24C2, 0x24C2 }, // E0.6   [1] (Ⓜ️)       circled M
    Interval{ 0x25AA, 0x25AB }, // E0.6   [2] (▪️..▫️)    black small square..white small square
    Interval{ 0x25B6, 0x25B6 }, // E0.6   [1] (▶️)       play button
    Interval{ 0x25C0, 0x25C0 }, // E0.6   [1] (◀️)       reverse button
    Interval{ 0x25FB, 0x25FE }, // E0.6   [4] (◻️..◾)    white medium square..black medium-small square
    Interval{ 0x2600, 0x2601 }, // E0.6   [2] (☀️..☁️)    sun..cloud
    Interval{ 0x2602, 0x2603 }, // E0.7   [2] (☂️..☃️)    umbrella..snowman
    Interval{ 0x2604, 0x2604 }, // E1.0   [1] (☄️)       comet
    Interval{ 0x260E, 0x260E }, // E0.6   [1] (☎️)       telephone
    Interval{ 0x2611, 0x2611 }, // E0.6   [1] (☑️)       check box with check
    Interval{ 0x2614, 0x2615 }, // E0.6   [2] (☔..☕)    umbrella with rain drops..hot beverage
    Interval{ 0x2618, 0x2618 }, // E1.0   [1] (☘️)       shamrock
    Interval{ 0x261D, 0x261D }, // E0.6   [1] (☝️)       index pointing up
    Interval{ 0x2620, 0x2620 }, // E1.0   [1] (☠️)       skull and crossbones
    Interval{ 0x2622, 0x2623 }, // E1.0   [2] (☢️..☣️)    radioactive..biohazard
    Interval{ 0x2626, 0x2626 }, // E1.0   [1] (☦️)       orthodox cross
    Interval{ 0x262A, 0x262A }, // E0.7   [1] (☪️)       star and crescent
    Interval{ 0x262E, 0x262E }, // E1.0   [1] (☮️)       peace symbol
    Interval{ 0x262F, 0x262F }, // E0.7   [1] (☯️)       yin yang
    Interval{ 0x2638, 0x2639 }, // E0.7   [2] (☸️..☹️)    wheel of dharma..frowning face
    Interval{ 0x263A, 0x263A }, // E0.6   [1] (☺️)       smiling face
    Interval{ 0x2640, 0x2640 }, // E4.0   [1] (♀️)       female sign
    Interval{ 0x2642, 0x2642 }, // E4.0   [1] (♂️)       male sign
    Interval{ 0x2648, 0x2653 }, // E0.6  [12] (♈..♓)    Aries..Pisces
    Interval{ 0x265F, 0x265F }, // E11.0  [1] (♟️)       chess pawn
    Interval{ 0x2660, 0x2660 }, // E0.6   [1] (♠️)       spade suit
    Interval{ 0x2663, 0x2663 }, // E0.6   [1] (♣️)       club suit
    Interval{ 0x2665, 0x2666 }, // E0.6   [2] (♥️..♦️)    heart suit..diamond suit
    Interval{ 0x2668, 0x2668 }, // E0.6   [1] (♨️)       hot springs
    Interval{ 0x267B, 0x267B }, // E0.6   [1] (♻️)       recycling symbol
    Interval{ 0x267E, 0x267E }, // E11.0  [1] (♾️)       infinity
    Interval{ 0x267F, 0x267F }, // E0.6   [1] (♿)       wheelchair symbol
    Interval{ 0x2692, 0x2692 }, // E1.0   [1] (⚒️)       hammer and pick
    Interval{ 0x2693, 0x2693 }, // E0.6   [1] (⚓)       anchor
    Interval{ 0x2694, 0x2694 }, // E1.0   [1] (⚔️)       crossed swords
    Interval{ 0x2695, 0x2695 }, // E4.0   [1] (⚕️)       medical symbol
    Interval{ 0x2696, 0x2697 }, // E1.0   [2] (⚖️..⚗️)    balance scale..alembic
    Interval{ 0x2699, 0x2699 }, // E1.0   [1] (⚙️)       gear
    Interval{ 0x269B, 0x269C }, // E1.0   [2] (⚛️..⚜️)    atom symbol..fleur-de-lis
    Interval{ 0x26A0, 0x26A1 }, // E0.6   [2] (⚠️..⚡)    warning..high voltage
    Interval{ 0x26A7, 0x26A7 }, // E13.0  [1] (⚧️)       transgender symbol
    Interval{ 0x26AA, 0x26AB }, // E0.6   [2] (⚪..⚫)    white circle..black circle
    Interval{ 0x26B0, 0x26B1 }, // E1.0   [2] (⚰️..⚱️)    coffin..funeral urn
    Interval{ 0x26BD, 0x26BE }, // E0.6   [2] (⚽..⚾)    soccer ball..baseball
    Interval{ 0x26C4, 0x26C5 }, // E0.6   [2] (⛄..⛅)    snowman without snow..sun behind cloud
    Interval{ 0x26C8, 0x26C8 }, // E0.7   [1] (⛈️)       cloud with lightning and rain
    Interval{ 0x26CE, 0x26CE }, // E0.6   [1] (⛎)       Ophiuchus
    Interval{ 0x26CF, 0x26CF }, // E0.7   [1] (⛏️)       pick
    Interval{ 0x26D1, 0x26D1 }, // E0.7   [1] (⛑️)       rescue worker’s helmet
    Interval{ 0x26D3, 0x26D3 }, // E0.7   [1] (⛓️)       chains
    Interval{ 0x26D4, 0x26D4 }, // E0.6   [1] (⛔)       no entry
    Interval{ 0x26E9, 0x26E9 }, // E0.7   [1] (⛩️)       shinto shrine
    Interval{ 0x26EA, 0x26EA }, // E0.6   [1] (⛪)       church
    Interval{ 0x26F0, 0x26F1 }, // E0.7   [2] (⛰️..⛱️)    mountain..umbrella on ground
    Interval{ 0x26F2, 0x26F3 }, // E0.6   [2] (⛲..⛳)    fountain..flag in hole
    Interval{ 0x26F4, 0x26F4 }, // E0.7   [1] (⛴️)       ferry
    Interval{ 0x26F5, 0x26F5 }, // E0.6   [1] (⛵)       sailboat
    Interval{ 0x26F7, 0x26F9 }, // E0.7   [3] (⛷️..⛹️)    skier..person bouncing ball
    Interval{ 0x26FA, 0x26FA }, // E0.6   [1] (⛺)       tent
    Interval{ 0x26FD, 0x26FD }, // E0.6   [1] (⛽)       fuel pump
    Interval{ 0x2702, 0x2702 }, // E0.6   [1] (✂️)       scissors
    Interval{ 0x2705, 0x2705 }, // E0.6   [1] (✅)       check mark button
    Interval{ 0x2708, 0x270C }, // E0.6   [5] (✈️..✌️)    airplane..victory hand
    Interval{ 0x270D, 0x270D }, // E0.7   [1] (✍️)       writing hand
    Interval{ 0x270F, 0x270F }, // E0.6   [1] (✏️)       pencil
    Interval{ 0x2712, 0x2712 }, // E0.6   [1] (✒️)       black nib
    Interval{ 0x2714, 0x2714 }, // E0.6   [1] (✔️)       check mark
    Interval{ 0x2716, 0x2716 }, // E0.6   [1] (✖️)       multiply
    Interval{ 0x271D, 0x271D }, // E0.7   [1] (✝️)       latin cross
    Interval{ 0x2721, 0x2721 }, // E0.7   [1] (✡️)       star of David
    Interval{ 0x2728, 0x2728 }, // E0.6   [1] (✨)       sparkles
    Interval{ 0x2733, 0x2734 }, // E0.6   [2] (✳️..✴️)    eight-spoked asterisk..eight-pointed star
    Interval{ 0x2744, 0x2744 }, // E0.6   [1] (❄️)       snowflake
    Interval{ 0x2747, 0x2747 }, // E0.6   [1] (❇️)       sparkle
    Interval{ 0x274C, 0x274C }, // E0.6   [1] (❌)       cross mark
    Interval{ 0x274E, 0x274E }, // E0.6   [1] (❎)       cross mark button
    Interval{ 0x2753, 0x2755 }, // E0.6   [3] (❓..❕)    red question mark..white exclamation mark
    Interval{ 0x2757, 0x2757 }, // E0.6   [1] (❗)       red exclamation mark
    Interval{ 0x2763, 0x2763 }, // E1.0   [1] (❣️)       heart exclamation
    Interval{ 0x2764, 0x2764 }, // E0.6   [1] (❤️)       red heart
    Interval{ 0x2795, 0x2797 }, // E0.6   [3] (➕..➗)    plus..divide
    Interval{ 0x27A1, 0x27A1 }, // E0.6   [1] (➡️)       right arrow
    Interval{ 0x27B0, 0x27B0 }, // E0.6   [1] (➰)       curly loop
    Interval{ 0x27BF, 0x27BF }, // E1.0   [1] (➿)       double curly loop
    Interval{ 0x2934, 0x2935 }, // E0.6   [2] (⤴️..⤵️)    right arrow curving up..right arrow curving down
    Interval{ 0x2B05, 0x2B07 }, // E0.6   [3] (⬅️..⬇️)    left arrow..down arrow
    Interval{ 0x2B1B, 0x2B1C }, // E0.6   [2] (⬛..⬜)    black large square..white large square
    Interval{ 0x2B50, 0x2B50 }, // E0.6   [1] (⭐)       star
    Interval{ 0x2B55, 0x2B55 }, // E0.6   [1] (⭕)       hollow red circle
    Interval{ 0x3030, 0x3030 }, // E0.6   [1] (〰️)       wavy dash
    Interval{ 0x303D, 0x303D }, // E0.6   [1] (〽️)       part alternation mark
    Interval{ 0x3297, 0x3297 }, // E0.6   [1] (㊗️)       Japanese “congratulations” button
    Interval{ 0x3299, 0x3299 }, // E0.6   [1] (㊙️)       Japanese “secret” button
    Interval{ 0x1F004, 0x1F004 }, // E0.6   [1] (🀄)       mahjong red dragon
    Interval{ 0x1F0CF, 0x1F0CF }, // E0.6   [1] (🃏)       joker
    Interval{ 0x1F170, 0x1F171 }, // E0.6   [2] (🅰️..🅱️)    A button (blood type)..B button (blood type)
    Interval{ 0x1F17E, 0x1F17F }, // E0.6   [2] (🅾️..🅿️)    O button (blood type)..P button
    Interval{ 0x1F18E, 0x1F18E }, // E0.6   [1] (🆎)       AB button (blood type)
    Interval{ 0x1F191, 0x1F19A }, // E0.6  [10] (🆑..🆚)    CL button..VS button
    Interval{ 0x1F1E6, 0x1F1FF }, // E0.0  [26] (🇦..🇿)    regional indicator symbol letter a..regional indicator symbol letter z
    Interval{ 0x1F201, 0x1F202 }, // E0.6   [2] (🈁..🈂️)    Japanese “here” button..Japanese “service charge” button
    Interval{ 0x1F21A, 0x1F21A }, // E0.6   [1] (🈚)       Japanese “free of charge” button
    Interval{ 0x1F22F, 0x1F22F }, // E0.6   [1] (🈯)       Japanese “reserved” button
    Interval{ 0x1F232, 0x1F23A }, // E0.6   [9] (🈲..🈺)    Japanese “prohibited” button..Japanese “open for business” button
    Interval{ 0x1F250, 0x1F251 }, // E0.6   [2] (🉐..🉑)    Japanese “bargain” button..Japanese “acceptable” button
    Interval{ 0x1F300, 0x1F30C }, // E0.6  [13] (🌀..🌌)    cyclone..milky way
    Interval{ 0x1F30D, 0x1F30E }, // E0.7   [2] (🌍..🌎)    globe showing Europe-Africa..globe showing Americas
    Interval{ 0x1F30F, 0x1F30F }, // E0.6   [1] (🌏)       globe showing Asia-Australia
    Interval{ 0x1F310, 0x1F310 }, // E1.0   [1] (🌐)       globe with meridians
    Interval{ 0x1F311, 0x1F311 }, // E0.6   [1] (🌑)       new moon
    Interval{ 0x1F312, 0x1F312 }, // E1.0   [1] (🌒)       waxing crescent moon
    Interval{ 0x1F313, 0x1F315 }, // E0.6   [3] (🌓..🌕)    first quarter moon..full moon
    Interval{ 0x1F316, 0x1F318 }, // E1.0   [3] (🌖..🌘)    waning gibbous moon..waning crescent moon
    Interval{ 0x1F319, 0x1F319 }, // E0.6   [1] (🌙)       crescent moon
    Interval{ 0x1F31A, 0x1F31A }, // E1.0   [1] (🌚)       new moon face
    Interval{ 0x1F31B, 0x1F31B }, // E0.6   [1] (🌛)       first quarter moon face
    Interval{ 0x1F31C, 0x1F31C }, // E0.7   [1] (🌜)       last quarter moon face
    Interval{ 0x1F31D, 0x1F31E }, // E1.0   [2] (🌝..🌞)    full moon face..sun with face
    Interval{ 0x1F31F, 0x1F320 }, // E0.6   [2] (🌟..🌠)    glowing star..shooting star
    Interval{ 0x1F321, 0x1F321 }, // E0.7   [1] (🌡️)       thermometer
    Interval{ 0x1F324, 0x1F32C }, // E0.7   [9] (🌤️..🌬️)    sun behind small cloud..wind face
    Interval{ 0x1F32D, 0x1F32F }, // E1.0   [3] (🌭..🌯)    hot dog..burrito
    Interval{ 0x1F330, 0x1F331 }, // E0.6   [2] (🌰..🌱)    chestnut..seedling
    Interval{ 0x1F332, 0x1F333 }, // E1.0   [2] (🌲..🌳)    evergreen tree..deciduous tree
    Interval{ 0x1F334, 0x1F335 }, // E0.6   [2] (🌴..🌵)    palm tree..cactus
    Interval{ 0x1F336, 0x1F336 }, // E0.7   [1] (🌶️)       hot pepper
    Interval{ 0x1F337, 0x1F34A }, // E0.6  [20] (🌷..🍊)    tulip..tangerine
    Interval{ 0x1F34B, 0x1F34B }, // E1.0   [1] (🍋)       lemon
    Interval{ 0x1F34C, 0x1F34F }, // E0.6   [4] (🍌..🍏)    banana..green apple
    Interval{ 0x1F350, 0x1F350 }, // E1.0   [1] (🍐)       pear
    Interval{ 0x1F351, 0x1F37B }, // E0.6  [43] (🍑..🍻)    peach..clinking beer mugs
    Interval{ 0x1F37C, 0x1F37C }, // E1.0   [1] (🍼)       baby bottle
    Interval{ 0x1F37D, 0x1F37D }, // E0.7   [1] (🍽️)       fork and knife with plate
    Interval{ 0x1F37E, 0x1F37F }, // E1.0   [2] (🍾..🍿)    bottle with popping cork..popcorn
    Interval{ 0x1F380, 0x1F393 }, // E0.6  [20] (🎀..🎓)    ribbon..graduation cap
    Interval{ 0x1F396, 0x1F397 }, // E0.7   [2] (🎖️..🎗️)    military medal..reminder ribbon
    Interval{ 0x1F399, 0x1F39B }, // E0.7   [3] (🎙️..🎛️)    studio microphone..control knobs
    Interval{ 0x1F39E, 0x1F39F }, // E0.7   [2] (🎞️..🎟️)    film frames..admission tickets
    Interval{ 0x1F3A0, 0x1F3C4 }, // E0.6  [37] (🎠..🏄)    carousel horse..person surfing
    Interval{ 0x1F3C5, 0x1F3C5 }, // E1.0   [1] (🏅)       sports medal
    Interval{ 0x1F3C6, 0x1F3C6 }, // E0.6   [1] (🏆)       trophy
    Interval{ 0x1F3C7, 0x1F3C7 }, // E1.0   [1] (🏇)       horse racing
    Interval{ 0x1F3C8, 0x1F3C8 }, // E0.6   [1] (🏈)       american football
    Interval{ 0x1F3C9, 0x1F3C9 }, // E1.0   [1] (🏉)       rugby football
    Interval{ 0x1F3CA, 0x1F3CA }, // E0.6   [1] (🏊)       person swimming
    Interval{ 0x1F3CB, 0x1F3CE }, // E0.7   [4] (🏋️..🏎️)    person lifting weights..racing car
    Interval{ 0x1F3CF, 0x1F3D3 }, // E1.0   [5] (🏏..🏓)    cricket game..ping pong
    Interval{ 0x1F3D4, 0x1F3DF }, // E0.7  [12] (🏔️..🏟️)    snow-capped mountain..stadium
    Interval{ 0x1F3E0, 0x1F3E3 }, // E0.6   [4] (🏠..🏣)    house..Japanese post office
    Interval{ 0x1F3E4, 0x1F3E4 }, // E1.0   [1] (🏤)       post office
    Interval{ 0x1F3E5, 0x1F3F0 }, // E0.6  [12] (🏥..🏰)    hospital..castle
    Interval{ 0x1F3F3, 0x1F3F3 }, // E0.7   [1] (🏳️)       white flag
    Interval{ 0x1F3F4, 0x1F3F4 }, // E1.0   [1] (🏴)       black flag
    Interval{ 0x1F3F5, 0x1F3F5 }, // E0.7   [1] (🏵️)       rosette
    Interval{ 0x1F3F7, 0x1F3F7 }, // E0.7   [1] (🏷️)       label
    Interval{ 0x1F3F8, 0x1F407 }, // E1.0  [16] (🏸..🐇)    badminton..rabbit
    Interval{ 0x1F408, 0x1F408 }, // E0.7   [1] (🐈)       cat
    Interval{ 0x1F409, 0x1F40B }, // E1.0   [3] (🐉..🐋)    dragon..whale
    Interval{ 0x1F40C, 0x1F40E }, // E0.6   [3] (🐌..🐎)    snail..horse
    Interval{ 0x1F40F, 0x1F410 }, // E1.0   [2] (🐏..🐐)    ram..goat
    Interval{ 0x1F411, 0x1F412 }, // E0.6   [2] (🐑..🐒)    ewe..monkey
    Interval{ 0x1F413, 0x1F413 }, // E1.0   [1] (🐓)       rooster
    Interval{ 0x1F414, 0x1F414 }, // E0.6   [1] (🐔)       chicken
    Interval{ 0x1F415, 0x1F415 }, // E0.7   [1] (🐕)       dog
    Interval{ 0x1F416, 0x1F416 }, // E1.0   [1] (🐖)       pig
    Interval{ 0x1F417, 0x1F429 }, // E0.6  [19] (🐗..🐩)    boar..poodle
    Interval{ 0x1F42A, 0x1F42A }, // E1.0   [1] (🐪)       camel
    Interval{ 0x1F42B, 0x1F43E }, // E0.6  [20] (🐫..🐾)    two-hump camel..paw prints
    Interval{ 0x1F43F, 0x1F43F }, // E0.7   [1] (🐿️)       chipmunk
    Interval{ 0x1F440, 0x1F440 }, // E0.6   [1] (👀)       eyes
    Interval{ 0x1F441, 0x1F441 }, // E0.7   [1] (👁️)       eye
    Interval{ 0x1F442, 0x1F464 }, // E0.6  [35] (👂..👤)    ear..bust in silhouette
    Interval{ 0x1F465, 0x1F465 }, // E1.0   [1] (👥)       busts in silhouette
    Interval{ 0x1F466, 0x1F46B }, // E0.6   [6] (👦..👫)    boy..woman and man holding hands
    Interval{ 0x1F46C, 0x1F46D }, // E1.0   [2] (👬..👭)    men holding hands..women holding hands
    Interval{ 0x1F46E, 0x1F4AC }, // E0.6  [63] (👮..💬)    police officer..speech balloon
    Interval{ 0x1F4AD, 0x1F4AD }, // E1.0   [1] (💭)       thought balloon
    Interval{ 0x1F4AE, 0x1F4B5 }, // E0.6   [8] (💮..💵)    white flower..dollar banknote
    Interval{ 0x1F4B6, 0x1F4B7 }, // E1.0   [2] (💶..💷)    euro banknote..pound banknote
    Interval{ 0x1F4B8, 0x1F4EB }, // E0.6  [52] (💸..📫)    money with wings..closed mailbox with raised flag
    Interval{ 0x1F4EC, 0x1F4ED }, // E0.7   [2] (📬..📭)    open mailbox with raised flag..open mailbox with lowered flag
    Interval{ 0x1F4EE, 0x1F4EE }, // E0.6   [1] (📮)       postbox
    Interval{ 0x1F4EF, 0x1F4EF }, // E1.0   [1] (📯)       postal horn
    Interval{ 0x1F4F0, 0x1F4F4 }, // E0.6   [5] (📰..📴)    newspaper..mobile phone off
    Interval{ 0x1F4F5, 0x1F4F5 }, // E1.0   [1] (📵)       no mobile phones
    Interval{ 0x1F4F6, 0x1F4F7 }, // E0.6   [2] (📶..📷)    antenna bars..camera
    Interval{ 0x1F4F8, 0x1F4F8 }, // E1.0   [1] (📸)       camera with flash
    Interval{ 0x1F4F9, 0x1F4FC }, // E0.6   [4] (📹..📼)    video camera..videocassette
    Interval{ 0x1F4FD, 0x1F4FD }, // E0.7   [1] (📽️)       film projector
    Interval{ 0x1F4FF, 0x1F502 }, // E1.0   [4] (📿..🔂)    prayer beads..repeat single button
    Interval{ 0x1F503, 0x1F503 }, // E0.6   [1] (🔃)       clockwise vertical arrows
    Interval{ 0x1F504, 0x1F507 }, // E1.0   [4] (🔄..🔇)    counterclockwise arrows button..muted speaker
    Interval{ 0x1F508, 0x1F508 }, // E0.7   [1] (🔈)       speaker low volume
    Interval{ 0x1F509, 0x1F509 }, // E1.0   [1] (🔉)       speaker medium volume
    Interval{ 0x1F50A, 0x1F514 }, // E0.6  [11] (🔊..🔔)    speaker high volume..bell
    Interval{ 0x1F515, 0x1F515 }, // E1.0   [1] (🔕)       bell with slash
    Interval{ 0x1F516, 0x1F52B }, // E0.6  [22] (🔖..🔫)    bookmark..water pistol
    Interval{ 0x1F52C, 0x1F52D }, // E1.0   [2] (🔬..🔭)    microscope..telescope
    Interval{ 0x1F52E, 0x1F53D }, // E0.6  [16] (🔮..🔽)    crystal ball..downwards button
    Interval{ 0x1F549, 0x1F54A }, // E0.7   [2] (🕉️..🕊️)    om..dove
    Interval{ 0x1F54B, 0x1F54E }, // E1.0   [4] (🕋..🕎)    kaaba..menorah
    Interval{ 0x1F550, 0x1F55B }, // E0.6  [12] (🕐..🕛)    one o’clock..twelve o’clock
    Interval{ 0x1F55C, 0x1F567 }, // E0.7  [12] (🕜..🕧)    one-thirty..twelve-thirty
    Interval{ 0x1F56F, 0x1F570 }, // E0.7   [2] (🕯️..🕰️)    candle..mantelpiece clock
    Interval{ 0x1F573, 0x1F579 }, // E0.7   [7] (🕳️..🕹️)    hole..joystick
    Interval{ 0x1F57A, 0x1F57A }, // E3.0   [1] (🕺)       man dancing
    Interval{ 0x1F587, 0x1F587 }, // E0.7   [1] (🖇️)       linked paperclips
    Interval{ 0x1F58A, 0x1F58D }, // E0.7   [4] (🖊️..🖍️)    pen..crayon
    Interval{ 0x1F590, 0x1F590 }, // E0.7   [1] (🖐️)       hand with fingers splayed
    Interval{ 0x1F595, 0x1F596 }, // E1.0   [2] (🖕..🖖)    middle finger..vulcan salute
    Interval{ 0x1F5A4, 0x1F5A4 }, // E3.0   [1] (🖤)       black heart
    Interval{ 0x1F5A5, 0x1F5A5 }, // E0.7   [1] (🖥️)       desktop computer
    Interval{ 0x1F5A8, 0x1F5A8 }, // E0.7   [1] (🖨️)       printer
    Interval{ 0x1F5B1, 0x1F5B2 }, // E0.7   [2] (🖱️..🖲️)    computer mouse..trackball
    Interval{ 0x1F5BC, 0x1F5BC }, // E0.7   [1] (🖼️)       framed picture
    Interval{ 0x1F5C2, 0x1F5C4 }, // E0.7   [3] (🗂️..🗄️)    card index dividers..file cabinet
    Interval{ 0x1F5D1, 0x1F5D3 }, // E0.7   [3] (🗑️..🗓️)    wastebasket..spiral calendar
    Interval{ 0x1F5DC, 0x1F5DE }, // E0.7   [3] (🗜️..🗞️)    clamp..rolled-up newspaper
    Interval{ 0x1F5E1, 0x1F5E1 }, // E0.7   [1] (🗡️)       dagger
    Interval{ 0x1F5E3, 0x1F5E3 }, // E0.7   [1] (🗣️)       speaking head
    Interval{ 0x1F5E8, 0x1F5E8 }, // E2.0   [1] (🗨️)       left speech bubble
    Interval{ 0x1F5EF, 0x1F5EF }, // E0.7   [1] (🗯️)       right anger bubble
    Interval{ 0x1F5F3, 0x1F5F3 }, // E0.7   [1] (🗳️)       ballot box with ballot
    Interval{ 0x1F5FA, 0x1F5FA }, // E0.7   [1] (🗺️)       world map
    Interval{ 0x1F5FB, 0x1F5FF }, // E0.6   [5] (🗻..🗿)    mount fuji..moai
    Interval{ 0x1F600, 0x1F600 }, // E1.0   [1] (😀)       grinning face
    Interval{ 0x1F601, 0x1F606 }, // E0.6   [6] (😁..😆)    beaming face with smiling eyes..grinning squinting face
    Interval{ 0x1F607, 0x1F608 }, // E1.0   [2] (😇..😈)    smiling face with halo..smiling face with horns
    Interval{ 0x1F609, 0x1F60D }, // E0.6   [5] (😉..😍)    winking face..smiling face with heart-eyes
    Interval{ 0x1F60E, 0x1F60E }, // E1.0   [1] (😎)       smiling face with sunglasses
    Interval{ 0x1F60F, 0x1F60F }, // E0.6   [1] (😏)       smirking face
    Interval{ 0x1F610, 0x1F610 }, // E0.7   [1] (😐)       neutral face
    Interval{ 0x1F611, 0x1F611 }, // E1.0   [1] (😑)       expressionless face
    Interval{ 0x1F612, 0x1F614 }, // E0.6   [3] (😒..😔)    unamused face..pensive face
    Interval{ 0x1F615, 0x1F615 }, // E1.0   [1] (😕)       confused face
    Interval{ 0x1F616, 0x1F616 }, // E0.6   [1] (😖)       confounded face
    Interval{ 0x1F617, 0x1F617 }, // E1.0   [1] (😗)       kissing face
    Interval{ 0x1F618, 0x1F618 }, // E0.6   [1] (😘)       face blowing a kiss
    Interval{ 0x1F619, 0x1F619 }, // E1.0   [1] (😙)       kissing face with smiling eyes
    Interval{ 0x1F61A, 0x1F61A }, // E0.6   [1] (😚)       kissing face with closed eyes
    Interval{ 0x1F61B, 0x1F61B }, // E1.0   [1] (😛)       face with tongue
    Interval{ 0x1F61C, 0x1F61E }, // E0.6   [3] (😜..😞)    winking face with tongue..disappointed face
    Interval{ 0x1F61F, 0x1F61F }, // E1.0   [1] (😟)       worried face
    Interval{ 0x1F620, 0x1F625 }, // E0.6   [6] (😠..😥)    angry face..sad but relieved face
    Interval{ 0x1F626, 0x1F627 }, // E1.0   [2] (😦..😧)    frowning face with open mouth..anguished face
    Interval{ 0x1F628, 0x1F62B }, // E0.6   [4] (😨..😫)    fearful face..tired face
    Interval{ 0x1F62C, 0x1F62C }, // E1.0   [1] (😬)       grimacing face
    Interval{ 0x1F62D, 0x1F62D }, // E0.6   [1] (😭)       loudly crying face
    Interval{ 0x1F62E, 0x1F62F }, // E1.0   [2] (😮..😯)    face with open mouth..hushed face
    Interval{ 0x1F630, 0x1F633 }, // E0.6   [4] (😰..😳)    anxious face with sweat..flushed face
    Interval{ 0x1F634, 0x1F634 }, // E1.0   [1] (😴)       sleeping face
    Interval{ 0x1F635, 0x1F635 }, // E0.6   [1] (😵)       face with crossed-out eyes
    Interval{ 0x1F636, 0x1F636 }, // E1.0   [1] (😶)       face without mouth
    Interval{ 0x1F637, 0x1F640 }, // E0.6  [10] (😷..🙀)    face with medical mask..weary cat
    Interval{ 0x1F641, 0x1F644 }, // E1.0   [4] (🙁..🙄)    slightly frowning face..face with rolling eyes
    Interval{ 0x1F645, 0x1F64F }, // E0.6  [11] (🙅..🙏)    person gesturing NO..folded hands
    Interval{ 0x1F680, 0x1F680 }, // E0.6   [1] (🚀)       rocket
    Interval{ 0x1F681, 0x1F682 }, // E1.0   [2] (🚁..🚂)    helicopter..locomotive
    Interval{ 0x1F683, 0x1F685 }, // E0.6   [3] (🚃..🚅)    railway car..bullet train
    Interval{ 0x1F686, 0x1F686 }, // E1.0   [1] (🚆)       train
    Interval{ 0x1F687, 0x1F687 }, // E0.6   [1] (🚇)       metro
    Interval{ 0x1F688, 0x1F688 }, // E1.0   [1] (🚈)       light rail
    Interval{ 0x1F689, 0x1F689 }, // E0.6   [1] (🚉)       station
    Interval{ 0x1F68A, 0x1F68B }, // E1.0   [2] (🚊..🚋)    tram..tram car
    Interval{ 0x1F68C, 0x1F68C }, // E0.6   [1] (🚌)       bus
    Interval{ 0x1F68D, 0x1F68D }, // E0.7   [1] (🚍)       oncoming bus
    Interval{ 0x1F68E, 0x1F68E }, // E1.0   [1] (🚎)       trolleybus
    Interval{ 0x1F68F, 0x1F68F }, // E0.6   [1] (🚏)       bus stop
    Interval{ 0x1F690, 0x1F690 }, // E1.0   [1] (🚐)       minibus
    Interval{ 0x1F691, 0x1F693 }, // E0.6   [3] (🚑..🚓)    ambulance..police car
    Interval{ 0x1F694, 0x1F694 }, // E0.7   [1] (🚔)       oncoming police car
    Interval{ 0x1F695, 0x1F695 }, // E0.6   [1] (🚕)       taxi
    Interval{ 0x1F696, 0x1F696 }, // E1.0   [1] (🚖)       oncoming taxi
    Interval{ 0x1F697, 0x1F697 }, // E0.6   [1] (🚗)       automobile
    Interval{ 0x1F698, 0x1F698 }, // E0.7   [1] (🚘)       oncoming automobile
    Interval{ 0x1F699, 0x1F69A }, // E0.6   [2] (🚙..🚚)    sport utility vehicle..delivery truck
    Interval{ 0x1F69B, 0x1F6A1 }, // E1.0   [7] (🚛..🚡)    articulated lorry..aerial tramway
    Interval{ 0x1F6A2, 0x1F6A2 }, // E0.6   [1] (🚢)       ship
    Interval{ 0x1F6A3, 0x1F6A3 }, // E1.0   [1] (🚣)       person rowing boat
    Interval{ 0x1F6A4, 0x1F6A5 }, // E0.6   [2] (🚤..🚥)    speedboat..horizontal traffic light
    Interval{ 0x1F6A6, 0x1F6A6 }, // E1.0   [1] (🚦)       vertical traffic light
    Interval{ 0x1F6A7, 0x1F6AD }, // E0.6   [7] (🚧..🚭)    construction..no smoking
    Interval{ 0x1F6AE, 0x1F6B1 }, // E1.0   [4] (🚮..🚱)    litter in bin sign..non-potable water
    Interval{ 0x1F6B2, 0x1F6B2 }, // E0.6   [1] (🚲)       bicycle
    Interval{ 0x1F6B3, 0x1F6B5 }, // E1.0   [3] (🚳..🚵)    no bicycles..person mountain biking
    Interval{ 0x1F6B6, 0x1F6B6 }, // E0.6   [1] (🚶)       person walking
    Interval{ 0x1F6B7, 0x1F6B8 }, // E1.0   [2] (🚷..🚸)    no pedestrians..children crossing
    Interval{ 0x1F6B9, 0x1F6BE }, // E0.6   [6] (🚹..🚾)    men’s room..water closet
    Interval{ 0x1F6BF, 0x1F6BF }, // E1.0   [1] (🚿)       shower
    Interval{ 0x1F6C0, 0x1F6C0 }, // E0.6   [1] (🛀)       person taking bath
    Interval{ 0x1F6C1, 0x1F6C5 }, // E1.0   [5] (🛁..🛅)    bathtub..left luggage
    Interval{ 0x1F6CB, 0x1F6CB }, // E0.7   [1] (🛋️)       couch and lamp
    Interval{ 0x1F6CC, 0x1F6CC }, // E1.0   [1] (🛌)       person in bed
    Interval{ 0x1F6CD, 0x1F6CF }, // E0.7   [3] (🛍️..🛏️)    shopping bags..bed
    Interval{ 0x1F6D0, 0x1F6D0 }, // E1.0   [1] (🛐)       place of worship
    Interval{ 0x1F6D1, 0x1F6D2 }, // E3.0   [2] (🛑..🛒)    stop sign..shopping cart
    Interval{ 0x1F6D5, 0x1F6D5 }, // E12.0  [1] (🛕)       hindu temple
    Interval{ 0x1F6D6, 0x1F6D7 }, // E13.0  [2] (🛖..🛗)    hut..elevator
    Interval{ 0x1F6D8, 0x1F6D8 }, // E17.0  [1] (🛘)       landslide
    Interval{ 0x1F6DC, 0x1F6DC }, // E15.0  [1] (🛜)       wireless
    Interval{ 0x1F6DD, 0x1F6DF }, // E14.0  [3] (🛝..🛟)    playground slide..ring buoy
    Interval{ 0x1F6E0, 0x1F6E5 }, // E0.7   [6] (🛠️..🛥️)    hammer and wrench..motor boat
    Interval{ 0x1F6E9, 0x1F6E9 }, // E0.7   [1] (🛩️)       small airplane
    Interval{ 0x1F6EB, 0x1F6EC }, // E1.0   [2] (🛫..🛬)    airplane departure..airplane arrival
    Interval{ 0x1F6F0, 0x1F6F0 }, // E0.7   [1] (🛰️)       satellite
    Interval{ 0x1F6F3, 0x1F6F3 }, // E0.7   [1] (🛳️)       passenger ship
    Interval{ 0x1F6F4, 0x1F6F6 }, // E3.0   [3] (🛴..🛶)    kick scooter..canoe
    Interval{ 0x1F6F7, 0x1F6F8 }, // E5.0   [2] (🛷..🛸)    sled..flying saucer
    Interval{ 0x1F6F9, 0x1F6F9 }, // E11.0  [1] (🛹)       skateboard
    Interval{ 0x1F6FA, 0x1F6FA }, // E12.0  [1] (🛺)       auto rickshaw
    Interval{ 0x1F6FB, 0x1F6FC }, // E13.0  [2] (🛻..🛼)    pickup truck..roller skate
    Interval{ 0x1F7E0, 0x1F7EB }, // E12.0 [12] (🟠..🟫)    orange circle..brown square
    Interval{ 0x1F7F0, 0x1F7F0 }, // E14.0  [1] (🟰)       heavy equals sign
    Interval{ 0x1F90C, 0x1F90C }, // E13.0  [1] (🤌)       pinched fingers
    Interval{ 0x1F90D, 0x1F90F }, // E12.0  [3] (🤍..🤏)    white heart..pinching hand
    Interval{ 0x1F910, 0x1F918 }, // E1.0   [9] (🤐..🤘)    zipper-mouth face..sign of the horns
    Interval{ 0x1F919, 0x1F91E }, // E3.0   [6] (🤙..🤞)    call me hand..crossed fingers
    Interval{ 0x1F91F, 0x1F91F }, // E5.0   [1] (🤟)       love-you gesture
    Interval{ 0x1F920, 0x1F927 }, // E3.0   [8] (🤠..🤧)    cowboy hat face..sneezing face
    Interval{ 0x1F928, 0x1F92F }, // E5.0   [8] (🤨..🤯)    face with raised eyebrow..exploding head
    Interval{ 0x1F930, 0x1F930 }, // E3.0   [1] (🤰)       pregnant woman
    Interval{ 0x1F931, 0x1F932 }, // E5.0   [2] (🤱..🤲)    breast-feeding..palms up together
    Interval{ 0x1F933, 0x1F93A }, // E3.0   [8] (🤳..🤺)    selfie..person fencing
    Interval{ 0x1F93C, 0x1F93E }, // E3.0   [3] (🤼..🤾)    people wrestling..person playing handball
    Interval{ 0x1F93F, 0x1F93F }, // E12.0  [1] (🤿)       diving mask
    Interval{ 0x1F940, 0x1F945 }, // E3.0   [6] (🥀..🥅)    wilted flower..goal net
    Interval{ 0x1F947, 0x1F94B }, // E3.0   [5] (🥇..🥋)    1st place medal..martial arts uniform
    Interval{ 0x1F94C, 0x1F94C }, // E5.0   [1] (🥌)       curling stone
    Interval{ 0x1F94D, 0x1F94F }, // E11.0  [3] (🥍..🥏)    lacrosse..flying disc
    Interval{ 0x1F950, 0x1F95E }, // E3.0  [15] (🥐..🥞)    croissant..pancakes
    Interval{ 0x1F95F, 0x1F96B }, // E5.0  [13] (🥟..🥫)    dumpling..canned food
    Interval{ 0x1F96C, 0x1F970 }, // E11.0  [5] (🥬..🥰)    leafy green..smiling face with hearts
    Interval{ 0x1F971, 0x1F971 }, // E12.0  [1] (🥱)       yawning face
    Interval{ 0x1F972, 0x1F972 }, // E13.0  [1] (🥲)       smiling face with tear
    Interval{ 0x1F973, 0x1F976 }, // E11.0  [4] (🥳..🥶)    partying face..cold face
    Interval{ 0x1F977, 0x1F978 }, // E13.0  [2] (🥷..🥸)    ninja..disguised face
    Interval{ 0x1F979, 0x1F979 }, // E14.0  [1] (🥹)       face holding back tears
    Interval{ 0x1F97A, 0x1F97A }, // E11.0  [1] (🥺)       pleading face
    Interval{ 0x1F97B, 0x1F97B }, // E12.0  [1] (🥻)       sari
    Interval{ 0x1F97C, 0x1F97F }, // E11.0  [4] (🥼..🥿)    lab coat..flat shoe
    Interval{ 0x1F980, 0x1F984 }, // E1.0   [5] (🦀..🦄)    crab..unicorn
    Interval{ 0x1F985, 0x1F991 }, // E3.0  [13] (🦅..🦑)    eagle..squid
    Interval{ 0x1F992, 0x1F997 }, // E5.0   [6] (🦒..🦗)    giraffe..cricket
    Interval{ 0x1F998, 0x1F9A2 }, // E11.0 [11] (🦘..🦢)    kangaroo..swan
    Interval{ 0x1F9A3, 0x1F9A4 }, // E13.0  [2] (🦣..🦤)    mammoth..dodo
    Interval{ 0x1F9A5, 0x1F9AA }, // E12.0  [6] (🦥..🦪)    sloth..oyster
    Interval{ 0x1F9AB, 0x1F9AD }, // E13.0  [3] (🦫..🦭)    beaver..seal
    Interval{ 0x1F9AE, 0x1F9AF }, // E12.0  [2] (🦮..🦯)    guide dog..white cane
    Interval{ 0x1F9B0, 0x1F9B9 }, // E11.0 [10] (🦰..🦹)    red hair..supervillain
    Interval{ 0x1F9BA, 0x1F9BF }, // E12.0  [6] (🦺..🦿)    safety vest..mechanical leg
    Interval{ 0x1F9C0, 0x1F9C0 }, // E1.0   [1] (🧀)       cheese wedge
    Interval{ 0x1F9C1, 0x1F9C2 }, // E11.0  [2] (🧁..🧂)    cupcake..salt
    Interval{ 0x1F9C3, 0x1F9CA }, // E12.0  [8] (🧃..🧊)    beverage box..ice
    Interval{ 0x1F9CB, 0x1F9CB }, // E13.0  [1] (🧋)       bubble tea
    Interval{ 0x1F9CC, 0x1F9CC }, // E14.0  [1] (🧌)       troll
    Interval{ 0x1F9CD, 0x1F9CF }, // E12.0  [3] (🧍..🧏)    person standing..deaf person
    Interval{ 0x1F9D0, 0x1F9E6 }, // E5.0  [23] (🧐..🧦)    face with monocle..socks
    Interval{ 0x1F9E7, 0x1F9FF }, // E11.0 [25] (🧧..🧿)    red envelope..nazar amulet
    Interval{ 0x1FA70, 0x1FA73 }, // E12.0  [4] (🩰..🩳)    ballet shoes..shorts
    Interval{ 0x1FA74, 0x1FA74 }, // E13.0  [1] (🩴)       thong sandal
    Interval{ 0x1FA75, 0x1FA77 }, // E15.0  [3] (🩵..🩷)    light blue heart..pink heart
    Interval{ 0x1FA78, 0x1FA7A }, // E12.0  [3] (🩸..🩺)    drop of blood..stethoscope
    Interval{ 0x1FA7B, 0x1FA7C }, // E14.0  [2] (🩻..🩼)    x-ray..crutch
    Interval{ 0x1FA80, 0x1FA82 }, // E12.0  [3] (🪀..🪂)    yo-yo..parachute
    Interval{ 0x1FA83, 0x1FA86 }, // E13.0  [4] (🪃..🪆)    boomerang..nesting dolls
    Interval{ 0x1FA87, 0x1FA88 }, // E15.0  [2] (🪇..🪈)    maracas..flute
    Interval{ 0x1FA89, 0x1FA89 }, // E16.0  [1] (🪉)       harp
    Interval{ 0x1FA8A, 0x1FA8A }, // E17.0  [1] (🪊)       trombone
    Interval{ 0x1FA8E, 0x1FA8E }, // E17.0  [1] (🪎)       treasure chest
    Interval{ 0x1FA8F, 0x1FA8F }, // E16.0  [1] (🪏)       shovel
    Interval{ 0x1FA90, 0x1FA95 }, // E12.0  [6] (🪐..🪕)    ringed planet..banjo
    Interval{ 0x1FA96, 0x1FAA8 }, // E13.0 [19] (🪖..🪨)    military helmet..rock
    Interval{ 0x1FAA9, 0x1FAAC }, // E14.0  [4] (🪩..🪬)    mirror ball..hamsa
    Interval{ 0x1FAAD, 0x1FAAF }, // E15.0  [3] (🪭..🪯)    folding hand fan..khanda
    Interval{ 0x1FAB0, 0x1FAB6 }, // E13.0  [7] (🪰..🪶)    fly..feather
    Interval{ 0x1FAB7, 0x1FABA }, // E14.0  [4] (🪷..🪺)    lotus..nest with eggs
    Interval{ 0x1FABB, 0x1FABD }, // E15.0  [3] (🪻..🪽)    hyacinth..wing
    Interval{ 0x1FABE, 0x1FABE }, // E16.0  [1] (🪾)       leafless tree
    Interval{ 0x1FABF, 0x1FABF }, // E15.0  [1] (🪿)       goose
    Interval{ 0x1FAC0, 0x1FAC2 }, // E13.0  [3] (🫀..🫂)    anatomical heart..people hugging
    Interval{ 0x1FAC3, 0x1FAC5 }, // E14.0  [3] (🫃..🫅)    pregnant man..person with crown
    Interval{ 0x1FAC6, 0x1FAC6 }, // E16.0  [1] (🫆)       fingerprint
    Interval{ 0x1FAC8, 0x1FAC8 }, // E17.0  [1] (🫈)       hairy creature
    Interval{ 0x1FACD, 0x1FACD }, // E17.0  [1] (🫍)       orca
    Interval{ 0x1FACE, 0x1FACF }, // E15.0  [2] (🫎..🫏)    moose..donkey
    Interval{ 0x1FAD0, 0x1FAD6 }, // E13.0  [7] (🫐..🫖)    blueberries..teapot
    Interval{ 0x1FAD7, 0x1FAD9 }, // E14.0  [3] (🫗..🫙)    pouring liquid..jar
    Interval{ 0x1FADA, 0x1FADB }, // E15.0  [2] (🫚..🫛)    ginger root..pea pod
    Interval{ 0x1FADC, 0x1FADC }, // E16.0  [1] (🫜)       root vegetable
    Interval{ 0x1FADF, 0x1FADF }, // E16.0  [1] (🫟)       splatter
    Interval{ 0x1FAE0, 0x1FAE7 }, // E14.0  [8] (🫠..🫧)    melting face..bubbles
    Interval{ 0x1FAE8, 0x1FAE8 }, // E15.0  [1] (🫨)       shaking face
    Interval{ 0x1FAE9, 0x1FAE9 }, // E16.0  [1] (🫩)       face with bags under eyes
    Interval{ 0x1FAEA, 0x1FAEA }, // E17.0  [1] (🫪)       distorted face
    Interval{ 0x1FAEF, 0x1FAEF }, // E17.0  [1] (🫯)       fight cloud
    Interval{ 0x1FAF0, 0x1FAF6 }, // E14.0  [7] (🫰..🫶)    hand with index finger and thumb crossed..heart hands
    Interval{ 0x1FAF7, 0x1FAF8 }, // E15.0  [2] (🫷..🫸)    leftwards pushing hand..rightwards pushing hand
}; // }}}
auto static const Emoji_Component = std::array<Interval, 10>{ // {{{
    Interval{ 0x0023, 0x0023 }, // E0.0   [1] (#️)       hash sign
    Interval{ 0x002A, 0x002A }, // E0.0   [1] (*️)       asterisk
    Interval{ 0x0030, 0x0039 }, // E0.0  [10] (0️..9️)    digit zero..digit nine
    Interval{ 0x200D, 0x200D }, // E0.0   [1] (‍)        zero width joiner
    Interval{ 0x20E3, 0x20E3 }, // E0.0   [1] (⃣)       combining enclosing keycap
    Interval{ 0xFE0F, 0xFE0F }, // E0.0   [1] ()        VARIATION SELECTOR-16
    Interval{ 0x1F1E6, 0x1F1FF }, // E0.0  [26] (🇦..🇿)    regional indicator symbol letter a..regional indicator symbol letter z
    Interval{ 0x1F3FB, 0x1F3FF }, // E1.0   [5] (🏻..🏿)    light skin tone..dark skin tone
    Interval{ 0x1F9B0, 0x1F9B3 }, // E11.0  [4] (🦰..🦳)    red hair..white hair
    Interval{ 0xE0020, 0xE007F }, // E0.0  [96] (󠀠..󠁿)      tag space..cancel tag
}; // }}}
auto static const Emoji_Modifier = std::array<Interval, 1>{ // {{{
    Interval{ 0x1F3FB, 0x1F3FF }, // E1.0   [5] (🏻..🏿)    light skin tone..dark skin tone
}; // }}}
auto static const Emoji_Modifier_Base = std::array<Interval, 50>{ // {{{
    Interval{ 0x261D, 0x261D }, // E0.6   [1] (☝️)       index pointing up
    Interval{ 0x26F9, 0x26F9 }, // E0.7   [1] (⛹️)       person bouncing ball
    Interval{ 0x270A, 0x270C }, // E0.6   [3] (✊..✌️)    raised fist..victory hand
    Interval{ 0x270D, 0x270D }, // E0.7   [1] (✍️)       writing hand
    Interval{ 0x1F385, 0x1F385 }, // E0.6   [1] (🎅)       Santa Claus
    Interval{ 0x1F3C2, 0x1F3C4 }, // E0.6   [3] (🏂..🏄)    snowboarder..person surfing
    Interval{ 0x1F3C7, 0x1F3C7 }, // E1.0   [1] (🏇)       horse racing
    Interval{ 0x1F3CA, 0x1F3CA }, // E0.6   [1] (🏊)       person swimming
    Interval{ 0x1F3CB, 0x1F3CC }, // E0.7   [2] (🏋️..🏌️)    person lifting weights..person golfing
    Interval{ 0x1F442, 0x1F443 }, // E0.6   [2] (👂..👃)    ear..nose
    Interval{ 0x1F446, 0x1F450 }, // E0.6  [11] (👆..👐)    backhand index pointing up..open hands
    Interval{ 0x1F466, 0x1F46B }, // E0.6   [6] (👦..👫)    boy..woman and man holding hands
    Interval{ 0x1F46C, 0x1F46D }, // E1.0   [2] (👬..👭)    men holding hands..women holding hands
    Interval{ 0x1F46E, 0x1F478 }, // E0.6  [11] (👮..👸)    police officer..princess
    Interval{ 0x1F47C, 0x1F47C }, // E0.6   [1] (👼)       baby angel
    Interval{ 0x1F481, 0x1F483 }, // E0.6   [3] (💁..💃)    person tipping hand..woman dancing
    Interval{ 0x1F485, 0x1F487 }, // E0.6   [3] (💅..💇)    nail polish..person getting haircut
    Interval{ 0x1F48F, 0x1F48F }, // E0.6   [1] (💏)       kiss
    Interval{ 0x1F491, 0x1F491 }, // E0.6   [1] (💑)       couple with heart
    Interval{ 0x1F4AA, 0x1F4AA }, // E0.6   [1] (💪)       flexed biceps
    Interval{ 0x1F574, 0x1F575 }, // E0.7   [2] (🕴️..🕵️)    person in suit levitating..detective
    Interval{ 0x1F57A, 0x1F57A }, // E3.0   [1] (🕺)       man dancing
    Interval{ 0x1F590, 0x1F590 }, // E0.7   [1] (🖐️)       hand with fingers splayed
    Interval{ 0x1F595, 0x1F596 }, // E1.0   [2] (🖕..🖖)    middle finger..vulcan salute
    Interval{ 0x1F645, 0x1F647 }, // E0.6   [3] (🙅..🙇)    person gesturing NO..person bowing
    Interval{ 0x1F64B, 0x1F64F }, // E0.6   [5] (🙋..🙏)    person raising hand..folded hands
    Interval{ 0x1F6A3, 0x1F6A3 }, // E1.0   [1] (🚣)       person rowing boat
    Interval{ 0x1F6B4, 0x1F6B5 }, // E1.0   [2] (🚴..🚵)    person biking..person mountain biking
    Interval{ 0x1F6B6, 0x1F6B6 }, // E0.6   [1] (🚶)       person walking
    Interval{ 0x1F6C0, 0x1F6C0 }, // E0.6   [1] (🛀)       person taking bath
    Interval{ 0x1F6CC, 0x1F6CC }, // E1.0   [1] (🛌)       person in bed
    Interval{ 0x1F90C, 0x1F90C }, // E13.0  [1] (🤌)       pinched fingers
    Interval{ 0x1F90F, 0x1F90F }, // E12.0  [1] (🤏)       pinching hand
    Interval{ 0x1F918, 0x1F918 }, // E1.0   [1] (🤘)       sign of the horns
    Interval{ 0x1F919, 0x1F91E }, // E3.0   [6] (🤙..🤞)    call me hand..crossed fingers
    Interval{ 0x1F91F, 0x1F91F }, // E5.0   [1] (🤟)       love-you gesture
    Interval{ 0x1F926, 0x1F926 }, // E3.0   [1] (🤦)       person facepalming
    Interval{ 0x1F930, 0x1F930 }, // E3.0   [1] (🤰)       pregnant woman
    Interval{ 0x1F931, 0x1F932 }, // E5.0   [2] (🤱..🤲)    breast-feeding..palms up together
    Interval{ 0x1F933, 0x1F939 }, // E3.0   [7] (🤳..🤹)    selfie..person juggling
    Interval{ 0x1F93C, 0x1F93E }, // E3.0   [3] (🤼..🤾)    people wrestling..person playing handball
    Interval{ 0x1F977, 0x1F977 }, // E13.0  [1] (🥷)       ninja
    Interval{ 0x1F9B5, 0x1F9B6 }, // E11.0  [2] (🦵..🦶)    leg..foot
    Interval{ 0x1F9B8, 0x1F9B9 }, // E11.0  [2] (🦸..🦹)    superhero..supervillain
    Interval{ 0x1F9BB, 0x1F9BB }, // E12.0  [1] (🦻)       ear with hearing aid
    Interval{ 0x1F9CD, 0x1F9CF }, // E12.0  [3] (🧍..🧏)    person standing..deaf person
    Interval{ 0x1F9D1, 0x1F9DD }, // E5.0  [13] (🧑..🧝)    person..elf
    Interval{ 0x1FAC3, 0x1FAC5 }, // E14.0  [3] (🫃..🫅)    pregnant man..person with crown
    Interval{ 0x1FAF0, 0x1FAF6 }, // E14.0  [7] (🫰..🫶)    hand with index finger and thumb crossed..heart hands
    Interval{ 0x1FAF7, 0x1FAF8 }, // E15.0  [2] (🫷..🫸)    leftwards pushing hand..rightwards pushing hand
}; // }}}
auto static const Emoji_Presentation = std::array<Interval, 296>{ // {{{
    Interval{ 0x231A, 0x231B }, // E0.6   [2] (⌚..⌛)    watch..hourglass done
    Interval{ 0x23E9, 0x23EC }, // E0.6   [4] (⏩..⏬)    fast-forward button..fast down button
    Interval{ 0x23F0, 0x23F0 }, // E0.6   [1] (⏰)       alarm clock
    Interval{ 0x23F3, 0x23F3 }, // E0.6   [1] (⏳)       hourglass not done
    Interval{ 0x25FD, 0x25FE }, // E0.6   [2] (◽..◾)    white medium-small square..black medium-small square
    Interval{ 0x2614, 0x2615 }, // E0.6   [2] (☔..☕)    umbrella with rain drops..hot beverage
    Interval{ 0x2648, 0x2653 }, // E0.6  [12] (♈..♓)    Aries..Pisces
    Interval{ 0x267F, 0x267F }, // E0.6   [1] (♿)       wheelchair symbol
    Interval{ 0x2693, 0x2693 }, // E0.6   [1] (⚓)       anchor
    Interval{ 0x26A1, 0x26A1 }, // E0.6   [1] (⚡)       high voltage
    Interval{ 0x26AA, 0x26AB }, // E0.6   [2] (⚪..⚫)    white circle..black circle
    Interval{ 0x26BD, 0x26BE }, // E0.6   [2] (⚽..⚾)    soccer ball..baseball
    Interval{ 0x26C4, 0x26C5 }, // E0.6   [2] (⛄..⛅)    snowman without snow..sun behind cloud
    Interval{ 0x26CE, 0x26CE }, // E0.6   [1] (⛎)       Ophiuchus
    Interval{ 0x26D4, 0x26D4 }, // E0.6   [1] (⛔)       no entry
    Interval{ 0x26EA, 0x26EA }, // E0.6   [1] (⛪)       church
    Interval{ 0x26F2, 0x26F3 }, // E0.6   [2] (⛲..⛳)    fountain..flag in hole
    Interval{ 0x26F5, 0x26F5 }, // E0.6   [1] (⛵)       sailboat
    Interval{ 0x26FA, 0x26FA }, // E0.6   [1] (⛺)       tent
    Interval{ 0x26FD, 0x26FD }, // E0.6   [1] (⛽)       fuel pump
    Interval{ 0x2705, 0x2705 }, // E0.6   [1] (✅)       check mark button
    Interval{ 0x270A, 0x270B }, // E0.6   [2] (✊..✋)    raised fist..raised hand
    Interval{ 0x2728, 0x2728 }, // E0.6   [1] (✨)       sparkles
    Interval{ 0x274C, 0x274C }, // E0.6   [1] (❌)       cross mark
    Interval{ 0x274E, 0x274E }, // E0.6   [1] (❎)       cross mark button
    Interval{ 0x2753, 0x2755 }, // E0.6   [3] (❓..❕)    red question mark..white exclamation mark
    Interval{ 0x2757, 0x2757 }, // E0.6   [1] (❗)       red exclamation mark
    Interval{ 0x2795, 0x2797 }, // E0.6   [3] (➕..➗)    plus..divide
    Interval{ 0x27B0, 0x27B0 }, // E0.6   [1] (➰)       curly loop
    Interval{ 0x27BF, 0x27BF }, // E1.0   [1] (➿)       double curly loop
    Interval{ 0x2B1B, 0x2B1C }, // E0.6   [2] (⬛..⬜)    black large square..white large square
    Interval{ 0x2B50, 0x2B50 }, // E0.6   [1] (⭐)       star
    Interval{ 0x2B55, 0x2B55 }, // E0.6   [1] (⭕)       hollow red circle
    Interval{ 0x1F004, 0x1F004 }, // E0.6   [1] (🀄)       mahjong red dragon
    Interval{ 0x1F0CF, 0x1F0CF }, // E0.6   [1] (🃏)       joker
    Interval{ 0x1F18E, 0x1F18E }, // E0.6   [1] (🆎)       AB button (blood type)
    Interval{ 0x1F191, 0x1F19A }, // E0.6  [10] (🆑..🆚)    CL button..VS button
    Interval{ 0x1F1E6, 0x1F1FF }, // E0.0  [26] (🇦..🇿)    regional indicator symbol letter a..regional indicator symbol letter z
    Interval{ 0x1F201, 0x1F201 }, // E0.6   [1] (🈁)       Japanese “here” button
    Interval{ 0x1F21A, 0x1F21A }, // E0.6   [1] (🈚)       Japanese “free of charge” button
    Interval{ 0x1F22F, 0x1F22F }, // E0.6   [1] (🈯)       Japanese “reserved” button
    Interval{ 0x1F232, 0x1F236 }, // E0.6   [5] (🈲..🈶)    Japanese “prohibited” button..Japanese “not free of charge” button
    Interval{ 0x1F238, 0x1F23A }, // E0.6   [3] (🈸..🈺)    Japanese “application” button..Japanese “open for business” button
    Interval{ 0x1F250, 0x1F251 }, // E0.6   [2] (🉐..🉑)    Japanese “bargain” button..Japanese “acceptable” button
    Interval{ 0x1F300, 0x1F30C }, // E0.6  [13] (🌀..🌌)    cyclone..milky way
    Interval{ 0x1F30D, 0x1F30E }, // E0.7   [2] (🌍..🌎)    globe showing Europe-Africa..globe showing Americas
    Interval{ 0x1F30F, 0x1F30F }, // E0.6   [1] (🌏)       globe showing Asia-Australia
    Interval{ 0x1F310, 0x1F310 }, // E1.0   [1] (🌐)       globe with meridians
    Interval{ 0x1F311, 0x1F311 }, // E0.6   [1] (🌑)       new moon
    Interval{ 0x1F312, 0x1F312 }, // E1.0   [1] (🌒)       waxing crescent moon
    Interval{ 0x1F313, 0x1F315 }, // E0.6   [3] (🌓..🌕)    first quarter moon..full moon
    Interval{ 0x1F316, 0x1F318 }, // E1.0   [3] (🌖..🌘)    waning gibbous moon..waning crescent moon
    Interval{ 0x1F319, 0x1F319 }, // E0.6   [1] (🌙)       crescent moon
    Interval{ 0x1F31A, 0x1F31A }, // E1.0   [1] (🌚)       new moon face
    Interval{ 0x1F31B, 0x1F31B }, // E0.6   [1] (🌛)       first quarter moon face
    Interval{ 0x1F31C, 0x1F31C }, // E0.7   [1] (🌜)       last quarter moon face
    Interval{ 0x1F31D, 0x1F31E }, // E1.0   [2] (🌝..🌞)    full moon face..sun with face
    Interval{ 0x1F31F, 0x1F320 }, // E0.6   [2] (🌟..🌠)    glowing star..shooting star
    Interval{ 0x1F32D, 0x1F32F }, // E1.0   [3] (🌭..🌯)    hot dog..burrito
    Interval{ 0x1F330, 0x1F331 }, // E0.6   [2] (🌰..🌱)    chestnut..seedling
    Interval{ 0x1F332, 0x1F333 }, // E1.0   [2] (🌲..🌳)    evergreen tree..deciduous tree
    Interval{ 0x1F334, 0x1F335 }, // E0.6   [2] (🌴..🌵)    palm tree..cactus
    Interval{ 0x1F337, 0x1F34A }, // E0.6  [20] (🌷..🍊)    tulip..tangerine
    Interval{ 0x1F34B, 0x1F34B }, // E1.0   [1] (🍋)       lemon
    Interval{ 0x1F34C, 0x1F34F }, // E0.6   [4] (🍌..🍏)    banana..green apple
    Interval{ 0x1F350, 0x1F350 }, // E1.0   [1] (🍐)       pear
    Interval{ 0x1F351, 0x1F37B }, // E0.6  [43] (🍑..🍻)    peach..clinking beer mugs
    Interval{ 0x1F37C, 0x1F37C }, // E1.0   [1] (🍼)       baby bottle
    Interval{ 0x1F37E, 0x1F37F }, // E1.0   [2] (🍾..🍿)    bottle with popping cork..popcorn
    Interval{ 0x1F380, 0x1F393 }, // E0.6  [20] (🎀..🎓)    ribbon..graduation cap
    Interval{ 0x1F3A0, 0x1F3C4 }, // E0.6  [37] (🎠..🏄)    carousel horse..person surfing
    Interval{ 0x1F3C5, 0x1F3C5 }, // E1.0   [1] (🏅)       sports medal
    Interval{ 0x1F3C6, 0x1F3C6 }, // E0.6   [1] (🏆)       trophy
    Interval{ 0x1F3C7, 0x1F3C7 }, // E1.0   [1] (🏇)       horse racing
    Interval{ 0x1F3C8, 0x1F3C8 }, // E0.6   [1] (🏈)       american football
    Interval{ 0x1F3C9, 0x1F3C9 }, // E1.0   [1] (🏉)       rugby football
    Interval{ 0x1F3CA, 0x1F3CA }, // E0.6   [1] (🏊)       person swimming
    Interval{ 0x1F3CF, 0x1F3D3 }, // E1.0   [5] (🏏..🏓)    cricket game..ping pong
    Interval{ 0x1F3E0, 0x1F3E3 }, // E0.6   [4] (🏠..🏣)    house..Japanese post office
    Interval{ 0x1F3E4, 0x1F3E4 }, // E1.0   [1] (🏤)       post office
    Interval{ 0x1F3E5, 0x1F3F0 }, // E0.6  [12] (🏥..🏰)    hospital..castle
    Interval{ 0x1F3F4, 0x1F3F4 }, // E1.0   [1] (🏴)       black flag
    Interval{ 0x1F3F8, 0x1F407 }, // E1.0  [16] (🏸..🐇)    badminton..rabbit
    Interval{ 0x1F408, 0x1F408 }, // E0.7   [1] (🐈)       cat
    Interval{ 0x1F409, 0x1F40B }, // E1.0   [3] (🐉..🐋)    dragon..whale
    Interval{ 0x1F40C, 0x1F40E }, // E0.6   [3] (🐌..🐎)    snail..horse
    Interval{ 0x1F40F, 0x1F410 }, // E1.0   [2] (🐏..🐐)    ram..goat
    Interval{ 0x1F411, 0x1F412 }, // E0.6   [2] (🐑..🐒)    ewe..monkey
    Interval{ 0x1F413, 0x1F413 }, // E1.0   [1] (🐓)       rooster
    Interval{ 0x1F414, 0x1F414 }, // E0.6   [1] (🐔)       chicken
    Interval{ 0x1F415, 0x1F415 }, // E0.7   [1] (🐕)       dog
    Interval{ 0x1F416, 0x1F416 }, // E1.0   [1] (🐖)       pig
    Interval{ 0x1F417, 0x1F429 }, // E0.6  [19] (🐗..🐩)    boar..poodle
    Interval{ 0x1F42A, 0x1F42A }, // E1.0   [1] (🐪)       camel
    Interval{ 0x1F42B, 0x1F43E }, // E0.6  [20] (🐫..🐾)    two-hump camel..paw prints
    Interval{ 0x1F440, 0x1F440 }, // E0.6   [1] (👀)       eyes
    Interval{ 0x1F442, 0x1F464 }, // E0.6  [35] (👂..👤)    ear..bust in silhouette
    Interval{ 0x1F465, 0x1F465 }, // E1.0   [1] (👥)       busts in silhouette
    Interval{ 0x1F466, 0x1F46B }, // E0.6   [6] (👦..👫)    boy..woman and man holding hands
    Interval{ 0x1F46C, 0x1F46D }, // E1.0   [2] (👬..👭)    men holding hands..women holding hands
    Interval{ 0x1F46E, 0x1F4AC }, // E0.6  [63] (👮..💬)    police officer..speech balloon
    Interval{ 0x1F4AD, 0x1F4AD }, // E1.0   [1] (💭)       thought balloon
    Interval{ 0x1F4AE, 0x1F4B5 }, // E0.6   [8] (💮..💵)    white flower..dollar banknote
    Interval{ 0x1F4B6, 0x1F4B7 }, // E1.0   [2] (💶..💷)    euro banknote..pound banknote
    Interval{ 0x1F4B8, 0x1F4EB }, // E0.6  [52] (💸..📫)    money with wings..closed mailbox with raised flag
    Interval{ 0x1F4EC, 0x1F4ED }, // E0.7   [2] (📬..📭)    open mailbox with raised flag..open mailbox with lowered flag
    Interval{ 0x1F4EE, 0x1F4EE }, // E0.6   [1] (📮)       postbox
    Interval{ 0x1F4EF, 0x1F4EF }, // E1.0   [1] (📯)       postal horn
    Interval{ 0x1F4F0, 0x1F4F4 }, // E0.6   [5] (📰..📴)    newspaper..mobile phone off
    Interval{ 0x1F4F5, 0x1F4F5 }, // E1.0   [1] (📵)       no mobile phones
    Interval{ 0x1F4F6, 0x1F4F7 }, // E0.6   [2] (📶..📷)    antenna bars..camera
    Interval{ 0x1F4F8, 0x1F4F8 }, // E1.0   [1] (📸)       camera with flash
    Interval{ 0x1F4F9, 0x1F4FC }, // E0.6   [4] (📹..📼)    video camera..videocassette
    Interval{ 0x1F4FF, 0x1F502 }, // E1.0   [4] (📿..🔂)    prayer beads..repeat single button
    Interval{ 0x1F503, 0x1F503 }, // E0.6   [1] (🔃)       clockwise vertical arrows
    Interval{ 0x1F504, 0x1F507 }, // E1.0   [4] (🔄..🔇)    counterclockwise arrows button..muted speaker
    Interval{ 0x1F508, 0x1F508 }, // E0.7   [1] (🔈)       speaker low volume
    Interval{ 0x1F509, 0x1F509 }, // E1.0   [1] (🔉)       speaker medium volume
    Interval{ 0x1F50A, 0x1F514 }, // E0.6  [11] (🔊..🔔)    speaker high volume..bell
    Interval{ 0x1F515, 0x1F515 }, // E1.0   [1] (🔕)       bell with slash
    Interval{ 0x1F516, 0x1F52B }, // E0.6  [22] (🔖..🔫)    bookmark..water pistol
    Interval{ 0x1F52C, 0x1F52D }, // E1.0   [2] (🔬..🔭)    microscope..telescope
    Interval{ 0x1F52E, 0x1F53D }, // E0.6  [16] (🔮..🔽)    crystal ball..downwards button
    Interval{ 0x1F54B, 0x1F54E }, // E1.0   [4] (🕋..🕎)    kaaba..menorah
    Interval{ 0x1F550, 0x1F55B }, // E0.6  [12] (🕐..🕛)    one o’clock..twelve o’clock
    Interval{ 0x1F55C, 0x1F567 }, // E0.7  [12] (🕜..🕧)    one-thirty..twelve-thirty
    Interval{ 0x1F57A, 0x1F57A }, // E3.0   [1] (🕺)       man dancing
    Interval{ 0x1F595, 0x1F596 }, // E1.0   [2] (🖕..🖖)    middle finger..vulcan salute
    Interval{ 0x1F5A4, 0x1F5A4 }, // E3.0   [1] (🖤)       black heart
    Interval{ 0x1F5FB, 0x1F5FF }, // E0.6   [5] (🗻..🗿)    mount fuji..moai
    Interval{ 0x1F600, 0x1F600 }, // E1.0   [1] (😀)       grinning face
    Interval{ 0x1F601, 0x1F606 }, // E0.6   [6] (😁..😆)    beaming face with smiling eyes..grinning squinting face
    Interval{ 0x1F607, 0x1F608 }, // E1.0   [2] (😇..😈)    smiling face with halo..smiling face with horns
    Interval{ 0x1F609, 0x1F60D }, // E0.6   [5] (😉..😍)    winking face..smiling face with heart-eyes
    Interval{ 0x1F60E, 0x1F60E }, // E1.0   [1] (😎)       smiling face with sunglasses
    Interval{ 0x1F60F, 0x1F60F }, // E0.6   [1] (😏)       smirking face
    Interval{ 0x1F610, 0x1F610 }, // E0.7   [1] (😐)       neutral face
    Interval{ 0x1F611, 0x1F611 }, // E1.0   [1] (😑)       expressionless face
    Interval{ 0x1F612, 0x1F614 }, // E0.6   [3] (😒..😔)    unamused face..pensive face
    Interval{ 0x1F615, 0x1F615 }, // E1.0   [1] (😕)       confused face
    Interval{ 0x1F616, 0x1F616 }, // E0.6   [1] (😖)       confounded face
    Interval{ 0x1F617, 0x1F617 }, // E1.0   [1] (😗)       kissing face
    Interval{ 0x1F618, 0x1F618 }, // E0.6   [1] (😘)       face blowing a kiss
    Interval{ 0x1F619, 0x1F619 }, // E1.0   [1] (😙)       kissing face with smiling eyes
    Interval{ 0x1F61A, 0x1F61A }, // E0.6   [1] (😚)       kissing face with closed eyes
    Interval{ 0x1F61B, 0x1F61B }, // E1.0   [1] (😛)       face with tongue
    Interval{ 0x1F61C, 0x1F61E }, // E0.6   [3] (😜..😞)    winking face with tongue..disappointed face
    Interval{ 0x1F61F, 0x1F61F }, // E1.0   [1] (😟)       worried face
    Interval{ 0x1F620, 0x1F625 }, // E0.6   [6] (😠..😥)    angry face..sad but relieved face
    Interval{ 0x1F626, 0x1F627 }, // E1.0   [2] (😦..😧)    frowning face with open mouth..anguished face
    Interval{ 0x1F628, 0x1F62B }, // E0.6   [4] (😨..😫)    fearful face..tired face
    Interval{ 0x1F62C, 0x1F62C }, // E1.0   [1] (😬)       grimacing face
    Interval{ 0x1F62D, 0x1F62D }, // E0.6   [1] (😭)       loudly crying face
    Interval{ 0x1F62E, 0x1F62F }, // E1.0   [2] (😮..😯)    face with open mouth..hushed face
    Interval{ 0x1F630, 0x1F633 }, // E0.6   [4] (😰..😳)    anxious face with sweat..flushed face
    Interval{ 0x1F634, 0x1F634 }, // E1.0   [1] (😴)       sleeping face
    Interval{ 0x1F635, 0x1F635 }, // E0.6   [1] (😵)       face with crossed-out eyes
    Interval{ 0x1F636, 0x1F636 }, // E1.0   [1] (😶)       face without mouth
    Interval{ 0x1F637, 0x1F640 }, // E0.6  [10] (😷..🙀)    face with medical mask..weary cat
    Interval{ 0x1F641, 0x1F644 }, // E1.0   [4] (🙁..🙄)    slightly frowning face..face with rolling eyes
    Interval{ 0x1F645, 0x1F64F }, // E0.6  [11] (🙅..🙏)    person gesturing NO..folded hands
    Interval{ 0x1F680, 0x1F680 }, // E0.6   [1] (🚀)       rocket
    Interval{ 0x1F681, 0x1F682 }, // E1.0   [2] (🚁..🚂)    helicopter..locomotive
    Interval{ 0x1F683, 0x1F685 }, // E0.6   [3] (🚃..🚅)    railway car..bullet train
    Interval{ 0x1F686, 0x1F686 }, // E1.0   [1] (🚆)       train
    Interval{ 0x1F687, 0x1F687 }, // E0.6   [1] (🚇)       metro
    Interval{ 0x1F688, 0x1F688 }, // E1.0   [1] (🚈)       light rail
    Interval{ 0x1F689, 0x1F689 }, // E0.6   [1] (🚉)       station
    Interval{ 0x1F68A, 0x1F68B }, // E1.0   [2] (🚊..🚋)    tram..tram car
    Interval{ 0x1F68C, 0x1F68C }, // E0.6   [1] (🚌)       bus
    Interval{ 0x1F68D, 0x1F68D }, // E0.7   [1] (🚍)       oncoming bus
    Interval{ 0x1F68E, 0x1F68E }, // E1.0   [1] (🚎)       trolleybus
    Interval{ 0x1F68F, 0x1F68F }, // E0.6   [1] (🚏)       bus stop
    Interval{ 0x1F690, 0x1F690 }, // E1.0   [1] (🚐)       minibus
    Interval{ 0x1F691, 0x1F693 }, // E0.6   [3] (🚑..🚓)    ambulance..police car
    Interval{ 0x1F694, 0x1F694 }, // E0.7   [1] (🚔)       oncoming police car
    Interval{ 0x1F695, 0x1F695 }, // E0.6   [1] (🚕)       taxi
    Interval{ 0x1F696, 0x1F696 }, // E1.0   [1] (🚖)       oncoming taxi
    Interval{ 0x1F697, 0x1F697 }, // E0.6   [1] (🚗)       automobile
    Interval{ 0x1F698, 0x1F698 }, // E0.7   [1] (🚘)       oncoming automobile
    Interval{ 0x1F699, 0x1F69A }, // E0.6   [2] (🚙..🚚)    sport utility vehicle..delivery truck
    Interval{ 0x1F69B, 0x1F6A1 }, // E1.0   [7] (🚛..🚡)    articulated lorry..aerial tramway
    Interval{ 0x1F6A2, 0x1F6A2 }, // E0.6   [1] (🚢)       ship
    Interval{ 0x1F6A3, 0x1F6A3 }, // E1.0   [1] (🚣)       person rowing boat
    Interval{ 0x1F6A4, 0x1F6A5 }, // E0.6   [2] (🚤..🚥)    speedboat..horizontal traffic light
    Interval{ 0x1F6A6, 0x1F6A6 }, // E1.0   [1] (🚦)       vertical traffic light
    Interval{ 0x1F6A7, 0x1F6AD }, // E0.6   [7] (🚧..🚭)    construction..no smoking
    Interval{ 0x1F6AE, 0x1F6B1 }, // E1.0   [4] (🚮..🚱)    litter in bin sign..non-potable water
    Interval{ 0x1F6B2, 0x1F6B2 }, // E0.6   [1] (🚲)       bicycle
    Interval{ 0x1F6B3, 0x1F6B5 }, // E1.0   [3] (🚳..🚵)    no bicycles..person mountain biking
    Interval{ 0x1F6B6, 0x1F6B6 }, // E0.6   [1] (🚶)       person walking
    Interval{ 0x1F6B7, 0x1F6B8 }, // E1.0   [2] (🚷..🚸)    no pedestrians..children crossing
    Interval{ 0x1F6B9, 0x1F6BE }, // E0.6   [6] (🚹..🚾)    men’s room..water closet
    Interval{ 0x1F6BF, 0x1F6BF }, // E1.0   [1] (🚿)       shower
    Interval{ 0x1F6C0, 0x1F6C0 }, // E0.6   [1] (🛀)       person taking bath
    Interval{ 0x1F6C1, 0x1F6C5 }, // E1.0   [5] (🛁..🛅)    bathtub..left luggage
    Interval{ 0x1F6CC, 0x1F6CC }, // E1.0   [1] (🛌)       person in bed
    Interval{ 0x1F6D0, 0x1F6D0 }, // E1.0   [1] (🛐)       place of worship
    Interval{ 0x1F6D1, 0x1F6D2 }, // E3.0   [2] (🛑..🛒)    stop sign..shopping cart
    Interval{ 0x1F6D5, 0x1F6D5 }, // E12.0  [1] (🛕)       hindu temple
    Interval{ 0x1F6D6, 0x1F6D7 }, // E13.0  [2] (🛖..🛗)    hut..elevator
    Interval{ 0x1F6D8, 0x1F6D8 }, // E17.0  [1] (🛘)       landslide
    Interval{ 0x1F6DC, 0x1F6DC }, // E15.0  [1] (🛜)       wireless
    Interval{ 0x1F6DD, 0x1F6DF }, // E14.0  [3] (🛝..🛟)    playground slide..ring buoy
    Interval{ 0x1F6EB, 0x1F6EC }, // E1.0   [2] (🛫..🛬)    airplane departure..airplane arrival
    Interval{ 0x1F6F4, 0x1F6F6 }, // E3.0   [3] (🛴..🛶)    kick scooter..canoe
    Interval{ 0x1F6F7, 0x1F6F8 }, // E5.0   [2] (🛷..🛸)    sled..flying saucer
    Interval{ 0x1F6F9, 0x1F6F9 }, // E11.0  [1] (🛹)       skateboard
    Interval{ 0x1F6FA, 0x1F6FA }, // E12.0  [1] (🛺)       auto rickshaw
    Interval{ 0x1F6FB, 0x1F6FC }, // E13.0  [2] (🛻..🛼)    pickup truck..roller skate
    Interval{ 0x1F7E0, 0x1F7EB }, // E12.0 [12] (🟠..🟫)    orange circle..brown square
    Interval{ 0x1F7F0, 0x1F7F0 }, // E14.0  [1] (🟰)       heavy equals sign
    Interval{ 0x1F90C, 0x1F90C }, // E13.0  [1] (🤌)       pinched fingers
    Interval{ 0x1F90D, 0x1F90F }, // E12.0  [3] (🤍..🤏)    white heart..pinching hand
    Interval{ 0x1F910, 0x1F918 }, // E1.0   [9] (🤐..🤘)    zipper-mouth face..sign of the horns
    Interval{ 0x1F919, 0x1F91E }, // E3.0   [6] (🤙..🤞)    call me hand..crossed fingers
    Interval{ 0x1F91F, 0x1F91F }, // E5.0   [1] (🤟)       love-you gesture
    Interval{ 0x1F920, 0x1F927 }, // E3.0   [8] (🤠..🤧)    cowboy hat face..sneezing face
    Interval{ 0x1F928, 0x1F92F }, // E5.0   [8] (🤨..🤯)    face with raised eyebrow..exploding head
    Interval{ 0x1F930, 0x1F930 }, // E3.0   [1] (🤰)       pregnant woman
    Interval{ 0x1F931, 0x1F932 }, // E5.0   [2] (🤱..🤲)    breast-feeding..palms up together
    Interval{ 0x1F933, 0x1F93A }, // E3.0   [8] (🤳..🤺)    selfie..person fencing
    Interval{ 0x1F93C, 0x1F93E }, // E3.0   [3] (🤼..🤾)    people wrestling..person playing handball
    Interval{ 0x1F93F, 0x1F93F }, // E12.0  [1] (🤿)       diving mask
    Interval{ 0x1F940, 0x1F945 }, // E3.0   [6] (🥀..🥅)    wilted flower..goal net
    Interval{ 0x1F947, 0x1F94B }, // E3.0   [5] (🥇..🥋)    1st place medal..martial arts uniform
    Interval{ 0x1F94C, 0x1F94C }, // E5.0   [1] (🥌)       curling stone
    Interval{ 0x1F94D, 0x1F94F }, // E11.0  [3] (🥍..🥏)    lacrosse..flying disc
    Interval{ 0x1F950, 0x1F95E }, // E3.0  [15] (🥐..🥞)    croissant..pancakes
    Interval{ 0x1F95F, 0x1F96B }, // E5.0  [13] (🥟..🥫)    dumpling..canned food
    Interval{ 0x1F96C, 0x1F970 }, // E11.0  [5] (🥬..🥰)    leafy green..smiling face with hearts
    Interval{ 0x1F971, 0x1F971 }, // E12.0  [1] (🥱)       yawning face
    Interval{ 0x1F972, 0x1F972 }, // E13.0  [1] (🥲)       smiling face with tear
    Interval{ 0x1F973, 0x1F976 }, // E11.0  [4] (🥳..🥶)    partying face..cold face
    Interval{ 0x1F977, 0x1F978 }, // E13.0  [2] (🥷..🥸)    ninja..disguised face
    Interval{ 0x1F979, 0x1F979 }, // E14.0  [1] (🥹)       face holding back tears
    Interval{ 0x1F97A, 0x1F97A }, // E11.0  [1] (🥺)       pleading face
    Interval{ 0x1F97B, 0x1F97B }, // E12.0  [1] (🥻)       sari
    Interval{ 0x1F97C, 0x1F97F }, // E11.0  [4] (🥼..🥿)    lab coat..flat shoe
    Interval{ 0x1F980, 0x1F984 }, // E1.0   [5] (🦀..🦄)    crab..unicorn
    Interval{ 0x1F985, 0x1F991 }, // E3.0  [13] (🦅..🦑)    eagle..squid
    Interval{ 0x1F992, 0x1F997 }, // E5.0   [6] (🦒..🦗)    giraffe..cricket
    Interval{ 0x1F998, 0x1F9A2 }, // E11.0 [11] (🦘..🦢)    kangaroo..swan
    Interval{ 0x1F9A3, 0x1F9A4 }, // E13.0  [2] (🦣..🦤)    mammoth..dodo
    Interval{ 0x1F9A5, 0x1F9AA }, // E12.0  [6] (🦥..🦪)    sloth..oyster
    Interval{ 0x1F9AB, 0x1F9AD }, // E13.0  [3] (🦫..🦭)    beaver..seal
    Interval{ 0x1F9AE, 0x1F9AF }, // E12.0  [2] (🦮..🦯)    guide dog..white cane
    Interval{ 0x1F9B0, 0x1F9B9 }, // E11.0 [10] (🦰..🦹)    red hair..supervillain
    Interval{ 0x1F9BA, 0x1F9BF }, // E12.0  [6] (🦺..🦿)    safety vest..mechanical leg
    Interval{ 0x1F9C0, 0x1F9C0 }, // E1.0   [1] (🧀)       cheese wedge
    Interval{ 0x1F9C1, 0x1F9C2 }, // E11.0  [2] (🧁..🧂)    cupcake..salt
    Interval{ 0x1F9C3, 0x1F9CA }, // E12.0  [8] (🧃..🧊)    beverage box..ice
    Interval{ 0x1F9CB, 0x1F9CB }, // E13.0  [1] (🧋)       bubble tea
    Interval{ 0x1F9CC, 0x1F9CC }, // E14.0  [1] (🧌)       troll
    Interval{ 0x1F9CD, 0x1F9CF }, // E12.0  [3] (🧍..🧏)    person standing..deaf person
    Interval{ 0x1F9D0, 0x1F9E6 }, // E5.0  [23] (🧐..🧦)    face with monocle..socks
    Interval{ 0x1F9E7, 0x1F9FF }, // E11.0 [25] (🧧..🧿)    red envelope..nazar amulet
    Interval{ 0x1FA70, 0x1FA73 }, // E12.0  [4] (🩰..🩳)    ballet shoes..shorts
    Interval{ 0x1FA74, 0x1FA74 }, // E13.0  [1] (🩴)       thong sandal
    Interval{ 0x1FA75, 0x1FA77 }, // E15.0  [3] (🩵..🩷)    light blue heart..pink heart
    Interval{ 0x1FA78, 0x1FA7A }, // E12.0  [3] (🩸..🩺)    drop of blood..stethoscope
    Interval{ 0x1FA7B, 0x1FA7C }, // E14.0  [2] (🩻..🩼)    x-ray..crutch
    Interval{ 0x1FA80, 0x1FA82 }, // E12.0  [3] (🪀..🪂)    yo-yo..parachute
    Interval{ 0x1FA83, 0x1FA86 }, // E13.0  [4] (🪃..🪆)    boomerang..nesting dolls
    Interval{ 0x1FA87, 0x1FA88 }, // E15.0  [2] (🪇..🪈)    maracas..flute
    Interval{ 0x1FA89, 0x1FA89 }, // E16.0  [1] (🪉)       harp
    Interval{ 0x1FA8A, 0x1FA8A }, // E17.0  [1] (🪊)       trombone
    Interval{ 0x1FA8E, 0x1FA8E }, // E17.0  [1] (🪎)       treasure chest
    Interval{ 0x1FA8F, 0x1FA8F }, // E16.0  [1] (🪏)       shovel
    Interval{ 0x1FA90, 0x1FA95 }, // E12.0  [6] (🪐..🪕)    ringed planet..banjo
    Interval{ 0x1FA96, 0x1FAA8 }, // E13.0 [19] (🪖..🪨)    military helmet..rock
    Interval{ 0x1FAA9, 0x1FAAC }, // E14.0  [4] (🪩..🪬)    mirror ball..hamsa
    Interval{ 0x1FAAD, 0x1FAAF }, // E15.0  [3] (🪭..🪯)    folding hand fan..khanda
    Interval{ 0x1FAB0, 0x1FAB6 }, // E13.0  [7] (🪰..🪶)    fly..feather
    Interval{ 0x1FAB7, 0x1FABA }, // E14.0  [4] (🪷..🪺)    lotus..nest with eggs
    Interval{ 0x1FABB, 0x1FABD }, // E15.0  [3] (🪻..🪽)    hyacinth..wing
    Interval{ 0x1FABE, 0x1FABE }, // E16.0  [1] (🪾)       leafless tree
    Interval{ 0x1FABF, 0x1FABF }, // E15.0  [1] (🪿)       goose
    Interval{ 0x1FAC0, 0x1FAC2 }, // E13.0  [3] (🫀..🫂)    anatomical heart..people hugging
    Interval{ 0x1FAC3, 0x1FAC5 }, // E14.0  [3] (🫃..🫅)    pregnant man..person with crown
    Interval{ 0x1FAC6, 0x1FAC6 }, // E16.0  [1] (🫆)       fingerprint
    Interval{ 0x1FAC8, 0x1FAC8 }, // E17.0  [1] (🫈)       hairy creature
    Interval{ 0x1FACD, 0x1FACD }, // E17.0  [1] (🫍)       orca
    Interval{ 0x1FACE, 0x1FACF }, // E15.0  [2] (🫎..🫏)    moose..donkey
    Interval{ 0x1FAD0, 0x1FAD6 }, // E13.0  [7] (🫐..🫖)    blueberries..teapot
    Interval{ 0x1FAD7, 0x1FAD9 }, // E14.0  [3] (🫗..🫙)    pouring liquid..jar
    Interval{ 0x1FADA, 0x1FADB }, // E15.0  [2] (🫚..🫛)    ginger root..pea pod
    Interval{ 0x1FADC, 0x1FADC }, // E16.0  [1] (🫜)       root vegetable
    Interval{ 0x1FADF, 0x1FADF }, // E16.0  [1] (🫟)       splatter
    Interval{ 0x1FAE0, 0x1FAE7 }, // E14.0  [8] (🫠..🫧)    melting face..bubbles
    Interval{ 0x1FAE8, 0x1FAE8 }, // E15.0  [1] (🫨)       shaking face
    Interval{ 0x1FAE9, 0x1FAE9 }, // E16.0  [1] (🫩)       face with bags under eyes
    Interval{ 0x1FAEA, 0x1FAEA }, // E17.0  [1] (🫪)       distorted face
    Interval{ 0x1FAEF, 0x1FAEF }, // E17.0  [1] (🫯)       fight cloud
    Interval{ 0x1FAF0, 0x1FAF6 }, // E14.0  [7] (🫰..🫶)    hand with index finger and thumb crossed..heart hands
    Interval{ 0x1FAF7, 0x1FAF8 }, // E15.0  [2] (🫷..🫸)    leftwards pushing hand..rightwards pushing hand
}; // }}}
auto static const Extended_Pictographic = std::array<Interval, 451>{ // {{{
    Interval{ 0x00A9, 0x00A9 }, // E0.6   [1] (©️)       copyright
    Interval{ 0x00AE, 0x00AE }, // E0.6   [1] (®️)       registered
    Interval{ 0x203C, 0x203C }, // E0.6   [1] (‼️)       double exclamation mark
    Interval{ 0x2049, 0x2049 }, // E0.6   [1] (⁉️)       exclamation question mark
    Interval{ 0x2122, 0x2122 }, // E0.6   [1] (™️)       trade mark
    Interval{ 0x2139, 0x2139 }, // E0.6   [1] (ℹ️)       information
    Interval{ 0x2194, 0x2199 }, // E0.6   [6] (↔️..↙️)    left-right arrow..down-left arrow
    Interval{ 0x21A9, 0x21AA }, // E0.6   [2] (↩️..↪️)    right arrow curving left..left arrow curving right
    Interval{ 0x231A, 0x231B }, // E0.6   [2] (⌚..⌛)    watch..hourglass done
    Interval{ 0x2328, 0x2328 }, // E1.0   [1] (⌨️)       keyboard
    Interval{ 0x23CF, 0x23CF }, // E1.0   [1] (⏏️)       eject button
    Interval{ 0x23E9, 0x23EC }, // E0.6   [4] (⏩..⏬)    fast-forward button..fast down button
    Interval{ 0x23ED, 0x23EE }, // E0.7   [2] (⏭️..⏮️)    next track button..last track button
    Interval{ 0x23EF, 0x23EF }, // E1.0   [1] (⏯️)       play or pause button
    Interval{ 0x23F0, 0x23F0 }, // E0.6   [1] (⏰)       alarm clock
    Interval{ 0x23F1, 0x23F2 }, // E1.0   [2] (⏱️..⏲️)    stopwatch..timer clock
    Interval{ 0x23F3, 0x23F3 }, // E0.6   [1] (⏳)       hourglass not done
    Interval{ 0x23F8, 0x23FA }, // E0.7   [3] (⏸️..⏺️)    pause button..record button
    Interval{ 0x24C2, 0x24C2 }, // E0.6   [1] (Ⓜ️)       circled M
    Interval{ 0x25AA, 0x25AB }, // E0.6   [2] (▪️..▫️)    black small square..white small square
    Interval{ 0x25B6, 0x25B6 }, // E0.6   [1] (▶️)       play button
    Interval{ 0x25C0, 0x25C0 }, // E0.6   [1] (◀️)       reverse button
    Interval{ 0x25FB, 0x25FE }, // E0.6   [4] (◻️..◾)    white medium square..black medium-small square
    Interval{ 0x2600, 0x2601 }, // E0.6   [2] (☀️..☁️)    sun..cloud
    Interval{ 0x2602, 0x2603 }, // E0.7   [2] (☂️..☃️)    umbrella..snowman
    Interval{ 0x2604, 0x2604 }, // E1.0   [1] (☄️)       comet
    Interval{ 0x260E, 0x260E }, // E0.6   [1] (☎️)       telephone
    Interval{ 0x2611, 0x2611 }, // E0.6   [1] (☑️)       check box with check
    Interval{ 0x2614, 0x2615 }, // E0.6   [2] (☔..☕)    umbrella with rain drops..hot beverage
    Interval{ 0x2618, 0x2618 }, // E1.0   [1] (☘️)       shamrock
    Interval{ 0x261D, 0x261D }, // E0.6   [1] (☝️)       index pointing up
    Interval{ 0x2620, 0x2620 }, // E1.0   [1] (☠️)       skull and crossbones
    Interval{ 0x2622, 0x2623 }, // E1.0   [2] (☢️..☣️)    radioactive..biohazard
    Interval{ 0x2626, 0x2626 }, // E1.0   [1] (☦️)       orthodox cross
    Interval{ 0x262A, 0x262A }, // E0.7   [1] (☪️)       star and crescent
    Interval{ 0x262E, 0x262E }, // E1.0   [1] (☮️)       peace symbol
    Interval{ 0x262F, 0x262F }, // E0.7   [1] (☯️)       yin yang
    Interval{ 0x2638, 0x2639 }, // E0.7   [2] (☸️..☹️)    wheel of dharma..frowning face
    Interval{ 0x263A, 0x263A }, // E0.6   [1] (☺️)       smiling face
    Interval{ 0x2640, 0x2640 }, // E4.0   [1] (♀️)       female sign
    Interval{ 0x2642, 0x2642 }, // E4.0   [1] (♂️)       male sign
    Interval{ 0x2648, 0x2653 }, // E0.6  [12] (♈..♓)    Aries..Pisces
    Interval{ 0x265F, 0x265F }, // E11.0  [1] (♟️)       chess pawn
    Interval{ 0x2660, 0x2660 }, // E0.6   [1] (♠️)       spade suit
    Interval{ 0x2663, 0x2663 }, // E0.6   [1] (♣️)       club suit
    Interval{ 0x2665, 0x2666 }, // E0.6   [2] (♥️..♦️)    heart suit..diamond suit
    Interval{ 0x2668, 0x2668 }, // E0.6   [1] (♨️)       hot springs
    Interval{ 0x267B, 0x267B }, // E0.6   [1] (♻️)       recycling symbol
    Interval{ 0x267E, 0x267E }, // E11.0  [1] (♾️)       infinity
    Interval{ 0x267F, 0x267F }, // E0.6   [1] (♿)       wheelchair symbol
    Interval{ 0x2692, 0x2692 }, // E1.0   [1] (⚒️)       hammer and pick
    Interval{ 0x2693, 0x2693 }, // E0.6   [1] (⚓)       anchor
    Interval{ 0x2694, 0x2694 }, // E1.0   [1] (⚔️)       crossed swords
    Interval{ 0x2695, 0x2695 }, // E4.0   [1] (⚕️)       medical symbol
    Interval{ 0x2696, 0x2697 }, // E1.0   [2] (⚖️..⚗️)    balance scale..alembic
    Interval{ 0x2699, 0x2699 }, // E1.0   [1] (⚙️)       gear
    Interval{ 0x269B, 0x269C }, // E1.0   [2] (⚛️..⚜️)    atom symbol..fleur-de-lis
    Interval{ 0x26A0, 0x26A1 }, // E0.6   [2] (⚠️..⚡)    warning..high voltage
    Interval{ 0x26A7, 0x26A7 }, // E13.0  [1] (⚧️)       transgender symbol
    Interval{ 0x26AA, 0x26AB }, // E0.6   [2] (⚪..⚫)    white circle..black circle
    Interval{ 0x26B0, 0x26B1 }, // E1.0   [2] (⚰️..⚱️)    coffin..funeral urn
    Interval{ 0x26BD, 0x26BE }, // E0.6   [2] (⚽..⚾)    soccer ball..baseball
    Interval{ 0x26C4, 0x26C5 }, // E0.6   [2] (⛄..⛅)    snowman without snow..sun behind cloud
    Interval{ 0x26C8, 0x26C8 }, // E0.7   [1] (⛈️)       cloud with lightning and rain
    Interval{ 0x26CE, 0x26CE }, // E0.6   [1] (⛎)       Ophiuchus
    Interval{ 0x26CF, 0x26CF }, // E0.7   [1] (⛏️)       pick
    Interval{ 0x26D1, 0x26D1 }, // E0.7   [1] (⛑️)       rescue worker’s helmet
    Interval{ 0x26D3, 0x26D3 }, // E0.7   [1] (⛓️)       chains
    Interval{ 0x26D4, 0x26D4 }, // E0.6   [1] (⛔)       no entry
    Interval{ 0x26E9, 0x26E9 }, // E0.7   [1] (⛩️)       shinto shrine
    Interval{ 0x26EA, 0x26EA }, // E0.6   [1] (⛪)       church
    Interval{ 0x26F0, 0x26F1 }, // E0.7   [2] (⛰️..⛱️)    mountain..umbrella on ground
    Interval{ 0x26F2, 0x26F3 }, // E0.6   [2] (⛲..⛳)    fountain..flag in hole
    Interval{ 0x26F4, 0x26F4 }, // E0.7   [1] (⛴️)       ferry
    Interval{ 0x26F5, 0x26F5 }, // E0.6   [1] (⛵)       sailboat
    Interval{ 0x26F7, 0x26F9 }, // E0.7   [3] (⛷️..⛹️)    skier..person bouncing ball
    Interval{ 0x26FA, 0x26FA }, // E0.6   [1] (⛺)       tent
    Interval{ 0x26FD, 0x26FD }, // E0.6   [1] (⛽)       fuel pump
    Interval{ 0x2702, 0x2702 }, // E0.6   [1] (✂️)       scissors
    Interval{ 0x2705, 0x2705 }, // E0.6   [1] (✅)       check mark button
    Interval{ 0x2708, 0x270C }, // E0.6   [5] (✈️..✌️)    airplane..victory hand
    Interval{ 0x270D, 0x270D }, // E0.7   [1] (✍️)       writing hand
    Interval{ 0x270F, 0x270F }, // E0.6   [1] (✏️)       pencil
    Interval{ 0x2712, 0x2712 }, // E0.6   [1] (✒️)       black nib
    Interval{ 0x2714, 0x2714 }, // E0.6   [1] (✔️)       check mark
    Interval{ 0x2716, 0x2716 }, // E0.6   [1] (✖️)       multiply
    Interval{ 0x271D, 0x271D }, // E0.7   [1] (✝️)       latin cross
    Interval{ 0x2721, 0x2721 }, // E0.7   [1] (✡️)       star of David
    Interval{ 0x2728, 0x2728 }, // E0.6   [1] (✨)       sparkles
    Interval{ 0x2733, 0x2734 }, // E0.6   [2] (✳️..✴️)    eight-spoked asterisk..eight-pointed star
    Interval{ 0x2744, 0x2744 }, // E0.6   [1] (❄️)       snowflake
    Interval{ 0x2747, 0x2747 }, // E0.6   [1] (❇️)       sparkle
    Interval{ 0x274C, 0x274C }, // E0.6   [1] (❌)       cross mark
    Interval{ 0x274E, 0x274E }, // E0.6   [1] (❎)       cross mark button
    Interval{ 0x2753, 0x2755 }, // E0.6   [3] (❓..❕)    red question mark..white exclamation mark
    Interval{ 0x2757, 0x2757 }, // E0.6   [1] (❗)       red exclamation mark
    Interval{ 0x2763, 0x2763 }, // E1.0   [1] (❣️)       heart exclamation
    Interval{ 0x2764, 0x2764 }, // E0.6   [1] (❤️)       red heart
    Interval{ 0x2795, 0x2797 }, // E0.6   [3] (➕..➗)    plus..divide
    Interval{ 0x27A1, 0x27A1 }, // E0.6   [1] (➡️)       right arrow
    Interval{ 0x27B0, 0x27B0 }, // E0.6   [1] (➰)       curly loop
    Interval{ 0x27BF, 0x27BF }, // E1.0   [1] (➿)       double curly loop
    Interval{ 0x2934, 0x2935 }, // E0.6   [2] (⤴️..⤵️)    right arrow curving up..right arrow curving down
    Interval{ 0x2B05, 0x2B07 }, // E0.6   [3] (⬅️..⬇️)    left arrow..down arrow
    Interval{ 0x2B1B, 0x2B1C }, // E0.6   [2] (⬛..⬜)    black large square..white large square
    Interval{ 0x2B50, 0x2B50 }, // E0.6   [1] (⭐)       star
    Interval{ 0x2B55, 0x2B55 }, // E0.6   [1] (⭕)       hollow red circle
    Interval{ 0x3030, 0x3030 }, // E0.6   [1] (〰️)       wavy dash
    Interval{ 0x303D, 0x303D }, // E0.6   [1] (〽️)       part alternation mark
    Interval{ 0x3297, 0x3297 }, // E0.6   [1] (㊗️)       Japanese “congratulations” button
    Interval{ 0x3299, 0x3299 }, // E0.6   [1] (㊙️)       Japanese “secret” button
    Interval{ 0x1F004, 0x1F004 }, // E0.6   [1] (🀄)       mahjong red dragon
    Interval{ 0x1F02C, 0x1F02F }, // E0.0   [4] (🀬..🀯)    <reserved-1F02C>..<reserved-1F02F>
    Interval{ 0x1F094, 0x1F09F }, // E0.0  [12] (🂔..🂟)    <reserved-1F094>..<reserved-1F09F>
    Interval{ 0x1F0AF, 0x1F0B0 }, // E0.0   [2] (🂯..🂰)    <reserved-1F0AF>..<reserved-1F0B0>
    Interval{ 0x1F0C0, 0x1F0C0 }, // E0.0   [1] (🃀)       <reserved-1F0C0>
    Interval{ 0x1F0CF, 0x1F0CF }, // E0.6   [1] (🃏)       joker
    Interval{ 0x1F0D0, 0x1F0D0 }, // E0.0   [1] (🃐)       <reserved-1F0D0>
    Interval{ 0x1F0F6, 0x1F0FF }, // E0.0  [10] (🃶..🃿)    <reserved-1F0F6>..<reserved-1F0FF>
    Interval{ 0x1F170, 0x1F171 }, // E0.6   [2] (🅰️..🅱️)    A button (blood type)..B button (blood type)
    Interval{ 0x1F17E, 0x1F17F }, // E0.6   [2] (🅾️..🅿️)    O button (blood type)..P button
    Interval{ 0x1F18E, 0x1F18E }, // E0.6   [1] (🆎)       AB button (blood type)
    Interval{ 0x1F191, 0x1F19A }, // E0.6  [10] (🆑..🆚)    CL button..VS button
    Interval{ 0x1F1AE, 0x1F1E5 }, // E0.0  [56] (🆮..🇥)    <reserved-1F1AE>..<reserved-1F1E5>
    Interval{ 0x1F201, 0x1F202 }, // E0.6   [2] (🈁..🈂️)    Japanese “here” button..Japanese “service charge” button
    Interval{ 0x1F203, 0x1F20F }, // E0.0  [13] (🈃..🈏)    <reserved-1F203>..<reserved-1F20F>
    Interval{ 0x1F21A, 0x1F21A }, // E0.6   [1] (🈚)       Japanese “free of charge” button
    Interval{ 0x1F22F, 0x1F22F }, // E0.6   [1] (🈯)       Japanese “reserved” button
    Interval{ 0x1F232, 0x1F23A }, // E0.6   [9] (🈲..🈺)    Japanese “prohibited” button..Japanese “open for business” button
    Interval{ 0x1F23C, 0x1F23F }, // E0.0   [4] (🈼..🈿)    <reserved-1F23C>..<reserved-1F23F>
    Interval{ 0x1F249, 0x1F24F }, // E0.0   [7] (🉉..🉏)    <reserved-1F249>..<reserved-1F24F>
    Interval{ 0x1F250, 0x1F251 }, // E0.6   [2] (🉐..🉑)    Japanese “bargain” button..Japanese “acceptable” button
    Interval{ 0x1F252, 0x1F25F }, // E0.0  [14] (🉒..🉟)    <reserved-1F252>..<reserved-1F25F>
    Interval{ 0x1F266, 0x1F2FF }, // E0.0 [154] (🉦..🋿)    <reserved-1F266>..<reserved-1F2FF>
    Interval{ 0x1F300, 0x1F30C }, // E0.6  [13] (🌀..🌌)    cyclone..milky way
    Interval{ 0x1F30D, 0x1F30E }, // E0.7   [2] (🌍..🌎)    globe showing Europe-Africa..globe showing Americas
    Interval{ 0x1F30F, 0x1F30F }, // E0.6   [1] (🌏)       globe showing Asia-Australia
    Interval{ 0x1F310, 0x1F310 }, // E1.0   [1] (🌐)       globe with meridians
    Interval{ 0x1F311, 0x1F311 }, // E0.6   [1] (🌑)       new moon
    Interval{ 0x1F312, 0x1F312 }, // E1.0   [1] (🌒)       waxing crescent moon
    Interval{ 0x1F313, 0x1F315 }, // E0.6   [3] (🌓..🌕)    first quarter moon..full moon
    Interval{ 0x1F316, 0x1F318 }, // E1.0   [3] (🌖..🌘)    waning gibbous moon..waning crescent moon
    Interval{ 0x1F319, 0x1F319 }, // E0.6   [1] (🌙)       crescent moon
    Interval{ 0x1F31A, 0x1F31A }, // E1.0   [1] (🌚)       new moon face
    Interval{ 0x1F31B, 0x1F31B }, // E0.6   [1] (🌛)       first quarter moon face
    Interval{ 0x1F31C, 0x1F31C }, // E0.7   [1] (🌜)       last quarter moon face
    Interval{ 0x1F31D, 0x1F31E }, // E1.0   [2] (🌝..🌞)    full moon face..sun with face
    Interval{ 0x1F31F, 0x1F320 }, // E0.6   [2] (🌟..🌠)    glowing star..shooting star
    Interval{ 0x1F321, 0x1F321 }, // E0.7   [1] (🌡️)       thermometer
    Interval{ 0x1F324, 0x1F32C }, // E0.7   [9] (🌤️..🌬️)    sun behind small cloud..wind face
    Interval{ 0x1F32D, 0x1F32F }, // E1.0   [3] (🌭..🌯)    hot dog..burrito
    Interval{ 0x1F330, 0x1F331 }, // E0.6   [2] (🌰..🌱)    chestnut..seedling
    Interval{ 0x1F332, 0x1F333 }, // E1.0   [2] (🌲..🌳)    evergreen tree..deciduous tree
    Interval{ 0x1F334, 0x1F335 }, // E0.6   [2] (🌴..🌵)    palm tree..cactus
    Interval{ 0x1F336, 0x1F336 }, // E0.7   [1] (🌶️)       hot pepper
    Interval{ 0x1F337, 0x1F34A }, // E0.6  [20] (🌷..🍊)    tulip..tangerine
    Interval{ 0x1F34B, 0x1F34B }, // E1.0   [1] (🍋)       lemon
    Interval{ 0x1F34C, 0x1F34F }, // E0.6   [4] (🍌..🍏)    banana..green apple
    Interval{ 0x1F350, 0x1F350 }, // E1.0   [1] (🍐)       pear
    Interval{ 0x1F351, 0x1F37B }, // E0.6  [43] (🍑..🍻)    peach..clinking beer mugs
    Interval{ 0x1F37C, 0x1F37C }, // E1.0   [1] (🍼)       baby bottle
    Interval{ 0x1F37D, 0x1F37D }, // E0.7   [1] (🍽️)       fork and knife with plate
    Interval{ 0x1F37E, 0x1F37F }, // E1.0   [2] (🍾..🍿)    bottle with popping cork..popcorn
    Interval{ 0x1F380, 0x1F393 }, // E0.6  [20] (🎀..🎓)    ribbon..graduation cap
    Interval{ 0x1F396, 0x1F397 }, // E0.7   [2] (🎖️..🎗️)    military medal..reminder ribbon
    Interval{ 0x1F399, 0x1F39B }, // E0.7   [3] (🎙️..🎛️)    studio microphone..control knobs
    Interval{ 0x1F39E, 0x1F39F }, // E0.7   [2] (🎞️..🎟️)    film frames..admission tickets
    Interval{ 0x1F3A0, 0x1F3C4 }, // E0.6  [37] (🎠..🏄)    carousel horse..person surfing
    Interval{ 0x1F3C5, 0x1F3C5 }, // E1.0   [1] (🏅)       sports medal
    Interval{ 0x1F3C6, 0x1F3C6 }, // E0.6   [1] (🏆)       trophy
    Interval{ 0x1F3C7, 0x1F3C7 }, // E1.0   [1] (🏇)       horse racing
    Interval{ 0x1F3C8, 0x1F3C8 }, // E0.6   [1] (🏈)       american football
    Interval{ 0x1F3C9, 0x1F3C9 }, // E1.0   [1] (🏉)       rugby football
    Interval{ 0x1F3CA, 0x1F3CA }, // E0.6   [1] (🏊)       person swimming
    Interval{ 0x1F3CB, 0x1F3CE }, // E0.7   [4] (🏋️..🏎️)    person lifting weights..racing car
    Interval{ 0x1F3CF, 0x1F3D3 }, // E1.0   [5] (🏏..🏓)    cricket game..ping pong
    Interval{ 0x1F3D4, 0x1F3DF }, // E0.7  [12] (🏔️..🏟️)    snow-capped mountain..stadium
    Interval{ 0x1F3E0, 0x1F3E3 }, // E0.6   [4] (🏠..🏣)    house..Japanese post office
    Interval{ 0x1F3E4, 0x1F3E4 }, // E1.0   [1] (🏤)       post office
    Interval{ 0x1F3E5, 0x1F3F0 }, // E0.6  [12] (🏥..🏰)    hospital..castle
    Interval{ 0x1F3F3, 0x1F3F3 }, // E0.7   [1] (🏳️)       white flag
    Interval{ 0x1F3F4, 0x1F3F4 }, // E1.0   [1] (🏴)       black flag
    Interval{ 0x1F3F5, 0x1F3F5 }, // E0.7   [1] (🏵️)       rosette
    Interval{ 0x1F3F7, 0x1F3F7 }, // E0.7   [1] (🏷️)       label
    Interval{ 0x1F3F8, 0x1F3FA }, // E1.0   [3] (🏸..🏺)    badminton..amphora
    Interval{ 0x1F400, 0x1F407 }, // E1.0   [8] (🐀..🐇)    rat..rabbit
    Interval{ 0x1F408, 0x1F408 }, // E0.7   [1] (🐈)       cat
    Interval{ 0x1F409, 0x1F40B }, // E1.0   [3] (🐉..🐋)    dragon..whale
    Interval{ 0x1F40C, 0x1F40E }, // E0.6   [3] (🐌..🐎)    snail..horse
    Interval{ 0x1F40F, 0x1F410 }, // E1.0   [2] (🐏..🐐)    ram..goat
    Interval{ 0x1F411, 0x1F412 }, // E0.6   [2] (🐑..🐒)    ewe..monkey
    Interval{ 0x1F413, 0x1F413 }, // E1.0   [1] (🐓)       rooster
    Interval{ 0x1F414, 0x1F414 }, // E0.6   [1] (🐔)       chicken
    Interval{ 0x1F415, 0x1F415 }, // E0.7   [1] (🐕)       dog
    Interval{ 0x1F416, 0x1F416 }, // E1.0   [1] (🐖)       pig
    Interval{ 0x1F417, 0x1F429 }, // E0.6  [19] (🐗..🐩)    boar..poodle
    Interval{ 0x1F42A, 0x1F42A }, // E1.0   [1] (🐪)       camel
    Interval{ 0x1F42B, 0x1F43E }, // E0.6  [20] (🐫..🐾)    two-hump camel..paw prints
    Interval{ 0x1F43F, 0x1F43F }, // E0.7   [1] (🐿️)       chipmunk
    Interval{ 0x1F440, 0x1F440 }, // E0.6   [1] (👀)       eyes
    Interval{ 0x1F441, 0x1F441 }, // E0.7   [1] (👁️)       eye
    Interval{ 0x1F442, 0x1F464 }, // E0.6  [35] (👂..👤)    ear..bust in silhouette
    Interval{ 0x1F465, 0x1F465 }, // E1.0   [1] (👥)       busts in silhouette
    Interval{ 0x1F466, 0x1F46B }, // E0.6   [6] (👦..👫)    boy..woman and man holding hands
    Interval{ 0x1F46C, 0x1F46D }, // E1.0   [2] (👬..👭)    men holding hands..women holding hands
    Interval{ 0x1F46E, 0x1F4AC }, // E0.6  [63] (👮..💬)    police officer..speech balloon
    Interval{ 0x1F4AD, 0x1F4AD }, // E1.0   [1] (💭)       thought balloon
    Interval{ 0x1F4AE, 0x1F4B5 }, // E0.6   [8] (💮..💵)    white flower..dollar banknote
    Interval{ 0x1F4B6, 0x1F4B7 }, // E1.0   [2] (💶..💷)    euro banknote..pound banknote
    Interval{ 0x1F4B8, 0x1F4EB }, // E0.6  [52] (💸..📫)    money with wings..closed mailbox with raised flag
    Interval{ 0x1F4EC, 0x1F4ED }, // E0.7   [2] (📬..📭)    open mailbox with raised flag..open mailbox with lowered flag
    Interval{ 0x1F4EE, 0x1F4EE }, // E0.6   [1] (📮)       postbox
    Interval{ 0x1F4EF, 0x1F4EF }, // E1.0   [1] (📯)       postal horn
    Interval{ 0x1F4F0, 0x1F4F4 }, // E0.6   [5] (📰..📴)    newspaper..mobile phone off
    Interval{ 0x1F4F5, 0x1F4F5 }, // E1.0   [1] (📵)       no mobile phones
    Interval{ 0x1F4F6, 0x1F4F7 }, // E0.6   [2] (📶..📷)    antenna bars..camera
    Interval{ 0x1F4F8, 0x1F4F8 }, // E1.0   [1] (📸)       camera with flash
    Interval{ 0x1F4F9, 0x1F4FC }, // E0.6   [4] (📹..📼)    video camera..videocassette
    Interval{ 0x1F4FD, 0x1F4FD }, // E0.7   [1] (📽️)       film projector
    Interval{ 0x1F4FF, 0x1F502 }, // E1.0   [4] (📿..🔂)    prayer beads..repeat single button
    Interval{ 0x1F503, 0x1F503 }, // E0.6   [1] (🔃)       clockwise vertical arrows
    Interval{ 0x1F504, 0x1F507 }, // E1.0   [4] (🔄..🔇)    counterclockwise arrows button..muted speaker
    Interval{ 0x1F508, 0x1F508 }, // E0.7   [1] (🔈)       speaker low volume
    Interval{ 0x1F509, 0x1F509 }, // E1.0   [1] (🔉)       speaker medium volume
    Interval{ 0x1F50A, 0x1F514 }, // E0.6  [11] (🔊..🔔)    speaker high volume..bell
    Interval{ 0x1F515, 0x1F515 }, // E1.0   [1] (🔕)       bell with slash
    Interval{ 0x1F516, 0x1F52B }, // E0.6  [22] (🔖..🔫)    bookmark..water pistol
    Interval{ 0x1F52C, 0x1F52D }, // E1.0   [2] (🔬..🔭)    microscope..telescope
    Interval{ 0x1F52E, 0x1F53D }, // E0.6  [16] (🔮..🔽)    crystal ball..downwards button
    Interval{ 0x1F549, 0x1F54A }, // E0.7   [2] (🕉️..🕊️)    om..dove
    Interval{ 0x1F54B, 0x1F54E }, // E1.0   [4] (🕋..🕎)    kaaba..menorah
    Interval{ 0x1F550, 0x1F55B }, // E0.6  [12] (🕐..🕛)    one o’clock..twelve o’clock
    Interval{ 0x1F55C, 0x1F567 }, // E0.7  [12] (🕜..🕧)    one-thirty..twelve-thirty
    Interval{ 0x1F56F, 0x1F570 }, // E0.7   [2] (🕯️..🕰️)    candle..mantelpiece clock
    Interval{ 0x1F573, 0x1F579 }, // E0.7   [7] (🕳️..🕹️)    hole..joystick
    Interval{ 0x1F57A, 0x1F57A }, // E3.0   [1] (🕺)       man dancing
    Interval{ 0x1F587, 0x1F587 }, // E0.7   [1] (🖇️)       linked paperclips
    Interval{ 0x1F58A, 0x1F58D }, // E0.7   [4] (🖊️..🖍️)    pen..crayon
    Interval{ 0x1F590, 0x1F590 }, // E0.7   [1] (🖐️)       hand with fingers splayed
    Interval{ 0x1F595, 0x1F596 }, // E1.0   [2] (🖕..🖖)    middle finger..vulcan salute
    Interval{ 0x1F5A4, 0x1F5A4 }, // E3.0   [1] (🖤)       black heart
    Interval{ 0x1F5A5, 0x1F5A5 }, // E0.7   [1] (🖥️)       desktop computer
    Interval{ 0x1F5A8, 0x1F5A8 }, // E0.7   [1] (🖨️)       printer
    Interval{ 0x1F5B1, 0x1F5B2 }, // E0.7   [2] (🖱️..🖲️)    computer mouse..trackball
    Interval{ 0x1F5BC, 0x1F5BC }, // E0.7   [1] (🖼️)       framed picture
    Interval{ 0x1F5C2, 0x1F5C4 }, // E0.7   [3] (🗂️..🗄️)    card index dividers..file cabinet
    Interval{ 0x1F5D1, 0x1F5D3 }, // E0.7   [3] (🗑️..🗓️)    wastebasket..spiral calendar
    Interval{ 0x1F5DC, 0x1F5DE }, // E0.7   [3] (🗜️..🗞️)    clamp..rolled-up newspaper
    Interval{ 0x1F5E1, 0x1F5E1 }, // E0.7   [1] (🗡️)       dagger
    Interval{ 0x1F5E3, 0x1F5E3 }, // E0.7   [1] (🗣️)       speaking head
    Interval{ 0x1F5E8, 0x1F5E8 }, // E2.0   [1] (🗨️)       left speech bubble
    Interval{ 0x1F5EF, 0x1F5EF }, // E0.7   [1] (🗯️)       right anger bubble
    Interval{ 0x1F5F3, 0x1F5F3 }, // E0.7   [1] (🗳️)       ballot box with ballot
    Interval{ 0x1F5FA, 0x1F5FA }, // E0.7   [1] (🗺️)       world map
    Interval{ 0x1F5FB, 0x1F5FF }, // E0.6   [5] (🗻..🗿)    mount fuji..moai
    Interval{ 0x1F600, 0x1F600 }, // E1.0   [1] (😀)       grinning face
    Interval{ 0x1F601, 0x1F606 }, // E0.6   [6] (😁..😆)    beaming face with smiling eyes..grinning squinting face
    Interval{ 0x1F607, 0x1F608 }, // E1.0   [2] (😇..😈)    smiling face with halo..smiling face with horns
    Interval{ 0x1F609, 0x1F60D }, // E0.6   [5] (😉..😍)    winking face..smiling face with heart-eyes
    Interval{ 0x1F60E, 0x1F60E }, // E1.0   [1] (😎)       smiling face with sunglasses
    Interval{ 0x1F60F, 0x1F60F }, // E0.6   [1] (😏)       smirking face
    Interval{ 0x1F610, 0x1F610 }, // E0.7   [1] (😐)       neutral face
    Interval{ 0x1F611, 0x1F611 }, // E1.0   [1] (😑)       expressionless face
    Interval{ 0x1F612, 0x1F614 }, // E0.6   [3] (😒..😔)    unamused face..pensive face
    Interval{ 0x1F615, 0x1F615 }, // E1.0   [1] (😕)       confused face
    Interval{ 0x1F616, 0x1F616 }, // E0.6   [1] (😖)       confounded face
    Interval{ 0x1F617, 0x1F617 }, // E1.0   [1] (😗)       kissing face
    Interval{ 0x1F618, 0x1F618 }, // E0.6   [1] (😘)       face blowing a kiss
    Interval{ 0x1F619, 0x1F619 }, // E1.0   [1] (😙)       kissing face with smiling eyes
    Interval{ 0x1F61A, 0x1F61A }, // E0.6   [1] (😚)       kissing face with closed eyes
    Interval{ 0x1F61B, 0x1F61B }, // E1.0   [1] (😛)       face with tongue
    Interval{ 0x1F61C, 0x1F61E }, // E0.6   [3] (😜..😞)    winking face with tongue..disappointed face
    Interval{ 0x1F61F, 0x1F61F }, // E1.0   [1] (😟)       worried face
    Interval{ 0x1F620, 0x1F625 }, // E0.6   [6] (😠..😥)    angry face..sad but relieved face
    Interval{ 0x1F626, 0x1F627 }, // E1.0   [2] (😦..😧)    frowning face with open mouth..anguished face
    Interval{ 0x1F628, 0x1F62B }, // E0.6   [4] (😨..😫)    fearful face..tired face
    Interval{ 0x1F62C, 0x1F62C }, // E1.0   [1] (😬)       grimacing face
    Interval{ 0x1F62D, 0x1F62D }, // E0.6   [1] (😭)       loudly crying face
    Interval{ 0x1F62E, 0x1F62F }, // E1.0   [2] (😮..😯)    face with open mouth..hushed face
    Interval{ 0x1F630, 0x1F633 }, // E0.6   [4] (😰..😳)    anxious face with sweat..flushed face
    Interval{ 0x1F634, 0x1F634 }, // E1.0   [1] (😴)       sleeping face
    Interval{ 0x1F635, 0x1F635 }, // E0.6   [1] (😵)       face with crossed-out eyes
    Interval{ 0x1F636, 0x1F636 }, // E1.0   [1] (😶)       face without mouth
    Interval{ 0x1F637, 0x1F640 }, // E0.6  [10] (😷..🙀)    face with medical mask..weary cat
    Interval{ 0x1F641, 0x1F644 }, // E1.0   [4] (🙁..🙄)    slightly frowning face..face with rolling eyes
    Interval{ 0x1F645, 0x1F64F }, // E0.6  [11] (🙅..🙏)    person gesturing NO..folded hands
    Interval{ 0x1F680, 0x1F680 }, // E0.6   [1] (🚀)       rocket
    Interval{ 0x1F681, 0x1F682 }, // E1.0   [2] (🚁..🚂)    helicopter..locomotive
    Interval{ 0x1F683, 0x1F685 }, // E0.6   [3] (🚃..🚅)    railway car..bullet train
    Interval{ 0x1F686, 0x1F686 }, // E1.0   [1] (🚆)       train
    Interval{ 0x1F687, 0x1F687 }, // E0.6   [1] (🚇)       metro
    Interval{ 0x1F688, 0x1F688 }, // E1.0   [1] (🚈)       light rail
    Interval{ 0x1F689, 0x1F689 }, // E0.6   [1] (🚉)       station
    Interval{ 0x1F68A, 0x1F68B }, // E1.0   [2] (🚊..🚋)    tram..tram car
    Interval{ 0x1F68C, 0x1F68C }, // E0.6   [1] (🚌)       bus
    Interval{ 0x1F68D, 0x1F68D }, // E0.7   [1] (🚍)       oncoming bus
    Interval{ 0x1F68E, 0x1F68E }, // E1.0   [1] (🚎)       trolleybus
    Interval{ 0x1F68F, 0x1F68F }, // E0.6   [1] (🚏)       bus stop
    Interval{ 0x1F690, 0x1F690 }, // E1.0   [1] (🚐)       minibus
    Interval{ 0x1F691, 0x1F693 }, // E0.6   [3] (🚑..🚓)    ambulance..police car
    Interval{ 0x1F694, 0x1F694 }, // E0.7   [1] (🚔)       oncoming police car
    Interval{ 0x1F695, 0x1F695 }, // E0.6   [1] (🚕)       taxi
    Interval{ 0x1F696, 0x1F696 }, // E1.0   [1] (🚖)       oncoming taxi
    Interval{ 0x1F697, 0x1F697 }, // E0.6   [1] (🚗)       automobile
    Interval{ 0x1F698, 0x1F698 }, // E0.7   [1] (🚘)       oncoming automobile
    Interval{ 0x1F699, 0x1F69A }, // E0.6   [2] (🚙..🚚)    sport utility vehicle..delivery truck
    Interval{ 0x1F69B, 0x1F6A1 }, // E1.0   [7] (🚛..🚡)    articulated lorry..aerial tramway
    Interval{ 0x1F6A2, 0x1F6A2 }, // E0.6   [1] (🚢)       ship
    Interval{ 0x1F6A3, 0x1F6A3 }, // E1.0   [1] (🚣)       person rowing boat
    Interval{ 0x1F6A4, 0x1F6A5 }, // E0.6   [2] (🚤..🚥)    speedboat..horizontal traffic light
    Interval{ 0x1F6A6, 0x1F6A6 }, // E1.0   [1] (🚦)       vertical traffic light
    Interval{ 0x1F6A7, 0x1F6AD }, // E0.6   [7] (🚧..🚭)    construction..no smoking
    Interval{ 0x1F6AE, 0x1F6B1 }, // E1.0   [4] (🚮..🚱)    litter in bin sign..non-potable water
    Interval{ 0x1F6B2, 0x1F6B2 }, // E0.6   [1] (🚲)       bicycle
    Interval{ 0x1F6B3, 0x1F6B5 }, // E1.0   [3] (🚳..🚵)    no bicycles..person mountain biking
    Interval{ 0x1F6B6, 0x1F6B6 }, // E0.6   [1] (🚶)       person walking
    Interval{ 0x1F6B7, 0x1F6B8 }, // E1.0   [2] (🚷..🚸)    no pedestrians..children crossing
    Interval{ 0x1F6B9, 0x1F6BE }, // E0.6   [6] (🚹..🚾)    men’s room..water closet
    Interval{ 0x1F6BF, 0x1F6BF }, // E1.0   [1] (🚿)       shower
    Interval{ 0x1F6C0, 0x1F6C0 }, // E0.6   [1] (🛀)       person taking bath
    Interval{ 0x1F6C1, 0x1F6C5 }, // E1.0   [5] (🛁..🛅)    bathtub..left luggage
    Interval{ 0x1F6CB, 0x1F6CB }, // E0.7   [1] (🛋️)       couch and lamp
    Interval{ 0x1F6CC, 0x1F6CC }, // E1.0   [1] (🛌)       person in bed
    Interval{ 0x1F6CD, 0x1F6CF }, // E0.7   [3] (🛍️..🛏️)    shopping bags..bed
    Interval{ 0x1F6D0, 0x1F6D0 }, // E1.0   [1] (🛐)       place of worship
    Interval{ 0x1F6D1, 0x1F6D2 }, // E3.0   [2] (🛑..🛒)    stop sign..shopping cart
    Interval{ 0x1F6D5, 0x1F6D5 }, // E12.0  [1] (🛕)       hindu temple
    Interval{ 0x1F6D6, 0x1F6D7 }, // E13.0  [2] (🛖..🛗)    hut..elevator
    Interval{ 0x1F6D8, 0x1F6D8 }, // E17.0  [1] (🛘)       landslide
    Interval{ 0x1F6D9, 0x1F6DB }, // E0.0   [3] (🛙..🛛)    <reserved-1F6D9>..<reserved-1F6DB>
    Interval{ 0x1F6DC, 0x1F6DC }, // E15.0  [1] (🛜)       wireless
    Interval{ 0x1F6DD, 0x1F6DF }, // E14.0  [3] (🛝..🛟)    playground slide..ring buoy
    Interval{ 0x1F6E0, 0x1F6E5 }, // E0.7   [6] (🛠️..🛥️)    hammer and wrench..motor boat
    Interval{ 0x1F6E9, 0x1F6E9 }, // E0.7   [1] (🛩️)       small airplane
    Interval{ 0x1F6EB, 0x1F6EC }, // E1.0   [2] (🛫..🛬)    airplane departure..airplane arrival
    Interval{ 0x1F6ED, 0x1F6EF }, // E0.0   [3] (🛭..🛯)    <reserved-1F6ED>..<reserved-1F6EF>
    Interval{ 0x1F6F0, 0x1F6F0 }, // E0.7   [1] (🛰️)       satellite
    Interval{ 0x1F6F3, 0x1F6F3 }, // E0.7   [1] (🛳️)       passenger ship
    Interval{ 0x1F6F4, 0x1F6F6 }, // E3.0   [3] (🛴..🛶)    kick scooter..canoe
    Interval{ 0x1F6F7, 0x1F6F8 }, // E5.0   [2] (🛷..🛸)    sled..flying saucer
    Interval{ 0x1F6F9, 0x1F6F9 }, // E11.0  [1] (🛹)       skateboard
    Interval{ 0x1F6FA, 0x1F6FA }, // E12.0  [1] (🛺)       auto rickshaw
    Interval{ 0x1F6FB, 0x1F6FC }, // E13.0  [2] (🛻..🛼)    pickup truck..roller skate
    Interval{ 0x1F6FD, 0x1F6FF }, // E0.0   [3] (🛽..🛿)    <reserved-1F6FD>..<reserved-1F6FF>
    Interval{ 0x1F7DA, 0x1F7DF }, // E0.0   [6] (🟚..🟟)    <reserved-1F7DA>..<reserved-1F7DF>
    Interval{ 0x1F7E0, 0x1F7EB }, // E12.0 [12] (🟠..🟫)    orange circle..brown square
    Interval{ 0x1F7EC, 0x1F7EF }, // E0.0   [4] (🟬..🟯)    <reserved-1F7EC>..<reserved-1F7EF>
    Interval{ 0x1F7F0, 0x1F7F0 }, // E14.0  [1] (🟰)       heavy equals sign
    Interval{ 0x1F7F1, 0x1F7FF }, // E0.0  [15] (🟱..🟿)    <reserved-1F7F1>..<reserved-1F7FF>
    Interval{ 0x1F80C, 0x1F80F }, // E0.0   [4] (🠌..🠏)    <reserved-1F80C>..<reserved-1F80F>
    Interval{ 0x1F848, 0x1F84F }, // E0.0   [8] (🡈..🡏)    <reserved-1F848>..<reserved-1F84F>
    Interval{ 0x1F85A, 0x1F85F }, // E0.0   [6] (🡚..🡟)    <reserved-1F85A>..<reserved-1F85F>
    Interval{ 0x1F888, 0x1F88F }, // E0.0   [8] (🢈..🢏)    <reserved-1F888>..<reserved-1F88F>
    Interval{ 0x1F8AE, 0x1F8AF }, // E0.0   [2] (🢮..🢯)    <reserved-1F8AE>..<reserved-1F8AF>
    Interval{ 0x1F8BC, 0x1F8BF }, // E0.0   [4] (🢼..🢿)    <reserved-1F8BC>..<reserved-1F8BF>
    Interval{ 0x1F8C2, 0x1F8CF }, // E0.0  [14] (🣂..🣏)    <reserved-1F8C2>..<reserved-1F8CF>
    Interval{ 0x1F8D9, 0x1F8FF }, // E0.0  [39] (🣙..🣿)    <reserved-1F8D9>..<reserved-1F8FF>
    Interval{ 0x1F90C, 0x1F90C }, // E13.0  [1] (🤌)       pinched fingers
    Interval{ 0x1F90D, 0x1F90F }, // E12.0  [3] (🤍..🤏)    white heart..pinching hand
    Interval{ 0x1F910, 0x1F918 }, // E1.0   [9] (🤐..🤘)    zipper-mouth face..sign of the horns
    Interval{ 0x1F919, 0x1F91E }, // E3.0   [6] (🤙..🤞)    call me hand..crossed fingers
    Interval{ 0x1F91F, 0x1F91F }, // E5.0   [1] (🤟)       love-you gesture
    Interval{ 0x1F920, 0x1F927 }, // E3.0   [8] (🤠..🤧)    cowboy hat face..sneezing face
    Interval{ 0x1F928, 0x1F92F }, // E5.0   [8] (🤨..🤯)    face with raised eyebrow..exploding head
    Interval{ 0x1F930, 0x1F930 }, // E3.0   [1] (🤰)       pregnant woman
    Interval{ 0x1F931, 0x1F932 }, // E5.0   [2] (🤱..🤲)    breast-feeding..palms up together
    Interval{ 0x1F933, 0x1F93A }, // E3.0   [8] (🤳..🤺)    selfie..person fencing
    Interval{ 0x1F93C, 0x1F93E }, // E3.0   [3] (🤼..🤾)    people wrestling..person playing handball
    Interval{ 0x1F93F, 0x1F93F }, // E12.0  [1] (🤿)       diving mask
    Interval{ 0x1F940, 0x1F945 }, // E3.0   [6] (🥀..🥅)    wilted flower..goal net
    Interval{ 0x1F947, 0x1F94B }, // E3.0   [5] (🥇..🥋)    1st place medal..martial arts uniform
    Interval{ 0x1F94C, 0x1F94C }, // E5.0   [1] (🥌)       curling stone
    Interval{ 0x1F94D, 0x1F94F }, // E11.0  [3] (🥍..🥏)    lacrosse..flying disc
    Interval{ 0x1F950, 0x1F95E }, // E3.0  [15] (🥐..🥞)    croissant..pancakes
    Interval{ 0x1F95F, 0x1F96B }, // E5.0  [13] (🥟..🥫)    dumpling..canned food
    Interval{ 0x1F96C, 0x1F970 }, // E11.0  [5] (🥬..🥰)    leafy green..smiling face with hearts
    Interval{ 0x1F971, 0x1F971 }, // E12.0  [1] (🥱)       yawning face
    Interval{ 0x1F972, 0x1F972 }, // E13.0  [1] (🥲)       smiling face with tear
    Interval{ 0x1F973, 0x1F976 }, // E11.0  [4] (🥳..🥶)    partying face..cold face
    Interval{ 0x1F977, 0x1F978 }, // E13.0  [2] (🥷..🥸)    ninja..disguised face
    Interval{ 0x1F979, 0x1F979 }, // E14.0  [1] (🥹)       face holding back tears
    Interval{ 0x1F97A, 0x1F97A }, // E11.0  [1] (🥺)       pleading face
    Interval{ 0x1F97B, 0x1F97B }, // E12.0  [1] (🥻)       sari
    Interval{ 0x1F97C, 0x1F97F }, // E11.0  [4] (🥼..🥿)    lab coat..flat shoe
    Interval{ 0x1F980, 0x1F984 }, // E1.0   [5] (🦀..🦄)    crab..unicorn
    Interval{ 0x1F985, 0x1F991 }, // E3.0  [13] (🦅..🦑)    eagle..squid
    Interval{ 0x1F992, 0x1F997 }, // E5.0   [6] (🦒..🦗)    giraffe..cricket
    Interval{ 0x1F998, 0x1F9A2 }, // E11.0 [11] (🦘..🦢)    kangaroo..swan
    Interval{ 0x1F9A3, 0x1F9A4 }, // E13.0  [2] (🦣..🦤)    mammoth..dodo
    Interval{ 0x1F9A5, 0x1F9AA }, // E12.0  [6] (🦥..🦪)    sloth..oyster
    Interval{ 0x1F9AB, 0x1F9AD }, // E13.0  [3] (🦫..🦭)    beaver..seal
    Interval{ 0x1F9AE, 0x1F9AF }, // E12.0  [2] (🦮..🦯)    guide dog..white cane
    Interval{ 0x1F9B0, 0x1F9B9 }, // E11.0 [10] (🦰..🦹)    red hair..supervillain
    Interval{ 0x1F9BA, 0x1F9BF }, // E12.0  [6] (🦺..🦿)    safety vest..mechanical leg
    Interval{ 0x1F9C0, 0x1F9C0 }, // E1.0   [1] (🧀)       cheese wedge
    Interval{ 0x1F9C1, 0x1F9C2 }, // E11.0  [2] (🧁..🧂)    cupcake..salt
    Interval{ 0x1F9C3, 0x1F9CA }, // E12.0  [8] (🧃..🧊)    beverage box..ice
    Interval{ 0x1F9CB, 0x1F9CB }, // E13.0  [1] (🧋)       bubble tea
    Interval{ 0x1F9CC, 0x1F9CC }, // E14.0  [1] (🧌)       troll
    Interval{ 0x1F9CD, 0x1F9CF }, // E12.0  [3] (🧍..🧏)    person standing..deaf person
    Interval{ 0x1F9D0, 0x1F9E6 }, // E5.0  [23] (🧐..🧦)    face with monocle..socks
    Interval{ 0x1F9E7, 0x1F9FF }, // E11.0 [25] (🧧..🧿)    red envelope..nazar amulet
    Interval{ 0x1FA58, 0x1FA5F }, // E0.0   [8] (🩘..🩟)    <reserved-1FA58>..<reserved-1FA5F>
    Interval{ 0x1FA6E, 0x1FA6F }, // E0.0   [2] (🩮..🩯)    <reserved-1FA6E>..<reserved-1FA6F>
    Interval{ 0x1FA70, 0x1FA73 }, // E12.0  [4] (🩰..🩳)    ballet shoes..shorts
    Interval{ 0x1FA74, 0x1FA74 }, // E13.0  [1] (🩴)       thong sandal
    Interval{ 0x1FA75, 0x1FA77 }, // E15.0  [3] (🩵..🩷)    light blue heart..pink heart
    Interval{ 0x1FA78, 0x1FA7A }, // E12.0  [3] (🩸..🩺)    drop of blood..stethoscope
    Interval{ 0x1FA7B, 0x1FA7C }, // E14.0  [2] (🩻..🩼)    x-ray..crutch
    Interval{ 0x1FA7D, 0x1FA7F }, // E0.0   [3] (🩽..🩿)    <reserved-1FA7D>..<reserved-1FA7F>
    Interval{ 0x1FA80, 0x1FA82 }, // E12.0  [3] (🪀..🪂)    yo-yo..parachute
    Interval{ 0x1FA83, 0x1FA86 }, // E13.0  [4] (🪃..🪆)    boomerang..nesting dolls
    Interval{ 0x1FA87, 0x1FA88 }, // E15.0  [2] (🪇..🪈)    maracas..flute
    Interval{ 0x1FA89, 0x1FA89 }, // E16.0  [1] (🪉)       harp
    Interval{ 0x1FA8A, 0x1FA8A }, // E17.0  [1] (🪊)       trombone
    Interval{ 0x1FA8B, 0x1FA8D }, // E0.0   [3] (🪋..🪍)    <reserved-1FA8B>..<reserved-1FA8D>
    Interval{ 0x1FA8E, 0x1FA8E }, // E17.0  [1] (🪎)       treasure chest
    Interval{ 0x1FA8F, 0x1FA8F }, // E16.0  [1] (🪏)       shovel
    Interval{ 0x1FA90, 0x1FA95 }, // E12.0  [6] (🪐..🪕)    ringed planet..banjo
    Interval{ 0x1FA96, 0x1FAA8 }, // E13.0 [19] (🪖..🪨)    military helmet..rock
    Interval{ 0x1FAA9, 0x1FAAC }, // E14.0  [4] (🪩..🪬)    mirror ball..hamsa
    Interval{ 0x1FAAD, 0x1FAAF }, // E15.0  [3] (🪭..🪯)    folding hand fan..khanda
    Interval{ 0x1FAB0, 0x1FAB6 }, // E13.0  [7] (🪰..🪶)    fly..feather
    Interval{ 0x1FAB7, 0x1FABA }, // E14.0  [4] (🪷..🪺)    lotus..nest with eggs
    Interval{ 0x1FABB, 0x1FABD }, // E15.0  [3] (🪻..🪽)    hyacinth..wing
    Interval{ 0x1FABE, 0x1FABE }, // E16.0  [1] (🪾)       leafless tree
    Interval{ 0x1FABF, 0x1FABF }, // E15.0  [1] (🪿)       goose
    Interval{ 0x1FAC0, 0x1FAC2 }, // E13.0  [3] (🫀..🫂)    anatomical heart..people hugging
    Interval{ 0x1FAC3, 0x1FAC5 }, // E14.0  [3] (🫃..🫅)    pregnant man..person with crown
    Interval{ 0x1FAC6, 0x1FAC6 }, // E16.0  [1] (🫆)       fingerprint
    Interval{ 0x1FAC7, 0x1FAC7 }, // E0.0   [1] (🫇)       <reserved-1FAC7>
    Interval{ 0x1FAC8, 0x1FAC8 }, // E17.0  [1] (🫈)       hairy creature
    Interval{ 0x1FAC9, 0x1FACC }, // E0.0   [4] (🫉..🫌)    <reserved-1FAC9>..<reserved-1FACC>
    Interval{ 0x1FACD, 0x1FACD }, // E17.0  [1] (🫍)       orca
    Interval{ 0x1FACE, 0x1FACF }, // E15.0  [2] (🫎..🫏)    moose..donkey
    Interval{ 0x1FAD0, 0x1FAD6 }, // E13.0  [7] (🫐..🫖)    blueberries..teapot
    Interval{ 0x1FAD7, 0x1FAD9 }, // E14.0  [3] (🫗..🫙)    pouring liquid..jar
    Interval{ 0x1FADA, 0x1FADB }, // E15.0  [2] (🫚..🫛)    ginger root..pea pod
    Interval{ 0x1FADC, 0x1FADC }, // E16.0  [1] (🫜)       root vegetable
    Interval{ 0x1FADD, 0x1FADE }, // E0.0   [2] (🫝..🫞)    <reserved-1FADD>..<reserved-1FADE>
    Interval{ 0x1FADF, 0x1FADF }, // E16.0  [1] (🫟)       splatter
    Interval{ 0x1FAE0, 0x1FAE7 }, // E14.0  [8] (🫠..🫧)    melting face..bubbles
    Interval{ 0x1FAE8, 0x1FAE8 }, // E15.0  [1] (🫨)       shaking face
    Interval{ 0x1FAE9, 0x1FAE9 }, // E16.0  [1] (🫩)       face with bags under eyes
    Interval{ 0x1FAEA, 0x1FAEA }, // E17.0  [1] (🫪)       distorted face
    Interval{ 0x1FAEB, 0x1FAEE }, // E0.0   [4] (🫫..🫮)    <reserved-1FAEB>..<reserved-1FAEE>
    Interval{ 0x1FAEF, 0x1FAEF }, // E17.0  [1] (🫯)       fight cloud
    Interval{ 0x1FAF0, 0x1FAF6 }, // E14.0  [7] (🫰..🫶)    hand with index finger and thumb crossed..heart hands
    Interval{ 0x1FAF7, 0x1FAF8 }, // E15.0  [2] (🫷..🫸)    leftwards pushing hand..rightwards pushing hand
    Interval{ 0x1FAF9, 0x1FAFF }, // E0.0   [7] (🫹..🫿)    <reserved-1FAF9>..<reserved-1FAFF>
    Interval{ 0x1FC00, 0x1FFFD }, // E0.0[1022] (🰀..🿽)    <reserved-1FC00>..<reserved-1FFFD>
}; // }}}
} // end namespace tables

bool is_emoji(char32_t codepoint) noexcept {
    return contains(tables::Emoji, codepoint);
}

bool is_emoji_component(char32_t codepoint) noexcept {
    return contains(tables::Emoji_Component, codepoint);
}

bool is_emoji_modifier(char32_t codepoint) noexcept {
    return contains(tables::Emoji_Modifier, codepoint);
}

bool is_emoji_modifier_base(char32_t codepoint) noexcept {
    return contains(tables::Emoji_Modifier_Base, codepoint);
}

bool is_emoji_presentation(char32_t codepoint) noexcept {
    return contains(tables::Emoji_Presentation, codepoint);
}

bool is_extended_pictographic(char32_t codepoint) noexcept {
    return contains(tables::Extended_Pictographic, codepoint);
}

namespace tables {
auto static const Bidi_Mirrored = std::array<Interval, 114>{ // {{{
    Interval{ 0x0028, 0x0029 },
    Interval{ 0x003C, 0x003C },
    Interval{ 0x003E, 0x003E },
    Interval{ 0x005B, 0x005B },
    Interval{ 0x005D, 0x005D },
    Interval{ 0x007B, 0x007B },
    Interval{ 0x007D, 0x007D },
    Interval{ 0x00AB, 0x00AB },
    Interval{ 0x00BB, 0x00BB },
    Interval{ 0x0F3A, 0x0F3D },
    Interval{ 0x169B, 0x169C },
    Interval{ 0x2039, 0x203A },
    Interval{ 0x2045, 0x2046 },
    Interval{ 0x207D, 0x207E },
    Interval{ 0x208D, 0x208E },
    Interval{ 0x2140, 0x2140 },
    Interval{ 0x2201, 0x2204 },
    Interval{ 0x2208, 0x220D },
    Interval{ 0x2211, 0x2211 },
    Interval{ 0x2215, 0x2216 },
    Interval{ 0x221A, 0x221D },
    Interval{ 0x221F, 0x2222 },
    Interval{ 0x2224, 0x2224 },
    Interval{ 0x2226, 0x2226 },
    Interval{ 0x222B, 0x2233 },
    Interval{ 0x2239, 0x2239 },
    Interval{ 0x223B, 0x224C },
    Interval{ 0x2252, 0x2255 },
    Interval{ 0x225F, 0x2260 },
    Interval{ 0x2262, 0x2262 },
    Interval{ 0x2264, 0x226B },
    Interval{ 0x226D, 0x228C },
    Interval{ 0x228F, 0x2292 },
    Interval{ 0x2298, 0x2298 },
    Interval{ 0x22A2, 0x22A3 },
    Interval{ 0x22A6, 0x22B8 },
    Interval{ 0x22BE, 0x22BF },
    Interval{ 0x22C9, 0x22CD },
    Interval{ 0x22D0, 0x22D1 },
    Interval{ 0x22D6, 0x22ED },
    Interval{ 0x22F0, 0x22FF },
    Interval{ 0x2308, 0x230B },
    Interval{ 0x2320, 0x2321 },
    Interval{ 0x2329, 0x232A },
    Interval{ 0x2768, 0x2775 },
    Interval{ 0x27C0, 0x27C0 },
    Interval{ 0x27C3, 0x27C6 },
    Interval{ 0x27C8, 0x27C9 },
    Interval{ 0x27CB, 0x27CD },
    Interval{ 0x27D3, 0x27D6 },
    Interval{ 0x27DC, 0x27DE },
    Interval{ 0x27E2, 0x27EF },
    Interval{ 0x2983, 0x2998 },
    Interval{ 0x299B, 0x29A0 },
    Interval{ 0x29A2, 0x29AF },
    Interval{ 0x29B8, 0x29B8 },
    Interval{ 0x29C0, 0x29C5 },
    Interval{ 0x29C9, 0x29C9 },
    Interval{ 0x29CE, 0x29D2 },
    Interval{ 0x29D4, 0x29D5 },
    Interval{ 0x29D8, 0x29DC },
    Interval{ 0x29E1, 0x29E1 },
    Interval{ 0x29E3, 0x29E5 },
    Interval{ 0x29E8, 0x29E9 },
    Interval{ 0x29F4, 0x29F9 },
    Interval{ 0x29FC, 0x29FD },
    Interval{ 0x2A0A, 0x2A1C },
    Interval{ 0x2A1E, 0x2A21 },
    Interval{ 0x2A24, 0x2A24 },
    Interval{ 0x2A26, 0x2A26 },
    Interval{ 0x2A29, 0x2A29 },
    Interval{ 0x2A2B, 0x2A2E },
    Interval{ 0x2A34, 0x2A35 },
    Interval{ 0x2A3C, 0x2A3E },
    Interval{ 0x2A57, 0x2A58 },
    Interval{ 0x2A64, 0x2A65 },
    Interval{ 0x2A6A, 0x2A6D },
    Interval{ 0x2A6F, 0x2A70 },
    Interval{ 0x2A73, 0x2A74 },
    Interval{ 0x2A79, 0x2AA3 },
    Interval{ 0x2AA6, 0x2AAD },
    Interval{ 0x2AAF, 0x2AD6 },
    Interval{ 0x2ADC, 0x2ADC },
    Interval{ 0x2ADE, 0x2ADE },
    Interval{ 0x2AE2, 0x2AE6 },
    Interval{ 0x2AEC, 0x2AEE },
    Interval{ 0x2AF3, 0x2AF3 },
    Interval{ 0x2AF7, 0x2AFB },
    Interval{ 0x2AFD, 0x2AFD },
    Interval{ 0x2BFE, 0x2BFE },
    Interval{ 0x2E02, 0x2E05 },
    Interval{ 0x2E09, 0x2E0A },
    Interval{ 0x2E0C, 0x2E0D },
    Interval{ 0x2E1C, 0x2E1D },
    Interval{ 0x2E20, 0x2E29 },
    Interval{ 0x2E55, 0x2E5C },
    Interval{ 0x3008, 0x3011 },
    Interval{ 0x3014, 0x301B },
    Interval{ 0xFE59, 0xFE5E },
    Interval{ 0xFE64, 0xFE65 },
    Interval{ 0xFF08, 0xFF09 },
    Interval{ 0xFF1C, 0xFF1C },
    Interval{ 0xFF1E, 0xFF1E },
    Interval{ 0xFF3B, 0xFF3B },
    Interval{ 0xFF3D, 0xFF3D },
    Interval{ 0xFF5B, 0xFF5B },
    Interval{ 0xFF5D, 0xFF5D },
    Interval{ 0xFF5F, 0xFF60 },
    Interval{ 0xFF62, 0xFF63 },
    Interval{ 0x1D6DB, 0x1D6DB },
    Interval{ 0x1D715, 0x1D715 },
    Interval{ 0x1D74F, 0x1D74F },
    Interval{ 0x1D789, 0x1D789 },
    Interval{ 0x1D7C3, 0x1D7C3 },
}; // }}}
auto static const Bidi_Mirroring_Glyph = std::array<std::pair<char32_t, char32_t>, 428>{ // {{{
    std::pair<char32_t, char32_t>{ 0x0028, 0x0029 },
    std::pair<char32_t, char32_t>{ 0x0029, 0x0028 },
    std::pair<char32_t, char32_t>{ 0x003C, 0x003E },
    std::pair<char32_t, char32_t>{ 0x003E, 0x003C },
    std::pair<char32_t, char32_t>{ 0x005B, 0x005D },
    std::pair<char32_t, char32_t>{ 0x005D, 0x005B },
    std::pair<char32_t, char32_t>{ 0x007B, 0x007D },
    std::pair<char32_t, char32_t>{ 0x007D, 0x007B },
    std::pair<char32_t, char32_t>{ 0x00AB, 0x00BB },
    std::pair<char32_t, char32_t>{ 0x00BB, 0x00AB },
    std::pair<char32_t, char32_t>{ 0x0F3A, 0x0F3B },
    std::pair<char32_t, char32_t>{ 0x0F3B, 0x0F3A },
    std::pair<char32_t, char32_t>{ 0x0F3C, 0x0F3D },
    std::pair<char32_t, char32_t>{ 0x0F3D, 0x0F3C },
    std::pair<char32_t, char32_t>{ 0x169B, 0x169C },
    std::pair<char32_t, char32_t>{ 0x169C, 0x169B },
    std::pair<char32_t, char32_t>{ 0x2039, 0x203A },
    std::pair<char32_t, char32_t>{ 0x203A, 0x2039 },
    std::pair<char32_t, char32_t>{ 0x2045, 0x2046 },
    std::pair<char32_t, char32_t>{ 0x2046, 0x2045 },
    std::pair<char32_t, char32_t>{ 0x207D, 0x207E },
    std::pair<char32_t, char32_t>{ 0x207E, 0x207D },
    std::pair<char32_t, char32_t>{ 0x208D, 0x208E },
    std::pair<char32_t, char32_t>{ 0x208E, 0x208D },
    std::pair<char32_t, char32_t>{ 0x2208, 0x220B },
    std::pair<char32_t, char32_t>{ 0x2209, 0x220C },
    std::pair<char32_t, char32_t>{ 0x220A, 0x220D },
    std::pair<char32_t, char32_t>{ 0x220B, 0x2208 },
    std::pair<char32_t, char32_t>{ 0x220C, 0x2209 },
    std::pair<char32_t, char32_t>{ 0x220D, 0x220A },
    std::pair<char32_t, char32_t>{ 0x2215, 0x29F5 },
    std::pair<char32_t, char32_t>{ 0x221F, 0x2BFE },
    std::pair<char32_t, char32_t>{ 0x2220, 0x29A3 },
    std::pair<char32_t, char32_t>{ 0x2221, 0x299B },
    std::pair<char32_t, char32_t>{ 0x2222, 0x29A0 },
    std::pair<char32_t, char32_t>{ 0x2224, 0x2AEE },
    std::pair<char32_t, char32_t>{ 0x223C, 0x223D },
    std::pair<char32_t, char32_t>{ 0x223D, 0x223C },
    std::pair<char32_t, char32_t>{ 0x2243, 0x22CD },
    std::pair<char32_t, char32_t>{ 0x2245, 0x224C },
    std::pair<char32_t, char32_t>{ 0x224C, 0x2245 },
    std::pair<char32_t, char32_t>{ 0x2252, 0x2253 },
    std::pair<char32_t, char32_t>{ 0x2253, 0x2252 },
    std::pair<char32_t, char32_t>{ 0x2254, 0x2255 },
    std::pair<char32_t, char32_t>{ 0x2255, 0x2254 },
    std::pair<char32_t, char32_t>{ 0x2264, 0x2265 },
    std::pair<char32_t, char32_t>{ 0x2265, 0x2264 },
    std::pair<char32_t, char32_t>{ 0x2266, 0x2267 },
    std::pair<char32_t, char32_t>{ 0x2267, 0x2266 },
    std::pair<char32_t, char32_t>{ 0x2268, 0x2269 },
    std::pair<char32_t, char32_t>{ 0x2269, 0x2268 },
    std::pair<char32_t, char32_t>{ 0x226A, 0x226B },
    std::pair<char32_t, char32_t>{ 0x226B, 0x226A },
    std::pair<char32_t, char32_t>{ 0x226E, 0x226F },
    std::pair<char32_t, char32_t>{ 0x226F, 0x226E },
    std::pair<char32_t, char32_t>{ 0x2270, 0x2271 },
    std::pair<char32_t, char32_t>{ 0x2271, 0x2270 },
    std::pair<char32_t, char32_t>{ 0x2272, 0x2273 },
    std::pair<char32_t, char32_t>{ 0x2273, 0x2272 },
    std::pair<char32_t, char32_t>{ 0x2274, 0x2275 },
    std::pair<char32_t, char32_t>{ 0x2275, 0x2274 },
    std::pair<char32_t, char32_t>{ 0x2276, 0x2277 },
    std::pair<char32_t, char32_t>{ 0x2277, 0x2276 },
    std::pair<char32_t, char32_t>{ 0x2278, 0x2279 },
    std::pair<char32_t, char32_t>{ 0x2279, 0x2278 },
    std::pair<char32_t, char32_t>{ 0x227A, 0x227B },
    std::pair<char32_t, char32_t>{ 0x227B, 0x227A },
    std::pair<char32_t, char32_t>{ 0x227C, 0x227D },
    std::pair<char32_t, char32_t>{ 0x227D, 0x227C },
    std::pair<char32_t, char32_t>{ 0x227E, 0x227F },
    std::pair<char32_t, char32_t>{ 0x227F, 0x227E },
    std::pair<char32_t, char32_t>{ 0x2280, 0x2281 },
    std::pair<char32_t, char32_t>{ 0x2281, 0x2280 },
    std::pair<char32_t, char32_t>{ 0x2282, 0x2283 },
    std::pair<char32_t, char32_t>{ 0x2283, 0x2282 },
    std::pair<char32_t, char32_t>{ 0x2284, 0x2285 },
    std::pair<char32_t, char32_t>{ 0x2285, 0x2284 },
    std::pair<char32_t, char32_t>{ 0x2286, 0x2287 },
    std::pair<char32_t, char32_t>{ 0x2287, 0x2286 },
    std::pair<char32_t, char32_t>{ 0x2288, 0x2289 },
    std::pair<char32_t, char32_t>{ 0x2289, 0x2288 },
    std::pair<char32_t, char32_t>{ 0x228A, 0x228B },
    std::pair<char32_t, char32_t>{ 0x228B, 0x228A },
    std::pair<char32_t, char32_t>{ 0x228F, 0x2290 },
    std::pair<char32_t, char32_t>{ 0x2290, 0x228F },
    std::pair<char32_t, char32_t>{ 0x2291, 0x2292 },
    std::pair<char32_t, char32_t>{ 0x2292, 0x2291 },
    std::pair<char32_t, char32_t>{ 0x2298, 0x29B8 },
    std::pair<char32_t, char32_t>{ 0x22A2, 0x22A3 },
    std::pair<char32_t, char32_t>{ 0x22A3, 0x22A2 },
    std::pair<char32_t, char32_t>{ 0x22A6, 0x2ADE },
    std::pair<char32_t, char32_t>{ 0x22A8, 0x2AE4 },
    std::pair<char32_t, char32_t>{ 0x22A9, 0x2AE3 },
    std::pair<char32_t, char32_t>{ 0x22AB, 0x2AE5 },
    std::pair<char32_t, char32_t>{ 0x22B0, 0x22B1 },
    std::pair<char32_t, char32_t>{ 0x22B1, 0x22B0 },
    std::pair<char32_t, char32_t>{ 0x22B2, 0x22B3 },
    std::pair<char32_t, char32_t>{ 0x22B3, 0x22B2 },
    std::pair<char32_t, char32_t>{ 0x22B4, 0x22B5 },
    std::pair<char32_t, char32_t>{ 0x22B5, 0x22B4 },
    std::pair<char32_t, char32_t>{ 0x22B6, 0x22B7 },
    std::pair<char32_t, char32_t>{ 0x22B7, 0x22B6 },
    std::pair<char32_t, char32_t>{ 0x22B8, 0x27DC },
    std::pair<char32_t, char32_t>{ 0x22C9, 0x22CA },
    std::pair<char32_t, char32_t>{ 0x22CA, 0x22C9 },
    std::pair<char32_t, char32_t>{ 0x22CB, 0x22CC },
    std::pair<char32_t, char32_t>{ 0x22CC, 0x22CB },
    std::pair<char32_t, char32_t>{ 0x22CD, 0x2243 },
    std::pair<char32_t, char32_t>{ 0x22D0, 0x22D1 },
    std::pair<char32_t, char32_t>{ 0x22D1, 0x22D0 },
    std::pair<char32_t, char32_t>{ 0x22D6, 0x22D7 },
    std::pair<char32_t, char32_t>{ 0x22D7, 0x22D6 },
    std::pair<char32_t, char32_t>{ 0x22D8, 0x22D9 },
    std::pair<char32_t, char32_t>{ 0x22D9, 0x22D8 },
    std::pair<char32_t, char32_t>{ 0x22DA, 0x22DB },
    std::pair<char32_t, char32_t>{ 0x22DB, 0x22DA },
    std::pair<char32_t, char32_t>{ 0x22DC, 0x22DD },
    std::pair<char32_t, char32_t>{ 0x22DD, 0x22DC },
    std::pair<char32_t, char32_t>{ 0x22DE, 0x22DF },
    std::pair<char32_t, char32_t>{ 0x22DF, 0x22DE },
    std::pair<char32_t, char32_t>{ 0x22E0, 0x22E1 },
    std::pair<char32_t, char32_t>{ 0x22E1, 0x22E0 },
    std::pair<char32_t, char32_t>{ 0x22E2, 0x22E3 },
    std::pair<char32_t, char32_t>{ 0x22E3, 0x22E2 },
    std::pair<char32_t, char32_t>{ 0x22E4, 0x22E5 },
    std::pair<char32_t, char32_t>{ 0x22E5, 0x22E4 },
    std::pair<char32_t, char32_t>{ 0x22E6, 0x22E7 },
    std::pair<char32_t, char32_t>{ 0x22E7, 0x22E6 },
    std::pair<char32_t, char32_t>{ 0x22E8, 0x22E9 },
    std::pair<char32_t, char32_t>{ 0x22E9, 0x22E8 },
    std::pair<char32_t, char32_t>{ 0x22EA, 0x22EB },
    std::pair<char32_t, char32_t>{ 0x22EB, 0x22EA },
    std::pair<char32_t, char32_t>{ 0x22EC, 0x22ED },
    std::pair<char32_t, char32_t>{ 0x22ED, 0x22EC },
    std::pair<char32_t, char32_t>{ 0x22F0, 0x22F1 },
    std::pair<char32_t, char32_t>{ 0x22F1, 0x22F0 },
    std::pair<char32_t, char32_t>{ 0x22F2, 0x22FA },
    std::pair<char32_t, char32_t>{ 0x22F3, 0x22FB },
    std::pair<char32_t, char32_t>{ 0x22F4, 0x22FC },
    std::pair<char32_t, char32_t>{ 0x22F6, 0x22FD },
    std::pair<char32_t, char32_t>{ 0x22F7, 0x22FE },
    std::pair<char32_t, char32_t>{ 0x22FA, 0x22F2 },
    std::pair<char32_t, char32_t>{ 0x22FB, 0x22F3 },
    std::pair<char32_t, char32_t>{ 0x22FC, 0x22F4 },
    std::pair<char32_t, char32_t>{ 0x22FD, 0x22F6 },
    std::pair<char32_t, char32_t>{ 0x22FE, 0x22F7 },
    std::pair<char32_t, char32_t>{ 0x2308, 0x2309 },
    std::pair<char32_t, char32_t>{ 0x2309, 0x2308 },
    std::pair<char32_t, char32_t>{ 0x230A, 0x230B },
    std::pair<char32_t, char32_t>{ 0x230B, 0x230A },
    std::pair<char32_t, char32_t>{ 0x2329, 0x232A },
    std::pair<char32_t, char32_t>{ 0x232A, 0x2329 },
    std::pair<char32_t, char32_t>{ 0x2768, 0x2769 },
    std::pair<char32_t, char32_t>{ 0x2769, 0x2768 },
    std::pair<char32_t, char32_t>{ 0x276A, 0x276B },
    std::pair<char32_t, char32_t>{ 0x276B, 0x276A },
    std::pair<char32_t, char32_t>{ 0x276C, 0x276D },
    std::pair<char32_t, char32_t>{ 0x276D, 0x276C },
    std::pair<char32_t, char32_t>{ 0x276E, 0x276F },
    std::pair<char32_t, char32_t>{ 0x276F, 0x276E },
    std::pair<char32_t, char32_t>{ 0x2770, 0x2771 },
    std::pair<char32_t, char32_t>{ 0x2771, 0x2770 },
    std::pair<char32_t, char32_t>{ 0x2772, 0x2773 },
    std::pair<char32_t, char32_t>{ 0x2773, 0x2772 },
    std::pair<char32_t, char32_t>{ 0x2774, 0x2775 },
    std::pair<char32_t, char32_t>{ 0x2775, 0x2774 },
    std::pair<char32_t, char32_t>{ 0x27C3, 0x27C4 },
    std::pair<char32_t, char32_t>{ 0x27C4, 0x27C3 },
    std::pair<char32_t, char32_t>{ 0x27C5, 0x27C6 },
    std::pair<char32_t, char32_t>{ 0x27C6, 0x27C5 },
    std::pair<char32_t, char32_t>{ 0x27C8, 0x27C9 },
    std::pair<char32_t, char32_t>{ 0x27C9, 0x27C8 },
    std::pair<char32_t, char32_t>{ 0x27CB, 0x27CD },
    std::pair<char32_t, char32_t>{ 0x27CD, 0x27CB },
    std::pair<char32_t, char32_t>{ 0x27D5, 0x27D6 },
    std::pair<char32_t, char32_t>{ 0x27D6, 0x27D5 },
    std::pair<char32_t, char32_t>{ 0x27DC, 0x22B8 },
    std::pair<char32_t, char32_t>{ 0x27DD, 0x27DE },
    std::pair<char32_t, char32_t>{ 0x27DE, 0x27DD },
    std::pair<char32_t, char32_t>{ 0x27E2, 0x27E3 },
    std::pair<char32_t, char32_t>{ 0x27E3, 0x27E2 },
    std::pair<char32_t, char32_t>{ 0x27E4, 0x27E5 },
    std::pair<char32_t, char32_t>{ 0x27E5, 0x27E4 },
    std::pair<char32_t, char32_t>{ 0x27E6, 0x27E7 },
    std::pair<char32_t, char32_t>{ 0x27E7, 0x27E6 },
    std::pair<char32_t, char32_t>{ 0x27E8, 0x27E9 },
    std::pair<char32_t, char32_t>{ 0x27E9, 0x27E8 },
    std::pair<char32_t, char32_t>{ 0x27EA, 0x27EB },
    std::pair<char32_t, char32_t>{ 0x27EB, 0x27EA },
    std::pair<char32_t, char32_t>{ 0x27EC, 0x27ED },
    std::pair<char32_t, char32_t>{ 0x27ED, 0x27EC },
    std::pair<char32_t, char32_t>{ 0x27EE, 0x27EF },
    std::pair<char32_t, char32_t>{ 0x27EF, 0x27EE },
    std::pair<char32_t, char32_t>{ 0x2983, 0x2984 },
    std::pair<char32_t, char32_t>{ 0x2984, 0x2983 },
    std::pair<char32_t, char32_t>{ 0x2985, 0x2986 },
    std::pair<char32_t, char32_t>{ 0x2986, 0x2985 },
    std::pair<char32_t, char32_t>{ 0x2987, 0x2988 },
    std::pair<char32_t, char32_t>{ 0x2988, 0x2987 },
    std::pair<char32_t, char32_t>{ 0x2989, 0x298A },
    std::pair<char32_t, char32_t>{ 0x298A, 0x2989 },
    std::pair<char32_t, char32_t>{ 0x298B, 0x298C },
    std::pair<char32_t, char32_t>{ 0x298C, 0x298B },
    std::pair<char32_t, char32_t>{ 0x298D, 0x2990 },
    std::pair<char32_t, char32_t>{ 0x298E, 0x298F },
    std::pair<char32_t, char32_t>{ 0x298F, 0x298E },
    std::pair<char32_t, char32_t>{ 0x2990, 0x298D },
    std::pair<char32_t, char32_t>{ 0x2991, 0x2992 },
    std::pair<char32_t, char32_t>{ 0x2992, 0x2991 },
    std::pair<char32_t, char32_t>{ 0x2993, 0x2994 },
    std::pair<char32_t, char32_t>{ 0x2994, 0x2993 },
    std::pair<char32_t, char32_t>{ 0x2995, 0x2996 },
    std::pair<char32_t, char32_t>{ 0x2996, 0x2995 },
    std::pair<char32_t, char32_t>{ 0x2997, 0x2998 },
    std::pair<char32_t, char32_t>{ 0x2998, 0x2997 },
    std::pair<char32_t, char32_t>{ 0x299B, 0x2221 },
    std::pair<char32_t, char32_t>{ 0x29A0, 0x2222 },
    std::pair<char32_t, char32_t>{ 0x29A3, 0x2220 },
    std::pair<char32_t, char32_t>{ 0x29A4, 0x29A5 },
    std::pair<char32_t, char32_t>{ 0x29A5, 0x29A4 },
    std::pair<char32_t, char32_t>{ 0x29A8, 0x29A9 },
    std::pair<char32_t, char32_t>{ 0x29A9, 0x29A8 },
    std::pair<char32_t, char32_t>{ 0x29AA, 0x29AB },
    std::pair<char32_t, char32_t>{ 0x29AB, 0x29AA },
    std::pair<char32_t, char32_t>{ 0x29AC, 0x29AD },
    std::pair<char32_t, char32_t>{ 0x29AD, 0x29AC },
    std::pair<char32_t, char32_t>{ 0x29AE, 0x29AF },
    std::pair<char32_t, char32_t>{ 0x29AF, 0x29AE },
    std::pair<char32_t, char32_t>{ 0x29B8, 0x2298 },
    std::pair<char32_t, char32_t>{ 0x29C0, 0x29C1 },
    std::pair<char32_t, char32_t>{ 0x29C1, 0x29C0 },
    std::pair<char32_t, char32_t>{ 0x29C4, 0x29C5 },
    std::pair<char32_t, char32_t>{ 0x29C5, 0x29C4 },
    std::pair<char32_t, char32_t>{ 0x29CF, 0x29D0 },
    std::pair<char32_t, char32_t>{ 0x29D0, 0x29CF },
    std::pair<char32_t, char32_t>{ 0x29D1, 0x29D2 },
    std::pair<char32_t, char32_t>{ 0x29D2, 0x29D1 },
    std::pair<char32_t, char32_t>{ 0x29D4, 0x29D5 },
    std::pair<char32_t, char32_t>{ 0x29D5, 0x29D4 },
    std::pair<char32_t, char32_t>{ 0x29D8, 0x29D9 },
    std::pair<char32_t, char32_t>{ 0x29D9, 0x29D8 },
    std::pair<char32_t, char32_t>{ 0x29DA, 0x29DB },
    std::pair<char32_t, char32_t>{ 0x29DB, 0x29DA },
    std::pair<char32_t, char32_t>{ 0x29E8, 0x29E9 },
    std::pair<char32_t, char32_t>{ 0x29E9, 0x29E8 },
    std::pair<char32_t, char32_t>{ 0x29F5, 0x2215 },
    std::pair<char32_t, char32_t>{ 0x29F8, 0x29F9 },
    std::pair<char32_t, char32_t>{ 0x29F9, 0x29F8 },
    std::pair<char32_t, char32_t>{ 0x29FC, 0x29FD },
    std::pair<char32_t, char32_t>{ 0x29FD, 0x29FC },
    std::pair<char32_t, char32_t>{ 0x2A2B, 0x2A2C },
    std::pair<char32_t, char32_t>{ 0x2A2C, 0x2A2B },
    std::pair<char32_t, char32_t>{ 0x2A2D, 0x2A2E },
    std::pair<char32_t, char32_t>{ 0x2A2E, 0x2A2D },
    std::pair<char32_t, char32_t>{ 0x2A34, 0x2A35 },
    std::pair<char32_t, char32_t>{ 0x2A35, 0x2A34 },
    std::pair<char32_t, char32_t>{ 0x2A3C, 0x2A3D },
    std::pair<char32_t, char32_t>{ 0x2A3D, 0x2A3C },
    std::pair<char32_t, char32_t>{ 0x2A64, 0x2A65 },
    std::pair<char32_t, char32_t>{ 0x2A65, 0x2A64 },
    std::pair<char32_t, char32_t>{ 0x2A79, 0x2A7A },
    std::pair<char32_t, char32_t>{ 0x2A7A, 0x2A79 },
    std::pair<char32_t, char32_t>{ 0x2A7B, 0x2A7C },
    std::pair<char32_t, char32_t>{ 0x2A7C, 0x2A7B },
    std::pair<char32_t, char32_t>{ 0x2A7D, 0x2A7E },
    std::pair<char32_t, char32_t>{ 0x2A7E, 0x2A7D },
    std::pair<char32_t, char32_t>{ 0x2A7F, 0x2A80 },
    std::pair<char32_t, char32_t>{ 0x2A80, 0x2A7F },
    std::pair<char32_t, char32_t>{ 0x2A81, 0x2A82 },
    std::pair<char32_t, char32_t>{ 0x2A82, 0x2A81 },
    std::pair<char32_t, char32_t>{ 0x2A83, 0x2A84 },
    std::pair<char32_t, char32_t>{ 0x2A84, 0x2A83 },
    std::pair<char32_t, char32_t>{ 0x2A85, 0x2A86 },
    std::pair<char32_t, char32_t>{ 0x2A86, 0x2A85 },
    std::pair<char32_t, char32_t>{ 0x2A87, 0x2A88 },
    std::pair<char32_t, char32_t>{ 0x2A88, 0x2A87 },
    std::pair<char32_t, char32_t>{ 0x2A89, 0x2A8A },
    std::pair<char32_t, char32_t>{ 0x2A8A, 0x2A89 },
    std::pair<char32_t, char32_t>{ 0x2A8B, 0x2A8C },
    std::pair<char32_t, char32_t>{ 0x2A8C, 0x2A8B },
    std::pair<char32_t, char32_t>{ 0x2A8D, 0x2A8E },
    std::pair<char32_t, char32_t>{ 0x2A8E, 0x2A8D },
    std::pair<char32_t, char32_t>{ 0x2A8F, 0x2A90 },
    std::pair<char32_t, char32_t>{ 0x2A90, 0x2A8F },
    std::pair<char32_t, char32_t>{ 0x2A91, 0x2A92 },
    std::pair<char32_t, char32_t>{ 0x2A92, 0x2A91 },
    std::pair<char32_t, char32_t>{ 0x2A93, 0x2A94 },
    std::pair<char32_t, char32_t>{ 0x2A94, 0x2A93 },
    std::pair<char32_t, char32_t>{ 0x2A95, 0x2A96 },
    std::pair<char32_t, char32_t>{ 0x2A96, 0x2A95 },
    std::pair<char32_t, char32_t>{ 0x2A97, 0x2A98 },
    std::pair<char32_t, char32_t>{ 0x2A98, 0x2A97 },
    std::pair<char32_t, char32_t>{ 0x2A99, 0x2A9A },
    std::pair<char32_t, char32_t>{ 0x2A9A, 0x2A99 },
    std::pair<char32_t, char32_t>{ 0x2A9B, 0x2A9C },
    std::pair<char32_t, char32_t>{ 0x2A9C, 0x2A9B },
    std::pair<char32_t, char32_t>{ 0x2A9D, 0x2A9E },
    std::pair<char32_t, char32_t>{ 0x2A9E, 0x2A9D },
    std::pair<char32_t, char32_t>{ 0x2A9F, 0x2AA0 },
    std::pair<char32_t, char32_t>{ 0x2AA0, 0x2A9F },
    std::pair<char32_t, char32_t>{ 0x2AA1, 0x2AA2 },
    std::pair<char32_t, char32_t>{ 0x2AA2, 0x2AA1 },
    std::pair<char32_t, char32_t>{ 0x2AA6, 0x2AA7 },
    std::pair<char32_t, char32_t>{ 0x2AA7, 0x2AA6 },
    std::pair<char32_t, char32_t>{ 0x2AA8, 0x2AA9 },
    std::pair<char32_t, char32_t>{ 0x2AA9, 0x2AA8 },
    std::pair<char32_t, char32_t>{ 0x2AAA, 0x2AAB },
    std::pair<char32_t, char32_t>{ 0x2AAB, 0x2AAA },
    std::pair<char32_t, char32_t>{ 0x2AAC, 0x2AAD },
    std::pair<char32_t, char32_t>{ 0x2AAD, 0x2AAC },
    std::pair<char32_t, char32_t>{ 0x2AAF, 0x2AB0 },
    std::pair<char32_t, char32_t>{ 0x2AB0, 0x2AAF },
    std::pair<char32_t, char32_t>{ 0x2AB1, 0x2AB2 },
    std::pair<char32_t, char32_t>{ 0x2AB2, 0x2AB1 },
    std::pair<char32_t, char32_t>{ 0x2AB3, 0x2AB4 },
    std::pair<char32_t, char32_t>{ 0x2AB4, 0x2AB3 },
    std::pair<char32_t, char32_t>{ 0x2AB5, 0x2AB6 },
    std::pair<char32_t, char32_t>{ 0x2AB6, 0x2AB5 },
    std::pair<char32_t, char32_t>{ 0x2AB7, 0x2AB8 },
    std::pair<char32_t, char32_t>{ 0x2AB8, 0x2AB7 },
    std::pair<char32_t, char32_t>{ 0x2AB9, 0x2ABA },
    std::pair<char32_t, char32_t>{ 0x2ABA, 0x2AB9 },
    std::pair<char32_t, char32_t>{ 0x2ABB, 0x2ABC },
    std::pair<char32_t, char32_t>{ 0x2ABC, 0x2ABB },
    std::pair<char32_t, char32_t>{ 0x2ABD, 0x2ABE },
    std::pair<char32_t, char32_t>{ 0x2ABE, 0x2ABD },
    std::pair<char32_t, char32_t>{ 0x2ABF, 0x2AC0 },
    std::pair<char32_t, char32_t>{ 0x2AC0, 0x2ABF },
    std::pair<char32_t, char32_t>{ 0x2AC1, 0x2AC2 },
    std::pair<char32_t, char32_t>{ 0x2AC2, 0x2AC1 },
    std::pair<char32_t, char32_t>{ 0x2AC3, 0x2AC4 },
    std::pair<char32_t, char32_t>{ 0x2AC4, 0x2AC3 },
    std::pair<char32_t, char32_t>{ 0x2AC5, 0x2AC6 },
    std::pair<char32_t, char32_t>{ 0x2AC6, 0x2AC5 },
    std::pair<char32_t, char32_t>{ 0x2AC7, 0x2AC8 },
    std::pair<char32_t, char32_t>{ 0x2AC8, 0x2AC7 },
    std::pair<char32_t, char32_t>{ 0x2AC9, 0x2ACA },
    std::pair<char32_t, char32_t>{ 0x2ACA, 0x2AC9 },
    std::pair<char32_t, char32_t>{ 0x2ACB, 0x2ACC },
    std::pair<char32_t, char32_t>{ 0x2ACC, 0x2ACB },
    std::pair<char32_t, char32_t>{ 0x2ACD, 0x2ACE },
    std::pair<char32_t, char32_t>{ 0x2ACE, 0x2ACD },
    std::pair<char32_t, char32_t>{ 0x2ACF, 0x2AD0 },
    std::pair<char32_t, char32_t>{ 0x2AD0, 0x2ACF },
    std::pair<char32_t, char32_t>{ 0x2AD1, 0x2AD2 },
    std::pair<char32_t, char32_t>{ 0x2AD2, 0x2AD1 },
    std::pair<char32_t, char32_t>{ 0x2AD3, 0x2AD4 },
    std::pair<char32_t, char32_t>{ 0x2AD4, 0x2AD3 },
    std::pair<char32_t, char32_t>{ 0x2AD5, 0x2AD6 },
    std::pair<char32_t, char32_t>{ 0x2AD6, 0x2AD5 },
    std::pair<char32_t, char32_t>{ 0x2ADE, 0x22A6 },
    std::pair<char32_t, char32_t>{ 0x2AE3, 0x22A9 },
    std::pair<char32_t, char32_t>{ 0x2AE4, 0x22A8 },
    std::pair<char32_t, char32_t>{ 0x2AE5, 0x22AB },
    std::pair<char32_t, char32_t>{ 0x2AEC, 0x2AED },
    std::pair<char32_t, char32_t>{ 0x2AED, 0x2AEC },
    std::pair<char32_t, char32_t>{ 0x2AEE, 0x2224 },
    std::pair<char32_t, char32_t>{ 0x2AF7, 0x2AF8 },
    std::pair<char32_t, char32_t>{ 0x2AF8, 0x2AF7 },
    std::pair<char32_t, char32_t>{ 0x2AF9, 0x2AFA },
    std::pair<char32_t, char32_t>{ 0x2AFA, 0x2AF9 },
    std::pair<char32_t, char32_t>{ 0x2BFE, 0x221F },
    std::pair<char32_t, char32_t>{ 0x2E02, 0x2E03 },
    std::pair<char32_t, char32_t>{ 0x2E03, 0x2E02 },
    std::pair<char32_t, char32_t>{ 0x2E04, 0x2E05 },
    std::pair<char32_t, char32_t>{ 0x2E05, 0x2E04 },
    std::pair<char32_t, char32_t>{ 0x2E09, 0x2E0A },
    std::pair<char32_t, char32_t>{ 0x2E0A, 0x2E09 },
    std::pair<char32_t, char32_t>{ 0x2E0C, 0x2E0D },
    std::pair<char32_t, char32_t>{ 0x2E0D, 0x2E0C },
    std::pair<char32_t, char32_t>{ 0x2E1C, 0x2E1D },
    std::pair<char32_t, char32_t>{ 0x2E1D, 0x2E1C },
    std::pair<char32_t, char32_t>{ 0x2E20, 0x2E21 },
    std::pair<char32_t, char32_t>{ 0x2E21, 0x2E20 },
    std::pair<char32_t, char32_t>{ 0x2E22, 0x2E23 },
    std::pair<char32_t, char32_t>{ 0x2E23, 0x2E22 },
    std::pair<char32_t, char32_t>{ 0x2E24, 0x2E25 },
    std::pair<char32_t, char32_t>{ 0x2E25, 0x2E24 },
    std::pair<char32_t, char32_t>{ 0x2E26, 0x2E27 },
    std::pair<char32_t, char32_t>{ 0x2E27, 0x2E26 },
    std::pair<char32_t, char32_t>{ 0x2E28, 0x2E29 },
    std::pair<char32_t, char32_t>{ 0x2E29, 0x2E28 },
    std::pair<char32_t, char32_t>{ 0x2E55, 0x2E56 },
    std::pair<char32_t, char32_t>{ 0x2E56, 0x2E55 },
    std::pair<char32_t, char32_t>{ 0x2E57, 0x2E58 },
    std::pair<char32_t, char32_t>{ 0x2E58, 0x2E57 },
    std::pair<char32_t, char32_t>{ 0x2E59, 0x2E5A },
    std::pair<char32_t, char32_t>{ 0x2E5A, 0x2E59 },
    std::pair<char32_t, char32_t>{ 0x2E5B, 0x2E5C },
    std::pair<char32_t, char32_t>{ 0x2E5C, 0x2E5B },
    std::pair<char32_t, char32_t>{ 0x3008, 0x3009 },
    std::pair<char32_t, char32_t>{ 0x3009, 0x3008 },
    std::pair<char32_t, char32_t>{ 0x300A, 0x300B },
    std::pair<char32_t, char32_t>{ 0x300B, 0x300A },
    std::pair<char32_t, char32_t>{ 0x300C, 0x300D },
    std::pair<char32_t, char32_t>{ 0x300D, 0x300C },
    std::pair<char32_t, char32_t>{ 0x300E, 0x300F },
    std::pair<char32_t, char32_t>{ 0x300F, 0x300E },
    std::pair<char32_t, char32_t>{ 0x3010, 0x3011 },
    std::pair<char32_t, char32_t>{ 0x3011, 0x3010 },
    std::pair<char32_t, char32_t>{ 0x3014, 0x3015 },
    std::pair<char32_t, char32_t>{ 0x3015, 0x3014 },
    std::pair<char32_t, char32_t>{ 0x3016, 0x3017 },
    std::pair<char32_t, char32_t>{ 0x3017, 0x3016 },
    std::pair<char32_t, char32_t>{ 0x3018, 0x3019 },
    std::pair<char32_t, char32_t>{ 0x3019, 0x3018 },
    std::pair<char32_t, char32_t>{ 0x301A, 0x301B },
    std::pair<char32_t, char32_t>{ 0x301B, 0x301A },
    std::pair<char32_t, char32_t>{ 0xFE59, 0xFE5A },
    std::pair<char32_t, char32_t>{ 0xFE5A, 0xFE59 },
    std::pair<char32_t, char32_t>{ 0xFE5B, 0xFE5C },
    std::pair<char32_t, char32_t>{ 0xFE5C, 0xFE5B },
    std::pair<char32_t, char32_t>{ 0xFE5D, 0xFE5E },
    std::pair<char32_t, char32_t>{ 0xFE5E, 0xFE5D },
    std::pair<char32_t, char32_t>{ 0xFE64, 0xFE65 },
    std::pair<char32_t, char32_t>{ 0xFE65, 0xFE64 },
    std::pair<char32_t, char32_t>{ 0xFF08, 0xFF09 },
    std::pair<char32_t, char32_t>{ 0xFF09, 0xFF08 },
    std::pair<char32_t, char32_t>{ 0xFF1C, 0xFF1E },
    std::pair<char32_t, char32_t>{ 0xFF1E, 0xFF1C },
    std::pair<char32_t, char32_t>{ 0xFF3B, 0xFF3D },
    std::pair<char32_t, char32_t>{ 0xFF3D, 0xFF3B },
    std::pair<char32_t, char32_t>{ 0xFF5B, 0xFF5D },
    std::pair<char32_t, char32_t>{ 0xFF5D, 0xFF5B },
    std::pair<char32_t, char32_t>{ 0xFF5F, 0xFF60 },
    std::pair<char32_t, char32_t>{ 0xFF60, 0xFF5F },
    std::pair<char32_t, char32_t>{ 0xFF62, 0xFF63 },
    std::pair<char32_t, char32_t>{ 0xFF63, 0xFF62 },
}; // }}}
} // end namespace tables

bool is_mirrored(char32_t codepoint) noexcept {
    return contains(tables::Bidi_Mirrored, codepoint);
}

char32_t bidi_mirroring_glyph(char32_t codepoint) noexcept {
    auto const it = std::lower_bound(
        tables::Bidi_Mirroring_Glyph.begin(),
        tables::Bidi_Mirroring_Glyph.end(),
        codepoint,
        [](auto const& pair, char32_t cp) { return pair.first < cp; });
    if (it != tables::Bidi_Mirroring_Glyph.end() && it->first == codepoint)
        return it->second;
    return codepoint;
}

std::optional<Age> make_age(std::string_view value) noexcept {
    using namespace std::string_view_literals;
    // clang-format off
    static constexpr auto mappings = std::array<std::pair<std::string_view, Age>, 28>{
        std::pair{ "1.1"sv, Age::V1_1 },
        std::pair{ "10.0"sv, Age::V10_0 },
        std::pair{ "11.0"sv, Age::V11_0 },
        std::pair{ "12.0"sv, Age::V12_0 },
        std::pair{ "12.1"sv, Age::V12_1 },
        std::pair{ "13.0"sv, Age::V13_0 },
        std::pair{ "14.0"sv, Age::V14_0 },
        std::pair{ "15.0"sv, Age::V15_0 },
        std::pair{ "15.1"sv, Age::V15_1 },
        std::pair{ "16.0"sv, Age::V16_0 },
        std::pair{ "17.0"sv, Age::V17_0 },
        std::pair{ "2.0"sv, Age::V2_0 },
        std::pair{ "2.1"sv, Age::V2_1 },
        std::pair{ "3.0"sv, Age::V3_0 },
        std::pair{ "3.1"sv, Age::V3_1 },
        std::pair{ "3.2"sv, Age::V3_2 },
        std::pair{ "4.0"sv, Age::V4_0 },
        std::pair{ "4.1"sv, Age::V4_1 },
        std::pair{ "5.0"sv, Age::V5_0 },
        std::pair{ "5.1"sv, Age::V5_1 },
        std::pair{ "5.2"sv, Age::V5_2 },
        std::pair{ "6.0"sv, Age::V6_0 },
        std::pair{ "6.1"sv, Age::V6_1 },
        std::pair{ "6.2"sv, Age::V6_2 },
        std::pair{ "6.3"sv, Age::V6_3 },
        std::pair{ "7.0"sv, Age::V7_0 },
        std::pair{ "8.0"sv, Age::V8_0 },
        std::pair{ "9.0"sv, Age::V9_0 },
    };
    // clang-format on
    for (auto const& [key, val] : mappings)
        if (key == value)
            return val;
    return std::nullopt;
}

} // namespace unicode
